Package: familiar
Title: End-to-End Automated Machine Learning and Model Evaluation
Version: 2.0.2
Authors@R: c(
    person("Alex", "Zwanenburg",
           email = "alexander.zwanenburg@nct-dresden.de",
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-0342-9545")),
    person("Steffen", "Löck", role="aut"),
    person("German Cancer Research Center (DKFZ)", role="cph"),
    person("Technische Universität Dresden", role="cph"))
Description: Single unified interface for end-to-end modelling of regression, 
    categorical and time-to-event (survival) outcomes. Models created using
    familiar are self-containing, and their use does not require additional
    information such as baseline survival, feature clustering, or feature
    transformation and normalisation parameters. Model performance,
    calibration, risk group stratification, (permutation) variable importance,
    individual conditional expectation, partial dependence, and more, are
    assessed automatically as part of the evaluation process and exported in
    tabular format and plotted, and may also be computed manually using export
    and plot functions. Where possible, metrics and values obtained during the
    evaluation process come with confidence intervals.
URL: https://github.com/oncoray/familiar
BugReports: https://github.com/oncoray/familiar/issues
Depends: R (>= 4.0.0)
License: EUPL
Encoding: UTF-8
VignetteBuilder: knitr
Imports: data.table, methods, rlang (>= 1.0.0), rstream, survival
Suggests: BART, callr (>= 3.4.3), cluster, CORElearn, coro,
        dynamicTreeCut, e1071 (>= 1.7.5), fastcluster, fastglm, ggplot2
        (>= 4.0.0), glmnet, gtable, harmonicmeanp, isotree (>= 0.6.0),
        knitr, labeling, laGP, maxstat, microbenchmark, nnet,
        paletteer, power.transform, praznik, proxy, randomForestSRC,
        ranger, rmarkdown, scales, testthat (>= 3.0.0), xml2, xgboost
        (>= 3.0.0)
Collate: 'FamiliarS4Classes.R' 'FamiliarS4Generics.R'
        'BatchNormalisation.R' 'BootstrapConfidenceInterval.R'
        'CheckArguments.R' 'CheckHyperparameters.R' 'CheckPackages.R'
        'ClassBalance.R' 'ClusteringMethod.R' 'Clustering.R'
        'ClusterRepresentation.R' 'Normalisation.R'
        'CombatNormalisation.R' 'LearnerS4Naive.R' 'DataObject.R'
        'DataParameterChecks.R' 'DataPreProcessing.R'
        'DataProcessing.R' 'DataServerBackend.R' 'ErrorMessages.R'
        'ExperimentData.R' 'ExperimentSetup.R' 'Familiar.R'
        'FamiliarCollection.R' 'FamiliarCollectionExport.R'
        'FamiliarData.R' 'FamiliarDataComputation.R'
        'FamiliarDataComputationPredictionData.R' 'PredictionTable.R'
        'FamiliarDataComputationAUCCurves.R'
        'FamiliarDataComputationCalibrationData.R'
        'FamiliarDataComputationCalibrationInfo.R'
        'FamiliarDataComputationConfusionMatrix.R'
        'FamiliarDataComputationDecisionCurveAnalysis.R'
        'FamiliarDataComputationFeatureExpression.R'
        'FamiliarDataComputationFeatureSimilarity.R'
        'FamiliarDataComputationHyperparameters.R'
        'FamiliarDataComputationICE.R'
        'FamiliarDataComputationModelPerformance.R'
        'FamiliarDataComputationPermutationVimp.R'
        'FamiliarDataComputationRiskStratificationData.R'
        'FamiliarDataComputationRiskStratificationInfo.R'
        'FamiliarDataComputationSHAP.R'
        'FamiliarDataComputationSampleSimilarity.R'
        'FamiliarDataComputationUnivariateAnalysis.R'
        'FamiliarDataComputationUtilities.R'
        'FamiliarDataComputationVimp.R' 'FamiliarDataElement.R'
        'FamiliarEnsemble.R' 'FamiliarHyperparameterLearner.R'
        'FamiliarModel.R' 'FamiliarNoveltyDetector.R'
        'FamiliarObjectConversion.R' 'Transformation.R'
        'FamiliarObjectUpdate.R' 'FamiliarSharedS4Methods.R'
        'FamiliarVimpMethod.R' 'FeatureInfo.R'
        'FeatureInfoParameters.R' 'FunctionWrapperUtilities.R'
        'HyperparameterOptimisation.R'
        'HyperparameterOptimisationMetaLearners.R'
        'HyperparameterOptimisationUtilities.R'
        'HyperparameterS4BayesianAdditiveRegressionTrees.R'
        'HyperparameterS4GaussianProcess.R'
        'HyperparameterS4RandomSearch.R' 'HyperparameterS4Ranger.R'
        'Imputation.R' 'Iterations.R' 'LearnerMain.R'
        'LearnerRecalibration.R' 'LearnerRiskStratification.R'
        'LearnerS4Cox.R' 'LearnerS4GLM.R' 'LearnerS4GLMnet.R'
        'LearnerS4KNN.R' 'LearnerS4MBoost.R' 'LearnerS4NaiveBayes.R'
        'LearnerS4RFSRC.R' 'LearnerS4Ranger.R' 'LearnerS4SVM.R'
        'LearnerS4SurvivalRegression.R' 'LearnerS4XGBoost.R'
        'LearnerSurvivalProbability.R' 'Logger.R' 'MetricS4.R'
        'MetricS4AUC.R' 'MetricS4Brier.R' 'MetricS4ConcordanceIndex.R'
        'MetricS4ConfusionMatrixMetrics.R' 'MetricS4Regression.R'
        'NoveltyDetectorMain.R' 'NoveltyDetectorS4IsolationTree.R'
        'NoveltyDetectorS4NoneNoveltyDetector.R' 'OutcomeInfo.R'
        'PairwiseSimilarity.R' 'ParallelFunctions.R' 'ParseData.R'
        'ParseSettings.R' 'PlotAUCcurves.R' 'PlotAll.R'
        'PlotCalibration.R' 'PlotColours.R' 'PlotConfusionMatrix.R'
        'PlotDecisionCurves.R' 'PlotFamiliarPlot.R'
        'PlotFeatureRanking.R' 'PlotFeatureSimilarity.R' 'PlotGTable.R'
        'PlotICE.R' 'PlotInputArguments.R' 'PlotKaplanMeier.R'
        'PlotModelPerformance.R' 'PlotPermutationVariableImportance.R'
        'PlotSampleClustering.R' 'PlotShapDependence.R'
        'PlotShapForce.R' 'PlotShapSummary.R' 'PlotShapWaterfall.R'
        'PlotUnivariateImportance.R' 'PlotUtilities.R'
        'PredictS4Methods.R' 'ProcessTimeUtilities.R' 'Random.R'
        'RankBordaAggregation.R' 'RankMain.R' 'RankSimpleAggregation.R'
        'RankStabilityAggregation.R' 'SocketServer.R'
        'StringUtilities.R' 'TaskEvaluate.R' 'TaskFeatureInfo.R'
        'TaskLearn.R' 'TaskLearnerHyperparameters.R' 'TaskMain.R'
        'TaskNoveltyDetector.R' 'TaskNoveltyDetectorHyperparameters.R'
        'TaskVimp.R' 'TaskVimpHyperparameters.R' 'TestDataCreators.R'
        'TestFeatureInfo.R' 'TestFunctions.R' 'TestTrain.R'
        'TestTrainNovelty.R' 'TestVimp.R' 'TrimUtilities.R'
        'Utilities.R' 'UtilitiesS4.R' 'VimpMain.R'
        'VimpS4Concordance.R' 'VimpS4CoreLearn.R' 'VimpS4Correlation.R'
        'VimpS4MutualInformation.R' 'VimpS4OtherMethods.R'
        'VimpS4Regression.R' 'VimpTable.R' 'aaa.R'
Config/testthat/parallel: true
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-01 15:43:01 UTC; alexz
Author: Alex Zwanenburg [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0342-9545>),
  Steffen Löck [aut],
  German Cancer Research Center (DKFZ) [cph],
  Technische Universität Dresden [cph]
Maintainer: Alex Zwanenburg <alexander.zwanenburg@nct-dresden.de>
Repository: CRAN
Date/Publication: 2026-06-01 17:30:02 UTC
Built: R 4.6.0; ; 2026-06-01 18:07:31 UTC; unix

# Generated by roxygen2: do not edit by hand

export(as_prediction_table)
export(coef)
export(get_xml_config)
export(plot_feature_selection_occurrence)
export(plot_feature_selection_variable_importance)
export(plot_model_signature_occurrence)
export(plot_model_signature_variable_importance)
export(precompute_data_assignment)
export(precompute_feature_info)
export(precompute_vimp)
export(summon_familiar)
export(theme_familiar)
export(train_familiar)
export(vcov)
export(waiver)
exportClasses(experimentData)
exportClasses(familiarCollection)
exportClasses(familiarData)
exportClasses(familiarDataElement)
exportClasses(familiarEnsemble)
exportClasses(familiarHyperparameterLearner)
exportClasses(familiarMetric)
exportClasses(familiarModel)
exportClasses(familiarNoveltyDetector)
exportClasses(familiarVimpMethod)
exportClasses(featureInfo)
exportClasses(featureInfoParameters)
exportClasses(outcomeInfo)
exportClasses(vimpTable)
exportMethods(aggregate_vimp_table)
exportMethods(as_data_object)
exportMethods(as_familiar_collection)
exportMethods(as_familiar_data)
exportMethods(as_familiar_ensemble)
exportMethods(export_all)
exportMethods(export_auc_data)
exportMethods(export_calibration_data)
exportMethods(export_calibration_info)
exportMethods(export_confusion_matrix_data)
exportMethods(export_decision_curve_analysis_data)
exportMethods(export_feature_expressions)
exportMethods(export_feature_similarity)
exportMethods(export_fs_vimp)
exportMethods(export_hyperparameters)
exportMethods(export_ice_data)
exportMethods(export_model_performance)
exportMethods(export_model_vimp)
exportMethods(export_partial_dependence_data)
exportMethods(export_permutation_vimp)
exportMethods(export_prediction_data)
exportMethods(export_risk_stratification_data)
exportMethods(export_risk_stratification_info)
exportMethods(export_sample_similarity)
exportMethods(export_univariate_analysis_data)
exportMethods(get_class_names)
exportMethods(get_data_set_names)
exportMethods(get_feature_names)
exportMethods(get_learner_names)
exportMethods(get_risk_group_names)
exportMethods(get_vimp_method_names)
exportMethods(get_vimp_table)
exportMethods(plot_auc_precision_recall_curve)
exportMethods(plot_auc_roc_curve)
exportMethods(plot_calibration_data)
exportMethods(plot_confusion_matrix)
exportMethods(plot_decision_curve)
exportMethods(plot_feature_similarity)
exportMethods(plot_ice)
exportMethods(plot_kaplan_meier)
exportMethods(plot_model_performance)
exportMethods(plot_pd)
exportMethods(plot_permutation_variable_importance)
exportMethods(plot_sample_clustering)
exportMethods(plot_shap_dependence)
exportMethods(plot_shap_force)
exportMethods(plot_shap_summary)
exportMethods(plot_shap_waterfall)
exportMethods(plot_univariate_importance)
exportMethods(plot_variable_importance)
exportMethods(predict)
exportMethods(set_class_names)
exportMethods(set_data_set_names)
exportMethods(set_feature_names)
exportMethods(set_learner_names)
exportMethods(set_risk_group_names)
exportMethods(set_vimp_method_names)
exportMethods(summary)
exportMethods(update_model_dir_path)
exportMethods(update_object)
import(data.table)
import(methods)
importClassesFrom(rstream,rstream.runif)
importFrom(rlang,"%<~%")
importFrom(rlang,enquo)
importFrom(rlang,enquos)
importFrom(rlang,ensym)
importFrom(rlang,ensyms)
importFrom(rlang,parse_expr)
importFrom(rlang,parse_exprs)
importFrom(rlang,quo)
importFrom(rlang,quos)
importFrom(rlang,sym)
importFrom(rlang,syms)
importFrom(stats,coef)
importFrom(stats,predict)
importFrom(stats,vcov)
importFrom(survival,Surv)
importFrom(survival,coxph)
importFrom(survival,survreg)
importFrom(utils,head)
importFrom(utils,tail)

# Familiar

A list of tools to help create D&D applications.

## Installation

Run the following to install:

```python

pip install familiar
```

## Usage

```python
from familiar import dice_roll

# dice_roll takes a number of dice to roll, number of sides on dice, and a modifier and returns a random result.

roll = dice_roll(2,6, 1)
print(roll)
```

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_decision_curve (generic) ------------------------------------------------

#' @title Plot decision curves.
#'
#' @description This method creates decision curves based on data in a
#'   familiarCollection object.
#'
#' @param dir_path (*optional*) Path to the directory where created decision
#'   curve plots are saved to. Output is saved in the `decision_curve_analysis`
#'   subdirectory. If `NULL`, figures are written to the folder, but are
#'   returned instead.
#'
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_decision_curve_data -object
#'
#' @details This function generates plots for decision curves.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, `data_set` and
#'   `positive_class` (categorical outcomes) or `evaluation_time` (survival
#'   outcomes). By default, the data is split by `vimp_method` and `learner`, with
#'   faceting by `data_set` and colouring by `positive_class` or
#'   `evaluation_time`.
#'
#'   Bootstrap confidence intervals of the decision curve (if present) can be
#'   shown using various styles set by `conf_int_style`:
#'
#'  * `ribbon` (default): confidence intervals are shown as a ribbon with an
#'   opacity of `conf_int_alpha` around the point estimate of the decision
#'   curve.
#'
#'  * `step` (default): confidence intervals are shown as a step function around
#'   the point estimate of the decision curve.
#'
#'  * `none`: confidence intervals are not shown. The point estimate of the
#'   decision curve is shown as usual.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#' @references 1. Vickers, A. J. & Elkin, E. B. Decision curve analysis: a novel
#'   method for evaluating prediction models. Med. Decis. Making 26, 565–574
#'   (2006).
#'
#'   1. Vickers, A. J., Cronin, A. M., Elkin, E. B. & Gonen, M. Extensions to
#'   decision curve analysis, a novel method for evaluating diagnostic tests,
#'   prediction models and molecular markers. BMC Med. Inform. Decis. Mak. 8, 53
#'   (2008).
#'
#'   1. Vickers, A. J., van Calster, B. & Steyerberg, E. W. A simple,
#'   step-by-step guide to interpreting decision curve analysis. Diagn Progn Res
#'   3, 18 (2019).
#' @exportMethod plot_decision_curve
#' @md
#' @rdname plot_decision_curve-methods
setGeneric(
  "plot_decision_curve",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_decision_curve")
  }
)



# plot_decision_curve (general) ------------------------------------------------

#' @rdname plot_decision_curve-methods
setMethod(
  "plot_decision_curve", 
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "decision_curve_analyis"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_decision_curve,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_decision_curve (collection) ---------------------------------------------

#' @rdname plot_decision_curve-methods
setMethod(
  "plot_decision_curve",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    curve_type <- ci_low <- ci_up <- net_benefit <- NULL

    # Make sure the collection object is updated.
    object <- update_object(object = object)

    # Get input data.
    x <- export_decision_curve_analysis_data(
      object = object,
      aggregate_results = TRUE
    )

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check that the data are not evaluated at the model level.
    if (all(sapply(x, function(x) (x@detail_level == "model")))) {
      ..warning_no_comparison_between_models()
      return(NULL)
    }

    # Obtain data element from list.
    if (is.list(x)) {
      if (is_empty(x)) return(NULL)

      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_decision_curve: list of data elements contains unmerged elements."
        )
      }

      # Get x directly.
      x <- x[[1L]]
    }

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to plot decision curves",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Check input arguments ----------------------------------------------------

    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # x_label
    if (is.waive(x_label)) x_label <- "threshold probability"

    # y_label
    if (is.waive(y_label)) y_label <- "net benefit"

    # x_range
    if (is.null(x_range)) x_range <- c(0.0, 1.0)

    # x_breaks
    if (is.null(x_breaks)) {
      .check_input_plot_args(x_n_breaks = x_n_breaks)

      # Create breaks
      x_breaks <- labeling::extended(
        m = x_n_breaks,
        dmin = x_range[1L],
        dmax = x_range[2L],
        only.loose = TRUE
      )
    }

    # conf_int_style
    if (length(conf_int_style) > 1L) {
      conf_int_style <- head(conf_int_style, n = 1L)
    }

    # Set the style of the confidence interval to none, in case no confidence
    # interval data is present.
    if (!x@estimation_type %in% c("bci", "bootstrap_confidence_interval")) {
      conf_int_style <- "none"
    }

    # y_range
    if (is.null(y_range)) {
      if (conf_int_style != "none") {
        # Base the y-range on the confidence intervals.
        y_range <- c(
          min(x@data[curve_type == "model" & is.finite(ci_low)]$ci_low),
          max(x@data[curve_type == "model" & is.finite(ci_up)]$ci_up)
        )
        
      } else {
        # Base the y-range on the range of the benefit.
        y_range <- c(
          min(c(0.0, min(x@data[curve_type == "model" & is.finite(net_benefit)]$net_benefit))),
          max(c(0.0, max(x@data[curve_type == "model" & is.finite(net_benefit)]$net_benefit)))
        )
      }
    }

    # y_breaks
    if (is.null(y_breaks)) {
      .check_input_plot_args(y_n_breaks = y_n_breaks)

      # Create breaks
      y_breaks <- labeling::extended(
        m = y_n_breaks,
        dmin = y_range[1L],
        dmax = y_range[2L],
        only.loose = TRUE
      )

      # Adapt the y-range.
      y_range <- c(
        head(y_breaks, n = 1L),
        tail(y_breaks, n = 1L)
      )
    }

    if (object@outcome_type %in% c("binomial", "multinomial")) {
      split_variable <- "positive_class"
    } else if (object@outcome_type %in% c("survival")) {
      split_variable <- "evaluation_time"
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }

    # Splitting variables
    if (is.null(split_by) && is.null(facet_by) && is.null(color_by)) {
      # Determine the number of learners and feature_selection methods.
      n_learner <- nlevels(x@data$learner)
      n_vimp_method <- nlevels(x@data$vimp_method)

      if (object@outcome_type %in% c("multinomial")) {
        n_class_or_time <- nlevels(x@data$positive_class)
      } else if (object@outcome_type %in% c("binomial")) {
        n_class_or_time <- 1L
      } else if (object@outcome_type %in% c("survival")) {
        n_class_or_time <- nlevels(x@data$evaluation_time)
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      if (n_learner > 1L && n_vimp_method > 1L) {
        # Split by learner and feature selection method.
        split_by <- c("vimp_method", "learner")

        if (n_class_or_time > 1L) {
          color_by <- split_variable
          facet_by <- "data_set"
        } else {
          color_by <- c("data_set", split_variable)
        }
        
      } else if (n_learner > 1L) {
        # Implying n_vimp_method == 1

        if (n_class_or_time > 1L) {
          split_by <- c("vimp_method", "learner")
          color_by <- split_variable
          facet_by <- "data_set"
        } else {
          split_by <- c("vimp_method")
          color_by <- c("learner")
          facet_by <- c("data_set", split_variable)
        }
        
      } else if (n_vimp_method > 1L) {
        # Implying n_learner == 1

        if (n_class_or_time > 1L) {
          split_by <- c("vimp_method", "learner")
          color_by <- split_variable
          facet_by <- "data_set"
        } else {
          split_by <- "learner"
          color_by <- "vimp_method"
          facet_by <- c("data_set", split_variable)
        }
        
      } else {
        # Implying n_learner == n_vimp_method == 1
        split_by <- c("vimp_method", "learner")

        if (n_class_or_time > 1L) {
          color_by <- split_variable
          facet_by <- "data_set"
        } else {
          color_by <- c("data_set", split_variable)
        }
      }
    }

    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      color_by = color_by,
      facet_by = facet_by,
      available = c("vimp_method", "learner", "data_set", split_variable)
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    color_by <- split_var_list$color_by
    facet_by <- split_var_list$facet_by

    # Create a legend label
    legend_label <- .create_plot_legend_title(
      user_label = legend_label,
      color_by = color_by
    )

    # Check input arguments for validity.
    .check_input_plot_args(
      x_range = x_range,
      y_range = y_range,
      x_breaks = x_breaks,
      y_breaks = y_breaks,
      conf_int_alpha = conf_int_alpha,
      conf_int_style = conf_int_style,
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption
    )

    # Create plots -------------------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data
    if (!is.null(split_by)) {
      x_split <- split(x@data, by = split_by)
    } else {
      x_split <- list("null.name" = x@data)
    }

    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over splits
    for (ii in names(x_split)) {
      # Skip empty datasets
      if (is_empty(x_split[[ii]])) next

      if (is.waive(plot_title)) plot_title <- "Decision curve"

      # Declare subtitle components.
      additional_subtitle <- NULL

      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, color_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(x_split[[ii]]$evaluation_time[1L])
        )
      }

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = x_split[[ii]]
        )
      }

      # Generate plot
      p <- .plot_decision_curve_plot(
        x = x_split[[ii]],
        color_by = color_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        x_range = x_range,
        x_breaks = x_breaks,
        y_range = y_range,
        y_breaks = y_breaks,
        conf_int_style = conf_int_style,
        conf_int_alpha = conf_int_alpha
      )

      # Check empty output
      if (is.null(p)) next

      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_decision_curve_plot_dimensions(
          x = x_split[[ii]],
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "decision_curve_analysis",
              "subtype" = "decision_curve",
              "x" = x_split[[ii]],
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_decision_curve_plot <- function(
    x,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    y_range,
    y_breaks,
    conf_int_style,
    conf_int_alpha
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  curve_type <- NULL

  # Generate a guide table.
  guide_list <- .create_plot_guide_table(
    x = x, color_by = color_by,
    discrete_palette = discrete_palette
  )

  # Extract data
  x <- guide_list$data

  # Create basic plot
  p <- ggplot2::ggplot(
    data = x[curve_type == "model"],
    mapping = ggplot2::aes(
      x = !!sym("threshold_probability"),
      y = !!sym("net_benefit")
    )
  )

  # Add theme
  p <- p + ggtheme

  # Add decision curve, the curve for intervention for all, and the curve for no
  # intervention.
  if (is.null(color_by)) {
    # Model-based curve without colour-based splitting.
    p <- p + ggplot2::geom_line()

    # Intervention for all.
    p <- p + ggplot2::geom_line(
      data = x[curve_type == "intervention_all"],
      mapping = ggplot2::aes(
        x = !!sym("threshold_probability"),
        y = !!sym("net_benefit")
      )
    )

    # Intervention for none.
    p <- p + ggplot2::geom_hline(yintercept = 0.0)
    
  } else {
    # Model-based curve with colour-based splitting.
    p <- p + ggplot2::geom_line(mapping = ggplot2::aes(colour = !!sym("color_breaks")))

    # Intervention for all.
    # Intervention for all.
    p <- p + ggplot2::geom_line(
      data = x[curve_type == "intervention_all"],
      mapping = ggplot2::aes(
        x = !!sym("threshold_probability"),
        y = !!sym("net_benefit"),
        colour = !!sym("color_breaks")
      )
    )

    # Intervention for none.
    p <- p + ggplot2::geom_hline(yintercept = 0.0)

    # Extract guidetable for color
    g_color <- guide_list$guide_color

    # Set colour and fill (fill may be unused)
    p <- p + ggplot2::scale_colour_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )

    p <- p + ggplot2::scale_fill_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )
  }

  # Plot confidence intervals
  if (conf_int_style[1L] != "none") {
    if (conf_int_style[1L] == "step") {
      if (is.null(color_by)) {
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(y = !!sym("ci_low")),
          linetype = "dashed"
        )

        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(y = !!sym("ci_up")),
          linetype = "dashed"
        )
        
      } else {
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_low"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed"
        )

        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_up"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed"
        )
      }

      # Remove linetype from the legend.
      p <- p + ggplot2::scale_linetype(guide = "none")
      
    } else if (conf_int_style[1L] == "ribbon") {
      if (is.null(color_by)) {
        p <- p + ggplot2::geom_ribbon(
          mapping = ggplot2::aes(
            ymin = !!sym("ci_low"),
            ymax = !!sym("ci_up")
          ),
          alpha = conf_int_alpha
        )
        
      } else {
        p <- p + ggplot2::geom_ribbon(
          mapping = ggplot2::aes(
            ymin = !!sym("ci_low"),
            ymax = !!sym("ci_up"),
            fill = !!sym("color_breaks")
          ),
          alpha = conf_int_alpha
        )
      }
    }
  }

  # Update x and y scales
  p <- p + ggplot2::scale_x_continuous(breaks = x_breaks)
  p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)

  # Labels
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context",
        drop = TRUE
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context",
        drop = TRUE
      )
    }
  }

  # Prevent clipping of confidence intervals.
  p <- p + ggplot2::coord_cartesian(xlim = x_range, ylim = y_range)

  return(p)
}



.determine_decision_curve_plot_dimensions <- function(
    x,
    facet_by,
    facet_wrap_cols
) {
  # Obtain faceting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )

  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- 4.0

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))

  return(c(height, width))
}

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "survival",
  features = c("feature_4"),
  sample_limit = 20L,
  n_sample_points = 10L,
  debug = debug_flag
)

# 1D plot without novelty, and with anchored values.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "survival",
  features = c("feature_1"),
  plot_args = list(
    "anchor_values" = list("feature_1" = 0.4),
    "value_scales" = "figure"),
  sample_limit = 20L,
  n_sample_points = 10L,
  debug = debug_flag
)

# 1D plot with novelty, and with anchored values.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "survival",
  features = c("feature_1"),
  plot_args = list(
    "anchor_values" = list("feature_1" = 0.4),
    "value_scales" = "figure"),
  sample_limit = 20L,
  n_sample_points = 10L,
  create_novelty_detector = TRUE,
  debug = debug_flag
)

# 1D plot with novelty, but no ICE
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "survival",
  features = c("feature_1"),
  plot_args = list("show_ice" = FALSE),
  sample_limit = 20L,
  n_sample_points = 10L,
  create_novelty_detector = TRUE,
  debug = debug_flag
)

# 1D plot with novelty, but no PD
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "survival",
  features = c("feature_1"),
  plot_args = list("show_pd" = FALSE),
  sample_limit = 20L,
  n_sample_points = 10L,
  create_novelty_detector = TRUE,
  debug = debug_flag
)


# 1D plot with novelty and colouring, and with anchored values.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "multinomial",
  features = c("feature_1"),
  plot_args = list(
    "anchor_values" = list("feature_1" = 0.4),
    "value_scales" = "figure",
    "color_by" = "positive_class",
    "facet_by" = c("data_set", "vimp_method", "learner")),
  sample_limit = 20L,
  n_sample_points = 10L,
  create_novelty_detector = TRUE,
  debug = debug_flag
)


# 1D plot with novelty and colouring, with anchored values and scaling per
# facet.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "multinomial",
  features = c("feature_1"),
  plot_args = list(
    "anchor_values" = list("feature_1" = 0.40),
    "value_scales" = "facet",
    "color_by" = "positive_class",
    "facet_by" = c("data_set", "vimp_method", "learner")),
  sample_limit = 20L,
  n_sample_points = 10L,
  create_novelty_detector = TRUE,
  debug = debug_flag
)


# 1D plot with novelty and colouring, and with anchored values.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "multinomial",
  features = c("feature_1"),
  plot_args = list(
    "anchor_values" = list("feature_1" = 0.40),
    "value_scales" = "figure",
    "split_by" = c("vimp_method", "learner"),
    "color_by" = "positive_class",
    "facet_by" = "data_set"),
  sample_limit = 20L,
  n_sample_points = 10L,
  create_novelty_detector = TRUE,
  debug = debug_flag
)

# 2D plot without novelty, and with anchored values.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "survival",
  features = c("feature_1", "feature_4"),
  plot_args = list(
    "anchor_values" = list(
      "feature_1" = 0.4,
      "feature_4" = "better"),
    "value_scales" = "figure"),
  sample_limit = 20L,
  n_sample_points = 10L,
  debug = debug_flag
)

# 2D plot with novelty, and with anchored values.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "survival",
  features = c("feature_1", "feature_4"),
  plot_args = list(
    "anchor_values" = list(
      "feature_1" = 0.4,
      "feature_4" = "better"),
    "value_scales" = "figure"),
  sample_limit = 20L,
  n_sample_points = 10L,
  create_novelty_detector = TRUE,
  debug = debug_flag
)

# 2D plot with novelty, and with anchored values.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_ice,
  data_element = "ice_data",
  outcome_type_available = "multinomial",
  features = c("feature_1", "feature_2a"),
  plot_args = list(
    "anchor_values" = list(
      "feature_1" = 0.4,
      "feature_2a" = 0.4),
    "value_scales" = "figure"),
  sample_limit = 20L,
  n_sample_points = 10L,
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarDataElementUnivariateAnalysis object ---------------------------------
setClass(
  "familiarDataElementUnivariateAnalysis",
  contains = "familiarDataElement",
  prototype = methods::prototype(
    detail_level = "ensemble",
    estimation_type = "point"
  )
)

# familiarDataElementRobustness object -----------------------------------------
setClass(
  "familiarDataElementRobustness",
  contains = "familiarDataElement",
  slots = list("icc_type" = "character"),
  prototype = methods::prototype(
    detail_level = "ensemble",
    estimation_type = "point",
    icc_type = "1"
  )
)



# extract_univariate_analysis (generic) ----------------------------------------

#'@title Internal function to extract data from a univariate analysis.
#'
#'@description Computes and extracts univariate analysis for the features used
#'  in a `familiarEnsemble` object. This assessment includes the computation of
#'  p-values, as well as robustness (in case of repeated measurements).
#'
#'@inheritParams .extract_data
#'
#'@return A list with a data.table containing information concerning the
#'  univariate analysis of important features.
#'@md
#'@keywords internal
setGeneric(
  "extract_univariate_analysis",
  function(
    object,
    data,
    cl = NULL,
    icc_type = waiver(),
    feature_similarity = NULL,
    feature_cluster_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_linkage_method = waiver(),
    feature_similarity_threshold = waiver(),
    feature_similarity_metric = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_univariate_analysis")
  }
)


# extract_univariate_analysis (familiarEnsemble) -------------------------------
setMethod(
  "extract_univariate_analysis",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    icc_type = waiver(),
    feature_similarity = NULL,
    feature_cluster_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_linkage_method = waiver(),
    feature_similarity_threshold = waiver(),
    feature_similarity_metric = waiver(),
    message_indent = 0L,
    verbose = FALSE
  ) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    p_value <- NULL
    
    # Message extraction start
    logger_message(
      paste0("Extracting univariate analysis information."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Obtain from settings, if unset.
    if (is.waive(icc_type)) icc_type <- object@settings$icc_type
    
    # Check icc_type
    .check_parameter_value_is_valid(
      x = icc_type, 
      var_name = "icc_type",
      values = .get_available_icc_types()
    )
    
    # Get and process the input data
    data <- process_input_data(
      object = object,
      data = data, 
      stop_at = "normalisation"
    )
    
    # Check if the data object is empty.
    if (is_empty(data)) return(NULL)
    
    # Maintain only important features. The current set is based on the required
    # features.
    data <- filter_features(
      data = data,
      available_features = object@model_features
    )
    
    # Determine feature columns
    feature_columns <- get_feature_columns(x = data)
    
    # Check if there are any features in the model.
    if (length(feature_columns) == 0L) return(NULL)
    
    ## Univariate p-values -----------------------------------------------------
    # Remove data with missing outcomes.
    feature_data <- filter_missing_outcome(data)
    
    if (is_empty(feature_data)) {
      # Check that data are not empty
      univariate_data <- NULL
      
    } else if (get_n_samples(feature_data) <= 5L) {
      # Check if the number of samples is sufficient (>5).
      univariate_data <- NULL
      
    } else {
      # Calculate univariate P values, based on aggregated data
      regression_p_values <- compute_univariable_p_values(
        cl = cl,
        data_obj = aggregate_data(data = feature_data),
        feature_columns = feature_columns
      )
      
      # Find and replace non-finite values
      regression_p_values[!is.finite(regression_p_values)] <- NA_real_
      
      # Collect to table
      univariate_data <- data.table::data.table(
        "feature" = names(regression_p_values),
        "p_value" = regression_p_values
      )[order(p_value)]
      
      # Set univariate data.
      univariate_data <- methods::new(
        "familiarDataElementUnivariateAnalysis",
        data = univariate_data,
        value_column = "p_value",
        grouping_column = "feature"
      )
      
      # Add model name.
      univariate_data <- add_model_name(univariate_data, object)
    }
    
    ## Feature robustness ------------------------------------------------------
    
    if (!all(data@data$repetition_id == 1L)) {
      
      # Determine which columns actually contains numeric data
      numeric_columns <- feature_columns[sapply(
        feature_columns,
        function(ii, data) (is.numeric(data@data[[ii]])),
        data = data
      )]
      
      if (length(numeric_columns) == 0L) {
        icc_data <- NULL
        
      } else {
        # Compute ICC values
        icc_data <- fam_mapply(
          cl = cl,
          assign = NULL,
          FUN = compute_icc,
          x = data@data[, mget(numeric_columns)],
          feature = numeric_columns,
          progress_bar = FALSE,
          MoreArgs = list(
            "id_data" = data@data[, mget(get_id_columns(id_depth = "repetition"))],
            "type" = icc_type
          )
        )
        
        # Compute values
        icc_data <- data.table::rbindlist(icc_data, use.names = TRUE)
        
        # Create object.
        icc_data <- methods::new(
          "familiarDataElementRobustness",
          data = icc_data,
          value_column = c("icc", "icc_low", "icc_up", "icc_panel", "icc_panel_low", "icc_panel_up"),
          grouping_column = "feature"
        )
        
        # Add model name.
        icc_data <- add_model_name(icc_data, object)
      }
      
    } else {
      icc_data <- NULL
    }
    
    return(list(univariate_data, icc_data))
  }
)



# extract_univariate_analysis (prediction table) -------------------------------
setMethod(
  "extract_univariate_analysis",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("univariate feature importance")
    
    return(NULL)
  }
)



# export_univariate_analysis_data (generic) ------------------------------------

#'@title Extract and export univariate analysis data of features.
#'
#'@description Extract and export univariate analysis data of features for data
#'  in a familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams plot_univariate_importance
#'
#'@inheritDotParams extract_univariate_analysis
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Univariate analysis includes the computation of p-values, as well as
#'  robustness (in case of repeated measurements). p-values are derived from
#'  Wald's test.
#'
#'@return A data.table (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_univariate_analysis_data
#'@md
#'@rdname export_univariate_analysis_data-methods
setGeneric(
  "export_univariate_analysis_data",
  function(
    object,
    dir_path = NULL,
    p_adjustment_method = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_univariate_analysis_data")
  }
)



# export_univariate_analysis_data (collection) ---------------------------------

#'@rdname export_univariate_analysis_data-methods
setMethod(
  "export_univariate_analysis_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    p_adjustment_method = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Set default adjust method.
    if (is.waive(p_adjustment_method)) p_adjustment_method <- "fdr"
    if (is.null(p_adjustment_method)) p_adjustment_method <- "none"
    
    # Obtain data from the univariate analysis.
    univariate_data <- .export(
      x = object,
      data_slot = "univariate_analysis",
      dir_path = dir_path,
      aggregate_results = FALSE,
      type = "variable_importance",
      subtype = "univariate",
      object_class = "familiarDataElementUnivariateAnalysis"
    )
    
    # Compute the adjusted p-values.
    univariate_data <- lapply(
      univariate_data,
      .compute_adjusted_univariate_p_value,
      method = p_adjustment_method
    )
    
    # Obtain robustness data.
    icc_data <- .export(
      x = object,
      data_slot = "univariate_analysis",
      dir_path = dir_path,
      aggregate_results = FALSE,
      type = "variable_importance",
      subtype = "robustness",
      object_class = "familiarDataElementRobustness"
    )
    
    # Set data list.
    data_list <- list(
      "univariate" = univariate_data,
      "icc" = icc_data
    )
    
    if (!is.null(dir_path)) data_list <- NULL
    if (export_collection) data_list <- c(data_list, list("collection" = object))
    
    return(data_list)
  }
)



# export_univariate_analysis_data (general) ------------------------------------

#'@rdname export_univariate_analysis_data-methods
setMethod(
  "export_univariate_analysis_data",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    p_adjustment_method = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "univariate_analysis"
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_univariate_analysis_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "p_adjustment_method" = p_adjustment_method,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)



.compute_adjusted_univariate_p_value <- function(x, method = "fdr") {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  p_value <- NULL
  
  # If the method is not a correction method, return the data as is.
  if (!method %in% stats::p.adjust.methods) return(x)
  
  # Check that the data are not empty.
  if (is_empty(x)) return(x)
  
  # Determine grouping columns.
  grouping_column <- setdiff(x@grouping_column, "feature")
  
  # Make a local copy of the data.
  data <- data.table::copy(x@data)
  
  if (length(grouping_column) == 0L) {
    data[, "adjusted_p_value" := stats::p.adjust(p = p_value, method = method)]
  } else {
    data[
      ,
      "adjusted_p_value" := stats::p.adjust(p = p_value, method = method),
      by = c(grouping_column)
    ]
  }
  
  # Add to data element.
  x@data <- data
  
  return(x)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R


create_outcome_info <- function(settings) {
  outcome_info <- methods::new(
    "outcomeInfo",
    name = settings$data$outcome_name,
    outcome_type = settings$data$outcome_type,
    outcome_column = settings$data$outcome_col
  )

  if (outcome_info@outcome_type %in% c("binomial", "multinomial")) {
    # Set class levels
    outcome_info@levels <- settings$data$class_levels

    # Set flag indicating that the outcome is ordinal (currently not enabled)
    outcome_info@ordered <- FALSE

    # Set reference level of the outcome
    outcome_info@reference <- settings$data$class_levels[1L]
  }

  if (outcome_info@outcome_type %in% c("survival", "competing_risk")) {
    # Set maximum time
    outcome_info@time <- ifelse(is.numeric(settings$eval$time_max), settings$eval$time_max, Inf)

    # Set indicator for censoring
    outcome_info@censored <- as.character(settings$data$censoring_indicator)

    # Set indicator for events
    outcome_info@event <- as.character(settings$data$event_indicator)
  }

  if (outcome_info@outcome_type %in% c("competing_risk")) {
    # Set indicator for competing risks
    outcome_info@competing_risk <- as.character(settings$data$competing_risk_indicator)
  }

  # Add familiar version.
  outcome_info <- add_package_version(outcome_info)
  
  return(outcome_info)
}


# .create_outcome_info (model, ensemble) ---------------------------------------
setMethod(
  ".create_outcome_info",
  signature(object = "familiarModelUnion"),
  function(object, ...) {
    return(object@outcome_info)
  }
)


# .create_outcome_info (data object) -------------------------------------------
setMethod(
  ".create_outcome_info",
  signature(object = "dataObject"),
  function(object, ...) {
    
    outcome_info <- methods::new(
      "outcomeInfo",
      name = character(0L),
      outcome_type = object@outcome_type,
      outcome_column = get_outcome_columns(x = object)
    )
    
    if (outcome_info@outcome_type %in% c("binomial", "multinomial")) {
      # Set class levels
      outcome_info@levels <- get_outcome_class_levels(x = object)
      
      # Set flag indicating that the outcome is ordinal (currently not enabled)
      outcome_info@ordered <- is.ordered(object@data[[outcome_info@outcome_column]])
      
      # Set reference level of the outcome
      outcome_info@reference <- outcome_info@levels[1L]
    }
    
    if (outcome_info@outcome_type %in% c("survival", "competing_risk")) {
      # Set maximum time
      outcome_info@time <- Inf
      
      # Set indicator for censoring
      outcome_info@censored <- "0"
      
      # Set indicator for events
      outcome_info@event <- "1"
    }
    
    if (outcome_info@outcome_type %in% c("competing_risk")) {
      # Set indicator for competing risks
      outcome_info@competing_risk <- as.character(setdiff(
        unique_na(object@data[[outcome_info@outcome_column[2L]]]), c(0L, 1L)
      ))
    }
    
    # Add familiar version.
    outcome_info <- add_package_version(outcome_info)
    
    return(outcome_info)
  }
)



# .create_outcome_info (data.table) --------------------------------------------
setMethod(
  ".create_outcome_info",
  signature(object = "data.table"),
  function(object, outcome_type, ...) {
    outcome_info <- methods::new(
      "outcomeInfo",
      name = character(0L),
      outcome_type = outcome_type,
      outcome_column = get_outcome_columns(outcome_type = outcome_type)
    )
    
    if (outcome_info@outcome_type %in% c("binomial", "multinomial")) {
      # Set class levels
      outcome_info@levels <- get_outcome_class_levels(x = object, outcome_type = outcome_info@outcome_type)
      
      # Set flag indicating that the outcome is ordinal (currently not enabled)
      outcome_info@ordered <- is.ordered(object[[outcome_info@outcome_column]])
      
      # Set reference level of the outcome
      outcome_info@reference <- outcome_info@levels[1L]
    }
    
    if (outcome_info@outcome_type %in% c("survival", "competing_risk")) {
      # Set maximum time
      outcome_info@time <- Inf
      
      # Set indicator for censoring
      outcome_info@censored <- "0"
      
      # Set indicator for events
      outcome_info@event <- "1"
    }
    
    if (outcome_info@outcome_type %in% c("competing_risk")) {
      # Set indicator for competing risks
      outcome_info@competing_risk <- as.character(setdiff(
        unique_na(object[[outcome_info@outcome_column[2L]]]), c(0L, 1L)
      ))
    }
    
    # Add familiar version.
    outcome_info <- add_package_version(outcome_info)
    
    return(outcome_info)
  }
)


# .create_outcome_info (prediction table) --------------------------------------
setMethod(
  ".create_outcome_info",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    outcome_info <- methods::new(
      "outcomeInfo",
      name = character(0L),
      outcome_type = object@outcome_type,
      outcome_column = get_outcome_columns(object@outcome_type)
    )
    
    if (.hasSlot(object, "classes")) {
      outcome_info@levels <- object@classes
      outcome_info@ordered <- is.ordered(object@classes)
      outcome_info@reference <- outcome_info@levels[1L]
    }
    
    if (.hasSlot(object, "time")) {
      outcome_info@time <- object@time
    }
    
    if (.hasSlot(object, "event")) {
      outcome_info@event <- object@event
    }
    
    if (.hasSlot(object, "censored")) {
      outcome_info@censored <- object@censored
    }
    
    # Add familiar version.
    outcome_info <- add_package_version(outcome_info)
    
    return(outcome_info)
  }
)



.assign_outcome_info_to_global <- function(outcome_info) {
  # Assign outcome_info in the familiar global environment
  assign("outcome_info", outcome_info, envir = familiar_global_env)
}



get_outcome_info_from_backend <- function() {
  # Retrieve the paths to files and directories
  if (exists("familiar_global_env")) {
    if (exists("outcome_info", where = familiar_global_env)) {
      data_env <- familiar_global_env
    } else if (exists("outcome_info", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
    } else {
      ..error("The outcomeInfo object was not found in backend environment.")
    }
  } else if (exists("outcome_info", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
  } else {
    ..error("The outcomeInfo object was not found in backend environment.")
  }

  return(get("outcome_info", envir = data_env))
}



.get_outcome_info <- function(x = NULL) {
  # Function to retrieve outcome_info in a generic manner.

  # Placeholder outcome_info
  outcome_info <- NULL

  # First, attempt to obtain from familiarModel and similar objects.
  if (inherits(x, c("familiarModel", "familiarEnsemble", "familiarData", "familiarCollection"))) {
    if (!is.null(x@outcome_info)) {
      return(x@outcome_info)
    }
  }

  # Second attempt to get from backend
  outcome_info <- tryCatch(
    get_outcome_info_from_backend(),
    error = identity
  )

  if (!inherits(outcome_info, "error")) {
    return(outcome_info)
  }

  # Third, attempt to infer from settings.
  settings <- tryCatch(
    get_settings(),
    error = identity
  )

  if (!inherits(settings, "error")) {
    return(create_outcome_info(settings = settings))
  }

  # Finally, attempt to infer from dataObject
  if (inherits(x, "dataObject")) {
    return(.create_outcome_info(x))
  }

  if (is.null(outcome_info)) {
    ..error("The requested outcomeInfo object could not be read or created on the fly.")
  }
}



.aggregate_outcome_info <- function(x) {
  # Suppress NOTES due to non-standard evaluation in data.table
  min <- Q1 <- median <- Q3 <- max <- count <- incidence <- NULL
  survival_probability <- survival_probability <- NULL

  # Remove empty elements from x.
  x <- x[!sapply(x, is.null)]
  if (length(x) == 0L) return(NULL)
  
  # Copy the first outcome info object
  outcome_info <- x[[1L]]

  # Find distribution items
  distribution_items <- names(outcome_info@distribution)

  if (!is.null(distribution_items)) {
    # Placeholder list
    distr_list <- list()

    # Iterate over items in the distribution list
    for (item in distribution_items) {
      if (grepl(pattern = "fivenum", x = item, fixed = TRUE)) {
        # Aggregate from list
        fivenum_values <- lapply(
          x, 
          function(outcome_info_object, item) {
            outcome_info_object@distribution[[item]]
          },
          item = item
        )
        
        # Combine all the data.tables
        fivenum_values <- data.table::rbindlist(fivenum_values)

        # Check for zero-length lists.
        if (is_empty(fivenum_values)) next()

        # Summarise
        fivenum_values <- fivenum_values[, list(
          "min" = min(min),
          "Q1" = mean(Q1),
          "median" = mean(median),
          "Q3" = mean(Q3),
          "max" = max(max)
        ), ]

        # Add to list
        distr_list[[item]] <- fivenum_values
      } else if (grepl(pattern = "frequency", x = item, fixed = TRUE)) {
        # Aggregate from list
        frequency_values <- lapply(
          x,
          function(outcome_info_object, item) {
            outcome_info_object@distribution[[item]]
          },
          item = item
        )
        
        # Combine all the data.tables
        frequency_values <- data.table::rbindlist(frequency_values)

        if (is_empty(frequency_values)) next

        # Summarise and add to list
        distr_list[[item]] <- frequency_values[, list("count" = mean(count)), by = "outcome"]
        
      } else if (grepl(pattern = "incidence", x = item, fixed = TRUE)) {
        # Identify unique times.
        unique_times <- sort(unique(unlist(lapply(
          x,
          function(outcome_info_object, item) {
            outcome_info_object@distribution[[item]]$time
          },
          item = item
        ))))
        
        # Interpolate at the unique times.
        incidence_table <- lapply(
          x,
          function(outcome_info_object, item, unique_times) {
            # Check that the object is not empty.
            if (is_empty(outcome_info_object@distribution[[item]])) {
              return(NULL)
            }
            
            # Check that there is at least unique times.
            if (data.table::uniqueN(outcome_info_object@distribution[[item]]$time) < 2L) {
              return(NULL)
            }
            
            # Interpolate the data at the unique time points.
            incidence <- stats::approx(
              x = outcome_info_object@distribution[[item]]$time,
              y = outcome_info_object@distribution[[item]][[item]],
              xout = unique_times,
              method = "linear",
              rule = 2L,
              yleft = 0.0
            )$y
            
            return(data.table::data.table(
              "time" = unique_times,
              "incidence" = incidence
            ))
          },
          item = item,
          unique_times = unique_times
        )
        
        # Aggregate the incidence table.
        incidence_table <- data.table::rbindlist(incidence_table)
        
        # Compute the mean incidence by each time point.
        incidence_table <- incidence_table[, list("incidence" = mean(incidence)), by = "time"]
        
        # Update the column name.
        data.table::setnames(
          x = incidence_table, 
          old = "incidence",
          new = item
        )

        # Add to list
        distr_list[[item]] <- incidence_table
        
      } else if (grepl(pattern = "survival_probability", x = item, fixed = TRUE)) {
        # Identify unique times.
        unique_times <- sort(unique(unlist(lapply(
          x,
          function(outcome_info_object, item) {
            outcome_info_object@distribution[[item]]$time
          },
          item = item
        ))))
        
        # Interpolate at the unique times.
        survival_table <- lapply(
          x,
          function(outcome_info_object, item, unique_times) {
            # Check that the object is not empty.
            if (is_empty(outcome_info_object@distribution[[item]])) {
              return(NULL)
            }
            
            # Check that there is at least unique times.
            if (data.table::uniqueN(outcome_info_object@distribution[[item]]$time) < 2L) {
              return(NULL)
            }
            
            # Interpolate the data at the unique time points.
            survival_probability <- stats::approx(
              x = outcome_info_object@distribution[[item]]$time,
              y = outcome_info_object@distribution[[item]][[item]],
              xout = unique_times,
              method = "linear",
              rule = 2L,
              yleft = 1.0
            )$y
            
            return(data.table::data.table(
              "time" = unique_times,
              "survival_probability" = survival_probability
            ))
          },
          item = item,
          unique_times = unique_times
        )
        
        # Aggregate the survival probability table.
        survival_table <- data.table::rbindlist(survival_table)
        
        # Compute the mean survival prob by each time point.
        survival_table <- survival_table[
          ,
          list("survival_probability" = mean(survival_probability)),
          by = "time"
        ]
        
        # Update the column name.
        data.table::setnames(
          x = survival_table,
          old = "survival_probability",
          new = item
        )

        # Add to list
        distr_list[[item]] <- survival_table
        
      } else {
        # Find mean value
        distr_list[[item]] <- mean(unlist(extract_from_slot(
          x, "distribution", item, na.rm = TRUE
        )))
      }
    }

    # Update distribution slot
    outcome_info@distribution <- distr_list
  }

  # Update transformation parameters
  if (!is.null(outcome_info@transformation_parameters)) {
    transform_method <- get_mode(extract_from_slot(
      object_list = x,
      slot_name = "transformation_parameters",
      slot_element = "transform_method"
    ))
    transform_lambda <- get_mode(extract_from_slot(
      object_list = x,
      slot_name = "transformation_parameters",
      slot_element = "transform_lambda"
    ))

    outcome_info@transformation_parameters <- list(
      "transform_method" = transform_method,
      "transform_lambda" = transform_lambda
    )
  }

  # Update normalisation parameters
  if (!is.null(outcome_info@normalisation_parameters)) {
    normalisation_method <- get_mode(extract_from_slot(
      object_list = x,
      slot_name = "normalisation_parameters",
      slot_element = "norm_method"
    ))
    normalisation_shift <- mean(extract_from_slot(
      object_list = x, 
      slot_name = "normalisation_parameters",
      slot_element = "norm_shift"
    ))
    normalisation_scale <- mean(extract_from_slot(
      object_list = x,
      slot_name = "normalisation_parameters",
      slot_element = "norm_scale"
    ))

    outcome_info@normalisation_parameters <- list(
      "norm_method" = normalisation_method,
      "norm_shift" = normalisation_shift,
      "norm_scale" = normalisation_scale
    )
  }

  return(outcome_info)
}



.compute_outcome_distribution_data <- function(object, data) {
  # Suppress NOTES due to non-standard evaluation in data.table
  event <- censored <- competing <- NULL
  interval_survival <- interval_incidence_rate <- interval_incidence_rate <- NULL
  outcome_event <- outcome_time <- n <- interval_non_censoring_rate <- survival_probability <- NULL
  
  # Check for empty datasets, and return without setting distribution info.
  if (is_empty(data, allow_no_features = TRUE)) return(object)

  # Placeholder distribution list
  distr_list <- list()

  # Find outcome data
  x <- data.table::copy(unique(
    data@data,
    by = get_id_columns(id_depth = "series")
  ))

  if (object@outcome_type %in% c("binomial", "multinomial")) {
    # Number of samples
    distr_list[["n"]] <- nrow(x)

    # Number of instances for each class
    distr_list[["frequency"]] <- x[, list("count" = .N), by = "outcome"]
    
  } else if (object@outcome_type %in% c("continuous")) {
    # Number of samples
    distr_list[["n"]] <- nrow(x)

    # Five-number summary of outcome values
    distr_list[["fivenum"]] <- fivenum_summary(x$outcome, na.rm = TRUE)

    # Mean value
    distr_list[["mean"]] <- mean(x$outcome, na.rm = TRUE)

    distr_list[["median"]] <- stats::median(x$outcome, na.rm = TRUE)
    
  } else if (object@outcome_type %in% c("survival", "competing_risk")) {
    # Baseline survival
    surv_group <- data.table::copy(x)

    # Count events and censoring at each time point.
    surv_group <- surv_group[
      ,
      list(
        "event" = sum(outcome_event == 1L),
        "censored" = sum(outcome_event == 0L),
        "competing" = sum(outcome_event > 1L)
      ),
      by = "outcome_time"
    ][order(outcome_time)]

    # Add group sizes at the start of each interval.
    surv_group[, "n" := nrow(x) - data.table::shift(
      cumsum(event + censored + competing),
      n = 1L,
      fill = 0L,
      type = "lag"
    )]

    # Compute the incidence and censoring rates in the interval
    surv_group[, ":="(
      "interval_survival" = 1.0 - (event + competing) / n,
      "interval_incidence_rate" = event / n,
      "interval_non_censoring_rate" = 1.0 - censored / n
    )]

    # Compute the Kaplan-Meier survival estimator and censoring estimator.
    surv_group[, ":="(
      "survival_probability" = cumprod(interval_survival),
      "cumulative_censoring" = 1.0 - cumprod(interval_non_censoring_rate)
    )]

    # Compute cumulative incidence and censoring rates.
    surv_group[
      ,
      ":="("cumulative_incidence" = cumsum(data.table::shift(
        survival_probability,
        n = 1L,
        fill = 1.0,
        type = "lag"
      ) * interval_incidence_rate))
    ]

    # Rename the outcome_time column.
    data.table::setnames(
      x = surv_group, 
      old = "outcome_time",
      new = "time"
    )

    # Retain only import
    surv_group <- surv_group[, c(
      "time", "survival_probability", "cumulative_incidence", "cumulative_censoring"
    )]

    if (!any(surv_group$time == 0.0)) {
      surv_group <- rbind(
        data.table::data.table(
          "time" = 0.0,
          "survival_probability" = 1.0,
          "cumulative_incidence" = 0.0,
          "cumulative_censoring" = 0.0
        ),
        surv_group
      )
    }

    # Number of samples
    distr_list[["n"]] <- nrow(x)

    # Number of events
    distr_list[["n_event"]] <- sum(x$outcome_event == 1L, na.rm = TRUE)

    # Five-number summary of follow-up
    distr_list[["follow_up_fivenum"]] <- fivenum_summary(x$outcome_time, na.rm = TRUE)

    # Five-number summary of event
    distr_list[["event_fivenum"]] <- fivenum_summary(x[outcome_event == 1L, ]$outcome_time, na.rm = TRUE)

    # Survival probability
    distr_list[["survival_probability"]] <- unique(surv_group[, c("time", "survival_probability")])

    # Cumulative incidence
    distr_list[["cumulative_incidence"]] <- unique(surv_group[, c("time", "cumulative_incidence")])

    # Censoring rate
    distr_list[["censoring_incidence"]] <- unique(surv_group[, c("time", "cumulative_censoring")])
  } else {
    ..error_no_known_outcome_type(object@outcome_type)
  }

  object@distribution <- distr_list

  return(object)
}



# show (outcomeInfo) -----------------------------------------------------------
setMethod(
  "show",
  signature(object = "outcomeInfo"),
  function(object) {
    # Basic outcome information.
    outcome_name <- object@name
    if (length(outcome_name) == 0L) outcome_name <- "outcome"

    # Form basic string.
    outcome_str <- paste0(outcome_name, " (", object@outcome_type, ")")

    # Adapt string per outcome-type
    if (object@outcome_type %in% c("binomial")) {
      # Show the reference class.
      outcome_classes <- object@levels
      outcome_classes[1L] <- paste0(outcome_classes[1L], " (reference)")

      # Add to the outcome string.
      outcome_str <- paste0(
        outcome_str, ", with classes: ",
        paste_s(outcome_classes), ".\n"
      )
      
    } else if (object@outcome_type %in% c("multinomial")) {
      # Show the outcome classes.
      outcome_str <- paste0(
        outcome_str, ", with classes: ",
        paste_s(object@levels), ".\n"
      )
      
    } else if (object@outcome_type %in% c("continuous")) {
      # No further details provided.
      outcome_str <- paste0(outcome_str, ".\n")
      
    } else if (object@outcome_type %in% c("survival")) {
      # Show the censoring and event values.
      if (length(object@censored) > 0L) {
        censoring_str <- paste0("censoring: ", paste_s(object@censored))
        event_str <- paste0("event: ", paste_s(object@event))

        # Add to outcome string.
        outcome_str <- paste0(
          outcome_str, ", with ", censoring_str,
          "; and ", event_str, ".\n"
        )
        
      } else {
        event_str <- paste0("event: ", paste_s(object@event))

        # Add to outcome string.
        outcome_str <- paste0(
          outcome_str, ", without censoring; and ", event_str, ".\n"
        )
      }
      
    } else if (object@outcome_type %in% c("competing_risk")) {
      if (length(object@censored) > 0L) {
        censoring_str <- paste0("with censoring: ", paste_s(object@censored))
      } else {
        censoring_str <- "without censoring"
      }

      if (length(object@competing_risk) > 0L) {
        competing_risk_str <- paste0(
          ifelse(
            length(object@competing_risk) > 1L,
            "competing risks: ",
            "competing risk: "
          ),
          paste_s(object@competing_risk)
        )
        
      } else {
        competing_risk_str <- "no competing risks"
      }

      event_str <- paste0("event: ", paste_s(object@event))

      # Add to outcome string.
      outcome_str <- paste0(
        outcome_str, ", ",
        censoring_str, "; ", 
        event_str, "; and ",
        competing_risk_str, ".\n"
      )
      
    } else if (object@outcome_type == "unsupervised") {
      # No further details provided.
      outcome_str <- paste0(outcome_str, ".\n")
      
    } else {
      ..error_no_known_outcome_type(object@outcome_type)
    }

    # Print to terminal.
    cat(paste0(outcome_str))
  }
)



# add_package_version (outcomeInfo) --------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "outcomeInfo"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)

.get_available_bootstrap_confidence_interval_methods <- function() {
  return(c("bc", "percentile"))
}



..bootstrap_ci <- function(
    x,
    x_0 = NULL,
    confidence_level = NULL,
    percentiles = NULL,
    bootstrap_ci_method = "percentile",
    ...
) {
  
  # Test confidence level
  if (!is.null(confidence_level)) {
    if (confidence_level >= 1.0) {
      ..error(
        "A 100% confidence interval does not exist.",
        error_class = "input_argument_error"
      )
      
    } else if (confidence_level <= 0.0) {
      ..error(
        "The confidence interval cannot be smaller than 0.",
        error_class = "input_argument_error"
      )
    }
    
    # Compute percentiles from confidence level, and add the median.
    percentiles <- c(
      0.5,
      (1.0 - confidence_level) / 2.0,
      1.0 - (1.0 - confidence_level) / 2.0
    )
  }
  
  if (!is.null(percentiles)) {
    if (any(percentiles > 1.0)) {
      ..error(
        "Percentiles cannot be greater than 1.",
        error_class = "input_argument_error"
      )
      
    } else if (any(percentiles < 0.0)) {
      ..error(
        "Percentiles cannot be smaller than 0.",
        error_class = "input_argument_error"
      )
    }
  }
  
  if (is.null(confidence_level) && is.null(percentiles)) {
    ..error_reached_unreachable_code(
      "..bootstrap_ic: confidence_interval and percentiles arguments cannot both be NULL."
    )
  }
  
  # For values that are not numeric, return mode. Note that the name "median" is
  # used for compatibility purposes.
  if (!is.numeric(x)) return(list("median" = get_mode(x)))
  
  # Set up percentile names.
  if (!is.null(confidence_level)) {
    # When a confidence level is provided, use standard naming.
    percentile_names <- c("median", "ci_low", "ci_up")
    
  } else {
    # Select unique percentiles as a precaution.
    percentiles <- unique(percentiles)
    
    # When confidence intervals are not provided, use naming based on the
    # percentiles. Determine the number of decimal digits that should be used to
    # parse the numbers (minimum 2).
    percentile_names_digits <- nchar(sub(".*\\.", "", percentiles))
    percentile_names_digits <- max(c(percentile_names_digits, 2L))
    
    # Set percentile names by passing them through the column name
    # checker.
    percentile_names <- .replace_illegal_column_name(
      paste0("q_", format(percentiles, nsmall = percentile_names_digits))
    )
  }
  
  # Create an empty list.
  empty_list <- as.list(rep(NA_real_, length(percentile_names)))
  names(empty_list) <- percentile_names
  
  # Select finite values.
  x <- x[is.finite(x)]
  
  if (length(x) == 0L) return(empty_list)
  
  # If no x_0 is provided, the percentile method should be used.
  if (length(x_0) == 0L) bootstrap_ci_method <- "percentile"
  
  if (bootstrap_ci_method == "percentile") {
    # Follows the percentile method of Efron, B. & Hastie, T. Computer Age
    # Statistical Inference. (Cambridge University Press, 2016).
   
    # Compute percentiles within the data set
    percentile_values <- stats::quantile(
      x,
      probs = percentiles,
      names = FALSE
    )
    
    # Generate a summary list
    summary_list <- as.list(percentile_values)
    names(summary_list) <- percentile_names
    
  } else if (bootstrap_ci_method == "bc") {
    # Follows the bias-corrected (BC) method of Efron, B. & Hastie, T. Computer
    # Age Statistical Inference. (Cambridge University Press, 2016).
    if (length(x_0) > 1L) {
      ..error(
        "The full-data point estimate should have length 1. Found: length ", length(x_0), "."
      )
    } 
    
    if (!is.finite(x_0)) return(empty_list)
    
    # Compute the bias-correction value. In absence of bias, z_0 equals 0.
    z_0 <- stats::qnorm(sum(x <= x_0) / length(x))
    
    if (!is.finite(z_0)) {
      # This means that we are dealing with an edge case where sum(x <= x_0) is
      # either 0 or the length of x. In that case we determine the confidence
      # interval with the percentile method.
      summary_list <- ..bootstrap_ci(
        x = x,
        x_0 = x_0,
        confidence_level = confidence_level,
        percentiles = percentiles,
        bootstrap_ci_method = "percentile"
      )
      
      return(summary_list)
    }
    
    # Define the z-statistic for bounds of the confidence interval.
    z_alpha <- stats::qnorm(percentiles)
    
    # Define bias-corrected percentiles.
    bc_percentiles <- stats::pnorm(2.0 * z_0 + z_alpha)
    
    # Compute percentiles within the data set
    percentile_values <- stats::quantile(
      x,
      probs = bc_percentiles,
      names = FALSE
    )
    
    # Generate a summary list
    summary_list <- as.list(percentile_values)
    names(summary_list) <- percentile_names
  }
  
  return(summary_list)
}


..bootstrap_bias_correction <- function(x) {
  
  # Define empty summary list.
  empty_list <- list("median" = NA_real_)
  
  # Select finite values.
  x <- x[is.finite(x)]
  
  if (length(x) == 0L) return(empty_list)
  
  # Compute the median value over the bootstraps.
  if (is.numeric(x)) {
    summary_list <- list("median" = stats::median(x))
    
  } else if (is.ordered(x)) {
    # Determine the mean average risk group. This requires discretisation
    # as rounding toward the nearest group would overinflate center groups.
    x_levels <- levels(x)
    n <- nlevels(x)
    
    # Discretise bins floor((mu - 1) / ((n-1) / n)) + 1. See fixed bin size
    # discretisation.
    x_num <- floor(n * (mean(as.numeric(x), na.rm = TRUE) - 1.0) / (n - 1.0)) + 1.0
    
    # Check if the x_num still falls within the range.
    x_num <- ifelse(x_num > n, n, x_num)
    
    # Re-encode and add to list.
    summary_list <- list(
      "median" = factor(x_levels[x_num], levels = x_levels, ordered = TRUE)
    )
    
  } else {
    summary_list <- list("median" = get_mode(x))
  }
  
  return(summary_list)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarDataElementConfusionMatrix object ------------------------------------

setClass(
  "familiarDataElementConfusionMatrix",
  contains = "familiarDataElement",
  prototype = methods::prototype(
    value_column = "count",
    grouping_column = c("observed_outcome", "expected_outcome")
  )
)



# extract_confusion_matrix (generic) -------------------------------------------

#'@title Internal function to extract the confusion matrix.
#'
#'@description Computes and extracts the confusion matrix for predicted and
#'  observed categorical outcomes used in a `familiarEnsemble` object.
#'
#'@inheritParams .extract_data
#'
#'@return A data.table containing predicted and observed outcome data together
#'  with a co-occurence count.
#'@md
#'@keywords internal
setGeneric(
  "extract_confusion_matrix",
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    detail_level = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_confusion_matrix")
  } 
)



# extract_confusion_matrix (familiarEnsemble) ----------------------------------
setMethod(
  "extract_confusion_matrix",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    detail_level = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE
  ) {
    
    # Don't compute a confusion matrix if there is nothing to be computed.
    if (!object@outcome_type %in% c("binomial", "multinomial")) return(NULL)
    
    # Message extraction start
    logger_message(
      paste0("Computing confusion matrix."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (is.waive(ensemble_method)) ensemble_method <- object@settings$ensemble_method

    proto_data_element <- .create_extract_confusion_matrix_object(
      object = object,
      ensemble_method = ensemble_method,
      detail_level = detail_level
    )
    
    # Generate elements to send to dispatch.
    confusion_matrix_data <- extract_dispatcher(
      FUN = .extract_confusion_matrix,
      has_internal_bootstrap = FALSE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      aggregate_results = TRUE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(confusion_matrix_data)
  }
)



# extract_confusion_matrix (prediction table) ----------------------------------
setMethod(
  "extract_confusion_matrix",
  signature("familiarDataElementPredictionTable"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    detail_level = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE
  ) {
    if (is_empty(object)) return(NULL)
    
    if (!is(object, "predictionTableClassification")) {
      ..warning_no_data_extraction_from_prediction_table("confusion matrix")
      
      return(NULL)
    }
    
    # Reference labels should be present.
    if (!.has_reference_data(object)) {
      ..warning_prediction_table_lacks_reference("confusion matrix")
      return(NULL)
    }
    
    # Message extraction start
    logger_message(
      paste0("Computing confusion matrix."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (is.waive(ensemble_method)) {
      ensemble_method <- "median"
      if (methods::.hasSlot(object, "ensemble_method")) ensemble_method <- object@ensemble_method
    }
    if (is.waive(detail_level)) detail_level <- "ensemble"
    
    # Copy object to prevent changing the provided object by reference.
    object <- .copy(object)
    
    proto_data_element <- .create_extract_confusion_matrix_object(
      object = object,
      ensemble_method = ensemble_method,
      detail_level = detail_level
    )
    
    # Generate elements to send to dispatch.
    confusion_matrix_data <- extract_dispatcher(
      FUN = .extract_confusion_matrix,
      has_internal_bootstrap = FALSE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      aggregate_results = TRUE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(confusion_matrix_data)
  }
)



.create_extract_confusion_matrix_object <- function(
    object,
    detail_level,
    ensemble_method
) {
  # Check the level detail.
  detail_level <- .parse_detail_level(
    x = detail_level,
    object = object,
    default = "ensemble",
    data_element = "confusion_matrix"
  )
  
  # Check ensemble_method argument
  .check_parameter_value_is_valid(
    x = ensemble_method,
    var_name = "ensemble_method",
    values = .get_available_ensemble_prediction_methods()
  )
  
  proto_data_element <- new(
    "familiarDataElementConfusionMatrix",
    detail_level = detail_level,
    estimation_type = "point"
  )
  
  return(proto_data_element)
}



.extract_confusion_matrix <- function(
    object,
    proto_data_element,
    ...
) {
  
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  if (!object@outcome_type %in% c("binomial", "multinomial")) {
    ..error_outcome_type_not_implemented(object@outcome_type)
  }

  # Add model name.
  data_element <- add_model_name(proto_data_element, object = object)
  
  return(..extract_confusion_matrix(
    object = object,
    data_element = data_element,
    ...
  ))
}



# ..extract_confusion_matrix (generic) -----------------------------------------
setGeneric(
  "..extract_confusion_matrix",
  function(object, ...) standardGeneric("..extract_confusion_matrix")
)



# ..extract_confusion_matrix (character) ---------------------------------------
setMethod(
  "..extract_confusion_matrix",
  signature(object = "character"),
  function(
    object,
    ...
  ){
    # Ensure that the object is loaded
    object <- load_familiar_object(object)
    
    return(..extract_confusion_matrix(object = object, ...))
  }
)



# ..extract_confusion_matrix (model, ensemble) ---------------------------------
setMethod(
  "..extract_confusion_matrix",
  signature(object = "familiarModelUnion"),
  function(
    object,
    data,
    ensemble_method,
    is_pre_processed,
    ...
  ) {
    # Predict class probabilities.
    prediction_data <- .predict(
      object = object,
      data = data,
      ensemble_method = ensemble_method,
      is_pre_processed = is_pre_processed
    )
    
    if (is_empty(prediction_data)) return(NULL)
    
    return(..extract_confusion_matrix(
      object = prediction_data,
      ...
    ))
  }
)



# ..extract_confusion (classification) -----------------------------------------
setMethod(
  "..extract_confusion_matrix",
  signature(object = "predictionTableClassification"),
  function(
    object,
    data_element,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    count <- NULL
    
    if (!all_predictions_valid(object)) return(NULL)
    
    # Remove data with missing outcomes.
    object <- filter_missing_outcome(object)
    
    # Check that any prediction data remain.
    if (is_empty(object)) return(NULL)
    
    # Make a local copy with only the required data
    data <- .copy(object)
    data <- .as_data_table(.complete(object))
    data <- data[, c("outcome", "predicted_class")]
    
    # Rename outcome columns
    data.table::setnames(
      x = data,
      old = c("outcome", "predicted_class"),
      new = c("observed_outcome", "expected_outcome")
    )
    
    # Sum pairs of observed and expected outcome categories.
    data <- data[
      , list("count" = .N),
      by = c("observed_outcome", "expected_outcome")
    ]
    
    # Find class levels in the data
    class_levels <- get_outcome_class_levels(object)
    
    # Construct an empty matrix 
    empty_matrix <- data.table::data.table(expand.grid(
      list(
        "observed_outcome" = class_levels,
        "expected_outcome" = class_levels
      ),
      stringsAsFactors = FALSE
    ))
    empty_matrix[, "count" := 0L]
    
    # Combine data with the empty matrix to add in combinations that appear 0
    # times.
    data <- data.table::rbindlist(list(data, empty_matrix), use.names = TRUE)
    
    # Use a max operation to remove any combinations that appear twice in the
    # table.
    data <- data[
      ,
      list("count" = max(count)),
      by = c("observed_outcome", "expected_outcome")
    ]
    
    # Set data element.
    data_element@data <- data
    
    return(data_element)
  }
)



# ..extract_confusion (NULL) ---------------------------------------------------
setMethod(
  "..extract_confusion_matrix",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



# export_confusion_matrix_data (generic) ---------------------------------------

#'@title Extract and export confusion matrices.
#'
#'@description Extract and export confusion matrics for models in a
#'  familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams plot_univariate_importance
#'
#'@inheritDotParams extract_confusion_matrix
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Confusion matrices are exported for individual and ensemble models.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_confusion_matrix_data
#'@md
#'@rdname export_confusion_matrix_data-methods
setGeneric(
  "export_confusion_matrix_data",
  function(
    object,
    dir_path = NULL,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_confusion_matrix_data")
  }
)



# export_confusion_matrix_data (collection) ------------------------------------

#'@rdname export_confusion_matrix_data-methods
setMethod(
  "export_confusion_matrix_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(.export(
      x = object,
      data_slot = "confusion_matrix",
      dir_path = dir_path,
      aggregate_results = TRUE,
      type = "performance",
      subtype = "confusion_matrix",
      export_collection = export_collection
    ))
  }
)



# export_confusion_matrix_data (general) ---------------------------------------

#'@rdname export_confusion_matrix_data-methods
setMethod(
  "export_confusion_matrix_data",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "confusion_matrix"
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_confusion_matrix_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
#' @include PredictionTable.R
NULL


# familiarDataElementCalibrationData object ------------------------------------
setClass(
  "familiarDataElementCalibrationData",
  contains = "familiarDataElement"
)

# familiarDataElementCalibrationLinearFit object -------------------------------
setClass(
  "familiarDataElementCalibrationLinearFit",
  contains = "familiarDataElement"
)

# familiarDataElementCalibrationGoodnessOfFit object ---------------------------
setClass(
  "familiarDataElementCalibrationGoodnessOfFit",
  contains = "familiarDataElement"
)

# familiarDataElementCalibrationDensity object ---------------------------------
setClass(
  "familiarDataElementCalibrationDensity",
  contains = "familiarDataElement"
)



# extract_calibration_data (generic) -------------------------------------------

#'@title Internal function to extract calibration data.
#'
#'@description Computes calibration data from a `familiarEnsemble` object.
#'  Calibration tests are performed based on expected (predicted) and observed
#'  outcomes. For all outcomes, calibration-at-the-large and calibration slopes
#'  are determined. Furthermore, for all but survival outcomes, a repeated,
#'  randomised grouping Hosmer-Lemeshow test is performed. For survival
#'  outcomes, the Nam-D'Agostino and Greenwood-Nam-D'Agostino tests are
#'  performed.
#'
#'@inheritParams .extract_data
#'
#'@return A list with data.tables containing calibration test information for
#'  the ensemble model.
#'@md
#'@keywords internal
setGeneric(
  "extract_calibration_data",
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_calibration_data")
  }
)


# extract_calibration_data (ensemble) ------------------------------------------
setMethod(
  "extract_calibration_data",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Assessing model calibration."),
      indent = message_indent,
      verbose = verbose
    )
    
    require_package(
      x = "harmonicmeanp",
      purpose = "to compute p-values for model calibration tests",
      message_type = "warning"
    )
    
    # Load evaluation_times from the object settings attribute, if it is not
    # provided.
    if (is.waive(evaluation_times)) evaluation_times <- object@settings$eval_times
    if (is.waive(ensemble_method)) ensemble_method <- object@settings$ensemble_method
    if (is.waive(confidence_level)) confidence_level <- object@settings$confidence_level
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- object@settings$bootstrap_ci_method
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    if (!model_is_trained(object = object)) return(NULL)
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "calibration_data"
    )
    
    proto_data_element <- .create_extract_calibration_data_object(
      object = object,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    # Generate elements to send to dispatch.
    calibration_data <- extract_dispatcher(
      FUN = .extract_calibration_data,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(calibration_data)
  }
)



# extract_calibration_data (prediction table) ----------------------------------
setMethod(
  "extract_calibration_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {

    # Message extraction start
    logger_message(
      paste0("Assessing model calibration."),
      indent = message_indent,
      verbose = verbose
    )
    
    require_package(
      x = "harmonicmeanp",
      purpose = "to compute p-values for model calibration tests",
      message_type = "warning"
    )
    
    if (is.waive(evaluation_times) && methods::.hasSlot(object, "time")) {
      evaluation_times <- object@time
    }
    if (is.waive(ensemble_method)) {
      ensemble_method <- "median"
      if (methods::.hasSlot(object, "ensemble_method")) ensemble_method <- object@ensemble_method
    } 
    
    # Default Values.
    if (is.waive(detail_level)) detail_level <- "ensemble"
    if (is.waive(estimation_type)) estimation_type <- "bootstrap_confidence_interval" 
    if (is.waive(confidence_level)) confidence_level <- 0.95
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- "bc"
    if (is.waive(aggregate_results)) aggregate_results <- TRUE
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "calibration_data"
    )
    
    proto_data_element <- .create_extract_calibration_data_object(
      object = object,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    # Generate elements to send to dispatch.
    calibration_data <- extract_dispatcher(
      FUN = .extract_calibration_data,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(calibration_data)
  }
)



.create_extract_calibration_data_object <- function(
    object,
    ensemble_method,
    evaluation_times,
    detail_level,
    estimation_type,
    confidence_level,
    bootstrap_ci_method
) {
    # Check evaluation_times argument
  if (object@outcome_type %in% c("survival")) {
    sapply(
      evaluation_times,
      .check_number_in_valid_range,
      var_name = "evaluation_times",
      range = c(0.0, Inf),
      closed = c(FALSE, TRUE)
    )
  }
  
  # Check ensemble_method argument
  .check_parameter_value_is_valid(
    x = ensemble_method,
    var_name = "ensemble_method",
    values = .get_available_ensemble_prediction_methods()
  )
  
  # Check confidence_level input argument
  .check_number_in_valid_range(
    x = confidence_level,
    var_name = "confidence_level",
    range = c(0.0, 1.0),
    closed = c(FALSE, FALSE)
  )
  
  # Check boostrap method
  .check_parameter_value_is_valid(
    x = bootstrap_ci_method,
    var_name = "bootstrap_ci_method",
    values = .get_available_bootstrap_confidence_interval_methods()
  )
  
  # Check the level detail.
  detail_level <- .parse_detail_level(
    x = detail_level,
    object = object,
    default = "hybrid",
    data_element = "calibration_data"
  )
  
  # Check the estimation type.
  estimation_type <- .parse_estimation_type(
    x = estimation_type,
    object = object,
    default = "bootstrap_confidence_interval",
    data_element = "calibration_data",
    detail_level = detail_level,
    has_internal_bootstrap = TRUE
  )
  
  # Generate a prototype data element.
  proto_data_element <- new(
    "familiarDataElementCalibrationData",
    detail_level = detail_level,
    estimation_type = estimation_type,
    confidence_level = confidence_level,
    bootstrap_ci_method = bootstrap_ci_method
  )
  
  return(proto_data_element)
}



.extract_calibration_data <- function(
    object,
    proto_data_element,
    evaluation_times = NULL,
    aggregate_results,
    cl,
    ...
) {
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  proto_data_element <- add_model_name(proto_data_element, object = object)
  
  # Add evaluation time as a identifier to the data element.
  if (length(evaluation_times) > 0L && object@outcome_type == "survival") {
    data_elements <- add_data_element_identifier(
      x = proto_data_element,
      evaluation_time = evaluation_times
    )
    
  } else {
    data_elements <- list(proto_data_element)
  }
  
  # Iterate over data elements.
  calibration_data <- mapply(
    ..extract_calibration_data,
    data_element = data_elements,
    MoreArgs = list(
      object = object,
      aggregate_results = aggregate_results,
      cl = cl,
      ...
    ),
    SIMPLIFY = FALSE
  )
  
  return(calibration_data)
}



# ..extract_calibration_data (generic) -----------------------------------------
setGeneric(
  "..extract_calibration_data",
  function(object, ...) standardGeneric("..extract_calibration_data")
)



# ..extract_calibration_data (character) ---------------------------------------
setMethod(
  "..extract_calibration_data",
  signature(object = "character"),
  function(
    object,
    ...
  ){
    # Ensure that the object is loaded
    object <- load_familiar_object(object)
    
    return(..extract_calibration_data(object = object, ...))
  }
)



# ..extract_calibration_data (model, ensemble) ---------------------------------
setMethod(
  "..extract_calibration_data",
  signature(object = "familiarModelUnion"),
  function(
    object,
    data,
    data_element,
    is_pre_processed,
    ensemble_method,
    ...
  ) {
    # Aggregate data.
    data <- aggregate_data(data)
    
    if (object@outcome_type %in% c("survival", "competing_risk")) {
      # Predict class probabilities or regression values.
      prediction_data <- .predict(
        object = object,
        data = data,
        type = "survival_probability",
        time = data_element@identifiers$evaluation_time,
        ensemble_method = ensemble_method,
        is_pre_processed = is_pre_processed
      )
      
    } else {
      # Predict class probabilities or regression values.
      prediction_data <- .predict(
        object = object,
        data = data,
        ensemble_method = ensemble_method,
        is_pre_processed = is_pre_processed
      )
    }
    
    return(..extract_calibration_data(
      object = prediction_data,
      data_element = data_element,
      ...
    ))
  }
)



# ..extract_calibration_data (prediction table) --------------------------------
setMethod(
  "..extract_calibration_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(
    object,
    data_element,
    aggregate_results,
    cl = NULL,
    progress_bar = FALSE,
    verbose = FALSE,
    message_indent,
    ...
  ) {
    if (!(
      is(object, "predictionTableSurvivalProbability") || 
      is(object, "predictionTableClassification") ||
      is(object, "predictionTableRegression")
    )) {
      ..warning_no_data_extraction_from_prediction_table("calibration data")
      
      return(NULL)
    }
    
    # Message the user concerning the time at which metrics are computed. This
    # is only relevant for survival analysis.
    if (length(data_element@identifiers$evaluation_time) > 0L && progress_bar) {
      logger_message(
        paste0(
          "Assessing model calibration at time ",
          data_element@identifiers$evaluation_time, "."
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    
    # Check if any predictions are valid.
    if (!all_predictions_valid(object)) return(NULL)
    
    # Remove data with missing outcomes.
    object <- filter_missing_outcome(object)
    
    # Check that any prediction data remain.
    if (is_empty(object)) return(NULL)
    
    # Add positive class as an identifier.
    if (object@outcome_type %in% c("binomial")) {
      data_element <- add_data_element_identifier(
        x = data_element,
        positive_class = get_outcome_class_levels(object)[2L]
      )
      
    } else if (object@outcome_type %in% c("multinomial")) {
      data_element <- add_data_element_identifier(
        x = data_element,
        positive_class =  get_outcome_class_levels(object)
      )
    }

    # Add bootstrap data.
    bootstrap_data <- add_data_element_bootstrap(
      x = data_element,
      ...
    )
    
    # Add distribution data.
    if (!is.list(data_element)) data_element <- list(data_element)
    
    density_data <- lapply(
      data_element,
      .compute_calibration_data_density,
      object = object
    )
    
    # Iterate over elements.
    data_elements <- fam_mapply(
      cl = cl,
      assign = NULL,
      FUN = .compute_calibration_data,
      data_element = bootstrap_data$data_element,
      bootstrap = bootstrap_data$bootstrap,
      bootstrap_seed = bootstrap_data$seed,
      MoreArgs = list("object" = object),
      progress_bar = progress_bar && verbose,
      chopchop = TRUE
    )
    
    # Flatten list of data elements.
    data_elements <- unlist(data_elements)
    if (!is.list(data_elements)) data_elements <- list(data_elements)

    # Add in density data elements.
    data_elements <- c(data_elements, density_data)
    
    # Merge data elements
    data_elements <- merge_data_elements(data_elements)
    
    if (aggregate_results) data_elements <- .compute_data_element_estimates(x = data_elements)
    
    return(data_elements)
  }
)



# ..extract_calibration_data (NULL) --------------------------------------------
setMethod(
  "..extract_calibration_data",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



.compute_calibration_data <- function(
    data_element,
    object,
    bootstrap,
    bootstrap_seed
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  observed <- NULL
  
  # Start random number generator.
  rstream_object <- .start_random_number_stream(seed = ifelse(is.finite(bootstrap_seed), bootstrap_seed, 19L))
  
  # Bootstrap the data.
  if (bootstrap) {
    object <- get_bootstrap_sample(data = object, rstream_object = rstream_object)
  }
  
  # Compute calibration, linear and gof test data.
  calibration_data <- ..compute_calibration_data(
    object,
    data_element = data_element,
    rstream_object = rstream_object
  )
  
  # Extract data.
  calibration_at_large <- calibration_data$linear
  calibration_gof_test <- calibration_data$gof_test
  calibration_data <- calibration_data$calibration

  if (!is_empty(calibration_data) && data_element@estimation_type != "point") {
    calibration_data <- ..compute_calibration_data_interpolated(
      data = calibration_data
    )
    
    # Keep only finite predictions,
    calibration_data <- calibration_data[is.finite(observed)]
  }
  
  # Create additional data elements to contain linear fit data and GoF test
  # data.
  linear_fit_data_element <- methods::new(
    "familiarDataElementCalibrationLinearFit",
    data_element
  )  
  gof_data_element <- methods::new(
    "familiarDataElementCalibrationGoodnessOfFit",
    data_element
  )
  
  # Set for calibration data.
  data_element@data <- calibration_data
  data_element@value_column <- "observed"
  data_element@grouping_column <- "expected"
  
  # Set linear fit data, i.e. calibration-in-the-large and calibration slope.
  linear_fit_data_element@data <- calibration_at_large
  linear_fit_data_element@value_column <- "value"
  linear_fit_data_element@grouping_column <- "type"
  
  # Set goodness of fit data.
  gof_data_element@data <- calibration_gof_test
  gof_data_element@value_column <- "p_value"
  gof_data_element@grouping_column <- "type"
  
  return(list(
    data_element,
    linear_fit_data_element,
    gof_data_element
  ))
}



# ..compute_calibration_data (generic) -----------------------------------------
setGeneric(
  "..compute_calibration_data",
  function(object, ...) standardGeneric("..compute_calibration_data")
)




# ...compute_calibration_data (generic) ----------------------------------------
setGeneric(
  "...compute_calibration_data",
  function(object, ...) standardGeneric("...compute_calibration_data")
)



# ..compute_calibration_data (survival probability) ----------------------------
setMethod(
  "..compute_calibration_data",
  signature(object = "predictionTableSurvivalProbability"),
  function(object, data_element, rstream_object = NULL) {
    # Generate baseline calibration data for survival models from an input table
    # (data) containing survival probabilities for each sample and
    # corresponding time and event status.
    
    # Suppress NOTES due to non-standard evaluation in data.table
    exp_prob <- NULL
    
    # Set time.
    time <- data_element@identifiers$evaluation_time

    # Find sample identifiers.
    sample_identifiers <- get_id_columns(id_depth = "series")
    
    # Rename the survival column to standard name.
    data <- data.table::copy(.as_data_table(object))
    data.table::setnames(
      data,
      old = c("outcome_time", "outcome_event", "predicted_outcome"),
      new = c("time", "event", "exp_prob")
    )
    
    # Sort by survival probability.
    data <- data[order(exp_prob)]
    
    # Randomly split into groups. The number of groups is determined using
    # sturges rule.
    group_data <- .calibration_create_randomised_groups(
      x = data$exp_prob,
      y = data$event,
      sample_identifiers = data[, mget(sample_identifiers)],
      n_min_y_in_group = 4L,
      rstream_object = rstream_object
    )
    
    if (is_empty(group_data)) return(NULL)
    
    # Merge with prediction table.
    group_data <- merge(
      x = data,
      y = group_data,
      by = sample_identifiers,
      allow.cartesian = TRUE
    )
    
    # Iterate over groups and add details by comparing the kaplan-meier survival
    # curve within each group at time with the mean survival probability in
    # the group.
    calibration_data <- lapply(
      split(group_data, by = "group_id"),
      function(x, object, time) {
        ...compute_calibration_data(
          object = object,
          data = x,
          time = time
        )
      },
      object = object,
      time = time
    )
    
    # Combine to a single list.
    calibration_data <- data.table::rbindlist(
      calibration_data,
      use.names = TRUE
    )
    
    # Prevent processing of empty tables.
    if (is_empty(calibration_data)) return(NULL)
    
    # Set column order.
    data.table::setcolorder(
      x = calibration_data,
      neworder = c("expected", "observed", "km_var", "n_g")
    )
    
    # Calibration-in-the-large and calibration slope
    calibration_at_large <- .compute_calibration_linear_fit(
      calibration_data = calibration_data,
      outcome_type = object@outcome_type
    )
    
    # Nam-D'Agostino tests
    calibration_gof_test <- .compute_calibration_nam_dagostino(
      calibration_data = calibration_data
    )
    
    return(list(
      "calibration" = calibration_data,
      "linear" = calibration_at_large,
      "gof_test" = calibration_gof_test
    ))
  }
)



# ..compute_calibration_data (classification) ----------------------------------
setMethod(
  "..compute_calibration_data",
  signature(object = "predictionTableClassification"),
  function(object, data_element, rstream_object = NULL) {
    # For assessing the calibration of categorical outcomes, we require expected
    # and observed probabilities for 1 (binomial) or all classes (multinomial).
    # Expected probabilities are easy, as the model predicts them. Observed
    # probabilities are however based on class proportion within a group. This
    # means that we need to define groups, ordered by predicted probability. The
    # same groups need to be defined for the Hosmer-Lemeshow test, so we might
    # as well obtain all the data here.
    #
    # Groups are defined for each class level because the samples should be
    # ordered according to predicted class probability to create the groups.
    
    # Suppress NOTES due to non-standard evaluation in data.table
    observed_class <- exp_prob <- NULL
    
    positive_class <- data_element@identifiers$positive_class

    # Identify the real outcome columns
    outcome_column <- get_outcome_columns(x = "binomial")
    
    # Find sample identifiers.
    sample_identifiers <- get_id_columns(id_depth = "series")
    
    # Make local copy.
    data <- data.table::copy(
      .as_data_table(object)[, mget(c(sample_identifiers, outcome_column, positive_class))]
    )
    
    # Rename columns to standard names
    data.table::setnames(
      data,
      old = c(outcome_column, positive_class),
      new = c("observed_class", "exp_prob")
    )
    
    # Mask class so that positive outcomes are TRUE and the rest FALSE
    data[, "observed_class" := observed_class == positive_class]
    
    # Sort by probability
    data <- data[order(exp_prob)]
    
    # Randomly split into groups. The number of groups is determined using
    # sturges rule.
    group_data <- .calibration_create_randomised_groups(
      x = data$exp_prob,
      sample_identifiers = data[, mget(sample_identifiers)],
      rstream_object = rstream_object
    )
    
    if (is_empty(group_data)) return(NULL)
    
    # Merge with prediction table.
    group_data <- merge(
      x = data,
      y = group_data,
      by = sample_identifiers,
      allow.cartesian = TRUE
    )
    
    # Compute calibration data from each group.
    calibration_data <- lapply(
      split(group_data, by = "group_id"),
      function(x, object) {
        ...compute_calibration_data(
          object = object,
          data = x
        )
      },
      object = object
    )
    
    # Combine to a single list.
    calibration_data <- data.table::rbindlist(
      calibration_data,
      use.names = TRUE
    )
    
    # Check that any calibration data was generated.
    if (is_empty(calibration_data)) return(NULL)
    
    # Set column order
    data.table::setcolorder(
      x = calibration_data,
      neworder = c("expected", "observed", "n_g", "n_pos", "n_neg")
    )
    
    # Calibration-in-the-large and calibration slope
    calibration_at_large <- .compute_calibration_linear_fit(
      calibration_data = calibration_data,
      outcome_type = object@outcome_type
    )
    
    # Hosmer-Lemeshow tests
    calibration_gof_test <- .compute_calibration_hosmer_lemeshow(
      calibration_data = calibration_data
    )
    
    return(list(
      "calibration" = calibration_data,
      "linear" = calibration_at_large,
      "gof_test" = calibration_gof_test
    ))
  }
)


# ..compute_calibration_data (regression) --------------------------------------
setMethod(
  "..compute_calibration_data",
  signature(object = "predictionTableRegression"),
  function(object, data_element, rstream_object = NULL) {
    # Calibration for regression problems is pretty straightforward. However, for
    # goodness-of-fit tests, we need to constrain expected and observed value
    # ranges to [0, 1]. To do so, we use the range of outcome values from the
    # development data, as stored in the calibration_info slot of the input
    # object.
    
    # Suppress NOTES due to non-standard evaluation in data.table
    outcome <- predicted_outcome <- NULL
    n_groups <- 1L
    
    # Extract data.
    data <- data.table::copy(.as_data_table(object))
    
    # Find sample identifiers.
    sample_identifiers <- get_id_columns(id_depth = "series")
    
    # Determine the outcome range
    outcome_range <- object@observed_value_range
    if (anyNA(outcome_range)) outcome_range <- range(data$outcome, na.rm = TRUE, finite = TRUE)
    
    # Get the shift and scale parameters.
    norm_shift <- outcome_range[1L]
    norm_scale <- diff(outcome_range)
    
    # Set scale parameters equal to 0.0 to 1.0 to avoid division by 0.
    if (norm_scale == 0.0) norm_scale <- 1.0
    
    # Apply normalisation.
    data[, ":="(
      "expected" = (predicted_outcome - norm_shift) / norm_scale,
      "observed" = (outcome - norm_shift) / norm_scale
    )]
    
    # Randomly split into groups. The number of groups is determined using
    # sturges rule.
    group_data <- .calibration_create_randomised_groups(
      x = data$expected,
      sample_identifiers = data[, mget(sample_identifiers)],
      rstream_object = rstream_object
    )
    
    if (is_empty(group_data)) return(NULL)
    
    # Merge with prediction table.
    group_data <- merge(
      x = data,
      y = group_data,
      by = sample_identifiers,
      allow.cartesian = TRUE
    )
    
    # Compute calibration data from each group.
    calibration_data <- lapply(
      split(group_data, by = "group_id"),
      function(x, object) {
        ...compute_calibration_data(
          object = object,
          data = x
        )
      },
      object = object
    )
    
    # Combine to a single list.
    calibration_data <- data.table::rbindlist(
      calibration_data,
      use.names = TRUE
    )
    
    # Check if the resulting table contains data.
    if (is_empty(calibration_data)) return(NULL)
    
    # Make columns match the expected order
    data.table::setcolorder(
      x = calibration_data,
      neworder = c("expected", "observed", "n_g")
    )
    
    # Calibration-in-the-large and calibration slope
    calibration_at_large <- .compute_calibration_linear_fit(
      calibration_data = calibration_data,
      outcome_type = object@outcome_type
    )
    
    # Hosmer-Lemeshow tests
    calibration_gof_test <- .compute_calibration_hosmer_lemeshow(
      calibration_data = calibration_data
    )
    
    return(list(
      "calibration" = calibration_data,
      "linear" = calibration_at_large,
      "gof_test" = calibration_gof_test
    ))
  }
)



# ...compute_calibration_data (survival probability) ---------------------------
setMethod(
  "...compute_calibration_data",
  signature(object = "predictionTableSurvivalProbability"),
  function(
    object,
    data,
    time
  ) {
    # Set default values.
    obs_prob <- exp_prob <- km_var <- NA_real_
    n_g <- NA_integer_
    
    if (nrow(data) >= 2L) {
      
      # Fit a Kaplan-Meier curve for the current group
      km_fit <- survival::survfit(Surv(time, event) ~ 1, data = data)
      
      if (length(km_fit$time) >= 2L) {
        
        # Get observed probability
        obs_prob <- stats::approx(
          x = km_fit$time,
          y = km_fit$surv,
          xout = time,
          method = "linear",
          rule = 2L
        )$y
        
        # Get expected probability
        exp_prob <- mean(data$exp_prob)
        
        # Get group size
        n_g <- nrow(data)
        
        # Get greenwood variance estimate.
        km_var <- stats::approx(
          x = km_fit$time,
          y = km_fit$std.err,
          xout = time,
          method = "linear",
          rule = 2L
        )$y^2.0
        
      } 
    }
    
    # Create table.
    calibration_table <- list(
      "expected" = exp_prob,
      "observed" = obs_prob,
      "n_g" = n_g,
      "km_var" = km_var
    )
    
    return(calibration_table)
  }
)



# ...compute_calibration_data (classification) ---------------------------------
setMethod(
  "...compute_calibration_data",
  signature(object = "predictionTableClassification"),
  function(
    object,
    data
  ) {
    # Check that the groups list contains at least one entry.
    if (is_empty(data)) return(NULL)
  
    # Create table
    calibration_table <- list(
      # Mean expected probability in a group.
      "expected" = mean(data$exp_prob),
      # Observed proportion of positive class in a group.
      "observed" = mean(data$observed_class),
      # Number of samples in the group
      "n_g" = nrow(data),
      # Number of samples with the positive class in each group.
      "n_pos" = sum(data$observed_class),
      # Number of samples with the negative class in each group.
      "n_neg" = sum(!data$observed_class)
    )
    
    return(calibration_table)
  }
)
    


# ...compute_calibration_data (regression) ---------------------------------
setMethod(
  "...compute_calibration_data",
  signature(object = "predictionTableRegression"),
  function(
    object,
    data
  ) {
    # Check that the groups list contains at least one entry.
    if (is_empty(data)) return(NULL)
    
    # Create table
    calibration_table <- list(
      # Mean expected value in the group.
      "expected" = mean(data$expected),
      # Observed value in the group.
      "observed" = mean(data$observed),
      # Number of samples in the group
      "n_g" = nrow(data)
    )
    
    return(calibration_table)
  }
)



#' Create randomised groups Creates randomised groups, e.g. for tests that
#' depend on splitting (continuous) data into groups, such as the
#' Hosmer-Lemeshow test
#'
#' The default fast mode is based on random sampling, whereas the slow mode is
#' based on probabilistic joining of adjacent groups. As the name suggests, fast
#' mode operates considerably more efficient.
#'
#' @param x Vector with data used for sorting. Groups are formed based on
#'   adjacent values.
#' @param y Vector with markers, e.g. the events. Should be 0 or 1 (for an
#'   event).
#' @param sample_identifiers data.table with sample_identifiers. If provide, a
#'   list of grouped sample_identifiers will be returned, and integers
#'   otherwise.
#' @param n_max_groups Maximum number of groups that need to be formed.
#' @param n_min_groups Minimum number of groups that need to be formed.
#' @param n_min_y_in_group Minimum number of y=1 in each group for a valid
#'   group.
#'
#' @details Creates randomised groups, e.g. for tests that depend on splitting
#'   (continuous) data into groups, such as the Hosmer-Lemeshow test
#'
#' @return data.table with sample identifiers and random group ids.
#' @md
#' @keywords internal
.calibration_create_randomised_groups <- function(
    x, 
    y = NULL, 
    sample_identifiers, 
    n_max_groups = NULL, 
    n_min_groups = NULL, 
    n_min_y_in_group = NULL, 
    unique_samples_only = TRUE,
    rstream_object = NULL
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  group_id <- cum_y <- weight <- cum_prob_lower <- cum_prob_upper <- NULL
  exclude <- y_in_group <- NULL
  
  # Populate the generic table.
  data <- data.table::copy(sample_identifiers[, mget(get_id_columns(id_depth = "series"))])
  data[, "x" := x]
  if (is.null(y)) {
    data[, "y" := 0L]
  } else {
    data[, "y" := y]
  }
  if (unique_samples_only) data <- unique(data)
  
  # - Get number of x
  n_x <- nrow(data)
  
  # - Get number of y
  n_y <- sum(data$y)
  
  # - Get the maximum of groups
  if (is.null(n_max_groups)) {
    n_max_groups <- ceiling(2.5 * n_x^(1.0 / 3.0))
  }
  
  # - Update maximum number of groups
  if (n_y > 0.0 && !is.null(n_min_y_in_group)) {
    n_max_groups <- min(c(n_max_groups, floor(n_y / n_min_y_in_group)))
  }
  
  # - Update mininum number of groups based on n_max_groups
  if (is.null(n_min_groups)) {
    n_min_groups <- min(c(n_max_groups, ceiling(1.0 * n_x^(1.0 / 3.0))))
  }
  
  # Some checks
  if (n_max_groups < n_min_groups || n_max_groups > n_x || n_max_groups < 2L) {
    return(NULL)
  }
  data <- data[order(x)]
  
  # Initial loop counter
  loop_iter <- 0L
  
  while (TRUE) {
    # Draw a random number of groups between n_min_groups and n_max_groups
    if (n_min_groups == n_max_groups) {
      n_group_draw <- n_max_groups
    } else {
      n_group_draw <- fam_sample(
        x = seq.int(from = n_min_groups, to = n_max_groups, by = 1L),
        size = 1L, 
        replace = FALSE,
        rstream_object = rstream_object
      )
    }
    
    # Draw a randomised groups assignment
    random_group_id <- sort(
      fam_sample(
        x = seq_len(n_group_draw),
        size = n_x, 
        replace = TRUE,
        rstream_object = rstream_object
      )
    )
    
    # Assign group id
    data[, "group_id" := random_group_id]
    
    # Check if all groups have the minimum required y (e.g. events/group size)
    if (n_y > 0L && !is.null(n_min_y_in_group)) {
      dt_y <- data[, list(y_in_group = sum(y)), by = group_id]
      if (all(dt_y$y_in_group >= n_min_y_in_group)) break
      
    } else {
      # If there is no minimum number of events required, only run one
      # iteration
      break
    }
    
    # Update loop_iter in case the sample was not good enough
    loop_iter <- loop_iter + 1L
    
    # Break from outer loop after 10 unsuccessful random samplings and create
    # a static split instead
    if (loop_iter >= 10L) {
      # Cumulative sum over y
      data[, "cum_y" := cumsum(y)]
      
      # Assign static group ids
      data[, "group_id" := findInterval(
        x = cum_y,
        vec = c(
          -Inf,
          stats::quantile(x = cum_y, (1L:(n_min_groups - 1L)) / n_min_groups),
          Inf
        )
      )]
      
      # Break from loop
      break
    }
  }
  
  # Get sample identifiers for each group
  return(data[, mget(c(get_id_columns(id_depth = "series"), "group_id"))])
}



.compute_calibration_data_density <- function(
    data_element,
    object
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  predicted_outcome <- NULL
  
  # Copy data element.
  data_element <- methods::new(
    "familiarDataElementCalibrationDensity",
    data_element
  )
  
  # Copy data
  data <- data.table::copy(.as_data_table(object))
  
  if (object@outcome_type %in% c("survival", "competing_risk")) {
    # Rename the survival column to standard name.
    data.table::setnames(
      data,
      old = "predicted_outcome",
      new = "expected"
    )
    
  } else if (object@outcome_type %in% c("binomial", "multinomial")) {
    # Rename class probability columns to standard names.
    data.table::setnames(
      data,
      old = data_element@identifiers$positive_class,
      new = "expected"
    )
    
  } else if (object@outcome_type %in% c("continuous")) {
    # Determine the outcome range
    outcome_range <- range(object@observed_value_range)
    
    # Get the shift and scale parameters.
    norm_shift <- outcome_range[1L]
    norm_scale <- diff(outcome_range)
    
    # Set scale parameters equal to 0.0 to 1.0 to avoid division by 0.
    if (norm_scale == 0.0) norm_scale <- 1.0
    
    # Apply normalisation.
    data[, ":="("expected" = (predicted_outcome - norm_shift) / norm_scale)]
    
  } else {
    ..error_outcome_type_not_implemented(object@outcome_type)
  }
  
  # Get expected values.
  expected <- data$expected
  
  # Set interpolation range for expected values. This allows for aggregating
  # observed values based on the expected value.
  expected_interpolated <- seq(
    from = 0.000,
    to = 1.000,
    by = 0.005
  )
  
  # Determine the bin for expected.
  n <- length(expected_interpolated)
  
  # Discretise bins.
  group_bin <- as.integer(floor((n - 1L) * expected)) + 1L
  
  # Check if the group_bin still falls within the valid range.
  group_bin <- ifelse(group_bin > n, n, group_bin)
  group_bin <- ifelse(group_bin < 1L, 1L, group_bin)
  
  # Compute the frequency of each entry.
  data <- data.table::data.table("expected" = expected_interpolated[group_bin])
  data <- data[, list("frequency" = .N / length(group_bin)), by = "expected"]
  
  # Set linear fit data, i.e. calibration-in-the-large and calibration slope.
  data_element@data <- data
  data_element@value_column <- "frequency"
  data_element@grouping_column <- "expected"
  
  return(data_element)
}



.compute_calibration_linear_fit <- function(
    calibration_data, 
    outcome_type
) {
  # Tests calibration-at-the-large and calibration slope
  
  # Suppress NOTES due to non-standard evaluation in data.table.
  expected <- observed <- p_value <- NULL
  
  # Check that calibration data are not empty.
  if (is_empty(calibration_data)) return(NULL)
  
  # Filter out non-finite expected and observed values.
  calibration_data <- calibration_data[is.finite(expected) & is.finite(observed)]
  
  # Check if the input is empty or only contains one entry
  if (nrow(calibration_data) < 2L) return(NULL)
  
  # Fit per repeated measurement.
  fit_data <- ..compute_calibration_fit(
    calibration_data,
    outcome_type = outcome_type
  )
  
  # Remove NaNs
  fit_data <- fit_data[is.finite(p_value)]
  if (is_empty(fit_data)) return(NULL)
  
  return(fit_data)
}



..compute_calibration_fit <- function(
    calibration_data,
    outcome_type
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table.
  type <- value <- ci_low <- ci_up <- NULL
  
  # Perform a linear fit. Note that the slope is offset by 1*expected to
  # directly compare with the expected slope of 1.
  fit <- stats::lm(
    observed ~ expected + 1 + offset(expected),
    data = calibration_data
  )
  
  # Get coefficients for intercept and slope
  fit_coef <- stats::coef(fit)
  
  # Determine if there is no slope.
  no_slope <- is.na(fit_coef["expected"]) && data.table::uniqueN(calibration_data$expected) == 1L
  if (no_slope) fit_coef["expected"] <- 0.0
  
  # Get confidence intervals
  fit_conf_int <- suppressWarnings(stats::confint(fit))
  if (no_slope) fit_conf_int["expected", ] <- c(-Inf, Inf)
  
  # Get summary
  fit_summary <- suppressWarnings(stats::summary.lm(fit))$coefficients
  if (no_slope) {
    fit_summary <- rbind(
      fit_summary,
      matrix(data = c(0.0, Inf, 0.0, 1.0), ncol = 4L)
    )
    rownames(fit_summary) <- c("(Intercept)", "expected")
  }
  
  # Correct for having multiple overlapping groups in binomial and multinomial
  # tests.
  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    
    # Determine the number of groups.
    n_groups <- 1L
    
    # Recompute the standard deviation
    fit_summary[, 2L] <- fit_summary[, 2L] * sqrt(n_groups)
    
    # Recompute the t-score
    fit_summary[, 3L] <- fit_summary[, 1L] / fit_summary[, 2L]
    fit_summary[, 3L][fit_summary[, 2L] == 0.0] <- Inf
    
    # Recompute the p-value
    fit_summary[, 4L] <- 2.0 * stats::pt(abs(fit_summary[, 3L]), fit$df.residual, lower.tail = FALSE)
    
    # Adapt the confidence intervals
    fit_conf_int[, 1L] <- fit_summary[, 1L] + fit_summary[, 2L] * stats::qnorm(0.025)
    fit_conf_int[, 2L] <- fit_summary[, 1L] + fit_summary[, 2L] * stats::qnorm(0.975)
  }
  
  # Construct data table and update the slope so that the expected slope
  # (slope+1) is shown instead.
  calibration_at_large <- data.table::data.table(
    "type" = c("offset", "slope"),
    "value" = fit_coef + c(0.0, 1.0),
    "ci_low" = fit_conf_int[, 1L] + c(0.0, 1.0),
    "ci_up" = fit_conf_int[, 2L] + c(0.0, 1.0),
    "p_value" = fit_summary[, 4L]
  )
  
  return(calibration_at_large)
}



.compute_calibration_hosmer_lemeshow <- function(calibration_data) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  observed <- expected <- n_g <- hm_group <- statistic <- n_groups <- NULL
  
  # Check for empty calibration tables
  if (is_empty(calibration_data)) return(NULL)
  
  # Make local copy of the calibration data.
  calibration_data <- data.table::copy(calibration_data)
  
  # Remove NA predictions.
  calibration_data <- calibration_data[is.finite(expected)]
  if (is_empty(calibration_data)) return(NULL)
  
  # Compute test statistic for each group
  calibration_data[
    ,
    "hm_group" := ..compute_calibration_test_statistic(observed, expected, n_g),
    by = seq_len(nrow(calibration_data))
  ]
  
  # Remove NA values in hm_group
  calibration_data <- calibration_data[is.finite(hm_group)]
  if (is_empty(calibration_data)) return(NULL)
  
  # Compute test statistic for each time point
  gof_table <- calibration_data[
    ,
    list("statistic" = sum(hm_group, na.rm = TRUE), n_groups = .N)
  ]
  
  # Remove all entries with n_groups < 3
  gof_table <- gof_table[n_groups >= 3L]
  
  if (is_empty(gof_table)) return(NULL)
  
  # Perform chi-square test
  gof_table <- gof_table[
    ,
    list("p_value" = stats::pchisq(
      q = statistic,
      df = n_groups - 2L,
      lower.tail = FALSE
    ))
  ]
  
  # Add type.
  gof_table[, "type" := "hosmer_lemeshow"]

  # Reorder columns
  data.table::setcolorder(gof_table, c("type", "p_value"))
  
  return(gof_table)
}



..compute_calibration_test_statistic <- function(
    observed,
    expected,
    n_g,
    method = "approximate"
) {
  
  if (!method %in% c("exact", "approximate")) {
    ..error_reached_unreachable_code(
      "..compute_hm_test_statistic: method should be exact or approximate"
    )
  }
  
  # Check for infinite or NA values.
  if (!is.finite(expected) || !is.finite(observed)) return(NA_real_)
  
  if (method == "exact") {
    # Use exact computation for comparison. Note that this has asymptotic
    # behaviour for expected -> 0 and expected -> 1.
    value <- (observed - expected)^2.0 * n_g / (expected * (1.0 - expected))
    
    return(value)
    
  } else if (method == "approximate") {
    # Use exact computation for comparison.
    exact_value <- (observed - expected)^2.0 * n_g / (expected * (1.0 - expected))
    if (!is.finite(exact_value)) exact_value <- Inf
    
    if (expected <= 0.05) {
      
      # Use Taylor series for 1/(x * (1-x)) to second order at 0.05.
      value <- n_g * (expected - observed)^2.0 * 400.0 *
        (1.0 / 19.0 - 360.0 / 361.0 * (expected - 1.0 / 20.0))
      
      if (exact_value < value) value <- exact_value
      if (value < 0.0) value <- 0.0
      
    } else if (expected >= 0.95) {
      
      # Use Taylor series for 1/(x * (1-x)) to second order at 0.95.
      value <- n_g * (expected - observed)^2.0 * 400.0 *
        (1.0 / 19.0 + 360.0 / 361.0 * (expected - 19.0 / 20.0))
      
      if (exact_value < value) value <- exact_value
      if (value < 0.0) value <- 0.0
      
    } else {
      # Use exact computation in between.
      value <- exact_value
    }
    
    return(value)
  }
}



.compute_calibration_nam_dagostino <- function(calibration_data) {
  # Nam-D'Agostino and Greenwood-Nam-D'Agostino tests. See Demler et al. 2015.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  observed <- expected <- n_g <- km_var <- nd_group <- gnd_group <- n_groups <- NULL
  nam_dagostino <- greenwood_nam_dagostino <- NULL
  
    # Check for empty calibration tables
  if (is_empty(calibration_data)) return(NULL)
  
  # Make local copy of calibration_data to avoid updating the main copy by
  # reference.
  calibration_data <- data.table::copy(calibration_data)
  
  # Remove NA predictions.
  calibration_data <- calibration_data[is.finite(expected)]
  if (is_empty(calibration_data)) return(NULL)
  
  # Compute test statistic for each group.
  calibration_data[
    ,
    ":="(
      "nd_group" = ..compute_calibration_test_statistic(observed, expected, n_g),
      "gnd_group" = (observed - expected)^2.0 / km_var
    ),
    by = seq_len(nrow(calibration_data))
  ]
  
  # Remove NA values in hm_group
  calibration_data <- calibration_data[is.finite(nd_group) & is.finite(gnd_group)]
  if (is_empty(calibration_data)) return(NULL)
  
  # Compute test statistic for each time point
  gof_table <- calibration_data[
    ,
    list(
      "nam_dagostino" = sum(nd_group, na.rm = TRUE),
      "greenwood_nam_dagostino" = sum(gnd_group, na.rm = TRUE),
      "n_groups" = .N
    )
  ]
  
  # Compute test score for both test tests.
  gof_table <- gof_table[
    n_groups > 1L,
    list(
      "nam_dagostino" = stats::pchisq(
        q = nam_dagostino,
        df = n_groups - 1L,
        lower.tail = FALSE
      ),
      "greenwood_nam_dagostino" = stats::pchisq(
        q = greenwood_nam_dagostino,
        df = n_groups - 1L,
        lower.tail = FALSE
      )
    )
  ]
  
  # Check for an empty goodness-of-fit table.
  if (is_empty(gof_table)) return(NULL)
  
  # Melt so that each test is on a separate row
  gof_table <- data.table::melt(
    gof_table,
    variable.name = "type",
    value.name = "p_value",
    measure.vars = colnames(gof_table),
    variable.factor = FALSE
  )
  
  # Reorder columns
  data.table::setcolorder(
    x = gof_table,
    neworder = c("type", "p_value")
  )
  
  return(gof_table)
}



..compute_calibration_data_interpolated <- function(
    data,
    method = "loess"
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  observed <- expected <- NULL
  
  # Set interpolation range for expected values. This allows for aggregating
  # observed values based on the expected value.
  expected_interpolated <- seq(
    from = 0.000,
    to = 1.000,
    by = 0.005
  )
  
  # Select only unique entries.
  data <- unique(data)
  
  # Select only instances where expected and observed are finite values.
  data <- data[is.finite(expected) & is.finite(observed)]
  
  # Determine the number of unique expected values.
  n_instances <- data.table::uniqueN(data, by = "expected")
  
  # Switch interpolation method for small number of predictions.
  if (method == "loess") {
    if (n_instances >= 4L) {
      degree <- 2L
      
    } else if (n_instances == 3L) {
      degree <- 1L
      
    } else if (n_instances == 2L) {
      method <- "linear"
      
    } else {
      return(list(
        "expected" = NA_real_,
        "observed" = NA_real_
      ))
    }
    
  } else if (method == "linear") {
    
    if (n_instances < 2L) {
      return(list(
        "expected" = NA_real_,
        "observed" = NA_real_
      ))
    }
    
  } else {
    ..error_reached_unreachable_code(paste0(
      "..compute_calibration_data_interpolated: unexpected interpolation ",
      "method encountered: ", method
    ))
  }
  
  if (method == "linear") {
    # Interpolate observed values. Return NA values outside the known expected
    # range.
    observed_interpolated <- suppressWarnings(stats::approx(
      x = data$expected,
      y = data$observed,
      xout = expected_interpolated,
      method = "linear",
      rule = 1L
    )$y)
    
  } else if (method == "loess") {
    # Set up the loess model.
    loess_predictor <- suppressWarnings(stats::loess(
      observed ~ expected,
      data = data,
      span = 1.0,
      degree = degree,
      normalize = FALSE,
      control = stats::loess.control(surface = "direct")
    ))
    
    # Predicted interpolated values.
    observed_interpolated <- suppressWarnings(predict(
      loess_predictor,
      newdata = expected_interpolated
    ))
  }
  
  # Set expected values and the corresponding interpolated observed values.
  return(data.table::data.table(
    "expected" = expected_interpolated,
    "observed" = observed_interpolated
  ))
}



# ..compute_data_element_estimates (familiarDataElementCalibrationData) --------
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElementCalibrationData"),
  function(x, x_list = NULL, ...) {
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Identify the estimation types of the current data elements.
    estimation_type <- sapply(x, function(x) (x@estimation_type))
    
    # Determine the bootstrap_ci_method and the aggregation function
    if (any(estimation_type %in% c("bci", "bootstrap_confidence_interval"))) {
      
      # Point estimates need to be for all expected probability values,
      # which means that we have to compute it from the data.
      x <- .add_point_estimate_from_elements(
        x = x[estimation_type %in% c("bci", "bootstrap_confidence_interval")]
      )
    }
    
    # Check that x is not empty.
    if (is_empty(x)) return(NULL)
    
    return(callNextMethod(x_list = x))
  }
)



# ..compute_data_element_estimates (familiarDataElementCalibrationLinearFit) ----
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElementCalibrationLinearFit"),
  function(x, x_list = NULL, ...) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    p_value <- NULL
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Process data normally.
    y <- callNextMethod()
    
    # Identify the estimation types of the current data elements.
    estimation_type <- sapply(x, function(x) (x@estimation_type))
    
    # Determine the bootstrap_ci_method and the aggregation function
    if (any(estimation_type %in% c("bci", "bootstrap_confidence_interval"))) {
      
      # Obtain data from bootstrapped data.
      data <- x[estimation_type %in% c("bci", "bootstrap_confidence_interval")][[1L]]@data
      
    } else {
      data <- x[[1L]]@data
    }
    
    # Get the grouping column.
    grouping_column <- x[[1L]]@grouping_column
    
    # Compute p-value by grouping column.
    p_value <- data[
      ,
      list("p_value" = harmonic_p_value(p_value)),
      by = c(grouping_column)
    ]
    
    # Merge data into 
    y@data <- merge(
      x = y@data,
      y = p_value,
      by = grouping_column
    )
    
    # Update value column
    y@value_column <- setdiff(
      names(y@data),
      y@grouping_column
    )
    
    return(y)
  }
)


# ..compute_data_element_estimates (familiarDataElementCalibrationGoodnessOfFit) -----
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElementCalibrationGoodnessOfFit"),
  function(x, x_list = NULL, ...) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    p_value <- NULL
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Identify the estimation types of the current data elements.
    estimation_type <- sapply(x, function(x) (x@estimation_type))
    
    # Determine the bootstrap_ci_method and the aggregation function
    if (any(estimation_type %in% c("bci", "bootstrap_confidence_interval"))) {
      
      # Obtain data from bootstrapped data.
      x <- x[estimation_type %in% c("bci", "bootstrap_confidence_interval")][[1L]]
      
    } else {
      x <- x[[1L]]
    }
    
    # Get the grouping column.
    grouping_column <- x@grouping_column
    
    # Since samples overlap, combination methods that assume independence cannot
    # be used (e.g. Fisher's method). Hence we calculate a harmonic mean p value
    # according to Wilson (2019), 10.1073/pnas.1814092116.
    x@data <- x@data[
      ,
      list("p_value" = harmonic_p_value(p_value)),
      by = c(grouping_column)
    ]
    
    # Update value column
    x@value_column <- setdiff(names(x@data), x@grouping_column)
    
    return(x)
  }
)


# ..compute_data_element_estimates (familiarDataElementCalibrationDensity) -----
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElementCalibrationDensity"),
  function(x, x_list = NULL, ...) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    frequency <- expected <- NULL
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Identify the estimation types of the current data elements.
    estimation_type <- sapply(x, function(x) (x@estimation_type))
    
    # Determine the bootstrap_ci_method and the aggregation function
    if (any(estimation_type %in% c("bci", "bootstrap_confidence_interval"))) {
      
      # Obtain data from bootstrapped data.
      x <- x[estimation_type %in% c("bci", "bootstrap_confidence_interval")][[1L]]
      
    } else {
      x <- x[[1L]]
    }
    
    # Get the grouping column.
    grouping_column <- x@grouping_column
    
    # Compute frequency by grouping column.
    x@data <- x@data[
      ,
      list("frequency" = sum(frequency)),
      by = c(grouping_column)
    ]
    
    # Normalise frequency by all grouping columns minus expected.
    grouping_column <- setdiff(grouping_column, "expected")
    
    if (length(grouping_column) > 0L) {
      x@data <- x@data[
        ,
        "frequency" := frequency / sum(frequency),
        by = c(grouping_column)
      ][order(expected)]
      
    } else {
      x@data <- x@data[, "frequency" := frequency / sum(frequency)][order(expected)]
    }
    
    # Update value column
    x@value_column <- setdiff(names(x@data), x@grouping_column)
    
    return(x)
  }
)



# export_calibration_data (generic) --------------------------------------------

#'@title Extract and export calibration and goodness-of-fit tests.
#'
#'@description Extract and export calibration and goodness-of-fit tests for data
#'  in a familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams extract_calibration_data
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Calibration tests are performed based on expected (predicted) and observed
#'  outcomes. For all outcomes, calibration-at-the-large and calibration slopes
#'  are determined. Furthermore, for all but survival outcomes, a repeated,
#'  randomised grouping Hosmer-Lemeshow test is performed. For survival
#'  outcomes, the Nam-D'Agostino and Greenwood-Nam-D'Agostino tests are
#'  performed.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_calibration_data
#'@md
#'@rdname export_calibration_data-methods
setGeneric(
  "export_calibration_data",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_calibration_data")
  }
)



# export_calibration_data (collection) -----------------------------------------

#'@rdname export_calibration_data-methods
setMethod(
  "export_calibration_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Obtain calibration data.
    calibration_data <- .export(
      x = object,
      data_slot = "calibration_data",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "calibration",
      subtype = "data",
      object_class = "familiarDataElementCalibrationData"
    )
    
    # Obtain linear fit data
    linear_fit_data <- .export(
      x = object,
      data_slot = "calibration_data",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "calibration",
      subtype = "at_large",
      object_class = "familiarDataElementCalibrationLinearFit"
    )
    
    # Obtain goodness of fit data
    gof_data <- .export(
      x = object,
      data_slot = "calibration_data",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "calibration",
      subtype = "gof_test",
      object_class = "familiarDataElementCalibrationGoodnessOfFit"
    )
    
    # Obtain density data
    density_data <- .export(
      x = object,
      data_slot = "calibration_data",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "calibration",
      subtype = "density",
      object_class = "familiarDataElementCalibrationDensity"
    )
    
    # Set data list.
    data_list <- list(
      "data" = calibration_data,
      "density" = density_data,
      "linear_test" = linear_fit_data,
      "gof_test" = gof_data
    )
    
    if (!is.null(dir_path)) data_list <- NULL
    if (export_collection) data_list <- c(data_list, list("collection" = object))
    
    return(data_list)
  }
)



# export_calibration_data (general) --------------------------------------------

#'@rdname export_calibration_data-methods
setMethod(
  "export_calibration_data",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "calibration_data",
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_calibration_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
#' @include PredictionTable.R
NULL

# familiarDataElementAUCCurve object definition --------------------------------

setClass(
  "familiarDataElementAUCCurve",
  contains = "familiarDataElement",
  prototype = methods::prototype(
    value_column = "y",
    grouping_column = "x"
  )
)



# extract_auc_data (generic) ---------------------------------------------------

#'@title Internal function to extract area under the ROC curve information.
#'
#'@description Computes the ROC curve from a `familiarEnsemble`.
#''
#'@inheritParams .extract_data
#'
#'@details This function also computes credibility intervals for the ROC curve
#'  for the ensemble model, at the level of `confidence_level`. In the case of
#'  multinomial outcomes, an AUC curve is computed per class in a
#'  one-against-all fashion.
#'
#'  To allow plotting of multiple AUC curves in the same plot and the use of
#'  ensemble models, the AUC curve is evaluated at 0.01 (1-specificity) intervals.
#'
#'@return A list with data.tables for single and ensemble model ROC curve data.
#'@md
#'@keywords internal
setGeneric(
  "extract_auc_data",
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_auc_data")
  }
)



# extract_auc_data (familiarEnsemble) ------------------------------------------
setMethod(
  "extract_auc_data",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    # Extract data for plotting AUC curves.
    
    # AUC data can only be prepared for binomial and multinomial outcomes
    if (!object@outcome_type %in% c("binomial", "multinomial")) return(NULL)
    
    # Message start of auc computations
    logger_message(
      paste0("Computing receiver-operating characteristic curves."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (is.waive(ensemble_method)) ensemble_method <- object@settings$ensemble_method
    if (is.waive(confidence_level)) confidence_level <- object@settings$confidence_level
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- object@settings$bootstrap_ci_method
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "auc_data"
    )
    
    proto_data_element <- .create_extract_auc_data_object(
      object = object,
      ensemble_method = ensemble_method,
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    # Check the level detail.
    detail_level <- .parse_detail_level(
      x = detail_level,
      object = object,
      default = "hybrid",
      data_element = "auc_data"
    )
    
    # Determine whether a single curve is obtained for point estimates.
    # When more than one model exists, these may be averaged for hybrid
    # estimation types.
    is_single_curve <- detail_level == "ensemble" || length(object@model_list) == 1L
    
    # Generate elements to send to dispatch.
    roc_data <- extract_dispatcher(
      FUN = .extract_roc_curve_data,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      aggregate_results = aggregate_results,
      is_single_curve = is_single_curve,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(roc_data)
  }
)



# extract_auc_data (prediction table) ------------------------------------------

setMethod(
  "extract_auc_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    # Extract data for plotting AUC curves.
    
    if (is_empty(object)) return(NULL)
    
    # AUC data can only be prepared for binomial and multinomial outcomes
    if (!is(object, "predictionTableClassification")) {
      ..warning_no_data_extraction_from_prediction_table("AUC curve data")
      
      return(NULL)
    }
    
    # Reference labels should be present.
    if (!.has_reference_data(object)) {
      ..warning_prediction_table_lacks_reference("AUC curve data")
      return(NULL)
    }

    # Message start of auc computations
    logger_message(
      paste0("Computing receiver-operating characteristic curves."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (is.waive(ensemble_method)) {
      ensemble_method <- "median"
      if (methods::.hasSlot(object, "ensemble_method")) ensemble_method <- object@ensemble_method
    }
    
    # Default Values.
    if (is.waive(detail_level)) detail_level <- "ensemble"
    if (is.waive(estimation_type)) estimation_type <- "bootstrap_confidence_interval" 
    if (is.waive(confidence_level)) confidence_level <- 0.95
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- "bc"
    if (is.waive(aggregate_results)) aggregate_results <- TRUE
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "auc_data"
    )
    
    # Copy object to prevent changing the provided object by reference.
    object <- .copy(object)
    
    proto_data_element <- .create_extract_auc_data_object(
      object = object,
      ensemble_method = ensemble_method,
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    # Determine whether a single curve is obtained for point estimates.
    is_single_curve <- TRUE
    
    # Generate elements to send to dispatch.
    roc_data <- extract_dispatcher(
      FUN = .extract_roc_curve_data,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      aggregate_results = aggregate_results,
      is_single_curve = is_single_curve,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(roc_data)
  }
)



.create_extract_auc_data_object <- function(
    object,
    ensemble_method,
    detail_level,
    estimation_type,
    confidence_level,
    bootstrap_ci_method
) {
  # Check confidence_level input argument.
  .check_number_in_valid_range(
    x = confidence_level, 
    var_name = "confidence_level",
    range = c(0.0, 1.0),
    closed = c(FALSE, FALSE)
  )
  
  # Check bootstrap_ci_method argument.
  .check_parameter_value_is_valid(
    x = bootstrap_ci_method, 
    var_name = "bootstrap_ci_method",
    values = .get_available_bootstrap_confidence_interval_methods()
  )
  
  # Check ensemble_method argument.
  .check_parameter_value_is_valid(
    x = ensemble_method,
    var_name = "ensemble_method",
    values = .get_available_ensemble_prediction_methods()
  )
  
  # Check the level detail.
  detail_level <- .parse_detail_level(
    x = detail_level,
    object = object,
    default = "hybrid",
    data_element = "auc_data"
  )
  
  # Check the estimation type.
  estimation_type <- .parse_estimation_type(
    x = estimation_type,
    object = object,
    default = "bootstrap_confidence_interval",
    data_element = "auc_data",
    detail_level = detail_level,
    has_internal_bootstrap = TRUE
  )
  
  # Generate a prototype data element.
  proto_data_element <- new(
    "familiarDataElementAUCCurve",
    detail_level = detail_level,
    estimation_type = estimation_type,
    confidence_level = confidence_level,
    bootstrap_ci_method = bootstrap_ci_method
  )
  
  return(proto_data_element)
}



.extract_roc_curve_data <- function(
    object,
    proto_data_element,
    is_single_curve,
    cl = NULL,
    ...
) {
  
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  proto_data_element <- add_model_name(
    proto_data_element,
    object = object
  )
  
  # Update is_single_curve
  is_single_curve <- proto_data_element@estimation_type == "point" && is_single_curve
  
  if (!object@outcome_type %in% c("binomial", "multinomial")) {
    ..error_outcome_type_not_implemented(object@outcome_type)
  }
  
  return(..extract_roc_curve_data(
    object = object,
    data_element = proto_data_element,
    is_single_curve = is_single_curve,
    cl = cl,
    ...
  ))
}



# ..extract_roc_curve_data (generic) -------------------------------------------
setGeneric(
  "..extract_roc_curve_data",
  function(object, ...) standardGeneric("..extract_roc_curve_data")
)



# ..extract_roc_curve_data (character) -----------------------------------------
setMethod(
  "..extract_roc_curve_data",
  signature(object = "character"),
  function(
    object,
    ...
  ){
    # Ensure that the object is loaded
    object <- load_familiar_object(object)
    
    return(..extract_roc_curve_data(object = object, ...))
  }
)



# ..extract_roc_curve_data (model, ensemble) -----------------------------------
setMethod(
  "..extract_roc_curve_data",
  signature(object = "familiarModelUnion"),
  function(
    object,
    data,
    ensemble_method,
    is_pre_processed,
    ...
  ) {
    
    # Predict class probabilities.
    prediction_data <- .predict(
      object = object,
      data = data,
      ensemble_method = ensemble_method,
      is_pre_processed = is_pre_processed
    )
    
    return(..extract_roc_curve_data(
      object = prediction_data,
      ...
    ))
  }
)



# ..extract_roc_curve_data (classification) ------------------------------------
setMethod(
  "..extract_roc_curve_data",
  signature(object = "predictionTableClassification"),
  function(
    object,
    data_element,
    cl,
    is_single_curve,
    ...
  ) {
    
    # Check if any predictions are valid.
    if (!all_predictions_valid(object)) return(NULL)
    
    # Remove data with missing outcomes.
    object <- filter_missing_outcome(object)
    
    # Check that any prediction data remain.
    if (is_empty(object)) return(NULL)
    
    # Determine class levels
    outcome_class_levels <- get_outcome_class_levels(object)
    
    # Select only one outcome type for binomial outcomes.
    if (object@outcome_type == "binomial") outcome_class_levels <- outcome_class_levels[2L]
    
    # Add positive class as an identifier.
    data_elements <- add_data_element_identifier(
      x = data_element,
      positive_class = outcome_class_levels
    )
    
    # Compute ROC data.
    roc_data <- lapply(
      data_elements,
      .compute_auc_data_categorical,
      data = object,
      cl = cl,
      is_single_curve = is_single_curve,
      ...
    )
    
    return(roc_data)
  }
)



# ..extract_roc_curve_data (NULL) ----------------------------------------------
setMethod(
  "..extract_roc_curve_data",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



.compute_auc_data_categorical <- function(
    data_element,
    data,
    aggregate_results,
    is_single_curve,
    cl = NULL,
    progress_bar = FALSE,
    verbose = FALSE,
    message_indent = 0L,
    ...
) {
  
  # Check if the data has more than 1 row.
  if (get_n_samples(data) <= 1L) return(NULL)
  
  if (length(data_element@identifiers$positive_class) > 0L && progress_bar) {
    logger_message(
      paste0(
        "Computing ROC and Precision-Recall curves for the \"",
        data_element@identifiers$positive_class, "\" class."
      ),
      indent = message_indent,
      verbose = verbose
    )
  }
  
  # Set test probabilities
  threshold_probabilities <- seq(
    from = 0.000,
    to = 1.000,
    by = 0.005
  )
  
  # Add bootstrap data.
  bootstrap_data <- add_data_element_bootstrap(
    x = data_element,
    ...
  )
  
  # Iterate over elements.
  data_elements <- fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = ..compute_auc_data_categorical,
    data_element = bootstrap_data$data_element,
    bootstrap = bootstrap_data$bootstrap,
    bootstrap_seed = bootstrap_data$seed,
    MoreArgs = list(
      "data" = data,
      "x" = threshold_probabilities,
      "is_single_curve" = is_single_curve
    ),
    progress_bar = progress_bar && verbose,
    chopchop = TRUE
  )
  
  # Merge data elements
  data_elements <- merge_data_elements(data_elements)
  
  if (aggregate_results) data_elements <- .compute_data_element_estimates(x = data_elements)
  
  return(data_elements)
}


..compute_auc_data_categorical <- function(
    data_element,
    data,
    x,
    is_single_curve,
    bootstrap,
    bootstrap_seed
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- probability <- ppv <- tpr <- fpr <- is_positive <- NULL
  n_true_positive <- n_false_positive <- sorting_index <- NULL
  
  # Get the positive class.
  positive_class <- data_element@identifiers$positive_class
  
  # Bootstrap the data.
  if (bootstrap) {
    data <- get_bootstrap_sample(
      data = data,
      seed = bootstrap_seed
    )
  }
  
  # Make a local copy
  data <- data.table::copy(.as_data_table(data))
  data.table::setnames(
    x = data,
    old = positive_class,
    new = "probability"
  )
  
  # Determine the number of positive and negative outcomes.
  n_positive <- sum(data$outcome == positive_class)
  n_negative <- sum(data$outcome != positive_class)
  
  # Determine whether the observed outcome was positive.
  data[, "is_positive" := outcome == positive_class]
  
  # Keep only probability and is_positive
  data <- data[, c("probability", "is_positive")]
    
  # Order by inverse probability.
  data <- data[order(-probability)]
  
  # Determine the number of true and false positives.
  data[, ":="(
    "n_true_positive" = cumsum(is_positive),
    "n_false_positive" = cumsum(!is_positive),
    "n_true_negative" = n_negative - cumsum(!is_positive),
    "n_false_negative" = n_positive - cumsum(is_positive)
  )]
  
  # Insert initial values.
  data <- data.table::rbindlist(
    list(
      data.table::data.table(
        "n_true_positive" = 0L,
        "n_false_positive" = 0L,
        "n_true_negative" = n_negative,
        "n_false_negative" = n_positive
      ),
      data
    ),
    use.names = TRUE,
    fill = TRUE
  )
  
  # Select unique data.
  data <- unique(
    data,
    by = c(
      "n_true_positive", "n_false_positive",
      "n_true_negative", "n_false_negative"
    )
  )
  
  # Compute TPR / recall (sensitivity)
  if (n_positive > 0L) {
    data[, "tpr" := n_true_positive / n_positive]
    
  } else {
    data[, "tpr" := 0.0]
  }
  
  # Compute FPR (1-specificity)
  if (max(data$n_false_positive) > 0L) {
    data[, "fpr" := n_false_positive / n_negative]
    
  } else {
    data[, "fpr" := 0.0]
  }
  
  # Compute precision / positive predictive value
  if (max(data$n_true_positive + data$n_false_positive) > 0L) {
    data[, "ppv" := n_true_positive / (n_true_positive + n_false_positive)]
    
    # By convention, the initial value is 1.0.
    data[n_true_positive + n_false_positive == 0L, "ppv" := 1.0]
  } else {
    data[, "ppv" := 0.0]
  }
  
  # Insert a sorting index.
  data[, "sorting_index" := .I]
  
  # Copy the data element as prototype for ROC curve data.
  data_element_roc <- data_element
  
  # Copy relevant columns.
  data_auc_roc <- data[, mget(c("tpr", "fpr"))]
  
  # Add start and end points.
  data_auc_roc <- data.table::rbindlist(
    list(
      data.table::data.table("tpr" = 0.0, "fpr" = 0.0),
      data_auc_roc,
      data.table::data.table("tpr" = 1.0, "fpr" = 1.0)
    ),
    use.names = TRUE
  )
  
  # Select minimum and maximum sensitivity at each fpr.
  data_auc_roc <- unique(data_auc_roc[, list("tpr" = c(min(tpr), max(tpr))), by = "fpr"])
  
  # Select minimum and maximum fpr at each sensitivity to find edge values.
  data_auc_roc <- unique(data_auc_roc[, list("fpr" = c(min(fpr), max(fpr))), by = "tpr"])
  
  # Merge fpr and tpr with data to get the sorting index.
  data_auc_roc <- merge(
    x = data_auc_roc,
    y = data[, mget(c("tpr", "fpr", "sorting_index"))],
    by = c("tpr", "fpr")
  )
  
  # Fill out potentially missing sorting indices.
  data_auc_roc[tpr == 0.0 & fpr == 0.0 & is.na(sorting_index), "sorting_index" := 0L]
  data_auc_roc[tpr == 1.0 & fpr == 1.0 & is.na(sorting_index), "sorting_index" := n_positive + n_negative + 1L]
  
  # Order data by sorting index.
  data_auc_roc <- data_auc_roc[order(sorting_index)]
  
  if (is_single_curve) {
    # Extract data.
    x <- data_auc_roc$fpr
    y_auc_roc <- data_auc_roc$tpr
    
    # Make sure these curves are not aggregated further.
    data_element_roc@is_aggregated <- TRUE
    
  } else {
    # Interpolate at the values in x. This follows Davis J, Goadrich M. The
    # relationship between Precision-Recall and ROC curves. Proceedings of the
    # 23rd international conference on Machine learning. New York, NY, USA:
    # Association for Computing Machinery; 2006. pp. 233–240.
    y_auc_roc <- suppressWarnings(stats::approx(
      x = data_auc_roc$fpr,
      y = data_auc_roc$tpr,
      xout = x,
      yleft = 0.0,
      yright = 1.0,
      method = "linear",
      ties = "ordered"
    )$y)
  }
  
  # Set ROC curve data.
  data_element_roc@data <- data.table::data.table(
    "x" = x,
    "y" = y_auc_roc
  )
  
  # Add curve type as an identifier.
  data_element_roc <- add_data_element_identifier(
    data_element_roc,
    curve_type = "roc"
  )
  
  # Copy the data element as prototype for precisions-recall curve data.
  data_element_pr <- data_element
  
  # Get relevant columns
  data_auc_pr <- data[, mget(c("tpr", "ppv"))]
  
  # Find edge values.
  data_auc_pr <- unique(data_auc_pr[, list("ppv" = c(min(ppv), max(ppv))), by = "tpr"])
  data_auc_pr <- unique(data_auc_pr[, list("tpr" = c(min(tpr), max(tpr))), by = "ppv"])
  
  # Merge ppv and tpr with data to get the sorting index.
  data_auc_pr <- merge(
    x = data_auc_pr,
    y = data[, mget(c("tpr", "ppv", "sorting_index"))],
    by = c("tpr", "ppv")
  )
  
  # Order data by sorting index.
  data_auc_pr <- data_auc_pr[order(sorting_index)]
  
  if (is_single_curve) {
    # Extract required values.
    x <- data_auc_pr$tpr
    y_auc_pr <- data_auc_pr$ppv
    
    # Make sure these curves are not aggregated further.
    data_element_pr@is_aggregated <- TRUE
    
  } else {
    # Interpolate at the values in x.
    y_auc_pr <- suppressWarnings(stats::approx(
      x = data_auc_pr$tpr,
      y = data_auc_pr$ppv,
      xout = x,
      yleft = 1.0,
      yright = n_positive / (n_positive + n_negative),
      method = "linear",
      ties = "ordered"
    )$y)
  }
  
  # Set PR curve data.
  data_element_pr@data <- data.table::data.table(
    "x" = x,
    "y" = y_auc_pr
  )
  
  # Add curve type as an identifier.
  data_element_pr <- add_data_element_identifier(
    data_element_pr,
    curve_type = "pr"
  )
  
  return(c(
    data_element_roc,
    data_element_pr
  ))
}



# export_auc_data (generic) ----------------------------------------------------

#'@title Extract and export ROC and Precision-Recall curves.
#'
#'@description Extract and export ROC and Precision-Recall curves for models in
#'  a familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams extract_auc_data
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  ROC curve data are exported for individual and ensemble models. For ensemble
#'  models, a credibility interval for the ROC curve is determined using
#'  bootstrapping for each metric. In case of multinomial outcomes, ROC-curves
#'  are computed for each class, using a one-against-all approach.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_auc_data
#'@md
#'@rdname export_auc_data-methods
setGeneric(
  "export_auc_data",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_auc_data")
  }
)



# export_auc_data (collection) -------------------------------------------------

#'@rdname export_auc_data-methods
setMethod(
  "export_auc_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(.export(
      x = object,
      data_slot = "auc_data",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "performance",
      subtype = "auc_curves",
      export_collection = export_collection
    ))
  }
)

# export_auc_data (general) ----------------------------------------------------

#'@rdname export_auc_data-methods
setMethod(
  "export_auc_data",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "auc_data",
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_auc_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

# Don't perform any further tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

familiar:::test_export(
  export_function = familiar:::export_prediction_data,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  data_element = "prediction_data",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  detail_level = "ensemble",
  create_novelty_detector = TRUE,
  debug = debug_flag
)

familiar:::test_export(
  export_function = familiar:::export_prediction_data,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  data_element = "prediction_data",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  detail_level = "hybrid",
  estimation_type = "bci",
  confidence_level = 0.80,
  aggregate_results = TRUE,
  create_novelty_detector = TRUE,
  n_models = 100,
  test_config = "normal",
  debug = debug_flag
)

familiar:::test_export(
  export_function = familiar:::export_prediction_data,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  data_element = "prediction_data",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  detail_level = "hybrid",
  estimation_type = "bias_correction",
  confidence_level = 0.80,
  aggregate_results = TRUE,
  create_novelty_detector = TRUE,
  n_models = 20,
  test_config = "normal",
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# extract_predictions (generic) ------------------------------------------------

#'@title Internal function to extract predicted values from models.
#'
#'@description Collects predicted values from models in a `familiarEnsemble`.
#'
#'@inheritParams .extract_data
#'
#'@return A list with single-model and ensemble predictions.
#'@md
#'@keywords internal
setGeneric(
  "extract_predictions",
  function(
    object,
    data,
    cl = NULL,
    is_pre_processed = FALSE,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_predictions")
  }
)



# extract_predictions (familiarEnsemble) ---------------------------------------
setMethod(
  "extract_predictions",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    is_pre_processed = FALSE,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    # Extract predictions from the data using the models in the ensemble. Note:
    # we do not call the predict function on the familiarEnsemble directly as
    # this would cause predict to become highly convoluted, in particular with
    # generating both single and ensemble predictions.
    
    # Message extraction start
    logger_message(
      paste0("Computing ensemble predictions for the dataset."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Load evaluation_times from the object settings attribute, if it is not
    # provided.
    if (is.waive(evaluation_times)) {
      evaluation_times <- object@settings$eval_times
    }
    
    # Check evaluation_times argument
    if (object@outcome_type %in% c("survival")) {
      sapply(
        evaluation_times, 
        .check_number_in_valid_range, 
        var_name = "evaluation_times",
        range = c(0.0, Inf),
        closed = c(FALSE, TRUE)
      )
    }
    
    # Load confidence alpha from object settings attribute if not provided
    # externally.
    if (is.waive(confidence_level)) {
      confidence_level <- object@settings$confidence_level
    }
    
    # Check confidence_level input argument
    .check_number_in_valid_range(
      x = confidence_level,
      var_name = "confidence_level",
      range = c(0.0, 1.0),
      closed = c(FALSE, FALSE)
    )
    
    # Obtain ensemble method from stored settings, if required.
    if (is.waive(ensemble_method)) {
      ensemble_method <- object@settings$ensemble_method
    }
    
    # Check ensemble_method argument
    .check_parameter_value_is_valid(
      x = ensemble_method, 
      var_name = "ensemble_method",
      values = .get_available_ensemble_prediction_methods()
    )
    
    # Check the level detail.
    detail_level <- .parse_detail_level(
      x = detail_level,
      object = object,
      default = "ensemble",
      data_element = "prediction_data"
    )
    
    # Check the estimation type.
    estimation_type <- .parse_estimation_type(
      x = estimation_type,
      object = object,
      default = "point",
      data_element = "prediction_data",
      detail_level = detail_level,
      has_internal_bootstrap = FALSE
    )
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "prediction_data"
    )
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Aggregate data. It does not make sense to keep duplicate rows here.
    data <- aggregate_data(data = data)
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "familiarDataElementPredictionTable",
      detail_level = detail_level,
      confidence_level = confidence_level,
      estimation_type = estimation_type
    )
    
    # Generate elements to send to dispatch.
    performance_data <- extract_dispatcher(
      FUN = .extract_predictions,
      has_internal_bootstrap = FALSE,
      aggregate_results = aggregate_results,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(performance_data)
  }
)



# extract_predictions (prediction table) ---------------------------------------
setMethod(
  "extract_predictions",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    # A prediction table is already a prediction table :)
    return(.complete(object))
  }
)



.extract_predictions <- function(
    object,
    proto_data_element,
    evaluation_times = NULL,
    cl,
    ...
) {
    
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  proto_data_element <- add_model_name(
    proto_data_element,
    object = object
  )
  
  # Add evaluation time as a identifier to the data element.
  if (length(evaluation_times) > 0L && object@outcome_type == "survival") {
    data_elements <- add_data_element_identifier(
      x = proto_data_element,
      evaluation_time = evaluation_times
    )
    
  } else {
    data_elements <- list(proto_data_element)
  }
  
  # Iterate over data elements.
  prediction_data <- lapply(
    data_elements,
    ..extract_predictions,
    object = object,
    cl = cl,
    ...
  )
  
  return(prediction_data)
}



..extract_predictions <- function(
    data_element,
    object,
    data,
    cl = NULL,
    is_pre_processed,
    ensemble_method,
    ...) {
  
  if (object@outcome_type %in% c("survival", "competing_risk")) {
    type <- .get_available_prediction_type_arguments()
    
  } else {
    type <- c("default", "novelty")
  }
  
  data_element_list <- list()
  for (ii in seq_along(type)) {
    data_element_list[[ii]] <- .predict(
      object = object,
      data = data,
      ensemble_method = ensemble_method,
      time = data_element@identifiers$evaluation_time,
      type = type[ii],
      is_pre_processed = is_pre_processed,
      aggregate_results = FALSE
    )
  }
  
  data_element_list <- unlist(data_element_list)
  data_element_list <- data_element_list[!sapply(data_element_list, is_empty)]
  if (is_empty(data_element_list)) return(NULL)
  
  # Inherit from data_element
  data_element_list <- lapply(
    data_element_list,
    function(x, ref) {
      x@ensemble_method <- ref@ensemble_method
      x@identifiers <- c(x@identifiers, ref@identifiers)
      x@detail_level <- ref@detail_level
      x@estimation_type <- ref@estimation_type
      x@confidence_level <- ref@confidence_level
      x@bootstrap_ci_method <- ref@bootstrap_ci_method
      
      return(.merge_slots_into_data(x))
      
    },
    ref = data_element
  )
  
  return(data_element_list)
}



# ..compute_data_element_estimates (familiarDataElementPredictionTable) --------
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElementPredictionTable"),
  function(
    x,
    x_list = NULL,
    object,
    ...
  ) {
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)

    # Identify the estimation types of the current data elements.
    estimation_type <- sapply(x, function(x) (x@estimation_type))
    
    # Check percentiles and confidence level.
    if (
      any(estimation_type %in% c("bci", "bootstrap_confidence_interval")) &&
      is.null(x[[1L]]@percentiles) &&
      is.null(x[[1L]]@confidence_level)
    ) {
      ..error_reached_unreachable_code(paste0(
        "..compute_data_element_estimates: percentiles and confidence_level ",
        "cannot both be NULL."
      ))
    }

    # Determine the bootstrap_ci_method and the aggregation function
    if (any(estimation_type %in% c("bci", "bootstrap_confidence_interval"))) {
      ensemble_method <- ifelse(x[[1L]]@ensemble_method == "median", "percentile", "bc")
      
    } else {
      ensemble_method <- x[[1L]]@ensemble_method
    }
    
    # Collate the data.
    if (any(estimation_type %in% c("bci", "bootstrap_confidence_interval"))) {
      
      # Check that a point estimation is present.
      if (!any(estimation_type %in% c("point"))) {
        # Add point estimate, if not already present.
        x <- c(x, list(.add_point_estimate_from_elements(x)))
        
        # Update estimation type.
        estimation_type <- sapply(x, function(x) (x@estimation_type))
      }
      
      # Select point estimate.
      point_values <- data.table::as.data.table(x[estimation_type == "point"][[1L]]@data)
      point_values[, "estimation_type" := "point"]
      
      # Select bootstrap values.
      bootstrap_values <- data.table::as.data.table(
        x[estimation_type %in% c("bci", "bootstrap_confidence_interval")][[1L]]@data
      )
      bootstrap_values[, "estimation_type" := "bootstrap_confidence_interval"]
      
      # Combine to single table.
      data <- data.table::rbindlist(
        list(point_values, bootstrap_values),
        use.names = TRUE,
        fill = TRUE
      )
      
      # Copy the familiarDataElement.
      y <- x[estimation_type %in% c("bci", "bootstrap_confidence_interval")][[1L]]
      y@data <- NULL
      
    } else {
      # Select values.
      data <- data.table::as.data.table(x[[1L]]@data)
      data[, "estimation_type" := "ensemble"]
      
      # Copy the familiarDataElement.
      y <- x[[1L]]
      y@data <- NULL
    }
    
    # Compute prediction data.
    y@data <- ..compute_ensemble_prediction_estimates(
      x = y,
      data = data,
      ensemble_method = ensemble_method
    )
    
    # Update value column
    y@value_column <- setdiff(names(y@data), y@grouping_column)
    
    return(y)
  }
)


# ..compute_ensemble_prediction_estimates (generic) ----------------------------
setGeneric(
  "..compute_ensemble_prediction_estimates",
  function(x, ...) standardGeneric("..compute_ensemble_prediction_estimates")
)



# export_prediction_data (generic) ---------------------------------------------

#'@title Extract and export predicted values.
#'
#'@description Extract and export the values predicted by single and ensemble
#'  models in a familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams extract_predictions
#'@inheritDotParams as_familiar_collection
#'
#'@details Data, such as model performance and calibration information, is
#'  usually collected from a `familiarCollection` object. However, you can also
#'  provide one or more `familiarData` objects, that will be internally
#'  converted to a `familiarCollection` object. It is also possible to provide a
#'  `familiarEnsemble` or one or more `familiarModel` objects together with the
#'  data from which data is computed prior to export. Paths to the previous
#'  files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Both single and ensemble predictions are exported.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_prediction_data
#'@md
#'@rdname export_prediction_data-methods
setGeneric(
  "export_prediction_data",
  function(
    object,
    dir_path = NULL,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_prediction_data")
  }
)



# export_prediction_data (general) ---------------------------------------------

#'@rdname export_prediction_data-methods
setMethod(
  "export_prediction_data",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "prediction_data"
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_prediction_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarXGBoost --------------------------------------------------------------
setClass(
  "familiarXGBoost",
  contains = "familiarModel",
  slots = list(
    "encoding_reference_table" = "ANY",
    "feature_order" = "character"
  ),
  prototype = list(
    "encoding_reference_table" = NULL,
    "feature_order" = character()
  )
)

setClass("familiarXGBoostLM", contains = "familiarXGBoost")
setClass("familiarXGBoostTree", contains = "familiarXGBoost")
setClass("familiarXGBoostDart", contains = "familiarXGBoost")



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarXGBoost"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- "xgboost"

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarXGBoost"),
  function(object, ...) {
    # Extract learner and outcome_type from the familiarModel object.
    outcome_type <- object@outcome_type

    # Strip booster type from the learner.
    learner <- sub_all_patterns(
      x = object@learner,
      pattern = c("xgboost_lm", "xgboost_tree", "xgboost_dart"),
      replacement = "",
      fixed = TRUE
    )
    
    if (startsWith(learner, "_")) {
      learner <- sub(
        x = learner,
        pattern = "_",
        replacement = "",
        fixed = TRUE
      )
    }

    if (
      outcome_type == "continuous" &&
      learner %in% c("", "gaussian", "gamma", "poisson")
    ) {
      return(TRUE)
      
    } else if (outcome_type == "multinomial" && learner %in% c("", "logistic")) {
      return(TRUE)
      
    } else if (outcome_type == "binomial" && learner %in% c("", "logistic")) {
      return(TRUE)
      
    } else if (outcome_type == "survival" && learner %in% c("", "cox")) {
      return(TRUE)
      
    } else if (outcome_type == "count" && learner %in% c("", "poisson", "gaussian")) {
      ..deprecation_count()
      return(FALSE)
    }
     
    return(FALSE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarXGBoost"),
  function(
    object, 
    data = NULL, 
    user_list = NULL,
    ...
  ) {
    # Initialise list and declare hyperparameter entries
    param <- list()
    param$sign_size <- list()
    param$learn_objective <- list()

    # General parameters
    param$n_boost <- list()
    param$learning_rate <- list()
    param$lambda <- list()
    param$alpha <- list()
    param$sample_weighting <- list()
    param$sample_weighting_beta <- list()

    # Tree specific parameters
    if (is(object, "familiarXGBoostTree") || is(object, "familiarXGBoostDart")) {
      param$tree_depth <- list()
      param$sample_size <- list()
      param$min_child_weight <- list()
      param$gamma <- list()
    }

    # DART-specific parameters
    if (is(object, "familiarXGBoostDart")) {
      param$sample_type <- list()
      param$rate_drop <- list()
    }

    # If data is explicitly NULL, return the list with hyperparameter names
    # only.
    if (is.null(data)) return(param)

    # Extract outcome type.
    outcome_type <- object@outcome_type

    # Determine number of samples.
    n_samples <- get_n_samples(data)

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data)

    # model objective ----------------------------------------------------------

    # xgboost has several families, or rather objectives, that can be used.

    # Read family string by parsing learner.
    fam <- sub_all_patterns(
      x = object@learner,
      pattern = c("xgboost_lm", "xgboost_tree", "xgboost_dart"),
      replacement = "",
      fixed = TRUE
    )
    
    if (fam != "") {
      fam <- sub(
        x = fam, 
        pattern = "_",
        replacement = "", 
        fixed = TRUE
      )
    }

    # Define the objective based on the name of the learner.
    if (fam == "") {
      # No specific objective is provided.
      if (outcome_type == "continuous") {
        learn_objective_default <- c("gaussian", "gamma")
        
      } else if (outcome_type == "binomial") {
        learn_objective_default <- "binomial_logistic"
        
      } else if (outcome_type == "multinomial") {
        learn_objective_default <- "multinomial_logistic"
        
      } else if (outcome_type == "survival") {
        learn_objective_default <- "cox"
      }
      
    } else if (fam == "logistic") {
      # Logistic learner objectives form a collection, and are now uniquely
      # defined according to the type of outcome.
      if (outcome_type == "binomial") {
        learn_objective_default <- "binomial_logistic"
        
      } else if (outcome_type == "multinomial") {
        learn_objective_default <- "multinomial_logistic"
      }
      
    } else {
      # Other objectives are uniquely defined.
      learn_objective_default <- fam
    }

    # Set the learn_objective parameter
    param$learn_objective <- .set_hyperparameter(
      default = learn_objective_default,
      type = "factor",
      range = learn_objective_default,
      randomise = length(learn_objective_default) > 1L
    )

    # number of boosting iterations --------------------------------------------

    # This hyper-parameter is expressed on the log 10 scale. It is called
    # nrounds in xgboost.
    param$n_boost <- .set_hyperparameter(
      default = c(0.0, 1.0, 2.0, 3.0),
      type = "numeric",
      range = c(0.0, 3.0),
      valid_range = c(0.0, Inf),
      randomise = TRUE
    )

    # learning rate ------------------------------------------------------------

    # Learning rate is on a log10 scale from -5 to 0 (0.00001 - 1), and
    # determines how fast the algorithm tries to learn. Lower values typically
    # lead to better models, but take longer to learn. This parameter is called
    # eta by xgboost.
    param$learning_rate <- .set_hyperparameter(
      default = c(-3.0, -2.0, -1.0),
      type = "numeric",
      range = c(-5.0, 0.0),
      valid_range = c(-Inf, 0.0),
      randomise = TRUE
    )

    # L2 regularisation term ---------------------------------------------------

    # The L2 regularisation term lambda lies in the half-open range [0, inf).
    # This term is implemented as a power(10) with a 10^-6 offset.
    param$lambda <- .set_hyperparameter(
      default = c(-6.0, -3.0, -1.0, 1.0, 3.0),
      type = "numeric",
      range = c(-6.0, 3.0),
      valid_range = c(-6.0, Inf),
      randomise = TRUE
    )

    # L1 regularisation term ---------------------------------------------------

    # The L1 regularisation term alpha is implemented as the L2 regularisation
    # term.
    param$alpha <- .set_hyperparameter(
      default = c(-6.0, -3.0, -1.0, 1.0, 3.0),
      type = "numeric",
      range = c(-6.0, 3.0),
      valid_range = c(-6.0, Inf),
      randomise = TRUE
    )

    # sample weighting method --------------------------------------------------
    
    # Class imbalances may lead to learning majority classes. This can be
    # partially mitigated by increasing weight of minority classes.
    param$sample_weighting <- .get_default_sample_weighting_method(outcome_type = object@outcome_type)

    # effective number of samples beta -----------------------------------------
    
    # Specifies the beta parameter for effective number sample weighting method.
    # See Cui et al. (2019).
    param$sample_weighting_beta <- .get_default_sample_weighting_beta(
      method = c(
        param$sample_weighting$init_config,
        user_list$sample_weighting
      ),
      outcome_type = object@outcome_type
    )

    # Parameters for tree-based gradient boosting
    if (is(object, "familiarXGBoostTree") || is(object, "familiarXGBoostDart")) {
      # maximum tree depth -----------------------------------------------------

      # This hyperparameter is only used by tree models. The parameter is called
      # max_depth by xgboost. Larger depths increase the risk of overfitting.
      param$tree_depth <- .set_hyperparameter(
        default = c(1L, 2L, 3L, 7L),
        type = "integer",
        range = c(1L, 10L),
        valid_range = c(1L, Inf),
        randomise = TRUE
      )

      # data subsampling fraction ----------------------------------------------

      # Trees may be grown using only a subset of the data to limit overfitting.
      # The parameter is called subsample in xgboost.
      param$sample_size <- .set_hyperparameter(
        default = c(0.30, 0.50, 0.70, 1.00),
        type = "numeric",
        range = c(2.0 / n_samples, 1.0),
        valid_range = c(0.0, 1.0),
        randomise = TRUE
      )

      # minimum sum of instance weight -----------------------------------------

      # Minimum sum of instance weight (hessian) needed in a child. If the tree
      # partition step results in a leaf node with the sum of instance weight
      # less than min_child_weight, then the building process will give up
      # further partitioning. In linear regression mode, this simply corresponds
      # to minimum number of instances needed to be in each node. The larger,
      # the more conservative the algorithm will be. (source:xgboost
      # documentation)
      #
      # We implement this on a power(10) scale, with -1 offset.
      param$min_child_weight <- .set_hyperparameter(
        default = c(0.0, 1.0, 2.0),
        type = "numeric",
        range = c(0.0, 2.0),
        valid_range = c(0.0, Inf),
        randomise = TRUE
      )

      # minimum splitting error reduction --------------------------------------

      # Minimum error reduction required for splitting. This hyper-parameters is
      # called gamma or min_split_loss in xgboost. We implement it on the
      # power(10) scale, with 10^-6 offset.
      #
      # For continuous type outcomes, this parameter can be a bit tricky due to
      # a wide range in possible scales, and thus in error values. This is
      # resolved by normalising the outcome to the [0, 1] range.
      param$gamma <- .set_hyperparameter(
        default = c(-6.0, -3.0, -1.0, 1.0, 3.0),
        type = "numeric",
        range = c(-6.0, 3.0),
        valid_range = c(-6.0, Inf),
        randomise = TRUE
      )
    }

    # Parameters for dart tree-based gradient boosting
    if (is(object, "familiarXGBoostDart")) {
      ## dart booster sample type -----------------------------------------------

      # Select the sample algorithm used by Dart booster.
      param$sample_type <- .set_hyperparameter(
        default = c("uniform", "weighted"),
        type = "factor",
        range = c("uniform", "weighted"),
        randomise = TRUE
      )

      ## dart booster tree drop rate --------------------------------------------

      # Fraction of previous trees to drop during dropout.
      param$rate_drop <- .set_hyperparameter(
        default = c(0.0, 0.1, 0.3),
        type = "numeric",
        range = c(0.0, 1.0),
        randomise = TRUE
      )
    }

    # Return hyper-parameters
    return(param)
  }
)



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarXGBoost"),
  function(object, type = "default") {
    
    if (object@outcome_type != "survival") return(callNextMethod())

    # The prediction type is a bit more complicated for xgboost methods.
    if (type == "default") {
      if (as.character(object@hyperparameters$learn_objective %in% c("cox"))) {
        return("hazard_ratio")
      }
      
    } else if (type == "survival_probability") {
      return("survival_probability")
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "get_prediction_type,familiarXGBoost: unknown type (", type,
        ") for the current objective (",
        as.character(object@hyperparameters$learn_objective), ")."
      ))
    }
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarXGBoost",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    outcome <- NULL

    # Aggregate repeated measurement data - xgboost does not facilitate repeated
    # measurements.
    data <- aggregate_data(data = data)

    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object, 
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Use effect coding to convert categorical data into encoded data - this is
    # required to deal with factors with missing/new levels between training and
    # test data sets.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )

    # Find feature columns in the data.
    feature_columns <- get_feature_columns(x = encoded_data$encoded_data)

    # Find outcome columns in data table.
    outcome_columns <- get_outcome_columns(x = object)

    # Set y (response variable)
    if (object@outcome_type %in% c("binomial", "multinomial", "continuous")) {
      y <- encoded_data$encoded_data@data[[outcome_columns]]
      
    } else if (object@outcome_type == "survival") {
      y <- survival::Surv(
        encoded_data$encoded_data@data[[outcome_columns[1L]]],
        encoded_data$encoded_data@data[[outcome_columns[2L]]]
      )
    }

    # Set weights.
    weights <- create_instance_weights(
      data = encoded_data$encoded_data,
      method = object@hyperparameters$sample_weighting,
      beta = ..compute_effective_number_of_samples_beta(
        object@hyperparameters$sample_weighting_beta
      ),
      normalisation = "average_one"
    )

    # Create data object
    x <- encoded_data$encoded_data@data[, mget(feature_columns)]

    # Identify the booster to use.
    if (is(object, "familiarXGBoostLM")) {
      booster <- "gblinear"
    } else if (is(object, "familiarXGBoostTree")) {
      booster <- "gbtree"
    } else if (is(object, "familiarXGBoostDart")) {
      booster <- "dart"
    } else {
      ..error_reached_unreachable_code(paste0(
        "..train,familiarXGBoost: could not set booster for object of unknown class: ",
        paste_s(class(object))
      ))
    }

    # Select shared arguments.
    learner_arguments <- list(
      "booster" = booster,
      "nthreads" = 1L,
      "learning_rate" = 10.0^object@hyperparameters$learning_rate,
      "reg_lambda" = 10.0^object@hyperparameters$lambda - 10.0^-6.0,
      "reg_alpha" = 10.0^object@hyperparameters$alpha - 10.0^-6.0,
      "objective" = ..get_distribution_family(object),
      "x" = x,
      "y" = y,
      "weight" = weights,
      "nrounds" = round(10.0^object@hyperparameters$n_boost),
      "verbosity" = 0L
    )

    if (is(object, "familiarXGBoostTree") || is(object, "familiarXGBoostDart")) {
      learner_arguments <- c(
        learner_arguments,
        list(
          "max_depth" = object@hyperparameters$tree_depth,
          "subsample" = object@hyperparameters$sample_size,
          "min_child_weight" = 10.0^object@hyperparameters$min_child_weight - 1.0,
          "min_split_loss" = 10.0^object@hyperparameters$gamma - 10.0^-6.0
        )
      )
    }

    if (is(object, "familiarXGBoostDart")) {
      learner_arguments <- c(
        learner_arguments,
        list(
          "sample_type" = as.character(object@hyperparameters$sample_type),
          "rate_drop" = object@hyperparameters$rate_drop
        )
      )
    }

    # Train the model.
    model <- do.call_with_handlers(
      xgboost::xgboost,
      args = learner_arguments
    )

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) return(callNextMethod(object = object))

    # Add model. Note that we do not use xgboost::xgb.save.raw to save raw model
    # data, since this prevents the model being used for variable importance.
    object@model <- model

    # Add the contrast references to model_list
    object@encoding_reference_table <- encoded_data$reference_table

    # Add feature order
    object@feature_order <- feature_columns

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarXGBoost",
    data = "dataObject"
  ),
  function(object, data, ...) {
    if (object@outcome_type %in% c("continuous", "binomial", "multinomial")) {
      # Turn into a Naive model.
      object <- methods::new("familiarNaiveModel", object)
      
    } else if (object@outcome_type %in% c("survival")) {
      # Turn into a Naive Cox model.
      object <- methods::new("familiarNaiveCoxModel", object)
    }

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarXGBoost",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "default", 
    time = NULL,
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) return(callNextMethod())
    
    # Check if the data is empty.
    if (is_empty(data)) return(callNextMethod())
    
    # Load model through unserialisation.
    if (inherits(object@model, "raw")) {
      object@model <- xgboost::xgb.load.raw(object@model)
    }
    
    # Encode data so that the features are the same as in the training.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )
    
    if (type == "default") {
      # default ----------------------------------------------------------------

      model_predictions <- predict(
        object = object@model,
        newdata = as.matrix(encoded_data$encoded_data@data[, mget(object@feature_order)])
      )
      
      if (object@outcome_type == "binomial") {
        # binomial outcomes ----------------------------------------------------
        
        # Obtain class levels.
        class_levels <- get_outcome_class_levels(x = object)

        # Add class probabilities (glm always gives probability for the second
        # class).
        prediction_list <- list()
        prediction_list[[tail(class_levels, n = 1L)]] <- model_predictions
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type == "multinomial") {
        # multinomial outcomes -------------------------------------------------
        
        # Obtain class levels.
        class_levels <- get_outcome_class_levels(x = object)

        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          if (is.matrix(model_predictions)) {
            prediction_list[[class_levels[ii]]] <- model_predictions[, ii]
            
          } else {
            prediction_list[[class_levels[ii]]] <- model_predictions[ii]
          }
        }
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("continuous")) {
        # Numerical outcomes ---------------------------------------------------

        # Map predictions back to original scale.
        model_predictions <- model_predictions

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = model_predictions,
          type = "regression",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("survival")) {
        # Survival outcomes ----------------------------------------------------
        
        # Store as prediction table. Note that the raw prediction output needs
        # to be recalibrated.
        prediction_table <- as_prediction_table(
          x = model_predictions,
          type = "survival",
          data = data,
          model_object = object
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      return(prediction_table)
    
    } else if (type == "survival_probability" && object@outcome_type == "survival") {
      
      # Only compute for cox-like objectives.
      if (!as.character(object@hyperparameters$learn_objective) %in% c("cox")) {
        return(callNextMethod())
      }
      
      # If time is unset, read the max time stored by the model.
      if (is.null(time)) time <- object@settings$time_max
      
      return(.survival_probability_relative_risk(
        object = object,
        data = data, 
        time = time
      ))
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified method --------------------------------------------------

      # Note that xgboost:::predict.xgb.Booster does not have a type argument.
      return(predict(
        object = object@model,
        newdata = encoded_data$encoded_data@data[, mget(object@feature_order)],
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)


# ..get_prediction_table_type --------------------------------------------------
setMethod(
  "..get_prediction_table_type",
  signature(object = "familiarXGBoost"),
  function(object, type, ...) {
    prediction_table_type <- NULL
    if (object@outcome_type %in% c("survival") && type == "default") {
      prediction_table_type <- "survival"
    } else {
      prediction_table_type <- callNextMethod()
    }
    
    return(prediction_table_type)
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarXGBoost"),
  function(object, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    Weight <- NULL

    # Check if the model has been trained upon retry.
    if (!model_is_trained(object)) return(callNextMethod())

    # Check that required packages are loaded and installed.
    require_package(object, "vimp")

    # Load model through unserialisation. Note that this currently does not
    # produce a model that can be used to infer variable importance.
    if (inherits(object@model, "raw")) {
      object@model <- xgboost::xgb.load.raw(object@model)
    }

    # Use xgboost::xgb.importance function from xgboost to extract a
    # data.table.
    xgboost_score <- tryCatch(
      xgboost::xgb.importance(model = object@model),
      error = identity
    )

    if (inherits(xgboost_score, "error")) return(callNextMethod())
    
    if (is(object, "familiarXGBoostTree") || is(object, "familiarXGBoostDart")) {
      # Process to variable importance table.
      vimp_table <- data.table::data.table(
        "score" = xgboost_score$Gain,
        "name" = xgboost_score$Feature
      )
      
    } else if (is(object, "familiarXGBoostLM")) {
      # Output are linear coefficients, which may be negative. We keep
      # the maximum weight of each feature.
      xgboost_score <- xgboost_score[, list(Weight = max(abs(Weight))), by = "Feature"]

      # Parse score to data.table
      vimp_table <- data.table::data.table(
        "score" = xgboost_score$Weight,
        "name" = xgboost_score$Feature
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "..vimp,familiarXGBoost: could not process vimp for object of unknown class: ",
        paste_s(class(object))
      ))
    }

    # Create variable importance object.
    vimp_object <- methods::new("vimpTable",
      vimp_table = vimp_table,
      encoding_table = object@encoding_reference_table,
      score_aggregation = "max",
      invert = TRUE
    )

    return(vimp_object)
  }
)



# ..set_calibration_info -------------------------------------------------------
setMethod(
  "..set_calibration_info",
  signature(object = "familiarXGBoost"),
  function(object, data) {
    # Check if calibration info already.
    if (has_calibration_info(object)) return(object)

    if (object@outcome_type == "survival") {
      # Determine baseline survival.
      object@calibration_info <- get_baseline_survival(data = data)
      
    } else {
      return(callNextMethod())
    }

    return(object)
  }
)



# ..get_distribution_family ----------------------------------------------------
setMethod(
  "..get_distribution_family",
  signature(object = "familiarXGBoost"),
  function(object) {
    # Obtain family from the hyperparameters.
    objective <- object@hyperparameters$learn_objective

    # Check that the learn_objective hyperparameter exists.
    if (!is.character(objective) && !is.factor(objective)) {
      ..error_reached_unreachable_code(paste0(
        "..get_distribution_family,familiarXGBoost: learn_objective ",
        "hyperparameter was not set."
      ))
    }

    # Load objective for extreme gradient boosting.
    if (objective == "gaussian") {
      boost_objective <- "reg:squarederror"
    } else if (objective == "multinomial_logistic") {
      boost_objective <- "multi:softprob"
    } else if (objective == "binomial_logistic") {
      boost_objective <- "binary:logistic"
    } else if (objective == "poisson") {
      boost_objective <- "count:poisson"
    } else if (objective == "gamma") {
      boost_objective <- "reg:gamma"
    } else if (objective == "cox") {
      boost_objective <- "survival:cox"
    } else {
      ..error_reached_unreachable_code(paste0(
        "..get_distribution_family,familiarXGBoost: unknown learn objective: ",
        objective
      ))
    }

    return(boost_objective)
  }
)



# ..set_recalibration_model ----------------------------------------------------
setMethod(
  "..set_recalibration_model",
  signature(
    object = "familiarXGBoost",
    data = "dataObject"
  ),
  function(object, data, time = NULL) {
    # Recalibration is performed using standard methods
    if (object@outcome_type %in% c("survival")) {
      # Calibrate the models.
      object@calibration_model <- .set_recalibration(
        object = object,
        data = data,
        time = time
      )

      # Return object.
      return(object)
      
    } else {
      return(callNextMethod())
    }
  }
)


# .trim_model ------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarXGBoost"),
  function(object, ...) {
    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE
    
    # Prevent trimming of raw, serialised model information.
    if (inherits(object@model, "raw")) return(object)

    # Add show.
    object <- .capture_show(object)

    # Default method for models that lack a more specific method.
    return(object)
  }
)



.get_available_xgboost_lm_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c(
    "xgboost_lm", "xgboost_lm_logistic", "xgboost_lm_gaussian",
    "xgboost_lm_poisson", "xgboost_lm_gamma", "xgboost_lm_cox"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("xgboost_lm"))
  }
  
  return(learners)
}



.get_available_xgboost_tree_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c(
    "xgboost_tree", "xgboost_tree_logistic", "xgboost_tree_gaussian",
    "xgboost_tree_poisson", "xgboost_tree_gamma", "xgboost_tree_cox"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("xgboost_tree"))
  }
  
  return(learners)
}



.get_available_xgboost_dart_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c(
    "xgboost_dart", "xgboost_dart_logistic", "xgboost_dart_gaussian",
    "xgboost_dart_poisson", "xgboost_dart_gamma", "xgboost_dart_cox"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("xgboost_dart"))
  }
  
  return(learners)
}

# First test if all selectable learners are also available
familiar:::test_all_learners_available(
  learners = familiar:::.get_available_xgboost_dart_learners(show_general = TRUE)
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_all_learners_train_predict_vimp(
  learners = familiar:::.get_available_xgboost_dart_learners(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list(
      "n_boost" = 2,
      "learning_rate" = -1,
      "lambda" = 0.0,
      "alpha" = -6.0,
      "min_child_weight" = 1.04,
      "tree_depth" = 3,
      "sample_size" = 1.0,
      "gamma" = -6.0,
      "sample_type" = "uniform",
      "rate_drop" = 0.0
    ),
    "binomial" = list(
      "n_boost" = 2,
      "learning_rate" = -1,
      "lambda" = 0.0,
      "alpha" = -6.0,
      "min_child_weight" = 1.04,
      "tree_depth" = 3,
      "sample_size" = 1.0,
      "gamma" = -6.0,
      "sample_type" = "uniform",
      "rate_drop" = 0.0
    ),
    "multinomial" = list(
      "n_boost" = 2,
      "learning_rate" = -1,
      "lambda" = 0.0,
      "alpha" = -6.0,
      "min_child_weight" = 1.04,
      "tree_depth" = 3,
      "sample_size" = 1.0,
      "gamma" = -6.0,
      "sample_type" = "uniform",
      "rate_drop" = 0.0
    ),
    "survival" = list(
      "n_boost" = 2,
      "learning_rate" = -1,
      "lambda" = 0.0,
      "alpha" = -6.0,
      "min_child_weight" = 1.04,
      "tree_depth" = 3,
      "sample_size" = 1.0,
      "gamma" = -6.0,
      "sample_type" = "uniform",
      "rate_drop" = 0.0
    )
  )
)

familiar:::test_all_learners_parallel_train_predict_vimp(
  learners = familiar:::.get_available_xgboost_dart_learners(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list(
      "n_boost" = 2,
      "learning_rate" = -1,
      "lambda" = 0.0,
      "alpha" = -6.0,
      "min_child_weight" = 1.04,
      "tree_depth" = 3,
      "sample_size" = 1.0,
      "gamma" = -6.0,
      "sample_type" = "uniform",
      "rate_drop" = 0.0
    ),
    "binomial" = list(
      "n_boost" = 2,
      "learning_rate" = -1,
      "lambda" = 0.0,
      "alpha" = -6.0,
      "min_child_weight" = 1.04,
      "tree_depth" = 3,
      "sample_size" = 1.0,
      "gamma" = -6.0,
      "sample_type" = "uniform",
      "rate_drop" = 0.0
    ),
    "multinomial" = list(
      "n_boost" = 2,
      "learning_rate" = -1,
      "lambda" = 0.0,
      "alpha" = -6.0,
      "min_child_weight" = 1.04,
      "tree_depth" = 3,
      "sample_size" = 1.0,
      "gamma" = -6.0,
      "sample_type" = "uniform",
      "rate_drop" = 0.0
    ),
    "survival" = list(
      "n_boost" = 2,
      "learning_rate" = -1,
      "lambda" = 0.0,
      "alpha" = -6.0,
      "min_child_weight" = 1.04,
      "tree_depth" = 3,
      "sample_size" = 1.0,
      "gamma" = -6.0,
      "sample_type" = "uniform",
      "rate_drop" = 0.0
    )
  )
)


# Continuous outcome tests------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("continuous")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "n_boost" = 2,
    "learning_rate" = -1,
    "lambda" = 0.0,
    "alpha" = -6.0,
    "min_child_weight" = 1.04,
    "tree_depth" = 3,
    "sample_size" = 1.0,
    "gamma" = -6.0,
    "sample_type" = "uniform",
    "rate_drop" = 0.0
  ),
  learner = "xgboost_dart_gaussian"
)

testthat::test_that("Extreme gradient boosting dart tree model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # Check that no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("Extreme gradient boosting dart tree model has variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)
  
  # Expect that the vimp table has six rows.
  testthat::expect_lte(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(vimp_table$name %in% familiar:::get_feature_columns(good_data))
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Binomial tests----------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("binomial")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "n_boost" = 2,
    "learning_rate" = -1,
    "lambda" = 0.0,
    "alpha" = -6.0,
    "min_child_weight" = 1.04,
    "tree_depth" = 3,
    "sample_size" = 1.0,
    "gamma" = -6.0,
    "sample_type" = "uniform",
    "rate_drop" = 0.0
  ),
  learner = "xgboost_dart_logistic"
)

testthat::test_that("Extreme gradient boosting dart tree model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # Check that no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("Extreme gradient boosting dart tree model has variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)
  
  # Expect that the vimp table has six rows.
  testthat::expect_lte(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(vimp_table$name %in% familiar:::get_feature_columns(good_data))
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Multinomial tests-------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("multinomial")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "n_boost" = 2,
    "learning_rate" = -1,
    "lambda" = 0.0,
    "alpha" = -6.0,
    "min_child_weight" = 1.04,
    "tree_depth" = 3,
    "sample_size" = 1.0,
    "gamma" = -6.0,
    "sample_type" = "uniform",
    "rate_drop" = 0.0
  ),
  learner = "xgboost_dart_logistic"
)

testthat::test_that("Extreme gradient boosting dart tree model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # Check that no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})


testthat::test_that("Extreme gradient boosting dart tree model has variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)
  
  # Expect that the vimp table has six rows.
  testthat::expect_lte(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(vimp_table$name %in% familiar:::get_feature_columns(good_data))
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Survival tests----------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("survival")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "n_boost" = 2,
    "learning_rate" = -1,
    "lambda" = 0.0,
    "alpha" = -6.0,
    "min_child_weight" = 1.04,
    "tree_depth" = 3,
    "sample_size" = 1.0,
    "gamma" = -6.0,
    "sample_type" = "uniform",
    "rate_drop" = 0.0
  ),
  time_max = 3.5,
  learner = "xgboost_dart_cox"
)

testthat::test_that("Extreme gradient boosting dart tree model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # Check that no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("Extreme gradient boosting dart tree model has variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)
  
  # Expect that the vimp table has six rows.
  testthat::expect_lte(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(vimp_table$name %in% familiar:::get_feature_columns(good_data))
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


familiar:::test_hyperparameter_optimisation(
  learners = "xgboost_dart",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)


testthat::skip("Skip hyperparameter optimisation, unless manual.")

familiar:::test_hyperparameter_optimisation(
  learners = familiar:::.get_available_xgboost_dart_learners(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# as_familiar_ensemble (generic) -----------------------------------------------

#' @title Conversion to familiarEnsemble object.
#'
#' @description Creates `familiarEnsemble` a object from `familiarModel`
#'   objects.
#'
#' @param object A `familiarEnsemble` object, or one or more `familiarModel`
#'   objects that will be internally converted to a `familiarEnsemble` object.
#'   Paths to such objects can also be provided.
#' @param ... Unused arguments.
#'
#' @return A `familiarEnsemble` object.
#' @exportMethod as_familiar_ensemble
#' @md
#' @rdname as_familiar_ensemble-methods
setGeneric(
  "as_familiar_ensemble",
  function(object, ...) standardGeneric("as_familiar_ensemble")
)



## as_familiar_ensemble (ensemble) ---------------------------------------------

#' @rdname as_familiar_ensemble-methods
setMethod(
  "as_familiar_ensemble",
  signature(object = "familiarEnsemble"),
  function(object, ...) {
    return(object)
  }
)



## as_familiar_ensemble (model) ------------------------------------------------

#' @rdname as_familiar_ensemble-methods
setMethod(
  "as_familiar_ensemble",
  signature(object = "familiarModel"),
  function(object, ...) {
    # A separate familiar model is encapsulated in a list, and then transformed.
    return(do.call(
      as_familiar_ensemble,
      args = list("object" = list(object))
    ))
  }
)


## as_familiar_ensemble (novelty) ----------------------------------------------

#' @rdname as_familiar_ensemble-methods
setMethod(
  "as_familiar_ensemble",
  signature(object = "familiarNoveltyDetector"),
  function(object, ...) {
    # A separate familiar novelty detector is encapsulated in a list, and then
    # transformed.
    return(do.call(
      as_familiar_ensemble,
      args = list("object" = list(object))
    ))
  }
)


## as_familiar_ensemble (list) -------------------------------------------------

#' @rdname as_familiar_ensemble-methods
setMethod(
  "as_familiar_ensemble",
  signature(object = "list"),
  function(object, ...) {
    # Load familiar objects. This does nothing if the list already contains only
    # familiar S4 objects, but will load any files from the path and will check
    # uniqueness of classes.
    object <- load_familiar_object(object = object)

    # Return the object if it contains a single familiarEnsemble.
    if (length(object) == 1L && all(sapply(object, is, "familiarEnsemble"))) {
      return(object[[1L]])
      
    } else if (
      !all(sapply(object, is, "familiarModel")) &&
      !all(sapply(object, is, "familiarNoveltyDetector"))
    ) {
      ..error(paste0(
        "familiarEnsemble objects can only be constructed from familiarModel ",
        "or familiarNoveltyDetector objects."
      ))
    }

    # Generate a placeholder pooling table
    run_table <- data.table::data.table(
      "data_id" = 0L,
      "run_id" = 0L,
      "can_pre_process" = TRUE, 
      "perturbation" = "new_data",
      "perturb_level" = 0L
    )

    vimp_method <- ifelse(
      methods::.hasSlot(object[[1L]], "vimp_method"),
      object[[1L]]@vimp_method,
      "none"
    )
    
    # Generate a skeleton familiarEnsemble
    fam_ensemble <- methods::new(
      "familiarEnsemble",
      model_list = object,
      learner = object[[1L]]@learner,
      vimp_method = vimp_method,
      run_table = run_table
    )

    # Add package version.
    fam_ensemble <- add_package_version(object = fam_ensemble)

    # Complete the ensemble using information provided by the model(s)
    fam_ensemble <- complete_familiar_ensemble(
      object = fam_ensemble
    )

    return(fam_ensemble)
  }
)



## as_familiar_ensemble (character) --------------------------------------------

#' @rdname as_familiar_ensemble-methods
setMethod(
  "as_familiar_ensemble",
  signature(object = "character"),
  function(object, ...) {
    # Interpret character as if it is a path, and pass to the same method for
    # list objects.
    return(do.call(
      as_familiar_ensemble,
      args = list("object" = as.list(object))
    ))
  }
)



## as_familiar_ensemble (general) ----------------------------------------------

#' @rdname as_familiar_ensemble-methods
setMethod(
  "as_familiar_ensemble", signature(object = "ANY"),
  function(object, ...) {
    # There familiar ensembles can only be generated from one of the above
    # functions.
    ..error_cannot_convert_to_familiar_object(
      object = object,
      expected_class = "familiarEnsemble"
    )
  }
)


# as_familiar_data (generic) ---------------------------------------------------

#' @title Conversion to familiarData object.
#'
#' @description Creates `familiarData` a object from `familiarEnsemble` or
#'   `familiarModel` objects.
#'
#' @param object A `familiarData` object, or a `familiarEnsemble` or
#'   `familiarModel` objects that will be internally converted to a
#'   `familiarData` object. Paths to such objects can also be provided.
#'
#' @param name Name of the `familiarData` object. If not set, a name is
#'   automatically generated.
#'
#' @inheritDotParams .extract_data
#'
#' @details The `data` argument is required if `familiarEnsemble` or
#'   `familiarModel` objects are provided.
#'
#' @return A `familiarData` object.
#' @exportMethod as_familiar_data
#' @md
#' @rdname as_familiar_data-methods
setGeneric("as_familiar_data", function(object, ...) standardGeneric("as_familiar_data"))



## as_familiar_data (data) -----------------------------------------------------

#' @rdname as_familiar_data-methods
setMethod(
  "as_familiar_data",
  signature(object = "familiarData"),
  function(object, ...) {
    return(object)
  }
)



## as_familiar_data (ensemble) -------------------------------------------------

#' @rdname as_familiar_data-methods
setMethod(
  "as_familiar_data",
  signature(object = "familiarEnsemble"),
  function(object, name = NULL, ...) {
    # Familiar data
    fam_data <- do.call(
      extract_data,
      args = c(
        list("object" = object),
        list(...)
      )
    )
    
    # Set a placeholder name or a user-provided name for the familiarData
    # object.
    fam_data <- set_object_name(x = fam_data, new = name)

    return(fam_data)
  }
)



## as_familiar_data (prediction table) -----------------------------------------

#' @rdname as_familiar_data-methods
setMethod(
  "as_familiar_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, name = NULL, ...) {
    # Familiar data
    fam_data <- do.call(
      extract_data,
      args = c(
        list("object" = object),
        list(...)
      )
    )
    
    # Set a placeholder name or a user-provided name for the familiarData
    # object.
    fam_data <- set_object_name(x = fam_data, new = name)
    
    return(fam_data)
  }
)


## as_familiar_data (dataObject) -----------------------------------------------

#' @rdname as_familiar_data-methods
setMethod(
  "as_familiar_data",
  signature(object = "dataObject"),
  function(object, name = NULL, ...) {

    # Familiar data
    fam_data <- do.call(
      extract_data,
      args = c(
        list("object" = object),
        list(...)
      )
    )
    
    # Set name of the current batch as name.
    if (is.null(name)) name <- as.character(object@data[[get_id_columns("batch")]][1L])
    
    # Set a name derived from the batch identifier or a user-provided name for
    # the familiarData object.
    fam_data <- set_object_name(x = fam_data, new = name)
    
    return(fam_data)
  }
)


## as_familiar_data (model) ----------------------------------------------------

#' @rdname as_familiar_data-methods
setMethod(
  "as_familiar_data",
  signature(object = "familiarModel"),
  function(object, ...) {
    # Push to the same method for lists. This creates a familiarEnsemble and
    # then allows for creation of a familiarData object.
    return(do.call(
      as_familiar_data,
      args = c(
        list("object" = list(object)),
        list(...)
      )
    ))
  }
)



## as_familiar_data (list) -----------------------------------------------------

#' @rdname as_familiar_data-methods
setMethod(
  "as_familiar_data",
  signature(object = "list"),
  function(object, ...) {
    # Load familiar objects. This does nothing if the list already contains only
    # familiar S4 objects, but will load any files from the path and will check
    # uniqueness of classes.
    object <- load_familiar_object(object = object)

    # Return the object if it contains a single familiarData object.
    if (length(object) == 1L && all(sapply(object, is, "familiarData"))) {
      return(object[[1L]])
    }

    # Parse prediction table.
    if (all(sapply(object, is, "familiarDataElementPredictionTable"))) {
      return(lapply(object, as_familiar_data, ...))
    }
    
    # Parse dataObject.
    if (all(sapply(object, is, "dataObject"))) {
      # Split by batch-id.
      object <- lapply(object, .split_data_by_batch_id)
      
      # Flatten list.
      object <- unlist(object, recursive = FALSE)
      if (!rlang::is_bare_list(object)) object <- list(object)
      
      return(lapply(object, as_familiar_data, ...))
    }
    
    # Convert familiarModel(s) to familiarEnsemble.
    if (all(sapply(object, is, "familiarModel"))) {
      object <- list(as_familiar_ensemble(object = object))
    }
    
    # Check if a single familiarEnsemble has been supplied or generated.
    if (!all(sapply(object, is, "familiarEnsemble")) || length(object) > 1L) {
      ..error(paste0(
        "A familiarData object can only be constructed from a ",
        "single familiarEnsemble object."
      ))
      
    } else {
      object <- object[[1L]]
    }

    return(do.call(
      as_familiar_data,
      args = c(
        list("object" = object),
        list(...)
      )
    ))
  }
)



## as_familiar_data (character) ------------------------------------------------

#' @rdname as_familiar_data-methods
setMethod(
  "as_familiar_data",
  signature(object = "character"),
  function(object, ...) {
    # Pass to as_familiar_data method for lists to load objects there.
    return(do.call(
      as_familiar_data,
      args = c(
        list("object" = as.list(object)),
        list(...)
      )
    ))
  }
)



# as_familiar_data (general) ---------------------------------------------------

#' @rdname as_familiar_data-methods
setMethod(
  "as_familiar_data",
  signature(object = "ANY"),
  function(object, ...) {
    # There familiar ensembles can only be generated from one of the above
    # functions.
    ..error_cannot_convert_to_familiar_object(
      object = object,
      expected_class = "familiarData"
    )
  }
)



# as_familiar_collection (generic) ---------------------------------------------

#' @title Conversion to familiarCollection object.
#'
#' @description Creates a `familiarCollection` objects from `familiarData`,
#'   `familiarEnsemble` or `familiarModel` objects.
#'
#' @param object `familiarCollection` object, or one or more `familiarData`
#'   objects, that will be internally converted to a `familiarCollection`
#'   object. It is also possible to provide a `familiarEnsemble` or one or more
#'   `familiarModel` objects together with the data from which data is computed
#'   prior to export. Paths to such files can also be provided.
#'   
#'   Additionally, some `familiarData` objects can be created from prediction
#'   tables (`familiarDataElementPredictionTable`). Other `familiarData` objects
#'   can be created from data (`dataObject`, or `data.table`). Please
#'   check *details* for more information.
#' @param familiar_data_names Names of the dataset(s). Only used if the `object`
#'   parameter is one or more `familiarData` objects.
#' @param collection_name Name of the collection.
#'
#' @inheritDotParams .extract_data
#'
#' @details A `data` argument is expected if the `object` argument is a
#'   `familiarEnsemble` object or one or more `familiarModel` objects.
#'
#' @return A `familiarCollection` object.
#' @exportMethod as_familiar_collection
#' @md
#' @rdname as_familiar_collection-methods
setGeneric(
  "as_familiar_collection", 
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    standardGeneric("as_familiar_collection")
  }
)



## as_familiar_collection (collection) -----------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "familiarCollection"),
  function(object, ...) {
    return(object)
  }
)



## as_familiar_collection (data) -----------------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "familiarData"),
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    # Pass to as_familiar_collection for lists to load and process objects
    # there.
    return(do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = list(object),
          "familiar_data_names" = familiar_data_names,
          "collection_name" = collection_name
        ),
        list(...)
      )
    ))
  }
)



## as_familiar_collection (ensemble) -------------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "familiarEnsemble"),
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    # Pass to as_familiar_collection for lists to load and process objects
    # there.
    return(do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = list(object),
          "familiar_data_names" = familiar_data_names,
          "collection_name" = collection_name
        ),
        list(...)
      )
    ))
  }
)



## as_familiar_collection (model) ----------------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "familiarModel"),
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    # Pass to as_familiar_collection for lists to load and process objects
    # there.
    return(do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = list(object),
          "familiar_data_names" = familiar_data_names,
          "collection_name" = collection_name
        ),
        list(...)
      )
    ))
  }
)


## as_familiar_collection (prediction table) -----------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "familiarDataElementPredictionTable"),
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    # Pass to as_familiar_collection for lists to load and process objects
    # there.
    return(do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = list(object),
          "familiar_data_names" = familiar_data_names,
          "collection_name" = collection_name
        ),
        list(...)
      )
    ))
  }
)



## as_familiar_collection (dataObject) -----------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "dataObject"),
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    # Pass to as_familiar_collection
    return(do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = list(object),
          "familiar_data_names" = familiar_data_names,
          "collection_name" = collection_name
        ),
        list(...)
      ))
    ) 
  }
)


## as_familiar_collection (data.table) -----------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "data.table"),
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    dots <- list(...)
    
    # Extract outcome_type and outcome_column to provide overrides.
    outcome_type <- dots$outcome_type
    dots$outcome_type <- NULL
    if (is.null(outcome_type)) outcome_type <- waiver()
    
    outcome_column <- dots$outcome_column
    dots$outcome_column <- NULL
    if (is.null(outcome_column)) outcome_column <- waiver()
    
    if (is.waive(outcome_column) || is.waive(outcome_type)) outcome_type <- "unsupervised"
    
    # Extract .no_features_required to override checks on features in as_data_object.
    .no_features_required <- dots$.no_features_required
    dots$.no_features_required <- NULL
    if (is.null(.no_features_required)) .no_features_required <- FALSE
    
    # Convert to dataObject.
    object <- do.call(
      as_data_object,
      args = c(
        list(
          "data" = object,
          "outcome_column" = outcome_column,
          "outcome_type" = outcome_type,
          ".no_features_required" = .no_features_required
        ),
        dots
      )
    )
    
    # Pass to method for dataObject.
    return(do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "familiar_data_names" = familiar_data_names,
          "collection_name" = collection_name
        ),
        dots
      )
    ))
  }
)


## as_familiar_collection (list) -----------------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "list"),
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    # Load familiar objects. This does nothing if the list already contains only
    # familiar S4 objects, but will load any files from the path and will check
    # uniqueness of classes.
    object <- load_familiar_object(object = object)

    # Return the object if it contains a single familiarCollection.
    if (length(object) == 1L && all(sapply(object, is, class2 = "familiarCollection"))) {
      return(object[[1L]])
      
    } else if (all(sapply(object, is, class2 = "familiarCollection"))) {
      ..error("Only a single familiarCollection can be returned.")
    }

    # Convert to familiarModel(s) to familiarData
    if (all(sapply(object, is, class2 = "familiarModel"))) {
      object <- do.call(
        as_familiar_data,
        args = c(
          list("object" = object),
          list(...)
        )
      )
      
      # Store in list, if required
      if (!is(object, "list")) object <- list(object)
    }

    # Convert familiarEnsemble to familiarData
    if (all(sapply(object, is, class2 = "familiarEnsemble")) && length(object) == 1L) {
      object <- do.call(
        as_familiar_data,
        args = c(
          list("object" = object),
          list(...)
        )
      )
      
      # Store in list, if required.
      if (!is(object, "list")) object <- list(object)
      
    } else if (all(sapply(object, is, class2 = "familiarEnsemble"))) {
      ..error("A familiarData object can only be constructed from a single familiarEnsemble object.")
    }

    # Convert prediction table objects to familiarData.
    if (all(sapply(object, is, class2 = "familiarDataElementPredictionTable"))) {
      object <- do.call(
        as_familiar_data,
        args = c(
          list("object" = object),
          list(...)
        )
      )
      
      # Store in list, if required.
      if (!is(object, "list")) object <- list(object)
    }
    
    # Convert dataObject objects to familiarData.
    if (all(sapply(object, is, class2 = "dataObject"))) {
      object <- do.call(
        as_familiar_data,
        args = c(
          list("object" = object),
          list(...)
        )
      )
      
      # Store in list, if required.
      if (!is(object, "list")) object <- list(object)
    }
    
    # Check if all objects at this moments are familiarData objects.
    if (!all(sapply(object, is, class2 = "familiarData"))) {
      stop("Only familiarData objects can be used to construct a familiarCollection object.")
    }

    # Obtain names of the familiarData objects.
    object_names <- sapply(object, function(fam_data_obj) (fam_data_obj@name))

    # Check if names for the data are externally provided, and obtain them from
    # the familiarData objects otherwise.
    if (is.null(familiar_data_names)) {
      familiar_data_names <- object_names
    }

    # Set data names as a factor.
    if (!is.factor(familiar_data_names)) {
      familiar_data_names <- factor(familiar_data_names, levels = unique(familiar_data_names))
    }
    
    # Get names ordered correctly without duplicates.
    data_set_table <- data.table::data.table(
      name_object = object_names,
      name_used = familiar_data_names
    )
    data_set_table <- unique(data_set_table)

    # Check if the collection has a name
    if (is.null(collection_name)) {
      collection_name <- "collection"
    } else {
      collection_name <- as.character(collection_name)
    }
    
    # Generate data names
    fam_collect <- methods::new("familiarCollection",
      name = collection_name,
      data_sets = sapply(
        object,
        function(fam_data_obj) (fam_data_obj@name)
      ),
      outcome_type = object[[1L]]@outcome_type,
      outcome_info = .aggregate_outcome_info(x = lapply(
        object, 
        function(list_elem) (if (methods::.hasSlot(list_elem, "outcome_info")) return(list_elem@outcome_info))
      )),
      fs_vimp = collect(
        x = object,
        data_slot = "fs_vimp",
        identifiers = c("vimp_method")
      ),
      model_vimp = collect(
        x = object, 
        data_slot = "model_vimp",
        identifiers = c("vimp_method", "learner")
      ),
      permutation_vimp = collect(
        x = object,
        data_slot = "permutation_vimp"
      ),
      hyperparameters = collect(
        x = object,
        data_slot = "hyperparameters",
        identifiers = c("vimp_method", "learner")
      ),
      hyperparameter_data = NULL,
      required_features = unique(unlist(lapply(
        object,
        function(fam_data_obj) (fam_data_obj@required_features)
      ))),
      model_features = unique(unlist(extract_from_slot(
        object_list = object,
        slot_name = "model_features",
        na.rm = TRUE
      ))),
      learner = unique(sapply(
        object, 
        function(fam_data_obj) (fam_data_obj@learner)
      )),
      vimp_method = unique(sapply(
        object,
        function(fam_data_obj) (fam_data_obj@vimp_method)
      )),
      prediction_data = collect(
        x = object,
        data_slot = "prediction_data"
      ),
      confusion_matrix = collect(
        x = object, 
        data_slot = "confusion_matrix"
      ),
      decision_curve_data = collect(
        x = object,
        data_slot = "decision_curve_data"
      ),
      calibration_info = collect(
        x = object,
        data_slot = "calibration_info",
        identifiers = c("vimp_method", "learner")
      ),
      calibration_data = collect(
        x = object,
        data_slot = "calibration_data"
      ),
      model_performance = collect(
        x = object,
        data_slot = "model_performance"
      ),
      km_info = collect(
        x = object, 
        data_slot = "km_info",
        identifiers = c("vimp_method", "learner")
      ),
      km_data = collect(
        x = object,
        data_slot = "km_data"
      ),
      auc_data = collect(
        x = object, 
        data_slot = "auc_data"
      ),
      univariate_analysis = collect(
        x = object,
        data_slot = "univariate_analysis"
      ),
      feature_expressions = collect(
        x = object,
        data_slot = "feature_expressions"
      ),
      feature_similarity = collect(
        x = object, 
        data_slot = "feature_similarity"
      ),
      sample_similarity = collect(
        x = object,
        data_slot = "sample_similarity"
      ),
      ice_data = collect(
        x = object,
        data_slot = "ice_data"
      ),
      shap_data = collect(
        x = object,
        data_slot = "shap_data"
      ),
      project_id = object[[1L]]@project_id
    )

    # Add a package version to the familiarCollection object
    fam_collect <- add_package_version(object = fam_collect)

    # Create labels for the data names for correct ordering of plots etc.
    fam_collect <- set_data_set_names(
      x = fam_collect,
      old = data_set_table$name_object,
      new = as.character(data_set_table$name_used),
      order = levels(data_set_table$name_used)
    )

    return(fam_collect)
  }
)


## as_familiar_collection (character) ------------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "character"),
  function(
    object,
    familiar_data_names = NULL,
    collection_name = NULL,
    ...
  ) {
    # Pass to as_familiar_collection for lists to load and process objects
    # there.
    return(do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = as.list(object),
          "familiar_data_names" = familiar_data_names,
          "collection_name" = collection_name
        ),
        list(...)
      )
    ))
  }
)


## as_familiar_collection (generic) --------------------------------------------

#' @rdname as_familiar_collection-methods
setMethod(
  "as_familiar_collection",
  signature(object = "ANY"),
  function(object, ...) {
    # There familiar ensembles can only be generated from objects defined in the
    # previous methods.
    ..error_cannot_convert_to_familiar_object(
      object = object, 
      expected_class = "familiarCollection"
    )
  }
)


# load_familiar_object (character) ---------------------------------------------
setMethod(
  "load_familiar_object", 
  signature(object = "character"),
  function(object) {
    # Determine if file(s) exist
    existing_files <- sapply(object, file.exists)
    if (!all(existing_files)) {
      ..error(paste0(
        "Not all files could be found: ",
        paste_s(object[!existing_files])
      ))
    }

    # Load object
    fam_object <- lapply(object, readRDS)

    # Check that all objects have the correct class.
    if (!(
      all(sapply(fam_object, is, class2 = "familiarModel")) ||
      all(sapply(fam_object, is, class2 = "familiarNoveltyDetector")) ||
      all(sapply(fam_object, is, class2 = "familiarEnsemble")) ||
      all(sapply(fam_object, is, class2 = "familiarData")) ||
      all(sapply(fam_object, is, class2 = "familiarCollection"))
    )) {
      ..error(paste0(
        "Could not load familiar objects because they are not uniquely ",
        "familiarModel, familiarNoveltyDetector, familiarEnsemble, familiarData or ",
        "familiarCollection objects."
      ))
    }

    # Update the objects for backward compatibility
    fam_object <- lapply(fam_object, update_object)

    # If all the object(s) are familiarEnsemble, check the model list.
    if (all(sapply(fam_object, is, class2 = "familiarEnsemble"))) {
      fam_object <- mapply(
        ..update_model_list, 
        object = fam_object, 
        dir_path = object
      )
    }

    # Unlist if the input is singular.
    if (length(object) == 1L) fam_object <- fam_object[[1L]]

    return(fam_object)
  }
)



# load_familiar_object (list) --------------------------------------------------
setMethod(
  "load_familiar_object",
  signature(object = "list"),
  function(object) {
    # Load all objects in the list.
    fam_object <- lapply(object, load_familiar_object)

    # Check that all objects have the correct class.
    if (!(
      all(sapply(fam_object, is, class2 = "familiarModel")) ||
      all(sapply(fam_object, is, class2 = "familiarNoveltyDetector")) ||
      all(sapply(fam_object, is, class2 = "familiarEnsemble")) ||
      all(sapply(fam_object, is, class2 = "familiarData")) ||
      all(sapply(fam_object, is, class2 = "familiarCollection")) ||
      all(sapply(fam_object, is, class2 = "dataObject")) ||
      all(sapply(fam_object, is, class2 = "familiarDataElementPredictionTable"))
    )) {
      ..error(paste0(
        "Could not load familiar objects because they are not uniquely ",
        "familiarDataElementPredictionTable", "dataObject",
        "familiarModel, familiarNoveltyDetector, familiarEnsemble, familiarData ",
        "or familiarCollection objects. Do not mix objects with different classes."
      ))
    }

    # Update the objects for backward compatibility
    fam_object <- lapply(fam_object, update_object)

    return(fam_object)
  }
)


# load_familiar_object (prediction table) --------------------------------------
setMethod(
  "load_familiar_object",
  signature(object = "familiarDataElementPredictionTable"),
  function(object) {
    return(object)
  }
)



# load_familiar_object (dataObject)
setMethod(
  "load_familiar_object",
  signature(object = "dataObject"),
  function(object) {
    return(object)
  }
)



# load_familiar_object (general) -----------------------------------------------
setMethod(
  "load_familiar_object",
  signature(object = "ANY"),
  function(object) {
    # Return the object if it is a familiar S4 class object that has already
    # been loaded. Else throw an error.

    if (is_any(object, class2 = c(
      "familiarModel", "familiarNoveltyDetector",
      "familiarEnsemble", "familiarData", "familiarCollection"
    ))) {
      # Make sure the S4 object is updated.
      object <- update_object(object = object)

      return(object)
      
    } else {
      ..error(paste0(
        "The loaded object is not a familiar S4 object. Found: ",
        paste_s(class(object))
      ))
    }
  }
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  debug = debug_flag,
  not_available_no_samples = FALSE
)

# The default limit for the number of important features exceeds the number of
# features in the data, which means that no actual feature selection has 
# to take place. To test all path ways, we set the number of important features
# to one here. Because the test routine relies on "none" as variable importance
# method, this triggers ad-hoc feature selection when computing the dataset.
# This will fail if outcome data are absent, mostly absent, or singular.
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  data_element = "shap",
  shap_max_iterations = 10L,
  n_important_features = 1L,
  debug = debug_flag
)


# Swarm plot -------------------------------------------------------------------
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  test_config = "normal",
  debug = debug_flag
)

# With a limit to the number of features for which SHAP values are computed.
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  test_config = "normal",
  n_important_features = 2L,
  debug = debug_flag
)

# With a limit to the number of features shown.
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  test_config = "normal",
  plot_args = list("limit_n_features" = 2L),
  debug = debug_flag
)

# Test with single instance
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  test_config = "single instance",
  debug = debug_flag
)

# Test with absolute values
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  plot_args = list("value_representation" = "abs"),
  test_config = "normal",
  debug = debug_flag
)


# Bar plot ---------------------------------------------------------------------

# By specifying barplot as plot type.
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  plot_args = list("plot_type" = "barplot"),
  test_config = "normal",
  debug = debug_flag
)

# By specifying value_representation as abs_mean.
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  plot_args = list("value_representation" = "abs_mean"),
  test_config = "normal",
  debug = debug_flag
)

# By specifying value_representation as abs_min.
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  plot_args = list("value_representation" = "abs_min"),
  test_config = "normal",
  debug = debug_flag
)

# By specifying value_representation as abs_max.
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  plot_args = list("value_representation" = "abs_max"),
  test_config = "normal",
  debug = debug_flag
)


# Box plot ---------------------------------------------------------------------
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  plot_args = list("plot_type" = "boxplot"),
  test_config = "normal",
  debug = debug_flag
)


# Violin plot ------------------------------------------------------------------
familiar:::test_plots(
  plot_function = familiar::plot_shap_summary,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  plot_args = list("plot_type" = "violinplot"),
  test_config = "normal",
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_calibration_data (generic) ----------------------------------------------

#' @title Plot calibration figures.
#'
#' @description This method creates calibration plots from calibration data
#'  stored in a familiarCollection object. For this figures, the expected
#'  (predicted) values are plotted against the observed values. A
#'  well-calibrated model should be close to the identity line.
#'
#' @param dir_path (*optional*) Path to the directory where created calibration
#'  plots are saved to. Output is saved in the `calibration` subdirectory. If
#'  `NULL` no figures are saved, but are returned instead.
#' @param show_density (*optional*) Show point density in top margin of the
#'  figure. If `color_by` is set, this information will not be shown.
#' @param show_calibration_fit (*optional*) Specifies whether the calibration in
#'  the large and calibration slope are annotated in the plot. If `color_by` is
#'  set, this information will not be shown.
#' @param show_goodness_of_fit (*optional*) Specifies whether a the results of
#'  goodness of fit tests are annotated in the plot. If `color_by` is set, this
#'  information will not be shown.
#' @param density_plot_height (*optional*) Height of the density plot. The height
#'  is 1.5 cm by default. Height is expected to be grid unit (see `grid::unit`),
#'  which also allows for specifying relative heights. Will be ignored if
#'  `show_density` is `FALSE`.
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_calibration_data -object
#'
#' @details This function generates a calibration plot for each model in each
#'  dataset. Any data used for calibration (e.g. baseline survival) is obtained
#'  during model creation.
#'
#'  Available splitting variables are: `vimp_method`, `learner`, `data_set` and
#'  `evaluation_time` (survival analysis only) and `positive_class` (multinomial
#'  endpoints only). By default, separate figures are created for each
#'  combination of `vimp_method` and `learner`, with facetting by `data_set`.
#'
#'  Calibration in survival analysis is performed at set time points so that
#'  survival probabilities can be computed from the model, and compared with
#'  observed survival probabilities. This is done differently depending on the
#'  underlying model. For Cox partial hazards regression models, the base
#'  survival (of the development samples) are used, whereas accelerated failure
#'  time models (e.g. Weibull) and survival random forests can be used to
#'  directly predict survival probabilities at a given time point. For survival
#'  analysis, `evaluation_time` is an additional facet variable (by default).
#'
#'  Calibration for multinomial endpoints is performed in a one-against-all
#'  manner. This yields calibration information for each individual class of the
#'  endpoint. For such endpoints, `positive_class` is an additional facet variable
#'  (by default).
#'
#'  Calibration plots have a density plot in the margin, which shows the density
#'  of the plotted points, ordered by the expected probability or value. For
#'  binomial and multinomial outcomes, the density for positive and negative
#'  classes are shown separately. Note that this information is only provided in
#'  when `color_by` is not used as a splitting variable (i.e. one calibration
#'  plot per facet).
#'
#'  Calibration plots are annotated with the intercept and the slope of a linear
#'  model fitted to the sample points. A well-calibrated model has an intercept
#'  close to 0.0 and a slope of 1.0. Intercept and slope are shown with their
#'  respective 95% confidence intervals. In addition, goodness-of-fit tests may
#'  be shown. For most endpoints these are based on the Hosmer-Lemeshow (HL)
#'  test, but for survival endpoints both the Nam-D'Agostino (ND) and the
#'  Greenwood-Nam-D'Agostino (GND) tests are shown. Note that this information
#'  is only annotated when `color_by` is not used as a splitting variable (i.e.
#'  one calibration plot per facet).
#'
#'  Labelling methods such as `set_risk_group_names` or `set_data_set_names` can
#'  be applied to the `familiarCollection` object to update labels, and order
#'  the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @references 1. Hosmer, D. W., Hosmer, T., Le Cessie, S. & Lemeshow, S. A
#'  comparison of goodness-of-fit tests for the logistic regression model. Stat.
#'  Med. 16, 965–980 (1997).
#'
#'  1. D’Agostino, R. B. & Nam, B.-H. Evaluation of the Performance of Survival
#'  Analysis Models: Discrimination and Calibration Measures. in Handbook of
#'  Statistics vol. 23 1–25 (Elsevier, 2003).
#'
#'  1. Demler, O. V., Paynter, N. P. & Cook, N. R. Tests of calibration and
#'  goodness-of-fit in the survival setting. Stat. Med. 34, 1659–1680 (2015).
#'
#' @exportMethod plot_calibration_data
#' @md
#' @rdname plot_calibration_data-methods
setGeneric(
  "plot_calibration_data",
  function(
      object,
      draw = FALSE,
      dir_path = NULL,
      split_by = NULL,
      color_by = NULL,
      facet_by = NULL,
      facet_wrap_cols = NULL,
      ggtheme = NULL,
      discrete_palette = NULL,
      x_label = waiver(),
      x_label_shared = "column",
      y_label = waiver(),
      y_label_shared = "row",
      legend_label = waiver(),
      plot_title = waiver(),
      plot_sub_title = waiver(),
      caption = NULL,
      x_range = NULL,
      x_n_breaks = 5L,
      x_breaks = NULL,
      y_range = NULL,
      y_n_breaks = 5L,
      y_breaks = NULL,
      conf_int_style = c("ribbon", "step", "none"),
      conf_int_alpha = 0.2,
      show_density = TRUE,
      show_calibration_fit = TRUE,
      show_goodness_of_fit = TRUE,
      density_plot_height = grid::unit(1.0, "cm"),
      width = waiver(),
      height = waiver(),
      units = waiver(),
      export_collection = FALSE,
      ...
  ) {
    standardGeneric("plot_calibration_data")
  }
)



# plot_calibration_data (general) ----------------------------------------------

#' @rdname plot_calibration_data-methods
setMethod(
  "plot_calibration_data", 
  signature(object = "ANY"),
  function(
      object,
      draw = FALSE,
      dir_path = NULL,
      split_by = NULL,
      color_by = NULL,
      facet_by = NULL,
      facet_wrap_cols = NULL,
      ggtheme = NULL,
      discrete_palette = NULL,
      x_label = waiver(),
      x_label_shared = "column",
      y_label = waiver(),
      y_label_shared = "row",
      legend_label = waiver(),
      plot_title = waiver(),
      plot_sub_title = waiver(),
      caption = NULL,
      x_range = NULL,
      x_n_breaks = 5L,
      x_breaks = NULL,
      y_range = NULL,
      y_n_breaks = 5L,
      y_breaks = NULL,
      conf_int_style = c("ribbon", "step", "none"),
      conf_int_alpha = 0.2,
      show_density = TRUE,
      show_calibration_fit = TRUE,
      show_goodness_of_fit = TRUE,
      density_plot_height = grid::unit(1.0, "cm"),
      width = waiver(),
      height = waiver(),
      units = waiver(),
      export_collection = FALSE,
      ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list("object" = object, "data_element" = "calibration_data"),
        list(...)
      )
    )

    return(do.call(
      plot_calibration_data,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "x_label_shared" = x_label_shared,
        "y_label" = y_label,
        "y_label_shared" = y_label_shared,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "show_density" = show_density,
        "show_calibration_fit" = show_calibration_fit,
        "show_goodness_of_fit" = show_goodness_of_fit,
        "density_plot_height" = density_plot_height,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_calibration_data (collection) -------------------------------------------

#' @rdname plot_calibration_data-methods
setMethod(
  "plot_calibration_data",
  signature(object = "familiarCollection"),
  function(
      object,
      draw = FALSE,
      dir_path = NULL,
      split_by = NULL,
      color_by = NULL,
      facet_by = NULL,
      facet_wrap_cols = NULL,
      ggtheme = NULL,
      discrete_palette = NULL,
      x_label = waiver(),
      x_label_shared = "column",
      y_label = waiver(),
      y_label_shared = "row",
      legend_label = waiver(),
      plot_title = waiver(),
      plot_sub_title = waiver(),
      caption = NULL,
      x_range = NULL,
      x_n_breaks = 5L,
      x_breaks = NULL,
      y_range = NULL,
      y_n_breaks = 5L,
      y_breaks = NULL,
      conf_int_style = c("ribbon", "step", "none"),
      conf_int_alpha = 0.2,
      show_density = TRUE,
      show_calibration_fit = TRUE,
      show_goodness_of_fit = TRUE,
      density_plot_height = grid::unit(1.0, "cm"),
      width = waiver(),
      height = waiver(),
      units = waiver(),
      export_collection = FALSE,
      ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    .NATURAL <- NULL

    # Make sure the collection object is updated.
    object <- update_object(object = object)

    # Get input data
    x <- export_calibration_data(
      object = object,
      aggregate_results = TRUE
    )

    # Check that the data are not empty.
    if (is_empty(x$data)) return(NULL)

    # Check that the data are not evaluated at the model level.
    if (all(sapply(x$data, function(x) (x@detail_level == "model")))) {
      ..warning_no_comparison_between_models()
      return(NULL)
    }

    # Get data
    calibration_data <- x$data[[1L]]@data
    density_data <- x$density[[1L]]@data
    linear_test_data <- x$linear_test[[1L]]@data
    gof_test_data <- x$gof_test[[1L]]@data

    # Check that the calibration data are not empty.
    if (is_empty(calibration_data)) return(NULL)

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = TRUE),
      purpose = "to create calibration plots",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Check input arguments ----------------------------------------------------

    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # x_label_shared
    if (!is.waive(x_label_shared)) {
      .check_input_plot_args(x_label_shared = x_label_shared)
    } else {
      x_label_shared <- "column"
    }

    # y_label_shared
    if (!is.waive(y_label_shared)) {
      .check_input_plot_args(y_label_shared = y_label_shared)
    } else {
      y_label_shared <- "row"
    }

    # x_range
    if (is.null(x_range)) x_range <- c(0.0, 1.0)

    # x_breaks
    if (is.null(x_breaks)) {
      # Check input arguments.
      .check_input_plot_args(
        x_range = x_range,
        x_n_breaks = x_n_breaks
      )

      # Create breaks and update x_range
      x_breaks <- labeling::extended(
        m = x_n_breaks,
        dmin = x_range[1L],
        dmax = x_range[2L],
        only.loose = TRUE
      )

      x_range <- c(head(x_breaks, n = 1L), tail(x_breaks, n = 1L))
    }

    # y_range
    if (is.null(y_range)) y_range <- c(0.0, 1.0)

    # y_breaks
    if (is.null(y_breaks)) {
      # Check input arguments.
      .check_input_plot_args(
        y_range = y_range,
        y_n_breaks = y_n_breaks
      )

      # Create breaks and update y_range
      y_breaks <- labeling::extended(
        m = y_n_breaks,
        dmin = y_range[1L],
        dmax = y_range[2L],
        only.loose = TRUE
      )

      y_range <- c(head(y_breaks, n = 1L), tail(y_breaks, n = 1L))
    }

    # x_label
    if (is.waive(x_label)) {
      if (object@outcome_type %in% c("binomial", "multinomial")) {
        x_label <- "expected probability"
      } else if (object@outcome_type %in% c("continuous")) {
        x_label <- "expected value"
      } else if (object@outcome_type %in% c("survival")) {
        x_label <- "expected survival probability"
      } else {
        ..error_no_known_outcome_type(object@outcome_type)
      }
    }

    # y_label
    if (is.waive(y_label)) {
      if (object@outcome_type %in% c("binomial", "multinomial")) {
        y_label <- "observed proportion"
      } else if (object@outcome_type %in% c("continuous")) {
        y_label <- "observed value"
      } else if (object@outcome_type %in% c("survival")) {
        y_label <- "observed survival proportion"
      } else {
        ..error_no_known_outcome_type(object@outcome_type)
      }
    }

    # conf_int_style
    if (length(conf_int_style) > 1L) {
      conf_int_style <- head(conf_int_style, n = 1L)
    }

    # Set the style of the confidence interval to none, in case no confidence
    # interval data is present.
    if (!x$data[[1L]]@estimation_type %in% c("bci", "bootstrap_confidence_interval")) {
      conf_int_style <- "none"
    }

    # Add default splitting variables.
    if (is.null(split_by) & is.null(color_by) & is.null(facet_by)) {
      split_by <- c("vimp_method", "learner")

      # Set faceting variables
      facet_by <- c("data_set")

      # Add "evaluation_time" as a faceting variable in case of survival
      # analysis.
      if (object@outcome_type %in% c("survival")) {
        facet_by <- c(facet_by, "evaluation_time")
        
      } else if (object@outcome_type %in% c("multinomial")) {
        facet_by <- c(facet_by, "positive_class")
      }
    }

    # Obtain available splitting variables. This differ per type of endpoint
    # studied. For survival outcomes, evaluation time is an additional splitting
    # point. For multinomial outcomes, the positive class indicator is an
    # additional splitting variable.
    available_splitting_vars <- c("vimp_method", "learner", "data_set")
    if (object@outcome_type %in% c("survival")) {
      available_splitting_vars <- c(available_splitting_vars, "evaluation_time")
    } else if (object@outcome_type %in% c("multinomial")) {
      available_splitting_vars <- c(available_splitting_vars, "positive_class")
    }

    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = calibration_data,
      split_by = split_by,
      color_by = color_by,
      facet_by = facet_by,
      available = available_splitting_vars
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    color_by <- split_var_list$color_by
    facet_by <- split_var_list$facet_by

    # Create a legend label
    legend_label <- .create_plot_legend_title(
      user_label = legend_label,
      color_by = color_by
    )

    # Update show_density, show_goodness_of_fit and show_calibration_fit
    # variables so that they are FALSE in case color_by is set.
    if (!is.null(color_by)) {
      show_calibration_fit <- FALSE
      show_goodness_of_fit <- FALSE
      show_density <- FALSE
    }

    # Check input for show_* arguments.
    .check_parameter_value_is_valid(
      x = show_calibration_fit,
      var_name = "show_calibration_fit",
      values = c(FALSE, TRUE)
    )
    
    .check_parameter_value_is_valid(
      x = show_goodness_of_fit,
      var_name = "show_goodness_of_fit",
      values = c(FALSE, TRUE)
    )
    
    .check_parameter_value_is_valid(
      x = show_density,
      var_name = "show_density",
      values = c(FALSE, TRUE)
    )

    # Check density_plot_height
    .check_plot_grid_unit(
      x = density_plot_height,
      var_name = "density_plot_height"
    )

    # Check input arguments for validity.
    .check_input_plot_args(
      x_range = x_range,
      y_range = y_range,
      x_breaks = x_breaks,
      y_breaks = y_breaks,
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption
    )

    # Create plots -------------------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data and supporting data.
    if (!is.null(split_by)) {
      data_split <- split(
        unique(calibration_data[, mget(split_by)]),
        by = split_by,
        drop = TRUE
      )
      
    } else {
      data_split <- list(NULL)
    }

    # Store plots to list in case no dir_path is provided
    if (is.null(dir_path)) plot_list <- list()

    # Find grouping columns
    grouping_column <- x$data[[1L]]@grouping_column
    grouping_column <- setdiff(grouping_column, "expected")

    # Iterate over splits
    for (current_split in data_split) {
      # Default values
      calibration_data_split <- NULL
      density_data_split <- NULL
      linear_test_data_split <- NULL
      gof_test_data_split <- NULL

      if (!is.null(current_split)) {
        calibration_data_split <- calibration_data[current_split, on = .NATURAL]

        # Get split data. This prevents issues in case a dataset with
        # only identical samples is present.
        split_data <- unique(calibration_data_split[, mget(grouping_column)])

        if (show_density && !is_empty(density_data)) {
          density_data_split <- density_data[split_data, on = .NATURAL]
        }
        
        if (show_calibration_fit && !is_empty(linear_test_data)) {
          linear_test_data_split <- linear_test_data[split_data, on = .NATURAL]
        }
        
        if (show_goodness_of_fit && !is_empty(gof_test_data)) {
          gof_test_data_split <- gof_test_data[split_data, on = .NATURAL]
        }
        
      } else {
        calibration_data_split <- calibration_data

        if (show_density && !is_empty(density_data)) {
          density_data_split <- density_data
        }
        
        if (show_calibration_fit && !is_empty(linear_test_data)) {
          linear_test_data_split <- linear_test_data
        }
        
        if (show_goodness_of_fit && !is_empty(gof_test_data)) {
          gof_test_data_split <- gof_test_data
        }
      }

      if (is_empty(calibration_data_split)) next

      if (is.waive(plot_title)) plot_title <- "Calibration plot"

      # Declare subtitle components.
      additional_subtitle <- NULL

      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, color_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(calibration_data_split$evaluation_time[1L])
        )
      }

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = current_split
        )
      }

      # Generate plot
      p <- .plot_calibration_plot(
        x = calibration_data_split,
        color_by = color_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        x_label = x_label,
        x_label_shared = x_label_shared,
        y_label = y_label,
        y_label_shared = y_label_shared,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        x_range = x_range,
        x_breaks = x_breaks,
        y_range = y_range,
        y_breaks = y_breaks,
        conf_int_alpha = conf_int_alpha,
        conf_int_style = conf_int_style,
        show_density = show_density,
        show_calibration_fit = show_calibration_fit,
        show_goodness_of_fit = show_goodness_of_fit,
        density_plot_height = density_plot_height,
        linear_test = linear_test_data_split,
        gof_test = gof_test_data_split,
        density = density_data_split,
        outcome_type = object@outcome_type,
        grouping_column = x$linear_test[[1L]]@grouping_column,
        is_point = x$data[[1L]]@estimation_type %in% c("point")
      )

      # Check empty output
      if (is.null(p)) next

      # Draw plot
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_calibration_plot_dimensions(
          x = calibration_data_split,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols,
          show_density = show_density
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "calibration",
              "x" = current_split,
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list and export
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_calibration_plot <- function(
    x,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    x_label_shared,
    y_label,
    y_label_shared,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    y_range,
    y_breaks,
    conf_int_alpha,
    conf_int_style,
    show_density,
    show_calibration_fit,
    show_goodness_of_fit,
    density_plot_height,
    linear_test,
    gof_test,
    density,
    outcome_type,
    grouping_column,
    is_point
) {
  # Split by facet. This generates a list of data splits with faceting
  # information that allows for positioning.
  plot_layout_table <- .get_plot_layout_table(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Used for ordering of composite figures.
  layout_split <- split(
    plot_layout_table,
    by = c("col_id", "row_id"),
    sorted = TRUE
  )

  # Split data into facets. This is done by row.
  data_facet_list <- .split_data_by_plot_facet(
    x = x,
    plot_layout_table = plot_layout_table
  )
  linear_test_facet_list <- .split_data_by_plot_facet(
    x = linear_test,
    plot_layout_table = plot_layout_table
  )
  gof_test_facet_list <- .split_data_by_plot_facet(
    x = gof_test,
    plot_layout_table = plot_layout_table
  )
  density_facet_list <- .split_data_by_plot_facet(
    x = density,
    plot_layout_table = plot_layout_table
  )

  # Placeholders for plots.
  figure_list <- list()
  extracted_element_list <- list()

  # Iterate over facets
  for (ii in names(layout_split)) {
    
    if (is_empty(data_facet_list[[ii]])) next
    
    # Create calibration plot.
    p_calibration <- .create_calibration_plot(
      x = data_facet_list[[ii]],
      color_by = color_by,
      facet_by = facet_by,
      facet_wrap_cols = facet_wrap_cols,
      ggtheme = ggtheme,
      discrete_palette = discrete_palette,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      x_range = x_range,
      x_breaks = x_breaks,
      y_range = y_range,
      y_breaks = y_breaks,
      conf_int_alpha = conf_int_alpha,
      conf_int_style = conf_int_style,
      show_calibration_fit = show_calibration_fit,
      show_goodness_of_fit = show_goodness_of_fit,
      linear_test = linear_test_facet_list[[ii]],
      gof_test = gof_test_facet_list[[ii]],
      outcome_type = outcome_type,
      grouping_column = grouping_column,
      is_point = is_point
    )
    
    # Rename plot elements.
    g_calibration <- .rename_plot_grobs(
      g = .convert_to_grob(p_calibration),
      extension = "main"
    )
    if (!gtable::is.gtable(g_calibration)) next

    if (
      show_density && !is_empty(density_facet_list[[ii]])
    ) {
      # Procedure for normal density plots.
      p_margin <- .create_calibration_density_subplot(
        x = density_facet_list[[ii]],
        ggtheme = ggtheme,
        x_range = x_range,
        x_breaks = x_breaks,
        flip = FALSE,
        plot_height = density_plot_height
      )

      # Extract the panel element from the density plot.
      # Convert to gtable
      g_margin <- .rename_plot_grobs(
        g = .convert_to_grob(p_margin),
        extension = "density"
      )
      
      g_calibration <- .gtable_insert(
        g = g_calibration,
        g_new = .gtable_extract_grob(g_margin, element = "panel-density"),
        where = c("above", "panel-main"),
        grob_name = "panel-density",
        spacer = .get_plot_panel_spacing(ggtheme = ggtheme, axis = "y")
      )
    }
    
    # Attach to figure list.
    figure_list[[paste0(layout_split[[ii]]$row_id, ".", layout_split[[ii]]$col_id)]] <- as_familiar_plot(
      g = g_calibration,
      layout = layout_split[[ii]]
    )
  }

  # Compose the final figure. Magic.
  g <- .compose_figure(
    figure_list = figure_list,
    plot_layout_table = plot_layout_table,
    x_text_shared = x_label_shared,
    x_label_shared = x_label_shared,
    y_text_shared = y_label_shared,
    y_label_shared = y_label_shared,
    facet_wrap_cols = facet_wrap_cols,
    ggtheme = ggtheme
  )
  
  return(g)
}



.create_calibration_plot <- function(
    x,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    y_range,
    y_breaks,
    conf_int_alpha,
    conf_int_style,
    show_calibration_fit,
    show_goodness_of_fit,
    grouping_column,
    linear_test,
    gof_test,
    outcome_type,
    is_point
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  type <- NULL

  # Define formula for casting the linear test data wide. This data will be used
  # to plot linear fits in the plot. The "evaluation_time" column will only
  # appear for survival endpoints, and the "positive_class" column only in
  # binomial/multinomial outcomes, but will only matter for multinomial
  # outcomes.
  cast_formula <- stats::as.formula(paste0(
    paste(setdiff(grouping_column, "type"), collapse = "+"),
    "~type"
  ))
  
  if (!is_empty(linear_test)) {
    # Cast wide on the value of the intercept or slope coefficient.
    linear_fit <- data.table::dcast(
      data = linear_test,
      cast_formula,
      value.var = "value"
    )
    
  } else {
    linear_fit <- NULL
  }


  # Generate a guide table for the data
  x_guide_list <- .create_plot_guide_table(
    x = x,
    color_by = color_by,
    discrete_palette = discrete_palette
  )
  fit_guide_list <- .create_plot_guide_table(
    x = linear_fit,
    color_by = color_by,
    discrete_palette = discrete_palette
  )

  # Extract data
  x <- x_guide_list$data

  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("expected"),
      y = !!sym("observed")
    )
  )

  # Add theme
  p <- p + ggtheme

  # Add identity line
  p <- p + ggplot2::geom_abline(
    slope = 1.0,
    intercept = 0.0,
    colour = "grey80",
    linetype = "dashed"
  )

  # Add fill colors
  if (!is.null(color_by)) {
    if (is_point) {
      # Add scatter for individual data points.
      p <- p + ggplot2::geom_point(
        mapping = ggplot2::aes(colour = !!sym("color_breaks"))
      )
      
    } else {
      # Add line for individual data points.
      p <- p + ggplot2::geom_line(
        mapping = ggplot2::aes(colour = !!sym("color_breaks")),
        linewidth = ..get_plot_theme_linewidth(ggtheme = ggtheme) * 3.0
      )
    }

    # Add fit
    if (!is_empty(fit_guide_list$data)) {
      p <- p + ggplot2::geom_abline(
        data = fit_guide_list$data,
        mapping = ggplot2::aes(
          colour = !!sym("color_breaks"),
          intercept = !!sym("offset"),
          slope = !!sym("slope")
        )
      )
    }

    # Set colour.
    p <- p + ggplot2::scale_colour_manual(
      name = legend_label$guide_color,
      values = x_guide_list$guide_color$color_values,
      breaks = x_guide_list$guide_color$color_breaks,
      drop = FALSE
    )

    p <- p + ggplot2::scale_fill_manual(
      name = legend_label$guide_color,
      values = x_guide_list$guide_color$color_values,
      breaks = x_guide_list$guide_color$color_breaks,
      drop = FALSE
    )
    
  } else {
    if (is_point) {
      p <- p + ggplot2::geom_point()
      
    } else {
      # Add scatter for individual data points.
      p <- p + ggplot2::geom_line(
        linewidth = ..get_plot_theme_linewidth(ggtheme = ggtheme) * 3.0
      )
    }

    # Add fit
    if (!is_empty(fit_guide_list$data)) {
      p <- p + ggplot2::geom_abline(
        data = fit_guide_list$data,
        mapping = ggplot2::aes(
          intercept = !!sym("offset"),
          slope = !!sym("slope")
        )
      )
    }
  }

  # Plot confidence intervals
  if (conf_int_style[1L] != "none") {
    if (conf_int_style[1L] == "step") {
      if (is.null(color_by)) {
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(y = !!sym("ci_low")),
          linetype = "dashed"
        )

        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(y = !!sym("ci_up")),
          linetype = "dashed"
        )
        
      } else {
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_low"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed"
        )

        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_up"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed"
        )
      }

      # Remove linetype from the legend.
      p <- p + ggplot2::scale_linetype(guide = "none")
      
    } else if (conf_int_style[1L] == "ribbon") {
      if (is.null(color_by)) {
        p <- p + ggplot2::geom_ribbon(
          mapping = ggplot2::aes(
            ymin = !!sym("ci_low"),
            ymax = !!sym("ci_up")
          ),
          alpha = conf_int_alpha
        )
        
      } else {
        p <- p + ggplot2::geom_ribbon(
          mapping = ggplot2::aes(
            ymin = !!sym("ci_low"),
            ymax = !!sym("ci_up"),
            fill = !!sym("color_breaks")
          ),
          alpha = conf_int_alpha
        )
      }
    }
  }

  # Set breaks and  limits on the x and y-axis
  p <- p + ggplot2::scale_x_continuous(breaks = x_breaks)
  p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)

  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context"
      )
    }
  }

  # Annotate calibration information.
  if (
    (show_calibration_fit && !is_empty(linear_test)) ||
    (show_goodness_of_fit && !is_empty(gof_test))
  ) {
    # Merge test data
    label <- character(0L)

    if (show_calibration_fit && !is_empty(linear_test)) {
      # Add calibration-in-the-large.
      if (all(c("ci_low", "ci_up") %in% colnames(linear_test))) {
        label <- c(
          label,
          paste0(
            "intercept: ",
            format(round(linear_test[type == "offset"]$value, 2L), nsmall = 2L),
            " (",
            format(round(linear_test[type == "offset"]$ci_low, 2L), nsmall = 2L),
            "\u2013",
            ifelse(round(linear_test[type == "offset"]$ci_up, 2L) < 0.0, " ", ""),
            format(round(linear_test[type == "offset"]$ci_up, 2L), nsmall = 2L),
            ")"
          )
        )

        # Add calibration slope.
        label <- c(
          label,
          paste0(
            "slope: ",
            format(round(linear_test[type == "slope"]$value, 2L), nsmall = 2L),
            " (",
            format(round(linear_test[type == "slope"]$ci_low, 2L), nsmall = 2L),
            "\u2013",
            ifelse(round(linear_test[type == "slope"]$ci_up, 2L) < 0.0, " ", ""),
            format(round(linear_test[type == "slope"]$ci_up, 2L), nsmall = 2L),
            ")"
          )
        )
        
      } else {
        label <- c(
          label,
          paste0(
            "intercept: ",
            format(round(linear_test[type == "offset"]$value, 2L), nsmall = 2L)
          )
        )

        # Add calibration slope.
        label <- c(
          label,
          paste0(
            "slope: ",
            format(round(linear_test[type == "slope"]$value, 2L), nsmall = 2L)
          )
        )
      }
    }

    if (show_goodness_of_fit && !is_empty(gof_test)) {
      # Hosmer-Lemeshow test
      if ("hosmer_lemeshow" %in% c(gof_test$type)) {
        label <- c(
          label,
          paste0(
            "HL-test p: ",
            format(signif(gof_test[type == "hosmer_lemeshow"]$p_value, 2L), nsmall = 2L)
          )
        )
      }

      # Nam-D'Agostino test
      if ("nam_dagostino" %in% c(gof_test$type)) {
        label <- c(
          label,
          paste0(
            "ND-test p: ",
            format(signif(gof_test[type == "nam_dagostino"]$p_value, 2L), nsmall = 2L)
          )
        )
      }

      # Greenwood-Nam-D'Agostino test
      if ("greenwood_nam_dagostino" %in% c(gof_test$type)) {
        label <- c(
          label,
          paste0(
            "GND-test p: ",
            format(signif(gof_test[type == "greenwood_nam_dagostino"]$p_value, 2L), nsmall = 2L)
          )
        )
      }
    }

    # Combine all label elements, and use for annotation
    if (length(label) > 0L) {
      label <- paste(label, collapse = "\n")

      # Obtain default settings.
      text_settings <- .get_plot_geom_text_settings(ggtheme = ggtheme)

      # Show in plot
      p <- p + ggplot2::annotate(
        "text",
        x = x_range[1L],
        y = y_range[2L],
        label = label,
        colour = text_settings$colour,
        family = text_settings$family,
        fontface = text_settings$face,
        size = text_settings$geom_text_size,
        vjust = "inward",
        hjust = "inward"
      )
    }
  }

  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Prevent clipping of confidence intervals.
  p <- p + ggplot2::coord_cartesian(xlim = x_range, ylim = y_range)

  return(p)
}



.create_calibration_density_subplot <- function(
    x,
    ggtheme,
    x_range,
    x_breaks,
    flip = FALSE,
    plot_height
) {
  # Create plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("expected"),
      weight = !!sym("frequency")
    )
  )
  
  p <- p + ggplot2::geom_density(bw = 0.075)
  p <- p + ggplot2::scale_x_continuous(
    breaks = x_breaks,
    limits = x_range
  )

  # Set main theme
  p <- p + ggtheme

  # Remove some theme elements and reduce margins
  p <- p + ggplot2::theme(
    panel.grid = ggplot2::element_blank(),
    panel.background = ggplot2::element_blank(),
    panel.border = ggplot2::element_blank(),
    axis.line = ggplot2::element_blank(),
    axis.ticks.y = ggplot2::element_blank(),
    axis.text.x = ggplot2::element_blank(),
    axis.text.y = ggplot2::element_blank(),
    axis.ticks.x = ggplot2::element_blank(),
    axis.title.x = ggplot2::element_blank(),
    axis.title.y = ggplot2::element_blank()
  )

  if (flip) {
    p <- p + ggplot2::scale_y_reverse()
  }

  # Ensure that panel heights are set to the plot object.
  p$custom_grob <- list("heights" = list("name" = "panel", "height" = plot_height))
  class(p) <- c("familiar_ggplot", class(p))

  return(p)
}



.determine_calibration_plot_dimensions <- function(
    x,
    facet_by,
    facet_wrap_cols,
    show_density
) {
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- ifelse(show_density, 6.0, 5.5)

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))

  return(c(height, width))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL




.get_generic_feature_info <- function(
    data,
    outcome_type,
    descriptor = NULL,
    project_id
) {
  # Initialises feature_info objects

  # Expect that data is a data.table.
  if (is(data, "dataObject")) data <- data@data

  # Identify feature columns
  feature_columns <- get_feature_columns(
    x = data,
    outcome_type = outcome_type
  )

  # Iterate over feature columns and create a list of feature_info objects
  feature_info_list <- lapply(
    feature_columns,
    function(ii, data, descriptor, project_id) {
      # Determine feature type
      if (is.factor(data[[ii]])) {
        feature_type <- "factor"
      } else if (is.numeric(data[[ii]])) {
        feature_type <- "numeric"
      } else {
        feature_type <- "unknown"
      }
      
      # Initialise featureInfo object
      feature_info <- methods::new(
        "featureInfo",
        name = ii,
        set_descriptor = ifelse(is.null(descriptor), NA_character_, as.character(descriptor)),
        feature_type = feature_type
      )
      
      # Set factor levels for future reproducibility
      if (feature_type == "factor") {
        feature_info@levels <- levels(data[[ii]])
        feature_info@ordered <- is.ordered(data[[ii]])
      }
      
      # Mark "unknown" feature types for removal
      if (feature_type == "unknown") {
        feature_info@removed <- TRUE
        feature_info@removed_unknown_type <- TRUE
      }
      
      # Add project_id.
      feature_info@project_id
      
      # Update the familiar version.
      feature_info <- add_package_version(object = feature_info)
      
      return(feature_info)
    },
    data = data,
    descriptor = descriptor,
    project_id = project_id
  )
  
  # Set names in the list of featureInfo objects
  names(feature_info_list) <- feature_columns
  
  return(feature_info_list)
}



add_control_info <- function(
    feature_info_list,
    data_id,
    run_id,
    project_id
) {
  # Control information is added to every feature regardless of "removed'.
  
  # Make sure that both identifiers are integers
  data_id <- as.integer(data_id)
  run_id <- as.integer(run_id)
  
  # Update the feature info list.
  feature_info_list <- lapply(
    feature_info_list,
    function(object, data_id, run_id, project_id) {
      # Add data and run ids.
      object@data_id <- data_id
      object@run_id <- run_id
      object@project_id <- project_id
      
      return(object)
    },
    data_id = data_id,
    run_id = run_id,
    project_id = project_id
  )
  
  return(feature_info_list)
}



add_signature_info <- function(
    feature_info_list,
    signature = NULL
) {
  # Sets the in_signature flag on features in the signature variable.
  
  # Check if there is a signature
  if (is.null(signature)) return(feature_info_list)

  # Set signature status
  upd_list <- lapply(
    signature,
    function(signature_feature, feature_info_list) {
      # Obtain object
      object <- feature_info_list[[signature_feature]]
      
      # Mark signature
      object@in_signature <- TRUE
      
      # Update removed status
      object <- update_removed_status(object = object)
      
      return(object)
    },
    feature_info_list = feature_info_list
  )
  
  # Update the names
  names(upd_list) <- signature
  
  # Copy into the list
  feature_info_list[signature] <- upd_list
  
  return(feature_info_list)
}



add_novelty_info <- function(
    feature_info_list, 
    novelty_features = NULL
) {
  # Sets the in_novelty flag on features in the novelty_features variable.

  # Check if there is a signature
  if (is.null(novelty_features)) return(feature_info_list)

  # Set signature status
  upd_list <- lapply(
    novelty_features,
    function(novelty_feature, feature_info_list) {
      # Obtain object
      object <- feature_info_list[[novelty_feature]]
      
      # Mark signature
      object@in_novelty <- TRUE
      
      # Update removed status
      object <- update_removed_status(object = object)
      
      return(object)
    },
    feature_info_list = feature_info_list
  )
  
  # Update the names
  names(upd_list) <- novelty_features
  
  # Copy into the list
  feature_info_list[novelty_features] <- upd_list
  
  return(feature_info_list)
}



add_fairness_info <- function(
    feature_info_list,
    fairness_features = NULL
) {
  if (is.null(fairness_features)) return(feature_info_list)
  
  # TODO: Implement fairness features.
  browser()
}



add_missing_value_fractions <- function(
    cl = NULL,
    feature_info_list,
    data,
    threshold
) {
  # Add the fraction of missing values for features

  # Identify the feature columns in the data
  feature_columns <- get_feature_columns(x = data)

  # Determine number of missing values per column
  n_valid_val <- fam_sapply(
    cl = cl,
    assign = NULL,
    X = data@data[, mget(feature_columns)],
    FUN = function(data) (return(sum(is_valid_data(data)))),
    progress_bar = FALSE,
    chopchop = TRUE
  )

  # Determine fraction of missing values
  missing_frac <- 1.0 - n_valid_val / nrow(data@data)
  
  upd_list <- lapply(
    seq_len(length(feature_columns)),
    function(ii, feature_columns, feature_info_list, missing_frac, threshold) {
      # Extract the featureInfo object
      object <- feature_info_list[[feature_columns[ii]]]
      
      # Add missing value fraction
      object@fraction_missing <- unname(missing_frac[ii])
      
      # Compare with threshold and set removal status
      if (missing_frac[ii] >= threshold) {
        object@removed_missing_values <- TRUE
        
        # Update removed status
        object <- update_removed_status(object = object)
      }
      
      return(object)
    },
    feature_columns = feature_columns,
    feature_info_list = feature_info_list,
    missing_frac = missing_frac,
    threshold = threshold
  )
  
  # Set names of the updated list
  names(upd_list) <- feature_columns

  # Update the list
  feature_info_list[feature_columns] <- upd_list

  return(feature_info_list)
}



add_required_features <- function(feature_info_list) {
  # Find features that are not removed
  available_features <- get_available_features(feature_info_list = feature_info_list)
  
  # Add required features slot
  upd_list <- lapply(
    available_features,
    function(ii, feature_info_list) {
      # featureInfo object for the current feature
      object <- feature_info_list[[ii]]
      
      # A required feature is the feature itself
      required_features <- object@name
      
      # Required features for imputation
      if (is(object@imputation_parameters, "featureInfoParametersImputation")) {
        required_features <- union(
          required_features, object@imputation_parameters@required_features
        )
      }
      
      # Required features for clustering
      if (is(object@cluster_parameters, "featureInfoParametersCluster")) {
        required_features <- union(
          required_features, object@cluster_parameters@required_features
        )
      }
      
      # Add required features
      object@required_features <- unique(required_features)
      
      return(object)
    },
    feature_info_list = feature_info_list
  )
  
  # Set names of the updated list
  names(upd_list) <- available_features
  
  # Update the list
  feature_info_list[available_features] <- upd_list
  
  return(feature_info_list)
}



# Add feature distribution data
compute_feature_distribution_data <- function(
    cl,
    feature_info_list,
    data
) {
  # Identify the feature columns in the data
  feature_columns <- get_feature_columns(x = data)

  # Compute feature distributions
  updated_feature_info <- fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = .compute_feature_distribution_data,
    object = feature_info_list[feature_columns],
    x = data@data[, mget(feature_columns)],
    progress_bar = FALSE,
    chopchop = TRUE
  )

  if (length(feature_columns) > 0L) {
    feature_info_list[feature_columns] <- updated_feature_info
  }

  return(feature_info_list)
}



.compute_feature_distribution_data <- function(object, x) {
  # Placeholder distribution list
  distr_list <- list()

  if (object@feature_type %in% c("factor")) {
    # Number of samples
    distr_list[["n"]] <- length(x)

    # Number of instances for each class
    data <- data.table::data.table("factor_level" = x)
    distr_list[["frequency"]] <- data[, list("count" = .N), by = "factor_level"]
    
  } else if (object@feature_type %in% c("numeric")) {
    # Number of samples
    distr_list[["n"]] <- length(x)

    if (anyNA(x)) x <- x[!is.na(x)]
    
    # Five-number summary of feature values
    distr_list[["fivenum"]] <- fivenum_summary(x, na.rm = FALSE)

    # Mean value
    distr_list[["mean"]] <- mean(x, na.rm = FALSE)
    
    # Percentile summary of feature values
    distr_list[["pctl"]] <- stats::spline(
      x = (seq_along(x) - 1L) / (length(x) - 1L),
      y = sort(x),
      method = "hyman",
      xout = (seq_len(101L) - 1L) / 100L
    )$y
    
  } else {
    ..error_reached_unreachable_code(
      ".compute_feature_distribution_data: unknown feature type encountered."
    )
  }

  # Add to slot
  object@distribution <- distr_list

  return(object)
}



find_invariant_features <- function(
    cl = NULL,
    feature_info_list,
    data
) {
  # Find features that are invariant. Such features are ill-behaved and should
  # be removed.

  # Identify the feature columns in the data
  feature_columns <- get_feature_columns(x = data)

  # Shorthand.
  singular_features <- fam_sapply(
    cl = cl,
    assign = NULL,
    X = data@data[, mget(feature_columns)],
    FUN = is_singular_data,
    progress_bar = FALSE,
    chopchop = TRUE
  )

  singular_features <- feature_columns[singular_features]

  # Iterate over singular features and mark for removal
  if (length(singular_features) > 0L) {
    upd_list <- lapply(
      singular_features,
      function(ii, feature_info_list) {
        # Get the featureInfo for the current feature
        object <- feature_info_list[[ii]]
        
        # Set flag to TRUE (default FALSE)
        object@removed_no_variance <- TRUE
        
        # Update the removed flag
        object <- update_removed_status(object = object)
        
        return(object)
      },
      feature_info_list = feature_info_list
    )
    
    # Add names to the elements of upd_list
    names(upd_list) <- singular_features
    
    # Update the list
    feature_info_list[singular_features] <- upd_list
  }
  
  return(feature_info_list)
}



find_low_variance_features <- function(
    cl = NULL,
    feature_info_list,
    data,
    settings
) {
  # Determine which features have a very low variance and remove these

  # Suppress NOTES due to non-standard evaluation in data.table
  variance <- NULL

  # Identify the feature columns in the data
  feature_columns <- get_feature_columns(x = data)

  # Determine which columns actually contains numeric data
  is_numeric <- sapply(
    feature_columns,
    function(ii, data) (is.numeric(data@data[[ii]])),
    data = data
  )
  numeric_columns <- feature_columns[is_numeric]

  # Skip if there are no numeric columns
  if (length(numeric_columns) == 0L) return(feature_info_list)

  # Determine variance.
  feature_variances <- fam_sapply(
    cl = cl,
    assign = NULL,
    X = data@data[, mget(numeric_columns)],
    FUN = stats::var,
    progress_bar = FALSE,
    na.rm = TRUE,
    chopchop = TRUE
  )

  # Define a data table containing the variances
  variance_data <- data.table::data.table(
    "name" = numeric_columns,
    "variance" = feature_variances
  )

  # Set missing parameters
  if (is.null(settings$prep$low_var_threshold)) {
    # If unset, potentially include all data, pending
    # low_var_max_feature_set_size.
    settings$prep$low_var_threshold <- -1.0
  }

  if (is.null(settings$prep$low_var_max_feature_set_size)) {
    # If unset, potentially allow all data, pending low_var_threshold.
    settings$prep$low_var_max_feature_set_size <- nrow(variance_data)
  }

  # Determine cutoff variance required to obtain the maximally sized feature set
  if (settings$prep$low_var_max_feature_set_size > nrow(variance_data)) {
    # If the number allowed features in the feature set exceeds the actual
    # number of features, set the mimimum variance threshold to 0.
    min_var_thresh <- 0.0
    
  } else {
    # Else, set the minimum variance threshold to the level that corresponds to
    # maximum feature set size.
    min_var_thresh <- sort(
      variance_data$variance,
      decreasing = TRUE
    )[settings$prep$max_var_feat_set_size]
  }

  # Set the variance threshold
  sel_var_thresh <- max(min_var_thresh, settings$prep$low_var_threshold)

  # Determine the low-variance features that are to be removed
  low_variance_features <- variance_data[variance < sel_var_thresh]$name

  # Set removal status
  if (length(low_variance_features) > 0L) {
    upd_list <- lapply(
      low_variance_features,
      function(ii, feature_info_list) {
        # Get the featureInfo for the current feature
        object <- feature_info_list[[ii]]
        
        # Set flag to TRUE (default FALSE)
        object@removed_low_variance <- TRUE
        
        # Update the removed flag
        object <- update_removed_status(object = object)
        
        return(object)
      },
      feature_info_list = feature_info_list
    )
    
    # Add names to the elements of upd_list
    names(upd_list) <- low_variance_features
    
    # Update the list
    feature_info_list[low_variance_features] <- upd_list
  }
  
  return(feature_info_list)
}



find_non_robust_features <- function(
    cl = NULL,
    feature_info_list,
    data,
    settings
) {
  # Determine which features lack robustness and are to be removed. This is only
  # possible for repeated measurements.

  # Check if repeated measurements are present, otherwise return feature info
  # list as is..
  if (all(data@data$repetition_id == 1L)) return(feature_info_list)

  # Determine which columns contain feature data
  feature_columns <- get_feature_columns(x = data)

  # Determine which columns actually contains numeric data
  is_numeric <- sapply(
    feature_columns,
    function(ii, data) (is.numeric(data@data[[ii]])),
    data = data
  )
  numeric_columns <- feature_columns[is_numeric]

  # Skip if there are no columns with numeric data.
  if (length(numeric_columns) == 0L) return(feature_info_list)

  # Read several items from settings
  icc_type <- settings$prep$robustness_icc_type
  icc_filter_column <- settings$prep$robustness_threshold_param
  icc_threshold <- settings$prep$robustness_threshold_value

  # Compute ICC values.
  icc_list <- fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = compute_icc,
    x = data@data[, mget(numeric_columns)],
    feature = numeric_columns,
    progress_bar = FALSE,
    MoreArgs = list(
      "id_data" = data@data[, mget(get_id_columns())],
      "type" = icc_type
    ),
    chopchop = TRUE
  )

  # Combine ICC data from list
  icc_table <- data.table::rbindlist(icc_list)

  # Identify the features with low robustness
  low_robustness_features <- icc_table[get(icc_filter_column) < icc_threshold]$feature

  # Set removal flags for features with low robustness
  if (length(low_robustness_features) > 0L) {
    upd_list <- lapply(
      low_robustness_features, 
      function(ii, feature_info_list) {
        # Get the featureInfo for the current feature
        object <- feature_info_list[[ii]]
        
        # Set flag to TRUE (default FALSE)
        object@removed_low_robustness <- TRUE
        
        # Update the removed flag
        object <- update_removed_status(object = object)
        
        return(object)
      },
      feature_info_list = feature_info_list
    )
    
    # Add names to the elements of upd_list
    names(upd_list) <- low_robustness_features
    
    # Update the list
    feature_info_list[low_robustness_features] <- upd_list
  }
  
  return(feature_info_list)
}



find_unimportant_features <- function(
    cl = NULL,
    feature_info_list,
    data,
    settings
) {
  # Find which features are not important for the current endpoint.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  p_full <- p_val <- name <- p_median <- p_sel <- NULL

  # Base calculations on medians/modes for repeated data. Repeated
  # measurements are not independent and may inflate statistics.
  data <- aggregate_data(data = data)

  # Determine feature columns
  feature_columns <- get_feature_columns(x = data)

  # Set p-value to 1 if none is provided
  if (is.null(settings$prep$univar_threshold)) {
    # If NULL, allow potential selection of all features, pending
    # univar_feat_set_size.
    settings$prep$univar_threshold <- 1.0
  }

  if (is.null(settings$prep$univar_feat_set_size)) {
    # If NULL, allow potential selection of all features, pending
    # univar_threshold
    settings$prep$univar_feat_set_size <- length(feature_columns)
  }

  # Generate bootstraps
  n_iter <- 10L
  iter_list <- .create_bootstraps(
    n_iter = n_iter,
    settings = settings,
    data = data@data,
    stratify = TRUE
  )

  # Calculate p-values of the coefficients
  regr_pval <- compute_univariable_p_values(
    cl = cl,
    data_obj = data,
    feature_columns = feature_columns
  )

  # Find and replace non-finite values
  regr_pval[!is.finite(regr_pval)] <- 1.0

  # Add to table.
  regression_data <- data.table::data.table(
    "name" = names(regr_pval),
    "p_full" = regr_pval
  )
  regression_data[!is.finite(p_full), "p_full" := 1.0]

  if (settings$prep$univar_metric == "p_value") {
    # Filter out features with high p-value
    feature_columns <- feature_columns[
      feature_columns %in% regression_data[
        p_full <= settings$prep$univar_threshold
        ,
      ]$name
    ]
  }

  # Initiate storage list
  regression_data_bootstrap <- list()

  # Iterate over bootstraps
  for (ii in 1L:n_iter) {
    # Bootstrap data
    bootstrap_samples <- select_data_from_samples(
      data = data,
      samples = iter_list$train_list[[ii]]
    )

    # Calculate p-values for the current bootstrap
    if (length(feature_columns) > 0L) {
      regr_pval <- compute_univariable_p_values(
        cl = cl, 
        data_obj = bootstrap_samples,
        feature_columns = feature_columns
      )
      
      regression_data_bootstrap[[ii]] <- data.table::data.table(
        "name" = names(regr_pval),
        "p_val" = regr_pval,
        "iter_id" = ii
      )
      
      regression_data_bootstrap[[ii]][!is.finite(p_val), "p_val" := 1.0]
      
      rm(regr_pval)
      
    } else {
      regression_data_bootstrap[[ii]] <- data.table::data.table(
        "name" = character(0L),
        "p_val" = numeric(0L), 
        "iter_id" = numeric(0L)
      )
    }

    rm(bootstrap_samples)
  }

  # Combine into single data table
  regression_data_bootstrap <- data.table::rbindlist(regression_data_bootstrap)

  # Calculate median metric values of the bootstraps
  regression_data_bootstrap <- regression_data_bootstrap[
    ,
    list("p_median" = stats::median(p_val, na.rm = TRUE)),
    by = name
  ]

  # Merge median metric value table and the full table
  regression_data_bootstrap <- merge(
    x = regression_data_bootstrap,
    y = regression_data,
    by = "name",
    all = TRUE
  )

  # Address non-finite entries
  regression_data_bootstrap[!is.finite(p_median), "p_median" := 1.0]

  # Find the worst entries among the bootstraps and the full analysis
  regression_data_bootstrap[
    ,
    ":="("p_sel" = pmax(p_median, p_full)), 
    by = name
  ]

  rm(regression_data)

  # Determine cutoff required to obtain the maximally sized feature set
  if (settings$prep$univar_feat_set_size > nrow(regression_data_bootstrap)) {
    max_thresh <- 1.0
  } else if (settings$prep$univar_metric == "p_value") {
    max_thresh <- sort(regression_data_bootstrap$p_sel)[settings$prep$univar_feat_set_size]
  }

  # Set the actual q threshold (the lower of max_q_thresh and settings$prep$univar_threshold)
  sel_thresh <- min(max_thresh, settings$prep$univar_threshold)

  # Return features which have a p-value above the selected threshold
  if (settings$prep$univar_metric == "p_value") {
    unimportant_features <- regression_data_bootstrap[p_sel > sel_thresh, ]$name
  } else {
    unimportant_features <- character(0L)
  }

  if (length(unimportant_features) > 0L) {
    upd_list <- lapply(
      unimportant_features, 
      function(ii, feature_info_list) {
        # Get the featureInfo for the current feature
        object <- feature_info_list[[ii]]
        
        # Set flag to TRUE (default FALSE)
        object@removed_low_importance <- TRUE
        
        # Update the removed flag
        object <- update_removed_status(object = object)
        
        return(object)
      },
      feature_info_list = feature_info_list
    )
    
    # Add names to the elements of upd_list
    names(upd_list) <- unimportant_features
    
    # Update the list
    feature_info_list[unimportant_features] <- upd_list
  }
  
  return(feature_info_list)
}



get_available_features <- function(
    feature_info_list,
    data_obj = NULL,
    exclude_signature = FALSE,
    exclude_novelty = FALSE
) {
  # Determine the intersect of features a removed slot == FALSE and available
  # columns in dt (if not NULL).

  # Check that any features are available.
  if (length(feature_info_list) == 0L) return(NULL)

  available_list_features <- names(feature_info_list)[sapply(feature_info_list, is_available)]

  if (!is.null(data_obj)) {
    available_data_features <- get_feature_columns(x = data_obj)

    # The set of available features is the intersect of both.
    available_features <- intersect(available_list_features, available_data_features)
    
  } else {
    # The set of available features is equal to available_list_features.
    available_features <- available_list_features
  }

  if (exclude_signature) {
    # Determine the features in the signature
    signature_features <- names(feature_info_list)[sapply(feature_info_list, is_in_signature)]

    # Exclude these from the available features, e.g. for operations that only
    # work on non-signature features.
    available_features <- setdiff(available_features, signature_features)
  }

  if (exclude_novelty) {
    # Determine the features that are specifically novelty features.
    novelty_features <- names(feature_info_list)[sapply(feature_info_list, is_in_novelty)]

    # Exclude these from the available features, e.g. for operations that only
    # work on non-signature features.
    available_features <- setdiff(available_features, novelty_features)
  }

  return(available_features)
}



find_novelty_features <- function(
    model_features = NULL, 
    feature_info_list
) {
  # Find additional features that should be used for novelty detection.
  novelty_features <- unlist(lapply(
    feature_info_list, 
    function(feature_info) {
      # Check if the feature is a novelty feature.
      if (!feature_info@in_novelty) return(NULL)
      
      return(feature_info@name)
    }
  ))

  return(union(model_features, novelty_features))
}



trim_unused_features_from_list <- function(feature_info_list) {
  # Iterate over features to find all required features
  required_features <- unlist(lapply(
    feature_info_list,
    function(feature_info) {
      # Check if the feature was removed.
      if (feature_info@removed) return(NULL)
      
      return(feature_info@required_features)
    }
  ))
  
  # All required features.
  required_features <- unique(required_features)
  
  # All signature features.
  signature_features <- unlist(lapply(
    feature_info_list,
    function(feature_info) {
      # Check if the feature is in the signature.
      if (!feature_info@in_signature) return(NULL)
      
      return(feature_info@name)
    }
  ))
  
  # All novelty features.
  novelty_features <- unlist(lapply(
    feature_info_list,
    function(feature_info) {
      # Check if the feature is a novelty feature.
      if (!feature_info@in_novelty) return(NULL)
      
      return(feature_info@name)
    }
  ))
  
  features_kept <- unique(c(
    required_features,
    signature_features,
    novelty_features
  ))
  
  return(feature_info_list[features_kept])
}



.collect_and_aggregate_feature_info <- function(
    feature, 
    object, 
    model_list, 
    stop_at = "imputation"
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  min <- Q1 <- median <- Q3 <- max <- count <- NULL

  # Find all featureInfo objects for the current feature
  feature_info_list <- lapply(
    model_list, 
    function(fam_model, feature) {
      
      if (is.null(fam_model@feature_info[[feature]])) {
        return(NULL)
      } else {
        return(fam_model@feature_info[[feature]])
      }
    },
    feature = feature
  )
  
  # Remove NULL entries
  feature_info_list[sapply(feature_info_list, is.null)] <- NULL
  
  # Create a skeleton
  feature_info <- methods::new("featureInfo",
    name = feature,
    set_descriptor = feature_info_list[[1L]]@set_descriptor,
    feature_type = feature_info_list[[1L]]@feature_type,
    levels = feature_info_list[[1L]]@levels,
    data_id = as.integer(object@data_id),
    run_id = as.integer(object@run_id),
    in_signature = feature_info_list[[1L]]@in_signature
  )

  # Add package version.
  feature_info <- add_package_version(object = feature_info)

  # Compute average freaction of data missing
  feature_info@fraction_missing <- mean(extract_from_slot(
    object_list = feature_info_list, 
    slot_name = "fraction_missing",
    na.rm = TRUE
  ))

  if (
    !all_empty_slot(
      object_list = feature_info_list, 
      slot_name = "robustness"
    )
  ) {
    
    # Compute average robustness  
    feature_info@robustness <- mean(extract_from_slot(
      object_list = feature_info_list, 
      slot_name = "robustness",
      na.rm = TRUE
    ))
  }

  
  if (
    !all_empty_slot(
      object_list = feature_info_list,
      slot_name = "univariate_importance"
    )
  ) {
    # Compute average univariate importance
    feature_info@univariate_importance <- mean(extract_from_slot(
      object_list = feature_info_list, 
      slot_name = "univariate_importance",
      na.rm = TRUE
    ))
  }

  # Find distribution items
  distribution_items <- names(feature_info_list[[1L]]@distribution)

  if (!is.null(distribution_items)) {
    # Placeholder list
    distr_list <- list()

    # Iterate over items in the distribution list
    for (item in distribution_items) {
      if (grepl(pattern = "fivenum", x = item, fixed = TRUE)) {
        # Aggregate from list
        fivenum_values <- lapply(
          feature_info_list, 
          function(feature_info, item) (feature_info@distribution[[item]]),
          item = item
        )

        # Combine all the data.tables
        fivenum_values <- data.table::rbindlist(fivenum_values)

        # Check for zero-length lists.
        if (is_empty(fivenum_values)) next

        # Summarise
        fivenum_values <- fivenum_values[, list(
          "min" = min(min),
          "Q1" = mean(Q1),
          "median" = mean(median),
          "Q3" = mean(Q3),
          "max" = max(max)
        ), ]

        # Add to list
        distr_list[[item]] <- fivenum_values
        
      } else if (grepl(pattern = "frequency", x = item, fixed = TRUE)) {
        # Aggregate from list
        frequency_values <- lapply(
          feature_info_list, 
          function(feature_info, item) (feature_info@distribution[[item]]),
          item = item
        )

        # Combine all the data.tables
        frequency_values <- data.table::rbindlist(frequency_values)

        if (is_empty(frequency_values)) next

        # Summarise and add to list
        distr_list[[item]] <- frequency_values[, list("count" = mean(count)), by = "factor_level"]
      
      } else if (grepl(pattern = "pctl", x = item, fixed = TRUE)) {
        
        # Aggregate from list
        pctl_values <- unlist(
          lapply(
            feature_info_list, 
            function(feature_info, item) (feature_info@distribution[[item]]),
            item = item
          ),
          use.names = FALSE
        )
        
        # Check for zero-length lists.
        if (is_empty(pctl_values)) next
        
        # Add to list
        distr_list[[item]] <- unique(sort(pctl_values))
        
      } else {
        # Find mean value
        distr_list[[item]] <- mean(extract_from_slot(
          object_list = feature_info_list,
          slot_name = "distribution",
          slot_element = item,
          na.rm = TRUE
        ))
      }
    }

    # Update distribution slot
    feature_info@distribution <- distr_list
  }

  # Determine if instances of the feature were removed.
  instance_mask <- sapply(feature_info_list, is_available)

  # Extract transformation parameter data.
  transformation_parameter_data <- ..collect_and_aggregate_transformation_info(
    feature_info_list = feature_info_list,
    instance_mask = instance_mask,
    feature_name = feature
  )

  # Set the aggregated transformation parameters.
  feature_info@transformation_parameters <- transformation_parameter_data$parameters
  instance_mask <- instance_mask & transformation_parameter_data$instance_mask

  if (stop_at == "transformation") return(feature_info)

  # Extract normalisation parameter data.
  normalisation_parameter_data <- ..collect_and_aggregate_normalisation_info(
    feature_info_list = feature_info_list,
    instance_mask = instance_mask,
    feature_name = feature
  )

  # Set the aggregated normalisation methods.
  feature_info@normalisation_parameters <- normalisation_parameter_data$parameters
  instance_mask <- instance_mask & transformation_parameter_data$instance_mask

  if (stop_at == "normalisation") return(feature_info)

  # Extract batch normalisation parameter data.
  batch_normalisation_parameter_data <- ..collect_and_aggregate_batch_normalisation_info(
    feature_info_list = feature_info_list,
    instance_mask = instance_mask,
    feature_name = feature
  )

  # Update batch normalisation parameter data.
  feature_info@batch_normalisation_parameters <- batch_normalisation_parameter_data$parameters
  instance_mask <- instance_mask & transformation_parameter_data$instance_mask

  if (stop_at == "batch_normalisation") return(feature_info)

  # Extract imputation data
  imputation_parameter_data <- ..collect_and_aggregate_imputation_info(
    feature_info_list = feature_info_list,
    feature_name = feature,
    feature_type = feature_info@feature_type
  )

  # Add to slot.
  feature_info@imputation_parameters <- imputation_parameter_data$parameters

  # Set required features.
  feature_info@required_features <- feature_info@imputation_parameters@required_features

  return(feature_info)
}



.show_simple_feature_info <- function(object, line_end = "") {
  # Determine the feature type.
  if (object@feature_type == "factor") {
    if (object@ordered) {
      feature_type <- "ordinal"
    } else {
      feature_type <- "categorical"
    }
  } else {
    feature_type <- "numeric"
  }

  # Initialise the feature string.
  feature_str <- paste0(object@name, " (", feature_type, ")")

  if (feature_type == "categorical") {
    # Show levels, including which level is the reference.
    classes_str <- object@levels
    classes_str[1L] <- paste0(classes_str[1L], " (reference)")

    feature_str <- paste0(
      feature_str, ", with levels: ",
      paste_s(classes_str)
    )
    
  } else if (feature_type == "ordinal") {
    # Show ordered levels of the ordinal.
    feature_str <- paste0(
      feature_str, ", with levels: ",
      paste0(object@levels, collapse = " < ")
    )
  }

  return(paste0(feature_str, line_end))
}


# show (featureInfo) -----------------------------------------------------------
setMethod(
  "show", signature(object = "featureInfo"),
  function(object) {
    # Make sure the model object is updated.
    object <- update_object(object = object)

    # Create basic feature string.
    feature_str <- .show_simple_feature_info(object = object)

    # Transformation parameters.
    transform_str <- character(0L)

    # Attempt to create an actual descriptor, if meaningful.
    transform_parameters <- object@transformation_parameters
    if (!is.null(transform_parameters)) {
      if (is(transform_parameters, "featureInfoParametersTransformationPowerTransform")) {
        if (transform_parameters@method != "none") {
          transform_str <- paste0(
            "  transformation (",
            transform_parameters@method,
            ") with \u03BB = ",
            power.transform::get_lambda(transform_parameters@transformer),
            ", shift = ",
            power.transform::get_shift(transform_parameters@transformer),
            ", and scale = ",
            power.transform::get_scale(transform_parameters@transformer),
            ".\n"
          )
        }
      }
    }

    # Normalisation parameters
    normalisation_str <- character(0L)

    # Attempt to create an actual descriptor, if meaningful.
    normalisation_parameters <- object@normalisation_parameters
    if (!is.null(normalisation_parameters)) {
      if (is(normalisation_parameters, "featureInfoParametersNormalisationShiftScale")) {
        if (normalisation_parameters@shift != 0.0 || normalisation_parameters@scale != 1.0) {
          normalisation_str <- paste0(
            "  normalisation (",
            normalisation_parameters@method,
            ") with shift = ",
            normalisation_parameters@shift,
            " and scale = ",
            normalisation_parameters@scale,
            ".\n"
          )
        }
        
      } else if (is(normalisation_parameters, "featureInfoParametersNormalisationShift")) {
        if (normalisation_parameters@shift != 0.0) {
          normalisation_str <- paste0(
            "  normalisation (",
            normalisation_parameters@method,
            ") with shift = ",
            normalisation_parameters@shift,
            ".\n"
          )
        }
      }
    }

    # Batch normalisation parameters
    batch_norm_str <- character(0L)

    # Attempt to create an actual descriptor, if meaningful.
    batch_parameters <- object@batch_normalisation_parameters
    if (!is.null(batch_parameters)) {
      if (batch_parameters@method != "none") {
        # Iterate over normalisation parameters stored within the
        # container.
        for (normalisation_parameters in batch_parameters@batch_parameters) {
          if (is(normalisation_parameters, "featureInfoParametersNormalisationShiftScale")) {
            if (normalisation_parameters@shift != 0.0 || normalisation_parameters@scale != 1.0) {
              batch_norm_str <- c(
                batch_norm_str, 
                paste0(
                  "  [",
                  normalisation_parameters@batch,
                  "] normalisation (",
                  normalisation_parameters@method,
                  ") with shift = ",
                  normalisation_parameters@shift,
                  " and scale = ",
                  normalisation_parameters@scale,
                  ".\n"
                )
              )
            }
            
          } else if (is(normalisation_parameters, "featureInfoParametersNormalisationShift")) {
            if (normalisation_parameters@shift != 0.0) {
              batch_norm_str <- c(
                batch_norm_str,
                paste0(
                  "  [",
                  normalisation_parameters@batch,
                  "] normalisation (",
                  normalisation_parameters@method,
                  ") with shift = ",
                  normalisation_parameters@shift,
                  ".\n"
                )
              )
            }
          }
        }
      }
    }


    # Clustering
    cluster_str <- character(0L)

    # Attempt to create an actual descriptor, if meaningful.
    if (is(object@cluster_parameters, "featureInfoParametersCluster")) {
      if (object@cluster_parameters@cluster_size > 1L) {
        
        # Find the feature(s) required to form the cluster.
        cluster_feature_names <- object@cluster_parameters@required_features

        # Find the clustering method.
        if (is(object@cluster_parameters@method, "clusterMethod")) {
          cluster_method_str <- paste0(
            "(", object@cluster_parameters@method@method, ") "
          )
          
        } else if (is(object@cluster_parameters@method, "character")) {
          cluster_method_str <- paste0(
            "(", object@cluster_parameters@method, ") "
          )
          
        } else {
          cluster_method_str <- NULL
        }

        if (length(cluster_feature_names) == 1L) {
          # Only one feature is required to form the cluster.
          if (cluster_feature_names == object@name) {
            # The current feature is the reference feature.
            cluster_str <- paste0(
              "  forms cluster ",
              cluster_method_str,
              "with ",
              object@cluster_parameters@cluster_size - 1L,
              " other features, and is the reference feature."
            )
            
            # Determine other features that are clustered with the current
            # feature.
            other_features <- setdiff(object@cluster_parameters@cluster_features, object@name)
            
          } else {
            # The current feature is not the reference feature.
            cluster_str <- paste0(
              "  forms cluster ",
              cluster_method_str,
              "with ",
              object@cluster_parameters@cluster_size - 1L,
              " other features, with ",
              cluster_feature_names,
              " as the reference feature."
            )
            
            # Determine other features that are clustered with the current
            # feature.
            other_features <- setdiff(
              object@cluster_parameters@cluster_features,
              c(object@name, cluster_feature_names)
            )
          }
          
        } else {
          # Multiple features are required to form the cluster.
          cluster_str <- paste0(
            "  forms cluster ",
            cluster_method_str,
            "with ",
            paste_s(setdiff(cluster_feature_names, object@name)),
            "."
          )
          
          # Determine other features that are clustered with the current
          # feature.
          other_features <- setdiff(
            object@cluster_parameters@cluster_features,
            c(object@name, cluster_feature_names)
          )
        }
        
        # Mention additional features in the cluster.
        if (length(other_features) > 0L) {
          cluster_str <- c(
            cluster_str,
            paste0(" Other feature(s) in the cluster: ", paste_sh(other_features))
          )
        }
        
        # Close cluster part of the string.
        cluster_str <- c(cluster_str, "\n")
      }
    }

    # Make the feature string complete, depending on additional information.
    if (
      length(transform_str) == 0L &&
      length(normalisation_str) == 0L &&
      length(batch_norm_str) == 0L &&
      length(cluster_str) == 0L
    ) {
      feature_str <- paste0(feature_str, ".\n")
      
    } else {
      feature_str <- paste0(feature_str, ":\n")
    }

    # Export to console.
    cat(feature_str)
    if (length(transform_str) > 0L) cat(transform_str, sep = "")
    if (length(normalisation_str) > 0L) cat(normalisation_str, sep = "")
    if (length(batch_norm_str) > 0L) cat(batch_norm_str, sep = "")
    if (length(cluster_str) > 0L) cat(cluster_str, sep = "")
  }
)



# feature_info_complete (feature info) -----------------------------------------
setMethod(
  "feature_info_complete",
  signature(object = "featureInfo"),
  function(object, level = "clustering") {
    # Check if the feature is removed.
    if (!is_available(object)) return(TRUE)

    if (level == "none") return(TRUE)
    
    if (level == "signature") return(TRUE)

    if (!feature_info_complete(object@transformation_parameters)) return(FALSE)
    if (level == "transformation") return(TRUE)

    if (!feature_info_complete(object@normalisation_parameters)) return(FALSE)
    if (level == "normalisation") return(TRUE)
    
    if (!feature_info_complete(object@batch_normalisation_parameters)) return(FALSE)
    if (level == "batch_normalisation") return(TRUE)

    if (!feature_info_complete(object@imputation_parameters)) return(FALSE)
    if (level == "imputation") return(TRUE)

    if (!feature_info_complete(object@cluster_parameters)) return(FALSE)
    if (level == "clustering") return(TRUE)
    
    return(TRUE)
  }
)



# feature_info_complete (list) -------------------------------------------------
setMethod(
  "feature_info_complete",
  signature(object = "list"),
  function(object, level = "clustering") {
    return(all(sapply(
      object, 
      feature_info_complete, 
      level = level
    )))
  }
)



# feature_info_complete (NULL) -------------------------------------------------
setMethod(
  "feature_info_complete",
  signature(object = "NULL"),
  function(object, level = "clustering") {
    # No feature information found.
    return(FALSE)
  }
)



# feature_info_complete (featureInfoParameters) --------------------------------
setMethod(
  "feature_info_complete",
  signature(object = "featureInfoParameters"),
  function(object, ...) {
    # Check the 'complete' attribute.
    return(object@complete)
  }
)



# is_in_signature --------------------------------------------------------------
setMethod(
  "is_in_signature",
  signature(object = "featureInfo"),
  function(object) {
    return(object@in_signature)
  }
)


# is_in_novelty ----------------------------------------------------------------
setMethod(
  "is_in_novelty",
  signature(object = "featureInfo"),
  function(object) {
    return(object@in_novelty)
  }
)


# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "featureInfo"),
  function(object) {
    # Checks whether a feature is marked as being available
    return(!object@removed)
  }
)



# update_removed_status --------------------------------------------------------
setMethod(
  "update_removed_status",
  signature(object = "featureInfo"),
  function(object) {
    # Updates the "removed" slot based on other slots

    if (object@removed_unknown_type) {
      object@removed <- TRUE
    } else if (object@in_signature) {
      object@removed <- FALSE
    } else if (object@in_novelty) {
      object@removed <- FALSE
    } else if (object@removed_missing_values) {
      object@removed <- TRUE
    } else if (object@removed_no_variance) {
      object@removed <- TRUE
    } else if (object@removed_low_variance) {
      object@removed <- TRUE
    } else if (object@removed_low_robustness) {
      object@removed <- TRUE
    } else if (object@removed_low_importance) {
      object@removed <- TRUE
    }

    return(object)
  }
)



# add_package_version (feature info) -------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "featureInfo"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R

setClass("familiarHyperparameterLearnerBART",
  contains = "familiarHyperparameterLearner",
  slots = list(
    "encoding_reference_table" = "ANY",
    "hyperparameter_order" = "character"
  ),
  prototype = list(
    "encoding_reference_table" = NULL,
    "hyperparameter_order" = character()
  )
)




# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarHyperparameterLearnerBART"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- "BART"

    # Set name
    .Object@name <- "Bayesian additive regression trees (BART) hyperparameter optimisation model"

    return(.Object)
  }
)



.get_available_bart_hyperparameter_learners <- function() {
  return(c("bayesian_additive_regression_trees", "bart"))
}



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarHyperparameterLearnerBART"),
  function(object, ...) {
    return(c("default", "sd", "percentile", "raw"))
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarHyperparameterLearnerBART",
    data = "data.table"
  ),
  function(object, data, ...) {
    # Check if the training data is ok.
    if (has_bad_training_data(object = object, data = data)) {
      return(callNextMethod())
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Encode categorical variables.
    encoded_data <- encode_categorical_variables(
      data = data[, mget(object@target_hyperparameters)],
      object = NULL,
      encoding_method = "dummy",
      drop_levels = FALSE
    )

    # Get the hyperparameter names from the encoded data.table.
    hyperparameter_names <- colnames(encoded_data$encoded_data)

    # Identify invariant columns in the encoded data. These will be
    # dropped prior to training.
    invariant_columns <- sapply(encoded_data$encoded_data, is_singular_data)
    hyperparameter_names <- hyperparameter_names[!invariant_columns]

    # Check that are any non-invariant hyperparameters.
    if (length(hyperparameter_names) == 0L) {
      return(callNextMethod())
    }

    # Get optimisation score as response.
    y <- data$optimisation_score

    # Create model. Note that prediction by BART is expensive for large
    # ndpost.
    quiet(model <- BART::wbart(
      x.train = data.frame(encoded_data$encoded_data[, mget(hyperparameter_names)]),
      y.train = y,
      ntree = 100L,
      ndpost = 50L
    ))

    # Add model
    object@model <- model

    # Add the contrast references to the object.
    object@encoding_reference_table <- encoded_data$reference_table

    # Add hyperparameters that were used in the end.
    object@hyperparameter_order <- hyperparameter_names

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarHyperparameterLearnerBART",
    data = "data.table"
  ),
  function(
    object,
    data, 
    type = "default",
    percentile = NULL,
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) return(callNextMethod())

    # Check if the data is empty.
    if (is_empty(data)) return(callNextMethod())

    # Get an empty prediction table.
    prediction_table <- .get_placeholder_hyperparameter_prediction_table(
      object = object,
      data = data,
      type = type
    )

    # Encode categorical variables.
    x_encoded <- encode_categorical_variables(
      data = data,
      object = NULL,
      encoding_method = "dummy",
      drop_levels = FALSE
    )$encoded_data

    # Get predicted values. object@hyperparameter_order is used to drop
    # any columns that were invariant during training.
    quiet(predicted_scores <- predict(
      object = object@model,
      data.frame(x_encoded[, mget(object@hyperparameter_order)])
    ))

    # Compute mean and standard deviation.
    score_mean <- apply(predicted_scores, MARGIN = 2L, mean)
    score_sd <- apply(predicted_scores, MARGIN = 2L, stats::sd)

    # Separate by type
    if (type == "default") {
      prediction_table[, "mu" := score_mean]
      
    } else if (type == "sd") {
      prediction_table[, ":="(
        "mu" = score_mean,
        "sigma" = score_sd
      )]
      
    } else if (type == "percentile") {
      # Compute the requested percentile.
      score_percentile <- apply(
        predicted_scores,
        MARGIN = 2L,
        stats::quantile,
        probs = percentile,
        names = FALSE
      )

      prediction_table[, "percentile" := score_percentile]
      
    } else if (type == "raw") {
      # Drop raw_1 from the placeholder prediction table.
      prediction_table[, "raw_1" := NULL]

      # Parse raw data to a data.table with the expected output. Note
      # that we transpose the data to have row instances instead of
      # column instances.
      raw_data <- data.table::data.table(t(predicted_scores))

      # Set colnames.
      data.table::setnames(
        x = raw_data,
        new = paste0("raw_", seq_len(ncol(raw_data)))
      )

      # Combine with the placeholder prediction table.
      prediction_table <- cbind(prediction_table, raw_data)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "..predict,familiarHyperparameterLearnerBART,data.table: ",
        "Encountered an unknown prediction type: ", type
      ))
    }

    return(prediction_table)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_permutation_variable_importance (generic) -------------------------------

#' @title Plot permutation variable importance.
#'
#' @description This function plots the data on permutation variable importance
#'   stored in a familiarCollection object.
#'
#' @param draw (*optional*) Draws the plot if TRUE.
#' @param dir_path (*optional*) Path to the directory where created figures are
#'   saved to. Output is saved in the `variable_importance` subdirectory. If
#'   NULL no figures are saved, but are returned instead.
#' @param ggtheme (*optional*) `ggplot` theme to use for plotting.
#' @param height (*optional*) Height of the plot. A default value is derived
#'   from the number of features and the number of facets.
#' @param width (*optional*) Width of the plot. A default value is derived from
#'   the number of facets.
#' @param units (*optional*) Plot size unit. Either `cm` (default), `mm` or
#'   `in`.
#'
#' @inheritParams as_familiar_collection
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritParams plot_univariate_importance
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_permutation_vimp -object
#'
#' @details This function generates a horizontal barplot that lists features by
#'   the estimated model improvement over that of a dataset where the respective
#'   feature is randomly permuted.
#'
#'   The following splitting variables are available for `split_by`, `color_by`
#'   and `facet_by`:
#'
#'  * `vimp_method`: variable importance methods.
#'
#'  * `learner`: learners.
#'
#'  * `data_set`: data sets.
#'
#'  * `metric`: the model performance metrics.
#'
#'  * `evaluation_time`: the evaluation times (survival outcomes only).
#'
#'  * `similarity_threshold`: the similarity threshold used to identify groups
#'   of features to permute simultaneously.
#'
#'   By default, the data is split by `vimp_method`, `learner` and `metric`,
#'   faceted by `data_set` and `evaluation_time`, and coloured by
#'   `similarity_threshold`.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_feature_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#'   Bootstrap confidence intervals (if present) can be shown using various
#'   styles set by `conf_int_style`:
#'
#'  * `point_line` (default): confidence intervals are shown as lines, on which
#'   the point estimate is likewise shown.
#'
#'  * `line` (default): confidence intervals are shown as lines, but the point
#'   estimate is not shown.
#'
#'  * `bar_line`: confidence intervals are shown as lines, with the point
#'   estimate shown as a bar plot with the opacity of `conf_int_alpha`.
#'
#'  * `none`: confidence intervals are not shown. The point estimate is shown as
#'   a bar plot.
#'
#'   For metrics where lower values indicate better model performance, more
#'   negative permutation variable importance values indicate features that are
#'   more important. Because this may cause confusion, values obtained for these
#'   metrics are mirrored around 0.0 for plotting (but not any tabular data
#'   export).
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_permutation_variable_importance
#' @md
#' @rdname plot_permutation_variable_importance-methods
setGeneric(
  "plot_permutation_variable_importance",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = "feature",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    conf_int_style = c("point_line", "line", "bar_line", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_permutation_variable_importance")
  }
)



# plot_permutation_variable_importance (generic) -------------------------------

#' @rdname plot_permutation_variable_importance-methods
setMethod(
  "plot_permutation_variable_importance",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = "feature",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    conf_int_style = c("point_line", "line", "bar_line", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "permutation_vimp"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_permutation_variable_importance,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "limit_n_features" = limit_n_features,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "conf_int_alpha" = conf_int_alpha,
        "conf_int_style" = conf_int_style,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_permutation_variable_importance (collection) ----------------------------

#' @rdname plot_permutation_variable_importance-methods
setMethod(
  "plot_permutation_variable_importance",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = "feature",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    conf_int_style = c("point_line", "line", "bar_line", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    value <- ci_low <- ci_up <- NULL
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)

    # Check input arguments ----------------------------------------------------

    # Get input data.
    x <- export_permutation_vimp(
      object = object,
      aggregate_results = TRUE
    )

    # Check that the data are not empty (e.g. NULL).
    if (is_empty(x)) return(NULL)

    # Check that the data are not evaluated at the model level.
    if (all(sapply(x, function(x) (x@detail_level == "model")))) {
      ..warning_no_comparison_between_models()
      return(NULL)
    }

    # Obtain data element from list.
    if (is.list(x)) {
      if (is_empty(x)) return(NULL)

      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_model_performance: list of data elements contains unmerged elements."
        )
      }

      # Get x directly.
      x <- x[[1L]]
    }

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to create permutation variable importance plots",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # conf_int_style
    if (length(conf_int_style) > 1L) {
      conf_int_style <- head(conf_int_style, n = 1L)
    }

    # Set the style of the confidence interval to none, in case no confidence
    # interval data is present.
    if (!x@estimation_type %in% c("bci", "bootstrap_confidence_interval")) {
      conf_int_style <- "none"
    }

    # Encode the similarity_threshold as a factor.
    if (all(is.finite(x@data$similarity_threshold))) {
      # In this case the data originates from dendrograms that have been cut at
      # a certain height.

      # Obtain unique similarity thresholds.
      similarity_values <- rev(sort(unique(x@data$similarity_threshold)))

      # Convert to factor, and add nicely formatted labels.
      x@data$similarity_threshold <- factor(
        x = x@data$similarity_threshold,
        levels = similarity_values,
        labels = format(x = similarity_values, nsmall = 1L)
      )
      
    } else if (all(is.infinite(x@data$similarity_threshold))) {
      # This happens when data is not based on fixed cuts of a dendrogram.
      x@data$similarity_threshold <- factor(
        x = x@data$similarity_threshold,
        levels = c(-Inf, Inf),
        labels = c("clustered", "individual")
      )

      # Remove unused levels.
      x@data$similarity_threshold <- droplevels(x@data$similarity_threshold)
    } else {
      ..error(paste0(
        "Combinations of results from different types of clustering algorithms ",
        "cannot plotted in one figure."
      ))
    }

    # Set default splitting variables if none are provided.
    if (is.null(split_by) && is.null(color_by) && is.null(facet_by)) {
      split_by <- c("vimp_method", "learner", "metric")

      facet_by <- c("data_set")
      if (object@outcome_type %in% c("survival")) {
        facet_by <- c(facet_by, "evaluation_time")
      }

      color_by <- c("similarity_threshold")
    }

    # Set available splitting variables.
    available_splitting_variables <- c(
      "vimp_method", "learner", "data_set", "metric", "similarity_threshold"
    )
    if (object@outcome_type %in% c("survival")) {
      available_splitting_variables <- c(available_splitting_variables, "evaluation_time")
    }

    # Check splitting variables and generate sanitised output.
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      color_by = color_by,
      facet_by = facet_by,
      available = available_splitting_variables
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    color_by <- split_var_list$color_by
    facet_by <- split_var_list$facet_by

    # Parse legend label
    if (is.waive(legend_label)) {
      legend_label <- .create_plot_legend_title(
        user_label = legend_label,
        color_by = color_by
      )

      # Update "similarity threshold" in the legend label to be more specific.
      if (!is.null(color_by)) {
        if (grepl(
          pattern = "similarity threshold",
          x = legend_label$guide_color,
          fixed = TRUE
        )) {
          if (all(levels(x@data$similarity_threshold) %in% c("clustered", "individual"))) {
            legend_label$guide_color <- sub(
              pattern = "similarity threshold",
              replacement = "clustering",
              x = legend_label$guide_color,
              fixed = TRUE
            )
            
          } else {
            legend_label$guide_color <- sub(
              pattern = "similarity threshold",
              replacement = "threshold",
              x = legend_label$guide_color,
              fixed = TRUE
            )
          }
        }
      }
    }


    # Iterate over the data to identify data that should be reworked. For
    # metrics where higher scores indicate worse performance, permutation
    # variable importance is higher the more negative the value is. For other
    # metrics permutation variable is better the more positive a value is. To
    # facilitate comparisons and avoid confusion, we mirror values for the first
    # type of metrics around 0.0 here.
    x@data <- lapply(
      split(x@data, by = "metric"),
      function(x, outcome_type) {
        
        if (!is_higher_better(
          metric = as.character(x$metric[1L]),
          outcome_type = outcome_type
        )) {
          # For metrics where lower scores mark better model performance, a
          # feature is more important when the variable importance is more
          # negative.
          x[, "value" := -value]
          
          if (!is.null(x$ci_low) && !is.null(x$ci_up)) {
            # Rename confidence interval columns by exchanging upper and lower
            # bounds, and then mirror them around 0.0.
            data.table::setnames(
              x = x,
              old = c("ci_low", "ci_up"),
              new = c("ci_up", "ci_low")
            )
            
            x[, ":="(
              "ci_low" = -ci_low,
              "ci_up" = -ci_up
            )]
          }
        }
        
        return(x)
      },
      outcome_type = object@outcome_type
    )
    
    # Recombine dataset.
    x@data <- data.table::rbindlist(
      x@data,
      use.names = TRUE
    )

    if ("metric" %in% facet_by || "metric" %in% color_by) {
      available_metrics <- "combined"
    } else {
      available_metrics <- levels(x@data$metric)
    }

    # x_range depends on the 95% confidence intervals of individual metrics (if
    # split by metric), the overall range (if not split by metric), or their
    # respective point estimates (when confidence interval data are absent).
    if (is.null(x_range)) {
      # Iterate over metrics to determine the interval.
      x_range <- lapply(
        split(x@data, by = "metric"),
        function(x, conf_int_style) {
          if (conf_int_style == "none") {
            interval <- data.table::data.table(
              "min_value" = min(c(x$value, 0.0), na.rm = TRUE),
              "max_value" = max(c(x$value, 0.0), na.rm = TRUE)
            )
          } else {
            interval <- data.table::data.table(
              "min_value" = min(c(x$ci_low, 0.0), na.rm = TRUE),
              "max_value" = max(c(x$ci_up, 0.0), na.rm = TRUE)
            )
          }
          
          return(interval)
        },
        conf_int_style = conf_int_style
      )
      
      # In case multiple metrics are combined in the same plot
      if ("metric" %in% facet_by || "metric" %in% color_by) {
        # Concatenate to a single data.table.
        x_range <- data.table::rbindlist(x_range)
        x_range <- list("combined" = data.table::data.table(
          "min_value" = min(x_range$min_value, na.rm = TRUE),
          "max_value" = max(x_range$max_value, na.rm = TRUE)
        ))
      }
      
    } else if (is.list(x_range)) {
      # Check whether all metrics are present in the data provided by the user.
      .check_parameter_value_is_valid(
        x = names(x_range),
        var_name = "x_range",
        values = available_metrics
      )

      .check_argument_length(
        x = unique(names(x_range)),
        var_name = "x_range",
        min = length(available_metrics),
        max = length(available_metrics)
      )

      # Convert to the correct
      x_range <- lapply(
        x_range, 
        function(x_range) {
          .check_input_plot_args(x_range = x_range)
          
          return(data.table::data.table(
            "min_value" = min(x_range),
            "max_value" = max(x_range)
          ))
        }
      )
      
    } else {
      # For user-provided input.
      .check_input_plot_args(x_range = x_range)
      
      # Use the same range for each
      x_range <- lapply(
        available_metrics,
        function(metric, x_range) {
          return(data.table::data.table(
            "min_value" = min(x_range),
            "max_value" = max(x_range)
          ))
        },
        x_range = x_range
      )
      
      # Update names of the list elements
      names(x_range) <- available_metrics
    }
    
    # x_breaks
    if (is.null(x_breaks)) {
      .check_input_plot_args(x_n_breaks = x_n_breaks)

      # Create x_breaks.
      x_breaks <- lapply(
        x_range, 
        function(x_range, x_n_breaks) {
          # Create breaks
          x_breaks <- labeling::extended(
            m = x_n_breaks,
            dmin = x_range$min_value,
            dmax = x_range$max_value,
            only.loose = TRUE
          )
          
          return(x_breaks)
        },
        x_n_breaks = x_n_breaks
      )
      
    } else if (is.list(x_breaks)) {
      # Check whether all metrics are present in the data provided by the user.
      .check_parameter_value_is_valid(
        x = names(x_breaks),
        var_name = "x_breaks",
        values = available_metrics
      )

      .check_argument_length(
        x = unique(names(x_breaks)),
        var_name = "x_breaks",
        min = length(available_metrics),
        max = length(available_metrics)
      )

      # Check breaks.
      sapply(
        x_breaks, 
        function(x_breaks) {
          .check_input_plot_args(x_breaks = x_breaks)
        }
      )
      
    } else {
      .check_input_plot_args(x_breaks = x_breaks)
    }

    # Update x_range based on x_breaks.
    x_range <- lapply(
      available_metrics, 
      function(metric, x_range, x_breaks) {
        x_range[[metric]]$min_value <- head(x_breaks[[metric]], n = 1L)
        x_range[[metric]]$max_value <- tail(x_breaks[[metric]], n = 1L)
        
        return(x_range[[metric]])
      },
      x_range = x_range,
      x_breaks = x_breaks
    )
    
    # Set names.
    names(x_range) <- available_metrics

    # Check general input arguments
    .check_input_plot_args(
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      facet_wrap_cols = facet_wrap_cols,
      conf_int_alpha = conf_int_alpha,
      conf_int_style = conf_int_style,
      conf_int_default = c("point_line", "line", "bar_line", "none"),
      limit_n_features = limit_n_features
    )

    # Create plots -------------------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data
    if (!is.null(split_by)) {
      x_split <- split(x@data, by = split_by)
    } else {
      x_split <- list(x@data)
    }

    # Store plots to list in case no dir_path is provided
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over splits
    for (x_sub in x_split) {
      # Check that the table is not empty.
      if (is_empty(x_sub)) next

      # Check that the table contains finite values.
      if (all(is.na(x_sub$value))) next

      if (is.waive(plot_title)) plot_title <- "Permutation variable importance"

      # Declare subtitle components. The similarity metric is always shown.
      additional_subtitle <- list("similarity metric" = x@similarity_metric)

      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, color_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(x_sub$evaluation_time[1L])
        )
      }

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = x_sub
        )
      }

      # Generate plot
      p <- .plot_permutation_variable_importance(
        x = x_sub,
        color_by = color_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        limit_n_features = limit_n_features,
        conf_int_alpha = conf_int_alpha,
        conf_int_style = conf_int_style,
        x_range = x_range,
        x_breaks = x_breaks
      )

      # Check empty output
      if (is.null(p)) next

      # Draw plot
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_permutation_importance_plot_dimensions(
          x = x_sub,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "variable_importance",
              "subtype" = "permutation",
              "x" = x_sub,
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list and export
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



#' Internal plotting function for permutation variable importance plots
#'
#' @inheritParams plot_permutation_variable_importance
#'
#' @return ggplot plot object.
#'
#' @md
#' @keywords internal
.plot_permutation_variable_importance <- function(
    x,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    limit_n_features,
    conf_int_style,
    conf_int_alpha,
    x_range,
    x_breaks
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  value <- metric <- similarity_threshold <- order_id <- i.order_id <- NULL
  data_set <- learner <- vimp_method <- NULL

  # Create local copy
  x <- data.table::copy(x)

  # x_label depends on whether a single metric is shown in each plot, or
  # multiple metrics are combined.
  if (is.waive(x_label)) {
    if ("metric" %in% facet_by || "metric" %in% color_by) {
      x_label <- "variable importance"
    } else {
      x_label <- as.character(x$metric[1L])
    }
  }

  # Check the label
  .check_input_plot_args(x_label = x_label)

  # Determine available metrics
  if ("metric" %in% color_by || "metric" %in% facet_by) {
    available_metric <- "combined"
  } else {
    available_metric <- as.character(x$metric[1L])
  }

  # Sort features. In the outer loop iterate over metrics. In the inner loop
  # iterate over threshold values (in reverse). Resolve until order_id is unique
  # for all features.
  x[, "order_id" := 1L]
  for (current_data_set in levels(x$data_set)) {
    # Break in case all features have an unique order id.
    if (data.table::uniqueN(x$order_id) == data.table::uniqueN(x$feature)) break

    for (current_vimp_method in levels(x$vimp_method)) {
      # Break in case all features have an unique order id.
      if (data.table::uniqueN(x$order_id) == data.table::uniqueN(x$feature)) break

      for (current_learner in levels(x$learner)) {
        # Break in case all features have an unique order id.
        if (data.table::uniqueN(x$order_id) == data.table::uniqueN(x$feature)) break

        for (current_metric in levels(x$metric)) {
          # Break in case all features have an unique order id.
          if (data.table::uniqueN(x$order_id) == data.table::uniqueN(x$feature)) break

          for (current_threshold in rev(levels(x$similarity_threshold))) {
            for (id_table in split(
              x[
                data_set == current_data_set &
                vimp_method == current_vimp_method &
                learner == current_learner &
                metric == current_metric &
                similarity_threshold == current_threshold
              ],
              by = "order_id"
            )) {
              
              if (nrow(id_table) < 2L) next

              # Local copy
              id_table <- data.table::copy(id_table)

              # Rank by descending value.
              id_table[, "order_id" := order_id + data.table::frank(-value, ties.method = "min") - 1L]
              id_table <- id_table[, mget(c("feature", "order_id"))]

              # Update order id in x.
              x[id_table, "order_id" := i.order_id, on = "feature"]
            }

            # Break in case all features have an unique order id.
            if (data.table::uniqueN(x$order_id) == data.table::uniqueN(x$feature)) break
          }
        }
      }
    }
  }
  
  # Limit the number of features appearing in the plot.
  if (is.numeric(limit_n_features)) {
    x <- x[order_id <= limit_n_features]
  }
  
  # Order features by order_id
  x$feature <- factor(
    x = x$feature,
    levels = rev(unique(x[, mget(c("feature", "order_id"))])[order(order_id)][["feature"]])
  )

  # Generate a guide table
  guide_list <- .create_plot_guide_table(
    x = x,
    color_by = color_by,
    discrete_palette = discrete_palette
  )

  # Extract data
  x <- guide_list$data

  # Create basic plot.
  if (!is.null(color_by)) {
    # Extract guide_table for colour
    g_color <- guide_list$guide_color

    if (conf_int_style %in% c("none")) {
      p <- ggplot2::ggplot(
        data = x, 
        mapping = ggplot2::aes(
          x = !!sym("feature"),
          y = !!sym("value"),
          fill = !!sym("color_breaks")
        )
      )

      p <- p + ggplot2::scale_fill_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        guide = ggplot2::guide_legend(reverse = TRUE),
        drop = FALSE
      )
      
    } else if (conf_int_style %in% c("bar_line")) {
      p <- ggplot2::ggplot(
        data = x, 
        mapping = ggplot2::aes(
          x = !!sym("feature"),
          y = !!sym("value"),
          fill = !!sym("color_breaks"),
          color = !!sym("color_breaks")
        )
      )
      
      p <- p + ggplot2::scale_fill_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        guide = ggplot2::guide_legend(reverse = TRUE),
        drop = FALSE
      )
      
      p <- p + ggplot2::scale_colour_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        guide = ggplot2::guide_legend(reverse = TRUE),
        drop = FALSE
      )
      
    } else if (conf_int_style %in% c("line", "point_line")) {
      p <- ggplot2::ggplot(
        data = x, 
        mapping = ggplot2::aes(
          x = !!sym("feature"),
          y = !!sym("value"),
          color = !!sym("color_breaks")
        )
      )
      
      p <- p + ggplot2::scale_colour_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        guide = ggplot2::guide_legend(reverse = TRUE),
        drop = FALSE
      )
      
    } else {
      ..error_reached_unreachable_code(
        ".plot_permutation_variable_importance: unknown confidence interval style."
      )
    }
  } else {
    # Basic plot.
    p <- ggplot2::ggplot(
      data = x, 
      mapping = ggplot2::aes(
        x = !!sym("feature"),
        y = !!sym("value")
      )
    )
  }

  # Add theme.
  p <- p + ggtheme

  # Add main plotting elements.
  if (conf_int_style %in% c("bar_line")) {
    p <- p + ggplot2::geom_col(
      alpha = conf_int_alpha,
      position = "dodge"
    )

    p <- p + ggplot2::geom_linerange(
      mapping = ggplot2::aes(
        ymin = !!sym("ci_low"),
        ymax = !!sym("ci_up")
      ),
      position = ggplot2::position_dodge(width = 0.9)
    )
    
  } else if (conf_int_style %in% c("none")) {
    p <- p + ggplot2::geom_col(position = "dodge")
    
  } else if (conf_int_style %in% c("line")) {
    p <- p + ggplot2::geom_linerange(
      mapping = ggplot2::aes(
        ymin = !!sym("ci_low"),
        ymax = !!sym("ci_up")
      ),
      position = ggplot2::position_dodge(width = 0.8)
    )
    
  } else if (conf_int_style %in% c("point_line")) {
    p <- p + ggplot2::geom_pointrange(
      mapping = ggplot2::aes(
        ymin = !!sym("ci_low"),
        ymax = !!sym("ci_up")
      ),
      position = ggplot2::position_dodge(width = 0.9)
    )
    
  } else {
    ..error_reached_unreachable_code(
      ".plot_permutation_variable_importance: unknown confidence interval style."
    )
  }

  # Set breaks and limits
  x_range <- c(
    x_range[[available_metric]]$min_value,
    x_range[[available_metric]]$max_value
  )

  p <- p + ggplot2::scale_y_continuous(breaks = x_breaks[[available_metric]])
  p <- p + ggplot2::coord_flip(ylim = x_range)

  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context"
      )
    }
  }

  # Add a line to indicate 0.0, if 0.0 is included in the range.
  if (x_range[1L] <= 0.0 && x_range[2L] >= 0.0) {
    p <- p + ggplot2::geom_hline(
      yintercept = 0.0,
      linetype = "dotted"
    )
  }

  # Update labels. Note that the inversion of x_label and y_label is correct, as
  # the coordinates were flipped
  p <- p + ggplot2::labs(
    x = y_label,
    y = x_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  return(p)
}



.determine_permutation_importance_plot_dimensions <- function(
    x, 
    facet_by, 
    facet_wrap_cols
) {
  # Get plot layout dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Determine the number of features within each facet.
  n_features <- data.table::uniqueN(x = x$feature)
  longest_name <- max(sapply(levels(x$feature), nchar))

  # Assume each feature takes up about 14 points (~5mm) with 2 point (0.7mm)
  # spacing. Then add some room for other plot elements.
  default_height <- n_features * 0.5 + (n_features - 1L) * 0.07 + 1.0

  # Set a default height. Assume that the typical width of a character is about
  # 5 points (1.8mm).
  default_width <- 6.0 + longest_name * 0.18

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))

  return(c(height, width))
}

# Don't perform any further tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test.
familiar:::test_plots(
  plot_function = familiar:::plot_model_signature_occurrence,
  not_available_no_samples = FALSE,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  data_element = "model_vimp",
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_signature_occurrence,
  data_element = "model_vimp",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_signature_occurrence,
  data_element = "model_vimp",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner"),
    "color_by" = c("vimp_method")),
  debug = debug_flag
)

# Generic test.
familiar:::test_plots(
  plot_function = familiar:::plot_model_signature_variable_importance,
  not_available_no_samples = FALSE,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  data_element = "model_vimp",
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_signature_variable_importance,
  data_element = "model_vimp",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_signature_variable_importance,
  data_element = "model_vimp",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner"),
    "color_by" = c("vimp_method")),
  debug = debug_flag
)


# With a limit to the number of features shown.
familiar:::test_plots(
  plot_function = familiar::plot_model_signature_variable_importance,
  data_element = "model_vimp",
  test_config = "normal",
  plot_args = list("limit_n_features" = 2L),
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarMetricBrier ----------------------------------------------------------
setClass(
  "familiarMetricBrier",
  contains = "familiarMetric",
  prototype = methods::prototype(
    metric = "brier",
    outcome_type = NA_character_,
    name = "Brier Score",
    value_range = c(0.0, 1.0),
    higher_better = FALSE
  )
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMetricBrier"),
  function(object, ...) {
    return(object@outcome_type %in% c("binomial", "multinomial"))
  }
)



# compute_metric_score ---------------------------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricBrier"),
  function(metric, data, ...) {
    
    if (is_empty(data)) return(callNextMethod())
    
    if (!is(data, "predictionTableClassification")) {
      ..error_data_not_prediction_table(data, "predictionTableClassification")
    }
    
    # Get the classes and number of classes in data.
    outcome_classes <- get_outcome_class_levels(data)
    n_classes <- length(outcome_classes)

    # Skip calculation if there are no other classes.
    if (n_classes <= 1L) return(callNextMethod())

    # Remove any entries that lack valid predictions.
    data <- remove_invalid_predictions(data)

    # Remove any entries that lack observed values.
    data <- filter_missing_outcome(data)
    if (is_empty(data)) return(callNextMethod())
    
    # Identify the outcome column.
    outcome_column <- get_outcome_columns(x = data@outcome_type)
    data <- .as_data_table(data)

    # Create empty brier score vector.
    brier_score <- vector(mode = "numeric", length = n_classes)

    # Iterate over classes.
    for (ii in seq_along(outcome_classes)) {
      # Find the current positive class
      positive_class <- outcome_classes[ii]

      # Prepare data to compute a Brier score.
      brier_data <- data.table::copy(data[, mget(c(outcome_column, positive_class))])

      # Mark positive class as 1, and the rest as 0.
      brier_data[, "positive_outcome" := 0L]
      brier_data[get(outcome_column) == positive_class, "positive_outcome" := 1L]

      # Calculate uncorrected brier score for the current positive class.
      brier_score[ii] <- sum((brier_data[[positive_class]] - brier_data$positive_outcome)^2.0)
    }

    # Calculate overall brier score
    brier_score <- sum(brier_score) / nrow(data)

    # Correct score for binomial and multinomial outcomes so that it falls in
    # the (0, 1) range
    brier_score <- brier_score / 2.0

    return(brier_score)
  }
)



.get_available_brier_metrics <- function() {
  return(c("brier"))
}

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()

# power.transform and other packages are required.
if (!rlang::is_installed("power.transform")) testthat::skip()

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE
)

# Test that the duplicated data check ------------------------------------------

# Duplicate the first row.
duplicated_data <- data.table::copy(data)
duplicated_data[2L, ] <- duplicated_data[1L, ]

# Create data assignment object.
testthat::expect_warning(
  experiment_data_assignment <- familiar::precompute_data_assignment(
    data = duplicated_data,
    experimental_design = "bs(fs+mb,3)",
    outcome_type = "binomial",
    outcome_column = "outcome",
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    class_levels = c("red", "green"),
    verbose = FALSE
  ),
  class = "familiar_data_check"
)


# Test the invariant feature check ---------------------------------------------

# Add invariant feature
invariant_data <- data.table::copy(data)
invariant_data[, "feature_5" := 3.0]

testthat::expect_warning(
  experiment_data_assignment <- familiar::precompute_data_assignment(
    data = invariant_data,
    experimental_design = "bs(fs+mb,3)",
    outcome_type = "binomial",
    outcome_column = "outcome",
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    class_levels = c("red", "green"),
    verbose = FALSE
  ),
  class = "familiar_data_check"
)


# Test the one-to-one predictor check ------------------------------------------

# Add one-to-one predictor
one_to_one_data <- data.table::copy(data)
one_to_one_data[, "feature_5" := 0.0]
one_to_one_data[outcome == "red", "feature_5" := 1.0]

testthat::expect_warning(
  experiment_data_assignment <- familiar::precompute_data_assignment(
    data = one_to_one_data,
    experimental_design = "bs(fs+mb,3)",
    outcome_type = "binomial",
    outcome_column = "outcome",
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    class_levels = c("red", "green"),
    verbose = FALSE
  ),
  class = "familiar_data_check"
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_shap_force (generic) ----------------------------------------------------

#' @title Create SHAP force plot
#'
#' @description This method creates plots that show stacked SHAP force values
#'   obtained from the data stored in a familiarCollection object.
#'
#' @param dir_path (*optional*) Path to the directory where created SHAP force
#'   plots are saved to. Output is saved in the `explanation` subdirectory. If
#'   `NULL` no figures are saved, but are returned instead.
#' @param discrete_palette (*optional*) Discrete palette used to colour the
#'   elements of force plots. `familiar` has a default palette. Other palettes
#'   are supported by the `paletteer` package, `grDevices::palette.pals()`
#'   (requires R >= 4.0.0), `grDevices::hcl.pals()` (requires R >= 3.6.0). You
#' may also specify your own palette by providing a vector of colour names
#' listed by `grDevices::colors()` or through hexadecimal RGB strings.
#' @param highlight_feature (*optional*) Name of one or more features that
#'   should be highlighted in the force plot.
#' @param sample_order (*optional*) Ordering of samples, one of:
#'
#'   * `prediction`: samples are ordered by increasing predicted value. Sample
#'   order between facets may differ.
#'
#'   * `original`: samples retain the original ordering. Sample order between
#'   facets is consistent.
#' 
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams extract_performance -object
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#'
#' @details This function plots model performance based on empirical bootstraps,
#'   using various plot representations.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, `data_set`,
#'   `evaluation_time` (survival outcome only) and `positive_class` (categorical
#'   outcomes). The default for is to facet by `evaluation_time` or
#'   `positive_class`, and split by `vimp_method`, `learner` and `data_set`.
#'   `color_by` is not used.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_learner_names`
#'   can be applied to the `familiarCollection` object to update labels, and
#'   order the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_shap_force
#' @md
#' @rdname plot_shap_force-methods
setGeneric(
  "plot_shap_force",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    highlight_feature = NULL,
    sample_order = "prediction",
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_shap_force")
  }
)



# plot_shap_force (general) ------------------------------------------------

#' @rdname plot_shap_force-methods
setMethod(
  "plot_shap_force",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    highlight_feature = NULL,
    sample_order = "prediction",
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "shap"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_shap_force,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "x_axis_by" = x_axis_by,
        "y_axis_by" = y_axis_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "x_label_shared" = x_label_shared,
        "y_label" = y_label,
        "y_label_shared" = y_label_shared,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "highlight_feature" = highlight_feature,
        "sample_order" = sample_order,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_shap_force (collection) -----------------------------------------------

#' @rdname plot_shap_force-methods
setMethod(
  "plot_shap_force",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    highlight_feature = NULL,
    sample_order = "prediction",
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Check input arguments ----------------------------------------------------
    
    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)
    
    # Load the data.
    x <- export_shap(object = object)
    
    x <- x$shap_force
    if (is_empty(x)) return(NULL)
    
    # Obtain single data element from list.
    if (is.list(x)) {
      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_shap_force: list of data elements contains unmerged elements."
        )
      }
      x <- x[[1L]]
    }
    
    # Check that the data are not evaluated at the model level.
    if (x@detail_level == "model") {
      ..warning_no_comparison_between_models()
      return(NULL)
    }
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    # Ensure that we work with a copy of the data.
    x@data <- data.table::copy(x@data)
    
    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to create a SHAP waterfall plot",
      message_type = "warning"
    )) {
      return(NULL)
    }
    
    # Add evaluation time or class as a splitting variable.
    additional_variable <- NULL
    if (object@outcome_type %in% c("survival")) {
      additional_variable <- "evaluation_time"
      data.table::setnames(x@data, old = "shap_outcome", new = "evaluation_time")
      
    } else if (object@outcome_type %in% c("multinomial")) {
      additional_variable <- "positive_class"
      data.table::setnames(x@data, old = "shap_outcome", new = "positive_class")
    }
    
    # Add default splitting variables.
    if (
      is.null(split_by) &&
      is.null(facet_by)
    ) {
      # Split by vimp_method, learner.
      split_by <- c("vimp_method", "learner", "data_set")
      facet_by <- additional_variable
    }
    
    all_variables <- c("vimp_method", "learner", "data_set", additional_variable)
    
    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      facet_by = facet_by,
      available = all_variables
    )
    
    # Update splitting variables
    split_by <- split_var_list$split_by
    facet_by <- split_var_list$facet_by
    
    # x_label
    if (is.waive(x_label)) {
      x_label <- "sample"
    }
    
    # y_label
    if (is.waive(y_label)) {
      y_label <- "predicted value"
    }
    
    # x_label_shared
    if (!is.waive(x_label_shared)) {
      .check_input_plot_args(x_label_shared = x_label_shared)
    } else {
      x_label_shared <- "column"
    }
    
    # y_label_shared
    if (!is.waive(y_label_shared)) {
      .check_input_plot_args(y_label_shared = y_label_shared)
    } else {
      y_label_shared <- "row"
    }
    
    .check_input_plot_args(
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption
    )
    
    # Check that highlight_feature appears as a feature.
    if (!is.null(highlight_feature)) {
      features_in_data <- highlight_feature %in% levels(x@data$feature_name)
      
      if (!all(features_in_data)) {
        ..warning(
          paste0(
            "Not all features to highlight for SHAP force plots were found in the dataset. Missing: ",
            paste_s(highlight_feature[!features_in_data])
          )
        )
      }
    }
    
    # sample_order
    .check_parameter_value_is_valid(
      x = sample_order, 
      var_name = "sample_order", 
      values = c("prediction", "original")
    )
    
    # Create plots -------------------------------------------------------------
    
    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)
    
    # Split data.
    if (!is.null(split_by)) {
      x_split <- split(
        x@data, 
        by = split_by, 
        drop = FALSE
      )
      
    } else {
      x_split <- list("null.name" = x@data)
    }
    
    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()
    
    # Iterate over data splits.
    for (ii in names(x_split)) {
      # Skip empty datasets.
      if (is_empty(x_split[[ii]])) next
      
      if (is.waive(plot_title)) plot_title <- "SHAP force"
      
      # Declare subtitle components.
      additional_subtitle <- NULL
      
      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(x_split[[ii]]$evaluation_time[1L])
        )
      }
      
      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = x_split[[ii]]
        )
      }
      
      p <- .plot_shap_force_plot(
        x = x_split[[ii]],
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        x_label = x_label,
        x_label_shared = x_label_shared,
        y_label = y_label,
        y_label_shared = y_label_shared,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        y_range = y_range,
        y_n_breaks = y_n_breaks,
        y_breaks = y_breaks,
        highlight_feature = highlight_feature,
        sample_order = sample_order
      )
      
      # Check empty output
      if (is.null(p)) next
      
      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)
      
      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_shap_force_plot_dimensions(
          x = x_split[[ii]],
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols
        )
        
        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "explanation",
              "subtype" = "shap_force",
              "x" = x_split[[ii]],
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }
    
    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_shap_force_plot <- function(
    x,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    x_label_shared,
    y_label,
    y_label_shared,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    y_range,
    y_n_breaks,
    y_breaks,
    highlight_feature,
    sample_order
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  shap_value <- prediction <- NULL
  
  # Split by facet. This generates a list of data splits with faceting
  # information that allows for positioning.
  plot_layout_table <- .get_plot_layout_table(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Set the y-range, as this should be fixed across facets.
  if (is.null(y_range)) {
    # Find the correct y-range
    y_range_data <- x[
      ,
      list(
        "y_min" = prediction - sum(pmax(shap_value, 0.0)),
        "y_max" = prediction - sum(pmin(shap_value, 0.0))
      ),
      by = c("sample_id", facet_by)
    ]
    y_range <- c(min(y_range_data$y_min, na.rm = TRUE), max(y_range_data$y_max, na.rm = TRUE))
    if (y_range[1L] == y_range[2L]) y_range <- y_range + c(-0.1, 0.1)
  }

  .check_input_plot_args(y_range = y_range)
  
  # x_breaks
  if (is.null(y_breaks)) {
    .check_input_plot_args(
      y_range = y_range,
      y_n_breaks = y_n_breaks
    )
    
    # Create breaks and update x_range
    y_breaks <- labeling::extended(
      m = y_n_breaks,
      dmin = y_range[1L],
      dmax = y_range[2L],
      only.loose = TRUE
    )
    
    y_range <- c(
      head(y_breaks, n = 1L),
      tail(y_breaks, n = 1L)
    )
    
  } else {
    .check_input_plot_args(y_breaks = y_breaks)
  }
  
  # Split data into facets. This is done by row.
  data_facet_list <- .split_data_by_plot_facet(
    x = x,
    plot_layout_table = plot_layout_table
  )
  
  # Used for ordering of composite figures.
  layout_split <- split(
    plot_layout_table,
    by = c("col_id", "row_id"),
    sorted = TRUE
  )
  
  # Placeholders for plots.
  figure_list <- list()
  extracted_element_list <- list()
  
  # Iterate over facets
  for (ii in names(layout_split)) {
    # Create calibration plot.
    p_shap_force <- .create_shap_force_plot(
      x = data_facet_list[[ii]],
      facet_by = facet_by,
      facet_wrap_cols = facet_wrap_cols,
      ggtheme = ggtheme,
      discrete_palette = discrete_palette,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      y_range = y_range,
      y_breaks = y_breaks,
      highlight_feature = highlight_feature,
      sample_order = sample_order
    )
    
    # Rename plot elements.
    g_shap_force <- .rename_plot_grobs(
      g = .convert_to_grob(p_shap_force),
      extension = "main"
    )
    if (!gtable::is.gtable(g_shap_force)) next
    
    # Attach to figure list.
    figure_list[[paste0(layout_split[[ii]]$row_id, ".", layout_split[[ii]]$col_id)]] <- as_familiar_plot(
      g = g_shap_force,
      layout = layout_split[[ii]]
    )
  }
  
  # Compose the final figure. Magic.
  g <- .compose_figure(
    figure_list = figure_list,
    plot_layout_table = plot_layout_table,
    x_text_shared = x_label_shared,
    x_label_shared = x_label_shared,
    y_text_shared = y_label_shared,
    y_label_shared = y_label_shared,
    facet_wrap_cols = facet_wrap_cols,
    ggtheme = ggtheme
  )
  
  return(g)
}



.create_shap_force_plot <- function(
    x,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    y_range,
    y_breaks,
    highlight_feature,
    sample_order
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  shap_value <- vimp <- feature_value <- feature_name <- NULL
  feature_label <- prediction <- y <- label_text <- NULL
  
  # The force plot has two axes: a sample axis, and a prediction axis. By
  # default, samples are sorted according to the prediction value within the
  # facet. For each sample, the marginal feature value contributions to the
  # prediction are sorted by the absolute value of the contribution.
  
  # Set sample order in each facet.
  if (sample_order == "prediction") {
    
    # Order samples by increasing predicted value.
    prediction_table <- unique(x[, mget(c("prediction", "sample_id"))])
    prediction_table[, "sample_order" := order(order(prediction, decreasing = FALSE))]
    
    x <- merge(
      x = x,
      y = prediction_table[, "prediction" := NULL],
      by = "sample_id"
    )
    
  } else if (sample_order == "original") {
    sample_table <- unique(x[, mget(c("sample_id"))])
    sample_table[, "sample_order" := .I]
    
    x <- merge(
      x = x,
      y = sample_table,
      by = "sample_id"
    )
    
  } else {
    ..error_reached_unreachable_code(paste0("encountered invalid value for sample_order: ", sample_order))
  }
  
  # Update start and end positions for force elements.
  x[
    ,
    (c("shap_start", "shap_end")) := ..set_shap_force_positions(shap_value, prediction),
    by = c(facet_by, "sample_id")
  ]

  
  # Create a legend label.
  legend_label <- .create_plot_legend_title(
    user_label = legend_label,
    color_by = "shap_value"
  )
  
  # Check remaining input arguments.
  .check_input_plot_args(
    legend_label = legend_label
  )
  
  # Add gradient palette.
  discrete_palette <- .get_palette(
    x = discrete_palette, 
    palette_type = "qualitative",
    n = 2L
  )
  x[, "shap_positive" := shap_value >= 0.0]
  x[, "shap_highlight" := feature_name %in% highlight_feature]
  
  # Set up basic force plot.
  p <- ggplot2::ggplot(data = x)
  p <- p + ggtheme
  p <- p + geom_fam_force_shap(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("sample_order"),
      xmin = !!sym("sample_order") - 0.4,
      xmax = !!sym("sample_order") + 0.4,
      y = !!sym("prediction"),
      ymin = !!sym("shap_start"),
      ymax = !!sym("shap_end"),
      fill = !!sym("shap_positive"),
      colour = !!sym("shap_positive"),
      alpha = !!sym("shap_highlight")
    )
  )
  
  # Set alpha values.
  p <- p + ggplot2::scale_alpha_manual(
    values = c("TRUE" = 1.0, "FALSE" = 0.4),
    guide = "none"
  )
  
  # Set colour and fill.
  p <- p + ggplot2::scale_fill_manual(
    values = c("FALSE" = discrete_palette[1L], "TRUE" = discrete_palette[2L]),
    guide = "none",
    aesthetics = c("colour", "fill")
  )
  
  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )
  
  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context"
      )
    }
  }
  
  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label, 
    y = y_label, 
    title = plot_title, 
    subtitle = plot_sub_title, 
    caption = caption
  )
  
  # Set breaks and  limits on the x and y-axis
  # p <- p + ggplot2::scale_x_continuous(breaks = x_breaks)
  p <- p + ggplot2::scale_y_continuous(breaks = y_breaks, limits = y_range)
  
  return(p)
}



..set_shap_force_positions <- function(x, predictions) {
  # Prevent notes.
  feature_value <- density <- y_offset <- NULL
  
  # Initialise.
  x_start <- x_end <- numeric(length(x))

  # Set feature order based on the absolute shap value.
  feature_order <- order(abs(x), decreasing = TRUE)

  # Initialise positions.
  previous_end_pos <- previous_end_neg <- utils::head(predictions, n = 1L)
  
  # We fill start and end positions starting with the most
  # important feature first.
  for (ii in feature_order) {
    if (x[ii] >= 0.0){
      x_start[ii] <- previous_end_pos
      previous_end_pos <- x_end[ii] <- previous_end_pos - x[ii]
      
    } else {
      x_start[ii] <- previous_end_neg
      previous_end_neg <- x_end[ii] <- previous_end_neg - x[ii]
    }
  }

  return(list(
    "shap_start" = x_start,
    "shap_end" = x_end
  ))
}

# 
# 
# ..set_shap_waterfall_feature_name_value <- function(
#     feature_name,
#     feature_value, 
#     feature_label
# ) {
#   actual_label <- feature_label
#   actual_label[is.na(feature_label)] <- signif(feature_value[is.na(feature_label)], 3L)
#   
#   return(paste0(feature_name, ": ", actual_label))
# }



.determine_shap_force_plot_dimensions <- function(
    x,
    x_axis_by,
    y_axis_by,
    facet_by,
    facet_wrap_cols
) {
  
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- 4.0
  
  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))
  
  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))
  
  return(c(height, width))
}

# GeomSHAPForce ----------------------------------------------------------------

# Placeholder to prevent NOTES if ggplot2 is not installed.
GeomSHAPForce <- NULL
if (rlang::is_installed("ggplot2")) {
  GeomSHAPForce <- ggplot2::ggproto(
    "GeomPolygon",
    ggplot2::Geom,
    required_aes = c("x", "xmin", "xmax", "y", "ymin", "ymax"),
    default_aes = ggplot2::aes(
      colour = NA,
      fill = "grey35",
      linewidth = 0.5,
      linetype = 1,
      alpha = NA
    ),
    draw_key = ggplot2::draw_key_polygon,
    draw_panel = function(
      data,
      panel_params,
      coord,
      lineend = "butt",
      linejoin = "round",
      linemitre = 10
    ) {
      # Compute coordinates based on data.
      coords <- coord$transform(data, panel_params)

      # Instantiate parameters to feed to grid::polygonGrob. These vectors are 
      # sufficiently large to hold all polygons with taper.
      x <- y <- numeric(nrow(coords) * 6L)
      id <- integer(nrow(coords) * 6L)
      
      # Iterate over features.
      idx_offset <- 0L
      for (ii in seq_len(nrow(coords))) {
        # x is over the samples; y is over the predictions.
        x_min <- coords$xmin[ii]
        x_max <- coords$xmax[ii]
        y_min <- coords$ymin[ii]
        y_max <- coords$ymax[ii]
        
        idx <- idx_offset + (1L:6L)
        
        if (y_max <= y_min) {
          y_max_flank <- y_max - 0.01
          y_min_flank <- y_min - 0.01
          
        } else {
          y_max_flank <- y_max + 0.01
          y_min_flank <- y_min + 0.01
        }
        
        x_mid <- (x_min + x_max) / 2.0
        
        x[idx] <- c(x_min, x_mid, x_max, x_max, x_mid, x_min)
        y[idx] <- c(y_max_flank, y_max, y_max_flank, y_min_flank, y_min, y_min_flank)
        
        # Set grouping.
        id[idx] <- ii
        
        # Update offset.
        idx_offset <- idx_offset + length(idx)
      }
      
      # Select elements which are correctly set.
      valid_idx <- id > 0L
      
      return(grid::polygonGrob(
        x[valid_idx],
        y[valid_idx],
        id = id[valid_idx],
        default.units = "native",
        gp = grid::gpar(
          col = coords$colour,
          fill = ggplot2::fill_alpha(coords$fill, coords$alpha),
          lwd = coords$linewidth * ggplot2::.pt,
          lty = coords$linetype,
          lineend = lineend,
          linejoin = linejoin,
          linemitre = linemitre
        )
      ))
    }
  )
}



geom_fam_force_shap <- function(
    mapping = NULL,
    data = NULL,
    stat = "identity",
    position = "identity",
    na.rm = FALSE,
    show.legend = NA,
    inherit.aes = TRUE,
    ...
) {
  ggplot2::layer(
    geom = GeomSHAPForce,
    mapping = mapping,
    data = data, 
    stat = stat, 
    position = position, 
    show.legend = show.legend, 
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, ...)
  )
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarNoneNoveltyDetector --------------------------------------------------
setClass(
  "familiarNoneNoveltyDetector",
  contains = "familiarNoveltyDetector"
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarNoneNoveltyDetector"),
  function(object, ...) {
    # We can always not create a novelty detector.
    return(TRUE)
  }
)



.get_available_none_detectors <- function(show_general = TRUE) {
  return(c("none", "false", "no"))
}

#' familiar: Fully Automated Machine Learning with Interpretable Analysis of Results
#'
#' End-to-end, automated machine learning package for creating 
#' trustworthy and interpretable machine learning models. Familiar supports
#' modelling of regression, categorical and time-to-event (survival) outcomes.
#' Models created using familiar are self-containing, and their use does not
#' require additional information such as baseline survival, feature
#' clustering, or feature transformation and normalisation parameters. In
#' addition, an novelty or out-of-distribution detector is trained
#' simultaneously and contained with every model. Model performance,
#' calibration, risk group stratification, (permutation) variable importance,
#' individual conditional expectation, partial dependence, and more, are
#' assessed automatically as part of the evaluation process and exported in
#' tabular format and plotted, and may also be computed manually using export
#' and plot functions. Where possible, metrics and values obtained during the
#' evaluation process come with confidence intervals.
#'
#' @docType package
#' @name familiar
#' @import data.table
#' @import methods
#' @importClassesFrom rstream rstream.runif
#' @importFrom stats predict coef vcov
#' @importFrom survival Surv coxph survreg
#' @importFrom utils head tail
#' @importFrom rlang quo quos enquo enquos sym syms ensym ensyms parse_expr parse_exprs %<~%
"_PACKAGE"



#' Perform end-to-end machine learning and data analysis
#'
#' @param formula An R formula. The formula can only contain feature names and
#'   dot (`.`). The `*` and `+1` operators are not supported as these refer to
#'   columns that are not present in the data set.
#'   
#'   Use of the formula interface is optional.
#'
#' @param data A `data.table` object, a `data.frame` object, list containing
#'   multiple `data.table` or `data.frame` objects, or paths to data files.
#'
#'   `data` should be provided if no file paths are provided to the `data_files`
#'   argument. If both are provided, only `data` will be used.
#'
#'   All data is expected to be in wide format, and ideally has a sample
#'   identifier (see `sample_id_column`), batch identifier (see `cohort_column`)
#'   and outcome columns (see `outcome_column`).
#'
#'   In case paths are provided, the data should be stored as `csv`, `rds` or
#'   `RData` files. See documentation for the `data_files` argument for more
#'   information.
#'   
#' @param experiment_data Experimental data may provided in the form of the 
#'   output of `precompute_data_assignment`, `precompute_feature_info` or
#'   `precompute_vimp`. This allows for warm-starting experiments.
#' 
#' @param cl Cluster created using the `parallel` package. This cluster is then
#'   used to speed up computation through parallelisation. When a cluster is not
#'   provided, parallelisation is performed by setting up a cluster on the local
#'   machine.
#'
#'   This parameter has no effect if the `parallel` argument is set to `FALSE`.
#' @param config List containing configuration parameters, or path to an `xml`
#'   file containing these parameters. An empty configuration file can obtained
#'   using the `get_xml_config` function.
#'
#'   All parameters can also be set programmatically. These supersede any
#'   arguments derived from the configuration list.
#'
#' @param config_id Identifier for the configuration in case the list or `xml`
#'   table indicated by `config` contains more than one set of configurations.
#' @param verbose Indicates verbosity of the results. Default is TRUE, and all
#'   messages and warnings are returned.
#' @param .stop_after Variable for internal use.
#' @param .force_output Generates output even if results have been written to
#'   the file system.
#'
#' @inheritDotParams .parse_file_paths -config -verbose
#' @inheritDotParams .parse_experiment_settings -config
#' @inheritDotParams .parse_setup_settings -config
#' @inheritDotParams .parse_preprocessing_settings -config -data -parallel -outcome_type
#' @inheritDotParams .parse_variable_importance_settings -config -data -parallel -outcome_type
#' @inheritDotParams .parse_model_development_settings -config -data -parallel -outcome_type
#' @inheritDotParams .parse_hyperparameter_optimisation_settings -config -parallel -outcome_type
#' @inheritDotParams .parse_evaluation_settings -config -data -parallel -outcome_type -hpo_metric -development_batch_id -vimp_aggregation_rank_threshold -vimp_aggregation_method -prep_cluster_method -prep_cluster_linkage_method -prep_cluster_cut_method -prep_cluster_similarity_threshold -prep_cluster_similarity_metric
#'
#' @return Nothing. All output is written to the experiment directory. If the 
#'   experiment directory is in a temporary location, a list with all
#'   familiarModel, familiarEnsemble, familiarData and familiarCollection
#'   objects will be returned.
#'
#' @inherit .parse_general_settings references
#'
#' @export
#' @md
summon_familiar <- function(
    formula = NULL,
    data = NULL,
    experiment_data = NULL,
    cl = NULL,
    config = NULL,
    config_id = 1L,
    verbose = TRUE,
    .stop_after = "export",
    .force_output = FALSE,
    ...
) {
  
  # Set options.
  # Disable randomForestSRC OpenMP core use.
  options(rf.cores = 1L)
  on.exit(options(rf.cores = -1L), add = TRUE)
  
  # Disable multithreading on data.table to prevent reduced performance due to
  # resource collisions with familiar parallelisation.
  data.table::setDTthreads(1L)
  on.exit(data.table::setDTthreads(0L), add = TRUE)
  
  # Check .stop_after
  .check_parameter_value_is_valid(
    x = .stop_after,
    var_name = ".stop_after",
    values = c("setup", "preprocessing", "vimp", "training", "evaluation", "export")
  )
  
  
  # Load configuration file ----------------------------------------------------
  config <- .load_configuration_file(
    config = config,
    config_id = config_id
  )
  
  
  # Test arguments provided by ... and config ----------------------------------
  .check_dots_is_parameter(dots = names(list(...)))
  .check_configuration_tag_is_parameter(config = config)
  
  
  # File paths -----------------------------------------------------------------
  # Parse file paths
  file_paths  <- do.call(
    .parse_file_paths,
    args = c(
      list("config" = config, "verbose" = verbose),
      list(...)
    )
  )
  
  # Set paths to data
  if (!is.null(file_paths$data) && is.null(data)) {
    data <- file_paths$data
  } 
  
  # Make sure to remove the directory if it is on the temporary path.
  if (file_paths$is_temporary) {
    on.exit(
      unlink(file_paths$experiment_dir, recursive = TRUE),
      add = TRUE
    )
  } 
  
  # Load data ------------------------------------------------------------------
  dots <- list(...)
  
  # Add placeholder for experimental_design in case experiment_data is set.
  if (is.null(dots$experimental_design) && !is.null(experiment_data)) {
    dots$experimental_design <- "fs+mb"
  }
  
  # Parse experiment and data settings
  settings <- do.call(
    .parse_initial_settings,
    args = c(list("config" = config), dots)
  )
  
  if (is(data, "dataObject")) {
    # Reconstitute settings from the data.
    new_settings <- extract_settings_from_data(data)
    
    # Replace settings with new settings.
    settings$data[names(new_settings$data)] <- new_settings$data
    
    # Extract data as a data.table.
    data <- data@data
    
  } else {
    # Load data.
    data <- do.call(
      .load_data,
      args = c(list("data" = data), settings$data)
    )
    
    # Update settings
    settings <- .update_initial_settings(
      formula = formula,
      data = data,
      settings = settings
    )
  }
  
  # Parse data
  data <- .finish_data_preparation(
    data = data,
    sample_id_column = settings$data$sample_col,
    batch_id_column = settings$data$batch_col,
    series_id_column = settings$data$series_col,
    outcome_column = settings$data$outcome_col,
    outcome_type = settings$data$outcome_type,
    include_features = settings$data$include_features,
    class_levels = settings$data$class_levels,
    censoring_indicator = settings$data$censoring_indicator,
    event_indicator = settings$data$event_indicator,
    competing_risk_indicator = settings$data$competing_risk_indicator,
    reference_method = settings$data$reference_method
  )
  
  # Initialise workers ---------------------------------------------------------
  # Identify if an external cluster is provided, and required.
  is_external_cluster <- FALSE
  if (settings$run$parallel) {
    is_external_cluster <- inherits(cl, "cluster")
  }
  
  # Assign parallel options to global familiar environment.
  .assign_backend_options_to_global(
    backend_type = settings$run$backend_type,
    server_port = settings$run$server_port
  )
  .assign_parallel_options_to_global(
    is_external_cluster = is_external_cluster,
    restart_cluster = settings$run$restart_cluster,
    n_cores = settings$run$parallel_nr_cores,
    cluster_type = settings$run$cluster_type
  )
  
  # Make sure that backend server will close after the process finishes.
  on.exit(
    shutdown_backend_server(
      backend_type = settings$run$backend_type,
      server_port = settings$run$server_port
    ),
    add = TRUE
  )
  
  # Start workers, if required.
  if (
    settings$run$parallel &&
    !settings$run$restart_cluster &&
    !is_external_cluster
  ) {
    # Start local cluster in the overall process. Note that at this point we
    # cannot assign anything relevant -- this will be handled later.
    cl <- .restart_cluster(cl = NULL, assign = "none")
    on.exit(.terminate_cluster(cl), add = TRUE)
    
  } else if (
    settings$run$parallel &&
    settings$run$restart_cluster &&
    !is_external_cluster
  ) {
    # Start processes locally.
    cl <- waiver()
    
  } else if (!settings$run$parallel) {
    # No cluster is created when 
    cl <- NULL
  }
  
  # Data plausibility checks ---------------------------------------------------
  
  # Check data plausibility.
  .check_data_plausibility(
    data = data,
    settings = settings,
    verbose = verbose,
    cl = cl
  )
  
  # Load experiment data -------------------------------------------------------
  if (!is.null(experiment_data)) {
    
    # Write experiment data to the file system
    experiment_data <- load_experiment_data(
      experiment_data,
      file_paths = file_paths
    )
    
    # Force the use of existing experiment data.
    settings$data$exp_design <- .get_iteration_file_name(
      project_id = experiment_data@project_id,
      file_paths = file_paths
    )
  }
  
  # Experimental setup and settings --------------------------------------------
  
  # Derive experimental design
  experiment_setup <- extract_experimental_setup(
    experimental_design = settings$data$exp_design,
    file_dir = file_paths$iterations_dir,
    verbose = verbose
  )
  
  # Check experiment settings
  settings <- .update_experimental_design_settings(
    section_table = experiment_setup,
    data = data,
    settings = settings,
    verbose = verbose
  )
  
  # Import remaining settings
  settings <- .parse_general_settings(
    config = config,
    data = data,
    settings = settings,
    ...
  )
  
  # Create a generic outcome object
  outcome_info <- create_outcome_info(settings = settings)
  
  # Define iterations etc.
  project_info <- .get_iteration_data(
    file_paths = file_paths,
    data = data,
    experiment_setup = experiment_setup,
    settings = settings,
    verbose = verbose
  )
  
  # In case the iterations are loaded from a iterations file provided by the
  # user, perform some checks on the experimental design given the current data
  # set.
  if (is.waive(experiment_setup)) {
    # Replace experiment_setup with the setup that was loaded from the
    # iterations file.
    experiment_setup <- project_info$experiment_setup
    
    # Check experiment settings for the current dataset.
    settings <- .update_experimental_design_settings(
      section_table = experiment_setup,
      data = data,
      settings = settings,
      verbose = verbose
    )
  }
  
  # Update the number of runs based on the iteration list.
  experiment_setup <- .set_experimental_design_n_runs(
    section_table = experiment_setup,
    iteration_list = project_info$iter_list
  )
  
  # Backend and parallellisation -----------------------------------------------
  
  # Assign objects that should be accessible everywhere to the familiar global
  # environment. Note that .assign_data_to_backend will also start backend
  # server processes.
  .assign_settings_to_global(settings = settings)
  .assign_file_paths_to_global(file_paths = file_paths)
  .assign_project_info_to_global(project_info = project_info)
  .assign_outcome_info_to_global(outcome_info = outcome_info)
  .assign_data_to_backend(
    data = data,
    backend_type = settings$run$backend_type,
    server_port = settings$run$server_port
  )

  if (settings$run$parallel) {
    # Make sure that everything is present on the cluster.
    cl <- .update_cluster(cl = cl, assign = "all")
  } 
  
  # Clean familiar environment on exit. This is run last to avoid cleaning up
  # the familiar environment prior to shutting down the socket server process.
  on.exit(.clean_familiar_environment(), add = TRUE)

  experiment_data <- set_experiment_data(
    x = experiment_data,
    project_id = project_info$project_id,
    experiment_setup = experiment_setup,
    iteration_list = project_info$iter_list
  )
  
  if (.stop_after %in% c("setup")) {
    return(experiment_data)
  }

  # Setup tasks
  if (.stop_after == "training") {
    tasks <- .generate_trainer_tasks(
      experiment_data = experiment_data,
      optimisation_determine_vimp = settings$hpo$hpo_determine_vimp,
      vimp_methods = settings$vimp$vimp_methods,
      learners = settings$mb$learners,
      file_paths = file_paths
    )
    
  } else if (.stop_after == "vimp") {
    tasks <- .generate_vimp_tasks(
      experiment_data = experiment_data,
      vimp_methods = settings$vimp$vimp_methods,
      file_paths = file_paths
    )
    
  } else if (.stop_after == "preprocessing") {
    tasks <- c(
      .generate_vimp_data_preprocessing_tasks(
        experiment_data = experiment_data,
        file_paths = file_paths
      ),
      .generate_learner_data_preprocessing_tasks(
        experiment_data = experiment_data,
        file_paths = file_paths
      )
    )
    
  } else {
    tasks <- .generate_evaluation_tasks(
      experiment_data = experiment_data,
      optimisation_determine_vimp = settings$hpo$hpo_determine_vimp,
      vimp_methods = settings$vimp$vimp_methods,
      learners = settings$mb$learners,
      pool_only = settings$eval$pool_only,
      file_paths = file_paths
    )
  }

  # Select and sort unique tasks.
  tasks <- .sort_tasks(tasks)
  
  # Pre-processing -------------------------------------------------------------
  .run_preprocessing(
    cl = cl,
    tasks = tasks,
    experiment_data = experiment_data,
    settings = settings,
    outcome_info = outcome_info,
    file_paths = file_paths,
    verbose = verbose
  )
  
  # Check if the process should be stopped at this point.
  if (.stop_after %in% c("preprocessing")) {
    return(set_experiment_data(
      x = experiment_data,
      feature_info = get_feature_info_from_backend(
        data_id = waiver(),
        run_id = waiver()
      )
    ))
  }
  
  # Variable importance --------------------------------------------------------
  .run_variable_importance_computation(
    cl = cl,
    tasks = tasks,
    experiment_data = experiment_data,
    settings = settings,
    outcome_info = outcome_info,
    file_paths = file_paths,
    verbose = verbose
  )
  
  # Check if the process should be stopped at this point.
  if (.stop_after %in% c("vimp")) {
    feature_info <- NULL
    if (!is_empty(tasks$feature_info)) {
      feature_info <- get_feature_info_from_backend(
        data_id = waiver(),
        run_id = waiver()
      )
    }
    
    vimp_hyperparameters <- NULL
    if (!is_empty(tasks$hyperparameters_vimp)) {
      vimp_hyperparameters <- lapply(
        tasks$hyperparameters_vimp,
        function(x) readRDS(x@file)
      )
    }
    
    vimp_tables <- NULL
    if (!is_empty(tasks$vimp)) {
      vimp_tables <- lapply(
        tasks$vimp,
        function(x) readRDS(x@file)
      )
    }
    
    experiment_data <- set_experiment_data(
      x = experiment_data,
      feature_info = feature_info,
      vimp_hyperparameter_list = vimp_hyperparameters,
      vimp_table_list = vimp_tables
    )
    
    return(experiment_data)
  }
  
  # Training -------------------------------------------------------------------
  
  .run_learner(
    cl = cl,
    tasks = tasks,
    experiment_data = experiment_data,
    settings = settings,
    outcome_info = outcome_info,
    file_paths = file_paths,
    verbose = verbose
  )
  
  # Check if the process should be stopped at this point.
  if (.stop_after %in% c("training")) {
    return(.import_all_familiar_objects(file_paths = file_paths)$familiarModel)
  }
  
  # Explanation and evaluation -------------------------------------------------
  
  .run_evaluation(
    cl = cl,
    tasks = tasks,
    experiment_data = experiment_data,
    settings = settings,
    outcome_info = outcome_info,
    file_paths = file_paths,
    verbose = verbose
  )
  
  # Check if the process should be stopped at this point.
  if (.stop_after %in% c("evaluation") || file_paths$is_temporary) {
    return(.import_all_familiar_objects(file_paths = file_paths))
  }
  
  # Export ---------------------------------------------------------------------
  
  .run_export(
    tasks = tasks,
    file_paths = file_paths,
    verbose = verbose
  )
  
  if (.force_output) {
    return(.import_all_familiar_objects(file_paths = file_paths))
  }
  
  return(invisible(TRUE))
}



#' Pre-compute data assignment
#'
#' @description Creates data assignment.
#'
#' @param experimental_design (**required**) Defines what the experiment looks
#'   like, e.g. `cv(bt(fs,20)+mb,3,2)` for 2 times repeated 3-fold
#'   cross-validation with nested variable importance computation on 20
#'   bootstraps and model-building. The basic workflow components are:
#'
#'   * `fs`: (optional) variable importance computation step. If not explicitly
#'      declared, feature selection will be done just in time for hyperparameter
#'      optimisation.
#'
#'   * `mb`: (required) model building step.
#'
#'   * `ev`: (optional) external validation. If validation batches or cohorts
#'   are present in the dataset (`data`), these should be indicated in the
#'   `validation_batch_id` argument.
#'
#'   The different components are linked using `+`.
#'
#'   Different subsampling methods can be used in conjunction with the basic
#'   workflow components:
#'
#'   * `bs(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. In contrast to `bt`, feature pre-processing parameters and
#'   hyperparameter optimisation are conducted on individual bootstraps.
#'
#'   * `bt(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. Unlike `bs` and other subsampling methods, no separate
#'   pre-processing parameters or optimised hyperparameters will be determined
#'   for each bootstrap.
#'
#'   * `cv(x,n,p)`: (stratified) `n`-fold cross-validation, repeated `p` times.
#'   Pre-processing parameters are determined for each iteration.
#'
#'   * `lv(x)`: leave-one-out-cross-validation. Pre-processing parameters are
#'   determined for each iteration.
#'
#'   * `ip(x)`: imbalance partitioning for addressing class imbalances on the
#'   data set. Pre-processing parameters are determined for each partition. The
#'   number of partitions generated depends on the imbalance correction method
#'   (see the `imbalance_correction_method` parameter).
#'
#'   As shown in the example above, sampling algorithms can be nested.
#'
#'   Though neither variable importance is determined nor models are learned
#'   within `precompute_data_assignment`, the corresponding elements are still
#'   required to prevent issues when using the resulting `experimentData` object
#'   to warm-start the experiments.
#'
#'   The simplest valid experimental design is `fs+mb`. This is the default in
#'   `precompute_data_assignment`, and will simply assign all instances to the
#'   training set.
#'
#' @inheritParams summon_familiar
#' @inheritParams .parse_variable_importance_settings
#' @inheritDotParams .parse_experiment_settings -config
#' @inheritDotParams .parse_setup_settings -config
#' @inheritDotParams .parse_preprocessing_settings -config -data -parallel
#'   -outcome_type
#'
#' @details This is a thin wrapper around `summon_familiar`, and functions like
#'   it, but automatically skips computation of variable importance, learning
#'   and subsequent evaluation steps.
#'
#'   The function returns an `experimentData` object, which can be used to
#'   warm-start other experiments by providing it to the `experiment_data`
#'   argument.
#'
#' @return An `experimentData` object.
#'
#' @export
precompute_data_assignment <- function(
    formula = NULL,
    data = NULL,
    experiment_data = NULL,
    cl = NULL,
    experimental_design = "fs+mb",
    verbose = TRUE,
    ...
) {
  
  # Isolate dots.
  dots <- list(...)
  
  # Drop skip_evaluation_elements, vimp_method and learner if present.
  dots$skip_evaluation_elements <- NULL
  dots$vimp_method <- NULL
  dots$learner <- NULL
  
  # Summon a familiar and compute everything up to variable importance data.
  experiment_data <- do.call(
    summon_familiar,
    args = c(
      list(
        "formula" = formula,
        "data" = data,
        "experiment_data" = experiment_data,
        "cl" = cl,
        "experimental_design" = experimental_design,
        "vimp_method" = "none",
        "learner" = "glm",
        "skip_evaluation_elements" = "all",
        "verbose" = verbose,
        ".stop_after" = "setup"
      ),
      dots
    )
  )
  
  # Extract familiar models.
  return(experiment_data)
}



#' Pre-compute feature information
#'
#' @description Creates data assignment and subsequently extracts feature
#'   information such as normalisation and clustering parameters.
#'
#' @param experimental_design (**required**) Defines what the experiment looks
#'   like, e.g. `cv(bt(fs,20)+mb,3,2)` for 2 times repeated 3-fold
#'   cross-validation with nested variable importance computation on 20
#'   bootstraps and model-building. The basic workflow components are:
#'
#'   * `fs`: (optional) variable importance computation step. If not explicitly
#'      declared, feature selection will be done just in time for hyperparameter
#'      optimisation.
#'
#'   * `mb`: (required) model building step.
#'
#'   * `ev`: (optional) external validation. If validation batches or cohorts
#'   are present in the dataset (`data`), these should be indicated in the
#'   `validation_batch_id` argument.
#'
#'   The different components are linked using `+`.
#'
#'   Different subsampling methods can be used in conjunction with the basic
#'   workflow components:
#'
#'   * `bs(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. In contrast to `bt`, feature pre-processing parameters and
#'   hyperparameter optimisation are conducted on individual bootstraps.
#'
#'   * `bt(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. Unlike `bs` and other subsampling methods, no separate
#'   pre-processing parameters or optimised hyperparameters will be determined
#'   for each bootstrap.
#'
#'   * `cv(x,n,p)`: (stratified) `n`-fold cross-validation, repeated `p` times.
#'   Pre-processing parameters are determined for each iteration.
#'
#'   * `lv(x)`: leave-one-out-cross-validation. Pre-processing parameters are
#'   determined for each iteration.
#'
#'   * `ip(x)`: imbalance partitioning for addressing class imbalances on the
#'   data set. Pre-processing parameters are determined for each partition. The
#'   number of partitions generated depends on the imbalance correction method
#'   (see the `imbalance_correction_method` parameter).
#'
#'   As shown in the example above, sampling algorithms can be nested.
#'
#'   Though neither variable importance is determined nor models are learned
#'   within `precompute_feature_info`, the corresponding elements are still
#'   required to prevent issues when using the resulting `experimentData` object
#'   to warm-start the experiments.
#'
#'   The simplest valid experimental design is `fs+mb`. This is the default in
#'   `precompute_feature_info`, and will determine feature parameters over the
#'   entire dataset.
#'   
#'   This argument is ignored if the `experiment_data` argument is set.
#'
#' @inheritParams summon_familiar
#' @inheritParams .parse_variable_importance_settings
#' @inheritDotParams .parse_experiment_settings -config
#' @inheritDotParams .parse_setup_settings -config
#' @inheritDotParams .parse_preprocessing_settings -config -data -parallel
#'   -outcome_type
#'
#' @details This is a thin wrapper around `summon_familiar`, and functions like
#'   it, but automatically skips computation of variable importance, learning
#'   and subsequent evaluation steps.
#'
#'   The function returns an `experimentData` object, which can be used to
#'   warm-start other experiments by providing it to the `experiment_data`
#'   argument.
#'
#' @return An `experimentData` object.
#'
#' @export
precompute_feature_info <- function(
    formula = NULL,
    data = NULL,
    experiment_data = NULL,
    cl = NULL,
    experimental_design = "fs+mb",
    verbose = TRUE,
    ...
) {
  
  # Isolate dots.
  dots <- list(...)
  
  # Drop skip_evaluation_elements, vimp_method and learner if present.
  dots$skip_evaluation_elements <- NULL
  dots$vimp_method <- NULL
  dots$learner <- NULL
  
  # Summon a familiar and compute everything up to variable importance data.
  experiment_data <- do.call(
    summon_familiar,
    args = c(
      list(
        "formula" = formula,
        "data" = data,
        "experiment_data" = experiment_data,
        "cl" = cl,
        "experimental_design" = experimental_design,
        "vimp_method" = "none",
        "learner" = "glm",
        "skip_evaluation_elements" = "all",
        "verbose" = verbose,
        ".stop_after" = "preprocessing"
      ),
      dots
    )
  )
  
  # Extract familiar models.
  return(experiment_data)
}



#' Pre-compute variable importance
#'
#' @description Creates data assignment, extracts feature information and
#'   subsequently computes variable importance.
#'
#' @param experimental_design (**required**) Defines what the experiment looks
#'   like, e.g. `cv(bt(fs,20)+mb,3,2)` for 2 times repeated 3-fold
#'   cross-validation with nested variable importance computation on 20
#'   bootstraps and model-building. The basic workflow components are:
#'
#'   * `fs`: (required) variable importance computation step. No variable
#'   importances will be prepared if this step is not explicitly used, instead,
#'   feature selection will be done just in time for hyperparameter
#'   optimisation.
#'
#'   * `mb`: (required) model building step. Though models are not learned by
#'   `precompute_vimp`, this element is still required to prevent issues when
#'   using the resulting `experimentData` object to warm-start the experiments.
#'
#'   * `ev`: (optional) external validation. If validation batches or cohorts
#'   are present in the dataset (`data`), these should be indicated in the
#'   `validation_batch_id` argument.
#'
#'   The different components are linked using `+`.
#'
#'   Different subsampling methods can be used in conjunction with the basic
#'   workflow components:
#'
#'   * `bs(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. In contrast to `bt`, feature pre-processing parameters and
#'   hyperparameter optimisation are conducted on individual bootstraps.
#'
#'   * `bt(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. Unlike `bs` and other subsampling methods, no separate
#'   pre-processing parameters or optimised hyperparameters will be determined
#'   for each bootstrap.
#'
#'   * `cv(x,n,p)`: (stratified) `n`-fold cross-validation, repeated `p` times.
#'   Pre-processing parameters are determined for each iteration.
#'
#'   * `lv(x)`: leave-one-out-cross-validation. Pre-processing parameters are
#'   determined for each iteration.
#'
#'   * `ip(x)`: imbalance partitioning for addressing class imbalances on the
#'   data set. Pre-processing parameters are determined for each partition. The
#'   number of partitions generated depends on the imbalance correction method
#'   (see the `imbalance_correction_method` parameter).
#'
#'   As shown in the example above, sampling algorithms can be nested.
#'
#'   The simplest valid experimental design is `fs+mb`. This is the default in
#'   `precompute_vimp`, and will compute variable importance over the entire
#'   dataset.
#'   
#'   This argument is ignored if the `experiment_data` argument is set.
#'
#' @inheritParams summon_familiar
#' @inheritParams .parse_variable_importance_settings
#' @inheritDotParams .parse_experiment_settings -config
#' @inheritDotParams .parse_setup_settings -config
#' @inheritDotParams .parse_preprocessing_settings -config -data -parallel -outcome_type
#' @inheritDotParams .parse_variable_importance_settings -parallel_vimp
#'
#' @details This is a thin wrapper around `summon_familiar`, and functions like
#'   it, but automatically skips learning and subsequent evaluation steps.
#'
#'   The function returns an `experimentData` object, which can be used to
#'   warm-start other experiments by providing it to the `experiment_data`
#'   argument. Variable importance may be retrieved from this object using the
#'   `get_vimp_table` and `aggregate_vimp_table` methods.
#'
#' @return An `experimentData` object.
#'
#' @seealso \code{\link{get_vimp_table}}, \code{\link{aggregate_vimp_table}}
#'
#' @export
precompute_vimp <- function(
    formula = NULL,
    data = NULL,
    experiment_data = NULL,
    cl = NULL,
    experimental_design = "fs+mb",
    vimp_method = NULL,
    vimp_method_parameter = NULL,
    verbose = TRUE,
    ...
) {
  
  # Check that a single learner is present.
  vimp_method <- .parse_arg(
    x_config = NULL,
    x_var = vimp_method,
    var_name = "vimp_method",
    type = "character_list",
    optional = FALSE
  )
  
  # Hyperparameters may be interpreted as belonging to the specified learner.
  vimp_method_parameter <- .parse_arg(
    x_config = NULL,
    x_var = vimp_method_parameter,
    var_name = "vimp_method_parameter",
    type = "list",
    optional = TRUE,
    default = list()
  )
  
  # Encode hyperparameter as expected by parsing it to a nested list.
  if (length(vimp_method_parameter) > 0L && length(vimp_method) == 1L) {
    if (is.null(vimp_method_parameter[[vimp_method]])) {
      vimp_method_parameter_list <- list()
      vimp_method_parameter_list[[vimp_method]] <- vimp_method_parameter
      vimp_method_parameter <- vimp_method_parameter_list
    }
  }
  
  # Isolate dots.
  dots <- list(...)
  
  # Drop skip_evaluation_elements, and learner if present.
  dots$skip_evaluation_elements <- NULL
  dots$learner <- NULL
  
  # Summon a familiar and compute everything up to variable importance data.
  experiment_data <- do.call(
    summon_familiar,
    args = c(
      list(
        "formula" = formula,
        "data" = data,
        "experiment_data" = experiment_data,
        "cl" = cl,
        "experimental_design" = experimental_design,
        "vimp_method" = vimp_method,
        "vimp_method_parameter" = vimp_method_parameter,
        "learner" = "glm",
        "skip_evaluation_elements" = "all",
        "verbose" = verbose,
        ".stop_after" = "vimp"
      ),
      dots
    )
  )
  
  # Extract familiar models.
  return(experiment_data)
}



#' Create models using end-to-end machine learning
#'
#' @description Train models using familiar. Evaluation is not performed.
#'
#' @param experimental_design (**required**) Defines what the experiment looks
#'   like, e.g. `cv(bt(fs,20)+mb,3,2)` for 2 times repeated 3-fold
#'   cross-validation with nested variable importance computation on 20
#'   bootstraps and model-building. The basic workflow components are:
#'
#'   * `fs`: (optional) variable importance computation step. If not explicitly
#'   declared, feature selection will be done just in time for hyperparameter
#'   optimisation.
#'
#'   * `mb`: (required) model building step.
#'
#'   * `ev`: (optional) external validation. Setting this is not required for
#'   `train_familiar`, but if validation batches or cohorts are present in the
#'   dataset (`data`), these should be indicated in the `validation_batch_id`
#'   argument.
#'
#'   The different components are linked using `+`.
#'
#'   Different subsampling methods can be used in conjunction with the basic
#'   workflow components:
#'
#'   * `bs(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. In contrast to `bt`, feature pre-processing parameters and
#'   hyperparameter optimisation are conducted on individual bootstraps.
#'
#'   * `bt(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. Unlike `bs` and other subsampling methods, no separate
#'   pre-processing parameters or optimised hyperparameters will be determined
#'   for each bootstrap.
#'
#'   * `cv(x,n,p)`: (stratified) `n`-fold cross-validation, repeated `p` times.
#'   Pre-processing parameters are determined for each iteration.
#'
#'   * `lv(x)`: leave-one-out-cross-validation. Pre-processing parameters are
#'   determined for each iteration.
#'
#'   * `ip(x)`: imbalance partitioning for addressing class imbalances on the
#'   data set. Pre-processing parameters are determined for each partition. The
#'   number of partitions generated depends on the imbalance correction method
#'   (see the `imbalance_correction_method` parameter).
#'
#'   As shown in the example above, sampling algorithms can be nested.
#'
#'   The simplest valid experimental design is `fs+mb`. This is the default in
#'   `train_familiar`, and will create one model for each variable importance
#'   method in `vimp_method`. To create more models, a subsampling method should
#'   be introduced, e.g. `bs(fs+mb,20)` to create 20 models based on bootstraps
#'   of the data.
#'   
#'   This argument is ignored if the `experiment_data` argument is set.
#'
#' @param learner (**required**) Name of the learner used to develop a model. A
#'   sizeable number learners is supported in `familiar`. Please see the
#'   vignette on learners for more information concerning the available
#'   learners. Unlike the `summon_familiar` function, `train_familiar` only
#'   allows for a single learner.
#' @param hyperparameter (*optional*) List, or nested list containing
#'   hyperparameters for learners. If a nested list is provided, each sublist
#'   should have the name of the learner method it corresponds to, with list
#'   elements being named after the intended hyperparameter, e.g.
#'   \code{"glm_logistic"=list("sign_size"=3)}
#'
#'   All learners have hyperparameters. Please refer to the vignette on learners
#'   for more details. If no parameters are provided, sequential model-based
#'   optimisation is used to determine optimal hyperparameters.
#'
#'   Hyperparameters provided by the user are never optimised. However, if more
#'   than one value is provided for a single hyperparameter, optimisation will
#'   be conducted using these values.
#'
#' @inheritParams summon_familiar
#' @inheritDotParams .parse_experiment_settings -config
#' @inheritDotParams .parse_setup_settings -config
#' @inheritDotParams .parse_preprocessing_settings -config -data -parallel -outcome_type
#' @inheritDotParams .parse_variable_importance_settings -config -data -parallel -outcome_type
#' @inheritDotParams .parse_model_development_settings -config -data -parallel -outcome_type
#' @inheritDotParams .parse_hyperparameter_optimisation_settings -config -parallel -outcome_type
#'
#' @details This is a thin wrapper around `summon_familiar`, and functions like
#'   it, but automatically skips all evaluation steps. Only a single learner is
#'   allowed.
#'
#' @return One or more familiarModel objects.
#'   
#' @export
train_familiar <- function(
    formula = NULL,
    data = NULL,
    experiment_data = NULL,
    cl = NULL,
    experimental_design = "fs+mb",
    learner = NULL,
    hyperparameter = NULL,
    verbose = TRUE,
    ...
) {
  
  # Check that a single learner is present.
  learner <- .parse_arg(
    x_config = NULL,
    x_var = learner,
    var_name = "learner",
    type = "character",
    optional = FALSE
  )
  
  # Hyperparameters may be interpreted as belonging to the specified learner.
  hyperparameter <- .parse_arg(
    x_config = NULL,
    x_var = hyperparameter,
    var_name = "hyperparameter",
    type = "list",
    optional = TRUE,
    default = list()
  )
  
  # Encode hyperparameter as expected by parsing it to a nested list.
  if (length(hyperparameter) > 0L && is.null(hyperparameter[[learner]])) {
    hyperparameter_list <- list()
    hyperparameter_list[[learner]] <- hyperparameter
    hyperparameter <- hyperparameter_list
  }
  
  # Isolate dots.
  dots <- list(...)
  
  # Drop skip_evaluation_elements, project_dir, and experiment_dir if present.
  dots$skip_evaluation_elements <- NULL
  dots$project_dir <- NULL
  dots$experiment_dir <- NULL
  
  # Summon a familiar.
  familiar_models <- do.call(
    summon_familiar,
    args = c(
      list(
        "formula" = formula,
        "data" = data,
        "experiment_data" = experiment_data,
        "cl" = cl,
        "experimental_design" = experimental_design,
        "learner" = learner,
        "hyperparameter" = hyperparameter,
        "experiment_dir" = NULL,
        "project_dir" = NULL,
        "skip_evaluation_elements" = "all",
        "verbose" = verbose,
        ".stop_after" = "training"
      ),
      dots
    )
  )
  
  # Extract familiar models.
  return(familiar_models)
}



.is_absolute_path <- function(x) {
  return(dir.exists(paste0(
    unlist(strsplit(x, split = .Platform$file.sep))[1L],
    .Platform$file.sep
  )))
}



.load_configuration_file <- function(config, config_id = 1L) {
  
  if (!is.null(config)) {
    if (is.character(config)) {
      if (length(config) > 1L) {
        ..error(
          paste0(
            "Configuration: the path to the configuration file is expected ",
            "to be a single character string. Multiple strings were found."
          ),
          error_class = "input_argument_error"
        )
      }
      
      # Normalise file paths
      config <- normalizePath(config)
      
      # Check that the xml2 package is installed
      require_package("xml2", "to configure familiar using a configuration file")
      
      # Read xml file, parse to list and remove comments
      config <- xml2::as_list(xml2::read_xml(config))[[1L]][[config_id]]
      config <- .clean_configuration_comments(config = config)
      
    } else {
      if (!is.list(config) || !is.recursive(config)) {
        ..error("Configuration: the input configuration data is not a list of lists.")
      }
    }
  } else {
    config <- NULL
  }
  
  return(config)
}



.clean_configuration_comments <- function(config) {
  
  cleaning_cycle <- function(conf_list) {
    # Nested function for iterating through configuration list
    
    # If the current position is no longer a list, skip the following steps
    if (is.list(conf_list)) {
      
      # Retain only those list entries that are not comments "[ comment ]".
      conf_list <- Filter(Negate(function(x) ((is.character(x[1L]) & x[1L] == "[ comment ]"))), conf_list)
      
      # Go one level deeper by applying this function to the list entries of the current list
      conf_list <- lapply(conf_list, cleaning_cycle)
    }
    
    return(conf_list)
  }
  
  # Remove comments by running the iterative cleaningCycle script
  config_new <- cleaning_cycle(conf_list = config)
  
  return(config_new)
}



#' Create an empty xml configuration file
#'
#' This function creates an empty configuration xml file in the directory
#' specified by `dir_path`. This provides an alternative to the use of input
#' arguments for familiar.
#'
#' @param dir_path Path to the directory where the configuration file should be
#'  created. The directory should exist, and no file named `config.xml` should
#'  be present.
#'
#' @return Nothing. A file named `config.xml` is created in the directory
#' indicated by `dir_path`.
#' @export
#'
#' @examples
#' \dontrun{
#' # Creates a config.xml file in the working directory
#' get_xml_config(dir_path=getwd())
#' }
#' @md
#' @keywords IO
get_xml_config <- function(dir_path) {
  
  file.copy(
    from = system.file("config.xml", package = "familiar"),
    to = dir_path,
    overwrite = FALSE
  )
  
  return(invisible(TRUE))
}



.assign_settings_to_global <- function(settings) {
  # Assign settings into the familiar global, environment
  assign(
    x = "settings",
    value = settings,
    envir = familiar_global_env
  )
  
  return(invisible(TRUE))
}



get_settings <- function() {
  
  # Retrieve settings from the backend
  if (exists("familiar_global_env")) {
    if (exists("settings", where = familiar_global_env)) {
      data_env <- familiar_global_env
      
    } else if (exists("settings", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
      
    } else {
      ..error("Settings not found in backend.")
    }
    
  } else if (exists("settings", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
    
  } else {
    ..error("Settings not found in backend.")
  }
  
  return(get("settings", envir = data_env))
}



.assign_file_paths_to_global <- function(file_paths) {
  # Put file_paths into the familiar environment
  assign(
    x = "file_paths",
    value = file_paths,
    envir = familiar_global_env
  )
  
  return(invisible(TRUE))
}



get_file_paths <- function() {
  
  # Retrieve the paths to files and directories
  if (exists("familiar_global_env")) {
    if (exists("file_paths", where = familiar_global_env)) {
      data_env <- familiar_global_env
      
    } else if (exists("file_paths", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
      
    } else {
      ..error("File paths were not found in backend.")
    }
    
  } else if (exists("file_paths", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
    
  } else {
    ..error("File paths were not found in backend.")
  }
  
  return(get("file_paths", envir = data_env))
}



.assign_project_info_to_global <- function(project_info) {
  # Put project_info_list into the familiar environment
  assign(
    x = "project_info_list",
    value = project_info,
    envir = familiar_global_env
  )
  
  return(invisible(TRUE))
}



get_project_list <- function() {
  
  # Retrieve the project list
  if (exists("familiar_global_env")) {
    if (exists("project_info_list", where = familiar_global_env)) {
      data_env <- familiar_global_env
      
    } else if (exists("project_info_list", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
      
    } else {
      ..error("Project list not found in backend.")
    }
  } else if (exists("project_info_list", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
    
  } else {
    ..error("Project list not found in backend.")
  }
  
  return(get("project_info_list", envir = data_env))
}



.import_all_familiar_objects <- function(file_paths) {
  familiar_list <- list()
  
  # Find familiarModel files
  model_files <- list.files(
    path = file_paths$mb_dir,
    pattern = "model.RDS",
    full.names = TRUE
  )
  familiar_list$familiarModel <- load_familiar_object(model_files)
  
  # Find familiarData files
  data_files <- list.files(
    path = file_paths$fam_data_dir,
    pattern = "data.RDS",
    full.names = TRUE
  )
  familiar_list$familiarData <- load_familiar_object(data_files)
  
  # Find familiarCollection files
  coll_files <- list.files(
    path = file_paths$fam_coll_dir,
    pattern = "collection.RDS",
    full.names = TRUE
  )
  familiar_list$familiarCollection <- load_familiar_object(coll_files)
  
  return(familiar_list)
}



.clean_familiar_environment <- function() {
  # Cleans all objects assigned to the familiar global environment.
  if (exists("familiar_global_env")) {
    rm(
      list = ls(envir = familiar_global_env),
      envir = familiar_global_env
    )
  }
  
  return(invisible(TRUE))
}

# Create test dataset.
data <- familiar:::test_create_good_data(outcome_type = "binomial")

weights <- familiar:::create_instance_weights(
  data = data,
  method = "inverse_number_of_samples")

# Check that weights sum to the number of classes (2)
testthat::test_that("Weights sum to the number of classes.", {
  testthat::expect_equal(sum(weights), 2)
})

weights <- familiar:::create_instance_weights(
  data = data,
  method = "inverse_number_of_samples",
  normalisation = "average_one")

# Check that the mean weight is 1.0.
testthat::test_that("Mean weight is 1.0.", {
  testthat::expect_equal(mean(weights), 1.0)
})

weights <- familiar:::create_instance_weights(
  data = data,
  method = "inverse_number_of_samples",
  normalisation = "sum_one")

# Check that weights sum to the number of classes (2)
testthat::test_that("Weights sum to 1.0.", {
  testthat::expect_equal(sum(weights), 1.0)
})

# Check that with beta equal to 0.9, weights between classes are similar.
weights <- familiar:::create_instance_weights(
  data = data,
  method = "effective_number_of_samples",
  beta = 0.9)

testthat::test_that("Weights between classes are almost the same.", {
  testthat::expect_equal(abs(diff(unique(weights))) < 10^-4, TRUE)
})

# Check that with beta equal to 0.99999, weights between classes are similar to
# inverse number of samples.
inv_weights <- familiar:::create_instance_weights(
  data = data,
  method = "inverse_number_of_samples")

weights <- familiar:::create_instance_weights(
  data = data,
  method = "effective_number_of_samples",
  beta = 0.99999)

testthat::test_that("Weights are similar to inverse number of samples.", {
  testthat::expect_equal(all(abs(inv_weights - weights) < 10^-4), TRUE)
})

weights <- familiar:::create_instance_weights(
  data = data,
  method = "effective_number_of_samples",
  beta = 0.9,
  normalisation = "average_one")

# Check that the mean weight is 1.0.
testthat::test_that("Mean weight is 1.0.", {
  testthat::expect_equal(mean(weights), 1.0)
})

weights <- familiar:::create_instance_weights(
  data = data,
  method = "effective_number_of_samples",
  beta = 0.9,
  normalisation = "sum_one")

# Check that weights sum to the number of classes (2)
testthat::test_that("Weights sum to 1.0.", {
  testthat::expect_equal(sum(weights), 1.0)
})


weights <- familiar:::create_instance_weights(
  data = data,
  method = "effective_number_of_samples",
  beta = 0.9)

comp_weights <- familiar:::create_instance_weights(
  data = data,
  method = "effective_number_of_samples",
  beta = familiar:::..compute_effective_number_of_samples_beta(-1))

testthat::test_that("Weights are the same.", {
  testthat::expect_equal(all(weights - comp_weights == 0.0), TRUE)
})

testthat::skip("Skip hyperparameter optimisation, unless manual.")

# Check that variable importance functions correctly.
model <- familiar::train_familiar(
  data = data,
  vimp_method = "mrmr",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "cell_malignancy",
  sample_id_column = "id",
  class_levels = c("benign", "malignant"),
  verbose = FALSE
)

# Check that sample weighting is inverse_number_of_samples (default).
testthat::test_that("Default method is inverse_number_of_samples.", {
  testthat::expect_equal(
    as.character(model@hyperparameters$sample_weighting),
    "inverse_number_of_samples")
})

# Check that effective_number_of_samples functions.
model <- familiar::train_familiar(
  data = data,
  vimp_method = "mrmr",
  learner = "glm_logistic",
  hyperparameter = list("sample_weighting" = "effective_number_of_samples"),
  outcome_type = "binomial",
  outcome_column = "cell_malignancy",
  sample_id_column = "id",
  class_levels = c("benign", "malignant"),
  verbose = FALSE
)

testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

data <- familiar:::test_create_good_data(outcome_type = "binomial", to_data_object = FALSE)
data[101L:150L, "batch_id" := "test"]


# With unpooled collections ----------------------------------------------------

# Set evaluate_top_level_only to FALSE evaluate underlying data divisions.
results <- familiar::summon_familiar(
  data = data,
  experimental_design = "cv(bs(mb, 2), 3) + ev",
  evaluate_top_level_only = FALSE,
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  validation_batch_id = "test",
  vimp_method = "mim",
  learner = "glm_logistic",
  estimation_type = "point",
  shap_max_iterations = 10L,
  iteration_seed = 9L,
  parallel = FALSE,
  verbose = debug_flag
)


testthat::test_that("cv-only with nested bootstraps experiment is correctly created", {
  testthat::expect_length(results$familiarModel, 6L)
  testthat::expect_length(results$familiarData, 12L)
  testthat::expect_length(results$familiarCollection, 4L)
  testthat::expect_setequal(
    sapply(results$familiarData, function(x) (x@name)),
    c("development", "internal_validation", "external_validation")
  )
  
  pooled_collection <- results$familiarCollection[
    sapply(results$familiarCollection, function(x) (endsWith(x@name, "pooled_collection")))
  ][[1L]]
  prediction_data <- familiar::export_prediction_data(pooled_collection)
  prediction_data <- prediction_data$classification[[1L]]@data
  ext_val_samples <- prediction_data[data_set == "ext. validation"]$sample_id
  int_val_samples <- prediction_data[data_set == "int. validation"]$sample_id
  dev_samples <- prediction_data[data_set == "development"]$sample_id
  
  testthat::expect_length(intersect(ext_val_samples, dev_samples), 0L)
  testthat::expect_length(intersect(ext_val_samples, int_val_samples), 0L)
  testthat::expect_length(ext_val_samples, nrow(data[batch_id == "test"]))
  testthat::expect_length(int_val_samples, nrow(data[batch_id == "basic"]))
  testthat::expect_lt(length(dev_samples), nrow(data[batch_id == "basic"]))
  
  cv_1_collection <- results$familiarCollection[
    sapply(results$familiarCollection, function(x) (endsWith(x@name, "2_1_collection")))
  ][[1L]]
  prediction_data <- familiar::export_prediction_data(cv_1_collection)
  prediction_data <- prediction_data$classification[[1L]]@data
  ext_val_samples_1 <- prediction_data[data_set == "ext. validation"]$sample_id
  int_val_samples_1 <- prediction_data[data_set == "int. validation"]$sample_id
  dev_samples_1 <- prediction_data[data_set == "development"]$sample_id
  
  testthat::expect_length(intersect(ext_val_samples_1, dev_samples_1), 0L)
  testthat::expect_length(intersect(ext_val_samples_1, int_val_samples_1), 0L)
  testthat::expect_length(intersect(int_val_samples_1, dev_samples_1), 0L)
  testthat::expect_length(ext_val_samples_1, nrow(data[batch_id == "test"]))
  testthat::expect_lt(length(int_val_samples_1) + length(dev_samples_1), nrow(data[batch_id == "basic"]))
  
  cv_2_collection <- results$familiarCollection[
    sapply(results$familiarCollection, function(x) (endsWith(x@name, "2_2_collection")))
  ][[1L]]
  prediction_data <- familiar::export_prediction_data(cv_2_collection)
  prediction_data <- prediction_data$classification[[1L]]@data
  ext_val_samples_2 <- prediction_data[data_set == "ext. validation"]$sample_id
  int_val_samples_2 <- prediction_data[data_set == "int. validation"]$sample_id
  dev_samples_2 <- prediction_data[data_set == "development"]$sample_id
  
  testthat::expect_length(intersect(ext_val_samples_2, dev_samples_2), 0L)
  testthat::expect_length(intersect(ext_val_samples_2, int_val_samples_2), 0L)
  testthat::expect_length(intersect(int_val_samples_2, dev_samples_2), 0L)
  testthat::expect_length(ext_val_samples_2, nrow(data[batch_id == "test"]))
  testthat::expect_lt(length(int_val_samples_2) + length(dev_samples_2), nrow(data[batch_id == "basic"]))
  
  cv_3_collection <- results$familiarCollection[
    sapply(results$familiarCollection, function(x) (endsWith(x@name, "2_3_collection")))
  ][[1L]]
  prediction_data <- familiar::export_prediction_data(cv_3_collection)
  prediction_data <- prediction_data$classification[[1L]]@data
  ext_val_samples_3 <- prediction_data[data_set == "ext. validation"]$sample_id
  int_val_samples_3 <- prediction_data[data_set == "int. validation"]$sample_id
  dev_samples_3 <- prediction_data[data_set == "development"]$sample_id
  
  testthat::expect_length(intersect(ext_val_samples_3, dev_samples_3), 0L)
  testthat::expect_length(intersect(ext_val_samples_3, int_val_samples_3), 0L)
  testthat::expect_length(intersect(int_val_samples_3, dev_samples_3), 0L)
  testthat::expect_length(ext_val_samples_3, nrow(data[batch_id == "test"]))
  testthat::expect_lt(length(int_val_samples_3) + length(dev_samples_3), nrow(data[batch_id == "basic"]))
  
  # Internal validation folds between experiments do not overlap.
  testthat::expect_length(intersect(int_val_samples_1, int_val_samples_2), 0L)
  testthat::expect_length(intersect(int_val_samples_1, int_val_samples_3), 0L)
  testthat::expect_length(intersect(int_val_samples_2, int_val_samples_3), 0L)
  
  # External validation folds between experiments are the same.
  testthat::expect_setequal(ext_val_samples_1, ext_val_samples_2)
  testthat::expect_setequal(ext_val_samples_1, ext_val_samples_3)
  testthat::expect_setequal(ext_val_samples_2, ext_val_samples_3)
  
  # Pooled performance data.
  performance_data <- familiar::export_model_performance(
    results$familiarCollection[[4L]],
    aggregate_results = FALSE
  )[[1L]]@data
  
  # Expect that the values are not the same.
  dev_values <- performance_data[data_set == "development"]$value
  int_values <- performance_data[data_set == "int. validation"]$value
  ext_values <- performance_data[data_set == "ext. validation"]$value
  testthat::expect_length(dev_values, 6L)
  testthat::expect_length(int_values, 6L)
  testthat::expect_length(ext_values, 6L)
  testthat::expect_false(setequal(dev_values, int_values))
  testthat::expect_false(setequal(dev_values, ext_values))
  
  for (ii in seq_len(3L)) {
    # Get performance for individual folds.
    performance_data <- familiar::export_model_performance(
      results$familiarCollection[[ii]],
      aggregate_results = FALSE
    )[[1L]]@data
    
    # Expect that the values are not the same. There are two models per fold.
    dev_values <- performance_data[data_set == "development"]$value
    int_values <- performance_data[data_set == "int. validation"]$value
    ext_values <- performance_data[data_set == "ext. validation"]$value
    testthat::expect_length(dev_values, 2L)
    testthat::expect_length(int_values, 2L)
    testthat::expect_length(ext_values, 2L)
    testthat::expect_false(setequal(dev_values, int_values))
    testthat::expect_false(setequal(dev_values, ext_values))
  }
})



# Set evaluate_top_level_only to FALSE evaluate underlying data divisions.
results <- familiar::summon_familiar(
  data = data[c(1L:30L, 101L:150L),],
  experimental_design = "lv(mb) + ev",
  evaluate_top_level_only = FALSE,
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  validation_batch_id = "test",
  vimp_method = "mim",
  learner = "glm_logistic",
  estimation_type = "point",
  shap_max_iterations = 10L,
  iteration_seed = 9L,
  parallel = FALSE,
  verbose = FALSE
)


testthat::test_that("loocv with external validation experiment is correctly created", {
  testthat::expect_length(results$familiarModel, 30L)
  testthat::expect_length(results$familiarData, 93L)
  testthat::expect_length(results$familiarCollection, 31L)
  testthat::expect_setequal(
    sapply(results$familiarData, function(x) (x@name)),
    c("development", "internal_validation", "external_validation")
  )
  
  # This is the pooled data.
  performance_data <- familiar::export_model_performance(
    results$familiarCollection[[31L]],
    aggregate_results = FALSE
  )[[1L]]@data
  
  # Expect that the values are not the same. Note that the detail-level is
  # automatically changed to ensemble because of the limited number of values
  # in the cross-validation.
  dev_values <- performance_data[data_set == "development"]$value
  int_values <- performance_data[data_set == "int. validation"]$value
  ext_values <- performance_data[data_set == "ext. validation"]$value
  testthat::expect_length(dev_values, 1L)
  testthat::expect_length(int_values, 1L)
  testthat::expect_length(dev_values, 1L)
  testthat::expect_false(setequal(dev_values, int_values))
  testthat::expect_false(setequal(dev_values, ext_values))
  
  # For individual folds of the leave-one-out cross-validation set. Note that
  # performance data for internal validation data should be missing -- you
  # cannot compute performance for internal validation (1 sample).
  for (ii in seq_len(30L)) {
    performance_data <- familiar::export_model_performance(
      results$familiarCollection[[ii]],
      aggregate_results = FALSE
    )[[1L]]@data
    
    # Expect that the values are not the same. Note that the detail-level is
    # automatically changed to ensemble because of the limited number of values
    # in the cross-validation.
    dev_values <- performance_data[data_set == "development"]$value
    int_values <- performance_data[data_set == "int. validation"]$value
    ext_values <- performance_data[data_set == "ext. validation"]$value
    testthat::expect_length(dev_values, 1L)
    testthat::expect_length(int_values, 0L)
    testthat::expect_length(dev_values, 1L)
    testthat::expect_false(setequal(dev_values, int_values))
    testthat::expect_false(setequal(dev_values, ext_values))
  }
})

testthat::skip_on_cran()
testthat::skip_on_ci()

# binomial ---------------------------------------------------------------------
# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE
)

# Check that train_familiar functions correctly.
model <- familiar::train_familiar(
  data = data,
  vimp_method = "mrmr",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Logistic model can be trained using train_familiar", {
  testthat::expect_s4_class(model, "familiarGLM")
  testthat::expect_true(familiar:::model_is_trained(model))
  testthat::expect_s3_class(summary(model), "summary.glm")
  testthat::expect_false(is.null(familiar::coef(model)))
  
  # Assert that between 1 and 6 features are present, aside from the intercept.
  testthat::expect_gte(length(model@model_features), 1L)
  testthat::expect_lte(length(model@model_features), 6L)
  testthat::expect_lte(length(model@model_features), model@hyperparameters$sign_size)
})

# multinomial ------------------------------------------------------------------
# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "multinomial",
  to_data_object = FALSE
)

# Check that train_familiar functions correctly.
model <- familiar::train_familiar(
  data = data,
  vimp_method = "none",
  learner = "glm_multinomial",
  outcome_type = "multinomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green", "blue"),
  verbose = FALSE
)

testthat::test_that("Logistic model can be trained using train_familiar", {
  testthat::expect_s4_class(model, "familiarGLM")
  testthat::expect_true(familiar:::model_is_trained(model))
  testthat::expect_s3_class(summary(model), "summary.multinom")
  testthat::expect_false(is.null(familiar::coef(model)))
  testthat::expect_false(is.null(familiar::vcov(model)))
})

# continuous -------------------------------------------------------------------

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "continuous",
  to_data_object = FALSE)

# Check that train_familiar functions correctly.
model <- familiar::train_familiar(
  data = data,
  vimp_method = "mrmr",
  learner = "glm_gaussian",
  outcome_type = "continuous",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  verbose = FALSE
)

testthat::test_that("Gaussian model can be trained using train_familiar", {
  testthat::expect_s4_class(model, "familiarGLM")
  testthat::expect_true(familiar:::model_is_trained(model))
  testthat::expect_s3_class(summary(model), "summary.glm")
  testthat::expect_false(is.null(familiar::coef(model)))
})

# survival ---------------------------------------------------------------------

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "survival",
  to_data_object = FALSE)

# Check that train_familiar functions correctly.
model <- familiar::train_familiar(
  data = data,
  vimp_method = "none",
  learner = "cox",
  outcome_type = "survival",
  outcome_column = c("outcome_time", "outcome_event"),
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  verbose = FALSE
)

testthat::test_that("Cox proportional hazards model can be trained using train_familiar", {
  testthat::expect_s4_class(model, "familiarCoxPH")
  testthat::expect_true(familiar:::model_is_trained(model))
  testthat::expect_s3_class(summary(model), "summary.coxph")
  testthat::expect_false(is.null(familiar::coef(model)))
  testthat::expect_false(is.null(familiar::vcov(model)))
})

# Use experiment data ----------------------------------------------------------
# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE)

# Create data assignment.
experiment_data <- familiar::precompute_data_assignment(
  data = data,
  experimental_design = "bt(fs+mb,5)",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

# Check that train_familiar functions correctly.
model <- familiar::train_familiar(
  data = data,
  experiment_data = experiment_data,
  vimp_method = "mrmr",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Logistic model can be trained using train_familiar", {
  # Assert that 5 models are trained.
  testthat::expect_equal(length(model), 5L)

  # Assert that the project ids match.
  testthat::expect_equal(model[[1]]@project_id, experiment_data@project_id)
})

# Check "none" variable importance method --------------------------------------

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE)

# Check that train_familiar functions correctly.
model <- familiar::train_familiar(
  data = data,
  cluster_method = "none",
  vimp_method = "none",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Assert that all features are used for \"none\" variable importance method.", {
  # Assert that all features are included in the model.
  testthat::expect_equal(length(model@model_features), 6L)
})

# Check "signature_only" variable importance method ----------------------------

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE)

# Check that train_familiar functions correctly.
model <- familiar::train_familiar(
  data = data,
  signature = c("feature_1", "feature_2a"),
  vimp_method = "signature_only",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Assert that all features are used for \"signature_only\" variable importance method.", {
  # Assert that only signature features are included in the model.
  testthat::expect_equal(length(model@model_features), 2L)
  testthat::expect_setequal(
    c("feature_1", "feature_2a"),
    model@model_features
  )
})

# Check interaction between signature and other features -----------------------

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE
)

# Check that train_familiar functions correctly.
model <- familiar::train_familiar(
  data = data,
  signature = c("feature_1", "feature_2a"),
  vimp_method = "mrmr",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Assert that signature features are used in the model.", {
  # Assert that only signature features are included in the model.
  testthat::expect_gte(length(model@model_features), 2L)
  testthat::expect_lte(length(model@model_features), 6L)
  testthat::expect_true(all(c("feature_1", "feature_2a") %in% model@model_features))
  testthat::expect_lte(length(model@model_features), model@hyperparameters$sign_size)
})

testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE


# With external validation -----------------------------------------------------

data <- familiar:::test_create_good_data(outcome_type = "binomial", to_data_object = FALSE)
data[101L:150L, "batch_id" := "test"]

# Training + external validation
results <- familiar::summon_familiar(
  data = data,
  experimental_design = "mb+ev",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  validation_batch_id = "test",
  vimp_method = "mim",
  learner = "glm_logistic",
  estimation_type = "point",
  shap_max_iterations = 10L,
  parallel = FALSE,
  verbose = FALSE
)

testthat::test_that("development + evaluation experiment is correctly created", {
  testthat::expect_length(results$familiarModel, 1L)
  testthat::expect_length(results$familiarData, 2L)
  testthat::expect_equal(results$familiarData[[1L]]@name, "development")
  testthat::expect_equal(results$familiarData[[2L]]@name, "external_validation")

  performance_data <- familiar::export_model_performance(
    results$familiarCollection,
    aggregate_results = FALSE
  )[[1L]]@data
  
  # Expect that the values are not the same.
  dev_values <- performance_data[data_set == "development"]$value
  int_values <- performance_data[data_set == "int. validation"]$value
  ext_values <- performance_data[data_set == "ext. validation"]$value
  testthat::expect_length(dev_values, 1L)
  testthat::expect_length(int_values, 0L)
  testthat::expect_length(ext_values, 1L)
  testthat::expect_false(setequal(dev_values, int_values))
  testthat::expect_false(setequal(dev_values, ext_values))
})


# Internal bootstraps (incomplete) + external validation
results <- familiar::summon_familiar(
  data = data,
  experimental_design = "bt(mb, 3)+ev",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  validation_batch_id = "test",
  vimp_method = "mim",
  learner = "glm_logistic",
  estimation_type = "point",
  shap_max_iterations = 10L,
  parallel = FALSE,
  verbose = debug_flag
)

testthat::test_that("incomplete bootstrap-only + evaluation experiment is correctly created", {
  testthat::expect_length(results$familiarModel, 3L)
  testthat::expect_length(results$familiarData, 2L)
  testthat::expect_equal(results$familiarData[[1L]]@name, "development")
  testthat::expect_equal(results$familiarData[[2L]]@name, "external_validation")

  performance_data <- familiar::export_model_performance(
    results$familiarCollection,
    aggregate_results = FALSE
  )[[1L]]@data
  
  # Expect that the values are not the same.
  dev_values <- performance_data[data_set == "development"]$value
  int_values <- performance_data[data_set == "int. validation"]$value
  ext_values <- performance_data[data_set == "ext. validation"]$value
  testthat::expect_length(dev_values, 3L)
  testthat::expect_length(int_values, 0L)
  testthat::expect_length(ext_values, 3L)
  testthat::expect_false(setequal(dev_values, int_values))
  testthat::expect_false(setequal(dev_values, ext_values))
})


# Internal cross-validation
results <- familiar::summon_familiar(
  data = data,
  experimental_design = "cv(mb, 3)+ev",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  validation_batch_id = "test",
  vimp_method = "mim",
  learner = "glm_logistic",
  estimation_type = "point",
  shap_max_iterations = 10L,
  parallel = FALSE,
  verbose = debug_flag
)

testthat::test_that("cv + evaluation experiment is correctly created", {
  testthat::expect_length(results$familiarModel, 3L)
  testthat::expect_length(results$familiarData, 3L)
  testthat::expect_setequal(
    sapply(results$familiarData, function(x) (x@name)),
    c("development", "internal_validation", "external_validation")
  )
  
  performance_data <- familiar::export_model_performance(
    results$familiarCollection,
    aggregate_results = FALSE
  )[[1L]]@data
  
  # Expect that the values are not the same.
  dev_values <- performance_data[data_set == "development"]$value
  int_values <- performance_data[data_set == "int. validation"]$value
  ext_values <- performance_data[data_set == "ext. validation"]$value
  testthat::expect_length(dev_values, 3L)
  testthat::expect_length(int_values, 3L)
  testthat::expect_length(ext_values, 3L)
  testthat::expect_false(setequal(dev_values, int_values))
  testthat::expect_false(setequal(dev_values, ext_values))
})


# Internal bootstraps (full)
results <- familiar::summon_familiar(
  data = data,
  experimental_design = "bs(mb, 3)+ev",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  validation_batch_id = "test",
  vimp_method = "mim",
  learner = "glm_logistic",
  estimation_type = "point",
  shap_max_iterations = 10L,
  parallel = FALSE,
  verbose = debug_flag
)

testthat::test_that("bootstrap + evaluation experiment is correctly created", {
  testthat::expect_length(results$familiarModel, 3L)
  testthat::expect_length(results$familiarData, 3L)
  testthat::expect_setequal(
    sapply(results$familiarData, function(x) (x@name)),
    c("development", "internal_validation", "external_validation")
  )
  
  performance_data <- familiar::export_model_performance(
    results$familiarCollection,
    aggregate_results = FALSE
  )[[1L]]@data
  
  # Expect that the values are not the same.
  dev_values <- performance_data[data_set == "development"]$value
  int_values <- performance_data[data_set == "int. validation"]$value
  ext_values <- performance_data[data_set == "ext. validation"]$value
  testthat::expect_length(dev_values, 3L)
  testthat::expect_length(int_values, 3L)
  testthat::expect_length(ext_values, 3L)
  testthat::expect_false(setequal(dev_values, int_values))
  testthat::expect_false(setequal(dev_values, ext_values))
})



results <- familiar::summon_familiar(
  data = data[c(1L:30L, 101L:150L),],
  experimental_design = "lv(mb) + ev",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  validation_batch_id = "test",
  vimp_method = "mim",
  learner = "glm_logistic",
  estimation_type = "point",
  shap_max_iterations = 10L,
  parallel = FALSE,
  verbose = debug_flag
)

testthat::test_that("loocv-only experiment is correctly created", {
  testthat::expect_length(results$familiarModel, 30L)
  testthat::expect_length(results$familiarData, 3L)
  testthat::expect_setequal(
    sapply(results$familiarData, function(x) (x@name)),
    c("development", "internal_validation", "external_validation")
  )
  
  performance_data <- familiar::export_model_performance(
    results$familiarCollection,
    aggregate_results = FALSE
  )[[1L]]@data
  
  # Expect that the values are not the same. Note that the detail-level is
  # automatically changed to ensemble because of the limited number of values
  # in the internal validation set (1 per fold.)
  dev_values <- performance_data[data_set == "development"]$value
  int_values <- performance_data[data_set == "int. validation"]$value
  ext_values <- performance_data[data_set == "ext. validation"]$value
  testthat::expect_length(dev_values, 1L)
  testthat::expect_length(int_values, 1L)
  testthat::expect_length(ext_values, 1L)
  testthat::expect_false(setequal(dev_values, int_values))
  testthat::expect_false(setequal(dev_values, ext_values))
})



# Internal cross-validation with nested (full) bootstraps
results <- familiar::summon_familiar(
  data = data,
  experimental_design = "cv(bs(mb, 2), 3) + ev",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  validation_batch_id = "test",
  vimp_method = "mim",
  learner = "glm_logistic",
  estimation_type = "point",
  shap_max_iterations = 10L,
  iteration_seed = 9L,
  parallel = FALSE,
  verbose = debug_flag
)

# Get predicted probabilities for red. The bootstraps might not visit all
# training data. More over the probabilities should generally be different
# because different models are used to predict each sample.
prediction_data <- merge(
  x = results$familiarData[[1L]]@prediction_data[[1L]]@data[, mget(c("sample_id", "red"))],
  y =  results$familiarData[[3L]]@prediction_data[[1L]]@data[, mget(c("sample_id", "red"))],
  by = "sample_id",
  suffixes = c("_dev", "_int"),
  all = FALSE
)

testthat::test_that("cv-only with nested bootstraps experiment is correctly created", {
  testthat::expect_length(results$familiarModel, 6L)
  testthat::expect_length(results$familiarData, 3L)
  testthat::expect_setequal(
    sapply(results$familiarData, function(x) (x@name)),
    c("development", "internal_validation", "external_validation")
  )
  
  performance_data <- familiar::export_model_performance(
    results$familiarCollection,
    aggregate_results = FALSE
  )[[1L]]@data
  
  # Expect that the values are not the same.
  dev_values <- performance_data[data_set == "development"]$value
  int_values <- performance_data[data_set == "int. validation"]$value
  ext_values <- performance_data[data_set == "ext. validation"]$value
  testthat::expect_length(dev_values, 6L)
  testthat::expect_length(int_values, 6L)
  testthat::expect_length(ext_values, 6L)
  testthat::expect_false(setequal(dev_values, int_values))
  testthat::expect_false(setequal(dev_values, ext_values))
  
  # Expect that fewer than 150 samples appear in the training dataset. If this
  # fails, check that the iteration seed correctly generates the same sample
  # set consistently.
  testthat::expect_lt(
    nrow(results$familiarData[[1L]]@prediction_data[[1L]]@data),
    nrow(data[batch_id == "basic"])
  )
  
  # Expect that predicted probabilities are not all the same.
  testthat::expect_false(all(prediction_data$red_dev == prediction_data$red_int))
  
  # Expect that there is no overlap between development and external validation.
  testthat::expect_equal(
    nrow(merge(
      x = results$familiarData[[1L]]@prediction_data[[1L]]@data[, mget(c("sample_id", "red"))],
      y =  results$familiarData[[2L]]@prediction_data[[1L]]@data[, mget(c("sample_id", "red"))],
      by = "sample_id",
      suffixes = c("_dev", "_ext"),
      all = FALSE
    )),
    0L
  )
  
  # Expect that there is no overlap between internal and external development.
  testthat::expect_equal(
    nrow(merge(
      x = results$familiarData[[3L]]@prediction_data[[1L]]@data[, mget(c("sample_id", "red"))],
      y =  results$familiarData[[2L]]@prediction_data[[1L]]@data[, mget(c("sample_id", "red"))],
      by = "sample_id",
      suffixes = c("_int", "_ext"),
      all = FALSE
    )),
    0L
  )
})

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE
)

# Create data assignment object.
experiment_data_assignment <- familiar::precompute_data_assignment(
  data = data,
  experimental_design = "bs(fs+mb,3)",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Assert that data assignment performs correctly", {
  testthat::expect_false(is.null(experiment_data_assignment@experiment_setup))
  testthat::expect_false(is.null(experiment_data_assignment@iteration_list))
  testthat::expect_true(is.null(experiment_data_assignment@feature_info))
  testthat::expect_true(is.null(experiment_data_assignment@vimp_table_list))
})

# Create feature info object.
experiment_feature_info <- familiar::precompute_feature_info(
  data = data,
  experiment_data = experiment_data_assignment,
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Assert that feature info computation performs correctly", {
  # Check that 4 subsets are present, namely generic + complete
  testthat::expect_equal(length(experiment_feature_info@feature_info), 4L)
  
  # Check that there are 6 features.
  testthat::expect_equal(length(experiment_feature_info@feature_info$generic), 6L)

  for (feature_info_set in names(experiment_feature_info@feature_info)) {
    for (feature_info in experiment_feature_info@feature_info[[feature_info_set]]) {
      # Assert that feature info has the correct class.
      testthat::expect_s4_class(feature_info, "featureInfo")

      # Assert that the feature information is complete.
      testthat::expect_equal(
        familiar:::feature_info_complete(feature_info), 
        feature_info_set != "generic"
      )
    }
  }

  # Test contents.
  testthat::expect_true(is.null(experiment_feature_info@vimp_table_list))
  testthat::expect_equal(
    experiment_data_assignment@experiment_setup,
    experiment_feature_info@experiment_setup,
    ignore_attr = TRUE
  )
  testthat::expect_equal(
    experiment_data_assignment@iteration_list,
    experiment_feature_info@iteration_list,
    ignore_attr = TRUE
  )
  testthat::expect_equal(
    experiment_data_assignment@project_id,
    experiment_feature_info@project_id
  )
})

# Create variable importance
experiment_vimp <- familiar::precompute_vimp(
  data = data,
  experiment_data = experiment_feature_info,
  vimp_method = "univariate_regression",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Assert that variable importance computation performs correctly", {
  # Assert that variable importance data are correct.
  testthat::expect_length(experiment_vimp@vimp_table_list, 3L)
  for (vimp_table in experiment_vimp@vimp_table_list) {
    testthat::expect_s4_class(vimp_table, "vimpTable")
  }

  # Test contents.
  testthat::expect_equal(
    experiment_data_assignment@experiment_setup,
    experiment_vimp@experiment_setup,
    ignore_attr = TRUE
  )
  testthat::expect_equal(
    experiment_data_assignment@iteration_list,
    experiment_vimp@iteration_list,
    ignore_attr = TRUE
  )
  testthat::expect_equal(
    experiment_feature_info@feature_info,
    experiment_vimp@feature_info,
    ignore_attr = TRUE
  )
  testthat::expect_equal(
    experiment_data_assignment@project_id,
    experiment_vimp@project_id
  )

  # Test that get_vimp_table produces three tables.
  vimp_table <- familiar::get_vimp_table(experiment_vimp@vimp_table_list)
  testthat::expect_length(vimp_table, 3L)

  # Test that aggregate_vimp_table produces one table.
  vimp_table <- familiar::aggregate_vimp_table(
    experiment_vimp@vimp_table_list, 
    aggregation_method = "borda"
  )
  testthat::expect_s4_class(vimp_table, "vimpTable")
  
  vimp_table <- familiar::get_vimp_table(vimp_table)
  testthat::expect_s3_class(vimp_table, "data.table")
})

# Add additional variable importance methods.
experiment_vimp_3 <- familiar::precompute_vimp(
  data = data,
  experiment_data = experiment_vimp,
  vimp_method = c("univariate_regression", "mim", "mrmr"),
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

testthat::test_that("Addition variable importance computation performs correctly", {
  # Assert that the vimp tables for the univariate_regression method are not
  # altered.
  for (ii in 1L:3L) {
    testthat::expect_equal(
      experiment_vimp_3@vimp_table_list[[ii]],
      experiment_vimp@vimp_table_list[[ii]],
      ignore_attr = TRUE
    )
  }

  # Test that 9 variable importance tables were created.
  testthat::expect_length(
    familiar::get_vimp_table(experiment_vimp_3@vimp_table_list), 9L
  )

  # Test that 3 aggregated variable importance tables are created.
  testthat::expect_length(
    familiar::get_vimp_table(
      familiar::aggregate_vimp_table(experiment_vimp_3@vimp_table_list, "borda")
    ),
    3L
  )

  testthat::expect_equal(
    experiment_vimp@project_id,
    experiment_vimp_3@project_id)
})

# Don't perform tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

# debug flag
debug <- FALSE

# Create datasets
full_data <- familiar:::test_create_good_data("binomial")
full_data@data <- full_data@data[1L:20L]

# Set learner
learner <- "lasso"
  
# Parse hyperparameter list
hyperparameters <- list(
  "lasso" = list(
    "sign_size" = familiar:::get_n_features(full_data),
    "family" = "binomial"
  )
)
            
output <- familiar:::summon_familiar(
  data = full_data,
  experimental_design = "lv(fs+mb)",
  vimp_method = "none",
  cluster_method = "none",
  imputation_method = "simple",
  learner = learner,
  hyperparameter = hyperparameters,
  skip_evaluation_elements = c(
    "auc_data", "calibration_data", "calibration_info",
    "decision_curve_analyis", "feature_expressions", "feature_similarity",
    "fs_vimp", "hyperparameters", "ice_data", "shap",
    "model_vimp", "permutation_vimp", "sample_similarity",
    "risk_stratification_data", "risk_stratification_info", "univariate_analysis"
  ),
  parallel = FALSE,
  verbose = debug
)

testthat::test_that("LOOCV is performed correctly.", {
  
  # Expect that model performance is computed for all data.
  testthat::expect_length(output$familiarCollection@model_performance, 4L)
  testthat::expect_length(output$familiarData, 2L)
  testthat::expect_length(output$familiarData[[1L]]@model_performance, 2L)
  testthat::expect_length(output$familiarData[[2L]]@model_performance, 2L)
  
})

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

verbose <- FALSE

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE
)

# Create data assignment object.
experiment_data_assignment <- familiar::precompute_data_assignment(
  data = data,
  experimental_design = "bs(fs+mb,3)",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

# Create feature info object.
experiment_feature_info <- familiar::precompute_feature_info(
  data = data,
  experiment_data = experiment_data_assignment,
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("feature information is present", {
  testthat::expect_false(familiar:::is_empty(experiment_feature_info@feature_info))
})


# Create variable importance
experiment_vimp <- familiar::precompute_vimp(
  data = data,
  experiment_data = experiment_feature_info,
  vimp_method = "mim",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("variable importance data is present", {
  testthat::expect_false(familiar:::is_empty(experiment_vimp@feature_info))
  testthat::expect_false(familiar:::is_empty(experiment_vimp@vimp_table_list))
})


# Train model
model <- familiar::train_familiar(
  data = data,
  experiment_data = experiment_vimp,
  vimp_method = "mim",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("all models are present", {
  testthat::expect_true(all(sapply(model, familiar:::model_is_trained)))
  testthat::expect_false(any(sapply(model, function(x) (is.null(x@vimp_table)))))
})

# Check without explicit variable importance computation -----------------------
# Create variable importance
experiment_vimp <- familiar::precompute_vimp(
  data = data,
  experimental_design = "bs(mb,3)",
  vimp_method = "mim",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("variable importance data is absent", {
  testthat::expect_null(experiment_vimp@feature_info)
  testthat::expect_null(experiment_vimp@vimp_table_list)
})

# Train model
model <- familiar::train_familiar(
  data = data,
  experiment_data = experiment_vimp,
  vimp_method = "mim",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("all models are present", {
  testthat::expect_true(all(sapply(model, familiar:::model_is_trained)))
  testthat::expect_false(any(sapply(model, function(x) (is.null(x@vimp_table)))))
})


# Check using variable importance from feature selection -----------------------
experiment_vimp <- familiar::precompute_vimp(
  data = data,
  experimental_design = "bs(fs,3)+bs(mb,3)",
  vimp_method = "mim",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("variable importance data is present", {
  testthat::expect_false(familiar:::is_empty(experiment_vimp@feature_info))
  testthat::expect_false(familiar:::is_empty(experiment_vimp@vimp_table_list))
})

# Train model
model <- familiar::train_familiar(
  data = data,
  experiment_data = experiment_vimp,
  optimisation_determine_vimp = FALSE,
  vimp_method = "mim",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("all models are present", {
  testthat::expect_true(all(sapply(model, familiar:::model_is_trained)))
  testthat::expect_false(any(sapply(model, function(x) (is.null(x@vimp_table)))))
})


# Including evaluation ---------------------------------------------------------

# Defining an experiment directory ensures that models and other data are
# stored for inspection.
data <- familiar:::test_create_small_good_data("binomial")
exp_dir <- tempdir()

familiar::summon_familiar(
  data = data,
  experiment_dir = exp_dir,
  experimental_design = "bs(fs,3)+bs(mb, 3)",
  evaluation_elements = "auc_data",
  vimp_method = "mim",
  learner = "glm_logistic",
  evaluate_top_level_only = FALSE,
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("all output is present", {
  # Models
  model_files <- list.files(
    file.path(exp_dir, "trained_models"),
    pattern = "model.RDS"
  )
  testthat::expect_length(model_files, 3L)
  
  # Evaluation data
  data_files <- list.files(
    file.path(exp_dir, "familiar_data"),
    pattern = "data.RDS"
  )
  testthat::expect_length(data_files, 8L)
  
  # Collections
  collection_files <- list.files(
    file.path(exp_dir, "familiar_collections"),
    pattern = "collection.RDS"
  )
  testthat::expect_length(collection_files, 4L)
  
  # Collection export
  export_dirs <- list.dirs(
    file.path(exp_dir, "results"),
    recursive = FALSE
  )
  testthat::expect_length(export_dirs, 4L)
})



# Train with special feature selection methods ---------------------------------

data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE
)

# Train baseline model
model <- familiar::train_familiar(
  data = data,
  experimental_design = "mb",
  vimp_method = "mim",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)

testthat::test_that("model is correctly formed", {
  testthat::expect_true(familiar:::model_is_trained(model))
  testthat::expect_false(any(sapply(model@vimp_table, is.null)))
})

# Train model using none.
model <- familiar::train_familiar(
  data = data,
  experimental_design = "mb",
  vimp_method = "none",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)
testthat::test_that("model is correctly formed", {
  testthat::expect_true(familiar:::model_is_trained(model))
  testthat::expect_true(is.null(model@vimp_table))
})

# Train model using random.
model <- familiar::train_familiar(
  data = data,
  experimental_design = "mb",
  vimp_method = "random",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)
testthat::test_that("model is correctly formed", {
  testthat::expect_true(familiar:::model_is_trained(model))
  testthat::expect_true(is.null(model@vimp_table))
})

# Train model using random.
model <- familiar::train_familiar(
  data = data,
  signature = c("feature_1", "feature_2a"),
  experimental_design = "mb",
  vimp_method = "signature_only",
  learner = "glm_logistic",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = verbose,
  parallel = FALSE
)
testthat::test_that("model is correctly formed", {
  testthat::expect_true(familiar:::model_is_trained(model))
  testthat::expect_true(is.null(model@vimp_table))
  testthat::expect_equal(model@hyperparameters$sign_size, 2L)
  testthat::expect_setequal(c("feature_1", "feature_2a"), model@model_features)
})

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

familiar:::test_plots(
  plot_function = familiar::plot_shap_waterfall,
  data_element = "shap",
  shap_max_iterations = 10L,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  debug = debug_flag,
  not_available_no_samples = FALSE
)

# The default limit for the number of important features exceeds the number of
# features in the data, which means that no actual feature selection has 
# to take place. To test all path ways, we set the number of important features
# to one here. Because the test routine relies on "none" as variable importance
# method, this triggers ad-hoc feature selection when computing the dataset.
# This will fail if outcome data are absent, mostly absent, or singular.
familiar:::test_plots(
  plot_function = familiar::plot_shap_waterfall,
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  data_element = "shap",
  shap_max_iterations = 10L,
  n_important_features = 1L,
  debug = debug_flag
)

# Test only bog-standard data: no edge cases.
familiar:::test_plots(
  plot_function = familiar::plot_shap_waterfall,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  test_config = "normal",
  debug = debug_flag
)

# Test with single instance
familiar:::test_plots(
  plot_function = familiar::plot_shap_waterfall,
  data_element = "shap",
  shap_max_iterations = 10L,
  evaluation_time = c(1.0, 2.0, 3.5),
  test_config = "single instance",
  debug = debug_flag
)

# Test with single instance and limit to the number of features.
familiar:::test_plots(
  plot_function = familiar::plot_shap_waterfall,
  data_element = "shap",
  shap_max_iterations = 10L,
  test_config = "single instance",
  plot_args = list("limit_n_features" = 2L),
  debug = debug_flag
)

# Test plotting for specific samples.
for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  data <- familiar:::test_create_good_data(outcome_type = outcome_type)
  
  model <- familiar::train_familiar(
    data = data,
    learner = switch(
      outcome_type,
      "binomial" = "glm_logistic",
      "multinomial" = "glm",
      "continuous" = "glm_gaussian",
      "survival" = "cox"
    ),
    vimp_method = "mim",
    parallel = FALSE,
    verbose = FALSE
  )
  
  plot <- familiar::plot_shap_waterfall(
    object = model,
    data = data@data[1L, ],
    shap_phi_0 = switch(
      outcome_type,
      "binomial" = 0.5,
      "multinomial" = c(0.4, 0.3, 0.3),
      "continuous" = 0.25,
      "survival" = c(0.2, 0.1)
    ),
    evaluation_times = switch(
      outcome_type,
      "survival" = c(2.0, 3.0),
      NULL
    ),
    verbose = FALSE,
    draw = debug_flag
  )
  
  testthat::test_that("Waterfall plot was correctly created.", {
    testthat::expect_true(ggplot2::is_ggplot(plot[[1L]]))
  })
  
  # With only 2 features shown.
  plot <- familiar::plot_shap_waterfall(
    object = model,
    data = data@data[1L, ],
    shap_phi_0 = switch(
      outcome_type,
      "binomial" = 0.5,
      "multinomial" = c(0.4, 0.3, 0.3),
      "continuous" = 0.25,
      "survival" = c(0.2, 0.1)
    ),
    evaluation_times = switch(
      outcome_type,
      "survival" = c(2.0, 3.0),
      NULL
    ),
    limit_n_features = 2L,
    verbose = FALSE,
    draw = debug_flag
  )
  
  testthat::test_that("Waterfall plot was correctly created with a limited number of features shown.", {
    testthat::expect_true(ggplot2::is_ggplot(plot[[1L]]))
  })
}

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE
exp_dir <- file.path(tempdir(), "familiar_intro_test")

testthat::test_that(
  "Basic example works",
  {
    if (!debug_flag) on.exit(unlink(exp_dir, recursive = TRUE), add = TRUE)
    
    testthat::expect_no_error(
      suppressWarnings(
        familiar::summon_familiar(
          data = datasets::iris,
          experiment_dir = exp_dir,
          outcome_type = "multinomial",
          outcome_column = "Species",
          experimental_design = "fs+mb",
          vimp_method = "mrmr",
          learner = "glm",
          parallel = FALSE,
          verbose = debug_flag
        )
      )
    )
  }
)

unlink(exp_dir, recursive = TRUE)

testthat::test_that(
  "Basic example with formula interface",
  {
    if (!debug_flag) on.exit(unlink(exp_dir, recursive = TRUE), add = TRUE)
    
    testthat::expect_no_error(
      suppressWarnings(
        familiar::summon_familiar(
          Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
          data = datasets::iris,
          experiment_dir = exp_dir,
          outcome_type = "multinomial",
          experimental_design = "fs+mb",
          vimp_method = "mrmr",
          learner = "glm",
          parallel = FALSE,
          verbose = debug_flag
        )
      )
    )
  }
)

unlink(exp_dir, recursive = TRUE)



testthat::test_that(
  "Example with warm start",
  {
    if (!debug_flag) on.exit(unlink(exp_dir, recursive = TRUE), add = TRUE)
    
    testthat::expect_no_error(
      suppressWarnings(
        experiment_data <- familiar::precompute_feature_info(
          data = iris,
          experiment_dir = exp_dir,
          outcome_type = "multinomial",
          outcome_column = "Species",
          experimental_design = "bs(fs+mb,5)",
          parallel = FALSE,
          verbose = debug_flag
        )
      )
    )
    unlink(exp_dir, recursive = TRUE)
    
    testthat::expect_no_error(
      suppressWarnings(
        familiar::summon_familiar(
          data = iris,
          experiment_data = experiment_data,
          experiment_dir = exp_dir,
          outcome_type = "multinomial",
          outcome_column = "Species",
          vimp_method = "mrmr",
          learner = "glm",
          parallel = FALSE,
          verbose = debug_flag
        )
      )
    )
  }
)

unlink(exp_dir, recursive = TRUE)

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

for (outcome_type in c("continuous", "multinomial", "survival")) {
  # Get data.
  data <- familiar:::test_create_good_data(outcome_type = outcome_type)

  # Select simple regression models to train with.
  learner <- switch(
    outcome_type,
    "continuous" = "glm_gaussian",
    "multinomial" = "glm_multinomial",
    "survival" = "cox"
  )

  # Data elements to skip.
  skip_data_elements <- c("ice_data", "permutation_vimp", "shap")

  # Create experiment data.
  experiment_data <- familiar::summon_familiar(
    data = data,
    experimental_design = "fs+mb",
    vimp_method = "mim",
    imputation_method = "simple",
    learner = learner,
    estimation_type = "point",
    skip_evaluation_elements = skip_data_elements,
    parallel = FALSE,
    verbose = debug_flag
  )

  # Test both familiarCollection and familiarModel objects.
  familiar_collection_list <- list(
    "collection" = experiment_data$familiarCollection,
    "model" = familiar::as_familiar_collection(
      object = experiment_data$familiarModel,
      data = data,
      familiar_data_names = "development",
      estimation_type = "point",
      data_element = setdiff(familiar:::.get_available_data_elements(), skip_data_elements)
    )
  )

  # class names ----------------------------------------------------------------
  for (collection in familiar_collection_list) {
    class_names <- familiar::get_class_names(collection)

    testthat::test_that(
      "Class names are correctly read and updated.",
      {
        if (outcome_type %in% c("binomial", "multinomial")) {
          # Test equality of the set.
          testthat::expect_setequal(class_names, levels(data@data$outcome))

          # Replace class names.
          new_class_names <- paste0("class_", seq_along(class_names))
          collection <- familiar::set_class_names(
            collection,
            old = rev(class_names),
            new = rev(new_class_names)
          )

          # Expect that the labels are the same and have the same order as the
          # original labels.
          testthat::expect_equal(
            familiar::get_class_names(collection),
            new_class_names
          )

          # Reorder levels.
          collection <- familiar::set_class_names(
            collection,
            order = rev(new_class_names)
          )

          # Expect that the labels are now re-ordered.
          testthat::expect_equal(
            familiar::get_class_names(collection),
            rev(new_class_names)
          )
          
        } else {
          testthat::expect_equal(class_names, character(0L))
        }
      }
    )
  }

  # data set names -------------------------------------------------------------
  for (collection in familiar_collection_list) {
    data_set_names <- familiar::get_data_set_names(collection)

    testthat::test_that(
      "Dataset names are correct",
      {
        testthat::expect_setequal(data_set_names, "development")

        # Replace data set names.
        new_data_set_names <- paste0("data_set", seq_along(data_set_names))
        collection <- familiar::set_data_set_names(
          collection,
          old = "development",
          new = new_data_set_names
        )

        # Expect that the labels are the same and have the same order as the
        # original labels.
        testthat::expect_equal(
          familiar::get_data_set_names(collection),
          new_data_set_names
        )
      }
    )
  }

  # feature names --------------------------------------------------------------
  for (collection in familiar_collection_list) {
    feature_names <- familiar::get_feature_names(collection)

    testthat::test_that(
      "Feature names are correct",
      {
        testthat::expect_true(all(feature_names %in% familiar:::get_feature_columns(data)))

        # Replace feature names.
        new_feature_names <- paste0("feature_", seq_along(feature_names))
        collection <- familiar::set_feature_names(
          collection,
          old = rev(feature_names),
          new = rev(new_feature_names)
        )

        # Expect that the labels are the same and have the same order as the
        # original labels.
        testthat::expect_equal(
          familiar::get_feature_names(collection),
          new_feature_names
        )

        # Reorder levels.
        collection <- familiar::set_feature_names(
          collection,
          order = rev(new_feature_names)
        )

        # Expect that the labels are now re-ordered.
        testthat::expect_equal(
          familiar::get_feature_names(collection),
          rev(new_feature_names)
        )
      }
    )
  }

  # vimp names -----------------------------------------------------------------
  for (collection in familiar_collection_list) {
    vimp_names <- familiar::get_vimp_method_names(collection)

    testthat::test_that(
      "VIMP names are correct",
      {
        testthat::expect_equal(vimp_names, "mim")

        # Replace vimp-method names.
        new_vimp_names <- paste0("vimp_method_", seq_along(vimp_names))
        collection <- familiar::set_vimp_method_names(
          collection,
          old = vimp_names,
          new = new_vimp_names
        )

        # Expect that the labels are the same and have the same order as the
        # original labels.
        testthat::expect_equal(
          familiar::get_vimp_method_names(collection),
          new_vimp_names
        )
      }
    )
  }

  # learner names --------------------------------------------------------------
  for (collection in familiar_collection_list) {
    learner_names <- familiar::get_learner_names(collection)

    testthat::test_that(
      "Learner names are correct",
      {
        testthat::expect_equal(learner_names, learner)

        # Replace learner names.
        new_learner_names <- paste0("learner_", seq_along(learner_names))
        collection <- familiar::set_learner_names(
          collection,
          old = learner_names,
          new = new_learner_names
        )

        # Expect that the labels are the same and have the same order as the
        # original labels.
        testthat::expect_equal(
          familiar::get_learner_names(collection),
          new_learner_names
        )
      }
    )
  }

  # risk group names -----------------------------------------------------------
  for (collection in familiar_collection_list) {
    risk_group_names <- familiar::get_risk_group_names(collection)

    testthat::test_that(
      "Risk group names are correct",
      {
        if (outcome_type %in% c("survival")) {
          testthat::expect_setequal(
            risk_group_names,
            c("low", "moderate", "high")
          )

          # Replace risk-group names.
          new_risk_group_names <- paste0("risk_group", seq_along(risk_group_names))
          collection <- familiar::set_risk_group_names(
            collection,
            old = rev(risk_group_names),
            new = rev(new_risk_group_names)
          )

          # Expect that the labels are the same and have the same order as the
          # original labels.
          testthat::expect_equal(
            familiar::get_risk_group_names(collection),
            new_risk_group_names
          )

          # Reorder levels.
          collection <- familiar::set_risk_group_names(
            collection,
            order = rev(new_risk_group_names)
          )

          # Expect that the labels are now re-ordered.
          testthat::expect_equal(
            familiar::get_risk_group_names(collection),
            rev(new_risk_group_names)
          )
          
        } else {
          testthat::expect_equal(
            risk_group_names,
            c("low", "moderate", "high")
          )
        }
      }
    )
  }
}

# Don't perform these tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

outcome_type_available = c("continuous", "binomial", "multinomial", "survival")

for (outcome_type in outcome_type_available) {
  # Create dataset
  full_data <- familiar:::test_create_good_data(outcome_type)
  full_data@data[, "batch_id" := "train"]
  full_data@data[121:150, "batch_id" := "test"]
  
  # Set variable importance method.
  vimp_method <- c("concordance", "mim")
  
  # Set learner
  learner_glm <- switch(
    outcome_type,
    "continuous" = "glm_gaussian",
    "binomial" = "glm_logistic",
    "multinomial" = "glm_multinomial",
    "survival" = "cox"
  )
  learner_lasso <- switch(
    outcome_type,
    "continuous" = "lasso_gaussian",
    "binomial" = "lasso_binomial",
    "multinomial" = "lasso_multinomial",
    "survival" = "lasso_cox"
  )
  
  # Get output.
  output <- suppressWarnings(familiar::summon_familiar(
    data = full_data,
    experimental_design = "mb + ev",
    validation_batch_id = "test",
    vimp_method = vimp_method,
    learner = c(learner_glm, learner_lasso),
    estimation_type = "point",
    time_max = 3.5,
    verbose = debug_flag,
    parallel = FALSE
  ))
  
  testthat::test_that("Output is correctly formed.", {
    testthat::expect_length(output$familiarModel, 4L)
    testthat::expect_length(output$familiarData, 8L)
    testthat::expect_length(output$familiarCollection, 1L)
  })
}

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()

# power.transform and other packages are required.
if (!rlang::is_installed("power.transform")) testthat::skip()

# Create data.table.
data <- familiar:::test_create_good_data(
  outcome_type = "binomial",
  to_data_object = FALSE
)

# Check reproducibility of sample assignment -----------------------------------

# Create data assignment object without fixed seed.
experiment_data_assignment_random <- familiar::precompute_data_assignment(
  data = data,
  experimental_design = "bs(fs+mb,3)",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  verbose = FALSE
)

# Create data assignment object with fixed seed.
experiment_data_assignment_a <- familiar::precompute_data_assignment(
  data = data,
  experimental_design = "bs(fs+mb,3)",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  iteration_seed = 19L,
  verbose = FALSE
)

# Create data assignment object with fixed seed.
experiment_data_assignment_b <- familiar::precompute_data_assignment(
  data = data,
  experimental_design = "bs(fs+mb,3)",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  iteration_seed = 19L,
  verbose = FALSE
)

# Create data assignment object with fixed seed.
experiment_data_assignment_different <- familiar::precompute_data_assignment(
  data = data,
  experimental_design = "bs(fs+mb,3)",
  outcome_type = "binomial",
  outcome_column = "outcome",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  class_levels = c("red", "green"),
  iteration_seed = 20L,
  verbose = FALSE
)

# Test that the same fixed seed leads to the same sample assignment.
testthat::expect_equal(
  experiment_data_assignment_a@iteration_list,
  experiment_data_assignment_b@iteration_list,
  ignore_attr = TRUE
)

# Test that the random seed leads to a different sample assignment.
testthat::expect_false(identical(
  experiment_data_assignment_random@iteration_list,
  experiment_data_assignment_a@iteration_list
))

# Test that a different fixed seed leads to a different sample assignment.
testthat::expect_false(identical(
  experiment_data_assignment_different@iteration_list,
  experiment_data_assignment_a@iteration_list
))

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# extract_data (generic) -------------------------------------------------------
setGeneric(
  "extract_data",
  function(object, ...) standardGeneric("extract_data")
)



# extract_data (familiarEnsemble) ----------------------------------------------
setMethod(
  "extract_data",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    data_element = waiver(),
    is_pre_processed = FALSE,
    dynamic_model_loading = FALSE,
    ...
  ) {
    # Generates a familiarData object from the ensemble.
    
    if (is.waive(data_element)) data_element <- .get_available_data_elements()
    
    # Check the data_element argument.
    if (length(data_element) > 0L) {
      .check_parameter_value_is_valid(
        x = data_element,
        var_name = "data_element",
        values = .get_available_data_elements()
      )
    }
    
    # Check the dynamic_model_loading argument because it is used here.
    .check_parameter_value_is_valid(
      x = dynamic_model_loading,
      var_name = "dynamic_model_loading",
      values = c(FALSE, TRUE)
    )
    
    # Set auto-detach here. Note that, if TRUE, load_models may reset it to
    # FALSE if models cannot be detached.
    object@auto_detach <- dynamic_model_loading
    
    # Check whether data is a dataObject, and create one otherwise.
    if (!is(data, "dataObject")) {
      data <- as_data_object(
        data = data,
        object = object,
        check_stringency = "external_warn"
      )
      
      # Set pre-processing level.
      data@preprocessing_level <- ifelse(is_pre_processed, "clustering", "none")
    }
    
    # Load models, and drop any models that were not trained.
    object <- load_models(object = object, drop_untrained = TRUE)
    
    # Pass to .extract_data
    return(.extract_data(
      object = object,
      data = data,
      data_element = data_element,
      ...
    ))
  }
)



# extract_data (prediction table) ----------------------------------------------
setMethod(
  "extract_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(
    object,
    data_element = waiver(),
    is_pre_processed = FALSE,
    ...
  ) {
    # Generates a familiarData object from a prediction table.
    
    if (is.waive(data_element)) {
      data_element <- .get_available_data_elements(check_from_prediction_table = TRUE)
    }
    
    # Check the data_element argument.
    if (length(data_element) > 0L) {
      .check_parameter_value_is_valid(
        x = data_element,
        var_name = "data_element",
        values = .get_available_data_elements(check_from_prediction_table = TRUE)
      )
    }
    
    # Pass to .extract_data
    return(.extract_data(
      object = object,
      data_element = data_element,
      ...
    ))
  }
)


# extract_data (dataObject) ----------------------------------------------------
setMethod(
  "extract_data",
  signature(object = "dataObject"),
  function(
    object,
    data_element = waiver(),
    is_pre_processed = TRUE,
    ...
  ) {
    # Generates a familiarData object from a dataObject.
    if (is.waive(data_element)) {
      data_element <- .get_available_data_elements(check_from_data_object = TRUE)
    }
    
    # Check the data_element argument.
    if (length(data_element) > 0L) {
      .check_parameter_value_is_valid(
        x = data_element,
        var_name = "data_element",
        values = .get_available_data_elements(check_from_data_object = TRUE)
      )
    }
    
    # Pass to .extract_data
    return(.extract_data(
      object = object,
      data_element = data_element,
      ...
    ))
  }
)



#'@title Internal function to create a familiarData object.
#'
#'@description Compute various data related to model performance and calibration
#'  from the provided dataset and `familiarEnsemble` object and store it as a
#'  `familiarData` object.
#'
#'@param object A `familiarEnsemble`, which is an ensemble of one or more
#'  `familiarModel` objects, or a `familiarDataElementPredictionTable` object
#'  that contains prediction data.
#'@param data A `dataObject` object, `data.table` or `data.frame` that
#'  constitutes the data that are assessed.
#'@param is_pre_processed Flag that indicates whether the data was already
#'  pre-processed externally, e.g. normalised and clustered. Only used if the
#'  `data` argument is a `data.table` or `data.frame`.
#'@param cl Cluster created using the `parallel` package. This cluster is then
#'  used to speed up computation through parallellisation.
#'@param time_max Time point which is used as the benchmark for e.g. cumulative
#'  risks generated by random forest, or the cut-off value for Uno's concordance
#'  index. If not provided explicitly, this parameter is read from settings used
#'  at creation of the underlying `familiarModel` objects. Only used for
#'  `survival` outcomes.
#'@param evaluation_times One or more time points that are used for in analysis
#'  of survival problems when data has to be assessed at a set time, e.g.
#'  calibration. If not provided explicitly, this parameter is read from
#'  settings used at creation of the underlying `familiarModel` objects. Only
#'  used for `survival` outcomes.
#'@param aggregation_method Method for aggregating variable importances for the
#'  purpose of evaluation. Variable importances are determined during feature
#'  selection steps and after training the model. Both types are evaluated, but
#'  feature selection variable importance is only evaluated at run-time.
#'
#'  See the documentation for the `vimp_aggregation_method` argument in
#'  `summon_familiar` for information concerning the different available
#'  methods.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'@param rank_threshold The threshold used to  define the subset of highly
#'  important features during evaluation.
#'
#'  See the documentation for the `vimp_aggregation_rank_threshold` argument in
#'  `summon_familiar` for more information.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'@param ensemble_method Method for ensembling predictions from models for the
#'  same sample. Available methods are:
#'
#'  * `median` (default): Use the median of the predicted values as the ensemble
#'  value for a sample.
#'
#'  * `mean`: Use the mean of the predicted values as the ensemble value for a
#'  sample.
#'
#'@param metric One or more metrics for assessing model performance. See the
#'  vignette on performance metrics for the available metrics. If not provided
#'  explicitly, this parameter is read from settings used at creation of the
#'  underlying `familiarModel` objects.
#'@param feature_cluster_method The method used to perform clustering. These are
#'  the same methods as for the `cluster_method` configuration parameter:
#'  `none`, `hclust`, `agnes`, `diana` and `pam`.
#'
#'  `none` cannot be used when extracting data regarding mutual correlation or
#'  feature expressions.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'
#'@param feature_linkage_method The method used for agglomerative clustering in
#'  `hclust` and `agnes`. These are the same methods as for the
#'  `cluster_linkage_method` configuration parameter: `average`, `single`,
#'  `complete`, `weighted`, and `ward`.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'
#'@param feature_cluster_cut_method The method used to divide features into
#'  separate clusters. The available methods are the same as for the
#'  `cluster_cut_method` configuration parameter: `silhouette`, `fixed_cut` and
#'  `dynamic_cut`.
#'
#'  `silhouette` is available for all cluster methods, but `fixed_cut` only
#'  applies to methods that create hierarchical trees (`hclust`, `agnes` and
#'  `diana`). `dynamic_cut` requires the `dynamicTreeCut` package and can only
#'  be used with `agnes` and `hclust`.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'
#'@param feature_similarity_threshold The threshold level for pair-wise
#'  similarity that is required to form feature clusters with the `fixed_cut`
#'  method.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'
#'@param feature_similarity_metric Metric to determine pairwise similarity
#'  between features. Similarity is computed in the same manner as for
#'  clustering, and `feature_similarity_metric` therefore has the same options
#'  as `cluster_similarity_metric`: `mcfadden_r2`, `cox_snell_r2`,
#'  `nagelkerke_r2`, `spearman`, `kendall` and `pearson`.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'
#'@param sample_cluster_method The method used to perform clustering based on
#'  distance between samples. These are the same methods as for the
#'  `cluster_method` configuration parameter: `hclust`, `agnes`, `diana` and
#'  `pam`.
#'
#'  `none` cannot be used when extracting data for feature expressions.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'
#'@param sample_linkage_method The method used for agglomerative clustering in
#'  `hclust` and `agnes`. These are the same methods as for the
#'  `cluster_linkage_method` configuration parameter: `average`, `single`,
#'  `complete`, `weighted`, and `ward`.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'
#'@param sample_similarity_metric Metric to determine pairwise similarity
#'  between samples. Similarity is computed in the same manner as for
#'  clustering, but `sample_similarity_metric` has different options that are
#'  better suited to computing distance between samples instead of between
#'  features: `gower`, `euclidean`.
#'
#'  The underlying feature data is scaled to the \eqn{[0, 1]} range (for
#'  numerical features) using the feature values across the samples. The
#'  normalisation parameters required can optionally be computed from feature
#'  data with the outer 5% (on both sides) of feature values trimmed or
#'  winsorised. To do so append `_trim` (trimming) or `_winsor` (winsorising) to
#'  the metric name. This reduces the effect of outliers somewhat.
#'
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'
#'@param icc_type String indicating the type of intraclass correlation
#'  coefficient (`1`, `2` or `3`) that should be used to compute robustness for
#'  features in repeated measurements during the evaluation of univariate
#'  importance. These types correspond to the types in Shrout and Fleiss (1979).
#'  If not provided explicitly, this parameter is read from settings used at
#'  creation of the underlying `familiarModel` objects.
#'@param verbose Flag to indicate whether feedback should be provided on the
#'  computation and extraction of various data elements.
#'@param message_indent Number of indentation steps for messages shown during
#'  computation and extraction of various data elements.
#'@param data_element String indicating which data elements are to be extracted.
#'  Default is `all`, but specific elements can be specified to speed up
#'  computations if not all elements are to be computed. This is an internal
#'  parameter that is set by, e.g. the `export_model_vimp` method.
#'@param ... Unused arguments.
#'
#'@inheritParams .parse_evaluation_settings
#'
#'@return A `familiarData` object.
#'@references 1. Shrout, P. E. & Fleiss, J. L. Intraclass correlations: uses in
#'  assessing rater reliability. Psychol. Bull. 86, 420–428 (1979).
#'@md
#'@keywords internal
.extract_data <- function(
    object,
    data = NULL,
    data_element,
    is_pre_processed = FALSE,
    cl = NULL,
    time_max = waiver(),
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    ensemble_method = waiver(),
    stratification_method = waiver(),
    evaluation_times = waiver(),
    metric = waiver(),
    feature_cluster_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_linkage_method = waiver(),
    feature_similarity_metric = waiver(),
    feature_similarity_threshold = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    sample_limit = waiver(),
    n_important_features = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    icc_type = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
) {
  
  ## Compute distance between features -----------------------------------------
  feature_similarity <- NULL
  if (any(c("model_vimp", "feature_similarity", "univariate_analysis",
            "feature_expressions", "permutation_vimp") %in% data_element)) {
    # Not for the fs_vimp data elements. This is because the subset of
    # features in the ensemble is generally smaller than that assessed within
    # feature selection..
    
    # Compute a table containing the pairwise distance between features.
    feature_similarity <- extract_feature_similarity(
      object = object,
      data = data,
      cl = cl,
      estimation_type = estimation_type,
      aggregate_results = aggregate_results,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      is_pre_processed = is_pre_processed,
      feature_cluster_method = feature_cluster_method,
      feature_linkage_method = feature_linkage_method,
      feature_cluster_cut_method = feature_cluster_cut_method,
      feature_similarity_threshold = feature_similarity_threshold,
      feature_similarity_metric = feature_similarity_metric,
      verbose = verbose,
      message_indent = message_indent
    )
  }
  
  ## Compute distance between samples ----------------------------------------
  sample_similarity <- NULL
  if (any(c("sample_similarity", "feature_expressions") %in% data_element)) {
    
    # Compute a table containing the pairwise distance between samples.
    sample_similarity <- extract_sample_similarity(
      object = object,
      data = data,
      cl = cl,
      is_pre_processed = is_pre_processed,
      sample_limit = sample_limit,
      sample_similarity_metric = sample_similarity_metric,
      sample_cluster_method = sample_cluster_method,
      sample_linkage_method = sample_linkage_method,
      verbose = verbose,
      message_indent = message_indent
    )
  }
  
  ## Aggregate feature selection variable importance -------------------------
  fs_vimp_info <- NULL
  if (any(c("fs_vimp") %in% data_element)) {
    fs_vimp_info <- extract_fs_vimp(
      object = object,
      aggregation_method = aggregation_method,
      rank_threshold = rank_threshold,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute model-specific variable importance ------------------------------
  model_vimp_info <- NULL
  if (any(c("model_vimp") %in% data_element)) {
    model_vimp_info <- extract_model_vimp(
      object = object,
      data = data,
      aggregation_method = aggregation_method,
      rank_threshold = rank_threshold,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute permutation variable importance ---------------------------------
  permutation_vimp <- NULL
  if (any(c("permutation_vimp") %in% data_element)) {
    permutation_vimp <- extract_permutation_vimp(
      object = object,
      data = data,
      cl = cl,
      feature_similarity = feature_similarity,
      metric = metric,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      n_important_features = n_important_features,
      sample_limit = sample_limit,
      detail_level = detail_level,
      estimation_type = estimation_type,
      aggregate_results = aggregate_results,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute feature expression heatmap --------------------------------------
  expression_info <- NULL
  if (any(c("feature_expressions") %in% data_element)) {
    expression_info <- extract_feature_expression(
      object = object,
      data = data,
      feature_similarity = feature_similarity,
      sample_similarity = sample_similarity,
      feature_cluster_method = feature_cluster_method,
      feature_linkage_method = feature_linkage_method,
      feature_similarity_metric = feature_similarity_metric,
      sample_cluster_method = sample_cluster_method,
      sample_linkage_method = sample_linkage_method,
      sample_similarity_metric = sample_similarity_metric,
      evaluation_times = evaluation_times,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute univariate feature importance -----------------------------------
  univar_info <- NULL
  if (any(c("univariate_analysis") %in% data_element)) {
    univar_info <- extract_univariate_analysis(
      object = object,
      data = data,
      cl = cl,
      icc_type = icc_type,
      feature_similarity = feature_similarity,
      feature_cluster_method = feature_cluster_method,
      feature_cluster_cut_method = feature_cluster_cut_method,
      feature_linkage_method = feature_linkage_method,
      feature_similarity_threshold = feature_similarity_threshold,
      feature_similarity_metric = feature_similarity_metric,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Aggregate model hyper-parameters ----------------------------------------
  hyperparameter_info <- NULL
  if (any(c("hyperparameters") %in% data_element)) {
    hyperparameter_info <- extract_hyperparameters(
      object = object,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute model predictions -----------------------------------------------
  prediction_data <- NULL
  if (any(c("prediction_data") %in% data_element)) {
    prediction_data <- extract_predictions(
      object = object,
      data = data,
      cl = cl,
      ensemble_method = ensemble_method,
      detail_level = detail_level,
      estimation_type = estimation_type,
      aggregate_results = aggregate_results,
      confidence_level = confidence_level,
      evaluation_times = evaluation_times,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute model performance metrics ---------------------------------------
  model_performance_data <- NULL
  if (any(c("model_performance") %in% data_element)) {
    model_performance_data <- extract_performance(
      object = object,
      data = data,
      cl = cl,
      metric = metric,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      aggregate_results = aggregate_results,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute decision curve analysis data ------------------------------------
  decision_curve_data <- NULL
  if (any(c("decision_curve_analyis") %in% data_element)) {
    decision_curve_data <- extract_decision_curve_data(
      object = object,
      data = data,
      cl = cl,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      aggregate_results = aggregate_results,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Aggregate stratification data -------------------------------------------
  stratification_info <- NULL
  if (any(c("risk_stratification_info") %in% data_element)) {
    stratification_info <- extract_risk_stratification_info(
      object = object,
      detail_level = detail_level,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute risk group stratification ---------------------------------------
  stratification_data <- NULL
  if (any(c("risk_stratification_data") %in% data_element)) {
    stratification_data <- extract_risk_stratification_data(
      object = object,
      data = data,
      cl = cl,
      ensemble_method = ensemble_method,
      stratification_method = stratification_method,
      detail_level = detail_level,
      confidence_level = confidence_level,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
  }
  
  ## Aggregate calibration information ---------------------------------------
  calibration_info <- NULL
  if (any(c("calibration_info") %in% data_element)) {
    calibration_info <- extract_calibration_info(
      object = object,
      detail_level = detail_level,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute calibration data ------------------------------------------------
  calibration_data <- NULL
  if (any(c("calibration_data") %in% data_element)) {
    calibration_data <- extract_calibration_data(
      object = object,
      data = data,
      cl = cl,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      aggregate_results = aggregate_results,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute AUC-ROC and AUC-PR ----------------------------------------------
  auc_data <- NULL
  if (any(c("auc_data") %in% data_element)) {
    auc_data <- extract_auc_data(
      object = object,
      data = data,
      cl = cl,
      ensemble_method = ensemble_method,
      detail_level = detail_level,
      estimation_type = estimation_type,
      aggregate_results = aggregate_results,
      bootstrap_ci_method = bootstrap_ci_method,
      confidence_level = confidence_level,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute confusion matrix ------------------------------------------------
  confusion_matrix_info <- NULL
  if (any(c("confusion_matrix") %in% data_element)) {
    confusion_matrix_info <- extract_confusion_matrix(
      object = object,
      data = data,
      cl = cl,
      ensemble_method = ensemble_method,
      detail_level = detail_level,
      message_indent = message_indent,
      verbose = verbose
    )
  }
  
  ## Compute individual conditional expectation ------------------------------
  ice_data <- NULL
  if (any(c("ice_data") %in% data_element)) {
    ice_data <- extract_ice(
      object = object,
      data = data,
      cl = cl,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      sample_limit = sample_limit,
      n_important_features = n_important_features,
      detail_level = detail_level,
      estimation_type = estimation_type,
      aggregate_results = aggregate_results,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      is_pre_processed = is_pre_processed,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
  }
  
  ## Compute SHAP values for features ------------------------------------------
  shap_data <- NULL
  if ("shap" %in% data_element) {
    
    shap_data <- extract_shap(
      object = object,
      data = data,
      cl = cl,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      sample_limit = sample_limit,
      detail_level = detail_level,
      aggregate_results = aggregate_results,
      n_important_features = n_important_features,
      is_pre_processed = is_pre_processed,
      verbose = verbose,
      message_indent = message_indent,
      ...
    )
  }
  
  # Set up outcome information object.
  outcome_info <- .create_outcome_info(object)

  # Create a familiarData object
  fam_data <- methods::new(
    "familiarData",
    outcome_type = object@outcome_type,
    outcome_info = outcome_info,
    fs_vimp = fs_vimp_info,
    model_vimp = model_vimp_info,
    permutation_vimp = permutation_vimp,
    hyperparameters = hyperparameter_info,
    hyperparameter_data = NULL,
    required_features = .optional_from_slot(object, "required_features", alternative = NULL),
    model_features = .optional_from_slot(object, "model_features", alternative = NULL),
    learner = .optional_from_slot(object, "learner", alternative = "none"),
    vimp_method = .optional_from_slot(object, "vimp_method", alternative = "none"),
    prediction_data = prediction_data,
    confusion_matrix = confusion_matrix_info,
    decision_curve_data = decision_curve_data,
    calibration_info = calibration_info,
    calibration_data = calibration_data,
    model_performance = model_performance_data,
    km_info = stratification_info,
    km_data = stratification_data,
    auc_data = auc_data,
    univariate_analysis = univar_info,
    feature_expressions = expression_info,
    feature_similarity = feature_similarity,
    sample_similarity = sample_similarity,
    ice_data = ice_data,
    shap_data = shap_data
  )
  
  # Add package version to the data set 
  fam_data <- add_package_version(object = fam_data)
  
  # Return data
  return(fam_data)
}

.set_recalibration <- function(
    object, 
    data, 
    time = NULL, 
    ...
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- predicted_outcome <- NULL
  
  # Initial empty calibration list
  calibration_model_list <- list()

  # Predict the outcome using the current model.
  model_predictions <- .predict(
    object = object,
    data = data,
    allow_recalibration = FALSE,
    time = time
  )
  
  if (!all_predictions_valid(model_predictions)) return(NULL)
  
  # Convert to data.table.
  model_predictions <- .as_data_table(model_predictions)

  if (object@outcome_type %in% c("binomial", "multinomial")) {
    # Get class levels and class probability column names
    class_levels <- get_outcome_class_levels(x = object)

    # Build a logistic model on top of the predicted class probabilities for
    # each predicted class probability
    for (ii in seq_along(class_levels)) {
      # Set positive class flag
      model_predictions[, "positive_class" := outcome == class_levels[ii]]

      # Parse formula
      model_formula <- stats::reformulate(
        termlabels = class_levels[ii],
        response = "positive_class"
      )

      # Remove NA from the data.table
      model_predictions <- model_predictions[is.finite(get(class_levels[ii])), ]

      # If the prediction data table returns no or just one (valid) entry,
      # calibration is not possible.
      if (nrow(model_predictions) <= 1L) return(NULL)

      # Generate the calibration model.
      calibration_model <- tryCatch(
        stats::glm(
          model_formula,
          data = model_predictions,
          family = stats::binomial(link = "logit")
        ),
        error = identity
      )

      # Check if the calibration model was created.
      if (inherits(calibration_model, "error")) calibration_model <- NULL

      # Create calibration model
      calibration_model_list[[class_levels[ii]]] <- calibration_model
    }
    
  } else if (object@outcome_type == "survival") {
    # Parse formula
    formula <- stats::reformulate(
      termlabels = "predicted_outcome",
      response = quote(survival::Surv(outcome_time, outcome_event))
    )
    
    # Remove NA from the table
    model_predictions <- model_predictions[is.finite(predicted_outcome)]

    # If the prediction data table returns no or just one (valid) entry,
    # calibration is not possible.
    if (nrow(model_predictions) <= 1L) return(NULL)

    # Generate model
    model_control <- survival::coxph.control(iter.max = 100L)
    calibration_model <- tryCatch(
      survival::coxph(
        formula,
        data = model_predictions,
        control = model_control,
        y = FALSE
      ),
      error = identity
    )
    
    # Check if the model trained at all.
    if (inherits(calibration_model, "error")) return(NULL)

    # Check if the model fitter converged in time.
    if (calibration_model$iter >= 100L) return(NULL)

    # Store calibration model.
    calibration_model_list[[1L]] <- calibration_model
  }

  # Return list of calibration models
  return(calibration_model_list)
}



.apply_recalibration <- function(object, prediction_table, data) {
  # Suppress NOTES due to non-standard evaluation in data.table
  prob_sum <- NULL

  # Return predictions if calibration models are missing
  if (is_empty(object@calibration_model)) return(prediction_table)

  # Return predictions if it is empty
  if (is_empty(prediction_table)) return(prediction_table)

  # Convert to data.table.
  predictions <- .as_data_table(prediction_table)
  
  if (object@outcome_type %in% c("binomial", "multinomial")) {
    class_levels <- get_outcome_class_levels(x = object)
    prediction_list <- list()
    
    # Iterate over the calibration model for each class and obtain the class
    # probabilities.
    for (ii in seq_along(class_levels)) {

      # Skip if no calibration model is provided.
      if (is.null(object@calibration_model[[class_levels[ii]]])) {
        prediction_list[[class_levels[ii]]] <- predictions[[class_levels[ii]]]
        next
      }

      # Predict calibrated probabilities using the calibration model for the
      # current column.
      prediction_list[[class_levels[ii]]] <- stats::predict.glm(
        object = object@calibration_model[[class_levels[ii]]],
        newdata = predictions,
        type = "response"
      )
    }
    
    # Convert list of class probabilities to a data.table.
    predictions <- data.table::as.data.table(prediction_list)
    
    # Normalise predicted probabilities to 1.0.
    predictions[, "prob_sum" := rowSums(.SD, na.rm = TRUE), .SDcols = class_levels]
    predictions[, (class_levels) := lapply(.SD, "/", prob_sum), .SDcols = class_levels]
    predictions[, "prob_sum" := NULL]

    # Create new prediction table object.
    prediction_table <- as_prediction_table(
      x = predictions,
      type = "classification",
      data = data,
      model_object = object
    )
    
  } else if (object@outcome_type == "survival") {
    # Predict cox PH relative risk
    predicted_outcome_value <- predict(
      object = object@calibration_model[[1L]],
      newdata = predictions,
      type = "risk"
    )

    # Create new prediction table object.
    prediction_table <- as_prediction_table(
      x = predicted_outcome_value,
      type = "hazard_ratio",
      data = data,
      model_object = object
    )
  }

  return(prediction_table)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarSurvRegr -------------------------------------------------------------
setClass(
  "familiarSurvRegr",
  contains = "familiarModel",
  slots = list("encoding_reference_table" = "ANY"),
  prototype = list("encoding_reference_table" = NULL)
)

# initialize -------------------------------------------------------------------
setMethod(
  "initialize", 
  signature(.Object = "familiarSurvRegr"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set the required package
    .Object@package <- "survival"

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available", signature(object = "familiarSurvRegr"),
  function(object, ...) {
    # Survival regression is only available if for survival outcomes.
    return(object@outcome_type == "survival")
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarSurvRegr"),
  function(object, data = NULL, ...) {
    # Initialise list and declare hyperparameter entries
    param <- list()
    param$sign_size <- list()
    param$distribution <- list()

    # If data is explicitly set to NULL, return the list with hyperparameter
    # names only
    if (is.null(data)) return(param)

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(
      data = data,
      restrict_samples = TRUE
    )

    # outcome distribution -----------------------------------------------------

    # Randomisation of distribution depends on selected learner.
    if (object@learner == "survival_regr") {
      distribution_default <- c(
        "weibull", "exponential", "gaussian", "logistic", "loglogistic", "lognormal"
      )
      
    } else {
      distribution_default <- sub(
        x = object@learner,
        pattern = "survival_regr_",
        replacement = "",
        fixed = TRUE
      )
    }

    # Set the distribution parameter
    param$distribution <- .set_hyperparameter(
      default = distribution_default,
      type = "factor",
      range = distribution_default,
      randomise = length(distribution_default) > 1L
    )

    # Return hyper-parameters
    return(param)
  }
)



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarSurvRegr"),
  function(object, type = "default") {
    # Survival regression models predict an expected survival time by default.
    if (type == "default") {
      return("expected_survival_time")
    } else if (type == "survival_probability") {
      return("survival_probability")
    } else {
      ..error_reached_unreachable_code(
        "get_prediction_type,familiarSurvRegr: unknown type"
      )
    }
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarSurvRegr",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object,
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Use effect coding to convert categorical data into encoded data - this is
    # required to deal with factors with missing/new levels between training and
    # test data sets.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )

    # Find feature columns in the data.
    feature_columns <- get_feature_columns(x = encoded_data$encoded_data)

    # Parse formula
    formula <- stats::reformulate(
      termlabels = feature_columns,
      response = quote(survival::Surv(outcome_time, outcome_event))
    )

    # Set limits to the number of iterations that can be performed by
    # survival regression.
    model_control <- survival::survreg.control(iter.max = 100L)

    # Train the model.
    model <- do.call_with_handlers(
      survival::survreg,
      args = list(formula,
        "data" = encoded_data$encoded_data@data,
        "control" = model_control,
        "y" = FALSE,
        "dist" = as.character(object@hyperparameters$distribution)
      )
    )

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) return(callNextMethod(object = object))

    # Check if the model fitter converged in time.
    if (model$iter >= 100L) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        "Model fitter ran out of iterations and did not converge."
      )))
    }

    # Check if all coefficients could not be estimated. Sometimes models could
    # be trained with a large number of features whose coefficients fail to
    # converge. This would sometimes lead to overly large signature sizes being
    # selected during hyperparameter optimisation, especially in situations
    # where there is not a lot of signal. Checking for non-finite coefficients
    # is an easy way to figure out if the model is not properly trained.
    if (!all(sapply(stats::coef(model), is.finite))) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_failed_model_coefficient_estimation()
      )))
    }

    # Add model
    object@model <- model

    # Add the contrast references to model_list
    object@encoding_reference_table <- encoded_data$reference_table

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarSurvRegr",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Turn into a Naive model.
    object <- methods::new("familiarNaiveSurvivalTimeModel", object)

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarSurvRegr",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "default",
    time = NULL, 
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) return(callNextMethod())
    
    # Check if the data is empty.
    if (is_empty(data)) return(callNextMethod())
    
    # Encode data so that the features are the same as in the training.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )
    
    if (type %in% c("default", "survival_probability")) {
      # default ----------------------------------------------------------------

      if (object@outcome_type == "survival") {
        if (type == "default") {
          # Use the model to predict expected survival time.
          model_predictions <- predict(
            object = object@model,
            newdata = encoded_data$encoded_data@data,
            type = "response"
          )

          # Store as prediction table.
          prediction_table <- as_prediction_table(
            x = model_predictions,
            type = "expected_survival_time",
            data = data,
            model_object = object
          )
          
        } else if (type == "survival_probability") {
          # To predict survival probability we first compute survival quantiles,
          # which are survival probabilities.

          # If time is unset, read the max time stored by the model.
          if (is.null(time)) time <- object@settings$time_max
          
          # Survival quantiles from 1.00 to 0.01
          survival_quantiles <- seq(from = 1.00, to = 0.01, by = -0.01)

          # Get estimated failure times
          failure_matrix <- predict(
            object = object@model,
            newdata = encoded_data$encoded_data@data,
            type = "quantile",
            p = 1.00 - survival_quantiles
          )
          
          # Set id columns
          id_columns <- get_id_columns()

          # Convert event_matrix to a matrix.
          if (!is.matrix(failure_matrix)) {
            failure_matrix <- matrix(
              data = failure_matrix,
              ncol = length(failure_matrix)
            )
          }

          # Combine with identifiers and cast to table.
          failure_table <- cbind(
            data@data[, mget(id_columns)],
            data.table::as.data.table(failure_matrix)
          )

          # Remove duplicate entries
          failure_table <- unique(failure_table, by = id_columns)

          # Melt to a long format.
          failure_table <- data.table::melt(
            failure_table,
            id.vars = id_columns,
            variable.name = "quantile_variable",
            value.name = "survival_time"
          )

          # Create conversion table to convert temporary variables into
          # the event times.
          conversion_table <- data.table::data.table(
            "quantile_variable" = paste0("V", seq_along(survival_quantiles)),
            "survival_quantile" = survival_quantiles
          )

          # Add in
          failure_table <- merge(
            x = failure_table, 
            y = conversion_table, 
            by = "quantile_variable"
          )

          # Drop the time_variable column
          failure_table[, "quantile_variable" := NULL]

          # Now, interpolate at the given time point.
          failure_table <- lapply(
            split(failure_table, by = id_columns),
            function(sample_table, time, id_columns) {
              # Interpolate values at the given time.
              value <- stats::approx(
                x = sample_table$survival_time,
                y = sample_table$survival_quantile,
                xout = time,
                rule = 2L
              )$y
              
              # Create an output table
              output_table <- data.table::copy(sample_table[1L, mget(id_columns)])
              output_table[, "survival_probability" := value]
              
              return(output_table)
            },
            time = time, 
            id_columns = id_columns
          )
          
          # Concatenate to single table.
          failure_table <- data.table::rbindlist(failure_table)
          
          # Then merge the event table into the prediction table.
          prediction_table <- merge(
            x = data@data[, mget(id_columns)],
            y = failure_table, 
            by = id_columns,
            sort = FALSE
          )
          
          # Store as prediction table
          prediction_table <- as_prediction_table(
            x = prediction_table$survival_probability,
            type = "survival_probability",
            data = data,
            time = time,
            model_object = object
          )
        }
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      return(prediction_table)
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified method --------------------------------------------------

      # Use the model to predict expected survival time.
      return(predict(
        object = object@model,
        newdata = encoded_data$encoded_data@data,
        type = type,
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



# ..get_prediction_table_type --------------------------------------------------
setMethod(
  "..get_prediction_table_type",
  signature(object = "familiarSurvRegr"),
  function(object, type, ...) {
    prediction_table_type <- NULL
    if (object@outcome_type %in% c("survival") && type == "default") {
      prediction_table_type <- "expected_survival_time"
    } else {
      prediction_table_type <- callNextMethod()
    }
    
    return(prediction_table_type)
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarSurvRegr"),
  function(object, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL

    if (!model_is_trained(object)) return(callNextMethod())

    # Check that required packages are loaded and installed.
    require_package(object, "vimp")

    # Define p-values
    coefficient_z_values <- tryCatch(
      .compute_z_statistic(object),
      error = identity
    )

    if (inherits(coefficient_z_values, "error")) return(callNextMethod())

    # Remove any intercept from the data.
    coefficient_z_values <- coefficient_z_values[
      names(coefficient_z_values) != "(Intercept)"
    ]

    if (length(coefficient_z_values) == 0L) return(callNextMethod())

    # Assign to variable importance table.
    vimp_table <- data.table::data.table(
      "score" = coefficient_z_values,
      "name" = names(coefficient_z_values)
    )

    # Remove NA values
    vimp_table <- vimp_table[is.finite(score)]

    # Create variable importance object.
    vimp_object <- methods::new(
      "vimpTable",
      vimp_table = vimp_table,
      encoding_table = object@encoding_reference_table,
      score_aggregation = "max",
      invert = TRUE
    )

    return(vimp_object)
  }
)



# ..set_calibration_info -------------------------------------------------------
setMethod(
  "..set_calibration_info",
  signature(object = "familiarSurvRegr"),
  function(object, data) {
    # Check if calibration info already.
    if (has_calibration_info(object)) return(object)

    if (object@outcome_type == "survival") {
      # Determine baseline survival.
      object@calibration_info <- get_baseline_survival(data = data)
      
    } else {
      return(callNextMethod())
    }

    return(object)
  }
)



# ..set_vimp_parameters --------------------------------------------------------
setMethod(
  "..set_vimp_parameters",
  signature(object = "familiarSurvRegr"),
  function(object, method, ...) {
    # Randomisation of distribution depends on selected learner.
    if (method == "survival_regr") {
      distribution_default <- "weibull"
      
    } else {
      distribution_default <- sub(
        x = method,
        pattern = "survival_regr_",
        replacement = "",
        fixed = TRUE
      )
    }

    # Set the distribution parameter
    object@hyperparameters$distribution <- distribution_default

    return(object)
  }
)



# .trim_model-------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarSurvRegr"),
  function(object, ...) {
    # Update model by removing the call.
    object@model$call <- call("trimmed")

    # Add show.
    object <- .capture_show(object)

    # Remove .Environment.
    object@model$terms <- .replace_environment(object@model$terms)

    # Remove elements that contain sample-specific values.
    object@model$linear.predictors <- NULL

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    # Default method for models that lack a more specific method.
    return(object)
  }
)



.get_available_survival_regression_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c(
    "survival_regr", "survival_regr_weibull", "survival_regr_exponential",
    "survival_regr_gaussian", "survival_regr_logistic",
    "survival_regr_lognormal", "survival_regr_loglogistic"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("survival_regr"))
  }
  
  return(learners)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarDataElementFeatureSimilarity object ----------------------------------
setClass(
  "familiarDataElementFeatureSimilarity",
  contains = "familiarDataElement",
  slots = list(
    "similarity_metric" = "character",
    "cluster_method" = "character",
    "linkage_method" = "character",
    "cluster_cut_method" = "character",
    "similarity_threshold" = "ANY",
    "dendrogram" = "ANY"
  ),
  prototype = methods::prototype(
    detail_level = "ensemble",
    similarity_metric = NA_character_,
    cluster_method = NA_character_,
    linkage_method = NA_character_,
    cluster_cut_method = NA_character_,
    similarity_threshold = NULL,
    dendrogram = NULL,
    value_column = "value",
    grouping_column = c("feature_name_1", "feature_name_2")
  )
)


# extract_feature_similarity (generic) -----------------------------------------

#'@title Internal function to extract the feature distance table.
#'
#'@description Computes and extracts the feature distance table for features
#'  used in a `familiarEnsemble` object. This table can be used to cluster
#'  features, and is exported directly by `export_feature_similarity`.
#'
#'@inheritParams .extract_data
#'
#'@return A data.table containing pairwise distance between features. This data
#'  is only the upper triangular of the complete matrix (i.e. the sparse
#'  unitriangular representation). Diagonals will always be 0.0 and the lower
#'  triangular is mirrored.
#'@md
#'@keywords internal
setGeneric(
  "extract_feature_similarity",
  function(
    object,
    data,
    cl = NULL,
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    feature_similarity_metric = waiver(),
    verbose = FALSE,
    message_indent = 0L,
    ...
  ) {
    standardGeneric("extract_feature_similarity")
  }
)


# extract_feature_similarity (familiarEnsemble) --------------------------------
setMethod(
  "extract_feature_similarity",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    feature_similarity_metric = waiver(),
    verbose = FALSE,
    message_indent = 0L,
    ...
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Computing pairwise similarity between features."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Obtain cluster method from stored settings, if required.
    if (is.waive(feature_cluster_method)) {
      feature_cluster_method <- object@settings$feature_cluster_method
    }
    
    # Obtain linkage function from stored settings, if required.
    if (is.waive(feature_linkage_method)) {
      feature_linkage_method <- object@settings$feature_linkage_method
    } 
    
    # Obtain feature cluster cut method from stored settings, if required.
    if (is.waive(feature_cluster_cut_method)) {
      feature_cluster_cut_method <- object@settings$feature_cluster_cut_method
    } 
    
    # Obtain cluster similarity threshold from stored settings, if required.
    if (is.waive(feature_similarity_threshold)) {
      feature_similarity_threshold <- object@settings$feature_similarity_threshold
    }
    
    # Obtain similarity metric from stored settings, if required.
    if (is.waive(feature_similarity_metric)) {
      feature_similarity_metric <- object@settings$feature_similarity_metric
    }
    
    # Replace feature cluster method == "none" with "hclust"
    if (feature_cluster_method == "none") {
      feature_cluster_method <- "hclust"
    } 
    
    .check_cluster_parameters(
      cluster_method = feature_cluster_method,
      cluster_linkage = feature_linkage_method,
      cluster_cut_method = feature_cluster_cut_method,
      cluster_similarity_threshold = feature_similarity_threshold,
      cluster_similarity_metric = feature_similarity_metric,
      data_type = "feature"
    )
    
    # Obtain confidence level from the settings file stored with the
    # familiarEnsemble object.
    if (is.waive(confidence_level)) {
      confidence_level <- object@settings$confidence_level
    }
    
    # Check alpha
    .check_number_in_valid_range(
      x = confidence_level,
      var_name = "confidence_level",
      range = c(0.0, 1.0),
      closed = c(FALSE, FALSE)
    )
    
    # Load the bootstrap method
    if (is.waive(bootstrap_ci_method)) {
      bootstrap_ci_method <- object@settings$bootstrap_ci_method
    }
    
    .check_parameter_value_is_valid(
      x = bootstrap_ci_method,
      var_name = "bootstrap_ci_methpd",
      values = .get_available_bootstrap_confidence_interval_methods()
    )
    
    # Check the estimation type.
    estimation_type <- .parse_estimation_type(
      x = estimation_type,
      object = object,
      default = "point",
      data_element = "feature_similarity",
      detail_level = "ensemble",
      has_internal_bootstrap = TRUE
    )
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "feature_similarity"
    )
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "familiarDataElementFeatureSimilarity",
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      similarity_metric = feature_similarity_metric,
      cluster_method = feature_cluster_method,
      linkage_method = feature_linkage_method,
      cluster_cut_method = feature_cluster_cut_method,
      similarity_threshold = feature_similarity_threshold
    )
    
    # Generate elements to send to dispatch.
    similarity_data <- extract_dispatcher(
      FUN = .extract_feature_similarity,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(similarity_data)
  }
)


# extract_feature_similarity (familiarDataElementPredictionTable) --------------
setMethod(
  "extract_feature_similarity",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("feature similarity")
    
    return(NULL)
  }
)



# extract_feature_similarity (dataObject) --------------------------------------

setMethod(
  "extract_feature_similarity",
  signature(object = "dataObject"),
  function(
    object,
    cl = NULL,
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    feature_similarity_metric = waiver(),
    verbose = FALSE,
    message_indent = 0L,
    ...
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Computing pairwise similarity between features."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Parse input.
    settings <- .parse_feature_clustering(
      feature_cluster_method = feature_cluster_method,
      feature_linkage_method = feature_linkage_method,
      feature_cluster_cut_method = feature_cluster_cut_method,
      feature_similarity_metric = feature_similarity_metric,
      feature_similarity_threshold = feature_similarity_threshold
    )
    
    # There are no settings attached to dataObject, so we pass these through.
    # Set default cluster method.
    if (is.waive(feature_cluster_method)) {
      feature_cluster_method <- settings$feature_cluster_method
    }
    
    # Set default linkage method, if required.
    if (is.waive(feature_linkage_method)) {
      feature_linkage_method <- settings$feature_linkage_method
    } 
    
    # Set default cluster cut method, if required.
    if (is.waive(feature_cluster_cut_method)) {
      feature_cluster_cut_method <- settings$feature_cluster_cut_method
    } 
    
    # Set default cluster similarity threshold, if required.
    if (is.waive(feature_similarity_threshold)) {
      feature_similarity_threshold <- settings$feature_similarity_threshold
    }
    
    # Obtain default similarity metric, if required.
    if (is.waive(feature_similarity_metric)) {
      feature_similarity_metric <- settings$feature_similarity_metric
    }
    
    # Replace feature cluster method == "none" with "hclust"
    if (feature_cluster_method == "none") feature_cluster_method <- "hclust"
    
    # Set default confidence level.
    if (is.waive(confidence_level)) confidence_level <- 0.95
    
    # Check alpha
    .check_number_in_valid_range(
      x = confidence_level,
      var_name = "confidence_level",
      range = c(0.0, 1.0),
      closed = c(FALSE, FALSE)
    )
    
    # Load the bootstrap method
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- "percentile"
    
    .check_parameter_value_is_valid(
      x = bootstrap_ci_method,
      var_name = "bootstrap_ci_methpd",
      values = .get_available_bootstrap_confidence_interval_methods()
    )
    
    # Check the estimation type.
    estimation_type <- .parse_estimation_type(
      x = estimation_type,
      object = object,
      default = "point",
      data_element = "feature_similarity",
      detail_level = "ensemble",
      has_internal_bootstrap = TRUE
    )
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "feature_similarity"
    )
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "familiarDataElementFeatureSimilarity",
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      similarity_metric = feature_similarity_metric,
      cluster_method = feature_cluster_method,
      linkage_method = feature_linkage_method,
      cluster_cut_method = feature_cluster_cut_method,
      similarity_threshold = feature_similarity_threshold
    )
    
    # Generate feature info by running the familiarTaskGenericFeatureInfo task.
    generic_feature_info_task <- new("familiarTaskGenericFeatureInfo")
    object@feature_info <- .perform_task(
      object = generic_feature_info_task,
      data = object
    )
    
    # Generate elements to send to dispatch.
    similarity_data <- extract_dispatcher(
      FUN = .extract_feature_similarity,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = object,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(similarity_data)
  }
)



.extract_feature_similarity <- function(
    object,
    data,
    proto_data_element,
    cl = NULL,
    is_pre_processed,
    progress_bar = FALSE,
    aggregate_results,
    n_bootstraps,
    message_indent = 0L,
    verbose = FALSE,
    ...
) {
  
  # Add the name of the ensemble model
  data_element <- add_model_name(data = proto_data_element, object = object)
  
  # Retrieve input data.
  data <- process_input_data(
    object = object,
    data = data,
    stop_at = "imputation",
    is_pre_processed = is_pre_processed
  )
  
  # Check if the input data is not empty
  if (is_empty(data)) return(NULL)
  
  # Check if the number of samples is sufficient (>5), and return an empty table
  # if not.
  if (get_n_samples(data, "series") <= 5L) return(NULL)
  
  if (is(object, "familiarEnsemble")) {
    # Maintain only important features. The current set is based on the required
    # features.
    data <- filter_features(
      data = data,
      available_features = get_model_features(object)
    )
  }

  # Identify eligible columns.
  feature_columns <- get_feature_columns(x = data)
  
  # Break if there are not at least 2 features present between which correlation
  # can be compared.
  if (length(feature_columns) < 2L) return(data_element)
  
  # Add bootstrap data.
  bootstrap_data <- add_data_element_bootstrap(
    x = data_element,
    n_bootstraps = n_bootstraps,
    ...
  )
  
  # Iterate over elements.
  if (n_bootstraps > 1L) {
    data_elements <- fam_mapply(
      cl = cl,
      assign = NULL,
      FUN = ..extract_feature_similarity,
      data_element = bootstrap_data$data_element,
      bootstrap = bootstrap_data$bootstrap,
      bootstrap_seed = bootstrap_data$seed,
      MoreArgs = list(
        "data" = data,
        "feature_info_list" = object@feature_info
      ),
      progress_bar = progress_bar && verbose,
      chopchop = TRUE
    )
    
  } else {
    data_elements <- fam_mapply(
      cl = NULL,
      assign = NULL,
      FUN = ..extract_feature_similarity,
      data_element = bootstrap_data$data_element,
      bootstrap = bootstrap_data$bootstrap,
      bootstrap_seed = bootstrap_data$seed,
      MoreArgs = list(
        "data" = data,
        "feature_info_list" = object@feature_info,
        "cl" = cl,
        "verbose" = verbose,
        "message_indent" = message_indent
      ),
      progress_bar = FALSE
    )
    
  }
  
  # Merge data elements
  data_elements <- merge_data_elements(data_elements)
  
  if (aggregate_results) data_elements <- .compute_data_element_estimates(x = data_elements)
  
  return(data_elements)
}



..extract_feature_similarity <- function(
    cl = NULL,
    data_element,
    data,
    feature_info_list,
    bootstrap,
    bootstrap_seed,
    message_indent = 0L,
    verbose = FALSE
) {
  
  # Bootstrap the data.
  if (bootstrap) {
    data <- get_bootstrap_sample(
      data = data,
      seed = bootstrap_seed
    )
  }
  
  # Check if the number of samples is sufficient (>5), and return an
  # empty table if not.
  if (get_n_samples(data, "series") <= 5L) return(NULL)
  
  # Identify eligible columns.
  feature_columns <- get_feature_columns(x = data)
  
  # Compute the similarity table
  data_element@data <- set_similarity_table(
    data = data,
    feature_info_list = feature_info_list[feature_columns],
    similarity_metric = data_element@similarity_metric,
    data_type = "feature",
    cl = cl,
    message_indent = message_indent + 1L,
    verbose = verbose
  )
  
  return(data_element)
}            



.append_feature_similarity_dendrogram <- function(x) {
  
  if (is_empty(x)) return(x)
  
  # Create a cluster method object using data stored in x.
  cluster_method_object <- .create_feature_similarity_cluster_method_object(x = x)
  
  if (is.null(cluster_method_object)) return(x)
  
  # Create the cluster object.
  object <- apply_cluster_method(object = cluster_method_object)
  
  # Attach to data element.
  x@dendrogram <- object@object
  
  return(x)
}



.append_feature_similarity_clusters <- function(x) {
  
  if (is_empty(x)) return(x)
  
  # Generate the clustering table.
  cluster_table <- .compute_feature_similarity_cluster_table(x = x)
  
  # Check for empty cluster tables.
  if (is_empty(cluster_table)) return(x)
  
  # Keep only name and cluster_id columns
  cluster_table <- cluster_table[, mget(c("name", "cluster_id"))]
  
  # Compute cluster size
  cluster_table[, "cluster_size" := .N, by = "cluster_id"]
  
  # Rename name column to 
  data.table::setnames(
    x = cluster_table,
    old = "name",
    new = "feature"
  )
  
  # Set cluster info as data.
  x@data <- cluster_table
  
  # Reset value and grouping columns.
  x@value_column <- c("cluster_id", "cluster_size")
  x@grouping_column <- "feature"
  
  return(x)
}



.append_feature_similarity_clustering <- function(x) {
  
  if (is_empty(x)) return(x)
  
  # Generate the clustering table.
  cluster_table <- .compute_feature_similarity_cluster_table(x = x)
  
  # Check for empty cluster tables.
  if (is_empty(cluster_table)) return(x)
  
  # Keep only name and label_order columns,
  cluster_table <- cluster_table[, mget(c("name", "label_order"))]
  
  # Merge ordering into feature_similarity_table. The table is first
  # merged on feature_name_1 and then on feature_name_2.
  mutual_correlation_table <- data.table::copy(x@data)
  mutual_correlation_table <- merge(
    x = mutual_correlation_table,
    y = cluster_table,
    by.x = "feature_name_1",
    by.y = "name",
    all.x = TRUE,
    all.y = FALSE
  )
  
  mutual_correlation_table <- merge(
    x = mutual_correlation_table,
    y = cluster_table,
    by.x = "feature_name_2",
    by.y = "name",
    all.x = TRUE,
    all.y = FALSE
  )
  
  # Rename columns
  data.table::setnames(
    x = mutual_correlation_table,
    old = c("label_order.x", "label_order.y"),
    new = c("label_order_1", "label_order_2")
  )
  
  # Add to data element.
  x@data <- mutual_correlation_table
  
  # Add grouping column.
  x@grouping_column <- c(x@grouping_column, "label_order_1", "label_order_2")
  
  return(x)
}



.compute_feature_similarity_cluster_table <- function(x) {
  # Computes the feature similarity cluster table from the similarity table in
  # x.
  
  # Create a cluster method object using data stored in x.
  cluster_method_object <- .create_feature_similarity_cluster_method_object(x = x)
  
  if (is.null(x)) return(NULL)
  
  # Compute the cluster table.
  cluster_table <- create_clusters(
    object = cluster_method_object,
    as_cluster_object = FALSE
  )
  
  return(cluster_table)
}



.create_feature_similarity_cluster_method_object <- function(x) {
  
  if (is_empty(x)) return(NULL)
  
  if (length(x@similarity_threshold) > 1L) {
    # Remove 1.0 because that does not yield clustering info.
    available_thresholds <- setdiff(x@similarity_threshold, 1.0)
    
    # Select the maximum threshold.
    x@similarity_threshold <- max(available_thresholds)
  }
  
  # Create cluster method object.
  cluster_method_object <- create_cluster_method_object(
    cluster_method = x@cluster_method,
    data_type = "feature",
    cluster_linkage = x@linkage_method,
    cluster_cut_method = x@cluster_cut_method,
    cluster_similarity_threshold = x@similarity_threshold,
    cluster_similarity_metric = x@similarity_metric,
    cluster_representation_method = "none"
  )
  
  # Attach the similarity table to the cluster_method_object.
  cluster_method_object@similarity_table <- methods::new(
    "similarityTable",
    data = x@data[, mget(c("feature_name_1", "feature_name_2", "value"))],
    similarity_metric = x@similarity_metric,
    data_type = cluster_method_object@data_type
  )
  
  return(cluster_method_object)
}



# export_feature_similarity (generic) ------------------------------------------

#'@title Extract and export mutual correlation between features.
#'
#'@description Extract and export mutual correlation between features in a
#'  familiarCollection.
#'
#'@param export_dendrogram Add dendrogram in the data element objects.
#'@param export_ordered_data Add feature label ordering to data in the data
#'  element objects.
#'@param export_clustering Add clustering information to data.
#'
#'@inheritParams export_all
#'@inheritParams .extract_data
#'@inheritParams plot_univariate_importance
#'
#'@inheritDotParams as_familiar_collection
#'@inheritDotParams as_data_object
#'
#'@details
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'  
#'  Feature similarity data can be created from `dataObject`, or `data.table` objects.
#'  For `data.table`, see \code{\link{as_data_object}} for additional arguments.
#'
#'@return A list containing a data.table (if `dir_path` is not provided), or
#'  nothing, as all data is exported to `csv` files.
#'@exportMethod export_feature_similarity
#'@md
#'@rdname export_feature_similarity-methods
setGeneric(
  "export_feature_similarity",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    export_dendrogram = FALSE,
    export_ordered_data = FALSE,
    export_clustering = FALSE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_feature_similarity")
  } 
)

# export_feature_similarity (collection) ---------------------------------------

#'@rdname export_feature_similarity-methods
setMethod(
  "export_feature_similarity",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    export_dendrogram = FALSE,
    export_ordered_data = FALSE,
    export_clustering = FALSE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Extract data.
    x <- object@feature_similarity
    
    if (export_ordered_data && export_clustering) {
      ..error("Cannot simultaneously export cluster information and ordering of features.")
    }
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    if (!is.waive(feature_cluster_method)) {
      
      # Set clustering method.
      x <- lapply(
        x,
        function(x, feature_cluster_method) {
          x@cluster_method <- feature_cluster_method
          return(x)
        },
        feature_cluster_method = feature_cluster_method
      )
    }
    
    if (!is.waive(feature_linkage_method)) {
      
      # Set linkage method.
      x <- lapply(
        x,
        function(x, feature_linkage_method) {
          x@linkage_method <- feature_linkage_method
          return(x)
        },
        feature_linkage_method = feature_linkage_method
      )
    }
    
    if (!is.waive(feature_cluster_cut_method)) {
      
      # Set cut method
      x <- lapply(
        x,
        function(x, feature_cluster_cut_method) {
          x@cluster_cut_method <- feature_cluster_cut_method
          return(x)
        },
        feature_cluster_cut_method = feature_cluster_cut_method
      )
    }
    
    if (!is.waive(feature_similarity_threshold)) {
      
      # Set similarity threshold.
      x <- lapply(
        x,
        function(x, feature_similarity_threshold) {
          x@similarity_threshold <- feature_similarity_threshold
          return(x)
        },
        feature_similarity_threshold = feature_similarity_threshold
      )
    }
    
    # Check whether the input parameters are valid and create a cluster object.
    .check_cluster_parameters(
      cluster_method = x[[1L]]@cluster_method,
      data_type = "feature",
      cluster_linkage = x[[1L]]@linkage_method,
      cluster_cut_method = x[[1L]]@cluster_cut_method,
      cluster_similarity_threshold = x[[1L]]@similarity_threshold,
      cluster_similarity_metric = x[[1L]]@similarity_metric,
      cluster_representation_method = "none"
    )
    
    if (
      aggregate_results ||
      export_dendrogram ||
      export_ordered_data ||
      export_clustering
    ) {
      x <- .compute_data_element_estimates(x)
      
      if (export_dendrogram || export_ordered_data || export_clustering) {
        # Add dendrogram and other cluster objects.
        x <- lapply(x, .append_feature_similarity_dendrogram)
      }
      
      if (export_clustering) {
        x <- lapply(x, .append_feature_similarity_clusters)
      }
      
      if (export_ordered_data) {
        # Add clustering information.
        x <- lapply(x, .append_feature_similarity_clustering)
      }
    }
    
    return(.export(
      x = object,
      data_elements = x,
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "feature_similarity",
      subtype = x[[1L]]@similarity_metric,
      export_dendrogram = export_dendrogram,
      export_ordered_data = export_ordered_data,
      export_collection = export_collection
    ))
  }
)



# export_feature_similarity (general) ------------------------------------------

#'@rdname export_feature_similarity-methods
setMethod(
  "export_feature_similarity",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "feature_similarity",
          "aggregate_results" = aggregate_results,
          "feature_cluster_method" = feature_cluster_method,
          "feature_linkage_method" = feature_linkage_method,
          "feature_cluster_cut_method" = feature_cluster_cut_method,
          "feature_similarity_threshold" = feature_similarity_threshold
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_feature_similarity,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "feature_cluster_method" = feature_cluster_method,
          "feature_linkage_method" = feature_linkage_method,
          "feature_cluster_cut_method" = feature_cluster_cut_method,
          "feature_similarity_threshold" = feature_similarity_threshold,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)



#.export (familiarDataElementFeatureSimilarity)---------------------------------
setMethod(
  ".export",
  signature(x = "familiarDataElementFeatureSimilarity"),
  function(
    x,
    x_list,
    aggregate_results = FALSE,
    export_dendrogram,
    export_ordered_data, 
    ...
  ) {
    # This is like .export,familiarDataElement, but the elements are merged
    # prior to computing estimates.
    
    # Only merge if dendrograms are missing for all entries.
    if (!export_dendrogram && !export_ordered_data) {
      # Merge data elements.
      x <- merge_data_elements(
        x = x_list,
        as_data = "all",
        as_grouping_column = TRUE,
        force_data_table = TRUE
      )
      
    } else {
      x <- x_list
    }
    
    if (aggregate_results) {
      x <- .compute_data_element_estimates(x)
    }
    
    return(x)
  }
)

#' @include PredictionTable.R

.get_available_stratification_methods <- function() {
  return(c("median", "fixed", "optimised", "mean", "mean_trim", "mean_winsor"))
}



.find_survival_grouping_thresholds <- function(object, data) {
  if (!object@outcome_type %in% c("survival")) {
    ..error_reached_unreachable_code(paste0(
      ".find_survival_grouping_thresholds: only available for ",
      "survival outcome. Found: ", object@outcome_type
    ))
  }
  
  # Load settings to find survival thresholds
  settings <- get_settings()

  # Set time_max
  time_max <- settings$eval$time_max

  # Generate prediction table
  prediction_table <- .predict(
    object = object,
    data = data,
    allow_recalibration = TRUE,
    time = time_max
  )

  # Check if any predictions are valid.
  if (!any_predictions_valid(prediction_table)) return(NULL)
  
  # Remove data with missing predictions.
  prediction_table <- remove_invalid_predictions(prediction_table)
  km_info_list <- list()

  # Iterate over stratification methods
  for (cut_off_method in settings$eval$strat_method) {
    if (cut_off_method == "median") {
      # Identify threshold
      cutoff <- .find_quantile_threshold(
        x = prediction_table,
        quantiles = 0.5
      )
      
    } else if (cut_off_method == "fixed") {
      # Identify thresholds
      cutoff <- .find_quantile_threshold(
        x = prediction_table,
        quantiles = settings$eval$strat_quant_threshold
      )
      
    } else if (cut_off_method == "optimised") {
      # Identify threshold
      cutoff <- .find_maxstat_threshold(x = prediction_table)
      
    } else if (cut_off_method %in% c("mean", "mean_winsor", "mean_trim")) {
      # Identify threshold
      cutoff <- .find_mean_threshold(
        x = prediction_table,
        method = cut_off_method
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        ".find_survival_grouping_thresholds: encountered an unknown ",
        "threshold type:", cut_off_method
      ))
    }

    # Find corresponding sizes of the generated groups
    risk_group <- .apply_risk_threshold(
      x = prediction_table,
      cutoff = cutoff
    )
    
    group_size <- .get_risk_group_sizes(risk_group = risk_group)

    # Populate method_list
    method_list <- list(
      "method" = cut_off_method,
      "cutoff" = cutoff,
      "group_size" = group_size
    )

    # Attach method_list to the general km_info_list
    km_info_list[[cut_off_method]] <- method_list
  }

  # Add stratification methods
  out_list <- list(
    "stratification_method" = settings$eval$strat_method,
    "parameters" = km_info_list,
    "time_max" = time_max
  )

  return(out_list)
}


# .find_mean_threshold ---------------------------------------------------------
setGeneric(".find_mean_threshold", function(x, ...) standardGeneric(".find_mean_threshold"))

setMethod(
  ".find_mean_threshold",
  signature(x = "predictionTableSurvival"),
  function(x, method, ...) {
    data_slot <- ifelse(.is_merged_prediction_table(x), "data", "prediction_data")
    values <- slot(x, data_slot)$predicted_outcome
    
    if (method == "mean") {
      return(mean(values))
      
    } else if (method == "mean_trim") {
      return(trim(values))
      
    } else if (method == "mean_winsor") {
      return(winsor(values))
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "The provided method for setting the mean risk stratification method was not recognised: ",
        method
      ))
    }
  }
)


# .find_quantile_threshold -----------------------------------------------------
setGeneric(".find_quantile_threshold", function(x, ...) standardGeneric(".find_quantile_threshold"))

setMethod(
  ".find_quantile_threshold",
  signature(x = "predictionTableSurvival"),
  function(x, quantiles, ...) {
    
    # Order quantiles in ascending order.
    quantiles <- quantiles[order(quantiles)]
    
    # Extract values
    data_slot <- ifelse(.is_merged_prediction_table(x), "data", "prediction_data")
    values <- slot(x, data_slot)$predicted_outcome
    
    # Return threshold values
    return(stats::quantile(
      x = values,
      probs = quantiles,
      names = FALSE,
      na.rm = TRUE
    ))
  }
)

setMethod(
  ".find_quantile_threshold",
  signature(x = "predictionTableSurvivalTime"),
  function(x, quantiles, ...) {
    # For time-like predictions, we should use the complements of the provided
    # quantiles.
    quantiles <- abs(1.0 - quantiles)
    
    # Pass to main method with updated quantiles.
    return(callNextMethod(x = x, quantiles = quantiles))
  }
)


# .find_maxstat_threshold ------------------------------------------------------
setGeneric(".find_maxstat_threshold", function(x, ...) standardGeneric(".find_maxstat_threshold"))

setMethod(
  ".find_maxstat_threshold",
  signature(x = "predictionTableSurvival"),
  function(x, ...) {
    require_package(
      x = "maxstat",
      purpose = "to determine an optimal risk threshold"
    )
    
    # Convert prediction_table to a data.table.
    x <- .as_data_table(x)
    
    # Check for invariant features.
    if (stats::var(x$predicted_outcome) == 0.0) {
      return(x$predicted_outcome[1L])
    }
    
    # Perform maxstat test
    h <- tryCatch(
      maxstat::maxstat.test(
        survival::Surv(outcome_time, outcome_event) ~ predicted_outcome,
        data = x,
        smethod = "LogRank",
        minprop = 0.10,
        maxprop = 0.90
      ),
      error = identity
    )
    
    # Check that maxstat.test did not produce an error.
    if (inherits(h, "error")) {
      return(mean(x$predicted_outcome, na.rm = TRUE))
    }
    
    # Check if at least 4 unique values are present for the smoothing spline
    if (length(h$cuts) < 4L) {
      return(unname(h$estimate))
    }
    
    # Smoothed scores
    spline_fit <- tryCatch(
      stats::smooth.spline(x = h$cuts, y = h$stats)$fit,
      error = identity
    )
    
    # Capture error.
    if (inherits(spline_fit, "error")) {
      return(unname(h$estimate))
    }
    
    # Predict scores on a fine grid
    x_sample_cuts <- seq(
      from = min(h$cuts),
      to = max(h$cuts),
      length.out = 100L
    )
    
    test_scores <- stats::predict(
      object = spline_fit, 
      x = x_sample_cuts
    )$y
    
    return(x_sample_cuts[which.max(test_scores)])
  }
)


# .apply_risk_threshold --------------------------------------------------------
setGeneric(".apply_risk_threshold", function(x, ...) standardGeneric(".apply_risk_threshold"))

setMethod(
  ".apply_risk_threshold",
  signature(x = "predictionTableSurvival"),
  function(x, cutoff, invert = FALSE) {
    
    # Convert to data table and get the predicted values.
    x <- .as_data_table(x)$predicted_outcome
    
    # Initialise risk group
    risk_group <- rep.int(1L, times = length(x))
    
    # Iterate over cutoffs and define risk groups
    for (current_cutoff in cutoff) {
      # We assume that risk groups go from 1 (low risk) to k (high risk), with
      # k-1 being the number of provided cutoff values.
      if (invert) {
        risk_group <- risk_group + as.numeric(x < current_cutoff)
      } else {
        risk_group <- risk_group + as.numeric(x >= current_cutoff)
      }
    }
    
    # Convert to factor
    risk_group <- .assign_risk_group_names(
      risk_group = risk_group,
      cutoff = cutoff
    )
    
    # Replace non-finite predicted values by NA.
    risk_group[!is.finite(x)] <- NA
    
    # Return risk groups
    return(risk_group)
  }
)

setMethod(
  ".apply_risk_threshold",
  signature(x = "predictionTableSurvivalTime"),
  function(x, cutoff, invert = TRUE) {
    # For prediction of survival times, pass to  
    return(
      callNextMethod(
        x = x,
        cutoff = cutoff,
        invert = TRUE
      )
    )
  }
)



.get_risk_group_sizes <- function(risk_group) {
  # Suppress NOTES due to non-standard evaluation in data.table
  indicated_group <- NULL

  # Find group sizes
  group_table <- data.table::data.table("indicated_group" = risk_group)
  group_table <- group_table[, list("group_size" = .N), by = "indicated_group"][order(indicated_group)]

  # Get group sizes and set names
  group_sizes <- group_table$group_size / length(risk_group)
  names(group_sizes) <- group_table$indicated_group

  return(group_sizes)
}




.assign_risk_group_names <- function(risk_group, cutoff) {
  # Determine the number of risk groups
  n_groups <- length(cutoff) + 1L

  if (n_groups == 2L) {
    # Stratification into low and high-risk groups
    y <- factor(
      x = risk_group, 
      levels = seq_len(n_groups), 
      labels = c("low", "high"), 
      ordered = TRUE
    )
    
  } else if (n_groups == 3L) {
    # Stratification into low, moderate and high-risk groups
    y <- factor(
      x = risk_group,
      levels = seq_len(n_groups),
      labels = c("low", "moderate", "high"),
      ordered = TRUE
    )
    
  } else {
    # Assign numbers
    y <- factor(
      x = risk_group, 
      levels = seq_len(n_groups),
      ordered = TRUE
    )
  }

  return(y)
}



get_mean_risk_group <- function(risk_group) {
  # Determine the mean average risk group. This requires discretisation
  # as rounding toward the nearest group would overinflate center groups.
  group_names <- levels(risk_group)
  n <- length(group_names)

  # Discretise bins floor((mu - 1) / ((n-1) / n)) + 1. See fixed bin size
  # discretisation.
  risk_group_num <- as.integer(floor(n * (mean(as.numeric(risk_group), na.rm = TRUE) - 1.0) / (n - 1.0))) + 1L

  # Check if the risk_group_num still falls within the range
  risk_group_num <- ifelse(risk_group_num > n, n, risk_group_num)

  return(factor(
    x = group_names[risk_group_num],
    levels = group_names, 
    ordered = TRUE
  ))
}

# Avoid warnings due to non-standard evaluation in data.table.
outcome <- outcome_time <- NULL

for (outcome_type in c("continuous", "binomial", "multinomial", "survival")) {
  # Generate data.
  data <- familiar:::test_create_good_data(outcome_type = outcome_type)
  data@data[, "batch_id" := "A"]
  data_b <- familiar:::test_create_good_data(outcome_type = outcome_type)
  data_b@data[, "batch_id" := "B"]
  data_c <- familiar:::test_create_good_data(outcome_type = outcome_type)
  data_c@data[, "batch_id" := "C"]
  data@data <- rbind(data@data, data_b@data, data_c@data)
  
  # Introduce data with offset.
  data_offset <- data
  data_offset@data <- data.table::copy(data_offset@data)
  if (outcome_type == "continuous") {
    data_offset@data[batch_id == "A", "outcome" := outcome + 50.0]
    
  } else if (outcome_type %in% c("binomial", "multinomial")) {
    data_offset@data[batch_id == "A", "outcome" := "red"]
    
  } else if (outcome_type == "survival") {
    data_offset@data[batch_id == "A", "outcome_time" := outcome_time + 1000.0]
  }
  
  # Test results
  testthat::expect_no_condition(
    familiar:::.check_batch_normalisation_assumptions(
      data = data,
      normalisation_method = "combat"
    )
  )
  
  testthat::expect_warning(
    familiar:::.check_batch_normalisation_assumptions(
      data = data_offset,
      normalisation_method = "combat"
    ),
    class = "familiar_batch_outcome_difference"
  )
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

.parse_hyperparameters <- function(
    data,
    parameter_list,
    outcome_type,
    vimp_method = NULL,
    learner = NULL,
    detector = NULL
) {
  
  # Check if any hyperparameters have been set
  if (length(parameter_list) == 0L) return(parameter_list)
  
  if (is.null(vimp_method) && is.null(learner) && is.null(detector)) {
    ..error_reached_unreachable_code(".parse_hyperparameters: one of vimp_method, learner, detector should not be NULL")
  }
  
  if (!is.null(vimp_method)) {
    all_functions <- vimp_method
    type <- "variable importance methods"
    
  } else if (!is.null(learner)) {
    all_functions <- learner
    type <- "learners"
    
  } else if (!is.null(detector)) {
    all_functions <- detector
    type <- "novelty detector"
  }
  
  # Find hyperparameter names.
  user_function_names <- names(parameter_list)
  
  # Check whether all names in the hyperparameter list correspond to a
  # vimp_method, learner or detector.
  unknown_function <- setdiff(user_function_names, all_functions)
  if (length(unknown_function) > 0L) {
    ..error(
      paste0(
        "Could not match all entries in the parameter list to ", type, ".\n",
        "Failed to match: ", paste_s(unknown_function)
      ),
      error_class = "input_argument_error"
    )
  }
  
  # Check for duplicates.
  if (anyDuplicated(user_function_names)) {
    ..error(
      "Parameters for one or more of the ", type, " are defined more than once.",
      error_class = "input_argument_error"
    )
  }
  
  # Package data into a data object.
  data <- methods::new(
    "dataObject",
    data = data,
    preprocessing_level = "none",
    outcome_type = outcome_type
  )
  
  # Iterate over the feature selection methods or learners to parse and check the hyperparameters
  out_list <- lapply(
    user_function_names,
    .check_hyperparameters,
    in_list = parameter_list,
    data = data,
    vimp_method = vimp_method,
    learner = learner,
    detector = detector
  )
  
  # Set names of the input parameters
  names(out_list) <- user_function_names

  return(out_list)
}



.check_hyperparameters <- function(
    user_function_name,
    in_list,
    data,
    vimp_method = NULL,
    learner = NULL,
    detector = NULL
) {
  
  # Select current parameter list
  in_list <- in_list[[user_function_name]]
  
  # Obtain preset values for hyperparameters.
  if (!is.null(vimp_method)) {
    preset_list <- .get_preset_hyperparameters(
      data = data,
      vimp_method = user_function_name,
      names_only = FALSE
    )
    
    type <- "variable importance method"
    
  } else if (!is.null(learner)) {
    preset_list <- .get_preset_hyperparameters(
      data = data,
      learner = user_function_name,
      names_only = FALSE
    )
    
    type <- "learner"
    
  } else {
    preset_list <- .get_preset_hyperparameters(
      data = data,
      detector = user_function_name,
      names_only = FALSE
    )
    
    type <- "novelty detector"
  }

  # Update user function name for passing into error messages
  user_function_name <- paste(user_function_name, type)
  
  # Check if the desired function actually has parameters.
  if (length(preset_list) == 0L) {
    ..error(
      "The ", user_function_name, " has no associated parameters.",
      error_class = "input_argument_error"
    )
  }
  
  # Check if there are any parameters in the user-provided list that do not appear in the preset list
  unknown_parameter <- setdiff(names(in_list), names(preset_list))
  if (length(unknown_parameter) > 0L) {
    ..error(
      paste0(
        "Could not match all parameters provided for the ",
        user_function_name,
        " to a valid parameter. Failed to match: ",
        paste_s(unknown_parameter)
      ),
      error_class = "input_argument_error"
    )
  }
  
  # Check if there are any duplicate parameters
  if (anyDuplicated(names(in_list))) {
    ..error(
      paste0(
        "One or more parameters provided for the ",
        user_function_name,
        " appear more than once: ",
        names(in_list)[duplicated(names(in_list))]
      ),
      error_class = "input_argument_error"
    )
  }
  
  # Iterate over the individual parameters and parse/check individual parameters
  out_list <- lapply(
    names(in_list),
    function(parameter_name, in_list, user_function_name, preset_list) {
      .check_single_hyperparameter(
        parameter_name,
        user_function_name = user_function_name,
        x = in_list[[parameter_name]],
        preset_list = preset_list[[parameter_name]]
      )
    },
    user_function_name = user_function_name,
    in_list = in_list,
    preset_list = preset_list
  )
  
  # Set names of the out_list
  names(out_list) <- names(in_list)

  return(out_list)
}



.check_single_hyperparameter <- function(
    parameter_name,
    user_function_name,
    x,
    preset_list
) {

  if (is.list(x) && length(x) == 1L) {
    # Unlist list. This happens when the hyperparameter settings come from a
    # configuration file.
    x <- unlist(x) 
  }
  
  if (is.character(x)) {
    # Divide by comma
    x <- strsplit(
      x = x,
      split = ",",
      fixed = TRUE
    )[[1L]]
    
    # Remove whitespace
    x <- gsub(
      x = x,
      pattern = " ",
      replacement = "",
      fixed = TRUE
    )
  }
  
  # Update the parameter_name for passing into error warnings
  parameter_name <- paste0(parameter_name, " of the ", user_function_name)
  
  # Convert input values
  x <- .perform_type_conversion(
    x = x,
    to_type = preset_list$type,
    var_name = parameter_name,
    req_length = 1L,
    allow_more = TRUE
  )
  
  # Check if the parameter has the allowed values
  if (!is.null(preset_list$valid_range)) {
    valid_range <- preset_list$valid_range
    
  } else {
    valid_range <- preset_list$range
  }

  if (preset_list$type %in% c("numeric", "integer")) {
    sapply(
      x,
      .check_number_in_valid_range,
      var_name = parameter_name,
      range = valid_range
    )
    
  } else {
    .check_parameter_value_is_valid(
      x = x,
      var_name = parameter_name,
      values = valid_range
    )
  }
  
  return(x)
}



.get_preset_hyperparameters <- function(
    data = NULL,
    vimp_method = NULL,
    learner = NULL,
    detector = NULL,
    outcome_type = NULL,
    names_only = FALSE
) {
  
  if (is.null(vimp_method) && is.null(learner) && is.null(detector)) {
    ..error_reached_unreachable_code(
      ".get_preset_hyperparameters: one of vimp_method, learner, detector should not be NULL"
    )
  }

  # Internal error checks. We should be able to obtain the outcome_type.
  if (is.null(detector)) {
    if (is.null(data) && is.null(outcome_type)) {
      ..error_reached_unreachable_code(".get_preset_hyperparameters: outcome_type is missing.")
      
    } else if (!is(data, "dataObject") && is.null(outcome_type)) {
      ..error_reached_unreachable_code(".get_preset_hyperparameters: outcome_type is missing.")
    }
    
    if (is.null(outcome_type)) outcome_type <- data@outcome_type
  }
  
  # Find the parameter list
  if (!is.null(vimp_method)) {
    preset_list <- .get_vimp_hyperparameters(
      data = data,
      method = vimp_method,
      outcome_type = outcome_type, 
      names_only = names_only
    )
    
  } else if (!is.null(learner)) {
    preset_list <- .get_learner_hyperparameters(
      data = data,
      learner = learner,
      outcome_type = outcome_type,
      names_only = names_only
    )
    
  } else if (!is.null(detector)) {
    preset_list <- .get_detector_hyperparameters(
      data = data,
      detector = detector,
      names_only = names_only
    )
    
  } else {
    ..error_reached_unreachable_code(
      ".get_preset_hyperparameters: one of vimp_method, learner, detector should not be NULL"
    )
  }
  
  return(preset_list)
}



.update_hyperparameters <- function(
    parameter_list,
    user_list = NULL,
    n_features = NULL
) {
  
  # Check if any parameters are provided
  if (length(parameter_list) == 0L) return(parameter_list)
  
  for (parameter_name in names(parameter_list)) {
    
    if (!is.null(user_list[[parameter_name]])) {
      
      # Obtain values provided by the user
      user_values <- user_list[[parameter_name]]
      
      # The signature size depends on the number of features, which may not be a
      # priori known by the user. To avoid errors, we force sign_size to be in
      # the dynamically determined preset range. All values over the maximum
      # range are silently set to the edge of the range. We don't enforce
      # uniqueness, as the number of values affects how this parameter is
      # interpreted.
      if (parameter_name == "sign_size") {
        user_values[user_values > n_features] <- n_features
      }
      
      if (length(user_values) == 1L) {
        # User provides one value for a parameter
        parameter_list[[parameter_name]]$init_config <- user_values
        parameter_list[[parameter_name]]$randomise <- FALSE
        
      } else if (length(user_values) > 1L) {

        if (parameter_list[[parameter_name]]$type %in% c("numeric", "integer")) {
          
          if (length(user_values) == 2L) {
            # Find initial values from the default and from the user-provided
            # range. Only values within the latter range are used.
            initial_values <- sort(unique(c(
              parameter_list[[parameter_name]]$init_config,
              user_values
            )))
            initial_values <- initial_values[initial_values >= user_values[1L] & initial_values <= user_values[2L]]
            
          } else {
            # For more than 2 values, copy the user values directly.
            initial_values <- user_values
          }
          
          parameter_list[[parameter_name]]$init_config <- initial_values
          parameter_list[[parameter_name]]$range <- user_values
          parameter_list[[parameter_name]]$randomise <- TRUE
          
        } else {
          # User provides multiple values for a parameter
          parameter_list[[parameter_name]]$init_config <- user_values
          parameter_list[[parameter_name]]$range <- user_values
          parameter_list[[parameter_name]]$randomise <- TRUE
        }
      
      } else {
        # This code should never be reached as such cases should be captures by
        # .parse_hyperparameters earlier in the workflow.
        ..error_reached_unreachable_code(paste0(
          ".update_hyperparameters: at least one user_values is expected."
        ))
      }
    }
    
    # Identify parameters that only allow for a single value.
    if (length(unique(parameter_list[[parameter_name]]$range)) == 1L) {
      parameter_list[[parameter_name]]$init_config <- parameter_list[[parameter_name]]$range[1L]
      parameter_list[[parameter_name]]$range <- parameter_list[[parameter_name]]$range[1L]
      parameter_list[[parameter_name]]$randomise <- FALSE
    }
    
    # Check if the parameters fall within the valid range.
    if (parameter_list[[parameter_name]]$type %in% c("numeric", "integer")) {
      sapply(
        parameter_list[[parameter_name]]$init_config,
        .check_number_in_valid_range,
        var_name = parameter_name,
        range = parameter_list[[parameter_name]]$valid_range
      )
      
    } else {
      sapply(
        parameter_list[[parameter_name]]$init_config,
        .check_parameter_value_is_valid,
        var_name = parameter_name,
        values = parameter_list[[parameter_name]]$valid_range
      )
    }
    
    if (!parameter_list[[parameter_name]]$randomise) {
      if (length(parameter_list[[parameter_name]]$init_config) > 1L) {
        ..error_reached_unreachable_code(paste0(
          ".update_hyperparameters: non-randomised hyperparameter (",
          parameter_name, ") contains more than one initial value."
        ))
      }
    }
  }
  
  return(parameter_list)
}



.any_randomised_hyperparameters <- function(parameter_list) {
  # Hyper-parameters are randomised by the sequential model boosting algorithm.
  # However, some hyperparameters may not require randomisation, or settings
  # were provided by the user. In case no hyper-parameters are to be randomised,
  # SMBO is not required, and the algorithm is the stop early. Therefore, this
  # function is used to check whether there are any randomisable
  # hyper-parameters.
  
  if (length(parameter_list) == 0L) return(FALSE)

  # Determine if any parameters require randomisation
  requires_randomisation <- sapply(
    parameter_list,
    function(list_entry) (list_entry$randomise)
  )
  
  # Return FALSE if no feature is randomised, and TRUE if any feature is randomised.
  return(sum(requires_randomisation) > 0L)
}


.set_hyperparameter <- function(
    default,
    type,
    range,
    valid_range = NULL,
    randomise = FALSE,
    distribution = NULL
) {
  
  # Initialise list
  hyperparameter <- list()
  
  # Set default values
  hyperparameter$init_config <- default
  
  # Set type of hyperparameter
  hyperparameter$type <- type
  
  # Check for accidental typos in type.
  if (!type %in% c("numeric", "integer", "factor", "logical")) {
    ..error_reached_unreachable_code(paste0(
      ".set_hyperparameter: type does not match one of the expected types: ", type
    ))
  }
  
  # Set range.
  hyperparameter$range <- range
  
  # Set the range of valid values.
  if (is.null(valid_range)) {
    hyperparameter$valid_range <- range
    
  } else {
    hyperparameter$valid_range <- valid_range
  }
  
  # Set randomisation.
  hyperparameter$randomise <- randomise
  
  if (!is.null(distribution)) {
    hyperparameter$rand_distr <- distribution
    
    # Check for typos in distribution.
    if (!distribution %in% c("log")) {
      ..error_reached_unreachable_code(paste0(
        ".set_hyperparameter: distribution does not match one of the expected distributions: ",
        distribution
      ))
    }
  }
  
  return(hyperparameter)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# plot_univariate_importance (generic) -----------------------------------------

#' @title Plot univariate importance.
#'
#' @description This function plots the univariate analysis data stored in a
#'   familiarCollection object.
#'
#' @param draw (*optional*) Draws the plot if TRUE.
#' @param dir_path (*optional*) Path to the directory where created figures are
#'   saved to. Output is saved in the `variable_importance` subdirectory. If
#'   NULL no figures are saved, but are returned instead.
#' @param p_adjustment_method (*optional*) Indicates type of p-value that is
#'   shown. One of `holm`, `hochberg`, `hommel`, `bonferroni`, `BH`, `BY`,
#'   `fdr`, `none`, `p_value` or `q_value` for adjusted p-values, uncorrected
#'   p-values
#' @param show_cluster (*optional*) Show which features were clustered together.
#' @param ggtheme (*optional*) `ggplot` theme to use for plotting.
#' @param discrete_palette (*optional*) Palette for colouring the plot elements
#'   according to the groupings indicated by the `color_by` argument (if any).
#'   `familiar` has a default palette. Other palettes are supported by
#'   `paletteer`, `grDevices::palette.pals()` (requires R >= 4.0.0),
#'   `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`, `heat.colors`,
#'   `terrain.colors`, `topo.colors` and `cm.colors`, which correspond to the
#'   palettes of the same name in `grDevices`. You may also specify your own
#'   palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param gradient_palette (*optional*) Palette to use for filling the bars in
#'   case the `color_by` argument is not set. The bars are then coloured
#'   according to their importance. By default, no gradient is used, and the
#'   bars are not filled according to importance. Use `NULL` to fill the bars
#'   using the default palette in `familiar`. Other palettes are supported by
#'   the `paletteer` package, `grDevices::palette.pals()` (requires R >= 4.0.0),
#'   `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`, `heat.colors`,
#'   `terrain.colors`, `topo.colors` and `cm.colors`, which correspond to the
#'   palettes of the same name in `grDevices`. You may also specify your own
#'   palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param significance_level_shown Position(s) to draw vertical lines indicating
#'   a significance level, e.g. 0.05. Can be NULL to not draw anything.
#' @param height (*optional*) Height of the plot. A default value is derived
#'   from the number of features and the number of facets.
#' @param width (*optional*) Width of the plot. A default value is derived from
#'   the number of facets.
#' @param units (*optional*) Plot size unit. Either `cm` (default), `mm` or
#'   `in`.
#' @param verbose Flag to indicate whether feedback should be provided for the
#'   plotting.
#' @param export_collection (*optional*) Exports the collection if TRUE.
#' @inheritParams export_univariate_analysis_data
#' @inheritParams export_feature_similarity
#' @inheritParams as_familiar_collection
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_univariate_analysis -object -feature_cluster_method -feature_linkage_method -feature_cluster_cut_method -verbose
#'
#' @details This function generates a horizontal barplot with the length of the
#'   bars corresponding to the 10-logarithm of the (multiple-testing corrected)
#'   p-value.
#'
#'   Features are assessed univariately using one-sample location t-tests after
#'   fitting a suitable regression model. The fitted model coefficient and the
#'   covariance matrix are then used to compute a p-value.
#'
#'   The following splitting variables are available for `split_by`, `color_by`
#'   and `facet_by`:
#'
#'  * `vimp_method`: variable importance methods
#'
#'  * `learner`: learners
#'
#'  * `data_set`: data sets
#'
#'   Unlike for plots of feature ranking in initial variable importance
#'   computation and after modelling (as assessed by model-specific routines),
#'   clusters of features are now found during creation of underlying
#'   `familiarData` objects, instead of through consensus clustering. Hence,
#'   clustering results may differ due to differences in the underlying
#'   datasets.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_feature_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_univariate_importance
#' @md
#' @rdname plot_univariate_importance-methods
setGeneric(
  "plot_univariate_importance",
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    p_adjustment_method = waiver(),
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    show_cluster = TRUE,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = waiver(),
    x_label = waiver(),
    y_label = "feature",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    significance_level_shown = 0.05,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    verbose = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_univariate_importance")
  }
)



# plot_univariate_importance (general) -----------------------------------------

#' @rdname plot_univariate_importance-methods
setMethod(
  "plot_univariate_importance",
  signature(object = "ANY"),
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    p_adjustment_method = waiver(),
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    show_cluster = TRUE,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = waiver(),
    x_label = waiver(),
    y_label = "feature",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    significance_level_shown = 0.05,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    verbose = TRUE,
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "univariate_analysis",
          "feature_cluster_method" = feature_cluster_method,
          "feature_linkage_method" = feature_linkage_method,
          "feature_cluster_cut_method" = feature_cluster_cut_method,
          "feature_similarity_threshold" = feature_similarity_threshold,
          "verbose" = verbose
        ),
        list(...)
      )
    )

    return(do.call(
      plot_univariate_importance,
      args = list(
        "object" = object,
        "feature_cluster_method" = feature_cluster_method,
        "feature_linkage_method" = feature_linkage_method,
        "feature_cluster_cut_method" = feature_cluster_cut_method,
        "feature_similarity_threshold" = feature_similarity_threshold,
        "draw" = draw,
        "dir_path" = dir_path,
        "p_adjustment_method" = p_adjustment_method,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "show_cluster" = show_cluster,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "gradient_palette" = gradient_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "limit_n_features" = limit_n_features,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "significance_level_shown" = significance_level_shown,
        "width" = width,
        "height" = height,
        "units" = units,
        "verbose" = verbose,
        "export_collection" = export_collection
      )
    ))
  }
)

# plot_univariate_importance (collection) --------------------------------------

#' @rdname plot_univariate_importance-methods
setMethod(
  "plot_univariate_importance",
  signature(object = "familiarCollection"),
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    p_adjustment_method = waiver(),
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    show_cluster = TRUE,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = waiver(),
    x_label = waiver(),
    y_label = "feature",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    significance_level_shown = 0.05,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    verbose = TRUE,
    export_collection = FALSE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    value <- .NATURAL <- NULL

    # Make sure the collection object is updated.
    object <- update_object(object = object)

    # Check data ---------------------------------------------------------------

    # Get input data.
    x <- export_univariate_analysis_data(
      object = object,
      p_adjustment_method = p_adjustment_method
    )

    # Use only the univariate data.
    x <- x$univariate

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check that the data are not evaluated at the model level.
    if (all(sapply(x, function(x) (x@detail_level == "model")))) {
      ..warning_no_comparison_between_models()
      return(NULL)
    }

    # Obtain data element from list.
    if (is.list(x)) {
      if (is_empty(x)) return(NULL)

      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_univariate_importance: list of data elements contains unmerged elements."
        )
      }

      # Get x directly.
      x <- x[[1L]]
    }

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check show_cluster
    .check_parameter_value_is_valid(
      x = show_cluster, 
      var_name = "show_cluster",
      values = c(FALSE, TRUE)
    )

    if (show_cluster) {
      # Get feature similarity data.
      feature_similarity <- export_feature_similarity(
        object = object,
        feature_cluster_method = feature_cluster_method,
        feature_linkage_method = feature_linkage_method,
        feature_cluster_cut_method = feature_cluster_cut_method,
        feature_similarity_threshold = feature_similarity_threshold,
        export_dendrogram = FALSE,
        export_ordered_data = FALSE,
        export_clustering = TRUE
      )[[1L]]
      
    } else {
      feature_similarity <- NULL
    }

    # Update data --------------------------------------------------------------

    # Set default adjust method.
    if (is.waive(p_adjustment_method)) p_adjustment_method <- "fdr"

    # Determine the metric column.
    column_data <- .plot_univariate_check_p_adjustment_method(
      method = p_adjustment_method,
      x = x@data,
      verbose = verbose
    )
    
    if (is.null(column_data)) return(NULL)

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to plot univariate variable importance",
      message_type = "warning"
    )) {
      return(NULL)
    }

    data.table::setnames(
      x = x@data,
      old = column_data$value_column, 
      new = "value"
    )

    # Update p-values below the machine precision (i.e. p=0.0).
    x@data[value <= 0.0, "value" := 2.0 * .Machine$double.eps]

    # Convert p-value column to logarithmic scale
    x@data[, "log_value" := -log10(value)]
    
    # Check input arguments ----------------------------------------------------
    
    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # significance_level_shown
    if (!is.null(significance_level_shown)) {
      sapply(
        significance_level_shown, 
        .check_number_in_valid_range,
        var_name = "significance_level_shown",
        range = c(0.0, 1.0),
        closed = c(FALSE, TRUE)
      )
    }

    # x_range
    if (is.null(x_range) && is.null(significance_level_shown)) {
      x_range <- c(0.0, max(x@data$log_value, na.rm = TRUE))
      
    } else if (is.null(x_range) && !is.null(significance_level_shown)) {
      x_range <- c(
        0.0,
        max(c(
          max(x@data$log_value, na.rm = TRUE),
          max(-log10(significance_level_shown))
        ))
      )
    }

    # x_breaks
    if (is.null(x_breaks)) {
      .check_input_plot_args(
        x_range = x_range, 
        x_n_breaks = x_n_breaks
      )

      # Create breaks and update x_range
      x_breaks <- labeling::extended(
        m = x_n_breaks,
        dmin = x_range[1L],
        dmax = x_range[2L],
        only.loose = TRUE
      )
      
      x_range <- c(0.0, tail(x_breaks, n = 1L))
    }

    # x_label Set default name for x-axis label
    if (is.waive(x_label)) {
      label_name <- column_data$label

      x_label <- bquote(-log[10] * "(" * .(label_name) * ")")
    }

    # Clusters cannot be generated in case no cluster information is present.
    if (is_empty(feature_similarity)) {
      show_cluster <- FALSE
    }

    # Set default splitting variables if none are provided.
    if (is.null(split_by) && is.null(color_by) && is.null(facet_by)) {
      split_by <- c("vimp_method", "learner")
      color_by <- c("data_set")
    }

    # Check splitting variables and generate sanitised output.
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      color_by = color_by,
      facet_by = facet_by,
      available = c("vimp_method", "learner", "data_set")
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    color_by <- split_var_list$color_by
    facet_by <- split_var_list$facet_by

    # Parse legend label
    legend_label <- .create_plot_legend_title(
      user_label = legend_label, 
      color_by = color_by
    )

    # Check general input arguments
    .check_input_plot_args(
      x_range = x_range,
      x_breaks = x_breaks,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      facet_wrap_cols = facet_wrap_cols,
      limit_n_features = limit_n_features
    )

    # Create plots -------------------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data
    if (!is.null(split_by)) {
      x_split <- split(x@data, by = split_by)
    } else {
      x_split <- list(x@data)
    }

    # Store plots to list in case no dir_path is provided
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over splits
    for (x_sub in x_split) {
      if (is_empty(x_sub)) next

      # Join cluster and univariate data.
      if (show_cluster) {
        # Join on common columns. These are feature plus any column in split_by.
        x_temporary <- feature_similarity@data[x_sub, on = .NATURAL]

        # Check that the resulting data is not empty, because this would mean
        # that there is e.g. only one feature.
        if (is_empty(x_temporary)) {
          x_temporary <- data.table::copy(x_sub)
          x_temporary[, ":="(
            "cluster_id" = .I,
            "cluster_size" = 1L
          )]
        }

        # Replace x_sub
        x_sub <- x_temporary
      }

      if (is.waive(plot_title)) plot_title <- "Univariate variable importance"

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          x = x_sub
        )
      }

      # Generate plot
      p <- .plot_univariate_importance(
        x = x_sub,
        color_by = color_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        show_cluster = show_cluster,
        discrete_palette = discrete_palette,
        gradient_palette = gradient_palette,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        limit_n_features = limit_n_features,
        x_range = x_range,
        x_breaks = x_breaks,
        significance_level_shown = significance_level_shown
      )

      # Check empty output
      if (is.null(p)) next
      
      # Draw plot
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_univariate_importance_plot_dimensions(
          x = x_sub,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "variable_importance",
              "subtype" = "univariate",
              "x" = x_sub,
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list and export
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



#' Internal plotting function for univariate plots
#'
#' @inheritParams plot_univariate_importance
#'
#' @return ggplot plot object.
#'
#' @seealso
#' * \code{\link{plot_univariate_importance}} for the user interface.
#' @md
#' @keywords internal
.plot_univariate_importance <- function(
    x,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    show_cluster,
    discrete_palette,
    gradient_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    limit_n_features,
    x_range,
    x_breaks,
    significance_level_shown
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  log_value <- feature <- NULL
  
  # Create local copy
  x <- data.table::copy(x)

  # Drop levels and reorder table so that features are sorted by the log-values
  x$feature <- droplevels(x$feature)
  x <- x[order(log_value)]
  x$feature <- factor(x$feature, levels = unique(x$feature))

  # Generate a guide table
  guide_list <- .create_plot_guide_table(
    x = x,
    color_by = color_by,
    discrete_palette = discrete_palette
  )

  # Extract data
  x <- guide_list$data
  
  # Determine the features that need to be plotted.
  if (is.numeric(limit_n_features)) {
    # Find the features with the highest log-p-values, and select these.
    x_agg <- x[, list("log_value" = max(log_value)), by = c("feature")]
    
    # Explicitly select features based on threshold value instead of simply 
    # selecting the best features: features may have the same value because they
    # belong to the same cluster.
    threshold_value <- min(tail(unique(sort(x_agg$log_value)), n = limit_n_features))
    selected_features <- x_agg[log_value >= threshold_value]$feature
    
    # Make slice.
    x <- x[feature %in% selected_features, ]
  }
  
  # Check if cluster information should be shown.
  if (show_cluster) {
    x <- .add_plot_cluster_name(
      x = x,
      color_by = color_by,
      facet_by = facet_by
    )
  }

  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("feature"), 
      y = !!sym("log_value")
    )
  )
  p <- p + ggtheme

  # Add fill colors
  if (!is.null(color_by)) {
    # Extract guide_table for color
    g_color <- guide_list$guide_color

    p <- p + ggplot2::geom_bar(
      stat = "identity",
      mapping = ggplot2::aes(fill = !!sym("color_breaks")),
      position = "dodge"
    )

    p <- p + ggplot2::scale_fill_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )
    
  } else if (!is.waive(gradient_palette)) {
    # Coloring by log of the p-value
    p <- p + ggplot2::geom_bar(
      stat = "identity",
      mapping = ggplot2::aes(fill = !!sym("log_value")),
      show.legend = FALSE
    )
    p <- p + ggplot2::scale_fill_gradientn(
      colors = .get_palette(
        x = gradient_palette,
        palette_type = "sequential"
      ),
      limits = x_range
    )
    
  } else {
    # No colouring of the bars.
    p <- p + ggplot2::geom_bar(stat = "identity")
  }

  # Set breaks and limits
  p <- p + ggplot2::scale_y_continuous(breaks = x_breaks)
  p <- p + ggplot2::coord_flip(ylim = x_range)

  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context"
      )
    }
  }

  # Show clusters
  if (show_cluster) {
    # Obtain default settings.
    text_settings <- .get_plot_geom_text_settings(ggtheme = ggtheme)

    if (is.null(color_by)) {
      p <- p + ggplot2::geom_text(
        ggplot2::aes(
          label = !!sym("cluster_name"),
          y = 0.0,
          angle = 270.0
        ),
        colour = text_settings$colour,
        family = text_settings$family,
        fontface = text_settings$face,
        size = text_settings$geom_text_size,
        vjust = -0.50
      )
      
    } else {
      p <- p + ggplot2::geom_text(
        ggplot2::aes(
          label = !!sym("cluster_name"),
          y = 0.0,
          group = !!sym("color_breaks"),
          angle = 270.0
        ),
        colour = text_settings$colour,
        family = text_settings$family,
        fontface = text_settings$face,
        size = text_settings$geom_text_size,
        vjust = -0.50,
        position = ggplot2::position_dodge(width = 0.9)
      )
    }
  }
  
  # Add vertical lines for significance levels.
  if (!is.null(significance_level_shown)) {
    for (curr_signif_level in significance_level_shown) {
      p <- p + ggplot2::geom_hline(
        yintercept = -log10(curr_signif_level),
        linetype = "dotted"
      )
    }
  }

  # Update labels. Note that the inversion of x_label and y_label is correct, as
  # the coordinates were flipped
  p <- p + ggplot2::labs(
    x = y_label,
    y = x_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  return(p)
}



.determine_univariate_importance_plot_dimensions <- function(
    x,
    facet_by, 
    facet_wrap_cols
) {
  
  # Get plot layout dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Determine the number of features within each facet.
  n_features <- data.table::uniqueN(x = x$feature)
  longest_name <- max(sapply(levels(x$feature), nchar))

  # Assume each feature takes up about 14 points (~5mm) with 2 point (0.7mm)
  # spacing. Then add some room for other plot elements.
  default_height <- n_features * 0.5 + (n_features - 1L) * 0.07 + 1.0

  # Set a default height. Assume that the typical width of a character is about
  # 5 points (1.8mm).
  default_width <- 6.0 + longest_name * 0.18

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))

  return(c(height, width))
}



.plot_univariate_check_p_adjustment_method <- function(
    x, 
    method, 
    verbose = TRUE
) {
  .check_parameter_value_is_valid(
    x = method,
    var_name = "p_adjustment_method",
    values = c(stats::p.adjust.methods, "p_value", "q_value")
  )

  value_column <- switch(
    method,
    "holm" = "adjusted_p_value",
    "hochberg" = "adjusted_p_value",
    "hommel" = "adjusted_p_value",
    "bonferroni" = "adjusted_p_value",
    "BH" = "adjusted_p_value",
    "BY" = "adjusted_p_value",
    "fdr" = "adjusted_p_value",
    "none" = "p_value",
    "p_value" = "p_value"
  )

  label_name <- switch(
    method,
    "holm" = "Holm corrected p-value",
    "hochberg" = "Hochberg corrected p-value",
    "hommel" = "Hommel corrected p-value",
    "bonferroni" = "Bonferroni corrected p-value",
    "BH" = "Benjamini-Hochberg corrected p-value",
    "BY" = "Benjamini-Yekutieli corrected p-value",
    "fdr" = "FDR-corrected p-value",
    "none" = "p-value",
    "p_value" = "p-value"
  )

  # Check if the column is present.
  if (is.null(x[[value_column]])) {
    ..error(paste0(
      "Univariate importance can not be plotted as the values for ",
      "the requested p adjustment method (", value_column, 
      ") were not found in the data."
    ))
  }

  # Check if any values are valid.
  if (!any(is.finite(x[[value_column]]))) {
    if (verbose) {
      ..warning("Univariate importance can not be plotted as all values are NA.")
    }

    return(NULL)
  }

  return(list(
    "value_column" = value_column,
    "label" = label_name
  ))
}

# Simple ensemble aggregation, as in Wald R, Khoshgoftaar TM, Dittman D, Awada
# W, Napolitano A. An extensive comparison of feature ranking aggregation
# techniques in bioinformatics. 2012 IEEE 13th Int. Conf. Inf. Reuse Integr.,
# IEEE; 2012, p. 377-84.

.compute_rank_mean_score <- function(x) {
  # Aggregate using mean score.

  # Suppress NOTES due to non-standard evaluation in data.table
  score <- NULL

  # Compute mean score
  x@vimp_table <- x@vimp_table[, list("score" = mean(score)), by = c("name")]

  return(x)
}



.compute_rank_mean_rank <- function(x) {
  # Aggregate using mean rank

  # Suppress NOTES due to non-standard evaluation in data.table
  rank <- NULL

  # Calculate mean rank
  x@vimp_table <- x@vimp_table[, list("score" = mean(rank)), by = c("name")]

  # Set correct invert value.
  x@invert <- FALSE

  return(x)
}



.compute_rank_median_rank <- function(x) {
  # Aggregate using median rank

  # Suppress NOTES due to non-standard evaluation in data.table
  rank <- NULL

  # Calculate median rank
  x@vimp_table <- x@vimp_table[, list("score" = as.double(stats::median(rank))), by = c("name")]

  # Set correct invert value.
  x@invert <- FALSE

  return(x)
}



.compute_rank_best_rank <- function(x) {
  # Aggregate using the best rank achieved.

  # Suppress NOTES due to non-standard evaluation in data.table
  rank <- NULL

  # Calculate best rank (i.e. lowest rank)
  x@vimp_table <- x@vimp_table[, list("score" = min(rank)), by = c("name")]

  # Set correct invert value.
  x@invert <- FALSE

  return(x)
}



.compute_rank_worst_rank <- function(x) {
  # Aggregate using the worst rank achieved

  # Suppress NOTES due to non-standard evaluation in data.table
  rank <- NULL

  # Calculate worst rank (i.e. highest rank)
  x@vimp_table <- x@vimp_table[, list("score" = max(rank)), by = c("name")]

  # Set correct invert value.
  x@invert <- FALSE

  return(x)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

setClass(
  "familiarRandomVimp",
  contains = "familiarVimpMethod"
)

setClass(
  "familiarNoneVimp",
  contains = "familiarVimpMethod"
)

setClass(
  "familiarSignatureVimp",
  contains = "familiarVimpMethod"
)

setClass(
  "familiarNoFeaturesVimp",
  contains = "familiarVimpMethod"
)



.get_available_random_vimp_methods <- function(show_general = TRUE) {
  return("random")
}

.get_available_none_vimp_methods <- function(show_general = TRUE) {
  return("none")
}

.get_available_signature_only_vimp_methods <- function(show_general = TRUE) {
  return("signature_only")
}

.get_available_no_features_vimp_methods <- function(show_general = TRUE) {
  return("no_features")
}



# is_available (random) --------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarRandomVimp"),
  function(object, ...) {
    return(TRUE)
  }
)



# is_available (none) ----------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarNoneVimp"),
  function(object, ...) {
    return(TRUE)
  }
)

# is_available (signature only) ------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarSignatureVimp"),
  function(object, ...) {
    return(TRUE)
  }
)

# is_available (signature only) ------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarNoFeaturesVimp"),
  function(object, ...) {
    return(TRUE)
  }
)

do.call_strict <- function(
    what,
    args,
    quote = FALSE,
    envir = parent.frame()
) {
  # Only pass fully matching arguments. Side effect is that ... is always empty.
  # Use with care in case arguments need to be passed to another function.

  # Get arguments that can be passed.
  passable_argument_names <- intersect(
    names(formals(what)),
    names(args)
  )

  return(do.call(
    what, 
    args = args[passable_argument_names],
    quote = quote, 
    envir = envir
  ))
}



do.call_with_timeout <- function(
    what, 
    args, 
    timeout, 
    quote = FALSE, 
    envir = parent.frame(),
    muffle = TRUE, 
    additional_packages = NULL
) {
  
  # Set up the muffler that captures all output.
  muffle_fun <- identity
  if (muffle) muffle_fun <- quiet

  if (!require_package(
    "callr",
    purpose = "to allow for executing functions with timeout",
    message_type = "warning"
  )) {
    # If callr cannot be loaded, execute the function as is.
    muffle_fun(
      value <- do.call(
        what = what,
        args = args, 
        envir = envir
      )
    )

    # Return data.
    return(list(
      "value" = value,
      "timeout" = FALSE
    ))
    
  } else if (!is.finite(timeout)) {
    # If there is no finite timeout, execute the function as is.
    muffle_fun(
      value <- do.call(
        what = what, 
        args = args, 
        envir = envir
      )
    )

    # Return data.
    return(list(
      "value" = value,
      "timeout" = FALSE
    ))
    
  } else {
    bg_function_wrapper <- function(what, args, additional_packages) {
      if (!is.null(bg_function_wrapper)) {
        for (package in additional_packages) {
          requireNamespace(package)
        }
      }

      do.call(what = what, args = args)
    }

    # Launch as background to allow for non-blocking timeout checks.
    bg_process <- callr::r_bg(
      func = bg_function_wrapper,
      args = list(
        "what" = what,
        "args" = args,
        "additional_packages" = additional_packages
      ),
      package = TRUE
    )

    # Wait until the timeout, or until the background process completes,
    # whichever happens first.
    bg_process$wait(timeout)

    # If the process is still running past the timeout, kill it.
    if (bg_process$is_alive()) {
      bg_process$kill()

      return(list("value" = NULL, "timeout" = TRUE))
      
    } else {
      return(list("value" = bg_process$get_result(), "timeout" = FALSE))
    }
  }
}



do.call_with_handlers_timeout <- function(
    what, 
    args, 
    timeout, 
    quote = FALSE, 
    envir = parent.frame(), 
    muffle = TRUE, 
    additional_packages = NULL
) {
  # Pass to do.call_with_timeout, which then calls do.call_with_handlers on
  # "what" with specified arguments "args".
  results <- do.call_with_timeout(
    what = do.call_with_handlers,
    args = list(
      "what" = what,
      "args" = args,
      "quote" = quote,
      "envir" = envir,
      "muffle" = muffle
    ),
    timeout = timeout,
    muffle = muffle,
    additional_packages = additional_packages
  )

  # Flatten results.
  if (!is.null(results$value)) {
    results$error <- results$value$error
    results$warning <- results$value$warning
    results$value <- results$value$value
  }

  return(results)
}



do.call_with_handlers <- function(
    what, 
    args, 
    quote = FALSE, 
    envir = parent.frame(),
    muffle = TRUE
) {
  # Set up the muffler that captures all output.
  muffle_fun <- identity
  if (muffle) muffle_fun <- quiet

  # Set up initial warning and error logs. Note that the error log is only
  # written after the tryCatch.
  warn_logs <- error_logs <- NULL

  muffle_fun(
    value <- tryCatch(
      withCallingHandlers(
        do.call(what = what, args = args, envir = envir),
        warning = function(w) {
          # Since warn_logs is found in the enclosing frame, we either have to
          # use <<- to access and update this variable. However, this causes
          # CRAN notes, because <<- can be a global assignment, i.e. in the user
          # space. Hence we explicitly use get and assign to update the
          # variable.
          warn_logs <- assign(
            "warn_logs",
            append(get("warn_logs"), condition_parser(w)),
            inherits = TRUE
          )
          
          # This prevents any warnings from being generated outside the
          # function.
          invokeRestart("muffleWarning")
        }
      ),
      error = identity
    )
  )
  
  if (inherits(value, "error")) {
    error_logs <- condition_parser(value)
    value <- NULL
  }

  return(list(
    "value" = value,
    "warning" = warn_logs,
    "error" = error_logs
  ))
}



do.call_external <- function(
    what, 
    args, 
    additional_packages = NULL
) {
  
  bg_function_wrapper <- function(
    what, 
    args, 
    additional_packages
  ) {
    if (!is.null(bg_function_wrapper)) {
      for (package in additional_packages) {
        requireNamespace(package)
      }
    }
    
    do.call(
      what = what,
      args = args
    )
  }
  
  # Launch as background.
  bg_process <- callr::r_bg(
    func = bg_function_wrapper,
    args = list(
      "what" = what,
      "args" = args,
      "additional_packages" = additional_packages
    ),
    package = TRUE
  )
  
  # Wait until the background process completes.
  bg_process$wait()
  
  # Check if the external process crashed.
  if (bg_process$get_exit_status() == 0L) {
    # If TRUE, results can be read.
    results <- bg_process$get_result()
    
  } else {
    # If FALSE, the external process disconnected.
    results <- list("error" = paste0("External process was lost with exit status", bg_process$get_exit_status()))
  }
  
  return(results)
}



condition_parser <- function(x, ...) {
  # Based on print.condition

  # Find information on the condition message and call.
  condition_message <- conditionMessage(x)
  condition_call <- conditionCall(x)

  # Parse condition to string.
  if (!is.null(condition_call)) {
    # Get the condition call. Also, just in case, select only the first section
    # that contains the function itself.
    deparsed_condition_call <- deparse1(condition_call)[1L]

    # We remove the function arguments from the call to prevent leaking data, if
    # any. Identify the first parenthesis that opens the argument section.
    parenthesis_position <- regexpr(
      text = deparsed_condition_call,
      pattern = "(",
      fixed = TRUE
    )[1L]

    # Check if a parenthesis appears in the initial section of the call.
    if (parenthesis_position > -1L) {
      # Select the function name which appears prior to the parenthesis.
      deparsed_condition_call <- substr(
        deparsed_condition_call,
        start = 1L,
        stop = parenthesis_position - 1L
      )
      
    } else {
      # If no parenthesis appears, the function call was likely cut off prior to
      # the argument section. We mark this abbreviation.
      deparsed_condition_call <- paste0(deparsed_condition_call, "[...]")
    }

    if (nchar(deparsed_condition_call) > 0L) {
      # Add in the deparsed condition call, and combine.
      parsed_condition <- paste0(
        deparsed_condition_call, ": ",
        condition_message
      )
      
    } else {
      parsed_condition <- condition_message
    }
    
  } else {
    # Set message.
    parsed_condition <- condition_message
  }

  return(parsed_condition)
}



condition_summary <- function(x) {
  # Convert x to a data.table.
  data <- data.table::data.table("condition_message" = x)
  
  # Count the number of times each unique condition message appears.
  data <- data[, list("n" = .N, "id" = .GRP), by = "condition_message"]
  
  # Parse each line.
  summary_vector <- sapply(
    split(data, by = "condition_message"),
    function(x) {
      # Initialise using the id, e.g. "3. "
      message_string <- paste0(as.character(x$id), ". ")
      
      # Show the number of times the condition has appeared, but ignore if it
      # only appeared once.
      if (x$n > 1L) {
        message_string <- c(message_string, paste0("(", x$n, "x) "))
      }
      
      # Add the condition message itself.
      message_string <- c(message_string, x$condition_message)
      
      return(paste0(message_string, collapse = ""))
    },
    USE.NAMES = FALSE
  )
  
  return(unname(summary_vector))
}



suppress_warnings <- function(
    expr,
    classes = "warning",
    regexp = NULL
) {
  withCallingHandlers(
    expr,
    warning = function(w) {
      # Attempt to silence a warning. A warning is silenced if it is one of the
      # provided classes. In addition, if regexp is set, it is silenced only
      # when the warning matches one of the classes, and has the regexp in its
      # message.
      if (rlang::inherits_any(w, classes)) {
        if (is.null(regexp)) {
          tryInvokeRestart("muffleWarning")
          
        } else if (any(sapply(regexp, grepl, x = conditionMessage(w)))) {
          tryInvokeRestart("muffleWarning")
        }
      }
    }
  )
}



quiet <- function(x) {
  # Removes all output to console.

  sink(nullfile())
  on.exit(sink())

  invisible(utils::capture.output(x, file = nullfile(), type = "message"))
}

testthat::skip_on_cran()
testthat::skip_on_ci()

test_fun_normal <- function(x = "default") {
  Sys.sleep(1)
  return(x)
}

test_fun_warning <- function(x = "default") {
  warning("warning")
  Sys.sleep(1)
  return(x)
}

test_fun_error <- function(x = "default") {
  warning("warning")
  Sys.sleep(1)
  stop("error")
  return(x)
}

testthat::test_that("Processes that complete within the timeout function correctly", {
  results_normal <- familiar:::do.call_with_timeout(
    test_fun_normal, 
    list("x" = "done"),
    timeout = 10000)
  
  testthat::expect_equal(results_normal$value, "done")
  testthat::expect_equal(results_normal$timeout, FALSE)
  
  results_warning <- familiar:::do.call_with_timeout(
    test_fun_warning, 
    list("x" = "done"),
    timeout = 10000)
  
  testthat::expect_equal(results_warning$value, "done")
  testthat::expect_equal(results_warning$timeout, FALSE)
  
  testthat::expect_error(
    familiar:::do.call_with_timeout(
      test_fun_error, 
      list("x" = "done"), 
      timeout = 10000), 
    "error")
})

testthat::test_that("Processes that complete outside the timeout fail", {
  results_normal <- familiar:::do.call_with_timeout(
    test_fun_normal, 
    list("x" = "done"),
    timeout = 500)

  testthat::expect_equal(is.null(results_normal$value), TRUE)
  testthat::expect_equal(results_normal$timeout, TRUE)

  results_warning <- familiar:::do.call_with_timeout(
    test_fun_warning, 
    list("x" = "done"),
    timeout = 500)

  testthat::expect_equal(is.null(results_warning$value), TRUE)
  testthat::expect_equal(results_warning$timeout, TRUE)

  results_error <- familiar:::do.call_with_timeout(
    test_fun_error, 
    list("x" = "done"),
    timeout = 500)

  testthat::expect_equal(is.null(results_error$value), TRUE)
  testthat::expect_equal(results_error$timeout, TRUE)
})


testthat::test_that("Processes that are called with handles function correctly", {
  results_normal <- familiar:::do.call_with_handlers(
    test_fun_normal, 
    list("x" = "done"))

  testthat::expect_equal(results_normal$value, "done")
  testthat::expect_equal(is.null(results_normal$warning), TRUE)
  testthat::expect_equal(is.null(results_normal$error), TRUE)

  results_warning <- familiar:::do.call_with_handlers(
    test_fun_warning, 
    list("x" = "done"))

  testthat::expect_equal(results_warning$value, "done")
  testthat::expect_equal(results_warning$warning, "warning")
  testthat::expect_equal(is.null(results_warning$error), TRUE)

  results_error <- familiar:::do.call_with_handlers(
    test_fun_error, 
    list("x" = "done"))

  testthat::expect_equal(is.null(results_error$value), TRUE)
  testthat::expect_equal(results_error$warning, "warning")
  testthat::expect_equal(results_error$error, "error")
})

testthat::test_that("Processes that are called with handles and timeout function correctly", {
  results_normal <- familiar:::do.call_with_handlers_timeout(
    test_fun_normal, 
    list("x" = "done"),
    timeout = 10000)

  testthat::expect_equal(results_normal$value, "done")
  testthat::expect_equal(is.null(results_normal$warning), TRUE)
  testthat::expect_equal(is.null(results_normal$error), TRUE)
  testthat::expect_equal(results_normal$timeout, FALSE)

  results_warning <- familiar:::do.call_with_handlers_timeout(
    test_fun_warning, 
    list("x" = "done"), 
    timeout = 10000)

  testthat::expect_equal(results_warning$value, "done")
  testthat::expect_equal(results_warning$warning, "warning")
  testthat::expect_equal(is.null(results_warning$error), TRUE)
  testthat::expect_equal(results_warning$timeout, FALSE)

  results_error <- familiar:::do.call_with_handlers_timeout(
    test_fun_error, 
    list("x" = "done"), 
    timeout = 10000)

  testthat::expect_equal(is.null(results_error$value), TRUE)
  testthat::expect_equal(results_error$warning, "warning")
  testthat::expect_equal(results_error$error, "error")
  testthat::expect_equal(results_error$timeout, FALSE)
})

testthat::test_that("Processes that are called with handles and infinite timeout function correctly", {
  results_normal <- familiar:::do.call_with_handlers_timeout(
    test_fun_normal,
    list("x" = "done"),
    timeout = Inf)

  testthat::expect_equal(results_normal$value, "done")
  testthat::expect_equal(is.null(results_normal$warning), TRUE)
  testthat::expect_equal(is.null(results_normal$error), TRUE)
  testthat::expect_equal(results_normal$timeout, FALSE)

  results_warning <- familiar:::do.call_with_handlers_timeout(
    test_fun_warning, 
    list("x" = "done"),
    timeout = Inf)

  testthat::expect_equal(results_warning$value, "done")
  testthat::expect_equal(results_warning$warning, "warning")
  testthat::expect_equal(is.null(results_warning$error), TRUE)
  testthat::expect_equal(results_warning$timeout, FALSE)

  results_error <- familiar:::do.call_with_handlers_timeout(
    test_fun_error, 
    list("x" = "done"), 
    timeout = Inf)

  testthat::expect_equal(is.null(results_error$value), TRUE)
  testthat::expect_equal(results_error$warning, "warning")
  testthat::expect_equal(results_error$error, "error")
  testthat::expect_equal(results_error$timeout, FALSE)
})



testthat::test_that("Processes that complete outside timeout fail", {
  results_normal <- familiar:::do.call_with_handlers_timeout(
    test_fun_normal, 
    list("x" = "done"),
    timeout = 500)

  testthat::expect_equal(is.null(results_normal$value), TRUE)
  testthat::expect_equal(is.null(results_normal$warning), TRUE)
  testthat::expect_equal(is.null(results_normal$error), TRUE)
  testthat::expect_equal(results_normal$timeout, TRUE)

  results_warning <- familiar:::do.call_with_handlers_timeout(
    test_fun_warning, 
    list("x" = "done"), 
    timeout = 500)

  testthat::expect_equal(is.null(results_warning$value), TRUE)
  testthat::expect_equal(is.null(results_warning$warning), TRUE)
  testthat::expect_equal(is.null(results_warning$error), TRUE)
  testthat::expect_equal(results_warning$timeout, TRUE)

  results_error <- familiar:::do.call_with_handlers_timeout(
    test_fun_error, 
    list("x" = "done"), 
    timeout = 500)

  testthat::expect_equal(is.null(results_error$value), TRUE)
  testthat::expect_equal(is.null(results_error$warning), TRUE)
  testthat::expect_equal(is.null(results_error$error), TRUE)
  testthat::expect_equal(results_error$timeout, TRUE)
})

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

familiar:::test_export(
  export_function = familiar:::export_calibration_info,
  data_element = "calibration_info",
  not_available_no_samples = FALSE,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("survival"),
  debug = debug_flag
)

familiar:::test_export(
  export_function = familiar:::export_calibration_info,
  data_element = "calibration_info",
  not_available_no_samples = FALSE,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("survival"),
  export_args = list("aggregate_results" = FALSE),
  debug = debug_flag
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test.
familiar:::test_plots(
  plot_function = familiar:::plot_decision_curve,
  outcome_type_available = c("binomial", "multinomial", "survival"),
  data_element = "decision_curve_analyis",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  not_available_extreme_probability = TRUE,
  debug = debug_flag
)

# Test with step-wise confidence interval
familiar:::test_plots(
  plot_function = familiar:::plot_decision_curve,
  outcome_type_available = c("binomial", "multinomial", "survival"),
  data_element = "decision_curve_analyis",
  test_config = "normal",
  plot_args = list("conf_int_style" = "step"),
  debug = debug_flag
)

# Test without confidence interval shown
familiar:::test_plots(
  plot_function = familiar:::plot_decision_curve,
  outcome_type_available = c("binomial", "multinomial", "survival"),
  data_element = "decision_curve_analyis",
  test_config = "normal",
  plot_args = list("conf_int_style" = "none"),
  debug = debug_flag
)

# Test without confidence interval
familiar:::test_plots(
  plot_function = familiar:::plot_decision_curve,
  outcome_type_available = c("binomial", "multinomial", "survival"),
  data_element = "decision_curve_analyis",
  estimation_type = "point",
  test_config = "normal",
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_decision_curve,
  data_element = "decision_curve_analyis",
  estimation_type = "point",
  outcome_type_available = c("binomial", "multinomial", "survival"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data. Survival requires
# survival probability and cannot be generated.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_decision_curve,
  data_element = "decision_curve_analyis",
  estimation_type = "point",
  outcome_type_available = c("binomial", "multinomial", "survival"),
  test_config = "normal",
  prediction_type = list("survival" = "survival_probability"),
  debug = debug_flag
)


# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_decision_curve,
  data_element = "decision_curve_analyis",
  outcome_type_available = c("binomial"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method"),
    "color_by" = "data_set"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_decision_curve,
  data_element = "decision_curve_analyis",
  outcome_type_available = c("multinomial"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method"),
    "color_by" = c("data_set", "positive_class")),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_decision_curve,
  data_element = "decision_curve_analyis",
  outcome_type_available = c("survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method"),
    "color_by" = c("data_set")),
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_shap_waterfall (generic) ------------------------------------------------

#' @title Create SHAP waterfall plot
#'
#' @description This method creates plots that show a waterfall of SHAP values
#'   obtained from the data stored in a familiarCollection object.
#'
#' @param dir_path (*optional*) Path to the directory where created 
#'   plots are saved to. Output is saved in the `explanation` subdirectory. If
#'   `NULL` no figures are saved, but are returned instead.
#' @param gradient_palette (*optional*) Divergent palette used to
#'   colour the elements of waterfall plots. `familiar` has a default
#'   palette. Other palettes are supported by the `paletteer` package,
#'   `grDevices::palette.pals()` (requires R >= 4.0.0), `grDevices::hcl.pals()`
#'   (requires R >= 3.6.0) and `rainbow`, `heat.colors`, `terrain.colors`,
#'   `topo.colors` and `cm.colors`, which correspond to the palettes of the same
#'   name in `grDevices`. You may also specify your own palette by providing a
#'   vector of colour names listed by `grDevices::colors()` or through
#'   hexadecimal RGB strings.
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams extract_performance -object
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#'
#' @details This function creates SHAP waterfall plots, which show the
#'   individual marginal contributions of feature values to the predicted value.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, `data_set`,
#'   `evaluation_time` (survival outcome only) and `positive_class` (categorical
#'   outcomes), `sample_id`. The default for is to facet by `evaluation_time` or 
#'   `positive_class`, and split by `vimp_method`,
#'   `learner`, `data_set`, and `sample_id`. `color_by` is not used.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_learner_names`
#'   can be applied to the `familiarCollection` object to update labels, and
#'   order the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_shap_waterfall
#' @md
#' @rdname plot_shap_waterfall-methods
setGeneric(
  "plot_shap_waterfall",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_shap_waterfall")
  }
)



# plot_shap_waterfall (general) ------------------------------------------------

#' @rdname plot_shap_waterfall-methods
setMethod(
  "plot_shap_waterfall",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "shap"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_shap_waterfall,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "x_axis_by" = x_axis_by,
        "y_axis_by" = y_axis_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "gradient_palette" = gradient_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "limit_n_features" = limit_n_features,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_shap_waterfall (collection) -----------------------------------------------

#' @rdname plot_shap_waterfall-methods
setMethod(
  "plot_shap_waterfall",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Check input arguments ----------------------------------------------------
    
    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)
    
    # Load the data.
    x <- export_shap(object = object)
    
    x <- x$shap_force
    if (is_empty(x)) return(NULL)
    
    # Obtain single data element from list.
    if (is.list(x)) {
      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_shap_waterfall: list of data elements contains unmerged elements."
        )
      }
      x <- x[[1L]]
    }
    
    # Check that the data are not evaluated at the model level.
    if (x@detail_level == "model") {
      ..warning_no_comparison_between_models()
      return(NULL)
    }
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    # Ensure that we work with a copy of the data.
    x@data <- data.table::copy(x@data)
    
    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to create a SHAP waterfall plot",
      message_type = "warning"
    )) {
      return(NULL)
    }
    
    # Add evaluation time or class as a splitting variable.
    additional_variable <- NULL
    if (object@outcome_type %in% c("survival")) {
      additional_variable <- "evaluation_time"
      data.table::setnames(x@data, old = "shap_outcome", new = "evaluation_time")
      
    } else if (object@outcome_type %in% c("multinomial")) {
      additional_variable <- "positive_class"
      data.table::setnames(x@data, old = "shap_outcome", new = "positive_class")
    }
    
    # Add default splitting variables.
    if (
      is.null(split_by) &&
      is.null(facet_by)
    ) {
      # Split by vimp_method, learner and sample id.
      split_by <- c("vimp_method", "learner", "data_set", "sample_id")
      facet_by <- additional_variable
    }
    
    all_variables <- c("vimp_method", "learner", "data_set", "sample_id", additional_variable)
    
    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      facet_by = facet_by,
      available = all_variables
    )
    
    # Update splitting variables
    split_by <- split_var_list$split_by
    facet_by <- split_var_list$facet_by
    
    # x_label
    if (is.waive(x_label)) {
      x_label <- "predicted value"
    }
    
    # y_label
    if (is.waive(y_label)) {
      y_label <- "feature"
    }
    
    .check_input_plot_args(
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      limit_n_features = limit_n_features
    )
    
    # Create plots -------------------------------------------------------------
    
    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)
    
    # Split data.
    if (!is.null(split_by)) {
      x_split <- split(
        x@data, 
        by = split_by, 
        drop = FALSE
      )
      
    } else {
      x_split <- list("null.name" = x@data)
    }
    
    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()
    
    # Iterate over data splits.
    for (ii in names(x_split)) {
      # Skip empty datasets.
      if (is_empty(x_split[[ii]])) next
      
      if (is.waive(plot_title)) plot_title <- "SHAP waterfall"
      
      # Declare subtitle components.
      additional_subtitle <- NULL
      
      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(x_split[[ii]]$evaluation_time[1L])
        )
      }
      
      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = x_split[[ii]]
        )
      }
      
      # Generate plot
      p <- .plot_shap_waterfall_plot(
        x = x_split[[ii]],
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        gradient_palette = gradient_palette,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        limit_n_features = limit_n_features,
        x_range = x_range,
        x_n_breaks = x_n_breaks,
        x_breaks = x_breaks
      )
      
      # Check empty output
      if (is.null(p)) next
      
      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)
      
      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_shap_waterfall_plot_dimensions(
          x = x_split[[ii]],
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols
        )
        
        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "explanation",
              "subtype" = "shap_waterfall",
              "x" = x_split[[ii]],
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }
    
    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_shap_waterfall_plot <- function(
    x,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    gradient_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    limit_n_features,
    x_range,
    x_n_breaks,
    x_breaks
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  shap_value <- vimp <- feature_value <- feature_name <- NULL
  feature_label <- prediction <- y <- label_text <- phi_0 <- NULL
  
  # Sort features by importance (mean absolute SHAP).
  feature_importance <- x[, list("vimp" = mean(abs(shap_value))), by = c(facet_by, "feature_name")]
  feature_importance <- feature_importance[, list("vimp" = max(vimp)), by = "feature_name"][order(vimp)]
  
  # Determine the features that need to be plotted.
  if (is.numeric(limit_n_features)) {

    # Explicitly select features based on threshold value instead of simply 
    # selecting the best features: features may have the same value because they
    # belong to the same cluster.
    threshold_value <- min(tail(unique(sort(feature_importance$vimp)), n = limit_n_features))
    selected_features <- feature_importance[vimp >= threshold_value]$feature_name
    
    # Determine contribution of features that were not selected (if any)
    not_selected_features <- setdiff(feature_importance$feature_name, selected_features)
    
    if (!is_empty(not_selected_features)) {
      # Use a template to prevent having to manually fill out columns that are
      # not updated.
      x_template <- data.table::copy(x[feature_name == not_selected_features[1L]])
      x_template[, "feature_name" := "other"]
      x_template[, "feature_value" := NA_real_]
      x_template[, "feature_label" := NA_character_]
      x_template[, "shap_value" := NULL]
      
      other_shap_value <- x[
        feature_name %in% not_selected_features,
        list(
          "feature_name" = "other",
          "shap_value" = sum(shap_value)
        ),
        by = facet_by
      ]
      
      x_template <- merge(
        x = x_template,
        y = other_shap_value,
        by = c("feature_name", facet_by)
      )
      
      x <- x[feature_name %in% selected_features, ]
      x <- data.table::rbindlist(list(x, x_template), use.names = TRUE)
      
      feature_importance <- feature_importance[vimp >= threshold_value][order(vimp)]

      x$feature_name <- factor(
        x = x$feature_name,
        levels = c("other", as.character(feature_importance$feature_name))
      )
      
    } else {
      # All features were selected: use default procedure.
      x$feature_name <- factor(
        x = x$feature_name,
        levels = feature_importance$feature_name
      )
    }
    
  } else {
    # Default procedure without selection.
    x$feature_name <- factor(
      x = x$feature_name,
      levels = feature_importance$feature_name
    )
  }
  
  # Add y (for positioning).
  x[, "y" := as.numeric(feature_name)]
  
  y_label_table <- unique(x[, mget(c("feature_name", "feature_value", "feature_label", "y"))])
  y_label_table[, "feature_name_value" := ..set_shap_waterfall_feature_name_value(feature_name, feature_value, feature_label)]
  y_label_table <- y_label_table[order(y)]

  # Common base for formatting prediction and shap values.
  common_base <- ..format_get_common_base(c(x$shap_value, x$prediction))
  n_small <- max(c(-(common_base - 2L), 0.0))

  # Update start and end positions for force elements.
  if (!is.null(facet_by)) {
    x[, (c("x_start", "x_end")) := ..set_shap_waterfall_positions(shap_value, prediction, feature_name), by = facet_by]
    
  } else {
    x[, (c("x_start", "x_end")) := ..set_shap_waterfall_positions(shap_value, prediction, feature_name)]
  }
  
  # Derive information for the average prediction and extract the instance
  # prediction.
  if (!is.null(facet_by)) {
    f_average_data <- x[, list("prediction" = max(phi_0)), by = facet_by]
  } else {
    f_average_data <- x[, list("prediction" = max(phi_0))]
  }
  
  f_instance_data <- unique(x[, mget(c("prediction", facet_by))])
  
  # Check x-range.
  if (!is.null(x_range)) {
    .check_input_plot_args(x_range = x_range)
    
    # x_breaks
    if (is.null(x_breaks)) {
      .check_input_plot_args(
        x_range = x_range,
        x_n_breaks = x_n_breaks
      )
      
      # Create breaks and update x_range
      x_breaks <- labeling::extended(
        m = x_n_breaks,
        dmin = x_range[1L],
        dmax = x_range[2L],
        only.loose = TRUE
      )
      
      x_range <- c(
        head(x_breaks, n = 1L),
        tail(x_breaks, n = 1L)
      )
      
    } else {
      .check_input_plot_args(x_breaks = x_breaks)
    }
  }

  # Create a legend label.
  legend_label <- .create_plot_legend_title(
    user_label = legend_label,
    color_by = "shap_value"
  )
  
  # Check remaining input arguments.
  .check_input_plot_args(
    legend_label = legend_label
  )
  
  # Add gradient palette.
  gradient_colours <- .get_palette(
    x = gradient_palette, 
    palette_type = "divergent"
  )
  
  # Set up shap value labels
  x[, "label_align" := data.table::fifelse(shap_value >= 0.0, yes = "left", no = "right")]
  x[, "label_colour" := data.table::fifelse(shap_value >= 0.0, yes = tail(gradient_colours, n = 1L), no = head(gradient_colours, n = 1L))]
  x[, "label_text" := format(round(shap_value, digits = n_small), nsmall = n_small)]
  x[, "label_text" := paste0(" ", label_text, " ")]
  
  # Set up basic waterfall plot.
  p <- ggplot2::ggplot(data = x)
  p <- p + ggtheme
  p <- p + geom_waterfall_shap(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("x_start"),
      xend = !!sym("x_end"),
      y = !!sym("y"),
      ymin = !!sym("y") - 0.4,
      ymax = !!sym("y") + 0.4,
      fill = !!sym("shap_value")
    )
  )
  
  # Set labels for y-axis.
  p <- p + ggplot2::scale_y_continuous(
    breaks = y_label_table$y,
    labels = y_label_table$feature_name_value
  )
  
  # Add spacing for text values.
  p <- p + ggplot2::scale_x_continuous(
    expand = ggplot2::expansion(mult = 0.2)
  )
  
  p <- p + ggplot2::scale_fill_gradientn(
    name = legend_label,
    colors = gradient_colours,
    limits = c(-max(abs(x$shap_value)), max(abs(x$shap_value)))
  )
  
  # Add instance and group average prediction.
  p <- p + ggplot2::geom_vline(
    data = f_average_data,
    mapping = ggplot2::aes(xintercept = !!sym("prediction")),
    color = "grey80",
    linetype = 2L
  )
  p <- p + ggplot2::geom_vline(
    data = f_instance_data,
    mapping = ggplot2::aes(xintercept = !!sym("prediction")),
    color = "grey80",
    linetype = 2L
  )
  
  # Use geom-segment to connect segments.
  p <- p + ggplot2::geom_segment(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("x_start"),
      xend = !!sym("x_start"),
      y = !!sym("y") - 0.5,
      yend = !!sym("y") + 0.45,
    ),
    color = "grey60"
  )
  p <- p + ggplot2::geom_segment(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("x_end"),
      xend = !!sym("x_end"),
      y = !!sym("y") - 0.45,
      yend = !!sym("y") + 0.5,
    ),
    color = "grey60"
  )
  
  text_settings <- .get_plot_geom_text_settings(ggtheme = ggtheme)
  
  # Add shap label.
  for (x_text in split(x, by = "label_colour", drop = TRUE)) {
    p <- p + ggplot2::geom_text(
      data = x_text,
      mapping = ggplot2::aes(
        x = !!sym("x_end"),
        y = !!sym("y"),
        label = !!sym("label_text"),
        hjust = !!sym("label_align")
      ),
      colour = head(x_text$label_colour, n = 1L),
      family = text_settings$family,
      fontface = text_settings$face,
      size = text_settings$geom_text_size,
      show.legend = FALSE
    )
  }
  
  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )
  
  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows, 
        cols = facet_by_list$facet_cols,
        scales = ifelse(is.null(x_range), "free_x", "fixed"),
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by, 
        scales = ifelse(is.null(x_range), "free_x", "fixed"),
        labeller = "label_context"
      )
    }
  }
  
  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label, 
    y = y_label, 
    title = plot_title, 
    subtitle = plot_sub_title, 
    caption = caption
  )
  
  # Prevent clipping of confidence intervals.
  if (!is.null(x_range)) p <- p + ggplot2::coord_cartesian(xlim = x_range)
  
  return(p)
}



..set_shap_waterfall_positions <- function(x, predictions, feature_name) {
  # Prevent notes.
  feature_value <- density <- y_offset <- NULL
  
  # Initialise.
  x_start <- x_end <- y <- y_seg_start <- y_seg_end <- numeric(length(x))
  
  # Set feature order.
  feature_order <- order(feature_name, decreasing = TRUE)
  
  # We fill start and end positions in reverse order, beginning with the most
  # important feature.
  previous_start <- utils::head(predictions, n = 1L)
  for (ii in feature_order) {
    x_end[ii] <- previous_start
    x_start[ii] <- previous_start - x[ii]
    previous_start <- x_start[ii]
  }

  return(list(
    "x_start" = x_start,
    "x_end" = x_end
  ))
}



..set_shap_waterfall_feature_name_value <- function(
    feature_name,
    feature_value, 
    feature_label
) {
  actual_label <- feature_label
  actual_label[is.na(feature_label)] <- signif(feature_value[is.na(feature_label)], 3L)
  actual_label[!is.na(actual_label)] <- paste0(": ", actual_label[!is.na(actual_label)])
  actual_label[is.na(actual_label)] <- ""
  
  return(paste0(feature_name, actual_label))
}



.determine_shap_waterfall_plot_dimensions <- function(
    x,
    x_axis_by,
    y_axis_by,
    facet_by,
    facet_wrap_cols
) {
  
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- 4.0
  
  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))
  
  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))
  
  return(c(height, width))
}


# GeomSHAPWaterfall ------------------------------------------------------------
# Placeholder to prevent NOTES if ggplot2 is not installed.
GeomSHAPWaterfall <- NULL
if (rlang::is_installed("ggplot2")) {
  GeomSHAPWaterfall <- ggplot2::ggproto(
    "GeomPolygon",
    ggplot2::Geom,
    required_aes = c("x", "xend", "y", "ymin", "ymax"),
    default_aes = ggplot2::aes(
      colour = NA,
      fill = "grey35",
      linewidth = 0.5,
      linetype = 1,
      alpha = NA
    ),
    draw_key = ggplot2::draw_key_polygon,
    draw_panel = function(
    data,
    panel_params,
    coord,
    lineend = "butt", linejoin = "round", linemitre = 10
    ) {
      # Compute coordinates based on data.
      coords <- coord$transform(data, panel_params)
      
      # Instantiate parameters to feed to grid::polygonGrob. These vectors are 
      # sufficiently large to hold all polygons with taper.
      x <- y <- numeric(nrow(coords) * 5L)
      id <- integer(nrow(coords) * 5L)
      
      # Iterate over features.
      idx_offset <- 0L
      for (ii in seq_len(nrow(coords))) {
        # Set up coordinates.
        x_start = coords$x[ii]
        x_end = coords$xend[ii]
        y_down <- coords$ymin[ii]
        y_up <- coords$ymax[ii]
        
        if (abs(x_start - x_end) > 0.05) {
          # Add taper if there is sufficient room on the plot.
          x_mid <- ifelse(x_start < x_end, x_end - 0.05, x_end + 0.05)
          y_mid <- (y_up + y_down) * 0.5
          
          # Define coordinates for polygon.
          idx <- idx_offset + (1L:5L)
          x[idx] <- c(x_start, x_start, x_mid, x_end, x_mid)
          y[idx] <- c(y_down, y_up, y_up, y_mid, y_down)
          
        } else {
          # Avoid taper if there is not sufficient room.
          idx <- idx_offset + (1L:4L)
          x[idx] <- c(x_start, x_start, x_end, x_end)
          y[idx] <- c(y_down, y_up, y_up, y_down)
        }
        
        # Set grouping.
        id[idx] <- ii
        
        # Update offset.
        idx_offset <- idx_offset + length(idx)
      }
      
      # Select elements which are correctly set.
      valid_idx <- id > 0L
      
      return(grid::polygonGrob(
        x[valid_idx],
        y[valid_idx],
        id = id[valid_idx],
        default.units = "native",
        gp = grid::gpar(
          col = coords$colour,
          fill = ggplot2::fill_alpha(coords$fill, coords$alpha),
          lwd = coords$linewidth * ggplot2::.pt,
          lty = coords$linetype,
          lineend = lineend,
          linejoin = linejoin,
          linemitre = linemitre
        )
      ))
    }
  )
}



geom_waterfall_shap <- function(
    mapping = NULL,
    data = NULL,
    stat = "identity",
    position = "identity",
    na.rm = FALSE,
    show.legend = NA,
    inherit.aes = TRUE,
    ...
) {
  ggplot2::layer(
    geom = GeomSHAPWaterfall,
    mapping = mapping,
    data = data, 
    stat = stat, 
    position = position, 
    show.legend = show.legend, 
    inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, ...)
  )
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarDataElementSHAP object -----------------------------------------------
setClass(
  "familiarDataElementSHAP",
  contains = "familiarDataElement",
  slots = list(
    "sample_identifiers" = "ANY",
    "data_mapping" = "ANY",
    "predicted_values" = "ANY",
    "lookup_table" = "ANY"
  ),
  prototype = methods::prototype(
    detail_level = "ensemble",
    estimation_type = "point",
    value_column = "shap_value",
    grouping_column = c("feature_name", "feature_value_mapping", "sample_id", "phi_0"),
    sample_identifiers = NULL,
    data_mapping = NULL,
    predicted_values = NULL,
    lookup_table = NULL
  )
)


# familiarDataElementSHAPSummary object ----------------------------------------

# Objects for creating SHAP summary plots. These are created at run-time from
# data included in familiarDataElementSHAP objects.
setClass(
  "familiarDataElementSHAPSummary",
  contains = "familiarDataElement"
)


# familiarDataElementSHAPForce object ------------------------------------------

# Objects for creating SHAP force plots. These are created at run-time from
# data included in familiarDataElementSHAP objects.
setClass(
  "familiarDataElementSHAPForce",
  contains = "familiarDataElement"
)


# familiarDataElementSHAPDependence object -------------------------------------

# Objects for creating SHAP dependence plots. These are created at run-time from
# data included in familiarDataElementSHAP objects.
setClass(
  "familiarDataElementSHAPDependence",
  contains = "familiarDataElement"
)




# extract_shap (generic) -------------------------------------------------------

#'@title Internal function for computing SHAP values.
#'
#'@description Computes SHAP values for feature values using a
#'  `familiarEnsemble`.
#'
#'@param features Features for whose values SHAP values need to be computed.
#'  defaults to all features in the model.
#'@param n_sample_points Minimum number of values to sample for numeric
#'  features. By default, this is based on input dataset. But if the number of
#'  values of a feature within that dataset is too low, additional values are
#'  drawn from the feature distribution (stored with the model).
#'@param shap_tolerance Relative tolerance for convergence of SHAP values. The
#'  tolerance is scaled with the range in SHAP values. Default: 0.05.
#'@param shap_max_iterations Maximum iterations for convergence of SHAP values.
#'  Default: 1000
#'@param shap_phi_0 Reference predicted value(s). Determined from data by
#'  default.
#'
#'@inheritParams .extract_data
#'
#'@return A list of familiarDataElements with SHAP values.
#'@md
#'@keywords internal
setGeneric(
  "extract_shap",
  function(
    object,
    data,
    cl = NULL,
    features = NULL,
    n_sample_points = 20L,
    shap_tolerance = waiver(),
    shap_max_iterations = waiver(),
    shap_phi_0 = waiver(),
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    sample_limit = waiver(),
    detail_level = waiver(),
    aggregate_results = waiver(),
    n_important_features = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_shap")
  }
)



# extract_shap (familiarEnsemble) ----------------------------------------------
setMethod(
  "extract_shap",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    features = NULL,
    n_sample_points = 20L,
    shap_tolerance = waiver(),
    shap_max_iterations = waiver(),
    shap_phi_0 = NULL,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    sample_limit = waiver(),
    detail_level = waiver(),
    aggregate_results = waiver(),
    n_important_features = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    # Compute SHAP values.
    
    # Message extraction start
    logger_message(
      paste0("Extracting SHAP values for the ensemble."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (is.null(features)) {
      logger_message(
        paste0(
          "Computing SHAP values for features in the dataset."
        ),
        indent = message_indent + 1L,
        verbose = verbose
      )
      
    } else {
      logger_message(
        paste0(
          "Computing SHAP values for selected features: ", paste_s(features), "."
        ),
        indent = message_indent + 1L,
        verbose = verbose
      )
    }
    
    # Load evaluation_times from the object settings attribute, if it is not provided.
    if (is.waive(evaluation_times)) evaluation_times <- object@settings$eval_times
    
    # Check evaluation_times argument
    if (object@outcome_type %in% c("survival")) {
      sapply(
        evaluation_times,
        .check_number_in_valid_range,
        var_name = "evaluation_times",
        range = c(0.0, Inf),
        closed = c(FALSE, TRUE)
      )
    }
    
    # Check n_sample_points argument. This defines the maximum number of feature
    # values for which SHAP values are computed.
    .check_number_in_valid_range(
      x = n_sample_points,
      var_name = "n_sample_points",
      range = c(2L, Inf),
      closed = c(TRUE, TRUE)
    )
    
    # Check shap_tolerance argument. This sets stopping criteria for convergence
    # of SHAP values.
    if (is.waive(shap_tolerance)) shap_tolerance <- object@settings$shap_tolerance
    
    .check_number_in_valid_range(
      x = shap_tolerance,
      var_name = "shap_tolerance",
      range = c(0.0, Inf),
      closed = c(FALSE, FALSE)
    )
    
    # Check shap_max_iterations argument. This sets the maximum number of
    # iterations for computing SHAP values.
    if (is.waive(shap_max_iterations)) shap_max_iterations <- object@settings$shap_max_iterations
    
    .check_number_in_valid_range(
      x = shap_max_iterations,
      var_name = "shap_max_iterations",
      range = c(1L, Inf),
      closed = c(TRUE, TRUE)
    )
    
    # Check shap_phi_0.
    if (is.waive(shap_phi_0)) shap_phi_0 <- object@settings$shap_phi_0
    
    if (!is.null(shap_phi_0)) {
      if (object@outcome_type == "survival") {
        .check_argument_length(
          x = shap_phi_0,
          var_name = "shap_phi_0",
          min = length(evaluation_times),
          max = length(evaluation_times)
        )
        
        .check_is_numeric(
          x = shap_phi_0,
          var_name = "shap_phi_0"
        )
        
      } else if (object@outcome_type == "multinomial") {
        .check_argument_length(
          x = shap_phi_0,
          var_name = "shap_phi_0",
          min = length(get_outcome_class_levels(object)),
          max = length(get_outcome_class_levels(object))
        )
        
        .check_is_numeric(
          x = shap_phi_0,
          var_name = "shap_phi_0"
        )
        
      } else {
        .check_is_numeric(
          x = shap_phi_0,
          var_name = "shap_phi_0"
        )
      }
    }
    
    # Obtain ensemble method from stored settings, if required.
    if (is.waive(ensemble_method)) ensemble_method <- object@settings$ensemble_method
    
    # Check ensemble_method argument
    .check_parameter_value_is_valid(
      x = ensemble_method,
      var_name = "ensemble_method",
      values = .get_available_ensemble_prediction_methods()
    )
    
    # Check the sample limit. This defines the subset of samples that are being
    # assessed (if real data is used instead of a minimum subset).
    sample_limit <- .parse_sample_limit(
      x = sample_limit,
      object = object,
      default = 200L,
      data_element = "shap"
    )
    
    # Check the level detail.
    detail_level <- .parse_detail_level(
      x = detail_level,
      object = object,
      default = "ensemble",
      data_element = "shap"
    )
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = FALSE,
      data_element = "shap"
    )
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Test if any model in the ensemble was successfully trained.
    if (!model_is_trained(object = object)) return(NULL)
    
    # Check the number of important features.
    n_important_features <- .parse_n_important_features(
      x = n_important_features,
      object = object,
      default = 20,
      data_element = "shap"
    )
    
    # Set features to be assessed using SHAP.
    important_features <- .select_important_features(
      object = object,
      data = data,
      n_important_features = n_important_features
    )
    
    # Get and process the input data. Since we define SHAP values using the
    # features in their original scales, we need to apply only minimal
    # pre-processing.
    data <- process_input_data(
      object = object,
      data = data, 
      stop_at = "signature"
    )
    
    # Use sample limit to cap the number of samples that are assessed.
    data <- get_subsample(
      data = data,
      size = sample_limit,
      seed = 0L
    )
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "familiarDataElementSHAP",
      detail_level = detail_level
    )
    
    # Generate elements to send to dispatch.
    shap_data <- extract_dispatcher(
      FUN = .extract_shap,
      has_internal_bootstrap = FALSE,
      cl = cl,
      object = object,
      data = data,
      features = features,
      n_sample_points = n_sample_points,
      tolerance = shap_tolerance,
      n_max_iter = shap_max_iterations,
      phi_0 = shap_phi_0,
      proto_data_element = proto_data_element,
      important_features = important_features,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
  }
)



# extract_shap (prediction table) ----------------------------------------------
setMethod(
  "extract_shap",
  signature(object = "familiarDataElementSHAP"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("SHAP values")
    
    return(NULL)
  }
)



.extract_shap <- function(
    object,
    data = NULL,
    proto_data_element,
    important_features,
    evaluation_times = NULL,
    features = NULL,
    n_sample_points,
    aggregate_results,
    is_pre_processed = FALSE,
    ensemble_method,
    cl,
    tolerance = 0.005,
    n_max_iter = 1000L,
    phi_0 = NULL,
    mapping_method = "fixed",
    sampling_method = "importance",
    message_indent = 0L,
    verbose = FALSE,
    progress_bar = FALSE,
    ...
) {
  # Step 1: Determine feature values that are to be sampled for determining SHAP
  # values.
  #
  # Step 2: Determine the minimum sampleset X required to determine SHAP values.
  # The number of samples (n) is equal to the feature with the largest number of 
  # values (m_i) to sample. Feature values can be randomly ordered for each
  # feature. Features with m_i < n can randomly draw additional features.
  #
  # Alternative: Use the actual dataset X (trigger on function argument).
  #
  # Step 3: Create coalition sets for coalitions with all but one off and all but
  # one on.
  #
  # Step 4: Iterate samples in sampleset X. Generate samples corresponding to
  # coalitions for each sample. Concatenate generated samples and add to table
  # with previously generated samples X_gen.
  #
  # Step 5: Select samples without existing predictions.
  #
  # Step 6: Predict samples. Merge new predictions into existing predictions.
  #
  # Step 7: Compute average predicted value (phi_0).
  #
  # Step 8: For each sample in sampleset X, determine coalition represented by
  # each sample in X_gen. Compute kernel weight based on coalition. Compute SHAP
  # values by solving linear equation.
  #
  # Step 9: Average SHAP value for each feature value.
  #
  # Step 10: Determine convergence and repeat steps 4-9 until convergence is
  # reached, or capacity is exhausted.
  #
  # Parallel processing: perform steps 4-6 multiple times within a parallel loop.
  # This allows for faster convergence.
  
  shap_value <- NULL

  # Check that the model requires any features.
  if (is_empty(important_features)) return(NULL)
  
  # Get set of feature values.
  feature_set <- .get_shap_feature_set(
    data = data,
    features = object@model_features,
    feature_info = object@feature_info[object@model_features],
    n_sample_points = n_sample_points
  )
  
  # Generate data if absent.
  if (is_empty(data)) {
    data <- .get_shap_sample_set(
      object = object,
      feature_set = feature_set
    )
  }
  
  # Ensure that data is unique.
  data <- select_unique_data(data = data)
  
  # Get sample identifiers.
  sample_identifiers <- get_unique_row_names(data)
  
  # From here, work with mapping representations of the data (h).
  mapping_input <- .shap_data_to_mapping(
    data = data,
    feature_set = feature_set
  )
  
  # Predict outcome values from the input data. Output may be more than one
  # column.
  predicted_values_input <- .predict_from_coalition(
    mapping = mapping_input,
    feature_set = feature_set,
    object = object,
    ensemble_method = ensemble_method,
    evaluation_time = evaluation_times
  )
  
  if (is_empty(predicted_values_input)) return(NULL)
  
  # Compute phi_0.
  if (is.null(phi_0)) phi_0 <- colMeans(predicted_values_input)
  
  if (length(important_features) == 1L) {
    # Single feature (shapley) -------------------------------------------------
    shap_values <- .compute_shap_value_single_feature(
      important_features = important_features,
      mapping = mapping_input,
      sample_id = sample_identifiers,
      feature_set = feature_set,
      predicted_values = predicted_values_input,
      phi_0 = phi_0
    )

  } else {
    # Multiple features (kernel) -----------------------------------------------
    
    # Generate coalitions (Z)
    input_coalitions <- .get_shap_coalitions(
      important_features = important_features,
      depth = 1L
    )
    
    # Check that coalitions are not empty: this happens if the data contains a
    # single feature: SHAP values cannot be computed.
    if (is.null(input_coalitions)) return(NULL)
    
    # Provide the initial set of coalitions.
    coalitions <- list(input_coalitions)
    
    # Compute weights for each coalition in a coalition set.
    kernel_weights <- .compute_shap_kernel_weights(
      n = ncol(input_coalitions),
      individual_coalition = TRUE
    )
    
    # Looping variables.
    iter_id <- 0L
    all_shap_converged <- FALSE
    shap_values <- NULL
    shap_matrices <- NULL
    n_parallel <- max(c(length(cl), 1L))
    
    while (!all_shap_converged && iter_id < n_max_iter) {
      # Determine the iteration identifiers.
      current_ids <- iter_id + seq_len(n_parallel)
      
      # Distribute computation.
      shap_values_iter <- fam_lapply(
        cl = cl,
        X = current_ids,
        FUN = .compute_shap_iterative,
        object = object,
        ensemble_method = ensemble_method,
        evaluation_times = evaluation_times,
        mapping_input = mapping_input,
        sample_identifiers = sample_identifiers,
        sample_predictions = predicted_values_input,
        important_features = important_features,
        phi_0 = phi_0,
        coalitions = coalitions,
        kernel_weights = kernel_weights,
        feature_set = feature_set,
        mapping_method = mapping_method
      )
      
      # Combine with previous shap_values.
      shap_values <- data.table::rbindlist(c(list(shap_values), shap_values_iter))
      
      # Compute variance of each SHAP value.
      shap_variance <- shap_values[
        ,
        list("shap_var" = stats::var(shap_value), "n" = .N),
        by = c("sample_id", "feature_name", "feature_value_mapping", "shap_outcome")
      ]
      
      # Check convergence.
      all_shap_converged <- .evaluate_shap_convergence(
        shap_variance = shap_variance,
        tolerance = max(c(
          tolerance * diff(range(shap_values$shap_value)),
          tolerance * diff(range(c(predicted_values_input))) / sqrt(length(feature_set))
        ))
      )
      
      # Update coalitions.
      coalitions <- .sample_shap_coalitions(
        coalitions = input_coalitions,
        kernel_weights = kernel_weights,
        shap_variance = shap_variance,
        sampling_method = sampling_method,
        seed = 19L + iter_id
      )
      
      # Update iteration id.
      iter_id <- tail(current_ids, n = 1L)
    }
  }
  
  # Compute final SHAP values.
  shap_values <- shap_values[
    ,
    list("shap_value" = mean(shap_value)),
    by = c("sample_id", "feature_name", "feature_value_mapping", "shap_outcome")
  ]
  
  # Add model name to data element.
  proto_data_element <- add_model_name(
    proto_data_element,
    object = object
  )
  
  # Store data mapping of feature values for input data.
  proto_data_element@data_mapping <- mapping_input
  
  # Store lookup-table translate feature mapping back to feature values.
  proto_data_element@lookup_table <- feature_set

  # Add predictions for input data.
  proto_data_element@predicted_values <- predicted_values_input
  
  # Add sample identifiers.
  proto_data_element@sample_identifiers <- sample_identifiers
  
  # Store shap data. Value column is "shap_value", grouping columns are
  # "feature_name" and "feature_value_mapping". For multinomial and survival
  # outcomes, "shap_outcome" is an additional grouping column.
  if (object@outcome_type %in% c("multinomial", "survival")) {
    # Add phi_0
    if (object@outcome_type == "survival") {
      phi_0_data <- data.table::data.table(
        "shap_outcome" = as.character(evaluation_times),
        "phi_0" = phi_0
      )
        
    } else if (object@outcome_type == "multinomial") {
      phi_0_data <- data.table::data.table(
        "shap_outcome" = get_outcome_class_levels(object),
        "phi_0" = phi_0
      )
    }
    
    shap_values <- merge(
      x = shap_values,
      y = phi_0_data,
      by = "shap_outcome"
    )
    
    proto_data_element@data <- data.table::copy(
      shap_values[, mget(c("feature_name", "feature_value_mapping", "shap_outcome", "shap_value", "phi_0", "sample_id"))]
    )
    
    # Add shap_outcome as additional grouping level.
    proto_data_element@grouping_column <- c(proto_data_element@grouping_column, "shap_outcome")
    
    if (object@outcome_type %in% c("multinomial")) {
      # Convert shap_outcome to categorical values corresponding to the levels
      # in the modelled endpoint.
      proto_data_element@data$shap_outcome <- factor(
        proto_data_element@data$shap_outcome,
        levels = get_outcome_class_levels(object)
      )
      
    } else if (object@outcome_type == "survival") {
      # Convert shap_outcome to categorical values corresponding to the
      # evaluation time
      proto_data_element@data$shap_outcome <- factor(
        proto_data_element@data$shap_outcome,
        levels = as.character(evaluation_times)
      )
    }
    
  } else {
    # Add phi_0.
    shap_values[, "phi_0" := phi_0]
    
    proto_data_element@data <- data.table::copy(
      shap_values[, mget(c("feature_name", "feature_value_mapping", "shap_value", "phi_0", "sample_id"))]
    )
  }
  
  return(proto_data_element)
}



.compute_shap_iterative <- function(
    iter_id,
    object,
    ensemble_method,
    evaluation_times,
    mapping_input,
    important_features,
    sample_identifiers,
    sample_predictions,
    phi_0,
    coalitions,
    kernel_weights,
    feature_set,
    mapping_method
) {
  # Determine additional mapping.
  mapping_iter <- .shap_randomise_mapping_from_coalition(
    important_features = important_features,
    samples = mapping_input,
    coalitions = coalitions,
    feature_set = feature_set,
    seed = iter_id,
    mapping_method = mapping_method
  )
  
  # Predict from new unique mappings.
  predicted_values_iter <- .predict_from_coalition(
    mapping = mapping_iter,
    feature_set = feature_set,
    object = object,
    ensemble_method = ensemble_method,
    evaluation_time = evaluation_times
  )
  
  if (is.null(predicted_values_iter)) return(NULL)
  
  # Compute and update A and b matrices.
  shap_matrices <- .compute_shap_matrices(
    important_features = important_features,
    samples = mapping_input,
    sample_predictions = sample_predictions,
    sample_id = sample_identifiers,
    mapping = mapping_iter,
    predicted_values = predicted_values_iter,
    phi_0 = phi_0,
    kernel_weights = kernel_weights
  )
  
  # Compute SHAP values for this iteration.
  shap_values <- .compute_shap_value(
    shap_matrices = shap_matrices
  )
  
  if (is.null(shap_values)) return(NULL)
  
  return(shap_values)
}



.get_shap_feature_set <- function(
    data = NULL,
    features,
    feature_info,
    n_sample_points
) {
  # Gets set of feature values for the features of interest. For categorical
  # features, all levels are used. For numerical features, the data is sampled
  # (if available), and additional values are drawn based on the known
  # distribution of feature values for each feature.
  feature_set <- list()
  for (feature in features) {
    # For categorical features, use all levels.
    if (feature_info[[feature]]@feature_type == "factor") {
      feature_set[[feature]] <- factor(feature_info[[feature]]@levels, levels = feature_info[[feature]]@levels)
      next
    }
    
    # Select (numeric) feature values from the data.
    feature_values <- NULL
    if (!is_empty(data)) {
      feature_values <- unique_na(data@data[[feature]])
    }
    
    # Check number of values to sample.
    n_to_sample <- n_sample_points - length(feature_values)
    if (n_to_sample <= 0L) {
      feature_set[[feature]] <- feature_values
      next
    }
    
    # Add feature values by sampling distribution.
    feature_set[[feature]] <- unique(c(
      feature_values,
      stats::spline(
        x = (seq_along(feature_info[[feature]]@distribution$pctl) - 1L) / 
          (length(feature_info[[feature]]@distribution$pctl) - 1L),
        y = as.numeric(feature_info[[feature]]@distribution$pctl),
        xout = get_percentiles(n_to_sample),
        method = "hyman"
      )$y
    ))
  }
  
  return(feature_set)
}



.get_shap_sample_set <- function(
    object,
    feature_set
) {
  # Determine the number of samples that are required.
  n_samples <- max(lengths(feature_set))
  
  # Fill sample set. Feature values are randomly ordered and then distributed
  # over features.
  sample_set <- list()
  for (ii in seq_along(feature_set)) {
    feature <- names(feature_set)[ii]
    feature_values <- feature_set[[feature]]
    feature_values <- feature_values[fam_sample(
      seq_along(feature_values),
      n = length(feature_values),
      replace = FALSE,
      seed = ii
    )]
    sample_set[[feature]] <- rep_len(feature_values, length.out = n_samples)
  }
  
  # Convert to data.table and batch and sample identifiers.
  data <- data.table::as.data.table(sample_set)
  data[, ":="(
    "batch_id" = "generated",
    "sample_id" = seq_len(nrow(data))
  )]
  
  return(as_data_object(
    data = data,
    object = object,
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    check_stringency = "external"
  ))
}



.get_shap_coalitions <- function(
    important_features,
    depth = 1L
) {
  # Get initial set of coalitions. Antithetic coalitions are created within
  # ...shap_randomise_mapping_from_coalition
  
  # Helper function that inserts TRUE at the indices indicated by ones.
  ..fun <- function(ones, n_features) {
    x <- logical(n_features)
    x[ones] <- TRUE
    return(x)
  }

  # Set number of features.
  n_features <- length(important_features)
  
  # Check that depth is at most n_features - 1L.
  if (depth >= n_features) depth <- n_features - 1L
  if (depth < 1L) return(NULL)
  
  z <- list()
  for (ii in seq_len(depth)) {
    z <- c(
      z,
      utils::combn(
        n_features,
        m = ii,
        FUN = ..fun,
        simplify = FALSE,
        n_features = n_features
      )
    )
  }
  
  # To matrix. Data are stored row-wise.
  z <- matrix(unlist(z), ncol = n_features, byrow = TRUE)
  colnames(z) <- important_features
  
  return(z)
}



.sample_shap_coalitions <- function(
    coalitions,
    kernel_weights,
    shap_variance,
    sampling_method,
    seed
) {
  shap_var <- coalition_id <- NULL
  if (sampling_method == "fixed") {
    sampled_coalitions <- list(coalitions)
    
  } else if (sampling_method == "importance") {
    # Probabilistic selection of coalitions. This is very much similar to
    # importance sampling in MCMC. The central concept is that SHAP values for
    # some features are more difficult to determine than those of others. The
    # more difficult features will have a larger overall variance. Therefore we
    # want to focus more on coalitions where these features are present.
    
    # Check if shap_variance contains useful info. Particularly, it cannot be
    # empty, or contain NA or inf values. In that case, just return the fixed
    # coalitions
    if (is_empty(shap_variance)) return(list(coalitions))
    if (any(!is.finite(shap_variance$shap_var))) return(list(coalitions))
    
    # Determine the cost of each input coalition, normalised by their total
    # cost. We want to sample coalitions until the budget (1.0) is exceeded.
    # This is the same budget available to the input coalition. The cost is
    # equal to the corresponding kernel weight for the individual coalition.
    
    coalition_size <- rowSums(coalitions)
    coalition_cost <- kernel_weights[coalition_size + 1L]
    coalition_cost <- coalition_cost / sum(coalition_cost)
    
    # The selection probability for each individual coalition is the variance(s)
    # of the on-feature(s) times the kernel weight, normalised by the total
    # probability over all input coalitions.
    shap_variance <- shap_variance[, list("feature_shap_var" = sum(shap_var)), by = "feature_name"]
    
    # Ensure that feature shap variance and the respective features in the
    # coalition matrix are ordered the same way.
    feature_shap_variance <- shap_variance$feature_shap_var
    names(feature_shap_variance) <- shap_variance$feature_name
    feature_shap_variance <- feature_shap_variance[colnames(coalitions)]
    
    # Compute selection likelihood.
    coalition_probability <- colSums(t(coalitions * kernel_weights[coalition_size + 1L]) * feature_shap_variance)
    coalition_probability <- coalition_probability / sum(coalition_probability)
    
    # Draw 1 / min(coalition_cost) coalitions with resampling, and use these up to
    # and including the coalition where the budget is exceeded.
    n_to_sample <- ceiling(1.0 / min(coalition_cost))
    u_sampled <- fam_runif(n = n_to_sample, seed = seed)
    
    selected_coalition <- sapply(
      u_sampled,
      function(x, u) (which.min(u < x)),
      u = cumsum(coalition_probability)
    )
    
    selected_coalition <- head(
      selected_coalition,
      n = which.min(cumsum(coalition_cost[selected_coalition]) < 1.0)
    )
    
    # Form subsets of coalitions so that each individual coalition only appears
    # once in its bag. This is important so that a different sample will be drawn
    # when the fixed mapping method in ...shap_randomise_mapping_from_coalition is
    # used.
    coalition_table <- data.table::data.table("coalition_id" = selected_coalition)
    coalition_table[, "bag_id" := data.table::rowid(coalition_id)]
    
    sampled_coalitions <- lapply(
      split(coalition_table, by = "bag_id"),
      function(x, coalitions){
        coalitions[x$coalition_id, , drop = FALSE]
      },
      coalitions = coalitions
    )
    
    # Note that antithetic sampling is done later in
    # ...shap_randomise_mapping_from_coalition.
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown sampling_method: ", sampling_method))
  }
  
  return(sampled_coalitions)
}



.shap_data_to_mapping <- function(
  data,
  feature_set
) {
  # Convert data to mapping matrix. This uses the fact that all the values in
  # the feature set
  mapping <- list()
  
  # Maps data to a matrix of integers that establishes a mapping to the feature
  # values in feature set.
  for (feature in names(feature_set)) {
    mapping[[feature]] <- match(data@data[[feature]], feature_set[[feature]])
  }
  
  # Create matrix.
  h <- matrix(unlist(mapping), ncol = length(feature_set))
  colnames(h) <- names(feature_set)
  
  return(h)
}



.shap_mapping_to_data <- function(
    mapping,
    feature_set,
    object
) {
  ..fun <- function(feature, y, x) {
    # Use column in mapping matrix x to lookup value from y.
    return(y[x[, feature]])
  }
  
  # Use lookup to fill data.
  data <- mapply(
    ..fun,
    feature = names(feature_set),
    y = feature_set,
    MoreArgs = list("x" = mapping),
    SIMPLIFY = FALSE
  )
  
  # Set names of list elements.
  names(data) <- names(feature_set)
  
  # Convert to data.table and add identifiers.
  data <- data.table::as.data.table(data)
  
  data <- as_data_object(
    data = data,
    object = object,
    check_stringency = "external"
  )
  
  # Update pre-processing level from none to signature, because we are strictly
  # working with model features here.
  data@preprocessing_level <- "signature"
  
  return(data)
}



.shap_mapping_to_feature_list <- function(
    feature,
    mapping_value, 
    lookup_table
) {
  y <- lookup_table[[feature[1L]]]
  if (is.factor(y)) {
    feature_value <- as.numeric(y)[mapping_value]
    feature_label <- as.character(y)[mapping_value]
  } else {
    feature_value <- y[mapping_value]
    feature_label <- rep_len(NA_character_, length(mapping_value))
  }
  
  return(list(
    "feature_value" = feature_value,
    "feature_label" = feature_label
  ))
}



.shap_randomise_mapping_from_coalition <- function(
    important_features,
    samples,
    coalitions,
    feature_set,
    seed,
    n_min_mappings = 300L,
    mapping_method = "fixed"
) {
  # Determine the number of feature values for each value.
  n_feature_values <- lengths(feature_set)

  # Start random stream
  rstream_object <- .start_random_number_stream(seed)
  
  mapping <- list()
  n_mappings <- 0L
  
  # Ensure that sufficient mappings are generated to limit the effect of
  # overhead on the computation of SHAP values.
  while (n_mappings < n_min_mappings) {
    
    # Generate random values.
    if (mapping_method == "fixed") {
      # Generate a random number for each feature and each sample and coalition
      # set.
      n_random <- nrow(samples) * length(coalitions) * length(important_features)
      
    } else if (mapping_method == "random") {
      # Generate a random number for each feature and each sample and each
      # single coalition. Due to antithetic sampling, the number of coalitions
      # is doubled.
      n_random <- nrow(samples) * 2L * sum(sapply(coalitions, nrow)) * length(important_features)
      
    } else {
      ..error_reached_unreachable_code(paste0("unknown mapping_method: ", mapping_method))
    }
    
    # Stream random numbers.
    x_random <- fam_runif(n = n_random, rstream_object = rstream_object)
    
    random_mapping <- mapply(
      FUN = ..shap_randomise_mapping_from_coalition,
      x = asplit(samples, 1L),
      x_random = split(x_random, fam_cut(seq_along(x_random), nrow(samples))),
      MoreArgs = list(
        "coalitions" = coalitions,
        "n_feature_values" = n_feature_values,
        "mapping_method" = mapping_method
      ),
      USE.NAMES = FALSE,
      SIMPLIFY = FALSE
    )
    
    mapping <- c(mapping, random_mapping)
    n_mappings <- n_mappings + sum(sapply(random_mapping, nrow))
  }
  
  # Concatenate by rows.
  mapping <- do.call(rbind, mapping)

  # Remove duplicates.
  mapping <- unique(mapping)
  
  return(mapping)
}



..shap_randomise_mapping_from_coalition <- function(
  x, 
  x_random,
  coalitions,
  n_feature_values,
  mapping_method
) {
  # Split random values.
  if (length(coalitions) == 1L) {
    x_random <- list(x_random)
    
  } else {
    if (mapping_method == "fixed") {
      # Each coalition set requires the same number of random values.
      x_random <- split(x_random, fam_cut(seq_along(x_random), length(coalitions)))
      
    } else if (mapping_method == "random") {
      # Due to antithetic sampling, the number of coalitions is doubled.
      n_coalitions_in_bag <- 2L * sapply(coalitions, nrow)
      x_random <- split(
        x_random, 
        rep(seq_along(coalitions), n_coalitions_in_bag * length(n_feature_values))
      )
    }
  }
  
  # Loop over coalition sets.
  mapping <- mapply(
    FUN = ...shap_randomise_mapping_from_coalition,
    coalitions = coalitions,
    x_random = x_random,
    MoreArgs = list(
      "x" = x,
      "n_feature_values" = n_feature_values,
      "mapping_method" = mapping_method
    ),
    USE.NAMES = FALSE,
    SIMPLIFY = FALSE
  )
  
  return(do.call(rbind, mapping))
}



...shap_randomise_mapping_from_coalition <- function(
    coalitions,
    x_random,
    x,
    n_feature_values,
    mapping_method
) {
  
  ..select_shap_feature_mapping <- function(
    available_feature_values,
    random_values
  ) {
    # Sample with replacement.
    x_indices <- as.integer(ceiling(random_values * length(available_feature_values)))
    x_indices[x_indices == 0L] <- 1L
    
    return(available_feature_values[x_indices])
  }
  
  # Generate antithetic coalitions from input.
  coalitions <- rbind(coalitions, !coalitions)
  
  # Get important features.
  important_features <- colnames(coalitions)
  unimportant_features <- setdiff(names(n_feature_values), important_features)
  
  # The mapping method determines how values are drawn.
  #
  # - "fixed": a single sample, not_x, that is fully distinct from x is 
  #    randomly drawn. on-features are copied from x, and off-features from
  #    not_x.
  # - "random": off-features are drawn randomly.
  
  mapping <- list()
  if (mapping_method == "fixed") {
    for (ii in seq_along(important_features)) {
      # Determine eligible features from in-coalition (on) and off-coalition
      # (off) features.
      feature <- important_features[ii]
      on_feature_set <- unname(x[feature])
      off_feature_set <- seq_len(n_feature_values[feature])[-on_feature_set]
      
      # Sample a single value from the off-feature set, and append to the
      # on-feature set. This forms the look-up table for forming coalitions.
      feature_set <- c(
        on_feature_set,
        ..select_shap_feature_mapping(
          available_feature_values = off_feature_set,
          random_values = x_random[ii]
        )
      )
      
      # Determine which value from feature_set should be used. Index 1L
      # corresponds to the in-coalition feature value, and index 2L to the
      # off-coalition feature value.
      lookup_vector <- 1L + !coalitions[, feature]
      
      # Add features to mapping.
      mapping[[feature]] <- feature_set[lookup_vector]
    }
    
    for (ii in seq_along(unimportant_features)) {
      # Unimportant features are simply repeated without coalition.
      feature <- unimportant_features[ii]
      mapping[[feature]] <- rep(unname(x[feature]), nrow(coalitions))
    }
      
  } else if (mapping_method == "random") {
    # Determine the number of feature values to samples for off-coalition
    # features. This should be the same number for each feature in antithetical
    # sampling.
    n_to_draw <- colSums(!coalitions)
    jj <- 0L
    
    for (ii in seq_along(important_features)) {
      # Determine eligible features from in-coalition (on) and off-coalition
      # (off) features.
      feature <- important_features[ii]
      on_feature_set <- unname(x[feature])
      off_feature_set <- seq_len(n_feature_values[feature])[-on_feature_set]
      
      # Sample the off-feature set, and append to the on-feature set. This forms
      # the look-up table for forming coalitions.
      feature_set <- c(
        on_feature_set,
        ..select_shap_feature_mapping(
          available_feature_values = off_feature_set,
          random_values = x_random[(1L:n_to_draw[feature]) + jj]
        )
      )
      
      # Update offset.
      jj <- jj + n_to_draw[feature]
      
      # Accumulate off-coalition elements. E.g. with coalitions (across samples
      # for each feature) [0, 1, 1, 0], the lookup-vector is [1, 1, 1, 2].
      lookup_vector <- cumsum(!coalitions[, feature])
      
      # Reset in-coalition elements of the lookup vector, yielding, e.g. [1, 0, 0,
      # 2], and increment by 1. This results in indices (e.g. [2, 1, 1, 3])
      # referring to the feature set, with index 1 corresponding to the
      # in-coalition value.
      lookup_vector <- 1L + lookup_vector * !coalitions[, feature]
      
      # Add features to mapping.
      mapping[[feature]] <- feature_set[lookup_vector]
    }
    
    for (ii in seq_along(unimportant_features)) {
      # Unimportant features are simply repeated without coalition.
      feature <- unimportant_features[ii]
      mapping[[feature]] <- rep(unname(x[feature]), nrow(coalitions))
    }
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown mapping_method: ", mapping_method))
  }
  
  # Convert to matrix. Mapping consists of columns (that are first ordered
  # correctly before being flattened), and the matrix is then filled by column.
  mapping <- matrix(
    unlist(mapping[names(n_feature_values)]),
    ncol = length(n_feature_values)
  )
  colnames(mapping) <- names(n_feature_values)
  
  return(mapping)
}



.compute_shap_kernel_weights <- function(n, individual_coalition = FALSE) {
  # Form a lookup-table for kernel weights.
  n_present <- seq_len(n + 1L) - 1L
  n_permutations <- choose(n, n_present)
  kernel_weights <- (n - 1.0) / (n_permutations * n_present * (n - n_present))
  kernel_weights[!is.finite(kernel_weights)] <- 0.0
  
  # Normalise kernel-weights to 1.
  kernel_weights <- kernel_weights / sum(kernel_weights)
  
  if (individual_coalition) {
    # Determine weights for individual unique coalitions.
    kernel_weights <- kernel_weights / n_permutations
  }
  
  return(kernel_weights)
}



.compute_shap_matrices <- function(
  important_features,
  samples,
  sample_predictions,
  sample_id,
  mapping,
  predicted_values,
  phi_0,
  kernel_weights
) {
  # Replace NA in samples and mapping.
  samples[is.na(samples)] <- 0L
  mapping[is.na(mapping)] <- 0L
  
  # Select only important features.
  samples <- samples[, important_features, drop = FALSE]
  mapping <- mapping[, important_features, drop = FALSE]
  
  # Compute A and b matrices for each sample.
  new_matrices <- mapply(
    ..compute_shap_matrices,
    x = asplit(samples, MARGIN = 1L),
    v_0 = asplit(sample_predictions, MARGIN = 1L),
    sample_id = sample_id,
    MoreArgs = list(
      "kernel_weights" = kernel_weights,
      "mapping" = mapping,
      "predicted_values" = predicted_values,
      "phi_0" = phi_0
    ),
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )
  
  return(new_matrices)
  
  # Update full matrix. THIS IS CURRENTLY NOT USED.
  # We follow the recipe by Covert and Lee (2021), which means that we update
  # the A and b matrices each iteration.
  # if (is.null(matrices)) {
  #   matrices <- new_matrices
  #   names(matrices) <- sample_id
  #   matrices <- lapply(
  #     matrices,
  #     function(x) {
  #       x$n_iter <- 1L
  #       return(x)
  #     }
  #   )
  #   
  # } else {
  #   matrices <- mapply(
  #     FUN = .update_shap_matrices,
  #     old = matrices,
  #     new = new_matrices,
  #     SIMPLIFY = FALSE,
  #     USE.NAMES = FALSE
  #   )
  # }
  
  # Return both the full and temporary matrices.
}



..compute_shap_matrices <- function(
    x,
    v_0,
    sample_id,
    kernel_weights,
    mapping,
    predicted_values,
    phi_0
) {
  # Prevent notes due to data.table.
  weight <- NULL
  
  # x is the mapping corresponding to the sample. First we determine the 
  # coalitions pertaining to current sample. Since `==` is operating by column,
  # we can simply transpose the mapping matrix so that rows become columns. Then
  # the comparison is performed on the columns representing each row, and the
  # result is transposed again.
  coalitions <- t(t(mapping) == c(x))
  
  # Compute the number of features "present" in each coalition. 
  n_coalition_size <- rowSums(coalitions)
  weights <- kernel_weights[n_coalition_size + 1L]
  non_zero_weights <- weights > 0.0
  
  # Check for empty weights.  
  if (!any(non_zero_weights)) return(NULL)
  
  # Weighted least squares solves for coefficients beta as follows:
  # beta = (t(X) W X)^-1 t(X)W y
  # In the context of kernelSHAP, this means:
  #    beta = phi
  #       X = Z (coalitions),
  #       W = diag(pi) (kernel_weights)
  #   and y = f(h(z)) - phi_0
  X <- coalitions[non_zero_weights, , drop = FALSE]
  
  # This ensures that phi_0 is subtracted row-wise.
  y <- t(t(predicted_values[non_zero_weights, , drop = FALSE] - phi_0))
  
  # kernelSHAP originally was defined by sampling coalitions, with each
  # coalition drawn probabilistically according to the SHAP kernel weights. This
  # means that the appearance of coalitions would stochastically mirror their
  # probabilities. Here we have a fixed set of coalitions, and coalitions that
  # are randomly formed from the entire set of decisions that should be
  # explained (and their fixed coalitions). This means that our coalitions are
  # not distributed as expected, and cannot be used without updating the
  # weights.
  
  # First determine how often individual coalitions appear. This is currently
  # the most expensive part of this function. I thought of solutions to speed up
  # the process, other than data.table::frank. Hashing each row using
  # rlang::hash or paste0 is several times slower. Using a filter technique
  # where we count which individual coalitions are present is not scalable for
  # large feature sets due to large number of permutations of coalitions. Even
  # though that solution could be quite fast when few features are present (and
  # thus a non-sparse coalition set is present in `coalitions`), the likely
  # gains will be minimal. Another alternative is to encode coalitions as a sum
  # of positional powers of 2 ( feature 1: TRUE / FALSE * 2^0; feature 2: TRUE /
  # FALSE 2^1; etc.), but that would still be problematic with large numbers of
  # features due to integer bit precision.
  coalition_id <- data.table::frank(data.table::as.data.table(X), ties.method = "dense")
  n_coalition_instances <- integer(length(coalition_id))
  for (ii in seq_len(max(coalition_id))) {
    instance_in_coalition <- coalition_id == ii
    n_coalition_instances[instance_in_coalition] <- sum(instance_in_coalition)
  }
  
  # Then re-weight probability of individual coalitions, and normalise.
  w <- weights[non_zero_weights]
  w <- w / n_coalition_instances
  w <- w / sum(w)
  
  # Instead of computing a diagonal matrix, we rely on equivalent element-wise
  # multiplications (which are considerably cheaper).
  return(list(
    "A" = t(X) %*% (X * w),
    "b" = t(X) %*% (y * w),
    "v_0" = as.numeric(v_0),
    "phi_0" = phi_0,
    "sample_id" = sample_id,
    "sample_mapping" = x
  ))
}



.update_shap_matrices <- function(
    old,
    new
) {
  old$A <- old$A + new$A
  old$b <- old$b + new$b
  old$n_iter <- old$n_iter + 1L
  old$phi_0 <- new$phi_0
  
  return(old)
}



.compute_shap_value <- function(
  shap_matrices
) {
  # Compute shap values.
  shap_values <- lapply(
    shap_matrices,
    ..compute_shap_value
  )
  shap_values <- data.table::rbindlist(shap_values)
  
  return(shap_values)
}



..compute_shap_value <- function(x) {
  if (is.null(x)) return(NULL)
  
  A_inv <- matrix_pseudo_inverse(x$A)
  b <- x$b
  
  # Compute initial coefficients.
  phi <- A_inv %*% b
  
  # Due to the local accuracy criterion the sum of of the SHAP values plus phi_0
  # should be equal to the predicted value. We estimate the SHAP values under
  # this constraint.
  phi <- A_inv %*% t(t(b) - (colSums(phi) - (x$v_0 - x$phi_0)) / sum(A_inv))
  
  shap_values <- list(
    "sample_id" = x$sample_id,
    "feature_name" = rep(colnames(x$A), times = ncol(x$b)),
    "feature_value_mapping" = rep(x$sample_mapping, times = ncol(x$b)),
    "shap_value" = c(phi),
    "shap_outcome" = rep(colnames(x$b), each = ncol(x$A))
  )
  
  return(shap_values)
}



.compute_shap_value_single_feature <- function(
    important_features,
    mapping,
    sample_id,
    feature_set,
    predicted_values,
    phi_0
) {
  # Get value mapping.
  mapping <- mapping[, important_features]
  
  data <- data.table::data.table(
    "sample_id" = sample_id,
    "feature_name" = important_features,
    "feature_value_mapping" = rep(mapping, times = ncol(predicted_values)),
    "shap_value" = c(t(t(predicted_values) - phi_0)),
    "shap_outcome" = rep(colnames(predicted_values), each = nrow(predicted_values))
  )
  
  return(data)
}



.evaluate_shap_convergence <- function(
    shap_variance,
    tolerance
) {
  # Compute sample error of the mean for each shap value.
  sem_values <- sqrt(shap_variance$shap_var / shap_variance$n)
  if (any(!is.finite(sem_values))) return(FALSE)
  # TODO: remove
  # cat(paste0("sum SEM: ", sum(sem_values), " ; total converged: ", sum(sem_values <= tolerance), "\n"))
  return(all(sem_values <= tolerance))
}



.predict_from_coalition <- function(
    mapping,
    feature_set,
    object,
    ensemble_method,
    evaluation_time
) {
  # Set prediction type.
  prediction_type <- ifelse(
    object@outcome_type %in% c("survival", "competing_risk"),
    "survival_probability", 
    "default"
  )
  
  # Convert input to dataObject
  data <- .shap_mapping_to_data(
    mapping = mapping,
    feature_set = feature_set,
    object = object
  )
  
  if (object@outcome_type == "survival") {
    prediction_list <- list()
    
    for (ii in seq_along(evaluation_time)) {
      # Predict input data
      prediction_data <- predict(
        object = object,
        newdata = data,
        ensemble_method = ensemble_method,
        time = evaluation_time[ii],
        type = prediction_type,
        .as_prediction_table = TRUE
      )
      
      # Check if all predictions are valid.
      if (!all_predictions_valid(prediction_data)) return(NULL)
      
      # Convert to data.table.
      prediction_data <- .as_data_table(prediction_data)[, mget(prediction_data@value_column)]
      prediction_list[[ii]] <- matrix(prediction_data$predicted_outcome, ncol = 1L)
    }
    
    prediction_data <- do.call(cbind, prediction_list)
    colnames(prediction_data) <- as.character(evaluation_time)
    
  } else {
    # Predict input data
    prediction_data <- predict(
      object = object,
      newdata = data,
      ensemble_method = ensemble_method,
      type = prediction_type,
      .as_prediction_table = TRUE
    )
    
    # Check if all predictions are valid.
    if (!all_predictions_valid(prediction_data)) return(NULL)
    
    # Convert to data.table.
    prediction_data <- .as_data_table(prediction_data)[, mget(prediction_data@value_column)]
    
    if (object@outcome_type == "continuous") {
      prediction_data <- matrix(prediction_data$predicted_outcome, ncol = 1L)
      colnames(prediction_data) <- "predicted_outcome"
      
    } else if (object@outcome_type %in% c("multinomial")) {
      probability_columns <- get_outcome_class_levels(object)
      prediction_data <- as.matrix(prediction_data[, mget(probability_columns)])
      
    } else if (object@outcome_type %in% c("binomial")) {
      probability_column <- utils::tail(get_outcome_class_levels(object), n = 1L)
      prediction_data <- as.matrix(prediction_data[, mget(probability_column)])
      
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }
  }
  
  return(prediction_data)
}



.hash_mapping <- function(x) {
  return(apply(
    x,
    MARGIN = 1L,
    FUN = rlang::hash,
    simplify = TRUE
  ))
}



.extract_shap_summary <- function(
    x
) {
  # Prevent NOTES due to non-standard evaluation
  feature_name <- feature_value_mapping <- NULL
  
  # Generate object using the incoming familiarDataElementSHAP object as a
  # template.
  data_element <- methods::new(
    "familiarDataElementSHAPSummary",
    x
  )
  
  # Clean reporting elements.
  data_element@data <- NULL
  
  if (is_empty(x@data)) return(data_element)
  
  # Get data_mapping and turn into a long data.table.
  mapping_data <- data.table::as.data.table(x@data_mapping)
  mapping_data[, "sample_id" := x@sample_identifiers]
  mapping_data <- data.table::melt(
    data = mapping_data,
    id.vars = "sample_id",
    variable.name = "feature_name",
    value.name = "feature_value_mapping"
  )
  
  # Insert feature values in mapping data.
  mapping_data[
    ,
    c("feature_value", "feature_label") := .shap_mapping_to_feature_list(
      feature = feature_name,
      mapping_value = feature_value_mapping, 
      lookup_table = x@lookup_table
    ),
    by = "feature_name"
  ]
  
  # Cartesian merge.
  summary_data <- merge(
    x = x@data,
    y = mapping_data,
    by = c("feature_name", "feature_value_mapping", "sample_id"),
    allow.cartesian = TRUE
  )
  
  # Drop unused columns.
  summary_data[, feature_value_mapping := NULL]
  
  # Set data.
  data_element@data <- summary_data
  
  # Set identifiers
  data_element@grouping_column <- c(
    setdiff(data_element@grouping_column, "feature_value_mapping"),
    c("feature_value", "feature_label")
  )
  
  return(data_element)
}



.extract_shap_force <- function(
    x
) {
  # Prevent NOTES due to non-standard evaluation
  feature_name <- feature_value_mapping <- NULL
  
  # Generate object using the incoming familiarDataElementSHAP object as a
  # template.
  data_element <- methods::new(
    "familiarDataElementSHAPForce",
    x
  )
  
  # Clean reporting elements.
  data_element@data <- NULL
  
  if (is_empty(x@data)) return(data_element)
  
  # Get data_mapping and turn into a long data.table.
  mapping_data <- data.table::as.data.table(x@data_mapping)
  mapping_data[, "sample_id" := x@sample_identifiers]
  mapping_data <- data.table::melt(
    data = mapping_data,
    id.vars = "sample_id",
    variable.name = "feature_name",
    value.name = "feature_value_mapping"
  )
  
  # Insert feature values in mapping data.
  mapping_data[
    ,
    c("feature_value", "feature_label") := .shap_mapping_to_feature_list(
      feature = feature_name,
      mapping_value = feature_value_mapping, 
      lookup_table = x@lookup_table
    ),
    by = "feature_name"
  ]
  
  # Prediction data
  prediction_data <- data.table::as.data.table(x@predicted_values)
  prediction_data[, "sample_id" := x@sample_identifiers]
  prediction_data <- data.table::melt(
    data = prediction_data,
    id.vars = "sample_id",
    variable.name = "shap_outcome",
    value.name = "prediction"
  )
  
  # Cartesian merge.
  force_data <- merge(
    x = x@data,
    y = mapping_data,
    by = c("feature_name", "feature_value_mapping", "sample_id"),
    allow.cartesian = TRUE
  )
  
  merge_cols <- "sample_id"
  if ("shap_outcome" %in% colnames(force_data)) merge_cols <- c(merge_cols, "shap_outcome")
  
  force_data <- merge(
    x = force_data,
    y = prediction_data,
    by = merge_cols
  )
  
  # Drop unused columns.
  force_data[, feature_value_mapping := NULL]

  # Set data.
  data_element@data <- force_data
  
  # Set identifiers
  data_element@grouping_column <- c(
    setdiff(data_element@grouping_column, "feature_value_mapping"),
    c("feature_value", "feature_label", "prediction")
  )
  
  return(data_element)
}



.extract_shap_dependence <- function(
    x,
    feature_x,
    feature_y
) {
  data_element_list <- list()
  iter_id <- 1L
  for (current_feature_x in feature_x) {
    if (!current_feature_x %in% names(x@lookup_table)) {
      ..warning(paste0(
        current_feature_x, " is not part of the feature set used by the model. ",
        "The following features were used: ", paste_s(names(x@lookup_table))
      ))
      next
    }
      
    for (current_feature_y in feature_y) {
      if (!current_feature_y %in% names(x@lookup_table)) {
        ..warning(paste0(
          current_feature_y, " is not part of the feature set used by the model. ",
          "The following features were used: ", paste_s(names(x@lookup_table))
        ))
        next
      }
      
      data_element_list[[iter_id]] <- ..extract_shap_dependence(
        x = x,
        feature_x = current_feature_x,
        feature_y = current_feature_y
      )
      
      iter_id <- iter_id + 1L
    }
  }
  
  if (is_empty(data_element_list)) return(NULL)
  
  return(data_element_list)
}



..extract_shap_dependence <- function(
    x,
    feature_x,
    feature_y
) {
  # Prevent NOTES due to non-standard evaluation
  feature_name <- feature_value_mapping <- NULL
  
  # Generate object using the incoming familiarDataElementSHAP object as a
  # template.
  data_element <- methods::new(
    "familiarDataElementSHAPDependence",
    x
  )
  
  # Add feature x and feature y as identifiers to the dataset.
  data_element <- add_data_element_identifier(
    x = data_element,
    feature_x = feature_x
  )[[1L]]
  data_element <- add_data_element_identifier(
    x = data_element,
    feature_y = feature_y
  )[[1L]]
  
  # Clean reporting elements.
  data_element@data <- NULL
  
  if (is_empty(x@data)) return(data_element)
  
  # Get data_mapping and turn into a long data.table.
  mapping_data <- data.table::as.data.table(x@data_mapping)
  mapping_data[, "sample_id" := x@sample_identifiers]
  mapping_data <- data.table::melt(
    data = mapping_data,
    id.vars = "sample_id",
    variable.name = "feature_name",
    value.name = "feature_value_mapping"
  )
  
  # Insert feature values in mapping data.
  mapping_data[
    ,
    c("feature_value", "feature_label") := .shap_mapping_to_feature_list(
      feature = feature_name,
      mapping_value = feature_value_mapping, 
      lookup_table = x@lookup_table
    ),
    by = "feature_name"
  ]
  
  # To create a dependence plot, we need for:
  # feature x: its value and its SHAP value.
  # feature y: its value
  # These are linked by the sample identifier.
  
  feature_x_data <- data.table::copy(mapping_data[feature_name == feature_x])
  feature_y_data <- data.table::copy(mapping_data[feature_name == feature_y])
  
  # Cartesian merge.
  dependence_data <- merge(
    x = x@data,
    y = feature_x_data,
    by = c("feature_name", "feature_value_mapping", "sample_id"),
    allow.cartesian = TRUE
  )
  
  # Update column names.
  data.table::setnames(
    dependence_data,
    old = c("feature_value", "feature_label"),
    new = c("feature_value_x", "feature_label_x")
  )
  dependence_data[, feature_value_mapping := NULL]
  dependence_data[, feature_name := NULL]
  
  # Add feature_y_data.
  dependence_data <- merge(
    x = dependence_data,
    y = feature_y_data,
    by = "sample_id"
  )
  
  # Update column names.
  data.table::setnames(
    dependence_data,
    old = c("feature_value", "feature_label"),
    new = c("feature_value_y", "feature_label_y")
  )
  dependence_data[, feature_value_mapping := NULL]
  dependence_data[, feature_name := NULL]
  
  # Update column order.
  col_order <- c("sample_id", "feature_value_x", "feature_label_x")
  if ("shap_outcome" %in% colnames(dependence_data)) col_order <- c(col_order, "shap_outcome")
  col_order <- c(col_order, "shap_value", "feature_value_y", "feature_label_y")
  data.table::setcolorder(
    dependence_data,
    neworder = col_order
  )
  
  # Set data.
  data_element@data <- dependence_data
  
  # Update grouping columns.
  data_element@grouping_column <- c(
    setdiff(data_element@grouping_column, c("feature_name", "feature_value_mapping")),
    c("feature_value_x", "feature_label_x", "feature_value_y", "feature_label_y")
  )
  
  return(data_element)
}


# export_shap (generic) --------------------------------------------------------

#'@title Extract and export individual conditional expectation data.
#'
#'@description Extract and export individual conditional expectation data.
#'
#'@param feature_x (*optional*) Feature(s) whose SHAP values are used for
#'  determining dependence.
#'@param feature_y (*optional*) Feature(s) whose values are used to show 
#'  interaction with the feature(s) in `feature_x`.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_ice_data
#'@md
#'@rdname export_shap-methods
setGeneric(
  "export_shap",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    feature_x = NULL,
    feature_y = NULL,
    ...
  ) {
    standardGeneric("export_shap")
  }
)



# export_shap (collection) -----------------------------------------------------

#'@rdname export_shap-methods
setMethod(
  "export_shap",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    feature_x = NULL,
    feature_y = NULL,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)

    # Generate data for summary plots.
    summary_data_elements <- lapply(
      object@shap_data,
      .extract_shap_summary
    )
    
    # Export summary data.
    summary_data <- .export(
      x = object,
      data_elements = summary_data_elements,
      dir_path = dir_path,
      aggregate_results = TRUE,
      object_class = "familiarDataElementSHAPSummary",
      type = "explanation",
      subtype = "shap_summary"
    )
    
    # Generate data for force plots.
    force_data_elements <- lapply(
      object@shap_data,
      .extract_shap_force
    )
    
    # Export data for force plots.
    force_data <- .export(
      x = object,
      data_elements = force_data_elements,
      dir_path = dir_path,
      aggregate_results = TRUE,
      object_class = "familiarDataElementSHAPForce",
      type = "explanation",
      subtype = "shap_force"
    )
    
    dependence_data <- NULL
    if (!is.null(feature_x) && !is.null(feature_y)) {
      # Generate data for SHAP dependence plots.
      dependence_data_elements <- lapply(
        object@shap_data,
        .extract_shap_dependence,
        feature_x = feature_x,
        feature_y = feature_y
      )
      
      dependence_data <- .export(
        x = object,
        data_elements = dependence_data_elements,
        dir_path = dir_path,
        aggregate_results = TRUE,
        object_class = "familiarDataElementSHAPDependence",
        type = "explanation",
        subtype = "shap_dependence"
      )
    }
    
    # Add to list.
    data_list <- c(
      "shap_summary" = summary_data,
      "shap_force" = force_data,
      "shap_dependence" = dependence_data
    )
    
    if (export_collection) {
      data_list <- c(
        data_list,
        list("collection" = object)
      )
    } 
    
    return(data_list)
  }
)



# export_shap (general) ----------------------------------------------------

#'@rdname export_shap-methods
setMethod(
  "export_shap",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "export_shap",
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_shap,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

.get_run_list <- function(
    iteration_list,
    data_id,
    run_id = NULL
) {
  
  # Check if data_id has any length.
  if (length(data_id) == 0L) return(list())
  
  # Return an empty list if data_id equals 0.
  if (data_id == 0L) return(list())
  
  # Return part of the run list.
  if (is.null(run_id)) {
    return(iteration_list[[as.character(data_id)]]$run)
    
  } else {
    return(iteration_list[[as.character(data_id)]]$run[[as.character(run_id)]])
  }
}



.get_sample_identifiers <- function(
    run = NULL,
    iteration_list = NULL,
    data_id = NULL,
    run_id = NULL,
    train_or_validate
) {
  
  # Get run from iter_list if not provided directly.
  if (is.null(run)) {
    run <- .get_run_list(
      iteration_list = iteration_list,
      data_id = data_id,
      run_id = run_id
    )
  }
  
  # Extract training or validation data - note that run$valid_samples can be
  # NULL.
  if (train_or_validate == "train") {
    samples <- run$train_samples
    
  } else {
    samples <- run$valid_samples
  }
  
  if (!data.table::is.data.table(samples)) {
    samples <- data.table::data.table("sample_id" = samples)
  }
  
  return(samples)
}



.get_iteration_identifiers <- function(run, perturb_level = NULL) {
  # Extract the runs
  run_table <- run$run_table
  
  # Set the perturbation level that will be returned.
  if (is.null(perturb_level)) {
    indicated_perturb_level <- tail(run_table, n = 1L)$perturb_level
    
  } else {
    indicated_perturb_level <- perturb_level
  }
  
  # Get the corresponding row
  run_table <- run_table[perturb_level == indicated_perturb_level, ]
  
  # Extract data and run identifiers.
  return(list(
    "data" = run_table$data_id[1L],
    "run" = run_table$run_id[1L],
    "perturb_level" = run_table$perturb_level[1L],
    "perturbation" = run_table$perturbation[1L]
  ))
}



.get_preprocessing_iteration_identifiers <- function(run) {
  # Find the identifiers for the data and run identifiers that allow for
  # pre-processing.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  can_pre_process <- NULL
  
  # Extract the run table
  run_table <- run$run_table
  
  # Find the last entry that is available for pre-processing
  run_table <- tail(run_table[can_pre_process == TRUE, ], n = 1L)
  
  # Extract data and run identifiers.
  return(list(
    "data" = run_table$data_id[1L],
    "run" = run_table$run_id[1L],
    "perturb_level" = run_table$perturb_level[1L],
    "perturbation" = run_table$perturbation[1L]
  ))
}



.get_process_step_data_identifier <- function(project_info, process_step) {
  # Get the main data id for a step in the overall modelling process.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  vimp <- train <- external_validation <- NULL
  
  # Load experiment data table
  experiment_table <- project_info$experiment_setup
  
  if (process_step == "vimp") {
    # Find row on where feature selection takes place and extract the main data
    # id.
    main_data_id <- experiment_table[vimp == TRUE, ]$main_data_id[1L]
    
  } else if (process_step %in% c("mb")) {
    # Find row where model building takes place and extract the main data id.
    main_data_id <- experiment_table[train == TRUE, ]$main_data_id[1L]
    
  } else if (process_step == "ev") {
    # Check if external validation is present; otherwise return an illegal main
    # data id.
    if (any(experiment_table$external_validation)) {
      main_data_id <- experiment_table[external_validation == TRUE, ]$main_data_id[1L]
      
    } else {
      main_data_id <- -1L
    }
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".get_process_step_data_identifier: encountered unknown process step code: ",
      process_step
    ))
  }
  
  return(main_data_id)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarMetricConcordanceIndex -----------------------------------------------
setClass(
  "familiarMetricConcordanceIndex",
  contains = "familiarMetric",
  slots = list("time" = "numeric"),
  prototype = list("time" = Inf)
)



# Harrell's Concordance Index --------------------------------------------------
setClass(
  "familiarMetricConcordanceIndexHarrell",
  contains = "familiarMetricConcordanceIndex",
  prototype = methods::prototype(
    name = "Concordance Index",
    baseline_value = 0.5,
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarMetricConcordanceIndex"),
  function(
    .Object,
    time = NULL,
    object = NULL,
    prediction_type = NULL,
    ...
  ) {
    # Update with parent class first.
    .Object <- callNextMethod(.Object, ...)

    # Attempt to set time.
    if (!is.null(time)) {
      .Object@time <- time
      
    } else if (
      is(object, "familiarModel") ||
      is(object, "familiarEnsemble") ||
      is(object, "familiarVimpMethod")
    ) {
      # Obtain time from the outcome info.
      .Object@time <- object@outcome_info@time
    }

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMetricConcordanceIndex"),
  function(object, ...) {
    return(object@outcome_type %in% c("survival", "competing_risk"))
  }
)



# compute_metric_score (Harrell's Concordance Index) ---------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricConcordanceIndexHarrell"),
  function(metric, data, ...) {
    # Compute a standard concordance index.
    score <- .compute_concordance_index(
      metric = metric,
      data = data,
      ...
    )

    # Invert the score for risk-like predictions.
    if (!is(data, "predictionTableSurvivalTime")) {
      score <- 1.0 - score
    }
    
    return(score)
  }
)



.compute_concordance_index <- function(
    metric,
    data,
    time = NULL,
    weight_function = NULL
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome_time <- outcome_event <- NULL

  if (is_empty(data)) return(NA_real_)
  
  if (!is(data, "predictionTableSurvival")) {
    ..error_data_not_prediction_table(data, "predictionTableSurvival")
  }
  
  # Remove any entries that lack valid predictions.
  data <- remove_invalid_predictions(data)
  
  # Remove any entries that lack observed values.
  data <- filter_missing_outcome(data)
  if (is_empty(data)) return(NA_real_)

  # Determine maximum time for censoring. There are multiple ways to obtain
  # maximum time, in order of preference:
  #
  # 1. From function input.
  #
  # 2. From the metric object (metric), which may have inherited it from the
  # model or model ensemble.
  #
  # 3. From the prediction table object (data).
  #
  # As fallback, max time is set to infinite.
  max_time <- time
  if (is.null(max_time) && is.finite(metric@time)) max_time <- metric@time
  if (is.null(max_time) && methods::.hasSlot(data, "time")) {
    if (is.finite(data@time)) max_time <- data@time 
  }
  if (is.null(max_time)) max_time <- Inf
  
  # Check max_time
  .check_number_in_valid_range(
    x = max_time,
    var_name = "time",
    range = c(0.0, Inf),
    closed = c(FALSE, TRUE)
  )
  
  # Apply max time. Everything beyond max time is censored for the purpose of
  # computing a concordance index.
  data <- .as_data_table(data)
  data[outcome_time > max_time, "outcome_event" := 0L]

  # All competing risks are treated as censoring for computing the concordance
  # index.
  data[outcome_event > 1L, "outcome_event" := 0L]

  # Check that there are any events.
  if (nrow(data[outcome_event == 1L]) == 0L) return(NA_real_)

  # Remove any samples that are censored prior to the first event.
  earliest_event <- min(data[outcome_event == 1L]$outcome_time)

  # Keep only data from the earliest event onward.
  data <- data[outcome_time >= earliest_event]

  # Check that sufficient data is remaining.
  if (nrow(data) < 2L) return(NA_real_)

  # Compute a concordance index score
  score <- ..compute_concordance_index(
    x = data$predicted_outcome,
    time = data$outcome_time,
    event = data$outcome_event,
    weight_function = weight_function
  )

  return(score)
}



..compute_concordance_index <- function(
    x,
    time,
    event, 
    weight_function = NULL,
    ...
) {
  # Compute using concordance in the survival package.
  ci <- survival::concordance(
    Surv(time, event) ~ x,
    data = list("x" = x, "time" = time, "event" = event)
  )$concordance
  
  # Check if the concordance index is valid
  if (!is.finite(ci)) ci <- NA_real_
  
  return(ci)
}



.get_available_concordance_index_metrics <- function() {
  return(c(.get_available_concordance_index_harrell()))
}


.get_available_concordance_index_harrell <- function() {
  return(c("concordance_index", "c_index", "concordance_index_harrell", "c_index_harrell"))
}

familiar:::test_all_metrics_available(
  metrics = familiar:::.get_available_concordance_index_metrics()
)

# Skip remainder on CRAN due to runtimes.
testthat::skip_on_cran()

# power.transform and other packages are required.
if (!rlang::is_installed("power.transform")) testthat::skip()

familiar:::test_all_metrics(
  metrics = familiar:::.get_available_concordance_index_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE
)



.as_risk <- function(x) {
  familiar::as_prediction_table(
    x = x$predicted_outcome,
    y = list("outcome_time" = x$outcome_time, "outcome_event" = x$outcome_event),
    type = "hazard_ratio"
  )
}

.as_lifetime <- function(x) {
  familiar::as_prediction_table(
    x = x$predicted_outcome,
    y = list("outcome_time" = x$outcome_time, "outcome_event" = x$outcome_event),
    type = "expected_survival_time"
  )
}

data_good_no_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 1, 1, 1),
  "predicted_outcome" = c(10, 8, 6, 4, 2)
)

data_inv_no_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 1, 1, 1),
  "predicted_outcome" = c(2, 4, 6, 8, 10)
)

data_bad_no_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 1, 1, 1),
  "predicted_outcome" = c(5, 5, 5, 5, 5)
)

data_moderate_no_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 1, 1, 1),
  "predicted_outcome" = c(10, 6, 8, 2, 4)
)

data_good_init_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(0, 1, 1, 1, 1),
  "predicted_outcome" = c(10, 8, 6, 4, 2)
)

data_inv_init_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(0, 1, 1, 1, 1),
  "predicted_outcome" = c(2, 4, 6, 8, 10)
)

data_bad_init_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(0, 1, 1, 1, 1),
  "predicted_outcome" = c(5, 5, 5, 5, 5)
)

data_moderate_init_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(0, 1, 1, 1, 1),
  "predicted_outcome" = c(10, 6, 8, 2, 4)
)

data_good_end_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 1, 1, 0),
  "predicted_outcome" = c(10, 8, 6, 4, 2)
)

data_inv_end_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 1, 1, 0),
  "predicted_outcome" = c(2, 4, 6, 8, 10)
)

data_bad_end_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 1, 1, 0),
  "predicted_outcome" = c(5, 5, 5, 5, 5)
)

data_moderate_end_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 1, 1, 0),
  "predicted_outcome" = c(10, 6, 8, 2, 4)
)

data_good_mid_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 0, 0, 1),
  "predicted_outcome" = c(10, 8, 6, 4, 2)
)

data_inv_mid_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 0, 0, 1),
  "predicted_outcome" = c(2, 4, 6, 8, 10)
)

data_bad_mid_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 0, 0, 1),
  "predicted_outcome" = c(5, 5, 5, 5, 5)
)

data_moderate_mid_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(1, 1, 0, 0, 1),
  "predicted_outcome" = c(10, 6, 8, 2, 4)
)

data_all_censoring_risk <- data.table::data.table(
  "outcome_time" = c(1, 2, 3, 4, 5),
  "outcome_event" = c(0, 0, 0, 0, 0),
  "predicted_outcome" = c(10, 8, 6, 4, 2)
)

data_list <- list(
  "good_no_censoring_risk" = data_good_no_censoring_risk,
  "inv_no_censoring_risk" = data_inv_no_censoring_risk,
  "bad_no_censoring_risk" = data_bad_no_censoring_risk,
  "moderate_no_censoring_risk" = data_moderate_no_censoring_risk,
  "good_init_censoring_risk" = data_good_init_censoring_risk,
  "inv_init_censoring_risk" = data_inv_init_censoring_risk,
  "bad_init_censoring_risk" = data_bad_init_censoring_risk,
  "moderate_init_censoring_risk" = data_moderate_init_censoring_risk,
  "good_end_censoring_risk" = data_good_end_censoring_risk,
  "inv_end_censoring_risk" = data_inv_end_censoring_risk,
  "bad_end_censoring_risk" = data_bad_end_censoring_risk,
  "moderate_end_censoring_risk" = data_moderate_end_censoring_risk,
  "good_mid_censoring_risk" = data_good_mid_censoring_risk,
  "inv_mid_censoring_risk" = data_inv_mid_censoring_risk,
  "bad_mid_censoring_risk" = data_bad_mid_censoring_risk,
  "moderate_mid_censoring_risk" = data_moderate_mid_censoring_risk,
  "all_censoring_risk" = data_all_censoring_risk
)



# Test for risk-like predictions -----------------------------------------------
testthat::test_that("Concordance index for risk-like predictions is correct", {
  expected_score <- c(
    1.0, 0.0, 0.5, 0.8, 1.0, 0.0, 
    0.5, 2 / 3, 1.0, 0.0, 0.5, 0.8,
    1.0, 0.0, 0.5, 6 / 7, NA)
  expected_objective <- c(
    1.0, -1.0, 0.0, 0.6, 1.0, -1.0,
    0.0, 1 / 3, 1.0, -1.0, 0.0, 0.6,
    1.0, -1.0, 0.0, 5 / 7, NA)

  # Create metric object.
  metric_object <- familiar:::as_metric(
    metric = "concordance_index",
    outcome_type = "survival"
  )

  # Iterate over the data sets.
  for (ii in seq_along(data_list)) {
    # Check that the metric is available
    testthat::expect_true(familiar:::is_available(metric_object))

    # Compute the metric value.
    score <- familiar:::compute_metric_score(
      metric = metric_object,
      data = .as_risk(data_list[[ii]])
    )

    # Compute the objective score.
    objective_score <- familiar:::compute_objective_score(
      metric = metric_object,
      data = .as_risk(data_list[[ii]])
    )

    # Test the values.
    testthat::expect_equal(score, expected_score[ii])
    testthat::expect_equal(objective_score, expected_objective[ii])
  }
})

# Test for t-imelike predictions ----------------------------------------------
testthat::test_that("Concordance index for time-like predictions is correct", {
  expected_score <- c(
    0.0, 1.0, 0.5, 0.2, 0.0, 1.0,
    0.5, 1 / 3, 0.0, 1.0, 0.5, 0.2,
    0.0, 1.0, 0.5, 1 / 7, NA)
  expected_objective <- c(
    -1.0, 1.0, 0.0, -0.6, -1.0, 1.0, 
    0.0, -1 / 3, -1.0, 1.0, 0.0, -0.6,
    -1.0, 1.0, 0.0, -5 / 7, NA)

  # Create metric object.
  metric_object <- familiar:::as_metric(
    metric = "concordance_index",
    outcome_type = "survival"
  )

  # Iterate over the data sets.
  for (ii in seq_along(data_list)) {
    # Check that the metric is available
    testthat::expect_true(familiar:::is_available(metric_object))

    # Compute the metric value.
    score <- familiar:::compute_metric_score(
      metric = metric_object,
      data = .as_lifetime(data_list[[ii]])
    )

    # Compute the objective score.
    objective_score <- familiar:::compute_objective_score(
      metric = metric_object,
      data = .as_lifetime(data_list[[ii]])
    )

    # Test the values.
    testthat::expect_equal(score, expected_score[ii])
    testthat::expect_equal(objective_score, expected_objective[ii])
  }
})

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarTaskVimp -------------------------------------------------------------
setClass(
  "familiarTaskVimp",
  contains = "familiarTask",
  slots = list(
    "vimp_method" = "character",
    "hyperparameter_file" = "character",
    "feature_info_file" = "character"
  ),
  prototype = methods::prototype(
    vimp_method = NA_character_,
    hyperparameter_file = NA_character_,
    feature_info_file = NA_character_,
    task_name = "compute_variable_importance"
  )
)



# .set_file_name (vimp task) ---------------------------------------------------
setMethod(
  ".set_file_name",
  signature(object = "familiarTaskVimp"),
  function(object, file_paths = NULL) {
    if (is.null(file_paths)) return(object)
    
    # Generate file name of variable importance table
    object@file <- get_object_file_name(
      object_type = "vimpTable",
      data_id = object@data_id,
      run_id = object@run_id,
      vimp_method = object@vimp_method,
      project_id = object@project_id,
      dir_path = file_paths$vimp_dir
    )
    
    return(object)
  }
)



# .get_task_descriptor (vimp task) ---------------------------------------------
setMethod(
  ".get_task_descriptor",
  signature(object = "familiarTaskVimp"),
  function(object, ...) {
    return(paste0(object@task_name, "_", object@data_id, "_", object@run_id, "_", object@vimp_method))
  }
)



# .perform_task (vimp task , NULL) ---------------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskVimp",
    data = "NULL"
  ),
  function(
    object,
    data,
    experiment_data = NULL,
    outcome_info = NULL,
    ...
  ) {
    # This method is called when "data" is expected to be available somewhere in
    # the backend.
    
    if (is.null(experiment_data)) {
      ..error_reached_unreachable_code("experiment_data is required for retrieving data from the backend.")
    }
    if (is.null(outcome_info)) {
      ..error_reached_unreachable_code("outcome_info is required.")
    }
    
    # Find the run list.
    run_list <- .get_run_list(
      iteration_list = experiment_data@iteration_list,
      data_id = object@data_id,
      run_id = object@run_id
    )
    
    # Select unique samples.
    sample_identifiers <- .get_sample_identifiers(
      run = run_list,
      train_or_validate = "train"
    )
    sample_identifiers <- unique(sample_identifiers)
    
    # Create a dataObject.
    data <- methods::new(
      "dataObject",
      data = get_data_from_backend(sample_identifiers = sample_identifiers),
      preprocessing_level = "none",
      outcome_type = outcome_info@outcome_type,
      outcome_info = outcome_info
    )
    
    # Pass to method that dispatches with dataObject for further processing.
    return(.perform_task(
      object = object,
      data = data,
      experiment_data = experiment_data,
      ...
    ))
  }
)


# .perform_task (vimp task, dataObject) ----------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskVimp",
    data = "dataObject"
  ),
  function(
    object,
    data,
    vimp_aggregation_method = NULL,
    vimp_rank_threshold = NULL,
    experiment_data = NULL,
    settings = NULL,
    feature_info_list = NULL,
    hyperparameters = NULL,
    message_indent = 0L,
    verbose = FALSE,
    cl = NULL,
    return_results = TRUE,
    ...
  ) {

    # Check if the desired data already exist elsewhere.
    results_exist <- FALSE
    if (is(experiment_data, "experimentData")) {
      if (!is_empty(experiment_data@vimp_table_list)) {
        # Identify if the intended vimp_table already exists. Variable
        # importance tables are stored in a flat list in the vimp_table_list
        # attribute. The intended variable importance table must match the
        # variable importance method (vimp_method), data_id and run_id of the
        # current task.
        matching <- sapply(
          experiment_data@vimp_table_list,
          function(x, vimp_method, data_id, run_id) {
            if (x@vimp_method != vimp_method) return(FALSE)
            
            run_data <- tail(x@run_table, n = 1L)
            if (run_data$data_id != data_id) return(FALSE)
            if (run_data$run_id != run_id) return(FALSE)
            
            return(TRUE)
          },
          vimp_method = object@vimp_method,
          data_id = object@data_id,
          run_id = object@run_id,
          simplify = TRUE,
          USE.NAMES = FALSE
        )
        
        if (any(matching)) {
          vimp_table <- experiment_data@vimp_table_list[matching][[1L]]
          results_exist <- TRUE
        }
      }
    }
    
    if (file.exists(object@file)) {
      vimp_table <- update_object(object = readRDS(object@file))
      results_exist <- TRUE
    }
    
    if (results_exist) {
      if (!is.na(object@file)) {
        saveRDS(vimp_table, file = object@file)
      }
      
      if (return_results) {
        return(vimp_table)
      }
      
      return(invisible(TRUE))
    }
    
    # Check that outcome_info is present on data
    if (!is(data@outcome_info, "outcomeInfo")) {
      ..error_reached_unreachable_code(
        "outcome_info attribute of data (dataObject) does not contain an outcomeInfo object"
      )
    }
    
    # Set vimp aggregation method and vimp_rank_threshold based on settings.
    if (!is.null(settings)) {
      if (is.null(vimp_aggregation_method)) {
        vimp_aggregation_method <- settings$vimp$aggregation
      }
      if (is.null(vimp_rank_threshold)) {
        vimp_rank_threshold <- settings$vimp$aggr_rank_threshold
      }
    }
    
    # Check and retrieve feature info list.
    feature_info_list <- .get_feature_info_list(
      object = object,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings,
      message_indent = message_indent,
      verbose = verbose,
      cl = cl,
      ...
    )
    
    # Check and retrieve hyperparameters.
    hyperparameters <- .get_hyperparameters(
      object = object,
      hyperparameters = hyperparameters,
      feature_info_list = feature_info_list,
      vimp_aggregation_method = vimp_aggregation_method,
      vimp_rank_threshold = vimp_rank_threshold,
      data = data,
      settings = settings,
      message_indent = message_indent,
      verbose = verbose,
      cl = cl,
      ...
    )
    
    # Create the variable importance method object or familiar model object to
    # compute variable importance with.
    vimp_object <- methods::new(
      "familiarVimpMethod",
      outcome_type = data@outcome_type,
      hyperparameters = hyperparameters,
      vimp_method = object@vimp_method,
      vimp_aggregation_method = vimp_aggregation_method,
      vimp_rank_threshold = vimp_rank_threshold,
      outcome_info = data@outcome_info,
      run_table = .get_current_run_table(object = object),
      project_id = object@project_id
    )
    
    # Promote to the correct subclass.
    vimp_object <- promote_vimp_method(object = vimp_object)
    
    # Set multivariate methods.
    if (is(vimp_object, "familiarModel")) is_multivariate <- TRUE
    if (is(vimp_object, "familiarVimpMethod")) is_multivariate <- vimp_object@multivariate
    
    # Find required features. Exclude the signature features at this point, as
    # these will have been dropped from the variable importance table.
    required_features <- get_required_features(
      x = data,
      feature_info_list = feature_info_list,
      exclude_signature = !is_multivariate
    )
    
    # Limit to required features.
    vimp_object@required_features <- required_features
    vimp_object@feature_info <- feature_info_list[required_features]
    
    # Make sure the input data is processed.
    data <- process_input_data(
      object = vimp_object,
      data = data
    )
    
    logger_message(
      paste0(
        "Variable importance: Starting variable importance computation using the \"",
        object@vimp_method, "\" method for run ",
        object@task_id, " of ",
        object@n_tasks, "."
      ),
      indent = message_indent,
      verbose = verbose
    )
    
    # Compute variable importance.
    vimp_table <- .vimp(
      object = vimp_object, 
      data = data
    )
    
    # Set data id and run id.
    vimp_table@data_id <- object@data_id
    vimp_table@run_id <- object@run_id
    
    if (!is.na(object@file)) {
      saveRDS(vimp_table, file = object@file)
    }
    
    if (return_results) {
      return(vimp_table)
    }
    
    return(invisible(TRUE))
  }
)



# .get_hyperparameters (vimp task) ---------------------------------------------
setMethod(
  ".get_hyperparameters",
  signature(object = "familiarTaskVimp"),
  function(
    object,
    hyperparameters,
    file_paths = NULL,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    can_pre_process <- NULL
    
    if (is.null(hyperparameters) && !is.null(object@run_table)) {
      # This routine loads hyperparameters from disk, and is used when an
      # experiment is run using summon_familiar.
      
      # This check exists to make sure that the standard workflow passes the
      # correct objects.
      if (is.null(file_paths)) {
        ..error_reached_unreachable_code("file_paths was expected, but not provided.")
      }
      
      # Find the last entry on the run table that is marked as available for
      # pre-processing. This is what hyperparameters are based on.
      hyperparameter_run <- tail(
        object@run_table[[paste0(object@data_id, ".", object@run_id)]][can_pre_process == TRUE, ],
        n = 1L
      )
      
      # Find the file name.
      hyperparameter_file <- get_object_file_name(
        project_id = object@project_id,
        data_id = hyperparameter_run$data_id[1L],
        run_id = hyperparameter_run$run_id[1L],
        vimp_method = object@vimp_method,
        object_type = "hyperparametersVimp",
        dir_path = file_paths$vimp_dir
      )
      
      if (file.exists(hyperparameter_file)) {
        hyperparameter_object <- update_object(readRDS(hyperparameter_file))
        hyperparameters <- hyperparameter_object@hyperparameters
      }
    }
    
    
    if (is.null(hyperparameters) && is.na(object@hyperparameter_file)) {
      # Create an ad-hoc list of hyperparameters
      
      # Set up task, and explicitly don't write to file.
      hyperparameter_task <- methods::new(
        "familiarTaskVimpHyperparameters",
        project_id = object@project_id,
        vimp_method = object@vimp_method,
        file = NA_character_
      )
      
      # Execute the task.
      hyperparameter_object <- .perform_task(
        object = hyperparameter_task,
        ...
      )
      
      hyperparameters <- hyperparameter_object@hyperparameters
      
    } else if (is.null(hyperparameters)) {
      # Assume that the hyperparameter_file attribute contains the path to the
      # file containing the vimp method hyperparameters.
      if (!file.exists(object@hyperparameter_file)) {
        ..error(paste0("hyperparameter file does not exist at location: ", object@hyperparameter_file))
      }
      hyperparameter_object <- update_object(readRDS(object@hyperparameter_file))
      hyperparameters <- hyperparameter_object@hyperparameters
      
    } else if (is.character(hyperparameters)) {
      # If hyperparameters is a string, interpret this as a path to the
      # file containing the vimp method hyperparameters.
      if (!file.exists(hyperparameters)) {
        ..error(paste0("hyperparameter file does not exist at location: ", hyperparameters))
      }
      hyperparameter_object <- update_object(readRDS(hyperparameters))
      hyperparameters <- hyperparameter_object@hyperparameters
    }
    
    if (!(rlang::is_bare_list(hyperparameters))) {
      ..error("No hyperparameters were found.")
    }
    
    return(hyperparameters)
  }
)



# .get_variable_importance_table (vimp task) -----------------------------------
setMethod(
  ".get_variable_importance_table",
  signature(object = "familiarTaskVimp"),
  function(
    object,
    ...
  ) {
    ..error_reached_unreachable_code(".get_variable_importance_table does not exist for this task")
  }
)




.generate_vimp_tasks <- function(
    experiment_data,
    vimp_methods,
    file_paths,
    skip_existing = FALSE
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  vimp <- main_data_id <- can_pre_process <- NULL
  
  # Find the data_id related to computing variable importance.
  data_id <- experiment_data@experiment_setup[vimp == TRUE, ]$main_data_id[1L]
  if (is.na(data_id)) return(NULL)
  
  # Initialise empty list.
  task_list <- list()
  ii <- 1L
  run_tables <- .collect_run_tables(iteration_list = experiment_data@iteration_list)
  
  # vimp tasks -----------------------------------------------------------------
  
  # Get run ids.
  run_ids <- seq_len(experiment_data@experiment_setup[main_data_id == data_id]$n_runs[1L])
  
  # Set up variable importance computation task.
  for (vimp_method in vimp_methods) {
    for (run_id in run_ids) {
      
      # Create task to generate run-specific feature info.
      vimp_task <- methods::new(
        "familiarTaskVimp",
        data_id = data_id,
        run_id = run_id,
        vimp_method = vimp_method,
        run_table = run_tables,
        project_id = experiment_data@project_id
      )
      
      # Add file names.
      vimp_task <- .set_file_name(
        object = vimp_task,
        file_paths = file_paths
      )
      
      # Add to list, if the file does not exist on disk.
      if (!skip_existing || !.file_exists(vimp_task)) {
        task_list[[ii]] <- vimp_task
        ii <- ii + 1L
      }
    }
  }
  
  # Check if any vimp-related tasks are required.
  if (length(task_list) == 0L) return(NULL)
  
  # vimp hyperparameter tasks --------------------------------------------------
  
  # Identify which data id corresponds to computing hyperparameters.
  vimp_run_table <- .get_run_table_from_experiment_setup(
    data_id = data_id,
    experiment_setup = experiment_data@experiment_setup
  )
  
  vimp_hyperparameter_data_id <- tail(
    vimp_run_table[main_data_id <= data_id & can_pre_process == TRUE, ],
    n = 1L
  )$main_data_id[1L]
  
  # Get run ids.
  run_ids <- seq_len(vimp_run_table[main_data_id == vimp_hyperparameter_data_id, ]$n_runs[1L])
  
  for (vimp_method in vimp_methods) {
    for (run_id in run_ids) {
      # Create task to generate run-specific feature info.
      vimp_hyperparameter_task <- methods::new(
        "familiarTaskVimpHyperparameters",
        data_id = vimp_hyperparameter_data_id,
        run_id = run_id,
        vimp_method = vimp_method,
        run_table = run_tables,
        project_id = experiment_data@project_id
      )
      
      # Add file names.
      vimp_hyperparameter_task <- .set_file_name(
        object = vimp_hyperparameter_task,
        file_paths = file_paths
      )
      
      # Add to list, if the file does not exist on disk.
      if (!skip_existing || !.file_exists(vimp_hyperparameter_task)) {
        task_list[[ii]] <- vimp_hyperparameter_task
        ii <- ii + 1L
      }
    }
  }

  # Add tasks related to data processing for vimp methods.
  task_list <- c(
    task_list, 
    .generate_vimp_data_preprocessing_tasks(
      experiment_data = experiment_data,
      file_paths = file_paths
    )
  )
  
  return(task_list)
}



.run_variable_importance_computation <- function(
  cl,
  tasks,
  message_indent = 0L,
  verbose,
  ...
) {

  # Check that any tasks are available for processing.
  if (is_empty(tasks$hyperparameters_vimp) || is_empty(tasks$vimp)) return(invisible(FALSE))
  
  # Determine which variable importance hyperparameters need to be found.
  finished_tasks <- sapply(tasks$hyperparameters_vimp, .file_exists)
  unfinished_tasks <- tasks$hyperparameters_vimp[!finished_tasks]
  finished_tasks <- tasks$hyperparameters_vimp[finished_tasks]
  
  # Process any unfinished tasks.
  if (length(unfinished_tasks) > 0L) {
    ..run_variable_importance_computation_hyperparameters(
      cl = cl,
      tasks = unfinished_tasks,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
  }
  
  # Determine which variable importance tasks are required.
  finished_tasks <- sapply(tasks$vimp, .file_exists)
  unfinished_tasks <- tasks$vimp[!finished_tasks]
  finished_tasks <- tasks$vimp[finished_tasks]
  
  # Process any unfinished tasks.
  if (length(unfinished_tasks) > 0L) {
    ..run_variable_importance_computation(
      cl = cl,
      tasks = unfinished_tasks,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
  }
  
  return(invisible(TRUE))
}



..run_variable_importance_computation <- function(
    tasks,
    cl,
    message_indent = 0L,
    verbose,
    ...
) {
  
  # Message that variable importances computation is starting.
  logger_message(
    paste0(
      "Variable importance: Starting variable importance computation."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  fam_mapply_lb(
    cl = cl,
    assign = "all",
    FUN = .perform_task,
    progress_bar = FALSE,
    object = tasks,
    MoreArgs = list(
      "data" = NULL,
      "return_results" = FALSE,
      "message_indent" = message_indent + 1L,
      "verbose" = verbose,
      ...
    )
  )
  
  # Message that variable importances have been computed.
  logger_message(
    paste0(
      "Variable importance: Variable importances have been computed.\n"
    ),
    indent = message_indent,
    verbose = verbose
  )
}

library(testthat)
library(familiar)

on_cran <- function() {
  !interactive() && identical(Sys.getenv("NOT_CRAN"), "")
}

# Prevent thread overuse (through data.table?) when running tests on CRAN.
if (on_cran()) {
  Sys.setenv("OMP_THREAD_LIMIT" = 2L)
}

suppressWarnings(
  testthat::test_check("familiar"),
  classes = c("deprecation_warning")
)

testthat::skip("Testing object updates should be done manually.")

# List files in tests/old_experiments.
archive_files <- list.files(
  path = "./tests/old_experiments",
  pattern = ".zip",
  full.names = TRUE)

# Set path to the directory where the archive should be unpacked.
archive_directory <- file.path(".", "tests", "old_experiments", "_experiment")

test_generate_experiment_parameters <- coro::generator(function(outcome_type) {
  for (current_outcome_type in outcome_type) {
    # Set learner
    learner <- switch(
      current_outcome_type,
      "binomial" = c("glm_logistic", "lasso"),
      "multinomial" = c("glm", "lasso"),
      "count" = c("glm", "lasso"),
      "continuous" = c("glm_gaussian", "lasso"),
      "survival" = c("cox", "survival_regr_weibull"))

    coro::yield(list(
      "experimental_design" = "bs(fs+mb,3)",
      "outcome_type" = current_outcome_type,
      "vimp_method" = c("mim", "concordance"),
      "learner" = learner))
  }
})

# Iterate over archive files.
for (archive_file in archive_files) {
  # Determine outcome type by stripping the extension, underscore and version.
  outcome_type <- gsub(
    x = basename(archive_file),
    pattern = "[[:digit:]]+|[.]zip$|[_]",
    replacement = "")

  # Determine the version by stripping the extension and all alpha-characters.
  familiar_version <- gsub(
    x = basename(archive_file),
    pattern = "[[:alpha:]]+|[.]zip$",
    replacement = "")

  # Strip final underscore.
  familiar_version <- gsub(
    x = familiar_version,
    pattern = "[_]$",
    replacement = "")

  # Replace underscores by period and convert to package version object.
  familiar_version <- as.package_version(gsub(
    x = familiar_version,
    pattern = "[_]",
    replacement = "."))

  # Avoid running tests if the archived version for some reason exceeds the
  # installed familiar version.
  if (utils::packageVersion("familiar") < familiar_version) {
    warning(
      "The installed version of familiar (",
      utils::packageVersion("familiar"),
      ") is older than the archive being tested (",
      familiar_version, ").")

    next
  }

  # Unzip directory.
  utils::unzip(archive_file,
    exdir = archive_directory,
    overwrite = TRUE)

  # Yield current set of parameters.
  config_parameters <- test_generate_experiment_parameters(outcome_type = outcome_type)()

  # Recreate data.
  data <- familiar:::test_create_good_data_random_missing(outcome_type = outcome_type)

  # Test 1: Running the experiment again ---------------------------------------

  message(paste0(
    "-------------------------------------------------------------\n",
    outcome_type, " (", familiar_version, "): ",
    "test direct export\n",
    "-------------------------------------------------------------\n"
  ))

  # Remove the results directory.
  unlink(
    file.path(archive_directory, "results"),
    recursive = TRUE)

  # Run the experiment to extract the data again.
  do.call(
    familiar::summon_familiar,
    args = c(
      list(
        "data" = data,
        "experiment_dir" = archive_directory,
        "experiment_data" = list.files(
          path = archive_directory,
          full.names = TRUE,
          pattern = "iterations.RDS"),
        "parallel" = FALSE),
      config_parameters))

  # Test 2: Running the experiment again, starting with familiar data objects ------

  message(paste0(
    "-------------------------------------------------------------\n",
    outcome_type, " (", familiar_version, "): ",
    "test export after collecting familiarData objects\n",
    "-------------------------------------------------------------\n"
  ))

  # Remove the results directory.
  unlink(
    file.path(archive_directory, "results"),
    recursive = TRUE)

  # Remove the familiar_collections directory.
  unlink(
    file.path(archive_directory, "familiar_collections"),
    recursive = TRUE)

  # Run the experiment to extract the data again.
  do.call(
    familiar::summon_familiar,
    args = c(
      list(
        "data" = data,
        "experiment_dir" = archive_directory,
        "experiment_data" = list.files(
          path = archive_directory,
          full.names = TRUE,
          pattern = "iterations.RDS"),
        "parallel" = FALSE),
      config_parameters))

  # Test 3: Running the experiment again, starting with ensembles --------------

  if (familiar_version > "1.1.4") {
    # The isolation forests within the historic version 1.1.4 are not backward
    # compatible, and we need to skip these tests for version 1.1.4.
    # Remove temporary directory.
    unlink(
      file.path(archive_directory),
      recursive = TRUE
    )
    next
  }
  
  message(paste0(
    "-------------------------------------------------------------\n",
    outcome_type, " (", familiar_version, "): ",
    "test export after creating familiarData objects\n",
    "-------------------------------------------------------------\n"
  ))

  # Remove temporary directory.
  unlink(
    file.path(archive_directory),
    recursive = TRUE)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# plot_all (generic) -----------------------------------------------------------
setGeneric(
  "plot_all",
  function(object, ...) standardGeneric("plot_all")
)



# plot_all (collection) --------------------------------------------------------
setMethod(
  "plot_all",
  signature(object = "familiarCollection"),
  function(object, dir_path = NULL, ...) {
    
    if (!is.null(dir_path)) dir_path <- encapsulate_path(dir_path)
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Feature univariate p-values (horizontal bars)
    do.call(
      plot_univariate_importance,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Feature occurrence (unclustered)
    do.call(
      plot_feature_selection_occurrence,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Feature ranking (unclustered)
    do.call(
      plot_feature_selection_variable_importance,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Model signature occurrence (unclustered)
    do.call(
      plot_model_signature_occurrence,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Model signature ranking (unclustered)
    do.call(
      plot_model_signature_variable_importance,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Permutation variable importance
    do.call(
      plot_permutation_variable_importance,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Feature similarity heatmap
    do.call(
      plot_feature_similarity,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Calibration curves
    do.call(
      plot_calibration_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Model performance
    do.call(
      plot_model_performance,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # AUC-ROC curve
    do.call(
      plot_auc_roc_curve,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # AUC-PR curve
    do.call(
      plot_auc_precision_recall_curve,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Decision curve
    do.call(
      plot_decision_curve,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Kaplan-Meier curves
    do.call(
      plot_kaplan_meier,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Feature expressions
    do.call(
      plot_sample_clustering,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Confusion matrix
    do.call(
      plot_confusion_matrix,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Individual conditional expectation
    do.call(
      plot_ice,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # Partial dependence
    do.call(
      plot_pd,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # SHAP summary.
    do.call(
      plot_shap_summary,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # SHAP waterfall.
    do.call(
      plot_shap_waterfall,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # SHAP force.
    do.call(
      plot_shap_force,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    # SHAP dependence.
    do.call(
      plot_shap_dependence,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    )
    
    return(invisible(NULL))
  }
)

# plot_all (general) -----------------------------------------------------------
setMethod(
  "plot_all",
  signature(object = "ANY"),
  function(object, dir_path = NULL, ...) {
    # Attempt conversion to familiarCollection object.
    
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "all"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_all,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

setClass(
  "familiarConcordanceVimp",
  contains = "familiarVimpMethod"
)


.get_available_concordance_vimp_method <- function(show_general = TRUE) {
  return("concordance")
}


# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarConcordanceVimp"),
  function(object, ...) {
    
    if (object@outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(TRUE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarConcordanceVimp"),
  function(object, data = NULL, ...) {
    return(list())
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarConcordanceVimp"),
  function(
    object,
    data,
    cl = NULL,
    ...
  ) {
    
    if (is_empty(data)) return(callNextMethod())
    
    # Use concordance-based measures for variable importance:
    # - Gini index for binomial and multinomial outcomes
    # - Kendall's Tau for continuous and counts outcomes
    # - Concordance index for survival features
    
    if (object@outcome_type %in% c("binomial", "multinomial")) {
      # Compute gini index for categorical outcomes.
      
      # Use effect coding to convert categorical data into encoded data - this
      # is required to deal with factors with missing/new levels between
      # training and test data sets.
      encoded_data <- encode_categorical_variables(
        data = data,
        object = object,
        encoding_method = "dummy",
        drop_levels = FALSE
      )
      
      # Find feature columns in the data.
      feature_columns <- get_feature_columns(x = encoded_data$encoded_data)
      
      # Compute auc-roc
      auc_roc <- fam_sapply(
        cl = cl,
        X = encoded_data$encoded_data@data[, mget(feature_columns)],
        FUN = .compute_auc_roc,
        y_obs = encoded_data$encoded_data@data$outcome,
        chopchop = TRUE
      )
      
      # Map to [0, 1] range.
      score <- abs(auc_roc - 0.5) * 2.0
      
      # Create variable importance object.
      vimp_object <- methods::new(
        "vimpTable",
        vimp_table = data.table::data.table(
          "score" = score,
          "name" = feature_columns
        ),
        encoding_table = encoded_data$reference_table,
        score_aggregation = "max",
        invert = TRUE
      )
      
      return(vimp_object)
      
    } else if (object@outcome_type %in% c("continuous")) {
      # For continuous outcomes use kendall's tau.

      # Create a new vimp object, and replace the vimp_method slot.
      new_vimp_object <- methods::new("familiarCorrelationVimp", object)
      new_vimp_object@vimp_method <- "kendall"

      return(..vimp(
        object = new_vimp_object,
        data = data,
        cl = cl
      ))
      
    } else if (object@outcome_type == "survival") {
      # Compute the concordance index for each feature.

      # Use effect coding to convert categorical data into encoded data - this
      # is required to deal with factors with missing/new levels between
      # training and test data sets.
      encoded_data <- encode_categorical_variables(
        data = data,
        object = object,
        encoding_method = "dummy",
        drop_levels = FALSE
      )

      # Find feature columns in the data.
      feature_columns <- get_feature_columns(x = encoded_data$encoded_data)

      # Compute concordance indices
      c_index <- fam_sapply(
        X = encoded_data$encoded_data@data[, mget(feature_columns)],
        FUN = ..compute_concordance_index,
        time = encoded_data$encoded_data@data$outcome_time,
        event = encoded_data$encoded_data@data$outcome_event,
        chopchop = TRUE
      )

      # Map to [0, 1] range.
      score <- abs(c_index - 0.5) * 2.0
      
      # Create variable importance object.
      vimp_object <- methods::new("vimpTable",
        vimp_table = data.table::data.table(
          "score" = score,
          "name" = feature_columns
        ),
        encoding_table = encoded_data$reference_table,
        score_aggregation = "max",
        invert = TRUE
      )

      return(vimp_object)
      
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }
  }
)

.determine_preprocessing_parameters <- function(
    cl = NULL,
    data,
    feature_info_list,
    settings,
    message_indent = 0L,
    verbose = FALSE
) {
  
  if (!is(data, "dataObject")) {
    ..error_reached_unreachable_code(
      ".determine_preprocessing_parameters: data is not a dataObject."
    )
  }
  
  if (is_empty(data)) ..error_data_set_is_empty()
  if (!has_feature_data(data)) ..error_data_set_has_no_features()
  
  # Remove samples with missing outcome data -----------------------------------
  n_samples_current <- get_n_samples(data)
  
  logger_message(
    paste0("Pre-processing: ", n_samples_current, " samples were initially available."),
    indent = message_indent,
    verbose = verbose
  )
  
  # Remove all samples with missing outcome data
  data <- filter_missing_outcome(data = data, is_validation = FALSE)
  if (is_empty(data)) ..error_data_set_has_no_outcome()
  
  n_samples_remain <- get_n_samples(data)
  n_samples_removed <- n_samples_current - n_samples_remain
  
  logger_message(
    paste0(
      "Pre-processing: ", n_samples_removed,
      " samples were removed because of missing outcome data. ",
      n_samples_remain, " samples remain."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  
  #Remove features with a large fraction of missing values ---------------------
  n_features_current <- get_n_features(data)
  logger_message(
    paste0("Pre-processing: ", n_features_current, " features were initially available."),
    indent = message_indent,
    verbose = verbose
  )
  
  # Determine the fraction of missing values
  feature_info_list <- add_missing_value_fractions(
    cl = cl,
    feature_info_list = feature_info_list,
    data = data,
    threshold = settings$prep$feature_max_fraction_missing
  )
  
  # Find features that are not missing too many values.
  available_features <- get_available_features(feature_info_list = feature_info_list)
  
  # Remove features with a high fraction of missing values
  data <- filter_features(data = data, available_features = available_features)
  
  if (!has_feature_data(data)) {
    ..error(
      paste0(
        "The provided dataset lacks features with sufficient available values. ",
        "Please investigate missing values in the dataset or increase the missingness ",
        "threshold by increasing the feature_max_fraction_missing configuration parameter."
      ),
      error_class = "dataset_error"
    )
  } 
  
  # Message how many features were removed
  logger_message(
    paste0(
      "Pre-processing: ", n_features_current - length(available_features),
      " features were removed because of a high fraction of missing values. ",
      length(available_features), " features remain."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  n_samples_current <- n_samples_remain
  
  
  # Remove samples with a large fraction of missing values ---------------------
  
  data <- filter_bad_samples(
    data = data,
    threshold = settings$prep$sample_max_fraction_missing
  )
  
  if (is_empty(data)) {
    ..error(
      paste0(
        "The provided dataset lacks samples with sufficient available feature values. ",
        "Please investigate missing values in the dataset or increase the missingness ",
        "threshold by increasing the sample_max_fraction_missing configuration parameter."
      ),
      error_class = "dataset_error"
    )
  }
  
  # Message how many samples were removed
  n_samples_remain <- get_n_samples(data)
  
  logger_message(
    paste0(
      "Pre-processing: ", n_samples_current - n_samples_remain,
      " samples were removed because of missing feature data. ",
      n_samples_remain, " samples remain."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  n_features_current <- length(available_features)
  
  
  # Remove invariant features --------------------------------------------------
  
  # Filter features that are invariant.
  feature_info_list  <- find_invariant_features(
    cl = cl,
    feature_info_list = feature_info_list,
    data = data
  )
  
  # Find available features.
  available_features <- get_available_features(feature_info_list = feature_info_list)
  
  # Remove invariant features from the data
  data <- filter_features(data = data, available_features = available_features)
  
  if (!has_feature_data(data)) {
    ..error(
      paste0(
        "Remaining features in the dataset only have a single value for all samples ",
        "and cannot be used for training."
      ),
      error_class = "dataset_error"
    )
  }
  
  # Message number of features removed by the no-variance filter.
  logger_message(
    paste0(
      "Pre-processing: ", n_features_current - length(available_features),
      " features were removed due to invariance. ",
      length(available_features), " features remain."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  
  # Add feature distribution data ----------------------------------------------
  logger_message(
    paste0("Pre-processing: Adding value distribution statistics to features."),
    indent = message_indent,
    verbose = verbose
  )
    
  # Add feature distribution data
  feature_info_list <- compute_feature_distribution_data(
    cl = cl,
    feature_info_list = feature_info_list,
    data = data
  )
  
  
  # Transform features ---------------------------------------------------------
  logger_message(
    "Pre-processing: Performing transformations to normalise feature value distributions.",
    indent = message_indent,
    verbose = verbose && settings$prep$transform_method != "none"
  )
  
  # Add skeletons to the feature information list.
  feature_info_list <- create_transformation_parameter_skeleton(
    feature_info_list = feature_info_list,
    transformation_method = settings$prep$transform_method,
    transformation_optimisation_criterion = settings$prep$transformation_optimisation_criterion,
    transformation_gof_p_value = settings$prep$transformation_gof_test_p_value
  )
  
  # Add transformation parameters to the feature information list
  feature_info_list <- add_transformation_parameters(
    cl = cl,
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose
  )
  
  # Apply transformation.
  data <- transform_features(
    data = data,
    feature_info_list = feature_info_list
  )
  
  logger_message(
    "Pre-processing: Feature distributions have been transformed for normalisation.",
    indent = message_indent,
    verbose = verbose & settings$prep$transform_method != "none"
  )
  

  # Remove low-variance features -----------------------------------------------
  if ("low_variance" %in% settings$prep$filter_method) {
    n_features_current <- length(available_features)
    
    # Filter features that are invariant.
    feature_info_list <- find_low_variance_features(
      cl = cl,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings
    )
    
    # Check available features.
    available_features <- get_available_features(feature_info_list = feature_info_list)
    
    # Remove invariant features from the data
    data <- filter_features(data = data, available_features = available_features)
    
    if (!has_feature_data(data)) {
      ..error(
        paste0(
          "Remaining features in the dataset have a variance that is lower than the threshold ",
          "and were therefore all removed. Please investigate your data, or increase the threshold ",
          "through the low_var_minimum_variance_threshold configuration parameter."
        ),
        error_class = "dataset_error"
      )
    } 
    
    # Message number of features removed by the low-variance filter.
    logger_message(
      paste0(
        "Pre-processing: ", n_features_current - length(available_features),
        " features were removed due to low variance. ",
        length(available_features), " features remain."
      ),
      indent = message_indent,
      verbose = verbose
    )
  }
  
  
  
  # Normalise features ---------------------------------------------------------
  logger_message(
    "Pre-processing: Extracting normalisation parameters from feature data.",
    indent = message_indent,
    verbose = verbose && settings$prep$normalisation_method != "none"
  )
  
  
  # Add skeletons to the feature information list.
  feature_info_list <- create_normalisation_parameter_skeleton(
    feature_info_list = feature_info_list,
    normalisation_method = settings$prep$normalisation_method
  )
  
  # Add normalisation parameters to the feature information list.
  feature_info_list <- add_normalisation_parameters(
    cl = cl,
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose
  )

  # Apply normalisation to data before clustering
  data <- normalise_features(
    data = data,
    feature_info_list = feature_info_list
  )
  
  logger_message(
    "Pre-processing: Feature data were normalised.",
    indent = message_indent,
    verbose = verbose && settings$prep$normalisation_method != "none"
  )
  
  
  # Batch normalise features ---------------------------------------------------
  logger_message(
    "Pre-processing: Extracting batch normalisation parameters from feature data.",
    indent = message_indent,
    verbose = verbose && settings$prep$batch_normalisation_method != "none"
  )
  
  # Check that assumptions for batch normalisation are fulfilled.
  .check_batch_normalisation_assumptions(
    data = data,
    normalisation_method = settings$prep$batch_normalisation_method
  )
  
  # Add batch normalisation skeletons.
  feature_info_list <- create_batch_normalisation_parameter_skeleton(
    feature_info_list = feature_info_list,
    normalisation_method = settings$prep$batch_normalisation_method
  )
  
  # Add batch normalisation parameters to the feature information list.
  feature_info_list <- add_batch_normalisation_parameters(
    cl = cl, 
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose
  )
  
  # Batch-normalise feature values
  data <- batch_normalise_features(
    data = data,
    feature_info_list = feature_info_list
  )
  
  logger_message(
    "Pre-processing: Feature data were batch-normalised.",
    indent = message_indent,
    verbose = verbose && settings$prep$batch_normalisation_method != "none"
  )
  
  
  # Remove non-robust features -------------------------------------------------
  if ("robustness" %in% settings$prep$filter_method) {
    n_features_current <- length(available_features)
    
    # Filter features that are not robust
    feature_info_list  <- find_non_robust_features(
      cl = cl,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings
    )
    
    available_features <- get_available_features(feature_info_list = feature_info_list)
    
    # Remove non-robust features from the data
    data <- filter_features(data = data, available_features = available_features)
    
    if (!has_feature_data(data)) {
      ..error(
        paste0(
          "Remaining features in the dataset have a robustness that is lower than the threshold ",
          "and were therefore all removed. Please investigate your data, or decrease the threshold ",
          "through the robustness_threshold_value configuration parameter."
        ),
        error_class = "dataset_error"
      )
    }
    
    # Message number of features removed by the robustness filter.
    logger_message(
      paste0(
        "Pre-processing: ", n_features_current - length(available_features),
        " features were removed due to low robustness. ",
        length(available_features), " features remain."
      ),
      indent = message_indent,
      verbose = verbose
    )
  }
  
  
  # Remove unimportant features ------------------------------------------------
  if ("univariate_test" %in% settings$prep$filter_method) {
    n_features_current <- length(available_features)
    
    # Filter features that are not relevant
    feature_info_list  <- find_unimportant_features(
      cl = cl,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings
    )
    
    available_features <- get_available_features(feature_info_list = feature_info_list)
    
    # Remove unimportant features from the data
    data <- filter_features(data = data, available_features = available_features)
    
    if (!has_feature_data(data)) {
      ..error(
        paste0(
          "Remaining features in the dataset have a p-value that is higher than the threshold ",
          "and were therefore all removed. Please investigate your data, or increase the threshold ",
          "through the univariate_test_threshold configuration parameter."
        ),
        error_class = "dataset_error"
      )
    }
    
    # Message number of features removed by the importance filter.
    logger_message(
      paste0(
        "Pre-processing: ", n_features_current - length(available_features),
        " features were removed due to low importance. ",
        length(available_features), " features remain."
      ),
      indent = message_indent,
      verbose = verbose
    )
  }
  
  
  # Impute missing values ------------------------------------------------------
  logger_message(
    "Pre-processing: Adding imputation information to features.",
    indent = message_indent,
    verbose = verbose
  )
  
  # Add imputation skeletons.
  feature_info_list <- create_imputation_parameter_skeleton(
    feature_info_list = feature_info_list,
    imputation_method = settings$prep$imputation_method
  )
  
  # Add imputation info
  feature_info_list <- add_imputation_info(
    cl = cl,
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose
  )
  
  # Impute features with censored data prior to clustering
  data <- impute_features(
    data = data,
    feature_info_list = feature_info_list
  )
  
  
  # Cluster features -----------------------------------------------------------
  logger_message(
    "Pre-processing: Starting clustering of redundant features",
    indent = message_indent,
    verbose = verbose && settings$prep$cluster_method != "none"
  )
  
  # Determine the number of features prior to clustering.
  n_features_current <- length(get_available_features(feature_info_list = feature_info_list))
  
  # Add clustering skeletons.
  feature_info_list <- create_cluster_parameter_skeleton(
    feature_info_list = feature_info_list,
    cluster_method = settings$prep$cluster_method,
    cluster_linkage = settings$prep$cluster_linkage,
    cluster_cut_method = settings$prep$cluster_cut_method,
    cluster_similarity_threshold = settings$prep$cluster_similarity_threshold,
    cluster_similarity_metric = settings$prep$cluster_similarity_metric,
    cluster_representation_method = settings$prep$cluster_representation_method
  )
  
  # Extract clustering information
  feature_info_list  <- add_cluster_info(
    cl = cl,
    feature_info_list = feature_info_list,
    data = data,
    message_indent = message_indent + 1L,
    verbose = verbose
  )
  
  # Build cluster table.
  cluster_table <- .create_clustering_table(feature_info_list = feature_info_list)
  
  # Further summarise the clusters by grouping.
  cluster_table <- cluster_table[, list("cluster_size" = .N), by = "cluster_name"]
  
  logger_message(
    paste0(
      nrow(cluster_table),
      ifelse(nrow(cluster_table) == 1L, " feature cluster was", " feature clusters were"),
      " created from ", n_features_current,
      ifelse(n_features_current == 1L, " feature. ", " features. "),
      sum(cluster_table$cluster_size > 1L),
      ifelse(sum(cluster_table$cluster_size > 1L) == 1L, " cluster contains", " clusters contain"),
      " more than one feature. The remaining ",
      sum(cluster_table$cluster_size == 1L),
      ifelse(sum(cluster_table$cluster_size == 1L) == 1L, " cluster is", " clusters are"),
      " singular."
    ),
    indent = message_indent + 1L,
    verbose = verbose && settings$prep$cluster_method != "none"
  )
  
  # Add required features
  feature_info_list <- add_required_features(feature_info_list = feature_info_list)
  
  # Filter features that are not required from the list
  feature_info_list <- trim_unused_features_from_list(feature_info_list = feature_info_list)
  
  # Return list of featureInfo objects
  return(feature_info_list)
}



combine_feature_info_list <- function(
    preferred = NULL,
    custom = NULL,
    generic = NULL
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  name <- present <- list_name <- complete <- NULL
  
  # Get all features.
  feature_names <- unique(c(
    names(preferred),
    names(custom),
    names(generic)
  ))
  
  # Identify which features appear where.
  data <- mapply(
    FUN = function(x, x_name, feature_names) {
      
      # Default dataset.
      data <- data.table::data.table(
        "list_name" = x_name,
        "name" = feature_names,
        "present" = FALSE,
        "complete" = FALSE
      )
      
      # Mark feature names that are present in the current dataset.
      data[name %in% names(x), "present" := TRUE]
      
      # Check whether data are present, are complete.
      data[
        present == TRUE,
        "complete" := feature_info_complete(object = x[[name]]),
        by = "name"
      ]
      
      return(data)
    },
    x = list(
      "preferred" = preferred,
      "custom" = custom,
      "generic" = generic
    ),
    x_name = c("preferred", "custom", "generic"),
    MoreArgs = list("feature_names" = feature_names),
    SIMPLIFY = FALSE
  )
  
  # Combine list.
  data <- data.table::rbindlist(data)
  
  # Start new feature list.
  new_feature_list <- NULL
  
  # Preference to get features with complete information first from the
  # preferred set.
  selected_feature_names <- data[list_name == "preferred" & complete == TRUE]$name
  
  # Add to feature list and remove from set of features.
  if (length(selected_feature_names) > 0L) {
    new_feature_list <- c(new_feature_list, preferred[selected_feature_names])
    feature_names <- setdiff(feature_names, selected_feature_names)
  }
  
  # Then, preference to get complete information from the custom set.
  selected_feature_names <- data[name %in% feature_names & list_name == "custom" & complete == TRUE, ]$name
  
  # Add to feature list and remove from set of features.
  if (length(selected_feature_names) > 0L) {
    new_feature_list <- c(new_feature_list, custom[selected_feature_names])
    feature_names <- setdiff(feature_names, selected_feature_names)
  }
  
  # Then, preference to get incomplete information first from the preferred set.
  selected_feature_names <- data[name %in% feature_names & list_name == "preferred" & present == TRUE, ]$name
  
  # Add to feature list and remove from set of features.
  if (length(selected_feature_names) > 0L) {
    new_feature_list <- c(new_feature_list, preferred[selected_feature_names])
    feature_names <- setdiff(feature_names, selected_feature_names)
  }
  
  # Then, preference to get incomplete information first from the custom set.
  selected_feature_names <- data[name %in% feature_names & list_name == "custom" & present == TRUE, ]$name
  
  # Add to feature list and remove from set of features.
  if (length(selected_feature_names) > 0L) {
    new_feature_list <- c(new_feature_list, custom[selected_feature_names])
    feature_names <- setdiff(feature_names, selected_feature_names)
  }
  
  # Then, preference to get incomplete information first from the generic set.
  selected_feature_names <- data[name %in% feature_names & list_name == "generic" & present == TRUE, ]$name
  
  # Add to feature list and remove from set of features.
  if (length(selected_feature_names) > 0L) {
    new_feature_list <- c(new_feature_list, generic[selected_feature_names])
    feature_names <- setdiff(feature_names, selected_feature_names)
  }
  
  return(new_feature_list)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


load_experiment_data <- function(x, file_paths) {
  # This function restores the content of the experimentData object to file
  # system - basically allowing for a reproducible hot start.
  
  # Attempt to load from file.
  project_id <- NULL
  if (is.character(x)) {
    project_id <- gsub(
      x = basename(x),
      pattern = "[[:alpha:]]|[.]RDS$|[_]",
      replacement = ""
    )
    
    # Read from file system.
    x <- readRDS(x)
  } 
  
  # Users may have added a configuration
  if (is.list(x)) {
    if (all(c("iteration_list", "experiment_setup") %in% names(x))) {
      x <- methods::new(
        "experimentData",
        iteration_list = x$iteration_list,
        experiment_setup = x$experiment_setup,
        project_id = project_id
      )
    }
  }
  
  # Expect that the file is an experimentData object.
  if (!is(x, "experimentData")) {
    ..error(
      paste0(
        "An experimentData object was expected. Found: a ",
        paste_s(class(x)), " object."
      ),
      error_class = "input_argument_error"
    )
  }
  
  # Update the experimentData object.
  x <- update_object(x)
  
  # Start writing the contents of the object to the working directory to deploy
  # from there.
  if (!is.null(x@experiment_setup) && !is.null(x@iteration_list)) {
    
    # Set file name
    file_name <- .get_iteration_file_name(
      file_paths = file_paths,
      project_id = x@project_id
    )
    
    # Check if the directory exists, and create it otherwise.
    if (!dir.exists(file_paths$iterations_dir)) {
      dir.create(file_paths$iterations_dir, recursive = TRUE)
    } 
    
    # Save both files to the expected location.
    saveRDS(
      list(
        "iteration_list" = x@iteration_list,
        "experiment_setup" = x@experiment_setup
      ),
      file = file_name
    )
  }
  
  # Start writing feature information.
  if (!is.null(x@feature_info)) {
    
    for (feature_info_name in names(x@feature_info)) {
      feature_info <- x@feature_info[[feature_info_name]]
      
      # Set file name.
      if (feature_info_name == "generic") {
        file_name <- get_object_file_name(
          object_type = "genericFeatureInfo",
          project_id = feature_info[[1L]]@project_id,
          dir_path = file_paths$process_data_dir
        )
        
      } else {
        file_name <- get_object_file_name(
          object_type = "featureInfo",
          project_id = feature_info[[1L]]@project_id,
          data_id = feature_info[[1L]]@data_id,
          run_id = feature_info[[1L]]@run_id,
          dir_path = file_paths$process_data_dir
        )
      }
      
      # Check if the directory exists, and create it otherwise.
      if (!dir.exists(file_paths$process_data_dir)) {
        dir.create(file_paths$process_data_dir, recursive = TRUE)
      } 
    
      # Write to file.
      saveRDS(feature_info, file = file_name)
    }
  }
  
  # Write variable importance information.
  if (!is.null(x@vimp_table_list)) {
    for (vimp_table in x@vimp_table_list) {
      
      # Set file name
      file_name <- get_object_file_name(
        object_type = "vimpTable",
        data_id = vimp_table@data_id,
        run_id = vimp_table@run_id,
        vimp_method = vimp_table@vimp_method,
        project_id = vimp_table@project_id,
        dir_path = file_paths$vimp_dir
      )
      
      # Check if the directory exists, and create it otherwise.
      if (!dir.exists(dirname(file_name))) {
        dir.create(dirname(file_name), recursive = TRUE)
      } 
      
      # Write to file.
      saveRDS(vimp_table, file = file_name)
    }
  }
  
  return(x)
}



set_experiment_data <- function(
    x = NULL,
    project_id,
    experiment_setup = NULL,
    iteration_list = NULL,
    feature_info = NULL,
    vimp_hyperparameter_list = NULL,
    vimp_table_list = NULL
) {

  # Create new object.
  if (!is(x, "experimentData")) {
    x <- methods::new(
      "experimentData",
      project_id = project_id
    )
    
    # Add package version
    x <- add_package_version(x)
  }
  
  if (is.null(x@experiment_setup) && !is.null(experiment_setup)) x@experiment_setup <- experiment_setup
  if (is.null(x@iteration_list) && !is.null(iteration_list)) x@iteration_list <- iteration_list
  if (is.null(x@feature_info) && !is.null(feature_info)) x@feature_info <- feature_info
  if (!is.null(vimp_hyperparameter_list)) x@vimp_hyperparameter_list <- vimp_hyperparameter_list
  if (!is.null(vimp_table_list)) x@vimp_table_list <- vimp_table_list
  
  return(x)
}



# show (experimentData) --------------------------------------------------------
setMethod(
  "show",
  signature(object = "experimentData"),
  function(object) {
    
    # Make sure the model object is updated.
    object <- update_object(object = object)
    
    # Experiment data is always present.
    content_str <- c("experiment data")
    
    # Check if feature info is present.
    if (!is.null(object@feature_info)) {
      if (length(object@feature_info) > 1L) {
        content_str <- c(
          content_str,
          "basic and extended feature information"
        )
        
      } else {
        content_str <- c(
          content_str,
          "basic feature information"
        )
      }
    }
    
    # Check if variable importance information is present.
    if (!is.null(object@vimp_table_list)) {
      content_str <- c(
        content_str,
        paste0("variable importance (", paste_s(names(object@vimp_table_list)), ")")
      )
    }
    
    cat(paste0(
      "Experiment data object (", .familiar_version_string(object), ") with project id ",
      object@project_id, " containing ", paste_s(content_str), ".\n"
    ))
  }
)


  
# add_package_version (experiment data) ----------------------------------------
setMethod(
  "add_package_version",
  signature(object = "experimentData"),
  function(object) {
    
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarDataElementVimpData object -------------------------------------------
setClass(
  "familiarDataElementVimpData",
  contains = "familiarDataElement",
  slots = list(
    "rank_threshold" = "numeric",
    "rank_aggregation_method" = "character"
  ),
  prototype = methods::prototype(
    detail_level = "hybrid",
    estimation_type = "point",
    rank_threshold = NA_real_,
    rank_aggregation_method = NA_character_
  )
)



# extract_model_vimp (generic) -------------------------------------------------

#'@title Internal function to extract variable importance from models.
#'
#'@description Aggregate variable importance from models in a
#'  `familiarEnsemble`.
#'
#'@inheritParams .extract_data
#'
#'@return A list containing variable importance information.
#'@md
#'@keywords internal
setGeneric(
  "extract_model_vimp",
  function(
    object,
    data,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_model_vimp")
  }
)



# extract_model_vimp (familiarEnsemble) ----------------------------------------
setMethod(
  "extract_model_vimp",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Extracting variable importance obtained from the models."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Obtain aggregation method from stored settings, if required.
    if (is.waive(aggregation_method)) {
      if (length(object@model_list) > 1L) {
        aggregation_method <- object@settings$aggregation
        
      } else {
        # If only one model is evaluated, do not aggregate data.
        aggregation_method <- "none" 
      }
    } 
    
    # Check aggregation method
    .check_parameter_value_is_valid(
      x = aggregation_method,
      var_name = "aggregation_method",
      values = .get_available_rank_aggregation_methods()
    )
    
    # Obtain rank thresholds from stored settings, if required
    if (is.waive(rank_threshold)) rank_threshold <- object@settings$aggr_rank_threshold
    
    # Check rank threshold.
    if (!is.waive(rank_threshold)) {
      .check_number_in_valid_range(
        x = rank_threshold,
        var_name = "rank_threshold",
        range = c(1L, Inf)
      )
    }
    
    # Generate a prototype data element
    proto_data_element <- methods::new(
      "familiarDataElementVimpData",
      rank_threshold = rank_threshold,
      rank_aggregation_method = aggregation_method
    )
    
    # Generate elements to send to dispatch.
    vimp_info <- extract_dispatcher(
      FUN = .extract_model_vimp,
      cl = NULL,
      has_internal_bootstrap = FALSE,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      aggregation_method = aggregation_method,
      aggregate_results = FALSE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(vimp_info)
  }
)



# extract_model_vimp (prediction table) ----------------------------------------
setMethod(
  "extract_model_vimp",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("model-based variable importance")
    
    return(NULL)
  }
)



.extract_model_vimp <- function(
    object,
    data,
    proto_data_element,
    ...
) {
  
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Check that the model has been trained.
  if (!model_is_trained(object)) return(NULL)
  
  # Extract variable importance from the models
  vimp_table <- get_vimp_table(
    x = object,
    data = data,
    as_object = TRUE
  )
  
  # Check that the variable importance table is not empty.
  if (is_empty(vimp_table)) return(NULL)
  
  # Add data.
  data_element <- proto_data_element
  data_element@data <- vimp_table

  return(data_element)
}



# extract_fs_vimp (generic) ----------------------------------------------------

#'@title Internal function to extract feature selection variable importance.
#'
#'@description Aggregate variable importance obtained during feature selection.
#'  This information can only be obtained as part of the main `summon_familiar`
#'  process.
#'
#'@inheritParams .extract_data
#'
#'@return A list containing feature selection variable importance information.
#'@md
#'@keywords internal
setGeneric(
  "extract_fs_vimp",
  function(
    object,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_fs_vimp")
  }
)


# extract_fs_vimp (familiarEnsemble) -------------------------------------------
setMethod(
  "extract_fs_vimp",
  signature(object = "familiarEnsemble"),
  function(
    object,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Obtain aggregation method from stored settings, if required.
    if (is.waive(aggregation_method)) {
      aggregation_method <- object@settings$aggregation
    }
    
    # Check aggregation method
    .check_parameter_value_is_valid(
      x = aggregation_method,
      var_name = "aggregation_method",
      values = .get_available_rank_aggregation_methods()
    )
    
    # Obtain rank thresholds from stored settings, if required
    if (is.waive(rank_threshold)) {
      rank_threshold <- object@settings$aggr_rank_threshold
    }
    
    # Check rank threshold.
    if (!is.waive(rank_threshold)) {
      .check_number_in_valid_range(
        x = rank_threshold,
        var_name = "rank_threshold",
        range = c(1L, Inf)
      )
    }
    
    # Extract variable importance tables from models.
    vimp_table_list <- unlist(
      lapply(
        object@model_list,
        function(x) {
          x <- load_familiar_object(x)
          return(x@vimp_table)
        }
      ),
      use.names = FALSE
    )
    if (!rlang::is_bare_list(vimp_table_list)) vimp_table_list <- list(vimp_table_list)
    
    # Remove missing entries.
    vimp_table_list <- vimp_table_list[sapply(vimp_table_list, is, class2 = "vimpTable")]
    if (length(vimp_table_list) == 0L) return(NULL)
    
    # Select only non-duplicate variable importance tables.
    table_identifiers <- sapply(
      vimp_table_list,
      function(x) paste(x@project_id, x@data_id, x@run_id, x@vimp_method, sep = "_")
    )
    vimp_table_list <- vimp_table_list[!duplicated(table_identifiers)]
    
    # Create list of data elements.
    data_element_list <- lapply(
      vimp_table_list, 
      function(x, rank_threshold, aggregation_method) {
        return(methods::new(
          "familiarDataElementVimpData",
          data = x,
          rank_threshold = rank_threshold,
          rank_aggregation_method = aggregation_method
        ))
      },
      rank_threshold = rank_threshold,
      aggregation_method = aggregation_method
    )
    
    # Merge data elements
    data_element <- merge_data_elements(data_element_list)
    
    return(data_element)
  }
)


# extract_fs_vimp (prediction table) -------------------------------------------
setMethod(
  "extract_fs_vimp",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("variable importance")
    
    return(NULL)
  }
)



# merge_data_elements (familiarDataElementVimpData) ----------------------------
setMethod(
  "merge_data_elements",
  signature(x = "familiarDataElementVimpData"),
  function(
    x,
    x_list,
    ...
  ) {
    
    # Identify items that can be joined.
    id_table <- identify_element_sets(x = x_list, ...)
    
    # Identify the element identifiers that should be grouped.
    grouped_data_element_ids <- lapply(
      split(id_table[, c("element_id", "group_id")], by = "group_id"),
      function(id_table) (id_table$element_id)
    )
    
    # List of data elements.
    data_elements <- list()
    
    for (current_group_data_element_ids in grouped_data_element_ids) {
      # Copy the first data element in the group and use it as a
      # prototype.
      prototype_data_element <- x_list[[current_group_data_element_ids[1L]]]
      
      if (any(sapply(x_list[current_group_data_element_ids], function(x) (is(x@data, "vimpTable"))))) {
        
        # Data attribute contains a variable importance table object.
        data_attribute <- lapply(
          x_list[current_group_data_element_ids],
          function(x) (x@data)
        )
        
        # Set data attribute.
        prototype_data_element@data <- data_attribute
        
      } else if (length(current_group_data_element_ids) != 1L) {
        ..error_reached_unreachable_code(
          "merge_data_elements,familiarDataElementVimpData: exactly one element is expected."
        )
      }
      
      # Add merged data element to the list.
      data_elements <- c(data_elements, list(prototype_data_element))
    }
    
    return(data_elements)
  }
)



# ..compute_data_elements_estimates (familiarDataElementHyperparameters) -------
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElementVimpData"),
  function(
    x,
    x_list = NULL,
    ...
  ) {
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Remove empty entries.
    x <- x[!sapply(x, is_empty)]
    if (is_empty(x)) return(NULL)
    
    # Collect all the stored variable importance table objects.
    vimp_table_list <- unlist(lapply(x, function(x) x@data))
    
    # Aggregate list of vimp tables.
    vimp_object <- aggregate_vimp_table(
      x = vimp_table_list,
      aggregation_method = x[[1L]]@rank_aggregation_method,
      rank_threshold = x[[1L]]@rank_threshold
    )
    
    # Copy data element.
    y <- x[[1L]]
    y@data <- vimp_object
    
    return(y)
  }
)



# .export (familiarDataElementVimpData) ----------------------------------------
setMethod(
  ".export",
  signature(x = "familiarDataElementVimpData"),
  function(
    x,
    x_list,
    aggregate_results = FALSE,
    ...
  ) {
    
    if (aggregate_results) {
      x_list <- .compute_data_element_estimates(x_list)
    }
    
    # Extract variable importance tables, and add identifiers.
    x_list <- lapply(x_list, ..export_vimp_data_elements)
    
    # Extract the data.table and merge to a single data.table.
    vimp_table <- lapply(x_list, function(x) x@data)
    vimp_table <- data.table::rbindlist(vimp_table, use.names = TRUE)
    
    # Form prototype.
    x <- x_list[[1L]]
    x@data <- vimp_table
    
    return(x)
  }
)



..export_vimp_data_elements <- function(x) {
  # Until export, all variable importance data elements are vimpTable objects.
  # We need to export these as data.tables/
  
  # Convert to data.table internally.
  if (is.list(x@data)) {
    vimp_table_list <- lapply(x@data, get_vimp_table)
    
  } else {
    vimp_table_list <- list(get_vimp_table(x@data))
  }
  
  # Combine to a single table.
  vimp_table_list <- data.table::rbindlist(vimp_table_list, use.names = TRUE)
  
  # Check if the resulting combined table is empty.
  if (is_empty(vimp_table_list)) vimp_table_list <- NULL
  
  # Update the data attribute.
  x@data <- vimp_table_list
  
  # Update grouping and value columns.
  x@grouping_column <- "name"
  x@value_column <- c("score", "rank")
  
  # Add data attributes to the table
  x <- .identifier_as_data_attribute(
    x = x,
    identifier = "all",
    as_grouping_column = TRUE
  )
  
  return(x)
}


# export_model_vimp (generic) --------------------------------------------------

#'@title Extract and export model-based variable importance.
#'
#'@description Extract and export model-based variable importance from a
#'  familiarCollection.
#'
#'@param aggregation_method (*optional*) The method used to aggregate variable
#'  importances over different data subsets, e.g. bootstraps. The following
#'  methods can be selected:
#'
#'  * `mean` (default): Use the mean rank of a feature over the subsets to
#'  determine the aggregated feature rank.
#'
#'  * `median`: Use the median rank of a feature over the subsets to determine
#'  the aggregated feature rank.
#'
#'  * `best`: Use the best rank the feature obtained in any subset to determine
#'  the aggregated feature rank.
#'
#'  * `worst`: Use the worst rank the feature obtained in any subset to
#'  determine the aggregated feature rank.
#'
#'  * `stability`: Use the frequency of the feature being in the subset of
#'  highly ranked features as measure for the aggregated feature rank
#'  (Meinshausen and Buehlmann, 2010).
#'
#'  * `exponential`: Use a rank-weighted frequence of occurrence in the subset
#'  of highly ranked features as measure for the aggregated feature rank (Haury
#'  et al., 2011).
#'
#'  * `borda`: Use the borda count as measure for the aggregated feature rank
#'  (Wald et al., 2012).
#'
#'  * `enhanced_borda`: Use an occurrence frequency-weighted borda count as
#'  measure for the aggregated feature rank (Wald et al., 2012).
#'
#'  * `truncated_borda`: Use borda count computed only on features within the
#'  subset of highly ranked features.
#'
#'  * `enhanced_truncated_borda`: Apply both the enhanced borda method and the
#'  truncated borda method and use the resulting borda count as the aggregated
#'  feature rank.
#'
#'@param rank_threshold (*optional*) The threshold used to define the subset of
#'  highly important features. If not set, this threshold is determined by
#'  maximising the variance in the occurrence value over all features over the
#'  subset size.
#'
#'  This parameter is only relevant for `stability`, `exponential`,
#'  `enhanced_borda`, `truncated_borda` and `enhanced_truncated_borda` methods.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams as_familiar_collection
#'
#'@details Data, such as model performance and calibration information, is
#'  usually collected from a `familiarCollection` object. However, you can also
#'  provide one or more `familiarData` objects, that will be internally
#'  converted to a `familiarCollection` object. It is also possible to provide a
#'  `familiarEnsemble` or one or more `familiarModel` objects together with the
#'  data from which data is computed prior to export. Paths to the previous
#'  files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Variable importance is based on the ranking produced by model-specific
#'  variable importance routines, e.g. permutation for random forests. If such a
#'  routine is absent, variable importance is based on the feature selection
#'  method that led to the features included in the model. In case multiple
#'  models (`familiarModel` objects) are combined, feature ranks are first
#'  aggregated using the method defined by the `aggregation_method`, some of
#'  which require a `rank_threshold` to indicate a subset of most important
#'  features.
#'
#'  Information concerning highly similar features that form clusters is
#'  provided as well. This information is based on consensus clustering of the
#'  features that were used in the signatures of the underlying models. This
#'  clustering information is also used during aggregation to ensure that
#'  co-clustered features are only taken into account once.
#'
#'@return A data.table (if `dir_path` is not provided), or nothing, as all data
#'  is exported to `csv` files.
#'@exportMethod export_model_vimp
#'@md
#'@rdname export_model_vimp-methods
setGeneric(
  "export_model_vimp",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_model_vimp")
  }
)



# export_model_vimp (collection) -----------------------------------------------

#'@rdname export_model_vimp-methods
setMethod(
  "export_model_vimp",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Extract data.
    x <- object@model_vimp
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    if (!is.waive(aggregation_method)) {
      # Check if the aggregation method is valid.
      .check_parameter_value_is_valid(
        x = aggregation_method,
        var_name = "aggregation_method",
        values = .get_available_rank_aggregation_methods()
      )
      
      # Set aggregation method.
      x <- lapply(
        x, 
        function(x, aggregation_method) {
          x@rank_aggregation_method <- aggregation_method
          return(x)
        },
        aggregation_method = aggregation_method
      )
    }
    
    if (!is.waive(rank_threshold)) {
      # Check if the threshold is valid.
      .check_number_in_valid_range(
        x = rank_threshold,
        var_name = "rank_threshold",
        range = c(1L, Inf)
      )
      
      # Set threshold.
      x <- lapply(
        x,
        function(x, rank_threshold) {
          x@rank_threshold <- rank_threshold
          return(x)
        },
        rank_threshold = rank_threshold
      )
    }
    
    subtype <- paste("learner", x[[1L]]@rank_aggregation_method, sep = "_")
    
    return(.export(
      x = object,
      data_elements = x,
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "variable_importance",
      subtype = subtype,
      export_collection = export_collection
    ))
  }
)



# export_model_vimp (general) --------------------------------------------------

#'@rdname export_model_vimp-methods
setMethod(
  "export_model_vimp",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object. Note that we pass
    # on aggregation_method and rank_threshold values.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "model_vimp",
          "aggregation_method" =  aggregation_method,
          "rank_threshold" = rank_threshold
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_model_vimp,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "aggregation_method" = aggregation_method,
          "rank_threshold" = rank_threshold,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)



# export_fs_vimp (generic) -----------------------------------------------------

#'@title Extract and export feature selection variable importance.
#'
#'@description Extract and export feature selection variable importance from a
#'  familiarCollection.
#'
#'@param aggregation_method (*optional*) The method used to aggregate variable
#'  importances over different data subsets, e.g. bootstraps. The following
#'  methods can be selected:
#'
#'  * `mean` (default): Use the mean rank of a feature over the subsets to
#'  determine the aggregated feature rank.
#'
#'  * `median`: Use the median rank of a feature over the subsets to determine
#'  the aggregated feature rank.
#'
#'  * `best`: Use the best rank the feature obtained in any subset to determine
#'  the aggregated feature rank.
#'
#'  * `worst`: Use the worst rank the feature obtained in any subset to
#'  determine the aggregated feature rank.
#'
#'  * `stability`: Use the frequency of the feature being in the subset of
#'  highly ranked features as measure for the aggregated feature rank
#'  (Meinshausen and Buehlmann, 2010).
#'
#'  * `exponential`: Use a rank-weighted frequence of occurrence in the subset
#'  of highly ranked features as measure for the aggregated feature rank (Haury
#'  et al., 2011).
#'
#'  * `borda`: Use the borda count as measure for the aggregated feature rank
#'  (Wald et al., 2012).
#'
#'  * `enhanced_borda`: Use an occurrence frequency-weighted borda count as
#'  measure for the aggregated feature rank (Wald et al., 2012).
#'
#'  * `truncated_borda`: Use borda count computed only on features within the
#'  subset of highly ranked features.
#'
#'  * `enhanced_truncated_borda`: Apply both the enhanced borda method and the
#'  truncated borda method and use the resulting borda count as the aggregated
#'  feature rank.
#'
#'@param rank_threshold (*optional*) The threshold used to define the subset of
#'  highly important features. If not set, this threshold is determined by
#'  maximising the variance in the occurrence value over all features over the
#'  subset size.
#'
#'  This parameter is only relevant for `stability`, `exponential`,
#'  `enhanced_borda`, `truncated_borda` and `enhanced_truncated_borda` methods.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams as_familiar_collection
#'
#'@details Data, such as model performance and calibration information, is
#'  usually collected from a `familiarCollection` object. However, you can also
#'  provide one or more `familiarData` objects, that will be internally
#'  converted to a `familiarCollection` object. Paths to the previous files can
#'  also be provided.
#'
#'  Unlike other export function, export using `familiarEnsemble` or
#'  `familiarModel` objects is not possible. This is because feature selection
#'  variable importance is not stored within `familiarModel` objects.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Variable importance is based on the ranking produced by feature selection
#'  routines. In case feature selection was performed repeatedly, e.g. using
#'  bootstraps, feature ranks are first aggregated using the method defined by
#'  the `aggregation_method`, some of which require a `rank_threshold` to
#'  indicate a subset of most important features.
#'
#'  Information concerning highly similar features that form clusters is
#'  provided as well. This information is based on consensus clustering of the
#'  features. This clustering information is also used during aggregation to
#'  ensure that co-clustered features are only taken into account once.
#'
#'@return A data.table (if `dir_path` is not provided), or nothing, as all data
#'  is exported to `csv` files.
#'@exportMethod export_fs_vimp
#'@md
#'@rdname export_fs_vimp-methods
setGeneric(
  "export_fs_vimp",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_fs_vimp")
  }
)



# export_fs_vimp (collection) --------------------------------------------------

#'@rdname export_fs_vimp-methods
setMethod(
  "export_fs_vimp",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Extract data.
    x <- object@fs_vimp
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    if (!is.waive(aggregation_method)) {
      # Check if the aggregation method is valid.
      .check_parameter_value_is_valid(
        x = aggregation_method,
        var_name = "aggregation_method",
        values = .get_available_rank_aggregation_methods()
      )
      
      # Set aggregation method.
      x <- lapply(
        x,
        function(x, aggregation_method) {
          x@rank_aggregation_method <- aggregation_method
          return(x)
        },
        aggregation_method = aggregation_method
      )
    }
    
    if (!is.waive(rank_threshold)) {
      # Check if the threshold is valid.
      .check_number_in_valid_range(
        x = rank_threshold,
        var_name = "rank_threshold",
        range = c(1L, Inf)
      )
      
      # Set threshold.
      x <- lapply(
        x,
        function(x, rank_threshold) {
          x@rank_threshold <- rank_threshold
          return(x)
        },
        rank_threshold = rank_threshold
      )
    }
    
    # Get subtype
    subtype <- paste("feature_selection", x[[1L]]@rank_aggregation_method, sep = "_")
    
    return(.export(
      x = object,
      data_elements = x,
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "variable_importance",
      subtype = subtype,
      export_collection = export_collection
    ))
  }
)


# export_fs_vimp (general) -----------------------------------------------------

#'@rdname export_fs_vimp-methods
setMethod(
  "export_fs_vimp",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object. Note that we pass on
    # aggregation_method and rank_threshold values.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "fs_vimp",
          "aggregation_method" = aggregation_method,
          "rank_threshold" = rank_threshold
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_fs_vimp,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "aggregation_method" = aggregation_method,
          "rank_threshold" = rank_threshold,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# plot_variable_importance (generic) -------------------------------------------

#' @title Plot variable importance scores of features during feature selection
#'   or after training a model.
#'
#' @description This function plots variable importance based data obtained
#'   during feature selection or after training a model, which are stored in a
#'   `familiarCollection` object.
#'
#' @param type Determine what variable importance should be shown. Can be
#'   `feature_selection` or `model` for the variable importance after the
#'   feature selection step and after the model training step, respectively.
#' @param dir_path (*optional*) Path to the directory where created figures are
#'   saved to. Output is saved in the `variable_importance` subdirectory. If
#'   `NULL` no figures are saved, but are returned instead.
#' @param show_cluster (*optional*) Show which features were clustered together.
#'   Currently not available in combination with variable importance obtained
#'   during feature selection.
#' @param gradient_palette (*optional*) Palette for filling bars if the
#'   `color_by` argument is not set. By default, bars are not coloured. If
#'   `gradient_palette` is set, the palette will colour bars according to
#'   feature importance. Use `NULL` to fill the bars using the `familiar`
#'   default palette. Other palettes are supported by the `paletteer` package,
#'   `grDevices::palette.pals()` (requires R >= 4.0.0), `grDevices::hcl.pals()`
#'   (requires R >= 3.6.0) and `rainbow`, `heat.colors`, `terrain.colors`,
#'   `topo.colors` and `cm.colors`, which correspond to the palettes of the same
#'   name in `grDevices`. You may also specify your own palette by providing a
#'   vector of colour names listed by `grDevices::colors()` or through
#'   hexadecimal RGB strings.
#' @param height (*optional*) Height of the plot. A default value is derived
#'   from number of facets, and the length of the longest feature name (if
#'   `rotate_x_tick_labels` is `TRUE`).
#' @param width (*optional*) Width of the plot. A default value is derived from
#'   the number of facets and the number of features.
#' @inheritParams export_fs_vimp
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_fs_vimp -object -aggregation_method -rank_threshold
#'
#' @details This function generates a barplot based on variable importance of
#'   features.
#'
#'   The only allowed values for `split_by`, `color_by` or `facet_by` are
#'   `vimp_method` and `learner`, but note that `learner` has no effect when
#'   plotting variable importance of features acquired during feature selection.
#'
#'   Labeling methods such as `set_feature_names` or `set_vimp_method_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_variable_importance
#' @export plot_feature_selection_occurrence
#' @export plot_feature_selection_variable_importance
#' @export plot_model_signature_occurrence
#' @export plot_model_signature_variable_importance
#' @md
#' @rdname plot_variable_importance-methods
setGeneric(
  "plot_variable_importance",
  function(
    object,
    type,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    show_cluster = TRUE,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = waiver(),
    x_label = "feature",
    rotate_x_tick_labels = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_variable_importance")
  }
)



# plot_variable_importance (general) -------------------------------------------

#' @rdname plot_variable_importance-methods
setMethod(
  "plot_variable_importance", 
  signature(object = "ANY"),
  function(
    object,
    type,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    show_cluster = TRUE,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = waiver(),
    x_label = "feature",
    rotate_x_tick_labels = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Set the data element.
    data_element <- switch(
      type,
      "feature_selection" = "fs_vimp",
      "model" = "model_vimp"
    )
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = data_element,
          "feature_cluster_method" = feature_cluster_method,
          "feature_linkage_method" = feature_linkage_method,
          "feature_cluster_cut_method" = feature_cluster_cut_method,
          "feature_similarity_threshold" = feature_similarity_threshold,
          "aggregation_method" = aggregation_method,
          "rank_threshold" = rank_threshold
        ),
        list(...)
      )
    )

    return(do.call(
      plot_variable_importance,
      args = c(
        list(
          "object" = object,
          "type" = type,
          "feature_cluster_method" = feature_cluster_method,
          "feature_linkage_method" = feature_linkage_method,
          "feature_cluster_cut_method" = feature_cluster_cut_method,
          "feature_similarity_threshold" = feature_similarity_threshold,
          "aggregation_method" = aggregation_method,
          "rank_threshold" = rank_threshold,
          "draw" = draw,
          "dir_path" = dir_path,
          "split_by" = split_by,
          "color_by" = color_by,
          "facet_by" = facet_by,
          "facet_wrap_cols" = facet_wrap_cols,
          "show_cluster" = show_cluster,
          "ggtheme" = ggtheme,
          "discrete_palette" = discrete_palette,
          "gradient_palette" = gradient_palette,
          "x_label" = x_label,
          "rotate_x_tick_labels" = rotate_x_tick_labels,
          "y_label" = y_label,
          "legend_label" = legend_label,
          "plot_title" = plot_title,
          "plot_sub_title" = plot_sub_title,
          "caption" = caption,
          "limit_n_features" = limit_n_features,
          "y_range" = y_range,
          "y_n_breaks" = y_n_breaks,
          "y_breaks" = y_breaks,
          "width" = width,
          "height" = height,
          "units" = units,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)



# plot_variable_importance (familiarCollection) --------------------------------

#' @rdname plot_variable_importance-methods
setMethod(
  "plot_variable_importance",
  signature(object = "familiarCollection"),
  function(
    object,
    type,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    aggregation_method = waiver(),
    rank_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    show_cluster = TRUE,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = waiver(),
    x_label = "feature",
    rotate_x_tick_labels = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(.plot_variable_importance(
      object = object,
      type = type,
      feature_cluster_method = feature_cluster_method,
      feature_linkage_method = feature_linkage_method,
      feature_cluster_cut_method = feature_cluster_cut_method,
      feature_similarity_threshold = feature_similarity_threshold,
      aggregation_method = aggregation_method,
      rank_threshold = rank_threshold,
      draw = draw,
      dir_path = dir_path,
      split_by = split_by,
      color_by = color_by,
      facet_by = facet_by,
      facet_wrap_cols = facet_wrap_cols,
      show_cluster = show_cluster,
      ggtheme = ggtheme,
      discrete_palette = discrete_palette,
      gradient_palette = gradient_palette,
      x_label = x_label,
      rotate_x_tick_labels = rotate_x_tick_labels,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      limit_n_features = limit_n_features,
      y_range = y_range,
      y_n_breaks = y_n_breaks,
      y_breaks = y_breaks,
      width = width,
      height = height,
      units = units,
      export_collection = export_collection,
      ...
    ))
  }
)



#' @rdname plot_variable_importance-methods
plot_feature_selection_occurrence <- function(...) {
  return(do.call(
    plot_variable_importance,
    args = c(
      list(
        "type" = "feature_selection",
        "aggregation_method" = "stability"
      ),
      list(...)
    )
  ))
}



#' @rdname plot_variable_importance-methods
plot_feature_selection_variable_importance <- function(...) {
  return(do.call(
    plot_variable_importance,
    args = c(
      list("type" = "feature_selection"),
      list(...)
    )
  ))
}



#' @rdname plot_variable_importance-methods
plot_model_signature_occurrence <- function(...) {
  return(do.call(
    plot_variable_importance,
    args = c(
      list(
        "type" = "model",
        "aggregation_method" = "stability"
      ),
      list(...)
    )
  ))
}


#' @rdname plot_variable_importance-methods
plot_model_signature_variable_importance <- function(...) {
  return(do.call(
    plot_variable_importance,
    args = c(
      list("type" = "model"),
      list(...)
    )
  ))
}



.plot_variable_importance <- function(
    object,
    type,
    feature_cluster_method,
    feature_linkage_method,
    feature_cluster_cut_method,
    feature_similarity_threshold,
    aggregation_method,
    rank_threshold,
    draw,
    dir_path,
    split_by,
    color_by,
    facet_by,
    facet_wrap_cols,
    show_cluster,
    ggtheme,
    discrete_palette,
    gradient_palette,
    x_label,
    rotate_x_tick_labels,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    limit_n_features,
    y_range,
    y_n_breaks,
    y_breaks,
    width,
    height,
    units,
    export_collection = FALSE,
    ...
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  data_set <- NULL

  # Get input data.
  if (type == "feature_selection") {
    x <- export_fs_vimp(
      object = object,
      aggregation_method = aggregation_method,
      rank_threshold = rank_threshold,
      aggregate_results = TRUE
    )

    available_splitting <- "vimp_method"
    
  } else if (type == "model") {
    x <- export_model_vimp(
      object = object,
      aggregation_method = aggregation_method,
      rank_threshold = rank_threshold,
      aggregate_results = TRUE
    )

    available_splitting <- c("vimp_method", "learner")
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".plot_variable_importance: unknown value for type (", type, ")"
    ))
  }

  # Check that the data are not empty.
  if (is_empty(x)) return(NULL)

  # Obtain data element from list.
  if (is.list(x)) {
    if (is_empty(x)) return(NULL)

    if (length(x) > 1L) {
      ..error_reached_unreachable_code(
        ".plot_variable_importance: list of data elements contains unmerged elements."
      )
    }

    # Get x directly.
    x <- x[[1L]]
  }

  # Check that the data are not empty.
  if (is_empty(x)) return(NULL)

  # Check package requirements for plotting.
  if (!require_package(
    x = ..required_plotting_packages(extended = FALSE),
    purpose = ifelse(
      type == "feature_selection",
      "to plot variable importance determined through feature selection methods",
      "to plot model-based variable importance"
    ),
    message_type = "warning"
  )) {
    return(NULL)
  }

  # Check input arguments ------------------------------------------------------
  
  # Check show_cluster
  .check_parameter_value_is_valid(
    x = show_cluster, 
    var_name = "show_cluster",
    values = c(FALSE, TRUE)
  )

  if (show_cluster) {
    # Get feature similarity data.
    feature_similarity <- export_feature_similarity(
      object = object,
      feature_cluster_method = feature_cluster_method,
      feature_linkage_method = feature_linkage_method,
      feature_cluster_cut_method = feature_cluster_cut_method,
      feature_similarity_threshold = feature_similarity_threshold,
      export_dendrogram = FALSE,
      export_ordered_data = FALSE,
      export_clustering = TRUE
    )[[1L]]
    
  } else {
    feature_similarity <- NULL
  }

  # ggtheme
  ggtheme <- .check_ggtheme(ggtheme)

  # rotate_x_tick_labels
  if (is.waive(rotate_x_tick_labels)) {
    rotate_x_tick_labels <- FALSE
  }

  # Clusters cannot be generated in case no cluster information is
  # present.
  if (is_empty(feature_similarity) || type == "feature_selection") {
    show_cluster <- FALSE
  }

  # y_label
  if (is.waive(y_label)) {
    y_label <- ifelse(x@rank_aggregation_method == "stability", "occurrence", "score")
  }

  # y_range
  if (is.null(y_range) && x@rank_aggregation_method == "stability") {
    # for occurrence plots
    y_range <- c(0.0, 1.0)
    
  } else if (is.null(y_range)) {
    # for variable importance score-based plots
    y_range <- c(0.0, max(x@data$score, na.rm = TRUE))
  }

  # Set y_breaks
  if (is.null(y_breaks)) {
    .check_input_plot_args(
      y_range = y_range,
      y_n_breaks = y_n_breaks
    )

    # Create breaks and update x_range
    y_breaks <- labeling::extended(
      m = y_n_breaks,
      dmin = y_range[1L],
      dmax = y_range[2L],
      only.loose = TRUE
    )
    
    y_range <- c(0.0, tail(y_breaks, n = 1L))
  }

  # Add default parameters.
  if (is.null(split_by) && is.null(color_by) && is.null(facet_by)) {
    split_by <- available_splitting
  }

  # Check splitting variables and generate sanitised output.
  split_var_list <- .check_plot_splitting_variables(
    x = x@data,
    split_by = split_by,
    color_by = color_by,
    facet_by = facet_by,
    available = available_splitting
  )

  # Update splitting variables
  split_by <- split_var_list$split_by
  color_by <- split_var_list$color_by
  facet_by <- split_var_list$facet_by

  # legend_label
  legend_label <- .create_plot_legend_title(
    user_label = legend_label,
    color_by = color_by
  )

  # Perform last checks prior to plotting
  .check_input_plot_args(
    x_label = x_label,
    y_label = y_label,
    legend_label = legend_label,
    plot_title = plot_title,
    plot_sub_title = plot_sub_title,
    caption = caption,
    rotate_x_tick_labels = rotate_x_tick_labels,
    facet_wrap_cols = facet_wrap_cols,
    y_range = y_range,
    y_breaks = y_breaks,
    limit_n_features = limit_n_features
  )

  # Create plots ---------------------------------------------------------------

  # Determine if subtitle should be generated.
  autogenerate_plot_subtitle <- is.waive(plot_sub_title)

  # Split data
  if (!is.null(split_by)) {
    x_split <- split(x@data, by = split_by)
  } else {
    x_split <- list(x@data)
  }

  # Store plots to list in case no dir_path is provided
  if (is.null(dir_path)) plot_list <- list()

  # Iterate over splits
  for (x_sub in x_split) {
    if (is_empty(x_sub)) next

    # Rename "name" column to "feature".
    x_sub <- data.table::copy(x_sub)
    data.table::setnames(
      x = x_sub,
      old = "name",
      new = "feature"
    )

    # Join cluster and univariate data.
    if (show_cluster) {
      if (type == "model") {
        x_temporary <- merge(
          x = x_sub,
          y = feature_similarity@data,
          by.x = c("feature", available_splitting, "ensemble_model_name"),
          by.y = c("feature", available_splitting, "ensemble_model_name"),
          allow.cartesian = TRUE
        )

        # Model-based ranking does not aggregate along data-set.
        x_temporary <- x_temporary[data_set %in% c(unique(x_temporary$data_set)[1L])]
      }

      # Check that the resulting data is not empty, because this would
      # mean that there is e.g. only one feature.
      if (is_empty(x_temporary)) {
        x_temporary <- data.table::copy(x_sub)
        x_temporary[, ":="(
          "cluster_id" = .I,
          "cluster_size" = 1L
        )]
      }

      # Replace x_sub
      x_sub <- x_temporary
    }

    if (is.waive(plot_title)) {
      plot_title <- ifelse(
        type == "feature_selection",
        "Feature selection-based variable importance",
        "Model-based variable importance"
      )
    }

    if (autogenerate_plot_subtitle) {
      plot_sub_title <- .create_plot_subtitle(
        split_by = split_by,
        additional = list("aggregation_method" = x@rank_aggregation_method),
        x = x_sub
      )
    }

    # Generate plot
    p <- .create_feature_rank_plot(
      x = x_sub,
      type = type,
      aggregation_method = x@rank_aggregation_method,
      color_by = color_by,
      facet_by = facet_by,
      facet_wrap_cols = facet_wrap_cols,
      show_cluster = show_cluster,
      ggtheme = ggtheme,
      discrete_palette = discrete_palette,
      gradient_palette = gradient_palette,
      x_label = x_label,
      rotate_x_tick_labels = rotate_x_tick_labels,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      limit_n_features = limit_n_features,
      y_range = y_range,
      y_breaks = y_breaks
    )

    # Check empty output
    if (is.null(p)) next

    # Draw plot
    if (draw) .draw_plot(plot_or_grob = p)

    # Save and export
    if (!is.null(dir_path)) {
      # Obtain decent default values for the plot.
      def_plot_dims <- .determine_feature_ranking_plot_dimensions(
        x = x_sub,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        rotate_x_tick_labels = rotate_x_tick_labels
      )

      # Save to file.
      do.call(
        .save_plot_to_file,
        args = c(
          list(
            "plot_or_grob" = p,
            "object" = object,
            "dir_path" = dir_path,
            "type" = "variable_importance",
            "subtype" = ifelse(type == "feature_selection", "feature_selection", "learner"),
            "x" = x_sub,
            "split_by" = split_by,
            "additional" = list("aggregation_method" = x@rank_aggregation_method),
            "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
            "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
            "units" = ifelse(is.waive(units), "cm", units)
          ),
          list(...)
        )
      )
    } else {
      # Store as list and export
      plot_list <- c(plot_list, list(p))
    }
  }

  # Generate output
  return(.get_plot_results(
    dir_path = dir_path,
    plot_list = plot_list,
    export_collection = export_collection,
    object = object
  ))
}



.create_feature_rank_plot <- function(
    x,
    type,
    aggregation_method,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    show_cluster,
    discrete_palette,
    gradient_palette,
    x_label,
    rotate_x_tick_labels,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    limit_n_features,
    y_range,
    y_breaks
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  rank <- feature <- NULL

  # Create a local copy of x prior to making changes based on plot_data
  x <- data.table::copy(x)
  
  if (is.numeric(limit_n_features)) {
    # Find the features with the lowest rank, and select these.
    x_agg <- x[, list("rank" = min(rank)), by = c("feature")]
    
    # Explicitly select features based on threshold value instead of simply 
    # selecting the best features: features may have the same value because they
    # belong to the same cluster.
    threshold_value <- max(head(unique(sort(x_agg$rank)), n = limit_n_features))
    selected_features <- x_agg[rank <= threshold_value]$feature
    
    # Make slice.
    x <- x[feature %in% selected_features, ]
  }
  
  # We need to remove categories related to features that are no longer
  # relevant.
  x$feature <- droplevels(x$feature)
  x <- x[order(rank)]

  # Update the ordering of features so that the features are ordered by
  # increasing score or occurrence
  x$feature <- factor(x$feature, levels = unique(x$feature))

  # Generate a guide table
  guide_list <- .create_plot_guide_table(
    x = x,
    color_by = color_by,
    discrete_palette = discrete_palette
  )

  # Extract data
  x <- guide_list$data

  # Check if cluster information should be shown:
  if (show_cluster) {
    x <- .add_plot_cluster_name(
      x = x,
      color_by = color_by,
      facet_by = facet_by
    )
  }

  # Perform last checks prior to plotting
  .check_input_plot_args(
    y_range = y_range,
    y_breaks = y_breaks
  )

  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("feature"),
      y = !!sym("score")
    )
  )
  p <- p + ggtheme
  
  # Add fill colours.
  if (!is.null(color_by)) {
    # Extract guide_table for color
    g_color <- guide_list$guide_color

    p <- p + ggplot2::geom_bar(
      stat = "identity",
      mapping = ggplot2::aes(fill = !!sym("color_breaks")),
      position = "dodge"
    )

    p <- p + ggplot2::scale_fill_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )
    
  } else if (!is.waive(gradient_palette)) {
    # A gradient palette is used to colour the bars by value.
    p <- p + ggplot2::geom_bar(
      stat = "identity",
      mapping = ggplot2::aes(fill = !!sym("value")),
      show.legend = FALSE
    )

    # Determine gradient order. This is so that bars of more important features
    # are always colored with the high-range colors, independent of the
    # orientation of the score. The correct gradient order will be set using the
    # trans argument, which defines the transformations.
    gradient_order <- "identity"

    # Determine best and worst scores.
    best_score <- x[rank == 1L]$score[1L]
    worst_score <- x[rank == max(x$rank)]$score[1L]

    # Invert gradient if the worst score is higher than the best score.
    if (best_score < worst_score) gradient_order <- "reverse"

    # Get gradient colours
    gradient_colours <- .get_palette(
      x = gradient_palette,
      palette_type = "sequential"
    )

    p <- p + ggplot2::scale_fill_gradientn(
      colors = gradient_colours,
      limits = y_range,
      trans = gradient_order
    )
    
  } else {
    # Bars are not coloured based on occurrence/importance.
    p <- p + ggplot2::geom_bar(stat = "identity")
  }

  # Set breaks and  limits on the y-axis
  p <- p + ggplot2::scale_y_continuous(
    breaks = y_breaks,
    limits = y_range
  )

  # Determine how things are facetted
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context"
      )
    }
  }

  # Add clustering information.
  if (show_cluster) {
    # Obtain default settings.
    text_settings <- .get_plot_geom_text_settings(ggtheme = ggtheme)

    if (is.null(color_by)) {
      p <- p + ggplot2::geom_text(
        ggplot2::aes(
          label = !!sym("cluster_name"),
          y = 0.0
        ),
        colour = text_settings$colour,
        family = text_settings$family,
        fontface = text_settings$face,
        size = text_settings$geom_text_size,
        vjust = "inward"
      )
      
    } else {
      p <- p + ggplot2::geom_text(
        ggplot2::aes(
          label = !!sym("cluster_name"),
          y = 0.0,
          group = !!sym("color_breaks")
        ),
        colour = text_settings$colour,
        family = text_settings$family,
        fontface = text_settings$face,
        size = text_settings$geom_text_size,
        vjust = "inward",
        position = ggplot2::position_dodge(width = 0.9)
      )
    }
  }

  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Rotate x-axis ticks
  if (rotate_x_tick_labels) {
    p <- p + ggplot2::theme(
      axis.text.x = ggplot2::element_text(
        vjust = 0.25, 
        hjust = 1.0, 
        angle = 90.0
      )
    )
  }

  return(p)
}



.determine_feature_ranking_plot_dimensions <- function(
    x,
    facet_by,
    facet_wrap_cols,
    rotate_x_tick_labels
) {
  # Get plot layout dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Determine the number of features within each facet.
  n_features <- data.table::uniqueN(x = x$feature)
  longest_name <- max(sapply(levels(x$feature), nchar))

  # Assume each feature takes up about 14 points (~5mm) with 2 point (0.07mm)
  # spacing. Then add some room for other plot elements.
  default_width <- n_features * 0.5 + (n_features - 1L) * 0.07 + 1.0
  default_width <- max(c(4.0, default_width))

  # Set default height.
  default_height <- 4.0

  # Reserve space for x-axis tick labels. Assume that the typical width of a
  # character is about 5 points (1.8 mm). For the x-axis we only reserve extra
  # space in case the ticks are rotated, otherwise we just assume a typical
  # height of 10 points (3.6 mm).
  x_tick_space <- ifelse(rotate_x_tick_labels, longest_name * 0.18, 0.36)

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height + x_tick_space, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))

  return(c(height, width))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarTaskCollect ----------------------------------------------------------
setClass(
  "familiarTaskCollect",
  contains = "familiarTask",
  slots = list(
    "data_file" = "character"
  ),
  prototype = methods::prototype(
    data_file = NA_character_,
    task_name = "collect_data"
  )
)



# .set_file_name (collection task) ---------------------------------------------
setMethod(
  ".set_file_name",
  signature(object = "familiarTaskCollect"),
  function(object, file_paths = NULL) {
    if (is.null(file_paths)) return(object)
    
    name <- NULL
    if (object@data_id == 1L && object@run_id == 1L) {
      name <- "pooled"
    }
    
    # Generate file name of the model.
    object@file <- get_object_file_name(
      object_type = "familiarCollection",
      data_id = object@data_id,
      run_id = object@run_id,
      name = name,
      project_id = object@project_id,
      dir_path = file_paths$fam_coll_dir
    )
    
    return(object)
  }
)



# .get_task_descriptor (collection task) ---------------------------------------
setMethod(
  ".get_task_descriptor",
  signature(object = "familiarTaskCollect"),
  function(object, ...) {
    return(paste0(
      object@task_name, "_",
      object@data_id, "_", 
      object@run_id
    ))
  }
)



# .perform_task (collection task, NULL) ----------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskCollect",
    data = "NULL"
  ),
  function(
    object,
    data,
    return_results = TRUE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    internal <- label_order <- NULL
    
    # Set collection name.
    if (!is.na(object@data_id) && !is.na(object@run_id)) {
      name <- NULL
      if (object@data_id == 1L && object@run_id == 1L) {
        name <- "pooled"
      }
      
      # Generate file name of the model.
      collection_name <- get_object_file_name(
        object_type = "familiarCollection",
        data_id = object@data_id,
        run_id = object@run_id,
        name = name,
        project_id = object@project_id,
        with_extension = FALSE
      )
      
    } else {
      collection_name <- "collection"
    }
    
    # Process collection.
    collection_object <- suppressWarnings(
      as_familiar_collection(
        object = object@data_file,
        collection_name = collection_name
      )
    )
    
    # Update labels of datasets, which affect how datasets are ordered in
    # plots and tables. Labels are only updated and ordered when the internal
    # naming system is used.
    standard_names <- data.table::data.table(
      "internal" = c("development", "internal_validation", "external_validation"),
      "label" = c("development", "int. validation", "ext. validation"),
      "label_order" = seq_len(3L)
    )
    old_names <- get_data_set_names(collection_object)
    if (all(old_names %in% standard_names$internal)) {
      standard_names <- standard_names[internal %in% old_names][order(label_order)]
      collection_object <- set_data_set_names(
        x = collection_object,
        old = standard_names$internal,
        new = standard_names$label,
        order = standard_names$label
      )
    }
    
    if (!is.na(object@file)) {
      saveRDS(collection_object, file = object@file)
    }
    
    if (return_results) {
      return(collection_object)
    }
  }
)



# .perform_task (collection task, ANY) -----------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskCollect",
    data = "ANY"
  ),
  function(
    object,
    data,
    return_results = TRUE,
    ...
  ) {
    # Process collection.
    collection_object <- suppressWarnings(
      as_familiar_collection(
        object = data
      )
    )
    
    if (!is.na(object@file)) {
      saveRDS(collection_object, file = object@file)
    }
    
    if (return_results) {
      return(collection_object)
    }
    
    return(TRUE)
  }
)



# familiarTaskEvaluate ---------------------------------------------------------
setClass(
  "familiarTaskEvaluate",
  contains = "familiarTask",
  slots = list(
    "validation" = "logical",
    "ensemble_data_id" = "integer",
    "ensemble_run_id" = "integer",
    "predict_data_id" = "integer",
    "force_ensemble_detail_level" = "logical",
    "vimp_method" = "character",
    "learner" = "character",
    "data_set_name" = "character",
    "model_files" = "character"
  ),
  prototype = methods::prototype(
    validation = NA,
    # Ensemble data id and run ids are *only* used for generating unique file
    # names.
    ensemble_data_id = NA_integer_,
    ensemble_run_id = NA_integer_,
    # Whereas data_id describes where the overall data comes from, the
    # ensemble_data_id describes where ensembles are formed.
    predict_data_id = NA_integer_,
    # If individual models do not have sufficient data to perform hybrid
    # analysis (each model is used to compute part of the bootstraps when
    # computing confidence intervals) for an evaluation step, that evaluation
    # step may fail, even though over all models, sufficient data are present.
    # In that case, we need to force an ensemble detail level.
    force_ensemble_detail_level = FALSE,
    vimp_method = NA_character_,
    learner = NA_character_,
    data_set_name = NA_character_,
    model_files = NA_character_,
    task_name = "evaluate"
  )
)



# .set_file_name (evaluation task) ---------------------------------------------
setMethod(
  ".set_file_name",
  signature(object = "familiarTaskEvaluate"),
  function(object, file_paths = NULL) {
    if (is.null(file_paths)) return(object)
    
    # Generate file name of the model.
    object@file <- get_object_file_name(
      object_type = "familiarData",
      data_id = object@data_id,
      run_id = object@run_id,
      ensemble_data_id = object@ensemble_data_id,
      ensemble_run_id = object@ensemble_run_id,
      learner = object@learner,
      vimp_method = object@vimp_method,
      name = object@data_set_name,
      project_id = object@project_id,
      dir_path = file_paths$fam_data_dir
    )
    
    return(object)
  }
)



# .get_task_descriptor (evaluation task) ---------------------------------------
setMethod(
  ".get_task_descriptor",
  signature(object = "familiarTaskEvaluate"),
  function(object, ...) {
    return(paste0(
      object@task_name, "_",
      object@data_id, "_", 
      object@run_id, "_",
      object@ensemble_data_id, "_",
      object@ensemble_run_id, "_",
      object@vimp_method, "_", 
      object@learner, "_",
      object@data_set_name
    ))
  }
)



# .perform_task (evaluation task , NULL) --------------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskEvaluate",
    data = "NULL"
  ),
  function(
    object,
    data,
    experiment_data = NULL,
    outcome_info = NULL,
    ...
  ) {
    # This method is called when "data" is expected to be available somewhere in
    # the backend.
    
    if (is.null(experiment_data)) {
      ..error_reached_unreachable_code("experiment_data is required for retrieving data from the backend.")
    }
    if (is.null(outcome_info)) {
      ..error_reached_unreachable_code("outcome_info is required.")
    }
    
    # Set up a dataObject that allows for delayed loading. Importantly, we set
    # run_id to NA_integer_. This allows for loading the data based on the
    # context provided by data_id and the ensemble or its underlying models.
    data <- methods::new(
      "delayedDataObject",
      data = NULL,
      data_id = object@predict_data_id,
      run_id = NA_integer_,
      preprocessing_level = "none",
      outcome_type = outcome_info@outcome_type,
      outcome_info = outcome_info,
      validation = object@validation,
      aggregate_on_load = FALSE
    )
    
    # Pass to method that dispatches with dataObject for further processing.
    return(.perform_task(
      object = object,
      data = data,
      experiment_data = experiment_data,
      ...
    ))
  }
)




# .perform_task (evaluation task, dataObject) ----------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskEvaluate",
    data = "dataObject"
  ),
  function(
    object,
    data,
    settings,
    cl = NULL,
    message_indent = 0L,
    verbose = FALSE,
    return_results = TRUE,
    ...
  ) {
    # Prevent notes.
    n <- data_id <- NULL
    
    # Signal evaluation start for the current task.
    logger_message(
      paste0(
        "Evaluation: Starting evaluation for the \"", object@learner,
        "\" learner and the \"", object@vimp_method,
        "\" variable importance method for ",
        object@data_set_name, " data (data_id: ",
        object@data_id, "; run_id: ", object@run_id,
        "). This is task ",
        object@task_id, " of ",
        object@n_tasks, "."
      ),
      indent = message_indent,
      verbose = verbose
    )
    # Form an ensemble using the associated or provided models.
    
    # Check which detail level should be provided based on the number of
    # available instances for each model.
    fam_ensemble <- methods::new(
      "familiarEnsemble",
      model_list = as.list(object@model_files),
      learner = object@learner,
      vimp_method = object@vimp_method,
      data_id = object@data_id,
      run_id = object@run_id
    )
    
    # Add package version.
    fam_ensemble <- add_package_version(object = fam_ensemble)
    
    # Load models and prevent auto-detaching.
    fam_ensemble <- load_models(
      object = fam_ensemble,
      suppress_auto_detach = TRUE
    )
    
    # Create a run table. Select only the runs that consistently appear for
    # all models.
    run_table <- data.table::rbindlist(lapply(
      fam_ensemble@model_list,
      function(fam_model) fam_model@run_table)
    )
    
    # Count how many times each run appears. Run tables can start bifurcating
    # for underlying models, e.g. through bootstraps.
    run_table <- run_table[, list("n" = .N), by = eval(colnames(run_table))]
    fam_ensemble@run_table <- run_table[n == length(fam_ensemble@model_list)][order(data_id)]

    # Complete the ensemble using information provided by the model
    fam_ensemble <- complete_familiar_ensemble(
      object = fam_ensemble,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    # Set evaluation level.
    detail_level <- settings$eval$detail_level
    if (object@force_ensemble_detail_level) {
      detail_level <- "ensemble"
    }
    
    # Compute evaluation data.
    evaluation_data <- extract_data(
      object = fam_ensemble,
      data = data,
      cl = cl,
      data_element = settings$eval$evaluation_data_elements,
      time_max = settings$eval$time_max,
      evaluation_times = settings$eval$eval_times,
      sample_limit = settings$eval$sample_limit,
      n_important_features = settings$eval$n_important_features,
      detail_level = detail_level,
      estimation_type = settings$eval$estimation_type,
      aggregate_results = settings$eval$aggregate_results,
      aggregation_method = settings$eval$aggregation,
      rank_threshold = settings$eval$aggr_rank_threshold,
      ensemble_method = settings$eval$ensemble_method,
      stratification_method = settings$eval$strat_method,
      metric = settings$eval$metric,
      feature_cluster_method = settings$eval$feature_cluster_method,
      feature_cluster_cut_method = settings$eval$feature_cluster_cut_method,
      feature_linkage_method = settings$eval$feature_linkage_method,
      feature_similarity_metric = settings$eval$feature_similarity_metric,
      feature_similarity_threshold = settings$eval$feature_similarity_threshold,
      sample_cluster_method = settings$eval$sample_cluster_method,
      sample_linkage_method = settings$eval$sample_linkage_method,
      sample_similarity_metric = settings$eval$sample_similarity_metric,
      confidence_level = settings$eval$confidence_level,
      bootstrap_ci_method = settings$eval$bootstrap_ci_method,
      dynamic_model_loading = settings$eval$auto_detach,
      icc_type = settings$eval$icc_type,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    # Add additional details.
    evaluation_data@name <- object@data_set_name
    evaluation_data@project_id <- object@project_id
    
    if (!is.na(object@file)) {
      saveRDS(evaluation_data, file = object@file)
    }
    
    if (return_results) {
      return(evaluation_data)
    }
    
    return(TRUE)
  }
)



.generate_evaluation_tasks <- function(
  experiment_data,
  vimp_methods,
  learners,
  file_paths,
  pool_only,
  skip_existing = FALSE,
  ...
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  train <- internal_validation <- external_validation <- NULL
  can_pre_process <- perturbation_level <- main_data_id <- NULL
  data_id <- run_id <- NULL
  
  # collection tasks -----------------------------------------------------------
  # Always created the top-layer pooled collection.
  collect_task_list <- list(
    methods::new(
      "familiarTaskCollect",
      data_id = 1L,
      run_id = 1L,
      project_id = experiment_data@project_id
    )
  )
  
  # Find the data_id related to ensembling of models.
  train_data_id <- experiment_data@experiment_setup[train == TRUE, ]$main_data_id[1L]
  internal_validation_data_id <- experiment_data@experiment_setup[internal_validation == TRUE, ]$main_data_id[1L]
  external_validation_data_id <- experiment_data@experiment_setup[external_validation == TRUE, ]$main_data_id[1L]
  
  # If there is no train data id, we have an issue.
  if (is.na(train_data_id)) return(NULL)
  
  # Determine which parts of the experiment can be used for internal validation..
  run_table <- .get_run_table_from_experiment_setup(
    data_id = train_data_id,
    experiment_setup = experiment_data@experiment_setup
  )
  
  if (!pool_only && !is.na(internal_validation_data_id)) {
    # Determine the collections at the last experimental level that can
    # pre-process and is part of the model-building branch.
    collection_run_ids <- seq_len(run_table[main_data_id == internal_validation_data_id]$n_runs[1L])
    
    ii <- 2L
    for (run_id in collection_run_ids) {
      collect_task_list[[ii]] <- methods::new(
        "familiarTaskCollect",
        task_id = ii + 1L,
        data_id = internal_validation_data_id,
        run_id = run_id,
        project_id = experiment_data@project_id
      )
      
      ii <- ii + 1L
    }
  }
  
  # evaluation tasks -----------------------------------------------------------
  
  n_min_model_instances <- Inf
  
  # External validation: the top level has associated validation data.
  if (!is.na(external_validation_data_id)) {
    n_min_model_instances <- min(
      c(experiment_data@experiment_setup[main_data_id == external_validation_data_id]$max_validation_instances),
      n_min_model_instances
    )
  }
  
  # Internal validation: the lowest model ensembling level has associated
  # validation data.
  if (!is.na(internal_validation_data_id)) {
    n_min_model_instances <- min(
      c(experiment_data@experiment_setup[main_data_id == internal_validation_data_id]$max_validation_instances),
      n_min_model_instances
    )
  }
  
  # Development data: the lowest model ensembling level has associated
  # development data. NOTE: this should always be the case.
  if (!is.na(train_data_id)) {
    n_min_model_instances <- min(
      c(experiment_data@experiment_setup[main_data_id == train_data_id]$max_training_instances),
      n_min_model_instances
    )
  }
  
  # Check model instances and determine if we need to force ensemble detail
  # level for evaluation.
  force_ensemble_detail_level <- n_min_model_instances < 10L
  
  # Then select the run-tables which contain this data_id as their final
  # (bottom) level.
  run_tables <- .collect_run_tables(iteration_list = experiment_data@iteration_list)
  run_tables <- run_tables[sapply(
    run_tables,
    function(x, data_id) {
      return (tail(x, n = 1L)$data_id == data_id)
    },
    data_id = train_data_id
  )]
  
  # Use collection tasks to set up the evaluation tasks, including for internal
  # validation.
  evaluate_task_list <- list()
  ii <- 1L
  
  for (jj in seq_along(collect_task_list)) {
    
    # Select run tables where the ensemble data and run identifiers appear.
    selected_run_tables <- run_tables[
      sapply(
        X = run_tables,
        FUN = function(x, task_data_id, task_run_id) {
          return(!is_empty(x[data_id == task_data_id & run_id == task_run_id]))
        },
        task_data_id = collect_task_list[[jj]]@data_id,
        task_run_id = collect_task_list[[jj]]@run_id
      )
    ]
    
    # Initialise file names.
    data_file_names <- NULL
    for (learner in learners) {
      for (vimp_method in vimp_methods) {
        
        # Set model files.
        model_files <- unname(sapply(
          selected_run_tables,
          function(x, ...) {
            get_object_file_name(
              object_type = "familiarModel",
              data_id = tail(x, n = 1L)$data_id,
              run_id = tail(x, n = 1L)$run_id,
              ...
            )
          },
          learner = learner,
          vimp_method = vimp_method,
          project_id = experiment_data@project_id,
          dir_path = file_paths$mb_dir
        ))
        
        ## external validation -------------------------------------------------
        
        if (!is.na(external_validation_data_id)) {
          # Initialise task.
          evaluate_task <- methods::new(
            "familiarTaskEvaluate",
            data_id = external_validation_data_id,
            run_id = 1L,
            validation = TRUE,
            ensemble_data_id = collect_task_list[[jj]]@data_id,  # Only used for naming.
            ensemble_run_id = collect_task_list[[jj]]@run_id,  # Only used for naming.
            predict_data_id = external_validation_data_id,  # Determines which data are dynamically loaded.
            force_ensemble_detail_level = force_ensemble_detail_level,
            learner = learner,
            vimp_method = vimp_method,
            model_files = model_files,
            data_set_name = "external_validation",
            project_id = experiment_data@project_id
          )
          
          # Set file name.
          evaluate_task <-.set_file_name(
            object = evaluate_task,
            file_paths = file_paths
          )
          
          # Make task and associated file names available.
          data_file_names <- c(data_file_names, evaluate_task@file)
          evaluate_task_list[[ii]] <- evaluate_task
          
          ii <- ii + 1L
        }
        
        # internal validation --------------------------------------------------
        
        if (!is.na(internal_validation_data_id)) {
          # Initialise task.
          evaluate_task <- methods::new(
            "familiarTaskEvaluate",
            data_id = collect_task_list[[jj]]@data_id,
            run_id = collect_task_list[[jj]]@run_id,
            validation = TRUE,
            ensemble_data_id = collect_task_list[[jj]]@data_id,
            ensemble_run_id = collect_task_list[[jj]]@run_id,
            predict_data_id = internal_validation_data_id,
            force_ensemble_detail_level = force_ensemble_detail_level,
            learner = learner,
            vimp_method = vimp_method,
            model_files = model_files,
            data_set_name = "internal_validation",
            project_id = experiment_data@project_id
          )
          
          # Set file name.
          evaluate_task <-.set_file_name(
            object = evaluate_task,
            file_paths = file_paths
          )
          
          # Make task and associated file names available.
          data_file_names <- c(data_file_names, evaluate_task@file)
          evaluate_task_list[[ii]] <- evaluate_task
          
          ii <- ii + 1L
        }
        
        # development ----------------------------------------------------------
        
        if (!is.na(train_data_id)) {
          # Initialise task.
          evaluate_task <- methods::new(
            "familiarTaskEvaluate",
            data_id = collect_task_list[[jj]]@data_id,
            run_id = collect_task_list[[jj]]@run_id,
            validation = FALSE,
            ensemble_data_id = collect_task_list[[jj]]@data_id,
            ensemble_run_id = collect_task_list[[jj]]@run_id,
            predict_data_id = train_data_id,
            force_ensemble_detail_level = force_ensemble_detail_level,
            learner = learner,
            vimp_method = vimp_method,
            model_files = model_files,
            data_set_name = "development",
            project_id = experiment_data@project_id
          )
          
          # Set file name.
          evaluate_task <-.set_file_name(
            object = evaluate_task,
            file_paths = file_paths
          )
          
          # Make task and associated file names available.
          data_file_names <- c(data_file_names, evaluate_task@file)
          evaluate_task_list[[ii]] <- evaluate_task
          
          ii <- ii + 1L
          
        } else {
          ..error_reached_unreachable_code("development data are always present.")
        }
      }
    }
    
    # Update collection tasks by adding file paths to 
    collect_task_list[[jj]]@data_file <- data_file_names
    collect_task_list[[jj]] <- .set_file_name(
      object = collect_task_list[[jj]],
      file_paths = file_paths
    )
  }

  # train and variable importance tasks ----------------------------------------
  task_list <- .generate_trainer_tasks(
    experiment_data = experiment_data,
    learners = learners,
    vimp_methods = vimp_methods,
    file_paths = file_paths,
    skip_existing = skip_existing,
    ...
  )

  return(c(
    task_list,
    evaluate_task_list,
    collect_task_list
  ))
}



.run_evaluation <- function(
    cl,
    tasks,
    message_indent = 0L,
    verbose,
    ...
) {
  
  # Check that any tasks are available for processing.
  if (is_empty(tasks$evaluate) || is_empty(tasks$collect)) return(invisible(FALSE))
  
  # Determine which evaluation tasks need to be performed.
  finished_tasks <- sapply(tasks$evaluate, .file_exists)
  unfinished_tasks <- tasks$evaluate[!finished_tasks]
  finished_tasks <- tasks$evaluate[finished_tasks]
  
  # Process any unfinished tasks.
  if (length(unfinished_tasks) > 0L) {
    ..run_evaluation(
      cl = cl,
      tasks = unfinished_tasks,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
  }
  
  # Determine which collection tasks are required.
  finished_tasks <- sapply(tasks$collect, .file_exists)
  unfinished_tasks <- tasks$collect[!finished_tasks]
  finished_tasks <- tasks$collect[finished_tasks]
  
  # Process any unfinished tasks.
  if (length(unfinished_tasks) > 0L) {
    ..run_collection(
      tasks = unfinished_tasks,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
  }
  
  return(invisible(TRUE))
}



..run_evaluation <- function(
    tasks,
    cl,
    settings,
    message_indent = 0L,
    verbose,
    ...
) {
  
  # Message that evaluation is starting.
  logger_message(
    paste0(
      "Evaluation: Starting model evaluation."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  # Set outer vs. inner loop parallelisation.
  if (settings$eval$do_parallel %in% c("TRUE", "inner")) {
    cl_inner <- cl
    cl_outer <- NULL
    
  } else if (settings$eval$do_parallel %in% c("outer")) {
    cl_inner <- NULL
    cl_outer <- cl
    
    logger_message(
      paste0(
        "Evaluation: Parallel processing is done in the outer loop. ",
        "No progress can be displayed."
      ),
      indent = message_indent,
      verbose = verbose && !is.null(cl_outer)
    )
    
  } else {
    cl_inner <- cl_outer <- NULL
  }
  
  fam_mapply_lb(
    cl = cl_outer,
    assign = "all",
    FUN = .perform_task,
    progress_bar = !is.null(cl_outer),
    object = tasks,
    MoreArgs = list(
      "data" = NULL,
      "return_results" = FALSE,
      "settings" = settings,
      "message_indent" = message_indent + 1L,
      "verbose" = verbose && is.null(cl_outer),
      "cl" = cl_inner,
      ...
    )
  )

  # Message that models were evaluated.
  logger_message(
    paste0(
      "Evaluation: Models were evaluated.\n"
    ),
    indent = message_indent,
    verbose = verbose
  )
}



..run_collection <- function(
  tasks,
  message_indent = 0L,
  verbose,
  ...
) {
  
  # Message that evaluation is starting.
  logger_message(
    paste0(
      "Evaluation: Starting collection of evaluation datasets."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  fam_mapply_lb(
    cl = NULL,
    assign = "all",
    FUN = .perform_task,
    progress_bar = verbose,
    object = tasks,
    MoreArgs = list(
      "data" = NULL,
      "return_results" = FALSE,
      "message_indent" = message_indent + 1L,
      "verbose" = verbose,
      ...
    )
  )
  
  # Message that variable importances have been computed.
  logger_message(
    paste0(
      "Evaluation: Evaluation datasets were collected.\n"
    ),
    indent = message_indent,
    verbose = verbose
  )
}



.run_export <- function(
    tasks,
    message_indent = 0L,
    verbose,
    file_paths
) {
  # Message start of export.
  logger_message(
    paste0(
      "Export: exporting collected evaluation data to plots and tables."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  for (task in tasks$collect) {
    # Set path for directory where data are exported to.
    dir_path <- file.path(
      file_paths$results_dir,
      tools::file_path_sans_ext(basename(task@file))
    )
    
    # Load collection locally.
    fam_collection <- load_familiar_object(task@file)
    
    # Export tables.
    export_all(
      object = fam_collection,
      dir_path = dir_path
    )
    
    # Export plots.
    plot_all(
      object = fam_collection,
      dir_path = dir_path
    )
  }
  
  logger_message(
    paste0(
      "Export: all plots and tables were exported to ",
      file_paths$results_dir,
      ".\n"
    ),
    indent = message_indent,
    verbose = verbose
  )
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarDataElementIndividualConditionalExpectation object -------------------
setClass(
  "familiarDataElementIndividualConditionalExpectation",
  contains = "familiarDataElement"
)



# familiarDataElementPartialDependence object ----------------------------------
setClass(
  "familiarDataElementPartialDependence",
  contains = "familiarDataElement"
)



# extract_ice (generic) --------------------------------------------------------

#'@title Internal function to extract data for individual conditional
#'  expectation plots.
#'
#'@description Computes data for individual conditional expectation plots and
#'  partial dependence plots for the model(s) in a `familiarEnsemble` object.
#'
#'@param features Names of the feature or features (2) assessed simultaneously.
#'  By default `NULL`, which means that all features are assessed one-by-one.
#'@param feature_x_range When one or two features are defined using `features`,
#'  `feature_x_range` can be used to set the range of values for the first
#'  feature. For numeric features, a vector of two values is assumed to indicate
#'  a range from which `n_sample_points` are uniformly sampled. A vector of more
#'  than two values is interpreted as is, i.e. these represent the values to be
#'  sampled. For categorical features, values should represent a (sub)set of
#'  available levels.
#'@param feature_y_range As `feature_x_range`, but for the second feature in
#'  case two features are defined.
#'@param n_sample_points Number of points used to sample continuous features.
#'@inheritParams .extract_data
#'
#'@return A data.table containing individual conditional expectation plot data.
#'@md
#'@keywords internal
setGeneric(
  "extract_ice",
  function(
    object,
    data,
    cl = NULL,
    features = NULL,
    feature_x_range = NULL,
    feature_y_range = NULL,
    n_sample_points = 20L,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    sample_limit = waiver(),
    n_important_features = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_ice")
  }
)



# extract_ice (familiarEnsemble) -----------------------------------------------
setMethod(
  "extract_ice",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    features = NULL,
    feature_x_range = NULL,
    feature_y_range = NULL,
    n_sample_points = 20L,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    sample_limit = waiver(),
    n_important_features = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    
    # Message extraction start
    if (is.null(features)) {
      logger_message(
        paste0(
          "Computing individual conditional expectation and partial dependence ",
          "data for features in the dataset."
        ),
        indent = message_indent,
        verbose = verbose
      )
      
    } else {
      logger_message(
        paste0(
          "Computing individual conditional expectation and partial dependence ",
          "data for the selected features: ", paste_s(features), "."
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    # Load evaluation_times from the object settings attribute, if it is not provided.
    if (is.waive(evaluation_times)) evaluation_times <- object@settings$eval_times
    
    # Check evaluation_times argument
    if (object@outcome_type %in% c("survival")) {
      sapply(
        evaluation_times,
        .check_number_in_valid_range,
        var_name = "evaluation_times",
        range = c(0.0, Inf),
        closed = c(FALSE, TRUE)
      )
    }
    
    # Check n_sample_points argument
    .check_number_in_valid_range(
      x = n_sample_points,
      var_name = "n_sample_points",
      range = c(1L, Inf)
    )
    
    # Obtain ensemble method from stored settings, if required.
    if (is.waive(ensemble_method)) ensemble_method <- object@settings$ensemble_method
    
    # Check ensemble_method argument
    .check_parameter_value_is_valid(
      x = ensemble_method,
      var_name = "ensemble_method",
      values = .get_available_ensemble_prediction_methods()
    )
    
    # Load confidence alpha from object settings attribute if not provided
    # externally.
    if (is.waive(confidence_level)) confidence_level <- object@settings$confidence_level
    
    # Check confidence_level input argument
    .check_number_in_valid_range(
      x = confidence_level,
      var_name = "confidence_level",
      range = c(0.0, 1.0),
      closed = c(FALSE, FALSE)
    )
    
    # Load the bootstrap method
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- object@settings$bootstrap_ci_method
    
    .check_parameter_value_is_valid(
      x = bootstrap_ci_method,
      var_name = "bootstrap_ci_method",
      values = .get_available_bootstrap_confidence_interval_methods()
    )
    
    # Check the sample limit.
    sample_limit <- .parse_sample_limit(
      x = sample_limit,
      object = object,
      default = Inf,
      data_element = "ice_data"
    )
    
    # Check the level detail.
    detail_level <- .parse_detail_level(
      x = detail_level,
      object = object,
      default = "hybrid",
      data_element = "ice_data"
    )
    
    # Check the estimation type.
    estimation_type <- .parse_estimation_type(
      x = estimation_type,
      object = object,
      default = "bootstrap_confidence_interval",
      data_element = "ice_data",
      detail_level = detail_level,
      has_internal_bootstrap = FALSE
    )
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "ice_data"
    )
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Test if any model in the ensemble was successfully trained.
    if (!model_is_trained(object = object)) return(NULL)
    
    important_features <- NULL
    if (is.null(features)) {
      # Check the number of important features.
      n_important_features <- .parse_n_important_features(
        x = n_important_features,
        object = object,
        default = 20,
        data_element = "ice_data"
      )
      
      # Set features to be assessed using ICE, which are the most important
      # features.
      important_features <- .select_important_features(
        object = object,
        data = data,
        n_important_features = n_important_features
      )
    }
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "familiarDataElementIndividualConditionalExpectation",
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    # Generate elements to send to dispatch.
    ice_data <- extract_dispatcher(
      FUN = .extract_ice,
      has_internal_bootstrap = FALSE,
      cl = cl,
      object = object,
      data = data,
      features = features,
      feature_x_range = feature_x_range,
      feature_y_range = feature_y_range,
      sample_limit = sample_limit,
      important_features = important_features,
      n_sample_points = n_sample_points,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      aggregate_results = TRUE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(ice_data)
  }
)



# extract_ice (prediction table) -----------------------------------------------
setMethod(
  "extract_ice",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("individual conditional expectation")
    
    return(NULL)
  }
)



.extract_ice <- function(
    object,
    data,
    proto_data_element,
    evaluation_times = NULL,
    features = NULL,
    sample_limit,
    important_features = NULL,
    aggregate_results,
    n_models,
    is_pre_processed = FALSE,
    cl,
    message_indent = 0L,
    verbose = FALSE,
    progress_bar = FALSE,
    ...
) {
  
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  proto_data_element <- add_model_name(proto_data_element, object = object)
  
  # Retrieve input data.
  data <- process_input_data(
    object = object,
    data = data,
    stop_at = "signature",
    is_pre_processed = is_pre_processed
  )
  
  # Maintain only important features. The current set is based on the
  # required features.
  data <- filter_features(
    data = data,
    available_features = object@model_features
  )
  
  # Check if the input data is not empty
  if (is_empty(data)) return(NULL)
  
  # Select samples up to sample_limit.
  data <- get_subsample(
    data = data,
    size = sample_limit,
    seed = 0L
  )
  
  # Aggregate data.
  data <- aggregate_data(data)
  
  # Add evaluation time as a identifier to the data element.
  if (length(evaluation_times) > 0L && object@outcome_type == "survival") {
    data_elements <- add_data_element_identifier(
      x = proto_data_element, 
      evaluation_time = evaluation_times
    )
    
  } else {
    data_elements <- list(proto_data_element)
  }
  
  if (!is.null(features)) {
    
    # Check that the features exist in the data set.
    if (!all(features %in% c(object@model_features))) {
      ..warning(paste0(
        "Data for individual conditional expectation or partial dependence plots ",
        "could not be computed for ",
        paste_s(setdiff(features, object@model_features)),
        " feature(s) as they are not used by the model."
      ))
    }
    
    # Add features as identifier.
    if (length(features) == 1L) {
      data_elements <- add_data_element_identifier(
        x = data_elements,
        feature_x = features
      )
      
    } else if (length(features) == 2L) {
      
      if (length(unique(features)) != 2L) {
        ..error(paste0(
          "Data for individual conditional expectation or partial dependence plots ",
          "could not be computed as the provided features are not unique: ",
          paste_s(features), "."
        ))
      }
      
      data_elements <- add_data_element_identifier(
        x = data_elements,
        feature_x = features[1L]
      )
      data_elements <- add_data_element_identifier(
        x = data_elements,
        feature_y = features[2L]
      )
      
    } else {
      ..error(paste0(
        "Data for individual conditional expectation or partial dependence plots cannot ",
        "be computed for more than 2 features simultaneously. Found: ",
        paste_s(features), "."
      ))
    }
    
  } else {
    # Add important features as identifier.
    data_elements <- add_data_element_identifier(
      x = data_elements,
      feature_x = important_features
    )
  }
  
  # Iterate over elements.
  data_elements <- fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = ..extract_ice_data,
    data_element = data_elements,
    MoreArgs = c(
      list(
        "data" = data,
        "object" = object,
        "verbose" = verbose && !progress_bar && n_models == 1L,
        "message_indent" = message_indent
      ),
      list(...)
    ),
    progress_bar = progress_bar && verbose,
    chopchop = TRUE
  )
  
  # Flatten list of data elements.
  data_elements <- unlist(data_elements)
  if (!is.list(data_elements)) data_elements <- list(data_elements)
  
  # Merge data elements
  data_elements <- merge_data_elements(data_elements)
  
  if (aggregate_results) data_elements <- .compute_data_element_estimates(x = data_elements)
  
  return(data_elements)
}



..extract_ice_data <- function(
    cl = NULL,
    data_element,
    data,
    object,
    aggregate_results,
    feature_x_range = NULL,
    feature_y_range = NULL,
    n_sample_points,
    ensemble_method,
    verbose = FALSE,
    message_indent,
    ...
) {
  # Divide feature(s) into points.
  
  # Check that feature is present for the model.
  if (!data_element@identifiers$feature_x %in% names(object@feature_info)) return(NULL)
  
  # Generate range
  feature_x_range <- .create_feature_range(
    feature_info = object@feature_info,
    feature = data_element@identifiers$feature_x,
    column_type = class(data@data[[data_element@identifiers$feature_x]]),
    feature_range = feature_x_range,
    n = n_sample_points
  )
  
  # Add feature values.
  data_elements <- add_data_element_identifier(
    x = data_element,
    feature_x_value = feature_x_range
  )
  
  # Mention feature.
  message_str <- paste0(
    "Computing ICE / PD curves for \"", data_element@identifiers$feature_x, "\""
  )
  
  if (!is.null(data_element@identifiers$feature_y)) {
    # Check that feature is present for the model.
    if (!data_element@identifiers$feature_y %in% names(object@feature_info)) return(NULL)
    
    feature_y_range <- .create_feature_range(
      feature_info = object@feature_info,
      feature = data_element@identifiers$feature_y,
      column_type = class(data@data[[data_element@identifiers$feature_y]]),
      feature_range = feature_y_range,
      n = n_sample_points
    )
    
    # Add feature values.
    data_elements <- add_data_element_identifier(
      x = data_elements,
      feature_y_value = feature_y_range
    )
    
    # Mention feature.
    message_str <- c(
      message_str, 
      paste0(" and \"", data_element@identifiers$feature_y, "\"")
    )
  }
  
  # Add evaluation time.
  if (length(data_element@identifiers$evaluation_time) > 0L) {
    message_str <- c(
      message_str,
      paste0(" at time ", data_element@identifiers$evaluation_time, ".")
    )
    
  } else {
    message_str <- c(message_str, ".")
  }
  
  logger_message(
    paste0(message_str, collapse = ""),
    indent = message_indent,
    verbose = verbose
  )
  
  # Iterate over elements.
  data_elements <- lapply(
    data_elements,
    ...extract_ice_data,
    data = data,
    object = object,
    ensemble_method = ensemble_method
  )
  
  return(data_elements)
}



...extract_ice_data <- function(
    data_element,
    data,
    object,
    ensemble_method
) {
  
  # Make a local copy of the data.
  data@data <- data.table::copy(data@data)
  
  # Replace the feature indicated by feature_x by the value in feature_x_value.
  data@data[, (data_element@identifiers$feature_x) := data_element@identifiers$feature_x_value]
  
  # Replace the feature indicated by feature_y by the value in feature_y_value.
  if (!is.null(data_element@identifiers$feature_y)) {
    data@data[, (data_element@identifiers$feature_y) := data_element@identifiers$feature_y_value]
  }

  # Predict both primary outcomes and novelty
  type <- ifelse(
    object@outcome_type %in% c("survival", "competing_risk"),
    "survival_probability", 
    "default"
  )
  
  # Compute performance data.
  prediction_data <- .predict(
    object = object,
    data = data,
    ensemble_method = ensemble_method,
    time = data_element@identifiers$evaluation_time,
    type = type,
    aggregate_results = TRUE
  )

  # Compute novelty values.
  novelty_data <- .predict(
    object = object,
    data = data,
    ensemble_method = ensemble_method,
    time = data_element@identifiers$evaluation_time,
    type = "novelty",
    aggregate_results = TRUE
  )
  
  # Check that valid prediction data were generated.
  if (!any_predictions_valid(prediction_data)) return(NULL)
  prediction_data <- .drop_reference_data(prediction_data)
  prediction_data <- .merge_slots_into_data(prediction_data)
  prediction_data <- remove_invalid_predictions(prediction_data)
  
  # Check if removing invalid predictions leaves any data.
  if (is_empty(prediction_data)) return(NULL)
  
  novelty_data <- .drop_reference_data(novelty_data)
  novelty_data <- .merge_slots_into_data(novelty_data)
  novelty_data <- remove_invalid_predictions(novelty_data)
  
  if (object@outcome_type %in% c("binomial", "multinomial")) {
    
    # Determine class levels.
    class_levels <- get_outcome_class_levels(object)
    
    if (object@outcome_type == "binomial") {
      used_class_levels <- class_levels[2L]
      prediction_data@value_column <- c(
        setdiff(prediction_data@value_column, class_levels),
        used_class_levels
      )
      
    } else {
      used_class_levels <- class_levels
    }
    
    # Make sure that probability is returned.
    prediction_data <- .convert_value_to_grouping_column(
      prediction_data,
      new_grouping_column = used_class_levels, 
      new_grouping_column_name = "positive_class", 
      new_value_column_name = "probability"
    )
  }
  
  # Create ice and pd plot data.
  data_elements <- .create_ice_and_pd_objects(
    data_element,
    prediction_data = prediction_data,
    novelty_data = novelty_data
  )
  
  return(data_elements)
}



.create_feature_range <- function(
    feature_info,
    feature, 
    feature_range, 
    n, 
    column_type
) {
  
  # Find the feature information associated with the feature.
  feature_info <- feature_info[[feature]]
  
  # Check that the feature info is present.
  if (is.null(feature_info)) {
    ..error(
      paste0("Feature information could not be found for the ", feature, " feature.")
    )
  }
  
  # Determine if the feature is categorical or numerical.
  if (feature_info@feature_type == "factor") {
    
    # Get the levels.
    feature_levels <- feature_info@levels
    
    if (is.null(feature_range)) {
      feature_range <- factor(feature_levels, levels = feature_levels)
      
    } else if (!all(feature_range %in% feature_levels)) {
      ..error(paste0(
        "One or more levels defined in the feature range for creating ",
        "individual conditional expectation and partial dependence plots do not ",
        "match levels found in training data: ",
        paste_s(setdiff(feature_range, feature_levels)),
        ". Check for misspelled levels."
      ))
    }
    # If not null, and no mismatches occur, use feature_range directly.
    
  } else if (feature_info@feature_type == "numeric") {
    
    if (!is.null(feature_range)) {
      # Check that values are numeric.
      if (!is.numeric(feature_range)) {
        ..error(paste0(
          "Numeric values are required to define the feature range for creating ",
          "individual conditional expectation and partial dependence plots."
        ))
      } 
      
      # Check that all values are finite.
      if (!all(is.finite(feature_range))) {
        ..error(paste0(
          "Numeric values for creating individual conditional expectation and partial dependence ",
          " plots should be finite. NA and infinite values are not allowed."
        ))
      }
        
      # Sort values.
      feature_range <- sort(unique(feature_range))
      
      # If two values are defined, interpret as range, and sample from it.
      # If not, use feature_range directly.
      if (length(feature_range) == 2L) {
        feature_range <- stats::approx(
          x = c(0.00, 1.00),
          y = feature_range,
          n = n,
          method = "linear"
        )$y
      }
      
    } else {
      # Create the range of values from the feature distribution.
      feature_range <- stats::spline(
        x = (seq_along(feature_info@distribution$pctl) - 1L) / 
          (length(feature_info@distribution$pctl) - 1L),
        y = as.numeric(feature_info@distribution$pctl),
        xout = get_percentiles(n),
        method = "hyman"
      )$y
    }
    
    # Convert to integer if required.
    if (any(column_type == "integer")) feature_range <- as.integer(feature_range)
    
    # Select unique values.
    feature_range <- unique(feature_range)
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".create_feature_range: encountered unknown feature type (",
      feature_info@feature_type, ") for the ", feature, " feature."
    ))
  }
  
  return(feature_range)
}



.create_ice_and_pd_objects <- function(
    data_element,
    prediction_data,
    novelty_data
) {
  # Create ice and pd data elements.
  ice_data_element <- data_element

  if (is_empty(novelty_data)) {
    ice_data_element@data <- data.table::copy(.as_data_table(prediction_data))
    ice_data_element@grouping_column <- prediction_data@grouping_column
    ice_data_element@value_column <- prediction_data@value_column
    
  } else {
    data <- merge(
      x = data.table::copy(.as_data_table(prediction_data)),
      y = data.table::copy(.as_data_table(novelty_data)),
      by = intersect(prediction_data@grouping_column, novelty_data@grouping_column),
      all = TRUE
    )
    
    ice_data_element@data <- data
    ice_data_element@grouping_column <- union(prediction_data@grouping_column, novelty_data@grouping_column)
    ice_data_element@value_column <- union(prediction_data@value_column, novelty_data@value_column)
  }
  
  # Update pd data element.
  pd_data_element <- .create_pd_object(ice_data_element)
  
  return(list(ice_data_element, pd_data_element))
}
  


.create_pd_object <- function(ice_data_element) {
  
  # Create partial dependence data.
  pd_data_element <- methods::new(
    "familiarDataElementPartialDependence",
    ice_data_element
  )
  
  # Select grouping columns.
  grouping_columns <- setdiff(
    ice_data_element@grouping_column, get_id_columns()
  )
  
  if (length(grouping_columns) == 0L) grouping_columns <- NULL
  pd_data_element@grouping_column <- grouping_columns
  
  # Average data.
  if (length(grouping_columns) > 0L) {
    pd_data_element@data <- pd_data_element@data[
      ,
      lapply(.SD, mean, na.rm = TRUE),
      .SDcols = ice_data_element@value_column,
      by = c(grouping_columns)
    ]
    
  } else {
    pd_data_element@data <- pd_data_element@data[
      ,
      lapply(.SD, mean, na.rm = TRUE),
      .SDcols = ice_data_element@value_column
    ]
  }
  
  return(pd_data_element)
}



.update_ice_and_pd_output <- function(
    ice_data,
    pd_data,
    outcome_type,
    class_levels = NULL,
    anchor_values = NULL,
    n_samples = NULL,
    seed
) {
  
  # Prevent NOTES
  positive_class <- NULL
  
  if (is_empty(ice_data)) {
    return(list(
      "ice_data" = ice_data,
      "pd_data" = pd_data
    ))
  } 
  
  # Find anchor value for the x-feature. It will be NULL if the current feature
  # does does not appear in anchor_values.
  x_anchor <- tryCatch(
    anchor_values[[ice_data@identifiers$feature_x]],
    error = function(err) (return(NULL))
  )
  
  if (!is.null(x_anchor)) {
    if (length(x_anchor) > 1L) {
      ..error(paste0(
        "Only a single value can be provided as an anchor value for the ",
        ice_data@identifiers$feature_x, " feature."
      ))
    } 
  }
  
  # Find anchor value for the y-feature. It will be NULL if the current feature
  # does does not appear in anchor_values, or there is not y-feature associated
  # with the data.
  y_anchor <- tryCatch(
    anchor_values[[ice_data@identifiers$feature_y]],
    error = function(err) (return(NULL))
  )
  
  if (!is.null(y_anchor)) {
    if (length(y_anchor) > 1L) {
      ..error(paste0(
        "Only a single value can be provided as an anchor value for the ",
        ice_data@identifiers$feature_y, " feature."
      ))
    }
  }
  
  # Rename main value column to a consist name.
  if (outcome_type %in% c("binomial", "multinomial")) {
    old_value_column <- "probability"
    
  } else if (outcome_type %in% c("continuous", "survival")) {
    old_value_column <- "predicted_outcome"
    
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }
  
  # Update confidence interval names.
  if (ice_data@estimation_type %in% c("bci", "bootstrap_confidence_interval")) {
    old_value_column <- paste0(old_value_column, c("", "_ci_low", "_ci_up"))
    new_value_column <- c("value", "value_ci_low", "value_ci_up")
    
  } else {
    new_value_column <- "value"
  }
  
  # Replace column names.
  data.table::setnames(
    x = ice_data@data,
    old = old_value_column,
    new = new_value_column
  )
  data.table::setnames(
    x = pd_data@data, 
    old = old_value_column,
    new = new_value_column
  )
  
  # Update value column attributes.
  ice_data@value_column <- c(
    new_value_column,
    setdiff(ice_data@value_column, old_value_column)
  )
  pd_data@value_column <- c(
    new_value_column,
    setdiff(pd_data@value_column, old_value_column)
  )
  
  if (!is.null(x_anchor) || !is.null(y_anchor)) {
    
    # Update individual conditional expectation data.
    ice_data <- ..update_ice_and_pd_output(
      x = ice_data,
      x_anchor = x_anchor,
      y_anchor = y_anchor,
      value_column = new_value_column,
      outcome_type = outcome_type,
      anchor_values = anchor_values
    )
    
    # Update partial dependence data.
    pd_data <- .create_pd_object(ice_data)
  }
  
  if (!is.null(n_samples)) {
    # Select up to n_samples. These samples are random between experiments, but
    # fixed within one, as we explicitly set the random seed to be used.
    cropped_ice_data <- lapply(
      split(ice_data@data, by = "data_set", drop = TRUE),
      ..restrict_ice_samples,
      n_samples = n_samples,
      seed = seed
    )
    
    # Replace data attribute with the limited sample list.
    ice_data@data <- data.table::rbindlist(
      cropped_ice_data,
      use.names = TRUE
    )
  }
  
  # Retain only positive class for binomial outcomes.
  if (outcome_type == "binomial") {
    selected_class <- tail(class_levels, n = 1L)
    ice_data@data <- ice_data@data[positive_class == selected_class]
    pd_data@data <- pd_data@data[positive_class == selected_class]
  }
  
  return(list(
    "ice_data" = ice_data,
    "pd_data" = pd_data
  ))
}



..update_ice_and_pd_output <- function(
    x,
    x_anchor = NULL,
    y_anchor = NULL, 
    value_column, 
    outcome_type, 
    anchor_values = NULL
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  feature_x_value <- feature_x_value <- feature_y_value <- value <- NULL
  
  # Subtract anchor values.
  if (!is.null(x_anchor) && !is.null(y_anchor)) {
  
    # Set grouping columns.
    grouping_columns <- setdiff(x@grouping_column, c("feature_x_value", "feature_y_value"))
    
    # Subtract anchor value for 2D plots with x and y anchors.
    if (length(grouping_columns) > 0L) {
      x@data[
        ,
        (value_column) := lapply(
          .SD,
          ..anchor_ice_values_2D,
          x = feature_x_value,
          x_anchor = x_anchor,
          y = feature_y_value,
          y_anchor = y_anchor,
          value_offset = value,
          x_name = x@identifiers$feature_x,
          y_name = x@identifiers$feature_y
        ),
        by = c(grouping_columns),
        .SDcols = value_column
      ]
      
    } else {
      x@data[
        ,
        (value_column) := lapply(
          .SD,
          ..anchor_ice_values_1D,
          x = feature_x_value,
          x_anchor = x_anchor,
          y = feature_y_value,
          y_anchor = y_anchor,
          value_offset = value,
          x_name = x@identifiers$feature_x,
          y_name = x@identifiers$feature_y
        ),
        .SDcols = value_column
      ]
    }
    
  } else if (!is.null(x_anchor)) {
    # Set grouping columns.
    grouping_columns <- setdiff(x@grouping_column, c("feature_x_value"))
    
    # Subtract anchor value for x anchor.
    if (length(grouping_columns) > 0L) {
      x@data[
        ,
        (value_column) := lapply(
          .SD,
          ..anchor_ice_values_1D,
          x = feature_x_value,
          x_anchor = x_anchor,
          value_offset = value,
          name = x@identifiers$feature_x
        ),
        by = c(grouping_columns),
        .SDcols = value_column
      ]
      
    } else {
      x@data[
        ,
        (value_column) := lapply(
          .SD,
          ..anchor_ice_values_1D,
          x = feature_x_value,
          x_anchor = x_anchor,
          value_offset = value,
          name = x@identifiers$feature_x
        ),
        .SDcols = value_column
      ]
    }
    
  } else if (!is.null(y_anchor)) {
    # Set grouping columns.
    grouping_columns <- setdiff(x@grouping_column, c("feature_y_value"))
    
    # Subtract anchor value for y anchor.
    if (length(grouping_columns) > 0L) {
      x@data[
        ,
        (value_column) := lapply(
          .SD,
          ..anchor_ice_values_1D,
          x = feature_y_value,
          x_anchor = y_anchor,
          value_offset = value,
          name = x@identifiers$feature_y
        ),
        by = c(grouping_columns),
        .SDcols = value_column
      ]
      
    } else {
      x@data[
        ,
        (value_column) := lapply(
          .SD,
          ..anchor_ice_values_1D,
          x = feature_y_value,
          x_anchor = y_anchor,
          value_offset = value,
          name = x@identifiers$feature_y
        ),
        .SDcols = value_column
      ]
    }
    
  } else {
    # No anchor values were set.
    return(x)
  }
  
  return(x)
}



..restrict_ice_samples <- function(x, n_samples, seed = NULL) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  sample <- .NATURAL <- NULL
  
  if (is.null(n_samples)) return(x)

  # Check if the number of samples allows for sampling.
  n_samples_present <- get_n_samples(x, id_depth = "sample")
  if (n_samples_present < n_samples) n_samples <- n_samples_present
  
  # Select samples.
  selected_samples <- fam_sample(
    x = x, 
    size = n_samples, 
    replace = FALSE, 
    seed = seed
  )
  
  if (is_empty(selected_samples)) return(NULL)
  
  # Select only selected samples
  x <- x[selected_samples, on = .NATURAL]
  
  return(x)
}



..anchor_ice_values_1D <- function(
    value_in, 
    x, 
    x_anchor, 
    value_offset, 
    name
) {
  
  # Check anchor value.
  x_anchor <- .check_anchor_value(
    x = x,
    x_anchor = x_anchor,
    name = name
  )
  
  if (is.numeric(x)) {
    # Set anchor value
    if (x_anchor %in% x) {
      anchor_value <- value_offset[x == x_anchor]
        
    } else {
      # Interpolate.
      anchor_value <- stats::spline(
        x = x,
        y = value_offset,
        xout = x_anchor
      )$y
    }
  } else {
    # Set anchor value by selecting the value corresponding to x_anchor.
    anchor_value <- value_offset[x == x_anchor]
  }
  
  # Subtract the anchor value.
  return(value_in - anchor_value)
}



..anchor_ice_values_2D <- function(
    value_in, 
    x,
    x_anchor, 
    y, 
    y_anchor, 
    value_offset, 
    x_name, 
    y_name
) {
  # This is a bit more tricky since we need to do a surface interpolation which
  # R does not support out of the box. The akima package has a weird licence.
  
  # Check anchor values.
  x_anchor <- .check_anchor_value(
    x = x,
    x_anchor = x_anchor, 
    name = x_name
  )
  y_anchor <- .check_anchor_value(
    x = y, 
    x_anchor = y_anchor, 
    name = y_name
  )
  
  if (x_anchor %in% x && y_anchor %in% y) {
    # Simple - no interpolation needed.
    anchor_value <- value_offset[x == x_anchor & y == y_anchor]
    
  } else if (x_anchor %in% x) {
    # Interpolation in y.
    anchor_value <- stats::spline(
      x = y[x == x_anchor],
      y = value_offset[x == x_anchor],
      xout = y_anchor
    )$y
    
  } else if (y_anchor %in% y) {
    # Interpolation in x.
    anchor_value <- stats::spline(
      x = x[y == y_anchor],
      y = value_offset[y == y_anchor],
      xout = x_anchor
    )$y
    
  } else {
    require_package(
      x = "ranger",
      purpose = "to anchor ICE/PD plot curves"
    )
    
    # Interpolation using random forest.
    model <- ranger::ranger(
      value ~ x + y,
      data = data.table::data.table(
        "x" = x,
        "y" = y,
        "value" = value_offset
      ),
      num.trees = 100L,
      num.threads = 1L,
      seed = 1L
    )
    
    anchor_value <- predict(
      model,
      data.table::data.table(
        "x" = x_anchor,
        "y" = y_anchor
      ),
      num.threads = 1L
    )$predictions
  }
  
  # Subtract the anchor value.
  return(value_in - anchor_value)
}



.check_anchor_value <- function(x, x_anchor, name) {
  # Checks the anchor value, and updates it if needed.
  
  if (is.numeric(x)) {
    if (!is.numeric(x_anchor)) {
      ..error(paste0("Anchor value of the ", name, " feature should be numeric."))
    }
    
    # Check for out-of-range anchor values.
    if (x_anchor < min(x) || x_anchor > max(x)) {
      x_anchor_old <- x_anchor
      
      if (x_anchor < min(x)) x_anchor <- min(x)
      if (x_anchor > max(x)) x_anchor <- max(x)
      
      ..warning(paste0(
        "Anchor value (", x_anchor_old, ") of the ", name,
        " feature lies outside computed range. ",
        "The nearest value (", x_anchor, ") is used instead."
      ))
    }
    
  } else if (is.factor(x)) {
    
    # Check that the anchor value appears in the levels of x.
    if (!x_anchor %in% levels(x)) {
      ..error(paste0(
        "Anchor value (", x_anchor, ") of the ", name,
        " feature was not found among the computed levels (",
        paste_s(levels(x)), ")."
      ))
    }
    
  } else {
    if (!x_anchor %in% x) {
      ..error(paste0(
        "Anchor value (", x_anchor, ") of the ", name,
        " feature was not found among the available values."
      ))
    }
  }
  
  return(x_anchor)
}



# export_ice_data (generic) ----------------------------------------------------

#'@title Extract and export individual conditional expectation data.
#'
#'@description Extract and export individual conditional expectation data.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams extract_ice
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_ice_data
#'@md
#'@rdname export_ice_data-methods
setGeneric(
  "export_ice_data",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_ice_data")
  }
)



# export_ice_data (collection) -------------------------------------------------

#'@rdname export_ice_data-methods
setMethod(
  "export_ice_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Obtain individual conditional expectation plots.
    return(.export(
      x = object,
      data_slot = "ice_data",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "explanation",
      subtype = "ice",
      object_class = "familiarDataElementIndividualConditionalExpectation",
      export_collection = export_collection
    ))
  }
)



# export_ice_data (general) ----------------------------------------------------

#'@rdname export_ice_data-methods
setMethod(
  "export_ice_data",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "ice_data",
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_ice_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)



# .export (familiarDataElementIndividualConditionalExpectation) ----------------
setMethod(
  ".export",
  signature(x = "familiarDataElementIndividualConditionalExpectation"),
  function(
    x,
    x_list, 
    aggregate_results = FALSE,
    ...
  ) {

    if (aggregate_results) {
      x_list <- .compute_data_element_estimates(x_list)
    }
    
    # Determine identifiers that should be merged. Since the feature values of
    # the x and y features may be different (e.g. numeric and factor), merging
    # them would cause features values to merged incorrectly.
    merging_identifiers <- setdiff(names(x@identifiers), c("feature_x", "feature_y"))
    
    # Merge data elements.
    x <- merge_data_elements(
      x = x_list,
      as_data = merging_identifiers,
      as_grouping_column = TRUE,
      force_data_table = TRUE
    )
    
    return(x)
  }
)



# export_partial_dependence_data (generic) -------------------------------------

#'@title Extract and export partial dependence data.
#'
#'@description Extract and export partial dependence data.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams extract_ice
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_partial_dependence_data
#'@md
#'@rdname export_partial_dependence_data-methods
setGeneric(
  "export_partial_dependence_data",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_partial_dependence_data")
  }
)



# export_partial_dependence_data (collection) ----------------------------------

#'@rdname export_partial_dependence_data-methods
setMethod(
  "export_partial_dependence_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Obtain partial dependence
    return(.export(
      x = object,
      data_slot = "ice_data",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "explanation",
      subtype = "pd",
      object_class = "familiarDataElementPartialDependence",
      export_collection = export_collection
    ))
  }
)



# export_partial_dependence_data (general) -------------------------------------

#'@rdname export_partial_dependence_data-methods
setMethod(
  "export_partial_dependence_data",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "ice_data",
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_partial_dependence_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)


# .export (familiarDataElementPartialDependence) -------------------------------
setMethod(
  ".export",
  signature(x = "familiarDataElementPartialDependence"),
  function(
    x,
    x_list,
    aggregate_results = FALSE,
    ...
  ) {
    
    if (aggregate_results) {
      x_list <- .compute_data_element_estimates(x_list)
    }
    
    # Determine identifiers that should be merged. Since the feature values of
    # the x and y features may be different (e.g. numeric and factor), merging
    # them would cause features values to merged incorrectly.
    merging_identifiers <- setdiff(names(x@identifiers), c("feature_x", "feature_y"))
    
    # Merge data elements.
    x <- merge_data_elements(
      x = x_list,
      as_data = merging_identifiers,
      as_grouping_column = TRUE,
      force_data_table = TRUE
    )
    
    return(x)
  }
)

# Save method for model saving. The save function appears in base and is always
# explicity imported.
setGeneric("save")

setGeneric(".complete", function(object, ...) standardGeneric(".complete"))

setGeneric(".copy", function(object, ...) standardGeneric(".copy"))

setGeneric(".predict", function(object, data, ...) standardGeneric(".predict"))

setGeneric(".predict_novelty", function(object, data, ...) standardGeneric(".predict_novelty"))

setGeneric(".predict_risk_stratification", function(object, data, ...) standardGeneric(".predict_risk_stratification"))

setGeneric(".train", function(object, data, ...) standardGeneric(".train"))

setGeneric("require_package", function(x, ...) standardGeneric("require_package"))

setGeneric("set_package_version", function(object, ...) standardGeneric("set_package_version"))

setGeneric("check_package_version", function(object, ...) standardGeneric("check_package_version"))

setGeneric("get_signature_features", function(object, ...) standardGeneric("get_signature_features"))

setGeneric("model_is_trained", function(object, ...) standardGeneric("model_is_trained"))

setGeneric("complete_familiar_ensemble", function(object, ...) standardGeneric("complete_familiar_ensemble"))

setGeneric("..get_model_file_path", function(ii, object, ...) standardGeneric("..get_model_file_path"))

setGeneric("..get_model", function(ii, object, ...) standardGeneric("..get_model"))

setGeneric("..can_detach_models", function(ii, object, ...) standardGeneric("..can_detach_models"))

setGeneric("..update_model_list", function(object, ...) standardGeneric("..update_model_list"))

setGeneric("load_models", function(object, ...) standardGeneric("load_models"))

setGeneric("is_model_loaded", function(object, ...) standardGeneric("is_model_loaded"))

setGeneric("detach_models", function(object, ...) standardGeneric("detach_models"))

setGeneric("add_identifiers", function(data, object, ...) standardGeneric("add_identifiers"))

setGeneric("add_model_name", function(data, object, ...) standardGeneric("add_model_name"))

setGeneric("get_object_name", function(object, ...) standardGeneric("get_object_name"))

setGeneric("process_input_data", function(object, data, ...) standardGeneric("process_input_data"))

setGeneric("load_delayed_data", function(data, object, ...) standardGeneric("load_delayed_data"))

setGeneric("select_data_from_samples", function(data, samples, ...) standardGeneric("select_data_from_samples"))

setGeneric("filter_features", function(data, ...) standardGeneric("filter_features"))

setGeneric("filter_missing_outcome", function(data, ...) standardGeneric("filter_missing_outcome"))

setGeneric("filter_bad_samples", function(data, ...) standardGeneric("filter_bad_samples"))

setGeneric("transform_features", function(data, ...) standardGeneric("transform_features"))

setGeneric("normalise_features", function(data, ...) standardGeneric("normalise_features"))

setGeneric("batch_normalise_features", function(data, ...) standardGeneric("batch_normalise_features"))

setGeneric("impute_features", function(data, ...) standardGeneric("impute_features"))

setGeneric("cluster_features", function(data, ...) standardGeneric("cluster_features"))

setGeneric("aggregate_data", function(data, ...) standardGeneric("aggregate_data"))

setGeneric("select_unique_data", function(data, ...) standardGeneric("select_unique_data"))

setGeneric("select_features", function(data, ...) standardGeneric("select_features"))

setGeneric("preprocess_data", function(data, object, ...) standardGeneric("preprocess_data"))

setGeneric("postprocess_data", function(data, object, ...) standardGeneric("postprocess_data"))

setGeneric("update_with_replacement", function(data, ...) standardGeneric("update_with_replacement"))

setGeneric("add_package_version", function(object, ...) standardGeneric("add_package_version"))

setGeneric("add_data_column_info", function(object, ...) standardGeneric("add_data_column_info"))

setGeneric("has_calibration_info", function(object) standardGeneric("has_calibration_info"))

setGeneric("extract_settings_from_data", function(data, ...) standardGeneric("extract_settings_from_data"))

setGeneric("set_object_parameters", function(object, ...) standardGeneric("set_object_parameters"))

setGeneric("get_required_features", function(x, ...) standardGeneric("get_required_features"))

setGeneric("get_model_features", function(x, ...) standardGeneric("get_model_features"))

setGeneric("set_model_features", function(object, ...) standardGeneric("set_model_features"))



# clustering methods -----------------------------------------------------------
setGeneric("set_similarity_table", function(object, data, ...) standardGeneric("set_similarity_table"))

setGeneric(".set_similarity_table", function(object, data, ...) standardGeneric(".set_similarity_table"))

setGeneric("get_similarity_names", function(object, ...) standardGeneric("get_similarity_names"))

setGeneric("get_distance_table", function(object, ...) standardGeneric("get_distance_table"))

setGeneric("get_distance_matrix", function(object, ...) standardGeneric("get_distance_matrix"))

setGeneric("apply_cluster_method", function(object, ...) standardGeneric("apply_cluster_method"))

setGeneric("create_clusters", function(object, ...) standardGeneric("create_clusters"))

setGeneric(".cluster_by_silhouette", function(object, ...) standardGeneric(".cluster_by_silhouette"))

setGeneric(".cluster_by_fixed_cut", function(object, ...) standardGeneric(".cluster_by_fixed_cut"))

setGeneric(".cluster_by_dynamic_cut", function(object, ...) standardGeneric(".cluster_by_dynamic_cut"))

setGeneric(".cluster_by_generic", function(object, ...) standardGeneric(".cluster_by_generic"))



# export methods ---------------------------------------------------------------
# Additional methods are found in FamiliarCollectionExport.R
setGeneric(".export_to_file", function(data, object, dir_path, ...) standardGeneric(".export_to_file"))

setGeneric(".export", function(x, ...) standardGeneric(".export"))



# conversion & loading ---------------------------------------------------------
# Additional methods are documented in FamiliarObjectConversion.R
setGeneric("load_familiar_object", function(object, ...) standardGeneric("load_familiar_object"))


# utilities --------------------------------------------------------------------
setGeneric("is_empty", function(x, ...) standardGeneric("is_empty"))

setGeneric("get_outcome_name", function(x, ...) standardGeneric("get_outcome_name"))

setGeneric("get_outcome_class_levels", function(x, ...) standardGeneric("get_outcome_class_levels"))

setGeneric("get_outcome_columns", function(x, ...) standardGeneric("get_outcome_columns"))

setGeneric("get_non_feature_columns", function(x, ...) standardGeneric("get_non_feature_columns"))

setGeneric("get_feature_columns", function(x, ...) standardGeneric("get_feature_columns"))

setGeneric("get_n_features", function(x, ...) standardGeneric("get_n_features"))

setGeneric("get_n_samples", function(x, ...) standardGeneric("get_n_samples"))

setGeneric("has_feature_data", function(x, ...) standardGeneric("has_feature_data"))

setGeneric("get_unique_row_names", function(x, ...) standardGeneric("get_unique_row_names"))

setGeneric("get_class_probability_name", function(x, ...) standardGeneric("get_class_probability_name"))

setGeneric("encode_categorical_variables", function(object, data, ...) standardGeneric("encode_categorical_variables"))

setGeneric("has_bad_training_data", function(object, data, ...) standardGeneric("has_bad_training_data"))

setGeneric("has_optimised_hyperparameters", function(object, ...) standardGeneric("has_optimised_hyperparameters"))

setGeneric("fam_sample", function(x, ...) standardGeneric("fam_sample"))

setGeneric("get_bootstrap_sample", function(data, ...) standardGeneric("get_bootstrap_sample"))

setGeneric("get_subsample", function(data, ...) standardGeneric("get_subsample"))

setGeneric("create_instance_weights", function(data, ...) standardGeneric("create_instance_weights"))

setGeneric(".create_outcome_info", function(object, ...) standardGeneric(".create_outcome_info"))



# familiarModel methods --------------------------------------------------------
setGeneric("promote_learner", function(object, ...) standardGeneric("promote_learner"))

setGeneric("get_default_hyperparameters", function(object, ...) standardGeneric("get_default_hyperparameters"))

setGeneric("..train", function(object, data, ...) standardGeneric("..train"))

setGeneric("..train_naive", function(object, data, ...) standardGeneric("..train_naive"))

setGeneric("..predict", function(object, data, ...) standardGeneric("..predict"))

setGeneric("..set_recalibration_model", function(object, data, ...) standardGeneric("..set_recalibration_model"))

setGeneric("..vimp", function(object, ...) standardGeneric("..vimp"))

setGeneric("get_prediction_type", function(object, ...) standardGeneric("get_prediction_type"))

setGeneric("..set_calibration_info", function(object, data, ...) standardGeneric("..set_calibration_info"))

setGeneric(
  "..set_risk_stratification_thresholds",
  function(object, data, ...) {
    standardGeneric("..set_risk_stratification_thresholds")
  }
)

setGeneric("..set_vimp_parameters", function(object, ...) standardGeneric("..set_vimp_parameters"))

setGeneric("..get_distribution_family", function(object, ...) standardGeneric("..get_distribution_family"))

setGeneric("..update_outcome", function(object, data, ...) standardGeneric("..update_outcome"))

setGeneric("..update_warnings", function(object, ...) standardGeneric("..update_warnings"))

setGeneric("..update_errors", function(object, ...) standardGeneric("..update_errors"))

setGeneric("optimise_hyperparameters", function(object, data, ...) standardGeneric("optimise_hyperparameters"))

setGeneric("trim_model", function(object, ...) standardGeneric("trim_model"))

setGeneric(".trim_model", function(object, ...) standardGeneric(".trim_model"))

setGeneric("requires_naive_model", function(object, ...) standardGeneric("requires_naive_model"))

setGeneric("..get_prediction_table_type", function(object, ...) standardGeneric("..get_prediction_table_type"))


# familiarNoveltyDetector methods ----------------------------------------------
setGeneric("promote_detector", function(object, ...) standardGeneric("promote_detector"))



# familiarVimpMethod methods ---------------------------------------------------
setGeneric(".vimp", function(object, ...) standardGeneric(".vimp"))

setGeneric("promote_vimp_method", function(object, ...) standardGeneric("promote_vimp_method"))

setGeneric("test_create_vimp_method", function(data, ...) standardGeneric("test_create_vimp_method"))



# familiarMetric methods -------------------------------------------------------
setGeneric("is_higher_better", function(metric, ...) standardGeneric("is_higher_better"))

setGeneric("compute_metric_score", function(metric, ...) standardGeneric("compute_metric_score"))

setGeneric("compute_objective_score", function(metric, ...) standardGeneric("compute_objective_score"))

setGeneric("set_metric_baseline_value", function(metric, ...) standardGeneric("set_metric_baseline_value"))



# familiarCollection methods ---------------------------------------------------
setGeneric(".set_labels", function(x, ...) standardGeneric(".set_labels"))

setGeneric(".get_labels", function(x, ...) standardGeneric(".get_labels"))

setGeneric(".construct_label_table", function(x, ...) standardGeneric(".construct_label_table"))

setGeneric(".apply_labels", function(data, object, ...) standardGeneric(".apply_labels"))

setGeneric("set_object_name", function(x, ...) standardGeneric("set_object_name"))

setGeneric("set_data_set_names", function(x, ...) standardGeneric("set_data_set_names"))

setGeneric("set_learner_names", function(x, ...) standardGeneric("set_learner_names"))

setGeneric("set_vimp_method_names", function(x, ...) standardGeneric("set_vimp_method_names"))

setGeneric("set_feature_names", function(x, ...) standardGeneric("set_feature_names"))

setGeneric("set_risk_group_names", function(x, ...) standardGeneric("set_risk_group_names"))

setGeneric("set_class_names", function(x, ...) standardGeneric("set_class_names"))

setGeneric("get_data_set_names", function(x, ...) standardGeneric("get_data_set_names"))

setGeneric("get_learner_names", function(x, ...) standardGeneric("get_learner_names"))

setGeneric("get_vimp_method_names", function(x, ...) standardGeneric("get_vimp_method_names"))

setGeneric("get_feature_names", function(x, ...) standardGeneric("get_feature_names"))

setGeneric("get_risk_group_names", function(x, ...) standardGeneric("get_risk_group_names"))

setGeneric("get_class_names", function(x, ...) standardGeneric("get_class_names"))

setGeneric("get_data_set_name_levels", function(x, ...) standardGeneric("get_data_set_name_levels"))

setGeneric("get_learner_name_levels", function(x, ...) standardGeneric("get_learner_name_levels"))

setGeneric("get_vimp_method_name_levels", function(x, ...) standardGeneric("get_vimp_method_name_levels"))

setGeneric("get_feature_name_levels", function(x, ...) standardGeneric("get_feature_name_levels"))

setGeneric("get_risk_group_name_levels", function(x, ...) standardGeneric("get_risk_group_name_levels"))

setGeneric("get_class_name_levels", function(x, ...) standardGeneric("get_class_name_levels"))



# familiarDataElement methods --------------------------------------------------
setGeneric("extract_dispatcher", function(object, proto_data_element, ...) standardGeneric("extract_dispatcher"))

setGeneric("identify_element_sets", function(x, ...) standardGeneric("identify_element_sets"))

setGeneric("merge_data_elements", function(x, ...) standardGeneric("merge_data_elements"))

setGeneric("add_data_element_identifier", function(x, ...) standardGeneric("add_data_element_identifier"))

setGeneric("add_data_element_bootstrap", function(x, ...) standardGeneric("add_data_element_bootstrap"))

setGeneric(".add_point_estimate_from_elements", function(x, ...) standardGeneric(".add_point_estimate_from_elements"))

setGeneric(".identifier_as_data_attribute", function(x, ...) standardGeneric(".identifier_as_data_attribute"))

setGeneric(".compute_data_element_estimates", function(x, ...) standardGeneric(".compute_data_element_estimates"))

setGeneric("..compute_data_element_estimates", function(x, ...) standardGeneric("..compute_data_element_estimates"))

setGeneric(".merge_slots_into_data", function(x, ...) standardGeneric(".merge_slots_into_data"))

setGeneric(".extract_slots_from_data", function(x, ...) standardGeneric(".extract_slots_from_data"))

setGeneric(".as_data_table", function(x, ...) standardGeneric(".as_data_table"))

setGeneric(
  ".convert_value_to_grouping_column",
  function(x, ...) {
    standardGeneric(".convert_value_to_grouping_column")
  }
)

setGeneric("collect", function(x, ...) standardGeneric("collect"))


# vimpTable methods ------------------------------------------------------------
setGeneric("decode_vimp_table", function(x, ...) standardGeneric("decode_vimp_table"))

setGeneric("decluster_vimp_table", function(x, ...) standardGeneric("decluster_vimp_table"))

setGeneric("recluster_vimp_table", function(x, ...) standardGeneric("recluster_vimp_table"))

setGeneric("rank_vimp_table", function(x, ...) standardGeneric("rank_vimp_table"))

setGeneric("preprocess_vimp_table", function(x, ...) standardGeneric("preprocess_vimp_table"))

setGeneric("remove_signature_features", function(x, ...) standardGeneric("remove_signature_features"))

setGeneric("update_vimp_table_to_reference", function(x, ...) standardGeneric("update_vimp_table_to_reference"))

setGeneric("collect_vimp_table", function(x, ...) standardGeneric("collect_vimp_table"))



# featureInfo methods ----------------------------------------------------------
setGeneric("is_available", function(object, ...) standardGeneric("is_available"))

setGeneric("is_in_signature", function(object, ...) standardGeneric("is_in_signature"))

setGeneric("is_in_novelty", function(object, ...) standardGeneric("is_in_novelty"))

setGeneric("update_removed_status", function(object, ...) standardGeneric("update_removed_status"))

setGeneric("feature_info_complete", function(object, ...) standardGeneric("feature_info_complete"))

setGeneric("add_feature_info_parameters", function(object, data, ...) standardGeneric("add_feature_info_parameters"))

setGeneric("apply_feature_info_parameters", function(object, data, ...) standardGeneric("apply_feature_info_parameters"))



# task methods -----------------------------------------------------------------
setGeneric(".set_file_name", function(object, ...) standardGeneric(".set_file_name"))

setGeneric(".file_exists", function(object, ...) standardGeneric(".file_exists"))

setGeneric(".perform_task", function(object, data, ...) standardGeneric(".perform_task"))

setGeneric(".get_task_descriptor", function(object, ...) standardGeneric(".get_task_descriptor"))

setGeneric(".get_feature_info_list", function(object, ...) standardGeneric(".get_feature_info_list"))

setGeneric(".get_hyperparameters", function(object, ...) standardGeneric(".get_hyperparameters"))

setGeneric(".get_variable_importance_table", function(object, ...) standardGeneric(".get_variable_importance_table"))

setGeneric(".get_current_run_table", function(object, ...) standardGeneric(".get_current_run_table"))

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
#' @include LearnerS4Naive.R
NULL


# as_data_object methods -------------------------------------------------------

#'@title Creates a valid data object from input data.
#'
#'@description Creates a `dataObject` object from input data. Input data can be
#'  a `data.frame` or `data.table`, a path to such tables on a local or network
#'  drive, or a path to tabular data that may be converted to these formats.
#'
#'  In addition, a `familiarEnsemble` or `familiarModel` object can be passed
#'  along to check whether the data are formatted correctly, e.g. by checking
#'  the levels of categorical features, whether all expected columns are
#'  present, etc.
#'
#'@param data A `data.frame` or `data.table`, a path to such tables on a local
#'  or network drive, or a path to tabular data that may be converted to these
#'  formats.
#'
#'@param object A `familiarEnsemble` or `familiarModel` object that is used to
#'  check consistency of these objects.
#'
#'@param check_stringency Specifies stringency of various checks. This is mostly:
#'
#'  * `strict`: default value used for `summon_familiar`. Thoroughly checks
#'  input data. Used internally for checking development data.
#'  
#'  * `external_warn`: value used for `extract_data` and related methods. Less
#'  stringent checks, but will warn for possible issues. Used internally for
#'  checking data for evaluation and explanation.
#'  
#'  * `external`: value used for external methods such as `predict`. Less
#'  stringent checks, particularly for identifier and outcome columns, which may
#'  be completely absent. Used internally for `predict`.
#'
#'@param .no_features_required Internal flag to signify that data without
#'  features is allowed. Default: FALSE (most processing steps require features).
#'
#'@inheritParams .parse_experiment_settings
#'
#'@details You can specify settings for your data manually, e.g. the column for
#'  sample identifiers (`sample_id_column`). This prevents you from having to
#'  change the column name externally. In the case you provide a `familiarModel`
#'  or `familiarEnsemble` for the `object` argument, any parameters you provide
#'  take precedence over parameters specified by the object.
#'
#'@return A `dataObject` object.
#'@exportMethod as_data_object
#'@md
#'@rdname as_data_object-methods



## as_data_object (generic) ----------------------------------------------------
setGeneric(
  "as_data_object",
  function(data, ...) standardGeneric("as_data_object")
)



## as_data_object (dataObject) -------------------------------------------------
#'@rdname as_data_object-methods
setMethod(
  "as_data_object",
  signature(data = "dataObject"),
  function(data, object = NULL, ...) {
    return(data)
  }
)



## as_data_object (data.table) -------------------------------------------------
#'@rdname as_data_object-methods
setMethod(
  "as_data_object",
  signature(data = "data.table"),
  function(
    data,
    object = NULL,
    batch_id_column = waiver(),
    sample_id_column = waiver(),
    series_id_column = waiver(),
    development_batch_id = waiver(),
    validation_batch_id = waiver(),
    outcome_name = waiver(),
    outcome_column = waiver(),
    outcome_type = waiver(),
    event_indicator = waiver(),
    censoring_indicator = waiver(),
    competing_risk_indicator = waiver(),
    class_levels = waiver(),
    exclude_features = waiver(),
    include_features = waiver(),
    reference_method = waiver(),
    check_stringency = "strict",
    .no_features_required = FALSE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    type <- outcome_event <- NULL
    
    # Determine whether the object contains data concerning columns, and
    # outcome. Note that user-provided names always take precedence.
    has_model_object <- is(object, "familiarModel") ||
      is(object, "familiarEnsemble") ||
      is(object, "familiarNoveltyDetector")
    
    # Check whether a model potentially has outcome information.
    if (has_model_object) {
      has_outcome_info_slot <- methods::.hasSlot(object, "outcome_info")
      
    } else {
      has_outcome_info_slot <- FALSE
    }
    
    if (check_stringency != "strict") {
      if (!has_model_object) ..error("Dummy columns cannot be set without a model or ensemble object.")
    } 
    
    # Attempt to identify a sample identifier column.
    if (is.waive(sample_id_column)) {
      if (has_model_object) {
        if (!is_empty(object@data_column_info)) {
          # Find the sample id column stored with the model.
          model_sample_id_column <- object@data_column_info[type == "sample_id_column"]$external
          
          # Check that the model actually has a column name (not character(0))
          # that is not NA, and set this column name.
          if (length(model_sample_id_column) > 0L) {
            if (!is.na(model_sample_id_column)) sample_id_column <- model_sample_id_column
          }
        }
      }
    }
    
    # Attempt to identify a batch identifier column.
    if (is.waive(batch_id_column)) {
      if (has_model_object) {
        if (!is_empty(object@data_column_info)) {
          # Find the batch id column stored with the model.
          model_batch_id_column <- object@data_column_info[type == "batch_id_column"]$external
          
          # Check that the model actually has a column name (not character(0))
          # that is not NA, and set this column name.
          if (length(model_batch_id_column) > 0L) {
            if (!is.na(model_batch_id_column)) batch_id_column <- model_batch_id_column
          }
        }
      }
    }
    
    # Attempt to identify a series identifier column.
    if (is.waive(series_id_column)) {
      if (has_model_object) {
        if (!is_empty(object@data_column_info)) {
          # Find the series id column stored with the model.
          model_series_id_column <- object@data_column_info[type == "series_id_column"]$external
          
          # Check that the model actually has a column name (not character(0))
          # that is not NA, and set this column name.
          if (length(model_series_id_column) > 0L) {
            if (!is.na(model_series_id_column)) series_id_column <- model_series_id_column
          }
        }
      }
    }
    
    # Development and validation batch ids are not incorporated into
    # familiarModel or familiarEnsemble objects.
    
    # Attempt to identify the type of outcome.
    if (is.waive(outcome_type)) {
      if (is(object, "familiarNoveltyDetector")) {
        outcome_type <- "unsupervised"
        
      } else if (has_model_object) {
        outcome_type <- object@outcome_type
      }
    }
    
    if (outcome_type != "unsupervised") {
      # Attempt to identify the name of the outcome.
      if (is.waive(outcome_name)) {
        if (has_model_object && has_outcome_info_slot) {
          if (is(object@outcome_info, "outcomeInfo")) {
            
            # Check that the outcome name is not empty.
            if (length(object@outcome_info@name) >= 1L) outcome_name <- object@outcome_info@name
          }
        }
      }
      
      # Attempt to identify the outcome columns.
      if (is.waive(outcome_column)) {
        if (has_model_object && has_outcome_info_slot) {
          if (!is_empty(object@data_column_info)) {
            # Find the model columns.
            outcome_column <- object@data_column_info[type == "outcome_column"]$external
          }
        }
      }
      
      # Attempt to identify the event indicator.
      if (is.waive(event_indicator)) {
        if (has_model_object && has_outcome_info_slot) {
          if (is(object@outcome_info, "outcomeInfo")) {
            if (length(object@outcome_info@event) > 0L) {
              if (!is.na(object@outcome_info@event)) {
                event_indicator <- object@outcome_info@event
              } 
            }
          }
        }
      }
      
      # Attempt to identify the censoring indicator.
      if (is.waive(censoring_indicator)) {
        if (has_model_object && has_outcome_info_slot) {
          if (is(object@outcome_info, "outcomeInfo")) {
            if (length(object@outcome_info@censored) > 0L) {
              if (!is.na(object@outcome_info@censored)) {
                censoring_indicator <- object@outcome_info@censored
              } 
            }
          }
        }
      }
      
      # Attempt to identify the competing risk indicator.
      if (is.waive(competing_risk_indicator)) {
        if (has_model_object && has_outcome_info_slot) {
          if (is(object@outcome_info, "outcomeInfo")) {
            if (length(object@outcome_info@competing_risk) > 0L) {
              if (!is.na(object@outcome_info@competing_risk)) {
                competing_risk_indicator <- object@outcome_info@competing_risk
              } 
            }
          }
        }
      }
      
      # Attempt to identify class levels of the outcome.
      if (is.waive(class_levels)) {
        if (has_model_object && has_outcome_info_slot) {
          if (is(object@outcome_info, "outcomeInfo")) {
            if (length(object@outcome_info@levels) > 0L) {
              class_levels <- object@outcome_info@levels
            }
          }
        }
      }
    }
    
    # Load settings from input.
    settings <- do.call(
      .parse_initial_settings,
      args = c(
        list(
          "experimental_design" = "fs+mb",
          "sample_id_column" = sample_id_column,
          "batch_id_column" = batch_id_column,
          "series_id_column" = series_id_column,
          "development_batch_id" = development_batch_id,
          "validation_batch_id" = validation_batch_id,
          "outcome_name" = outcome_name,
          "outcome_column" = outcome_column,
          "outcome_type" = outcome_type,
          "event_indicator" = event_indicator,
          "censoring_indicator" = censoring_indicator,
          "competing_risk_indicator" = competing_risk_indicator,
          "class_levels" = class_levels,
          "exclude_features" = exclude_features,
          "include_features" = include_features,
          "reference_method" = reference_method
        ),
        list(...)
      )
    )
    
    # Prepare data.table.
    data <- .load_data(
      data = data,
      sample_id_column = settings$data$sample_col,
      batch_id_column = settings$data$batch_col,
      series_id_column = settings$data$series_col
    )
    
    # Update settings
    settings <- .update_initial_settings(
      data = data,
      settings = settings,
      check_stringency = check_stringency
    )
    
    # Parse data
    data <- .finish_data_preparation(
      data = data,
      sample_id_column = settings$data$sample_col,
      batch_id_column = settings$data$batch_col,
      series_id_column = settings$data$series_col,
      outcome_column = settings$data$outcome_col,
      outcome_type = settings$data$outcome_type,
      include_features = settings$data$include_features,
      class_levels = settings$data$class_levels,
      censoring_indicator = settings$data$censoring_indicator,
      event_indicator = settings$data$event_indicator,
      competing_risk_indicator = settings$data$competing_risk_indicator,
      check_stringency = check_stringency,
      reference_method = settings$data$reference_method,
      .no_features_required = .no_features_required
    )
    
    # Update the dataset according to the feature info list.
    if (has_model_object) data <- update_data_set(data = data, object = object)
    
    # Add outcome information, preferentially from the familiarModel or
    # familiarEnsemble, as it is more complete.
    if (has_model_object && has_outcome_info_slot) {
      outcome_info <- object@outcome_info
      
    } else {
      if (settings$data$outcome_type %in% c("survival", "competing_risk")) {
        # Update time max.
        if (is.null(settings$eval$time_max)) {
          settings$eval$time_max <- .get_default_time_max(data[outcome_event == 1L]$outcome_time)
        }
      }
      outcome_info <- create_outcome_info(settings = settings)
    }
    
    if (has_model_object) {
      if (!is_empty(object@data_column_info)) {
        data_info <- object@data_column_info
        
      } else {
        data_info <- create_data_column_info(settings = settings)
      }
      
    } else {
      data_info <- create_data_column_info(settings = settings)
    }
    
    # Convert to dataObject
    data <- methods::new(
      "dataObject",
      data = data,
      preprocessing_level = "none",
      outcome_type = settings$data$outcome_type,
      outcome_info = outcome_info,
      data_column_info = data_info
    )
    
    return(data)
  }
)



## as_data_object (ANY) --------------------------------------------------------
#'@rdname as_data_object-methods
setMethod(
  "as_data_object",
  signature(data = "ANY"),
  function(
    data,
    object = NULL,
    sample_id_column = waiver(),
    batch_id_column = waiver(),
    series_id_column = waiver(),
    ...
  ) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    type <- NULL
    
    # Determine whether the object contains data concerning columns. Note that
    # user-provided names always take precedence.
    has_model_object <- FALSE
    if (is(object, "familiarModel") || is(object, "familiarEnsemble")) {
      if (!is_empty(object@data_column_info)) has_model_object <- TRUE
    }
    
    # Create a local copy of sample_id_column to pass on to .load_data.
    if (is.waive(sample_id_column)) {
      sample_id_column_local <- NULL
      
      if (has_model_object) {
        # Find the sample id column stored with the model.
        model_sample_id_column <- object@data_column_info[type == "sample_id_column"]$external
        
        # Check that the model actually has a column name (not character(0))
        # that is not NA, and set this column name.
        if (length(model_sample_id_column) > 0L) {
          if (!is.na(model_sample_id_column)) sample_id_column_local <- model_sample_id_column
        }
      }
      
    } else {
      sample_id_column_local <- sample_id_column
    }
    
    # Create a local copy of batch_id_column to pass on to .load_data.
    if (is.waive(batch_id_column)) {
      batch_id_column_local <- NULL
      
      if (has_model_object) {
        # Find the batch id column stored with the model.
        model_batch_id_column <- object@data_column_info[type == "batch_id_column"]$external
        
        # Check that the model actually has a column name (not character(0))
        # that is not NA, and set this column name.
        if (length(model_batch_id_column) > 0L) {
          if (!is.na(model_batch_id_column)) batch_id_column_local <- model_batch_id_column
        }
      }
      
    } else {
      batch_id_column_local <- batch_id_column
    }
    
    # Create a local copy of series_id_column to pass on to .load_data
    if (is.waive(series_id_column)) {
      series_id_column_local <- NULL
      
      if (has_model_object) {
        # Find the series id column stored with the model.
        model_series_id_column <- object@data_column_info[type == "series_id_column"]$external
        
        # Check that the model actually has a column name (not
        # character(0)) that is not NA, and set this column name.
        if (length(model_series_id_column) > 0L) {
          if (!is.na(model_series_id_column)) series_id_column_local <- model_series_id_column
        }
      }
      
    } else {
      series_id_column_local <- series_id_column
    }
    
    # Load data and convert to data.table
    data <- .load_data(
      data = data,
      sample_id_column = sample_id_column_local,
      batch_id_column = batch_id_column_local,
      series_id_column = series_id_column_local
    )
    
    # Pass on to data.table method.
    return(do.call(
      as_data_object,
      args = c(
        list(
          "data" = data,
          "object" = object,
          "sample_id_column" = sample_id_column,
          "batch_id_column" = batch_id_column,
          "series_id_column" = series_id_column
        ),
        list(...)
      )
    ))
  }
)



# extract_settings_from_data ---------------------------------------------------
setMethod(
  "extract_settings_from_data",
  signature(data = "dataObject"),
  function(
    data,
    settings = NULL,
    signature = NULL
  ) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    type <- NULL
    
    if (is.null(settings)) settings <- list("data" = list())
    
    # Placeholders
    sample_id_column <- batch_id_column <- series_id_column <- outcome_columns <- NULL
    
    if (!is_empty(data@data_column_info)) {
      # Sample identifier
      sample_id_column <- ..set_identifier_column(
        current = sample_id_column,
        data = data@data,
        internal = data@data_column_info[type == "sample_id_column"]$internal,
        external = data@data_column_info[type == "sample_id_column"]$external
      )
      
      # Batch identifier
      batch_id_column <- ..set_identifier_column(
        current = batch_id_column,
        data = data@data,
        internal = data@data_column_info[type == "batch_id_column"]$internal,
        external = data@data_column_info[type == "batch_id_column"]$external
      )
      
      # Series identifier
      series_id_column <- ..set_identifier_column(
        current = series_id_column,
        data = data@data,
        internal = data@data_column_info[type == "series_id_column"]$internal,
        external = data@data_column_info[type == "series_id_column"]$external
      )
      
      # Outcome columns
      outcome_columns <- ..set_identifier_column(
        current = outcome_columns,
        data = data@data,
        internal = data@data_column_info[type == "outcome_column"]$internal,
        external = data@data_column_info[type == "outcome_column"]$external
      )
    }
    
    if (is.null(outcome_columns)) get_outcome_columns(data)
    
    # Sample identifier column
    settings$data$sample_col <- sample_id_column
    settings$data$batch_col <- batch_id_column
    settings$data$series_col <- series_id_column
    settings$data$outcome_col <- outcome_columns
    settings$data$outcome_type <- data@outcome_type
    settings$data$outcome_name <- get_outcome_name(data@outcome_info)
    settings$data$class_levels <- get_outcome_class_levels(data@outcome_info)
    settings$data$event_indicator <- data@outcome_info@event
    settings$data$censoring_indicator <- data@outcome_info@censored
    settings$data$competing_risk_indicator <- data@outcome_info@competing_risk
    settings$data$signature <- signature
    settings$data$include_features <- get_feature_columns(data)
    
    return(settings)
  }
)



..set_identifier_column <- function(
    current = NULL,
    data = NULL,
    internal,
    external
) {
  
  if (!(is.waive(current) || is.null(current))) return(current)
  
  temporary <- NULL
  
  # Prefer external before internal, as long as it is present in data.
  if (all(lengths(external) > 0L)) {
    if (!any(sapply(external, is.na))) {
      if (data.table::is.data.table(data)) {
        if (all(external %in% colnames(data))) temporary <- external
      }
    }
  }
  
  # Prefer internal if it is present in data.
  if (data.table::is.data.table(data) && is.null(temporary)) {
    if (all(internal %in% colnames(data))) temporary <- internal
  }
  
  # Use external if internal is not present in data.
  if (all(lengths(external) > 0L) && is.null(temporary)) {
    if (!any(sapply(external, is.na))) temporary <- internal
  }
  
  if (is.null(temporary)) {
    return(current)
    
  } else {
    return(temporary)
  }
}


.split_data_by_batch_id <- function(data) {
  # Don't split if there is no data.
  if (is_empty(data, allow_no_features = TRUE)) return(NULL)
  
  # Split by batch id.
  split_data <- split(
    data@data,
    by = get_id_columns(id_depth = "batch")
  )
  
  # Update dataObject.
  data <- lapply(
    split_data,
    function(x, data) {
      data@data <- x
      return(data)
    },
    data = data
  )
  
  return(data)
}



# load_delayed_data methods ----------------------------------------------------

## load_delayed_data (dataObject, ANY) -----------------------------------------
setMethod(
  "load_delayed_data",
  signature(
    data = "dataObject",
    object = "ANY"
  ),
  function(
    data,
    object,
    ...
  ) {
    # If data is a dataObject, the data already was loaded.
    return(data)
  }
)



## load_delayed_data (dataObject, NULL) ----------------------------------------
setMethod(
  "load_delayed_data",
  signature(
    data = "delayedDataObject",
    object = "NULL"
  ),
  function(
    data,
    object,
    column_names = NULL,
    ...
  ) {
    if (is.na(data@validation) && is_empty(data@sample_set_on_load)) {
      ..error_reached_unreachable_code("validation attribute was not set")
    }
    
    # Determine which samples are required.
    if (is_empty(data@sample_set_on_load)) {
      # Check if data_id and run_id were set.
      if (is.na(data@data_id)) {
        ..error_reached_unreachable_code("data_id attribute was not set")
      }
      if (is.na(data@run_id)) {
        ..error_reached_unreachable_code("run_id attribute was not set")
      }
      
      # From iteration list.
      sample_identifiers <- .get_sample_identifiers(
        iteration_list = get_project_list()$iter_list,
        data_id = data@data_id,
        run_id = data@run_id,
        train_or_validate = ifelse(data@validation, "valid", "train")
      )
      
    } else {
      # From data object.
      sample_identifiers <- data@sample_set_on_load
    }
    
    # Resample sample identifiers.
    if (!is.na(data@sample_seed)) {
      sample_identifiers <- fam_sample(
        x = sample_identifiers,
        seed = data@sample_seed
      )
    }
    
    # Initialise dataObject
    new_data <- methods::new(
      "dataObject",
      data = NULL,
      preprocessing_level = "none",
      outcome_type = data@outcome_type,
      outcome_info = data@outcome_info,
      data_column_info = data@data_column_info,
      data_id = data@data_id,
      run_id = data@run_id,
      validation = data@validation,
      sample_seed = data@sample_seed
    )
    
    # Select only unique samples from the backend.
    if (!is_empty(sample_identifiers)) {
      unique_sample_identifiers <- unique(sample_identifiers)
      
    } else {
      # Return the dataObject placeholder.
      return(new_data)
    }
    
    # Attach data.
    new_data@data <- get_data_from_backend(
      sample_identifiers = unique_sample_identifiers,
      column_names = column_names
    )

    # Recreate iteration. Note that we here also use duplicate samples to
    # recreate e.g. bootstraps.
    new_data <- select_data_from_samples(
      data = new_data,
      samples = sample_identifiers
    )

    return(new_data)
  }
)


## load_delayed_data (delayedDataObject, model) --------------------------------
setMethod(
  "load_delayed_data",
  signature(
    data = "delayedDataObject",
    object = "familiarModel"
  ),
  function(
    data,
    object,
    stop_at,
    keep_novelty = FALSE,
    ...
  ) {
    # Prevent notes.
    data_id <- NULL
    
    # Read required features.
    required_features <- object@required_features
    
    if (is.na(data@run_id)) {
      data@run_id <- object@run_table[data_id == data@data_id, ]$run_id[1L]
    }
    
    # Get columns in data frame which are not features, but identifiers and
    # outcome instead.
    non_feature_columns <- get_non_feature_columns(x = object)
    
    # Explicitly refer to the method for object = NULL.
    new_data <- load_delayed_data(
      data = data,
      object = NULL,
      column_names = c(non_feature_columns, required_features)
    )
    
    # Pre-process data, if needed.
    new_data <- preprocess_data(
      data = new_data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty
    )
    
    # Aggregate data, if needed.
    if (data@aggregate_on_load) {
      new_data <- aggregate_data(data = new_data)
    }
    
    return(new_data)
  }
)



## load_delayed_data (delayedDataObject, novelty detector) ---------------------
setMethod(
  "load_delayed_data",
  signature(
    data = "delayedDataObject",
    object = "familiarNoveltyDetector"
  ),
  function(
    data,
    object,
    stop_at,
    ...
  ) {
    # Prevent notes.
    data_id <- NULL
    
    # Read required features.
    required_features <- object@required_features
    
    if (is.na(data@run_id)) {
      data@run_id <- object@run_table[data_id == data@data_id, ]$run_id[1L]
    }
    
    # Get columns in data frame which are not features, but identifiers instead.
    non_feature_columns <- get_non_feature_columns(x = object)
    
    # Explicitly refer to the method for object = NULL.
    new_data <- load_delayed_data(
      data = data,
      object = NULL,
      column_names = c(non_feature_columns, required_features)
    )
    
    # Pre-process data, if needed.
    new_data <- preprocess_data(
      data = new_data,
      object = object,
      stop_at = stop_at,
      keep_novelty = TRUE
    )
    
    # Aggregate data, if needed.
    if (data@aggregate_on_load) {
      new_data <- aggregate_data(data = new_data)
    }
    
    return(new_data)
  }
)



## load_delayed_data (delayedDataObject, vimp method) --------------------------
setMethod(
  "load_delayed_data",
  signature(
    data = "delayedDataObject",
    object = "familiarVimpMethod"
  ),
  function(
    data,
    object,
    stop_at,
    keep_novelty = FALSE,
    ...
  ) {
    # Prevent notes.
    data_id <- NULL
    
    # Read required features.
    required_features <- object@required_features
    
    if (is.na(data@run_id)) {
      data@run_id <- object@run_table[data_id == data@data_id, ]$run_id[1L]
    }
    
    # Get columns in data frame which are not features, but identifiers and
    # outcome instead.
    non_feature_columns <- get_non_feature_columns(x = object)
    
    # Explicitly refer to the method for object = NULL.
    new_data <- load_delayed_data(
      data = data,
      object = NULL,
      column_names = c(non_feature_columns, required_features)
    )
    
    # Pre-process data, if needed.
    new_data <- preprocess_data(
      data = new_data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty
    )
    
    # Aggregate data, if needed.
    if (data@aggregate_on_load) {
      new_data <- aggregate_data(data = new_data)
    }
    
    return(new_data)
  }
)



## load_delayed_data (delayedDataObject, ensemble) -----------------------------
setMethod(
  "load_delayed_data",
  signature(
    data = "delayedDataObject",
    object = "familiarEnsemble"
  ),
  function(
    data,
    object,
    stop_at,
    keep_novelty = FALSE,
    ...
  ) {
    # Prevent notes.
    data_id <- NULL
    
    # Read required features.
    required_features <- object@required_features
    
    # Leaving the run id NA signals that the run_id is determined by the data_id
    # of the data. For ensembles, this data_id may not actually be contained
    # in the run table (which is a composite of the underlying models).
    if (is.na(data@run_id)) {
      run_table <- tail(object@run_table[data_id <= data@data_id, ], n = 1L)
      
      if (run_table$data_id == data@data_id) {
        data@run_id <- run_table$run_id
        
      } else {
        # In this case, we are considering a higher subset of data, and
        # we should only consider the training set, because the actually 
        # requested subset is contained in the training set at that level --
        # but not within the validation subset.
        data@data_id <- run_table$data_id
        data@run_id <- run_table$run_id
        data@validation <- FALSE
      }
    }
    
    # Get columns in data frame which are not features, but identifiers and
    # outcome instead.
    non_feature_columns <- get_non_feature_columns(x = object)
    
    # Explicitly refer to the method for object = NULL.
    new_data <- load_delayed_data(
      data = data,
      object = NULL,
      column_names = c(non_feature_columns, required_features)
    )
    
    # Pre-process data, if needed.
    new_data <- preprocess_data(
      data = new_data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty
    )
    
    # Aggregate data, if needed.
    if (data@aggregate_on_load) {
      new_data <- aggregate_data(data = new_data)
    }
    
    return(new_data)
  }
)


# preprocess_data methods ------------------------------------------------------

## preprocess_data (delayedDataObject, ANY) ------------------------------------
setMethod(
  "preprocess_data",
  signature(
    data = "delayedDataObject",
    object = "ANY"
  ),
  function(
    data,
    object,
    stop_at = "clustering",
    keep_novelty = FALSE,
    ...
  ) {
    
    # Load data. This automatically pre-processes the data.
    data <- load_delayed_data(
      data = data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty,
      ...
    )
    
    # Perform any object-specific processing.
    data <- preprocess_data(
      data = data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty,
      ...
    )
    
    return(data)
  }
)



## preprocess_data (vimp method) -----------------------------------------------
setMethod(
  "preprocess_data",
  signature(
    data = "dataObject",
    object = "familiarVimpMethod"
  ),
  function(
    data,
    object,
    stop_at = "clustering",
    keep_novelty = FALSE, ...
  ) {
    
    # Pre-process the data.
    data <-  .preprocess_data(
      data = data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty
    )
    
    return(data)
  }
)



## preprocess_data (model) -----------------------------------------------------
setMethod(
  "preprocess_data",
  signature(
    data = "dataObject",
    object = "familiarModel"
  ),
  function(
    data,
    object,
    stop_at = "clustering",
    keep_novelty = FALSE,
    ...
  ) {
    
    # Pre-process the data.
    data <- .preprocess_data(
      data = data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty
    )
    
    # Post-process the data to select the correct feature and identifier set.
    data <- postprocess_data(
      data = data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty,
      ...
    )
    
    return(data)
  }
)



## preprocess_data (familiarNaiveModel) ----------------------------------------
setMethod(
  "preprocess_data",
  signature(
    data = "dataObject",
    object = "familiarNaiveModel"
  ),
  function(
    data,
    object,
    stop_at = "clustering",
    keep_novelty = FALSE,
    ...
  ) {
    
    if (.as_preprocessing_level(data@preprocessing_level) > .as_preprocessing_level(stop_at)) {
      ..error_reached_unreachable_code(paste0(
        "preprocess_data,dataObject,familiarNaiveModel: data were preprocessed ",
        "at a higher level than required by stop_at."
      ))
    }
    
    # familiarNaiveModels do not have any features, so we remove them.
    data <- select_features(data = data, features = NULL)
    
    # Set the pre-processing level.
    data@preprocessing_level <- stop_at  
    
    return(data)
  }
)



## preprocess_data (novelty detector) ------------------------------------------
setMethod(
  "preprocess_data",
  signature(
    data = "dataObject",
    object = "familiarNoveltyDetector"
  ),
  function(
    data,
    object,
    stop_at = "clustering",
    ...
  ) {
    
    # Note that keep_novelty is always false to prevent reading novelty_features
    # slot. Novelty features are stored in the model_features slot of
    # familiarNoveltyDetector objects.
    
    # Assign "unsupervised" outcome type to the data to prevent outcome columns
    # being selected.
    data@outcome_type <- "unsupervised"
    
    # Pre-process the data.
    data <- .preprocess_data(
      data = data,
      object = object,
      stop_at = stop_at,
      keep_novelty = FALSE
    )
    
    # Post-process the data to select the correct feature and identifier set.
    data <- postprocess_data(
      data = data,
      object = object,
      stop_at = stop_at,
      ...
    )
    
    return(data)
  }
)


## preprocess_data (ensemble) --------------------------------------------------
setMethod(
  "preprocess_data",
  signature(
    data = "dataObject",
    object = "familiarEnsemble"
  ),
  function(
    data,
    object,
    stop_at = "clustering",
    keep_novelty = FALSE,
    ...
  ) {
    
    # Pre-process the data.
    data <- .preprocess_data(
      data = data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty
    )
    
    return(data)
  }
) 



.preprocess_data <- function(
    data,
    object,
    stop_at,
    keep_novelty = FALSE
) {
  # Convert the preprocessing_level attained and the requested stopping level to
  # ordinals.
  preprocessing_level_attained <- .as_preprocessing_level(data@preprocessing_level)
  stop_at <- .as_preprocessing_level(stop_at)
  
  # Check whether pre-processing is required
  if (preprocessing_level_attained == stop_at) {
    return(data)
    
  } else if (preprocessing_level_attained > stop_at) {
    ..error_reached_unreachable_code(paste0(
      "preprocess_data,dataObject,ANY: data were preprocessed at a higher level", 
      " than required by stop_at."
    ))
  }
  
  if (preprocessing_level_attained < "signature" && stop_at >= "signature") { 
    # Apply the signature.
    data <- select_features(data = data, features = object@required_features)
    
    # Update pre-processing level externally as it is not limited to
    # pre-processing per sé.
    data@preprocessing_level <- "signature"
    
  } else if (preprocessing_level_attained == "signature" && stop_at >= "signature") {
    
    required_features <- object@required_features
    selected_features <- object@model_features
    if (keep_novelty) selected_features <- union(selected_features, object@novelty_features)
    
    if (length(required_features) > 0L && length(selected_features) > 0L && has_feature_data(data)) {
      
      # Select available features specific to the object.
      if (all(required_features %in% get_feature_columns(data))) {
        data <- select_features(data = data, features = required_features)
        
      } else if (all(selected_features %in% get_feature_columns(data))) {
        data <- select_features(data = data, features = selected_features)
        
      } else {
        ..error_reached_unreachable_code(
          ".preprocess_data: could not identify overlapping features"
        )
      }
    }
  }
  
  if (preprocessing_level_attained < "transformation" && stop_at >= "transformation") {
    # Transform the features.
    data <- transform_features(
      data = data,
      feature_info_list = object@feature_info
    )
  }
  
  if (preprocessing_level_attained < "normalisation" && stop_at >= "normalisation") {
    # Normalise feature values.
    data <- normalise_features(
      data = data,
      feature_info_list = object@feature_info
    )
  }
  
  if (preprocessing_level_attained < "batch_normalisation" && stop_at >= "batch_normalisation") {
    # Batch-normalise feature values
    data <- batch_normalise_features(
      data = data,
      feature_info_list = object@feature_info
    )
  }
  
  if (preprocessing_level_attained < "imputation" && stop_at >= "imputation") {
    # Impute missing values
    data  <- impute_features(
      data = data,
      feature_info_list = object@feature_info
    )
  }
  
  if (preprocessing_level_attained < "clustering" && stop_at >= "clustering") {
    # Cluster features
    data <- cluster_features(
      data = data,
      feature_info_list = object@feature_info
    )
  }
  
  return(data)
}


# postprocess_data methods -----------------------------------------------------

## postprocess_data (model)-----------------------------------------------------
setMethod(
  "postprocess_data",
  signature(
    data = "dataObject",
    object = "familiarModel"
  ),
  function(
    data,
    object,
    stop_at = "clustering",
    keep_novelty = FALSE,
    force_check = FALSE,
    ...
  ) {
    
    # Convert the preprocessing_level attained and the requested stopping level
    # to ordinals.
    preprocessing_level_attained <- .as_preprocessing_level(data@preprocessing_level)
    stop_at <- .as_preprocessing_level(stop_at)
    
    if (stop_at < "clustering" || preprocessing_level_attained < "clustering") return(data)
    
    # Select features.
    features <- object@model_features
    if (keep_novelty) features <- union(features, object@novelty_features)
    
    # Return data if there are no features.
    if (length(features) == 0L && !force_check) return(data)
    
    # Determine the features after clustering.
    features <- features_after_clustering(
      features = features,
      feature_info_list = object@feature_info
    )
    
    # Create a slice of the data for the feature set.
    data <- select_features(
      data = data,
      features = features
    )
    
    return(data)
  }
)



## postprocess_data (novelty detector)------------------------------------------
setMethod(
  "postprocess_data",
  signature(
    data = "dataObject",
    object = "familiarNoveltyDetector"
  ),
  function(
    data,
    object,
    stop_at = "clustering",
    force_check = FALSE,
    ...
  ) {
    
    # Convert the preprocessing_level attained and the requested stopping level
    # to ordinals.
    preprocessing_level_attained <- .as_preprocessing_level(data@preprocessing_level)
    stop_at <- .as_preprocessing_level(stop_at)
    
    if (stop_at < "clustering" || preprocessing_level_attained < "clustering") return(data)
    
    # Select features.
    features <- object@model_features
    
    # Return data if there are no features.
    if (length(features) == 0L && !force_check) return(data)
    
    # Determine the features after clustering.
    features <- features_after_clustering(
      features = features,
      feature_info_list = object@feature_info
    )
    
    # Create a slice of the data for the feature set.
    data <- select_features(
      data = data,
      features = features
    )
    
    return(data)
  }
)



## process_input_data (vimp method) --------------------------------------------
setMethod(
  "process_input_data",
  signature(
    object = "familiarVimpMethod",
    data = "ANY"
  ),
  function(
    object,
    data,
    is_pre_processed = FALSE,
    stop_at = "clustering",
    keep_novelty = FALSE,
    ...
  ) {
    
    data <- .process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = stop_at,
      keep_novelty = keep_novelty,
      ...
    )
    
    return(data)
  }
)

## process_input_data (novelty detector) ---------------------------------------
setMethod(
  "process_input_data",
  signature(
    object = "familiarNoveltyDetector",
    data = "ANY"
  ),
  function(
    object,
    data,
    is_pre_processed = FALSE, 
    stop_at = "clustering",
    ...
  ) {
    
    # Ensure that the outcome_type of data is changed to "unsupervised". We
    # don't need any outcome columns.
    data@outcome_type <- "unsupervised"
    
    # Process data.
    data <- .process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = stop_at,
      keep_novelty = FALSE,
      ...
    )
    
    return(data)
  }
)



## process_input_data (model) --------------------------------------------------
setMethod(
  "process_input_data",
  signature(
    object = "familiarModel",
    data = "ANY"
  ),
  function(
    object,
    data,
    is_pre_processed = FALSE,
    stop_at = "clustering",
    keep_novelty = FALSE,
    ...
  ) {
    
    data <- .process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = stop_at,
      keep_novelty = keep_novelty,
      ...
    )
    
    return(data)
  }
)



## process_input_data (ensemble) -----------------------------------------------
setMethod(
  "process_input_data",
  signature(
    object = "familiarEnsemble",
    data = "ANY"
  ),
  function(
    object,
    data,
    is_pre_processed = FALSE,
    stop_at = "clustering",
    keep_novelty = FALSE
  ) {
    
    data <- .process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = stop_at,
      keep_novelty = keep_novelty
    )
    
    return(data)
  }
)


## process_input_data (dataObject) -----------------------------------------------
setMethod(
  "process_input_data",
  signature(
    object = "dataObject",
    data = "ANY"
  ),
  function(
    object,
    data,
    is_pre_processed = FALSE,
    stop_at = "clustering",
    keep_novelty = FALSE
  ) {
    return(data)
  }
)



.process_input_data <- function(
    object,
    data,
    is_pre_processed,
    stop_at,
    keep_novelty = FALSE,
    ...
) {
  
  # Check whether data is a dataObject, and create one otherwise
  if (is(data, "delayedDataObject")) {
    data <- load_delayed_data(
      data = data,
      object = object,
      stop_at = stop_at,
      keep_novelty = keep_novelty,
      ...
    )
    
  } else if (!is(data, "dataObject")) {
    data <- as_data_object(
      data = data,
      object = object
    )
    
    # Set pre-processing level.
    data@preprocessing_level <- ifelse(is_pre_processed, "clustering", "none")
  }
  
  # Pre-process data in case it has not been pre-processed
  data <- preprocess_data(
    data = data,
    object = object,
    stop_at = stop_at,
    keep_novelty = keep_novelty,
    ...
  )
  
  # Return data
  return(data)
}



# select_data_from_samples (dataObject) ----------------------------------------
setMethod(
  "select_data_from_samples",
  signature(
    data = "dataObject",
    samples = "ANY"
  ),
  function(data, samples = NULL) {
    
    # Determine the names of the id-columns, up to the series level.
    id_columns <- get_id_columns(id_depth = "series")
    
    if (is_empty(samples)) {
      # Return an empty data set if no samples are provided
      data@data <- head(data@data, n = 0L)
      
    } else {
      # Use samples from the samples function argument. allow.cartesian is set
      # to true to allow use with repeated measurements.
      if (all(id_columns %in% colnames(samples))) {
        data@data <- merge(
          x = samples,
          y = data@data,
          by = id_columns,
          all = FALSE,
          allow.cartesian = TRUE
        )
        
      } else {
        data@data <- merge(
          x = samples,
          y = data@data,
          by = get_id_columns(id_depth = "sample"),
          all = FALSE,
          allow.cartesian = TRUE
        )
      }
    } 

    return(data)
  }
)



# select_data_from_samples (delayedDataObject) ---------------------------------
setMethod(
  "select_data_from_samples",
  signature(
    data = "delayedDataObject",
    samples = "ANY"
  ),
  function(data, samples = NULL) {
    
    # Set the sample set that should be loaded.
    data@sample_set_on_load <- samples
    
    return(data)
  }
)




# aggregate_data (dataObject) --------------------------------------------------
setMethod(
  "aggregate_data",
  signature(data = "dataObject"),
  function(data) {
    
    # Check if the data is empty
    if (is_empty(data)) return(data)
    
    # Drop any duplicates (e.g. from bootstraps).
    data@data <- unique(data@data)
    
    # Identify the columns containing outcome, series, sample, and batch
    # identifiers.
    id_cols <- get_non_feature_columns(x = data, id_depth = "series")
    if (all(data@data$repetition_id == 1L)) return(data)
    
    # Identify feature columns for repeated measurement data.
    feature_columns <- get_feature_columns(x = data)
    
    # Identify class of features
    column_class <- lapply(
      feature_columns,
      function(ii, data) (class(data[[ii]])[1L]),
      data = data@data
    )
    
    # Determine numerical features
    numeric_features <- sapply(
      column_class,
      function(selected_column_class) (any(selected_column_class %in% c("numeric", "integer")))
    )
    
    numeric_features <- feature_columns[numeric_features]
    
    # Determine categorical features
    categorical_features <- sapply(
      column_class,
      function(selected_column_class) (any(selected_column_class %in% c("logical", "character", "factor")))
    )
    
    categorical_features <- feature_columns[categorical_features]
    
    # Find non-duplicate id columns
    aggregated_data <- unique(data@data[, mget(id_cols)])
    
    # Add aggregated numeric columns
    if (length(numeric_features) > 0L) {
      numeric_data <- data@data[
        , lapply(.SD, stats::median),
        by = id_cols,
        .SDcols = numeric_features
      ]
      
      aggregated_data <- merge(
        x = aggregated_data,
        y = numeric_data,
        by = id_cols
      )
    }
    
    # Add aggregated factor columns
    if (length(categorical_features) > 0L) {
      categorical_data <- data@data[
        ,
        lapply(.SD, get_mode),
        by = id_cols,
        .SDcols = categorical_features
      ]
      
      aggregated_data <- merge(
        x = aggregated_data,
        y = categorical_data,
        by = id_cols
      )
    }
    
    # Add in repetition_id column again
    aggregated_data[, "repetition_id" := -1L]
    
    # Reorder columns so that it matches the input
    data.table::setcolorder(
      aggregated_data,
      neworder = colnames(data@data)
    )
    
    data@data <- aggregated_data
    
    return(data)
  }
)



# aggregate_data (delayedDataObject) -------------------------------------------
setMethod(
  "aggregate_data",
  signature(data = "delayedDataObject"),
  function(data) {
    # Mark for future aggregation after loading the data.
    data@aggregate_on_load <- TRUE
    return(data)
  }
)



# select_unique_data (dataObject) ----------------------------------------------
setMethod(
  "select_unique_data",
  signature(data = "dataObject"),
  function(data) {
    # Check if the data is empty
    if (is_empty(data)) return(data)
    
    # Drop any duplicates (e.g. from bootstraps).
    data@data <- unique(data@data)
    
    return(data)
  }
)



# filter_features --------------------------------------------------------------
setMethod(
  "filter_features",
  signature(data = "dataObject"),
  function(
    data,
    remove_features = NULL,
    available_features = NULL
  ) {
    # Removes features from a data set
    
    # If both are provided, use remove_features
    if (!is.null(remove_features) && !is.null(available_features)) {
      available_features <- NULL
    }
    
    # Do not do anything if both are null
    if (is.null(remove_features) && is.null(available_features)) {
      return(data)
      
      # Based on remove_features input
    } else if (!is.null(remove_features)) {
      
      # Determine which remove_features are actually present in data
      remove_features <- intersect(remove_features, get_feature_columns(x = data))
      
      # Based on available_features input
    } else if (!is.null(available_features)) {
      # Determine which features should be removed
      remove_features <- setdiff(get_feature_columns(x = data), available_features)
      
      # Only keep features that are in data.
      available_features <- intersect(get_feature_columns(x = data), available_features)
      
    } else {
      ..error_reached_unreachable_code(
        "This point should never be reachable. Check for inconsistencies if it does."
      )
    }
    
    # Make a copy to prevent updating by reference.
    data@data <- data.table::copy(data@data)
    
    # Remove features from data if there is 1 or more feature to remove
    if (length(remove_features) > 0L) data@data[, (remove_features) := NULL]
    
    # Make sure that the column order is the same as available_features
    if (!is.null(available_features)) {
      data.table::setcolorder(
        x = data@data,
        neworder = c(get_non_feature_columns(x = data), available_features)
      )
    }
    
    return(data)
  }
)


# filter_missing_outcome (dataObject) ------------------------------------------
setMethod(
  "filter_missing_outcome",
  signature(data = "dataObject"),
  function(data, is_validation = FALSE, ...) {
    
    # Check if data is empty
    if (is_empty(data)) return(data)
    
    # Pass to method for data.table.
    data@data <- filter_missing_outcome(
      data = data@data,
      outcome_type = data@outcome_type,
      is_validation = is_validation,
      ...
    )
    
    return(data)
  }
)


# filter_missing_outcome (data.table) ------------------------------------------
setMethod(
  "filter_missing_outcome",
  signature(data = "data.table"),
  function(
    data, 
    outcome_type, 
    is_validation = FALSE, 
    ...
  ) {
    if (is_empty(data)) return(data)
    
    # Check predicted outcome columns.
    if (outcome_type %in% c("survival", "competing_risk")) {
      outcome_is_valid <- is_valid_data(data[["outcome_time"]]) & is_valid_data(data[["outcome_event"]])
      
    } else if (outcome_type %in% c("continuous")) {
      outcome_is_valid <- is_valid_data(data[["outcome"]])
      
    } else if (outcome_type %in% c("binomial", "multinomial")) {
      outcome_is_valid <- is_valid_data(data[["outcome"]])
      
    } else {
      ..error_no_known_outcome_type(outcome_type)
    }
    
    if (is_validation) {
      # Check whether all outcome information is missing for validation. It may
      # be a prospective study. In that case, keep all data.
      if (!any(outcome_is_valid)) outcome_is_valid <- !outcome_is_valid
    }
    
    # Keep only data for which the outcome exists
    data <- data[(outcome_is_valid), ]
    
    return(data)
  }
)



# filter_bad_samples -----------------------------------------------------------
setMethod(
  "filter_bad_samples",
  signature(data = "dataObject"),
  function(data, threshold) {
    
    # Check if data is empty
    if (is_empty(data)) return(data)
    
    # Find the columns containing features
    feature_columns <- get_feature_columns(x = data)
    
    # Still very fast, but is much friendlier to our poor memory.
    n_missing <- numeric(nrow(data@data))
    for (feature in feature_columns) {
      n_missing <- n_missing + !is_valid_data(data@data[[feature]])
    }
    
    # Set threshold number, as threshold is a fraction.
    threshold_number <- length(feature_columns) * threshold
    
    # Only keep samples with a number of missing values below the threshold
    data@data <- data@data[(n_missing <= threshold_number)]
    
    return(data)
  }
)


# transform_features methods ---------------------------------------------------

## transform_features (dataObject) ---------------------------------------------
setMethod(
  "transform_features",
  signature(data = "dataObject"),
  function(
    data,
    feature_info_list,
    invert = FALSE
  ) {
    
    # Check if transformation was already performed.
    if (!invert && .as_preprocessing_level(data) >= "transformation") {
      ..error_reached_unreachable_code(
        "transform_features,dataObject: attempting to transform data that are already transformed."
      )
    }
    
    # Update the preprocessing level.
    if (!invert) data@preprocessing_level <- "transformation"
    if (invert) data@preprocessing_level <- "signature"
    
    # Check if data is empty
    if (is_empty(data)) return(data)
    
    # Find the columns containing features
    feature_columns <- get_feature_columns(x = data)
    
    # Transform features
    data@data <- transform_features(
      data = data@data,
      feature_info_list = feature_info_list,
      features = feature_columns,
      invert = invert
    )
    
    return(data)
  }
)


## transform_features (data.table) ---------------------------------------------
setMethod(
  "transform_features",
  signature(data = "data.table"),
  function(
    data,
    feature_info_list,
    features,
    invert = FALSE
  ) {
    
    # Check if data is empty
    if (is_empty(data)) return(data)
    
    # Apply transformations
    transformed_list <- lapply(
      features,
      function(ii, data, feature_info_list, invert) {
        x <- apply_feature_info_parameters(
          object = feature_info_list[[ii]]@transformation_parameters,
          data = data[[ii]],
          invert = invert
        )
        
        return(x)
      },
      data = data,
      feature_info_list = feature_info_list,
      invert = invert
    )
    
    # Update name of data in columns
    names(transformed_list) <- features
    
    # Update with replacement in the data object
    data <- update_with_replacement(
      data = data,
      replacement_list = transformed_list
    )
    
    return(data)
  }
)


# normalise_features methods ---------------------------------------------------

## normalise_features (dataObject) ---------------------------------------------
setMethod(
  "normalise_features",
  signature(data = "dataObject"),
  function(
    data,
    feature_info_list,
    invert = FALSE
  ) {
    
    # Check if normalisation was already performed.
    if (!invert && .as_preprocessing_level(data) >= "normalisation") {
      ..error_reached_unreachable_code(
        "normalise_features,dataObject: attempting to normalise data that are already normalised."
      )
    }
    
    # Check if the previous step (transformation) was conducted.
    if (!invert && .as_preprocessing_level(data) < "transformation") {
      ..error_reached_unreachable_code(
        "normalise_features,dataObject: data should be transformed prior to normalisation."
      )
    }
    
    # Update the preprocessing_level.
    if (!invert) data@preprocessing_level <- "normalisation"
    if (invert) data@preprocessing_level <- "transformation"
    
    # Check if data is empty
    if (is_empty(data)) return(data)
    
    # Find the columns containing features
    feature_columns <- get_feature_columns(x = data)
    
    # Apply normalisation
    data@data <- normalise_features(
      data = data@data,
      feature_info_list = feature_info_list,
      features = feature_columns,
      invert = invert
    )
    
    return(data)
  }
)


## normalise_features (data.table) ---------------------------------------------
setMethod(
  "normalise_features",
  signature(data = "data.table"),
  function(
    data,
    feature_info_list,
    features,
    invert = FALSE
  ) {
    
    # Check if data is empty.
    if (is_empty(data)) return(data)
    
    # Apply normalisation.
    normalised_list <- lapply(
      features,
      function(ii, data, feature_info_list, invert) {
        x <- apply_feature_info_parameters(
          object = feature_info_list[[ii]]@normalisation_parameters,
          data = data[[ii]],
          invert = invert
        )
        
        return(x)
      },
      data = data,
      feature_info_list = feature_info_list,
      invert = invert
    )
    
    # Update name of data in columns.
    names(normalised_list) <- features
    
    # Update with replacement in the data object.
    data <- update_with_replacement(
      data = data,
      replacement_list = normalised_list
    )
    
    return(data)
  }
)


# batch_normalise_features -----------------------------------------------------

## batch_normalise_features (dataObject) ---------------------------------------
setMethod(
  "batch_normalise_features",
  signature(data = "dataObject"),
  function(
    data,
    feature_info_list,
    cl = NULL,
    invert = FALSE
  ) {
    
    # Check if batch normalisation was already performed.
    if (!invert && .as_preprocessing_level(data) >= "batch_normalisation") {
      ..error_reached_unreachable_code(paste0(
        "batch_normalise_features,dataObject: attempting to batch normalise data ",
        "that are already batch normalised."
      ))
    }
    
    # Check if the previous step (normalisation) was conducted.
    if (!invert && .as_preprocessing_level(data) < "normalisation") {
      ..error_reached_unreachable_code(paste0(
        "batch_normalise_features,dataObject: data should be normalised globally ",
        "prior to batch normalisation."
      ))
    }
    
    # Update the preprocessing_level.
    if (!invert) data@preprocessing_level <- "batch_normalisation"
    if (invert) data@preprocessing_level <- "normalisation"
    
    # Check if data is empty
    if (is_empty(data)) return(data)
    
    # Find the columns containing features
    features <- get_feature_columns(x = data)
    
    # Update feature_info_list by adding info for missing batches
    feature_info_list <- add_batch_normalisation_parameters(
      feature_info_list = feature_info_list[features],
      data = data
    )
    
    # Apply normalisation
    data@data <- batch_normalise_features(
      data = data@data,
      feature_info_list = feature_info_list,
      features = features,
      invert = invert
    )
    
    return(data)
  }
)



## batch_normalise_features (data.table) ---------------------------------------
setMethod(
  "batch_normalise_features",
  signature(data = "data.table"),
  function(
    data,
    feature_info_list,
    features,
    invert = FALSE
  ) {
    
    # Check if data is empty.
    if (is_empty(data)) return(data)
    
    # Apply batch normalisation.
    batch_normalised_list <- lapply(
      features,
      function(ii, data, batch_data, feature_info_list, invert) {
        x <- apply_feature_info_parameters(
          object = feature_info_list[[ii]]@batch_normalisation_parameters,
          data = data[[ii]],
          batch_data = batch_data,
          invert = invert
        )
        
        return(x)
      },
      data = data,
      batch_data = data[[get_id_columns("batch")]],
      feature_info_list = feature_info_list,
      invert = invert
    )
    
    # Update name of data in columns.
    names(batch_normalised_list) <- features
    
    # Update with replacement in the data object.
    data <- update_with_replacement(
      data = data,
      replacement_list = batch_normalised_list
    )
    
    return(data)
  }
)


# impute_features --------------------------------------------------------------
setMethod(
  "impute_features",
  signature(data = "dataObject"),
  function(
    data,
    feature_info_list,
    cl = NULL
  ) {
    
    # Check if imputation was already performed.
    if (.as_preprocessing_level(data) >= "imputation") {
      ..error_reached_unreachable_code(
        "impute_features,dataObject: attempting to impute data that already have been imputed."
      )
    }
    
    # Check if the previous step (batch normalisation) was conducted.
    if (.as_preprocessing_level(data) < "batch_normalisation") {
      ..error_reached_unreachable_code(
        "impute_features,dataObject: data should be batch normalised prior to imputation."
      )
    }
    
    # Update the attained processing level.
    data@preprocessing_level <- "imputation"
    
    # Check if data is empty
    if (is_empty(data)) return(data)
    
    # Check if data has features
    if (!has_feature_data(x = data)) return(data)
    
    # Apply univariate results to the dataset.
    imputed_data <- .impute_features(
      data = data,
      feature_info_list = feature_info_list,
      initial_imputation = TRUE
    )
    
    # Apply multivariate model to the dataset.
    data <- .impute_features(
      data = imputed_data,
      feature_info_list = feature_info_list,
      initial_imputation = FALSE,
      mask_data = data
    )
    
    return(data)
  }
)


# cluster_features -------------------------------------------------------------
setMethod(
  "cluster_features",
  signature(data = "dataObject"),
  function(data, feature_info_list) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    feature_required <- feature_name <- NULL
    
    if (.as_preprocessing_level(data) >= "clustering") {
      ..error_reached_unreachable_code(
        "cluster_features,dataObject: attempting to cluster data that already have been clustered."
      )
    }
    
    # Check if the previous step (imputation) was conducted.
    if (.as_preprocessing_level(data) < "imputation") {
      ..error_reached_unreachable_code(
        "cluster_features,dataObject: data should be imputed prior to clustering."
      )
    }
    
    # Update the attained processing level.
    data@preprocessing_level <- "clustering"
    
    # Check if data is empty
    if (is_empty(data)) return(data)
    
    # Check if data has features
    if (!has_feature_data(x = data)) return(data)
    
    # Find the columns containing features
    feature_columns <- get_feature_columns(x = data)
    
    # Derive clustering table.
    cluster_table <- .create_clustering_table(
      feature_info_list = feature_info_list,
      selected_features = feature_columns
    )
    
    # Update data using the clustering table. Note that only features
    # that are both required and present are processed.
    clustered_data <- lapply(
      split(
        cluster_table[feature_required == TRUE & feature_name %in% feature_columns],
        by = "cluster_name"
      ),
      set_clustered_data,
      data = data,
      feature_info_list = feature_info_list
    )
    
    # Attach the clustered data.
    data@data <- cbind(
      data@data[, mget(get_non_feature_columns(data))],
      data.table::setDT(clustered_data)
    )
    
    return(data)
  }
)



# update_with_replacement methods ----------------------------------------------

## update_with_replacement (dataObject) ----------------------------------------
setMethod(
  "update_with_replacement",
  signature(data = "dataObject"),
  function(data, replacement_list) {
    
    # Replace data by passing to the data.table method.
    data@data <- update_with_replacement(
      data = data@data,
      replacement_list = replacement_list
    )
    
    return(data)
  }
)



## update_with_replacement (data.table) ----------------------------------------
setMethod(
  "update_with_replacement",
  signature(data = "data.table"),
  function(data, replacement_list) {
    # Updates columns of a data table with replacement data from the replacement
    # list.
    replacement_table <- data.table::copy(data)
    
    # Find feature names corresponding to columns to be replaced
    replace_features <- names(replacement_list)
    
    # Iterate over replacement list entries
    for (current_feature in replace_features) {
      replacement_table[, (current_feature) := replacement_list[[current_feature]]]
    }
    
    return(replacement_table)
  }
)


# select_features --------------------------------------------------------------
setMethod(
  "select_features",
  signature(data = "dataObject"),
  function(data, features) {
    # Allows for slicing the data.
    
    # Find non-feature columns
    non_feature_columns <- get_non_feature_columns(x = data)
    
    # Check if features are present as column name
    if (length(features) > 0L) {
      if (!all(features %in% colnames(data@data))) {
        logger_stop("Not all features were found in the data set.")
      }
    }
    
    # Define the selected columns
    selected_columns <- unique(c(non_feature_columns, features))
    
    # Check if all columns are already present in the data. In that case we do
    # not need to copy the dataset.
    if (!setequal(selected_columns, colnames(data@data))) {
      # Select features
      data@data <- data.table::copy(data@data[, mget(selected_columns)])
    }
    
    return(data)
  }
)



# get_required_features methods ------------------------------------------------

## get_required_features (dataObject) ------------------------------------------
setMethod(
  "get_required_features",
  signature(x = "dataObject"),
  function(
    x,
    feature_info_list,
    features = NULL,
    exclude_signature = FALSE,
    exclude_novelty = FALSE,
    ...
  ) {
    
    # Check if features are provided externally.
    is_external <- !is.null(features)
    
    if (!is_external) {
      # Get features directly.
      features <- get_feature_columns(x)
    }
    
    # Pass to underlying function.
    return(.get_required_features(
      features = features,
      feature_info_list,
      is_clustered = .as_preprocessing_level(x@preprocessing_level) == "clustering",
      is_external = is_external,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty
    ))
  }
)


## get_required_features (delayedDataObject) -----------------------------------
setMethod(
  "get_required_features",
  signature(x = "delayedDataObject"),
  function(
    x,
    feature_info_list,
    features = NULL,
    exclude_signature = FALSE,
    exclude_novelty = FALSE,
    ...
  ) {
    # Check if features are provided externally.
    is_external <- !is.null(features)

    if (is_external) {
      ..error_reached_unreachable_code("features should remain unset (NULL).")
    }
    
    features <- get_available_features(
      feature_info_list = feature_info_list,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty
    )
    
    return(.get_required_features(
      features = features,
      feature_info_list,
      is_clustered = .as_preprocessing_level(x@preprocessing_level) == "clustering",
      is_external = is_external,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty
    ))
  }
)



## get_required_features (character) -------------------------------------------
setMethod(
  "get_required_features",
  signature(x = "character"),
  function(
    x,
    feature_info_list,
    is_clustered,
    exclude_signature = FALSE,
    exclude_novelty = FALSE,
    ...
  ) {
    # Method intended for directly providing features.
    
    if (length(x) == 0L) return(NULL)
    
    return(.get_required_features(
      features = x,
      feature_info_list = feature_info_list,
      is_clustered = is_clustered,
      is_external = TRUE,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty
    ))
  }
)



## get_required_features (list) ------------------------------------------------
setMethod(
  "get_required_features",
  signature(x = "list"),
  function(
    x,
    exclude_signature = FALSE,
    exclude_novelty = FALSE,
    ...
  ) {
    # Method intended for directly providing a list of featureInfo objects.
    if (is_empty(x)) return(NULL)
    
    # Sanity check. x should be a list of 
    if (!all(sapply(x, is, "featureInfo"))) {
      ..error_reached_unreachable_code(
        "get_required_features,list: expected a list of featureInfo objects."
      )
    }
    
    # Get features from the featureInfo objects.
    features <- get_available_features(
      feature_info_list = x,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty
    )
    
    return(.get_required_features(
      features = features,
      feature_info_list = x,
      is_clustered = FALSE,
      is_external = FALSE,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty
    ))
  }
)



## get_required_features (NULL) ------------------------------------------------
setMethod(
  "get_required_features",
  signature(x = "NULL"),
  function(x, ...) {
    return(NULL)            
  }
)



.get_required_features <- function(
    features,
    feature_info_list,
    is_clustered,
    is_external,
    exclude_signature = FALSE,
    exclude_novelty = FALSE,
    exclude_imputation = FALSE,
    ...
) {
  # What are required features? Required features are features that are:
  #
  # 1. Directly used for clustering (representative features).
  #
  # 2. Used for imputation of features mentioned in 1).
  
  # Suppress NOTES due to non-standard evaluation in data.table
  cluster_name <- feature_name <- feature_required <- NULL
  
  # If the features are empty
  if (is_empty(features) && is_external) return(NULL)
  
  # Generate a cluster table
  cluster_table <- .create_clustering_table(feature_info_list = feature_info_list)
  
  # Identify which features are required to form clusters.
  if (is_clustered) {
    # Data are clustered.
    required_features <- cluster_table[
      cluster_name %in% features & feature_required == TRUE
    ]$feature_name
    
    # Sanity check. All externally provided features should be present
    # in the cluster table.
    if (is_external) {
      if (!all(features %in% cluster_table$cluster_name)) {
        ..error_reached_unreachable_code(paste0(
          ".get_required_features: some features do not appear in the ",
          " cluster table. Potentially, an incomplete feature_info_list was passed."
        ))
      }
    }
    
  } else {
    # Data are not yet clustered.
    required_features <- cluster_table[
      feature_name %in% features & feature_required == TRUE
    ]$feature_name
    
    # Sanity check. All externally provided features should be present
    # in the cluster table.
    if (is_external) {
      if (!all(features %in% cluster_table$feature_name)) {
        ..error_reached_unreachable_code(paste0(
          ".get_required_features: some features do not appear in the ",
          " cluster table. Potentially, an incomplete feature_info_list was passed."
        ))
      }
    }
  }
  
  # Determine available features from feature info list.
  available_features <- get_available_features(
    feature_info_list = feature_info_list,
    exclude_signature = exclude_signature,
    exclude_novelty = exclude_novelty
  )
  
  # Sanity check. All features that we provide externally should be
  # present in the available features.
  if (is_external) {
    if (length(required_features) == 0L) {
      ..error_reached_unreachable_code(paste0(
        "get_required_features,dataObject: there are no required features, ",
        "whereas at least one is expected."
      ))
    }
    
    if (!all(required_features %in% available_features)) {
      ..error_reached_unreachable_code(paste0(
        "get_required_features,dataObject: there are required features for clustering ",
        "that are not available."
      ))
    }
  }
  
  # Determine the intersect of the available features and features.
  required_features <- intersect(available_features, required_features)
  
  # Return NULL if no features are required.
  if (length(required_features) == 0L) return(NULL)
  
  if (!exclude_imputation) {
    # Now, additionally select which features are required for imputation.
    required_features <- unique(unlist(lapply(
      feature_info_list[required_features],
      function(x) x@required_features
    )))
  }
  
  return(required_features)
}



# get_model_features methods ---------------------------------------------------

## get_model_features (dataObject) ---------------------------------------------
setMethod(
  "get_model_features",
  signature(x = "dataObject"),
  function(
    x,
    feature_info_list,
    features = NULL,
    exclude_signature = FALSE,
    exclude_novelty = FALSE,
    ...
  ) {
    
    # Check if features are provided externally.
    is_external <- !is.null(features)
    
    if (!is_external) {
      # Get features directly.
      features <- get_feature_columns(x)
    }
    
    # Pass to underlying function.
    return(.get_required_features(
      features = features,
      feature_info_list,
      is_clustered = .as_preprocessing_level(x@preprocessing_level) == "clustering",
      is_external = is_external,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty,
      exclude_imputation = TRUE,
      ...
    ))
  }
)



## get_model_features (delayedDataObject) --------------------------------------
setMethod(
  "get_model_features",
  signature(x = "delayedDataObject"),
  function(
    x,
    feature_info_list,
    features = NULL,
    exclude_signature = FALSE,
    exclude_novelty = FALSE,
    ...
  ) {
    # Check if features are provided externally.
    is_external <- !is.null(features)
    
    if (is_external) {
      ..error_reached_unreachable_code("features should remain unset (NULL).")
    }
    
    features <- get_available_features(
      feature_info_list = feature_info_list,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty
    )
    
    return(.get_required_features(
      features = features,
      feature_info_list,
      is_clustered = .as_preprocessing_level(x@preprocessing_level) == "clustering",
      is_external = is_external,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty,
      exclude_imputation = TRUE,
      ...
    ))
  }
)




## get_model_features (character) ----------------------------------------------
setMethod(
  "get_model_features",
  signature(x = "character"),
  function(
    x,
    feature_info_list,
    is_clustered,
    exclude_signature = FALSE,
    exclude_novelty = FALSE,
    ...
  ) {
    return(.get_required_features(
      features = x,
      feature_info_list = feature_info_list,
      is_clustered = is_clustered,
      is_external = TRUE,
      exclude_signature = exclude_signature,
      exclude_novelty = exclude_novelty,
      exclude_imputation = TRUE,
      ...
    ))
  }
)



## get_model_features (NULL) ---------------------------------------------------
setMethod(
  "get_model_features",
  signature(x = "NULL"),
  function(x, ...) {
    return(NULL)            
  }
)



## get_model_features (familiarEnsemble) ---------------------------------------
setMethod(
  "get_model_features",
  signature(x = "familiarEnsemble"),
  function(x, ...) {
    return(x@model_features)
  }
)



## get_model_features (familiarModel) ------------------------------------------
setMethod(
  "get_model_features",
  signature(x = "familiarModel"),
  function(x, ...) {
    return(x@model_features)
  }
)


create_data_column_info <- function(settings) {
  
  # Read from settings. If not set, these will be NULL.
  sample_id_column <- settings$data$sample_col
  batch_id_column <- settings$data$batch_col
  series_id_column <- settings$data$series_col
  
  # Replace any missing.
  if (is.null(sample_id_column)) sample_id_column <- NA_character_
  if (is.null(batch_id_column)) batch_id_column <- NA_character_
  if (is.null(series_id_column)) series_id_column <- NA_character_
  
  # Repetition column ids are only internal.
  repetition_id_column <- NA_character_
  
  # Create table
  data_info_table <- data.table::data.table(
    "type" = c(
      "batch_id_column",
      "sample_id_column",
      "series_id_column",
      "repetition_id_column"
    ),
    "internal" = get_id_columns(),
    "external" = c(
      batch_id_column,
      sample_id_column,
      series_id_column,
      repetition_id_column
    )
  )
  
  if (settings$data$outcome_type %in% c("survival", "competing_risk")) {
    
    # Find internal and external outcome column names.
    internal_outcome_columns <- get_outcome_columns(settings$data$outcome_type)
    external_outcome_columns <- settings$data$outcome_col
    
    if (is.null(external_outcome_columns)) {
      external_outcome_columns <- c(NA_character_, NA_character_)
    }
    
    # Add to table
    outcome_info_table <- data.table::data.table(
      "type" = c("outcome_column", "outcome_column"),
      "internal" = internal_outcome_columns,
      "external" = external_outcome_columns
    )
    
  } else if (settings$data$outcome_type %in% c("binomial", "multinomial", "continuous")) {
    
    # Find internal and external outcome column names.
    internal_outcome_columns <- get_outcome_columns(settings$data$outcome_type)
    external_outcome_columns <- settings$data$outcome_col
    
    if (is.null(external_outcome_columns)) {
      external_outcome_columns <- c(NA_character_)
    }
    
    # Add to table
    outcome_info_table <- data.table::data.table(
      "type" = "outcome_column",
      "internal" = internal_outcome_columns,
      "external" = external_outcome_columns
    )
    
  } else if (settings$data$outcome_type %in% c("unsupervised")) {
    
    # There is no outcome for unsupervised learners.
    outcome_info_table <- NULL
    
  } else {
    ..error_no_known_outcome_type(outcome_type = settings$data$outcome_type)
  }
  
  # Combine into one table and add to object
  return(rbind(data_info_table, outcome_info_table))
}



# add_model_name (familiarDataElement, dataObject)------------------------
setMethod(
  "add_model_name",
  signature(
    data = "familiarDataElement",
    object = "dataObject"
  ),
  function(data, object) {
    
    data <- add_data_element_identifier(
      x = data,
      model_name = "custom_data"
    )[[1L]]
    
    return(data)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarRFSRC ----------------------------------------------------------------
setClass("familiarRFSRC",
  contains = "familiarModel",
  slots = list("seed" = "ANY"),
  prototype = list("seed" = NULL)
)



# familiarRFSRCDefault ---------------------------------------------------------
setClass("familiarRFSRCDefault",
  contains = "familiarRFSRC"
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarRFSRC"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set package
    .Object@package <- "randomForestSRC"

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available", signature(object = "familiarRFSRC"),
  function(object, ...) {
    # Random forests exists for all outcome types and variable importance
    # methods.
    if (object@outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(TRUE)
  }
)



# get_default_hyperparameters (familiarRFSRC) ----------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarRFSRC"),
  function(object, data = NULL, user_list = NULL, ...) {
    # Initialise list and declare hyperparameter entries
    param <- list()
    param$sign_size <- list()
    param$n_tree <- list()
    param$sample_size <- list()
    param$m_try <- list()
    param$node_size <- list()
    param$tree_depth <- list()
    param$n_split <- list()
    param$split_rule <- list()
    param$sample_weighting <- list()
    param$sample_weighting_beta <- list()

    # Variable importance only parameters (are not optimised by hyperparameter
    # optimisation)
    param$fs_vimp_method <- list()

    # If datat is not provided, return the list with hyperparameter names only
    if (is.null(data)) return(param)

    # Get the number of samples
    n_samples <- data.table::uniqueN(data@data, by = get_id_columns(id_depth = "series"))

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data)

    # number of trees ----------------------------------------------------------

    # Note that the number of trees is defined in powers of 2, based on Oshiro,
    # T. M., Perez, P. S., & Baranauskas, J. A. (2012, July). How many trees in
    # a random forest?. In MLDM (pp. 154-168).
    param$n_tree <- .set_hyperparameter(
      default = c(4L, 8L, 10L),
      type = "integer",
      range = c(4L, 10L),
      valid_range = c(0L, Inf),
      randomise = TRUE
    )

    # sample size --------------------------------------------------------------

    # Note that the sample size is here noted as a fraction, which corresponds
    # to the usage in random forest SRC
    param$sample_size <- .set_hyperparameter(
      default = c(0.30, 0.50, 0.70, 0.90),
      type = "numeric",
      range = c(2.0 / n_samples, 1.0),
      valid_range = c(0.0, 1.0),
      randomise = TRUE
    )

    # number of candidate features selected at node ----------------------------

    # Note that the number of features is here noted as a fraction, but is used
    # in randomforestSRC as an integer. Familiar ensures that always at least 1
    # feature is available as a candidate.
    param$m_try <- .set_hyperparameter(
      default = c(0.1, 0.3, 0.5, 1.0),
      type = "numeric",
      range = c(0.0, 1.0),
      randomise = TRUE
    )
    
    # terminal node size -------------------------------------------------------

    # Number of instances in the terminal node. Larger terminal node sizes limit
    # tree depth and overfitting.

    # Define the default range.
    node_size_range <- c(5L, as.integer(max(c(5L, floor(n_samples / 3L)))))

    # Define the default values.
    node_size_default <- c(5L, 10L, 20L, 50L)
    node_size_default <- node_size_default[
      node_size_default >= node_size_range[1L] &
        node_size_default <= node_size_range[2L]
    ]

    # Set the node_size parameter.
    param$node_size <- .set_hyperparameter(
      default = node_size_default,
      type = "integer",
      range = node_size_range,
      valid_range = c(1L, Inf),
      randomise = TRUE
    )

    # maximum tree depth -------------------------------------------------------

    # Determines the depth trees are allowed to grow to. Larger depths increase
    # the risk of overfitting.
    param$tree_depth <- .set_hyperparameter(
      default = c(1L, 2L, 3L, 7L),
      type = "integer",
      range = c(1L, 10L),
      valid_range = c(1L, Inf),
      randomise = TRUE
    )

    # number of split points ---------------------------------------------------

    # The number of split points for each candidate variable is not randomised
    # by default, and deterministic splitting is used.
    param$n_split <- .set_hyperparameter(
      default = 0L,
      type = "integer",
      range = c(0L, 10L),
      valid_range = c(0L, Inf),
      randomise = FALSE
    )

    # splitting rule -----------------------------------------------------------

    # Splitting rule is dependent on the outcome
    if (data@outcome_type %in% c("binomial", "multinomial")) {
      split_rule_range <- c("gini", "auc", "entropy")
      split_rule_default <- "gini"
    } else if (data@outcome_type %in% c("continuous")) {
      split_rule_range <- c("mse", "quantile.regr", "la.quantile.regr")
      split_rule_default <- "mse"
    } else if (data@outcome_type == "survival") {
      split_rule_range <- c("logrank", "logrankscore", "bs.gradient")
      split_rule_default <- "logrank"
    } else if (data@outcome_type == "competing_risk") {
      split_rule_range <- c("logrankCR", "logrank")
      split_rule_default <- "logrankCR"
    } else {
      ..error_no_known_outcome_type(data@outcome_type)
    }

    # Set the split_rule parameter.
    param$split_rule <- .set_hyperparameter(
      default = split_rule_default,
      type = "factor",
      range = split_rule_range,
      randomise = FALSE
    )

    # sample weighting method --------------------------------------------------
    
    # Class imbalances may lead to learning majority classes. This can be
    # partially mitigated by increasing weight of minority classes.
    param$sample_weighting <- .get_default_sample_weighting_method(
      outcome_type = object@outcome_type
    )


    # effective number of samples beta -----------------------------------------
    
    # Specifies the beta parameter for effective number sample weighting method.
    # See Cui et al. (2019).
    param$sample_weighting_beta <- .get_default_sample_weighting_beta(
      method = c(
        param$sample_weighting$init_config,
        user_list$sample_weighting
      ),
      outcome_type = object@outcome_type
    )

    # variable importance method -----------------------------------------------
    param$fs_vimp_method <- .set_hyperparameter(
      default = "permutation",
      type = "factor",
      range = c("permutation", "holdout"),
      randomise = FALSE
    )

    return(param)
  }
)



# get_default_hyperparameters (familiarRFSRCDefault) ---------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarRFSRCDefault"),
  function(
    object, 
    data = NULL, 
    user_list = NULL, 
    ...
  ) {
    # Initialise list and declare hyperparameter entries
    param <- list()
    param$sign_size <- list()
    param$sample_weighting <- list()
    param$sample_weighting_beta <- list()

    # Variable importance only parameters (are not optimised by hyperparameter
    # optimisation)
    param$fs_vimp_method <- list()

    # If data is not provided, return the list with hyperparameter names only.
    if (is.null(data)) return(param)

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data)

    # sample weighting method --------------------------------------------------
    
    # Class imbalances may lead to learning majority classes. This can be
    # partially mitigated by increasing weight of minority classes.
    param$sample_weighting <- .get_default_sample_weighting_method(
      outcome_type = object@outcome_type
    )

    # effective number of samples beta -----------------------------------------
    
    # Specifies the beta parameter for effective number sample weighting method.
    # See Cui et al. (2019).
    param$sample_weighting_beta <- .get_default_sample_weighting_beta(
      method = c(
        param$sample_weighting$init_config,
        user_list$sample_weighting
      ),
      outcome_type = object@outcome_type
    )

    # variable importance method -----------------------------------------------
    param$fs_vimp_method <- .set_hyperparameter(
      default = "permutation",
      type = "factor",
      range = c("permutation", "holdout"),
      randomise = FALSE
    )

    return(param)
  }
)



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarRFSRC"),
  function(object, type = "default") {
    if (object@outcome_type == "survival") {
      if (type == "default") {
        # Standard predictions.
        return("cumulative_hazard")
      } else if (type %in% c("survival", "survival_probability")) {
        return("survival_probability")
      } else if (type %in% c("response", "cumulative_hazard")) {
        return("cumulative_hazard")
      } else {
        ..error(paste0("Prediction type is not implemented: ", type))
      }
      
    } else {
      return(callNextMethod())
    }
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train", 
  signature(
    object = "familiarRFSRC",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    anonymous = TRUE, 
    ...
  ) {
    # Aggregate repeated measurement data - randomForestSRC does not facilitate
    # repeated measurements.
    data <- aggregate_data(data = data)

    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object, 
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Find feature columns in data table
    feature_columns <- get_feature_columns(x = data)

    # Parse formula.
    if (object@outcome_type == "survival") {
      Surv <- survival::Surv
      formula <- stats::reformulate(
        termlabels = feature_columns,
        response = quote(Surv(outcome_time, outcome_event))
      )
      
    } else if (object@outcome_type %in% c("binomial", "multinomial", "continuous")) {
      formula <- stats::reformulate(
        termlabels = feature_columns,
        response = quote(outcome)
      )
      
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }

    # Extract hyperparameters from the model object.
    param <- object@hyperparameters

    # Determine the sample size
    sample_size <- ceiling(param$sample_size * nrow(data@data))
    sample_type <- ifelse(sample_size == nrow(data@data), "swr", "swor")

    # Set weights.
    weights <- create_instance_weights(
      data = data,
      method = object@hyperparameters$sample_weighting,
      beta = ..compute_effective_number_of_samples_beta(
        object@hyperparameters$sample_weighting_beta
      ),
      normalisation = "average_one"
    )

    # Set forest seed. If object comes with a defined seed use the seed.
    forest_seed <- ifelse(
      is.null(object@seed),
      as.integer(stats::runif(1L, -100000L, -1L)),
      object@seed
    )

    # Use anonymised version for the forest, if available.
    if (utils::packageVersion("randomForestSRC") >= "2.11.0" && anonymous) {
      forest_function <- randomForestSRC::rfsrc.anonymous
    } else {
      forest_function <- randomForestSRC::rfsrc
    }

    # Set learner arguments.
    learner_arguments <- list(formula,
      "data" = data@data,
      "case.wt" = weights,
      "seed" = forest_seed
    )

    # Add additional hyperparameters for models that use them.
    if (!is(object, "familiarRFSRCDefault")) {
      learner_arguments <- c(
        learner_arguments,
        list(
          "ntree" = 2L^param$n_tree,
          "samptype" = sample_type,
          "sampsize" = sample_size,
          "mtry" = max(c(1L, ceiling(param$m_try * length(feature_columns)))),
          "nodesize" = param$node_size,
          "nodedepth" = param$tree_depth,
          "nsplit" = param$n_split,
          "splitrule" = as.character(param$split_rule)
        )
      )
    }
    
    # Train the model.
    model <- do.call_with_handlers(
      forest_function,
      args = learner_arguments
    )

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) return(callNextMethod(object = object))

    # Add model to the object.
    object@model <- model

    # Store the seed used to create the object. This helps recreate the object
    # in case the model is an anonymous forest.
    object@seed <- forest_seed

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarRFSRC",
    data = "dataObject"
  ),
  function(object, data, ...) {
    if (object@outcome_type %in% c("continuous", "binomial", "multinomial")) {
      # Turn into a Naive model.
      object <- methods::new("familiarNaiveModel", object)
      
    } else if (object@outcome_type %in% c("survival")) {
      # Turn into a Naive cumulative incidence model.
      object <- methods::new("familiarNaiveCumulativeIncidenceModel", object)
    }

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarRFSRC",
    data = "dataObject"
  ),
  function(
    object,
    data, 
    type = "default",
    time = NULL, 
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(NULL)
    }
    
    # Check if the data is empty.
    if (is_empty(data)) {
      return(NULL)
    }
    
    if (type %in% c("default", "survival_probability")) {
      # Default method ---------------------------------------------------------

      # Make predictions using the model.
      model_predictions <- predict(
        object = object@model,
        newdata = data@data
      )

      if (object@outcome_type %in% c("binomial", "multinomial")) {
        # categorical outcomes -------------------------------------------------

        # Obtain class levels.
        class_levels <- colnames(model_predictions$predicted)
        
        # Add class probabilities.
        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          prediction_list[[class_levels[ii]]] <- model_predictions$predicted[, ii]
        }
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("continuous")) {
        # numerical outcomes ---------------------------------------------------

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = as.numeric(model_predictions$predicted),
          type = "regression",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("survival")) {
        # survival outcomes ----------------------------------------------------

        # Get the unique event times
        event_times <- model_predictions$time.interest

        # Set default time, if not provided.
        time <- ifelse(is.null(time), max(event_times), time)

        if (type == "default") {
          # Cumulative hazard.

          # Get the cumulative hazards at the given time point.
          prediction_table <- .random_forest_survival_predictions(
            object = object,
            event_matrix = model_predictions$chf,
            event_times = event_times,
            data = data,
            time = time,
            type = "cumulative_hazard"
          )
          
        } else if (type == "survival_probability") {
          # Survival probability.

          # Get the survival probability at the given time point.
          prediction_table <- .random_forest_survival_predictions(
            object = object,
            event_matrix = model_predictions$survival,
            event_times = event_times,
            data = data,
            time = time,
            type = "survival"
          )
          
        } else {
          ..error_outcome_type_not_implemented(object@outcome_type)
        }
      }

      return(prediction_table)
      
    } else if (!.is_available_prediction_type(type)) {
      # User-specified method --------------------------------------------------

      # Make predictions using the model.
      return(predict(
        object = object@model,
        newdata = data@data,
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



# ..get_prediction_table_type --------------------------------------------------
setMethod(
  "..get_prediction_table_type",
  signature(object = "familiarRFSRC"),
  function(object, type, ...) {
    prediction_table_type <- NULL
    if (object@outcome_type %in% c("survival") && type == "default") {
      prediction_table_type <- "cumulative_hazard"
    } else {
      prediction_table_type <- callNextMethod()
    }
    
    return(prediction_table_type)
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarRFSRC"),
  function(object, data = NULL, ...) {
    
    # Attempt to train the model if it has not been trained yet.
    if (!model_is_trained(object)) {
      object <- .train(
        object = object,
        data = data,
        get_additional_info = FALSE,
        trim_model = FALSE,
        anonymous = FALSE
      )
    }

    # Check if the model has been trained upon retry.
    if (!model_is_trained(object)) return(callNextMethod())

    # Check that required packages are loaded and installed.
    require_package(object, "vimp")

    # Find the vimp_method specified.
    vimp_method <- object@hyperparameters$fs_vimp_method
    if (is.null(vimp_method)) {
      ..error_reached_unreachable_code(paste0(
        "..vimp,familiarRFSRC: vimp method was not specified as a ",
        "hyperparameter (fs_vimp_method)"
      ))
    }
    
    vimp_method <- as.character(vimp_method)

    # Extract the variable importance score
    if (vimp_method == "permutation") {
      # Check if the model is anonymous, and rebuild if it is. VIMP does not
      # work otherwise.
      if (inherits(object@model, "anonymous")) {
        object <- .train(
          object = object,
          data = data,
          get_additional_info = FALSE,
          trim_model = FALSE,
          anonymous = FALSE
        )

        # Check that the non-anonymous model is trained.
        if (!model_is_trained(object)) return(callNextMethod())
      }

      # Determine permutation variable importance
      vimp_score <- randomForestSRC::vimp(
        object = object@model,
        importance = "permute"
      )$importance

      # Check that the variable importance score is not empty.
      if (is_empty(vimp_score)) return(callNextMethod())

      # The variable importance score for binomial and multinomial
      # outcomes is per class
      if (is.matrix(vimp_score)) {
        vimp_score_names <- rownames(vimp_score)
        vimp_score <- vimp_score[, 1L]
        
      } else {
        vimp_score_names <- names(vimp_score)
      }

      # Create variable importance object.
      vimp_object <- methods::new(
        "vimpTable",
        vimp_table = data.table::data.table(
          "score" = vimp_score, 
          "name" = vimp_score_names
        ),
        score_aggregation = "max",
        invert = TRUE
      )

      return(vimp_object)
      
    } else if (vimp_method == "minimum_depth") {
      ..deprecation_rfsrc_minimum_depth()
      return(callNextMethod())
      
    } else if (vimp_method == "variable_hunting") {
      ..deprecation_rfsrc_variable_hunting()
      return(callNextMethod())
      
    } else if (vimp_method == "holdout") {
      # Check that data is present.
      if (is_empty(data)) {
        warning(paste0(
          "Data is required to assess variable importance using ",
          "randomForestRFSRC::holdout.vimp"
        ))
        return(callNextMethod())
      }

      if (get_n_features(data) <= 1L) {
        warning(paste0(
          "Variable importance using randomForestRFSRC::holdout.vimp ",
          "requires more than 1 feature."
        ))

        # Create variable importance object.
        vimp_object <- methods::new(
          "vimpTable",
          vimp_table = data.table::data.table(
            "score" = 0.0, 
            "name" = get_feature_columns(x = data)
          ),
          score_aggregation = "max",
          invert = TRUE
        )

        return(vimp_object)
      }

      # Find feature columns in the data.
      feature_columns <- get_feature_columns(x = data)

      # Parse formula.
      if (object@outcome_type == "survival") {
        # Creating a local instance of survival::Surv() prevents an
        # error.
        Surv <- survival::Surv
        formula <- stats::reformulate(
          termlabels = feature_columns,
          response = quote(Surv(outcome_time, outcome_event))
        )
        
      } else if (object@outcome_type %in% c("binomial", "multinomial", "continuous")) {
        formula <- stats::reformulate(
          termlabels = feature_columns,
          response = quote(outcome)
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      # Get arguments for the holdout.vimp learner.
      learner_arguments <- list(
        formula,
        "data" = data@data,
        "verbose" = FALSE
      )

      # Get additional variables for the optimised learner.
      if (!is(object, "familiarRFSRCDefault")) {
        # Determine the sample size
        sample_size <- ceiling(object@hyperparameters$sample_size * nrow(data@data))
        sample_type <- ifelse(sample_size == nrow(data@data), "swr", "swor")

        learner_arguments <- c(
          learner_arguments,
          list(
            "mtry" = max(c(1L, ceiling(object@hyperparameters$m_try * get_n_features(data)))),
            "samptype" = sample_type,
            "sampsize" = sample_size,
            "nodesize" = object@hyperparameters$node_size,
            "nodedepth" = object@hyperparameters$tree_depth,
            "nsplit" = object@hyperparameters$n_split,
            "splitrule" = as.character(object@hyperparameters$split_rule)
          )
        )
      }

      # Perform holdout variable importance.
      vimp_score <- do.call(
        randomForestSRC::holdout.vimp,
        args = learner_arguments
      )$importance

      # Check that the variable importance score is not empty.
      if (is_empty(vimp_score)) return(callNextMethod())

      # Select the "all" column, which is the first column
      if (is.matrix(vimp_score)) {
        vimp_score_names <- rownames(vimp_score)
        vimp_score <- vimp_score[, 1L]
        
      } else {
        vimp_score_names <- names(vimp_score)
      }

      # Create variable importance object.
      vimp_object <- methods::new(
        "vimpTable",
        vimp_table = data.table::data.table(
          "score" = vimp_score,
          "name" = vimp_score_names
        ),
        score_aggregation = "max",
        invert = TRUE
      )

      return(vimp_object)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "..vimp,familiarRFSRC: unknown vimp method was specified: ", vimp_method
      ))
    }

    return(NULL)
  }
)



# ..set_calibration_info -------------------------------------------------------
setMethod(
  "..set_calibration_info",
  signature(object = "familiarRFSRC"),
  function(object, data) {
    # Check if calibration info already.
    if (has_calibration_info(object)) return(object)

    if (object@outcome_type == "survival") {
      # Determine baseline survival.
      object@calibration_info <- get_baseline_survival(data = data)
      
    } else {
      return(callNextMethod())
    }

    return(object)
  }
)



# ..set_vimp_parameters --------------------------------------------------------
setMethod(
  "..set_vimp_parameters",
  signature(object = "familiarRFSRC"),
  function(
    object,
    method,
    ...
  ) {
    # Determine variable importance method
    if (startswith_any(
      method,
      prefix = c(
        "random_forest_permutation",
        "random_forest_rfsrc_permutation"
      )
    )) {
      vimp_method <- "permutation"
      
    } else if (startswith_any(
      method,
      prefix = c(
        "random_forest_minimum_depth",
        "random_forest_rfsrc_minimum_depth"
      )
    )) {
      ..deprecation_rfsrc_minimum_depth(as_error = TRUE)
      
    } else if (startswith_any(
      method,
      prefix = c(
        "random_forest_variable_hunting",
        "random_forest_rfsrc_variable_hunting"
      )
    )) {
      ..deprecation_rfsrc_variable_hunting(as_error = TRUE)
      
    } else if (startswith_any(
      method,
      prefix = c(
        "random_forest_holdout",
        "random_forest_rfsrc_holdout"
      )
    )) {
      vimp_method <- "holdout"
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "..set_vimp_parameters,familiarRFSRC: unknown feature selection ",
        "method was specified."
      ))
    }

    # Check if a list of hyperparameters is already present.
    if (is.null(object@hyperparameters)) {
      hyperparameters <- list()
    } else {
      hyperparameters <- object@hyperparameters
    }

    # Update the fs_vimp_method hyperparameters.
    hyperparameters$fs_vimp_method <- vimp_method

    # Store in the object
    object@hyperparameters <- hyperparameters

    return(object)
  }
)



# .trim_model-------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarRFSRC"),
  function(object, ...) {
    # Update model by removing the call.
    object@model$call <- call("trimmed")

    # Add show.
    object <- .capture_show(object)

    # Remove the predictions.
    object@model$predicted <- NULL
    object@model$predicted.oob <- NULL

    # Remove yvar due to redundancy.
    object@model$yvar <- NULL

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    return(object)
  }
)



.get_available_rfsrc_learners <- function(show_general = TRUE) {
  return(c("random_forest", "random_forest_rfsrc"))
}



.get_available_rfsrc_default_learners <- function(show_general = TRUE) {
  return(paste0(.get_available_rfsrc_learners(show_general = show_general), "_default"))
}



.get_available_rfsrc_vimp_methods <- function(show_general = TRUE) {
  return(c(
    "random_forest_permutation", "random_forest_rfsrc_permutation",
    "random_forest_holdout", "random_forest_rfsrc_holdout"
  ))
}



.get_available_rfsrc_default_vimp_methods <- function(show_general = TRUE) {
  return(paste0(.get_available_rfsrc_vimp_methods(show_general = show_general), "_default"))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarCoxPH object ---------------------------------------------------------
setClass("familiarCoxPH",
  contains = "familiarModel",
  slots = list("encoding_reference_table" = "ANY"),
  prototype = list("encoding_reference_table" = NULL)
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarCoxPH"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set the required package
    .Object@package <- "survival"

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarCoxPH"),
  function(object, ...) {
    # CoxPH is only available if for survival outcomes.
    return(object@outcome_type == "survival")
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarCoxPH"),
  function(object, data = NULL, ...) {
    # Initialise list and declare hyperparameter entries
    param <- list()
    param$sign_size <- list()

    # If data is explicitly NULL, return the list with hyperparameter names only
    if (is.null(data)) return(param)

    ## signature size ----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data, restrict_samples = TRUE)

    return(param)
  }
)



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarCoxPH"),
  function(object, type = "default") {
    # Cox proportional hazards models predict relative risks
    if (type == "default") {
      return("hazard_ratio")
      
    } else if (type == "survival_probability") {
      return("survival_probability")
      
    } else {
      ..error_reached_unreachable_code(
        "get_prediction_type,familiarCoxPH: unknown type"
      )
    }
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarCoxPH",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object,
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Use effect coding to convert categorical data into encoded data - this is
    # required to deal with factors with missing/new levels between training and
    # test data sets.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )

    # Find feature columns in the data.
    feature_columns <- get_feature_columns(x = encoded_data$encoded_data)

    # Parse formula
    formula <- stats::reformulate(
      termlabels = feature_columns,
      response = quote(survival::Surv(outcome_time, outcome_event))
    )

    # Generate model.
    model_control <- survival::coxph.control(iter.max = 100L)

    # Train the model
    model <- do.call_with_handlers(
      survival::coxph,
      args = list(
        formula,
        "data" = encoded_data$encoded_data@data,
        "control" = model_control,
        "y" = FALSE
      )
    )

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) {
      return(callNextMethod(object = object))
    }

    # Check if the model fitter converged in time.
    if (model$iter >= 100L) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        "Model fitter ran out of iterations and did not converge."
      )))
    }

    # Check if all coefficients could not be estimated. Sometimes models could
    # be trained with a large number of features whose coefficients fail to
    # converge. This would sometimes lead to overly large signature sizes being
    # selected during hyperparameter optimisation, especially in situations
    # where there is not a lot of signal. Checking for non-finite coefficients
    # is an easy way to figure out if the model is not properly trained.
    if (!all(sapply(stats::coef(model), is.finite))) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_failed_model_coefficient_estimation()
      )))
    }

    # Add model
    object@model <- model

    # Add the contrast references to model_list
    object@encoding_reference_table <- encoded_data$reference_table

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarCoxPH",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Turn into a Naive model.
    object <- methods::new("familiarNaiveCoxModel", object)

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarCoxPH",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "default",
    time = NULL,
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")
    
    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }
    
    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }
    
    if (type == "default") {
      # default ----------------------------------------------------------------

      # Encode data so that the features are the same as in the training.
      encoded_data <- encode_categorical_variables(
        data = data,
        object = object,
        encoding_method = "dummy",
        drop_levels = FALSE
      )

      # Use the model for prediction.
      model_predictions <- predict(
        object = object@model,
        newdata = encoded_data$encoded_data@data,
        type = "risk"
      )
      
      # Store as prediction table.
      prediction_table <- as_prediction_table(
        x = model_predictions,
        type = "hazard_ratio",
        data = data,
        model_object = object
      )
      
      return(prediction_table)
      
    } else if (type == "survival_probability") {
      # survival probability ---------------------------------------------------
      
      # If time is unset, read the max time stored by the model.
      if (is.null(time)) time <- object@settings$time_max
      
      return(.survival_probability_relative_risk(
        object = object,
        data = data,
        time = time
      ))
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified ---------------------------------------------------------

      # Encode data so that the features are the same as in the training.
      encoded_data <- encode_categorical_variables(
        data = data,
        object = object,
        encoding_method = "dummy",
        drop_levels = FALSE
      )

      return(predict(
        object = object@model,
        newdata = encoded_data$encoded_data@data,
        type = type,
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



#  ..vimp ----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarCoxPH"),
  function(object, ...) {
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }

    # Check that required packages are loaded and installed.
    require_package(object, "vimp")

    # Define p-values
    coefficient_z_values <- tryCatch(
      .compute_z_statistic(object),
      error = identity
    )

    if (inherits(coefficient_z_values, "error")) {
      return(callNextMethod())
    }

    # Remove any coefficients for the intercept.
    coefficient_z_values <- coefficient_z_values[names(coefficient_z_values) != "(Intercept)"]

    if (length(coefficient_z_values) == 0L) {
      return(callNextMethod())
    }

    # Create variable importance object.
    vimp_object <- methods::new("vimpTable",
      vimp_table = data.table::data.table(
        "score" = abs(coefficient_z_values),
        "name" = names(coefficient_z_values)
      ),
      encoding_table = object@encoding_reference_table,
      score_aggregation = "max",
      invert = TRUE
    )

    return(vimp_object)
  }
)



# ..set_calibration_info -------------------------------------------------------
setMethod(
  "..set_calibration_info",
  signature(object = "familiarCoxPH"),
  function(object, data) {
    # Check if calibration info already.
    if (has_calibration_info(object)) {
      return(object)
    }

    if (object@outcome_type == "survival") {
      # Determine baseline survival.
      object@calibration_info <- get_baseline_survival(data = data)
      
    } else {
      return(callNextMethod())
    }

    return(object)
  }
)



# .trim_model ------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarCoxPH"),
  function(object, ...) {
    # Update model by removing the call.
    object@model$call <- call("trimmed")

    # Add show.
    object <- .capture_show(object)

    # Remove .Environment.
    object@model$terms <- .replace_environment(object@model$terms)
    object@model$formula <- .replace_environment(object@model$formula)

    # Remove elements that contain sample-specific values.
    object@model$linear.predictors <- NULL
    object@model$residuals <- NULL

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    # Default method for models that lack a more specific method.
    return(object)
  }
)



.get_available_cox_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c("cox")
  
  return(learners)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


as_familiar_plot <- function(
    p = NULL,
    g = NULL,
    layout
) {
  fam_plot <- methods::new("familiarPlot")
  
  if (is.null(g) && ggplot2::is_ggplot(p)) {
    # Get gtable from ggplot2 object.
    fam_plot@gtable <- .convert_to_grob(p)
    
  } else if (gtable::is.gtable(g)) {
    fam_plot@gtable <- g
  }
  
  # Set column and row id.
  fam_plot@row_id <- layout$row_id
  fam_plot@col_id <- layout$col_id
  
  # Add global plot elements.
  fam_plot@global_elements <- .extract_global_plot_elements(g)
  
  # Make panels inherit heights and widths, if they don't have any. This is done
  # to ensure that panels retain heights and widths, even if supporting elements
  # such as the axis text and label elements are stripped on figure composition.
  fam_plot@gtable <- .gtable_update_panel_aspects(fam_plot@gtable)
  
  return(fam_plot)
}



.extract_global_plot_elements <- function(g) {
  
  element_list <- list()
  
  # Export list of elements.
  if (is.null(g)) return(element_list)
  
  # Find names of all existing elements.
  elements_names <- g$layout$name
  
  # Set names of all global elements.
  global_elements <- c(
    .all_gtable_guide_names(),
    .all_gtable_strip_x_names(),
    .all_gtable_strip_y_names(),
    .all_gtable_label_x_names(),
    .all_gtable_label_y_names(),
    .all_gtable_title_names()
  )
  
  # Identify which global elements are present.
  present_elements <- elements_names[sapply(
    elements_names, 
    startswith_any, 
    prefix = global_elements
  )]
  if (length(present_elements) == 0L) return(element_list)
  
  # Add elements that are present in the table and are related to the global
  # elements.
  for (present_element in present_elements) {
    element_list[[present_element]] <- .gtable_extract_grob(
      g = g,
      element = present_element
    )
  }
  
  # # Filter zeroGrob and nullGrob classes, which represent empty elements.
  element_list <- element_list[!
    sapply(
      lapply(element_list, class),
      function(ii) any(ii %in% c("zeroGrob", "nullGrob"))
    )
  ] 
  
  return(element_list)
}



.create_placeholder_figure <- function(
    template_figure_row,
    template_figure_col,
    row_id,
    col_id
) {
  # Creates placeholder for missing figures in faceted panel, e.g. because no
  # data were present.
  if (
    !is(template_figure_row, "familiarPlot") ||
    !is(template_figure_col, "familiarPlot")
  ) {
    ..error_reached_unreachable_code("both templates should be familiarPlot objects.")
  }
  
  # Use the row item as the initial template.
  figure <- template_figure_row
  
  # Ensure that panels are removed.
  figure@remove_panel <- TRUE
  
  # Drop global plot elements -- we will extract these again later.
  figure@global_elements <- NULL
  
  # We need to update elements from the column template, e.g. axis-t, xlab-t,
  # and strip-t-1.
  col_element_names <- c(
    .all_gtable_strip_x_names(),
    .all_gtable_label_x_names(),
    .all_gtable_axis_x_names()
  )
  
  # Find names of all existing elements.
  updatable_elements <- figure@gtable$layout$name
  updatable_elements <- updatable_elements[sapply(
    updatable_elements, 
    startswith_any, 
    prefix = col_element_names
  )]
  
  for (update_element in updatable_elements) {
    figure@gtable <- .gtable_insert(
      g = figure@gtable,
      g_new = .gtable_extract_grob(template_figure_col@gtable, element = update_element),
      where = c("replace", update_element)
    )
  }
  
  # Update row_id and col_id.
  figure@row_id <- row_id
  figure@col_id <- col_id
  
  # Add global elements again.
  figure@global_elements <- .extract_global_plot_elements(figure@gtable)
  
  return(figure)
}



.set_figure_element_removal <- function(
  object,
  top_row_id,
  bottow_row_id,
  left_col_id,
  right_col_id,
  x_text_shared,
  y_text_shared,
  x_label_shared,
  y_label_shared
) {
  is_top_row <- object@row_id == top_row_id
  is_bottom_row <- object@row_id == bottow_row_id
  is_left_col <- object@col_id == left_col_id
  is_right_col <- object@col_id == right_col_id
  
  # Facet strips
  if (!is_top_row) {
    object@remove_strip_x <- TRUE
  }
  if (!is_right_col) {
    object@remove_strip_y <- TRUE
  }
  
  # x-axis text. "individual" and "FALSE" do not lead to removal.
  if (x_text_shared %in% c("overall", "TRUE")) {
    object@remove_axis_text_x <- TRUE
    
  } else if (x_text_shared == "column" && !is_bottom_row) {
    object@remove_axis_text_x <- TRUE
  }
  
  # x-axis label. "individual" and "FALSE" do not lead to removal.
  if (x_label_shared %in% c("overall", "TRUE")) {
    object@remove_axis_label_x <- TRUE
    
  } else if (x_label_shared == "column" && !is_bottom_row) {
    object@remove_axis_label_x <- TRUE
  }
  
  # y-axis text. "individual" and "FALSE" do not lead to removal.
  if (y_text_shared %in% c("overall", "TRUE")) {
    object@remove_axis_text_y <- TRUE
    
  } else if (y_text_shared == "row" && !is_left_col) {
    object@remove_axis_text_y <- TRUE
  }
  
  # x-axis label. "individual" and "FALSE" do not lead to removal.
  if (y_label_shared %in% c("overall", "TRUE")) {
    object@remove_axis_label_y <- TRUE
    
  } else if (y_label_shared == "row" && !is_left_col) {
    object@remove_axis_label_y <- TRUE
  }
  
  return(object)
}



.remove_figure_elements <- function(
  object,
  replace_by_zero_grob = FALSE
) {
  
  # Always remove guide and title.
  base_elements <- c(
    .all_gtable_guide_names(),
    .all_gtable_title_names()
  )
  
  # First determine which stuff can be removed, and then match any elements in
  # the gtable.
  if (object@remove_strip_x) {
    base_elements <- c(base_elements, .all_gtable_strip_x_names())
  }
  if (object@remove_strip_y) {
    base_elements <- c(base_elements, .all_gtable_strip_y_names())
  }
  if (object@remove_axis_text_x) {
    base_elements <- c(base_elements, .all_gtable_axis_x_names())
  }
  if (object@remove_axis_text_y) {
    base_elements <- c(base_elements, .all_gtable_axis_y_names())
  }
  if (object@remove_axis_label_x) {
    base_elements <- c(base_elements, .all_gtable_label_x_names())
  }
  if (object@remove_axis_label_y) {
    base_elements <- c(base_elements, .all_gtable_label_y_names())
  }
  if (object@remove_panel) {
    base_elements <- c(base_elements, .all_gtable_panel_names())
  }
  
  removable_elements <- object@gtable$layout$name
  removable_elements <- removable_elements[sapply(
    removable_elements, 
    startswith_any, 
    prefix = base_elements
  )]
  
  # Iterate to remove or replace with zeroGrob. Any zeroGrobs that remain will
  # be removed when composing the figure (.compose_figure).
  zeroGrob <- ggplot2::zeroGrob()
  for (removable_element in removable_elements) {
    if (replace_by_zero_grob) {
      object@gtable <- .gtable_insert(
        g = object@gtable,
        g_new = list(zeroGrob),
        where = c("replace", removable_element)
      )
      
    } else {
      object@gtable <- .gtable_remove(
        g = object@gtable,
        removed_element = removable_element
      )
    }
  }
  
  # Update widths and heights.
  object@gtable <- .gtable_update_layout(g = object@gtable)
  
  return(object)
}

..trim_functions <- function() {
  return(list(
    "coef" = list(
      "FUN" = stats::coef,
      "with_timeout" = FALSE
    ),
    "vcov" = list(
      "FUN" = stats::vcov,
      "with_timeout" = TRUE
    ),
    "summary" = list(
      "FUN" = summary,
      "with_timeout" = TRUE
    )
  ))
}



.capture_show <- function(object) {
  # Check that required packages are loaded and installed.
  require_package(object, "show")

  # Capture message
  captured_message <- suppressWarnings(tryCatch(
    utils::capture.output(show(object@model), file = NULL),
    error = identity
  ))

  # Check if show generated an error.
  if (inherits(captured_message, "error")) return(object)

  # Remove call.
  call_present <- startsWith(x = captured_message, prefix = "Call:")
  if (any(call_present)) {
    call_present <- which(call_present)
    call_present <- c(call_present - 1L, call_present)
    call_present <- setdiff(call_present, 0L)

    # Remove lines from message.
    captured_message <- captured_message[-c(call_present)]
  }

  # Add to trimmed functions.
  object@trimmed_function <- c(
    object@trimmed_function,
    list("show" = captured_message)
  )

  return(object)
}



.replace_broken_functions <- function(object, trimmed_object, timeout = 60000.0) {
  # Find all methods that are trimmable.
  trimmable_methods <- names(..trim_functions())
  
  # Load packages associated with the model into the namespace so that all
  # associated methods may be found.
  require_package(object)

  # Find all methods associated with the model object itself.
  class_methods <- .get_class_methods(object@model)

  # Special check for S3 methods defined outside of standard packages, or
  # dependent packages. There seems to be an issue with visibility of such
  # methods when the package is only loaded, not attached. I don't want to
  # attach packages because that may alter the search space of the user.
  if (inherits(object@model, "mboost")) {
    class_methods <- unique(c(class_methods, "varimp"))
  }

  # Find those methods that are actually associated with the class.
  trimmable_methods <- trimmable_methods[trimmable_methods %in% class_methods]

  # Return trimmed object if there are no associated trimmable functions.
  if (is_empty(trimmable_methods)) return(trimmed_object)

  trimmable_methods <- ..trim_functions()[trimmable_methods]

  # Generate replacement functions, if required.
  replacement_functions <- lapply(trimmable_methods,
    ..replace_broken_function,
    object = object,
    trimmed_object = trimmed_object,
    timeout = timeout
  )

  # Get non-null items and add to existing functions.
  trimmed_object@trimmed_function <- c(
    trimmed_object@trimmed_function,
    replacement_functions[!sapply(replacement_functions, is.null)]
  )

  return(trimmed_object)
}



..replace_broken_function <- function(method_list, object, trimmed_object, timeout) {
  # Isolate FUN and with_timeout:
  FUN <- method_list$FUN
  if (!method_list$with_timeout) timeout <- Inf

  # Apply function using the original, untrimmed object.
  initial_info <- do.call_with_handlers_timeout(
    FUN,
    args = list(object@model),
    timeout = timeout,
    additional_packages = object@package
  )

  # If an error occurs or the project times out the required function is not
  # considered implemented for the object.
  if (!is.null(initial_info$error) || initial_info$timeout == TRUE) {
    return(NULL)
  }
  initial_info <- initial_info$value
  
  # Attempt to extract the results from the trimmed object.
  new_info <- do.call_external(
    what = do.call_with_handlers,
    args = list(
      "what" = FUN,
      "args" = list(trimmed_object@model)
    ),
    additional_packages = object@package
  )

  # If an error occurs, it means that the information required to create the
  # function is no longer available due to object trimming, or recreating the
  # object takes a long time.
  if (!is.null(new_info$error)) {
    # Check for elements that contain stuff.
    if (is.list(initial_info)) {
      # This is the generic check for S3-methods.
      if (!is.null(names(initial_info))) {
        # Check for call.
        if (is.call(initial_info$call)) {
          initial_info$call <- call("trimmed")
        }

        # Check for formula.
        if (inherits(initial_info$formula, "formula")) {
          initial_info$formula <- .replace_environment(initial_info$formula)
        }

        # Check for terms.
        if (inherits(initial_info$terms, "terms")) {
          initial_info$terms <- .replace_environment(initial_info$terms)
        }
      }
      
    } else if (is(initial_info, "summary.vglm")) {
      # This is the check for the summary.vglm S4 object.

      # Check for call.
      if (is.call(initial_info@call)) initial_info@call <- call("trimmed")

      # Check for terms.
      initial_info@terms <- lapply(
        initial_info@terms,
        function(x) {
          if (!inherits(x, "terms")) return(x)

          return(.replace_environment(x))
        }
      )

      # Check for formula.
      if (inherits(initial_info@misc$formula, "formula")) {
        initial_info@misc$formula <- .replace_environment(initial_info@misc$formula)
      }

      # Replace qr attribute.
      initial_info@qr <- list()
    }

    return(initial_info)
  }

  return(NULL)
}



.replace_environment <- function(x) {
  if (".Environment" %in% names(attributes(x))) {
    attr(x, ".Environment") <- rlang::base_env()
  }

  return(x)
}



.remove <- function(..., envir) {
  # Remove a variable from a specific environment.

  dots <- unlist(list(...))

  for (current_object in dots) {
    if (is_empty(current_object)) next

    if (exists(current_object, envir = envir, inherits = FALSE)) {
      remove(
        list = current_object, 
        envir = envir
      )
    }
  }
}



.duplicate_environment <- function(envir, parent = parent.env(envir)) {
  # Duplicate environment
  duplicated_environment <- list2env(
    as.list.environment(envir,
      all.names = TRUE,
      sorted = FALSE
    ),
    parent = parent
  )

  return(duplicated_environment)
}



.change_environment <- function(x, old_env, new_env, recursive = TRUE) {
  if (is.environment(x)) {
    # Iterate over elements in the environment.
    sapply(
      ls(x),
      function(y, x, old_env, new_env, recursive) {
        # Import locally.
        local_x <- get(y, envir = x)
        
        # Change environment on variable.
        local_x <- .change_environment(
          x = local_x,
          old_env = old_env,
          new_env = new_env,
          recursive = recursive
        )
        
        # Re-assign to environment.
        assign(y, local_x, envir = x)
        
        return(invisible(NULL))
      },
      x = x,
      old_env = old_env,
      new_env = new_env,
      recursive = recursive
    )

    return(invisible(NULL))
    
  } else if (is.list(x)) {
    if (length(x) == 0L) return(x)

    # Determine class (if any).
    object_class <- class(x)

    # Skip if x is an extree_data object.
    if (inherits(x, "extree_data")) return(x)

    # Iterate over list elements.
    x <- lapply(
      x, 
      function(x, old_env, new_env, recursive) {
        # Update the environment of functions.
        if (is.function(x)) {
          if (identical(environment(x), old_env)) environment(x) <- new_env
          
        } else if (rlang::is_bare_list(x) && recursive) {
          x <- .change_environment(
            x,
            old_env = old_env,
            new_env = new_env,
            recursive = recursive
          )
        }
        
        return(x)
      },
      old_env = old_env,
      new_env = new_env,
      recursive = recursive
    )
    
    class(x) <- object_class
    
  } else {
    if (is.function(x)) {
      if (identical(environment(x), old_env)) environment(x) <- new_env
    }
  }

  return(x)
}

..create_initial_hyperparameter_set <- function(
    parameter_list,
    grid_initialisation_method,
    n_random_sets = 200L
) {
  
  if (length(parameter_list) == 0L) return(NULL)

  if (!.any_randomised_hyperparameters(parameter_list = parameter_list)) {
    # All variables have been fixed. No optimisation is required.

    # Get values for each parameter.
    value_list <- lapply(
      parameter_list, 
      function(list_entry) (list_entry$init_config)
    )

    # Convert to a data table.
    parameter_table <- data.table::as.data.table(value_list)
    
  } else {
    
    if (grid_initialisation_method %in% "random") {
      # Create completely random hyperparameter sets.

      # Create random hyperparameter sets.
      random_list <- lapply(
        seq_len(n_random_sets),
        function(ii, parameter_list) {
          return(..randomise_hyperparameter_set(
            parameter_list = parameter_list,
            local = FALSE
          ))
        },
        parameter_list = parameter_list
      )

      # Combine list into a single data table.
      parameter_table <- data.table::rbindlist(random_list, use.names = TRUE)

      # Allow only unique parameter sets.
      parameter_table <- unique(parameter_table, by = names(parameter_list))
    }

    if (grid_initialisation_method %in% c("fixed", "fixed_subsample")) {
      # Create a fixed grid for the parameters. This may have the advantage of
      # spanning a number of suitable configurations.

      # Get initial values for each parameter.
      value_list <- lapply(
        parameter_list,
        function(list_entry) (list_entry$init_config)
      )

      # Generate a table with all permutations of the grid points.
      parameter_table <- expand.grid(
        value_list,
        stringsAsFactors = FALSE,
        KEEP.OUT.ATTRS = FALSE
      )
      parameter_table <- data.table::as.data.table(parameter_table)

      # Allow only unique parameter sets.
      parameter_table <- unique(parameter_table, by = names(parameter_list))
    }

    if (grid_initialisation_method %in% "fixed_subsample") {
      # Subsample the fixed grid.

      # Randomly select up to n_random_sets initial parameter sets from the
      # fixed grid.
      selected_row_id <- sample(
        x = seq_len(nrow(parameter_table)),
        size = min(c(nrow(parameter_table), n_random_sets)),
        replace = FALSE
      )

      parameter_table <- parameter_table[selected_row_id, ]
    }
  }

  # Add parameter id.
  parameter_table[, "param_id" := .I]

  return(parameter_table)
}



..randomise_hyperparameter_set <- function(
    parameter_table = NULL,
    parameter_list,
    local = TRUE
) {
  
  if (local && is_empty(parameter_table)) {
    ..error_reached_unreachable_code(
      "..randomise_hyperparameter_set: no values for local search."
    )
  }

  if (local) {
    # Parameters are locally randomised.

    # Create the updated list from the parameter table.
    updated_list <- as.list(parameter_table)

    # Remove param_id and expected improvement from updated list.
    updated_list$param_id <- NULL
    updated_list$expected_improvement <- NULL

    # Select one hyperparameter from the list of randomisable hyperparameters to
    # randomly update.
    random_parameters <- sapply(
      parameter_list, 
      function(list_entry) (list_entry$randomise)
    )
    selected_parameter <- fam_sample(
      names(parameter_list)[random_parameters],
      size = 1L
    )

    # Get current value of the selected hyperparameter from the table.
    current_parameter_value <- parameter_table[[selected_parameter]]

    # Get type and range of current hyperparameter
    parameter_type <- parameter_list[[selected_parameter]]$type
    parameter_range <- parameter_list[[selected_parameter]]$range

    if (parameter_type %in% c("integer", "numeric")) {
      if (length(parameter_range) == 2L) {
        # Check that the range is not 0.
        if (max(parameter_range) != min(parameter_range)) {
          # Convert to float in [0,1] range
          hp_val_float <- (current_parameter_value - min(parameter_range)) /
            (max(parameter_range) - min(parameter_range))

          # Draw a new random position in [0,1]
          rand_valid <- FALSE
          while (!rand_valid) {
            # Draw 20 random numbers from a normal distribution with
            # mean = hp_val_float and sd = 0.2
            hp_rand <- stats::rnorm(20L, mean = hp_val_float, sd = 0.1)
            rand_valid <- any(hp_rand >= 0.0 & hp_rand <= 1.0)
          }

          # Select new value in [0,1] and convert to original range
          new_hp_val_float <- hp_rand[hp_rand >= 0.0 & hp_rand <= 1.0][1L]
          new_hp_val_float <- new_hp_val_float * 
            (max(parameter_range) - min(parameter_range)) + min(parameter_range)
          
        } else {
          # Range is 0. Just copy the current parameter value.
          new_hp_val_float <- current_parameter_value
        }
        
      } else {
        # Treat a range such as c(0,1,3) as if only these values can be selected.
        new_hp_val_float <- fam_sample(parameter_range, size = 1L)
      }

      # Set new value as integer or numeric float
      if (parameter_type == "integer") {
        updated_list[[selected_parameter]] <- as.integer(round(new_hp_val_float))
        
      } else {
        updated_list[[selected_parameter]] <- new_hp_val_float
      }
      
    } else if (parameter_type %in% c("factor")) {
      # Find range of available options
      available_parameter_values <- parameter_range[parameter_range != current_parameter_value]

      # Randomly select one option
      updated_list[[selected_parameter]] <- factor(
        fam_sample(available_parameter_values, size = 1L),
        levels = parameter_range
      )
      
    } else if (parameter_type %in% c("logical")) {
      # Find range of available options
      available_parameter_values <- parameter_range[parameter_range != current_parameter_value]

      # Randomly select one option
      updated_list[[selected_parameter]] <- fam_sample(available_parameter_values, size = 1L)
      
    } else {
      ..error_reached_unreachable_code("randomise_hyperparameter_set_unknown_type")
    }
    
  } else {
    # Parameters are globally randomised.

    # Generate an updated list from parameter_list for randomisation
    updated_list <- lapply(
      parameter_list,
      function(list_entry) (list_entry$init_config[1L])
    )

    # Select hyperparameters to be randomised
    random_parameters <- sapply(
      parameter_list,
      function(list_entry) (list_entry$randomise)
    )
    random_parameters <- names(parameter_list)[random_parameters]

    # Iterate over hyperparameters
    for (selected_parameter in random_parameters) {
      # Get type and range of current hyperparameter
      parameter_type <- parameter_list[[selected_parameter]]$type
      parameter_range <- parameter_list[[selected_parameter]]$range
      parameter_distribution <- parameter_list[[selected_parameter]]$rand_distr

      if (parameter_type %in% c("integer", "numeric")) {
        if (length(parameter_range) == 2L) {
          if (is.null(parameter_distribution)) {
            # Select new value in [0,1] and convert to original range
            hp_rand <- stats::runif(1L)
            new_hp_val_float <- hp_rand * 
              (max(parameter_range) - min(parameter_range)) + min(parameter_range)
            
          } else if (parameter_distribution == "log") {
            # Select new value in [0,1] and convert to log transform of original
            # range
            hp_log_rand <- stats::runif(1L) *
              (log(max(parameter_range)) - log(min(parameter_range))) + log(min(parameter_range))

            # Transform back from log transformation: note this emphasises
            # smaller values
            new_hp_val_float <- exp(hp_log_rand)
          }
        } else {
          # Treat a range such as c(0,1,3) as if only these values can be
          # selected.
          new_hp_val_float <- fam_sample(parameter_range, size = 1L)
        }

        # Set new value as integer or numeric float
        if (parameter_type == "integer") {
          updated_list[[selected_parameter]] <- as.integer(round(new_hp_val_float))
        } else {
          updated_list[[selected_parameter]] <- new_hp_val_float
        }
        
      } else if (parameter_type %in% c("factor")) {
        # Randomly select one option
        updated_list[[selected_parameter]] <- factor(
          fam_sample(parameter_range, size = 1L),
          levels = parameter_range
        )
        
      } else if (parameter_type %in% c("logical")) {
        updated_list[[selected_parameter]] <- fam_sample(parameter_range, size = 1L)
      }
    }
  }

  # Return a randomised configuration as a data.table
  return(data.table::as.data.table(updated_list))
}



..create_hyperparameter_run_table <- function(
    run_ids,
    parameter_ids = NULL,
    measure_time = FALSE,
    score_table = NULL,
    parameter_table = NULL,
    optimisation_model = NULL,
    n_max_bootstraps = NULL
) {
  # Creates a run table from the run identifiers and parameter identifiers.

  # Suppress NOTES due to non-standard evaluation in data.table
  time_taken <- NULL

  if (measure_time) {
    # Filter by time. This filters out the hyperparameter sets that took very
    # long to complete, while not delivering good enough performance.

    # Compute optimisation score based on the presented scores.
    optimisation_score_table <- .compute_hyperparameter_optimisation_score(
      score_table = score_table,
      optimisation_function = optimisation_model@optimisation_function
    )

    # Find data related to the most promising set of hyperparameters.
    incumbent_set <- get_best_hyperparameter_set(
      score_table = optimisation_score_table,
      parameter_table = parameter_table,
      optimisation_model = optimisation_model,
      n_max_bootstraps = n_max_bootstraps
    )

    # Select fitting hyperparameters.
    parameter_ids <- optimisation_score_table[time_taken <= incumbent_set$max_time]$param_id
  }

  if (is.null(parameter_ids)) {
    ..error_reached_unreachable_code(
      "..create_hyperparameter_run_table: parameter_ids cannot be NULL."
    )
  }

  # Create a run table consisting of parameter and run identifiers.
  run_table <- data.table::as.data.table(expand.grid(
    param_id = parameter_ids,
    run_id = run_ids,
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  ))

  return(run_table)
}



.optimisation_process_time_available <- function(
    process_clock,
    time_limit = NULL,
    message_indent = 0L,
    verbose = FALSE
) {
  
  # Check that there is time-limit to obey.
  if (is.null(time_limit)) return(TRUE)

  # Compute time spent optimising.
  optimisation_time <- process_clock$time(units = "mins")

  # Check if there still is time left.
  if (optimisation_time < time_limit) return(TRUE)

  logger_message(
    paste0(
      "Hyperparameter optimisation: Optimisation stopped because the ",
      "optimisation process exceeded the allotted time: ",
      "time spent: ", optimisation_time, " mins; ",
      "time allotted: ", time_limit, " mins."
    ),
    indent = message_indent,
    verbose = verbose
  )

  return(FALSE)
}



.compute_hyperparameter_variable_importance <- function(
    cl = NULL,
    object,
    data,
    bootstraps,
    verbose,
    message_indent,
    ...
) {
  
  if (
    object@vimp_method %in% .get_available_random_vimp_methods() ||
    object@vimp_method %in% .get_available_none_vimp_methods() ||
    object@vimp_method %in% .get_available_signature_only_vimp_methods() ||
    object@vimp_method %in% .get_available_no_features_vimp_methods()
  ) {
    return(NULL)
    
  } else if (is(object@vimp_table, "vimpTable") || rlang::is_bare_list(object@vimp_table)) {
    # Use existing vimp_tables.
    vimp_table <- lapply(
      bootstraps,
      function(x, vimp_table) (vimp_table),
      vimp_table = object@vimp_table
    ) 
    
  } else {
    
    logger_message(
      paste0(
        "Computing variable importance for ",
        length(bootstraps), " bootstraps."
      ),
      indent = message_indent + 1L,
      verbose = verbose
    )
    
    # Spawn task to obtain variable importance tables.
    vimp_task <- methods::new(
      "familiarTaskVimp",
      project_id = object@project_id,
      vimp_method = object@vimp_method,
      file = NA_character_
    )
    
    vimp_table <- fam_lapply(
      cl = cl,
      assign = "all",
      X = bootstraps,
      FUN = ..compute_hyperparameter_variable_importance,
      vimp_task = vimp_task,
      vimp_aggregation_method = object@vimp_aggregation_method,
      vimp_rank_threshold = object@vimp_rank_threshold,
      data = data,
      feature_info = object@feature_info,
      progress_bar = verbose,
      chopchop = TRUE
    )
  }

  return(vimp_table)
}



..compute_hyperparameter_variable_importance <- function(
    train_samples, 
    vimp_task, 
    data,
    vimp_aggregation_method,
    vimp_rank_threshold,
    feature_info
) {
  # Select bootstrap data.
  data <- select_data_from_samples(
    data = data,
    samples = train_samples
  )

  # Execute the task to compute variable importance table for the current
  # bootstrap.
  vimp_table <- .perform_task(
    object = vimp_task,
    feature_info_list = feature_info,
    vimp_aggregation_method = vimp_aggregation_method,
    vimp_rank_threshold = vimp_rank_threshold,
    data = data
  )
  
  return(vimp_table)
}



.compute_hyperparameter_model_performance <- function(
    cl = NULL,
    object,
    run_table,
    parameter_table,
    bootstraps,
    data,
    vimp_table_list,
    metric_objects,
    iteration_id,
    time_optimisation_model = NULL,
    overhead_time = NULL,
    verbose = FALSE
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  param_id <- NULL

  # Check that the run table is not empty. This may occur if, for instance, the
  # initial search turned up nothing.
  if (is_empty(run_table)) return(NULL)
  
  # Prepare training and validation samples for the separate runs.
  training_list <- lapply(
    run_table$run_id,
    function(ii, training_list) (training_list[[ii]]),
    training_list = bootstraps$train_list
  )

  validation_list <- lapply(
    run_table$run_id,
    function(ii, validation_list) (validation_list[[ii]]),
    validation_list = bootstraps$valid_list
  )

  # Generate parameter sets.
  parameter_list <- lapply(
    run_table$param_id,
    function(ii, parameter_table) (parameter_table[param_id == ii, ]),
    parameter_table = parameter_table
  )

  # Replicate variable importance table objects.
  if (is(vimp_table_list, "vimpTable")) vimp_table_list <- list(vimp_table_list)
  
  vimp_table_list <- lapply(
    run_table$run_id,
    function(ii, vimp_table_list) (vimp_table_list[[ii]]),
    vimp_table_list = vimp_table_list
  )
  
  if (is.null(time_optimisation_model)) {
    # Create a scoring table, with accompanying information.
    score_results <- fam_mapply_lb(
      cl = cl,
      assign = NULL,
      FUN = ..compute_hyperparameter_model_performance,
      run_id = run_table$run_id,
      training_samples = training_list,
      validation_samples = validation_list,
      vimp_table = vimp_table_list,
      parameter_table = parameter_list,
      progress_bar = verbose,
      MEASURE.TIME = TRUE,
      MoreArgs = list(
        "object" = object,
        "data" = data,
        "metric_objects" = metric_objects
      )
    )
    
  } else {
    # Predict expected process time for each parameter set.
    process_time <- sapply(parameter_list,
      .compute_expected_train_time,
      time_model = time_optimisation_model
    )

    # Create a scoring table, with accompanying information.
    score_results <- fam_mapply(
      cl = cl,
      assign = NULL,
      FUN = ..compute_hyperparameter_model_performance,
      run_id = run_table$run_id,
      training_samples = training_list,
      validation_samples = validation_list,
      vimp_table = vimp_table_list,
      parameter_table = parameter_list,
      progress_bar = verbose,
      process_time = process_time,
      overhead_time = overhead_time,
      chopchop = TRUE,
      MEASURE.TIME = TRUE,
      MoreArgs = list(
        "object" = object,
        "data" = data,
        "metric_objects" = metric_objects
      )
    )
  }

  # Aggregate the results, and the score table
  aggregated_results <- list("results" = data.table::rbindlist(
    lapply(score_results$results, function(x) (x$score_table)),
    use.names = TRUE
  ))

  # Add in process times to the results.
  aggregated_results$results[, "time_taken" := rep(
    score_results$process_time, each = 2L * length(metric_objects)
  )]

  # Add in iteration id.
  aggregated_results$results[, "iteration_id" := iteration_id]

  # Aggregate errors.
  aggregated_results$error <- unlist(lapply(
    score_results$results, 
    function(x) (x$error)
  ))

  # Return aggregated results.
  return(aggregated_results)
}



..compute_hyperparameter_model_performance <- function(
    object,
    data,
    run_id,
    training_samples,
    validation_samples,
    parameter_table,
    vimp_table,
    metric_objects,
    signature_features = NULL
) {
  
  if (!is(object, "familiarModel")) {
    ..error_reached_unreachable_code(paste0(
      "..compute_hyperparameter_model_performance: object is ",
      "not a familiarModel object."
    ))
  }

  # Find parameter id and run id for the current run
  parameter_list <- as.list(parameter_table[, -c("param_id")])
  param_id <- parameter_table$param_id[1L]

  # Select training (in-bag) and validation (out-of-bag) data for current run,
  data_training <- select_data_from_samples(
    data = data,
    samples = training_samples
  )

  data_validation <- select_data_from_samples(
    data = data,
    samples = validation_samples
  )
  
  # Update the familiar model (for variable importance)
  object@hyperparameters <- parameter_list

  # Attach variable importance table to object.
  object@vimp_table <- vimp_table
  
  # Set signature.
  object <- set_model_features(
    object = object,
    minimise_footprint = TRUE
  )

  # Train model with the set of hyperparameters.
  object <- .train(
    object = object,
    data = data_training,
    get_additional_info = FALSE,
    trim_model = FALSE,
    approximate = TRUE
  )

  # Generate scores.
  score_table <- mapply(
    function(
      data, 
      data_set, 
      object, 
      metric_objects, 
      settings
    ) {
      # Get metric names.
      metric_names <- sapply(metric_objects, function(metric_object) metric_object@metric)

      # Set time_max.
      if (!is.null(object@settings$time_max)) {
        # Derive from evaluation settings attached to the familiarModel object.
        time_max <- object@settings$time_max
      } else {
        time_max <- object@outcome_info@time
      }

      # Predict for the in-bag and out-of-bag datasets.
      prediction_table <- .predict(
        object = object,
        data = data,
        time = time_max
      )

      # Compute metric scores.
      metrics_values <- sapply(
        metric_objects,
        compute_metric_score,
        data = prediction_table
      )

      # Compute objective scores.
      metrics_objective_score <- mapply(
        compute_objective_score,
        metric = metric_objects,
        value = metrics_values,
        SIMPLIFY = TRUE
      )

      # Return as data.table.
      return(data.table::data.table(
        "metric" = metric_names,
        "data_set" = data_set,
        "value" = metrics_values,
        "objective_score" = metrics_objective_score
      ))
    },
    data = list(data_training, data_validation),
    data_set = c("training", "validation"),
    MoreArgs = list(
      "object" = object,
      "metric_objects" = metric_objects
    ),
    SIMPLIFY = FALSE
  )

  # Aggregate to a single table.
  score_table <- data.table::rbindlist(score_table, use.names = TRUE)

  # Add parameter id and run id.
  score_table[, ":="(
    "param_id" = param_id,
    "run_id" = run_id
  )]
  
  # Set the column order.
  data.table::setcolorder(
    x = score_table,
    neworder = c("param_id", "run_id")
  )

  return(list(
    "score_table" = score_table,
    "error" = object@messages$error
  ))
}



.compute_hyperparameter_optimisation_score <- function(
    score_table,
    optimisation_function
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  .NATURAL <- NULL

  if (is_empty(score_table)) return(NULL)

  # Compute optimisation score.
  optimisation_score_table <- .compute_metric_optimisation_score(
    score_table = score_table,
    optimisation_function = optimisation_function
  )

  # Reinsert time_taken.
  optimisation_score_table <- optimisation_score_table[
    unique(score_table[, c("param_id", "run_id", "time_taken")]),
    on = .NATURAL
  ]

  return(optimisation_score_table)
}



.compute_hyperparameter_summary_score <- function(
    score_table,
    parameter_table,
    optimisation_model
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  optimisation_score <- time_taken <- .NATURAL <- mu <- sigma <- NULL
  summary_score <- score_estimate <- NULL

  # Check that the table is not empty.
  if (is_empty(score_table)) return(NULL)

  # Check if summary score has already been created.
  if ("summary_score" %in% colnames(score_table)) return(score_table)

  # Extract the optimisation function.
  optimisation_function <- optimisation_model@optimisation_function

  # Check if the score table already contains optimisation scores.
  if (!"optimisation_score" %in% colnames(score_table)) {
    score_table <- .compute_hyperparameter_optimisation_score(
      score_table = score_table,
      optimisation_function = optimisation_function
    )
  }

  # Make sure to return both the summary score, as well as the mean optimisation
  # score (or its model estimate if the optimisation function allows it). The
  # mean score is important for acquisition functions, where we work with
  # predictions of the optimisation score -- not the summary score. In addition,
  # compute time_taken.
  if (optimisation_function %in% c(
    "validation", "max_validation", "balanced", "stronger_balance"
  )) {
    # Here we just use the mean score.
    data <- score_table[, list(
      "summary_score" = mean(optimisation_score, na.rm = TRUE),
      "score_estimate" = mean(optimisation_score, na.rm = TRUE),
      "time_taken" = stats::median(time_taken)
    ), by = "param_id"]
    
  } else if (optimisation_function %in% c("validation_minus_sd")) {
    # Here we need to compute the standard deviation, with an exception for
    # single subsamples.
    data <- score_table[, list(
      "summary_score" = ..optimisation_function_validation_minus_sd(optimisation_score),
      "score_estimate" = mean(optimisation_score, na.rm = TRUE),
      "time_taken" = stats::median(time_taken)
    ), by = "param_id"]
    
  } else if (optimisation_function %in% c("validation_25th_percentile")) {
    # Summary score is formed by the 25th percentile of the optimisation scores.
    data <- score_table[, list(
      "summary_score" = stats::quantile(optimisation_score, probs = 0.25, names = FALSE, na.rm = TRUE),
      "score_estimate" = mean(optimisation_score, na.rm = TRUE),
      "time_taken" = stats::median(time_taken)
    ), by = "param_id"]
    
  } else if (optimisation_function %in% c(
    "model_estimate", "model_estimate_minus_sd",
    "model_balanced_estimate", "model_balanced_estimate_minus_sd"
  )) {
    # Model based summary scores. The main difference with other optimisation
    # functions is that a hyperparameter model is used to both infer the summary
    # scores and the score estimate.

    # Check if the model is trained.
    if (!model_is_trained(optimisation_model)) {
      optimisation_model <- .train(
        object = optimisation_model,
        data = score_table,
        parameter_data = parameter_table
      )
    }

    # Get parameter data that is used in the score table.
    parameter_data <- parameter_table[
      unique(score_table[, mget("param_id")]),
      on = .NATURAL
    ]

    # Model was successfully trained.
    prediction_table <- .predict(
      object = optimisation_model,
      data = parameter_data,
      type = ifelse(
        optimisation_function %in% c("model_estimate", "model_balanced_estimate"),
        "default", "sd"
      )
    )

    # Check the prediction table has returned sensible information.
    if (any(is.finite(prediction_table$mu))) {
      # Prepare the score estimate and time taken.
      data <- score_table[
        , list("time_taken" = stats::median(time_taken)),
        by = "param_id"
      ]

      # Fold the prediction table into data.
      data <- merge(
        x = data,
        y = prediction_table,
        all.x = TRUE,
        all.y = FALSE,
        by = "param_id"
      )

      if (optimisation_function %in% c("model_estimate", "model_balanced_estimate")) {
        # Copy mu as score estimate.
        data[, "score_estimate" := mu]

        # Rename mu to summary_score.
        data.table::setnames(
          x = data,
          old = "mu",
          new = "summary_score"
        )
        
      } else if (optimisation_function %in% c(
        "model_estimate_minus_sd", "model_balanced_estimate_minus_sd"
      )) {
        # Compute summary score.
        data[, "summary_score" := mu - sigma]

        # Rename mu to score_estimate.
        data.table::setnames(
          x = data,
          old = "mu",
          new = "score_estimate"
        )
        
      } else {
        ..error_reached_unreachable_code(paste0(
          ".compute_hyperparameter_summary_score: encountered ",
          "unknown optimisation function: ", optimisation_function
        ))
      }

      # Check that any predictions make sense at all.
      if (!any(is.finite(score_table$optimisation_score))) {
        data[, ":="(
          summary_score = NA_real_,
          score_estimate = NA_real_
        )]
      }
      
    } else if (optimisation_function %in% c(
      "model_estimate", "model_balanced_estimate"
    )) {
      # In absence of suitable data, use the model-less equivalent of
      # model_estimate or model_balanced_estimate, namely "validation".
      data <- score_table[, list(
        "summary_score" = mean(optimisation_score, na.rm = TRUE),
        "score_estimate" = mean(optimisation_score, na.rm = TRUE),
        "time_taken" = stats::median(time_taken)
      ), by = "param_id"]
      
    } else if (optimisation_function %in% c(
      "model_estimate_minus_sd", "model_balanced_estimate_minus_sd"
    )) {
      # In absence of suitable data, use the model-less equivalent of
      # model_estimate_minus_sd or model_balanced_estimate_minus_sd,
      # namely "validation_minus_sd".
      data <- score_table[, list(
        "summary_score" = ..optimisation_function_validation_minus_sd(optimisation_score),
        "score_estimate" = mean(optimisation_score, na.rm = TRUE),
        "time_taken" = stats::median(time_taken)
      ), by = "param_id"]
      
    } else {
      ..error_reached_unreachable_code(paste0(
        ".compute_hyperparameter_summary_score: encountered unknown ",
        "optimisation function: ", optimisation_function
      ))
    }
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".compute_hyperparameter_summary_score: encountered an unknown ",
      "optimisation function"
    ))
  }

  # Format data by selecting only relevant columns. This also orders the data.
  data <- data[, mget(c("param_id", "summary_score", "score_estimate", "time_taken"))]

  # Update missing scores.
  data[!is.finite(summary_score), "summary_score" := ..get_replacement_optimisation_score()]
  data[!is.finite(score_estimate), "score_estimate" := ..get_replacement_optimisation_score()]

  return(data)
}



..optimisation_function_validation_minus_sd <- function(x) {
  # Here we need to switch based on the length of x.

  # The default behaviour is when multiple subsamples exist, and the standard
  # deviation can be computed.
  if (length(x) > 1L) {
    return(mean(x, na.rm = TRUE) - stats::sd(x, na.rm = TRUE))
  }

  # The exception for x with length 1, where the standard deviation does not
  # exist.
  return(mean(x, na.rm = TRUE))
}



get_best_hyperparameter_set <- function(
    score_table,
    parameter_table,
    optimisation_model,
    n_max_bootstraps,
    n = 1L,
    parameter_id_set = NULL
) {
  # The best set has several interesting values that can be computed and used
  # for acquisition functions, information for the user etc.

  # Suppress NOTES due to non-standard evaluation in data.table
  summary_score <- param_id <- NULL

  # Compute summary scores.
  data <- .compute_hyperparameter_summary_score(
    score_table = score_table,
    parameter_table = parameter_table,
    optimisation_model = optimisation_model
  )

  # Select only the parameter sets of interest. This happens during the run-off
  # between the incumbent and challenger hyperparameter sets.
  if (!is.null(parameter_id_set)) {
    data <- data[param_id %in% parameter_id_set]
  }

  # Find the best hyperparameter set based on summary scores in data.
  data <- data[order(-summary_score)]
  data <- head(data, n = n)

  # Compute round t, i.e. the highest number of bootstraps observed across the
  # parameters.
  t <- max(unique(score_table[, mget(c("param_id", "run_id"))])[, list("n" = .N), by = "param_id"]$n)

  # Compute the maximum allowed time.
  max_time <- (5.0 - 4.0 * t / n_max_bootstraps) * data$time_taken[1L]

  # Check that the maximum time is not trivially small (i.e. less than 10
  # seconds).
  if (is.na(max_time)) {
    max_time <- Inf
  } else if (max_time < 10.0) {
    max_time <- 10.0
  }

  # Extract and update summary score and score estimate
  summary_score <- data$summary_score
  if (!is.finite(summary_score)) {
    summary_score <- ..get_replacement_optimisation_score()
  }

  score_estimate <- data$score_estimate
  if (!is.finite(score_estimate)) {
    score_estimate <- ..get_replacement_optimisation_score()
  }

  if (!"optimisation_score" %in% colnames(score_table)) {
    # Compute mean validation score as a summary score, as well as mean validation
    # scores of the raw metric values.
    additional_scores <- .compute_hyperparameter_additional_scores(
      score_table = score_table[param_id %in% data$param_id]
    )

    # Merge additional scores with summary data and sort again to prevent any
    # merge issues.
    data <- merge(
      x = data,
      y = additional_scores,
      by = "param_id"
    )
    data <- data[order(-summary_score)]

    # Find metric columns
    metric_columns <- unique(score_table$metric)

    # Extract the summary score, score estimate, and time taken, and return with
    # other information.
    return(list(
      "param_id" = data$param_id,
      "t" = t,
      "time" = data$time_taken,
      "max_time" = max_time,
      "summary_score" = summary_score,
      "score_estimate" = score_estimate,
      "validation_score" = data$validation_score,
      "metric_score" = data[, mget(c(metric_columns))],
      "n" = n_max_bootstraps
    ))
    
  } else {
    return(list(
      "param_id" = data$param_id,
      "t" = t,
      "time" = data$time_taken,
      "max_time" = max_time,
      "summary_score" = summary_score,
      "score_estimate" = score_estimate,
      "n" = n_max_bootstraps
    ))
  }
}



.compute_hyperparameter_additional_scores <- function(score_table) {
  # Suppress NOTES due to non-standard evaluation in data.table
  data_set <- optimisation_score <- value <- NULL

  # Compute mean validation score, which is the typical objective score for
  # hyperparameter optimisation.
  mean_validation_scores <- .compute_hyperparameter_optimisation_score(
    score_table = score_table,
    optimisation_function = "validation"
  )
  
  mean_validation_scores <- mean_validation_scores[
    ,
    list("validation_score" = mean(optimisation_score, na.rm = TRUE)),
    by = c("param_id")
  ]

  # Compute mean metric values.
  mean_metric_scores <- score_table[
    data_set == "validation",
    list("average_metric_value" = mean(value, na.rm = TRUE)),
    by = c("param_id", "metric")
  ]
  mean_metric_scores <- data.table::dcast(
    mean_metric_scores,
    param_id ~ metric,
    value.var = "average_metric_value"
  )

  return(merge(
    x = mean_validation_scores,
    y = mean_metric_scores,
    by = "param_id"
  ))
}



..parse_optimisation_summary_to_string <- function(
    parameter_set,
    parameter_table,
    parameter_list
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  param_id <- NULL

  # Set up string.
  message_str <- character(0L)

  # Add summary score.
  message_str <- c(
    message_str,
    paste0(
      "summary score: ",
      sprintf("%.4f", parameter_set$summary_score[1L])
    )
  )

  # Add validation optimisation score.
  message_str <- c(
    message_str,
    paste0(
      "validation optimisation score: ",
      sprintf("%.4f", parameter_set$validation_score[1L])
    )
  )


  # Add metric scores.
  for (metric in colnames(parameter_set$metric_score)) {
    message_str <- c(
      message_str,
      paste0(
        metric, ": ",
        sprintf("%.4f", parameter_set$metric_score[[metric]][1L])
      )
    )
  }

  # Add hyperparameters. Iterate through parameter list and identify parameters
  # that are being optimised.
  for (current_parameter in names(parameter_list)) {
    # Check if the current parameter is randomised.
    if (parameter_list[[current_parameter]]$randomise == TRUE) {
      # Determine the value of the parameter for the set identified by the id
      # variable.
      optimal_value <- parameter_table[param_id == parameter_set$param_id, ][[current_parameter]][1L]

      # Append to string.
      message_str <- c(
        message_str,
        paste0(current_parameter, ": ", optimal_value)
      )
    }
  }

  return(paste0(message_str, collapse = "; "))
}



..update_hyperparameter_optimisation_stopping_criteria <- function(
    set_data,
    stop_data = NULL,
    tolerance = 1E-2
) {
  
  # Store the validation score of the incumbent set. Note that the latest score
  # is always appended.
  validation_scores <- c(
    stop_data$score,
    set_data$validation_score
  )

  # Store the parameter id of the incumbent set. Note that the most recent
  # hyperparameter set identifier is always appended.
  incumbent_parameter_id <- c(
    stop_data$parameter_id,
    set_data$param_id
  )

  # Read the convergence counters. Can be NULL initially.
  convergence_counter_score <- stop_data$convergence_counter_score
  if (is.null(convergence_counter_score)) convergence_counter_score <- 0L

  # Convergence counter for parameter identifiers.
  convergence_counter_parameter_id <- stop_data$convergence_counter_parameter_id
  if (is.null(convergence_counter_parameter_id)) convergence_counter_parameter_id <- 0L

  # Update convergence counter for parameter identifiers if there is no change
  # in the incumbent parameter set. Skip on the first run-through.
  if (length(incumbent_parameter_id) >= 2L) {
    if (all(tail(incumbent_parameter_id, n = 3L) == set_data$param_id)) {
      # Update the convergence counter.
      convergence_counter_parameter_id <- convergence_counter_parameter_id + 1L
    } else {
      # Reset the convergence counter.
      convergence_counter_parameter_id <- 0L
    }
  }

  # Assess convergence of the summary scores. This is skipped on the first
  # run-through.
  if (length(validation_scores) >= 2L) {
    # Compute absolute deviation from the mean.
    max_abs_deviation <- max(abs(
      tail(validation_scores, n = 3L) - mean(tail(validation_scores, n = 3L))
    ))

    if (is.na(max_abs_deviation)) {
      convergence_counter_score <- 0L
    } else if (max_abs_deviation <= tolerance) {
      convergence_counter_score <- convergence_counter_score + 1L
    } else {
      convergence_counter_score <- 0L
    }
  }

  # Read the no-naive-improvement counter. Can be NULL initially. This reduces
  # needless search for meaningless hyperparameter sets.
  no_naive_improvement_counter <- stop_data$no_naive_improvement_counter
  if (is.null(no_naive_improvement_counter)) no_naive_improvement_counter <- 0L

  if (set_data$validation_score <= 0.0) {
    # The best known model does not predict better than a naive_model.
    no_naive_improvement_counter <- no_naive_improvement_counter + 1L
  } else {
    # The best known model predicts better than a naive model.
    no_naive_improvement_counter <- 0L
  }

  # Return list with stopping parameters.
  return(list(
    "score" = validation_scores,
    "parameter_id" = incumbent_parameter_id,
    "convergence_counter_score" = convergence_counter_score,
    "convergence_counter_parameter_id" = convergence_counter_parameter_id,
    "no_naive_improvement_counter" = no_naive_improvement_counter
  ))
}



..create_hyperparameter_intensify_run_table <- function(
    parameter_id_incumbent,
    parameter_id_challenger,
    score_table,
    n_max_bootstraps,
    n_intensify_step_bootstraps,
    exploration_method
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  param_id <- sampled <- run_id <- to_sample <- n <- NULL

  # Combine all parameter ids.
  parameter_ids <- c(parameter_id_incumbent, parameter_id_challenger)

  # Make a copy of the score table and keep only relevant parameter and run
  # identifiers.
  score_table <- unique(data.table::copy(
    score_table[param_id %in% parameter_ids, c("param_id", "run_id")]
  ))

  # Add a sampled column that marks those elements that have been previously
  # sampled.
  score_table[, "sampled" := TRUE]

  # Generate run data table
  run_table <- data.table::as.data.table(expand.grid(
    param_id = parameter_ids,
    run_id = seq_len(n_max_bootstraps),
    KEEP.OUT.ATTRS = FALSE,
    stringsAsFactors = FALSE
  ))

  # Mark those runs that have not been sampled yet. These have an NA-value for
  # the sampled column.
  run_table <- merge(
    x = run_table,
    y = score_table,
    by = c("param_id", "run_id"),
    all.x = TRUE
  )
  run_table[is.na(sampled), "sampled" := FALSE]

  # Add a to_sample column to mark new samples to be made.
  run_table[, "to_sample" := FALSE]

  # Find runs that have only been fully, partially or not sampled by the hyperparameter sets.
  sampled_runs <- run_table[sampled == TRUE, list("n" = .N), by = "run_id"]

  # Identify run identifiers.
  fully_sampled_runs <- sampled_runs[n == length(parameter_ids)]$run_id
  partially_sampled_runs <- sampled_runs[n > 0L & n < length(parameter_ids)]$run_id
  unsampled_runs <- setdiff(seq_len(n_max_bootstraps), sampled_runs[n > 0L]$run_id)

  # Check that there are any runs to be sampled.
  if (length(fully_sampled_runs) == n_max_bootstraps) return(NULL)

  # Compute the number of runs that could be completely new. This is 1/3rd of
  # n_max_bootstraps, with a minimum of 1. This makes the selection of new runs
  # more conservative, saving time and resources. When the exploration method
  # equals none, we select up to n_intensify_step_bootstraps of new runs.
  if (exploration_method == "none") {
    n_new_runs <- n_intensify_step_bootstraps
  } else if (exploration_method == "single_shot") {
    n_new_runs <- 1L
  } else {
    n_new_runs <- max(c(1L, as.integer(floor(n_intensify_step_bootstraps / 3.0))))
  }

  # An exception should be made if there are no partially sampled runs. Then up
  # to n_intensify_step_bootstraps should be sampled.
  if (length(partially_sampled_runs) == 0L) n_new_runs <- n_intensify_step_bootstraps

  # Iterate over the parameter identifiers and identify what runs should be
  # sampled.
  for (parameter_id in parameter_ids) {
    # Find partially sampled runs that have not been sampled for the current
    # hyperparameter set.
    current_partially_sampled_runs <- setdiff(
      partially_sampled_runs,
      run_table[sampled == TRUE & param_id == parameter_id, ]$run_id
    )

    # Select runs to be sampled from among those that have been sampled by other
    # hyperparameter sets.
    if (length(current_partially_sampled_runs) > 0L) {
      run_table[
        run_id %in% head(current_partially_sampled_runs, n = n_intensify_step_bootstraps) &
          param_id == parameter_id,
        "to_sample" := TRUE
      ]
    }

    # Add completely new runs if there is room.
    if (
      length(current_partially_sampled_runs) < n_intensify_step_bootstraps &&
      length(unsampled_runs) > 0L
    ) {
      # Determine the number of new runs that should be added.
      n_current_new_runs <- min(c(
        n_new_runs,
        n_intensify_step_bootstraps - length(current_partially_sampled_runs)
      ))

      # Select up to current_new_runs to sample.
      run_table[
        run_id %in% head(unsampled_runs, n = n_current_new_runs) &
          param_id == parameter_id,
        "to_sample" := TRUE
      ]
    }
  }

  # Select only runs that should be sampled, and return run_id and param_id
  # columns.
  run_table <- run_table[to_sample == TRUE, c("param_id", "run_id")]

  return(run_table)
}



.compare_hyperparameter_sets <- function(
    score_table,
    parameter_table,
    optimisation_model,
    parameter_id_incumbent,
    parameter_id_challenger,
    exploration_method = "successive_halving",
    intensify_stop_p_value
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  param_id <- p_value <- summary_score <- NULL

  if (exploration_method == "successive_halving") {
    # Compute summary scores.
    data <- .compute_hyperparameter_summary_score(
      score_table = score_table,
      parameter_table = parameter_table,
      optimisation_model = optimisation_model
    )

    # Select only data concerning the incumbent and challenger sets
    data <- data[param_id %in% c(parameter_id_incumbent, parameter_id_challenger)]

    # Order the relevant data according to summary score.
    data <- data[order(-summary_score)]

    # Remove the worst half of performers. For simplicity we increase the number
    # of selectable hyperparameters by one to account for the challenger.
    n_selectable <- as.integer(floor((length(parameter_id_challenger)) / 2.0))

    # Select the new incumbent, which may be identical to the old incumbent.
    param_id_incumbent_new <- head(data, n = 1L)$param_id

    # Select the remaining challengers, which temporarily includes the new
    # incumbent.
    param_id_challenger_new <- head(data, n = n_selectable + 1L)$param_id
    
  } else if (exploration_method == "stochastic_reject") {
    # Compute hyperparameter optimisation scores. First check if the score table
    # already contains optimisation scores.
    if (!"optimisation_score" %in% colnames(score_table)) {
      data <- .compute_hyperparameter_optimisation_score(
        score_table = score_table[param_id %in% c(parameter_id_incumbent, parameter_id_challenger)],
        optimisation_function = optimisation_model@optimisation_function
      )
      
    } else {
      data <- data.table::copy(score_table)
    }

    # Compute summary scores.
    summary_data <- .compute_hyperparameter_summary_score(
      score_table = data,
      parameter_table = parameter_table,
      optimisation_model = optimisation_model
    )

    # Select only data concerning the incumbent and challenger sets
    summary_data <- summary_data[param_id %in% c(parameter_id_incumbent, parameter_id_challenger)]

    # Order the relevant data according to summary score.
    summary_data <- summary_data[order(-summary_score)]

    # Select the new incumbent, which may be identical to the old incumbent.
    param_id_incumbent_new <- head(summary_data, n = 1L)$param_id

    # Select the preliminary set of challengers.
    param_id_challenger_new <- setdiff(
      union(parameter_id_incumbent, parameter_id_challenger),
      param_id_incumbent_new
    )

    # Compare optimisation scores and compute p-values.
    p_value_table <- data[
      param_id %in% param_id_challenger_new,
      ..compare_hyperparameter_optimisation_scores(
        challenger_score_table = .SD,
        incumbent_score_table = data[param_id == param_id_incumbent_new]
      ),
      by = c("param_id")
    ]

    # Select parameter identifiers with p-values above the thresholds.
    param_id_challenger_new <- p_value_table[p_value > intensify_stop_p_value, ]$param_id
    
  } else if (exploration_method %in% c("none", "single_shot")) {
    # Keep the incumbent set.
    param_id_incumbent_new <- parameter_id_incumbent

    # Keep the challengers.
    param_id_challenger_new <- parameter_id_challenger
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".compare_hyperparameter_sets: encountered unknown exploration method: ",
      exploration_method
    ))
  }

  # Update param_id_challenger_new by removing the incumbent set.
  param_id_challenger_new <- setdiff(
    union(param_id_incumbent_new, param_id_challenger_new),
    param_id_incumbent_new
  )

  return(list(
    "parameter_id_incumbent" = param_id_incumbent_new,
    "parameter_id_challenger" = param_id_challenger_new
  ))
}



..compare_hyperparameter_optimisation_scores <- function(
    challenger_score_table,
    incumbent_score_table
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  run_id <- optimisation_score <- NULL

  # Find matching bootstrap ids
  run_id_match <- intersect(challenger_score_table$run_id, incumbent_score_table$run_id)

  # Select rows that match. Note that optimisation score is updated locally to
  # avoid update by reference on protected slices of the score table in
  # .compare_hyperparameter_optimisation_scores.
  challenger_score_table <- data.table::copy(challenger_score_table[run_id %in% run_id_match, ])
  challenger_score_table <- challenger_score_table[is.na(optimisation_score), "optimisation_score" := -1.0]
  incumbent_score_table <- data.table::copy(incumbent_score_table[run_id %in% run_id_match, ])
  incumbent_score_table <- incumbent_score_table[is.na(optimisation_score), "optimisation_score" := -1.0]

  # Find p-value for matching means using paired wilcoxon rank test.
  p_value <- suppressWarnings(stats::wilcox.test(
    x = challenger_score_table[order(run_id_match)]$optimisation_score,
    y = incumbent_score_table[order(run_id_match)]$optimisation_score,
    paired = TRUE,
    alternative = "less"
  )$p.value)
  
  # Return list with p-values.
  return(list("p_value" = p_value))
}



.encode_categorical_hyperparameters <- function(parameter_list) {
  # Iterate over hyperparameters.
  parameter_list <- lapply(
    parameter_list,
    ..encode_categorical_hyperparameters
  )

  return(parameter_list)
}



..encode_categorical_hyperparameters <- function(x) {
  # Encodes the init_config element of x in case x is a categorical
  # hyperparameter.

  # Skip if x is not categorical.
  if (x$type != "factor") return(x)

  if (x$randomise) {
    # Assign default range + initial value as levels, in case the hyperparameter
    # is randomised.
    x$init_config <- factor(x$init_config, levels = union(x$range, x$init_config))
  } else {
    # Assign only the selected level as factor.
    x$init_config <- factor(x$init_config, levels = x$init_config)
  }

  return(x)
}



.set_signature_size <- function(
    object,
    vimp_table_list,
    suggested_range = NULL
) {
  if (is.null(suggested_range)) suggested_range <- c(1L, Inf)

  # Some variable importance methods fail to produce a list (e.g. none,
  # signature_only, random). We create a single element list in that case.
  if (is(vimp_table_list, "vimpTable")) vimp_table_list <- list(vimp_table_list)
  if (is_empty(vimp_table_list)) vimp_table_list <- list(NULL)

  # Update minimum signature size in object.
  object@hyperparameters$sign_size <- min(suggested_range)
  
  # Determine the lower range of the signature. This includes any signature
  # features that may exist.
  signature_list <- lapply(
    vimp_table_list,
    function(vimp_table, object) {
      object@vimp_table <- vimp_table
      return(get_signature_features(object))
    },
    object = object
  )
  
  # Determine minimum signature size.
  min_signature_size <- min(lengths(signature_list))
  
  # Update maximum signature size in the list.
  object@hyperparameters$sign_size <- max(suggested_range)

  signature_list <- lapply(
    vimp_table_list,
    function(vimp_table, object) {
      object@vimp_table <- vimp_table
      return(get_signature_features(object))
    },
    object = object
  )
  
  # Determine maximum signature size.  
  max_signature_size <- max(lengths(signature_list))
  
  # Add minimum and maximum signature size, if necessary.
  if (any(suggested_range < min_signature_size)) {
    suggested_range <- c(suggested_range, min_signature_size)
  }
  
  if (any(suggested_range > max_signature_size)) {
    suggested_range <- c(suggested_range, max_signature_size)
  }

  # Limit range to unique values within the range.
  suggested_range <- suggested_range[
    suggested_range >= min_signature_size & suggested_range <= max_signature_size
  ]
  
  suggested_range <- unique(suggested_range)

  return(suggested_range)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarDataElementCalibrationInfo object ------------------------------------

setClass(
  "familiarDataElementCalibrationInfo",
  contains = "familiarDataElement",
  slots = list("outcome_type" = "ANY"),
  prototype = methods::prototype(
    estimation_type = "point",
    outcome_type = NULL
  )
)



# extract_calibration_info (generic) -------------------------------------------

#'@title Internal function to extract calibration info from data.
#'
#'@description Collects .
#'
#'@inheritParams .extract_data
#'
#'@return A list of familiarDataElements with hyperparameters.
#'@md
#'@keywords internal
setGeneric(
  "extract_calibration_info",
  function(
    object,
    detail_level = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_calibration_info")
  }
)



# extract_calibration_info (familiarEnsemble) ----------------------------------
setMethod(
  "extract_calibration_info",
  signature(object = "familiarEnsemble"),
  function(
    object,
    detail_level = waiver(),
    message_indent = 0L,
    verbose = FALSE
  ) {
    
    # Extracts calibration info for survival outcomes. Note that some routines
    # for continuous outcomes are available, but not used.
    if (!object@outcome_type %in% c("survival")) return(NULL)
    
    # Message extraction start
    logger_message(
      paste0("Extracting calibration information."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Check the level detail.
    detail_level <- .parse_detail_level(
      x = detail_level,
      object = object,
      default = "ensemble",
      data_element = "calibration_info"
    )
    
    proto_data_element <- methods::new(
      "familiarDataElementCalibrationInfo",
      detail_level = detail_level
    )
    
    # Generate elements to send to dispatch.
    calibration_info <- extract_dispatcher(
      FUN = .extract_calibration_info,
      cl = NULL,
      has_internal_bootstrap = FALSE,
      object = object,
      proto_data_element = proto_data_element,
      aggregate_results = FALSE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(calibration_info)
  }
)



# extract_calibration_info (prediction table) ----------------------------------
setMethod(
  "extract_calibration_info",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("extract calibration info")
    
    return(NULL)
  }
)



.extract_calibration_info <- function(
    object,
    proto_data_element,
    ...
) {
  
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  data_element <- add_model_name(proto_data_element, object = object)
  
  # Add data.
  data_element@data <- object@calibration_info
  
  # Add outcome type.
  data_element@outcome_type <- object@outcome_type
  
  # Check if the element is empty.
  if (is_empty(data_element@data)) return(NULL)
  
  # Set value and grouping columns.
  if (object@outcome_type == "survival") {
    data_element@value_column <- setdiff(colnames(data_element@data), "time")
    data_element@grouping_column <- "time"
    
  } else if (object@outcome_type %in% c("continuous")) {
    data_element@value_column <- colnames(data_element@data)
    
  } else {
    ..error_outcome_type_not_implemented(object@outcome_type)
  }
  
  return(data_element)
}



# ..compute_data_elements_estimates (familiarDataElementHyperparameters) -------
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElementCalibrationInfo"),
  function(x, x_list = NULL, ...) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    min_value <- max_value <- time <- NULL
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Remove empty entries.
    x <- x[!sapply(x, is_empty)]
    if (is_empty(x)) return(NULL)
    
    # Collect all data.
    data <- data.table::rbindlist(
      lapply(x, function(x) (x@data)),
      use.names = TRUE,
      fill = TRUE
    )
    
    if (x[[1L]]@outcome_type %in% c("continuous")) {
      
      # Extract min and max values of the range for continuous type outcomes.
      if (is.null(x[[1L]]@grouping_column)) {
        data <- data[
          ,
          list(
            "min_value" = min(min_value),
            "max_value" = max(max_value)
          )
        ]
        
      } else {
        data <- data[
          ,
          list(
            "min_value" = min(min_value),
            "max_value" = max(max_value)
          ),
          by = x[[1L]]@grouping_column
        ]
      }
      
    } else if (x[[1L]]@outcome_type %in% c("survival")) {
      
      # Identify all unique time points.
      new_time <- sort(unique(data$time))
      
      # Find non-time grouping columns.
      grouping_column <- setdiff(
        colnames(data), c("time", x[[1L]]@value_column)
      )
      
      # Select only unique data points to prevent warnings during approximation.
      data <- unique(data)
      
      if (length(grouping_column) > 0L) {
        # Interpolate survival columns by grouping column.
        data <- data[
          ,
          ..interpolate_survival_data(.SD, time = time, new_time = new_time),
          by = c(grouping_column),
          .SDcols = x[[1L]]@value_column
        ]
        
      } else {
        # Interpolate survival columns.
        data <- data[
          ,
          ..interpolate_survival_data(.SD, time = time, new_time = new_time),
          .SDcols = x[[1L]]@value_column
        ]
      }
      
      # Compute mean value at each time point, by group.
      data <- data[
        ,
        lapply(.SD, mean, na.rm = TRUE),
        by = c(x[[1L]]@grouping_column),
        .SDcols = c(x[[1L]]@value_column)
      ]
      
    } else {
      ..error_outcome_type_not_implemented(outcome_type = x[[1L]]@outcome_type)
    }
    
    # Copy data element.
    y <- x[[1L]]
    y@data <- data
    
    return(y)
  }
)



..interpolate_survival_data <- function(
    data,
    time,
    new_time,
    extrapolate = FALSE
) {
  
  # Get columns names.
  value_columns <- colnames(data)
  
  # Interpolate survival data.
  data <- data[
    ,
    lapply(
      .SD,
      function(y, x, x_out, extrapolate) {
        return(stats::approx(
          x = x,
          y = y,
          xout = x_out,
          rule = ifelse(extrapolate, 2L, 1L),
          method = "linear"
        )$y)
      },
      x = time,
      x_out = new_time,
      extrapolate = extrapolate
    )
  ]
  
  # Insert time column.
  data[, "time" := new_time]
  
  # Select only finite values. This removes extrapolated values in case
  # extrapolate equals FALSE.
  data <- data[!is.na(value_columns[1L])]
  
  return(as.list(data))
}



# export_calibration_info (generic) --------------------------------------------

#'@title Extract and export calibration information.
#'
#'@description Extract and export calibration information (e.g. baseline
#'  survival) for data in a familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Currently only baseline survival is exported as supporting calibration
#'  information. See `export_calibration_data` for export of direct assessment
#'  of calibration, including calibration and goodness-of-fit tests.
#'
#'@return A data.table (if `dir_path` is not provided), or nothing, as all data
#'  is exported to `csv` files.
#'@exportMethod export_calibration_info
#'@md
#'@rdname export_calibration_info-methods
setGeneric(
  "export_calibration_info",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_calibration_info")
  }
)

# export_calibration_info (collection) -----------------------------------------

#'@rdname export_calibration_info-methods
setMethod(
  "export_calibration_info",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    if (object@outcome_type %in% c("binomial", "multinomial")) {
      return(NULL)
      
    } else if (object@outcome_type %in% c("continuous")) {
      subtype <- "observed_value_range"
      
    } else if (object@outcome_type %in% c("survival")) {
      subtype <- "baseline_value"
      
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }
    
    return(.export(
      x = object,
      data_slot = "calibration_info",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "calibration",
      subtype = subtype,
      export_collection = export_collection
    ))
  }
)



# export_calibration_info (general) --------------------------------------------

#'@rdname export_calibration_info-methods
setMethod(
  "export_calibration_info",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "calibration_info"
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_calibration_info,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarTaskGenericFeatureInfo -----------------------------------------------
setClass(
  "familiarTaskGenericFeatureInfo",
  contains = "familiarTask",
  prototype = methods::prototype(
    task_name = "create_generic_feature_info",
    data_id = 1L,
    run_id = 1L
  )
)


# .set_file_name (generic feature info task) -----------------------------------
setMethod(
  ".set_file_name",
  signature(object = "familiarTaskGenericFeatureInfo"),
  function(object, file_paths = NULL) {
    if (is.null(file_paths)) return(object)
    
    # Generate file name of pre-processing file
    object@file <- get_object_file_name(
      object_type = "genericFeatureInfo",
      project_id = object@project_id,
      dir_path = file_paths$process_data_dir
    )
    
    return(object)
  }
)



# .get_task_descriptor (generic feature info task) -----------------------------
setMethod(
  ".get_task_descriptor",
  signature(object = "familiarTaskGenericFeatureInfo"),
  function(object, ...) {
    return(object@task_name)
  }
)



# .perform_task (generic feature info task, NULL) ------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskGenericFeatureInfo",
    data = "NULL"
  ),
  function(
    object,
    data,
    outcome_info = NULL,
    ...
  ) {
    # This method is called when "data" is expected to be available somewhere in
    # the backend.
    if (is.null(outcome_info)) {
      ..error_reached_unreachable_code("outcome_info is required.")
    }
    
    # Create a dataObject.
    data <- methods::new(
      "dataObject",
      data = get_data_from_backend(),
      preprocessing_level = "none",
      outcome_type = outcome_info@outcome_type,
      outcome_info = outcome_info
    )
    
    # Pass to .perform_task for dataObject.
    return(.perform_task(
      object = object,
      data = data,
      ...
    ))
  }
)



# .perform_task (generic feature info task, dataObject) ------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskGenericFeatureInfo",
    data = "dataObject"
  ),
  function(
    object,
    data,
    descriptor = NULL,
    experiment_data = NULL,
    return_results = TRUE,
    ...
  ) {
    
    # Check if the desired data already exist elsewhere.
    results_exist <- FALSE
    
    if (is(experiment_data, "experimentData")) {
      if (!is_empty(experiment_data@feature_info[["generic"]])) {
        feature_info_list <- experiment_data@feature_info[["generic"]]
        results_exist <- TRUE
      }
    }
    
    if (file.exists(object@file)) {
      feature_info_list <- update_object(object = readRDS(object@file))
      results_exist <- TRUE
    }
    
    if (results_exist) {
      if (!is.na(object@file)) {
        saveRDS(feature_info_list, file = object@file)
      }
      
      if (return_results) {
        return(feature_info_list)
      }
      
      return(TRUE)
    }
    
    # If this point is reached, results will be created de novo.
    # Extract basic feature information from the data.
    feature_info_list <- .get_generic_feature_info(
      data = data,
      outcome_type = data@outcome_type,
      descriptor = descriptor,
      project_id = object@project_id
    )
    
    # Write to file or return.
    if (!is.na(object@file)) {
      saveRDS(feature_info_list, file = object@file)
    }
    
    if (return_results) {
      return(feature_info_list)
    }
    
    return(TRUE)
  }
)



# .get_feature_info_list (generic feature info task) ---------------------------
setMethod(
  ".get_feature_info_list",
  signature(object = "familiarTaskGenericFeatureInfo"),
  function(object, feature_info_list, ...) {
    ..error_reached_unreachable_code(".get_feature_info_list does not exist for this task")
  }
)



# familiarTaskFeatureInfo ------------------------------------------------------
setClass(
  "familiarTaskFeatureInfo",
  contains = "familiarTask",
  prototype = methods::prototype(
    task_name = "create_feature_info"
  )
)



# .set_file_name (feature info task) -------------------------------------------
setMethod(
  ".set_file_name",
  signature(object = "familiarTaskFeatureInfo"),
  function(object, file_paths = NULL) {
    if (is.null(file_paths)) return(object)
    
    # Generate file name of pre-processing file
    object@file <- get_object_file_name(
      object_type = "featureInfo",
      project_id = object@project_id,
      data_id = object@data_id,
      run_id = object@run_id,
      dir_path = file_paths$process_data_dir
    )
    
    return(object)
  }
)



# .get_task_descriptor (feature info task) -------------------------------------
setMethod(
  ".get_task_descriptor",
  signature(object = "familiarTaskFeatureInfo"),
  function(object, ...) {
    return(paste0(object@task_name, "_", object@data_id, "_", object@run_id))
  }
)



# .perform_task (feature info task , NULL) -------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskFeatureInfo",
    data = "NULL"
  ),
  function(
    object,
    data,
    experiment_data = NULL,
    outcome_info = NULL,
    ...
  ) {
    # This method is called when "data" is expected to be available somewhere in
    # the backend.
    if (is.null(experiment_data)) {
      ..error_reached_unreachable_code("project_info is required for retrieving data from the backend.")
    }
    if (is.null(outcome_info)) {
      ..error_reached_unreachable_code("outcome_info is required.")
    }
    
    # Find the run list.
    run_list <- .get_run_list(
      iteration_list = experiment_data@iteration_list,
      data_id = object@data_id,
      run_id = object@run_id
    )
    
    # Select unique samples.
    sample_identifiers <- .get_sample_identifiers(
      run = run_list,
      train_or_validate = "train"
    )
    sample_identifiers <- unique(sample_identifiers)
    
    # Create a dataObject.
    data <- methods::new(
      "dataObject",
      data = get_data_from_backend(sample_identifiers = sample_identifiers),
      preprocessing_level = "none",
      outcome_type = outcome_info@outcome_type,
      outcome_info = outcome_info
    )
    
    # Pass to method that dispatches with dataObject for further processing.
    return(.perform_task(
      object = object,
      data = data,
      experiment_data = experiment_data,
      ...
    ))
  }
)


# .perform_task (feature info task, dataObject) --------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskFeatureInfo",
    data = "dataObject"
  ),
  function(
    object,
    data,
    experiment_data = NULL,
    settings = NULL,
    feature_info_list = NULL,
    message_indent = 0L,
    verbose = FALSE,
    cl = NULL,
    signature_features = NULL,
    novelty_features = NULL,
    fairness_features = NULL,
    return_results = TRUE,
    ...
  ) {
    logger_message(
      paste0(
        "Pre-processing: Starting preprocessing for run ",
        object@task_id, " of ",
        object@n_tasks, "."
      ),
      indent = message_indent,
      verbose = verbose
    )
    
    # Check if the desired data already exist elsewhere.
    results_exist <- FALSE
    
    if (is(experiment_data, "experimentData")) {
      if (!is_empty(experiment_data@feature_info[[paste0(object@data_id, ".", object@run_id)]])) {
        feature_info_list <- experiment_data@feature_info[[paste0(object@data_id, ".", object@run_id)]]
        results_exist <- TRUE
      }
    }
    
    if (file.exists(object@file)) {
      feature_info_list <- update_object(object = readRDS(object@file))
      results_exist <- TRUE
    }
    
    if (results_exist) {
      if (!is.na(object@file)) {
        saveRDS(feature_info_list, file = object@file)
      }
      
      if (return_results) {
        return(feature_info_list)
      }
      
      return(TRUE)
    }
    
    # Check that a feature info list is provided, otherwise create an ad-hoc
    # list as an template.
    if (is.null(feature_info_list)) {
      # Set up task, and explicitly don't write to file.
      generic_feature_info_task <- methods::new(
        "familiarTaskGenericFeatureInfo",
        project_id = object@project_id,
        file = NA_character_
      )
      
      # Execute the task.
      feature_info_list <- .perform_task(
        object = generic_feature_info_task,
        data = data
      )
    }
    
    # Add workflow control info.
    feature_info_list <- add_control_info(
      feature_info_list = feature_info_list,
      data_id = object@data_id,
      run_id = object@run_id,
      project_id = object@project_id
    )
    
    # Add signature feature info.
    if (is.null(signature_features)) signature_features <- settings$data$signature
    feature_info_list <- add_signature_info(
      feature_info_list = feature_info_list,
      signature = signature_features
    )
    
    # Add novelty feature info.
    if (is.null(novelty_features)) novelty_features <- settings$data$novelty_features
    feature_info_list <- add_novelty_info(
      feature_info_list = feature_info_list,
      novelty_features = novelty_features
    )
    
    # Add fairness feature info.
    if (is.null(fairness_features)) fairness_features <- settings$data$fairness_features
    feature_info_list <- add_fairness_info(
      feature_info_list = feature_info_list,
      fairness_features = fairness_features
    )
    
    # Find currently available features.
    available_features <- get_available_features(feature_info_list = feature_info_list)
    
    # Remove unavailable features from the data object.
    data <- filter_features(
      data = data,
      available_features = available_features
    )
    
    # Use data to determine pre-processing parameters.
    feature_info_list <- .determine_preprocessing_parameters(
      cl = cl,
      data = data,
      feature_info_list = feature_info_list,
      settings = settings,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    if (!is.na(object@file)) {
      saveRDS(feature_info_list, file = object@file)
    }
    
    if (return_results) {
      return(feature_info_list)
    }
    
    return(TRUE)
  }
)



# .get_feature_info_list (feature info task) -----------------------------------
setMethod(
  ".get_feature_info_list",
  signature(object = "familiarTaskFeatureInfo"),
  function(object, feature_info_list, ...) {
    ..error_reached_unreachable_code(".get_feature_info_list does not exist for this task")
  }
)


# .get_variable_importance_table (feature info task) ---------------------------
setMethod(
  ".get_variable_importance_table",
  signature(object = "familiarTaskFeatureInfo"),
  function(
    object,
    ...
  ) {
    ..error_reached_unreachable_code(".get_variable_importance_table does not exist for this task")
  }
)


.generate_data_preprocessing_tasks <- function(
    data_ids,
    run_ids,
    file_paths,
    project_id,
    skip_existing = FALSE
  ) {
  task_list <- list()
  
  # Create task to generic feature_info.
  generic_info_task <- methods::new(
    "familiarTaskGenericFeatureInfo",
    project_id = project_id
  )
  
  # Add file names.
  generic_info_task <- .set_file_name(
    object = generic_info_task,
    file_paths = file_paths
  )
  
  # Add to list, if the file does not exist on disk.
  if (!skip_existing || !.file_exists(generic_info_task)) {
    task_list[[1L]] <- generic_info_task
  }
  
  ii <- 2L
  for (data_id in data_ids) {
    for (run_id in run_ids) {
      # Create task to generate run-specific feature info.
      run_info_task <- methods::new(
        "familiarTaskFeatureInfo",
        data_id = data_id,
        run_id = run_id,
        project_id = project_id
      )
      
      # Add file names.
      run_info_task <- .set_file_name(
        object = run_info_task,
        file_paths = file_paths
      )
      
      # Add to list, if the file does not exist on disk.
      if (!skip_existing || !.file_exists(run_info_task)) {
        task_list[[ii]] <- run_info_task
        ii <- ii + 1L
      }
    }
  }
  
  return(task_list)
}



.run_preprocessing <- function(
    tasks,
    experiment_data,
    ...
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  data_id <- run_id <- list_name <- complete <- NULL
  
  # Create or load generic feature info.
  if (!is_empty(tasks$generic_feature_info)) {
    if (!.file_exists(tasks$generic_feature_info[[1L]])) {
      generic_feature_info <- .perform_task(
        object = tasks$generic_feature_info[[1L]],
        data = NULL,
        experiment_data = experiment_data,
        ...
      )
      
    } else {
      generic_feature_info <- readRDS(tasks$generic_feature_info[[1L]]@file)
    }
    
  } else {
    generic_feature_info <- NULL
  }
  
  # Check that any feature info tasks are required.
  if (is_empty(tasks$feature_info)) return(invisible(FALSE))
  
  # Determine which feature info objects need to be obtained.
  finished_tasks <- sapply(tasks$feature_info, .file_exists)
  unfinished_tasks <- tasks$feature_info[!finished_tasks]
  finished_tasks <- tasks$feature_info[finished_tasks]
  
  # Process any unfinished tasks.
  if (length(unfinished_tasks) > 0L) {
    ..run_preprocessing(
      tasks = unfinished_tasks,
      generic_feature_info = generic_feature_info,
      experiment_data = experiment_data,
      ...
    )
  }
  
  # Load processed data from files.
  feature_info_list <- lapply(
    tasks$feature_info,
    function(x) (readRDS(x@file))
  )
  
  # Set names.
  names(feature_info_list) <- sapply(
    tasks$feature_info,
    function(x) (paste0(x@data_id, ".", x@run_id))
  )
  
  # Attach generic feature information.
  feature_info_list[["generic"]] <- generic_feature_info
  
  # Attach feature preprocessing information to the backend.
  .assign_feature_info_to_backend(feature_info_list = feature_info_list)
  
  return(invisible(TRUE))
}



..run_preprocessing <- function(
    tasks,
    generic_feature_info,
    settings,
    cl,
    message_indent = 0L,
    verbose,
    ...
) {
  logger_message(
    paste0(
      "Pre-processing: Start identifying data processing parameters."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  # Determine how parallel processing takes place.
  if (settings$prep$do_parallel %in% c("TRUE", "inner")) {
    # Parallel processing in inner function, i.e. within each data subset.
    cl_inner <- cl
    cl_outer <- NULL
    
  } else if (settings$prep$do_parallel %in% c("outer")) {
    # Parallel processing in outer loop, i.e. over all data subsets.
    cl_inner <- NULL
    cl_outer <- cl
    
    if (!is.null(cl_outer)) {
      logger_message(
        paste0(
          "Pre-processing: Load-balanced parallel processing is done in the outer loop. ",
          "No progress can be displayed."
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
  } else {
    # No parallel processing.
    cl_inner <- cl_outer <- NULL
  }
  
  # Iterate over data subsets for which parameters have not yet been set.
  fam_mapply_lb(
    cl = cl_outer,
    assign = "data",
    FUN = .perform_task,
    progress_bar = !is.null(cl_outer),
    object = tasks,
    MoreArgs = list(
      "cl" = cl_inner,
      "data" = NULL,
      "feature_info_list" = generic_feature_info,
      "settings" = settings,
      "message_indent" = message_indent + 1L,
      "verbose" = verbose && is.null(cl_outer),
      "return_results" = FALSE,
      ...
    )
  )
  
  logger_message(
    paste0(
      "Pre-processing: Completed identifying data processing parameters.",
      "\n"
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  return(invisible(TRUE))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarDataElementSampleSimilarity object -----------------------------------
setClass(
  "familiarDataElementSampleSimilarity",
  contains = "familiarDataElement",
  slots = list(
    "similarity_metric" = "character",
    "cluster_method" = "character",
    "linkage_method" = "character",
    "cluster_cut_method" = "character",
    "similarity_threshold" = "ANY",
    "dendrogram" = "ANY"
  ),
  prototype = methods::prototype(
    detail_level = "ensemble",
    estimation_type = "point",
    similarity_metric = NA_character_,
    cluster_method = NA_character_,
    linkage_method = NA_character_,
    cluster_cut_method = NA_character_,
    similarity_threshold = NULL,
    dendrogram = NULL,
    value_column = "value",
    grouping_column = c("sample_1", "sample_2")
  )
)



# extract_sample_similarity (generic) ------------------------------------------

#'@title Internal function to extract the sample distance table.
#'
#'@description Computes and extracts the sample distance table for samples
#'  analysed using a `familiarEnsemble` object to form a `familiarData` object.
#'  This table can be used to cluster samples, and is exported directly by
#'  `extract_feature_expression`.
#'
#'@inheritParams .extract_data
#'
#'@return An object containing pairwise distance between samples. This data
#'  is only the upper triangular of the complete matrix (i.e. the sparse
#'  unitriangular representation). Diagonals will always be 0.0 and the lower
#'  triangular is mirrored.
#'  
#'@md
#'@keywords internal
setGeneric(
  "extract_sample_similarity",
  function(
    object,
    data,
    cl = NULL,
    is_pre_processed = FALSE,
    sample_limit = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    verbose = FALSE,
    message_indent = 0L,
    ...
  ) {
    standardGeneric("extract_sample_similarity")
  }
)



# extract_sample_similarity (familiarEnsemble) ---------------------------------
setMethod(
  "extract_sample_similarity",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    is_pre_processed = FALSE,
    sample_limit = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    verbose = FALSE,
    message_indent = 0L,
    ...
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Computing pairwise similarity between samples."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Obtain sample cluster method from stored settings, if required.
    if (is.waive(sample_cluster_method)) {
      sample_cluster_method <- object@settings$sample_cluster_method
    }
    
    # Obtain sample linkage function from stored settings, if required.
    if (is.waive(sample_linkage_method)) {
      sample_linkage_method <- object@settings$sample_linkage_method
    }
    
    # Obtain sample similarity metric from stored settings, if required.
    if (is.waive(sample_similarity_metric)) {
      sample_similarity_metric <- object@settings$sample_similarity_metric
    }
    
    # Replace sample cluster method == "none" with "hclust"
    if (sample_cluster_method == "none") {
      sample_cluster_method <- "hclust"
    }
    
    .check_cluster_parameters(
      cluster_method = sample_cluster_method,
      cluster_linkage = sample_linkage_method,
      cluster_similarity_metric = sample_similarity_metric,
      data_type = "sample"
    )
    
    # Check the sample limit.
    sample_limit <- .parse_sample_limit(
      x = sample_limit,
      object = object,
      default = Inf,
      data_element = "sample_similarity"
    )
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "familiarDataElementSampleSimilarity",
      similarity_metric = sample_similarity_metric,
      cluster_method = sample_cluster_method,
      linkage_method = sample_linkage_method
    )
    
    # Generate elements to send to dispatch.
    similarity_data <- extract_dispatcher(
      FUN = .extract_sample_similarity,
      has_internal_bootstrap = FALSE,
      cl = cl,
      object = object,
      data = data,
      sample_limit = sample_limit,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      aggregate_results = TRUE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(similarity_data)
  }
)



# extract_sample_similarity (familiarDataElementPredictionTable) ---------------
setMethod(
  "extract_sample_similarity",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("sample similarity")
    
    return(NULL)
  }
)



# extract_sample_similarity (dataObject) ---------------------------------------
setMethod(
  "extract_sample_similarity",
  signature(object = "dataObject"),
  function(
    object,
    data,
    cl = NULL,
    is_pre_processed = FALSE,
    sample_limit = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    verbose = FALSE,
    message_indent = 0L,
    ...
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Computing pairwise similarity between samples."),
      indent = message_indent,
      verbose = verbose
    )
    
    settings <- .parse_sample_clustering(
      sample_cluster_method = sample_cluster_method,
      sample_linkage_method = sample_linkage_method,
      sample_similarity_metric = sample_similarity_metric
    )
    
    # Set default cluster method, if required.
    if (is.waive(sample_cluster_method)) {
      sample_cluster_method <- settings$sample_cluster_method
    }
    
    # Set default linkage function, if required.
    if (is.waive(sample_linkage_method)) {
      sample_linkage_method <- settings$sample_linkage_method
    }
    
    # Set default similarity metric, if required.
    if (is.waive(sample_similarity_metric)) {
      sample_similarity_metric <- settings$sample_similarity_metric
    }
    
    # Replace sample cluster method == "none" with "hclust"
    if (sample_cluster_method == "none") {
      sample_cluster_method <- "hclust"
    }
    
    # Check the sample limit.
    sample_limit <- .parse_sample_limit(
      x = sample_limit,
      object = object,
      default = Inf,
      data_element = "sample_similarity"
    )
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "familiarDataElementSampleSimilarity",
      similarity_metric = sample_similarity_metric,
      cluster_method = sample_cluster_method,
      linkage_method = sample_linkage_method
    )
    
    # Assume data is transformed.
    object@preprocessing_level <- "transformation"
    
    # Generate feature info by running the familiarTaskGenericFeatureInfo task.
    generic_feature_info_task <- new("familiarTaskGenericFeatureInfo")
    object@feature_info <- .perform_task(
      object = generic_feature_info_task,
      data = object
    )
    
    # Generate elements to send to dispatch.
    similarity_data <- extract_dispatcher(
      FUN = .extract_sample_similarity,
      has_internal_bootstrap = FALSE,
      cl = cl,
      object = object,
      data = object,
      sample_limit = sample_limit,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      aggregate_results = TRUE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(similarity_data)
  }
)



.extract_sample_similarity <- function(
    object,
    data,
    proto_data_element,
    cl = NULL,
    sample_limit,
    is_pre_processed,
    message_indent,
    aggregate_results = TRUE,
    verbose = FALSE,
    ...
) {
  
  # Add the name of the ensemble model
  data_element <- add_model_name(
    data = proto_data_element,
    object = object
  )
  
  # Retrieve input data.
  data <- process_input_data(
    object = object,
    data = data,
    stop_at = "imputation",
    is_pre_processed = is_pre_processed
  )
  
  # Check if the input data is not empty
  if (is_empty(data)) return(NULL)
  
  # Check if the number of samples is sufficient to form pairs (>= 2), and
  # return an empty table if not.
  if (get_n_samples(data) < 2L) return(data_element)
  
  # Select samples up to sample_limit.
  data <- get_subsample(
    data = data,
    size = sample_limit,
    seed = 0L
  )
  
  if (is(object, "familiarEnsemble")){
    # Maintain only important features when assessing data from the perspective
    # of an ensemble. The current set is based on the required features.
    data <- filter_features(
      data = data,
      available_features = object@model_features
    )
    
    # Aggregate features.
    data <- aggregate_data(data = data)
  }
  
  # Identify eligible columns.
  feature_columns <- get_feature_columns(x = data)
  
  # Compute the similarity table
  data_element@data <- set_similarity_table(
    data = data,
    feature_info_list = object@feature_info[feature_columns],
    similarity_metric = data_element@similarity_metric,
    data_type = "sample",
    cl = cl,
    message_indent = message_indent + 1L,
    verbose = verbose
  )
  
  # Merge data elements
  data_elements <- merge_data_elements(list(data_element))
  
  if (aggregate_results) data_elements <- .compute_data_element_estimates(x = data_elements)
  
  return(data_elements)
}



.append_sample_similarity_dendrogram <- function(x) {
  
  if (is_empty(x)) return(x)
  
  # Create a cluster method object using data stored in x.
  cluster_method_object <- .create_sample_similarity_cluster_method_object(x = x)
  
  if (is.null(cluster_method_object)) return(x)
  
  # Create the cluster object.
  object <- apply_cluster_method(object = cluster_method_object)
  
  # Attach to data element.
  x@dendrogram <- object@object
  
  return(x)
}



.compute_sample_similarity_cluster_table <- function(x) {
  # Computes the sample similarity cluster table from the similarity table in x.
  
  # Create a cluster method object using data stored in x.
  cluster_method_object <- .create_sample_similarity_cluster_method_object(x = x)
  
  if (is.null(cluster_method_object)) return(NULL)
  
  # Compute the cluster table.
  cluster_table <- create_clusters(
    object = cluster_method_object,
    as_cluster_object = FALSE
  )
  
  return(cluster_table)
}



.create_sample_similarity_cluster_method_object <- function(x) {
  
  if (is_empty(x)) return(NULL)
  
  if (length(x@similarity_threshold) > 1L) {
    # Remove 1.0 because that does not yield clustering info.
    available_thresholds <- setdiff(x@similarity_threshold, 1.0)
    
    # Select the maximum threshold.
    x@similarity_threshold <- max(available_thresholds)
  }
  
  # Create cluster method object.
  cluster_method_object <- create_cluster_method_object(
    cluster_method = x@cluster_method,
    data_type = "sample",
    cluster_linkage = x@linkage_method,
    cluster_cut_method = "none",
    cluster_similarity_metric = x@similarity_metric,
    cluster_representation_method = "none"
  )
  
  # Attach the similarity table to the cluster_method_object.
  cluster_method_object@similarity_table <- methods::new(
    "similarityTable",
    data = x@data[, mget(c("sample_1", "sample_2", "value"))],
    similarity_metric = x@similarity_metric,
    data_type = cluster_method_object@data_type
  )
  
  return(cluster_method_object)
}




..limit_sample_similarity_samples <- function(x, sample_limit) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  sample_1 <- sample_2 <- NULL
  
  if (is_empty(x)) return(x)
  
  # Find the names of samples.
  sample_names <- unique(c(x@data$sample_1, x@data$sample_2))
  
  # Check the number of samples against the limit.
  if (length(sample_names) <= sample_limit) return(x)
  
  # Select samples.
  sample_names <- fam_sample(
    sample_names,
    size = sample_limit,
    replace = FALSE,
    seed = 0L
  )
  
  # Select only the selected samples.
  x@data <- x@data[sample_1 %in% sample_names & sample_2 %in% sample_names]
  
  return(x)
}



# export_sample_similarity (generic) -------------------------------------------

#'@title Extract and export mutual correlation between features.
#'
#'@description Extract and export mutual correlation between features in a
#'  familiarCollection.
#'
#'@param export_dendrogram Add dendrogram in the data element objects.
#'
#'@inheritParams export_all
#'@inheritParams .extract_data
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams as_familiar_collection
#'@inheritDotParams as_data_object
#'
#'@details 
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Sample similarity data can be created from `dataObject`, or `data.table` objects.
#'  For `data.table`, see \code{\link{as_data_object}} for additional arguments.
#'
#'@return A list containing a data.table (if `dir_path` is not provided), or
#'  nothing, as all data is exported to `csv` files.
#'@exportMethod export_sample_similarity
#'@md
#'@rdname export_sample_similarity-methods
setGeneric(
  "export_sample_similarity",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    sample_limit = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    export_dendrogram = FALSE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_sample_similarity")
  }
)



# export_sample_similarity (collection) ----------------------------------------

#'@rdname export_sample_similarity-methods
setMethod(
  "export_sample_similarity",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    sample_limit = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    export_dendrogram = FALSE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Extract data.
    x <- object@sample_similarity
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    # Check sample cluster method.
    if (!is.waive(sample_cluster_method)) {
      
      # Set clustering method.
      x <- lapply(
        x,
        function(x, sample_cluster_method) {
          x@cluster_method <- sample_cluster_method
          return(x)
        },
        sample_cluster_method = sample_cluster_method
      )
    }
    
    # Check sample linkage method.
    if (!is.waive(sample_linkage_method)) {
      
      # Set linkage method.
      x <- lapply(
        x,
        function(x, sample_linkage_method) {
          x@linkage_method <- sample_linkage_method
          return(x)
        },
        sample_linkage_method = sample_linkage_method
      )
    }
    
    # Check the sample limit.
    if (!is.waive(sample_limit)) {
      .check_number_in_valid_range(
        x = sample_limit,
        var_name = "sample_limit",
        range = c(20L, Inf)
      )
      
    } else {
      sample_limit <- Inf
    }
    
    # Check whether the input parameters are valid and create a cluster
    # object.
    .check_cluster_parameters(
      cluster_method = x[[1L]]@cluster_method,
      data_type = "sample",
      cluster_linkage = x[[1L]]@linkage_method,
      cluster_cut_method = "none",
      cluster_similarity_metric = x[[1L]]@similarity_metric,
      cluster_representation_method = "none"
    )
    
    if (aggregate_results || export_dendrogram) {
      x <- .compute_data_element_estimates(x)
      
      # Limit the number of samples.
      if (is.finite(sample_limit)) {
        x <- lapply(
          x,
          ..limit_sample_similarity_samples,
          sample_limit = sample_limit
        )
      }
      
      # Add clustering information.
      if (export_dendrogram) {
        x <- lapply(x, .append_sample_similarity_dendrogram)
      }
    }
    
    return(.export(
      x = object,
      data_elements = x,
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "sample_similarity",
      subtype = x[[1L]]@similarity_metric,
      export_dendrogram = export_dendrogram,
      export_collection = export_collection
    ))
  }
)


# export_sample_similarity (general) -------------------------------------------

#'@rdname export_sample_similarity-methods
setMethod(
  "export_sample_similarity",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    sample_limit = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "feature_similarity",
          "sample_limit" = sample_limit,
          "aggregate_results" = aggregate_results,
          "sample_cluster_method" = sample_cluster_method,
          "sample_linkage_method" = sample_linkage_method
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_sample_similarity,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "sample_cluster_method" = sample_cluster_method,
          "sample_linkage_method" = sample_linkage_method,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)



# .export (familiarDataElementFeatureSimilarity) -------------------------------
setMethod(
  ".export",
  signature(x = "familiarDataElementSampleSimilarity"),
  function(
    x,
    x_list,
    aggregate_results = FALSE, 
    export_dendrogram,
    ...
  ) {
    # This is like .export,familiarDataElement, but the elements are merged
    # prior to computing estimates.
    
    # Only merge if dendrograms are missing for all entries.
    if (!export_dendrogram) {
      # Merge data elements.
      x <- merge_data_elements(
        x = x_list,
        as_data = "all",
        as_grouping_column = TRUE,
        force_data_table = TRUE
      )
      
    } else {
      x <- x_list
    }
    
    if (aggregate_results) {
      x <- .compute_data_element_estimates(x)
    }
    
    return(x)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
#' @include PredictionTable.R
NULL


# familiarDataElementRiskStrata (object) ---------------------------------------
setClass(
  "familiarDataElementRiskStrata",
  contains = "familiarDataElement"
)

# familiarDataElementRiskHazardRatio (object) ----------------------------------
setClass(
  "familiarDataElementRiskHazardRatio",
  contains = "familiarDataElement"
)

# familiarDataElementRiskLogrank (object) --------------------------------------
setClass(
  "familiarDataElementRiskLogrank",
  contains = "familiarDataElement"
)



# extract_risk_stratification_data (generic) -----------------------------------

#'@title Internal function to extract stratification data.
#'
#'@description Computes and extracts stratification data from a
#'  `familiarEnsemble` object. This includes the data required to draw
#'  Kaplan-Meier plots, as well as logrank and hazard-ratio tests between the
#'  respective risk groups.
#'
#'@inheritParams .extract_data
#'
#'@return A list with data.tables containing information concerning risk group
#'  stratification.
#'@md
#'@keywords internal
setGeneric(
  "extract_risk_stratification_data",
  function(
    object,
    data,
    cl = NULL,
    is_pre_processed = FALSE,
    ensemble_method = waiver(),
    detail_level = waiver(),
    confidence_level = waiver(),
    message_indent = 0L,
    verbose = FALSE, 
    ...
  ) {
    standardGeneric("extract_risk_stratification_data")
  }
)



# extract_risk_stratification_data (familiarEnsemble) --------------------------
setMethod(
  "extract_risk_stratification_data",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    is_pre_processed = FALSE,
    ensemble_method = waiver(),
    detail_level = waiver(),
    confidence_level = waiver(),
    stratification_method = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    
    # This mostly follows the same routines as extract_prediction_data. In
    # addition, tests are created during export.
    
    # Only assess stratification for survival outcomes.
    if (!object@outcome_type %in% c("survival")) return(NULL)
    
    # Message extraction start
    logger_message(
      paste0("Assessing stratification into risk groups."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Load confidence alpha from object settings attribute if not provided
    # externally.
    if (is.waive(confidence_level)) {
      confidence_level <- object@settings$confidence_level
    }
    
    # Check confidence_level input argument
    .check_number_in_valid_range(
      x = confidence_level, 
      var_name = "confidence_level",
      range = c(0.0, 1.0),
      closed = c(FALSE, FALSE)
    )
    
    # Obtain ensemble method from stored settings, if required.
    if (is.waive(ensemble_method)) {
      ensemble_method <- object@settings$ensemble_method
    }
    
    # Check ensemble_method argument
    .check_parameter_value_is_valid(
      x = ensemble_method, 
      var_name = "ensemble_method",
      values = .get_available_ensemble_prediction_methods()
    )
    
    # Check the level detail.
    detail_level <- .parse_detail_level(
      x = detail_level,
      object = object,
      default = "ensemble",
      data_element = "risk_stratification_data"
    )
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Load models.
    model_list <- ..get_model(object = object)
    if (is_empty(model_list)) return(NULL)
    
    # Check available stratification methods.
    available_stratification_method <- unique(unlist(lapply(
      model_list,
      function(fam_model) (fam_model@km_info$stratification_method)
    )))
    
    # Check that any are available.
    if (is.null(available_stratification_method)) return(NULL)
    
    # Prepare the stratification_method variable for comparison. The default is
    # to use all methods present in the dataset.
    if (is.waive(stratification_method)) {
      stratification_method <- available_stratification_method
    }
    
    # Check that the stratification method is valid.
    .check_parameter_value_is_valid(
      x = stratification_method,
      var_name = "stratification_method",
      values = available_stratification_method
    )

    # Aggregate data. It does not make sense to keep duplicate rows here.
    data <- aggregate_data(data = data)
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "predictionTableRiskGroups",
      detail_level = detail_level,
      confidence_level = confidence_level,
      estimation_type = "point"
    )
    
    # Generate elements to send to dispatch.
    performance_data <- extract_dispatcher(
      FUN = .extract_risk_stratification_data,
      has_internal_bootstrap = FALSE,
      cl = cl,
      object = object,
      data = data,
      aggregate_results = TRUE,
      stratification_method = stratification_method,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      time = object@settings$time_max,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(performance_data)
  }
)



# extract_risk_stratification_data (risk groups) -------------------------------
setMethod(
  "extract_risk_stratification_data",
  signature(object = "predictionTableRiskGroups"),
  function(
    object,
    confidence_level = waiver(),
    ...
  ) {
    if (is_empty(object)) return(NULL)
    
    # Reference labels should be present.
    if (!.has_reference_data(object)) {
      ..warning_prediction_table_lacks_reference("risk stratification")
      return(NULL)
    }
    
    # Ensure that all slots are merged into data.
    object <- .copy(object)
    object <- .merge_slots_into_data(object)
    
    # Load confidence alpha from object settings attribute if not provided
    # externally.
    if (is.waive(confidence_level)) {
      confidence_level <- 0.95
    }
    
    # Check confidence_level input argument
    .check_number_in_valid_range(
      x = confidence_level, 
      var_name = "confidence_level",
      range = c(0.0, 1.0),
      closed = c(FALSE, FALSE)
    )
    
    # Set confidence level.
    object@confidence_level <- confidence_level
    
    return(object)
  }
)



# extract_risk_stratification_data (dataObject) --------------------------------
setMethod(
  "extract_risk_stratification_data",
  signature(object = "dataObject"),
  function(
    object,
    data,
    cl = NULL,
    is_pre_processed = FALSE,
    ensemble_method = waiver(),
    detail_level = waiver(),
    confidence_level = waiver(),
    stratification_method = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    risk_group_column = waiver(),
    ...
  ) {
    
    # This mostly follows the same routines as extract_prediction_data. In
    # addition, tests are created during export.
    
    # Only assess stratification for survival outcomes.
    if (!object@outcome_type %in% c("survival")) return(NULL)
    
    # Message extraction start
    logger_message(
      paste0("Assessing stratification into risk groups."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Set default confidence alpha from object settings attribute if not provided
    # externally.
    if (is.waive(confidence_level)) confidence_level <- 0.95
    
    # Check confidence_level input argument
    .check_number_in_valid_range(
      x = confidence_level, 
      var_name = "confidence_level",
      range = c(0.0, 1.0),
      closed = c(FALSE, FALSE)
    )

    if (is.waive(risk_group_column)) risk_group_column <- get_id_columns("batch")
    
    # Check if only one column is provided as risk_group_column.
    .check_argument_length(
      x = risk_group_column,
      var_name = "risk_group_column",
      min = 1L,
      max = 1L
    )
    
    # Check that the risk group column can be found in the dataset.
    if (!risk_group_column %in% colnames(object@data)) {
      ..error(paste0(risk_group_column, " is not a column in the data."))
    }
    
    # Convert to prediction table.
    prediction_data <- as_prediction_table(
      x = object@data[[risk_group_column]],
      type = "risk_stratification",
      data = object
    )
    prediction_data <- .merge_slots_into_data(prediction_data)
    
    # Manually set attributes for this dataElement.
    prediction_data@detail_level <- "ensemble"
    prediction_data@estimation_type <- "point"
    prediction_data@confidence_level <- confidence_level
    prediction_data@bootstrap_ci_method <- "percentile"
    prediction_data@is_aggregated <- FALSE
    prediction_data@learner <- "none"
    prediction_data@vimp_method <- "none"
    
    # Add model name.
    prediction_data <- add_model_name(prediction_data, object = object)
    
    # Add stratification methods.
    prediction_data <- add_data_element_identifier(
      x = prediction_data,
      stratification_method = "external"
    )
    
    return(prediction_data)
  }
)



# extract_risk_stratification_data (prediction table) --------------------------
setMethod(
  "extract_risk_stratification_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("risk stratification")
    
    return(NULL)
  }
)



# extract_risk_stratification_data (NULL) --------------------------------------
setMethod(
  "extract_risk_stratification_data",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



.extract_risk_stratification_data <- function(
    object,
    proto_data_element,
    cl = NULL,
    stratification_method,
    ...
) {
  # Add model name.
  proto_data_element <- add_model_name(proto_data_element, object = object)
  
  # Add stratification methods.
  data_elements <- add_data_element_identifier(
    x = proto_data_element,
    stratification_method = stratification_method
  )
  
  # Iterate over data elements.
  data_elements <- lapply(
    data_elements,
    ..extract_risk_stratification_data,
    object = object,
    cl = cl,
    ...
  )
  
  return(data_elements)
}



..extract_risk_stratification_data <- function(
    data_element,
    object,
    data,
    cl = NULL,
    is_pre_processed,
    ensemble_method,
    time,
    ...
) {
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Compute performance data.
  prediction_data <- .predict(
    object = object,
    data = data,
    ensemble_method = ensemble_method,
    time = time,
    type = "risk_stratification",
    stratification_method = data_element@identifiers$stratification_method,
    is_pre_processed = is_pre_processed
  )

  # Ensure that all slots are merged into data.
  prediction_data <- .merge_slots_into_data(prediction_data)
  
  if (is_empty(prediction_data)) return(NULL)
  
  # We will actually use prediction_data from here, but need to update some
  # attributes.
  prediction_data@detail_level <- data_element@detail_level
  prediction_data@estimation_type <- data_element@estimation_type
  prediction_data@confidence_level <- data_element@confidence_level
  prediction_data@bootstrap_ci_method <- data_element@bootstrap_ci_method
  prediction_data@is_aggregated <- FALSE
  prediction_data@identifiers <- c(prediction_data@identifiers, data_element@identifiers)
  
  return(prediction_data)
}



.compute_risk_stratification_curves <- function(x, time_range = NULL) {
  
  if (is_empty(x)) return(NULL)
  if (!all_predictions_valid(x)) return(NULL)
  
  if (!x@is_aggregated) {
    ..error_reached_unreachable_code(
      ".compute_risk_stratification_curves: expecting aggregated data."
    )
  }

  # Make local copy.
  data <- .copy(x)
  
  # Remove data with missing predictions.
  data <- remove_invalid_predictions(data)
  
  # Remove data with missing outcomes.
  data <- filter_missing_outcome(data, outcome_type = "survival")
  
  # Check that any prediction data remain.
  if (is_empty(data)) return(NULL)
  
  # Create new element with strata based on x.
  data_element <- methods::new(
    "familiarDataElementRiskStrata",
    x,
    grouping_column = c(
      setdiff(x@grouping_column, c(get_non_feature_columns(x = "survival"))),
      "group"
    ),
    value_column = "survival"
  )
  
  # Collect strata
  data <- .merge_slots_into_data(data)
  strata <- data@data[
    ,
    ..compute_risk_stratification_curves(
      .SD, 
      confidence_level = data_element@confidence_level,
      time_range = time_range
    ),
    by = c(data_element@grouping_column),
    .SDcols = c("outcome_time", "outcome_event")
  ]
  
  # Check that strata are not empty.
  if (is_empty(strata)) return(NULL)
  
  # Store strata data.
  data_element@data <- strata
  
  return(data_element)
}



..compute_risk_stratification_curves <- function(x, confidence_level, time_range) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  survival <- ci_low <- ci_up <- time <- NULL
  
  if (is_empty(x)) return(NULL)
  
  # Obtain Kaplan-Meier survival curves.
  km_fit <- tryCatch(
    survival::survfit(
      Surv(outcome_time, outcome_event) ~ 1,
      data = x,
      conf.int = confidence_level
    ),
    error = identity
  )
  
  # Check if the survival curve could be generated at all. Causes could be lack
  # of events, no events beyond the first time point, etc.
  if (inherits(km_fit, "error")) return(NULL)
  
  # Extract plotting information
  km_data <- data.table::data.table(
    "time" = km_fit$time,
    "group_size" = km_fit$n.risk,
    "n_event" = km_fit$n.event,
    "n_censor" = km_fit$n.censor,
    "survival" = km_fit$surv,
    "ci_low" = km_fit$lower,
    "ci_up" = km_fit$upper
  )
  
  # Add in data at time = 0 if necessary.
  if (min(km_data$time) > 0.0) {
    km_data <- rbind(
      data.table::data.table(
        "time" = 0.0,
        "group_size" = km_fit$n,
        "n_event" = 0L,
        "n_censor" = 0L,
        "survival" = 1.00,
        "ci_low" = km_fit$lower[1L],
        "ci_up" = 1.0
      ),
      km_data
    )
  }
  
  # Update absent censoring (notably when survival is 0).
  km_data[survival == 0.0 & is.na(ci_low), "ci_low" := 0.0]
  km_data[survival == 0.0 & is.na(ci_up), "ci_up" := 0.0]
  
  # In rare circumstances (i.e. single-sample risk groups), ci_low may be
  # missing at the initial time point.
  km_data[time == 0.0 & is.na(ci_low), "ci_low" := 0.0]
  
  if (!is.null(time_range)) {
    # Add an entry at the proximal and distal range edges. This prevents the
    # curve from being cut off prematurely, or starting too late.
    if (
      is_empty(km_data[time == time_range[1L]]) &&
      time_range[1L] > min(km_data$time) &&
      time_range[1L] < max(km_data$time)
    ) {
      # Select the closest entry prior to the proximal edge, make changes and
      # introduce it back into the data.
      proximal_data <- tail(km_data[time < time_range[1L]][order(time)], n = 1L)
      proximal_data[, ":="(
        "time" = time_range[1L],
        "n_event" = 0L,
        "n_censor" = 0L
      )]
      km_data <- rbind(km_data, proximal_data)[order(time)]
    }
    
    if (
      is_empty(km_data[time == time_range[2L]]) &&
      time_range[2L] > min(km_data$time) &&
      time_range[2L] < max(km_data$time)
    ) {
      # Do the same for the distal edge.
      distal_data <- tail(km_data[time < time_range[2L]][order(time)], n = 1L)
      distal_data[, ":="(
        "time" = time_range[2L],
        "n_event" = 0L,
        "n_censor" = 0L
      )]
      km_data <- rbind(km_data, distal_data)[order(time)]
    }
    
    # Limit strata to the time range.
    km_data <- km_data[time >= time_range[1L] & time <= time_range[2L]]
  }
  
  return(as.list(km_data))
}



.compute_risk_stratification_tests <- function(x, time_range) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome_time <- NULL
  
  if (is_empty(x)) return(NULL)
  
  if (!x@is_aggregated) {
    ..error_reached_unreachable_code(
      ".compute_risk_stratification_test: expecting aggregated data."
    )
  }
  
  # Make local copy.
  x <- .copy(x)
  
  # Remove data with missing predictions or missing outcomes.
  x <- remove_invalid_predictions(x)
  x <- filter_missing_outcome(x, outcome_type = "survival")
  
  # Check that any prediction data remain.
  if (is_empty(x)) return(NULL)
  
  # Right-censor the data to the desired time window.
  if (!is.null(time_range)) {
    # Right censor data past the time range.
    x@data[outcome_time > time_range[2L], "outcome_event" := 0L]
  } 

  # Check that 2 (or more risk groups) are present.
  if (length(x@groups) < 2L) return(NULL)
  
  # Get data from logrank tests.
  logrank_data <- .compute_risk_stratification_logrank_test(
    x = x,
    time_range = time_range
  )
  
  # Get data from hazard ratio tests.
  hr_data <- .compute_risk_stratification_hazard_ratio_test(
    x = x,
    confidence_level = x@confidence_level
  )
  
  # Return test results
  return(list(
    "logrank" = logrank_data,
    "hazard_ratio" = hr_data
  ))
}



.compute_risk_stratification_logrank_test <- function(x, time_range) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  p_value <- NULL
  
  # Create new element with strata based on x.
  data_element <- methods::new(
    "familiarDataElementRiskLogrank",
    x,
    grouping_column = c(setdiff(
      x@grouping_column,
      c(get_non_feature_columns(x = "survival"))
    )),
    value_column = c("p_value", "p_value_adjusted")
  )
  
  # Compute logrank test over all risk groups.
  data <- .as_data_table(x)
  overall_results <- data[
    ,
    ..compute_risk_stratification_logrank_test(x = .SD),
    by = c(data_element@grouping_column),
    .SDcols = c("outcome_time", "outcome_event", "group")
  ]
  
  # Get all pairs
  pairs <- utils::combn(
    x = x@groups,
    m = 2L,
    simplify = FALSE
  )
  
  # Compute logrank test results over all risk groups.
  pairwise_results <- data[
    ,
    dmapply(
      ..compute_risk_stratification_logrank_test,
      selected_groups = pairs,
      MoreArgs = list("x" = .SD)
    ),
    by = c(data_element@grouping_column),
    .SDcols = c("outcome_time", "outcome_event", "group")
  ]
  
  # Compute multiple-testing corrected p-value.
  pairwise_results[
    ,
    "p_value_adjusted" := stats::p.adjust(p_value, method = "holm"),
    by = c(data_element@grouping_column)
  ]
  
  # Combine to single table and return.
  data_element@data <- data.table::rbindlist(
    list(overall_results, pairwise_results), use.names = TRUE, fill = TRUE
  )
  
  return(data_element)
}



..compute_risk_stratification_logrank_test <- function(x, selected_groups = NULL) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  group <- NULL
  
  # Select indicated risk groups and make sure they are both present.
  if (!is.null(selected_groups)) x <- x[group %in% selected_groups]
  
  # Determine the number of groups.
  n_groups <- data.table::uniqueN(x$group)
  
  # Check that 2 (or more risk groups) are present.
  if (n_groups < 2L) return(NULL)
  
  # Determine chi-square of log-rank test
  chi_sq <- tryCatch(
    survival::survdiff(
      survival::Surv(time = outcome_time, event = outcome_event) ~ group,
      data = x,
      subset = NULL,
      na.action = "na.omit"
    )$chisq,
    error = identity
  )
  
  # Check if the test statistic could be computed. Causes could be lack of
  # events, no events beyond the first time point, etc.
  if (inherits(chi_sq, "error")) return(NULL)
  
  # Derive  p-value
  p_value  <- stats::pchisq(
    q = chi_sq,
    df = n_groups - 1L,
    lower.tail = FALSE
  )
  
  # In case all groups should be compared, add a place-holder name.
  if (is.null(selected_groups)) {
    group_1 <- group_2 <- "all"
    
  } else {
    group_1 <- selected_groups[1L]
    group_2 <- selected_groups[2L]
  }
  
  # Set data.
  test_data <- data.table::data.table(
    "group_1" = group_1,
    "group_2" = group_2,
    "p_value" = p_value
  )
  
  return(as.list(test_data))
}



.compute_risk_stratification_hazard_ratio_test <- function(x, confidence_level) {
  # Suppress NOTES due to non-standard evaluation in data.table
  p_value <- NULL
  
  # Create new element with strata based on x.
  data_element <- methods::new(
    "familiarDataElementRiskHazardRatio",
    x,
    grouping_column = c(setdiff(
      x@grouping_column,
      get_non_feature_columns(x = "survival")
    )),
    value_column = c("p_value", "p_value_adjusted")
  )
  
  # Isolate data.
  data <- .as_data_table(x)
  data$group <- droplevels(data$group)
  
  # Use the first group as reference for ordinal variables.
  reference_groups <- levels(data$group)[1L]
  
  # Compute hazard test results over all risk groups.
  hr_results <- data[
    ,
    dmapply(
      ..compute_risk_stratification_hazard_ratio_test,
      reference_group = reference_groups,
      MoreArgs = list(
        "x" = .SD,
        "confidence_level" = confidence_level
      )
    ),
    by = c(data_element@grouping_column),
    .SDcols = c("outcome_time", "outcome_event", "group")
  ]
  
  # Compute multiple-testing corrected p-value.
  hr_results[
    ,
    "p_value_adjusted" := stats::p.adjust(p_value, method = "holm"),
    by = c(data_element@grouping_column)
  ]
  
  # Store to data element.
  data_element@data <- hr_results
  
  return(data_element)
}



..compute_risk_stratification_hazard_ratio_test <- function(
    x,
    reference_group,
    confidence_level
) {
  
  # Determine the number of groups.
  n_groups <- data.table::uniqueN(x$group)
  
  # Check that 2 (or more risk groups) are present.
  if (n_groups < 2L) return(NULL)
  
  # Check if the reference group is present.
  if (!any(x$group == reference_group)) return(NULL)
  
  # Make local copy of x.
  x <- data.table::copy(x)
  
  # Force the group column to be categorical.
  if (is.factor(x$group)) {
    x$group <- droplevels(x$group)
    
  } else {
    x$group <- as.factor(x$group)
  }
  
  # Reset the reference risk group
  if (!is.ordered(x$group)) {
    x$group <- stats::relevel(
      x$group,
      ref = reference_group
    )
  }
  
  # Create Cox proportional hazards model.
  model <- suppressWarnings(tryCatch(
    survival::coxph(
      survival::Surv(time = outcome_time, event = outcome_event) ~ group,
      data = x
    ),
    error = identity
  ))
  
  # Check if the Cox PH model could not be computed. Causes could be lack of
  # events, no events beyond the first time point, etc.
  if (inherits(model, "error")) return(NULL)
  
  # Extract summary information concerning hazard ratios
  summary_info <- summary(model, confint = confidence_level)
  
  # Fill test data.
  test_data <- data.table::data.table(
    "reference_group" = reference_group,
    "group" = levels(x$group)[-1L],
    "hazard_ratio" = summary_info$conf.int[, 1L],
    "ci_low" = summary_info$conf.int[, 3L],
    "ci_up" = summary_info$conf.int[, 4L],
    "p_value" = summary_info$coefficients[, 5L]
  )
  
  # Return test results
  return(as.list(test_data))
}



# export_risk_stratification_data (generic) ------------------------------------

#'@title Extract and export sample risk group stratification and associated
#'  tests.
#'
#'@description Extract and export sample risk group stratification and
#'  associated tests for data in a familiarCollection.
#'
#'@param export_strata Flag that determines whether the raw data or strata are
#'  exported.
#'@param time_range Time range for which strata should be created. If `NULL`,
#'  the full time range is used.
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams extract_risk_stratification_data
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Three tables are exported in a list:
#'
#'  * `data`: Contains the assigned risk group for a given sample, along with
#'  its reported survival time and censoring status.
#'
#'  * `hr_ratio`: Contains the hazard ratio between different risk groups.
#'
#'  * `logrank`: Contains the results from the logrank test between different
#'  risk groups.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_risk_stratification_data
#'@md
#'@rdname export_risk_stratification_data-methods
setGeneric(
  "export_risk_stratification_data",
  function(
    object,
    dir_path = NULL,
    export_strata = TRUE,
    time_range = NULL,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_risk_stratification_data")
  }
)



# export_risk_stratification_data (collection) ---------------------------------

#'@rdname export_risk_stratification_data-methods
setMethod(
  "export_risk_stratification_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    export_strata = TRUE,
    time_range = NULL,
    export_collection = FALSE,
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    if (!is.null(time_range)) {
      # Check that time_range is valid.
      .check_argument_length(
        time_range, 
        var_name = "time_range",
        min = 2L,
        max = 2L
      )
      
      sapply(
        time_range,
        .check_number_in_valid_range,
        var_name = "time_range",
        range = c(0.0, Inf)
      )
    }
    
    if (export_strata) {
      # Compute kaplan-meier curves.
      strata_data <- lapply(
        object@km_data,
        .compute_risk_stratification_curves,
        time_range = time_range
      )
      
      # Determine hazard ratio and logrank tests. We do this here because the
      # data needs to be aggregated.
      test_data <- lapply(
        object@km_data,
        .compute_risk_stratification_tests,
        time_range = time_range
      )
      
      # Export raw data.
      raw_data <- .export(
        x = object,
        data_slot = "km_data",
        dir_path = dir_path,
        aggregate_results = TRUE,
        type = "stratification",
        subtype = "data"
      )
      
      # Export strata.
      strata_data <- .export(
        x = object,
        data_elements = strata_data,
        dir_path = dir_path,
        aggregate_results = TRUE,
        type = "stratification",
        subtype = "strata"
      )
      
      # Export logrank data.
      logrank_data <- .export(
        x = object,
        data_elements = test_data,
        dir_path = dir_path,
        aggregate_results = TRUE,
        object_class = "familiarDataElementRiskLogrank",
        type = "stratification",
        subtype = "logrank"
      )
      
      # Export hazard ratio data.
      hazard_ratio_data <- .export(
        x = object,
        data_elements = test_data,
        dir_path = dir_path,
        aggregate_results = TRUE,
        object_class = "familiarDataElementRiskHazardRatio",
        type = "stratification",
        subtype = "hazard_ratio"
      )
      
      data_list <- list(
        "data" = raw_data,
        "strata" = strata_data,
        "logrank" = logrank_data,
        "hazard_ratio_data" = hazard_ratio_data
      )
      
      if (export_collection) {
        data_list <- c(
          data_list,
          list("collection" = object)
        )
      } 
      
      return(data_list)
      
    } else {
      return(.export(
        x = object,
        data_slot = "km_data",
        dir_path = dir_path,
        aggregate_results = TRUE,
        type = "stratification",
        subtype = "data",
        export_collection = export_collection
      ))
    }
  }
)



# export_risk_stratification_data (general) ------------------------------------

#'@rdname export_risk_stratification_data-methods
setMethod(
  "export_risk_stratification_data",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    export_strata = TRUE,
    time_range = NULL,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "risk_stratification_data"
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_risk_stratification_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "export_strata" = export_strata,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- TRUE

results <- familiar:::test_export_specific(
  export_function = familiar:::export_shap,
  outcome_type_available = "multinomial",
  data_element = "shap",
  create_novelty_detector = FALSE,
  debug = debug_flag,
  export_args = list(
    feature_x = "feature_1",
    feature_y = "feature_2a"
  )
)



# Default
results <- familiar:::test_export_specific(
  export_function = familiar:::export_shap,
  outcome_type_available = "multinomial",
  data_element = "shap",
  create_novelty_detector = FALSE,
  debug = debug_flag
)

.get_palette <- function(
    x = NULL,
    palette_type,
    n = NULL, 
    invert = FALSE,
    use_alternative = FALSE,
    diverge_to_white = TRUE
) {
  
  # Check whether the provided palette is known, a set of colours, or a default.
  if (is.null(x)) {
    colours <- .get_default_palette(
      n = n,
      palette_type = palette_type,
      invert = invert,
      use_alternative = use_alternative,
      diverge_to_white = diverge_to_white
    )
    
  } else if (all(is.character(x))) {
    if (length(x) > 1L) {
      # Check that all elements are colours.
      valid_colours <- sapply(x, .is_colour)

      if (!all(valid_colours)) {
        ..error(paste0(
          "The following palette colours could not be interpreted: ",
          paste_s(x[!valid_colours]),
          " . A valid colour is either a hexadecimal string (e.g. \"#4F94CD\"), ",
          "a colour specified in grDevices::colors() (e.g. \"steelblue3\"), ",
          "or \"transparant\". Alternatively, a palette can be specified by name."
        ))
      }

      colours <- x
      
    } else if (length(x) == 1L) {
      if (palette_type %in% c("divergent", "sequential") && is.null(n)) {
        # Set n if unspecified.
        n <- 31L
        
      } else if (palette_type == "qualitative" && is.null(n)) {
        # In case of qualitative palettes, n should be set explicitly by familiar.
        ..error_reached_unreachable_code("Qualitative palettes requires the number of discrete cases n.")
      }
      
      # Obtain colours from a predefined palette.
      colours <- .palette_to_colour(
        x = x,
        n = n
      )
    }
    
  } else {
    ..error(paste0(
      "The requested palette are neither colours nor a palette: ",
      paste_s(x)
    ))
  }

  return(colours)
}



.is_colour <- function(x) {
  return(
    x %in% grDevices::colors() ||
      x == "transparant" ||
      grepl(pattern = "^#(\\d|[a-f]){6,8}$", x, ignore.case = TRUE)
  )
}



.palette_to_colour <- function(x, n) {
  # Determine if the string ends with _, _r or _rev.
  invert_colours <- grepl(pattern = "_$|_r$|_rev$", x, ignore.case = TRUE)

  # Strip from x
  x <- gsub(pattern = "_$|_r$|_rev$", replacement = "", x)

  colours <- NULL
  # Try paletteer.
  if (rlang::is_installed("paletteer")) {
    # Continuous palette
    colours <- tryCatch(
      paletteer::paletteer_c(
        palette = x,
        n = n
      ),
      error = function(err) (NULL)
    )
    
    # Dynamic palette
    if (is.null(colours)) {
      colours <- tryCatch(
        paletteer::paletteer_dynamic(
          palette = x,
          n = n
        ),
        error = function(err) (NULL)
      )
    }
    
    # Discrete palette
    if (is.null(colours)) {
      colours <- tryCatch(
        paletteer::paletteer_d(
          palette = x,
          n = n
        ),
        error = function(err) (NULL)
      )
    }
  }
  
  # Try grDevices::palette (requires R version >= 4.0.0)
  if (is.null(colours)) {
    colours <- tryCatch(
      grDevices::palette.colors(n = n, palette = x),
      error = function(err) (NULL)
    )
  }

  # Try grDevices::hcl.colors (requires R version >= 3.6.0)
  if (is.null(colours)) {
    colours <- tryCatch(
      grDevices::hcl.colors(n = n, palette = x),
      error = function(err) (NULL)
    )
  }
  
  
  # Palettes that are always available.
  if (is.null(colours)) {
    if (x == "default") {
      colours <- grDevices::palette()
    } else if (x == "rainbow") {
      colours <- grDevices::rainbow(n = n)
    } else if (x == "heat.colors") {
      colours <- grDevices::heat.colors(n = n)
    } else if (x == "terrain.colors") {
      colours <- grDevices::terrain.colors(n = n)
    } else if (x == "topo.colors") {
      colours <- grDevices::topo.colors(n = n)
    } else if (x == "cm.colors") {
      colours <- grDevices::cm.colors(n = n)
    }
  }

  if (is.null(colours)) {
    ..error(paste0(
      "The palette was not recognised: ", x,
      ". Please check the spelling. Note that some options may not be available prior ",
      "to R 4.0.0 (grDevices::palette.pals(), and R 3.6.0 (grDevices::hcl.pals()))."
    ))
  }

  if (invert_colours) colours <- rev(colours)

  return(colours)
}


.get_default_palette <- function(
    n,
    palette_type,
    invert,
    use_alternative = FALSE,
    diverge_to_white = TRUE
) {
  
  .check_parameter_value_is_valid(
    x = palette_type, var_name = "palette_type",
    values = c("qualitative", "sequential", "divergent")
  )

  if (palette_type == "qualitative") {
    # Palette derived from:
    #
    # colorspace::diverge_hcl(n = 21, h = c(245, 35), c = c(130, 85), l = c(65,
    # 80), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
    #
    # colorspace::diverge_hcl(n = 21, h = c(185, 10), c = c(130, 85), l = c(60,
    # 80), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
    #
    # colorspace::diverge_hcl(n = 21, h = c(65, 125), c = c(100, 85), l = c(70,
    # 80), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
    #
    # colorspace::diverge_hcl(n = 21, h = c(210, 320), c = c(60, 85), l = c(60,
    # 80), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
    #
    # colorspace::diverge_hcl(n = 21, h = c(100, 0), c = c(50, 85), l = c(70,
    # 80), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
    
    palette_10 <- c(
      "#00A8FF",
      "#F77D00",
      "#00BCAB",
      "#FE535E",
      "#D4A600",
      "#40C41C",
      "#00A2B7",
      "#C973B6",
      "#9FB368",
      "#E495A5"
    )
    
    palette_20 <- c(
      "#00A8FF",
      "#93BDF0",
      "#F77D00",
      "#E9AD8D",
      "#00BCAB",
      "#96C9C5",
      "#FE535E",
      "#DFB6B7",
      "#D4A600",
      "#CEC2AC",
      "#40C41C",
      "#B2C9B0",
      "#00A2B7",
      "#8AC4D1",
      "#C973B6",
      "#DCADD0",
      "#9FB368",
      "#B3C680",
      "#E495A5",
      "#ECB1BC"
    )
    
    if (!use_alternative) {
      if (n <= 10L) {
        colours <- palette_10[1L:n]
        
      } else if (n <= 20L) {
        colours <- palette_20[1L:n]
        
      } else {
        ..error(paste0(
          "The required number (", n, ") of discrete colors is too large for the ",
          "default qualitative score (max 20). "
        ))
      }
    } else {
      # Alternative colour schemes were the blue and orange colours come last.
      # This is to avoid confusion with other gradients that may be used in the
      # plot.
      if (n <= 10L) {
        colours <- rev(palette_10)[1L:n]
        
      } else if (n <= 20L) {
        colours <- rev(palette_20)[1L:n]
        
      } else {
        ..error(paste0(
          "The required number (", n, ") of discrete colors is too large for ",
          "the default qualitative score (max 20). "
        ))
      }
    }
  } else if (palette_type == "sequential") {
    if (!use_alternative) {
      # A palette with the same hue (blue) as the first color of the qualitative
      # palette. Based on the blue part of colorspace::diverge_hcl(n = 21, h =
      # c(245, 35), c = c(130, 85), l = c(65, 80), power = c(1, 1.3), gamma =
      # NULL, fixup = TRUE)
      colours <- c(
        "#C6C6C6", "#BCC5D2", "#B0C3DD", "#A3C0E7", "#93BDF0", "#81BAF8",
        "#6AB7FF", "#49B3FF", "#00B0FF", "#00ACFF", "#00A8FF"
      )
      
    } else {
      # Alternative reddish colour scheme that avoids the use of blues and
      # orange tones that may have been used as a primary palette. Based on the
      # red part of colorspace::diverge_hcl(n = 21, h = c(185, 10), c = c(130,
      # 85), l = c(60, 80), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
      colours <- c(
        "#C6C6C6", "#D4BFBF", "#DFB6B7", "#E8ACAE", "#EFA1A4", "#F5969A",
        "#F98A8F", "#FC7E84", "#FE7178", "#FE636C", "#FE535E"
      )
    }

    if (invert) colours <- rev(colours)
    
  } else if (palette_type == "divergent") {
    if (!use_alternative) {
      # A palette with the same hues (blue and orange) as the first two colors
      # of the qualitative palette.
      if (diverge_to_white) {
        # Bright center.
        #
        # colorspace::diverge_hcl(n = 21, h = c(245, 35), c = c(130, 85), l =
        # c(65, 80), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
        colours <- c(
          "#00A8FF", "#00ACFF", "#00B0FF", "#49B3FF", "#6AB7FF", "#81BAF8",
          "#93BDF0", "#A3C0E7", "#B0C3DD", "#BCC5D2",
          "#C6C6C6",
          "#D2C1BA", "#DBBBAC", "#E3B49D", "#E9AD8D", "#EEA57C", "#F29D68",
          "#F49650", "#F68E2F", "#F78600", "#F77D00"
        )
        
      } else {
        # Dark centre.
        #
        # colorspace::diverge_hcl(n = 21, h = c(245, 35), c = c(130, 85), l =
        # c(65, 30), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
        colours <- c(
          "#00A8FF", "#009BFF", "#008FEF", "#0083D8", "#0078C1", "#056DAB",
          "#266495", "#345B81", "#3C526D", "#424C59",
          "#474747",
          "#584740", "#694A39", "#794E32", "#8A522A", "#9A581E", "#AC5E07",
          "#BE6500", "#D06D00", "#E37500" ,"#F77D00"
        )
      }
      
    } else {
      # A palette based on the same hues the first two colours of the
      # alternative qualitative palette.
      if (diverge_to_white) {
        # Bright centre.
        #
        # colorspace::diverge_hcl(n = 21, h = c(185, 10), c = c(130, 85), l =
        # c(60, 80), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
        colours <- c(
          "#00BCAB", "#00BFAF", "#00C2B4", "#00C4B7", "#00C6BB", "#00C8BE",
          "#40C9C1", "#74C9C3", "#96C9C5", "#B0C9C6",
          "#C6C6C6",
          "#D4BFBF", "#DFB6B7", "#E8ACAE", "#EFA1A4", "#F5969A", "#F98A8F",
          "#FC7E84", "#FE7178", "#FE636C", "#FE535E"
        )
        
      } else {
        # Dark centre.
        #
        # colorspace::diverge_hcl(n = 21, h = c(185, 10), c = c(130, 85), l =
        # c(60, 30), power = c(1, 1.3), gamma = NULL, fixup = TRUE)
        colours <- c(
          "#00BCAB", "#00AE9F", "#00A093", "#009387", "#00877C", "#007A71",
          "#006F67", "#00645E", "#065955", "#334F4D",
          "#474747",
          "#5A4546", "#6D4446", "#7E4547", "#904649", "#A1474C", "#B3494F",
          "#C54B52", "#D74D56", "#EB505A", "#FE535E"
        )
      }
    }

    if (invert) colours <- rev(colours)
  }

  return(colours)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarDataElementDecisionCurve object --------------------------------------

setClass(
  "familiarDataElementDecisionCurve",
  contains = "familiarDataElement",
  prototype = methods::prototype(
    value_column = "net_benefit",
    grouping_column = "threshold_probability"
  )
)

# exract_decision_curve_data (generic) -----------------------------------------

#'@title Internal function to extract decision curve analysis data.
#'
#'@description Computes decision curve analysis data from a `familiarEnsemble`
#'  object. Calibration tests are performed based on expected (predicted) and
#'  observed outcomes. For all outcomes, calibration-at-the-large and
#'  calibration slopes are determined. Furthermore, for all but survival
#'  outcomes, a repeated, randomised grouping Hosmer-Lemeshow test is performed.
#'  For survival outcomes, the Nam-D'Agostino and Greenwood-Nam-D'Agostino tests
#'  are performed.
#'
#'@inheritParams .extract_data
#'
#'@return A list with data.tables containing calibration test information for
#'  the ensemble model.
#'@md
#'@keywords internal
setGeneric(
  "extract_decision_curve_data",
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_decision_curve_data")
  }
)


# extract_decision_curve_data (familiarEnsemble) -------------------------------
setMethod(
  "extract_decision_curve_data",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    
    # Decision curve analysis is only available for categorical and survival
    # outcomes.
    if (!object@outcome_type %in% c("binomial", "multinomial", "survival")) return(NULL)
    
    # Message extraction start
    logger_message(
      paste0("Computing data for decision curve analysis."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (is.waive(evaluation_times)) evaluation_times <- object@settings$eval_times
    if (is.waive(ensemble_method)) ensemble_method <- object@settings$ensemble_method
    if (is.waive(confidence_level)) confidence_level <- object@settings$confidence_level
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- object@settings$bootstrap_ci_method
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "decision_curve_analyis"
    )
    
    # Generate a prototype data element.
    proto_data_element <- .create_extract_decision_curve_object(
      object = object,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    # Generate elements to send to dispatch.
    dca_data <- extract_dispatcher(
      FUN = .extract_decision_curve_data,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(dca_data)
  }
)



# extract_decision_curve_data (prediction table) -------------------------------
setMethod(
  "extract_decision_curve_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    if (is_empty(object)) return(NULL)
    
    # Decision curve analysis is only available for categorical and survival
    # outcomes.
    if (!(
      is(object, "predictionTableSurvivalProbability") || 
      is(object, "predictionTableClassification")
    )) {
      ..warning_no_data_extraction_from_prediction_table("decision curves")
      
      return(NULL)
    } 
    # Reference labels should be present.
    if (!.has_reference_data(object)) {
      ..warning_prediction_table_lacks_reference("decision curves")
      return(NULL)
    }
    
    # Message extraction start
    logger_message(
      paste0("Computing data for decision curve analysis."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (is.waive(evaluation_times) && methods::.hasSlot(object, "time")) {
      evaluation_times <- object@time
    }
    if (is.waive(ensemble_method)) {
      ensemble_method <- "median"
      if (methods::.hasSlot(object, "ensemble_method")) ensemble_method <- object@ensemble_method
    } 
    
    # Default Values.
    if (is.waive(detail_level)) detail_level <- "ensemble"
    if (is.waive(estimation_type)) estimation_type <- "bootstrap_confidence_interval" 
    if (is.waive(confidence_level)) confidence_level <- 0.95
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- "bc"
    if (is.waive(aggregate_results)) aggregate_results <- TRUE
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "decision_curve_analyis"
    )
    
    # Copy object to prevent changing the provided object by reference.
    object <- .copy(object)
    
    # Generate a prototype data element.
    proto_data_element <- .create_extract_decision_curve_object(
      object = object,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    if (is(object, "predictionTableSurvivalProbability")) {
      proto_data_element <- add_data_element_identifier(
        x = proto_data_element,
        evaluation_time = object@time
      )[[1L]]
    }
    
    # Generate elements to send to dispatch.
    dca_data <- extract_dispatcher(
      FUN = .extract_decision_curve_data,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(dca_data)
  }
)



.create_extract_decision_curve_object <- function(
    object,
    ensemble_method,
    evaluation_times,
    detail_level,
    estimation_type,
    confidence_level,
    bootstrap_ci_method
) {
  
  # Check evaluation_times argument
  if (object@outcome_type %in% c("survival")) {
    sapply(
      evaluation_times,
      .check_number_in_valid_range,
      var_name = "evaluation_times",
      range = c(0.0, Inf),
      closed = c(FALSE, TRUE)
    )
  }
  
  # Check confidence_level input argument.
  .check_number_in_valid_range(
    x = confidence_level, 
    var_name = "confidence_level",
    range = c(0.0, 1.0),
    closed = c(FALSE, FALSE)
  )
  
  # Check ensemble_method argument.
  .check_parameter_value_is_valid(
    x = ensemble_method,
    var_name = "ensemble_method",
    values = .get_available_ensemble_prediction_methods()
  )
  
  # bootstrap_ci_method
  .check_parameter_value_is_valid(
    x = bootstrap_ci_method,
    var_name = "bootstrap_ci_method",
    values = .get_available_bootstrap_confidence_interval_methods()
  )
  
  # Check the level detail.
  detail_level <- .parse_detail_level(
    x = detail_level,
    object = object,
    default = "hybrid",
    data_element = "decision_curve_analyis"
  )
  
  # Check the estimation type.
  estimation_type <- .parse_estimation_type(
    x = estimation_type,
    object = object,
    default = "bootstrap_confidence_interval",
    data_element = "decision_curve_analyis",
    detail_level = detail_level,
    has_internal_bootstrap = TRUE
  )

  # Generate a prototype data element.
  proto_data_element <- new(
    "familiarDataElementDecisionCurve",
    detail_level = detail_level,
    estimation_type = estimation_type,
    confidence_level = confidence_level,
    bootstrap_ci_method = bootstrap_ci_method
  )
  
  return(proto_data_element)
}



# .extract_decision_curve_data (generic) --------------------------------------
setGeneric(
  ".extract_decision_curve_data",
  function(object, ...) standardGeneric(".extract_decision_curve_data")
)


# .extract_decision_curve_data (character) -------------------------------------
setMethod(
  ".extract_decision_curve_data",
  signature(object = "character"),
  function(
    object,
    ...
  ){
    # Ensure that the object is loaded
    object <- load_familiar_object(object)
    
    return(.extract_decision_curve_data(object = object, ...))
  }
)



# .extract_decision_curve_data (familiarModelUnion) ----------------------------
setMethod(
  ".extract_decision_curve_data",
  signature(object = "familiarModelUnion"),
  function(
    object,
    data,
    proto_data_element,
    evaluation_times = NULL,
    cl,
    ensemble_method,
    is_pre_processed,
    ...
  ) {
    # Add model name.
    proto_data_element <- add_model_name(proto_data_element, object = object)
    
    if (length(evaluation_times) > 0L && object@outcome_type == "survival") {
      # Add evaluation time as a identifier to the data element.
      data_elements <- add_data_element_identifier(
        x = proto_data_element,
        evaluation_time = evaluation_times
      )
      
    } else {
      data_elements <- list(proto_data_element)
    }
    
    if (object@outcome_type == "survival") {
      dca_data <- lapply(
        data_elements,
        function(data_element, object, data, ensemble_method, is_pre_processed, ...) {
          prediction_table <- .predict(
            object = object,
            data = data,
            time = data_element@identifiers$evaluation_time,
            ensemble_method = ensemble_method,
            is_pre_processed = is_pre_processed,
            type = "survival_probability"
          )
          
          return(.extract_decision_curve_data(
            object = prediction_table,
            proto_data_element = data_element,
            ...
          ))
        },
        object = object,
        data = data,
        ensemble_method = ensemble_method,
        is_pre_processed = is_pre_processed,
        ...
      )
      
    } else if (object@outcome_type %in% c("binomial", "multinomial")) {
      prediction_table <- .predict(
        object = object,
        data = data,
        ensemble_method = ensemble_method,
        is_pre_processed = is_pre_processed
      )
      
      dca_data <- mapply(
        .extract_decision_curve_data,
        proto_data_element = data_elements,
        MoreArgs = c(
          list("object" = prediction_table),
          list(...)
        ),
        SIMPLIFY = FALSE
      )
      
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }
    
    return(dca_data)
  }
)



# .extract_decision_curve_data (classification) --------------------------------
setMethod(
  ".extract_decision_curve_data",
  signature(object = "predictionTableClassification"),
  function(
    object,
    proto_data_element,
    aggregate_results,
    cl = NULL,
    progress_bar = FALSE,
    verbose = FALSE,
    message_indent = 0L,
    ...
  ) {
    # Check if any predictions are valid.
    if (!all_predictions_valid(object)) return(NULL)
    
    # Remove data with missing outcomes.
    object <- filter_missing_outcome(object)
    if (is_empty(object)) return(NULL)
    
    data <- .as_data_table(object)
    if (nrow(data) <= 1L) return(NULL)
    
    # Determine class levels
    outcome_class_levels <- get_outcome_class_levels(object)
    
    # Select only one outcome class for binomial outcomes.
    if (object@outcome_type == "binomial") outcome_class_levels <- outcome_class_levels[2L]
    
    # Add positive class as an identifier.
    data_elements <- add_data_element_identifier(
      x = proto_data_element,
      positive_class = outcome_class_levels
    )
    
    # Add bootstrap data.
    bootstrap_data <- add_data_element_bootstrap(x = data_elements, ...)
    
    if (length(bootstrap_data) > 1L && progress_bar) {
      logger_message(
        paste0(
          "Computing decision curves for the ",
          paste_s(outcome_class_levels),
          ifelse(length(outcome_class_levels) == 1L, " class.", " classes.")
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    # Set test probabilities
    threshold_probabilities <- seq(
      from = 0.000, 
      to = 1.000,
      by = 0.005
    )
    
    # Iterate over elements.
    data_elements <- fam_mapply(
      cl = cl,
      assign = NULL,
      FUN = .compute_dca_data_categorical_model,
      data_element = bootstrap_data$data_element,
      bootstrap = bootstrap_data$bootstrap,
      bootstrap_seed = bootstrap_data$seed,
      MoreArgs = list(
        "data" = data,
        "threshold_probabilities" = threshold_probabilities
      ),
      progress_bar = progress_bar && verbose,
      chopchop = TRUE
    )
    
    # Merge data elements
    data_elements <- merge_data_elements(data_elements)
    
    if (aggregate_results) data_elements <- .compute_data_element_estimates(x = data_elements)
    
    return(data_elements)
  }
)



# .extract_decision_curve_data (survival probability) --------------------------
setMethod(
  ".extract_decision_curve_data",
  signature(object = "predictionTableSurvivalProbability"),
  function(
    object,
    proto_data_element,
    aggregate_results,
    cl = NULL,
    progress_bar = FALSE,
    verbose = FALSE,
    message_indent = 0L,
    ...
  ) {

    # Check if any predictions are valid.
    if (!all_predictions_valid(object)) return(NULL)
    
    # Remove data with missing outcomes.
    object <- filter_missing_outcome(object)
    if (is_empty(object)) return(NULL)
    
    data <- .as_data_table(object)
    if (nrow(data) <= 1L) return(NULL)
    
    # Add bootstrap data.
    bootstrap_data <- add_data_element_bootstrap(x = proto_data_element, ...)
    
    # Message the user concerning the time at which the decision curves are
    # computed. This is only relevant for survival analysis, where survival
    # probability is time depend.
    if (length(bootstrap_data) > 0L && progress_bar) {
      message_str <- "Computing decision curves"
      if (!is.null(proto_data_element@identifiers$evaluation_time)) {
        message_str <- c(
          message_str,
          " at time ", proto_data_element@identifiers$evaluation_time, "."
        )
        
      } else {
        message_str <- c(message_str, ".")
      }
      
      logger_message(
        paste0(message_str, collapse = ""),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    # Set test probabilities
    threshold_probabilities <- seq(
      from = 0.000,
      to = 1.000,
      by = 0.005
    )
    
    # Iterate over elements.
    data_elements <- fam_mapply(
      cl = cl,
      assign = NULL,
      FUN = .compute_dca_data_survival_model,
      data_element = bootstrap_data$data_element,
      bootstrap = bootstrap_data$bootstrap,
      bootstrap_seed = bootstrap_data$seed,
      MoreArgs = list(
        "data" = data,
        "threshold_probabilities" = threshold_probabilities
      ),
      progress_bar = progress_bar && verbose,
      chopchop = TRUE
    )
    
    # Merge data elements
    data_elements <- merge_data_elements(data_elements)
    
    if (aggregate_results) data_elements <- .compute_data_element_estimates(x = data_elements)
    
    return(data_elements)
  }
)


# .extract_decision_curve_data (NULL) ------------------------------------------
setMethod(
  ".extract_decision_curve_data",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)


.compute_dca_data_categorical_model <- function(
    data_element,
    data,
    threshold_probabilities,
    bootstrap,
    bootstrap_seed
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- probability <- is_positive <- NULL
  
  # Get the positive class.
  positive_class <- data_element@identifiers$positive_class
  
  # Bootstrap the data.
  if (bootstrap) {
    data <- get_bootstrap_sample(
      data = data,
      seed = bootstrap_seed
    )
  } 
  
  # Make a local copy
  data <- data.table::copy(data)
  data.table::setnames(
    x = data,
    old = positive_class,
    new = "probability"
  )
  
  # Determine positive output.
  data[, "is_positive" := outcome == positive_class]
  
  # Keep only probability and is_positive
  data <- data[, c("probability", "is_positive")]
  
  # Order by inverse probability.
  data <- data[order(-probability)]
  
  # Determine the number of samples.
  n <- nrow(data)
  
  # Compute intervention data.
  # Determine maximum number of true and false positives.
  n_max_true_positive <- sum(data$is_positive)
  n_max_false_positive <- sum(!data$is_positive)
  
  # Compute benefit for the situation an intervention always happens.
  intervention_net_benefit <- n_max_true_positive / n - 
    n_max_false_positive / n * (threshold_probabilities / (1.0 - threshold_probabilities))
  
  # Copy data element for intervention.
  intervention_data_element <- data_element
  
  # Set the data attribute.
  intervention_data_element@data <- data.table::data.table(
    "threshold_probability" = threshold_probabilities,
    "net_benefit" = intervention_net_benefit
  )
  
  # Set the curve type identifier.
  intervention_data_element <- add_data_element_identifier(
    x = intervention_data_element,
    curve_type = "intervention_all"
  )
  
  # Determine the number of true and false positives to determine the net
  # benefit of the model.
  data[, ":="(
    "n_true_positive" = cumsum(is_positive),
    "n_false_positive" = cumsum(!is_positive)
  )]
  
  # Compute benefit for the model.
  model_net_benefit <- ..compute_dca_data_net_benefit(data, threshold_probabilities)
  
  if (is.null(model_net_benefit)) return(NULL)
  
  # Set the data attribute
  data_element@data <- data.table::data.table(
    "threshold_probability" = threshold_probabilities,
    "net_benefit" = model_net_benefit
  )
  
  # Set the curve type identifier.
  data_element <- add_data_element_identifier(
    x = data_element,
    curve_type = "model"
  )
  
  return(c(
    data_element,
    intervention_data_element
  ))
}



.compute_dca_data_survival_model <- function(
    data_element,
    data,
    threshold_probabilities,
    bootstrap,
    bootstrap_seed
) {
  
  # Bootstrap the data.
  if (bootstrap) {
    data <- get_bootstrap_sample(
      data = data,
      seed = bootstrap_seed
    )
  }
  
  # Compute benefit for the situation an intervention always happens.
  intervention_net_benefit <- ..compute_dca_data_net_benefit_survival(
    data = data,
    x = threshold_probabilities,
    evaluation_time = data_element@identifiers$evaluation_time,
    intervention = TRUE
  )
  
  # Copy data element for intervention.
  intervention_data_element <- data_element
  
  # Set the data attribute.
  intervention_data_element@data <- data.table::data.table(
    "threshold_probability" = threshold_probabilities,
    "net_benefit" = intervention_net_benefit
  )
  
  # Set the curve type identifier.
  intervention_data_element <- add_data_element_identifier(
    x = intervention_data_element,
    curve_type = "intervention_all"
  )
  
  # Compute benefit for the model.
  model_net_benefit <- ..compute_dca_data_net_benefit_survival(
    data = data,
    x = threshold_probabilities,
    evaluation_time = data_element@identifiers$evaluation_time,
    intervention = FALSE
  )
  
  if (is.null(model_net_benefit)) return(NULL)
  
  # Set the data attribute
  data_element@data <- data.table::data.table(
    "threshold_probability" = threshold_probabilities,
    "net_benefit" = model_net_benefit
  )
  
  # Set the curve type identifier.
  data_element <- add_data_element_identifier(
    x = data_element,
    curve_type = "model"
  )
  
  return(c(
    data_element,
    intervention_data_element
  ))
}



..compute_dca_data_net_benefit <- function(data, x) {
  # Compute net benefit for models.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  n_true_positive <- n_false_positive <- probability <- net_benefit <- NULL
  
  # Determine maximum number of true and false positives.
  n_max_true_positive <- max(data$n_true_positive)
  
  # Determine the number of samples.
  n <- nrow(data)
  
  # Determine net benefit.
  data[, ":="(
    "net_benefit" = n_true_positive / n -
      n_false_positive / n * (probability / (1.0 - probability))
  )]
  
  # Net benefit should be numeric.
  data <- data[is.finite(net_benefit)]
  
  # Check if the data has more than 1 row.
  if (nrow(data) <= 1L) return(NULL)
  
  # If the predicted probability occurs more than once, select the lowest net
  # benefit.
  data <- data[, list("net_benefit" = min(net_benefit)), by = "probability"]

  # Check if the data has more than 1 row.
  if (nrow(data) <= 1L) return(NULL)
  
  # Compute net benefit at the test probabilities.
  net_benefit <- suppressWarnings(stats::approx(
    x = data$probability,
    y = data$net_benefit,
    xout = x,
    yleft = n_max_true_positive / n,
    yright = -Inf,
    method = "linear"
  )$y)
  
  return(net_benefit)
}



..compute_dca_data_net_benefit_survival <- function(
    data,
    x,
    evaluation_time,
    intervention = FALSE
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  predicted_outcome <- outcome_event <- outcome_time <- NULL
  death <- censored <- n <- NULL
  
  # Prepare net benefit.
  net_benefit <- numeric(length(x))
  
  # Determine the total group size.
  n_group_size <- nrow(data)
  
  # Define probability thresholds that will be evaluated. When the intervention
  # curve is required, only one threshold (0.0) is used.
  if (intervention) {
    p_threshold <- 0.0
    
  } else {
    p_threshold <- x
  }
  
  # We want to avoid running too many computations. Therefore, we will only
  # compute the number of true positives and false positives when the group size
  # changes.
  previous_group_size <- n_group_size + 1L
  for (ii in seq_along(p_threshold)) {
    # Select the group of patients
    surv_group <- data.table::copy(data[predicted_outcome >= x[ii]])
    
    # Get the total group size of the group where predicted survival probability
    # exceeds the threshold probability.
    n_surv_group <- nrow(surv_group)
    
    if (n_surv_group == previous_group_size) {
      # True positive and false positive numbers did not change. This is always
      # skipped in the first iteration, because the previous group size is set
      # to a value that lies beyond the possible survival group size.
      n_true_positive <- n_true_positive
      n_false_positive <- n_false_positive
      
    } else if (n_surv_group == 0L) {
      # There are no true and false positives, because there are no positives
      # with survival probability greater than the threshold.
      n_true_positive <- n_false_positive <- 0L
      
    } else {
      # Create the basic part of the Kaplan-Meier data by summing the number of
      # deaths and censored patients at the end of each interval. We limit
      # ourselves to those samples that were censored or had an event prior to
      # the evaluation time. The remaining patients do not affect the survival
      # probability.
      surv_group <- surv_group[
        outcome_time <= evaluation_time,
        list(
          "death" = sum(outcome_event == 1L),
          "censored" = sum(outcome_event == 0L)
        ),
        by = "outcome_time"
      ][order(outcome_time)]
      
      if (nrow(surv_group) > 0L) {
        # Add group sizes at the start of each interval.
        surv_group[
          , "n" := n_surv_group - data.table::shift(cumsum(death + censored), n = 1L, fill = 0L, type = "lag")
        ]
        
        # Compute the probability of survival in the interval
        surv_group[, "survival_in_interval" := (n - death) / n]
        
        # Compute survival probability.
        survival_probability <- prod(surv_group$survival_in_interval)
        
      } else {
        # All samples survived longer than evaluation_time.
        survival_probability <- 1.0
      }
      
      # Compute the number of true and false positives according to Vickers et
      # al. 2008 (10.1186/1147-6947-8-53)
      n_true_positive <- (1.0 - survival_probability) * n_surv_group
      n_false_positive <- survival_probability * n_surv_group
    }
    
    # Update the previous group size.
    previous_group_size <- n_surv_group
    
    # Compute net benefit.
    if (intervention) {
      net_benefit <- n_true_positive / n_group_size - 
        n_false_positive / n_group_size * (x / (1.0 - x))
      
    } else {
      net_benefit[ii] <- n_true_positive / n_group_size - 
        n_false_positive / n_group_size * (p_threshold[ii] / (1.0 - p_threshold[ii]))
      
      if (p_threshold[ii] == 1.0) net_benefit[ii] <- 0.0
    }
  }
  
  return(net_benefit)
}


# export_decision_curve_analysis_data (generic) --------------------------------

#'@title Extract and export decision curve analysis data.
#'
#'@description Extract and export decision curve analysis data in a
#'  familiarCollection.
#'
#'@inheritParams export_all
#'
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'  
#'  Decision curve analysis data is computed for categorical outcomes, i.e.
#'  binomial and multinomial, as well as survival outcomes.
#'
#'@return A list of data.table (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_decision_curve_analysis_data
#'@md
#'@rdname export_decision_curve_analysis_data-methods
setGeneric(
  "export_decision_curve_analysis_data",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    ...
  ) {
    standardGeneric("export_decision_curve_analysis_data")
  }
)



# export_decision_curve_analysis_data (collection) -----------------------------

#'@rdname export_decision_curve_analysis_data-methods
setMethod(
  "export_decision_curve_analysis_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE, 
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(.export(
      x = object,
      data_slot = "decision_curve_data",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "decision_curve_analysis",
      subtype = "data"
    ))
  }
)



# export_decision_curve_analysis_data (general) --------------------------------

#'@rdname export_decision_curve_analysis_data-methods
setMethod(
  "export_decision_curve_analysis_data",
  signature(object = "ANY"),
  function(
    object, 
    dir_path = NULL, 
    aggregate_results = TRUE, 
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "decision_curve_analyis",
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_decision_curve_analysis_data,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_sample_clustering (generic) ---------------------------------------------

#' @title Plot heatmaps for pairwise similarity between features.
#'
#' @description This method creates a heatmap based on data stored in a
#'   `familiarCollection` object. Features in the heatmap are ordered so that
#'   more similar features appear together.
#'
#' @param dir_path (*optional*) Path to the directory where created performance
#'   plots are saved to. Output is saved in the `feature_similarity`
#'   subdirectory. If `NULL` no figures are saved, but are returned instead.
#' @param gradient_palette (*optional*) Sequential or divergent palette used to
#'   colour the similarity or distance between features in a heatmap. `familiar`
#'   has a default palette. Other palettes are supported by the `paletteer`
#'   package, `grDevices::palette.pals()` (requires R >= 4.0.0),
#'   `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`, `heat.colors`,
#'   `terrain.colors`, `topo.colors` and `cm.colors`, which correspond to the
#'   palettes of the same name in `grDevices`. You may also specify your own
#'   palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param gradient_palette_range (*optional*) Numerical range used to span the
#'   gradient. This should be a range of two values, e.g. `c(0, 1)`. Lower or
#'   upper boundary can be unset by using `NA`. If not set, the full
#'   metric-specific range is used.
#' @param outcome_palette (*optional*) Sequential (`continuous` outcomes) or
#'   qualitative (other outcome types) palette used to show outcome values. This
#'   argument is ignored if the outcome is not shown.
#' @param outcome_palette_range (*optional*) Numerical range used to span the
#'   gradient of numeric (`continuous`) outcome values. This argument is ignored
#'   for other outcome types or if the outcome is not shown.
#' @param outcome_legend_label (*optional*) Label to provide to the legend for
#'   outcome data. If NULL, the legend will not have a name. By default,
#'   `class`, `value` and `event` are used for `binomial` and `multinomial`,
#'   `continuous`, and `survival` outcome types, respectively.
#' @param show_feature_dendrogram (*optional*) Show feature dendrogram around
#'   the main panel. Can be `TRUE`, `FALSE`, `NULL`, or a position, i.e. `top`,
#'   `bottom`, `left` and `right`.
#'
#'   If a position is specified, it should be appropriate with regard to the
#'   `x_axis_by` or `y_axis_by` argument. If `x_axis_by` is `sample` (default),
#'   the only valid positions are `top` (default) and `bottom`. Alternatively,
#'   if `y_axis_by` is `feature`, the only valid positions are `right` (default)
#'   and `left`.
#'
#'   A dendrogram can only be drawn from cluster methods that produce
#'   dendograms, such as `hclust`. A dendogram can for example not be
#'   constructed using the partioning around medioids method (`pam`).
#'
#' @param show_sample_dendrogram (*optional*) Show sample dendrogram around the
#'   main panel. Can be `TRUE`, `FALSE`, `NULL`, or a position, i.e. `top`,
#'   `bottom`, `left` and `right`.
#'
#'   If a position is specified, it should be appropriate with regard to the
#'   `x_axis_by` or `y_axis_by` argument. If `y_axis_by` is `sample` (default),
#'   the only valid positions are `right` (default) and `left`. Alternatively,
#'   if `x_axis_by` is `sample`, the only valid positions are `top` (default)
#'   and `bottom`.
#'
#'   A dendrogram can only be drawn from cluster methods that produce
#'   dendograms, such as `hclust`. A dendogram can for example not be
#'   constructed using the partioning around medioids method (`pam`).
#' @param show_normalised_data (*optional*) Flag that determines whether the
#'   data shown in the main heatmap is normalised using the same settings as
#'   within the analysis (`fixed`; default), using a standardisation method
#'   (`set_normalisation`) that is applied separately to each dataset, or not at
#'   all (`none`), which shows the data at the original scale, albeit with
#'   batch-corrections.
#'
#'   Categorial variables are plotted to span 90% of the entire numerical value
#'   range, i.e. the levels of categorical variables with 2 levels are
#'   represented at 5% and 95% of the range, with 3 levels at 5%, 50%, and 95%,
#'   etc.
#' @param show_outcome (*optional*) Show outcome column(s) or row(s) in the
#'   graph. Can be `TRUE`, `FALSE`, `NULL` or a poistion, i.e. `top`, `bottom`,
#'   `left` and `right`.
#'
#'   If a position is specified, it should be appropriate with regard to the
#'   `x_axis_by` or `y_axis_by` argument. If `y_axis_by` is `sample` (default),
#'   the only valid positions are `left` (default) and `right`. Alternatively,
#'   if `x_axis_by` is `sample`, the only valid positions are `top` (default)
#'   and `bottom`.
#'
#'   The outcome data will be drawn between the main panel and the sample
#'   dendrogram (if any).
#'
#' @param dendrogram_height (*optional*) Height of the dendrogram. The height is
#'   1.5 cm by default. Height is expected to be grid unit (see `grid::unit`),
#'   which also allows for specifying relative heights.
#' @param outcome_height (*optional*) Height of an outcome data column/row. The
#'   height is 0.3 cm by default. Height is expected to be a grid unit (see
#'   `grid::unit`), which also allows for specifying relative heights. In case
#'   of `survival` outcome data with multipe `evaluation_times`, this height is
#'   multiplied by the number of time points.
#' @param evaluation_times (*optional*) Times at which the event status of
#'   time-to-event survival outcomes are determined. Only used for `survival`
#'   outcome. If not specified, the values used when creating the underlying
#'   `familiarData` objects are used.
#' @inheritParams export_feature_similarity
#' @inheritParams export_sample_similarity
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_feature_expression -object -feature_cluster_method -feature_linkage_method -sample_cluster_method -sample_linkage_method
#'
#' @details This function generates area under the ROC curve plots.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, and `data_set`.
#'   By default, the data is split by `vimp_method` and `learner` and `data_set`,
#'   since the number of samples will typically differ between data sets, even
#'   for the same variable importance method and learner.
#'
#'   The `x_axis_by` and `y_axis_by` arguments determine what data are shown
#'   along which axis. Each argument takes one of `feature` and `sample`, and
#'   both arguments should be unique. By default, features are shown along the
#'   x-axis and samples along the y-axis.
#'
#'   Note that similarity is determined based on the underlying data. Hence the
#'   ordering of features may differ between facets, and tick labels are
#'   maintained for each panel.
#'
#'   Labeling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'   
#'   This plot can be created from `dataObject`, or `data.table` objects.
#'   For `data.table`, see \code{\link{as_data_object}} for additional arguments.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_sample_clustering
#' @md
#' @rdname plot_sample_clustering-methods
setGeneric(
  "plot_sample_clustering",
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_limit = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    gradient_palette_range = waiver(),
    outcome_palette = NULL,
    outcome_palette_range = waiver(),
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    outcome_legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 3L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 3L,
    y_breaks = NULL,
    rotate_x_tick_labels = waiver(),
    show_feature_dendrogram = TRUE,
    show_sample_dendrogram = TRUE,
    show_normalised_data = TRUE,
    show_outcome = TRUE,
    dendrogram_height = grid::unit(1.5, "cm"),
    outcome_height = grid::unit(0.3, "cm"),
    evaluation_times = waiver(),
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    verbose = TRUE,
    ...
  ) {
    standardGeneric("plot_sample_clustering")
  }
)



# plot_sample_clustering (general) ---------------------------------------------

#' @rdname plot_sample_clustering-methods
setMethod(
  "plot_sample_clustering",
  signature(object = "ANY"),
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_limit = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    gradient_palette_range = waiver(),
    outcome_palette = NULL,
    outcome_palette_range = waiver(),
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    outcome_legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 3L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 3L,
    y_breaks = NULL,
    rotate_x_tick_labels = waiver(),
    show_feature_dendrogram = TRUE,
    show_sample_dendrogram = TRUE,
    show_normalised_data = TRUE,
    show_outcome = TRUE,
    dendrogram_height = grid::unit(1.5, "cm"),
    outcome_height = grid::unit(0.3, "cm"),
    evaluation_times = waiver(),
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    verbose = TRUE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "feature_expressions",
          "sample_limit" = sample_limit,
          "feature_cluster_method" = feature_cluster_method,
          "feature_linkage_method" = feature_linkage_method,
          "sample_cluster_method" = sample_cluster_method,
          "sample_linkage_method" = sample_linkage_method
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_sample_clustering,
      args = list(
        "object" = object,
        "feature_cluster_method" = feature_cluster_method,
        "feature_linkage_method" = feature_linkage_method,
        "sample_cluster_method" = sample_cluster_method,
        "sample_linkage_method" = sample_linkage_method,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "x_axis_by" = x_axis_by,
        "y_axis_by" = y_axis_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "gradient_palette" = gradient_palette,
        "gradient_palette_range" = gradient_palette_range,
        "outcome_palette" = outcome_palette,
        "outcome_palette_range" = outcome_palette_range,
        "x_label" = x_label,
        "x_label_shared" = x_label_shared,
        "y_label" = y_label,
        "y_label_shared" = y_label_shared,
        "legend_label" = legend_label,
        "outcome_legend_label" = outcome_legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "rotate_x_tick_labels" = rotate_x_tick_labels,
        "show_feature_dendrogram" = show_feature_dendrogram,
        "show_sample_dendrogram" = show_sample_dendrogram,
        "show_normalised_data" = show_normalised_data,
        "show_outcome" = show_outcome,
        "dendrogram_height" = dendrogram_height,
        "outcome_height" = outcome_height,
        "evaluation_times" = evaluation_times,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection,
        "verbose" = verbose
      )
    ))
  }
)



# plot_sample_clustering (collection) ------------------------------------------

#' @rdname plot_sample_clustering-methods
setMethod(
  "plot_sample_clustering",
  signature(object = "familiarCollection"),
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_limit = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    gradient_palette_range = waiver(),
    outcome_palette = NULL,
    outcome_palette_range = waiver(),
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    outcome_legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 3L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 3L,
    y_breaks = NULL,
    rotate_x_tick_labels = waiver(),
    show_feature_dendrogram = TRUE,
    show_sample_dendrogram = TRUE,
    show_normalised_data = TRUE,
    show_outcome = TRUE,
    dendrogram_height = grid::unit(1.5, "cm"),
    outcome_height = grid::unit(0.3, "cm"),
    evaluation_times = waiver(),
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    verbose = TRUE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    .NATURAL <- NULL
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Get feature expression data
    feature_expression <- export_feature_expressions(
      object = object,
      evaluation_time = evaluation_times
    )
    
    # Get feature similarity data.
    feature_similarity <- export_feature_similarity(
      object = object,
      feature_cluster_method = feature_cluster_method,
      feature_linkage_method = feature_linkage_method,
      export_dendrogram = FALSE,
      export_ordered_data = FALSE
    )[[1L]]

    # Get feature similarity data.
    sample_similarity <- export_sample_similarity(
      object = object,
      sample_limit = sample_limit,
      sample_cluster_method = sample_cluster_method,
      sample_linkage_method = sample_linkage_method
    )[[1L]]

    # Check that the data are not empty.
    if (is_empty(feature_expression)) return(NULL)
    if (all(sapply(feature_expression, is_empty))) return(NULL)
    
    feature_expression <- feature_expression[!sapply(feature_expression, is_empty)]

    # Build an identifier table.
    identifier_table <- lapply(
      feature_expression, 
      function(x) {
        return(unique(x@data[, mget(x@grouping_column)]))
      }
    )

    # Combine to table.
    identifier_table <- data.table::rbindlist(
      identifier_table, 
      use.names = TRUE
    )

    # Add row identifiers to make it easier to track the list elements for
    # feature expression.
    identifier_table[, "list_id" := .I]

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = TRUE),
      purpose = "to create sample clustering heatmaps",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Check input arguments ----------------------------------------------------

    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # rotate_x_tick_labels
    if (is.waive(rotate_x_tick_labels)) {
      rotate_x_tick_labels <- FALSE
    }

    # x_label_shared
    if (!is.waive(x_label_shared)) {
      .check_input_plot_args(x_label_shared = x_label_shared)
    } else {
      x_label_shared <- "column"
    }

    # y_label_shared
    if (!is.waive(y_label_shared)) {
      .check_input_plot_args(y_label_shared = y_label_shared)
    } else {
      y_label_shared <- "row"
    }

    # x_label
    if (is.waive(x_label)) x_label <- NULL

    # y_label
    if (is.waive(y_label)) y_label <- NULL

    # rotate_x_tick_labels
    if (is.waive(rotate_x_tick_labels)) rotate_x_tick_labels <- TRUE

    # show_normalised_data
    if (is.logical(show_normalised_data)) {
      if (show_normalised_data) {
        show_normalised_data <- "fixed"
      } else {
        show_normalised_data <- "none"
      }
      
    } else {
      .check_parameter_value_is_valid(
        x = show_normalised_data,
        var_name = "show_normalised_data",
        values = c("none", "fixed", "set_normalisation")
      )
    }

    # legend_label
    if (is.waive(legend_label)) {
      legend_label <- ifelse(show_normalised_data == "none", "value", "norm. value")
    }

    # outcome_legend_label
    if (is.waive(outcome_legend_label)) {
      # Assign default label to the
      if (object@outcome_type %in% c("binomial", "multinomial")) {
        outcome_legend_label <- "class"
      } else if (object@outcome_type %in% c("continuous")) {
        outcome_legend_label <- "value"
      } else if (object@outcome_type %in% c("survival", "competing_risk")) {
        outcome_legend_label <- "event"
      } else if (object@outcome_type == "unsupervised") {
        outcome_legend_label <- "unset"
        show_outcome <- FALSE
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }
    }

    # check outcome_legend_label
    .check_input_plot_label(
      label_var = outcome_legend_label,
      var_name = "outcome_legend_label"
    )

    # x_axis_by and y_axis_by
    available_axis_variables <- c("feature", "sample")
    if (is.null(x_axis_by) && is.null(y_axis_by)) {
      # Set both variables.
      x_axis_by <- available_axis_variables[1L]
      y_axis_by <- available_axis_variables[2L]
      
    } else if (is.null(x_axis_by)) {
      .check_parameter_value_is_valid(
        x = y_axis_by, 
        var_name = "y_axis_by",
        values = available_axis_variables
      )
      
      x_axis_by <- setdiff(available_axis_variables, y_axis_by)
      
    } else if (is.null(y_axis_by)) {
      .check_parameter_value_is_valid(
        x = x_axis_by, 
        var_name = "x_axis_by",
        values = available_axis_variables
      )
      
      y_axis_by <- setdiff(available_axis_variables, x_axis_by)
    } else {
      .check_parameter_value_is_valid(
        x = x_axis_by, 
        var_name = "x_axis_by", 
        values = available_axis_variables
      )
      
      .check_parameter_value_is_valid(
        x = y_axis_by, 
        var_name = "y_axis_by",
        values = available_axis_variables
      )
    }

    .check_value_not_shared(
      x_axis_by, y_axis_by,
      "x_axis_by", "y_axis_by"
    )

    # Check length of x_axis_by and y_axis_by variables.
    .check_argument_length(
      x = x_axis_by,
      var_name = "x_axis_by",
      min = 1L,
      max = 1L
    )
    .check_argument_length(
      x = y_axis_by,
      var_name = "y_axis_by",
      min = 1L,
      max = 1L
    )

    # show_feature_dendrogram
    if (is.logical(show_feature_dendrogram)) {
      if (show_feature_dendrogram) {
        show_feature_dendrogram <- ifelse(x_axis_by == "feature", "top", "right")
      } else {
        show_feature_dendrogram <- NULL
      }
      
    } else if (length(show_feature_dendrogram) == 0L) {
      show_feature_dendrogram <- NULL
      
    } else {
      if (x_axis_by == "feature") {
        .check_parameter_value_is_valid(
          x = show_feature_dendrogram,
          var_name = "show_feature_dendrogram",
          values = c("top", "bottom")
        )
        
      } else {
        .check_parameter_value_is_valid(
          x = show_feature_dendrogram,
          var_name = "show_feature_dendrogram",
          values = c("left", "right")
        )
      }

      .check_argument_length(
        x = show_feature_dendrogram,
        var_name = "show_feature_dendrogram",
        min = 1L, 
        max = 1L
      )
    }

    # Check that the data allows for creating a dendrogram.
    if (is_empty(feature_similarity)) {
      show_feature_dendrogram <- NULL
    } else if (!feature_similarity@cluster_method %in% c("hclust", "agnes", "diana")) {
      show_feature_dendrogram <- NULL
    }

    # show_sample_dendrogram
    if (is.logical(show_sample_dendrogram)) {
      if (show_sample_dendrogram) {
        show_sample_dendrogram <- ifelse(x_axis_by == "sample", "top", "right")
      } else {
        show_sample_dendrogram <- NULL
      }
      
    } else if (length(show_sample_dendrogram) == 0L) {
      show_sample_dendrogram <- NULL
      
    } else {
      if (x_axis_by == "sample") {
        .check_parameter_value_is_valid(
          x = show_sample_dendrogram, 
          var_name = "show_sample_dendrogram",
          values = c("top", "bottom")
        )
        
      } else {
        .check_parameter_value_is_valid(
          x = show_sample_dendrogram, 
          var_name = "show_sample_dendrogram",
          values = c("left", "right")
        )
      }

      .check_argument_length(
        x = show_sample_dendrogram, 
        var_name = "show_sample_dendrogram",
        min = 1L,
        max = 1L
      )
    }

    # Check that the data allows for creating a dendrogram.
    if (is_empty(sample_similarity)) {
      show_sample_dendrogram <- NULL
    } else if (!sample_similarity@cluster_method %in% c("hclust", "agnes", "diana")) {
      show_sample_dendrogram <- NULL
    }

    # Check if the dendrogram_height argument is correct.
    if (!is.null(show_sample_dendrogram) || !is.null(show_feature_dendrogram)) {
      .check_plot_grid_unit(
        x = dendrogram_height,
        var_name = "dendrogram_height"
      )
    }

    # Check if show_outcome is specified correctly.
    if (is.logical(show_outcome)) {
      if (show_outcome) {
        show_outcome <- ifelse(x_axis_by == "sample", "top", "left")
        
      } else {
        show_outcome <- NULL
      }
      
    } else if (length(show_outcome) == 0L) {
      show_outcome <- NULL
      
    } else {
      if (x_axis_by == "sample") {
        .check_parameter_value_is_valid(
          x = show_outcome,
          var_name = "show_outcome",
          values = c("top", "bottom")
        )
        
      } else {
        .check_parameter_value_is_valid(
          x = show_outcome, var_name = "show_outcome",
          values = c("left", "right")
        )
      }

      .check_argument_length(
        x = show_outcome,
        var_name = "show_outcome",
        min = 1L, 
        max = 1L
      )
    }

    # Check if the outcome_height argument is correct.
    if (!is.null(show_outcome)) {
      .check_plot_grid_unit(
        x = outcome_height, 
        var_name = "outcome_height"
      )
    }

    # Add default splitting variables
    if (is.null(split_by) & is.null(facet_by)) {
      # Split by variable importance method, learner and data set
      split_by <- c("vimp_method", "learner", "data_set")

      # Do not use facets.
      facet_by <- NULL
    }

    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = identifier_table,
      split_by = split_by,
      facet_by = facet_by,
      available = c("data_set", "vimp_method", "learner")
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    facet_by <- split_var_list$facet_by

    # Check input arguments
    .check_input_plot_args(
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      rotate_x_tick_labels = rotate_x_tick_labels
    )

    # Create plots -------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data.
    if (!is.null(split_by)) {
      x_split <- split(
        identifier_table,
        by = split_by, 
        drop = FALSE
      )
      
    } else {
      x_split <- list("null.name" = identifier_table)
    }

    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over splits
    for (x_sub in x_split) {
      if (is_empty(x_sub)) next

      # Declare subtitle components.
      additional_subtitle <- NULL

      # Select relevant feature expressions from list.
      feature_expression_split <- feature_expression[x_sub$list_id]

      # Select data for feature similarity.
      feature_similarity_split <- NULL
      if (!is_empty(feature_similarity)) {
        feature_similarity_split <- methods::new(
          "familiarDataElementFeatureSimilarity",
          feature_similarity,
          data = feature_similarity@data[x_sub, on = .NATURAL, nomatch = NULL]
        )

        # Add similarity metric.
        additional_subtitle <- c(
          additional_subtitle,
          list("metric (features)" = feature_similarity_split@similarity_metric)
        )
      }

      # Select data for sample similarity
      sample_similarity_split <- NULL
      if (!is_empty(sample_similarity)) {
        sample_similarity_split <- methods::new(
          "familiarDataElementSampleSimilarity",
          sample_similarity,
          data = sample_similarity@data[x_sub, on = .NATURAL, nomatch = NULL]
        )

        # Add similarity metric.
        additional_subtitle <- c(
          additional_subtitle,
          list("metric (samples)" = sample_similarity_split@similarity_metric)
        )
      }

      if (is.waive(plot_title)) plot_title <- "Sample clustering"

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = x_sub
        )
      }

      # Generate plot
      p <- .plot_sample_clustering_plot(
        x = x_sub,
        data = feature_expression_split,
        feature_similarity = feature_similarity_split,
        sample_similarity = sample_similarity_split,
        outcome_type = object@outcome_type,
        x_axis_by = x_axis_by,
        y_axis_by = y_axis_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        gradient_palette = gradient_palette,
        gradient_palette_range = gradient_palette_range,
        outcome_palette = outcome_palette,
        outcome_palette_range = outcome_palette_range,
        x_label = x_label,
        x_label_shared = x_label_shared,
        y_label = y_label,
        y_label_shared = y_label_shared,
        legend_label = legend_label,
        outcome_legend_label = outcome_legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        x_range = x_range,
        x_n_breaks = x_n_breaks,
        x_breaks = x_breaks,
        y_range = y_range,
        y_n_breaks = y_n_breaks,
        y_breaks = y_breaks,
        rotate_x_tick_labels = rotate_x_tick_labels,
        show_feature_dendrogram = show_feature_dendrogram,
        show_sample_dendrogram = show_sample_dendrogram,
        show_normalised_data = show_normalised_data,
        show_outcome = show_outcome,
        dendrogram_height = dendrogram_height,
        outcome_height = outcome_height
      )

      # Check empty output
      if (is.null(p)) next

      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Find unique features
        features <- lapply(feature_expression_split, function(x) (x@value_column))
        features <- unique(unlist(features))

        # Find unique samples
        samples <- lapply(feature_expression_split, function(x) (x@data$sample_name))
        samples <- unique(unlist(samples))

        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_sample_clustering_plot_dimensions(
          x = x_sub,
          x_axis_by = x_axis_by,
          y_axis_by = y_axis_by,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols,
          features = features,
          samples = samples,
          show_feature_dendrogram = show_feature_dendrogram,
          show_sample_dendrogram = show_sample_dendrogram,
          rotate_x_tick_labels = rotate_x_tick_labels
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "sample_clustering",
              "x" = x_sub,
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_sample_clustering_plot <- function(
    x,
    data,
    feature_similarity,
    sample_similarity,
    outcome_type,
    x_axis_by,
    y_axis_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    gradient_palette,
    gradient_palette_range,
    outcome_palette,
    outcome_palette_range,
    x_label,
    x_label_shared,
    y_label,
    y_label_shared,
    legend_label,
    outcome_legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_n_breaks,
    x_breaks,
    y_range,
    y_n_breaks,
    y_breaks,
    rotate_x_tick_labels,
    show_feature_dendrogram,
    show_sample_dendrogram,
    show_normalised_data,
    show_outcome,
    dendrogram_height,
    outcome_height
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  .NATURAL <- NULL

  # Define elements that need to be shared. Note that "guide" and "strip_y" may
  # be absent.
  elements <- c("guide", "strip_x", "strip_y")
  if (x_label_shared == "overall") elements <- c(elements, "axis_title_x")
  if (y_label_shared == "overall") elements <- c(elements, "axis_title_y")


  # gradient_palette_range and expression_data ---------------------------------

  # Determine the range of the gradient palette.
  if (is.waive(gradient_palette_range)) {
    if (show_normalised_data == "none") {
      # Default to an empty range.
      gradient_palette_range <- c(NA, NA)
      
    } else if (show_normalised_data == "fixed") {
      # Identify the normalisation method used to convert the data.
      normalisation_method <- lapply(
        data,
        function(data) {
          # Obtain the normalisation methods for each feature in the current
          # dataset.
          normalisation_method <- sapply(
            data@feature_info,
            function(feature) (.optional_from_slot(feature@normalisation_parameters, "method", alternative = "standardisation_winsor"))
          )
          
          return(unname(normalisation_method))
        }
      )
      
      # Select the normalisation methods that we should consider.
      normalisation_method <- unique(unlist(normalisation_method))
      
      if (all(normalisation_method == "none")) {
        normalisation_method <- "none"
      } else {
        normalisation_method <- setdiff(normalisation_method, "none")
      }

      # Find the default value.
      gradient_palette_range <- .get_default_normalisation_range_for_plotting(
        norm_method = normalisation_method
      )
      
    } else if (show_normalised_data == "set_normalisation") {
      # By default show range -3 to 3 standard deviations
      gradient_palette_range <- .get_default_normalisation_range_for_plotting(
        norm_method = "standardisation_winsor"
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        ".plot_sample_clustering_plot: encountered unknown value for ",
        "show_normalised_data: ",
        show_normalised_data
      ))
    }
    
  } else if (is.null(gradient_palette_range)) {
    # Default to an empty range.
    gradient_palette_range <- c(NA, NA)
  }

  # Normalise expression data
  data <- lapply(
    data,
    .normalise_expression_data,
    show_normalised_data = show_normalised_data
  )

  # Ensure that a gradient palette range is set. This is required because the
  # legend is shared between all facets and .
  if (!all(is.finite(gradient_palette_range))) {
    # Iterate over expression data to find minimum and maximum
    feature_ranges <- lapply(
      data, 
      function(data) {
        if (is_empty(data)) {
          return(data.table::data.table(
            "min_value" = numeric(0L),
            "max_value" = numeric(0L)
          ))
        }
        
        # Find feature value ranges in the current expression data.
        feature_ranges <- lapply(
          data@feature_info, 
          function(feature, x) {
            # Only numeric features have a range.
            if (feature@feature_type == "numeric") {
              # Find feature values that are finite.
              feature_data <- x[[feature@name]]
              feature_data <- feature_data[is.finite(feature_data)]
              
              if (length(feature_data) == 0L) {
                return(data.table::data.table(
                  "min_value" = numeric(0L),
                  "max_value" = numeric(0L)
                ))
                
              } else {
                return(data.table::data.table(
                  "min_value" = as.double(min(feature_data, na.rm = FALSE)),
                  "max_value" = as.double(max(feature_data, na.rm = FALSE))
                ))
              }
            } else {
              return(data.table::data.table(
                "min_value" = numeric(0L),
                "max_value" = numeric(0L)
              ))
            }
          },
          x = data@data
        )
        
        # Combine ranges for all features.
        feature_ranges <- data.table::rbindlist(
          feature_ranges, 
          use.names = TRUE
        )
        
        return(feature_ranges)
      }
    )
    
    # Combine ranges
    feature_ranges <- data.table::rbindlist(
      feature_ranges, 
      use.names = TRUE
    )
    
    if (is_empty(feature_ranges)) {
      # Set a default if all features are categorical.
      gradient_palette_range <- c(-1.0, 0.0, 1.0)
      
    } else {
      # Find a nice range for missing values of the palette range.
      gradient_palette_range <- .format_plot_number_nice_range(
        input_range = gradient_palette_range,
        x = c(
          min(feature_ranges$min_value),
          max(feature_ranges$max_value)
        )
      )
    }
  }

  # outcome_palette_range and outcome_plot_data --------------------------------

  # Set outcome_palette_range for continuous outcome types.
  if (is.waive(outcome_palette_range)) {
    outcome_palette_range <- c(NA, NA)
  }

  # Process outcome plot data for plotting.
  if (!is.null(show_outcome)) {
    # Iterate over data elements to create outcome plot data.
    outcome_plot_data <- lapply(
      data,
      function(data, outcome_type) {
        if (is_empty(data)) return(NULL)
        
        if (outcome_type %in% c("survival", "competing_risk")) {
          evaluation_times <- data@evaluation_time
          
          outcome_plot_data <- .process_expression_survival_outcome(
            x = data@data,
            evaluation_times = evaluation_times
          )
        } else {
          outcome_plot_data <- .process_expression_generic_outcome(x = data@data)
        }
        
        return(outcome_plot_data)
      },
      outcome_type = outcome_type
    )
  } else {
    outcome_plot_data <- NULL
  }
  
  # Update the outcome palette range based on data present.
  if (
    !all(is.finite(outcome_palette_range)) &&
    outcome_type %in% c("continuous") &&
    !is.null(show_outcome)
  ) {
    # Iterate over outcome_plot_data to find minimum and maximum values.
    outcome_ranges <- lapply(
      outcome_plot_data,
      function(x) {
        if (is_empty(x)) {
          return(data.table::data.table(
            "min_value" = numeric(0L),
            "max_value" = numeric(0L)
          ))
          
        } else if (!any(is.finite(x$value))) {
          return(data.table::data.table(
            "min_value" = numeric(0L),
            "max_value" = numeric(0L)
          ))
          
        } else {
          return(data.table::data.table(
            "min_value" = as.double(min(x$value, na.rm = TRUE)),
            "max_value" = as.double(max(x$value, na.rm = TRUE))
          ))
        }
      }
    )
    
    # Combine outcome ranges
    outcome_ranges <- data.table::rbindlist(
      outcome_ranges,
      use.names = TRUE
    )
    
    if (is_empty(outcome_ranges)) {
      # Set a default if all features are categorical.
      outcome_palette_range <- c(0.0, 1.0)
    } else {
      # Find a nice range for missing values of the palette range.
      outcome_palette_range <- .format_plot_number_nice_range(
        input_range = outcome_palette_range,
        x = c(
          min(outcome_ranges$min_value),
          max(outcome_ranges$max_value)
        )
      )
    }
  }

  # plot creation --------------------------------------------------------------

  # Update the list identifiers in x.
  x <- data.table::copy(x)[, "list_id" := .I]

  # Split by facet. This generates a list of data splits with faceting
  # information that allows for positioning.
  plot_layout_table <- .get_plot_layout_table(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Define the split in data required for faceting.
  layout_split <- split(
    plot_layout_table,
    by = c("col_id", "row_id"),
    sorted = TRUE
  )

  # Create plots to join
  figure_list <- list()
  extracted_element_list <- list()

  for (current_split in layout_split) {
    # Get current split on the identifier table.
    if (is.null(facet_by)) {
      x_split <- x
    } else {
      x_split <- x[current_split, on = .NATURAL]
    }

    # Get expression data split and outcome data split.
    expression_data_split <- data[x_split$list_id]
    outcome_plot_data_split <- outcome_plot_data[x_split$list_id]

    # Check data for a single facet is present.
    if (
      length(expression_data_split) > 1L ||
      length(outcome_plot_data_split) > 1L
    ) {
      ..error_reached_unreachable_code(
        ".plot_sample_clustering_plot: cannot process data from multiple facets simultaneously."
      )
    }

    # Extract data elements from list.
    expression_data_split <- expression_data_split[[1L]]
    outcome_plot_data_split <- outcome_plot_data_split[[1L]]

    # Split feature similarity.
    if (is.null(facet_by) || is_empty(feature_similarity)) {
      feature_similarity_split <- feature_similarity
      
    } else {
      feature_similarity_split <- methods::new(
        "familiarDataElementFeatureSimilarity",
        feature_similarity,
        data = feature_similarity@data[current_split, on = .NATURAL, nomatch = NULL]
      )
    }

    # Split sample_similarity.
    if (is.null(facet_by) || is_empty(sample_similarity)) {
      sample_similarity_split <- sample_similarity
      
    } else {
      sample_similarity_split <- methods::new(
        "familiarDataElementSampleSimilarity",
        sample_similarity,
        data = sample_similarity@data[current_split, on = .NATURAL, nomatch = NULL]
      )
    }

    # Add cluster objects to feature and sample similarity data.
    feature_similarity_split <- .append_feature_similarity_dendrogram(feature_similarity_split)
    sample_similarity_split <- .append_sample_similarity_dendrogram(sample_similarity_split)

    # Complete the expression data
    plot_data <- .complete_expression_table(
      x = expression_data_split,
      feature_similarity = feature_similarity_split,
      sample_similarity = sample_similarity_split,
      gradient_palette_range = gradient_palette_range
    )

    # Create expression heatmap
    p_heatmap <- .create_expression_heatmap(
      x = plot_data,
      x_axis_by = x_axis_by,
      y_axis_by = y_axis_by,
      facet_by = facet_by,
      facet_wrap_cols = facet_wrap_cols,
      ggtheme = ggtheme,
      gradient_palette = gradient_palette,
      gradient_palette_range = gradient_palette_range,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      rotate_x_tick_labels = rotate_x_tick_labels,
      show_feature_dendrogram = show_feature_dendrogram,
      show_sample_dendrogram = show_sample_dendrogram
    )
    
    # Convert to gtable and append "main" to grob names.
    g_heatmap <- .rename_plot_grobs(
      g = .convert_to_grob(p_heatmap),
      extension = "main"
    )
    if (!gtable::is.gtable(g_heatmap)) next

    # Add sample dendogram
    if (!is.null(sample_similarity_split)) {
      if (
        !is.null(show_sample_dendrogram) &&
        inherits(sample_similarity_split@dendrogram, "hclust")
      ) {
        # Obtain dendogram plotting data as line segments.
        dendro_data <- .convert_dendrogram_to_table(
          h = sample_similarity_split@dendrogram,
          similarity_metric = sample_similarity_split@similarity_metric
        )

        # Find the right axes settings.
        if (show_sample_dendrogram %in% c("left", "right")) {
          dist_range <- x_range
          dist_n_breaks <- x_n_breaks
          dist_breaks <- x_breaks
        } else {
          dist_range <- y_range
          dist_n_breaks <- y_n_breaks
          dist_breaks <- y_breaks
        }

        # Plot dendrogram
        p_dendro <- .create_expression_dendrogram_plot(
          x = dendro_data,
          position = show_sample_dendrogram,
          ggtheme = ggtheme,
          dist_range = dist_range,
          dist_n_breaks = dist_n_breaks,
          dist_breaks = dist_breaks,
          plot_height = dendrogram_height,
          rotate_x_tick_labels = rotate_x_tick_labels
        )

        dendro_extension <- paste0("dendro-", show_sample_dendrogram)
        panel_element_name <- paste0("panel-", dendro_extension)
        axis_element_name <- ifelse(show_sample_dendrogram %in% c("top", "bottom"), "axis-l", "axis-b")
        axis_element_name <- paste0(axis_element_name, "-", dendro_extension)
        
        # Convert to gtable
        g_dendro <- .rename_plot_grobs(
          g = .convert_to_grob(p_dendro),
          extension = dendro_extension
        )
        
        where_panel <- switch(
          show_sample_dendrogram,
          "top" = c("above", "panel-main"),
          "bottom" = c("below", "panel-main"),
          "left" = c("left", "panel-main"),
          "right" = c("right", "panel-main")
        )
        
        # Insert panel-main next to panel-dendro.
        g_heatmap <- .gtable_insert(
          g = g_heatmap,
          g_new = .gtable_extract_grob(g_dendro, element = panel_element_name),
          where = where_panel,
          grob_name = panel_element_name,
          spacer = .get_plot_panel_spacing(
            ggtheme = ggtheme, 
            axis = ifelse(show_sample_dendrogram %in% c("top", "bottom"), "y", "x")
          )
        )
        
        where_axis_element <- switch(
          show_sample_dendrogram,
          "top" = c("intersect", "above", "axis-l-main", "left", panel_element_name),
          "bottom" = c("intersect", "below", "axis-l-main", "left", panel_element_name),
          "left" = c("intersect", "below", panel_element_name, "left", "axis-b-main"),
          "right" = c("intersect", "below", panel_element_name, "right", "axis-b-main")
        )
        
        # Insert the axis element at the intersect of dendro-panel and the
        # corresponding axis element of the main plot.
        g_heatmap <- .gtable_insert(
          g = g_heatmap,
          g_new = .gtable_extract_grob(g_dendro, element = axis_element_name),
          where = where_axis_element,
          grob_name = axis_element_name
        )
      }
    }

    # Add feature dendrogram
    if (!is.null(feature_similarity_split)) {
      if (
        !is.null(show_feature_dendrogram) &&
        inherits(feature_similarity_split@dendrogram, "hclust")
      ) {
        # Obtain dendrogram plotting data as line segments.
        dendro_data <- .convert_dendrogram_to_table(
          h = feature_similarity_split@dendrogram,
          similarity_metric = feature_similarity_split@similarity_metric
        )

        # Find the right axes settings.
        if (show_feature_dendrogram %in% c("left", "right")) {
          dist_range <- x_range
          dist_n_breaks <- x_n_breaks
          dist_breaks <- x_breaks
        } else {
          dist_range <- y_range
          dist_n_breaks <- y_n_breaks
          dist_breaks <- y_breaks
        }

        # Plot dendogram
        p_dendro <- .create_expression_dendrogram_plot(
          x = dendro_data,
          position = show_feature_dendrogram,
          ggtheme = ggtheme,
          dist_range = dist_range,
          dist_n_breaks = dist_n_breaks,
          dist_breaks = dist_breaks,
          plot_height = dendrogram_height,
          rotate_x_tick_labels = rotate_x_tick_labels
        )

        dendro_extension <- paste0("dendro-", show_feature_dendrogram)
        panel_element_name <- paste0("panel-", dendro_extension)
        axis_element_name <- ifelse(show_feature_dendrogram %in% c("top", "bottom"), "axis-l", "axis-b")
        axis_element_name <- paste0(axis_element_name, "-", dendro_extension)
        
        # Convert to gtable
        g_dendro <- .rename_plot_grobs(
          g = .convert_to_grob(p_dendro),
          extension = dendro_extension
        )
        
        where_panel <- switch(
          show_feature_dendrogram,
          "top" = c("above", "panel-main"),
          "bottom" = c("below", "panel-main"),
          "left" = c("left", "panel-main"),
          "right" = c("right", "panel-main")
        )
        
        # Insert panel-main next to panel-dendro.
        g_heatmap <- .gtable_insert(
          g = g_heatmap,
          g_new = .gtable_extract_grob(g_dendro, element = panel_element_name),
          where = where_panel,
          grob_name = panel_element_name,
          spacer = .get_plot_panel_spacing(
            ggtheme = ggtheme, 
            axis = ifelse(show_feature_dendrogram %in% c("top", "bottom"), "y", "x")
          )
        )
        
        where_axis_element <- switch(
          show_feature_dendrogram,
          "top" = c("intersect", "above", "axis-l-main", "left", panel_element_name),
          "bottom" = c("intersect", "below", "axis-l-main", "left", panel_element_name),
          "left" = c("intersect", "below", panel_element_name, "left", "axis-b-main"),
          "right" = c("intersect", "below", panel_element_name, "right", "axis-b-main")
        )
        
        # Insert the axis element at the intersect of dendro-panel and the
        # corresponding axis element of the main plot.
        g_heatmap <- .gtable_insert(
          g = g_heatmap,
          g_new = .gtable_extract_grob(g_dendro, element = axis_element_name),
          where = where_axis_element,
          grob_name = axis_element_name
        )
      }
    }

    if (!is.null(show_outcome) && gtable::is.gtable(g_heatmap)) {
      # Create expression outcome plot.
      p_outcome <- .create_expression_outcome_plot(
        x = outcome_plot_data_split,
        ggtheme = ggtheme,
        position = show_outcome,
        outcome_type = outcome_type,
        outcome_palette = outcome_palette,
        outcome_palette_range = outcome_palette_range,
        outcome_legend_label = outcome_legend_label,
        plot_height = outcome_height,
        sample_similarity = sample_similarity_split,
        rotate_x_tick_labels = rotate_x_tick_labels
      )

      # Convert to gtable
      g_outcome <- .rename_plot_grobs(
        g = .convert_to_grob(p_outcome),
        extension = "outcome"
      )
      axis_element_name <- ifelse(show_outcome %in% c("top", "bottom"), "axis-l-outcome", "axis-b-outcome")
      
      where_panel <- switch(
        show_outcome,
        "top" = c("above", "panel-main"),
        "bottom" = c("below", "panel-main"),
        "left" = c("left", "panel-main"),
        "right" = c("right", "panel-main")
      )
      
      # Insert panel-main next to panel-outcome.
      g_heatmap <- .gtable_insert(
        g = g_heatmap,
        g_new = .gtable_extract_grob(g_outcome, element = "panel-outcome"),
        where = where_panel,
        grob_name = "panel-outcome"
      )
      
      where_axis_element <- switch(
        show_outcome,
        "top" = c("intersect", "above", "axis-l-main", "left", "panel-outcome"),
        "bottom" = c("intersect", "below", "axis-l-main", "left", "panel-outcome"),
        "left" = c("intersect", "below", "panel-outcome", "left", "axis-b-main"),
        "right" = c("intersect", "below", "panel-outcome", "right", "axis-b-main")
      )
      
      # Insert the axis element at the intersect of panel-outcome and the
      # corresponding axis element of the main plot.
      g_heatmap <- .gtable_insert(
        g = g_heatmap,
        g_new = .gtable_extract_grob(g_outcome, element = axis_element_name),
        where = where_axis_element,
        grob_name = axis_element_name
      )
      
      # Combine guides,
      g_heatmap <- .combine_plot_elements(
        g_main = g_heatmap,
        g_new = g_outcome,
        element_name = .all_gtable_guide_names(),
        spacer = .get_plot_legend_spacing(ggtheme = ggtheme, axis = "y")
      )
    }

    # Attach to figure list.
    figure_list[[paste0(current_split$row_id, ".", current_split$col_id)]] <- as_familiar_plot(
      g = g_heatmap,
      layout = current_split
    )
  }
  
  # Compose the final figure. Magic.
  g <- .compose_figure(
    figure_list = figure_list,
    plot_layout_table = plot_layout_table,
    x_text_shared = FALSE,
    x_label_shared = x_label_shared,
    y_text_shared = FALSE,
    y_label_shared = y_label_shared,
    facet_wrap_cols = facet_wrap_cols,
    ggtheme = ggtheme
  )

  return(g)
}



.create_expression_heatmap <- function(
    x,
    x_axis_by,
    y_axis_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    gradient_palette,
    gradient_palette_range,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    rotate_x_tick_labels,
    show_feature_dendrogram,
    show_sample_dendrogram
) {
  # Determine whether a sequential or divergent palette should be used by
  # default.
  palette_type <- ifelse(
    length(gradient_palette_range) > 2L,
    "divergent",
    "sequential"
  )
  
  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym(x_axis_by),
      y = !!sym(y_axis_by),
      fill = !!sym("value")
    )
  )
  p <- p + ggtheme

  if (!is_empty(x)) {
    p <- p + ggplot2::geom_raster()
  } else {
    # Insert blank for empty datasets.
    p <- p + ggplot2::geom_blank()
  }

  # Colors
  gradient_colours <- .get_palette(
    x = gradient_palette,
    palette_type = palette_type,
    diverge_to_white = TRUE
  )

  # Add gradient palette. If the legend is not shown, legend_label equals NULL.
  p <- p + ggplot2::scale_fill_gradientn(
    name = legend_label,
    colors = gradient_colours,
    limits = range(gradient_palette_range),
    oob = scales::squish
  )

  # Create show_dendrogram that combines show_feature_dendrogram and
  # show_sample_dendrogram.
  show_dendrogram <- c(show_feature_dendrogram, show_sample_dendrogram)

  # Show dendrogram determines where tick labels and axis labels are placed. By
  # default axes are located at the bottom and right.
  x_axis_position <- "bottom"
  y_axis_position <- "left"
  if (!is.null(show_dendrogram)) {
    if ("bottom" %in% show_dendrogram) {
      x_axis_position <- "top"
    }
    if ("left" %in% show_dendrogram) {
      y_axis_position <- "right"
    }
  }

  # Specify both axes. Note that only the heatmap is shown, without additional
  # space between the plot area and the axes.
  p <- p + ggplot2::scale_x_discrete(position = x_axis_position, expand = c(0.0, 0.0))
  p <- p + ggplot2::scale_y_discrete(position = y_axis_position, expand = c(0.0, 0.0))

  # Set labels.
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Determine how plots are faceted. The actual facets are created in the
  # calling function, not here.
  facet_by_list <- .parse_plot_facet_by(
    x = x@data,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context",
        drop = TRUE
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context",
        drop = TRUE
      )
    }
  }

  # Rotate x-axis ticks
  if (rotate_x_tick_labels) {
    p <- p + ggplot2::theme(
      axis.text.x = ggplot2::element_text(
        vjust = 0.25, 
        hjust = 1.0, 
        angle = 90.0
      )
    )
  }

  return(p)
}



.create_expression_dendrogram_plot <- function(
    x,
    position,
    ggtheme,
    dist_range,
    dist_n_breaks,
    dist_breaks,
    plot_height,
    rotate_x_tick_labels
) {
  # Check if there is any data to plot.
  if (is_empty(x)) return(NULL)
  
  # Define the range along the x-axis.
  x_range <- range(x$x_1, na.rm = TRUE)
  x_range <- c(x_range[1L] - 0.5, x_range[2L] + 0.5)

  # y_range
  if (is.null(dist_range)) {
    dist_range <- range(c(x$y_1, x$y_2), na.rm = TRUE)
  }

  # y_breaks
  if (is.null(dist_breaks)) {
    if (diff(dist_range) == 0.0) {
      dist_range[1L] <- dist_range[1L] - 0.1
      dist_range[2L] <- dist_range[2L] + 0.1
    }
    
    .check_input_plot_args(
      y_range = dist_range,
      y_n_breaks = dist_n_breaks
    )

    # Create breaks and update y_range
    dist_breaks <- labeling::extended(
      m = dist_n_breaks,
      dmin = dist_range[1L],
      dmax = dist_range[2L],
      only.loose = TRUE
    )

    dist_range <- c(
      head(dist_breaks, n = 1L),
      tail(dist_breaks, n = 1L)
    )
  }

  .check_input_plot_args(
    y_range = dist_range,
    y_breaks = dist_breaks
  )

  # Create basic plot
  p <- ggplot2::ggplot(
    data = x, 
    mapping = ggplot2::aes(
      x = !!sym("x_1"),
      y = !!sym("y_1"),
      xend = !!sym("x_2"),
      yend = !!sym("y_2")
    )
  )
  p <- p + ggtheme

  # Plot line segments.
  p <- p + ggplot2::geom_segment(lineend = "round")

  if (position == "right") {
    p <- p + ggplot2::scale_x_continuous(
      limits = x_range,
      expand = c(0.0, 0.0)
    )
    p <- p + ggplot2::scale_y_continuous(
      limits = dist_range,
      breaks = dist_breaks
    )
    p <- p + ggplot2::coord_flip()
    
  } else if (position == "bottom") {
    p <- p + ggplot2::scale_x_continuous(
      limits = x_range, 
      expand = c(0.0, 0.0)
    )
    p <- p + ggplot2::scale_y_reverse(
      limits = rev(dist_range),
      breaks = rev(dist_breaks)
    )
    
  } else if (position == "left") {
    p <- p + ggplot2::scale_x_continuous(
      limits = x_range, 
      expand = c(0.0, 0.0)
    )
    p <- p + ggplot2::scale_y_reverse(
      limits = rev(dist_range),
      breaks = rev(dist_breaks)
    )
    p <- p + ggplot2::coord_flip()
    
  } else if (position == "top") {
    p <- p + ggplot2::scale_x_continuous(
      limits = x_range, 
      expand = c(0.0, 0.0)
    )
    p <- p + ggplot2::scale_y_continuous(
      limits = dist_range, 
      breaks = dist_breaks
    )
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".create_expression_dendrogram_plot: unknown position encountered: ", position
    ))
  }

  # Remove some theme elements and reduce margins. The histogram height is left.
  p <- p + ggplot2::theme(
    panel.grid = ggplot2::element_blank(),
    panel.background = ggplot2::element_blank(),
    panel.border = ggplot2::element_blank(),
    axis.title.x = ggplot2::element_blank(),
    axis.title.y = ggplot2::element_blank()
  )

  if (position %in% c("top", "bottom")) {
    # Remove x-axis
    p <- p + ggplot2::theme(
      axis.line.x = ggplot2::element_blank(),
      axis.ticks.x = ggplot2::element_blank(),
      axis.text.x = ggplot2::element_blank()
    )
    
  } else if (position %in% c("left", "right")) {
    # Remove y-axis (rotated x-axis in plot)
    p <- p + ggplot2::theme(
      axis.line.y = ggplot2::element_blank(),
      axis.ticks.y = ggplot2::element_blank(),
      axis.text.y = ggplot2::element_blank()
    )
    
    # Rotate x-labels
    if (rotate_x_tick_labels) {
      p <- p + ggplot2::theme(
        axis.text.x = ggplot2::element_text(
          vjust = 0.25,
          hjust = 1.0, 
          angle = 90.0
        )
      )
    }
  }

  # Ensure that panel heights/widths are set to the plot object.
  if (position %in% c("top", "bottom")) {
    p$custom_grob <- list("heights" = list(
      "name" = "panel", 
      "height" = plot_height
    ))
  } else if (position %in% c("left", "right")) {
    p$custom_grob <- list("widths" = list(
      "name" = "panel", 
      "width" = plot_height
    ))
  }

  class(p) <- c("familiar_ggplot", class(p))

  return(p)
}



.create_expression_outcome_plot <- function(
    x,
    ggtheme,
    position,
    outcome_type,
    outcome_palette,
    outcome_palette_range,
    outcome_legend_label,
    plot_height,
    sample_similarity,
    rotate_x_tick_labels
) {
  # Set type of palette that is to be used for default palettes.
  if (outcome_type %in% c("binomial", "multinomial")) {
    palette_type <- "qualitative"
  } else if (outcome_type %in% c("continuous")) {
    palette_type <- "sequential"
  } else if (outcome_type %in% c("survival", "competing_risk")) {
    palette_type <- "qualitative"
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }

  # Determine sample order.
  if (is.null(sample_similarity)) {
    # Placeholder in case the sample_similarity object is NULL.
    sample_order <- data.table::data.table("name" = unique(x$sample))
    sample_order[, "label_order" := .I]
  } else if (is.null(sample_similarity@dendrogram)) {
    # Placeholder in case the dendrogram object is NULL.
    sample_order <- data.table::data.table("name" = unique(x$sample))
    sample_order[, "label_order" := .I]
  } else {
    # Default option.
    sample_order <- .compute_sample_similarity_cluster_table(x = sample_similarity)
  }

  # Correctly order the samples
  x$sample <- factor(
    x$sample,
    levels = sample_order$name[order(sample_order$label_order)]
  )

  if (position %in% c("left", "right")) {
    p <- ggplot2::ggplot(
      data = x,
      mapping = ggplot2::aes(
        x = !!sym("evaluation_point"),
        y = !!sym("sample"),
        fill = !!sym("value")
      )
    )
    
    # Limit margins along the axis with samples so it will fit one-to-one with
    # the main heatmap.
    p <- p + ggplot2::scale_y_discrete(expand = c(0.0, 0.0))
    
  } else {
    p <- ggplot2::ggplot(
      data = x,
      mapping = ggplot2::aes(
        x = !!sym("sample"),
        y = !!sym("evaluation_point"),
        fill = !!sym("value")
      )
    )
    
    # Limit margins along the axis with samples so it will fit one-to-one with
    # the main heatmap.
    p <- p + ggplot2::scale_x_discrete(expand = c(0.0, 0.0))
  }
  # Create basic plot
  
  # Add plot theme
  p <- p + ggtheme

  # Plot heatmap
  p <- p + ggplot2::geom_raster()

  # Specify the colours.
  if (outcome_type %in% c("continuous")) {
    # Colors
    outcome_colours <- .get_palette(
      x = outcome_palette,
      palette_type = palette_type,
      use_alternative = TRUE
    )

    # Set the gradient
    p <- p + ggplot2::scale_fill_gradientn(
      name = outcome_legend_label,
      colors = outcome_colours,
      limits = range(outcome_palette_range),
      oob = scales::squish
    )
    
  } else {
    # Colors
    outcome_colours <- .get_palette(
      x = outcome_palette,
      n = nlevels(x$value),
      palette_type = palette_type,
      use_alternative = TRUE
    )

    # Set the qualitative scale
    p <- p + ggplot2::scale_fill_manual(
      name = outcome_legend_label,
      values = outcome_colours[seq_along(levels(x$value))],
      breaks = levels(x$value),
      drop = FALSE
    )
  }

  # Remove some theme elements and reduce margins. The histogram height is left.
  p <- p + ggplot2::theme(
    panel.grid = ggplot2::element_blank(),
    panel.background = ggplot2::element_blank(),
    panel.border = ggplot2::element_blank(),
    axis.title.x = ggplot2::element_blank(),
    axis.title.y = ggplot2::element_blank()
  )

  if (outcome_type %in% c("survival", "competing_risk")) {
    # Survival and competing risk outcomes leave the height axis intact

    if (position %in% c("top", "bottom")) {
      # Remove x-axis
      p <- p + ggplot2::theme(
        axis.line.x = ggplot2::element_blank(),
        axis.ticks.x = ggplot2::element_blank(),
        axis.text.x = ggplot2::element_blank()
      )
      
    } else if (position %in% c("left", "right")) {
      # Remove y-axis (rotated x-axis in plot)
      p <- p + ggplot2::theme(
        axis.line.y = ggplot2::element_blank(),
        axis.ticks.y = ggplot2::element_blank(),
        axis.text.y = ggplot2::element_blank()
      )
      
      # Rotate x-labels
      if (rotate_x_tick_labels) {
        p <- p + ggplot2::theme(
          axis.text.x = ggplot2::element_text(
            vjust = 0.25, 
            hjust = 1.0, 
            angle = 90.0
          )
        )
      }
    }
  } else {
    # For the remaining outcomes, remove all axes.
    p <- p + ggplot2::theme(
      axis.line.x = ggplot2::element_blank(),
      axis.ticks.x = ggplot2::element_blank(),
      axis.text.x = ggplot2::element_blank(),
      axis.line.y = ggplot2::element_blank(),
      axis.ticks.y = ggplot2::element_blank(),
      axis.text.y = ggplot2::element_blank()
    )
  }

  # Adapt plot_height so that it scales with the number of evaluation points.
  if (outcome_type %in% c("survival", "competing_risk")) {
    plot_height <- nlevels(x$evaluation_point) * plot_height
  }

  # Ensure that panel heights/widths are set to the plot object.
  if (position %in% c("top", "bottom")) {
    p$custom_grob <- list("heights" = list(
      "name" = "panel",
      "height" = plot_height
    ))
  } else if (position %in% c("left", "right")) {
    p$custom_grob <- list("widths" = list(
      "name" = "panel",
      "width" = plot_height
    ))
  }

  class(p) <- c("familiar_ggplot", class(p))

  return(p)
}



.determine_sample_clustering_plot_dimensions <- function(
    x,
    x_axis_by,
    y_axis_by,
    facet_by,
    facet_wrap_cols,
    features,
    samples,
    show_feature_dendrogram,
    show_sample_dendrogram,
    rotate_x_tick_labels
) {
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Determine the number of elements along the x-axis.
  if (x_axis_by == "feature") {
    x_n_elements <- length(features)
    x_longest_element <- max(sapply(features, nchar))
    y_n_elements <- length(samples)
    y_longest_element <- max(sapply(samples, nchar))
  } else {
    x_n_elements <- length(samples)
    x_longest_element <- max(sapply(samples, nchar))
    y_n_elements <- length(features)
    y_longest_element <- max(sapply(features, nchar))
  }

  # Assume each x-axis element takes up about 0.5 cm. Then add some room for
  # other plot elements.
  default_width <- x_n_elements * 0.5 + 1.0

  # Assume each y-axis element takes up about 0.5 cm as well.
  default_height <- y_n_elements * 0.5 + 1.0

  # Reserve space for y-axis and x-axis tick labels. Assume that the typical
  # width of a character is about 5 points (1.8 mm). For the x-axis we only
  # reserve extra space in case the ticks are rotated, otherwise we just
  # assume a typical height of 10 points (3.6 mm).
  y_tick_space <- y_longest_element * 0.18
  x_tick_space <- ifelse(rotate_x_tick_labels, x_longest_element * 0.18, 0.36)

  # Reserve space for the dendrograms (1.5 cm)
  dendro_height <- ifelse(
    any(c("top", "bottom") %in% c(show_feature_dendrogram, show_sample_dendrogram)),
    1.5, 0.0
  )
  dendro_width <- ifelse(
    any(c("left", "right") %in% c(show_feature_dendrogram, show_sample_dendrogram)),
    1.5, 0.0
  )

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * (default_height + x_tick_space + dendro_height), 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm). We leave some
  # room for the legend on the right.
  width <- min(c(5.0 + plot_dims[2L] * (default_width + y_tick_space + dendro_width), 19.0))

  return(c(height, width))
}



.normalise_expression_data <- function(x, show_normalised_data) {
  # Check for empty data
  if (is_empty(x)) return(NULL)

  # Make local copy of data element to avoid updating by reference.
  x@data <- data.table::copy(x@data)

  # Apply normalisation
  if (show_normalised_data == "fixed") {
    # Fixed normalization applies normalisation and transformation parameters
    # derived during model creation. Note that extract_feature_expression method
    # already corrects batch differences (if present), and we do not apply batch
    # normalisation as a consequence.

    # Transform features
    x@data <- transform_features(
      data = x@data,
      feature_info_list = x@feature_info,
      features = names(x@feature_info),
      invert = FALSE
    )

    # Normalise features
    x@data <- normalise_features(
      data = x@data,
      feature_info_list = x@feature_info,
      features = names(x@feature_info),
      invert = FALSE
    )
    
  } else if (show_normalised_data == "set_normalisation") {
    # Normalise features within the current dataset.
    for (curr_feat in names(x@feature_info)) {
      x@data[
        ,
        (curr_feat) := .normalise(
          get(curr_feat),
          normalisation_method = "standardisation_winsor"
        )
      ]
    }
  }

  return(x)
}



.complete_expression_table <- function(
    x,
    feature_similarity,
    sample_similarity,
    gradient_palette_range
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  feature <- sample <- NULL

  if (is_empty(x)) return(NULL)

  # Replace categorical features by numerical values and scale to 0.05-.95 of
  # the z-range.
  categorical_features <- lapply(
    x@feature_info, 
    function(feature) {
      if (feature@feature_type == "factor") {
        return(feature@name)
      } else {
        return(NULL)
      }
    }
  )

  # Identify categorical features
  categorical_features <- unique(unlist(categorical_features))

  # Convert to numerical and rescale so that the categorical values lie within
  # the range of the remaining numerical data.
  if (length(categorical_features) > 0L) {
    # Determine the output value range
    output_value_range <- numeric(2L)
    output_value_range[1L] <- head(gradient_palette_range, n = 1L) + 
      0.05 * diff(range(gradient_palette_range))
    output_value_range[2L] <- head(gradient_palette_range, n = 1L) + 
      0.95 * diff(range(gradient_palette_range))

    # Make a local copy of x to prevent warnings raised by data.table.
    x <- data.table::copy(x)

    for (feature in categorical_features) {
      # Convert categorical data to numerical data.
      numeric_data <- as.numeric(x@data[[feature]])

      # Determine the range of the input value range.
      input_value_range <- c(1.0, length(x@feature_info[[feature]]@levels))

      # Convert numeric data to the output range.
      numeric_data <- (numeric_data - input_value_range[1L]) / 
        diff(input_value_range) * diff(output_value_range) + 
        output_value_range[1L]

      # Update column
      x@data[[feature]] <- numeric_data
    }
  }

  # Copy x and revert feature_name_1 and feature_name_2 so that all pairs are
  # present.
  data <- data.table::melt(
    data = x@data,
    measure.vars = names(x@feature_info),
    variable.name = "feature",
    value.name = "value",
    variable.factor = TRUE,
    value.factor = FALSE
  )

  # Change sample_name to sample
  data.table::setnames(
    x = data, 
    old = "sample_name",
    new = "sample"
  )

  # Determine feature order.
  if (is.null(feature_similarity)) {
    # Placeholder in case the feature_similarity object is NULL.
    feature_order <- data.table::data.table("name" = unique(data$feature))
    feature_order[, "label_order" := .I]
  } else if (is.null(feature_similarity@dendrogram)) {
    # Placeholder in case the dendrogram object is NULL.
    feature_order <- data.table::data.table("name" = unique(data$feature))
    feature_order[, "label_order" := .I]
  } else {
    # Default option.
    feature_order <- .compute_feature_similarity_cluster_table(x = feature_similarity)
  }

  # Determine sample order.
  if (is.null(sample_similarity)) {
    # Placeholder in case the sample_similarity object is NULL.
    sample_order <- data.table::data.table("name" = x@data$sample_name)
    sample_order[, "label_order" := .I]
  } else if (is.null(sample_similarity@dendrogram)) {
    # Placeholder in case the dendrogram object is NULL.
    sample_order <- data.table::data.table("name" = x@data$sample_name)
    sample_order[, "label_order" := .I]
  } else {
    # Default option.
    sample_order <- .compute_sample_similarity_cluster_table(x = sample_similarity)
  }

  # Keep only features and samples that are in feature order and sample order.
  data <- data[feature %in% feature_order$name & sample %in% sample_order$name]

  # Set feature and sample order
  data$feature <- factor(data$feature,
    levels = feature_order$name[order(feature_order$label_order)]
  )
  data$sample <- factor(data$sample,
    levels = sample_order$name[order(sample_order$label_order)]
  )

  return(data)
}



.process_expression_generic_outcome <- function(x) {
  # Keep only one copy for each sample.
  x <- data.table::copy(x[, c("sample_name", "outcome")])

  # Rename columns
  data.table::setnames(
    x = x,
    old = c("sample_name", "outcome"),
    new = c("sample", "value")
  )

  # Set evaluation point (which is on the y-axis)
  x$evaluation_point <- factor("1", levels = "1")

  return(x)
}



.process_expression_survival_outcome <- function(x, evaluation_times) {
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome_time <- outcome_event <- NULL

  # Check if evaluation_times fall in a valid range.
  sapply(
    evaluation_times, 
    .check_number_in_valid_range, 
    var_name = "evaluation_times",
    range = c(0.0, Inf), 
    closed = c(FALSE, TRUE)
  )

  # Keep only one copy for each sample.
  x <- data.table::copy(x[, c("sample_name", "outcome_time", "outcome_event")])

  plot_data <- lapply(
    evaluation_times, 
    function(eval_time, x) {
      # Make a copy with the relevant data.
      y <- data.table::copy(x)
      
      # Mark all as no event initially.
      y[, "value" := "no"]
      
      # Mark all data with missing outcome_time or outcome_event as missing.
      y[!is.finite(outcome_time) | is.na(outcome_event), "value" := NA_character_]
      
      # All samples that have an event before or at the evaluation time.
      y[outcome_time <= eval_time & outcome_event == 1L, "value" := "yes"]
      
      # All samples that were censored (lost to follow-up) at the evaluation
      # time.
      y[outcome_time <= eval_time & outcome_event == 0L, "value" := "cens."]
      
      # Set as factor
      y$value <- factor(y$value, levels = c("no", "yes", "cens."))
      
      # Set evaluation time point
      y[, "evaluation_point" := eval_time]
      
      return(y)
    }, 
    x = x
  )
  
  # Combine plot data
  plot_data <- data.table::rbindlist(plot_data)
  
  # Set evaluation point as a factor
  plot_data$evaluation_point <- factor(
    x = plot_data$evaluation_point,
    levels = evaluation_times
  )
  
  # Change subject_id to sample
  data.table::setnames(
    x = plot_data, 
    old = "sample_name",
    new = "sample"
  )

  return(plot_data)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_shap_summary (generic) --------------------------------------------------

#' @title Plot SHAP summary.
#'
#' @description This method creates plots that show a summary of SHAP values
#'   obtained from the data stored in a familiarCollection object.
#'
#' @param dir_path (*optional*) Path to the directory where created SHAP summary
#'   plots are saved to. Output is saved in the `explanation` subdirectory. If
#'   `NULL` no figures are saved, but are returned instead.
#' @param plot_type (*optional*) Type of plot to draw. This is one of
#'   `swarmplot` (draws a beeswarm plot), `barplot` (draws a barplot),
#'   `boxplot` (draws a boxplot) and `violinplot` (draws a violin plot).
#'   Defaults to `boxplot` if a single SHAP value is available for each feature,
#'   and `swarmplot` otherwise.
#'
#'   The choice for `plot_type` affects several other arguments.
#' @param discrete_palette (*optional*) Palette for colouring plot elements
#'   indicated by the `color_by` argument (if any). Only used if `plot_type` is
#'   not `swarmplot`. `familiar` has a default palette. Other palettes are
#'   supported by the `paletteer` package, `grDevices::palette.pals()` (requires
#'   R >= 4.0.0), `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`,
#'   `heat.colors`, `terrain.colors`, `topo.colors` and `cm.colors`, which
#'   correspond to the palettes of the same name in `grDevices`. You may also
#'   specify your own palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param gradient_palette (*optional*) Sequential or divergent palette used to
#'   colour the points  the raster in the default `swarmplot` plots. This
#'   argument is not used for other `plot_type` value. `familiar` has a default
#'   palette. Other palettes are supported by the `paletteer` package,
#'   `grDevices::palette.pals()` (requires R >= 4.0.0), `grDevices::hcl.pals()`
#'   (requires R >= 3.6.0) and `rainbow`, `heat.colors`, `terrain.colors`,
#'   `topo.colors` and `cm.colors`, which correspond to the palettes of the same
#'   name in `grDevices`. You may also specify your own palette by providing a
#'   vector of colour names listed by `grDevices::colors()` or through
#'   hexadecimal RGB strings.
#' @param value_representation (*optional*) Indicates how SHAP values are
#'   represented, with the following options:
#'
#'   * `raw` (default for `swarmplot`, `boxplot`, and `violinplot` plot 
#'     types): uses SHAP values as they are.
#'
#'   * `abs`: uses absolute value of SHAP values.
#'
#'   * `abs_mean` (default for `barplot` plot type): uses mean absolute value of
#'     SHAP values. Only used by `barplot`.
#'
#'   * `abs_max`: uses maximum absolute value of SHAP values. Only used by 
#'     `barplot`.
#'
#'   * `abs_min`: uses minimum absolute value of SHAP values. Only used by 
#'     `barplot`.
#'   
#'   If `abs_mean`, `abs_max` or `abs_min` are chosen, `plot_type` automatically
#'   switches to `barplot`.
#'
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams extract_performance -object
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#'
#' @details This function creates SHAP summary plots, which provide an overview
#'   of marginal contributions of feature values to the predicted values.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, `data_set`,
#'   `evaluation_time` (survival outcome only) and `positive_class` (categorical
#'   outcomes). The default for `value_representation = "raw"` is to facet by
#'   `evaluation_time` or `positive_class`, and split by `vimp_method` and
#'   `learner`. `color_by` is not used. The default for other
#'   `value_representation` is to `color_by` `evaluation_time` or
#'   `positive_class`, and split by `vimp_method`, `learner` and `data_set`.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_learner_names`
#'   can be applied to the `familiarCollection` object to update labels, and
#'   order the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_shap_summary
#' @md
#' @rdname plot_shap_summary-methods
setGeneric(
  "plot_shap_summary",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    plot_type = NULL,
    value_representation = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_shap_summary")
  }
)



# plot_shap_summary (general) --------------------------------------------------

#' @rdname plot_shap_summary-methods
setMethod(
  "plot_shap_summary",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    plot_type = NULL,
    value_representation = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "shap"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_shap_summary,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "x_axis_by" = x_axis_by,
        "y_axis_by" = y_axis_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "plot_type" = plot_type,
        "value_representation" = value_representation,
        "discrete_palette" = discrete_palette,
        "gradient_palette" = gradient_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "limit_n_features" = limit_n_features,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_shap_summary (collection) -----------------------------------------------

#' @rdname plot_shap_summary-methods
setMethod(
  "plot_shap_summary",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    plot_type = NULL,
    value_representation = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    limit_n_features = waiver(),
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Check input arguments ----------------------------------------------------
    
    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)
    
    # Load the data.
    x <- export_shap(object = object)
    x <- x$shap_summary
    if (is_empty(x)) return(NULL)
    
    # Obtain single data element from list.
    if (is.list(x)) {
      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_shap_summary: list of data elements contains unmerged elements."
        )
      }
      x <- x[[1L]]
    }
    
    # Check that the data are not evaluated at the model level.
    if (x@detail_level == "model") {
      ..warning_no_comparison_between_models()
      return(NULL)
    }
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    # Ensure that we work with a copy of the data.
    x@data <- data.table::copy(x@data)
    
    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to create a SHAP summary plot",
      message_type = "warning"
    )) {
      return(NULL)
    }
    
    # Check whether one or multiple SHAP-values are available for each feature.
    value_data <- x@data[, list("n" = .N), by = setdiff(x@grouping_column, c("sample_id", "feature_value", "feature_label"))]
    is_single_sample <- all(value_data$n == 1L)
    
    # Set initial estimates for plot-type. If only a single SHAP value is
    # present, use barplot. Otherwise, if unset, use swarmplot.
    if (is_single_sample) {
      plot_type <- "barplot"
      if (is.waive(x_label) && is.null(value_representation)) {
        x_label <- "SHAP"
      }
    } else if (is.null(plot_type)) {
      plot_type <- "swarmplot"
    }
    
    # Check value representation. If unset, the default for barplots is
    # abs_mean; and raw otherwise. Furthermore, ensure that color_by
    # is reserved for feature values when using "raw".
    if (is.null(value_representation)) {
      if (plot_type == "barplot") {
        value_representation <- "abs_mean"
        
      } else {
        value_representation <- "raw"
        color_by <- NULL
      }
    }
    .check_parameter_value_is_valid(
      x = value_representation,
      var_name = "value_representation",
      values = c("raw", "abs", "abs_mean", "abs_max", "abs_min")
    )
    
    # For abs_mean, abs_max and abs_min, force plot_type to be "bar_plot"
    if (value_representation %in% c("abs_mean", "abs_max", "abs_min")) {
      plot_type <- "barplot"
    }
    
    # Check plot_type.
    .check_parameter_value_is_valid(
      x = plot_type,
      var_name = "plot_type",
      values = c("swarmplot", "barplot", "boxplot", "violinplot")
    )
    
    # Add evaluation time as a splitting variable.
    additional_variable <- NULL
    if (object@outcome_type %in% c("survival")) {
      additional_variable <- "evaluation_time"
      data.table::setnames(x@data, old = "shap_outcome", new = "evaluation_time")
      
    } else if (object@outcome_type %in% c("multinomial")) {
      additional_variable <- "positive_class"
      data.table::setnames(x@data, old = "shap_outcome", new = "positive_class")
    }
    
    # Add default splitting variables.
    all_variables <- c("vimp_method", "learner", "data_set", additional_variable)
    if (
      is.null(split_by) &&
      is.null(color_by) &&
      is.null(facet_by)
    ) {
      # Split by vimp_method and learner.
      split_by <- c("vimp_method", "learner", "data_set")
      
      # Set facetting and color_by variables.
      if (plot_type == "bar_plot") {
        color_by <- additional_variable
        
      } else if (value_representation == "raw" && plot_type == "swarmplot") {
        color_by <- NULL
        facet_by <- additional_variable

      } else {
        color_by <- additional_variable
      }
    }
    
    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      color_by = color_by,
      facet_by = facet_by,
      available = all_variables
    )
    
    # Update splitting variables
    split_by <- split_var_list$split_by
    color_by <- split_var_list$color_by
    facet_by <- split_var_list$facet_by
    
    # x_label
    if (is.waive(x_label)) {
      x_label <- switch(
        value_representation,
        raw = "SHAP",
        abs = "|SHAP|",
        abs_mean = "mean(|SHAP|)",
        abs_max = "max(|SHAP|)",
        abs_min = "min(|SHAP|)"
      )
    }
    
    # y_label
    if (is.waive(y_label)) {
      y_label <- "feature"
    }
    
    .check_input_plot_args(
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      limit_n_features = limit_n_features
    )
    
    # Create plots -------------------------------------------------------------
    
    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)
    
    # Split data.
    if (!is.null(split_by)) {
      x_split <- split(
        x@data, 
        by = split_by, 
        drop = FALSE
      )
      
    } else {
      x_split <- list("null.name" = x@data)
    }
    
    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()
    
    # Iterate over data splits.
    for (ii in names(x_split)) {
      # Skip empty datasets.
      if (is_empty(x_split[[ii]])) next
      
      if (is.waive(plot_title)) plot_title <- "SHAP summary"
      
      # Declare subtitle components.
      additional_subtitle <- NULL
      
      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, color_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(x_split[[ii]]$evaluation_time[1L])
        )
      }
      
      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = x_split[[ii]]
        )
      }
      
      # Generate plot
      p <- .plot_shap_summary_plot(
        x = x_split[[ii]],
        color_by = color_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        plot_type = plot_type,
        value_representation = value_representation,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        gradient_palette = gradient_palette,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        limit_n_features = limit_n_features,
        x_range = x_range,
        x_n_breaks = x_n_breaks,
        x_breaks = x_breaks,
        outcome_type = object@outcome_type
      )
      
      # Check empty output
      if (is.null(p)) next
      
      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)
      
      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_shap_summary_plot_dimensions(
          x = x_split[[ii]],
          plot_type = plot_type,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols
        )
        
        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "explanation",
              "subtype" = paste0("shap", "_", plot_type, "_", value_representation),
              "x" = x_split[[ii]],
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }
    
    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_shap_summary_plot <- function(
    x,
    color_by,
    facet_by,
    facet_wrap_cols,
    plot_type,
    value_representation,
    ggtheme,
    discrete_palette,
    gradient_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    limit_n_features,
    x_range,
    x_n_breaks,
    x_breaks,
    outcome_type
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  shap_value <- vimp <- feature_value <- feature_name <- NULL
  
  value_group_columns <- c("vimp_method", "learner", "feature_name")
  if ("evaluation_time" %in% colnames(x)) value_group_columns <- c(value_group_columns, "evaluation_time")
  if ("positive_class" %in% colnames(x)) value_group_columns <- c(value_group_columns, "positive_class")
  
  # Apply representation.
  if (value_representation == "abs") {
    x[, "shap_value" := abs(shap_value)]
    
  } else if (value_representation == "abs_mean") {
    x <- x[, list("shap_value" = mean(abs(shap_value))), by = value_group_columns]
    
  } else if (value_representation == "abs_max") {
    x <- x[, list("shap_value" = max(abs(shap_value))), by = value_group_columns]
    
  } else if (value_representation == "abs_min") {
    x <- x[, list("shap_value" = min(abs(shap_value))), by = value_group_columns]
  }
  
  # Map feature values to (-1, 1) range for raw representation.
  if (value_representation == "raw") {
    x[
      ,
      "feature_value" := 2.0 * (feature_value - min(feature_value)) / (max(feature_value) - min(feature_value))- 1.0,
      by = "feature_name"
    ]
  }
  
  # Sort features by importance (mean absolute SHAP).
  feature_importance <- x[, list("vimp" = mean(abs(shap_value))), by = value_group_columns]
  feature_importance <- feature_importance[, list("vimp" = max(vimp)), by = "feature_name"][order(vimp)]
  
  # Determine the features that need to be plotted.
  if (is.numeric(limit_n_features)) {
    # Explicitly select features based on threshold value instead of simply 
    # selecting the best features: features may have the same value because they
    # belong to the same cluster.
    threshold_value <- min(tail(unique(sort(feature_importance$vimp)), n = limit_n_features))
    selected_features <- feature_importance[vimp >= threshold_value]$feature_name
    
    # Make slice.
    x <- x[feature_name %in% selected_features, ]
    feature_importance <- feature_importance[feature_name %in% selected_features, ][order(vimp)]
  }
  
  x$feature_name <- factor(
    x = x$feature_name,
    levels = feature_importance$feature_name
  )
  
  # Check x-range.
  if (is.null(x_range)) {
    if (value_representation == "raw") {
      x_range <- c(min(x$shap_value, na.rm = TRUE), max(x$shap_value, na.rm = TRUE))
      
    } else {
      x_range <- c(0.0, max(x$shap_value, na.rm = TRUE))
    }
    
  } else {
    .check_input_plot_args(x_range = x_range)
  }
  
  # Check that the range is not closed.
  if (all(x_range == 0.0)) {
    x_range[2L] <- 0.1
  }
  
  # x_breaks
  if (is.null(x_breaks)) {
    .check_input_plot_args(
      x_range = x_range,
      x_n_breaks = x_n_breaks
    )
    
    # Create breaks and update x_range
    x_breaks <- labeling::extended(
      m = x_n_breaks,
      dmin = x_range[1L],
      dmax = x_range[2L],
      only.loose = TRUE
    )
    
    x_range <- c(
      head(x_breaks, n = 1L),
      tail(x_breaks, n = 1L)
    )
    
  } else {
    .check_input_plot_args(x_breaks = x_breaks)
  }

  # Create a legend label.
  legend_label <- .create_plot_legend_title(
    user_label = legend_label,
    color_by = if(value_representation == "raw") "feature_value" else color_by
  )
  
  # Check remaining input arguments.
  .check_input_plot_args(
    legend_label = legend_label
  )
  
  # Generate a guide table
  guide_list <- .create_plot_guide_table(
    x = x,
    color_by = color_by,
    discrete_palette = discrete_palette
  )
  
  # Extract data
  x <- guide_list$data
  
  # Extract guide_table for color.
  g_color <- guide_list$guide_color
  
  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("shap_value"),
      y = !!sym("feature_name")
    ))
  p <- p + ggtheme
  
  # Set breaks and range.
  p <- p + ggplot2::scale_x_continuous(breaks = x_breaks)
  p <- p + ggplot2::coord_cartesian(xlim = x_range)
  
  if (plot_type == "swarmplot") {
    # Swarm plot ---------------------------------------------------------------
    
    # Determine the density of points for each feature as function of the
    # shap-value.
    grouping_variables <- c("feature_name", facet_by)
    if (!is.null(color_by)) {
      grouping_variables <- c(grouping_variables, "color_breaks")
    }
    
    x[
      ,
      "y_offset" := ..set_shap_swarmplot_jitter(shap_value, feature_value, value_representation = value_representation),
      by = c("feature_name", facet_by, color_by)
    ]
    
    if (value_representation == "raw") {
      p <- p + ggplot2::geom_point(
        mapping = ggplot2::aes(color = !!sym("feature_value")),
        position = ggplot2::position_nudge(y = x$y_offset)
      )
      
      # Colors
      gradient_colours <- .get_palette(
        x = gradient_palette, 
        palette_type = "divergent"
      )
      
      # Add gradient palette.
      p <- p + ggplot2::scale_colour_gradientn(
        name = legend_label,
        colors = gradient_colours,
        limits = c(-1.0, 1.0)
      )
      
    } else if (is.null(color_by)) {
      p <- p + ggplot2::geom_point(position = ggplot2::position_nudge(y = x$y_offset))
      
    } else {
      p <- p + ggplot2::geom_jitter(
        mapping = ggplot2::aes(color = !!sym("color_breaks")),
        position = ggplot2::position_nudge(y = x$y_offset)
      )
      
      # Set fill colours.
      p <- p + ggplot2::scale_color_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        drop = FALSE
      )
    }
    
  } else if (plot_type == "barplot") {
    # Barplot ------------------------------------------------------------------
    
    if (is.null(color_by)) {
      p <- p + ggplot2::geom_bar(
        stat = "identity",
        position = ggplot2::position_dodge(width = 0.9),
      )
      
    } else {
      # Add barplot.
      p <- p + ggplot2::geom_bar(
        mapping = ggplot2::aes(
          fill = !!sym("color_breaks")
        ),
        stat = "identity",
        position = ggplot2::position_dodge(width = 0.9)
      )
      
      # Set fill colours.
      p <- p + ggplot2::scale_fill_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        drop = FALSE
      )
    }
    
  } else if (plot_type == "boxplot") {
    # Boxplot ------------------------------------------------------------------
    
    if (is.null(color_by)) {
      p <- p + ggplot2::geom_boxplot()
      
    } else {
      p <- p + ggplot2::geom_boxplot(
        mapping = ggplot2::aes(
          colour = !!sym("color_breaks")
        )
      )
      
      # Set fill colours.
      p <- p + ggplot2::scale_colour_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        drop = FALSE
      )
    }
    
  } else if (plot_type == "violinplot") {
    # Violinplot ---------------------------------------------------------------
    
    if (is.null(color_by)) {
      # Create boxplot.
      p <- p + ggplot2::geom_violin(
        quantiles = c(0.025, 0.5, 0.975),
        quantile.linetype = ggtheme$line$linetype,
        scale = "width",
        position = ggplot2::position_dodge(width = 1.0)
      )
      
    } else {
      # Create boxplot.
      p <- p + ggplot2::geom_violin(
        mapping = ggplot2::aes(
          fill = !!sym("color_breaks")
        ),
        quantiles = c(0.025, 0.5, 0.975),
        quantile.linetype = ggtheme$line$linetype,
        scale = "width",
        position = ggplot2::position_dodge(width = 1.0)
      )
      
      # Set fill colours.
      p <- p + ggplot2::scale_fill_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        drop = FALSE
      )
    }
  }
  
  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )
  
  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows, 
        cols = facet_by_list$facet_cols, 
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by, 
        labeller = "label_context"
      )
    }
  }
  
  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label, 
    y = y_label, 
    title = plot_title, 
    subtitle = plot_sub_title, 
    caption = caption
  )
  
  return(p)
}



..set_shap_swarmplot_jitter <- function(x, value, value_representation = "raw") {
  # Prevent notes.
  feature_value <- density <- y_offset <- NULL
  
  # Get density. In edge cases, this will fail (e.g. too few samples to
  # establish a bandwidth for the filter.)
  density_object <- tryCatch(
    stats::density(x = x),
    error = identity
  )
  
  if (inherits(density_object, "density")) {
    # Find density at every point and normalise.
    data <- data.table::data.table(
      "original_order" = seq_along(x),
      "shap_value" = x,
      "feature_value" = value,
      "density" = stats::approx(
        x = density_object$x,
        y = density_object$y, 
        xout = x
      )$y
    )
    data[, "density" := density / max(density)]
    
  } else if (length(x) == 1L) {
    data <- data.table::data.table(
      "original_order" = seq_along(x),
      "shap_value" = x,
      "feature_value" = value,
      "density" = 0.0
    )
    
  } else {
    data <- data.table::data.table(
      "original_order" = seq_along(x),
      "shap_value" = x,
      "feature_value" = value,
      "density" = 1.0
    )
  }
  
  if (value_representation == "raw") {
    offset <- stats::rnorm(n = length(x), mean = 0.0, sd = 0.1)
    offset[offset < -0.3] <- -0.3
    offset[offset > 0.3] <- 0.3
    data[, "y_offset" := (feature_value + offset) * density]
    data[, "y_offset" := 0.25 * y_offset / max(abs(y_offset))]
    
  } else {
    offset <- stats::runif(n = length(x), min = -1.0, max = 1.0)
    data[, "y_offset" := offset * density]
    data[, "y_offset" := 0.25 * y_offset / max(abs(y_offset))]
  }
  
  # Replace NaN-values.
  data[is.na(y_offset), "y_offset" := 0.0]
  
  return(data$y_offset)
}



.determine_shap_summary_plot_dimensions <- function(
    x,
    plot_type,
    x_axis_by,
    y_axis_by,
    facet_by,
    facet_wrap_cols
) {
  
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- 4.0
  
  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))
  
  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))
  
  return(c(height, width))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarKNN ------------------------------------------------------------------
setClass(
  "familiarKNN",
  contains = "familiarModel"
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarKNN"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- c("e1071", "proxy")

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarKNN"),
  function(object, ...) {
    # k-nearest neighbours is only available for categorical outcomes.
    if (object@outcome_type %in% c("binomial", "multinomial", "continuous")) {
      return(TRUE)
      
    } else if (object@outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(FALSE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarKNN"),
  function(object, data = NULL, ...) {
    # Initialise list and declare hyperparameter entries.
    param <- list()
    param$sign_size <- list()
    param$k <- list()
    param$distance_metric <- list()

    # If data is explicitly NULL, return the list with hyperparameter names
    # only.
    if (is.null(data)) return(param)

    # Get the number of unique series.
    n_samples <- data.table::uniqueN(
      data@data,
      by = get_id_columns(id_depth = "series")
    )

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data)

    # number of nearest neighbours k -------------------------------------------

    # Define the range for the number of nearest neighbour clusters.
    k_range <- c(1L, max(c(1L, as.integer(ceiling(2L * n_samples^(1.0 / 3.0))))))

    # Define the default value.
    k_default <- sort(unique(c(1L, 2L, 5L, 10L, 20L, k_range)))
    k_default <- k_default[k_default >= k_range[1L] & k_default <= k_range[2L]]

    param$k <- .set_hyperparameter(
      default = k_default,
      type = "integer",
      range = k_range,
      valid_range = c(1L, Inf),
      randomise = TRUE
    )

    # distance metric ----------------------------------------------------------

    # Read distance_metric string by parsing the learner.
    distance_metric <- sub_all_patterns(
      x = object@learner,
      pattern = c("knn", "k_nearest_neighbours"), 
      replacement = "", 
      fixed = TRUE
    )
    
    if (distance_metric != "") {
      distance_metric <- sub(
        x = distance_metric, 
        pattern = "_", 
        replacement = "", 
        fixed = TRUE
      )
    }

    if (distance_metric == "") {
      # Detect feature columns present in the data.
      feature_columns <- get_feature_columns(x = data)

      # Set default values based on the data. For fully numeric data, use
      # euclidean by default, and gower otherwise.
      if (all(sapply(
        feature_columns, 
        function(ii, data) (is.numeric(data@data[[ii]])),
        data = data
      ))) {
        distance_metric_default <- "euclidean"
        
      } else {
        distance_metric_default <- "gower"
      }
      
    } else {
      # A distance metric was unambiguously defined.
      distance_metric_default <- distance_metric
    }

    # Set default and valid ranges.
    distance_metric_default_range <- c("euclidean", "manhattan", "gower")
    # Do not generate additional distance metrics if proxy is not available.
    if (requireNamespace("proxy", quietly = TRUE)) {
      distance_metric_valid_range <- tolower(unname(unlist(lapply(
        proxy::pr_DB$get_entries(), 
        function(list_elem) (list_elem$names)
      ))))
      
    } else {
      distance_metric_valid_range <- distance_metric_default_range
    }

    # Set distance metric.
    param$distance_metric <- .set_hyperparameter(
      default = distance_metric_default,
      type = "factor",
      range = distance_metric_default_range,
      valid_range = distance_metric_valid_range,
      randomise = distance_metric == ""
    )

    return(param)
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarKNN",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object,
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Find feature columns in the data.
    feature_columns <- get_feature_columns(x = data)

    # Parse formula
    formula <- stats::reformulate(
      termlabels = feature_columns,
      response = quote(outcome)
    )

    # Train model. Disable scaling because of bad interaction with
    # predicting single instances.
    model <- do.call_with_handlers(
      e1071::gknn,
      args = list(formula,
        "data" = data@data,
        "k" = object@hyperparameters$k,
        "method" = as.character(object@hyperparameters$distance_metric),
        "scale" = FALSE
      )
    )

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) {
      return(callNextMethod(object = object))
    }

    # Add model to the familiarModel object.
    object@model <- model

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarKNN",
    data = "dataObject"
  ),
  function(object, data, ...) {
    if (object@outcome_type %in% c("continuous", "binomial", "multinomial")) {
      # Turn into a naive model.
      object <- methods::new("familiarNaiveModel", object)
    }

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarKNN",
    data = "dataObject"
  ),
  function(object, data, type = "default", ...) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }
    
    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }
    
    if (type == "default") {
      # default ----------------------------------------------------------------

      if (object@outcome_type %in% c("binomial", "multinomial")) {
        # Binomial and multinomial outcomes-------------------------------------

        # Use the model to predict class probabilities.
        model_predictions <- predict(
          object = object@model,
          newdata = data@data,
          type = "prob"
        )

        # Obtain class levels and set class probabilities.
        class_levels <- get_outcome_class_levels(x = object)
        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          prediction_list[[class_levels[ii]]] <- model_predictions[, class_levels[ii]]
        }
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("continuous")) {
        # Continuous outcomes --------------------------------------------------

        # Use the model for prediction.
        model_predictions <- predict(
          object = object@model,
          newdata = data@data
        )

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = model_predictions,
          type = "regression",
          data = data,
          model_object = object
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      return(prediction_table)
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified method --------------------------------------------------
      
      # Use the model for prediction.
      return(predict(
        object = object@model,
        newdata = data@data,
        type = type,
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



# ..vimp -----------------------------------------------------------------------
# KNN does not have an associated variable importance method.


# .trim_model ------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarKNN"),
  function(object, ...) {
    # Note that knn by design always contains a copy of the dataset: it cannot
    # identify nearest neighbours otherwise.

    # Update model by removing the call.
    object@model$call <- call("trimmed")

    # Add show.
    object <- .capture_show(object)

    # Remove .Environment.
    object@model$terms <- .replace_environment(object@model$terms)

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    # Default method for models that lack a more specific method.
    return(object)
  }
)




.get_available_knn_learners <- function(
    show_general = TRUE,
    show_default = FALSE) {
  # General names without pre-specified distance metrics.
  general_learners <- c("k_nearest_neighbours", "knn")
  
  # Do not generate additional learners if proxy is not available.
  if (!requireNamespace("proxy", quietly = TRUE)) {
    return(general_learners)
  }
  
  # Define all distance metrics and the smaller default set.
  distance_metrics <- tolower(unname(unlist(
    lapply(
      proxy::pr_DB$get_entries(),
      function(list_elem) (list_elem$names)
    )
  )))
  
  distance_metrics_default <- c("euclidean", "manhattan", "gower")
  
  # Append distance metrics.
  all_learners <- c(sapply(general_learners, paste, distance_metrics, sep = "_"))
  default_learners <- c(sapply(general_learners, paste, distance_metrics_default, sep = "_"))
  
  # Add general learners, if needed.
  if (show_general) {
    all_learners <- c(general_learners, all_learners)
    default_learners <- c(general_learners, default_learners)
  }
  
  # Return only default learners -- this is for unit testing.
  if (show_default) {
    return(default_learners)
  }
  
  return(all_learners)
}

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# FDR-corrected p-values. Note that failure to predict survival probabilities
# does not influence the produced plots.
familiar:::test_plots(
  plot_function = familiar::plot_univariate_importance,
  data_element = "univariate_analysis",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list("verbose" = FALSE),
  debug = debug_flag
)

# Uncorrected p-values.
familiar:::test_plots(
  plot_function = familiar::plot_univariate_importance,
  data_element = "univariate_analysis",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "verbose" = FALSE,
    "p_adjustment_method" = "p_value"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar::plot_univariate_importance,
  data_element = "univariate_analysis",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "verbose" = FALSE,
    "p_adjustment_method" = "holm",
    "facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar::plot_univariate_importance,
  data_element = "univariate_analysis",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "verbose" = FALSE,
    "p_adjustment_method" = "BY",
    "facet_by" = c("data_set", "vimp_method"),
    "color_by" = "learner"),
  debug = debug_flag
)


# With a limit to the number of features shown.
familiar:::test_plots(
  plot_function = familiar::plot_univariate_importance,
  data_element = "univariate_analysis",
  test_config = "normal",
  plot_args = list("limit_n_features" = 2L),
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# promote_detector -------------------------------------------------------------
setMethod(
  "promote_detector",
  signature(object = "familiarNoveltyDetector"),
  function(object) {
    learner <- object@learner

    if (learner %in% .get_available_isolation_forest_detectors()) {
      # Isolation forest
      object <- methods::new("familiarIsolationForest", object)
      
    } else if (learner %in% .get_available_none_detectors()) {
      object <- methods::new("familiarNoneNoveltyDetector", object)
    }

    # Add package version.
    object <- add_package_version(object = object)

    # Returned object can be a standard familiarNoveltyDetector.
    return(object)
  }
)



.get_detector_hyperparameters <- function(
    data,
    detector,
    names_only = FALSE
) {
  # Create familiarNoveltyDetector
  fam_detector <- methods::new(
    "familiarNoveltyDetector",
    learner = detector
  )

  # Set up the specific model.
  fam_detector <- promote_detector(object = fam_detector)

  # Model hyperparameters
  detector_hyperparameters <- get_default_hyperparameters(
    object = fam_detector,
    data = data
  )

  # Extract names from parameter list
  if (names_only) detector_hyperparameters <- names(detector_hyperparameters)

  return(detector_hyperparameters)
}



.check_novelty_detector_available <- function(detector, as_flag = FALSE) {
  # Create familiarNoveltyDetector
  fam_detector <- methods::new(
    "familiarNoveltyDetector",
    learner = detector
  )

  # Set up the specific novelty detector
  fam_detector <- promote_detector(fam_detector)

  # Check validity.
  detector_available <- is_available(fam_detector)

  if (as_flag) return(detector_available)

  # Check if the familiar model has been successfully promoted.
  if (!is_subclass(class(fam_detector)[1L], "familiarNoveltyDetector")) {
    ..error(
      paste0(
        detector, " is not a valid novelty detector. ",
        "Please check the documentation for available novelty detectors."
      ),
      error_class = "input_argument_error"
    )
  }

  # Check that the required package can be loaded.
  require_package(
    x = fam_detector,
    purpose = paste0("to train a novelty detector using the ", detector, " algorithm"),
    message_type = "backend_error"
  )
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# test_train (generic) ---------------------------------------------------------
setGeneric(
  "test_train",
  function(data, ...) standardGeneric("test_train")
)



# test_train (data.table) ------------------------------------------------------
setMethod(
  "test_train",
  signature(data = "data.table"),
  function(
    data,
    data_bypass = NULL,
    learner,
    hyperparameter_list = list(),
    create_bootstrap = FALSE,
    ...
  ) {
    if (!is.null(data_bypass)) {
      # Convert data_bypass to dataObject.
      data_bypass <- do.call(
        as_data_object,
        args = c(
          list("data" = data_bypass),
          list(...)
        )
      )
    }

    # Convert data to dataObject.
    data <- do.call(
      as_data_object,
      args = c(
        list("data" = data),
        list(...)
      )
    )

    return(do.call(
      test_train, 
      args = c(
        list(
          "data" = data,
          "data_bypass" = data_bypass,
          "learner" = learner,
          "hyperparameter_list" = hyperparameter_list,
          "create_bootstrap" = create_bootstrap
        ),
        list(...)
      )
    ))
  }
)



# test_train (dataObject) ------------------------------------------------------
setMethod(
  "test_train",
  signature(data = "dataObject"),
  function(
    data,
    data_bypass = NULL,
    learner,
    hyperparameter_list = list(),
    create_bootstrap = FALSE,
    create_novelty_detector = FALSE,
    create_naive = FALSE,
    cl = NULL,
    trim_model = FALSE,
    verbose = FALSE,
    ...
  ) {
    # The bypass data allows for bypassing important aspects of the
    # pre-processing pipeline, e.g. the preprocessing checks. This enables
    # testing of very rare cases where preprocessing may run fine, but the
    # subsample does not allow for training.
    if (is.null(data_bypass)) data_bypass <- data
    
    # Prepare settings ---------------------------------------------------------
    
    # Reconstitute settings from the data.
    settings <- extract_settings_from_data(data_bypass)
    
    # Update some missing settings that can be fixed within this method.
    settings$data$train_cohorts <- unique(data_bypass@data[[get_id_columns(single_column = "batch")]])
    
    # Parse the remaining settings that are important. Remove outcome_type from
    # ... This prevents an error caused by multiple matching arguments.
    dots <- list(...)
    dots$parallel <- NULL
    dots$vimp_method <- NULL

    # Determine if a naive model should be forced.
    vimp_method <- ifelse(create_naive, "no_features", "none")
    
    settings <- do.call(
      .parse_general_settings,
      args = c(
        list(
          "settings" = settings,
          "data" = data_bypass@data,
          "parallel" = FALSE,
          "vimp_method" = vimp_method,
          "learner" = learner
        ),
        dots
      )
    )
    
    # Push settings to the backend as some functions require this.
    .assign_settings_to_global(settings = settings)
    
    
    # Prepare hyperparameters --------------------------------------------------
    
    # Get default hyperparameters.
    param_list <- .get_preset_hyperparameters(
      data = data,
      learner = learner,
      names_only = FALSE
    )
    
    # Update with user-provided settings.
    param_list <- .update_hyperparameters(
      parameter_list = param_list,
      user_list = hyperparameter_list
    )
    
    # Determine which hyperparameters still need to be specified.
    unset_parameters <- sapply(
      param_list,
      function(hyperparameter_entry) hyperparameter_entry$randomise
    )
    
    # Raise an error if any hyperparameters were not set.
    if (any(unset_parameters)) {
      ..error(paste0(
        "The following hyperparameters need to be specified: ",
        paste_s(names(unset_parameters)[unset_parameters])
      ))
    }
    
    # Obtain the final list of hyperparameters.
    param_list <- lapply(
      param_list, 
      function(hyperparameter_entry) hyperparameter_entry$init_config
    )
    
    
    # Create feature information list ------------------------------------------
    
    feature_info_task <- methods::new(
      "familiarTaskFeatureInfo"
    )
    
    # Feature information objects are created from the bypass dataset.
    feature_info <- .perform_task(
      object = feature_info_task,
      data = data_bypass,
      settings = settings,
      verbose = verbose
    )
    
    
    # Create learner -----------------------------------------------------------
    
    train_task <- methods::new(
      "familiarTaskTrain",
      vimp_method = vimp_method,
      learner = learner
    )
    
    if (create_bootstrap) {
      data <- select_data_from_samples(
        data = data,
        samples = fam_sample(
          x = data@data,
          replace = TRUE,
          seed = 19L
        )
      )
    }
    
    # Learners are trained using the actual data.
    object <- .perform_task(
      object = train_task,
      data = data,
      settings = settings,
      feature_info_list = feature_info,
      hyperparameters = param_list,
      novelty_detector = ifelse(create_novelty_detector, "isolation_forest", "none"),
      detector_parameters = NULL,
      trim_model = trim_model,
      verbose = verbose
    )
    
    return(object)
  }
)

# familiarModel object ---------------------------------------------------------

#' Familiar model.
#'
#' A familiarModel object is a self-contained model that can be applied to
#' generate predictions for a dataset. familiarModel objects form the parent
#' class of learner-specific child classes.
#'
#' @slot name Name of the familiarModel object.
#' @slot model The actual model trained using a specific algorithm, e.g. a
#'   random forest from the `ranger` package, or a LASSO model from `glmnet`.
#' @slot outcome_type Outcome type of the data used to create the object.
#' @slot outcome_info Outcome information object, which contains additional
#'   information concerning the outcome, such as class levels.
#' @slot feature_info List of objects containing feature information, e.g.,
#'   name, class levels, transformation, normalisation and clustering
#'   parameters.
#' @slot data_column_info Data information object containing information
#'   regarding identifier column names and outcome column names.
#' @slot hyperparameters Set of hyperparameters used to train the model.
#' @slot hyperparameter_data Information generated during hyperparameter
#'   optimisation.
#' @slot calibration_model One or more models used to recalibrate the model
#'   output. Currently only used by some models.
#' @slot novelty_detector A familiarNoveltyDetector object that can be used to
#'   detect out-of-distribution samples.
#' @slot learner Learning algorithm used to create the model.
#' @slot vimp_method Method used to determine variable importance for the model.
#' @slot vimp_table Variable importance table or list of variable importance
#'   tables for the model.
#' @slot vimp_aggregation_method Method used for aggregating variable importance
#'   tables if more than one is present.)
#' @slot vimp_rank_threshold Threshold used for some variable importance
#'   aggregation methods.
#' @slot required_features The set of features required for complete
#'   reproduction, i.e. with imputation.
#' @slot model_features The set of features that is used to train the model,
#' @slot novelty_features The set of features that is used to train all novelty
#'   detectors in the ensemble.
#' @slot calibration_info Calibration information, e.g. baseline survival in the
#'   development cohort.
#' @slot km_info Data concerning stratification into risk groups.
#' @slot data_id Internal identifier for the dataset used to train the model.
#' @slot run_id Internal identifier for the specific subset of the dataset used
#'   used to train the model.
#' @slot run_table Run table for the data used to train the model. Used
#'   internally.
#' @slot settings A copy of the evaluation configuration parameters used at
#'   model creation. These are used as default parameters when evaluating the
#'   model (technically, familiarEnsemble) to create a familiarData object.
#' @slot is_trimmed Flag that indicates whether the model, stored in the `model`
#'   slot, has been trimmed.
#' @slot trimmed_function List of functions whose output has been captured prior
#'   to trimming the model.
#' @slot messages List of warning and error messages generated during training.
#' @slot project_id Identifier of the project that generated the familiarModel
#'   object.
#' @slot familiar_version Version of the familiar package.
#' @slot package Name of package(s) required to executed the model itself, e.g.
#'   `ranger` or `glmnet`.
#' @slot package_version Version of the packages mentioned in the `package`
#'   attribute.
#'
#' @export

setClass("familiarModel",
  slots = list(
    # Model name.
    name = "character",
    # Model container
    model = "ANY",
    # Outcome type
    outcome_type = "character",
    # Outcome info, such as class levels, mean values etc.
    outcome_info = "ANY",
    # Data required for feature pre-processing
    feature_info = "ANY",
    # Info related to the columns in the dataset.
    data_column_info = "ANY",
    # Hyper-parameters (typically stored in the model as well)
    hyperparameters = "ANY",
    # Hyperparameter data, e.g. for visualising the hyperparameter space.
    hyperparameter_data = "ANY",
    # Models used for recalibration
    calibration_model = "ANY",
    # Model used for novelty detection
    novelty_detector = "ANY",
    # Name of learner
    learner = "character",
    # Name of variable importance method
    vimp_method = "character",
    # Variable importance table
    vimp_table = "ANY",
    # Vimp aggregation method.
    vimp_aggregation_method = "character",
    # Vimp rank threshold.
    vimp_rank_threshold = "integer",
    # Required features for complete reconstruction, including imputation.
    required_features = "ANY",
    # Features that are required for the model.
    model_features = "ANY",
    # Features that are required for novelty detection.
    novelty_features = "ANY",
    # data_id for the data used to train the model.
    data_id = "integer",
    # run_id for the data used to train the model.
    run_id = "integer",
    # Run table for the current model
    run_table = "ANY",
    # Information required to assess model calibrations (e.g. baseline survival)
    calibration_info = "ANY",
    # Information required to do perform a Kaplan-Meier analysis using the model
    km_info = "ANY",
    # Evaluation settings. This allows default values for external use of
    # existing models.
    settings = "ANY",
    # Flags trimming of the model
    is_trimmed = "logical",
    # Restores functions lost due to model trimming, such as coef or vcov.
    trimmed_function = "list",
    # List of warning and error messages encountered during training.
    messages = "list",
    # Project identifier for consistency tracking
    project_id = "ANY",
    # Package version for backward compatibility
    familiar_version = "ANY",
    # Name of the package required to train the learner.
    package = "ANY",
    # Version of the learner for reproducibility.
    package_version = "ANY"
  ),
  prototype = list(
    name = character(0L),
    model = NULL,
    outcome_type = NA_character_,
    outcome_info = NULL,
    feature_info = NULL,
    data_column_info = NULL,
    hyperparameters = NULL,
    hyperparameter_data = NULL,
    calibration_model = NULL,
    novelty_detector = NULL,
    learner = NA_character_,
    vimp_method = NA_character_,
    vimp_table = NULL,
    vimp_aggregation_method = NA_character_,
    vimp_rank_threshold = NA_integer_,
    required_features = NULL,
    model_features = NULL,
    novelty_features = NULL,
    calibration_info = NULL,
    km_info = NULL,
    data_id = NA_integer_,
    run_id = NA_integer_,
    run_table = NULL,
    settings = NULL,
    is_trimmed = FALSE,
    trimmed_function = list(),
    messages = list(),
    project_id = NULL,
    familiar_version = NULL,
    package = NULL,
    package_version = NULL
  )
)


# familiarEnsemble object ------------------------------------------------------

#' Ensemble of familiar models.
#'
#' A familiarEnsemble object contains one or more familiarModel objects.
#'
#' @slot name Name of the familiarEnsemble object.
#' @slot model_list List of attached familiarModel objects, or paths to these
#'   objects. Familiar attaches familiarModel objects when required.
#' @slot outcome_type Outcome type of the data used to create the object.
#' @slot outcome_info Outcome information object, which contains additional
#'   information concerning the outcome, such as class levels.
#' @slot data_column_info Data information object containing information
#'   regarding identifier column names and outcome column names.
#' @slot learner Learning algorithm used to create the models in the ensemble.
#' @slot vimp_method Method used to determine variable importance for the models
#'   in the ensemble.
#' @slot feature_info List of objects containing feature information, e.g.,
#'   name, class levels, transformation, normalisation and clustering
#'   parameters.
#' @slot required_features The set of features required for complete
#'   reproduction, i.e. with imputation.
#' @slot model_features The combined set of features that is used to train the
#'   models in the ensemble,
#' @slot novelty_features The combined set of features that is used to train all
#'   novelty detectors in the ensemble.
#' @slot data_id Internal identifier for the dataset used to train or
#'   evaluate the ensemble.
#' @slot run_id Internal identifier for the specific subset of the dataset used
#'   used to train or evaluate the ensemble.
#' @slot run_table Run table for the data used to train the ensemble. Used
#'   internally.
#' @slot calibration_info Calibration information, e.g. baseline survival in the
#'   development cohort.
#' @slot model_dir_path Path to folder containing the familiarModel objects. Can
#'   be updated using the `update_model_dir_path` method.
#' @slot auto_detach Flag used to determine whether models should be detached
#'   from the model after use, or not. Used internally.
#' @slot settings A copy of the evaluation configuration parameters used at
#'   model creation. These are used as default parameters when evaluating the
#'   ensemble to create a familiarData object.
#' @slot project_id Identifier of the project that generated the underlying
#'   familiarModel object(s).
#' @slot familiar_version Version of the familiar package.
#'
#' @export

setClass("familiarEnsemble",
  slots = list(
    # Ensemble name
    name = "character",
    # Model container.
    model_list = "ANY",
    # Model outcome type.
    outcome_type = "character",
    # Outcome info, such as class levels, mean values etc.
    outcome_info = "ANY",
    # Info related to the columns in the dataset.
    data_column_info = "ANY",
    # Name of learner.
    learner = "character",
    # Name of variable importance method.
    vimp_method = "character",
    # Data required for feature pre-processing.
    feature_info = "ANY",
    # Required features for complete reconstruction, including imputation.
    required_features = "ANY",
    # Features that are required for reconstruction, without imputation (i.e.
    # features that are in the signature directly or as part of a cluster)
    model_features = "ANY",
    # Features that are required for novelty detection.
    novelty_features = "ANY",
    # data_id for the data used to train the model.
    data_id = "integer",
    # run_id for the data used to train the model.
    run_id = "integer",
    # data_id for predictions. This forces the models to predict at the subsets
    # in this layer, and overrides data_id (but only for predictions).
    run_table = "ANY",
    # Information required to assess model calibrations (e.g. baseline survival)
    calibration_info = "ANY",
    # Path to the model directory. Required for auto-detaching.
    model_dir_path = "character",
    # Flag that signals auto-detaching. This means that models are loaded and
    # discarded one-by-one. This saves memory, but comes at the cost of IO
    # overhead. Moreover, its not possible if the models are not stored on drive
    # in the first place.
    auto_detach = "logical",
    # Evaluation settings. This allows default values for external use of
    # existing models.
    settings = "ANY",
    # Project identifier for consistency tracking.
    project_id = "ANY",
    # Package version for backward compatibility checks.
    familiar_version = "ANY"
  ),
  prototype = list(
    name = character(0L),
    model_list = NULL,
    outcome_type = NA_character_,
    outcome_info = NULL,
    data_column_info = NULL,
    learner = NA_character_,
    vimp_method = NA_character_,
    feature_info = NULL,
    required_features = NULL,
    model_features = NULL,
    novelty_features = NULL,
    data_id = NA_integer_,
    run_id = NA_integer_,
    run_table = NULL,
    calibration_info = NULL,
    model_dir_path = NA_character_,
    auto_detach = FALSE,
    settings = NULL,
    project_id = NULL,
    familiar_version = NULL
  )
)


setClassUnion(
  "familiarModelUnion",
  members = c("familiarModel", "familiarEnsemble")
)


# familiarData object ----------------------------------------------------------

#' Dataset obtained after evaluating models on a dataset.
#'
#' A familiarData object is created by evaluating familiarEnsemble or
#' familiarModel objects on a dataset. Multiple familiarData objects are
#' aggregated in a familiarCollection object.
#'
#' @slot name Name of the dataset, e.g. training or internal validation.
#' @slot outcome_type Outcome type of the data used to create the object.
#' @slot outcome_info Outcome information object, which contains additional
#'   information concerning the outcome, such as class levels.
#' @slot fs_vimp Data collected for variable importance methods.
#' @slot model_vimp Variable importance data collected from model-specific
#'   algorithms implemented by models created by familiar.
#' @slot permutation_vimp Data collected for permutation variable importance.
#' @slot hyperparameters Hyperparameters collected from created models.
#' @slot hyperparameter_data Additional data concerning hyperparameters. This is
#'   currently not used yet.
#' @slot required_features The set of features required for complete
#'   reproduction, i.e. with imputation.
#' @slot model_features The set of features that are required for using the
#'   model or ensemble of models, but without imputation.
#' @slot learner Learning algorithm used to create the model or ensemble of
#'   models.
#' @slot vimp_method Method used to determine variable importance for the model
#'   or ensemble of models.
#' @slot pooling_table Run table for the data underlying the familiarData
#'   object. Used internally.
#' @slot prediction_data Model predictions for a model or ensemble of models for
#'   the underlying dataset.
#' @slot confusion_matrix Confusion matrix for a model or ensemble of models,
#'   based on the underlying dataset.
#' @slot decision_curve_data Decision curve analysis data for a model or
#'   ensemble of models, based on the underlying dataset.
#' @slot calibration_info Calibration information, e.g. baseline survival in the
#'   development cohort.
#' @slot calibration_data Calibration data for a model or ensemble of models,
#'   based on the underlying dataset.
#' @slot model_performance Model performance data for a model or ensemble of
#'   models, based on the underlying dataset.
#' @slot km_info Information concerning risk-stratification cut-off values..
#' @slot km_data Kaplan-Meier survival data for a model or ensemble of models,
#'   based on the underlying dataset.
#' @slot auc_data AUC-ROC and AUC-PR data for a model or ensemble of models,
#'   based on the underlying dataset.
#' @slot ice_data Individual conditional expectation data for features included
#'   in a model or ensemble of models, based on the underlying dataset. Partial
#'   dependence data are computed on the fly from these data.
#' @slot shap_data SHAP values for features included in a model or ensemble of 
#'   models.
#' @slot univariate_analysis Univariate analysis of the underlying dataset.
#' @slot feature_expressions Feature expression values of the underlying
#'   dataset.
#' @slot feature_similarity Feature similarity information of the underlying
#'   dataset.
#' @slot sample_similarity Sample similarity information of the underlying
#'   dataset.
#' @slot project_id Identifier of the project that generated the familiarData
#'   object.
#' @slot familiar_version Version of the familiar package.
#'
#'   familiarData objects contain information obtained by evaluating a single
#'   model or single ensemble of models on a dataset.
#'
#' @export

setClass("familiarData",
  slots = list(
    # Name of the familiar data set
    name = "character",
    # Model outcome type
    outcome_type = "character",
    # Outcome info, such as class levels, mean values etc.
    outcome_info = "ANY",
    # Variable importance method
    fs_vimp = "ANY",
    # Model variable importance
    model_vimp = "ANY",
    # Permutation variable importance
    permutation_vimp = "ANY",
    # Hyper-parameters
    hyperparameters = "ANY",
    # Hyperparameter data, e.g. for visualising the hyperparameter space.
    hyperparameter_data = "ANY",
    # Required features to update the data
    required_features = "ANY",
    # Features that are required for reconstruction, without imputation (i.e.
    # features that are in the signature directly or as part of a cluster)
    model_features = "ANY",
    # Name of learner
    learner = "character",
    # Name of variable importance method
    vimp_method = "character",
    # Run table for the current data
    pooling_table = "ANY",
    # Model predictions for later reference
    prediction_data = "ANY",
    # Confusion matrix for categorical outcomes
    confusion_matrix = "ANY",
    # Data for decision curve analysis
    decision_curve_data = "ANY",
    # Calibration information, e.g. baseline survival
    calibration_info = "ANY",
    # Calibration test information
    calibration_data = "ANY",
    # Model performance metrics
    model_performance = "ANY",
    # Kaplan-Meier cut-offs
    km_info = "ANY",
    # Kaplan-Meier data
    km_data = "ANY",
    # AUC data (for plotting)
    auc_data = "ANY",
    # Information concerning the univariate importance of features
    univariate_analysis = "ANY",
    # Information concerning feature expression for individual samples
    feature_expressions = "ANY",
    # Information concerning mutual correlations between features
    feature_similarity = "ANY",
    # Information concerning similarity between samples.
    sample_similarity = "ANY",
    # Information on individual conditional expectation
    ice_data = "ANY",
    # Information on SHAP values
    shap_data = "ANY",
    # Project identifier
    project_id = "ANY",
    # Package version for backward compatibility
    familiar_version = "ANY"
  ),
  prototype = list(
    name = character(0L),
    outcome_type = NA_character_,
    outcome_info = NULL,
    fs_vimp = NULL,
    model_vimp = NULL,
    permutation_vimp = NULL,
    hyperparameters = NULL,
    hyperparameter_data = NULL,
    required_features = NULL,
    model_features = NULL,
    learner = NA_character_,
    vimp_method = NA_character_,
    pooling_table = NULL,
    prediction_data = NULL,
    confusion_matrix = NULL,
    decision_curve_data = NULL,
    calibration_info = NULL,
    calibration_data = NULL,
    model_performance = NULL,
    km_info = NULL,
    km_data = NULL,
    auc_data = NULL,
    univariate_analysis = NULL,
    feature_expressions = NULL,
    feature_similarity = NULL,
    sample_similarity = NULL,
    ice_data = NULL,
    shap_data = NULL,
    project_id = NULL,
    familiar_version = NULL
  )
)


# familiarCollection object ----------------------------------------------------

#' Collection of familiar data.
#'
#' A familiarCollection object aggregates data from one or more familiarData
#' objects.
#'
#' @slot name Name of the collection.
#' @slot data_sets Name of the individual underlying datasets.
#' @slot outcome_type Outcome type for which the collection was created.
#' @slot outcome_info Outcome information object, which contains information
#'   concerning the outcome, such as class levels.
#' @slot fs_vimp collected for variable importance methods.
#' @slot model_vimp Variable importance data collected from model-specific
#'   algorithms implemented by models created by familiar.
#' @slot permutation_vimp Data collected for permutation variable importance.
#' @slot hyperparameters Hyperparameters collected from created models.
#' @slot hyperparameter_data Additional data concerning hyperparameters. This is
#'   currently not used yet.
#' @slot required_features The set of features required for complete
#'   reproduction, i.e. with imputation.
#' @slot model_features The set of features that are required for using the
#'   model, but without imputation.
#' @slot learner Learning algorithm(s) used for data in the collection.
#' @slot vimp_method Variable importance method(s) used for data in the
#'   collection.
#' @slot prediction_data Model predictions for the data in the collection.
#' @slot confusion_matrix Confusion matrix information for the data in the
#'   collection.
#' @slot decision_curve_data Decision curve analysis data for the data in the
#'   collection.
#' @slot calibration_info Calibration information, e.g. baseline survival in the
#'   development cohort.
#' @slot calibration_data Model calibration data collected from data in the
#'   collection.
#' @slot model_performance Collection of model performance data for data in the
#'   collection.
#' @slot km_info Information concerning risk-stratification cut-off values for
#'   data in the collection.
#' @slot km_data Kaplan-Meier survival data for data in the collection.
#' @slot auc_data AUC-ROC and AUC-PR data for data in the collection.
#' @slot ice_data Individual conditional expectation data for data in the
#'   collection. Partial dependence data are computed on the fly from these
#'   data.
#' @slot shap_data SHAP values for features included in a model or ensemble of 
#'   models.
#' @slot univariate_analysis Univariate analysis results of data in the
#'   collection.
#' @slot feature_expressions Feature expression values for data in the
#'   collection.
#' @slot feature_similarity Feature similarity information for data in the
#'   collection.
#' @slot sample_similarity Sample similarity information for data in the
#'   collection.
#' @slot data_set_labels Labels for the different datasets in the collection.
#'   See `get_data_set_names` and `set_data_set_names`.
#' @slot learner_labels Labels for the different learning algorithms used to
#'   create the collection. See `get_learner_names` and `set_learner_names`.
#' @slot vimp_method_labels Labels for the different variable importance methods
#'   used to create the collection. See `get_vimp_method_names` and
#'   `set_vimp_method_names`.
#' @slot feature_labels Labels for the features in this collection. See
#'   `get_feature_names` and `set_feature_names`.
#' @slot km_group_labels Labels for the risk strata in this collection. See
#'   `get_risk_group_names` and `set_risk_group_names`.
#' @slot class_labels Labels of the response variable. See `get_class_names` and
#'   `set_class_names`.
#' @slot project_id Identifier of the project that generated this collection.
#' @slot familiar_version Version of the familiar package.
#'
#'   familiarCollection objects collect data from one or more familiarData
#'   objects. This objects are important, as all plotting and export functions
#'   use it. The fact that one can supply familiarModel, familiarEnsemble and
#'   familiarData objects as arguments for these methods, is because familiar
#'   internally converts these into familiarCollection objects prior to
#'   executing the method.
#'
#' @export

setClass("familiarCollection",
  slots = list(
    # Name of the collection
    name = "character",
    # Name of the underlying data sets
    data_sets = "character",
    # Model outcome type
    outcome_type = "character",
    # Outcome info, such as class levels, mean values etc.
    outcome_info = "ANY",
    # Feature selection variable importance
    fs_vimp = "ANY",
    # Model variable importance
    model_vimp = "ANY",
    # Permutation variable importance
    permutation_vimp = "ANY",
    # Hyper-parameters
    hyperparameters = "ANY",
    # Hyperparameter data, e.g. for visualising the hyperparameter space.
    hyperparameter_data = "ANY",
    # Required features to update the data
    required_features = "ANY",
    # Important features, e.g. features that ended up in a signature
    # individually or as part of a cluster
    model_features = "ANY",
    # Name of learner
    learner = "character",
    # Name of variable importance method
    vimp_method = "character",
    # Model predictions for later reference
    prediction_data = "ANY",
    # Confusion matrix for categorical outcomes
    confusion_matrix = "ANY",
    # Data for decision curve analysis
    decision_curve_data = "ANY",
    # Calibration information, e.g. baseline survival
    calibration_info = "ANY",
    # Calibration test information
    calibration_data = "ANY",
    # Model performance metrics
    model_performance = "ANY",
    # Kaplan-Meier cut-offs
    km_info = "ANY",
    # Kaplan-Meier data
    km_data = "ANY",
    # AUC data (for plotting)
    auc_data = "ANY",
    # Information concerning the univariate importance of features
    univariate_analysis = "ANY",
    # Information concerning feature expression for individual samples
    feature_expressions = "ANY",
    # Information concerning mutual correlations between features
    feature_similarity = "ANY",
    # Information concerning similarity between samples.
    sample_similarity = "ANY",
    # Information on individual conditional expectation
    ice_data = "ANY",
    # Information on SHAP values
    shap_data = "ANY",
    # Label and order of data names
    data_set_labels = "ANY",
    # Label and order of learners
    learner_labels = "ANY",
    # Label and order of variable importance methods
    vimp_method_labels = "ANY",
    # Label and order of features
    feature_labels = "ANY",
    # Label and order of kaplan-meier groups
    km_group_labels = "ANY",
    # Label and order of outcome classes
    class_labels = "ANY",
    # Project identifier
    project_id = "ANY",
    # Package version for backward compatibility
    familiar_version = "ANY"
  ),
  prototype = list(
    name = character(0L),
    data_sets = character(0L),
    outcome_type = NA_character_,
    outcome_info = NULL,
    fs_vimp = NULL,
    model_vimp = NULL,
    permutation_vimp = NULL,
    hyperparameters = NULL,
    hyperparameter_data = NULL,
    required_features = NULL,
    model_features = NULL,
    learner = NA_character_,
    vimp_method = NA_character_,
    prediction_data = NULL,
    confusion_matrix = NULL,
    decision_curve_data = NULL,
    calibration_info = NULL,
    calibration_data = NULL,
    model_performance = NULL,
    km_info = NULL,
    km_data = NULL,
    auc_data = NULL,
    univariate_analysis = NULL,
    feature_expressions = NULL,
    feature_similarity = NULL,
    sample_similarity = NULL,
    ice_data = NULL,
    shap_data = NULL,
    data_set_labels = NULL,
    learner_labels = NULL,
    vimp_method_labels = NULL,
    feature_labels = NULL,
    km_group_labels = NULL,
    class_labels = NULL,
    project_id = NULL,
    familiar_version = NULL
  )
)



# dataObject object ------------------------------------------------------------

#' Data object
#'
#' The dataObject class is used to resolve the issue of keeping track of
#' pre-processing status and data loading inside complex workflows, e.g. nested
#' predict functions inside a calibration function.
#'
#' @slot data NULL or data table containing the data. This is the data which
#'   will be read and used.
#' @slot preprocessing_level character indicating the level of pre-processing
#'   already conducted.
#' @slot outcome_type character, determines the outcome type.
#' @slot outcome_info Outcome information object, which contains additional
#'   information concerning the outcome, such as class levels.
#' @slot feature_info List of objects containing feature information, e.g.,
#'   name, class levels, transformation, normalisation and clustering
#'   parameters. Optional.
#' @slot data_column_info Object containing column information.
#' @slot data_id Data identifier for dataset. Set using internal routines if the
#'   `dataObject` was created from a `delayedDataObject`
#' @slot run_id Run identifier for dataset. Set using internal routines if the
#'   `dataObject` was created from a `delayedDataObject`
#' @slot validation Identifies if validation or development samples were loaded.
#'   Set using internal routines if the `dataObject` was created from a
#'   `delayedDataObject`.
#' @slot sample_seed Seed used for creating a bootstrap of the data.
setClass(
  "dataObject",
  slots = list(
    # Data
    data = "ANY",
    # Level to which pre-processing has been conducted.
    preprocessing_level = "character",
    # Outcome type
    outcome_type = "character",
    # Outcome info, such as class levels, mean values etc.
    outcome_info = "ANY",
    # Info related to features in the data.
    feature_info = "ANY",
    # Info related to the other columns in the dataset.
    data_column_info = "ANY",
    # Data id
    data_id = "integer",
    # Run id
    run_id = "integer",
    # Validation marker.
    validation = "logical",
    # Sample seed
    sample_seed = "integer"
  ),
  prototype = list(
    data = NULL,
    preprocessing_level = "none",
    outcome_type = NA_character_,
    outcome_info = NULL,
    feature_info = NULL,
    data_column_info = NULL,
    data_id = NA_integer_,
    run_id = NA_integer_,
    validation = NA,
    sample_seed = NA_integer_
  )
)



# delayedDataObject object -----------------------------------------------------

#' Data object with delayed loading
#'
#' The delayed loading object provides an interface to the backend data. This
#' data object is typically used within the evaluation pipeline to load data
#' when needed.
#'
#' @slot data NULL or data table containing the data. If present (not `NULL`),
#'   data is considered loaded. This should not happen -- load_data_object auto-
#'   matically creates a dataObject from the delayedDataObject.
#' @slot preprocessing_level character indicating the level of pre-processing
#'   already conducted. `"none"` by default.
#' @slot outcome_type character, determines the outcome type.
#' @slot outcome_info Outcome information object, which contains additional
#'   information concerning the outcome, such as class levels.
#' @slot feature_info List of objects containing feature information, e.g.,
#'   name, class levels, transformation, normalisation and clustering
#'   parameters. Optional.
#' @slot data_column_info Object containing column information.
#' @slot data_id integer. Defines the data_id of the dataset that should be
#'   loaded.
#' @slot run_id integer. Defines the run_id of the dataset that should be load.
#'   Together with data_id, run_id and validation allows for looking up the
#'   sample set. If run_id is left unset (NA_integer_), this will force the
#'   run_id to be set using the model, vimp_method or ensemble object. This is
#'   used during the evaluation process to load data specifically related to
#'   training, internal validation and external validation. The run-tables
#'   (which contain information about data partitioning) associated with these
#'   objects are used to look-up the run_id based on the data_id (that is always
#'   explicitly set). The perform_task method for familiarTaskEvaluate uses this
#'   aspect explicitly.
#' @slot validation logical. This determines which internal data set will be
#'   loaded. If TRUE, the validation data will be loaded, whereas FALSE loads
#'   the development data.
#' @slot aggregate_on_load logical. Determines whether data is aggregated after
#'   loading.
#' @slot sample_set_on_load NULL or vector of sample identifiers to be loaded.
#'   Overrides any `sample_seed` that may have been provided.
setClass(
  "delayedDataObject",
  contains = "dataObject",
  slots = list(
    # Flag for aggregation after loading and pre-processing
    aggregate_on_load = "logical",
    # Samples to be loaded. 
    sample_set_on_load = "ANY"
  ),
  prototype = list(
    aggregate_on_load = NA,
    sample_set_on_load = NULL
  )
)



# featureInfo object -----------------------------------------------------------

#' Feature information object.
#'
#' A featureInfo object contains information for a single feature. This
#' information is used to check data prospectively for consistency and for data
#' preparation. These objects are, for instance, attached to a familiarModel
#' object so that data can be pre-processed in the same way as the development
#' data.
#'
#' @slot name Name of the feature, which by default is the column name of the
#'   feature.
#' @slot set_descriptor Character string describing the set to which the feature
#'   belongs. Currently not used.
#' @slot feature_type Describes the feature type, i.e. `factor` or `numeric`.
#' @slot levels The class levels of categorical features. This is used to check
#'   prospective datasets.
#' @slot ordered Specifies whether the
#' @slot distribution Five-number summary (numeric) or class frequency
#'   (categorical).
#' @slot data_id Internal identifier for the dataset used to derive the feature
#'   information.
#' @slot run_id Internal identifier for the specific subset of the dataset used
#'   to derive the feature information.
#' @slot in_signature Specifies whether the feature is included in the model
#'   signature.
#' @slot in_novelty Specifies whether the feature is included in the novelty
#'   detector.
#' @slot removed Specifies whether the feature was removed during
#'   pre-processing.
#' @slot removed_unknown_type Specifies whether the feature was removed during
#'   pre-processing because the type was neither factor nor numeric..
#' @slot removed_missing_values Specifies whether the feature was removed during
#'   pre-processing because it contained too many missing values.
#' @slot removed_no_variance Specifies whether the feature was removed during
#'   pre-processing because it did not contain more than 1 unique value.
#' @slot removed_low_variance Specifies whether the feature was removed during
#'   pre-processing because the variance was too low. Requires applying
#'   `low_variance` as a `filter_method`.
#' @slot removed_low_robustness Specifies whether the feature was removed during
#'   pre-processing because it lacks robustness. Requires applying `robustness`
#'   as a `filter_method`, as well as repeated measurement.
#' @slot removed_low_importance Specifies whether the feature was removed during
#'   pre-processing because it lacks relevance. Requires applying
#'   `univariate_test` as a `filter_method`.
#' @slot fraction_missing Specifies the fraction of missing values.
#' @slot robustness Specifies robustness of the feature, if measured.
#' @slot univariate_importance Specifies the univariate p-value of the feature,
#'   if measured.
#' @slot transformation_parameters Details parameters for power transformation
#'   of numeric features.
#' @slot normalisation_parameters Details parameters for (global) normalisation
#'   of numeric features.
#' @slot batch_normalisation_parameters Details parameters for batch
#'   normalisation of numeric features.
#' @slot imputation_parameters Details parameters or models for imputation of
#'   missing values.
#' @slot cluster_parameters Details parameters for forming clusters with other
#'   features.
#' @slot required_features Details features required for clustering or
#'   imputation.
#' @slot project_id Identifier of the project that generated this collection.
#' @slot familiar_version Version of the familiar package.
#'
#' @export

setClass("featureInfo",
  slots = list(
    name = "character",
    set_descriptor = "character",
    feature_type = "character",
    levels = "ANY",
    ordered = "logical",
    distribution = "ANY",
    data_id = "integer",
    run_id = "integer",
    in_signature = "logical",
    in_novelty = "logical",
    removed = "logical",
    removed_unknown_type = "logical",
    removed_missing_values = "logical",
    removed_no_variance = "logical",
    removed_low_variance = "logical",
    removed_low_robustness = "logical",
    removed_low_importance = "logical",
    fraction_missing = "numeric",
    robustness = "ANY",
    univariate_importance = "ANY",
    transformation_parameters = "ANY",
    normalisation_parameters = "ANY",
    batch_normalisation_parameters = "ANY",
    imputation_parameters = "ANY",
    cluster_parameters = "ANY",
    required_features = "ANY",
    project_id = "ANY",
    familiar_version = "ANY"
  ),
  prototype = list(
    name = NA_character_,
    set_descriptor = NA_character_,
    feature_type = NA_character_,
    levels = NULL,
    ordered = FALSE,
    distribution = NULL,
    data_id = NA_integer_,
    run_id = NA_integer_,
    in_signature = FALSE,
    in_novelty = FALSE,
    removed = FALSE,
    removed_unknown_type = FALSE,
    removed_missing_values = FALSE,
    removed_no_variance = FALSE,
    removed_low_variance = FALSE,
    removed_low_robustness = FALSE,
    removed_low_importance = FALSE,
    fraction_missing = NA_real_,
    robustness = NULL,
    univariate_importance = NULL,
    transformation_parameters = NULL,
    normalisation_parameters = NULL,
    batch_normalisation_parameters = NULL,
    imputation_parameters = NULL,
    cluster_parameters = NULL,
    required_features = NULL,
    project_id = NULL,
    familiar_version = NULL
  )
)



# featureInfoParameters object -------------------------------------------------

#' Feature information parameters object.
#'
#' A featureInfo object contains information for a single feature. Some
#' information, for example concerning clustering and transformation contains
#' various parameters that allow for applying the data transformation correctly.
#' These are stored in featureInfoParameters objects.
#'
#' @slot name Name of the feature, which by default is the column name of the
#'   feature. Typically used to correctly assign the data.
#' @slot complete Flags whether the parameters have been completely set.
#' @slot familiar_version Version of the familiar package.
#'
#' @details featureInfoParameters is normally a parent class for specific
#'   classes, such as featureInfoParametersTransformation.
#'
#' @export
setClass("featureInfoParameters",
  slots = list(
    name = "character",
    complete = "logical",
    familiar_version = "ANY"
  ),
  prototype = list(
    name = NA_character_,
    complete = FALSE,
    familiar_version = NULL
  )
)



# vimpTable object -------------------------------------------------------------

#' Variable importance table
#'
#' A vimpTable object contains information concerning variable importance of one
#' or more features. These objects are created during variable importance
#' computation..
#'
#' @slot vimp_table Table containing features with corresponding scores.
#' @slot vimp_method Method used to compute variable importance scores for each
#'   feature.
#' @slot data_id Internal identifier for the dataset used to derive the variable
#'   importance table.
#' @slot run_id Internal identifier for the specific subset of the dataset used
#'   to derive the variable importance table.
#' @slot run_table Run table for the data used to compute variable importances
#'   from. Used internally.
#' @slot score_aggregation Method used to aggregate the score of contrasts for
#'   each categorical feature, if any,
#' @slot encoding_table Table used to relate categorical features to their
#'   contrasts, if any. Not used for all variable importance methods.
#' @slot cluster_table Table used to relate original features with features
#'   after clustering. Variable importance is determined after feature
#'   processing, which includes clustering.
#' @slot invert Determines whether increasing score corresponds to increasing
#'   (`FALSE`) or decreasing rank (`TRUE`). Used internally to determine how
#'   ranks should be formed.
#' @slot project_id Identifier of the project that generated the vimpTable
#'   object.
#' @slot familiar_version Version of the familiar package used to create this
#'   table.
#' @slot state State of the variable importance table. The object can have the
#'   following states:
#'
#'   * `initial`: initial state, directly after the variable importance table is
#'   filled.
#'
#'   * `decoded`: depending on the variable importance method, the initial
#'   variable importance table may contain the scores of individual contrasts
#'   for categorical variables. When decoded, data in the `encoding_table`
#'   attribute has been used to aggregate scores from all contrasts into a
#'   single score for each feature.
#'
#'   * `declustered`: variable importance is determined from fully processed
#'   features, which includes clustering. This means that a single feature in
#'   the variable importance table may represent multiple original features.
#'   When a variable importance table has been declustered, all clusters have
#'   been turned into their constituent features.
#'
#'   * `reclustered`: When the table is reclustered, features are replaced by
#'   their respective clusters. This is actually used when updating the cluster
#'   table to ensure it fits to a local context. This prevents issues when
#'   attempting to aggregate or apply variable importance tables in data with
#'   different feature preprocessing, and as a result, different clusters.
#'
#'   * `ranked`: The scores have been used to create ranks, with lower ranks
#'   indicating better features.
#'
#'   * `aggregated`: Score and ranks from multiple variable importance tables
#'   were aggregated.
#'
#' @details vimpTable objects exists in various states. These states are
#'   generally incremental, i.e. one cannot turn a declustered table into the
#'   initial version. Some methods such as aggregation internally do some state
#'   reshuffling.
#'
#'   This object replaces the ad-hoc lists with information that were used in
#'   versions prior to familiar 1.2.0.
#' @seealso \code{\link{get_vimp_table}}, \code{\link{aggregate_vimp_table}}
#' @export

setClass("vimpTable",
  slots = list(
    # Variable importance table.
    vimp_table = "ANY",
    # Variable importance method that generated the current variable
    # importance table.
    vimp_method = "character",
    data_id = "integer",
    run_id = "integer",
    # Run table for the current table.
    run_table = "ANY",
    # Set how scores from encoded features should be aggregated.
    score_aggregation = "character",
    # Table that can be used to merge encoded features back into
    # singleton features, if necessary.
    encoding_table = "ANY",
    # Table that can be used to decluster the current table.
    cluster_table = "ANY",
    # Whether scores should be inverted for ranking.
    invert = "logical",
    # Project identifier.
    project_id = "ANY",
    # Version of familiar used to create the object.
    familiar_version = "ANY",
    # State of the object.
    state = "character"
  ),
  prototype = list(
    vimp_table = NULL,
    vimp_method = NA_character_,
    data_id = NA_integer_,
    run_id = NA_integer_,
    run_table = NULL,
    score_aggregation = NA_character_,
    encoding_table = NULL,
    cluster_table = NULL,
    invert = FALSE,
    project_id = NULL,
    familiar_version = NULL,
    state = "initial"
  )
)



# outcomeInfo object -----------------------------------------------------------

#' Outcome information object.
#'
#' An outcome information object stores data concerning an outcome. This is used
#' to prospectively check data.
#'
#' @slot name Name of the outcome, inherited from the original column name by
#'   default.
#' @slot outcome_type Type of outcome.
#' @slot outcome_column Name of the outcome column in data.
#' @slot levels Specifies class levels of categorical outcomes.
#' @slot ordered Specifies whether categorical outcomes are ordered.
#' @slot reference Class level used as reference.
#' @slot time Maximum time, as set by the `time_max` configuration parameter.
#' @slot censored Censoring indicators for survival outcomes.
#' @slot event Event indicators for survival outcomes.
#' @slot competing_risk Indicators for competing risks in survival outcomes.
#' @slot distribution Five-number summary (numeric outcomes), class frequency
#'   (categorical outcomes), or survival distributions.
#' @slot data_id Internal identifier for the dataset used to derive the outcome
#'   information.
#' @slot run_id Internal identifier for the specific subset of the dataset used
#'   to derive the outcome information.
#' @slot transformation_parameters Parameters used for transforming a numeric
#'   outcomes. Currently unused.
#' @slot normalisation_parameters Parameters used for normalising numeric
#'   outcomes. Currently unused.
#' @slot familiar_version Version of the familiar package used to create this
#'   object.
#'
#' @export

setClass("outcomeInfo",
  slots = list(
    # Name of the outcome
    name = "character",
    # Outcome type
    outcome_type = "character",
    # Outcome column
    outcome_column = "ANY",
    # Class levels of categorical outcomes.
    levels = "ANY",
    # Flag for ordinal categorical outcomes.
    ordered = "logical",
    # Reference class of categorical outcomes.
    reference = "ANY",
    # Max time for the outcome.
    time = "numeric",
    # Censor indicator for survival outcomes, e.g. alive.
    censored = "character",
    # Event indicator for survival outcomes, e.g. recurrent disease.
    event = "character",
    # Competing risk indicator(s) for survival outcomes, e.g. dead.
    competing_risk = "character",
    # Distribution information of outcome variables.
    distribution = "ANY",
    # Data id to which this outcome data belongs.
    data_id = "integer",
    # Run id to which this outcome data belongs.
    run_id = "integer",
    # Transformation parameters for the outcome data.
    transformation_parameters = "ANY",
    # Normalisation parameters for the outcome data.
    normalisation_parameters = "ANY",
    # Version of familiar used to create the object.
    familiar_version = "ANY"
  ),
  prototype = list(
    name = NA_character_,
    outcome_type = NA_character_,
    outcome_column = NULL,
    levels = NULL,
    ordered = FALSE,
    reference = NA_character_,
    time = Inf,
    censored = NA_character_,
    event = NA_character_,
    competing_risk = NA_character_,
    distribution = NULL,
    data_id = NA_integer_,
    run_id = NA_integer_,
    transformation_parameters = NULL,
    normalisation_parameters = NULL,
    familiar_version = NULL
  )
)



# familiarVimpMethod object ----------------------------------------------------

#' Variable importance method object.
#'
#' The familiarVimpMethod class is the parent class for all variable importance
#' methods in familiar.
#'
#' @slot outcome_type Outcome type of the data to be evaluated using the object.
#' @slot hyperparameters Set of hyperparameters for the variable importance
#'   method.
#' @slot vimp_method The character string indicating the variable importance
#'   method.
#' @slot vimp_aggregation_method Method used for aggregating variable importance
#'   tables if more than one is present.)
#' @slot vimp_rank_threshold Threshold used for some variable importance
#'   aggregation methods.
#' @slot multivariate Flags whether the variable importance method is
#'   multivariate vs. univariate.
#' @slot outcome_info Outcome information object, which contains additional
#'   information concerning the outcome, such as class levels.
#' @slot feature_info List of objects containing feature information, e.g.,
#'   name, class levels, transformation, normalisation and clustering
#'   parameters.
#' @slot required_features The set of features to be assessed by the variable
#'   importance method.
#' @slot package Name of the package(s) required to execute the variable
#'   importance method.
#' @slot run_table Run table for the data to be assessed by the variable
#'   importance method. Used internally.
#' @slot project_id Identifier of the project that generated the
#'   familiarVimpMethod object.
#' @slot familiar_version Version of the familiar package used to create this
#'   object.
#'
#' @export
setClass("familiarVimpMethod",
  slots = list(
    # Outcome type
    outcome_type = "character",
    # Hyper-parameters (typically stored in the model as well)
    hyperparameters = "ANY",
    # Name of variable importance method
    vimp_method = "character",
    # Vimp aggregation method.
    vimp_aggregation_method = "character",
    # Vimp rank threshold.
    vimp_rank_threshold = "integer",
    # Indicates whether the method is a univariate or multivariate
    # method.
    multivariate = "logical",
    # Outcome info, such as class levels, mean values etc.
    outcome_info = "ANY",
    # Data required for feature pre-processing
    feature_info = "ANY",
    # Required features for complete reconstruction, including imputation
    required_features = "ANY",
    # Name of the package required to perform variable importance.
    package = "ANY",
    # Run table for the current vimp method
    run_table = "ANY",
    # Project identifier for consistency tracking
    project_id = "ANY",
    # Version of familiar used to create the object.
    familiar_version = "ANY"
  ),
  prototype = list(
    outcome_type = NA_character_,
    hyperparameters = NULL,
    vimp_method = NA_character_,
    vimp_aggregation_method = NA_character_,
    vimp_rank_threshold = NA_integer_,
    multivariate = FALSE,
    outcome_info = NULL,
    feature_info = NULL,
    required_features = NULL,
    package = NULL,
    run_table = NULL,
    project_id = NULL,
    familiar_version = NULL
  )
)



# familiarNoveltyDetector object -----------------------------------------------

#' Novelty detector.
#'
#' A familiarNoveltyDetector object is a self-contained model that can be
#' applied to generate out-of-distribution predictions for instances in a
#' dataset.
#'
#' @slot name Name of the familiarNoveltyDetector object.
#' @slot model The actual novelty detector trained using a specific algorithm,
#'   e.g. a isolation forest from the `isotree` package.
#' @slot feature_info List of objects containing feature information, e.g.,
#'   name, class levels, transformation, normalisation and clustering
#'   parameters.
#' @slot data_column_info Data information object containing information
#'   regarding identifier column names.
#' @slot hyperparameters Set of hyperparameters used to train the detector.
#' @slot hyperparameter_data Information generated during hyperparameter
#'   optimisation. Currently not used.
#' @slot calibration_model Model used to convert raw output to statistical
#'   probability of being out-of-distribution. Currently not used.
#' @slot learner Learning algorithm used to create the novelty detector.
#' @slot vimp_method Method used to determine variable importance for the
#'   novelty detector.
#' @slot vimp_table Variable importance table or list of variable importance
#'   tables for the model.
#' @slot vimp_aggregation_method Method used for aggregating variable importance
#'   tables if more than one is present.)
#' @slot vimp_rank_threshold Threshold used for some variable importance
#'   aggregation methods.
#' @slot required_features The set of features required for complete
#'   reproduction, i.e. with imputation.
#' @slot model_features The set of features that is used to train the detector.
#' @slot data_id Internal identifier for the dataset used to train the detector.
#' @slot run_id Internal identifier for the specific subset of the dataset used
#'   used to train the detector.
#' @slot run_table Run table for the data used to train the detector. Used
#'   internally.
#' @slot is_trimmed Flag that indicates whether the detector, stored in the
#'   `model` slot, has been trimmed.
#' @slot trimmed_function List of functions whose output has been captured prior
#'   to trimming the model.
#' @slot messages List of warning and error messages generated during training.
#' @slot project_id Identifier of the project that generated the
#'   familiarNoveltyDetector object.
#' @slot familiar_version Version of the familiar package.
#' @slot package Name of package(s) required to executed the detector itself,
#'   e.g. `isotree`.
#' @slot package_version Version of the packages mentioned in the `package`
#'   attribute.
#'
#' @details Note that these objects do not contain any data concerning outcome,
#'   as this not relevant for (prospective) out-of-distribution detection.
#'
#' @export

setClass("familiarNoveltyDetector",
  slots = list(
    
    # Model name.
    name = "character",
    # Model container
    model = "ANY",
    # Data required for feature pre-processing
    feature_info = "ANY",
    # Info related to the columns in the dataset.
    data_column_info = "ANY",
    # Hyper-parameters (typically stored in the model as well)
    hyperparameters = "ANY",
    # Hyperparameter data, e.g. for visualising the hyperparameter space.
    hyperparameter_data = "ANY",
    # Models used for recalibration
    calibration_model = "ANY",
    # Name of learner
    learner = "character",
    # Name of variable importance method
    vimp_method = "character",
    # Variable importance table
    vimp_table = "ANY",
    # Vimp aggregation method.
    vimp_aggregation_method = "character",
    # Vimp rank threshold.
    vimp_rank_threshold = "integer",
    # Required features for complete reconstruction, including imputation.
    required_features = "ANY",
    # Features that are required for the model.
    model_features = "ANY",
    # data_id for the data used to train the model.
    data_id = "integer",
    # run_id for the data used to train the model.
    run_id = "integer",
    # Run table for the current model
    run_table = "ANY",
    # Flags trimming of the model
    is_trimmed = "logical",
    # Restores functions lost due to model trimming, such as coef or vcov.
    trimmed_function = "list",
    # List of warning and error messages encountered during training.
    messages = "list",
    # Project identifier for consistency tracking
    project_id = "ANY",
    # Package version for backward compatibility
    familiar_version = "ANY",
    # Name of the package required to train the learner.
    package = "ANY",
    # Version of the learner for reproducibility.
    package_version = "ANY"
  ),
  prototype = list(
    name = character(0L),
    model = NULL,
    feature_info = NULL,
    data_column_info = NULL,
    hyperparameters = NULL,
    hyperparameter_data = NULL,
    calibration_model = NULL,
    learner = NA_character_,
    vimp_method = NA_character_,
    vimp_table = NULL,
    vimp_aggregation_method = NA_character_,
    vimp_rank_threshold = NA_integer_,
    required_features = NULL,
    model_features = NULL,
    data_id = NA_integer_,
    run_id = NA_integer_,
    run_table = NULL,
    is_trimmed = FALSE,
    trimmed_function = list(),
    messages = list(),
    project_id = NULL,
    familiar_version = NULL,
    package = NULL,
    package_version = NULL
  )
)



# familiarHyperparameterLearner object -----------------------------------------

#' Hyperparameter learner.
#'
#' A familiarHyperparameterLearner object is a self-contained model that can be
#' applied to predict optimisation scores for a set of hyperparameters.
#'
#' @slot name Name of the familiarHyperparameterLearner object.
#' @slot learner Algorithm used to create the hyperparameter learner.
#' @slot target_learner Algorithm for which the hyperparameters are being
#'   learned.
#' @slot target_outcome_type Outcome type of the learner for which
#'   hyperparameters are being modeled. Used to determine the target
#'   hyperparameters.
#' @slot optimisation_metric One or metrics used to generate the optimisation
#'   score.
#' @slot optimisation_function Function used to generate the optimisation score.
#' @slot model The actual model trained using the specific algorithm, e.g. a
#'   isolation forest from the `isotree` package.
#' @slot target_hyperparameters The names of the hyperparameters that are used
#'   to train the hyperparameter learner.
#' @slot project_id Identifier of the project that generated the
#'   familiarHyperparameterLearner object.
#' @slot familiar_version Version of the familiar package.
#' @slot package Name of package(s) required to executed the hyperparameter
#'   learner itself, e.g. `laGP`.
#' @slot package_version Version of the packages mentioned in the `package`
#'   attribute.
#'
#' @details Hyperparameter learners are used to infer the optimisation score for
#'   sets of hyperparameters. These are then used to either infer utility using
#'   acquisition functions or to generate summary scores to identify the optimal
#'   model.
#'
#' @export

setClass("familiarHyperparameterLearner",
  slots = list(
    # Model name.
    name = "character",
    # Hyperparameter learner
    learner = "character",
    # Learner for which the hyperparameters are being learned.
    target_learner = "character",
    # Outcome type for the above learner.
    target_outcome_type = "character",
    # Metric(s) used to generate the input data for optimisation score
    # that is being learned.
    optimisation_metric = "character",
    # Function used to generate the optimisation score.
    optimisation_function = "character",
    # Model container
    model = "ANY",
    # Names of the hyperparameters that are being learned.
    target_hyperparameters = "ANY",
    # Project identifier for consistency tracking
    project_id = "ANY",
    # Package version for backward compatibility.
    familiar_version = "ANY",
    # Name of the package required to train the learner.
    package = "ANY",
    # Version of the learner for reproducibility.
    package_version = "ANY"
  ),
  prototype = list(
    name = character(0L),
    learner = NA_character_,
    target_learner = NA_character_,
    target_outcome_type = NA_character_,
    optimisation_metric = NA_character_,
    optimisation_function = NA_character_,
    model = NULL,
    target_hyperparameters = NULL,
    project_id = NULL,
    familiar_version = NULL,
    package = NULL,
    package_version = NULL
  )
)



# familiarMetric object --------------------------------------------------------

#' Model performance metric.
#'
#' Superclass for model performance objects.
#'
#' @slot metric Performance metric.
#' @slot outcome_type Type of outcome being predicted.
#' @slot name Name of the performance metric.
#' @slot value_range Range of the performance metric. Can be half-open.
#' @slot baseline_value Value of the metric for trivial models, e.g. models that
#'   always predict the median value, the majority class, or the mean hazard,
#'   etc.
#' @slot higher_better States whether higher metric values correspond to better
#'   predictive model performance (e.g. accuracy) or not (e.g. root mean squared
#'   error).
#'
#' @export

setClass(
  "familiarMetric",
  slots = list(
    # The metric itself.
    metric = "character",
    # The outcome type associated with the metric.
    outcome_type = "character",
    # The name of the metric, e.g. for plotting.
    name = "character",
    # The potential value range of the metric.
    value_range = "numeric",
    # The baseline value of the metric, e.g. to derive an objective
    # function from.
    baseline_value = "ANY",
    # Flag that sets whether higher values denote better performance.
    higher_better = "logical"
  ),
  prototype = list(
    metric = NA_character_,
    outcome_type = NA_character_,
    name = NA_character_,
    value_range = c(NA_real_, NA_real_),
    baseline_value = NULL,
    higher_better = TRUE
  )
)



# familiarDataElement object ---------------------------------------------------

#' Data container for evaluation data.
#'
#' Most attributes of the familiarData object are objects of the
#' familiarDataElement class. This (super-)class is used to allow for
#' standardised aggregation and processing of evaluation data.
#'
#' @slot data Evaluation data, typically a data.table or list.
#' @slot identifiers Identifiers of the data, e.g. the generating model name,
#'   learner, etc.
#' @slot detail_level Sets the level at which results are computed and
#'   aggregated.
#'
#'  * `ensemble`: Results are computed at the ensemble level, i.e. over all
#'   models in the ensemble. This means that, for example, bias-corrected
#'   estimates of model performance are assessed by creating (at least) 20
#'   bootstraps and computing the model performance of the ensemble model for
#'   each bootstrap.
#'
#'  * `hybrid` (default): Results are computed at the level of models in an
#'   ensemble. This means that, for example, bias-corrected estimates of model
#'   performance are directly computed using the models in the ensemble. If
#'   there are at least 20 trained models in the ensemble, performance is
#'   computed for each model, in contrast to `ensemble` where performance is
#'   computed for the ensemble of models. If there are less than 20 trained
#'   models in the ensemble, bootstraps are created so that at least 20 point
#'   estimates can be made.
#'
#'  * `model`: Results are computed at the model level. This means that, for
#'   example, bias-corrected estimates of model performance are assessed by
#'   creating (at least) 20 bootstraps and computing the performance of the
#'   model for each bootstrap.
#'
#'   Note that each level of detail has a different interpretation for bootstrap
#'   confidence intervals. For `ensemble` and `model` these are the confidence
#'   intervals for the ensemble and an individual model, respectively. That is,
#'   the confidence interval describes the range where an estimate produced by a
#'   respective ensemble or model trained on a repeat of the experiment may be
#'   found with the probability of the confidence level. For `hybrid`, it
#'   represents the range where any single model trained on a repeat of the
#'   experiment may be found with the probability of the confidence level. By
#'   definition, confidence intervals obtained using `hybrid` are at least as
#'   wide as those for `ensemble`. `hybrid` offers the correct interpretation if
#'   the goal of the analysis is to assess the result of a single, unspecified,
#'   model.
#'
#'   Some child classes do not use this parameter.
#' @slot estimation_type Sets the type of estimation that should be possible.
#'   This has the following options:
#'
#'  * `point`: Point estimates.
#'
#'  * `bias_correction` or `bc`: Bias-corrected estimates. A bias-corrected
#'   estimate is computed from (at least) 20 point estimates, and `familiar` may
#'   bootstrap the data to create them.
#'
#'  * `bootstrap_confidence_interval` or `bci` (default): Bias-corrected
#'   estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
#'   number of point estimates required depends on the `confidence_level`
#'   parameter, and `familiar` may bootstrap the data to create them.
#'
#'   Some child classes do not use this parameter.
#' @slot confidence_level (*optional*) Numeric value for the level at which
#'   confidence intervals are determined. In the case bootstraps are used to
#'   determine the confidence intervals bootstrap estimation, `familiar` uses
#'   the rule of thumb \eqn{n = 20 / ci.level} to determine the number of
#'   required bootstraps.
#' @slot bootstrap_ci_method Method used to determine bootstrap confidence
#'   intervals (Efron and Hastie, 2016). The following methods are implemented:
#'
#'  * `percentile` (default): Confidence intervals obtained using the percentile
#'   method.
#'
#'  * `bc`: Bias-corrected confidence intervals.
#'
#'   Note that the standard method is not implemented because this method is
#'   often not suitable due to non-normal distributions. The bias-corrected and
#'   accelerated (BCa) method is not implemented yet.
#' @slot value_column Identifies column(s) in the `data` attribute presenting
#'   values.
#' @slot grouping_column Identifies column(s) in the `data` attribute presenting
#'   identifier columns for grouping during aggregation. Familiar will
#'   automatically assign items from the `identifiers` attribute to the data and
#'   this attribute when combining multiple familiarDataElements of the same
#'   (child) class.
#' @slot is_aggregated Defines whether the object was aggregated.
#'
#' @references 1. Efron, B. & Hastie, T. Computer Age Statistical Inference.
#'   (Cambridge University Press, 2016).
#'
#' @export

setClass(
  "familiarDataElement",
  slots = list(
    # The primary results.
    data = "ANY",
    # Identifiers of the data, e.g. the generating model name, the variable
    # importance method and learner.
    identifiers = "ANY",
    # The level of detail at which the data was computed.
    detail_level = "character",
    # The kind of estimation for which the data was computed, e.g.
    # bias-corrected estimates.
    estimation_type = "character",
    # The confidence level for which data was computed. Only set if the
    # correct estimation type is set.
    confidence_level = "ANY",
    # The method used to compute the bootstrap confidence intervals from
    # the data.
    bootstrap_ci_method = "character",
    # The column that contains the relevant data. Useful for merging and
    # identifying bootstraps.
    value_column = "character",
    # The column(s) required for grouping the data. Useful for determining confidence intervals.
    grouping_column = "ANY",
    # Flag that signals whether the data is aggregated, e.g. by computing
    # confidence intervals and a bias-corrected value.
    is_aggregated = "logical"
  ),
  prototype = list(
    data = NULL,
    identifiers = NULL,
    detail_level = NA_character_,
    estimation_type = NA_character_,
    confidence_level = NULL,
    bootstrap_ci_method = NA_character_,
    value_column = NA_character_,
    grouping_column = NULL,
    is_aggregated = FALSE
  )
)



# experimentData object --------------------------------------------------------

#' Experiment data
#'
#' An experimentData object contains information concerning the experiment.
#' These objects can be used to instantiate multiple experiments using the same
#' iterations, feature information and variable importance.
#'
#' @slot experiment_setup Contains regarding the experimental setup that is used
#'   to generate the iteration list.
#' @slot iteration_list List of iteration data that determines which instances
#'   are assigned to training, validation and test sets.
#' @slot feature_info Feature information objects. Only available if the
#'   experimentData object was generated using the `precompute_feature_info` or
#'   `precompute_vimp` functions.
#' @slot vimp_hyperparameter_list List of hyperparameters for variable importance
#'   objects.  Only available if the experimentData object was created using the
#'   `precompute_vimp` function.
#' @slot vimp_table_list List of variable importance table objects. Only
#'   available if the experimentData object was created using the
#'   `precompute_vimp` function.
#' @slot project_id Identifier of the project that generated the experimentData
#'   object.
#' @slot familiar_version Version of the familiar package used to create this
#'   experimentData.
#'
#' @details experimentData objects are primarily used to improve
#'   reproducibility, since these allow for training models on a shared
#'   foundation.
#'
#' @seealso \code{\link{precompute_data_assignment}}
#'   \code{\link{precompute_feature_info}}, \code{\link{precompute_vimp}}
#' @export

setClass(
  "experimentData",
  slots = list(
    # Experimental design.
    experiment_setup = "ANY",
    # List of iteration data.
    iteration_list = "ANY",
    # List of feature information objects.
    feature_info = "ANY",
    # List of variable importance hyperparameters.
    vimp_hyperparameter_list = "ANY",
    # List of variable importance tables.
    vimp_table_list = "ANY",
    # Project identifier for consistency tracking
    project_id = "ANY",
    # Package version for backward compatibility
    familiar_version = "ANY"
  ),
  prototype = list(
    experiment_setup = NULL,
    iteration_list = NULL,
    feature_info = NULL,
    vimp_hyperparameter_list = NULL,
    vimp_table_list = NULL,
    project_id = NULL,
    familiar_version = NULL
  )
)



# familiarDataElementPredictionTable object ------------------------------------
setClass(
  "familiarDataElementPredictionTable",
  contains = "familiarDataElement",
  slots = list(
    "learner" = "character",
    "vimp_method" = "character",
    "ensemble_method" = "character",
    "percentiles" = "ANY",
    "outcome_type" = "ANY",
    "identifier_data" = "ANY",
    "reference_data" = "ANY",
    "prediction_data" = "ANY"
  ),
  prototype = methods::prototype(
    bootstrap_ci_method = "percentile",
    learner = "custom_learner",
    vimp_method = "custom_vimp_method",
    ensemble_method = "median",
    percentiles = NULL,
    outcome_type = NULL,
    identifier_data = NULL,
    reference_data = NULL, 
    prediction_data = NULL
  )
)



# familiarTask object ----------------------------------------------------------
setClass(
  "familiarTask",
  slots = list(
    "task_name" = "character",
    "task_id" = "integer",
    "n_tasks" = "integer",
    "data_id" = "integer",
    "run_id" = "integer",
    "run_table" = "ANY",
    "file" = "character",
    "project_id" = "ANY"
  ),
  prototype = methods::prototype(
    task_name = NA_character_,
    task_id = 1L,
    n_tasks = 1L,
    data_id = NA_integer_,
    run_id = NA_integer_,
    run_table = NULL,
    file = NA_character_,
    project_id = NULL
  )
)


# familiarPlot object ----------------------------------------------------------
setClass(
  "familiarPlot",
  slots = list(
    "gtable" = "ANY",
    "global_elements" = "ANY",
    "row_id" = "integer",
    "col_id" = "integer",
    "remove_strip_x" = "logical",
    "remove_strip_y" = "logical",
    "remove_axis_text_x" = "logical",
    "remove_axis_text_y" = "logical",
    "remove_axis_label_x" = "logical",
    "remove_axis_label_y" = "logical",
    "remove_panel" = "logical"
  ),
  prototype = methods::prototype(
    gtable = NULL,
    global_elements = NULL,
    row_id = NA_integer_,
    col_id = NA_integer_,
    remove_strip_x = FALSE,
    remove_strip_y = FALSE,
    remove_axis_text_x = FALSE,
    remove_axis_text_y = FALSE,
    remove_axis_label_x = FALSE,
    remove_axis_label_y = FALSE,
    remove_panel = FALSE
  )
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
#' @include Transformation.R
NULL

# update_object (generic) ------------------------------------------------------

#' @title Update familiar S4 objects to the most recent version.
#'
#' @description Provides backward compatibility for familiar objects exported to
#'   a file. This mitigates compatibility issues when working with files that
#'   become outdated as new versions of familiar are released, e.g. because
#'   slots have been removed.
#'   
#'   Major version releases (e.g., versions 1.0.0, 2.0.0) introduce breaking
#'   changes which can result in objects that cannot be updated.
#'
#' @param object A `familiarModel`, a `familiarEnsemble`, a `familiarData` or
#'   `familiarCollection` object.
#' @param ... Unused arguments.
#'
#' @return An up-to-date version of the respective S4 object.
#' @exportMethod update_object
#' @md
#' @rdname update_object-methods
setGeneric("update_object", function(object, ...) standardGeneric("update_object"))



## update_object (familiarModel) -----------------------------------------------

#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "familiarModel"),
  function(object, ...) {
    if (tail(object@familiar_version, n = 1L) < "2.0.0") {
      ..error_cannot_update_object(object, "2.0.0")
    }
    
    if (!methods::validObject(object)) ..error_updated_object_invalid(object)

    # Update package version.
    object <- add_package_version(object = object)

    return(object)
  }
)



## update_object (familiarEnsemble) --------------------------------------------

#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "familiarEnsemble"),
  function(object, ...) {
    if (tail(object@familiar_version, n = 1L) < "2.0.0") {
      ..error_cannot_update_object(object, "2.0.0")
    }
    
    if (!methods::validObject(object)) ..error_updated_object_invalid(object)
    
    # Update package version.
    object <- add_package_version(object = object)

    return(object)
  }
)


## update_object (familiarData) ------------------------------------------------

#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "familiarData"),
  function(object, ...) {
    if (tail(object@familiar_version, n = 1L) < "2.0.0") {
      ..error_cannot_update_object(object, "2.0.0")
    }
    
    if (!methods::validObject(object)) ..error_updated_object_invalid(object)
    
    # Update package version.
    object <- add_package_version(object = object)

    return(object)
  }
)



## update_object (familiarCollection) ------------------------------------------

#' @rdname update_object-methods
setMethod(
  "update_object", signature(object = "familiarCollection"),
  function(object, ...) {
    if (tail(object@familiar_version, n = 1L) < "2.0.0") {
      ..error_cannot_update_object(object, "2.0.0")
    }
    
    if (!methods::validObject(object)) ..error_updated_object_invalid(object)
    
    # Update package version.
    object <- add_package_version(object = object)
    
    return(object)
  }
)



## update_object (vimpTable) ---------------------------------------------------
#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "vimpTable"),
  function(object, ...) {
    # Update package version.
    object <- add_package_version(object = object)

    return(object)
  }
)



## update_object (familiarNoveltyDetector) -------------------------------------
#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "familiarNoveltyDetector"),
  function(object, ...) {
    if (tail(object@familiar_version, n = 1L) < "2.0.0") {
      ..error_cannot_update_object(object, "2.0.0")
    }
    
    if (!methods::validObject(object)) ..error_updated_object_invalid(object)
    
    # Update package version.
    object <- add_package_version(object = object)

    return(object)
  }
)



## update_object (featureInfo) -------------------------------------------------
#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "featureInfo"),
  function(object, ...) {
    if (tail(object@familiar_version, n = 1L) < "2.0.0") {
      ..error_cannot_update_object(object, "2.0.0")
    }
    
    if (!methods::validObject(object)) ..error_updated_object_invalid(object)
    
    # Update package version.
    object <- add_package_version(object = object)

    return(object)
  }
)



## update_object (featureInfoParametersTransformationPowerTransform) -----------

#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "featureInfoParametersTransformationPowerTransform"),
  function(object, ...) {
    
    if (tail(object@familiar_version, n = 1L) < "2.0.0") {
      ..error_cannot_update_object(object, "2.0.0")
    }
    
    if (!methods::validObject(object)) ..error_updated_object_invalid(object)
    
    # Update package version.
    object <- add_package_version(object = object)
    
    return(object)
  }
)



## update_object (experimentData) ----------------------------------------------

#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "experimentData"),
  function(object, ...) {
    if (tail(object@familiar_version, n = 1L) < "2.0.0") {
      ..error_cannot_update_object(object, "2.0.0")
    }
    
    if (!methods::validObject(object)) ..error_updated_object_invalid(object)
    
    # Update package version.
    object <- add_package_version(object = object)

    return(object)
  }
)



## update_object (list) ------------------------------------------------------

#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "list"),
  function(object, ...) {
    # Pass to underlying methods.
    object <- lapply(
      object,
      update_object,
      ...
    )

    return(object)
  }
)



## update_object (ANY) ------------------------------------------------------
#' @rdname update_object-methods
setMethod(
  "update_object",
  signature(object = "ANY"),
  function(object, ...) {
    # Fallback method for missing or unknown items.
    return(object)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R

setClass("familiarHyperparameterRandomSearch",
  contains = "familiarHyperparameterLearner"
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarHyperparameterRandomSearch"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set name
    .Object@name <- "Random search"

    return(.Object)
  }
)



.get_available_random_search_hyperparameter_learners <- function() {
  return(c("random", "random_search"))
}



# model_is_trained -------------------------------------------------------------
setMethod(
  "model_is_trained",
  signature(object = "familiarHyperparameterRandomSearch"),
  function(object) {
    # Override default model_is_trained since random search is technically not a
    # model at all.
    return(TRUE)
  }
)

testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Simple test
vimp_methods <- c("none", "mim", "mim")
experimental_designs <- c("fs+mb", "fs+mb", "mb")
for (ii in seq_along(vimp_methods)) {
  familiar:::integrated_test(
    experimental_design = experimental_designs[ii],
    vimp_method = vimp_methods[ii],
    parallel = FALSE,
    estimation_type = "point",
    n_important_features = 2L,
    outcome_type_available = "binomial",
    debug = debug_flag
  )
  
  # Using all features for evaluation.
  familiar:::integrated_test(
    experimental_design = experimental_designs[ii],
    vimp_method = vimp_methods[ii],
    parallel = FALSE,
    estimation_type = "point",
    outcome_type_available = "binomial",
    debug = debug_flag
  )
}


# Bootstrap (without optimisation within bootstraps)
experimental_designs <- c("bt(fs+mb, 3)", "bt(fs+mb, 3)", "bt(mb, 3)")
for (ii in seq_along(vimp_methods)) {
  familiar:::integrated_test(
    experimental_design = experimental_designs[ii],
    vimp_method = vimp_methods[ii],
    parallel = FALSE,
    estimation_type = "point",
    n_important_features = 2L,
    outcome_type_available = "binomial",
    debug = debug_flag
  )
}


# Bootstrap (with pre-processing and optimisation within bootstraps)
experimental_designs <- c("bs(fs+mb, 3)", "bs(fs+mb, 3)", "bs(mb, 3)")
for (ii in seq_along(vimp_methods)) {
  familiar:::integrated_test(
    experimental_design = experimental_designs[ii],
    vimp_method = vimp_methods[ii],
    parallel = FALSE,
    estimation_type = "point",
    n_important_features = 2L,
    outcome_type_available = "binomial",
    debug = debug_flag
  )
}


# Cross-validation
experimental_designs <- c("cv(fs+mb, 3)", "cv(fs+mb, 3)", "cv(mb, 3)")
for (ii in seq_along(vimp_methods)) {
  familiar:::integrated_test(
    experimental_design = experimental_designs[ii],
    vimp_method = vimp_methods[ii],
    parallel = FALSE,
    estimation_type = "point",
    n_important_features = 2L,
    outcome_type_available = "binomial",
    debug = debug_flag
  )
  
  # Using all features for evaluation.
  familiar:::integrated_test(
    experimental_design = experimental_designs[ii],
    vimp_method = vimp_methods[ii],
    parallel = FALSE,
    estimation_type = "point",
    outcome_type_available = "binomial",
    debug = debug_flag
  )
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

.get_available_imputation_methods <- function() {
  return(c(
    .get_available_simple_imputation_methods(),
    .get_available_lasso_imputation_methods()
  ))
}

.get_available_simple_imputation_methods <- function() {
  return("simple")
}

.get_available_lasso_imputation_methods <- function() {
  return("lasso")
}

setClass(
  "featureInfoParametersImputation",
  contains = "featureInfoParameters",
  slots = list(
    "method" = "character",
    "type" = "character",
    "required_features" = "ANY"
  ),
  prototype = list(
    "method" = NA_character_,
    "type" = NA_character_,
    "required_features" = NULL
  )
)

setClass(
  "featureInfoParametersImputationNone",
  contains = "featureInfoParametersImputation",
  slots = list("reason" = "ANY"),
  prototype = list("reason" = NULL)
)

setClass(
  "featureInfoParametersImputationSimple",
  contains = "featureInfoParametersImputation",
  slots = list("model" = "ANY"),
  prototype = list("model" = NULL)
)

setClass(
  "featureInfoParametersImputationLasso",
  contains = "featureInfoParametersImputation",
  slots = list(
    "simple" = "ANY",
    "model" = "ANY"
  ),
  prototype = list(
    "simple" = NULL,
    "model" = NULL
  )
)

setClass(
  "featureInfoParametersImputationContainer",
  contains = "featureInfoParameters",
  slots = list(
    "type" = "character",
    "required_features" = "ANY",
    "model" = "ANY"
  ),
  prototype = list(
    "type" = NA_character_,
    "required_features" = NULL,
    "model" = NULL
  )
)



create_imputation_parameter_skeleton <- function(
    feature_info_list,
    feature_names = NULL,
    imputation_method,
    .override_existing = FALSE
) {
  # Creates a skeleton for the provided imputation method.
  
  # Determine feature names from the feature info list, if provided.
  if (is.null(feature_names)) feature_names <- names(feature_info_list)
  
  # Select only features that appear in the feature info list.
  feature_names <- intersect(
    names(feature_info_list),
    feature_names
  )
  
  # Skip step if no feature info objects are updated.
  if (is_empty(feature_names)) return(feature_info_list)
  
  # Check that method is applicable.
  .check_parameter_value_is_valid(
    x = imputation_method,
    var_name = "imputation_method",
    values = .get_available_imputation_methods()
  )
  
  # Update familiar info objects with a feature normalisation skeleton.
  updated_feature_info <- fam_lapply(
    X = feature_info_list[feature_names],
    FUN = .create_imputation_parameter_skeleton,
    method = imputation_method,
    .override_existing = .override_existing
  )
  
  # Provide names for the updated feature info objects.
  names(updated_feature_info) <- feature_names
  
  # Replace list elements.
  feature_info_list[feature_names] <- updated_feature_info
  
  return(feature_info_list)
}



.create_imputation_parameter_skeleton <- function(
    feature_info,
    method,
    .override_existing = FALSE
) {
  # Check if imputation data was already completed, and does not require being
  # determined anew.
  if (
    feature_info_complete(feature_info@imputation_parameters)
    && !.override_existing
  ) {
    return(feature_info)
  }
  
  # Pass to underlying function that constructs the skeleton.
  object <- ..create_imputation_parameter_skeleton(
    feature_name = feature_info@name,
    feature_type = feature_info@feature_type,
    available = is_available(feature_info),
    method = method
  )
  
  # Update imputation_parameters slot.
  feature_info@imputation_parameters <- object
  
  return(feature_info)
}



..create_imputation_parameter_skeleton <- function(
    feature_name,
    feature_type = "numeric",
    available = TRUE,
    method
) {
  # This is the lowest level function for creating imputation parameter
  # skeletons.
  
  # Create the relevant objects.
  if (!available || method == "none") {
    object <- methods::new(
      "featureInfoParametersImputationNone",
      reason = "feature was omitted prior to transformation"
    )
    
  } else if (method %in% .get_available_simple_imputation_methods()) {
    object <- methods::new(
      "featureInfoParametersImputationSimple",
      "method" = method
    )
    
  } else if (method %in% .get_available_lasso_imputation_methods()) {
    object <- methods::new(
      "featureInfoParametersImputationLasso",
      "method" = method
    )
    
  } else if (method == "container") {
    object <- methods::new("featureInfoParametersImputationContainer")
    
  } else {
    ..error_reached_unreachable_code(paste0(
      "..create_imputation_parameter_skeleton: encountered an unknown imputation method: ",
      paste_s(method)
    ))
  }
  
  # Set the name and type of the object.
  object@name <- feature_name
  object@type <- feature_type
  
  # Update the familiar version.
  object <- add_package_version(object = object)
  
  return(object)
}



add_imputation_info <- function(
    cl = NULL,
    feature_info_list,
    data,
    verbose = FALSE
) {
  # Determine normalisation parameters and add them to the feature_info_list.
  
  # Find feature columns.
  feature_names <- get_feature_columns(x = data)
  
  # Sanity check.
  if (!(setequal(feature_names, get_available_features(feature_info_list = feature_info_list)))) {
    ..error_reached_unreachable_code(paste0(
      "add_imputation_info: features in data and the feature info list ",
      "are expect to be the same, but were not."
    ))
  }
  
  # Iterate over features and train univariate imputation.
  updated_feature_info <- fam_mapply(
    cl = cl,
    FUN = .add_imputation_info,
    feature_info = feature_info_list[feature_names],
    mask_data = data@data[, mget(feature_names)],
    MoreArgs = list("data" = data),
    progress_bar = verbose,
    chopchop = TRUE
  )
  
  # Provide names for the updated feature info objects.
  names(updated_feature_info) <- feature_names
  
  # Replace list elements.
  feature_info_list[feature_names] <- updated_feature_info
  
  # Apply initial results to the dataset.
  imputed_data <- .impute_features(
    data = data,
    feature_info_list = feature_info_list,
    initial_imputation = TRUE
  )
  
  # Check if there are any features that lack imputation (are of the
  # featureInfoParametersImputationNone class). These are subsequently skipped
  # to avoid having to rely on potentially missing data to infer other features.
  feature_names <- unname(feature_names[!sapply(
    feature_info_list[feature_names],
    function(x) (is(x@imputation_parameters, "featureInfoParametersImputationNone"))
  )])
  
  if (length(feature_names) > 0L) {
    # Iterate over features again to train multivariate imputation models.
    updated_feature_info <- fam_mapply(
      cl = cl,
      FUN = .add_imputation_info,
      feature_info = feature_info_list[feature_names],
      mask_data = data@data[, mget(feature_names)],
      MoreArgs = list("data" = imputed_data),
      progress_bar = verbose,
      chopchop = TRUE
    )
    
    # Provide names for the updated feature info objects.
    names(updated_feature_info) <- feature_names
    
    # Replace list elements.
    feature_info_list[feature_names] <- updated_feature_info
  }
  
  return(feature_info_list)
}



.add_imputation_info <- function(
    feature_info,
    data,
    mask_data
) {
  
  # Pass to underlying function that adds the feature info.
  object <- add_feature_info_parameters(
    object = feature_info@imputation_parameters,
    data = data,
    mask_data = mask_data
  )
  
  # Update normalisation_parameters slot.
  feature_info@imputation_parameters <- object
  
  return(feature_info)
}



.impute_features <- function(
    cl = NULL,
    data,
    feature_info_list,
    initial_imputation,
    mask_data = NULL,
    verbose = FALSE
) {
  
  # Check if data is empty
  if (is_empty(data)) return(data)
  
  # Check if data has features
  if (!has_feature_data(x = data)) return(data)
  
  # Find the columns containing features
  feature_names <- get_feature_columns(x = data)
  
  # Use data variable for masking uncensored data.
  if (is.null(mask_data)) mask_data <- data
  
  # Check if any features have missing values, otherwise we can just skip
  # imputation.
  has_missing_values <- sapply(
    X = mask_data@data[, mget(feature_names)],
    FUN = function(x) (return(sum(!is_valid_data(x))))
  ) > 0L
  
  if (!any(has_missing_values)) return(data)
  
  # Impute data.  
  imputation_list <- fam_mapply(
    cl = cl,
    FUN = ..impute_features,
    feature_info = feature_info_list[feature_names],
    mask_data = mask_data@data[, mget(feature_names)],
    MoreArgs = list(
      "data" = data,
      "initial_imputation" = initial_imputation
    ),
    progress_bar = verbose,
    chopchop = TRUE
  )
  
  # Update name of data in columns.
  names(imputation_list) <- feature_names
  
  # Update with replacement in the data object.
  data <- update_with_replacement(
    data = data,
    replacement_list = imputation_list
  )
  
  return(data)
}



..impute_features <- function(
    feature_info,
    data,
    mask_data,
    initial_imputation
) {
  
  # Find uncensored data.
  uncensored_data <- .mask_data_to_mask(
    mask_data = mask_data,
    type = feature_info@imputation_parameters@type
  )
  
  # Return data as is, if there is no censored data.
  if (all(uncensored_data)) {
    return(data@data[[feature_info@imputation_parameters@name]])
  }
  
  # Infer missing values.
  y <- apply_feature_info_parameters(
    object = feature_info@imputation_parameters,
    data = data,
    mask_data = mask_data,
    initial_imputation = initial_imputation
  )
  
  return(y)
}



# initialize (none) ------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "featureInfoParametersImputationNone"),
  function(.Object, ...) {
    
    # Update with parent class first.
    .Object <- callNextMethod()
    
    # The parameter set is by definition complete when no normalisation
    # is performed.
    .Object@complete <- TRUE
    
    return(.Object)
  }
)



# add_feature_info_parameters (generic imputation, ANY) ------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersImputation",
    data = "ANY"
  ),
  function(
    object,
    data,
    mask_data,
    ...
  ) {
    
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Check that any data is present.
    if (is_empty(data)) {
      # Return a none-class object.
      object <- ..create_imputation_parameter_skeleton(
        feature_name = object@name,
        feature_type = object@type,
        method = "none"
      )
      
      object@reason <- "insufficient data to infer imputation parameters"
      
      return(object)
    }
    
    # Find uncensored data.
    mask_data <- .mask_data_to_mask(
      mask_data = mask_data,
      type = object@type
    )
    
    # If there are no uncensored data, we cannot determine imputation
    # parameters.
    if (!any(mask_data)) {
      # Return a none-class object.
      object <- ..create_imputation_parameter_skeleton(
        feature_name = object@name,
        feature_type = object@type,
        method = "none"
      )
      
      object@reason <- "insufficient data to infer imputation parameters"
      
      return(object)
    }
    
    return(object)
  }
)



# add_feature_info_parameters (simple imputation, data object) -----------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersImputationSimple",
    data = "dataObject"
  ),
  function(
    object,
    data,
    mask_data,
    ...
  ) {
    
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Run general checks for imputation. This may yield none-class objects which
    # are complete by default.
    object <- callNextMethod()
    
    # Check if all required parameters have been set now.
    if (feature_info_complete(object)) return(object)
    
    # Find uncensored data.
    mask_data <- .mask_data_to_mask(
      mask_data = mask_data,
      type = object@type
    )
    
    # Select data.
    y <- data@data[mask_data, mget(object@name)][[1L]]
    
    # Select value.
    if (object@type == "numeric") {
      # Select median values
      common_value <- stats::median(y)
      
    } else if (object@type == "factor") {
      # Select the mode as the replacement value
      common_value <- get_mode(y)
    }
    
    # Add value to object.
    object@model <- common_value
    
    # Set required feature.
    object@required_features <- object@name
    
    # Set complete.
    object@complete <- TRUE
    
    return(object)
  }
)



# add_feature_info_parameters (lasso imputation, data object) ------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersImputationLasso",
    data = "dataObject"
  ),
  function(
    object, 
    data,
    mask_data,
    ...
  ) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    name <- NULL
    
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Run general checks for imputation. This may yield none-class
    # objects which are complete by default.
    object <- callNextMethod()
    
    # Check if all required parameters have been set now.
    if (feature_info_complete(object)) return(object)
    
    # Check if simple inference values have been set. For multivariate inference
    # methods, an initial univariate step is used to fill any holes in the data,
    # prior to prediction. This method is therefore first called to set
    # parameters of the univariate, prior to determining the LASSO model.
    if (!feature_info_complete(object@simple)) {
      object@simple <- ..create_imputation_parameter_skeleton(
        feature_name = object@name,
        feature_type = object@type,
        method = "simple"
      )
      
      # Determine parameters.
      object@simple <- add_feature_info_parameters(
        object = object@simple,
        data = data,
        mask_data = mask_data
      )
      
      return(object)
    }
    
    # Determine if more than one feature is present.
    if (get_n_features(data) == 1L) {
      # Switch to simple  univariate inference only, in case only one feature is
      # present.
      return(object@simple)
    }
    
    # Find uncensored data.
    mask_data <- .mask_data_to_mask(
      mask_data = mask_data,
      type = object@type
    )
    
    # Select finite data.
    distribution <- ifelse(object@type == "numeric", "gaussian", "multinomial")
    
    # Check that the glmnet package is installed.
    require_package(
      x = "glmnet",
      purpose = "to impute data using lasso regression"
    )
    
    # Select known data as response variable.
    y <- data@data[[object@name]][mask_data]
    if (distribution == "multinomial") y <- droplevels(y)
    
    # Select features.
    x <- filter_features(data, remove_features = object@name)
    
    # Use effect coding to convert categorical data into encoded data.
    encoded_data <- encode_categorical_variables(
      data = x,
      object = NULL,
      encoding_method = "dummy",
      drop_levels = FALSE
    )
    
    # Extract data table with contrasts.
    x <- encoded_data$encoded_data
    
    # Perform a cross-validation to derive a optimal lambda. This may rarely
    # fail if y is numeric but does not change in a particular fold. In that
    # case, skip.
    lasso_model <- suppressWarnings(tryCatch(
      glmnet::cv.glmnet(
        x = as.matrix(x@data[mask_data, mget(get_feature_columns(x))]),
        y = y,
        family = distribution,
        alpha = 1L,
        standardize = FALSE,
        nfolds = min(sum(mask_data), 20L),
        parallel = FALSE
      ),
      error = identity
    ))
    
    if (inherits(lasso_model, "error")) return(object@simple)
    
    # Determine the features with non-zero coefficients. We want to have the
    # minimal required support for the model to minimise the effort that is
    # required for external model validation. GLMNET by definition wants to have
    # the exact same complete input space, even though but a few features are
    # used. That is a bad idea for portability. Therefore we will be jumping
    # through some hoops to make a model with minimum support.
    if (distribution == "multinomial") {
      # Read coefficient lists
      coef_list <- tryCatch(
        coef(lasso_model, s = "lambda.1se"),
        error = identity
      )
      
      if (inherits(coef_list, "error")) return(object@simple)
      
      # Parse into matrix and retrieve row names
      coef_mat <- sapply(coef_list, as.matrix)
      rownames(coef_mat) <- dimnames(coef_list[[1L]])[[1L]]
      
      # Calculate score
      score <- apply(abs(coef_mat), 1L, max)
      
    } else {
      # Read coefficient lists
      coef_list <- tryCatch(
        coef(lasso_model, s = "lambda.1se"),
        error = identity
      )
      
      if (inherits(coef_list, "error")) return(object@simple)
      
      # Read coefficient matrix
      coef_mat <- as.matrix(coef_list)
      
      # Calculate score
      score <- abs(coef_mat)[, 1L]
    }
    
    # Parse score to data.table
    vimp_table <- data.table::data.table(
      "score" = score,
      "name" = names(score)
    )
    
    # Throw out the intercept and elements with 0.0 coefficients
    vimp_table <- vimp_table[name != "(Intercept)" & score != 0.0]
    
    # Create variable importance object.
    vimp_object <- methods::new(
      "vimpTable",
      vimp_table = vimp_table,
      encoding_table = encoded_data$reference_table,
      score_aggregation = "max",
      invert = TRUE
    )
    
    # Find the original names
    vimp_table <- get_vimp_table(vimp_object, "decoded")
    
    # Check that the optimal model complexity lambda is connected to at least
    # one feature. If not, we have to use the simple estimate.
    if (is_empty(vimp_table)) return(object@simple)
    
    # Derive required features
    required_features <- vimp_table$name
    
    # Select features.
    x <- filter_features(data, available_features = required_features)
    
    # Use effect coding to convert categorical data into encoded data.
    encoded_data <- encode_categorical_variables(
      data = x,
      object = NULL,
      encoding_method = "dummy",
      drop_levels = FALSE
    )
    
    # Extract data table with contrasts.
    x <- encoded_data$encoded_data@data[
      mask_data,
      mget(get_feature_columns(encoded_data$encoded_data))
    ]
    
    # Check the number of columns in train_data. glmnet wants at least two
    # columns.
    if (ncol(x) == 1L) x[, "bogus__variable__" := 0.0]
    
    # Force x locally, otherwise an error may occur.
    x <- as.matrix(x)
    
    # Train a small model at this lambda.1se.
    lasso_model <- suppressWarnings(tryCatch(
      glmnet::glmnet(
        x = x,
        y = y,
        family = distribution,
        lambda = lasso_model$lambda.1se,
        standardize = FALSE
      ),
      error = identity
    ))
    
    if (inherits(coef_list, "error")) return(object@simple)
    
    # Remove extraneous information from the model.
    lasso_model <- ..trim_glmnet(lasso_model)
    
    # Add LASSO regression model to object.
    object@model <- lasso_model
    object@required_features <- required_features
    
    # Set complete
    object@complete <- TRUE
    
    return(object)
  }
)



# apply_feature_info_parameters (none, data object) ----------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersImputationNone",
    data = "dataObject"
  ),
  function(
    object, 
    data,
    mask_data,
    ...
  ) {
    return(data@data[[object@name]])
  }
)


# apply_feature_info_parameters (simple, data object) --------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersImputationSimple",
    data = "dataObject"
  ),
  function(
    object, 
    data,
    mask_data,
    ...
  ) {
    
    # Find uncensored data.
    mask_data <- .mask_data_to_mask(
      mask_data = mask_data,
      type = object@type
    )
    
    # Return data as is, if there is no censored data.
    if (all(mask_data)) return(data@data[[object@name]])
    
    # Get intended data.
    y <- data@data[[object@name]]
    
    # Replace censored values.
    y[!mask_data] <- object@model
    
    return(y)
  }
)



# apply_feature_info_parameters (none, data object) ----------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersImputationLasso",
    data = "dataObject"
  ),
  function(
    object, 
    data,
    mask_data,
    initial_imputation,
    ...
  ) {
    
    # For initial imputation use univariate imputer.
    if (initial_imputation) {
      return(apply_feature_info_parameters(
        object = object@simple,
        data = data,
        mask_data = mask_data,
        initial_imputation = initial_imputation
      ))
    }
    
    # Find uncensored data.
    mask_data <- .mask_data_to_mask(
      mask_data = mask_data,
      type = object@type
    )
    
    # Return data as is, if there is no censored data.
    if (all(mask_data)) return(data@data[[object@name]])
    
    # Check that features required for imputation are present in data.
    if (!all(object@required_features %in% get_feature_columns(data))) {
      return(apply_feature_info_parameters(
        object = object@simple,
        data = data,
        mask_data = mask_data,
        initial_imputation = initial_imputation
      ))
    }
    
    # Check that the glmnet package is installed.
    require_package(
      x = "glmnet",
      purpose = "to impute data using lasso regression"
    )
    
    # Get intended data.
    y <- data@data[[object@name]]
    
    # Select features.
    x <- filter_features(data, available_features = object@required_features)
    
    # Use effect coding to convert categorical data into encoded data.
    encoded_data <- encode_categorical_variables(
      data = x,
      object = NULL,
      encoding_method = "dummy",
      drop_levels = FALSE
    )
    
    # Extract data table with contrasts.
    x <- encoded_data$encoded_data@data[
      !mask_data,
      mget(get_feature_columns(encoded_data$encoded_data))
    ]
    
    # Check if the validation data has two or more columns
    if (ncol(x) == 1L) x[, "bogus__variable__" := 0.0]
    
    # Force x locally, otherwise an error may occur.
    x <- as.matrix(x)
    
    # Get the type of response for glmnet predict
    response_type <- ifelse(object@type == "numeric", "response", "class")
    
    # Impute values for censored values.
    y[!mask_data] <- drop(predict(
      object = object@model,
      newx = x,
      type = response_type
    ))
    
    return(y)
  }
)



# apply_feature_info_parameters (container, data object) -----------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersImputationContainer",
    data = "dataObject"
  ),
  function(
    object, 
    data,
    mask_data,
    initial_imputation,
    ...
  ) {
    
    # Find uncensored data.
    uncensored_data <- .mask_data_to_mask(
      mask_data = mask_data,
      type = object@type
    )
    
    # Return data as is, if there is no censored data.
    if (all(uncensored_data)) return(data@data[[object@name]])
    
    # Get intended data.
    y <- data@data[[object@name]]
    
    # Push only rows of data to be imputed.
    censored_data <- data
    censored_data@data <- data.table::copy(censored_data@data[!uncensored_data, ])
    
    # Set aggregation function for 
    if (object@type == "numeric") {
      aggregation_function <- mean
      
    } else if (object@type == "factor") {
      aggregation_function <- get_mode
    }
    
    # Dispatch to imputers in the container.
    imputed_values <- lapply(
      object@model,
      apply_feature_info_parameters,
      data = censored_data,
      mask_data = mask_data[!uncensored_data],
      initial_imputation = initial_imputation
    )
    
    # Set as data.table.
    imputed_values <- data.table::as.data.table(imputed_values)
    
    # Aggregate by row.
    imputed_values <- imputed_values[
      ,
      list("value" = aggregation_function(do.call(c, .SD))),
      .SDcols = colnames(imputed_values),
      by = seq_len(nrow(imputed_values))
    ]
    
    # Replace censored values in y.
    y[!uncensored_data] <- imputed_values$value
    
    return(y)
  }
)



.mask_data_to_mask <- function(mask_data, type) {
  if (type == "numeric") {
    mask_data <- is.finite(mask_data)
    
  } else if (type == "factor") {
    mask_data <- !is.na(mask_data)
    
  } else {
    ..error_reached_unreachable_code(paste0(".mask_data_to_mask: encountered unknown feature type: ", type))
  }
  
  return(mask_data)
}



..collect_and_aggregate_imputation_info <- function(
    feature_info_list,
    feature_name,
    feature_type
) {
  # Aggregate transformation parameters. This function exists so that it can be
  # tested as part of a unit test.
  
  # Create container object.
  container_object <- ..create_imputation_parameter_skeleton(
    feature_name = feature_name,
    feature_type = feature_type,
    method = "container"
  )
  
  # Extract and store imputation objects.
  container_object@model <- lapply(
    feature_info_list,
    function(x) {
      # Filter out "none" class imputation
      # objects.
      if (is(x@imputation_parameters, "featureInfoParametersImputationNone")) return(NULL)
      
      return(x@imputation_parameters)
    }
  )
  
  if (is_empty(container_object@model)) {
    container_object@model <- list(
      ..create_imputation_parameter_skeleton(
        feature_name = feature_name,
        feature_type = feature_type,
        method = "none"
      )
    )
  }
  
  # Set required parameters.
  container_object@required_features <- unique(unlist(lapply(
    container_object@model,
    function(x) (x@required_features)
  )))
  
  # Mark as complete.
  container_object@complete <- TRUE
  
  return(list("parameters" = container_object))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarMetricCM -------------------------------------------------------------
setClass(
  "familiarMetricCM",
  contains = "familiarMetric"
)



# familiarMetricCMAveraging ----------------------------------------------------
setClass(
  "familiarMetricCMAveraging",
  contains = "familiarMetricCM",
  slots = list("averaging" = "character"),
  prototype = list("averaging" = "macro")
)


# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarMetricCMAveraging"),
  function(.Object, metric, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    if (endsWith(x = metric, suffix = "_macro")) {
      .Object@averaging <- "macro"
      .Object@metric <- sub_last(
        x = metric,
        pattern = "_macro",
        replacement = "",
        fixed = TRUE
      )
      
    } else if (endsWith(x = metric, suffix = "_micro")) {
      .Object@averaging <- "micro"
      .Object@metric <- sub_last(
        x = metric,
        pattern = "_micro",
        replacement = "",
        fixed = TRUE
      )
      
    } else if (endsWith(x = metric, suffix = "_weighted")) {
      .Object@averaging <- "weighted"
      .Object@metric <- sub_last(
        x = metric,
        pattern = "_weighted",
        replacement = "",
        fixed = TRUE
      )
      
    } else {
      # Default setting.
      .Object@averaging <- "macro"
      .Object@metric <- metric
    }

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMetricCM"),
  function(object, ...) {
    return(object@outcome_type %in% c("binomial", "multinomial"))
  }
)



# accuracy ---------------------------------------------------------------------
setClass(
  "familiarMetricAccuracy",
  contains = "familiarMetricCM",
  prototype = methods::prototype(
    name = "Accuracy",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (accuracy) ---------------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricAccuracy"),
  function(metric, data, ...) {
    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Compute accuracy
    return(sum(cm$tp) / cm$n_samples)
  }
)



# balanced accuracy ------------------------------------------------------------
setClass(
  "familiarMetricBalancedAccuracy",
  contains = "familiarMetricCM",
  prototype = methods::prototype(
    name = "Balanced Accuracy",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (balanced accuracy) ------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricBalancedAccuracy"),
  function(metric, data, ...) {
    # Balanced accuracy [Brodersen, K. H., Ong, C. S., Stephan, K. E. & Buhmann,
    # J. M. The Balanced Accuracy and Its Posterior Distribution. in 2010 20th
    # International Conference on Pattern Recognition 3121–3124 (2010).]

    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Compute balanced accuracy
    return(sum(cm$tp / (cm$tp + cm$fn)) / cm$n_classes)
  }
)



# balanced error rate ----------------------------------------------------------
setClass(
  "familiarMetricBalancedErrorRate",
  contains = "familiarMetricCM",
  prototype = methods::prototype(
    name = "Balanced Error Rate",
    value_range = c(0.0, 1.0),
    higher_better = FALSE
  )
)



## compute_metric_score (balanced error) ---------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricBalancedErrorRate"),
  function(metric, data, ...) {
    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Compute balanced error rate
    return(sum(cm$fn / (cm$tp + cm$fn)) / cm$n_classes)
  }
)



# Cohen's kappa ----------------------------------------------------------------
setClass(
  "familiarMetricCohenKappa",
  contains = "familiarMetricCM",
  prototype = methods::prototype(
    name = "Cohen's Kappa",
    value_range = c(-1.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (Cohen's kappa) ----------------------------------------
setMethod(
  "compute_metric_score", signature(metric = "familiarMetricCohenKappa"),
  function(metric, data, ...) {
    # Cohen's kappa [Cohen, J. A Coefficient of Agreement for Nominal Scales.
    # Educ. Psychol. Meas. 20, 37–46 (1960).]
    #
    # Implementation after
    # https://stats.stackexchange.com/questions/251165/cohens-kappa-with-three-categories-of-variable

    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Compute the observed agreement
    obs_agreement <- sum(cm$tp) / cm$n_samples

    # Agreement expected at random
    rand_agreement <- sum((cm$tp + cm$fp) * (cm$tp + cm$fn)) / cm$n_samples^2.0

    if (rand_agreement == 1.0 && obs_agreement == 1.0) {
      # Prevent division by zero for perfect predictions.
      return(1.0)
      
    } else {
      # Compute kappa
      return((obs_agreement - rand_agreement) / (1.0 - rand_agreement))
    }
  }
)


# F1 score ---------------------------------------------------------------------
setClass(
  "familiarMetricF1Score",
  contains = "familiarMetricCMAveraging",
  prototype = methods::prototype(
    name = "F1 score",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (F1 score) ---------------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricF1Score"),
  function(metric, data, ...) {
    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Aggregate confusion matrix.
    cm <- ..aggregate_confusion_matrix(
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )

    # Compute score
    score <- 2.0 * cm$tp / (2.0 * cm$tp + cm$fp + cm$fn)

    # Aggregate score.
    score <- ..aggregate_confusion_matrix_score(
      score = score,
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )

    return(score)
  }
)



# false discovery rate ---------------------------------------------------------
setClass(
  "familiarMetricFDR",
  contains = "familiarMetricCMAveraging",
  prototype = methods::prototype(
    name = "False Discovery Rate",
    value_range = c(0.0, 1.0),
    higher_better = FALSE
  )
)



## compute_metric_score (false discovery rate) ---------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricFDR"),
  function(metric, data, ...) {
    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Aggregate confusion matrix.
    cm <- ..aggregate_confusion_matrix(
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )

    # Compute score.
    score <- cm$fp / (cm$tp + cm$fp)

    # Aggregate score.
    score <- ..aggregate_confusion_matrix_score(
      score = score,
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )
    
    return(score)
  }
)



# informedness -----------------------------------------------------------------
setClass(
  "familiarMetricInformedness",
  contains = "familiarMetricCM",
  prototype = methods::prototype(
    name = "Informedness",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (informedness) -----------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricInformedness"),
  function(metric, data, ...) {
    # Informedness [Powers, D. M. Evaluation: from precision, recall and
    # F-measure to ROC, informedness, markedness and correlation. International
    # Journal of Machine Learning Technology 2, 37–63 (2011).]
    #
    # See equations 28, 42.

    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)
    
    if (is.null(cm)) return(callNextMethod())

    # Compute the score
    return(sum(cm$prevalence / (1.0 - cm$prevalence) * (cm$tp / (cm$tp + cm$fn) - cm$bias)))
  }
)



# markedness -------------------------------------------------------------------
setClass(
  "familiarMetricMarkedness",
  contains = "familiarMetricCM",
  prototype = methods::prototype(
    name = "Markedness",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



##### compute_metric_score (Markedness)#####
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricMarkedness"),
  function(metric, data, ...) {
    # Markedness [Powers, D. M. Evaluation: from precision, recall and F-measure
    # to ROC, informedness, markedness and correlation. International Journal of
    # Machine Learning Technology 2, 37–63 (2011).]
    #
    # See equations 29, 43.

    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Compute the score
    return(sum(cm$bias / (1.0 - cm$bias) * (cm$tp / (cm$tp + cm$fp) - cm$prevalence)))
  }
)



# Matthews` correlation coefficient --------------------------------------------
setClass(
  "familiarMetricMCC",
  contains = "familiarMetricCM",
  prototype = methods::prototype(
    name = "Matthews' Correlation Coefficient",
    value_range = c(-1.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (Matthews` correlation coefficient) --------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricMCC"),
  function(metric, data, ...) {
    # See [Jurman, G., Riccadonna, S. & Furlanello, C. A comparison of MCC and
    # CEN error measures in multi-class prediction. PLoS One 7, e41882 (2012);
    # Gorodkin, J. Comparing two K-category assignments by a K-category
    # correlation coefficient. Comput. Biol. Chem. 28, 367–374 (2004).]
    #
    # Mirrors scikit-learn implementation.

    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Compute the score
    cov_obs_pred <- sum(cm$tp) / cm$n_samples - sum(cm$prevalence * cm$bias)
    cov_obs_obs <- 1.0 - sum(cm$prevalence * cm$prevalence)
    cov_pred_pred <- 1.0 - sum(cm$bias * cm$bias)

    if (cov_obs_obs == 0.0 || cov_pred_pred == 0.0) {
      return(0.0)
      
    } else {
      return(cov_obs_pred / sqrt(cov_obs_obs * cov_pred_pred))
    }
  }
)



# negative predictive value ----------------------------------------------------
setClass("familiarMetricNPV",
  contains = "familiarMetricCMAveraging",
  prototype = methods::prototype(
    name = "Negative Predictive Value",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (negative predictive value) ----------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricNPV"),
  function(metric, data, ...) {
    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Aggregate confusion matrix.
    cm <- ..aggregate_confusion_matrix(
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )

    # Compute NPV
    score <- cm$tn / (cm$tn + cm$fn)

    # Aggregate score.
    score <- ..aggregate_confusion_matrix_score(
      score = score,
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )

    return(score)
  }
)



# positive predictive value ----------------------------------------------------
setClass(
  "familiarMetricPPV",
  contains = "familiarMetricCMAveraging",
  prototype = methods::prototype(
    name = "Positive Predictive Value",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)




## compute_metric_score (Positive predictive value) ----------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricPPV"),
  function(metric, data, ...) {
    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Aggregate confusion matrix.
    cm <- ..aggregate_confusion_matrix(
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )

    # Compute PPV.
    score <- cm$tp / (cm$tp + cm$fp)

    # Aggregate score.
    score <- ..aggregate_confusion_matrix_score(
      score = score,
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )
    
    return(score)
  }
)



# sensitivity ------------------------------------------------------------------
setClass(
  "familiarMetricSensitivity",
  contains = "familiarMetricCMAveraging",
  prototype = methods::prototype(
    name = "Sensitivity",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (Sensitivity) ------------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricSensitivity"),
  function(metric, data, ...) {
    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Aggregate confusion matrix.
    cm <- ..aggregate_confusion_matrix(
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )

    # Compute sensitivity
    score <- cm$tp / (cm$tp + cm$fn)

    # Aggregate score.
    score <- ..aggregate_confusion_matrix_score(
      score = score,
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )
    
    return(score)
  }
)



# specificity ------------------------------------------------------------------
setClass(
  "familiarMetricSpecificity",
  contains = "familiarMetricCMAveraging",
  prototype = methods::prototype(
    name = "Specificity",
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (specificity) ------------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricSpecificity"),
  function(metric, data, ...) {
    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Aggregate confusion matrix.
    cm <- ..aggregate_confusion_matrix(
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )
    
    # Compute specificity.
    score <- cm$tn / (cm$tn + cm$fp)

    # Aggregate score.
    score <- ..aggregate_confusion_matrix_score(
      score = score,
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )

    return(score)
  }
)


# Youden's index ---------------------------------------------------------------
setClass(
  "familiarMetricYouden",
  contains = "familiarMetricCMAveraging",
  prototype = methods::prototype(
    name = "Youden's J Statistic",
    value_range = c(-1.0, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (Youden's index) ---------------------------------------
setMethod(
  "compute_metric_score", 
  signature(metric = "familiarMetricYouden"),
  function(metric, data, ...) {
    # Youden's J index [Youden, W. J. Index for rating diagnostic tests. Cancer
    # 3, 32–35 (1950).]

    # Compute data from the confusion metrics.
    cm <- ..compute_confusion_matrix_data(data = data)

    if (is.null(cm)) return(callNextMethod())

    # Aggregate confusion matrix.
    cm <- ..aggregate_confusion_matrix(
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )
    
    # Compute Youden index.
    score <- cm$tp / (cm$tp + cm$fn) + cm$tn / (cm$tn + cm$fp) - 1.0

    # Aggregate score.
    score <- ..aggregate_confusion_matrix_score(
      score = score,
      cm = cm,
      averaging = metric@averaging,
      outcome_type = metric@outcome_type
    )
    
    return(score)
  }
)



..compute_confusion_matrix_data <- function(data) {
  # Suppress NOTES due to non-standard evaluation in data.table
  predicted_class <- NULL
  
  if (is_empty(data)) return(NULL)
  
  if (!is(data, "predictionTableClassification")) {
    ..error_data_not_prediction_table(data, "predictionTableClassification")
  }
  
  # Get the classes and number of classes in data.
  outcome_classes <- get_outcome_class_levels(data)
  n_classes <- length(outcome_classes)

  # Skip calculation if there are no other classes.
  if (n_classes <= 1L) return(NULL)
  
  # Remove any entries that lack valid predictions.
  data <- remove_invalid_predictions(data)
  
  # Remove any entries that lack observed values.
  data <- filter_missing_outcome(data)
  if (is_empty(data)) return(NULL)
  
  # Identify the outcome column.
  outcome_column <- get_outcome_columns(x = data@outcome_type)
  data <- .as_data_table(.complete(data))
  
  # Create empty scalars
  tp <- tn <- fp <- fn <- prevalence <- bias <- numeric(n_classes)

  # Total number of samples
  n_samples <- nrow(data)

  # Iterate over positive classes
  for (ii in seq_along(outcome_classes)) {
    # Determine true positives, true negatives, false positives and false negatives
    tp[ii] <- nrow(data[predicted_class == outcome_classes[ii] & get(outcome_column) == outcome_classes[ii], ])
    tn[ii] <- nrow(data[predicted_class != outcome_classes[ii] & get(outcome_column) != outcome_classes[ii], ])
    fp[ii] <- nrow(data[predicted_class == outcome_classes[ii] & get(outcome_column) != outcome_classes[ii], ])
    fn[ii] <- nrow(data[predicted_class != outcome_classes[ii] & get(outcome_column) == outcome_classes[ii], ])

    # Prevalence (fraction of observed outcome_column class)
    prevalence[ii] <- nrow(data[get(outcome_column) == outcome_classes[ii]]) / n_samples

    # Bias (fraction of predicted positive class)
    bias[ii] <- nrow(data[predicted_class == outcome_classes[ii]]) / n_samples
  }

  return(list(
    "tp" = tp,
    "tn" = tn,
    "fp" = fp,
    "fn" = fn,
    "prevalence" = prevalence,
    "bias" = bias,
    "n_samples" = n_samples,
    "n_classes" = n_classes
  ))
}



..aggregate_confusion_matrix <- function(cm, averaging, outcome_type) {
  if (outcome_type == "binomial") {
    # Use second class as positive class
    return(list(
      "tp" = cm$tp[2L],
      "tn" = cm$tn[2L],
      "fp" = cm$fp[2L],
      "fn" = cm$fn[2L],
      "prevalence" = 1.0,
      "bias" = 1.0,
      "n_samples" = cm$n_samples,
      "n_classes" = 1.0
    ))
    
  } else if (outcome_type == "multinomial") {
    if (averaging == "micro") {
      # Merge confusion matrix for micro-averaging.
      return(list(
        "tp" = sum(cm$tp),
        "tn" = sum(cm$tn),
        "fp" = sum(cm$fp),
        "fn" = sum(cm$fn),
        "prevalence" = 1.0,
        "bias" = 1.0,
        "n_samples" = cm$n_classes * cm$n_samples,
        "n_classes" = 1.0
      ))
      
    } else {
      # No changes are needed.
      return(cm)
    }
    
  } else {
    ..error_no_known_outcome_type(outcome_type)
  }
}



..aggregate_confusion_matrix_score <- function(score, cm, averaging, outcome_type) {
  # Replace all non-finite values.
  score[!is.finite(score)] <- NA_real_

  if (all(is.na(score))) return(NA_real_)

  # Aggregation is only required for multinomial outcomes.
  if (outcome_type == "multinomial") {
    if (averaging == "weighted") {
      # Weighted averaging
      score <- sum(cm$prevalence * score, na.rm = TRUE)
      
    } else if (averaging == "macro") {
      # Compute the average.
      score <- mean(score, na.rm = TRUE)
      
    } else if (averaging == "micro") {
      # Micro averaging
      score <- score
      
    } else {
      ..error_reached_unreachable_code(
        "..aggregate_confusion_matrix_score: encountered unknown averaging method."
      )
    }
  }

  return(score)
}



.get_available_confusion_matrix_metrics <- function() {
  return(c(
    .get_available_accuracy_metrics(),
    .get_available_balanced_accuracy_metrics(),
    .get_available_balanced_error_rate_metrics(),
    .get_available_cohen_kappa_metrics(),
    .get_available_f1_score_metrics(),
    .get_available_fdr_metrics(),
    .get_available_informedness_metrics(),
    .get_available_markedness_metrics(),
    .get_available_mcc_metrics(),
    .get_available_npv_metrics(),
    .get_available_ppv_metrics(),
    .get_available_sensitivity_metrics(),
    .get_available_specificity_metrics(),
    .get_available_youden_metrics()
  ))
}



.get_available_accuracy_metrics <- function() {
  return(c("accuracy"))
}

.get_available_balanced_accuracy_metrics <- function() {
  return(c("balanced_accuracy", "bac"))
}

.get_available_balanced_error_rate_metrics <- function() {
  return(c("balanced_error_rate", "ber"))
}

.get_available_cohen_kappa_metrics <- function() {
  return(c("cohen_kappa", "kappa"))
}

.get_available_f1_score_metrics <- function() {
  return(paste0("f1_score", c("", "_micro", "_macro", "_weighted")))
}

.get_available_fdr_metrics <- function() {
  return(c(
    paste0("fdr", c("", "_micro", "_macro", "_weighted")),
    paste0("false_discovery_rate", c("", "_micro", "_macro", "_weighted"))
  ))
}

.get_available_informedness_metrics <- function() {
  return("informedness")
}

.get_available_markedness_metrics <- function() {
  return("markedness")
}

.get_available_mcc_metrics <- function() {
  return(c("mcc", "matthews_correlation_coefficient"))
}

.get_available_npv_metrics <- function() {
  return(c(
    paste0("npv", c("", "_micro", "_macro", "_weighted")),
    paste0("negative_predictive_value", c("", "_micro", "_macro", "_weighted"))
  ))
}

.get_available_ppv_metrics <- function() {
  return(c(
    paste0("ppv", c("", "_micro", "_macro", "_weighted")),
    paste0("positive_predictive_value", c("", "_micro", "_macro", "_weighted")),
    paste0("precision", c("", "_micro", "_macro", "_weighted"))
  ))
}

.get_available_sensitivity_metrics <- function() {
  return(c(
    paste0("sensitivity", c("", "_micro", "_macro", "_weighted")),
    paste0("recall", c("", "_micro", "_macro", "_weighted")),
    paste0("tpr", c("", "_micro", "_macro", "_weighted")),
    paste0("true_positive_rate", c("", "_micro", "_macro", "_weighted"))
  ))
}

.get_available_specificity_metrics <- function() {
  return(c(
    paste0("specificity", c("", "_micro", "_macro", "_weighted")),
    paste0("tnr", c("", "_micro", "_macro", "_weighted")),
    paste0("true_negative_rate", c("", "_micro", "_macro", "_weighted"))
  ))
}

.get_available_youden_metrics <- function() {
  return(c(
    paste0("youden_j", c("", "_micro", "_macro", "_weighted")),
    paste0("youden_index", c("", "_micro", "_macro", "_weighted"))
  ))
}

familiar:::test_all_metrics_available(
  metrics = familiar:::.get_available_confusion_matrix_metrics()
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

# power.transform and other packages are required.
if (!rlang::is_installed("power.transform")) testthat::skip()

cm_metric_test <- function(
    metric, 
    data_list, 
    baseline_value, 
    expected_score, 
    expected_objective = NULL
) {
  # For several metrics the objective score and the expected score are the same.
  if (is.null(expected_objective)) expected_objective <- expected_score

  # Extract the correct expected_score and expected_objective variables.
  if (endsWith(x = metric, suffix = "_macro")) {
    expected_score <- expected_score$macro
    expected_objective <- expected_objective$macro
  } else if (endsWith(x = metric, suffix = "_micro")) {
    expected_score <- expected_score$micro
    expected_objective <- expected_objective$micro
  } else if (endsWith(x = metric, suffix = "_weighted")) {
    expected_score <- expected_score$weighted
    expected_objective <- expected_objective$weighted
  } else if (is.list(expected_score)) {
    expected_score <- expected_score$macro
    expected_objective <- expected_objective$macro
  }

  for (ii in seq_along(data_list)) {
    # Create metric object
    metric_object <- familiar:::as_metric(
      metric = metric,
      outcome_type = data_list[[ii]]@outcome_type
    )

    # Set baseline-value explicitly.
    metric_object@baseline_value <- baseline_value

    # Check that the metric is available
    testthat::expect_true(familiar:::is_available(metric_object))

    # Compute the metric value.
    score <- familiar:::compute_metric_score(
      metric = metric_object,
      data = data_list[[ii]]
    )

    # Compute the objective score.
    objective_score <- familiar:::compute_objective_score(
      metric = metric_object,
      data = data_list[[ii]]
    )

    # Test the values.
    testthat::expect_equal(score, expected_score[ii])
    testthat::expect_equal(objective_score, expected_objective[ii])
  }
}
data_good_binomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1, 1, 1, 1, 1, 0, 0, 0, 0, 0),
    "b" = c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "a", "a", "b", "b", "b", "b", "b"), levels = c("a", "b"))),
  type = "classification"
)

data_ok_binomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1.0, 0.9, 0.8, 0.7, 0.3, 0.7, 0.6, 0.2, 0.1, 0.0),
    "b" = c(0.0, 0.1, 0.2, 0.3, 0.7, 0.3, 0.4, 0.8, 0.9, 1.0)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "a", "a", "b", "b", "b", "b", "b"), levels = c("a", "b"))),
  type = "classification"
)

data_bad_binomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1),
    "b" = c(1, 1, 1, 1, 1, 0, 0, 0, 0, 0)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "a", "a", "b", "b", "b", "b", "b"), levels = c("a", "b"))),
  type = "classification"
)

data_good_multinomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1, 1, 1, 0, 0, 0, 0, 0, 0, 0),
    "b" = c(0, 0, 0, 1, 1, 1, 0, 0, 0, 0),
    "c" = c(0, 0, 0, 0, 0, 0, 1, 1, 1, 1)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "b", "b", "b", "c", "c", "c", "c"), levels = c("a", "b", "c"))),
  type = "classification"
)

data_ok_multinomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1.0, 0.4, 0.3, 0.7, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0),
    "b" = c(0.0, 0.3, 0.4, 0.2, 1.0, 0.4, 0.0, 0.4, 0.4, 0.3),
    "c" = c(0.0, 0.3, 0.3, 0.1, 0.0, 0.5, 1.0, 0.6, 0.6, 0.7)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "b", "b", "b", "c", "c", "c", "c"), levels = c("a", "b", "c"))),
  type = "classification"
)

data_bad_multinomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(0, 0, 0, 1, 1, 1, 0, 0, 0, 0),
    "b" = c(0, 0, 0, 0, 0, 0, 1, 1, 1, 1),
    "c" = c(1, 1, 1, 0, 0, 0, 0, 0, 0, 0)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "b", "b", "b", "c", "c", "c", "c"), levels = c("a", "b", "c"))),
  type = "classification"
)


# Package data in a list for easier iterative tests
data_list <- list(
  "good_binomial" = data_good_binomial,
  "bad_binomial" = data_bad_binomial,
  "ok_binomial" = data_ok_binomial,
  "good_multinomial" = data_good_multinomial,
  "bad_multinomial" = data_bad_multinomial,
  "ok_multinomial" = data_ok_multinomial
)


# Confusion matrix -------------------------------------------------------------
testthat::test_that("confusion matrix is correct", {
  # Good binomial
  cm_gb <- familiar:::..compute_confusion_matrix_data(data = data_good_binomial)

  testthat::expect_equal(cm_gb$tp, c(5, 5))
  testthat::expect_equal(cm_gb$tn, c(5, 5))
  testthat::expect_equal(cm_gb$fp, c(0, 0))
  testthat::expect_equal(cm_gb$fn, c(0, 0))
  testthat::expect_equal(cm_gb$prevalence, c(0.5, 0.5))
  testthat::expect_equal(cm_gb$bias, c(0.5, 0.5))
  testthat::expect_equal(cm_gb$n_samples, 10)
  testthat::expect_equal(cm_gb$n_classes, 2)

  # Bad binomial
  cm_bb <- familiar:::..compute_confusion_matrix_data(data = data_bad_binomial)

  testthat::expect_equal(cm_bb$tp, c(0, 0))
  testthat::expect_equal(cm_bb$tn, c(0, 0))
  testthat::expect_equal(cm_bb$fp, c(5, 5))
  testthat::expect_equal(cm_bb$fn, c(5, 5))
  testthat::expect_equal(cm_bb$prevalence, c(0.5, 0.5))
  testthat::expect_equal(cm_bb$bias, c(0.5, 0.5))
  testthat::expect_equal(cm_bb$n_samples, 10)
  testthat::expect_equal(cm_bb$n_classes, 2)

  # Ok binomial
  cm_ob <- familiar:::..compute_confusion_matrix_data(data = data_ok_binomial)

  testthat::expect_equal(cm_ob$tp, c(4, 3))
  testthat::expect_equal(cm_ob$tn, c(3, 4))
  testthat::expect_equal(cm_ob$fp, c(2, 1))
  testthat::expect_equal(cm_ob$fn, c(1, 2))
  testthat::expect_equal(cm_ob$prevalence, c(0.5, 0.5))
  testthat::expect_equal(cm_ob$bias, c(0.6, 0.4))
  testthat::expect_equal(cm_ob$n_samples, 10)
  testthat::expect_equal(cm_ob$n_classes, 2)

  # Good multinomial
  cm_gm <- familiar:::..compute_confusion_matrix_data(data = data_good_multinomial)

  testthat::expect_equal(cm_gm$tp, c(3, 3, 4))
  testthat::expect_equal(cm_gm$tn, c(7, 7, 6))
  testthat::expect_equal(cm_gm$fp, c(0, 0, 0))
  testthat::expect_equal(cm_gm$fn, c(0, 0, 0))
  testthat::expect_equal(cm_gm$prevalence, c(0.3, 0.3, 0.4))
  testthat::expect_equal(cm_gm$bias, c(0.3, 0.3, 0.4))
  testthat::expect_equal(cm_gm$n_samples, 10)
  testthat::expect_equal(cm_gm$n_classes, 3)

  # Bad multinomial
  cm_bm <- familiar:::..compute_confusion_matrix_data(data = data_bad_multinomial)

  testthat::expect_equal(cm_bm$tp, c(0, 0, 0))
  testthat::expect_equal(cm_bm$tn, c(4, 3, 3))
  testthat::expect_equal(cm_bm$fp, c(3, 4, 3))
  testthat::expect_equal(cm_bm$fn, c(3, 3, 4))
  testthat::expect_equal(cm_bm$prevalence, c(0.3, 0.3, 0.4))
  testthat::expect_equal(cm_bm$bias, c(0.3, 0.4, 0.3))
  testthat::expect_equal(cm_bm$n_samples, 10)
  testthat::expect_equal(cm_bm$n_classes, 3)

  # Ok multinomial
  cm_om <- familiar:::..compute_confusion_matrix_data(data = data_ok_multinomial)

  testthat::expect_equal(cm_om$tp, c(2, 1, 4))
  testthat::expect_equal(cm_om$tn, c(6, 6, 5))
  testthat::expect_equal(cm_om$fp, c(1, 1, 1))
  testthat::expect_equal(cm_om$fn, c(1, 2, 0))
  testthat::expect_equal(cm_om$prevalence, c(0.3, 0.3, 0.4))
  testthat::expect_equal(cm_om$bias, c(0.3, 0.2, 0.5))
  testthat::expect_equal(cm_om$n_samples, 10)
  testthat::expect_equal(cm_om$n_classes, 3)
})

# Accuracy ---------------------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_accuracy_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("accuracy is correct", {
  for (metric in familiar:::.get_available_accuracy_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = c(1.0, 0.0, 7 / 10, 1.0, 0.0, 7 / 10)
    )
  }
})

# Balanced accuracy ------------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_balanced_accuracy_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE
)

testthat::test_that("balanced accuracy is correct", {
  for (metric in familiar:::.get_available_balanced_accuracy_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = c(1.0, 0.0, 7 / 10, 1.0, 0.0, 2 / 3))
  }
})

# Balanced error rate ----------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_balanced_error_rate_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE)

testthat::test_that("balanced error rate is correct", {
  for (metric in familiar:::.get_available_balanced_error_rate_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 1.0,
      expected_score = c(0.0, 1.0, 3 / 10, 0.0, 1.0, 1 / 3),
      expected_objective = c(1.0, 0.0, 7 / 10, 1.0, 0.0, 2 / 3)
    )
  }
})

# Cohen's kappa ----------------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_cohen_kappa_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE)

testthat::test_that("Cohen's kappa is correct", {
  for (metric in familiar:::.get_available_cohen_kappa_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = c(1.0, -1.0, 2 / 5, 1.0, -33 / 67, 35 / 65)
    )
  }
})

# F1 score ---------------------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_f1_score_metrics(),
  not_available_single_sample = "binomial",
  not_available_all_samples_identical = "binomial")

testthat::test_that("F1-score is correct", {
  for (metric in familiar:::.get_available_f1_score_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = list(
        "macro" = c(1.0, 0.0, 2 / 3, 1.0, 0.0, 88 / 135),
        "micro" = c(1.0, 0.0, 2 / 3, 1.0, 0.0, 7 / 10),
        "weighted" = c(1.0, 0.0, 2 / 3, 1.0, 0.0, 304 / 450)
      )
    )
  }
})

# False detection rate ---------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_fdr_metrics(),
  not_available_single_sample = "binomial",
  not_available_all_samples_identical = "binomial",
  not_available_all_predictions_identical = "binomial")

testthat::test_that("False detection rate is correct", {
  for (metric in familiar:::.get_available_fdr_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 1.0,
      expected_score = list(
        "macro" = c(0.0, 1.0, 1 / 4, 0.0, 1.0, 31 / 90),
        "micro" = c(0.0, 1.0, 1 / 4, 0.0, 1.0, 3 / 10),
        "weighted" = c(0.0, 1.0, 1 / 4, 0.0, 1.0, 33 / 100)
      ),
      expected_objective = list(
        "macro" = c(1.0, 0.0, 3 / 4, 1.0, 0.0, 59 / 90),
        "micro" = c(1.0, 0.0, 3 / 4, 1.0, 0.0, 7 / 10),
        "weighted" = c(1.0, 0.0, 3 / 4, 1.0, 0.0, 67 / 100)
      )
    )
  }
})

# Informedness -----------------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_informedness_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE)

testthat::test_that("Informedness is correct", {
  for (metric in familiar:::.get_available_informedness_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = c(1.0, -1.0, 2 / 5, 1.0, -1 / 2, 23 / 42)
    )
  }
})

# Markedness -------------------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_markedness_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE,
  not_available_all_predictions_identical = TRUE
)

testthat::test_that("Markedness is correct", {
  for (metric in familiar:::.get_available_markedness_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = c(1.0, -1.0, 5 / 12, 1.0, -1 / 2, 17 / 28)
    )
  }
})

# Matthews' correlation coefficient --------------------------------------------
familiar:::test_all_metrics(metrics = familiar:::.get_available_mcc_metrics())

testthat::test_that("Matthew's correlation coefficient is correct", {
  for (metric in familiar:::.get_available_mcc_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = c(1.0, -1.0, 1 / sqrt(6), 1.0, -1 / 2, 35 / sqrt(4092))
    )
  }
})

# Negative predictive value ----------------------------------------------------
familiar:::test_all_metrics(metrics = familiar:::.get_available_npv_metrics())

testthat::test_that("Negative predictive value is correct", {
  for (metric in familiar:::.get_available_npv_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = list(
        "macro" = c(1.0, 0.0, 2 / 3, 1.0, 1 / 2, 73 / 84),
        "micro" = c(1.0, 0.0, 2 / 3, 1.0, 1 / 2, 17 / 20),
        "weighted" = c(1.0, 0.0, 2 / 3, 1.0, 69 / 140, 247 / 280)
      )
    )
  }
})

# Positive predictive value ----------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_ppv_metrics(),
  not_available_single_sample = "binomial",
  not_available_all_samples_identical = "binomial",
  not_available_all_predictions_identical = "binomial")

testthat::test_that("Positive predictive value is correct", {
  for (metric in familiar:::.get_available_ppv_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = list(
        "macro" = c(1.0, 0.0, 3 / 4, 1.0, 0.0, 59 / 90),
        "micro" = c(1.0, 0.0, 3 / 4, 1.0, 0.0, 7 / 10),
        "weighted" = c(1.0, 0.0, 3 / 4, 1.0, 0.0, 67 / 100)
      )
    )
  }
})

# Recall -----------------------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_sensitivity_metrics(),
  not_available_single_sample = "binomial",
  not_available_all_samples_identical = "binomial")

testthat::test_that("Recall is correct", {
  for (metric in familiar:::.get_available_sensitivity_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = list(
        "macro" = c(1.0, 0.0, 3 / 5, 1.0, 0.0, 2 / 3),
        "micro" = c(1.0, 0.0, 3 / 5, 1.0, 0.0, 7 / 10),
        "weighted" = c(1.0, 0.0, 3 / 5, 1.0, 0.0, 7 / 10)
      )
    )
  }
})

# Specificity ------------------------------------------------------------------
familiar:::test_all_metrics(metrics = familiar:::.get_available_specificity_metrics())

testthat::test_that("Specificity is correct", {
  for (metric in familiar:::.get_available_specificity_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = list(
        "macro" = c(1.0, 0.0, 4 / 5, 1.0, 1 / 2, 107 / 126),
        "micro" = c(1.0, 0.0, 4 / 5, 1.0, 1 / 2, 17 / 20),
        "weighted" = c(1.0, 0.0, 4 / 5, 1.0, 1 / 2, 89 / 105)
      )
    )
  }
})

# Youden's J statistic ---------------------------------------------------------
# Test results depend on averaging method.
metric_youden <- familiar:::.get_available_youden_metrics()
metric_non_micro <- metric_youden[!endsWith(x = metric_youden, suffix = "_micro")]
metric_micro <- metric_youden[endsWith(x = metric_youden, suffix = "_micro")]

familiar:::test_all_metrics(
  metrics = metric_non_micro,
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE
)

familiar:::test_all_metrics(
  metrics = metric_micro,
  not_available_single_sample = "binomial",
  not_available_all_samples_identical = "binomial"
)

testthat::test_that("Youden's J statistic is correct", {
  for (metric in familiar:::.get_available_youden_metrics()) {
    cm_metric_test(
      data_list = data_list,
      metric = metric,
      baseline_value = 0.0,
      expected_score = list(
        "macro" = c(1.0, -1.0, 2 / 5, 1.0, -1 / 2, 65 / 126),
        "micro" = c(1.0, -1.0, 2 / 5, 1.0, -1 / 2, 11 / 20),
        "weighted" = c(1.0, -1.0, 2 / 5, 1.0, -1 / 2, 23 / 42)
      )
    )
  }
})

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarMetricAUCROC
setClass(
  "familiarMetricAUCROC",
  contains = "familiarMetric",
  prototype = methods::prototype(
    metric = "auc_roc",
    outcome_type = NA_character_,
    name = "Area under Receiver Operating Characteristic Curve",
    baseline_value = 0.5,
    value_range = c(0.0, 1.0),
    higher_better = TRUE
  )
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMetricAUCROC"),
  function(object, ...) {
    return(object@outcome_type %in% c("binomial", "multinomial"))
  }
)



# compute_metric_score ---------------------------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricAUCROC"),
  function(metric, data, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    outcome <- NULL

    if (is_empty(data)) return(callNextMethod())

    if (!is(data, "predictionTableClassification")) {
      ..error_data_not_prediction_table(data, "predictionTableClassification")
    }

    # Get the classes and number of classes in data.
    outcome_classes <- get_outcome_class_levels(data)
    n_classes <- length(outcome_classes)

    # Skip calculation if an AUC cannot be computed.
    if (n_classes <= 1L) return(callNextMethod())

    # Remove any entries that lack valid predictions.
    data <- remove_invalid_predictions(data)
    
    # Remove any entries that lack observed values.
    data <- filter_missing_outcome(data)
    if (is_empty(data)) return(callNextMethod())
    
    data <- .as_data_table(data)
    if (nrow(data) <= 1L) return(callNextMethod())

    # Define class combinations (>1 in case of multinomial outcomes)
    class_combinations <- utils::combn(outcome_classes, 2L)
    n_class_combinations <- ncol(class_combinations)

    # Generate empty auc vector AUC of the ROC is calculated according to Hand,
    # D.J, and Till, R.J. A simple generalisation of the area under the ROC
    # curve for multiple class classification problems, Machine Learning 45
    # 171-186 (2001)
    auc_score <- vector(mode = "numeric", length = n_class_combinations)

    # Iterate over combinations
    for (ii in seq_len(n_class_combinations)) {
      # Find the current positive and negative classes
      positive_class <- class_combinations[1L, ii]
      negative_class <- class_combinations[2L, ii]
      
      # Get the probabilities that correspond to the positive and # negative
      # class in outcome (g and f in Hand et al.).
      class_probability_positive <- data[outcome == positive_class, ][[positive_class]]
      class_probability_negative <- data[outcome == negative_class, ][[positive_class]]
      
      # Set x and y_obs
      x <- c(class_probability_positive, class_probability_negative)
      
      y_obs <- c(
        rep_len(TRUE, length(class_probability_positive)),
        rep_len(FALSE, length(class_probability_negative))
      )
      
      # Compute AUC value for the current combination.
      auc_score[ii] <- ..compute_auc_roc(
        x = x,
        y_obs = y_obs
      )
    }

    # Calculate mean AUC (eq. 7 from Hand et al.). This has no effect for
    # binomial AUC.
    if (!any(is.finite(auc_score))) return(callNextMethod())
    auc_score <- mean(auc_score, na.rm = TRUE)

    return(auc_score)
  }
)


.compute_auc_roc <- function(x, y_obs) {
  if (!is.factor(y_obs)) ..error("y_obs should be a categorical value")
  # Called from functions other then compute_metric_score.
  outcome_classes <- levels(y_obs)
  
  # Define class combinations (>1 in case of multinomial outcomes)
  class_combinations <- utils::combn(outcome_classes, 2L)
  n_class_combinations <- ncol(class_combinations)
  
  # Generate empty auc vector AUC of the ROC is calculated according to Hand,
  # D.J, and Till, R.J. A simple generalisation of the area under the ROC
  # curve for multiple class classification problems, Machine Learning 45
  # 171-186 (2001)
  auc_score <- vector(mode = "numeric", length = n_class_combinations)
  
  # Iterate over combinations
  for (ii in seq_len(n_class_combinations)) {
    # Find the current positive and negative classes
    positive_class <- class_combinations[1L, ii]
    negative_class <- class_combinations[2L, ii]
    
    # Get the probabilities that correspond to the positive and # negative
    # class in outcome (g and f in Hand et al.).
    x_positive <- x[y_obs == positive_class]
    x_negative <- x[y_obs == negative_class]
    
    # Compute AUC value for the current combination.
    auc_score[ii] <- ..compute_auc_roc(
      x = c(x_positive, x_negative),
      y_obs = c(
        rep_len(TRUE, length(x_positive)),
        rep_len(FALSE, length(x_negative))
      )
    )
  }
  
  # Calculate mean AUC (eq. 7 from Hand et al.). This has no effect for
  # binomial AUC.
  if (!any(is.finite(auc_score))) return(NA_real_)
  auc_score <- mean(auc_score, na.rm = TRUE)
  
  return(auc_score)
}


..compute_auc_roc <- function(x, y_obs) {
  # Called directly from compute_metric_score
  if (!is.logical(y_obs)) ..error("y_obs should contain only boolean values")
  n_positive <- sum(y_obs)
  n_negative <- length(y_obs) - n_positive
  
  if (n_positive > 0L & n_negative > 0L) {
    
    # Determine ranks
    sample_rank <- data.table::frank(
      x = x,
      ties.method = "average"
    )
    
    # Calculate AUC
    auc <- (sum(sample_rank[seq_len(n_positive)]) - n_positive * (n_positive + 1L) / 2.0) /
      (n_positive * n_negative)
    
  } else {
    auc <- NA_real_
  }
  
  return(auc)
}



.get_available_auc_roc_metrics <- function() {
  return(c("auc", "auc_roc"))
}

familiar:::test_all_metrics_available(metrics = familiar:::.get_available_auc_roc_metrics())

# Skip remainder on CRAN due to runtimes.
testthat::skip_on_cran()

# power.transform and other packages are required.
if (!rlang::is_installed("power.transform")) testthat::skip()

familiar:::test_all_metrics(
  metrics = familiar:::.get_available_auc_roc_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE
)


data_good_binomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1, 1, 1, 1, 1, 0, 0, 0, 0, 0),
    "b" = c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "a", "a", "b", "b", "b", "b", "b"), levels = c("a", "b"))),
  type = "classification"
)

data_bad_binomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(.5, .5, .5, .5, .5, .5, .5, .5, .5, .5),
    "b" = c(.5, .5, .5, .5, .5, .5, .5, .5, .5, .5)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "a", "a", "b", "b", "b", "b", "b"), levels = c("a", "b"))),
  type = "classification"
)

data_ok_binomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1.0, 0.9, 0.8, 0.4, 0.3, 0.7, 0.6, 0.2, 0.1, 0.0),
    "b" = c(0.0, 0.1, 0.2, 0.6, 0.7, 0.3, 0.4, 0.8, 0.9, 1.0)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "a", "a", "b", "b", "b", "b", "b"), levels = c("a", "b"))),
  type = "classification"
)

data_inv_binomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1),
    "b" = c(1, 1, 1, 1, 1, 0, 0, 0, 0, 0)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "a", "a", "b", "b", "b", "b", "b"), levels = c("a", "b"))),
  type = "classification"
)

data_good_multinomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1, 1, 1, 0, 0, 0, 0, 0, 0, 0),
    "b" = c(0, 0, 0, 1, 1, 1, 0, 0, 0, 0),
    "c" = c(0, 0, 0, 0, 0, 0, 1, 1, 1, 1)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "b", "b", "b", "c", "c", "c", "c"), levels = c("a", "b", "c"))),
  type = "classification"
)

data_bad_multinomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3),
    "b" = c(1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3),
    "c" = c(1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3, 1 / 3)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "b", "b", "b", "c", "c", "c", "c"), levels = c("a", "b", "c"))),
  type = "classification"
)

data_ok_multinomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(1.0, 0.2, 0.4, 0.0, 0.4, 0.1, 0.0, 0.0, 0.0, 0.1),
    "b" = c(0.0, 0.5, 0.3, 1.0, 0.3, 0.5, 0.0, 0.6, 0.4, 0.3),
    "c" = c(0.0, 0.3, 0.3, 0.0, 0.3, 0.4, 1.0, 0.4, 0.6, 0.7)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "b", "b", "b", "c", "c", "c", "c"), levels = c("a", "b", "c"))),
  type = "classification"
)

data_inv_multinomial <- familiar::as_prediction_table(
  x = data.table::data.table(
    "a" = c(0, 0, 0, 0, 0, 0, 1, 1, 1, 1),
    "b" = c(1, 1, 1, 0, 0, 0, 0, 0, 0, 0),
    "c" = c(0, 0, 0, 1, 1, 1, 0, 0, 0, 0)
  ),
  y = list("outcome" = factor(c("a", "a", "a", "b", "b", "b", "c", "c", "c", "c"), levels = c("a", "b", "c"))),
  type = "classification"
)

# Package data in a list for easier iterative tests
data_list <- list(
  "good_binomial" = data_good_binomial,
  "bad_binomial" = data_bad_binomial,
  "ok_binomial" = data_ok_binomial,
  "inv_binomial" = data_inv_binomial,
  "good_multinomial" = data_good_multinomial,
  "bad_multinomial" = data_bad_multinomial,
  "ok_multinomial" = data_ok_multinomial,
  "inv_multinomial" = data_inv_multinomial
)

# Area under the curve ---------------------------------------------------------
testthat::test_that("AUC-ROC is correct", {
  expected_score <- c(1.0, 0.5, 21 / 25, 0.0, 1.0, 0.5, 61 / 72, 1 / 3)
  expected_objective <- c(1.0, 0.0, 17 / 25, -1.0, 1.0, 0.0, 25 / 36, -1 / 3)
  
  # Iterate over the data sets.
  for (ii in seq_along(data_list)) {
    # Create metric object.
    metric_object <- familiar:::as_metric(
      metric = "auc_roc",
      outcome_type = data_list[[ii]]@outcome_type
    )
    
    # Set baseline-value explicitly.
    metric_object@baseline_value <- 0.5
    
    # Check that the metric is available
    testthat::expect_true(familiar:::is_available(metric_object))
    
    # Compute the metric value.
    score <- familiar:::compute_metric_score(
      metric = metric_object,
      data = data_list[[ii]]
    )
    
    # Compute the objective score.
    objective_score <- familiar:::compute_objective_score(
      metric = metric_object,
      data = data_list[[ii]]
    )
    
    # Test the values.
    testthat::expect_equal(score, expected_score[ii])
    testthat::expect_equal(objective_score, expected_objective[ii])
  }
})

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test.
familiar:::test_plots(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = c("binomial", "multinomial", "survival"),
  not_available_single_sample = c("binomial", "multinomial", "survival"),
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  debug = debug_flag
)

# Violin-plot (categorical) ----------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list("facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list("facet_by" = c("data_set", "learner", "vimp_method")),
  use_prediction_table = TRUE,
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "facet_by" = c("data_set", "learner", "vimp_method"),
    "x_axis_by" = "metric"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method"),
    "x_axis_by" = "data_set",
    "color_by" = "metric"),
  debug = debug_flag
)

# Violin-plot (numeric) --------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("continuous"),
  plot_args = list("facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("continuous"),
  plot_args = list("facet_by" = c("data_set", "learner", "vimp_method")),
  use_prediction_table = TRUE,
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list(
    "facet_by" = c("data_set", "learner", "vimp_method"),
    "x_axis_by" = "metric"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method"),
    "x_axis_by" = "data_set",
    "color_by" = "metric"),
  debug = debug_flag
)

# Violin-plot (survival) -------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("survival"),
  plot_args = list("facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("survival"),
  plot_args = list("facet_by" = c("data_set", "learner", "vimp_method")),
  use_prediction_table = TRUE,
  debug = debug_flag
)

# Bar plot (categorical) -------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "plot_type" = "barplot",
    "facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "plot_type" = "barplot",
    "facet_by" = c("data_set", "learner", "vimp_method"),
    "x_axis_by" = "metric"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "plot_type" = "barplot",
    "facet_by" = c("learner", "vimp_method"),
    "x_axis_by" = "data_set",
    "color_by" = "metric"),
  debug = debug_flag
)

# Bar plot (numeric) -----------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("continuous"),
  plot_args = list(
    "plot_type" = "barplot",
    "facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list(
    "plot_type" = "barplot",
    "facet_by" = c("data_set", "learner", "vimp_method"),
    "x_axis_by" = "metric"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list(
    "plot_type" = "barplot",
    "facet_by" = c("learner", "vimp_method"),
    "x_axis_by" = "data_set",
    "color_by" = "metric"),
  debug = debug_flag
)

# Bar plot (survival) ----------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("survival"),
  plot_args = list(
    "plot_type" = "barplot",
    "facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

# Box plot (categorical) -------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "plot_type" = "boxplot",
    "facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "plot_type" = "boxplot",
    "facet_by" = c("data_set", "learner", "vimp_method"),
    "x_axis_by" = "metric"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "plot_type" = "boxplot",
    "facet_by" = c("learner", "vimp_method"),
    "x_axis_by" = "data_set",
    "color_by" = "metric"),
  debug = debug_flag
)

# Box plot (numeric) -----------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("continuous"),
  plot_args = list(
    "plot_type" = "boxplot",
    "facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list(
    "plot_type" = "boxplot",
    "facet_by" = c("data_set", "learner", "vimp_method"),
    "x_axis_by" = "metric"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list(
    "plot_type" = "boxplot",
    "facet_by" = c("learner", "vimp_method"),
    "x_axis_by" = "data_set",
    "color_by" = "metric"),
  debug = debug_flag
)

# Box plot (survival) ----------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("survival"),
  plot_args = list(
    "plot_type" = "boxplot",
    "facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

# Heatmap (categorical) --------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list("plot_type" = "heatmap"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list("plot_type" = "heatmap"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "plot_type" = "heatmap",
    "facet_by" = c("data_set", "metric"),
    "x_axis_by" = "learner",
    "y_axis_by" = "vimp_method"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "plot_type" = "heatmap",
    "annotate_performance" = "value_ci",
    "facet_by" = c("data_set", "metric"),
    "x_axis_by" = "learner",
    "y_axis_by" = "vimp_method"),
  debug = debug_flag
)

# Heatmap (numeric) ------------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list("plot_type" = "heatmap"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list("plot_type" = "heatmap"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list(
    "plot_type" = "heatmap",
    "facet_by" = c("data_set", "metric"),
    "x_axis_by" = "learner",
    "y_axis_by" = "vimp_method"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  metric = c("rmse", "mae"),
  outcome_type_available = c("continuous"),
  plot_args = list(
    "plot_type" = "heatmap",
    "annotate_performance" = "value_ci",
    "facet_by" = c("data_set", "metric"),
    "x_axis_by" = "learner",
    "y_axis_by" = "vimp_method"),
  debug = debug_flag
)

# Heatmap (survival) -----------------------------------------------------------
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_model_performance,
  data_element = "model_performance",
  outcome_type_available = c("survival"),
  plot_args = list("plot_type" = "heatmap"),
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

setClass(
  "familiarCorrelationVimp",
  contains = "familiarVimpMethod"
)



.get_available_correlation_vimp_methods <- function(show_general = TRUE) {
  return(c("pearson", "spearman", "kendall"))
}



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarCorrelationVimp"),
  function(object, ...) {
    
    if (object@outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(object@outcome_type %in% c("continuous", "survival"))
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarCorrelationVimp"),
  function(object, data = NULL, ...) {
    return(list())
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarCorrelationVimp"),
  function(
    object,
    data,
    cl = NULL,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    outcome_event <- NULL

    if (is_empty(data)) return(callNextMethod())

    # Drop non-event data for censored data analysis for calculating correlation
    # and set outcome column.
    if (object@outcome_type == "survival") {
      data@data <- data@data[outcome_event == 1L, ]

      # Check whether the filtered data does not allow for assessing variable
      # importance.
      if (has_bad_training_data(object = object, data = data)) {
        return(callNextMethod)
      }
    }

    # Use effect coding to convert categorical data into encoded data - this is
    # required to deal with factors with missing/new levels between training and
    # test data sets.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )

    # Find feature columns in the data.
    feature_columns <- get_feature_columns(x = encoded_data$encoded_data)

    if (object@outcome_type =="survival") {
      outcome <- encoded_data$encoded_data@data$outcome_time
      
    } else {
      outcome <- encoded_data$encoded_data@data$outcome
    }
    
    # Compute correlation coefficients.
    correlation_coefficients <- fam_sapply(
      cl = cl,
      X = encoded_data$encoded_data@data[, mget(feature_columns)],
      FUN = stats::cor,
      y = outcome,
      method = object@vimp_method,
      chopchop = TRUE
    )

    # Create variable importance object.
    vimp_object <- methods::new("vimpTable",
      vimp_table = data.table::data.table(
        "score" = abs(correlation_coefficients),
        "name" = feature_columns
      ),
      encoding_table = encoded_data$reference_table,
      score_aggregation = "max",
      invert = TRUE
    )

    return(vimp_object)
  }
)

familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_correlation_vimp_methods(show_general = TRUE)
)

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_correlation_vimp_methods(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE
)
familiar:::test_all_vimp_methods(
  familiar:::.get_available_correlation_vimp_methods(show_general = FALSE)
)
familiar:::test_all_vimp_methods_parallel(
  familiar:::.get_available_correlation_vimp_methods(show_general = FALSE)
)

# Continuous outcome -----------------------------------------------------------
data <- familiar:::test_create_good_data("continuous")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "spearman",
  vimp_method_parameter_list = NULL,
  outcome_type = "continuous",
  cluster_method = "none",
  imputation_method = "simple"
)

testthat::test_that(paste0("Spearman correlation correctly ranks continuous data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)
  ))

  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

# Survival outcome -------------------------------------------------------------
data <- familiar:::test_create_good_data("survival")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "spearman",
  vimp_method_parameter_list = NULL,
  outcome_type = "survival",
  cluster_method = "none",
  imputation_method = "simple"
)

testthat::test_that(paste0(
  "Spearman correlation correctly ranks survival outcome data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)
  ))

  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

.load_data <- function(
    data, 
    sample_id_column = NULL,
    batch_id_column = NULL, 
    series_id_column = NULL,
    ...
) {
  # Parse the input sample_id_column
  if (!is.null(sample_id_column)) {
    sample_id_column <- .replace_illegal_column_name(sample_id_column)
  }

  # Parse the input batch_id_column
  if (!is.null(batch_id_column)) {
    batch_id_column <- .replace_illegal_column_name(batch_id_column)
  }

  # Parse the input series_id_column
  if (!is.null(series_id_column)) {
    series_id_column <- .replace_illegal_column_name(series_id_column)
  }

  if (data.table::is.data.table(data)) {
    # Keep data as is.
    data <- data.table::copy(data)

    # Update column names using a fixed routine
    data.table::setnames(data, .replace_illegal_column_name(colnames(data)))
  } else if (is.data.frame(data)) {
    
    # Convert to data.table
    data <- data.table::as.data.table(data)

    # Update column names using a fixed routine
    data.table::setnames(data, .replace_illegal_column_name(colnames(data)))
  } else if (is.character(data) && length(data) == 1L) {
    
    # Read from path
    if (!file.exists(data)) {
      ..error(paste0("The requested data file does not exist: ", data))
    }

    # Load data based on file extension
    file_extension <- tolower(.file_extension(data))

    if (file_extension == "csv") {
      data <- .load_csv(data)
    } else if (file_extension == "rdata") {
      data <- .load_rdata(data)
    } else if (file_extension == "rds") {
      data <- .load_rds(data)
    } else {
      ..error(paste(
        "File extension", file_extension, "was not recognised as a loadable data type. Please load the",
        "data manually."
      ))
    }

    # Update column names using a fixed routine
    data.table::setnames(data, .replace_illegal_column_name(colnames(data)))
    
  } else if ((is.atomic(data) && length(data) > 1L) || (is.list(data))) {
    
    # Read data to a list by calling the function recursively
    data_list <- lapply(data, .load_data)

    if (length(data_list) > 1L) {
      new_data_list <- list()
      joined_id <- integer()

      # Attempt to bind rows
      for (ii in seq_len(length(data_list))) {
        # Check if the current data list entry was already joined previously
        if (ii %in% joined_id) next

        # Get current data
        current_data <- data_list[[ii]]

        # Iterate over remaining datasets
        for (jj in seq(ii + 1L, length(data_list))) {
          # Check that index jj does not access non-existing data
          if (jj > length(data_list)) break

          # Find colnames in current data
          current_data_cols <- colnames(current_data)

          # Find column names of the new dataset
          new_data_cols <- colnames(data_list[[jj]])

          # Determine overlap in columns
          overlap_cols <- intersect(current_data_cols, new_data_cols)

          # Check if there is at least 90% overlap in column names
          if (length(overlap_cols) / min(c(length(current_data_cols), length(new_data_cols))) < 0.90) {
            next
          }

          # Check if types for each column are consistent
          matching_types <- sapply(
            overlap_cols,
            function(col, x, y) {
              if (is.logical(x[[col]]) && is.logical(y[[col]])) {
                return(TRUE)
              } else if (is.character(x[[col]]) && is.character(y[[col]])) {
                return(TRUE)
              } else if (is.factor(x[[col]]) && is.factor(y[[col]])) {
                return(TRUE)
              } else if (is.numeric(x[[col]]) && is.numeric(y[[col]])) {
                return(TRUE)
              } else {
                return(FALSE)
              }
            },
            x = current_data,
            y = data_list[[jj]]
          )
          
          # Throw an error if any class is not consistent.
          if (!all(matching_types)) {
            ..error(paste0(
              "Mismatching column classes were found between datasets ",
              ii, " and ", jj, ". Differences were found in columns:",
              paste_s(overlap_cols[!matching_types])
            ))
          }

          # Bind the datasets together
          current_data <- data.table::rbindlist(
            list(current_data, data_list[[jj]]),
            use.names = TRUE,
            fill = TRUE
          )

          # Mark the dataset as joined to prevent double usage.
          joined_id <- c(joined_id, jj)
        }

        # Append new data list with current_data
        new_data_list <- c(new_data_list, list(current_data))
      }

      # Replace data_list with new_data_list
      data_list <- new_data_list

      # Check if it makes sense to attempt to join datasets by index.
      if (length(data_list) > 1L) {
        # Get the left-hand dataset for merging
        data < data_list[[ii]]

        # Iterate over remaining datasets
        for (ii in seq(2L, length(data_list))) {
          if (is.null(sample_id_column)) {
            # Attempt cbind if both datasets have the same number of rows and
            # non-overlapping column names
            if (nrow(data) != nrow(data_list[[ii]])) {
              ..error(paste(
                "datasets could not be joined row-wise because the number of samples is different.",
                "Please provide a sample id column or merge the datasets yourself prior to input."
              ))
            }

            # Check for column overlap
            overlap_cols <- intersect(colnames(data), colnames(data_list[[ii]]))
            if (length(overlap_cols) > 0L) {
              ..error(paste(
                "datasets could not be joined row-wise as one or more columns with the same name",
                "appear in both the left and right-hand datasets:", paste_s(overlap_cols)
              ))
            }

            # Throw a warning, as row-wise binding is dangerous
            ..warning(paste(
              "datasets could be joined row-wise. Integrity of the data could not be ensured.",
              "Please ensure that each sample occupies same row for consistency. Alternatively,",
              "provide a sample id column or merge the datasets yourself prior to input."
            ))

            # Combine row-wise
            data <- cbind(data, data_list[[ii]])
            
          } else if (
            !is.null(sample_id_column) &&
            !is.null(batch_id_column) && 
            is.null(series_id_column)
          ) {
            # Attempt full join on sample_id_column and batch_id_column,
            # provided that other column names are not overlapping, and all
            # sample_ids for a batch are unique.

            # Check if the sample identifier columns is presents in the data
            if (
              !sample_id_column %in% colnames(data) ||
              !sample_id_column %in% colnames(data_list[[ii]])
            ) {
              ..error(paste(
                "The specified column of sample identifiers", sample_id_column,
                "was not found in one or more of the datasets."
              ))
            }

            # Check if the batch identifier column is present in the data
            if (!batch_id_column %in% colnames(data) ||
              !batch_id_column %in% colnames(data_list[[ii]])) {
              ..error(paste(
                "The specified column of batch identifiers", batch_id_column,
                "was not found in one or more of the datasets."
              ))
            }

            # Check uniqueness of identifiers
            if (
              anyDuplicated(data[, mget(sample_id_column, batch_id_column)]) > 0L ||
              anyDuplicated(data_list[[ii]][, mget(sample_id_column, batch_id_column)]) > 0L
            ) {
              ..error(paste(
                "Sample identifiers were not uniquely specified within each batch.",
                "In case this is intentional, i.e. for repeated measurements, please merge the",
                "datasets yourself prior to input."
              ))
            }

            # Check for column overlap
            overlap_cols <- setdiff(
              intersect(colnames(data), colnames(data_list[[ii]])),
              c(sample_id_column, batch_id_column)
            )

            if (length(overlap_cols) > 0L) {
              ..error(paste(
                "datasets could not be merged by sample and batch identifiers",
                "as one or more columns with the same name appear in both the left and",
                "right-hand datasets:", paste_s(overlap_cols)
              ))
            }

            # Perform full join
            data <- merge(
              x = data,
              y = data_list[[ii]],
              on = c(sample_id_column, batch_id_column),
              all = TRUE
            )
            
          } else if (
            !is.null(sample_id_column) &&
            !is.null(batch_id_column) &&
            !is.null(series_id_column)
          ) {
            # Attempt full join on sample_id_column, batch_id_column, and
            # series_id_column provided that other column names are not
            # overlapping, and all sample_ids and series_ids for a batch are
            # unique.

            # Check if the sample identifier columns is presents in the data.
            if (
              !sample_id_column %in% colnames(data) ||
              !sample_id_column %in% colnames(data_list[[ii]])
            ) {
              ..error(paste(
                "The specified column of sample identifiers", sample_id_column,
                "was not found in one or more of the datasets."
              ))
            }

            # Check if the batch identifier column is present in the data.
            if (
              !batch_id_column %in% colnames(data) ||
              !batch_id_column %in% colnames(data_list[[ii]])
            ) {
              ..error(paste(
                "The specified column of batch identifiers", batch_id_column,
                "was not found in one or more of the datasets."
              ))
            }

            # Check if the series identifier column is present in the data.
            if (
              !series_id_column %in% colnames(data) ||
              !series_id_column %in% colnames(data_list[[ii]])
            ) {
              ..error(paste(
                "The specified column of series identifiers", series_id_column,
                "was not found in one or more of the datasets."
              ))
            }

            # Check uniqueness of identifiers
            if (
              anyDuplicated(data[, mget(sample_id_column, batch_id_column, series_id_column)]) > 0L ||
              anyDuplicated(data_list[[ii]][, mget(sample_id_column, batch_id_column, series_id_column)]) > 0L
            ) {
              ..error(paste(
                "Sample identifiers were not uniquely specified within each batch.",
                "In case this is intentional, i.e. for repeated measurements, please merge the",
                "datasets yourself prior to input."
              ))
            }

            # Check for column overlap
            overlap_cols <- setdiff(
              intersect(colnames(data), colnames(data_list[[ii]])),
              c(sample_id_column, batch_id_column, series_id_column)
            )

            if (length(overlap_cols) > 0L) {
              ..error(paste(
                "datasets could not be merged by sample, batch and series identifiers",
                "as one or more columns with the same name appear in both the left and",
                "right-hand datasets:", paste_s(overlap_cols)
              ))
            }

            # Perform full join
            data <- merge(
              x = data,
              y = data_list[[ii]],
              on = c(sample_id_column, batch_id_column, series_id_column),
              all = TRUE
            )
            
          } else {
            # Attempt full join on sample_id_column, provided that other column
            # names do not overlap, and all sample ids are unique.

            # Check if the sample identifier columns is presents in the data
            if (
              !sample_id_column %in% colnames(data) ||
              !sample_id_column %in% colnames(data_list[[ii]])
            ) {
              ..error(paste(
                "The specified column of sample identifiers", sample_id_column,
                "was not found in one or more of the datasets."
              ))
            }

            # Determine if all sample identifiers are unique
            if (
              anyDuplicated(data[, mget(sample_id_column)]) > 0L ||
              anyDuplicated(data_list[[ii]][, mget(sample_id_column)]) > 0L
            ) {
              ..error(paste0(
                "Sample identifiers were not uniquely specified. In case this is intentional, ",
                "i.e. for repeated measurements, please merge the datasets yourself prior to input."
              ))
            }

            # Check for column overlap.
            overlap_cols <- setdiff(
              intersect(colnames(data), colnames(data_list[[ii]])),
              sample_id_column
            )

            if (length(overlap_cols) > 0L) {
              ..error(paste0(
                "Datasets could not be merged by sample identifiers as one or more ",
                "columns with the same name appear in both the left and ",
                "right-hand datasets: ", paste_s(overlap_cols)
              ))
            }

            # Perform full join.
            data <- merge(
              x = data,
              y = data_list[[ii]],
              on = sample_id_column,
              all = TRUE
            )
          }
        }
        
      } else {
        data <- data_list[[1L]]
      }
      
    } else {
      data <- data_list[[1L]]
    }
  } else {
    ..error(paste0(
      "Data is expected to be a data.table, data.frame, ",
      "a path towards a file or a vector or list of the above."
    ))
  }

  if (is_empty(data)) ..error_data_set_is_empty()

  return(data)
}



.load_rdata <- function(file_path) {
  # Loads an RData file, and returns its contents
  load(file_path)

  # Find data in the local environment
  data <- get(ls()[ls() != "file_path"])

  # Cast to data.table
  return(data.table::as.data.table(data))
}



.load_csv <- function(file_path) {
  return(data.table::fread(file = file_path))
}



.load_rds <- function(file_path) {
  return(data.table::as.data.table(readRDS(file_path)))
}



#' Internal function for finalising generic data processing
#'
#' @param data data.table with feature data
#' @inheritParams .parse_experiment_settings
#' @inheritParams as_data_object
#'
#' @details This function is used to update data.table provided by loading the
#'   data. When part of the main familiar workflow, this function is used after
#'   .parse_initial_settings --> .load_data --> .update_initial_settings.
#'
#'   When used to parse external data (e.g. in conjunction with familiarModel)
#'   it follows after .load_data. Hence the function contains several checks
#'   which are otherwise part of .update_initial_settings.
#'
#' @return data.table with expected column names.
#'
#' @md
#' @keywords internal
.finish_data_preparation <- function(
    data,
    sample_id_column,
    batch_id_column,
    series_id_column,
    outcome_column,
    outcome_type,
    include_features,
    class_levels,
    censoring_indicator,
    event_indicator,
    competing_risk_indicator,
    check_stringency = "strict",
    reference_method = "auto",
    .no_features_required = FALSE
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  sample_id <- batch_id <- n <- NULL
  
  # Check if the input data has any samples
  if (is_empty(data)) ..error_data_set_is_empty()

  # Set sample identifier column
  if (!is.null(sample_id_column)) {
    # Check input -- note this may be double, but this function may be called
    # when parsing external data as well, e.g. as argument to a predict method
    # that is called using external data.
    .check_input_identifier_column(
      id_column = sample_id_column,
      data = data,
      include_features = include_features,
      col_type = "sample"
    )

    # Rename column
    data.table::setnames(
      x = data,
      old = sample_id_column,
      new = "sample_id"
    )
    
  } else {
    # Create new column with sample ids.
    data[, "sample_id" := .I]
  }

  # Set batch identifier column
  if (!is.null(batch_id_column)) {
    # Check input
    .check_input_identifier_column(
      id_column = batch_id_column,
      data = data,
      include_features = include_features,
      col_type = "batch"
    )

    # Rename column
    data.table::setnames(
      x = data,
      old = batch_id_column,
      new = "batch_id"
    )

    # Check data type of the batch_id column and change to character. Cohort
    # names are parsed as characters, not integers.
    if (!is.character(data$batch_id)) data$batch_id <- as.character(data$batch_id)
    
  } else {
    # Create a cohort id column with placeholder.
    data[, "batch_id" := "placeholder"]
  }

  # Set outcome column
  if (!is.null(outcome_column)) {
    # Check plausibility of outcome type given the data
    .check_outcome_type_plausibility(
      data = data,
      outcome_type = outcome_type,
      outcome_column = outcome_column,
      censoring_indicator = censoring_indicator,
      event_indicator = event_indicator,
      competing_risk_indicator = competing_risk_indicator,
      check_stringency = check_stringency
    )

    if (outcome_type %in% c("survival")) {
      # Add survival columns

      if (check_stringency == "strict") {
        # Identify survival status columns
        event_cols <- sapply(
          outcome_column, 
          .is_survival_status_col,
          data = data,
          censoring_indicator = censoring_indicator,
          event_indicator = event_indicator,
          competing_risk_indicator = competing_risk_indicator
        )

        # The plausibility already took care of consistency checking. This means
        # that there is one and only one event status column and the other
        # column contains survival time.
        time_column <- outcome_column[!event_cols]
        event_column <- outcome_column[event_cols]
        
      } else {
        # For other stringency levels, outcome columns are assumed to be
        # extracted from familiarModel or familiarEnsemble objects, which
        # already contain ordered values.
        time_column <- outcome_column[1L]
        event_column <- outcome_column[2L]
      }

      # Rename columns
      data.table::setnames(
        x = data,
        old = c(time_column, event_column),
        new = get_outcome_columns(x = outcome_type)
      )
      
    } else {
      # Rename outcome column
      data.table::setnames(
        x = data,
        old = outcome_column,
        new = get_outcome_columns(x = outcome_type)
      )
    }
    
  } else if (outcome_type %in% c("survival", "competing_risk")) {
    # Generate outcome columns with NA

    # Find outcome column names
    outcome_column <- get_outcome_columns(x = outcome_type)

    # Create new columns and set to NA
    for (current_outcome_col in outcome_column) {
      data[, (current_outcome_col) := NA_real_]
    }
    
  } else if (outcome_type %in% c("binomial", "multinomial")) {
    # Find outcome column names
    outcome_column <- get_outcome_columns(x = outcome_type)

    # Generate outcome column with NA values
    data[, (outcome_column) := NA_character_]
    
  } else if (outcome_type %in% c("continuous")) {
    # Find outcome column names
    outcome_column <- get_outcome_columns(x = outcome_type)

    # Generate outcome column with NA values
    data[, (outcome_column) := NA_real_]
    
  } else if (outcome_type %in% c("unsupervised")) {
    # Outcome column is NULL, as unsupervised data do not have outcome.
    outcome_column <- NULL
    
  } else {
    ..error_no_known_outcome_type(outcome_type)
  }

  # Set class levels
  if (outcome_type %in% c("binomial", "multinomial")) {
    if (!is.null(class_levels)) {
      # Perform checks on class levels
      .check_class_level_plausibility(
        data = data,
        outcome_type = outcome_type,
        outcome_column = "outcome",
        class_levels = class_levels,
        check_stringency = check_stringency
      )

      # Set class levels. This may involve reordering class levels in case the
      # outcome already was a factor.
      data$outcome <- factor(
        x = data$outcome,
        levels = class_levels,
        exclude = NA
      )
      
    } else if (!is.factor(data$outcome)) {
      # Convert to factors
      data$outcome <- factor(
        x = data$outcome,
        levels = unique(data$outcome),
        exclude = c(NA, "NA", "NAN", "na", "nan", "NaN")
      )
    }
  }

  # Check survival time for positivity.
  if (outcome_type %in% c("survival", "competing_risk")) {
    time_column <- get_outcome_columns(x = outcome_type)[1L]

    .check_survival_time_plausibility(
      data = data,
      outcome_column = time_column,
      outcome_type = outcome_type,
      check_stringency = check_stringency
    )

    # Convert outcome_event to 0s and 1s.
    replacement_outcome_event <- numeric(length(data$outcome_event))
    replacement_outcome_event[data$outcome_event %in% censoring_indicator] <- 0L
    replacement_outcome_event[data$outcome_event %in% event_indicator] <- 1L

    # Update competing risk indicators.
    if (length(competing_risk_indicator) > 0L) {
      for (ii in seq_along(competing_risk_indicator)) {
        replacement_outcome_event[data$outcome_event %in% competing_risk_indicator[ii]] <- ii + 1L
      }
    }

    # Update outcome event column.
    data$outcome_event <- replacement_outcome_event
  }

  # Find outcome columns.
  outcome_cols <- get_outcome_columns(x = outcome_type)

  # Set series identifier.
  if (!is.null(series_id_column)) {
    # Check input
    .check_input_identifier_column(
      id_column = series_id_column,
      data = data,
      include_features = include_features,
      col_type = "series"
    )

    # Rename column
    data.table::setnames(
      x = data,
      old = series_id_column,
      new = "series_id"
    )

    # Check data type of the series_id column and change to character. Series
    # identifiers are parsed as characters, not integers.
    if (!is.character(data$series_id)) data$series_id <- as.character(data$series_id)
    
  } else {
    # Assign series ID per unique outcome for each sample.
    temp_data <- unique(data[, mget(c("sample_id", "batch_id", outcome_cols))])
    temp_data[, "series_id" := seq_len(.N), by = c("sample_id", "batch_id")]

    # Merge the series_id column into the main dataset.
    data <- merge(
      x = data,
      y = temp_data,
      by = c("sample_id", "batch_id", outcome_cols)
    )
  }

  # Add repetition identifiers
  data[
    ,
    "repetition_id" := seq_len(.N),
    by = c("sample_id", "batch_id", "series_id", outcome_cols)
  ]

  # Check that there are all combinations of sample_id, batch_id and series_id
  # have the same outcome.
  single_outcome_samples <- unique(data, by = c("sample_id", "batch_id", "series_id", outcome_cols))
  single_outcome_samples <- single_outcome_samples[, list("n" = .N), by = c("sample_id", "batch_id", "series_id")]
  if (any(single_outcome_samples$n > 1L)) {
    single_outcome_samples <- single_outcome_samples[n > 1L]
    single_outcome_samples[, "descriptor" := paste0(sample_id, " (", batch_id, ")")]
    
    ..error(paste0(
      "One or more samples with the same identifier do not have the same outcome value: ",
      paste_s(single_outcome_samples$descriptor)
    ))
  }

  # Determine which columns to maintain
  if (!is.null(include_features)) {
    # Check presence of features in include_features in the data
    .check_feature_availability(
      data = data, 
      feature = include_features
    )

    # Select only features marked for inclusion, as well as identifier and outcome columns
    data <- data[, mget(c(get_non_feature_columns(x = outcome_type), include_features))]
  }

  # Perform feature processing and checks. Skip if features are not required
  # to be present.
  if (!.no_features_required) {
    # Check if the data actually contains any features at this point
    if (!has_feature_data(x = data, outcome_type = outcome_type)) {
      ..error_data_set_has_no_features()
    }
    
    # Convert data to categorical features
    data <- .parse_categorical_features(
      data = data,
      outcome_type = outcome_type,
      reference_method = reference_method
    )
    
    # Convert integer data to double. This prevents rare errors later e.g., 
    # when aggregating data by computing a median value (that is not guaranteed to
    # be an integer).
    data <- .parse_integer_features(
      data = data,
      outcome_type = outcome_type
    )
  }
  
  return(data)
}



.check_data_plausibility <- function(
    data,
    settings,
    message_indent = 0L,
    verbose = TRUE,
    cl = NULL
) {
  # Avoid CRAN NOTE due to non-standard evaluation in data.table.
  score <- NULL
  
  logger_message(
    "Setup report: Preliminary data-checks.",
    indent = message_indent,
    verbose = verbose
  )
  
  feature_cols <- get_feature_columns(data, outcome_type = settings$data$outcome_type)
  outcome_cols <- get_outcome_columns(settings$data$outcome_type)
  
  # Plausibility checks: duplicate rows.
  if (anyDuplicated(data[, mget(c(feature_cols, outcome_cols))]) > 0L) {
    logger_warning(
      paste0(
        "Ignoring identifiers, one or more rows in the dataset may contain duplicated data. ",
        "This means the same combination of feature values and outcome appears multiple times."
      ),
      warn_class = "familiar_data_check"
    )
  }
  
  
  # Plausibility checks: invariant features.
  invariant_features <- fam_sapply(
    cl = cl,
    assign = NULL,
    X = data[, mget(feature_cols)],
    FUN = is_singular_data,
    progress_bar = FALSE,
    chopchop = TRUE
  )
  
  invariant_features <- feature_cols[invariant_features]
  if (length(invariant_features) > 0L) {
    logger_warning(
      paste0(
        "The following features are invariant: ",
        paste_s(invariant_features),
        "."
      ),
      warn_class = "familiar_data_check"
    )
  }
  
  
  # Plausibility checks: one-to-one predictors
  
  # Convert to dataObject. We pretend that the data are preprocessed up to
  # clustering.
  data <- methods::new(
    "dataObject",
    data = data,
    preprocessing_level = "clustering",
    outcome_type = settings$data$outcome_type,
    outcome_info = create_outcome_info(settings = settings),
    data_column_info = create_data_column_info(settings = settings)
  )
  
  # Add generic feature information.
  generic_feature_info_task <- new("familiarTaskGenericFeatureInfo")
  generic_feature_info <- .perform_task(
    object = generic_feature_info_task,
    data = data
  )
  
  # Set up vimp object and promote to concordance.
  vimp_object <- promote_vimp_method(methods::new(
    "familiarVimpMethod",
    outcome_type = data@outcome_type,
    outcome_info = data@outcome_info,
    feature_info = generic_feature_info,
    vimp_method = "concordance"
  ))
  
  # Compute concordance.
  vimp_table <- suppressWarnings(get_vimp_table(.vimp(
    object = vimp_object,
    data = data,
    cl = cl
  )))
  
  # Identify features with perfect concordance (1.0) or discordance (also 1.0),
  # and warn.
  if (any(vimp_table$score == 1.0)) {
    logger_warning(
      paste0(
        "The following features are perfect predictors for the outcome: ",
        paste_s(vimp_table[score == 1.0]$name),
        ". Please ensure that there are no outcome data included as features."
      ),
      warn_class = "familiar_data_check"
    )
  }
  
  return(invisible(TRUE))
}



#' Internal function for converting integer features
#'
#' @param data data.table with feature data
#' @param outcome_type character, indicating the type of outcome
#'
#' @details This function parses columns containing integer feature data to
#'   features to double. This prevents, e.g., errors when the result of an
#'   operation on the feature data yields a non-integer (i.e. floating point)
#'   result.
#'
#' @return data.table with integer features converted to double.
#'
#' @md
#' @keywords internal
.parse_integer_features <- function(data, outcome_type) {
  # Replace columns types so that only numeric and categorical features remain
  
  # Check presence of feature columns
  if (!has_feature_data(x = data, outcome_type = outcome_type)) ..error_data_set_has_no_features()
  
  # Get feature columns
  feature_columns <- get_feature_columns(x = data, outcome_type = outcome_type)
  
  # Identify features that consist of integer values.
  integer_features <- sapply(
    feature_columns,
    function(feature, data) (is.integer(data[[feature]])),
    data = data
  )
  integer_features <- feature_columns[integer_features]
  
  # Do not update data if there are no columns with integer features.
  if (length(integer_features) == 0L) return(data)
  
  # Update to integer features to double.
  for (feature in integer_features) {
    data.table::set(data, j = feature, value = as.double(data[[feature]]))
  }
  
  return(data)
}



#' Internal function for setting categorical features
#'
#' @param data data.table with feature data
#' @param outcome_type character, indicating the type of outcome
#' @param reference_method character, indicating the type of method used to set
#'   the reference level.
#'
#' @details This function parses columns containing feature data to factors if
#'   the data contained therein have logical (TRUE, FALSE), character, or factor
#'   classes.  Unless passed as feature names with `reference`, numerical data,
#'   including integers, are not converted to factors.
#'
#' @return data.table with several features converted to factor.
#'
#' @md
#' @keywords internal
.parse_categorical_features <- function(
    data, 
    outcome_type,
    reference_method = "auto"
) {
  # Replace columns types so that only numeric and categorical features remain

  # Check presence of feature columns
  if (!has_feature_data(x = data, outcome_type = outcome_type)) ..error_data_set_has_no_features()

  # Get feature columns
  feature_columns <- get_feature_columns(x = data, outcome_type = outcome_type)

  # Find column classes
  column_class <- lapply(
    feature_columns,
    function(ii, data) (class(data[[ii]])),
    data = data
  )

  # Identify categorical columns
  categorical_columns <- sapply(
    column_class,
    function(selected_column_class) {
      any(selected_column_class %in% c("logical", "character", "factor"))
    }
  )
  categorical_columns <- feature_columns[categorical_columns]

  # Do not update data if there are no columns for categorical data
  if (length(categorical_columns) == 0L) return(data)

  # Generate a list of warnings.
  warning_list <- list()

  # Iterate over categorical columns
  for (ii in categorical_columns) {
    # The default option is to parse categorical features by setting the most
    # frequent level as reference.
    factor_fun <- ..parse_categorical_most_frequent

    # Check if the feature was externally set to be a factor.
    is_external <- is.factor(data[[ii]])

    if (reference_method == "auto") {
      # Under "auto" mode, the most frequent level is used for categorical
      # features that are not externally set, whereas features that are
      # externally set are not re-ordered.
      if (is_external) factor_fun <- ..parse_categorical_as_is
      
    } else if (reference_method == "never") {
      # Under "never" mode, the  categorical features that are not externally
      # set are simply ordered, whereas features that are externally set are not
      # re-ordered.
      factor_fun <- if (is_external) ..parse_categorical_as_is else ..parse_categorical_sorted
    }

    # Exclude any potential re-ordering of known ordinal features under all
    # conditions.
    if (is_external) {
      if (is.ordered(data[[ii]])) factor_fun <- ..parse_categorical_as_is
    }

    # Create a categorical variable for the current column.
    data.table::set(data, j = ii, value = factor_fun(data[[ii]]))
  }

  # Raise an error in case there are missing class levels for any feature.
  if (length(warning_list) > 0L) {
    ..error(paste(warning_list, collapse = "\n"))
  }

  return(data)
}



..parse_categorical_as_is <- function(x) {
  # Return as is.
  return(x)
}



..parse_categorical_sorted <- function(x) {
  # Find class levels.
  class_levels <- if (is.factor(x)) levels(x) else unique_na(x)

  # Sort class levels.
  class_levels <- sort(class_levels)

  return(factor(x = x, levels = class_levels))
}



..parse_categorical_most_frequent <- function(x) {
  # Suppress NOTES due to non-standard evaluation in data.table
  n <- NULL

  # Created sorted table
  class_level_data <- data.table::data.table("x" = x)
  class_level_data <- class_level_data[, list("n" = .N), by = "x"][order(-n, x)]

  # Keep only non-NA values.
  class_level_data <- class_level_data[!is.na(x)]

  # Convert to character.
  class_levels <- as.character(class_level_data$x)

  # Get the reference level.
  reference_level <- head(class_levels, n = 1L)

  # Order class levels.
  class_levels <- sort(class_levels)

  # Insert the reference level in the first position.
  class_levels <- setdiff(class_levels, reference_level)
  class_levels <- c(reference_level, class_levels)

  return(factor(x = x, levels = class_levels))
}



update_data_set <- function(data, object) {
  # Check if the classes of the input is correct.
  if (!data.table::is.data.table(data)) {
    ..error("update_data_set: data is not a data.table")
  }
  
  if (!(
    is(object, "familiarModel") ||
    is(object, "familiarEnsemble") ||
    is(object, "familiarNoveltyDetector")
  )) {
    ..error(paste0(
      "update_data_set: object is not a familiarModel, familiarNoveltyDetector ",
      "or a familiarEnsemble."
    ))
  }

  # Find the outcome type.
  if (is(object, "familiarNoveltyDetector")) {
    outcome_type <- "unsupervised"
  } else {
    outcome_type <- object@outcome_type
  }

  # Find the outcome column
  outcome_column <- get_outcome_columns(outcome_type)

  # Start warning list.
  warning_list <- NULL

  # Check outcome --------------------------------------------------------------

  # Checks for categorical / ordinal outcomes.
  if (outcome_type %in% c("binomial", "multinomial")) {
    if (is(object@outcome_info, "outcomeInfo")) {
      # Update the outcome column if the outcome data is ordinal.
      if (object@outcome_info@ordered && !is.ordered(data[[outcome_column]])) {
        data[[outcome_column]] <- ordered(
          x = data[[outcome_column]],
          levels = object@outcome_info@levels
        )
        
      } else if (!object@outcome_info@ordered && !is.factor(data[[outcome_column]])) {
        data[[outcome_column]] <- factor(
          x = data[[outcome_column]],
          levels = object@outcome_info@levels
        )
      }

      # Check that the data does not have extra levels.
      extra_levels <- setdiff(
        levels(data[[outcome_column]]),
        object@outcome_info@levels
      )

      if (length(extra_levels) > 0L) {
        warning_list <- c(
          warning_list,
          paste0(
            "The outcome column contains the following ",
            ifelse(length(extra_levels) > 1L, "levels", "level"),
            " that were not found in the original dataset: ",
            paste_s(extra_levels), "; original: ", paste_s(object@outcome_info@levels)
          )
        )
        
      } else {
        # Ensure that order is correct.
        data[[outcome_column]] <- factor(
          x = data[[outcome_column]],
          levels = object@outcome_info@levels,
          ordered = object@outcome_info@ordered
        )
      }
    }
  }

  # TODO: When we start supporting transformation and normalisation
  # parameters for outcome, process the data here.
  if (outcome_type %in% c("continuous")) {
    if (is(object@outcome_info, "outcomeInfo")) {
      if (
        !is.null(object@outcome_info@transformation_parameters) ||
        !is.null(object@outcome_info@normalisation_parameters)
      ) {
        browser()
      }
    }
  }

  # Check columns --------------------------------------------------------------

  # Get all column names.
  all_columns <- colnames(data)

  # Check that the non-feature columns are present.
  non_feature_columns <- get_non_feature_columns(outcome_type)
  missing_non_feature_columns <- setdiff(non_feature_columns, all_columns)

  if (length(missing_non_feature_columns) > 0L) {
    warning_list <- c(
      warning_list,
      paste0(
        "The following non-feature ",
        ifelse(length(missing_non_feature_columns) > 1L, "columns are", "column is"),
        " missing in the dataset: ",
        paste_s(missing_non_feature_columns)
      )
    )
  }

  # Remove non-feature columns from the check.
  all_columns <- setdiff(all_columns, non_feature_columns)

  # Check that the feature columns for required_features, and if not, for the
  # union of model_features and novelty_features are present.
  required_features <- object@required_features

  # Novelty detectors do not have separate novelty feature attributes.
  if (is(object, "familiarNoveltyDetector")) {
    model_and_novelty_features <- object@model_features
  } else {
    model_and_novelty_features <- union(object@model_features, object@novelty_features)
  }

  # Check presence of features.
  if (length(required_features) > 0L && length(model_and_novelty_features) > 0L) {
    if (all(required_features %in% all_columns)) {
      available_features <- required_features
      
    } else if (all(model_and_novelty_features %in% all_columns)) {
      available_features <- model_and_novelty_features
      
    } else {
      # At least one model / novelty feature is missing.
      missing_feature_columns <- setdiff(model_and_novelty_features, all_columns)
      
      warning_list <- c(
        warning_list,
        paste0(
          "The following feature ",
          ifelse(length(missing_feature_columns) > 1L, "columns are", "column is"),
          " missing in the dataset: ",
          paste_s(missing_feature_columns)
        )
      )

      # Select features that are available.
      available_features <- intersect(model_and_novelty_features, all_columns)

      if (length(available_features) == 0L) {
        warning_list <- c(
          warning_list,
          paste0(
            "No additional feature-specific details could be assessed because ",
            "none of the features appear in the dataset."
          )
        )
      }
    }
  } else {
    available_features <- NULL
  }

  # Check features -------------------------------------------------------------
  feature_info_list <- object@feature_info[available_features]

  # Iterate over features.
  for (feature in available_features) {
    # Select the feature info object for the current feature.
    feature_info <- feature_info_list[[feature]]

    if (feature_info@feature_type == "numeric") {
      # For numeric features determine whether the feature in the data is numeric.
      if (!is.numeric(data[[feature]])) {
        warning_list <- c(
          warning_list,
          paste0(
            "The ", feature, " column contain a numeric feature. Found: ",
            typeof(data[[feature]])
          )
        )
      }
      
    } else if (feature_info@feature_type == "factor") {
      # For categorical and ordinal features determine whether there are any
      # unknown levels in the data.
      if (is.factor(data[[feature]])) {
        levels_present <- levels(data[[feature]])
      } else {
        levels_present <- unique(data[[feature]])
      }

      # Check for extra levels. Note that fewer levels is fine.
      extra_levels <- setdiff(levels_present, feature_info@levels)

      if (length(extra_levels) > 0L) {
        warning_list <- c(
          warning_list,
          paste0(
            "The ", feature, " column contains the following ",
            ifelse(length(extra_levels) > 1L, "levels", "level"),
            " that were not found in the original dataset: ",
            paste_s(extra_levels), "; original: ", paste_s(feature_info@levels)
          )
        )
        
      } else {
        # Ensure that order of levels in the data is correct.
        data[[feature]] <- factor(data[[feature]],
          levels = feature_info@levels,
          ordered = feature_info@ordered
        )
      }
    } else {
      ..error_reached_unreachable_code(paste0(
        "update_data_set: unknown feature type encountered: ",
        feature_info@feature_type
      ))
    }
  }

  # Raise an error in case there was any error for any feature.
  if (length(warning_list) > 0L) ..error(paste(warning_list, collapse = "\n\n"))

  return(data)
}



.add_data_dummy_columns <- function(
    data,
    sample_id_column,
    batch_id_column,
    series_id_column,
    outcome_column) {
  # Add dummy sample identifier column if absent.
  if (!sample_id_column %in% colnames(data)) {
    data[, (sample_id_column) := .I]
  }

  # Add dummy batch identifier column if absent.
  if (!batch_id_column %in% colnames(data)) {
    data[, (batch_id_column) := "placeholder"]
  }

  # Add dummy series identifier column if absent.
  if (!series_id_column %in% colnames(data)) {
    data[, (series_id_column) := seq_len(.N), by = c(sample_id_column, batch_id_column)]
  }

  # Add dummy outcome columns, if absent.
  for (current_outcome_column in outcome_column) {
    if (!current_outcome_column %in% colnames(data)) {
      data[, (current_outcome_column) := NA]
    }
  }

  return(data)
}

# Don't perform any further tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Default
familiar:::test_export(
  export_function = familiar:::export_risk_stratification_info,
  not_available_some_predictions_fail = FALSE,
  not_available_no_samples = FALSE,
  data_element = "risk_stratification_info",
  outcome_type_available = c("survival"),
  debug = debug_flag
)

# With ensemble detail level.
familiar:::test_export(
  export_function = familiar:::export_risk_stratification_info,
  not_available_some_predictions_fail = FALSE,
  not_available_no_samples = FALSE,
  data_element = "risk_stratification_info",
  detail_level = "ensemble",
  outcome_type_available = c("survival"),
  debug = debug_flag
)

# Without aggregation
familiar:::test_export(
  export_function = familiar:::export_risk_stratification_info,
  not_available_some_predictions_fail = FALSE,
  not_available_no_samples = FALSE,
  data_element = "risk_stratification_info",
  outcome_type_available = c("survival"),
  export_args = list("aggregate_results" = FALSE),
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# Cluster parameter object -----------------------------------------------------
setClass(
  "featureInfoParametersCluster",
  contains = "featureInfoParameters",
  slots = list(
    "type" = "character",
    "method" = "ANY",
    "weight" = "numeric",
    "invert" = "logical",
    "cluster_name" = "character",
    "cluster_size" = "integer",
    "cluster_features" = "character",
    "required_features" = "ANY"
  ),
  prototype = list(
    "type" = NA_character_,
    "method" = NULL,
    "weight" = 1.0,
    "invert" = FALSE,
    "cluster_name" = NA_character_,
    "cluster_size" = NA_integer_,
    "cluster_features" = NA_character_,
    "required_features" = NULL
  )
)


# Representation object --------------------------------------------------------
setClass(
  "clusterRepresentationObject",
  slots = list(
    "name" = "character",
    "weight" = "numeric",
    "invert" = "logical",
    "cluster_name" = "character",
    "cluster_size" = "integer",
    "cluster_features" = "character",
    "required_features" = "ANY"
  ),
  prototype = list(
    "weight" = NA_real_,
    "invert" = NA,
    "cluster_name" = NA_character_,
    "cluster_size" = NA_integer_,
    "cluster_features" = NA_character_,
    "required_features" = NULL
  )
)



create_cluster_parameter_skeleton <- function(
    feature_info_list,
    feature_names = NULL,
    cluster_method,
    cluster_linkage = NULL,
    cluster_cut_method = NULL,
    cluster_similarity_threshold = NULL,
    cluster_similarity_metric = NULL,
    cluster_representation_method = NULL,
    .override_existing = FALSE
) {
  # Creates a skeleton for the provided cluster method.
  
  # Determine feature names from the feature info list, if provided.
  if (is.null(feature_names)) feature_names <- names(feature_info_list)
  
  # Select only features that appear in the feature info list.
  feature_names <- intersect(names(feature_info_list), feature_names)
  
  # Skip step if no feature info objects are updated.
  if (is_empty(feature_names)) return(feature_info_list)
  
  # Check that method is applicable.
  .check_parameter_value_is_valid(
    x = cluster_method,
    var_name = "cluster_method",
    values = .get_available_cluster_methods()
  )
  
  # Update familiar info objects with a feature clustering skeleton.
  updated_feature_info <- fam_lapply(
    X = feature_info_list[feature_names],
    FUN = .create_cluster_parameter_skeleton,
    method = cluster_method,
    cluster_linkage = cluster_linkage,
    cluster_cut_method = cluster_cut_method,
    cluster_similarity_threshold = cluster_similarity_threshold,
    cluster_similarity_metric = cluster_similarity_metric,
    cluster_representation_method = cluster_representation_method,
    .override_existing = .override_existing
  )
  
  # Provide names for the updated feature info objects.
  names(updated_feature_info) <- feature_names
  
  # Replace list elements.
  feature_info_list[feature_names] <- updated_feature_info
  
  return(feature_info_list)
}



.create_cluster_parameter_skeleton <- function(
    feature_info,
    method,
    ...,
    .override_existing = FALSE
) {
  
  # Check if clustering data was already completed, and does not require being
  # determined anew.
  if (feature_info_complete(feature_info@cluster_parameters) && !.override_existing) {
    return(feature_info)
  }
  
  # Create cluster parameter object.
  object <- methods::new("featureInfoParametersCluster", method = method)
  
  # Set feature name
  object@name <- feature_info@name
  
  # Set feature type
  object@type <- feature_info@feature_type
  
  # Update the familiar version.
  object <- add_package_version(object = object)
  
  # Check that the feature was not removed and is not set as a signature.
  if (!is_available(feature_info) || is_in_signature(feature_info)) method <- "none"
  
  # Create the cluster method object.
  object@method <- create_cluster_method_object(
    cluster_method = method,
    data_type = "cluster",
    ...
  )
  
  # Update imputation_parameters slot.
  feature_info@cluster_parameters <- object
  
  return(feature_info)
}



add_cluster_info <- function(
    cl = NULL,
    feature_info_list,
    data,
    message_indent = 0L,
    verbose = FALSE
) {
  
  # Find feature columns.
  feature_names <- get_feature_columns(x = data)
  
  # Sanity check.
  if (!(setequal(feature_names, get_available_features(feature_info_list = feature_info_list)))) {
    ..error_reached_unreachable_code(paste0(
      "add_cluster_info: features in data and the feature info list are expected ",
      "to be the same, but were not."
    ))
  }
  
  # Skeletons should be present. Now we need to identify which features can be
  # grouped by their method slot. This is done by iterating, and identifying
  # which features form groups. Though this may seem to a bit convoluted, we do
  # this to allow for proper processing of features that may have different
  # parameters assigned on purpose, e.g. signature features or externally
  # provided features.
  
  # Eliminate features that are already complete.
  feature_names <- feature_names[!sapply(
    feature_info_list[feature_names],
    function(x) (feature_info_complete(x@cluster_parameters))
  )]
  
  # Skip any further processing if all parameter information has already been
  # completed.
  if (length(feature_names) == 0L) return(feature_info_list)  
  
  # Set unassigned features.
  unassigned_features <- feature_names
  
  # Iterate to eliminate any groups that would be smaller than 2.
  while (TRUE) {
    # Break once all features have been assigned.
    if (length(unassigned_features) == 0L) break
    
    # Get the cluster method object of the first feature that still needs to be
    # sorted.
    cluster_method_object <- feature_info_list[[unassigned_features[1L]]]@cluster_parameters@method
    
    # Find other unsorted features that have the same method.
    same_method_features <- unassigned_features[sapply(
      feature_info_list[unassigned_features],
      function(x, y) {
        return(identical(x@cluster_parameters@method, y))
      },
      y = cluster_method_object
    )]
    
    # Perform 2 checks:
    #
    # 1, Check that at least 2 features are present. Otherwise no sensible
    # clusters may be formed.
    #
    # 2, Check that there are at least 5 instances present. We cannot form
    # clusters when similarity is difficult to assess.
    if (length(same_method_features) < 2L || get_n_samples(data, id_depth = "repetition") < 5L) {
      feature_info_list[same_method_features] <- create_cluster_parameter_skeleton(
        feature_info_list[same_method_features],
        cluster_method = "none"
      )
    }
    
    # Update the number of features that have not been assigned.
    unassigned_features <- setdiff(unassigned_features, same_method_features)
  }
  
  # Set unassigned features and the initial grouping list
  unassigned_features <- feature_names
  
  # Iterate to create feature groups and add feature information.
  while (TRUE) {
    # Break once all features have been assigned.
    if (length(unassigned_features) == 0L) break()
    
    # Get the cluster method object of the first feature that still needs to be
    # sorted.
    cluster_method_object <- feature_info_list[[unassigned_features[1L]]]@cluster_parameters@method
    
    # Find other unsorted features that have the same method.
    same_method_features <- unassigned_features[sapply(
      feature_info_list[unassigned_features],
      function(x, y) {
        return(identical(x@cluster_parameters@method, y))
      },
      y = cluster_method_object
    )]
    
    # Message that computations are starting.
    if (cluster_method_object@method != "none") {
      logger_message(
        paste0(
          "Computing similarity between ", length(same_method_features), " features ",
          "using the ", cluster_method_object@similarity_metric, " metric ",
          "for clustering using the ", cluster_method_object@method, " method."
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    # Compute similarity.
    cluster_method_object <- set_similarity_table(
      object = cluster_method_object,
      data = filter_features(
        data = data,
        available_features = same_method_features
      ),
      feature_info_list = feature_info_list[same_method_features],
      cl = cl,
      verbose = verbose
    )
    
    # Create clustering objects. These are used to update the
    # feature_info_lists.
    clustering_objects <- unname(create_clusters(object = cluster_method_object))
    
    # Update feature info lists.
    updated_feature_info <- fam_lapply(
      cl = cl,
      X = clustering_objects,
      FUN = .add_cluster_info,
      cluster_method_object = cluster_method_object,
      feature_info_list = feature_info_list,
      data = data,
      chopchop = TRUE,
      progress_bar = FALSE
    )
    
    # Flatten lists with feature info.
    updated_feature_info <- unlist(updated_feature_info, recursive = FALSE)
    
    # Replace list elements without reordering.
    feature_info_list[same_method_features] <- updated_feature_info[same_method_features]
    
    # Remove features from the list of unassigned features.
    unassigned_features <- setdiff(unassigned_features, same_method_features)
  }
  
  return(feature_info_list)
}



.add_cluster_info <- function(
    clustering_object,
    cluster_method_object,
    feature_info_list,
    data
) {
  
  # Limit feature info list and data to the features in the cluster
  data <- filter_features(
    data = data,
    available_features = clustering_object@cluster_features
  )
  
  # Find representation.
  representation_objects <- add_feature_info_parameters(
    object = clustering_object,
    data = data,
    feature_info = feature_info_list[clustering_object@cluster_features],
    cluster_method_object = cluster_method_object
  )
  
  # Update cluster_parameters attribute in the feature info objects.
  updated_feature_info <- fam_mapply(
    FUN = ..add_cluster_info,
    feature_info = feature_info_list[clustering_object@cluster_features],
    representation_object = representation_objects[clustering_object@cluster_features]
  )
  
  # Update names.
  names(updated_feature_info) <- sapply(updated_feature_info, function(x) (x@name))
  
  return(updated_feature_info)
}



..add_cluster_info <- function(
    feature_info,
    representation_object) {
  
  # Use the representation object to update the featureInfoParametersCluster
  # object in cluster_parameters.
  object <- add_feature_info_parameters(
    object = feature_info@cluster_parameters,
    data = representation_object
  )
  
  # Attach updated information object.
  feature_info@cluster_parameters <- object
  
  return(feature_info)
}



# add_feature_info_parameters (cluster info, representation object) ------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersCluster",
    data = "clusterRepresentationObject"
  ),
  function(object, data, ...) {
    
    # Sanity check: check that names are correct.
    if (object@name != data@name) {
      ..error_reached_unreachable_code(paste0(
        "add_feature_info_parameters,featureInfoParametersCluster,clusterRepresentationObject: ",
        "the cluster information object and representation object were not specified for ",
        "the same feature: ", object@name, " (info) and ", data@name, " (representation)"
      ))
    }
    
    # Copy contents
    object@weight <- data@weight
    object@invert <- data@invert
    object@cluster_name <- data@cluster_name
    object@cluster_size <- data@cluster_size
    object@cluster_features <- data@cluster_features
    object@required_features <- data@required_features
    
    # Set complete.
    object@complete <- TRUE
    
    return(object)
  }
)



# apply_feature_info_parameters (cluster info, dataObject) ---------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersCluster",
    data = "dataObject"
  ),
  function(object, data, ...) {
    
    # Determine if the current feature is required, and skip if not.
    if (!object@name %in% object@required_features) return(NULL)
    
    # Determine if weight is not 0.0, and skip if it is.
    weight <- object@weight
    if (weight == 0.0) return(NULL)
    
    # Invert weights
    if (object@invert) weight <- -1.0 * weight
    
    # Only apply weights to numeric features.
    if (object@type == "numeric") {
      return(data@data[[object@name]] * weight)
      
    } else {
      return(data@data[[object@name]])
    }
  }
)



.create_clustering_table <- function(
    feature_info_list,
    selected_features = NULL,
    show_weights = FALSE) {
  
  # Select only requested features.
  if (!is.null(selected_features)) {
    feature_info_list <- feature_info_list[selected_features]
  }
  
  cluster_table <- lapply(
    feature_info_list,
    function(x, show_weights) {
      
      if (is.null(x@cluster_parameters)) {
        # If cluster parameters are not set (e.g. for familiarEnsemble objects),
        # treat features as singular.
        cluster_data <- list(
          "cluster_name" = x@name,
          "feature_name" = x@name,
          "feature_required" = TRUE
        )
        
        if (show_weights) cluster_data <- c(cluster_data, list("weight" = 1.0))
        
      } else {
        cluster_data <- list(
          "cluster_name" = x@cluster_parameters@cluster_name,
          "feature_name" = x@cluster_parameters@cluster_features,
          "feature_required" = x@cluster_parameters@cluster_features %in% x@cluster_parameters@required_features
        )
        
        if (show_weights) cluster_data <- c(cluster_data, list("weight" = x@cluster_parameters@weight))
      }
      
      return(cluster_data)
    },
    show_weights = show_weights
  )
  
  # Remove duplicate entries.
  cluster_table <- unique(data.table::rbindlist(
    cluster_table,
    use.names = TRUE
  ))
  
  return(cluster_table)
}



features_before_clustering <- function(
    features,
    cluster_table = NULL,
    feature_info_list = NULL,
    representative_only = FALSE
) {
  # Convert input features to original features
  
  # Suppress NOTES due to non-standard evaluation in data.table
  cluster_name <- feature_required <- NULL
  
  # Create a cluster table if it is not provided
  if (is.null(cluster_table) && is.null(feature_info_list)) {
    ..error_reached_unreachable_code(paste0(
      "features_before_clustering: if no cluster_table is provided, ",
      "feature_info_list cannot be empty."
    ))
    
  } else if (is.null(cluster_table)) {
    # Expressly set selected_features to NULL: selected_features is used to
    # create a subset of the feature_info_list, which leads to missing data for
    # features that are clustered.
    cluster_table <- .create_clustering_table(
      feature_info_list = feature_info_list,
      selected_features = NULL
    )
  }
  
  # Find original features.
  if (representative_only) {
    original_features <- unique(
      cluster_table[cluster_name %in% features & feature_required == TRUE]$feature_name
    )
    
  } else {
    original_features <- unique(
      cluster_table[cluster_name %in% features]$feature_name
    )
  }
  
  # Return original features.
  return(original_features)
}



features_after_clustering <- function(
    features,
    cluster_table = NULL,
    feature_info_list = NULL
) {
  # Convert input features to features after clustering
  
  # Check if features are set.
  if (is_empty(features)) return(NULL)
  
  # Create a cluster table if it is not provided
  if (is.null(cluster_table) && is.null(feature_info_list)) {
    ..error_reached_unreachable_code(paste0(
      "features_after_clustering: if no cluster_table is provided, ",
      "feature_info_list may be empty."
    ))
    
  } else if (is.null(cluster_table)) {
    cluster_table <- .create_clustering_table(
      feature_info_list = feature_info_list,
      selected_features = features
    )
  }
  
  # Find and return feature names after clustering.
  return(unique(cluster_table$cluster_name[cluster_table$feature_name %in% features]))
}



set_clustered_data <- function(
    cluster_table,
    data,
    feature_info_list
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  feature_required <- NULL
  
  # Only use required data.
  if (!all(cluster_table$feature_required)){
    cluster_table <- cluster_table[feature_required == TRUE]
  }
  
  # For singular clusters or clusters represented by a single feature, simply
  # return the data in the respective column.
  if (nrow(cluster_table) == 1L) return(data@data[[cluster_table$feature_name]])
  
  # Add weighted. Instantiate with 0s.
  feature_values <- numeric(nrow(data@data))
  
  # Iterate over features.
  for (current_feature in cluster_table$feature_name) {
    feature_values <- feature_values + apply_feature_info_parameters(
      object = feature_info_list[[current_feature]]@cluster_parameters,
      data = data
    )
  }
  
  return(feature_values)
}



.get_available_cluster_methods <- function() {
  return(c("none", "pam", "agnes", "diana", "hclust"))
}



.get_available_linkage_methods <- function(cluster_method) {
  if (cluster_method %in% c("agnes", "diana", "hclust")) {
    return(c("average", "single", "complete", "weighted", "ward"))
    
  } else {
    return("none")
  }
}



.get_available_cluster_cut_methods <- function(cluster_method) {
  # Note that "none" is used to prevent clustering. It's not a valid choice per
  # se.
  
  if (cluster_method == "none") {
    cut_methods <- "none"
    
  } else if (cluster_method == "pam") {
    cut_methods <- c("none", "silhouette")
    
  } else if (cluster_method %in% c("agnes", "diana")) {
    cut_methods <- c("none", "fixed_cut", "silhouette")
    
  } else if (cluster_method %in% c("hclust")) {
    cut_methods <- c("none", "fixed_cut", "silhouette", "dynamic_cut")
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".get_available_cluster_cut_methods: encountered unknown cluster method: ",
      cluster_method
    ))
  }
  
  return(cut_methods)
  
}



.get_available_cluster_representation_methods <- function(cluster_method) {
  # Note that "none" is used to prevent forming cluster representations. It's
  # not a valid choice per se.
  
  if (cluster_method == "none") {
    representation_methods <- "none"
    
  } else if (cluster_method == "pam") {
    representation_methods <- c("none", "medioid")
    
  } else if (cluster_method %in% c("agnes", "diana", "hclust")) {
    representation_methods <- c("none", "medioid", "best_predictor", "mean")
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".get_available_cluster_representation_methods: ",
      "encountered unknown cluster representation method: ",
      cluster_method
    ))
  }
  
  return(representation_methods)
}

testthat::skip_on_cran()

# power.transform is also required for transformation_method = "none".
if (!rlang::is_installed("power.transform")) testthat::skip()

# Basic tests ------------------------------------------------------------------

data <- familiar:::test_create_synthetic_correlated_data(
  outcome_type = "continuous",
  n_numeric = 2,
  cluster_size = c(1, 1, 2, 3))

vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "pearson",
  outcome_type = "continuous",
  transformation_method = "none",
  normalisation_method = "none",
  cluster_method = "hclust",
  cluster_cut_method = "fixed_cut",
  cluster_similarity_metric = "mcfadden_r2",
  cluster_similarity_threshold = 0.99,
  imputation_method = "simple")

vimp_table_object <- suppressWarnings(familiar:::.vimp(vimp_object, data))

testthat::test_that("Test that the variable importance table in its initial state is correct.", {
  vimp_table <- familiar:::get_vimp_table(vimp_table_object, "initial")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name"))

  # Assert that there are two features that end with cluster.
  testthat::expect_equal(sum(sapply(vimp_table$name, endsWith, "cluster")), 2L)

  # Assert that there are three entries for feature_1.
  testthat::expect_equal(sum(sapply(vimp_table$name, startsWith, "feature_1")), 3L)

  # Assert that there are four entries for feature_2.
  testthat::expect_equal(sum(sapply(vimp_table$name, startsWith, "feature_2")), 4L)
})


testthat::test_that("Test that the variable importance table in its encoded state is correct.", {
  vimp_table <- familiar:::get_vimp_table(vimp_table_object, "decoded")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name"))

  # Assert that there are two features that end with cluster.
  testthat::expect_equal(sum(sapply(vimp_table$name, endsWith, "cluster")), 2L)

  # Assert that there is one entry for feature_1.
  testthat::expect_equal(sum(vimp_table$name == "feature_1"), 1L)

  # Assert that there is one entry for feature_2.
  testthat::expect_equal(sum(vimp_table$name == "feature_2"), 1L)
})


testthat::test_that("Test that the variable importance table in its declustered state is correct.", {
  vimp_table <- familiar:::get_vimp_table(vimp_table_object, "declustered")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name"))

  # Assert that there is one entry for feature_1.
  testthat::expect_equal(sum(vimp_table$name == "feature_1"), 1L)

  # Assert that there is one entry for feature_2.
  testthat::expect_equal(sum(vimp_table$name == "feature_2"), 1L)

  # Assert that there are two entries for feature_3.
  testthat::expect_equal(sum(sapply(vimp_table$name, startsWith, "feature_3")), 2L)

  # Assert that there are three entries for feature_4.
  testthat::expect_equal(sum(sapply(vimp_table$name, startsWith, "feature_4")), 3L)
})


testthat::test_that("Test that the variable importance table in its re-clustered state is correct.", {
  # Compute reclustered vimp table.
  reclustered_vimp_table_object <- familiar:::recluster_vimp_table(vimp_table_object)
  reclustered_vimp_table <- familiar:::get_vimp_table(reclustered_vimp_table_object, "reclustered")

  # Compute decoded variable importance table. The reclustered table and the
  # decoded table should have the same contents.
  vimp_table <- familiar:::get_vimp_table(vimp_table_object, "decoded")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name"))

  # Assert that there are two features that end with cluster.
  testthat::expect_equal(sum(sapply(vimp_table$name, endsWith, "cluster")), 2L)

  # Assert that there is one entry for feature_1.
  testthat::expect_equal(sum(vimp_table$name == "feature_1"), 1L)

  # Assert that there is one entry for feature_2.
  testthat::expect_equal(sum(vimp_table$name == "feature_2"), 1L)

  # Assert that scores match.
  testthat::expect_equal(
    reclustered_vimp_table[order(name)]$score,
    vimp_table[order(name)]$score)
})


testthat::test_that("Test that the variable importance table in its ranked state is correct.", {
  vimp_table <- familiar:::get_vimp_table(vimp_table_object, "ranked")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name", "rank"))

  # Assert that there is one entry for feature_1.
  testthat::expect_equal(sum(vimp_table$name == "feature_1"), 1L)

  # Assert that there is one entry for feature_2.
  testthat::expect_equal(sum(vimp_table$name == "feature_2"), 1L)

  # Assert that there are two entries for feature_3.
  testthat::expect_equal(sum(sapply(vimp_table$name, startsWith, "feature_3")), 2L)

  # Assert that there are three entries for feature_4.
  testthat::expect_equal(sum(sapply(vimp_table$name, startsWith, "feature_4")), 3L)

  # Assert that there 4 ranks.
  testthat::expect_equal(sort(unique(vimp_table$rank)), seq_len(4L))

  # Assert that feature_3 has the same rank for each entry.
  testthat::expect_equal(data.table::uniqueN(vimp_table[startsWith(name, "feature_3")]$rank), 1L)

  # Assert that feature_4 has the same rank for each entry.
  testthat::expect_equal(data.table::uniqueN(vimp_table[startsWith(name, "feature_4")]$rank), 1L)
})

testthat::test_that("Test that the re-clustered variable importance table in its ranked state is correct.", {
  reclustered_vimp_table_object <- familiar:::recluster_vimp_table(vimp_table_object)
  vimp_table <- familiar:::get_vimp_table(reclustered_vimp_table_object, "ranked")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name", "rank"))

  # Assert that there are two features that end with cluster.
  testthat::expect_equal(sum(sapply(vimp_table$name, endsWith, "cluster")), 2L)

  # Assert that there is one entry for feature_1.
  testthat::expect_equal(sum(vimp_table$name == "feature_1"), 1L)

  # Assert that there is one entry for feature_2.
  testthat::expect_equal(sum(vimp_table$name == "feature_2"), 1L)
})

# Test updating with different cluster table -----------------------------------

data_2 <- familiar:::test_create_synthetic_correlated_data(
  outcome_type = "continuous",
  n_numeric = 2,
  cluster_size = c(1, 1, 4, 2))

vimp_object_2 <- familiar:::test_create_vimp_method(
  data = data_2,
  vimp_method = "pearson",
  outcome_type = "continuous",
  transformation_method = "none",
  normalisation_method = "none",
  cluster_method = "hclust",
  cluster_cut_method = "fixed_cut",
  cluster_similarity_metric = "mcfadden_r2",
  cluster_similarity_threshold = 0.99,
  imputation_method = "simple")

reference_cluster_table <- familiar:::.create_clustering_table(
  vimp_object_2@feature_info)

# Update the variable importance object using the reference cluster table.
updated_vimp_table_object <- familiar:::update_vimp_table_to_reference(
  vimp_table_object,
  reference_cluster_table)

testthat::test_that("Updating variable importance tables functions correctly.", {
  vimp_table <- familiar:::get_vimp_table(updated_vimp_table_object, "ranked")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name", "rank"))

  # Assert that there is one entry for feature_1.
  testthat::expect_equal(sum(vimp_table$name == "feature_1"), 1L)

  # Assert that there is one entry for feature_2.
  testthat::expect_equal(sum(vimp_table$name == "feature_2"), 1L)

  # Assert that there are two entries for feature_3.
  testthat::expect_equal(sum(sapply(vimp_table$name, startsWith, "feature_3")), 4L)

  # Assert that there are three entries for feature_4.
  testthat::expect_equal(sum(sapply(vimp_table$name, startsWith, "feature_4")), 2L)

  # Assert that there 4 ranks.
  testthat::expect_equal(sort(unique(vimp_table$rank)), seq_len(4L))

  # Assert that feature_3 has the same rank for each entry.
  testthat::expect_equal(data.table::uniqueN(vimp_table[startsWith(name, "feature_3")]$rank), 1L)

  # Assert that feature_4 has the same rank for each entry.
  testthat::expect_equal(data.table::uniqueN(vimp_table[startsWith(name, "feature_4")]$rank), 1L)
})

testthat::test_that("Updating variable importance tables with reclustering functions correctly.", {
  reclustered_vimp_table_object <- familiar:::recluster_vimp_table(
    updated_vimp_table_object)
  vimp_table <- familiar:::get_vimp_table(reclustered_vimp_table_object, "ranked")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name", "rank"))

  # Assert that there are two features that end with cluster.
  testthat::expect_equal(sum(sapply(vimp_table$name, endsWith, "cluster")), 2L)

  # Assert that there is one entry for feature_1.
  testthat::expect_equal(sum(vimp_table$name == "feature_1"), 1L)

  # Assert that there is one entry for feature_2.
  testthat::expect_equal(sum(vimp_table$name == "feature_2"), 1L)
})

# Test updating table where all features are different -------------------------

data_3 <- familiar:::test_create_synthetic_correlated_data(
  outcome_type = "continuous",
  n_numeric = 2,
  cluster_size = c(1, 1, 1, 1))

# Rename features.
data.table::setnames(
  x = data_3@data,
  old = c("feature_1", "feature_2", "feature_3", "feature_4"),
  new = c("feature_A", "feature_B", "feature_C", "feature_D"))

vimp_object_3 <- familiar:::test_create_vimp_method(
  data = data_3,
  vimp_method = "pearson",
  outcome_type = "continuous",
  transformation_method = "none",
  normalisation_method = "none",
  cluster_method = "hclust",
  cluster_cut_method = "fixed_cut",
  cluster_similarity_metric = "mcfadden_r2",
  cluster_similarity_threshold = 0.99,
  imputation_method = "simple")

reference_cluster_table <- familiar:::.create_clustering_table(
  vimp_object_3@feature_info)

# Update the variable importance object using the reference cluster table.
updated_vimp_table_object <- familiar:::update_vimp_table_to_reference(
  vimp_table_object,
  reference_cluster_table)

testthat::test_that("Updating variable importance tables with mismatching reference cluster tables.", {
  vimp_table <- familiar:::get_vimp_table(updated_vimp_table_object, "ranked")

  testthat::expect_equal(familiar:::is_empty(updated_vimp_table_object), TRUE)

  testthat::expect_equal(familiar:::is_empty(vimp_table), TRUE)
})


testthat::test_that("Updating variable importance tables with reclustering functions correctly.", {
  reclustered_vimp_table_object <- familiar:::recluster_vimp_table(updated_vimp_table_object)
  vimp_table <- familiar:::get_vimp_table(reclustered_vimp_table_object, "ranked")

  testthat::expect_equal(familiar:::is_empty(updated_vimp_table_object), TRUE)

  testthat::expect_equal(familiar:::is_empty(vimp_table), TRUE)
})

# Test updating table where the variable importance table could not be created. -----

empty_vimp_table_object <- vimp_table_object
empty_vimp_table_object@vimp_table <- NULL

# Update the variable importance object using the reference cluster table.
updated_vimp_table_object <- familiar:::update_vimp_table_to_reference(
  empty_vimp_table_object,
  reference_cluster_table)

testthat::test_that("Updating variable importance tables with mismatching reference cluster tables.", {
  vimp_table <- familiar:::get_vimp_table(updated_vimp_table_object, "ranked")

  testthat::expect_equal(familiar:::is_empty(updated_vimp_table_object), TRUE)

  testthat::expect_equal(familiar:::is_empty(vimp_table), TRUE)
})

testthat::test_that("Updating variable importance tables with reclustering functions correctly.", {
  reclustered_vimp_table_object <- familiar:::recluster_vimp_table(updated_vimp_table_object)
  vimp_table <- familiar:::get_vimp_table(reclustered_vimp_table_object, "ranked")

  testthat::expect_equal(familiar:::is_empty(updated_vimp_table_object), TRUE)

  testthat::expect_equal(familiar:::is_empty(vimp_table), TRUE)
})

# Test with signature features -------------------------------------------------

data <- familiar:::test_create_synthetic_correlated_data(
  outcome_type = "continuous",
  n_numeric = 2,
  cluster_size = c(1, 1, 1, 1))

vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "pearson",
  outcome_type = "continuous",
  transformation_method = "none",
  normalisation_method = "none",
  cluster_method = "hclust",
  cluster_cut_method = "fixed_cut",
  cluster_similarity_metric = "mcfadden_r2",
  cluster_similarity_threshold = 0.99,
  signature = c("feature_1"),
  imputation_method = "simple")

vimp_table_object <- suppressWarnings(familiar:::.vimp(vimp_object, data))

testthat::test_that("A signature feature does not appear in the variable importance table.", {
  vimp_table <- familiar:::get_vimp_table(vimp_table_object, "ranked")

  # Assert that colnames are correct.
  testthat::expect_setequal(colnames(vimp_table), c("score", "name", "rank"))

  # Assert that there is no entry for feature_1.
  testthat::expect_equal(sum(vimp_table$name == "feature_1"), 0L)

  # Assert that there is one entry for feature_2.
  testthat::expect_equal(sum(vimp_table$name == "feature_2"), 1L)

  # Assert that there is one entry for feature_3.
  testthat::expect_equal(sum(vimp_table$name == "feature_3"), 1L)

  # Assert that there is one entry for feature_4.
  testthat::expect_equal(sum(vimp_table$name == "feature_4"), 1L)

  # Assert that there 3 ranks.
  testthat::expect_equal(sort(unique(vimp_table$rank)), seq_len(3L))
})

# Test with all signature features ---------------------------------------------

vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "pearson",
  outcome_type = "continuous",
  transformation_method = "none",
  normalisation_method = "none",
  cluster_method = "hclust",
  cluster_cut_method = "fixed_cut",
  cluster_similarity_metric = "mcfadden_r2",
  cluster_similarity_threshold = 0.99,
  signature = c("feature_1", "feature_2", "feature_3", "feature_4"),
  imputation_method = "simple")

vimp_table_object <- suppressWarnings(familiar:::.vimp(vimp_object, data))

testthat::test_that("A signature feature does not appear in the variable importance table.", {
  vimp_table <- familiar:::get_vimp_table(vimp_table_object, "ranked")

  # Assert that both the variable importance table object and the variable
  # importance table are empty.
  testthat::expect_equal(familiar:::is_empty(vimp_table_object), TRUE)
  testthat::expect_equal(familiar:::is_empty(vimp_table), TRUE)
})

# Skip on CRAN as this test takes quite some time.
testthat::skip_on_cran()
testthat::skip_on_ci()

verbose <- FALSE

# Create test parameters. This is actually a generator, which makes it easier to
# nest all the different combinations that should be checked.
generate_test_parameters <- coro::generator(function(
    similarity_metric,
    similarity_threshold,
    n_numeric_features,
    cluster_method = NULL,
    cluster_linkage = NULL,
    cluster_cut_method = NULL,
    cluster_representation_method = NULL) {
  for (n_numeric_features in n_numeric_features) {
    if (is.null(cluster_method)) {
      local_cluster_method <- familiar:::.get_available_cluster_methods()
    } else {
      local_cluster_method <- intersect(
        cluster_method,
        familiar:::.get_available_cluster_methods())
    }
    
    for (current_cluster_method in local_cluster_method) {
      if (is.null(cluster_linkage)) {
        local_cluster_linkage <- familiar:::.get_available_linkage_methods(
          cluster_method = current_cluster_method)
      } else {
        local_cluster_linkage <- intersect(
          cluster_linkage,
          familiar:::.get_available_linkage_methods(
            cluster_method = current_cluster_method))
      }
      
      for (current_cluster_linkage in local_cluster_linkage) {
        if (is.null(cluster_cut_method)) {
          local_cluster_cut_method <- familiar:::.get_available_cluster_cut_methods(
            cluster_method = current_cluster_method)
        } else {
          local_cluster_cut_method <- intersect(
            cluster_cut_method,
            familiar:::.get_available_cluster_cut_methods(
              cluster_method = current_cluster_method))
        }
        
        for (current_cluster_cut_method in local_cluster_cut_method) {
          if (is.null(cluster_representation_method)) {
            local_cluster_representation_method <- familiar:::.get_available_cluster_representation_methods(
              cluster_method = current_cluster_method)
          } else {
            local_cluster_representation_method <- intersect(
              cluster_representation_method,
              familiar:::.get_available_cluster_representation_methods(
                cluster_method = current_cluster_method))
          }
          
          for (current_cluster_representation_method in local_cluster_representation_method) {
            # Set outcome type to avoid redundant tests where outcome type does
            # not matter.
            if (current_cluster_representation_method %in% c("best_predictor")) {
              available_outcome_type <- c("survival", "binomial", "multinomial", "continuous")
            } else {
              available_outcome_type <- c("binomial")
            }
            
            for (outcome_type in available_outcome_type) {
              for (current_similarity_metric in similarity_metric) {
                for (current_similarity_threshold in similarity_threshold) {
                  coro::yield(list(
                    "outcome_type" = outcome_type,
                    "n_numeric" = n_numeric_features,
                    "cluster_method" = current_cluster_method,
                    "cluster_linkage" = current_cluster_linkage,
                    "cluster_cut_method" = current_cluster_cut_method,
                    "cluster_representation_method" = current_cluster_representation_method,
                    "similarity_metric" = current_similarity_metric,
                    "similarity_threshold" = current_similarity_threshold))
                }
              }
            }
          }
        }
      }
    }
  }
})



# Generic test -----------------------------------------------------------------

# Note: use of spearman or other correlation metrics causes failing tests
# because categorical variables are encoded for such metrics.
#
# Note: only test all-numeric, 1 mixed and all-categorical options to limit the
# number of tests that have to be performed.

generic_test <- generate_test_parameters(
  similarity_metric = "mcfadden_r2",
  similarity_threshold = 0.99,
  n_numeric_features = c(4, 3, 0))

while (TRUE) {
  # Generate parameters.
  parameters <- generic_test()
  if (coro::is_exhausted(parameters)) break

  # Set cluster size
  cluster_size <- c(2, 3, 3, 3)

  # Create data,
  data <- familiar:::test_create_synthetic_correlated_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Create a list of featureInfo objects.
  feature_info_list <- test_create_generic_info(data = data)

  # Create cluster skeletons
  feature_info_list <- familiar:::create_cluster_parameter_skeleton(
    feature_info_list,
    cluster_method = parameters$cluster_method,
    cluster_linkage = parameters$cluster_linkage,
    cluster_cut_method = parameters$cluster_cut_method,
    cluster_similarity_threshold = parameters$similarity_threshold,
    cluster_similarity_metric = parameters$similarity_metric,
    cluster_representation_method = parameters$cluster_representation_method)

  # Update feature info with cluster parameters.
  feature_info_list <- familiar:::add_cluster_info(
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose)

  # Assume that the data are imputed.
  data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_data <- familiar:::cluster_features(
    data = data,
    feature_info_list = feature_info_list)

  # Create clustering table.
  cluster_table <- familiar:::.create_clustering_table(
    feature_info_list = feature_info_list)

  # Create single-sample data.
  one_sample_data <- familiar:::test_create_synthetic_correlated_one_sample_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Assume that data are imputed.
  one_sample_data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_one_sample_data <- familiar:::cluster_features(
    data = one_sample_data,
    feature_info_list = feature_info_list)

  # Perform tests
  testthat::test_that(paste0(
    "Clustering works correctly for the ", parameters$cluster_method, " clustering method, ",
    "with cluster cut method \"", parameters$cluster_cut_method, "\" and ",
    "cluster representation method \"", parameters$cluster_representation_method, "\" for ",
    "normal synthetic data with a ", parameters$outcome_type, " outcome."
  ), {
    # Assert that the number of created clusters is correct.
    if (parameters$cluster_cut_method == "none" ||
        parameters$cluster_representation_method == "none") {
      testthat::expect_equal(
        familiar:::get_n_features(clustered_data),
        sum(cluster_size))
    } else {
      testthat::expect_equal(
        familiar:::get_n_features(clustered_data),
        length(cluster_size))
    }
    
    if (parameters$cluster_cut_method != "none" &&
        parameters$cluster_representation_method != "none") {
      # Assert that the individual clusters have their expected size.
      for (ii in seq_along(cluster_size)) {
        # Get the base feature name
        base_feature_name <- paste0("feature_", ii)
        
        # Compute the number of features in each cluster that start with the
        # base feature name.
        n_matching_features <- sapply(
          split(cluster_table, by = "cluster_name"),
          function(x, base_feature_name) {
            sum(grepl(
              pattern = base_feature_name,
              x = x$feature_name))
          },
          base_feature_name = base_feature_name)
        
        # Test that the number of features that start with the base feature name
        # match the expected cluster size.
        testthat::expect_equal(any(n_matching_features == cluster_size[ii]), TRUE)
      }
      
      # Check representative feature.
      if (!parameters$cluster_representation_method %in% c("mean", "none")) {
        for (ii in seq_along(cluster_size)) {
          # Get the base feature name
          base_feature_name <- paste0("feature_", ii)
          
          # Get cluster corresponding to the base feature name.
          x <- cluster_table[grepl(pattern = base_feature_name, x = cluster_name)]
          
          # Check that only one required feature per cluster exists.
          testthat::expect_equal(sum(x$feature_required), 1L)
          
          # Create name for representative features and cluster name.
          representative_feature <- x[feature_required == TRUE]$feature_name[1]
          current_cluster_name <- paste0(representative_feature, "_cluster")
          
          # Test that the weight for the representative feature is 1.0.
          testthat::expect_equal(
            feature_info_list[[representative_feature]]@cluster_parameters@weight,
            1.0)
          
          # Test that the cluster size is as advertised.
          testthat::expect_equal(
            feature_info_list[[representative_feature]]@cluster_parameters@cluster_size,
            cluster_size[ii])
          
          # Test again that the cluster size is correct, but now using x.
          testthat::expect_equal(nrow(x), cluster_size[ii])
        }
      }
      
      # Check weights.
      if (!parameters$cluster_representation_method %in% c("none")) {
        for (ii in seq_along(cluster_size)) {
          # Get the base feature name
          base_feature_name <- paste0("feature_", ii)
          
          # Get cluster corresponding to the base feature name.
          x <- cluster_table[grepl(pattern = base_feature_name, x = cluster_name)]
          current_cluster_name <- x$cluster_name[1]
          
          # Obtain weights from all features in the cluster.
          cluster_features <- x$feature_name
          feature_weights <- sapply(
            feature_info_list[cluster_features],
            function(x) (x@cluster_parameters@weight))
          
          # Assert that the sum of the weights is 1.0.
          testthat::expect_equal(sum(feature_weights), 1.0)
          
          if (parameters$cluster_representation_method == "mean") {
            if (any(sapply(feature_info_list[cluster_features], function(x) (x@feature_type)) == "factor")) {
              # For clusters consisting of categorical or mixed features, there
              # is one feature with non-zero weight. "mean" is not really
              # definable for such feature sets.
              testthat::expect_equal(nrow(x[feature_required == TRUE]), 1L)
              
            } else {
              # Assert that all weights are the same.
              testthat::expect_equal(all(feature_weights == feature_weights[1]), TRUE)
            }
          }
        }
      }
    }
    
    # Check whether data is correctly presented.
    for (x in split(cluster_table, by = "cluster_name")) {
      # Do not check if more than feature is required.
      if (sum(x$feature_required) > 1L) next
      
      # Find the representative feature.
      representative_feature <- x[feature_required == TRUE]$feature_name
      current_cluster_name <- x$cluster_name[1]
      
      # Check that the representative feature has weight 1.0.
      testthat::expect_equal(
        feature_info_list[[representative_feature]]@cluster_parameters@weight,
        1.0)
      
      # Test that the cluster data for the data and clustered_data data sets are
      # the same.
      testthat::expect_equal(
        clustered_data@data[[current_cluster_name]],
        data@data[[representative_feature]])
      
      # Test that the cluster data for the one-sample dataset and the clustered
      # data are the same.
      testthat::expect_equal(
        clustered_one_sample_data@data[[current_cluster_name]],
        one_sample_data@data[[representative_feature]])
    }
  }
  )
}


# One feature dataset ----------------------------------------------------------
generic_test <- generate_test_parameters(
  similarity_metric = "mcfadden_r2",
  similarity_threshold = 0.99,
  n_numeric_features = c(4, 0))

while (TRUE) {
  # Generate parameters.
  parameters <- generic_test()
  if (coro::is_exhausted(parameters)) break

  # Set cluster size.
  cluster_size <- c(1, 1, 1, 1)

  # Create data,
  data <- familiar:::test_create_synthetic_correlated_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Select only first feature.
  data <- familiar:::select_features(
    data = data, 
    features = "feature_1")

  # Create a list of featureInfo objects.
  feature_info_list <- test_create_generic_info(data = data)

  # Create cluster skeletons
  feature_info_list <- familiar:::create_cluster_parameter_skeleton(
    feature_info_list,
    cluster_method = parameters$cluster_method,
    cluster_linkage = parameters$cluster_linkage,
    cluster_cut_method = parameters$cluster_cut_method,
    cluster_similarity_threshold = parameters$similarity_threshold,
    cluster_similarity_metric = parameters$similarity_metric,
    cluster_representation_method = parameters$cluster_representation_method)

  # Update feature info with cluster parameters.
  feature_info_list <- familiar:::add_cluster_info(
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose)

  # Assume that the data is imputed.
  data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_data <- familiar:::cluster_features(
    data = data,
    feature_info_list = feature_info_list)

  # Create clustering table.
  cluster_table <- familiar:::.create_clustering_table(
    feature_info_list = feature_info_list)

  # Perform tests
  testthat::test_that(paste0(
    "Clustering works correctly for the ", parameters$cluster_method, " clustering method, ",
    "with cluster cut method \"", parameters$cluster_cut_method, "\" and ",
    "cluster representation method \"", parameters$cluster_representation_method, "\" for ",
    "normal synthetic data with a single feature column and a ", parameters$outcome_type, " outcome."
  ), {
    # Test that exactly one cluster is formed.
    testthat::expect_equal(familiar:::get_n_features(clustered_data), 1L)

    # Check that the cluster is formed by one feature.
    testthat::expect_equal(nrow(cluster_table), 1L)

    # Check that the representative feature has weight 1.0.
    testthat::expect_equal(
      feature_info_list[["feature_1"]]@cluster_parameters@weight, 
      1.0)

    # Test that the cluster data for the data and clustered_data data sets are
    # the same.
    testthat::expect_equal(
      clustered_data@data[["feature_1"]],
      data@data[["feature_1"]])
  })
}


# Two-feature dataset (uncorrelated) -------------------------------------------
generic_test <- generate_test_parameters(
  similarity_metric = "mcfadden_r2",
  similarity_threshold = 0.99,
  n_numeric_features = c(4, 3, 0))

while (TRUE) {
  # Generate parameters.
  parameters <- generic_test()
  if (coro::is_exhausted(parameters)) break

  # Set cluster size.
  cluster_size <- c(1, 1, 1, 1)

  # Create data,
  data <- familiar:::test_create_synthetic_correlated_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Select the first two features.
  data <- familiar:::select_features(
    data = data, 
    features = c("feature_1", "feature_2"))

  # Create a list of featureInfo objects.
  feature_info_list <- test_create_generic_info(data = data)

  # Create cluster skeletons
  feature_info_list <- familiar:::create_cluster_parameter_skeleton(
    feature_info_list,
    cluster_method = parameters$cluster_method,
    cluster_linkage = parameters$cluster_linkage,
    cluster_cut_method = parameters$cluster_cut_method,
    cluster_similarity_threshold = parameters$similarity_threshold,
    cluster_similarity_metric = parameters$similarity_metric,
    cluster_representation_method = parameters$cluster_representation_method)

  # Update feature info with cluster parameters.
  feature_info_list <- familiar:::add_cluster_info(
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose)

  # Assume that the data is imputed.
  data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_data <- familiar:::cluster_features(
    data = data,
    feature_info_list = feature_info_list)

  # Create clustering table.
  cluster_table <- familiar:::.create_clustering_table(
    feature_info_list = feature_info_list)

  # Perform tests
  testthat::test_that(paste0(
    "Clustering works correctly for the ", parameters$cluster_method, " clustering method, ",
    "with cluster cut method \"", parameters$cluster_cut_method, "\" and ",
    "cluster representation method \"", parameters$cluster_representation_method, "\" for ",
    "normal synthetic data with two uncorrelated feature columns and a ", parameters$outcome_type, " outcome."
  ), {
    # Test that exactly two clusters are formed.
    testthat::expect_equal(familiar:::get_n_features(clustered_data), 2L)

    # Check whether data is correctly presented.
    for (x in split(cluster_table, by = "cluster_name")) {
      # Find the representative feature.
      representative_feature <- x[feature_required == TRUE]$feature_name
      current_cluster_name <- x$cluster_name[1]

      # Check that the clusters are singular.
      testthat::expect_equal(nrow(x), 1L)

      # Check that the representative feature has weight 1.0.
      testthat::expect_equal(
        feature_info_list[[representative_feature]]@cluster_parameters@weight,
        1.0)

      # Test that the cluster data for the data and
      # clustered_data data sets are the same.
      testthat::expect_equal(
        clustered_data@data[[current_cluster_name]],
        data@data[[representative_feature]])
    }
  })
}



# Two-feature dataset (correlated) ---------------------------------------------
generic_test <- generate_test_parameters(
  similarity_metric = "mcfadden_r2",
  similarity_threshold = 0.99,
  n_numeric_features = c(4, 0))

while (TRUE) {
  # Generate parameters.
  parameters <- generic_test()
  if (coro::is_exhausted(parameters)) break

  # Set cluster size.
  cluster_size <- c(2, 1, 1, 1)

  # Create data,
  data <- familiar:::test_create_synthetic_correlated_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Select only first two feature.
  data <- familiar:::select_features(
    data = data, 
    features = c("feature_1_A", "feature_1_B"))

  # Create a list of featureInfo objects.
  feature_info_list <- test_create_generic_info(data = data)
  
  # Create cluster skeletons
  feature_info_list <- familiar:::create_cluster_parameter_skeleton(
    feature_info_list,
    cluster_method = parameters$cluster_method,
    cluster_linkage = parameters$cluster_linkage,
    cluster_cut_method = parameters$cluster_cut_method,
    cluster_similarity_threshold = parameters$similarity_threshold,
    cluster_similarity_metric = parameters$similarity_metric,
    cluster_representation_method = parameters$cluster_representation_method)

  # Update feature info with cluster parameters.
  feature_info_list <- familiar:::add_cluster_info(
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose)

  # Assume that the data is imputed.
  data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_data <- familiar:::cluster_features(
    data = data,
    feature_info_list = feature_info_list)

  # Create clustering table.
  cluster_table <- familiar:::.create_clustering_table(
    feature_info_list = feature_info_list)

  # Perform tests
  testthat::test_that(paste0(
    "Clustering works correctly for the ", parameters$cluster_method, " clustering method, ",
    "with cluster cut method \"", parameters$cluster_cut_method, "\" and ",
    "cluster representation method \"", parameters$cluster_representation_method, "\" for ",
    "normal synthetic data with two correlated feature columns and a ", parameters$outcome_type, " outcome."
  ), {
    if (parameters$cluster_method %in% c("none") ||
        parameters$cluster_cut_method == "none" ||
        parameters$cluster_representation_method == "none") {
      # Test that exactly two clusters of single features are formed.
      testthat::expect_equal(familiar:::get_n_features(clustered_data), 2L)
      
    } else {
      # Test that exactly one cluster of two features is formed.
      testthat::expect_equal(familiar:::get_n_features(clustered_data), 1L)
    }
  })
}



# Dataset with signature features ----------------------------------------------

generic_test <- generate_test_parameters(
  similarity_metric = "mcfadden_r2",
  similarity_threshold = 0.99,
  n_numeric_features = c(4, 0))

while (TRUE) {
  # Generate parameters.
  parameters <- generic_test()
  if (coro::is_exhausted(parameters)) break

  # Set cluster size.
  cluster_size <- c(2, 2, 2, 2)

  # Create data,
  data <- familiar:::test_create_synthetic_correlated_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Create a list of featureInfo objects.
  feature_info_list <- test_create_generic_info(data = data)

  # Set signature features
  feature_info_list <- familiar:::add_signature_info(
    feature_info_list = feature_info_list,
    signature = c("feature_1_A", "feature_1_B"))

  # Create cluster skeletons
  feature_info_list <- familiar:::create_cluster_parameter_skeleton(
    feature_info_list,
    cluster_method = parameters$cluster_method,
    cluster_linkage = parameters$cluster_linkage,
    cluster_cut_method = parameters$cluster_cut_method,
    cluster_similarity_threshold = parameters$similarity_threshold,
    cluster_similarity_metric = parameters$similarity_metric,
    cluster_representation_method = parameters$cluster_representation_method)

  # Update feature info with cluster parameters.
  feature_info_list <- familiar:::add_cluster_info(
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose)

  # Assume that the data is imputed.
  data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_data <- familiar:::cluster_features(
    data = data,
    feature_info_list = feature_info_list)

  # Create clustering table.
  cluster_table <- familiar:::.create_clustering_table(
    feature_info_list = feature_info_list)

  # Perform tests
  testthat::test_that(paste0(
    "Clustering works correctly for the ", parameters$cluster_method, " clustering method, ",
    "with cluster cut method \"", parameters$cluster_cut_method, "\" and ",
    "cluster representation method \"", parameters$cluster_representation_method, "\" for ",
    "normal synthetic data with two signature features and a ", parameters$outcome_type, " outcome."
  ), {
    if (parameters$cluster_method %in% c("none") ||
        parameters$cluster_cut_method == "none" ||
        parameters$cluster_representation_method == "none") {
      # Test that exactly eight clusters of single features are formed when
      # clustering is disabled.
      testthat::expect_equal(familiar:::get_n_features(clustered_data), 8L)
      
    } else {
      # Test that exactly fives clusters are formed.
      testthat::expect_equal(familiar:::get_n_features(clustered_data), 5L)

      # Test that feature_1_A and feature_1_B appear in separate, singular
      # clusters.
      for (x in split(cluster_table, by = "cluster_name")) {
        if ("feature_1_A" %in% x$feature_name) {
          testthat::expect_equal(nrow(x), 1L)

          # Check that the signature feature values remain unaltered.
          testthat::expect_equal(
            data@data[["feature_1_A"]],
            clustered_data@data[["feature_1_A"]])
          
        } else if ("feature_1_B" %in% x$feature_name) {
          testthat::expect_equal(nrow(x), 1L)

          # Check that the signature feature values remain unaltered.
          testthat::expect_equal(
            data@data[["feature_1_B"]],
            clustered_data@data[["feature_1_B"]])
          
        } else {
          # Assert that clusters are formed by two features with the same base
          # name.
          base_feature_name <- substr(x$feature_name[1], 1, 9)

          testthat::expect_equal(nrow(x), 2L)
          testthat::expect_equal(
            sum(grepl(pattern = base_feature_name, x = x$feature_name)),
            2L)
        }
      }
    }
  })
}



# Dataset with one invariant feature -------------------------------------------

generic_test <- generate_test_parameters(
  similarity_metric = "mcfadden_r2",
  similarity_threshold = 0.99,
  n_numeric_features = c(4, 0))

while (TRUE) {
  # Generate parameters.
  parameters <- generic_test()
  if (coro::is_exhausted(parameters)) break

  # Set cluster size. Note that the second feature is made invariant.
  cluster_size <- c(2, 1, 2, 2)

  # Create data,
  data <- familiar:::test_create_synthetic_correlated_one_feature_invariant_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Create a list of featureInfo objects.
  feature_info_list <- test_create_generic_info(data = data)

  # Create cluster skeletons
  feature_info_list <- familiar:::create_cluster_parameter_skeleton(
    feature_info_list,
    cluster_method = parameters$cluster_method,
    cluster_linkage = parameters$cluster_linkage,
    cluster_cut_method = parameters$cluster_cut_method,
    cluster_similarity_threshold = parameters$similarity_threshold,
    cluster_similarity_metric = parameters$similarity_metric,
    cluster_representation_method = parameters$cluster_representation_method)

  # Update feature info with cluster parameters.
  feature_info_list <- familiar:::add_cluster_info(
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose)

  # Assume that the data is imputed.
  data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_data <- familiar:::cluster_features(
    data = data,
    feature_info_list = feature_info_list)

  # Create clustering table.
  cluster_table <- familiar:::.create_clustering_table(
    feature_info_list = feature_info_list)

  # Perform tests
  testthat::test_that(paste0(
    "Clustering works correctly for the ", parameters$cluster_method, " clustering method, ",
    "with cluster cut method \"", parameters$cluster_cut_method, "\" and ",
    "cluster representation method \"", parameters$cluster_representation_method, "\" for ",
    "a synthetic dataset with an invariant feature and a ", parameters$outcome_type, " outcome."
  ), {
    # The expectation is that clustering functions normally.

    if (parameters$cluster_method %in% c("none") ||
        parameters$cluster_cut_method == "none" ||
        parameters$cluster_representation_method == "none") {
      # Test that exactly eight clusters of single features are formed when
      # clustering is disabled.
      testthat::expect_equal(
        familiar:::get_n_features(clustered_data),
        sum(cluster_size))
      
    } else {
      # Test that the correct number of clusters is formed.
      testthat::expect_equal(
        familiar:::get_n_features(clustered_data),
        length(cluster_size))

      for (x in split(cluster_table, by = "cluster_name")) {
        if ("feature_2" %in% x$feature_name) {
          # Test that feature_2 appears in a separate, singular cluster.

          testthat::expect_equal(nrow(x), 1L)

          # Check that the signature feature values remain unaltered.
          testthat::expect_equal(
            data@data[["feature_2"]],
            clustered_data@data[["feature_2"]])
          
        } else {
          # Assert that clusters are formed by two features
          # with the same base name.
          base_feature_name <- substr(x$feature_name[1], 1, 9)

          testthat::expect_equal(nrow(x), 2L)
          testthat::expect_equal(
            sum(grepl(pattern = base_feature_name, x = x$feature_name)),
            2L)
        }
      }
    }
  })
}



#### One-sample dataset --------------------------------------------------------

generic_test <- generate_test_parameters(
  similarity_metric = "mcfadden_r2",
  similarity_threshold = 0.99,
  n_numeric_features = c(4, 0))

while (TRUE) {
  # Generate parameters.
  parameters <- generic_test()
  if (coro::is_exhausted(parameters)) break

  # Set cluster size.
  cluster_size <- c(2, 2, 2, 2)

  # Create data,
  data <- familiar:::test_create_synthetic_correlated_one_sample_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Create a list of featureInfo objects.
  feature_info_list <- test_create_generic_info(data = data)

  # Create cluster skeletons
  feature_info_list <- familiar:::create_cluster_parameter_skeleton(feature_info_list,
    cluster_method = parameters$cluster_method,
    cluster_linkage = parameters$cluster_linkage,
    cluster_cut_method = parameters$cluster_cut_method,
    cluster_similarity_threshold = parameters$similarity_threshold,
    cluster_similarity_metric = parameters$similarity_metric,
    cluster_representation_method = parameters$cluster_representation_method)

  # Update feature info with cluster parameters.
  feature_info_list <- familiar:::add_cluster_info(
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose)

  # Assume that the data is imputed.
  data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_data <- familiar:::cluster_features(
    data = data,
    feature_info_list = feature_info_list)

  # Create clustering table.
  cluster_table <- familiar:::.create_clustering_table(
    feature_info_list = feature_info_list)

  # Perform tests
  testthat::test_that(paste0(
    "Clustering works correctly for the ", parameters$cluster_method, " clustering method, ",
    "with cluster cut method \"", parameters$cluster_cut_method, "\" and ",
    "cluster representation method \"", parameters$cluster_representation_method, "\" for ",
    "a single-sample synthetic dataset with a ", parameters$outcome_type, " outcome."
  ), {
    # The expectation is that clustering is skipped because one cannot really
    # determine similarity if there is only a single value present.

    # Assert that the correct number of clusters is formed.
    testthat::expect_equal(
      familiar:::get_n_features(clustered_data),
      sum(cluster_size))
    testthat::expect_equal(
      data.table::uniqueN(cluster_table$cluster_name),
      sum(cluster_size))
  })
}


# Invariant outcome dataset ----------------------------------------------------

generic_test <- generate_test_parameters(
  similarity_metric = "mcfadden_r2",
  similarity_threshold = 0.99,
  n_numeric_features = c(4, 0))

while (TRUE) {
  # Generate parameters.
  parameters <- generic_test()
  if (coro::is_exhausted(parameters)) break

  # Set cluster size.
  cluster_size <- c(2, 2, 2, 2)

  # Create data,
  data <- familiar:::test_create_synthetic_correlated_one_outcome_data(
    outcome_type = parameters$outcome_type,
    n_numeric = parameters$n_numeric,
    cluster_size = cluster_size)

  # Create a list of featureInfo objects.
  feature_info_list <- test_create_generic_info(data = data)

  # Create cluster skeletons
  feature_info_list <- familiar:::create_cluster_parameter_skeleton(
    feature_info_list,
    cluster_method = parameters$cluster_method,
    cluster_linkage = parameters$cluster_linkage,
    cluster_cut_method = parameters$cluster_cut_method,
    cluster_similarity_threshold = parameters$similarity_threshold,
    cluster_similarity_metric = parameters$similarity_metric,
    cluster_representation_method = parameters$cluster_representation_method)

  # Update feature info with cluster parameters.
  feature_info_list <- familiar:::add_cluster_info(
    feature_info_list = feature_info_list,
    data = data,
    verbose = verbose)

  # Assume that the data is imputed.
  data@preprocessing_level <- "imputation"

  # Cluster features.
  clustered_data <- familiar:::cluster_features(
    data = data,
    feature_info_list = feature_info_list)

  # Create clustering table.
  cluster_table <- familiar:::.create_clustering_table(
    feature_info_list = feature_info_list)

  # Perform tests
  testthat::test_that(paste0(
    "Clustering works correctly for the ", parameters$cluster_method, " clustering method, ",
    "with cluster cut method \"", parameters$cluster_cut_method, "\" and ",
    "cluster representation method \"", parameters$cluster_representation_method, "\" for ",
    "a synthetic dataset with an invariant ", parameters$outcome_type, " outcome."
  ), {
    # The expectation is that clustering functions normally.

    if (parameters$cluster_method %in% c("none") ||
        parameters$cluster_cut_method == "none" ||
        parameters$cluster_representation_method == "none") {
      # Test that exactly eight clusters of single features are formed when
      # clustering is disabled.
      testthat::expect_equal(
        familiar:::get_n_features(clustered_data),
        sum(cluster_size))
      
    } else {
      # Test that the correct number of clusters is formed.
      testthat::expect_equal(
        familiar:::get_n_features(clustered_data),
        length(cluster_size))

      for (x in split(cluster_table, by = "cluster_name")) {
        # Assert that clusters are formed by two features
        # with the same base name.
        base_feature_name <- substr(x$feature_name[1], 1, 9)

        testthat::expect_equal(nrow(x), 2L)
        testthat::expect_equal(
          sum(grepl(pattern = base_feature_name, x = x$feature_name)),
          2L)
      }
    }
  })
}

.start_random_number_stream <- function(seed) {
  if (is.null(seed)) {
    return(NULL)
  }

  r <- methods::new("rstream.runif", seed = seed)

  return(r)
}



..fam_sample <- function(
    x, 
    size, 
    replace, 
    seed = NULL, 
    rstream_object = NULL
) {
  # Create a random generator stream to avoid touching seeds globally.

  if (is.null(seed) && is.null(rstream_object)) {
    ..error_reached_unreachable_code(
      "..fam_sample: seed and rstream_object cannot both be NULL"
    )
  }

  if (!is.null(seed) && is.null(rstream_object)) {
    rstream_object <- .start_random_number_stream(seed = seed)
  }

  if (!is(rstream_object, "rstream.runif")) {
    ..error_reached_unreachable_code(
      "..fam_sample: rstream_object is not an rstream.runif object."
    )
  }

  # Determine the number of random values to generate. When samples are
  # replaced, "size" is sufficient. Otherwise we need to draw a number of random
  # values up to x.
  n_random_values <- ifelse(replace, size, length(x))

  # Obtain random values from the interval [0, 1].
  random_values <- rstream::rstream.sample(rstream_object, n_random_values)

  if (replace) {
    # Generate sample indices.
    x_indices <- as.integer(ceiling(random_values * length(x)))

    # Replace index 0, if present.
    x_indices[x_indices == 0L] <- 1L
    
  } else {
    # Create sample indices by ranking indices.
    x_indices <- data.table::frank(random_values, ties.method = "first")

    # Select the first x_indices (up to size)
    x_indices <- head(x_indices, n = size)
  }

  # Return sampled values.
  return(x[x_indices])
}



fam_rnorm <- function(
    n,
    mean = 0.0, 
    sd = 1.0, 
    seed = NULL, 
    rstream_object = NULL) {
  # Based on Marsaglia polar method.

  if (is.null(seed) && is.null(rstream_object)) {
    return(stats::rnorm(n = n, mean = mean, sd = sd))
  }

  return(.fam_rnorm(
    n = n, 
    mean = mean, 
    sd = sd, 
    seed = seed, 
    rstream_object = rstream_object
  ))
}



.fam_rnorm <- function(
    n,
    mean = 0.0, 
    sd = 1.0, 
    seed = NULL,
    rstream_object = NULL
) {
  # Based on Marsaglia polar method.

  if (is.null(seed) && is.null(rstream_object)) {
    ..error_reached_unreachable_code(
      "..fam_rnorm: seed and rstream_object cannot both be NULL"
    )
  }

  if (!is.null(seed) && is.null(rstream_object)) {
    rstream_object <- .start_random_number_stream(seed = seed)
  }

  if (!is(rstream_object, "rstream.runif")) {
    ..error_reached_unreachable_code(
      "..fam_rnorm: rstream_object is not an rstream.runif object."
    )
  }

  x <- sapply(
    seq_len(n),
    ..fam_rnorm, 
    rstream_object = rstream_object
  )

  return(mean + sd * x)
}


..fam_rnorm <- function(ii, rstream_object) {
  # Based on Marsaglia polar method.

  while (TRUE) {
    x_1 <- rstream::rstream.sample(rstream_object, 1L) * 2.0 - 1.0
    x_2 <- rstream::rstream.sample(rstream_object, 1L) * 2.0 - 1.0

    s <- x_1^2.0 + x_2^2.0

    if (s < 1.0) break
  }

  if (s == 0.0) return(0.0)

  return(x_1 * sqrt(-2.0 * log(s) / s))
}



fam_runif <- function(
    n, 
    min = 0.0, 
    max = 1.0, 
    seed = NULL, 
    rstream_object = NULL
) {
  
  if (is.null(seed) && is.null(rstream_object)) {
    return(stats::runif(
      n = n, 
      min = min, 
      max = max
    ))
  }

  return(.fam_runif(
    n = n, 
    min = min, 
    max = max, 
    seed = seed,
    rstream_object = rstream_object
  ))
}



.fam_runif <- function(
    n, 
    min = 0.0, 
    max = 1.0, 
    seed = NULL, 
    rstream_object = NULL
) {
  if (is.null(seed) && is.null(rstream_object)) {
    ..error_reached_unreachable_code(
      "..fam_runif: seed and rstream_object cannot both be NULL"
    )
  }

  if (!is.null(seed) && is.null(rstream_object)) {
    rstream_object <- .start_random_number_stream(seed = seed)
  }

  if (!is(rstream_object, "rstream.runif")) {
    ..error_reached_unreachable_code(
      "..fam_runif: rstream_object is not an rstream.runif object."
    )
  }

  # Create uniformly sampled data.
  x <- rstream::rstream.sample(rstream_object, n)

  return(min + x * (max - min))
}

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

results_fun <- function(t, x, model) {
  predictions <- familiar::predict(
    object = model,
    newdata = x,
    type = "survival_probability",
    time = t
  )
  
  return(predictions$predicted_outcome)
} 

# Create test data set ---------------------------------------------------------

# Create random stream object so that the same numbers are produced every
# time.
r <- familiar:::.start_random_number_stream(seed = 1963L)
n_series_instances <- 1000L

# Draw random numbers for 1 feature.
feature_1 <- familiar:::fam_rnorm(n_series_instances, rstream_object = r)

# Simulate event times using exponential function and lambda = 0.5
outcome_time <- - log(familiar:::fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)) / 
  (0.5 * exp(feature_1))

# Simulate censoring times using exponential function and lambda = 0.1
censor_time <- - log(familiar:::fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)) / 0.1
outcome_event <- rep_len(1L, length.out = n_series_instances)
outcome_event[censor_time < outcome_time] <- 0L
outcome_time[outcome_event == 0L] <- censor_time[outcome_event == 0L]

# Create basic table.
data <- data.table::data.table(
  "batch_id" = "basic",
  "sample_id" = paste0("sample_", seq_len(n_series_instances)),
  "series_id" = 1L,
  "feature_1" = feature_1,
  "outcome_time" = outcome_time,
  "outcome_event" = outcome_event
)

# Turn into a data object.
data <- familiar::as_data_object(
  data = data,
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  outcome_column = c("outcome_time", "outcome_event"),
  outcome_type = "survival"
)

# Single test-dataset
test_data <- data.table::data.table(
  "batch_id" = "basic",
  "sample_id" = 1001L,
  "series_id" = 1L,
  "feature_1" = 0.0,
  "outcome_time" = 1.386,
  "outcome_event" = 1L
)

test_data <- familiar::as_data_object(
  data = test_data,
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  outcome_column = c("outcome_time", "outcome_event"),
  outcome_type = "survival"
)

# Train models -----------------------------------------------------------------

# Cox PH
cox_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  normalisation_method = "none",
  transformation_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "cox"
)

# Random forest model (ranger)
rf_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  normalisation_method = "none",
  transformation_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "random_forest_ranger_default"
)

# Survival regression model
surv_reg_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  normalisation_method = "none",
  transformation_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "survival_regr_weibull"
)

# Evaluate predictions ---------------------------------------------------------

time_points <- (seq_len(51L) - 1L) / 10.0

# Get Kaplan-Meier data.
km_fit <- survival::survfit(
  Surv(outcome_time, outcome_event) ~ 1,
  data = data@data
)

km_fit_prob <- stats::approx(
  x = km_fit$time,
  y = km_fit$surv,
  xout = time_points,
  method = "linear",
  rule = 2L,
  yleft = 1.0
)$y


# Cox PH
cox_predictions <- sapply(
  time_points,
  results_fun,
  x = test_data,
  model = cox_model
)

testthat::test_that(
  "Cox PH predictions are plausible.", 
  {
    # For the current data, predictions should be very close to the KM-curve, 
    # since feature = 0 means a relative risk of approximately 0.
    testthat::expect_true(all(abs(km_fit_prob - cox_predictions) < 0.01))
  }
)


# Weibull model
surv_reg_predictions <- sapply(
  time_points,
  results_fun,
  x = test_data,
  model = surv_reg_model
)

testthat::test_that(
  "Weibull regression predictions are plausible.",
  {
    # For the current data, predictions should be very close to the KM-curve, 
    # since feature = 0 means a relative risk of approximately 0. However,
    # the Weibull model does not rely on the observed survival curve, and thus
    # will deviate from the Kaplan-Meier curve somewhat.
    testthat::expect_true(all(abs(km_fit_prob - surv_reg_predictions) < 0.10))
  }
)


# RF model (ranger)
rf_predictions <- sapply(
  time_points,
  results_fun,
  x = test_data,
  model = rf_model
)

testthat::test_that(
  "Random forest (ranger)-based predictions are plausible.",
  {
    # For the current data, predictions should be very close to the KM-curve, 
    # since feature = 0 means a relative risk of approximately 0. However,
    # the RF model does not rely on the observed survival curve, and thus
    # will deviate from the Kaplan-Meier curve somewhat.
    testthat::expect_true(all(abs(km_fit_prob - rf_predictions) < 0.15))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# promote_learner (model) ------------------------------------------------------
setMethod(
  "promote_learner",
  signature(object = "familiarModel"),
  function(object) {
    learner <- object@learner

    if (learner %in% .get_available_naive_bayes_learners()) {
      # Naive bayes model
      object <- methods::new("familiarNaiveBayes", object)
    } else if (learner %in% .get_available_knn_learners()) {
      # k-nearest neighbours model with radial kernel
      object <- methods::new("familiarKNN", object)
    } else if (learner %in% .get_available_svm_c_learners()) {
      # C-classification support vector machines.
      object <- methods::new("familiarSVMC", object)
    } else if (learner %in% .get_available_svm_nu_learners()) {
      # Nu-classification and regression support vector machines.
      object <- methods::new("familiarSVMNu", object)
    } else if (learner %in% .get_available_svm_eps_learners()) {
      # Epsilon regression support vector machines.
      object <- methods::new("familiarSVMEps", object)
    } else if (learner %in% .get_available_glm_learners()) {
      # Generalised linear models
      object <- methods::new("familiarGLM", object)
    } else if (learner %in% .get_available_glmnet_ridge_learners()) {
      # Ridge penalised regression models
      object <- methods::new("familiarGLMnetRidge", object)
    } else if (learner %in% .get_available_glmnet_lasso_learners()) {
      # Lasso penalised regression models
      object <- methods::new("familiarGLMnetLasso", object)
    } else if (learner %in% .get_available_glmnet_elastic_net_learners()) {
      # Elastic net penalised regression models.
      object <- methods::new("familiarGLMnetElasticNet", object)
    } else if (learner %in% .get_available_mboost_lm_learners()) {
      # Boosted generalised linear models
      object <- methods::new("familiarMBoostLM", object)
    } else if (learner %in% .get_available_xgboost_lm_learners()) {
      # Extreme gradient boosted linear models
      object <- methods::new("familiarXGBoostLM", object)
    } else if (learner %in% .get_available_cox_learners()) {
      # Cox proportional hazards regression model
      object <- methods::new("familiarCoxPH", object)
    } else if (learner %in% .get_available_survival_regression_learners()) {
      # Fully parametric survival regression model
      object <- methods::new("familiarSurvRegr", object)
    } else if (learner %in% .get_available_rfsrc_learners()) {
      # Random forests for survival, regression, and classification
      object <- methods::new("familiarRFSRC", object)
    } else if (learner %in% .get_available_rfsrc_default_learners()) {
      # Random forests for survival, regression, and classification
      object <- methods::new("familiarRFSRCDefault", object)
    } else if (learner %in% .get_available_ranger_learners()) {
      # Ranger random forests
      object <- methods::new("familiarRanger", object)
    } else if (learner %in% .get_available_ranger_default_learners()) {
      # Ranger random forests
      object <- methods::new("familiarRangerDefault", object)
    } else if (learner %in% .get_available_mboost_tree_learners()) {
      # Boosted regression trees
      object <- methods::new("familiarMBoostTree", object)
    } else if (learner %in% .get_available_xgboost_tree_learners()) {
      # Extreme gradient boosted trees
      object <- methods::new("familiarXGBoostTree", object)
    } else if (learner %in% .get_available_xgboost_dart_learners()) {
      # Extreme gradient boosted trees
      object <- methods::new("familiarXGBoostDart", object)
    } else if (learner %in% .get_available_glmnet_lasso_learners_test_all_fail()) {
      # Lasso penalised regression models for testing purposes.
      object <- methods::new("familiarGLMnetLassoTestAllFail", object)
    } else if (learner %in% .get_available_glmnet_lasso_learners_test_some_fail()) {
      # Lasso penalised regression models for testing purposes.
      object <- methods::new("familiarGLMnetLassoTestSomeFail", object)
    } else if (learner %in% .get_available_glmnet_lasso_learners_test_extreme()) {
      # Lasso penalised regression models for testing purposes.
      object <- methods::new("familiarGLMnetLassoTestAllExtreme", object)
    }

    # Add package version.
    object <- add_package_version(object = object)

    # Returned object can be a standard familiarModel
    return(object)
  }
)



.get_learner_hyperparameters <- function(
    data, 
    learner, 
    outcome_type, 
    names_only = FALSE
) {
  # Get the outcome type from the data object, if available
  if (!is.null(data)) outcome_type <- data@outcome_type

  # Create familiarModel
  fam_model <- methods::new("familiarModel",
    learner = learner,
    outcome_type = outcome_type
  )

  # Set up the specific model
  fam_model <- promote_learner(fam_model)

  # Model hyperparameters
  model_hyperparameters <- get_default_hyperparameters(fam_model, data = data)

  # Extract names from parameter list
  if (names_only) model_hyperparameters <- names(model_hyperparameters)

  # Return hyperparameter list, or hyperparameter names
  return(model_hyperparameters)
}



.check_learner_outcome_type <- function(
    learner, 
    outcome_type, 
    as_flag = FALSE
) {
  # Create familiarModel
  fam_model <- methods::new("familiarModel",
    learner = learner,
    outcome_type = outcome_type
  )

  # Set up the specific model
  fam_model <- promote_learner(fam_model)

  # Check validity.
  learner_available <- is_available(fam_model)

  if (as_flag) return(learner_available)

  # Check if the familiar model has been successfully promoted.
  if (!is_subclass(class(fam_model)[1L], "familiarModel")) {
    ..error(
      paste0(
        learner, " is not a valid learner. ",
        "Please check the vignette for available learners."
      ),
      error_class = "input_argument_error"
    )
  }

  # Check if the learner is available.
  if (!learner_available) {
    ..error(
      paste0(learner, " is not available for \"", outcome_type, "\" outcomes."),
      error_class = "input_argument_error"
    )
  }

  # Check that the required package can be loaded.
  require_package(
    x = fam_model,
    purpose = paste0("to train models using the ", learner, " learner"),
    message_type = "backend_error"
  )
}



#' Internal function for obtaining a default signature size parameter
#'
#' @param data dataObject class object which contains the data on which the
#'   preset parameters are determined.
#' @param restrict_samples Logical indicating whether the signature size should
#'   be limited by the number of samples in addition to the number of available
#'   features. This may help convergence of OLS-based methods.
#'
#' @return List containing the preset values for the signature size parameter.
#'
#' @md
#' @keywords internal
.get_default_sign_size <- function(
    data,
    restrict_samples = FALSE
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome_event <- NULL

  # Determine the outcome type
  outcome_type <- data@outcome_type

  # Determine the number of samples and features
  n_samples <- data.table::uniqueN(
    data@data,
    by = get_id_columns(id_depth = "series")
  )
  n_features <- get_n_features(data)

  # Determine the actual range of features dynamically.
  if (restrict_samples && n_samples > 1L) {
    sign_size_range <- c(1L, min(n_samples - 1L, n_features))
  } else if (restrict_samples && n_samples <= 1L) {
    sign_size_range <- c(1L, 1L)
  } else {
    sign_size_range <- c(1L, n_features)
  }

  if (outcome_type %in% c("binomial", "multinomial")) {
    # Get the number of outcome classes
    n_classes <- nlevels(data@data$outcome)

    # Determine the range
    sign_size_default <- unique(c(1L, 2L, 5L, 10L, max(c(1.0, floor(n_samples / (n_classes * 7.5))))))
    
  } else if (outcome_type %in% c("survival")) {
    # Get the number of events
    n_events <- nrow(data@data[outcome_event == 1L, ])

    # Determine the range
    sign_size_default <- unique(c(1L, 2L, 5L, 10L, as.integer(max(c(1.0, floor(n_events / 15.0))))))
    
  } else if (outcome_type %in% c("continuous")) {
    # Determine the range
    sign_size_default <- unique(c(1L, 2L, 5L, 10L, as.integer(max(c(1.0, floor(n_samples / 15.0))))))
    
  } else {
    ..error_no_known_outcome_type(outcome_type)
  }

  # Limit default to those values that fall within the range.
  sign_size_default <- sign_size_default[
    sign_size_default >= sign_size_range[1L] &
      sign_size_default <= sign_size_range[2L]
  ]
  
  return(.set_hyperparameter(
    default = sign_size_default,
    type = "integer",
    range = sign_size_range,
    valid_range = c(0L, Inf),
    randomise = TRUE,
    distribution = "log"
  ))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_kaplan_meier (generic) --------------------------------------------------

#' @title Plot Kaplan-Meier survival curves.
#'
#' @description This function creates Kaplan-Meier survival curves from
#'   stratification data stored in a familiarCollection object.
#'
#' @param dir_path (*optional*) Path to the directory where created figures are
#'   saved to. Output is saved in the `stratification` subdirectory. If `NULL`
#'   no figures are saved, but are returned instead.
#' @param censoring (*optional*) Flag to indicate whether censored samples
#'   should be indicated on the survival curve.
#' @param censor_shape (*optional*) Shape used to indicate censored samples on
#'   the survival curve. Available shapes are documented in the `ggplot2`
#'   vignette *Aesthetic specifications*. By default a plus shape is used.
#' @param show_logrank (*optional*) Specifies whether the results of a logrank
#'   test to assess differences between the risk strata is annotated in the
#'   plot. A log-rank test can only be shown when `color_by` and `linestyle_by`
#'   are either unset, or only contain `group`.
#' @param show_survival_table (*optional*) Specifies whether a survival table is
#'   shown below the Kaplan-Meier survival curves. Survival in the risk strata
#'   is assessed for each of the breaks in `x_breaks`.
#' @param height (*optional*) Height of the plot. A default value is derived
#'   from number of facets and the inclusion of survival tables.
#' @param confidence_level (*optional*) Confidence level for the strata in the
#'   plot.
#'
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_risk_stratification_data -object
#'
#' @details This function generates a Kaplan-Meier survival plot based on risk
#'   group stratification by the learners.
#'
#'   `familiar` does not determine what units the x-axis has or what kind of
#'   survival the y-axis represents. It is therefore recommended to provide
#'   `x_label` and `y_label` arguments.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, `data_set`,
#'   `group` and `stratification_method`. By default, separate figures are
#'   created for each combination of `vimp_method` and `learner`, with faceting by
#'   `data_set`, colouring of the strata in each individual plot by
#'   `group`.
#'
#'   Greenwood confidence intervals of the Kaplan-Meier curve can be shown using
#'   various styles set by `conf_int_style`:
#'
#'  * `ribbon` (default): confidence intervals are shown as a ribbon with an
#'   opacity of `conf_int_alpha` around the point estimate of the Kaplan-Meier
#'   curve.
#'
#'  * `step` (default): confidence intervals are shown as a step function around
#'   the point estimate of the Kaplan-Meier curve.
#'
#'  * `none`: confidence intervals are not shown. The point estimate of the ROC
#'   curve is shown as usual.
#'
#'   Labelling methods such as `set_risk_group_names` or `set_data_set_names`
#'   can be applied to the `familiarCollection` object to update labels, and
#'   order the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_kaplan_meier
#' @md
#' @rdname plot_kaplan_meier-methods
setGeneric(
  "plot_kaplan_meier",
  function(
      object,
      draw = FALSE,
      dir_path = NULL,
      split_by = NULL,
      color_by = NULL,
      linetype_by = NULL,
      facet_by = NULL,
      facet_wrap_cols = NULL,
      combine_legend = TRUE,
      ggtheme = NULL,
      discrete_palette = NULL,
      x_label = "time",
      x_label_shared = "column",
      y_label = "survival probability",
      y_label_shared = "row",
      legend_label = waiver(),
      plot_title = waiver(),
      plot_sub_title = waiver(),
      caption = NULL,
      x_range = NULL,
      x_n_breaks = 5L,
      x_breaks = NULL,
      y_range = c(0.0, 1.0),
      y_n_breaks = 5L,
      y_breaks = NULL,
      confidence_level = NULL,
      conf_int_style = c("ribbon", "step", "none"),
      conf_int_alpha = 0.2,
      censoring = TRUE,
      censor_shape = "plus",
      show_logrank = TRUE,
      show_survival_table = TRUE,
      width = waiver(),
      height = waiver(),
      units = waiver(),
      export_collection = FALSE,
      ...
  ) {
    standardGeneric("plot_kaplan_meier")
  }
)



# plot_kaplan_meier (general) --------------------------------------------------

#' @rdname plot_kaplan_meier-methods
setMethod(
  "plot_kaplan_meier",
  signature(object = "ANY"),
  function(
      object,
      draw = FALSE,
      dir_path = NULL,
      split_by = NULL,
      color_by = NULL,
      linetype_by = NULL,
      facet_by = NULL,
      facet_wrap_cols = NULL,
      combine_legend = TRUE,
      ggtheme = NULL,
      discrete_palette = NULL,
      x_label = "time",
      x_label_shared = "column",
      y_label = "survival probability",
      y_label_shared = "row",
      legend_label = waiver(),
      plot_title = waiver(),
      plot_sub_title = waiver(),
      caption = NULL,
      x_range = NULL,
      x_n_breaks = 5L,
      x_breaks = NULL,
      y_range = c(0.0, 1.0),
      y_n_breaks = 5L,
      y_breaks = NULL,
      confidence_level = NULL,
      conf_int_style = c("ribbon", "step", "none"),
      conf_int_alpha = 0.2,
      censoring = TRUE,
      censor_shape = "plus",
      show_logrank = TRUE,
      show_survival_table = TRUE,
      width = waiver(),
      height = waiver(),
      units = waiver(),
      export_collection = FALSE,
      ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "risk_stratification_data",
          ".no_features_required" = TRUE
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_kaplan_meier,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "linetype_by" = linetype_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "combine_legend" = combine_legend,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "x_label_shared" = x_label_shared,
        "y_label" = y_label,
        "y_label_shared" = y_label_shared,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "confidence_level" = confidence_level,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "censoring" = censoring,
        "censor_shape" = censor_shape,
        "show_logrank" = show_logrank,
        "show_survival_table" = show_survival_table,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_kaplan_meier (collection) -----------------------------------------------

#' @rdname plot_kaplan_meier-methods
setMethod(
  "plot_kaplan_meier",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    linetype_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    combine_legend = TRUE,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = "time",
    x_label_shared = "column",
    y_label = "survival probability",
    y_label_shared = "row",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = c(0.0, 1.0),
    y_n_breaks = 5L,
    y_breaks = NULL,
    confidence_level = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    censoring = TRUE,
    censor_shape = "plus",
    show_logrank = TRUE,
    show_survival_table = TRUE,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    outcome_event <- .NATURAL <- NULL
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)

    # Get input data
    x <- export_risk_stratification_data(
      object = object,
      export_strata = FALSE
    )
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check that the data are not evaluated at the model level.
    if (all(sapply(x, function(x) (x@detail_level == "model")))) {
      ..warning_no_comparison_between_models()
      return(NULL)
    }

    # Obtain data element from list.
    if (is.list(x)) {
      if (is_empty(x)) return(NULL)

      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_kaplan_meier: list of data elements contains unmerged elements."
        )
      }

      # Get x directly.
      x <- x[[1L]]
    }
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    if (!all_predictions_valid(x)) return(NULL)

    # Remove non-valid risk groups
    x <- remove_invalid_predictions(x)

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = TRUE),
      purpose = "to create kaplan-meier survival curves",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Check input arguments ----------------------------------------------------

    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # x_label_shared
    if (!is.waive(x_label_shared)) {
      .check_input_plot_args(x_label_shared = x_label_shared)
    } else {
      x_label_shared <- "column"
    }

    # y_label_shared
    if (!is.waive(y_label_shared)) {
      .check_input_plot_args(y_label_shared = y_label_shared)
    } else {
      y_label_shared <- "row"
    }
    
    # x_range
    if (is.null(x_range)) {
      time_max <- object@outcome_info@time
      if (is.infinite(time_max)) {
        time_max <- .get_default_time_max(x@data[outcome_event == 1L, ]$outcome_time)
      }
      x_range <- c(0.0, time_max)
    }

    # x_breaks
    if (is.null(x_breaks)) {
      .check_input_plot_args(x_n_breaks = x_n_breaks)

      # Create breaks and update x_range
      x_breaks <- labeling::extended(
        m = x_n_breaks,
        dmin = x_range[1L],
        dmax = x_range[2L],
        only.loose = TRUE
      )

      x_range <- c(0.0, tail(x_breaks, n = 1L))
    }

    # y_range
    if (is.null(y_range)) y_range <- c(0.0, 1.0)

    # y_breaks
    if (is.null(y_breaks)) {
      .check_input_plot_args(y_n_breaks = y_n_breaks)

      # Create breaks and update y_range
      y_breaks <- labeling::extended(
        m = y_n_breaks,
        dmin = y_range[1L],
        dmax = y_range[2L]
      )

      y_range <- c(0.0, tail(y_breaks, n = 1L))
    }

    # confidence level
    if (!is.null(confidence_level)) {
      .check_number_in_valid_range(
        x = confidence_level,
        var_name = "confidence_level",
        range = c(0.0, 1.0),
        closed = c(FALSE, FALSE)
      )

      # Attach to object.
      x@confidence_level <- confidence_level
    }


    # conf_int_style
    if (length(conf_int_style) > 1L) {
      conf_int_style <- head(conf_int_style, n = 1L)
    }

    # Store plots to list in case no dir_path is provided
    if (is.null(dir_path)) plot_list <- list()

    # Add default splitting variables.
    if (is.null(split_by) &&
        is.null(color_by) &&
        is.null(linetype_by) &&
        is.null(facet_by)) {
      split_by <- c("vimp_method", "learner", "stratification_method")
      color_by <- c("group")
      facet_by <- c("data_set")
    }

    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      color_by = color_by,
      linetype_by = linetype_by,
      facet_by = facet_by,
      available = c(
        "vimp_method", "learner", "data_set", "group", "stratification_method"
      )
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    facet_by <- split_var_list$facet_by
    color_by <- split_var_list$color_by
    linetype_by <- split_var_list$linetype_by

    # Create a legend label
    legend_label <- .create_plot_legend_title(
      user_label = legend_label,
      color_by = color_by,
      linetype_by = linetype_by,
      combine_legend = combine_legend
    )

    # Set show_logrank to FALSE in case color_by and linetype_by
    # contain more than group at most.
    if (!is.null(union(color_by, linetype_by))) {
      if (!all(union(color_by, linetype_by) == "group")) {
        show_logrank <- FALSE
      }
    }

    # Check input arguments for validity.
    .check_input_plot_args(
      x_range = x_range,
      y_range = y_range,
      x_breaks = x_breaks,
      y_breaks = y_breaks,
      conf_int_alpha = conf_int_alpha,
      conf_int_style = conf_int_style,
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption
    )

    # Create plots -------------------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data and supporting data.
    if (!is.null(split_by)) {
      data_split <- split(
        unique(x@data[, mget(split_by)]),
        by = split_by,
        drop = TRUE
      )
      
    } else {
      data_split <- list(NULL)
    }

    # Store plots in a list in case no dir_path is provided
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over splits
    for (current_split in data_split) {
      # Generate split.
      if (!is.null(current_split)) {
        x_split <- methods::new(
          "predictionTableRiskGroups",
          x,
          data = x@data[current_split, on = .NATURAL]
        )
        
      } else {
        x_split <- x
      }

      if ("stratification_method" %in% split_by) {
        # Check that only relevant risk groups are present.
        x_split@data$group <- droplevels(x_split@data$group)
      }

      if (is_empty(x_split)) next
      if (!any_predictions_valid(x_split)) next

      if (is.waive(plot_title)) {
        plot_title <- "Kaplan-Meier survival curve"
      }

      # Declare subtitle components.
      additional_subtitle <- NULL

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = current_split
        )
      }

      # Create plot
      p <- .plot_kaplan_meier(
        x = x_split,
        color_by = color_by,
        linetype_by = linetype_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        combine_legend = combine_legend,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        x_label = x_label,
        x_label_shared = x_label_shared,
        y_label = y_label,
        y_label_shared = y_label_shared,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        x_range = x_range,
        x_breaks = x_breaks,
        y_range = y_range,
        y_breaks = y_breaks,
        conf_int_style = conf_int_style,
        conf_int_alpha = conf_int_alpha,
        censoring = censoring,
        censor_shape = censor_shape,
        show_logrank = show_logrank,
        show_survival_table = show_survival_table
      )

      # Check empty output
      if (is.null(p)) next

      # Draw plot
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Obtain default plot dimensions
        def_plot_dims <- .determine_km_plot_default_dimensions(
          x = x_split@data,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols,
          show_survival_table = show_survival_table
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "stratification",
              "x" = current_split,
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_kaplan_meier <- function(
    x,
    color_by,
    linetype_by,
    facet_by,
    facet_wrap_cols,
    combine_legend,
    ggtheme,
    discrete_palette,
    x_label,
    x_label_shared,
    y_label,
    y_label_shared,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    y_range,
    y_breaks,
    conf_int_style,
    conf_int_alpha,
    censoring,
    censor_shape,
    show_logrank,
    show_survival_table
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  .NATURAL <- NULL

  # Split by facet. This generates a list of data splits with faceting
  # information that allows for positioning.
  plot_layout_table <- .get_plot_layout_table(
    x = x@data,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Define the split in data required for faceting.
  layout_split <- split(
    plot_layout_table,
    by = c("col_id", "row_id"),
    sorted = TRUE
  )
  
  # Create plots to join
  figure_list <- list()
  extracted_element_list <- list()
  for (current_split in layout_split) {
    # Generate the split in case there is a faceting variable.
    if (!is.null(facet_by)) {
      x_split <- methods::new(
        "predictionTableRiskGroups",
        x,
        data = x@data[current_split, on = .NATURAL]
      )
      
    } else {
      x_split <- x
    }
    
    # Skip if the split doesn't contain any information.
    if (is_empty(x_split@data) || all(is.na(x_split@data$outcome_time))) next
    
    # Compute strata for the current split.
    strata <- .compute_risk_stratification_curves(
      x = x_split,
      time_range = x_range
    )

    # Compute logrank test value.
    if (show_logrank) {
      test_data <- .compute_risk_stratification_tests(
        x = x_split,
        time_range = x_range
      )$logrank
    } else {
      test_data <- NULL
    }

    # Kaplan-Meier plots
    p_kaplan_meier <- .create_km_subplot(
      x = strata,
      h = test_data,
      color_by = color_by,
      linetype_by = linetype_by,
      facet_by = facet_by,
      facet_wrap_cols = facet_wrap_cols,
      combine_legend = combine_legend,
      ggtheme = ggtheme,
      discrete_palette = discrete_palette,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      x_range = x_range,
      x_breaks = x_breaks,
      y_range = y_range,
      y_breaks = y_breaks,
      conf_int_style = conf_int_style,
      conf_int_alpha = conf_int_alpha,
      censoring = censoring,
      censor_shape = censor_shape,
      show_logrank = show_logrank
    )

    # Rename plot elements and convert to gtable.
    g_kaplan_meier <- .rename_plot_grobs(
      g = .convert_to_grob(p_kaplan_meier),
      extension = "main"
    )
    if (!gtable::is.gtable(g_kaplan_meier)) next

    if (show_survival_table) {
      # Survival tables
      p_survival_table <- .create_survival_table_subplot(
        x = strata,
        color_by = color_by,
        linetype_by = linetype_by,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        x_range = x_range,
        x_breaks = x_breaks
      )
      
      # Convert to gtable
      g_survival_table <- .rename_plot_grobs(
        g = .convert_to_grob(p_survival_table),
        extension = "surv"
      )
        
      # Insert panel-surv below xlab-b-main.
      g_kaplan_meier <- .gtable_insert(
        g = g_kaplan_meier,
        g_new = .gtable_extract_grob(g_survival_table, element = "panel-surv"),
        where = c("below", "xlab-b-main"),
        grob_name = "panel-surv",
        spacer = .get_plot_panel_spacing(ggtheme = ggtheme, axis = "y")
      )
      
      # Insert axis-l-surv at the intersection of panel-surv and axis-l-main
      g_kaplan_meier <- .gtable_insert(
        g = g_kaplan_meier,
        g_new = .gtable_extract_grob(g_survival_table, element = "axis-l-surv"),
        grob_name = "axis-l-surv",
        where = c("intersect", "below", "axis-l-main", "left", "panel-surv")
      )
    }

    # Attach to figure list.
    figure_list[[paste0(current_split$row_id, ".", current_split$col_id)]] <- as_familiar_plot(
      g = g_kaplan_meier,
      layout = current_split
    )
  }
  
  # Compose the final figure. Magic.
  g <- .compose_figure(
    figure_list = figure_list,
    plot_layout_table = plot_layout_table,
    x_text_shared = x_label_shared,
    x_label_shared = x_label_shared,
    y_text_shared = y_label_shared,
    y_label_shared = y_label_shared,
    facet_wrap_cols = facet_wrap_cols,
    ggtheme = ggtheme
  )

  return(g)
}



.create_km_subplot <- function(
    x,
    h,
    color_by,
    linetype_by,
    facet_by,
    facet_wrap_cols,
    combine_legend,
    ggtheme,
    discrete_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    y_range,
    y_breaks,
    conf_int_style,
    conf_int_alpha,
    censoring,
    censor_shape,
    show_logrank
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  n_censor <- group_1 <- NULL
  
  if (!is_empty(x)) x <- x@data
  if (!is_empty(h)) h <- h@data

  # Generate a guide table to allow integration of guides into a single legend
  guide_list <- .create_plot_guide_table(
    x = x,
    color_by = color_by,
    linetype_by = linetype_by,
    discrete_palette = discrete_palette,
    combine_legend = combine_legend
  )
  
  # Extract data
  x <- guide_list$data

  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("time"),
      y = !!sym("survival")
    )
  )
  
  p <- p + ggtheme

  # Create step function
  if (is.null(color_by) && is.null(linetype_by)) {
    p <- p + ggplot2::geom_step()
    
  } else if (!is.null(color_by) && is.null(linetype_by)) {
    p <- p + ggplot2::geom_step(
      mapping = ggplot2::aes(
        colour = !!sym("color_breaks")
      ),
      show.legend = TRUE
    )
    
  } else if (is.null(color_by) && !is.null(linetype_by)) {
    p <- p + ggplot2::geom_step(
      mapping = ggplot2::aes(
        linetype = !!sym("linetype_breaks")
      ),
      show.legend = TRUE
    )
    
  } else {
    p <- p + ggplot2::geom_step(
      mapping = ggplot2::aes(
        colour = !!sym("color_breaks"),
        linetype = !!sym("linetype_breaks")
      ),
      show.legend = TRUE
    )
  }

  # Set colour
  if (!is.null(color_by)) {
    # Extract guide_table for color
    g_color <- guide_list$guide_color

    p <- p + ggplot2::scale_colour_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )

    p <- p + ggplot2::scale_fill_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )
  }

  # Set line style
  if (!is.null(linetype_by)) {
    # Extract guide_table for linetype
    g_linetype <- guide_list$guide_linetype

    p <- p + ggplot2::scale_linetype_manual(
      name = legend_label$guide_linetype,
      values = g_linetype$linetype_values,
      breaks = g_linetype$linetype_breaks,
      drop = FALSE
    )
  }

  if (show_logrank && !is_empty(h)) {
    # Parse p-value
    p_value_label <- paste0(
      "p: ", as.character(signif(h[group_1 == "all"]$p_value, 2L))
    )

    # Obtain default settings.
    text_settings <- .get_plot_geom_text_settings(ggtheme = ggtheme)

    # Show in plot
    p <- p + ggplot2::annotate(
      "text",
      x = x_range[1L],
      y = y_range[1L],
      label = p_value_label,
      colour = text_settings$colour,
      family = text_settings$family,
      fontface = text_settings$face,
      size = text_settings$geom_text_size,
      vjust = "inward",
      hjust = "inward"
    )
  }
  
  # Plot confidence intervals
  if (conf_int_style[1L] != "none" && !is_empty(x)) {
    if (conf_int_style[1L] == "step") {
      if (is.null(color_by)) {
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_low")
          ),
          linetype = "dashed",
          na.rm = TRUE
        )

        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_up")
          ),
          linetype = "dashed",
          na.rm = TRUE
        )
        
      } else {
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_low"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed",
          na.rm = TRUE
        )

        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_up"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed",
          na.rm = TRUE
        )
      }
      
    } else if (conf_int_style[1L] == "ribbon") {
      if (is.null(color_by)) {
        # Create special data for ribbon so that it becomes a step ribbon.
        x_ribbon <- .prepare_km_conf_int_plot_data(x = x)

        p <- p + ggplot2::geom_ribbon(
          data = x_ribbon,
          mapping = ggplot2::aes(
            x = !!sym("time"),
            ymin = !!sym("ci_low"),
            ymax = !!sym("ci_up")
          ),
          alpha = conf_int_alpha,
          na.rm = TRUE
        )
        
      } else {
        # Create special data for ribbon so that it becomes a step ribbon.
        x_ribbon <- data.table::rbindlist(lapply(
          split(x, by = "color_breaks"),
          .prepare_km_conf_int_plot_data
        ))
        
        p <- p + ggplot2::geom_ribbon(
          data = x_ribbon,
          mapping = ggplot2::aes(
            x = !!sym("time"),
            ymin = !!sym("ci_low"),
            ymax = !!sym("ci_up"),
            fill = !!sym("color_breaks")
          ),
          alpha = conf_int_alpha,
          na.rm = TRUE,
          show.legend = TRUE
        )
      }
    }
  }

  # Censoring indicators
  if (censoring) {
    if (is.null(color_by)) {
      p <- p + ggplot2::geom_point(
        data = x[n_censor > 0L],
        shape = censor_shape,
        show.legend = FALSE
      )
      
    } else {
      p <- p + ggplot2::geom_point(
        data = x[n_censor > 0L],
        mapping = ggplot2::aes(
          colour = !!sym("color_breaks"),
          fill = !!sym("color_breaks")
        ),
        shape = censor_shape,
        show.legend = FALSE
      )
    }
  }

  # Update x and y scales
  p <- p + ggplot2::scale_x_continuous(breaks = x_breaks)
  p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)

  # Labels
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Determine how things are facetted.
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context",
        drop = TRUE
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context",
        drop = TRUE
      )
    }
  }

  # Plot to Cartesian coordinates.
  p <- p + ggplot2::coord_cartesian(xlim = x_range, ylim = y_range)

  return(p)
}



.create_survival_table_subplot <- function(
    x,
    color_by,
    linetype_by,
    ggtheme,
    discrete_palette,
    x_range,
    x_breaks
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  id <- missing_entry <- NULL

  # Find the names for the groups
  if (is.null(c(color_by, linetype_by))) {
    # Parse the table to obtain group sizes at x_breaks
    survival_table <- ..get_survival_breakpoints(
      x@data,
      x_breaks = x_breaks
    )

    survival_table[, ":="("group_name" = "")]
    
  } else {
    # Parse the table to obtain group sizes at x_breaks
    survival_list <- lapply(
      split(x@data, by = c(color_by, linetype_by)),
      ..get_survival_breakpoints,
      x_breaks = x_breaks
    )

    # Combine into list
    survival_table <- data.table::rbindlist(
      survival_list, 
      use.names = TRUE
    )

    unique_vars <- unique(c(color_by, linetype_by))

    # Generate a guide table
    guide_table <- data.table::data.table(expand.grid(lapply(
      rev(unique_vars),
      function(ii, x) (levels(x[[ii]])),
      x = survival_table
    )))

    # Rename variables
    data.table::setnames(
      x = guide_table,
      rev(unique_vars)
    )

    # Convert to factors
    for (ii in unique_vars) {
      guide_table[[ii]] <- factor(
        x = guide_table[[ii]],
        levels = levels(x@data[[ii]])
      )
    }

    # Order columns according to unique_vars
    data.table::setcolorder(
      x = guide_table,
      neworder = unique_vars
    )

    # Order data set by columns
    data.table::setorderv(
      x = guide_table,
      cols = unique_vars
    )

    # Set breaks
    breaks <- apply(guide_table, 1L, paste, collapse = ", ")

    # Update guide_table
    guide_table$group_name <- factor(breaks, levels = breaks)

    # Identify missing entries in the guide table. First add an identifier to
    # the guide-table, then determine which entries are empty.
    guide_table[, "id" := .I]
    guide_table$missing_entry <- sapply(
      split(guide_table, by = "id"),
      function(x, y, by) {
        return(is_empty(merge(
          x = x,
          y = y,
          by = by,
          all = FALSE
        )))
      },
      y = survival_table,
      by = unique_vars
    )

    if (any(guide_table$missing_entry)) {
      # Find a suitable prototype and then create a prototype table.
      available_prototype <- head(guide_table[missing_entry == FALSE]$id, 1L)

      proto_table <- merge(
        x = survival_table,
        y = guide_table[id == available_prototype, mget(unique_vars)]
      )

      # Iterate over missing entries.
      new_survival_list <- lapply(
        split(guide_table[missing_entry == TRUE], by = "id"),
        function(x, proto_table, unique_vars) {
          # Make a copy of the prototype and replace the group size by 0.
          y <- data.table::copy(proto_table)[, "group_size" := 0L]
          
          # Insert the correct values for each plotting variable that was
          # previously missing.
          for (var in unique_vars) {
            y[, (var) := x[[var]]]
          }

          return(y)
        },
        proto_table = proto_table,
        unique_vars = unique_vars
      )

      # Add the placeholder entries to the table.
      survival_table <- data.table::rbindlist(
        c(list(survival_table), new_survival_list),
        use.names = TRUE
      )

      # Remove superfluous columns from guide_table
      guide_table[, ":="(
        "id" = NULL,
        "missing_entry" = NULL
      )]
    }

    # Combine the guide table with the survival table to add in the group names.
    survival_table <- merge(
      x = survival_table,
      y = guide_table,
      by = unique_vars
    )
  }

  # Obtain default settings.
  text_settings <- .get_plot_geom_text_settings(ggtheme = ggtheme)
  fontsize <- text_settings$fontsize
  fontsize_rel <- text_settings$fontsize_rel
  lineheight <- text_settings$lineheight

  # Create plot
  p <- ggplot2::ggplot(
    data = survival_table,
    mapping = ggplot2::aes(
      x = !!sym("time"),
      y = !!sym("group_name"),
      label = !!sym("group_size")
    )
  )

  # Annotate survival in strata.
  p <- p + ggplot2::geom_text(
    colour = text_settings$colour,
    family = text_settings$family,
    fontface = text_settings$face,
    size = text_settings$geom_text_size
  )

  # Adapt axes.
  p <- p + ggplot2::scale_x_continuous(
    breaks = x_breaks,
    limits = x_range
  )

  p <- p + ggplot2::scale_y_discrete(
    breaks = rev(levels(survival_table$group_name)),
    limits = rev(levels(survival_table$group_name))
  )

  # Set main theme
  p <- p + ggtheme

  # Remove some theme elements and reduce margins
  p <- p + ggplot2::theme(
    panel.grid = ggplot2::element_blank(),
    panel.border = ggplot2::element_blank(),
    axis.line = ggplot2::element_blank(),
    axis.ticks.y = ggplot2::element_blank(),
    axis.text.x = ggplot2::element_blank(),
    axis.ticks.x = ggplot2::element_blank(),
    axis.title.x = ggplot2::element_blank(),
    axis.title.y = ggplot2::element_blank()
  )

  # Convert to gtable
  g <- ggplot2::ggplotGrob(p)

  # Set the height on the panel element to n * fontsize + (n-1) * lineheight
  n_lines <- data.table::uniqueN(survival_table$group_name)
  grob_height <- n_lines * fontsize * fontsize_rel + 
    (n_lines - 1L) * fontsize * fontsize_rel * lineheight + 4.0

  # Find the panel
  panel_row_t <- g$layout[g$layout$name == "panel", "t"]

  # Set heights of the row
  g$heights[panel_row_t] <- grid::unit(x = grob_height, "pt")

  # Set to p
  p$custom_grob <- list("heights" = list(
    "name" = "panel",
    "height" = grid::unit(grob_height, "pt")
  ))

  class(p) <- c("familiar_ggplot", class(p))

  return(p)
}



..get_survival_breakpoints <- function(x, x_breaks) {
  # Suppress NOTES due to non-standard evaluation in data.table
  time <- survival <- NULL

  # Iterate over the breaks to obtain the group size at this time point.
  survival_list <- lapply(
    x_breaks,
    function(break_time, x) {
      # Find the nearest candidate entry
      candidate_entry <- data.table::copy(tail(x[time <= break_time, ], n = 1L))

      # In case the number of survivors is 0, the group size is also 0.
      candidate_entry[survival == 0L, "group_size" := 0L]

      return(candidate_entry)
    },
    x = x
  )

  # Convert to a new data.table
  survival_table <- data.table::rbindlist(
    survival_list,
    use.names = TRUE
  )

  # Update time to x_breaks
  survival_table[, "time" := x_breaks]

  return(survival_table)
}



.determine_km_plot_default_dimensions <- function(
    x,
    facet_by,
    facet_wrap_cols,
    show_survival_table
) {
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- ifelse(show_survival_table, 6.0, 4.0)

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))

  return(c(height, width))
}



.prepare_km_conf_int_plot_data <- function(x) {
  # Suppress NOTES due to non-standard evaluation in data.table
  time <- ci_low <- ci_up <- NULL

  if (is_empty(x)) return(x)

  # Make sure that the surv_lower and surv_upper coordinate sets are correct.
  x <- data.table::copy(x)[order(time)]
  y <- data.table::copy(x)[1L:(nrow(x) - 1L)]
  y[, "time" := x$time[2L:nrow(x)]]

  # Combine and order correctly.
  x <- rbind(x, y)[order(time, -ci_low, -ci_up)]

  return(x)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarDataElementModelPerformance object -----------------------------------
setClass(
  "familiarDataElementModelPerformance",
  contains = "familiarDataElement",
  prototype = methods::prototype(value_column = "value")
)

# extract_performance (generic) ------------------------------------------------

#'@title Internal function to extract performance metrics.
#'
#'@description Computes and collects discriminative performance metrics from a
#'  `familiarEnsemble`.
#'
#'@inheritParams .extract_data
#'
#'@details This method computes credibility intervals for the ensemble model, at
#'  the level of `confidence_level`. This is a general method. Metrics with
#'  known, theoretically derived confidence intervals, nevertheless have a
#'  credibility interval computed.
#'
#'@return A list with data.tables for single and ensemble model assessments.
#'@md
#'@keywords internal
setGeneric(
  "extract_performance",
  function(
    object,
    data,
    cl = NULL,
    metric = waiver(),
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_performance")
  }
)

# extract_performance (familiarEnsemble) ---------------------------------------
setMethod(
  "extract_performance",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    metric = waiver(),
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Computing model performance metrics on the dataset."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (is.waive(evaluation_times)) evaluation_times <- object@settings$eval_times
    if (is.waive(ensemble_method)) ensemble_method <- object@settings$ensemble_method
    if (is.waive(confidence_level)) confidence_level <- object@settings$confidence_level
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- object@settings$bootstrap_ci_method
    if (is.waive(metric)) metric <- object@settings$metric
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = FALSE,
      data_element = "model_performance"
    )
    
    proto_data_element <- .create_extract_model_performance_object(
      object = object,
      metric = metric,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    # Generate elements to send to dispatch.
    performance_data <- extract_dispatcher(
      FUN = .extract_model_performance_data,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      metric = metric,
      evaluation_times = evaluation_times,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(performance_data)
  }
)


# extract_performance (prediction table) ---------------------------------------
setMethod(
  "extract_performance",
  signature(object = "familiarDataElementPredictionTable"),
  function(
    object,
    data,
    cl = NULL,
    metric = waiver(),
    ensemble_method = waiver(),
    evaluation_times = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    
    if (is_empty(object)) return(NULL)
    
    # Message extraction start
    logger_message(
      paste0("Computing model performance metrics from the prediction table."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Reference labels should be present.
    if (!.has_reference_data(object)) {
      ..warning_prediction_table_lacks_reference("model performance metrics")
      return(NULL)
    }
    
    # Load evaluation_times and ensemble method from the prediction table.
    if (is.waive(evaluation_times) && methods::.hasSlot(object, "time")) {
      evaluation_times <- object@time
    }
    if (is.waive(ensemble_method)) {
      ensemble_method <- "median"
      if (methods::.hasSlot(object, "ensemble_method")) ensemble_method <- object@ensemble_method
    } 
    
    # Set default metric
    if (is.waive(metric)) metric <- .get_default_metric(outcome_type = object@outcome_type)

    # Default Values.
    if (is.waive(detail_level)) detail_level <- "ensemble"
    if (is.waive(estimation_type)) estimation_type <- "bootstrap_confidence_interval" 
    if (is.waive(confidence_level)) confidence_level <- 0.95
    if (is.waive(bootstrap_ci_method)) bootstrap_ci_method <- "bc"
    if (is.waive(aggregate_results)) aggregate_results <- FALSE
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = FALSE,
      data_element = "model_performance"
    )
    
    # Copy object to prevent changing the provided object by reference.
    object <- .copy(object)
    
    proto_data_element <- .create_extract_model_performance_object(
      object = object,
      metric = metric,
      ensemble_method = ensemble_method,
      evaluation_times = evaluation_times,
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    performance_data <- extract_dispatcher(
      FUN = .extract_model_performance_data,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      metric = metric,
      evaluation_times = evaluation_times,
      aggregate_results = aggregate_results,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(performance_data)
  }
)



.create_extract_model_performance_object <- function(
    object,
    metric,
    ensemble_method,
    evaluation_times,
    detail_level,
    estimation_type,
    confidence_level,
    bootstrap_ci_method
) {

  # Check evaluation_times argument
  if (object@outcome_type %in% c("survival")) {
    sapply(
      evaluation_times, 
      .check_number_in_valid_range, 
      var_name = "evaluation_times",
      range = c(0.0, Inf),
      closed = c(FALSE, TRUE)
    )
  }
  
  # Check confidence_level input argument
  .check_number_in_valid_range(
    x = confidence_level, 
    var_name = "confidence_level",
    range = c(0.0, 1.0),
    closed = c(FALSE, FALSE)
  )
  
  # Check ensemble_method argument
  .check_parameter_value_is_valid(
    x = ensemble_method,
    var_name = "ensemble_method",
    values = .get_available_ensemble_prediction_methods()
  )
  
  # Check bootstrap_ci_method argument
  .check_parameter_value_is_valid(
    x = bootstrap_ci_method,
    var_name = "bootstrap_ci_method",
    values = .get_available_bootstrap_confidence_interval_methods()
  )
  
  # Check the level detail.
  detail_level <- .parse_detail_level(
    x = detail_level,
    object = object,
    default = "hybrid",
    data_element = "model_performance"
  )
  
  # Check the estimation type.
  estimation_type <- .parse_estimation_type(
    x = estimation_type,
    object = object,
    default = "bootstrap_confidence_interval",
    data_element = "model_performance",
    detail_level = detail_level,
    has_internal_bootstrap = TRUE
  )

  # Check metric input argument
  sapply(
    metric,
    .check_metric_outcome_type,
    object = object
  )

  # Generate a prototype data element.
  proto_data_element <- new(
    "familiarDataElementModelPerformance",
    detail_level = detail_level,
    estimation_type = estimation_type,
    confidence_level = confidence_level,
    bootstrap_ci_method = bootstrap_ci_method
  )
  
  return(proto_data_element)
}



.extract_model_performance_data <- function(
    object,
    proto_data_element,
    evaluation_times = NULL,
    aggregate_results,
    cl,
    ...) {
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  proto_data_element <- add_model_name(proto_data_element, object = object)
  
  # Add evaluation time as a identifier to the data element.
  if (length(evaluation_times) > 0L && object@outcome_type == "survival") {
    data_elements <- add_data_element_identifier(
      x = proto_data_element,
      evaluation_time = evaluation_times
    )
    
  } else {
    data_elements <- list(proto_data_element)
  }
  
  # Iterate over data elements.
  performance_data <- lapply(
    data_elements,
    ..extract_model_performance_data,
    object = object,
    aggregate_results = aggregate_results,
    cl = cl,
    ...
  )
  
  return(performance_data)
}


# ..extract_model_performance_data (generic) -----------------------------------
setGeneric(
  "..extract_model_performance_data",
  function(object, ...) standardGeneric("..extract_model_performance_data")
)



# ..extract_model_performance_data (character) ---------------------------------
setMethod(
  "..extract_model_performance_data",
  signature(object = "character"),
  function(
    object,
    ...
  ){
    # Ensure that the object is loaded
    object <- load_familiar_object(object)
    
    return(..extract_model_performance_data(object = object, ...))
  }
)



# ..extract_model_performance_data (model, ensemble) ---------------------------
setMethod(
  "..extract_model_performance_data",
  signature(object = "familiarModelUnion"),
  function(
    object,
    data_element,
    data,
    is_pre_processed,
    ensemble_method,
    progress_bar = FALSE,
    verbose = FALSE,
    message_indent,
    ...
  ) {
    # Ensure that the object is loaded
    object <- load_familiar_object(object)
    
    # Message the user concerning the time at which metrics are computed. This is
    # only relevant for survival analysis.
    if (length(data_element@identifiers$evaluation_time) > 0L && progress_bar) {
      logger_message(
        paste0(
          "Computing metric value at time ",
          data_element@identifiers$evaluation_time, "."
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    # Predict class probabilities.
    prediction_data <- .predict(
      object = object,
      data = data,
      time = data_element@identifiers$evaluation_time,
      ensemble_method = ensemble_method,
      is_pre_processed = is_pre_processed
    )
    
    # Capture empty or invalid prediction tables.
    if (!all_predictions_valid(prediction_data)) return(NULL)
    
    return(..extract_model_performance_data(
      object = prediction_data,
      data_element = data_element,
      ...
    ))
  }
)



# ..extract_model_performance_data (prediction_table) --------------------------
setMethod(
  "..extract_model_performance_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(
    object,
    data_element,
    data = NULL,
    cl = NULL,
    metric,
    aggregate_results,
    verbose = FALSE,
    progress_bar = FALSE,
    ...
  ) {
    # Check if that all predictions are valid.
    if (!all_predictions_valid(object)) return(NULL)
    
    # Remove data with missing outcomes.
    object <- filter_missing_outcome(data = object)
    
    # Check that any prediction data remain.
    if (is_empty(object)) return(NULL)
    
    # Add metric as an identifier.
    data_elements <- add_data_element_identifier(
      x = data_element,
      metric = metric
    )
    
    # Add bootstrap data.
    bootstrap_data <- add_data_element_bootstrap(x = data_elements, ...)
    
    # Iterate over elements.
    data_elements <- fam_mapply(
      cl = cl,
      assign = NULL,
      FUN = .compute_model_performance,
      data_element = bootstrap_data$data_element,
      bootstrap = bootstrap_data$bootstrap,
      bootstrap_seed = bootstrap_data$seed,
      MoreArgs = list("data" = object),
      progress_bar = progress_bar && verbose,
      chopchop = TRUE
    )
    
    # Merge data elements
    data_elements <- merge_data_elements(data_elements)
    
    if (aggregate_results) {
      data_elements <- .compute_data_element_estimates(x = data_elements)
    }

    return(data_elements)
  }
)



# ..extract_model_performance_data (NULL) --------------------------------------
setMethod(
  "..extract_model_performance_data",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



.compute_model_performance <- function(
    data_element,
    data, 
    bootstrap, 
    bootstrap_seed
) {
  
  # Bootstrap the data.
  if (bootstrap) {
    data <- get_bootstrap_sample(
      data = data,
      seed = bootstrap_seed
    )
  }
  
  # Compute the score
  score <- compute_metric_score(
    metric = data_element@identifiers$metric,
    data = data,
    time = data_element@identifiers$evaluation_time
  )
  
  # Set the value, if finite.
  if (is.finite(score)) {
    data_element@data <- list("value" = unname(score))
  }
  
  return(data_element)
}


# export_model_performance (generic) -------------------------------------------

#'@title Extract and export metrics for model performance.
#'
#'@description Extract and export metrics for model performance of models in a
#'  familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams extract_performance
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Performance of individual and ensemble models is exported. For ensemble
#'  models, a credibility interval is determined using bootstrapping for each
#'  metric.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_model_performance
#'@md
#'@rdname export_model_performance-methods
setGeneric(
  "export_model_performance",
  function(
    object,
    dir_path = NULL,
    aggregate_results = FALSE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_model_performance")
  }
)



# export_model_performance (collection) ----------------------------------------

#'@rdname export_model_performance-methods
setMethod(
  "export_model_performance",
  signature(object = "familiarCollection"),
  function(
    object, 
    dir_path = NULL,
    aggregate_results = FALSE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(.export(
      x = object,
      data_slot = "model_performance",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "performance",
      subtype = "metric",
      export_collection = export_collection
    ))
  }
)



# export_model_performance (general) -------------------------------------------

#'@rdname export_model_performance-methods
setMethod(
  "export_model_performance",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = FALSE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "model_performance",
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_model_performance,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

familiar:::test_plots(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  debug = debug_flag
)

# With the ensemble detail level.
familiar:::test_plots(
  plot_function = familiar:::plot_calibration_data,
  detail_level = "ensemble",
  data_element = "calibration_data",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  debug = debug_flag
)

# For bias-corrected data
familiar:::test_plots(
  plot_function = familiar:::plot_calibration_data,
  estimation_type = "bias_correction",
  data_element = "calibration_data",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  debug = debug_flag
)

# For point estimates
familiar:::test_plots(
  plot_function = familiar:::plot_calibration_data,
  estimation_type = "point",
  data_element = "calibration_data",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  debug = debug_flag
)

# Test without density plot
familiar:::test_plots(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  test_config = "normal",
  plot_args = list("show_density" = FALSE),
  debug = debug_flag
)

# Test without the goodness of fit data
familiar:::test_plots(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  test_config = "normal",
  plot_args = list("show_goodness_of_fit" = FALSE),
  debug = debug_flag
)

# Test without the calibration fit data
familiar:::test_plots(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  test_config = "normal",
  plot_args = list("show_calibration_fit" = FALSE),
  debug = debug_flag
)

# Test without either fit data
familiar:::test_plots(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  test_config = "normal",
  plot_args = list(
    "show_calibration_fit" = FALSE,
    "show_goodness_of_fit" = FALSE),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  debug = debug_flag
)

# Test alignment of different plots from prediction tables.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  prediction_type = list("survival" = "survival_probability"),
  use_prediction_table = TRUE,
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  outcome_type_available = c("continuous"),
  plot_args = list(
    "facet_by" = c("vimp_method", "learner"),
    "color_by" = c("data_set")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  outcome_type_available = c("binomial"),
  plot_args = list(
    "facet_by" = c("vimp_method", "learner"),
    "color_by" = c("data_set")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  outcome_type_available = c("multinomial"),
  plot_args = list(
    "facet_by" = c("vimp_method", "learner"),
    "color_by" = c("data_set", "positive_class")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  outcome_type_available = c("survival"),
  plot_args = list(
    "facet_by" = c("vimp_method", "learner"),
    "color_by" = c("data_set")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  outcome_type_available = c("continuous", "binomial", "survival"),
  plot_args = list("facet_by" = c("learner", "vimp_method", "data_set")),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_calibration_data,
  data_element = "calibration_data",
  outcome_type_available = c("multinomial"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "color_by" = "positive_class"),
  debug = debug_flag
)

# Test if detectors are available.
familiar:::test_all_novelty_detectors_available(
  detectors = familiar:::.get_available_none_detectors())

# Don't perform any further tests on CRAN due to time of running the test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_all_novelty_detectors(
  detectors = familiar:::.get_available_none_detectors(),
  except_train = familiar:::.get_available_none_detectors(),
  can_trim = FALSE)

familiar:::test_all_novelty_detectors_parallel(
  detectors = familiar:::.get_available_none_detectors(),
  except_train = familiar:::.get_available_none_detectors())

.start_cluster <- function(n_cores = NULL, cluster_type = NULL) {
  # Start a cluster of workers

  # Load from familiar global environment if unset.
  if (is.null(n_cores)) n_cores <- .get_desired_n_cores()
  if (is.null(cluster_type)) cluster_type <- .get_selected_cluster_type()

  if (cluster_type == "none") return(NULL)

  # Check the number of available connections
  n_available_connections <- .available_connections()

  # Find the number of available cores.
  n_available_cores <- parallel::detectCores()
  n_cores <- max(c(1L, min(c(n_available_cores - 1L, n_cores))))

  if (n_available_connections < 2L && n_cores > 2L && cluster_type %in% c("psock", "sock", "fork")) {
    ..error(paste0(
      "R has insufficient available connections to perform parallel processing. ",
      "You may close connections using closeAllConnections() to free up connections. ",
      "If this does not resolve the issue, disable parallelisation for familiar."
    ))
    
  } else if (n_cores > 2L && n_available_connections < n_cores) {
    ..warning(paste0(
      "R has insufficient available connections to use all available cores for parallel processing. ",
      n_available_connections, " are used instead."
    ))

    # Update the number of cores.
    n_cores <- n_available_connections
  }

  # Return NULL if the number of cores is 1 or less.
  if (n_cores <= 1L) return(NULL)

  # Start cluster.
  if (cluster_type == "psock") {
    cl <- parallel::makePSOCKcluster(n_cores)
  } else if (cluster_type == "fork") {
    cl <- parallel::makeForkCluster(n_cores)
  } else if (cluster_type %in% c("mpi", "nws", "sock")) {
    cl <- parallel::makeCluster(type = toupper(cluster_type), 2L)
  }

  return(cl)
}



.terminate_cluster <- function(cl) {
  # Stops a cluster and releases workers

  # Do not stop a cluster if it is external.
  if (.is_external_cluster()) return(cl)

  if (inherits(cl, "cluster")) parallel::stopCluster(cl)

  return(NULL)
}



.available_connections <- function() {
  # TODO Add check for R version in case the number of available connections
  # changes.

  # The standard is that R allows 128 connections. 3 are in standard use by R.
  # Another socket is used by the default socket_server data backend.
  R_available <- 124L

  # Find the number of currently available connections.
  n_available_connections <- R_available - nrow(showConnections())

  if (n_available_connections < 0L) n_available_connections <- 0L

  return(n_available_connections)
}



.check_cluster_type_availability <- function(cluster_type) {
  # Check if the provided cluster type is available.

  if (get_os() == "windows" && cluster_type == "fork") {
    ..error("FORK clusters are not available for windows OS.")
  }

  installed_libs <- utils::installed.packages()[, 1L]

  if (!"snow" %in% installed_libs && cluster_type %in% c("mpi", "nws", "sock")) {
    ..error("The parallel package requires the snow package to work with MPI, NWS or SOCK clusters.")
  }

  if (!"Rmpi" %in% installed_libs && cluster_type == "mpi") {
    ..error("The parallel packages requires the Rmpi package to work with MPI clusters.")
  }
  
  return(invisible(TRUE))
}



.assign_parallel_options_to_global <- function(
    is_external_cluster = FALSE,
    restart_cluster = FALSE,
    n_cores = NULL,
    cluster_type = NULL
) {
  # We don't need a backend if the cluster is external.
  if (is.null(cluster_type) && !is_external_cluster) cluster_type <- "psock"

  # Assign parallelisation options to the familiar global environment.
  assign("is_external_cluster", is_external_cluster, envir = familiar_global_env)
  assign("restart_cluster", restart_cluster, envir = familiar_global_env)
  assign("n_cores", n_cores, envir = familiar_global_env)
  assign("cluster_type", cluster_type, envir = familiar_global_env)
  
  return(invisible(NULL))
}



.needs_cluster_restart <- function() {
  if (exists("familiar_global_env")) {
    if (exists("restart_cluster", where = familiar_global_env)) {
      data_env <- familiar_global_env
    } else if (exists("restart_cluster", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
    } else {
      return(FALSE)
    }
  } else if (exists("restart_cluster", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
  } else {
    return(FALSE)
  }

  return(get("restart_cluster", envir = data_env))
}



.is_external_cluster <- function() {
  if (exists("familiar_global_env")) {
    if (exists("is_external_cluster", where = familiar_global_env)) {
      data_env <- familiar_global_env
    } else if (exists("is_external_cluster", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
    } else {
      return(TRUE)
    }
  } else if (exists("is_external_cluster", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
  } else {
    return(TRUE)
  }

  return(get("is_external_cluster", envir = data_env))
}



.get_desired_n_cores <- function() {
  return(get("n_cores", envir = familiar_global_env))
}



.get_selected_cluster_type <- function() {
  return(get("cluster_type", envir = familiar_global_env))
}



.restart_cluster <- function(cl = NULL, assign = NULL, n_nodes = NULL) {
  # Determine the number of nodes that we need to assign.
  if (inherits(cl, "cluster") && is.null(n_nodes)) n_nodes <- length(cl)

  # Terminate old cluster (if necessary)
  cl <- .terminate_cluster(cl = cl)

  # Start a new cluster
  cl <- .start_cluster(n_cores = n_nodes)

  # If the cluster doesn't start, return a NULL
  if (is.null(cl)) return(NULL)

  # Ensure that all required data are present on the cluster.
  cl <- .update_cluster(cl = cl, assign = assign)

  return(cl)
}



.update_cluster <- function(cl = NULL, assign = NULL) {
  # Set library paths to avoid issues with non-standard library locations.
  libs <- .libPaths()
  parallel::clusterExport(cl = cl, varlist = "libs", envir = environment())
  parallel::clusterEvalQ(cl = cl, .libPaths(libs))

  # Load familiar and data.table libraries to each cluster node.
  parallel::clusterEvalQ(cl = cl, library(familiar))
  parallel::clusterEvalQ(cl = cl, library(data.table))

  # Set options on each cluster node.
  parallel::clusterEvalQ(cl = cl, options(rf.cores = 1L))
  parallel::clusterEvalQ(cl = cl, data.table::setDTthreads(1L))

  if ("none" %in% assign) return(cl)
  
  # Check if anything needs to be loaded
  if (any(c("all", "data") %in% assign)) {
    # Only add master data to the global environment of the clusters when
    # running a non-rserve data backend.
    if (.get_selected_backend_type() == "none") {
      parallel::clusterExport(cl = cl, varlist = "master_data", envir = familiar_global_env)
    }
  }

  # Export the feature_info list
  if (any(c("all", "feature_info") %in% assign)) {
    if (.get_selected_backend_type() == "none") {
      if (exists("master_feature_info_list", envir = familiar_global_env)) {
        parallel::clusterExport(cl = cl, varlist = "master_feature_info_list", envir = familiar_global_env)
      }
    }
  }

  # Export the project_info list
  if (any(c("all", "project_info") %in% assign)) {
    parallel::clusterExport(cl = cl, varlist = "project_info_list", envir = familiar_global_env)
  }

  # Export smaller objects directly, so that we don't have to worry about them.
  parallel::clusterExport(cl = cl, varlist = "settings", envir = familiar_global_env)
  parallel::clusterExport(cl = cl, varlist = "file_paths", envir = familiar_global_env)
  parallel::clusterExport(cl = cl, varlist = "outcome_info", envir = familiar_global_env)
  parallel::clusterExport(cl = cl, varlist = "server_port", envir = familiar_global_env)
  parallel::clusterExport(cl = cl, varlist = "backend_type", envir = familiar_global_env)

  return(cl)
}



.update_info_on_cluster <- function(cl, assign, backend_type = NULL) {
  # Not all variables in the familiar global environment are static. For
  # example, the feature_info_list may be updated by pre-processing. This
  # function updates the clusters with the latest version of such variables.

  if (is.null(backend_type)) backend_type <- .get_selected_backend_type()

  # Re-export the feature_info list.
  if (any(c("all", "feature_info") %in% assign)) {
    if (
      backend_type == "none" && 
      exists("master_feature_info_list", envir = familiar_global_env)
    ) {
      parallel::clusterExport(cl = cl, varlist = "master_feature_info_list", envir = familiar_global_env)
    }
  }

  return(cl)
}



fam_sapply <- function(
    cl = NULL,
    assign = NULL,
    X,
    FUN,
    ...,
    progress_bar = FALSE,
    MoreArgs = NULL,
    SIMPLIFY = TRUE,
    USE.NAMES = TRUE,
    chopchop = FALSE,
    overhead_time = NULL,
    process_time = NULL,
    MEASURE.TIME = FALSE
) {
  # Determine the output format.
  output_format <- ifelse(SIMPLIFY, "vector", "list")
  
  # Get the name of the initial argument.
  initial_argument_name <- head(names(formals(FUN)), n = 1L)
  
  # Upgrade X to a named list.
  X <- list(X)
  names(X) <- initial_argument_name
  
  # Make call to .fam_apply.
  y <- do.call(
    .fam_apply,
    args = c(
      list(
        "cl" = cl,
        "assign" = assign,
        "FUN" = FUN,
        "additional_arguments" = list(...),
        "progress_bar" = progress_bar,
        "chopchop" = chopchop,
        "overhead_time" = overhead_time,
        "process_time" = process_time,
        "measure_time" = MEASURE.TIME,
        "process_scheduling" = "static",
        "output_format" = output_format,
        "use_names" = USE.NAMES
      ),
      X
    )
  )
  
  return(y)
}



fam_lapply <- function(
    cl = NULL,
    assign = NULL,
    X,
    FUN,
    ...,
    progress_bar = FALSE,
    MoreArgs = NULL,
    SIMPLIFY = FALSE,
    chopchop = FALSE,
    overhead_time = NULL,
    process_time = NULL,
    MEASURE.TIME = FALSE
) {
  # Determine the output format.
  output_format <- ifelse(SIMPLIFY, "vector", "list")

  # Get the name of the initial argument.
  initial_argument_name <- head(names(formals(FUN)), n = 1L)

  # Upgrade X to a named list.
  X <- list(X)
  names(X) <- initial_argument_name

  # Make call to .fam_apply.
  y <- do.call(
    .fam_apply,
    args = c(
      list(
        "cl" = cl,
        "assign" = assign,
        "FUN" = FUN,
        "additional_arguments" = list(...),
        "progress_bar" = progress_bar,
        "chopchop" = chopchop,
        "overhead_time" = overhead_time,
        "process_time" = process_time,
        "measure_time" = MEASURE.TIME,
        "process_scheduling" = "static",
        "output_format" = output_format
      ),
      X
    )
  )
  
  return(y)
}




fam_mapply <- function(
    cl = NULL,
    assign = NULL,
    FUN,
    ...,
    progress_bar = FALSE,
    MoreArgs = NULL,
    SIMPLIFY = FALSE,
    USE.NAMES = TRUE,
    chopchop = FALSE,
    overhead_time = NULL,
    process_time = NULL,
    MEASURE.TIME = FALSE
) {
  # Determine the output format.
  output_format <- ifelse(SIMPLIFY, "vector", "list")
  
  # Make call to .fam_apply.
  y <- do.call(
    .fam_apply,
    args = c(
      list(
        "cl" = cl,
        "assign" = assign,
        "FUN" = FUN,
        "additional_arguments" = MoreArgs,
        "progress_bar" = progress_bar,
        "chopchop" = chopchop,
        "overhead_time" = overhead_time,
        "process_time" = process_time,
        "measure_time" = MEASURE.TIME,
        "process_scheduling" = "static",
        "output_format" = output_format,
        "use_names" = USE.NAMES
      ),
      list(...)
    )
  )

  return(y)
}



fam_sapply_lb <- function(
    cl = NULL,
    assign = NULL,
    X,
    FUN,
    ...,
    progress_bar = FALSE,
    SIMPLIFY = TRUE,
    USE.NAMES = TRUE,
    MEASURE.TIME = FALSE
) {
  # Determine the output format.
  output_format <- ifelse(SIMPLIFY, "vector", "list")
  
  # Get the name of the initial argument.
  initial_argument_name <- head(names(formals(FUN)), n = 1L)

  # Upgrade X to a named list.
  X <- list(X)
  names(X) <- initial_argument_name

  # Make call to .fam_apply.
  y <- do.call(
    .fam_apply,
    args = c(
      list(
        "cl" = cl,
        "assign" = assign,
        "FUN" = FUN,
        "additional_arguments" = list(...),
        "progress_bar" = progress_bar,
        "process_scheduling" = "dynamic",
        "output_format" = output_format,
        "use_names" = USE.NAMES,
        "measure_time" = MEASURE.TIME
      ),
      X
    )
  )
  
  return(y)
}



fam_lapply_lb <- function(
    cl = NULL,
    assign = NULL,
    X,
    FUN,
    ...,
    progress_bar = FALSE,
    SIMPLIFY = FALSE,
    MEASURE.TIME = FALSE
) {
  # Determine the output format.
  output_format <- ifelse(SIMPLIFY, "vector", "list")
  
  # Get the name of the initial argument.
  initial_argument_name <- head(names(formals(FUN)), n = 1L)

  # Upgrade X to a named list.
  X <- list(X)
  names(X) <- initial_argument_name

  # Make call to .fam_apply.
  y <- do.call(
    .fam_apply,
    args = c(
      list(
        "cl" = cl,
        "assign" = assign,
        "FUN" = FUN,
        "additional_arguments" = list(...),
        "progress_bar" = progress_bar,
        "process_scheduling" = "dynamic",
        "output_format" = output_format,
        "measure_time" = MEASURE.TIME
      ),
      X
    )
  )
  return(y)
}



fam_mapply_lb <- function(
    cl = NULL,
    assign = NULL,
    FUN,
    ...,
    progress_bar = FALSE,
    MoreArgs = NULL,
    SIMPLIFY = FALSE,
    USE.NAMES = TRUE,
    MEASURE.TIME = FALSE
) {
  # Determine the output format.
  output_format <- ifelse(SIMPLIFY, "vector", "list")
  
  # Make call to .fam_apply.
  y <- do.call(
    .fam_apply,
    args = c(
      list(
        "cl" = cl,
        "assign" = assign,
        "FUN" = FUN,
        "additional_arguments" = MoreArgs,
        "progress_bar" = progress_bar,
        "process_scheduling" = "dynamic",
        "output_format" = output_format,
        "use_names" = USE.NAMES,
        "measure_time" = MEASURE.TIME
      ),
      list(...)
    )
  )
  
  return(y)
}



.fam_apply <- function(
    cl = NULL,
    assign = NULL,
    FUN,
    ...,
    additional_arguments = NULL,
    progress_bar = FALSE,
    chopchop = FALSE,
    overhead_time = NULL,
    process_time = NULL,
    measure_time = FALSE,
    process_scheduling = "static",
    output_format = "list",
    use_names = TRUE
) {
  .check_parameter_value_is_valid(
    x = process_scheduling,
    var_name = "process_scheduling",
    values = c("static", "dynamic")
  )
  
  .check_parameter_value_is_valid(
    x = output_format,
    var_name = "output_format",
    values = c("list", "vector")
  )
  
  # Check the length of the dots argument.
  n_x <- .dots_arg_length(...)
  if (n_x == 0L) {
    return(list())
  }

  # Adapt cl to the iterable data in ...
  if (inherits(cl, "cluster")) {
    if (n_x == 1L) cl <- NULL
  }

  # Check that there is more than one node.
  if (inherits(cl, "cluster")) {
    if (length(cl) == 1L) cl <- NULL
  }

  if ((chopchop || measure_time) && inherits(cl, "cluster")) {
    if (!require_package(
      x = "microbenchmark",
      purpose = "to make use of optimised parallel processing",
      as_error = FALSE
    )) {
      # If package is not installed, do not do anything fancy.
      chopchop <- measure_time <- FALSE
    }
  }

  # Check whether optimisation of the apply distribution is required and useful.
  require_optimisation <- chopchop && 
    inherits(cl, "cluster") && 
    (is.null(overhead_time) ||
       is.null(process_time))
  require_process_stacking <- chopchop && 
    inherits(cl, "cluster") && 
    length(process_time) == n_x && 
    !is.null(overhead_time)

  # Set initial values.
  y_initial <- NULL
  skip_element <- NULL
  stacking_data <- NULL

  # Adapt cl based on optimised overhead times.
  if (chopchop && inherits(cl, "cluster")) {
    if (require_optimisation) {
      # Send to a chunking function for chopping into a list with up to n_cl
      # elements. The first element may be skipped, because we already analysed
      # it.
      dots <- .chop_args(
        args = list(...),
        n_cl = 1L,
        n_x = 1L
      )

      # Measure process start
      overhead_start <- microbenchmark::get_nanotime()

      # Setup initial cluster.
      if (.needs_cluster_restart() && !.is_external_cluster()) {
        cl_test <- .restart_cluster(
          cl = cl[1L],
          assign = assign,
          n_nodes = 1L
        )
      } else {
        cl_test <- .update_info_on_cluster(
          cl = cl[1L],
          assign = assign
        )
      }

      # Iterate over dots and perform analysis.
      y <- do.call(
        parallel::clusterMap,
        args = list(
          "fun" = .chopped_apply,
          "cl" = cl_test,
          "dots" = dots,
          "SIMPLIFY" = FALSE,
          "MoreArgs" = list(
            "wrapper_fun" = .wrapper_fun_time,
            "FUN2" = FUN,
            "measure_time" = TRUE,
            "additional_arguments" = additional_arguments
          )
        )
      )
      
      # Measure process end
      overhead_end <- microbenchmark::get_nanotime()
      
      # Collect data and concatenate to a flat list.
      y <- unlist(y, recursive = FALSE)
      y <- .flatten_nested_fam_apply(y)

      # Compute overhead and process times. Overhead time is computed from the
      # time passed minus the time of the longest subprocess.
      overhead_time <- (overhead_end - overhead_start) / 1.0E9 - y$process_time_total

      # Obtain initial
      y_initial <- y

      # Skip analysis of the first iterable element.
      skip_element <- 1L
      n_x <- n_x - 1L

      # Update the first node that was used to run the test.
      cl[1L] <- cl_test
      
    } else if (require_process_stacking) {
      # Determine how to optimally stack the processes.
      stacking_data <- .auto_stacker(
        n_nodes = length(cl),
        overhead_time = overhead_time,
        process_time = process_time
      )
    }

    # Set the optimal number of nodes.
    if (require_process_stacking) {
      if (is.null(stacking_data)) {
        n_nodes_optimal <- 0L
      } else {
        n_nodes_optimal <- max(stacking_data$node_id)
      }
      
    } else {
      # Compute the optimal number of nodes. This is based on computing the
      # derivative of the process time curve, and determining the optima. Here t
      # = f(n, m) = m * overhead_time + n / m * process_time, with n the number
      # of processes and m the number of available nodes.
      n_nodes_optimal <- as.integer(
        floor(sqrt(n_x * stats::median(c(process_time, y$process_time)) / overhead_time))
      )
    }

    # Select the required nodes.
    if (n_nodes_optimal <= 1L) {
      cl <- NULL
    } else if (n_nodes_optimal < length(cl)) {
      cl <- cl[1L:n_nodes_optimal]
    }
  }

  # Adapt cl based on the number of iterable elements.
  if (inherits(cl, "cluster")) {
    if (n_x < length(cl)) cl <- cl[1L:n_x]

    if (measure_time) startup_overhead_start <- microbenchmark::get_nanotime()

    # Update and restart clusters.
    if (.needs_cluster_restart() && !.is_external_cluster()) {
      if (require_optimisation) {
        cl[-1L] <- .restart_cluster(
          cl = cl[-1L],
          assign = assign
        )
      } else {
        cl <- .restart_cluster(
          cl = cl,
          assign = assign
        )
      }

      # Close cluster on exit.
      on.exit(.terminate_cluster(cl))
      
    } else {
      if (require_optimisation) {
        cl[-1L] <- .update_info_on_cluster(
          cl = cl[-1L],
          assign = assign
        )
      } else {
        cl <- .update_info_on_cluster(
          cl = cl,
          assign = assign
        )
      }
    }

    if (measure_time) {
      startup_overhead_end <- microbenchmark::get_nanotime()
      startup_overhead_time <- (startup_overhead_end - startup_overhead_start) / 1.0E9
      
    } else {
      startup_overhead_time <- 0.0
    }
  }

  # Set up the wrapper.
  wrapper_fun <- .wrapper_fun
  if (measure_time) wrapper_fun <- .wrapper_fun_time
  if (progress_bar && !measure_time && !inherits(cl, "cluster")) wrapper_fun <- .wrapper_fun_progress
  if (progress_bar && measure_time && !inherits(cl, "cluster")) wrapper_fun <- .wrapper_fun_time_progress

  # Start progress bar.
  if (progress_bar && !inherits(cl, "cluster")) {
    pb_conn <- utils::txtProgressBar(min = 0L, max = n_x, style = 3L)
  } else {
    pb_conn <- NULL
  }

  if (!inherits(cl, "cluster")) {
    # Check if sequential processing was selected despite cluster nodes being
    # present initially. For simplicity, we recompute the first element, as the
    # lack of cluster indicates that the time spent by the process is
    # negligible.
    if (require_optimisation) n_x <- n_x + 1L

    # Perform the sequential apply as mapply.
    y <- do.call(
      mapply,
      args = c(
        list("FUN" = wrapper_fun),
        list(...),
        list(
          "II" = seq_len(n_x),
          "SIMPLIFY" = FALSE,
          "MoreArgs" = list(
            "FUN2" = FUN,
            "pb_conn" = pb_conn,
            "MoreArgs" = additional_arguments
          )
        )
      )
    )
    
    # Collect data and concatenate to a flat list.
    y <- .flatten_nested_fam_apply(y)
    
  } else if (inherits(cl, "cluster") && !chopchop) {
    if (measure_time) overhead_start <- microbenchmark::get_nanotime()

    # Perform the sequential apply as clusterMap.
    y <- do.call(
      parallel::clusterMap,
      args = c(
        list(
          "cl" = cl,
          "fun" = wrapper_fun
        ),
        list(...),
        list(
          "SIMPLIFY" = FALSE,
          ".scheduling" = process_scheduling,
          "MoreArgs" = list(
            "FUN2" = FUN,
            "MoreArgs" = additional_arguments
          )
        )
      )
    )

    if (measure_time) overhead_end <- microbenchmark::get_nanotime()

    # Collect data and concatenate to a flat list.
    y <- .flatten_nested_fam_apply(y)

    if (measure_time) {
      # Compute overhead and process times. Overhead time is computed from the
      # time passed minus the time of the longest subprocess.
      overhead_time <- (
        startup_overhead_time + (overhead_end - overhead_start) / 1.0E9 - sum(y$process_time) / length(cl)
      ) / length(cl)
      
    }
  } else if (inherits(cl, "cluster") && chopchop) {
    # Send to a chunking function for chopping into a list with up to n_cl
    # elements. The first element may be skipped, because we already analysed
    # it.
    dots <- .chop_args(
      args = list(...),
      n_cl = length(cl),
      stacking_data = stacking_data,
      skip_elements = skip_element
    )

    if (measure_time) overhead_start <- microbenchmark::get_nanotime()

  
    # Iterate over dots and perform analysis.
    y <- do.call(
      parallel::clusterMap,
      args = list(
        "fun" = .chopped_apply,
        "cl" = cl,
        "dots" = dots,
        "MoreArgs" = list(
          "wrapper_fun" = wrapper_fun,
          "FUN2" = FUN,
          "measure_time" = measure_time,
          "additional_arguments" = additional_arguments
        )
      )
    )
    
    if (measure_time) overhead_end <- microbenchmark::get_nanotime()
    
    # Collect data and concatenate to a flat list.
    y <- unlist(y, recursive = FALSE)
    y <- .flatten_nested_fam_apply(y)

    # Reorder results to original order. This is because the stacking algorithm
    # will assign based on expected time passed.
    if (!is.null(stacking_data)) {
      # Reorder results to input order.
      y$results[stacking_data$original_index] <- y$results

      # Reorder process time to input order.
      if (!is.null(y$process_time)) y$process_time[stacking_data$original_index] <- y$process_time
    }

    if (measure_time) {
      # Compute overhead and process times. Overhead time is computed from the
      # time passed minus the time of the longest subprocess.
      overhead_time <- (
        startup_overhead_time + (overhead_end - overhead_start) / 1.0E9 - max(y$process_time_total)
      ) / length(cl)
    }

    # Add in initial list.
    y <- .flatten_nested_list(c(list(y_initial), list(y)), flatten = TRUE)
    
  } else {
    ..error_reached_unreachable_code(".fam_apply: no correct combination found.")
  }

  # Simplify to output format.
  if (output_format == "vector") y$results <- simplify2array(y$results)

  # Set names.
  if (use_names) {
    dots <- list(...)

    # Try to get names.
    element_names <- names(dots[[1L]])
    if (is.null(element_names) && is.character(dots[[1L]])) {
      names(y$results) <- dots[[1L]]
    } else if (!is.null(element_names)) {
      names(y$results) <- element_names
    }
  }

  # Parse the return data.
  if (measure_time) {
    y <- list(
      "results" = y$results,
      "overhead_time" = overhead_time,
      "process_time" = y$process_time
    )
  } else {
    y <- y$results
  }

  if (!is.null(pb_conn)) close(pb_conn)

  return(y)
}



.wrapper_fun <- function(
    FUN2, 
    II = NULL, 
    pb_conn = NULL, 
    ...,
    MoreArgs = NULL
) {
  # Execute function
  y <- do.call(FUN2, args = c(
    list(...),
    MoreArgs
  ))

  return(list("results" = y))
}



.wrapper_fun_time <- function(
    FUN2, 
    II = NULL, 
    pb_conn = NULL, 
    ..., 
    MoreArgs = NULL) {
  # Get process start time.
  process_start <- microbenchmark::get_nanotime()

  # Execute function
  y <- do.call(FUN2, args = c(
    list(...),
    MoreArgs
  ))

  # Get process end time.
  process_end <- microbenchmark::get_nanotime()

  return(list(
    "results" = y,
    "process_time" = (process_end - process_start) / 1.0E9
  ))
}



.wrapper_fun_progress <- function(
    FUN2, 
    II = NULL, 
    pb_conn = NULL, 
    ...,
    MoreArgs = NULL
) {
  # Execute function
  y <- do.call(FUN2, args = c(
    list(...),
    MoreArgs
  ))

  # Update the progress bar
  utils::setTxtProgressBar(pb = pb_conn, value = II)

  return(list("results" = y))
}



.wrapper_fun_time_progress <- function(
    FUN2, 
    II = NULL, 
    pb_conn = NULL, 
    ..., 
    MoreArgs = NULL
) {
  # Get process start time.
  process_start <- microbenchmark::get_nanotime()

  # Execute function
  y <- do.call(FUN2, args = c(
    list(...),
    MoreArgs
  ))

  # Get process end time.
  process_end <- microbenchmark::get_nanotime()

  # Update the progress bar
  utils::setTxtProgressBar(pb = pb_conn, value = II)

  return(list(
    "results" = y,
    "process_time" = (process_end - process_start) / 1.0E9
  ))
}



.dots_arg_length <- function(...) {
  # Get the length of the dots argument.
  n_x <- unique(lengths(list(...)))

  if (length(n_x) != 1L) {
    ..error_reached_unreachable_code(
      ".dots_arg_length: iterable arguments do not have the same length."
    )
  } 

  return(n_x)
}



.chopped_apply <- function(
    wrapper_fun,
    FUN2,
    dots,
    additional_arguments = NULL,
    measure_time = FALSE
) {
  if (measure_time) process_total_start <- microbenchmark::get_nanotime()

  # Perform the sequential apply as mapply.
  y <- do.call(
    mapply,
    args = c(
      list("FUN" = wrapper_fun),
      dots,
      list(
        "SIMPLIFY" = FALSE,
        "USE.NAMES" = FALSE,
        "MoreArgs" = list(
          "FUN2" = FUN2,
          "MoreArgs" = additional_arguments
        )
      )
    )
  )

  if (measure_time) process_total_end <- microbenchmark::get_nanotime()

  # Set total process time (which may vary between nodes.)
  if (measure_time) y[[1L]]$process_time_total <- (process_total_end - process_total_start) / 1.0E9

  return(y)
}



.flatten_nested_fam_apply <- function(x) {
  x <- .flatten_nested_list(x)

  if (!is.null(x$process_time)) x$process_time <- unlist(x$process_time)
  if (!is.null(x$process_time_total)) x$process_time_total <- unlist(x$process_time_total)

  return(x)
}



.chop_args <- function(
    args,
    n_cl,
    n_x = NULL,
    skip_elements = NULL,
    stacking_data = NULL
) {
  # Determine n_x if not provided. n_x may be provided when we attempt to
  # determine process and overhead time.
  if (is.null(n_x)) n_x <- unique(lengths(args))

  if (length(n_x) != 1L) ..error_reached_unreachable_code(".chop_args: arguments do not have the same length.")

  # Set up the sequence of values.
  n_x <- seq_len(n_x)

  # Skip if there elements that should be skipped. This is useful if, for
  # example, the optimisation process needs to determine overhead time.
  if (!is.null(skip_elements)) n_x <- setdiff(n_x, skip_elements)

  # Make sure not to generate empty entries. Therefore the number of nodes can
  # only be up to the number of available elements.
  if (n_cl > length(n_x)) n_cl <- length(n_x)

  if (length(n_x) == 0L || n_cl == 0L) {
    index <- list()
    
  } else if (length(n_x) == 1L || n_cl == 1L) {
    index <- list(n_x)
    
  } else if (!is.null(stacking_data)) {
    index <- unname(lapply(
      split(stacking_data, by = "node_id", sorted = TRUE),
      function(list_element) (list_element$original_index)
    ))
    
  } else {
    index <- structure(
      split(n_x, cut(n_x, n_cl)),
      names = NULL
    )
  }

  # Split arguments internally.
  chopped_args <- lapply(
    index,
    function(ii, x) {
      return(
        lapply(
          x,
          function(x, ii) {
            if (data.table::is.data.table(x)) {
              column_names <- colnames(x)[ii]
              return(x[, mget(column_names)])
            } else {
              return(x[ii])
            }
          },
          ii = ii
        )
      )
    },
    x = args
  )
  
  return(chopped_args)
}



.auto_stacker <- function(
    n_nodes,
    overhead_time,
    process_time,
    min_decrease = 0.05
) {
  # Check if there are fewer processes than nodes.
  if (n_nodes > length(process_time)) n_nodes <- length(process_time)

  # Check if there is any reason to parallellise.
  if (n_nodes < 2L) return(NULL)

  # Determine the baseline time.
  baseline_time <- sum(process_time)

  # Compute the minimal decrease.
  min_decrease <- min_decrease * baseline_time

  # Initialise parameters for the while loop.
  previous_time <- baseline_time
  previous_stacking_data <- NULL

  # Initial nodes
  current_nodes <- 2L

  while (current_nodes <= n_nodes) {
    # Check if the overhead time does not exceed the previous time.
    if (current_nodes * overhead_time >= previous_time) break

    # Get current stacking data
    current_stacking_data <- ..auto_stack(
      n_nodes = current_nodes,
      process_time = process_time,
      overhead_time = overhead_time
    )

    # Check that an increase in nodes does not negatively affect the time spent
    # processing. Also, ensure that the achieved decrease is at least
    # min_decrease with regard to the previous time.
    if (
      previous_time <= current_stacking_data$node_time || 
      previous_time - current_stacking_data$node_time < min_decrease
    ) {
      break
    }

    # Replace stacking data for the next iteration.
    previous_stacking_data <- current_stacking_data

    # Update other variables.
    previous_time <- current_stacking_data$node_time
    current_nodes <- current_nodes + 1L
  }

  # Return process data belonging to the process prior to
  return(previous_stacking_data$process_data)
}



..auto_stack <- function(n_nodes, process_time, overhead_time) {
  # Suppress NOTES due to non-standard evaluation in data.table
  node_id <- NULL

  # Initiate a placeholder for the time allocated to each node.
  node_time <- numeric(n_nodes)

  # Sort process times. Assign the longest processes first.
  process_data <- data.table::data.table("process_time" = process_time)
  process_data[, ":="("original_index" = .I, "node_id" = 0L)]
  process_data <- process_data[order(process_time, decreasing = TRUE)]

  for (ii in seq_along(process_time)) {
    # Select the node that has the least process time assigned to it.
    selected_node <- which.min(node_time)

    # Assign node to process
    process_data[ii, "node_id" := selected_node]

    # Update time on the node
    node_time[selected_node] <- node_time[selected_node] + process_data[ii, ]$process_time
  }

  # Set element order.
  process_data <- process_data[order(node_id)]
  process_data[, "process_index" := .I]

  # Return the maximum node time (which is of course)
  return(list(
    "node_time" = max(node_time) + overhead_time * n_nodes,
    "process_data" = process_data
  ))
}

# Don't perform any further tests on CRAN due to parallelisation issues.
testthat::skip_on_cran()
testthat::skip_on_ci()

apply_vector_fun <- function(x) {
  Sys.sleep(x)

  return(x)
}

apply_vector_fun_2 <- function(x, y) {
  Sys.sleep(x)

  return(y)
}

apply_list_fun <- function(x) {
  Sys.sleep(x)

  return(list("x" = x))
}

apply_list_fun_2 <- function(x, y) {
  Sys.sleep(x)

  return(list("x" = x, "y" = y))
}

apply_list_fun_3 <- function(x, y, z) {
  Sys.sleep(x)

  return(list("x" = x, "y" = y, "z" = z))
}

# Start local cluster in the overall process.
cl <- familiar:::.test_start_cluster(2L)


# Generate random values.
x_sequential <- stats::rnorm(5, mean = 1.0, sd = 0.2)
x_sequential[x_sequential < 0.2] <- 0.2

x_sequential_named <- x_sequential
names(x_sequential_named) <- letters[seq_along(x_sequential_named)]

x_single <- x_sequential[1]

x_parallel <- stats::rnorm(20, mean = 1.0, sd = 0.2)
x_parallel[x_parallel < 0.2] <- 0.2

x_parallel_named <- x_parallel
names(x_parallel_named) <- letters[seq_along(x_parallel_named)]

# fam_sapply -------------------------------------------------------------------

testthat::test_that("fam_sapply works correctly.", {
  # Simple sequential fam_sapply
  x <- familiar:::fam_sapply(
    cl = NULL,
    assign = NULL,
    X = x_sequential,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, x_sequential)

  # Simple sequential fam_sapply with names
  x <- familiar:::fam_sapply(
    cl = NULL,
    assign = NULL,
    X = x_sequential_named,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, x_sequential_named)

  # Sequential fam_sapply with extra argument.
  x <- familiar:::fam_sapply(
    cl = NULL,
    assign = NULL,
    X = x_sequential,
    FUN = apply_vector_fun_2,
    y = "A")

  testthat::expect_length(x, length(x_sequential))
  testthat::expect_setequal(x, "A")

  # Sequential fam_sapply with mini-batching (does nothing)
  x <- familiar:::fam_sapply(
    cl = NULL,
    assign = NULL,
    X = x_sequential,
    FUN = apply_vector_fun,
    chopchop = TRUE)

  testthat::expect_equal(x, x_sequential)

  # Simple parallel fam_sapply.
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = x_sequential,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, x_sequential)

  # Simple parallel fam_sapply with names.
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = x_sequential_named,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, x_sequential_named)

  # Parallel fam_sapply with extra argument.
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = x_sequential,
    FUN = apply_vector_fun_2,
    y = "A")

  testthat::expect_length(x, length(x_sequential))
  testthat::expect_setequal(x, "A")

  # Parallel fam_sapply with mini-batching.
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = x_parallel,
    FUN = apply_vector_fun,
    chopchop = TRUE)

  testthat::expect_equal(x, x_parallel)

  # Parallel fam_sapply with mini-batching and "predicted" process times.
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = x_parallel,
    FUN = apply_vector_fun,
    chopchop = TRUE,
    process_time = x_parallel,
    overhead_time = 0.1)

  testthat::expect_equal(x, x_parallel)

  # Parallel fam_sapply with mini-batching and "predicted" process times.
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = x_parallel_named,
    FUN = apply_vector_fun,
    chopchop = TRUE,
    process_time = x_parallel,
    overhead_time = 0.1)

  testthat::expect_equal(x, x_parallel_named)

  # Parallel, load-balanced sapply.
  x <- familiar:::fam_sapply_lb(
    cl = cl,
    assign = NULL,
    X = x_sequential,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, x_sequential)

  # Parallel, load-balanced sapply with time measurement.
  x <- familiar:::fam_sapply_lb(
    cl = cl,
    assign = NULL,
    X = x_sequential,
    FUN = apply_vector_fun_2,
    y = "A",
    MEASURE.TIME = TRUE)

  testthat::expect_length(x$results, length(x_sequential))
  testthat::expect_setequal(x$results, "A")

  testthat::expect_equal(x$process_time, x_sequential, tolerance = 0.1)

  # Sequential sapply with NULL
  x <- familiar:::fam_sapply(
    cl = NULL,
    assign = NULL,
    X = NULL,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, list())

  # Parallel sapply with NULL
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = NULL,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, list())

  # Parallel sapply with mini-batching and NULL
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = NULL,
    FUN = apply_vector_fun,
    chopchop = TRUE)

  testthat::expect_equal(x, list())

  # Sequential sapply with single value
  x <- familiar:::fam_sapply(
    cl = NULL,
    assign = NULL,
    X = x_single,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, x_single)

  # Parallel sapply with single value
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = x_single,
    FUN = apply_vector_fun)

  testthat::expect_equal(x, x_single)

  # Parallel sapply with mini-batching and single value.
  x <- familiar:::fam_sapply(
    cl = cl,
    assign = NULL,
    X = x_single,
    FUN = apply_vector_fun,
    chopchop = TRUE)

  testthat::expect_equal(x, x_single)
})

# fam_lapply -------------------------------------------------------------------

testthat::test_that("fam_lapply works correctly.", {
  # Simple sequential lapply.
  x <- familiar:::fam_lapply(
    cl = NULL,
    assign = NULL,
    X = x_sequential,
    FUN = apply_list_fun)

  testthat::expect_equal(unname(unlist(x)), x_sequential)

  # Simple sequential fam_sapply with names
  x <- familiar:::fam_lapply(
    cl = NULL,
    assign = NULL,
    X = x_sequential_named,
    FUN = apply_list_fun)

  testthat::expect_equal(names(x), names(x_sequential_named))
  testthat::expect_equal(unname(unlist(x)), x_sequential)

  # Simple sequential lapply with extra argument.
  x <- familiar:::fam_lapply(
    cl = NULL,
    assign = NULL,
    X = x_sequential,
    FUN = apply_list_fun_2,
    y = "A")

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_sequential)
  testthat::expect_setequal(sapply(x, function(x) (x$y)), "A")

  # Sequential lapply with mini-batching (no effect).
  x <- familiar:::fam_lapply(
    cl = NULL,
    assign = NULL,
    X = x_sequential,
    FUN = apply_list_fun,
    chopchop = TRUE)

  testthat::expect_equal(unname(unlist(x)), x_sequential)

  # Parallel lapply.
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = x_sequential,
    FUN = apply_list_fun_2,
    y = "A")

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_sequential)
  testthat::expect_setequal(sapply(x, function(x) (x$y)), "A")

  # Parallel lapply with mini-batching.
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = x_parallel,
    FUN = apply_list_fun,
    chopchop = TRUE)

  testthat::expect_equal(unname(unlist(x)), x_parallel)

  # Parallel lapply with mini-batching and named elements.
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = x_parallel_named,
    FUN = apply_list_fun,
    chopchop = TRUE)

  testthat::expect_equal(names(x), names(x_parallel_named))
  testthat::expect_equal(unname(unlist(x)), x_parallel)

  # Parallel lapply with mini-batching and "predicted" process times.
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = x_parallel,
    FUN = apply_list_fun,
    chopchop = TRUE,
    process_time = x_parallel,
    overhead_time = 0.1)

  testthat::expect_equal(unname(unlist(x)), x_parallel)

  # Parallel lapply with mini-batching and "predicted" process times.
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = x_parallel_named,
    FUN = apply_list_fun,
    chopchop = TRUE,
    process_time = x_parallel,
    overhead_time = 0.1)

  testthat::expect_equal(names(x), names(x_parallel_named))
  testthat::expect_equal(unname(unlist(x)), x_parallel)

  # Parallel lapply with mini-batching, named elements and vector output.
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = x_parallel_named,
    FUN = apply_vector_fun,
    chopchop = TRUE)

  testthat::expect_equal(names(x), names(x_parallel_named))
  testthat::expect_equal(unlist(x), x_parallel_named)

  # Parallel lapply with load balancing.
  x <- familiar:::fam_lapply_lb(
    cl = cl,
    assign = NULL,
    X = x_sequential,
    FUN = apply_list_fun)

  testthat::expect_equal(unname(unlist(x)), x_sequential)

  # Parallel lapply with load balancing and time measurement.
  x <- familiar:::fam_lapply_lb(
    cl = cl,
    assign = NULL,
    X = x_sequential,
    FUN = apply_list_fun_2,
    y = "A",
    MEASURE.TIME = TRUE)

  testthat::expect_equal(sapply(x$results, function(x) (x$x)), x_sequential)
  testthat::expect_setequal(sapply(x$results, function(x) (x$y)), "A")

  testthat::expect_equal(x$process_time, x_sequential, tolerance = 0.1)

  # Sequential lapply with NULL
  x <- familiar:::fam_lapply(
    cl = NULL,
    assign = NULL,
    X = NULL,
    FUN = apply_list_fun)

  testthat::expect_equal(x, list())

  # Parallel lapply with NULL
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = NULL,
    FUN = apply_list_fun)

  testthat::expect_equal(x, list())

  # Parallel lapply with mini-batching and NULL
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = NULL,
    FUN = apply_list_fun,
    chopchop = TRUE)

  testthat::expect_equal(x, list())

  # Sequential lapply with single value
  x <- familiar:::fam_lapply(
    cl = NULL,
    assign = NULL,
    X = x_single,
    FUN = apply_list_fun)

  testthat::expect_equal(unname(unlist(x)), x_single)

  # Parallel lapply with single value
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = x_single,
    FUN = apply_list_fun)

  testthat::expect_equal(unname(unlist(x)), x_single)

  # Parallel lapply with mini-batching and single value.
  x <- familiar:::fam_lapply(
    cl = cl,
    assign = NULL,
    X = x_single,
    FUN = apply_list_fun,
    chopchop = TRUE)

  testthat::expect_equal(unname(unlist(x)), x_single)
})



# fam_mapply -------------------------------------------------------------------

testthat::test_that("fam_mapply works correctly.", {
  # Simple sequential mapply,
  x <- familiar:::fam_mapply(
    cl = NULL,
    assign = NULL,
    FUN = apply_list_fun,
    x = x_sequential)

  testthat::expect_equal(unname(unlist(x)), x_sequential)

  # Simple sequential mapply, with additional argument.
  x <- familiar:::fam_mapply(
    cl = NULL,
    assign = NULL,
    FUN = apply_list_fun_2,
    x = x_sequential,
    y = seq_along(x_sequential))

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_sequential)
  testthat::expect_equal(sapply(x, function(x) (x$y)), seq_along(x_sequential))

  # Simple sequential mapply, with two additional arguments.
  x <- familiar:::fam_mapply(
    cl = NULL,
    assign = NULL,
    FUN = apply_list_fun_3,
    x = x_sequential,
    y = seq_along(x_sequential),
    MoreArgs = list("z" = "A"))

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_sequential)
  testthat::expect_equal(sapply(x, function(x) (x$y)), seq_along(x_sequential))
  testthat::expect_setequal(sapply(x, function(x) (x$z)), "A")

  # Simple sequential mapply, with two additional arguments, and named input.
  x <- familiar:::fam_mapply(
    cl = NULL,
    assign = NULL,
    FUN = apply_list_fun_3,
    x = x_sequential_named,
    y = seq_along(x_sequential_named),
    MoreArgs = list("z" = "A"))

  testthat::expect_equal(names(x), names(x_sequential_named))
  testthat::expect_equal(unname(sapply(x, function(x) (x$x))), x_sequential)
  testthat::expect_equal(unname(sapply(x, function(x) (x$y))), seq_along(x_sequential))
  testthat::expect_setequal(unname(sapply(x, function(x) (x$z))), "A")

  # Sequential mapply with mini-batching (no effect)
  x <- familiar:::fam_mapply(
    cl = NULL,
    assign = NULL,
    FUN = apply_list_fun_2,
    x = x_sequential,
    y = seq_along(x_sequential),
    chopchop = TRUE)

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_sequential)
  testthat::expect_equal(sapply(x, function(x) (x$y)), seq_along(x_sequential))

  # Simple parallel mapply.
  x <- familiar:::fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun_2,
    x = x_sequential,
    y = seq_along(x_sequential))

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_sequential)
  testthat::expect_equal(sapply(x, function(x) (x$y)), seq_along(x_sequential))

  # Parallel mapply with mini-batching.
  x <- familiar:::fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun_2,
    x = x_parallel,
    y = seq_along(x_parallel),
    chopchop = TRUE)

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_parallel)
  testthat::expect_equal(sapply(x, function(x) (x$y)), seq_along(x_parallel))

  # Parallel mapply with mini-batching and "predicted" process times.
  x <- familiar:::fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun_2,
    x = x_parallel,
    y = seq_along(x_parallel),
    process_time = x_parallel,
    overhead_time = 0.1,
    chopchop = TRUE)

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_parallel)
  testthat::expect_equal(sapply(x, function(x) (x$y)), seq_along(x_parallel))

  # Parallel mapply with mini-batching, "predicted" process times and additional
  # arguments.
  x <- familiar:::fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun_3,
    x = x_parallel,
    y = seq_along(x_parallel),
    process_time = x_parallel,
    overhead_time = 0.1,
    MoreArgs = list("z" = "A"),
    chopchop = TRUE)

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_parallel)
  testthat::expect_equal(sapply(x, function(x) (x$y)), seq_along(x_parallel))
  testthat::expect_setequal(sapply(x, function(x) (x$z)), "A")

  # Parallel mapply with load balancing.
  x <- familiar:::fam_mapply_lb(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun_2,
    x = x_sequential,
    y = seq_along(x_sequential))

  testthat::expect_equal(sapply(x, function(x) (x$x)), x_sequential)
  testthat::expect_equal(sapply(x, function(x) (x$y)), seq_along(x_sequential))

  # Parallel mapply with load balancing and time measurement.
  x <- familiar:::fam_mapply_lb(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun_2,
    x = x_sequential,
    y = seq_along(x_sequential),
    MEASURE.TIME = TRUE)

  testthat::expect_equal(sapply(x$results, function(x) (x$x)), x_sequential)
  testthat::expect_equal(sapply(x$results, function(x) (x$y)), seq_along(x_sequential))

  testthat::expect_equal(x$process_time, x_sequential, tolerance = 0.1)

  # Sequential mapply with NULL
  x <- familiar:::fam_mapply(
    cl = NULL,
    assign = NULL,
    FUN = apply_list_fun,
    x = NULL)

  testthat::expect_equal(x, list())

  # Parallel mapply with NULL
  x <- familiar:::fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun,
    x = NULL)

  testthat::expect_equal(x, list())

  # Parallel mapply with mini-batching and NULL
  x <- familiar:::fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun,
    x = NULL,
    chopchop = TRUE)

  testthat::expect_equal(x, list())

  # Sequential mapply with single value
  x <- familiar:::fam_mapply(
    cl = NULL,
    assign = NULL,
    FUN = apply_list_fun,
    x = x_single)

  testthat::expect_equal(unname(unlist(x)), x_single)

  # Parallel mapply with single value
  x <- familiar:::fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun,
    x = x_single)

  testthat::expect_equal(unname(unlist(x)), x_single)

  # Parallel mapply with mini-batching and single value.
  x <- familiar:::fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = apply_list_fun,
    x = x_single,
    chopchop = TRUE)

  testthat::expect_equal(unname(unlist(x)), x_single)
})

familiar:::.terminate_cluster(cl)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# plot_feature_similarity (generic) --------------------------------------------

#' @title Plot heatmaps for pairwise similarity between features.
#'
#' @description This method creates a heatmap based on data stored in a
#'   `familiarCollection` object. Features in the heatmap are ordered so that
#'   more similar features appear together.
#'
#' @param dir_path (*optional*) Path to the directory where created performance
#'   plots are saved to. Output is saved in the `feature_similarity`
#'   subdirectory. If `NULL` no figures are saved, but are returned instead.
#' @param gradient_palette (*optional*) Sequential or divergent palette used to
#'   colour the similarity or distance between features in a heatmap. 
#'   `familiar` has a default palette. Other palettes are supported by the
#'   `paletteer` package, `grDevices::palette.pals()` (requires R >= 4.0.0),
#'   `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`, `heat.colors`,
#'   `terrain.colors`, `topo.colors` and `cm.colors`, which correspond to the
#'   palettes of the same name in `grDevices`. You may also specify your own
#'   palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param gradient_palette_range (*optional*) Numerical range used to span the
#'   gradient. This should be a range of two values, e.g. `c(0, 1)`. Lower or
#'   upper boundary can be unset by using `NA`. If not set, the full
#'   metric-specific range is used.
#' @param show_dendrogram (*optional*) Show dendrogram around the main panel.
#'   Can be `TRUE`, `FALSE`, `NULL`, or a position, i.e. `top`, `bottom`, `left`
#'   and `right`. Up to two positions may be provided, but only as long as the
#'   dendrograms are not on opposite sides of the heatmap: `top` and `bottom`,
#'   and `left` and `right` cannot be used together.
#'
#'   A dendrogram can only be drawn from cluster methods that produce
#'   dendrograms, such as `hclust`. A dendrogram can for example not be
#'   constructed using the partitioning around medioids method (`pam`).
#'
#'   By default, a dendrogram is drawn to the top and right of the panel.
#' @param dendrogram_height (*optional*) Height of the dendrogram. The height is
#'   1.5 cm by default. Height is expected to be grid unit (see `grid::unit`),
#'   which also allows for specifying relative heights.
#' @inheritParams export_feature_similarity
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_feature_similarity -object -feature_cluster_method -feature_linkage_method -feature_cluster_cut_method -feature_similarity_threshold
#'
#' @details This function generates area under the ROC curve plots.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, and `data_set`.
#'   By default, the data is split by `vimp_method`, `learner` and `data_set`,
#'   since the features may be ordered differently for each data set.
#'
#'   Note that similarity is determined based on the underlying data. Hence the
#'   ordering of features may differ between facets, and tick labels are
#'   maintained for each panel.
#'
#'   Labeling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'   
#'   This plot can be created from `dataObject`, or `data.table` objects.
#'   For `data.table`, see \code{\link{as_data_object}} for additional arguments.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_feature_similarity
#' @md
#' @rdname plot_feature_similarity-methods
setGeneric(
  "plot_feature_similarity",
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    gradient_palette_range = NULL,
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    y_range = NULL,
    y_n_breaks = 3L,
    y_breaks = NULL,
    rotate_x_tick_labels = waiver(),
    show_dendrogram = c("top", "right"),
    dendrogram_height = grid::unit(1.5, "cm"),
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_feature_similarity")
  }
)



# plot_feature_similarity (general) --------------------------------------------

#' @rdname plot_feature_similarity-methods
setMethod(
  "plot_feature_similarity",
  signature(object = "ANY"),
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    gradient_palette_range = NULL,
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    y_range = NULL,
    y_n_breaks = 3L,
    y_breaks = NULL,
    rotate_x_tick_labels = waiver(),
    show_dendrogram = c("top", "right"),
    dendrogram_height = grid::unit(1.5, "cm"),
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "feature_similarity",
          "feature_cluster_method" = feature_cluster_method,
          "feature_linkage_method" = feature_linkage_method,
          "feature_cluster_cut_method" = feature_cluster_cut_method,
          "feature_similarity_threshold" = feature_similarity_threshold
        ),
        list(...)
      )
    )

    return(do.call(
      plot_feature_similarity,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "gradient_palette" = gradient_palette,
        "gradient_palette_range" = gradient_palette_range,
        "x_label" = x_label,
        "x_label_shared" = x_label_shared,
        "y_label" = y_label,
        "y_label_shared" = y_label_shared,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "rotate_x_tick_labels" = rotate_x_tick_labels,
        "show_dendrogram" = show_dendrogram,
        "dendrogram_height" = dendrogram_height,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_feature_similarity (collection) -----------------------------------------

#' @rdname plot_feature_similarity-methods
setMethod(
  "plot_feature_similarity",
  signature(object = "familiarCollection"),
  function(
    object,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_threshold = waiver(),
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    gradient_palette = NULL,
    gradient_palette_range = NULL,
    x_label = waiver(),
    x_label_shared = "column",
    y_label = waiver(),
    y_label_shared = "row",
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    y_range = NULL,
    y_n_breaks = 3L,
    y_breaks = NULL,
    rotate_x_tick_labels = waiver(),
    show_dendrogram = c("top", "right"),
    dendrogram_height = grid::unit(1.5, "cm"),
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Get input data
    x <- export_feature_similarity(
      object = object,
      feature_cluster_method = feature_cluster_method,
      feature_linkage_method = feature_linkage_method,
      feature_cluster_cut_method = feature_cluster_cut_method,
      feature_similarity_threshold = feature_similarity_threshold,
      export_dendrogram = FALSE,
      export_ordered_data = FALSE
    )

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Obtain data element from list.
    if (is.list(x)) {
      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_feature_similarity: list of data elements contains unmerged elements."
        )
      }

      # Get x directly.
      x <- x[[1L]]
    }

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = TRUE),
      purpose = "to plot feature similarity heatmaps",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Check input arguments ----------------------------------------------------

    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # rotate_x_tick_labels
    if (is.waive(rotate_x_tick_labels)) {
      rotate_x_tick_labels <- FALSE
    }

    # x_label_shared
    if (!is.waive(x_label_shared)) {
      .check_input_plot_args(x_label_shared = x_label_shared)
    } else {
      x_label_shared <- "column"
    }

    # y_label_shared
    if (!is.waive(y_label_shared)) {
      .check_input_plot_args(y_label_shared = y_label_shared)
    } else {
      y_label_shared <- "row"
    }

    # x_label
    if (is.waive(x_label)) x_label <- NULL

    # y_label
    if (is.waive(y_label)) y_label <- NULL

    # legend_label
    if (is.waive(legend_label)) legend_label <- NULL

    # show_dendrogram
    if (is.logical(show_dendrogram)) {
      if (show_dendrogram) {
        show_dendrogram <- c("top", "right")
      } else {
        show_dendrogram <- NULL
      }
      
    } else if (length(show_dendrogram) == 0L) {
      show_dendrogram <- NULL
      
    } else {
      .check_parameter_value_is_valid(
        x = show_dendrogram,
        var_name = "show_dendrogram",
        values = c("top", "bottom", "left", "right")
      )

      # Check that bottom and top, and left and right do not appear together.
      if (all(c("top", "bottom") %in% show_dendrogram)) {
        ..error("Dendrograms can not be drawn both above and below the plot.")
      } else if (all(c("left", "right") %in% show_dendrogram)) {
        ..error("Dendrograms can not be drawn both to the left and right of the plot.")
      }
    }

    if (!x@cluster_method %in% c("hclust", "agnes", "diana")) show_dendrogram <- NULL

    # Check if the dendrogram_height argument is correct.
    if (!is.null(show_dendrogram)) {
      .check_plot_grid_unit(x = dendrogram_height, var_name = "dendrogram_height")
    }

    # Add default splitting variables
    if (is.null(split_by) & is.null(facet_by)) {
      # Split by variable importance method, learner and data set
      split_by <- c("vimp_method", "learner", "data_set")
      
      # Do not use facets.
      facet_by <- NULL
    }

    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      facet_by = facet_by,
      available = c("data_set", "vimp_method", "learner")
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    facet_by <- split_var_list$facet_by

    # Check input arguments
    .check_input_plot_args(
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      rotate_x_tick_labels = rotate_x_tick_labels
    )

    ##### Create plots -------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data.
    if (!is.null(split_by)) {
      x_split <- split(x@data, by = split_by, drop = FALSE)
    } else {
      x_split <- list("null.name" = x@data)
    }

    # Store plots to list in case no dir_path is provided
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over splits
    for (x_sub in x_split) {
      if (is_empty(x_sub)) next

      if (is.waive(plot_title)) plot_title <- "Feature similarity"

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = list("metric" = x@similarity_metric),
          x = x_sub
        )
      }

      # Generate plot
      p <- .plot_feature_similarity_plot(
        x = x_sub,
        data = x,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        gradient_palette = gradient_palette,
        gradient_palette_range = gradient_palette_range,
        x_label = x_label,
        x_label_shared = x_label_shared,
        y_label = y_label,
        y_label_shared = y_label_shared,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        y_range = y_range,
        y_n_breaks = y_n_breaks,
        y_breaks = y_breaks,
        rotate_x_tick_labels = rotate_x_tick_labels,
        show_dendrogram = show_dendrogram,
        dendrogram_height = dendrogram_height
      )

      # Check empty output
      if (is.null(p)) next

      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Identify unique features:
        features <- unique(c(x_sub$feature_name_1, x_sub$feature_name_2))

        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_feature_similarity_plot_dimensions(
          x = x_sub,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols,
          features = as.character(features),
          show_dendrogram = show_dendrogram,
          rotate_x_tick_labels = rotate_x_tick_labels
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "feature_similarity",
              "subtype" = "similarity",
              "x" = x_sub,
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_feature_similarity_plot <- function(
    x,
    data,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    gradient_palette,
    gradient_palette_range,
    x_label,
    x_label_shared,
    y_label,
    y_label_shared,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    y_range,
    y_n_breaks,
    y_breaks,
    rotate_x_tick_labels,
    show_dendrogram,
    dendrogram_height
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  .NATURAL <- NULL
  
  # Split by facet. This generates a list of data splits with faceting
  # information that allows for positioning.
  plot_layout_table <- .get_plot_layout_table(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Define the split in data required for faceting.
  layout_split <- split(
    plot_layout_table,
    by = c("col_id", "row_id"),
    sorted = TRUE
  )

  # Create plots to join
  figure_list <- list()
  extracted_element_list <- list()
  for (current_split in layout_split) {
    # Generate the split in case there is a faceting variable.
    if (!is.null(facet_by)) {
      x_split <- methods::new(
        "familiarDataElementFeatureSimilarity",
        data,
        data = x[current_split, on = .NATURAL, nomatch = NULL]
      )
      
    } else {
      x_split <- methods::new(
        "familiarDataElementFeatureSimilarity",
        data,
        data = x
      )
    }

    # Add in clustering information and a dendrogram.
    x_split <- .append_feature_similarity_dendrogram(x_split)
    x_split <- .append_feature_similarity_clustering(x_split)

    # Find the cluster object
    dendrogram <- x_split@dendrogram

    # Complete the similarity data
    similarity_data <- .complete_feature_similarity_table(
      x = x_split@data,
      similarity_metric = x_split@similarity_metric
    )

    # Create similarity heatmap
    p_heatmap <- .create_feature_similarity_heatmap(
      x = similarity_data,
      facet_by = facet_by,
      facet_wrap_cols = facet_wrap_cols,
      ggtheme = ggtheme,
      gradient_palette = gradient_palette,
      gradient_palette_range = gradient_palette_range,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      rotate_x_tick_labels = rotate_x_tick_labels,
      show_dendrogram = show_dendrogram,
      similarity_metric = x_split@similarity_metric
    )

    # Convert to gtable and append "main" to grob names.
    g_heatmap <- .rename_plot_grobs(
      g = .convert_to_grob(p_heatmap),
      extension = "main"
    )
    if (!gtable::is.gtable(g_heatmap)) next

    # Add dendrogram
    if (!is.null(show_dendrogram) && inherits(dendrogram, "hclust")) {
      # Obtain dendrogram plotting data as line segments.
      dendro_data <- .convert_dendrogram_to_table(
        h = dendrogram,
        similarity_metric = x_split@similarity_metric
      )
      
      for (position in show_dendrogram) {
        # Plot dendrogram
        p_dendro <- .create_feature_similarity_dendrogram_plot(
          x = dendro_data,
          position = position,
          ggtheme = ggtheme,
          y_range = y_range,
          y_n_breaks = y_n_breaks,
          y_breaks = y_breaks,
          plot_height = dendrogram_height,
          rotate_x_tick_labels = rotate_x_tick_labels
        )
        
        dendro_extension <- paste0("dendro-", position)
        panel_element_name <- paste0("panel-", dendro_extension)
        axis_element_name <- ifelse(position %in% c("top", "bottom"), "axis-l", "axis-b")
        axis_element_name <- paste0(axis_element_name, "-", dendro_extension)
        
        # Convert to gtable
        g_dendro <- .rename_plot_grobs(
          g = .convert_to_grob(p_dendro),
          extension = dendro_extension
        )
        
        where_panel <- switch(
          position,
          "top" = c("above", "panel-main"),
          "bottom" = c("below", "panel-main"),
          "left" = c("left", "panel-main"),
          "right" = c("right", "panel-main")
        )
        
        # Insert panel-main next to panel-dendro.
        g_heatmap <- .gtable_insert(
          g = g_heatmap,
          g_new = .gtable_extract_grob(g_dendro, element = panel_element_name),
          where = where_panel,
          grob_name = panel_element_name,
          spacer = .get_plot_panel_spacing(
            ggtheme = ggtheme, 
            axis = ifelse(position %in% c("top", "bottom"), "y", "x")
          )
        )
        
        where_axis_element <- switch(
          position,
          "top" = c("intersect", "above", "axis-l-main", "left", panel_element_name),
          "bottom" = c("intersect", "below", "axis-l-main", "left", panel_element_name),
          "left" = c("intersect", "below", panel_element_name, "left", "axis-b-main"),
          "right" = c("intersect", "below", panel_element_name, "right", "axis-b-main")
        )
        
        # Insert the axis element at the intersect of dendro-panel and the
        # corresponding axis element of the main plot.
        g_heatmap <- .gtable_insert(
          g = g_heatmap,
          g_new = .gtable_extract_grob(g_dendro, element = axis_element_name),
          where = where_axis_element,
          grob_name = axis_element_name
        )
      }
    }
    
    # Attach to figure list.
    figure_list[[paste0(current_split$row_id, ".", current_split$col_id)]] <- as_familiar_plot(
      g = g_heatmap,
      layout = current_split
    )
  }
  # Compose the final figure.
  g <- .compose_figure(
    figure_list = figure_list,
    plot_layout_table = plot_layout_table,
    x_text_shared = x_label_shared,
    x_label_shared = x_label_shared,
    y_text_shared = y_label_shared,
    y_label_shared = y_label_shared,
    facet_wrap_cols = facet_wrap_cols,
    ggtheme = ggtheme
  )

  return(g)
}



.create_feature_similarity_heatmap <- function(
    x,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    gradient_palette,
    gradient_palette_range,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    rotate_x_tick_labels,
    show_dendrogram,
    similarity_metric
) {
  
  if (is.null(gradient_palette_range) && !is.null(similarity_metric)) {
    # Find the palette range.
    gradient_palette_range <- get_similarity_range(similarity_metric = similarity_metric)
  }

  # Determine whether a sequential or divergent palette should be used by default.
  palette_type <- ifelse(
    length(gradient_palette_range) > 2L,
    "divergent",
    "sequential"
  )

  # Should the palette be inverted? This is because for some metrics, clusters
  # are those with least distance, not highest similarity.
  invert_palette <- FALSE
  if (!is.null(similarity_metric)) {
    invert_palette <- is_default_distance(similarity_metric = similarity_metric)
  }

  # Create basic plot
  p <- ggplot2::ggplot(data = x, mapping = ggplot2::aes(
    x = !!sym("feature_name_1"),
    y = !!sym("feature_name_2"),
    fill = !!sym("value")
  ))
  p <- p + ggtheme

  if (!is_empty(x)) {
    p <- p + ggplot2::geom_raster()
  } else {
    # Insert blank for empty datasets.
    p <- p + ggplot2::geom_blank()
  }


  # Colors
  gradient_colours <- .get_palette(
    x = gradient_palette,
    palette_type = palette_type
  )
  if (invert_palette) gradient_colours <- rev(gradient_colours)

  if (length(gradient_palette_range) > 0L) {
    transform <- "identity"
    limits <- range(gradient_palette_range)
    if (similarity_metric == "mutual_information") {
      transform <- "log"
      limits[1L] <- 1E-3
    }
    
    # Add gradient palette. If the legend is not shown, legend_label equals
    # NULL.
    p <- p + ggplot2::scale_fill_gradientn(
      name = legend_label,
      colors = gradient_colours,
      limits = limits,
      oob = scales::squish,
      transform = transform
    )
  }

  # Show dendrogram determines where tick labels and axis labels are placed. By
  # default axes are located at the bottom and right.
  x_axis_position <- "bottom"
  y_axis_position <- "left"
  if (!is.null(show_dendrogram)) {
    if ("bottom" %in% show_dendrogram) {
      x_axis_position <- "top"
    }

    if ("left" %in% show_dendrogram) {
      y_axis_position <- "right"
    }
  }

  # Specify both axes. Note that only the heatmap is shown, without additional
  # space between the plot area and the axes.
  p <- p + ggplot2::scale_x_discrete(position = x_axis_position, expand = c(0.0, 0.0))
  p <- p + ggplot2::scale_y_discrete(position = y_axis_position, expand = c(0.0, 0.0))

  # Set labels.
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Determine how plots are facetted. The actual facets are created in the
  # calling function, not here.
  facet_by_list <- .parse_plot_facet_by(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context",
        drop = TRUE
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context",
        drop = TRUE
      )
    }
  }

  # Rotate x-axis ticks
  if (rotate_x_tick_labels) {
    p <- p + ggplot2::theme(
      axis.text.x = ggplot2::element_text(
        vjust = 0.25, 
        hjust = 1.0, 
        angle = 90.0
      )
    )
  }

  return(p)
}



.create_feature_similarity_dendrogram_plot <- function(
    x,
    position,
    ggtheme,
    y_range,
    y_n_breaks,
    y_breaks,
    plot_height,
    rotate_x_tick_labels
) {
  # Check if there is any data to plot.
  if (is_empty(x)) return(NULL)
  
  # Define the range along the x-axis.
  x_range <- range(x$x_1, na.rm = TRUE)
  x_range <- c(x_range[1L] - 0.5, x_range[2L] + 0.5)

  # y_range
  if (is.null(y_range)) y_range <- range(c(x$y_1, x$y_2), na.rm = TRUE)

  # y_breaks
  if (is.null(y_breaks)) {
    .check_input_plot_args(
      y_range = y_range,
      y_n_breaks = y_n_breaks
    )

    # Create breaks and update y_range
    y_breaks <- labeling::extended(
      m = y_n_breaks,
      dmin = y_range[1L],
      dmax = y_range[2L],
      only.loose = TRUE
    )

    y_range <- c(
      head(y_breaks, n = 1L),
      tail(y_breaks, n = 1L)
    )
  }

  .check_input_plot_args(
    y_range = y_range,
    y_breaks = y_breaks
  )

  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("x_1"),
      y = !!sym("y_1"),
      xend = !!sym("x_2"),
      yend = !!sym("y_2")
    )
  )
  p <- p + ggtheme
  
  # Plot line segments.
  p <- p + ggplot2::geom_segment(lineend = "round")
  
  if (position == "right") {
    p <- p + ggplot2::scale_x_continuous(
      limits = x_range, 
      expand = c(0.0, 0.0)
    )
    p <- p + ggplot2::scale_y_continuous(
      limits = y_range, 
      breaks = y_breaks
    )
    p <- p + ggplot2::coord_flip()
    
  } else if (position == "bottom") {
    p <- p + ggplot2::scale_x_continuous(
      limits = x_range, 
      expand = c(0.0, 0.0)
    )
    p <- p + ggplot2::scale_y_reverse(
      limits = rev(y_range),
      breaks = rev(y_breaks)
    )
    
  } else if (position == "left") {
    p <- p + ggplot2::scale_x_continuous(
      limits = x_range, 
      expand = c(0.0, 0.0)
    )
    
    p <- p + ggplot2::scale_y_reverse(
      limits = rev(y_range),
      breaks = rev(y_breaks)
    )
    p <- p + ggplot2::coord_flip()
    
  } else if (position == "top") {
    p <- p + ggplot2::scale_x_continuous(
      limits = x_range, 
      expand = c(0.0, 0.0)
    )
    
    p <- p + ggplot2::scale_y_continuous(
      limits = y_range, 
      breaks = y_breaks
    )
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".create_feature_similarity_dendrogram_plot: unknown position encountered: ",
      position
    ))
  }

  # Remove some theme elements and reduce margins. The histogram height is left.
  p <- p + ggplot2::theme(
    panel.grid = ggplot2::element_blank(),
    panel.background = ggplot2::element_blank(),
    panel.border = ggplot2::element_blank(),
    axis.title.x = ggplot2::element_blank(),
    axis.title.y = ggplot2::element_blank()
  )

  if (position %in% c("top", "bottom")) {
    # Remove x-axis
    p <- p + ggplot2::theme(
      axis.line.x = ggplot2::element_blank(),
      axis.ticks.x = ggplot2::element_blank(),
      axis.text.x = ggplot2::element_blank()
    )
  } else if (position %in% c("left", "right")) {
    # Remove y-axis (rotated x-axis in plot)
    p <- p + ggplot2::theme(
      axis.line.y = ggplot2::element_blank(),
      axis.ticks.y = ggplot2::element_blank(),
      axis.text.y = ggplot2::element_blank()
    )
    # Rotate x-labels
    if (rotate_x_tick_labels) {
      p <- p + ggplot2::theme(
        axis.text.x = ggplot2::element_text(
          vjust = 0.25, 
          hjust = 1.0, 
          angle = 90.0
        )
      )
    }
  }

  # Ensure that panel heights/widths are set to the plot object.
  if (position %in% c("top", "bottom")) {
    p$custom_grob <- list("heights" = list("name" = "panel", "height" = plot_height))
  } else if (position %in% c("left", "right")) {
    p$custom_grob <- list("widths" = list("name" = "panel", "width" = plot_height))
  }

  class(p) <- c("familiar_ggplot", class(p))

  return(p)
}



.determine_feature_similarity_plot_dimensions <- function(
    x,
    facet_by,
    facet_wrap_cols,
    features,
    rotate_x_tick_labels,
    show_dendrogram
) {
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )

  # Determine the number of elements along the x-axis.
  x_n_elements <- y_n_elements <- length(features)
  x_longest_element <- y_longest_element <- max(sapply(features, nchar))

  # Assume each x-axis element takes up about 0.5 cm. Then add some room for
  # other plot elements.
  default_width <- x_n_elements * 0.5 + 1.0

  # Assume each y-axis element takes up about 0.5 cm as well.
  default_height <- y_n_elements * 0.5 + 1.0

  # Reserve space for y-axis and x-axis tick labels. Assume that the typical
  # width of a character is about 5 points (1.8 mm). For the x-axis we only
  # reserve extra space in case the ticks are rotated, otherwise we just
  # assume a typical height of 10 points (3.6 mm).
  y_tick_space <- y_longest_element * 0.18
  x_tick_space <- ifelse(rotate_x_tick_labels, x_longest_element * 0.18, 0.36)

  # Reserve space for the dendrograms (1.5 cm)
  dendro_height <- ifelse(any(c("top", "bottom") %in% show_dendrogram), 1.5, 0.0)
  dendro_width <- ifelse(any(c("left", "right") %in% show_dendrogram), 1.5, 0.0)

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * (default_height + x_tick_space + dendro_height), 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * (default_width + y_tick_space + dendro_width), 19.0))

  return(c(height, width))
}


.complete_feature_similarity_table <- function(x, similarity_metric) {
  # Check for empty file.
  if (is_empty(x)) {
    return(NULL)
  }

  # Copy x and revert feature_name_1 and feature_name_2 so that all pairs are
  # present.
  y <- data.table::copy(x)
  data.table::setnames(y,
    old = c("feature_name_1", "feature_name_2", "label_order_1", "label_order_2"),
    new = c("feature_name_2", "feature_name_1", "label_order_2", "label_order_1")
  )

  # Combine.
  x <- rbind(x, y, use.names = TRUE)

  # Determine order of features.
  feature_1_order <- unique(x[, c("feature_name_1", "label_order_1")])
  feature_2_order <- unique(x[, c("feature_name_2", "label_order_2")])

  # Add self-paired features.
  features <- unique(x$feature_name_1)
  y <- x[rep(1.0, length(features))]
  y[, ":="("feature_name_1" = features,
    "feature_name_2" = features,
    "value" = 1.0,
    "label_order_1" = NULL,
    "label_order_2" = NULL
  )]
  y <- merge(x = y, y = feature_1_order, by = "feature_name_1", all = FALSE)
  y <- merge(x = y, y = feature_2_order, by = "feature_name_2", all = FALSE)

  # Combine.
  x <- rbind(x, y, use.names = TRUE)

  # Reorder features
  x$feature_name_1 <- factor(
    x = x$feature_name_1, 
    levels = feature_1_order$feature_name_1[order(feature_1_order$label_order_1)]
  )
  x$feature_name_2 <- factor(
    x = x$feature_name_2,
    levels = feature_2_order$feature_name_2[order(feature_2_order$label_order_2)]
  )

  return(x)
}

.get_available_rank_aggregation_methods <- function() {
  return(c(
    "none",
    "mean",
    "median",
    "best",
    "worst",
    "stability",
    "exponential",
    "borda",
    "enhanced_borda",
    "truncated_borda",
    "enhanced_truncated_borda"
  ))
}



.optimise_feature_occurrence_threshold <- function(vimp_table) {
  if (is(vimp_table, "vimpTable")) vimp_table <- vimp_table@vimp_table

  # Determine number of features, runs and the maximum rank
  n_features <- data.table::uniqueN(vimp_table, by = "name")
  n_runs <- data.table::uniqueN(vimp_table, by = "run_id")
  max_rank <- max(vimp_table$rank)

  # Check if there is only one feature or only one run
  if (n_features == 1L) return(1L)
  if (n_runs == 1L) return(max_rank)

  # Set max threshold to either 50 or max_rank, if lower.
  max_threshold <- min(c(max_rank, 50L))

  # We need to calculate for all possible cut-off thresholds which threshold
  # maximises variance in occurrence
  threshold_var <- sapply(
    seq_len(max_threshold),
    function(threshold, n_runs, vimp_table) {
      return(stats::var(
        .compute_feature_occurrence(
          vimp_table = vimp_table, 
          n_runs = n_runs, 
          threshold = threshold
        )$occurrence
      ))
    },
    vimp_table = vimp_table, 
    n_runs = n_runs
  )

  # Return optimal threshold
  return(which.max(threshold_var))
}



.compute_feature_occurrence <- function(vimp_table, n_runs, threshold) {
  # Suppress NOTES due to non-standard evaluation in data.table
  name <- NULL

  # Make a local copy
  vimp_table <- data.table::copy(vimp_table)

  # Determine the occurrence for features with rank less or equal to the given threshold.
  vimp_table <- vimp_table[, list(occurrence = sum(rank <= threshold) / n_runs), by = name]

  return(vimp_table)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarIsolationForest ------------------------------------------------------
setClass(
  "familiarIsolationForest",
  contains = "familiarNoveltyDetector"
)


# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarIsolationForest"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- "isotree"

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarIsolationForest"),
  function(object, ...) {
    # Isolation forests exists for all outcome types and data.
    return(TRUE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarIsolationForest"),
  function(
    object, 
    data = NULL, 
    ...
  ) {
    # Initialise list and declare hyperparameter entries.
    param <- list()
    param$n_tree <- list()
    param$sample_size <- list()
    param$m_try <- list()
    param$tree_depth <- list()
    param$n_dim <- list()

    # If data is not provided, return the list with hyperparameter names only.
    if (is.null(data)) return(param)

    # Get the number of instances. Note that we are not interested in samples
    # per se, as we are assessing novelty.
    n_samples <- nrow(data@data)

    # Get the number of features.
    n_features <- get_n_features(x = data)

    # number of trees ----------------------------------------------------------
    
    # We limit the number of trees to limit memory footprint.
    default_n_trees <- log2(max(c(64L, ceiling(sqrt(n_samples)))))

    # Note that the number of trees is defined in powers of 2, based on Oshiro,
    # T. M., Perez, P. S., & Baranauskas, J. A. (2012, July). How many trees in
    # a random forest?. In MLDM (pp. 154-168).
    param$n_tree <- .set_hyperparameter(
      default = default_n_trees,
      type = "integer",
      range = c(4L, 10L),
      valid_range = c(0L, Inf),
      randomise = FALSE
    )


    # sample size --------------------------------------------------------------
    
    # We limit the number of samples in each tree to limit memory footprint.
    default_sample_size <- max(c(128L, ceiling(sqrt(n_samples))))
    if (n_samples < default_sample_size) default_sample_size <- n_samples

    # Express as fraction.
    default_sample_size <- default_sample_size / n_samples

    # Note that the sample size is here noted as a fraction, which corresponds
    # to the usage in ranger.
    param$sample_size <- .set_hyperparameter(
      default = default_sample_size,
      type = "numeric",
      range = c(2.0 / n_samples, 1.0),
      valid_range = c(0.0, 1.0),
      randomise = FALSE
    )

    # number of candidate features selected at node ----------------------------

    default_m_try <- 3.0 / n_features
    if (default_m_try > 1.0) default_m_try <- 1.0

    # Note that the number of features is here noted as a fraction, but is used
    # in the isolation forest as an integer. Familiar ensures that always at
    # least 1 feature is available as a candidate.
    param$m_try <- .set_hyperparameter(
      default = default_m_try,
      type = "numeric",
      range = c(0.0, 1.0),
      randomise = FALSE
    )

    # maximum tree depth -------------------------------------------------------

    default_tree_depth <- as.integer(ceiling(log2(default_sample_size * n_samples)))
    if (default_tree_depth < 1L) default_tree_depth <- 1L

    # Determines the depth trees are allowed to grow to. Larger depths increase
    # the risk of overfitting.
    param$tree_depth <- .set_hyperparameter(
      default = default_tree_depth,
      type = "integer",
      range = c(1L, 10L),
      valid_range = c(1L, Inf),
      randomise = FALSE
    )

    # number of splitting dimensions -------------------------------------------

    # Switch between extended and conventional isolation trees.
    if (object@learner %in% c("isolation_forest", "extended_isolation_forest")) {
      default_n_dim <- 3L
    } else if (object@learner %in% c("simple_isolation_forest")) {
      default_n_dim <- 1L
    }

    # The number of splitting dimensions cannot be larger than the number of
    # features.
    if (default_n_dim > n_features && n_features > 0L) default_n_dim <- n_features

    param$n_dim <- .set_hyperparameter(
      default = default_n_dim,
      type = "integer",
      range = c(1L, 3L),
      valid_range = c(1L, Inf),
      randomise = FALSE
    )

    return(param)
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarIsolationForest",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Check if the training data is ok.
    if (has_bad_training_data(object = object, data = data)) return(callNextMethod())

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) return(callNextMethod())

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Replace any ordered variables by factors. We do this because
    # ordered features can not be handled using isotree.
    ordered_features <- colnames(data@data)[sapply(data@data, is.ordered)]
    
    for (current_feature in ordered_features) {
      data@data[[current_feature]] <- factor(
        x = data@data[[current_feature]],
        levels = levels(data@data[[current_feature]]),
        ordered = FALSE
      )
    }
    
    # Check that the current ndim parameter is not larger than the number of
    # features.
    features <- get_feature_columns(data)
    if (object@hyperparameters$n_dim > length(features)) {
      object@hyperparameters$n_dim <- length(features)
    }
    
    # Extract hyperparameters from the object.
    param <- object@hyperparameters
    
    # Create an isolation forest. Note that in addition to specifying the number
    # of trees and the number of samples assessed for each tree, missing_action
    # is set to "fail" -- this decreases model footprint and is not necessary as
    # familiar has its own imputation routines.
    detector <- isotree::isolation.forest(
      data = data@data[, mget(features)],
      sample_size = ceiling(param$sample_size * nrow(data@data)),
      ntrees = as.integer(ceiling(2.0^(param$n_tree))),
      ndim = param$n_dim,
      ntry = max(c(1L, as.integer(ceiling(param$m_try * length(features))))),
      max_depth = param$tree_depth,
      nthreads = 1L,
      missing_action = "fail"
    )

    # Add model
    object@model <- detector

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarIsolationForest",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "novelty",
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) return(callNextMethod())

    # Check if the data is empty.
    if (is_empty(data)) return(callNextMethod())

    # Find and replace ordered features.
    ordered_features <- colnames(data@data)[sapply(data@data, is.ordered)]
    for (current_feature in ordered_features) {
      data@data[[current_feature]] <- factor(
        x = data@data[[current_feature]],
        levels = levels(data@data[[current_feature]]),
        ordered = FALSE
      )
    }

    # Find novelty values.
    novelty_values <- predict(
      object = object@model,
      newdata = data@data,
      nthreads = 1L
    )
    
    prediction_table <- as_prediction_table(
      x = novelty_values,
      type = "novelty",
      data = data,
      model_object = object
    )

    return(prediction_table)
  }
)



# .trim_model ------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarIsolationForest"),
  function(object, ...) {
    # Add show.
    object <- .capture_show(object)

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    return(object)
  }
)




# has_bad_training_data --------------------------------------------------------
setMethod(
  "has_bad_training_data", 
  signature(
    object = "familiarNoveltyDetector",
    data = "dataObject"
  ),
  function(
    object,
    data,
    allow_no_features = FALSE,
    ...
  ) {
    # Checks the data for consistency and usability. Any errors are passed as
    # attributes
    
    # Call general, less stringent, routine first.
    return_value <- callNextMethod()
    if (return_value) return(return_value)
    
    if (nrow(data@data) < 3L) {
      return_value <- TRUE
      attr(return_value, "error") <- paste0(
        "Too few samples were available to train the isolation forest."
      )
      
      return(return_value)
    }
    
    return(FALSE)
  }
)



.get_available_isolation_forest_detectors <- function(show_general = TRUE) {
  return(c("isolation_forest", "simple_isolation_forest", "extended_isolation_forest"))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# add_feature_info_parameters (NULL, ANY) --------------------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "NULL",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # This is when feature info parameters are completely missing.
    return(object)
  }
)



# add_feature_info_parameters (feature info parameters, ANY) -------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParameters",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Default behaviour is to return the object as is.
    return(object)
  }
)


# apply_feature_info_parameters (NULL, ANY) ------------------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "NULL",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # This is when feature info parameters are completely missing.
    return(data)
  }
)



# apply_feature_info_parameters (feature info parameters, ANY) -----------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParameters",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Default behaviour is to return the data as is.
    return(data)
  }
)



# add_package_version (feature info parameters) --------------------------------
setMethod(
  "add_package_version",
  signature(object = "featureInfoParameters"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarGLM object -----------------------------------------------------------
setClass("familiarGLM",
  contains = "familiarModel",
  slots = list(
    "encoding_reference_table" = "ANY",
    "feature_order" = "character"
  ),
  prototype = list(
    "encoding_reference_table" = NULL,
    "feature_order" = character()
  )
)

# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarGLM"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    if (.Object@outcome_type == "multinomial") {
      .Object@package <- "nnet"
    } else if (.Object@outcome_type == "survival") {
      .Object@package <- "survival"
    } else {
      # The default option is to use "fastglm::fastglm" for speed, but if this
      # package is not available, switch to the "stats::glm" backup option.
      .Object@package <- if (is_package_installed("fastglm")) c("stats", "fastglm") else "stats"
    }

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarGLM"),
  function(object, ...) {
    # Extract outcome type and learner from the model object.
    outcome_type <- object@outcome_type
    learner <- object@learner

    # Check outcome type and learner.
    if (
      outcome_type == "binomial" &&
      learner %in% c("glm", "glm_logistic", "glm_probit", "glm_cauchy", "glm_loglog")
    ) {
      return(TRUE)
      
    } else if (
      outcome_type == "multinomial" &&
      learner %in% c("glm", "glm_multinomial")
    ) {
      return(TRUE)
      
    } else if (
      outcome_type == "continuous" &&
      learner %in% c(
        "glm", "glm_log", "glm_gaussian", "glm_log_gaussian",
        "glm_inv_gaussian", "glm_poisson", "glm_log_poisson"
      )
    ) {
      return(TRUE)
      
    } else if (
      outcome_type == "survival" &&
      learner %in% c("glm")
    ) {
      return(TRUE)
      
    } else if (
      outcome_type == "count" &&
      learner %in% c("glm", "glm_poisson", "glm_log_poisson")
    ) {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(FALSE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarGLM"),
  function(object, data = NULL, user_list = NULL, ...) {
    # Initialise list and declare hyperparameter entries
    param <- list()
    param$sign_size <- list()
    param$family <- list()
    param$sample_weighting <- list()
    param$sample_weighting_beta <- list()

    # If no data object is not provided, return the list with hyperparameter
    # names only
    if (is.null(data)) return(param)

    # Get the outcome type
    outcome_type <- data@outcome_type

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(
      data = data, 
      restrict_samples = TRUE
    )

    # model family -------------------------------------------------------------

    # Read family string by parsing the learner.
    fam <- sub(
      x = object@learner,
      pattern = "glm",
      replacement = "",
      fixed = TRUE
    )
    if (fam != "") {
      fam <- sub(
        x = fam,
        pattern = "_",
        replacement = "",
        fixed = TRUE
      )
    }

    # Determine the family or families.
    if (fam == "") {
      # If no family is specified, the default behaviour is to identify the
      # family through optimisation.
      if (outcome_type == "binomial") {
        family_default <- c("logistic", "probit", "loglog", "cauchy")
      } else if (outcome_type == "continuous") {
        family_default <- c("gaussian", "log_gaussian", "inv_gaussian", "poisson", "log_poisson")
      } else if (outcome_type == "multinomial") {
        family_default <- "multinomial"
      } else if (outcome_type == "survival") {
        family_default <- "cox"
      } else {
        ..error_outcome_type_not_implemented(outcome_type)
      }
      
    } else if (fam == "log") {
      # "log" is a collection of different families, that should be specified
      # according to the outcome type.
      if (outcome_type == "continuous") {
        family_default <- c("log_gaussian", "log_poisson")
      }
      
    } else {
      # A family was unambiguously specified.
      family_default <- fam
    }

    # Set family parameter.
    param$family <- .set_hyperparameter(
      default = family_default,
      type = "factor",
      range = family_default,
      randomise = length(family_default) > 1L
    )

    # sample weighting method --------------------------------------------------
    # Class imbalances may lead to learning majority classes. This can be
    # partially mitigated by increasing weight of minority classes.
    param$sample_weighting <- .get_default_sample_weighting_method(
      outcome_type = outcome_type
    )

    # effective number of samples beta -----------------------------------------
    # Specifies the beta parameter for effective number sample weighting method.
    # See Cui et al. (2019).
    param$sample_weighting_beta <- .get_default_sample_weighting_beta(
      method = c(
        param$sample_weighting$init_config,
        user_list$sample_weighting
      ),
      outcome_type = outcome_type
    )

    return(param)
  }
)



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarGLM"),
  function(object, type = "default") {
    if (object@outcome_type != "survival") {
      return(callNextMethod())
    }

    # This is a backup in case glm is used to refer to CoxPH methods.
    if (type == "default") {
      return("hazard_ratio")
    } else if (type == "survival_probability") {
      return("survival_probability")
    } else {
      ..error_reached_unreachable_code("get_prediction_type,familiarGLM: unknown type")
    }
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarGLM",
    data = "dataObject"
  ),
  function(object, data, approximate = FALSE, ...) {
    # For survival outcomes, switch to familiarCoxPH.
    if (object@outcome_type == "survival") {
      # Create a familiarCoxPH object.
      object <- methods::new("familiarCoxPH", object)

      return(..train(
        object = object,
        data = data,
        ...
      ))
    }

    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object, 
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Use effect coding to convert categorical data into encoded data - this is
    # required to deal with factors with missing/new levels between training and
    # test data sets.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )

    # Find feature columns in the data.
    feature_columns <- get_feature_columns(x = encoded_data$encoded_data)

    # Parse formula
    formula <- stats::reformulate(
      termlabels = feature_columns, 
      response = quote(outcome)
    )

    # Get family for glm, which determines how the response and predictors are
    # linked.
    family <- ..get_distribution_family(object)

    # Set weights
    weights <- create_instance_weights(
      data = encoded_data$encoded_data,
      method = object@hyperparameters$sample_weighting,
      beta = ..compute_effective_number_of_samples_beta(
        object@hyperparameters$sample_weighting_beta
      ),
      normalisation = "average_one"
    )
    
    if (object@outcome_type %in% c("binomial", "continuous")) {
      # Faster implementation using fastglm. We keep the stats::glm
      # implementation in case fastglm disappears from CRAN at some point.

      version <- ifelse(
        require_package("fastglm", message_type = "silent"),
        "fastglm",
        "stats"
      )

      if (version == "fastglm") {
        outcome_data <- encoded_data$encoded_data@data$outcome
        if (object@outcome_type == "binomial") {
          # Convert levels to [0, 1] range.
          outcome_data <- as.numeric(outcome_data) - 1.0
        }

        # Add intercept.
        encoded_data$encoded_data@data[, "intercept__" := 1.0]

        # Fit using fastglm using the LDLT Cholesky method.
        model <- do.call_with_handlers(
          fastglm::fastglm,
          args = list(
            "x" = as.matrix(encoded_data$encoded_data@data[, mget(c(feature_columns, "intercept__"))]),
            "y" = matrix(outcome_data, ncol = 1L),
            "weights" = weights,
            "family" = family,
            "method" = 3L
          )
        )
        
      } else {
        model <- do.call_with_handlers(
          stats::glm,
          args = list(formula,
            "data" = encoded_data$encoded_data@data,
            "weights" = weights,
            "family" = family,
            "model" = FALSE,
            "x" = FALSE,
            "y" = FALSE
          )
        )
      }
    } else if (object@outcome_type == "multinomial") {
      max_iterations <- ifelse(approximate, 100L, 500L)
      absolute_tolerance <- ifelse(approximate, 1.0e-2, 1.0e-4)
      relative_tolerance <- ifelse(approximate, 1.0e-2, 1.0e-8)

      # Fit multinomial logistic models using nnet::multinom.
      model <- do.call_with_handlers(
        nnet::multinom,
        args = list(formula,
          "data" = encoded_data$encoded_data@data,
          "weights" = weights,
          "maxit" = max_iterations,
          "abstol" = absolute_tolerance,
          "reltol" = relative_tolerance,
          "MaxNWts" = Inf
        )
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "..train,familiarGLM: unknown outcome type: ",
        object@outcome_type
      ))
    }

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) {
      return(callNextMethod(object = object))
    }

    # Check if all coefficients could not be estimated. Sometimes models could
    # be trained with a large number of features whose coefficients fail to
    # converge. This would sometimes lead to overly large signature sizes being
    # selected during hyperparameter optimisation, especially in situations
    # where there is not a lot of signal. Checking for non-finite coefficients
    # is an easy way to figure out if the model is not properly trained.
    if (!all(sapply(stats::coef(model), is.finite))) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_failed_model_coefficient_estimation()
      )))
    }

    # Add model
    object@model <- model

    # Add the contrast references to model_list
    object@encoding_reference_table <- encoded_data$reference_table

    # Add feature order
    object@feature_order <- feature_columns

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarGLM",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # For survival outcomes, switch to familiarCoxPH.
    if (object@outcome_type == "survival") {
      # Create a familiarCoxPH object.
      object <- methods::new("familiarCoxPH", object)

      return(..train_naive(
        object = object,
        data = data,
        ...
      ))
    }

    # Turn into a Naive model.
    object <- methods::new("familiarNaiveModel", object)

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarGLM",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "default", 
    time = NULL,
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }
    
    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }
    
    # Encode data so that the features are the same as in the training.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )
    
    if (inherits(object@model, "fastglm")) {
      encoded_data$encoded_data@data[, "intercept__" := 1.0]
    }
    
    if (type == "default") {
      # default ----------------------------------------------------------------

      if (object@outcome_type == "binomial") {
        ## Binomial outcomes ---------------------------------------------------

        class_levels <- get_outcome_class_levels(x = object)
        prediction_list <- list()

        if (inherits(object@model, "fastglm")) {
          # For fastglm::fastglm models.
          model_predictions <- suppressWarnings(predict(
            object = object@model,
            newdata = as.matrix(
              encoded_data$encoded_data@data[, mget(c(object@feature_order, "intercept__"))]
            ),
            type = "response"
          ))
          
        } else {
          # For stats::glm models.
          model_predictions <- suppressWarnings(predict(
            object = object@model,
            newdata = encoded_data$encoded_data@data,
            type = "response"
          ))
        }
        
        # Set as list so that the positive class can be directly inferred
        # without throwing a warning.
        prediction_list[[tail(class_levels, n = 1L)]] <- model_predictions
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type == "multinomial") {
        ## Multinomial outcomes ------------------------------------------------

        if (inherits(object@model, "nnet")) {
          # For nnet::multinomial models.
          model_predictions <- predict(
            object@model,
            newdata = encoded_data$encoded_data@data[, mget(object@feature_order)],
            type = "probs"
          )
        } else {
          # For VGAM::vglm
          ..deprecation_vgam()
          return(callNextMethod())
        }

        class_levels <- get_outcome_class_levels(x = object)
        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          if (is.matrix(model_predictions)) {
            prediction_list[[class_levels[ii]]] <- model_predictions[, ii]
          } else {
            prediction_list[[class_levels[ii]]] <- model_predictions[ii]
          }
        }

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("continuous")) {
        ## Continuous outcomes -------------------------------------------------

        if (inherits(object@model, "fastglm")) {
          # For fastglm::fastglm models.
          model_predictions <- suppressWarnings(
            predict(
              object = object@model,
              newdata = as.matrix(
                encoded_data$encoded_data@data[, mget(c(object@feature_order, "intercept__"))]
              ),
              type = "response"
            )
          )
          
        } else {
          # For stats::glm models.
          model_predictions <- suppressWarnings(
            predict(
              object = object@model,
              newdata = encoded_data$encoded_data@data,
              type = "response"
            )
          )
        }

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = model_predictions,
          type = "regression",
          data = data,
          model_object = object
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      return(prediction_table)
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified method --------------------------------------------------

      if (object@outcome_type %in% c("continuous", "binomial")) {
        ## Binomial and continuous outcomes ------------------------------------

        if (inherits(object@model, "fastglm")) {
          return(predict(
            object = object@model,
            newdata = as.matrix(
              encoded_data$encoded_data@data[, mget(c(object@feature_order, "intercept__"))]
            ),
            type = type,
            ...
          ))
          
        } else {
          # Use the model for prediction.
          return(predict(
            object = object@model,
            newdata = encoded_data$encoded_data@data,
            type = type,
            ...
          ))
        }
        
      } else if (object@outcome_type == "multinomial") {
        ## Multinomial outcomes ------------------------------------------------

        if (inherits(object@model, "nnet")) {
          # For nnet::multinomial models.
          return(predict(object@model,
            newdata = encoded_data$encoded_data@data[, mget(c(object@feature_order, "intercept__"))],
            type = type,
            ...
          ))
          
        } else {
          # For VGAM::vglm
          ..deprecation_vgam()
          return(callNextMethod())
        }
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarGLM"),
  function(object, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL

    if (!model_is_trained(object)) {
      return(callNextMethod())
    }

    # Check that required packages are loaded and installed.
    require_package(object, "vimp")

    # Compute z-values
    coefficient_z_values <- tryCatch(
      .compute_z_statistic(object, fix_all_missing = TRUE),
      error = identity
    )

    if (inherits(coefficient_z_values, "error")) {
      return(callNextMethod())
    }

    if (is(object@model, "vglm")) {
      ..deprecation_vgam()
      return(callNextMethod())

    } else if (inherits(object@model, "nnet")) {
      coefficient_names <- colnames(coefficient_z_values)
      coefficient_names <- rep(coefficient_names, each = nrow(coefficient_z_values))
      coefficient_z_values <- as.vector(coefficient_z_values)
      names(coefficient_z_values) <- coefficient_names
    }

    # Remove intercept from the coefficients.
    coefficient_z_values <- coefficient_z_values[
      !names(coefficient_z_values) %in% c("(Intercept)", "intercept__")
    ]
    
    if (length(coefficient_z_values) == 0L) {
      return(callNextMethod())
    }

    # Assign to variable importance table.
    vimp_table <- data.table::data.table(
      "score" = abs(coefficient_z_values),
      "name" = names(coefficient_z_values)
    )

    # Merge by name (vglm coefficients can occur multiple times for the same
    # feature).
    vimp_table <- vimp_table[, list("score" = max(score)), by = "name"]

    # Create variable importance object.
    vimp_object <- methods::new("vimpTable",
      vimp_table = vimp_table,
      encoding_table = object@encoding_reference_table,
      score_aggregation = "max",
      invert = TRUE
    )

    return(vimp_object)
  }
)



# ..get_distribution_family ----------------------------------------------------
setMethod(
  "..get_distribution_family",
  signature(object = "familiarGLM"),
  function(object) {
    # Obtain family from the hyperparameters.
    family <- object@hyperparameters$family

    # Check that required packages are loaded and installed.
    require_package(object, "distribution")

    # Check that the family hyperparameter exists.
    if (!is.character(family) && !is.factor(family)) {
      ..error_reached_unreachable_code(
        "..get_distribution_family,familiarGLM: family hyperparameter was not set."
      )
    }

    # Load families for linear regression
    if (family %in% c("logistic", "binomial")) {
      family_fun <- stats::binomial(link = "logit")
    } else if (family == "probit") {
      family_fun <- stats::binomial(link = "probit")
    } else if (family == "cauchy") {
      family_fun <- stats::binomial(link = "cauchit")
    } else if (family == "loglog") {
      family_fun <- stats::binomial(link = "cloglog")
    } else if (family == "gaussian") {
      family_fun <- stats::gaussian(link = "identity")
    } else if (family == "log_gaussian") {
      family_fun <- stats::gaussian(link = "log")
    } else if (family == "inv_gaussian") {
      family_fun <- stats::gaussian(link = "inverse")
    } else if (family == "poisson") {
      family_fun <- stats::poisson(link = "identity")
    } else if (family == "log_poisson") {
      family_fun <- stats::poisson(link = "log")
    } else if (family == "multinomial") {
      family_fun <- "_placeholder_"
    } else {
      ..error_reached_unreachable_code(paste0(
        "..get_distribution_family,familiarGLM: unknown family.", family
      ))
    }

    return(family_fun)
  }
)



# ..set_vimp_parameters --------------------------------------------------------
setMethod(
  "..set_vimp_parameters",
  signature(object = "familiarGLM"),
  function(object, method, ...) {
    # Read family string by parsing the learner.
    family_str <- sub(
      x = method,
      pattern = "glm",
      replacement = "",
      fixed = TRUE
    )
    
    if (family_str != "") {
      family_str <- sub(
        x = family_str, 
        pattern = "_",
        replacement = "",
        fixed = TRUE
      )
    }

    # Determine the family or families.
    if (family_str == "") {
      # If no family is specified, the default behaviour is to identify the
      # family through optimisation.
      if (object@outcome_type == "binomial") {
        family_default <- c("logistic")
      } else if (object@outcome_type == "continuous") {
        family_default <- c("gaussian")
      } else if (object@outcome_type == "multinomial") {
        family_default <- "multinomial"
      } else if (object@outcome_type == "survival") {
        family_default <- "cox"
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }
      
    } else if (family_str == "log") {
      # "log" is a collection of different families, that should be specified
      # according to the outcome type.
      if (object@outcome_type == "continuous") {
        family_default <- c("log_gaussian")
      }
      
    } else {
      # A family was unambiguously specified.
      family_default <- family_str
    }

    # Update family hyperparameter.
    object@hyperparameters$family <- family_default

    return(object)
  }
)



# .trim_model ------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarGLM"),
  function(object, ...) {
    if (inherits(object@model, "fastglm")) {
      # Update model by removing the call.
      object@model$call <- call("trimmed")

      # Add show.
      object <- .capture_show(object)

      object@model$fitted.values <- NULL
      object@model$linear.predictors <- NULL
      object@model$weights <- NULL
      object@model$prior.weights <- NULL
      object@model$y <- NULL
      object@model$n <- NULL
      object@model$residuals <- NULL
      
    } else if (inherits(object@model, "glm")) {
      # Update model by removing the call.
      object@model$call <- call("trimmed")

      # Add show.
      object <- .capture_show(object)

      # Remove .Environment.
      object@model$terms <- .replace_environment(object@model$terms)
      object@model$formula <- .replace_environment(object@model$formula)

      # Remove elements that contain sample-specific values.
      object@model$fitted.values <- NULL
      object@model$data <- NULL
      object@model$linear.predictors <- NULL
      object@model$prior.weights <- NULL
      object@model$weights <- NULL
      object@model$qr$qr <- NULL
      object@model$residuals <- NULL
      object@model$effects <- NULL
      
    } else if (inherits(object@model, "nnet")) {
      # Update model by removing the call.
      object@model$call <- call("trimmed")

      # Add show.
      object <- .capture_show(object)

      # Remove .Environment.
      object@model$terms <- .replace_environment(object@model$terms)

      # Remove elements that contain sample-specific values.
      object@model$fitted.values <- NULL
      object@model$weights <- NULL
      object@model$residuals <- NULL
      
    } else if (inherits(object@model, "vglm")) {
      ..deprecation_vgam()
    }

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    # Default method for models that lack a more specific method.
    return(object)
  }
)



.get_available_glm_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c(
    "glm", "glm_logistic", "glm_probit", "glm_cauchy",
    "glm_log", "glm_loglog", "glm_multinomial", "glm_gaussian",
    "glm_log_gaussian", "glm_inv_gaussian", "glm_poisson",
    "glm_log_poisson"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("glm", "glm_log"))
  }
  
  return(learners)
}

# First test if all selectable learners are also available
familiar:::test_all_learners_available(
  learners = familiar:::.get_available_glm_learners(show_general = TRUE)
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_all_learners_train_predict_vimp(
  learners = familiar:::.get_available_glm_learners(show_general = FALSE)
)

familiar:::test_all_learners_parallel_train_predict_vimp(
  learners = familiar:::.get_available_glm_learners(show_general = FALSE)
)


# Continuous outcome tests------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("continuous")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(good_data)),
  learner = "glm_gaussian"
)

testthat::test_that("Generalised linear model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # That no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})


testthat::test_that("Generalised linear model has variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)

  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(good_data) %in% vimp_table$name)
  )

  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Binomial tests----------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("binomial")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(good_data)),
  learner = "glm_logistic"
)

testthat::test_that("Generalised linear model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # Check that no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})


testthat::test_that("Generalised linear model has variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)
  
  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(good_data) %in% vimp_table$name)
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Multinomial tests-------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("multinomial")

# Train the model using the good dataset.
good_model <- suppressWarnings(familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(good_data)),
  learner = "glm_multinomial"
))

testthat::test_that("Generalised linear model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # That no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning, "deprec")

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("Generalised linear model has variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)
  
  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(good_data) %in% vimp_table$name)
  )
  
  # Feature 1 is among the most important.
  testthat::expect_true("feature_1" %in% vimp_table[rank <= 2, ]$name)
})


# Survival tests----------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("survival")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(good_data)),
  time_max = 3.5,
  learner = "glm"
)

testthat::test_that("Generalised linear model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # That no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})


testthat::test_that("Generalised linear model has variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)
  
  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(good_data) %in% vimp_table$name)
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


familiar:::test_hyperparameter_optimisation(
  learners = "glm",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)


testthat::skip("Skip hyperparameter optimisation, unless manual.")

# Test hyperparameters
familiar:::test_hyperparameter_optimisation(
  learners = familiar:::.get_available_glm_learners(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test.
familiar:::test_plots(
  plot_function = familiar:::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  estimation_type = "point",
  debug = debug_flag
)

# Point line confidence intervals.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  estimation_type = "bci",
  confidence_level = 0.50,
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("multinomial"),
  plot_args = list("conf_int_style" = "point_line"),
  debug = debug_flag
)

# Line style confidence intervals.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  estimation_type = "bci",
  confidence_level = 0.50,
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("multinomial"),
  plot_args = list("conf_int_style" = "line"),
  debug = debug_flag
)

# Bar + line style confidence intervals.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  estimation_type = "bci",
  confidence_level = 0.50,
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("multinomial"),
  plot_args = list("conf_int_style" = "bar_line"),
  debug = debug_flag
)

# No confidence intervals.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  estimation_type = "bci",
  confidence_level = 0.50,
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("multinomial"),
  plot_args = list("conf_int_style" = "none"),
  debug = debug_flag
)

familiar:::test_plot_ordering(
  plot_function = familiar:::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  estimation_type = "point",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("multinomial"),
  plot_args = list(
    "facet_by" = c("data_set", "learner", "vimp_method"),
    "color_by" = c("metric", "similarity_threshold")),
  debug = debug_flag
)

# Different clustering method.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  feature_cluster_method = "pam",
  feature_cluster_cut_method = "silhouette",
  estimation_type = "point",
  metric = c("auc_roc", "accuracy"),
  outcome_type_available = c("multinomial"),
  debug = debug_flag
)

# With limiting the number of features during calculation.
familiar:::test_plots(
  plot_function = familiar:::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  test_config = "normal",
  n_important_features = 2L,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  estimation_type = "point",
  debug = debug_flag
)

# With limiting the number of features during plotting.
familiar:::test_plots(
  plot_function = familiar::plot_permutation_variable_importance,
  data_element = "permutation_vimp",
  test_config = "normal",
  plot_args = list("limit_n_features" = 2L),
  estimation_type = "point",
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarTaskLearnerHyperparameters -------------------------------------------
setClass(
  "familiarTaskLearnerHyperparameters",
  contains = "familiarTask",
  slots = list(
    "learner" = "character",
    "vimp_method" = "character",
    "use_vimp" = "character",
    "feature_info_file" = "character",
    "vimp_table_file" = "character"
  ),
  prototype = methods::prototype(
    learner = NA_character_,
    vimp_method = NA_character_,
    use_vimp = "use_hpo_vimp",
    feature_info_file = NA_character_,
    vimp_table_file = NA_character_,
    task_name = "set_learner_hyperparameters"
  )
)



# .set_file_name (learner hyperparameters task) --------------------------------
setMethod(
  ".set_file_name",
  signature(object = "familiarTaskLearnerHyperparameters"),
  function(object, file_paths = NULL) {
    if (is.null(file_paths)) return(object)
    
    # Generate file name of variable importance table
    object@file <- get_object_file_name(
      object_type = "hyperparametersLearner",
      data_id = object@data_id,
      run_id = object@run_id,
      learner = object@learner,
      vimp_method = object@vimp_method,
      project_id = object@project_id,
      dir_path = file_paths$mb_dir
    )
    
    return(object)
  }
)



# .get_task_descriptor (learner hyperparameters task) --------------------------
setMethod(
  ".get_task_descriptor",
  signature(object = "familiarTaskLearnerHyperparameters"),
  function(object, ...) {
    return(paste0(
      object@task_name, "_",
      object@data_id, "_",
      object@run_id, "_",
      object@vimp_method, "_",
      object@learner
    ))
  }
)



# .perform_task (learner hyperparameters task , NULL) --------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskLearnerHyperparameters",
    data = "NULL"
  ),
  function(
    object,
    data,
    experiment_data = NULL,
    outcome_info = NULL,
    ...
  ) {
    # This method is called when "data" is expected to be available somewhere in
    # the backend.
    
    if (is.null(experiment_data)) {
      ..error_reached_unreachable_code("experiment_data is required for retrieving data from the backend.")
    }
    if (is.null(outcome_info)) {
      ..error_reached_unreachable_code("outcome_info is required.")
    }
    
    # Find the run list.
    run_list <- .get_run_list(
      iteration_list = experiment_data@iteration_list,
      data_id = object@data_id,
      run_id = object@run_id
    )
    
    # Select unique samples.
    sample_identifiers <- .get_sample_identifiers(
      run = run_list,
      train_or_validate = "train"
    )
    sample_identifiers <- unique(sample_identifiers)
    
    # Create a dataObject.
    data <- methods::new(
      "dataObject",
      data = get_data_from_backend(sample_identifiers = sample_identifiers),
      preprocessing_level = "none",
      outcome_type = outcome_info@outcome_type,
      outcome_info = outcome_info
    )
    
    # Pass to method that dispatches with dataObject for further processing.
    return(.perform_task(
      object = object,
      data = data,
      experiment_data = experiment_data,
      ...
    ))
  }
)


# .perform_task (learner hyperparameters task, dataObject) ---------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskLearnerHyperparameters",
    data = "dataObject"
  ),
  function(
    object,
    data,
    settings = NULL,
    feature_info_list = NULL,
    vimp_table = NULL,
    vimp_aggregation_method = NULL,
    vimp_rank_threshold = NULL,
    hyperparameters = NULL,
    message_indent = 0L,
    verbose = FALSE,
    cl = NULL,
    return_results = TRUE,
    ...
  ) {

    # Check that outcome_info is present on data
    if (!is(data@outcome_info, "outcomeInfo")) {
      ..error_reached_unreachable_code(
        "outcome_info attribute of data (dataObject) does not contain an outcomeInfo object"
      )
    }
    
    # Check and retrieve feature info list.
    feature_info_list <- .get_feature_info_list(
      object = object,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings,
      message_indent = message_indent,
      verbose = verbose,
      cl = cl,
      ...
    )

    # Check and retrieve variable importances.
    if (object@use_vimp == "use_main_vimp") {
      vimp_table <- .get_variable_importance_table(
        object = object,
        vimp_table = vimp_table,
        feature_info_list = feature_info_list,
        data = data,
        settings = settings,
        message_indent = message_indent,
        verbose = verbose,
        cl = cl,
        ...
      )
      
    } else if (object@use_vimp %in% c("use_hpo_vimp", "return_hpo_vimp")) {
      vimp_table <- NULL
      
    } else {
      ..error_reached_unreachable_code(paste0("use_vimp attribute has an unrecognised "))
    }
    
    # Set vimp aggregation method and vimp_rank_threshold based on settings.
    if (!is.null(settings)) {
      if (is.null(vimp_aggregation_method)) {
        vimp_aggregation_method <- settings$vimp$aggregation
      }
      if (is.null(vimp_rank_threshold)) {
        vimp_rank_threshold <- settings$vimp$aggr_rank_threshold
      }
    }
    
    # Get user-provided hyperparameters.
    if (is.null(hyperparameters)) {
      hyperparameters <- settings$mb$hyper_param[[object@learner]]
      
    } else if (rlang::is_bare_list(hyperparameters)) {
      if (object@learner %in% names(hyperparameters)) {
        hyperparameters <- hyperparameters[[object@learner]]
      }
    }
    
    # Create a model object to set hyperparameters.
    hyperparameter_object <- promote_learner(
      object = methods::new(
        "familiarModel",
        outcome_type = data@outcome_type,
        hyperparameters = NULL,
        learner = object@learner,
        vimp_method = object@vimp_method,
        vimp_table = vimp_table,
        vimp_aggregation_method = vimp_aggregation_method,
        vimp_rank_threshold = vimp_rank_threshold,
        outcome_info = data@outcome_info,
        run_table = .get_current_run_table(object = object),
        settings = settings,
        project_id = object@project_id
      )
    )
    
    # Find required features.
    required_features <- get_required_features(
      x = feature_info_list,
      exclude_signature = FALSE
    )
    
    # Limit to required features. This removes signature features which are not
    # assessed through variable importance.
    feature_info_list <- feature_info_list[required_features]
    hyperparameter_object@required_features <- required_features
    hyperparameter_object@feature_info <- feature_info_list
    
    # Make sure the input data is processed.
    data <- process_input_data(
      object = hyperparameter_object,
      data = data
    )
    
    # Ensure that available data have associated outcome data.
    data <- filter_missing_outcome(data = data)
    
    logger_message(
      paste0(
        "Hyperparameter optimisation: Starting hyperparameter optimisation for the \"",
        object@learner, "\" learner with the \"",
        object@vimp_method, "\" variable importance method for run ",
        object@task_id, " of ",
        object@n_tasks, "."
      ),
      indent = message_indent,
      verbose = verbose
    )
    
    # Compute hyperparameters. Function arguments to optimise_hyperparameters
    # are passed from the calling function.
    hyperparameter_object <- optimise_hyperparameters(
      object = hyperparameter_object,
      data = data,
      user_list = hyperparameters,
      verbose = verbose,
      message_indent = message_indent + 1L,
      save_in_place = FALSE,
      is_vimp = FALSE,
      cl = cl,
      ...
    )
    
    if (object@use_vimp != "return_hpo_vimp") {
      # Variable importance should not be passed using the hyperparameter
      # object, which means that any subsequent learner task will use the cached
      # variable importance tables instead.
      hyperparameter_object@vimp_table <- NULL
    }
    
    # Set familiar version.
    hyperparameter_object <- add_package_version(hyperparameter_object)
    
    if (!is.na(object@file)) {
      saveRDS(hyperparameter_object, file = object@file)
    }
    
    if (return_results) {
      return(hyperparameter_object)
    }
    
    return(TRUE)
  }
)



..run_learner_computation_hyperparameters <- function(
    tasks,  
    settings,
    cl,
    message_indent = 0L,
    verbose,
    ...
) {
  logger_message(
    paste0(
      "Hyperparameter optimisation: Starting parameter optimisation for learners."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  # Determine how parallel processing takes place.
  if (settings$hpo$do_parallel %in% c("TRUE", "inner")) {
    cl_inner <- cl
    cl_outer <- NULL
    
  } else if (settings$hpo$do_parallel %in% c("outer")) {
    cl_inner <- NULL
    cl_outer <- cl
    
    if (!is.null(cl_outer)) {
      logger_message(
        paste0(
          "Hyperparameter optimisation: Load-balanced parallel processing ",
          "is done in the outer loop. No progress can be displayed."
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
  } else {
    cl_inner <- cl_outer <- NULL
  }
  
  # Iterate over data subsets for which parameters have not yet been set.
  fam_mapply_lb(
    cl = cl_outer,
    assign = "all",
    FUN = .perform_task,
    progress_bar = !is.null(cl_outer),
    object = tasks,
    MoreArgs = list(
      "cl" = cl_inner,
      "data" = NULL,
      "settings" = settings,
      "vimp_aggregation_method" = settings$vimp$aggregation,
      "vimp_rank_threshold" = settings$vimp$aggr_rank_threshold,
      "metric" = settings$hpo$hpo_metric,
      "hyperparameters" = settings$mb$param,
      "optimisation_function" = settings$hpo$hpo_optimisation_function,
      "acquisition_function" = settings$hpo$hpo_acquisition_function,
      "grid_initialisation_method" = settings$hpo$hpo_grid_initialisation_method,
      "n_random_sets" = settings$hpo$hpo_n_grid_initialisation_samples,
      "exploration_method" = settings$hpo$hpo_exploration_method,
      "determine_vimp" = settings$hpo$hpo_determine_vimp,
      "measure_time" = TRUE,
      "hyperparameter_learner" = settings$hpo$hpo_hyperparameter_learner,
      "n_max_bootstraps" = settings$hpo$hpo_max_bootstraps,
      "n_initial_bootstraps" = settings$hpo$hpo_initial_bootstraps,
      "n_intensify_step_bootstraps" = settings$hpo$hpo_bootstraps,
      "n_max_optimisation_steps" = settings$hpo$hpo_smbo_iter_max,
      "n_max_intensify_steps" = settings$hpo$hpo_intensify_max_iter,
      "intensify_stop_p_value" = settings$hpo$hpo_alpha,
      "convergence_tolerance" = settings$hpo$hpo_convergence_tolerance,
      "convergence_stopping" = settings$hpo$hpo_conv_stop,
      "time_limit" = settings$hpo$hpo_time_limit,
      "message_indent" = message_indent + 1L,
      "verbose" = verbose && is.null(cl_outer),
      "return_results" = FALSE,
      ...
    )
  )
  
  logger_message(
    paste0(
      "Hyperparameter optimisation: Completed parameter optimisation for learners.",
      "\n"
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  return(invisible(TRUE))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarRanger object --------------------------------------------------------
setClass(
  "familiarRanger",
  contains = "familiarModel"
)

# familiarRangerDefault object --------------------------------------------------
setClass(
  "familiarRangerDefault",
  contains = "familiarRanger"
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarRanger"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- "ranger"

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarRanger"),
  function(object, ...) {
    # Ranger exists for all outcome types and variable importance methods,
    # including impurity for survival.
    if (object@outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(TRUE)
  }
)



# get_default_hyperparameters (familiarRanger) ---------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarRanger"),
  function(object, data = NULL, user_list = NULL, ...) {
    # Initialise list and declare hyperparameter entries.
    param <- list()
    param$sign_size <- list()
    param$n_tree <- list()
    param$sample_size <- list()
    param$m_try <- list()
    param$node_size <- list()
    param$tree_depth <- list()
    param$split_rule <- list()
    param$alpha <- list()
    param$sample_weighting <- list()
    param$sample_weighting_beta <- list()

    # The following hyperparameters are only used for feature selection.
    param$fs_forest_type <- list()
    param$fs_vimp_method <- list()

    # If dt is not provided, return the list with hyperparameter /names only.
    if (is.null(data)) return(param)

    # Get the number of samples
    n_samples <- get_n_samples(data, "series")

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data)

    # number of trees ----------------------------------------------------------

    # Note that the number of trees is defined in powers of 2, based on Oshiro,
    # T. M., Perez, P. S., & Baranauskas, J. A. (2012, July). How many trees in
    # a random forest?. In MLDM (pp. 154-168).
    param$n_tree <- .set_hyperparameter(
      default = c(4L, 8L, 10L),
      type = "integer",
      range = c(4L, 10L),
      valid_range = c(0L, Inf),
      randomise = TRUE
    )

    # sample size --------------------------------------------------------------

    # Note that the sample size is here noted as a fraction, which corresponds
    # to the usage in ranger.
    param$sample_size <- .set_hyperparameter(
      default = c(0.30, 0.50, 0.70, 1.00),
      type = "numeric",
      range = c(2.0 / n_samples, 1.0),
      valid_range = c(0.0, 1.0),
      randomise = TRUE
    )

    # number of candidate features selected at node ----------------------------

    # Note that the number of features is here noted as a fraction, but is used
    # in ranger as an integer. Familiar ensures that always at least 1 feature
    # is available as a candidate.
    param$m_try <- .set_hyperparameter(
      default = c(0.1, 0.3, 0.5, 1.0),
      type = "numeric",
      range = c(0.0, 1.0),
      randomise = TRUE
    )

    # terminal node size -------------------------------------------------------

    # Number of instances in the terminal node. Larger terminal node sizes limit
    # tree depth and overfitting.

    # Define the default range.
    node_size_range <- c(5L, as.integer(floor(n_samples / 3.0)))

    # Define the default values.
    node_size_default <- c(5L, 10L, 20L, 50L)
    node_size_default <- node_size_default[
      node_size_default >= node_size_range[1L] &
        node_size_default <= node_size_range[2L]
    ]

    # Set the node_size parameter.
    param$node_size <- .set_hyperparameter(
      default = node_size_default,
      type = "integer",
      range = node_size_range,
      valid_range = c(1L, Inf),
      randomise = TRUE
    )

    # maximum tree depth -------------------------------------------------------

    # Determines the depth trees are allowed to grow to. Larger depths increase
    # the risk of overfitting.
    param$tree_depth <- .set_hyperparameter(
      default = c(1L, 2L, 3L, 7L),
      type = "integer",
      range = c(1L, 10L),
      valid_range = c(1L, Inf),
      randomise = TRUE
    )

    # splitting rule -----------------------------------------------------------

    # Availability of splitting rules is dependent on the type of outcome.
    if (object@outcome_type %in% c("binomial", "multinomial")) {
      split_rule_range <- c("gini", "extratrees", "hellinger")
      split_rule_default <- "gini"
    } else if (object@outcome_type %in% c("continuous")) {
      split_rule_range <- c("variance", "extratrees", "maxstat", "beta")
      split_rule_default <- "maxstat"
    } else if (object@outcome_type == "survival") {
      split_rule_range <- c("logrank", "extratrees", "C", "maxstat")
      split_rule_default <- "maxstat"
    } else {
      ..error_no_known_outcome_type(object@outcome_type)
    }

    # Set the split_rule parameter.
    param$split_rule <- .set_hyperparameter(
      default = split_rule_default,
      type = "factor",
      range = split_rule_range,
      randomise = FALSE
    )

    # significance threshold for splitting -------------------------------------

    # Sets the significance level required to allow a split on a variable. It is
    # only used with maxstat.
    alpha_randomise <- split_rule_default == "maxstat"

    if (alpha_randomise) {
      default_values <- c(0.05, 0.1, 0.5, 1.0)
    } else {
      default_values <- c(0.05)
    }

    # Set the alpha parameter
    param$alpha <- .set_hyperparameter(
      default = default_values,
      type = "numeric",
      range = c(10.0^-6.0, 1.0),
      valid_range = c(0.0, 1.0),
      randomise = alpha_randomise,
      distribution = "log"
    )

    # sample weighting method --------------------------------------------------
    
    # Class imbalances may lead to learning majority classes. This can be
    # partially mitigated by increasing weight of minority classes.
    param$sample_weighting <- .get_default_sample_weighting_method(
      outcome_type = object@outcome_type
    )

    # effective number of samples beta -----------------------------------------
    
    # Specifies the beta parameter for effective number sample weighting method.
    # See Cui et al. (2019).
    param$sample_weighting_beta <- .get_default_sample_weighting_beta(
      method = c(
        param$sample_weighting$init_config,
        user_list$sample_weighting
      ),
      outcome_type = object@outcome_type
    )


    # feature selection forest type --------------------------------------------

    # Enables the construction of holdout forests. A conventional forest is
    # grown by default.
    param$fs_forest_type <- .set_hyperparameter(
      default = "standard",
      type = "factor",
      range = c("standard", "holdout"),
      randomise = FALSE
    )

    # feature selection variable importance method -----------------------------

    # Enables the use of different variable importance methods. The permutation
    # method is used by default.
    param$fs_vimp_method <- .set_hyperparameter(
      default = "permutation",
      type = "factor",
      range = c("permutation", "impurity", "impurity_corrected"),
      randomise = FALSE
    )

    return(param)
  }
)



# get_default_hyperparameters (familiarRangerDefault) --------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarRangerDefault"),
  function(object, data = NULL, user_list = NULL, ...) {
    # Initialise list and declare hyperparameter entries.
    param <- list()
    param$sign_size <- list()
    param$sample_weighting <- list()
    param$sample_weighting_beta <- list()
    param$split_rule <- list()

    # The following hyperparameters are only used for feature selection.
    param$fs_forest_type <- list()
    param$fs_vimp_method <- list()

    # If dt is not provided, return the list with hyperparameter names only.
    if (is.null(data)) return(param)

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data)

    # sample weighting method --------------------------------------------------
    # Class imbalances may lead to learning majority classes. This can be
    # partially mitigated by increasing weight of minority classes.
    param$sample_weighting <- .get_default_sample_weighting_method(
      outcome_type = object@outcome_type
    )

    # effective number of samples beta -----------------------------------------
    # Specifies the beta parameter for effective number sample weighting method.
    # See Cui et al. (2019).
    param$sample_weighting_beta <- .get_default_sample_weighting_beta(
      method = c(
        param$sample_weighting$init_config,
        user_list$sample_weighting
      ),
      outcome_type = object@outcome_type
    )

    # splitting rule -----------------------------------------------------------
    
    # Availability of splitting rules is dependent on the type of outcome.
    if (object@outcome_type %in% c("binomial", "multinomial")) {
      split_rule_range <- c("gini", "extratrees", "hellinger")
      split_rule_default <- "gini"
    } else if (object@outcome_type %in% c("continuous")) {
      split_rule_range <- c("variance", "extratrees", "maxstat", "beta")
      split_rule_default <- "maxstat"
    } else if (object@outcome_type == "survival") {
      split_rule_range <- c("logrank", "extratrees", "C", "maxstat")
      split_rule_default <- "maxstat"
    } else {
      ..error_no_known_outcome_type(object@outcome_type)
    }
    
    # Set the split_rule parameter.
    param$split_rule <- .set_hyperparameter(
      default = split_rule_default,
      type = "factor",
      range = split_rule_range,
      randomise = FALSE
    )
    
    # feature selection forest type --------------------------------------------
    # Enables the construction of holdout forests. A conventional forest is
    # grown by default.
    param$fs_forest_type <- .set_hyperparameter(
      default = "standard",
      type = "factor",
      range = c("standard", "holdout"),
      randomise = FALSE
    )

    # feature selection variable importance method -----------------------------
    # Enables the use of different variable importance methods. The permutation
    # method is used by default.
    param$fs_vimp_method <- .set_hyperparameter(
      default = "permutation",
      type = "factor",
      range = c("permutation", "impurity", "impurity_corrected"),
      randomise = FALSE
    )

    return(param)
  }
)



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarRanger"),
  function(object, type = "default") {
    if (object@outcome_type == "survival") {
      if (type == "default") {
        return("cumulative_hazard")
      } else if (type %in% c("survival", "survival_probability")) {
        return("survival_probability")
      } else if (type %in% c("response", "cumulative_hazard")) {
        return("cumulative_hazard")
      } else {
        stop(paste0("Prediction type is not implemented: ", type))
      }
    } else {
      return(callNextMethod())
    }
  }
)


# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarRanger",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Aggregate repeated measurement data - ranger does not facilitate repeated
    # measurements.
    data <- aggregate_data(data = data)

    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object,
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Find feature columns in data table
    feature_columns <- get_feature_columns(x = data)

    # Parse formula.
    if (object@outcome_type == "survival") {
      formula <- stats::reformulate(
        termlabels = feature_columns,
        response = quote(survival::Surv(outcome_time, outcome_event))
      )
      
    } else if (object@outcome_type %in% c("binomial", "multinomial", "continuous")) {
      formula <- stats::reformulate(
        termlabels = feature_columns,
        response = quote(outcome)
      )
      
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }

    # Train probability trees for binomial and multinomial trees
    fit_probability <- object@outcome_type %in% c("binomial", "multinomial")

    # Extract hyperparameters from the model
    param <- object@hyperparameters

    # Set weights.
    weights <- create_instance_weights(
      data = data,
      method = object@hyperparameters$sample_weighting,
      beta = ..compute_effective_number_of_samples_beta(
        object@hyperparameters$sample_weighting_beta
      ),
      normalisation = "average_one"
    )

    # Get the arguments which are shared between holdout and standard forests.
    learner_arguments <- list(formula,
      "data" = data@data,
      "probability" = fit_probability,
      "num.threads" = 1L,
      "verbose" = FALSE,
      "splitrule" = as.character(param$split_rule)
    )

    if (!is(object, "familiarRangerDefault")) {
      # Non-default random forests have more arguments.
      learner_arguments <- c(
        learner_arguments,
        list(
          "num.trees" = 2.0^param$n_tree,
          "mtry" = max(c(1.0, ceiling(param$m_try * length(feature_columns)))),
          "min.node.size" = param$node_size,
          "max.depth" = param$tree_depth,
          "alpha" = param$alpha
        )
      )
    }

    # Create random forest using ranger.
    if (param$fs_forest_type == "standard") {
      # Conventional random forest (used for model building and variable
      # importance estimations)

      if (!is(object, "familiarRangerDefault")) {
        # Non-default random forests have more arguments.
        learner_arguments <- c(
          learner_arguments,
          list("sample.fraction" = param$sample_size)
        )
      }

      model <- do.call_with_handlers(
        ranger::ranger,
        args = c(
          learner_arguments,
          list(
            "case.weights" = weights,
            "importance" = as.character(param$fs_vimp_method)
          )
        )
      )
      
    } else if (param$fs_forest_type == "holdout") {
      # Hold-out random forest (used only for variable importance estimations).
      model <- do.call_with_handlers(
        ranger::holdoutRF,
        args = learner_arguments
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "..train,familiarRanger: encountered unknown forest type: ",
        param$fs_forest_type
      ))
    }

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) {
      return(callNextMethod(object = object))
    }

    # Add model
    object@model <- model

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarRanger",
    data = "dataObject"
  ),
  function(object, data, ...) {
    if (object@outcome_type %in% c("continuous", "binomial", "multinomial")) {
      # Turn into a Naive model.
      object <- methods::new("familiarNaiveModel", object)
      
    } else if (object@outcome_type %in% c("survival")) {
      # Turn into a Naive cumulative incidence model.
      object <- methods::new("familiarNaiveCumulativeIncidenceModel", object)
    }

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarRanger",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "default",
    time = NULL, 
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }
    
    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }
    
    if (type %in% c("default", "survival_probability")) {
      # Default method ---------------------------------------------------------

      # Make predictions using the model.
      if (inherits(object@model, "ranger")) {
        model_predictions <- suppressWarnings(predict(
          object = object@model,
          data = data@data,
          type = "response",
          num.threads = 1L,
          verbose = FALSE
        ))
        
      } else if (inherits(object@model, "holdoutRF")) {
        model_predictions <- suppressWarnings(predict(
          object = object@model$rf1,
          data = data@data,
          type = "response",
          num.threads = 1L,
          verbose = FALSE
        ))
        
      } else {
        ..error_reached_unreachable_code(paste0(
          "..predict,familiarRanger,dataObject: unknown model class detected ",
          class(object@model), ". Expected: ranger, holdoutRF"
        ))
      }

      if (object@outcome_type %in% c("binomial", "multinomial")) {
        # categorical outcomes -------------------------------------------------

        # Obtain class levels.
        class_levels <- colnames(model_predictions$predictions)

        # Add class probabilities.
        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          prediction_list[[class_levels[ii]]] <- model_predictions$predictions[, ii]
        }
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("continuous")) {
        # numerical outcomes ---------------------------------------------------

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = model_predictions$predictions,
          type = "regression",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("survival")) {
        # survival outcomes ----------------------------------------------------

        # Get the unique event times
        if (inherits(object@model, "ranger")) {
          event_times <- ranger::timepoints(object@model)
        } else {
          event_times <- ranger::timepoints(object@model$rf1)
        }

        # Set default time, if not provided.
        time <- ifelse(is.null(time), max(event_times), time)

        if (type == "default") {
          # Cumulative hazard.

          # Get the cumulative hazards at the given time point.
          prediction_table <- .random_forest_survival_predictions(
            object = object,
            event_matrix = model_predictions$chf,
            event_times = event_times,
            data = data,
            time = time,
            type = "cumulative_hazard"
          )
          
        } else if (type == "survival_probability") {
          # Survival probability.
          
          # Get the survival probability at the given time point.
          prediction_table <- .random_forest_survival_predictions(
            object = object,
            event_matrix = model_predictions$survival,
            event_times = event_times,
            data = data,
            time = time,
            type = "survival"
          )
        }
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      return(prediction_table)
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified method --------------------------------------------------

      # Make predictions using the model.
      return(predict(
        object = object@model,
        data = data@data,
        type = type,
        num.threads = 1L,
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



# ..get_prediction_table_type --------------------------------------------------
setMethod(
  "..get_prediction_table_type",
  signature(object = "familiarRanger"),
  function(object, type, ...) {
    prediction_table_type <- NULL
    if (object@outcome_type %in% c("survival") && type == "default") {
      prediction_table_type <- "cumulative_hazard"
    } else {
      prediction_table_type <- callNextMethod()
    }
    
    return(prediction_table_type)
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarRanger"),
  function(object, data = NULL, ...) {
    # Attempt to train the model if it has not been trained yet.
    if (!model_is_trained(object)) {
      object <- .train(
        object = object,
        data = data,
        get_additional_info = FALSE,
        trim_model = FALSE
      )
    }

    # Check if the model has been trained upon retry.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }

    # Check that required packages are loaded and installed.
    require_package(object, "vimp")

    # Extract the variable importance score
    if (object@hyperparameters$fs_forest_type == "standard") {
      vimp_score <- ranger::importance(object@model)
    } else {
      vimp_score <- object@model$variable.importance
    }

    # Create variable importance object.
    vimp_object <- methods::new(
      "vimpTable",
      vimp_table = data.table::data.table(
        "score" = vimp_score, 
        "name" = names(vimp_score)
      ),
      score_aggregation = "max",
      invert = TRUE
    )

    return(vimp_object)
  }
)



# ..set_calibration_info -------------------------------------------------------
setMethod(
  "..set_calibration_info", 
  signature(object = "familiarRanger"),
  function(object, data) {
    # Check if calibration info already.
    if (has_calibration_info(object)) {
      return(object)
    }

    if (object@outcome_type == "survival") {
      # Determine baseline survival.
      object@calibration_info <- get_baseline_survival(data = data)
    } else {
      return(callNextMethod())
    }

    return(object)
  }
)



# ..set_vimp_parameters --------------------------------------------------------
setMethod(
  "..set_vimp_parameters",
  signature(object = "familiarRanger"),
  function(object, method, ...) {
    # Determine variable importance method
    if (startswith_any(
      method,
      prefix = c(
        "random_forest_ranger_holdout_permutation",
        "random_forest_ranger_permutation"
      )
    )) {
      vimp_method <- "permutation"
      
    } else if (startswith_any(
      method,
      prefix = c("random_forest_ranger_impurity")
    )) {
      vimp_method <- "impurity_corrected"
    }

    # Determine forest type
    if (startswith_any(
      method, 
      prefix = c("random_forest_ranger_holdout_permutation")
    )) {
      forest_type <- "holdout"
      
    } else if (startswith_any(
      method,
      prefix = c(
        "random_forest_ranger_permutation",
        "random_forest_ranger_impurity"
      )
    )) {
      forest_type <- "standard"
    }

    if (is.null(object@hyperparameters)) {
      hyperparameters <- list()
    } else {
      hyperparameters <- object@hyperparameters
    }

    # Update the fs_vimp_method and fs_forest_type hyperparameters
    hyperparameters$fs_vimp_method <- vimp_method
    hyperparameters$fs_forest_type <- forest_type

    # Store in the object
    object@hyperparameters <- hyperparameters

    return(object)
  }
)



# .trim_model ------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarRanger"),
  function(object, ...) {
    # Update model by removing the call.
    object@model$call <- call("trimmed")

    # Add show.
    object <- .capture_show(object)

    # Remove the predictions.
    object@model$predictions <- NULL

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    return(object)
  }
)



.get_available_ranger_learners <- function(show_general = TRUE) {
  return("random_forest_ranger")
}



.get_available_ranger_default_learners <- function(show_general = TRUE) {
  return("random_forest_ranger_default")
}



.get_available_ranger_vimp_methods <- function(show_general = TRUE) {
  return(c(
    "random_forest_ranger_holdout_permutation",
    "random_forest_ranger_permutation",
    "random_forest_ranger_impurity"
  ))
}



.get_available_ranger_default_vimp_methods <- function(show_general = TRUE) {
  return(paste0(
    .get_available_ranger_vimp_methods(show_general = show_general),
    "_default"
  ))
}



.random_forest_survival_predictions <- function(
    object,
    event_matrix, 
    event_times, 
    data,
    time, 
    type
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  event_time <- NULL

  # Set id columns
  id_columns <- get_id_columns()
  
  # Convert event_matrix to a matrix.
  if (!is.matrix(event_matrix)) {
    event_matrix <- matrix(
      data = event_matrix, 
      ncol = length(event_matrix)
    )
  }
  
  # Combine with identifiers and cast to table.
  event_table <- cbind(
    data@data[, mget(id_columns)],
    data.table::as.data.table(event_matrix)
  )
  
  # Remove duplicate entries
  event_table <- unique(event_table, by = id_columns)
  
  # Melt to a long format.
  event_table <- data.table::melt(
    event_table,
    id.vars = id_columns,
    variable.name = "time_variable",
    value.name = "value"
  )
  
  # Create conversion table to convert temporary variables into the event times.
  conversion_table <- data.table::data.table(
    "time_variable" = paste0("V", seq_along(event_times)),
    "event_time" = event_times
  )
  
  # Add in event times
  event_table <- merge(
    x = event_table, 
    y = conversion_table, 
    by = "time_variable"
  )
  
  # Drop the time_variable column
  event_table[, "time_variable" := NULL]
  
  if (time %in% event_times) {
    # Get the event time directly.
    event_table <- event_table[event_time == time, ]
    
    # Remove event_time column and rename the value column to predicted_outcome.
    event_table[, "event_time" := NULL]
    data.table::setnames(x = event_table, old = "value", new = "predicted_outcome")
    
  } else {
    # Add starting values.
    if (!0.0 %in% event_times) {
      # Create initial table
      initial_event_table <- data.table::copy(event_table[event_time == event_times[1L]])
      
      # Update values
      if (type == "cumulative_hazard") {
        initial_event_table[, ":="("value" = 0.0, "event_time" = 0.0)]
        
      } else if (type == "survival") {
        initial_event_table[, ":="("value" = 1.0, "event_time" = 0.0)]
        
      } else {
        ..error_reached_unreachable_code(paste0(
          ".random_forest_survival_predictions: type was not recognised: ",
          type
        ))
      }
      
      # Combine with the event table.
      event_table <- rbind(initial_event_table, event_table)
    }
    
    # Now, interpolate at the given time point.
    event_table <- lapply(
      split(event_table, by = id_columns), 
      function(sample_table, time, id_columns) {
        # Interpolate values at the given time.
        value <- stats::approx(
          x = sample_table$event_time,
          y = sample_table$value,
          xout = time,
          rule = 2L
        )$y
        
        # Create an output table
        output_table <- data.table::copy(sample_table[1L, mget(id_columns)])
        output_table[, "predicted_outcome" := value]
        
        return(output_table)
      },
      time = time, 
      id_columns = id_columns
    )
    
    # Concatenate to single table.
    event_table <- data.table::rbindlist(event_table)
  }
  
  # Sort as original data.
  event_table <- merge(
    x = data@data[, mget(id_columns)],
    y = event_table, 
    by = id_columns,
    sort = FALSE
  )
  
  # Convert to prediction table objects.
  if (type == "cumulative_hazard") {
    prediction_table <- as_prediction_table(
      x = event_table$predicted_outcome,
      type = "cumulative_hazard",
      data = data,
      time = time,
      model_object = object
    )
    
  } else {
    prediction_table <- as_prediction_table(
      x = event_table$predicted_outcome,
      type = "survival_probability",
      data = data,
      time = time,
      model_object = object
    )
  }
  
  return(prediction_table)
}

familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_ranger_vimp_methods(show_general = TRUE))
familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_ranger_default_vimp_methods(show_general = TRUE))

# Don't perform any further tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_all_vimp_methods(
  familiar:::.get_available_ranger_vimp_methods(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list(
      "n_tree" = 4,
      "sample_size" = 1.00,
      "m_try" = 0.3,
      "node_size" = 5,
      "tree_depth" = 5,
      "alpha" = 0.1
    ),
    "binomial" = list(
      "n_tree" = 4,
      "sample_size" = 1.00,
      "m_try" = 0.3,
      "node_size" = 5,
      "tree_depth" = 5,
      "alpha" = 0.1
    ),
    "multinomial" = list(
      "n_tree" = 4,
      "sample_size" = 1.00,
      "m_try" = 0.3,
      "node_size" = 5,
      "tree_depth" = 5,
      "alpha" = 0.1
    ),
    "survival" = list(
      "n_tree" = 4,
      "sample_size" = 1.00,
      "m_try" = 0.3,
      "node_size" = 5,
      "tree_depth" = 5,
      "alpha" = 0.1
    )
  )
)

familiar:::test_all_vimp_methods(
  familiar:::.get_available_ranger_default_vimp_methods(show_general = FALSE))

# Parallel test.
familiar:::test_all_vimp_methods_parallel(
  familiar:::.get_available_ranger_vimp_methods(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list(
      "n_tree" = 4,
      "sample_size" = 1.00,
      "m_try" = 0.3,
      "node_size" = 5,
      "tree_depth" = 5,
      "alpha" = 0.1
    ),
    "binomial" = list(
      "n_tree" = 4,
      "sample_size" = 1.00,
      "m_try" = 0.3,
      "node_size" = 5,
      "tree_depth" = 5,
      "alpha" = 0.1
    ),
    "multinomial" = list(
      "n_tree" = 4,
      "sample_size" = 1.00,
      "m_try" = 0.3,
      "node_size" = 5,
      "tree_depth" = 5,
      "alpha" = 0.1
    ),
    "survival" = list(
      "n_tree" = 4,
      "sample_size" = 1.00,
      "m_try" = 0.3,
      "node_size" = 5,
      "tree_depth" = 5,
      "alpha" = 0.1
    )
  )
)

# Continuous outcome -----------------------------------------------------------
data <- familiar:::test_create_good_data("continuous")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_impurity",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "continuous",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0("The ranger random forest impurity method correctly ranks continuous data."),
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_permutation",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "continuous",
  cluster_method = "none",
  imputation_method = "simple")


testthat::test_that(
  paste0("The ranger random forest permutation method correctly ranks continuous data."), 
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_holdout_permutation",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "continuous",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0("The ranger random forest hold-out permutation method correctly ranks continuous data."), 
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Binomial outcome -------------------------------------------------------------
data <- familiar:::test_create_good_data("binomial")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_impurity",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "binomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0("The ranger random forest impurity method correctly ranks binomial data."),
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_permutation",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "binomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0("The ranger random forest permutation method correctly ranks binomial data."), {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_holdout_permutation",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "binomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0("The ranger random forest hold-out permutation method correctly ranks binomial data."),
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Multinomial outcome ----------------------------------------------------------
data <- familiar:::test_create_good_data("multinomial")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_impurity",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "multinomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0("The ranger random forest impurity method correctly ranks multinomial outcome data."), {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_permutation",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "multinomial",
  cluster_method = "none",
  imputation_method = "simple")


testthat::test_that(
  paste0("The ranger random forest permutation method correctly ranks multinomial outcome data."),
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_holdout_permutation",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "multinomial",
  cluster_method = "none",
  imputation_method = "simple")


testthat::test_that(
  paste0(
    "The ranger random forest hold-out permutation method ",
    "correctly ranks multinomial outcome data."), 
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Survival outcome -------------------------------------------------------------
data <- familiar:::test_create_good_data("survival")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_impurity",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "survival",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0("The ranger random forest impurity method correctly ranks survival outcome data."),
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_permutation",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "survival",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0("The ranger random forest permutation method correctly ranks survival outcome data."), 
  {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "random_forest_ranger_holdout_permutation",
  vimp_method_parameter_list = list(
    "n_tree" = 4,
    "sample_size" = 1.00,
    "m_try" = 0.3,
    "node_size" = 5,
    "tree_depth" = 5,
    "alpha" = 0.1),
  outcome_type = "survival",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(
  paste0(
    "The ranger random forest hold-out permutation method correctly ranks ",
    "survival outcome data."
  ), {
    vimp_table <- suppressWarnings(familiar:::get_vimp_table(
      familiar:::.vimp(vimp_object, data)))
    
    # Expect that the vimp table has at most six rows.
    testthat::expect_lte(nrow(vimp_table), 6L)
    
    # Expect that the names are the same as that of the features.
    testthat::expect_true(
      all(vimp_table$name %in% familiar:::get_feature_columns(data)))
    
    # Feature 1 is most important.
    testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
  }
)


familiar:::test_hyperparameter_optimisation(
  vimp_methods = "random_forest_ranger_impurity",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)



testthat::skip("Skip hyperparameter optimisation, unless manual.")

familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_ranger_vimp_methods(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE)

.replace_illegal_column_name <- function(column_name) {
  # Remove spaces
  column_name <- gsub(pattern = " ", replacement = "_", column_name)

  # Remove less/equal/greater than signs
  column_name <- gsub(pattern = ">=", replacement = "_geq_", fixed = TRUE, column_name)
  column_name <- gsub(pattern = "<=", replacement = "_leq_", fixed = TRUE, column_name)
  column_name <- gsub(pattern = "!=", replacement = "_neq_", fixed = TRUE, column_name)
  column_name <- gsub(pattern = "<", replacement = "_l_", fixed = TRUE, column_name)
  column_name <- gsub(pattern = ">", replacement = "_g_", fixed = TRUE, column_name)
  column_name <- gsub(pattern = "=", replacement = "_eq_", fixed = TRUE, column_name)

  # Remove punctuation
  column_name <- gsub(pattern = "[[:punct:]]", replacement = "_", column_name)

  # Remove starting number
  column_name <- gsub(pattern = "^([0-9])", replacement = "n_\\1", column_name)

  # Replace illegal strings that will cause issues for stats::reformulate.
  illegal_names <- column_name %in% c("else", "for", "function", "if", "in", "while")
  if (any(illegal_names)) column_name[illegal_names] <- paste0(column_name[illegal_names], "_")
  
  return(column_name)
}



paste_s <- function(...) {
  # Function to collapse a series of strings into a summation in the form
  # "element_1, element_2, ..., and element_n".
  dots <- c(...)

  if (length(dots) > 2L) {
    # For more than 2 elements, split into an initial and final section.
    initial_string <- paste0(
      head(dots, n = length(dots) - 2L),
      collapse = ", "
    )

    final_string <- paste0(
      tail(dots, n = 2L), 
      collapse = " and "
    )

    return(paste0(c(initial_string, final_string), collapse = ", "))
    
  } else if (length(dots) == 2L) {
    # For exactly 2 elements, combine with "and".
    return(paste0(dots, collapse = " and "))
    
  } else {
    # For only one element, return as is.
    return(paste0(dots))
  }
}



paste_sh <- function(..., n = 6L) {
  # Function to collapse the first few elements of a series of strings into a
  # summation of the form: "element_1, element_2, element_3, ..."
  
  dots <- c(...)
  
  # Check if all elements can be collapsed.
  if (length(dots) <= n) return(paste_s(...))
  
  initial_string <- paste0(head(dots, n = n), collapse = ", ")
  
  return(paste0(initial_string, ", ..."))
}



paste_c <- function(x, y, collapse = NULL) {
  # Create combinations of strings in x and y.
  str_combinations <- unlist(lapply(x, paste0, y, collapse = collapse))

  return(str_combinations)
}




rstring <- function(n = 1L, character_set = "alphanumeric") {
  # Initialise the available set.
  available_characters <- NULL

  # Sanity check on n.
  if (n < 1L) stop("n cannot be smaller than 1.")

  # Uppercase characters
  if (character_set %in% c("uppercase", "alphanumeric", "letters")) {
    available_characters <- c(available_characters, LETTERS)
  }

  # Lowercase characters
  if (character_set %in% c("lowercase", "alphanumeric", "letters")) {
    available_characters <- c(available_characters, letters)
  }

  # Numbers
  if (character_set %in% (c("numbers", "numeric", "alphanumeric"))) {
    available_characters <- c(
      available_characters,
      "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
    )
  }

  # Draw random indices and convert to integer values by rounding up.
  random_indices <- as.integer(ceiling(stats::runif(
    n = n,
    min = 0L, 
    max = length(available_characters)
  )))

  # Replaces any 0 (which is improbable, but could happen).
  random_indices[random_indices == 0L] <- 1L

  return(paste0(
    available_characters[random_indices],
    collapse = ""
  ))
}



startswith_any <- function(x, prefix) {
  # Check that x starts with any of the prefixes. x can be multiple strings.
  return(sapply(
    x,
    function(x, prefix) {
      any(startsWith(x = x, prefix = prefix))
    },
    prefix = prefix,
    USE.NAMES = FALSE
  ))
}



endswith_any <- function(x, suffix) {
  # Check that x ends with any of the suffixes. x can be multiple strings.
  return(sapply(
    x,
    function(x, suffix) {
      any(endsWith(x = x, suffix = suffix))
    },
    suffix = suffix,
    USE.NAMES = FALSE
  ))
}



sub_all_patterns <- function(pattern, replacement, x, ...) {
  # Replace the first instance of multiple patterns in x with the same
  # replacement.
  for (current_pattern in pattern) {
    x <- sub(
      x = x, 
      replacement = replacement, 
      pattern = current_pattern, 
      ...
    )
  }

  return(x)
}



gsub_all_patterns <- function(pattern, replacement, x, ...) {
  # Replace all instances of multiple patterns in x with the same replacement.
  for (current_pattern in pattern) {
    x <- gsub(
      x = x, 
      replacement = replacement, 
      pattern = current_pattern, 
      ...
    )
  }

  return(x)
}



sub_last <- function(pattern, replacement, x, ...) {
  # Dispatch to underlying function for individual strings.
  return(sapply(
    x, 
    .sub_last, 
    pattern = pattern, 
    replacement = replacement, 
    ..., 
    USE.NAMES = FALSE
  ))
}



.sub_last <- function(x, pattern, replacement, ...) {
  # Replace last instance of a pattern in x.
  instances <- gregexpr(pattern = pattern, text = x, ...)[[1L]]

  # Skip if pattern has not been found.
  if (all(instances == -1L)) return(x)

  # Select last instance.
  instances <- tail(instances, n = 1L)

  # Replace the pattern with its replacement.
  initial_string <- substr(x, start = 1L, instances - 1L)
  final_string <- substr(x, start = instances + nchar(pattern), stop = nchar(x))

  return(paste0(
    initial_string, 
    replacement, 
    final_string, 
    collapse = ""
  ))
}



strsplit_all <- function(x, split, ...) {
  # Iterate over list or vector.
  y <- lapply(
    x, 
    function(x, split, ...) {
      return(strsplit(
        x = x, 
        split = split,
        ...
      )[[1L]])
    },
    split = split,
    ...
  )

  return(y)
}

.create_hyperparameter_challenger_sets <- function(
    score_table,
    parameter_list,
    parameter_table,
    time_optimisation_model,
    score_optimisation_model,
    acquisition_function,
    n_max_bootstraps,
    n_challengers = 20L,
    smbo_iter,
    measure_time,
    random_admixture_fraction = 0.20
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  param_id <- expected_train_time <- utility <- summary_score <- weight <- NULL
  is_observed <- .NATURAL <- i.param_id <- NULL

  # Set the maximum number of local steps based on the number of randomisable
  # hyperparameters.
  n_max_local_steps <- 2L * sum(sapply(parameter_list, function(x) (x$randomise)))

  # Train hyperparameter optimisation model.
  if (!model_is_trained(score_optimisation_model)) {
    score_optimisation_model <- .train(
      object = score_optimisation_model,
      data = score_table,
      parameter_data = parameter_table
    )
  }

  # Find information regarding the dataset that has the highest
  # optimisation score.
  incumbent_set <- get_best_hyperparameter_set(
    score_table = score_table,
    parameter_table = parameter_table,
    optimisation_model = score_optimisation_model,
    n_max_bootstraps = n_max_bootstraps,
    n = 1L
  )

  # Generate random sets--------------------------------------------------------

  # Increase random admixture for random search.
  if (is(score_optimisation_model, "familiarHyperparameterRandomSearch")) {
    random_admixture_fraction <- 1.0
  }

  # In general, draw about 10 times as much random sets then required.
  n_random_sets <- ceiling(n_challengers * random_admixture_fraction)

  # Generate random sets
  random_sets <- lapply(
    seq_len(n_random_sets * 10L),
    function(ii, parameter_list) {
      # Create random configuration.
      return(..randomise_hyperparameter_set(
        parameter_list = parameter_list,
        local = FALSE
      ))
    },
    parameter_list = parameter_list
  )

  # Combine random sets into a data.table.
  random_sets <- unique(data.table::rbindlist(random_sets, use.names = TRUE))

  # Compute expected improvement.
  if (
    !is_empty(random_sets) &&
    !is(score_optimisation_model, "familiarHyperparameterRandomSearch")
  ) {
    random_sets[, "utility" := .compute_utility_value(
      parameter_set = .SD,
      score_model = score_optimisation_model,
      acquisition_data = incumbent_set,
      acquisition_function = acquisition_function
    )]
  }

  # Compute expect train time.
  if (!is_empty(random_sets)) {
    random_sets[, "expected_train_time" := .compute_expected_train_time(
      parameter_set = .SD,
      time_model = time_optimisation_model
    )]
  }

  # Identify local sets --------------------------------------------------------
  n_local_sets <- ceiling(n_challengers * (1.0 - random_admixture_fraction))

  # Find local sets
  local_sets <- list()

  if (
    n_local_sets > 0L &&
    !is(score_optimisation_model, "familiarHyperparameterRandomSearch")
  ) {
    # Create a table of summary scores to identify the best sets.
    summary_score_table <- .compute_hyperparameter_summary_score(
      score_table = score_table,
      parameter_table = parameter_table,
      optimisation_model = score_optimisation_model
    )

    # Find the most promising samples by comparing against the median.
    summary_score_table[, "weight" := summary_score - stats::median(summary_score)]

    # If all parameters have the same optimisation score (an edge case), assign
    # weight 1 to each, that is, all hyperparameter sets are as likely to be
    # sampled.
    if (all(summary_score_table$weight == 0.0)) {
      summary_score_table[, "weight" := 1.0]
    }

    # Remove 0 and negative weights.
    summary_score_table <- summary_score_table[weight > 0.0]

    # Select the hyperparameter ids to sample, starting with the most promising
    # samples first.
    summary_score_table <- summary_score_table[, "n_select" := round(
      2.0 * n_local_sets * weight^2.0 / sum(weight^2.0)
    )][order(-summary_score)]
    
    selected_best_parameter_id <- unlist(mapply(
      function(x, n) (rep(x, times = n)),
      x = summary_score_table$param_id,
      n = summary_score_table$n_select
    ))

    # Perform a local search.
    for (best_param_id in selected_best_parameter_id) {
      # Initialise iterator variable
      iter_local_step <- 0L

      # Select initial configuration for local search
      selected_set <- parameter_table[param_id == best_param_id, ]

      while (iter_local_step < n_max_local_steps) {
        # Randomise configuration
        local_random_set <- ..randomise_hyperparameter_set(
          parameter_table = selected_set,
          parameter_list = parameter_list,
          local = TRUE
        )

        # Add configuration to list
        local_sets <- c(local_sets, list(local_random_set))

        # Continue search from current set
        selected_set <- local_random_set

        # Update iterator
        iter_local_step <- iter_local_step + 1L
      }
    }
  }

  # Combine local sets.
  local_sets <- unique(data.table::rbindlist(local_sets))

  if (!is_empty(local_sets)) {
    # Compute expected improvement.
    local_sets[, "utility" := .compute_utility_value(
      parameter_set = .SD,
      score_model = score_optimisation_model,
      acquisition_data = incumbent_set,
      acquisition_function = acquisition_function
    )]

    # Compute expected train time.
    local_sets[, "expected_train_time" := .compute_expected_train_time(
      parameter_set = .SD,
      time_model = time_optimisation_model
    )]
  }


  # Select challengers ---------------------------------------------------------

  local_challenger_sets <- list()
  random_challenger_sets <- list()

  if (is(score_optimisation_model, "familiarHyperparameterRandomSearch")) {
    # Preferentially select in the following order:
    # 1. Random sets that were not observed before.
    # 2. Random sets that were observed before.

    # Remove incumbent set.
    if (!is_empty(random_sets)) {
      random_sets <- random_sets[
        !parameter_table[param_id == incumbent_set$param_id],
        on = .NATURAL
      ]
    }

    # Select only parameter sets that can be evaluated within the allowed time.
    if (measure_time && !is_empty(random_sets)) {
      random_sets <- random_sets[expected_train_time < incumbent_set$max_time]
    }

    # Consider random sets.
    if (!is_empty(random_sets)) {
      # Set observed status
      random_sets[, "is_observed" := FALSE]
      random_sets[parameter_table, "is_observed" := TRUE, on = .NATURAL]

      # Random sets that were not observed before.
      if (!all(random_sets$is_observed) && n_random_sets > 0L) {
        # Select up to n_random_sets of random samples sets that were not
        # selected previously.
        selected_sets <- head(random_sets[is_observed == FALSE], n = n_random_sets)

        # Add to challenger sets and reduce the number of challengers to select.
        random_challenger_sets <- list(selected_sets)
        n_challengers <- n_challengers - nrow(selected_sets)
        n_random_sets <- min(c(n_random_sets - nrow(selected_sets), n_challengers))
      }

      # Random sets that were observed before.
      if (any(random_sets$is_observed) && n_random_sets > 0L) {
        # Select up to n_random_sets of random samples sets that were observed
        # selected previously.
        selected_sets <- head(random_sets[is_observed == TRUE], n = n_random_sets)

        # Add to challenger sets and reduce the number of challengers to select.
        random_challenger_sets <- c(random_challenger_sets, list(selected_sets))
        n_challengers <- n_challengers - nrow(selected_sets)
      }
    }

    if (!is_empty(random_challenger_sets)) {
      # Concatenate challenger sets.
      challenger_sets <- data.table::rbindlist(
        random_challenger_sets, 
        use.names = TRUE
      )

      # Remove expected time from challenger sets.
      challenger_sets[, ":="(
        "expected_train_time" = NULL,
        "is_observed" = NULL
      )]
      
    } else {
      challenger_sets <- NULL
    }
    
  } else {
    # Preferentially select in the following order:
    # 1. Local sets that were not observed before, with highest utility first.
    # 2. Local sets that were observed before, with highest utility first.
    # 3. Random sets that were not observed before, with highest utility first.
    # 4. Random sets that were observed before, with highest utility first.

    # First, remove incumbent.
    if (!is_empty(local_sets)) {
      local_sets <- local_sets[
        !parameter_table[param_id == incumbent_set$param_id],
        on = .NATURAL
      ]
    }

    # Select only parameter sets that can be evaluated within the allowed time.
    if (measure_time && !is_empty(local_sets)) {
      local_sets <- local_sets[expected_train_time < incumbent_set$max_time]
    }

    # Consider local sets.
    if (!is_empty(local_sets)) {
      # Set observed status
      local_sets[, "is_observed" := FALSE]
      local_sets[parameter_table, "is_observed" := TRUE, on = .NATURAL]

      # Local sets that were not observed before, with highest utility first.
      if (!all(local_sets$is_observed)) {
        # Select up to n_local_sets of locally sampled sets that were not
        # selected previously.
        selected_sets <- head(
          local_sets[is_observed == FALSE][order(-utility)],
          n = n_local_sets
        )

        # Add to challenger sets and reduce the number of challengers to select.
        local_challenger_sets <- list(selected_sets)
        n_challengers <- n_challengers - nrow(selected_sets)
        n_local_sets <- min(c(n_local_sets - nrow(selected_sets), n_challengers))
      }

      # Local sets that were observed before, with highest utility first.
      if (any(local_sets$is_observed) && n_local_sets > 0L) {
        # Select up to n_local_sets of locally samples sets that were observed
        # selected previously.
        selected_sets <- head(
          local_sets[is_observed == TRUE][order(-utility)],
          n = n_local_sets
        )

        # Add to challenger sets and reduce the number of challengers to select.
        local_challenger_sets <- c(local_challenger_sets, list(selected_sets))
        n_challengers <- n_challengers - nrow(selected_sets)
      }
    }

    if (!is_empty(local_challenger_sets)) {
      # Concatenate challenger sets.
      local_challenger_sets <- data.table::rbindlist(
        local_challenger_sets,
        use.names = TRUE
      )

      # Remove expected improvement and time from challenger sets.
      local_challenger_sets[, ":="(
        "utility" = NULL,
        "expected_train_time" = NULL,
        "is_observed" = NULL
      )]
    }
    
    # Remove incumbent and already selected challenger sets.
    if (!is_empty(random_sets)) {
      random_sets <- random_sets[
        !parameter_table[param_id == incumbent_set$param_id],
        on = .NATURAL
      ]
    }
    
    if (!is_empty(random_sets) && !is_empty(local_challenger_sets) > 0L) {
      random_sets <- random_sets[
        !local_challenger_sets,
        on = .NATURAL
      ]
    }

    # Select only parameter sets that can be evaluated within the allowed time.
    if (measure_time && !is_empty(random_sets)) {
      random_sets <- random_sets[expected_train_time < incumbent_set$max_time]
    }

    # Update n_random_sets to compensate for selected challengers.
    n_random_sets <- min(c(n_challengers, n_random_sets))

    # Consider random sets.
    if (!is_empty(random_sets)) {
      # Set observed status
      random_sets[, "is_observed" := FALSE]
      random_sets[parameter_table, "is_observed" := TRUE, on = .NATURAL]

      # Random sets that were not observed before, with highest utility first.
      if (!all(random_sets$is_observed) && n_random_sets > 0L) {
        # Select up to n_random_sets of random samples sets that were not
        # selected previously.
        selected_sets <- head(
          random_sets[is_observed == FALSE][order(-utility)],
          n = n_random_sets
        )

        # Add to challenger sets and reduce the number of challengers to select.
        random_challenger_sets <- list(selected_sets)
        n_challengers <- n_challengers - nrow(selected_sets)
        n_random_sets <- min(c(n_random_sets - nrow(selected_sets), n_challengers))
      }

      # Random sets that were observed before, with highest utility first.
      if (any(random_sets$is_observed) && n_random_sets > 0L) {
        # Select up to n_random_sets of random samples sets that were observed
        # selected previously.
        selected_sets <- head(
          random_sets[is_observed == TRUE][order(-utility)],
          n = n_random_sets
        )

        # Add to challenger sets and reduce the number of challengers to select.
        random_challenger_sets <- c(random_challenger_sets, list(selected_sets))
        n_challengers <- n_challengers - nrow(selected_sets)
      }
    }

    if (!is_empty(random_challenger_sets)) {
      # Concatenate challenger sets.
      random_challenger_sets <- data.table::rbindlist(
        random_challenger_sets, 
        use.names = TRUE
      )

      # Remove expected improvement and time from challenger sets.
      random_challenger_sets[, ":="(
        "utility" = NULL,
        "expected_train_time" = NULL,
        "is_observed" = NULL
      )]
    }

    # Combine to challenger sets.
    challenger_sets <- data.table::rbindlist(
      list(local_challenger_sets, random_challenger_sets),
      use.names = TRUE
    )
  }

  # Add parameter identifiers for known sets.
  if (!is_empty(challenger_sets)) {
    # Left-join with parameter table to add in known parameters
    challenger_sets <- challenger_sets[
      parameter_table,
      "param_id" := i.param_id,
      on = .NATURAL
    ]

    # Fill out missing parameter ids.
    if (anyNA(challenger_sets$param_id)) {
      # Find the largest existing parameter id.
      max_param_id <- max(parameter_table$param_id)

      # Generate parameter ids for new parameter sets.
      challenger_sets[is.na(param_id), "param_id" := max_param_id + .I]
    }
    
  } else {
    challenger_sets <- NULL
  }

  return(challenger_sets)
}



.create_hyperparameter_time_optimisation_model <- function(
    hyperparameter_learner = "random_forest",
    score_table,
    parameter_table,
    optimisation_function
) {
  
  # Check that score_table contains optimisation scores.
  if (!"optimisation_score" %in% colnames(score_table)) {
    score_table <- .compute_hyperparameter_optimisation_score(
      score_table = score_table,
      optimisation_function = optimisation_function
    )
  }

  # Train model to predict process time.
  model <- ..hyperparameter_random_forest_time_learner(
    score_table = score_table,
    parameter_table = parameter_table
  )

  return(model)
}



..hyperparameter_random_forest_time_learner <- function(
    score_table,
    parameter_table
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  time_taken <- NULL

  # Check if package is installed
  require_package(
    x = "ranger",
    purpose = "to predict process time of hyperparameter sets"
  )

  # Fill NA values with large, but finite values. Determine the maximum time.
  if (!all(is.na(score_table$time_taken))) {
    max_time_taken <- max(score_table$time_taken, na.rm = TRUE)
    
  } else {
    # Set max time to an improbable 86400 seconds, in case all time values are
    # NA. A single model trained in familiar should never take a day to finish.
    max_time_taken <- 86400.0
  }

  # Merge score and parameter data tables on param id.
  joint_table <- merge(
    x = score_table,
    y = parameter_table,
    by = "param_id",
    all = FALSE
  )

  # Replace NA entries with the minimum optimisation score.
  joint_table[is.na(time_taken), time_taken := 10.0 * max_time_taken]

  # Get parameter names.
  parameter_names <- setdiff(
    colnames(parameter_table),
    c("param_id", "run_id", "optimisation_score", "time_taken")
  )

  # Hyperparameters for the random forest.
  n_tree <- 400L
  n_train <- nrow(joint_table)
  sample_fraction <- max(c(0.3, min(c(1.0, 1.0 / (0.025 * n_train)))))

  # Parse formula.
  formula <- stats::reformulate(
    termlabels = parameter_names,
    response = "time_taken"
  )

  # Train random forest.
  rf_model <- ranger::ranger(formula,
    data = joint_table,
    num.trees = n_tree,
    num.threads = 1L,
    sample.fraction = sample_fraction,
    verbose = FALSE
  )

  return(rf_model)
}



.compute_utility_value <- function(
    parameter_set,
    score_model,
    acquisition_data,
    acquisition_function
) {
  # Check that the score_model is trained, and return NA if not.
  if (!model_is_trained(score_model)) {
    return(rep_len(NA_real_, nrow(parameter_set)))
  }

  # Dispatch to acquisition functions.
  acquisition_fun <- switch(
    acquisition_function,
    "improvement_probability" = acquisition_improvement_probability,
    "improvement_empirical_probability" = acquisition_improvement_empirical_probability,
    "expected_improvement" = acquisition_expected_improvement,
    "upper_confidence_bound" = acquisition_upper_confidence_bound,
    "bayes_upper_confidence_bound" = acquisition_bayes_upper_confidence_bound
  )
  
  utility_scores <- acquisition_fun(
    object = score_model,
    parameter_data = parameter_set,
    acquisition_data = acquisition_data
  )

  return(utility_scores)
}



acquisition_improvement_probability <- function(
    object,
    parameter_data,
    acquisition_data
) {
  # The definition of probability of improvement is found in Shahriari, B.,
  # Swersky, K., Wang, Z., Adams, R. P. & de Freitas, N. Taking the Human Out of
  # the Loop: A Review of Bayesian Optimization. Proc. IEEE 104, 148–175 (2016)

  # Get the score estimate from the incumbent hyperparameter set.
  tau <- acquisition_data$score_estimate

  # Predict the mean and standard deviation for the parameter sets in parameter
  # data.
  prediction_table <- .predict(
    object = object,
    data = parameter_data,
    type = "sd"
  )

  # Compute the probability of improvement.
  alpha <- stats::pnorm((prediction_table$mu - tau) / prediction_table$sigma)

  return(alpha)
}



acquisition_improvement_empirical_probability <- function(
    object,
    parameter_data,
    acquisition_data
) {
  # The definition of probability of improvement is found in Shahriari, B.,
  # Swersky, K., Wang, Z., Adams, R. P. & de Freitas, N. Taking the Human Out of
  # the Loop: A Review of Bayesian Optimization. Proc. IEEE 104, 148–175 (2016)

  # Get the incumbent score.
  tau <- acquisition_data$score_estimate

  if ("raw" %in% get_prediction_type(object)) {
    # Compute the empirical probability of improvement.

    # Define helper function.
    .acquisition_improvement_empirical_probability <- function(x, tau) {
      return(sum(x > tau) / length(x))
    }

    # Predict raw data for the hyperparameter sets.
    prediction_table <- .predict(
      object = object,
      data = parameter_data,
      type = "raw"
    )

    # Drop parameter id. This should leave only raw data.
    prediction_table[, "param_id" := NULL]

    # Compute utility.
    alpha <- apply(
      prediction_table,
      MARGIN = 1L,
      .acquisition_improvement_empirical_probability,
      tau = tau
    )
    
  } else {
    # Compute the stochastic probability of improvement.

    # Predict the mean and standard deviation for the parameter sets in
    # parameter data.
    prediction_table <- .predict(
      object = object,
      data = parameter_data,
      type = "sd"
    )

    # Stochastic values
    alpha <- stats::pnorm((prediction_table$mu - tau) / prediction_table$sigma)
  }

  return(alpha)
}



acquisition_expected_improvement <- function(
    object,
    parameter_data,
    acquisition_data
) {
  # The definition of expected improvement is found in Shahriari, B., Swersky,
  # K., Wang, Z., Adams, R. P. & de Freitas, N. Taking the Human Out of the
  # Loop: A Review of Bayesian Optimization. Proc. IEEE 104, 148–175 (2016).

  # Get the incumbent score.
  tau <- acquisition_data$score_estimate

  # Predict the mean and standard deviation for the parameter sets in parameter
  # data.
  prediction_table <- .predict(
    object = object,
    data = parameter_data,
    type = "sd"
  )

  # Compute a z-score.
  z <- (prediction_table$mu - tau) / prediction_table$sigma

  # Compute the expected improvement.
  alpha <- (prediction_table$mu - tau) * stats::pnorm(z) +
    prediction_table$sigma * stats::dnorm(z)

  return(alpha)
}



acquisition_upper_confidence_bound <- function(
    object,
    parameter_data,
    acquisition_data
) {
  # The definition of upper confidence bound is adapted from Srinivas, N.,
  # Krause, A., Kakade, S. M. & Seeger, M. W. Information-Theoretic Regret
  # Bounds for Gaussian Process Optimization in the Bandit Setting. IEEE
  # Trans. Inf. Theory 58, 3250–3265 (2012).

  # Find the number hyperparameters.
  n_parameters <- length(object@target_hyperparameters)

  # Predict the mean and standard deviation for the parameter sets in parameter
  # data.
  prediction_table <- .predict(
    object = object,
    data = parameter_data,
    type = "sd"
  )

  # Compute the beta parameter.
  beta <- 2.0 / 5.0 * log(5.0 / 3.0 * n_parameters * (acquisition_data$t + 1.0)^2.0 * pi^2.0)

  # Compute alpha
  alpha <- prediction_table$mu + beta * prediction_table$sigma

  return(alpha)
}



acquisition_bayes_upper_confidence_bound <- function(
    object,
    parameter_data,
    acquisition_data
) {
  # The definition of the Bayesian upper confidence bound is adapted from
  # Kaufmann, E., Cappé, O. & Garivier, A. On Bayesian upper confidence bounds
  # for bandit problems. in Artificial intelligence and statistics 592–600
  # (2012).

  # Compute the quantile we are interested in.
  q <- 1.0 - 1.0 / (acquisition_data$t + 1.0)

  if ("percentile" %in% get_prediction_type(object)) {
    # Compute empiric alpha.
    alpha <- .predict(
      object = object,
      data = parameter_data,
      type = "percentile",
      percentile = q
    )$percentile
    
  } else {
    # Predict the mean and standard deviation for the parameter sets in
    # parameter data.
    prediction_table <- .predict(
      object = object,
      data = parameter_data,
      type = "sd"
    )

    # Stochastic alpha.
    alpha <- stats::qnorm(
      q,
      mean = prediction_table$mu, 
      sd = prediction_table$sigma
    )
  }

  return(alpha)
}



.compute_expected_train_time <- function(
    parameter_set,
    time_model
) {
  # If no time model is present, return NA.
  if (is.null(time_model)) return(NA_real_)

  if (!data.table::is.data.table(parameter_set)) {
    parameter_set <- data.table::as.data.table(parameter_set)
  }

  # Check if package is installed
  require_package(
    x = "ranger",
    purpose = "to predict process time of hyperparameter sets"
  )

  predictions <- predict(time_model,
    data = parameter_set,
    type = "response",
    num.threads = 1L,
    verbose = FALSE
  )

  return(predictions$predictions)
}



.get_available_acquisition_functions <- function() {
  return(c(
    "improvement_probability",
    "improvement_empirical_probability",
    "expected_improvement",
    "upper_confidence_bound",
    "bayes_upper_confidence_bound"
  ))
}



.get_available_hyperparameter_learners <- function() {
  return(c(
    .get_available_ranger_hyperparameter_learners(),
    .get_available_lagp_hyperparameter_learners(),
    .get_available_bart_hyperparameter_learners(),
    .get_available_random_search_hyperparameter_learners()
  ))
}



.get_available_hyperparameter_exploration_methods <- function() {
  return(c(
    "single_shot",
    "successive_halving",
    "stochastic_reject",
    "none"
  ))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R



# show (vimpTable) -------------------------------------------------------------
setMethod(
  "show", 
  signature(object = "vimpTable"),
  function(object) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL

    # Make sure the object is updated.
    object <- update_object(object = object)

    # Create an initial descriptor.
    if (is_empty(object)) {
      cat(paste0(
        "An empty variable importance table for the ", object@vimp_method,
        " variable importance method ",
        "(", .familiar_version_string(object), ")."
      ))
      
    } else {
      cat(paste0(
        "A variable importance table for the ", object@vimp_method,
        " variable importance method ",
        "(", .familiar_version_string(object), "):\n\n"
      ))

      if (object@invert) {
        show(object@vimp_table[order(-score)])
      } else {
        show(object@vimp_table[order(score)])
      }
    }
  }
)



#' @title Extract variable importance table.
#'
#' @description This method retrieves and parses variable importance tables from
#'   their respective `vimpTable` objects.
#'
#' @param x Variable importance (`vimpTable`) object, a list thereof, or one or
#'   more paths to these objects. This method extracts the variable importance
#'   table from such objects.
#' @param state State of the returned variable importance table. This affects
#'   what contents are shown, and in which format. The variable importance table
#'   can be returned with the following states:
#'
#'  * `initial`: initial state, directly after the variable importance table is
#'   filled. The returned variable importance table shows the raw, un-processed
#'   data.
#'
#'  * `decoded`: depending on the variable importance method, the initial
#'   variable importance table may contain the scores of individual contrasts
#'   for categorical variables. When decoded, scores from all contrasts are
#'   aggregated to a single score for each feature.
#'
#'  * `declustered`: variable importance is determined from fully processed
#'   features, which includes clustering. This means that a single feature in
#'   the variable importance table may represent multiple original features.
#'   When a variable importance table has been declustered, all clusters have
#'   been turned into their constituent features.
#'
#'  * `ranked` (default): The scores have been used to create ranks, with lower
#'   ranks indicating better features.
#'
#'   Internally, the variable importance table will go through each state, i.e.
#'   an variable importance table in the initial state will be decoded,
#'   declustered and then ranked prior to returning the variable importance
#'   table.
#' @param data Internally used argument for use with `familiarModel` objects.
#' @param as_object Internally used argument for use with `familiarModel`
#'   objects.
#' @param ... Unused arguments.
#'
#' @return A `data.table` with variable importance scores and, with
#'   `state="ranked"`, the respective ranks.
#' @exportMethod get_vimp_table
#'
#' @md
#' @rdname get_vimp_table-methods
setGeneric(
  "get_vimp_table", function(
    x,
    state = "ranked",
    ...
  ) {
    standardGeneric("get_vimp_table")
  }
)



# get_vimp_table (list) --------------------------------------------------------

#' @rdname get_vimp_table-methods
setMethod(
  "get_vimp_table",
  signature(x = "list"),
  function(x, state = "ranked", ...) {
    # Dispatch to methods for underlying objects.
    return(lapply(x, get_vimp_table, state = state, ...))
  }
)



# get_vimp_table (character) ---------------------------------------------------

#' @rdname get_vimp_table-methods
setMethod(
  "get_vimp_table",
  signature(x = "character"),
  function(x, state = "ranked", ...) {
    # Dispatch to list, if x contains more than one element.
    if (length(x) > 1L) {
      return(get_vimp_table(as.list(x), state = state, ...))
    }

    # Attempt to read file.
    x <- readRDS(x)

    # Prior to version 1.2.0, files would contain a list of data.
    if (is.list(x)) {
      x <- as_vimp_table_object(
        x,
        project_id = ""
      )
    }

    # Make sure the object is updated.
    x <- update_object(object = x)

    # Dispatch to object-specific routines.
    return(get_vimp_table(x = x, state = state, ...))
  }
)



# get_vimp_table (vimpTable) ---------------------------------------------------

#' @rdname get_vimp_table-methods
setMethod(
  "get_vimp_table",
  signature(x = "vimpTable"),
  function(x, state = "ranked", ...) {
    # Check that x is not empty.
    if (is_empty(x)) return(NULL)

    # Ensure that the variable importance table is processed.
    x <- preprocess_vimp_table(x, stop_at = state)

    # Return a copy of the variable importance table attribute.
    return(data.table::copy(x@vimp_table))
  }
)



# get_vimp_table (NULL) --------------------------------------------------------

#' @rdname get_vimp_table-methods
setMethod(
  "get_vimp_table",
  signature(x = "NULL"),
  function(x, state = "ranked", ...) {
    return(NULL)
  }
)



# get_vimp_table (experimentData) ----------------------------------------------

#' @rdname get_vimp_table-methods
setMethod(
  "get_vimp_table",
  signature(x = "experimentData"),
  function(x, state = "ranked", ...) {
    # Check if the attribute has been set.
    if (is.null(x@vimp_table_list)) {
      ..warning("No variable importance tables are present.")

      return(NULL)
    }

    # Iterate over data for the different variable importance methods.
    return(lapply(
      x@vimp_table_list,
      get_vimp_table,
      state = state,
      ...
    ))
  }
)



# get_vimp_table (familiarModel) -----------------------------------------------

#' @rdname get_vimp_table-methods
setMethod(
  "get_vimp_table",
  signature(x = "familiarModel"),
  function(x, state = "ranked", data = NULL, as_object = FALSE, ...) {
    # This method is used to obtain post-hoc variable importance tables from a
    # trained model.

    # Determine variable importance using internal routines of the familiarModel
    # object.
    vimp_object <- ..vimp(
      object = x,
      data = data
    )

    # Set up a table with clustering information.
    cluster_table <- .create_clustering_table(x@feature_info)

    # Update variable importance table object.
    vimp_object@vimp_method <- x@learner
    vimp_object@run_table <- x@run_table
    vimp_object@cluster_table <- cluster_table
    vimp_object@project_id <- x@project_id

    # Set package version.
    vimp_object <- add_package_version(vimp_object)

    # Determine what should be returned.
    if (as_object) {
      # The as_object flag allows for returning the actual data.
      return(vimp_object)
      
    } else {
      return(get_vimp_table(
        x = vimp_object,
        state = state
      ))
    }
  }
)



# decode_vimp_table (vimpTable) ------------------------------------------------
setMethod(
  "decode_vimp_table", signature(x = "vimpTable"),
  function(x, ...) {
    # Use encoding table to parse variable importance table.

    # Suppress NOTES due to non-standard evaluation in data.table
    name <- original_name <- score <- NULL

    # Check if the table has already been decoded.
    if (.as_vimp_table_state(x@state) >= "decoded") {
      return(x)
    }

    # Update state.
    x@state <- "decoded"

    # Skip if the vimp table is empty.
    if (is_empty(x)) return(x)

    # Skip if there is no encoding table.
    if (is_empty(x@encoding_table)) return(x)

    # Split vimp_table into categorical and non-categorical features.
    data_non_categorical <- x@vimp_table[!name %in% x@encoding_table$reference_name, ]
    data_categorical <- x@vimp_table[name %in% x@encoding_table$reference_name, ]

    if (!is_empty(data_categorical)) {
      # Merge with reference.
      data_categorical <- merge(
        x = data_categorical,
        y = x@encoding_table,
        by.x = "name",
        by.y = "reference_name"
      )

      # Summarise score by single value according to the method indicated by the
      # score_aggregation attribute.
      if (x@score_aggregation == "max") {
        data_categorical <- suppressWarnings(
          data_categorical[, list(score = max(score, na.rm = TRUE)), by = original_name]
        )
        
      } else if (x@score_aggregation == "abs_max") {
        data_categorical <- suppressWarnings(
          data_categorical[, list(score = max(abs(score), na.rm = TRUE)), by = original_name]
        )
        
      } else if (x@score_aggregation == "min") {
        data_categorical <- suppressWarnings(
          data_categorical[, list(score = min(score, na.rm = TRUE)), by = original_name]
        )
        
      } else if (x@score_aggregation == "abs_min") {
        data_categorical <- suppressWarnings(
          data_categorical[, list(score = min(abs(score), na.rm = TRUE)), by = original_name]
        )
        
      } else if (x@score_aggregation == "mean") {
        data_categorical <- suppressWarnings(
          data_categorical[, list(score = mean(score, na.rm = TRUE)), by = original_name]
        )
        
      } else if (x@score_aggregation == "abs_mean") {
        data_categorical <- suppressWarnings(
          data_categorical[, list(score = mean(abs(score), na.rm = TRUE)), by = original_name]
        )
        
      } else if (x@score_aggregation == "median") {
        data_categorical <- suppressWarnings(
          data_categorical[, list(score = stats::median(score, na.rm = TRUE)), by = original_name]
        )

      } else if (x@score_aggregation == "abs_median") {
        data_categorical <- suppressWarnings(
          data_categorical[, list(score = stats::median(abs(score), na.rm = TRUE)), by = original_name]
        )
        
      } else {
        ..error_reached_unreachable_code(paste0(
          "decode_vimp_table,vimpTable: unknown score aggregation method"
        ))
      }

      # Change name of original_name column to name
      data.table::setnames(data_categorical, "original_name", "name")
    }

    # Combine to single data.table
    vimp_table <- rbind(
      data_non_categorical,
      data_categorical
    )

    # Replace infinite/nan/etc values by NA
    vimp_table[!is.finite(score), "score" := NA_real_]

    # Replace vimp_table attribute.
    x@vimp_table <- vimp_table

    return(x)
  }
)



# decode_vimp_table (NULL) -----------------------------------------------------
setMethod(
  "decode_vimp_table",
  signature(x = "NULL"),
  function(x, ...) {
    return(x)
  }
)



# decluster_vimp_table (list) --------------------------------------------------
setMethod(
  "decluster_vimp_table",
  signature(x = "list"),
  function(x, ...) {
    # If the list is empty, return NULL instead.
    if (is_empty(x)) return(NULL)
    
    # Dispatch to method for single variable importance tables.
    return(lapply(
      x,
      decluster_vimp_table
    ))
  }
)



# decluster_vimp_table (vimpTable) ---------------------------------------------
setMethod(
  "decluster_vimp_table",
  signature(x = "vimpTable"),
  function(x, show_weights = FALSE, show_cluster_name = FALSE, ...) {
    
    # Check if the table has already been declustered.
    if (.as_vimp_table_state(x@state) >= "declustered") return(x)

    # Pre-process the variable importance table.
    x <- preprocess_vimp_table(x, stop_at = "decoded")

    # Update state.
    x@state <- "declustered"

    # Skip if the vimp table is empty.
    if (is_empty(x)) return(x)

    # Skip if there is no cluster table to work with.
    if (is_empty(x@cluster_table)) return(x)

    # Merge the cluster table into the variable importance table..
    vimp_table <- merge(
      x = x@vimp_table,
      y = x@cluster_table,
      by.x = "name",
      by.y = "cluster_name",
      all.x = FALSE,
      all.y = FALSE
    )

    # Adapt table by removing the original name column and renaming the
    # feature name column.
    vimp_table[, ":="(
      "name" = NULL, 
      "feature_required" = NULL
    )]
    data.table::setnames(
      x = vimp_table,
      old = "feature_name",
      new = "name"
    )

    # Remove weights, unless explicitly stated.
    if (!show_weights && "weight" %in% colnames(vimp_table)) {
      vimp_table[, "weight" := NULL]
    }

    # Update vimp_table attribute.
    x@vimp_table <- vimp_table

    return(x)
  }
)



# decluster_vimp_table (NULL) --------------------------------------------------
setMethod(
  "decluster_vimp_table",
  signature(x = "NULL"),
  function(x, ...) {
    return(x)
  }
)



# recluster_vimp_table (list) --------------------------------------------------
setMethod(
  "recluster_vimp_table",
  signature(x = "list"),
  function(x, ...) {
    # If the list is empty, return NULL instead.
    if (is_empty(x)) return(NULL)

    # Dispatch to method for single variable importance tables.
    return(lapply(
      x,
      recluster_vimp_table
    ))
  }
)



# recluster_vimp_table (vimpTable) ---------------------------------------------
setMethod(
  "recluster_vimp_table",
  signature(x = "vimpTable"),
  function(x, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL

    # Check if the table has already been reclustered.
    if (.as_vimp_table_state(x@state) >= "reclustered") return(x)

    # Pre-process the variable importance table. Generally, this should not be
    # necessary. The idea is that one calls update_vimp_table_to_reference prior
    # to calling recluster_vimp_table. This makes it possible to form clusters
    # according in the local context -- even if the same clusters did not appear
    # when computing variable importance.
    x <- preprocess_vimp_table(x, stop_at = "declustered")

    # Update state.
    x@state <- "reclustered"

    # Skip if the vimp table is empty.
    if (is_empty(x)) return(x)

    # Skip if there is no cluster table to work with.
    if (is_empty(x@cluster_table)) return(x)

    # Merge variable importance table with reference cluster table.
    vimp_table <- merge(
      x = x@vimp_table[, mget(c("name", "score"))],
      y = x@cluster_table,
      by.x = "name",
      by.y = "feature_name",
      all.x = FALSE,
      all.y = TRUE
    )

    # Compute mean score for all features in the same cluster.
    vimp_table <- vimp_table[, list("score" = mean(score, na.rm = TRUE)), by = "cluster_name"]

    # Keep only relevant columns, and rename "cluster_name" to "name".
    vimp_table[, mget(c("score", "cluster_name"))]
    data.table::setnames(
      x = vimp_table,
      old = "cluster_name",
      new = "name"
    )

    # Update the vimp_table attribute.
    x@vimp_table <- vimp_table

    return(x)
  }
)



# recluster_vimp_table (NULL) --------------------------------------------------
setMethod(
  "recluster_vimp_table",
  signature(x = "NULL"),
  function(x, ...) {
    return(x)
  }
)



# rank_vimp_table (vimpTable) --------------------------------------------------
setMethod(
  "rank_vimp_table",
  signature(x = "vimpTable"),
  function(x, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL

    # Check if the table has already been ranked
    if (.as_vimp_table_state(x@state) >= "ranked") return(x)

    # Pre-process the variable importance table.
    x <- preprocess_vimp_table(x, stop_at = "declustered")

    # Update state.
    x@state <- "declustered"

    # Skip if the vimp table is empty.
    if (is_empty(x)) return(x)

    # Make copy to avoid updating by reference.
    vimp_table <- data.table::copy(x@vimp_table)

    # Remove NA values.
    vimp_table <- vimp_table[is.finite(score)]

    # Compute dense ranks, with NA ranked as NA.
    if (x@invert) {
      vimp_table[
        ,
        "rank" := data.table::frank(-score, ties.method = "dense", na.last = "keep")
      ]
      
    } else {
      vimp_table[
        ,
        "rank" := data.table::frank(score, ties.method = "dense", na.last = "keep")
      ]
    }

    # Update vimp_table attribute.
    x@vimp_table <- vimp_table

    return(x)
  }
)



# rank_vimp_table (NULL) -------------------------------------------------------
setMethod(
  "rank_vimp_table",
  signature(x = "NULL"),
  function(x, ...) {
    return(x)
  }
)



# preprocess_vimp_table (vimpTable) --------------------------------------------
setMethod(
  "preprocess_vimp_table",
  signature(x = "vimpTable"),
  function(x, stop_at = "ranked", ...) {
    # Convert the state attained and the requested state to ordinals.
    state_attained <- .as_vimp_table_state(x@state)
    stop_at <- .as_vimp_table_state(stop_at)

    if (state_attained < "decoded" && stop_at >= "decoded") {
      # Transform the features.
      x <- decode_vimp_table(x, ...)
    }

    if (state_attained < "declustered" && stop_at >= "declustered") {
      # Normalise feature values.
      x <- decluster_vimp_table(x, ...)
    }

    if (state_attained < "ranked" && stop_at >= "ranked") {
      # Batch-normalise feature values
      x <- rank_vimp_table(x, ...)
    }

    return(x)
  }
)



# preprocess_vimp_table (NULL) -------------------------------------------------
setMethod(
  "preprocess_vimp_table",
  signature(x = "NULL"),
  function(x, ...) {
    return(x)
  }
)



# remove_signature_features (vimpTable) ----------------------------------------
setMethod(
  "remove_signature_features",
  signature(x = "vimpTable"),
  function(x, features = NULL) {
    # Suppress NOTES due to non-standard evaluation in data.table
    original_name <- name <- NULL

    # Check state.
    if (x@state != "initial") {
      ..error_reached_unreachable_code(paste0(
        "remove_signature_features,vimpTable: can only remove signature ",
        "features from a variable importance table in its initial state."
      ))
    }

    # If there are no signature features, we don't need to remove them.
    if (is_empty(features)) return(x)

    # Check if any of the signature features are encoded.
    if (!is.null(x@encoding_table)) {
      # Find features that were encoded in the variable importance table.
      encoded_features <- features[features %in% x@encoding_table$original_name]

      # Find features that did not require encoding.
      non_encoded_features <- setdiff(features, encoded_features)

      # Find names of the encoded features as they appear in the variable
      # importance table.
      encoded_features <- x@encoding_table[
        original_name %in% encoded_features
      ]$reference_name

      # Combine encoded and non-encoded feature names.
      features <- c(encoded_features, non_encoded_features)
    }

    # Keep only features in the variable importance table that are not signature
    # features.
    x@vimp_table <- x@vimp_table[!name %in% features]

    return(x)
  }
)

# remove_signature_features (NULL) ---------------------------------------------
setMethod(
  "remove_signature_features",
  signature(x = "NULL"),
  function(x, features = NULL) {
    return(x)
  }
)



# update_vimp_table_to_reference (list) ----------------------------------------
setMethod(
  "update_vimp_table_to_reference", signature(x = "list"),
  function(x, reference_cluster_table = NULL, ...) {
    # If the list is empty, return NULL instead.
    if (is_empty(x)) return(NULL)

    # Dispatch to method for single variable importance tables.
    return(lapply(
      x,
      update_vimp_table_to_reference,
      reference_cluster_table = reference_cluster_table
    ))
  }
)

# update_vimp_table_to_reference (vimpTable) -----------------------------------
setMethod(
  "update_vimp_table_to_reference",
  signature(x = "vimpTable"),
  function(x, reference_cluster_table = NULL, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL

    # Force declustering.
    if (.as_vimp_table_state(x@state) == "reclustered") x@state <- "decoded"

    # Ensure that the vimp table is declustered.
    x <- preprocess_vimp_table(x, "declustered", ...)

    # Update the cluster table.
    x@cluster_table <- reference_cluster_table

    # Set the state to declustered, since ranking is removed..
    x@state <- "declustered"

    if (is_empty(x)) return(x)

    # Merge variable importance table with reference cluster table.
    vimp_table <- merge(
      x = x@vimp_table[, mget(c("name", "score"))],
      y = reference_cluster_table,
      by.x = "name",
      by.y = "feature_name",
      all.x = FALSE,
      all.y = TRUE
    )

    # Compute mean score for all features in the same cluster.
    vimp_table[, "score" := mean(score, na.rm = TRUE), by = "cluster_name"]

    # Select only those entries with a finite score (i.e. those where the merge
    # produces matches).
    vimp_table <- vimp_table[is.finite(score)]

    # Keep only relevant columns, and rename "feature_name" to "name". Update
    # the vimp_table attribute.
    x@vimp_table <- vimp_table[, mget(c("score", "name"))]

    return(x)
  }
)



# update_vimp_table_to_reference (NULL) ----------------------------------------
setMethod(
  "update_vimp_table_to_reference",
  signature(x = "NULL"),
  function(x, reference_cluster_table = NULL, ...) {
    return(x)
  }
)



# collect_vimp_table (list) ----------------------------------------------------
setMethod(
  "collect_vimp_table",
  signature(x = "list"),
  function(x, run_table = NULL, ...) {
    # Collect variable importance tables, potentially limiting to

    # Suppress NOTES due to non-standard evaluation in data.table
    data_id <- run_id <- NULL

    # Check that all variable importance tables are for the same variable
    # importance method. Check first which variable importance methods are
    # present.
    vimp_method <- unique(sapply(x, function(x) x@vimp_method))

    # Dispatch to separate lists.
    if (length(vimp_method) > 1L) {
      # Initialise list to store variable importance tables.
      vimp_table_list <- list()
      for (current_vimp_method in vimp_method) {
        # Identify elements which use the same variable importance method.
        same_vimp_method <- sapply(
          x, 
          function(x, vimp_method) x@vimp_method == current_vimp_method
        )

        # Store all elements with the same variable importance method as an
        # element in the vimp_table_list list.
        vimp_table_list[[current_vimp_method]] <- list(x[same_vimp_method])
      }

      # Dispatch collections of the lists.
      return(lapply(
        vimp_table_list, 
        aggregate_vimp_table, 
        run_table = run_table, 
        ...
      ))
    }

    # Select only elements that share the same dataset.
    if (!is.null(run_table)) {
      # Try and identify matching data and run ids between vimpTable object and
      # the run_table. Start at the bottom and try to match.
      for (ii in rev(run_table$perturb_level)) {
        # Get a list with identifiers.
        run_id_list <- .get_iteration_identifiers(
          run = list("run_table" = run_table),
          perturb_level = ii
        )

        # Check whether there are any matching data and run ids by checking if
        # matching creates a non-empty table.
        matching_elements <- sapply(
          x,
          function(x, run_id_list) {
            if (is_empty(x)) {
              return(FALSE)
            }
            
            return(!is_empty(x@run_table[
              data_id == run_id_list$data & run_id == run_id_list$run
            ]))
          },
          run_id_list = run_id_list
        )

        # If there is any match, we break from the loop
        if (any(matching_elements)) break
      }

      # Update matching_elements with entries where the run_table has
      # not been specified.
      matching_elements <- matching_elements | sapply(x, function(x) is_empty(x@run_table))

      # Shrink x to matching elements.
      x <- x[matching_elements]
    }

    # Check that x is not empty.
    if (is_empty(x)) return(NULL)

    return(x)
  }
)



# collect_vimp_table (vimpTable) -----------------------------------------------
setMethod(
  "collect_vimp_table",
  signature(x = "vimpTable"),
  function(x, run_table = NULL, ...) {
    # Parse to list.
    return(collect_vimp_table(
      x = list(x),
      run_table = run_table,
      ...
    ))
  }
)



# collect_vimp_table (NULL) ----------------------------------------------------
setMethod(
  "collect_vimp_table",
  signature(x = "NULL"),
  function(x, run_table = NULL, ...) {
    return(x)
  }
)



#' @title Aggregate variable importance from multiple variable importance
#'  objects.
#'
#' @description This methods aggregates variable importance from one or more
#'  `vimpTable` objects.
#'
#' @param x Variable importance (`vimpTable`) object, a list thereof, or one or
#'  more paths to these objects.
#' @param aggregation_method Method used to aggregate variable importance. The
#'  available methods are described in the *feature selection methods* vignette.
#' @param rank_threshold Rank threshold used within several aggregation methods.
#'  See the *feature selection methods* vignette for more details.
#' @param ... unused parameters.
#'
#' @return A `vimpTable` object with aggregated variable importance data.
#' @exportMethod aggregate_vimp_table
#'
#' @md
#' @rdname aggregate_vimp_table-methods
setGeneric(
  "aggregate_vimp_table",
  function(
    x,
    aggregation_method,
    rank_threshold = NULL,
    ...
  ) {
    standardGeneric("aggregate_vimp_table")
  }
)



# aggregate_vimp_table (list) --------------------------------------------------

#' @rdname aggregate_vimp_table-methods
setMethod(
  "aggregate_vimp_table",
  signature(x = "list"),
  function(x, aggregation_method, rank_threshold = NULL, ...) {
    # Check that the list itself is not empty.
    if (is_empty(x)) return(NULL)

    # Check that any character elements are interpreted as paths, and are
    # loaded.
    x <- lapply(
      x,
      function(x) {
        if (is.character(x)) {
          # Attempt to read from file.
          x <- readRDS(x)

          # Prior to version 1.2.0, files would contain a list of data.
          if (is.list(x)) {
            x <- as_vimp_table_object(x, project_id = "")
          }

          # Update object definitions.
          x <- update_object(object = x)
        }

        return(x)
      }
    )

    # Check that the list contents are not empty.
    empty_elements <- sapply(x, is_empty)
    if (all(empty_elements)) return(NULL)

    # Keep only non-empty elements.
    x <- x[!empty_elements]

    # Split by variable importance method.
    vimp_methods <- sapply(x, function(x) (x@vimp_method))
    if (length(unique(vimp_methods)) > 1L) {
      vimp_table_list <- lapply(
        unique(vimp_methods),
        function(current_vimp_method, vimp_methods, x) {
          return(x[vimp_methods == current_vimp_method])
        },
        vimp_methods = vimp_methods,
        x = x
      )
      return(lapply(
        vimp_table_list,
        aggregate_vimp_table,
        aggregation_method = aggregation_method,
        rank_threshold = rank_threshold
      ))
    }
    
    # Preprocess individual elements prior to merging.
    x <- lapply(x, preprocess_vimp_table, ...)

    # Add in unique run_id, and export the variable importance themselves.
    vimp_table <- mapply(
      function(x, ii) {
        # Prevent updating by reference.
        vimp_table <- data.table::copy(x@vimp_table)

        # Add run identifier.
        vimp_table[, "run_id" := ii]

        return(vimp_table)
      },
      x = x,
      ii = seq_along(x),
      SIMPLIFY = FALSE
    )

    # Combine all variable importance tables into a single table.
    vimp_table <- data.table::rbindlist(vimp_table, use.names = TRUE)

    # Create an aggregated variable importance table.
    aggregated_vimp_object <- methods::new("vimpTable", x[[1L]])

    # Attach the aggregated variable importance table.
    aggregated_vimp_object@vimp_table <- vimp_table

    if (aggregation_method == "none") {
      aggregated_vimp_object <- .compute_rank_mean_score(
        x = aggregated_vimp_object
      )
      
    } else if (aggregation_method %in% c("mean", "median", "best", "worst")) {
      # Perform aggregation using simple ensemble methods

      if (aggregation_method == "mean") {
        aggregated_vimp_object <- .compute_rank_mean_rank(
          x = aggregated_vimp_object
        )
        
      } else if (aggregation_method == "median") {
        aggregated_vimp_object <- .compute_rank_median_rank(
          x = aggregated_vimp_object
        )
        
      } else if (aggregation_method == "best") {
        aggregated_vimp_object <- .compute_rank_best_rank(
          x = aggregated_vimp_object
        )
        
      } else if (aggregation_method == "worst") {
        aggregated_vimp_object <- .compute_rank_worst_rank(
          x = aggregated_vimp_object
        )
      }
      
    } else if (aggregation_method %in% c("stability", "exponential")) {
      # Perform aggregation using occurence-based methods.
      if (is.null(rank_threshold)) {
        rank_threshold <- .optimise_feature_occurrence_threshold(
          vimp_table = aggregated_vimp_object
        )
      }

      if (aggregation_method == "stability") {
        aggregated_vimp_object <- .compute_rank_stability(
          x = aggregated_vimp_object,
          rank_threshold = rank_threshold
        )
        
      } else if (aggregation_method == "exponential") {
        aggregated_vimp_object <- .compute_rank_exponential(
          x = aggregated_vimp_object,
          rank_threshold = rank_threshold
        )
      }
    } else if (aggregation_method %in% c(
      "borda", "enhanced_borda", "truncated_borda", "enhanced_truncated_borda"
    )) {
      # Perform aggregation using Borda-count based methods
      if (is.null(rank_threshold)) {
        rank_threshold <- .optimise_feature_occurrence_threshold(
          vimp_table = aggregated_vimp_object
        )
      }

      if (aggregation_method == "borda") {
        aggregated_vimp_object <- .compute_rank_borda(
          x = aggregated_vimp_object,
          rank_threshold = rank_threshold,
          truncated = FALSE,
          enhanced = FALSE
        )

      } else if (aggregation_method == "enhanced_borda") {
        aggregated_vimp_object <- .compute_rank_borda(
          x = aggregated_vimp_object,
          rank_threshold = rank_threshold,
          truncated = FALSE,
          enhanced = TRUE
        )

      } else if (aggregation_method == "truncated_borda") {
        aggregated_vimp_object <- .compute_rank_borda(
          x = aggregated_vimp_object,
          rank_threshold = rank_threshold,
          truncated = TRUE,
          enhanced = FALSE
        )

      } else if (aggregation_method == "enhanced_truncated_borda") {
        aggregated_vimp_object <- .compute_rank_borda(
          x = aggregated_vimp_object,
          rank_threshold = rank_threshold,
          truncated = TRUE,
          enhanced = TRUE
        )
      }
    } else {
      ..error_reached_unreachable_code(paste0(
        "aggregate_vimp_table,list: encountered an unknown aggregation method:",
        aggregation_method
      ))
    }

    # Force ranking of the aggregated variable importance object.
    aggregated_vimp_object@state <- "declustered"

    # Rank aggregated scores.
    aggregated_vimp_object <- rank_vimp_table(aggregated_vimp_object)

    # Update state again to reflect the current state.
    aggregated_vimp_object@state <- "aggregated"

    return(aggregated_vimp_object)
  }
)



# aggregate_vimp_table (character) ---------------------------------------------

#' @rdname aggregate_vimp_table-methods
setMethod(
  "aggregate_vimp_table",
  signature(x = "character"),
  function(x, aggregation_method, rank_threshold = NULL, ...) {
    # Force x to a list, and pass to main method.
    return(aggregate_vimp_table(
      x = as.list(x),
      aggregation_method = aggregation_method,
      rank_threshold = rank_threshold,
      ...
    ))
  }
)



# aggregate_vimp_table (vimpTable) ---------------------------------------------

#' @rdname aggregate_vimp_table-methods
setMethod(
  "aggregate_vimp_table",
  signature(x = "vimpTable"),
  function(x, aggregation_method, rank_threshold = NULL, ...) {
    # Check if the variable importance table object already has been aggregated.
    if (.as_vimp_table_state(x@state) == "aggregated") return(x)

    return(aggregate_vimp_table(
      x = list(x),
      aggregation_method = aggregation_method,
      rank_threshold = rank_threshold,
      ...
    ))
  }
)



# aggregate_vimp_table (NULL) --------------------------------------------------

#' @rdname aggregate_vimp_table-methods
setMethod(
  "aggregate_vimp_table",
  signature(x = "NULL"),
  function(x, aggregation_method, rank_threshold = NULL, ...) {
    return(NULL)
  }
)



# aggregate_vimp_table (experimentData) ----------------------------------------

#' @rdname aggregate_vimp_table-methods
setMethod(
  "aggregate_vimp_table",
  signature(x = "experimentData"),
  function(x, aggregation_method, rank_threshold = NULL, ...) {
    # Check if the attribute has been set.
    if (is.null(x@vimp_table_list)) {
      ..warning("No variable importance tables are present.")

      return(NULL)
    }

    # Iterate over data for the different variable importance methods.
    return(lapply(
      x@vimp_table_list,
      aggregate_vimp_table,
      aggregation_method = aggregation_method,
      rank_threshold = rank_threshold,
      ...
    ))
  }
)



# add_package_version (vimpTable) ----------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "vimpTable"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)



.as_vimp_table_state <- function(x) {
  if (is(x, "vimpTable")) x <- x@state

  # Available states
  state_levels <- c(
    "initial",
    "decoded",
    "declustered",
    "reclustered",
    "ranked",
    "aggregated"
  )

  if (!all(x %in% state_levels)) {
    ..error_reached_unreachable_code(
      ".as_vimp_table_state: one or more of x could not be matched to vimp table states."
    )
  }

  return(factor(
    x = x,
    levels = state_levels,
    ordered = TRUE
  ))
}



as_vimp_table_object <- function(
    x,
    project_id
) {
  # Helper function for turning variable importance tables generated prior to
  # version 1.2.0 into vimpTable objects.
  
  vimp_table <- x$vimp
  if (!is_empty(vimp_table)) {
    # Determine if an increasing score leads to an increasing rank.
    invert <- stats::cor(
      vimp_table$score,
      vimp_table$rank,
      method = "spearman"
    ) < 0.0
    
    # Select only expected columns.
    vimp_table <- vimp_table[, mget(c("score", "name"))]
    
  } else {
    invert <- FALSE
  }

  # Create variable importance object.
  vimp_table <- methods::new(
    "vimpTable",
    vimp_table = vimp_table,
    vimp_method = x$vimp_method,
    run_table = x$run_table,
    score_aggregation = "max",
    encoding_table = NULL,
    cluster_table = NULL,
    invert = invert,
    project_id = project_id,
    familiar_version = "1.2.0",
    state = "declustered"
  )

  return(vimp_table)
}



prepare_vimp_table_object <- function(
    vimp_method,
    vimp_table = NULL,
    run_table = NULL,
    score_aggregation = "max",
    encoding_table = NULL,
    cluster_table = NULL,
    invert,
    state
) {
  # This function helps prepare a vimpTable object for unit testing. It is not
  # used as part of the main workflow.

  proto_vimp_table <- methods::new(
    "vimpTable",
    vimp_table = vimp_table,
    vimp_method = vimp_method,
    run_table = run_table,
    score_aggregation = score_aggregation,
    encoding_table = encoding_table,
    cluster_table = cluster_table,
    invert = invert,
    state = state
  )

  # Add package version.
  proto_vimp_table <- add_package_version(proto_vimp_table)

  return(proto_vimp_table)
}

# Create prototype.
proto_vimp_table <- familiar:::prepare_vimp_table_object(
  vimp_method = "mim",
  invert = TRUE,
  state = "declustered"
)

# Fill a list with multiple prototypes.
vimp_table_list <- list(
  proto_vimp_table,
  proto_vimp_table,
  proto_vimp_table
)

# Update with actual data.
vimp_table_list[[1]]@vimp_table <- data.table::data.table(
  "score" = c(1.0, 0.8, 0.6, 0.4, 0.2),
  "name" = c("a", "b", "c", "d", "e")
)
vimp_table_list[[2]]@vimp_table <- data.table::data.table(
  "score" = c(0.8, 1.0, 0.6),
  "name" = c("a", "b", "c")
)
vimp_table_list[[3]]@vimp_table <- data.table::data.table(
  "score" = c(0.6, 1.0),
  "name" = c("a", "e")
)

# Mean rank aggregation --------------------------------------------------------
testthat::test_that("Mean rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(5 / 3, 3 / 2, 3, 4, 3),
    "rank" = c(2, 1, 3, 4, 3)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "mean")

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Median rank aggregation ------------------------------------------------------
testthat::test_that("Median rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(2, 3 / 2, 3, 4, 3),
    "rank" = c(2, 1, 3, 4, 3)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "median")

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Best rank aggregation --------------------------------------------------------
testthat::test_that("Best rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(1, 1, 3, 4, 1),
    "rank" = c(1, 1, 2, 3, 1)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "best")

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Worst rank aggregation -------------------------------------------------------
testthat::test_that("Worst rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(2, 2, 3, 4, 5),
    "rank" = c(1, 1, 2, 3, 4)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "worst")

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Stability rank aggregation ---------------------------------------------------
testthat::test_that("Stability rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(1, 2 / 3, 2 / 3, 0, 1 / 3),
    "rank" = c(1, 2, 2, 4, 3)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "stability",
    rank_threshold = 3L)

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Exponential rank aggregation -------------------------------------------------
testthat::test_that("Exponential rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(
      exp(-1 / 3) + 2 * exp(-2 / 3),
      exp(-2 / 3) + exp(-1 / 3),
      2 * exp(-1),
      0.0,
      exp(-1 / 3)),
    "rank" = c(1, 2, 3, 5, 4)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "exponential",
    rank_threshold = 3L)

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Borda aggregation ------------------------------------------------------------
testthat::test_that("Borda rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(13 / 6, 9 / 5, 14 / 15, 2 / 5, 6 / 5),
    "rank" = c(1, 2, 4, 5, 3)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "borda",
    rank_threshold = 3L)

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Enhanced borda aggregation ---------------------------------------------------
testthat::test_that("Enhanced borda rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(13 / 6, 6 / 5, 28 / 45, 0, 2 / 5),
    "rank" = c(1, 2, 3, 5, 4)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "enhanced_borda",
    rank_threshold = 3L)

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Truncated borda aggregation --------------------------------------------------
testthat::test_that("Truncated borda rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(7 / 3, 5 / 3, 2 / 3, 0, 1),
    "rank" = c(1, 2, 4, 5, 3)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "truncated_borda",
    rank_threshold = 3L)

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

# Enhanced truncated borda aggregation -----------------------------------------
testthat::test_that("Enchanced truncated borda rank aggregation is correct", {
  expected_rank_table <- data.table::data.table(
    "name" = c("a", "b", "c", "d", "e"),
    "score" = c(7 / 3, 10 / 9, 4 / 9, 0, 1 / 3),
    "rank" = c(1, 2, 3, 5, 4)
  )

  aggregated_rank_table <- familiar:::aggregate_vimp_table(
    vimp_table_list,
    aggregation_method = "enhanced_truncated_borda",
    rank_threshold = 3L)

  testthat::expect_equal(
    familiar:::get_vimp_table(aggregated_rank_table),
    expected_rank_table,
    ignore_attr = TRUE)
})

logger_message <- function(
    mess_str,
    file_name = NULL,
    indent = 0L,
    verbose = TRUE
) {
  # Write message to console and file

  if (is.null(file_name)) file_name <- .get_log_file()

  # Derive an indent string using two spaces for each indentation.
  indent_str <- strrep("  ", indent)

  # Date and time string
  date_str <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")

  # Parse date and time with message string to log string
  log_str <- paste0(date_str, "\tMESSAGE\t", indent_str, mess_str)

  # Write message to log file
  if (!is.null(file_name)) {
    tryCatch(write(
      x = log_str,
      file = file_name,
      append = TRUE
    ))
  }

  # Write message to console
  if (verbose) message(paste0(indent_str, mess_str))
  
  return(invisible(NULL))
}



logger_warning <- function(
    warn_str,
    file_name = NULL,
    warn_class = NULL,
    call = rlang::caller_env()
) {
  # Write warning to console and file
  
  if (is.null(file_name)) file_name <- .get_log_file()

  # Date and time string
  date_str <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")

  # Parse date and time with warning string to log string
  log_str <- paste0(date_str, "\tWARNING\t", warn_str)

  # Write warning to log file
  if (!is.null(file_name)) {
    tryCatch(write(
      x = log_str,
      file = file_name,
      append = TRUE
    ))
  }

  # Write warning to console
  rlang::warn(
    message =  warn_str,
    class = union("familiar_warning", warn_class),
    call = call
  )
  
  return(invisible(NULL))
}



logger_stop <- function(
    err_str,
    file_name = NULL,
    error_class = NULL,
    call = rlang::caller_env()
) {
  # Write error to console and file

  if (is.null(file_name)) file_name <- .get_log_file()

  # Date and time string
  date_str <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")

  # Parse date and time with error string to log string
  log_str <- paste0(date_str, "\tERROR\t", err_str)

  # Write error to log file
  if (!is.null(file_name)) {
    tryCatch(write(
      x = log_str, 
      file = file_name, 
      append = TRUE
    ))
  }

  # Write error to console
  rlang::abort(
    message = err_str,
    class = union("familiar_error", error_class),
    call = call
  )
}



.get_log_file <- function() {
  # If the log file cannot be found return NULL.
  log_file <- tryCatch(
    get("log_file", envir = familiar_global_env),
    error = function(err) (NULL)
  )

  return(log_file)
}

testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Leave-one-out-cross-validation
familiar:::integrated_test(
  experimental_design = "lv(fs+mb)",
  vimp_method = "none",
  parallel = FALSE,
  skip_evaluation_elements = "all",
  outcome_type_available = "binomial",
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarNaiveBayes -----------------------------------------------------------
setClass(
  "familiarNaiveBayes",
  contains = "familiarModel"
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarNaiveBayes"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- "e1071"

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarNaiveBayes"),
  function(object, ...) {
    # Naive bayes is only available for categorical outcomes.
    if (object@outcome_type %in% c("binomial", "multinomial")) {
      return(TRUE)
    } else {
      return(FALSE)
    }
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarNaiveBayes"),
  function(object, data = NULL, ...) {
    # Initialise list and declare hyperparameter entries.
    param <- list()
    param$sign_size <- list()
    param$laplace <- list()

    # If data is explicitly set to NULL, return the list with
    # hyperparameter names only.
    if (is.null(data)) return(param)
    
    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(
      data = data, 
      restrict_samples = TRUE
    )

    # laplace smoothing parameter ----------------------------------------------
    param$laplace <- .set_hyperparameter(
      default = c(0.0, 1.0, 2.0, 5.0),
      type = "numeric",
      range = c(0.0, 10.0),
      valid_range = c(0.0, Inf),
      randomise = TRUE
    )

    return(param)
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarNaiveBayes",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object,
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Find feature columns in the data.
    feature_columns <- get_feature_columns(x = data)

    # Parse formula
    formula <- stats::reformulate(
      termlabels = feature_columns,
      response = quote(outcome)
    )

    # Train the model.
    model <- do.call_with_handlers(
      e1071::naiveBayes,
      args = list(
        formula,
        "data" = data@data,
        "laplace" = object@hyperparameters$laplace
      )
    )

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) {
      return(callNextMethod(object = object))
    }

    # Add model to the familiarModel object.
    object@model <- model

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarNaiveBayes",
    data = "dataObject"
  ),
  function(object, data, ...) {
    if (object@outcome_type %in% c("binomial", "multinomial")) {
      # Turn into a Naive model.
      object <- methods::new("familiarNaiveModel", object)
    }

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarNaiveBayes",
    data = "dataObject"
  ),
  function(object, data, type = "default", ...) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }
    
    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }
    
    if (type == "default") {
      # default ----------------------------------------------------------------
      
      if (object@outcome_type %in% c("binomial", "multinomial")) {
        
        # Use the model to predict class probabilities.
        model_predictions <- predict(
          object = object@model,
          newdata = data@data,
          type = "raw"
        )
        
        # Obtain class levels.
        class_levels <- get_outcome_class_levels(x = object)
        
        # Add class probabilities.
        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          prediction_list[[class_levels[ii]]] <- model_predictions[, class_levels[ii]]
        }
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      return(prediction_table)
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified method --------------------------------------------------

      # Use the model for prediction.
      return(predict(
        object = object@model,
        newdata = data@data,
        type = type,
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



# ..vimp------------------------------------------------------------------------
# Naive Bayes does not have an associated variable importance method.



# .trim_model-------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarNaiveBayes"),
  function(object, ...) {
    # Update model by removing the call.
    object@model$call <- call("trimmed")

    # Add show.
    object <- .capture_show(object)

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    # Default method for models that lack a more specific method.
    return(object)
  }
)



.get_available_naive_bayes_learners <- function(show_general = TRUE) {
  return("naive_bayes")
}

# First test if all selectable learners are also available
familiar:::test_all_learners_available(
  learners = familiar:::.get_available_naive_bayes_learners(show_general = TRUE)
)

# Don't perform any further tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_all_learners_train_predict_vimp(
  learners = familiar:::.get_available_naive_bayes_learners(show_general = FALSE),
  hyperparameter_list = list(
    "binomial" = list("laplace" = 0.0),
    "multinomial" = list("laplace" = 0.0)
  ),
  has_vimp = FALSE
)

familiar:::test_all_learners_parallel_train_predict_vimp(
  learners = familiar:::.get_available_naive_bayes_learners(show_general = FALSE),
  hyperparameter_list = list(
    "binomial" = list("laplace" = 0.0),
    "multinomial" = list("laplace" = 0.0)
  ),
  has_vimp = FALSE
)


# Binomial tests----------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("binomial")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "laplace" = 0.0
  ),
  learner = "naive_bayes"
)

testthat::test_that("Naive Bayes model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # That no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("Naive Bayes model has no variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)

  # Expect that the vimp table is empty.
  testthat::expect_true(familiar:::is_empty(vimp_table))
})


# Multinomial tests-------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("multinomial")

# Train the model using the good dataset.
good_model <- suppressWarnings(familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "laplace" = 0.0
  ),
  learner = "naive_bayes"
))

testthat::test_that("Naive Bayes model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # That no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("Naive Bayes model has no variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)

  # Expect that the vimp table is empty.
  testthat::expect_true(familiar:::is_empty(vimp_table))
})


familiar:::test_hyperparameter_optimisation(
  learners = "naive_bayes",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)


testthat::skip("Skip hyperparameter optimisation, unless manual.")

familiar:::test_hyperparameter_optimisation(
  learners = familiar:::.get_available_naive_bayes_learners(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
#' @include Normalisation.R
NULL



# Parametric ComBat object -----------------------------------------------------
setClass(
  "featureInfoParametersNormalisationParametricCombat",
  contains = "featureInfoParametersNormalisationShiftScale",
  slots = list("method" = "character"),
  prototype = list("method" = NA_character_)
)

# Non-parametric ComBat object -------------------------------------------------
setClass(
  "featureInfoParametersNormalisationNonParametricCombat",
  contains = "featureInfoParametersNormalisationShiftScale",
  slots = list("method" = "character"),
  prototype = list("method" = NA_character_)
)


# Only included in .get_available_batch_normalisation_methods
.get_available_combat_parametric_normalisation_methods <- function() {
  return(c("combat", "combat_p", "combat_parametric"))
}

# Only included in .get_available_batch_normalisation_methods
.get_available_combat_non_parametric_normalisation_methods <- function() {
  return(c("combat_np", "combat_non_parametric"))
}



# add_feature_info_parameters (parametric ComBat, numeric) ---------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationParametricCombat",
    data = "numeric"
  ),
  function(
    object, 
    data,
    batch_parameter_data,
    ...
  ) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    feature <- batch_id <- NULL
    
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Generate a replacement object that uses univariate
    # standardisation.
    replacement_object <- ..create_normalisation_parameter_skeleton(
      feature_name = object@name,
      method = "standardisation",
      batch = object@batch
    )
    
    # Select the current feature and batch from batch_parameter_data.
    batch_parameter_data <- batch_parameter_data[feature == object@name & batch_id == object@batch, ]
    
    # If the batch parameter data are empty, attempt to use univariate
    # standardisation.
    if (is_empty(batch_parameter_data)) {
      return(add_feature_info_parameters(
        object = replacement_object,
        data = data
      ))
    }
    
    # If shift and scale parameters are not finite, attempt to use univariate
    # standardisation.
    if (!is.finite(batch_parameter_data$norm_shift) || !is.finite(batch_parameter_data$norm_scale)) {
      return(add_feature_info_parameters(
        object = replacement_object,
        data = data
      ))
    }
    
    # Set shift and scale parameters.
    object@shift <- batch_parameter_data$norm_shift
    object@scale <- batch_parameter_data$norm_scale
    
    # Check for scales which are close to 0
    if (object@scale < 2.0 * .Machine$double.eps) object@scale <- 1.0
    
    # Mark complete.
    object@complete <- TRUE
    
    return(object)
  }
)



# add_feature_info_parameters (non-parametric ComBat, numeric) -----------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationNonParametricCombat",
    data = "numeric"
  ),
  function(
    object, 
    data,
    batch_parameter_data,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    feature <- batch_id <- NULL
    
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Generate a replacement object that uses univariate
    # standardisation.
    replacement_object <- ..create_normalisation_parameter_skeleton(
      feature_name = object@name,
      method = "standardisation",
      batch = object@batch
    )
    
    # Select the current feature and batch from batch_parameter_data.
    batch_parameter_data <- batch_parameter_data[feature == object@name & batch_id == object@batch, ]
    
    # If the batch parameter data are empty, attempt to use univariate
    # standardisation.
    if (is_empty(batch_parameter_data)) {
      return(add_feature_info_parameters(
        object = replacement_object,
        data = data
      ))
    }
    
    # If shift and scale parameters are not finite, attempt to use
    # univariate standardisation.
    if (!is.finite(batch_parameter_data$norm_shift) || !is.finite(batch_parameter_data$norm_scale)) {
      return(add_feature_info_parameters(
        object = replacement_object,
        data = data
      ))
    }
    
    # Set shift and scale parameters.
    object@shift <- batch_parameter_data$norm_shift
    object@scale <- batch_parameter_data$norm_scale
    
    # Check for scales which are close to 0
    if (object@scale < 2.0 * .Machine$double.eps) object@scale <- 1.0
    
    # Mark complete.
    object@complete <- TRUE
    
    return(object)
  }
)




.compute_combat_batch_normalisation_z_matrix <- function(data, feature_names) {
  # Suppress NOTES due to non-standard evaluation in data.table
  value <- n <- is_valid <- invariant <- n_features <- NULL
  gamma_hat <- .NATURAL <- NULL
  
  # Get the batch identifier column.
  batch_id_column <- get_id_columns(single_column = "batch")
  
  # Set a minimum threshold for valid sample sizes
  min_valid_samples <- 10L
  min_valid_features <- 3L
  
  # Convert the table from wide to long format
  x <- data.table::melt(
    data[, mget(c(batch_id_column, feature_names))],
    id.vars = batch_id_column,
    variable.name = "feature",
    value.name = "value"
  )
  
  # Sum the number of instances with a valid value, and determine whether all
  # values are singular.
  aggregate_table <- x[
    ,
    list(
      "n" = sum(is.finite(value)),
      "invariant" = is_singular_data(value)
    ),
    by = c(batch_id_column, "feature")
  ]
  
  # A feature/batch combination is valid if it is not invariant and has
  # sufficient instances with a finite value.
  aggregate_table[, "is_valid" := invariant == FALSE & n >= min_valid_samples]
  
  # Determine the number of features available in each batch.
  aggregate_table[, "n_features" := sum(is_valid), by = c(batch_id_column)]
  
  # If a batch does not have sufficient features (3 or more), mark as invalid.
  aggregate_table[n_features < min_valid_features, "is_valid" := FALSE]
  
  # Drop invariant and n_features columns, as these are no longer required.
  aggregate_table[, ":="("n_features" = NULL, "invariant" = NULL)]
  
  # First select for which combat can be used. This table is called z,
  # after the name for standardized data in Johnson et al. Note that overall
  # standardisation is conducted prior to batch-normalisation in familiar, and
  # is external to this function.
  z <- x[aggregate_table[is_valid == TRUE], on = .NATURAL]
  
  if (!is_empty(z)) {
    # Remove NA or infinite values.
    z <- z[is.finite(value)]
    
    # Determine the mean (gamma_hat) of each feature in each batch.
    z[, "gamma_hat" := mean(value), by = c(batch_id_column, "feature")]
    
    # Determine the variance (delta_hat_squared) of each feature in each batch.
    z[, "delta_hat_squared" := 1.0 / n * sum((value - gamma_hat)^2.0), by = c(batch_id_column, "feature")]
    
  } else {
    z <- NULL
  }
  
  return(z)
}



.combat_iterative_parametric_bayes_solver <- function(
    z,
    tolerance = 0.0001,
    max_iterations = 20L,
    cl = NULL,
    progress_bar = FALSE
) {
  # In the empirical Bayes approach with parametric priors, the posterior
  # estimations of gamma and delta_squared are iteratively optimised.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  gamma_hat <- delta_hat_squared <- NULL
  
  # Check that z is not empty.
  if (is_empty(z)) return(NULL)
  
  # Get the batch identifier column.
  batch_id_column <- get_id_columns(single_column = "batch")
  
  # Create new table by removing value and taking only unique rows.
  z_short <- unique(data.table::copy(z)[, "value" := NULL])

  # Determine gamma_bar, which is the average value of gamma_hat across
  # features within each batch. 1 value per batch.
  z_short[, "gamma_bar" := mean(gamma_hat), by = batch_id_column]
  
  # Determine tau_bar_squared, which is the prior for variance in the normal
  # distribution of gamma_hat. 1 value per batch.
  z_short[, "tau_bar_squared" := stats::var(gamma_hat), by = batch_id_column]
  
  # Determine lambda_bar (also called alpha) priors for the inverse gamma
  # distribution according to Johnson et al, supplement A3.1. 1 value per batch
  z_short[, "lambda_bar" := ..combat_lambda_prior(delta_hat_squared), by = batch_id_column]
  
  # Determine theta_bar (also called beta) priors according to Johnson et al.
  # supplement A3.1. 1 value per batch
  z_short[, "theta_bar" := ..combat_theta_prior(delta_hat_squared), by = batch_id_column]

  # Split by batch.
  batch_parameters <- fam_lapply_lb(
    cl = cl,
    assign = NULL,
    X = split(z, by = batch_id_column),
    FUN = ..combat_iterative_parametric_bayes_solver,
    progress_bar = progress_bar,
    z_short = z_short,
    tolerance = tolerance,
    max_iterations = max_iterations
  )
  
  # Concatenate to single table.
  batch_parameters <- data.table::rbindlist(batch_parameters)
  
  return(batch_parameters)
}



..combat_iterative_parametric_bayes_solver <- function(
    z,
    z_short,
    tolerance,
    max_iterations
) {
  # Perform the actual parametric estimations within each batch. Called from the
  # .combat_iterative_parametric_bayes_solver function
  
  # Suppress NOTES due to non-standard evaluation in data.table
  batch_id <- NULL
  value <- n <- gamma_hat <- tau_bar_squared <- gamma_bar <- gamma_star_new <- sum_squared_error <- NULL
  delta_star_squared_old <- delta_star_squared_new <- delta_hat_squared <- theta_bar <- lambda_bar <- NULL
  
  # Get the batch identifier column.
  batch_id_column <- get_id_columns(single_column = "batch")
  
  # Limit z_short to the current batch.
  current_batch_id <- z[[batch_id_column]][1L]
  z_short <- data.table::copy(z_short[batch_id == current_batch_id])
  
  # Initialise conditional posteriors. 1 value per feature per batch.
  z_short[, ":="(
    "gamma_star_old" = gamma_hat,
    "delta_star_squared_old" = delta_hat_squared
  )]
  
  # Convergence update value.
  convergence_update <- 1.0
  iteration_count <- 0L
  
  while (convergence_update > tolerance && iteration_count < max_iterations) {

    # Update the conditional gamma star posterior.
    z_short[, "gamma_star_new" := ..combat_gamma_posterior(
      n = n,
      tau_bar_squared = tau_bar_squared,
      gamma_hat = gamma_hat,
      delta_star_squared = delta_star_squared_old,
      gamma_bar = gamma_bar
    )]
    
    # Merge z_short back into z prior to computing the sum squared error.
    z_new <- merge(
      x = z[, mget(c(batch_id_column, "feature", "value"))],
      y = z_short[, mget(c(batch_id_column, "feature", "gamma_star_new"))],
      by = c(batch_id_column, "feature"),
      all = TRUE
    )
    
    # Compute sum squared error in equation 3.1 for the delta_star_squared posterior.
    z_new <- z_new[
      ,
      list("sum_squared_error" = sum((value - gamma_star_new)^2.0)),
      by = c(batch_id_column, "feature")
    ]
    
    z_short <- merge(
      x = z_short,
      y = z_new,
      by = c(batch_id_column, "feature"),
      all = TRUE
    )
    
    # Update the conditional delta_star_squared posterior.
    z_short[, "delta_star_squared_new" := ..combat_delta_squared_posterior(
      theta_bar = theta_bar,
      sum_squared_error = sum_squared_error,
      n = n,
      lambda_bar = lambda_bar
    )]
    
    # Update convergence
    convergence_update <- max(
      abs((z_short$gamma_star_new - z_short$gamma_star_old) / z_short$gamma_star_old),
      abs((z_short$delta_star_squared_new - z_short$delta_star_squared_old) / z_short$delta_star_squared_old)
    )
    
    # Replace old posterior values
    z_short[, ":="(
      "gamma_star_old" = gamma_star_new,
      "delta_star_squared_old" = delta_star_squared_new,
      "sum_squared_error" = NULL
    )]
    
    iteration_count <- iteration_count + 1L
  }
  
  # Store delta_star
  z_short[, "delta_star" := sqrt(delta_star_squared_new)]
  
  # Rename gamma_star_new and delta_star_squared_new
  data.table::setnames(
    x = z_short,
    old = c("gamma_star_new", "delta_star"),
    new = c("norm_shift", "norm_scale")
  )

  return(z_short[, mget(c(batch_id_column, "feature", "norm_shift", "norm_scale", "n"))])
}



.combat_non_parametric_bayes_solver <- function(
    z,
    n_sample_features = 50L,
    cl = NULL,
    progress_bar = TRUE
) {
  # Computes batch parameters using non-parametric priors. Sadly, this seems to
  # be an O(2) problem because it compares each feature with every other
  # feature. Johnson et al. We may be able to cheat a bit by subsampling the
  # feature space, so that we have O(n_features * n_sample_features) instead of
  # O(n_features * (n_features-1)).
  #
  # One of the datasets of Johnson et al. had 54000 features, which led to
  # 2.916E9 computations times 3 batches that took about an hour to complete.
  # Using n_sample_features=50, this would have been 3 * 2.7E6 computations,
  # which is more tractable, and scales way better.
  
  # Check if z is not empty.
  if (is_empty(z)) return(NULL)
  
  # Create a shortened table containing gamma_hat and delta_hat_squared
  # parameters for each feature and batch. If we make this function parallel at
  # one point, I prefer not to distribute z (in its entirety) to clusters, due
  # to memory issues.
  z_short <- unique(data.table::copy(z)[, "value" := NULL])
  
  # Iterate over batches and features to obtain batch parameters.
  batch_parameters <- fam_lapply(
    cl = cl,
    assign = NULL,
    X = split(
      z,
      by = c(get_id_columns(single_column = "batch"), "feature"),
      drop = TRUE
    ),
    FUN = ..combat_non_parametric_bayes_solver,
    progress_bar = progress_bar,
    z_short = z_short,
    n_sample_features = n_sample_features,
    chopchop = TRUE
  )
  
  # Concatenate to single table.
  batch_parameters <- data.table::rbindlist(batch_parameters)

  return(batch_parameters)
}



..combat_non_parametric_bayes_solver <- function(
    z,
    z_short,
    n_sample_features
) {

  # Suppress NOTES due to non-standard evaluation in data.table
  batch_id <- feature <- NULL
  
  # Get the batch identifier column.
  batch_id_column <- get_id_columns(single_column = "batch")
  
  # Limit z_short to the current batch.
  current_batch_id <- z[[batch_id_column]][1L]
  
  # Select the current cohort from z_short. Note that like in sva and other
  # codes, we interpret the g"=1...G in the supplement of Johnson et al. to
  # indicate that the current feature is not directly considered.
  z_short <- data.table::copy(z_short[batch_id == current_batch_id & feature != z$feature[1L]])
  
  # Find the number of features that are used to compute w_ig".
  features <- z_short$feature
  
  # Adapt the number of sample features.
  n_sample_features <- ifelse(n_sample_features > length(features), length(features), n_sample_features)
  
  # Find the selected features
  selected_features <- fam_sample(
    features,
    size = n_sample_features,
    replace = FALSE
  )
  
  # Make selection
  z_short <- droplevels(z_short[feature %in% selected_features])
  
  # Compute data for the weights w_ig that are used to reconstruct gamma_star
  # and delta_star_squared for the selected feature in the calling function.
  # (the selected_features variable contains the set difference of all
  # difference and the feature in the calling features.)
  weight_data <- lapply(
    split(z_short, by = "feature", sorted = FALSE),
    function(prior_pair, x) {
      # Compute sum((x-gamma_hat)^2) term in the product of the probability
      # density functions of the normal distribution over x.
      sum_squared_error <- sum((x - prior_pair$gamma_hat[1L])^2.0)
      
      # Compute w_ig" = L(Z_ig | gamma_hat_ig", delta_hat_squared_ig")
      # for the current feature in g".
      w_ig <- 1.0 / (2.0 * pi * prior_pair$delta_hat_squared[1L])^(length(x) / 2.0) * 
        exp(-sum_squared_error / (2.0 * prior_pair$delta_hat_squared[1L]))
      
      # Replace NaNs and other problematic values.
      if (!is.finite(w_ig)) w_ig <- 0.0
      
      return(data.table::data.table(
        "feature" = prior_pair$feature[1L],
        "w" = w_ig,
        "gamma_hat" = prior_pair$gamma_hat[1L],
        "delta_hat_squared" = prior_pair$delta_hat_squared[1L]
      ))
    },
    x = z$value
  )
  
  # Combine to data.table
  weight_data <- data.table::rbindlist(weight_data)
  
  # Compute gamma_star and delta_star_squared according to Johnson et al.
  gamma_star <- sum(weight_data$w * weight_data$gamma_hat) / sum(weight_data$w)
  delta_star_squared <- sum(weight_data$w * weight_data$delta_hat_squared) / sum(weight_data$w)
  
  return(data.table::data.table(
    "batch_id" = z[[batch_id_column]][1L],
    "feature" = z$feature[1L],
    "norm_shift" = gamma_star,
    "norm_scale" = sqrt(delta_star_squared),
    "n" = nrow(z)
  ))
}



..combat_lambda_prior <- function(delta_hat_squared) {
  # Computes the lambda prior from Johnson et al. See suppl. A3.1 for more
  # details. It seems that the supplement contains an error, where the prior is
  # created using V instead of V^2. Since V^2 appears in other code
  # implementations of combat (e.g. sva) and is actually found in other
  # publications, e.g. (https://arxiv.org/pdf/1605.01019.pdf). Manual
  # derivation confirms that V^2 should be used.
  v <- mean(delta_hat_squared)
  s_squared <- stats::var(x = delta_hat_squared)
  
  return((v^2.0 / s_squared) + 2.0)
}



..combat_theta_prior <- function(delta_hat_squared) {
  # Computes the theta prior from Johnson et al. See suppl. A3.1 for more
  # details.
  v <- mean(delta_hat_squared)
  s_squared <- stats::var(x = delta_hat_squared)
  
  return(v * (1.0 + (v^2.0 / s_squared)))
}



..combat_gamma_posterior <- function(
    n,
    tau_bar_squared,
    gamma_hat,
    delta_star_squared,
    gamma_bar) {
  # Computes conditional posterior for gamma, see equation 3.1 in Johnson et al.
  gamma_star <- (n * tau_bar_squared * gamma_hat + delta_star_squared * gamma_bar) /
    (n * tau_bar_squared + delta_star_squared)
  
  return(gamma_star)
}



..combat_delta_squared_posterior <- function(
    theta_bar,
    sum_squared_error,
    n,
    lambda_bar) {
  # Computes conditional posterior for delta_star_squared, see equation 3.1 in Johnson et al.
  delta_star_squared <- (theta_bar + 0.5 * sum_squared_error) / (0.5 * n + lambda_bar - 1.0)
  
  return(delta_star_squared)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# batch normalisation container ------------------------------------------------
setClass(
  "featureInfoParametersBatchNormalisationContainer",
  contains = "featureInfoParameters",
  slots = list(
    "method" = "character",
    "batch_parameters" = "ANY",
    "type" = "character",
    "available" = "logical"
  ),
  prototype = list(
    "method" = NA_character_,
    "batch_parameters" = NULL,
    "type" = NA_character_,
    "available" = NA
  )
)


.get_available_batch_normalisation_methods <- function(type = "all") {
  
  # Find all basic normalisation.
  available_methods <- .get_available_normalisation_methods(type = type)
  
  if (type %in% c("combat", "combat_non_parametric", "all")) {
    available_methods <- c(
      available_methods,
      .get_available_combat_parametric_normalisation_methods()
    )
  }
  
  if (type %in% c("combat", "combat_parametric", "all")) {
    available_methods <- c(
      available_methods,
      .get_available_combat_non_parametric_normalisation_methods()
    )
  }
  
  return(available_methods)
}



create_batch_normalisation_parameter_skeleton <- function(
    feature_info_list,
    feature_names = NULL,
    normalisation_method,
    .override_existing = FALSE
) {
  
  # Creates a skeleton for the provided batch normalisation method.
  
  # Determine feature names from the feature info list, if provided.
  if (is.null(feature_names)) feature_names <- names(feature_info_list)
  
  # Select only features that appear in the feature info list.
  feature_names <- intersect(names(feature_info_list), feature_names)
  
  # Skip step if no feature info objects are updated.
  if (is_empty(feature_names)) return(feature_info_list)
  
  # Check that method is applicable.
  .check_parameter_value_is_valid(
    x = normalisation_method,
    var_name = "normalisation_method",
    values = .get_available_batch_normalisation_methods()
  )
  
  # Update familiar info objects with a feature normalisation skeleton.
  updated_feature_info <- fam_lapply(
    X = feature_info_list[feature_names],
    FUN = .create_batch_normalisation_parameter_skeleton,
    method = normalisation_method,
    .override_existing = .override_existing
  )
  
  # Provide names for the updated feature info objects.
  names(updated_feature_info) <- feature_names
  
  # Replace list elements.
  feature_info_list[feature_names] <- updated_feature_info
  
  return(feature_info_list)
}



.create_batch_normalisation_parameter_skeleton <- function(
    feature_info,
    method,
    .override_existing = FALSE
) {
  
  # Check if normalisation data was already completed, and does not require
  # being determined anew.
  if (feature_info_complete(feature_info@batch_normalisation_parameters) && !.override_existing) {
    return(feature_info)
  } 
  
  # Pass to underlying function that constructs the skeleton.
  object <- ..create_batch_normalisation_parameter_skeleton(
    feature_name = feature_info@name,
    feature_type = feature_info@feature_type,
    method = method,
    available = is_available(feature_info)
  )
  
  # Update normalisation_parameters slot.
  feature_info@batch_normalisation_parameters <- object
  
  return(feature_info)
}



..create_batch_normalisation_parameter_skeleton <- function(
    method,
    feature_name,
    feature_type = "numeric",
    available = TRUE
) {
  # This is the lowest level function for creating batch normalisation parameter
  # skeletons.
  
  # If the feature type is not numeric, or the feature is not available, change
  # the main method to "none".
  if (feature_type != "numeric" || !available) method <- "none"
  
  # Generate container.
  object <- methods::new(
    "featureInfoParametersBatchNormalisationContainer",
    "method" = method,
    "type" = feature_type,
    "available" = available
  )
  
  # Set the name of the object.
  object@name <- feature_name
  
  # Update the familiar version.
  object <- add_package_version(object = object)
  
  return(object)
}



add_batch_normalisation_parameters <- function(
    cl = NULL,
    feature_info_list,
    data,
    verbose = FALSE
) {
  # Determine normalisation parameters and add them to the feature_info_list.
  
  # Find feature columns.
  feature_names <- get_feature_columns(x = data)
  
  # Sanity check.
  if (!(setequal(feature_names, get_available_features(feature_info_list = feature_info_list)))) {
    ..error_reached_unreachable_code(paste0(
      "add_transformation_parameters: features in data and in the feature_info_list are ",
      "expected to be the same, but were not."
    ))
  }
  
  # Identify methods specified by the containers for different features.
  container_batch_methods <- sapply(
    feature_info_list[feature_names],
    function(x) (x@batch_normalisation_parameters@method)
  )
  
  # Iterate over methods. The names of features that are assessed using the same
  # batch normalisation method are dispatched to ensure computation within that
  # feature set only.
  for (container_batch_method in unique(container_batch_methods)) {
    
    # Find feature names.
    batch_method_feature_names <- feature_names[container_batch_methods == container_batch_method]
    
    # Check if a ComBat method is used, and pre-compute the z-matrix, if it is.
    if (container_batch_method %in% .get_available_batch_normalisation_methods(type = "combat")) {
      
      # Pre-compute z-matrix. Otherwise we need to compute for each feature,
      # over and over again.
      z <- .compute_combat_batch_normalisation_z_matrix(
        data = data@data,
        feature_names = batch_method_feature_names
      )
      
      if (container_batch_method %in% .get_available_batch_normalisation_methods(type = "combat_parametric")) {
        # Obtain ComBat data using the parametric solver.
        batch_parameter_data <- .combat_iterative_parametric_bayes_solver(
          z = z,
          cl = cl,
          progress_bar = FALSE
        )
        
      } else {
        # Obtain ComBat data using the non-parametric solver.
        batch_parameter_data <- .combat_iterative_parametric_bayes_solver(
          z = z,
          cl = cl,
          progress_bar = FALSE
        )
      }
      
    } else {
      batch_parameter_data <- NULL
    }
    
    # Iterate over features.
    updated_feature_info <- fam_mapply(
      cl = cl,
      FUN = .add_batch_normalisation_parameters,
      feature_info = feature_info_list[batch_method_feature_names],
      data = data@data[, mget(batch_method_feature_names)],
      MoreArgs = list(
        "batch_column_data" = data@data[[get_id_columns("batch", single_column = TRUE)]],
        "batch_parameter_data" = batch_parameter_data
      ),
      progress_bar = verbose,
      chopchop = TRUE
    )
    
    # Provide names for the updated feature info objects.
    names(updated_feature_info) <- batch_method_feature_names
    
    # Replace list elements.
    feature_info_list[batch_method_feature_names] <- updated_feature_info
  }
  
  return(feature_info_list)
}



.add_batch_normalisation_parameters <- function(
    feature_info,
    data,
    batch_column_data,
    batch_parameter_data = NULL
) {
  # Pass to underlying function that adds the feature info.
  object <- add_feature_info_parameters(
    object = feature_info@batch_normalisation_parameters,
    data = data,
    batch_column_data = batch_column_data,
    batch_parameter_data = batch_parameter_data
  )
  
  # Update normalisation_parameters slot.
  feature_info@batch_normalisation_parameters <- object
  
  return(feature_info)
}



# add_feature_info_parameters (container, numeric) -----------------------------
setMethod(
  "add_feature_info_parameters",
  signature(object = "featureInfoParametersBatchNormalisationContainer", data = "numeric"),
  function(
    object,
    data,
    batch_column_data,
    ...
  ) {
    ## Populate container with batch parameter objects -------------------------
    
    # Find the batch identifiers in the data.
    batch_ids <- unique(batch_column_data)
    
    # Determine which normalisation objects already exist.
    visited_batch <- names(object@batch_parameters)
    
    # Remove batches that were already created.
    batch <- setdiff(batch_ids, visited_batch)
    
    if (length(batch) > 0L) {
      # Create batch objects.
      normalisation_objects <- fam_mapply(
        FUN = ..create_normalisation_parameter_skeleton,
        batch = batch,
        MoreArgs = list(
          "feature_name" = object@name,
          "feature_type" = object@type,
          "available" = object@available,
          "method" = object@method
        )
      )
      
      # Set names.
      names(normalisation_objects) <- batch
      
      # Append to the batch parameters.
      object@batch_parameters <- c(
        object@batch_parameters,
        normalisation_objects
      )
    }
    
    ## Determine batch parameters ----------------------------------------------
    
    # Select normalisation objects that have not been completed. 
    batch_to_update <- intersect(
      names(object@batch_parameters)[!sapply(object@batch_parameters, feature_info_complete)],
      batch_ids
    )
    
    if (length(batch_to_update) > 0L) {
      # Determine normalisation parameters within each batch.
      normalisation_objects <- fam_mapply(
        FUN = add_feature_info_parameters,
        object = object@batch_parameters[batch_to_update],
        data = split(data, f = batch_column_data)[batch_to_update],
        MoreArgs = list(...)
      )
      
      # Set names
      names(normalisation_objects) <- batch_to_update
      
      # Update batch parameters
      object@batch_parameters[batch_to_update] <- normalisation_objects
    }
    
    ## Check new none-class parameter sets -------------------------------------
    
    # Select parameter sets that are a) newly created, and b) were
    # updated. If these objects are none, for various reasons such as
    # small batches, attempt to replace them by an average object.
    batch_to_check <- intersect(batch, batch_to_update)
    
    if (length(batch_to_check) > 0L && object@method != "none") {
      # Generate a new object.
      replacement_normalisation_object <- ...collect_and_aggregate_normalisation_info(
        object_list = object@batch_parameters,
        instance_mask = rep_len(TRUE, length(object@batch_parameters)),
        feature_name = object@name
      )$parameters
      
      # Identify which new objects are none objects.
      batch_to_check <- batch_to_check[sapply(
        object@batch_parameters[batch_to_check],
        is,
        class2 = "featureInfoParametersNormalisationNone"
      )]
      
      if (length(batch_to_check) > 0L) {
        
        # Add batch attribute to normalisation objects.
        normalisation_objects <- lapply(
          batch_to_check,
          function(x, object) {
            # Add batch name
            object@batch <- x
            
            return(object)
          },
          object = replacement_normalisation_object
        )
        
        # Update names.
        names(normalisation_objects) <- batch_to_check
        
        # Replace these objects.
        object@batch_parameters[batch_to_check] <- normalisation_objects
      }
    }
    
    # Check that all objects are complete.
    if (!all(sapply(object@batch_parameters, feature_info_complete))) {
      ..error_reached_unreachable_code(paste0(
        "add_feature_info_parameters,featureInfoParametersBatchNormalisationContainer, data.table: ",
        "all normalisation objects in the container should be complete, but some were not."
      ))
    }
    
    # Mark complete.
    object@complete <- TRUE
    
    return(object)
  }
)



# apply_feature_info_parameters (container, ANY) -------------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(object = "featureInfoParametersBatchNormalisationContainer", data = "ANY"),
  function(
    object,
    data,
    batch_data,
    ...
  ) {
    
    if (object@method == "none") return(data)
    
    # Suppress NOTES due to non-standard evaluation in data.table
    batch_normalisation_ordering_id <- NULL
    
    # Find the batch identifiers in the data.
    batch_ids <- unique(batch_data)
    
    # Sanity check: batch normalisation objects are available for all
    # batch identifiers.
    if (!all(batch_ids %in% names(object@batch_parameters))) {
      ..error_reached_unreachable_code(paste0(
        "apply_feature_info_parameters,featureInfoParametersBatchNormalisationContainer,ANY: ",
        "Batch normalisation objects are missing for one or more batches."
      ))
    }
    
    # Sanity check: all relevant batch normalisation objects are
    # complete.
    if (!all(sapply(object@batch_parameters[batch_ids], feature_info_complete))) {
      ..error_reached_unreachable_code(paste0(
        "apply_feature_info_parameters,featureInfoParametersBatchNormalisationContainer,ANY: ",
        "One or more batch normalisation objects are incomplete and may be missing parameters."
      ))
    }
    
    # Avoid updating by reference.
    data <- data.table::data.table(
      "feature" = data,
      "batch_id" = batch_data
    )
    
    # Set feature name.
    data.table::setnames(data, old = "feature", new = object@name)
    
    # Insert ordering variable.
    data[, "batch_normalisation_ordering_id" := .I]
    
    # Transform per batch.
    data <- fam_mapply(
      FUN = apply_feature_info_parameters,
      object = object@batch_parameters[batch_ids],
      data = split(data, by = "batch_id")[batch_ids],
      MoreArgs = list(...)
    )
    
    # Combine into a single data.table.
    data <- data.table::rbindlist(data, use.names = TRUE)
    
    # Order by ordering variable to maintain original order.
    data <- data[order(batch_normalisation_ordering_id)]
    
    # Extract the feature values and return.
    return(data[[object@name]])
  }
)



..collect_and_aggregate_batch_normalisation_info <- function(
    feature_info_list,
    instance_mask,
    feature_name
) {
  # Aggregate batch normalisation parameters. This function exists so that it
  # can be tested as part of a unit test.
  
  # Find all batches in the parameters.
  batch_ids <- unique(unlist(lapply(
    feature_info_list,
    function(x) {
      # Return NULL if parameters are missing.
      if (is.null(x@batch_normalisation_parameters)) return(NULL)
      
      return(names(x@batch_normalisation_parameters@batch_parameters))
    }
  )))
  
  # Main batch method
  main_batch_method <- unique(unlist(lapply(
    feature_info_list,
    function(x) {
      # Return NULL if parameters are missing.
      if (is.null(x@batch_normalisation_parameters)) return(NULL)
      
      return(x@batch_normalisation_parameters@method)
    }
  )))
  
  # Create parameter container
  batch_parameter_container <- ..create_batch_normalisation_parameter_skeleton(
    method = main_batch_method[1L],
    feature_name = feature_name
  )
  
  if (!any(instance_mask)) {
    # Create placeholder objects.
    batch_normalisation_objects <- lapply(
      batch_ids,
      function(batch, feature_name) {
        ..create_normalisation_parameter_skeleton(
          feature_name = feature_name,
          method = "none",
          batch = batch
        )
      },
      feature_name = feature_name
    )
    
    # Add names.
    names(batch_normalisation_objects) <- batch_ids
    
    # Add to container.
    batch_parameter_container@batch_parameters <- batch_normalisation_objects
    
    return(list(
      "parameters" = batch_parameter_container,
      "instance_mask" = instance_mask
    ))
  }
  
  # Initialise list of objects.
  batch_normalision_object_list <- list()
  
  # Iterate over batch identifiers and collect all batch identifiers.
  for (batch_id in batch_ids) {
    
    # Collect batch normalisation objects.
    batch_normalisation_objects <- lapply(
      feature_info_list[instance_mask],
      function(x, batch_id) {
        # Return NULL if parameters are missing.
        if (is.null(x@batch_normalisation_parameters)) return(NULL)
        
        # Return parameters for the specific batch.
        return(x@batch_normalisation_parameters@batch_parameters[[batch_id]])
      },
      batch_id = batch_id
    )
    
    # Collect aggregate object.
    batch_normalision_object_list[[batch_id]] <- ...collect_and_aggregate_normalisation_info(
      object_list = batch_normalisation_objects,
      instance_mask = rep_len(TRUE, length(batch_normalisation_objects)),
      feature_name = feature_name
    )$parameters
    
    # Set batch name.
    batch_normalision_object_list[[batch_id]]@batch <- batch_id
  }
  
  # Add batch objects to container.
  batch_parameter_container@batch_parameters <- batch_normalision_object_list
  
  return(list(
    "parameters" = batch_parameter_container,
    "instance_mask" = instance_mask
  ))
}



.check_batch_normalisation_assumptions <- function(
  data,
  normalisation_method
) {
  # Check that batches do not differ in outcome data. We can use the following
  # tests: 
  #
  # * Continuous: Kruskal-Wallis-test.
  # * Binomial / multinomial: Chi-squared test.
  # * Survival: Log-rank test.
  
  if (all(normalisation_method == "none")) return(invisible(TRUE))
  
  x <- data@data[, mget(get_outcome_columns(data))]
  g <- data@data[[get_id_columns("batch", single_column = TRUE)]]
  
  # Determine the number of batches.
  n_groups <- data.table::uniqueN(g)
  # Check that 2 (or more groups) are present.
  if (n_groups < 2L) return(invisible(TRUE))
  
  if (data@outcome_type == "continuous") {
    h <- tryCatch(
      stats::kruskal.test(
        x = x[[1L]],
        g = g, 
        na.action = "na.omit"
      ),
      error = identity
    )
    
    # Check if the test statistic could be computed.
    if (inherits(h, "error")) return(invisible(TRUE))
    p_value <- h$p.value
    
  } else if (data@outcome_type %in% c("binomial", "multinomial")) {
    h <- tryCatch(
      stats::chisq.test(x = x, y = g),
      error = identity
    )
    
    # Check if the test statistic could be computed.
    if (inherits(h, "error")) return(invisible(TRUE))
    p_value <- h$p.value
    
  } else if (data@outcome_type == "survival") {
    
    # Determine chi-square of log-rank test
    chi_sq <- tryCatch(
      survival::survdiff(
        survival::Surv(time = outcome_time, event = outcome_event) ~ group,
        data = data.table::data.table(
          outcome_time = x$outcome_time,
          outcome_event = x$outcome_event,
          group = g
        ),
        subset = NULL,
        na.action = "na.omit"
      )$chisq,
      error = identity
    )
    
    # Check if the test statistic could be computed. Causes could be lack of
    # events, no events beyond the first time point, etc.
    if (inherits(chi_sq, "error")) return(invisible(TRUE))
    
    # Derive  p-value
    p_value  <- stats::pchisq(
      q = chi_sq,
      df = n_groups - 1L,
      lower.tail = FALSE
    )
    
  } else {
    ..error_outcome_type_not_implemented(data@outcome_type)
  }
  
  if (p_value < 0.05) {
    logger_warning(
      paste0(
        "One or more batches have a statistically significant (p < 0.05) different outcome ",
        "compared to other batches. Note: a statistically significant outcome does ",
        "not mean that the difference is actually relevant. However, please assert ",
        "that batch normalisation does not remove important differences between batches."
      ),
      warn_class = "familiar_batch_outcome_difference"
    )
    return(invisible(FALSE))
  }
  
  
  return(invisible(TRUE))
}

#' @include FamiliarS4Generics.R


stop_or_warn <- function(message, as_error = TRUE, call = rlang::caller_env()) {
  if (as_error) {
    ..error(message, call = call)
  } else {
    ..warning(message, call = call)
  }
}



is.unset <- function(x) {
  if (is.null(x)) {
    return(TRUE)
  } else if (is.waive(x)) {
    return(TRUE)
  } else if (length(x) == 0L) { 
    return(TRUE)
  } else if (all(is.na(x))) {
    return(TRUE)
  } else {
    return(FALSE)
  }
}



compute_univariable_p_values <- function(cl = NULL, data_obj, feature_columns) {
  outcome_type <- data_obj@outcome_type
  outcome_columns <- get_outcome_columns(data_obj)

  if (outcome_type == "survival") {
    univariate_fun <- .univariate_cox_regression_test
  } else if (outcome_type == "continuous") {
    univariate_fun <- .univariate_linear_regression_test
  } else if (outcome_type == "binomial") {
    univariate_fun <- .univariate_binomial_logistic_regression_test
  } else if (outcome_type == "multinomial") {
    univariate_fun <- .univariate_multinomial_logistic_regression_test
  } else if (outcome_type == "competing_risk") {
    ..error_outcome_type_not_implemented(outcome_type)
  } else {
    ..error_no_known_outcome_type(outcome_type)
  }

  # Obtain p-values
  coefficient_p_values <- fam_sapply(
    cl = cl,
    assign = NULL,
    X = data_obj@data[, mget(feature_columns)],
    FUN = univariate_fun,
    progress_bar = FALSE,
    outcome_data = data_obj@data[, mget(outcome_columns)],
    chopchop = TRUE
  )

  return(coefficient_p_values)
}



..univariate_test_variable_encoding <- function(x, insert_intercept = FALSE) {
  n <- length(x)

  if (is.factor(x)) {
    # Categorical variables are encoded as numeric levels (ordinal), or using
    # one-hot-encoding (nominal).
    if (is.ordered(x)) {
      x <- list(as.numeric(x))
    } else {
      # Drop unused levels.
      x <- droplevels(x)

      # Dummy encoding of categorical variable.
      level_names <- levels(x)
      level_count <- nlevels(x)

      x <- lapply(level_names[2L:level_count], function(ii, x) (as.numeric(x == ii)), x = x)
    }
  } else {
    # Numeric variables are only stored as a list.
    x <- list(x)
  }

  # Set names
  names(x) <- paste0("name_", seq_along(x))

  if (insert_intercept) {
    x <- c(x, list("intercept__" = numeric(n) + 1.0))
  }

  return(data.table::as.data.table(x))
}



.univariate_cox_regression_test <- function(x, outcome_data) {
  # Cox regression model for univariate analysis

  # Check if any data was provided.
  if (length(x) == 0L) {
    return(NA_real_)
  }

  # Extract response
  y_time <- outcome_data$outcome_time
  y_event <- outcome_data$outcome_event

  # Remove missing elements.
  valid_elements <- is.finite(x) & is.finite(y_time) & is.finite(y_event)
  if (sum(valid_elements) <= 1L) return(NA_real_)

  # Keep only valid elements.
  x <- x[valid_elements]
  y_time <- y_time[valid_elements]
  y_event <- y_event[valid_elements]

  # Check if the feature column is singular.
  if (is_singular_data(x)) return(NA_real_)

  # Bind data.
  data <- cbind(
    ..univariate_test_variable_encoding(x, insert_intercept = FALSE),
    data.table::data.table(
      "time" = y_time,
      "event" = y_event
    )
  )

  # Construct model
  model <- tryCatch(
    coxph(
      Surv(time, event) ~ .,
      data = data
    ),
    error = identity
  )

  # Check if the model did not converge.
  if (inherits(model, "error")) return(NA_real_)

  # Compute z-statistic.
  z <- .compute_z_statistic(model)

  # Compute the p-value from a t-distribution
  p_value <- 2.0 * (1.0 - stats::pt(abs(z), df = length(x)))

  # Check if the value is finite.
  if (!any(is.finite(p_value))) return(NA_real_)

  # Return p-value.
  return(unname(min(p_value, na.rm = TRUE)))
}



.univariate_linear_regression_test <- function(x, outcome_data) {
  # Gaussian regression for univariable analysis

  # Check if any data was provided.
  if (length(x) == 0L) return(NA_real_)

  # Extract response
  y <- outcome_data$outcome

  # Remove missing elements.
  valid_elements <- is.finite(x) & is.finite(y)
  if (sum(valid_elements) <= 1L) return(NA_real_)

  # Keep only valid elements.
  x <- x[valid_elements]
  y <- y[valid_elements]

  # Check if the feature column is singular.
  if (is_singular_data(x)) return(NA_real_)

  # Bind data.
  data <- cbind(
    ..univariate_test_variable_encoding(x, insert_intercept = TRUE),
    data.table::data.table("response" = y)
  )

  if (require_package("fastglm", message_type = "silent")) {
    # Using fastglm.

    predictors <- setdiff(colnames(data), "response")

    # Construct model.
    model <- do.call_with_handlers(
      fastglm::fastglm,
      args = list(
        "x" = as.matrix(data[, mget(predictors)]),
        "y" = data$response,
        "family" = stats::gaussian(link = "identity"),
        "method" = 3L
      )
    )

    # Check for errors.
    if (!is.null(model$error)) return(NA_real_)

    # Extract the model itself.
    model <- model$value
    
  } else {
    # Construct model
    model <- tryCatch(
      stats::glm(response ~ . - 1,
        data = data,
        family = stats::gaussian(link = "identity")
      ),
      error = identity
    )

    # Check if the model did not converge.
    if (inherits(model, "error")) return(NA_real_)
  }

  # Compute z-statistic.
  z <- .compute_z_statistic(model)

  # Remove intercept (if present).
  z <- z[!names(z) %in% "intercept__"]

  # Compute the p-value from a t-distribution
  p_value <- 2.0 * (1.0 - stats::pt(abs(z), df = length(x)))

  # Check if the value is finite.
  if (!any(is.finite(p_value))) return(NA_real_)

  # Return p-value.
  return(unname(min(p_value, na.rm = TRUE)))
}



.univariate_binomial_logistic_regression_test <- function(x, outcome_data) {
  # Binomial model for univariable analysis using logistic regression

  # Check if any data was provided.
  if (length(x) == 0L) return(NA_real_)

  # Extract response
  y <- outcome_data$outcome

  # Remove missing elements.
  valid_elements <- is.finite(x) & is.finite(y)
  if (sum(valid_elements) <= 1L) return(NA_real_)

  # Keep only valid elements.
  x <- x[valid_elements]
  y <- y[valid_elements]

  # Check if the feature column is singular.
  if (is_singular_data(x)) return(NA_real_)

  # Bind data.
  data <- cbind(
    ..univariate_test_variable_encoding(x, insert_intercept = TRUE),
    data.table::data.table("response" = y)
  )

  if (require_package("fastglm", message_type = "silent")) {
    # Using fastglm.
    predictors <- setdiff(colnames(data), "response")

    # Construct model.
    model <- do.call_with_handlers(
      fastglm::fastglm,
      args = list(
        "x" = as.matrix(data[, mget(predictors)]),
        "y" = as.numeric(data$response) - 1,
        "family" = stats::binomial(link = "logit"),
        "method" = 3L
      )
    )

    # Check for errors.
    if (!is.null(model$error)) return(NA_real_)

    # Extract the model itself.
    model <- model$value

    # Check for Hauck-Donner effect.
    if (approximately(model$deviance, 0.0)) {
      # Create a model without an intercept.
      model_2 <- do.call_with_handlers(
        fastglm::fastglm,
        args = list(
          "x" = as.matrix(data[, mget(setdiff(predictors, "intercept__"))]),
          "y" = as.numeric(data$response) - 1,
          "family" = stats::binomial(link = "logit"),
          "method" = 3L
        )
      )

      # Check for errors, and only attempt to replace model by model_2 if no
      # errors are present.
      if (is.null(model_2$error)) {
        # Extract the model itself.
        model_2 <- model_2$value

        # Replace by intercept-less model in case the deviance is higher.
        if (model_2$deviance > model$deviance) model <- model_2
      }
    }
    
  } else {
    # Construct model
    model <- suppressWarnings(tryCatch(
      stats::glm(
        response ~ . - 1,
        data = data,
        family = stats::binomial(link = "logit")
      ),
      error = identity
    ))

    # Check if the model did not converge
    if (inherits(model, "error")) return(NA_real_)
  }

  # Compute z-statistic
  z <- .compute_z_statistic(model)

  # Remove intercept (if present).
  z <- z[names(z) != "intercept__"]

  # Compute the p-value from a t-distribution
  p_value <- 2.0 * (1.0 - stats::pt(abs(z), df = length(x)))

  # Check if the value is finite.
  if (!any(is.finite(p_value))) return(NA_real_)

  # Return p-value.
  return(unname(min(p_value, na.rm = TRUE)))
}



.univariate_multinomial_logistic_regression_test <- function(x, outcome_data) {
  # Multinomial model for univariable analysis using logistic regression

  # Check if any data was provided.
  if (length(x) == 0L) return(NA_real_)

  # Extract response
  y <- outcome_data$outcome

  # Remove missing elements.
  valid_elements <- is.finite(x) & is.finite(y)
  if (sum(valid_elements) <= 1L) return(NA_real_)

  # Keep only valid elements.
  x <- x[valid_elements]
  y <- y[valid_elements]

  # Check if the feature column is singular.
  if (is_singular_data(x)) return(NA_real_)

  # Bind data.
  data <- cbind(
    ..univariate_test_variable_encoding(x, insert_intercept = TRUE),
    data.table::data.table("response" = y)
  )

  # Check if the package is installed and attached.
  require_package(
    x = "nnet",
    purpose = "to determine univariate p-values for multinomial outcomes"
  )

  # Construct model
  model <- do.call_with_handlers(
    nnet::multinom,
    args = list(
      response ~ . - 1,
      "data" = data,
      "maxit" = 500L
    )
  )

  # Check for errors.
  if (!is.null(model$error)) return(NA_real_)

  # Extract the model itself.
  model <- model$value

  # Check if the model converged. NNET might fail to converge if the
  # Hauck-Donner effect is present.
  if (model$convergence == 1L) {
    model_2 <- do.call_with_handlers(
      nnet::multinom,
      args = list(
        response ~ . - 1 - intercept__,
        "data" = data,
        "maxit" = 500L
      )
    )

    # Check for errors, and only attempt to replace model by model_2 if no
    # errors are present.
    if (is.null(model_2$error)) {
      # Extract the model itself.
      model_2 <- model_2$value

      # If the Hauck-Donner effect is indeed present, the new model without the
      # intercept should converge quickly.
      if (model_2$convergence == 0L) model <- model_2
    }
  }

  # Compute z-statistic
  z <- .compute_z_statistic(model)

  # Remove intercept (if present).
  predictors <- setdiff(colnames(data), c("response", "intercept__"))
  z <- if (is.matrix(z)) as.vector(z[, predictors]) else z[predictors]

  # Compute the p-value from a t-distribution
  p_value <- 2.0 * (1.0 - stats::pt(abs(z), df = length(x) * (nlevels(y) - 1L)))

  # Check if the value is finite.
  if (!any(is.finite(p_value))) return(NA_real_)

  # Return p-value.
  return(unname(min(p_value, na.rm = TRUE)))
}



.get_available_icc_types <- function() {
  return(c("1", "2", "3"))
}



compute_icc <- function(x, feature, id_data, type = "1") {
  # We start from the following equation: xij = mu + a_i + b_j + e_ij, with mu
  # the population mean, a_i the rater-dependent change, b_j the
  # subject-dependent change and eij an error with mean 0.
  #
  # * type = "1": ICC 1: We assume that each rated value is cased by a random
  # effect of the rater: i.e. the actual rotation is not associated with an
  # systematic change in value, and there is no interaction between rater and
  # subject. [Shrout 1979]. This means that a_i has mean 0, so that wij = ai +
  # eij. Put differently, raters are randomly chosen from a larger population
  # for each sample.
  #
  # * type = "2": ICC 2: There is a panel of raters, and the entire panel
  # evaluates each sample. However, the panel is assumed to be part of a larger
  # population of raters. Raters are assumed to be systematically biased.

  # * type = "3": ICC 3: There is a panel of raters and the entire panel evaluates each
  # subject. The panel is the entire population and raters are assumed to be
  # systematically biased.

  # Suppress NOTES due to non-standard evaluation in data.table
  value <- mu <- bj <- ai <- NULL

  # Determine identifier columns.
  sample_id_columns <- get_id_columns(id_depth = "series")
  repetition_id_column <- get_id_columns(single_column = "repetition_id")

  # Create data table from x and combine with id_data
  data <- data.table::data.table("value" = x)
  data <- cbind(id_data, data)

  # Calculate each parameter in the equation
  data[, "mu" := mean(value, na.rm = TRUE)]
  data[, "bj" := mean(value, na.rm = TRUE) - mu, by = sample_id_columns]
  data[, "ai" := mean(value, na.rm = TRUE) - mu, by = repetition_id_column]
  data[, "eij" := value - mu - bj - ai]

  # Calculate
  n_samples <- data.table::uniqueN(data, by = sample_id_columns)
  n_raters <- data.table::uniqueN(data, by = repetition_id_column)

  # Calculate mean squared errors: msb between subjects (bj), msj between raters
  # (ai), mse of error (eij) and msw of error with rater (ai + eij).
  if (n_samples > 1L) {
    msb <- sum(data$bj^2.0, na.rm = TRUE) / (n_samples - 1.0)
  }

  if (type == "1") {
    # Calculate mean squared of error with rater
    msw <- (sum(data$eij^2.0, na.rm = TRUE) + sum(data$ai^2.0, na.rm = TRUE)) /
      (n_samples * (n_raters - 1.0))

    # Calculate icc for individual rater and rater panel
    if (msb == 0.0 && msw == 0.0) {
      icc <- 1.0
      icc_panel <- 1.0
      icc_ci_low <- 1.0
      icc_ci_up <- 1.0
      icc_panel_ci_low <- 1.0
      icc_panel_ci_up <- 1.0
      
    } else {
      icc <- (msb - msw) / (msb + (n_raters - 1.0) * msw)
      icc_panel <- (msb - msw) / msb

      # Fisher score
      s_fisher <- msb / msw
      s_fisher_low <- s_fisher / stats::qf(0.975, n_samples - 1L, n_samples * (n_raters - 1L))
      s_fisher_up <- s_fisher / stats::qf(0.025, n_samples - 1L, n_samples * (n_raters - 1L))

      # Calcuate confidence intervals from fisher score
      icc_ci_low <- (s_fisher_low - 1.0) / (s_fisher_low + n_raters - 1.0)
      icc_ci_up <- (s_fisher_up - 1.0) / (s_fisher_up + n_raters - 1.0)
      icc_panel_ci_low <- 1.0 - 1.0 / s_fisher_low
      icc_panel_ci_up <- 1.0 - 1.0 / s_fisher_up
    }
  }

  if (type == "2") {
    # Calculate mean squared error (mse) and mean squared rater error (msj)
    msj <- sum(data$ai^2.0, na.rm = TRUE) / (n_raters - 1.0)
    mse <- sum(data$eij^2.0, na.rm = TRUE) / ((n_samples - 1.0) * (n_raters - 1.0))

    # Calculate icc for individual rater and rater panel
    if (msb == 0.0 && mse == 0.0) {
      icc <- 1.0
      icc_panel <- 1.0
      icc_ci_low <- 1.0
      icc_ci_up <- 1.0
      icc_panel_ci_low <- 1.0
      icc_panel_ci_up <- 1.0
      
    } else {
      icc <- (msb - mse) / (msb + (n_raters - 1.0) * mse + (n_raters / n_samples) * (msj - mse))
      icc_panel <- (msb - mse) / (msb + (msj - mse) / n_samples)

      # Determine confidence intervals
      vn <- (n_raters - 1.0) * (n_samples - 1.0) * 
        (n_raters * icc * msj / mse + n_samples * (1.0 + (n_raters - 1.0) * icc) - n_raters * icc)^2.0
      vd <- (n_samples - 1.0) * n_raters^2.0 * icc^2.0 * (msj / mse)^2.0 + 
        (n_samples * (1.0 + (n_raters - 1.0) * icc) - n_raters * icc)^2.0
      v <- vn / vd
      thresh_low <- stats::qf(0.975, n_samples - 1L, v)
      thresh_up <- stats::qf(0.025, n_samples - 1L, v)

      # Calcuate confidence intervals from fisher score
      icc_ci_low <- n_samples * (msb - thresh_low * mse) / 
        (thresh_low * (n_raters * msj + (n_raters * n_samples - n_raters - n_samples) * mse) + n_samples * msb)
      icc_ci_up <- n_samples * (msb - thresh_up * mse) / 
        (thresh_up * (n_raters * msj + (n_raters * n_samples - n_raters - n_samples) * mse) + n_samples * msb)
      icc_panel_ci_low <- icc_ci_low * n_raters / (1.0 + icc_ci_low * (n_raters - 1.0))
      icc_panel_ci_up <- icc_ci_up * n_raters / (1.0 + icc_ci_up * (n_raters - 1.0))
    }
  }

  if (type == "3") {
    # Calculate mean squared error (mse)
    mse <- sum(data$eij^2.0, na.rm = TRUE) / ((n_samples - 1.0) * (n_raters - 1.0))

    # Calculate icc for individual rater and rater panel
    if (msb == 0.0 && mse == 0.0) {
      icc <- 1.0
      icc_panel <- 1.0
      icc_ci_low <- 1.0
      icc_ci_up <- 1.0
      icc_panel_ci_low <- 1.0
      icc_panel_ci_up <- 1.0
      
    } else {
      icc <- (msb - mse) / (msb + (n_raters - 1.0) * mse)
      icc_panel <- (msb - mse) / msb

      # Fisher score
      s_fisher <- msb / mse
      s_fisher_low <- s_fisher / stats::qf(0.975, n_samples - 1L, (n_samples - 1L) * (n_raters - 1L))
      s_fisher_up <- s_fisher / stats::qf(0.025, n_samples - 1L, (n_samples - 1L) * (n_raters - 1L))

      # Calcuate confidence intervals from fisher score
      icc_ci_low <- (s_fisher_low - 1.0) / (s_fisher_low + n_raters - 1.0)
      icc_ci_up <- (s_fisher_up - 1.0) / (s_fisher_up + n_raters - 1.0)
      icc_panel_ci_low <- 1.0 - 1.0 / s_fisher_low
      icc_panel_ci_up <- 1.0 - 1.0 / s_fisher_up
    }
  }

  if (icc == 1.0) {
    if (!is.finite(icc_ci_low)) icc_ci_low <- 1.0
    if (!is.finite(icc_ci_up)) icc_ci_up <- 1.0
  }

  if (icc_panel == 1.0) {
    if (!is.finite(icc_ci_low)) icc_panel_ci_low <- 1.0
    if (!is.finite(icc_ci_up)) icc_panel_ci_up <- 1.0
  }

  return(data.table::data.table(
    "feature" = feature,
    "icc" = icc,
    "icc_low" = icc_ci_low,
    "icc_up" = icc_ci_up,
    "icc_panel" = icc_panel,
    "icc_panel_low" = icc_panel_ci_low,
    "icc_panel_up" = icc_panel_ci_up
  ))
}



.file_extension <- function(x) {
  # Find the file extension by extracting the substring to the right of the
  # final period.

  # Find the position of the period preceding the file extension.
  indicator <- tail(
    gregexpr(pattern = ".", text = x, fixed = TRUE)[[1L]],
    n = 1L
  )

  # Check when period (.) is not found in x.
  if (indicator == -1L) return("")

  # Find the extension
  extension <- substr(x, start = indicator + 1L, stop = nchar(x))

  return(extension)
}



harmonic_p_value <- function(x) {
  if (!is_package_installed("harmonicmeanp")) return(NA_real_)

  if (data.table::is.data.table(x)) {
    x <- x$p_value

    # Fix numeric issues due to very small p-values.
    x[x < 1.0E-15] <- 1.0E-15

    return(list("p_value" = harmonicmeanp::p.hmp(x, L = length(x))))
    
  } else {
    # Fix numeric issues due to very small p-values.
    x[x < 1.0E-15] <- 1.0E-15

    return(harmonicmeanp::p.hmp(x, L = length(x)))
  }
}



get_mode <- function(x) {
  # Ken Williams:
  # https://stackoverflow.com/questions/2547402/is-there-a-built-in-function-for-finding-the-mode
  ux <- unique(x)
  ux <- ux[which.max(tabulate(match(x, ux)))]

  return(ux)
}



get_estimate <- function(x, na.rm = TRUE) {
  # Remove NA values.
  if (na.rm) x <- x[!is.na(x)]

  if (is.numeric(x)) {
    # Determine the estimate.
    if (length(x) > 0L) {
      y <- mean(x)
    } else {
      y <- NA_real_
    }
    
  } else {
    # Determine the estimate.
    if (length(x) > 0L) {
      y <- get_mode(x)
    } else {
      y <- NA
    }
  }

  return(y)
}


.sanitise_dots <- function(class, ...) {
  dots <- list(...)

  if (length(dots) == 0L) return(dots)

  slot_names <- names(methods::getSlots(class))

  return(dots[intersect(names(dots), slot_names)])
}



get_placeholder_vimp_table <- function(
    vimp_method,
    run_table = NULL
) {
  # Set vimp method.
  vimp_object <- methods::new(
    "vimpTable",
    vimp_method = vimp_method
  )

  # Set run table.
  vimp_object@run_table <- run_table

  # Set package version.
  vimp_object <- add_package_version(vimp_object)

  return(vimp_object)
}



get_id_columns <- function(id_depth = "repetition", single_column = NULL) {
  # Check that until_depth is correctly specified.
  if (!id_depth %in% c("batch", "sample", "series", "repetition")) {
    ..error_value_not_allowed(
      x = id_depth, 
      var_name = "id_depth", 
      values = c("batch", "sample", "series", "repetition")
    )
  }

  if (is.null(single_column)) {
    # Generate the names of the non-feature columns
    id_columns <- switch(
      id_depth,
      "batch" = "batch_id",
      "sample" = c("batch_id", "sample_id"),
      "series" = c("batch_id", "sample_id", "series_id"),
      "repetition" = c("batch_id", "sample_id", "series_id", "repetition_id")
    )
    
  } else {
    id_columns <- switch(
      single_column,
      "batch" = "batch_id",
      "sample" = "sample_id",
      "series" = "series_id",
      "repetition" = "repetition_id"
    )
  }

  return(id_columns)
}



get_object_file_name <- function(
    project_id,
    data_id = NULL,
    run_id = NULL,
    ensemble_data_id = NULL,
    ensemble_run_id = NULL,
    learner = NULL,
    vimp_method = NULL,
    name = NULL,
    object_type,
    with_extension = TRUE,
    dir_path = NULL
) {

  if (object_type == "familiarModel") {
    # For familiarModel objects

    if (is.null(learner) || is.null(vimp_method)) {
      ..error_reached_unreachable_code("missing arguments")
    }
    
    output_str <- c(project_id, learner, vimp_method, data_id, run_id, "model")
    
  } else if (object_type == "familiarNoveltyDetector") {
    # For familiarNoveltyDetector objects
    
    if (is.null(learner) || is.null(vimp_method)) {
      ..error_reached_unreachable_code("missing arguments")
    }
    
    output_str <- c(project_id, learner, vimp_method, data_id, run_id, "detector")
    
  } else if (object_type == "familiarData") {
    # For familiarData objects

    if (
      is.null(learner) || is.null(vimp_method) ||
      is.null(data_id) || is.null(run_id) || is.null(name) ||
      is.null(ensemble_data_id) || is.null(ensemble_run_id)
    ) {
      ..error_reached_unreachable_code("missing arguments")
    }

    output_str <- c(
      project_id, learner, vimp_method, data_id, run_id, ensemble_data_id, ensemble_run_id, name, "data"
    )
  
  } else if (object_type == "familiarCollection") {
    
    if (
      ((is.null(data_id) || is.null(run_id)) && is.null(name))
    ) {
      ..error_reached_unreachable_code("missing arguments")
    }
    
    output_str <- c(project_id)
    if (is.null(name)) {
      output_str <- c(output_str, data_id, run_id)
    } else {
      output_str <- c(output_str, name)
    }
    
    output_str <- c(output_str, "collection")
    
  } else if (object_type == "vimpTable") {
    
    if (is.null(vimp_method)) {
      ..error_reached_unreachable_code("missing arguments")
    }
    
    output_str <- c(project_id, vimp_method, data_id, run_id, "vimp")
    
  } else if (object_type == "featureInfo") {
    # Complete feature info objects.
    output_str <- c(project_id, data_id, run_id, "feature_info")
    
  } else if (object_type == "genericFeatureInfo") {
    # Generic feature info objects.
    output_str <- c(project_id, "generic_feature_info")
  
  } else if (object_type == "hyperparametersVimp") {
    
    if (is.null(vimp_method)) {
      ..error_reached_unreachable_code("missing arguments")
    }
    
    output_str <- c(project_id, data_id, run_id, vimp_method, "vimp_hyperparameters")
    
  } else if (object_type == "hyperparametersLearner") {
    
    if (is.null(vimp_method) || is.null(learner)) {
      ..error_reached_unreachable_code("missing arguments")
    }
    
    output_str <- c(project_id, data_id, run_id, learner, vimp_method, "learner_hyperparameters")
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown object_type: ", object_type))
  }

  # Strip any NA entries, and collapse to single string.
  output_str <- output_str[!is.na(output_str)]
  output_str <- paste(output_str, collapse = "_")
  
  # Add extension
  if (with_extension) {
    output_str <- paste0(output_str, ".RDS")
  }

  # Join with dir_path
  if (!is.null(dir_path)) {
    output_str <- normalizePath(file.path(dir_path, output_str), mustWork = FALSE)
  }

  return(output_str)
}



get_object_dir_path <- function(
    dir_path, 
    object_type, 
    learner = NULL, 
    vimp_method = NULL
) {
  # Generate the directory path to an object

  if (!object_type %in% c(
    "familiarModel", "familiarEnsemble", "familiarData", "familiarCollection"
  )) {
    ..error("The object type was not recognised.")
  }

  if (object_type %in% c("familiarModel", "familiarEnsemble")) {
    if (is.null(learner) && is.null(vimp_method)) {
      return(normalizePath(file.path(dir_path), mustWork = FALSE))
    } else {
      return(normalizePath(file.path(dir_path, learner, vimp_method), mustWork = FALSE))
    }
    
  } else if (object_type %in% c("familiarData", "familiarCollection")) {
    return(normalizePath(file.path(dir_path)))
  }
}


extract_from_slot <- function(
    object_list, 
    slot_name, 
    slot_element = NULL, 
    na.rm = FALSE
) {
  # Extracts values from a slot in an object. Providing slot_element allows
  # extraction from a list in a slot

  if (is.null(slot_element)) {
    slot_values <- sapply(object_list, slot, name = slot_name)
    
  } else {
    slot_values <- sapply(
      object_list, 
      function(object, slot_name, slot_element) {
        list_element <- slot(object = object, name = slot_name)
        if (!is.list(list_element)) {
          return(NULL)
        } else {
          return(list_element[[slot_element]])
        }
      }, 
      slot_name = slot_name,
      slot_element = slot_element
    )
  }

  if (na.rm) {
    # First remove NULL
    slot_values <- slot_values[!sapply(slot_values, is.null)]

    # Then remove NA
    if (
      is.numeric(slot_values) ||
      is.logical(slot_values) ||
      is.character(slot_values)
    ) {
      slot_values <- slot_values[!sapply(slot_values, is.na)]
    }

    # Check if the slot values are numeric, and remove infinite values if so
    if (is.numeric(slot_values)) {
      slot_values <- slot_values[!sapply(slot_values, is.infinite)]
    }
  }

  return(slot_values)
}



all_empty_slot <- function(object_list, slot_name, slot_element = NULL) {
  # Determine if all slots of objects in the object list are empty

  # Extract slot values
  slot_values <- extract_from_slot(
    object_list = object_list,
    slot_name = slot_name,
    slot_element = slot_element,
    na.rm = TRUE
  )

  # Determine if there are any slot values that have a value.
  return(length(slot_values) == 0L)
}



.optional_from_slot <- function(object, slot_name, alternative = NULL) {
  
  if (methods::.hasSlot(object, slot_name)) {
    return(methods::slot(object, slot_name))
    
  } else {
    return(alternative)
  }
}



is_singular_data <- function(x) {
  # Checks if the input data is singular (i.e. only has one value)
  class_x <- class(x)

  # Drop any NA data.
  x <- x[!is.na(x)]
  if (length(x) <= 1L) return(TRUE)

  if (any(class_x %in% "factor")) {
    return(nlevels(droplevels(x)) == 1L)
    
  } else if (any(class_x %in% c("numeric", "integer", "logical"))) {
    # Drop any infinite data
    x <- x[is.finite(x)]
    if (length(x) <= 1L) return(TRUE)

    return(stats::var(x, na.rm = TRUE) == 0.0)
    
  } else if (any(class_x %in% c("character"))) {
    return(length(unique(x)) == 1L)
    
  } else {
    return(TRUE)
  }
}



is_valid_data <- function(x) {
  # Checks validity of input data x

  class_x <- class(x)

  if (any(class_x %in% c("numeric", "integer", "logical"))) {
    return(is.finite(x))
    
  } else if (any(class_x %in% c("factor"))) {
    return(!is.na(x))
    
  } else if (any(class_x %in% "character")) {
    return(!(is.na(x) | tolower(x) %in% c("na", "nan", "inf", "-inf")))
    
  } else {
    return(FALSE)
  }
}



.get_class_methods <- function(object) {
  # In R, methods are not really inherently associated with a class, but they
  # are functions that are dispatched to a specific class using magic. This
  # function identifies which methods are associated with an object.

  associated_methods <- NULL
  for (object_class in class(object)) {
    # The utils::methods function can return all known (i.e. package in
    # namespace) methods for an object.
    associated_methods <- c(
      associated_methods,
      attr(utils::methods(class = object_class), "info")$generic
    )
  }

  # Keep only unique methods.
  associated_methods <- unique(associated_methods)

  return(associated_methods)
}



.compute_z_statistic <- function(model, fix_all_missing = FALSE) {
  mu <- cov_matrix <- stdevs <- NULL

  if (is(model, "familiarModel")) {
    # Attempt to extract the estimates and the covariance matrix from the
    # familiarModel object.
    if (!is.null(model@trimmed_function$coef)) {
      mu <- model@trimmed_function$coef
    }

    if (!is.null(model@trimmed_function$vcov)) {
      cov_matrix <- model@trimmed_function$vcov
    }

    # Extract the model itself.
    model <- model@model
  }

  if (is(model, "vglm")) {
    # Check if the package is installed and attached.
    ..deprecation_vgam()
    return(NULL)
    
  } else if (inherits(model, "fastglm")) {
    require_package(
      x = "fastglm",
      purpose = "to determine z-scores for generalised linear regression models"
    )

    if (is.null(mu)) mu <- stats::coef(model)

    stdevs <- model$se
    names(stdevs) <- names(mu)
    
  } else if (inherits(model, "nnet")) {
    if (is.null(mu)) mu <- stats::coef(model)
    if (is.null(cov_matrix)) cov_matrix <- stats::vcov(model)
    
  } else {
    if (is.null(mu)) mu <- stats::coef(model)
    if (is.null(cov_matrix)) cov_matrix <- stats::vcov(model)
  }

  if (is.null(stdevs)) {
    if (is.matrix(mu)) {
      stdevs <- matrix(sqrt(diag(cov_matrix)), ncol = ncol(mu), byrow = TRUE)
    } else {
      stdevs <- sqrt(diag(cov_matrix))

      # Order standard deviations according to the estimates.
      if (is.null(names(stdevs))) {
        stdevs <- stdevs[seq_along(mu)]
      } else {
        stdevs <- stdevs[names(stdevs) %in% names(mu)][names(mu)]
      }
    }
  }

  # Compute z-score
  z <- mu / stdevs

  if (fix_all_missing && !any(is.finite(z))) z <- mu

  # Return z-score, as p-values can become very small.
  return(abs(z))
}



is_any <- function(object, class2) {
  # Tests whether the object is any of the classes in class2.
  return(any(sapply(
    class2,
    function(class, object) is(object, class2 = class),
    object = object
  )))
}



fivenum_summary <- function(x, na.rm = FALSE) {
  # Compute fivenumber summary
  y <- stats::fivenum(x = x, na.rm = na.rm)
  
  # Return as data.table
  return(data.table::data.table(
    "min" = y[1L],
    "Q1" = y[2L],
    "median" = y[3L],
    "Q3" = y[4L],
    "max" = y[5L]
  ))
}



get_percentiles <- function(n) {
  return((seq_len(n) - 1.0 / 3.0) / (n + 1.0 / 3.0))
}



trim <- function(x, fraction = 0.1) {
  if (fraction < 0.0 || fraction > 0.5) {
    ..error("Trimming fraction should be between 0.0 and 0.5.")
  }

  # Determine thresholds based on fraction.
  threshold <- stats::quantile(
    x = x, 
    probs = c(fraction, 1.0 - fraction), 
    na.rm = TRUE
  )

  # Set mask
  x_mask <- x >= threshold[1L] & x <= threshold[2L]

  # Return trimmed array
  return(x[x_mask])
}



winsor <- function(x, fraction = 0.1) {
  if (fraction < 0.0 || fraction > 0.5) {
    ..error("Winsoring fraction should be between 0.0 and 0.5.")
  }

  # Determine thresholds based on fraction.
  threshold <- stats::quantile(
    x = x, 
    probs = c(fraction, 1.0 - fraction),
    na.rm = TRUE
  )

  # Values outside valid range are assigned edge values
  x[x < threshold[1L]] <- threshold[1L]
  x[x > threshold[2L]] <- threshold[2L]

  return(x)
}



unique_na <- function(...) {
  values <- do.call(unique, args = list(...))

  return(values[!is.na(values)])
}



#' Create a waiver object
#'
#' This function is functionally identical to `ggplot2::waiver()` function and
#' creates a waiver object. A waiver object is an otherwise empty object that
#' serves the same purpose as `NULL`, i.e. as placeholder for a default value.
#' Because `NULL` can sometimes be a valid input argument, it can therefore not
#' be used to switch to an internal default value.
#'
#' @return waiver object
#' @export
#' @md
waiver <- function() {
  structure(list(), class = "waiver")
}



#' Internal test to see if an object is a waiver
#'
#' This function tests if the object was created by the `waiver` function. This
#' function is functionally identical to `ggplot2:::is.waive()`.
#'
#' @param x Object to be tested.
#'
#' @return `TRUE` for objects that are waivers, `FALSE` otherwise.
#' @md
#' @keywords internal
is.waive <- function(x) {
  return(inherits(x, "waiver"))
}



#' Encapsulate path
#'
#' This function is used to encapsulate paths to allow for behaviour switches.
#' One use is for example when plotting. The plot_all method will encapsulate a
#' path so that plots may be saved to a directory structure. Other plot methods,
#' e.g. plot_model_performance do not encapsulate a path, and if the user calls
#' these functions directly, the plot may be written to the provided path
#' instead of a directory structure.
#'
#' @return encapsulated_path object
#' @md
#' @keywords internal
encapsulate_path <- function(path) {
  structure(path, class = "encapsulated_path")
}



#' Internal test for encapsulated_path
#'
#' This function tests if the object is an `encapsulated_path` object.
#'
#' @param x Object to be tested.
#'
#' @return `TRUE` for objects that are `encapsulated_path`, `FALSE` otherwise.
#' @md
#' @keywords internal
is.encapsulated_path <- function(x) {
  return(inherits(x, "encapsulated_path"))
}



is_subclass <- function(class_1, class_2) {
  # Find out if class_1 is a subclass of class_2
  extending_classes <- methods::extends(class_1)

  # If class 2 is not in the classes from which class 1 inherits, class 1 cannot
  # be a subclass of class 2.
  if (!class_2 %in% extending_classes) return(FALSE)

  # If the classes are the same, class 1 cannot be a subclass of class 2.
  if (class_1 == class_2) return(FALSE)

  # The classes are ordered by distance. Therefore class 2 cannot be a subclass
  # of class 1 if it is less distant.
  if (extending_classes[1L] == class_2) return(FALSE)

  return(TRUE)
}



.as_preprocessing_level <- function(x) {
  if (is(x, "dataObject")) x <- x@preprocessing_level

  # Available pre-processing levels
  preprocessing_levels <- c(
    "none",
    "signature",
    "transformation",
    "normalisation",
    "batch_normalisation",
    "imputation",
    "clustering"
  )

  if (!all(x %in% preprocessing_levels)) {
    ..error_reached_unreachable_code(paste0(
      ".as_preprocessing_level: one or more of x could not be matched to ",
      "preprocessing levels."
    ))
  }

  return(factor(
    x = x,
    levels = preprocessing_levels,
    ordered = TRUE
  ))
}



.flatten_nested_list <- function(x, flatten = FALSE) {
  # Identify names of elements.
  element_names <- unique(unlist(lapply(x, names)))
  
  flattened_list <- lapply(
    element_names, 
    function(element_name, x, flatten) {
      
      element_content <- lapply(x, function(x) (x[[element_name]]))
      
      # Set names of elements
      if (!is.null(names(x))) {
        if (length(names(x)) == length(element_content)) {
          names(element_content) <- names(x)
        }
      }
      
      if (flatten) {
        element_content <- unlist(element_content, recursive = FALSE)
      }
      
      return(element_content)
    },
    x = x,
    flatten = flatten
  )
  
  names(flattened_list) <- element_names
  
  return(flattened_list)
}



dmapply <- function(FUN, ..., MoreArgs = NULL) {
  # mapply for use within data.table. This parses the result of FUN to a flat
  # list. data.table then adds the contents of the list as columns.

  # Apply function.
  x <- mapply(FUN, ..., MoreArgs = MoreArgs, SIMPLIFY = FALSE)

  # Combine to data.table.
  x <- data.table::rbindlist(x, use.names = TRUE)

  # Return as list.
  return(as.list(x))
}



near <- function(x, y, df = 2.0, tol = .Machine$double.eps) {
  # Determines whether floating points values in x are very close to another
  # value in y.
  .near <- function(x, tol) (x <= tol)

  if (length(x) != length(y) && length(y) != 1L) {
    ..error("x and y should have the same length, or y should be a single value.")
  }

  if (!is.numeric(x)) {
    ..error("x should be a numeric value.")
  }

  if (!is.numeric(y)) {
    ..error("y should be a numeric value.")
  }

  return(sapply(abs(x - y), .near, tol = df * tol))
}



approximately <- function(x, y, tol = sqrt(.Machine$double.eps)) {
  # Determines whether numeric values in x are approximately y. This check is
  # the same as all.equal but without the all additional and checking that
  # all.equal does.
  #
  # approximately wraps near.
  return(near(
    x = x,
    y = y,
    df = 1.0,
    tol = tol
  ))
}



huber_estimate <- function(x, k = 1.28, tol = 1E-4) {
  
  # Huber M-estimates require the power.transform package.
  if (require_package("power.transform", message_type = "silent")) {
    return(power.transform::huber_estimate(
      x = x,
      k = k,
      tol = tol
    ))
    
  } else {
    # Fallback option based on winsorised distributions.
    x <- winsor(x, fraction = 0.10)
    x <- x[is.finite(x)]
    
    if (length(x) <= 1L) {
      return(list("mu" = NA_real_, "sigma" = NA_real_))
    }
    
    return(list("mu" = mean(x), "sigma" = stats::sd(x)))
  }
  
  return(x)
}



matrix_pseudo_inverse <- function(x, tol = sqrt(.Machine$double.eps)) {
  # Compute matrix Moore-Penrose pseudo-inverse of x using singular value
  # decomposition.
  
  # Compute single value decomposition.
  svd_decomp <- svd(x)
  
  # Find all positive values in the vector of singular values d.
  positive <- svd_decomp$d > max(tol * svd_decomp$d[1L], 0.0)
  
  if (!any(positive)) {
    return(matrix(0.0, nrow = ncol(x), ncol = nrow(x)))
    
  } else {
    return(
      svd_decomp$v[, positive, drop = FALSE] %*% 
        ((1.0 / svd_decomp$d[positive]) * t(svd_decomp$u[, positive, drop = FALSE]))
    )
  }
}



fam_cut <- function(x, n) {
  if (n > 1L) {
    return(cut(x, n, labels = FALSE))
  }
  # n = 1
  return(rep.int(1L, length(x)))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_ice (generic) -----------------------------------------------------------

#' @title Plot individual conditional expectation plots.
#'
#' @description This method creates individual conditional expectation plots
#'   based on data in a familiarCollection object.
#'
#' @param dir_path (*optional*) Path to the directory where created individual
#'   conditional expectation plots are saved to. Output is saved in the
#'   `explanation` subdirectory. If `NULL`, figures are written to the folder,
#'   but are returned instead.
#' @param discrete_palette (*optional*) Palette for colouring plot elements
#'   indicated by the `color_by` argument (if any). For 2D individual
#'   conditional expectation plots without novelty, the initial colour
#'   determines the colour of the points indicating sample values. `familiar`
#'   has a default palette. Other palettes are supported by the `paletteer`
#'   package, `grDevices::palette.pals()` (requires R >= 4.0.0),
#'   `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`, `heat.colors`,
#'   `terrain.colors`, `topo.colors` and `cm.colors`, which correspond to the
#'   palettes of the same name in `grDevices`. You may also specify your own
#'   palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param gradient_palette (*optional*) Sequential or divergent palette used to
#'   colour the raster in 2D individual conditional expectation or partial
#'   dependence plots. This argument is not used for 1D plots. `familiar` has a
#'   default palette. Other palettes are supported by the `paletteer` package,
#'   `grDevices::palette.pals()` (requires R >= 4.0.0), `grDevices::hcl.pals()`
#'   (requires R >= 3.6.0) and `rainbow`, `heat.colors`, `terrain.colors`,
#'   `topo.colors` and `cm.colors`, which correspond to the palettes of the same
#'   name in `grDevices`. You may also specify your own palette by providing a
#'   vector of colour names listed by `grDevices::colors()` or through
#'   hexadecimal RGB strings.
#' @param gradient_palette_range (*optional*) Numerical range used to span the
#'   gradient for 2D plots. This should be a range of two values, e.g. `c(0,
#'   1)`. By default, values are determined from the data, dependent on the
#'   `value_scales` parameter. This parameter is ignored for 1D plots.
#' @param n_max_samples_shown (*optional*) Maximum number of samples shown in an
#'   individual conditional expectation plot. Defaults to 50. These samples are
#'   randomly picked from the samples present in the ICE data, but the same
#'   samples are consistently picked. Partial dependence is nonetheless computed
#'   from all available samples.
#' @param novelty_range (*optional*) Numerical range used to span the range of
#'   novelty values. This determines the size of the bubbles in 2D, and
#'   transparency of lines in 1D. This should be a range of two values, e.g.
#'   `c(0, 1)`. By default, values are determined from the data, dependent on
#'   the `value_scales` parameter. This parameter is ignored if
#'   `show_novelty=FALSE`.
#' @param value_scales (*optional*) Sets scaling of predicted values. This
#'   parameter has several options:
#'
#'  * `fixed` (default): The value axis for all features will have the same
#'   range.
#'
#'  * `feature`: The value axis for each feature will have the same range. This
#'   option is unavailable for 2D plots.
#'
#'  * `figure`: The value axis for all facets in a figure will have the same
#'   range.
#'
#'  * `facet`: Each facet has its own range. This option is unavailable for 2D
#'   plots.
#'
#'   For 1D plots, this option is ignored if the `y_range` is provided, whereas
#'   for 2D it is ignored if the `gradient_palette_range` is provided.
#' @param novelty_scales (*optional*) Sets scaling of novelty values, similar to
#'   the `value_scales` parameter, but with more limited options:
#'
#'  * `fixed` (default): The novelty will have the same range for all features.
#'
#'  * `figure`: The novelty will have the same range for all facets in a figure.
#'
#' @param ice_default_alpha (*optional*) Default transparency (value) of sample
#'   lines in an 1D plot. When novelty is shown, this is the transparency
#'   corresponding to the least novel points. The confidence interval alpha
#'   values is scaled by this value.
#' @param show_ice (*optional*) Sets whether individual conditional expectation
#'   plots should be created.
#' @param show_pd (*optional*) Sets whether partial dependence plots should be
#'   created. Note that if an anchor is set for a particular feature, its
#'   partial dependence cannot be shown.
#' @param show_novelty (*optional*) Sets whether novelty is shown in plots.
#' @param anchor_values (*optional*) A single value or a named list or array of
#'   values that are used to centre the individual conditional expectation plot.
#'   A single value is valid if and only if only a single feature is assessed.
#'   Otherwise, values Has no effect if the plot is not shown, i.e.
#'   `show_ice=FALSE`. A partial dependence plot cannot be shown for those
#'   features.
#'
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams export_ice_data -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_ice -object
#'
#' @details This function generates individual conditional expectation plots.
#'   These plots come in two varieties, namely 1D and 2D. 1D plots show the
#'   predicted value as function of a single feature, whereas 2D plots show the
#'   predicted value as a function of two features.
#'
#'   Available splitting variables are: `feature_x`, `feature_y` (2D only),
#'   `vimp_method`, `learner`, `data_set` and `positive_class` (categorical
#'   outcomes) or `evaluation_time` (survival outcomes). By default, for 1D ICE
#'   plots the data are split by `feature_x`, `vimp_method` and `learner`, with
#'   faceting by `data_set`, `positive_class` or `evaluation_time`. If only
#'   partial dependence is shown, `positive_class` and `evaluation_time` are
#'   used to set colours instead. For 2D plots, by default the data are split by
#'   `feature_x`, `vimp_method` and `learner`, with faceting by `data_set`,
#'   `positive_class` or `evaluation_time`. The `color_by` argument cannot be
#'   used with 2D plots, and attempting to do so causes an error. Attempting to
#'   specify `feature_x` or `feature_y` for `color_by` will likewise result in
#'   an error, as multiple features cannot be shown in the same facet.
#'
#'   Bootstrap confidence intervals of the partial dependence plots can be shown
#'   using various styles set by `conf_int_style`:
#'
#'  * `ribbon` (default): confidence intervals are shown as a ribbon with an
#'   opacity of `conf_int_alpha` around the point estimate of the partial
#'   dependence.
#'
#'  * `step` (default): confidence intervals are shown as a step function around
#'   the point estimate of the partial dependence.
#'
#'  * `none`: confidence intervals are not shown. The point estimate of the
#'   partial dependence is shown as usual.
#'
#'   Note that when bootstrap confidence intervals were computed, they were also
#'   computed for individual samples in individual conditional expectation
#'   plots. To avoid clutter, only point estimates for individual samples are
#'   shown.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#' @exportMethod plot_ice
#' @md
#' @rdname plot_ice-methods
setGeneric(
  "plot_ice",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    gradient_palette_range = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    novelty_range = NULL,
    value_scales = waiver(),
    novelty_scales = waiver(),
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    ice_default_alpha = 0.6,
    n_max_samples_shown = 50L,
    show_ice = TRUE,
    show_pd = TRUE,
    show_novelty = TRUE,
    anchor_values = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_ice")
  }
)



# plot_ice (general) -----------------------------------------------------------

#' @rdname plot_ice-methods
setMethod(
  "plot_ice",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    gradient_palette_range = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    novelty_range = NULL,
    value_scales = waiver(),
    novelty_scales = waiver(),
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    ice_default_alpha = 0.6,
    n_max_samples_shown = 50L,
    show_ice = TRUE,
    show_pd = TRUE,
    show_novelty = TRUE,
    anchor_values = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "ice_data"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_ice,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "gradient_palette" = gradient_palette,
        "gradient_palette_range" = gradient_palette_range,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "n_max_samples_shown" = n_max_samples_shown,
        "novelty_range" = novelty_range,
        "value_scales" = value_scales,
        "novelty_scales" = novelty_scales,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "ice_default_alpha" = ice_default_alpha,
        "show_ice" = show_ice,
        "show_pd" = show_pd,
        "show_novelty" = show_novelty,
        "anchor_values" = anchor_values,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_pd (generic) ------------------------------------------------------------

#' @title Plot partial dependence.
#'
#' @description This method creates partial dependence plots based on data in a
#'   familiarCollection object.
#'
#' @inheritParams plot_ice
#' @inheritDotParams export_ice_data -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_ice -object
#'
#' @details This function generates partial dependence plots. These plots come
#'   in two varieties, namely 1D and 2D. 1D plots show the predicted value as
#'   function of a single feature, whereas 2D plots show the predicted value as
#'   a function of two features.
#'
#'   Available splitting variables are: `feature_x`, `feature_y` (2D only),
#'   `vimp_method`, `learner`, `data_set` and `positive_class` (categorical
#'   outcomes) or `evaluation_time` (survival outcomes). By default, for 1D ICE
#'   plots the data are split by `feature_x`, `vimp_method` and `learner`, with
#'   faceting by `data_set`, `positive_class` or `evaluation_time`. If only
#'   partial dependence is shown, `positive_class` and `evaluation_time` are
#'   used to set colours instead. For 2D plots, by default the data are split by
#'   `feature_x`, `vimp_method` and `learner`, with faceting by `data_set`,
#'   `positive_class` or `evaluation_time`. The `color_by` argument cannot be
#'   used with 2D plots, and attempting to do so causes an error. Attempting to
#'   specify `feature_x` or `feature_y` for `color_by` will likewise result in
#'   an error, as multiple features cannot be shown in the same facet.
#'
#'   The splitting variables indicated by `color_by` are coloured according to
#'   the `discrete_palette` parameter. This parameter is therefore only used for
#'   1D plots. Available palettes for `discrete_palette` and `gradient_palette`
#'   are those listed by `grDevices::palette.pals()` (requires R >= 4.0.0),
#'   `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`, `heat.colors`,
#'   `terrain.colors`, `topo.colors` and `cm.colors`, which correspond to the
#'   palettes of the same name in `grDevices`. If not specified, a default
#'   palette based on palettes in Tableau are used. You may also specify your
#'   own palette by using colour names listed by `grDevices::colors()` or
#'   through hexadecimal RGB strings.
#'
#'   Bootstrap confidence intervals of the partial dependence plots can be shown
#'   using various styles set by `conf_int_style`:
#'
#'  * `ribbon` (default): confidence intervals are shown as a ribbon with an
#'   opacity of `conf_int_alpha` around the point estimate of the partial
#'   dependence.
#'
#'  * `step` (default): confidence intervals are shown as a step function around
#'   the point estimate of the partial dependence.
#'
#'  * `none`: confidence intervals are not shown. The point estimate of the
#'   partial dependence is shown as usual.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#' @exportMethod plot_ice
#' @md
#' @rdname plot_pd-methods
setGeneric(
  "plot_pd",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    gradient_palette_range = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    novelty_range = NULL,
    value_scales = waiver(),
    novelty_scales = waiver(),
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    show_novelty = TRUE,
    anchor_values = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_pd")
  }
)



# plot_pd (general) ------------------------------------------------------------

#' @rdname plot_pd-methods
setMethod(
  "plot_pd",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    gradient_palette_range = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    novelty_range = NULL,
    value_scales = waiver(),
    novelty_scales = waiver(),
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    show_novelty = TRUE,
    anchor_values = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object, 
          "data_element" = "ice_data"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_ice,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "gradient_palette" = gradient_palette,
        "gradient_palette_range" = gradient_palette_range,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "novelty_range" = novelty_range,
        "value_scales" = value_scales,
        "novelty_scales" = novelty_scales,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "show_ice" = FALSE,
        "show_pd" = TRUE,
        "show_novelty" = show_novelty,
        "anchor_values" = anchor_values,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_ice (collection) --------------------------------------------------------

#' @rdname plot_ice-methods
setMethod(
  "plot_ice",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    gradient_palette_range = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    novelty_range = NULL,
    value_scales = waiver(),
    novelty_scales = waiver(),
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    ice_default_alpha = 0.6,
    n_max_samples_shown = 50L,
    show_ice = TRUE,
    show_pd = TRUE,
    show_novelty = TRUE,
    anchor_values = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Get input data.
    ice_data <- export_ice_data(
      object = object,
      aggregate_results = TRUE
    )

    pd_data <- export_partial_dependence_data(
      object = object,
      aggregate_results = TRUE
    )

    # Check anchor values.
    if (length(ice_data) > 0L && !is.null(anchor_values)) {
      # Determine feature names.
      feature_names <- unique(c(
        sapply(ice_data, function(x) (x@identifiers$feature_x)),
        sapply(ice_data, function(x) (x@identifiers$feature_y))
      ))

      # Check if names are provided to anchor values.
      if (length(feature_names) > 1L && length(names(anchor_values)) == 0L) {
        ..error(paste0(
          "Data for plotting individual conditional expectation plots for more ",
          "than one feature are present. However, the anchor values for ",
          "centering the plots could not be assigned because they were not named. ",
          "Please provide a named vector or list."
        ))
      }

      # Check if the names provided to anchor values match existing values.
      if (length(setdiff(names(anchor_values), feature_names)) > 0L) {
        ..warning(paste0(
          "One or more feature names specified as anchor values do not exist ",
          "as features for which data for plotting individual conditional ",
          "expectation plots were computed: ",
          paste_s(setdiff(names(anchor_values), feature_names))
        ))
      }
    }

    # Check that the data are not empty.
    if (is_empty(ice_data)) return(NULL)

    # Check that the data are not evaluated at the model level.
    if (all(sapply(ice_data, function(x) (x@detail_level == "model")))) {
      ..warning_no_comparison_between_models()
      return(NULL)
    }

    # Obtain data element from list.
    if (is.list(ice_data)) {
      if (is_empty(ice_data)) return(NULL)
      if (all(sapply(ice_data, is_empty))) return(NULL)
    }

    # Check that the data are not empty.
    if (is_empty(ice_data)) return(NULL)

    # Check "n_max_samples_shown".
    if (!is.null(n_max_samples_shown)) {
      n_max_samples_shown <- as.integer(n_max_samples_shown)
      .check_number_in_valid_range(
        x = n_max_samples_shown, 
        var_name = "n_max_samples_shown",
        range = c(0L, Inf)
      )
    }

    # If the maximum number of ICE samples shown equal 0, set to NULL.
    if (n_max_samples_shown == 0L) {
      show_ice <- FALSE
      n_max_samples_shown <- NULL
    }
    
    # Update the output so that it is more consistent.
    data <- mapply(
      .update_ice_and_pd_output,
      ice_data = ice_data,
      pd_data = pd_data,
      MoreArgs = list(
        "outcome_type" = object@outcome_type,
        "class_levels" = get_outcome_class_levels(object),
        "anchor_values" = anchor_values,
        "n_samples" = n_max_samples_shown,
        "seed" = sample.int(n = 10000L, size = 1L)
      ),
      SIMPLIFY = FALSE
    )
    
    # Flatten nested list.
    data <- .flatten_nested_list(data)
    ice_data <- data$ice_data
    pd_data <- data$pd_data
    
    # Add in sample to ice_data, and drop identifier columns.
    ice_data <- lapply(
      ice_data,
      function(x) {
        x@data[, "sample" := get_unique_row_names(x = x@data)]
        for (id_column in get_id_columns()) {
          x@data[, (id_column) := NULL]
        }
        
        x@grouping_column <- setdiff(x@grouping_column, get_id_columns())
        x@grouping_column <- c(x@grouping_column, "sample")
        
        return(x)
      }
    )
    
    # Check that show_pd and show_ice are not both FALSE.
    if (!show_pd && !show_ice) {
      ..warning(paste0(
        "One or both of \"show_pd\" and \"show_ice\" should be TRUE to plot ",
        "individual conditional expectation and/or partial dependence plots."
      ))
      return(NULL)
    }

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = TRUE),
      purpose = "to create ICE/PD plots",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Check input arguments ----------------------------------------------------

    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # Determine whether 1D or 2D plots are shown.
    show_2d <- any(sapply(
      ice_data, 
      function(x) (!is.null(x@identifiers$feature_y))
    ))

    # If the y-axis will contain a feature, only the partial dependence
    # can be shown.
    if (show_2d) {
      show_pd <- TRUE
      show_ice <- FALSE
    }

    # conf_int_style
    if (length(conf_int_style) > 1L) {
      conf_int_style <- head(conf_int_style, n = 1L)
    }

    # Set the style of the confidence interval to none, in case no confidence
    # interval data is present.
    if (show_pd && !show_2d) {
      if (!all(sapply(
        pd_data,
        function(x) (x@estimation_type %in% c("bci", "bootstrap_confidence_interval"))
      ))) {
        conf_int_style <- "none"
      }
      
    } else {
      conf_int_style <- "none"
    }

    # Determine identifiers that should be dropped from plot_data.
    if (show_2d) {
      dropped_identifiers <- c("feature_x_value", "feature_y_value")
    } else if (show_ice) {
      dropped_identifiers <- c("sample", "feature_x_value")
    } else {
      dropped_identifiers <- c("feature_x_value")
    }
    
    # Determine splitting variables present in the dataset.
    if (show_ice) {
      plot_data <- identify_element_sets(
        ice_data,
        ignore_grouping_column = FALSE,
        ignore_list_identifier = FALSE,
        drop_identiers = dropped_identifiers
      )
      
    } else {
      plot_data <- identify_element_sets(
        pd_data,
        ignore_grouping_column = FALSE,
        ignore_list_identifier = FALSE,
        drop_identiers = dropped_identifiers
      )
    }

    # Set default splitting variables.
    if (is.null(split_by) && is.null(facet_by) && is.null(color_by)) {
      if (show_2d) {
        split_by <- c("vimp_method", "learner", "feature_x", "feature_y")

        facet_by <- c("data_set")
        if (object@outcome_type == "multinomial") {
          facet_by <- c(facet_by, "positive_class")
        }
        if (object@outcome_type == "survival") {
          facet_by <- c(facet_by, "evaluation_time")
        }

        color_by <- NULL
        
      } else if (show_ice) {
        split_by <- c("vimp_method", "learner", "feature_x")

        facet_by <- c("data_set")
        if (object@outcome_type == "multinomial") {
          facet_by <- c(facet_by, "positive_class")
        }
        if (object@outcome_type == "survival") {
          facet_by <- c(facet_by, "evaluation_time")
        }

        color_by <- NULL
        
      } else if (show_pd) {
        split_by <- c("vimp_method", "learner", "feature_x")
        facet_by <- c("data_set")

        color_by <- NULL
        if (object@outcome_type == "multinomial") {
          color_by <- c("positive_class")
        }
        if (object@outcome_type == "survival") {
          color_by <- c("evaluation_time")
        }
      }
    }

    # Determine splitting variables.
    available_splitting_vars <- c("vimp_method", "learner", "data_set", "feature_x")
    if (show_2d) {
      available_splitting_vars <- c(available_splitting_vars, "feature_y")
    }
    if (object@outcome_type %in% c("survival")) {
      available_splitting_vars <- c(available_splitting_vars, "evaluation_time")
    }
    if (object@outcome_type %in% c("multinomial")) {
      available_splitting_vars <- c(available_splitting_vars, "positive_class")
    }

    # Check that the color_by parameter is not set for 2d plots.
    if (show_2d && !is.null(color_by)) {
      ..error(paste0(
        "The \"color_by\" parameter cannot be used when creating 2D ",
        "individual conditional expection and partial dependence plots, ",
        "as additional colour-coding would making interpretation very difficult."
      ))
    }

    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = plot_data,
      split_by = split_by,
      color_by = color_by,
      facet_by = facet_by,
      available = available_splitting_vars
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    color_by <- split_var_list$color_by
    facet_by <- split_var_list$facet_by

    if ("feature_x" %in% color_by || "feature_y" %in% color_by) {
      ..error("Features cannot be used to specify colors.")
    }

    # Create a legend label
    if (show_2d & is.waive(legend_label)) {
      legend_label <- switch(
        object@outcome_type,
        "binomial" = "probability",
        "multinomial" = "probability",
        "count" = "value",
        "continuous" = "value",
        "survival" = "probability",
        "competing_risk" = "probability"
      )
      
    } else if (!show_2d) {
      legend_label <- .create_plot_legend_title(
        user_label = legend_label,
        color_by = color_by
      )
    }
    
    # Check input arguments for validity.
    .check_input_plot_args(
      conf_int_alpha = conf_int_alpha,
      conf_int_style = conf_int_style,
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption
    )

    # Set unused data to NULL.
    if (!show_ice) ice_data <- NULL
    if (!show_pd) pd_data <- NULL

    # Set value scales
    if (is.waive(value_scales)) value_scales <- "fixed"

    # Check if the value scales are correctly defined.
    if (show_2d) {
      .check_parameter_value_is_valid(
        x = value_scales, 
        var_name = "value_scales",
        values = c("fixed", "figure")
      )
      
    } else {
      .check_parameter_value_is_valid(
        x = value_scales,
        var_name = "value_scales",
        values = c("fixed", "figure", "feature", "facet")
      )
    }

    # Set value ranges.
    if (show_2d && !is.null(gradient_palette_range)) {
      # The value range is provided by the user.
      value_range <- unique(plot_data[, c("feature_x", "feature_y")])
      value_range[, ":="(
        "min_value" = gradient_palette_range[1L],
        "max_value" = gradient_palette_range[2L]
      )]
      
    } else if (!show_2d && !is.null(y_range)) {
      # The value range is provided by the user.
      value_range <- unique(plot_data[, c("feature_x")])
      value_range[, ":="(
        "min_value" = y_range[1L],
        "max_value" = y_range[2L]
      )]
      
    } else if (show_2d & value_scales == "fixed") {
      # The value range is the same for all features.
      value_range <- .create_ice_plot_value_range(
        x = pd_data,
        scale_method = "fixed",
        outcome_type = object@outcome_type,
        confidence_interval = conf_int_style != "none"
      )
      
    } else if (!show_2d & value_scales == "fixed") {
      # The value range is the same for all features.
      if (show_ice) {
        value_range <- .create_ice_plot_value_range(
          x = ice_data,
          scale_method = "fixed",
          outcome_type = object@outcome_type,
          confidence_interval = conf_int_style != "none"
        )
        
      } else {
        value_range <- .create_ice_plot_value_range(
          x = pd_data,
          scale_method = "fixed",
          outcome_type = object@outcome_type,
          confidence_interval = conf_int_style != "none"
        )
      }
      
    } else if (!show_2d & value_scales == "feature") {
      # Every feature has its value range.
      if (show_ice) {
        value_range <- .create_ice_plot_value_range(
          x = ice_data,
          scale_method = "feature",
          outcome_type = object@outcome_type,
          confidence_interval = conf_int_style != "none"
        )
        
      } else {
        value_range <- .create_ice_plot_value_range(
          x = pd_data,
          scale_method = "feature",
          outcome_type = object@outcome_type,
          confidence_interval = conf_int_style != "none"
        )
      }
      
    } else {
      # We will create value-ranges downstream.
      value_range <- NULL
    }

    # Set novelty scales
    if (is.waive(novelty_scales)) {
      novelty_scales <- "fixed"
    }

    # Check if the novelty scales are correctly defined.
    .check_parameter_value_is_valid(
      x = novelty_scales, 
      var_name = "novelty_scales",
      values = c("fixed", "figure")
    )

    if (!is.null(novelty_range)) {
      # The value range is provided by the user.
      novelty_range <- unique(plot_data[, c("feature_x", "feature_y")])[, ":="(
        "min_value" = novelty_range[1L],
        "max_value" = novelty_range[2L]
      )]
      
    } else if (novelty_scales == "fixed") {
      # The novelty range is the same for all features.
      if (show_ice) {
        novelty_range <- .create_ice_plot_novelty_range(
          x = ice_data,
          scale_method = "fixed",
          outcome_type = object@outcome_type
        )
        
      } else {
        novelty_range <- .create_ice_plot_novelty_range(
          x = pd_data,
          scale_method = "fixed",
          outcome_type = object@outcome_type
        )
      }
      
    } else {
      novelty_range <- NULL
    }

    # Check the transparency value,
    .check_number_in_valid_range(
      x = ice_default_alpha,
      var_name = "ice_default_alpha",
      range = c(0.0, 1.0)
    )

    # Set plot function.
    if (show_2d) {
      plot_function <- .create_2d_ice_plot
    } else {
      plot_function <- .create_1d_ice_plot
    }

    # Create plots -------------------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data
    if (!is.null(split_by)) {
      x_split <- split(plot_data, by = split_by)
    } else {
      x_split <- list("null.name" = plot_data)
    }

    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over splits
    for (ii in names(x_split)) {
      # Skip empty datasets
      if (is_empty(x_split[[ii]])) next

      if (is.waive(plot_title)) {
        plot_title <- ifelse(
          show_ice,
          "Individual conditional expectation",
          "Partial dependence"
        )
      }

      # Declare subtitle components.
      additional_subtitle <- NULL

      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, color_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(x_split[[ii]]$evaluation_time[1L])
        )
      }

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = x_split[[ii]]
        )
      }
      
      # Generate plot
      p <- .plot_ice(
        x = x_split[[ii]],
        ice_data = ice_data,
        pd_data = pd_data,
        plot_function = plot_function,
        color_by = color_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        gradient_palette = gradient_palette,
        gradient_palette_range = gradient_palette_range,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        x_range = x_range,
        x_breaks = x_breaks,
        x_n_breaks = x_n_breaks,
        y_range = y_range,
        y_breaks = y_breaks,
        y_n_breaks = y_n_breaks,
        value_scales = value_scales,
        value_range = value_range,
        novelty_scales = novelty_scales,
        novelty_range = novelty_range,
        conf_int_style = conf_int_style,
        conf_int_alpha = conf_int_alpha,
        ice_default_alpha = ice_default_alpha,
        show_novelty = show_novelty,
        show_ice = show_ice,
        show_pd = show_pd,
        show_2d = show_2d,
        outcome_type = object@outcome_type
      )

      # Check empty output
      if (is.null(p)) next

      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Set initial subtype.
        subtype <- ifelse(show_ice, "ice", "pd")
        if (show_2d) subtype <- c(subtype, "2d")

        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_ice_plot_dimensions(
          x = x_split[[ii]],
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "explanation",
              "subtype" = subtype,
              "x" = x_split[[ii]],
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_ice <- function(
    x,
    ice_data,
    pd_data,
    plot_function,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    gradient_palette,
    gradient_palette_range,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    x_n_breaks,
    x_label_shared = waiver(),
    y_range,
    y_breaks,
    y_n_breaks,
    y_label_shared = waiver(),
    value_scales,
    value_range,
    novelty_scales,
    novelty_range,
    conf_int_style,
    conf_int_alpha,
    ice_default_alpha,
    show_novelty,
    show_ice,
    show_pd,
    show_2d,
    outcome_type
) {
  # Split by facet. This generates a list of data splits with faceting
  # information that allows for positioning.
  plot_layout_table <- .get_plot_layout_table(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Split data into facets. This is done by row.
  data_facet_list <- .split_data_by_plot_facet(
    x = x,
    plot_layout_table = plot_layout_table
  )
  
  # Used for ordering of composite figures.
  layout_split <- split(
    plot_layout_table,
    by = c("col_id", "row_id"),
    sorted = TRUE
  )

  # Set value scales.
  if (is.null(value_range) && value_scales == "figure") {
    # Keep only data that are used to set the scale.
    plot_ice_data <- .select_ice_plot_data(x = x, data = ice_data)
    plot_pd_data <- .select_ice_plot_data(x = x, data = pd_data)

    if (!is.null(plot_ice_data)) {
      value_range <- .create_ice_plot_value_range(
        x = plot_ice_data,
        scale_method = "figure",
        outcome_type = outcome_type,
        confidence_interval = conf_int_style != "none"
      )
      
    } else {
      value_range <- .create_ice_plot_value_range(
        x = plot_pd_data,
        scale_method = "figure",
        outcome_type = outcome_type,
        confidence_interval = conf_int_style != "none"
      )
    }
  }
  
  # Set novelty scales
  if (is.null(novelty_range) && novelty_scales == "figure") {
    # Keep only data that are used to set the scale.
    plot_ice_data <- .select_ice_plot_data(x = x, data = ice_data)
    plot_pd_data <- .select_ice_plot_data(x = x, data = pd_data)

    if (!is.null(plot_ice_data)) {
      novelty_range <- .create_ice_plot_novelty_range(
        x = plot_ice_data,
        scale_method = "figure",
        outcome_type = outcome_type
      )
      
    } else {
      novelty_range <- .create_ice_plot_novelty_range(
        x = plot_pd_data,
        scale_method = "figure",
        outcome_type = outcome_type
      )
    }
  }

  # Set x_label_shared
  if (is.waive(x_label_shared)) {
    # Default value.
    x_label_shared <- "column"

    if (length(facet_by) > 1L) {
      if ("feature_x" %in% c(tail(facet_by, n = length(facet_by) - 1L))) {
        # Same feature is used row-wise, not column-wise.
        x_label_shared <- "individual"
      }
    }
  }

  # Set y_label_shared
  if (is.waive(y_label_shared)) {
    # Default value.
    y_label_shared <- "row"

    if (value_scales == "facet") {
      # If y-axis is scaled per individual facet
      y_label_shared <- "individual"
      
    } else if (show_2d && length(facet_by) > 0L) {
      if ("feature_y" == facet_by[1L]) {
        # Same feature is used column-wise, not row-wise.
        y_label_shared <- "individual"
      }
    }
  }

  # Check label sharing.
  .check_input_plot_args(
    x_label_shared = x_label_shared,
    y_label_shared = y_label_shared
  )

  # Placeholders for plots and plot elements.
  figure_list <- list()
  extracted_element_list <- list()

  # Iterate over facets
  for (ii in names(layout_split)) {
    # Set facet_data
    facet_data <- data_facet_list[[ii]]
    if (is_empty(data_facet_list[[ii]])) next

    # Keep only data that are used to set the scale.
    facet_ice_data <- .select_ice_plot_data(x = facet_data, data = ice_data)
    facet_pd_data <- .select_ice_plot_data(x = facet_data, data = pd_data)

    p_ice <- do.call(
      plot_function,
      args = list(
        "facet_data" = facet_data,
        "ice_data" = facet_ice_data,
        "pd_data" = facet_pd_data,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "gradient_palette" = gradient_palette,
        "gradient_palette_range" = gradient_palette_range,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_range" = x_range,
        "x_breaks" = x_breaks,
        "x_n_breaks" = x_n_breaks,
        "y_range" = y_range,
        "y_breaks" = y_breaks,
        "y_n_breaks" = y_n_breaks,
        "value_scales" = value_scales,
        "value_range" = value_range,
        "novelty_range" = novelty_range,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "ice_default_alpha" = ice_default_alpha,
        "show_novelty" = show_novelty,
        "show_ice" = show_ice,
        "show_pd" = show_pd,
        "outcome_type" = outcome_type
      )
    )
    
    # Rename plot elements.
    g_ice <- .rename_plot_grobs(
      g = .convert_to_grob(p_ice),
      extension = "main"
    )
    if (!gtable::is.gtable(g_ice)) next
    
    # Attach to figure list.
    figure_list[[paste0(layout_split[[ii]]$row_id, ".", layout_split[[ii]]$col_id)]] <- as_familiar_plot(
      g = g_ice,
      layout = layout_split[[ii]]
    )
  }
  
  # Compose the final figure. Magic.
  g <- .compose_figure(
    figure_list = figure_list,
    plot_layout_table = plot_layout_table,
    x_text_shared = x_label_shared,
    x_label_shared = x_label_shared,
    y_text_shared = y_label_shared,
    y_label_shared = y_label_shared,
    facet_wrap_cols = facet_wrap_cols,
    ggtheme = ggtheme
  )

  return(g)
}



.create_1d_ice_plot <- function(
    facet_data,
    ice_data,
    pd_data,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    x_n_breaks,
    y_breaks,
    y_n_breaks,
    value_scales,
    value_range,
    novelty_range,
    conf_int_style,
    conf_int_alpha,
    ice_default_alpha,
    show_novelty,
    show_ice,
    show_pd,
    outcome_type,
    ...
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  feature_x <- color_breaks <- sample <- NULL
  
  # Flag indicating presence of data.
  data_present <- TRUE
  
  # Get the data that determines the main plot characteristics.
  if (show_ice) {
    if (!is_empty(ice_data[[1L]])) {
      plot_data <- data.table::copy(ice_data[[1L]]@data)
    } else {
      plot_data <- facet_data
      data_present <- FALSE
    }
    
  } else {
    if (!is_empty(pd_data[[1L]])) {
      plot_data <- data.table::copy(pd_data[[1L]]@data)
    } else {
      plot_data <- facet_data
      data_present <- FALSE
    }
  }
  
  # Set value range
  if (is.null(value_range) && value_scales == "facet" && data_present) {
    if (show_ice) {
      value_range <- .create_ice_plot_value_range(
        x = ice_data,
        scale_method = "facet",
        outcome_type = outcome_type,
        confidence_interval = conf_int_style != "none"
      )
      
    } else {
      value_range <- .create_ice_plot_value_range(
        x = pd_data,
        scale_method = "facet",
        outcome_type = outcome_type,
        confidence_interval = conf_int_style != "none"
      )
    }
  }

  # x_range
  if (is.numeric(plot_data$feature_x_value)) {
    if (is.null(x_range)) {
      x_range <- c(
        min(plot_data$feature_x_value, na.rm = TRUE),
        max(plot_data$feature_x_value, na.rm = TRUE)
      )
    }

    # x_breaks
    if (is.null(x_breaks)) {
      .check_input_plot_args(x_n_breaks = x_n_breaks)

      # Create breaks and update x_range
      x_breaks <- labeling::extended(
        m = x_n_breaks,
        dmin = x_range[1L],
        dmax = x_range[2L],
        only.loose = TRUE
      )

      # Update x_range
      x_range <- c(
        head(x_breaks, n = 1L),
        tail(x_breaks, n = 1L)
      )
    }

    .check_input_plot_args(x_range = x_range)
    
  } else {
    x_range <- NULL
  }

  # Find the correct y-range
  value_range <- value_range[feature_x == as.character(plot_data$feature_x[1L])]
  if (is_empty(value_range)) {
    y_range <- c(NA_real_, NA_real_)
    
  } else {
    y_range <- c(value_range$min_value, value_range$max_value)

    # y_breaks
    if (is.null(y_breaks)) {
      .check_input_plot_args(y_n_breaks = y_n_breaks)

      # Create breaks and update y_range
      y_breaks <- labeling::extended(
        m = y_n_breaks,
        dmin = y_range[1L],
        dmax = y_range[2L],
        only.loose = TRUE
      )

      # Update y-range
      y_range <- c(
        head(y_breaks, n = 1L),
        tail(y_breaks, n = 1L)
      )
    }
  }

  .check_input_plot_args(y_range = y_range)

  # Set x-label
  if (is.waive(x_label)) {
    if (!is.na(plot_data$feature_x[1L])) {
      x_label <- as.character(plot_data$feature_x[1L])
    } else {
      x_label <- NULL
    }
  }

  # Set y-label
  if (is.waive(y_label)) {
    y_label <- switch(
      outcome_type,
      "binomial" = "probability",
      "multinomial" = "probability",
      "count" = "value",
      "continuous" = "value",
      "survival" = "survival probability",
      "competing_risk" = "survival probability"
    )
  }

  .check_input_plot_args(
    x_label = x_label,
    y_label = y_label
  )
  
  # Update show_novelty to check for non-finite values in the novelty data.
  show_novelty <- show_novelty && !is.null(plot_data$novelty) && all(is.finite(plot_data$novelty))

  if (show_novelty) {
    # Find the correct novelty-range
    novelty_range <- novelty_range[feature_x == as.character(plot_data$feature_x[1L])]

    if (is_empty(novelty_range)) {
      novelty_range <- c(NA_real_, NA_real_)
    } else {
      novelty_range <- c(novelty_range$min_value, novelty_range$max_value)
    }
  }

  # Generate a guide table for ice data to allow integration of guides into a
  # single legend.
  if (show_ice && data_present) {
    ice_guide_list <- .create_plot_guide_table(
      x = ice_data[[1L]]@data,
      color_by = color_by,
      discrete_palette = discrete_palette,
      combine_legend = FALSE
    )

    # Add column to force proper grouping of lines.
    ice_guide_list$data[, "color_breaks_sample" := paste0(color_breaks, "_", sample)]
    
  } else {
    ice_guide_list <- NULL
  }

  # Generate a guide table for pd data to allow integration of guides into a
  # single legend.
  if (show_pd && data_present) {
    pd_guide_list <- .create_plot_guide_table(
      x = pd_data[[1L]]@data,
      color_by = color_by,
      discrete_palette = discrete_palette,
      combine_legend = FALSE
    )

    # Set grouping variable to deal with point-group complaints.
    if (is.factor(pd_guide_list$data$feature_x_value)) {
      pd_group_variable <- 1L
      
    } else {
      pd_group_variable <- NULL
    }
    
  } else {
    pd_guide_list <- NULL
  }

  # Make pd line thicker than ice lines.
  ice_line_size <- 0.5 * ..get_plot_theme_linewidth(ggtheme = ggtheme)
  pd_line_size <- 6.0 * ice_line_size

  # In case only partial dependency plots are shown, update ice_default_alpha
  if (!show_ice) ice_default_alpha <- 1.0

  # Create basic plot.
  p <- ggplot2::ggplot(
    data = plot_data,
    mapping = ggplot2::aes(
      x = !!sym("feature_x_value"),
      y = !!sym("value")
    )
  )
  p <- p + ggtheme

  if (!all(is.finite(y_range))) {
    # Blank elements in case the value range is unset, e.g. because the model
    # cannot compute survival probabilities. This happens for some mboost
    # learners.
    p <- p + ggplot2::geom_blank()
    
  } else if (show_novelty) {
    # Plot with novelty.
    if (show_ice) {
      if (!is.null(color_by)) {
        # Create lines with alpha.
        p <- p + ggplot2::geom_line(
          data = ice_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value"),
            alpha = !!sym("novelty"),
            colour = !!sym("color_breaks"),
            group = !!sym("color_breaks_sample")
          ),
          linewidth = ice_line_size
        )
        
      } else {
        # Create lines with alpha.
        p <- p + ggplot2::geom_line(
          data = ice_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value"),
            group = !!sym("sample"),
            alpha = !!sym("novelty")
          ),
          linewidth = ice_line_size
        )
      }
    }
    
    if (show_pd) {
      if (!is.null(color_by)) {
        # Create lines with alpha.
        p <- p + ggplot2::geom_line(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value"),
            alpha = !!sym("novelty"),
            colour = !!sym("color_breaks"),
            group = pd_group_variable
          ),
          linewidth = pd_line_size
        )
        
      } else {
        # Create lines with alpha.
        p <- p + ggplot2::geom_line(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value"),
            alpha = !!sym("novelty"),
            group = pd_group_variable
          ),
          linewidth = pd_line_size
        )
      }
    }
    
    # Invert novelty values, since higher values indicates greater novelty.
    p <- p + ggplot2::scale_alpha(
      trans = "reverse",
      limits = novelty_range,
      range = c(0.1, 1.0) * ice_default_alpha
    )
    
  } else {
    # Plot without novelty.
  
    if (show_ice) {
      if (!is.null(color_by)) {
        # Create lines with alpha.
        p <- p + ggplot2::geom_line(
          data = ice_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value"),
            group = !!sym("color_breaks_sample"),
            colour = !!sym("color_breaks")
          ),
          linewidth = ice_line_size,
          alpha = ice_default_alpha
        )
        
      } else {
        # Create lines with alpha.
        p <- p + ggplot2::geom_line(
          data = ice_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value"),
            group = !!sym("sample")
          ),
          linewidth = ice_line_size,
          alpha = ice_default_alpha
        )
      }
    }

    if (show_pd) {
      if (!is.null(color_by)) {
        # Create lines with alpha.
        p <- p + ggplot2::geom_line(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value"),
            colour = !!sym("color_breaks"),
            group = pd_group_variable
          ),
          linewidth = pd_line_size
        )
        
      } else {
        # Create lines with alpha.
        p <- p + ggplot2::geom_line(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value"),
            group = pd_group_variable
          ),
          linewidth = pd_line_size
        )
      }
    }
  }

  # Set colour
  if (!is.null(color_by)) {
    # Extract guide_table for colour.
    if (show_ice) {
      g_color <- ice_guide_list$guide_color
    } else {
      g_color <- pd_guide_list$guide_color
    }

    p <- p + ggplot2::scale_colour_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )
  }

  # Update x and y scales
  if (!is.null(x_range)) p <- p + ggplot2::scale_x_continuous(breaks = x_breaks)
  if (!is.null(y_range)) p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)

  # Plot confidence intervals.
  if (conf_int_style[1L] != "none") {
    if (conf_int_style[1L] == "step") {
      if (is.null(color_by)) {
        p <- p + ggplot2::geom_step(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value_ci_low")
          ),
          linetype = "dashed"
        )

        p <- p + ggplot2::geom_step(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value_ci_up")
          ),
          linetype = "dashed"
        )
        
      } else {
        p <- p + ggplot2::geom_step(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value_ci_low"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed"
        )

        p <- p + ggplot2::geom_step(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            y = !!sym("value_ci_up"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed"
        )
      }


      # Remove linetype from the legend.
      p <- p + ggplot2::scale_linetype(guide = "none")
      
    } else if (conf_int_style[1L] == "ribbon") {
      if (show_novelty) conf_int_alpha <- conf_int_alpha * ice_default_alpha

      if (is.null(color_by)) {
        p <- p + ggplot2::geom_ribbon(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            ymin = !!sym("value_ci_low"),
            ymax = !!sym("value_ci_up")
          ),
          alpha = conf_int_alpha
        )
        
      } else {
        p <- p + ggplot2::geom_ribbon(
          data = pd_guide_list$data,
          mapping = ggplot2::aes(
            x = !!sym("feature_x_value"),
            ymin = !!sym("value_ci_low"),
            ymax = !!sym("value_ci_up"),
            fill = !!sym("color_breaks")
          ),
          alpha = conf_int_alpha
        )
      }
    }
  }

  # Determine how things are facetted.
  facet_by_list <- .parse_plot_facet_by(
    x = plot_data,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context"
      )
    }
  }

  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Plot to Cartesian coordinates.
  p <- p + ggplot2::coord_cartesian(xlim = x_range, ylim = y_range)

  return(p)
}



.create_2d_ice_plot <- function(
    facet_data,
    pd_data,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    gradient_palette,
    gradient_palette_range,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    x_n_breaks,
    y_range,
    y_breaks,
    y_n_breaks,
    value_scales,
    value_range,
    novelty_range,
    conf_int_style,
    conf_int_alpha,
    show_novelty,
    outcome_type,
    ...
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  feature_x <- feature_y <- feature_x_value <- feature_y_value <- NULL

  # Get the data that determines the plot characteristics.
  if (!is_empty(pd_data[[1L]])) {
    plot_data <- data.table::copy(pd_data[[1L]]@data)[order(feature_x_value, feature_y_value)]
  } else {
    plot_data <- facet_data
  }

  # Set value range
  if (is.null(value_range) && value_scales == "facet") {
    value_range <- .create_ice_plot_value_range(
      x = pd_data,
      scale_method = "facet",
      outcome_type = outcome_type,
      confidence_interval = conf_int_style != "none"
    )
    
  } else if (is.null(value_range)) {
    ..error_reached_unreachable_code(paste0(
      ".create_2d_ice_plot: No value range was set, and value_scales ",
      "was not recognised either."
    ))
  }

  # x_range
  if (is.numeric(plot_data$feature_x_value)) {
    if (is.null(x_range)) {
      x_range <- c(
        min(plot_data$feature_x_value, na.rm = TRUE),
        max(plot_data$feature_x_value, na.rm = TRUE)
      )
    }

    # x_breaks
    if (is.null(x_breaks)) {
      .check_input_plot_args(x_n_breaks = x_n_breaks)

      # Create breaks and update x_range
      x_breaks <- labeling::extended(
        m = x_n_breaks,
        dmin = x_range[1L],
        dmax = x_range[2L],
        only.loose = TRUE
      )

      # Update x_range
      x_range <- c(
        head(x_breaks, n = 1L),
        tail(x_breaks, n = 1L)
      )
    }

    .check_input_plot_args(x_range = x_range)
    
  } else {
    x_range <- NULL
  }

  # y_range
  if (is.numeric(plot_data$feature_y_value)) {
    if (is.null(y_range)) {
      y_range <- c(
        min(plot_data$feature_y_value, na.rm = TRUE),
        max(plot_data$feature_y_value, na.rm = TRUE)
      )
    }

    # y_breaks
    if (is.null(y_breaks)) {
      .check_input_plot_args(y_n_breaks = y_n_breaks)

      # Create breaks and update y_range
      y_breaks <- labeling::extended(
        m = y_n_breaks,
        dmin = y_range[1L],
        dmax = y_range[2L],
        only.loose = TRUE
      )

      # Update y_range
      y_range <- c(
        head(y_breaks, n = 1L),
        tail(y_breaks, n = 1L)
      )
    }

    .check_input_plot_args(y_range = y_range)
    
  } else {
    y_range <- NULL
  }

  # Find the correct value-range
  value_range <- value_range[
    feature_x == as.character(plot_data$feature_x[1L]) &
    feature_y == as.character(plot_data$feature_y[1L])
  ]
  
  if (is_empty(value_range)) {
    value_range <- c(NA_real_, NA_real_)
  } else {
    value_range <- c(value_range$min_value, value_range$max_value)
  }

  # Update show_novelty to check for non-finite values in the novelty data.
  show_novelty <- show_novelty && !is.null(plot_data$novelty) && all(is.finite(plot_data$novelty))

  if (show_novelty) {
    # Find the correct novelty-range
    novelty_range <- novelty_range[
      feature_x == as.character(plot_data$feature_x[1L]) &
      feature_y == as.character(plot_data$feature_y[1L])
    ]
    
    if (is_empty(novelty_range)) {
      novelty_range <- c(NA_real_, NA_real_)
    } else {
      novelty_range <- c(novelty_range$min_value, novelty_range$max_value)
    }
  }

  # Set x-label
  if (is.waive(x_label)) {
    if (!is.na(plot_data$feature_x[1L])) {
      x_label <- as.character(plot_data$feature_x[1L])
    } else {
      x_label <- NULL
    }
  }

  # Set y-label
  if (is.waive(y_label)) {
    if (!is.na(plot_data$feature_y[1L])) {
      y_label <- as.character(plot_data$feature_y[1L])
    } else {
      y_label <- NULL
    }
  }

  .check_input_plot_args(
    x_label = x_label,
    y_label = y_label
  )

  # Set colours used for plotting
  gradient_colours <- .get_palette(
    x = gradient_palette,
    palette_type = "sequential"
  )
  
  # Create basic plot.
  p <- ggplot2::ggplot(
    data = plot_data,
    mapping = ggplot2::aes(
      x = !!sym("feature_x_value"),
      y = !!sym("feature_y_value")
    )
  )
  p <- p + ggtheme
  
  if (!all(is.finite(value_range))) {
    # Blank elements in case the value range is unset, e.g. because the model
    # cannot compute survival probabilities.
    p <- p + ggplot2::geom_blank()
    p <- p + ggplot2::theme(
      axis.line.x = ggplot2::element_blank(),
      axis.ticks.x = ggplot2::element_blank(),
      axis.text.x = ggplot2::element_blank(),
      axis.line.y = ggplot2::element_blank(),
      axis.ticks.y = ggplot2::element_blank(),
      axis.text.y = ggplot2::element_blank()
    )
    
  } else if (show_novelty) {
    # Create point cloud with size of points by novelty -> bubblechart.
    p <- p + ggplot2::geom_point(
      data = plot_data,
      mapping = ggplot2::aes(
        colour = !!sym("value"),
        size = !!sym("novelty")
      )
    )

    # Invert novelty values, since higher values indicates greater novelty.
    p <- p + ggplot2::scale_size(
      trans = "reverse",
      limits = novelty_range
    )
    
    # Add gradient palette as fill.
    p <- p + ggplot2::scale_colour_gradientn(
      name = legend_label,
      colors = gradient_colours,
      limits = value_range
    )
    
  } else {
    # Set colour for value points.
    if (!is.null(discrete_palette)) {
      colour <- .get_palette(
        x = discrete_palette,
        n = 1L,
        palette_type = "qualitative"
      )
      
    } else {
      colour <- "white"
    }

    # Specify coordinates for rectangle.
    plot_data[
      ,
      c("xmin", "xmax") := ..set_edge_points(
        feature_x_value,
        range = x_range, 
        type = "x"
      ),
      by = "feature_y_value"
    ]
    plot_data[
      ,
      c("ymin", "ymax") := ..set_edge_points(
        feature_y_value,
        range = y_range, 
        type = "y"
      ),
      by = "feature_x_value"
    ]

    # Insert points. This will help set up the figure - otherwise the plot
    # cannot be drawn.
    p <- p + ggplot2::geom_point()

    # Create raster in case novelty is not or cannot be shown.
    p <- p + ggplot2::geom_rect(
      data = plot_data,
      mapping = ggplot2::aes(
        xmin = !!sym("xmin"),
        xmax = !!sym("xmax"),
        ymin = !!sym("ymin"),
        ymax = !!sym("ymax"),
        fill = !!sym("value")
      )
    )

    # Add gradient palette as fill.
    p <- p + ggplot2::scale_fill_gradientn(
      name = legend_label,
      colors = gradient_colours,
      limits = value_range
    )
    
    # Draw points on top.
    p <- p + ggplot2::geom_point(colour = colour)
  }

  # Update x and y scales
  if (!is.null(x_range)) p <- p + ggplot2::scale_x_continuous(breaks = x_breaks)
  if (!is.null(y_range)) p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)

  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = plot_data,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context"
      )
    }
  }

  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Plot to Cartesian coordinates.
  p <- p + ggplot2::coord_cartesian(xlim = x_range, ylim = y_range)

  return(p)
}



.determine_ice_plot_dimensions <- function(
    x,
    facet_by,
    facet_wrap_cols
) {
  # Obtain faceting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- 4.0

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))

  return(c(height, width))
}



.create_ice_plot_value_range <- function(
    x,
    scale_method, 
    outcome_type,
    confidence_interval = FALSE
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  min_value <- max_value <- NULL

  if (is_empty(x)) {
    return(c(NA_real_, NA_real_))
  }

  # Obtain value ranges.
  value_range <- lapply(
    x, 
    ..create_ice_plot_value_range,
    confidence_interval = confidence_interval
  )
  value_range <- data.table::rbindlist(
    value_range,
    use.names = TRUE
  )

  if (scale_method %in% c("fixed", "figure")) {
    if (outcome_type %in% c("binomial", "multinomial", "survival")) {
      # Check if probabilities all lie in the expected range.
      if (all(value_range$min_value >= 0.0) && all(value_range$max_value <= 1.0)) {
        value_range[, ":="(
          "min_value" = 0.0,
          "max_value" = 1.0
        )]
      } else {
        value_range[, ":="(
          "min_value" = min(min_value),
          "max_value" = max(max_value)
        )]
      }
    } else {
      value_range[, ":="(
        "min_value" = min(min_value),
        "max_value" = max(max_value)
      )]
    }
  } else if (scale_method == "facet") {
    # Facet scaling does not make the value ranges nice (i.e. map to (0,1) in
    # case of probabilities.)
    value_range[, ":="(
      "min_value" = min(min_value),
      "max_value" = max(max_value)
    )]
    
  } else if (scale_method == "feature") {
    value_range[
      ,
      ":="(
        "min_value" = min(min_value),
        "max_value" = max(max_value)
      ),
      by = c("feature_x")
    ]
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".create_ice_plot_value_range: unknown scale_method encountered: ",
      scale_method
    ))
  }

  return(value_range)
}



..create_ice_plot_value_range <- function(x, confidence_interval) {
  # Extract value range information.

  # Get the x-feature
  value_range <- list("feature_x" = x@identifiers$feature_x)

  # Get the y-feature, if present.
  if (!is.null(x@identifiers$feature_y)) {
    value_range <- c(
      value_range,
      list("feature_y" = x@identifiers$feature_y)
    )
  }

  # Get minimum and maximum values.
  if (confidence_interval) {
    value_range <- c(
      value_range, 
      list(
        "min_value" = min(x@data$value_ci_low),
        "max_value" = max(x@data$value_ci_up)
      )
    )
    
  } else {
    value_range <- c(
      value_range, 
      list(
        "min_value" = min(x@data$value),
        "max_value" = max(x@data$value)
      )
    )
  }

  # Return as data.table.
  return(data.table::as.data.table(value_range))
}



.create_ice_plot_novelty_range <- function(x, scale_method, outcome_type) {
  # Suppress NOTES due to non-standard evaluation in data.table
  min_novelty <- max_novelty <- NULL

  if (is_empty(x)) {
    return(c(NA_real_, NA_real_))
  }

  # Obtain value ranges.
  novelty_range <- lapply(x, ..create_ice_plot_novelty_range)
  novelty_range <- data.table::rbindlist(
    novelty_range, 
    use.names = TRUE
  )

  if (
    any(is.finite(novelty_range$min_novelty)) &&
    any(is.finite(novelty_range$max_novelty))
  ) {
    novelty_range[, ":="(
      "min_novelty" = min(min_novelty, na.rm = TRUE),
      "max_novelty" = max(max_novelty, na.rm = TRUE)
    )]
    
  } else {
    novelty_range[, ":="(
      "min_novelty" = NA_real_,
      "max_novelty" = NA_real_
    )]
  }
  
  return(novelty_range)
}



..create_ice_plot_novelty_range <- function(x) {
  # Extract novelty range information.

  # Get the x-feature
  novelty_range <- list("feature_x" = x@identifiers$feature_x)

  # Get the y-feature, if present.
  if (!is.null(x@identifiers$feature_y)) {
    novelty_range <- c(
      novelty_range, 
      list("feature_y" = x@identifiers$feature_y)
    )
  }

  if (any(is.finite(x@data$novelty))) {
    # Get minimum and maximum novelty features.
    novelty_range <- c(
      novelty_range,
      list(
        "min_novelty" = min(x@data$novelty, na.rm = TRUE),
        "max_novelty" = max(x@data$novelty, na.rm = TRUE)
      )
    )
  } else {
    novelty_range <- c(
      novelty_range,
      list(
        "min_novelty" = NA_real_,
        "max_novelty" = NA_real_
      )
    )
  }

  # Return as data.table.
  return(data.table::as.data.table(novelty_range))
}



.select_ice_plot_data <- function(x, data) {
  # Check that data is not empty.
  if (is_empty(data)) return(NULL)

  return(lapply(
    split(x, by = "list_id"),
    ..select_ice_plot_data,
    data = data
  ))
}



..select_ice_plot_data <- function(x, data) {
  # Suppress NOTES due to non-standard evaluation in data.table
  .NATURAL <- NULL

  # Get list identifier.
  selected_list_index <- x$list_id[1L]

  # Check if data exists or is NA.
  if (is.na(selected_list_index)) return(NULL)

  # Select the correct data element.
  data <- data[[selected_list_index]]

  # Select the subset of the data.
  data@data <- data@data[x, on = .NATURAL]

  return(data)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# .train -----------------------------------------------------------------------
setMethod(
  ".train",
  signature(
    object = "familiarNoveltyDetector",
    data = "dataObject"
  ),
  function(
    object,
    data,
    get_additional_info = FALSE,
    is_pre_processed = FALSE,
    trim_model = TRUE,
    timeout = 60000.0,
    ...
  ) {
    # Train method for novelty detectors.
    
    # Check if the class of object is a subclass of familiarNoveltyDetector..
    if (!is_subclass(class(object)[1L], "familiarNoveltyDetector")) {
      object <- promote_detector(object)
    }

    # Process data, if required.
    data <- process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = "clustering",
      force_check = TRUE
    )

    # Set the training flag
    can_train <- TRUE

    # Check if there are any data entries. The familiar model cannot be trained
    # otherwise
    if (is_empty(x = data)) can_train <- FALSE

    # Check the number of features in data; if it has no features, the familiar
    # model can not be trained
    if (!has_feature_data(x = data)) can_train <- FALSE

    # Check if the hyperparameters are plausible.
    if (!has_optimised_hyperparameters(object = object)) can_train <- FALSE
    
    # Train a new model based on data.
    if (can_train) object <- ..train(object = object, data = data)

    # Extract information required for external use.
    if (get_additional_info) {
      # Add column data
      object <- add_data_column_info(
        object = object,
        data = data
      )
    }
    
    if (trim_model) object <- trim_model(object = object, timeout = timeout)

    # Empty slots if a model can not be trained.
    if (!can_train) {
      object@required_features <- NULL
      object@model_features <- NULL
    }

    return(object)
  }
)



# show -------------------------------------------------------------------------
setMethod(
  "show",
  signature(object = "familiarNoveltyDetector"),
  function(object) {
    # Make sure the model object is updated.
    object <- update_object(object = object)

    if (!model_is_trained(object)) {
      cat(paste0(
        "A ", object@learner, " novelty detector (class: ", class(object)[1L],
        ") that was not successfully trained (", 
        .familiar_version_string(object), ").\n"
      ))
      
      return(invisible(NULL))
    }
    
    # Describe the learner and the version of familiar.
    message_str <- paste0(
      "A ", object@learner, " novelty detector (class: ", class(object)[1L],
      "; ", .familiar_version_string(object), ")"
    )
    
    # Describe the package(s), if any
    if (!is.null(object@package)) {
      message_str <- c(
        message_str,
        paste0(" trained using "),
        paste_s(mapply(
          ..message_package_version, 
          x = object@package, 
          version = object@package_version
        )),
        ifelse(length(object@package) > 1L, " packages", " package")
      )
    }
    
    # Complete message and write.
    message_str <- paste0(c(message_str, ".\n"), collapse = "")
    cat(message_str)
    
    cat(paste0("\n--------------- Detector details ---------------\n"))
    
    # Model details
    if (object@is_trimmed) {
      cat(object@trimmed_function$show, sep = "\n")
    } else {
      show(object@model)
    }
    
    cat(paste0("---------------------------------------------\n"))
    
    # Details concerning hyperparameters.
    cat("\nThe novelty detector was trained using the following hyperparameters:\n")
    invisible(lapply(
      names(object@hyperparameters),
      function(x, object) {
        cat(paste0("  ", x, ": ", object@hyperparameters[[x]], "\n"))
      },
      object = object
    ))
    
    # Details concerning model features:
    cat("\nThe following features were used for the novelty detector:\n")
    lapply(
      object@model_features, 
      function(x, object) show(object@feature_info[[x]]),
      object = object
    )
    
    # Check package version.
    check_package_version(object)
  }
)



# require_package --------------------------------------------------------------
setMethod(
  "require_package",
  signature(x = "familiarNoveltyDetector"),
  function(
    x, 
    purpose = NULL,
    message_type = "error",
    ...
  ) {
    # Skip if no package is required.
    if (is_empty(x@package)) return(invisible(TRUE))

    # Set standard purposes for common uses.
    if (!is.null(purpose)) {
      if (purpose %in% c("train", "predict")) {
        purpose <- switch(
          purpose,
          "train" = "to train a novelty detector",
          "predict" = "to assess novelty",
          "show" = "to capture output"
        )
      }
    }

    return(invisible(.require_package(
      x = x@package, 
      purpose = purpose, 
      message_type = message_type
    )))
  }
)



# set_package_version ----------------------------------------------------------
setMethod(
  "set_package_version",
  signature(object = "familiarNoveltyDetector"),
  function(object) {
    # Do not add package versions if there are no packages.
    if (is_empty(object@package)) return(object)

    # Obtain package versions.
    object@package_version <- sapply(
      object@package, 
      function(x) (as.character(utils::packageVersion(x)))
    )

    return(object)
  }
)



# check_package_version --------------------------------------------------------
setMethod(
  "check_package_version", 
  signature(object = "familiarNoveltyDetector"),
  function(object) {
    # Check whether installed packages are outdated or newer.
    .check_package_version(
      name = object@package,
      version = object@package_version,
      when = "when creating the novelty detector"
    )
  }
)



# model_is_trained (familiarNoveltyDetector) -----------------------------------
setMethod(
  "model_is_trained",
  signature(object = "familiarNoveltyDetector"),
  function(object) {
    # Check if a model was trained
    if (is.null(object@model)) {
      # Check if a model is present
      return(FALSE)
      
    } else {
      # Assume that the model is present if it is not specifically stated using
      # the model_trained element
      return(TRUE)
    }
  }
)



# is_available (familiarNoveltyDetector)----------------------------------------
setMethod(
  "is_available", 
  signature(object = "familiarNoveltyDetector"),
  function(object, ...) {
    return(FALSE)
  }
)



# get_default_hyperparameters (familiarNoveltyDetector) ------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarNoveltyDetector"),
  function(object, ...) {
    return(list())
  }
)



# ..train (familiarNoveltyDetector, dataObject) --------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarNoveltyDetector",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Set a NULL model
    object@model <- NULL

    return(object)
  }
)



# ..train (familiarNoveltyDetector, NULL) --------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarNoveltyDetector",
    data = "NULL"
  ),
  function(object, data, ...) {
    # Set a NULL model
    object@model <- NULL

    return(object)
  }
)



# ..predict (familiarNoveltyDetector, dataObject) ------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarNoveltyDetector",
    data = "dataObject"
  ),
  function(object, data, ...) {
    return(NULL)
  }
)



# trim_model (familiarNoveltyDetector) -----------------------------------------
setMethod(
  "trim_model",
  signature(object = "familiarNoveltyDetector"),
  function(object, timeout = 60000.0, ...) {
    # Do not trim the model if there is nothing to trim.
    if (!model_is_trained(object)) return(object)

    # Trim the model.
    trimmed_object <- .trim_model(object = object)

    # Skip further processing if the model object was not trimmed.
    if (!trimmed_object@is_trimmed) return(object)

    # Go over different functions.
    trimmed_object <- .replace_broken_functions(
      object = object,
      trimmed_object = trimmed_object,
      timeout = timeout
    )

    return(trimmed_object)
  }
)


# .trim_model (familiarNoveltyDetector) ----------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarNoveltyDetector"),
  function(object, ...) {
    # Default method for models that lack a more specific method.
    return(object)
  }
)


# add_package_version (familiarModel) ------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "familiarNoveltyDetector"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)


# add_data_column_info (familiarNoveltyDetector) -------------------------------
setMethod(
  "add_data_column_info",
  signature(object = "familiarNoveltyDetector"),
  function(
    object, 
    data = NULL, 
    sample_id_column = NULL, 
    batch_id_column = NULL, 
    series_id_column = NULL
  ) {
    # Don't determine new column information if this information is already
    # present.
    if (!is.null(object@data_column_info)) return(object)

    # Don't determine new column information if this information can be
    # inherited from a dataObject.
    if (is(data, "dataObject")) {
      if (!is_empty(data@data_column_info)) {
        object@data_column_info <- data@data_column_info

        return(object)
      }
    }

    # Load settings to find identifier columns
    settings <- get_settings()

    # Read from settings. If not set, these will be NULL.
    if (is.null(sample_id_column)) sample_id_column <- settings$data$sample_col
    if (is.null(batch_id_column)) batch_id_column <- settings$data$batch_col
    if (is.null(series_id_column)) series_id_column <- settings$data$series_col

    # Replace any missing.
    if (is.null(sample_id_column)) sample_id_column <- NA_character_
    if (is.null(batch_id_column)) batch_id_column <- NA_character_
    if (is.null(series_id_column)) series_id_column <- NA_character_

    # Repetition column ids are only internal.
    repetition_id_column <- NA_character_

    # Create table
    data_info_table <- data.table::data.table(
      "type" = c("batch_id_column", "sample_id_column", "series_id_column", "repetition_id_column"),
      "internal" = get_id_columns(),
      "external" = c(batch_id_column, sample_id_column, series_id_column, repetition_id_column)
    )

    # Combine into one table and add to object
    object@data_column_info <- data_info_table

    return(object)
  }
)



# set_model_features (familiarNoveltyDetector)----------------------------------
setMethod(
  "set_model_features",
  signature(object = "familiarNoveltyDetector"),
  function(
    object, 
    signature_features = NULL, 
    minimise_footprint = FALSE, 
    ...
  ) {
    
    if (is.null(signature_features)) {
      # Get signature features. This is a list features after clustering.
      signature_features <- get_signature_features(object = object)
    }
    
    # Find novelty features. The resulting list of features are features prior
    # to clustering.
    novelty_features <- find_novelty_features(
      model_features = features_before_clustering(
        features = signature_features,
        feature_info_list = object@feature_info,
        representative_only = TRUE
      ),
      feature_info_list = object@feature_info
    )
    
    if (minimise_footprint) {
      # Find only features that are required for running the model.
      required_features <- novelty_features
      
    } else {
      # Find features that are required for processing the data.
      required_features <- get_required_features(
        x = novelty_features,
        is_clustered = FALSE,
        feature_info_list = object@feature_info
      )
    }
    
    # Select only necessary feature info objects.
    available_feature_info <- names(object@feature_info) %in% required_features
    object@feature_info <- object@feature_info[available_feature_info]
    
    # Set feature-related attribute slots
    object@required_features <- required_features
    object@model_features <- novelty_features
    
    return(object)
  }
)



# get_signature_features (familiarNoveltyDetector)------------------------------
setMethod(
  "get_signature_features",
  signature(object = "familiarNoveltyDetector"),
  function(
    object
  ) {
    # Attempt to get signature directly from the object.
    if (!is_empty(object@model_features)) {
      # If model features have been set (i.e. using set_model_features), 
      # return those features instead. Since the model_features attribute stores
      # feature names prior to clustering, make sure to update these.
      return(features_after_clustering(
        features = object@model_features,
        feature_info_list = object@feature_info
      ))
    }
    
    # Get signature based on the stored feature information.
    return(.get_signature_features(object))
  }
)



# set_object_name (familiarNoveltyDetector) ------------------------------------

#' @title Set the name of a `familiarNoveltyDetector` object.
#'
#' @description Set the `name` slot using the object name.
#'
#' @param x A `familiarNoveltyDetector` object.
#'
#' @return A `familiarNoveltyDetector` object with a generated or a provided name.
#' @md
#' @keywords internal
setMethod(
  "set_object_name",
  signature(x = "familiarNoveltyDetector"),
  function(x, new = NULL) {
    
    if (is.null(x@project_id) && is.null(new)) {
      # Generate a random object name. A project_id of 0 means that the objects
      # was auto-generated (i.e. through object conversion). We randomly
      # generate characters and add a time stamp, so that collision is
      # practically impossible.
      slot(object = x, name = "name") <- paste0(
        as.character(as.numeric(format(Sys.time(), "%H%M%S"))),
        "_", rstring(n = 20L)
      )
      
    } else if (is.null(new)) {
      # Generate a sensible object name.
      slot(object = x, name = "name") <- get_object_name(object = x)
      
    } else {
      slot(object = x, name = "name") <- new
    }
    
    return(x)
  }
)



# get_object_name (model) ------------------------------------------------------
setMethod(
  "get_object_name",
  signature(object = "familiarNoveltyDetector"),
  function(object, abbreviated = FALSE) {
    
    if (abbreviated) {
      # Create an abbreviated name
      model_name <- paste0("model.", object@data_id, ".", object@run_id)
      
    } else {
      # Create the full name of the model
      model_name <- get_object_file_name(
        learner = object@learner,
        vimp_method = object@vimp_method,
        project_id = object@project_id,
        data_id = object@data_id,
        run_id = object@run_id,
        object_type = "familiarNoveltyDetector",
        with_extension = FALSE
      )
    }
    
    return(model_name)
  }
)

test_create_generic_info <- function(
    data
) {
  # Setup feature info task.
  feature_info_task <- methods::new(
    "familiarTaskGenericFeatureInfo"
  )
  
  # Feature information objects are created from the bypass dataset.
  feature_info <- .perform_task(
    object = feature_info_task,
    data = data
  )
  
  return(feature_info)
}



test_create_feature_info <- function(
    data,
    signature = NULL,
    ...
) {
  # This creates a list of featureInfo objects, with processing, based on data.
  # This code is primarily used within unit tests.
  
  # Reconstitute settings from the data.
  settings <- extract_settings_from_data(data = data)
  
  # Update some missing settings that can be fixed within this method.
  settings$data$train_cohorts <- unique(data@data[[get_id_columns(single_column = "batch")]])
  
  # Parse the remaining settings that are important.
  settings <- do.call(
    .parse_general_settings,
    args = c(
      list(
        "settings" = settings,
        "data" = data@data
      ),
      list(...)
    )
  )
  
  # Setup feature info task.
  feature_info_task <- methods::new(
    "familiarTaskFeatureInfo"
  )
  
  # Feature information objects are created from the bypass dataset.
  feature_info <- .perform_task(
    object = feature_info_task,
    data = data,
    settings = settings
  )
  
  return(feature_info)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

setClass("familiarCoreLearnVimp",
  contains = "familiarVimpMethod"
)

setClass(
  "familiarCoreLearnGiniVimp",
  contains = "familiarCoreLearnVimp"
)

setClass(
  "familiarCoreLearnMDLVimp",
  contains = "familiarCoreLearnVimp"
)

setClass(
  "familiarCoreLearnRelieffExpRankVimp",
  contains = "familiarCoreLearnVimp"
)

setClass(
  "familiarCoreLearnGainRatioVimp",
  contains = "familiarCoreLearnVimp"
)


# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarCoreLearnVimp"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Update package
    .Object@package <- "CORElearn"

    return(.Object)
  }
)


.get_available_corelearn_gini_vimp_method <- function(show_general = TRUE) {
  return("gini")
}

.get_available_corelearn_mdl_vimp_method <- function(show_general = TRUE) {
  return("mdl")
}

.get_available_corelearn_relieff_exp_rank_vimp_method <- function(show_general = TRUE) {
  return("relieff_exp_rank")
}

.get_available_corelearn_gain_ratio_vimp_method <- function(show_general = TRUE) {
  return("gain_ratio")
}



# is_available (gini) ----------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarCoreLearnGiniVimp"),
  function(object, ...) {
    return(object@outcome_type %in% c("binomial", "multinomial"))
  }
)

# is_available (mdl) -----------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarCoreLearnMDLVimp"),
  function(object, ...) {
    return(object@outcome_type %in% c("binomial", "multinomial"))
  }
)

# is_available (relieff) -------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarCoreLearnRelieffExpRankVimp"),
  function(object, ...) {
    
    if (object@outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(object@outcome_type %in% c("binomial", "multinomial", "continuous"))
  }
)

# is_available (gain ratio) ----------------------------------------------------
setMethod(
  "is_available", 
  signature(object = "familiarCoreLearnGainRatioVimp"),
  function(object, ...) {
    return(object@outcome_type %in% c("binomial", "multinomial"))
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarCoreLearnVimp"),
  function(object, data = NULL, ...) {
    return(list())
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarCoreLearnVimp"),
  function(object, data, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL

    if (is_empty(data)) return(callNextMethod())

    # Check that required packages are loaded and installed.
    require_package(object, "vimp")

    # Identify feature columns.
    feature_columns <- get_feature_columns(data)

    # Generate a formula.
    formula <- stats::reformulate(feature_columns, response = quote(outcome))

    if (is(object, "familiarCoreLearnGiniVimp")) {
      # Gini measure.
      score <- CORElearn::attrEval(
        formula, 
        data = data@data, 
        estimator = "Gini"
      )
      
    } else if (is(object, "familiarCoreLearnMDLVimp")) {
      # MDL method.
      score <- CORElearn::attrEval(
        formula, 
        data = data@data, 
        estimator = "MDL"
      )
      
    } else if (is(object, "familiarCoreLearnRelieffExpRankVimp")) {
      if (object@outcome_type %in% c("continuous")) {
        # RReliefFexpRank method.
        score <- CORElearn::attrEval(
          formula, 
          data = data@data, 
          estimator = "RReliefFexpRank"
        )
        
      } else if (object@outcome_type %in% c("binomial", "multinomial")) {
        # ReliefFexpRank method.
        score <- CORElearn::attrEval(
          formula, 
          data = data@data, 
          estimator = "ReliefFexpRank"
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }
    } else if (is(object, "familiarCoreLearnGainRatioVimp")) {
      # Gain ration method.
      score <- CORElearn::attrEval(
        formula, 
        data = data@data, 
        estimator = "GainRatio"
      )
      
    } else {
      ..error_reached_unreachable_code(
        "..vimp,familiarCoreLearnVimp: unknown class encountered."
      )
    }

    # Create variable importance object.
    vimp_object <- methods::new(
      "vimpTable",
      vimp_table = data.table::data.table(
        "score" = score, 
        "name" = names(score)
      ),
      score_aggregation = "max",
      invert = TRUE
    )

    return(vimp_object)
  }
)

familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_corelearn_gini_vimp_method(show_general = TRUE))
familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_corelearn_mdl_vimp_method(show_general = TRUE))
familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_corelearn_relieff_exp_rank_vimp_method(show_general = TRUE))
familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_corelearn_gain_ratio_vimp_method(show_general = TRUE))

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_corelearn_gini_vimp_method(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE)
familiar:::test_all_vimp_methods(
  familiar:::.get_available_corelearn_gini_vimp_method(show_general = FALSE))
familiar:::test_all_vimp_methods_parallel(
  familiar:::.get_available_corelearn_gini_vimp_method(show_general = FALSE))


familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_corelearn_mdl_vimp_method(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE)
familiar:::test_all_vimp_methods(
  familiar:::.get_available_corelearn_mdl_vimp_method(show_general = FALSE))
familiar:::test_all_vimp_methods_parallel(
  familiar:::.get_available_corelearn_mdl_vimp_method(show_general = FALSE))


familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_corelearn_relieff_exp_rank_vimp_method(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE)
familiar:::test_all_vimp_methods(
  familiar:::.get_available_corelearn_relieff_exp_rank_vimp_method(show_general = FALSE))
familiar:::test_all_vimp_methods_parallel(
  familiar:::.get_available_corelearn_relieff_exp_rank_vimp_method(show_general = FALSE))


familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_corelearn_gain_ratio_vimp_method(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE)
familiar:::test_all_vimp_methods(
  familiar:::.get_available_corelearn_gain_ratio_vimp_method(show_general = FALSE))
familiar:::test_all_vimp_methods_parallel(
  familiar:::.get_available_corelearn_gain_ratio_vimp_method(show_general = FALSE))

# Continuous outcome -----------------------------------------------------------
data <- familiar:::test_create_good_data("continuous")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "relieff_exp_rank",
  vimp_method_parameter_list = NULL,
  outcome_type = "continuous",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(paste0(
  "RReliefF exponentially decreasing ranks method correctly ranks continuous data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Binomial outcome -------------------------------------------------------------
data <- familiar:::test_create_good_data("binomial")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "relieff_exp_rank",
  vimp_method_parameter_list = NULL,
  outcome_type = "binomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(paste0(
  "RReliefF exponentially decreasing ranks method correctly ranks binomial data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "gini",
  vimp_method_parameter_list = NULL,
  outcome_type = "binomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(paste0("The Gini method correctly ranks binomial data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "gain_ratio",
  vimp_method_parameter_list = NULL,
  outcome_type = "binomial",
  cluster_method = "none",
  imputation_method = "simple"
)

testthat::test_that(paste0("The gain ratio method correctly ranks binomial data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "mdl",
  vimp_method_parameter_list = NULL,
  outcome_type = "binomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(paste0("The MDL method correctly ranks binomial data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Multinomial outcome ----------------------------------------------------------
data <- familiar:::test_create_good_data("multinomial")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "relieff_exp_rank",
  vimp_method_parameter_list = NULL,
  outcome_type = "multinomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(paste0(
  "RReliefF exponentially decreasing ranks method correctly ranks multinomial outcome data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "gini",
  vimp_method_parameter_list = NULL,
  outcome_type = "multinomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(paste0("The Gini method correctly ranks multinomial outcome data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "gain_ratio",
  vimp_method_parameter_list = NULL,
  outcome_type = "multinomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(paste0("The gain ratio method correctly ranks multinomial outcome data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "mdl",
  vimp_method_parameter_list = NULL,
  outcome_type = "multinomial",
  cluster_method = "none",
  imputation_method = "simple")

testthat::test_that(paste0("The MDL method correctly ranks multinomial outcome data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)))

  # Expect that the vimp table has six rows.
  testthat::expect_equal(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(familiar:::get_feature_columns(data) %in% vimp_table$name))
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

# Environment holding global variables and settings.
familiar_global_env <- new.env(parent = emptyenv())

# Prevents notes on co-routines using the coro package. The coro developers may
# be able to fix this issue at some point. See
# https://github.com/r-lib/coro/issues/40
utils::globalVariables(c(
  "generator_env",
  "exits"
))

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarGLMnet objects -------------------------------------------------------

## familiarGLMnet parent -------------------------------------------------------
setClass(
  "familiarGLMnet",
  contains = "familiarModel",
  slots = list(
    "encoding_reference_table" = "ANY",
    "feature_order" = "character"
  ),
  prototype = list(
    "encoding_reference_table" = NULL,
    "feature_order" = character()
  )
)

## familiarGLMnetRidge ---------------------------------------------------------
setClass(
  "familiarGLMnetRidge",
  contains = "familiarGLMnet"
)

## familiarGLMnetLasso ---------------------------------------------------------
setClass(
  "familiarGLMnetLasso",
  contains = "familiarGLMnet"
)

## familiarGLMnetElasticNet ----------------------------------------------------
setClass(
  "familiarGLMnetElasticNet",
  contains = "familiarGLMnet"
)

## familiarGLMnetLassoTest -----------------------------------------------------
setClass(
  "familiarGLMnetLassoTest",
  contains = "familiarGLMnetLasso"
)

## familiarGLMnetLassoTestAllFail ----------------------------------------------
# This class predicts NA for all samples.
setClass(
  "familiarGLMnetLassoTestAllFail",
  contains = "familiarGLMnetLassoTest"
)

## familiarGLMnetLassoTestSomeFail ---------------------------------------------
# This class predicts NA for some sample samples,but not all.
setClass(
  "familiarGLMnetLassoTestSomeFail",
  contains = "familiarGLMnetLassoTest"
)

## familiarGLMnetLassoTestAllExtreme -------------------------------------------
# This class predicts probabilities that are always exactly 0 or 1.
setClass(
  "familiarGLMnetLassoTestAllExtreme",
  contains = "familiarGLMnetLassoTest"
)


# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarGLMnet"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set the required package
    .Object@package <- "glmnet"

    return(.Object)
  }
)


# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarGLMnet"),
  function(object, ...) {
    # Extract learner and outcome_type.
    learner <- object@learner
    outcome_type <- object@outcome_type

    if (
      outcome_type == "survival" &&
      learner %in% c(
        "elastic_net", "elastic_net_cox", "lasso", "lasso_cox",
        "ridge", "ridge_cox"
      )
    ) {
      return(TRUE)
      
    } else if (
      outcome_type == "continuous" &&
      learner %in% c(
        "elastic_net", "elastic_net_gaussian", "elastic_net_poisson",
        "lasso", "lasso_gaussian", "lasso_poisson",
        "ridge", "ridge_gaussian", "ridge_poisson"
      )
    ) {
      return(TRUE)
      
    } else if (
      outcome_type == "multinomial" &&
      learner %in% c(
        "elastic_net", "elastic_net_multinomial",
        "lasso", "lasso_multinomial",
        "ridge", "ridge_multinomial"
      )
    ) {
      return(TRUE)
      
    } else if (
      outcome_type == "binomial" &&
      learner %in% c(
        "elastic_net", "elastic_net_binomial",
        "lasso", "lasso_binomial",
        "ridge", "ridge_binomial"
      )
    ) {
      return(TRUE)
      
    } else if (
      outcome_type == "count" && learner %in% c(
        "elastic_net", "elastic_net_poisson",
        "lasso", "lasso_poisson",
        "ridge", "ridge_poisson"
      )
    ) {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(FALSE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarGLMnet"),
  function(object, data = NULL, user_list = NULL, ...) {
    # Initialise list and declare hyperparameter entries.
    param <- list()
    param$sign_size <- list()
    param$family <- list()
    param$lambda_min <- list()
    param$n_folds <- list()
    param$normalise <- list()
    param$sample_weighting <- list()
    param$sample_weighting_beta <- list()

    if (is(object, "familiarGLMnetElasticNet")) param$alpha <- list()

    # If data is not provided, return the list with hyperparameter names only.
    if (is.null(data)) return(param)

    # Internal
    outcome_type <- data@outcome_type

    # Determine the family.
    fam <- sub_all_patterns(
      x = object@learner,
      pattern = c("elastic_net", "lasso", "ridge"), 
      replacement = "",
      fixed = TRUE
    )
    
    if (fam != "") {
      fam <- sub(
        x = fam, 
        pattern = "_",
        replacement = "",
        fixed = TRUE
      )
    }

    # Check for lasso_test
    if (object@learner %in% c("lasso_test_all_fail", "lasso_test_some_fail", "lasso_test_extreme")) {
      fam <- switch(
        outcome_type,
        "continuous" = "gaussian",
        "binomial" = "binomial",
        "multinomial" = "multinomial",
        "survival" = "cox"
      )
    }

    # Determine number of subjects
    n_samples <- data.table::uniqueN(
      data@data,
      by = get_id_columns(id_depth = "series")
    )

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data)

    # family -------------------------------------------------------------------
    if (fam == "") {
      if (outcome_type == "continuous") {
        family_default <- c("gaussian", "poisson")
      } else if (outcome_type == "binomial") {
        family_default <- "binomial"
      } else if (outcome_type == "multinomial") {
        family_default <- "multinomial"
      } else if (outcome_type == "survival") {
        family_default <- "cox"
      }
      
    } else {
      family_default <- fam
    }

    # Set family parameter
    param$family <- .set_hyperparameter(
      default = family_default,
      type = "factor",
      range = family_default,
      randomise = length(family_default) > 1L
    )

    # lambda indicating the optimal model complexity ---------------------------
    param$lambda_min <- .set_hyperparameter(
      default = "lambda.min",
      type = "factor",
      range = c("lambda.1se", "lambda.min"),
      randomise = FALSE
    )

    # number of cross-validation folds -----------------------------------------

    # glmnet requires at least 3 folds. The default number of cross-validation
    # folds may grow up to 20, for data sets > 200 samples.
    n_folds_default <- min(c(20L, max(c(3L, as.integer(floor(n_samples / 10.0))))))

    # Set the number of cross-validation folds.
    param$n_folds <- .set_hyperparameter(
      default = n_folds_default,
      type = "integer",
      range = c(3L, n_samples),
      valid_range = c(3L, Inf),
      randomise = FALSE
    )

    # feature normalisation ----------------------------------------------------

    # By default, normalisation is part of the pre-processing of familiar, but
    # the user may have disabled it. In that the case, the user can set
    # normalisation to TRUE to avoid complaints by glmnet.
    param$normalise <- .set_hyperparameter(
      default = FALSE,
      type = "logical",
      range = c(FALSE, TRUE),
      randomise = FALSE
    )
    
    # sample weighting method -------------------------------------------------
    
    # Class imbalances may lead to learning majority classes. This can be
    # partially mitigated by increasing weight of minority classes.
    param$sample_weighting <- .get_default_sample_weighting_method(outcome_type = outcome_type)

    # effective number of samples beta ----------------------------------------
    
    # Specifies the beta parameter for effective number sample weighting method.
    # See Cui et al. (2019).
    param$sample_weighting_beta <- .get_default_sample_weighting_beta(
      method = c(
        param$sample_weighting$init_config,
        user_list$sample_weighting
      ),
      outcome_type = outcome_type
    )

    if (is(object, "familiarGLMnetElasticNet")) {
      # elastic net mixing parameter -------------------------------------------
      
      # Set alpha parameter. Alpha = 1 is lasso, alpha = 0 is ridge. glmnet
      # requires alpha to be in the closed interval [0, 1].
      param$alpha <- .set_hyperparameter(
        default = c(0.0, 1.0 / 3.0, 2.0 / 3.0, 1.0),
        type = "numeric",
        range = c(0.0, 1.0),
        valid_range = c(0.0, 1.0),
        randomise = TRUE
      )
    }
    
    # Return hyperparameters
    return(param)
  }
)



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarGLMnet"),
  function(object, type = "default") {
    if (
      object@outcome_type != "survival" &&
      object@learner %in% c("elastic_net", "elastic_net_cox", "lasso", "lasso_cox", "ridge", "ridge_cox")
    ) {
      return(callNextMethod())
    }

    # Default are hazard ratios.
    if (type == "default") {
      return("hazard_ratio")
    } else if (type == "survival_probability") {
      return("survival_probability")
    } else {
      ..error_reached_unreachable_code("get_prediction_type,familiarGLMnet: unknown type")
    }
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarGLMnet",
    data = "dataObject"
  ),
  function(
    object,
    data,
    force_signature = FALSE, 
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    original_name <- NULL

    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object, 
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # For data with one feature, switch to familiarGLM.
    if (get_n_features(data) == 1L) {
      # Create a familiarGLM object.
      object <- methods::new("familiarGLM", object)

      return(..train(
        object = object,
        data = data,
        ...
      ))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Use effect coding to convert categorical data into encoded data - this is
    # required to deal with factors with missing/new levels between training and
    # test data sets.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )

    # Find feature columns in the data.
    feature_columns <- get_feature_columns(x = encoded_data$encoded_data)

    # Parse outcome data.
    if (object@outcome_type == "survival") {
      outcome_data <- survival::Surv(
        data@data$outcome_time,
        data@data$outcome_event
      )
      
    } else {
      outcome_data <- data@data$outcome
    }

    # Determine id columns
    id_columns <- get_id_columns("series")

    # Generate folds using our own fold generating algorithm to handle repeated
    # measurements
    fold_table <- .create_cv(
      n_folds = object@hyperparameters$n_folds,
      outcome_type = object@outcome_type,
      data = encoded_data$encoded_data@data,
      stratify = FALSE,
      return_fold_id = TRUE
    )

    # Order according to samples in encoded_data$encoded_data@data so that
    # fold_id corresponds to the correct rows.
    fold_table <- merge(
      x = fold_table,
      y = encoded_data$encoded_data@data[, mget(id_columns)],
      by = id_columns
    )

    if (force_signature) {
      # Find signature features.
      signature_feature <- names(object@feature_info)[sapply(object@feature_info, is_in_signature)]

      if (length(signature_feature) > 0.0) {
        # Initially mark all features for shrinkage.
        penalty_factor <- rep(1.0, length(feature_columns))

        # Update all signature features that were not encoded.
        penalty_factor[feature_columns %in% signature_feature] <- 0.0

        # Update all signatures features that were encoded.
        encoded_signature <- encoded_data$reference_table[
          original_name %in% signature_feature
        ]$reference_name
        penalty_factor[feature_columns %in% encoded_signature] <- 0.0
        
      } else {
        # Allow shrinking of each feature.
        penalty_factor <- rep(1.0, length(feature_columns))
      }
      
    } else {
      # Allow shrinking of each feature.
      penalty_factor <- rep(1.0, length(feature_columns))
    }

    # Set weights
    weights <- create_instance_weights(
      data = encoded_data$encoded_data,
      method = object@hyperparameters$sample_weighting,
      beta = ..compute_effective_number_of_samples_beta(
        object@hyperparameters$sample_weighting_beta
      ),
      normalisation = "average_one"
    )

    # Get the arguments which are shared between all different objects.
    learner_arguments <- list(
      "x" = as.matrix(encoded_data$encoded_data@data[, mget(feature_columns)]),
      "y" = outcome_data,
      "family" = as.character(object@hyperparameters$family),
      "weights" = weights,
      "standardize" = object@hyperparameters$normalise,
      "nfolds" = NULL,
      "foldid" = fold_table$fold_id,
      "parallel" = FALSE,
      "penalty.factor" = penalty_factor
    )

    # Set learner-specific arguments.
    if (is(object, "familiarGLMnetRidge")) {
      learner_arguments <- c(learner_arguments, list("alpha" = 0.0))
    } else if (is(object, "familiarGLMnetLasso")) {
      learner_arguments <- c(learner_arguments, list("alpha" = 1.0))
    } else if (is(object, "familiarGLMnetElasticNet")) {
      learner_arguments <- c(learner_arguments, list("alpha" = object@hyperparameters$alpha))
    } else {
      ..error_reached_unreachable_code(paste0(
        "..train,familiarGLMnet: encountered unknown learner of unknown class: ",
        paste_s(class(object))
      ))
    }

    # Train the model.
    model <- do.call_with_handlers(
      glmnet::cv.glmnet,
      args = learner_arguments
    )

    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) {
      return(callNextMethod(object = object))
    }

    # Add model
    object@model <- model

    # Add the contrast references to the object.
    object@encoding_reference_table <- encoded_data$reference_table

    # Add feature order
    object@feature_order <- feature_columns

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarGLMnet",
    data = "dataObject"
  ),
  function(object, data, ...) {
    if (object@outcome_type %in% c("continuous", "binomial", "multinomial")) {
      # Turn into a Naive model.
      object <- methods::new("familiarNaiveModel", object)
      
    } else if (object@outcome_type %in% c("survival")) {
      # Turn into a Naive model.
      object <- methods::new("familiarNaiveCoxModel", object)
    }

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarGLMnet",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "default",
    time = NULL,
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }
    
    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }
    
    # Encode data so that the features are the same as in the training.
    encoded_data <- encode_categorical_variables(
      data = data,
      object = object,
      encoding_method = "dummy",
      drop_levels = FALSE
    )
    
    if (type == "default") {
      # default ----------------------------------------------------------------

      if (object@outcome_type == "binomial") {
        # Binomial outcomes ----------------------------------------------------

        # Obtain class levels.
        class_levels <- get_outcome_class_levels(x = object)
        
        # Use the model to predict class probabilities.
        model_predictions <- predict(
          object = object@model,
          newx = as.matrix(
            encoded_data$encoded_data@data[, mget(object@feature_order)]
          ),
          s = as.character(object@hyperparameters$lambda_min),
          type = "response"
        )

        # Set as list so that the positive class can be directly inferred
        # without throwing a warning.
        prediction_list <- list()
        prediction_list[[tail(class_levels, n = 1L)]] <- model_predictions
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type == "multinomial") {
        # Multinomial outcomes -------------------------------------------------

        # Obtain class levels.
        class_levels <- get_outcome_class_levels(x = object)
        
        # Use the model to predict class probabilities.
        model_predictions <- predict(
          object = object@model,
          newx = as.matrix(
            encoded_data$encoded_data@data[, mget(object@feature_order)]
          ),
          s = as.character(object@hyperparameters$lambda_min),
          type = "response"
        )[, , 1L]

        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          if (is.matrix(model_predictions)) {
            prediction_list[[class_levels[ii]]] <- model_predictions[, class_levels[ii]]
          } else {
            prediction_list[[class_levels[ii]]] <- model_predictions[class_levels[ii]]
          }
        }
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type == "continuous") {
        # Continuous outcomes --------------------------------------------------

        # Use the model for prediction.
        model_predictions <- predict(
          object = object@model,
          newx = as.matrix(encoded_data$encoded_data@data[, mget(object@feature_order)]),
          s = as.character(object@hyperparameters$lambda_min),
          type = "response"
        )

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = model_predictions,
          type = "regression",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type == "survival") {
        # Survival outcomes ----------------------------------------------------
        
        # Use the model for prediction.
        model_predictions <- predict(
          object = object@model,
          newx = as.matrix(encoded_data$encoded_data@data[, mget(object@feature_order)]),
          s = as.character(object@hyperparameters$lambda_min),
          type = "response"
        )
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = model_predictions,
          type = "hazard_ratio",
          data = data,
          model_object = object
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      return(prediction_table)
      
    } else if (type == "survival_probability" && object@outcome_type == "survival") {
      # survival probability ---------------------------------------------------

      # If time is unset, read the max time stored by the model.
      if (is.null(time)) time <- object@settings$time_max
      
      return(.survival_probability_relative_risk(
        object = object,
        data = data,
        time = time
      ))  
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified method --------------------------------------------------

      # Use the model to predict class probabilities.
      return(predict(
        object = object@model,
        newx = as.matrix(
          encoded_data$encoded_data@data[, mget(object@feature_order)]
        ),
        s = as.character(object@hyperparameters$lambda_min),
        type = type,
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp", signature(object = "familiarGLMnet"),
  function(object, data = NULL, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL

    # Attempt to train the model if it has not been trained yet.
    if (!model_is_trained(object)) {
      object <- .train(
        object = object,
        data = data,
        get_additional_info = FALSE,
        trim_model = FALSE,
        force_signature = TRUE
      )
    }

    # Check if the model has been trained upon retry.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }

    # Check if the model is a familiarGLMnet object, and not familiarGLM (which
    # happens for one-feature datasets).
    if (!is(object, "familiarGLMnet")) {
      return(..vimp(object = object, data = data))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "vimp")

    if (object@hyperparameters$family == "multinomial") {
      # Read coefficient lists
      coefficient_list <- coef(
        object@model,
        s = as.character(object@hyperparameters$lambda_min)
      )

      # Parse into matrix and retrieve row names
      coefficient_matrix <- sapply(coefficient_list, as.matrix)
      rownames(coefficient_matrix) <- dimnames(coefficient_list[[1L]])[[1L]]

      # Compute variable importance score
      vimp_score <- apply(abs(coefficient_matrix), 1L, max)
      
    } else {
      # Read coefficient matrix
      coefficient_matrix <- as.matrix(coef(
        object@model,
        s = as.character(object@hyperparameters$lambda_min)
      ))

      # Compute variable importance score
      vimp_score <- abs(coefficient_matrix)[, 1L]
    }

    # Remove intercept from the variable importances.
    vimp_score <- vimp_score[names(vimp_score) != "(Intercept)"]
    
    if (length(vimp_score) == 0L) {
      return(callNextMethod())
    }

    # Assign to variable importance table.
    vimp_table <- data.table::data.table(
      "score" = vimp_score,
      "name" = names(vimp_score)
    )

    # Throw out elements with 0.0 coefficients
    vimp_table <- vimp_table[score != 0.0]

    # Check if any features remain.
    if (is_empty(vimp_table)) {
      return(callNextMethod())
    }

    # Create variable importance object.
    vimp_object <- methods::new("vimpTable",
      vimp_table = vimp_table,
      encoding_table = object@encoding_reference_table,
      score_aggregation = "max",
      invert = TRUE
    )

    return(vimp_object)
  }
)



# ..set_calibration_info--------------------------------------------------------
setMethod(
  "..set_calibration_info",
  signature(object = "familiarGLMnet"),
  function(object, data) {
    # Check if calibration info already.
    if (has_calibration_info(object)) {
      return(object)
    }

    if (object@outcome_type == "survival") {
      # Determine baseline survival.
      object@calibration_info <- get_baseline_survival(data = data)
    } else {
      return(callNextMethod())
    }

    return(object)
  }
)



# .trim_model-------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarGLMnet"),
  function(object, ...) {
    # Update model.
    object@model <- ..trim_glmnet(object@model)

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    # Default method for models that lack a more specific method.
    return(object)
  }
)



# is_available (test) ----------------------------------------------------------
setMethod(
  "is_available", signature(object = "familiarGLMnetLassoTest"),
  function(object, ...) {
    return(TRUE)
  }
)

# is_available (extreme test) --------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarGLMnetLassoTestAllExtreme"),
  function(object, ...) {
    return(object@outcome_type %in% c("binomial", "multinomial"))
  }
)



# ..predict (all fail) ---------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarGLMnetLassoTestAllFail",
    data = "dataObject"
  ),
  function(object, data, type = "default", ...) {
    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }

    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }

    # Get the prediction table using the model.
    prediction_table <- callNextMethod()
    
    if (is_empty(prediction_table)) {
      return(prediction_table)
    }
    
    # Insert NA values.
    for (current_column in colnames(prediction_table@prediction_data)) {
      data.table::set(
        prediction_table@prediction_data,
        j = current_column,
        value = NA
      )
    }
    
    return(prediction_table)
  }
)



# ..predict (some fail) --------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarGLMnetLassoTestSomeFail",
    data = "dataObject"
  ),
  function(object, data, type = "default", ...) {
    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }

    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }

    # Get a prediction table.
    prediction_table <- callNextMethod()

    if (is_empty(prediction_table)) {
      return(prediction_table)
    }
    
    # Insert NA values.
    for (current_column in colnames(prediction_table@prediction_data)) {
      data.table::set(
        prediction_table@prediction_data,
        i = 1L,
        j = current_column,
        value = NA
      )
    }

    return(prediction_table)
  }
)


# ..predict (extreme) --------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarGLMnetLassoTestAllExtreme",
    data = "dataObject"
  ),
  function(object, data, type = "default", ...) {
    ..set_max <- function(x) {
      y <- numeric(length(x))
      y[which.max(x)] <- 1.0
      names(y) <- names(x)
      return(as.list(y))
    }
    
    # Suppress NOTES due to non-standard evaluation in data.table
    predicted_outcome <- NULL
    
    # Check if the model was trained.
    if (!model_is_trained(object)) {
      return(callNextMethod())
    }
    
    # Check if the data is empty.
    if (is_empty(data)) {
      return(callNextMethod())
    }
    
    # Get a prediction table.
    prediction_table <- callNextMethod()
    
    if (is_empty(prediction_table)) {
      return(prediction_table)
    }
    
    if (object@outcome_type %in% c("binomial", "multinomial")) {
      # Get class levels
      class_levels <- get_outcome_class_levels(x = object)
      
      # Set probability to 1.0 for the column with the highest predicted
      # probability. First add maximum probability for each instance.
      x <- prediction_table@prediction_data[, ..set_max(.SD), by = .I]
      prediction_table@prediction_data <- x[, mget(class_levels)]
      
    } else if (object@outcome_type == "continuous") {
      # Set predicted outcome.
      prediction_table@prediction_data[, "predicted_outcome" := data@data$outcome]
      
    } else if (object@outcome_type == "survival") {
      # Set predicted outcome.
      prediction_table@prediction_data[, "predicted_outcome" := 1.0 / data@data$outcome_time]
      prediction_table@prediction_data[, "predicted_outcome" := predicted_outcome - mean(predicted_outcome) + 1.0]
      
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }
    
    return(prediction_table)
  }
)



# get_prediction_type (test) ---------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarGLMnetLassoTest"),
  function(object, type = "default") {
    if (object@outcome_type != "survival") {
      return(callNextMethod())
    }

    # Default are hazard ratios.
    if (type == "default") {
      return("hazard_ratio")
    } else if (type == "survival_probability") {
      return("survival_probability")
    } else {
      ..error_reached_unreachable_code("get_prediction_type,familiarGLMnetLassoTest: unknown type")
    }
  }
)



.get_available_glmnet_ridge_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c(
    "ridge", "ridge_gaussian", "ridge_poisson", "ridge_binomial",
    "ridge_multinomial", "ridge_cox"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("ridge"))
  }
  
  return(learners)
}



.get_available_glmnet_lasso_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c(
    "lasso", "lasso_gaussian", "lasso_poisson", "lasso_binomial",
    "lasso_multinomial", "lasso_cox"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("lasso"))
  }
  
  return(learners)
}



.get_available_glmnet_elastic_net_learners <- function(show_general = TRUE) {
  # Learners
  learners <- c(
    "elastic_net", "elastic_net_gaussian", "elastic_net_poisson",
    "elastic_net_binomial", "elastic_net_multinomial", "elastic_net_cox"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("elastic_net"))
  }
  
  return(learners)
}



.get_available_glmnet_ridge_vimp_methods <- function(show_general = TRUE) {
  return(.get_available_glmnet_ridge_learners(show_general = show_general))
}



.get_available_glmnet_lasso_vimp_methods <- function(show_general = TRUE) {
  return(.get_available_glmnet_lasso_learners(show_general = show_general))
}



.get_available_glmnet_elastic_net_vimp_methods <- function(show_general = TRUE) {
  return(.get_available_glmnet_elastic_net_learners(show_general = show_general))
}



.get_available_glmnet_lasso_learners_test_all_fail <- function(show_general = TRUE) {
  return("lasso_test_all_fail")
}



.get_available_glmnet_lasso_learners_test_some_fail <- function(show_general = TRUE) {
  return("lasso_test_some_fail")
}



.get_available_glmnet_lasso_learners_test_extreme <- function(show_general = TRUE) {
  return("lasso_test_extreme")
}



..trim_glmnet <- function(object) {
  # Function to trim glmnet objects.
  
  # Check if the object is a glmnet object.
  if (!(inherits(object, "glmnet") || inherits(object, "cv.glmnet"))) {
    return(object)
  }
  
  # Replace calls
  object$call <- call("nullcall")
  
  # Specific to cv.glmnet.
  if (!is.null(object$glmnet.fit)) {
    object$glmnet.fit$call <- call("nullcall")
  }
  
  return(object)
}

#' Internal function for parsing file paths
#'
#' @param config A list of settings, e.g. from an xml file.
#' @param project_dir (*optional*) Path to the project directory. `familiar`
#'   checks if the directory indicated by `experiment_dir` and data files in
#'   `data_file` are relative to the `project_dir`.
#'
#' @param experiment_dir (**recommended**) Path to the directory where all
#'   intermediate and final results produced by `familiar` are written to.
#'
#'   The `experiment_dir` can be a path relative to `project_dir` or an absolute
#'   path.
#'
#'   In case no project directory is provided and the experiment directory is
#'   not on an absolute path, a directory will be created in the temporary R
#'   directory indicated by `tempdir()`. This directory is deleted after closing
#'   the R session or once data analysis has finished. All information will be
#'   lost afterwards. Hence, it is recommended to provide either
#'   `experiment_dir` as an absolute path, or provide both `project_dir` and
#'   `experiment_dir`.
#' @param data_file (*optional*) Path to files containing data that should be
#'   analysed. The paths can be relative to `project_dir` or absolute paths. An
#'   error will be raised if the file cannot be found.
#'
#'   The following types of data are supported.
#'
#'   * `csv` files containing column headers on the first row, and samples per
#'   row. `csv` files are read using `data.table::fread`.
#'
#'   * `rds` files that contain a `data.table` or `data.frame` object. `rds`
#'   files are imported using `base::readRDS`.
#'
#'   * `RData` files that contain a single `data.table` or `data.frame` object.
#'   `RData` files are imported using `base::load`.
#'
#'   All data are expected in wide format, with sample information organised
#'   row-wise.
#'
#'   More than one data file can be provided. `familiar` will try to combine
#'   data files based on column names and identifier columns.
#'
#'   Alternatively, data can be provided using the `data` argument. These data
#'   are expected to be `data.frame` or `data.table` objects or paths to data
#'   files. The latter are handled in the same way as file paths provided to
#'   `data_file`.
#' @param verbose Sets verbosity.
#' @param ... Unused arguments
#'
#' @return List of paths to important directories and files.
#'
#' @md
#' @keywords internal
.parse_file_paths <- function(
    config = NULL,
    project_dir = waiver(),
    experiment_dir = waiver(),
    data_file = waiver(),
    verbose = TRUE,
    ...
) {
  # Initialise list of file paths
  file_paths <- list()

  # Read project directory; if none is provided through the configuration file
  # or the function input, defer to the R temp directory
  project_dir <- .parse_arg(
    x_config = config$paths$project_dir,
    x_var = project_dir,
    var_name = "project_dir",
    type = "character",
    optional = TRUE,
    default = NULL
  )

  # Read experiment directory path and create the directory if required
  experiment_dir <- .parse_arg(
    x_config = config$paths$experiment_dir,
    x_var = experiment_dir,
    var_name = "experiment_dir",
    type = "character",
    optional = TRUE,
    default = NULL
  )

  # Set temporary directory flag to FALSE by default.
  file_paths$is_temporary <- FALSE

  # Check how project_dir and experiment_dir settings should be interpreted.
  if (!is.null(project_dir)) {
    # Cases 1-3. project_dir exists.

    # Check that project_dir exists, and can be accessed.
    project_dir <- normalizePath(project_dir, mustWork = TRUE)

    if (is.null(experiment_dir)) {
      # Case 1. project_dir exists and experiment_dir does not.
      experiment_dir <- project_dir
      
    } else if (.is_absolute_path(experiment_dir)) {
      # Case 2. project_dir exists and experiment_dir is on its own path.
      experiment_dir <- normalizePath(experiment_dir, mustWork = FALSE)
      
    } else {
      # Case 3. project_dir exists and experiment_dir is a subdirectory.
      experiment_dir <- normalizePath(
        file.path(project_dir, experiment_dir), mustWork = FALSE
      )
    }
    
  } else {
    # Cases 4-6. project_dir does not exist.
    if (is.null(experiment_dir)) {
      # Case 4. project_dir does not exist, and neither does experiment_dir.

      # Work in a temporary directory.
      temporary_directory <- file.path(tempdir(), "familiar", rstring(n = 8L))
      experiment_dir <- project_dir <- normalizePath(temporary_directory, mustWork = FALSE)
      file_paths$is_temporary <- TRUE
      
    } else if (.is_absolute_path(experiment_dir)) {
      # Case 5. project_dir does not exist, and experiment_dir is on its own
      # path.
      project_dir <- experiment_dir <- normalizePath(experiment_dir, mustWork = FALSE)
      
    } else {
      # Case 6. project_dir does not exist, and experiment_dir is a
      # subdirectory. We ignore the fact that experiment_dir may be a
      # subdirectory, and assign a random name instead.

      # Work in a temporary directory.
      temporary_directory <- file.path(tempdir(), "familiar", rstring(n = 8L))
      experiment_dir <- project_dir <- normalizePath(temporary_directory, mustWork = FALSE)
      file_paths$is_temporary <- TRUE
    }
  }

  file_paths$experiment_dir <- experiment_dir
  if (!dir.exists(file_paths$experiment_dir)) {
    dir.create(file_paths$experiment_dir, recursive = TRUE)
  }

  # Log file - set as global variable as well
  if (!file_paths$is_temporary) {
    file_paths$log_file <- normalizePath(
      file.path(experiment_dir, "log.txt"),
      mustWork = FALSE
    )
  }

  # Assign to global environment.
  assign(
    "log_file", 
    file_paths$log_file, 
    envir = familiar_global_env
  )

  # Directory for iterations
  file_paths$iterations_dir <- normalizePath(
    experiment_dir,
    mustWork = FALSE
  )
  if (!dir.exists(file_paths$iterations_dir)) dir.create(file_paths$iterations_dir)

  # Directory for pre-processing
  file_paths$process_data_dir <- normalizePath(
    experiment_dir,
    mustWork = FALSE
  )
  if (!dir.exists(file_paths$process_data_dir)) dir.create(file_paths$process_data_dir)

  # Directory for variable importance
  file_paths$vimp_dir <- normalizePath(
    file.path(experiment_dir, "variable_importance"),
    mustWork = FALSE
  )
  if (!dir.exists(file_paths$vimp_dir)) dir.create(file_paths$vimp_dir)

  # Directory and files for model building
  file_paths$mb_dir <- normalizePath(
    file.path(experiment_dir, "trained_models"),
    mustWork = FALSE
  )
  if (!dir.exists(file_paths$mb_dir)) dir.create(file_paths$mb_dir)

  # Directory for familiarData objects
  file_paths$fam_data_dir <- normalizePath(
    file.path(experiment_dir, "familiar_data"),
    mustWork = FALSE
  )
  if (!dir.exists(file_paths$fam_data_dir)) dir.create(file_paths$fam_data_dir)

  # Directory for familiarDataCollection objects
  file_paths$fam_coll_dir <- normalizePath(
    file.path(experiment_dir, "familiar_collections"),
    mustWork = FALSE
  )
  if (!dir.exists(file_paths$fam_coll_dir)) dir.create(file_paths$fam_coll_dir)

  # Create results directory
  file_paths$results_dir <- normalizePath(
    file.path(experiment_dir, "results"),
    mustWork = FALSE
  )
  if (!dir.exists(file_paths$results_dir)) dir.create(file_paths$results_dir)

  # Data file location
  data_file <- .parse_arg(
    x_config = config$paths$data_file,
    x_var = data_file,
    var_name = "data_file",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(data_file)) {
    # Iterate over data file(s) to determine the full path
    file_paths$data <- sapply(
      data_file,
      function(curr_data_file, project_dir) {
        # Check if the current file is represented using an absolute path
        if (file.exists(curr_data_file)) {
          return(curr_data_file)
          
        } else if (file.exists(file.path(project_dir, curr_data_file))) {
          return(file.path(project_dir, curr_data_file))
          
        } else {
          ..error(paste(
            "Data could not be found. A file was expected at:",
            curr_data_file, "or at", file.path(project_dir, curr_data_file)
          ))
        }
      },
      project_dir = project_dir
    )
  }
  
  if (file_paths$is_temporary) {
    logger_message(
      paste0(
        "Configuration: A temporary R directory is created for the analysis: ",
        temporary_directory
      ),
      verbose = verbose
    )
  }

  return(file_paths)
}



#' Internal function for parsing settings required to parse the input data and
#' define the experiment
#'
#' This function parses settings required to parse the data set, e.g. determine
#' which columns are identfier columns, what column contains outcome data, which
#' type of outcome is it?
#'
#' @param config A list of settings, e.g. from an xml file.
#' @inheritDotParams .parse_experiment_settings -config
#' @inheritDotParams .parse_setup_settings -config
#'
#' @details Three variants of parameters exist:
#'
#' * required: this parameter is required and must be set by the user.
#'
#' * recommended: not setting this parameter might cause an error to be thrown,
#'   dependent on other input.
#'
#' * optional: these parameters have default values that may be altered if
#'   required.
#'
#' @return A list of settings to be used for configuring the experiments.
#' @md
#' @keywords internal
.parse_initial_settings <- function(config = NULL, ...) {
  # Generate settings list
  settings <- list()

  # Experiment and data settings
  settings$data <- do.call(
    .parse_experiment_settings,
    args = c(
      list("config" = config$data),
      list(...)
    )
  )
  
  # Computational setup settings
  settings$run <- do.call_strict(
    .parse_setup_settings,
    args = c(
      list("config" = config$run),
      list(...)
    )
  )

  return(settings)
}


#' Internal function for parsing settings that configure various aspects of the
#' worklow
#'
#' @param settings List of settings that was previously generated by
#'   `.parse_initial_settings`.
#' @param config A list of settings, e.g. from an xml file.
#' @param data Data set as loaded using the `.load_data` function.
#' @inheritDotParams .parse_setup_settings -config
#' @inheritDotParams .parse_preprocessing_settings -data -config -parallel -outcome_type
#' @inheritDotParams .parse_variable_importance_settings -data -config -parallel -outcome_type
#' @inheritDotParams .parse_model_development_settings -data -config -parallel -outcome_type
#' @inheritDotParams .parse_hyperparameter_optimisation_settings -config -parallel -outcome_type
#' @inheritDotParams .parse_evaluation_settings -config -data -parallel -outcome_type -hpo_metric -development_batch_id -vimp_aggregation_rank_threshold -vimp_aggregation_method -prep_cluster_method -prep_cluster_linkage_method -prep_cluster_cut_method -prep_cluster_similarity_threshold -prep_cluster_similarity_metric
#'
#' @return A list of settings to be used within the workflow
#'
#' @references
#'   1. Shrout, P. E. & Fleiss, J. L. Intraclass correlations: uses in assessing
#'   rater reliability. Psychol. Bull. 86, 420–428 (1979).
#'
#'   1. Koo, T. K. & Li, M. Y. A guideline of selecting and reporting intraclass
#'   correlation coefficients for reliability research. J. Chiropr. Med. 15,
#'   155–163 (2016).
#'
#'   1. Yeo, I. & Johnson, R. A. A new family of power transformations to
#'   improve normality or symmetry. Biometrika 87, 954–959 (2000).
#'
#'   1. Box, G. E. P. & Cox, D. R. An analysis of transformations. J. R. Stat.
#'   Soc. Series B Stat. Methodol. 26, 211–252 (1964).
#'
#'   1. Raymaekers, J., Rousseeuw,  P. J. Transforming variables to central
#'   normality. Mach Learn. (2021).
#'
#'   1. Zwanenburg, A., & Löck, S. (2026). Location and scale-invariant power
#'   transformations for transforming data to normality. Machine Learning, 
#'   115(3), 34.
#'  
#'   1. Park, M. Y., Hastie, T. & Tibshirani, R. Averaged gene expressions for
#'   regression. Biostatistics 8, 212–227 (2007).
#'
#'   1. Tolosi, L. & Lengauer, T. Classification with correlated features:
#'   unreliability of feature ranking and solutions. Bioinformatics 27,
#'   1986–1994 (2011).
#'
#'   1. Johnson, W. E., Li, C. & Rabinovic, A. Adjusting batch effects in
#'   microarray expression data using empirical Bayes methods. Biostatistics 8,
#'   118–127 (2007)
#'
#'   1. Kaufman, L. & Rousseeuw, P. J. Finding groups in data: an introduction
#'   to cluster analysis. (John Wiley & Sons, 2009).
#'
#'   1. Muellner, D. fastcluster: fast hierarchical, agglomerative clustering
#'   routines for R and Python. J. Stat. Softw. 53, 1–18 (2013).
#'
#'   1. Rousseeuw, P. J. Silhouettes: A graphical aid to the interpretation and
#'   validation of cluster analysis. J. Comput. Appl. Math. 20, 53–65 (1987).
#'
#'   1. Langfelder, P., Zhang, B. & Horvath, S. Defining clusters from a
#'   hierarchical cluster tree: the Dynamic Tree Cut package for R.
#'   Bioinformatics 24, 719–720 (2008).
#'
#'   1. McFadden, D. Conditional logit analysis of qualitative choice behavior.
#'   in Frontiers in Econometrics (ed. Zarembka, P.) 105–142 (Academic Press,
#'   1974).
#'
#'   1. Cox, D. R. & Snell, E. J. Analysis of binary data. (Chapman and Hall,
#'   1989).
#'
#'   1. Nagelkerke, N. J. D. A note on a general definition of the coefficient
#'   of determination. Biometrika 78, 691–692 (1991).
#'
#'   1. Meinshausen, N. & Buehlmann, P. Stability selection. J. R. Stat. Soc.
#'   Series B Stat. Methodol. 72, 417–473 (2010).
#'
#'   1. Haury, A.-C., Gestraud, P. & Vert, J.-P. The influence of feature
#'   selection methods on accuracy, stability and interpretability of molecular
#'   signatures. PLoS One 6, e28210 (2011).
#'
#'   1. Wald, R., Khoshgoftaar, T. M., Dittman, D., Awada, W. & Napolitano,A. An
#'   extensive comparison of feature ranking aggregation techniques in
#'   bioinformatics. in 2012 IEEE 13th International Conference on Information
#'   Reuse Integration (IRI) 377–384 (2012).
#'
#'   1. Hutter, F., Hoos, H. H. & Leyton-Brown, K. Sequential model-based
#'   optimization for general algorithm configuration. in Learning and
#'   Intelligent Optimization (ed. Coello, C. A. C.) 6683, 507–523 (Springer
#'   Berlin Heidelberg, 2011).
#'
#'   1. Shahriari, B., Swersky, K., Wang, Z., Adams, R. P. & de Freitas, N.
#'   Taking the Human Out of the Loop: A Review of Bayesian Optimization. Proc.
#'   IEEE 104, 148–175 (2016)
#'
#'   1. Srinivas, N., Krause, A., Kakade, S. M. & Seeger, M. W.
#'   Information-Theoretic Regret Bounds for Gaussian Process Optimization in
#'   the Bandit Setting. IEEE Trans. Inf. Theory 58, 3250–3265 (2012)
#'
#'   1. Kaufmann, E., Cappé, O. & Garivier, A. On Bayesian upper confidence
#'   bounds for bandit problems. in Artificial intelligence and statistics
#'   592–600 (2012).
#'
#'   1. Jamieson, K. & Talwalkar, A. Non-stochastic Best Arm Identification and
#'   Hyperparameter Optimization. in Proceedings of the 19th International
#'   Conference on Artificial Intelligence and Statistics (eds. Gretton, A. &
#'   Robert, C. C.) vol. 51 240–248 (PMLR, 2016).
#'
#'   1. Gramacy, R. B. laGP: Large-Scale Spatial Modeling via Local Approximate
#'   Gaussian Processes in R. Journal of Statistical Software 72, 1–46 (2016)
#'
#'   1. Sparapani, R., Spanbauer, C. & McCulloch, R. Nonparametric Machine
#'   Learning and Efficient Computation with Bayesian Additive Regression Trees:
#'   The BART R Package. Journal of Statistical Software 97, 1–66 (2021)
#'
#'   1. Davison, A. C. & Hinkley, D. V. Bootstrap methods and their application.
#'   (Cambridge University Press, 1997).
#'
#'   1. Efron, B. & Hastie, T. Computer Age Statistical Inference. (Cambridge
#'   University Press, 2016).
#'
#'   1. Lausen, B. & Schumacher, M. Maximally Selected Rank Statistics.
#'   Biometrics 48, 73 (1992).
#'
#'   1. Hothorn, T. & Lausen, B. On the exact distribution of maximally selected
#'   rank statistics. Comput. Stat. Data Anal. 43, 121–137 (2003).
#' @md
#' @keywords internal
.parse_general_settings <- function(settings, config = NULL, data, ...) {
  
  if (is.null(settings$run)) {
    # Run may already be set by .parse_initial_settings.
    settings$run <- do.call_strict(
      .parse_setup_settings,
      args = c(
        list("config" = config$run),
        list(...)
      )
    )
  }
  
  # Remove outcome_type, development_batch_id and parallel from ... This
  # prevents an error caused by multiple matching arguments.
  dots <- list(...)
  dots$parallel <- NULL
  dots$outcome_type <- NULL

  # Pre-processing settings
  settings$prep <- do.call_strict(
    .parse_preprocessing_settings,
    args = c(
      list(
        "config" = config$preprocessing,
        "data" = data,
        "parallel" = settings$run$parallel,
        "outcome_type" = settings$data$outcome_type
      ),
      dots
    )
  )

  # Variable importance settings
  settings$vimp <- do.call_strict(
    .parse_variable_importance_settings,
    args = c(
      list(
        "config" = config$variable_importance,
        "data" = data,
        "parallel" = settings$run$parallel,
        "outcome_type" = settings$data$outcome_type
      ),
      dots
    )
  )

  # Model development settings
  settings$mb <- do.call_strict(
    .parse_model_development_settings,
    args = c(
      list(
        "config" = config$model_development,
        "data" = data,
        "parallel" = settings$run$parallel,
        "outcome_type" = settings$data$outcome_type
      ),
      dots
    )
  )

  # Hyperparameter optimisation settings
  settings$hpo <- do.call_strict(
    .parse_hyperparameter_optimisation_settings,
    args = c(
      list(
        "config" = config$hyperparameter_optimisation,
        "parallel" = settings$run$parallel,
        "outcome_type" = settings$data$outcome_type
      ),
      dots
    )
  )

  # Remove development_batch_id, hpo_metric, vimp_aggregation_method and
  # vimp_aggregation_rank_threshold from ... This prevents an error caused by
  # multiple matching arguments.
  dots$development_batch_id <- NULL
  dots$hpo_metric <- NULL
  dots$vimp_aggregation_method <- NULL
  dots$vimp_aggregation_rank_threshold <- NULL

  # Evaluation settings
  settings$eval <- do.call_strict(
    .parse_evaluation_settings,
    args = c(
      list(
        "config" = config$evaluation,
        "data" = data,
        "parallel" = settings$run$parallel,
        "outcome_type" = settings$data$outcome_type,
        "hpo_metric" = settings$hpo$hpo_metric,
        "development_batch_id" = settings$data$train_cohorts,
        "vimp_aggregation_method" = settings$vimp$aggregation,
        "vimp_aggregation_rank_threshold" = settings$vimp$aggr_rank_threshold,
        "prep_cluster_method" = settings$prep$cluster_method,
        "prep_cluster_linkage_method" = settings$prep$cluster_linkage,
        "prep_cluster_cut_method" = settings$prep$cluster_cut_method,
        "prep_cluster_similarity_threshold" = settings$prep$cluster_similarity_threshold,
        "prep_cluster_similarity_metric" = settings$prep$cluster_similarity_metric
      ),
      dots
    )
  )

  # Set the general parallel switch to FALSE if all workflow steps disabled
  # parallel processing.
  settings$run$parallel <- settings$prep$do_parallel ||
    settings$vimp$do_parallel ||
    settings$mb$do_parallel ||
    settings$hpo$do_parallel %in% c("TRUE", "inner", "outer") ||
    settings$eval$do_parallel %in% c("TRUE", "inner", "outer")

  # Report any issues with missing packages that have been written to the
  # backend.
  .report_missing_package_messages()

  return(settings)
}


#' Internal function for parsing settings related to the experimental setup
#'
#' @param config A list of settings, e.g. from an xml file.
#' @param batch_id_column (**recommended**) Name of the column containing batch
#'   or cohort identifiers. This parameter is required if more than one dataset
#'   is provided, or if external validation is performed.
#'
#'   In familiar any row of data is organised by four identifiers:
#'
#'   * The batch identifier `batch_id_column`: This denotes the group to which a
#'   set of samples belongs, e.g. patients from a single study, samples measured
#'   in a batch, etc. The batch identifier is used for batch normalisation, as
#'   well as selection of development and validation datasets.
#'
#'   * The sample identifier `sample_id_column`: This denotes the sample level,
#'   e.g. data from a single individual. Subsets of data, e.g. bootstraps or
#'   cross-validation folds, are created at this level.
#'
#'   * The series identifier `series_id_column`: Indicates measurements on a
#'   single sample that may not share the same outcome value, e.g. a time
#'   series, or the number of cells in a view.
#'
#'   * The repetition identifier: Indicates repeated measurements in a single
#'   series where any feature values may differ, but the outcome does not.
#'   Repetition identifiers are always implicitly set when multiple entries for
#'   the same series of the same sample in the same batch that share the same
#'   outcome are encountered.
#'
#' @param sample_id_column (**recommended**) Name of the column containing
#'   sample or subject identifiers. See `batch_id_column` above for more
#'   details.
#'
#'   If unset, every row will be identified as a single sample.
#' @param series_id_column (*optional*) Name of the column containing series
#'   identifiers, which distinguish between measurements that are part of a
#'   series for a single sample. See `batch_id_column` above for more details.
#'
#'   If unset, rows which share the same batch and sample identifiers but have a
#'   different outcome are assigned unique series identifiers.
#' @param development_batch_id (*optional*) One or more batch or cohort
#'   identifiers to constitute data sets for development. Defaults to all, or
#'   all minus the identifiers in `validation_batch_id` for external validation.
#'   Required if external validation is performed and `validation_batch_id` is
#'   not provided.
#' @param validation_batch_id (*optional*) One or more batch or cohort
#'   identifiers to constitute data sets for external validation. Defaults to
#'   all data sets except those in `development_batch_id` for external
#'   validation, or none if not. Required if `development_batch_id` is not
#'   provided.
#'
#' @param outcome_name (*optional*) Name of the modelled outcome. This name will
#'   be used in figures created by `familiar`.
#'
#'   If not set, the column name in `outcome_column` will be used for
#'   `binomial`, `multinomial`, and `continuous` outcomes. For other outcomes
#'   (`survival` and `competing_risk`) no default is used.
#'
#' @param outcome_column (**recommended**) Name of the column containing the
#'   outcome of interest. May be identified from a formula, if a formula is
#'   provided as an argument. Otherwise an error is raised. Note that `survival`
#'   and `competing_risk` outcome type outcomes require two columns that
#'   indicate the time-to-event or the time of last follow-up and the event
#'   status.
#'
#' @param outcome_type (**recommended**) Type of outcome found in the outcome
#'   column. The outcome type determines many aspects of the overall process,
#'   e.g. the available variable importance methods and learners, but also the
#'   type of assessments that can be conducted to evaluate the resulting models.
#'   Implemented outcome types are:
#'
#'   * `binomial`: categorical outcome with 2 levels.
#'
#'   * `multinomial`: categorical outcome with 2 or more levels.
#'
#'   * `continuous`: general continuous numeric outcomes.
#'
#'   * `survival`: survival outcome for time-to-event data.
#'
#'   If not provided, the algorithm will attempt to obtain outcome_type from
#'   contents of the outcome column. This may lead to unexpected results, and we
#'   therefore advise to provide this information manually.
#'
#'   Note that `competing_risk` survival analysis are not fully supported, and
#'   is currently not a valid choice for `outcome_type`. The `count` outcome
#'   type was deprecated in version 2.0.0, and superseded by `continuous`.
#'
#' @param class_levels (*optional*) Class levels for `binomial` or `multinomial`
#'   outcomes. This argument can be used to specify the ordering of levels for
#'   categorical outcomes. These class levels must exactly match the levels
#'   present in the outcome column.
#'
#' @param event_indicator (**recommended**) Indicator for events in `survival`
#'   and `competing_risk` analyses. `familiar` will automatically recognise `1`,
#'   `true`, `t`, `y` and `yes` as event indicators, including different
#'   capitalisations. If this parameter is set, it replaces the default values.
#'
#' @param censoring_indicator (**recommended**) Indicator for right-censoring in
#'   `survival` and `competing_risk` analyses. `familiar` will automatically
#'   recognise `0`, `false`, `f`, `n`, `no` as censoring indicators, including
#'   different capitalisations. If this parameter is set, it replaces the
#'   default values.
#'
#' @param competing_risk_indicator (**recommended**) Indicator for competing
#'   risks in `competing_risk` analyses. There are no default values, and if
#'   unset, all values other than those specified by the `event_indicator` and
#'   `censoring_indicator` parameters are considered to indicate competing
#'   risks.
#'
#' @param signature (*optional*) One or more names of feature columns that are
#'   considered part of a specific signature. Features specified here will
#'   always be used for modelling. Ranking of variable importances has no effect
#'   for these features.
#'
#' @param novelty_features (*optional*) One or more names of feature columns
#'   that should be included for the purpose of novelty detection.
#'
#' @param exclude_features (*optional*) Feature columns that will be removed
#'   from the data set. Cannot overlap with features in `signature`,
#'   `novelty_features` or `include_features`.
#'
#' @param include_features (*optional*) Feature columns that are specifically
#'   included in the data set. By default all features are included. Cannot
#'   overlap with `exclude_features`, but may overlap `signature`. Features in
#'   `signature` and `novelty_features` are always included. If both
#'   `exclude_features` and `include_features` are provided, `include_features`
#'   takes precedence, provided that there is no overlap between the two.
#' @param reference_method (*optional*) Method used to set reference levels for
#'   categorical features. There are several options:
#'
#'   * `auto` (default): Categorical features that are not explicitly set by the
#'   user, i.e. columns containing boolean values or characters, use the most
#'   frequent level as reference. Categorical features that are explicitly set,
#'   i.e. as factors, are used as is.
#'
#'   * `always`: Both automatically detected and user-specified categorical
#'   features have the reference level set to the most frequent level. Ordinal
#'   features are not altered, but are used as is.
#'
#'   * `never`: User-specified categorical features are used as is.
#'   Automatically detected categorical features are simply sorted, and the
#'   first level is then used as the reference level. This was the behaviour
#'   prior to familiar version 1.3.0.
#'
#' @param experimental_design (**required**) Defines what the experiment looks
#'   like, e.g. `cv(bt(fs,20)+mb,3,2)+ev` for 2 times repeated 3-fold
#'   cross-validation with nested variable importance computation on 20
#'   bootstraps and model-building, and external validation. The basic workflow
#'   components are:
#'
#'   * `fs`: (optional) variable importance computation step. If not explicitly
#'      declared, feature selection will be done just in time for hyperparameter
#'      optimisation.
#'
#'   * `mb`: (required) model building step.
#'
#'   * `ev`: (optional) external validation. Note that internal validation due
#'   to subsampling will always be conducted if the subsampling methods create
#'   any validation data sets.
#'
#'   The different components are linked using `+`.
#'
#'   Different subsampling methods can be used in conjunction with the basic
#'   workflow components:
#'
#'   * `bs(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. In contrast to `bt`, feature pre-processing parameters and
#'   hyperparameter optimisation are conducted on individual bootstraps.
#'
#'   * `bt(x,n)`: (stratified) .632 bootstrap, with `n` the number of
#'   bootstraps. Unlike `bs` and other subsampling methods, no separate
#'   pre-processing parameters or optimised hyperparameters will be determined
#'   for each bootstrap.
#'
#'   * `cv(x,n,p)`: (stratified) `n`-fold cross-validation, repeated `p` times.
#'   Pre-processing parameters are determined for each iteration.
#'
#'   * `lv(x)`: leave-one-out-cross-validation. Pre-processing parameters are
#'   determined for each iteration.
#'
#'   * `ip(x)`: imbalance partitioning for addressing class imbalances on the
#'   data set. Pre-processing parameters are determined for each partition. The
#'   number of partitions generated depends on the imbalance correction method
#'   (see the `imbalance_correction_method` parameter). Imbalance partitioning
#'   does not generate validation sets.
#'
#'   As shown in the example above, sampling algorithms can be nested.
#'
#'   The simplest valid experimental design is `fs+mb`, which corresponds to a
#'   TRIPOD type 1a analysis. Type 1b analyses are only possible using
#'   bootstraps, e.g. `bt(fs+mb,100)`. Type 2a analyses can be conducted using
#'   cross-validation, e.g. `cv(bt(fs,100)+mb,10,1)`. Depending on the origin of
#'   the external validation data, designs such as `fs+mb+ev` or
#'   `cv(bt(fs,100)+mb,10,1)+ev` constitute type 2b or type 3 analyses. Type 4
#'   analyses can be done by obtaining one or more `familiarModel` objects from
#'   others and applying them to your own data set.
#'
#'   Alternatively, the `experimental_design` parameter may be used to provide a
#'   path to a file containing iterations, which is named `####_iterations.RDS`
#'   by convention. This path can be relative to the directory of the current
#'   experiment (`experiment_dir`), or an absolute path. The absolute path may
#'   thus also point to a file from a different experiment.
#'
#' @param imbalance_correction_method (*optional*) Type of method used to
#'   address class imbalances. Available options are:
#'
#'   * `full_undersampling` (default): All data will be used in an ensemble
#'   fashion. The full minority class will appear in each partition, but
#'   majority classes are undersampled until all data have been used.
#'
#'   * `random_undersampling`: Randomly undersamples majority classes. This is
#'   useful in cases where full undersampling would lead to the formation of
#'   many models due major overrepresentation of the largest class.
#'
#'   This parameter is only used in combination with imbalance partitioning in
#'   the experimental design, and `ip` should therefore appear in the string
#'   that defines the design.
#' @param imbalance_n_partitions (*optional*) Number of times random
#'   undersampling should be repeated. 10 undersampled subsets with balanced
#'   classes are formed by default.
#' @param iteration_seed (*optional*) Integer seed used in sampling algorithms
#'   specified by the `experimental_design` argument. This allows for creating
#'   the same sample assignments across different experiments -- of course
#'   provided that the same dataset is used. By default a random seed is used.
#' @param ... Unused arguments.
#'
#' @return List of parameters related to data parsing and the experiment.
#' @md
#' @keywords internal
.parse_experiment_settings <- function(
    config = NULL,
    batch_id_column = waiver(),
    sample_id_column = waiver(),
    series_id_column = waiver(),
    development_batch_id = waiver(),
    validation_batch_id = waiver(),
    outcome_name = waiver(),
    outcome_column = waiver(),
    outcome_type = waiver(),
    event_indicator = waiver(),
    censoring_indicator = waiver(),
    competing_risk_indicator = waiver(),
    class_levels = waiver(),
    signature = waiver(),
    novelty_features = waiver(),
    exclude_features = waiver(),
    include_features = waiver(),
    reference_method = waiver(),
    experimental_design = waiver(),
    imbalance_correction_method = waiver(),
    imbalance_n_partitions = waiver(),
    iteration_seed = waiver(),
    ...
) {
  settings <- list()

  # experimental_design --------------------------------------------------------
  # Experimental design
  settings$exp_design <- .parse_arg(
    x_config = config$experimental_design,
    x_var = experimental_design,
    var_name = "experimental_design",
    optional = FALSE,
    type = "character"
  )

  # imbalance_method -----------------------------------------------------------
  # Class imbalance correction method
  settings$imbalance_method <- .parse_arg(
    x_config = config$imbalance_correction_method,
    x_var = imbalance_correction_method,
    var_name = "imbalance_correction_method",
    type = "character",
    optional = TRUE,
    default = "full_undersampling"
  )

  .check_parameter_value_is_valid(
    x = settings$imbalance_method,
    var_name = "imbalance_correction_method",
    values = c("full_undersampling", "random_undersampling")
  )

  # imbalance_n_partitions -----------------------------------------------------
  # Number of imbalance partitions for random undersampling
  settings$imbalance_n_partitions <- .parse_arg(
    x_config = config$imbalance_n_partitions,
    x_var = imbalance_n_partitions,
    var_name = "imbalance_n_partitions",
    type = "integer",
    optional = TRUE,
    default = 10L
  )

  .check_number_in_valid_range(
    x = settings$imbalance_n_partitions,
    var_name = "imbalance_n_partitions",
    range = c(1L, Inf)
  )

  # iteration_seed -------------------------------------------------------------
  # Seed for sampling algorithms.
  settings$iteration_seed <- .parse_arg(
    x_config = config$iteration_seed,
    x_var = iteration_seed,
    var_name = "iteration_seed",
    type = "integer",
    optional = TRUE,
    default = NULL
  )
  
  
  # sample_id_column -----------------------------------------------------------
  # Sample identifier column
  settings$sample_col <- .parse_arg(
    x_config = config$sample_id_column,
    x_var = sample_id_column,
    var_name = "sample_id_colum",
    type = "character",
    optional = TRUE,
    default = NULL
  )

  # Update column name
  if (!is.null(settings$sample_col)) {
    settings$sample_col <- .replace_illegal_column_name(settings$sample_col)
  }

  # batch_id_column ------------------------------------------------------------
  # Batch identifier column
  settings$batch_col <- .parse_arg(
    x_config = config$batch_id_column,
    x_var = batch_id_column,
    var_name = "batch_id_column",
    type = "character",
    optional = TRUE,
    default = NULL
  )

  # Update column name
  if (!is.null(settings$batch_col)) {
    settings$batch_col <- .replace_illegal_column_name(settings$batch_col)
  }

  # series_id_column -----------------------------------------------------------
  # Series identifier column
  settings$series_col <- .parse_arg(
    x_config = config$series_id_column,
    x_var = series_id_column,
    var_name = "series_id_column",
    type = "character",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$series_col)) {
    settings$series_col <- .replace_illegal_column_name(settings$series_col)
  }

  # development_batch_id -------------------------------------------------------
  # Development cohort identifier
  settings$train_cohorts <- .parse_arg(
    x_config = config$development_batch_id,
    x_var = development_batch_id,
    var_name = "development_batch_id",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  # validation_batch_id --------------------------------------------------------
  # Validation cohort identifier
  settings$valid_cohorts <- .parse_arg(
    x_config = config$validation_batch_id,
    x_var = validation_batch_id,
    var_name = "validation_batch_id",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  # outcome_column -------------------------------------------------------------
  # Outcome column(s)
  settings$outcome_col <- .parse_arg(
    x_config = config$outcome_column,
    x_var = outcome_column,
    var_name = "outcome_column",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$outcome_col)) {
    settings$outcome_col <- .replace_illegal_column_name(settings$outcome_col)
  }

  # outcome_type ---------------------------------------------------------------
  # Outcome type - a check will be done later
  settings$outcome_type <- .parse_arg(
    x_config = config$outcome_type,
    x_var = outcome_type,
    var_name = "outcome_type",
    type = "character",
    optional = TRUE,
    default = NULL
  )

  # outcome_name ---------------------------------------------------------------
  # Outcome name
  settings$outcome_name <- .parse_arg(
    x_config = config$outcome_name,
    x_var = outcome_name,
    var_name = "outcome_name",
    type = "character",
    optional = TRUE,
    default = NULL
  )

  # class_levels ---------------------------------------------------------------
  # Class levels
  settings$class_levels <- .parse_arg(
    x_config = config$class_levels,
    x_var = class_levels,
    var_name = "class_levels",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  # event_indicator ------------------------------------------------------------
  # Event indicator
  settings$event_indicator <- .parse_arg(
    x_config = config$event_indicator,
    x_var = event_indicator,
    var_name = "event_indicator",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  # censoring_indicator --------------------------------------------------------
  # Censoring indicator
  settings$censoring_indicator <- .parse_arg(
    x_config = config$censoring_indicator,
    x_var = censoring_indicator,
    var_name = "censoring_indicator",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  # competing_risk_indicator ---------------------------------------------------
  # Competing risk indicator
  settings$competing_risk_indicator <- .parse_arg(
    x_config = config$competing_risk_indicator,
    x_var = competing_risk_indicator,
    var_name = "competing_risk_indicator",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  # signature ------------------------------------------------------------------
  # Signature features
  settings$signature <- .parse_arg(
    x_config = config$signature,
    x_var = signature,
    var_name = "signature",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$signature)) {
    settings$signature <- .replace_illegal_column_name(settings$signature)
  }

  # novelty_features -----------------------------------------------------------
  # Novelty features
  settings$novelty_features <- .parse_arg(
    x_config = config$novelty_features,
    x_var = novelty_features,
    var_name = "novelty_features",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$novelty_features)) {
    settings$novelty_features <- .replace_illegal_column_name(settings$novelty_features)
  }

  # include_features -----------------------------------------------------------
  # Included features
  settings$include_features <- .parse_arg(
    x_config = config$include_features,
    x_var = include_features,
    var_name = "include_features",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$include_features)) {
    settings$include_features <- .replace_illegal_column_name(settings$include_features)
  }

  # exclude_features -----------------------------------------------------------
  # Excluded features
  settings$exclude_features <- .parse_arg(
    x_config = config$exclude_features,
    x_var = exclude_features,
    var_name = "exclude_features",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$exclude_features)) {
    settings$exclude_features <- .replace_illegal_column_name(settings$exclude_features)
  }


  # reference_method -----------------------------------------------------------
  # Determine how reference levels for categorical features are set.
  settings$reference_method <- .parse_arg(
    x_config = config$reference_method,
    x_var = reference_method,
    var_name = "reference_method",
    type = "character",
    optional = TRUE,
    default = "auto"
  )

  .check_parameter_value_is_valid(
    x = settings$reference_method,
    var_name = "reference_method",
    values = c("auto", "always", "never")
  )

  return(settings)
}



#' Internal function for parsing settings related to the computational setup
#'
#' @param config A list of settings, e.g. from an xml file.
#' @param parallel (*optional*) Enable parallel processing. Defaults to `TRUE`.
#'   When set to `FALSE`, this disables all parallel processing, regardless of
#'   specific parameters such as `parallel_preprocessing`. However, when
#'   `parallel` is `TRUE`, parallel processing of different parts of the
#'   workflow can be disabled by setting respective flags to `FALSE`.
#' @param parallel_nr_cores (*optional*) Number of cores available for
#'   parallelisation. Defaults to 2. This setting does nothing if
#'   parallelisation is disabled.
#' @param restart_cluster (*optional*) Restart nodes used for parallel computing
#'   to free up memory prior to starting a parallel process. Note that it does
#'   take time to set up the clusters. Therefore setting this argument to `TRUE`
#'   may impact processing speed. This argument is ignored if `parallel` is
#'   `FALSE` or the cluster was initialised outside of familiar. Default is
#'   `FALSE`, which causes the clusters to be initialised only once.
#' @param cluster_type (*optional*) Selection of the cluster type for parallel
#'   processing. Available types are the ones supported by the parallel package
#'   that is part of the base R distribution: `psock` (default), `fork`, `mpi`,
#'   `nws`, `sock`. In addition, `none` is available, which also disables
#'   parallel processing.
#' @param backend_type (*optional*) Selection of the backend for distributing
#'   copies of the data. This backend ensures that only a single master copy is
#'   kept in memory. This limits memory usage during parallel processing.
#'
#'   Several backend options are available, notably `socket_server`, and `none`
#'   (default). `socket_server` is based on the callr package and R sockets,
#'   comes with `familiar` and is available for any OS. `none` uses the package
#'   environment of familiar to store data, and is available for any OS.
#'   However, `none` requires copying of data to any parallel process, and has a
#'   larger memory footprint.
#' @param server_port (*optional*) Integer indicating the port on which the
#'   socket server or RServe process should communicate. Defaults to port 6311.
#'   Note that ports 0 to 1024 and 49152 to 65535 cannot be used.
#' @param ... Unused arguments.
#'
#' @return List of parameters related to the computational setup.
#' @md
#' @keywords internal
.parse_setup_settings <- function(
    config = NULL,
    parallel = waiver(),
    parallel_nr_cores = waiver(),
    restart_cluster = waiver(),
    cluster_type = waiver(),
    backend_type = waiver(),
    server_port = waiver(),
    ...
) {
  settings <- list()

  # parallel -------------------------------------------------------------------
  # Parallelisation master switch
  settings$parallel <- .parse_arg(
    x_config = config$parallel,
    x_var = parallel,
    var_name = "parallel",
    type = "logical",
    optional = TRUE,
    default = TRUE
  )

  # parallel_nr_cores ----------------------------------------------------------
  # Maximum number of cores that a R may use
  settings$parallel_nr_cores <- .parse_arg(
    x_config = config$parallel_nr_cores,
    x_var = parallel_nr_cores,
    var_name = "parallel_nr_cores",
    type = "integer",
    optional = TRUE,
    default = 2L
  )

  if (!is.null(settings$parallel_nr_cores)) {
    .check_number_in_valid_range(
      x = settings$parallel_nr_cores,
      var_name = "parallel_nr_cores",
      range = c(1L, parallel::detectCores())
    )
  }

  # Set cores to 1 in case parallel processing is disabled.
  if (!settings$parallel) settings$parallel_nr_cores <- 1L

  # restart_cluster ------------------------------------------------------------
  # Restart clusters
  settings$restart_cluster <- .parse_arg(
    x_config = config$restart_cluster,
    x_var = restart_cluster,
    var_name = "restart_cluster",
    type = "logical",
    optional = TRUE,
    default = FALSE
  )

  if (!settings$parallel) settings$restart_cluster <- FALSE

  # cluster_type ---------------------------------------------------------------
  # Define the cluster type
  settings$cluster_type <- .parse_arg(
    x_config = config$cluster_type,
    x_var = cluster_type,
    var_name = "cluster_type",
    type = "character",
    optional = TRUE,
    default = "psock"
  )

  .check_parameter_value_is_valid(
    x = settings$cluster_type,
    var_name = "cluster_type",
    values = c("psock", "fork", "mpi", "nws", "sock", "none")
  )

  if (!settings$parallel) settings$cluster_type <- "none"

  .check_cluster_type_availability(cluster_type = settings$cluster_type)

  if (settings$cluster_type != "none") {
    require_package(
      x = "microbenchmark",
      purpose = "to make use of optimised parallel processing",
      message_type = "backend_warning"
    )
  }

  # backend_type ---------------------------------------------------------------
  # Data server backend - this is os- and package-dependent
  settings$backend_type <- .parse_arg(
    x_config = config$backend_type,
    x_var = backend_type,
    var_name = "backend_type",
    type = "character",
    optional = TRUE,
    default = "none"
  )

  .check_parameter_value_is_valid(
    x = settings$backend_type,
    var_name = "backend_type",
    values = .get_available_backend_types()
  )

  require_package(
    x = .required_packages_backend(settings$backend_type),
    purpose = "to use the requested backend (", settings$backend_type, ")",
    message_type = "backend_error"
  )

  # server_port ----------------------------------------------------------------
  # RServe communications port
  settings$server_port <- .parse_arg(
    x_config = config$server_port,
    x_var = server_port,
    var_name = "server_port",
    type = "integer",
    optional = TRUE,
    default = 6311L
  )

  .check_number_in_valid_range(
    x = settings$server_port,
    var_name = "server_port",
    range = c(1025L, 49151L)
  )

  return(settings)
}



#' Internal function for parsing settings related to preprocessing
#'
#' @param config A list of settings, e.g. from an xml file.
#' @param data Data set as loaded using the `.load_data` function.
#' @param parallel Logical value that whether familiar uses parallelisation. If
#'   `FALSE` it will override `parallel_preprocessing`.
#' @param outcome_type Type of outcome found in the data set.
#' @param feature_max_fraction_missing (*optional*) Numeric value between `0.0`
#'   and `0.95` that determines the meximum fraction of missing values that
#'   still allows a feature to be included in the data set. All features with a
#'   missing value fraction over this threshold are not processed further. The
#'   default value is `0.30`.
#' @param sample_max_fraction_missing (*optional*) Numeric value between `0.0`
#'   and `0.95` that determines the maximum fraction of missing values that
#'   still allows a sample to be included in the data set. All samples with a
#'   missing value fraction over this threshold are excluded and not processed
#'   further. The default value is `0.30`.
#' @param filter_method (*optional*) One or methods used to reduce
#'   dimensionality of the data set by removing irrelevant or poorly
#'   reproducible features.
#'
#'   Several method are available:
#'
#'   * `none` (default): None of the features will be filtered.
#'
#'   * `low_variance`: Features with a variance below the
#'   `low_var_minimum_variance_threshold` are filtered. This can be useful to
#'   filter, for example, genes that are not differentially expressed.
#'
#'   * `univariate_test`: Features undergo a univariate regression using an
#'   outcome-appropriate regression model. The p-value of the model coefficient
#'   is collected. Features with coefficient p-value above the
#'   `univariate_test_threshold` are subsequently filtered.
#'
#'   * `robustness`: Features that are not sufficiently robust according to the
#'   intraclass correlation coefficient are filtered. Use of this method
#'   requires that repeated measurements are present in the data set, i.e. there
#'   should be entries for which the sample and cohort identifiers are the same.
#'
#'   More than one method can be used simultaneously. Features with singular
#'   values are always filtered, as these do not contain information.
#' @param univariate_test_threshold (*optional*) Numeric value between `1.0` and
#'   `0.0` that determines which features are irrelevant and will be filtered by
#'   the `univariate_test`. p-values are compared to this threshold.
#'   All features with values above the threshold are filtered. The default
#'   value is `0.20`.
#' @param univariate_test_threshold_metric (*optional*) Metric used with the to
#'   compare the `univariate_test_threshold` against. The following metric can
#'   be chosen:
#'
#'   * `p_value` (default): The unadjusted p-value of each feature is used for
#'   to filter features.
#'
#' @param univariate_test_max_feature_set_size (*optional*) Maximum size of the
#'   feature set after the univariate test. P or q values of features are
#'   compared against the threshold, but if the resulting data set would be
#'   larger than this setting, only the most relevant features up to the desired
#'   feature set size are selected.
#'
#'   The default value is `NULL`, which causes features to be filtered based on
#'   their relevance only.
#' @param low_var_minimum_variance_threshold (required, if used) Numeric value
#'   that determines which features will be filtered by the `low_variance`
#'   method. The variance of each feature is computed and compared to the
#'   threshold. If it is below the threshold, the feature is removed.
#'
#'   This parameter has no default value and should be set if `low_variance` is
#'   used.
#' @param low_var_max_feature_set_size (*optional*) Maximum size of the feature
#'   set after filtering features with a low variance. All features are first
#'   compared against `low_var_minimum_variance_threshold`. If the resulting
#'   feature set would be larger than specified, only the most strongly varying
#'   features will be selected, up to the desired size of the feature set.
#'
#'   The default value is `NULL`, which causes features to be filtered based on
#'   their variance only.
#' @param robustness_icc_type (*optional*) String indicating the type of
#'   intraclass correlation coefficient (`1`, `2` or `3`) that should be used to
#'   compute robustness for features in repeated measurements. These types
#'   correspond to the types in Shrout and Fleiss (1979). The default value is
#'   `1`.
#' @param robustness_threshold_metric (*optional*) String indicating which
#'   specific intraclass correlation coefficient (ICC) metric should be used to
#'   filter features. This should be one of:
#'
#'   * `icc`: The estimated ICC value itself.
#'
#'   * `icc_low` (default): The estimated lower limit of the 95% confidence
#'   interval of the ICC, as suggested by Koo and Li (2016).
#'
#'   * `icc_panel`: The estimated ICC value over the panel average, i.e. the ICC
#'   that would be obtained if all repeated measurements were averaged.
#'
#'   * `icc_panel_low`: The estimated lower limit of the 95% confidence interval
#'   of the panel ICC.
#'
#' @param robustness_threshold_value (*optional*) The intraclass correlation
#'   coefficient value that is as threshold. The default value is `0.70`.
#' @param transformation_method (*optional*) The transformation method used to
#'   change the distribution of the data to be more normal-like. The following
#'   methods are available:
#'
#'   * `none`: This disables transformation of features.
#'
#'   * `yeo_johnson`: Transformation using the location and scale invariant
#'   version of the Yeo-Johnson transformation (Yeo and Johnson, 2000;
#'   Zwanenburg and Löck, 2026).
#'
#'   * `yeo_johnson_robust` (default): A robust version of `yeo_johnson`.
#'   This method is less sensitive to outliers.
#'
#'   * `yeo_johnson_conventional`: As `yeo_johnson`, but without optimisation of
#'   location and scale parameters. This method is equivalent to the original
#'   transformation proposed by Yeo and Johnson (2001).
#'
#'   * `box_cox`: Transformation using the location and scale invariant version
#'   of the Box-Cox transformation (Box and Cox, 1964; Zwanenburg and Löck,
#'   2026).
#'
#'   * `box_cox_robust`: A robust version of `yeo_johnson`. This method is less
#'   sensitive to outliers.
#'
#'   * `box_cox_conventional`: As `box_cox`, but without optimisation of
#'   location and scale parameters. This method is equivalent to the original
#'   transformation proposed by Box and Cox (1964). This method requires
#'   strictly positive feature values.
#'
#'   Transformation requires the `power.transform` package. Only features that
#'   contain numerical data are transformed. Transformation parameters obtained
#'   in development data are stored within `featureInfo` objects for later use
#'   with validation data sets.
#'
#' @param transformation_optimisation_criterion (*optional*) Transformation
#'   parameters are optimised using a criterion, conventionally
#'   maximum-likelihood-estimation. `power.transform` implements multiple
#'   optimisation criteria, of which the following are available:
#'
#'   * `mle` (default): Optimisation using maximum likelihood estimation.
#'
#'   * `cramer_von_mises`: Optimisation using the Cramér-von Mises
#'   criterion. Zwanenburg and Löck (2026) found that this criterion was
#'   relatively robust against outliers.
#'
#' @param transformation_gof_test_p_value (*optional*) Not all transformations
#'   will lead to features that are roughly normally distributed. Zwanenburg and
#'   Löck (2026) established a empirical goodness-of-fit test for central
#'   normality. This parameter sets the significance for rejecting the
#'   null-hypothesis that a feature distribution is centrally normal. When the
#'   null-hypothesis is rejected, no transformation is performed. The default
#'   value is `NULL`, which disables the test.
#'
#' @param normalisation_method (*optional*) The normalisation method used to
#'   improve the comparability between numerical features that may have very
#'   different scales. The following normalisation methods can be chosen:
#'
#'   * `none`: This disables feature normalisation.
#'
#'   * `standardisation`: Features are normalised by subtraction of their mean
#'   values and division by their standard deviations. This causes every feature
#'   to be have a center value of 0.0 and standard deviation of 1.0.
#'
#'   * `standardisation_trim`: As `standardisation`, but based on the set of
#'   feature values where the 5% lowest and 5% highest values are discarded.
#'   This reduces the effect of outliers.
#'
#'   * `standardisation_winsor`: As `standardisation`, but based on the set of
#'   feature values where the 5% lowest and 5% highest values are winsorised.
#'   This reduces the effect of outliers.
#'
#'   * `standardisation_robust` (default): A robust version of `standardisation`
#'   that relies on computing Huber's M-estimators for location and scale.
#'
#'   * `normalisation`: Features are normalised by subtraction of their minimum
#'   values and division by their ranges. This maps all feature values to a
#'   \eqn{[0, 1]} interval.
#'
#'   * `normalisation_trim`: As `normalisation`, but based on the set of feature
#'   values where the 5% lowest and 5% highest values are discarded. This
#'   reduces the effect of outliers.
#'
#'   * `normalisation_winsor`: As `normalisation`, but based on the set of
#'   feature values where the 5% lowest and 5% highest values are winsorised.
#'   This reduces the effect of outliers.
#'
#'   * `quantile`: Features are normalised by subtraction of their median values
#'   and division by their interquartile range.
#'
#'   * `mean_centering`: Features are centered by substracting the mean, but do
#'   not undergo rescaling.
#'
#'   Only features that contain numerical data are normalised. Normalisation
#'   parameters obtained in development data are stored within `featureInfo`
#'   objects for later use with validation data sets.
#' @param batch_normalisation_method (*optional*) The method used for batch
#'   normalisation. Available methods are:
#'
#'   * `none` (default): This disables batch normalisation of features.
#'
#'   * `standardisation`: Features within each batch are normalised by
#'   subtraction of the mean value and division by the standard deviation in
#'   each batch.
#'
#'   * `standardisation_trim`: As `standardisation`, but based on the set of
#'   feature values where the 5% lowest and 5% highest values are discarded.
#'   This reduces the effect of outliers.
#'
#'   * `standardisation_winsor`: As `standardisation`, but based on the set of
#'   feature values where the 5% lowest and 5% highest values are winsorised.
#'   This reduces the effect of outliers.
#'
#'   * `standardisation_robust`: A robust version of `standardisation` that
#'   relies on computing Huber's M-estimators for location and scale within each
#'   batch.
#'
#'   * `normalisation`: Features within each batch are normalised by subtraction
#'   of their minimum values and division by their range in each batch. This
#'   maps all feature values in each batch to a \eqn{[0, 1]} interval.
#'
#'   * `normalisation_trim`: As `normalisation`, but based on the set of feature
#'   values where the 5% lowest and 5% highest values are discarded. This
#'   reduces the effect of outliers.
#'
#'   * `normalisation_winsor`: As `normalisation`, but based on the set of
#'   feature values where the 5% lowest and 5% highest values are winsorised.
#'   This reduces the effect of outliers.
#'
#'   * `quantile`: Features in each batch are normalised by subtraction of the
#'   median value and division by the interquartile range of each batch.
#'
#'   * `mean_centering`: Features in each batch are centered on 0.0 by
#'   substracting the mean value in each batch, but are not rescaled.
#'
#'   * `combat_parametric`: Batch adjustments using parametric empirical Bayes
#'   (Johnson et al, 2007). `combat_p` leads to the same method.
#'
#'   * `combat_non_parametric`: Batch adjustments using non-parametric empirical
#'   Bayes (Johnson et al, 2007). `combat_np` and `combat` lead to the same
#'   method. Note that we reduced complexity from O(\eqn{n^2}) to O(\eqn{n}) by
#'   only computing batch adjustment parameters for each feature on a subset of
#'   50 randomly selected features, instead of all features.
#'
#'   Only features that contain numerical data are normalised using batch
#'   normalisation. Batch normalisation parameters obtained in development data
#'   are stored within `featureInfo` objects for later use with validation data
#'   sets, in case the validation data is from the same batch.
#'
#'   If validation data contains data from unknown batches, normalisation
#'   parameters are separately determined for these batches.
#'
#'   Note that for both empirical Bayes methods, the batch effect is assumed to
#'   produce results across the features. This is often true for things such as
#'   gene expressions, but the assumption may not hold generally.
#'
#'   When performing batch normalisation, it is moreover important to check that
#'   differences between batches or cohorts are not related to the studied
#'   endpoint.
#'
#' @param imputation_method (*optional*) Method used for imputing missing
#'   feature values. Two methods are implemented:
#'
#'   * `simple`: Simple replacement of a missing value by the median value (for
#'   numeric features) or the modal value (for categorical features).
#'
#'   * `lasso`: Imputation of missing value by lasso regression (using `glmnet`)
#'   based on information contained in other features.
#'
#'   `simple` imputation precedes `lasso` imputation to ensure that any missing
#'   values in predictors required for `lasso` regression are resolved. The
#'   `lasso` estimate is then used to replace the missing value.
#'
#'   The default value depends on the number of features in the dataset. If the
#'   number is lower than 100, `lasso` is used by default, and `simple`
#'   otherwise.
#'
#'   Only single imputation is performed. Imputation models and parameters are
#'   stored within `featureInfo` objects for later use with validation data
#'   sets.
#' @param cluster_method (*optional*) Clustering is performed to identify and
#'   replace redundant features, for example those that are highly correlated.
#'   Such features do not carry much additional information and may be removed
#'   or replaced instead (Park et al., 2007; Tolosi and Lengauer, 2011).
#'
#'   The cluster method determines the algorithm used to form the clusters. The
#'   following cluster methods are implemented:
#'
#'   * `none`: No clustering is performed.
#'
#'   * `hclust` (default): Hierarchical agglomerative clustering. If the
#'   `fastcluster` package is installed, `fastcluster::hclust` is used (Muellner
#'   2013), otherwise `stats::hclust` is used.
#'
#'   * `agnes`: Hierarchical clustering using agglomerative nesting (Kaufman and
#'   Rousseeuw, 1990). This algorithm is similar to `hclust`, but uses the
#'   `cluster::agnes` implementation.
#'
#'   * `diana`: Divisive analysis hierarchical clustering. This method uses
#'   divisive instead of agglomerative clustering (Kaufman and Rousseeuw, 1990).
#'   `cluster::diana` is used.
#'
#'   * `pam`: Partioning around medioids. This partitions the data into $k$
#'   clusters around medioids (Kaufman and Rousseeuw, 1990). $k$ is selected
#'   using the `silhouette` metric. `pam` is implemented using the
#'   `cluster::pam` function.
#'
#'   Clusters and cluster information is stored within `featureInfo` objects for
#'   later use with validation data sets. This enables reproduction of the same
#'   clusters as formed in the development data set.
#' @param cluster_linkage_method (*optional*) Linkage method used for
#'   agglomerative clustering in `hclust` and `agnes`. The following linkage
#'   methods can be used:
#'
#'   * `average` (default): Average linkage.
#'
#'   * `single`: Single linkage.
#'
#'   * `complete`: Complete linkage.
#'
#'   * `weighted`: Weighted linkage, also known as McQuitty linkage.
#'
#'   * `ward`: Linkage using Ward's minimum variance method.
#'
#'   `diana` and `pam` do not require a linkage method.
#' @param cluster_cut_method (*optional*) The method used to define the actual
#'   clusters. The following methods can be used:
#'
#'   * `silhouette`: Clusters are formed based on the silhouette score
#'   (Rousseeuw, 1987). The average silhouette score is computed from 2 to
#'   \eqn{n} clusters, with \eqn{n} the number of features. Clusters are only
#'   formed if the average silhouette exceeds 0.50, which indicates reasonable
#'   evidence for structure. This procedure may be slow if the number of
#'   features is large (>100s).
#'
#'   * `fixed_cut`: Clusters are formed by cutting the hierarchical tree at the
#'   point indicated by the `cluster_similarity_threshold`, e.g. where features
#'   in a cluster have an average Spearman correlation of 0.90. `fixed_cut` is
#'   only available for `agnes`, `diana` and `hclust`.
#'
#'   * `dynamic_cut`: Dynamic cluster formation using the cutting algorithm in
#'   the `dynamicTreeCut` package. This package should be installed to select
#'   this option. `dynamic_cut` can only be used with `agnes` and `hclust`.
#'
#'   The default options are `silhouette` for partioning around medioids (`pam`)
#'   and `fixed_cut` otherwise.
#'
#' @param cluster_similarity_metric (*optional*) Clusters are formed based on
#'   feature similarity. All features are compared in a pair-wise fashion to
#'   compute similarity, for example correlation. The resulting similarity grid
#'   is converted into a distance matrix that is subsequently used for
#'   clustering. The following metrics are supported to compute pairwise
#'   similarities:
#'
#'   * `mutual_information` (default): normalised mutual information.
#'
#'   * `mcfadden_r2`: McFadden's pseudo R-squared (McFadden, 1974).
#'
#'   * `cox_snell_r2`: Cox and Snell's pseudo R-squared (Cox and Snell, 1989).
#'
#'   * `nagelkerke_r2`: Nagelkerke's pseudo R-squared (Nagelkerke, 1991).
#'
#'   * `spearman`: Spearman's rank order correlation.
#'
#'   * `kendall`: Kendall rank correlation.
#'
#'   * `pearson`: Pearson product-moment correlation.
#'
#'   The pseudo R-squared metrics can be used to assess similarity between mixed
#'   pairs of numeric and categorical features, as these are based on the
#'   log-likelihood of regression models. In `familiar`, the more informative
#'   feature is used as the predictor and the other feature as the reponse
#'   variable. In numeric-categorical pairs, the numeric feature is considered
#'   to be more informative and is thus used as the predictor. In
#'   categorical-categorical pairs, the feature with most levels is used as the
#'   predictor.
#'
#'   In case any of the classical correlation coefficients (`pearson`,
#'   `spearman` and `kendall`) are used with (mixed) categorical features, the
#'   categorical features are one-hot encoded and the mean correlation over all
#'   resulting pairs is used as similarity.
#'
#' @param cluster_similarity_threshold (*optional*) The threshold level for
#'   pair-wise similarity that is required to form clusters using `fixed_cut`.
#'   This should be a numerical value between 0.0 and 1.0. Note however, that a
#'   reasonable threshold value depends strongly on the similarity metric. The
#'   following are the default values used:
#'
#'   * `mcfadden_r2` and `mutual_information`: `0.30`
#'
#'   * `cox_snell_r2` and `nagelkerke_r2`: `0.75`
#'
#'   * `spearman`, `kendall` and `pearson`: `0.90`
#'
#'   Alternatively, if the `fixed cut` method is not used, this value determines
#'   whether any clustering should be performed, because the data may not
#'   contain highly similar features. The default values in this situation are:
#'
#'   * `mcfadden_r2`  and `mutual_information`: `0.25`
#'
#'   * `cox_snell_r2` and `nagelkerke_r2`: `0.40`
#'
#'   * `spearman`, `kendall` and `pearson`: `0.70`
#'
#'   The threshold value is converted to a distance (1-similarity) prior to
#'   cutting hierarchical trees.
#' @param cluster_representation_method (*optional*) Method used to determine
#'   how the information of co-clustered features is summarised and used to
#'   represent the cluster. The following methods can be selected:
#'
#'   * `best_predictor` (default): The feature with the highest importance
#'   according to univariate regression with the outcome is used to represent
#'   the cluster.
#'
#'   * `medioid`: The feature closest to the cluster center, i.e. the feature
#'   that is most similar to the remaining features in the cluster, is used to
#'   represent the feature.
#'
#'   * `mean`: A meta-feature is generated by averaging the feature values for
#'   all features in a cluster. This method aligns all features so that all
#'   features will be positively correlated prior to averaging. Should a cluster
#'   contain one or more categorical features, the `medioid` method will be used
#'   instead, as averaging is not possible. Note that if this method is chosen,
#'   the `normalisation_method` parameter should be one of `standardisation`,
#'   `standardisation_trim`, `standardisation_winsor` or `quantile`.`
#'
#'   If the `pam` cluster method is selected, only the `medioid` method can be
#'   used. In that case 1 medioid is used by default.
#' @param parallel_preprocessing (*optional*) Enable parallel processing for the
#'   preprocessing workflow. Defaults to `TRUE`. When set to `FALSE`, this will
#'   disable the use of parallel processing while preprocessing, regardless of
#'   the settings of the `parallel` parameter. `parallel_preprocessing` is
#'   ignored if `parallel=FALSE`.
#' @param ... Unused arguments.
#'
#' @return List of parameters related to preprocessing.
#'
#' @references
#'   1. Shrout, P. E. & Fleiss, J. L. Intraclass correlations: uses in assessing
#'   rater reliability. Psychol. Bull. 86, 420–428 (1979).
#'
#'   1. Koo, T. K. & Li, M. Y. A guideline of selecting and reporting intraclass
#'   correlation coefficients for reliability research. J. Chiropr. Med. 15,
#'   155–163 (2016).
#'
#'   1. Yeo, I. & Johnson, R. A. A new family of power transformations to
#'   improve normality or symmetry. Biometrika 87, 954–959 (2000).
#'
#'   1. Box, G. E. P. & Cox, D. R. An analysis of transformations. J. R. Stat.
#'   Soc. Series B Stat. Methodol. 26, 211–252 (1964).
#'
#'   1. Raymaekers, J., Rousseeuw,  P. J. Transforming variables to central
#'   normality. Mach Learn. (2021).
#'
#'   1. Zwanenburg, A., & Löck, S. (2026). Location and scale-invariant power
#'   transformations for transforming data to normality. Machine Learning, 
#'   115(3), 34.
#'
#'   1. Park, M. Y., Hastie, T. & Tibshirani, R. Averaged gene expressions for
#'   regression. Biostatistics 8, 212–227 (2007).
#'
#'   1. Tolosi, L. & Lengauer, T. Classification with correlated features:
#'   unreliability of feature ranking and solutions. Bioinformatics 27,
#'   1986–1994 (2011).
#'
#'   1. Johnson, W. E., Li, C. & Rabinovic, A. Adjusting batch effects in
#'   microarray expression data using empirical Bayes methods. Biostatistics 8,
#'   118–127 (2007)
#'
#'   1. Kaufman, L. & Rousseeuw, P. J. Finding groups in data: an introduction
#'   to cluster analysis. (John Wiley & Sons, 2009).
#'
#'   1. Muellner, D. fastcluster: fast hierarchical, agglomerative clustering
#'   routines for R and Python. J. Stat. Softw. 53, 1–18 (2013).
#'
#'   1. Rousseeuw, P. J. Silhouettes: A graphical aid to the interpretation and
#'   validation of cluster analysis. J. Comput. Appl. Math. 20, 53–65 (1987).
#'
#'   1. Langfelder, P., Zhang, B. & Horvath, S. Defining clusters from a
#'   hierarchical cluster tree: the Dynamic Tree Cut package for R.
#'   Bioinformatics 24, 719–720 (2008).
#'
#'   1. McFadden, D. Conditional logit analysis of qualitative choice behavior.
#'   in Frontiers in Econometrics (ed. Zarembka, P.) 105–142 (Academic Press,
#'   1974).
#'
#'   1. Cox, D. R. & Snell, E. J. Analysis of binary data. (Chapman and Hall,
#'   1989).
#'
#'   1. Nagelkerke, N. J. D. A note on a general definition of the coefficient
#'   of determination. Biometrika 78, 691–692 (1991).
#'
#' @md
#' @keywords internal
.parse_preprocessing_settings <- function(
    config = NULL,
    data,
    parallel,
    outcome_type,
    feature_max_fraction_missing = waiver(),
    sample_max_fraction_missing = waiver(),
    filter_method = waiver(),
    univariate_test_threshold = waiver(),
    univariate_test_threshold_metric = waiver(),
    univariate_test_max_feature_set_size = waiver(),
    low_var_minimum_variance_threshold = waiver(),
    low_var_max_feature_set_size = waiver(),
    robustness_icc_type = waiver(),
    robustness_threshold_metric = waiver(),
    robustness_threshold_value = waiver(),
    transformation_method = waiver(),
    transformation_optimisation_criterion = waiver(),
    transformation_gof_test_p_value = waiver(),
    normalisation_method = waiver(),
    batch_normalisation_method = waiver(),
    imputation_method = waiver(),
    cluster_method = waiver(),
    cluster_linkage_method = waiver(),
    cluster_cut_method = waiver(),
    cluster_similarity_metric = waiver(),
    cluster_similarity_threshold = waiver(),
    cluster_representation_method = waiver(),
    parallel_preprocessing = waiver(),
    ...
) {
  settings <- list()

  # feature_max_fraction_missing -----------------------------------------------
  # Maximum fraction of data points missing for inclusion of a feature
  settings$feature_max_fraction_missing <- .parse_arg(
    x_config = config$feature_max_fraction_missing,
    x_var = feature_max_fraction_missing,
    var_name = "feature_max_fraction_missing",
    type = "numeric",
    optional = TRUE,
    default = 0.30
  )

  .check_number_in_valid_range(
    x = settings$feature_max_fraction_missing,
    var_name = "feature_max_fraction_missing",
    range = c(0.0, 0.95)
  )

  # sample_max_fraction_missing ------------------------------------------------
  # Maximum fraction of features missing for inclusion of a subject
  settings$sample_max_fraction_missing <- .parse_arg(
    x_config = config$sample_max_fraction_missing,
    x_var = sample_max_fraction_missing,
    var_name = "sample_max_fraction_missing",
    type = "numeric",
    optional = TRUE,
    default = 0.30
  )

  .check_number_in_valid_range(
    x = settings$sample_max_fraction_missing,
    var_name = "sample_max_fraction_missing",
    range = c(0.0, 0.95)
  )

  # filter_method --------------------------------------------------------------
  # Univariate filter methods
  settings$filter_method <- .parse_arg(
    x_config = config$filter_method,
    x_var = filter_method,
    var_name = "filter_method",
    type = "character_list",
    optional = TRUE,
    default = "none"
  )

  .check_parameter_value_is_valid(
    x = settings$filter_method,
    var_name = "filter_method",
    values = c("none", "low_variance", "univariate_test", "robustness")
  )

  if (outcome_type == "multinomial" && "univariate_test" %in% settings$filter_method) {
    # The nnet package is required for univariate tests with multinomial
    # endpoints.
    if (!require_package(
      x = "nnet",
      purpose = "to filter features using univariate tests",
      message_type = "backend_warning"
    )) {
      # If the nnet package is not present, avoid the univariate test.
      settings$filter_method <- setdiff(settings$filter_method, "univariate_test")
      if (length(settings$filter_method) == 0L) {
        settings$filter_method <- "none"
      }
    }
  }

  # univariate_test_threshold --------------------------------------------------
  # Univariate model filter threshold value
  settings$univar_threshold <- .parse_arg(
    x_config = config$univariate_test_threshold,
    x_var = univariate_test_threshold,
    var_name = "univariate_test_threshold",
    type = "numeric",
    optional = TRUE,
    default = 0.20
  )

  .check_number_in_valid_range(
    x = settings$univar_threshold,
    var_name = "univariate_test_threshold",
    range = c(0.0, 1.0),
    closed = c(FALSE, TRUE)
  )

  # univariate_test_threshold_metric -------------------------------------------
  # Univariate model threshold metric
  settings$univar_metric <- .parse_arg(
    x_config = config$univariate_test_threshold_metric,
    x_var = univariate_test_threshold_metric,
    var_name = "univariate_test_threshold_metric",
    type = "character",
    optional = TRUE,
    default = "p_value"
  )

  .check_parameter_value_is_valid(
    x = settings$univar_metric,
    var_name = "univariate_test_threshold_metric",
    values = c("p_value", "q_value")
  )

  # If q_value is provided, warn for deprecation and use p-values instead.
  if (settings$univar_metric == "q_value") {
    ..deprecation_qvalue()
    settings$univar_metric <- "p_value"
  }
  
  # univariate_test_max_feature_set_size ---------------------------------------
  # Maximum feature set size after univariate regression models.
  settings$univar_feat_set_size <- .parse_arg(
    x_config = config$univariate_test_max_feature_set_size,
    x_var = univariate_test_max_feature_set_size,
    var_name = "univariate_test_max_feature_set_size",
    type = "integer",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$univar_feat_set_size)) {
    .check_number_in_valid_range(
      x = settings$univar_feat_set_size,
      var_name = "univariate_test_max_feature_set_size",
      range = c(1L, Inf)
    )
  }

  # low_var_minimum_variance_threshold -----------------------------------------
  # Minimum amount of variance for inclusion of feature.
  if ("low_variance" %in% settings$filter_method) {
    settings$low_var_threshold <- .parse_arg(
      x_config = config$low_var_minimum_variance_threshold,
      x_var = low_var_minimum_variance_threshold,
      var_name = "low_var_minimum_variance_threshold",
      type = "numeric",
      optional = FALSE
    )

    .check_number_in_valid_range(
      x = settings$low_var_threshold,
      var_name = "low_var_minimum_variance_threshold",
      range = c(0.0, Inf)
    )
  }

  # low_var_max_feature_set_size -----------------------------------------------
  # Maximum feature set size after variance thresholding
  settings$low_var_max_feature_set_size <- .parse_arg(
    x_config = config$low_var_max_feature_set_size,
    x_var = low_var_max_feature_set_size,
    var_name = "low_var_max_feature_set_size",
    type = "integer",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$low_var_max_feature_set_size)) {
    .check_number_in_valid_range(
      x = settings$low_var_max_feature_set_size,
      var_name = "low_var_max_feature_set_size",
      range = c(1L, Inf)
    )
  }

  # robustness_icc_type --------------------------------------------------------
  # Intraclass correlation coefficient (ICC) type for robustness analysis
  settings$robustness_icc_type <- .parse_arg(
    x_config = config$robustness_icc_type,
    x_var = robustness_icc_type,
    var_name = "robustness_icc_type",
    type = "character",
    optional = TRUE,
    default = "1"
  )

  .check_parameter_value_is_valid(
    x = settings$robustness_icc_type,
    var_name = "robustness_icc_type",
    values = .get_available_icc_types()
  )

  # robustness_threshold_metric ------------------------------------------------
  # ICC parameter to use for thresholding. Can be icc (estimated icc), icc_low
  # (lower edge of the icc confidence interval), icc_panel (estimated panel
  # icc), icc_panel_low (lower edge of the panel icc confidence interval)
  settings$robustness_threshold_param <- .parse_arg(
    x_config = config$robustness_threshold_metric,
    x_var = robustness_threshold_metric,
    var_name = "robustness_threshold_metric",
    type = "character",
    optional = TRUE,
    default = "icc_low"
  )

  .check_parameter_value_is_valid(
    x = settings$robustness_threshold_param,
    var_name = "robustness_threshold_metric",
    values = c("icc", "icc_low", "icc_panel", "icc_panel_low")
  )

  # robustness_threshold_value -------------------------------------------------
  # ICC value for thresholding.
  settings$robustness_threshold_value <- .parse_arg(
    x_config = config$robustness_threshold_value,
    x_var = robustness_threshold_value,
    var_name = "robustness_threshold_value",
    type = "numeric",
    optional = TRUE,
    default = 0.70
  )

  .check_number_in_valid_range(
    x = settings$robustness_threshold_value,
    var_name = "robustness_threshold_value",
    range = c(-Inf, 1.0)
  )

  # imputation_method ----------------------------------------------------------
  # Data imputation method. For datasets smaller than 100 features we use lasso,
  # and simple imputation is used otherwise.
  default_imputation_method <- ifelse(
    get_n_features(data, outcome_type = outcome_type) < 100L,
    "lasso", "simple"
  )

  settings$imputation_method <- .parse_arg(
    x_config = config$imputation_method,
    x_var = imputation_method,
    var_name = "imputation_method",
    type = "character",
    optional = TRUE,
    default = default_imputation_method
  )
  
  .check_parameter_value_is_valid(
    x = settings$imputation_method,
    var_name = "imputation_method",
    values = .get_available_imputation_methods()
  )

  if (settings$imputation_method == "lasso") {
    # If glmnet is not installed, use simple imputation.
    if (!require_package(
      x = "glmnet",
      purpose = "to impute data using lasso regression",
      message_type = "backend_warning"
    )) {
      settings$imputation_method <- "simple"
    }
  }

  # transformation_method ------------------------------------------------------
  # Transformation method
  settings$transform_method <- .parse_arg(
    x_config = config$transformation_method,
    x_var = transformation_method,
    var_name = "transformation_method",
    type = "character",
    optional = TRUE,
    default = "yeo_johnson_robust"
  )

  .check_parameter_value_is_valid(
    x = settings$transform_method,
    var_name = "transformation_method",
    values = .get_available_transformation_methods()
  )
  
  # If power.transform is not installed, no transformation can be performed.
  if (!require_package(
    x = "power.transform",
    purpose = "to transform data",
    message_type = "backend_warning"
  )) {
    
    settings$transform_method <- "none"
  }

  # transformation_optimisation_criterion --------------------------------------
  # Optimisation criterion
  settings$transformation_optimisation_criterion <- .parse_arg(
    x_config = config$transformation_optimisation_criterion,
    x_var = transformation_optimisation_criterion,
    var_name = "transformation_optimisation_criterion",
    type = "character",
    optional = TRUE,
    default = "mle"
  )
  
  .check_parameter_value_is_valid(
    x = settings$transformation_optimisation_criterion,
    var_name = "transformation_optimisation_criterion",
    values = c("mle", "cramer_von_mises")
  )
  
  # transformation_gof_test_p_value --------------------------------------------
  # Significance threshold for empirical goodness-of-fit test for central
  # normality.
  settings$transformation_gof_test_p_value <- .parse_arg(
    x_config = config$transformation_gof_test_p_value,
    x_var = transformation_gof_test_p_value,
    var_name = "transformation_gof_test_p_value",
    type = "numeric",
    optional = TRUE,
    default = NULL
  )
  
  if (!is.null(settings$transformation_gof_test_p_value)) {
    .check_number_in_valid_range(
      x = settings$transformation_gof_test_p_value,
      var_name = "transformation_gof_test_p_value",
      range = c(0.0, 1.0)
    )
  }
  
  # normalisation_method -------------------------------------------------------
  # Normalisation method
  settings$normalisation_method <- .parse_arg(
    x_config = config$normalisation_method,
    x_var = normalisation_method,
    var_name = "normalisation_method",
    type = "character",
    optional = TRUE,
    default = "standardisation_robust"
  )

  .check_parameter_value_is_valid(
    x = settings$normalisation_method,
    var_name = "normalisation_method",
    values = .get_available_normalisation_methods()
  )

  # batch_normalisation_method -------------------------------------------------
  # Batch normalisation method
  settings$batch_normalisation_method <- .parse_arg(
    x_config = config$batch_normalisation_method,
    x_var = batch_normalisation_method,
    var_name = "batch_normalisation_method",
    type = "character",
    optional = TRUE,
    default = "none"
  )

  .check_parameter_value_is_valid(
    x = settings$batch_normalisation_method,
    var_name = "batch_normalisation_method",
    values = .get_available_batch_normalisation_methods()
  )

  # If the batch normalisation method is combat, pre-normalisation of the
  # entire data is required.
  if (
    settings$batch_normalisation_method %in% 
    .get_available_batch_normalisation_methods(type = "combat") &&
    settings$normalisation_method %in% c("none", "mean_centering")
  ) {
    settings$normalisation_method <- "standardisation"
  }

  # cluster_method -------------------------------------------------------------
  # Feature clustering
  settings$cluster_method <- .parse_arg(
    x_config = config$cluster_method,
    x_var = cluster_method,
    var_name = "cluster_method",
    type = "character",
    optional = TRUE,
    default = "hclust"
  )

  # Check that the cluster package is installed, and revert to if none.
  if (!settings$cluster_method %in% c("none", "hclust")) {
    if (!require_package(
      x = "cluster",
      purpose = "to cluster similar features together",
      message_type = "backend_warning"
    )) {
      settings$cluster_method <- "hclust"
    }
  }

  # Advise to install the fastcluster package.
  if (settings$cluster_method == "hclust") {
    require_package(
      x = "fastcluster",
      purpose = "to create clusters faster",
      message_type = "backend_warning"
    )
  }

  # cluster_linkage ------------------------------------------------------------
  # Feature cluster linkage method
  settings$cluster_linkage <- .parse_arg(
    x_config = config$cluster_linkage_method,
    x_var = cluster_linkage_method,
    var_name = "cluster_linkage_method",
    type = "character",
    optional = TRUE,
    default = "average"
  )

  # cluster_cut_method ---------------------------------------------------------
  # Feature cluster cut method
  default_cluster_cut_method <- ifelse(
    settings$cluster_method == "pam",
    "silhouette", "fixed_cut"
  )

  settings$cluster_cut_method <- .parse_arg(
    x_config = config$cluster_cut_method,
    x_var = cluster_cut_method,
    var_name = "cluster_cut_method",
    type = "character",
    optional = TRUE,
    default = default_cluster_cut_method
  )

  if (settings$cluster_cut_method == "dynamic_cut") {
    if (!require_package(
      x = "dynamicTreeCut",
      purpose = "to cut dendrograms dynamically",
      message_type = "backend_warning"
    )) {
      settings$cluster_cut_method <- "fixed_cut"
    }
  }

  # cluster_similarity_metric --------------------------------------------------
  # Feature similarity metric which expresses some sort of correlation between a
  # pair of features.
  settings$cluster_similarity_metric <- .parse_arg(
    x_config = config$cluster_similarity_metric,
    x_var = cluster_similarity_metric,
    var_name = "cluster_similarity_metric",
    type = "character",
    optional = TRUE,
    default = "mutual_information"
  )

  if (settings$cluster_similarity_metric %in%
      c("mcfadden_r2", "cox_snell_r2", "nagelkerke_r2")) {
    if (!require_package(
      x = "nnet",
      purpose = paste0(
        "to compute log-likelihood pseudo R2 similarity using the ",
        settings$cluster_similarity_metric, " metric"
      ),
      message_type = "backend_warning"
    )) {
      settings$cluster_similarity_metric <- "spearman"
    }
    
  } else if (settings$cluster_similarity_metric %in% c("mutual_information")) {
    if (!require_package(
      x = "praznik",
      purpose = paste0(
        "to compute similarity using the ",
        settings$cluster_similarity_metric, " metric"
      ),
      message_type = "backend_warning"
    )) {
      settings$cluster_similarity_metric <- "spearman"
    }
  }

  # cluster_similarity_threshold -----------------------------------------------
  # Feature cluster similarity that determines the similarity threshold for
  # features to be considered part of one cluster. Should be expressed in terms
  # of the similarity metric, e.g. 0.8 for spearman would consider all features
  # that have a pairwise correlation of 0.8 and over to belong to a cluster.
  settings$cluster_similarity_threshold <- .parse_arg(
    x_config = config$cluster_similarity_threshold,
    x_var = cluster_similarity_threshold,
    var_name = "cluster_similarity_threshold",
    type = "numeric",
    optional = TRUE,
    default = NULL
  )

  if (is.null(settings$cluster_similarity_threshold)) {
    if (settings$cluster_cut_method %in% c("fixed_cut")) {
      # Fixed cut requires stringent defaults, otherwise non-sense clusters will
      # be produced.
      if (settings$cluster_similarity_metric %in% c("mcfadden_r2")) {
        settings$cluster_similarity_threshold <- 0.30
      } else if (settings$cluster_similarity_metric %in% c("mutual_information")) {
        settings$cluster_similarity_threshold <- 0.23
      } else if (settings$cluster_similarity_metric %in% c("cox_snell_r2", "nagelkerke_r2")) {
        settings$cluster_similarity_threshold <- 0.75
      } else {
        settings$cluster_similarity_threshold <- 0.90
      }
      
    } else {
      # The similarity threshold is also used to determine if any clusters could
      # potentially be found. The threshold is set low so that other cut methods
      # can be explored.
      if (settings$cluster_similarity_metric %in% c("mcfadden_r2")) {
        settings$cluster_similarity_threshold <- 0.25
      } else if (settings$cluster_similarity_metric %in% c("mutual_information")) {
        settings$cluster_similarity_threshold <- 0.10
      } else if (settings$cluster_similarity_metric %in% c("cox_snell_r2", "nagelkerke_r2")) {
        settings$cluster_similarity_threshold <- 0.40
      } else {
        settings$cluster_similarity_threshold <- 0.70
      }
    }
  }

  # cluster_representation_method ----------------------------------------------
  # Method to select the feature that represents the cluster
  settings$cluster_representation_method <- .parse_arg(
    x_config = config$cluster_representation_method,
    x_var = cluster_representation_method,
    var_name = "cluster_representation_method",
    type = "character",
    optional = TRUE,
    default = "best_predictor"
  )

  # Partioning around medioids only allows the use of medioids for
  # representation.
  if (settings$cluster_method == "pam") {
    settings$cluster_representation_method <- "medioid"
  }

  # If mean is used, this requires the data to have some kind of standardisation
  # with centering at 0.
  if (settings$cluster_representation_method %in% c("mean")) {
    if (!settings$normalisation_method %in% c(
      "standardisation", "standardisation_trim", "standardisation_winsor",
      "standardisation_robust", "quantile"
    )) {
      ..warning(
        paste0(
          "When computing the meta-feature for a cluster using the mean value ",
          "of co-clustered features,  each feature is expected to be centered ",
          "at 0.0 and to have a standard scale. The provided normalisation method (",
          settings$normalisation_method, ") does not allow this, ",
          "and has been replaced by the \"standardisation_robust\" method."
        )
      )

      # Replace by standardisation_robust
      settings$normalisation_method <- "standardisation_robust"
    }
  }

  # Perform a check on all methods.
  .check_cluster_parameters(
    cluster_method = settings$cluster_method,
    cluster_cut_method = settings$cluster_cut_method,
    cluster_linkage = settings$cluster_linkage,
    cluster_similarity_threshold = settings$cluster_similarity_threshold,
    cluster_similarity_metric = settings$cluster_similarity_metric,
    cluster_representation_method = settings$cluster_representation_method,
    data_type = "cluster",
    message_type = "backend_error"
  )

  # parallel_preprocessing -----------------------------------------------------
  # Parallel processing
  settings$do_parallel <- .parse_arg(
    x_config = config$parallel_preprocessing,
    x_var = parallel_preprocessing,
    var_name = "parallel_preprocessing",
    type = "logical",
    optional = TRUE,
    default = TRUE
  )

  # Disable if parallel is FALSE.
  if (!parallel) settings$do_parallel <- FALSE

  return(settings)
}



#' Internal function for parsing settings related to variable importance
#' computation.
#'
#' @param config A list of settings, e.g. from an xml file.
#' @param data Data set as loaded using the `.load_data` function.
#' @param parallel Logical value that whether familiar uses parallelisation. If
#'   `FALSE` it will override `parallel_vimp`.
#' @param outcome_type Type of outcome found in the data set.
#' @param vimp_method (**required**) Variable importance method. `familiar`
#'   implements various variable importance methods. Please refer to the
#'   vignette on variable importance methods for more details.
#'
#'   More than one variable importance method can be chosen. The experiment will
#'   then repeated for each feature selection method.
#'
#'   Variable importance methods determine the ranking of features. Actual
#'   selection of features is done by optimising the signature size model
#'   hyperparameter during the hyperparameter optimisation step.
#' @param vimp_method_parameter (*optional*) List of lists containing parameters
#'   for feature selection methods. Each sublist should have the name of the
#'   feature selection method it corresponds to.
#'
#'   Most feature selection methods do not have parameters that can be set.
#'   Please refer to the vignette on feature selection methods for more details.
#'   Note that if the feature selection method is based on a learner (e.g. lasso
#'   regression), hyperparameter optimisation may be performed prior to
#'   assessing variable importance.
#' @param vimp_aggregation_method (*optional*) The method used to aggregate
#'   variable importances over different data subsets, e.g. bootstraps. The
#'   following methods can be selected:
#'
#'   * `none`: Don't aggregate ranks, but rather aggregate the variable
#'   importance scores themselves.
#'
#'   * `mean`: Use the mean rank of a feature over the subsets to
#'   determine the aggregated feature rank.
#'
#'   * `median`: Use the median rank of a feature over the subsets to determine
#'   the aggregated feature rank.
#'
#'   * `best`: Use the best rank the feature obtained in any subset to determine
#'   the aggregated feature rank.
#'
#'   * `worst`: Use the worst rank the feature obtained in any subset to
#'   determine the aggregated feature rank.
#'
#'   * `stability`: Use the frequency of the feature being in the subset of
#'   highly ranked features as measure for the aggregated feature rank
#'   (Meinshausen and Buehlmann, 2010).
#'
#'   * `exponential`: Use a rank-weighted frequence of occurrence in the subset
#'   of highly ranked features as measure for the aggregated feature rank (Haury
#'   et al., 2011).
#'
#'   * `borda` (default): Use the borda count as measure for the aggregated
#'   feature rank (Wald et al., 2012).
#'
#'   * `enhanced_borda`: Use an occurrence frequency-weighted borda count as
#'   measure for the aggregated feature rank (Wald et al., 2012).
#'
#'   * `truncated_borda`: Use borda count computed only on features within the
#'   subset of highly ranked features.
#'
#'   * `enhanced_truncated_borda`: Apply both the enhanced borda method and the
#'   truncated borda method and use the resulting borda count as the aggregated
#'   feature rank.
#'
#'   The *feature selection methods* vignette provides additional information.
#'
#' @param vimp_aggregation_rank_threshold (*optional*) The threshold used to
#'   define the subset of highly important features. If set to `NULL`, this
#'   threshold is determined by maximising the variance in the occurrence value
#'   over all features over the subset size. The default value is `5`.
#'
#'   This parameter is only relevant for `stability`, `exponential`,
#'   `enhanced_borda`, `truncated_borda` and `enhanced_truncated_borda` methods.
#' @param parallel_vimp (*optional*) Enable parallel processing for
#'   variable importance tasks. Defaults to `TRUE`. When set to `FALSE`,
#'   this will disable the use of parallel processing while performing feature
#'   selection, regardless of the settings of the `parallel` parameter.
#'   `parallel_vimp` is ignored if `parallel = FALSE`.
#' @param ... Unused arguments.
#'
#' @return List of parameters related to variable importance computation.
#'
#' @references 1. Wald, R., Khoshgoftaar, T. M., Dittman, D., Awada, W. &
#'   Napolitano, A. An extensive comparison of feature ranking aggregation
#'   techniques in bioinformatics. in 2012 IEEE 13th International Conference on
#'   Information Reuse Integration (IRI) 377–384 (2012).
#'
#'   1. Meinshausen, N. & Buehlmann, P. Stability selection. J. R. Stat. Soc.
#'   Series B Stat. Methodol. 72, 417–473 (2010).
#'
#'   1. Haury, A.-C., Gestraud, P. & Vert, J.-P. The influence of feature
#'   selection methods on accuracy, stability and interpretability of molecular
#'   signatures. PLoS One 6, e28210 (2011).
#' @md
#' @keywords internal
.parse_variable_importance_settings <- function(
    config = NULL,
    data,
    parallel,
    outcome_type,
    vimp_method = waiver(),
    vimp_method_parameter = waiver(),
    vimp_aggregation_method = waiver(),
    vimp_aggregation_rank_threshold = waiver(),
    parallel_vimp = waiver(),
    ...
) {
  settings <- list()
  
  # vimp_method ----------------------------------------------------------------
  # Variable importance methods
  settings$vimp_methods <- .parse_arg(
    x_config = config$vimp_method,
    x_var = vimp_method,
    var_name = "vimp_method",
    type = "character_list",
    optional = FALSE
  )

  sapply(
    settings$vimp_methods, 
    .check_vimp_outcome_type, 
    outcome_type = outcome_type
  )

  # vimp_method_parameter ------------------------------------------------------
  # Variable importance parameters
  settings$param <- .parse_arg(
    x_config = config$vimp_method_parameter,
    x_var = vimp_method_parameter,
    var_name = "vimp_method_parameter",
    type = "list",
    optional = TRUE,
    default = list()
  )

  settings$param <- .parse_hyperparameters(
    data = data,
    parameter_list = settings$param,
    outcome_type = outcome_type,
    vimp_method = settings$vimp_methods
  )

  # vimp_aggregation_method ----------------------------------------------------
  # Variable importance aggregation methods
  settings$aggregation <- .parse_arg(
    x_config = config$vimp_aggregation_method,
    x_var = vimp_aggregation_method,
    var_name = "vimp_aggregation_method",
    type = "character",
    optional = TRUE,
    default = "borda"
  )

  .check_parameter_value_is_valid(
    x = settings$aggregation,
    var_name = "vimp_aggregation_method",
    values = .get_available_rank_aggregation_methods()
  )

  # vimp_aggregation_rank_threshold --------------------------------------------
  # Variable importance rank threshold (used by some aggregation methods)
  settings$aggr_rank_threshold <- .parse_arg(
    x_config = config$vimp_aggregation_rank_threshold,
    x_var = vimp_aggregation_rank_threshold,
    var_name = "vimp_aggregation_rank_threshold",
    type = "integer",
    optional = TRUE,
    default = 5L
  )

  if (!is.null(settings$aggr_rank_threshold)) {
    .check_number_in_valid_range(
      x = settings$aggr_rank_threshold,
      var_name = "vimp_aggregation_rank_threshold",
      range = c(1L, Inf)
    )
  }

  # parallel_vimp --------------------------------------------------------------
  # Parallelisation switch for variable importance computation
  settings$do_parallel <- .parse_arg(
    x_config = config$parallel_vimp,
    x_var = parallel_vimp,
    var_name = "parallel_vimp",
    type = "logical",
    optional = TRUE,
    default = TRUE
  )

  # Disable if parallel is FALSE.
  if (!parallel) settings$do_parallel <- FALSE

  return(settings)
}



#' Internal function for parsing settings related to model development
#'
#' @param config A list of settings, e.g. from an xml file.
#' @param data Data set as loaded using the `.load_data` function.
#' @param parallel Logical value that whether familiar uses parallelisation. If
#'   `FALSE` it will override `parallel_model_development`.
#' @param outcome_type Type of outcome found in the data set.
#' @param learner (**required**) One or more algorithms used for model
#'   development. A sizeable number learners is supported in `familiar`. Please
#'   see the vignette on learners for more information concerning the available
#'   learners.
#' @param hyperparameter (*optional*) List of lists containing hyperparameters
#'   for learners. Each sublist should have the name of the learner method it
#'   corresponds to, with list elements being named after the intended
#'   hyperparameter, e.g. \code{"glm_logistic"=list("sign_size"=3)}
#'
#'   All learners have hyperparameters. Please refer to the vignette on learners
#'   for more details. If no parameters are provided, sequential model-based
#'   optimisation is used to determine optimal hyperparameters.
#'
#'   Hyperparameters provided by the user are never optimised. However, if more
#'   than one value is provided for a single hyperparameter, optimisation will
#'   be conducted using these values.
#' @param novelty_detector (*optional*) Specify the algorithm used for training
#'   a novelty detector. This detector can be used to identify
#'   out-of-distribution data prospectively.
#' @param detector_parameters (*optional*) List lists containing hyperparameters
#'   for novelty detectors. Currently not used.
#' @param parallel_model_development (*optional*) Enable parallel processing for
#'   the model development workflow. Defaults to `TRUE`. When set to `FALSE`,
#'   this will disable the use of parallel processing while developing models,
#'   regardless of the settings of the `parallel` parameter.
#'   `parallel_model_development` is ignored if `parallel=FALSE`.
#' @param ... Unused arguments.
#'
#' @return List of parameters related to model development.
#' @md
#' @keywords internal
.parse_model_development_settings <- function(
    config = NULL,
    data,
    parallel,
    outcome_type,
    learner = waiver(),
    hyperparameter = waiver(),
    novelty_detector = waiver(),
    detector_parameters = waiver(),
    parallel_model_development = waiver(),
    ...
) {
  settings <- list()
  
  # learner --------------------------------------------------------------------
  # Learners for model development
  settings$learners <- .parse_arg(
    x_config = config$learner, x_var = learner,
    var_name = "learner",
    type = "character_list",
    optional = FALSE
  )

  sapply(
    settings$learners,
    .check_learner_outcome_type, 
    outcome_type = outcome_type
  )

  # hyperparameters ------------------------------------------------------------
  # Model hyperparameters
  settings$hyper_param <- .parse_arg(
    x_config = config$hyperparameter,
    x_var = hyperparameter,
    var_name = "hyperparameter",
    type = "list",
    optional = TRUE,
    default = list()
  )

  settings$hyper_param <- .parse_hyperparameters(
    data = data,
    parameter_list = settings$hyper_param,
    outcome_type = outcome_type,
    learner = settings$learners
  )

  # novelty_detector -----------------------------------------------------------
  settings$novelty_detector <- .parse_arg(
    x_config = config$novelty_detector,
    x_var = novelty_detector,
    var_name = "novelty_detector",
    type = "character",
    optional = TRUE,
    default = "isolation_forest"
  )

  .check_novelty_detector_available(detector = settings$novelty_detector)

  # detector_parameters --------------------------------------------------------
  settings$detector_parameters <- .parse_arg(
    x_config = config$detector_parameters,
    x_var = detector_parameters,
    var_name = "detector_parameters",
    type = "list",
    optional = TRUE,
    default = list()
  )

  settings$detector_parameters <- .parse_hyperparameters(
    data = data,
    parameter_list = settings$detector_parameters,
    detector = settings$novelty_detector,
    outcome_type = "unsupervised"
  )

  # parallel_model_development -------------------------------------------------
  # Parallelisation switch for model building
  settings$do_parallel <- .parse_arg(
    x_config = config$parallel_model_development,
    x_var = parallel_model_development,
    var_name = "parallel_model_development",
    type = "logical",
    optional = TRUE,
    default = TRUE
  )

  # Disable if parallel is FALSE
  if (!parallel) settings$do_parallel <- FALSE

  return(settings)
}



#' Internal function for parsing settings related to hyperparameter optimisation
#'
#' @param config A list of settings, e.g. from an xml file.
#' @param parallel Logical value that whether familiar uses parallelisation. If
#'   `FALSE` it will override `parallel_hyperparameter_optimisation`.
#' @param outcome_type Type of outcome found in the data set.
#' @param optimisation_bootstraps (*optional*) Number of bootstraps that should
#'   be generated from the development data set. During the optimisation
#'   procedure one or more of these bootstraps (indicated by
#'   `smbo_step_bootstraps`) are used for model development using different
#'   combinations of hyperparameters. The effect of the hyperparameters is then
#'   assessed by comparing in-bag and out-of-bag model performance.
#'
#'   The default number of bootstraps is `50`. Hyperparameter optimisation may
#'   finish before exhausting the set of bootstraps.
#' @param optimisation_determine_vimp (*optional*) Logical value that indicates
#'   whether variable importance is determined separately for each of the
#'   bootstraps created during the optimisation process (`TRUE`) or the
#'   applicable results from the variable importance computation step are used
#'   (`FALSE`).
#'
#'   Determining variable importance increases the initial computational
#'   overhead. However, it prevents positive biases for the out-of-bag data due
#'   to overlap of these data with the development data set used for the feature
#'   selection step. In this case, any hyperparameters of the variable
#'   importance method are not determined separately for each bootstrap, but
#'   those obtained during the variable importance computation step are used
#'   instead. In case multiple of such hyperparameter sets could be applicable,
#'   the set that will be used is randomly selected for each bootstrap.
#'
#'   This parameter only affects hyperparameter optimisation of learners. The
#'   default is `TRUE`.
#' @param smbo_random_initialisation (*optional*) String indicating the
#'   initialisation method for the hyperparameter space. Can be one of
#'   `fixed_subsample` (default), `fixed`, or `random`. `fixed` and
#'   `fixed_subsample` first create hyperparameter sets from a range of default
#'   values set by familiar. `fixed_subsample` then randomly draws up to
#'   `smbo_n_random_sets` from the grid. `random` does not rely upon a fixed
#'   grid, and randomly draws up to `smbo_n_random_sets` hyperparameter sets
#'   from the hyperparameter space.
#' @param smbo_n_random_sets (*optional*) Number of random or subsampled
#'   hyperparameters drawn during the initialisation process. Default: `100`.
#'   Cannot be smaller than `10`. The parameter is not used when
#'   `smbo_random_initialisation` is `fixed`, as the entire pre-defined grid
#'   will be explored.
#'
#' @param max_smbo_iterations (*optional*) Maximum number of intensify
#'   iterations of the SMBO algorithm. During an intensify iteration a run-off
#'   occurs between the current *best* hyperparameter combination and either 10
#'   challenger combination with the highest expected improvement or a set of 20
#'   random combinations.
#'
#'   Run-off with random combinations is used to force exploration of the
#'   hyperparameter space, and is performed every second intensify iteration, or
#'   if there is no expected improvement for any challenger combination.
#'
#'   If a combination of hyperparameters leads to better performance on the same
#'   data than the incumbent *best* set of hyperparameters, it replaces the
#'   incumbent set at the end of the intensify iteration.
#'
#'   The default number of intensify iteration is `20`. Iterations may be
#'   stopped early if the incumbent set of hyperparameters remains the same for
#'   `smbo_stop_convergent_iterations` iterations, or performance improvement is
#'   minimal. This behaviour is suppressed during the first 4 iterations to
#'   enable the algorithm to explore the hyperparameter space.
#' @param smbo_stop_convergent_iterations (*optional*) The number of subsequent
#'   convergent SMBO iterations required to stop hyperparameter optimisation
#'   early. An iteration is convergent if the *best* parameter set has not
#'   changed or the optimisation score over the 4 most recent iterations has not
#'   changed beyond the tolerance level in `smbo_stop_tolerance`.
#'
#'   The default value is `3`.
#' @param smbo_stop_tolerance (*optional*) Tolerance for early stopping due to
#'   convergent optimisation score.
#'
#'   The default value depends on the square root of the number of samples (at
#'   the series level), and is `0.01` for 100 samples. This value is computed as
#'   `0.1 * 1 / sqrt(n_samples)`. The upper limit is `0.0001` for 1M or more
#'   samples.
#' @param smbo_time_limit (*optional*) Time limit (in minutes) for the
#'   optimisation process. Optimisation is stopped after this limit is exceeded.
#'   Time taken to determine variable importance for the optimisation process
#'   (see the `optimisation_determine_vimp` parameter) does not count.
#'
#'   The default is `NULL`, indicating that there is no time limit for the
#'   optimisation process. The time limit cannot be less than 1 minute.
#'
#' @param smbo_initial_bootstraps (*optional*) The number of bootstraps taken
#'   from the set of `optimisation_bootstraps` as the bootstraps assessed
#'   initially.
#'
#'   The default value is `1`. The value cannot be larger than
#'   `optimisation_bootstraps`.
#'
#' @param smbo_step_bootstraps (*optional*) The number of bootstraps taken from
#'   the set of `optimisation_bootstraps` bootstraps as the bootstraps assessed
#'   during the steps of each intensify iteration.
#'
#'   The default value is `3`. The value cannot be larger than
#'   `optimisation_bootstraps`.
#' @param smbo_intensify_steps (*optional*) The number of steps in each SMBO
#'   intensify iteration. Each step a new set of `smbo_step_bootstraps`
#'   bootstraps is drawn and used in the run-off between the incumbent *best*
#'   hyperparameter combination and its challengers.
#'
#'   The default value is `5`. Higher numbers allow for a more detailed
#'   comparison, but this comes with added computational cost.
#'
#' @param optimisation_metric (*optional*) One or more metrics used to compute
#'   performance scores. See the vignette on performance metrics for the
#'   available metrics.
#'
#'   If unset, the following metrics are used by default:
#'
#'   * `auc_roc`: For `binomial` and `multinomial` models.
#'
#'   * `mse`: Mean squared error for `continuous` models.
#'
#'   * `concordance_index`: For `survival` models.
#'
#'   Multiple optimisation metrics can be specified. Actual metric values are
#'   converted to an objective value by comparison with a baseline metric value
#'   that derives from a trivial model, i.e. majority class for binomial and
#'   multinomial outcomes, the median outcome for count and continuous outcomes
#'   and a fixed risk or time for survival outcomes.
#'
#' @param optimisation_function (*optional*) Type of optimisation function used
#'   to quantify the performance of a hyperparameter set. Model performance is
#'   assessed using the metric(s) specified by `optimisation_metric` on the
#'   in-bag (IB) and out-of-bag (OOB) samples of a bootstrap. These values are
#'   converted to objective scores with a standardised interval of 
#'   \eqn{[-1.0, 1.0]}. Each pair of objective is subsequently used to compute an
#'   optimisation score. The optimisation score across different bootstraps is
#'   than aggregated to a summary score. This summary score is used to rank
#'   hyperparameter sets, and select the optimal set.
#'
#'   The combination of optimisation score and summary score is determined by
#'   the optimisation function indicated by this parameter:
#'
#'   * `validation` or `max_validation` (default): seeks to maximise OOB score.
#'
#'   * `balanced`: seeks to balance IB and OOB score.
#'
#'   * `stronger_balance`: similar to `balanced`, but with stronger penalty for
#'   differences between IB and OOB scores.
#'
#'   * `validation_minus_sd`: seeks to optimise the average OOB score minus its
#'   standard deviation.
#'
#'   * `validation_25th_percentile`: seeks to optimise the 25th percentile of
#'   OOB scores, and is conceptually similar to `validation_minus_sd`.
#'
#'   * `model_estimate`: seeks to maximise the OOB score estimate predicted by
#'   the hyperparameter learner (not available for random search).
#'
#'   * `model_estimate_minus_sd`: seeks to maximise the OOB score estimate minus
#'   its estimated standard deviation, as predicted by the hyperparameter
#'   learner (not available for random search).
#'
#'   * `model_balanced_estimate`: seeks to maximise the estimate of the balanced
#'   IB and OOB score. This is similar to the `balanced` score, and in fact uses
#'   a hyperparameter learner to predict said score (not available for random
#'   search).
#'
#'   * `model_balanced_estimate_minus_sd`: seeks to maximise the estimate of the
#'   balanced IB and OOB score, minus its estimated standard deviation. This is
#'   similar to the `balanced` score, but takes into account its estimated
#'   spread.
#'
#'   Additional detail are provided in the *Learning algorithms and
#'   hyperparameter optimisation* vignette.
#'
#' @param hyperparameter_learner (*optional*) Any point in the hyperparameter
#'   space has a single, scalar, optimisation score value that is *a priori*
#'   unknown. During the optimisation process, the algorithm samples from the
#'   hyperparameter space by selecting hyperparameter sets and computing the
#'   optimisation score value for one or more bootstraps. For each
#'   hyperparameter set the resulting values are distributed around the actual
#'   value. The learner indicated by `hyperparameter_learner` is then used to
#'   infer optimisation score estimates for unsampled parts of the
#'   hyperparameter space.
#'
#'   The following models are available:
#'
#'   * `bayesian_additive_regression_trees` or `bart`: Uses Bayesian Additive
#'   Regression Trees (Sparapani et al., 2021) for inference. Unlike standard
#'   random forests, BART allows for estimating posterior distributions directly
#'   and can extrapolate.
#'
#'   * `gaussian_process` (default): Creates a localised approximate Gaussian
#'   process for inference (Gramacy, 2016). This allows for better scaling than
#'   deterministic Gaussian Processes.
#'
#'   * `random_forest`: Creates a random forest for inference. Originally
#'   suggested by Hutter et al. (2011). A weakness of random forests is their
#'   lack of extrapolation beyond observed values, which limits their usefulness
#'   in exploiting promising areas of hyperparameter space.
#'
#'   * `random` or `random_search`: Forgoes the use of models to steer
#'   optimisation. Instead, a random search is performed.
#'
#' @param acquisition_function (*optional*) The acquisition function influences
#'   how new hyperparameter sets are selected. The algorithm uses the model
#'   learned by the learner indicated by `hyperparameter_learner` to search the
#'   hyperparameter space for hyperparameter sets that are either likely better
#'   than the best known set (*exploitation*) or where there is considerable
#'   uncertainty (*exploration*). The acquisition function quantifies this
#'   (Shahriari et al., 2016).
#'
#'   The following acquisition functions are available, and are described in
#'   more detail in the *learner algorithms* vignette:
#'
#'   * `improvement_probability`: The probability of improvement quantifies the
#'   probability that the expected optimisation score for a set is better than
#'   the best observed optimisation score
#'
#'   * `improvement_empirical_probability`: Similar to
#'   `improvement_probability`, but based directly on optimisation scores
#'   predicted by the individual decision trees.
#'
#'   * `expected_improvement` (default): Computes expected improvement.
#'
#'   * `upper_confidence_bound`: This acquisition function is based on the upper
#'   confidence bound of the distribution (Srinivas et al., 2012).
#'
#'   * `bayes_upper_confidence_bound`: This acquisition function is based on the
#'   upper confidence bound of the distribution (Kaufmann et al., 2012).
#' @param exploration_method (*optional*) Method used to steer exploration in
#'   post-initialisation intensive searching steps. As stated earlier, each SMBO
#'   iteration step compares suggested alternative parameter sets with an
#'   incumbent **best** set in a series of steps. The exploration method
#'   controls how the set of alternative parameter sets is pruned after each
#'   step in an iteration. Can be one of the following:
#'
#'   * `single_shot` (default): The set of alternative parameter sets is not
#'   pruned, and each intensification iteration contains only a single
#'   intensification step that only uses a single bootstrap. This is the fastest
#'   exploration method, but only superficially tests each parameter set.
#'
#'   * `successive_halving`: The set of alternative parameter sets is
#'   pruned by removing the worst performing half of the sets after each step
#'   (Jamieson and Talwalkar, 2016).
#'
#'   * `stochastic_reject`: The set of alternative parameter sets is pruned by
#'   comparing the performance of each parameter set with that of the incumbent
#'   **best** parameter set using a paired Wilcoxon test based on shared
#'   bootstraps. Parameter sets that perform significantly worse, at an alpha
#'   level indicated by `smbo_stochastic_reject_p_value`, are pruned.
#'
#'   * `none`: The set of alternative parameter sets is not pruned.
#'
#' @param smbo_stochastic_reject_p_value (*optional*) The p-value threshold used
#'   for the `stochastic_reject` exploration method.
#'
#'   The default value is `0.05`.
#' @param parallel_hyperparameter_optimisation (*optional*) Enable parallel
#'   processing for hyperparameter optimisation. Defaults to `TRUE`. When set to
#'   `FALSE`, this will disable the use of parallel processing while performing
#'   optimisation, regardless of the settings of the `parallel` parameter. The
#'   parameter moreover specifies whether parallelisation takes place within the
#'   optimisation algorithm (`inner`, default), or in an outer loop ( `outer`)
#'   over learners, data subsamples, etc.
#'
#'   `parallel_hyperparameter_optimisation` is ignored if `parallel=FALSE`.
#' @param ... Unused arguments.
#'
#' @return List of parameters related to model hyperparameter optimisation.
#'
#' @references 1. Hutter, F., Hoos, H. H. & Leyton-Brown, K. Sequential
#'   model-based optimization for general algorithm configuration. in Learning
#'   and Intelligent Optimization (ed. Coello, C. A. C.) 6683, 507–523 (Springer
#'   Berlin Heidelberg, 2011).
#'
#'   1. Shahriari, B., Swersky, K., Wang, Z., Adams, R. P. & de Freitas, N.
#'   Taking the Human Out of the Loop: A Review of Bayesian Optimization. Proc.
#'   IEEE 104, 148–175 (2016)
#'
#'   1. Srinivas, N., Krause, A., Kakade, S. M. & Seeger, M. W.
#'   Information-Theoretic Regret Bounds for Gaussian Process Optimization in
#'   the Bandit Setting. IEEE Trans. Inf. Theory 58, 3250–3265 (2012)
#'
#'   1. Kaufmann, E., Cappé, O. & Garivier, A. On Bayesian upper confidence
#'   bounds for bandit problems. in Artificial intelligence and statistics
#'   592–600 (2012).
#'
#'   1. Jamieson, K. & Talwalkar, A. Non-stochastic Best Arm Identification and
#'   Hyperparameter Optimization. in Proceedings of the 19th International
#'   Conference on Artificial Intelligence and Statistics (eds. Gretton, A. &
#'   Robert, C. C.) vol. 51 240–248 (PMLR, 2016).
#'
#'   1. Gramacy, R. B. laGP: Large-Scale Spatial Modeling via Local Approximate
#'   Gaussian Processes in R. Journal of Statistical Software 72, 1–46 (2016)
#'
#'   1. Sparapani, R., Spanbauer, C. & McCulloch, R. Nonparametric Machine
#'   Learning and Efficient Computation with Bayesian Additive Regression Trees:
#'   The BART R Package. Journal of Statistical Software 97, 1–66 (2021)
#' @md
#' @keywords internal
.parse_hyperparameter_optimisation_settings <- function(
    config = NULL,
    parallel,
    outcome_type,
    optimisation_bootstraps = waiver(),
    optimisation_determine_vimp = waiver(),
    smbo_random_initialisation = waiver(),
    smbo_n_random_sets = waiver(),
    max_smbo_iterations = waiver(),
    smbo_stop_convergent_iterations = waiver(),
    smbo_stop_tolerance = waiver(),
    smbo_time_limit = waiver(),
    smbo_initial_bootstraps = waiver(),
    smbo_step_bootstraps = waiver(),
    smbo_intensify_steps = waiver(),
    smbo_stochastic_reject_p_value = waiver(),
    optimisation_function = waiver(),
    optimisation_metric = waiver(),
    acquisition_function = waiver(),
    exploration_method = waiver(),
    hyperparameter_learner = waiver(),
    parallel_hyperparameter_optimisation = waiver(),
    ...
) {
  settings <- list()

  # smbo_random_initialisation -------------------------------------------------
  # Randomisation of initial parameter grid
  settings$hpo_grid_initialisation_method <- .parse_arg(
    x_config = config$smbo_random_initialisation,
    x_var = smbo_random_initialisation,
    var_name = "smbo_random_initialisation",
    type = "character",
    optional = TRUE,
    default = "fixed_subsample"
  )

  .check_parameter_value_is_valid(
    x = settings$hpo_grid_initialisation_method,
    var_name = "smbo_random_initialisation",
    values = c("fixed_subsample", "fixed", "random")
  )

  # smbo_n_random_sets ---------------------------------------------------------
  # Number of samples in the initial parameter grid.
  settings$hpo_n_grid_initialisation_samples <- .parse_arg(
    x_config = config$smbo_n_random_sets,
    x_var = smbo_n_random_sets,
    var_name = "smbo_n_random_sets",
    type = "integer",
    optional = TRUE,
    default = 100L
  )

  .check_number_in_valid_range(
    x = settings$hpo_n_grid_initialisation_samples,
    var_name = "smbo_n_random_sets",
    range = c(10L, Inf)
  )

  # optimisation_determine_vimp ------------------------------------------------
  # Variable importance for the bootstraps
  settings$hpo_determine_vimp <- .parse_arg(
    x_config = config$optimisation_determine_vimp,
    x_var = optimisation_determine_vimp,
    var_name = "optimisation_determine_vimp",
    type = "logical",
    optional = TRUE,
    default = TRUE
  )

  # max_smbo_iterations --------------------------------------------------------
  # Maximum number of SMBO iterations before stopping
  settings$hpo_smbo_iter_max <- .parse_arg(
    x_config = config$max_smbo_iterations,
    x_var = max_smbo_iterations,
    var_name = "max_smbo_iterations",
    type = "integer",
    optional = TRUE,
    default = 20L
  )

  .check_number_in_valid_range(
    x = settings$hpo_smbo_iter_max,
    var_name = "max_smbo_iterations",
    range = c(1L, Inf)
  )

  # optimisation_bootstraps ----------------------------------------------------
  # Maximum number of bootstraps for hyperparameter evaluation
  settings$hpo_max_bootstraps <- .parse_arg(
    x_config = config$optimisation_bootstraps,
    x_var = optimisation_bootstraps,
    var_name = "optimisation_bootstraps",
    type = "integer",
    optional = TRUE,
    default = 20L
  )

  .check_number_in_valid_range(
    x = settings$hpo_max_bootstraps,
    var_name = "optimisation_bootstraps",
    range = c(10L, Inf)
  )

  # smbo_initial_bootstraps ----------------------------------------------------
  # Number of bootstraps evaluated initially
  settings$hpo_initial_bootstraps <- .parse_arg(
    x_config = config$smbo_initial_bootstraps,
    x_var = smbo_initial_bootstraps,
    var_name = "smbo_initial_bootstraps",
    type = "integer",
    optional = TRUE,
    default = 1L
  )

  .check_number_in_valid_range(
    x = settings$hpo_initial_bootstraps,
    var_name = "smbo_initial_bootstraps",
    range = c(1L, settings$hpo_max_bootstraps)
  )

  # smbo_step_bootstraps -------------------------------------------------------
  # Maximum number of bootstrap evaluated each intensify step of each SMBO
  # iteration.
  settings$hpo_bootstraps <- .parse_arg(
    x_config = config$smbo_step_bootstraps,
    x_var = smbo_step_bootstraps,
    var_name = "smbo_step_bootstraps",
    type = "integer",
    optional = TRUE,
    default = 3L
  )

  .check_number_in_valid_range(
    x = settings$hpo_bootstraps,
    var_name = "smbo_step_bootstraps",
    range = c(1L, settings$hpo_max_bootstraps)
  )

  # smbo_intensify_steps -------------------------------------------------------
  # Maximum number of intensify iterations during each SMBO iteration.
  settings$hpo_intensify_max_iter <- .parse_arg(
    x_config = config$smbo_intensify_steps,
    x_var = smbo_intensify_steps,
    var_name = "smbo_intensify_steps",
    type = "integer",
    optional = TRUE,
    default = 5L
  )

  .check_number_in_valid_range(
    x = settings$hpo_intensify_max_iter,
    var_name = "smbo_intensify_steps",
    range = c(1L, Inf)
  )

  # smbo_stochastic_reject_p_value ---------------------------------------------
  # Significance level for early stopping of intensity iterations. Only used for
  # stochastic rejection as an exploration method.
  settings$hpo_alpha <- .parse_arg(
    x_config = config$smbo_stochastic_reject_p_value,
    x_var = smbo_stochastic_reject_p_value,
    var_name = "smbo_stochastic_reject_p_value",
    type = "numeric",
    optional = TRUE,
    default = 0.05
  )

  .check_number_in_valid_range(
    x = settings$hpo_alpha,
    var_name = "smbo_stochastic_reject_p_value",
    range = c(0.0, 1.0),
    closed = c(FALSE, TRUE)
  )

  # smbo_stop_convergent_iterations --------------------------------------------
  # Number of converging iterations before stopping SMBO.
  settings$hpo_conv_stop <- .parse_arg(
    x_config = config$smbo_stop_convergent_iterations,
    x_var = smbo_stop_convergent_iterations,
    var_name = "smbo_stop_convergent_iterations",
    type = "integer",
    optional = TRUE,
    default = 3L
  )

  .check_number_in_valid_range(
    x = settings$hpo_conv_stop,
    var_name = "smbo_stop_convergent_iterations",
    range = c(1L, Inf)
  )

  # smbo_stop_tolerance --------------------------------------------------------
  # Convergence tolerance
  settings$hpo_convergence_tolerance <- .parse_arg(
    x_config = config$smbo_stop_tolerance,
    x_var = smbo_stop_tolerance,
    var_name = "smbo_stop_tolerance",
    type = "numeric",
    optional = TRUE,
    default = NULL
  )

  # Check provided settings. If NULL, convergence will be set by the algorithm.
  if (!is.null(settings$hpo_convergence_tolerance)) {
    .check_number_in_valid_range(
      x = settings$hpo_convergence_tolerance,
      var_name = "smbo_stop_tolerance",
      range = c(0.0, 1.0),
      closed = c(FALSE, TRUE)
    )
  }

  # smbo_time_limit ------------------------------------------------------------
  # Time limit for the optimisation process..
  settings$hpo_time_limit <- .parse_arg(
    x_config = config$smbo_time_limit,
    x_var = smbo_time_limit,
    var_name = "smbo_time_limit",
    type = "numeric",
    optional = TRUE,
    default = NULL
  )

  if (!is.null(settings$hpo_time_limit)) {
    .check_number_in_valid_range(
      x = settings$hpo_time_limit,
      var_name = "smbo_time_limit",
      range = c(0.0, Inf),
      closed = c(FALSE, TRUE)
    )

    require_package(
      c("callr", "microbenchmark"),
      purpose = "to measure hyperparameter optimisation time",
      message_type = "backend_warning"
    )
  }

  # hyperparameter_learner -----------------------------------------------------
  # Hyperparameter learner
  settings$hpo_hyperparameter_learner <- .parse_arg(
    x_config = config$hyperparameter_learner,
    x_var = hyperparameter_learner,
    var_name = "hyperparameter_learner",
    type = "character",
    optional = TRUE,
    default = "gaussian_process"
  )

  .check_parameter_value_is_valid(
    x = settings$hpo_hyperparameter_learner,
    var_name = "hyperparameter_learner",
    values = .get_available_hyperparameter_learners()
  )

  .check_hyperparameter_learner_available(
    hyperparameter_learner = settings$hpo_hyperparameter_learner
  )


  # optimisation_function ------------------------------------------------------
  # Objective function
  settings$hpo_optimisation_function <- .parse_arg(
    x_config = config$optimisation_function,
    x_var = optimisation_function,
    var_name = "optimisation_function",
    type = "character",
    optional = TRUE,
    default = "validation"
  )

  .check_parameter_value_is_valid(
    x = settings$hpo_optimisation_function,
    var_name = "optimisation_function",
    values = .get_available_optimisation_functions(
      hyperparameter_learner = settings$hpo_hyperparameter_learner
    )
  )


  # acquisition_function -------------------------------------------------------
  # Acquisition function
  settings$hpo_acquisition_function <- .parse_arg(
    x_config = config$acquisition_function,
    x_var = acquisition_function,
    var_name = "acquisition_function",
    type = "character",
    optional = TRUE,
    default = "expected_improvement"
  )

  .check_parameter_value_is_valid(
    x = settings$hpo_acquisition_function,
    var_name = "acquisition_function",
    values = .get_available_acquisition_functions()
  )


  # exploration_method ---------------------------------------------------------
  # Exploration method
  settings$hpo_exploration_method <- .parse_arg(
    x_config = config$exploration_method,
    x_var = exploration_method,
    var_name = "exploration_method",
    type = "character",
    optional = TRUE,
    default = "single_shot"
  )

  .check_parameter_value_is_valid(
    x = settings$hpo_exploration_method,
    var_name = "exploration_method",
    values = .get_available_hyperparameter_exploration_methods()
  )

  # optimisation_metric --------------------------------------------------------
  # Performance metric for hyperparameter optimisation
  settings$hpo_metric <- .parse_arg(
    x_config = config$optimisation_metric,
    x_var = optimisation_metric,
    var_name = "optimisation_metric",
    type = "character_list",
    optional = TRUE,
    default = NULL
  )

  # Set default metric
  if (is.null(settings$hpo_metric)) {
    settings$hpo_metric <- .get_default_metric(outcome_type = outcome_type)
  }

  # Check if the metric is ok. Packed into a for loop to enable multi-objective
  # optimisation.
  sapply(
    settings$hpo_metric,
    .check_metric_outcome_type,
    outcome_type = outcome_type
  )

  # parallel_hyperparameter_optimisation ---------------------------------------
  # Parallelisation switch for parallel processing
  settings$do_parallel <- .parse_arg(
    x_config = config$parallel_hyperparameter_optimisation,
    x_var = parallel_hyperparameter_optimisation,
    var_name = "parallel_hyperparameter_optimisation",
    type = "character",
    optional = TRUE,
    default = "TRUE"
  )

  .check_parameter_value_is_valid(
    x = settings$do_parallel,
    var_name = "parallel_hyperparameter_optimisation",
    values = c("TRUE", "FALSE", "inner", "outer")
  )

  # Disable if parallel is FALSE
  if (!parallel) settings$do_parallel <- "FALSE"

  return(settings)
}





#' Internal function for parsing settings related to model evaluation
#'
#' @param config A list of settings, e.g. from an xml file.
#' @param data Data set as loaded using the `.load_data` function.
#' @param parallel Logical value that whether familiar uses parallelisation. If
#'  `FALSE` it will override `parallel_evaluation`.
#' @param outcome_type Type of outcome found in the data set.
#' @param hpo_metric Metric defined for hyperparameter optimisation.
#' @param development_batch_id Identifiers of batches used for model development.
#'  These identifiers are used to determine the cohorts used to determine a
#'  setting for `time_max`, if the `outcome_type` is `survival`, and both
#'  `time_max` and `evaluation_times` are not provided.
#' @param vimp_aggregation_method Variable importance aggregation method.
#' @param vimp_aggregation_rank_threshold Rank threshold used for variable
#'   importance aggregation.
#' @param prep_cluster_method Cluster method used during pre-processing.
#' @param prep_cluster_linkage_method Cluster linkage method used during
#'  pre-processing.
#' @param prep_cluster_cut_method Cluster cut method used during pre-processing.
#' @param prep_cluster_similarity_threshold Cluster similarity threshold used
#'  during pre-processing.
#' @param prep_cluster_similarity_metric Cluster similarity metric used during
#'  pre-processing.
#' @param evaluate_top_level_only (*optional*) Flag that signals that only
#'  evaluation at the most global experiment level is required. Consider a
#'  cross-validation experiment with additional external validation. The global
#'  experiment level consists of data that are used for development, internal
#'  validation and external validation. The next lower experiment level are the
#'  individual cross-validation iterations.
#'
#'  When the flag is `true`, evaluations take place on the global level only,
#'  and no results are generated for the next lower experiment levels. In our
#'  example, this means that results from individual cross-validation iterations
#'  are not computed and shown. When the flag is `false`, results are computed
#'  from both the global layer and the next lower level.
#'
#'  Setting the flag to `true` saves computation time.
#' @param evaluation_elements (*optional*) Specifies which evaluation steps
#'  should be performed during the evaluation process. Defaults to `all`,
#'  indicating that all evaluation steps should be performed. It can have one or
#'  more of the following values:
#'
#'  * `all`, `true`: all evaluation steps are performed.
#'  
#'  * `none`, `false`: no evaluation is performed.
#'
#'  * `auc_data`: data for assessing and plotting the area under the receiver
#'  operating characteristic curve are computed.
#'
#'  * `calibration_data`: data for assessing and plotting model calibration are
#'  computed.
#'
#'  * `calibration_info`: data required to assess calibration, such as baseline
#'  survival curves, are collected. These data will still be present in the
#'  models.
#'
#'  * `confusion_matrix`: data for assessing and plotting a confusion matrix are
#'  collected.
#'
#'  * `decision_curve_analyis`: data for performing a decision curve analysis
#'  are computed.
#'
#'  * `feature_expressions`: data for assessing and plotting sample clustering
#'  are computed.
#'
#'  * `feature_similarity`: data for assessing and plotting feature clusters
#'  not computed.
#'
#'  * `fs_vimp`: data for assessing and plotting feature selection-based
#'  variable importance are collected.
#'
#'  * `hyperparameters`: data for assessing model hyperparameters are 
#'  collected. These data will still be present in the models.
#'
#'  * `ice_data`: data for individual conditional expectation and partial
#'  dependence plots are created.
#'
#'  * `model_performance`: data for assessing and visualising model performance
#'  are created.
#'
#'  * `model_vimp`: data for assessing and plotting model-based variable
#'  importance are collected.
#'
#'  * `permutation_vimp`: data for assessing and plotting model-agnostic
#'  permutation variable importance are computed.
#'
#'  * `prediction_data`: predictions for each sample are made and exported.
#'
#'  * `risk_stratification_data`: data for assessing and plotting Kaplan-Meier
#'  survival curves are collected.
#'
#'  * `risk_stratification_info`: data for assessing stratification into risk
#'  groups are computed.
#'  
#'  * `sample_similarity`: data for assessing sample similarity are computed.
#'  
#'  * `shap`: data for assessing marginal contributions of feature values in
#'  a SHAP analysis.
#'  
#'  * `univariate_analysis`: data for assessing and plotting univariate feature
#'  importance are computed.
#'  
#'  The logical intersect of the evaluation elements specified by
#'  `evaluation_elements` and `skip_evaluation_elements` is used to define which
#'  evaluation steps are performed.
#' @param skip_evaluation_elements (*optional*) Specifies which evaluation steps,
#'  if any, should be skipped as part of the evaluation process. Defaults to
#'  `none`, which means that all relevant evaluation steps are performed. It can
#'  have one or more of the same values as `evaluation_elements`.
#'  
#'  The logical intersect of the evaluation elements specified by
#'  `evaluation_elements` and `skip_evaluation_elements` is used to define which
#'  evaluation steps are performed.
#'
#' @param ensemble_method (*optional*) Method for ensembling predictions from
#'  models for the same sample. Available methods are:
#'
#'  * `median` (default): Use the median of the predicted values as the ensemble
#'  value for a sample.
#'
#'  * `mean`: Use the mean of the predicted values as the ensemble value for a
#'  sample.
#'
#'  This parameter is only used if `detail_level` is `ensemble`.
#'
#' @param evaluation_metric (*optional*) One or more metrics for assessing model
#'  performance. See the vignette on performance metrics for the available
#'  metrics.
#'
#'  Confidence intervals (or rather credibility intervals) are computed for each
#'  metric during evaluation. This is done using bootstraps, the number of which
#'  depends on the value of `confidence_level` (Davison and Hinkley, 1997).
#'
#'  If unset, the metric in the `optimisation_metric` variable is used.
#'
#' @param sample_limit (*optional*) Set the upper limit of the number of samples
#'  that are used during evaluation steps. Cannot be fewer than 20.
#'
#'  This setting can be specified per data element by providing a parameter
#'  value in a named list with data elements, e.g.
#'  `list("sample_similarity"=100, "permutation_vimp"=1000)`.
#'
#'  This parameter can be set for the following data elements:
#'  `sample_similarity`, `shap`, `permutation_vimp`, and `ice_data`.
#'  
#' @param n_important_features (*optional*) Set the number of features that are
#'  evaluated in evaluation steps. Cannot be 0 or fewer.
#'  
#'  This setting can be specified per data element by providing a parameter
#'  value in a named list with data elements, e.g.
#'  `list("ice_data"=10, "permutation_vimp"=5)`.
#'
#'  This parameter can be set for the following data elements:
#'  `ice_data`, `permutation_vimp`, and `shap`.
#'  
#' @param detail_level (*optional*) Sets the level at which results are computed
#'  and aggregated.
#'
#'  * `ensemble`: Results are computed at the ensemble level, i.e. over all
#'  models in the ensemble. This means that, for example, bias-corrected
#'  estimates of model performance are assessed by creating (at least) 20
#'  bootstraps and computing the model performance of the ensemble model for
#'  each bootstrap.
#'
#'  * `hybrid` (default): Results are computed at the level of models in an
#'  ensemble. This means that, for example, bias-corrected estimates of model
#'  performance are directly computed using the models in the ensemble. If there
#'  are at least 20 trained models in the ensemble, performance is computed for
#'  each model, in contrast to `ensemble` where performance is computed for the
#'  ensemble of models. If there are less than 20 trained models in the
#'  ensemble, bootstraps are created so that at least 20 point estimates can be
#'  made.
#'
#'  * `model`: Results are computed at the model level. This means that, for
#'  example, bias-corrected estimates of model performance are assessed by
#'  creating (at least) 20 bootstraps and computing the performance of the model
#'  for each bootstrap.
#'
#'  Note that each level of detail has a different interpretation for bootstrap
#'  confidence intervals. For `ensemble` and `model` these are the confidence
#'  intervals for the ensemble and an individual model, respectively. That is,
#'  the confidence interval describes the range where an estimate produced by a
#'  respective ensemble or model trained on a repeat of the experiment may be
#'  found with the probability of the confidence level. For `hybrid`, it
#'  represents the range where any single model trained on a repeat of the
#'  experiment may be found with the probability of the confidence level. By
#'  definition, confidence intervals obtained using `hybrid` are at least as
#'  wide as those for `ensemble`. `hybrid` offers the correct interpretation if
#'  the goal of the analysis is to assess the result of a single, unspecified,
#'  model.
#'
#'  `hybrid` is generally computationally less expensive then `ensemble`, which
#'  in turn is somewhat less expensive than `model`.
#'
#'  A non-default `detail_level` parameter can be specified for separate
#'  evaluation steps by providing a parameter value in a named list with data
#'  elements, e.g. `list("auc_data"="ensemble", "model_performance"="hybrid")`.
#'  This parameter can be set for the following data elements: `auc_data`,
#'  `decision_curve_analyis`, `model_performance`, `permutation_vimp`,
#'  `ice_data`, `prediction_data` and `confusion_matrix`.
#'  
#'  If results are computed from 10 samples or fewer, `ensemble` is
#'  automatically used. This prevents issues where evaluation steps do not have
#'  a required minimum number of samples for `hybrid` or `model`.
#'
#' @param estimation_type (*optional*) Sets the type of estimation that should be
#'  possible. This has the following options:
#'
#'  * `point`: Point estimates.
#'
#'  * `bias_correction` or `bc`: Bias-corrected estimates. A bias-corrected
#'  estimate is computed from (at least) 20 point estimates, and `familiar` may
#'  bootstrap the data to create them.
#'
#'  * `bootstrap_confidence_interval` or `bci` (default): Bias-corrected
#'  estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
#'  number of point estimates required depends on the `confidence_level`
#'  parameter, and `familiar` may bootstrap the data to create them.
#'
#'  As with `detail_level`, a non-default `estimation_type` parameter can be
#'  specified for separate evaluation steps by providing a parameter value in a
#'  named list with data elements, e.g. `list("auc_data"="bci",
#'  "model_performance"="point")`. This parameter can be set for the following
#'  data elements: `auc_data`, `decision_curve_analyis`, `model_performance`,
#'  `permutation_vimp`, `ice_data`, and `prediction_data`.
#'
#' @param aggregate_results (*optional*) Flag that signifies whether results
#'  should be aggregated during evaluation. If `estimation_type` is
#'  `bias_correction` or `bc`, aggregation leads to a single bias-corrected
#'  estimate. If `estimation_type` is `bootstrap_confidence_interval` or `bci`,
#'  aggregation leads to a single bias-corrected estimate with lower and upper
#'  boundaries of the confidence interval. This has no effect if
#'  `estimation_type` is `point`.
#'
#'  The default value is equal to `TRUE` except when assessing metrics to assess
#'  model performance, as the default violin plot requires underlying data.
#'
#'  As with `detail_level` and `estimation_type`, a non-default
#'  `aggregate_results` parameter can be specified for separate evaluation steps
#'  by providing a parameter value in a named list with data elements, e.g.
#'  `list("auc_data"=TRUE, , "model_performance"=FALSE)`. This parameter exists
#'  for the same elements as `estimation_type`.
#'
#' @param confidence_level (*optional*) Numeric value for the level at which
#'  confidence intervals are determined. In the case bootstraps are used to
#'  determine the confidence intervals bootstrap estimation, `familiar` uses the
#'  rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
#'  bootstraps.
#'
#'  The default value is `0.95`.
#'
#' @param bootstrap_ci_method (*optional*) Method used to determine bootstrap
#'  confidence intervals (Efron and Hastie, 2016). The following methods are
#'  implemented:
#'
#'  * `percentile` (default): Confidence intervals obtained using the percentile
#'  method.
#'
#'  * `bc`: Bias-corrected confidence intervals.
#'
#'  Note that the standard method is not implemented because this method is
#'  often not suitable due to non-normal distributions. The bias-corrected and
#'  accelerated (BCa) method is not implemented yet.
#'
#' @param shap_tolerance (*optional*) Relative tolerance for convergence of SHAP
#'   values. The tolerance is scaled with the range in SHAP values.
#'   
#'   The default value is `0.05`.
#'   
#' @param shap_max_iterations (*optional*) Maximum iterations for convergence of
#'   SHAP values.
#' 
#'   The default value is `1000` iterations.
#'
#' @param shap_phi_0 (*optional*) Value(s) to use for computing marginal
#'   contributions. By default, the average predicted value in the population is
#'   used. For `multinomial` and `survival` outcomes, this parameter requires a
#'   value for each class and evaluation time (`evaluation_times`),
#'   respectively.
#'
#' @param feature_cluster_method (*optional*) Method used to perform clustering
#'  of features. The same methods as for the `cluster_method` configuration
#'  parameter are available: `none`, `hclust`, `agnes`, `diana` and `pam`.
#'
#'  The value for the `cluster_method` configuration parameter is used by
#'  default. When generating clusters for the purpose of determining mutual
#'  correlation and ordering feature expressions, `none` is ignored and `hclust`
#'  is used instead.
#'
#' @param feature_linkage_method (*optional*) Method used for agglomerative
#'  clustering with `hclust` and `agnes`. Linkage determines how features are
#'  sequentially combined into clusters based on distance. The methods are
#'  shared with the `cluster_linkage_method` configuration parameter: `average`,
#'  `single`, `complete`, `weighted`, and `ward`.
#'
#'  The value for the `cluster_linkage_method` configuration parameters is used
#'  by default.
#'
#' @param feature_cluster_cut_method (*optional*) Method used to divide features
#'  into separate clusters. The available methods are the same as for the
#'  `cluster_cut_method` configuration parameter: `silhouette`, `fixed_cut` and
#'  `dynamic_cut`.
#'
#'  `silhouette` is available for all cluster methods, but `fixed_cut` only
#'  applies to methods that create hierarchical trees (`hclust`, `agnes` and
#'  `diana`). `dynamic_cut` requires the `dynamicTreeCut` package and can only
#'  be used with `agnes` and `hclust`.
#'
#'  The value for the `cluster_cut_method` configuration parameter is used by
#'  default.
#'
#' @param feature_similarity_metric (*optional*) Metric to determine pairwise
#'  similarity between features. Similarity is computed in the same manner as
#'  for clustering, and `feature_similarity_metric` therefore has the same
#'  options as `cluster_similarity_metric`: `mcfadden_r2`, `cox_snell_r2`,
#'  `nagelkerke_r2`, `mutual_information`, `spearman`, `kendall` and `pearson`.
#'
#'  The value used for the `cluster_similarity_metric` configuration parameter
#'  is used by default.
#'
#' @param feature_similarity_threshold (*optional*) The threshold level for
#'  pair-wise similarity that is required to form feature clusters with the
#'  `fixed_cut` method. This threshold functions in the same manner as the one
#'  defined using the `cluster_similarity_threshold` parameter.
#'
#'  By default, the value for the `cluster_similarity_threshold` configuration
#'  parameter is used.
#'
#'  Unlike for `cluster_similarity_threshold`, more than one value can be
#'  supplied here.
#'
#' @param sample_cluster_method (*optional*) The method used to perform
#'  clustering based on distance between samples. These are the same methods as
#'  for the `cluster_method` configuration parameter: `hclust`, `agnes`, `diana`
#'  and `pam`.
#'
#'  The value for the `cluster_method` configuration parameter is used by
#'  default. When generating clusters for the purpose of ordering samples in
#'  feature expressions, `none` is ignored and `hclust` is used instead.
#'
#' @param sample_linkage_method (*optional*) The method used for agglomerative
#'  clustering in `hclust` and `agnes`. These are the same methods as for the
#'  `cluster_linkage_method` configuration parameter: `average`, `single`,
#'  `complete`, `weighted`, and `ward`.
#'
#'  The value for the `cluster_linkage_method` configuration parameters is used
#'  by default.
#'
#' @param sample_similarity_metric (*optional*) Metric to determine pairwise
#'  similarity between samples. Similarity is computed in the same manner as for
#'  clustering, but `sample_similarity_metric` has different options that are
#'  better suited to computing distance between samples instead of between
#'  features. The following metrics are available.
#'
#'  * `gower` (default): compute Gower's distance between samples. By default,
#'  Gower's distance is computed based on winsorised data to reduce the effect
#'  of outliers (see below).
#'
#'  * `euclidean`: compute the Euclidean distance between samples.
#'
#'  The underlying feature data for numerical features is scaled to the
#'  \eqn{[0,1]} range using the feature values across the samples. The
#'  normalisation parameters required can optionally be computed from feature
#'  data with the outer 5% (on both sides) of feature values trimmed or
#'  winsorised. To do so append `_trim` (trimming) or `_winsor` (winsorising) to
#'  the metric name. This reduces the effect of outliers somewhat.
#'
#'  Regardless of metric, all categorical features are handled as for the
#'  Gower's distance: distance is 0 if the values in a pair of samples match,
#'  and 1 if they do not.
#'
#' @param eval_aggregation_method (*optional*) Method for aggregating variable
#'  importances for the purpose of evaluation. Variable importances are
#'  determined both during initial variable importance computation steps and
#'  after training the model. Both types are evaluated after training the
#'  model, but only the initial variable importances are evaluated at run-time.
#'
#'  See the documentation for the `vimp_aggregation_method` argument for
#'  information concerning the different methods available.
#' @param eval_aggregation_rank_threshold (*optional*) The threshold used to
#'  define the subset of highly important features during evaluation.
#'
#'  See the documentation for the `vimp_aggregation_rank_threshold` argument for
#'  more information.
#' @param eval_icc_type (*optional*) String indicating the type of intraclass
#'  correlation coefficient (`1`, `2` or `3`) that should be used to compute
#'  robustness for features in repeated measurements during the evaluation of
#'  univariate importance. These types correspond to the types in Shrout and
#'  Fleiss (1979). The default value is `1`.
#' @param stratification_method (*optional*) Method for determining the
#'  stratification threshold for creating survival groups. The actual,
#'  model-dependent, threshold value is obtained from the development data, and
#'  can afterwards be used to perform stratification on validation data.
#'
#'  The following stratification methods are available:
#'
#'  * `median` (default): The median predicted value in the development cohort
#'  is used to stratify the samples into two risk groups. For predicted outcome
#'  values that build a continuous spectrum, the two risk groups in the
#'  development cohort will be roughly equal in size.
#'
#'  * `mean`: The mean predicted value in the development cohort is used to
#'  stratify the samples into two risk groups.
#'
#'  * `mean_trim`: As `mean`, but based on the set of predicted values
#'  where the 5% lowest and 5% highest values are discarded. This reduces the
#'  effect of outliers.
#'
#'  * `mean_winsor`: As `mean`, but based on the set of predicted values where
#'  the 5% lowest and 5% highest values are winsorised. This reduces the effect
#'  of outliers.
#'
#'  * `fixed`: Samples are stratified based on the sample quantiles of the
#'  predicted values. These quantiles are defined using the
#'  `stratification_threshold` parameter.
#'
#'  * `optimised`: Use maximally selected rank statistics to determine the
#'  optimal threshold (Lausen and Schumacher, 1992; Hothorn et al., 2003) to
#'  stratify samples into two optimally separated risk groups.
#'
#'  One or more stratification methods can be selected simultaneously.
#'
#'  This parameter is only relevant for `survival` outcomes.
#' @param stratification_threshold (*optional*) Numeric value(s) signifying the
#'  sample quantiles for stratification using the `fixed` method. The number of
#'  risk groups will be the number of values +1.
#'
#'  The default value is `c(1/3, 2/3)`, which will yield two thresholds that
#'  divide samples into three equally sized groups. If `fixed` is not among the
#'  selected stratification methods, this parameter is ignored.
#'
#'  This parameter is only relevant for `survival` outcomes.
#'
#' @param time_max (*optional*) Time point which is used as the benchmark for
#'  e.g. cumulative risks generated by random forest, or the cutoff for Uno's
#'  concordance index.
#'
#'  If `time_max` is not provided, but `evaluation_times` is, the largest value
#'  of `evaluation_times` is used. If both are not provided, `time_max` is set
#'  to the 98th percentile of the distribution of survival times for samples
#'  with an event in the development data set.
#'
#'  This parameter is only relevant for `survival` outcomes.
#' @param evaluation_times (*optional*) One or more time points that are used for
#'  assessing calibration in survival problems. This is done as expected and
#'  observed survival probabilities depend on time.
#'
#'  If unset, `evaluation_times` will be equal to `time_max`.
#'
#'  This parameter is only relevant for `survival` outcomes.
#' @param dynamic_model_loading (*optional*) Enables dynamic loading of models
#'  during the evaluation process, if `TRUE`. Defaults to `FALSE`. Dynamic
#'  loading of models may reduce the overall memory footprint, at the cost of
#'  increased disk or network IO. Models can only be dynamically loaded if they
#'  are found at an accessible disk or network location. Setting this parameter
#'  to `TRUE` may help if parallel processing causes out-of-memory issues during
#'  evaluation.
#' @param parallel_evaluation (*optional*) Enable parallel processing for
#'  hyperparameter optimisation. Defaults to `TRUE`. When set to `FALSE`, this
#'  will disable the use of parallel processing while performing optimisation,
#'  regardless of the settings of the `parallel` parameter. The parameter
#'  moreover specifies whether parallelisation takes place within the evaluation
#'  process steps (`inner`, default), or in an outer loop ( `outer`) over
#'  learners, data subsamples, etc.
#'
#'  `parallel_evaluation` is ignored if `parallel=FALSE`.
#' @param ... Unused arguments.
#'
#' @return List of parameters related to model evaluation.
#'
#' @references 1. Davison, A. C. & Hinkley, D. V. Bootstrap methods and their
#'  application. (Cambridge University Press, 1997).
#'
#'  1. Efron, B. & Hastie, T. Computer Age Statistical Inference. (Cambridge
#'  University Press, 2016).
#'
#'  1. Lausen, B. & Schumacher, M. Maximally Selected Rank Statistics.
#'  Biometrics 48, 73 (1992).
#'
#'  1. Hothorn, T. & Lausen, B. On the exact distribution of maximally selected
#'  rank statistics. Comput. Stat. Data Anal. 43, 121–137 (2003).
#' @md
#' @keywords internal
.parse_evaluation_settings <- function(
    config = NULL,
    data,
    parallel,
    outcome_type,
    hpo_metric,
    development_batch_id,
    vimp_aggregation_method,
    vimp_aggregation_rank_threshold,
    prep_cluster_method,
    prep_cluster_linkage_method,
    prep_cluster_cut_method,
    prep_cluster_similarity_threshold,
    prep_cluster_similarity_metric,
    evaluate_top_level_only = waiver(),
    evaluation_elements = waiver(),
    skip_evaluation_elements = waiver(),
    ensemble_method = waiver(),
    evaluation_metric = waiver(),
    sample_limit = waiver(),
    n_important_features = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    shap_tolerance = waiver(),
    shap_max_iterations = waiver(),
    shap_phi_0 = waiver(),
    feature_cluster_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_linkage_method = waiver(),
    feature_similarity_metric = waiver(),
    feature_similarity_threshold = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    eval_aggregation_method = waiver(),
    eval_aggregation_rank_threshold = waiver(),
    eval_icc_type = waiver(),
    stratification_method = waiver(),
    stratification_threshold = waiver(),
    time_max = waiver(),
    evaluation_times = waiver(),
    dynamic_model_loading = waiver(),
    parallel_evaluation = waiver(),
    ...
) {
  settings <- list()
  
  # pool_only ------------------------------------------------------------------
  # Flag that limits the depth of the evaluation.
  settings$pool_only <- .parse_arg(
    x_config = config$evaluate_top_level_only,
    x_var = evaluate_top_level_only,
    var_name = "evaluate_top_level_only",
    type = "logical",
    optional = TRUE,
    default = TRUE
  )

  # evaluation_elements --------------------------------------------------------
  keep_evaluation_elements <- .parse_arg(
    x_config = config$evaluation_elements,
    x_var = evaluation_elements,
    var_name = "evaluation_elements",
    type = "character_list",
    optional = TRUE,
    default = "all"
  )
  
  keep_evaluation_elements <- tolower(keep_evaluation_elements)
  .check_parameter_value_is_valid(
    x = keep_evaluation_elements,
    var_name = "evaluation_elements",
    values = c(.get_available_data_elements(), "none", "false", "all", "true")
  )
  
  if (any(keep_evaluation_elements %in% c("all", "true"))) {
    keep_evaluation_elements <- .get_available_data_elements()
  }
  
  if (any(keep_evaluation_elements %in% c("none", "false"))) {
    keep_evaluation_elements <- NULL
  }
  
  # skip_evaluation_elements ---------------------------------------------------
  # Specify any specific elements of the evaluation to skip.
  skipped_evaluation_elements <- .parse_arg(
    x_config = config$skip_evaluation_elements,
    x_var = skip_evaluation_elements,
    var_name = "skip_evaluation_elements",
    type = "character_list",
    optional = TRUE,
    default = "none"
  )

  skipped_evaluation_elements <- tolower(skipped_evaluation_elements)
  .check_parameter_value_is_valid(
    x = skipped_evaluation_elements,
    var_name = "skip_evaluation_elements",
    values = c(.get_available_data_elements(), "none", "false", "all", "true")
  )

  if (any(skipped_evaluation_elements %in% c("all", "true"))) {
    skipped_evaluation_elements <- .get_available_data_elements()
  }

  if (any(skipped_evaluation_elements %in% c("none", "false"))) {
    skipped_evaluation_elements <- NULL
  }

  # Instead of specifying the elements to skip, we specify the elements to keep.
  skipped_evaluation_elements <- setdiff(
    .get_available_data_elements(),
    skipped_evaluation_elements
  )
  
  if ("calibration_data" %in% settings$evaluation_data_elements) {
    require_package(
      x = "harmonicmeanp",
      purpose = "to compute p-values for model calibration tests",
      message_type = "backend_warning"
    )
  }

  settings$evaluation_data_elements <- intersect(
    keep_evaluation_elements, skipped_evaluation_elements
  )
  
  if (length(settings$evaluation_data_elements) == 0L) {
    settings$evaluation_data_elements <- NULL
  }

  # Check whether plotting packages are available if any data elements are
  # computed.
  if (!is.null(settings$evaluation_data_element)) {
    require_package(
      x = ..required_plotting_packages(extended = TRUE),
      purpose = "to create plots",
      message_type = "backend_warning"
    )
  }

  # ensemble_method ------------------------------------------------------------
  # Method for ensemble predictions
  settings$ensemble_method <- .parse_arg(
    x_config = config$ensemble_method,
    x_var = ensemble_method,
    var_name = "ensemble_method",
    type = "character",
    optional = TRUE,
    default = "median"
  )

  .check_parameter_value_is_valid(
    x = settings$ensemble_method,
    var_name = "ensemble_method",
    values = .get_available_ensemble_prediction_methods()
  )

  # evaluation_metric ----------------------------------------------------------
  # List of performance metrics for evaluation
  settings$metric <- .parse_arg(
    x_config = config$evaluation_metric,
    x_var = evaluation_metric,
    var_name = "evaluation_metric",
    type = "character_list",
    optional = TRUE,
    default = hpo_metric
  )

  sapply(
    settings$metric,
    .check_metric_outcome_type, 
    outcome_type = outcome_type
  )

  # sample_limit ---------------------------------------------------------------
  # Number of samples that should be analysed.
  settings$sample_limit <- .parse_arg(
    x_config = config$sample_limit,
    x_var = sample_limit,
    var_name = "sample_limit",
    type = "list",
    optional = TRUE,
    default = list()
  )

  if (length(settings$sample_limit) == 0L) {
    # Default - use method-specific settings.
    settings$sample_limit <- NULL
    
  } else if (length(settings$sample_limit) == 1L && is.null(names(settings$sample_limit))) {
    # Check that the contents are a correctly specified. At least 20 samples
    # should be present.
    .check_number_in_valid_range(
      x = settings$sample_limit[[1L]],
      var_name = "sample_limit",
      range = c(20L, Inf)
    )

    # Add provided sample limit to each possible element.
    settings$sample_limit <- lapply(
      .get_available_data_elements(check_has_sample_limit = TRUE),
      function(x, sample_limit) (sample_limit),
      sample_limit = settings$sample_limit[[1L]]
    )

    # Add name of respective data elements.
    names(settings$sample_limit) <- .get_available_data_elements(
      check_has_sample_limit = TRUE
    )
    
  } else {
    # Check that the list elements are correctly specified.
    sapply(
      names(settings$sample_limit),
      .check_parameter_value_is_valid,
      var_name = "sample_limit (data element name)",
      values = .get_available_data_elements(check_has_sample_limit = TRUE)
    )

    # Check that the list contents are correctly specified.
    sapply(
      names(settings$sample_limit),
      function(element_name, x) {
        .check_number_in_valid_range(
          x = x[[element_name]],
          var_name = paste0("sample_limit (", element_name, ")"),
          range = c(20L, Inf)
        )
      },
      x = settings$sample_limit
    )
  }
  
  # n_important_features -------------------------------------------------------
  # Number of important features
  settings$n_important_features <- .parse_arg(
    x_config = config$n_important_features,
    x_var = n_important_features,
    var_name = "n_important_features",
    type = "list",
    optional = TRUE,
    default = list()
  )
  
  if (length(settings$n_important_features) == 0L) {
    # Default - use method-specific settings.
    settings$n_important_features <- NULL
    
  } else if (length(settings$n_important_features) == 1L && is.null(names(settings$n_important_features))) {
    # Check that the parameter has values between 1 and Inf.
    .check_number_in_valid_range(
      x = settings$n_important_features[[1L]],
      var_name = "n_important_features",
      range = c(1L, Inf)
    )
    
    # Add provided sample limit to each possible element.
    settings$n_important_features <- lapply(
      .get_available_data_elements(check_has_n_important_features = TRUE),
      function(x, n_important_features) (n_important_features),
      n_important_features = settings$n_important_features[[1L]]
    )
    
    # Add name of respective data elements.
    names(settings$n_important_features) <- .get_available_data_elements(
      check_has_n_important_features = TRUE
    )
    
  } else {
    # Check that the list elements are correctly specified.
    sapply(
      names(settings$n_important_features),
      .check_parameter_value_is_valid,
      var_name = "n_important_features (data element name)",
      values = .get_available_data_elements(check_has_sample_limit = TRUE)
    )
    
    # Check that the list contents are correctly specified.
    sapply(
      names(settings$n_important_features),
      function(element_name, x) {
        .check_number_in_valid_range(
          x = x[[element_name]],
          var_name = paste0("n_important_features (", element_name, ")"),
          range = c(1L, Inf)
        )
      },
      x = settings$sample_limit
    )
  }

  # detail_level ---------------------------------------------------------------
  # Level at which evaluations are computed.
  settings$detail_level <- .parse_arg(
    x_config = config$detail_level,
    x_var = detail_level,
    var_name = "detail_level",
    type = "list",
    optional = TRUE,
    default = list()
  )

  if (length(settings$detail_level) == 0L) {
    # Default - use method-specific settings.
    settings$detail_level <- NULL
    
  } else if (
    length(settings$detail_level) == 1L && 
    is.null(names(settings$detail_level))
  ) {
    # Check that the contents are a correctly specified, single string.
    .check_parameter_value_is_valid(
      x = settings$detail_level[[1L]],
      var_name = "detail_level",
      values = c("ensemble", "hybrid", "model")
    )

    # Add provided detail level to each possible element.
    settings$detail_level <- lapply(
      .get_available_data_elements(check_has_detail_level = TRUE),
      function(x, detail_level) (detail_level),
      detail_level = settings$detail_level[[1L]]
    )

    # Add name of respective data elements.
    names(settings$detail_level) <- .get_available_data_elements(
      check_has_detail_level = TRUE
    )
    
  } else {
    # Check that the list elements are correctly specified.
    sapply(
      names(settings$detail_level),
      .check_parameter_value_is_valid,
      var_name = "detail_level (data element name)",
      values = .get_available_data_elements(check_has_detail_level = TRUE)
    )

    # Check that the list contents are correctly specified.
    sapply(
      names(settings$detail_level),
      function(element_name, x) {
        .check_parameter_value_is_valid(
          x = x[[element_name]],
          var_name = paste0("detail_level (", element_name, ")"),
          values = c("ensemble", "hybrid", "model")
        )
      },
      x = settings$detail_level
    )
  }
  
  # estimation_type ------------------------------------------------------------
  # Type of estimation performed.
  settings$estimation_type <- .parse_arg(
    x_config = config$estimation_type,
    x_var = estimation_type,
    var_name = "estimation_type",
    type = "list",
    optional = TRUE,
    default = list()
  )

  if (length(settings$estimation_type) == 0L) {
    # Default - use method-specific settings.
    settings$estimation_type <- NULL
    
  } else if (
    length(settings$estimation_type) == 1L &&
    is.null(names(settings$estimation_type))
  ) {
    # Check that the contents are a correctly specified, single string.
    .check_parameter_value_is_valid(
      x = settings$estimation_type[[1L]],
      var_name = "estimation_type",
      values = c(
        "point", "bias_correction", "bc",
        "bootstrap_confidence_interval", "bci"
      )
    )

    # Add provided estimation type to each possible element.
    settings$estimation_type <- lapply(
      .get_available_data_elements(check_has_estimation_type = TRUE),
      function(x, estimation_type) (estimation_type),
      estimation_type = settings$estimation_type[[1L]]
    )
    
    # Add name of respective data elements.
    names(settings$estimation_type) <- .get_available_data_elements(
      check_has_estimation_type = TRUE
    )
    
  } else {
    # Check that the list elements are correctly specified.
    sapply(
      names(settings$estimation_type),
      .check_parameter_value_is_valid,
      var_name = "estimation_type (data element name)",
      values = .get_available_data_elements(check_has_estimation_type = TRUE)
    )

    # Check that the list contents are correctly specified.
    sapply(
      names(settings$estimation_type),
      function(element_name, x) {
        .check_parameter_value_is_valid(
          x = x[[element_name]],
          var_name = paste0("estimation_type (", element_name, ")"),
          values = c(
            "point", "bias_correction", "bc",
            "bootstrap_confidence_interval", "bci"
          )
        )
      },
      x = settings$estimation_type
    )
  }
  
  # aggregate_results ----------------------------------------------------------
  # Aggregate results.
  settings$aggregate_results <- .parse_arg(
    x_config = config$aggregate_results,
    x_var = aggregate_results,
    var_name = "aggregate_results",
    type = "list",
    optional = TRUE,
    default = list()
  )

  if (length(settings$aggregate_results) == 0L) {
    # Default - use method-specific settings.
    settings$aggregate_results <- NULL
    
  } else if (
    length(settings$aggregate_results) == 1L &&
    is.null(names(settings$aggregate_results))
  ) {
    # Check that the contents are a correctly specified, single string.
    .check_parameter_value_is_valid(
      x = tolower(settings$aggregate_results[[1L]]),
      var_name = "aggregate_results",
      values = c("true", "false", "none", "all", "default")
    )

    # Add provided aggregate_results value to each possible element.
    settings$aggregate_results <- lapply(
      .get_available_data_elements(check_has_estimation_type = TRUE),
      function(x, aggregate_results) (aggregate_results),
      aggregate_results = tolower(settings$aggregate_results[[1L]])
    )

    # Add name of respective data elements.
    names(settings$aggregate_results) <- .get_available_data_elements(
      check_has_estimation_type = TRUE
    )
    
  } else {
    # Check that the list elements are correctly specified.
    sapply(
      names(settings$aggregate_results),
      .check_parameter_value_is_valid,
      var_name = "aggregate_results (data element name)",
      values = .get_available_data_elements(check_has_estimation_type = TRUE)
    )

    # Check that the list contents are correctly specified.
    sapply(
      names(settings$aggregate_results),
      function(element_name, x) {
        .check_parameter_value_is_valid(
          x = tolower(x[[element_name]]),
          var_name = paste0("aggregate_results (", element_name, ")"),
          values = c("true", "false", "none", "all", "default")
        )
      },
      x = settings$aggregate_results
    )
  }

  # bootstrap_ci_method --------------------------------------------------------
  # Bootstrap confidence interval.
  settings$bootstrap_ci_method <- .parse_arg(
    x_config = config$bootstrap_ci_method,
    x_var = bootstrap_ci_method,
    var_name = "bootstrap_ci_method",
    type = "character",
    optional = TRUE,
    default = "percentile"
  )

  .check_parameter_value_is_valid(
    x = settings$bootstrap_ci_method,
    var_name = "bootstrap_ci_method",
    values = .get_available_bootstrap_confidence_interval_methods()
  )

  # confidence_level -----------------------------------------------------------
  # Width of the confidence intervals
  settings$confidence_level <- .parse_arg(
    x_config = config$confidence_level,
    x_var = confidence_level,
    var_name = "confidence_level",
    type = "numeric",
    optional = TRUE,
    default = 0.95
  )

  .check_number_in_valid_range(
    x = settings$confidence_level,
    var_name = "confidence_level",
    range = c(0.0, 1.0),
    closed = c(FALSE, FALSE)
  )

  # SHAP tolerance -------------------------------------------------------------
  settings$shap_tolerance <- .parse_arg(
    x_config = config$shap_tolerance,
    x_var = shap_tolerance,
    var_name = "shap_tolerance",
    type = "numeric",
    optional = TRUE,
    default = 0.05
  )
  
  .check_number_in_valid_range(
    x = settings$shap_tolerance,
    var_name = "shap_tolerance",
    range = c(0.0, Inf),
    closed = c(FALSE, FALSE)
  )
  
  # SHAP maximum iterations ----------------------------------------------------
  settings$shap_max_iterations <- .parse_arg(
    x_config = config$shap_max_iterations,
    x_var = shap_max_iterations,
    var_name = "shap_max_iterations",
    type = "integer",
    optional = TRUE,
    default = 1000L
  )
  
  .check_number_in_valid_range(
    x = settings$shap_max_iterations,
    var_name = "shap_max_iterations",
    range = c(1L, Inf),
    closed = c(TRUE, TRUE)
  )
  
  # SHAP reference value -------------------------------------------------------
  settings$shap_phi_0 <- .parse_arg(
    x_config = config$shap_phi_0,
    x_var = shap_phi_0,
    var_name = "shap_phi_0",
    type = "numeric_list",
    optional = TRUE,
    default = NULL
  )

  # feature cluster options ----------------------------------------------------
  settings <- c(
    settings,
    .parse_feature_clustering(
      config = config,
      default_cluster_method = prep_cluster_method,
      default_cluster_linkage_method = prep_cluster_linkage_method,
      default_cluster_cut_method = prep_cluster_cut_method,
      default_cluster_similarity_metric = prep_cluster_similarity_metric,
      default_cluster_similarity_threshold = prep_cluster_similarity_threshold,
      feature_cluster_method = feature_cluster_method,
      feature_linkage_method = feature_linkage_method,
      feature_cluster_cut_method = feature_cluster_cut_method,
      feature_similarity_metric = feature_similarity_metric,
      feature_similarity_threshold = feature_similarity_threshold,
      .check_more = any(c("feature_similarity", "univariate_analysis", "feature_expressions", "permutation_vimp") %in% settings$evaluation_data_elements)
    )
  )
  
  # sample cluster options -----------------------------------------------------
  settings <- c(
    settings,
    .parse_sample_clustering(
      config = config,
      default_cluster_method = prep_cluster_method,
      default_cluster_linkage_method = prep_cluster_linkage_method,
      sample_cluster_method = sample_cluster_method,
      sample_linkage_method = sample_linkage_method,
      sample_similarity_metric = sample_similarity_metric,
      .check_more = any(c("sample_similarity", "feature_expressions") %in% settings$evaluation_data_elements)
    )
  )

  # eval_aggregation_method ----------------------------------------------------
  # Variable importance aggregation methods
  settings$aggregation <- .parse_arg(
    x_config = config$eval_aggregation_method,
    x_var = eval_aggregation_method,
    var_name = "eval_aggregation_method",
    type = "character",
    optional = TRUE,
    default = vimp_aggregation_method
  )

  .check_parameter_value_is_valid(
    x = settings$aggregation,
    var_name = "eval_aggregation_method",
    values = .get_available_rank_aggregation_methods()
  )

  # eval_aggregation_rank_threshold --------------------------------------------
  # Variable importance rank threshold (used by some aggregation methods)
  settings$aggr_rank_threshold <- .parse_arg(
    x_config = config$eval_aggregation_rank_threshold,
    x_var = eval_aggregation_rank_threshold,
    var_name = "eval_aggregation_rank_threshold",
    type = "integer",
    optional = TRUE,
    default = vimp_aggregation_rank_threshold
  )

  if (!is.null(settings$aggr_rank_threshold)) {
    .check_number_in_valid_range(
      x = settings$aggr_rank_threshold,
      var_name = "eval_aggregation_rank_threshold",
      range = c(1L, Inf)
    )
  }

  # eval_icc_type --------------------------------------------------------------
  # Type of ICC computed for univariate analysis.
  settings$icc_type <- .parse_arg(
    x_config = config$eval_icc_type,
    x_var = eval_icc_type,
    var_name = "eval_icc_type",
    type = "character",
    optional = TRUE,
    default = "1"
  )

  .check_parameter_value_is_valid(
    x = settings$icc_type,
    var_name = "eval_icc_type",
    values = .get_available_icc_types()
  )

  # stratification_method ------------------------------------------------------
  # Method used to set stratification thresholds for Kaplan-Meier analysis
  settings$strat_method <- .parse_arg(
    x_config = config$stratification_method,
    x_var = stratification_method,
    var_name = "stratification_method",
    type = "character_list",
    optional = TRUE,
    default = "median"
  )

  .check_parameter_value_is_valid(
    x = settings$strat_method,
    var_name = "stratification_method",
    values = .get_available_stratification_methods()
  )

  if ("optimised" %in% settings$strat_method) {
    if (!require_package(
      x = "maxstat",
      purpose = "to determine an optimal risk threshold",
      message_type = "backend_warning"
    )) {
      settings$strat_method <- setdiff(settings$strat_method, "optimised")
      if (length(settings$strat_method) == 0L) settings$strat_method <- "median"
    }
  }

  # stratification_threshold ---------------------------------------------------
  # Quantile stratification thresholds for the "fixed" stratification method.
  # Note that c(0.333, 0.667) means that 3 groups are created: a low risk group
  # (up to 0.333), a moderate risk group (between 0.333 and 0.667), and a high
  # risk group (0.667)
  settings$strat_quant_threshold <- .parse_arg(
    x_config = config$stratification_threshold,
    x_var = stratification_threshold,
    var_name = "stratification_threshold",
    type = "numeric_list",
    optional = TRUE,
    default = c(1.0 / 3.0, 2.0 / 3.0)
  )

  sapply(
    settings$strat_quant_threshold,
    .check_number_in_valid_range,
    var_name = "stratification_threshold",
    range = c(0.0, 1.0),
    closed = c(FALSE, FALSE)
  )

  # evaluation time options ----------------------------------------------------
  settings <- c(
    settings,
    .parse_evaluation_time_settings(
      data = data,
      outcome_type = outcome_type,
      config = config,
      time_max = time_max,
      evaluation_times = evaluation_times,
      development_batch_id = development_batch_id
    )
  )

  # dynamic_model_loading ------------------------------------------------------
  # Dynamic loading of models during evaluation.
  settings$auto_detach <- .parse_arg(
    x_config = config$dynamic_model_loading,
    x_var = dynamic_model_loading,
    var_name = "dynamic_model_loading",
    type = "logical",
    optional = TRUE,
    default = FALSE
  )

  # parallel_evaluation --------------------------------------------------------
  # Parallelisation switch for parallel processing
  settings$do_parallel <- .parse_arg(
    x_config = config$parallel_evaluation,
    x_var = parallel_evaluation,
    var_name = "parallel_evaluation",
    type = "character",
    optional = TRUE,
    default = "TRUE"
  )

  .check_parameter_value_is_valid(
    x = settings$do_parallel,
    var_name = "parallel_evaluation",
    values = c("TRUE", "FALSE", "inner", "outer")
  )

  # Disable if parallel is FALSE
  if (!parallel) settings$do_parallel <- "FALSE"

  # Return list of settings
  return(settings)
}



.parse_feature_clustering <- function(
    config = NULL,
    default_cluster_method = waiver(),
    default_cluster_linkage_method = waiver(),
    default_cluster_cut_method = waiver(),
    default_cluster_similarity_metric = waiver(),
    default_cluster_similarity_threshold = waiver(),
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_metric = waiver(),
    feature_similarity_threshold = waiver(),
    .check_more = TRUE
) {
  settings <- list()
  
  # Set default cluster method.
  if (is.waive(default_cluster_method)) default_cluster_method <- "hclust"
  if (is.waive(default_cluster_linkage_method)) default_cluster_linkage_method <- "average"
  if (is.waive(default_cluster_cut_method)) {
    default_cluster_cut_method <- ifelse(default_cluster_method == "pam", "silhouette", "fixed_cut")
  }
  if (is.waive(default_cluster_similarity_metric)) default_cluster_similarity_metric <- "mutual_information"
  if (is.waive(default_cluster_similarity_threshold)) default_cluster_similarity_threshold <- NULL
  
  # feature_cluster_method -----------------------------------------------------
  # Feature cluster method
  settings$feature_cluster_method <- .parse_arg(
    x_config = config$feature_cluster_method,
    x_var = feature_cluster_method,
    var_name = "feature_cluster_method",
    type = "character",
    optional = TRUE,
    default = default_cluster_method
  )
  
  # feature_linkage_method -----------------------------------------------------
  # Feature linkage method
  settings$feature_linkage_method <- .parse_arg(
    x_config = config$feature_linkage_method,
    x_var = feature_linkage_method,
    var_name = "feature_linkage_method",
    type = "character",
    optional = TRUE,
    default = default_cluster_linkage_method
  )
  
  # feature_cluster_cut_method -------------------------------------------------
  # Feature cluster cluster cut method
  settings$feature_cluster_cut_method <- .parse_arg(
    x_config = config$feature_cut_method,
    x_var = feature_cluster_cut_method,
    var_name = "feature_cluster_cut_method",
    type = "character",
    optional = TRUE,
    default = default_cluster_cut_method
  )
  
  # feature_similarity_metric --------------------------------------------------
  # Feature similarity metric
  settings$feature_similarity_metric <- .parse_arg(
    x_config = config$feature_similarity_metric,
    x_var = feature_similarity_metric,
    var_name = "feature_similarity_metric",
    type = "character",
    optional = TRUE,
    default = default_cluster_similarity_metric
  )
  
  # Perform additional checks, if required.
  if (.check_more && settings$feature_similarity_metric %in% c("mcfadden_r2", "cox_snell_r2", "nagelkerke_r2")) {
    if (!require_package(
      x = "nnet",
      purpose = paste0(
        "to compute log-likelihood pseudo R2 similarity using the ",
        settings$feature_similarity_metric, " metric"
      ),
      message_type = "backend_warning"
    )) {
      settings$feature_similarity_metric <- "spearman"
    }
    
  } else if (.check_more && settings$feature_similarity_metric %in% c("mutual_information")) {
    if (!require_package(
      x = "praznik",
      purpose = paste0(
        "to compute similarity using the ",
        settings$feature_similarity_metric, " metric"
      ),
      message_type = "backend_warning"
    )) {
      settings$feature_similarity_metric <- "spearman"
    }
  }
  
  # feature_similarity_threshold -----------------------------------------------
  # Feature similarity threshold
  settings$feature_similarity_threshold <- .parse_arg(
    x_config = config$feature_similarity_threshold,
    x_var = feature_similarity_threshold,
    var_name = "feature_similarity_threshold",
    type = "numeric_list",
    optional = TRUE,
    default = default_cluster_similarity_threshold
  )
  
  if (is.null(settings$feature_similarity_threshold)) {
    if (settings$feature_cluster_cut_method %in% c("fixed_cut")) {
      # Fixed cut requires stringent defaults, otherwise non-sense clusters will
      # be produced.
      if (settings$feature_similarity_metric %in% c("mcfadden_r2")) {
        settings$feature_similarity_threshold <- 0.30
      } else if (settings$feature_similarity_metric %in% c("mutual_information")) {
        settings$feature_similarity_threshold <- 0.23
      } else if (settings$feature_similarity_metric %in% c("cox_snell_r2", "nagelkerke_r2")) {
        settings$feature_similarity_threshold <- 0.75
      } else {
        settings$feature_similarity_threshold <- 0.90
      }
      
    } else {
      # The similarity threshold is also used to determine if any clusters could
      # potentially be found. The threshold is set low so that other cut methods
      # can be explored.
      if (settings$feature_similarity_metric %in% c("mcfadden_r2")) {
        settings$feature_similarity_threshold <- 0.25
      } else if (settings$feature_similarity_metric %in% c("mutual_information")) {
        settings$feature_similarity_threshold <- 0.10
      } else if (settings$feature_similarity_metric %in% c("cox_snell_r2", "nagelkerke_r2")) {
        settings$feature_similarity_threshold <- 0.40
      } else {
        settings$feature_similarity_threshold <- 0.70
      }
    }
  }
  
  # Check the proposed cluster parameters.
  if (.check_more) {
    .check_cluster_parameters(
      cluster_method = settings$feature_cluster_method,
      cluster_linkage = settings$feature_linkage_method,
      cluster_cut_method = settings$feature_cluster_cut_method,
      cluster_similarity_threshold = settings$feature_similarity_threshold,
      cluster_similarity_metric = settings$feature_similarity_metric,
      data_type = "feature",
      message_type = "backend_error"
    )
  }
  
  return(settings)
}
  


.parse_sample_clustering <- function(
    config = NULL,
    default_cluster_method = waiver(),
    default_cluster_linkage_method = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    .check_more = TRUE
) {
  settings <- list()
  
  # Set default cluster method.
  if (is.waive(default_cluster_method)) default_cluster_method <- "hclust"
  if (is.waive(default_cluster_linkage_method)) default_cluster_linkage_method <- "average"
  
  # sample_cluster_method ------------------------------------------------------
  # Sample cluster method
  settings$sample_cluster_method <- .parse_arg(
    x_config = config$sample_cluster_method,
    x_var = sample_cluster_method,
    var_name = "sample_cluster_method",
    type = "character",
    optional = TRUE,
    default = default_cluster_method
  )
  
  # sample_linkage_method ------------------------------------------------------
  # Sample cluster linkage method
  settings$sample_linkage_method <- .parse_arg(
    x_config = config$sample_linkage_method,
    x_var = sample_linkage_method,
    var_name = "sample_linkage_method",
    type = "character",
    optional = TRUE,
    default = default_cluster_linkage_method
  )
  
  # sample_similarity_metric ---------------------------------------------------
  # Sample similarity metric
  settings$sample_similarity_metric <- .parse_arg(
    x_config = config$sample_similarity_metric,
    x_var = sample_similarity_metric,
    var_name = "sample_similarity_metric",
    type = "character",
    optional = TRUE,
    default = "gower_winsor"
  )
  
  # Check similarity metrics can be feasibly computed.
  if (
    .check_more &&
    settings$sample_similarity_metric %in% c("mcfadden_r2", "cox_snell_r2", "nagelkerke_r2")
  ) {
    if (!require_package(
      x = "nnet",
      purpose = paste0(
        "to compute log-likelihood pseudo R2 similarity using the ",
        settings$sample_similarity_metric, " metric"
      ),
      message_type = "backend_warning"
    )) {
      settings$sample_similarity_metric <- "gower_winsor"
    }
  }
  
  # Check the proposed cluster parameters.
  if (.check_more) {
    .check_cluster_parameters(
      cluster_method = settings$sample_cluster_method,
      cluster_linkage = settings$sample_linkage_method,
      cluster_similarity_metric = settings$sample_similarity_metric,
      data_type = "sample",
      message_type = "backend_error"
    )
  }
  
  return(settings)
}



.parse_evaluation_time_settings <- function(
    data,
    outcome_type,
    config = NULL,
    time_max = waiver(),
    evaluation_times = waiver(),
    development_batch_id = NULL
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome_event <- batch_id <- NULL
  
  settings <- list()
  
  # time_max -------------------------------------------------------------------
  # Study end time (this is used for plotting, and Uno's concordance index).
  settings$time_max <- .parse_arg(
    x_config = config$time_max,
    x_var = time_max,
    var_name = "time_max",
    type = "numeric",
    optional = TRUE,
    default = NULL
  )
  
  # evaluation_times -----------------------------------------------------------
  # Times at which calibration is evaluated for time-to-event (survival) data.
  settings$eval_times <- .parse_arg(
    x_config = config$evaluation_times,
    evaluation_times,
    var_name = "evaluation_time",
    type = "numeric_list",
    optional = TRUE,
    default = NULL
  )
  
  # Update time_max and eval_times only if we are dealing with survival
  # endpoints.
  if (outcome_type %in% c("survival")) {
    # Identify values for time_max if it has not been provided.
    if (is.null(settings$time_max)) {
      if (!is.null(settings$eval_times)) {
        # Use maximum evaluation time
        settings$time_max <- max(settings$eval_times)
        
      } else if (!is.null(development_batch_id)) {
        # 98th percentile of all outcome times in the training cohorts.
        settings$time_max <- .get_default_time_max(
          data[outcome_event == 1L & batch_id %in% development_batch_id]$outcome_time
        )
        
      } else {
        # 98th percentile of all outcome times in the training cohorts.
        settings$time_max <- .get_default_time_max(
          data[outcome_event == 1L]$outcome_time
        )
      }
    }
    
    .check_number_in_valid_range(
      settings$time_max,
      var_name = "time_max",
      range = c(0.0, Inf),
      closed = c(FALSE, TRUE)
    )
    
    # Identify evaluation times if they were not provided.
    if (is.null(settings$eval_times)) settings$eval_times <- settings$time_max
    
    sapply(
      settings$eval_times,
      .check_number_in_valid_range,
      var_name = "evaluation_times",
      range = c(0.0, Inf),
      closed = c(FALSE, TRUE)
    )
  }
  
  return(settings)
}



.get_all_parameter_names <- function() {
  # Find all possible arguments that could be matched.
  available_parameters <- unique(c(
    names(as.list(args(.parse_file_paths))),
    names(as.list(args(.parse_experiment_settings))),
    names(as.list(args(.parse_setup_settings))),
    names(as.list(args(.parse_preprocessing_settings))),
    names(as.list(args(.parse_variable_importance_settings))),
    names(as.list(args(.parse_model_development_settings))),
    names(as.list(args(.parse_hyperparameter_optimisation_settings))),
    names(as.list(args(.parse_evaluation_settings)))
  ))

  # Remove "" and "..." and other arguments that are not parameters that can be
  # specified using ... ..
  available_parameters <- setdiff(
    available_parameters,
    c("",
      "...",
      "config",
      "data",
      "hpo_metric",
      "prep_cluster_method",
      "prep_cluster_linkage_method",
      "prep_cluster_cut_method",
      "prep_cluster_similarity_threshold",
      "prep_cluster_similarity_metric",
      "verbose"
    )
  )

  return(available_parameters)
}



.get_all_configuration_parent_node_names <- function() {
  return(c(
    "paths", "data", "run", "preprocessing", "variable_importance",
    "model_development", "hyperparameter_optimisation", "evaluation"
  ))
}



.get_default_time_max <- function(x) {
  if (is_empty(x)) return(Inf)
  if (!any(is.finite(x))) return(Inf)
  
  x <- x[is.finite(x)]
  
  return(stats::quantile(
    x,
    probs = 0.98,
    na.rm = TRUE,
    names = FALSE
  ))
}

# xml2 is required to .
if (!rlang::is_installed("xml2")) testthat::skip()

# Find path to configuration file in package.
config <- system.file("config.xml", package = "familiar")

# Load the configuration file
config <- familiar:::.load_configuration_file(config)

# Check whether the main branches have the correct names.
# Find names of parent nodes.
config_node_names <- names(config)
expected_node_names <- familiar:::.get_all_configuration_parent_node_names()

testthat::test_that(
  "1. All parent nodes are present in the configuration file.", 
  {
    testthat::expect_setequal(config_node_names, expected_node_names)
  }
)

testthat::test_that(
  "2. All parameters are specified.",
  {
    config_args <- unique(unlist(sapply(config, names)))
    testthat::expect_setequal(config_args, familiar:::.get_all_parameter_names())
  }
)

for (parent_node in expected_node_names) {
  # Identify the parsing function for the node.
  parse_fun <- switch(parent_node,
    paths = familiar:::.parse_file_paths,
    data = familiar:::.parse_experiment_settings,
    run = familiar:::.parse_setup_settings,
    preprocessing = familiar:::.parse_preprocessing_settings,
    variable_importance = familiar:::.parse_variable_importance_settings,
    model_development = familiar:::.parse_model_development_settings,
    hyperparameter_optimisation = familiar:::.parse_hyperparameter_optimisation_settings,
    evaluation = familiar:::.parse_evaluation_settings
  )

  # Find the expected arguments.
  expected_config_args <- names(as.list(args(parse_fun)))

  # Remove "" and "..." and other arguments that are not parameters that can be
  # specified using ... .
  expected_config_args <- intersect(expected_config_args, familiar:::.get_all_parameter_names())

  # Remove specific arguments unless they are shared by a specific function.
  if (parent_node != "data") {
    expected_config_args <- setdiff(
      expected_config_args, c("outcome_type", "development_batch_id")
    )
  }

  if (parent_node != "run") {
    expected_config_args <- setdiff(
      expected_config_args, "parallel"
    )
  }

  if (parent_node != "variable_importance") {
    expected_config_args <- setdiff(
      expected_config_args,
      c("vimp_aggregation_rank_threshold", "vimp_aggregation_method")
    )
  }

  # Find the argument names.
  config_args <- names(config[[parent_node]])

  testthat::test_that(
    paste0("3. All parameters are specified for the \"", parent_node, "\" node."),
    {
      testthat::expect_setequal(config_args, expected_config_args)
    }
  )
}

test_create_good_data <- function(
    outcome_type,
    to_data_object = TRUE,
    seed = 1844L,
    rstream_object = NULL,
    one_relevant_feature = FALSE,
    two_groups = FALSE
) {
  
  # Create random stream object so that the same numbers are produced every
  # time.
  if (is.null(rstream_object)) {
    r <- .start_random_number_stream(seed = seed)
  } else {
    r <- rstream_object
  }
  
  n_series_instances <- 150L
  
  # Draw random numbers for four features.
  feature_1 <- fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)
  feature_2a <- fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)
  feature_2b <- feature_2a + fam_runif(n = n_series_instances, min = -0.05, max = 0.05, rstream_object = r)
  feature_3a <- fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)
  feature_3b <- feature_3a + fam_runif(n = n_series_instances, min = -0.20, max = 0.20, rstream_object = r)
  feature_4 <- fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)
  
  # Determine the raw outcome.
  if (one_relevant_feature) {
    # Only one feature is relevant.
    outcome_raw <- feature_1
  } else {
    # More than one feature is relevant (default).
    outcome_raw <- feature_1 + 0.1 * feature_2a + 0.3 * feature_3a + 0.1 * feature_4
  }
  
  
  if (outcome_type == "binomial") {
    # Because outcome_raw is strongly correlated with feature 1, this may lead to
    # numerical issues. To prevent this, we add some noise to the outcome.
    outcome_raw <- outcome_raw + fam_rnorm(n = n_series_instances, mean = 0.0, sd = 0.1, rstream_object = r)
    
    # Convert to 0, 1
    outcome_value <- outcome_raw > 0.75
    outcome_value <- factor(
      x = outcome_value,
      levels = c(FALSE, TRUE),
      labels = c("red", "green")
    )
    
  } else if (outcome_type == "multinomial") {
    outcome_value <- numeric(n_series_instances)
    
    # Convert to 0 (x < 0.58), 1 (0.58 < x < 0.92), 2 (0.92 < x < 1.5)
    outcome_value[outcome_raw < 0.58] <- 0.0
    outcome_value[outcome_raw >= 0.58 & outcome_raw < 0.92] <- 1.0
    outcome_value[outcome_raw >= 0.92] <- 2.0
    
    outcome_value <- factor(
      x = outcome_value,
      levels = c(0.0, 1.0, 2.0),
      labels = c("red", "green", "blue")
    )
    
  } else if (outcome_type == "continuous") {
    outcome_value <- outcome_raw
    
  } else if (outcome_type == "survival") {
    # Simulate event times using exponential function and lambda = 0.5
    outcome_time <- - log(fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)) / 
      (0.5 * exp(3.0 * (outcome_raw - mean(outcome_raw))))
    
    # Simulate censoring times using exponential function and lambda = 0.1
    censor_time <- - log(fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)) / 0.1
    outcome_event <- rep_len(1L, length.out = n_series_instances)
    outcome_event[censor_time < outcome_time] <- 0L
    outcome_time[outcome_event == 0L] <- censor_time[outcome_event == 0L]
    
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }
  
  # Create basic table.
  data <- data.table::data.table(
    "batch_id" = "basic",
    "sample_id" = paste0("sample_", seq_len(n_series_instances)),
    "series_id" = 1L,
    "feature_1" = feature_1,
    "feature_2a" = feature_2a,
    "feature_2b" = feature_2b,
    "feature_3a" = cut(
      x = feature_3a,
      breaks = c(0.0, 0.5, 1.0),
      labels = c("round", "square")
    ),
    "feature_3b" = cut(
      x = feature_3b,
      breaks = c(-1.0, 0.5, 2.0),
      labels = c("sphere", "cube")
    ),
    "feature_4" = cut(
      x = feature_4,
      breaks = c(0.0, 0.333, 0.667, 1.0),
      labels = c("good", "better", "best"),
      ordered_result = TRUE
    )
  )
  
  # Add outcome.
  if (outcome_type %in% "survival") {
    data[, ":="(
      "outcome_time" = outcome_time,
      "outcome_event" = outcome_event
    )]
    outcome_column <- c("outcome_time", "outcome_event")
    
  } else {
    data[, ":="("outcome" = outcome_value)]
    outcome_column <- "outcome"
  }
  
  # Update batch_id
  if (two_groups) {
    data[1L:75L, "batch_id" := "group_a"]
    data[76L:150L, "batch_id" := "group_b"]
  }
  
  # Convert to a data object.
  if (to_data_object) {
    data <- as_data_object(
      data = data,
      batch_id_column = "batch_id",
      sample_id_column = "sample_id",
      series_id_column = "series_id",
      outcome_column = outcome_column,
      outcome_type = outcome_type
    )
  }
  
  return(data)
}



test_create_small_good_data <- function(outcome_type) {
  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)

  # Now select a subset of the data.
  data@data <- data@data[fam_sample(
    seq_len(nrow(data@data)),
    size = 30L,
    replace = FALSE,
    seed = 1844L
  )]

  return(data)
}



test_create_reverse_column_order_data <- function(outcome_type) {
  # Create good dataset with reversed column order.
  data <- test_create_good_data(outcome_type = outcome_type)
  
  # Reverse column order.
  column_names <- colnames(data@data)
  data.table::setcolorder(x = data@data, neworder = rev(column_names))
  
  return(data)
}



test_create_invariant_good_data <- function(outcome_type) {
  # Create good dataset with one invariant feature.
  
  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)
  data@data[, "feature_2a" := 0.5]

  return(data)
}



test_create_good_data_without_censoring <- function(outcome_type) {
  # Dataset where none of the samples is censored.
  
  if (!outcome_type %in% c("survival", "competing_risk")) return(NULL)

  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)

  # Set all data to event.
  data@data[, "outcome_event" := 1L]

  return(data)
}



test_create_good_data_one_censored <- function(outcome_type) {
  # Dataset where just one of the samples is censored.
  
  if (!outcome_type %in% c("survival", "competing_risk")) return(NULL)

  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)

  # Set all data to event.
  data@data[, "outcome_event" := 1L]

  # Set one instance to censored.
  data@data[1L, "outcome_event" := 0L]

  return(data)
}



test_create_good_data_few_censored <- function(outcome_type) {
  # Dataset where a minor fraction of the samples is censored.
  if (!outcome_type %in% c("survival", "competing_risk")) return(NULL)

  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)

  # Set all data to event.
  data@data[, "outcome_event" := 1L]

  # Set a few instances to censored.
  data@data[seq_len(4L), "outcome_event" := 0L]

  return(data)
}



test_create_good_data_random_missing <- function(
    outcome_type,
    n_missing_frac = 0.05,
    seed = 1844L,
    rstream_object = NULL
) {
  # Some data points are NA, but not instances.
  if (!is.null(seed) && is.null(rstream_object)) {
    rstream_object <- .start_random_number_stream(seed = seed)
  }

  # Create test data.
  data <- test_create_good_data(outcome_type = outcome_type)

  # Get the number of rows and feature columns.
  n_rows <- nrow(data@data)
  feature_columns <- get_feature_columns(data)

  # Select the number of data points to randomise.
  n_randomise <- ceiling(n_missing_frac * n_rows * length(feature_columns))

  # Determine which data points should be randomised.
  random_data <- data.table::data.table(
    "row_id" = fam_sample(
      seq_len(n_rows),
      size = n_randomise,
      replace = TRUE,
      rstream_object = rstream_object
    ),
    "feature" = fam_sample(
      feature_columns,
      size = n_randomise,
      replace = TRUE,
      rstream_object = rstream_object
    )
  )
  
  # Select only unique data points.
  random_data <- unique(random_data)

  # Split for iteration.
  random_data <- split(
    random_data,
    by = "feature",
    drop = TRUE
  )

  # Update feature columns.
  for (feature in feature_columns) {
    if (!is.null(random_data[[feature]])) {
      # Split by numeric and factor features.
      if (is.factor(data@data[[feature]])) {
        data@data[random_data[[feature]]$row_id, (feature) := NA]
      } else {
        data@data[random_data[[feature]]$row_id, (feature) := NA_real_]
      }
    }
  }

  return(data)
}



test_create_single_relevant_feature_data <- function(outcome_type) {
  # Multiple features, but only one relevant.
  
  # Create dataset.
  data <- test_create_good_data(
    outcome_type = outcome_type,
    one_relevant_feature = TRUE
  )
  
  return(data)
}



test_create_empty_data <- function(outcome_type) {
  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)

  # Now empty the data.
  data@data <- head(data@data, n = 0L)

  return(data)
}



test_create_data_without_feature <- function(outcome_type, to_data_object = TRUE) {
  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type)

  # Remove features.
  data@data <- data@data[, mget(get_non_feature_columns(outcome_type))]

  if (!to_data_object) data <- data@data
  
  return(data)
}



test_create_bootstrapped_data <- function(outcome_type, seed = 1844L, to_data_object = TRUE) {
  # Suppress NOTES due to non-standard evaluation in data.table
  sample_id <- NULL

  # Create good dataset first and work from there.
  data <- test_create_good_data(
    outcome_type = outcome_type,
    to_data_object = to_data_object
  )

  # Now keep only the first sample.
  if (to_data_object) {
    data@data <- data@data[fam_sample(
      seq_len(nrow(data@data)),
      size = nrow(data@data),
      replace = TRUE,
      seed = seed
    )][order(sample_id)]
    
  } else {
    data@data <- data[fam_sample(
      seq_len(nrow(data)),
      size = nrow(data),
      replace = TRUE,
      seed = seed
    )][order(sample_id)]
  }

  return(data)
}



test_create_one_sample_data <- function(outcome_type, to_data_object = TRUE) {
  # Create good dataset first and work from there.
  data <- test_create_good_data(
    outcome_type = outcome_type,
    to_data_object = to_data_object
  )

  # Now keep only the first sample.
  if (to_data_object) {
    data@data <- head(data@data, n = 1L)
  } else {
    data <- head(data, n = 1L)
  }

  return(data)
}



test_create_all_identical_data <- function(outcome_type) {
  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)

  # Now keep only the first sample.
  data@data <- head(data@data, n = 1L)

  # Fill the dataset with the same sample.
  data@data <- data@data[rep.int(1L, 10L)]

  # Set unique subject ids.
  data@data[, "sample_id" := paste0("sample_", seq_len(nrow(data@data)))]

  return(data)
}



test_create_single_feature_data <- function(outcome_type) {
  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)
  
  data@data[, ":="(
    "feature_2a" = NULL,
    "feature_2b" = NULL,
    "feature_3a" = NULL,
    "feature_3b" = NULL,
    "feature_4" = NULL
  )]

  return(data)
}



test_create_single_feature_one_sample_data <- function(outcome_type) {
  # Create good dataset first and work from there.
  data <- test_create_single_feature_data(outcome_type = outcome_type)

  # Now keep only the first sample.
  data@data <- head(data@data, n = 1L)

  return(data)
}



test_create_single_feature_invariant_data <- function(outcome_type) {
  # Create good dataset first and work from there.
  data <- test_create_single_feature_data(outcome_type = outcome_type)

  # Get the feature column
  feature_column <- get_feature_columns(data)

  # Set the feature to a fixed value.
  data@data[, (feature_column) := data@data[[feature_column]][1L]]

  return(data)
}



test_create_single_feature_two_values_data <- function(outcome_type) {
  # Create good dataset first.
  data <- test_create_single_feature_data(outcome_type = outcome_type)

  # Get the feature columns
  feature_column <- get_feature_columns(data)

  # Find unique values of the feature and use the first 2.
  feature_values <- head(unique(data@data[[feature_column]]), n = 2L)

  # Fill all the rows while alternating the value.
  data@data[, (feature_column) := rep_len(feature_values, nrow(data@data))]

  return(data)
}



test_create_wide_data <- function(outcome_type) {
  # Create random stream object so that the same numbers are produced every
  # time.
  r <- .start_random_number_stream(seed = 1844L)

  data <- test_create_good_data(
    outcome_type = outcome_type,
    rstream_object = r
  )
  
  # Add twenty random features
  random_data <- lapply(
    seq_len(20L), 
    function(ii, n, r) fam_rnorm(n = n, rstream_object = r),
    n = nrow(data@data),
    r = r
  )
  names(random_data) <- paste0("random_feature_", seq_len(20L))
  
  # Add to dataset
  data@data <- cbind(data@data, data.table::as.data.table(random_data))


  return(data)
}



test_create_bad_data <- function(outcome_type, add_na_data = FALSE) {
  # add_na_data argument is intended for integration tests, where we have to
  # circumvent a check on the outcome classes. We do this by keeping these
  # classes in, and assigning NA to rows of one class, causing the data to pass
  # the check, but have the rows be removed afterwards.

  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- NULL

  # Create good dataset first and work from there.
  data <- test_create_good_data(outcome_type = outcome_type)

  if (outcome_type == "survival") {
    # For survival data it would be really bad if all data are censored.
    data@data[, "outcome_event" := 0L]
    
  } else if (outcome_type == "multinomial") {
    # For multinomial data, having not all classes is bad.

    if (add_na_data) {
      # Assign NA to the rows containing the "red" class.

      # Identify the feature columns.
      feature_columns <- get_feature_columns(data)

      # Update feature columns.
      for (feature in feature_columns) {
        if (is.factor(data@data[[feature]])) {
          data@data[outcome == "red", (feature) := NA]
        } else {
          data@data[outcome == "red", (feature) := NA_real_]
        }
      }
    } else {
      # Select 2 of 3 classes by leaving red out.
      data@data <- data@data[outcome %in% c("blue", "green"), ]
    }
    
  } else if (outcome_type == "binomial") {
    # For binomial data, having a single class is bad.

    if (add_na_data) {
      # Assign NA to the rows containing the red class.

      # Identify the feature columns.
      feature_columns <- get_feature_columns(data)

      # Update feature columns.
      for (feature in feature_columns) {
        if (is.factor(data@data[[feature]])) {
          data@data[outcome == "red", (feature) := NA]
        } else {
          data@data[outcome == "red", (feature) := NA_real_]
        }
      }
    } else {
      # Assign everything to the green class.
      data@data[, "outcome" := "green"]
    }
    
  } else if (outcome_type == "continuous") {
    # For continuous data, it would be bad if all outcome values are invariant.
    data@data[, "outcome" := 1.0]
    
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }

  return(data)
}


test_create_small_bad_data <- function(outcome_type) {
  # Create good dataset first and work from there.
  data <- test_create_bad_data(outcome_type = outcome_type)

  # Now select a subset of the data.
  data@data <- data@data[fam_sample(
    seq_len(nrow(data@data)),
    size = 30L,
    replace = FALSE,
    seed = 1844L
  )]

  return(data)
}



test_create_prospective_data <- function(outcome_type) {
  # Prospective data has NA for outcome.
  
  data <- test_create_good_data(outcome_type = outcome_type)
  
  if (outcome_type %in% c("survival", "competing_risk")) {
    data@data[
      ,
      ":="(
        "outcome_time" = NA,
        "outcome_event" = NA_integer_
      )
    ]
    
  } else {
    data@data[, ":="("outcome" = NA)]
  }
  
  return(data)
}



test_create_partially_prospective_data <- function(outcome_type) {
  # Prospective data has NA for outcome for a few samples, but not all.

  data <- test_create_good_data(outcome_type = outcome_type)

  if (outcome_type %in% c("survival", "competing_risk")) {
    data@data[
      c(1L, 2L),
      ":="(
        "outcome_time" = NA,
        "outcome_event" = NA_integer_
      )
    ]
    
  } else {
    data@data[c(1L, 2L), ":="("outcome" = NA)]
  }

  return(data)
}



test_create_mostly_prospective_data <- function(outcome_type) {
  # Prospective data has NA for all but one sample.

  # Suppress NOTES due to non-standard evaluation in data.table
  sample_id <- NULL

  data <- test_create_good_data(outcome_type = outcome_type)

  if (outcome_type %in% c("survival", "competing_risk")) {
    data@data[
      sample_id != "sample_1",
      ":="(
        "outcome_time" = NA,
        "outcome_event" = NA_integer_
      )
    ]
    
  } else {
    data@data[sample_id != "sample_1", ":="("outcome" = NA)]
  }

  return(data)
}



test_create_synthetic_series_data <- function(
    outcome_type,
    n_batch = 3L,
    n_samples = 10L,
    n_series = 3L,
    n_rep = 3L,
    n_numeric = 4L,
    rare_outcome = FALSE,
    seed = 1844L,
    rstream_object = NULL
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  batch_id <- feature_1 <- feature_2 <- feature_3 <- feature_4 <- NULL

  # Create random stream object so that the same numbers are produced every
  # time.
  if (is.null(rstream_object)) {
    r <- .start_random_number_stream(seed = seed)
  } else {
    r <- rstream_object
  }

  # Determine the number of series instances.
  n_series_instances <- n_batch * n_samples * n_series

  # Draw random numbers for four features.
  feature_1 <- fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)
  feature_2 <- fam_runif(n = n_series_instances, min = 0.0, max = 2.0, rstream_object = r)
  feature_3 <- fam_runif(n = n_series_instances, min = 0.0, max = 2.0, rstream_object = r)
  feature_4 <- fam_runif(n = n_series_instances, min = 0.0, max = 1.0, rstream_object = r)

  # Determine the raw outcome.
  outcome_raw <- feature_1 + feature_2 + feature_3 + feature_4

  if (outcome_type == "binomial") {
    # Convert to 0, 1
    outcome_value <- outcome_raw > 3.0
    outcome_value <- factor(
      x = outcome_value,
      levels = c(FALSE, TRUE),
      labels = c("0", "1")
    )
    
  } else if (outcome_type == "multinomial") {
    outcome_value <- numeric(n_series_instances)

    # Convert to 0 (x < 2.59), 1 (2.59 < x < 3.41), 2 (3.41 < x < 6)
    outcome_value[outcome_raw < 2.59] <- 0.0
    outcome_value[outcome_raw >= 2.59 & outcome_raw < 3.41] <- 1.0
    outcome_value[outcome_raw >= 3.41 & outcome_raw < 6.00] <- 2.0

    if (rare_outcome) {
      outcome_value[length(outcome_value)] <- 3.0
      outcome_value <- factor(
        x = outcome_value,
        levels = c(0.0, 1.0, 2.0, 3.0),
        labels = c("0", "1", "2", "3")
      )
      
    } else {
      outcome_value <- factor(
        x = outcome_value,
        levels = c(0.0, 1.0, 2.0),
        labels = c("0", "1", "2")
      )
      
    }
  } else if (outcome_type == "continuous") {
    outcome_value <- outcome_raw
    
  } else if (outcome_type == "survival") {
    # Outcome follows an exponential distribution.
    outcome_time <- exp(outcome_raw)
    outcome_event <- rep_len(1L, length.out = n_series_instances)
    
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }

  # Create basic table. Sample identifiers are explicitly repeated for different
  # batches.
  data <- data.table::data.table(
    "batch_id" = rep(seq_len(n_batch), each = n_samples * n_series),
    "sample_id" = rep(seq_len(n_samples), each = n_series, times = n_batch),
    "series_id" = rep(seq_len(n_series), times = n_batch * n_samples),
    "feature_1" = feature_1,
    "feature_2" = feature_2,
    "feature_3" = feature_3,
    "feature_4" = feature_4
  )

  # Add outcome.
  if (outcome_type %in% "survival") {
    data[, ":="(
      "outcome_time" = outcome_time,
      "outcome_event" = outcome_event
    )]
    outcome_column <- c("outcome_time", "outcome_event")
    
  } else {
    data[, ":="("outcome" = outcome_value)]
    outcome_column <- "outcome"
  }

  # Create batch-offsets
  data[, ":="(
    "feature_1" = feature_1 + batch_id - 1.0,
    "feature_2" = feature_2 + batch_id - 1.0,
    "feature_3" = feature_3 + batch_id - 1.0,
    "feature_4" = feature_4 + batch_id - 1.0
  )]

  # Create repetitions.
  if (n_rep > 1L) {
    repeated_rows <- rep(seq_len(n_series_instances), each = n_rep)
    data <- data[repeated_rows, ]

    # Add some noise to features.
    data[, ":="(
      "feature_1" = feature_1 + fam_rnorm(n = n_rep * n_series_instances, mean = 0.0, sd = 0.125, rstream_object = r),
      "feature_2" = feature_2 + fam_rnorm(n = n_rep * n_series_instances, mean = 0.0, sd = 0.125, rstream_object = r),
      "feature_3" = feature_3 + fam_rnorm(n = n_rep * n_series_instances, mean = 0.0, sd = 0.125, rstream_object = r),
      "feature_4" = feature_4 + fam_rnorm(n = n_rep * n_series_instances, mean = 0.0, sd = 0.125, rstream_object = r)
    )]

    data[feature_1 <= 0.0, "feature_1" := 0.01]
    data[feature_2 <= 0.0, "feature_2" := 0.01]
    data[feature_3 <= 0.0, "feature_3" := 0.01]
    data[feature_4 <= 0.0, "feature_4" := 0.01]
  }

  if (n_numeric < 4L) data$feature_1 <- factor(floor(data$feature_1))
  if (n_numeric < 3L) data$feature_2 <- factor(floor(data$feature_2))
  if (n_numeric < 2L) data$feature_3 <- factor(floor(data$feature_3))
  if (n_numeric < 1L) data$feature_4 <- factor(floor(data$feature_4))

  # Convert to a data object.
  data <- as_data_object(
    data = data,
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    outcome_column = outcome_column,
    outcome_type = outcome_type
  )

  return(data)
}



test_create_synthetic_series_one_outcome <- function(
    outcome_type,
    n_numeric = 4L,
    seed = 1844L,
    rstream_object = NULL
) {
  # Create test data.
  data <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric,
    seed = seed,
    rstream_object = rstream_object
  )

  if (outcome_type %in% c("binomial", "multinomial")) {
    data@data[, "outcome" := "0"]
  } else if (outcome_type %in% c("continuous")) {
    data@data[, "outcome" := 1L]
  } else if (outcome_type == "survival") {
    data@data[, ":="("outcome_time" = 1.25, "outcome_event" = 1L)]
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }

  return(data)
}



test_create_synthetic_series_one_sample_data <- function(
    outcome_type,
    n_numeric = 4L,
    seed = 1844L,
    rstream_object = NULL
) {
  # Create test data.
  data <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric,
    seed = seed,
    rstream_object = rstream_object
  )

  # Select the first instance
  data@data <- head(data@data, n = 1L)

  return(data)
}



test_create_synthetic_series_invariant_feature_data <- function(
    outcome_type,
    n_numeric = 4L,
    seed = 1844L,
    rstream_object = NULL
) {
  # Create test data.
  data <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric,
    seed = seed,
    rstream_object = rstream_object
  )

  # Select the first instance
  data@data$feature_1 <- data@data$feature_1[1L]
  data@data$feature_2 <- data@data$feature_2[1L]
  data@data$feature_3 <- data@data$feature_3[1L]
  data@data$feature_4 <- data@data$feature_4[1L]

  return(data)
}



test_create_synthetic_series_one_feature_invariant_data <- function(
    outcome_type,
    n_numeric = 4L,
    seed = 1844L,
    rstream_object = NULL
) {
  # Create test data.
  data <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric,
    seed = seed,
    rstream_object = rstream_object
  )

  # Select the first instance for feature 2.
  data@data$feature_2 <- data@data$feature_2[1L]

  return(data)
}



test_create_synthetic_series_na_data <- function(
    outcome_type,
    n_numeric = 4L,
    n_missing_frac = 0.1,
    seed = 1844L,
    rstream_object = NULL
) {
  # Some instances are completely NA.

  # Create test data.
  data <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric,
    seed = seed,
    rstream_object = rstream_object
  )

  # Select which rows will be updated.
  n_rows <- nrow(data@data)
  na_rows <- fam_sample(
    seq_len(n_rows),
    size = ceiling(n_missing_frac * n_rows),
    replace = FALSE,
    seed = seed,
    rstream_object = rstream_object
  )

  # Identify the feature columns.
  feature_columns <- get_feature_columns(data)

  # Update feature columns.
  for (feature in feature_columns) {
    if (is.factor(data@data[[feature]])) {
      data@data[na_rows, (feature) := NA]
    } else {
      data@data[na_rows, (feature) := NA_real_]
    }
  }

  return(data)
}



test_create_synthetic_series_random_na_data <- function(
    outcome_type,
    n_numeric = 4L,
    n_missing_frac = 0.1,
    seed = 1844L,
    rstream_object = NULL
) {
  # Some data points are NA, but not instances.

  if (!is.null(seed) && is.null(rstream_object)) {
    rstream_object <- .start_random_number_stream(seed = seed)
  }

  # Create test data.
  data <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric,
    rstream_object = rstream_object
  )

  # Get the number of rows and feature columns.
  n_rows <- nrow(data@data)
  feature_columns <- get_feature_columns(data)

  # Select the number of data points to randomise.
  n_randomise <- ceiling(n_missing_frac * n_rows * length(feature_columns))

  # Determine which data points should be randomised.
  random_data <- data.table::data.table(
    "row_id" = fam_sample(seq_len(n_rows),
      size = n_randomise,
      replace = TRUE,
      rstream_object = rstream_object
    ),
    "feature" = fam_sample(feature_columns,
      size = n_randomise,
      replace = TRUE,
      rstream_object = rstream_object
    )
  )
  
  # Select only unique data points.
  random_data <- unique(random_data)

  # Split for iteration.
  random_data <- split(random_data, by = "feature", drop = TRUE)

  # Update feature columns.
  for (feature in feature_columns) {
    if (!is.null(random_data[[feature]])) {
      # Split by numeric and factor features.
      if (is.factor(data@data[[feature]])) {
        data@data[random_data[[feature]]$row_id, (feature) := NA]
      } else {
        data@data[random_data[[feature]]$row_id, (feature) := NA_real_]
      }
    }
  }

  return(data)
}



test_create_synthetic_series_one_feature_all_na_data <- function(
    outcome_type,
    n_numeric = 4L,
    seed = 1844L,
    rstream_object = NULL
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  feature_2 <- NULL

  # Create test data.
  data <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric,
    seed = seed,
    rstream_object = rstream_object
  )

  # Set the first feature column to NA.
  if (is.factor(data@data[["feature_2"]])) {
    data@data[, feature_2 := NA]
  } else {
    data@data[, feature_2 := NA_real_]
  }

  return(data)
}



test_create_multiple_synthetic_series <- function(outcome_type) {
  # The idea here is to create multiple synthetic datasets that together
  # represent extreme variation in data composition.

  ..extend_feature_set <- function(data) {
    # Get feature columns.
    original_feature_columns <- get_feature_columns(data)

    # Find new feature columns for the correlated features
    new_feature_columns <- paste0(
      "feature_",
      seq_along(original_feature_columns) + length(original_feature_columns)
    )

    # Add in correlated features.
    for (ii in seq_along(original_feature_columns)) {
      data@data[, (new_feature_columns[ii]) := get(original_feature_columns[ii])]
    }

    return(data)
  }

  # Draw the first dataset.
  data_1 <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = 3L,
    n_samples = 20L,
    seed = 1L
  )

  # Add correlated features.
  data_1 <- ..extend_feature_set(data_1)

  # Draw the second dataset.
  data_2 <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = 3L,
    n_samples = 20L,
    seed = 2L
  )
  # Do not add correlated features to dataset 2.

  # Draw a third dataset.
  data_3 <- test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = 3L,
    n_samples = 20L,
    seed = 3L
  )

  # Add correlated features, but remove the original features.
  data_3 <- ..extend_feature_set(data_3)
  data_3@data[, ":="(
    "feature_1" = NULL,
    "feature_2" = NULL,
    "feature_3" = NULL,
    "feature_4" = NULL
  )]

  # Draw a fourth dataset that cannot be used for training, e.g. contains just
  # one sample.
  data_4 <- test_create_synthetic_series_one_outcome(
    outcome_type = outcome_type,
    n_numeric = 3L,
    seed = 4L
  )

  # Add correlated features.
  data_4 <- ..extend_feature_set(data_4)

  return(list(
    "set_1" = data_1,
    "set_2" = data_2,
    "set_3" = data_3,
    "set_4" = data_4
  ))
}




test_create_synthetic_correlated_data <- function(
    ...,
    seed = 1844L,
    rstream_object = NULL,
    cluster_size = c(1L, 1L, 1L, 1L),
    mix_feature_types = TRUE,
    allow_anti_correlation = TRUE
) {
  # Create random stream object so that the same numbers are produced every
  # time.
  if (is.null(rstream_object)) {
    r <- .start_random_number_stream(seed = seed)
  } else {
    r <- rstream_object
  }

  # Create basic data.
  base_data <- do.call(
    test_create_synthetic_series_data,
    args = c(
      list("rstream_object" = r),
      list(...)
    )
  )

  if (length(cluster_size) != get_n_features(base_data)) {
    ..error(paste0("The cluster_size argument should match the number of features."))
  }

  # Get feature names prior to extending the dataset.
  original_feature_names <- get_feature_columns(base_data)

  # Create copy of data.
  data <- base_data
  data@data <- data.table::copy(base_data@data)

  for (ii in seq_along(cluster_size)) {
    # Skip if no clusters are required.
    if (cluster_size[ii] == 1L) next

    # Isolate feature data. Force reading this data just in case.
    x <- base_data@data[[original_feature_names[ii]]]

    # Generate clusters.
    for (jj in seq_len(cluster_size[ii])) {
      # Create the name of the new feature.
      new_feature_name <- paste0(original_feature_names[ii], "_", LETTERS[jj])

      if (jj == 1L) {
        # The first feature should just be renamed.
        data.table::setnames(data@data,
          old = original_feature_names[ii],
          new = new_feature_name
        )
        
      } else if (jj == 2L && is.factor(x) && mix_feature_types) {
        # Add numeric feature when mixing data types.
        data@data[, (new_feature_name) := as.numeric(x) * 1.1]
        
      } else if (is.factor(x)) {
        # For categorical features, remix the levels.
        y <- factor(x,
          levels = levels(x),
          labels = levels(x)[
            fam_sample(
              seq_along(levels(x)),
              size = nlevels(x),
              replace = FALSE,
              rstream_object = r
            )
          ]
        )
        
        # Set feature.
        data@data[, (new_feature_name) := y]
        
      } else {
        # For numerical features, introduce an offset and scaling. We scale
        # between 0.5 and 1.5 or -0.5 and -1.5 to avoid multiplying by 0. The
        # negative values are used for strong anti-correlation, which
        if (jj %% 2L == 0L || !allow_anti_correlation) {
          r_scale <- fam_runif(n = 1L, min = 0.5, max = 1.5, rstream_object = r)
        } else {
          r_scale <- fam_runif(n = 1L, min = -1.5, max = -0.5, rstream_object = r)
        }

        r_shift <- fam_runif(n = 1L, min = -1.0, max = 1.0, rstream_object = r)

        y <- x * r_scale + r_shift

        data@data[, (new_feature_name) := y]
      }
    }
  }

  # Order columns nicely.
  data.table::setcolorder(
    x = data@data, 
    neworder = c(
      get_non_feature_columns(data),
      sort(get_feature_columns(data))
    )
  )

  return(data)
}



test_create_synthetic_correlated_one_feature_invariant_data <- function(
    ...,
    cluster_size = c(1L, 1L, 1L, 1L)
) {
  # Set the size of the second cluster to 1, always.
  cluster_size[2L] <- 1L

  # Create test data.
  data <- do.call(
    test_create_synthetic_correlated_data,
    args = c(
      list("cluster_size" = cluster_size),
      list(...)
    )
  )

  # Select the first instance for feature 2.
  data@data$feature_2 <- data@data$feature_2[1L]

  return(data)
}



test_create_synthetic_correlated_one_sample_data <- function(...) {
  # Create test data.
  data <- do.call(
    test_create_synthetic_correlated_data,
    args = list(...)
  )

  # Select the first instance
  data@data <- head(data@data, n = 1L)

  return(data)
}



test_create_synthetic_correlated_one_outcome_data <- function(
    ...,
    outcome_type
) {
  
  # Create test data.
  data <- do.call(
    test_create_synthetic_correlated_data,
    args = c(
      list("outcome_type" = outcome_type),
      list(...)
    )
  )

  if (outcome_type %in% c("binomial", "multinomial")) {
    data@data[, "outcome" := "0"]
  } else if (outcome_type %in% c("continuous")) {
    data@data[, "outcome" := 1.0]
  } else if (outcome_type == "survival") {
    data@data[, ":="("outcome_time" = 1.25, "outcome_event" = 1L)]
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }

  return(data)
}



test_create_synthetic_correlated_bad_outcome_data <- function(
    ...,
    outcome_type
) {
  # Create test data.
  data <- do.call(
    test_create_synthetic_correlated_data,
    args = c(
      list("outcome_type" = outcome_type),
      list(...)
    )
  )

  if (outcome_type %in% c("binomial", "multinomial")) {
    data@data[, "outcome" := "0"]
  } else if (outcome_type %in% c("continuous")) {
    data@data[, "outcome" := 1.0]
  } else if (outcome_type == "survival") {
    data@data[, ":="("outcome_event" = 0L)]
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }

  return(data)
}



test_data_drop_rare_feature_levels <- function(data) {
  # Replace rare levels in feature data.

  # Prevent NOTE by non-standard use in data.table.
  n <- NULL

  for (ii in get_feature_columns(data)) {
    # Skip numeric data.
    if (is.numeric(data@data[[ii]])) next

    # Find the rarest level of the categorical feature, and set replacement
    # level.
    level_data <- data@data[, list("n" = .N), by = c(ii)][order(n)]
    rare_level <- head(level_data[[ii]], n = 1L)
    replacement_level <- tail(level_data[[ii]], n = 1L)
    data@data[get(ii) == rare_level, (ii) := replacement_level]
  }

  return(data)
}

if (!rlang::is_installed("power.transform")) testthat::skip()

# Generic test -----------------------------------------------------------------
outcome_type <- "survival"

for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (transformation_method in familiar:::.get_available_transformation_methods()) {
    testthat::test_that(
      paste0(
        "Transformation is correctly performed using the ", transformation_method,
        " method and ", n_numeric_features, " numeric features."
      ),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Add skeletons.
        feature_info_list <- familiar:::create_transformation_parameter_skeleton(
          feature_info_list = feature_info_list,
          transformation_method = transformation_method)

        # Update the feature info list.
        feature_info_list <- familiar:::add_transformation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Perform the transformation.
        data_copy <- familiar:::transform_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Test whether the features are transformed (unless none).
        if (transformation_method == "none") {
          # Check that the data is not altered.
          testthat::expect_true(data.table::fsetequal(data_copy@data, data@data))
          
        } else {
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              # Expect that values are finite.
              testthat::expect_true(is.finite(power.transform::get_lambda(
                feature_info_list[[feature]]@transformation_parameters@transformer)))
              
            } else {
              # For categorical features test that the object method is "none".
              testthat::expect_equal(
                feature_info_list[[feature]]@transformation_parameters@method,
                "none")
            }
          }
        }

        # Assert that inverting the transformation produces the original
        # dataset.
        data_restored <- familiar:::transform_features(
          data = data_copy,
          feature_info_list = feature_info_list,
          invert = TRUE)

        # Iterate over features and compare. They should be equal
        for (feature in familiar:::get_feature_columns(data_restored)) {
          # Expect that values are similar to a tolerance.
          testthat::expect_equal(data_restored@data[[feature]], data@data[[feature]])
        }

        # Assert that aggregating the transformation parameters works as expected.
        aggr_transform_parameters <- familiar:::..collect_and_aggregate_transformation_info(
          feature_info_list = feature_info_list,
          instance_mask = rep_len(TRUE, length(feature_info_list)),
          feature_name = "test")

        if (n_numeric_features > 0 && transformation_method != "none") {
          # Expect that the selected transformation method matches the selected
          # method -- Note that the method attribute is only the transformation
          # method.
          testthat::expect_true(startsWith(
            x = transformation_method,
            prefix = aggr_transform_parameters$parameters@method))

          # Expect that the lambda is finite.
          testthat::expect_true(is.finite(power.transform::get_lambda(
            aggr_transform_parameters$parameters@transformer)))

          # Expect that instance mask is equal to one.
          testthat::expect_equal(sum(aggr_transform_parameters$instance_mask), 1)
          
        } else {
          # Expect that the selected transformation method matches the selected
          # method.
          testthat::expect_equal(aggr_transform_parameters$parameters@method, "none")

          # Expect that instance mask is equal to the number of numeric
          # features.
          testthat::expect_equal(
            sum(aggr_transform_parameters$instance_mask),
            familiar:::get_n_features(data_copy))
        }
      }
    )
  }
}

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

# NA-value test ----------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (transformation_method in familiar:::.get_available_transformation_methods()) {
    testthat::test_that(
      paste0(
        "Transformation is correctly performed using the ",
        transformation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with some NA data."
      ),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Add skeletons.
        feature_info_list <- familiar:::create_transformation_parameter_skeleton(
          feature_info_list = feature_info_list,
          transformation_method = transformation_method)

        # Update the feature info list.
        feature_info_list <- familiar:::add_transformation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Perform a transformation.
        data_copy <- familiar:::transform_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Test whether the features are transformed (unless none).
        if (transformation_method == "none") {
          # Check that the data is not altered.
          testthat::expect_true(data.table::fsetequal(data_copy@data, data@data))
        } else {
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              # Expect that values are finite.
              testthat::expect_true(is.finite(power.transform::get_lambda(
                feature_info_list[[feature]]@transformation_parameters@transformer)))
              
            } else {
              # For categorical features test that the none method is used.
              testthat::expect_equal(
                feature_info_list[[feature]]@transformation_parameters@method,
                "none")
            }
          }
        }
      }
    )
  }
}

# One feature NA test ----------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_feature_all_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (transformation_method in familiar:::.get_available_transformation_methods()) {
    testthat::test_that(
      paste0(
        "Transformation is correctly performed using the ",
        transformation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with one feature completely NA."
      ),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Add skeletons.
        feature_info_list <- familiar:::create_transformation_parameter_skeleton(
          feature_info_list = feature_info_list,
          transformation_method = transformation_method)

        # Update the feature info list.
        feature_info_list <- familiar:::add_transformation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Perform a normalisation.
        data_copy <- familiar:::transform_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Test whether the features are transformed (unless none).
        if (transformation_method == "none") {
          # Check that the data is not altered.
          testthat::expect_true(data.table::fsetequal(data_copy@data, data@data))
        } else {
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              if (feature == "feature_2") {
                # Expect that for the NA feature, the none method is used.
                testthat::expect_equal(
                  feature_info_list[[feature]]@transformation_parameters@method,
                  "none")
                
              } else {
                # Check that the lambda is finite.
                testthat::expect_true(is.finite(power.transform::get_lambda(
                  feature_info_list[[feature]]@transformation_parameters@transformer)))
              }
            } else {
              # For categorical features test that the none method is used.
              testthat::expect_equal(
                feature_info_list[[feature]]@transformation_parameters@method,
                "none")
            }
          }
        }
      }
    )
  }
}

# Invariant feature test -------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_invariant_feature_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (transformation_method in familiar:::.get_available_transformation_methods()) {
    testthat::test_that(
      paste0(
        "Transformation is correctly performed using the ",
        transformation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with invariant features."
      ),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Add skeletons.
        feature_info_list <- familiar:::create_transformation_parameter_skeleton(
          feature_info_list = feature_info_list,
          transformation_method = transformation_method)

        # Update the feature info list.
        feature_info_list <- familiar:::add_transformation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Perform the transformation.
        data_copy <- familiar:::transform_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Test whether the features are transformed (unless none).
        if (transformation_method == "none") {
          # Check that the data is not altered.
          testthat::expect_true(data.table::fsetequal(data_copy@data, data@data))
          
        } else {
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              # Check the none method is used for invariant features.
              testthat::expect_equal(
                feature_info_list[[feature]]@transformation_parameters@method,
                "none")
            } else {
              # For categorical features test that the none method is used.
              testthat::expect_equal(
                feature_info_list[[feature]]@transformation_parameters@method,
                "none")
            }
          }
        }
      }
    )
  }
}

# One feature invariant test ---------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_feature_invariant_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (transformation_method in familiar:::.get_available_transformation_methods()) {
    testthat::test_that(
      paste0(
        "Transformation is correctly performed using the ",
        transformation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with one invariant feature."
      ),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Add skeletons.
        feature_info_list <- familiar:::create_transformation_parameter_skeleton(
          feature_info_list = feature_info_list,
          transformation_method = transformation_method)

        # Update the feature info list.
        feature_info_list <- familiar:::add_transformation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Perform the transformation.
        data_copy <- familiar:::transform_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Test whether the features are transformed (unless none).
        if (transformation_method == "none") {
          # Check that the data is not altered.
          testthat::expect_true(data.table::fsetequal(data_copy@data, data@data))
          
        } else {
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              if (feature == "feature_2") {
                # Expect that the invariant feature uses the none method.
                testthat::expect_equal(
                  feature_info_list[[feature]]@transformation_parameters@method,
                  "none")
              } else {
                # Check that the lambda is finite.
                testthat::expect_true(is.finite(power.transform::get_lambda(
                  feature_info_list[[feature]]@transformation_parameters@transformer)))
              }
            } else {
              # For categorical features test that the none method is used.
              testthat::expect_equal(
                feature_info_list[[feature]]@transformation_parameters@method,
                "none")
            }
          }
        }
      }
    )
  }
}

# One-sample test --------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_sample_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (transformation_method in familiar:::.get_available_transformation_methods()) {
    testthat::test_that(
      paste0(
        "Transformation is correctly performed using the ",
        transformation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with one sample."
      ),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Add skeletons.
        feature_info_list <- familiar:::create_transformation_parameter_skeleton(
          feature_info_list = feature_info_list,
          transformation_method = transformation_method)

        # Update the feature info list.
        feature_info_list <- familiar:::add_transformation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Perform the transformation.
        data_copy <- familiar:::transform_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Test whether the features are transformed (unless none).
        if (transformation_method == "none") {
          # Check that the data is not altered.
          testthat::expect_true(data.table::fsetequal(data_copy@data, data@data))
          
        } else {
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              # Expect that the none-method is used for one-sample data.
              testthat::expect_equal(
                feature_info_list[[feature]]@transformation_parameters@method,
                "none")
            } else {
              # For categorical features test that the none method is used.
              testthat::expect_equal(
                feature_info_list[[feature]]@transformation_parameters@method,
                "none")
            }
          }
        }
      }
    )
  }
}

# Mix of methods ---------------------------------------------------------------

data <- familiar:::test_create_synthetic_series_data(
  outcome_type = outcome_type,
  n_numeric = 4)

testthat::test_that(
  paste0("Transformation is correctly performed when mixing transformation methods."),
  {
    # Make a copy of the data.
    data_copy <- data.table::copy(data)

    # Create a list of featureInfo objects.
    feature_info_list <- test_create_generic_info(data = data_copy)

    # Add skeletons for features 1 and 2.
    feature_info_list <- familiar:::create_transformation_parameter_skeleton(
      feature_info_list = feature_info_list,
      transformation_method = "box_cox",
      feature_names = c("feature_1", "feature_2"))

    # Check that this only updates features 1 and 2.
    testthat::expect_equal(
      feature_info_list$feature_1@transformation_parameters@fitting_parameters$method,
      "box_cox")
    testthat::expect_equal(
      feature_info_list$feature_2@transformation_parameters@fitting_parameters$method,
      "box_cox")
    testthat::expect_equal(feature_info_list$feature_3@transformation_parameters, NULL)
    testthat::expect_equal(feature_info_list$feature_4@transformation_parameters, NULL)

    # Add skeleton for feature 3.
    feature_info_list <- familiar:::create_transformation_parameter_skeleton(
      feature_info_list = feature_info_list,
      transformation_method = "yeo_johnson",
      feature_names = c("feature_3"))

    # Check that this only updates feature 3.
    testthat::expect_equal(
      feature_info_list$feature_1@transformation_parameters@fitting_parameters$method,
      "box_cox")
    testthat::expect_equal(
      feature_info_list$feature_2@transformation_parameters@fitting_parameters$method,
      "box_cox")
    testthat::expect_equal(
      feature_info_list$feature_3@transformation_parameters@fitting_parameters$method,
      "yeo_johnson")
    testthat::expect_equal(
      feature_info_list$feature_4@transformation_parameters,
      NULL)

    # Add skeletons for feature 4.
    feature_info_list <- familiar:::create_transformation_parameter_skeleton(
      feature_info_list = feature_info_list,
      transformation_method = "none",
      feature_names = c("feature_4"))
    
    # Check that this only updates feature 4.
    testthat::expect_equal(
      feature_info_list$feature_1@transformation_parameters@fitting_parameters$method,
      "box_cox")
    testthat::expect_equal(
      feature_info_list$feature_2@transformation_parameters@fitting_parameters$method,
      "box_cox")
    testthat::expect_equal(
      feature_info_list$feature_3@transformation_parameters@fitting_parameters$method,
      "yeo_johnson")
    testthat::expect_equal(
      feature_info_list$feature_4@transformation_parameters@fitting_parameters$method,
      "none")

    # Update the feature info list.
    feature_info_list <- familiar:::add_transformation_parameters(
      feature_info_list = feature_info_list,
      data = data_copy)

    # Perform the transformation.
    data_copy <- familiar:::transform_features(
      data = data_copy,
      feature_info_list = feature_info_list)

    # Assert that inverting the transformation produces the original dataset.
    data_restored <- familiar:::transform_features(
      data = data_copy,
      feature_info_list = feature_info_list,
      invert = TRUE)

    # Iterate over features and compare. They should be equal
    for (feature in familiar:::get_feature_columns(data_restored)) {
      # Expect that values are similar to a tolerance.
      testthat::expect_equal(data_restored@data[[feature]], data@data[[feature]])
    }
  }
)

# Skip on CRAN as this test takes about a 30 seconds.
testthat::skip_on_cran()
testthat::skip_on_ci()

outcome_type <- "survival"

# Generic test -----------------------------------------------------------------
for (n_numeric_features in c(4L, 3L, 2L, 1L, 0L)) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features
  )

  for (batch_normalisation_method in familiar:::.get_available_batch_normalisation_methods()) {
    testthat::test_that(paste0(
      "Batch normalisation is correctly performed using the ",
      batch_normalisation_method, " method and ",
      n_numeric_features, " numeric features."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)
      
      # Find batch names.
      batch_ids <- unique(data_copy@data[[familiar:::get_id_columns("batch")]])
      
      # Create a list of featureInfo objects.
      feature_info_list <- familiar:::test_create_generic_info(data = data_copy)

      # Combat requires global standardisation
      if (batch_normalisation_method %in% familiar:::.get_available_batch_normalisation_methods("combat")) {
        # Create normalisation skeletons.
        feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
          feature_info_list = feature_info_list,
          normalisation_method = "standardisation"
        )

        # Update the feature info list with global standardisation parameters.
        feature_info_list <- familiar:::add_normalisation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy
        )

        # Act as if the data has been transformed.
        data_copy@preprocessing_level <- "transformation"

        # Perform a global normalisation.
        data_copy <- familiar:::normalise_features(
          data = data_copy,
          feature_info_list = feature_info_list
        )
      }

      # Create batch normalisation container skeleton.
      feature_info_list <- familiar:::create_batch_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = batch_normalisation_method
      )

      # Add batch normalisation parameters.
      feature_info_list <- familiar:::add_batch_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy
      )

      # Assume that the data is pre-processed.
      data_copy@preprocessing_level <- "normalisation"

      # Attempt to batch normalise the data.
      data_copy <- familiar:::batch_normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list
      )

      # Test whether the features are normalised (unless none).
      if (batch_normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_true(data.table::fsetequal(data_copy@data, data@data))

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Test that the container method matches the batch normalisation
          # method.
          testthat::expect_equal(
            feature_info_list[[feature]]@batch_normalisation_parameters@method,
            batch_normalisation_method
          )

          # Expect that all batches are named.
          testthat::expect_setequal(
            names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
            batch_ids
          )

          # Expect that all batches parameter objects have the none class.
          for (batch_parameter_object in 
               feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters
          ) {
            testthat::expect_s4_class(
              batch_parameter_object,
              "featureInfoParametersNormalisationNone"
            )
            testthat::expect_equal(batch_parameter_object@name, feature)
            testthat::expect_true(batch_parameter_object@batch %in% batch_ids)
          }
        }
        
      } else {
        # Iterate over the different batches to determine if batch corrections
        # were performed correctly.
        for (x in split(data_copy@data, by = "batch_id")) {
          # Find the current batch identifier.
          current_batch_id <- x[["batch_id"]][1L]

          # Iterate over features.
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              # Compute mean.
              mu <- mean(x[[feature]], na.rm = TRUE)

              if (batch_normalisation_method %in% 
                  familiar:::.get_available_batch_normalisation_methods("normalisation")) {
                # Check that the feature is correctly centred around 0.5.
                
                testthat::expect_true(mu < 0.7 & mu > 0.3)
                
              } else if (batch_normalisation_method %in% 
                         familiar:::.get_available_batch_normalisation_methods("combat")) {
                # Check that the feature is correctly centred around 0, but with
                # wider margins for the non-parametric method Check also that
                # the overall mean is around 0.
                if (batch_normalisation_method %in% c("combat", "combat_np", "combat_non_parametric")) {
                  testthat::expect_true(mu < 0.5 & mu > -0.5)
                } else {
                  testthat::expect_true(mu < 0.2 & mu > -0.2)
                }

                # Check overall mean.
                mu_t <- mean(data_copy@data[[feature]], na.rm = TRUE)
                testthat::expect_true(mu_t < 0.2 & mu_t > -0.2)
                
              } else {
                # Check that the feature is correctly centred around 0.
                testthat::expect_true(mu < 0.2 & mu > -0.2)
              }

              # Test that the container method matches the batch normalisation
              # method.
              testthat::expect_equal(
                feature_info_list[[feature]]@batch_normalisation_parameters@method,
                batch_normalisation_method
              )

              # Expect that all batches are named.
              testthat::expect_setequal(
                names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                batch_ids
              )
              
            } else {
              # For categorical features test that the none batch normalisation
              # method is present.
              testthat::expect_equal(
                feature_info_list[[feature]]@batch_normalisation_parameters@method,
                "none"
              )

              # Expect that all batches are named.
              testthat::expect_setequal(
                names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                batch_ids
              )

              # Expect that all batches parameter objects have
              # the none class.
              for (batch_parameter_object in 
                   feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters
              ) {
                testthat::expect_s4_class(batch_parameter_object, "featureInfoParametersNormalisationNone")
                testthat::expect_equal(batch_parameter_object@name, feature)
                testthat::expect_true(batch_parameter_object@batch %in% batch_ids)
              }
            }
          }
        }
      }

      # Assert that inverting the transformation produces the original dataset.
      data_restored <- familiar:::batch_normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list,
        invert = TRUE
      )

      # Inverse the global normalisation.
      if (batch_normalisation_method %in% 
          familiar:::.get_available_batch_normalisation_methods("combat")) {
        # Perform a global normalisation.
        data_restored <- familiar:::normalise_features(
          data = data_restored,
          feature_info_list = feature_info_list,
          invert = TRUE
        )
      }

      # Iterate over features and compare. They should be equal.
      for (feature in familiar:::get_feature_columns(data_restored)) {
        # Expect that values are similar to a tolerance.
        testthat::expect_equal(data_restored@data[[feature]], data@data[[feature]])
      }

      # Assert that aggregating the batch normalisation parameters works as expected.
      aggr_batch_normalisation_parameters <- familiar:::..collect_and_aggregate_batch_normalisation_info(
        feature_info_list = feature_info_list,
        instance_mask = rep_len(TRUE, length(feature_info_list)),
        feature_name = "test"
      )$parameters

      for (batch_norm_parameters in aggr_batch_normalisation_parameters@batch_parameters) {
        if (n_numeric_features > 0L && batch_normalisation_method != "none") {
          # Expect that the selected transformation method matches the selected
          # method, except for combat methods.
          if (n_numeric_features <= 2L && 
              batch_normalisation_method %in% familiar:::.get_available_batch_normalisation_methods("combat")) {
            testthat::expect_s4_class(
              batch_norm_parameters,
              "featureInfoParametersNormalisationStandardisation")
            
          } else {
            testthat::expect_equal(
              batch_norm_parameters@method, 
              batch_normalisation_method)
          }

          # Expect that the shift and scale parameters are not NA.
          testthat::expect_true(is.finite(batch_norm_parameters@shift))

          if (is(batch_norm_parameters, "featureInfoParametersNormalisationShiftScale")) {
            testthat::expect_true(is.finite(batch_norm_parameters@scale))
          }
        } else {
          # Assert that for the categorical features, the none method is used.
          testthat::expect_s4_class(
            batch_norm_parameters,
            "featureInfoParametersNormalisationNone"
          )
        }
      }
    })
  }
}



# NA-value test ----------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (batch_normalisation_method in familiar:::.get_available_batch_normalisation_methods()) {
    testthat::test_that(paste0(
      "Batch normalisation is correctly performed using the ",
      batch_normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with some NA data."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Find batch names.
      batch_ids <- unique(data_copy@data[[familiar:::get_id_columns("batch")]])

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Combat requires global standardisation
      if (batch_normalisation_method %in% 
          familiar:::.get_available_batch_normalisation_methods("combat")) {
        # Create normalisation skeletons.
        feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
          feature_info_list = feature_info_list,
          normalisation_method = "standardisation")

        # Update the feature info list with global standardisation parameters.
        feature_info_list <- familiar:::add_normalisation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Act as if the data has been transformed.
        data_copy@preprocessing_level <- "transformation"

        # Perform a global normalisation.
        data_copy <- familiar:::normalise_features(
          data = data_copy,
          feature_info_list = feature_info_list)
      }

      # Create batch normalisation container skeleton.
      feature_info_list <- familiar:::create_batch_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = batch_normalisation_method)

      # Add batch normalisation parameters.
      feature_info_list <- familiar:::add_batch_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Assume that the data is pre-processed.
      data_copy@preprocessing_level <- "normalisation"

      # Attempt to batch normalise the data.
      data_copy <- familiar:::batch_normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (batch_normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Test that the container method matches the batch normalisation
          # method.
          testthat::expect_equal(
            feature_info_list[[feature]]@batch_normalisation_parameters@method,
            batch_normalisation_method)

          # Expect that all batches are named.
          testthat::expect_setequal(
            names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
            batch_ids)

          # Expect that all batches parameter objects have the none class.
          for (batch_parameter_object in
               feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
            testthat::expect_s4_class(
              batch_parameter_object,
              "featureInfoParametersNormalisationNone")
            testthat::expect_equal(batch_parameter_object@name, feature)
            testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
          }
        }
        
      } else {
        # Iterate over the different batches to determine if batch corrections
        # were performed correctly.
        for (x in split(data_copy@data, by = "batch_id")) {
          # Find the current batch identifier.
          current_batch_id <- x[["batch_id"]][1]

          # Iterate over features.
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              # Compute mean.
              mu <- mean(x[[feature]], na.rm = TRUE)

              if (batch_normalisation_method %in% 
                  familiar:::.get_available_batch_normalisation_methods("normalisation")) {
                # Check that the feature is correctly centred around 0.5.
                testthat::expect_equal(mu < 0.7 & mu > 0.3, TRUE)
                
              } else if (batch_normalisation_method %in% 
                         familiar:::.get_available_batch_normalisation_methods("combat")) {
                # Check that the feature is correctly centred around 0, but with
                # wider margins for the non-parametric method Check also that
                # the overall mean is around 0.
                if (batch_normalisation_method %in% c("combat", "combat_np", "combat_non_parametric")) {
                  testthat::expect_equal(mu < 0.5 & mu > -0.5, TRUE)
                } else {
                  testthat::expect_equal(mu < 0.2 & mu > -0.2, TRUE)
                }

                # Check overall mean.
                mu_t <- mean(data_copy@data[[feature]], na.rm = TRUE)
                testthat::expect_equal(mu_t < 0.2 & mu_t > -0.2, TRUE)
                
              } else {
                # Check that the feature is correctly centred around 0.
                testthat::expect_equal(mu < 0.2 & mu > -0.2, TRUE)
              }

              # Test that the container method matches the batch normalisation
              # method.
              testthat::expect_equal(
                feature_info_list[[feature]]@batch_normalisation_parameters@method,
                batch_normalisation_method)

              # Expect that all batches are named.
              testthat::expect_setequal(
                names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                batch_ids)
              
            } else {
              # For categorical features test that the none batch normalisation
              # method is present.
              testthat::expect_equal(
                feature_info_list[[feature]]@batch_normalisation_parameters@method,
                "none")

              # Expect that all batches are named.
              testthat::expect_setequal(
                names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                batch_ids)

              # Expect that all batches parameter objects have the none class.
              for (batch_parameter_object in 
                   feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                testthat::expect_s4_class(
                  batch_parameter_object,
                  "featureInfoParametersNormalisationNone")
                testthat::expect_equal(batch_parameter_object@name, feature)
                testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
              }
            }
          }
        }
      }
    })
  }
}



#### One feature NA test ##############################################################
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_feature_all_na_data(
    outcome_type = outcome_type, 
    n_numeric = n_numeric_features)

  for (batch_normalisation_method in familiar:::.get_available_batch_normalisation_methods()) {
    testthat::test_that(paste0(
      "Batch normalisation is correctly performed using the ",
      batch_normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with one feature entirely NA."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Find batch names.
      batch_ids <- unique(data_copy@data[[familiar:::get_id_columns("batch")]])

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Combat requires global standardisation
      if (batch_normalisation_method %in% familiar:::.get_available_batch_normalisation_methods("combat")) {
        # Create normalisation skeletons.
        feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
          feature_info_list = feature_info_list,
          normalisation_method = "standardisation")

        # Update the feature info list with global standardisation parameters.
        feature_info_list <- familiar:::add_normalisation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Act as if the data has been transformed.
        data_copy@preprocessing_level <- "transformation"

        # Perform a global normalisation.
        data_copy <- familiar:::normalise_features(
          data = data_copy,
          feature_info_list = feature_info_list)
      }

      # Create batch normalisation container skeleton.
      feature_info_list <- familiar:::create_batch_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = batch_normalisation_method)

      # Add batch normalisation parameters.
      feature_info_list <- familiar:::add_batch_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Assume that the data is pre-processed.
      data_copy@preprocessing_level <- "normalisation"

      # Attempt to batch normalise the data.
      data_copy <- familiar:::batch_normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (batch_normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Test that the container method matches the batch normalisation
          # method.
          testthat::expect_equal(
            feature_info_list[[feature]]@batch_normalisation_parameters@method,
            batch_normalisation_method)

          # Expect that all batches are named.
          testthat::expect_setequal(
            names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
            batch_ids)

          # Expect that all batches parameter objects have the none class.
          for (batch_parameter_object in 
               feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
            testthat::expect_s4_class(
              batch_parameter_object,
              "featureInfoParametersNormalisationNone")
            testthat::expect_equal(batch_parameter_object@name, feature)
            testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
          }
        }
      } else {
        # Iterate over the different batches to determine if batch corrections
        # were performed correctly.
        for (x in split(data_copy@data, by = "batch_id")) {
          # Find the current batch identifier.
          current_batch_id <- x[["batch_id"]][1]

          # Iterate over features.
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              if (feature == "feature_2") {
                # Expect that all batches are named.
                testthat::expect_setequal(
                  names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                  batch_ids)

                # Expect that all batch parameter objects have the none class.
                for (batch_parameter_object in 
                     feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                  testthat::expect_s4_class(
                    batch_parameter_object,
                    "featureInfoParametersNormalisationNone")
                  testthat::expect_equal(batch_parameter_object@name, feature)
                  testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
                }
                
              } else {
                # Compute mean.
                mu <- mean(x[[feature]], na.rm = TRUE)

                if (batch_normalisation_method %in% 
                    familiar:::.get_available_batch_normalisation_methods("normalisation")) {
                  # Check that the feature is correctly centred around 0.5.
                  testthat::expect_equal(mu < 0.7 & mu > 0.3, TRUE)
                  
                } else if (batch_normalisation_method %in% 
                           familiar:::.get_available_batch_normalisation_methods("combat")) {
                  # Check that the feature is correctly centred around 0, but
                  # with wider margins for the non-parametric method Check also
                  # that the overall mean is around 0.
                  if (batch_normalisation_method %in% c("combat", "combat_np", "combat_non_parametric")) {
                    testthat::expect_equal(mu < 0.5 & mu > -0.5, TRUE)
                  } else {
                    testthat::expect_equal(mu < 0.2 & mu > -0.2, TRUE)
                  }

                  # Check overall mean.
                  mu_t <- mean(data_copy@data[[feature]], na.rm = TRUE)
                  testthat::expect_equal(mu_t < 0.2 & mu_t > -0.2, TRUE)
                  
                } else {
                  # Check that the feature is correctly
                  # centred around 0.
                  testthat::expect_equal(mu < 0.2 & mu > -0.2, TRUE)
                }

                # Test that the container method matches the batch normalisation
                # method.
                testthat::expect_equal(
                  feature_info_list[[feature]]@batch_normalisation_parameters@method,
                  batch_normalisation_method)

                # Expect that all batches are named.
                testthat::expect_setequal(
                  names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                  batch_ids)
              }
              
            } else {
              # For categorical features test that the none batch normalisation
              # method is present.
              testthat::expect_equal(
                feature_info_list[[feature]]@batch_normalisation_parameters@method,
                "none")

              # Expect that all batches are named.
              testthat::expect_setequal(
                names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                batch_ids)

              # Expect that all batch parameter objects have
              # the none class.
              for (batch_parameter_object in 
                   feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                testthat::expect_s4_class(
                  batch_parameter_object,
                  "featureInfoParametersNormalisationNone")
                testthat::expect_equal(batch_parameter_object@name, feature)
                testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
              }
            }
          }
        }
      }
    })
  }
}



# Invariant feature test -------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_invariant_feature_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (batch_normalisation_method in familiar:::.get_available_batch_normalisation_methods()) {
    testthat::test_that(paste0(
      "Batch normalisation is correctly performed using the ",
      batch_normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with invariant features."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Find batch names.
      batch_ids <- unique(data_copy@data[[familiar:::get_id_columns("batch")]])

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Combat requires global standardisation
      if (batch_normalisation_method %in% familiar:::.get_available_batch_normalisation_methods("combat")) {
        # Create normalisation skeletons.
        feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
          feature_info_list = feature_info_list,
          normalisation_method = "standardisation")

        # Update the feature info list with global standardisation parameters.
        feature_info_list <- familiar:::add_normalisation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Act as if the data has been transformed.
        data_copy@preprocessing_level <- "transformation"

        # Perform a global normalisation.
        data_copy <- familiar:::normalise_features(
          data = data_copy,
          feature_info_list = feature_info_list)
      }

      # Create batch normalisation container skeleton.
      feature_info_list <- familiar:::create_batch_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = batch_normalisation_method)

      # Add batch normalisation parameters.
      feature_info_list <- familiar:::add_batch_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Assume that the data is pre-processed.
      data_copy@preprocessing_level <- "normalisation"

      # Attempt to batch normalise the data.
      data_copy <- familiar:::batch_normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (batch_normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Test that the container method matches the batch normalisation
          # method.
          testthat::expect_equal(
            feature_info_list[[feature]]@batch_normalisation_parameters@method,
            batch_normalisation_method)

          # Expect that all batches are named.
          testthat::expect_setequal(
            names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
            batch_ids)

          # Expect that all batches parameter objects have
          # the none class.
          for (batch_parameter_object in feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
            testthat::expect_s4_class(batch_parameter_object, "featureInfoParametersNormalisationNone")
            testthat::expect_equal(batch_parameter_object@name, feature)
            testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
          }
        }
      } else {
        # Iterate over the different batches to determine if batch corrections
        # were performed correctly.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          for (x in split(data_copy@data, by = "batch_id")) {
            # Find the current batch identifier.
            current_batch_id <- x[["batch_id"]][1]

            # Iterate over features.
            for (feature in familiar:::get_feature_columns(data_copy)) {
              # Determine if the feature is numeric.
              if (feature_info_list[[feature]]@feature_type == "numeric") {
                # Expect that all batches are named.
                testthat::expect_setequal(
                  names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                  batch_ids)

                # Expect that all batch parameter objects have the none class.
                for (batch_parameter_object in 
                     feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                  testthat::expect_s4_class(
                    batch_parameter_object,
                    "featureInfoParametersNormalisationNone")
                  testthat::expect_equal(batch_parameter_object@name, feature)
                  testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
                }
                
              } else {
                # For categorical features test that the none batch
                # normalisation method is present.
                testthat::expect_equal(
                  feature_info_list[[feature]]@batch_normalisation_parameters@method,
                  "none")

                # Expect that all batches are named.
                testthat::expect_setequal(
                  names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                  batch_ids)

                # Expect that all batches parameter objects have the none class.
                for (batch_parameter_object in 
                     feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                  testthat::expect_s4_class(
                    batch_parameter_object,
                    "featureInfoParametersNormalisationNone")
                  testthat::expect_equal(batch_parameter_object@name, feature)
                  testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
                }
              }
            }
          }
        }
      }
    })
  }
}



# One feature invariant test ---------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_feature_invariant_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (batch_normalisation_method in 
       familiar:::.get_available_batch_normalisation_methods()) {
    testthat::test_that(paste0(
      "Batch normalisation is correctly performed using the ",
      batch_normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with one invariant feature."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Find batch names.
      batch_ids <- unique(data_copy@data[[familiar:::get_id_columns("batch")]])

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Combat requires global standardisation
      if (batch_normalisation_method %in% 
          familiar:::.get_available_batch_normalisation_methods("combat")) {
        # Create normalisation skeletons.
        feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
          feature_info_list = feature_info_list,
          normalisation_method = "standardisation")

        # Update the feature info list with global standardisation parameters.
        feature_info_list <- familiar:::add_normalisation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Act as if the data has been transformed.
        data_copy@preprocessing_level <- "transformation"

        # Perform a global normalisation.
        data_copy <- familiar:::normalise_features(
          data = data_copy,
          feature_info_list = feature_info_list)
      }

      # Create batch normalisation container skeleton.
      feature_info_list <- familiar:::create_batch_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = batch_normalisation_method)

      # Add batch normalisation parameters.
      feature_info_list <- familiar:::add_batch_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Assume that the data is pre-processed.
      data_copy@preprocessing_level <- "normalisation"

      # Attempt to batch normalise the data.
      data_copy <- familiar:::batch_normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (batch_normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Test that the container method matches the batch normalisation
          # method.
          testthat::expect_equal(
            feature_info_list[[feature]]@batch_normalisation_parameters@method,
            batch_normalisation_method)

          # Expect that all batches are named.
          testthat::expect_setequal(
            names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
            batch_ids)

          # Expect that all batches parameter objects have the none class.
          for (batch_parameter_object in 
               feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
            testthat::expect_s4_class(
              batch_parameter_object,
              "featureInfoParametersNormalisationNone")
            testthat::expect_equal(batch_parameter_object@name, feature)
            testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
          }
        }
        
      } else {
        # Iterate over the different batches to determine if batch corrections
        # were performed correctly.
        for (x in split(data_copy@data, by = "batch_id")) {
          # Find the current batch identifier.
          current_batch_id <- x[["batch_id"]][1]

          # Iterate over features.
          for (feature in familiar:::get_feature_columns(data_copy)) {
            # Determine if the feature is numeric.
            if (feature_info_list[[feature]]@feature_type == "numeric") {
              if (feature == "feature_2") {
                # Expect that all batches are named.
                testthat::expect_setequal(
                  names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                  batch_ids)

                # Expect that all batch parameter objects have the none class.
                for (batch_parameter_object in 
                     feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                  testthat::expect_s4_class(
                    batch_parameter_object,
                    "featureInfoParametersNormalisationNone")
                  testthat::expect_equal(batch_parameter_object@name, feature)
                  testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
                }
                
              } else {
                # Compute mean.
                mu <- mean(x[[feature]], na.rm = TRUE)

                if (batch_normalisation_method %in% 
                    familiar:::.get_available_batch_normalisation_methods("normalisation")) {
                  # Check that the feature is correctly centred around 0.5.
                  testthat::expect_equal(mu < 0.7 & mu > 0.3, TRUE)
                  
                } else if (batch_normalisation_method %in% 
                           familiar:::.get_available_batch_normalisation_methods("combat")) {
                  # Check that the feature is correctly centred around 0, but
                  # with wider margins for the non-parametric method Check also
                  # that the overall mean is around 0.
                  if (batch_normalisation_method %in% c("combat", "combat_np", "combat_non_parametric")) {
                    testthat::expect_equal(mu < 0.5 & mu > -0.5, TRUE)
                  } else {
                    testthat::expect_equal(mu < 0.2 & mu > -0.2, TRUE)
                  }

                  # Check overall mean.
                  mu_t <- mean(data_copy@data[[feature]], na.rm = TRUE)
                  testthat::expect_equal(mu_t < 0.2 & mu_t > -0.2, TRUE)
                  
                } else {
                  # Check that the feature is correctly centred around 0.
                  testthat::expect_equal(mu < 0.2 & mu > -0.2, TRUE)
                }

                # Test that the container method matches the batch normalisation
                # method.
                testthat::expect_equal(
                  feature_info_list[[feature]]@batch_normalisation_parameters@method,
                  batch_normalisation_method)

                # Expect that all batches are named.
                testthat::expect_setequal(
                  names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                  batch_ids)
              }
            } else {
              # For categorical features test that the none batch normalisation
              # method is present.
              testthat::expect_equal(
                feature_info_list[[feature]]@batch_normalisation_parameters@method,
                "none")

              # Expect that all batches are named.
              testthat::expect_setequal(
                names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                batch_ids
              )

              # Expect that all batch parameter objects have the none class.
              for (batch_parameter_object in 
                   feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                testthat::expect_s4_class(
                  batch_parameter_object,
                  "featureInfoParametersNormalisationNone")
                testthat::expect_equal(batch_parameter_object@name, feature)
                testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
              }
            }
          }
        }
      }
    })
  }
}


# One-sample test --------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_sample_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (batch_normalisation_method in familiar:::.get_available_batch_normalisation_methods()) {
    testthat::test_that(paste0(
      "Batch normalisation is correctly performed using the ",
      batch_normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with invariant features."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Find batch names.
      batch_ids <- unique(data_copy@data[[familiar:::get_id_columns("batch")]])

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)
      
      # Combat requires global standardisation
      if (batch_normalisation_method %in% familiar:::.get_available_batch_normalisation_methods("combat")) {
        # Create normalisation skeletons.
        feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
          feature_info_list = feature_info_list,
          normalisation_method = "standardisation")

        # Update the feature info list with global standardisation parameters.
        feature_info_list <- familiar:::add_normalisation_parameters(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Act as if the data has been transformed.
        data_copy@preprocessing_level <- "transformation"

        # Perform a global normalisation.
        data_copy <- familiar:::normalise_features(
          data = data_copy,
          feature_info_list = feature_info_list)
      }

      # Create batch normalisation container skeleton.
      feature_info_list <- familiar:::create_batch_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = batch_normalisation_method)

      # Add batch normalisation parameters.
      feature_info_list <- familiar:::add_batch_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Assume that the data is pre-processed.
      data_copy@preprocessing_level <- "normalisation"

      # Attempt to batch normalise the data.
      data_copy <- familiar:::batch_normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (batch_normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Test that the container method matches the batch normalisation
          # method.
          testthat::expect_equal(
            feature_info_list[[feature]]@batch_normalisation_parameters@method,
            batch_normalisation_method)

          # Expect that all batches are named.
          testthat::expect_setequal(
            names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
            batch_ids)

          # Expect that all batches parameter objects have the none class.
          for (batch_parameter_object in 
               feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
            testthat::expect_s4_class(
              batch_parameter_object,
              "featureInfoParametersNormalisationNone")
            testthat::expect_equal(batch_parameter_object@name, feature)
            testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
          }
        }
        
      } else {
        # Iterate over the different batches to determine if batch corrections
        # were performed correctly.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          for (x in split(data_copy@data, by = "batch_id")) {
            # Find the current batch identifier.
            current_batch_id <- x[["batch_id"]][1]

            # Iterate over features.
            for (feature in familiar:::get_feature_columns(data_copy)) {
              # Determine if the feature is numeric.
              if (feature_info_list[[feature]]@feature_type == "numeric") {
                # Expect that all batches are named.
                testthat::expect_setequal(
                  names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                  batch_ids)

                # Expect that all batch parameter objects have the none class.
                for (batch_parameter_object in 
                     feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                  testthat::expect_s4_class(
                    batch_parameter_object,
                    "featureInfoParametersNormalisationNone")
                  testthat::expect_equal(batch_parameter_object@name, feature)
                  testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
                }
                
              } else {
                # For categorical features test that the none batch
                # normalisation method is present.
                testthat::expect_equal(
                  feature_info_list[[feature]]@batch_normalisation_parameters@method,
                  "none")

                # Expect that all batches are named.
                testthat::expect_setequal(
                  names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
                  batch_ids)

                # Expect that all batches parameter objects have the none class.
                for (batch_parameter_object in 
                     feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
                  testthat::expect_s4_class(
                    batch_parameter_object,
                    "featureInfoParametersNormalisationNone")
                  testthat::expect_equal(batch_parameter_object@name, feature)
                  testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
                }
              }
            }
          }
        }
      }
    })
  }
}



# One-batch NA test ------------------------------------------------------------
# Test what happens when all data for feature 1 in batch 1 are missing.
data <- familiar:::test_create_synthetic_series_data(outcome_type = outcome_type)
data@data[batch_id == "1", "feature_1" := NA_real_]

for (batch_normalisation_method in familiar:::.get_available_batch_normalisation_methods()) {
  testthat::test_that(paste0(
    "Batch normalisation is correctly performed using the ",
    batch_normalisation_method,
    " method, when data for feature_1 in batch 1 is missing."
  ), {
    # Make a copy of the data.
    data_copy <- data.table::copy(data)

    # Find batch names.
    batch_ids <- unique(data_copy@data[[familiar:::get_id_columns("batch")]])

    # Create a list of featureInfo objects.
    feature_info_list <- test_create_generic_info(data = data_copy)

    # Combat requires global standardisation
    if (batch_normalisation_method %in% 
        familiar:::.get_available_batch_normalisation_methods("combat")) {
      # Create normalisation skeletons.
      feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = "standardisation")

      # Update the feature info list with global standardisation parameters.
      feature_info_list <- familiar:::add_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Act as if the data has been transformed.
      data_copy@preprocessing_level <- "transformation"

      # Perform a global normalisation.
      data_copy <- familiar:::normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)
    }

    # Create batch normalisation container skeleton.
    feature_info_list <- familiar:::create_batch_normalisation_parameter_skeleton(
      feature_info_list = feature_info_list,
      normalisation_method = batch_normalisation_method)

    # Add batch normalisation parameters.
    feature_info_list <- familiar:::add_batch_normalisation_parameters(
      feature_info_list = feature_info_list,
      data = data_copy)

    # Assume that the data is pre-processed.
    data_copy@preprocessing_level <- "normalisation"

    # Attempt to batch normalise the data.
    data_copy <- familiar:::batch_normalise_features(
      data = data_copy,
      feature_info_list = feature_info_list)

    # Test whether the features are normalised (unless none).
    if (batch_normalisation_method == "none") {
      # Check that the data is not altered.
      testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

      # Iterate over features.
      for (feature in familiar:::get_feature_columns(data_copy)) {
        # Test that the container method matches the batch normalisation method.
        testthat::expect_equal(
          feature_info_list[[feature]]@batch_normalisation_parameters@method,
          batch_normalisation_method)

        # Expect that all batches are named.
        testthat::expect_setequal(
          names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
          batch_ids)

        # Expect that all batches parameter objects have the none class.
        for (batch_parameter_object in 
             feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
          testthat::expect_s4_class(
            batch_parameter_object,
            "featureInfoParametersNormalisationNone")
          testthat::expect_equal(batch_parameter_object@name, feature)
          testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
        }
      }
      
    } else {
      # Iterate over the different batches to determine if batch corrections
      # were performed correctly.
      for (x in split(data_copy@data, by = "batch_id")) {
        # Find the current batch identifier.
        current_batch_id <- x[["batch_id"]][1]

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          if (feature == "feature_1" && current_batch_id == "1") {
            # Check that parameter data are inferred.

            # Check that the normalisation object is not none.
            testthat::expect_equal(
              is(
                feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters[[current_batch_id]],
                "featureInfoParametersNormalisationNone"),
              FALSE)

            # Check that the normalisation object is complete.
            testthat::expect_equal(
              familiar:::feature_info_complete(
                feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters[[current_batch_id]]),
              TRUE)
            
          } else if (feature_info_list[[feature]]@feature_type == "numeric") {
            # Determine if the feature is numeric.

            # Compute mean.
            mu <- mean(x[[feature]], na.rm = TRUE)

            if (batch_normalisation_method %in% 
                familiar:::.get_available_batch_normalisation_methods("normalisation")) {
              # Check that the feature is correctly centred around 0.5.
              testthat::expect_equal(mu < 0.7 & mu > 0.3, TRUE)
              
            } else if (batch_normalisation_method %in% 
                       familiar:::.get_available_batch_normalisation_methods("combat")) {
              # Check that the feature is correctly centred around 0, but with
              # wider margins for the non-parametric method Check also that the
              # overall mean is around 0.
              if (batch_normalisation_method %in% c("combat", "combat_np", "combat_non_parametric")) {
                testthat::expect_equal(mu < 0.5 & mu > -0.5, TRUE)
              } else {
                testthat::expect_equal(mu < 0.2 & mu > -0.2, TRUE)
              }

              # Check overall mean.
              mu_t <- mean(data_copy@data[[feature]], na.rm = TRUE)
              testthat::expect_equal(mu_t < 0.2 & mu_t > -0.2, TRUE)
              
            } else {
              # Check that the feature is correctly centred around 0.
              testthat::expect_equal(mu < 0.2 & mu > -0.2, TRUE)
            }

            # Test that the container method matches the batch normalisation
            # method.
            testthat::expect_equal(
              feature_info_list[[feature]]@batch_normalisation_parameters@method,
              batch_normalisation_method)

            # Expect that all batches are named.
            testthat::expect_setequal(
              names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
              batch_ids)
            
          } else {
            # For categorical features test that the none batch normalisation
            # method is present.
            testthat::expect_equal(
              feature_info_list[[feature]]@batch_normalisation_parameters@method,
              "none")

            # Expect that all batches are named.
            testthat::expect_setequal(
              names(feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters),
              batch_ids)

            # Expect that all batches parameter objects have the none class.
            for (batch_parameter_object in 
                 feature_info_list[[feature]]@batch_normalisation_parameters@batch_parameters) {
              testthat::expect_s4_class(
                batch_parameter_object,
                "featureInfoParametersNormalisationNone")
              testthat::expect_equal(batch_parameter_object@name, feature)
              testthat::expect_equal(batch_parameter_object@batch %in% batch_ids, TRUE)
            }
          }
        }
      }
    }
  })
}

# Don't perform any further tests on CRAN due to time of running the test.
testthat::skip_on_cran()
testthat::skip_on_ci()

verbose <- FALSE

# Test acquisition functions for all hyperparameter learners -------------------
for (hyperparameter_learner in familiar:::.get_available_hyperparameter_learners()) {
  for (acquisition_function in familiar:::.get_available_acquisition_functions()) {
    familiar:::test_hyperparameter_optimisation(
      learners = "glm_logistic",
      outcome_type_available = "binomial",
      acquisition_function = acquisition_function,
      hyperparameter_learner = hyperparameter_learner,
      debug = FALSE,
      parallel = FALSE)
  }
}

# Test optimisation functions for one metric -----------------------------------
for (optimisation_function in familiar:::.get_available_optimisation_functions()) {
  familiar:::test_hyperparameter_optimisation(
    learners = "glm_logistic",
    outcome_type_available = "binomial",
    optimisation_function = optimisation_function,
    debug = FALSE,
    parallel = FALSE)
}

# Test optimisation functions for multiple metrics -----------------------------
for (optimisation_function in familiar:::.get_available_optimisation_functions()) {
  familiar:::test_hyperparameter_optimisation(
    learners = "glm_logistic",
    outcome_type_available = "binomial",
    optimisation_function = optimisation_function,
    metric = c("auc", "brier", "balanced_accuracy"),
    debug = FALSE,
    parallel = FALSE)
}

# Test hyperparameter learners for learner with only one hyperparameter --------
for (hyperparameter_learner in familiar:::.get_available_hyperparameter_learners()) {
  familiar:::test_hyperparameter_optimisation(
    learners = "cox",
    outcome_type_available = "survival",
    hyperparameter_learner = hyperparameter_learner,
    debug = FALSE,
    parallel = FALSE)
}


# Test without measuring time --------------------------------------------------
familiar:::test_hyperparameter_optimisation(
  learners = "glm_logistic",
  outcome_type_available = "binomial",
  measure_time = FALSE,
  debug = FALSE,
  parallel = FALSE)


# Test that "none" feature selection keeps all features ------------------------
..generate_hyperparameters <- function(
    data,
    learner,
    vimp_method,
    set_signature = FALSE,
    cluster_method = "none",
    cluster_similarity_metric = "mcfadden_r2",
    cluster_similarity_threshold = 0.90,
    use_vimp = "use_hpo_vimp",
    ...
) {
  # Reconstitute settings from the data.
  settings <- familiar:::extract_settings_from_data(data = data)
  
  # Update some missing settings that can be fixed within this method.
  settings$data$train_cohorts <- unique(data@data[[familiar:::get_id_columns(single_column = "batch")]])
  if (set_signature) {
    settings$data$signature <- familiar:::get_feature_columns(data)[1L:2L]
  }
  
  # Parse the remaining settings that are important.
  settings <- do.call(
    familiar:::.parse_general_settings,
    args = c(
      list(
        "settings" = settings,
        "data" = data@data,
        "vimp_method" = vimp_method,
        "learner" = learner,
        "cluster_method" = cluster_method,
        "cluster_similarity_metric" = cluster_similarity_metric,
        "cluster_similarity_threshold" = cluster_similarity_threshold
      )
    )
  )
  
  # Create task.
  task <- methods::new(
    "familiarTaskLearnerHyperparameters",
    vimp_method = vimp_method,
    learner = learner,
    use_vimp = use_vimp
  )
  
  # Optimise hyperparameters.
  new_object <- familiar:::.perform_task(
    object = task,
    data = data,
    settings = settings,
    ...
  )
  
  return(new_object)
}

# Create dataset.
data <- familiar:::test_create_good_data(outcome_type = "binomial")

# Optimise hyperparameters.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "none",
  learner = "elastic_net",
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 3L,
  n_max_intensify_steps = 2L,
  n_random_sets = 20L,
  n_challengers = 10L,
  verbose = verbose
)

testthat::test_that("Test that \"none\" feature selection keeps all features.", {
  testthat::expect_true(all(new_object@hyperparameter_data$parameter_table$sign_size == familiar:::get_n_features(data)))
})



# Test that "random" feature selection can select up to the maximum number of features ------

# Optimise hyperparameters.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "random",
  learner = "elastic_net",
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 3L,
  n_max_intensify_steps = 2L,
  n_random_sets = 20L,
  n_challengers = 10L,
  verbose = verbose
)

testthat::test_that("Test that \"random\" feature selection can select up to the maximum number of features.", {
  testthat::expect_true(all(new_object@hyperparameter_data$parameter_table$sign_size >= 1L))
  testthat::expect_true(all(new_object@hyperparameter_data$parameter_table$sign_size <= familiar:::get_n_features(data)))
})



# Test that "signature_only" keeps only signature features ---------------------

# Optimise hyperparameters.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "signature_only",
  learner = "elastic_net",
  set_signature = TRUE,
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 3L,
  n_max_intensify_steps = 2L,
  n_random_sets = 20L,
  n_challengers = 10L,
  verbose = verbose
)

testthat::test_that("Test that \"signature_only\" feature selection keeps only signature features.", {
  testthat::expect_true(all(new_object@hyperparameter_data$parameter_table$sign_size == 2L))
})



# Test that a range of signature sizes can be provided -------------------------

# Optimise hyperparameters.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "mim",
  learner = "elastic_net",
  set_signature = FALSE,
  hyperparameters = list("sign_size" = c(2, 5)),
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 3L,
  n_max_intensify_steps = 2L,
  n_random_sets = 20L,
  n_challengers = 10L,
  verbose = verbose
)

testthat::test_that("Test that setting a range of signature sizes works correctly.", {
  testthat::expect_true(all(new_object@hyperparameter_data$parameter_table$sign_size >= 2L))
  testthat::expect_true(all(new_object@hyperparameter_data$parameter_table$sign_size <= 5L))
  testthat::expect_true(all(unique(new_object@hyperparameter_data$parameter_table$sign_size) %in% 2L:5L))
})



# Test that a range of signature sizes can be provided -------------------------

# Optimise hyperparameters.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "mim",
  learner = "elastic_net",
  set_signature = FALSE,
  hyperparameters = list("sign_size" = c(1, 4, 6)),
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 3L,
  n_max_intensify_steps = 2L,
  n_random_sets = 20L,
  n_challengers = 10L,
  verbose = verbose
)

testthat::test_that("Test that \"signature_only\" feature selection keeps only signature features.", {
  testthat::expect_setequal(
    unique(new_object@hyperparameter_data$parameter_table$sign_size),
    c(1, 4, 6)
  )
})


# Test exploration methods -----------------------------------------------------

# Optimise hyperparameters.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "mim",
  learner = "elastic_net",
  set_signature = FALSE,
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 1L,
  n_max_intensify_steps = 4L,
  n_intensify_step_bootstraps = 1L,
  n_random_sets = 16L,
  n_challengers = 10L,
  exploration_method = "none",
  verbose = verbose
)

# Set expected range of rows. Upper and lower boundary are the same, as all runs
# are executed simultaneously.
expected_rows_lower <- expected_rows_upper <- (16 + 10 * 4 + 1 * 4) * 2

testthat::test_that(paste0(
  "Test that \"none\" exploration method does not prune any hyperparameter sets ",
  "during intensification"), {
    testthat::expect_lte(nrow(new_object@hyperparameter_data$score_table), expected_rows_upper)
    testthat::expect_gte(nrow(new_object@hyperparameter_data$score_table), expected_rows_lower)
  }
)



# Hyperparameter optimisation using successive_halving for pruning. Note that
# n_max_intensify_steps is 5, but only 4 will be steps are possible. Just as a
# test.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "mim",
  learner = "elastic_net",
  set_signature = FALSE,
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 1L,
  n_max_intensify_steps = 5L,
  n_intensify_step_bootstraps = 1L,
  n_random_sets = 16L,
  n_challengers = 10L,
  exploration_method = "successive_halving",
  verbose = verbose
)

# Set expected range of rows. 10 initial challengers decrease to 5, 2 and 1 in
# subsequent rounds. Upper and lower boundary are the same because here
# n_intensify_step_bootstraps = 1, and only one new run will be assessed for
# each parameter set.
expected_rows_lower <- expected_rows_upper <- (16 + 10 + 5 + 2 + 1 + 4) * 2

testthat::test_that(paste0(
  "Test that \"successive_halving\" exploration method may prune any ",
  "hyperparameter sets during intensification"), {
    testthat::expect_lte(nrow(new_object@hyperparameter_data$score_table), expected_rows_upper)
    testthat::expect_gte(nrow(new_object@hyperparameter_data$score_table), expected_rows_lower)
  }
)



# Hyperparameter optimisation using stochastic_reject for pruning.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "mim",
  learner = "elastic_net",
  set_signature = FALSE,
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 1L,
  n_max_intensify_steps = 4L,
  n_initial_bootstraps = 2L,
  n_intensify_step_bootstraps = 5L,
  n_random_sets = 16L,
  n_challengers = 10L,
  exploration_method = "stochastic_reject",
  verbose = verbose
)

# Set expected range of rows. The lowest boundary occurs when all challengers
# are rejected after one round, and only one new run is sampled. The upper
# boundary occurs when no challengers are rejected at all and 5 new runs are
# sampled.
expected_rows_lower <- (16 * 2 + 10 + 1) * 2 # initial + step 1 + incumbent
expected_rows_upper <- (16 * 2 + 10 * 5 * 4 + 1 * 5 * 4) * 2 # initial + steps 1-4 + incumbent

testthat::test_that(paste0(
  "Test that \"stochastic_reject\" exploration method may prune any ",
  "hyperparameter sets during intensification"), {
    testthat::expect_lte(nrow(new_object@hyperparameter_data$score_table), expected_rows_upper)
    testthat::expect_gte(nrow(new_object@hyperparameter_data$score_table), expected_rows_lower)
  }
)



# Single-shot hyperparameter optimisation. Note that n_intensify_step_bootstraps
# and n_max_intensify_steps should be set to 1L internally.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "mim",
  learner = "elastic_net",
  set_signature = FALSE,
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 1L,
  n_max_intensify_steps = 4L,
  n_intensify_step_bootstraps = 5L,
  n_random_sets = 16L,
  n_challengers = 10L,
  exploration_method = "single_shot",
  verbose = verbose
)

# Set expected range of rows. Upper and lower boundary are the same, as all runs
# are executed simultaneously.
expected_rows_lower <- expected_rows_upper <- (16 + 10 * 1 + 1 * 1) * 2

testthat::test_that(paste0(
  "Test that \"single_shot\" exploration method does not prune any ",
  "hyperparameter sets during intensification"), {
    testthat::expect_lte(nrow(new_object@hyperparameter_data$score_table), expected_rows_upper)
    testthat::expect_gte(nrow(new_object@hyperparameter_data$score_table), expected_rows_lower)
  }
)



# Test time truncation ---------------------------------------------------------

# Hyperparameter optimisation without pruning and marginal time limit. This
# should just complete the initial step.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "mim",
  learner = "elastic_net",
  set_signature = FALSE,
  time_limit = 0.000001,
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 1L,
  n_max_intensify_steps = 4L,
  n_intensify_step_bootstraps = 1L,
  n_random_sets = 16L,
  n_challengers = 10L,
  exploration_method = "none",
  verbose = verbose
)

testthat::test_that("Time limits are respected and only the initial bootstraps are run.", {
  testthat::expect_gte(new_object@hyperparameter_data$time_taken, 0.000001)
  testthat::expect_true(all(new_object@hyperparameter_data$score_table$iteration_id == 0))
})



# Test that clustered data are correctly handled -------------------------------

# Create data,
data <- familiar:::test_create_synthetic_correlated_data(
  outcome_type = "continuous",
  n_numeric = 4,
  cluster_size = c(3, 3, 3, 3)
)

# Optimise hyperparameters.
new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "mim",
  learner = "elastic_net",
  cluster_method = "hclust",
  cluster_similarity_metric = "mcfadden_r2",
  cluster_similarity_threshold = 0.90,
  n_max_bootstraps = 25L,
  n_max_optimisation_steps = 1L,
  n_max_intensify_steps = 5L,
  n_intensify_step_bootstraps = 1L,
  n_random_sets = 16L,
  n_challengers = 10L,
  exploration_method = "successive_halving",
  verbose = verbose
)

testthat::test_that("One to four features are assessed for clustered features.", {
  testthat::expect_true(all(new_object@hyperparameter_data$parameter_table$sign_size >= 1))
  testthat::expect_true(all(new_object@hyperparameter_data$parameter_table$sign_size <= 4))
  testthat::expect_true(any(new_object@hyperparameter_data$parameter_table$sign_size == 1))
  testthat::expect_true(any(new_object@hyperparameter_data$parameter_table$sign_size == 4))
})


# Test that data with only one relevant feature are correctly handled ----------

# Create data.
data <- familiar:::test_create_single_relevant_feature_data(outcome_type = "continuous")

new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "familiarMultivariateInfoVimpTestSingle",
  learner = "glm_gaussian",
  use_vimp = "return_hpo_vimp",
  verbose = verbose
)

testthat::test_that("Variable importance information is returned.", {
  testthat::expect_false(familiar:::is_empty(new_object@vimp_table))
})

new_object <- ..generate_hyperparameters(
  data = data,
  vimp_method = "familiarMultivariateInfoVimpTestSingle",
  learner = "glm_gaussian",
  use_vimp = "use_hpo_vimp",
  verbose = verbose
)

testthat::test_that("Variable importance information is not returned.", {
  testthat::expect_true(familiar:::is_empty(new_object@vimp_table))
})

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test
familiar:::test_plots(
  plot_function = familiar:::plot_feature_similarity,
  not_available_single_feature = TRUE,
  not_available_single_sample = TRUE,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  data_element = "feature_similarity",
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_feature_similarity,
  data_element = "feature_similarity",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_feature_similarity,
  data_element = "feature_similarity",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list("facet_by" = c("learner", "vimp_method", "data_set")),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_feature_similarity,
  data_element = "feature_similarity",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list("show_dendrogram" = c("left", "bottom")),
  debug = debug_flag
)


# Test normal tests.
familiar:::test_plots(
  plot_function = familiar:::plot_feature_similarity,
  feature_similarity_metric = "spearman",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  data_element = "feature_similarity",
  debug = debug_flag,
  test_config = "normal"
)


familiar:::test_plots(
  plot_function = familiar:::plot_feature_similarity,
  feature_similarity_metric = "mutual_information",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  data_element = "feature_similarity",
  debug = debug_flag,
  test_config = "normal"
)


# Test plotting from dataObject.
data <- familiar:::test_create_good_data(outcome_type = "continuous")
p <- familiar::plot_feature_similarity(object = data, feature_similarity_metric = "spearman")
testthat::test_that("Plotting feature similarity using dataObject works.", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})

# Test plotting from dataObject with two groups.
data <- familiar:::test_create_good_data(outcome_type = "continuous", two_groups = TRUE)
p <- familiar::plot_feature_similarity(object = data, feature_similarity_metric = "spearman")
testthat::test_that("Plotting feature similarity for two groups works.", {
  testthat::expect_length(p, 2L)
  testthat::expect_true(is(p[[1L]], "gtable"))
  testthat::expect_true(is(p[[2L]], "gtable"))
})


# Test plotting from data.table.
data <- familiar:::test_create_good_data(outcome_type = "continuous", to_data_object = FALSE)
p <- familiar::plot_feature_similarity(
  object = data,
  feature_similarity_metric = "spearman",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  outcome_type = "continuous",
  outcome_column = "outcome"
)
testthat::test_that("Plotting feature similarity using data.table works.", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})


# Test plotting from data.table without outcome data.
data <- familiar:::test_create_good_data(outcome_type = "continuous", to_data_object = FALSE)
data[, ":="("batch_id" = NULL, "sample_id" = NULL, "series_id" = NULL, "outcome" = NULL)]
p <- familiar::plot_feature_similarity(
  object = data,
  feature_similarity_metric = "spearman"
)
testthat::test_that("Plotting feature similarity using data.table works.", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test
familiar:::test_plots(
  plot_function = familiar:::plot_sample_clustering,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  data_element = "feature_expressions",
  plot_args = list(
    "verbose" = FALSE,
    "show_normalised_data" = "set_normalisation"),
  debug = debug_flag
)

# No extra elements
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "x_axis_by" = "sample",
    "y_axis_by" = "feature",
    "show_outcome" = FALSE,
    "show_feature_dendrogram" = FALSE,
    "show_sample_dendrogram" = FALSE,
    "verbose" = FALSE),
  debug = debug_flag
)

# No normalisation.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "show_normalised_data" = "none",
    "verbose" = FALSE),
  debug = debug_flag
)

# Normalisation per dataset.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "show_normalised_data" = "set_normalisation",
    "verbose" = FALSE),
  debug = debug_flag
)


# With sample limit
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "sample_limit" = 20L,
    "verbose" = FALSE),
  debug = debug_flag
)


# Test multiple evaluation times
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("survival"),
  plot_args = list(
    "evaluation_times" = c(500, 1000, 1500, 2000),
    "verbose" = FALSE),
  debug = debug_flag
)



# Test plotting from dataObject.
data <- familiar:::test_create_good_data(outcome_type = "survival")
p <- familiar::plot_sample_clustering(object = data, feature_similarity_metric = "spearman")
testthat::test_that("Plotting sample similarity using dataObject works (survival).", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})

data <- familiar:::test_create_good_data(outcome_type = "continuous")
p <- familiar::plot_sample_clustering(object = data, feature_similarity_metric = "spearman")
testthat::test_that("Plotting sample similarity using dataObject works (continuous).", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})


# Test plotting from dataObject with two groups.
data <- familiar:::test_create_good_data(outcome_type = "continuous", two_groups = TRUE)
p <- familiar::plot_sample_clustering(object = data, feature_similarity_metric = "spearman")
testthat::test_that("Plotting sample similarity for two groups works.", {
  testthat::expect_length(p, 2L)
  testthat::expect_true(is(p[[1L]], "gtable"))
  testthat::expect_true(is(p[[2L]], "gtable"))
})


# Test plotting from data.table.
data <- familiar:::test_create_good_data(outcome_type = "survival", to_data_object = FALSE)
p <- familiar::plot_sample_clustering(
  object = data,
  feature_similarity_metric = "spearman",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  outcome_type = "survival",
  outcome_column = c("outcome_time", "outcome_event")
)
testthat::test_that("Plotting sample similarity using data.table works (survival).", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})

data <- familiar:::test_create_good_data(outcome_type = "continuous", to_data_object = FALSE)
p <- familiar::plot_sample_clustering(
  object = data,
  feature_similarity_metric = "spearman",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  outcome_type = "continuous",
  outcome_column = "outcome"
)
testthat::test_that("Plotting sample similarity using data.table works (continuous).", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})


# Test plotting from data.table without outcome data.
data <- familiar:::test_create_good_data(outcome_type = "continuous", to_data_object = FALSE)
data[, ":="("batch_id" = NULL, "sample_id" = NULL, "series_id" = NULL, "outcome" = NULL)]
p <- familiar::plot_sample_clustering(
  object = data,
  feature_similarity_metric = "spearman"
)
testthat::test_that("Plotting sample similarity using data.table works.", {
  testthat::expect_true(is(p[[1]], "gtable"))
})

testthat::skip_on_cran()
testthat::skip_on_ci()

# Train a model to transfer settings.
data <- familiar:::test_create_good_data("survival")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "cox",
  create_novelty_detector = TRUE
)

# Feature levels are correctly ordered -----------------------------------------
data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

# Change order of features.
data$feature_3a <- factor(
  x = data$feature_3a,
  levels = rev(levels(data$feature_3a))
)

testthat::test_that("Feature levels are correctly ordered", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_equal(
      levels(parsed_data@data$feature_3a),
      c("round", "square")
    )
  }
})


data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

# Keep features as characters,
data$feature_3a <- as.character(data$feature_3a)

testthat::test_that("Feature levels are correctly set", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_equal(
      levels(parsed_data@data$feature_3a),
      c("round", "square")
    )
  }
})



# Ordered features are correctly set -------------------------------------------
data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

# Change order of features.
data$feature_4 <- factor(
  x = data$feature_4,
  levels = c("best", "better", "good"),
  ordered = TRUE
)

testthat::test_that("Feature levels are correctly ordered", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_equal(
      levels(parsed_data@data$feature_4),
      c("good", "better", "best")
    )
    testthat::expect_true(is.ordered(parsed_data@data$feature_4))
  }
})

data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

# Keep features as characters,
data$feature_4 <- as.character(data$feature_4)

testthat::test_that("Feature levels are correctly set", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness)

    testthat::expect_equal(
      levels(parsed_data@data$feature_4),
      c("good", "better", "best")
    )
    testthat::expect_true(is.ordered(parsed_data@data$feature_4))
  }
})



# Missing levels in categorical features ---------------------------------------

data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

# Remove a feature level.
data[feature_4 == "better", "feature_4" := "best"]

testthat::test_that("Feature levels are correctly ordered", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_equal(
      levels(parsed_data@data$feature_4),
      c("good", "better", "best")
    )
  }
})


# Additional levels in categorical features -----------------------------------
data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

# Add new level to categorical features.
data[c(1, 2, 3), "feature_3a" := "extra_square"]

testthat::test_that("Extra levels are detected", {
  for (strictness in c("strict", "external_warn", "external")) {
    testthat::expect_error(familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    ))
  }
})


# Test methods to set reference levels -----------------------------------------
data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)
data$feature_3a <- stats::relevel(data$feature_3a, ref = "square")

testthat::test_that("Auto-method does not re-order existing levels.", {
  parsed_data <- familiar::as_data_object(
    data = data.table::copy(data),
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    outcome_column = c("outcome_time", "outcome_event"),
    outcome_type = "survival",
    include_features = c("feature_3a", "feature_3b"),
    reference_method = "auto"
  )

  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3a), n = 1L), "square"
  )
  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3b), n = 1L), "sphere"
  )
})

testthat::test_that("Always-method re-orders existing levels.", {
  parsed_data <- familiar::as_data_object(
    data = data.table::copy(data),
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    outcome_column = c("outcome_time", "outcome_event"),
    outcome_type = "survival",
    include_features = c("feature_3a", "feature_3b"),
    reference_method = "always"
  )

  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3a), n = 1L), "round"
  )
  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3b), n = 1L), "sphere"
  )
})

testthat::test_that("Never-method does not re-order existing levels.", {
  parsed_data <- familiar::as_data_object(
    data = data.table::copy(data),
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    outcome_column = c("outcome_time", "outcome_event"),
    outcome_type = "survival",
    include_features = c("feature_3a", "feature_3b"),
    reference_method = "never"
  )

  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3a), n = 1L), "square"
  )
  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3b), n = 1L), "sphere"
  )
})

# Now for automatic conversion of categorical variables.
data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)
data$feature_3a <- as.character(data$feature_3a)
data[, "feature_3_extra" := feature_3a == "square"]

testthat::test_that("Auto-method re-orders existing levels.", {
  parsed_data <- familiar::as_data_object(
    data = data.table::copy(data),
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    outcome_column = c("outcome_time", "outcome_event"),
    outcome_type = "survival",
    include_features = c("feature_3a", "feature_3_extra"),
    reference_method = "auto"
  )
  
  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3a), n = 1L), "round"
  )
  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3_extra), n = 1L), "FALSE"
  )
})

testthat::test_that(
  "Always-method re-orders existing levels.", {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      batch_id_column = "batch_id",
      sample_id_column = "sample_id",
      series_id_column = "series_id",
      outcome_column = c("outcome_time", "outcome_event"),
      outcome_type = "survival",
      include_features = c("feature_3a", "feature_3_extra"),
      reference_method = "always"
    )
    
    testthat::expect_equal(
      head(levels(parsed_data@data$feature_3a), n = 1L), "round"
    )
    testthat::expect_equal(
      head(levels(parsed_data@data$feature_3_extra), n = 1L), "FALSE"
    )
})

testthat::test_that("Never-method sorts existing levels.", {
  parsed_data <- familiar::as_data_object(
    data = data.table::copy(data),
    batch_id_column = "batch_id",
    sample_id_column = "sample_id",
    series_id_column = "series_id",
    outcome_column = c("outcome_time", "outcome_event"),
    outcome_type = "survival",
    include_features = c("feature_3a", "feature_3_extra"),
    reference_method = "never"
  )
  
  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3a), n = 1L), "round"
  )
  testthat::expect_equal(
    head(levels(parsed_data@data$feature_3_extra), n = 1L), "FALSE"
  )
})



# Censoring and event identifiers are set --------------------------------------
data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

testthat::test_that("Censoring is correctly transferred", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_true(all(parsed_data@data$outcome_event %in% c(0, 1)))
  }
})



# Manual censoring and event identifiers ---------------------------------------
data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)
manual_data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

# Update status column to specific
manual_data$outcome_event <- as.character(data$outcome_event)
manual_data[outcome_event == "0", "outcome_event" := "alive"]
manual_data[outcome_event == "1", "outcome_event" := "dead"]

testthat::test_that("Censoring and event identifiers can be manually set", {
  for (strictness in c("strict", "external_warn", "external")) {
    # With specific manual event identifiers.
    manual_parsed_data <- familiar::as_data_object(
      data = data.table::copy(manual_data),
      object = fam_model,
      event_indicator = "dead",
      censoring_indicator = "alive",
      check_stringency = strictness
    )

    # With original event identifiers.
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_true(
      all(manual_parsed_data@data$outcome_event == parsed_data@data$outcome_event)
    )
  }
})



# Unknown censoring and event identifiers --------------------------------------
data <- familiar:::test_create_good_data("survival", to_data_object = FALSE)

# Update status column to specific
data$outcome_event <- as.character(data$outcome_event)
data[outcome_event == "0", "outcome_event" := "alive"]
data[outcome_event == "1", "outcome_event" := "dead"]

testthat::test_that("Censoring and event identifiers are not known", {
  for (strictness in c("strict", "external_warn", "external")) {
    # With specific manual event identifiers. However, now they are not manually
    # provided.

    if (strictness == "strict") {
      testthat::expect_error(familiar::as_data_object(
        data = data.table::copy(data),
        object = fam_model,
        check_stringency = strictness
      ))
      
    } else if (strictness == "external_warn") {
      warns <- testthat::capture_warnings(familiar::as_data_object(
        data = data.table::copy(data),
        object = fam_model,
        check_stringency = strictness
      ))

      testthat::expect_true(any(grepl("event indicator", warns)))
      testthat::expect_true(any(grepl("censoring indicator", warns)))
      
    } else if (strictness == "external") {
      # With original event identifiers.
      parsed_data <- familiar::as_data_object(
        data = data.table::copy(data),
        object = fam_model,
        check_stringency = strictness
      )
    }
  }
})



# Non-standard censoring and event identifiers are correctly propagated --------
# Train a model to transfer settings.
data <- familiar:::test_create_good_data("survival")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "cox",
  event_indicator = "dead",
  censoring_indicator = "alive",
  create_novelty_detector = TRUE
)

testthat::test_that("Non-standard censoring and event identifiers are automatically transferred", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_true(all(parsed_data@data$outcome_event %in% c(0, 1)))
  }
})


# Class levels are correctly ordered -------------------------------------------
# Train a model to transfer settings.
data <- familiar:::test_create_good_data("binomial")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "glm_logistic",
  create_novelty_detector = TRUE
)

# Create test dataset.
data <- familiar:::test_create_good_data("binomial", to_data_object = FALSE)

testthat::test_that("Class levels are correctly ordered", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_equal(
      levels(parsed_data@data$outcome),
      c("red", "green")
    )
  }
})



# Class levels are restored to expected order ----------------------------------
data <- familiar:::test_create_good_data("binomial", to_data_object = FALSE)

# Reorder class levels
data$outcome <- factor(
  x = data$outcome,
  levels = c("green", "red")
)

testthat::test_that("Class levels are ordered according to expectations", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )

    testthat::expect_equal(
      levels(parsed_data@data$outcome),
      c("red", "green")
    )
  }
})


# Missing class levels ---------------------------------------------------------
data <- familiar:::test_create_good_data("binomial", to_data_object = FALSE)
data[outcome == "red", outcome := "green"]

testthat::test_that("Class levels are ordered according to expectations", {
  for (strictness in c("strict", "external_warn", "external")) {
    if (strictness == "strict") {
      testthat::expect_error(familiar::as_data_object(
        data = data.table::copy(data),
        object = fam_model,
        check_stringency = strictness
      ))
      
    } else {
      # Class levels may be missing for external_warn and external strictness
      # levels.
      parsed_data <- familiar::as_data_object(
        data = data.table::copy(data),
        object = fam_model,
        check_stringency = strictness
      )

      testthat::expect_equal(
        levels(parsed_data@data$outcome),
        c("red", "green"))
    }
  }
})



# Additional class levels --------------------------------------------------
data <- familiar:::test_create_good_data("binomial", to_data_object = FALSE)

# Add new level
data[c(1, 2, 3), "outcome" := "blue"]

testthat::test_that("Additional class levels are detected", {
  for (strictness in c("strict", "external_warn", "external")) {
    if (strictness %in% c("strict", "external_warn")) {
      testthat::expect_error(familiar::as_data_object(
        data = data.table::copy(data),
        object = fam_model,
        check_stringency = strictness
      ))
      
    } else {
      parsed_data <- familiar::as_data_object(
        data = data.table::copy(data),
        object = fam_model,
        check_stringency = strictness
      )
    }
  }
})



# Naive model ------------------------------------------------------------------
# Train a model to transfer settings.
data <- familiar:::test_create_good_data("binomial")

# Train a naive model using the good dataset.
fam_naive_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "glm_logistic",
  create_novelty_detector = TRUE,
  create_naive = TRUE
)
  
data <- familiar:::test_create_good_data("binomial", to_data_object = FALSE)

testthat::test_that("Naive models can be used to convert data objects.", {
  for (strictness in c("strict", "external_warn", "external")) {
    parsed_data <- familiar::as_data_object(
      data = data.table::copy(data),
      object = fam_model,
      check_stringency = strictness
    )
    
    testthat::expect_equal(
      levels(parsed_data@data$outcome),
      c("red", "green")
    )
  }
})

# Skip on CRAN as this test takes some time.
testthat::skip_on_cran()
testthat::skip_on_ci()

outcome_type <- "survival"
n_numeric_features <- 1
imputation_method <- "lasso"

# Generic test -----------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features
  )

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}



# Rare categories test ---------------------------------------------------------
for (n_numeric_features in c(3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features
  )

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # At this point replace rare levels in the data.
        data_copy <- familiar:::test_data_drop_rare_feature_levels(data = data_copy)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}




# NA-instance test -------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with some NA instances"),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}



# NA random-value test ---------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_random_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features
  )

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with some NA datapoints."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}



# One feature NA test ----------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_feature_all_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with one feature entirely NA."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(
              feature_info_list[[feature]]@imputation_parameters),
            TRUE)
          
          if (feature == "feature_2") {
            # Feature 2 is completely missing data.
            testthat::expect_equal(
              any(familiar:::is_valid_data(data_copy@data[[feature]])),
              FALSE)

            # Check that a None-class imputer is trained.
            testthat::expect_s4_class(
              feature_info_list[[feature]]@imputation_parameters,
              "featureInfoParametersImputationNone")
            
          } else {
            # Check that all feature data are valid.
            testthat::expect_equal(
              all(familiar:::is_valid_data(data_copy@data[[feature]])),
              TRUE)
          }
        }
      }
    )
  }
}



# Invariant feature test -------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_invariant_feature_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features
  )

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with invariant features."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}



# One feature invariant test ---------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_feature_invariant_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with one invariant feature."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}


# One-sample test --------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_sample_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with invariant features."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}


# One-feature test -------------------------------------------------------------
for (n_numeric_features in c(1, 0)) {
  data <- familiar:::test_create_synthetic_series_random_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  # Select only the first feature.
  data@data <- data@data[, mget(
    c(familiar:::get_non_feature_columns(data), "feature_1"))]

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with a single ",
        "feature and some NA datapoints."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list,
          data = data_copy)

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}


# Collection test --------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_random_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (imputation_method in familiar:::.get_available_imputation_methods()) {
    testthat::test_that(
      paste0(
        "Imputation is correctly performed using the ",
        imputation_method, " method and ",
        n_numeric_features, " numeric features for a dataset with some NA ",
        "datapoints and a collection of imputers."),
      {
        # Make a copy of the data.
        data_copy <- data.table::copy(data)

        # Create a list of featureInfo objects.
        feature_info_list <- test_create_generic_info(data = data_copy)

        # Create imputation skeleton.
        feature_info_list_1 <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list_1 <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list_1,
          data = data_copy)

        # Create imputation skeleton.
        feature_info_list_2 <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list_2 <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list_2,
          data = data_copy)

        # Create imputation skeleton.
        feature_info_list_3 <- familiar:::create_imputation_parameter_skeleton(
          feature_info_list = feature_info_list,
          imputation_method = imputation_method)

        # Add imputer parameters.
        feature_info_list_3 <- familiar:::add_imputation_info(
          feature_info_list = feature_info_list_3,
          data = data_copy)

        # Collect features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          feature_info_list[[feature]]@imputation_parameters <- familiar:::..collect_and_aggregate_imputation_info(
            feature_info_list = list(
              feature_info_list_1[[feature]],
              feature_info_list_2[[feature]],
              feature_info_list_3[[feature]]),
            feature_type = feature_info_list[[feature]]@feature_type,
            feature_name = feature_info_list[[feature]]@name
          )$parameters
        }

        # Assume that the data is pre-processed.
        data_copy@preprocessing_level <- "batch_normalisation"

        # Attempt to impute the data.
        data_copy <- familiar:::impute_features(
          data = data_copy,
          feature_info_list = feature_info_list)

        # Iterate over features.
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Check that imputer parameters were set.
          testthat::expect_equal(
            familiar:::feature_info_complete(feature_info_list[[feature]]@imputation_parameters),
            TRUE)

          # Check that all feature data are valid.
          testthat::expect_equal(
            all(familiar:::is_valid_data(data_copy@data[[feature]])),
            TRUE)
        }
      }
    )
  }
}

testthat::skip_on_cran()

# Find available stratification methods.
stratification_methods <- familiar:::.get_available_stratification_methods()

if (!rlang::is_installed("power.transform")) testthat::skip()

# Test for good dataset.
for (stratification_method in stratification_methods) {
  # Create a dataset using the good dataset.
  data <- familiar:::test_create_good_data("survival")
  
  # Train a simple linear GLM using the good dataset.
  fam_model <- familiar:::test_train(
    data = data,
    cluster_method = "none",
    imputation_method = "simple",
    hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
    learner = "cox",
    stratification_method = stratification_method,
    create_novelty_detector = FALSE
  )
  
  # Risk stratification.
  predictions_risk <- familiar::predict(
    object = fam_model,
    newdata = data,
    type = "risk_stratification"
  )
  
  testthat::test_that("Risk groups are formed.", {
    testthat::expect_equal(fam_model@km_info$stratification_method, stratification_method)
    testthat::expect_true(!any(is.na(predictions_risk$group)))
  })
}

# Test for single-feature dataset.
for (stratification_method in stratification_methods) {
  # Create a dataset using the single-feature dataset.
  data <- familiar:::test_create_single_feature_data("survival")
  
  # Train a simple linear GLM using the good dataset.
  fam_model <- familiar:::test_train(
    data = data,
    cluster_method = "none",
    imputation_method = "simple",
    hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
    learner = "cox",
    stratification_method = stratification_method,
    create_novelty_detector = FALSE
  )
  
  # Risk stratification.
  predictions_risk <- familiar::predict(
    object = fam_model,
    newdata = data,
    type = "risk_stratification"
  )
  
  testthat::test_that("Risk groups are formed.", {
    testthat::expect_equal(fam_model@km_info$stratification_method, stratification_method)
    testthat::expect_true(!any(is.na(predictions_risk$group)))
  })
}

# Test for dataset with only one value for a feature.
for (stratification_method in stratification_methods) {
  # Create a dataset with a feature that only has two values.
  data <- familiar:::test_create_single_feature_two_values_data("survival")
  
  # Train a simple linear GLM using the feature set that only has two values.
  fam_model <- familiar:::test_train(
    data = data,
    cluster_method = "none",
    imputation_method = "simple",
    hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
    learner = "cox",
    stratification_method = stratification_method,
    create_novelty_detector = FALSE
  )
  
  # Risk stratification.
  predictions_risk <- familiar::predict(
    object = fam_model,
    newdata = data,
    type = "risk_stratification"
  )
  
  testthat::test_that("Risk groups are formed.", {
    testthat::expect_equal(fam_model@km_info$stratification_method, stratification_method)
    testthat::expect_true(!any(is.na(predictions_risk$group)))
  })
}

# Test that multiple risk stratifications methods can be applied simultaneously.

# Create a dataset using the good dataset.
data <- familiar:::test_create_good_data("survival")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "cox",
  stratification_method = stratification_methods,
  create_novelty_detector = FALSE
)

for (stratification_method in stratification_methods) {
  # Risk stratification.
  predictions_risk <- familiar::predict(
    object = fam_model,
    newdata = data,
    type = "risk_stratification",
    stratification_method = stratification_method
  )
  
  testthat::test_that("Risk groups are formed.", {
    testthat::expect_true(!any(is.na(predictions_risk$group)))
  })
}

# Test that creation fails for models that only yield invalid predictions
for (stratification_method in stratification_methods) {
  # Create a dataset using the good dataset.
  data <- familiar:::test_create_good_data("survival")
  
  # Train a simple linear GLM using the good dataset.
  fam_model <- familiar:::test_train(
    data = data,
    cluster_method = "none",
    imputation_method = "simple",
    hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
    learner = "lasso_test_all_fail",
    stratification_method = stratification_method,
    create_novelty_detector = FALSE
  )
  
  # Risk stratification.
  predictions_risk <- familiar::predict(
    object = fam_model,
    newdata = data,
    type = "risk_stratification"
  )
  
  testthat::test_that("Risk groups are formed.", {
    testthat::expect_equal(fam_model@km_info$stratification_method, NULL)
    testthat::expect_true(all(is.na(predictions_risk$group)))
  })
}

# Test that creation succeeds for models that yield some invalid predictions.
for (stratification_method in stratification_methods) {
  # Create a dataset using the good dataset.
  data <- familiar:::test_create_good_data("survival")
  
  # Train a simple linear GLM using the good dataset.
  fam_model <- familiar:::test_train(
    data = data,
    cluster_method = "none",
    imputation_method = "simple",
    hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
    learner = "lasso_test_some_fail",
    stratification_method = stratification_method,
    create_novelty_detector = FALSE
  )
  
  # Risk stratification.
  predictions_risk <- familiar::predict(
    object = fam_model,
    newdata = data,
    type = "risk_stratification"
  )
  
  testthat::test_that("Risk groups are formed.", {
    testthat::expect_equal(fam_model@km_info$stratification_method, stratification_method)
    testthat::expect_false(all(is.na(predictions_risk$group)))
    testthat::expect_true(any(is.na(predictions_risk$group)))
  })
}

testthat::skip_on_cran()

normalisation_test <- function(
    data_list,
    norm_method,
    expected_shift = NULL,
    expected_scale = NULL,
    expected_min,
    expected_max,
    expected_none) {
  # Iterate over the data sets.
  for (ii in seq_along(data_list)) {
    # Get values
    x <- data_list[[ii]]

    # Create skeleton
    object <- familiar:::..create_normalisation_parameter_skeleton(
      feature_name = "test",
      method = norm_method)

    # Acquire normalisation parameters.
    object <- familiar:::add_feature_info_parameters(object, data = x)

    if (expected_none[ii]) {
      testthat::expect_s4_class(object, "featureInfoParametersNormalisationNone")
      
    } else {
      # Test the shift and scale values.
      if (!is.null(expected_scale)) {
        # Scale may be absent for mean centering and similar methods.
        testthat::expect_equal(object@scale, expected_scale[ii])
      }

      # Test shift parameter
      testthat::expect_equal(object@shift, expected_shift[ii])
    }

    # Apply normalisation
    y <- familiar:::apply_feature_info_parameters(
      object = object,
      data = x)

    # Test that length is the same as input
    testthat::expect_equal(length(y), length(x))

    # Test that non-finite values are present and correctly positioned.
    testthat::expect_equal(is.finite(y), is.finite(x))

    # Remove non-finite values
    y <- y[is.finite(y)]

    # Check minimum value of normalised range, and its position (first).
    if (length(expected_min[[ii]]) > 0) {
      testthat::expect_equal(min(y), expected_min[[ii]])
      testthat::expect_equal(min(y), head(y, n = 1))
    } else {
      testthat::expect_equal(length(y), 0)
    }

    # Check maximum value of normalised range, and its position (last).
    if (length(expected_max[[ii]]) > 0) {
      testthat::expect_equal(max(y), expected_max[[ii]])
      testthat::expect_equal(max(y), tail(y, n = 1))
    } else {
      testthat::expect_equal(length(y), 0)
    }
  }
}


# Create data
good_data <- c(0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
bad_data <- c(-Inf, Inf, Inf, NA, NA, NA, -Inf, Inf, Inf, NA)
no_data <- numeric(0L)
realistic_data <- c(0, 1, 2, NA, 3, 4, 5, 6, 7, Inf, 8, 9)

# Combine to list
data_list <- list(
  "good" = good_data,
  "bad" = bad_data,
  "empty" = no_data,
  "realistic" = realistic_data)

# No normalisation -------------------------------------------------------------
testthat::test_that("None normalisation is correctly performed", {
  expected_min <- list(0.0, numeric(0), numeric(0), 0.0)
  expected_max <- list(9.0, numeric(0), numeric(0), 9.0)
  expected_none <- c(TRUE, TRUE, TRUE, TRUE)

  normalisation_test(
    data_list = data_list,
    norm_method = "none",
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})

# Normalisation ----------------------------------------------------------------
testthat::test_that("Normalisation is correctly performed", {
  expected_shift <- c(0.0, NA_real_, NA_real_, 0.0)
  expected_scale <- c(9.0, NA_real_, NA_real_, 9.0)
  expected_min <- list(0.0, numeric(0), numeric(0), 0.0)
  expected_max <- list(1.0, numeric(0), numeric(0), 1.0)
  expected_none <- c(FALSE, TRUE, TRUE, FALSE)

  normalisation_test(
    data_list = data_list,
    norm_method = "normalisation",
    expected_shift = expected_shift,
    expected_scale = expected_scale,
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})

# Normalisation (trimmed) ------------------------------------------------------
testthat::test_that("Normalisation (trimmed) is correctly performed", {
  expected_shift <- c(1.0, NA_real_, NA_real_, 1.0)
  expected_scale <- c(7.0, NA_real_, NA_real_, 7.0)
  expected_min <- list(-1 / 7, numeric(0), numeric(0), -1 / 7)
  expected_max <- list(8 / 7, numeric(0), numeric(0), 8 / 7)
  expected_none <- c(FALSE, TRUE, TRUE, FALSE)

  normalisation_test(
    data_list = data_list,
    norm_method = "normalisation_trim",
    expected_shift = expected_shift,
    expected_scale = expected_scale,
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})

# Normalisation (winsorised) ---------------------------------------------------
testthat::test_that("Normalisation (winsorised) is correctly performed", {
  expected_shift <- c(0.45, NA_real_, NA_real_, 0.45)
  expected_scale <- c(8.1, NA_real_, NA_real_, 8.1)
  expected_min <- list(-0.45 / 8.1, numeric(0), numeric(0), -0.45 / 8.1)
  expected_max <- list(8.55 / 8.1, numeric(0), numeric(0), 8.55 / 8.1)
  expected_none <- c(FALSE, TRUE, TRUE, FALSE)

  normalisation_test(
    data_list = data_list,
    norm_method = "normalisation_winsor",
    expected_shift = expected_shift,
    expected_scale = expected_scale,
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})


# Standardisation --------------------------------------------------------------
testthat::test_that("Standardisation is correctly performed", {
  # Set x.
  x <- c(0, 1, 2, 3, 4, 5, 6, 7, 8, 9)

  # Mean
  mu <- mean(x)

  # Standard deviation (without Bessel correction)
  sigma <- sqrt(sum((x - mu)^2) / length(x))

  expected_shift <- c(mu, NA_real_, NA_real_, mu)
  expected_scale <- c(sigma, NA_real_, NA_real_, sigma)
  expected_min <- list(-mu / sigma, numeric(0), numeric(0), -mu / sigma)
  expected_max <- list((9 - mu) / sigma, numeric(0), numeric(0), (9 - mu) / sigma)
  expected_none <- c(FALSE, TRUE, TRUE, FALSE)

  normalisation_test(
    data_list = data_list,
    norm_method = "standardisation",
    expected_shift = expected_shift,
    expected_scale = expected_scale,
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})


# Standardisation (trimmed) ----------------------------------------------------
testthat::test_that("Standardisation (trimmed) is correctly performed", {
  # Set x.
  x <- c(1, 2, 3, 4, 5, 6, 7, 8)

  # Mean
  mu <- mean(x)

  # Standard deviation (without Bessel correction)
  sigma <- sqrt(sum((x - mu)^2) / length(x))

  expected_shift <- c(mu, NA_real_, NA_real_, mu)
  expected_scale <- c(sigma, NA_real_, NA_real_, sigma)
  expected_min <- list(-mu / sigma, numeric(0), numeric(0), -mu / sigma)
  expected_max <- list((9 - mu) / sigma, numeric(0), numeric(0), (9 - mu) / sigma)
  expected_none <- c(FALSE, TRUE, TRUE, FALSE)

  normalisation_test(
    data_list = data_list,
    norm_method = "standardisation_trim",
    expected_shift = expected_shift,
    expected_scale = expected_scale,
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})


# Standardisation (winsorised) -------------------------------------------------
testthat::test_that("Standardisation (winsorised) is correctly performed", {
  # Set x.
  x <- c(0.45, 1, 2, 3, 4, 5, 6, 7, 8, 8.55)

  # Mean
  mu <- mean(x)

  # Standard deviation (without Bessel correction)
  sigma <- sqrt(sum((x - mu)^2) / length(x))

  expected_shift <- c(mu, NA_real_, NA_real_, mu)
  expected_scale <- c(sigma, NA_real_, NA_real_, sigma)
  expected_min <- list(-mu / sigma, numeric(0), numeric(0), -mu / sigma)
  expected_max <- list((9 - mu) / sigma, numeric(0), numeric(0), (9 - mu) / sigma)
  expected_none <- c(FALSE, TRUE, TRUE, FALSE)

  normalisation_test(
    data_list = data_list,
    norm_method = "standardisation_winsor",
    expected_shift = expected_shift,
    expected_scale = expected_scale,
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})


# Mean centering ---------------------------------------------------------------
testthat::test_that("Mean centering is correctly performed", {
  expected_shift <- c(4.5, NA_real_, NA_real_, 4.5)
  expected_min <- list(-4.5, numeric(0), numeric(0), -4.5)
  expected_max <- list(4.5, numeric(0), numeric(0), 4.5)
  expected_none <- c(FALSE, TRUE, TRUE, FALSE)

  normalisation_test(
    data_list = data_list,
    norm_method = "mean_centering",
    expected_shift = expected_shift,
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})


# Quantile ---------------------------------------------------------------------
testthat::test_that("Quantile normalisation is correctly performed", {
  expected_shift <- c(4.5, NA_real_, NA_real_, 4.5) # Median will be between 4.0 and 5.0: 4.5
  expected_scale <- c(4.5, NA_real_, NA_real_, 4.5) # IQR
  expected_min <- list(-1.0, numeric(0), numeric(0), -1.0)
  expected_max <- list(1.0, numeric(0), numeric(0), 1.0)
  expected_none <- c(FALSE, TRUE, TRUE, FALSE)

  normalisation_test(
    data_list = data_list,
    norm_method = "quantile",
    expected_shift = expected_shift,
    expected_scale = expected_scale,
    expected_min = expected_min,
    expected_max = expected_max,
    expected_none = expected_none)
})



# Generic test -----------------------------------------------------------------
outcome_type <- "survival"

for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type, 
    n_numeric = n_numeric_features)

  for (normalisation_method in familiar:::.get_available_normalisation_methods()) {
    testthat::test_that(paste0(
      "Normalisation is correctly performed using the ",
      normalisation_method, " method and ", n_numeric_features,
      " numeric features."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Add skeletons.
      feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = normalisation_method)

      # Update the feature info list.
      feature_info_list <- familiar:::add_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Act as if the data has been transformed.
      data_copy@preprocessing_level <- "transformation"

      # Perform a normalisation.
      data_copy <- familiar:::normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)
        
      } else {
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Determine if the feature is numeric.
          if (feature_info_list[[feature]]@feature_type == "numeric") {
            # Compute mean.
            x <- mean(data_copy@data[[feature]], na.rm = TRUE)

            if (normalisation_method %in% 
                c("normalisation", "normalisation_trim", "normalisation_winsor")) {
              # Check that the feature is correctly centred around 0.
              testthat::expect_equal(x < 0.7 & x > 0.3, TRUE)
              
            } else {
              # Check that the feature is correctly centred around 0.
              testthat::expect_equal(x < 0.2 & x > -0.2, TRUE)
            }
          } else {
            # For categorical features test that the object class is
            # featureInfoParametersNormalisationNone.
            testthat::expect_s4_class(
              feature_info_list[[feature]]@normalisation_parameters,
              "featureInfoParametersNormalisationNone")
          }
        }
      }

      # Assert that inverting the transformation produces the original dataset.
      data_restored <- familiar:::normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list,
        invert = TRUE)

      # Iterate over features and compare. They should be equal.
      for (feature in familiar:::get_feature_columns(data_restored)) {
        # Expect that values are similar to a tolerance.
        testthat::expect_equal(data_restored@data[[feature]], data@data[[feature]])
      }

      # Assert that aggregating the normalisation parameters works as expected.
      aggr_normalisation_parameters <- familiar:::..collect_and_aggregate_normalisation_info(
        feature_info_list = feature_info_list,
        instance_mask = rep_len(TRUE, length(feature_info_list)),
        feature_name = "test"
      )

      if (n_numeric_features > 0 && normalisation_method != "none") {
        # Expect that the selected normalisation method matches the selected
        # method.
        testthat::expect_equal(
          aggr_normalisation_parameters$parameters@method, 
          normalisation_method, 
          ignore_attr = TRUE)

        # Expect that the shift and scale parameters are finite.
        testthat::expect_equal(is.finite(aggr_normalisation_parameters$parameters@shift), TRUE)
        if (is(aggr_normalisation_parameters$parameters,
               "featureInfoParametersNormalisationShiftScale")) {
          testthat::expect_equal(is.finite(aggr_normalisation_parameters$parameters@scale), TRUE)
        }

        # Expect that instance mask is equal to the number of numeric features.
        testthat::expect_equal(sum(aggr_normalisation_parameters$instance_mask) > 0, TRUE)
        testthat::expect_lte(
          sum(aggr_normalisation_parameters$instance_mask),
          n_numeric_features)
        
      } else {
        # Expect that the selected normalisation method matches the selected
        # method.
        testthat::expect_equal(aggr_normalisation_parameters$parameters@method, "none")

        # Expect that instance mask is equal to the number of numeric features.
        testthat::expect_equal(
          sum(aggr_normalisation_parameters$instance_mask),
          familiar:::get_n_features(data_copy))
      }
    })
  }
}


# NA-value test ----------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_na_data(
    outcome_type = outcome_type,
    n_numeric = n_numeric_features)

  for (normalisation_method in familiar:::.get_available_normalisation_methods()) {
    testthat::test_that(paste0(
      "Normalisation is correctly performed using the ",
      normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with some NA data."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Add skeletons.
      feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = normalisation_method)

      # Update the feature info list.
      feature_info_list <- familiar:::add_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Act as if the data has been transformed.
      data_copy@preprocessing_level <- "transformation"

      # Perform a normalisation.
      data_copy <- familiar:::normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)
        
      } else {
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Determine if the feature is numeric.
          if (feature_info_list[[feature]]@feature_type == "numeric") {
            # Compute mean.
            x <- mean(data_copy@data[[feature]], na.rm = TRUE)

            if (normalisation_method %in% 
                c("normalisation", "normalisation_trim", "normalisation_winsor")) {
              # Check that the feature is correctly centred around 0.
              testthat::expect_equal(x < 0.7 & x > 0.3, TRUE)
            } else {
              # Check that the feature is correctly centred around 0.
              testthat::expect_equal(x < 0.2 & x > -0.2, TRUE)
            }
          } else {
            # Assert that for the categorical features, the none method is used.
            testthat::expect_s4_class(
              feature_info_list[[feature]]@normalisation_parameters,
              "featureInfoParametersNormalisationNone")
          }
        }
      }
    })
  }
}


# One feature NA test ----------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_feature_all_na_data(
    outcome_type = outcome_type, 
    n_numeric = n_numeric_features)

  for (normalisation_method in familiar:::.get_available_normalisation_methods()) {
    testthat::test_that(paste0(
      "Normalisation is correctly performed using the ",
      normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with one feature completely NA."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Add skeletons.
      feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = normalisation_method)

      # Update the feature info list.
      feature_info_list <- familiar:::add_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Act as if the data has been transformed.
      data_copy@preprocessing_level <- "transformation"

      # Perform a normalisation.
      data_copy <- familiar:::normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)
        
      } else {
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Determine if the feature is numeric.
          if (feature_info_list[[feature]]@feature_type == "numeric") {
            if (feature == "feature_2") {
              # Assert that for the NA feature, the none method is used.
              testthat::expect_s4_class(
                feature_info_list[[feature]]@normalisation_parameters,
                "featureInfoParametersNormalisationNone")
              
            } else {
              # Compute mean.
              x <- mean(data_copy@data[[feature]], na.rm = TRUE)

              if (normalisation_method %in% 
                  c("normalisation", "normalisation_trim", "normalisation_winsor")) {
                # Check that the feature is correctly centred around 0.
                testthat::expect_equal(x < 0.7 & x > 0.3, TRUE)
              } else {
                # Check that the feature is correctly centred around 0.
                testthat::expect_equal(x < 0.2 & x > -0.2, TRUE)
              }
            }
          } else {
            # Assert that for the categorical features, the none method is used.
            testthat::expect_s4_class(
              feature_info_list[[feature]]@normalisation_parameters,
              "featureInfoParametersNormalisationNone")
          }
        }
      }
    })
  }
}



# Invariant feature test -------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_invariant_feature_data(
    outcome_type = outcome_type, 
    n_numeric = n_numeric_features)

  for (normalisation_method in familiar:::.get_available_normalisation_methods()) {
    testthat::test_that(paste0(
      "Normalisation is correctly performed using the ",
      normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with invariant features."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Add skeletons.
      feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = normalisation_method)

      # Update the feature info list.
      feature_info_list <- familiar:::add_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Act as if the data has been transformed.
      data_copy@preprocessing_level <- "transformation"

      # Perform a normalisation.
      data_copy <- familiar:::normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Check that the data is not altered.
      testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

      for (feature in familiar:::get_feature_columns(data_copy)) {
        # Assert that the none method is used.
        testthat::expect_s4_class(
          feature_info_list[[feature]]@normalisation_parameters,
          "featureInfoParametersNormalisationNone")
      }
    })
  }
}



# One feature invariant test ---------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_feature_invariant_data(
    outcome_type = outcome_type, 
    n_numeric = n_numeric_features)

  for (normalisation_method in familiar:::.get_available_normalisation_methods()) {
    testthat::test_that(paste0(
      "Normalisation is correctly performed using the ",
      normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with one invariant feature."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Add skeletons.
      feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = normalisation_method)

      # Update the feature info list.
      feature_info_list <- familiar:::add_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Act as if the data has been transformed.
      data_copy@preprocessing_level <- "transformation"

      # Perform a normalisation.
      data_copy <- familiar:::normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Test whether the features are normalised (unless none).
      if (normalisation_method == "none") {
        # Check that the data is not altered.
        testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

      } else {
        for (feature in familiar:::get_feature_columns(data_copy)) {
          # Determine if the feature is numeric.
          if (feature_info_list[[feature]]@feature_type == "numeric") {
            if (feature == "feature_2") {
              # Assert that for the NA feature, the none method is used.
              testthat::expect_s4_class(
                feature_info_list[[feature]]@normalisation_parameters,
                  "featureInfoParametersNormalisationNone")

            } else {
              # Compute mean.
              x <- mean(data_copy@data[[feature]], na.rm = TRUE)

              if (normalisation_method %in% 
                  c("normalisation", "normalisation_trim", "normalisation_winsor")) {
                # Check that the feature is correctly centred around 0.
                testthat::expect_equal(x < 0.7 & x > 0.3, TRUE)
              } else {
                # Check that the feature is correctly centred around 0.
                testthat::expect_equal(x < 0.2 & x > -0.2, TRUE)
              }
            }
          } else {
            # Assert that for the categorical features, the none method is used.
            testthat::expect_s4_class(
              feature_info_list[[feature]]@normalisation_parameters,
              "featureInfoParametersNormalisationNone")
          }
        }
      }
    })
  }
}



# One-sample test --------------------------------------------------------------
for (n_numeric_features in c(4, 3, 2, 1, 0)) {
  data <- familiar:::test_create_synthetic_series_one_sample_data(
    outcome_type = outcome_type, 
    n_numeric = n_numeric_features)

  for (normalisation_method in familiar:::.get_available_normalisation_methods()) {
    testthat::test_that(paste0(
      "Normalisation is correctly performed using the ",
      normalisation_method, " method and ",
      n_numeric_features, " numeric features for a dataset with one sample."
    ), {
      # Make a copy of the data.
      data_copy <- data.table::copy(data)

      # Create a list of featureInfo objects.
      feature_info_list <- test_create_generic_info(data = data_copy)

      # Add skeletons.
      feature_info_list <- familiar:::create_normalisation_parameter_skeleton(
        feature_info_list = feature_info_list,
        normalisation_method = normalisation_method)

      # Update the feature info list.
      feature_info_list <- familiar:::add_normalisation_parameters(
        feature_info_list = feature_info_list,
        data = data_copy)

      # Act as if the data has been transformed.
      data_copy@preprocessing_level <- "transformation"

      # Perform a normalisation.
      data_copy <- familiar:::normalise_features(
        data = data_copy,
        feature_info_list = feature_info_list)

      # Check that the data is not altered.
      testthat::expect_equal(data.table::fsetequal(data_copy@data, data@data), TRUE)

      for (feature in familiar:::get_feature_columns(data_copy)) {
        # Assert that the none method is used.
        testthat::expect_s4_class(
          feature_info_list[[feature]]@normalisation_parameters,
          "featureInfoParametersNormalisationNone")
      }
    })
  }
}

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

# Survival data with multiple instances ----------------------------------------

# Create a dataset using the good dataset.
data <- familiar:::test_create_good_data("survival")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "cox",
  create_novelty_detector = TRUE
)

# Default predictions using data object.
predictions_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  .as_prediction_table = TRUE
)

# Default predictions using data.table.
predictions_2 <- familiar::predict(
  object = fam_model,
  newdata = familiar:::test_create_good_data("survival", to_data_object = FALSE),
  .as_prediction_table = TRUE
)

# Survival probability.
predictions_surv <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "survival_probability",
  .as_prediction_table = TRUE
)

# Survival probability at a different time points.
predictions_surv_early <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "survival_probability",
  time = 1.0,
  .as_prediction_table = TRUE
)

# Novelty.
predictions_novelty_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "novelty",
  .as_prediction_table = TRUE
)

# Novelty predictions using data.table.
predictions_novelty_2 <- familiar::predict(
  object = fam_model@novelty_detector,
  newdata = familiar:::test_create_good_data("survival", to_data_object = FALSE),
  .as_prediction_table = TRUE
)

# Novelty from list of novelty detectors.
predictions_novelty_3 <- familiar::predict(
  object = list(fam_model@novelty_detector),
  newdata = data,
  .as_prediction_table = TRUE
)

# Risk stratification.
predictions_risk <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "risk_stratification",
  .as_prediction_table = TRUE
)

testthat::test_that("Survival models can predict using external data.", {
  # Check that predictions 1 and 2 are equal.
  testthat::expect_equal(predictions_1, predictions_2, ignore_attr = TRUE)
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_2, ignore_attr = TRUE)
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_3, ignore_attr = TRUE)

  # Check that the number of rows is equal to the input data.
  testthat::expect_equal(nrow(predictions_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_surv@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_risk@data), nrow(data@data))

  # Check that all predictions are valid.
  testthat::expect_true(familiar:::any_predictions_valid(predictions_1))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_2))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_surv))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_1))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_2))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_risk))
  
  # Test on survival probability: that 
  testthat::expect_gt(predictions_surv@time, predictions_surv_early@time)
  testthat::expect_true(all(predictions_surv@data$predicted_outcome < predictions_surv_early@data$predicted_outcome))
})

# Survival data with one instance ----------------------------------------------

# Create one-sample dataset.
data <- familiar:::test_create_one_sample_data("survival")

# Default predictions using data object.
predictions_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  .as_prediction_table = TRUE
)

# Default predictions using data.table.
predictions_2 <- familiar::predict(
  object = fam_model,
  newdata = familiar:::test_create_one_sample_data(
    "survival", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

# Survival probability.
predictions_surv <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "survival_probability",
  .as_prediction_table = TRUE
)

# Novelty.
predictions_novelty_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "novelty",
  .as_prediction_table = TRUE
)

# Novelty predictions using data.table.
predictions_novelty_2 <- familiar::predict(
  object = fam_model@novelty_detector,
  newdata = familiar:::test_create_one_sample_data(
    "survival", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

# Risk stratification.
predictions_risk <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "risk_stratification",
  .as_prediction_table = TRUE
)

testthat::test_that("Surival models can predict single instances using external data.", {
  # Check that predictions 1 and 2 are equal.
  testthat::expect_equal(predictions_1, predictions_2, ignore_attr = TRUE)
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_2, ignore_attr = TRUE)

  # Check that the number of rows is equal to the input data.
  testthat::expect_equal(nrow(predictions_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_surv@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_risk@data), nrow(data@data))

  # Check that all predictions are valid.
  testthat::expect_true(familiar:::any_predictions_valid(predictions_1))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_2))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_surv))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_1))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_2))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_risk))
})

# Binomial data with multiple instances ----------------------------------------

# Create a dataset using the good dataset.
data <- familiar:::test_create_good_data("binomial")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "glm_logistic",
  create_novelty_detector = TRUE
)

# Default predictions using data object.
predictions_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  .as_prediction_table = TRUE
)

# Default predictions using data.table.
predictions_2 <- familiar::predict(
  object = fam_model,
  newdata = familiar:::test_create_good_data(
    "binomial", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

# Novelty.
predictions_novelty_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "novelty",
  .as_prediction_table = TRUE
)

# Novelty predictions using data.table.
predictions_novelty_2 <- familiar::predict(
  object = fam_model@novelty_detector,
  newdata = familiar:::test_create_good_data(
    "binomial", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

testthat::test_that("Binomial models can predict using external data.", {
  # Check that predictions 1 and 2 are equal.
  testthat::expect_equal(predictions_1, predictions_2, ignore_attr = TRUE)
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_2, ignore_attr = TRUE)

  # Check that the number of rows is equal to the input data.
  testthat::expect_equal(nrow(predictions_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_2@data), nrow(data@data))

  # Check that all predictions are valid.
  testthat::expect_true(familiar:::any_predictions_valid(predictions_1))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_2))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_1))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_2))
})

# Binomial data with one instance ----------------------------------------------

# Create one-sample dataset.
data <- familiar:::test_create_one_sample_data("binomial")

# Default predictions using data object.
predictions_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  .as_prediction_table = TRUE
)

# Default predictions using data.table.
predictions_2 <- familiar::predict(
  object = fam_model,
  newdata = familiar:::test_create_one_sample_data(
    "binomial", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

# Novelty.
predictions_novelty_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "novelty",
  .as_prediction_table = TRUE
)

# Novelty predictions using data.table.
predictions_novelty_2 <- familiar::predict(
  object = fam_model@novelty_detector,
  newdata = familiar:::test_create_one_sample_data(
    "binomial", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

testthat::test_that("Binomial models can predict single instances using external data.", {
  # Check that predictions 1 and 2 are equal.
  testthat::expect_equal(predictions_1, predictions_2, ignore_attr = TRUE)
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_2, ignore_attr = TRUE)

  # Check that the number of rows is equal to the input data.
  testthat::expect_equal(nrow(predictions_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_2@data), nrow(data@data))

  # Check that all predictions are valid.
  testthat::expect_true(familiar:::any_predictions_valid(predictions_1))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_2))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_1))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_2))
})

# Multinomial data with multiple instances -------------------------------------

# Create a dataset using the good dataset.
data <- familiar:::test_create_good_data("multinomial")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "lasso_multinomial",
  create_novelty_detector = TRUE
)

# Default predictions using data object.
predictions_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  .as_prediction_table = TRUE
)

# Default predictions using data.table.
predictions_2 <- familiar::predict(
  object = fam_model,
  newdata = familiar:::test_create_good_data(
    "multinomial", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

# Novelty.
predictions_novelty_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "novelty",
  .as_prediction_table = TRUE
)

# Novelty predictions using data.table.
predictions_novelty_2 <- familiar::predict(
  object = fam_model@novelty_detector,
  newdata = familiar:::test_create_good_data(
    "multinomial", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

testthat::test_that("Multinomial models can predict using external data.", {
  # Check that predictions 1 and 2 are equal.
  testthat::expect_equal(predictions_1, predictions_2, ignore_attr = TRUE)
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_2, ignore_attr = TRUE)

  # Check that the number of rows is equal to the input data.
  testthat::expect_equal(nrow(predictions_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_2@data), nrow(data@data))

  # Check that all predictions are valid.
  testthat::expect_true(familiar:::any_predictions_valid(predictions_1))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_2))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_1))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_2))
})

# Multinomial data with one instance -------------------------------------------

# Create one-sample dataset.
data <- familiar:::test_create_one_sample_data("multinomial")

# Default predictions using data object.
predictions_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  .as_prediction_table = TRUE
)

# Default predictions using data.table.
predictions_2 <- familiar::predict(
  object = fam_model,
  newdata = familiar:::test_create_one_sample_data(
    "multinomial", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

# Novelty.
predictions_novelty_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "novelty",
  .as_prediction_table = TRUE
)

# Novelty predictions using data.table.
predictions_novelty_2 <- familiar::predict(
  object = fam_model@novelty_detector,
  newdata = familiar:::test_create_one_sample_data(
    "multinomial", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

testthat::test_that("Multinomial models can predict single instances using external data.", {
  # Check that predictions 1 and 2 are equal.
  testthat::expect_equal(predictions_1, predictions_2, ignore_attr = TRUE)
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_2, ignore_attr = TRUE)

  # Check that the number of rows is equal to the input data.
  testthat::expect_equal(nrow(predictions_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_2@data), nrow(data@data))

  # Check that all predictions are valid.
  testthat::expect_true(familiar:::any_predictions_valid(predictions_1))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_2))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_1))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_2))
})

# Continuous data with multiple instances --------------------------------------

# Create a dataset using the good dataset.
data <- familiar:::test_create_good_data("continuous")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "lasso_gaussian",
  create_novelty_detector = TRUE
)

# Default predictions using data object.
predictions_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  .as_prediction_table = TRUE
)

# Default predictions using data.table.
predictions_2 <- familiar::predict(
  object = fam_model,
  newdata = familiar:::test_create_good_data(
    "continuous", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

# Novelty.
predictions_novelty_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "novelty",
  .as_prediction_table = TRUE
)

# Novelty predictions using data.table.
predictions_novelty_2 <- familiar::predict(
  object = fam_model@novelty_detector,
  newdata = familiar:::test_create_good_data(
    "continuous", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

testthat::test_that("Continuous models can predict using external data.", {
  # Check that predictions 1 and 2 are equal.
  # testthat::expect_equal(predictions_1, predictions_2, ignore_attr = TRUE)
  testthat::expect_equal(predictions_1@data, predictions_2@data, ignore_attr = TRUE)
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_2, ignore_attr = TRUE)

  # Check that the number of rows is equal to the input data.
  testthat::expect_equal(nrow(predictions_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_2@data), nrow(data@data))

  # Check that all predictions are valid.
  testthat::expect_true(familiar:::any_predictions_valid(predictions_1))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_2))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_1))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_2))
})

# Continuous data with one instance --------------------------------------------

# Create one-sample dataset.
data <- familiar:::test_create_one_sample_data("continuous")

# Default predictions using data object.
predictions_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  .as_prediction_table = TRUE
)

# Default predictions using data.table.
predictions_2 <- familiar::predict(
  object = fam_model,
  newdata = familiar:::test_create_one_sample_data(
    "continuous", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

# Novelty.
predictions_novelty_1 <- familiar::predict(
  object = fam_model,
  newdata = data,
  type = "novelty",
  .as_prediction_table = TRUE
)

# Novelty predictions using data.table.
predictions_novelty_2 <- familiar::predict(
  object = fam_model@novelty_detector,
  newdata = familiar:::test_create_one_sample_data(
    "continuous", to_data_object = FALSE
  ),
  .as_prediction_table = TRUE
)

testthat::test_that("Continuous models can predict single instances using external data.", {
  # Check that predictions 1 and 2 are equal.
  testthat::expect_equal(predictions_novelty_1, predictions_novelty_2, ignore_attr = TRUE)

  # Check that the number of rows is equal to the input data.
  testthat::expect_equal(nrow(predictions_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_2@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_1@data), nrow(data@data))
  testthat::expect_equal(nrow(predictions_novelty_2@data), nrow(data@data))

  # Check that all predictions are valid.
  testthat::expect_true(familiar:::any_predictions_valid(predictions_1))
  testthat::expect_true(familiar:::any_predictions_valid(predictions_2))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_1))
  testthat::expect_true(familiar:::all_predictions_valid(predictions_novelty_2))
})

# Don't perform any further tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar_data_creation_unit_test <- function(outcome_type) {
  if (outcome_type == "continuous") {
    learner <- "glm_gaussian"
  } else if (outcome_type == "binomial") {
    learner <- "glm_logistic"
  } else if (outcome_type == "multinomial") {
    learner <- "glm_multinomial"
  } else if (outcome_type == "survival") {
    learner <- "cox"
  } else {
    ..error_outcome_type_not_implemented(outcome_type)
  }

  # Create a dataset using the good dataset.
  data <- familiar:::test_create_good_data(outcome_type)

  # Train a simple linear GLM using the good dataset.
  fam_model <- suppressWarnings(familiar:::test_train(
    data = data,
    cluster_method = "none",
    imputation_method = "simple",
    hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
    learner = learner,
    time_max = 3.5))
  
  testthat::test_that(
    paste0(
      "1. familiarData for the ", outcome_type, " outcome can be created from ",
      "good data and the model created using the complete data set."),
    {
      # Test with small, but good dataset.
      fam_data <- suppressWarnings(familiar::as_familiar_data(
        object = fam_model,
        data = familiar:::test_create_small_good_data(outcome_type = outcome_type),
        estimation_type = "point",
        verbose = FALSE))
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )
  
  testthat::test_that(
    paste0(
      "2. familiarData for the ", outcome_type, " outcome can be created from ",
      "single-sample data and the model created using the complete data set."), 
    {
      # Test with one-sample dataset.
      fam_data <- familiar::as_familiar_data(
        object = fam_model,
        data = familiar:::test_create_one_sample_data(outcome_type = outcome_type),
        estimation_type = "point",
        verbose = FALSE
      )
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )

  testthat::test_that(
    paste0(
      "3. familiarData for the ", outcome_type, " outcome can be created from ",
      "empty data and the model created using the complete data set."), 
    {
    # Test with empty dataset
      fam_data <- familiar::as_familiar_data(
        object = fam_model,
        data = familiar:::test_create_empty_data(outcome_type = outcome_type),
        estimation_type = "point",
        verbose = FALSE)
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )

  testthat::test_that(
    paste0(
      "4. familiarData for the ", outcome_type, " outcome can be created from ",
      "the bad dataset and the model created using the complete data set."), 
    {
      # Test with bad dataset
      fam_data <- suppressWarnings(familiar::as_familiar_data(
        object = fam_model,
        data = familiar:::test_create_small_bad_data(outcome_type = outcome_type),
        estimation_type = "point",
        verbose = FALSE))
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )
  
  # Create a dataset using the one-feature dataset.
  data <- familiar:::test_create_single_feature_data(outcome_type = outcome_type)
  
  # Train a simple linear GLM using the one feature dataset.
  fam_model <- suppressWarnings(familiar:::test_train(
    data = data,
    cluster_method = "none",
    imputation_method = "simple",
    hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
    learner = learner,
    time_max = 3.5))

  testthat::test_that(
    paste0(
      "5. familiarData for the ", outcome_type, " outcome can be created from ",
      "the one-feature dataset and the model created using the same data set."),
    {
      # Test with one-feature dataset.
      fam_data <- suppressWarnings(familiar::as_familiar_data(
        object = fam_model,
        data = data,
        estimation_type = "point",
        verbose = FALSE))
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )

  testthat::test_that(
    paste0(
      "6. familiarData for the ", outcome_type, " outcome can be created from ",
      "the one-feature, one-sample dataset and the model created using the one-feature data set."),
    {
      # Test with one-sample dataset
      fam_data <- familiar::as_familiar_data(
        object = fam_model,
        data = familiar:::test_create_single_feature_one_sample_data(outcome_type = outcome_type),
        estimation_type = "point",
        verbose = FALSE)
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )
  
  # Create a dataset using the bad dataset.
  data <- familiar:::test_create_bad_data(outcome_type = outcome_type)

  # Train a simple linear GLM using the bad dataset. The model will fail to
  # build.
  fam_model <- suppressWarnings(familiar:::test_train(
    data = data,
    cluster_method = "none",
    imputation_method = "simple",
    hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
    learner = learner,
    time_max = 3.5))

  testthat::test_that(
    paste0(
      "7. familiarData for the ", outcome_type, " outcome can be created from ",
      "good data and the NULL model created using the bad data set."), 
    {
      # Test with small, but good dataset.
      fam_data <- suppressWarnings(familiar::as_familiar_data(
        object = fam_model,
        data = familiar:::test_create_small_good_data(outcome_type = outcome_type),
        estimation_type = "point",
        verbose = FALSE))
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )

  testthat::test_that(
    paste0(
      "8. familiarData for the ", outcome_type, " outcome can be created from ",
      "single-sample data and the NULL model created using the bad data set."),
    {
      # Test with one-sample dataset.
      fam_data <- familiar::as_familiar_data(
        object = fam_model,
        data = familiar:::test_create_one_sample_data(outcome_type = outcome_type),
        estimation_type = "point",
        verbose = FALSE)
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )
  
  testthat::test_that(
    paste0(
      "9. familiarData for the ", outcome_type, " outcome can be created from ",
      "empty data and the NULL model created using the bad data set."), 
    {
      # Test with empty dataset
      fam_data <- familiar::as_familiar_data(
        object = fam_model,
        data = familiar:::test_create_empty_data(outcome_type = outcome_type),
        estimation_type = "point",
        verbose = FALSE)
      
      # Test that an familiarDataObject is created.
      testthat::expect_s4_class(fam_data, "familiarData")
    }
  )
}

familiar_data_creation_unit_test("continuous")
familiar_data_creation_unit_test("binomial")
familiar_data_creation_unit_test("multinomial")
familiar_data_creation_unit_test("survival")

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test
# Note that one-sample kaplan-meier curves can be created.
familiar:::test_plots(
  plot_function = familiar:::plot_kaplan_meier,
  not_available_all_prospective = TRUE,
  outcome_type_available = c("survival"),
  data_element = "risk_stratification_data",
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_kaplan_meier,
  data_element = "risk_stratification_data",
  outcome_type_available = c("survival"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_kaplan_meier,
  data_element = "risk_stratification_data",
  outcome_type_available = c("survival"),
  use_prediction_table = TRUE,
  prediction_type = list("survival" = "risk_stratification"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_kaplan_meier,
  data_element = "risk_stratification_data",
  outcome_type_available = c("survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "color_by" = "group"),
  debug = debug_flag
)

# Fixed stratification with 3 groups works correctly.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_kaplan_meier,
  data_element = "risk_stratification_data",
  outcome_type_available = c("survival"),
  experiment_args = list(
    stratification_method = "fixed"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "color_by" = "group"),
  debug = debug_flag
)

# Fixed stratification with 5 groups works correctly.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_kaplan_meier,
  data_element = "risk_stratification_data",
  outcome_type_available = c("survival"),
  experiment_args = list(
    stratification_method = "fixed",
    stratification_threshold = c(0.20, 0.40, 0.60, 0.80)),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "color_by" = "group"),
  debug = debug_flag
)

# Fixed stratification with 5 groups works correctly.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_kaplan_meier,
  data_element = "risk_stratification_data",
  outcome_type_available = c("survival"),
  experiment_args = list(
    stratification_method = c("median", "fixed"),
    stratification_threshold = c(0.20, 0.40, 0.60, 0.80)),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "color_by" = "group",
    "split_by" = "stratification_method"),
  debug = debug_flag
)



# Test plotting from dataObject.
data <- familiar:::test_create_good_data(outcome_type = "survival")
p <- familiar::plot_kaplan_meier(object = data)
testthat::test_that("Plotting kaplan-meier curves using dataObject works (survival).", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})


# Test plotting from dataObject with two groups.
data <- familiar:::test_create_good_data(outcome_type = "survival", two_groups = TRUE)
p <- familiar::plot_kaplan_meier(object = data)
testthat::test_that("Plotting kaplan-meier curves for two groups works.", {
  testthat::expect_length(p, 1L)
  testthat::expect_true(is(p[[1L]], "gtable"))
})


# Test plotting from data.table.
data <- familiar:::test_create_good_data(outcome_type = "survival", to_data_object = FALSE)
p <- familiar::plot_kaplan_meier(
  object = data,
  feature_similarity_metric = "spearman",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  outcome_type = "survival",
  outcome_column = c("outcome_time", "outcome_event")
)
testthat::test_that("Plotting kaplan-meier curves using data.table works (survival).", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})


# Test plotting from data.table without feature data.
data <- familiar:::test_create_data_without_feature(outcome_type = "survival", to_data_object = FALSE)
p <- familiar::plot_kaplan_meier(
  object = data,
  feature_similarity_metric = "spearman",
  batch_id_column = "batch_id",
  sample_id_column = "sample_id",
  series_id_column = "series_id",
  outcome_type = "survival",
  outcome_column = c("outcome_time", "outcome_event")
)
testthat::test_that("Plotting kaplan-meier curves using data.table without any feature works (survival).", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})


# Test plotting from dataObject with a risk_group_column.
data <- familiar:::test_create_good_data(outcome_type = "survival")
data@data[, "risk_group" := "risk-group A"]
data@data[51L:100L, "risk_group" := "risk-group B"]
p <- familiar::plot_kaplan_meier(object = data, risk_group_column = "risk_group")
testthat::test_that("Plotting kaplan-meier curves using data.table without a risk_group_column works (survival).", {
  testthat::expect_true(is(p[[1L]], "gtable"))
})

# Don't perform any further tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

# Create a dataset using the good dataset.
data <- familiar:::test_create_small_good_data("survival")

# Train a simple linear GLM using the good dataset.
fam_model <- familiar:::test_train(
  data = data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list("sign_size" = familiar:::get_n_features(data)),
  learner = "cox",
  hpo_metric = "concordance_index",
  vimp_aggregation_method = "mean",
  vimp_aggregation_rank_threshold = 5)

# familiarModel ----------------------------------------------------------------
testthat::test_that("Conversion of single familiarModel to familiarEnsemble works", {
  testthat::expect_s4_class(
    familiar::as_familiar_ensemble(object = fam_model),
    "familiarEnsemble")
})

testthat::test_that("Conversion of multiple familiarModel objects to familiarEnsemble works", {
  testthat::expect_s4_class(
    familiar::as_familiar_ensemble(object = list(fam_model, fam_model)),
    "familiarEnsemble")
})

testthat::test_that("Conversion of single familiarModel to familiarData works", {
  testthat::expect_s4_class(
    familiar::as_familiar_data(
      object = fam_model,
      data = data,
      estimation_type = "point",
      verbose = FALSE),
    "familiarData")
})

testthat::test_that("Conversion of multiple familiarModel objects to familiarData works", {
  testthat::expect_s4_class(
    familiar::as_familiar_data(
      object = list(fam_model, fam_model),
      data = data,
      estimation_type = "point",
      verbose = FALSE),
    "familiarData")
})

testthat::test_that("Conversion of single familiarModel to familiarCollection works", {
  testthat::expect_s4_class(
    familiar::as_familiar_collection(
      object = fam_model,
      data = data,
      estimation_type = "point",
      verbose = FALSE),
    "familiarCollection")
})

testthat::test_that("Conversion of multiple familiarModel objects to familiarCollection works", {
  testthat::expect_s4_class(
    familiar::as_familiar_collection(
      object = list(fam_model, fam_model),
      data = data,
      estimation_type = "point",
      verbose = FALSE),
    "familiarCollection")
})

# familiarEnsemble -------------------------------------------------------------

# Create familiarEnsemble from fam_model objects.
fam_ensemble <- familiar::as_familiar_ensemble(object = list(fam_model, fam_model))

testthat::test_that("Conversion of familiarEnsemble to familiarEnsemble works", {
  testthat::expect_s4_class(
    familiar::as_familiar_ensemble(object = fam_ensemble),
    "familiarEnsemble")
})

testthat::test_that("Conversion of multiple familiarEnsemble objects to familiarEnsemble fails", {
  testthat::expect_error(
    familiar::as_familiar_ensemble(object = list(fam_ensemble, fam_ensemble)))
})

testthat::test_that("Conversion of familiarEnsemble to familiarData works", {
  testthat::expect_s4_class(
    familiar::as_familiar_data(
      object = fam_ensemble,
      data = data,
      estimation_type = "point",
      verbose = FALSE),
    "familiarData")
})

testthat::test_that("Conversion of multiple familiarEnsemble objects to familiarData fails", {
  testthat::expect_error(
    familiar::as_familiar_data(
      object = list(fam_ensemble, fam_ensemble),
      data = data, 
      verbose = FALSE))
})

testthat::test_that("Conversion of familiarEnsemble to familiarCollection works", {
  testthat::expect_s4_class(
    familiar::as_familiar_collection(
      object = fam_ensemble,
      data = data,
      estimation_type = "point",
      verbose = FALSE),
    "familiarCollection")
})

testthat::test_that("Conversion of multiple familiarEnsemble objects to familiarCollection fails", {
  testthat::expect_error(
    familiar::as_familiar_collection(
      object = list(fam_ensemble, fam_ensemble),
      data = data,
      verbose = FALSE))
})

# familiarData -----------------------------------------------------------------

# Create familiarData from fam_model objects.
fam_data_1 <- familiar::as_familiar_data(
  object = list(fam_model, fam_model),
  estimation_type = "point",
  data = data)
fam_data_2 <- familiar::as_familiar_data(
  object = list(fam_model, fam_model),
  estimation_type = "point",
  data = data)

testthat::test_that("Conversion of familiarData to familiarEnsemble fails", {
  testthat::expect_error(
    familiar::as_familiar_ensemble(object = fam_data_1))
})

testthat::test_that("Conversion of multiple familiarData objects to familiarEnsemble fails", {
  testthat::expect_error(
    familiar::as_familiar_ensemble(object = list(fam_data_1, fam_data_2)))
})

testthat::test_that("Conversion of familiarData to familiarData works", {
  testthat::expect_s4_class(
    familiar::as_familiar_data(object = fam_data_1),
    "familiarData")
})

testthat::test_that("Conversion of multiple familiarData objects to familiarData fails", {
  testthat::expect_error(
    familiar::as_familiar_data(object = list(fam_data_1, fam_data_2)))
})

testthat::test_that("Conversion of familiarData to familiarCollection works", {
  testthat::expect_s4_class(
    familiar::as_familiar_collection(object = fam_data_1),
    "familiarCollection")
})

testthat::test_that("Conversion of multiple familiarData objects to familiarCollection works", {
  testthat::expect_s4_class(
    familiar::as_familiar_collection(object = list(fam_data_1, fam_data_2)),
    "familiarCollection")
})

testthat::test_that("Conversion of multiple identical familiarData objects to familiarCollection fails", {
  testthat::expect_s4_class(
    familiar::as_familiar_collection(object = list(fam_data_1, fam_data_1)),
    "familiarCollection")
})

# familiarCollection -----------------------------------------------------------
# Create familiarCollection from fam_data objects.
fam_collection <- familiar::as_familiar_collection(object = list(fam_data_1, fam_data_2))

testthat::test_that("Conversion of familiarCollection to familiarEnsemble fails", {
  testthat::expect_error(
    familiar::as_familiar_ensemble(object = fam_collection))
})

testthat::test_that("Conversion of multiple familiarCollection objects to familiarEnsemble fails", {
  testthat::expect_error(
    familiar::as_familiar_ensemble(object = list(fam_collection, fam_collection)))
})

testthat::test_that("Conversion of familiarCollection to familiarData fails", {
  testthat::expect_error(
    familiar::as_familiar_data(object = fam_collection))
})

testthat::test_that("Conversion of multiple familiarCollection objects to familiarData fails", {
  testthat::expect_error(
    familiar::as_familiar_data(object = list(fam_collection, fam_collection)))
})

testthat::test_that("Conversion of familiarCollection to familiarCollection works", {
  testthat::expect_s4_class(
    familiar::as_familiar_collection(object = fam_collection),
    "familiarCollection")
})

testthat::test_that("Conversion of multiple familiarCollection objects to familiarCollection fails", {
  testthat::expect_error(
    familiar::as_familiar_collection(object = list(fam_collection, fam_collection)))
})

# Mixed elements ---------------------------------------------------------------
testthat::test_that("Conversion of mixed objects to familiarEnsemble fails", {
  testthat::expect_error(
    familiar::as_familiar_ensemble(object = list(fam_model, fam_ensemble)))
})

testthat::test_that("Conversion of mixed objects to familiarData fails", {
  testthat::expect_error(
    familiar::as_familiar_data(object = list(fam_model, fam_ensemble)))
})

testthat::test_that("Conversion of mixed objects to to familiarCollection fails", {
  testthat::expect_error(
    familiar::as_familiar_collection(object = list(fam_model, fam_ensemble)))
})

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# Cluster method objects -------------------------------------------------------
setClass(
  "clusterMethod",
  slots = list(
    "method" = "character",
    "data_type" = "character",
    "cluster_cut_method" = "character",
    "representation_method" = "character",
    "similarity_table" = "ANY",
    "object" = "ANY"
  ),
  prototype = list(
    "method" = NA_character_,
    "data_type" = NA_character_,
    "cluster_cut_method" = "none",
    "representation_method" = "none",
    "similarity_table" = NULL,
    "object" = NULL
  )
)

## clusterMethodHierarchical ---------------------------------------------------
setClass(
  "clusterMethodHierarchical",
  contains = "clusterMethod",
  slots = list(
    "similarity_metric" = "character",
    "similarity_threshold" = "numeric"
  ),
  prototype = list(
    "similarity_metric" = NA_character_,
    "similarity_threshold" = NA_real_
  )
)

## clusterMethodHClust ---------------------------------------------------------
setClass(
  "clusterMethodHClust",
  contains = "clusterMethodHierarchical",
  slots = list("linkage_method" = "character"),
  prototype = list("linkage_method" = NA_character_)
)

## clusterMethodAgnes ----------------------------------------------------------
setClass(
  "clusterMethodAgnes",
  contains = "clusterMethodHierarchical",
  slots = list("linkage_method" = "character"),
  prototype = list("linkage_method" = NA_character_)
)

## clusterMethodDiana ----------------------------------------------------------
setClass(
  "clusterMethodDiana",
  contains = "clusterMethodHierarchical"
)

## clusterMethodPAM ------------------------------------------------------------
setClass(
  "clusterMethodPAM",
  contains = "clusterMethod",
  slots = list("similarity_metric" = "character"),
  prototype = list("similarity_metric" = "character")
)

## clusterMethodNone -----------------------------------------------------------
setClass(
  "clusterMethodNone",
  contains = "clusterMethod"
)



# Similarity table object ------------------------------------------------------

## similarityTable -------------------------------------------------------------
setClass(
  "similarityTable",
  slots = list(
    "data" = "ANY",
    "similarity_metric" = "character",
    "data_type" = "character"
  ),
  prototype = list(
    "data" = NULL,
    "similarity_metric" = NA_character_,
    "data_type" = NA_character_
  )
)

## noSimilarityTable -----------------------------------------------------------
setClass(
  "noSimilarityTable",
  slots = list(
    "data" = "ANY",
    "data_type" = "character"
  ),
  prototype = list(
    "data" = NULL,
    "data_type" = NA_character_
  )
)



# clustering objects -----------------------------------------------------------

## clusteringObject ------------------------------------------------------------
setClass(
  "clusteringObject",
  slots = list(
    "representation_method" = "character",
    "cluster_features" = "character"
  ),
  prototype = list(
    "representation_method" = "none",
    "cluster_features" = NA_character_
  )
)

## singularClusteringObject ----------------------------------------------------
setClass(
  "singularClusteringObject",
  slots = list("cluster_features" = "character"),
  prototype = list("cluster_features" = NA_character_)
)



# set_object_parameters methods ------------------------------------------------

## set_object_parameters (none) ------------------------------------------------
setMethod(
  "set_object_parameters",
  signature(object = "clusterMethodNone"),
  function(object, ...) {
    
    return(object)
  }
)



## set_object_parameters (general) ---------------------------------------------
setMethod(
  "set_object_parameters",
  signature(object = "clusterMethod"),
  function(
    object,
    cluster_cut_method = NULL,
    cluster_representation_method = NULL,
    ...
  ) {
    
    # Check that data type is set correctly. This is used for many
    # checks.
    if (!object@data_type %in% c("feature", "sample", "cluster")) {
      ..error_reached_unreachable_code(paste0(
        "set_object_parameters,clusterMethod: the data_type attribute was not correctly set."
      ))
    }
    
    # Cut methods are optional, and default to "none".
    if (!is.null(cluster_cut_method)) {
      
      # Check cluster cut method.
      .check_parameter_value_is_valid(
        x = cluster_cut_method,
        var_name = ifelse(
          object@data_type == "cluster",
          "cluster_cut_method",
          paste0(object@data_type, "_cluster_cut_method")
        ),
        values = .get_available_cluster_cut_methods(object@method)
      )
      
      # Set cluster cut method.
      object@cluster_cut_method <- cluster_cut_method
    }
    
    # Check representation method.
    if (!is.null(cluster_representation_method) && object@cluster_cut_method != "none") {
      
      # Check representation method.
      .check_parameter_value_is_valid(
        x = cluster_representation_method,
        var_name = ifelse(
          object@data_type == "cluster",
          "cluster_representation_method",
          paste0(object@data_type, "_cluster_representation_method")
        ),
        values = .get_available_cluster_representation_methods(object@method)
      )
      
      # Set cluster representation method.
      object@representation_method <- cluster_representation_method
    }
    
    return(object)
  }
)



## set_object_parameters (PAM) -------------------------------------------------
setMethod(
  "set_object_parameters",
  signature(object = "clusterMethodPAM"),
  function(
    object,
    cluster_similarity_metric,
    ...
  ) {
    
    # Call the method for the parent class (clusterMethod) first.
    object <- methods::callNextMethod()
    
    # Check that similarity metric is valid.
    .check_parameter_value_is_valid(
      x = cluster_similarity_metric,
      var_name = paste0(object@data_type, "_similarity_metric"),
      values = .get_available_similarity_metrics(data_type = object@data_type)
    )
    
    # Set similarity metric.
    object@similarity_metric <- cluster_similarity_metric
    
    return(object)
  }
)



## set_object_parameters (general hierarchical) --------------------------------
setMethod(
  "set_object_parameters",
  signature(object = "clusterMethodHierarchical"),
  function(
    object,
    cluster_similarity_metric,
    cluster_similarity_threshold = NULL,
    ...
  ) {
    
    # Call the method for the parent class (clusterMethod) first.
    object <- methods::callNextMethod()
    
    # Check that similarity metric is valid.
    .check_parameter_value_is_valid(
      x = cluster_similarity_metric,
      var_name = paste0(object@data_type, "_similarity_metric"),
      values = .get_available_similarity_metrics(data_type = object@data_type)
    )
    
    # Set similarity metric.
    object@similarity_metric <- cluster_similarity_metric
    
    if (object@cluster_cut_method %in% c("fixed_cut", "dynamic_cut")) {
      # Check cutting height for fixed cut. Multiple cut heights are
      # possible. Use as_distance to get two-value ranges, but note that
      # these values are similarity otherwise.
      sapply(
        cluster_similarity_threshold,
        .check_number_in_valid_range,
        var_name = paste0(object@data_type, "_similarity_threshold"),
        range = get_similarity_range(
          similarity_metric = object@similarity_metric,
          as_distance = TRUE
        )
      )
      
      # Attach to object.
      object@similarity_threshold <- cluster_similarity_threshold
    }
    
    return(object)
  }
)



## set_object_parameters (hclust) ----------------------------------------------
setMethod(
  "set_object_parameters",
  signature(object = "clusterMethodHClust"),
  function(
    object,
    cluster_linkage_method,
    ...
  ) {
    
    # Call next method (clusterMethodHierarchical). This will also call the
    # method for its parent method (clusterMethod).
    object <- methods::callNextMethod()
    
    # Check that linkage method is valid.
    .check_parameter_value_is_valid(
      x = cluster_linkage_method,
      var_name = paste0(object@data_type, "_linkage_method"),
      values = .get_available_linkage_methods(cluster_method = object@method)
    )
    
    # attach to object.
    object@linkage_method <- cluster_linkage_method
    
    return(object)
  }
)



## set_object_parameters (agnes) -----------------------------------------------
setMethod(
  "set_object_parameters",
  signature(object = "clusterMethodAgnes"),
  function(
    object,
    cluster_linkage_method,
    ...
  ) {
    
    # Call next method (clusterMethodHierarchical). This will also call the
    # method for its parent method (clusterMethod).
    object <- methods::callNextMethod()
    
    # Check that linkage method is valid.
    .check_parameter_value_is_valid(
      x = cluster_linkage_method,
      var_name = paste0(object@data_type, "_linkage_method"),
      values = .get_available_linkage_methods(cluster_method = object@method)
    )
    
    # attach to object.
    object@linkage_method <- cluster_linkage_method
    
    return(object)
  }
)



## set_object_parameters (diana) -----------------------------------------------
setMethod(
  "set_object_parameters",
  signature(object = "clusterMethodDiana"),
  function(object, ...) {
    
    # Call next method (clusterMethodHierarchical). This will also call the
    # method for its parent method (clusterMethod).
    object <- methods::callNextMethod()
    
    return(object)
  }
)


# set_similarity_table methods -------------------------------------------------

## set_similarity_table (missing, dataObject -----------------------------------
setMethod(
  "set_similarity_table",
  signature(
    object = "missing",
    data = "dataObject"
  ),
  function(
    object,
    data,
    feature_info_list,
    similarity_metric,
    data_type,
    ...
  ) {
    
    # For calls where we just want to create the similarity table, e.g.
    # in ..extract_feature_similarity. Here we create a generic
    # clusterMethodPAM object just to create the similarity table.
    object <- methods::new(
      "clusterMethodPAM",
      data_type = data_type,
      similarity_metric = similarity_metric
    )
    
    # Pass to set_similarity_table.
    object <- set_similarity_table(
      object = object,
      data = data,
      feature_info_list = feature_info_list,
      ...
    )
    
    # Return the table itself.
    return(object@similarity_table@data)
  }
)



## set_similarity_table (none, dataObject) -------------------------------------
setMethod(
  "set_similarity_table",
  signature(
    object = "clusterMethodNone",
    data = "dataObject"
  ),
  function(
    object,
    data,
    feature_info_list,
    ...
  ) {
    
    # Try to get similarity table
    similarity_table <- .set_similarity_table(
      object = object,
      data = data,
      feature_info_list = feature_info_list,
      ...
    )
    
    # Set the similarity_table attribute.
    object@similarity_table <- similarity_table
    
    return(object)
  }
)



## set_similarity_table (clusterMethod, dataObject) ----------------------------
setMethod(
  "set_similarity_table",
  signature(
    object = "clusterMethod",
    data = "dataObject"
  ),
  function(
    object,
    data,
    feature_info_list,
    ...
  ) {
    
    # Try to get similarity table
    similarity_table <- .set_similarity_table(
      object = object,
      data = data,
      feature_info_list = feature_info_list,
      ...
    )
    
    # If setting the similarity table does not work, we cannot create clusters.
    # Switch to "none" method instead.
    if (is.null(similarity_table)) {
      object <- create_cluster_method_object(
        cluster_method = "none",
        data_type = object@data_type
      )
      
      return(set_similarity_table(
        object = object,
        data = data,
        feature_info_list = feature_info_list
      ))
    }
    
    # Set the similarity_table attribute.
    object@similarity_table <- similarity_table
    
    return(object)
  }
)


# .set_similarity_table methods ------------------------------------------------

## .set_similarity_table (none, dataObject) ------------------------------------
setMethod(
  ".set_similarity_table",
  signature(
    object = "clusterMethodNone",
    data = "dataObject"
  ),
  function(
    object,
    data,
    feature_info_list,
    ...
  ) {
    # Specific method for objects that indicate that no clustering should be
    # performed.
    
    # Check that the data are not empty.
    if (is_empty(data)) return(NULL)
    
    # Get feature columns.
    feature_columns <- get_feature_columns(data)
    
    # Sanity check.
    if (!(setequal(feature_columns, get_available_features(feature_info_list = feature_info_list)))) {
      ..error_reached_unreachable_code(paste0(
        ".set_similarity_table,clusterMethodNone,dataObject: features in data and the ",
        "feature info list are expect to be the same, but were not."
      ))
    }
    
    if (object@data_type %in% c("cluster", "feature")) {
      similarity_data <- feature_columns
      
    } else if (object@data_type == "sample") {
      # Create unique row names.
      similarity_data <- get_unique_row_names(x = data)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        ".set_similarity_table,clusterMethodNone,dataObject: ",
        "encountered an unknown data_type: ",
        object@data_type
      ))
    }
    
    # Create (no) similarity table.
    similarity_table <- methods::new(
      "noSimilarityTable",
      "data" = similarity_data,
      "data_type" = object@data_type
    )
    
    
    return(similarity_table)
  }
)



## .set_similarity_table (clusterMethod, dataObject) ---------------------------
setMethod(
  ".set_similarity_table",
  signature(
    object = "clusterMethod",
    data = "dataObject"
  ),
  function(
    object,
    data,
    feature_info_list,
    cl = NULL,
    verbose = FALSE,
    ...
  ) {
    
    # Check that the data are not empty.
    if (is_empty(data)) return(NULL)
    
    # Get feature columns.
    feature_columns <- get_feature_columns(data)
    
    # Sanity check.
    if (!(setequal(feature_columns, get_available_features(feature_info_list = feature_info_list)))) {
      ..error_reached_unreachable_code(paste0(
        ".set_similarity_table,clusterMethod,dataObject: features in data and ",
        "the feature info list are expect to be the same, but were not."
      ))
    }
    
    if (object@data_type %in% c("cluster", "feature")) {
      
      # Find similarity between features.
      similarity_data  <- compute_feature_similarity_metric(
        data = data,
        similarity_metric = object@similarity_metric,
        feature_info_list = feature_info_list,
        cl = cl,
        verbose = verbose
      )
      
    } else if (object@data_type == "sample") {
      
      # Find similarity between samples
      similarity_data <- compute_sample_similarity_metric(
        data = data,
        similarity_metric = object@similarity_metric,
        feature_info_list = feature_info_list,
        cl = cl,
        verbose = verbose
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        ".set_similarity_table,clusterMethod,dataObject: encountered an unknown data_type: ",
        object@data_type
      ))
    }
    
    # Create similarity table.
    similarity_table <- methods::new(
      "similarityTable",
      "data" = similarity_data,
      "similarity_metric" = object@similarity_metric,
      "data_type" = object@data_type
    )
    
    # Set similarity table,
    return(similarity_table)
  }
)


# get_similarity_names methods -------------------------------------------------

## get_similarity_names (NULL) -------------------------------------------------
setMethod(
  "get_similarity_names",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



## get_similarity_names (similarityTable) --------------------------------------
setMethod(
  "get_similarity_names",
  signature(object = "similarityTable"),
  function(object, ...) {
    
    element_names <- NULL
    if (object@data_type %in% c("cluster", "feature")) {
      if (!is.null(object@data)) {
        element_names <- unique(c(
          object@data$feature_name_1,
          object@data$feature_name_2
        ))
      }
      
    } else if (object@data_type == "sample") {
      if (!is.null(object@data)) {
        element_names <- unique(c(
          object@data$sample_1,
          object@data$sample_2
        ))
      }
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "get_similarity_names,similarityTable: encountered an unknown data_type: ",
        object@data_type
      ))
    }
    
    return(element_names)
  }
)



## get_similarity_names (noSimilarityTable) ------------------------------------
setMethod(
  "get_similarity_names",
  signature(object = "noSimilarityTable"),
  function(object, ...) {
    
    element_names <- NULL
    if (object@data_type %in% c("cluster", "feature", "sample")) {
      if (!is.null(object@data)) element_names <- object@data
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "get_similarity_names,noSimilarityTable: encountered an unknown data_type: ",
        object@data_type
      ))
    }
    
    return(element_names)
  }
)



# get_distance_table methods ---------------------------------------------------

## get_distance_table (clusterMethod) ------------------------------------------
setMethod(
  "get_distance_table",
  signature(object = "noSimilarityTable"),
  function(object, ...) {
    
    # Check that similarity table is set.
    if (is.null(object@similarity_table)) {
      # Create similarity table first.
      object <- do.call(
        set_similarity_table,
        args = c(
          list("object" = object),
          list(...)
        )
      )
    }
    
    # Push to get_distance_matrix for similarity tables.
    return(get_distance_table(object = object@similarity_table, ...))
  }
)



## get_distance_table (similarityTable) ----------------------------------------
setMethod(
  "get_distance_table",
  signature(object = "similarityTable"),
  function(
    object,
    include_diagonal = TRUE,
    ...
  ) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    value <- NULL
    
    # Extract similarity table.
    similarity_metric <- object@similarity_metric
    
    # Copy data from the similarity table.
    lower_triangle <- data.table::copy(object@data)
    
    # Determine whether the similarity table is for features (columns)
    # or samples (rows).
    element_names <- .get_cluster_data_type_element_name(data_type = object@data_type)
    
    # Find elements from the distance table.
    elements <- union(
      lower_triangle[[element_names[1L]]],
      lower_triangle[[element_names[2L]]]
    )
    
    # Convert similarity to distance.
    lower_triangle[, "value" := convert_similarity_to_distance(
      x = value,
      similarity_metric = similarity_metric
    )]
    
    # Add in other triangle of the table by switching around the columns.
    upper_triangle <- data.table::copy(lower_triangle)
    data.table::setnames(
      upper_triangle,
      old = element_names,
      new = rev(element_names)
    )
    
    if (include_diagonal) {
      # Create diagonals that always have distance 0.
      diagonal_table <- data.table::data.table(
        "element_1" = elements,
        "element_2" = elements,
        "value" = as.double(0.0)
      )
      
      # Add names to the diagonal table.
      data.table::setnames(
        diagonal_table,
        old = c("element_1", "element_2"),
        new = element_names
      )
      
      # Combine to single, long, table
      distance_table <- rbind(lower_triangle, diagonal_table, upper_triangle)
      
    } else {
      # Combine upper and lower triangles.
      distance_table <- rbind(lower_triangle, upper_triangle)
    }
    
    return(distance_table)
  }
)



## get_distance_table (noSimilarityTable) --------------------------------------
setMethod(
  "get_distance_table",
  signature(object = "noSimilarityTable"),
  function(object, ...) {
    # For noSimilarityTable objects, return NULL.
    
    return(NULL)
  }
)


# get_distance_matrix methods --------------------------------------------------

## get_distance_matrix (clusterMethod) -----------------------------------------
setMethod(
  "get_distance_matrix",
  signature(object = "clusterMethod"),
  function(object, ...) {
    
    # Check that similarity table is set.
    if (is.null(object@similarity_table)) {
      # Create similarity table first.
      object <- do.call(
        set_similarity_table,
        args = c(
          list("object" = object),
          list(...)
        )
      )
    }
    
    return(get_distance_matrix(object = object@similarity_table))
  }
)



## get_distance_matrix (similarityTable) ---------------------------------------
setMethod(
  "get_distance_matrix",
  signature(object = "similarityTable"),
  function(object, ...) {
    # Converts a similarity table into a distance matrix.
    
    # Determine whether the similarity table is for features (columns) or
    # samples (rows).
    element_names <- .get_cluster_data_type_element_name(data_type = object@data_type)
    
    # Convert a similarity table to a full distance table first.
    distance_table <- get_distance_table(object = object)
    
    # Check if the element_names appear as instance names. This will cause an
    # error if not handled correctly.
    if (length(intersect(
      c(distance_table[[element_names[1L]]], distance_table[[element_names[2L]]]),
      element_names
    )) > 0L) {
      # If any sample or feature names match the column name, add a
      # low-collision randomly generated string as a prefix to the column names.
      random_prefix <- paste0("A", rstring(64L), "_")

      data.table::setnames(
        distance_table,
        old = element_names,
        new = paste0(random_prefix, element_names)
      )
      
      element_names <- paste0(random_prefix, element_names)
    }
    
    # Create n x n table
    distance_table  <- data.table::dcast(
      distance_table,
      stats::as.formula(paste(element_names[1L], "~", element_names[2L])),
      value.var = "value"
    )
    
    # Ensure that the diagonal is formed by the pairwise distance of the same
    # feature, i.e. 0.0.
    data.table::setcolorder(
      distance_table,
      neworder = c(element_names[1L], as.character(distance_table[[element_names[1L]]]))
    )
    
    # Convert to dissimilarity matrix by first getting the labels
    # corresponding to each sample / feature, and then completing the
    # converted distance table.
    element_labels <- as.character(distance_table[[element_names[1L]]])
    distance_table <- as.matrix(distance_table[, mget(element_labels)])
    rownames(distance_table) <- element_labels
    distance_matrix <- stats::as.dist(distance_table)
    
    return(distance_matrix)
  }
)



## get_distance_matrix (noSimilarityTable) -------------------------------------
setMethod(
  "get_distance_matrix",
  signature(object = "noSimilarityTable"),
  function(object, ...) {
    
    return(NULL)
  }
)


# apply_cluster_method methods -------------------------------------------------

## apply_cluster_method (clusterMethod) ----------------------------------------
setMethod(
  "apply_cluster_method",
  signature(object = "clusterMethod"),
  function(object, ...) {
    # Generic method where no clustering object is generated.
    return(object)
  }
)




## apply_cluster_method (hclust) -----------------------------------------------
setMethod(
  "apply_cluster_method",
  signature(object = "clusterMethodHClust"),
  function(object, ...) {
    
    # Get distance matrix.
    distance_matrix <- do.call(
      get_distance_matrix,
      args = c(
        list("object" = object),
        list(...)
      )
    )
    
    # Skip if the distance matrix is NULL.
    if (is.null(distance_matrix)) return(object)
    
    # Convert general linkage names to stats::hclust linkage names.
    linkage_method <- object@linkage_method
    if (object@linkage_method == "ward")          {
      linkage_method <- "ward.D2"
      
    } else if (linkage_method == "weighted") {
      linkage_method <- "mcquitty"
    }
    
    if (is_package_installed("fastcluster")) {
      object@object <- fastcluster::hclust(
        d = distance_matrix,
        method = linkage_method
      )
      
    } else {
      object@object <- stats::hclust(
        d = distance_matrix,
        method = linkage_method
      )
    }
    
    return(object)
  }
)



## apply_cluster_method (agnes) ------------------------------------------------
setMethod(
  "apply_cluster_method",
  signature(object = "clusterMethodAgnes"),
  function(object, ...) {
    
    # Get distance matrix.
    distance_matrix <- do.call(
      get_distance_matrix,
      args = c(
        list("object" = object),
        list(...)
      )
    )
    
    # Skip if the distance matrix is NULL.
    if (is.null(distance_matrix)) return(object)
    
    require_package(
      x = "cluster",
      purpose = "to cluster similar features together"
    )
    
    # Compute agglomerative hierarchical clustering of the data set
    object@object <- stats::as.hclust(
      cluster::agnes(
        x = distance_matrix,
        method = object@linkage_method,
        keep.diss = FALSE,
        keep.data = FALSE
      )
    )
    
    return(object)
  }
)



## apply_cluster_method (diana) ------------------------------------------------
setMethod(
  "apply_cluster_method",
  signature(object = "clusterMethodDiana"),
  function(object,  ...) {
    
    # Get distance matrix.
    distance_matrix <- do.call(
      get_distance_matrix,
      args = c(
        list("object" = object),
        list(...)
      )
    )
    
    # Skip if the distance matrix is NULL.
    if (is.null(distance_matrix)) return(object)
    
    require_package(
      x = "cluster",
      purpose = "to cluster similar features together"
    )
    
    # Compute DIvisive ANAlysis hierarchical clustering of the data set
    object@object <- stats::as.hclust(
      cluster::diana(
        x = distance_matrix,
        keep.diss = FALSE,
        keep.data = FALSE
      )
    )
    
    return(object)
  }
)



# create_clusters methods ------------------------------------------------------

## create_clusters (generic hierarchical) --------------------------------------
setMethod(
  "create_clusters",
  signature(object = "clusterMethodHierarchical"),
  function(
    object,
    as_cluster_objects = TRUE,
    ...
  ) {
    
    cluster_table <- NULL
    if (object@cluster_cut_method == "silhouette") {
      # Cluster by silhouette.
      cluster_table <- .cluster_by_silhouette(object = object)
      
      # Attempt to create a cluster table if the cluster table returned empty.
      # Attempt to set singular clusters instead.
      if (is_empty(cluster_table)) {
        # Add names. Each cluster is singular.
        cluster_table <- .cluster_by_generic(object = object)
      }
      
    } else if (object@cluster_cut_method == "fixed_cut") {
      # Cluster by cutting the tree at a fixed height.
      cluster_table <- .cluster_by_fixed_cut(object = object)
      
      # Attempt to create a cluster table if the cluster table returned empty.
      # Attempt to set singular clusters instead.
      if (is_empty(cluster_table)) {
        # Add names. Each cluster is singular.
        cluster_table <- .cluster_by_generic(object = object)
      }
      
    } else if (object@cluster_cut_method == "none") {
      # Add names. Each cluster is singular.
      cluster_table <- .cluster_by_generic(object = object)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "create_clusters,clusterMethodHierarchical: encountered an unknown cluster cut method: ",
        object@cluster_method
      ))
    }
    
    # Determine how to return the results.
    if (as_cluster_objects) {
      return(.convert_cluster_table_to_cluster_objects(
        cluster_table = cluster_table,
        representation_method = object@representation_method
      ))
      
    } else {
      return(cluster_table)
    }
  }
)



## create_clusters (hclust) ----------------------------------------------------
setMethod(
  "create_clusters",
  signature(object = "clusterMethodHClust"),
  function(
    object,
    as_cluster_objects = TRUE,
    ...
  ) {
    
    if (object@cluster_cut_method %in% c("silhouette", "fixed_cut")) {
      # Silhouette and fixed cut are implemented for the parent class
      # (clusterMethodHierarchical).
      return(callNextMethod())
      
    } else if (object@cluster_cut_method == "dynamic_cut") {
      # Cluster by cutting the tree at a fixed height.
      cluster_table <- .cluster_by_dynamic_cut(object = object)
      
      # Attempt to create a cluster table if the cluster table returned empty.
      # Attempt to set singular clusters instead.
      if (is_empty(cluster_table)) {
        # Add names. Each cluster is singular.
        cluster_table <- .cluster_by_generic(object = object)
      }
      
    } else if (object@cluster_cut_method == "none") {
      # Add names. Each cluster is singular.
      cluster_table <- .cluster_by_generic(object = object)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "create_clusters,clusterMethodHClust: encountered an unknown cluster cut method: ",
        object@cluster_method
      ))
    }
    
    # Determine how to return the results.
    if (as_cluster_objects) {
      return(.convert_cluster_table_to_cluster_objects(
        cluster_table = cluster_table,
        representation_method = object@representation_method
      ))
      
    } else {
      return(cluster_table)
    }
  }
)



## create_clusters (PAM) -------------------------------------------------------
setMethod(
  "create_clusters",
  signature(object = "clusterMethodPAM"),
  function(
    object,
    as_cluster_objects = TRUE,
    ...
  ) {
    
    cluster_table <- NULL
    if (object@cluster_cut_method == "silhouette") {
      # Cluster by silhouette.
      cluster_table <- .cluster_by_silhouette(object = object)
      
      # Attempt to create a cluster table if the cluster table returned empty.
      # Attempt to set singular clusters instead.
      if (is_empty(cluster_table)) {
        # Add names. Each cluster is singular.
        cluster_table <- .cluster_by_generic(object = object)
      }
      
    } else if (object@cluster_cut_method == "none") {
      # Add names. Each cluster is singular.
      cluster_table <- .cluster_by_generic(object = object)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "create_clusters,clusterMethodPAM: encountered an unknown cluster cut method: ",
        object@cluster_method
      ))
    }
    
    # Determine how to return the results.
    if (as_cluster_objects) {
      return(.convert_cluster_table_to_cluster_objects(
        cluster_table = cluster_table,
        representation_method = object@representation_method
      ))
      
    } else {
      return(cluster_table)
    }
  }
)



## create_clusters (none) ------------------------------------------------------
setMethod(
  "create_clusters",
  signature(object = "clusterMethodNone"),
  function(
    object,
    as_cluster_objects = TRUE,
    ...
  ) {
    
    # Add names. Each cluster is singular.
    cluster_table <- .cluster_by_generic(object = object)
    
    # Determine how to return the results.
    if (as_cluster_objects) {
      return(.convert_cluster_table_to_cluster_objects(
        cluster_table = cluster_table,
        representation_method = "none"
      ))
      
    } else {
      return(cluster_table)
    }
  }
)



# .cluster_by_generic methods --------------------------------------------------

## .cluster_by_generic (generic hierarchical) -----------------------------------
setMethod(
  ".cluster_by_generic",
  signature(object = "clusterMethodHierarchical"),
  function(object, ...) {
    
    # Attempt to create the dendrogram.
    if (is.null(object@object)) object <- apply_cluster_method(object)
    
    # Find element names.
    element_names <- get_similarity_names(object@similarity_table)
    
    if (is.null(object@object)) {
      # A dendrogram wasn't formed, and we use the element names
      # instead.
      if (is.null(element_names)) {
        return(NULL)
        
      } else {
        # Set 
        return(data.table::data.table(
          "name" = element_names,
          "cluster_id" = seq_along(element_names),
          "label_order" = seq_along(element_names)
        ))
      }
      
    } else {
      return(data.table::data.table(
        "name" = object@object$labels[object@object$order],
        "cluster_id" = seq_along(object@object$labels),
        "label_order" = seq_along(object@object$labels)
      ))
    }
  }
)



## .cluster_by_generic (generic) -----------------------------------------------
setMethod(
  ".cluster_by_generic",
  signature(object = "clusterMethod"),
  function(object, ...) {
    
    # Find element names.
    element_names <- get_similarity_names(object@similarity_table)
    
    # Check that any element names are present.
    if (is.null(element_names)) return(NULL)
    
    return(data.table::data.table(
      "name" = element_names,
      "cluster_id" = seq_along(element_names),
      "label_order" = seq_along(element_names)
    ))
  }
)



# .cluster_by_silhouette methods -----------------------------------------------

## .cluster_by_silhouette (PAM) ------------------------------------------------
setMethod(
  ".cluster_by_silhouette",
  signature(object = "clusterMethodPAM"),
  function(object, ...) {
    
    # Get distance matrix.
    distance_matrix <- do.call(
      get_distance_matrix,
      args = c(
        list("object" = object),
        list(...)
      )
    )
    
    # Skip if the distance matrix is NULL.
    if (is.null(distance_matrix)) return(NULL)
    
    require_package(
      x = "cluster",
      purpose = "to cluster similar features together"
    )
    
    # Determine optimal numbers of clusters based on silhouette
    n_clusters <- .optimise_cluster_silhouette(
      object = object,
      distance_matrix = distance_matrix
    )
    
    # PAM clustering doesn't like it when you n_clusters is equal to the number
    # of features.
    if (n_clusters == length(get_similarity_names(object@similarity_table))) return(NULL)
    
    # Create clustering.
    cluster_object <- cluster::pam(
      x = distance_matrix,
      k = n_clusters,
      keep.diss = FALSE,
      keep.data = FALSE
    )
    
    return(data.table::data.table(
      "name" = names(cluster_object$clustering),
      "cluster_id" = cluster_object$clustering,
      "label_order" = seq_along(cluster_object$clustering)
    ))
  }
)



## .cluster_by_silhouette (generic hierarchical) -------------------------------
setMethod(
  ".cluster_by_silhouette",
  signature(object = "clusterMethodHierarchical"),
  function(object, ...) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    .NATURAL <- NULL
    
    # Get distance matrix.
    distance_matrix <- do.call(
      get_distance_matrix,
      args = c(
        list("object" = object),
        list(...)
      )
    )
    
    # Skip if the distance matrix is NULL.
    if (is.null(distance_matrix)) return(NULL)
    
    # Attempt to create the dendrogram.
    if (is.null(object@object)) {
      object <- apply_cluster_method(object)
    }
    
    # Check if a dendrogram could be created.
    if (is.null(object@object)) return(NULL)
    
    require_package(
      x = "cluster",
      purpose = "to cluster similar features together by silhouette"
    )
    
    # Determine optimal numbers of clusters based on silhouette
    n_clusters <- .optimise_cluster_silhouette(
      object = object,
      distance_matrix = distance_matrix
    )
    
    # PAM clustering doesn't like it when you n_clusters is equal to the
    # number of features.
    if (n_clusters == length(get_similarity_names(object@similarity_table))) return(NULL)
    
    # Cut the tree for the optimal number of clusters
    cluster_object <- stats::cutree(
      tree = stats::as.hclust(object@object),
      k = n_clusters
    )
    
    # Set initial cluster table.
    cluster_table <- data.table::data.table(
      "name" = names(cluster_object),
      "cluster_id" = cluster_object
    )
    
    # Get an ordering table.
    order_table <- data.table::data.table(
      "name" = object@object$labels[object@object$order],
      "label_order" = seq_along(object@object$labels)
    )
    
    # Insert label order into the cluster table.
    cluster_table <- cluster_table[order_table, on = .NATURAL]
    
    return(cluster_table)
  }
)



# .cluster_by_fixed_cut methods ------------------------------------------------

## .cluster_by_fixed_cut (generic hierarchical) --------------------------------
setMethod(
  ".cluster_by_fixed_cut",
  signature(object = "clusterMethodHierarchical"),
  function(object, ...) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    .NATURAL <- NULL
    
    # Attempt to create the dendrogram.
    if (is.null(object@object)) object <- apply_cluster_method(object)
    
    # Check if a dendrogram could be created.
    if (is.null(object@object)) return(NULL)
    
    # Compute the height at which the dendrogram should be cut.
    cut_height <- convert_similarity_to_distance(
      x = object@similarity_threshold,
      similarity_metric = object@similarity_metric
    )
    
    # Cut the dendrogram at the given height.
    cluster_object <- stats::cutree(
      tree = stats::as.hclust(object@object),
      h = cut_height
    )
    
    # Set initial cluster table.
    cluster_table <- data.table::data.table(
      "name" = names(cluster_object),
      "cluster_id" = cluster_object
    )
    
    # Get an ordering table.
    order_table <- data.table::data.table(
      "name" = object@object$labels[object@object$order],
      "label_order" = seq_along(object@object$labels)
    )
    
    # Insert label order into the cluster table.
    cluster_table <- cluster_table[order_table, on = .NATURAL]
    
    return(cluster_table)
  }
)

# .cluster_by_dynamic_cut methods ----------------------------------------------

## .cluster_by_dynamic_cut (hclust) --------------------------------------------
setMethod(
  ".cluster_by_dynamic_cut",
  signature(object = "clusterMethodHClust"),
  function(object, ...) {
    
    # Attempt to create the dendrogram.
    if (is.null(object@object)) object <- apply_cluster_method(object)
    
    # Check if a dendrogram could be created.
    if (is.null(object@object)) return(NULL)
    
    require_package(
      x = "dynamicTreeCut",
      purpose = "to cluster similar features together through dynamic dendrogram cutting"
    )
    
    # Compute the height at which the dendrogram should be cut anyway.
    cut_height <- convert_similarity_to_distance(
      x = object@similarity_threshold,
      similarity_metric = object@similarity_metric
    )
    
    if (length(get_similarity_names(object@similarity_table)) == 2L) {
      # For two features, dynamicTreeCut seems to ignore maxTreeHeight.
      if (
        convert_similarity_to_distance(
          x = object@similarity_table@data$value,
          similarity_metric = object@similarity_metric
        ) <= cut_height
      ) {
        cluster_ids <- c(1L, 1L)
        
      } else {
        cluster_ids <- c(1L, 2L)
      }
      
    } else {
      # From Langfelder P, Zhang B, Horvath S (2007) Defining clusters from a
      # hierarchical cluster tree: the Dynamic Tree Cut package for R.
      # Bioinformatics 2008 24(5):719-720
      cluster_ids <- tryCatch(
        dynamicTreeCut::cutreeDynamicTree(
          dendro = object@object,
          maxTreeHeight = cut_height,
          deepSplit = TRUE,
          minModuleSize = 1L
        ),
        error = identity
      )
      
      # Check that dynamic cutting does not produce an error.
      if (inherits(cluster_ids, "error")) return(.cluster_by_fixed_cut(object, ...))
      
      # Order the cluster identifiers correctly.
      cluster_ids <- cluster_ids[object@object$order]
    }
    
    # Create a clustering table.
    cluster_table <- data.table::data.table(
      "name" = object@object$labels[object@object$order],
      "cluster_id" = cluster_ids,
      "label_order" = seq_along(object@object$labels)
    )
    
    return(cluster_table)
  }
)



create_cluster_method_object <- function(
    cluster_method,
    data_type,
    cluster_linkage = NULL,
    cluster_cut_method = NULL,
    cluster_similarity_threshold = NULL,
    cluster_similarity_metric = NULL,
    cluster_representation_method = NULL
) {
  
  # Check that method is applicable.
  .check_parameter_value_is_valid(
    x = cluster_method,
    var_name = ifelse(
      data_type == "cluster",
      "cluster_method",
      paste0(data_type, "_cluster_method")
    ),
    values = .get_available_cluster_methods()
  )
  
  # Check that data_type is valid.
  .check_parameter_value_is_valid(
    x = data_type,
    var_name = "data_type",
    values = c("feature", "cluster", "sample")
  )
  
  if (cluster_method == "none") {
    object <- methods::new("clusterMethodNone")
    
  } else if (cluster_method == "pam") {
    object <- methods::new("clusterMethodPAM")
    
  } else if (cluster_method == "hclust") {
    object <- methods::new("clusterMethodHClust")
    
  } else if (cluster_method == "agnes") {
    object <- methods::new("clusterMethodAgnes")
    
  } else if (cluster_method == "diana") {
    object <- methods::new("clusterMethodDiana")
    
  } else {
    ..error_reached_unreachable_code(paste0(
      "create_cluster_method_object: encountered an unknown cluster method: ",
      cluster_method
    ))
  }
  
  # Cluster method and data type are always set.
  object@method <- cluster_method
  object@data_type <- data_type
  
  # Set cluster object method parameters as required.
  object <- set_object_parameters(
    object = object,
    cluster_linkage = cluster_linkage,
    cluster_cut_method = cluster_cut_method,
    cluster_similarity_threshold = cluster_similarity_threshold,
    cluster_similarity_metric = cluster_similarity_metric,
    cluster_representation_method = cluster_representation_method
  )
  
  return(object)
}



.convert_cluster_table_to_cluster_objects <- function(
    cluster_table,
    representation_method
) {
  
  # Check that the cluster table is not empty.
  if (is_empty(cluster_table)) return(NULL)
  
  # Check that the expect columns are present.
  if (!(all(c("name", "cluster_id") %in% colnames(cluster_table)))) {
    ..error_reached_unreachable_code(paste0(
      ".convert_cluster_table_to_cluster_objects: expected name and ",
      "cluster_id columns were not found."
    ))
  }
  
  return(lapply(
    split(cluster_table, by = "cluster_id"),
    ..convert_cluster_table_to_cluster_objects,
    representation_method = representation_method
  ))
}



..convert_cluster_table_to_cluster_objects <- function(
    cluster_table,
    representation_method
) {
  
  if (nrow(cluster_table) == 1L) {
    # Create singular cluster object.
    object <- methods::new(
      "singularClusteringObject",
      cluster_features = cluster_table$name
    )
    
  } else {
    # Create cluster object with multiple features or instances.
    object <- methods::new(
      "clusteringObject",
      cluster_features = cluster_table$name,
      representation_method = representation_method
    )
  }
  
  return(object)
}



.optimise_cluster_silhouette <- function(
    object,
    distance_matrix,
    tol = 0.01
) {
  
  # Determine the number of features.
  n_features <- length(get_similarity_names(object@similarity_table))
  
  highly_similar_distance <- convert_similarity_to_distance(
    x = get_high_similarity_threshold(similarity_metric = object@similarity_metric),
    similarity_metric = object@similarity_metric
  )
  
  # Check problematic values.
  if (n_features == 1L) {
    return(1L)
    
  } else if (n_features == 2L) {
    
    if (all(distance_matrix <= highly_similar_distance)) {
      # Zero distance can be safely imputed as being identical.
      return(1L)
      
    } else {
      # Otherwise interpret as singular clusters.
      return(n_features)
    }
  }
  
  # If all elements have distance 0, return 1 cluster.
  if (all(distance_matrix <= highly_similar_distance)) return(1L)
  
  # The optimiser doesn't like a singular interval, which occurs for n_features
  # == 3.
  if (n_features == 3L) return(2L)
  
  # Set k to test.
  k_test <- seq(2L, n_features - 1L)
  silhouette_score <- numeric(n_features - 2L)
  silhouette_gradient <- numeric(max(c(0L, n_features - 4L)))
  
  for (k in k_test) {
    # Compute average silhouette.
    silhouette_score[k - 1L] <- ..optimise_cluster_silhouette(
      k = k,
      distance_matrix = distance_matrix,
      object = object
    )
    
    # Compute gradient as a symmetric difference coefficient.
    if (k > 3L) {
      silhouette_gradient[k - 3L] <- 0.5 * (silhouette_score[k - 1L] - silhouette_score[k - 3L])
      
      # Stop if the gradient is 0 or negative 3 times.
      if (sum(silhouette_gradient[seq(1L, k - 3L)] <= 0.0) > 2L) break
    }
  }
  
  # Set the optimal number of clusters.
  k_optimal <- head(k_test[silhouette_score > max(silhouette_score) - tol], n = 1L)
  max_silhouette <- silhouette_score[k_test == k_optimal]
  
  # Determine if the silhoutte indicates reasonable structure (> 0.50), see
  # Kaufman and Rousseeuw: Finding groups in data.
  if (max_silhouette < 0.50) k_optimal <- n_features
  
  return(k_optimal)
}



..optimise_cluster_silhouette <- function(
    k,
    distance_matrix,
    object
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  sil_width <- cluster_size <- NULL
  
  # Round k to integer value.
  k <- round(k, digits = 0L)
  
  if (is(object, "clusterMethodPAM")) {
    # Generate partioning around medoids cluster
    cluster_object <- cluster::pam(
      x = distance_matrix,
      k = k,
      keep.diss = FALSE,
      keep.data = FALSE
    )
    
    # Extract silhouette table
    silhouette_table <- data.table::as.data.table(
      cluster_object$silinfo$widths,
      keep.rownames = FALSE
    )
    
    # Compute the size and average silhouette in each cluster
    silhouette_table <- silhouette_table[
      ,
      list(
        "average_cluster_silhouette" = mean(sil_width),
        "cluster_size" = .N
      ),
      by = "cluster"
    ]
    
    # Maintain only non-singular clusters
    silhouette_table <- silhouette_table[cluster_size > 1L]
    
    # Return average silhouette in the formed non-singular clusters.
    if (!is_empty(silhouette_table)) {
      return(mean(silhouette_table$average_cluster_silhouette))
      
    } else {
      return(0.0)
    }
    
  } else if (is(object, "clusterMethodHierarchical")) {
    
    # Compute silhouette.
    silhouette_matrix <- cluster::silhouette(
      x = stats::cutree(tree = stats::as.hclust(object@object), k = k),
      dist = distance_matrix
    )
    
    # Parse to matrix by changing the class. 
    class(silhouette_matrix) <- "matrix"
    
    # Extract silhouette table
    silhouette_table <- data.table::as.data.table(
      silhouette_matrix,
      keep.rownames = FALSE
    )
    
    # Compute the size and average silhouette in each cluster
    silhouette_table <- silhouette_table[
      ,
      list(
        "average_cluster_silhouette" = mean(sil_width),
        "cluster_size" = .N
      ),
      by = "cluster"
    ]
    
    # Maintain only non-singular clusters
    silhouette_table <- silhouette_table[cluster_size > 1L]
    
    # Return average silhouette in the formed non-singular clusters.
    if (!is_empty(silhouette_table)) {
      return(mean(silhouette_table$average_cluster_silhouette))
      
    } else {
      return(0.0)
    }
    
  } else {
    ..error_reached_unreachable_code(
      "..optimise_cluster_silhouette: unknown clustering object encountered."
    )
  }
}



.check_cluster_parameters <- function(
    cluster_method,
    cluster_linkage = NULL,
    cluster_cut_method = NULL,
    cluster_similarity_threshold = NULL,
    cluster_similarity_metric = NULL,
    cluster_representation_method = NULL,
    data_type = "cluster",
    test_required_packages = TRUE,
    message_type = "error"
) {
  
  # Perform checks by creating the relevant object. This checks whether the
  # applicable parameters are set.
  object <- create_cluster_method_object(
    cluster_method,
    data_type = data_type,
    cluster_linkage = cluster_linkage,
    cluster_cut_method = cluster_cut_method,
    cluster_similarity_threshold = cluster_similarity_threshold,
    cluster_similarity_metric = cluster_similarity_metric,
    cluster_representation_method = cluster_representation_method
  )
  
  if (test_required_packages) {
    # Check whether the cluster package has been installed.
    if (object@method %in% c("pam", "agnes", "diana")) {
      require_package(
        x = "cluster",
        purpose = "to compute similarity between features or instances",
        message_type = message_type
      )
    }
    
    # Check whether the dynamicTreeCut package has been installed.
    if (.hasSlot(object, "cluster_cut_method")) {
      if (object@cluster_cut_method == "dynamic_cut") {
        require_package(
          x = "dynamicTreeCut",
          purpose = "to cut dendrograms dynamically",
          message_type = message_type
        )
      }
    }
    
    # Check whether the nnet package has been installed.
    if (.hasSlot(object, "similarity_metric")) {
      if (object@similarity_metric %in% c("mcfadden_r2", "cox_snell_r2", "nagelkerke_r2")) {
        require_package(
          x = "nnet",
          purpose = paste0(
            "to compute log-likelihood pseudo R2 similarity using the ",
            object@similarity_metric, " metric"
          ),
          message_type = message_type
        )
      }
    }
  }
  
  return(invisible(TRUE))
}



.get_cluster_data_type_element_name <- function(data_type) {
  # Determine whether the similarity table is for features (columns)
  # or samples (rows).
  element_1 <- ifelse(data_type %in% c("feature", "cluster"), "feature_name_1", "sample_1")
  element_2 <- ifelse(data_type %in% c("feature", "cluster"), "feature_name_2", "sample_2")
  
  return(c(element_1, element_2))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# promote_vimp_method ----------------------------------------------------------
setMethod(
  "promote_vimp_method",
  signature(object = "familiarVimpMethod"),
  function(object) {
    # Extract vimp_method.
    method <- object@vimp_method

    if (method %in% .get_available_concordance_vimp_method()) {
      # Concordance-based methods.
      object <- methods::new("familiarConcordanceVimp", object)
    } else if (method %in% .get_available_univariate_mutual_information_vimp_method()) {
      # Mutual information maximisation.
      object <- methods::new("familiarUnivariateMutualInfoVimp", object)
    } else if (method %in% .get_available_multivariate_mutual_information_vimp_method()) {
      # Multivariate information methods.
      object <- methods::new("familiarMultivariateMutualInfoVimp", object)
    } else if (method %in% .get_available_multivariate_mutual_information_test_vimp_method()) {
      # Multivariate test method returning single features.
      object <- methods::new("familiarMultivariateInfoVimpTestSingle", object)
    } else if (method %in% .get_available_correlation_vimp_methods()) {
      # Correlation-based methods.
      object <- methods::new("familiarCorrelationVimp", object)
    } else if (method %in% .get_available_corelearn_gini_vimp_method()) {
      # Gini measure.
      object <- methods::new("familiarCoreLearnGiniVimp", object)
    } else if (method %in% .get_available_corelearn_mdl_vimp_method()) {
      # MDL variable importance method.
      object <- methods::new("familiarCoreLearnMDLVimp", object)
    } else if (method %in% .get_available_corelearn_relieff_exp_rank_vimp_method()) {
      # ReliefF with exponentially decreasing rank.
      object <- methods::new("familiarCoreLearnRelieffExpRankVimp", object)
    } else if (method %in% .get_available_corelearn_gain_ratio_vimp_method()) {
      # Gain ratio measure.
      object <- methods::new("familiarCoreLearnGainRatioVimp", object)
    } else if (method %in% .get_available_univariate_regression_vimp_methods()) {
      # Univariate regression-based methods.
      object <- methods::new("familiarUnivariateRegressionVimp", object)
    } else if (method %in% .get_available_multivariate_regression_vimp_methods()) {
      # Multivariate regression-based methods.
      object <- methods::new("familiarMultivariateRegressionVimp", object)
    } else if (method %in% .get_available_glmnet_ridge_vimp_methods()) {
      # Ridge penalised regression model-based methods.

      # Create a familiarModel and promote to the right class.
      object <- methods::new(
        "familiarModel",
        vimp_method = "none",
        vimp_aggregation_method = object@vimp_aggregation_method,
        vimp_rank_threshold = object@vimp_rank_threshold,
        learner = method,
        outcome_type = object@outcome_type,
        hyperparameters = object@hyperparameters,
        outcome_info = object@outcome_info,
        feature_info = object@feature_info,
        required_features = object@required_features,
        run_table = object@run_table,
        project_id = object@project_id
      )

      # Promote to the correct subclass.
      object <- promote_learner(object)
      
    } else if (method %in% .get_available_glmnet_lasso_vimp_methods()) {
      # Lasso penalised regression model-based methods.

      # Create a familiarModel and promote to the right class.
      object <- methods::new(
        "familiarModel",
        vimp_method = "none",
        vimp_aggregation_method = object@vimp_aggregation_method,
        vimp_rank_threshold = object@vimp_rank_threshold,
        learner = method,
        outcome_type = object@outcome_type,
        hyperparameters = object@hyperparameters,
        outcome_info = object@outcome_info,
        feature_info = object@feature_info,
        required_features = object@required_features,
        run_table = object@run_table,
        project_id = object@project_id
      )

      # Promote to the correct subclass.
      object <- promote_learner(object)
      
    } else if (method %in% .get_available_glmnet_elastic_net_vimp_methods()) {
      # Elastic net penalised regression model-based methods.

      # Create a familiarModel and promote to the right class.
      object <- methods::new("familiarModel",
        vimp_method = "none",
        vimp_aggregation_method = object@vimp_aggregation_method,
        vimp_rank_threshold = object@vimp_rank_threshold,
        learner = method,
        outcome_type = object@outcome_type,
        hyperparameters = object@hyperparameters,
        outcome_info = object@outcome_info,
        feature_info = object@feature_info,
        required_features = object@required_features,
        run_table = object@run_table,
        project_id = object@project_id
      )

      # Promote to the correct subclass.
      object <- promote_learner(object)
      
    } else if (
      method %in% .get_available_rfsrc_vimp_methods() ||
      method %in% .get_available_rfsrc_default_vimp_methods()
    ) {
      # Random forest variable importance methods.

      # Create a familiarModel and promote to the right class.
      object <- methods::new("familiarModel",
        vimp_method = "none",
        vimp_aggregation_method = object@vimp_aggregation_method,
        vimp_rank_threshold = object@vimp_rank_threshold,
        learner = ifelse(method %in% .get_available_rfsrc_vimp_methods(),
          "random_forest_rfsrc",
          "random_forest_rfsrc_default"
        ),
        outcome_type = object@outcome_type,
        hyperparameters = object@hyperparameters,
        outcome_info = object@outcome_info,
        feature_info = object@feature_info,
        required_features = object@required_features,
        run_table = object@run_table,
        project_id = object@project_id
      )

      # Promote to the correct subclass.
      object <- promote_learner(object)

      # Set the variable importance parameters for the method.
      object <- ..set_vimp_parameters(object, method = method)
      
    } else if (
      method %in% .get_available_ranger_vimp_methods() ||
      method %in% .get_available_ranger_default_vimp_methods()
    ) {
      # Ranger random forest variable importance methods.

      # Create a familiarModel and promote to the right class.
      object <- methods::new("familiarModel",
        vimp_method = "none",
        vimp_aggregation_method = object@vimp_aggregation_method,
        vimp_rank_threshold = object@vimp_rank_threshold,
        learner = ifelse(
          method %in% .get_available_ranger_vimp_methods(),
          "random_forest_ranger",
          "random_forest_ranger_default"
        ),
        outcome_type = object@outcome_type,
        hyperparameters = object@hyperparameters,
        outcome_info = object@outcome_info,
        feature_info = object@feature_info,
        required_features = object@required_features,
        run_table = object@run_table,
        project_id = object@project_id
      )

      # Promote to the correct subclass.
      object <- promote_learner(object)

      # Set the variable importance parameters for the method.
      object <- ..set_vimp_parameters(object, method = method)
      
    } else if (method %in% .get_available_random_vimp_methods()) {
      # Random variable importances.
      object <- methods::new("familiarRandomVimp", object)
      
    } else if (method %in% .get_available_none_vimp_methods()) {
      # No variable importance: all features are equally important.
      object <- methods::new("familiarNoneVimp", object)
      
    } else if (method %in% .get_available_signature_only_vimp_methods()) {
      # Signature only methods.
      object <- methods::new("familiarSignatureVimp", object)
      
    } else if (method %in% .get_available_no_features_vimp_methods()) {
      # No variable importance methods - no features are selected, leading to
      # naive models.
      object <- methods::new("familiarNoFeaturesVimp", object)
    }

    return(object)
  }
)



.get_vimp_hyperparameters <- function(
    data, 
    method, 
    outcome_type, 
    names_only = FALSE
) {
  # Get the outcome type from the data object, if available
  if (!is.null(data)) outcome_type <- data@outcome_type

  # Create familiarModel
  vimp_method_object <- methods::new(
    "familiarVimpMethod",
    vimp_method = method,
    outcome_type = outcome_type
  )

  # Set up the specific model
  vimp_method_object <- promote_vimp_method(vimp_method_object)

  # Variable importance hyperparameters
  model_hyperparameters <- get_default_hyperparameters(vimp_method_object, data = data)

  # Extract names from parameter list
  if (names_only == TRUE) {
    model_hyperparameters <- names(model_hyperparameters)
  }

  # Return hyperparameter list, or names of parameters
  return(model_hyperparameters)
}



.check_vimp_outcome_type <- function(
    method, 
    outcome_type, 
    as_flag = FALSE
) {
  # Create familiarModel
  vimp_method_object <- methods::new(
    "familiarVimpMethod",
    vimp_method = method,
    outcome_type = outcome_type
  )

  # Set up the specific model
  vimp_method_object <- promote_vimp_method(vimp_method_object)

  # Check validity.
  vimp_method_available <- is_available(vimp_method_object)

  if (as_flag) return(vimp_method_available)

  # Check if the vimp method or familiar model has been successfully promoted.
  if (
    !is_subclass(class(vimp_method_object)[1L], "familiarVimpMethod") &&
    !is_subclass(class(vimp_method_object)[1L], "familiarModel")
  ) {
    ..error(paste0(
      method, " is not a valid variable importance method. ",
      "Please check the vignette for available methods."
    ))
  }

  if (!vimp_method_available) {
    ..error(paste0(method, " is not available for \"", outcome_type, "\" outcomes."))
  }

  # Check that the required package can be loaded.
  require_package(
    x = vimp_method_object,
    purpose = paste0("to assess variable importance using the ", method, " method"),
    message_type = "backend_error"
  )
}



# add_package_version ----------------------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "familiarVimpMethod"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)

testthat::skip_on_cran()
testthat::skip_on_ci()

# Perform an integrated test. Note that learner is set automatically.
familiar:::integrated_test(
  dynamic_model_loading = TRUE,
  experimental_design = "fs+mb",
  vimp_method = "none",
  cluster_method = "none",
  imputation_method = "simple",
  parallel = FALSE,
  estimation_type = "point"
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R



# promote_learner (hyperparameter learner) -------------------------------------
setMethod(
  "promote_learner",
  signature(object = "familiarHyperparameterLearner"),
  function(object) {
    learner <- object@learner

    if (learner %in% .get_available_ranger_hyperparameter_learners()) {
      # Ranger random forest
      object <- methods::new("familiarHyperparameterLearnerRanger", object)
      
    } else if (learner %in% .get_available_lagp_hyperparameter_learners()) {
      # Localised approximate Gaussian process.
      object <- methods::new("familiarHyperparameterLearnerLAGP", object)
      
    } else if (learner %in% .get_available_bart_hyperparameter_learners()) {
      # Bayesian additive regression trees
      object <- methods::new("familiarHyperparameterLearnerBART", object)
      
    } else if (learner %in% .get_available_random_search_hyperparameter_learners()) {
      # Random search
      object <- methods::new("familiarHyperparameterRandomSearch", object)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "promote_learner,familiarHyperparameterLearner: encountered ",
        "unknown hyperparameter learner: ", learner
      ))
    }

    # Add package version.
    object <- add_package_version(object = object)

    # Return promoted object.
    return(object)
  }
)



# .train (hyperparameter learner) ----------------------------------------------
setMethod(
  ".train",
  signature(
    object = "familiarHyperparameterLearner",
    data = "data.table"
  ),
  function(
    object,
    data,
    parameter_data = NULL,
    ...
  ) {
    # Train method for training hyperparameter models

    # Suppress NOTES due to non-standard evaluation in data.table
    optimisation_score <- NULL

    # Check if the class of object is a subclass of
    # familiarHyperparameterLearner.
    if (!is_subclass(class(object)[1L], "familiarHyperparameterLearner")) {
      object <- promote_learner(object)
    }

    # Check whether data contains the optimisation_score column.
    if (!"optimisation_score" %in% colnames(data)) {
      data <- .compute_hyperparameter_optimisation_score(
        score_table = data,
        optimisation_function = object@optimisation_function
      )
    }

    # Merge data and parameter_data on param_id.
    if (!is.null(parameter_data)) {
      data <- merge(
        x = data,
        y = parameter_data,
        all.x = TRUE,
        all.y = FALSE,
        by = "param_id"
      )
    }

    # Replace NA entries with the minimum optimisation score.
    data[is.na(optimisation_score), optimisation_score := -1.0]

    # Get hyperparameter names from the target learner.
    default_hyperparameters <- .get_learner_hyperparameters(
      data = NULL,
      learner = object@target_learner,
      outcome_type = object@target_outcome_type,
      names_only = TRUE
    )

    # Set hyperparameters in data.
    target_hyperparameters <- intersect(colnames(data), default_hyperparameters)
    if (length(target_hyperparameters) == 0L) {
      ..error(
        paste0(
          "The provided hyperparameter data does not contain any of the ",
          "hyperparameters expected by the learner. The following hyperparameters ",
          "are used: ", paste_s(default_hyperparameters)
        ),
        error_class = "input_argument_error"
      )
    }
    
    object@target_hyperparameters <- target_hyperparameters

    if (has_bad_training_data(object = object, data = data)) {
      # Create a random search object in case training data are bad in
      # some way or form.
      object@learner <- "random_search"
      object <- methods::new("familiarHyperparameterRandomSearch", object)
    }

    # Train a new model based on data.
    object <- ..train(
      object = object,
      data = data,
      ...
    )

    # Add familiar version.
    object <- add_package_version(object)

    return(object)
  }
)


# .predict (hyperparameter learner) --------------------------------------------
setMethod(
  ".predict",
  signature(
    object = "familiarHyperparameterLearner",
    data = "data.table"
  ),
  function(
    object,
    data,
    type = "default",
    percentile = NULL,
    ...
  ) {
    # Predict method for hyperparameter models. The type argument can be default
    # (reports the model estimate), sd (reports the model estimate + standard
    # deviation), percentile, in which case the model estimate and the requested
    # percentile are returned, and raw, which returns the data in a wide format.
    # Not all learners support the same types. Use the get_prediction_type
    # method to figure out which type of output can be generated.

    # Dispatch to the ..predict methods of the child classes.
    prediction_table <- ..predict(
      object = object,
      data = data,
      type = type,
      percentile = percentile
    )

    return(prediction_table)
  }
)



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarHyperparameterLearner"),
  function(object, ...) {
    # Placeholder for hyperparameter learners.
    return(NULL)
  }
)



# show (hyperparameter learner) ------------------------------------------------
setMethod(
  "show",
  signature(object = "familiarHyperparameterLearner"),
  function(object) {
    if (!model_is_trained(object)) {
      # Describe the learner and the version of familiar.
      cat(paste0(
        "A ", object@learner, " model (class: ", class(object)[1L],
        ") for inferring hyperparameters of the ", object@target_learner, ". ",
        "This hyperparameter model could not successfully be trained (",
        .familiar_version_string(object), ").\n"
      ))
      
      return(invisible(NULL))
    }
    
    # Describe the learner and the version of familiar.
    message_str <- paste0(
      "A ", object@learner, " model (class: ", class(object)[1L],
      "; ", .familiar_version_string(object), ") ",
      "for inferring hyperparameters of the ", object@target_learner,
      " learner. "
    )
    
    # Describe the package(s), if any.
    if (!is.null(object@package)) {
      message_str <- c(
        message_str,
        paste0("This model was trained using "),
        paste_s(mapply(
          ..message_package_version,
          x = object@package,
          version = object@package_version
        )),
        ifelse(length(object@package) > 1L, " packages", " package")
      )
    }
    
    # Complete message and write.
    message_str <- paste0(c(message_str, ".\n"), collapse = "")
    cat(message_str)
    
    cat(paste0("\n--------------- Model details ---------------\n"))
    
    # Model details
    show(object@model)
    
    cat(paste0("---------------------------------------------\n"))
    
    # Details concerning hyperparameters.
    cat(paste0(
      "\nThe model was trained to infer the optimisation ",
      "score of the following hyperparameter set:\n",
      paste_s(object@target_hyperparameters), "\n"
    ))
    
    cat(paste0(
      "\nOptimisation scores were determined using the ",
      object@optimisation_function, " method, ",
      "based on assessment of model performance using the ",
      paste_s(object@optimisation_metric),
      ifelse(length(object@optimisation_metric) > 1L, " metrics", " metric"),
      ".\n"
    ))
    
    # Check package version.
    check_package_version(object)
  }
)



# require_package (hyperparameter learner) -------------------------------------
setMethod(
  "require_package",
  signature(x = "familiarHyperparameterLearner"),
  function(
    x,
    purpose = NULL, 
    message_type = "error",
    ...
  ) {
    # Skip if no package is required.
    if (is_empty(x@package)) return(invisible(TRUE))

    # Set standard purposes for common uses.
    if (!is.null(purpose)) {
      if (purpose %in% c("train", "predict", "show")) {
        purpose <- switch(purpose,
          "train" = "to train a model for inferring hyperparameter scores",
          "predict" = "to infer hyperparameter scores",
          "show" = "to capture output of the hyperparameter model"
        )
      }
    }

    return(invisible(.require_package(
      x = x@package,
      purpose = purpose,
      message_type = message_type
    )))
  }
)



# set_package_version (hyperparameter learner) ---------------------------------
setMethod(
  "set_package_version",
  signature(object = "familiarHyperparameterLearner"),
  function(object) {
    # Do not add package versions if there are no packages.
    if (is_empty(object@package)) return(object)

    # Obtain package versions.
    object@package_version <- sapply(
      object@package, 
      function(x) (as.character(utils::packageVersion(x)))
    )

    return(object)
  }
)



# check_package_version (hyperparameter learner) -------------------------------
setMethod(
  "check_package_version",
  signature(object = "familiarHyperparameterLearner"),
  function(object) {
    .check_package_version(
      name = object@package,
      version = object@package_version,
      when = "at model creation"
    )
  }
)



# model_is_trained (hyperparameter learner) ------------------------------------
setMethod(
  "model_is_trained",
  signature(object = "familiarHyperparameterLearner"),
  function(object) {
    # Check if a model was trained.
    if (is.null(object@model)) {
      # If no model is attached to the object, assume that no model was trained.
      return(FALSE)
      
    } else {
      # If the the model is not NULL, assume that it is present.
      return(TRUE)
    }
  }
)



# add_package_version (hyperparameter learner) ---------------------------------
setMethod(
  "add_package_version",
  signature(object = "familiarHyperparameterLearner"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)


# ..train (hyperparameter learner, data) ---------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarHyperparameterLearner",
    data = "ANY"
  ),
  function(object, data, ...) {
    # This method is basically a capture-all for when model training does not
    # succeed. Normally ..train calls refer to child classes. This method is
    # then only called when something goes wrong there.

    # Set a NULL model
    object@model <- NULL

    return(object)
  }
)


# ..predict (hyperparameter learner, data) -------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarHyperparameterLearner",
    data = "data.table"
  ),
  function(object, data, type = "default", ...) {
    # Like ..train, this method is a capture-all for when predictions fail.
    return(.get_placeholder_hyperparameter_prediction_table(
      object = object,
      data = data,
      type = type
    ))
  }
)


# has_bad_training_data --------------------------------------------------------
setMethod(
  "has_bad_training_data",
  signature(
    object = "familiarHyperparameterLearner",
    data = "data.table"
  ),
  function(object, data, ...) {
    # One cannot train without data or on a single sample.
    if (is_empty(data)) return(TRUE)

    # Get identifier columns in the data.
    id_columns <- intersect(c("param_id", "run_id"), colnames(data))
    if (length(id_columns) > 0L) {
      # Check that at least two unique entries are present.
      if (data.table::uniqueN(data, by = c(id_columns)) < 2L) return(TRUE)
      
    } else {
      # Check that at least two rows are present.
      if (nrow(data) < 2L) return(TRUE)
    }

    # Check if all data are non-singular.
    if (all(sapply(data[, mget(object@target_hyperparameters)], is_singular_data))) {
      return(TRUE)
    }

    return(FALSE)
  }
)



.check_hyperparameter_learner_available <- function(
    hyperparameter_learner,
    as_flag = FALSE
) {
  # Create familiarHyperparameterLearner object.
  fam_hyperparameter_model <- methods::new(
    "familiarHyperparameterLearner",
    learner = hyperparameter_learner
  )

  # Set up the specific novelty detector
  fam_hyperparameter_model <- promote_learner(fam_hyperparameter_model)

  # Check if the familiar model has been successfully promoted.
  if (!is_subclass(class(fam_hyperparameter_model)[1L], "familiarHyperparameterLearner")) {
    ..error(
      paste0(
        hyperparameter_learner, " is not a valid hyperparameter learner. ",
        "Please check the vignette for available hyperparameter learners."
      ),
      error_class = "input_argument_error"
    )
  }

  # Check that the required package can be loaded.
  require_package(
    x = fam_hyperparameter_model,
    purpose = "train",
    message_type = "backend_error"
  )
}



.get_placeholder_hyperparameter_prediction_table <- function(object, data, type = "default") {
  # Find the id columns.
  id_columns <- intersect(c("param_id", "run_id"), colnames(data))
  
  if (length(id_columns) > 0L) {
    # Create a placeholder by only keeping the identifier columns.
    prediction_table <- data.table::copy(data[, mget(id_columns)])
    
  } else {
    # Add a placeholder parameter identifier as scaffolding.
    prediction_table <- data.table::data.table(param_id = rep_len(NA_integer_, nrow(data)))
  }
  
  # Add placeholder columns.
  if (type == "default") {
    prediction_table[, "mu" := NA_real_]
    
  } else if (type == "sd") {
    prediction_table[, ":="(
      "mu" = NA_real_,
      "sigma" = NA_real_
    )]
    
  } else if (type == "percentile") {
    prediction_table[, "percentile" := NA_real_]
    
  } else if (type == "raw") {
    prediction_table[, "raw_1" := NA_real_]
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".get_placeholder_hyperparameter_prediction_table: ",
      "Encountered an unknown prediction type: ", type
    ))
  }
  
  return(prediction_table)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# plot_auc_roc_curve (generic) -------------------------------------------------

#' @title Plot the receiver operating characteristic curve.
#'
#' @description This method creates receiver operating characteristic curves
#'   based on data in a familiarCollection object.
#'
#' @param dir_path (*optional*) Path to the directory where the plots of
#'   receiver operating characteristic curves are saved to. Output is saved in
#'   the `performance` subdirectory. If `NULL` no figures are saved, but are
#'   returned instead.
#'
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_auc_data -object
#'
#' @details This function generates area under the ROC curve plots.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, `data_set` and
#'   `positive_class`. By default, the data is split by `vimp_method` and
#'   `learner`, with faceting by `data_set` and colouring by `positive_class`.
#'
#'   Bootstrap confidence intervals of the ROC curve (if present) can be shown
#'   using various styles set by `conf_int_style`:
#'
#'  * `ribbon` (default): confidence intervals are shown as a ribbon with an
#'   opacity of `conf_int_alpha` around the point estimate of the ROC curve.
#'
#'  * `step` (default): confidence intervals are shown as a step function around
#'   the point estimate of the ROC curve.
#'
#'  * `none`: confidence intervals are not shown. The point estimate of the ROC
#'   curve is shown as usual.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_auc_roc_curve
#' @md
#' @rdname plot_auc_roc_curve-methods
setGeneric(
  "plot_auc_roc_curve",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_auc_roc_curve")
  }
)

# plot_auc_roc_curve (general) -------------------------------------------------

#' @rdname plot_auc_roc_curve-methods
setMethod(
  "plot_auc_roc_curve",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "auc_data"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_auc_roc_curve,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)


# plot_auc_roc_curve (collection) ----------------------------------------------

#' @rdname plot_auc_roc_curve-methods
setMethod(
  "plot_auc_roc_curve",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(do.call(
      .plot_auc_curve,
      args = list(
        "object" = object,
        "curve_type" = "roc",
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_auc_precision_recall_curve (generic) ------------------------------------

#' @title Plot the precision-recall curve.
#'
#' @description This method creates precision-recall curves based on data in a
#'  familiarCollection object.
#'
#' @inheritParams plot_auc_roc_curve
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#'
#' @details This function generates area under the precision-recall curve plots.
#'
#'  Available splitting variables are: `vimp_method`, `learner`, `data_set` and
#'  `positive_class`. By default, the data is split by `vimp_method` and `learner`,
#'  with faceting by `data_set` and colouring by `positive_class`.
#'
#'  Available palettes for `discrete_palette` are those listed by
#'  `grDevices::palette.pals()` (requires R >= 4.0.0), `grDevices::hcl.pals()`
#'  (requires R >= 3.6.0) and `rainbow`, `heat.colors`, `terrain.colors`,
#'  `topo.colors` and `cm.colors`, which correspond to the palettes of the same
#'  name in `grDevices`. If not specified, a default palette based on palettes
#'  in Tableau are used. You may also specify your own palette by using colour
#'  names listed by `grDevices::colors()` or through hexadecimal RGB strings.
#'
#'  Bootstrap confidence intervals of the ROC curve (if present) can be shown
#'  using various styles set by `conf_int_style`:
#'
#'  * `ribbon` (default): confidence intervals are shown as a ribbon with an
#'  opacity of `conf_int_alpha` around the point estimate of the ROC curve.
#'
#'  * `step` (default): confidence intervals are shown as a step function around
#'  the point estimate of the ROC curve.
#'
#'  * `none`: confidence intervals are not shown. The point estimate of the ROC
#'  curve is shown as usual.
#'
#'  Labelling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'  be applied to the `familiarCollection` object to update labels, and order
#'  the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_auc_precision_recall_curve
#' @md
#' @rdname plot_auc_precision_recall_curve-methods
setGeneric(
  "plot_auc_precision_recall_curve",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_auc_precision_recall_curve")
  }
)



# plot_auc_precision_recall_curve (general) ------------------------------------

#' @rdname plot_auc_precision_recall_curve-methods
setMethod(
  "plot_auc_precision_recall_curve",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "auc_data"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_auc_precision_recall_curve,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_auc_precision_recall_curve (collection) ---------------------------------

#' @rdname plot_auc_precision_recall_curve-methods
setMethod(
  "plot_auc_precision_recall_curve",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(do.call(
      .plot_auc_curve,
      args = list(
        "object" = object,
        "curve_type" = "pr",
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "conf_int_style" = conf_int_style,
        "conf_int_alpha" = conf_int_alpha,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



.plot_auc_curve <- function(
    object,
    curve_type,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = NULL,
    plot_sub_title = NULL,
    caption = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    conf_int_style = c("ribbon", "step", "none"),
    conf_int_alpha = 0.2,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
) {
  
  # Get input data.
  x <- export_auc_data(
    object = object,
    aggregate_results = TRUE
  )
  
  # Check that the data are not empty.
  if (is_empty(x)) return(NULL)
  
  # Check that the data are not evaluated at the model level.
  if (all(sapply(x, function(x) (x@detail_level == "model")))) {
    ..warning_no_comparison_between_models()
    return(NULL)
  }
  
  # Obtain data element from list.
  if (is.list(x)) {
    if (is_empty(x)) return(NULL)
    
    if (length(x) > 1L) {
      ..error_reached_unreachable_code(
        ".plot_auc_curve: list of data elements contains unmerged elements."
      )
    }
    
    # Get x directly.
    x <- x[[1L]]
  }
  
  # Check that the data are not empty.
  if (is_empty(x)) return(NULL)
  
  # Check that curve type is correctly specified by the coder.
  .check_parameter_value_is_valid(
    x = curve_type,
    var_name = "curve_type",
    values = c("roc", "pr")
  )
  
  # Select the correct subset.
  if (curve_type == "roc") {
    x@data <- x@data[curve_type == "roc"]
  } else {
    x@data <- x@data[curve_type == "pr"]
  }
  
  # Check that the data are not empty after selecting the subset.
  if (is_empty(x)) return(NULL)
  
  # Check package requirements for plotting.
  if (!require_package(
    x = ..required_plotting_packages(extended = FALSE),
    purpose = ifelse(
      curve_type == "roc",
      "to create AUC-ROC plots",
      "to create AUC-PR plots"
    ),
    message_type = "warning"
  )) {
    return(NULL)
  }
  
  # Check input arguments ------------------------------------------------------
  
  # ggtheme
  ggtheme <- .check_ggtheme(ggtheme)
  
  # x_label
  if (is.waive(x_label) && curve_type == "roc") x_label <- "1 - specificity"
  if (is.waive(x_label) && curve_type == "pr") x_label <- "recall"
  
  # y_label
  if (is.waive(y_label) && curve_type == "roc") y_label <- "sensitivity"
  if (is.waive(y_label) && curve_type == "pr") y_label <- "precision"
  
  # x_range and y_range
  x_range <- y_range <- c(0.0, 1.0)
  
  # x_breaks
  if (is.null(x_breaks)) {
    .check_input_plot_args(x_n_breaks = x_n_breaks)
    
    # Create breaks
    x_breaks <- labeling::extended(
      m = x_n_breaks,
      dmin = x_range[1L],
      dmax = x_range[2L],
      only.loose = TRUE
    )
  }
  
  # y_breaks
  if (is.null(y_breaks)) {
    .check_input_plot_args(y_n_breaks = y_n_breaks)
    
    # Create breaks
    y_breaks <- labeling::extended(
      m = y_n_breaks,
      dmin = y_range[1L],
      dmax = y_range[2L],
      only.loose = TRUE
    )
  }
  
  # conf_int_style
  if (length(conf_int_style) > 1L) {
    conf_int_style <- head(conf_int_style, n = 1L)
  }
  
  # Set the style of the confidence interval to none, in case no confidence
  # interval data is present.
  if (!x@estimation_type %in% c("bci", "bootstrap_confidence_interval")) {
    conf_int_style <- "none"
  }
  
  # Splitting variables
  if (is.null(split_by) && is.null(facet_by) && is.null(color_by)) {
    # Determine the number of learners and feature_selection methods.
    n_learner <- nlevels(x@data$learner)
    n_vimp_method <- nlevels(x@data$vimp_method)
    n_positive_class <- ifelse(
      object@outcome_type == "binomial",
      1L, nlevels(x@data$positive_class)
    )
    
    if (n_learner > 1L && n_vimp_method > 1L) {
      split_by <- c("vimp_method", "learner")
      
      if (n_positive_class > 1L) {
        color_by <- "positive_class"
        facet_by <- "data_set"
        
      } else {
        color_by <- c("data_set", "positive_class")
      }
      
    } else if (n_learner > 1L) {
      # Implying n_vimp_method == 1
      
      if (n_positive_class > 1L) {
        split_by <- c("vimp_method", "learner")
        color_by <- "positive_class"
        facet_by <- "data_set"
        
      } else {
        split_by <- c("vimp_method")
        color_by <- c("learner")
        facet_by <- c("data_set", "positive_class")
      }
      
    } else if (n_vimp_method > 1L) {
      # Implying n_learner == 1
      
      if (n_positive_class > 1L) {
        split_by <- c("vimp_method", "learner")
        color_by <- "positive_class"
        facet_by <- "data_set"
      } else {
        split_by <- "learner"
        color_by <- "vimp_method"
        facet_by <- c("data_set", "positive_class")
      }
      
    } else {
      # Implying n_learner == n_vimp_method == 1
      split_by <- c("vimp_method", "learner")
      
      if (n_positive_class > 1L) {
        color_by <- "positive_class"
        facet_by <- "data_set"
      } else {
        color_by <- c("data_set", "positive_class")
      }
    }
  }
  
  # Check splitting variables and generate sanitised output
  split_var_list <- .check_plot_splitting_variables(
    x = x@data,
    split_by = split_by,
    color_by = color_by,
    facet_by = facet_by,
    available = c("vimp_method", "learner", "data_set", "positive_class")
  )
  
  # Update splitting variables
  split_by <- split_var_list$split_by
  color_by <- split_var_list$color_by
  facet_by <- split_var_list$facet_by
  
  # Create a legend label
  legend_label <- .create_plot_legend_title(
    user_label = legend_label,
    color_by = color_by
  )
  
  # Check input arguments for validity.
  .check_input_plot_args(
    x_range = x_range,
    y_range = y_range,
    x_breaks = x_breaks,
    y_breaks = y_breaks,
    conf_int_alpha = conf_int_alpha,
    conf_int_style = conf_int_style,
    facet_wrap_cols = facet_wrap_cols,
    x_label = x_label,
    y_label = y_label,
    legend_label = legend_label,
    plot_title = plot_title,
    plot_sub_title = plot_sub_title,
    caption = caption
  )
  
  # Create plots ---------------------------------------------------------------
  
  # Determine if subtitle should be generated.
  autogenerate_plot_subtitle <- is.waive(plot_sub_title)
  
  # Split data
  if (!is.null(split_by)) {
    x_split <- split(x@data, by = split_by)
  } else {
    x_split <- list("null.name" = x@data)
  }
  
  # Store plots to list in case dir_path is absent.
  if (is.null(dir_path)) plot_list <- list()
  
  # Iterate over splits
  for (ii in names(x_split)) {
    # Skip empty datasets
    if (is_empty(x_split[[ii]])) next
    
    if (is.waive(plot_title)) {
      plot_title <- ifelse(
        curve_type == "roc",
        "Receiver operating characteristic curve",
        "Precision-recall curve"
      )
    }
    
    if (autogenerate_plot_subtitle) {
      plot_sub_title <- .create_plot_subtitle(
        split_by = split_by,
        x = x_split[[ii]]
      )
    }
    
    # Generate plot
    p <- ..plot_auc_curve(
      x = x_split[[ii]],
      color_by = color_by,
      facet_by = facet_by,
      facet_wrap_cols = facet_wrap_cols,
      ggtheme = ggtheme,
      discrete_palette = discrete_palette,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      x_range = x_range,
      x_breaks = x_breaks,
      y_range = y_range,
      y_breaks = y_breaks,
      conf_int_style = conf_int_style,
      conf_int_alpha = conf_int_alpha
    )
    
    # Check empty output
    if (is.null(p)) next
    
    # Draw figure.
    if (draw) .draw_plot(plot_or_grob = p)
    
    # Save and export
    if (!is.null(dir_path)) {
      # Obtain decent default values for the plot.
      def_plot_dims <- .determine_auc_roc_plot_dimensions(
        x = x_split[[ii]],
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols
      )
      
      # Save to file.
      do.call(
        .save_plot_to_file,
        args = c(
          list(
            "plot_or_grob" = p,
            "object" = object,
            "dir_path" = dir_path,
            "type" = "performance",
            "subtype" = paste0("auc_", curve_type),
            "x" = x_split[[ii]],
            "split_by" = split_by,
            "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
            "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
            "units" = ifelse(is.waive(units), "cm", units)
          ),
          list(...)
        )
      )
      
    } else {
      # Store as list for export.
      plot_list <- c(plot_list, list(p))
    }
  }
  
  # Generate output
  return(.get_plot_results(
    dir_path = dir_path,
    plot_list = plot_list,
    export_collection = export_collection,
    object = object
  ))
}



..plot_auc_curve <- function(
    x,
    color_by,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_breaks,
    y_range,
    y_breaks,
    conf_int_style,
    conf_int_alpha
) {
  # Generate a guide table.
  guide_list <- .create_plot_guide_table(
    x = x, color_by = color_by,
    discrete_palette = discrete_palette
  )
  
  # Extract data
  x <- guide_list$data
  
  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("x"),
      y = !!sym("y")
    )
  )
  
  # Add theme
  p <- p + ggtheme
  
  # Add AUC curve
  if (is.null(color_by)) {
    # Without colour-based splitting.
    p <- p + ggplot2::geom_line()
    
  } else {
    # With colour-based splitting.
    p <- p + ggplot2::geom_line(
      mapping = ggplot2::aes(colour = !!sym("color_breaks"))
    )
    
    # Extract guidetable for color
    g_color <- guide_list$guide_color
    
    # Set colour and fill (fill may be unused)
    p <- p + ggplot2::scale_colour_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )
    
    p <- p + ggplot2::scale_fill_manual(
      name = legend_label$guide_color,
      values = g_color$color_values,
      breaks = g_color$color_breaks,
      drop = FALSE
    )
  }
  
  # Plot confidence intervals
  if (conf_int_style[1L] != "none") {
    if (conf_int_style[1L] == "step") {
      if (is.null(color_by)) {
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(y = !!sym("ci_low")),
          linetype = "dashed"
        )
        
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(y = !!sym("ci_up")),
          linetype = "dashed"
        )
        
      } else {
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_low"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed"
        )
        
        p <- p + ggplot2::geom_step(
          mapping = ggplot2::aes(
            y = !!sym("ci_up"),
            colour = !!sym("color_breaks")
          ),
          linetype = "dashed"
        )
      }
      
      # Do not show dashed lines in the legend.
      p <- p + ggplot2::scale_linetype(guide = "none")
      
    } else if (conf_int_style[1L] == "ribbon") {
      if (is.null(color_by)) {
        p <- p + ggplot2::geom_ribbon(
          mapping = ggplot2::aes(
            ymin = !!sym("ci_low"),
            ymax = !!sym("ci_up")
          ),
          alpha = conf_int_alpha
        )
        
      } else {
        p <- p + ggplot2::geom_ribbon(
          mapping = ggplot2::aes(
            ymin = !!sym("ci_low"),
            ymax = !!sym("ci_up"),
            fill = !!sym("color_breaks")
          ),
          alpha = conf_int_alpha
        )
      }
    }
  }
  
  # Update x and y scales
  p <- p + ggplot2::scale_x_continuous(breaks = x_breaks)
  p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)
  
  # Labels
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )
  
  # Determine how things are faceted
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )
  
  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context",
        drop = TRUE
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context",
        drop = TRUE
      )
    }
  }
  
  # Prevent removal of data outside or on plot limits.
  p <- p + ggplot2::coord_cartesian(
    xlim = x_range,
    ylim = y_range
  )
  
  return(p)
}



.determine_auc_roc_plot_dimensions <- function(
    x,
    facet_by,
    facet_wrap_cols
) {
  # Obtain faceting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- 4.0
  
  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))
  
  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))
  
  return(c(height, width))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarSVM ------------------------------------------------------------------
setClass(
  "familiarSVM",
  contains = "familiarModel"
)

setClass("familiarSVMC", contains = "familiarSVM")
setClass("familiarSVMNu", contains = "familiarSVM")
setClass("familiarSVMEps", contains = "familiarSVM")



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarSVM"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- "e1071"

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarSVM"),
  function(object, ...) {
    # Extract outcome type.
    outcome_type <- object@outcome_type

    if (outcome_type %in% c("continuous") &&
      (is(object, "familiarSVMNu") || is(object, "familiarSVMEps"))
    ) {
      return(TRUE)
      
    } else if (
      outcome_type %in% c("binomial", "multinomial") && 
      (is(object, "familiarSVMNu") || is(object, "familiarSVMC"))
    ) {
      return(TRUE)
    
    } else if (outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)  
      
    } else {
      return(FALSE)
    }
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarSVM"),
  function(object, data = NULL, ...) {
    # Find kernel type.
    svm_kernel <- ..find_kernel_type(learner = object@learner)

    # Initialise list and declare hyperparameter entries Note that some
    # hyperparameters may not be required, dependent on the kernel and type of
    # svm.
    param <- list()
    param$sign_size <- list()
    param$kernel <- list()
    param$c <- list()

    # Type-specific parameters.
    if (is(object, "familiarSVMNu") || is(object, "familiarSVMEps")) {
      param$epsilon <- list()
    }

    if (is(object, "familiarSVMNu")) {
      param$nu <- list()
    }

    # Kernel-specific parameters.
    if (svm_kernel %in% c("radial")) {
      param$gamma <- list()
    } else if (svm_kernel %in% c("polynomial")) {
      param$degree <- param$gamma <- param$offset <- list()
    } else if (svm_kernel %in% c("sigmoid")) {
      param$gamma <- param$offset <- list()
    }

    # If the data object is explicitly NULL, return the list with hyperparameter
    # names only.
    if (is.null(data)) return(param)

    # signature size -----------------------------------------------------------
    param$sign_size <- .get_default_sign_size(data = data)

    # svm kernel ---------------------------------------------------------------

    # A default kernel has been set earlier. Convert the svm_kernel to the
    # internal naming used by e1071::svm.

    # Set the svm kernel.
    param$kernel <- .set_hyperparameter(
      default = svm_kernel,
      type = "factor",
      range = svm_kernel,
      randomise = FALSE
    )

    # constraints violation cost C ---------------------------------------------

    # This parameter defines the cost for constraint violations. It is expressed
    # on a log10 scale.
    param$c <- .set_hyperparameter(
      default = c(-3.0, -1.0, -0.0, 1.0, 3.0),
      type = "numeric",
      range = c(-5.0, 3.0),
      valid_range = c(-Inf, Inf),
      randomise = TRUE
    )

    # error tolerance epsilon --------------------------------------------------
    if (is(object, "familiarSVMNu") || is(object, "familiarSVMEps")) {
      # This parameter defines the error tolerance for regression SVM. It is
      # expressed on a log10 scale.
      param$epsilon <- .set_hyperparameter(
        default = c(-5.0, -3.0, -1.0, 0.0, 1.0),
        type = "numeric",
        range = c(-5.0, 1.0),
        valid_range = c(-Inf, Inf),
        randomise = TRUE
      )
    }

    # error bounds parameter nu ------------------------------------------------
    if (is(object, "familiarSVMNu")) {
      # nu is expressed on a log10 scale.
      param$nu <- .set_hyperparameter(
        default = c(-5.0, -3.0, -1.0, 0.0, 1.0),
        type = "numeric",
        range = c(-5.0, 1.0),
        valid_range = c(-Inf, Inf),
        randomise = TRUE
      )
    }

    # inverse kernel width gamma -----------------------------------------------
    if (svm_kernel %in% c("radial", "polynomial", "sigmoid")) {
      # sigma is expressed on a log10 scale
      param$gamma <- .set_hyperparameter(
        default = c(-7.0, -5.0, -3.0, -1.0, 1.0),
        type = "numeric",
        range = c(-9.0, 3.0),
        valid_range = c(-Inf, Inf),
        randomise = TRUE
      )
    }

    # polynomial degree --------------------------------------------------------
    if (svm_kernel %in% c("polynomial")) {
      param$degree <- .set_hyperparameter(
        default = c(1L, 2L, 3L, 4L, 5L),
        type = "integer",
        range = c(1L, 5L),
        valid_range = c(1L, Inf),
        randomise = TRUE
      )
    }

    # kernel offset parameter --------------------------------------------------
    if (svm_kernel %in% c("polynomial", "sigmoid")) {
      # As feature data is rescaled internally by svm, we should not expect
      # offsets outside the [0, 1] range. Also, negative values are not allowed
      # for either kernel.
      param$offset <- .set_hyperparameter(
        default = c(0.0, 0.2, 0.5, 1.0),
        type = "numeric",
        range = c(0.0, 1.0),
        valid_range = c(0.0, Inf),
        randomise = TRUE
      )
    }

    return(param)
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarSVM",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Check if training data is ok.
    if (reason <- has_bad_training_data(object = object, data = data)) {
      return(callNextMethod(object = .why_bad_training_data(
        object = object,
        reason = reason
      )))
    }

    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Find feature columns in data table
    feature_columns <- get_feature_columns(x = data)

    # Parse the formula.
    formula <- stats::reformulate(
      termlabels = feature_columns,
      response = quote(outcome)
    )

    # Derive fitting parameters for fitting class probabilities.
    fit_probability <- object@outcome_type %in% c("binomial", "multinomial")

    # Derive svm type from object
    if (is(object, "familiarSVMC")) {
      svm_type <- "C-classification"
      
    } else if (
      is(object, "familiarSVMNu") &&
      object@outcome_type %in% c("binomial", "multinomial")
    ) {
      svm_type <- "nu-classification"
      
    } else if (
      is(object, "familiarSVMNu") &&
      object@outcome_type %in% c("continuous")
    ) {
      svm_type <- "nu-regression"
      
    } else if (is(object, "familiarSVMEps")) {
      svm_type <- "eps-regression"
      
    } else {
      ..error_reached_unreachable_code("..train,familiarSVM: can not set the type of SVM.")
    }

    # Set svm-related parameters.
    svm_parameter_list <- list(
      "kernel" = as.character(object@hyperparameters$kernel),
      "cost" = 10.0^(object@hyperparameters$c)
    )

    # Set nu-parameter (which not all svm types use).
    if (is(object, "familiarSVMNu")) {
      svm_parameter_list$nu <- 10.0^(object@hyperparameters$nu)
    }

    # Set epsilon parameter (which not all svm types use).
    if (is(object, "familiarSVMNu") || is(object, "familiarSVMEps")) {
      svm_parameter_list$epsilon <- 10.0^(object@hyperparameters$epsilon)
    }

    if (!is.null(object@hyperparameters$gamma)) {
      svm_parameter_list$gamma <- 10.0^object@hyperparameters$gamma
    }

    if (!is.null(object@hyperparameters$degree)) {
      svm_parameter_list$degree <- object@hyperparameters$degree
    }

    if (!is.null(object@hyperparameters$offset)) {
      svm_parameter_list$coef0 <- object@hyperparameters$offset
    }

    if (object@outcome_type %in% c("binomial", "multinomial")) {
      svm_parameter_list$class.weights <- "inverse"
    }

    # Train the model.
    model <- do.call_with_handlers(
      e1071::svm,
      args = c(
        list(formula,
          "data" = data@data,
          "type" = svm_type,
          "probability" = fit_probability,
          "fitted" = FALSE,
          "cross" = 0L
        ),
        svm_parameter_list
      )
    )
    
    # Extract values.
    object <- ..update_warnings(object = object, model$warning)
    object <- ..update_errors(object = object, model$error)
    model <- model$value

    # Check if the model trained at all.
    if (!is.null(object@messages$error)) return(callNextMethod(object = object))
    
    if (is.null(model)) {
      return(callNextMethod(object = ..update_errors(
        object = object,
        "SVM model returned as NULL."
      )))
    }

    # Add model
    object@model <- model

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..train_naive ----------------------------------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarSVM",
    data = "dataObject"
  ),
  function(object, data, ...) {
    if (object@outcome_type %in% c("continuous", "binomial", "multinomial")) {
      # Turn into a naive model.
      object <- methods::new("familiarNaiveModel", object)
    }

    return(..train(
      object = object,
      data = data,
      ...
    ))
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarSVM",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "default",
    ...
  ) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) return(callNextMethod())
    
    # Check if the data is empty.
    if (is_empty(data)) return(callNextMethod())
    
    if (type == "default") {
      # default ----------------------------------------------------------------

      # Make predictions using the model.
      model_predictions <- tryCatch(
        predict(
          object = object@model,
          newdata = data@data,
          probability = object@outcome_type %in% c("binomial", "multinomial")
        ),
        error = identity
      )

      # Check if the model trained at all.
      if (inherits(model_predictions, "error")) return(callNextMethod())

      if (object@outcome_type %in% c("binomial", "multinomial")) {
        # categorical outcomes -------------------------------------------------

        # Isolate probabilities.
        model_predictions <- attr(model_predictions, "probabilities")

        # Obtain class levels from the object.
        class_levels <- get_outcome_class_levels(x = object)
        
        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          if (is.matrix(model_predictions)) {
            prediction_list[[class_levels[ii]]] <- model_predictions[, class_levels[ii]]
          } else {
            prediction_list[[class_levels[ii]]] <- model_predictions[class_levels[ii]]
          }
        }

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("continuous")) {
        # numerical outcomes ---------------------------------------------------

        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = model_predictions,
          type = "regression",
          data = data,
          model_object = object
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }

      return(prediction_table)
      
    } else if (!.is_available_prediction_type(type)) {
      # user-specified method --------------------------------------------------
      return(predict(
        object = object@model,
        newdata = data@data,
        ...
      ))
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
  }
)



# ..vimp -----------------------------------------------------------------------
# SVM does not have an associated variable importance method.



# .trim_model-------------------------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarSVM"),
  function(object, ...) {
    # Update model by removing the call.
    object@model$call <- call("trimmed")

    # Add show.
    object <- .capture_show(object)

    # Remove .Environment.
    object@model$terms <- .replace_environment(object@model$terms)

    # Set is_trimmed to TRUE.
    object@is_trimmed <- TRUE

    # Default method for models that lack a more specific method.
    return(object)
  }
)



.get_available_svm_c_learners <- function(show_general = TRUE) {
  return(c("svm_c", paste("svm_c", ..get_available_svm_kernels(), sep = "_")))
}



.get_available_svm_nu_learners <- function(show_general = TRUE) {
  return(c("svm_nu", paste("svm_nu", ..get_available_svm_kernels(), sep = "_")))
}



.get_available_svm_eps_learners <- function(show_general = TRUE) {
  return(c("svm_eps", paste("svm_eps", ..get_available_svm_kernels(), sep = "_")))
}



..get_available_svm_kernels <- function() {
  return(c("linear", "radial", "polynomial", "sigmoid"))
}



..find_kernel_type <- function(learner) {
  # Find all available svm kernels.
  svm_kernels <- ..get_available_svm_kernels()
  
  # Find matches with end of learner string.
  kernel_matches <- sapply(
    svm_kernels,
    function(suffix, x) (endsWith(x = x, suffix = suffix)),
    x = learner
  )
  
  # If all are missing (e.g. "svm_eps), use default RBF kernel.
  if (!any(kernel_matches)) {
    return("radial")
  }
  
  # Else, return selected kernel.
  return(svm_kernels[kernel_matches])
}

# First test if all selectable learners are also available
familiar:::test_all_learners_available(
  learners = familiar:::.get_available_svm_c_learners(show_general = TRUE)
)
familiar:::test_all_learners_available(
  learners = familiar:::.get_available_svm_nu_learners(show_general = TRUE)
)
familiar:::test_all_learners_available(
  learners = familiar:::.get_available_svm_eps_learners(show_general = TRUE)
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_all_learners_train_predict_vimp(
  learners = familiar:::.get_available_svm_c_learners(show_general = FALSE),
  hyperparameter_list = list(
    "binomial" = list(
      "c" = -1.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    ),
    "multinomial" = list(
      "c" = -1.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    )
  ),
  has_vimp = FALSE
)


familiar:::test_all_learners_train_predict_vimp(
  learners = familiar:::.get_available_svm_nu_learners(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list(
      "c" = -1.0,
      "epsilon" = 0.0,
      "nu" = -4.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    ),
    "binomial" = list(
      "c" = -1.0,
      "epsilon" = 0.0,
      "nu" = -4.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    ),
    "multinomial" = list(
      "c" = -1.0,
      "epsilon" = 0.0,
      "nu" = -4.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    )
  ),
  has_vimp = FALSE
)


familiar:::test_all_learners_train_predict_vimp(
  learners = familiar:::.get_available_svm_eps_learners(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list(
      "c" = -1.0,
      "epsilon" = 0.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    )
  ),
  has_vimp = FALSE
)

# Parallel tests.
familiar:::test_all_learners_parallel_train_predict_vimp(
  learners = familiar:::.get_available_svm_c_learners(show_general = FALSE),
  hyperparameter_list = list(
    "binomial" = list(
      "c" = -1.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    ),
    "multinomial" = list(
      "c" = -1.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    )
  ),
  has_vimp = FALSE
)


familiar:::test_all_learners_parallel_train_predict_vimp(
  learners = familiar:::.get_available_svm_nu_learners(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list(
      "c" = -1.0,
      "epsilon" = 0.0,
      "nu" = -4.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    ),
    "binomial" = list(
      "c" = -1.0,
      "epsilon" = 0.0,
      "nu" = -4.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    ),
    "multinomial" = list(
      "c" = -1.0,
      "epsilon" = 0.0,
      "nu" = -4.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    )
  ),
  has_vimp = FALSE
)


familiar:::test_all_learners_parallel_train_predict_vimp(
  learners = familiar:::.get_available_svm_eps_learners(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list(
      "c" = -1.0,
      "epsilon" = 0.0,
      "gamma" = 0.1,
      "degree" = 2.0,
      "offset" = 0.0
    )
  ),
  has_vimp = FALSE
)


# Continuous outcome tests------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("continuous")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "c" = -1.0,
    "epsilon" = 0.0,
    "gamma" = 1.0
  ),
  learner = "svm_eps_radial"
)

testthat::test_that("SVM model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # Check that no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("SVM model has no variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)

  # Expect that the vimp table is empty.
  testthat::expect_true(is_empty(vimp_table))
})


# Binomial tests----------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("binomial")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "c" = -1.0,
    "gamma" = 1.0
  ),
  learner = "svm_c_radial"
)

testthat::test_that("SVM model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # Check that no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("SVM model has no variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)

  # Expect that the vimp table is empty.
  testthat::expect_true(is_empty(vimp_table))
})


# Multinomial tests-------------------------------------------------------------

# Create test data sets.
good_data <- familiar:::test_create_good_data("multinomial")

# Train the model using the good dataset.
good_model <- familiar:::test_train(
  data = good_data,
  cluster_method = "none",
  imputation_method = "simple",
  hyperparameter_list = list(
    "sign_size" = familiar:::get_n_features(good_data),
    "c" = -1.0,
    "gamma" = 1.0
  ),
  learner = "svm_c_radial"
)

testthat::test_that("SVM model trained correctly", {
  # Model trained
  testthat::expect_true(familiar:::model_is_trained(good_model))

  # Check that no deprecation warnings are given.
  familiar:::test_not_deprecated(good_model@messages$warning)

  # Test that no errors appear.
  testthat::expect_equal(good_model@messages$error, NULL)
})

testthat::test_that("SVM model has no variable importance", {
  # Extract the variable importance table.
  vimp_table <- familiar:::get_vimp_table(good_model)

  # Expect that the vimp table is empty.
  testthat::expect_true(is_empty(vimp_table))
})


familiar:::test_hyperparameter_optimisation(
  learners = "svm_c_radial",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)

familiar:::test_hyperparameter_optimisation(
  learners = "svm_nu_radial",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)

familiar:::test_hyperparameter_optimisation(
  learners = "svm_eps_radial",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)


testthat::skip("Skip hyperparameter optimisation, unless manual.")

familiar:::test_hyperparameter_optimisation(
  learners = familiar:::.get_available_svm_c_learners(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE
)

familiar:::test_hyperparameter_optimisation(
  learners = familiar:::.get_available_svm_nu_learners(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE
)

familiar:::test_hyperparameter_optimisation(
  learners = familiar:::.get_available_svm_eps_learners(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test
familiar:::test_plots(
  plot_function = familiar:::plot_sample_clustering,
  not_available_all_predictions_fail = FALSE,
  not_available_some_predictions_fail = FALSE,
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  data_element = "feature_expressions",
  plot_args = list("verbose" = FALSE),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list("verbose" = FALSE),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "verbose" = FALSE),
  debug = debug_flag
)

# Try swapping the axis.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "x_axis_by" = "sample",
    "y_axis_by" = "feature",
    "verbose" = FALSE),
  debug = debug_flag
)

# Try swapping the axis and putting the outcome below the heatmap
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_sample_clustering,
  data_element = "feature_expressions",
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  plot_args = list(
    "facet_by" = c("learner", "vimp_method", "data_set"),
    "x_axis_by" = "sample",
    "y_axis_by" = "feature",
    "show_outcome" = "bottom",
    "verbose" = FALSE),
  debug = debug_flag
)

testthat::skip_on_cran()
testthat::skip_on_ci()

# Perform an integrated test. Note that learner is set automatically.
familiar:::integrated_test(
  backend_type = "none",
  experimental_design = "bt(fs+mb,20)",
  parallel_nr_cores = 2,
  vimp_method = "none",
  cluster_method = "none",
  imputation_method = "simple",
  parallel = TRUE,
  skip_evaluation_elements = "all"
)

# Perform an integrated test using the socket backend.
familiar:::integrated_test(
  backend_type = "socket_server",
  experimental_design = "bt(fs+mb,20)",
  parallel_nr_cores = 2,
  vimp_method = "none",
  cluster_method = "none",
  imputation_method = "simple",
  parallel = TRUE,
  skip_evaluation_elements = "all"
)

# Test use of externally started cluster.

# Start cluster
cl <- familiar:::.start_cluster(n_cores = 2L, cluster_type = "psock")

familiar:::integrated_test(
  backend_type = "none",
  cl = cl,
  experimental_design = "bt(fs+mb,20)",
  vimp_method = "mrmr",
  cluster_method = "none",
  imputation_method = "simple",
  parallel = TRUE,
  skip_evaluation_elements = "all"
)

cl <- familiar:::.terminate_cluster(cl = cl)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

setClass(
  "featureInfoParametersTransformationPowerTransform",
  contains = "featureInfoParameters",
  slots = list(
    "method" = "character",
    "transformer" = "ANY",
    "fitting_parameters" = "ANY"
  ),
  prototype = list(
    "method" = "none",
    "transformer" = NULL,
    "fitting_parameters" = NULL
  )
)

# NOTE: the classes below were used prior to version 1.5.0. These classes should
# not be removed to make sure that the corresponding objects can be updated. All
# related methods have been deprecated.

setClass(
  "featureInfoParametersTransformationNone",
  contains = "featureInfoParameters",
  slots = list("reason" = "ANY"),
  prototype = list("reason" = NULL)
)

setClass(
  "featureInfoParametersTransformationBoxCox",
  contains = "featureInfoParametersTransformationPowerTransform",
  slots = list("lambda" = "numeric"),
  prototype = list("lambda" = NA_real_)
)

setClass(
  "featureInfoParametersTransformationYeoJohnson",
  contains = "featureInfoParametersTransformationPowerTransform",
  slots = list("lambda" = "numeric"),
  prototype = list("lambda" = NA_real_)
)



.get_available_none_transformation_methods <- function() {
  return("none")
}

.get_available_box_cox_transformation_methods <- function() {
  return(c("box_cox", "box_cox_robust", "box_cox_conventional"))
}

.get_available_yeo_johnson_transformation_methods <- function() {
  return(c("yeo_johnson", "yeo_johnson_robust", "yeo_johnson_conventional"))
}

.get_available_transformation_methods <- function(type = "all") {
  # Check type
  if (!type %in% c("all", "none", "box_cox", "yeo_johnson")) {
    ..error_reached_unreachable_code(paste0(
      ".get_available_transformation_methods: unspecified type: ", type
    ))
  }

  available_transformation_method <- NULL

  if (type %in% c("none", "all")) {
    available_transformation_method <- c(
      available_transformation_method,
      .get_available_none_transformation_methods()
    )
  }

  if (type %in% c("all", "box_cox")) {
    available_transformation_method <- c(
      available_transformation_method,
      .get_available_box_cox_transformation_methods()
    )
  }

  if (type %in% c("all", "yeo_johnson")) {
    available_transformation_method <- c(
      available_transformation_method,
      .get_available_yeo_johnson_transformation_methods()
    )
  }

  return(available_transformation_method)
}



create_transformation_parameter_skeleton <- function(
    feature_info_list,
    feature_names = NULL,
    transformation_method,
    transformation_lambda = NULL,
    transformation_optimisation_criterion = "mle",
    transformation_gof_p_value = NULL,
    .override_existing = FALSE
) {
  # Creates a skeleton for the provided transformation method. If
  # transformation_lambda is provided (typically not), this value is updated as
  # well.

  # Determine feature names from the feature info list, if provided.
  if (is.null(feature_names)) feature_names <- names(feature_info_list)

  # Select only features that appear in the feature info list.
  feature_names <- intersect(
    names(feature_info_list),
    feature_names
  )

  # Skip step if no feature info objects are updated.
  if (is_empty(feature_names)) return(feature_info_list)

  # Check that method is applicable.
  .check_parameter_value_is_valid(
    x = transformation_method,
    var_name = "transformation_method",
    values = .get_available_transformation_methods()
  )

  # Check that transformation_lambda is numeric. This is slightly redundant, as
  # this is also checked by the power.transform package.
  if (!is.null(transformation_lambda)) {
    .check_number_in_valid_range(
      x = transformation_lambda,
      var_name = "transformation_lambda",
      range = c(-Inf, Inf)
    )
  }

  # Update familiar info objects with a feature transformation skeleton.
  updated_feature_info <- fam_lapply(
    X = feature_info_list[feature_names],
    FUN = .create_transformation_parameter_skeleton,
    method = transformation_method,
    lambda = transformation_lambda,
    optimisation_criterion = transformation_optimisation_criterion,
    gof_p_value = transformation_gof_p_value,
    .override_existing = .override_existing
  )

  # Provide names for the updated feature info objects.
  names(updated_feature_info) <- feature_names

  # Replace list elements.
  feature_info_list[feature_names] <- updated_feature_info

  return(feature_info_list)
}



.create_transformation_parameter_skeleton <- function(
    feature_info, 
    method,
    optimisation_criterion = "mle",
    gof_p_value = NULL,
    lambda = NULL, 
    .override_existing = FALSE
) {
  # Check if transformation data was already completed, and does not require
  # being determined anew.
  if (
    feature_info_complete(feature_info@transformation_parameters) &&
    !.override_existing
  ) {
    return(feature_info)
  }

  # Pass to underlying function that constructs the skeleton.
  object <- ..create_transformation_parameter_skeleton(
    feature_name = feature_info@name,
    feature_type = feature_info@feature_type,
    available = is_available(feature_info),
    method = method,
    optimisation_criterion = optimisation_criterion,
    gof_p_value = gof_p_value,
    lambda = lambda
  )

  # Update transformation_parameters slot.
  feature_info@transformation_parameters <- object

  return(feature_info)
}



..create_transformation_parameter_skeleton <- function(
    feature_name,
    feature_type = "numeric",
    available = TRUE,
    method,
    optimisation_criterion = "mle",
    gof_p_value = NULL,
    lambda = NULL
) {
  # This is the lowest level function for creation transformation parameter
  # skeletons. This function always generates the same class of object. Fitting
  # parameters are passed on to power.transform::find_transformation_parameters.

  if (!require_package("power.transform", message_type = "silent")) {
    fitting_parameters <- list("method" = "none")
    
  } else if (
    feature_type != "numeric" || 
    !available || 
    (method %in% .get_available_none_transformation_methods())
  ) {
    fitting_parameters <- list("method" = "none")
    
  } else if (method %in% .get_available_box_cox_transformation_methods()) {
    fitting_parameters <- list("method" = "box_cox")
    
  } else if (method %in% .get_available_yeo_johnson_transformation_methods()) {
    fitting_parameters <- list("method" = "yeo_johnson")
    
  } else {
    ..error_reached_unreachable_code(paste0(
      "create_transformation_parameter_skeleton: encountered an unknown transformation method: ",
      paste_s(method)
    ))
  }
  
  # Set estimation method (optimisation criterion).
  fitting_parameters <- c(
    fitting_parameters,
    list("estimation_method" = optimisation_criterion)
  )
  
  # Set invariant argument.
  fitting_parameters <- c(
    fitting_parameters,
    list("invariant" = !grepl(pattern = "conventional", x = method))
  )
  
  # Set robust argument
  fitting_parameters <- c(
    fitting_parameters,
    list("robust" = grepl(pattern = "robust", x = method))
  )
  
  # Set significance level for the empirical goodness-of-fit test.
  if (!is.null(gof_p_value)) {
    fitting_parameters <- c(
      fitting_parameters,
      list("empirical_gof_normality_p_value" = gof_p_value)
    )
  } 
  
  # Provide lambda.
  if (!is.null(lambda)) {
    fitting_parameters <- c(
      fitting_parameters,
      list("lambda" = lambda)
    )
    
    # Do not determine location and scale parameters -- external lambda is only
    # set using external objects.
    fitting_parameters$invariant <- FALSE
  }
  
  object <- methods::new(
    "featureInfoParametersTransformationPowerTransform",
    fitting_parameters = fitting_parameters
  )
  
  # Set the name of the object.
  object@name <- feature_name

  # Update the familiar version.
  object <- add_package_version(object = object)

  return(object)
}



add_transformation_parameters <- function(
    cl = NULL,
    feature_info_list,
    data,
    verbose = FALSE
) {
  # Determine transformation parameters and add them to the feature_info_list.

  # Find feature columns.
  feature_names <- get_feature_columns(x = data)

  # Sanity check.
  if (!(setequal(
    feature_names,
    get_available_features(feature_info_list = feature_info_list)
  ))) {
    ..error_reached_unreachable_code(paste0(
      "add_transformation_parameters: features in data and the feature info ",
      "list are expect to be the same, but were not."
    ))
  }

  # Iterate over features.
  updated_feature_info <- fam_mapply(
    cl = cl,
    FUN = .add_transformation_parameters,
    feature_info = feature_info_list[feature_names],
    data = data@data[, mget(feature_names)],
    progress_bar = verbose,
    chopchop = TRUE
  )

  # Provide names for the updated feature info objects.
  names(updated_feature_info) <- feature_names

  # Replace list elements.
  feature_info_list[feature_names] <- updated_feature_info

  return(feature_info_list)
}



.add_transformation_parameters <- function(
    feature_info,
    data
) {
  # Pass to underlying function that adds the feature info.
  object <- add_feature_info_parameters(
    object = feature_info@transformation_parameters,
    data = data
  )

  # Update transformation_parameters slot.
  feature_info@transformation_parameters <- object

  return(feature_info)
}



# add_feature_info_parameters (general, ANY) ---------------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersTransformationPowerTransform",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Check if the power.transform package is present.
    if (!require_package("power.transform", message_type = "silent")) {
      object@method <- "none"
      object@complete <- TRUE
      
      return(object)
    }

    # The case where all data are missing is not handled by the power.transform
    # package, which (correctly) throws an error.
    if (is.numeric(data)) {
      if (!any(is.finite(data))) {
        object@fitting_parameters$method <- "none"
      }
    }
    
    # Create transformer using the power.transform package. Suppress specific
    # types of warnings related to the input data.
    transformer <- suppressWarnings(
      do.call(
        power.transform::find_transformation_parameters,
        args = c(
          list("x" = data),
          object@fitting_parameters
        )
      ),
      classes = c(
        "power_transform_no_transform",
        "power_transform_few_unique_values",
        "power_transform_transform_invalid_values"
      )
    )

    object@transformer <- transformer
    object@method <- power.transform::get_transformation_method(transformer)
    object@complete <- TRUE
    
    return(object)
  }
)



# apply_feature_info_parameters (general, ANY) ---------------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersTransformationPowerTransform", 
    data = "ANY"
  ),
  function(
    object,
    data,
    invert = FALSE,
    ...
  ) {
    
    # Check if the power.transform package is present.
    if (!require_package("power.transform", message_type = "silent")) {
      if (object@method == "none") {
        return(data)
        
      } else {
        require_package(
          "power.transform",
          purpose = "to transform features",
          message_type = "error"
        )
      }
    }
    
    if (invert) {
      return(suppressWarnings(
        power.transform::revert_power_transform(
          y = data,
          transformer = object@transformer
        ),
        classes = "power_transform_transform_invalid_values"
      ))
      
    } else {
      return(suppressWarnings(
        power.transform::power_transform(
          x = data,
          transformer = object@transformer,
          oob_action = "valid"
        ),
        classes = "power_transform_transform_invalid_values"
      ))
    }
  }
)



..collect_and_aggregate_transformation_info <- function(
    feature_info_list, 
    instance_mask, 
    feature_name
) {
  # Aggregate transformation parameters. This function exists so that it can be
  # tested as part of a unit test.

  # Suppress NOTES due to non-standard evaluation in data.table
  n <- method <- NULL
  
  none_object <- ..create_transformation_parameter_skeleton(
    feature_name = feature_name,
    method = "none"
  )
  none_object@transformer <- power.transform::create_transformer_skeleton(method = "none")
  
  if (!any(instance_mask)) {
    return(list(
      "parameters" = none_object,
      "instance_mask" = instance_mask
    ))
  }

  # Check the method of the transformation objects.
  object_method <- sapply(
    feature_info_list, 
    function(x) {
      if (is.null(x@transformation_parameters)) return("none")
      
      return(x@transformation_parameters@method)
    }
  )

  # Determine if there are any objects that exist and don't use the "none"
  # method.
  if (all(object_method[instance_mask] == "none")) {
    return(list(
      "parameters" = none_object,
      "instance_mask" = instance_mask
    ))
  }

  # For the remaining objects, check which class occurs most.
  method_table <- data.table::data.table(
    "method" = object_method[instance_mask]
  )[, list("n" = .N), by = "method"]
  
  # Drop none transformations.
  method_table <- method_table[!method == "none"]

  # Select the method that occurs most often.
  most_common_method <- method_table[n == max(method_table$n), ]$method[1L]

  # Update the instance mask.
  instance_mask <- instance_mask & object_method == most_common_method

  # Because both lambda, shift and scale parameters may be varied, select only
  # one instance.
  selected_instance <- head(which(instance_mask), n = 1L)
  instance_mask <- logical(length(instance_mask))
  instance_mask[selected_instance] <- TRUE

  return(list(
    "parameters" = feature_info_list[[selected_instance]]@transformation_parameters,
    "instance_mask" = instance_mask
  ))
}

#' Parse experimental design
#'
#' @param message_indent Spacing inserted before messages.
#' @param verbose Sets verbosity.
#' @inheritParams .parse_experiment_settings
#'
#' @details This function converts the experimental_design string
#'
#' @return data.table with subsampler information at different levels of the
#' experimental design.
#'
#' @md
#' @keywords internal
extract_experimental_setup <- function(
    experimental_design,
    file_dir,
    message_indent = 0L,
    verbose = TRUE
) {
  
  if (.experimental_design_is_file(
    file_dir = file_dir,
    experimental_design = experimental_design
  )) {
    return(waiver())
  }
  
  # Remove all whitespace
  experimental_design <- gsub(
    pattern = " ",
    replacement = "",
    x = experimental_design
  )
  
  # Generate a section table
  section_table <- .get_experimental_design_section_table(
    experimental_design = experimental_design
  )
  
  # Identify the subsampler algorithms
  section_table <- .complete_experimental_design_section_table(
    section_table = section_table,
    experimental_design = experimental_design
  )
  
  # Check consistency of the table, e.g. feature selection should only appear
  # once, etc.
  .check_experimental_design_section_table(
    section_table = section_table
  )
  
  # Identify the experimental levels where pre-processing could occur.
  .set_experimental_design_preprocessing(
    section_table = section_table
  )
  
  # Report experimental design to the user.
  .report_experimental_design(
    section_table = section_table,
    message_indent = message_indent,
    verbose = verbose
  )
  
  return(section_table)
}



.experimental_design_is_file <- function(
    file_dir,
    experimental_design
) {
  # Check if the experimental design argument is actually a path to a file.
  
  if (is.null(file_dir)) return(FALSE)
  
  # Check if the file exists at all.
  if (
    !file.exists(file.path(file_dir, experimental_design)) &&
    !file.exists(experimental_design)
  ) {
    return(FALSE)
  }
  
  # Check whether the file is an RDS file.
  return(tolower(.file_extension(basename(experimental_design))) == "rds")
}



.get_experimental_design_section_table <- function(experimental_design) {
  
  # Determine the number of experimental levels
  # First we locate the position of parentheses
  left_parenthesis <- gregexpr(
    pattern = "(",
    text = experimental_design,
    fixed = TRUE
  )[[1L]]
  right_parenthesis <- gregexpr(
    pattern = ")",
    text = experimental_design,
    fixed = TRUE
  )[[1L]]
  
  if (left_parenthesis[1L] == -1L) left_parenthesis <- integer(0L)
  if (right_parenthesis[1L] == -1L) right_parenthesis <- integer(0L)
  
  # Subsequently generate the corresponding experimental levels
  experiment_levels <- integer(nchar(experimental_design))
  for (ii in right_parenthesis) {
    experiment_levels[1L:ii] <- experiment_levels[1L:ii] + 1L
  }
  
  for (ii in left_parenthesis) {
    experiment_levels[1L:(ii - 1L)] <- experiment_levels[1L:(ii - 1L)] - 1L
  }
  
  # Generate setup sections
  sections <- rle(experiment_levels)
  sections$end <- cumsum(sections$lengths)
  sections$start <- sections$end - sections$lengths + 1L
  class(sections) <- NULL
  
  # Set up data table with experimental sections
  section_table <- data.table::as.data.table(sections)
  section_table[, "lengths" := NULL]
  data.table::setnames(
    x = section_table,
    old = c("values", "end", "start"),
    new = c("exp_level_id", "sect_end", "sect_start")
  )
  
  # Set up columns to be filled
  section_table[, ":="(
    "ref_data_id" = 0L,
    "main_data_id" = 0L,
    "vimp" = FALSE,
    "train" = FALSE,
    "internal_validation" = FALSE,
    "external_validation" = FALSE,
    "perturb_method" = "none",
    "perturb_n_rep" = 0L,
    "perturb_n_folds" = 0L
  )]
  
  return(section_table)
}



.get_available_subsample_methods <- function() {
  return(c(
    "main",
    "limited_bootstrap",
    "full_bootstrap",
    "cross_val",
    "loocv",
    "imbalance_partition"
  ))
}



.set_experimental_design_preprocessing <- function(section_table) {
  # Suppress NOTES due to non-standard evaluation in data.table
  perturb_method <- NULL
  
  # Add can_preprocess column.
  section_table[, "can_pre_process" := TRUE]
  
  # Set can_pre_process to FALSE for select perturbation methods.
  section_table[
    perturb_method %in% c("limited_bootstrap"),
    ":="(can_pre_process = FALSE, internal_validation = FALSE)
  ]
  
  return(section_table)
}



.set_experimental_design_n_runs <- function(
    section_table,
    iteration_list
) {
  # Determine the number of runs for each level in the experimental design based
  # on the actually realised iterations. This ensures that the number of runs
  # is consistent with the iteration list. Some perturbation methods, such as
  # LOOCV, don't have a number of runs that can be determined without inspecting
  # the data. Hence we set n_runs based off the iteration list.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  main_data_id <- NULL
  
  ...get_n_samples <- function(x, type) {
    if (is_empty(x[[type]])) return(0L)
    
    return(nrow(x[[type]]))
  }
  
  # Add perturbation level.
  section_table[, "perturbation_level" := 1L]
  for (data_id in section_table$main_data_id) {
    section_table[main_data_id == data_id, "perturbation_level" := tail(iteration_list[[as.character(data_id)]]$run[["1"]]$run_table, n = 1L)$perturb_level]
  }
  
  # Add number of runs.
  section_table[, "n_runs" := 1L]
  for (data_id in section_table$main_data_id) {
    section_table[main_data_id == data_id, "n_runs" := length(iteration_list[[as.character(data_id)]]$run)]
  }
  
  # Determine the number of instances available for development and validation.
  for (data_id in section_table$main_data_id) {
    
    n_run_training_samples <- sapply(iteration_list[[as.character(data_id)]]$run, ...get_n_samples, type = "train_samples")
    n_run_validation_samples <- sapply(iteration_list[[as.character(data_id)]]$run, ...get_n_samples, type = "valid_samples")
    
    section_table[main_data_id == data_id, ":="(
      "min_training_instances" = min(n_run_training_samples),
      "max_training_instances" = max(n_run_training_samples),
      "min_validation_instances" = min(n_run_validation_samples),
      "max_validation_instances" = max(n_run_validation_samples)
    )]
  }

  return(section_table)
}



.complete_experimental_design_section_table <- function(
    section_table,
    experimental_design
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  exp_level_id <- sect_start <- perturb_method <- train <- main_data_id <- NULL
  
  # Iterator
  main_data_id_iter <- 1L
  
  #Identify samplers------------------------------------------------------------
  # Iterate over sections to set main_data_id
  for (ii in seq_len(nrow(section_table))) {
    
    # Check if ip, bt, lv or cv preceeds the current section
    if (section_table$sect_start[ii] > 2L) {
      sampler_str <- substr(
        x = experimental_design,
        start = section_table$sect_start[ii] - 2L,
        stop = section_table$sect_start[ii] - 1L
      )
      
      # Check for imbalance partition (ip), limited bootstrap (bt), full
      # bootstrap (bs), cross-validation (cv) and leave-one-out-cross-validation
      # (lv)
      if (sampler_str == "bt") {
        section_table$perturb_method[ii] <- "limited_bootstrap"
        section_table$main_data_id[ii] <- main_data_id_iter
        main_data_id_iter <- main_data_id_iter + 1L
      
      } else if (sampler_str == "bs") {
        section_table$perturb_method[ii] <- "full_bootstrap"
        section_table$main_data_id[ii] <- main_data_id_iter
        main_data_id_iter <- main_data_id_iter + 1L
        
      } else if (sampler_str == "cv") {
        section_table$perturb_method[ii] <- "cross_val"
        section_table$main_data_id[ii] <- main_data_id_iter
        main_data_id_iter <- main_data_id_iter + 1L
        
      } else if (sampler_str == "lv") {
        section_table$perturb_method[ii] <- "loocv"
        section_table$main_data_id[ii] <- main_data_id_iter
        main_data_id_iter <- main_data_id_iter + 1L
        
      } else if (sampler_str == "ip") {
        section_table$perturb_method[ii] <- "imbalance_partition"
        section_table$main_data_id[ii] <- main_data_id_iter
        main_data_id_iter <- main_data_id_iter + 1L
      }
      
      rm(sampler_str)
    } else {
      section_table$perturb_method[ii] <- "main"
      section_table$main_data_id[ii] <- main_data_id_iter
      main_data_id_iter <- main_data_id_iter + 1L
    }
  }
  
  rm(main_data_id_iter)
  
  
  # Add main_data_id and ref_data_id -------------------------------------------
  # Iterate over sections to set main_data_id (where missing) and ref_data_id
  for (ii in seq_len(nrow(section_table))) {
    
    if (section_table$main_data_id[ii] == 0L) {
      # Make subselection of data at the same level and select only preceding sections
      dt_sub <- section_table[
        exp_level_id == section_table$exp_level_id[ii] &
          sect_start < section_table$sect_start[ii]
        ,
      ]
      
      # Set data id from nearest preceding lower level section
      section_table$main_data_id[ii] <- dt_sub$main_data_id[nrow(dt_sub)]
      
      rm(dt_sub)
    }
    
    # Set reference data id
    if (section_table$exp_level_id[ii] == 0L) {
      # At the lowest level (main) there is no reference
      section_table$ref_data_id[ii] <- 0L
      
    } else {
      # Make subselection of data one level higher and select only preceding sections
      dt_sub <- section_table[
        exp_level_id == section_table$exp_level_id[ii] - 1L &
          sect_start < section_table$sect_start[ii]
        ,
      ]
      
      # Set data id from nearest preceding lower level section
      section_table$ref_data_id[ii] <- dt_sub$main_data_id[nrow(dt_sub)]
      
      rm(dt_sub)
    }
  }
  
  # Complete details------------------------------------------------------------
  # Add details for perturbations and other sections
  for (ii in seq_len(nrow(section_table))) {
    
    if (section_table$perturb_method[ii] %in% .get_available_subsample_methods()) {
      
      # Create readable string for current data id
      curr_data_id_str <- NULL
      for (jj in which(section_table$main_data_id == section_table$main_data_id[ii])) {
        curr_data_id_str <- c(
          curr_data_id_str,
          substr(
            x = experimental_design,
            start = section_table$sect_start[jj],
            stop = section_table$sect_end[jj]
          )
        )
      }
      
      curr_data_id_str <- paste0(curr_data_id_str, collapse = "")
      
      # Drop parentheses and split string by comma
      curr_data_id_str <- gsub(
        pattern = "\\(|\\)",
        replacement = "",
        x = curr_data_id_str
      )
      curr_data_id_str <- strsplit(
        x = curr_data_id_str,
        split = ",",
        fixed = TRUE
      )[[1L]]

      # Check if feature selection is included in the current section
      if (grepl(pattern = "fs", x = curr_data_id_str[1L])) {
        section_table$vimp[ii] <- TRUE
      }
      
      # Check if model building is included in the current section
      if (grepl(pattern = "mb", x = curr_data_id_str[1L])) {
        section_table$train[ii] <- TRUE
      }
      
      # Check if external validation is included in the current section
      if (grepl(pattern = "ev", x = curr_data_id_str[1L])) {
        section_table$external_validation[ii] <- TRUE
      }
      
      # Read bootstrap data
      if (section_table$perturb_method[ii] %in% c("limited_bootstrap", "full_bootstrap")) {
        if (length(curr_data_id_str) < 2L) {
          ..error(
            paste0(
              "The number of bootstraps should be indicated when using the bt ",
              "(bootstrap) subsampler. None was found."
            ),
            error_class = "input_argument_error"
          )
        }
        
        # Determine the number of bootstraps
        n_reps <- .perform_type_conversion(
          x = curr_data_id_str[2L],
          to_type = "integer",
          var_name = "The number of bootstraps",
          req_length = 1L
        )
        
        # Check whether the number of bootstraps is at least 1
        .check_number_in_valid_range(
          x = n_reps,
          var_name = "The number of bootstraps",
          range = c(1L, Inf)
        )
        
        if (section_table$perturb_method[ii] == "full_bootstrap") {
          section_table$internal_validation[ii] <- TRUE
        }
        
        # Add the number of bootstraps to the section table
        section_table$perturb_n_rep[ii] <- n_reps
      }
      
      # Read cross-validation settings
      if (section_table$perturb_method[ii] == "cross_val") {
        if (length(curr_data_id_str) < 2L) {
          ..error(
            paste0(
              "The number of folds should be indicated when using the cv ",
              "(cross-validation) subsampler. None was found."
            ),
            error_class = "input_argument_error"
          )
        }
        
        # Determine the number of folds
        n_folds <- .perform_type_conversion(
          x = curr_data_id_str[2L],
          to_type = "integer",
          var_name = "The number of cross-validation folds",
          req_length = 1L
        )
        
        # Check whether the number of folds is at least 2
        .check_number_in_valid_range(
          x = n_folds,
          var_name = "The number of cross-validations folds",
          range = c(2L, Inf)
        )
        
        # Set internal validation.
        section_table$internal_validation[ii] <- TRUE
        
        # Add number of folds to the section_table
        section_table$perturb_n_folds[ii] <- n_folds
        
        # Check the number of repetitions
        if (length(curr_data_id_str) >= 3L) {
          
          n_reps <- .perform_type_conversion(
            x = curr_data_id_str[3L],
            to_type = "integer",
            var_name = "The number of cross-validation repetitions",
            req_length = 1L
          )
          
          # Check whether the number of CV repetitions is at least 1
          .check_number_in_valid_range(
            x = n_folds,
            var_name = "The number of cross-validations repetitions",
            range = c(1L, Inf)
          )
          
          # Add number of repetitions to the section table
          section_table$perturb_n_rep[ii] <- n_reps
          
        } else {
          # Set the number of repetitions to one
          section_table$perturb_n_rep[ii] <- 1L
        }
      }
      
      # Read leave-one-out-cross-validation settings
      if (section_table$perturb_method[ii] == "loocv") {
        # Set internal validation.
        section_table$internal_validation[ii] <- TRUE
        
        section_table$perturb_n_folds[ii] <- -1L
        section_table$perturb_n_rep[ii] <- 1L
      }
      
      rm(curr_data_id_str, jj)
    }
  }
  
  # Remove unnecessary rows and columns
  section_table <- section_table[perturb_method != "none", ]
  section_table[, ":="(
    "exp_level_id" = NULL,
    "sect_end" = NULL,
    "sect_start" = NULL
  )]
  
  # Internal validation can only be at the same level as train, or above.
  if (any(section_table$internal_validation)) {
    internal_validation_data_id <- 0L
    data_id <- section_table[train == TRUE]$main_data_id[1L]
    while (data_id > 1L) {
      if (section_table[main_data_id == data_id]$internal_validation[1L] == TRUE) {
        # Update internal_validation_data_id if internal validation could occur
        # at the current level. This gets overridden by lower (i.e.
        # higher-level) data ids.
        internal_validation_data_id <- data_id
      }
      
      # Go to the next level.
      data_id <- section_table[main_data_id == data_id]$ref_data_id[1L]
    }
    
    # Set all internal_validation to FALSE.
    section_table[, "internal_validation" := FALSE]
    if (internal_validation_data_id > 0L) {
      section_table[main_data_id == internal_validation_data_id, "internal_validation" := TRUE]
    }
  }
  
  return(section_table)
}



.report_experimental_design <- function(
    section_table,
    message_indent = 0L,
    verbose = TRUE
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  vimp <- train <- main_data_id <- NULL
  
  # Report on validation data:
  if (any(section_table$external_validation)) {
    logger_message(
      "Setup report: Validation is external.",
      indent = message_indent,
      verbose = verbose
    )
    
  } else {
    logger_message(
      "Setup report: Validation is internal only.",
      indent = message_indent,
      verbose = verbose
    )
  }
  
  # Report on model building and feature selection
  if (any(section_table$vimp * section_table$train)) {
    main_message <- "Setup report: Feature selection and model building on"
    
    # Iteratively append message
    dt_sub <- section_table[vimp == TRUE & train == TRUE, ]
    curr_ref_data_id <- dt_sub$main_data_id[1L]
    while (curr_ref_data_id > 0L) {
      dt_sub <- section_table[main_data_id == curr_ref_data_id, ]
      
      if (dt_sub$perturb_method[1L] == "main") {
        main_message <- c(
          main_message,
          "the training data."
        )
        
      } else if (dt_sub$perturb_method[1L] %in% c("limited_bootstrap", "full_bootstrap")) {
        main_message <- c(
          main_message,
          paste0(dt_sub$perturb_n_rep[1L], " bootstraps of")
        )
        
      } else if (dt_sub$perturb_method[1L] == "cross_val") {
        main_message <- c(
          main_message,
          paste0(
            dt_sub$perturb_n_rep[1L], " repetitions of ",
            dt_sub$perturb_n_folds, "-fold cross validation of"
          )
        )
      
      } else if (dt_sub$perturb_method[1L] == "loocv") {
        main_message <- c(
          main_message,
          "folds of leave-one-out-cross-validation of"
        )
        
      } else if (dt_sub$perturb_method[1L] == "imbalance_partition") {
        main_message <- c(
          main_message,
          "class-balanced partitions of"
        )
      } 
      
      curr_ref_data_id <- dt_sub$ref_data_id[1L]
    }
    
    logger_message(
      paste0(main_message, collapse = " "),
      indent = message_indent,
      verbose = verbose
    )
    
  } else {
    
    if (any(section_table$vimp)) {
      # Feature selection first
      main_message <- "Setup report: Feature selection on"
      
      # Iteratively append message
      dt_sub <- section_table[vimp == TRUE, ]
      curr_ref_data_id <- dt_sub$main_data_id[1L]
      
      while (curr_ref_data_id > 0L) {
        
        dt_sub <- section_table[main_data_id == curr_ref_data_id, ]
        
        if (dt_sub$perturb_method[1L] == "main") {
          main_message <- c(
            main_message,
            "the training data."
          )
          
        } else if (dt_sub$perturb_method[1L] %in% c("limited_bootstrap", "full_bootstrap")) {
          main_message <- c(
            main_message,
            paste0(dt_sub$perturb_n_rep[1L], " bootstraps of")
          )
          
        } else if (dt_sub$perturb_method[1L] == "cross_val") {
          main_message <- c(
            main_message,
            paste0(
              dt_sub$perturb_n_rep[1L], " repetitions of ",
              dt_sub$perturb_n_folds, "-fold cross validation of"
            )
          )
          
        } else if (dt_sub$perturb_method[1L] == "loocv") {
          main_message <- c(
            main_message,
            "folds of leave-one-out-cross-validation of"
          )
          
        } else if (dt_sub$perturb_method[1L] == "imbalance_partition") {
          main_message <- c(
            main_message,
            "class-balanced partitions of"
          )
        }
        
        curr_ref_data_id <- dt_sub$ref_data_id[1L]
      }
      
      logger_message(
        paste0(main_message, collapse = " "),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    # Model building second
    main_message <- "Setup report: Model building on"
    
    # Iteratively append message
    dt_sub <- section_table[train == TRUE, ]
    curr_ref_data_id <- dt_sub$main_data_id[1L]
    
    while (curr_ref_data_id > 0L) {
      
      dt_sub <- section_table[main_data_id == curr_ref_data_id, ]
      
      if (dt_sub$perturb_method[1L] == "main") {
        main_message <- c(
          main_message,
          "the training data."
        )
        
      } else if (dt_sub$perturb_method[1L] %in% c("limited_bootstrap", "full_bootstrap")) {
        main_message <- c(
          main_message,
          paste0(dt_sub$perturb_n_rep[1L], " bootstraps of")
        )
        
      } else if (dt_sub$perturb_method[1L] == "cross_val") {
        main_message <- c(
          main_message,
          paste0(
            dt_sub$perturb_n_rep[1L], " repetitions of ",
            dt_sub$perturb_n_folds, "-fold cross validation of"
          )
        )
        
      } else if (dt_sub$perturb_method[1L] == "loocv") {
        main_message <- c(
          main_message,
          "folds of leave-one-out-cross-validation of"
        )
        
      } else if (dt_sub$perturb_method[1L] == "imbalance_partition") {
        main_message <- c(
          main_message,
          "class-balanced partitions of"
        )
      }
      
      curr_ref_data_id <- dt_sub$ref_data_id[1L]
    }
    
    logger_message(
      paste0(main_message, collapse = " "),
      indent = message_indent,
      verbose = verbose
    )
  }
}



.check_experimental_design_section_table <- function(section_table) {
  if (sum(section_table$vimp) > 1L) {
    ..error(
      paste0(
        "The fs component for variable importance computation may only be used once ",
        "in the experimental design."
      ),
      error_class = "input_argument_error"
    )
  }
  
  if (sum(section_table$train) > 1L) {
    ..error(
      paste0(
        "The mb component for training may only be used once ",
        "in the experimental design."
      ),
      error_class = "input_argument_error"
    )
  }
  
  if (sum(section_table$train) == 0L) {
    ..error(
      paste0(
        "The mb component for training must appear in the ",
        "experimental design. It was not found."
      ),
      error_class = "input_argument_error"
    )
  }
  
  if (sum(section_table$external_validation) > 1L) {
    ..error(
      paste0(
        "The ev component for external validation can only appear once ",
        "in the experimental design."
      ),
      error_class = "input_argument_error"
    )
  }
  
  return(invisible(TRUE))
}



.get_run_table_from_experiment_setup <- function(
    data_id,
    experiment_setup
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  perturbation_level <- main_data_id <- NULL
  
  # Get the data id chain.
  data_id_chain <- reference_data_id <- data_id
  while (reference_data_id != 0L) {
    reference_data_id <- experiment_setup[main_data_id == reference_data_id, ]$ref_data_id
    if (reference_data_id == 0L) break
    data_id_chain <- c(data_id_chain, reference_data_id)
  }
  
  # Reconstruct run table.
  run_table <- data.table::copy(
    experiment_setup[main_data_id %in% data_id_chain, ]
  )[order(perturbation_level)]
  
  return(run_table)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

setClass(
  "familiarMutualInformationVimp",
  contains = "familiarVimpMethod"
)

setClass(
  "familiarUnivariateMutualInfoVimp",
  contains = "familiarMutualInformationVimp"
)

setClass(
  "familiarMultivariateMutualInfoVimp",
  contains = "familiarMutualInformationVimp",
  prototype = methods::prototype(multivariate = TRUE)
)

setClass(
  "familiarMultivariateInfoVimpTestSingle",
  contains = "familiarMultivariateMutualInfoVimp"
)


# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarMultivariateMutualInfoVimp"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()
    
    # Set the required package
    .Object@multivariate <- TRUE
    
    return(.Object)
  }
)



.get_available_univariate_mutual_information_vimp_method <- function(show_general = TRUE) {
  return("mim")
}

.get_available_multivariate_mutual_information_vimp_method <- function(show_general = TRUE) {
  return(c("mifs", "mrmr"))
}

.get_available_multivariate_mutual_information_test_vimp_method <- function(show_general = TRUE) {
  return(c("familiarMultivariateInfoVimpTestSingle"))
}



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMutualInformationVimp"),
  function(object, ...) {
    
    if (object@outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(TRUE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarMutualInformationVimp"),
  function(object, data = NULL, ...) {
    return(list())
  }
)



# ..vimp (univariate) ----------------------------------------------------------
setMethod(
  "..vimp", 
  signature(object = "familiarUnivariateMutualInfoVimp"),
  function(object, data, ...) {
    
    if (is_empty(data)) return(callNextMethod())
    
    # Identify feature columns.
    feature_columns <- get_feature_columns(data)
    
    # Calculate mutual information
    mutual_information <- .compute_mutual_information(
      data = data, 
      features = feature_columns
    )
    
    # Create variable importance object.
    vimp_object <- methods::new(
      "vimpTable",
      vimp_table = data.table::data.table(
        "score" = mutual_information,
        "name" = feature_columns
      ),
      score_aggregation = "max",
      invert = TRUE
    )
    
    return(vimp_object)
  }
)


# ..vimp (multivariate) --------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarMultivariateMutualInfoVimp"),
  function(object, data, ...) {
    # - mifs: Mutual information feature selection using greedy search (Battiti
    # 1994)
    # - mrmr: Minimum Redundancy Maximum Relevance feature selection using
    # greedy search (Peng 2005)
    
    # Suppress NOTES due to non-standard evaluation in data.table
    objective_score <- available <- name <- mi_redundancy <- mi <- selected <- NULL
    
    if (is_empty(data)) return(callNextMethod())
    
    # Find feature columns in data table.
    feature_columns <- get_feature_columns(data)
    
    # Find signature features.
    signature_feature <- names(object@feature_info)[
      sapply(object@feature_info, is_in_signature)
    ]
    
    # Calculate mutual information.
    mutual_information <- .compute_mutual_information(data = data)
    
    # Setup mutual information table.
    mi_data <- data.table::data.table(
      "name" = feature_columns,
      "mi" = mutual_information,
      "mi_redundancy" = 0.0,
      "objective_score" = mutual_information,
      "selected" = FALSE,
      "available" = TRUE,
      "selection_step" = 0L
    )
    
    if (length(signature_feature) > 0L) {
      # Iteration counter.
      ii <- 1L
      
      for (current_feature in signature_feature) {
        # Mark the current signature feature. Note that we iterate over the
        # features, and update mutual information to ensure that we get to the
        # correct starting point for the non-signature variables.
        best_feature <- current_feature
        mi_data[
          name %in% best_feature,
          ":="(
            "selected" = TRUE,
            "available" = FALSE,
            "selection_step" = ii
          )
        ]
        
        # Update Iteration counter.
        ii <- ii + 1L
        
        # Determine the features still available.
        available_features <- mi_data[available == TRUE, ]$name
        if (length(available_features) == 0L) break
        
        # Also skip if this is the last signature feature -- this is to prevent
        # overlap with the main while-cycle below.
        if (current_feature == tail(signature_feature, n = 1L)) break
        
        # Calculate mutual information of available features and the
        # current signature feature.
        current_redundancy_mi <- .compute_mutual_information(
          data, 
          features = available_features, 
          with_feature = best_feature
        )
        
        # Update redundancy
        mi_data[
          available == TRUE,
          "mi_redundancy" := mi_redundancy + current_redundancy_mi
        ]
        
        if (object@vimp_method == "mifs") {
          # Update the objective score by subtracting the redundancy score. MIFS
          # uses beta=1 (Battiti 1994)
          mi_data[
            available == TRUE,
            "objective_score" := mi - mi_redundancy
          ]
          
          # Only consider features with a valid, non-negative optimisation
          # score, as redundancy is strictly increasing. Note that signature
          # features are never removed until they are selected.
          mi_data[
            objective_score < 0.0 & selected == FALSE & !name %in% signature_feature,
            "available" := FALSE
          ]
          
        } else if (object@vimp_method == "mrmr") {
          # Calculate optimisation score. MRMR uses beta=1/(ii-1), with ii-1 the
          # size of the selected feature set.
          mi_data[
            available == TRUE,
            "objective_score" := mi - mi_redundancy / (ii - 1L)
          ]
          
          # Only consider features with a valid, non-negative optimisation
          # score, as redundancy is strictly increasing. Note that signature
          # features are never removed until they are selected.
          mi_data[
            objective_score < 0.0 & selected == FALSE & !name %in% signature_feature,
            "available" := FALSE
          ]
          
        } else {
          ..error_reached_unreachable_code(paste0(
            "vimp,familiarMultivariateMutualInfoVimp: encountered unknown vimp_method: ",
            object@vimp_method
          ))
        }
      }
    } else {
      # Select initial feature and update mi_data.
      max_objective_score <- max(mi_data$objective_score)
      best_feature <- mi_data[objective_score == max_objective_score, ]$name[1L]
      mi_data[
        name %in% best_feature,
        ":="(
          "selected" = TRUE,
          "available" = FALSE,
          "selection_step" = 1L
        )
      ]
      
      # Determine the features that were not selected.
      available_features <- mi_data[available == TRUE, ]$name
      
      # Iteration counter.
      ii <- 2L
    }
    
    
    while (length(available_features) > 0L) {
      # Calculate mutual information of available features and the selected feature
      current_redundancy_mi <- .compute_mutual_information(
        data, 
        features = available_features, 
        with_feature = best_feature
      )
      
      # Update redundancy
      mi_data[
        available == TRUE, 
        "mi_redundancy" := mi_redundancy + current_redundancy_mi
      ]
      
      if (object@vimp_method == "mifs") {
        # Update the objective score by subtracting the redundancy score. MIFS
        # uses beta=1 (Battiti 1994)
        mi_data[
          available == TRUE,
          "objective_score" := mi - mi_redundancy
        ]
        
        # Only consider features with a valid, non-negative optimisation score,
        # as redundancy is strictly increasing.
        mi_data[
          objective_score < 0.0 & selected == FALSE,
          "available" := FALSE
        ]
        
      } else if (object@vimp_method == "mrmr") {
        # Calculate optimisation score. MRMR uses beta=1/(ii-1), with ii-1 the
        # size of the selected feature set
        mi_data[
          available == TRUE,
          "objective_score" := mi - mi_redundancy / (ii - 1L)
        ]
        
        # Only consider features with a valid, non-negative optimisation score,
        # as redundancy is strictly increasing.
        mi_data[
          objective_score < 0.0 & selected == FALSE,
          "available" := FALSE
        ]
        
      } else {
        ..error_reached_unreachable_code(paste0(
          "vimp,familiarMultivariateMutualInfoVimp: encountered unknown vimp_method: ",
          object@vimp_method
        ))
      }
      
      # If there are no more available features (i.e. due to negative objective
      # score), break from the loop.
      if (!any(mi_data$available)) break
      
      # Select best feature and update mi_data.
      max_objective_score <- max(mi_data[available == TRUE]$objective_score)
      best_feature <- mi_data[available == TRUE & objective_score == max_objective_score, ]$name[1L]
      mi_data[
        name %in% best_feature,
        ":="(
          "selected" = TRUE, 
          "available" = FALSE,
          "selection_step" = ii
        )
      ]
      
      # Determine the features still available.
      available_features <- mi_data[available == TRUE, ]$name
      
      # Increment iteration counter.
      ii <- ii + 1L
    }
    
    # Create variable importance table from the score table.
    vimp_table <- mi_data[selected == TRUE, c("name", "objective_score", "selection_step")]
    
    # Create variable importance object.
    vimp_object <- methods::new(
      "vimpTable",
      vimp_table = data.table::data.table(
        "score" = vimp_table$selection_step,
        "name" = vimp_table$name
      ),
      score_aggregation = "max",
      invert = FALSE
    )
    
    return(vimp_object)
  }
)



# ..vimp (multivariate test) ---------------------------------------------------
setMethod(
  "..vimp", 
  signature(object = "familiarMultivariateInfoVimpTestSingle"),
  function(object, data, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- NULL
    
    # Return the single most important feature.
    if (is_empty(data)) return(callNextMethod())
    
    # Identify feature columns.
    feature_columns <- get_feature_columns(data)
    
    # Calculate mutual information
    mutual_information <- .compute_mutual_information(
      data = data, 
      features = feature_columns
    )
    
    # Set vimp_table, and select only the best feature.
    vimp_table <- data.table::data.table(
      "score" = mutual_information,
      "name" = feature_columns
    )
    vimp_table <- vimp_table[score == max(score)]
    
    # Create variable importance object.
    vimp_object <- methods::new(
      "vimpTable",
      vimp_table = vimp_table,
      score_aggregation = "max",
      invert = TRUE
    )
    
    return(vimp_object)
  }
)



.compute_mutual_information <- function(
    data, 
    features = NULL, 
    with_feature = NULL
) {
  # If the comparison feature is NULL, compare with outcome.
  if (is.null(with_feature)) {
    # Set with_feature.
    with_feature <- get_outcome_columns(data)
    
    # Set the type of y.
    y_type <- switch(
      data@outcome_type,
      survival = "survival",
      binomial = "discrete",
      multinomial = "discrete",
      continuous = "continuous"
    )
    
    if (data@outcome_type %in% c("survival")) {
      y <- survival::Surv(
        time = data@data[["outcome_time"]],
        event = data@data[["outcome_event"]]
      )
      
    } else {
      y <- data@data[["outcome"]]
    }
    
  } else {
    # Isolate y-feature
    y <- data@data[[with_feature]]
    
    # Set the type of y.
    y_type <- ..determine_mi_variable_type(y)
  }
  
  # Get features.
  if (is.null(features)) {
    features <- get_feature_columns(data)
  }
  
  # Compute mutual information.
  mutual_information <- sapply(
    features,
    function(feature, data, y, y_type) {
      return(..compute_mutual_information(
        x = data@data[[feature]],
        y = y,
        y_type = y_type
      ))
    },
    data = data,
    y = y, 
    y_type = y_type
  )
  
  return(mutual_information)
}



..compute_mutual_information <- function(x, y, x_type = NULL, y_type = NULL) {
  # Determine variable type.
  if (is.null(x_type)) x_type <- ..determine_mi_variable_type(x)
  if (is.null(y_type)) y_type <- ..determine_mi_variable_type(y)
  
  # Convert discrete variables to factors.
  if (x_type == "discrete" && !is.factor(x)) x <- as.factor(x)
  if (y_type == "discrete" && !is.factor(y)) y <- as.factor(y)
  
  # Drop unused levels from factors.
  if (x_type == "discrete") x <- droplevels(x)
  if (y_type == "discrete") y <- droplevels(y)
  
  if (x_type == "continuous" && is.integer(x)) x <- as.numeric(x)
  if (y_type == "continuous" && is.integer(y)) y <- as.numeric(y)
  
  if (require_package("praznik", message_type = "silent")) {
    # Praznik-based computation.
    
    if (y_type == "survival") {
      # Mutual information for survival.
      mutual_information <- ..compute_mutual_information_any_survival(
        x = x,
        y = y
      )
      
    } else {
      # Use praznik package.
      mutual_information <- praznik::miScores(
        X = x,
        Y = y,
        threads = 1L
      )
    }
    
  } else {
    # Computation using internal methods.
    
    if (y_type == "survival") {
      # Mutual information for survival.
      mutual_information <- ..compute_mutual_information_any_survival(
        x = x, 
        y = y
      )
      
    } else if (y_type == "discrete") {
      if (x_type == "continuous") {
        # Mutual information for continuous and discrete variables.
        mutual_information <- ..compute_mutual_information_continuous_discrete(
          x = x, 
          y = y
        )
        
      } else if (x_type == "discrete") {
        # Mutual information for two discrete variables.
        mutual_information <- ..compute_mutual_information_discrete_discrete(
          x = x, 
          y = y
        )
        
      } else {
        ..error_reached_unreachable_code(
          "..compute_mutual_information: unknown variable type encountered for variable x."
        )
      }
      
    } else if (y_type == "continuous") {
      if (x_type == "continuous") {
        # Mutual information for two continuous variables.
        mutual_information <- ..compute_mutual_information_continuous_continuous(
          x = x, 
          y = y
        )
        
      } else if (x_type == "discrete") {
        # Mutual information for continuous and discrete variables. Note that
        # the order is reversed so that y is the discrete variable.
        mutual_information <- ..compute_mutual_information_continuous_discrete(
          x = y, 
          y = x
        )
        
      } else {
        ..error_reached_unreachable_code(
          "..compute_mutual_information: unknown variable type encountered for variable x."
        )
      }
    } else {
      ..error_reached_unreachable_code(
        "..compute_mutual_information: unknown variable type encountered for variable y."
      )
    }
  }
  
  # Check for invalid values.
  if (!is.finite(mutual_information)) mutual_information <- 0.0
  
  # Correct negative values.
  if (mutual_information < 0.0) mutual_information <- 0.0
  
  return(mutual_information)
}



..determine_mi_variable_type <- function(x) {
  # Determine the variable type.
  if (is.factor(x) || is.logical(x) || is.character(x)) {
    return("discrete")
  } else {
    return("continuous")
  }
}



..compute_mutual_information_any_survival <- function(x, y) {
  if (is.factor(x)) {
    # Expand x by using dummy coding.
    x <- stats::model.matrix(~ . - 1, data.frame(x))
    
    # Calculate concordance indices
    ci <- apply(
      x,
      2L, 
      function(x_data, y_data) {
        return(..compute_concordance_index(
          x = x_data,
          time = y_data[, 1L],
          event = y_data[, 2L]
        ))
      },
      y_data = y
    )
    
    # Calculate mutual information from the concordance indices
    mi <- -0.5 * log(1.0 - (2.0 * (ci - 0.5))^2.0 + 1.0E-10)
    
    # Find the maximum mutual information.
    mi <- max(mi)
    
  } else {
    # Compute the concordance index.
    ci <- ..compute_concordance_index(
      x = x, 
      time = y[, 1L],
      event = y[, 2L]
    )
    
    # Compute mutual information from the concordance index.
    mi <- -0.5 * log(1.0 - (2.0 * (ci - 0.5))^2.0 + 1.0E-10)
  }
  
  return(mi)
}



..compute_mutual_information_continuous_discrete <- function(x, y) {
  # MI is calculated based on linear approximation by De Jay et al. 2013;
  # doi:10.1093/bioinformatics/btt383 and Gel'fand, I.M.; Yaglom, A.M. (1957).
  # "Calculation of amount of information about a random function contained in
  # another such function". American Mathematical Society Translations: Series
  # 2. 12: 199-246
  
  # Expand y by using dummy coding. x is the continuous variable.
  y <- stats::model.matrix(~ . - 1, data.frame(y))
  
  # Calculate mutual information using correlation.
  mi <- apply(
    y,
    2L,
    function(y, x) {
      return(
        -0.5 * log(1.0 - stats::cor(x = x, y = y, method = "spearman")^2.0 + 1.0E-10)
      )
    },
    x = x
  )
  
  # Select the maximum mutual information.
  mi <- max(mi)
  
  return(mi)
}



..compute_mutual_information_discrete_discrete <- function(x, y) {
  # Direct, exact, calculation of mutual information for binomial and
  # multinomial outcomes Mutual information is defined as: I(X,Y) = H(X) -
  # H(X|Y) = H(Y) - H(Y|X) = H(X) + H(Y) - H(X,Y), and we calculate entropies
  
  # Calculate probability distributions
  p <- ..compute_mutual_information_conditional_probability(x = x, y = y)
  
  # Calculate entropies for marginal and joint distributions
  h_j <- -sum(p$prob_j$prob_j * log(p$prob_j$prob_j + 1.0E-10))
  h_k <- -sum(p$prob_k$prob_k * log(p$prob_k$prob_k + 1.0E-10))
  h_jk <- -sum(p$prob_kj$prob_kj * log(p$prob_kj$prob_kj + 1.0E-10))
  
  # Compute mutual information
  mi <- h_j + h_k - h_jk
  
  return(mi)
}



..compute_mutual_information_continuous_continuous <- function(x, y) {
  # MI is calculated based on linear approximation by De Jay et al. 2013;
  # doi:10.1093/bioinformatics/btt383 and Gel'fand, I.M.; Yaglom, A.M. (1957).
  # "Calculation of amount of information about a random function contained in
  # another such function". American Mathematical Society Translations: Series
  # 2. 12: 199-246
  
  return(-0.5 * log(1.0 - stats::cor(x = x, y = y, method = "spearman")^2.0 + 1.0E-10))
}



..compute_mutual_information_conditional_probability <- function(x, y) {
  # Suppress NOTES due to non-standard evaluation in data.table
  value <- NULL
  
  # Total number of training instances
  n <- length(x)
  
  # Create data table from data
  data <- data.table::data.table(
    "value" = x, 
    "class" = y
  )
  
  # p(k,j): joint probability for the combination of class k and value j
  p_kj <- data[, list(prob_kj = .N / n), by = list(value, class)]
  
  # p(k): marginal probability for class k
  p_k <- data[, list(prob_k = .N / n), by = class]
  
  # p(j): marginal probability for value j
  p_j <- data[, list(prob_j = .N / n), by = value]
  
  return(list(
    "prob_kj" = p_kj,
    "prob_k" = p_k, 
    "prob_j" = p_j
  ))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# add_data_element_identifier methods ------------------------------------------

## add_data_element_identifier (list) ------------------------------------------
setMethod(
  "add_data_element_identifier",
  signature(x = "list"),
  function(x, ...) {
    # Add identifier to every data element.
    data_elements <- unlist(lapply(x, add_data_element_identifier, ...))
    
    return(data_elements)
  }
)



## add_data_element_identifier (familiarDataElement) ---------------------------
setMethod(
  "add_data_element_identifier",
  signature(x = "familiarDataElement"),
  function(x, ...) {
    
    # Get dots, which contains the identifier to be set.
    dots <- list(...)
    
    if (length(dots) > 1L) {
      ..error_reached_unreachable_code(
        "add_data_element_identifier: can only add one identifier at a time."
      )
    }
    
    # Find the name of the identifier.
    identifier_name <- names(dots)
    
    # Iterate over values and create a separate data element for each value.
    data_elements <- lapply(
      dots[[identifier_name]],
      function(value, x, identifier_name) {
        
        # Find the list of previous identifiers, or create a new one.
        identifier_list <- x@identifiers
        if (is.null(identifier_list)) identifier_list <- list()
        
        # Add value to identifier list.
        identifier_list[[identifier_name]] <- value
        
        # Combine old and new identifiers.
        x@identifiers <- identifier_list
        
        return(x)
      },
      x = x,
      identifier_name = identifier_name
    )
    
    return(data_elements)
  }
)



## add_data_element_identifier (general) ---------------------------------------
setMethod(
  "add_data_element_identifier",
  signature(x = "ANY"),
  function(x, ...) {
    return(NULL)
  }
)


# add_data_element_bootstrap methods -------------------------------------------

## add_data_element_bootstrap (list) -------------------------------------------
setMethod(
  "add_data_element_bootstrap",
  signature(x = "list"),
  function(
    x,
    n_bootstraps, 
    n_instances, 
    bootstrap_seed_offset, 
    ...
  ) {
    
    if (n_bootstraps > 0L) {
      # Repeat elements.
      data_element <- rep(x, each = n_bootstraps)
      
      # Set bootstrap to TRUE
      bootstrap <- rep(TRUE, times = length(x) * n_bootstraps)
      
      # Set the seeds for the bootstraps to allow for reproducibility.
      bootstrap_seed <- rep(seq(n_bootstraps) + bootstrap_seed_offset, times = length(x))
      
      # Iterate over elements to check whether a point estimate should
      # be computed in addition.
      for (current_element in x) {
        if (
          current_element@detail_level %in% c("ensemble", "model") &&
          current_element@estimation_type %in% c("bci", "bootstrap_confidence_interval")
        ) {
          # Add a new element that estimates the point estimate.
          new_element <- current_element
          new_element@estimation_type <- "point"
          
          # Add the element to the list of elements.
          data_element <- c(data_element, new_element)
          bootstrap <- c(bootstrap, FALSE)
          bootstrap_seed <- c(bootstrap_seed, NA_integer_)
        }
      }
      
    } else {
      # Use the list x of data_elements.
      data_element <- x
      
      # No bootstraps need to be created.
      bootstrap <- rep(FALSE, times = length(x))
      
      # No seed is set
      bootstrap_seed <- rep(NA_integer_, times = length(x))
    }
    
    return(list(
      "data_element" = data_element,
      "bootstrap" = bootstrap,
      "seed" = bootstrap_seed
    ))
  }
)



## add_data_element_bootstrap (familiarDataElement) ----------------------------
setMethod(
  "add_data_element_bootstrap",
  signature(x = "familiarDataElement"),
  function(x, ...) {
    
    return(add_data_element_bootstrap(x = list(x), ...))
  }
)



# .identifier_as_data_attribute ------------------------------------------------
setMethod(
  ".identifier_as_data_attribute",
  signature(x = "familiarDataElement"),
  function(
    x, 
    identifier, 
    as_grouping_column = TRUE
  ) {
    
    if (length(identifier) == 0L) {
      ..error_reached_unreachable_code(
        ".identifier_as_data_attribute: Cannot pass an empty identifier."
      )
    }
    
    # If an "all" value is passed (e.g. during export), all identifiers are
    # added to the data.
    if (any(identifier == "all")) identifier <- names(x@identifiers)
    
    # Determine which of the identifiers is actually present. If none are
    # present, return x.
    identifier_present <- intersect(identifier, names(x@identifiers))
    if (length(identifier_present) == 0L) return(x)
    
    if (as_grouping_column) {
      x@grouping_column <- unique(c(x@grouping_column, identifier_present))
    }
    
    # Determine the indices of the selected list elements.
    identifier_index <- which(names(x@identifiers) %in% identifier_present)
    
    # Find values.
    identifier_values <- x@identifiers[identifier_index]
    
    # Remove identifiers from list.
    x@identifiers[identifier_index] <- NULL
    
    if (data.table::is.data.table(x@data)) {
      # The data element is a data.table.
      
      # Make a local copy to avoid updating by reference.
      x@data <- data.table::copy(x@data)
      
      # Iterate over identifier names.
      for (id_name in names(identifier_values)) {
        # Add identifier to the dataset.
        data.table::set(
          x = x@data,
          j = id_name,
          value = identifier_values[[id_name]]
        )
      }
      
    } else if (is.list(x@data)) {
      # Determine the number of instances in x@data
      n_instances <- length(x@data[[1L]])
      
      new_data <- lapply(
        identifier_values,
        function(x, n) (rep(x, times = n)),
        n = n_instances
      )
      names(new_data) <- names(identifier_values)
      
      # Add identifiers to the list.
      x@data <- c(x@data, new_data)
    }
    
    return(x)
  }
)



# identify_element_sets methods ------------------------------------------------

## identify_element_sets (list) ------------------------------------------------
setMethod(
  "identify_element_sets",
  signature(x = "list"),
  function(x, drop_identiers = NULL, ...) {
    
    # Check that that the list is not empty.
    if (is_empty(x)) return(NULL)
    
    # Check for empty elements.
    empty_elements <- sapply(x, is_empty)
    if (all(empty_elements)) return(NULL)
    
    # Iterate over list.
    id_table <- mapply(
      identify_element_sets,
      x = x,
      ii = seq_along(x),
      MoreArgs = list(...),
      SIMPLIFY = FALSE
    )
    
    # Combine to table and add group ids and model ids.
    id_table <- data.table::rbindlist(id_table, use.names = TRUE, fill = TRUE)
    
    # Drop identifiers.
    if (!is.null(drop_identiers)) {
      
      # Check that 
      if (!all(drop_identiers %in% colnames(id_table))) {
        ..error(paste0(
          "One or more identifiers to be dropped were not found in the table with identifiers: ",
          paste_s(setdiff(drop_identiers, colnames(id_table)))
        ))
      }
      
      # Drop identifiers
      id_table[, (drop_identiers) := NULL]
      
      # Keep unique entries.
      id_table <- unique(id_table)
    }
    
    # Add group identifier.
    id_table[, "group_id" := .GRP, by = c(colnames(id_table))] 
    
    # Add element identifier.
    id_table[, "element_id" := .I]
    
    return(id_table)
  }
)



## identify_element_sets (familiarDataElement) ---------------------------------
setMethod(
  "identify_element_sets",
  signature(x = "familiarDataElement"),
  function(
    x,
    ii,
    ignore_estimation_type = FALSE,
    ignore_grouping_column = TRUE, 
    ignore_list_identifier = TRUE, 
    ...
  ) {
    
    # Get the identifiers and the detail level and combine to a list.
    id_list <- c(
      x@identifiers,
      list(
        "detail_level" = x@detail_level,
        "object_class" = class(x)[1L]
      )
    )
    
    # Add the estimation type if it is not to be ignored.
    if (!ignore_estimation_type) {
      id_list <- c(id_list, list("estimation_type" = x@estimation_type))
    } 
    
    # Add data from grouping columns, if they are not to be ignored.
    if (!ignore_grouping_column && !is.null(x@grouping_column)) {
      id_list <- c(id_list, unique(x@data[, mget(x@grouping_column)]))
    }
    
    # Add list identifier.
    if (!ignore_list_identifier) {
      id_list <- c(id_list, "list_id" = ii)
    }
    
    return(data.table::as.data.table(id_list))
  }
)



## identify_element_sets (NULL) ------------------------------------------------
setMethod(
  "identify_element_sets",
  signature(x = "NULL"),
  function(x, ignore_estimation_type = FALSE, ...) {
    return(NULL)
  }
)



# merge_data_elements methods --------------------------------------------------

## merge_data_elements (list) --------------------------------------------------
setMethod(
  "merge_data_elements",
  signature(x = "list"),
  function(
    x,
    ...
  ) {
    
    # Check that the list is not empty.
    if (is_empty(x)) return(NULL)
    
    # Flatten (nested) lists.
    x <- unlist(x)
    if (!is.list(x)) x <- list(x)
    
    # Check that at least one of the data elements in the list is not empty.
    if (all(sapply(x, is_empty))) return(NULL)
    
    # Create return list for data elements.
    data_element <- list()
    
    # Determine class of all elements
    element_classes <- sapply(x, class)
    
    # Iterate over unique classes.
    for (element_class in unique(element_classes)) {
      
      # Continue to next element class if all current data elements are empty.
      if (all(sapply(x[which(element_classes == element_class)], is_empty))) next
      
      # Create a proto data element to avoid having to pass larger objects
      # than required.
      proto_data_element <- x[which(element_classes == element_class)][[1L]]
      if (methods::.hasSlot(proto_data_element, "data")) {
        proto_data_element@data <- NULL
      }
      
      # Run familiarDataElement-specific analysis. This means that we pass
      # the prototype data element as x with the list of elements.
      data_element <- c(
        data_element,
        merge_data_elements(
          x = proto_data_element,
          x_list = x[which(element_classes == element_class)],
          ...
        )
      )
    }
    
    # Assign a NULL to empty data
    if (is_empty(data_element)) data_element <- NULL
    
    return(data_element)
  }
)



## merge_data_elements (familiarDataElement) -----------------------------------
setMethod(
  "merge_data_elements",
  signature(x = "familiarDataElement"),
  function(
    x,
    x_list,
    as_data = NULL,
    as_grouping_column = TRUE,
    force_data_table = FALSE,
    ...
  ) {
    
    # Move identifiers from the identifiers attribute to the data attribute. The
    # primary reason for doing so is to group and merge similar elements, byt
    # e.g. from different models.
    if (!is.null(as_data)) {
      x_list <- lapply(
        x_list,
        .identifier_as_data_attribute,
        identifier = as_data,
        as_grouping_column = as_grouping_column
      )
    }
    
    # Identify items that can be joined.
    id_table <- identify_element_sets(x = x_list, ...)
    
    # Identify the element identifiers that should be grouped.
    grouped_data_element_ids <- lapply(
      split(id_table[, c("element_id", "group_id")], by = "group_id"),
      function(id_table) (id_table$element_id)
    )
    
    # List of data elements.
    data_elements <- list()
    
    for (current_group_data_element_ids in grouped_data_element_ids) {
      # Copy the first data element in the group and use it as a prototype.
      prototype_data_element <- x_list[[current_group_data_element_ids[1L]]]
      
      # Check contents of the data elements.
      any_is_data_table <- any(sapply(
        x_list[current_group_data_element_ids],
        function(x) (data.table::is.data.table(x@data))
      ))
      any_is_list <- any(sapply(
        x_list[current_group_data_element_ids],
        function(x) (rlang::is_bare_list(x@data))
      ))
      all_is_empty <- all(sapply(
        x_list[current_group_data_element_ids],
        function(x) (is_empty(x@data))
      ))
      
      if (any_is_data_table) {
        
        # Data attribute contains data.table.
        data_attribute <- lapply(
          x_list[current_group_data_element_ids],
          function(x) (x@data)
        )
        
        # Combine data attributes.
        data_attribute <- suppressWarnings(
          data.table::rbindlist(
            data_attribute,
            use.names = TRUE,
            fill = TRUE
          )
        )
        
        # Set data attribute.
        prototype_data_element@data <- data_attribute
        
      } else if (any_is_list) {
        
        # Data attribute contains data.table.
        element_names <- unique(unlist(lapply(
          x_list[current_group_data_element_ids],
          function(x) (names(x@data))
        )))
        
        # Iterate over different names in the list.
        data_attribute <- lapply(
          element_names,
          function(ii, x) {
            # Find values for the element.
            element_values <- unlist(lapply(
              x,
              function(x, ii) (x@data[[ii]]),
              ii = ii
            ))
            
            return(element_values)
          },
          x = x_list[current_group_data_element_ids]
        )
        
        # Set names.
        names(data_attribute) <- element_names
        
        # Force to data attribute.
        if (force_data_table) {
          data_attribute <- data.table::as.data.table(data_attribute)
        }
        
        # Set data attribute.
        prototype_data_element@data <- data_attribute
        
      } else if (all_is_empty) {
        # All data attributes are unset. We don't need to do anything.
        
      } else {
        # Unknown data type.
        ..error_reached_unreachable_code(paste0(
          "merge_data_elements,familiarDataElement: data attribute is neither ",
          "data.table, list, or empty."
        ))
      }
      
      # Add merged data element to the list.
      data_elements <- c(data_elements, list(prototype_data_element))
    }
    
    return(data_elements)
  }
)



## merge_data_elements (NULL) --------------------------------------------------
setMethod(
  "merge_data_elements",
  signature(x = "NULL"),
  function(x, ...) {
    return(NULL)
  }
)



# collect methods --------------------------------------------------------------

## collect (list)-------------------------------------------------------------
setMethod(
  "collect",
  signature(x = "list"),
  function(
    x,
    data_slot,
    identifiers = c("data_set", "vimp_method", "learner"),
    ...
  ) {
    
    # Collect from all 
    collected_data_elements <- lapply(
      x,
      collect,
      data_slot = data_slot,
      identifiers = identifiers
    )
    
    # Flatten (nested) lists.
    collected_data_elements <- unlist(collected_data_elements)
    if (!is.list(collected_data_elements)) {
      collected_data_elements <- list(collected_data_elements)
    }
    
    # Select unique elements. First identify which elements are present.
    id_table <- identify_element_sets(collected_data_elements)
    
    # Check that the identifier table is not empty.
    if (is_empty(id_table)) return(NULL)
    
    # Identify the first element id in each group.
    unique_elements <- sapply(
      split(id_table, by = "group_id"),
      function(x) (head(x$element_id, n = 1L))
    )
    
    # Keep unique elements.
    collected_data_elements <- collected_data_elements[unique_elements]
    
    return(collected_data_elements)
  }
)



## collect (familiarData)-------------------------------------------------------
setMethod(
  "collect",
  signature(x = "familiarData"),
  function(x, data_slot, identifiers, ...) {
    
    # Collect data elements.
    collected_data_elements <- slot(x, name = data_slot)
    
    if (is_empty(collected_data_elements)) return(NULL)
    
    # Add elements
    if ("data_set" %in% identifiers) {
      collected_data_elements <- add_data_element_identifier(
        x = collected_data_elements,
        data_set = x@name
      )
    }
    
    if ("vimp_method" %in% identifiers) {
      collected_data_elements <- add_data_element_identifier(
        x = collected_data_elements,
        vimp_method = x@vimp_method
      )
    }
    
    if ("learner" %in% identifiers) {
      collected_data_elements <- add_data_element_identifier(
        x = collected_data_elements,
        learner = x@learner
      )
    }
    
    return(collected_data_elements)
  }
)



# .export methods --------------------------------------------------------------

## .export (familiarCollection)-------------------------------------------------
setMethod(
  ".export",
  signature(x = "familiarCollection"),
  function(
    x,
    data_elements = NULL,
    data_slot = NULL,
    dir_path = NULL,
    type,
    subtype = NULL,
    object_class = NULL,
    export_collection = FALSE,
    ...
  ) {
    
    # Obtain the data elements from the attribute slot indicated by data_slot.
    if (!is.null(data_slot)) data_elements <- slot(x, name = data_slot)
    
    # Check that the list is not empty.
    if (is_empty(data_elements)) return(NULL)
    
    # Flatten (nested) lists.
    data_elements <- unlist(data_elements)
    if (!is.list(data_elements)) data_elements <- list(data_elements)
    
    # Check that at least one of the data elements in the list is not
    # empty.
    empty_elements <- sapply(data_elements, is_empty)
    if (all(empty_elements)) return(NULL)
    
    # Remove empty elements.
    data_elements <- data_elements[!empty_elements]
    
    # Determine class of all non-empty elements.
    element_classes <- sapply(data_elements, class)
    
    if (!is.null(object_class)) {
      # Keep data elements that correspond 
      data_elements <- data_elements[element_classes == object_class]
      if (is_empty(data_elements)) return(NULL)
      
    } else if (data.table::uniqueN(element_classes) > 2L) {
      ..error_reached_unreachable_code(paste0(
        ".export,familiarCollection: multiple data elements with different ",
        "classes found, whereas only one is expected."
      ))
    }
    
    # Merge and aggregate data. Note that the method is dispatched based on the
    # first object.
    data_element <- .export(
      x = data_elements[[1L]],
      x_list = data_elements,
      ...
    )
    
    # Apply labels.
    data_element <- .apply_labels(
      data = data_element,
      object = x
    )
    
    # Check that the data variable is not empty
    if (is_empty(data_element)) return(NULL)
    
    if (is.null(dir_path)) {
      # Export data.
      if (export_collection) {
        return(list(
          "collection" = x,
          "data" = data_element
        ))
        
      } else {
        return(data_element)
      }
      
    } else {
      # Export to file
      .export_to_file(
        data = data_element,
        object = x,
        dir_path = dir_path,
        type = type,
        subtype = subtype
      )
      
      if (export_collection) {
        return(list("collection" = x))
        
      } else {
        return(NULL)
      }
    }
  }
)



## .export (familiarDataElement) -----------------------------------------------
setMethod(
  ".export",
  signature(x = "familiarDataElement"),
  function(x, x_list, aggregate_results = FALSE, ...) {
    
    if (aggregate_results) {
      x_list <- .compute_data_element_estimates(x_list)
    }
    
    # Merge data elements.
    x <- merge_data_elements(
      x = x_list,
      as_data = "all",
      as_grouping_column = TRUE,
      force_data_table = TRUE
    )
    
    return(x)
  }
)


# .copy (familiarDataElement) --------------------------------------------------
setMethod(
  ".copy",
  signature(object = "familiarDataElement"),
  function(object) {
    slots_present <- methods::slotNames(object)
    
    for (current_slot in slots_present) {
      # Make deepcopy of data.tables present in the dataset.
      if (data.table::is.data.table(methods::slot(object, current_slot))) {
        methods::slot(object, current_slot) <- data.table::copy(methods::slot(object, current_slot))
      }
    }
    
    return(object)
  }
)


# .copy (dataObject) -----------------------------------------------------------
setMethod(
  ".copy",
  signature(object = "dataObject"),
  function(object) {
    slots_present <- methods::slotNames(object)
    
    for (current_slot in slots_present) {
      # Make deepcopy of data.tables present in the dataset.
      if (data.table::is.data.table(methods::slot(object, current_slot))) {
        methods::slot(object, current_slot) <- data.table::copy(methods::slot(object, current_slot))
      }
    }
    
    return(object)
  }
)


# .copy (null) -----------------------------------------------------------------
setMethod(
  ".copy",
  signature(object = "NULL"),
  function(object) {
    return(object)
  }
)


# extract_dispatcher (familiarEnsemble) ----------------------------------------

#'@title Internal function to dispatch extraction functions.
#'
#'@description This function provides a unified access point to extraction
#'  functions. Some of these functions require bootstrapping and result
#'  aggregation, which are handled here.
#'
#'@param FUN Extraction function or method to which data and parameters are
#'  dispatched.
#'@param object A `familiarEnsemble` object.
#'@param proto_data_element A `familiarDataElement` object, or an object that
#'  inherits from it.
#'@param aggregate_results A logical flag indicating whether results should be
#'  aggregated.
#'@param has_internal_bootstrap A logical flag that indicates whether `FUN` has
#'  internal bootstrapping capabilities.
#'
#'@inheritParams .extract_data
#'
#'@details This function first determines how many data points need to be
#'  evaluated to complete the desired estimation, i.e. 1 for point estimates, 20
#'  for bias-corrected estimates, and 20 / (1 - confidence level) for bootstrap
#'  confidence intervals.
#'
#'  Subsequently, we determine the number of models. This number is used to set
#'  external or internal clusters, the number of bootstraps, and to evaluate
#'  whether the estimation can be done in case `FUN` does not support
#'  bootstrapping.
#'
#'@return A list of `familiarDataElement` objects.
#'@md
#'@keywords internal
setMethod(
  "extract_dispatcher",
  signature(
    object = "familiarEnsemble",
    proto_data_element = "familiarDataElement"
  ),
  function(
    cl = NULL,
    FUN,
    object,
    proto_data_element,
    aggregate_results,
    has_internal_bootstrap,
    ...,
    message_indent = 0L,
    verbose = TRUE
  ) {
    
    # Check that any models were trained.
    if (!model_is_trained(object)) return(NULL)
    
    # Determine the number of instances we need to find the estimates.
    if (proto_data_element@estimation_type == "point") {
      n_instances <- 1L
      
    } else if (proto_data_element@estimation_type %in% c("bias_correction", "bc")) {
      n_instances <- 20L
      
    } else if (proto_data_element@estimation_type %in% c("bootstrap_confidence_interval", "bci")) {
      n_instances <- as.integer(ceiling(signif(20.0 / (1.00 - proto_data_element@confidence_level))))
    }
    
    # Determine the number of models we need to evaluate.
    if (proto_data_element@detail_level == "ensemble") {
      n_models <- 1L
      
    } else {
      n_models <- length(object@model_list)
    }
    
    # Check if the proposed analysis can be executed.
    if (
      !has_internal_bootstrap &&
      n_instances > 1L &&
      !(proto_data_element@detail_level == "hybrid" && n_models >= n_instances)
    ) {
      # The required number of instances cannot be created from models alone.
      
      # Add a message
      if (verbose) {
        message(paste0(
          "extract_dispatcher,familiarEnsemble,familiarDataElement: ",
          "too few models to compute confidence intervals."
        ))
      } 
      
      # Set the detail level to ensemble.
      if (proto_data_element@detail_level == "hybrid") {
        # Only one ensemble model is formed.
        proto_data_element@detail_level <- "ensemble"
        n_models <- 1L
      } 
      
      # Set the estimation type to point estimates.
      proto_data_element@estimation_type <- "point"
      n_instances <- 1L
    } 
    
    # Determine the number of bootstraps that should be computed internally.
    if (has_internal_bootstrap) {
      if (proto_data_element@detail_level == "hybrid") {
        n_bootstraps <- ceiling(n_instances / n_models)
        n_instances <- n_models * n_bootstraps
        
      } else {
        n_bootstraps <- n_instances
      }
      
    } else {
      n_bootstraps <- 0L
    }
    
    # If one bootstrap is required, that means no bootstraps are required.
    if (n_bootstraps <= 1L) n_bootstraps <- 0L
    
    
    # Determine the number of parallel cluster nodes.
    n_nodes <- length(cl)
    
    # Determine whether we should perform parallel processing across
    # models or internally.
    if (n_nodes > 1L) {
      if (n_models == 1L) {
        # No need to perform parallelisation across models, when there is only 1
        # model.
        parallel_external <- FALSE
        
      } else if (n_bootstraps == 0L) {
        # No need to perform internal parallelisation in case this is not
        # necessary. This check is hit when has_internal_bootstrap is false.
        parallel_external <- TRUE
        
      } else if (n_models >= 0.80 * n_nodes) {
        # Perform external parallelisation if the number of models would occupy
        # at least 80% of the nodes. This is because the parallel overhead in
        # any internal bootstrapping takes up more time.
        parallel_external <- TRUE
        
      } else if (n_models > n_bootstraps) {
        # Perform external parallelisation if the number of bootstraps would
        # underutilize the nodes compared to the number of nodes.
        parallel_external <- TRUE
        
      } else {
        parallel_external <- FALSE
      }
      
    } else {
      # Back-up when the number of nodes is 1 or none.
      parallel_external <- FALSE
    }
    
    # Message user concerning evaluation
    # - Type of estimation
    # - Which model(s)
    # - How many bootstraps each (if n models > 1), as well as total.
    # - If parallelisation takes place, and where.
    ..message_dispatcher_details(
      estimation_type = proto_data_element@estimation_type,
      detail_level = proto_data_element@detail_level,
      n_bootstraps = n_bootstraps,
      n_instances = n_instances,
      n_models = n_models,
      n_nodes = n_nodes,
      parallel_external = parallel_external,
      message_indent = message_indent,
      verbose = verbose
    )
    
    # Dispatch for ensemble models.
    if (proto_data_element@detail_level == "ensemble") {
      # Dispatch for results aggregated at the ensemble level.
      x <- .extract_dispatcher_ensemble(
        cl = cl,
        FUN = FUN,
        object = object,
        proto_data_element = proto_data_element,
        aggregate_results = aggregate_results,
        n_instances = n_instances,
        n_bootstraps = n_bootstraps,
        n_models = n_models,
        parallel_external = parallel_external,
        message_indent = message_indent,
        verbose = verbose,
        ...
      )
      
    } else if (proto_data_element@detail_level == "hybrid") {
      # Dispatch for results aggregated with hybrid details.
      x <- .extract_dispatcher_hybrid(
        cl = cl,
        FUN = FUN,
        object = object,
        proto_data_element = proto_data_element,
        aggregate_results = aggregate_results,
        n_instances = n_instances,
        n_bootstraps = n_bootstraps,
        n_models = n_models,
        parallel_external = parallel_external,
        message_indent = message_indent,
        verbose = verbose,
        ...
      )
      
    } else if (proto_data_element@detail_level == "model") {
      # Dispatch for results aggregated at the model level.
      x <- .extract_dispatcher_model(
        cl = cl,
        FUN = FUN,
        object = object,
        proto_data_element = proto_data_element,
        aggregate_results = aggregate_results,
        n_instances = n_instances,
        n_bootstraps = n_bootstraps,
        n_models = n_models,
        parallel_external = parallel_external,
        message_indent = message_indent,
        verbose = verbose,
        ...
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "extract_dispatcher: encountered an unknown detail_level attribute: ",
        proto_data_element@detail_level
      ))
    }
    
    return(x)
  }
)


# extract_dispatcher (familiarDataElementPredictionTable) ----------------------
setMethod(
  "extract_dispatcher",
  signature(
    object = "familiarDataElementPredictionTable",
    proto_data_element = "familiarDataElement"
  ),
  function(
    cl = NULL,
    FUN,
    object,
    proto_data_element,
    aggregate_results,
    has_internal_bootstrap,
    ...,
    message_indent = 0L,
    verbose = TRUE
  ) {
    
    # Determine the number of instances we need to find the estimates.
    if (proto_data_element@estimation_type == "point") {
      n_instances <- 1L
      
    } else if (proto_data_element@estimation_type %in% c("bias_correction", "bc")) {
      n_instances <- 20L
      
    } else if (proto_data_element@estimation_type %in% c("bootstrap_confidence_interval", "bci")) {
      n_instances <- as.integer(ceiling(signif(20.0 / (1.00 - proto_data_element@confidence_level))))
    }
    
    # Determine the number of models we need to evaluate. For prediction tables,
    # this is always 1.
    n_models <- 1L
    
    # Check if the proposed analysis can be executed.
    if (
      !has_internal_bootstrap && n_instances > 1L &&
      !(proto_data_element@detail_level == "hybrid" && n_models >= n_instances)
    ) {
      # The required number of instances cannot be created from models alone.
      
      # Add a message
      if (verbose) {
        message(paste0(
          "extract_dispatcher,familiarDataElementPredictionTable,familiarDataElement: ",
          "too few models to compute confidence intervals."
        ))
      } 
      
      # Set the detail level to ensemble.
      if (proto_data_element@detail_level == "hybrid") {
        # Only one ensemble model is formed.
        proto_data_element@detail_level <- "ensemble"
        n_models <- 1L
      } 
      
      # Set the estimation type to point estimates.
      proto_data_element@estimation_type <- "point"
      n_instances <- 1L
    } 
    
    # Determine the number of bootstraps that should be computed internally.
    if (has_internal_bootstrap) {
      if (proto_data_element@detail_level == "hybrid") {
        n_bootstraps <- ceiling(n_instances / n_models)
        n_instances <- n_models * n_bootstraps
        
      } else {
        n_bootstraps <- n_instances
      }
      
    } else {
      n_bootstraps <- 0L
    }
    
    # If one bootstrap is required, that means no bootstraps are required.
    if (n_bootstraps <= 1L) n_bootstraps <- 0L
    
    
    # Determine the number of parallel cluster nodes.
    n_nodes <- length(cl)
    
    # Determine whether we should perform parallel processing across
    # models or internally.
    if (n_nodes > 1L) {
      if (n_models == 1L) {
        # No need to perform parallelisation across models, when there is only 1
        # model.
        parallel_external <- FALSE
        
      } else if (n_bootstraps == 0L) {
        # No need to perform internal parallelisation in case this is not
        # necessary. This check is hit when has_internal_bootstrap is false.
        parallel_external <- TRUE
        
      } else if (n_models >= 0.80 * n_nodes) {
        # Perform external parallelisation if the number of models would occupy
        # at least 80% of the nodes. This is because the parallel overhead in
        # any internal bootstrapping takes up more time.
        parallel_external <- TRUE
        
      } else if (n_models > n_bootstraps) {
        # Perform external parallelisation if the number of bootstraps would
        # underutilize the nodes compared to the number of nodes.
        parallel_external <- TRUE
        
      } else {
        parallel_external <- FALSE
      }
      
    } else {
      # Back-up when the number of nodes is 1 or none.
      parallel_external <- FALSE
    }
    
    # Dispatch for ensemble models.
    if (proto_data_element@detail_level == "ensemble") {
      # Dispatch for results aggregated at the ensemble level.
      x <- .extract_dispatcher_ensemble(
        cl = cl,
        FUN = FUN,
        object = object,
        proto_data_element = proto_data_element,
        aggregate_results = aggregate_results,
        n_instances = n_instances,
        n_bootstraps = n_bootstraps,
        n_models = n_models,
        parallel_external = parallel_external,
        message_indent = message_indent,
        verbose = verbose,
        ...
      )
      
    } else if (proto_data_element@detail_level == "hybrid") {
      # Dispatch for results aggregated with hybrid details.
      x <- .extract_dispatcher_hybrid(
        cl = cl,
        FUN = FUN,
        object = object,
        proto_data_element = proto_data_element,
        aggregate_results = aggregate_results,
        n_instances = n_instances,
        n_bootstraps = n_bootstraps,
        n_models = n_models,
        parallel_external = parallel_external,
        message_indent = message_indent,
        verbose = verbose,
        ...
      )
      
    } else if (proto_data_element@detail_level == "model") {
      # Dispatch for results aggregated at the model level.
      x <- .extract_dispatcher_model(
        cl = cl,
        FUN = FUN,
        object = object,
        proto_data_element = proto_data_element,
        aggregate_results = aggregate_results,
        n_instances = n_instances,
        n_bootstraps = n_bootstraps,
        n_models = n_models,
        parallel_external = parallel_external,
        message_indent = message_indent,
        verbose = verbose,
        ...
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "extract_dispatcher: encountered an unknown detail_level attribute: ",
        proto_data_element@detail_level
      ))
    }
    
    return(x)
  }
)


# extract_dispatcher (dataObject) ----------------------------------------------
setMethod(
  "extract_dispatcher",
  signature(
    object = "dataObject",
    proto_data_element = "familiarDataElement"
  ),
  function(
    cl = NULL,
    FUN,
    object,
    proto_data_element,
    aggregate_results,
    has_internal_bootstrap,
    ...,
    message_indent = 0L,
    verbose = TRUE
  ) {
    
    # Determine the number of instances we need to find the estimates.
    if (proto_data_element@estimation_type == "point") {
      n_instances <- 1L
      
    } else if (proto_data_element@estimation_type %in% c("bias_correction", "bc")) {
      n_instances <- 20L
      
    } else if (proto_data_element@estimation_type %in% c("bootstrap_confidence_interval", "bci")) {
      n_instances <- as.integer(ceiling(signif(20.0 / (1.00 - proto_data_element@confidence_level))))
    }
    
    if (proto_data_element@detail_level != "ensemble") {
      ..error_reached_unreachable_code("Only ensemble-level detail levels are possible for extract_dispatcher with dataObject.")
    }
    
    # Determine the number of bootstraps that should be computed internally.
    n_bootstraps <- 0L
    if (has_internal_bootstrap) n_bootstraps <- n_instances
    
    # If one bootstrap is required, that means no bootstraps are required.
    if (n_bootstraps <= 1L) n_bootstraps <- 0L
    
    # Determine the number of parallel cluster nodes.
    parallel_external <- FALSE
    
    # Dispatch for results aggregated at the ensemble level.
    x <- .extract_dispatcher_ensemble(
      cl = cl,
      FUN = FUN,
      object = object,
      proto_data_element = proto_data_element,
      aggregate_results = aggregate_results,
      n_instances = n_instances,
      n_bootstraps = n_bootstraps,
      n_models = 1L,
      parallel_external = FALSE,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
    
    return(x)
  }
)



.extract_dispatcher_ensemble <- function(
    cl = NULL,
    FUN,
    object,
    proto_data_element,
    aggregate_results,
    n_instances,
    n_bootstraps,
    n_models,
    parallel_external,
    ...,
    verbose = FALSE
) {
  
  # Add ensemble model name.
  proto_data_element <- add_model_name(
    proto_data_element,
    object = object
  )
  
  # Set flag for interval aggregation.
  aggregate_internal <- aggregate_results &&
    n_instances == n_bootstraps &&
    n_bootstraps > 0L
  
  # Never perform outer-loop parallelisation when dispatching for ensemble-level
  # details.
  x <- FUN(
    cl = cl,
    object = object,
    bootstrap_seed_offset = 0L,
    proto_data_element = proto_data_element,
    aggregate_results = aggregate_internal,
    n_instances = n_instances,
    n_bootstraps = n_bootstraps,
    n_models = n_models,
    verbose = verbose,
    progress_bar = verbose && n_bootstraps > 1L,
    ...
  )
  
  # Pack to list.
  if (!is.list(x)) x <- list(x)
  
  # Merge data elements together.
  x <- merge_data_elements(x)
  
  # Aggregate results if required.
  if (aggregate_results) {
    x <- .compute_data_element_estimates(x)
  }
  
  return(x)
}



.extract_dispatcher_hybrid <- function(
    cl = NULL,
    FUN,
    object,
    proto_data_element,
    aggregate_results,
    n_instances,
    n_bootstraps,
    n_models,
    parallel_external,
    ...,
    verbose = FALSE
) {
  
  # Add ensemble model name.
  proto_data_element <- add_model_name(proto_data_element, object = object)
  
  if (parallel_external) {
    x <- fam_mapply(
      cl = cl,
      FUN = FUN,
      object = object@model_list,
      bootstrap_seed_offset = seq(
        from = 0L, 
        by = n_bootstraps,
        length.out = n_models
      ),
      MoreArgs = c(
        list(
          "cl" = NULL,
          "proto_data_element" = proto_data_element,
          "aggregate_results" = FALSE,
          "n_instances" = n_instances,
          "n_bootstraps" = n_bootstraps,
          "n_models" = n_models,
          "verbose" = verbose,
          "progress_bar" = verbose && length(object@model_list) == 1L && n_bootstraps > 1L
        ),
        list(...)
      ),
      progress_bar = verbose && length(object@model_list) > 1L,
      chopchop = TRUE
    )
    
  } else {
    x <- fam_mapply(
      cl = NULL,
      FUN = FUN,
      object = object@model_list,
      bootstrap_seed_offset = seq(
        from = 0L,
        by = n_bootstraps,
        length.out = n_models
      ),
      MoreArgs = c(
        list(
          "cl" = cl,
          "proto_data_element" = proto_data_element,
          "aggregate_results" = FALSE,
          "n_instances" = n_instances,
          "n_bootstraps" = n_bootstraps,
          "n_models" = n_models,
          "verbose" = verbose,
          "progress_bar" = verbose && n_models == 1L && n_bootstraps > 1L
        ),
        list(...)
      ),
      progress_bar = verbose && n_models > 1L
    )
  }
  
  # Merge data elements together. The model_name identifier gets added as data
  # instead, but not as a grouping column.
  x <- merge_data_elements(
    x,
    as_data = "model_name",
    as_grouping_column = FALSE
  )
  
  # Create point estimate from the data.
  if (proto_data_element@estimation_type %in% c("bootstrap_confidence_interval", "bci")) {
    x <- .add_point_estimate_from_elements(x)
  }
  
  # Aggregate results if required.
  if (aggregate_results) {
    x <- .compute_data_element_estimates(x)
  }
  
  return(x)
}



.extract_dispatcher_model <- function(
    cl = NULL,
    FUN,
    object,
    proto_data_element,
    aggregate_results,
    n_instances,
    n_bootstraps,
    n_models,
    parallel_external,
    ...,
    verbose = FALSE
) {
  
  # Create a list of data-elements.
  proto_data_element <- rep.int(
    list(proto_data_element),
    times = length(object@model_list)
  )
  
  # Add model (not ensemble) names to the prototype data elements.
  proto_data_element <- mapply(
    add_model_name,
    data = proto_data_element,
    object = object@model_list
  )
  
  # Set flag for interval aggregation.
  aggregate_internal <- aggregate_results && n_instances == n_bootstraps && n_bootstraps > 0L
  
  if (parallel_external) {
    x <- fam_mapply(
      cl = cl,
      FUN = FUN,
      object = object@model_list,
      bootstrap_seed_offset = seq(
        from = 0L,
        by = n_bootstraps,
        length.out = n_models
      ),
      proto_data_element = proto_data_element,
      MoreArgs = c(
        list(
          "cl" = NULL,
          "aggregate_results" = aggregate_internal,
          "n_instances" = n_instances,
          "n_bootstraps" = n_bootstraps,
          "n_models" = n_models,
          "verbose" = verbose,
          "progress_bar" = verbose && n_models == 1L && n_bootstraps > 1L
        ),
        list(...)
      ),
      progress_bar = verbose & length(object@model_list) > 1L,
      chopchop = TRUE
    )
    
  } else {
    x <- fam_mapply(
      cl = NULL,
      FUN = FUN,
      object = object@model_list,
      bootstrap_seed_offset = seq(
        from = 0L,
        by = n_bootstraps,
        length.out = n_models
      ),
      proto_data_element = proto_data_element,
      MoreArgs = c(
        list(
          "cl" = cl,
          "aggregate_results" = aggregate_internal,
          "n_instances" = n_instances,
          "n_bootstraps" = n_bootstraps,
          "n_models" = n_models,
          "verbose" = verbose,
          "progress_bar" = verbose && n_models == 1L && n_bootstraps > 1L
        ),
        list(...)
      ),
      progress_bar = verbose && length(object@model_list) > 1L
    )
  }
  
  # Merge data elements together.
  x <- merge_data_elements(x)
  
  # Aggregate results if required.
  if (aggregate_results) {
    x <- .compute_data_element_estimates(x)
  }
  
  return(x)
}



# .add_point_estimate_from_elements methods ------------------------------------

## .add_point_estimate_from_elements (list) ------------------------------------
setMethod(
  ".add_point_estimate_from_elements",
  signature(x = "list"),
  function(x, ...) {
    # Create return list for data elements.
    data_element <- list()
    
    # Determine class of all elements
    element_classes <- sapply(x, class)
    
    # Iterate over unique classes.
    for (element_class in unique(element_classes)) {
      
      # Create a proto data element to avoid having to pass larger objects
      # than required.
      proto_data_element <- x[which(element_classes == element_class)][[1L]]
      proto_data_element@data <- NULL
      
      # Run familiarDataElement-specific analysis. This means that we pass
      # the prototype data element as x with the list of elements.
      data_element <- c(
        data_element,
        .add_point_estimate_from_elements(
          x = proto_data_element,
          x_list = x[which(element_classes == element_class)],
          ...
        )
      )
    }
    
    # Check that any data elements were added.
    if (is_empty(data_element)) return(NULL)
    
    return(data_element)
  }
)


## .add_point_estimate_from_elements (familiarDataElement) ---------------------
setMethod(
  ".add_point_estimate_from_elements",
  signature(x = "familiarDataElement"),
  function(x, x_list = NULL, ...) {
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Find any unique elements that have not been aggregated and are not empty.
    id_table <- identify_element_sets(x, ignore_estimation_type = TRUE)
    
    # If no identifier table is generated, this means that the dataset is empty,
    # and a NULL should be returned.
    if (is_empty(id_table)) return(NULL)
    
    # Identify the element identifiers that should be grouped.
    grouped_data_element_ids <- lapply(
      split(id_table[, c("element_id", "group_id")], by = "group_id"),
      function(id_table) (id_table$element_id)
    )
    
    # List of data elements.
    data_elements <- list()
    
    for (current_group_data_element_ids in grouped_data_element_ids) {
      
      # Check that there is no point estimate present in the current table.
      any_point_estimate <- any(sapply(
        x[current_group_data_element_ids],
        function(x) (x@estimation_type == "point")
      ))
      
      if (any_point_estimate) next
      
      # Set conversion back to list, in case this is required.
      data_as_list <- FALSE
      
      # Copy the first data element in the group and use it as a prototype.
      prototype_data_element <- x[[current_group_data_element_ids[1L]]]
      
      # Set point estimate.
      prototype_data_element@estimation_type <- "point"
      
      # Check if all data are empty.
      all_data_empty <- all(sapply(
        x[current_group_data_element_ids],
        function(x) (is_empty(x@data))
      ))
      
      if (all_data_empty) {
        # Add empty element to data_elements and skip to the next
        data_elements <- c(data_elements, list(prototype_data_element))
        next
      }
      
      # Extract the data.
      any_is_data_table <- any(sapply(
        x[current_group_data_element_ids],
        function(x) (data.table::is.data.table(x@data))
      ))
      any_is_list <- any(sapply(
        x[current_group_data_element_ids],
        function(x) (rlang::is_bare_list(x@data))
      ))
      
      if (any_is_data_table) {
        # Data attribute contains data.table.
        data <- lapply(
          x[current_group_data_element_ids],
          function(x) (x@data)
        )
        
      } else if (any_is_list) {
        # Convert all lists to data tables.
        data <- lapply(
          x[current_group_data_element_ids],
          function(x) (data.table::as.data.table(x@data))
        )
        
        # Convert back to list in the end.
        data_as_list <- TRUE
      }
      
      # Combine data attributes.
      data <- data.table::rbindlist(
        data,
        use.names = TRUE,
        fill = TRUE
      )
      
      if (length(prototype_data_element@grouping_column) > 0L) {
        # Compute the mean value as point estimate.
        data <- data[
          ,
          lapply(.SD, get_estimate, na.rm = TRUE),
          by = c(prototype_data_element@grouping_column),
          .SDcols = c(prototype_data_element@value_column)
        ]
        
      } else {
        data <- data[
          ,
          lapply(.SD, get_estimate, na.rm = TRUE),
          .SDcols = c(prototype_data_element@value_column)
        ]
      }
      
      # Convert to list again, if necessary.
      if (data_as_list) data <- as.list(data)
      
      # Update data attribute with point estimate.
      prototype_data_element@data <- data
      
      # Add merged data element to the list.
      data_elements <- c(data_elements, list(prototype_data_element))
    }
    
    return(c(x, data_elements))
  }
)


## .add_point_estimate_from_elements (NULL) ------------------------------------
setMethod(
  ".add_point_estimate_from_elements",
  signature(x = "NULL"),
  function(x, ...) {
    return(NULL)
  }
)



# .compute_data_element_estimates methods --------------------------------------

## .compute_data_element_estimates (list) --------------------------------------
setMethod(
  ".compute_data_element_estimates",
  signature(x = "list"),
  function(x, ...) {
    
    # Create return list for data elements.
    data_element <- list()
    
    # Determine class of all elements
    element_classes <- sapply(x, class)
    
    # Iterate over unique classes.
    for (element_class in unique(element_classes)) {
      
      # Create a proto data element to avoid having to pass larger objects than
      # required. First we copy the first instance of the particular class.
      proto_data_element <- x[which(element_classes == element_class)][[1L]]
      
      # Check that the proto_data_element is not NULL, because NULL objects
      # cannot be parsed.
      if (is.null(proto_data_element)) next
      
      # Remove data from the proto data element.
      proto_data_element@data <- NULL
      
      # Run familiarDataElement-specific analysis. This means that we pass the
      # prototype data element as x with the list of elements.
      data_element <- c(
        data_element,
        .compute_data_element_estimates(
          x = proto_data_element,
          x_list = x[which(element_classes == element_class)],
          ...
        )
      )
    }
    
    if (is_empty(data_element)) return(NULL)
    
    return(data_element)
  }
)



## .compute_data_element_estimates (familiarDataElement) -----------------------
setMethod(
  ".compute_data_element_estimates",
  signature(x = "familiarDataElement"),
  function(x, x_list = NULL, ...) {
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Merge data.
    data_elements <- merge_data_elements(x = x)
    
    # Find any data elements that were already aggregated and keep these apart.
    is_aggregated <- sapply(
      data_elements,
      function(x) (x@is_aggregated)
    )
    
    if (all(is_aggregated)) return(data_elements)
    
    # Continue with unaggregated data elements.
    unaggregated_data_elements <- data_elements[!is_aggregated]
    data_elements <- data_elements[is_aggregated]
    
    # Find any unique elements that have not been aggregated.
    id_table <- identify_element_sets(
      unaggregated_data_elements,
      ignore_estimation_type = TRUE
    )
    
    # Identify the element identifiers that should be grouped.
    grouped_data_element_ids <- lapply(
      split(id_table[, c("element_id", "group_id")], by = "group_id"),
      function(id_table) (id_table$element_id)
    )
    
    # Aggregate unaggregated data.
    for (current_group_data_element_ids in grouped_data_element_ids) {
      
      # Select data elements corresponding to the current group.
      current_data_elements <- unaggregated_data_elements[current_group_data_element_ids]
      
      # Compute estimates.
      aggregated_data_element <- ..compute_data_element_estimates(
        x = current_data_elements,
        ...
      )
      
      if (is.null(aggregated_data_element)) next
      
      # Update the is_aggregated attribute.
      aggregated_data_element@is_aggregated <- TRUE
      
      # Add aggregated element.
      data_elements <- c(data_elements, list(aggregated_data_element))
    }
    
    if (is_empty(data_elements)) return(NULL)
    
    return(data_elements)
  }
)


## .compute_data_element_estimates (NULL) --------------------------------------
setMethod(
  ".compute_data_element_estimates",
  signature(x = "NULL"),
  function(x, ...) {
    return(NULL)
  }
)



# ..compute_data_element_estimates methods -------------------------------------

## ..compute_data_element_estimates (list) -------------------------------------
setMethod(
  "..compute_data_element_estimates",
  signature(x = "list"),
  function(x, ...) {
    
    # Create a proto data element to avoid having to pass larger objects than
    # required.
    proto_data_element <- x[[1L]]
    proto_data_element@data <- NULL
    
    return(..compute_data_element_estimates(
      x = proto_data_element,
      x_list = x,
      ...
    ))
  }
)


## ..compute_data_element_estimates (familiarDataElement) ----------------------
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElement"),
  function(x, x_list = NULL, ...) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    n_group <- NULL
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Identify the estimation types of the current data elements.
    estimation_type <- sapply(x, function(x) (x@estimation_type))
    
    if (any(sapply(x, is_empty))) {
      # Don't aggregate empty elements.
      y <- x[[1L]]
      
    } else if (any(estimation_type %in% c("bci", "bootstrap_confidence_interval"))) {
      
      # Check the number of elements.
      if (length(estimation_type) != 2L) {
        ..error_reached_unreachable_code(paste0(
          ".compute_data_element_estimates: exactly two data elements are ",
          "required for bootstrap confidence intervals."
        ))
      }
      
      if (!any(estimation_type %in% c("point"))) {
        ..error_reached_unreachable_code(paste0(
          ".compute_data_element_estimates: a point estimate is required for ",
          "bootstrap confidence intervals."
        ))
      }
      
      # Select point estimate.
      point_values <- data.table::as.data.table(
        x[estimation_type == "point"][[1L]]@data
      )
      point_values[, "estimation_type" := "point"]
      
      # Select bootstrap values
      bootstrap_values <- data.table::as.data.table(
        x[estimation_type %in% c("bci", "bootstrap_confidence_interval")][[1L]]@data
      )
      bootstrap_values[, "estimation_type" := "bootstrap_confidence_interval"]
      
      # Combine to single table.
      data <- data.table::rbindlist(
        list(point_values, bootstrap_values),
        use.names = TRUE,
        fill = TRUE
      )
      
      if (length(x[[1L]]@grouping_column) > 0L) {
        
        # Split table by grouping column and compute estimate and confidence
        # intervals.
        data <- lapply(
          split(data, by = x[[1L]]@grouping_column, drop = TRUE),
          ..compute_bootstrap_confidence_estimate,
          confidence_level = x[[1L]]@confidence_level,
          bootstrap_ci_method = x[[1L]]@bootstrap_ci_method,
          value_column = x[[1L]]@value_column,
          grouping_column = x[[1L]]@grouping_column
        )
        
        # Combine to single table
        data <- data.table::rbindlist(data, use.names = TRUE, fill = TRUE)
        
      } else {
        # Compute in absence of grouping columns.
        data <- ..compute_bootstrap_confidence_estimate(
          x = data,
          confidence_level = x[[1L]]@confidence_level,
          bootstrap_ci_method = x[[1L]]@bootstrap_ci_method,
          value_column = x[[1L]]@value_column
        )
      }
      
      # Update the data attribute.
      y <- x[estimation_type %in% c("bci", "bootstrap_confidence_interval")][[1L]]
      y@data <- data
      
      # Update value column
      y@value_column <- setdiff(names(y@data), y@grouping_column)
      
    } else if (any(estimation_type %in% c("bc", "bias_correction"))) {
      
      # Check the number of elements.
      if (length(estimation_type) != 1L) {
        ..error_reached_unreachable_code(paste0(
          ".compute_data_element_estimates: exactly one data element is required ",
          "for bias corrected estimates."
        ))
      }
      
      # Select values.
      bootstrap_values <- data.table::as.data.table(
        x[estimation_type %in% c("bc", "bias_correction")][[1L]]@data
      )
      
      if (length(x[[1L]]@grouping_column) > 0L) {
        # Split table by grouping column and compute bias corrected estimate.
        data <- lapply(
          split(bootstrap_values, by = x[[1L]]@grouping_column, drop = TRUE),
          ..compute_bias_corrected_estimate,
          value_column = x[[1L]]@value_column,
          grouping_column = x[[1L]]@grouping_column
        )
        
        # Combine to single table
        data <- data.table::rbindlist(
          data,
          use.names = TRUE,
          fill = TRUE
        )
        
      } else {
        # Compute in absence of grouping columns.
        data <- ..compute_bias_corrected_estimate(
          x = bootstrap_values,
          value_column = x[[1L]]@value_column
        )
      }
      
      # Update the data attribute.
      y <- x[[1L]]
      y@data <- data
      
      # Update value column
      y@value_column <- setdiff(names(y@data), y@grouping_column)
      
    } else if (any(estimation_type %in% c("point"))) {
      # This follows the same procedure as for bias-corrected estimates. For
      # ensemble and hybrid detail levels a single value needs to be generated.
      # However, in the case of hybrid detail level, a point estimate is created
      # for each model, and requires aggregation.
      
      # Check the number of elements.
      if (length(estimation_type) != 1L) {
        ..error_reached_unreachable_code(paste0(
          ".compute_data_element_estimates: exactly one data element is ",
          "required for point estimates."
        ))
      }
      
      # Find grouping columns.
      grouping_columns <- x[[1L]]@grouping_column
      if (length(grouping_columns) == 0L) grouping_columns <- NULL
      
      # Find value columns.
      value_columns <- x[[1L]]@value_column
      
      # Select values.
      bootstrap_values <- data.table::as.data.table(
        x[estimation_type %in% c("point")][[1L]]@data
      )[, mget(c(grouping_columns, value_columns))]
      
      # Refine a bit so that only those entries with multiple values for the
      # same grouping columns are aggregated. This can save a lot of time,
      # because the point estimate typically is determined only on a single run.
      bootstrap_values[, "n_group" := .N, by = grouping_columns]
      
      # Select data based on single/multiple entries. Keep only relevant
      # columns, namely grouping and value columns, to ensure that both
      # unique_values and bootstrap_values will be processed the same way.
      unique_values <- bootstrap_values[n_group == 1L, mget(c(grouping_columns, value_columns))]
      bootstrap_values <- bootstrap_values[n_group > 1L, mget(c(grouping_columns, value_columns))]
      
      if (is_empty(bootstrap_values)) {
        # Data are unique values.
        data <- unique_values
        
      } else if (length(grouping_columns) > 0L) {
        # Split table by grouping column and compute bias corrected estimate.
        data <- lapply(
          split(bootstrap_values, by = grouping_columns, drop = TRUE),
          ..compute_bias_corrected_estimate,
          value_column = value_columns,
          grouping_column = grouping_columns
        )
        
        # Combine to single table
        data <- data.table::rbindlist(
          c(list(unique_values), data),
          use.names = TRUE,
          fill = TRUE
        )
        
      } else {
        # Compute in absence of grouping columns.
        data <- ..compute_bias_corrected_estimate(
          x = bootstrap_values,
          value_column = value_columns
        )
        
        # Combine to single table
        data <- data.table::rbindlist(
          c(list(unique_values), list(data)),
          use.names = TRUE,
          fill = TRUE
        )
      }
      
      # Update the data attribute.
      y <- x[[1L]]
      y@data <- data
      
      # Update value column
      y@value_column <- setdiff(names(y@data), y@grouping_column)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        ".compute_data_element_estimates: unknown estimation type: ",
        paste_s(estimation_type)
      ))
    }
    
    return(y)
  }
)


## ..compute_data_element_estimates (NULL) -------------------------------------
setMethod(
  "..compute_data_element_estimates",
  signature(x = "NULL"),
  function(x, ...) {
    return(NULL)
  }
)



..compute_bootstrap_confidence_estimate <- function(
    x,
    value_column, 
    confidence_level,
    bootstrap_ci_method,
    grouping_column = NULL
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  estimation_type <- NULL
  
  # Construct NULL table.
  ci_estimate <- data.table::data.table()
  
  # Fill value list.
  value_list <- list()
  value_column_names <- character(0L)
  for (ii in seq_along(value_column)) {
    # Compute the estimate and its bootstrap confidence interval.
    temp_estimate <- ..bootstrap_ci(
      x = x[estimation_type == "bootstrap_confidence_interval"][[value_column[ii]]],
      x_0 = x[estimation_type == "point"][[value_column[ii]]],
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method
    )
    
    # Determine column names that should be assigned.
    if (length(temp_estimate) == 3L) {
      
      if (length(value_column) == 1L) {
        value_column_names <- c(
          value_column_names,
          value_column[ii], "ci_low", "ci_up"
        )
        
      } else {
        value_column_names <- c(
          value_column_names,
          value_column[ii],
          paste0(value_column[ii], "_ci_low"),
          paste0(value_column[ii], "_ci_up")
        )
      }
      
    } else {
      value_column_names <- c(
        value_column_names,
        value_column[ii]
      )
    }
    
    # Set to value list
    value_list[[ii]] <- temp_estimate
  }

  # Flatten list and then re-list
  value_list <- unlist(value_list, recursive = FALSE)
  if (!is.list(value_list)) value_list <- as.list(value_list)
  
  # Assign to table.
  ci_estimate[, c(value_column_names) := value_list]
  
  # Add in grouping columns, if any.
  if (length(grouping_column) > 0L) {
    ci_estimate <- cbind(
      head(x[, mget(grouping_column)], n = 1L),
      ci_estimate
    )
  }
  
  return(ci_estimate)
}



..compute_bias_corrected_estimate <- function(
    x,
    value_column,
    grouping_column = NULL
) {
  
  # Construct NULL table.
  bc_estimate <- data.table::data.table()
  
  # Fill value list.
  value_list <- list()
  for (ii in seq_along(value_column)) {
    value_list[[ii]] <- ..bootstrap_bias_correction(x = x[[value_column[ii]]])$median
  }
  
  # Assign to table.
  bc_estimate[, c(value_column) := value_list]
  
  # Add in grouping columns, if any.
  if (length(grouping_column) > 0L) {
    bc_estimate <- cbind(
      head(x[, mget(grouping_column)], n = 1L),
      bc_estimate
    )
  }
  
  return(bc_estimate)
}



..message_dispatcher_details <- function(
    estimation_type,
    detail_level,
    n_bootstraps,
    n_instances,
    n_models,
    n_nodes,
    parallel_external,
    message_indent,
    verbose
) {
  
  # Skip if the dispatcher is not verbose.
  if (!verbose) return(NULL)
  
  # Set the estimator string.
  estimator_str <- switch(
    estimation_type,
    "point" = "point estimate",
    "bc" = "bias-corrected estimate",
    "bias_correction" = "bias-corrected estimate",
    "bci" = "bias-corrected estimate with confidence interval",
    "bootstrap_confidence_interval" = "bias-corrected estimate with confidence interval"
  )
  
  # Set the detail level string.
  detail_level_str <- switch(
    detail_level,
    "ensemble" = "the ensemble model as a whole. ",
    "hybrid" = paste0(
      "the ensemble model from the ",
      ifelse(
        n_models > 1L,
        paste0(n_models, " underlying models. "),
        paste0("single underlying model. ")
      )
    ),
    "model" = paste0(
      ifelse(
        n_models > 1L,
        paste0("each of the ", n_models, " models"),
        paste0("the single model")
      ),
      " in the ensemble. "
    )
  )
  
  # Bootstraps.
  if (n_bootstraps > 0L) {
    bootstrap_str <- paste0(
      n_bootstraps, " bootstrap samples are obtained ",
      ifelse(
        n_models > 1L,
        paste0("for each model (total: ", n_instances, "). "),
        "in total. "
      )
    )
    
  } else {
    bootstrap_str <- ""
  }
  
  # Parallelisation.
  if (n_nodes > 1L) {
    if (parallel_external) {
      parallel_str <- paste0("Computation is parallelised over models.")
      
    } else {
      parallel_str <- paste0("Computation is parallelised over bootstraps.")
    }
    
  } else {
    parallel_str <- ""
  }
  
  logger_message(
    paste0(
      "Computing the ",
      estimator_str,
      " of the value(s) of interest for ",
      detail_level_str,
      bootstrap_str,
      parallel_str
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  return(invisible(NULL))
}


# .convert_value_to_grouping_column (familiarDataElement) ----------------------
setMethod(
  ".convert_value_to_grouping_column",
  signature(x = "familiarDataElement"),
  function(
    x, 
    new_grouping_column, 
    new_grouping_column_name, 
    new_value_column_name,
    ...
  ) {
    
    if (is_empty(x)) return(x)
    if (!data.table::is.data.table(x@data)) return(x)
    if (!any(new_grouping_column %in% x@value_column)) {
      rlang::warn("None of the intended grouping columns currently appear as value column.")
      return(x)
    }

    new_grouping_column <- intersect(new_grouping_column, x@value_column)
    remaining_value_column <- setdiff(x@value_column, new_grouping_column)
    
    original_grouping_column <- x@grouping_column
    if (length(original_grouping_column) == 0L) {
      old_data <- data.table::copy(data@data)
      old_data[, "_index_id" := .I]
      original_grouping_column <- "_index_id"
      
    } else {
      old_data <- x@data
    }
    
    data <- data.table::melt(
      data = old_data,
      id.vars = original_grouping_column,
      measure_vars = new_grouping_column,
      variable.name = new_grouping_column_name,
      value.name = new_value_column_name
    )
    
    # If any other value columns remain, merge these into data.
    if (length(remaining_value_column) > 0L) {
      data <- merge(
        x = data,
        y = old_data[, mget(c(original_grouping_column, remaining_value_column))],
        by = original_grouping_column,
        all = TRUE
      )
    }
    
    # Drop placeholder grouping column
    if (any(original_grouping_column == "_index_id")) {
      data[, "_index_id" := NULL]
    }
    
    # Update data, value column and grouping column.
    x@data <- data
    x@value_column <- c(new_value_column_name, remaining_value_column)
    x@grouping_column <- c(x@grouping_column, new_grouping_column_name)
    
    return(x)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



run_hyperparameter_optimisation <- function(
    cl = NULL,
    project_list = NULL,
    data_id = NULL,
    settings = NULL,
    file_paths = NULL,
    vimp_method,
    learner = NULL,
    message_indent = 0L,
    verbose = TRUE
) {
  # Hyper-parameters are parameters of a learner, such as the signature size or
  # the number of trees in a random forest. The selection of hyper-parameters
  # influences predictive performance of a model, often to a large extent.
  # Therefore hyper-parameters require optimisation. There are various way to
  # deal with this issue, for example grid search or random search, both of
  # which can be computationally expensive for complex learners. An efficient,
  # successful alternative is using a model-based strategy. In general, building
  # a model using the learner of interest is the most time-consuming step in the
  # process. Therefore, model building should be done only when necessary. The
  # main idea behind model-based hyper-parameter optimisation is that a
  # hyper-parameter model is trained and used to assess random configurations.
  # Then the most interesting configurations, i.e. those which are expected tp
  # provide good objective scores or are located in an unexplored region of the
  # hyperparameter space, are actually build and assessed.
  #
  # The algorithm implemented here is SMAC ("Hutter, Frank, Holger H. Hoos, and
  # Kevin Leyton-Brown. "Sequential Model-Based Optimization for General
  # Algorithm Configuration." LION 5 (2011): 507-523"). The version implemented
  # here varies from the original in the following ways:
  #
  # 1. The intensify step, where promising configurations are compared with the
  # best known configuration, is done in parallel, not sequentially.
  #
  # 2. During the intensify step we assess the probability that a promising
  # challenger configuration is at least as good as the best known
  # configuration. If it is highly unlikely to be better (default: p>=0.05), we
  # skip further intensification with that challenger. The non-parametric
  # Wilcoxon signed-rank test is used to compare objective scores on the same
  # bootstraps of the data.
  #
  # 3. Early stopping criteria are included for converging objective scores and
  # unaltered best configurations over multiple iterations.
  #
  # 4. The initial grid may be randomised by setting
  # \code{settings$hpo$hpo_grid_initialisation_method}. User-provided hyper-parameter
  # settings are never randomised.
  #
  # If not provided by the user, hyper-parameters and corresponding meta-data
  # are sourced from the learner.
  
  # In absence of a learner, we assume that variable importance is computed.
  is_vimp <- is.null(learner)
  
  if (is.null(project_list)) project_list <- get_project_list()
  if (is.null(settings)) settings <- get_settings()
  if (is.null(file_paths)) file_paths <- get_file_paths()
  
  # Get the iteration list for the variable importance computation or model
  # development step for which hyperparameter optimisation is performed.
  hpo_id_list <- .get_preprocessing_iteration_identifiers(run = .get_run_list(
    iteration_list = project_list$iter_list,
    data_id = data_id,
    run_id = 1L
  ))
  
  iteration_list <- .get_run_list(
    iteration_list = project_list$iter_list,
    data_id = hpo_id_list$data
  )
  
  # Generate objects for hyperparameter optimisation.
  object_list <- lapply(
    iteration_list,
    .create_hyperparameter_optimisation_initial_objects,
    vimp_method = vimp_method,
    learner = learner,
    settings = settings,
    project_id = project_list$project_id
  )
  
  # Replace objects that already exist as a file.
  object_list <- lapply(
    object_list,
    .collect_hyperparameter_optimisation_completed_objects,
    vimp_method = vimp_method,
    learner = learner,
    file_paths = file_paths
  )
  
  # Check which objects have already been created.
  object_exist <- sapply(
    object_list,
    .exists_hyperparameter_optimisation_object,
    vimp_method = vimp_method,
    learner = learner,
    file_paths = file_paths
  )
  
  # Return objects as is in case they already exist.
  if (all(object_exist)) return(object_list)
  
  # Determine how parallel processing takes place.
  if (settings$hpo$do_parallel %in% c("TRUE", "inner")) {
    cl_inner <- cl
    cl_outer <- NULL
    
  } else if (settings$hpo$do_parallel %in% c("outer")) {
    cl_inner <- NULL
    cl_outer <- cl
    
    if (!is.null(cl_outer)) {
      logger_message(
        paste0(
          "Hyperparameter optimisation: Load-balanced parallel processing ",
          "is done in the outer loop. No progress can be displayed."
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
  } else {
    cl_inner <- cl_outer <- NULL
  }
  
  # Message start of hyperparameter optimisation.
  if (is_vimp) {
    logger_message(
      paste0(
        "Hyperparameter optimisation: Starting parameter optimisation for the ",
        vimp_method, " variable importance method."
      ),
      indent = message_indent,
      verbose = verbose
    )
    
  } else {
    logger_message(
      paste0(
        "Hyperparameter optimisation: Starting parameter optimisation for the ",
        learner, " learner, based on variable importances from the ",
        vimp_method, " variable importance method."
      ),
      indent = message_indent,
      verbose = verbose
    )
  }
  
  # Generate experiment info
  experiment_info <- lapply(
    which(!object_exist),
    function(ii, n_total) {
      return(list(
        "experiment_id" = ii,
        "n_experiment_total" = n_total
      ))
    },
    n_total = length(object_exist)
  )
  
  # Pass to fam_mapply_lb to allow for passing cluster objects to the
  # optimise_hyperparameters method.
  new_object_list <- fam_mapply_lb(
    cl = cl_outer,
    assign = "all",
    FUN = optimise_hyperparameters,
    object = object_list[!object_exist],
    experiment_info = experiment_info,
    progress_bar = !is.null(cl_outer),
    MoreArgs = list(
      "data" = NULL,
      "cl" = cl_inner,
      "user_list" = NULL,
      "metric" = settings$hpo$hpo_metric,
      "optimisation_function" = settings$hpo$hpo_optimisation_function,
      "acquisition_function" = settings$hpo$hpo_acquisition_function,
      "grid_initialisation_method" = settings$hpo$hpo_grid_initialisation_method,
      "n_random_sets" = settings$hpo$hpo_n_grid_initialisation_samples,
      "exploration_method" = settings$hpo$hpo_exploration_method,
      "determine_vimp" = settings$hpo$hpo_determine_vimp,
      "measure_time" = TRUE,
      "hyperparameter_learner" = settings$hpo$hpo_hyperparameter_learner,
      "n_max_bootstraps" = settings$hpo$hpo_max_bootstraps,
      "n_initial_bootstraps" = settings$hpo$hpo_initial_bootstraps,
      "n_intensify_step_bootstraps" = settings$hpo$hpo_bootstraps,
      "n_max_optimisation_steps" = settings$hpo$hpo_smbo_iter_max,
      "n_max_intensify_steps" = settings$hpo$hpo_intensify_max_iter,
      "intensify_stop_p_value" = settings$hpo$hpo_alpha,
      "convergence_tolerance" = settings$hpo$hpo_convergence_tolerance,
      "convergence_stopping" = settings$hpo$hpo_conv_stop,
      "time_limit" = settings$hpo$hpo_time_limit,
      "verbose" = verbose,
      "message_indent" = message_indent + 1L,
      "save_in_place" = TRUE,
      "is_vimp" = is.null(learner)
    )
  )
  
  # Fill in object list.
  object_list[!object_exist] <- new_object_list
  
  # Message completion of hyperparameter optimisation.
  if (is_vimp) {
    logger_message("", verbose = verbose)
    logger_message(
      paste0(
        "Hyperparameter optimisation: Completed parameter optimisation for the ",
        vimp_method, " variable importance method."
      ),
      indent = message_indent,
      verbose = verbose
    )
    
  } else {
    logger_message("", verbose = verbose)
    logger_message(
      paste0(
        "Hyperparameter optimisation: Completed parameter optimisation for the ",
        learner, " learner, based on variable importances from the ",
        vimp_method, " variable importance method."
      ),
      indent = message_indent,
      verbose = verbose
    )
  }
  
  # Return the object which contains the (optimised) hyperparameters.
  return(object_list)
}



# optimise_hyperparameters (ANY, NULL) -----------------------------------------
setMethod(
  "optimise_hyperparameters",
  signature(
    object = "ANY",
    data = "NULL"
  ),
  function(
    object,
    data,
    ...,
    save_in_place = FALSE,
    is_vimp = NULL
  ) {
    # Create dataset on the fly. This is the usual route for
    # summon_familiar.
    data <- methods::new(
      "dataObject",
      data = NULL,
      preprocessing_level = "none",
      outcome_type = object@outcome_type,
      delay_loading = TRUE,
      perturb_level = tail(object@run_table, n = 1L)$perturb_level,
      load_validation = FALSE,
      aggregate_on_load = FALSE,
      outcome_info = object@outcome_info
    )
    
    # Make sure the input data is processed.
    data <- process_input_data(
      object = object,
      data = data
    )
    
    # Check that any data is present.
    if (is_empty(data)) return(object)
    
    # Call proper routine.
    new_object <- do.call(
      optimise_hyperparameters,
      args = c(
        list(
          "object" = object,
          "data" = data
        ),
        list(...)
      )
    )
    
    # Check if the code is called downstream from summon_familiar.
    is_main_process <- !inherits(tryCatch(get_file_paths(), error = identity), "error")
    
    # Save object.
    if (save_in_place && is_main_process) {
      .create_hyperparameter_optimisation_directory(
        object = new_object,
        is_vimp = is_vimp
      )
      
      file_path <- .get_hyperparameter_optimisation_object_path(
        object = new_object,
        is_vimp = is_vimp
      )
      
      saveRDS(new_object, file = file_path)
    }
    
    return(new_object)
  }
)



# optimise_hyperparameters (familiarNoveltyDetector, dataObject) ---------------
setMethod(
  "optimise_hyperparameters",
  signature(
    object = "familiarNoveltyDetector",
    data = "dataObject"
  ),
  function(
    object,
    data,
    user_list = NULL,
    ...
  ) {
    # Obtain standard parameters.
    parameter_list <- get_default_hyperparameters(
      object = object,
      data = data
    )
    
    # Check that any parameters are present.
    if (is_empty(parameter_list)) return(object)
    
    # Set the user_list if it is not present, or set through hyperparameter
    # attribute.
    if (is.null(user_list) && is.null(object@hyperparameters)) {
      user_list <- list()
    } else if (is.null(user_list) && !is.null(object@hyperparameters)) {
      user_list <- object@hyperparameters
    }
    
    # Recreate the default parameter list with information from the
    # user-provided list, if any. This allows for changing some hyperparameter
    # settings that depend on other hyperparameters.
    parameter_list <- get_default_hyperparameters(
      object = object,
      data = data,
      user_list = user_list
    )
    
    # Update the parameter list With user-defined variables.
    parameter_list <- .update_hyperparameters(
      parameter_list = parameter_list,
      user_list = user_list
    )
    
    if (.any_randomised_hyperparameters(parameter_list = parameter_list)) {
      ..error_reached_unreachable_code(paste0(
        "optimise_hyperparameters,familiarNoveltyDetector,dataObject: ",
        "unset hyperparameters are present, but not expected."
      ))
    }
    
    # Update hyperparameters to set any fixed parameters.
    object@hyperparameters <- lapply(
      parameter_list,
      function(list_entry) list_entry$init_config
    )
    
    return(object)
  }
)



# optimise_hyperparameters (familiarVimpMethod, dataObject) --------------------
setMethod(
  "optimise_hyperparameters",
  signature(
    object = "familiarVimpMethod",
    data = "dataObject"
  ),
  function(
    object,
    data,
    user_list = NULL,
    ...
  ) {
    # Obtain standard parameters.
    parameter_list <- get_default_hyperparameters(
      object = object,
      data = data
    )
    
    # Check that any parameters are present.
    if (is_empty(parameter_list)) {
      object@hyperparameters <- list()
      return(object)
    }
    
    # Set the user_list if it is not present, or set through hyperparameter
    # attribute.
    if (is.null(user_list) && is.null(object@hyperparameters)) {
      user_list <- list()
    } else if (is.null(user_list) && !is.null(object@hyperparameters)) {
      user_list <- object@hyperparameters
    }
    
    # Set the signature size. This parameter may not be used by all variable
    # importance methods, and will be ignored in that case.
    user_list$sign_size <- get_n_features(x = data)
    
    # Recreate the default parameter list with information from the
    # user-provided list, if any. This allows for changing some hyperparameter
    # settings that depend on other hyperparameters.
    parameter_list <- get_default_hyperparameters(
      object = object,
      data = data,
      user_list = user_list
    )
    
    # Update the parameter list With user-defined variables.
    parameter_list <- .update_hyperparameters(
      parameter_list = parameter_list,
      user_list = user_list
    )
    
    if (.any_randomised_hyperparameters(parameter_list = parameter_list)) {
      ..error_reached_unreachable_code(paste0(
        "optimise_hyperparameters,familiarVimpMethod,dataObject: ",
        "unset hyperparameters are present, but not expected."
      ))
    }
    
    # Update hyperparameters to set any fixed parameters.
    object@hyperparameters <- lapply(
      parameter_list, 
      function(list_entry) list_entry$init_config
    )
    
    return(object)
  }
)


# optimise_hyperparameters (familiarModel, dataObject) -------------------------
setMethod(
  "optimise_hyperparameters",
  signature(
    object = "familiarModel",
    data = "dataObject"
  ),
  function(
    object,
    data,
    cl = NULL,
    experiment_info = NULL,
    user_list = NULL,
    metric = waiver(),
    optimisation_function = "validation",
    acquisition_function = "expected_improvement",
    grid_initialisation_method = "fixed_subsample",
    exploration_method = "successive_halving",
    n_random_sets = 100L,
    measure_time = TRUE,
    hyperparameter_learner = "gaussian_process",
    n_max_bootstraps = 20L,
    n_initial_bootstraps = 1L,
    n_intensify_step_bootstraps = 3L,
    n_max_optimisation_steps = 20L,
    n_max_intensify_steps = 5L,
    n_challengers = 20L,
    intensify_stop_p_value = 0.05,
    convergence_tolerance = NULL,
    convergence_stopping = 3L,
    no_decent_model_stopping = 4L,
    time_limit = NULL,
    verbose = TRUE,
    message_indent = 0L,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    param_id <- NULL
    
    if (!is.null(experiment_info)) {
      logger_message("", verbose = verbose)
      logger_message(
        paste0(
          "Starting hyperparameter optimisation for data subsample ",
          experiment_info$experiment_id, " of ",
          experiment_info$n_experiment_total, "."
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    # Set default metric
    if (is.waive(metric)) metric <- .get_default_metric(outcome_type = object@outcome_type)
    
    # Set convergence tolerance, if it has not been set.
    if (is.null(convergence_tolerance)) {
      # Get the number of samples.
      n_samples <- get_n_samples(data, "series")
      
      # Nothing to do if there are no samples.
      if (n_samples == 0L) return(object)
      
      # Set convergence tolerance.
      convergence_tolerance <- 0.1 * 1.0 / sqrt(n_samples)
      
      # Limit to -4
      if (convergence_tolerance < 1E-4) convergence_tolerance <- 1E-4
    }
    
    # Check if the metric is ok. Packed into a for loop to enable multi-metric
    # optimisation.
    sapply(metric, .check_metric_outcome_type, outcome_type = object@outcome_type)
    
    # Check if acquisition_function is correctly specified.
    .check_parameter_value_is_valid(
      x = acquisition_function,
      var_name = "acquisition_function",
      values = .get_available_acquisition_functions()
    )
    
    # Check if hyperparameter learner is correctly specified.
    .check_parameter_value_is_valid(
      x = hyperparameter_learner,
      var_name = "hyperparameter_learner",
      values = .get_available_hyperparameter_learners()
    )
    
    # Check if optimisation_function is correctly specified.
    .check_parameter_value_is_valid(
      x = optimisation_function,
      var_name = "optimisation_function",
      values = .get_available_optimisation_functions(hyperparameter_learner = hyperparameter_learner)
    )
    
    # Check if exploration_method is correctly specified.
    .check_parameter_value_is_valid(
      x = exploration_method,
      var_name = "exploration_method",
      values = .get_available_hyperparameter_exploration_methods()
    )
    
    # Report on methodology used.
    optimisation_description <- switch(
      optimisation_function,
      "max_validation" = "maximising out-of-bag performance",
      "validation" = "maximising out-of-bag performance",
      "balanced" = paste0(
        "maximising out-of-bag performance while constraining performance ",
        "differences between in-bag and out-of-bag data"
      ),
      "stronger_balance" = paste0(
        "maximising out-of-bag performance while strongly constraining ",
        "performance differences between in-bag and out-of-bag data"
      ),
      "validation_minus_sd" = paste0(
        "maximising out-of-bag performance while penalising variance ",
        "using the lower 1 standard deviation bound"
      ),
      "validation_25th_percentile" = paste0(
        "maximising out-of-bag performance while penalising variance ",
        "using the 25th percentile bound"
      ),
      "model_estimate" = "maximising performance estimates from the hyperparameter model",
      "model_estimate_minus_sd" = paste0(
        "maximising performances estimates while penalising variance ",
        "estimated by the model"
      ),
      "model_balanced_estimate" = paste0(
        "maximising performance under estimates of the performance differences ",
        "between in-bag and out-of-bag data"
      ),
      "model_balanced_estimate_minus_sd" = paste0(
        "maximising performance under estimates of the performance differences ",
        "between in-bag and out-of-bag data while penalising ",
        "estimated variance in performance differences"
      )
    )
    
    logger_message(
      paste0(
        "Hyperparameter optimisation is conducted using the ",
        paste_s(metric), ifelse(length(metric) > 1L, " metrics", " metric"),
        " by ", optimisation_description, "."
      ),
      indent = message_indent + 1L,
      verbose = verbose
    )
    
    inference_description <- switch(
      hyperparameter_learner,
      "random_forest" = "selected after inferring utility using a Random Forest",
      "gaussian_process" = paste0(
        "selected after inferring utility using a ",
        "localised approximate Gaussian Process"
      ),
      "bayesian_additive_regression_trees" = paste0(
        "selected after inferring utility using ",
        "Bayesian Additive Regression Trees"
      ),
      "bart" = paste0(
        "selected after inferring utility using ",
        "Bayesian Additive Regression Trees"
      ),
      "random" = "drawn randomly",
      "random_search" = "drawn randomly"
    )
    
    logger_message(
      paste0(
        "Candidate hyperparameter sets after the initial run are ",
        inference_description, "."
      ),
      indent = message_indent + 1L,
      verbose = verbose
    )
    
    if (!hyperparameter_learner %in% c("random", "random_search")) {
      utility_description <- switch(
        acquisition_function,
        "improvement_probability" = paste0(
          "probability of improvement over the best known hyperparameter set"
        ),
        "improvement_empirical_probability" = paste0(
          "empirical probability of improvement over ",
          "the best known hyperparameter set"
        ),
        "expected_improvement" = "expected improvement",
        "upper_confidence_bound" = "the upper regret bound",
        "bayes_upper_confidence_bound" = "the Bayesian upper confidence bound"
      )
      
      logger_message(
        paste0("Utility is measured as ", utility_description, "."),
        indent = message_indent + 1L,
        verbose = verbose
      )
    }
    
    ## Create and update hyperparameter sets -----------------------------------
    
    # Set the user_list if it is not present, or set through hyperparameter
    # attribute.
    if (is.null(user_list) && is.null(object@hyperparameters)) {
      user_list <- list()
    } else if (is.null(user_list) && !is.null(object@hyperparameters)) {
      user_list <- object@hyperparameters
    }
    
    # Create the default parameter list with information from the user-provided
    # list, if any. This allows for changing some hyperparameter settings that
    # depend on other hyperparameters.
    parameter_list <- get_default_hyperparameters(
      object = object,
      data = data,
      user_list = user_list
    )
    
    # Check that any parameters are present.
    if (is_empty(parameter_list)) return(object)
    
    # Update the parameter list With user-defined variables.
    parameter_list <- .update_hyperparameters(
      parameter_list = parameter_list,
      user_list = user_list,
      n_features = ifelse(
        object@vimp_method %in% .get_available_no_features_vimp_methods(),
        0L,
        get_n_features(data)
      )
    )
    
    # Check that any parameters can be randomised.
    if (!.any_randomised_hyperparameters(parameter_list = parameter_list)) {
      # Update hyperparameters to set any fixed parameters.
      object@hyperparameters <- lapply(
        parameter_list, 
        function(list_entry) list_entry$init_config
      )
      
      logger_message(
        paste0(
          "Hyperparameter optimisation: All hyperparameters are fixed. ",
          "No optimisation is required."
        ),
        indent = message_indent,
        verbose = verbose
      )
      
      # Only exit if variable importance is not required.
      if (!is_empty(object@vimp_table)) {
        return(object)
      }
    }
    
    ## Create bootstrap samples ------------------------------------------------
    
    # Generate bootstrap samples
    bootstraps <- tryCatch(
      .create_bootstraps(
        n_iter = n_max_bootstraps,
        outcome_type = object@outcome_type,
        data = data@data
      ),
      error = identity
    )
    
    # Check that bootstraps could be created. This may fail if the data set is
    # too small.
    if (inherits(bootstraps, "error")) {
      logger_message(
        paste0(
          "Hyperparameter optimisation: Failed to create bootstraps. ",
          "The dataset may be too small."
        ),
        indent = message_indent,
        verbose = verbose
      )
      
      # Remove any fixed hyperparameters.
      object@hyperparameters <- NULL
      
      return(object)
    }
    
    ## Create or obtain variable importance ------------------------------------
    vimp_table_list <- .compute_hyperparameter_variable_importance(
      cl = cl,
      object = object,
      data = data,
      bootstraps = bootstraps$train_list,
      metric = metric,
      measure_time = measure_time,
      optimisation_function = optimisation_function,
      acquisition_function = acquisition_function,
      grid_initialisation_method = grid_initialisation_method,
      n_random_sets = min(c(n_random_sets, 50L)),
      n_max_bootstraps = min(c(n_max_bootstraps, 20L)),
      n_max_optimisation_steps = min(c(n_max_optimisation_steps, 5L)),
      n_max_intensify_steps = min(c(n_max_intensify_steps, 3L)),
      n_intensify_step_bootstraps = min(c(n_intensify_step_bootstraps, 5L)),
      intensify_stop_p_value = intensify_stop_p_value,
      convergence_tolerance = convergence_tolerance,
      convergence_stopping = min(c(convergence_stopping, 3L)),
      time_limit = time_limit,
      verbose = verbose,
      message_indent = message_indent
    )
    
    # Attach variable importance tables.
    object@vimp_table <- vimp_table_list
    
    # Break if there is nothing to optimise, This continues the earlier check,
    # but after setting the variable importance table.
    if (!.any_randomised_hyperparameters(parameter_list = parameter_list)) {
      return(object)
    }
    
    ## Set signature size ------------------------------------------------------
    
    # Signature size depends on the variable importance method that is used.
    if (!is.null(parameter_list$sign_size)) {
      # Set signature size.
      user_list$sign_size <- .set_signature_size(
        object = object,
        vimp_table_list = vimp_table_list,
        suggested_range = user_list$sign_size
      )

      # Update the parameter list With user-defined variables.
      parameter_list <- .update_hyperparameters(
        parameter_list = parameter_list,
        user_list = user_list,
        n_features = get_n_features(data)
      )
      
      # Check that any parameters can be randomised.
      if (!.any_randomised_hyperparameters(parameter_list = parameter_list)) {
        # Update hyperparameters to set any fixed parameters.
        object@hyperparameters <- lapply(
          parameter_list,
          function(list_entry) list_entry$init_config
        )
        
        logger_message(
          paste0(
            "Hyperparameter optimisation: All hyperparameters are fixed. ",
            "No optimisation is required."
          ),
          indent = message_indent,
          verbose = verbose
        )
        
        return(object)
      }
    }
    
    ## Create metric objects ---------------------------------------------------
    
    # Update the outcome_info attribute of the familiar model. This is required
    # to set the metric baseline value.
    object@outcome_info <- .compute_outcome_distribution_data(
      object = object@outcome_info,
      data = data
    )
    
    # Create metric objects.
    metric_object_list <- lapply(
      metric,
      as_metric,
      outcome_type = object@outcome_type
    )
    
    # Add baseline values for each metric.
    metric_object_list <- lapply(
      metric_object_list,
      set_metric_baseline_value,
      object = object,
      data = data
    )
    
    ## Create initial set of randomised hyperparameters and bootstraps ---------
    
    # Update the parameter list to make sure that categorical variables
    # are encoded as factors.
    parameter_list <- .encode_categorical_hyperparameters(parameter_list)
    
    # Create initial set of configurations.
    parameter_table <- ..create_initial_hyperparameter_set(
      parameter_list = parameter_list,
      grid_initialisation_method = grid_initialisation_method,
      n_random_sets = n_random_sets
    )
    
    # Check that the parameter table is not empty.
    if (is_empty(parameter_table)) {
      logger_message(
        paste0(
          "Hyperparameter optimisation: No hyperparameters were found ",
          "to initialise the optimisation process."
        ),
        indent = message_indent,
        verbose = verbose
      )
      
      # Remove any fixed hyperparameters.
      object@hyperparameters <- NULL
      
      return(object)
    }
    
    ## Setup the hyperparameter model prototype --------------------------------
    
    # Set up the model prototype.
    optimisation_model_prototype <- methods::new(
      "familiarHyperparameterLearner",
      learner = hyperparameter_learner,
      target_learner = object@learner,
      target_outcome_type = object@outcome_type,
      optimisation_metric = metric,
      optimisation_function = optimisation_function
    )
    
    ## Perform initial set of computations -------------------------------------
    
    # Initialise the process clock.
    process_clock <- NULL
    if (!is.null(time_limit)) {
      if (require_package(
        c("callr", "microbenchmark"),
        purpose = "to measure hyperparameter optimisation time",
        message_type = "warning"
      )) {
        process_clock <- ProcessClock$new()
        on.exit(process_clock$close(), add = TRUE)
        
      } else {
        # If the required packages are not present, avoid early abortion of the
        # process.
        time_limit <- NULL
      }
    }
    
    if (measure_time & n_initial_bootstraps > 1L) {
      # Start with an initial run to measure performance and time.
      run_table <- ..create_hyperparameter_run_table(
        run_ids = 1L,
        parameter_ids = parameter_table$param_id
      )
      
      # Message begin.
      logger_message(
        paste0(
          "Compute initial model performance based on ",
          nrow(run_table), " hyperparameter sets."
        ),
        indent = message_indent + 1L,
        verbose = verbose
      )
      
      # Build and evaluate models. This creates a table with metric values,
      # objective scores for in-bag and out-of-bag data.
      score_results <- .compute_hyperparameter_model_performance(
        cl = cl,
        object = object,
        run_table = run_table,
        bootstraps = bootstraps,
        data = data,
        vimp_table_list = vimp_table_list,
        parameter_table = parameter_table,
        metric_objects = metric_object_list,
        iteration_id = 0L,
        verbose = verbose
      )
      
      # Get score table.
      score_table <- score_results$results
      
      # Get errors encountered during model training.
      train_errors <- score_results$error
      
      if (.optimisation_process_time_available(
        process_clock = process_clock,
        time_limit = time_limit,
        verbose = FALSE
      )) {
        
        # Set up the runs for the remaining initial computations.
        run_table <- ..create_hyperparameter_run_table(
          run_ids = setdiff(seq_len(n_initial_bootstraps), 1L),
          measure_time = measure_time,
          score_table = score_table,
          parameter_table = parameter_table,
          optimisation_model = optimisation_model_prototype,
          n_max_bootstraps = n_max_bootstraps
        )
        
        logger_message(
          paste0(
            "Compute initial model performance based on the second batch of ",
            nrow(run_table), " hyperparameter sets."
          ),
          indent = message_indent + 1L,
          verbose = verbose
        )
        
        # Create a model to predict the time a process takes to complete
        # training.
        time_optimisation_model <- .create_hyperparameter_time_optimisation_model(
          score_table = score_table,
          parameter_table = parameter_table,
          optimisation_function = optimisation_function
        )
        
        # Compute second batch of results.
        score_results <- .compute_hyperparameter_model_performance(
          cl = cl,
          object = object,
          run_table = run_table,
          bootstraps = bootstraps,
          data = data,
          vimp_table_list = vimp_table_list,
          parameter_table = parameter_table,
          metric_objects = metric_object_list,
          iteration_id = 0L,
          time_optimisation_model = time_optimisation_model,
          overhead_time = score_results$overhead_time,
          verbose = verbose
        )
        
        # Add new data to score and optimisation tables.
        score_table <- rbind(
          score_table,
          score_results$results,
          use.names = TRUE
        )
        
        # Add errors encountered during model training.
        train_errors <- c(train_errors, score_results$error)
      }
      
    } else {
      # Set up hyperparameter experiment runs
      run_table <- ..create_hyperparameter_run_table(
        run_ids = seq_len(n_initial_bootstraps),
        parameter_ids = parameter_table$param_id
      )
      
      logger_message(
        paste0(
          "Compute initial model performance based on ",
          nrow(run_table), " hyperparameter sets."
        ),
        indent = message_indent + 1L,
        verbose = verbose
      )
      
      # Build and evaluate models. This creates a table with metric values,
      # objective scores for in-bag and out-of-bag data.
      score_results <- .compute_hyperparameter_model_performance(
        cl = cl,
        object = object,
        run_table = run_table,
        bootstraps = bootstraps,
        data = data,
        vimp_table_list = vimp_table_list,
        parameter_table = parameter_table,
        iteration_id = 0L,
        metric_objects = metric_object_list,
        verbose = verbose
      )
      
      # Get score table.
      score_table <- score_results$results
      
      # Get errors encountered during model training.
      train_errors <- score_results$error
    }
    
    # Find information regarding the dataset that has the highest optimisation
    # score.
    incumbent_set <- get_best_hyperparameter_set(
      score_table = score_table,
      parameter_table = parameter_table,
      optimisation_model = optimisation_model_prototype,
      n_max_bootstraps = n_max_bootstraps,
      n = 1L
    )
    
    # Message the user concerning the initial optimisation score.
    logger_message(
      paste0(
        "Hyperparameter optimisation: Initialisation complete: ",
        ..parse_optimisation_summary_to_string(
          parameter_set = incumbent_set,
          parameter_table = parameter_table,
          parameter_list = parameter_list
        )
      ),
      indent = message_indent + 1L,
      verbose = verbose
    )
    
    # Check whether any combination yielded anything valid.
    skip_optimisation <- incumbent_set$summary_score <= ..get_replacement_optimisation_score()
    
    # Update n_intensify_step_bootstraps and n_max_intensify_steps when
    # exploration_method equals none. There is no reason to perform steps
    # sequentially as there is no pruning.
    if (exploration_method == "none") {
      n_intensify_step_bootstraps <- n_intensify_step_bootstraps * n_max_intensify_steps
      n_max_intensify_steps <- 1L
    } else if (exploration_method == "single_shot") {
      n_max_intensify_steps <- 1L
      n_intensify_step_bootstraps <- 1L
    }
    
    # Create list with stopping criteria based on initial runs.
    stop_list <- ..update_hyperparameter_optimisation_stopping_criteria(
      set_data = incumbent_set,
      stop_data = NULL,
      tolerance = convergence_tolerance
    )
    
    optimisation_step <- 0L
    while (optimisation_step < n_max_optimisation_steps && !skip_optimisation) {
      # Stop optimisation if there is no time left. This is an initial check
      # because the nested loop may break for different reasons. This check is
      # not verbose, pending a later, final check.
      if (!.optimisation_process_time_available(
        process_clock = process_clock,
        time_limit = time_limit,
        verbose = FALSE
      )) {
        break
      }
      
      ## SMBO - Intensify ------------------------------------------------------
      
      # Create a model to predict the time a process takes to complete training.
      time_optimisation_model <- .create_hyperparameter_time_optimisation_model(
        score_table = score_table,
        parameter_table = parameter_table,
        optimisation_function = optimisation_function
      )
      
      # Local neighbourhood + random hyperparameter randomisation for challenger
      # configurations. This selects challengers combinations.
      challenger_data <- .create_hyperparameter_challenger_sets(
        score_table = score_table,
        parameter_list = parameter_list,
        parameter_table = parameter_table,
        time_optimisation_model = time_optimisation_model,
        score_optimisation_model = optimisation_model_prototype,
        acquisition_function = acquisition_function,
        n_max_bootstraps = n_max_bootstraps,
        smbo_iter = optimisation_step,
        measure_time = measure_time,
        n_challengers = n_challengers
      )
      
      # Check that any challenger datasets were found.
      if (is_empty(challenger_data)) break
      
      # Add challenger parameters to the parameter table
      parameter_table <- rbind(
        parameter_table,
        challenger_data,
        use.names = TRUE
      )
      
      # Select only unique parameters
      parameter_table <- unique(parameter_table, by = "param_id")
      
      # Determine parameter ids from incumbent and challenger
      parameter_id_incumbent <- incumbent_set$param_id
      parameter_id_challenger <- challenger_data$param_id
      
      # Drop incumbent parameter id from the list of challengers
      parameter_id_challenger <- setdiff(parameter_id_challenger, parameter_id_incumbent)
      
      # Check if there are any challengers left
      if (length(parameter_id_challenger) == 0L) break
      
      # Start intensification rounds
      n_intensify_steps <- 0L
      while (n_intensify_steps < n_max_intensify_steps) {
        # Check if there is time left for intensification.
        if (!.optimisation_process_time_available(
          process_clock = process_clock,
          time_limit = time_limit,
          verbose = FALSE
        )) {
          break
        }
        
        # Create run table.
        run_table <- ..create_hyperparameter_intensify_run_table(
          parameter_id_incumbent = parameter_id_incumbent,
          parameter_id_challenger = parameter_id_challenger,
          score_table = score_table,
          n_max_bootstraps = n_max_bootstraps,
          n_intensify_step_bootstraps = n_intensify_step_bootstraps,
          exploration_method = exploration_method
        )
        
        # Check if there are any runs to perform.
        if (is_empty(run_table)) break
        
        # Message the user.
        logger_message(
          paste0(
            "Intensify step ", n_intensify_steps + 1L, " using ",
            length(parameter_id_challenger), " challenger hyperparameter sets."
          ),
          indent = message_indent + 1L,
          verbose = verbose
        )
        
        # Compute metric values for the bootstraps of the incumbent and
        # challenger parameter sets.
        score_results <- .compute_hyperparameter_model_performance(
          cl = cl,
          object = object,
          run_table = run_table,
          bootstraps = bootstraps,
          data = data,
          vimp_table_list = vimp_table_list,
          parameter_table = parameter_table,
          metric_objects = metric_object_list,
          iteration_id = optimisation_step + 1L,
          time_optimisation_model = time_optimisation_model,
          overhead_time = score_results$overhead_time,
          verbose = verbose
        )
        
        # Get score table. Add new data to score and optimisation tables.
        score_table <- rbind(
          score_table,
          score_results$results,
          use.names = TRUE
        )
        
        # Add errors encountered during model training.
        train_errors <- c(train_errors, score_results$error)
        
        # Find scores and return parameter ids for challenger and incumbent
        # hyperparameter sets. Compared to the original SMAC algorithm, we
        # actively eliminate unsuccessful challengers. To do so we determine the
        # probability that the optimisation score of a challenger does not
        # exceed the incumbent score.
        runoff_parameter_ids <- .compare_hyperparameter_sets(
          score_table = score_table,
          parameter_table = parameter_table,
          optimisation_model = optimisation_model_prototype,
          parameter_id_incumbent = parameter_id_incumbent,
          parameter_id_challenger = parameter_id_challenger,
          exploration_method = exploration_method,
          intensify_stop_p_value = intensify_stop_p_value
        )
        
        # Extract hyperparameter set identifiers.
        parameter_id_incumbent <- runoff_parameter_ids$parameter_id_incumbent
        parameter_id_challenger <- runoff_parameter_ids$parameter_id_challenger
        
        # Check if there are challengers remaining
        if (length(parameter_id_challenger) == 0L) break
        
        # Update intensify iterator
        n_intensify_steps <- n_intensify_steps + 1L
        
        # Update the model that predicts the time a process takes to complete
        # training.
        time_optimisation_model <- .create_hyperparameter_time_optimisation_model(
          score_table = score_table,
          parameter_table = parameter_table,
          optimisation_function = optimisation_function
        )
      }
      
      ## SMBO - Evaluate -------------------------------------------------------
      # We assess improvement to provide early stopping on non-improving
      # incumbents.
      
      # Get all runs and compute details for the incumbent set.
      incumbent_set <- get_best_hyperparameter_set(
        score_table = score_table,
        parameter_table = parameter_table,
        optimisation_model = optimisation_model_prototype,
        n_max_bootstraps = n_max_bootstraps
      )
      
      # Update list with stopping criteria
      stop_list <- ..update_hyperparameter_optimisation_stopping_criteria(
        set_data = incumbent_set,
        stop_data = stop_list,
        tolerance = convergence_tolerance
      )
      
      # Message progress.
      logger_message(
        paste0(
          "Hyperparameter optimisation: SMBO iteration ", optimisation_step + 1L, ": ",
          ..parse_optimisation_summary_to_string(
            parameter_set = incumbent_set,
            parameter_table = parameter_table,
            parameter_list = parameter_list
          )
        ),
        indent = message_indent + 1L,
        verbose = verbose
      )
      
      # Check time and finish the optimisation process if required. This will
      # also automatically generate a message if verbose is TRUE.
      if (!.optimisation_process_time_available(
        process_clock = process_clock,
        time_limit = time_limit,
        message_indent = message_indent,
        verbose = verbose
      )) {
        break
      }
      
      # Break if the convergence counter reaches a certain number
      if (
        stop_list$convergence_counter_score >= convergence_stopping ||
        stop_list$convergence_counter_parameter_id >= convergence_stopping
      ) {
        # Message convergence
        logger_message(
          paste0(
            "Hyperparameter optimisation: Optimisation stopped early ",
            "as convergence was achieved."
          ),
          indent = message_indent,
          verbose = verbose
        )
        
        # Stop SMBO
        break
      }
      
      # Break if no model better than the naive model was found.
      if (stop_list$no_naive_improvement_counter >= no_decent_model_stopping) {
        # Message early stopping.
        logger_message(
          paste0(
            "Hyperparameter optimisation: Optimisation stopped early because ",
            "no models were found to perform better than naive models."
          ),
          indent = message_indent,
          verbose = verbose
        )
        
        # Stop SMBO.
        break
      }
      
      # Update main iterator
      optimisation_step <- optimisation_step + 1L
    }
    
    ## SMBO - Wrap-up and report------------------------------------------------
    
    # Get all runs and determine incumbent parameter id.
    incumbent_set <- get_best_hyperparameter_set(
      score_table = score_table,
      parameter_table = parameter_table,
      optimisation_model = optimisation_model_prototype,
      n_max_bootstraps = n_max_bootstraps
    )
    
    # Add corresponding hyper parameters and remove redundant columns.
    optimal_set_table <- parameter_table[param_id == incumbent_set$param_id, ]
    optimal_set_table[, "param_id" := NULL]
    
    # Check that a suitable set of hyperparameters was found.
    if (incumbent_set$summary_score <= ..get_replacement_optimisation_score()) {
      # In this case, no suitable hyperparameters were found, for whatever
      # reason.
      
      logger_message(
        paste0(
          "Hyperparameter optimisation: No suitable set of ",
          "hyperparameters was found."
        ),
        indent = message_indent,
        verbose = verbose
      )
      
      # Report on errors.
      if (!is.null(train_errors)) {
        # Generate a summary of the errors.
        train_errors <- condition_summary(train_errors)
        
        # Notify that one or more errors were encountered.
        logger_message(
          paste0(
            "Hyperparameter optimisation: The following ",
            ifelse(length(train_errors) == 1L, "error was", "errors were"),
            " encountered while training models:"
          ),
          indent = message_indent,
          verbose = verbose
        )
        
        # Show errors.
        sapply(
          train_errors,
          logger_message,
          indent = message_indent + 1L,
          verbose = verbose
        )
      }
      
      # Set NULL.
      object@hyperparameters <- NULL
      
    } else if (
      incumbent_set$validation_score < 0.0 &&
      !object@vimp_method %in% c("none", "signature_only")
    ) {
      # In this case, no set of hyperparameters found that led to a model that
      # was better than the naive model. We then train a naive model instead, by
      # forcing sign_size to 0.
      
      object@hyperparameters <- as.list(optimal_set_table)
      object@hyperparameters$sign_size <- 0L
      
      logger_message(
        paste0(
          "Hyperparameter optimisation: No set of hyperparameters was identified that ",
          "leads to a model that performs better than a naive model."
        ),
        indent = message_indent,
        verbose = verbose
      )
      
    } else {
      # In this case the model trained properly.
      object@hyperparameters <- as.list(optimal_set_table)
      
      logger_message(
        paste0(
          "Hyperparameter optimisation: A suitable set of hyperparameters was identified: ",
          ..parse_optimisation_summary_to_string(
            parameter_set = incumbent_set,
            parameter_table = parameter_table,
            parameter_list = parameter_list
          )
        ),
        indent = message_indent,
        verbose = verbose
      )
    }
    
    time_taken <- NULL
    if (is(process_clock, "processClock")) time_taken <- process_clock$time(units = "mins")
    
    # Update attributes of object.
    object@hyperparameter_data <- list(
      "score_table" = score_table,
      "parameter_table" = parameter_table,
      "time_taken" = time_taken,
      "metric" = metric,
      "metric_object" = metric_object_list,
      "hyperparameter_learner" = hyperparameter_learner,
      "optimisation_function" = optimisation_function,
      "n_samples" = get_n_samples(data),
      "n_features" = get_n_features(data)
    )
    
    return(object)
  }
)



.create_hyperparameter_optimisation_initial_objects <- function(
    run,
    vimp_method,
    learner = NULL,
    settings,
    project_id
) {
  # Obtain feature information list.
  feature_info_list <- .get_feature_info_list(run = run)
  
  if (is.null(learner)) {
    # Create the variable importance met hod object or familiar model object to
    # compute variable importance with.
    object <- promote_vimp_method(object = methods::new(
      "familiarVimpMethod",
      outcome_type = settings$data$outcome_type,
      hyperparameters = settings$vimp$param[[vimp_method]],
      vimp_method = vimp_method,
      outcome_info = .get_outcome_info(),
      run_table = run$run_table,
      project_id = project_id
    ))
    
    # Set multivariate methods.
    if (is(object, "familiarModel")) is_multivariate <- TRUE
    if (is(object, "familiarVimpMethod")) is_multivariate <- object@multivariate
    
    # Find required features.
    required_features <- get_required_features(
      x = feature_info_list,
      exclude_signature = !is_multivariate
    )
    
    # Limit to required features. In principle, this removes signature features
    # which are not assessed through variable importance.
    feature_info_list <- feature_info_list[required_features]
    
    # Update the object.
    object@required_features <- required_features
    object@feature_info <- feature_info_list
    
  } else {
    # Create familiar model object. The following need to be updated:
    object <- promote_learner(object = methods::new(
      "familiarModel",
      outcome_type = settings$data$outcome_type,
      hyperparameters = settings$mb$hyper_param[[learner]],
      learner = learner,
      vimp_method = vimp_method,
      run_table = run$run_table,
      outcome_info = .get_outcome_info(),
      settings = settings$eval,
      project_id = project_id
    ))
    
    # Find required features.
    required_features <- get_required_features(
      x = feature_info_list,
      exclude_signature = FALSE
    )
    
    # Limit to required features. In principle, this removes signature features
    # which are not assessed through variable importance.
    feature_info_list <- feature_info_list[required_features]
    
    # Update the object.
    object@required_features <- required_features
    object@feature_info <- feature_info_list
  }
  
  return(object)
}



.create_hyperparameter_optimisation_directory <- function(
    object,
    is_vimp,
    file_paths = NULL
) {
  # Set file paths.
  if (is.null(file_paths)) file_paths <- get_file_paths()
  
  if (is(object, "familiarVimpMethod")) {
    vimp_method <- object@vimp_method
  } else if (is_vimp) {
    vimp_method <- object@learner
  } else {
    vimp_method <- object@vimp_method
  }
  
  # Set created hyperparameters.
  if (is_vimp) {
    dir_path <- file.path(file_paths$vimp_dir, vimp_method)
  } else {
    dir_path <- file.path(file_paths$mb_dir, object@learner, vimp_method)
  }
  
  # Create directory if it does not exist.
  if (!dir.exists(dir_path)) dir.create(dir_path, recursive = TRUE)
  
  return(invisible(NULL))
}



.collect_hyperparameter_optimisation_completed_objects <- function(
    object,
    vimp_method,
    learner,
    file_paths
) {
  
  # Identify file path to object.
  file_path <- .get_hyperparameter_optimisation_object_path(
    object = object,
    is_vimp = is.null(learner),
    file_paths = file_paths
  )
  
  if (file.exists(file_path)) object <- readRDS(file_path)
  
  return(object)
}



.exists_hyperparameter_optimisation_object <- function(
    object,
    vimp_method,
    learner,
    file_paths
) {
  
  # Identify file path to object.
  file_path <- .get_hyperparameter_optimisation_object_path(
    object = object,
    is_vimp = is.null(learner),
    file_paths = file_paths
  )
  
  return(file.exists(file_path))
}



.get_hyperparameter_optimisation_object_path <- function(
    object,
    is_vimp,
    file_paths = NULL
) {
  
  if (is.null(file_paths)) file_paths <- get_file_paths()
  
  # Extract data and run id
  object_data_id <- tail(object@run_table, n = 1L)$data_id
  object_run_id <- tail(object@run_table, n = 1L)$run_id
  
  # Get the variable importance method.
  if (is_vimp && is(object, "familiarVimpMethod")) {
    vimp_method <- object@vimp_method
  } else if (is_vimp) {
    vimp_method <- object@learner
  } else {
    vimp_method <- object@vimp_method
  }
  
  if (is_vimp) {
    dir_path <- file.path(file_paths$vimp_dir, vimp_method)
    file_name <- paste0(
      object@project_id, "_hyperparameters_",
      vimp_method, "_",
      object_data_id, "_",
      object_run_id, ".RDS"
    )
    
  } else {
    dir_path <- file.path(file_paths$mb_dir, object@learner, vimp_method)
    file_name <- paste0(
      object@project_id, "_hyperparameters_",
      object@learner, "_",
      vimp_method, "_",
      object_data_id, "_",
      object_run_id, ".RDS"
    )
  }
  
  file_path <- normalizePath(
    file.path(dir_path, file_name),
    mustWork = FALSE
  )
  
  return(file_path)
}



.find_hyperparameters_for_run <- function(
    run,
    hpo_list,
    allow_random_selection = FALSE,
    as_list = FALSE
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  data_id <- run_id <- NULL
  
  # Identify the right entry on hpo_list.
  for (ii in rev(run$run_table$perturb_level)) {
    run_id_list <- .get_iteration_identifiers(
      run = run, 
      perturb_level = ii
    )
    
    # Check whether there are any matching data and run ids by determining the
    # number of rows in the table after matching
    match_hpo <- sapply(
      hpo_list,
      function(iter_hpo, run_id_list) {
        # Determine if there are any rows in the run_table of the parameter list
        # that match the data and run identifiers of the current level.
        match_size <- nrow(iter_hpo@run_table[data_id == run_id_list$data & run_id == run_id_list$run])
        
        # Return TRUE if any matching rows are found.
        return(match_size > 0L)
      },
      run_id_list = run_id_list
    )
    
    # If there is a match, we step out of the loop
    if (any(match_hpo)) break
  }
  
  # Extract the table of parameters
  if (allow_random_selection && sum(match_hpo) > 1L) {
    random_set <- sample(which(match_hpo), size = 1L)
    
    object <- hpo_list[[random_set]]
    
  } else {
    object <- hpo_list[match_hpo][[1L]]
  }
  
  if (as_list) {
    return(object@hyperparameters)
  } else {
    return(object)
  }
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# plot_confusion_matrix (generic) ----------------------------------------------

#' @title Plot confusion matrix.
#'
#' @description This method creates confusion matrices based on data in a
#'   familiarCollection object.
#'
#' @param dir_path (*optional*) Path to the directory where created confusion
#'   matrixes are saved to. Output is saved in the `performance` subdirectory.
#'   If `NULL` no figures are saved, but are returned instead.
#' @param discrete_palette (*optional*) Palette Palette for colouring the cells
#'   of the confusion matrix. The colour depends on whether each cell of the
#'   confusion matrix is on the diagonal (observed outcome matched expected
#'   outcome) or not. `familiar` has a default palette. Other palettes are
#'   supported by the `paletteer` package, `grDevices::palette.pals()` (requires
#'   R >= 4.0.0), `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`,
#'   `heat.colors`, `terrain.colors`, `topo.colors` and `cm.colors`, which
#'   correspond to the palettes of the same name in `grDevices`. You may also
#'   specify your own palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param show_alpha (*optional*) Interpreting confusion matrices is made easier
#'   by setting the opacity of the cells. `show_alpha` takes the following
#'   values:
#'
#'  * `none`: Cell opacity is not altered. Diagonal and off-diagonal cells are
#'   completely opaque and transparent, respectively. Same as
#'   `show_alpha=FALSE`.
#'
#'  * `by_class`: Cell opacity is normalised by the number of instances for each
#'   observed outcome class in each confusion matrix.
#'
#'  * `by_matrix` (default): Cell opacity is normalised by the number of
#'   instances in the largest observed outcome class in each confusion matrix.
#'   Same as `show_alpha=TRUE`
#'
#'  * `by_figure`: Cell opacity is normalised by the number of instances in the
#'   largest observed outcome class across confusion matrices in different
#'   facets.
#'
#'  * `by_all`: Cell opacity is normalised by the number of instances in the
#'   largest observed outcome class across all confusion matrices.
#'
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#' @inheritDotParams extract_confusion_matrix -object
#'
#' @details This function generates area under the ROC curve plots.
#'
#'   Available splitting variables are: `vimp_method`, `learner` and `data_set`.
#'   By default, the data is split by `vimp_method` and `learner`, with facetting
#'   by `data_set`.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_confusion_matrix
#' @md
#' @rdname plot_confusion_matrix-methods
setGeneric(
  "plot_confusion_matrix",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    rotate_x_tick_labels = waiver(),
    show_alpha = TRUE,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_confusion_matrix")
  }
)



# plot_confusion_matrix (general) ----------------------------------------------

#' @rdname plot_confusion_matrix-methods
setMethod(
  "plot_confusion_matrix",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    rotate_x_tick_labels = waiver(),
    show_alpha = TRUE,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "confusion_matrix"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_confusion_matrix,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "rotate_x_tick_labels" = rotate_x_tick_labels,
        "show_alpha" = show_alpha,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)


# plot_confusion_matrix (collection) -------------------------------------------

#' @rdname plot_confusion_matrix-methods
setMethod(
  "plot_confusion_matrix",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    rotate_x_tick_labels = waiver(),
    show_alpha = TRUE,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    observed_outcome <- expected_outcome <- count <- class_matches <- total_observed <- NULL

    # Make sure the collection object is updated.
    object <- update_object(object = object)

    # Get input data.
    x <- export_confusion_matrix_data(object = object)

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Obtain data element from list.
    if (is.list(x)) {
      if (is_empty(x)) return(NULL)

      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_model_performance: list of data elements contains unmerged elements."
        )
      }

      # Get x directly.
      x <- x[[1L]]
    }

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to plot confusion matrices",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Check input arguments ----------------------------------------------------

    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)

    # rotate_x_tick_labels
    if (is.waive(rotate_x_tick_labels)) {
      rotate_x_tick_labels <- FALSE
    }

    outcome_name <- get_outcome_name(object)

    # x_label
    if (is.waive(x_label)) {
      x_label <- "expected"
      if (length(outcome_name) > 0L) {
        x_label <- paste(x_label, outcome_name, sep = " ")
      }
    }

    # y_label
    if (is.waive(y_label)) {
      y_label <- "observed"
      if (length(outcome_name) > 0L) {
        y_label <- paste(y_label, outcome_name, sep = " ")
      }
    }

    # show_alpha
    if (is.logical(show_alpha)) {
      show_alpha <- ifelse(show_alpha, "by_matrix", "none")
    }

    .check_parameter_value_is_valid(
      x = show_alpha,
      var_name = "show_alpha",
      values = c("none", "by_class", "by_matrix", "by_figure", "by_all")
    )

    # Splitting variables
    if (is.null(split_by) && is.null(facet_by)) {
      split_by <- c("vimp_method", "learner")
      facet_by <- c("data_set")
    }

    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      facet_by = facet_by,
      available = c("vimp_method", "learner", "data_set")
    )

    # Update splitting variables
    split_by <- split_var_list$split_by
    facet_by <- split_var_list$facet_by

    # Create a legend label
    legend_label <- .create_plot_legend_title(user_label = legend_label)

    # Check input arguments for validity.
    .check_input_plot_args(
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      y_label = y_label,
      legend_label = legend_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      rotate_x_tick_labels = rotate_x_tick_labels
    )

    # Add a class_matches column
    x@data[, "class_matches" := observed_outcome == expected_outcome]

    if (show_alpha == "none") {
      # Full opacity on diagonal, full transparancy for off-diagonal cells.
      x@data[, "alpha" := as.double(class_matches)]
      
    } else {
      # Determine the alpha level (opacity) of the fills in the confusion
      # matrix. This is determined by the number of instances of observed
      # classes.
      max_observations <- x@data[
        ,
        list("total_observed" = sum(count)),
        by = c("observed_outcome", facet_by, split_by)
      ]

      if (show_alpha == "by_class") {
        # Nothing extra needed
      } else if (show_alpha == "by_matrix") {
        # Find the maximum observations in each facet and split, i.e. normalise
        # per matrix.
        if (!is.null(facet_by) || !is.null(split_by)) {
          max_observations[, "total_observed" := max(total_observed), by = c(facet_by, split_by)]
        }
      } else if (show_alpha == "by_figure") {
        # Find the maximum observations over all facets.
        if (!is.null(split_by)) {
          max_observations[, "total_observed" := max(total_observed), by = c(split_by)]
        }
      } else if (show_alpha == "by_all") {
        # Find the maximum observations over all plots.
        max_observations[, "total_observed" := max(total_observed)]
        
      } else {
        ..error_reached_unreachable_code("plot_confusion_matrix: unknown show_alpha value.")
      }

      # Merge back into x and compute the alpha level (opacity).
      x@data <- merge(
        x = x@data,
        y = max_observations,
        by = c("observed_outcome", facet_by, split_by)
      )
      
      x@data[, "alpha" := count / total_observed]
    }
    
    
    # Create plots -------------------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data
    if (!is.null(split_by)) {
      x_split <- split(x@data, by = split_by)
    } else {
      x_split <- list("null.name" = x@data)
    }

    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over splits
    for (ii in names(x_split)) {
      # Skip empty datasets
      if (is_empty(x_split[[ii]])) next()

      if (is.waive(plot_title)) plot_title <- "Confusion matrix"

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          x = x_split[[ii]]
        )
      }

      # Generate plot
      p <- .plot_confusion_matrix_plot(
        x = x_split[[ii]],
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        rotate_x_tick_labels = rotate_x_tick_labels
      )

      # Check empty output
      if (is.null(p)) next

      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_confusion_matrix_plot_dimensions(
          x = x_split[[ii]],
          class_levels = get_outcome_class_levels(object),
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols,
          rotate_x_tick_labels = rotate_x_tick_labels
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "performance",
              "subtype" = "confusion_matrix",
              "x" = x_split[[ii]],
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_confusion_matrix_plot <- function(
    x,
    facet_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    rotate_x_tick_labels
) {
  # Find the colours
  discrete_colours <- .get_palette(
    x = discrete_palette,
    palette_type = "qualitative",
    n = 2L
  )

  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("expected_outcome"),
      y = !!sym("observed_outcome"),
      fill = !!sym("class_matches"),
      alpha = !!sym("alpha")
    )
  )
  
  # Add theme
  p <- p + ggtheme

  # Plot the heatmap
  p <- p + ggplot2::geom_raster()

  # Colour the diagonal in the heatmap
  p <- p + ggplot2::scale_fill_manual(
    values = discrete_colours,
    breaks = c(FALSE, TRUE),
    drop = FALSE
  )

  # Set alpha scale.
  p <- p + ggplot2::scale_alpha_continuous(
    range = c(0.0, 1.0),
    limits = c(0.0, 1.0)
  )

  # Labels
  p <- p + ggplot2::labs(
    x = x_label,
    y = y_label,
    title = plot_title,
    subtitle = plot_sub_title,
    caption = caption
  )

  # Obtain default settings.
  text_settings <- .get_plot_geom_text_settings(ggtheme = ggtheme)

  # Annotate the number of pairs in the figure.
  p <- p + ggplot2::geom_text(
    data = x,
    mapping = ggplot2::aes(
      x = !!sym("expected_outcome"),
      y = !!sym("observed_outcome"),
      label = !!sym("count"),
      alpha = 1.0
    ),
    colour = text_settings$colour,
    family = text_settings$family,
    fontface = text_settings$face,
    size = text_settings$geom_text_size
  )

  # Determine how things are facetted
  facet_by_list <- .parse_plot_facet_by(
    x = x,
    facet_by = facet_by,
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows,
        cols = facet_by_list$facet_cols,
        labeller = "label_context",
        drop = TRUE
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by,
        labeller = "label_context",
        drop = TRUE
      )
    }
  }

  # Rotate x-axis ticks
  if (rotate_x_tick_labels) {
    p <- p + ggplot2::theme(
      axis.text.x = ggplot2::element_text(
        vjust = 0.25,
        hjust = 1.0,
        angle = 90.0
      )
    )
  }

  # Suppress legends
  p <- p + ggplot2::theme(legend.position = "none")

  return(p)
}



.determine_confusion_matrix_plot_dimensions <- function(
    x,
    class_levels,
    facet_by,
    facet_wrap_cols,
    rotate_x_tick_labels
) {
  # Determine the number of elements along the x-axis.
  n_elements <- length(class_levels)
  longest_element <- max(sapply(class_levels, nchar))

  # Assume each x-axis element takes up about 0.5 cm. Then add some room for
  # other plot elements.
  default_width <- n_elements * 1.0 + 1.0

  # Assume each y-axis element takes up about 0.5 cm as well.
  default_height <- n_elements * 1.0 + 1.0

  # Reserve space for y-axis and x-axis tick labels. Assume that the typical
  # width of a character is about 5 points (1.8 mm). For the x-axis we only
  # reserve extra space in case the ticks are rotated, otherwise we just
  # assume a typical height of 10 points (3.6 mm).
  y_tick_space <- longest_element * 0.18
  x_tick_space <- ifelse(rotate_x_tick_labels, longest_element * 0.18, 0.36)

  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x,
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * (default_height) + x_tick_space, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm). We leave some
  # room for the legend on the right.
  width <- min(c(2.0 + plot_dims[2L] * (default_width) + y_tick_space, 19.0))

  return(c(height, width))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_shap_dependence (generic) -----------------------------------------------

#' @title Create SHAP dependence plot.
#'
#' @description This method creates a SHAP dependence plot that shows the
#'   dependence of the SHAP value of a feature against its value.
#'
#' @param dir_path (*optional*) Path to the directory where created SHAP
#'   dependence plots are saved to. Output is saved in the `explanation`
#'   subdirectory. If `NULL` no figures are saved, but are returned instead.
#' @param discrete_palette (*optional*) Divergent or sequential palette used to
#'   colour the elements of dependence plots for interactions with another
#'   **categorical** feature. `familiar` has a default palette. Other palettes
#'   are supported by the `paletteer` package, `grDevices::palette.pals()`
#'   (requires R >= 4.0.0), `grDevices::hcl.pals()` (requires R >= 3.6.0) and
#'   `rainbow`, `heat.colors`, `terrain.colors`, `topo.colors` and `cm.colors`,
#'   which correspond to the palettes of the same name in `grDevices`. You may
#'   also specify your own palette by providing a vector of colour names listed
#'   by `grDevices::colors()` or through hexadecimal RGB strings.
#'
#'   If no `interaction_feature` is set, or is a numerical feature, the gradient
#'   palette is not used.
#' @param gradient_palette (*optional*) Divergent or sequential palette used to
#'   colour the elements of dependence plots for interactions with another
#'   **numeric** feature. `familiar` has a default palette. Other palettes are
#'   supported by the `paletteer` package, `grDevices::palette.pals()` (requires
#'   R >= 4.0.0), `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`,
#'   `heat.colors`, `terrain.colors`, `topo.colors` and `cm.colors`, which
#'   correspond to the palettes of the same name in `grDevices`. You may also
#'   specify your own palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#'
#'   If no `interaction_feature` is set, or is a categorical feature, the
#'   gradient palette is not used.
#'
#' @param shap_feature (*optional*) Feature(s) whose SHAP values are used for
#'   creating the SHAP dependence plot.
#' @param interaction_feature (*optional*) Feature(s) whose values are used to
#'   colour points of the `shap_feature`.
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams extract_performance -object
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#'
#' @details This function creates SHAP dependence plots, which show how the
#'   marginal contributions of a feature to the predicted value depend on its
#'   value.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, `data_set`,
#'   `evaluation_time` (survival outcome only) and `positive_class` (categorical
#'   outcomes). The default is to facet by `evaluation_time` or
#'   `positive_class`, and split by `vimp_method`, `learner` and `data_set`.
#'   `color_by` is not used.
#'
#'   Labelling methods such as `set_vimp_method_names` or `set_learner_names`
#'   can be applied to the `familiarCollection` object to update labels, and
#'   order the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_shap_dependence
#' @md
#' @rdname plot_shap_dependence-methods
setGeneric(
  "plot_shap_dependence",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    shap_feature = NULL,
    interaction_feature = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_shap_dependence")
  }
)



# plot_shap_dependence (general) -----------------------------------------------

#' @rdname plot_shap_dependence-methods
setMethod(
  "plot_shap_dependence",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    shap_feature = NULL,
    interaction_feature = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "shap"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_shap_dependence,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "discrete_palette" = discrete_palette,
        "gradient_palette" = gradient_palette,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "x_range" = x_range,
        "x_n_breaks" = x_n_breaks,
        "x_breaks" = x_breaks,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "shap_feature" = shap_feature,
        "interaction_feature" = interaction_feature,
        "width" = width,
        "height" = height,
        "units" = units,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_shap_dependence (collection) --------------------------------------------

#' @rdname plot_shap_dependence-methods
setMethod(
  "plot_shap_dependence",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    x_range = NULL,
    x_n_breaks = 5L,
    x_breaks = NULL,
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    shap_feature = NULL,
    interaction_feature = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    export_collection = FALSE,
    ...
  ) {
    # Prevent NOTES.
    .NATURAL <- NULL
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Check input arguments ----------------------------------------------------
    
    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)
    
    # Load the data.
    x <- export_shap(object = object)
    x <- x$shap_summary
    if (is_empty(x)) return(NULL)
    
    # Obtain single data element from list.
    if (is.list(x)) {
      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_shap_dependence: list of data elements contains unmerged elements."
        )
      }
      x <- x[[1L]]
    }
    
    # Check that the data are not evaluated at the model level.
    if (x@detail_level == "model") {
      ..warning_no_comparison_between_models()
      return(NULL)
    }
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    # Ensure that we work with a copy of the data.
    x@data <- data.table::copy(x@data)
    
    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to create a SHAP dependence plot",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Add evaluation time or positive class as a splitting variable.
    additional_variable <- NULL
    if (object@outcome_type %in% c("survival")) {
      additional_variable <- "evaluation_time"
      data.table::setnames(x@data, old = "shap_outcome", new = "evaluation_time")
      
    } else if (object@outcome_type %in% c("multinomial")) {
      additional_variable <- "positive_class"
      data.table::setnames(x@data, old = "shap_outcome", new = "positive_class")
    }
    
    # Add default splitting variables.
    if (
      is.null(split_by) &&
      is.null(facet_by)
    ) {
      # Split by vimp_method, learner.
      split_by <- c("vimp_method", "learner", "data_set", "feature_name")
      facet_by <- additional_variable
    }
    all_variables <- c("vimp_method", "learner", "data_set", additional_variable, "feature_name")
    
    # Check splitting variables and generate sanitised output
    split_var_list <- .check_plot_splitting_variables(
      x = x@data,
      split_by = split_by,
      facet_by = facet_by,
      available = all_variables
    )
    
    # Update splitting variables
    split_by <- split_var_list$split_by
    facet_by <- split_var_list$facet_by
    
    # y_label
    if (is.waive(y_label)) {
      y_label <- "SHAP value"
    }
    
    .check_input_plot_args(
      facet_wrap_cols = facet_wrap_cols,
      y_label = y_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption
    )
    
    # Determine which features to use.
    selected_shap_features <- levels(x@data$feature_name)
    if (!is.null(shap_feature)) {
      feature_in_data <- shap_feature %in% selected_shap_features
      if (!all(feature_in_data)) {
        ..warning(paste0(
          "Not all features in shap_feature could be found in the data. Missing: ",
          paste_s(shap_feature[!feature_in_data])
        ))
      }
      selected_shap_features <- intersect(selected_shap_features, shap_feature)
    }
    if (length(selected_shap_features) == 0L) return(NULL)
    
    # Determine which interaction features exist.
    selected_interaction_features <- NULL
    if (!is.null(interaction_feature)) {
      feature_in_data <- interaction_feature %in% levels(x@data$feature_name)
      if (!all(feature_in_data)) {
        ..warning(paste0(
          "Not all features in interaction_feature could be found in the data. Missing: ",
          paste_s(interaction_feature[!feature_in_data])
        ))
      }
      selected_interaction_features <- intersect(interaction_feature, levels(x@data$feature_name))
      if (length(selected_interaction_features) == 0L) interaction_features <- NULL
    }
    
    # Create plots -------------------------------------------------------------
    
    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)
    
    # Split data.
    if (!is.null(split_by)) {
      x_split <- split(
        x@data, 
        by = split_by, 
        drop = FALSE
      )
      
    } else {
      x_split <- list("null.name" = x@data)
    }
    
    # Select plot data based on split.
    plot_data <- list()
    plot_data_index <- 1L
    
    for (ii in seq_along(x_split)) {
      main_data <- x_split[[ii]]
      if (is_empty(main_data)) next
      
      # Skip if current feature is not in selected_shap_features.
      current_shap_feature <- main_data$feature_name[1L]
      if (!current_shap_feature %in% selected_shap_features) next
      
      interaction_data <- NULL
      if (length(selected_interaction_features) > 0L) {
        for (current_interaction_feature in selected_interaction_features) {
          interaction_selection_data <- unique(main_data[, mget(split_by)])
          interaction_selection_data[, "feature_name" := current_interaction_feature]
          interaction_data <- x@data[interaction_selection_data, on = .NATURAL]
          
          plot_data[[plot_data_index]] <- list(
            "main_data" = main_data,
            "interaction_data" = interaction_data,
            "feature" = as.character(current_shap_feature),
            "interaction_feature" = as.character(current_interaction_feature)
          )
          plot_data_index <- plot_data_index + 1L
        }
        
      } else {
        plot_data[[plot_data_index]] <- list(
          "main_data" = main_data,
          "interaction_data" = NULL,
          "feature" = as.character(current_shap_feature),
          "interaction_feature" = NULL
        )
        plot_data_index <- plot_data_index + 1L
      }
    }
    
    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()
    
    # Iterate over data splits.
    for (current_data in plot_data) {
      if (is.waive(plot_title)) plot_title <- "SHAP dependence"
      
      # Declare subtitle components.
      additional_subtitle <- NULL
      
      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(current_data$main_data$evaluation_time[1L])
        )
      }
      
      # Add feature as subtitle component if it is not used otherwise.
      if (!"feature_name" %in% c(split_by)) {
        additional_subtitle <- c(
          additional_subtitle,
          list("feature" = current_data$feature)
        )
      }
      
      # Add interaction feature as subtitle component.
      if (!is.null(current_data$interaction_feature)) {
        additional_subtitle <- c(
          additional_subtitle,
          list("interaction" = current_data$interaction_feature)
        )
      }
      
      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = current_data$main_data
        )
      }
      
      # Generate plot
      p <- .plot_shap_dependence_plot(
        x = current_data$main_data,
        x_feature = current_data$feature,
        z = current_data$interaction_data,
        z_feature = current_data$interaction_feature,
        split_by = split_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        gradient_palette = gradient_palette,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        x_range = x_range,
        x_n_breaks = x_n_breaks,
        x_breaks = x_breaks,
        y_range = y_range,
        y_n_breaks = y_n_breaks,
        y_breaks = y_breaks
      )
      
      # Check empty output
      if (is.null(p)) next
      
      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)
      
      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_shap_dependence_plot_dimensions(
          x = current_data$main_data,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols
        )
        
        # Set subtype
        subtype <- "shap_dependence"
        if (!is.null(current_data$interaction_feature)){
          subtype <- paste0(subtype, "int_", current_data$interaction_feature)
        }
        
        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "explanation",
              "subtype" = subtype,
              "x" = current_data$main_data,
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }
    
    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_shap_dependence_plot <- function(
    x,
    x_feature,
    z,
    z_feature,
    facet_by,
    split_by,
    facet_wrap_cols,
    ggtheme,
    discrete_palette,
    gradient_palette,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    x_range,
    x_n_breaks,
    x_breaks,
    y_range,
    y_n_breaks,
    y_breaks
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  shap_value <- vimp <- feature_value <- NULL

  # Check that the interaction feature is not the same as the main feature.
  if (!is.null(z_feature)) {
     if (z_feature == x_feature) z <- NULL
  }
  
  # Make local copies to prevent updating by reference.
  x <- data.table::copy(x)
  if (!is_empty(z)) z <- data.table::copy(z)
  
  # x_label
  if (is.waive(x_label)) {
    x_label <- paste0(x_feature, " value")
  }
  
  # Check if the main feature is numeric.
  x_numeric <- all(is.na(x$feature_label))
  z_numeric <- TRUE
  if (!is_empty(z)) z_numeric <- all(is.na(z$feature_label))
  
  if (x_numeric) {
    # Check x_range.
    if (is.null(x_range)) {
      x_range <- range(x$feature_value, na.rm = TRUE)
      
      if (diff(x_range) == 0.0) {
        x_range <- c(x_range[1L] - 0.01, x_range[2L] + 0.01)
      }
      
    } else {
      .check_input_plot_args(x_range = x_range)
    }
    
    # x_breaks
    if (is.null(x_breaks)) {
      .check_input_plot_args(
        x_range = x_range,
        x_n_breaks = x_n_breaks
      )
      
      # Create breaks and update x_range
      x_breaks <- labeling::extended(
        m = x_n_breaks,
        dmin = x_range[1L],
        dmax = x_range[2L],
        only.loose = TRUE
      )
      
      x_range <- c(
        head(x_breaks, n = 1L),
        tail(x_breaks, n = 1L)
      )
      
    } else {
      .check_input_plot_args(x_breaks = x_breaks)
    }
  }
  
  # Check y_range.
  if (is.null(y_range)) {
    y_range <- range(x$shap_value, na.rm = TRUE)
    
    if (diff(y_range) == 0.0) {
      y_range <- c(y_range[1L] - 0.01, y_range[2L] + 0.01)
    }
    
  } else {
    .check_input_plot_args(y_range = y_range)
  }
  
  # y_breaks
  if (is.null(y_breaks)) {
    .check_input_plot_args(
      y_range = y_range,
      y_n_breaks = y_n_breaks
    )
    
    # Create breaks and update x_range
    y_breaks <- labeling::extended(
      m = y_n_breaks,
      dmin = y_range[1L],
      dmax = y_range[2L],
      only.loose = TRUE
    )
    
    y_range <- c(
      head(y_breaks, n = 1L),
      tail(y_breaks, n = 1L)
    )
    
  } else {
    .check_input_plot_args(y_breaks = y_breaks)
  }
  
  # Create a legend label.
  if (is.waive(legend_label) && !is_empty(z)) {
    legend_label <- z_feature
  }
  
  # Check remaining input arguments.
  .check_input_plot_args(
    x_label = x_label,
    legend_label = legend_label
  )
  
  # Set numeric or categorical feature values.
  if (x_numeric) {
    x[, "x_value" := feature_value]
  } else {
    feature_labels <- unique(x[, mget(c("feature_value", "feature_label"))])[order(feature_value)]
    x[, "x_value" := factor(
      feature_value,
      levels = feature_labels$feature_value,
      labels = feature_labels$feature_label
    )]
  }
  
  # Set numeric or categorical feature values for the interaction feature.
  if (!is_empty(z)) {
    if (z_numeric) {
      z[, "z_value" := feature_value]
    } else {
      feature_labels <- unique(z[, mget(c("feature_value", "feature_label"))])[order(feature_value)]
      z[, "z_value" := factor(
        feature_value,
        levels = feature_labels$feature_value,
        labels = feature_labels$feature_label
      )]
    }
    
    split_by <- setdiff(split_by, "feature_name")
    x <- merge(
      x = x, 
      y = z[, mget(c(split_by, facet_by, "sample_id", "z_value"))],
      by = c(split_by, facet_by, "sample_id")
    )
  }
  
  mapping <- ggplot2::aes(
    x = !!sym("x_value"),
    y = !!sym("shap_value")
  )
  if (!is_empty(z)) {
    mapping <- ggplot2::aes(
      x = !!sym("x_value"),
      y = !!sym("shap_value"),
      colour = !!sym("z_value")
    )
  }
  
  # Create basic plot
  p <- ggplot2::ggplot(
    data = x,
    mapping = mapping
  )
  p <- p + ggtheme
  
  # Set breaks and range.
  if (x_numeric) p <- p + ggplot2::scale_x_continuous(breaks = x_breaks, limits = x_range)
  p <- p + ggplot2::scale_y_continuous(breaks = y_breaks, limits = y_range)
  
  # Set main plot type.
  if (x_numeric) {
    p <- p + ggplot2::geom_point()
  } else {
    # Jitter only possible along x-axis for categorical features.
    p <- p + ggplot2::geom_jitter(height = 0.0)
  }
  
  
  # Set colours for interactions.
  if (!is_empty(z) && z_numeric) {
    # Gradient palette for numeric interaction features.
    gradient_colours <- .get_palette(
      x = gradient_palette, 
      palette_type = "divergent"
    )
    
    p <- p + ggplot2::scale_colour_gradientn(
      name = legend_label,
      colors = gradient_colours
    )
    
  } else if (!is_empty(z) && !z_numeric) {
    # Discrete palette for categorical interaction features.
    discrete_colours <- .get_palette(
      x = discrete_palette, 
      palette_type = "qualitative",
      n = nlevels(x$z_value)
    )
    
    p <- p + ggplot2::scale_colour_manual(
      name = legend_label,
      values = discrete_colours,
      breaks = levels(x$z_value)
    )
  }
  
  # Determine how things are faceted.
  facet_by_list <- .parse_plot_facet_by(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )
  
  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows, 
        cols = facet_by_list$facet_cols, 
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by, 
        labeller = "label_context"
      )
    }
  }
  
  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label, 
    y = y_label, 
    title = plot_title, 
    subtitle = plot_sub_title, 
    caption = caption
  )
  
  return(p)
}



.determine_shap_dependence_plot_dimensions <- function(
    x,
    x_axis_by,
    y_axis_by,
    facet_by,
    facet_wrap_cols
) {
  
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )
  
  # Set default height and width for each subplot (in cm).
  default_width <- 6.0
  default_height <- 4.0
  
  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height, 27.7))
  
  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width, 19.0))
  
  return(c(height, width))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# export_all methods -----------------------------------------------------------

## export_all (generic) --------------------------------------------------------

#'@title Extract and export all data.
#'
#'@description Extract and export all data from a familiarCollection.
#'
#'@param object A `familiarCollection` object, or other other objects from which
#'  a `familiarCollection` can be extracted. See details for more information.
#'@param dir_path Path to folder where extracted data should be saved. `NULL`
#'  will allow export as a structured list of data.tables.
#'@param aggregate_results Flag that signifies whether results should be
#'  aggregated for export.
#'
#'@inheritDotParams .extract_data
#'@inheritDotParams as_familiar_collection
#'
#'@details Data, such as model performance and calibration information, is
#'  usually collected from a `familiarCollection` object. However, you can also
#'  provide one or more `familiarData` objects, that will be internally
#'  converted to a `familiarCollection` object. It is also possible to provide a
#'  `familiarEnsemble` or one or more `familiarModel` objects together with the
#'  data from which data is computed prior to export. Paths to the previous
#'  files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'@return A list of data.tables (if `dir_path` is not provided), or nothing, as
#'  all data is exported to `csv` files.
#'@exportMethod export_all
#'@md
#'@rdname export_all-methods
setGeneric(
  "export_all",
  function(
    object, 
    dir_path = NULL, 
    aggregate_results = waiver(),
    ...
  ) {
    standardGeneric("export_all")
  }
)


## export_all (collection) -----------------------------------------------------

#'@rdname export_all-methods
setMethod(
  "export_all",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = waiver(),
    ...
  ) {
    
    if (is.waive(aggregate_results)) aggregate_results <- TRUE
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Export feature selection variable importance
    fs_vimp <- export_fs_vimp(
      object = object,
      dir_path = dir_path
    )
    
    # Export feature selection variable importance using stability.
    fs_vimp_occurrence <- export_fs_vimp(
      object = object,
      dir_path = dir_path,
      aggregation_method = "stability"
    )
    
    # Export model variable importance
    model_vimp <- export_model_vimp(
      object = object,
      dir_path = dir_path
    )
    
    # Export model variable importance using stability.
    model_vimp_occurrence <- export_model_vimp(
      object = object,
      dir_path = dir_path,
      aggregation_method = "stability"
    )
    
    # Export permutation variable importance.
    permutation_vimp <- export_permutation_vimp(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    # Export model hyperparameters
    hyperparameters <- export_hyperparameters(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    # Export prediction tables
    prediction_data <- export_prediction_data(
      object = object,
      dir_path = dir_path
    )
    
    # Export decision curve analysis data
    dca_data <- export_decision_curve_analysis_data(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    # Export calibration information
    calibration_info <- export_calibration_info(
      object = object,
      dir_path = dir_path
    )
    
    # Export calibration data
    calibration_data <- export_calibration_data(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    # Export model performance
    model_performance <- export_model_performance(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    # Export confusion matrix
    confusion_matrix <- export_confusion_matrix_data(
      object = object,
      dir_path = dir_path
    )
    
    # Export kaplan-meier info
    km_info <- export_risk_stratification_info(
      object = object,
      dir_path = dir_path
    )
    
    # Export stratification data
    km_data <- export_risk_stratification_data(
      object = object,
      dir_path = dir_path
    )
    
    # Export AUC data
    auc_data <- export_auc_data(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    # Export data from the univariate analysis
    univariate_analysis <- export_univariate_analysis_data(
      object = object,
      dir_path = dir_path
    )
    
    # Export data from feature expressions
    feature_expressions <- export_feature_expressions(
      object = object,
      dir_path = dir_path
    )
    
    # Export mutual-correlation data
    feature_similarity <- export_feature_similarity(
      object = object,
      dir_path = dir_path
    )
    
    # Export partial dependence data
    pd_data <- export_partial_dependence_data(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    # Export individual conditional expectation data
    ice_data <- export_ice_data(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    # Export SHAP data.
    shap_data <- export_shap(
      object = object,
      dir_path = dir_path,
      aggregate_results = aggregate_results
    )
    
    if (is.null(dir_path)) {
      return(list(
        "fs_vimp" = list(
          "default" = fs_vimp,
          "occurrence" = fs_vimp_occurrence
        ),
        "model_vimp" = list(
          "default" = model_vimp,
          "occurrence" = model_vimp_occurrence
        ),
        "permutation_vimp" = permutation_vimp,
        "hyperparameters" = hyperparameters,
        "prediction_data" = prediction_data,
        "calibration_info" = calibration_info,
        "calibration_data" = calibration_data,
        "model_performance" = model_performance,
        "confusion_matrix" = confusion_matrix,
        "decision_curve" = dca_data,
        "km_info" = km_info,
        "km_data" = km_data,
        "auc_data" = auc_data,
        "univariate_analysis" = univariate_analysis,
        "feature_expressions" = feature_expressions,
        "feature_similarity" = feature_similarity,
        "pd_data" = pd_data,
        "ice_data" = ice_data,
        "shap_data" = shap_data
      ))
    }
  }
)



## export_all (general) --------------------------------------------------------

#'@rdname export_all-methods
setMethod(
  "export_all",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = waiver(),
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_all,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    ))
  }
)



# export_to_file methods ------------------------------------------------------

## .export_to_file (familiarDataElement) ---------------------------------------
setMethod(
  ".export_to_file",
  signature(
    data = "familiarDataElement",
    object = "familiarCollection",
    dir_path = "character"
  ),
  function(
    data,
    object,
    dir_path,
    type,
    subtype = NULL
  ) {
    
    if (is_empty(data)) return(NULL)
    
    # Check if any identifiers remain, and add to the subtype.
    if (length(data@identifiers) > 0L) subtype <- c(subtype, unlist(data@identifiers))
    
    return(.export_to_file(
      data = data@data,
      object = object,
      dir_path = dir_path,
      type = type,
      subtype = subtype
    ))
  }
)


## .export_to_file (list) ------------------------------------------------------
setMethod(
  ".export_to_file",
  signature(
    data = "list",
    object = "familiarCollection",
    dir_path = "character"
  ),
  function(
    data,
    object, 
    dir_path, 
    type, 
    subtype = NULL
  ) {
    
    # Check if data exists
    if (is_empty(data)) return(NULL)
    
    return(lapply(
      data,
      .export_to_file,
      object = object,
      dir_path = dir_path,
      type = type,
      subtype = subtype
    ))
  }
)



## .export_to_file (data.table) ------------------------------------------------
setMethod(
  ".export_to_file",
  signature(
    data = "data.table",
    object = "familiarCollection",
    dir_path = "character"
  ),
  function(
    data,
    object,
    dir_path,
    type, 
    subtype = NULL
  ) {
    
    # Check if data exists.
    if (is_empty(data)) return(NULL)
    
    # Check if directory exists.
    file_dir <- normalizePath(file.path(dir_path, type), mustWork = FALSE)
    if (!dir.exists(file_dir)) dir.create(file_dir, recursive = TRUE)
    
    # Generate file name.
    if (length(subtype) == 0L) {
      base_file_name <- type
      
    } else {
      base_file_name <- paste(type, paste0(subtype, collapse = "_"), sep = "_")
    }
    
    file_name <- file.path(file_dir, paste0(base_file_name, ".csv"))
    
    # Write data to file.
    data.table::fwrite(
      x = data,
      file = file_name,
      sep = ";",
      dec = "."
    )
    
    return(invisible(NULL))
  }
)



## .export_to_file (character) -------------------------------------------------
setMethod(
  ".export_to_file",
  signature(
    data = "character",
    object = "familiarCollection",
    dir_path = "character"
  ),
  function(
    data, 
    object, 
    dir_path, 
    type, 
    subtype = NULL
  ) {
    
    # Check if data exists
    if (is_empty(data)) return(NULL)
    
    # Check if directory exists
    file_dir <- normalizePath(file.path(dir_path, type), mustWork = FALSE)
    if (!dir.exists(file_dir)) dir.create(file_dir, recursive = TRUE)
    
    # Generate file name
    if (length(subtype) == 0L) {
      base_file_name <- type
      
    } else {
      base_file_name <- paste(type, paste0(subtype, collapse = "_"), sep = "_")
    }
    
    file_name <- file.path(file_dir, paste0(base_file_name, ".txt"))
    
    # Write to text file with each element of x on a new line.
    write(
      x = data,
      file = file_name,
      append = FALSE,
      sep = ifelse(.Platform$OS.type == "windows", "\r\n", "\n")
    )
    
    return(NULL)
  }
)



# .apply_labels methods --------------------------------------------------------

## .apply_labels (familiarDataElement, familiarCollection) ---------------------
setMethod(
  ".apply_labels",
  signature(
    data = "familiarDataElement",
    object = "familiarCollection"
  ),
  function(data, object) {
    
    # Return NULL for empty input
    if (is_empty(data)) return(NULL)
    
    # Don't update data if it is not a data.table.
    if (!data.table::is.data.table(data@data)) return(data)
    
    # Make sure that a local copy is updated
    x <- data.table::copy(data@data)
    
    # Check which labels are present, based on column names
    columns <- colnames(x)
    
    # Determine whether certain columns are present.
    has_data_set <- "data_set" %in% columns
    has_learner <- "learner" %in% columns
    has_vimp_method <- "vimp_method" %in% columns
    has_feature <- any(c("name", "feature_name", "feature_name_1", "feature_name_2", "feature", "feature_x", "feature_y") %in% columns)
    has_risk_group <- any(c("reference_group", "group", "group_1", "group_2") %in% columns)
    has_multiclass_outcome <- any(c("pos_class", "positive_class", "outcome", "shap_outcome") %in% columns) &&
      object@outcome_type == "multinomial"
    has_categorical_outcome <- any(c("observed_outcome", "expected_outcome") %in% columns) &&
      object@outcome_type %in% c("binomial", "multinomial")
    has_evaluation_time <- any(c("evaluation_time", "eval_time", "shap_outcome") %in% columns) &&
      object@outcome_type %in% c("survival", "competing_risk")
    has_performance_metric <- any(c("metric") %in% columns)
    has_model_name <- any(c("ensemble_model_name", "model_name") %in% columns)
    
    # Apply levels
    if (has_data_set) {
      data.table::set(
        x,
        j = "data_set",
        value = factor(
          x = x$data_set,
          levels = get_data_set_name_levels(x = object),
          labels = get_data_set_names(x = object)
        )
      )
    }
    
    if (has_learner) {
      data.table::set(
        x,
        j = "learner",
        value = factor(
          x = x$learner,
          levels = get_learner_name_levels(x = object),
          labels = get_learner_names(x = object)
        )
      )
    }
    
    if (has_vimp_method) {
      data.table::set(
        x,
        j = "vimp_method",
        value = factor(
          x = x$vimp_method,
          levels = get_vimp_method_name_levels(x = object),
          labels = get_vimp_method_names(x = object)
        )
      )
    }
    
    if (has_feature) {
      for (current_column_name in c("name", "feature_name", "feature_name_1", "feature_name_2", "feature", "feature_x", "feature_y")) {
        
        if (!is.null(x[[current_column_name]])) {
          # Check if all feature names are actually in the object. Some features
          # may be missing for e.g. variable importance because they were not
          # required for the model.
          if (all(unique(x[[current_column_name]]) %in% get_feature_name_levels(x = object))) {
            data.table::set(
              x,
              j = current_column_name,
              value = factor(
                x = x[[current_column_name]],
                levels = get_feature_name_levels(x = object),
                labels = get_feature_names(x = object)
              )
            )
            
          } else {
            data.table::set(
              x,
              j = current_column_name,
              value = factor(
                x = x[[current_column_name]],
                levels = unique(x[[current_column_name]])
              )
            )
          }
        }
      }
    }
    
    if (has_risk_group) {
      for (current_column_name in c(
        "reference_group", "group", "group_1", "group_2"
      )) {
        
        if (!is.null(x[[current_column_name]])) {
          data.table::set(
            x,
            j = current_column_name,
            value = factor(
              x = x[[current_column_name]],
              levels = get_risk_group_name_levels(x = object),
              labels = get_risk_group_names(x = object)
            )
          )
        }
      }
    }
    
    if (has_multiclass_outcome) {
      for (current_column_name in c("pos_class", "positive_class", "outcome", "shap_outcome")) {
        
        if (!is.null(x[[current_column_name]])) {
          data.table::set(
            x,
            j = current_column_name,
            value = factor(
              x = x[[current_column_name]],
              levels = get_class_name_levels(x = object),
              labels = get_class_names(x = object)
            )
          )
        }
      }
    }
    
    if (has_categorical_outcome) {
      # For confusion matrices.
      for (current_column_name in c("observed_outcome", "expected_outcome")) {
        
        if (!is.null(x[[current_column_name]])) {
          data.table::set(
            x,
            j = current_column_name,
            value = factor(
              x = x[[current_column_name]],
              levels = get_class_name_levels(x = object),
              labels = get_class_names(x = object)
            )
          )
        }
      }
    }
    
    if (has_evaluation_time) {
      for (current_column_name in c("evaluation_time", "eval_time", "shap_outcome")) {
        
        if (!is.null(x[[current_column_name]])) {
          data.table::set(
            x,
            j = current_column_name,
            value = factor(
              x = x[[current_column_name]],
              levels = sort(unique(x[[current_column_name]]))
            )
          )
        }
      }
    }
    
    if (has_performance_metric) {
      for (current_column_name in c("metric")) {
        if (!is.null(x[[current_column_name]])) {
          data.table::set(
            x,
            j = current_column_name,
            value = factor(
              x = x[[current_column_name]],
              levels = sort(unique(x[[current_column_name]]))
            )
          )
        }
      }
    }
    
    if (has_model_name) {
      for (current_column_name in c("ensemble_model_name", "model_name")) {
        if (!is.null(x[[current_column_name]])) {
          data.table::set(
            x,
            j = current_column_name,
            value = factor(
              x = x[[current_column_name]],
              levels = sort(unique(x[[current_column_name]]))
            )
          )
        }
      }
    }
    
    # Order columns. Grouping columns appear on the left, whereas value
    # columns appear on the right. First we identify the grouping
    # columns. Note that not all 
    grouping_columns <- c(
      "data_set", "vimp_method", "learner",
      "ensemble_model_name", "model_name",
      "evaluation_time", "eval_time",
      "name", "feature_name", "feature_name_1", "feature_name_2", "feature",
      "pos_class", "positive_class",
      "metric"
    )
    
    # Find the grouping columns actually present
    grouping_columns <- intersect(grouping_columns, columns)
    
    # Then identify the value columns.
    value_columns <- data@value_column
    if (all(is.na(value_columns))) value_columns <- NULL
    
    # Find any remaining columns.
    remaining_columns <- setdiff(columns, c(grouping_columns, value_columns))
    
    # Order columns.
    data.table::setcolorder(
      x = x,
      neworder = c(grouping_columns, remaining_columns, value_columns)
    )
    
    # Drop unused levels.
    x <- droplevels(x)
    
    # Replace data attribute.
    data@data <- x
    
    return(data)
  }
)



## .apply_labels (list, familiarCollection) ------------------------------------
setMethod(
  ".apply_labels",
  signature(
    data = "list",
    object = "familiarCollection"
  ),
  function(data, object) {
    
    return(lapply(
      data,
      .apply_labels,
      object = object
    ))
  }
)



# .apply_labels (ANY, familiarCollection) --------------------------------------
setMethod(
  ".apply_labels",
  signature(
    data = "ANY",
    object = "familiarCollection"
  ),
  function(data, object) {
    # This is the fall-back option for empty data.
    
    return(NULL)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R

setClass("familiarHyperparameterLearnerRanger",
  contains = "familiarHyperparameterLearner"
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize", 
  signature(.Object = "familiarHyperparameterLearnerRanger"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- "ranger"

    # Set name
    .Object@name <- "Random Forest (ranger) hyperparameter optimisation model"

    return(.Object)
  }
)



.get_available_ranger_hyperparameter_learners <- function() {
  return("random_forest")
}



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarHyperparameterLearnerRanger"),
  function(object, ...) {
    return(c("default", "sd", "percentile", "raw"))
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarHyperparameterLearnerRanger",
    data = "data.table"
  ),
  function(object, data, ...) {
    # Check if the training data is ok.
    if (has_bad_training_data(object = object, data = data)) {
      return(callNextMethod())
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Parse formula.
    formula <- stats::reformulate(
      termlabels = object@target_hyperparameters,
      response = "optimisation_score"
    )

    # Set hyperparameters of the random forest.
    n_tree <- 400L
    n_train <- nrow(data)
    sample_fraction <- max(c(0.3, min(c(1.0, 1.0 / (0.025 * n_train)))))

    # Train a conventional random forest
    model <- ranger::ranger(formula,
      data = data,
      num.trees = n_tree,
      sample.fraction = sample_fraction,
      num.threads = 1L,
      verbose = FALSE
    )

    # Add model
    object@model <- model

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict", signature(
    object = "familiarHyperparameterLearnerRanger",
    data = "data.table"
  ),
  function(object, data, type = "default", percentile = NULL, ...) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) return(callNextMethod())

    # Check if the data is empty.
    if (is_empty(data)) return(callNextMethod())

    # Get an empty prediction table.
    prediction_table <- .get_placeholder_hyperparameter_prediction_table(
      object = object,
      data = data,
      type = type
    )

    # Make predictions.
    predicted_scores <- predict(
      object = object@model,
      data = data,
      predict.all = TRUE,
      num.threads = 1L,
      verbose = FALSE
    )$predictions

    # Compute mean and standard deviation.
    score_mean <- apply(predicted_scores, MARGIN = 1L, mean)
    score_sd <- apply(predicted_scores, MARGIN = 1L, stats::sd)

    # Separate by type
    if (type == "default") {
      prediction_table[, "mu" := score_mean]
      
    } else if (type == "sd") {
      prediction_table[, ":="(
        "mu" = score_mean,
        "sigma" = score_sd
      )]
      
    } else if (type == "percentile") {
      # Compute the requested percentile.
      score_percentile <- apply(
        predicted_scores, 
        MARGIN = 1L,
        stats::quantile,
        probs = percentile,
        names = FALSE
      )

      prediction_table[, "percentile" := score_percentile]
      
    } else if (type == "raw") {
      # Drop raw_1 from the placeholder prediction table.
      prediction_table[, "raw_1" := NULL]

      # Parse raw data to a data.table with the expected output.
      raw_data <- data.table::data.table(predicted_scores)

      # Set colnames.
      data.table::setnames(
        x = raw_data,
        new = paste0("raw_", seq_len(ncol(raw_data)))
      )

      # Combine with the placeholder prediction table.
      prediction_table <- cbind(prediction_table, raw_data)
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "..predict,familiarHyperparameterLearnerRanger,data.table: ",
        "Encountered an unknown prediction type: ", type
      ))
    }

    return(prediction_table)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarDataElementPermutationVimp object ------------------------------------
setClass(
  "familiarDataElementPermutationVimp",
  contains = "familiarDataElement",
  slots = list("similarity_metric" = "character"),
  prototype = methods::prototype(
    value_column = "value",
    grouping_column = c("feature", "metric"),
    similarity_metric = NA_character_
  )
)


# extract_permutation_vimp (generic) -------------------------------------------

#'@title Internal function to extract permutation variable importance.
#'
#'@description Computes and collects permutation variable importance from a
#'  `familiarEnsemble`.
#'
#'@inheritParams .extract_data
#'
#'@details This function also computes credibility intervals for the ensemble
#'  model, at the level of `confidence_level`.
#'
#'@return A list with data.tables for single and ensemble model assessments.
#'@md
#'@keywords internal
setGeneric(
  "extract_permutation_vimp",
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    feature_similarity,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_metric = waiver(),
    feature_similarity_threshold = waiver(),
    metric = waiver(),
    evaluation_times = waiver(),
    sample_limit = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    n_important_features = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_permutation_vimp")
  }
)



# extract_permutation_vimp (familiarEnsemble) ----------------------------------
setMethod(
  "extract_permutation_vimp",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    cl = NULL,
    ensemble_method = waiver(),
    feature_similarity,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_cluster_cut_method = waiver(),
    feature_similarity_metric = waiver(),
    feature_similarity_threshold = waiver(),
    metric = waiver(),
    evaluation_times = waiver(),
    sample_limit = waiver(),
    detail_level = waiver(),
    estimation_type = waiver(),
    aggregate_results = waiver(),
    confidence_level = waiver(),
    bootstrap_ci_method = waiver(),
    n_important_features = waiver(),
    is_pre_processed = FALSE,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    # Message extraction start
    logger_message(
      paste0("Computing permutation variable importance for models in the dataset."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Load evaluation_times from the object settings attribute, if it is not
    # provided.
    if (is.waive(evaluation_times)) {
      evaluation_times <- object@settings$eval_times
    }
    
    # Check evaluation_times argument
    if (object@outcome_type %in% c("survival")) {
      sapply(
        evaluation_times, 
        .check_number_in_valid_range, 
        var_name = "evaluation_times",
        range = c(0.0, Inf),
        closed = c(FALSE, TRUE)
      )
    }
    
    # Obtain ensemble method from stored settings, if required.
    if (is.waive(ensemble_method)) {
      ensemble_method <- object@settings$ensemble_method
    } 
    
    # Load confidence alpha from object settings attribute if not provided
    # externally.
    if (is.waive(confidence_level)) {
      confidence_level <- object@settings$confidence_level
    }
    
    # Check confidence_level input argument
    .check_number_in_valid_range(
      x = confidence_level,
      var_name = "confidence_level",
      range = c(0.0, 1.0),
      closed = c(FALSE, FALSE)
    )
    
    # Check ensemble_method argument
    .check_parameter_value_is_valid(
      x = ensemble_method,
      var_name = "ensemble_method",
      values = .get_available_ensemble_prediction_methods()
    )
    
    # Load the bootstrap method
    if (is.waive(bootstrap_ci_method)) {
      bootstrap_ci_method <- object@settings$bootstrap_ci_method
    }
    
    .check_parameter_value_is_valid(
      x = bootstrap_ci_method,
      var_name = "bootstrap_ci_method",
      values = .get_available_bootstrap_confidence_interval_methods()
    )
    
    # Check the level detail.
    detail_level <- .parse_detail_level(
      x = detail_level,
      object = object,
      default = "hybrid",
      data_element = "permutation_vimp"
    )
    
    # Check the estimation type.
    estimation_type <- .parse_estimation_type(
      x = estimation_type,
      object = object,
      default = "bootstrap_confidence_interval",
      data_element = "permutation_vimp",
      detail_level = detail_level,
      has_internal_bootstrap = TRUE
    )
    
    # Check whether results should be aggregated.
    aggregate_results <- .parse_aggregate_results(
      x = aggregate_results,
      object = object,
      default = TRUE,
      data_element = "permutation_vimp"
    )
    
    # Check the sample limit. This defines the subset of samples that are being
    # assessed (if real data is used instead of a minimum subset).
    sample_limit <- .parse_sample_limit(
      x = sample_limit,
      object = object,
      default = 200L,
      data_element = "permutation_vimp"
    )
    
    # Load metric(s) from the object settings attribute if not provided
    # externally.
    if (is.waive(metric)) metric <- object@settings$metric
    
    # Check metric input argument
    sapply(
      metric,
      .check_metric_outcome_type,
      object = object
    )
    
    # Aggregate feature similarity data elements.
    feature_similarity <- .compute_data_element_estimates(feature_similarity)
    feature_similarity <- feature_similarity[[1L]]
    
    # Obtain cluster method from stored settings, if required.
    if (is.waive(feature_cluster_method)) {
      if (is.null(feature_similarity)) {
        feature_cluster_method <- object@settings$feature_cluster_method
      } else {
        feature_cluster_method <- feature_similarity@cluster_method
      }
    } 
    
    # Obtain linkage function from stored settings, if required.
    if (is.waive(feature_linkage_method)) {
      if (is.null(feature_similarity)) {
        feature_linkage_method <- object@settings$feature_linkage_method
      } else {
        feature_linkage_method <- feature_similarity@linkage_method
      }
    } 
    
    # Obtain feature cluster cut method from stored settings, if required.
    if (is.waive(feature_cluster_cut_method)) {
      if (is.null(feature_similarity)) {
        feature_cluster_cut_method <- object@settings$feature_cluster_cut_method
      } else {
        feature_cluster_cut_method <- feature_similarity@cluster_cut_method
      }
    } 
    
    # Obtain cluster similarity threshold from stored settings, if required.
    if (is.waive(feature_similarity_threshold)) {
      if (is.null(feature_similarity)) {
        feature_similarity_threshold <- object@settings$feature_similarity_threshold
      } else {
        feature_similarity_threshold <- feature_similarity@similarity_threshold
      }
    } 
    
    # Obtain similarity metric from stored settings, if required.
    if (is.waive(feature_similarity_metric)) {
      if (is.null(feature_similarity)) {
        feature_similarity_metric <- object@settings$feature_similarity_metric
      } else {
        feature_similarity_metric <- feature_similarity@similarity_metric
      }
    } 
    
    # Replace feature cluster method == "none" with "hclust"
    if (feature_cluster_method == "none") {
      feature_cluster_method <- "hclust"
    }
    
    .check_cluster_parameters(
      cluster_method = feature_cluster_method,
      cluster_linkage = feature_linkage_method,
      cluster_cut_method = feature_cluster_cut_method,
      cluster_similarity_threshold = feature_similarity_threshold,
      cluster_similarity_metric = feature_similarity_metric,
      data_type = "feature"
    )
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Check the number of important features.
    n_important_features <- .parse_n_important_features(
      x = n_important_features,
      object = object,
      default = 20,
      data_element = "permutation_vimp"
    )
    
    # Set features to be shuffled for permutation.
    important_features <- .select_important_features(
      object = object,
      data = data,
      n_important_features = n_important_features
    )
    
    # Generate a prototype data element.
    proto_data_element <- new(
      "familiarDataElementPermutationVimp",
      detail_level = detail_level,
      estimation_type = estimation_type,
      confidence_level = confidence_level,
      bootstrap_ci_method = bootstrap_ci_method,
      similarity_metric = feature_similarity_metric
    )
    
    # Generate elements and dispatch.
    vimp_data <- extract_dispatcher(
      FUN = .extract_permutation_vimp,
      has_internal_bootstrap = TRUE,
      cl = cl,
      object = object,
      data = data,
      proto_data_element = proto_data_element,
      is_pre_processed = is_pre_processed,
      ensemble_method = ensemble_method,
      metric = metric,
      sample_limit = sample_limit,
      evaluation_times = evaluation_times,
      important_features = important_features,
      aggregate_results = aggregate_results,
      similarity_table = feature_similarity,
      cluster_method = feature_cluster_method,
      cluster_linkage = feature_linkage_method,
      cluster_cut_method = feature_cluster_cut_method,
      cluster_similarity_threshold = feature_similarity_threshold,
      cluster_similarity_metric = feature_similarity_metric,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(vimp_data)
  }
)



# extract_permutation_vimp (prediction table) ----------------------------------
setMethod(
  "extract_permutation_vimp",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("permutation variable importance")
    
    return(NULL)
  }
)



.extract_permutation_vimp <- function(
    object,
    proto_data_element,
    evaluation_times = NULL,
    aggregate_results,
    cl,
    ...
) {
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  proto_data_element <- add_model_name(proto_data_element, object = object)
  
  # Add evaluation time as a identifier to the data element.
  if (length(evaluation_times) > 0L && object@outcome_type == "survival") {
    data_elements <- add_data_element_identifier(
      x = proto_data_element,
      evaluation_time = evaluation_times
    )
    
  } else {
    data_elements <- list(proto_data_element)
  }
  
  # Iterate over data elements
  vimp_data <- lapply(
    data_elements,
    ..extract_permutation_vimp,
    object = object,
    aggregate_results = aggregate_results,
    cl = cl,
    ...
  )
  
  return(vimp_data)
}



..extract_permutation_vimp <- function(
    data_element,
    object,
    data,
    is_pre_processed,
    cl,
    ensemble_method,
    sample_limit,
    important_features,
    similarity_table,
    cluster_method,
    cluster_linkage,
    cluster_cut_method,
    cluster_similarity_threshold,
    cluster_similarity_metric,
    aggregate_results,
    progress_bar = FALSE,
    verbose = FALSE,
    message_indent,
    ...
) {
  
  # Compute prediction data.
  prediction_data <- .predict(
    object = object,
    data = data,
    time = data_element@identifiers$evaluation_time,
    ensemble_method = ensemble_method,
    is_pre_processed = is_pre_processed
  )
  
  # Check if any predictions are valid.
  if (!any_predictions_valid(prediction_data)) return(NULL)
  
  # Remove data with invalid predictions.
  prediction_data <- remove_invalid_predictions(prediction_data)
  
  # Remove data with missing outcomes.
  prediction_data <- filter_missing_outcome(prediction_data)
  
  # Check that any prediction data remain.
  if (is_empty(prediction_data)) return(NULL)
  
  if (get_n_samples(prediction_data) < 5L) return(NULL)
  
  # Explicitly load input data. We stop at the signature step because we want to
  # work with the unclustered input features, but may need to apply
  # model-specific preprocessing steps later on.
  data <- process_input_data(
    object = object,
    data = data,
    is_pre_processed = is_pre_processed,
    stop_at = "signature"
  )
  
  # Remove instances with missing outcomes.
  data <- filter_missing_outcome(data)
  
  # Perform some checks to see if there is sufficient data to perform a half-way
  # meaningful analysis.
  if (is_empty(data)) return(NULL)
  if (get_n_samples(data) < 5L) return(NULL)
  
  # Message the user concerning the time at which metrics are computed. This is
  # only relevant for survival analysis.
  if (length(data_element@identifiers$evaluation_time) > 0L && progress_bar) {
    logger_message(
      paste0(
        "Computing permutation variable importance at time ",
        data_element@identifiers$evaluation_time, "."
      ),
      indent = message_indent,
      verbose = verbose
    )
  }
  
  # Maintain only important features. The current set is based on the required
  # features.
  data <- filter_features(
    data = data,
    available_features = object@model_features
  )
  
  # Derive feature information
  feature_cluster_info <- .select_feature_clusters(
    important_features = important_features,
    available_features = get_feature_columns(data),
    similarity_table = similarity_table,
    cluster_method = cluster_method,
    cluster_linkage = cluster_linkage,
    cluster_cut_method = cluster_cut_method,
    cluster_similarity_threshold = cluster_similarity_threshold,
    cluster_similarity_metric = cluster_similarity_metric
  )
  
  # Create full list of all instances that should be evaluated.
  bootstrap_data <- lapply(
    feature_cluster_info$feature_clusters,
    .create_permutation_data_elements,
    data_element = data_element,
    similarity_threshold = feature_cluster_info$similarity_threshold,
    ...
  )
  
  # Flatten the instance list.
  bootstrap_data <- .flatten_nested_list(bootstrap_data, flatten = TRUE)
  
  # Iterate over elements.
  data_elements <- fam_mapply(
    cl = cl,
    assign = NULL,
    FUN = .compute_permutation_vimp,
    data_element = bootstrap_data$data_element,
    bootstrap = bootstrap_data$bootstrap,
    bootstrap_seed = bootstrap_data$seed,
    shuffled_features = bootstrap_data$feature,
    similarity_threshold = bootstrap_data$similarity_threshold,
    n_shuffles = bootstrap_data$n_shuffle,
    MoreArgs = c(
      list(
        "object" = object,
        "data" = data,
        "prediction_data" = prediction_data,
        "ensemble_method" = ensemble_method,
        "important_features" = important_features,
        "sample_limit" = sample_limit
      ),
      list(...)
    ),
    progress_bar = progress_bar && verbose,
    chopchop = TRUE
  )
  
  # Merge data elements
  data_elements <- merge_data_elements(data_elements)
  
  # Aggregate results, if required.
  if (aggregate_results) data_elements <- .compute_data_element_estimates(x = data_elements)
  
  return(data_elements)
}



.compute_permutation_vimp <- function(
    data_element,
    bootstrap,
    bootstrap_seed,
    sample_limit,
    shuffled_features,
    important_features,
    similarity_threshold,
    n_shuffles,
    object,
    data,
    prediction_data,
    ...) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  value <- unpermuted <- permuted <- NULL
  
  # Bootstrap the data.
  if (bootstrap) {
    data <- get_bootstrap_sample(
      data = data,
      seed = bootstrap_seed,
      n_samples = sample_limit
    )
  }
  
  # Bootstrap the prediction data.
  if (bootstrap) {
    prediction_data <- get_bootstrap_sample(
      data = prediction_data,
      seed = bootstrap_seed,
      n_samples = sample_limit
    )
  }
  
  # Check if any predictions are valid.
  if (!any_predictions_valid(prediction_data)) return(NULL)
  
  # Determine metric scores for the real (unshuffled) data
  scores <- list()
  scores[[1L]] <- ..compute_permutation_vimp(
    data_element = data_element,
    object = object,
    data = NULL,
    prediction_data = prediction_data,
    shuffled_features = NULL,
    ...
  )
  
  # Determine metric scores for the shuffled data.
  for (ii in seq_len(n_shuffles)) {
    scores[[ii + 1L]] <- ..compute_permutation_vimp(
      data_element = data_element,
      object = object,
      data = data,
      prediction_data = NULL,
      shuffled_features = shuffled_features,
      ...
    )
  }
  
  # Combine to single data.table
  scores <- data.table::rbindlist(scores, use.names = TRUE)
  
  # Determine median score.
  scores <- scores[
    ,
    list("value" = stats::median(value, na.rm = TRUE)),
    by = c("metric", "is_shuffled")
  ]
  
  # Cast wide and compute difference between unshuffled and shuffled values.
  scores <- data.table::dcast(
    data = scores,
    metric ~ is_shuffled,
    value.var = "value"
  )
  
  # Rename FALSE and TRUE columns to something that will not cause issues.
  data.table::setnames(
    x = scores,
    old = c("FALSE", "TRUE"),
    new = c("unpermuted", "permuted")
  )
  
  # Compute the difference between permuted and unpermuted values.
  scores[, "value" := unpermuted - permuted]
  
  # Determine the number of metrics.
  n_metrics <- nrow(scores)
  
  # Retain only important features for further mention.
  important_features <- intersect(important_features, shuffled_features)
  
  # Add features.
  scores <- scores[rep(seq_len(n_metrics), each = length(important_features))]
  scores[, ":="(
    "feature" = rep(important_features, times = n_metrics),
    "permuted" = NULL,
    "unpermuted" = NULL
  )]
  
  # Store to data element.
  data_element@data <- scores
  
  # Add similarity threshold as an identifier. This ensures that data are
  # unchanged are multiplied.
  data_elements <- add_data_element_identifier(
    x = data_element,
    similarity_threshold = similarity_threshold
  )
  
  return(data_elements)
}



..compute_permutation_vimp <- function(
    data_element,
    object,
    metric,
    data = NULL,
    prediction_data = NULL,
    shuffled_features = NULL,
    ensemble_method,
    is_pre_processed,
    ...
) {
  
  # Generate prediction data if it does not exist.
  if (is.null(prediction_data)) {
    
    if (!is.null(shuffled_features)) {
      
      # Make local copy of data to avoid updating by reference.
      data@data <- data.table::copy(data@data)
      
      # Shuffle row entries.
      shuffle_id <-  sample.int(n = nrow(data@data))
      
      # Iterate over features and shuffle them. Check that the feature appears
      # in the data.
      for (ii in shuffled_features) {
        if (ii %in% colnames(data@data)) {
          data.table::set(
            x = data@data,
            j = ii,
            value = data@data[[ii]][shuffle_id]
          )
        }
      }
    }
    
    # Create prediction based on data.
    prediction_data <- .predict(
      object = object,
      data = data,
      time = data_element@identifiers$evaluation_time,
      ensemble_method = ensemble_method
    )
  }
  
  # Compute score.
  score <- sapply(
    metric,
    compute_metric_score,
    object = object,
    data = prediction_data,
    time = data_element@identifiers$evaluation_time
  )
  
  return(data.table::data.table(
    "metric" = metric,
    "value" = score,
    "is_shuffled" = !is.null(shuffled_features)
  ))
}



.select_feature_clusters <- function(
    important_features,
    available_features,
    similarity_table,
    cluster_method,
    cluster_linkage,
    cluster_cut_method,
    cluster_similarity_threshold,
    cluster_similarity_metric
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  name <- cluster_id <- similarity_threshold <- NULL
  has_important_feature <- NULL
  
  if (is_empty(similarity_table)) {
    # Set the placeholder similarity threshold
    cluster_similarity_threshold <- ifelse(
      cluster_method %in% c("agnes", "diana", "hclust") && cluster_cut_method == "fixed_cut",
      1.0,
      Inf
    )
    
    # Create method object.
    cluster_method_object <- create_cluster_method_object(
      cluster_method = "none",
      data_type = "feature",
      cluster_representation_method = "none"
    )
    
    # Update (no) similarity table.
    cluster_method_object@similarity_table <- methods::new(
      "noSimilarityTable",
      data = available_features,
      data_type = cluster_method_object@data_type
    )
    
    # Generate the clustering table.
    cluster_table <- create_clusters(
      object = cluster_method_object,
      as_cluster_object = FALSE
    )
    
    # Add in the similarity threshold.
    cluster_table[, "similarity_threshold" := cluster_similarity_threshold]
    cluster_table[, "cluster_size" := .N, by = "cluster_id"]

  } else if (
    cluster_method %in% c("agnes", "diana", "hclust") &&
    cluster_cut_method == "fixed_cut"
  ) {
    # Solution for methods where multiple cuts are possible.
    
    # Create method object.
    cluster_method_object <- create_cluster_method_object(
      cluster_method = cluster_method,
      data_type = "feature",
      cluster_linkage = cluster_linkage,
      cluster_cut_method = cluster_cut_method,
      cluster_similarity_threshold = cluster_similarity_threshold,
      cluster_similarity_metric = cluster_similarity_metric,
      cluster_representation_method = "none"
    )
    
    # Update the similarity table and attach.
    cluster_method_object@similarity_table <- methods::new(
      "similarityTable",
      data = similarity_table@data[, mget(c("feature_name_1", "feature_name_2", "value"))],
      similarity_metric = cluster_similarity_metric,
      data_type = cluster_method_object@data_type
    )
    
    # Insert 1.0 if necessary.
    if (!1.0 %in% cluster_similarity_threshold) {
      cluster_similarity_threshold <- c(1.0, cluster_similarity_threshold)
    } 
    
    # Sort descending.
    cluster_similarity_threshold <- sort(
      cluster_similarity_threshold,
      decreasing = TRUE
    )
    
    # Obtain cluster tables for each threshold.
    cluster_table <- lapply(
      cluster_similarity_threshold,
      function(cluster_similarity_threshold, cluster_method_object) {
        
        # Update the similarity threshold in the object.
        cluster_method_object@similarity_threshold <- cluster_similarity_threshold
        
        # Generate the clustering table.
        cluster_table <- create_clusters(
          object = cluster_method_object,
          as_cluster_object = FALSE
        )
        
        # Add in the similarity threshold and cluster size.
        cluster_table[, "similarity_threshold" := cluster_similarity_threshold]
        cluster_table[, "cluster_size" := .N, by = "cluster_id"]
        
        return(cluster_table)
      },
      cluster_method_object = cluster_method_object
    )
    
    # Concatenate to single table
    cluster_table <- data.table::rbindlist(cluster_table, use.names = TRUE)
    
    # Remove features that are not available.
    cluster_table <- cluster_table[name %in% available_features]
    
  } else {
    # Solution for methods that do not allow for multiple cuts, such as PAM.
    
    # Create method object.
    cluster_method_object <- create_cluster_method_object(
      cluster_method = cluster_method,
      data_type = "feature",
      cluster_linkage = cluster_linkage,
      cluster_cut_method = cluster_cut_method,
      cluster_similarity_threshold = cluster_similarity_threshold,
      cluster_similarity_metric = cluster_similarity_metric,
      cluster_representation_method = "none"
    )
    
    # Update the similarity table and attach.
    cluster_method_object@similarity_table <- methods::new(
      "similarityTable",
      data = similarity_table@data[, mget(c("feature_name_1", "feature_name_2", "value"))],
      similarity_metric = cluster_similarity_metric,
      data_type = cluster_method_object@data_type
    )
    
    # Generate the clustering table.
    cluster_table <- create_clusters(
      object = cluster_method_object,
      as_cluster_object = FALSE
    )

    # Add the similarity threshold.
    cluster_table[, "similarity_threshold" := -Inf]
    cluster_table[, "cluster_size" := .N, by = "cluster_id"]
    
    # Remove features that are not available.
    cluster_table <- cluster_table[name %in% available_features]
  }
  
  # Identify the unique thresholds that were used.
  similarity_thresholds_used <- sort(
    unique(cluster_table$similarity_threshold),
    decreasing = TRUE
  )
  
  # Identify the unique clusters of features that should be computed.
  cluster_table <- cluster_table[
    ,
    list(
      "similarity_threshold" = max(similarity_threshold),
      "n_thresholds_same_cluster" = .N,
      "cluster_id" = min(cluster_id)
    ),
    by = c("name", "cluster_size")
  ]

  # Determine which clusters have at least one unique feature.
  cluster_table[, "has_important_feature" := name %in% important_features]
  cluster_table[, "has_important_feature" := any(has_important_feature), by = c(
    "similarity_threshold",
    "n_thresholds_same_cluster",
    "cluster_id",
    "cluster_size"
  )]
  
  cluster_table <- cluster_table[has_important_feature == TRUE][, "has_important_feature" := NULL]
  
  # Split the cluster table into clusters.
  cluster_list <- split(
    cluster_table,
    by = c(
      "similarity_threshold",
      "n_thresholds_same_cluster",
      "cluster_id",
      "cluster_size"
    )
  )
  
  return(list(
    "similarity_threshold" = similarity_thresholds_used,
    "feature_clusters" = cluster_list
  ))
}



.create_permutation_data_elements <- function(
    feature_cluster,
    data_element,
    similarity_threshold,
    ...
) {

  # Add bootstraps (if required).
  bootstrap_data <- add_data_element_bootstrap(x = data_element, ...)
  
  # Add features.
  bootstrap_data$features <- lapply(
    seq_along(bootstrap_data$data_element),
    function(ii, feature) (feature),
    feature = feature_cluster$name
  )
  
  # Identify the similarity threshold for which the current feature sets form
  # a cluster.
  n <- feature_cluster$n_thresholds_same_cluster[1L]
  used_similarity_threshold <- head(
    similarity_threshold[similarity_threshold <= feature_cluster$similarity_threshold[1L]],
    n = n
  )
  
  # Add similarity threshold.
  bootstrap_data$similarity_threshold <- lapply(
    seq_along(bootstrap_data$data_element),
    function(ii, similarity_threshold) (similarity_threshold),
    similarity_threshold = used_similarity_threshold
  )
  
  # Add number of shuffles. We don't introduce additional shuffles except for
  # point estimates.
  n_shuffles <- ifelse(data_element@estimation_type == "point", 20L, 1L)
  
  # Add shuffles.
  bootstrap_data$n_shuffles <- rep(n_shuffles, times = length(bootstrap_data$data_element))
  
  return(bootstrap_data)
}



# export_permutation_vimp (generic) --------------------------------------------

#'@title Extract and export permutation variable importance.
#'
#'@description Extract and export model-based variable importance from a
#'  familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams extract_permutation_vimp
#'@inheritDotParams as_familiar_collection
#'
#'@details Data, such as permutation variable importance and calibration
#'  information, is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previously mentioned files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Permutation Variable importance assesses the improvement in model
#'  performance due to a feature. For this purpose, the performance of the model
#'  is measured as normal, and is measured again with a dataset where the values
#'  of the feature in question have been randomly permuted. The difference
#'  between both performance measurements is the permutation variable
#'  importance.
#'
#'  In familiar, this basic concept is extended in several ways:
#'
#'  * Point estimates of variable importance are based on multiple (21) random
#'  permutations. The difference between model performance on the normal dataset
#'  and the median performance measurement of the randomly permuted datasets is
#'  used as permutation variable importance.
#'
#'  * Confidence intervals for the ensemble model are determined using bootstrap
#'  methods.
#'
#'  * Permutation variable importance is assessed for any metric specified using
#'  the `metric` argument.
#'
#'  * Permutation variable importance can take into account similarity between
#'  features and permute similar features simultaneously.
#'
#'@return A data.table (if `dir_path` is not provided), or nothing, as all data
#'  is exported to `csv` files.
#'@exportMethod export_permutation_vimp
#'@md
#'@rdname export_permutation_vimp-methods
setGeneric(
  "export_permutation_vimp",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_permutation_vimp")
  }
)



# export_permutation_vimp (collection) -----------------------------------------

#'@rdname export_permutation_vimp-methods
setMethod(
  "export_permutation_vimp",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(.export(
      x = object,
      data_slot = "permutation_vimp",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "variable_importance",
      subtype = "permutation",
      export_collection = export_collection
    ))
  }
)



# export_permutation_vimp (general) --------------------------------------------

#'@rdname export_permutation_vimp-methods
setMethod(
  "export_permutation_vimp",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "permutation_vimp",
          "aggregate_results" = aggregate_results
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_permutation_vimp,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

.all_gtable_guide_names <- function(type = "all") {
  if (type == "all") {
    return(c(
      "guide-box-right", "guide-box-left",
      "guide-box-top", "guide-box-bottom", "guide-box-inside"
    ))
    
  } else if (type == "right") {
    return("guide-box-right")
    
  } else if (type == "left") {
    return("guide-box-left")
    
  } else if (type == "top") {
    return("guide-box-top")
    
  } else if (type == "bottom") {
    return("guide-box-bottom")
    
  } else if (type == "inside") {
    return("guide-box-inside")
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown type: ", type))
  }
}



.all_gtable_title_names <- function(type = "all") {
  if (type == "all") {
    return(c("title", "subtitle", "caption"))
    
  } else if (type == "title") {
    return(c("title", "subtitle"))
    
  } else if (type == "caption") {
    return("caption")
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown type: ", type))
  }
}



.all_gtable_panel_names <- function() {
  return("panel")
}



.all_gtable_strip_x_names <- function() {
  return(c("strip-t", "strip-b"))
}



.all_gtable_strip_y_names <- function() {
  return(c("strip-l", "strip-r"))
}



.all_gtable_label_names <- function(type = "all") {
  if (type == "all") {
    return(c(.all_gtable_label_x_names(), .all_gtable_label_y_names()))
    
  } else if (type == "right") {
    return(.all_gtable_label_y_names("right"))
    
  } else if (type == "left") {
    return(.all_gtable_label_y_names("left"))
    
  } else if (type == "top") {
    return(.all_gtable_label_x_names("top"))
    
  } else if (type == "bottom") {
    return(.all_gtable_label_x_names("bottom"))
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown type: ", type))
  }
}



.all_gtable_label_x_names <- function(type = "all") {
  if (type == "all") {
    return(c("xlab-b", "xlab-t"))
    
  } else if (type == "top") {
    return("xlab-t")
    
  } else if (type == "bottom") {
    return("xlab-b")
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown type: ", type))
  }
}



.all_gtable_label_y_names <- function(type = "all") {
  if (type == "all") {
    return(c("ylab-l", "ylab-r"))
    
  } else if (type == "left") {
    return("ylab-l")
    
  } else if (type == "right") {
    return("ylab-r")
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown type: ", type))
  }
}



.all_gtable_axis_names <- function(type = "all") {
  if (type == "all") {
    return(c(.all_gtable_axis_x_names(), .all_gtable_axis_y_names()))
    
  } else if (type == "right") {
    return(.all_gtable_axis_y_names("right"))
    
  } else if (type == "left") {
    return(.all_gtable_axis_y_names("left"))
    
  } else if (type == "top") {
    return(.all_gtable_axis_x_names("top"))
    
  } else if (type == "bottom") {
    return(.all_gtable_axis_x_names("bottom"))
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown type: ", type))
  }
}



.all_gtable_axis_x_names <- function(type = "all") {
  if (type == "all") {
    return(c("axis-b", "axis-t"))
    
  } else if (type == "top") {
    return("axis-t")
    
  } else if (type == "bottom") {
    return("axis-b")
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown type: ", type))
  }
}



.all_gtable_axis_y_names <- function(type = "all") {
  if (type == "all") {
    return(c("axis-l", "axis-r"))
    
  } else if (type == "left") {
    return("axis-l")
    
  } else if (type == "right") {
    return("axis-r")
    
  } else {
    ..error_reached_unreachable_code(paste0("unknown type: ", type))
  }
}



.gtable_element_in_layout <- function(g, element, partial_match = FALSE) {
  if (partial_match) {
    return(any(grepl(pattern = element, x = g$layout$name)))
  } else {
    return(any(element %in% g$layout$name))
  }
}



.gtable_get_position <- function(
    g,
    element, 
    where = NULL,
    partial_match = FALSE, 
    allow_multiple = FALSE
) {
  # Find position.
  if (partial_match) {
    # Based on partial matching.
    position <- g$layout[
      grepl(pattern = element, x = g$layout$name),
      c("t", "l", "b", "r")
    ]
    
  } else {
    # Based on exact matching.
    position <- g$layout[
      g$layout$name == element,
      c("t", "l", "b", "r")
    ]
  }

  if (nrow(position) == 0L) {
    ..error_reached_unreachable_code(
      ".gtable_get_position: element not found in layout table."
    )
  }

  if (is.null(where)) {
    if (nrow(position) != 1L) {
      if (allow_multiple) {
        browser()
      } else {
        ..error_reached_unreachable_code(
          paste0("Multiple matches, please set where attribute.")
        )
      }
    }
  } else if (where == "top") {
    # Select the uppermost element.
    position <- position[position$t == min(position$t), ][1L, ]
    
  } else if (where == "bottom") {
    # Select the bottommost element.
    position <- position[position$b == max(position$b), ][1L, ]
    
  } else if (where == "left") {
    # Select the leftmost element
    position <- position[position$l == min(position$l), ][1L, ]
    
  } else if (where == "right") {
    # Select the rightmost element
    position <- position[position$r == max(position$r), ][1L, ]
    
  } else {
    ..error_value_not_allowed(
      x = where,
      var_name = "where",
      values = c("top", "bottom", "left", "right")
    )
  }

  # Return as array.
  position <- simplify2array(position)

  return(position)
}



.gtable_extract_grob <- function(
    g,
    element
) {
  grob_id <- which(g$layout$name == element)
  
  if (length(grob_id) == 0L) {
    ..error_reached_unreachable_code("element could not be found in gtable")
    
  } else if (length(grob_id) > 1L) {
    ..error_reached_unreachable_code("more than one element was found in gtable")
  }
  
  return(g$grobs[[grob_id]])
}



.gtable_insert_spacer <- function(
    g,
    position,
    width = NULL,
    height = NULL,
    make_space = FALSE,
    name = NULL
) {
  # Generate name.
  if (is.null(name)) name <- "spacer"
  
  # Create empty grob.
  spacer <- grid::grob(cl = "spacerGrob", name = name)
  
  # Set width and height.
  spacer$widths <- grid::unit(0.0, "cm")
  if (!is.null(width)) spacer$widths <- width
  spacer$heights <- grid::unit(0.0, "cm")
  if (!is.null(height)) spacer$heights <- height
  
  # If make_space is TRUE, insert a column or row at the given position, and then
  # place the spacer in the new column or row. Otherwise the spacer is inserted
  # in place.
  if (make_space) {
    if (!is.null(width)) {
      g <- gtable::gtable_add_cols(
        x = g,
        widths = width,
        pos = position[["l"]]
      )
    }
    if (!is.null(height)) {
      g <- gtable::gtable_add_rows(
        x = g,
        heights = height,
        pos = position[["t"]]
      )
    }
  }
  
  g <- gtable::gtable_add_grob(
    g,
    grobs = spacer,
    t = position[["t"]],
    b = position[["b"]],
    l = position[["l"]],
    r = position[["r"]],
    name = name
  )
  
  return(g)
}



.gtable_remove <- function(
    g,
    removed_element = NULL,
    trim = FALSE
) {
  matched_elements <- !(g$layout$name %in% removed_element)
  
  # gtable::gtable_filter uses partial matching, which leads to issues with,
  # e.g., "title" and "subtitle" elements.
  g$layout <- g$layout[matched_elements, , drop = FALSE]
  g$grobs <- g$grobs[matched_elements]
  
  if (trim) g <- gtable::gtable_trim(g)
  
  return(g)
}



.gtable_insert <- function(
    g,
    g_new,
    where,
    grob_name = NULL,
    spacer = NULL
) {
  # Intended for inserting elements that stretch multiple along_elements. It can
  # also be used for inserting elements directly (without along_elements) and/or
  # replacing existing elements (attempt_replace=TRUE)
  if (length(g_new) == 0L) {
    return(g)
  }
  
  if (where[1L] == "at") {
    position <- c(
      "t" = as.integer(where[2L]),
      "l" = as.integer(where[3L]),
      "b" = as.integer(where[4L]),
      "r" = as.integer(where[5L])
    )
    
    g <- ..gtable_insert_at(
      g = g,
      g_new = g_new,
      grob_name = grob_name,
      position = position
    )
    
  } else if (where[1L] == "replace") {
    g <- ..gtable_insert_replace(
      g = g,
      g_new = g_new,
      ref_element = where[2L]
    )
    
  } else if (where[1L] == "intersect") {
    g <- ..gtable_insert_intersect(
      g = g,
      g_new = g_new,
      where_1 = where[2L],
      ref_element_1 = where[3L],
      where_2 = where[4L],
      ref_element_2 = where[5L],
      grob_name = grob_name
    )
    
  } else if (where[1L] == "left") {
    g <- ..gtable_insert_left(
      g = g,
      g_new = g_new,
      ref_element = where[2L],
      grob_name = grob_name,
      spacer = spacer
    )
    
  } else if (where[1L] == "right") {
    g <- ..gtable_insert_right(
      g = g,
      g_new = g_new,
      ref_element = where[2L],
      grob_name = grob_name,
      spacer = spacer
    )
    
  } else if (where[1L] == "above") {
    g <- ..gtable_insert_above(
      g = g,
      g_new = g_new,
      ref_element = where[2L],
      grob_name = grob_name,
      spacer = spacer
    )
    
  } else if (where[1L] == "below") {
    g <- ..gtable_insert_below(
      g = g,
      g_new = g_new,
      ref_element = where[2L],
      grob_name = grob_name,
      spacer = spacer
    )
    
  } else {
    ..error_reached_unreachable_code(
      "The first element of where should be one of replace, intersect, left, right, above or below."
    )
  }

  # Update widths and heights.
  g <- .gtable_update_layout(g = g)
  
  return(g)
}



..gtable_insert_at <- function(g, g_new, grob_name = NULL, position) {
  # Set clip.
  clip <- "on"
  if (is(g_new, "TableGrob") || is(g_new, "gtable")) clip <- g_new$layout$clip[1L]
  
  # Set name.
  name <- grob_name
  if (is.null(name) && (is(g_new, "TableGrob") || is(g_new, "gtable"))) name <- g_new$layout$name[1L]
  
  # Insert new element.
  g <- gtable::gtable_add_grob(
    g,
    grobs = g_new,
    t = position[["t"]],
    l = position[["l"]],
    b = position[["b"]],
    r = position[["r"]],
    name = name,
    clip = clip
  )
  
  return(g)
}



..gtable_insert_replace <- function(g, g_new, ref_element) {
  # Find position where new element is to be inserted.
  position <- .gtable_get_position(
    g = g, 
    element = ref_element
  )
  
  # Remove original element.
  g <- .gtable_remove(g = g, removed_element = ref_element)
  
  # Insert at position.
  g <- ..gtable_insert_at(
    g = g,
    g_new = g_new,
    grob_name = ref_element,
    position = position
  )
  
  return(g)
}



..gtable_insert_intersect <- function(
    g,
    g_new,
    where_1,
    ref_element_1,
    where_2,
    ref_element_2,
    grob_name = NULL
) {
  
  # Check that where is correctly specified.
  if (
    all(c(where_1, where_2) %in% c("above", "below")) ||
    all(c(where_1, where_2) %in% c("left", "right"))
  ) {
    ..error_reached_unreachable_code("where needs to be orthogonal positions")
  }
  
  elem_position <- integer(4L)
  names(elem_position) <- c("t", "l", "b", "r")
  
  # Get reference positions for each reference element.
  ref_position_1 <- .gtable_get_position(
    g = g, 
    element = ref_element_1
  )
  
  ref_position_2 <- .gtable_get_position(
    g = g, 
    element = ref_element_2
  )
  
  if (where_1 %in% c("above", "below")) {
    # Inherit l and r from the first reference element and t and b from the
    # second.
    elem_position[["l"]] <- ref_position_1[["l"]]
    elem_position[["r"]] <- ref_position_1[["r"]]
    elem_position[["t"]] <- ref_position_2[["t"]]
    elem_position[["b"]] <- ref_position_2[["b"]]
    
  } else {
    # Inherit t and b from the first reference element and l and r from the
    # second.
    elem_position[["t"]] <- ref_position_1[["t"]]
    elem_position[["b"]] <- ref_position_1[["b"]]
    elem_position[["l"]] <- ref_position_2[["l"]]
    elem_position[["r"]] <- ref_position_2[["r"]]
  }
  
  # Add element to g.
  g <- ..gtable_insert_at(
    g = g,
    g_new = g_new,
    grob_name = grob_name,
    position = elem_position
  )
  
  return(g)
}


..gtable_insert_below <- function(
    g,
    g_new, 
    ref_element,
    grob_name = NULL,
    spacer = NULL
) {
  # Create an offset.
  spacer_ref_offset <- elem_ref_offset <- integer(4L)
  names(spacer_ref_offset) <- c("t", "l", "b", "r")
  names(elem_ref_offset) <- c("t", "l", "b", "r")
  
  # Find position of reference element.
  ref_position <- .gtable_get_position(
    g = g, 
    element = ref_element
  )
  
  # Force the "top" of the reference position to the bottom, because we don't
  # need to copy the number of rows of the reference object.
  ref_position[["t"]] <- ref_position[["b"]]
  
  # Add space between the reference element and the element to be inserted.
  if (!is.null(spacer)) {
    g <- gtable::gtable_add_rows(
      g,
      heights = spacer,
      pos = ref_position[["b"]]
    )
    
    # Update offsets for spacer and new grob.
    spacer_ref_offset[["t"]] <- spacer_ref_offset[["b"]] <- 1L
    elem_ref_offset[["t"]] <- elem_ref_offset[["b"]] <- 1L
    
    # Add spacer.
    g <- .gtable_insert_spacer(
      g = g,
      position = ref_position + spacer_ref_offset,
      height = spacer
    )
  }
  
  # Add row below b or below spacer.
  g <- gtable::gtable_add_rows(
    g,
    heights = g_new$heights,
    pos = ref_position[["b"]] + elem_ref_offset[["b"]]
  )
  
  elem_ref_offset[["t"]] <- elem_ref_offset[["t"]] + 1L
  elem_ref_offset[["b"]] <- elem_ref_offset[["b"]] + length(g_new$heights)
  
  # Set new position
  new_position <- ref_position + elem_ref_offset

  # Add element to g.
  g <- ..gtable_insert_at(
    g = g,
    g_new = g_new,
    grob_name = grob_name,
    position = new_position
  )
  
  return(g)
}



..gtable_insert_above <- function(
    g,
    g_new, 
    ref_element,
    grob_name = NULL,
    spacer = NULL
) {
  # Create an offset.
  spacer_ref_offset <- elem_ref_offset <- integer(4L)
  names(spacer_ref_offset) <- c("t", "l", "b", "r")
  names(elem_ref_offset) <- c("t", "l", "b", "r")
  
  # Find position of reference element.
  ref_position <- .gtable_get_position(
    g = g, 
    element = ref_element
  )
  
  # Force the "top" of the reference position to the bottom, because we don't
  # need to copy the number of rows of the reference object.
  ref_position[["t"]] <- ref_position[["b"]]
  
  # Add space between the element that should be inserted and the reference
  # element.
  if (!is.null(spacer)) {
    g <- gtable::gtable_add_rows(
      g,
      heights = spacer,
      pos = ref_position[["t"]] - 1L
    )
  
    # Add spacer.
    g <- .gtable_insert_spacer(
      g = g,
      position = ref_position,
      height = spacer
    )
  }
  
  g <- gtable::gtable_add_rows(
    g,
    heights = g_new$heights,
    pos = ref_position[["t"]] - 1L
  )
  
  # Set new position
  new_position <- ref_position
  new_position[["b"]] <- new_position[["b"]] + length(g_new$heights) - 1L
  
  # Add element to g.
  g <- ..gtable_insert_at(
    g = g,
    g_new = g_new,
    grob_name = grob_name,
    position = new_position
  )
  
  return(g)
}



..gtable_insert_left <- function(
    g,
    g_new, 
    ref_element,
    grob_name = NULL,
    spacer = NULL
) {
  # Create an offset.
  spacer_ref_offset <- elem_ref_offset <- integer(4L)
  names(spacer_ref_offset) <- c("t", "l", "b", "r")
  names(elem_ref_offset) <- c("t", "l", "b", "r")
  
  # Find position of reference element.
  ref_position <- .gtable_get_position(
    g = g, 
    element = ref_element
  )
  
  # Force the "left" of the reference position to the right, because we don't
  # need to copy the number of columns of the reference object.
  ref_position[["l"]] <- ref_position[["r"]]
  
  # Add space between the element that should be inserted and the reference
  # element.
  if (!is.null(spacer)) {
    g <- gtable::gtable_add_cols(
      g,
      widths = spacer,
      pos = ref_position[["l"]] - 1L
    )
    
    # Add spacer.
    g <- .gtable_insert_spacer(
      g = g,
      position = ref_position,
      width = spacer
    )
  }
  
  # Make room for the grob that we want to insert.
  g <- gtable::gtable_add_cols(
    g,
    widths = g_new$widths,
    pos = ref_position[["l"]] - 1L
  )
  
  # Set new position
  new_position <- ref_position
  new_position[["r"]] <- new_position[["r"]] + length(g_new$widths) - 1L
  
  # Add element to g.
  g <- ..gtable_insert_at(
    g = g,
    g_new = g_new,
    grob_name = grob_name,
    position = new_position
  )
  
  return(g)
}



..gtable_insert_right <- function(
    g,
    g_new, 
    ref_element,
    grob_name = NULL,
    spacer = NULL
) {
  # Create an offset.
  spacer_ref_offset <- elem_ref_offset <- integer(4L)
  names(spacer_ref_offset) <- c("t", "l", "b", "r")
  names(elem_ref_offset) <- c("t", "l", "b", "r")
  
  # Find position of reference element.
  ref_position <- .gtable_get_position(
    g = g, 
    element = ref_element
  )
  
  # Force the "left" of the reference position to the right, because we don't
  # need to copy the number of rows of the reference object.
  ref_position[["l"]] <- ref_position[["r"]]
  
  # Add space between the reference element and the element to be inserted.
  if (!is.null(spacer)) {
    g <- gtable::gtable_add_cols(
      g,
      widths = spacer,
      pos = ref_position[["r"]]
    )
    
    # Update offsets for spacer and new grob.
    spacer_ref_offset[["l"]] <- spacer_ref_offset[["r"]] <- 1L
    elem_ref_offset[["l"]] <- elem_ref_offset[["r"]] <- 1L
    
    # Add spacer.
    g <- .gtable_insert_spacer(
      g = g,
      position = ref_position + spacer_ref_offset,
      width = spacer
    )
  }
  
  # Add row below b or below spacer.
  g <- gtable::gtable_add_cols(
    g,
    widths = g_new$widths,
    pos = ref_position[["r"]] + elem_ref_offset[["r"]]
  )
  
  elem_ref_offset[["l"]] <- elem_ref_offset[["l"]] + 1L
  elem_ref_offset[["r"]] <- elem_ref_offset[["r"]] + length(g_new$widths)
  
  # Set new position
  new_position <- ref_position + elem_ref_offset
  
  # Add element to g.
  g <- ..gtable_insert_at(
    g = g,
    g_new = g_new,
    grob_name = grob_name,
    position = new_position
  )
  
  return(g)
}





.gtable_rename_element <- function(
    g,
    old,
    new,
    partial_match = FALSE,
    allow_missing = FALSE
) {
  if (!.gtable_element_in_layout(
    g = g,
    element = old,
    partial_match = partial_match
  )) {
    
    if (allow_missing) {
      return(g)
    }

    ..error(".gtable_rename_element: element not found in layout table.")
  }

  if (partial_match) {
    updated_element <- grepl(pattern = old, x = g$layout$name)
  } else {
    updated_element <- g$layout$name == old
  }

  if (sum(updated_element) > 1L) {
    ..warning(".gtable_rename_element: multiple elements will be updated.")
  }

  g$layout$name[updated_element] <- new

  return(g)
}



.gtable_update_panel_aspects <- function(g) {
    # Make panels inherit heights and widths, if they don't have any. This is done
  # to ensure that panels retain heights and widths, even if supporting elements
  # such as the axis text and label elements are stripped on figure composition.
  element_names <- g$layout$name
  panel_elements <- element_names[sapply(
    element_names, 
    startswith_any, 
    prefix = .all_gtable_panel_names()
  )]
  
  for (panel_element in panel_elements) {
    for (aspect in c("height", "width")) {
      grob_id <- which(element_names == panel_element)
      panel_size <- .gtable_get_aspect_size(
        grob_id = grob_id,
        g = g,
        aspect = aspect
      )
      
      # Only inherit aspect size if the panel element does not have its own
      # size set.
      if (is.null(panel_size)) {
        if (aspect == "height") {
          position <-g$layout[grob_id, "t", drop = TRUE]
          panel_size <- g$heights[position]
          g$grobs[[grob_id]]$height <- panel_size
          
        } else {
          position <- g$layout[grob_id, "l", drop = TRUE]
          panel_size <- g$widths[position]
          g$grobs[[grob_id]]$width <- panel_size
        }
      }
    }
  }
  
  return(g)
}



.gtable_update_layout <- function(g) {
  # gtable uses its heights and widths attributes to structure and plot an 
  # image. Because we sometimes work with composite plots in familiar, we
  # need to update the gtable object. In particular, we need to set heights and
  # widths attributes correctly.
  #
  # To do so, we read the heights and widths of individual objects in a column
  # or row, depending on the aspect. These sizes are then converted to absolute
  # measures (points), with the exception of objects that have non-zero npc or
  # null as size. In that case, null elements take precedence over absolute
  # values. For example, the figure panel has 1null height and 1null width,
  # making it resizable.
  
  ..get_updated_aspect <- function(g, aspect = "width") {
    if (aspect == "width") {
      new_sizes <- g$widths
      aspect_length <- ncol(g)
      
    } else if (aspect == "height") {
      new_sizes <- g$heights
      aspect_length <- nrow(g)
      
    } else {
      ..error_reached_unreachable_code("aspect was not height or width")
    }
    
    for (ii in seq_len(aspect_length)) {
      # Select candidates.
      if (aspect == "width") {
        candidates <- which(g$layout$l == ii & g$layout$r == ii)
      } else {
        candidates <- which(g$layout$t == ii & g$layout$b == ii)
      }
      
      # Skip if there are no candidates.
      if(length(candidates) == 0L) {
        new_sizes[ii] <- grid::unit(0.0, "points")
        next
      } 
      # Identify the aspect size of the grobs.
      grob_sizes <- lapply(candidates, .gtable_get_aspect_size, g = g, aspect = aspect)
      grob_sizes <- grob_sizes[sapply(grob_sizes, grid::is.unit)]
      
      # Skip if there are no valid aspect sizes.
      if (length(grob_sizes) == 0L) {
        new_sizes[ii] <- grid::unit(0.0, "points")
        next
      }
      
      # Set grob sizes.
      grob_sizes <- do.call(grid::unit.c, grob_sizes)
      grob_sizes <- ..gtable_filter_aspect_sizes(grob_sizes)
      
      # Check if there any non-zero sizes.
      if (all(as.numeric(grob_sizes) == 0.0)) {
        new_sizes[ii] <- grid::unit(0.0, "points")
        next
      }
      
      if (length(grob_sizes) > 1L) grob_sizes <- max(grob_sizes)
      
      # Update size based on elements that are present, as long as the
      # previously provided size did not contain "null" size types.
      new_sizes[ii] <- grob_sizes
    }
    
    return(new_sizes)
  }
  
  new_heights <- ..get_updated_aspect(g = g, aspect = "height")
  new_widths <- ..get_updated_aspect(g = g, aspect = "width")
  
  # Update heights and widths in the table.
  g$heights <- new_heights
  g$widths <- new_widths
  
  return(g)
}



.gtable_get_aspect_size <- function(grob_id, g, aspect = "width") {
  
  grob_size <- .gtable_get_grob_aspect_size(
    grob = g$grobs[[grob_id]],
    aspect = aspect
  )
  
  return(grob_size)
}



.gtable_get_grob_aspect_size <- function(grob, aspect = "width") {
  
  if (aspect == "width") {
    aspect_names <- c("widths", "width")
    
  } else if (aspect == "height") {
    aspect_names <- c("heights", "height")
    
  } else {
    ..error_reached_unreachable_code("aspect was not height or width")
  }
  
  grob_size <- NULL
  
  # Attempt to get the size of the grob in absolute units, but maintain size
  # if this is null or npc.
  for (aspect_name in aspect_names) {
    current_size <- grob[[aspect_name]]
    if (!is.null(current_size)) {
      if (all(grid::unitType(current_size) %in% c("null", "npc"))) {
        grob_size <- current_size
        break
        
      } else {
        grob_size <- sum(grid::convertUnit(current_size, "points"))
        break
      }
    }
  }
  
  return(grob_size)
}



..gtable_filter_aspect_sizes <- function(x) {
  if (length(x) == 0L) return(NULL)
  
  # Filter items that have no dimension.
  grob_zero <- as.numeric(x) == 0.0
  if (length(x) == 1L && all(grob_zero)) return(grid::unit(0.0, "points"))
  
  x <- x[!grob_zero]
  if (length(x) == 0L) return(grid::unit(0.0, "points"))
  
  # null overrides everything else.
  grob_null <- grid::unitType(x) == "null"
  if (any(grob_null)) {
    x <- x[grob_null]
    return(grid::unit(max(as.numeric(x)), "null"))
  }
  
  # npc is only relevant if there are no other grobs with more specific unit
  # types.
  grob_npc <- grid::unitType(x) == "npc"
  if (any(grob_npc) && !all(grob_npc)) x <- x[!grob_npc]
  
  return(x)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarDataElementFeatureExpression object ----------------------------------
setClass(
  "familiarDataElementFeatureExpression",
  contains = "familiarDataElement",
  slots = list(
    "feature_info" = "ANY",
    "evaluation_time" = "ANY"
  ),
  prototype = methods::prototype(
    detail_level = "ensemble",
    estimation_type = "point",
    feature_info = NULL,
    evaluation_time = NULL
  )
)


# extract_feature_expression (generic) -----------------------------------------

#'@title Internal function to extract feature expressions.
#'
#'@description Computes and extracts feature expressions for features
#'  used in a `familiarEnsemble` object.
#'
#'@param feature_similarity Table containing pairwise distance between
#'  sample. This is used to determine cluster information, and indicate which
#'  samples are similar. The table is created by the
#'  `extract_sample_similarity` method.
#'@inheritParams .extract_data
#'
#'@return A list with a data.table containing feature expressions.
#'@md
#'@keywords internal
setGeneric(
  "extract_feature_expression",
  function(
    object,
    data,
    feature_similarity,
    sample_similarity,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_similarity_metric = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    evaluation_times = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_feature_expression")
  }
)



# extract_feature_expression (familiarEnsemble) --------------------------------
setMethod(
  "extract_feature_expression",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    feature_similarity,
    sample_similarity,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_similarity_metric = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    evaluation_times = waiver(),
    message_indent = 0L,
    verbose = FALSE
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Compute feature expression."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Obtain evaluation times from the data.
    if (
      is.waive(evaluation_times) &&
      object@outcome_type %in% c("survival", "competing_risk")
    ) {
      evaluation_times <- object@settings$eval_times
      
    } else if (is.waive(evaluation_times)) {
      evaluation_times <- NULL
    }
    
    # Check if evaluation_times is correct.
    if (object@outcome_type %in% c("survival", "competing_risk")) {
      sapply(
        evaluation_times,
        .check_number_in_valid_range,
        var_name = "evaluation_times",
        range = c(0.0, Inf),
        closed = c(FALSE, TRUE)
      )
    }
    
    # Aggregate data
    data <- aggregate_data(data = data)
    
    # Retrieve input data. Note that we batch_normalisation is taken into
    # account even though the data are converted back to their original scale to
    # derive expressions.
    data <- process_input_data(
      object = object,
      data = data,
      stop_at = "batch_normalisation"
    )
    
    if (is_empty(data)) return(NULL)
    
    # Determine signature features
    model_features <- object@model_features
    
    # Maintain only important features. The current set is based on the
    # important features of the model, i.e. those that end up in the
    # model (potentially as a cluster).
    expression_data <- filter_features(data = data, available_features = model_features)
    
    # Perform inverse normalisation
    expression_data <- normalise_features(
      data = expression_data,
      feature_info_list = object@feature_info,
      invert = TRUE
    )
    
    # Perform inverse transformation
    expression_data <- transform_features(
      data = expression_data,
      feature_info_list = object@feature_info,
      invert = TRUE
    )
    
    # Add sample_name to expression_data
    row_names <- get_unique_row_names(expression_data)
    expression_data@data[, ":="(
      "sample_name" = row_names,
      "batch_id" = NULL,
      "sample_id" = NULL,
      "series_id" = NULL,
      "repetition_id" = NULL
    )]
    
    # Set expression data.
    expression_data <- methods::new(
      "familiarDataElementFeatureExpression",
      data = expression_data@data,
      feature_info = object@feature_info[model_features],
      evaluation_time = evaluation_times,
      value_column = model_features
    )
    
    # Add model name.
    expression_data <- add_model_name(expression_data, object)
    
    return(list(expression_data))
  }
)



# extract_feature_expression (prediction table) --------------------------------
setMethod(
  "extract_feature_expression",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("feature expression")
    
    return(NULL)
  }
)



# extract_feature_expression (dataObject) --------------------------------------
setMethod(
  "extract_feature_expression",
  signature(object = "dataObject"),
  function(
    object,
    data,
    feature_similarity,
    sample_similarity,
    feature_cluster_method = waiver(),
    feature_linkage_method = waiver(),
    feature_similarity_metric = waiver(),
    sample_cluster_method = waiver(),
    sample_linkage_method = waiver(),
    sample_similarity_metric = waiver(),
    evaluation_times = waiver(),
    message_indent = 0L,
    verbose = FALSE
  ) {
    
    # Message extraction start
    logger_message(
      paste0("Compute feature expression."),
      indent = message_indent,
      verbose = verbose
    )
    
    if (
      is.waive(evaluation_times) &&
      object@outcome_type %in% c("survival", "competing_risk")
    ) {
      # Get default evaluation times.
      settings <- .parse_evaluation_time_settings(
        data = object@data,
        outcome_type = object@outcome_type
      )
      
      evaluation_times <- settings$eval_times
      
    } else if (is.waive(evaluation_times)) {
      evaluation_times <- NULL
    }
    
    # Check if evaluation_times is correct.
    if (object@outcome_type %in% c("survival", "competing_risk")) {
      sapply(
        evaluation_times,
        .check_number_in_valid_range,
        var_name = "evaluation_times",
        range = c(0.0, Inf),
        closed = c(FALSE, TRUE)
      )
    }
    
    if (is_empty(object)) return(NULL)

    # Generate feature info by running the familiarTaskGenericFeatureInfo task.
    generic_feature_info_task <- new("familiarTaskGenericFeatureInfo")
    object@feature_info <- .perform_task(
      object = generic_feature_info_task,
      data = object
    )
    
    # Perform inverse transformation
    expression_data <- .copy(object)
    
    # Add sample_name to expression_data
    row_names <- get_unique_row_names(expression_data)
    expression_data@data[, ":="(
      "sample_name" = row_names,
      "batch_id" = NULL,
      "sample_id" = NULL,
      "series_id" = NULL,
      "repetition_id" = NULL
    )]
    
    # Set expression data.
    expression_data <- methods::new(
      "familiarDataElementFeatureExpression",
      data = expression_data@data,
      feature_info = object@feature_info,
      evaluation_time = evaluation_times,
      value_column = get_feature_columns(object)
    )
    
    # Add model name.
    expression_data <- add_model_name(expression_data, object)
    
    return(list(expression_data))
  }
)



# export_feature_expressions (generic) -----------------------------------------

#'@title Extract and export feature expressions.
#'
#'@description Extract and export feature expressions for the features in a
#'  familiarCollection.
#'
#'@param evaluation_time One or more time points that are used to create the
#'  outcome columns in expression plots. If not provided explicitly, this
#'  parameter is read from settings used at creation of the underlying
#'  `familiarData` objects. Only used for `survival` outcomes.
#'
#'@inheritParams export_all
#'@inheritParams plot_univariate_importance
#'
#'@inheritDotParams extract_feature_expression
#'@inheritDotParams as_familiar_collection
#'@inheritDotParams as_data_object
#'
#'@details 
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'  
#'  Feature similarity data can be created from `dataObject`, or `data.table` objects.
#'  For `data.table`, see \code{\link{as_data_object}} for additional arguments.
#'
#'@return A data.table (if `dir_path` is not provided), or nothing, as all data
#'  is exported to `csv` files.
#'@exportMethod export_feature_expressions
#'@md
#'@rdname export_feature_expressions-methods
setGeneric(
  "export_feature_expressions",
  function(
    object,
    dir_path = NULL,
    evaluation_time = waiver(),
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_feature_expressions")
  }
)



# export_feature_expressions (collection) --------------------------------------

#'@rdname export_feature_expressions-methods
setMethod(
  "export_feature_expressions",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    evaluation_time = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Extract data.
    x <- object@feature_expressions
    
    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)
    
    if (!is.waive(evaluation_time)) {
      
      # Check values.
      sapply(
        evaluation_time,
        .check_number_in_valid_range,
        var_name = "evaluation_time",
        range = c(0.0, Inf)
      )
      
      # Set clustering method.
      x <- lapply(
        x,
        function(x, evaluation_time) {
          x@evaluation_time <- evaluation_time
          return(x)
        },
        evaluation_time = evaluation_time
      )
    }
    
    return(.export(
      x = object,
      data_elements = x,
      dir_path = dir_path,
      aggregate_results = FALSE,
      type = "feature_expression",
      subtype = NULL,
      export_collection = export_collection
    ))
  }
)



# export_feature_expressions (general) -----------------------------------------

#'@rdname export_feature_expressions-methods
setMethod(
  "export_feature_expressions",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    evaluation_time = waiver(),
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "feature_expressions",
          "evaluation_times" = evaluation_time
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_feature_expressions,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "evaluation_time" = evaluation_time,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)



# .export (familiarDataElementFeatureExpression) -------------------------------
setMethod(
  ".export",
  signature(x = "familiarDataElementFeatureExpression"),
  function(
    x,
    x_list,
    aggregate_results = FALSE,
    ...
  ) {
    
    # Add grouping columns to data. Note that we do not merge the data elements.
    x <- lapply(
      x_list,
      .identifier_as_data_attribute,
      identifier = "all",
      as_grouping_column = TRUE
    )
    
    return(x)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R

as_metric <- function(metric, outcome_type, ...) {
  
  if (metric %in% .get_available_auc_roc_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricAUCROC",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricAUCROC", ...)
      )
    )
    
  } else if (metric %in% .get_available_brier_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricBrier",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricBrier", ...)
      )
    )
    
  } else if (metric %in% .get_available_accuracy_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricAccuracy",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricAccuracy", ...)
      )
    )
    
  } else if (metric %in% .get_available_balanced_accuracy_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricBalancedAccuracy",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricBalancedAccuracy", ...)
      )
    )
    
  } else if (metric %in% .get_available_balanced_error_rate_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricBalancedErrorRate",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricBalancedErrorRate", ...)
      )
    )
    
  } else if (metric %in% .get_available_cohen_kappa_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricCohenKappa",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricCohenKappa", ...)
      )
    )
    
  } else if (metric %in% .get_available_f1_score_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricF1Score",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricF1Score", ...)
      )
    )
    
  } else if (metric %in% .get_available_fdr_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricFDR",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricFDR", ...)
      )
    )
    
  } else if (metric %in% .get_available_informedness_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricInformedness",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricInformedness", ...)
      )
    )
    
  } else if (metric %in% .get_available_markedness_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricMarkedness",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricMarkedness", ...)
      )
    )
    
  } else if (metric %in% .get_available_mcc_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricMCC",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricMCC", ...)
      )
    )
    
  } else if (metric %in% .get_available_npv_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricNPV",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricNPV", ...)
      )
    )
    
  } else if (metric %in% .get_available_ppv_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricPPV",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricPPV", ...)
      )
    )
    
  } else if (metric %in% .get_available_sensitivity_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricSensitivity",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricSensitivity", ...)
      )
    )
    
  } else if (metric %in% .get_available_specificity_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricSpecificity",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricSpecificity", ...)
      )
    )
    
  } else if (metric %in% .get_available_youden_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricYouden",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricYouden", ...)
      )
    )
    
  } else if (metric %in% .get_available_mae_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricMAE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricMAE", ...)
      )
    )
    
  } else if (metric %in% .get_available_rae_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricRAE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricRAE", ...)
      )
    )
    
  } else if (metric %in% .get_available_mlae_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricMLAE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricMLAE", ...)
      )
    )
    
  } else if (metric %in% .get_available_mse_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricMSE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricMSE", ...)
      )
    )
    
  } else if (metric %in% .get_available_rse_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricRSE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricRSE", ...)
      )
    )
    
  } else if (metric %in% .get_available_msle_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricMSLE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricMSLE", ...)
      )
    )
    
  } else if (metric %in% .get_available_medea_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricMedianAE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricMedianAE", ...)
      )
    )
    
  } else if (metric %in% .get_available_rmse_metrics()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricRMSE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricRMSE", ...)
      )
    )
    
  } else if (metric %in% .get_available_rrse_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricRRSE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricRRSE", ...)
      )
    )
    
  } else if (metric %in% .get_available_rmsle_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricRMSLE",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricRMSLE", ...)
      )
    )
    
  } else if (metric %in% .get_available_r_squared_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricR2",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricR2", ...)
      )
    )
    
  } else if (metric %in% .get_available_explained_variance_metrics()) {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetricExplainedVariance",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricExplainedVariance", ...)
      )
    )
    
  } else if (metric %in% .get_available_concordance_index_harrell()) {
    metric_object <- do.call(
      methods::new, 
      args = c(
        list(
          "Class" = "familiarMetricConcordanceIndexHarrell",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetricConcordanceIndexHarrell", ...)
      )
    )
    
  } else {
    metric_object <- do.call(
      methods::new,
      args = c(
        list(
          "Class" = "familiarMetric",
          "metric" = metric,
          "outcome_type" = outcome_type
        ),
        .sanitise_dots("familiarMetric", ...)
      )
    )
  }
  
  return(metric_object)
}



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMetric"),
  function(object, ...) {
    return(FALSE)
  }
)



# is_higher_better -------------------------------------------------------------
setMethod(
  "is_higher_better",
  signature(metric = "familiarMetric"),
  function(metric, ...) {
    return(metric@higher_better)
  }
)



# is_higher_better -------------------------------------------------------------
setMethod(
  "is_higher_better",
  signature(metric = "character"),
  function(metric, ...) {
    # Create metric objects.
    metric_object_list <- lapply(
      metric,
      function(metric, dots) {
        return(do.call(
          as_metric,
          args = c(list("metric" = metric), dots)
        ))
      },
      dots = list(...)
    )

    # Check that the metrics are available.
    if (!all(sapply(metric_object_list, is_available))) {
      ..error(paste0(
        "is_higher_better: the following metrics are not available for ",
        metric_object_list[[1L]]@outcome_type, " outcomes: ",
        paste_s(metric[!sapply(metric_object_list, is_available)])
      ))
    }

    # Determine which metrics have a higher value that is better.
    higher_better_flags <- lapply(metric_object_list, is_higher_better)

    # Set metric names.
    names(higher_better_flags) <- metric

    # Return flags.
    return(unlist(higher_better_flags))
  }
)



# compute_metric_score (metric object) -----------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetric"),
  function(metric, data, ...) {
    # This is a fall-back method. This method should be defined for all
    # subclasses.
    return(NA_real_)
  }
)



# compute_metric_score (character) ---------------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "character"),
  function(metric, data, object = NULL, ...) {
    if (is(data, "dataObject")) {
      if (!is(object, "familiarModel") && !is(object, "familiarEnsemble")) {
        ..error(paste0(
          "compute_metric_score: object should be a familiarModel ",
          "or familiarEnsemble object."
        ))
      }
      
      # Get outcome type.
      outcome_type <- object@outcome_type
      
      # Compute prediction table.
      data <- .predict(
        object = object,
        data = data,
        ...
      )
      
    } else if (is(data, "familiarDataElementPredictionTable")) {
      # Get outcome type.
      outcome_type <- data@outcome_type
      
    } else {
      ..error_reached_unreachable_code("data are neither a data object or a prediction table")
    }
    
    # Create metric objects.
    metric_object_list <- lapply(
      metric,
      as_metric,
      outcome_type = outcome_type
    )

    # Check that the metrics are available.
    if (!all(sapply(metric_object_list, is_available))) {
      ..error(paste0(
        "compute_metric_score: the following metrics are not available for ",
        object@outcome_type, " outcomes: ",
        paste_s(metric[!sapply(metric_object_list, is_available)])
      ))
    }

    # Compute metric values.
    metric_values <- lapply(
      metric_object_list,
      compute_metric_score,
      data = data,
      ...
    )
    
    # Set names.
    names(metric_values) <- metric

    return(unlist(metric_values))
  }
)



# compute_objective_score ------------------------------------------------------
setMethod(
  "compute_objective_score",
  signature(metric = "familiarMetric"),
  function(metric, data = NULL, value = NULL, ...) {
    # Check that a baseline value was set
    if (is.null(metric@baseline_value)) {
      # Set the baseline value.
      metric <- set_metric_baseline_value(
        metric = metric,
        data = data,
        ...
      )
      
      # Check again
      if (is.null(metric@baseline_value)) {
        ..error_reached_unreachable_code(
          "compute_objective_score: baseline_value was not set."
        )
      } 
    }

    # Compute the value, if not provided.
    if (is.null(value)) {
      value <- compute_metric_score(
        metric = metric,
        data = data
      )
    }

    # Get the baseline_value
    baseline_value <- metric@baseline_value

    if (!is.finite(baseline_value)) {
      # Set a NA value for the objective.
      objective_value <- NA_real_
      
    } else {
      # Determine the optimal value, i.e. the best value attainable.
      optimal_value <- ifelse(
        is_higher_better(metric),
        max(metric@value_range),
        min(metric@value_range)
      )
      
      # If the baseline value is already perfect, use a small offset instead.
      if (baseline_value == optimal_value) {
        baseline_value <- ifelse(
          is_higher_better(metric),
          optimal_value - 1E-5,
          optimal_value + 1E-5
        )
      }

      # Compute the objective_value
      objective_value <- ifelse(
        is_higher_better(metric),
        (value - baseline_value) / (optimal_value - baseline_value),
        (baseline_value - value) / (baseline_value - optimal_value)
      )
    }

    # Ensure that all objective scores fall in the [-1, 1] range.
    if (!is.finite(objective_value)) {
      objective_value <- NA_real_
      
    } else if (objective_value < -1.0) {
      objective_value <- -1.0
      
    } else if (objective_value > 1.0) {
      ..error_reached_unreachable_code(paste0(
        "compute_objective_score: objective value exceeds the maximum of 1.0: ",
        objective_value
      ))
    }

    return(objective_value)
  }
)



# set_metric_baseline_value ----------------------------------------------------
setMethod(
  "set_metric_baseline_value",
  signature(metric = "familiarMetric"),
  function(metric, object = NULL, data) {
    # Obtain or create
    if (
      is(object, "familiarModel") ||
      is(object, "familiarVimpMethod") ||
      is(object, "familiarEnsemble")
    ) {
      outcome_info <- object@outcome_info
      
    } else if (is(data, "dataObject")) {
      
      if (is(data@outcome_info, "outcomeInfo")) {
        outcome_info <- object@outcome_info
        
      } else {
        # Compute outcome information from scratch.
        outcome_info <- .create_outcome_info(data)
        outcome_info <- .compute_outcome_distribution_data(
          object = outcome_info,
          data = data@data
        )
      }
      
    } else if (data.table::is.data.table(data)) {
      # Compute outcome information from scratch.
      outcome_info <- .create_outcome_info(data, outcome_type = metric@outcome_type)
      outcome_info <- .compute_outcome_distribution_data(
        object = outcome_info,
        data = data
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "set_metric_baseline_value: baseline_value could not be set ",
        "using the provided data."
      ))
    }

    # We need to identify the data source for determining baseline values.
    if (metric@outcome_type %in% c("binomial", "multinomial")) {
      # Get the frequency table and find the class with the majority.
      frequency_table <- outcome_info@distribution$frequency
      majority_class <- frequency_table$outcome[which.max(frequency_table$count)]
      
      # Get class levels.
      class_levels <- get_outcome_class_levels(object)
      
      prediction_list <- list()
      for (ii in seq_along(class_levels)) {
        # Update the predicted probabilities with 1.0 for the majority
        # class and 0.0 for minority classes.
        if (class_levels[ii] == majority_class) {
          prediction_list[[class_levels[ii]]] <- rep.int(
            1.0,
            times = get_n_samples(data, "repetition", count_unique = FALSE)
          )
          
        } else {
          prediction_list[[class_levels[ii]]] <- rep.int(
            0.0,
            times = get_n_samples(data, "repetition", count_unique = FALSE)
          )
        }
      }
      
      prediction_table <- as_prediction_table(
        x = prediction_list,
        type = "classification",
        data = data,
        model_object = object
      )
      
    } else if (metric@outcome_type %in% c("continuous")) {
      # Baseline median value.
      median_value <- outcome_info@distribution$median
      
      # Fill the prediction_table.
      prediction_table <- as_prediction_table(
        x = rep.int(
          median_value,
          times = get_n_samples(data, "repetition", count_unique = FALSE)
        ),
        type = "regression",
        data = data,
        model_object = object
      )
      
    } else if (metric@outcome_type %in% c("survival")) {
      # Median baseline survival
      if (!is.null(outcome_info@distribution$survival_probability)) {
        mean_survival_probability <- sum(c(
          min(outcome_info@distribution$survival_probability$survival_probability),
          max(outcome_info@distribution$survival_probability$survival_probability)
        )) / 2.0
        
      } else {
        mean_survival_probability <- NA_real_
      }
      
      # Fill the prediction_table.
      prediction_table <- as_prediction_table(
        x = rep.int(
          mean_survival_probability,
          times = get_n_samples(data, "repetition", count_unique = FALSE)
        ),
        type = "survival_probability",
        data = data,
        model_object = object
      )

    } else {
      ..error_outcome_type_not_implemented(metric@outcome_type)
    }

    # Compute metric value
    metric@baseline_value <- compute_metric_score(
      metric = metric,
      data = prediction_table
    )
    
    # Check the baseline value is a finite value. If it isn't set the value to
    # the extreme value of the range.
    if (!is.finite(metric@baseline_value)) {
      metric@baseline_value <- ifelse(
        is_higher_better(metric),
        min(metric@value_range),
        max(metric@value_range)
      )
    }

    return(metric)
  }
)



.check_metric_outcome_type <- function(
    metric,
    object = NULL, 
    outcome_type = NULL,
    as_flag = FALSE
) {
  
  # Obtain outcome_type
  if (is.null(outcome_type) && !is.null(object)) {
    outcome_type <- object@outcome_type
  }

  # Initialise metric
  metric_object <- as_metric(
    metric = metric,
    outcome_type = outcome_type
  )

  # Check if the metric is available.
  metric_available <- is_available(metric_object)

  if (as_flag) return(metric_available)

  # Check if the metric is available.
  if (!is_subclass(class(metric_object)[1L], "familiarMetric")) {
    ..error(paste0(
      metric, " is not a valid metric. ",
      "Please check the vignette for available performance metrics."
    ))
    
  } else if (!metric_available) {
    ..error(paste0(
      "The ", metric, " metric is not available for ",
      outcome_type, " outcomes."
    ))
  }
  
  return(invisible(TRUE))
}



.get_metric_default_range <- function(
    metric,
    object = NULL,
    outcome_type = NULL
) {
  # Get default range of metric scores, e.g. for plotting metric values.

  # Obtain outcome_type
  if (is.null(outcome_type) && !is.null(object)) {
    outcome_type <- object@outcome_type
  }

  # Initialise metric object.
  metric_object <- as_metric(
    metric = metric,
    outcome_type = outcome_type
  )

  return(metric_object@value_range)
}



.compute_metric_optimisation_score <- function(
    score_table,
    optimisation_function,
    replace_na = TRUE
) {
  # Compute an optimisation score from validation and training scores. This
  # optimisation score is typically computed for each set of hyperparameters
  # (param_id) and subsample (run_id).
  #
  # For hyperparameter optimisation scores are aggregated as follows:
  # validation and training scores --> optimisation score --> summary score.

  # Suppress NOTES due to non-standard evaluation in data.table
  optimisation_score <- training <- validation <- NULL

  # Select the correct optimisation function.
  optimisation_fun <- switch(
    optimisation_function,
    "max_validation" = ..optimisation_score_max_validation,
    "validation" = ..optimisation_score_max_validation,
    "balanced" = ..optimisation_score_balanced,
    "stronger_balance" = ..optimisation_score_stronger_balance,
    "validation_minus_sd" = ..optimisation_score_max_validation,
    "validation_25th_percentile" = ..optimisation_score_max_validation,
    "model_estimate" = ..optimisation_score_max_validation,
    "model_estimate_minus_sd" = ..optimisation_score_max_validation,
    "model_balanced_estimate" = ..optimisation_score_balanced,
    "model_balanced_estimate_minus_sd" = ..optimisation_score_balanced
  )

  # Find identifier columns.
  id_columns <- intersect(colnames(score_table), c("param_id", "run_id"))

  # Create formula
  formula <- stats::reformulate(
    termlabels = "data_set",
    response = paste0(c(id_columns, "metric"), collapse = " + ")
  )

  # Cast objective score wide by data_set.
  optimisation_table <- data.table::dcast(
    data = score_table[, mget(c(id_columns, "metric", "data_set", "objective_score"))],
    formula,
    value.var = "objective_score"
  )

  # Compute optimisation score based on objective scores.
  optimisation_table <- optimisation_table[
    ,
    list(
      "optimisation_score" = optimisation_fun(
        training = training,
        validation = validation
      )
    ),
    by = c(id_columns, "metric")
  ]
  
  # Replace NA entries with the minimum optimisation score.
  if (replace_na) {
    optimisation_table[
      is.na(optimisation_score),
      optimisation_score := ..get_replacement_optimisation_score()
    ]
  }

  # Average optimisation score over metrics.
  optimisation_table <- optimisation_table[
    ,
    list("optimisation_score" = mean(optimisation_score, na.rm = TRUE)),
    by = id_columns
  ]
  
  return(optimisation_table)
}



.summarise_metric_optimisation_score <- function(
    score_table,
    method,
    replace_na = TRUE
) {
  # Compute a summary score either directly from optimisation scores, or using a
  # model. This optimisation score is typically computed for each set of
  # hyperparameters.
  #
  # For hyperparameter optimisation scores are aggregated as follows:
  # validation and training scores --> optimisation score --> summary score.

  # Suppress NOTES due to non-standard evaluation in data.table
  optimisation_score <- NULL

  # Find identifier columns.
  id_columns <- intersect(colnames(score_table), "param_id")

  # Obtain the aggregation method.
  aggregation_method <- switch(
    method,
    "improvement_empirical_probability" = stats::median,
    "improvement_probability" = mean,
    "expected_improvement" = mean,
    "upper_confidence_bound" = mean,
    "bayes_upper_confidence_bound" = mean,
    "median" = stats::median,
    "mean" = mean,
    "max" = max,
    "min" = min
  )

  # Compute the mean optimisation score, overall, or per parameter id.
  score_table <- score_table[
    ,
    list(
      "optimisation_score" = aggregation_method(optimisation_score, na.rm = TRUE)
    ),
    by = id_columns
  ]

  # Replace NA entries with the minimum optimisation score.
  if (replace_na) {
    score_table[
      is.na(optimisation_score), 
      optimisation_score := ..get_replacement_optimisation_score()
    ]
  }

  return(score_table)
}



..optimisation_score_max_validation <- function(training = NULL, validation) {
  return(validation)
}



..optimisation_score_balanced <- function(training, validation) {
  # Start with the validation score.
  value <- validation

  # Penalise by difference between training and validation.
  value <- value - abs(validation - training)

  # Check that the value is finite.
  if (!is.finite(value)) return(value)

  # Add penalty term to models that perform worse than naive models on the
  # training data, i.e. have a objective score below 0.0. We could also write
  # value + training, but I think this way its clearer that a penalty is
  # intended.
  if (training < 0.0) value <- value - abs(training)

  return(value)
}



..optimisation_score_stronger_balance <- function(training, validation) {
  # Start with the validation score.
  value <- validation

  # Penalise by difference between training and validation.
  value <- value - 2.0 * abs(validation - training)

  # Check that the value is finite.
  if (!is.finite(value)) return(value)

  # Add penalty term to models that perform worse than naive models on the
  # training data, i.e. have a objective score below 0.0. We could also write
  # value + training, but I think this way its clearer that a penalty is
  # intended.
  if (training < 0.0) value <- value - 5.0 * abs(training)

  return(value)
}



.get_available_optimisation_functions <- function(hyperparameter_learner = NULL) {
  # All optimisation functions.
  all_optimisation_functions <- c(
    "validation", "max_validation", "balanced", "stronger_balance",
    "validation_minus_sd", "validation_25th_percentile", "model_estimate",
    "model_estimate_minus_sd", "model_balanced_estimate", "model_balanced_estimate_minus_sd"
  )

  if (is.null(hyperparameter_learner)) {
    return(all_optimisation_functions)
    
  } else if (hyperparameter_learner %in% c("random", "random_search")) {
    # Random search does not return an estimate that can be used for
    # optimisation.
    return(setdiff(
      all_optimisation_functions,
      c("model_estimate", "model_estimate_minus_sd", "model_balanced_estimate",
        "model_balanced_estimate_minus_sd"
      )
    ))
  }

  return(all_optimisation_functions)
}



..get_replacement_optimisation_score <- function() {
  return(-9.0)
}



.get_all_metrics <- function() {
  # Returns a list of all metrics.
  
  metrics <- c(
    .get_available_auc_roc_metrics(),
    .get_available_brier_metrics(),
    .get_available_confusion_matrix_metrics(),
    .get_available_regression_metrics(),
    .get_available_concordance_index_metrics()
  )
  
  return(metrics)
}



.get_default_metric <- function(outcome_type) {
  if (outcome_type %in% c("binomial", "multinomial")) {
    default_metric <- "auc_roc"
  } else if (outcome_type == "continuous") {
    default_metric <- "mse"
  } else if (outcome_type == "survival") {
    default_metric <- "concordance_index"
  } else if (outcome_type == "competing_risk") {
    ..error_outcome_type_not_implemented(outcome_type)
  } else {
    ..error_no_known_outcome_type(outcome_type)
  }
  
  return(default_metric)
}

socket_server <- function(
    host = "local_host",
    port = 6311L,
    timeout = 2678400.0) {
  # The socket server basically runs until a client gives a shutdown command.
  server_status <- "active"

  # Activate server
  socket <- serverSocket(port = port)
  on.exit(close(socket))

  while (!server_status == "shutdown") {
    # Reset server status, unless shutdown
    server_status <- "active"

    # Check if connection to server can be accepted from a client.
    repeat {
      connection_available <- socketSelect(list(socket), FALSE, NULL)

      if (connection_available) {
        con <- socketAccept(socket = socket, blocking = TRUE, open = "a+b")
        break
      }
    }

    # Try handshake by serializing to connecting client?
    if (socket_server_handshake(con)) {
      # Run execution loop. server_status may be set to close_connection or
      # shutdown by the client. Both options close the connection at the server
      # side, with shutdown additionaly breaking the server loop.
      while (server_status == "active") {
        server_status <- socket_server_execution_loop(con = con)
      }
    }

    # Close existing connection.
    close(con)
  }
}



socket_server_handshake <- function(con) {
  # Attempt to send instructions to obtain the process id of the client process.
  success <- tryCatch(
    serialize(
      list(
        "FUN" = eval, 
        "args" = list(quote(Sys.getpid()))
      ),
      connection = con
    ),
    error = identity
  )

  # Check if the attempt was successful, otherwise return FALSE, which closes
  # the connection.
  if (inherits(success, "error")) return(FALSE)

  # Attempt to read the process id of the client.
  pid <- tryCatch(
    unserialize(con),
    error = identity
  )

  # Return TRUE or FALSE. TRUE indicates that the connection is accepted and
  # works. FALSE closes the connection.
  return(is.integer(pid))
}



socket_server_execution_loop <- function(con) {
  # Attempt to read serialised data that has been pushed to the socket.
  input <- tryCatch(
    unserialize(connection = con),
    warning = identity,
    error = identity
  )

  # Break from loop in case a warning or error is given.
  if (inherits(input, "warning") || inherits(input, "error")) {
    return("active")
  }

  if (input$type == "EXEC") {
    # Execute a function

    # Execute the function
    output <- do.call(input$FUN, args = input$args)

    # Send back output of the function.
    output_send <- tryCatch(
      serialize(object = output, connection = con),
      error = identity
    )

    return("active")
    
  } else if (input$type == "CLOSE") {
    # Signal that the socket connection should be closed.
    return("close_connection")
    
  } else if (input$type == "SHUTDOWN") {
    # Signal that the server itself should shutdown.
    return("shutdown")
  }
}



socket_client_open_connection <- function(
    host = "localhost",
    port = 6311L
) {
  # Open a client-side connection.

  repeat {
    # Run a socketConnection with the shortest possible timeout, so that in case
    # of asynchronous activation (we attempt to start the client before a
    # server is listening), we will not be stuck for long.
    con <- tryCatch(
      socketConnection(
        host = host,
        port = port, 
        blocking = TRUE,
        server = FALSE, 
        open = "a+b",
        timeout = 1.0
      ),
      warning = identity,
      error = identity
    )

    if (inherits(con, "sockconn")) {
      # Try to perform the handshake with the server process. The connection is
      # closed if this is not successful and it will attempt to start a new
      # connection.
      if (socket_client_handshake(con)) {
        break
      } else {
        close(con)
      }
    }
  }

  return(con)
}



socket_client_handshake <- function(con) {
  # The server will send a simple function through serialisation that request
  # the client pid.
  received <- tryCatch(
    unserialize(con),
    error = identity
  )

  # If the socket connection cannot be read, return a FALSE. This closes the
  # connection in the socket_client_open_connection.
  if (inherits(received, "error")) return(FALSE)

  # Execute the function received from the server process, and find the process
  # id.
  pid <- do.call(received$FUN, args = received$args)

  # Try to send the pid to the server.
  success <- tryCatch(
    serialize(pid, connection = con),
    error = identity
  )

  return(!inherits(success, "error"))
}



socket_client_do_call <- function(con, FUN, args = NULL) {
  if (!inherits(con, "sockconn")) {
    stop("Connection is not a socket connection.")
  }

  # Send call to server to execute the function.
  serialize(
    object = list(
      "type" = "EXEC",
      "FUN" = FUN,
      "args" = args
    ),
    connection = con
  )

  # Wait for output to be received from server.
  repeat {
    output <- tryCatch(
      unserialize(connection = con),
      error = identity
    )

    # Break from the loop if the output is not an error.
    if (!inherits(output, "error")) break
  }

  return(output)
}


socket_client_close_connection <- function(con) {
  if (!inherits(con, "sockconn")) {
    stop("Connection is not a socket connection.")
  }

  # Skip if already closed.
  if (isOpen(con)) {
    serialize(object = list("type" = "CLOSE"), connection = con)

    # Close connection from the client side.
    close(con)
  }
  
  return(invisible(NULL))
}



socket_client_server_shutdown <- function(
    con = NULL,
    host = "localhost",
    port = 6311L
) {
  
  if (!inherits(con, "sockconn")) {
    # Open a connection
    con <- socket_client_open_connection(host = host, port = port)
    on.exit(close(con))
    
  } else if (!isOpen(con)) {
    # Attempt to shut client-side connection and re-open it.
    close(con)

    # Open new connection
    con <- socket_client_open_connection(host = host, port = port)
    on.exit(close(con))
    
  } else {
    on.exit(close(con))
  }

  # Signal the server to close the connection and shutdown.
  serialize(
    list("type" = "SHUTDOWN"),
    connection = con
  )
  
  return(invisible(NULL))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# plot_model_performance (generic) ---------------------------------------------

#' @title Plot model performance.
#'
#' @description This method creates plots that show model performance from the
#'   data stored in a familiarCollection object. This method may create several
#'   types of plots, as determined by `plot_type`.
#'
#' @param dir_path (*optional*) Path to the directory where created performance
#'   plots are saved to. Output is saved in the `performance` subdirectory. If
#'   `NULL` no figures are saved, but are returned instead.
#' @param plot_type (*optional*) Type of plot to draw. This is one of `heatmap`
#'   (draws a heatmap), `barplot` (draws a barplot with confidence intervals),
#'   `boxplot` (draws a boxplot) and `violinplot` (draws a violin plot).
#'   Defaults to `violinplot`.
#'
#'   The choice for `plot_type` affects several other arguments, e.g. `color_by`
#'   is not used for `heatmap` and `y_axis_by` is only used by `heatmap`.
#' @param discrete_palette (*optional*) Palette for colouring plot elements
#'   indicated by the `color_by` argument (if any). Only used if `plot_type` is
#'   not `heatmap`. `familiar` has a default palette. Other palettes are
#'   supported by the `paletteer` package, `grDevices::palette.pals()` (requires
#'   R >= 4.0.0), `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`,
#'   `heat.colors`, `terrain.colors`, `topo.colors` and `cm.colors`, which
#'   correspond to the palettes of the same name in `grDevices`. You may also
#'   specify your own palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param gradient_palette (*optional*) Sequential or divergent palette used to
#'   color the raster in `heatmap` plots. This argument is not used for other
#'   `plot_type` value. `familiar` has a default palette. Other palettes are
#'   supported by the `paletteer` package, `grDevices::palette.pals()` (requires
#'   R >= 4.0.0), `grDevices::hcl.pals()` (requires R >= 3.6.0) and `rainbow`,
#'   `heat.colors`, `terrain.colors`, `topo.colors` and `cm.colors`, which
#'   correspond to the palettes of the same name in `grDevices`. You may also
#'   specify your own palette by providing a vector of colour names listed by
#'   `grDevices::colors()` or through hexadecimal RGB strings.
#' @param gradient_palette_range (*optional*) Numerical range used to span the
#'   gradient. This should be a range of two values, e.g. `c(0, 1)`. Lower or
#'   upper boundary can be unset by using `NA`. If not set, the full
#'   metric-specific range is used.
#' @param annotate_performance (*optional*) Indicates whether performance in
#'   heatmaps should be annotated with text. Can be `none`, `value` (default),
#'   or `value_ci` (median value plus 95% credibility intervals).
#'
#' @inheritParams as_familiar_collection
#' @inheritParams plot_univariate_importance
#' @inheritParams .check_input_plot_args
#' @inheritParams .check_plot_splitting_variables
#' @inheritDotParams extract_performance -object
#' @inheritDotParams as_familiar_collection -object
#' @inheritDotParams ggplot2::ggsave -height -width -units -path -filename -plot
#'
#' @details This function plots model performance based on empirical bootstraps,
#'   using various plot representations.
#'
#'   Available splitting variables are: `vimp_method`, `learner`, `data_set`,
#'   `evaluation_time` (survival outcome only) and `metric`. The default for
#'   `heatmap` is to split by `metric`, facet by `data_set` and
#'   `evaluation_time`, position `learner` along the x-axis and `vimp_method`
#'   along the y-axis. The `color_by` argument is not used. The only valid
#'   options for `x_axis_by` and `y_axis_by` are `learner` and `vimp_method`.
#'
#'   For other plot types (`barplot`, `boxplot` and `violinplot`), depends on
#'   the number of learners and feature selection methods:
#'
#'  * *one feature selection method and one learner*: the default is to split by
#'   `metric`, and have `data_set` along the x-axis.
#'
#'  * *one feature selection and multiple learners*: the default is to split by
#'   `metric`, facet by `data_set` and have `learner` along the x-axis.
#'
#'  * *multiple feature selection methods and one learner*: the default is to
#'   split by `metric`, facet by `data_set` and have `vimp_method` along the
#'   x-axis.
#'
#'  * *multiple feature selection methods and learners*: the default is to split
#'   by `metric`, facet by `data_set`, colour by `vimp_method` and have `learner`
#'   along the x-axis.
#'
#'   If applicable, additional faceting is performed for `evaluation_time`.
#'
#'   Labeling methods such as `set_vimp_method_names` or `set_data_set_names` can
#'   be applied to the `familiarCollection` object to update labels, and order
#'   the output in the figure.
#'
#' @return `NULL` or list of plot objects, if `dir_path` is `NULL`.
#'
#' @exportMethod plot_model_performance
#' @md
#' @rdname plot_model_performance-methods
setGeneric(
  "plot_model_performance",
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    plot_type = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    gradient_palette_range = waiver(),
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    rotate_x_tick_labels = waiver(),
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    annotate_performance = NULL,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("plot_model_performance")
  }
)



# plot_model_performance (general) ---------------------------------------------

#' @rdname plot_model_performance-methods
setMethod(
  "plot_model_performance",
  signature(object = "ANY"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    plot_type = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    gradient_palette_range = waiver(),
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    rotate_x_tick_labels = waiver(),
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    annotate_performance = NULL,
    export_collection = FALSE,
    ...
  ) {
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "model_performance"
        ),
        list(...)
      )
    )
    
    return(do.call(
      plot_model_performance,
      args = list(
        "object" = object,
        "draw" = draw,
        "dir_path" = dir_path,
        "split_by" = split_by,
        "x_axis_by" = x_axis_by,
        "y_axis_by" = y_axis_by,
        "color_by" = color_by,
        "facet_by" = facet_by,
        "facet_wrap_cols" = facet_wrap_cols,
        "ggtheme" = ggtheme,
        "plot_type" = plot_type,
        "discrete_palette" = discrete_palette,
        "gradient_palette" = gradient_palette,
        "gradient_palette_range" = gradient_palette_range,
        "x_label" = x_label,
        "y_label" = y_label,
        "legend_label" = legend_label,
        "plot_title" = plot_title,
        "plot_sub_title" = plot_sub_title,
        "caption" = caption,
        "rotate_x_tick_labels" = rotate_x_tick_labels,
        "y_range" = y_range,
        "y_n_breaks" = y_n_breaks,
        "y_breaks" = y_breaks,
        "width" = width,
        "height" = height,
        "units" = units,
        "annotate_performance" = annotate_performance,
        "export_collection" = export_collection
      )
    ))
  }
)



# plot_model_performance (collection) ------------------------------------------

#' @rdname plot_model_performance-methods
setMethod(
  "plot_model_performance",
  signature(object = "familiarCollection"),
  function(
    object,
    draw = FALSE,
    dir_path = NULL,
    split_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    color_by = NULL,
    facet_by = NULL,
    facet_wrap_cols = NULL,
    plot_type = NULL,
    ggtheme = NULL,
    discrete_palette = NULL,
    gradient_palette = NULL,
    gradient_palette_range = waiver(),
    x_label = waiver(),
    y_label = waiver(),
    legend_label = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = NULL,
    rotate_x_tick_labels = waiver(),
    y_range = NULL,
    y_n_breaks = 5L,
    y_breaks = NULL,
    width = waiver(),
    height = waiver(),
    units = waiver(),
    annotate_performance = NULL,
    export_collection = FALSE,
    ...
  ) {

    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Check input arguments ----------------------------------------------------
    
    # ggtheme
    ggtheme <- .check_ggtheme(ggtheme)
    
    # rotate_x_tick_labels
    if (is.waive(rotate_x_tick_labels)) {
      rotate_x_tick_labels <- FALSE
    }
    
    # Check plot type.
    if (!is.null(plot_type)) {
      .check_parameter_value_is_valid(
        x = plot_type,
        var_name = "plot_type",
        values = c("heatmap", "barplot", "boxplot", "violinplot")
      )
      
    } else {
      # Set default to violin plot.
      plot_type <- "violinplot"
    }

    if (plot_type == "heatmap") {
      # For the heatmap we require aggregated data.

      # Load the data.
      x <- export_model_performance(
        object = object, 
        aggregate_results = TRUE
      )

      # Check that the data are not empty.
      if (is_empty(x)) return(NULL)
      
    } else {
      # Load the data.
      x <- export_model_performance(
        object = object, 
        aggregate_results = FALSE
      )

      # Check that the data are not empty.
      if (is_empty(x)) return(NULL)

      # Check that data are disaggregated
      if (any(sapply(x, function(x) (x@is_aggregated)))) {
        ..warning(paste0(
          "Creating a ", plot_type, " requires de-aggregated data, ",
          "which are not available."
        ))
        
        return(NULL)
      }

      # Check that data are not point estimates.
      if (all(sapply(x, function(x) (x@estimation_type == "point")))) {
        ..warning(paste0(
          "Creating a ", plot_type, " requires bias-corrected estimates or ",
          "bootstrap confidence interval estimates instead of point estimates."
        ))
        
        return(NULL)
      }

      # For bootstrap confidence interval select only the bci data.
      bci_data <- sapply(x, function(x) {
        x@estimation_type %in% c("bci", "bootstrap_confidence_interval")
      })
      
      if (any(bci_data)) x <- x[bci_data]
    }

    # Check that the data are not evaluated at the model level.
    if (all(sapply(x, function(x) (x@detail_level == "model")))) {
      ..warning_no_comparison_between_models()
      return(NULL)
    }

    # Obtain data element from list.
    if (is.list(x)) {
      if (is_empty(x)) return(NULL)

      if (length(x) > 1L) {
        ..error_reached_unreachable_code(
          "plot_model_performance: list of data elements contains unmerged elements."
        )
      }

      # Get x directly.
      x <- x[[1L]]
    }

    # Check that the data are not empty.
    if (is_empty(x)) return(NULL)

    # Check package requirements for plotting.
    if (!require_package(
      x = ..required_plotting_packages(extended = FALSE),
      purpose = "to plot model performance",
      message_type = "warning"
    )) {
      return(NULL)
    }

    # Add evaluation time as a splitting variable.
    if (object@outcome_type %in% c("survival")) {
      split_variable <- "evaluation_time"
    } else {
      split_variable <- NULL
    }

    # Add default splitting variables.
    if (
      is.null(split_by) &&
      is.null(color_by) &&
      is.null(facet_by) &&
      is.null(x_axis_by) &&
      is.null(y_axis_by)
    ) {
      if (plot_type == "heatmap") {
        # Split by metric.
        split_by <- c("metric")

        # Set faceting variables.
        facet_by <- c("data_set", split_variable)

        # Set x-axis variables.
        x_axis_by <- c("learner")

        # Set y-axis variables. This splitting variable is only used in
        # heatmaps.
        y_axis_by <- c("vimp_method")
        
      } else {
        # Determine the number of learners and variable importance methods.
        n_learner <- nlevels(x@data$learner)
        n_vimp_method <- nlevels(x@data$vimp_method)

        # Split by metric.
        split_by <- c("metric")

        # Set facetting variables.
        if (n_learner > 1L || n_vimp_method > 1L) {
          facet_by <- c("data_set")
        }

        # Set color variables. This splitting variable is only used in
        # non-heatmap plots.
        if (n_learner > 1L && n_vimp_method > 1L) {
          color_by <- c("vimp_method")
        }

        # Set x-axis variables.
        if (n_learner == 1L && n_vimp_method == 1L) {
          x_axis_by <- c("data_set")
        } else if (n_learner == 1L && n_vimp_method > 1L) {
          x_axis_by <- c("vimp_method")
        } else {
          x_axis_by <- c("learner")
        }

        # Add split variable (if any) to facet_by.
        facet_by <- c(facet_by, split_variable)
      }
    }

    if (plot_type == "heatmap") {
      # Check splitting variables and generate sanitised output

      # Check if the color_by argument is provided.
      if (!is.null(color_by)) {
        ..warning("The color_by argument is ignored for heatmaps.")
        color_by <- NULL
      }

      split_var_list <- .check_plot_splitting_variables(
        x = x@data,
        split_by = split_by,
        facet_by = facet_by,
        x_axis_by = x_axis_by,
        y_axis_by = y_axis_by,
        available = c(
          "metric", "data_set", "vimp_method",
          "learner", split_variable
        )
      )
    } else {
      # Check if the y_axis_by argument is provided.
      if (!is.null(y_axis_by)) {
        ..warning("The y_axis_by argument is ignored for non-heatmap plots.")
      }

      # Check splitting variables and generate sanitised output
      split_var_list <- .check_plot_splitting_variables(
        x = x@data,
        split_by = split_by,
        color_by = color_by,
        facet_by = facet_by,
        x_axis_by = x_axis_by,
        available = c(
          "metric", "data_set", "vimp_method",
          "learner", split_variable
        )
      )
    }

    # Update splitting variables
    split_by <- split_var_list$split_by
    color_by <- split_var_list$color_by
    facet_by <- split_var_list$facet_by
    if (!is.null(split_var_list$x_axis_by)) {
      x_axis_by <- split_var_list$x_axis_by
    }
    if (!is.null(split_var_list$y_axis_by)) {
      y_axis_by <- split_var_list$y_axis_by
    }

    if (plot_type == "heatmap") {
      # Check that x_axis_by and y_axis_by only take vimp_method or learner.
      if (!x_axis_by %in% c("vimp_method", "learner", "data_set", split_variable)) {
        ..error("The x_axis_by argument should be one of vimp_method, learner or data_set.")
      }
      if (!y_axis_by %in% c("vimp_method", "learner", "data_set", split_variable)) {
        ..error("The y_axis_by argument should be one of vimp_method, learner or data_set.")
      }
    }

    if (is.null(x_axis_by)) {
      x_axis_by <- setdiff(
        c("metric", "data_set", "vimp_method", "learner", split_variable),
        c(split_by, color_by, facet_by, y_axis_by)
      )
      if (length(x_axis_by) == 0L) ..error("The x_axis_by argument should be set.")
      if (length(x_axis_by) > 1L && "metric" %in% c(x_axis_by)) {
        x_axis_by <- "metric"
      } else {
        x_axis_by <- x_axis_by[1L]
      }
    }

    # x_label
    if (is.waive(x_label)) {
      x_label <- switch(
        x_axis_by,
        learner = "learner",
        vimp_method = "feature selection method",
        data_set = "dataset",
        metric = "metric",
        evaluation_time = "time"
      )
    }

    # annotate_performance
    if (is.null(annotate_performance)) {
      annotate_performance <- "value"
    } else if (is.logical(annotate_performance)) {
      annotate_performance <- ifelse(annotate_performance, "value", "none")
    }
    .check_parameter_value_is_valid(
      x = annotate_performance,
      var_name = "annotate_performance",
      values = c("none", "value", "value_ci")
    )

    .check_input_plot_args(
      facet_wrap_cols = facet_wrap_cols,
      x_label = x_label,
      plot_title = plot_title,
      plot_sub_title = plot_sub_title,
      caption = caption,
      rotate_x_tick_labels = rotate_x_tick_labels
    )

    # Create plots -------------------------------------------------------------

    # Determine if subtitle should be generated.
    autogenerate_plot_subtitle <- is.waive(plot_sub_title)

    # Split data.
    if (!is.null(split_by)) {
      x_split <- split(
        x@data, 
        by = split_by, 
        drop = FALSE
      )
      
    } else {
      x_split <- list("null.name" = x@data)
    }

    # Store plots to list in case dir_path is absent.
    if (is.null(dir_path)) plot_list <- list()

    # Iterate over data splits.
    for (ii in names(x_split)) {
      # Skip empty datasets.
      if (is_empty(x_split[[ii]])) next

      if (is.waive(plot_title)) plot_title <- "Model performance"

      # Declare subtitle components.
      additional_subtitle <- NULL

      # Add evaluation time as subtitle component if it is not used
      # otherwise.
      if (
        !"evaluation_time" %in% c(split_by, color_by, facet_by) &&
        object@outcome_type %in% c("survival")
      ) {
        additional_subtitle <- c(
          additional_subtitle,
          .add_time_to_plot_subtitle(x_split[[ii]]$evaluation_time[1L])
        )
      }

      if (autogenerate_plot_subtitle) {
        plot_sub_title <- .create_plot_subtitle(
          split_by = split_by,
          additional = additional_subtitle,
          x = x_split[[ii]]
        )
      }

      # Generate plot
      p <- .plot_model_performance_plot(
        x = x_split[[ii]],
        x_axis_by = x_axis_by,
        y_axis_by = y_axis_by,
        color_by = color_by,
        facet_by = facet_by,
        facet_wrap_cols = facet_wrap_cols,
        plot_type = plot_type,
        ggtheme = ggtheme,
        discrete_palette = discrete_palette,
        gradient_palette = gradient_palette,
        gradient_palette_range = gradient_palette_range,
        x_label = x_label,
        y_label = y_label,
        legend_label = legend_label,
        plot_title = plot_title,
        plot_sub_title = plot_sub_title,
        caption = caption,
        rotate_x_tick_labels = rotate_x_tick_labels,
        y_range = y_range,
        y_n_breaks = y_n_breaks,
        y_breaks = y_breaks,
        annotate_performance = annotate_performance,
        outcome_type = object@outcome_type
      )

      # Check empty output
      if (is.null(p)) next

      # Draw figure.
      if (draw) .draw_plot(plot_or_grob = p)

      # Save and export
      if (!is.null(dir_path)) {
        # Obtain decent default values for the plot.
        def_plot_dims <- .determine_model_performance_plot_dimensions(
          x = x_split[[ii]],
          plot_type = plot_type,
          x_axis_by = x_axis_by,
          y_axis_by = y_axis_by,
          facet_by = facet_by,
          facet_wrap_cols = facet_wrap_cols,
          rotate_x_tick_labels = rotate_x_tick_labels
        )

        # Save to file.
        do.call(
          .save_plot_to_file,
          args = c(
            list(
              "plot_or_grob" = p,
              "object" = object,
              "dir_path" = dir_path,
              "type" = "performance",
              "subtype" = plot_type,
              "x" = x_split[[ii]],
              "split_by" = split_by,
              "height" = ifelse(is.waive(height), def_plot_dims[1L], height),
              "width" = ifelse(is.waive(width), def_plot_dims[2L], width),
              "units" = ifelse(is.waive(units), "cm", units)
            ),
            list(...)
          )
        )
        
      } else {
        # Store as list for export.
        plot_list <- c(plot_list, list(p))
      }
    }

    # Generate output
    return(.get_plot_results(
      dir_path = dir_path,
      plot_list = plot_list,
      export_collection = export_collection,
      object = object
    ))
  }
)



.plot_model_performance_plot <- function(
    x,
    x_axis_by,
    y_axis_by,
    color_by,
    facet_by,
    facet_wrap_cols,
    plot_type,
    ggtheme,
    discrete_palette,
    gradient_palette,
    gradient_palette_range,
    x_label,
    y_label,
    legend_label,
    plot_title,
    plot_sub_title,
    caption,
    rotate_x_tick_labels,
    y_range,
    y_n_breaks,
    y_breaks,
    annotate_performance,
    outcome_type
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  value <- metric <- median <- ci_low <- ci_up <- NULL
  
  # Identify metrics in the current dataset.
  metrics <- as.character(unique(x$metric))

  # Check y-range for all plots except heatmaps.
  if (plot_type != "heatmap") {
    if (is.null(y_range)) {
      # Obtain default ranges for the metrics.
      metric_ranges <- lapply(
        metrics, 
        .get_metric_default_range, 
        outcome_type = outcome_type
      )

      # Give a name to the list elements.
      names(metric_ranges) <- metrics

      # Placeholder range
      y_range <- c(Inf, -Inf)

      for (current_metric in metrics) {
        metric_range <- metric_ranges[[current_metric]]

        # Replace any positive infinite value by the max range in the data.
        if (any(metric_range == Inf)) {
          metric_range[metric_range == Inf] <- max(
            x[metric == current_metric, value], na.rm = TRUE
          )
        }

        # Replace any negative infinite value by the min range in the data.
        if (any(metric_range == -Inf)) {
          metric_range[metric_range == -Inf] <- min(
            x[metric == current_metric, value], na.rm = TRUE
          )
        }

        if (y_range[1L] > min(metric_range)) {
          y_range[1L] <- min(metric_range)
        }

        if (y_range[2L] < max(metric_range)) {
          y_range[2L] <- max(metric_range)
        }
      }
    } else {
      .check_input_plot_args(y_range = y_range)
    }

    # y_breaks
    if (is.null(y_breaks)) {
      .check_input_plot_args(
        y_range = y_range,
        y_n_breaks = y_n_breaks
      )

      # Create breaks and update x_range
      y_breaks <- labeling::extended(
        m = y_n_breaks,
        dmin = y_range[1L],
        dmax = y_range[2L],
        only.loose = TRUE
      )
      
      y_range <- c(0.0, tail(y_breaks, n = 1L))
      
    } else {
      .check_input_plot_args(y_breaks = y_breaks)
    }

    # y_label for non-heatmap plots
    if (is.waive(y_label)) {
      y_label <- ifelse(length(metrics) == 1L, metrics, "value")
    }

    # Create a legend label
    legend_label <- .create_plot_legend_title(
      user_label = legend_label,
      color_by = color_by
    )
    
  } else {
    # y-label for heatmap plots
    if (is.waive(y_label)) {
      y_label <- switch(
        y_axis_by,
        learner = "learner",
        vimp_method = "feature selection method",
        data_set = "dataset",
        metric = "metric",
        evaluation_time = "time"
      )
    }

    # gradient_palette_range
    if (is.waive(gradient_palette_range)) {
      if (length(metrics) == 1L) {
        gradient_palette_range <- .get_metric_default_range(
          metric = metrics,
          outcome_type = outcome_type
        )

        # Replace a positive infinite value by the max range in the data.
        if (gradient_palette_range[2L] == Inf) {
          gradient_palette_range[2L] <- max(x[metric == metrics, value], na.rm = TRUE)
        }

        # Replace any negative infinite value by the min range in the data.
        if (gradient_palette_range[1L] == -Inf) {
          gradient_palette_range[1L] <- min(x[metric == metrics, value], na.rm = TRUE)
        }

        gradient_was_provided <- FALSE
        
      } else {
        # If metric for whatever reason is not a single metric.
        gradient_palette_range <- c(NA, NA)
        gradient_was_provided <- FALSE
      }
      
    } else {
      # Check for NULL.
      if (is.null(gradient_palette_range)) {
        gradient_palette_range <- c(NA, NA)
      }

      gradient_was_provided <- TRUE
    }

    # Create a legend label
    legend_label <- ifelse(length(metrics) == 1L && is.waive(legend_label), metrics, "value")
  }

  # Check remaining input arguments.
  .check_input_plot_args(
    y_label = y_label,
    legend_label = legend_label
  )

  # Create basic plot
  p <- ggplot2::ggplot()
  p <- p + ggtheme

  if (plot_type == "heatmap") {
    # Heatmap ------------------------------------------------------------------

    # Create summary data.
    x_bar <- x[
      ,
      list(
        "median" = stats::median(value, na.rm = TRUE),
        "ci_up" = stats::quantile(value, probs = 0.975, na.rm = TRUE, names = FALSE),
        "ci_low" = stats::quantile(value, probs = 0.025, na.rm = TRUE, names = FALSE)
      ),
      by = c("metric", "data_set", "vimp_method", "learner")
    ]

    # Determine what direction a metric has.
    if (length(metrics) == 1L) {
      invert_scale <- !is_higher_better(
        metric = metrics,
        outcome_type = outcome_type
      )
      
    } else {
      invert_scale <- FALSE
    }

    # Determine the type of sequential colorscale. This has no effect if the
    # user provides a colorscale.
    if (length(metrics) == 1L && !gradient_was_provided) {
      palette_type <- ifelse(
        length(gradient_palette_range) > 2L,
        "divergent",
        "sequential"
      )
      
    } else {
      palette_type <- "sequential"
    }

    # Form heatmap raster.
    p <- p + ggplot2::geom_raster(
      data = x_bar,
      mapping = ggplot2::aes(
        x = !!sym(x_axis_by),
        y = !!sym(y_axis_by),
        fill = !!sym("median")
      )
    )

    # Colors
    gradient_colours <- .get_palette(
      x = gradient_palette, 
      palette_type = palette_type
    )
    
    if (invert_scale) gradient_colours <- rev(gradient_colours)

    # Add gradient palette.
    p <- p + ggplot2::scale_fill_gradientn(
      name = legend_label,
      colors = gradient_colours,
      limits = range(gradient_palette_range)
    )

    # Obtain default settings.
    text_settings <- .get_plot_geom_text_settings(ggtheme = ggtheme)

    # Show performance value as text.
    if (annotate_performance == "value") {
      # Show median value.
      x_bar[is.finite(median), "performance_text" := .format_plot_number(median, min_common_base = 0L)]

      # Add to figure.
      p <- p + ggplot2::geom_text(
        data = x_bar,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          y = !!sym(y_axis_by),
          label = !!sym("performance_text")
        ),
        colour = text_settings$colour,
        family = text_settings$family,
        fontface = text_settings$face,
        size = text_settings$geom_text_size
      )
      
    } else if (annotate_performance == "value_ci") {
      # Show median value and credibility interval
      x_bar[
        is.finite(median),
        "performance_text" := paste0(
          .format_plot_number(median),
          "\n(",
          .format_plot_number(ci_low),
          "\u2013",
          .format_plot_number(ci_up),
          ")"
        )
      ]

      # Add to figure.
      p <- p + ggplot2::geom_text(
        data = x_bar,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          y = !!sym(y_axis_by),
          label = !!sym("performance_text")
        ),
        colour = text_settings$colour,
        family = text_settings$family,
        fontface = text_settings$face,
        size = text_settings$geom_text_size
      )
    }
    
  } else if (plot_type == "barplot") {
    # Barplot ------------------------------------------------------------------

    # Create data for bar
    x_bar <- x[
      ,
      list(
        "median" = stats::median(value, na.rm = TRUE),
        "ci_up" = stats::quantile(value, probs = 0.975, na.rm = TRUE, names = FALSE),
        "ci_low" = stats::quantile(value, probs = 0.025, na.rm = TRUE, names = FALSE)
      ),
      by = c("metric", "data_set", "vimp_method", "learner")
    ]
    
    # Generate a guide table
    guide_list <- .create_plot_guide_table(
      x = x_bar,
      color_by = color_by,
      discrete_palette = discrete_palette
    )

    # Extract data
    x_bar <- guide_list$data

    # Set breaks.
    p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)

    if (is.null(color_by)) {
      # Add barplot
      p <- p + ggplot2::geom_bar(
        data = x_bar,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          y = !!sym("median")
        ),
        stat = "identity",
        position = "dodge"
      )

      # Add error bars
      p <- p + ggplot2::geom_errorbar(
        data = x_bar,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          ymin = !!sym("ci_low"),
          ymax = !!sym("ci_up")
        ),
        position = ggplot2::position_dodge(width = 0.9),
        width = 0.20
      )
      
    } else {
      # Extract guide_table for color
      g_color <- guide_list$guide_color

      # Add barplot.
      p <- p + ggplot2::geom_bar(
        data = x_bar,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          y = !!sym("median"),
          fill = !!sym("color_breaks")
        ),
        stat = "identity",
        position = ggplot2::position_dodge(width = 0.9)
      )

      # Add error bars
      p <- p + ggplot2::geom_errorbar(
        data = x_bar,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          ymin = !!sym("ci_low"),
          ymax = !!sym("ci_up"),
          group = !!sym("color_breaks")
        ),
        position = ggplot2::position_dodge(width = 0.9),
        width = 0.20
      )

      # Set fill colours.
      p <- p + ggplot2::scale_fill_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        drop = FALSE
      )
    }

    # Plot to Cartesian coordinates.
    p <- p + ggplot2::coord_cartesian(ylim = y_range)
    
  } else if (plot_type == "boxplot") {
    # Boxplot ------------------------------------------------------------------

    # Generate a guide table
    guide_list <- .create_plot_guide_table(
      x = x,
      color_by = color_by,
      discrete_palette = discrete_palette
    )

    # Extract data
    x <- guide_list$data

    # Set breaks.
    p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)

    if (is.null(color_by)) {
      # Create boxplot.
      p <- p + ggplot2::geom_boxplot(
        data = x,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          y = !!sym("value")
        ),
        outlier.alpha = 0.1
      )
      
    } else {
      # Extract guide_table for color
      g_color <- guide_list$guide_color

      # Create boxplot.
      p <- p + ggplot2::geom_boxplot(
        data = x,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          y = !!sym("value"),
          colour = !!sym("color_breaks")
        ),
        outlier.alpha = 0.1
      )

      # Set fill colours.
      p <- p + ggplot2::scale_colour_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        drop = FALSE
      )
    }

    # Plot to Cartesian coordinates.
    p <- p + ggplot2::coord_cartesian(ylim = y_range)
    
  } else if (plot_type == "violinplot") {
    # Violinplot ---------------------------------------------------------------

    # Generate a guide table
    guide_list <- .create_plot_guide_table(
      x = x,
      color_by = color_by,
      discrete_palette = discrete_palette
    )

    # Extract data
    x <- guide_list$data

    # Set breaks.
    p <- p + ggplot2::scale_y_continuous(breaks = y_breaks)

    if (is.null(color_by)) {
      # Create boxplot.
      p <- p + ggplot2::geom_violin(
        data = x,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          y = !!sym("value")
        ),
        quantiles = c(0.025, 0.5, 0.975),
        quantile.linetype = ggtheme$line$linetype,
        scale = "width",
        position = ggplot2::position_dodge(width = 1.0)
      )
      
    } else {
      # Extract guide_table for color
      g_color <- guide_list$guide_color

      # Create boxplot.
      p <- p + ggplot2::geom_violin(
        data = x,
        mapping = ggplot2::aes(
          x = !!sym(x_axis_by),
          y = !!sym("value"),
          fill = !!sym("color_breaks")
        ),
        quantiles = c(0.025, 0.5, 0.975),
        quantile.linetype = ggtheme$line$linetype,
        scale = "width",
        position = ggplot2::position_dodge(width = 1.0)
      )

      # Set fill colours.
      p <- p + ggplot2::scale_fill_manual(
        name = legend_label$guide_color,
        values = g_color$color_values,
        breaks = g_color$color_breaks,
        drop = FALSE
      )
    }

    # Plot to Cartesian coordinates.
    p <- p + ggplot2::coord_cartesian(ylim = y_range)
  }

  # Determine how things are facetted
  facet_by_list <- .parse_plot_facet_by(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )

  if (!is.null(facet_by)) {
    if (is.null(facet_wrap_cols)) {
      # Use a grid
      p <- p + ggplot2::facet_grid(
        rows = facet_by_list$facet_rows, 
        cols = facet_by_list$facet_cols, 
        labeller = "label_context"
      )
      
    } else {
      p <- p + ggplot2::facet_wrap(
        facets = facet_by_list$facet_by, 
        labeller = "label_context"
      )
    }
  }

  # Update labels.
  p <- p + ggplot2::labs(
    x = x_label, 
    y = y_label, 
    title = plot_title, 
    subtitle = plot_sub_title, 
    caption = caption
  )

  # Rotate x-axis ticks
  if (rotate_x_tick_labels) {
    p <- p + ggplot2::theme(
      axis.text.x = ggplot2::element_text(
        vjust = 0.25, 
        hjust = 1.0, 
        angle = 90.0
      )
    )
  }

  return(p)
}



.determine_model_performance_plot_dimensions <- function(
    x,
    plot_type,
    x_axis_by,
    y_axis_by,
    facet_by,
    facet_wrap_cols,
    rotate_x_tick_labels
) {
  
  # Obtain facetting dimensions
  plot_dims <- .get_plot_layout_dims(
    x = x, 
    facet_by = facet_by, 
    facet_wrap_cols = facet_wrap_cols
  )

  # Determine the number of elements along the x-axis.
  x_elements <- as.character(unique(x[[x_axis_by]]))
  x_n_elements <- length(x_elements)
  x_longest_element <- max(sapply(x_elements, nchar))

  if (plot_type == "heatmap") {
    # For heatmaps.

    # Determine the number of elements along the y-axis.
    y_elements <- as.character(unique(x[[y_axis_by]]))
    y_n_elements <- length(y_elements)
    y_longest_element <- max(sapply(y_elements, nchar))

    # Assume each x-axis element takes up about 0.8 cm. Then add some room for
    # other plot elements.
    default_width <- x_n_elements * 0.8 + 1.0

    # Assume each y-axis element takes up about 0.8 cm as well.
    default_height <- y_n_elements * 0.8 + 1.0

    # Reserve space for y-axis and x-axis tick labels. Assume that the typical
    # width of a character is about 5 points (1.8 mm). For the x-axis we only
    # reserve extra space in case the ticks are rotated, otherwise we just
    # assume a typical height of 10 points (3.6 mm).
    y_tick_space <- y_longest_element * 0.18
    x_tick_space <- ifelse(rotate_x_tick_labels, x_longest_element * 0.18, 0.36)
  } else {
    # For non-heatmap plots.

    # Assume each x-axis element takes up about 0.8 cm. Then add some room for
    # other plot elements.
    default_width <- x_n_elements * 0.8 + 1.0
    default_width <- max(c(4.0, default_width))

    # Set default height.
    default_height <- 4.0

    # Set tick space for the x-axis and y-axis. Assume that the y-axis tick
    # labels contain 4 digits.
    y_tick_space <- 4.0 * 0.18

    # For the x-axis we only reserve extra space in case the ticks are rotated,
    # otherwise we just assume a typical height of 10 points (3.6 mm).
    x_tick_space <- ifelse(rotate_x_tick_labels, x_longest_element * 0.18, 0.36)
  }

  # Set overall plot height, but limit to small-margin A4 (27.7 cm)
  height <- min(c(2.0 + plot_dims[1L] * default_height + x_tick_space, 27.7))

  # Set overall plot width, but limit to small-margin A4 (19 cm)
  width <- min(c(2.0 + plot_dims[2L] * default_width + y_tick_space, 19.0))

  return(c(height, width))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarMetricRegression -----------------------------------------------------
setClass(
  "familiarMetricRegression",
  contains = "familiarMetric",
  slots = list("robust" = "character"),
  prototype = list("robust" = "none")
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarMetricRegression"),
  function(.Object, metric, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    if (endsWith(x = metric, suffix = "_trim")) {
      .Object@robust <- "trim"
      .Object@metric <- sub_last(
        x = metric,
        pattern = "_trim",
        replacement = "",
        fixed = TRUE
      )
      
    } else if (endsWith(x = metric, suffix = "_winsor")) {
      .Object@robust <- "winsor"
      .Object@metric <- sub_last(
        x = metric,
        pattern = "_winsor",
        replacement = "",
        fixed = TRUE
      )
      
    } else {
      # Default setting.
      .Object@robust <- "none"
      .Object@metric <- metric
    }

    return(.Object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMetricRegression"),
  function(object, ...) {
    return(object@outcome_type %in% c("continuous"))
  }
)



# mean absolute error ----------------------------------------------------------
setClass(
  "familiarMetricMAE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Mean Absolute Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (mean absolute error) ----------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricMAE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    # Compute the mean absolute error.
    score <- sum(abs(data$outcome - data$predicted_outcome)) / nrow(data)

    return(score)
  }
)


# relative absolute error ------------------------------------------------------
setClass(
  "familiarMetricRAE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Relative Absolute Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (relative absolute error) ------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricRAE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    y_mean <- mean(data$outcome)
    nom <- sum(abs(data$outcome - data$predicted_outcome))
    denom <- sum(abs(y_mean - data$outcome))

    # Check if denominator is not 0.
    if (!denom == 0.0) {
      # Denominator not 0
      score <- nom / denom
      
    } else if (nom == denom) {
      # Denominator and nominator are both 0.
      score <- 0.0
      
    } else {
      # Denominator = 0. Would be inf otherwise.
      return(callNextMethod())
    }

    return(score)
  }
)



# mean log absolute error ------------------------------------------------------
setClass(
  "familiarMetricMLAE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Mean Log Absolute Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (mean log absolute error) ------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricMLAE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    # Compute the mean log absolute error.
    score <- sum(log1p(abs(data$outcome - data$predicted_outcome))) / nrow(data)

    return(score)
  }
)



# mean squared error -----------------------------------------------------------
setClass(
  "familiarMetricMSE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Mean Squared Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (mean squared error) -----------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricMSE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    # Compute the mean squared error.
    score <- sum((data$outcome - data$predicted_outcome)^2.0) / nrow(data)

    return(score)
  }
)


# relative squared error -------------------------------------------------------
setClass(
  "familiarMetricRSE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Relative Squared Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (root relative squared error) --------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricRSE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    # Compute the mean squared error.
    y_mean <- mean(data$outcome)
    nom <- sum((data$outcome - data$predicted_outcome)^2.0)
    denom <- sum((y_mean - data$outcome)^2.0)

    # Check if denominator is not 0.
    if (!denom == 0.0) {
      # Denominator not 0
      score <- nom / denom
      
    } else if (nom == denom) {
      # Denominator and nominator are both 0.
      score <- 0.0
      
    } else {
      # Denominator = 0. Would be inf otherwise.
      return(callNextMethod())
    }

    return(score)
  }
)



# mean squared log error -------------------------------------------------------
setClass(
  "familiarMetricMSLE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Mean Squared Log Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (mean squared log error) -------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricMSLE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    # Compute the mean squared log error.
    score <- suppressWarnings(sum((log1p(data$outcome) - log1p(data$predicted_outcome))^2.0, na.rm = TRUE)) / nrow(data)

    if (!is.finite(score)) return(callNextMethod())

    return(score)
  }
)



# median absolute error --------------------------------------------------------
setClass(
  "familiarMetricMedianAE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Median Absolute Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (median absolute error) --------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricMedianAE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )
    
    if (is_empty(data)) return(callNextMethod())

    # Compute the median absolute error.
    score <- stats::median(x = abs(data$outcome - data$predicted_outcome))

    return(score)
  }
)



# root mean square error -------------------------------------------------------
setClass(
  "familiarMetricRMSE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Root Mean Square Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (root mean square error) -------------------------------
setMethod(
  "compute_metric_score", 
  signature(metric = "familiarMetricRMSE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )
    
    if (is_empty(data)) return(callNextMethod())

    # Compute the root mean square error.
    score <- sqrt(sum((data$outcome - data$predicted_outcome)^2.0) / nrow(data))

    return(score)
  }
)



# root relative squared error --------------------------------------------------
setClass(
  "familiarMetricRRSE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Root Relative Squared Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (root relative squared error) --------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricRRSE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    # Compute the mean squared error.
    y_mean <- mean(data$outcome)
    nom <- sum((data$outcome - data$predicted_outcome)^2.0)
    denom <- sum((y_mean - data$outcome)^2.0)

    # Check if denominator is not 0.
    if (!denom == 0.0) {
      # Denominator not 0
      score <- sqrt(nom / denom)
      
    } else if (nom == denom) {
      # Denominator and nominator are both 0.
      score <- 0.0
      
    } else {
      # Denominator = 0. Would be inf otherwise.
      return(callNextMethod())
    }

    return(score)
  }
)



# root mean square log error ---------------------------------------------------
setClass(
  "familiarMetricRMSLE",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Root Mean Square Log Error",
    value_range = c(0.0, Inf),
    higher_better = FALSE
  )
)



## compute_metric_score (root mean square log error) ---------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricRMSLE"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    # Compute the root mean square log error.
    score <- sqrt(sum((log1p(data$outcome) - log1p(data$predicted_outcome))^2.0) / nrow(data))

    if (!is.finite(score)) return(callNextMethod())

    return(score)
  }
)



# r-squared --------------------------------------------------------------------
setClass(
  "familiarMetricR2",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "R squared",
    value_range = c(-Inf, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (r-squared) --------------------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricR2"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    y_mean <- mean(data$outcome)
    nom <- sum((data$outcome - data$predicted_outcome)^2.0)
    denom <- sum((data$outcome - y_mean)^2.0)

    # Check if denominator is not 0.
    if (!denom == 0.0) {
      # Denominator not 0
      score <- 1.0 - nom / denom
      
    } else if (nom == denom) {
      # Denominator and nominator are both 0.
      score <- 1.0
      
    } else {
      # Denominator = 0. Would be -inf otherwise.
      return(callNextMethod())
    }

    return(score)
  }
)



# explained variance -----------------------------------------------------------
setClass(
  "familiarMetricExplainedVariance",
  contains = "familiarMetricRegression",
  prototype = methods::prototype(
    name = "Explained Variance",
    value_range = c(-Inf, 1.0),
    higher_better = TRUE
  )
)



## compute_metric_score (explained variance) -----------------------------------
setMethod(
  "compute_metric_score",
  signature(metric = "familiarMetricExplainedVariance"),
  function(metric, data, ...) {
    # Prepare data for computing metric values.
    data <- ..process_data_for_regression_metrics(
      metric = metric,
      data = data
    )

    if (is_empty(data)) return(callNextMethod())

    # Explained variance
    y_err_mean <- mean(data$outcome - data$predicted_outcome)
    y_mean <- mean(data$outcome)
    nom <- sum((data$outcome - data$predicted_outcome - y_err_mean)^2.0)
    denom <- sum((data$outcome - y_mean)^2.0)

    # Check if denominator is not 0.
    if (!denom == 0.0) {
      # Denominator not 0
      score <- 1.0 - nom / denom
      
    } else if (nom == denom) {
      # Denominator and nominator are both 0.
      score <- 1.0
      
    } else {
      # Denominator = 0. Would be -inf otherwise.
      return(callNextMethod())
    }

    return(score)
  }
)



..process_data_for_regression_metrics <- function(
    metric,
    data,
    fraction = 0.05) {
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- predicted_outcome <- NULL

  if (is_empty(data)) return(NULL)
  
  if (!is(data, "predictionTableRegression")) {
    ..error_data_not_prediction_table(data, "predictionTableRegression")
  }
  
  # Remove any entries that lack valid predictions.
  data <- remove_invalid_predictions(data)
  
  # Remove any entries that lack observed values.
  data <- filter_missing_outcome(data)
  if (is_empty(data)) return(NULL)
  
  data <- .as_data_table(data)
  
  if (metric@robust == "trim") {
    # Compute absolute prediction error
    abs_error <- abs(data$outcome - data$predicted_outcome)

    # Determine threshold for truncating the data.
    threshold <- stats::quantile(
      x = abs_error,
      probs = 1.0 - fraction,
      na.rm = TRUE
    )

    # Set mask to identify instances that should be kept.
    mask <- abs_error <= threshold[1L]

    # Filter data to only include instances within the mask.
    data <- data[mask, ]
    
  } else if (metric@robust == "winsor") {
    # Compute absolute prediction error
    abs_error <- abs(data$outcome - data$predicted_outcome)

    # Determine threshold for truncating the data.
    threshold <- stats::quantile(
      x = abs_error,
      probs = 1.0 - fraction,
      na.rm = TRUE
    )

    # Set mask to identify instances that should be updated.
    mask <- !(abs_error <= threshold[1L])

    # Update data by winsorising outlier data.
    data[mask & outcome > predicted_outcome, "predicted_outcome" := outcome - threshold]
    data[mask & outcome < predicted_outcome, "predicted_outcome" := outcome + threshold]
  }

  if (is_empty(data)) return(NULL)
  
  return(data)
}



.get_available_regression_metrics <- function() {
  return(c(
    .get_available_mae_metrics(),
    .get_available_mlae_metrics(),
    .get_available_mse_metrics(),
    .get_available_msle_metrics(),
    .get_available_medea_metrics(),
    .get_available_rmse_metrics(),
    .get_available_rmsle_metrics(),
    .get_available_r_squared_metrics(),
    .get_available_explained_variance_metrics(),
    .get_available_rse_metrics(),
    .get_available_rrse_metrics(),
    .get_available_rae_metrics()
  ))
}

.get_available_mae_metrics <- function() {
  return(paste_c(c("mae", "mean_absolute_error"), c("", "_trim", "_winsor")))
}

.get_available_rae_metrics <- function() {
  return(paste_c(c("rae", "relative_absolutive_error"), c("", "_trim", "_winsor")))
}

.get_available_mlae_metrics <- function() {
  return(paste_c(c("mlae", "mean_log_absolute_error"), c("", "_trim", "_winsor")))
}

.get_available_mse_metrics <- function() {
  return(paste_c(c("mse", "mean_squared_error"), c("", "_trim", "_winsor")))
}

.get_available_rse_metrics <- function() {
  return(paste_c(c("rse", "relative_squared_error"), c("", "_trim", "_winsor")))
}

.get_available_msle_metrics <- function() {
  return(paste_c(c("msle", "mean_squared_log_error"), c("", "_trim", "_winsor")))
}

.get_available_medea_metrics <- function() {
  return(paste_c(c("medae", "median_absolute_error"), c("", "_trim", "_winsor")))
}

.get_available_rmse_metrics <- function() {
  return(paste_c(c("rmse", "root_mean_square_error"), c("", "_trim", "_winsor")))
}

.get_available_rrse_metrics <- function() {
  return(paste_c(c("rrse", "root_relative_squared_error"), c("", "_trim", "_winsor")))
}

.get_available_rmsle_metrics <- function() {
  return(paste_c(c("rmsle", "root_mean_square_log_error"), c("", "_trim", "_winsor")))
}

.get_available_r_squared_metrics <- function() {
  return(paste_c(c("r2_score", "r_squared"), c("", "_trim", "_winsor")))
}

.get_available_explained_variance_metrics <- function() {
  return(paste_c(c("explained_variance"), c("", "_trim", "_winsor")))
}

familiar:::test_all_metrics_available(
  metrics = familiar:::.get_available_regression_metrics()
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

# power.transform and other packages are required.
if (!rlang::is_installed("power.transform")) testthat::skip()

regr_metric_test <- function(
    metric,
    data_list, 
    expected_score, 
    expected_objective = NULL) {
  # For several metrics the objective score and the expected score are the same.
  if (is.null(expected_objective)) expected_objective <- expected_score

  # Create metric object
  metric_object <- familiar:::as_metric(
    metric = metric,
    outcome_type = "continuous"
  )

  # Set baseline-value explicitly.
  metric_object@baseline_value <- familiar:::compute_metric_score(
    metric = metric_object,
    data = data_list[["no_slope"]]
  )

  for (ii in seq_along(data_list)) {
    # Check that the metric is available
    testthat::expect_true(familiar:::is_available(metric_object))

    # Compute the metric value.
    score <- familiar:::compute_metric_score(
      metric = metric_object,
      data = data_list[[ii]]
    )

    # Compute the objective score.
    objective_score <- familiar:::compute_objective_score(
      metric = metric_object,
      data = data_list[[ii]]
    )

    # Test the values.
    testthat::expect_equal(score, expected_score[ii])
    testthat::expect_equal(objective_score, expected_objective[ii])
  }
}

good_data <- familiar::as_prediction_table(
  x = c(1, 2, 3, 4, 5),
  y = c(1, 2, 3, 4, 5),
  type = "regression"
)

bad_data <- familiar::as_prediction_table(
  x = c(5, 4, 3, 2, 1),
  y = c(1, 2, 3, 4, 5),
  type = "regression"
)

no_slope_data <- familiar::as_prediction_table(
  x = c(3, 3, 3, 3, 3),
  y = c(1, 2, 3, 4, 5),
  type = "regression"
)

bias_offset_data <- familiar::as_prediction_table(
  x = c(0, 1, 2, 3, 4),
  y = c(1, 2, 3, 4, 5),
  type = "regression"
)

data_list <- list(
  "good" = good_data,
  "bad" = bad_data,
  "no_slope" = no_slope_data,
  "bias_offset" = bias_offset_data
)


# Mean absolute error ----------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_mae_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("Mean absolute error is correct", {
  for (metric in familiar:::.get_available_mae_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(0.0, 12 / 5, 6 / 5, 1.0),
      expected_objective = c(1.0, -1.0, 0.0, 1 / 6)
    )
  }
})

# Relative absolute error ------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_rae_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE
)

testthat::test_that("Relative absolute error is correct", {
  for (metric in familiar:::.get_available_rae_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(0.0, 2.0, 1.0, 5 / 6),
      expected_objective = c(1.0, -1.0, 0.0, 1 / 6)
    )
  }
})

# Mean log absolute error ------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_mlae_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("Mean log absolute error is correct", {
  for (metric in familiar:::.get_available_mlae_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(
        0.0,
        2 / 5 * log(5) + 2 / 5 * log(3),
        2 / 5 * log(3) + 2 / 5 * log(2),
        log(2)
      ),
      expected_objective = c(
        1.0,
        1 - (2 / 5 * log(5) + 2 / 5 * log(3)) / (2 / 5 * log(3) + 2 / 5 * log(2)),
        0.0,
        1 - (log(2)) / (2 / 5 * log(3) + 2 / 5 * log(2))
      )
    )
  }
})


# Mean squared error -----------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_mse_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("Mean squared error is correct", {
  for (metric in familiar:::.get_available_mse_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(0.0, 8.0, 2.0, 1.0),
      expected_objective = c(1.0, -1.0, 0.0, 1 / 2)
    )
  }
})

# Relative squared error -------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_rse_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE
)

testthat::test_that("Relative squared error is correct", {
  for (metric in familiar:::.get_available_rse_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(0.0, 4.0, 1.0, 1 / 2),
      expected_objective = c(1.0, -1.0, 0.0, 1 / 2)
    )
  }
})

# Mean squared log error -------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_msle_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("Mean squared log error is correct", {
  expected_score <- c(
    0.0,
    (log(2 / 6)^2 + log(3 / 5)^2 + log(5 / 3)^2 + log(6 / 2)^2) / 5,
    (log(2 / 4)^2 + log(3 / 4)^2 + log(5 / 4)^2 + log(6 / 4)^2) / 5,
    (log(2 / 1)^2 + log(3 / 2)^2 + log(4 / 3)^2 + log(5 / 4)^2 + log(6 / 5)^2) / 5
  )
  expected_objective <- c(1.0, -1.0, 0.0, 1 - expected_score[4] / expected_score[3])

  for (metric in familiar:::.get_available_msle_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = expected_score,
      expected_objective = expected_objective
    )
  }
})

# Median absolute error --------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_medea_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("Median absolute error is correct", {
  for (metric in familiar:::.get_available_medea_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = "median_absolute_error",
      expected_score = c(0.0, 2.0, 1.0, 1.0),
      expected_objective = c(1.0, -1.0, 0.0, 0.0)
    )
  }
})

# Root mean square error -------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_rmse_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("Root mean square error is correct", {
  for (metric in familiar:::.get_available_rmse_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(0.0, 2 * sqrt(2), sqrt(2), 1.0),
      expected_objective = c(1.0, -1.0, 0.0, 1 - 1 / sqrt(2))
    )
  }
})

# Root relative squared error --------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_rrse_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE
)

testthat::test_that("Root relative squared error is correct", {
  for (metric in familiar:::.get_available_rrse_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(0.0, 2.0, 1.0, 1.0 / sqrt(2)),
      expected_objective = c(1.0, -1.0, 0.0, 1 - 1 / sqrt(2))
    )
  }
})

# Root mean square log error ---------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_rmsle_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("Root mean square log error is correct", {
  expected_score <- sqrt(c(
    0.0,
    (log(2 / 6)^2 + log(3 / 5)^2 + log(5 / 3)^2 + log(6 / 2)^2) / 5,
    (log(2 / 4)^2 + log(3 / 4)^2 + log(5 / 4)^2 + log(6 / 4)^2) / 5,
    (log(2 / 1)^2 + log(3 / 2)^2 + log(4 / 3)^2 + log(5 / 4)^2 + log(6 / 5)^2) / 5
  ))

  expected_objective <- c(
    1.0,
    1.0 - expected_score[2] / expected_score[3],
    0.0,
    1.0 - expected_score[4] / expected_score[3]
  )

  for (metric in familiar:::.get_available_rmsle_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = expected_score,
      expected_objective = expected_objective
    )
  }
})

# Explained variance -----------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_explained_variance_metrics(),
  not_available_single_sample = FALSE,
  not_available_all_samples_identical = FALSE
)

testthat::test_that("Explained variance is correct", {
  for (metric in familiar:::.get_available_explained_variance_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(1.0, -3.0, 0.0, 1.0),
      expected_objective = c(1.0, -1.0, 0.0, 1.0)
    )
  }
})

# R2 score ---------------------------------------------------------------------
familiar:::test_all_metrics(
  metrics = familiar:::.get_available_r_squared_metrics(),
  not_available_single_sample = TRUE,
  not_available_all_samples_identical = TRUE
)

testthat::test_that("R2 score is correct", {
  for (metric in familiar:::.get_available_r_squared_metrics()) {
    regr_metric_test(
      data_list = data_list,
      metric = metric,
      expected_score = c(1.0, -3.0, 0.0, 0.5),
      expected_objective = c(1.0, -1.0, 0.0, 0.5)
    )
  }
})

#' Internal function for creating or retrieving iteration data
#'
#' @param file_paths Path to directory where iteration files are stored.
#' @param data Data set as loaded using the `.load_data` function.
#' @param experiment_setup data.table with subsampler information at different
#'   levels of the experimental design.
#' @param settings List of parameter settings. Some of these parameters are
#'   relevant to creating iterations.
#' @param message_indent Indenting of messages.
#' @param verbose Sets verbosity.
#'
#' @return A list with the following elements:
#'
#' * `iter_list`: A list containing iteration data at the different levels of
#'   the experiment.
#'
#' * `project_id`: The unique project identifier.
#'
#' * `experiment_setup`: data.table with subsampler information at different
#'   levels of the experimental design.
#'
#' @md
#' @keywords internal
.get_iteration_data <- function(
    file_paths,
    data,
    experiment_setup,
    settings,
    message_indent = 0L,
    verbose = TRUE
) {
  # Get project iterations and project id

  if (.experimental_design_is_file(
    file_dir = file_paths$iterations_dir,
    experimental_design = settings$data$exp_design
  )) {
    # Load list from user-provided file.
    iteration_list <- .load_iterations(
      file_dir = file_paths$iterations_dir,
      iteration_file = settings$data$exp_design
    )

    # Extract iterations, project id and experiment setup from the file.
    new_iteration_list <- iteration_list$iteration_list
    project_id <- iteration_list$project_id
    experiment_setup <- iteration_list$experiment_setup
    
  } else {
    # Check whether iter_list already exists as a file
    iteration_list <- .load_iterations(file_dir = file_paths$iterations_dir)

    if (iteration_list$iteration_file_exists) {
      # Update iterations to check for new external validation data
      new_iteration_list <- .create_iterations(
        data = data,
        experiment_setup = experiment_setup,
        settings = settings,
        override_external_validation = TRUE,
        iteration_list = iteration_list$iteration_list,
        message_indent = message_indent,
        verbose = verbose
      )

      # Save to file and without generating new project id
      project_id <- .save_iterations(
        file_paths = file_paths,
        iteration_list = new_iteration_list,
        project_id = iteration_list$project_id,
        experiment_setup = experiment_setup,
        message_indent = message_indent,
        verbose = verbose
      )
      
    } else {
      # Create training and validation data iterations
      new_iteration_list <- .create_iterations(
        data = data,
        experiment_setup = experiment_setup,
        settings = settings,
        message_indent = message_indent,
        verbose = verbose
      )

      # Save to file and generate project id
      project_id <- .save_iterations(
        file_paths = file_paths,
        iteration_list = new_iteration_list,
        experiment_setup = experiment_setup,
        message_indent = message_indent,
        verbose = verbose
      )
    }
  }

  return(list(
    "iter_list" = new_iteration_list,
    "project_id" = project_id,
    "experiment_setup" = experiment_setup
  ))
}



.get_iteration_file_name <- function(
    file_paths,
    project_id
) {
  
  return(file.path(
    file_paths$iterations_dir,
    paste0(project_id, "_iterations.RDS")
  ))
}



#' Internal function for loading iteration data from the file system
#'
#' Loads iterations generated by `.create_iterations` that were created in a
#' previous session. If these are not available, this is indicated by setting a
#' return flag.
#'
#' @param file_dir Path to directory where iteration files are stored.
#'
#' @return List containing:
#'
#'  * `iteration_file_exists`: An indicator whether an iteration file was found.
#'
#'  * `iteration_list`: The list of iterations (if available).
#'
#'  * `project_id`: The unique project identifier (if available).
#'
#' @md
#' @keywords internal
.load_iterations <- function(file_dir, iteration_file = NULL) {
  # Loads data iterations

  # Suppress NOTES due to non-standard evaluation in data.table
  file_name <- file_prefix_num <- NULL

  if (is.null(iteration_file)) {
    # Find files in directory
    iter_files <- list.files(
      path = file_dir,
      pattern = "_iterations.RDS",
      full.names = FALSE
    )

    # If no files match the iteration files, return a FALSE
    if (length(iter_files) == 0L) {
      return(list("iteration_file_exists" = FALSE))
    }

    # Extract date strings from file name
    file_table <- data.table::data.table("file_name" = iter_files)
    file_table[, "file_prefix_num" := strsplit(x = file_name, split = "_", fixed = TRUE)[[1L]][1L], by = file_name]
    file_table <- file_table[!is.na(file_prefix_num), ]

    # If no date strings were found (i.e. backup files), return a FALSE
    if (nrow(file_table) == 0L) {
      return(list("iteration_file_exists" = FALSE))
    }

    # Select most recent file
    file_table[, "file_prefix_num" := as.numeric(file_prefix_num)]

    # From list of iteration files present in the data.
    select_file <- file_table[file_prefix_num == max(file_prefix_num), ]$file_name[1L]
    project_id <- file_table[file_prefix_num == max(file_prefix_num), ]$file_prefix_num[1L]

    # Ask whether a new file should be used
    input_answer <- "dummy"
    while (!tolower(input_answer) %in% c("y", "n")) {
      input_answer <- readline(prompt = paste0(
        "Latest iteration file found is: ", select_file, 
        ". Do you wish to create a new iterations file [y/n]?: "
      ))
      
      if (!tolower(input_answer) %in% c("y", "n")) {
        message("Only y (yes) or n (no) are valid entries. Please retry.")
      }
    }

    # If the answer is yes, create a new iterations while
    if (input_answer == "y") {
      return(list("iteration_file_exists" = FALSE))
      
    } else {
      # If the answer is no: load file
      iteration_list <- readRDS(file = file.path(file_dir, select_file))

      # For backward compatibility: iteration_list only contains an
      # iteration_list, but the latter is not nested.
      if (is.null(iteration_list$iteration_list)) {
        return(list(
          "iteration_file_exists" = TRUE,
          "iteration_list" = iteration_list,
          "project_id" = project_id
        ))
        
      } else {
        return(list(
          "iteration_file_exists" = TRUE,
          "iteration_list" = iteration_list$iteration_list,
          "project_id" = project_id
        ))
      }
    }
  } else {
    # From user-provided iteration file
    select_file <- basename(iteration_file)
    project_id <- as.numeric(strsplit(x = select_file, split = "_", fixed = TRUE)[[1L]][1L])

    # Attempt to load the user-provided iteration file.
    if (file.exists(iteration_file)) {
      # From an absolute path.
      iteration_list <- readRDS(file = iteration_file)
      
    } else if (file.exists(file.path(file_dir, select_file))) {
      # From a path relative to the current (experiment/iteration) directory.
      iteration_list <- readRDS(file = file.path(file_dir, select_file))
      
    } else {
      ..error(paste0("External iteration file could not be found: ", iteration_file))
    }

    if (is.null(iteration_list$experiment_setup)) {
      ..error(paste0(
        "The provided external iteration file lacks an experiment setup table ",
        "and cannot be used. The most likely cause is that the iteration file ",
        "was generated with familiar version < 0.0.0.41."
      ))
    }

    return(list(
      "iteration_file_exists" = TRUE,
      "iteration_list" = iteration_list$iteration_list,
      "project_id" = project_id,
      "experiment_setup" = iteration_list$experiment_setup
    ))
  }
}


.create_iterations <- function(
    data,
    experiment_setup,
    settings,
    override_external_validation = FALSE,
    iteration_list = NULL,
    message_indent = 0L,
    verbose = TRUE
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  main_data_id <- batch_id <- NULL

  # Get unique main data ids
  main_data_ids <- unique(experiment_setup$main_data_id)

  # Get names of the standard id columns
  id_columns <- get_id_columns(id_depth = "series")
  batch_id_column <- get_id_columns(id_depth = "batch")

  # Start random stream
  if (is.null(settings$data$iteration_seed)) {
    rstream_object <- NULL
  } else {
    rstream_object <- .start_random_number_stream(seed = settings$data$iteration_seed)
  }
  
  # Generate new iterations ----------------------------------------------------

  if (is.null(iteration_list)) {
    iteration_list <- list()

    logger_message(
      "Creating iterations: Starting creation of iterations.",
      indent = message_indent,
      verbose = verbose
    )

    while (length(iteration_list) < length(main_data_ids)) {
      for (curr_main_data_id in main_data_ids) {
        # Set up the subset_table
        subset_table <- experiment_setup[main_data_id == curr_main_data_id, ]

        # Check if the current data has already been processed
        if (!is.null(iteration_list[[as.character(curr_main_data_id)]])) next

        # Check if the required reference data exists
        curr_ref_data_id <- subset_table$ref_data_id[1L]
        curr_pert_method <- subset_table$perturb_method[1L]
        if (
          is.null(iteration_list[[as.character(curr_ref_data_id)]]) &&
          curr_pert_method != "main"
        ) {
          next
        }

        # Determine number of reference iterations
        ref_run_list <- .get_run_list(
          iteration_list = iteration_list,
          data_id = curr_ref_data_id
        )

        # Create an empty run list
        run_list <- list()

        # Check if the current perturbation method is "main"
        if (curr_pert_method == "main") {
          
          ## New main data -----------------------------------------------------

          # Determine if external validation is required
          external_validation_required <- subset_table$external_validation[1L]

          # In case of external validation, find cohorts matching the
          # train_cohorts.
          all_cohorts <- unique(data[[batch_id_column]])

          # Check whether the requested train cohorts are provided
          if (is.null(settings$data$train_cohorts) && external_validation_required) {
            logger_stop(paste0(
              "Creating iterations: Training cohorts were not provided, ",
              "but are required for external validation."
            ))
            
          } else if (is.null(settings$data$train_cohorts)) {
            # Assume that all data is used for training when external validation
            # is not required, but no specific cohorts are specified.
            train_cohorts <- unique(data[[batch_id_column]])
            
          } else {
            # This is the ideal situation: training cohort names are provided.
            train_cohorts <- settings$data$train_cohorts

            # Check if all training cohorts actually exist.
            if (!all(train_cohorts %in% all_cohorts)) {
              ..warning_missing_cohorts(x = train_cohorts[!train_cohorts %in% all_cohorts])
            }
          }

          # Handle validation cohorts
          if (external_validation_required == TRUE) {
            # Check whether validation cohorts are provided.
            if (!is.null(settings$data$valid_cohorts)) {
              # Extract validation cohorts from settings.
              validation_cohorts <- settings$data$valid_cohorts

              # Check if all validation cohorts actually exist.
              if (!all(validation_cohorts %in% all_cohorts)) {
                ..warning_missing_cohorts(x = validation_cohorts[!validation_cohorts %in% all_cohorts])
              }
              
            } else {
              # Extract validation cohorts from data
              validation_cohorts <- all_cohorts[!all_cohorts %in% train_cohorts]
              if (length(validation_cohorts) == 0L) {
                logger_warning("Creating iterations: No validation cohorts could be found.")
              }
            }
            
          } else {
            # Don't use validation cohorts if only training data is available
            validation_cohorts <- character()
          }

          # Extract lists of training and validation subjects
          train_samples <- list()
          valid_samples <- list()

          # Add training cohort subjects to train_samples
          train_samples[[1L]] <- unique(data[batch_id %in% train_cohorts, mget(id_columns)])

          # Add validation cohort subjects to valid_samples (if present)
          if (length(validation_cohorts) > 0L) {
            valid_samples[[1L]] <- unique(data[batch_id %in% validation_cohorts, mget(id_columns)])
          }

          # Generate a run list for the "main" data perturbation
          run_list <- .create_iteration_run_list(
            data_id = curr_main_data_id,
            train_samples = train_samples,
            valid_samples = valid_samples,
            can_pre_process = TRUE,
            perturbation = curr_pert_method
          )

          # Clean variables
          rm(
            external_validation_required, train_cohorts, train_samples,
            validation_cohorts, valid_samples
          )
        }

        # Check if the current perturbation method is "imbalance part"
        if (curr_pert_method == "imbalance_partition") {
          ## New imbalance partitions ------------------------------------------

          # Iterate over runs of the reference data
          for (run in ref_run_list) {
            # Generate partitions for generating balanced data sets. Note that
            # these partitions do not generate any validation data.
            partition <- .create_balanced_partitions(
              sample_identifiers = .get_sample_identifiers(
                run = run,
                train_or_validate = "train"
              ),
              settings = settings,
              data = data,
              rstream_object = rstream_object
            )

            # Append runs to the run list
            run_list <- c(
              run_list,
              .add_iteration_to_run(run,
                data_id = curr_main_data_id,
                run_id_offset = length(run_list),
                train_samples = partition,
                valid_samples = list(),
                can_pre_process = TRUE,
                perturbation = curr_pert_method
              )
            )
          }
          
          # Clean variables
          rm(partition, run)
        }

        # Check if the current perturbation method is "limited_bootstrap" or
        # "full_bootstrap".
        if (curr_pert_method %in% c("limited_bootstrap", "full_bootstrap")) {
          ## New bootstrap data ------------------------------------------------

          # Determine number of bootstrap iterations
          n_iter <- subset_table$perturb_n_rep[1L]

          # Iterate over runs of the reference data
          for (run in ref_run_list) {
            # Create bootstraps. Bootstraps contain both training and validation
            # data. However, they are not separately pre-processed.
            bt_iter_list <- .create_bootstraps(
              sample_identifiers = .get_sample_identifiers(
                run = run,
                train_or_validate = "train"
              ),
              n_iter = n_iter,
              settings = settings,
              data = data,
              rstream_object = rstream_object
            )
            
            # Append runs to the run list
            run_list <- c(
              run_list,
              .add_iteration_to_run(run,
                data_id = curr_main_data_id,
                run_id_offset = length(run_list),
                train_samples = bt_iter_list$train_list,
                valid_samples = bt_iter_list$valid_list,
                can_pre_process = curr_pert_method == "full_bootstrap",
                perturbation = curr_pert_method
              )
            )
          }

          # Clean variables
          rm(bt_iter_list, run, n_iter)
        }

        # Check if the current perturbation method is "cross_val"
        if (curr_pert_method == "cross_val") {
          ## New cross-validation data -----------------------------------------

          # Determine number of cross-validation repetitions and folds
          n_rep <- subset_table$perturb_n_rep[1L]
          n_folds <- subset_table$perturb_n_folds[1L]

          # Iterate over runs of the reference data
          for (run in ref_run_list) {
            # Create repeated cross-validations
            cv_iter_list <- .create_repeated_cv(
              sample_identifiers = .get_sample_identifiers(
                run = run,
                train_or_validate = "train"
              ),
              n_rep = n_rep,
              n_folds = n_folds,
              settings = settings,
              data = data,
              rstream_object = rstream_object
            )

            # Append runs to the run list
            run_list <- c(
              run_list,
              .add_iteration_to_run(run,
                data_id = curr_main_data_id,
                run_id_offset = length(run_list),
                train_samples = cv_iter_list$train_list,
                valid_samples = cv_iter_list$valid_list,
                can_pre_process = TRUE,
                perturbation = curr_pert_method
              )
            )
          }
          
          # Clean variables
          rm(cv_iter_list, run, n_rep, n_folds)
        }

        # Check if the current perturbation method is "loocv"
        if (curr_pert_method == "loocv") {
          ## New leave-one-out cross-validation data ---------------------------

          # Iterate over runs of the reference data
          for (run in ref_run_list) {
            # Find the available samples.
            sample_identifiers <- .get_sample_identifiers(
              run = run,
              train_or_validate = "train"
            )

            # Create leave-one-out cross-validation.
            cv_iter_list <- .create_loocv(
              sample_identifiers = sample_identifiers,
              settings = settings,
              data = data,
              rstream_object = rstream_object
            )

            # Append runs to the run list
            run_list <- c(
              run_list,
              .add_iteration_to_run(run,
                data_id = curr_main_data_id,
                run_id_offset = length(run_list),
                train_samples = cv_iter_list$train_list,
                valid_samples = cv_iter_list$valid_list,
                can_pre_process = TRUE,
                perturbation = curr_pert_method
              )
            )
          }

          # Clean variables
          rm(cv_iter_list, sample_identifiers)
        }

        # Add to iteration list
        iteration_list[[as.character(curr_main_data_id)]] <- list(
          "run" = run_list,
          "main_data_id" = curr_main_data_id
        )

        # Clean variables
        rm(run_list, ref_run_list, curr_main_data_id)
      }
    }

    logger_message(
      "Creating iterations: Finished creation of iterations.",
      indent = message_indent,
      verbose = verbose
    )
    
  } else if (override_external_validation) {
    if (is.null(iteration_list)) stop("iteration_list should be provided.")
    # Update external validation iterations ------------------------------------

    # Update external validation
    # Determine if external validation is required and skip otherwise
    external_validation_required <- experiment_setup[main_data_id == 1L, ]$external_validation[1L]
    if (external_validation_required) {
      # In case of external validation, find cohorts matching the train_cohorts
      all_cohorts <- unique(data[[batch_id_column]])

      # Check whether validation cohorts are provided
      if (!is.null(settings$data$valid_cohorts)) {
        # Extract validation cohorts from settings
        validation_cohorts <- settings$data$valid_cohorts
        if (!all(validation_cohorts %in% all_cohorts)) {
          ..warning_missing_cohorts(
            x = validation_cohorts[!validation_cohorts %in% all_cohorts]
          )
        }
        
      } else {
        # Extract validation cohorts from data by selecting all non-training
        # cohorts
        if (is.null(settings$data$train_cohorts)) {
          logger_stop("Creating iterations: Training cohorts were not provided.")
        }

        # Extract train cohorts from settings
        train_cohorts <- settings$data$train_cohorts
        if (!all(train_cohorts %in% all_cohorts)) {
          ..warning_missing_cohorts(x = train_cohorts[!train_cohorts %in% all_cohorts])
        }

        # Extract validation cohorts from data
        validation_cohorts <- all_cohorts[!all_cohorts %in% train_cohorts]
        if (length(validation_cohorts) == 0L) {
          logger_warning("Creating iterations: No validation cohorts could be found.")
        }
      }

      # Update validation subjects to iter list
      iteration_list[[as.character(1L)]]$run[[as.character(1L)]]$valid_samples <- unique(
        data[batch_id %in% validation_cohorts, mget(id_columns)]
      )
      
    } else {
      # Remove external validation subject
      iteration_list[[as.character(1L)]]$run[[as.character(1L)]]$valid_samples <- NULL
    }
  }

  return(iteration_list)
}



.save_iterations <- function(
    file_paths,
    iteration_list,
    project_id = NULL,
    experiment_setup,
    message_indent = 0L,
    verbose = TRUE
) {
  # Check if an existing project identifier is provided, otherwise generate a
  # new one.
  if (is.null(project_id)) {
    # Generate project id from system time
    project_id <- as.numeric(format(Sys.time(), "%Y%m%d%H%M%S"))

    # Message project identifier
    logger_message(
      paste0("Creating iterations: New project id is: \'", project_id, "\'."),
      indent = message_indent,
      verbose = verbose
    )
  }

  # Set file name
  file_name <- .get_iteration_file_name(
    file_paths = file_paths,
    project_id = project_id
  )

  # Attach both iteration list and experiment setup.
  save_iteration_list <- list(
    "iteration_list" = iteration_list,
    "experiment_setup" = experiment_setup
  )

  # Save both files to the folder
  saveRDS(save_iteration_list, file = file_name)

  return(project_id)
}



.create_iteration_run_list <- function(
    data_id,
    train_samples,
    valid_samples,
    can_pre_process,
    perturbation
) {
  # Start an empty run list
  run_list <- list()
  
  for (ii in seq_len(length(train_samples))) {
    # Create new run table
    run_table <- data.table::data.table(
      "data_id" = data_id,
      "run_id" = ii,
      "can_pre_process" = can_pre_process,
      "perturbation" = perturbation,
      "perturb_level" = 1L
    )

    # Add run_table to the list.
    run_list[[as.character(ii)]]$run_table <- run_table

    # Add samples for model development to the list.
    run_list[[as.character(ii)]]$train_samples <- train_samples[[ii]]

    # Add validation samples to the list.
    if (length(valid_samples) >= ii) {
      run_list[[as.character(ii)]]$valid_samples <- valid_samples[[ii]]
    } else {
      run_list[[as.character(ii)]]$valid_samples <- NULL
    }
  }

  return(run_list)
}



.add_iteration_to_run <- function(
    run,
    data_id = NULL,
    run_id_offset = 0L,
    train_samples, 
    valid_samples,
    can_pre_process,
    perturbation
) {
  # Determine the number of runs
  n_iter <- length(train_samples)

  # Start an empty run list
  run_list <- list()

  # If data_id is not provided, the run_list will be treated a custom run, and
  # will receive a data_id of -1
  if (is.null(data_id)) data_id <- -1L
  
  # Find the current perturbation level based on the input run
  curr_perturb_level <- max(run$run_table$perturb_level) + 1L

  for (ii in seq_len(n_iter)) {
    # Update run_id
    run_id <- run_id_offset + ii

    # Add run to existing run table.
    run_table <- rbind(
      run$run_table,
      data.table::data.table(
        "data_id" = data_id,
        "run_id" = run_id,
        "can_pre_process" = can_pre_process,
        "perturbation" = perturbation,
        "perturb_level" = curr_perturb_level
      )
    )
    
    # Add run_table to the list.
    run_list[[as.character(run_id)]]$run_table <- run_table

    # Add samples for model development to the list.
    run_list[[as.character(run_id)]]$train_samples <- train_samples[[ii]]

    # Add validation samples to the list.
    if (length(valid_samples) >= ii) {
      run_list[[as.character(run_id)]]$valid_samples <- valid_samples[[ii]]
    } else {
      run_list[[as.character(run_id)]]$valid_samples <- NULL
    }
  }

  return(run_list)
}



.create_repeated_cv <- function(
    data,
    sample_identifiers = NULL,
    n_rep,
    n_folds,
    settings = NULL,
    outcome_type = NULL,
    stratify = TRUE,
    rstream_object = NULL
) {
  # This function wraps the .create_cv function

  # Initiate lists for training and validation data
  train_list <- list()
  valid_list <- list()

  # Iterate over iterations
  for (ii in 1L:n_rep) {
    # Create cross-validation subsets.
    cv_iter_list <- .create_cv(
      sample_identifiers = sample_identifiers,
      n_folds = n_folds,
      settings = settings,
      outcome_type = outcome_type,
      data = data,
      stratify = stratify,
      rstream_object = rstream_object
    )

    # Add new subsets to the list.
    train_list <- c(train_list, cv_iter_list$train_list)
    valid_list <- c(valid_list, cv_iter_list$valid_list)
  }

  return(list(
    "train_list" = train_list,
    "valid_list" = valid_list
  ))
}


.create_loocv <- function(
    data,
    sample_identifiers = NULL,
    settings = NULL,
    outcome_type = NULL,
    rstream_object = NULL
) {
  # Determine the number of samples.
  if (is.null(sample_identifiers)) {
    # Obtain id columns
    id_columns <- get_id_columns(id_depth = "series")

    # Identify unique sample identifiers.
    sample_identifiers <- unique(data[, mget(id_columns)])
  }

  # Determine sample-id columns
  sample_id_columns <- get_id_columns(id_depth = "sample")

  # Determine the number of samples.
  n_samples <- data.table::uniqueN(sample_identifiers, by = sample_id_columns)

  # Create iterations.
  loocv_iter_list <- .create_cv(
    sample_identifiers = sample_identifiers,
    data = data,
    n_folds = n_samples,
    settings = settings,
    outcome_type = outcome_type,
    stratify = FALSE,
    rstream_object = rstream_object
  )

  return(loocv_iter_list)
}



.create_cv <- function(
    data,
    n_folds,
    sample_identifiers = NULL,
    settings = NULL,
    outcome_type = NULL,
    stratify = TRUE,
    return_fold_id = FALSE,
    rstream_object = NULL
) {
  # Cross-validation

  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- fold_id <- sample_order_id <- NULL
  n <- i.n <- assign_to_training <- .NATURAL <- NULL

  # Obtain id columns
  id_columns <- get_id_columns(id_depth = "series")

  # Set outcome_type from settings
  if (is.null(outcome_type)) outcome_type <- settings$data$outcome_type

  # Check stratification for continuous data
  if (outcome_type %in% c("continuous")) stratify <- FALSE

  # Do not stratify absent data
  if (is_empty(data)) stratify <- FALSE

  # Set sample identifiers, if not provided. Note that is only the case during
  # unit testing.
  if (is.null(sample_identifiers)) {
    sample_identifiers <- unique(data[, mget(id_columns)])
  }

  if (outcome_type %in% c("continuous")) {
    # Select data based on sample id - note that even if duplicate
    # sample_identifiers exist, only unique sample_identifiers are maintained -
    # this is intentional.
    subset_table <- merge(
      x = unique(data[, mget(id_columns)]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )
    
  } else if (outcome_type == "survival") {
    # For stratifying survival data we require the event status.
    # Event status (including NA) are transcoded.
    subset_table <- merge(
      x = unique(data[, mget(c(id_columns, "outcome_event"))]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )

    data.table::setnames(
      x = subset_table,
      old = "outcome_event",
      new = "outcome"
    )

    # Transcode event status
    subset_table[, "outcome" := factor(outcome)]
    subset_table$outcome <- addNA(subset_table$outcome, ifany = TRUE)
    subset_table$outcome <- as.numeric(subset_table$outcome)
    
  } else if (outcome_type %in% c("binomial", "multinomial")) {
    # For stratifying categorical data we require the outcome data as is.
    # Redundant factors are dropped and remaining factors (including NA) are
    # transcoded
    subset_table <- merge(
      x = unique(data[, mget(c(id_columns, "outcome"))]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )

    # Transcode classes
    subset_table$outcome <- addNA(subset_table$outcome, ifany = TRUE)
    subset_table$outcome <- droplevels(subset_table$outcome)
    subset_table$outcome <- as.numeric(subset_table$outcome)
    
  } else if (outcome_type %in% c("competing_risk", "multi_label")) {
    ..error_no_known_outcome_type(outcome_type)
  }

  # Determine sample-id columns
  sample_id_columns <- get_id_columns(id_depth = "sample")

  # Determine the number of samples.
  n_samples <- data.table::uniqueN(subset_table, by = sample_id_columns)

  if (n_folds < 2L) {
    ..error("The number of cross-validation folds should be at least 2.")
  }

  # Check if the number of folds exceeds the number of data points
  if (n_folds > n_samples) {
    ..error(
      paste0(
        "The number of cross-validation folds (",
        n_folds,
        ") exceeds the number of available data points (",
        n_samples,
        ")."
      ),
      error_Class = "input_argument_error"
    )
    
  } else if (stratify) {
    # Determine levels in stratified data
    if (n_folds > n_samples - data.table::uniqueN(subset_table$outcome) + 1L) {
      ..warning("Cannot perform stratified cross-validation as the number of folds is too high.")
      
      stratify <- FALSE
    }
  }

  # Add fold id to each entry
  subset_table[, "fold_id" := 0L]

  # Determine fold size
  fold_size <- n_samples %/% n_folds

  # Initiate training and validation lists
  train_list <- valid_list <- list()

  # Make sure that any rare outcome levels are always assigned to the training
  # folds, not validation folds. This concerns all samples where an outcome
  # level appears so rarely that all instances of it could be assigned to a
  # single fold. In practice this means the training folds should be
  # pre-assigned at least one instance of every outcome level. This is done
  # below.
  #
  # Note that the number of formable folds may be affected.
  if (outcome_type %in% c("binomial", "multinomial", "survival", "multi_label")) {
    # Iterate, because by hard-assigning samples to the training set and
    # decreasing the fold size, we may actually encounter other outcome levels
    # that are at risk. We break from the loop if there are no such outcome
    # levels.
    while (TRUE) {
      # Determine the frequency samples with each outcome, leaving out any
      # pre-assigned samples (fold-id equal to -1).
      level_frequency <- subset_table[
        fold_id == 0L,
        list("n" = .N > 0L),
        by = c(id_columns, "outcome")
      ]
      level_frequency <- level_frequency[, list("n" = sum(n)), by = c("outcome")][order(n)]

      # Determine which outcome levels occur in the same number or fewer samples
      # than the fold size, and could therefore be entirely assigned to a single
      # (validation) fold.
      level_frequency[, "assign_to_training" := n <= fold_size]

      # Ensure that these outcomes have not been pre-assigned (fold-id equal to
      # -1)
      level_frequency[
        subset_table[fold_id == -1L],
        "assign_to_training" := FALSE, 
        on = .NATURAL
      ]

      # Break from loop if no samples need to be selected.
      if (all(level_frequency$assign_to_training == FALSE)) break

      # Select the first outcome at risk. This is the minority outcome class on
      # the first iteration.
      outcome_level_training <- level_frequency[assign_to_training == TRUE]$outcome[1L]

      # Randomly select one sample with the given outcome.
      sample_training <- fam_sample(
        subset_table[fold_id == 0L & outcome == outcome_level_training],
        size = 1L,
        rstream_object = rstream_object
      )

      # Pre-assign the selected sample.
      subset_table[
        sample_training,
        "fold_id" := -1L,
        on = .NATURAL
      ]

      # Update n_samples and n_folds
      n_samples <- n_samples - 1L

      # Check that if n_folds is now larger than n_samples (e.g. LOOCV), that
      # n_folds is equalised again so that there is always one validation
      # sample.
      if (n_samples < n_folds) n_folds <- n_samples

      # Update fold size.
      fold_size <- n_samples %/% n_folds
    }
  }

  if (!stratify) {
    # Iterate over the folds.
    for (ii in 1L:n_folds) {
      # Choose subject id for current fold
      current_train_id <- fam_sample(
        x = subset_table[fold_id == 0L],
        size = fold_size,
        replace = FALSE,
        rstream_object = rstream_object
      )

      # Join subset-table based on selected samples, and set the fold id.
      subset_table[current_train_id, "fold_id" := ii, on = .NATURAL]
    }

    # Update remaining samples by random assignment to a fold.
    unassigned_data <- unique(
      subset_table[fold_id == 0L, mget(sample_id_columns)],
      by = sample_id_columns
    )
    
    if (nrow(unassigned_data) > 0L) {
      # Randomly assign fold identifier.
      unassigned_data[, "fold_id" := sample.int(
        n_folds,
        size = nrow(unassigned_data),
        replace = FALSE
      )]

      # Merge back the series identifier back into unassigned_data, and keep the
      # updated fold identifier.
      unassigned_data <- merge(
        x = unassigned_data,
        y = subset_table[fold_id == 0L, mget(id_columns)],
        by = sample_id_columns,
        all = FALSE
      )

      # Add unassigned samples to the subset table.
      subset_table <- data.table::rbindlist(
        list(
          subset_table[fold_id != 0L, mget(c("fold_id", id_columns))],
          unassigned_data
        ),
        use.names = TRUE
      )
    }

    # Assign training and validation folds
    for (ii in 1L:n_folds) {
      train_id <- subset_table[fold_id != ii, mget(id_columns)]
      valid_id <- subset_table[fold_id == ii, mget(id_columns)]

      train_list[[ii]] <- train_id
      valid_list[[ii]] <- valid_id
    }
    
  } else {
    # Determine the frequency of sample outcomes per fold.
    level_frequency <- subset_table[
      fold_id == 0L,
      list("n" = .N %/% n_folds),
      by = "outcome"
    ]

    # Iterate over folds
    for (ii in seq_len(n_folds)) {
      # Copy the fold_level frequency.
      fold_level_frequency <- data.table::copy(level_frequency)

      # Get all available data.
      available_data <- subset_table[fold_id == 0L]

      # Check that any data are available.
      if (nrow(available_data) == 0L) next

      # Determine the frequency of outcome levels for each sample.
      available_data <- available_data[, list("n" = .N), by = c(sample_id_columns, "outcome")]

      # Order available data randomly.
      sample_order <- fam_sample(
        x = available_data,
        replace = FALSE,
        rstream_object = rstream_object
      )

      # Set sample_order_id. This will be used to order available_data.
      sample_order[, "sample_order_id" := .I]

      # Merge with available_data
      available_data <- merge(
        x = available_data,
        y = sample_order,
        by = sample_id_columns
      )[order(sample_order_id)]

      # Iterate over samples. For each sample we check whether it can be added
      # to the fold.
      for (current_sample in split(available_data, by = "sample_order_id")) {
        # Determine if the sample can be added.
        safe_to_add <- all(fold_level_frequency[
          current_sample,
          list("n" = n - i.n),
          on = "outcome"
        ]$n >= 0L)

        if (!safe_to_add) next

        # If the check passes, we need to add the sample to the fold, and update
        # level_frequency_fold.
        fold_level_frequency[current_sample, "n" := n - i.n, on = "outcome"]

        # Add to fold.
        current_sample <- unique(current_sample[, mget(sample_id_columns)])
        subset_table[current_sample, "fold_id" := ii, on = .NATURAL]

        # Skip further samples if no more samples need to be added to the fold.
        if (all(fold_level_frequency$n == 0L)) break
      }
    }

    # Determine if all data was assigned.
    unassigned_data <- unique(
      subset_table[fold_id == 0L, mget(sample_id_columns)],
      by = sample_id_columns
    )
    
    if (nrow(unassigned_data) > 0L) {
      # Randomly order unassigned data.
      unassigned_data <- fam_sample(unassigned_data,
        replace = FALSE,
        rstream_object = rstream_object
      )

      # Assign fold id.
      unassigned_data[, "fold_id" := rep_len(
        x = seq_len(n_folds),
        length.out = nrow(unassigned_data)
      )]

      # Merge back the series identifier back into unassigned_data, and keep the
      # updated fold identifier.
      unassigned_data <- merge(
        x = unassigned_data,
        y = subset_table[fold_id == 0L, mget(id_columns)],
        by = sample_id_columns,
        all = FALSE
      )

      # Add unassigned samples to the subset table.
      subset_table <- data.table::rbindlist(
        list(
          subset_table[fold_id != 0L, mget(c(id_columns, "fold_id"))],
          unassigned_data
        ),
        use.names = TRUE
      )
    }

    # Assign to training and validation sets. Note that any unassigned samples
    # are assigned to the training folds.
    for (ii in 1L:n_folds) {
      train_list[[ii]] <- subset_table[fold_id != ii, mget(id_columns)]
      valid_list[[ii]] <- subset_table[fold_id == ii, mget(id_columns)]
    }
  }

  if (return_fold_id) {
    # Assign any instances with fold_id == -1 to the first fold to prevent very
    # small folds from being formed.
    subset_table[fold_id == -1L, "fold_id" := 1L]

    return(subset_table[, mget(c(id_columns, "fold_id"))])
    
  } else {
    return(list(
      "train_list" = train_list,
      "valid_list" = valid_list
    ))
  }
}



.create_bootstraps <- function(
    data,
    n_iter,
    sample_identifiers = NULL,
    settings = NULL,
    outcome_type = NULL,
    stratify = TRUE,
    rstream_object = NULL
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- outcome_present <- .NATURAL <- NULL
  keep <- sample_order_id <- cumulative_n <- i.n <- n <- NULL

  # Obtain id columns
  id_columns <- get_id_columns(id_depth = "series")

  # Set outcome_type from settings
  if (is.null(outcome_type)) outcome_type <- settings$data$outcome_type

  # Set sample identifiers, if not provided. Note that is only the case during
  # unit testing.
  if (is.null(sample_identifiers)) {
    sample_identifiers <- unique(data[, mget(id_columns)])
  }

  # Check stratification for continuous data
  if (outcome_type %in% c("continuous")) stratify <- FALSE

  # Check stratification for absent data
  if (is_empty(data)) stratify <- FALSE

  if (outcome_type %in% c("continuous")) {
    # Select data based on sample id - note that even if duplicate
    # sample_identifiers exist, only unique sample_identifiers are maintained -
    # this is intentional.
    subset_table <- merge(
      x = unique(data[, mget(c(id_columns, "outcome"))]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )
    
  } else if (outcome_type == "survival") {
    # For stratifying survival data we require the event status. Event status
    # (including NA) are transcoded.
    subset_table <- merge(
      x = unique(data[, mget(c(id_columns, "outcome_event"))]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )

    data.table::setnames(
      x = subset_table,
      old = "outcome_event",
      new = "outcome"
    )

    # Transcode event status
    subset_table[, "outcome" := factor(outcome)]
    subset_table$outcome <- addNA(subset_table$outcome, ifany = TRUE)
    
  } else if (outcome_type %in% c("binomial", "multinomial")) {
    # For stratifying categorical data we require the outcome data as is.
    # Redundant factors are dropped and remaining factors (including NA) are
    # transcoded
    subset_table <- merge(
      x = unique(data[, mget(c(id_columns, "outcome"))]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )

    # Transcode classes
    subset_table$outcome <- addNA(subset_table$outcome, ifany = TRUE)
    subset_table$outcome <- droplevels(subset_table$outcome)
    
  } else if (outcome_type %in% c("competing_risk", "multi_label")) {
    ..error_no_known_outcome_type(outcome_type)
  }

  # Determine sample-id columns
  sample_id_columns <- get_id_columns(id_depth = "sample")

  # Initiate training and validation lists
  train_list <- valid_list <- list()

  # Iterate over iterations
  ii <- jj <- 1L
  while (ii <= n_iter && jj <= 2L * n_iter) {
    if (!stratify) {
      # Sample training data with replacement
      train_id <- fam_sample(
        x = subset_table,
        replace = TRUE,
        rstream_object = rstream_object
      )

      # Merge train_id with subset_table.
      train_id <- merge(
        x = train_id,
        y = subset_table,
        by = sample_id_columns,
        all = FALSE,
        allow.cartesian = TRUE
      )

      # Check that any rare outcome classes are actually present in the training
      # data. This prevents issues with modelling and model evaluation later on.
      if (outcome_type %in% c("binomial", "multinomial", "survival")) {
        # Check if all outcome data are presents.
        missing_levels <- setdiff(
          levels(subset_table$outcome), unique(train_id$outcome)
        )

        if (length(missing_levels) > 0L) {
          # Integrate outcome levels in a table to keep track.
          missing_level_data <- data.table::data.table(
            "outcome" = levels(subset_table$outcome),
            "outcome_present" = TRUE
          )

          # Flag outcome levels as not present if they are missing.
          missing_level_data[outcome %in% missing_levels, "outcome_present" := FALSE]

          # Iterate over missing levels and remove them by adding a sample that
          # contains an outcome of the level.
          while (any(missing_level_data$outcome_present == FALSE)) {
            # Select the missing outcome level.
            missing_level <- missing_level_data[outcome_present == FALSE]$outcome[1L]

            # Select an additional sample that contains the missing level.
            additional_data <- fam_sample(
              subset_table[outcome == missing_level],
              size = 1L,
              rstream_object = rstream_object
            )

            # Select the data corresponding to the sample.
            additional_data <- merge(
              x = additional_data,
              y = subset_table,
              by = sample_id_columns,
              all = FALSE,
              allow.cartesian = TRUE
            )

            # Flag any missing levels that are now included in the data as
            # present.
            missing_level_data[
              additional_data[, "outcome"],
              "outcome_present" := TRUE, 
              on = .NATURAL
            ]

            # Update train_id.
            train_id <- data.table::rbindlist(
              list(train_id, additional_data)
            )
          }
        }
      }

      # Select only relevant columns.
      train_id <- train_id[, mget(id_columns)]

      # Determine the out-of-bag data.
      valid_id <- data.table::fsetdiff(
        x = subset_table[, mget(id_columns)],
        y = train_id
      )

      # Check for empty out-of-bag sets and re-run experiment if this happens.
      if (nrow(valid_id) == 0L) {
        jj <- jj + 1L
        next
      }

      # Store to list
      train_list[[ii]] <- train_id
      valid_list[[ii]] <- valid_id
      
    } else {
      # From here on we select the samples by iterating over classes, from minor
      # to major. In each iteration we do the following.
      #
      # * Select all samples that contain the currently selected class.
      #
      # * Remove all samples that would not fit the current partition, because
      # adding the sample would increase the frequency of any class beyond what
      # is required.
      #
      # * Resample with replacement. The order is fixed by setting a
      # sample_order_id column.
      #
      # * Select those samples that are valid in the sense that they would not
      # cause the number of instances for any outcome to be exceeded. This is
      # done in the order established above according to the sample_order_id.
      # These samples are assigned to the in-bag partition.
      #
      # * Update the frequency for each class that should still be obtained to
      # complete the outcome stratfication.
      #
      # * Use the updated frequency to remove all samples that were not
      # previously selected and that would not fit the current partition any
      # more.
      #
      # * Iterate over any remaining samples to add them to the partition.
      #
      # * Repeat the above for the next smallest class. We start with the
      # smallest classes so that there can be no out-of-bag data that contains a
      # "new" class.

      # Determine the frequency of sample outcomes, and order by increasing
      # frequency.
      level_frequency <- subset_table[, list("n" = .N), by = "outcome"][order(n)]
      class_order <- level_frequency$outcome

      # Initially mark all data as available.
      available_data <- data.table::copy(subset_table)

      # Create list to store data in.
      partition_train_list <- list()

      # Iterate over majority classes.
      for (current_class in class_order) {
        # Skip if the number of required samples for the outcome class is 0.
        if (level_frequency[outcome == current_class]$n == 0L) next

        # Select only data where the current class is present.
        partition_data <- available_data[
          unique(available_data[outcome == current_class, mget(sample_id_columns)]),
          on = .NATURAL
        ]

        # Determine the frequency of outcome levels for each sample.
        partition_data <- partition_data[, list("n" = .N), by = c(sample_id_columns, "outcome")]

        # Remove any samples that would exceed the required total instances for
        # each level. First mark the instances that do not exceed the required
        # number of samples.
        partition_data[level_frequency, "keep" := n <= i.n, on = "outcome"]
        partition_data[, "keep" := all(keep), by = c(sample_id_columns)]
        partition_data <- partition_data[keep == TRUE]

        if (is_empty(partition_data)) next

        # Randomly order samples. Note that we cannot directly select and insert
        # samples by outcome level, because a sample may contain multiple series
        # with different outcomes.

        # Order available data randomly, with replacement.
        sample_order <- fam_sample(
          x = partition_data,
          size = level_frequency[outcome == current_class]$n,
          replace = TRUE,
          rstream_object = rstream_object
        )

        # Set sample_order_id. This will be used to order partition_data.
        sample_order[, "sample_order_id" := .I]

        # Order the samples in partition data.
        partition_data <- merge(
          x = partition_data,
          y = sample_order,
          by = sample_id_columns,
          allow.cartesian = TRUE
        )[order(sample_order_id)]

        # Now, we select those samples which will surely be included. First,
        # determine the cumulative sum for each outcome.
        partition_data[, "cumulative_n" := cumsum(n), by = "outcome"]

        # Mark the samples where the cumulative sum of outcomes would not exceed
        # the required number.
        partition_data[level_frequency, "keep" := cumulative_n <= i.n, on = "outcome"]

        # Mark the corresponding samples.
        partition_data[, "keep" := all(keep), by = c(sample_id_columns, "sample_order_id")]

        # Add the selected data to the list. Note that the partition_data can
        # have multiple rows for the same sample_order_id, and we therefore
        # first select unique rows for each selected sample_order_id and from
        # this dataset then select the sample identifier columns.
        partition_train_list <- c(
          partition_train_list,
          list(
            unique(
              partition_data[
                keep == TRUE,
                mget(c(sample_id_columns, "sample_order_id"))
              ]
            )[, mget(sample_id_columns)]
          )
        )
        
        # Update partition level frequency
        level_frequency[
          partition_data[
            keep == TRUE,
            list("n" = sum(n)),
            by = "outcome"
          ],
          "n" := n - i.n,
          on = "outcome"
        ]

        # Determine if the outcome level was filled to the required level.
        if (level_frequency[outcome == current_class]$n == 0L) next

        # Select remaining data to complete
        partition_data <- partition_data[keep == FALSE]

        # Check that any data actually has been selected.
        if (nrow(partition_data) == 0L) next

        # Remove any samples that would exceed the required total instances for
        # each level. First mark the instances that do not exceed the required
        # number of samples.
        partition_data[
          level_frequency,
          "keep" := n <= i.n,
          on = "outcome"
        ]

        # Then mark the corresponding samples, and only keep those.
        partition_data[, "keep" := all(keep), by = c(sample_id_columns, "sample_order_id")]
        partition_data <- partition_data[keep == TRUE, ]

        # Check that any data actually has been selected.
        if (nrow(partition_data) == 0L) next

        # Iterate over the remaining samples. For each sample we check whether
        # it can be added to the partition.
        for (current_sample in split(partition_data, by = "sample_order_id")) {
          # Determine if the sample can be added.
          safe_to_add <- all(level_frequency[
            current_sample,
            list("n" = n - i.n),
            on = "outcome"
          ]$n >= 0L)
          
          # Skip to next sample, if the current sample cannot be added due to
          # constraints.
          if (!safe_to_add) next

          # If the check passes, we need to add the sample to the fold, and the
          # update level frequency.
          level_frequency[current_sample, "n" := n - i.n, on = "outcome"]

          # Add the selected sample to the list.
          partition_train_list <- c(
            partition_train_list,
            list(unique(current_sample[, mget(sample_id_columns)]))
          )

          # Skip further samples if no more samples need to be added to the fold.
          if (any(level_frequency$n == 0L)) break
        }

        # Mark all samples that lack the current outcome as available; or rather
        # mark all samples that have the current outcome as unavailable.
        available_data[, "is_available" := !any(outcome == current_class), by = c(sample_id_columns)]
      }

      # Select samples.
      train_id <- data.table::rbindlist(partition_train_list)

      # Merge train_id with subset_table.
      train_id <- merge(
        x = train_id,
        y = subset_table,
        by = sample_id_columns,
        all = FALSE,
        allow.cartesian = TRUE
      )

      # Select only relevant columns.
      train_id <- train_id[, mget(id_columns)]

      # Determine the out-of-bag data.
      valid_id <- data.table::fsetdiff(
        x = subset_table[, mget(id_columns)],
        y = train_id
      )

      # Check for empty out-of-bag sets and re-run sampling if this happens.
      if (nrow(valid_id) == 0L) {
        jj <- jj + 1L
        next
      }

      # Store to list
      train_list[[ii]] <- train_id
      valid_list[[ii]] <- valid_id
    }

    # Update iterators
    ii <- ii + 1L
    jj <- jj + 1L
  }

  if (ii != n_iter + 1L) {
    ..error(
      paste0(
        "Could not form ", n_iter, " bootstraps with out-of-bag data. ",
        "The data set may be too small."
      ),
      error_Class = "dataset_error"
    )
  }

  return(list(
    "train_list" = train_list,
    "valid_list" = valid_list
  ))
}



.create_balanced_partitions <- function(
    data,
    sample_identifiers = NULL,
    settings = NULL,
    outcome_type = NULL,
    imbalance_method = NULL,
    imbalance_n_partitions = NULL,
    rstream_object = NULL
) {
  # Methods to address class imbalance

  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- n <- partition <- i.n <- sample_order_id <- NULL
  keep <- cumulative_n <- .NATURAL <- NULL

  if (is.null(outcome_type)) outcome_type <- settings$data$outcome_type
  if (is.null(imbalance_method)) imbalance_method <- settings$data$imbalance_method
  if (is.null(imbalance_n_partitions)) imbalance_n_partitions <- settings$data$imbalance_n_partitions

  # Class imbalance should only be addressed if the outcome data is categorical.
  if (!outcome_type %in% c("binomial", "multinomial")) {
    logger_stop(paste0(
      "Creating iterations: Imbalance partitions (ip) are only available ",
      "for binomial and multinomial outcomes."
    ))
  }

  # Obtain id columns
  id_columns <- get_id_columns(id_depth = "series")
  sample_id_columns <- get_id_columns(id_depth = "sample")

  # Set sample identifiers, if not provided. Note that is only the case during
  # unit testing.
  if (is.null(sample_identifiers)) {
    sample_identifiers <- unique(data[, mget(id_columns)])
  }

  # For creating imbalance partitions based on categorical data we require the
  # outcome data as is. Redundant factors are dropped and remaining factors
  # (including NA) are transcoded.
  subset_table <- merge(
    x = unique(data[, mget(c(id_columns, "outcome"))]),
    y = sample_identifiers,
    by = id_columns,
    all = FALSE
  )

  # Transcode classes
  subset_table$outcome <- addNA(subset_table$outcome, ifany = TRUE)
  subset_table$outcome <- droplevels(subset_table$outcome)

  # Determine the frequency of sample outcomes, and order by frequency.
  level_frequency <- subset_table[, list("n" = .N), by = "outcome"][order(n)]
  class_order <- level_frequency$outcome

  if (imbalance_method %in% c("full_undersampling", "random_undersampling")) {
    # Create empty train_list
    train_list <- list()

    # Get the number of each class and identify the number of partitions.
    n_small <- min(level_frequency$n)
    n_large <- max(level_frequency$n)

    if (n_small / (n_large + n_small) > 0.3) {
      ..warning(paste0(
        "Imbalance partitions are not required as data ",
        "are not severely imbalanced."
      ))
    }

    # Set up the basic partition by selecting all data that are required for the
    # minority class. First determine the smallest class(es) and then the
    # majority classes (if any).
    minority_class <- level_frequency[n == n_small]$outcome
    majority_class <- setdiff(class_order, minority_class)

    # Update subset_table by assigning all samples that contain the minority
    # class to the base partition.
    subset_table[
      subset_table[outcome %in% minority_class, mget(sample_id_columns)],
      "partition" := "base",
      on = .NATURAL
    ]

    # Determine the number of partitions while allowing the majority class to
    # be up to 10% bigger. This should leave sufficient margin for a slight
    # discrepancy in class levels, when adding any remaining random elements.
    class_ratio <- n_large / n_small
    n_partitions <- ceiling(class_ratio - floor(class_ratio) * 0.1)

    # Determine the number of partitions.
    if (imbalance_method == "random_undersampling") {
      # Set randomisation flag.
      randomise <- n_partitions > 1L

      if (n_partitions > 1L) {
        # Use setting provided by the user.
        n_partitions <- imbalance_n_partitions
        
      } else if (n_partitions != imbalance_n_partitions) {
        # In case the number of unique partitions is 1, and the user-provided
        # setting is not 1, create only one partition and warn the user.
        ..warning("Only one unique partition can be created.")
      }
    } else {
      # Set randomise to FALSE.
      randomise <- FALSE
    }

    # Determine how much of the levels are occupied by the base partition, and
    # how many levels remain to be set.
    level_frequency[, "n" := n_small]
    level_frequency[
      subset_table[
        partition == "base",
        list("n" = .N),
        by = "outcome"
      ],
      "n" := n - i.n, 
      on = "outcome"
    ]
    level_frequency[n < 0L, "n" := 0L]
    
    # Throw an error if no further partitions can be formed aside from the base.
    # This basically means that the minority class is too small.
    if (all(level_frequency$n == 0L) && anyNA(subset_table$partition)) {
      ..error(
        paste0(
          "The minority class may contain too few instances (", n_small,
          ") to allow for balanced partitioning. ",
          "This error occurs when all samples containing the minority class ",
          "also completely fill out the allowed space for majority classes. ",
          "Thus, no other samples that contain a non-minority class could be ",
          "selected. We would recommend to either increase the data available, ",
          "or remove all instances of the minority class."
        ),
        error_class = "dataset_error"
      )
    }

    # Initialise iterator.
    ii <- 0L

    while (TRUE) {
      # Update the iterator
      ii <- ii + 1L

      # Set a warning flag that tracks classes for which no samples were
      # assigned.
      no_class_assigned <- NULL

      # Copy the fold_level frequency.
      partition_level_frequency <- data.table::copy(level_frequency)

      # Get all available data.
      #
      # Undersampling should use any data not assigned to the base minority
      # class dataset. This includes samples that may have been assigned to
      # other partitions. Such assigned samples will be treated at the same
      # priority for random partitions, and a lower priority for full
      # partitions.
      available_data <- subset_table[is.na(partition) | partition != "base"]

      # Break if partitions have been exhausted.
      if (imbalance_method == "random_undersampling" && ii > n_partitions) break

      # Check that any data are available, or that we can create a meaningful
      # partition using additional data.
      if (
        nrow(available_data) == 0L ||
        all(partition_level_frequency$n == 0L) ||
        length(majority_class) == 0L
      ) {
        # If no data are available in the first iteration, copy the dataset that
        # was based on the minority class(es), and break from the loop.
        if (ii == 1L) {
          train_list[[ii]] <- subset_table[partition == "base", mget(id_columns)]
        }
        
        break
      }

      # Break if all data have been assigned.
      if (imbalance_method == "full_undersampling" && !anyNA(subset_table$partition)) {
        break
      }

      # Iterate over majority classes.
      for (current_class in majority_class) {
        # Select only data where the current class is present.
        partition_data <- available_data[
          unique(available_data[outcome == current_class, mget(sample_id_columns)]),
          on = .NATURAL
        ]

        # Randomly order samples.
        #
        # In case of random partitioning, all available samples are randomly
        # ordered. Otherwise, for full partitioning, we first randomly order any
        # unassigned samples (if any), and then order any previously assigned
        # samples (if any). This is done in order that any intermediate classes
        # will be sufficiently represented in the data.
        #
        # Note that we cannot directly select and insert samples by outcome level,
        # because a sample may contain multiple series with different outcomes.
        if (randomise) {
          # For random partitioning.

          # Order available data randomly.
          sample_order <- fam_sample(
            x = partition_data,
            replace = FALSE,
            rstream_object = rstream_object
          )

          # Set sample_order_id. This will be used to order partition_data.
          sample_order[, "sample_order_id" := .I]
          
        } else {
          # For full partitioning.
          unassigned_partition_data <- partition_data[is.na(partition)]
          assigned_partition_data <- data.table::fsetdiff(
            partition_data, unassigned_partition_data
          )

          # Initial sample order offset.
          sample_order_offset <- 0L

          # Initialise datasets.
          assigned_sample_order <- unassigned_sample_order <- NULL

          if (nrow(unassigned_partition_data) > 0L) {
            # Order available data randomly.
            unassigned_sample_order <- fam_sample(
              x = unassigned_partition_data,
              replace = FALSE,
              rstream_object = rstream_object
            )

            # Set sample_order_id. This will be used to order available_data.
            unassigned_sample_order[, "sample_order_id" := .I]

            # Update the sample order offset.
            sample_order_offset <- nrow(unassigned_sample_order)
          }

          if (nrow(assigned_partition_data) > 0L) {
            # Order available data randomly.
            assigned_sample_order <- fam_sample(
              x = assigned_partition_data,
              replace = FALSE,
              rstream_object = rstream_object
            )

            # Set sample_order_id. This will be used to order available_data.
            assigned_sample_order[, "sample_order_id" := .I + sample_order_offset]
          }

          # Determine sample order.
          sample_order <- data.table::rbindlist(
            list(unassigned_sample_order, assigned_sample_order)
          )
        }

        # Merge with available_data
        partition_data <- merge(
          x = partition_data,
          y = sample_order,
          by = sample_id_columns
        )[order(sample_order_id)]

        # Determine the frequency of outcome levels for each sample.
        partition_data <- partition_data[
          , list("n" = .N),
          by = c(sample_id_columns, "outcome", "sample_order_id")
        ]
        
        # From here on we select the samples and add them to the partition. This
        # proceeds according to the following steps:
        #
        # * Remove all samples that would not fit the current partition, because
        # adding the sample would increase the frequency of any class beyond
        # what is required.
        #
        # * Select those samples that are valid in the sense that they would not
        # cause the number of instances for any outcome to be exceeded. This is
        # done in the order established above according to the sample_order_id.
        # These samples are assigned to the partition.
        #
        # * Update the frequency for each class that should still be obtained to
        # complete the outcome.
        #
        # * Use the updated frequency to remove all samples that were not
        # previously selected and that would not fit the current partition any
        # more.
        #
        # * Iterate over any remaining samples to add them to the partition.

        # Remove any samples that would exceed the required total instances for
        # each level. First mark the instances that do not exceed the required
        # number of samples.
        partition_data[
          partition_level_frequency, 
          "keep" := n <= i.n, 
          on = "outcome"
        ]

        # Then mark the corresponding samples, and only keep those.
        partition_data[, "keep" := all(keep), by = c(sample_id_columns, "sample_order_id")]
        partition_data <- partition_data[keep == TRUE]

        if (nrow(partition_data) == 0L) {
          # At this point none of the samples with the current class have been
          # assigned. If partition_data is empty, that means that the samples
          # with this class were skipped, perhaps because samples selected from
          # previous classes already created a balanced set.
          no_class_assigned <- c(no_class_assigned, current_class)
          next
        }

        # Now, we select those samples which will surely be included. First,
        # determine the cumulative sum for each outcome.
        partition_data[, "cumulative_n" := cumsum(n), by = "outcome"]

        # Mark the samples where the cumulative sum of outcomes would not exceed
        # the required number.
        partition_data[
          partition_level_frequency,
          "keep" := cumulative_n <= i.n, 
          on = "outcome"
        ]

        # Mark the corresponding samples.
        partition_data[, "keep" := all(keep), by = c(sample_id_columns, "sample_order_id")]

        # Mark the data in the subset table.
        subset_table[
          partition_data[keep == TRUE, mget(sample_id_columns)],
          "partition" := as.character(ii),
          on = .NATURAL
        ]

        # Update partition level frequency
        partition_level_frequency[
          partition_data[
            keep == TRUE,
            list("n" = sum(n)),
            by = "outcome"
          ],
          "n" := n - i.n,
          on = "outcome"
        ]

        # Determine if the outcome level was filled to the required level.
        if (partition_level_frequency[outcome == current_class]$n == 0L) next

        # Select remaining data to complete
        partition_data <- partition_data[keep == FALSE]

        # Check that any data actually has been selected.
        if (nrow(partition_data) == 0L) next

        # Remove any samples that would exceed the required total instances for
        # each level. First mark the instances that do not exceed the required
        # number of samples.
        partition_data[
          partition_level_frequency,
          "keep" := n <= i.n,
          on = "outcome"
        ]

        # Then mark the corresponding samples, and only keep those.
        partition_data[, "keep" := all(keep), by = c(sample_id_columns, "sample_order_id")]
        partition_data <- partition_data[keep == TRUE, ]

        # Check that any data actually has been selected.
        if (nrow(partition_data) == 0L) next

        # Iterate over the remaining samples. For each sample we check whether
        # it can be added to the partition.
        for (current_sample in split(partition_data, by = "sample_order_id")) {
          # Determine if the sample can be added.
          safe_to_add <- all(partition_level_frequency[
            current_sample, 
            list("n" = n - i.n),
            on = "outcome"
          ]$n >= 0L)

          # Skip to next sample, if the current sample cannot be added due to
          # constraints.
          if (!safe_to_add) next

          # If the check passes, we need to add the sample to the fold, and
          # update level_frequency_fold.
          partition_level_frequency[current_sample, "n" := n - i.n, on = "outcome"]

          # Mark the sample as being assigned. Note that randomised
          # undersampling ignores this.
          subset_table[
            unique(current_sample[, mget(sample_id_columns)]),
            "partition" := as.character(ii),
            on = .NATURAL
          ]

          # Skip further samples if no more samples need to be added to the fold.
          if (all(partition_level_frequency$n == 0L)) break
        }
      }

      # Select samples.
      train_list[[ii]] <- subset_table[
        partition %in% c("base", as.character(ii)),
        mget(id_columns)
      ]

      # Check that any unassigned samples are not completely covered by
      # no_class_assigned.
      if (!is.null(no_class_assigned) && imbalance_method == "full_undersampling") {
        if (anyNA(subset_table$partition)) {
          if (all(subset_table[is.na(partition)]$outcome %in% no_class_assigned)) {
            # Determine the number of samples that cannot be assigned.
            n_samples_not_assigned <- data.table::uniqueN(
              subset_table[is.na(partition), mget(sample_id_columns)]
            )

            # Throw warning.
            logger_warning(
              paste0(
                n_samples_not_assigned,
                ifelse(n_samples_not_assigned > 1L, " samples", " sample"),
                " could not be assigned during undersampling for balance correction."
              )
            )
            
            break
            
          } else if (ii > 1L) {
            # Check if the generated dataset is actually unique.
            flag_infinite_loop <- FALSE

            for (jj in seq_len(ii - 1L)) {
              if (data.table::fsetequal(train_list[[ii]], train_list[[jj]])) {
                # Remove non-unique dataset from train_list and break from the
                # loop.
                train_list[[ii]] <- NULL
                flag_infinite_loop <- TRUE

                break
              }
            }

            if (flag_infinite_loop) {
              # Determine the number of samples that cannot be assigned.
              n_samples_not_assigned <- data.table::uniqueN(
                subset_table[is.na(partition), mget(sample_id_columns)]
              )

              # Throw warning.
              logger_warning(paste0(
                n_samples_not_assigned,
                ifelse(n_samples_not_assigned > 1L, " samples", " sample"),
                " could not be assigned during undersampling for balance correction."
              ))
              
              break
            }
          }
        }
      }
    }

    # Update the number of partitions to the actual number of partitions.
    n_partitions <- length(train_list)

    # Return list with partitions
    return(train_list)
  }
}



.create_subsample <- function(
    data,
    size = NULL,
    n_iter = 1L,
    sample_identifiers = NULL,
    settings = NULL,
    outcome_type = NULL,
    stratify = TRUE,
    tolerance = 0.05,
    rstream_object = NULL
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  outcome <- frequency <- n <- .NATURAL <- NULL
  i.frequency <- i.n <- order_id <- cum_n <- frequency_diff <- NULL

  # Obtain id columns
  id_columns <- get_id_columns(id_depth = "series")

  # Set outcome_type from settings
  if (is.null(outcome_type)) outcome_type <- settings$data$outcome_type

  # Set sample identifiers, if not provided. Note that is only the case during
  # unit testing.
  if (is.null(sample_identifiers)) {
    sample_identifiers <- unique(data[, mget(id_columns)])
  }

  # Check stratification for continuous data
  if (outcome_type %in% c("continuous")) stratify <- FALSE

  # Check stratification for absent data
  if (is_empty(data)) stratify <- FALSE

  if (outcome_type %in% c("continuous")) {
    # Select data based on sample id - note that even if duplicate
    # sample_identifiers exist, only unique sample_identifiers are maintained -
    # this is intentional.
    subset_table <- merge(
      x = unique(data[, mget(c(id_columns, "outcome"))]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )
    
  } else if (outcome_type == "survival") {
    # For stratifying survival data we require the event status.
    # Event status (including NA) are transcoded.
    subset_table <- merge(
      x = unique(data[, mget(c(id_columns, "outcome_event"))]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )

    data.table::setnames(
      x = subset_table,
      old = "outcome_event",
      new = "outcome"
    )

    # Transcode event status
    subset_table[, "outcome" := factor(outcome)]
    subset_table$outcome <- addNA(subset_table$outcome, ifany = TRUE)
    
  } else if (outcome_type %in% c("binomial", "multinomial")) {
    # For stratifying categorical data we require the outcome data as is.
    # Redundant factors are dropped and remaining factors (including NA) are
    # transcoded
    subset_table <- merge(
      x = unique(data[, mget(c(id_columns, "outcome"))]),
      y = sample_identifiers,
      by = id_columns,
      all = FALSE
    )

    # Transcode classes
    subset_table$outcome <- addNA(subset_table$outcome, ifany = TRUE)
    subset_table$outcome <- droplevels(subset_table$outcome)
    
  } else if (outcome_type %in% c("competing_risk", "multi_label")) {
    ..error_no_known_outcome_type(outcome_type)
  }

  # Determine sample-id columns
  sample_id_columns <- get_id_columns(id_depth = "sample")

  # Get the number of samples.
  n_samples <- data.table::uniqueN(subset_table[, mget(sample_id_columns)])

  # Check size
  if (is.null(size)) size <- n_samples
  if (size > n_samples) size <- n_samples

  # Initiate training and validation lists
  train_list <- valid_list <- list()

  # Iterate
  for (ii in seq_len(n_iter)) {
    # Set initial size. This will decrease when assigning samples.
    size_remaining <- size

    # Make a copy of the data.
    available_data <- data.table::copy(subset_table)

    # Set availability status.
    available_data[, "is_available" := TRUE]

    if (outcome_type %in% c("survival", "binomial", "multinomial")) {
      # Determine the frequency of sample outcomes, and order by increasing
      # frequency.
      level_frequency <- subset_table[, list("n" = .N), by = "outcome"][order(n)]
      level_frequency[, "frequency" := n / sum(n)]
      class_order <- level_frequency$outcome

      # Pick initial samples.
      for (current_class in class_order) {
        # Check if any samples can be selected.
        if (size_remaining == 0L) break

        # Check if the class has already been selected.
        if (any(available_data[is_available == FALSE]$outcome == current_class)) next

        # Select data that contains the current data.
        partition_data <- available_data[
          unique(available_data[
            outcome == current_class & is_available == TRUE,
            mget(sample_id_columns)
          ]),
          on = .NATURAL
        ]
        
        # Select a single sample from the potential datasets.
        train_id <- fam_sample(partition_data,
          size = 1L,
          replace = FALSE,
          rstream_object = rstream_object
        )

        # Mark as being selected.
        available_data[train_id, "is_available" := FALSE, on = .NATURAL]

        # Reduce size.
        size_remaining <- size_remaining - 1L
      }
    }

    if (!stratify && size_remaining > 0L) {
      # Unstratified assignment.

      # Select a samples from the remaining available samples.
      train_id <- fam_sample(
        available_data[is_available == TRUE, mget(sample_id_columns)],
        size = size_remaining,
        replace = FALSE,
        rstream_object = rstream_object
      )

      # Mark as being selected.
      available_data[train_id, "is_available" := FALSE, on = .NATURAL]

      # Reduce size.
      size_remaining <- 0L
      
    } else if (stratify && size_remaining > 0L) {
      # Stratified assignment. This is more complex. First we attempt to
      # randomly assign samples, for which we check tolerance with the expected
      # stratification. Starting with the largest randomly selected sample set
      # for which class frequencies were acceptable, we then statically add any
      # suitable samples that are within tolerance, or improve the situation,
      # until we collected all the required samples.

      # Correct for already selected samples.
      selected_frequency <- available_data[
        is_available == FALSE,
        list("n" = .N),
        by = "outcome"
      ][order(n)]

      # Select samples that have not been selected previously.
      partition_data <- available_data[
        unique(available_data[is_available == TRUE, mget(sample_id_columns)]),
        on = .NATURAL
      ]
      
      # Set null data.
      null_data <- data.table::data.table(expand.grid(
        "order_id" = seq_len(size_remaining),
        "outcome" = levels(subset_table$outcome),
        "n" = 0L
      ))

      # Randomly selected samples.
      random_data <- fam_sample(partition_data,
        size = size_remaining,
        replace = FALSE,
        rstream_object = rstream_object
      )

      # Add in indices.
      random_data[, "order_id" := .I]

      # Update null data.
      null_data <- null_data[random_data, on = .NATURAL]

      # Compute the number of outcomes in random data
      random_data <- partition_data[random_data, on = .NATURAL]
      random_data <- random_data[, list("n" = .N), by = c("order_id", sample_id_columns, "outcome")]

      # Update missing levels.
      random_data <- data.table::rbindlist(
        list(random_data, null_data),
        use.names = TRUE
      )
      random_data <- random_data[, list("n" = max(n)), by = c("order_id", sample_id_columns, "outcome")]
      random_data <- random_data[order(order_id)]

      # Compute the cumulative number of outcomes.
      random_data[, "cum_n" := cumsum(n), by = "outcome"]

      # Add previously selected frequencies.
      random_data[selected_frequency, "cum_n" := cum_n + i.n, on = "outcome"]

      # Compute the frequency.
      random_data[, "frequency" := cum_n / sum(cum_n), by = "order_id"]

      # Compare with the required frequency.
      random_data[
        level_frequency, 
        "frequency_diff" := abs(frequency - i.frequency),
        on = "outcome"
      ]

      # Check where the random assignment was reasonably ok.
      frequency_check <- random_data[
        ,
        list("difference_ok" = all(frequency_diff <= tolerance)),
        by = "order_id"
      ]
      last_ok_sample <- tail(which(frequency_check$difference_ok), n = 1L)

      if (length(last_ok_sample) > 0L) {
        # Mark data.
        available_data[
          unique(random_data[order_id <= last_ok_sample, mget(sample_id_columns)]),
          "is_available" := FALSE,
          on = .NATURAL
        ]

        # Update remaining.
        size_remaining <- size_remaining - last_ok_sample
      }

      # Determine selected_frequency for already selected samples.
      selected_frequency <- available_data[
        is_available == FALSE, 
        list("n" = .N),
        by = "outcome"
      ][order(n)]
      selected_frequency[, "frequency" := n / sum(n)]
      selected_frequency[
        level_frequency, 
        "frequency_diff" := abs(frequency - i.frequency),
        on = "outcome"
      ]

      # Static selection.
      while (size_remaining > 0L) {
        # Select available data.
        partition_data <- available_data[
          unique(available_data[
            is_available == TRUE,
            mget(sample_id_columns)
          ]),
          on = .NATURAL
        ]
        
        # Shuffle partition_data.
        partition_data <- fam_sample(
          partition_data,
          size = data.table::uniqueN(partition_data[, mget(sample_id_columns)]),
          replace = FALSE,
          rstream_object = rstream_object
        )

        # Add in outcome data to the number of outcomes in random data
        partition_data <- available_data[partition_data, on = .NATURAL]

        # Compute the number of outcomes.
        partition_data <- partition_data[, list("n" = .N), by = c(sample_id_columns, "outcome")]

        # Flag to check if any samples were added.
        any_samples_added <- FALSE

        for (current_sample in split(partition_data, by = sample_id_columns)) {
          # Check if adding the sample would improve the frequency, or at least
          # not reduce it too much.
          new_frequency <- data.table::copy(selected_frequency)
          new_frequency[current_sample, "n" := n + i.n, on = "outcome"]
          new_frequency[, "frequency" := n / sum(n)]
          new_frequency[
            level_frequency,
            "frequency_diff" := abs(frequency - i.frequency),
            on = "outcome"
          ]

          # Check difference of previously selected
          previous_diff <- sum(selected_frequency$frequency_diff)
          new_diff <- sum(new_frequency$frequency_diff)

          # Check if difference was previously ok.
          previous_diff_ok <- all(selected_frequency$frequency_diff <= tolerance)
          new_diff_ok <- all(new_frequency$frequency_diff <= tolerance)

          assign_sample <- FALSE
          
          if (new_diff_ok) {
            # Assign sample if the new difference is still ok.
            assign_sample <- TRUE
            
          } else if (!previous_diff_ok && !new_diff_ok && new_diff <= previous_diff) {
            # Assign sample if the both the previous difference and new
            # difference are not ok, but the new difference is better.
            assign_sample <- TRUE
          }

          if (assign_sample) {
            # Assign the current sample.
            available_data[
              current_sample[, mget(sample_id_columns)],
              "is_available" := FALSE,
              on = .NATURAL
            ]
            
            # Update remaining.
            size_remaining <- size_remaining - 1L

            # Flag that samples were in fact added.
            any_samples_added <- TRUE

            # Use the new frequency as the selected frequency.
            selected_frequency <- data.table::copy(new_frequency)

            if (size_remaining == 0L) break
          }
        }

        # Check if any new samples were added. If not, simply draw from the
        # dataset.
        if (!any_samples_added) {
          # Select a single sample from the potential datasets.
          train_id <- fam_sample(
            partition_data,
            size = size_remaining,
            replace = FALSE,
            rstream_object = rstream_object
          )

          # Mark as being selected.
          available_data[train_id, "is_available" := FALSE, on = .NATURAL]

          # Reduce size.
          size_remaining <- 0L
        }
      }
    }

    # Select only relevant columns from selected samples.
    train_id <- unique(available_data[is_available == FALSE, mget(id_columns)])

    # Determine the out-of-bag data.
    valid_id <- data.table::fsetdiff(
      x = subset_table[, mget(id_columns)],
      y = train_id
    )

    # Store to list
    train_list[[ii]] <- train_id
    valid_list[[ii]] <- valid_id

    # Update iterators
    ii <- ii + 1L
  }

  return(list(
    "train_list" = train_list,
    "valid_list" = valid_list
  ))
}



.collect_run_tables <- function(iteration_list) {
  run_tables <- list()
  ii <- 1L
  for (iteration in iteration_list) {
    for (run in iteration$run) {
      run_tables[[ii]] <- run$run_table
      ii <- ii + 1L
    }
  }
  
  # Add names.
  names(run_tables) <- sapply(
    run_tables,
    function(x) {
      data_id <- tail(x, n = 1L)$data_id[[1L]]
      run_id <- tail(x, n = 1L)$run_id[[1L]]
      
      return(paste0(data_id, ".", run_id))
    },
    simplify = TRUE,
    USE.NAMES = FALSE
  )
  
  return(run_tables)
}

# Skip on CRAN as this test takes about a minute.
testthat::skip_on_cran()
testthat::skip_on_ci()

# Create random number generator stream for reproducibility.
r <- familiar:::.start_random_number_stream(seed = 1863)

# Subsampling ------------------------------------------------------------------
for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    rare_outcome = TRUE,
    rstream_object = r
  )
  n_rep <- 3L

  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    available_stratify_options <- c(FALSE, TRUE)
  } else {
    available_stratify_options <- FALSE
  }

  for (stratify in available_stratify_options) {
    testthat::test_that(paste0(
      "Subsampling ", ifelse(stratify, "(stratified) ", ""),
      "for ", outcome_type, " functions correctly."
    ), {
      # Create subsample.
      subsample_data <- familiar:::.create_subsample(
        data = data@data,
        n_iter = 20,
        size = 20,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r
      )

      # Check that none of in-bag datasets is the same.
      for (ii in 1:(length(subsample_data$train_list) - 1)) {
        for (jj in (ii + 1):length(subsample_data$train_list)) {
          testthat::expect_false(data.table::fsetequal(
            subsample_data$train_list[[ii]],
            subsample_data$train_list[[jj]]
          ))
        }
      }

      for (ii in seq_along(subsample_data$train_list)) {
        # Check that there is no overlap between in-bag and out-of-bag data.
        testthat::expect_equal(
          nrow(data.table::fintersect(
            unique(subsample_data$train_list[[ii]]),
            unique(subsample_data$valid_list[[ii]]))),
          0L)

        # Check that the combination of in-bag and out-of-bag data is the same
        # as the input dataset.
        testthat::expect_true(data.table::fsetequal(
          unique(rbind(subsample_data$train_list[[ii]], subsample_data$valid_list[[ii]])),
          unique(data@data[, mget(familiar:::get_id_columns(id_depth = "series"))])
        ))

        # Check that sampling creates a dataset identical to the development
        # dataset.
        train_data <- familiar:::select_data_from_samples(
          data = data,
          samples = subsample_data$train_list[[ii]])

        # Test that the samples and series are selected.
        testthat::expect_true(data.table::fsetequal(
          train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
          subsample_data$train_list[[ii]]))
        
        # Test that repetitions are likewise selected.
        testthat::expect_equal(
          nrow(train_data@data),
          n_rep * nrow(subsample_data$train_list[[ii]]))

        # Test that the subsample has the required size.
        testthat::expect_equal(
          data.table::uniqueN(train_data@data, by = c("batch_id", "sample_id")), 20)

        # Check that sampling creates a dataset identical to the validation
        # dataset.
        validation_data <- familiar:::select_data_from_samples(
          data = data,
          samples = subsample_data$valid_list[[ii]])

        # Test that the samples and series are selected.
        testthat::expect_true(
          data.table::fsetequal(
            validation_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
            subsample_data$valid_list[[ii]]))

        # Test that repetitions are likewise selected.
        testthat::expect_equal(
          nrow(validation_data@data),
          n_rep * nrow(subsample_data$valid_list[[ii]]))

        # If stratified, check that occurrence of event or categories is similar
        # between discovery and the entire dataset.
        if (stratify & outcome_type %in% c("binomial", "multinomial")) {
          # Determine the frequency of outcome classes in the original dataset.
          input_frequency <- data@data[, list(
            "frequency_original" = .N / nrow(data@data)), by = "outcome"]
          train_frequency <- train_data@data[, list(
            "frequency_bootstrap" = .N / nrow(train_data@data)), by = "outcome"]

          # Update the frequency table.
          frequency_table <- merge(
            x = input_frequency,
            y = train_frequency,
            by = "outcome")

          # Check that the data is correctly stratified.
          frequency_table[, "similar" := data.table::between(
            frequency_bootstrap,
            lower = frequency_original - 0.05,
            upper = frequency_original + 0.05
          )]

          testthat::expect_true(all(frequency_table$similar))
          
        } else if (stratify & outcome_type %in% c("survival")) {
          # Determine the frequency of censored data points and events in
          # classes in the original dataset.
          input_frequency <- data@data[, list(
            "frequency_original" = .N / nrow(data@data)), by = "outcome_event"]
          train_frequency <- train_data@data[, list(
            "frequency_bootstrap" = .N / nrow(train_data@data)), by = "outcome_event"]

          # Update the frequency table.
          frequency_table <- merge(
            x = input_frequency,
            y = train_frequency,
            by = "outcome_event")

          # Check that the data is correctly stratified.
          frequency_table[, "similar" := data.table::between(
            frequency_bootstrap,
            lower = frequency_original - 0.05,
            upper = frequency_original + 0.05
          )]

          testthat::expect_true(all(frequency_table$similar))
        }

        # Check that the rare outcome is found in the training data. This
        # prevent issues with training data.
        if (outcome_type == "multinomial") {
          testthat::expect_gt(nrow(train_data@data[outcome == "3"]), 0)
        }

        # Assert that all outcome levels in the validation folds also appear in
        # the training folds.
        if (outcome_type %in% c("binomial", "multinomial", "survival")) {
          testthat::expect_equal(
            length(setdiff(
              unique(validation_data@data$outcome),
              unique(train_data@data$outcome))),
            0)
        }
      }
    })
  }
}


for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  # Create synthetic dataset with one outcome.

  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    available_stratify_options <- c(FALSE, TRUE)
  } else {
    available_stratify_options <- FALSE
  }

  for (stratify in available_stratify_options) {
    testthat::test_that(paste0(
      "Bootstrap ", ifelse(stratify, "(stratified) ", ""),
      "for ", outcome_type, " with odd data functions correctly."
    ), {
      # One outcome-data
      data <- familiar:::test_create_synthetic_series_one_outcome(
        outcome_type = outcome_type,
        rstream_object = r)

      # Create subsample.
      subsample_data <- familiar:::.create_subsample(
        data = data@data,
        n_iter = 20,
        size = 20,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r)

      # Expect a list. This is sort of a placeholder because subsampling should
      # work even when the outcome value is singular.
      testthat::expect_type(subsample_data, "list")

      # One sample data.
      data <- familiar:::test_create_synthetic_series_one_sample_data(
        outcome_type = outcome_type,
        rstream_object = r)

      # Create subsample. We don't expect an error because you can't do
      # cross-validation with a single sample.
      subsample_data <- familiar:::.create_subsample(
        data = data@data,
        n_iter = 20,
        size = 20,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r)

      testthat::expect_type(subsample_data, "list")
    })
  }
}

# Full undersampling -----------------------------------------------------------
for (outcome_type in c("binomial", "multinomial")) {
  # Create synthetic dataset.
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    rare_outcome = FALSE,
    rstream_object = r)
  n_rep <- 3L

  testthat::test_that(paste0(
    "Full undersampling for correcting outcome imbalances for ",
    outcome_type, " functions correctly."
  ), {
    # Create subsample.
    subsample_data <- suppressWarnings(familiar:::.create_balanced_partitions(
      data = data@data,
      outcome_type = outcome_type,
      imbalance_method = "full_undersampling",
      rstream_object = r))

    # Check that none of the training folds are the same.
    if (length(subsample_data) > 1) {
      for (ii in 1:(length(subsample_data) - 1)) {
        for (jj in (ii + 1):length(subsample_data)) {
          testthat::expect_false(data.table::fsetequal(
            subsample_data[[ii]],
            subsample_data[[jj]]))
        }
      }
    }

    for (ii in seq_along(subsample_data)) {
      # Assert that all samples in the subsample are unique (not duplicated).
      testthat::expect_equal(anyDuplicated(subsample_data[[ii]]), 0)

      # Check that sampling creates a dataset identical to the development
      # subsample.
      train_data <- familiar:::select_data_from_samples(
        data = data,
        samples = subsample_data[[ii]])

      # Test that the samples and series are selected.
      testthat::expect_true(
        data.table::fsetequal(
          train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
          subsample_data[[ii]]))

      # Test that repetitions are likewise selected.
      testthat::expect_equal(
        nrow(train_data@data),
        n_rep * nrow(subsample_data[[ii]]))

      # Assert that outcomes the minority class is now selected as least as
      # often as other classes.
      original_table <- unique(data@data, by = familiar:::get_id_columns(id_depth = "series"))
      original_table <- original_table[, list("n" = .N), by = "outcome"][order(n)]
      minority_class_n <- min(original_table$n)
      minority_class <- original_table[n == minority_class_n]$outcome[1]

      frequency_table <- unique(train_data@data, by = familiar:::get_id_columns(id_depth = "series"))
      frequency_table <- frequency_table[, list("partition_occurrence" = .N), by = "outcome"]

      # Assert that all instances of the minority class are selected.
      testthat::expect_equal(
        frequency_table[outcome == minority_class]$partition_occurrence,
        minority_class_n)

      # Assert that all instances similar to the minority class are selected.
      testthat::expect_true(all(frequency_table$partition_occurrence <= minority_class_n))
    }
  })
}

for (outcome_type in c("binomial", "multinomial")) {
  # Create synthetic dataset.
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_series = 1L,
    n_samples = 30,
    rare_outcome = FALSE,
    rstream_object = r
  )
  n_rep <- 3L

  testthat::test_that(paste0(
    "Full undersampling for correcting outcome imbalances for ",
    outcome_type, " without multiple series functions correctly."
  ), {
    # Create subsample.
    subsample_data <- suppressWarnings(familiar:::.create_balanced_partitions(
      data = data@data,
      outcome_type = outcome_type,
      imbalance_method = "full_undersampling",
      rstream_object = r))

    # Check that none of the training folds are the same.
    if (length(subsample_data) > 1) {
      for (ii in 1:(length(subsample_data) - 1)) {
        for (jj in (ii + 1):length(subsample_data)) {
          testthat::expect_false(data.table::fsetequal(
            subsample_data[[ii]],
            subsample_data[[jj]]))
        }
      }
    }

    # The union of the datasets is the original dataset.
    testthat::expect_true(data.table::fsetequal(
      unique(data.table::rbindlist(subsample_data)),
      unique(data@data[, mget(familiar:::get_id_columns(id_depth = "series"))])))
    
    for (ii in seq_along(subsample_data)) {
      # Assert that all samples in the subsample are unique (not duplicated).
      testthat::expect_equal(anyDuplicated(subsample_data[[ii]]), 0)

      # Check that sampling creates a dataset identical to the development
      # subsample.
      train_data <- familiar:::select_data_from_samples(
        data = data,
        samples = subsample_data[[ii]])

      # Test that the samples and series are selected.
      testthat::expect_true(data.table::fsetequal(
        train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
        subsample_data[[ii]]))
      
      # Test that repetitions are likewise selected.
      testthat::expect_equal(
        nrow(train_data@data),
        n_rep * nrow(subsample_data[[ii]]))

      # Assert that outcomes the minority class is now selected as least as
      # often as other classes.
      original_table <- unique(data@data, by = familiar:::get_id_columns(id_depth = "series"))
      original_table <- original_table[, list("n" = .N), by = "outcome"][order(n)]
      minority_class_n <- min(original_table$n)
      minority_class <- original_table[n == minority_class_n]$outcome[1]

      frequency_table <- unique(train_data@data, by = familiar:::get_id_columns(id_depth = "series"))
      frequency_table <- frequency_table[, list("partition_occurrence" = .N), by = "outcome"]

      # Assert that all instances of the minority class are selected.
      testthat::expect_equal(
        frequency_table[outcome == minority_class]$partition_occurrence,
        minority_class_n)

      # Assert that all instances similar to the minority class are selected.
      testthat::expect_true(all(frequency_table$partition_occurrence == minority_class_n))
    }
  })
}


for (outcome_type in c("binomial", "multinomial")) {
  # Create synthetic dataset with one outcome.

  testthat::test_that(paste0(
    "Full undersampling for correcting outcome imbalances for ",
    outcome_type, " with odd data functions correctly."
  ), {
    # One outcome-data
    data <- familiar:::test_create_synthetic_series_one_outcome(
      outcome_type = outcome_type,
      rstream_object = r)

    # Create subsample.
    subsample_data <- suppressWarnings(familiar:::.create_balanced_partitions(
      data = data@data,
      outcome_type = outcome_type,
      imbalance_method = "full_undersampling",
      rstream_object = r))
    
    # Expect a list. This is sort of a placeholder because the partitioning
    # should work.
    testthat::expect_type(subsample_data, "list")

    # One sample data.
    data <- familiar:::test_create_synthetic_series_one_sample_data(
      outcome_type = outcome_type,
      rstream_object = r)

    # Create subsample
    subsample_data <- suppressWarnings(familiar:::.create_balanced_partitions(
      data = data@data,
      outcome_type = outcome_type,
      imbalance_method = "full_undersampling",
      rstream_object = r))

    # Expect a list. This is sort of a placeholder because the partitioning
    # should work.
    testthat::expect_type(subsample_data, "list")
  })
}

# Random undersampling ---------------------------------------------------------
for (outcome_type in c("binomial", "multinomial")) {
  # Create synthetic dataset.
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    rare_outcome = FALSE,
    rstream_object = r)
  
  n_rep <- 3L

  testthat::test_that(paste0(
    "Random undersampling for correcting outcome imbalances for ",
    outcome_type, " functions correctly."
  ), {
    # Create subsample.
    subsample_data <- suppressWarnings(familiar:::.create_balanced_partitions(
      data = data@data,
      outcome_type = outcome_type,
      imbalance_n_partitions = 3L,
      imbalance_method = "random_undersampling",
      rstream_object = r))

    # Check that none of the training folds are the same.
    if (length(subsample_data) > 1) {
      for (ii in 1:(length(subsample_data) - 1)) {
        for (jj in (ii + 1):length(subsample_data)) {
          testthat::expect_false(data.table::fsetequal(
            subsample_data[[ii]],
            subsample_data[[jj]]))
        }
      }
    }

    # Check that at most 3 (the number specified) partitions are created.
    testthat::expect_lte(length(subsample_data), 3L)

    for (ii in seq_along(subsample_data)) {
      # Assert that all samples in the subsample are unique (not duplicated).
      testthat::expect_equal(anyDuplicated(subsample_data[[ii]]), 0)

      # Check that sampling creates a dataset identical to the development
      # subsample.
      train_data <- familiar:::select_data_from_samples(
        data = data,
        samples = subsample_data[[ii]])

      # Test that the samples and series are selected.
      testthat::expect_true(data.table::fsetequal(
        train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
        subsample_data[[ii]]
      ))

      # Test that repetitions are likewise selected.
      testthat::expect_equal(
        nrow(train_data@data),
        n_rep * nrow(subsample_data[[ii]]))

      # Assert that outcomes the minority class is now selected as least as
      # often as other classes.
      original_table <- unique(data@data, by = familiar:::get_id_columns(id_depth = "series"))
      original_table <- original_table[, list("n" = .N), by = "outcome"][order(n)]
      minority_class_n <- min(original_table$n)
      minority_class <- original_table[n == minority_class_n]$outcome[1]

      frequency_table <- unique(train_data@data, by = familiar:::get_id_columns(id_depth = "series"))
      frequency_table <- frequency_table[, list("partition_occurrence" = .N), by = "outcome"]

      # Assert that all instances of the minority class are selected.
      testthat::expect_equal(
        frequency_table[outcome == minority_class]$partition_occurrence, 
        minority_class_n)

      # Assert that all instances similar to the minority class are selected.
      testthat::expect_true(all(frequency_table$partition_occurrence <= minority_class_n))
    }
  })
}

for (outcome_type in c("binomial", "multinomial")) {
  # Create synthetic dataset.
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    n_series = 1L,
    n_samples = 30,
    rare_outcome = FALSE,
    rstream_object = r)
  
  n_rep <- 3L

  testthat::test_that(paste0(
    "Random undersampling for correcting outcome imbalances for ",
    outcome_type, " without multiple series functions correctly."
  ), {
    # Create subsample.
    subsample_data <- suppressWarnings(familiar:::.create_balanced_partitions(
      data = data@data,
      outcome_type = outcome_type,
      imbalance_n_partitions = 3L,
      imbalance_method = "random_undersampling",
      rstream_object = r))

    # Check that none of the training folds are the same.
    if (length(subsample_data) > 1) {
      for (ii in 1:(length(subsample_data) - 1)) {
        for (jj in (ii + 1):length(subsample_data)) {
          testthat::expect_false(data.table::fsetequal(
            subsample_data[[ii]],
            subsample_data[[jj]]))
        }
      }
    }

    for (ii in seq_along(subsample_data)) {
      # Assert that all samples in the subsample are unique (not duplicated).
      testthat::expect_equal(anyDuplicated(subsample_data[[ii]]), 0)

      # Check that sampling creates a dataset identical to the development
      # subsample.
      train_data <- familiar:::select_data_from_samples(
        data = data,
        samples = subsample_data[[ii]])

      # Test that the samples and series are selected.
      testthat::expect_true(data.table::fsetequal(
        train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
        subsample_data[[ii]]))

      # Test that repetitions are likewise selected.
      testthat::expect_equal(
        nrow(train_data@data),
        n_rep * nrow(subsample_data[[ii]]))

      # Assert that outcomes the minority class is now selected as least as
      # often as other classes.
      original_table <- unique(data@data, by = familiar:::get_id_columns(id_depth = "series"))
      original_table <- original_table[, list("n" = .N), by = "outcome"][order(n)]
      minority_class_n <- min(original_table$n)
      minority_class <- original_table[n == minority_class_n]$outcome[1]

      frequency_table <- unique(train_data@data, by = familiar:::get_id_columns(id_depth = "series"))
      frequency_table <- frequency_table[, list("partition_occurrence" = .N), by = "outcome"]

      # Assert that all instances of the minority class are selected.
      testthat::expect_equal(
        frequency_table[outcome == minority_class]$partition_occurrence, 
        minority_class_n)

      # Assert that all instances similar to the minority class are selected.
      testthat::expect_true(all(frequency_table$partition_occurrence == minority_class_n))
    }
  })
}

for (outcome_type in c("binomial", "multinomial")) {
  # Create synthetic dataset with one outcome.

  testthat::test_that(paste0(
    "Random undersampling for correcting outcome imbalances for ",
    outcome_type, " with odd data functions correctly."
  ), {
    # One outcome-data
    data <- familiar:::test_create_synthetic_series_one_outcome(
      outcome_type = outcome_type,
      rstream_object = r)

    # Create subsample.
    subsample_data <- suppressWarnings(familiar:::.create_balanced_partitions(
      data = data@data,
      outcome_type = outcome_type,
      imbalance_n_partitions = 3L,
      imbalance_method = "random_undersampling",
      rstream_object = r))
    
    # Expect a list. This is sort of a placeholder because the partitioning
    # should work.
    testthat::expect_type(subsample_data, "list")

    # One sample data.
    data <- familiar:::test_create_synthetic_series_one_sample_data(
      outcome_type = outcome_type,
      rstream_object = r)

    # Create subsample
    subsample_data <- suppressWarnings(familiar:::.create_balanced_partitions(
      data = data@data,
      outcome_type = outcome_type,
      imbalance_n_partitions = 3L,
      imbalance_method = "random_undersampling",
      rstream_object = r))

    # Expect a list. This is sort of a placeholder because the partitioning
    # should work.
    testthat::expect_type(subsample_data, "list")
  })
}

# Cross-validation -------------------------------------------------------------
for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    rare_outcome = TRUE,
    rstream_object = r)
  
  n_rep <- 3L
  n_folds <- 3L

  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    available_stratify_options <- c(FALSE, TRUE)
  } else {
    available_stratify_options <- FALSE
  }

  for (stratify in available_stratify_options) {
    testthat::test_that(paste0(
      "Cross-validation ", ifelse(stratify, "(stratified) ", ""),
      "for ", outcome_type, " functions correctly."
    ), {
      # Create subsample.
      subsample_data <- familiar:::.create_cv(
        data = data@data,
        n_folds = n_folds,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r)

      # Check that none of the training folds are the same.
      for (ii in 1:(length(subsample_data$train_list) - 1)) {
        for (jj in (ii + 1):length(subsample_data$train_list)) {
          testthat::expect_false(data.table::fsetequal(
            subsample_data$train_list[[ii]],
            subsample_data$train_list[[jj]]
          ))
        }
      }

      for (ii in seq_along(subsample_data$train_list)) {
        # Check that there is no overlap between training folds and the
        # validation fold.
        testthat::expect_equal(
          nrow(data.table::fintersect(
            unique(subsample_data$train_list[[ii]]),
            unique(subsample_data$valid_list[[ii]]))),
          0L)
        
        # Check that the union of training and validation folds is the input
        # dataset.
        testthat::expect_true(data.table::fsetequal(
          unique(rbind(subsample_data$train_list[[ii]], subsample_data$valid_list[[ii]])),
          unique(data@data[, mget(familiar:::get_id_columns(id_depth = "series"))])
        ))
        
        # Assert that all samples in the training and validation folds are
        # unique (not duplicated).
        testthat::expect_equal(anyDuplicated(subsample_data$train_list[[ii]]), 0)
        testthat::expect_equal(anyDuplicated(subsample_data$valid_list[[ii]]), 0)

        # Check that sampling creates a dataset identical to the development
        # subsample.
        train_data <- familiar:::select_data_from_samples(
          data = data,
          samples = subsample_data$train_list[[ii]])

        # Test that the samples and series are selected.
        testthat::expect_true(data.table::fsetequal(
          train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
          subsample_data$train_list[[ii]]))
        
        # Test that repetitions are likewise selected.
        testthat::expect_equal(
          nrow(train_data@data),
          n_rep * nrow(subsample_data$train_list[[ii]]))

        # Test that the size of the training folds is about (n_folds - 1) /
        # n_folds of the complete set.
        testthat::expect_gte(nrow(train_data@data), nrow(data@data) * (n_folds - 1) / n_folds - 9)
        testthat::expect_lte(nrow(train_data@data), nrow(data@data) * (n_folds - 1) / n_folds + 9)

        # Check that sampling creates a dataset identical to the validation subsample.
        validation_data <- familiar:::select_data_from_samples(
          data = data,
          samples = subsample_data$valid_list[[ii]])

        # Test that the samples and series are selected.
        testthat::expect_true(data.table::fsetequal(
          validation_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
          subsample_data$valid_list[[ii]]))
        
        # Test that repetitions are likewise selected.
        testthat::expect_equal(
          nrow(validation_data@data),
          n_rep * nrow(subsample_data$valid_list[[ii]]))

        # Test that the size of the validation fold is about 1 / n_folds of
        # the complete set.
        testthat::expect_gte(nrow(validation_data@data), nrow(data@data) * 1 / n_folds - 9)
        testthat::expect_lte(nrow(validation_data@data), nrow(data@data) * 1 / n_folds + 9)

        # Assert that data are correctly stratified.
        if (stratify && outcome_type %in% c("binomial", "multinomial")) {
          # Determine the frequency of outcome classes in the original dataset.
          input_frequency <- data@data[, list(
            "frequency_original" = .N / nrow(data@data)), by = "outcome"]
          train_frequency <- train_data@data[, list(
            "frequency_bootstrap" = .N / nrow(train_data@data)), by = "outcome"]

          # Update the frequency table.
          frequency_table <- merge(
            x = input_frequency,
            y = train_frequency,
            by = "outcome")

          # Check that the data is correctly stratified.
          frequency_table[, "similar" := data.table::between(
            frequency_bootstrap,
            lower = frequency_original - 0.05,
            upper = frequency_original + 0.05
          )]

          testthat::expect_true(all(frequency_table$similar))
          
        } else if (stratify & outcome_type %in% c("survival")) {
          # Determine the frequency of censored data points and events in
          # classes in the original dataset.
          input_frequency <- data@data[, list(
            "frequency_original" = .N / nrow(data@data)), by = "outcome_event"]
          train_frequency <- train_data@data[, list(
            "frequency_bootstrap" = .N / nrow(train_data@data)), by = "outcome_event"]

          # Update the frequency table.
          frequency_table <- merge(
            x = input_frequency,
            y = train_frequency,
            by = "outcome_event")

          # Check that the data is correctly stratified.
          frequency_table[, "similar" := data.table::between(
            frequency_bootstrap,
            lower = frequency_original - 0.05,
            upper = frequency_original + 0.05
          )]

          testthat::expect_true(all(frequency_table$similar))
        }

        # Check that the rare outcome is found in the training data. This
        # prevent issues with training data.
        if (outcome_type == "multinomial") {
          testthat::expect_gt(nrow(train_data@data[outcome == "3"]), 0)
        }

        # Assert that all outcome levels in the validation folds also appear in
        # the training folds.
        if (outcome_type %in% c("binomial", "multinomial", "survival")) {
          testthat::expect_equal(
            length(setdiff(
              unique(validation_data@data$outcome),
              unique(train_data@data$outcome))),
            0)
        }
      }
    })
  }
}

for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  # Create synthetic dataset with one outcome.

  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    available_stratify_options <- c(FALSE, TRUE)
  } else {
    available_stratify_options <- FALSE
  }

  n_folds <- 3L

  for (stratify in available_stratify_options) {
    testthat::test_that(paste0(
      "Cross-validation ", ifelse(stratify, "(stratified) ", ""),
      "for ", outcome_type, " with odd data functions correctly."
    ), {
      # One outcome-data
      data <- familiar:::test_create_synthetic_series_one_outcome(
        outcome_type = outcome_type,
        rstream_object = r)

      # Create subsample.
      subsample_data <- suppressWarnings(familiar:::.create_cv(
        data = data@data,
        n_folds = n_folds,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r))
      
      # Expect a list. This is sort of a placeholder because cross-validation
      # should work even when the outcome value is singular.
      testthat::expect_type(subsample_data, "list")

      # One sample data.
      data <- familiar:::test_create_synthetic_series_one_sample_data(
        outcome_type = outcome_type,
        rstream_object = r)

      # Create subsample. We expect an error because you can't do
      # cross-validation with a single sample.
      subsample_data <- testthat::expect_error(suppressWarnings(familiar:::.create_cv(
        data = data@data,
        n_folds = n_folds,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r)))
    })
  }
}

# Repeated cross-validation ----------------------------------------------------
for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    rare_outcome = TRUE,
    rstream_object = r)
  
  n_rep <- 3L
  n_folds <- 3L

  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    available_stratify_options <- c(FALSE, TRUE)
  } else {
    available_stratify_options <- FALSE
  }

  for (stratify in available_stratify_options) {
    testthat::test_that(paste0(
      "Repeated cross-validation ", ifelse(stratify, "(stratified) ", ""),
      "for ", outcome_type, " functions correctly."
    ), {
      # Create subsample.
      subsample_data <- familiar:::.create_repeated_cv(
        data = data@data,
        n_rep = 3L,
        n_folds = n_folds,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r)

      # Check that none of the training folds are the same.
      for (ii in 1:(length(subsample_data$train_list) - 1)) {
        for (jj in (ii + 1):length(subsample_data$train_list)) {
          testthat::expect_false(data.table::fsetequal(
            subsample_data$train_list[[ii]],
            subsample_data$train_list[[jj]]))
        }
      }

      for (ii in seq_along(subsample_data$train_list)) {
        # Check that there is no overlap between training folds and the
        # validation fold.
        testthat::expect_equal(
          nrow(data.table::fintersect(
            unique(subsample_data$train_list[[ii]]),
            unique(subsample_data$valid_list[[ii]]))),
          0L)

        # Check that the union of training and validation folds is the input
        # dataset.
        testthat::expect_true(data.table::fsetequal(
          unique(rbind(subsample_data$train_list[[ii]], subsample_data$valid_list[[ii]])),
          unique(data@data[, mget(familiar:::get_id_columns(id_depth = "series"))])))

        # Assert that all samples in the training and validation folds are
        # unique (not duplicated).
        testthat::expect_equal(anyDuplicated(subsample_data$train_list[[ii]]), 0)
        testthat::expect_equal(anyDuplicated(subsample_data$valid_list[[ii]]), 0)

        # Check that sampling creates a dataset identical to the development
        # subsample.
        train_data <- familiar:::select_data_from_samples(
          data = data,
          samples = subsample_data$train_list[[ii]])

        # Test that the samples and series are selected.
        testthat::expect_true(data.table::fsetequal(
          train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
          subsample_data$train_list[[ii]]))
        
        # Test that repetitions are likewise selected.
        testthat::expect_equal(
          nrow(train_data@data),
          n_rep * nrow(subsample_data$train_list[[ii]]))

        # Test that the size of the training folds is about (n_folds - 1) /
        # n_folds of the complete set.
        testthat::expect_gte(nrow(train_data@data), nrow(data@data) * (n_folds - 1) / n_folds - 9)
        testthat::expect_lte(nrow(train_data@data), nrow(data@data) * (n_folds - 1) / n_folds + 9)

        # Check that sampling creates a dataset identical to the validation
        # subsample.
        validation_data <- familiar:::select_data_from_samples(
          data = data,
          samples = subsample_data$valid_list[[ii]])

        # Test that the samples and series are selected.
        testthat::expect_true(data.table::fsetequal(
          validation_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
          subsample_data$valid_list[[ii]]))
        
        # Test that repetitions are likewise selected.
        testthat::expect_equal(
          nrow(validation_data@data),
          n_rep * nrow(subsample_data$valid_list[[ii]]))

        # Test that the size of the validation fold is about 1 / n_folds of the
        # complete set.
        testthat::expect_gte(nrow(validation_data@data), nrow(data@data) * 1 / n_folds - 9)
        testthat::expect_lte(nrow(validation_data@data), nrow(data@data) * 1 / n_folds + 9)

        # Assert that data are correctly stratified.
        if (stratify && outcome_type %in% c("binomial", "multinomial")) {
          # Determine the frequency of outcome classes in the original dataset.
          input_frequency <- data@data[, list(
            "frequency_original" = .N / nrow(data@data)), by = "outcome"]
          train_frequency <- train_data@data[, list(
            "frequency_bootstrap" = .N / nrow(train_data@data)), by = "outcome"]

          # Update the frequency table.
          frequency_table <- merge(
            x = input_frequency,
            y = train_frequency,
            by = "outcome")

          # Check that the data is correctly stratified.
          frequency_table[, "similar" := data.table::between(
            frequency_bootstrap,
            lower = frequency_original - 0.05,
            upper = frequency_original + 0.05
          )]

          testthat::expect_true(all(frequency_table$similar))
          
        } else if (stratify && outcome_type %in% c("survival")) {
          # Determine the frequency of censored data points and events in
          # classes in the original dataset.
          input_frequency <- data@data[, list(
            "frequency_original" = .N / nrow(data@data)), by = "outcome_event"]
          train_frequency <- train_data@data[, list(
            "frequency_bootstrap" = .N / nrow(train_data@data)), by = "outcome_event"]

          # Update the frequency table.
          frequency_table <- merge(
            x = input_frequency,
            y = train_frequency,
            by = "outcome_event")

          # Check that the data is correctly stratified.
          frequency_table[, "similar" := data.table::between(
            frequency_bootstrap,
            lower = frequency_original - 0.05,
            upper = frequency_original + 0.05
          )]

          testthat::expect_true(all(frequency_table$similar))
        }

        # Check that the rare outcome is found in the training data. This
        # prevent issues with training data.
        if (outcome_type == "multinomial") {
          testthat::expect_gt(nrow(train_data@data[outcome == "3"]), 0)
        }

        # Assert that all outcome levels in the validation folds also appear in
        # the training folds.
        if (outcome_type %in% c("binomial", "multinomial", "survival")) {
          testthat::expect_equal(
            length(setdiff(
              unique(validation_data@data$outcome),
              unique(train_data@data$outcome))),
            0)
        }
      }
    })
  }
}

for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  # Create synthetic dataset with one outcome.

  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    available_stratify_options <- c(FALSE, TRUE)
  } else {
    available_stratify_options <- FALSE
  }

  n_folds <- 3L

  for (stratify in available_stratify_options) {
    testthat::test_that(paste0(
      "Repeated cross-validation ", ifelse(stratify, "(stratified) ", ""),
      "for ", outcome_type, " with odd data functions correctly."
    ), {
      # One outcome-data
      data <- familiar:::test_create_synthetic_series_one_outcome(
        outcome_type = outcome_type,
        rstream_object = r)

      # Create subsample.
      subsample_data <- suppressWarnings(familiar:::.create_repeated_cv(
        data = data@data,
        n_rep = 3L,
        n_folds = n_folds,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r))
      
      # Expect a list. This is sort of a placeholder because cross-validation
      # should work even when the outcome value is singular.
      testthat::expect_type(subsample_data, "list")

      # One sample data.
      data <- familiar:::test_create_synthetic_series_one_sample_data(
        outcome_type = outcome_type,
        rstream_object = r)

      # Create subsample. We expect an error because you can't do
      # cross-validation with a single sample.
      subsample_data <- testthat::expect_error(suppressWarnings(familiar:::.create_repeated_cv(
        data = data@data,
        n_rep = 3L,
        n_folds = n_folds,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r)))
    })
  }
}



# Leave-one-out cross-validation -----------------------------------------------
for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    rare_outcome = TRUE,
    rstream_object = r)
  
  n_rep <- 3L

  testthat::test_that(paste0(
    "Leave-one-out cross-validation for ",
    outcome_type, " functions correctly."
  ), {
    # Create subsample
    subsample_data <- familiar:::.create_loocv(
      data = data@data,
      outcome_type = outcome_type,
      rstream_object = r)

    # Check that none of the training folds are the same.
    for (ii in 1:(length(subsample_data$train_list) - 1)) {
      for (jj in (ii + 1):length(subsample_data$train_list)) {
        testthat::expect_false(data.table::fsetequal(
          subsample_data$train_list[[ii]],
          subsample_data$train_list[[jj]]))
      }
    }

    # Assert (for LOOCV) that the number of run pairs is the number of samples,
    # or the number of samples - 1 for multinomial (due to pre-assignment of a
    # sample with the rare outcome level).
    if (outcome_type == "multinomial") {
      testthat::expect_equal(
        length(subsample_data$train_list),
        data.table::uniqueN(
          data@data, by = familiar:::get_id_columns(id_depth = "sample")) - 1L)
      
    } else {
      testthat::expect_equal(
        length(subsample_data$train_list),
        data.table::uniqueN(
          data@data, by = familiar:::get_id_columns(id_depth = "sample")))
    }

    # Iterate over the subsamples.
    for (ii in seq_along(subsample_data$train_list)) {
      # Check that there is no overlap between training folds and the validation
      # fold.
      testthat::expect_equal(
        nrow(data.table::fintersect(
          unique(subsample_data$train_list[[ii]]),
          unique(subsample_data$valid_list[[ii]]))),
        0L)

      # Check that the union of training and validation folds is the input
      # dataset.
      testthat::expect_true(data.table::fsetequal(
        unique(rbind(subsample_data$train_list[[ii]], subsample_data$valid_list[[ii]])),
        unique(data@data[, mget(familiar:::get_id_columns(id_depth = "series"))])))

      # Assert that all samples in the training fold are unique (not
      # duplicated).
      testthat::expect_equal(anyDuplicated(subsample_data$train_list[[ii]]), 0)

      # Assert that there is only one sample in the validation fold.
      testthat::expect_equal(
        data.table::uniqueN(subsample_data$valid_list[[ii]],
          by = familiar:::get_id_columns(id_depth = "sample")),
        1)

      # Check that sampling creates a dataset identical to the development
      # subsample.
      train_data <- familiar:::select_data_from_samples(
        data = data,
        samples = subsample_data$train_list[[ii]])

      # Test that the samples and series are selected.
      testthat::expect_true(data.table::fsetequal(
        train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
        subsample_data$train_list[[ii]]))
      
      # Test that repetitions are likewise selected.
      testthat::expect_equal(
        nrow(train_data@data),
        n_rep * nrow(subsample_data$train_list[[ii]]))

      # Assert that the size of the training set is equal to the number of
      # samples - 1.
      testthat::expect_equal(
        data.table::uniqueN(train_data@data, by = familiar:::get_id_columns(id_depth = "sample")),
        data.table::uniqueN(data@data, by = familiar:::get_id_columns(id_depth = "sample")) - 1L)
      
      # Check that sampling creates a dataset identical to the validation
      # subsample.
      validation_data <- familiar:::select_data_from_samples(
        data = data,
        samples = subsample_data$valid_list[[ii]])

      # Test that the samples and series are selected.
      testthat::expect_true(data.table::fsetequal(
        validation_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
        subsample_data$valid_list[[ii]]))
      
      # Test that repetitions are likewise selected.
      testthat::expect_equal(
        nrow(validation_data@data),
        n_rep * nrow(subsample_data$valid_list[[ii]]))

      # Test that the number of samples in the the validation fold is 1.
      testthat::expect_equal(
        data.table::uniqueN(validation_data@data, by = familiar:::get_id_columns(id_depth = "sample")),
        1L)
      
      # Check that the rare outcome is found in the training data. This prevent
      # issues with training data.
      if (outcome_type == "multinomial") {
        testthat::expect_gt(nrow(train_data@data[outcome == "3"]), 0)
      }

      # Assert that all outcome levels in the validation folds also appear in
      # the training folds.
      if (outcome_type %in% c("binomial", "multinomial", "survival")) {
        testthat::expect_equal(
          length(setdiff(
            unique(validation_data@data$outcome),
            unique(train_data@data$outcome))),
          0)
      }
    }
  })
}

for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  testthat::test_that(paste0(
    "Repeated cross-validation for ",
    outcome_type, " with odd data functions correctly."
  ), {
    # One outcome-data
    data <- familiar:::test_create_synthetic_series_one_outcome(
      outcome_type = outcome_type,
      rstream_object = r)

    # Create subsample.
    subsample_data <- suppressWarnings(familiar:::.create_loocv(
      data = data@data,
      outcome_type = outcome_type,
      rstream_object = r))
    
    # Expect a list. This is sort of a placeholder because cross-validation
    # should work even when the outcome value is singular.
    testthat::expect_type(subsample_data, "list")

    # One sample data.
    data <- familiar:::test_create_synthetic_series_one_sample_data(
      outcome_type = outcome_type,
      rstream_object = r)

    # Create subsample. We expect an error because you can't do cross-validation
    # with a single sample.
    subsample_data <- testthat::expect_error(suppressWarnings(familiar:::.create_loocv(
      data = data@data,
      outcome_type = outcome_type,
      rstream_object = r)))
  })
}

# Bootstraps -------------------------------------------------------------------
for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  data <- familiar:::test_create_synthetic_series_data(
    outcome_type = outcome_type,
    rare_outcome = TRUE,
    rstream_object = r)
  
  n_rep <- 3L

  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    available_stratify_options <- c(FALSE, TRUE)
  } else {
    available_stratify_options <- FALSE
  }

  for (stratify in available_stratify_options) {
    testthat::test_that(paste0(
      "Bootstrap resampling ", ifelse(stratify, "(stratified) ", ""),
      "for ", outcome_type, " functions correctly."
    ), {
      # Create subsample.
      subsample_data <- familiar:::.create_bootstraps(
        data = data@data,
        n_iter = 20,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r)

      # Check that none of in-bag datasets is the same.
      for (ii in 1:(length(subsample_data$train_list) - 1)) {
        for (jj in (ii + 1):length(subsample_data$train_list)) {
          testthat::expect_false(data.table::fsetequal(
            subsample_data$train_list[[ii]],
            subsample_data$train_list[[jj]]))
        }
      }

      for (ii in seq_along(subsample_data$train_list)) {
        # Check that there is no overlap between in-bag and out-of-bag data.
        testthat::expect_equal(
          nrow(data.table::fintersect(
            unique(subsample_data$train_list[[ii]]),
            unique(subsample_data$valid_list[[ii]]))),
          0L)

        # Check that the combination of in-bag and out-of-bag data is the same
        # as the input dataset.
        testthat::expect_true(data.table::fsetequal(
          unique(rbind(subsample_data$train_list[[ii]], subsample_data$valid_list[[ii]])),
          unique(data@data[, mget(familiar:::get_id_columns(id_depth = "series"))])))
        
        # Check that sampling creates a dataset identical to the development
        # dataset.
        train_data <- familiar:::select_data_from_samples(
          data = data,
          samples = subsample_data$train_list[[ii]])

        # Test that the samples and series are selected.
        testthat::expect_true(data.table::fsetequal(
          train_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
          subsample_data$train_list[[ii]]))
        
        # Test that repetitions are likewise selected.
        testthat::expect_equal(
          nrow(train_data@data),
          n_rep * nrow(subsample_data$train_list[[ii]]))

        # Test that the subsample has the same size (or is slightly larger) than
        # the original dataset. It can be slightly larger because samples with
        # rare outcomes are added to the in-bag data.
        testthat::expect_gte(nrow(train_data@data), nrow(data@data) - 30)
        testthat::expect_lte(nrow(train_data@data), nrow(data@data) + 9)

        # Check that sampling creates a dataset identical to the validation
        # dataset.
        validation_data <- familiar:::select_data_from_samples(
          data = data,
          samples = subsample_data$valid_list[[ii]])

        # Test that the samples and series are selected.
        testthat::expect_true(data.table::fsetequal(
          validation_data@data[repetition_id == 1L, mget(familiar:::get_id_columns(id_depth = "series"))],
          subsample_data$valid_list[[ii]]))
        
        # Test that repetitions are likewise selected.
        testthat::expect_equal(
          nrow(validation_data@data),
          n_rep * nrow(subsample_data$valid_list[[ii]]))

        # If stratified, check that occurrence of event or categories is similar
        # between discovery and the entire dataset.
        if (stratify && outcome_type %in% c("binomial", "multinomial")) {
          # Determine the frequency of outcome classes in the original dataset.
          input_frequency <- data@data[, list(
            "frequency_original" = .N / nrow(data@data)), by = "outcome"]
          train_frequency <- train_data@data[, list(
            "frequency_bootstrap" = .N / nrow(train_data@data)), by = "outcome"]

          # Update the frequency table.
          frequency_table <- merge(
            x = input_frequency,
            y = train_frequency,
            by = "outcome")

          # Check that the data is correctly stratified.
          frequency_table[, "similar" := data.table::between(
            frequency_bootstrap,
            lower = frequency_original - 0.05,
            upper = frequency_original + 0.05
          )]

          testthat::expect_true(all(frequency_table$similar))
          
        } else if (stratify & outcome_type %in% c("survival")) {
          # Determine the frequency of censored data points and events in
          # classes in the original dataset.
          input_frequency <- data@data[, list(
            "frequency_original" = .N / nrow(data@data)), by = "outcome_event"]
          train_frequency <- train_data@data[, list(
            "frequency_bootstrap" = .N / nrow(train_data@data)), by = "outcome_event"]

          # Update the frequency table.
          frequency_table <- merge(
            x = input_frequency,
            y = train_frequency,
            by = "outcome_event")

          # Check that the data is correctly stratified.
          frequency_table[, "similar" := data.table::between(
            frequency_bootstrap,
            lower = frequency_original - 0.05,
            upper = frequency_original + 0.05
          )]

          testthat::expect_equal(all(frequency_table$similar), TRUE)
        }

        # Check that the rare outcome is found in the training data. This
        # prevent issues with training data.
        if (outcome_type == "multinomial") {
          testthat::expect_gt(nrow(train_data@data[outcome == "3"]), 0)
        }

        # Assert that all outcome levels in the validation folds also appear in
        # the training folds.
        if (outcome_type %in% c("binomial", "multinomial", "survival")) {
          testthat::expect_equal(
            length(setdiff(
              unique(validation_data@data$outcome),
              unique(train_data@data$outcome))),
            0)
        }
      }
    })
  }
}

for (outcome_type in c("binomial", "multinomial", "continuous", "survival")) {
  # Create synthetic dataset with one outcome.

  if (outcome_type %in% c("binomial", "multinomial", "survival")) {
    available_stratify_options <- c(FALSE, TRUE)
  } else {
    available_stratify_options <- FALSE
  }

  for (stratify in available_stratify_options) {
    testthat::test_that(paste0(
      "Bootstrap ", ifelse(stratify, "(stratified) ", ""),
      "for ", outcome_type, " with odd data functions correctly."
    ), {
      # One outcome-data
      data <- familiar:::test_create_synthetic_series_one_outcome(
        outcome_type = outcome_type,
        rstream_object = r)

      # Create subsample.
      subsample_data <- familiar:::.create_bootstraps(
        data = data@data,
        n_iter = 20,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r)

      # Expect a list. This is sort of a placeholder because cross-validation
      # should work even when the outcome value is singular.
      testthat::expect_type(subsample_data, "list")

      # One sample data.
      data <- familiar:::test_create_synthetic_series_one_sample_data(
        outcome_type = outcome_type,
        rstream_object = r)

      # Create subsample. We expect an error because you can't do
      # cross-validation with a single sample.
      subsample_data <- testthat::expect_error(familiar:::.create_bootstraps(
        data = data@data,
        n_iter = 20,
        stratify = stratify,
        outcome_type = outcome_type,
        rstream_object = r))
    })
  }
}

get_baseline_survival <- function(data) {
  # Determines baseline survival based on relative risk (i.e. proportional
  # hazards models, such as Cox) This is a general method that only requires
  # that a model produces relative risks, or is calibrated to produce relative
  # risks. Based on Cox and Oakes (1984)

  # Suppress NOTES due to non-standard evaluation in data.table
  time <- km_survival_var <- NULL

  if (!is(data, "dataObject")) {
    ..error_reached_unreachable_code(
      "get_baseline_survival: data is not a dataObject object."
    )
  }

  if (!data@outcome_type %in% c("survival")) {
    ..error_reached_unreachable_code(
      "get_baseline_survival: outcome_type is not survival."
    )
  }

  # Extract relevant information regarding survival.
  survival_data <- unique(data@data[, mget(c(
    get_id_columns(id_depth = "series"), "outcome_time", "outcome_event"
  ))])

  # Make a local copy.
  survival_data <- data.table::copy(survival_data)

  # Get the survival estimate from a Kaplan-Meier fit.
  km_fit <- survival::survfit(
    Surv(outcome_time, outcome_event) ~ 1,
    data = survival_data
  )

  # Add censoring rate
  cens_fit <- survival::survfit(
    Surv(outcome_time, outcome_event == 0L) ~ 1,
    data = survival_data
  )

  # Complete a Kaplan-Meier table including censoring rates.
  kaplan_meier_table <- data.table::data.table(
    "time" = km_fit$time,
    "km_survival" = km_fit$surv,
    "km_survival_var" = km_fit$std.err^2.0,
    "cens_distr" = cens_fit$surv
  )

  # Add time 0.
  if (min(kaplan_meier_table$time) > 0.0) {
    kaplan_meier_table <- rbind(
      kaplan_meier_table,
      data.table::data.table(
        "time" = 0.0,
        "km_survival" = 1.0,
        "km_survival_var" = 0.0,
        "cens_distr" = 1.0
      )
    )
  }

  # Replace inf variance by 1.0
  kaplan_meier_table[
    is.infinite(km_survival_var),
    "km_survival_var" := 1.0
  ]

  # Sort by time.
  kaplan_meier_table <- kaplan_meier_table[order(time)]

  return(kaplan_meier_table)
}


# .survival_probability_relative_risk (generic) --------------------------------
setGeneric(
  ".survival_probability_relative_risk",
  function(object, ...) setGeneric(".survival_probability_relative_risk")
)


# .survival_probability_relative_risk (model) ----------------------------------
setMethod(
  ".survival_probability_relative_risk",
  signature(object = "familiarModel"),
  function(
    object,
    data,
    time,
    ...
  ) {
    if (!is(data, "dataObject")) {
      ..error_reached_unreachable_code(
        ".survival_probability_relative_risk: object is not a dataObject object."
      )
    }
    
    # Predict relative risks.
    prediction_table <- .predict(
      object = object,
      data = data,
      type = "default",
      allow_recalibration = TRUE,
      time = time
    )
    
    # Convert to probability.
    prediction_table <- .survival_probability_relative_risk(
      object = prediction_table,
      data = data,
      time = time,
      model = object,
      ...
    )
    
    return(prediction_table)
  }
)


# .survival_probability_relative_risk (prediction_table) -----------------------
setMethod(
  ".survival_probability_relative_risk",
  signature(object = "predictionTableSurvivalHazardRatio"),
  function(
    object,
    data,
    time,
    model,
    ...
  ) {
    
    if (!is(data, "dataObject")) {
      ..error_reached_unreachable_code(
        ".survival_probability_relative_risk: object is not a dataObject object."
      )
    }
    
    # Check for several issues that prevent survival probabilities from being
    # predicted.
    if (is_empty(object)) return(NULL)
    if (!any_predictions_valid(object)) return(NULL)
    if (!has_calibration_info(model)) return(NULL)
    
    # Survival in the group is based on proportional hazards assumption, and
    # uses baseline cumulative hazard and the group's predicted relative risks.
    # This evaluation comes in handy when performing, e.g. the Nam-D'Agostino
    # test. It avoids recalculating the baseline hazard. Following Demler,
    # Paynter and Cook. (Stat. Med. 2015), we compute the survival probability
    # at t=time_max for each sample.
    survival_probabilities <- ..survival_probability_relative_risk(
      object = model,
      relative_risk = object@prediction_data$predicted_outcome,
      time = time
    )
    
    # Create prediction table.
    prediction_table <- as_prediction_table(
      x = survival_probabilities,
      type = "survival_probability",
      data = data,
      time = time,
      model_object = model
    )
    
    return(prediction_table)
  }
)



..survival_probability_relative_risk <- function(object, relative_risk, time) {
  # Survival in the group is based on proportional hazards assumption, and
  # uses baseline cumulative hazard and the group's predicted relative risks.
  # This evaluation comes in handy when performing, e.g. the Nam-D'Agostino
  # test. It avoids recalculating the baseline hazard. Following Demler,
  # Paynter and Cook. (Stat. Med. 2015), we compute the survival probability
  # at t=time_max for each sample.

  # Interpolate the baseline survival function at the fitting times
  baseline_surv <- stats::approx(
    x = object@calibration_info$time,
    y = object@calibration_info$km_survival,
    xout = time,
    method = "linear",
    rule = 2L
  )$y

  # Create a n_sample x n_times matrix
  return(sapply(
    relative_risk, 
    function(rr, s0) (s0^rr), 
    s0 = baseline_surv
  ))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarDataElementHyperparameters object ------------------------------------
setClass(
  "familiarDataElementHyperparameters",
  contains = "familiarDataElement",
  prototype = methods::prototype(
    detail_level = "hybrid",
    estimation_type = "point"
  )
)



# extract_hyperparameters (generic) --------------------------------------------

#'@title Internal function to extract hyperparameters from models.
#'
#'@description Collects hyperparameters from models in a `familiarEnsemble`.
#'
#'@inheritParams .extract_data
#'
#'@return A list of familiarDataElements with hyperparameters.
#'@md
#'@keywords internal
setGeneric(
  "extract_hyperparameters",
  function(
    object,
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_hyperparameters")
  }
)



# extract_hyperparameters (familiarEnsemble) -----------------------------------
setMethod(
  "extract_hyperparameters",
  signature(object = "familiarEnsemble"),
  function(
    object,
    message_indent = 0L,
    verbose = FALSE
  ) {
    # Extracts hyper-parameters from each model and collects them.
    
    # Message extraction start
    logger_message(
      paste0("Extracting hyperparameters from the models in the ensemble."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Test if models are properly loaded
    if (!is_model_loaded(object = object)) ..error_ensemble_models_not_loaded()
    
    # Test if the any of the models in the ensemble were trained.
    if (!model_is_trained(object)) return(NULL)
    
    proto_data_element <- methods::new("familiarDataElementHyperparameters")
    
    # Generate elements to send to dispatch.
    hyperparameter_data <- extract_dispatcher(
      FUN = .extract_hyperparameters,
      cl = NULL,
      has_internal_bootstrap = FALSE,
      object = object,
      proto_data_element = proto_data_element,
      aggregate_results = FALSE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(hyperparameter_data)
  }
)



# extract_hyperparameters (prediction table) -----------------------------------
setMethod(
  "extract_hyperparameters",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("model hyperparameters")
    
    return(NULL)
  }
)



.extract_hyperparameters <- function(
    object,
    proto_data_element,
    ...
) {
  
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  data_element <- add_model_name(proto_data_element, object = object)
  
  # Test if the any of the models in the ensemble were trained.
  if (!model_is_trained(object)) return(NULL)
  
  # Parse hyperparameters as data.table
  data_element@data <- data.table::as.data.table(object@hyperparameters)
  
  # Set value columns
  data_element@value_column <- names(object@hyperparameters)
  
  return(data_element)
}



..hyperparameter_to_string <- function(x) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  instance <- n <- NULL
  
  if (is.character(x) || is.factor(x) || is.logical(x)) {
    
    # Count instances and sort by number.
    data <- data.table::data.table("instance" = x)
    data <- data[, list(n = .N), by = instance][order(n)]
    
    # Parse to a vector
    export_vec <- sapply(
      seq_len(nrow(data)),
      function(ii, data) (paste0(data$instance[ii], " (", data$n[ii], ")")),
      data = data
    )
    
    export_vec <- paste(export_vec, collapse = "; ")
    
  } else if (is.numeric(x)) {
    export_vec <- paste0(
      stats::quantile(x = x, probs = 0.5, na.rm = TRUE, type = 1L, names = FALSE),
      " [",
      min(x, na.rm = TRUE), ", ",
      max(x, na.rm = TRUE), "]"
    )
  }
  
  return(export_vec)
}



# .identifier_as_data_attribute (familiarDataElementHyperparameters) -----------
setMethod(
  ".identifier_as_data_attribute",
  signature(x = "familiarDataElementHyperparameters"),
  function(
    x,
    identifier,
    as_grouping_column = TRUE
  ) {
    
    if (length(identifier) == 0L) {
      ..error_reached_unreachable_code(
        ".identifier_as_data_attribute: Cannot pass an empty identifier."
      )
    }
    
    # Different learners have different hyperparameters. We therefore
    # keep splitting by learner.
    if (any(identifier == "all")) {
      
      identifier <- names(x@identifiers)
      
      # Remove learner, if present.
      identifier <- setdiff(identifier, "learner")
      
      if (length(identifier) == 0L) return(x)
      
      return(callNextMethod(
        x = x,
        identifier = identifier,
        as_grouping_column = as_grouping_column
      ))
      
    } else {
      return(callNextMethod())
    }
  }
)


# ..compute_data_elements_estimates (familiarDataElementHyperparameters)--------
setMethod(
  "..compute_data_element_estimates",
  signature(x = "familiarDataElementHyperparameters"),
  function(
    x,
    x_list = NULL,
    ...
  ) {
    
    # It might be that x was only used to direct to this method.
    if (!is.null(x_list)) x <- x_list
    if (!is.list(x)) x <- list(x)
    
    # Remove empty entries.
    x <- x[!sapply(x, is_empty)]
    if (is_empty(x)) return(NULL)
    
    # Collect all data.
    data <- data.table::rbindlist(
      lapply(x, function(x) (x@data)),
      use.names = TRUE,
      fill = TRUE
    )
    
    # Split by vimp_method
    data <- split(data, by = "vimp_method")
    
    learner <- x[[1L]]@identifiers$learner
    if (is.null(learner)) learner <- x[[1L]]@data$learner[1L]
    
    # Set learner.
    parameter_string <- paste0("learner\t", learner)
    
    for (current_data in data) {
      # Determine the vimp_method
      vimp_method <- current_data$vimp_method[1L]
      
      # Set vimp method.
      parameter_string <- c(
        parameter_string,
        paste0("vimp_method\t", vimp_method),
        "---------------------"
      )
      
      # Parse data.
      parameter_string <- c(
        parameter_string,
        sapply(
          x[[1L]]@value_column,
          function(hyperparameter, data) {
            return(paste0(
              hyperparameter, "\t",
              ..hyperparameter_to_string(data[[hyperparameter]])
            ))
          },
          data = current_data
        )
      )
      
      parameter_string <- c(parameter_string, " ")
      
    }
    
    # Collapse to text.
    parameter_string <- paste0(parameter_string, collapse = "\n")
    
    # Copy data element.
    y <- x[[1L]]
    y@data <- parameter_string
    
    # Update value column
    y@value_column <- NA_character_
    
    return(y)
  }
)



# .export (familiarDataElementHyperparameters) ---------------------------------
setMethod(
  ".export",
  signature(x = "familiarDataElementHyperparameters"),
  function(
    x,
    x_list, 
    aggregate_results = FALSE, 
    ...
  ) {
    # This is like .export,familiarDataElement, but the elements are
    # merged prior to computing estimates.
    
    # Merge data elements.
    x <- merge_data_elements(
      x = x_list,
      as_data = "all",
      as_grouping_column = TRUE,
      force_data_table = TRUE
    )
    
    if (aggregate_results) {
      x <- .compute_data_element_estimates(x)
    }
    
    return(x)
  }
)



# export_hyperparameters (generic) ---------------------------------------------

#'@title Extract and export model hyperparameters.
#'
#'@description Extract and export model hyperparameters from models in a
#'  familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams as_familiar_collection
#'
#'@details Data, such as model performance and calibration information, is
#'  usually collected from a `familiarCollection` object. However, you can also
#'  provide one or more `familiarData` objects, that will be internally
#'  converted to a `familiarCollection` object. It is also possible to provide a
#'  `familiarEnsemble` or one or more `familiarModel` objects together with the
#'  data from which data is computed prior to export. Paths to the previous
#'  files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Many model hyperparameters are optimised using sequential model-based
#'  optimisation. The extracted hyperparameters are those that were selected to
#'  construct the underlying models (`familiarModel` objects).
#'
#'@return A data.table (if `dir_path` is not provided), or nothing, as all data
#'  is exported to `csv` files. In case of the latter, hyperparameters are
#'  summarised.
#'@exportMethod export_hyperparameters
#'@md
#'@rdname export_hyperparameters-methods
setGeneric(
  "export_hyperparameters",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_hyperparameters")
  }
)



# export_hyperparameters (collection) ------------------------------------------

#'@rdname export_hyperparameters-methods
setMethod(
  "export_hyperparameters",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    if (aggregate_results) {
      subtype <- "summary"
      
    } else {
      subtype <- NULL
    }
    
    return(.export(
      x = object,
      data_slot = "hyperparameters",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "hyperparameter",
      subtype = subtype,
      export_collection = export_collection
    ))
  }
)



# export_hyperparameters (general) ---------------------------------------------

#'@rdname export_hyperparameters-methods
setMethod(
  "export_hyperparameters",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "hyperparameters"
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_hyperparameters,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_results" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# familiarDataElementRiskStratificationInfo object -----------------------------
setClass(
  "familiarDataElementRiskStratificationInfo",
  contains = "familiarDataElement",
  prototype = methods::prototype(estimation_type = "point")
)

# extract_risk_stratification_info (generic) -----------------------------------

#'@title Internal function to extract risk stratification info from data.
#'
#'@description Collects risk stratification information.
#'
#'@inheritParams .extract_data
#'
#'@return A list of familiarDataElements with risk stratification information.
#'@md
#'@keywords internal
setGeneric(
  "extract_risk_stratification_info",
  function(
    object,
    detail_level = waiver(),
    message_indent = 0L,
    verbose = FALSE,
    ...
  ) {
    standardGeneric("extract_risk_stratification_info")
  }
)



# extract_risk_stratification_info (familiarEnsemble) --------------------------
setMethod(
  "extract_risk_stratification_info",
  signature(object = "familiarEnsemble"),
  function(
    object,
    detail_level = waiver(),
    message_indent = 0L,
    verbose = FALSE
  ) {
    
    # Test if the outcome type is survival. Other outcome types do not have risk
    # stratification data associated with them.
    if (!object@outcome_type %in% c("survival")) return(NULL)
    
    # Message extraction start
    logger_message(
      paste0("Extracting stratification information."),
      indent = message_indent,
      verbose = verbose
    )
    
    # Check the level detail.
    detail_level <- .parse_detail_level(
      x = detail_level,
      object = object,
      default = "hybrid",
      data_element = "risk_stratification_info"
    )
    
    proto_data_element <- methods::new(
      "familiarDataElementRiskStratificationInfo",
      detail_level = detail_level
    )
    
    # Generate elements to send to dispatch.
    stratification_info <- extract_dispatcher(
      FUN = .extract_risk_stratification_info,
      cl = NULL,
      has_internal_bootstrap = FALSE,
      object = object,
      proto_data_element = proto_data_element,
      aggregate_results = FALSE,
      message_indent = message_indent + 1L,
      verbose = verbose
    )
    
    return(stratification_info)
  }
)



# extract_risk_stratification_info (prediction table) --------------------------
setMethod(
  "extract_risk_stratification_info",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    ..warning_no_data_extraction_from_prediction_table("risk stratification info")
    
    return(NULL)
  }
)



.extract_risk_stratification_info <- function(
    object,
    proto_data_element,
    ...
) {
  # Ensure that the object is loaded
  object <- load_familiar_object(object)
  
  # Add model name.
  data_element <- add_model_name(proto_data_element, object = object)
  
  # Test if the any of the models in the ensemble were trained.
  if (!model_is_trained(object)) return(NULL)
  
  if (is(object, "familiarModel")) {
    # Iterate over stratification parameters
    data <- lapply(
      object@km_info$parameters,
      function(x) {
        # Extract risk stratification information.
        data <- data.table::data.table(
          "stratification_method" = x$method,
          "cutoff" = x$cutoff,
          "group_id" = seq_len(length(x$cutoff))
        )
        
        return(data)
      }
    )
    
    # Combine to single list
    data <- data.table::rbindlist(data, use.names = TRUE)
    if (is_empty(data)) return(NULL)
    
  } else {
    # Compute risk stratification data.
    risk_stratification_data <- extract_risk_stratification_info(
      object = object,
      detail_level = "hybrid",
      verbose = FALSE
    )
    
    risk_stratification_data <- .compute_data_element_estimates(risk_stratification_data)
    if (is_empty(risk_stratification_data)) return(NULL)
    
    # Extract data.
    data <- risk_stratification_data[[1L]]@data
  }
  
  # Attach data to the corresponding attribute.
  data_element@data <- data
  
  # Set value columns
  data_element@value_column <- c("cutoff")
  
  # Set grouping columns
  data_element@grouping_column <- c("stratification_method", "group_id")
  
  return(data_element)
}



# export_risk_stratification_info (generic) ------------------------------------ 

#'@title Extract and export cut-off values for risk group stratification.
#'
#'@description Extract and export cut-off values for risk group stratification
#'  by models in a familiarCollection.
#'
#'@inheritParams export_all
#'@inheritParams export_univariate_analysis_data
#'
#'@inheritDotParams as_familiar_collection
#'
#'@details Data is usually collected from a `familiarCollection` object.
#'  However, you can also provide one or more `familiarData` objects, that will
#'  be internally converted to a `familiarCollection` object. It is also
#'  possible to provide a `familiarEnsemble` or one or more `familiarModel`
#'  objects together with the data from which data is computed prior to export.
#'  Paths to the previous files can also be provided.
#'
#'  All parameters aside from `object` and `dir_path` are only used if `object`
#'  is not a `familiarCollection` object, or a path to one.
#'
#'  Stratification cut-off values are determined when creating a model, using
#'  one of several methods set by the `stratification_method` parameter. These
#'  values are then used to stratify samples in any new dataset. The available
#'  methods are:
#'
#'  * `median` (default): The median predicted value in the development cohort
#'  is used to stratify the samples into two risk groups.
#'
#'  * `fixed`: Samples are stratified based on the sample quantiles of the
#'  predicted values. These quantiles are defined using the
#'  `stratification_threshold` parameter.
#'
#'  * `optimised`: Use maximally selected rank statistics to determine the
#'  optimal threshold (Lausen and Schumacher, 1992; Hothorn et al., 2003) to
#'  stratify samples into two optimally separated risk groups.
#'
#'@return A data.table (if `dir_path` is not provided), or nothing, as all data
#'  is exported to `csv` files.
#'@references 1. Lausen, B. & Schumacher, M. Maximally Selected Rank Statistics.
#'  Biometrics 48, 73 (1992).
#'
#'  1. Hothorn, T. & Lausen, B. On the exact distribution of maximally selected
#'  rank statistics. Comput. Stat. Data Anal. 43, 121–137 (2003).
#'@exportMethod export_risk_stratification_info
#'@md
#'@rdname export_risk_stratification_info-methods
setGeneric(
  "export_risk_stratification_info",
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    standardGeneric("export_risk_stratification_info")
  }
)



# export_risk_stratification_info (collection) ---------------------------------

#'@rdname export_risk_stratification_info-methods
setMethod(
  "export_risk_stratification_info",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    return(.export(
      x = object,
      data_slot = "km_info",
      dir_path = dir_path,
      aggregate_results = aggregate_results,
      type = "stratification",
      export_collection = export_collection
    ))
  }
)



# export_risk_stratification_info (general) ------------------------------------

#'@rdname export_risk_stratification_info-methods
setMethod(
  "export_risk_stratification_info",
  signature(object = "ANY"),
  function(
    object,
    dir_path = NULL,
    aggregate_results = TRUE,
    export_collection = FALSE,
    ...
  ) {
    
    # Attempt conversion to familiarCollection object.
    object <- do.call(
      as_familiar_collection,
      args = c(
        list(
          "object" = object,
          "data_element" = "risk_stratification_info"
        ),
        list(...)
      )
    )
    
    return(do.call(
      export_risk_stratification_info,
      args = c(
        list(
          "object" = object,
          "dir_path" = dir_path,
          "aggregate_resuls" = aggregate_results,
          "export_collection" = export_collection
        ),
        list(...)
      )
    ))
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# featureInfoParametersNormalisationNone ---------------------------------------
setClass(
  "featureInfoParametersNormalisationNone",
  contains = "featureInfoParameters",
  slots = list(
    "method" = "character",
    "batch" = "ANY",
    "reason" = "ANY"
  ),
  prototype = list(
    "method" = "none",
    "batch" = NULL,
    "reason" = NULL
  )
)



# featureInfoParametersNormalisationShiftScale ---------------------------------
setClass(
  "featureInfoParametersNormalisationShiftScale",
  contains = "featureInfoParameters",
  slots = list(
    "batch" = "ANY",
    "shift" = "numeric",
    "scale" = "numeric"
  ),
  prototype = list(
    "batch" = NULL,
    "shift" = NA_real_,
    "scale" = NA_real_
  )
)



# featureInfoParametersNormalisationShift --------------------------------------
setClass(
  "featureInfoParametersNormalisationShift",
  contains = "featureInfoParameters",
  slots = list(
    "batch" = "ANY",
    "shift" = "numeric"
  ),
  prototype = list(
    "batch" = NULL,
    "shift" = NA_real_
  )
)



# featureInfoParametersNormalisationStandardisation ----------------------------
setClass(
  "featureInfoParametersNormalisationStandardisation",
  contains = "featureInfoParametersNormalisationShiftScale",
  slots = list("method" = "character"),
  prototype = list("method" = NA_character_)
)



# featureInfoParametersNormalisationQuantile -----------------------------------
setClass(
  "featureInfoParametersNormalisationQuantile",
  contains = "featureInfoParametersNormalisationShiftScale",
  slots = list("method" = "character"),
  prototype = list("method" = NA_character_)
)



# featureInfoParametersNormalisationNormalisation ------------------------------
setClass(
  "featureInfoParametersNormalisationNormalisation",
  contains = "featureInfoParametersNormalisationShiftScale",
  slots = list("method" = "character"),
  prototype = list("method" = NA_character_)
)



# featureInfoParametersNormalisationMeanCentering ------------------------------
setClass(
  "featureInfoParametersNormalisationMeanCentering",
  contains = "featureInfoParametersNormalisationShift",
  slots = list("method" = "character"),
  prototype = list("method" = NA_character_)
)



.get_available_none_normalisation_methods <- function() {
  return("none")
}

.get_available_standardisation_normalisation_methods <- function() {
  return(c(
    "standardisation", "standardisation_trim",
    "standardisation_winsor", "standardisation_robust"
  ))
}

.get_available_quantile_normalisation_methods <- function() {
  return(c("quantile"))
}

.get_available_normalisation_normalisation_methods <- function() {
  return(c("normalisation", "normalisation_trim", "normalisation_winsor"))
}

.get_available_mean_centering_normalisation_methods <- function() {
  return(c("mean_centering"))
}

.get_available_normalisation_methods <- function(type = "all") {
  available_normalisation_method <- NULL

  if (type %in% c("none", "all")) {
    available_normalisation_method <- c(
      available_normalisation_method,
      .get_available_none_normalisation_methods()
    )
  }

  if (type %in% c("all", "standardisation")) {
    available_normalisation_method <- c(
      available_normalisation_method,
      .get_available_standardisation_normalisation_methods()
    )
  }

  if (type %in% c("all", "quantile")) {
    available_normalisation_method <- c(
      available_normalisation_method,
      .get_available_quantile_normalisation_methods()
    )
  }

  if (type %in% c("all", "mean_centering")) {
    available_normalisation_method <- c(
      available_normalisation_method,
      .get_available_mean_centering_normalisation_methods()
    )
  }

  if (type %in% c("all", "normalisation")) {
    available_normalisation_method <- c(
      available_normalisation_method,
      .get_available_normalisation_normalisation_methods()
    )
  }

  return(available_normalisation_method)
}



create_normalisation_parameter_skeleton <- function(
    feature_info_list,
    feature_names = NULL,
    batch = NULL,
    normalisation_method,
    normalisation_shift = NULL,
    normalisation_scale = NULL,
    .override_existing = FALSE
) {
  # Creates a skeleton for the provided normalisation method. If
  # normalisation_shift and / or normalisation_scale are provided (typically
  # not), these values are updated as well.

  # Determine feature names from the feature info list, if provided.
  if (is.null(feature_names)) feature_names <- names(feature_info_list)

  # Select only features that appear in the feature info list.
  feature_names <- intersect(names(feature_info_list), feature_names)

  # Skip step if no feature info objects are updated.
  if (is_empty(feature_names)) return(feature_info_list)

  # Check that method is applicable.
  .check_parameter_value_is_valid(
    x = normalisation_method,
    var_name = "normalisation_method",
    values = .get_available_normalisation_methods()
  )

  # Check that normalisation_shift is numeric.
  if (!is.null(normalisation_shift)) {
    .check_number_in_valid_range(
      x = normalisation_shift,
      var_name = "normalisation_shift",
      range = c(-Inf, Inf)
    )
  }

  # Check that normalisation_scale is numeric.
  if (!is.null(normalisation_scale)) {
    .check_number_in_valid_range(
      x = normalisation_scale,
      var_name = "normalisation_scale",
      range = c(-Inf, Inf)
    )
  }

  # Update familiar info objects with a feature normalisation skeleton.
  updated_feature_info <- fam_lapply(
    X = feature_info_list[feature_names],
    FUN = .create_normalisation_parameter_skeleton,
    method = normalisation_method,
    batch = batch,
    shift = normalisation_shift,
    scale = normalisation_scale,
    .override_existing = .override_existing
  )

  # Provide names for the updated feature info objects.
  names(updated_feature_info) <- feature_names

  # Replace list elements.
  feature_info_list[feature_names] <- updated_feature_info

  return(feature_info_list)
}



.create_normalisation_parameter_skeleton <- function(
    feature_info,
    method,
    batch = NULL,
    shift = NULL,
    scale = NULL,
    .override_existing = FALSE
) {
  # Check if normalisation data was already completed, and does not require
  # being determined anew.
  if (feature_info_complete(feature_info@normalisation_parameters) && !.override_existing) {
    return(feature_info)
  }

  # Pass to underlying function that constructs the skeleton.
  object <- ..create_normalisation_parameter_skeleton(
    feature_name = feature_info@name,
    feature_type = feature_info@feature_type,
    available = is_available(feature_info),
    method = method,
    batch = batch,
    shift = shift,
    scale = scale
  )

  # Update normalisation_parameters slot.
  feature_info@normalisation_parameters <- object

  return(feature_info)
}



..create_normalisation_parameter_skeleton <- function(
    feature_name,
    feature_type = "numeric",
    available = TRUE,
    method,
    batch = NULL,
    shift = NULL,
    scale = NULL
) {
  # This is the lowest level function for creating normalisation parameter
  # skeletons.

  # Create the relevant objects.
  if (feature_type != "numeric") {
    object <- methods::new(
      "featureInfoParametersNormalisationNone",
      reason = "not a numeric feature"
    )
    
  } else if (!available) {
    object <- methods::new(
      "featureInfoParametersNormalisationNone",
      reason = "feature was omitted prior to transformation"
    )
    
  } else if (method %in% .get_available_none_normalisation_methods()) {
    object <- methods::new("featureInfoParametersNormalisationNone")
    
  } else if (method %in% .get_available_standardisation_normalisation_methods()) {
    object <- methods::new(
      "featureInfoParametersNormalisationStandardisation",
      "method" = method
    )
    
  } else if (method %in% .get_available_quantile_normalisation_methods()) {
    object <- methods::new(
      "featureInfoParametersNormalisationQuantile",
      "method" = method
    )
    
  } else if (method %in% .get_available_normalisation_normalisation_methods()) {
    object <- methods::new(
      "featureInfoParametersNormalisationNormalisation",
      "method" = method
    )
    
  } else if (method %in% .get_available_mean_centering_normalisation_methods()) {
    object <- methods::new(
      "featureInfoParametersNormalisationMeanCentering",
      "method" = method
    )
    
  } else if (method %in% .get_available_combat_parametric_normalisation_methods()) {
    object <- methods::new(
      "featureInfoParametersNormalisationParametricCombat",
      "method" = method
    )
    
  } else if (method %in% .get_available_combat_non_parametric_normalisation_methods()) {
    object <- methods::new(
      "featureInfoParametersNormalisationNonParametricCombat",
      "method" = method
    )
    
  } else {
    ..error_reached_unreachable_code(paste0(
      "..create_normalisation_parameter_skeleton: encountered an unknown ",
      "normalisation method: ", paste_s(method)
    ))
  }

  # Set the name of the object.
  object@name <- feature_name

  # Add batch, if provided.
  if (!is.null(batch)) object@batch <- batch

  # Add shift and/or scale, when provided.
  if (!is.null(shift) || !is.null(scale)) {
    object <- add_feature_info_parameters(
      object,
      data = NULL,
      batch = batch,
      shift = shift,
      scale = scale
    )
  }

  # Update the familiar version.
  object <- add_package_version(object = object)

  return(object)
}



add_normalisation_parameters <- function(
    cl = NULL,
    feature_info_list,
    data,
    verbose = FALSE
) {
  # Determine normalisation parameters and add them to the feature_info_list.

  # Find feature columns.
  feature_names <- get_feature_columns(x = data)

  # Sanity check.
  if (!(setequal(feature_names, get_available_features(feature_info_list = feature_info_list)))) {
    ..error_reached_unreachable_code(paste0(
      "add_normalisation_parameters: features in data and the feature info list ",
      "are expect to be the same, but were not."
    ))
  }

  # Iterate over features.
  updated_feature_info <- fam_mapply(
    cl = cl,
    FUN = .add_normalisation_parameters,
    feature_info = feature_info_list[feature_names],
    data = data@data[, mget(feature_names)],
    progress_bar = verbose,
    chopchop = TRUE
  )

  # Provide names for the updated feature info objects.
  names(updated_feature_info) <- feature_names

  # Replace list elements.
  feature_info_list[feature_names] <- updated_feature_info

  return(feature_info_list)
}



.add_normalisation_parameters <- function(
    feature_info,
    data
) {
  # Pass to underlying function that adds the feature info.
  object <- add_feature_info_parameters(
    object = feature_info@normalisation_parameters,
    data = data
  )

  # Update normalisation_parameters slot.
  feature_info@normalisation_parameters <- object

  return(feature_info)
}



# initialize (none) ------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "featureInfoParametersNormalisationNone"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # The parameter set is by definition complete when no normalisation is
    # performed.
    .Object@complete <- TRUE

    return(.Object)
  }
)



# add_feature_info_parameters (none, NULL) -------------------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationNone",
    data = "NULL"
  ),
  function(
    object,
    data,
    batch = NULL,
    ...
  ) {
    
    if (!is.null(batch)) object@batch <- batch

    return(object)
  }
)



# add_feature_info_parameters (shift normalisation, NULL) ----------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationShift",
    data = "NULL"
  ),
  function(
    object,
    data,
    batch = NULL,
    shift = NULL,
    ...
  ) {
    
    if (!is.null(batch)) {
      object@batch <- batch
    } else if (!is.null(object@batch)) {
      # If batch has been set as part of the object, get it here.
      batch <- object@batch
    }
    
    if (is.numeric(shift)) {
      if (is.finite(shift)) {
        # Shift is numeric, and not NA. This is typical when shift is set
        # externally. We then update the shift parameter.
        object@shift <- shift
        object@complete <- TRUE

        return(object)
        
      } else {
        # Shift is numeric, but NA or Inf. We then return a none-class
        # normalisation instead. This is typical when updating versions prior to
        # familiar v1.2.0.
        object <- ..create_normalisation_parameter_skeleton(
          feature_name = object@name,
          batch = batch,
          method = "none"
        )

        object@reason <- "shift was NA or infinite"

        return(object)
      }
    }

    # If shift is not set, but data is NULL, shift cannot be determined.
    if (is.null(shift)) {
      object <- ..create_normalisation_parameter_skeleton(
        feature_name = object@name,
        batch = batch,
        method = "none"
      )

      object@reason <- "insufficient data to determine shift"

      return(object)
    }

    # If shift is not numeric, it can still be NA. We then return a none-class
    # normalisation instead. This is typical when updating versions prior to
    # familiar v1.2.0.
    if (is.na(shift)) {
      object <- ..create_normalisation_parameter_skeleton(
        feature_name = object@name,
        batch = batch,
        method = "none"
      )

      object@reason <- "shift was NA"

      return(object)
    }

    # Any other reasons why shift cannot be set directly.
    object <- ..create_normalisation_parameter_skeleton(
      feature_name = object@name,
      batch = batch,
      method = "none"
    )

    object@reason <- "shift could not be determined for an unknown reason"

    return(object)
  }
)



# add_feature_info_parameters (shift normalisation, ANY) -----------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationShift",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # This method is targeted when directly determining the normalisation
    # parameters. It is usually called from the child functions.

    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)

    # Check if data is not empty, and return none-class normalisation
    # object, if it is. This is done by calling the method with
    # signature data=NULL to handle setting the data.
    if (is_empty(data)) {
      return(add_feature_info_parameters(object = object, data = NULL))
    }

    # Remove non-finite values from data.
    data <- data[is.finite(data)]

    # Again, check if data is not empty.
    if (is_empty(data)) {
      return(add_feature_info_parameters(object = object, data = NULL))
    }

    # Check that at least three unique values are present.
    if (length(unique(data)) <= 3L) {
      return(add_feature_info_parameters(object = object, data = NULL))
    }

    # For batch normalisation, check that at least five instances are present.
    if (!is.null(object@batch)) {
      if (length(data) < 5L) {
        return(add_feature_info_parameters(object = object, data = NULL))
      }
    }

    return(object)
  }
)



# add_feature_info_parameters (shift and scale normalisation, NULL) --------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationShiftScale",
    data = "NULL"
  ),
  function(
    object,
    data,
    batch = NULL,
    shift = NULL,
    scale = NULL,
    ...
  ) {
    if (!is.null(batch)) {
      object@batch <- batch
    } else if (!is.null(object@batch)) {
      # If batch has been set as part of the object, get it here.
      batch <- object@batch
    }

    if (is.numeric(shift) && is.numeric(scale)) {
      if (is.finite(shift) && is.finite(scale)) {
        # Shift and scale are numeric, and not NA. This is typical when shift
        # and scale are set externally. We then update the shift and scale
        # parameters.
        object@shift <- shift
        object@scale <- scale
        object@complete <- TRUE

        return(object)
        
      } else {
        # Shift and scale are numeric, but NA or Inf. We then return a
        # none-class normalisation instead. This is typical when updating
        # versions prior to familiar v1.2.0.
        object <- ..create_normalisation_parameter_skeleton(
          feature_name = object@name,
          batch = batch,
          method = "none"
        )

        object@reason <- "shift or scale were NA or infinite"

        return(object)
      }
    }

    # If shift or scale are not set, but data is NULL, so shift and scale cannot
    # be determined.
    if (is.null(shift) || is.null(scale)) {
      object <- ..create_normalisation_parameter_skeleton(
        feature_name = object@name,
        batch = batch,
        method = "none"
      )

      object@reason <- "insufficient data to determine shift and scale"

      return(object)
    }

    # If shift and scale are not numeric, they can still be NA. We then return a
    # none-class normalisation instead. This is typical when updating versions
    # prior to familiar v1.2.0.
    if (is.na(shift) || is.na(scale)) {
      object <- ..create_normalisation_parameter_skeleton(
        feature_name = object@name,
        batch = batch,
        method = "none"
      )

      object@reason <- "shift or scale were NA"

      return(object)
    }

    # Any other reasons why shift and scale cannot be set directly.
    object <- ..create_normalisation_parameter_skeleton(
      feature_name = object@name,
      batch = batch,
      method = "none"
    )

    object@reason <- "shift and scale could not be determined for an unknown reason"

    return(object)
  }
)



# add_feature_info_parameters (shift and scale normalisation, ANY) -------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationShiftScale",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # This method is targeted when directly determining the normalisation
    # parameters. It is usually called from the child functions.

    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)

    # Check if data is not empty, and return none-class normalisation object, if
    # it is. This is done by calling the method with signature data=NULL to
    # handle setting the data.
    if (is_empty(data)) {
      return(add_feature_info_parameters(object = object, data = NULL))
    }

    # Remove non-finite values from data.
    data <- data[is.finite(data)]

    # Again, check if data is not empty.
    if (is_empty(data)) {
      return(add_feature_info_parameters(object = object, data = NULL))
    }

    # Check that at least three unique values are present.
    if (length(unique(data)) <= 3L) {
      return(add_feature_info_parameters(object = object, data = NULL))
    }

    # For batch normalisation, check that at least five instances are present.
    if (!is.null(object@batch)) {
      if (length(data) < 5L) {
        return(add_feature_info_parameters(object = object, data = NULL))
      }
    }

    return(object)
  }
)



# add_feature_info_parameters (standardisation, data.table) --------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationStandardisation",
    data = "data.table"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Pass to non-data.table method.
    return(add_feature_info_parameters(
      object = object,
      data = data[[object@name]],
      ...
    ))
  }
)



# add_feature_info_parameters (standardisation, ANY) ---------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationStandardisation",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)

    # Run general checks for shift and scale transforms. This may yield
    # none-transforms which are complete by default.
    object <- callNextMethod()

    # Check if all required parameters have been set now.
    if (feature_info_complete(object)) return(object)

    # Remove any non-finite values.
    data <- data[is.finite(data)]

    # Check if data is not empty after removing non-finite data and return
    # none-class normalisation object, if it is. This is done by calling the
    # method with signature data=NULL to handle setting the data.
    if (is_empty(data)) {
      object <- add_feature_info_parameters(object = object, data = NULL)
      object@reason <- "no finite values left for normalisation"

      return(object)
    }

    # Trimming and winsoring of input data.
    if (object@method %in% c("standardisation_trim")) {
      data <- trim(data, fraction = 0.05)
      
    } else if (object@method %in% c("standardisation_winsor")) {
      data <- winsor(data, fraction = 0.05)
    }

    # Determine mean (shift) and standard deviation (scale)
    if (object@method %in% c("standardisation_robust")) {
      # Using Huber's M-estimators for location and scale.
      robust_estimates <- huber_estimate(data)

      shift <- robust_estimates$mu
      scale <- robust_estimates$sigma
      
    } else {
      # Using conventional estimators.
      shift <- mean(data)
      scale <- sqrt(sum((data - shift)^2.0) / length(data))
    }

    # Check for scales which are close to 0, or could not be computed.
    if (!is.finite(scale)) scale <- 1.0
    if (scale < 2.0 * .Machine$double.eps) scale <- 1.0

    # Set shift and scale parameters.
    object@shift <- shift
    object@scale <- scale
    object@complete <- TRUE

    return(object)
  }
)



# add_feature_info_parameters (quantile, data.table) ---------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationQuantile",
    data = "data.table"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Pass to non-data.table method.
    return(add_feature_info_parameters(
      object = object,
      data = data[[object@name]],
      ...
    ))
  }
)



# add_feature_info_parameters (quantile, ANY) ----------------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationQuantile",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Run general checks for shift and scale transforms. This may yield
    # none-transforms which are complete by default.
    object <- callNextMethod()
    
    # Check if all required parameters have been set now.
    if (feature_info_complete(object)) return(object)

    # Remove any non-finite values.
    data <- data[is.finite(data)]

    # Check if data is not empty after removing non-finite data and return
    # none-class normalisation object, if it is. This is done by calling the
    # method with signature data=NULL to handle setting the data.
    if (is_empty(data)) {
      object <- add_feature_info_parameters(object = object, data = NULL)
      object@reason <- "no finite values left for normalisation"

      return(object)
    }

    # Determine median (shift) and interquartile range (scale)
    shift <- stats::median(data)
    scale <- unname(diff(stats::quantile(data, probs = c(0.25, 0.75))))

    # Check for scales which are close to 0
    if (scale < 2.0 * .Machine$double.eps) scale <- 1.0

    # Set shift and scale parameters.
    object@shift <- shift
    object@scale <- scale
    object@complete <- TRUE

    return(object)
  }
)



# add_feature_info_parameters (normalisation, data.table) ----------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationNormalisation", 
    data = "data.table"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Pass to non-data.table method.
    return(add_feature_info_parameters(
      object = object,
      data = data[[object@name]],
      ...
    ))
  }
)



# add_feature_info_parameters (normalisation, ANY) -----------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationNormalisation",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Run general checks for shift and scale transforms. This may yield
    # none-transforms which are complete by default.
    object <- callNextMethod()

    # Check if all required parameters have been set now.
    if (feature_info_complete(object)) return(object)

    # Remove any non-finite values.
    data <- data[is.finite(data)]

    # Check if data is not empty after removing non-finite data and return
    # none-class normalisation object, if it is. This is done by calling the
    # method with signature data=NULL to handle setting the data.
    if (is_empty(data)) {
      object <- add_feature_info_parameters(object = object, data = NULL)
      object@reason <- "no finite values left for normalisation"

      return(object)
    }

    # Trimming and winsoring of input data.
    if (object@method %in% c("normalisation_trim")) {
      data <- trim(data, fraction = 0.05)
      
    } else if (object@method %in% c("normalisation_winsor")) {
      data <- winsor(data, fraction = 0.05)
    }

    # Determine max and min values.
    x_min <- min(data, na.rm = TRUE)
    x_max <- max(data, na.rm = TRUE)

    # Set shift and scale parameters.
    shift <- x_min
    scale <- x_max - x_min

    # Check for scales which are close to 0.
    if (scale < 2.0 * .Machine$double.eps) scale <- 1.0

    # Set shift and scale parameters.
    object@shift <- shift
    object@scale <- scale
    object@complete <- TRUE

    return(object)
  }
)



# add_feature_info_parameters (mean centering, data.table) ---------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationMeanCentering",
    data = "data.table"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Pass to non-data.table method.
    return(add_feature_info_parameters(
      object = object,
      data = data[[object@name]],
      ...
    ))
  }
)



# add_feature_info_parameters (mean centering, ANY) ----------------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationMeanCentering",
    data = "ANY"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Check if all required parameters have been set.
    if (feature_info_complete(object)) return(object)
    
    # Run general checks for shift and scale transforms. This may yield
    # none-transforms which are complete by default.
    object <- callNextMethod()
    
    # Check if all required parameters have been set now.
    if (feature_info_complete(object)) return(object)
    
    # Remove any non-finite values.
    data <- data[is.finite(data)]
    
    # Check if data is not empty after removing non-finite data and
    # return none-class normalisation object, if it is. This is done by
    # calling the method with signature data=NULL to handle setting the
    # data.
    if (is_empty(data)) {
      object <- add_feature_info_parameters(object = object, data = NULL)
      object@reason <- "no finite values left for normalisation"
      
      return(object)
    }
    
    # Determine mean as shift.
    shift <- mean(data)

    # Set shift parameter.
    object@shift <- shift
    object@complete <- TRUE

    return(object)
  }
)



# apply_feature_info_parameters (shift and scale, data.table) ------------------
setMethod(
  "apply_feature_info_parameters", 
  signature(
    object = "featureInfoParametersNormalisationShiftScale",
    data = "data.table"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Pass to method with signature data="ANY" to perform the normalisation.
    y <- apply_feature_info_parameters(object = object, data[[object@name]], ...)
    
    # Avoid updating by reference.
    data <- data.table::copy(data)
    
    # Update the respective column with the transformed values.
    data[, (object@name) := y]
    
    return(data)
  }
)



# apply_feature_info_parameters (shift and scale, ANY) -------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationShiftScale",
    data = "ANY"
  ),
  function(
    object,
    data,
    invert = FALSE,
    ...
  ) {
    if (invert) {
      # Invert normalisation by multiplying input x by the scale and adding the
      # shift.
      return(data * object@scale + object@shift)
      
    } else {
      # Apply normalisation by subtracting the shift parameter and dividing by
      # the scale.
      return((data - object@shift) / (object@scale))
    }
  }
)



# apply_feature_info_parameters (shift, data.table) ----------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationShift",
    data = "data.table"
  ),
  function(
    object,
    data,
    ...
  ) {
    # Pass to method with signature data="ANY" to perform the normalisation.
    y <- apply_feature_info_parameters(object = object, data[[object@name]], ...)
    
    # Avoid updating by reference.
    data <- data.table::copy(data)
    
    # Update the respective column with the transformed values.
    data[, (object@name) := y]
    
    return(data)
  }
)



# apply_feature_info_parameters (shift, ANY) -----------------------------------
setMethod(
  "apply_feature_info_parameters",
  signature(
    object = "featureInfoParametersNormalisationShift",
    data = "ANY"
  ),
  function(
    object,
    data,
    invert = FALSE,
    ...
  ) {
    if (invert) {
      # Invert normalisation by adding the shift.
      return(data + object@shift)
    } else {
      # Apply normalisation by subtracting the shift parameter.
      return(data - object@shift)
    }
  }
)



.normalise <- function(x, normalisation_method, range = NULL) {
  # Direct normalisation, no questions asked.
  
  # Create a skeleton object.
  object <- ..create_normalisation_parameter_skeleton(
    feature_name = "x",
    feature_type = ifelse(is.numeric(x), "numeric", "factor"),
    available = TRUE,
    method = normalisation_method
  )

  # Obtain normalisation parameters.
  object <- add_feature_info_parameters(
    object = object,
    data = x
  )

  # Apply normalisation parameters.
  y <- apply_feature_info_parameters(
    object = object,
    data = x
  )

  # Output to certain range (if provided).
  if (!is.null(range)) {
    y[y < range[1L]] <- range[1L]
    y[y > range[2L]] <- range[2L]
  }

  return(y)
}



.get_default_normalisation_range_for_plotting <- function(norm_method) {
  # Find a default range for normalised values during plotting.
  if (norm_method == "none") {
    return(c(NA, NA))
    
  } else if (norm_method == "mean_centering") {
    return(c(NA, NA))
    
  } else if (norm_method == "quantile") {
    return(c(-1.5, 0.0, 1.5))
    
  } else if (
    norm_method %in% c(
      "standardisation", "standardisation_trim",
      "standardisation_winsor", "standardisation_robust"
    )
  ) {
    return(c(-3.0, 0.0, 3.0))
  } else if (
    norm_method %in% c("normalisation", "normalisation_trim", "normalisation_winsor")
  ) {
    return(c(0.0, 1.0))
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".get_default_normalisation_range_for_plotting: unknown ",
      "normalisation method: ", norm_method
    ))
  }
}



..collect_and_aggregate_normalisation_info <- function(
    feature_info_list,
    instance_mask,
    feature_name
) {
  # Aggregate normalisation parameters. This function exists so that it can be
  # tested as part of a unit test.

  return(...collect_and_aggregate_normalisation_info(
    object_list = lapply(
      feature_info_list,
      function(x) (x@normalisation_parameters)
    ),
    instance_mask = instance_mask,
    feature_name = feature_name
  ))
}



...collect_and_aggregate_normalisation_info <- function(
    object_list,
    instance_mask,
    feature_name
) {
  # This is used to collect from a list of objects. It serves both to aggregate
  # data for the ensemble and when creating replacement data for batch
  # normalisation.

  # Suppress NOTES due to non-standard evaluation in data.table
  n <- NULL

  if (!any(instance_mask)) {
    return(list(
      "parameters" = ..create_normalisation_parameter_skeleton(
        feature_name = feature_name,
        method = "none"
      ),
      "instance_mask" = instance_mask
    ))
  }

  # Check the class of the transformation objects.
  object_class <- sapply(object_list, function(x) (class(x)[1L]))

  # Determine if there are any objects that are not NULL or
  # featureInfoParametersNormalisationNone.
  if (
    all(object_class[instance_mask] %in% c("NULL", "featureInfoParametersNormalisationNone"))
  ) {
    return(list(
      "parameters" = ..create_normalisation_parameter_skeleton(
        feature_name = feature_name,
        method = "none"
      ),
      "instance_mask" = instance_mask
    ))
  }

  # For the remaining objects, check which class occurs most.
  class_table <- data.table::data.table(
    "class" = object_class[instance_mask]
  )[, list("n" = .N), by = "class"]

  # Drop NULL and none transformations.
  class_table <- class_table[!class %in% c("NULL", "featureInfoParametersNormalisationNone")]

  # Select the object that occurs most often.
  most_common_class <- class_table[n == max(class_table$n), ]$class[1L]

  # Update the instance mask.
  instance_mask <- instance_mask & object_class == most_common_class

  if (most_common_class %in% c(
    "featureInfoParametersNormalisationStandardisation",
    "featureInfoParametersNormalisationQuantile",
    "featureInfoParametersNormalisationNormalisation",
    "featureInfoParametersNormalisationParametricCombat",
    "featureInfoParametersNormalisationNonParametricCombat"
  )) {
    # Aggregate shift values, and select median value.
    all_shift <- sapply(object_list[instance_mask], function(x) (x@shift))
    selected_shift <- stats::median(all_shift)

    # Aggregate scale values, and select mean value.
    all_scale <- sapply(object_list[instance_mask], function(x) (x@scale))
    selected_scale <- stats::median(all_scale)

    # Identify the first method -- its not that crucial.
    selected_method <- sapply(object_list[instance_mask], function(x) (x@method))[1L]
    
  } else if (most_common_class %in% c("featureInfoParametersNormalisationMeanCentering")) {
    # Aggregate shift values, and select median value.
    all_shift <- sapply(object_list[instance_mask], function(x) (x@shift))
    selected_shift <- stats::median(all_shift)

    # Identify the first method -- its not that crucial.
    selected_method <- sapply(object_list[instance_mask], function(x) (x@method))[1L]
    
  } else {
    ..error_reached_unreachable_code(paste0(
      "..collect_and_aggregate_normalisation_info: encountered ",
      "an unknown class of objects: ", paste_s(most_common_class)
    ))
  }

  return(list(
    "parameters" = ..create_normalisation_parameter_skeleton(
      feature_name = feature_name,
      method = unname(selected_method),
      shift = unname(selected_shift),
      scale = unname(selected_scale)
    ),
    "instance_mask" = instance_mask
  ))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# show (familiarData) ----------------------------------------------------------
setMethod(
  "show",
  signature(object = "familiarData"),
  function(object) {
    
    # Make sure the data object is updated.
    object <- update_object(object = object)
    
    # Create an initial descriptor.
    cat(paste0(
      "A dataset (", object@name, "; ",
      .familiar_version_string(object),
      ").\n"
    ))
    
    # Details concerning the generating ensemble.
    cat(paste0(
      "\nThe ensemble that created this dataset contained of one or more ",
      object@learner,
      " models with variable importance computed by the ",
      object@vimp_method,
      " variable importance method.\n"
    ))
    
    # Outcome details
    cat("\nThe following outcome was modelled:\n")
    show(object@outcome_info)
  }
)



# save (familiarData) ----------------------------------------------------------
setMethod(
  "save",
  signature(
    list = "familiarData",
    file = "character"
  ),
  function(list, file) {
    .save(object = list, dir_path = file)
  }
)





# add_package_version (familiarData) -------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "familiarData"),
  function(object) {
    
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)



# add_identifiers --------------------------------------------------------------
setMethod(
  "add_identifiers",
  signature(
    data = "ANY",
    object = "familiarData"
  ),
  function(
    data,
    object,
    more_identifiers = NULL
  ) {
    # Adds identifying columns to a table
    
    if (is_empty(data)) return(NULL)
    
    if (!inherits(data, "data.table")) {
      ..error_reached_unreachable_code("\"data\" should be a data.table.")
    }
    
    # Check which identifiers should be added
    if (is.null(more_identifiers)) {
      if (!all(more_identifiers %in% c("vimp_method", "learner"))) {
        ..error_reached_unreachable_code(paste0(
          "Only variable importance methods (\"vimp_method\") and learners (\"learner\") ",
          "can be added as additional identifiers."
        ))
      }
    }
    id_order <- c("data_set", more_identifiers)
    
    if (nrow(data) >= 1L) {
      
      # Get the name of the data
      data_set <- object@name
      
      # Insert "model_name" column
      data[, "data_set" := data_set]
      
      if (any(id_order == "vimp_method")) data[, "vimp_method" := object@vimp_method]
      
      if (any(id_order == "learner")) data[, "learner" := object@learner]
      
      # Reorder columns and move model_name to the front
      data.table::setcolorder(data, neworder = id_order)
      
      return(data)
      
    } else {
      # In case the table is empty, return an empty table with the model name
      # attached.
      empty_table <- data.table::data.table("data_set" = character(0L))
      
      if (any(id_order == "vimp_method")) {
        empty_table[, "vimp_method" := character(0L)]
      }
      
      if (any(id_order == "learner")) {
        empty_table[, "learner" := character(0L)]
      }
      
      return(cbind(empty_table, data))
    }
  }
)



# set_object_name (familiarData) -----------------------------------------------

#' @title Set the name of a `familiarData` object.
#'  
#' @description Set the `name` slot using the object name.
#'
#' @param x A `familiarData` object.
#' 
#' @return A `familiarData` object with a generated or a provided name.
#' @md
#' @keywords internal
setMethod(
  "set_object_name",
  signature(x = "familiarData"),
  function(x, new = NULL) {
    
    if (is.null(x@project_id) && is.null(new)) {
      # Generate a random object name. A project_id of 0 means that the objects
      # was auto-generated (i.e. through object conversion). We randomly
      # generate chracters and add a time stamp, so that collision is
      # practically impossible.
      slot(object = x, name = "name") <- paste0(
        as.character(as.numeric(format(Sys.time(), "%H%M%S"))),
        "_", rstring(n = 20L)
      )
      
    } else if (is.null(new)) {
      # Generate a sensible object name.
      slot(object = x, name = "name") <- get_object_name(object = x)
      
    } else {
      slot(object = x, name = "name") <- new
    }
    
    return(x)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

setClass("familiarMBoost", contains = "familiarModel")

setClass(
  "familiarMBoostLM",
  contains = "familiarMBoost",
  slots = list(
    "encoding_reference_table" = "ANY",
    "feature_order" = "character"
  ),
  prototype = list(
    "encoding_reference_table" = NULL,
    "feature_order" = character()
  )
)

setClass(
  "familiarMBoostTree",
  contains = "familiarMBoost",
  slots = list(
    "encoding_reference_table" = "ANY",
    "feature_order" = "character"
  ),
  prototype = list(
    "encoding_reference_table" = NULL,
    "feature_order" = character()
  )
)



.get_available_mboost_lm_learners <- function(show_general = TRUE) {
  
  # Learners
  learners <- c(
    "boosted_glm", "boosted_glm_logistic",
    "boosted_glm_probit", "boosted_glm_loglog", "boosted_glm_cauchy", "boosted_glm_log",
    "boosted_glm_auc", "boosted_glm_gaussian", "boosted_glm_huber", "boosted_glm_laplace",
    "boosted_glm_poisson", "boosted_glm_cox", "boosted_glm_surv",
    "boosted_glm_weibull", "boosted_glm_lognormal", "boosted_glm_gehan", "boosted_glm_cindex"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("boosted_glm", "boosted_glm_surv"))
  }
  
  return(learners)
}


.get_available_mboost_tree_learners <- function(show_general = TRUE) {
  
  # Learners
  learners <- c(
    "boosted_tree", "boosted_tree_logistic", "boosted_tree_probit",
    "boosted_tree_loglog", "boosted_tree_cauchy", "boosted_tree_log",
    "boosted_tree_auc", "boosted_tree_gaussian", "boosted_tree_huber",
    "boosted_tree_laplace", "boosted_tree_poisson", "boosted_tree_cox", "boosted_tree_surv",
    "boosted_tree_weibull", "boosted_tree_lognormal", "boosted_tree_gehan", "boosted_tree_cindex"
  )
  
  if (!show_general) {
    learners <- setdiff(learners, c("boosted_tree", "boosted_tree_surv"))
  }
  
  return(learners)
}


# is_available,familiarMBoostLM ------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMBoostLM"),
  function(object, ...) {
    
    # Extract learner and outcome_type from the familiarModel object.
    learner <- object@learner
    outcome_type <- object@outcome_type
    
    if (
      outcome_type == "survival" &&
      learner %in% c(
        "boosted_glm", "boosted_glm_cox", "boosted_glm_surv",
        "boosted_glm_loglog", "boosted_glm_weibull", "boosted_glm_lognormal",
        "boosted_glm_gehan", "boosted_glm_cindex"
      )
    ) {
      ..deprecation_mboost()
      return(FALSE)
      
    } else if (
      outcome_type == "continuous" &&
      learner %in% c(
        "boosted_glm", "boosted_glm_gaussian", "boosted_glm_huber",
        "boosted_glm_laplace", "boosted_glm_poisson"
      )
    ) {
      ..deprecation_mboost()
      return(FALSE)
      
    } else if (
      outcome_type == "binomial" &&
      learner %in% c(
        "boosted_glm", "boosted_glm_logistic", "boosted_glm_probit", "boosted_glm_loglog",
        "boosted_glm_cauchy", "boosted_glm_log", "boosted_glm_auc"
      )
    ) {
      ..deprecation_mboost()
      return(FALSE)
      
    } else if (outcome_type == "count" && learner %in% c("boosted_glm", "boosted_glm_poisson")) {
      ..deprecation_mboost()
      ..deprecation_count()
      return(FALSE)
      
    } else {
      return(FALSE)
    }
  }
)



# is_available,familiarMBoostTree ----------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarMBoostTree"),
  function(object, ...) {
            
    # Extract learner and outcome_type from the familiarModel object.
    learner <- object@learner
    outcome_type <- object@outcome_type
    
    if (
      outcome_type == "survival" && 
      learner %in% c(
        "boosted_tree", "boosted_tree_cox", "boosted_tree_surv",
        "boosted_tree_loglog", "boosted_tree_weibull", "boosted_tree_lognormal",
        "boosted_tree_gehan", "boosted_tree_cindex"
      )
    ) {
      ..deprecation_mboost()
      return(FALSE)
      
    } else if (
      outcome_type == "continuous" &&
      learner %in% c(
        "boosted_tree", "boosted_tree_gaussian", "boosted_tree_huber",
        "boosted_tree_laplace", "boosted_tree_poisson"
      )
    ) {
      ..deprecation_mboost()
      return(FALSE)
      
    } else if (
      outcome_type == "binomial" &&
      learner %in% c(
        "boosted_tree", "boosted_tree_logistic", "boosted_tree_probit",
        "boosted_tree_loglog", "boosted_tree_cauchy", "boosted_tree_log",
        "boosted_tree_auc"
      )
    ) {
      ..deprecation_mboost()
      return(FALSE)
      
    } else if (outcome_type == "count" && learner %in% c("boosted_tree", "boosted_tree_poisson")) {
      ..deprecation_mboost()
      ..deprecation_count()
      return(FALSE)
      
    } else {
      return(FALSE)
    }
  }
)


# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarMBoost",
    data = "dataObject"
  ),
  function(object, data, type = "default", ...) {
    ..deprecation_mboost()
    return(callNextMethod())
  }
)

testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Simple test
vimp_methods <- c("none", "mim")

# Imbalance corrections using full undersampling
experimental_designs <- c("ip(fs+mb)", "ip(mb)")
for (ii in seq_along(vimp_methods)) {
  familiar:::integrated_test(
    experimental_design = experimental_designs[ii],
    imbalance_correction_method = "full_undersampling",
    vimp_method = vimp_methods[ii],
    parallel = FALSE,
    estimation_type = "point",
    n_important_features = 2L,
    outcome_type_available = "binomial",
    debug = debug_flag
  )
  
  familiar:::integrated_test(
    experimental_design = experimental_designs[ii],
    imbalance_correction_method = "random_undersampling",
    imbalance_n_partitions = 3,
    vimp_method = vimp_methods[ii],
    parallel = FALSE,
    skip_evaluation_elements = "all",
    outcome_type_available = "binomial",
    debug = debug_flag
  )
}

# First test if all selectable learners are also available
familiar:::test_all_vimp_methods_available(
  vimp_methods = familiar:::.get_available_glmnet_ridge_vimp_methods(show_general = TRUE)
)
familiar:::test_all_vimp_methods_available(
  vimp_methods = familiar:::.get_available_glmnet_lasso_vimp_methods(show_general = TRUE)
)
familiar:::test_all_vimp_methods_available(
  vimp_methods = familiar:::.get_available_glmnet_elastic_net_vimp_methods(show_general = TRUE)
)

# Don't perform any further tests on CRAN due to time of running the complete test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_all_vimp_methods(
  vimp_methods = familiar:::.get_available_glmnet_ridge_vimp_methods(show_general = FALSE)
)
familiar:::test_all_vimp_methods(
  vimp_methods = familiar:::.get_available_glmnet_lasso_vimp_methods(show_general = FALSE)
)
familiar:::test_all_vimp_methods(
  vimp_methods = familiar:::.get_available_glmnet_elastic_net_vimp_methods(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list("alpha" = 0.50),
    "binomial" = list("alpha" = 0.50),
    "multinomial" = list("alpha" = 0.50),
    "survival" = list("alpha" = 0.50)
  )
)

# Parallel tests
familiar:::test_all_vimp_methods_parallel(
  vimp_methods = familiar:::.get_available_glmnet_ridge_vimp_methods(show_general = FALSE)
)
familiar:::test_all_vimp_methods_parallel(
  vimp_methods = familiar:::.get_available_glmnet_lasso_vimp_methods(show_general = FALSE)
)
familiar:::test_all_vimp_methods_parallel(
  vimp_methods = familiar:::.get_available_glmnet_elastic_net_vimp_methods(show_general = FALSE),
  hyperparameter_list = list(
    "continuous" = list("alpha" = 0.50),
    "binomial" = list("alpha" = 0.50),
    "multinomial" = list("alpha" = 0.50),
    "survival" = list("alpha" = 0.50)
  )
)


familiar:::test_hyperparameter_optimisation(
  vimp_methods = "ridge",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)
familiar:::test_hyperparameter_optimisation(
  vimp_methods = "lasso",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)
familiar:::test_hyperparameter_optimisation(
  vimp_methods = "elastic_net",
  debug = FALSE,
  parallel = FALSE,
  test_specific_config = TRUE
)


testthat::skip("Skip hyperparameter optimisation, unless manual.")

familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_glmnet_ridge_vimp_methods(show_general = FALSE),
  debug = TRUE,
  parallel = FALSE
)
familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_glmnet_lasso_vimp_methods(show_general = FALSE),
  debug = TRUE,
  parallel = FALSE
)
familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_glmnet_elastic_net_vimp_methods(show_general = TRUE),
  debug = TRUE,
  parallel = FALSE
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

setClass(
  "familiarRegressionVimp",
  contains = "familiarVimpMethod"
)

setClass(
  "familiarUnivariateRegressionVimp",
  contains = "familiarRegressionVimp"
)

setClass(
  "familiarMultivariateRegressionVimp",
  contains = "familiarRegressionVimp",
  prototype = methods::prototype(multivariate = TRUE)
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarMultivariateRegressionVimp"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()
    
    # Set the required package
    .Object@multivariate <- TRUE
    
    return(.Object)
  }
)


.get_available_univariate_regression_vimp_methods <- function(show_general = TRUE) {
  return("univariate_regression")
}

.get_available_multivariate_regression_vimp_methods <- function(show_general = TRUE) {
  return("multivariate_regression")
}

# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarRegressionVimp"),
  function(object, ...) {
    
    if (object@outcome_type == "count") {
      ..deprecation_count()
      return(FALSE)
    }
    
    return(TRUE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters", 
  signature(object = "familiarRegressionVimp"),
  function(object, data = NULL, ...) {
    param <- list()
    param$n_bootstrap <- list()
    param$learner <- list()
    param$metric <- list()
    param$drop_rate <- list()
    
    if (is.null(data)) return(param)
    
    # number of bootstrap iterations -------------------------------------------
    param$n_bootstrap <- .set_hyperparameter(
      default = 10L,
      type = "integer",
      range = c(1L, Inf),
      randomise = FALSE
    )
    
    # rate at which least performing features are dropped ----------------------
    param$drop_rate <- .set_hyperparameter(
      default = 0.333, 
      type = "numeric",
      range = c(0.0, 1.0),
      randomise = FALSE
    )
    
    # distribution family to use for linear regression -------------------------
    learner_default <- switch(
      object@outcome_type,
      "binomial" = "glm_logistic",
      "multinomial" = "glm_multinomial",
      "continuous" = "glm_gaussian",
      "survival" = "cox"
    )
    
    # Set the learner range, i.e. all generalised linear models.
    learner_range <- unique(c(
      .get_available_glm_learners(),
      .get_available_cox_learners(),
      .get_available_survival_regression_learners()
    ))
    
    # Determine which learners are available for the outcome_type
    learner_is_available <- sapply(
      learner_range,
      .check_learner_outcome_type,
      outcome_type = object@outcome_type,
      as_flag = TRUE
    )
    
    # Create the learner hyperparameter
    param$learner <- .set_hyperparameter(
      default = learner_default,
      type = "factor",
      range = learner_range[learner_is_available],
      randomise = FALSE
    )
    
    # performance metric -------------------------------------------------------
    metric_default <- switch(
      object@outcome_type,
      "binomial" = "auc_roc",
      "multinomial" = "auc_roc",
      "continuous" = "mse",
      "survival" = "concordance_index"
    )
    
    # Get all available metrics.
    metric_range <- .get_all_metrics()
    
    # Determine which of the metrics is available for the outcome type.
    metric_is_available <- sapply(
      metric_range,
      .check_metric_outcome_type,
      outcome_type = object@outcome_type,
      as_flag = TRUE
    )
    
    # Create the metric hyperparameter
    param$metric <- .set_hyperparameter(
      default = metric_default,
      type = "factor",
      range = metric_range[metric_is_available],
      randomise = FALSE
    )
    
    return(param)
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarRegressionVimp"),
  function(object, data, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    score <- name <- available <- selected <- NULL
    
    if (is_empty(data)) return(callNextMethod())
    
    # Aggregate data.
    data <- aggregate_data(data)
    
    # Find feature columns in data table.
    feature_columns <- get_feature_columns(x = data)
    
    # Generate iteration list.
    iteration_list <- .create_bootstraps(
      n_iter = object@hyperparameters$n_bootstrap,
      outcome_type = object@outcome_type,
      data = data@data
    )
    
    # Create a generic model.
    fam_model <- promote_learner(
      object = methods::new(
        "familiarModel",
        outcome_type = object@outcome_type,
        learner = as.character(object@hyperparameters$learner),
        vimp_method = "none",
        feature_info = object@feature_info,
        outcome_info = .compute_outcome_distribution_data(
          object = object@outcome_info, 
          data = data
        )
      )
    )
    
    # Create metric objects.
    metric_object_list <- lapply(
      as.character(object@hyperparameters$metric),
      as_metric,
      outcome_type = object@outcome_type
    )
    
    # Add baseline values for each metric.
    metric_object_list <- lapply(
      metric_object_list,
      set_metric_baseline_value,
      object = fam_model,
      data = data
    )
    
    # Generate score table.
    score_table <- data.table::data.table(
      "name" = feature_columns,
      "available" = TRUE,
      "selected" = FALSE,
      "select_step" = 0L,
      "score" = NA_real_
    )
    
    # Find signature features, if any.
    signature_feature <- names(object@feature_info)[sapply(object@feature_info, is_in_signature)]
    
    if (
      length(signature_feature) > 0L &&
      is(object, "familiarMultivariateRegressionVimp")
    ) {
      # For multivariate regression with features in a signature.
      
      # Get the objective score for the signature.
      objective_score <- ..regression_vimp_assess_feature(
        feature = NULL,
        fam_model = fam_model,
        metric_objects = metric_object_list,
        data = data,
        iteration_list = iteration_list,
        fixed_set = signature_feature
      )
      
      # Set max_objective_score.
      max_objective_score <- objective_score$score
      
      # Update the score table.
      score_table[
        name %in% signature_feature,
        ":="(
          "score" = max_objective_score,
          "available" = FALSE,
          "selected" = TRUE,
          "select_step" = seq_len(length(signature_feature))
        )
      ]
      
      # Iteration counter.
      ii <- length(signature_feature) + 1L
      
    } else {
      # For multivariate regression without signature features, or univariate
      # regression.
      
      # Compute objective score for every feature.
      objective_score <- lapply(
        feature_columns,
        ..regression_vimp_assess_feature,
        fam_model = fam_model,
        metric_objects = metric_object_list,
        data = data,
        iteration_list = iteration_list,
        fixed_set = NULL
      )
      
      # Combine into data.table.
      objective_score <- data.table::rbindlist(objective_score)
      
      # In case of univariate regression, return at this point.
      if (is(object, "familiarUnivariateRegressionVimp")) {
        # Create variable importance object.
        vimp_object <- methods::new(
          "vimpTable",
          vimp_table = data.table::data.table(
            "score" = objective_score$score,
            "name" = objective_score$name
          ),
          score_aggregation = "max",
          invert = TRUE
        )
        
        return(vimp_object)
      }
      
      # Proceed with forward selection and multivariate regression.
      
      # Find the best performing feature.
      max_objective_score <- max(objective_score$score)
      best_feature <- objective_score[score == max_objective_score]$name
      
      # Update the score_table.
      score_table[
        name %in% best_feature,
        ":="(
          "score" = max_objective_score,
          "available" = FALSE,
          "selected" = TRUE,
          "select_step" = 1L
        )
      ]
      
      # Identify features to be dropped.
      n_available <- nrow(objective_score) - length(best_feature)
      n_dropped <- floor(object@hyperparameters$drop_rate * n_available)
      
      # Make dropped features unavailable.
      if (n_dropped > 0L) {
        # Identify the worst performing features.
        bad_features <- tail(objective_score[(order(-score))], n = n_dropped)$name
        score_table[name %in% bad_features, "available" := FALSE]
      }
      
      # Iteration counter.
      ii <- 2L
    }
    
    # Determine the features still available.
    available_features <- score_table[available == TRUE, ]$name
    selected_features <- score_table[selected == TRUE, ]$name
    
    # Set current objective score.
    previous_objective_score <- max_objective_score
    
    
    while (length(available_features) > 0L) {
      # Generate new iteration list.
      iteration_list <- .create_bootstraps(
        n_iter = object@hyperparameters$n_bootstrap,
        outcome_type = object@outcome_type,
        data = data@data
      )
      
      # Compute objective score for every feature.
      objective_score <- lapply(
        available_features,
        ..regression_vimp_assess_feature,
        fam_model = fam_model,
        metric_objects = metric_object_list,
        data = data,
        iteration_list = iteration_list,
        fixed_set = selected_features
      )
      
      # Combine into data.table.
      objective_score <- data.table::rbindlist(objective_score)
      
      # Find the best performing feature.
      max_objective_score <- max(objective_score$score)
      best_feature <- objective_score[score == max_objective_score]$name
      
      # Only continue adding, if the expected value changed.
      if (max_objective_score <= previous_objective_score) break
      
      # Update the score_table.
      score_table[
        name %in% best_feature, 
        ":="(
          "score" = max_objective_score,
          "available" = FALSE,
          "selected" = TRUE,
          "select_step" = ii
        )
      ]
      
      # Identify features to be dropped.
      n_available <- nrow(objective_score) - length(best_feature)
      n_dropped <- floor(object@hyperparameters$drop_rate * n_available)
      
      # Make dropped features unavailable.
      if (n_dropped > 0L) {
        # Identify the worst performing features.
        bad_features <- tail(objective_score[(order(-score))], n = n_dropped)$name
        score_table[name %in% bad_features, "available" := FALSE]
      }
      
      # In addition, drop any features that do not improve upon the previous
      # best score.
      bad_features <- objective_score[score < previous_objective_score]$name
      score_table[name %in% bad_features, "available" := FALSE]
      
      # Determine the features still available.
      available_features <- score_table[available == TRUE, ]$name
      selected_features <- score_table[selected == TRUE, ]$name
      
      # Update iteration counter.
      ii <- ii + 1L
      
      # Set current objective score.
      previous_objective_score <- max_objective_score
    }
    
    # Create variable importance table from the score table.
    vimp_table <- score_table[selected == TRUE, c("name", "score", "select_step")]
    
    # Create variable importance object.
    vimp_object <- methods::new(
      "vimpTable",
      vimp_table = data.table::data.table(
        "score" = vimp_table$select_step,
        "name" = vimp_table$name
      ),
      score_aggregation = "max",
      invert = FALSE
    )
    
    return(vimp_object)
  }
)



..regression_vimp_assess_feature <- function(
    feature,
    fam_model,
    metric_objects,
    data,
    iteration_list,
    fixed_set
) {
  # Make local copy of the data prior to filtering features.
  data <- data.table::copy(data)
  
  # Remove features that are neither in feature nor fixed_set.
  data <- filter_features(
    data,
    available_features = c(feature, fixed_set)
  )
  
  # Find features that are required for processing the data and building the
  # model.
  fam_model@required_features <- get_required_features(
    x = data,
    feature_info_list = fam_model@feature_info
  )
  fam_model@model_features <- get_model_features(
    x = data,
    feature_info_list = fam_model@feature_info
  )
  
  # Iterate over bootstraps.
  performance_data <- lapply(
    seq_along(iteration_list$train_list),
    function(bootstrap_id, data, fam_model, metric_objects, iteration_list) {
      # Select train and test data.
      train_data <- select_data_from_samples(
        data = data,
        samples = iteration_list$train_list[[bootstrap_id]]
      )
      test_data <- select_data_from_samples(
        data = data,
        samples = iteration_list$valid_list[[bootstrap_id]]
      )
      
      # Update the familiar model.
      fam_model@hyperparameters$sign_size <- length(c(feature, fixed_set))
      
      # Set missing parameters.
      parameter_list <- get_default_hyperparameters(
        object = fam_model,
        data = train_data,
        user_list = fam_model@hyperparameters
      )
      
      # Update the parameter list With user-defined variables.
      parameter_list <- .update_hyperparameters(
        parameter_list = parameter_list,
        user_list = fam_model@hyperparameters
      )
      
      # Update hyperparameters to set any fixed parameters.
      fam_model@hyperparameters <- lapply(
        parameter_list,
        function(list_entry) list_entry$init_config
      )
      
      # Set additional parameters, e.g. the learner family.
      fam_model <- ..set_vimp_parameters(
        object = fam_model,
        method = fam_model@learner
      )
      
      # Train the model using the train data.
      fam_model <- suppressWarnings(.train(
        object = fam_model,
        data = train_data,
        get_additional_info = FALSE,
        trim_model = FALSE
      ))
      
      # Compute score using the metrics.
      score_table <- mapply(
        function(data, data_set, object, metric_objects) {
          # Get metric names.
          metric_names <- sapply(
            metric_objects, 
            function(metric_object) metric_object@metric
          )
          
          # Predict for the in-bag and out-of-bag datasets.
          prediction_table <- .predict(
            object = object,
            data = data
          )
          
          # Compute objective scores.
          metrics_objective_score <- sapply(
            metric_objects,
            compute_objective_score,
            data = prediction_table
          )
          
          # Return as data.table.
          return(data.table::data.table(
            "metric" = metric_names,
            "data_set" = data_set,
            "objective_score" = metrics_objective_score
          ))
        },
        data = list(train_data, test_data),
        data_set = c("training", "validation"),
        MoreArgs = list(
          "object" = fam_model,
          "metric_objects" = metric_objects
        ),
        SIMPLIFY = FALSE
      )
      
      # Aggregate to a single table.
      score_table <- data.table::rbindlist(score_table, use.names = TRUE)
      
      # Add run id.
      score_table[, ":="("run_id" = bootstrap_id)]
      
      # Set the column order.
      data.table::setcolorder(
        x = score_table,
        neworder = c("run_id")
      )
      
      return(score_table)
    },
    data = data,
    fam_model = fam_model,
    metric_objects = metric_objects,
    iteration_list = iteration_list
  )
  
  # Combine performance data to a single table and compute optimisation scores.
  performance_data <- .compute_metric_optimisation_score(
    score_table = data.table::rbindlist(performance_data),
    optimisation_function = "max_validation"
  )
  
  # Compute the summary score.
  performance_data <- .summarise_metric_optimisation_score(
    score_table = performance_data,
    method = "median"
  )
  
  # Return sample mean and standard deviation of the objective score.
  return(data.table::data.table(
    "name" = feature,
    "score" = performance_data$optimisation_score
  ))
}

familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_univariate_regression_vimp_methods(show_general = TRUE)
)
familiar:::test_all_vimp_methods_available(
  familiar:::.get_available_multivariate_regression_vimp_methods(show_general = TRUE)
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_univariate_regression_vimp_methods(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE,
  not_available_no_samples = FALSE
)

familiar:::test_all_vimp_methods(
  familiar:::.get_available_univariate_regression_vimp_methods(show_general = FALSE)
)
familiar:::test_all_vimp_methods_parallel(
  familiar:::.get_available_univariate_regression_vimp_methods(show_general = FALSE)
)

familiar:::test_hyperparameter_optimisation(
  vimp_methods = familiar:::.get_available_multivariate_regression_vimp_methods(show_general = TRUE),
  debug = FALSE,
  parallel = FALSE,
  not_available_no_samples = FALSE
)

familiar:::test_all_vimp_methods(
  familiar:::.get_available_multivariate_regression_vimp_methods(show_general = FALSE)
)
familiar:::test_all_vimp_methods_parallel(
  familiar:::.get_available_multivariate_regression_vimp_methods(show_general = FALSE)
)


# Continuous outcome -----------------------------------------------------------
data <- familiar:::test_create_good_data("continuous")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "multivariate_regression",
  vimp_method_parameter_list = NULL,
  outcome_type = "continuous",
  cluster_method = "none",
  imputation_method = "simple"
)

testthat::test_that(paste0("Multivariate regression correctly ranks continuous data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data)
  ))

  # Expect that the vimp table has at most six rows.
  testthat::expect_lte(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(vimp_table$name %in% familiar:::get_feature_columns(data))
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Binomial outcome -------------------------------------------------------------
data <- familiar:::test_create_good_data("binomial")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "multivariate_regression",
  vimp_method_parameter_list = NULL,
  outcome_type = "binomial",
  cluster_method = "none",
  imputation_method = "simple"
)

testthat::test_that(paste0("Multivariate regression correctly ranks binomial data.."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data))
  )
  
  # Expect that the vimp table has at most six rows.
  testthat::expect_lte(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(vimp_table$name %in% familiar:::get_feature_columns(data))
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Multinomial outcome ----------------------------------------------------------
data <- familiar:::test_create_good_data("multinomial")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "multivariate_regression",
  vimp_method_parameter_list = NULL,
  outcome_type = "multinomial",
  cluster_method = "none",
  imputation_method = "simple"
)

testthat::test_that(paste0("Multivariate regression correctly ranks multinomial outcome data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data))
  )
  
  # Expect that the vimp table has at most six rows.
  testthat::expect_lte(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(vimp_table$name %in% familiar:::get_feature_columns(data))
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})


# Survival outcome -------------------------------------------------------------
data <- familiar:::test_create_good_data("survival")

# Process dataset.
vimp_object <- familiar:::test_create_vimp_method(
  data = data,
  vimp_method = "multivariate_regression",
  vimp_method_parameter_list = NULL,
  outcome_type = "survival",
  cluster_method = "none",
  imputation_method = "simple"
)

testthat::test_that(paste0("Multivariate regression correctly ranks survival outcome data."), {
  vimp_table <- suppressWarnings(familiar:::get_vimp_table(
    familiar:::.vimp(vimp_object, data))
  )
  
  # Expect that the vimp table has at most six rows.
  testthat::expect_lte(nrow(vimp_table), 6L)
  
  # Expect that the names are the same as that of the features.
  testthat::expect_true(
    all(vimp_table$name %in% familiar:::get_feature_columns(data))
  )
  
  # Feature 1 is most important.
  testthat::expect_equal(vimp_table[rank == 1, ]$name, "feature_1")
})

#' Internal checks on common plot input arguments
#'
#' @param x_range (*optional*) Value range for the x-axis.
#' @param y_range (*optional*) Value range for the y-axis.
#' @param x_n_breaks (*optional*) Number of breaks to show on the x-axis of the
#'   plot. `x_n_breaks` is used to determine the `x_breaks` argument in case it
#'   is unset.
#' @param y_n_breaks (*optional*) Number of breaks to show on the y-axis of the
#'   plot. `y_n_breaks` is used to determine the `y_breaks` argument in case it
#'   is unset.
#' @param x_breaks (*optional*) Break points on the x-axis of the plot.
#' @param y_breaks (*optional*) Break points on the y-axis of the plot.
#' @param conf_int (*optional*)
#' @param conf_int_alpha (*optional*) Alpha value to determine transparency of
#'   confidence intervals or, alternatively, other plot elements with which the
#'   confidence interval overlaps. Only values between 0.0 (fully transparent)
#'   and 1.0 (fully opaque) are allowed.
#' @param conf_int_style (*optional*) Confidence interval style. See details for
#'   allowed styles.
#' @param conf_int_default Sets the default options for the confidence interval.
#' @param limit_n_features (*optional*) The number of features that should be
#'   included in the plot. Only the most important features are shown. By 
#'   default, the number of features is not limited.
#' @param facet_wrap_cols (*optional*) Number of columns to generate when facet
#'   wrapping. If NULL, a facet grid is produced instead.
#' @param x_label (*optional*) Label to provide to the x-axis. If NULL, no label
#'   is shown.
#' @param y_label (*optional*) Label to provide to the y-axis. If NULL, no label
#'   is shown.
#' @param legend_label (*optional*) Label to provide to the legend. If NULL, the
#'   legend will not have a name.
#' @param combine_legend (*optional*) Flag to indicate whether the same legend
#'   is to be shared by multiple aesthetics, such as those specified by
#'   `color_by` and `linetype_by` arguments.
#' @param plot_title (*optional*) Label to provide as figure title. If NULL, no
#'   title is shown.
#' @param plot_sub_title (*optional*) Label to provide as figure subtitle. If
#'   NULL, no subtitle is shown.
#' @param caption (*optional*) Label to provide as figure caption. If NULL, no
#'   caption is shown.
#' @param x_label_shared (*optional*) Sharing of x-axis labels between facets.
#'   One of three values:
#'
#'   * `overall`: A single label is placed at the bottom of the figure. Tick
#'   text (but not the ticks themselves) is removed for all but the bottom facet
#'   plot(s).
#'
#'   * `column`: A label is placed at the bottom of each column. Tick text (but
#'   not the ticks themselves) is removed for all but the bottom facet plot(s).
#'
#'   * `individual`: A label is placed below each facet plot. Tick text is kept.
#'
#' @param y_label_shared (*optional*) Sharing of y-axis labels between facets.
#'   One of three values:
#'
#'   * `overall`: A single label is placed to the left of the figure. Tick text
#'   (but not the ticks themselves) is removed for all but the left-most facet
#'   plot(s).
#'
#'   * `row`: A label is placed to the left of each row. Tick text (but not the
#'   ticks themselves) is removed for all but the left-most facet plot(s).
#'
#'   * `individual`: A label is placed below each facet plot. Tick text is kept.
#'
#' @param rotate_x_tick_labels (*optional*) Rotate tick labels on the x-axis by
#'   90 degrees. Defaults to `TRUE`. Rotation of x-axis tick labels may also be
#'   controlled through the `ggtheme`. In this case, `FALSE` should be provided
#'   explicitly.
#'
#' @param rotate_y_tick_labels (*optional*) Rotate tick labels on the y-axis by
#'   45 degrees.
#' @md
#' @keywords internal
.check_input_plot_args <- function(
    x_range = waiver(),
    y_range = waiver(),
    x_n_breaks = waiver(),
    y_n_breaks = waiver(),
    x_breaks = waiver(),
    y_breaks = waiver(),
    conf_int = waiver(),
    conf_int_alpha = waiver(),
    conf_int_style = waiver(),
    conf_int_default = c("step", "ribbon", "none"),
    limit_n_features = waiver(),
    facet_wrap_cols = waiver(),
    x_label = waiver(),
    y_label = waiver(),
    x_label_shared = waiver(),
    y_label_shared = waiver(),
    rotate_x_tick_labels = waiver(),
    rotate_y_tick_labels = waiver(),
    legend_label = waiver(),
    combine_legend = waiver(),
    plot_title = waiver(),
    plot_sub_title = waiver(),
    caption = waiver()
) {
  
  # x and y ranges
  if (!is.waive(x_range)) {
    .check_input_plot_range(
      range_var = x_range, 
      var_name = "x_range"
    )
  }
  if (!is.waive(y_range)) {
    .check_input_plot_range(
      range_var = y_range, 
      var_name = "y_range"
    )
  }

  # x and y number of breaks
  if (!is.waive(x_n_breaks)) {
    .check_input_plot_n_breaks(
      n_break_var = x_n_breaks,
      var_name = "x_n_breaks"
    )
  }
  if (!is.waive(y_n_breaks)) {
    .check_input_plot_n_breaks(
      n_break_var = y_n_breaks, 
      var_name = "y_n_breaks"
    )
  }

  # x and y breaks
  if (!is.waive(x_breaks)) {
    .check_input_plot_break(
      break_var = x_breaks, 
      var_name = "x_breaks"
    )
  }
  if (!is.waive(y_breaks)) {
    .check_input_plot_break(
      break_var = y_breaks, 
      var_name = "y_breaks"
    )
  }

  # labels
  if (!is.waive(x_label)) {
    .check_input_plot_label(
      label_var = x_label, 
      var_name = "x_label"
    )
  }
  if (!is.waive(y_label)) {
    .check_input_plot_label(
      label_var = y_label, 
      var_name = "y_label"
    )
  }
  if (!is.waive(legend_label)) {
    .check_input_plot_legend(
      label_var = legend_label, 
      var_name = "legend_label"
    )
  }
  if (!is.waive(plot_title)) {
    .check_input_plot_label(
      label_var = plot_title, 
      var_name = "plot_title"
    )
  }
  if (!is.waive(plot_sub_title)) {
    .check_input_plot_label(
      label_var = plot_sub_title, 
      var_name = "plot_sub_title"
    )
  }
  if (!is.waive(caption)) {
    .check_input_plot_label(
      label_var = caption, 
      var_name = "caption"
    )
  }

  # Legend combination flag
  if (!is.waive(combine_legend)) {
    .check_parameter_value_is_valid(
      x = combine_legend, 
      var_name = "combine_legend",
      values = c(FALSE, TRUE)
    )
  }

  # Transparency values
  if (!is.waive(conf_int_alpha)) {
    .check_number_in_valid_range(
      x = conf_int_alpha, 
      var_name = "conf_int_alpha", 
      range = c(0.0, 1.0), 
      closed = c(TRUE, TRUE)
    )
  }

  # Confidence intervals. Note that a confidence interval of 0.0 (equals FALSE)
  # is allowed.
  if (!is.waive(conf_int)) {
    .check_number_in_valid_range(
      x = conf_int, 
      var_name = "conf_int",
      range = c(0.0, 1.0),
      closed = c(TRUE, FALSE)
    )
  }
  
  # Limits to number of shown features.
  if (!is.waive(limit_n_features) && !is.null(limit_n_features)) {
    .check_is_integerish(
      x = limit_n_features,
      var_name = "limit_n_features"
    )
    .check_number_in_valid_range(
      x = limit_n_features,
      var_name = "limit_n_features",
      range = c(1L, Inf)
    )
  }
  
  # Facet wrap cols
  if (!is.waive(facet_wrap_cols) && !is.null(facet_wrap_cols)) {
    .check_number_in_valid_range(
      x = facet_wrap_cols,
      var_name = "facet_wrap_cols",
      range = c(1L, Inf), 
      closed = c(TRUE, TRUE)
    )
  }

  # Style of confidence intervals
  if (!is.waive(conf_int_style)) {
    .check_parameter_value_is_valid(
      x = conf_int_style, 
      var_name = "conf_int_style", 
      values = conf_int_default
    )
  }

  # Sharing of x-axis labels
  if (!is.waive(x_label_shared)) {
    .check_parameter_value_is_valid(
      x = x_label_shared, 
      var_name = "x_label_shared", 
      values = c("overall", "column", "individual")
    )
  }

  # Sharing of y-axis labels
  if (!is.waive(y_label_shared)) {
    .check_parameter_value_is_valid(
      x = y_label_shared, 
      var_name = "y_label_shared",
      values = c("overall", "row", "individual")
    )
  }

  # Rotation of tick labels on the x-axis by 90 degrees
  if (!is.waive(rotate_x_tick_labels)) {
    .check_parameter_value_is_valid(
      x = rotate_x_tick_labels, 
      var_name = "rotate_x_tick_labels",
      values = c(FALSE, TRUE)
    )
  }

  # Rotation of tick labels on the y-axis by 45 degrees
  if (!is.waive(rotate_y_tick_labels)) {
    .check_parameter_value_is_valid(
      x = rotate_y_tick_labels, 
      var_name = "rotate_y_tick_labels",
      values = c(FALSE, TRUE)
    )
  }
  
  return(invisible(TRUE))
}



.check_input_plot_range <- function(range_var, var_name) {
  # Generic range check
  if (!is.numeric(range_var)) {
    ..error(
      paste0(
        "The ", var_name, "argument should be a numerical vector of length 2, ",
        "indicating min and max of the range. Use NA to indicate an unset upper ",
        "or lower boundary."
      ),
      error_class = "input_argument_error"
    )
    
  } else if (length(range_var) != 2L) {
    ..error(
      paste0(
        "The ", var_name, " argument should be a numerical vector of length 2, ",
        "indicating min and max of the range. Use NA to indicate an unset upper ",
        "or lower boundary."
      ),
      error_class = "input_argument_error"
    )
    
  } else if (any(is.infinite(range_var))) {
    ..error(
      paste0(
        "The range provided by the ", var_name, " argument cannot have Inf or -Inf values. ",
        "Use NA to indicate an unset upper or lower boundary."
      ),
      error_class = "input_argument_error"
    )
    
  } else if (!anyNA(range_var)) {
    if (range_var[1L] == range_var[2L]) {
      ..error(
        paste0(
          "The range provided by the ", var_name, " argument cannot have the ",
          "same value for the lower and upper boundary."
        ),
        error_class = "input_argument_error"
      )
      
    } else if (range_var[1L] > range_var[2L]) {
      ..error(
        paste0(
          "The range provided by the ", var_name, " argument cannot have a lower ",
          "boundary with a higher value than the upper boundary."
        ),
        error_class = "input_argument_error"
      )
    }
  }
  
  return(invisible(TRUE))
}



.check_input_plot_n_breaks <- function(n_break_var, var_name) {
  # Generic number of breaks check
  if (!is.numeric(n_break_var)) {
    ..error(
      paste0("The ", var_name, " argument should be a single integer value of 2 or larger."),
      error_class = "input_argument_error"
    )
    
  } else if (length(n_break_var) != 1L) {
    ..error(
      paste0("The ", var_name, " argument should be a single integer value of 2 or larger."),
      error_class = "input_argument_error"
    )
    
  } else if (!is.finite(n_break_var)) {
    ..error(
      paste0(
        "The ", var_name, " argument should be a single integer value of 2 or larger. ",
        "It cannot be infinite or NA."
      ),
      error_class = "input_argument_error"
    )
    
  } else if (n_break_var < 2L) {
    ..error(
      paste0("The ", var_name, " argument should be a single integer value of 2 or larger."),
      error_class = "input_argument_error"
    )
  }
  
  return(invisible(TRUE))
}



.check_input_plot_break <- function(break_var, var_name) {
  # Generic break check
  if (is.null(break_var)) return(invisible(TRUE))

  if (is.numeric(break_var) && !all(is.finite(break_var))) {
    ..error(
      paste0("The ", var_name, " argument contains NA or infite values."),
      error_class = "input_argument_error"
    )
    
  } else if (anyNA(break_var)) {
    ..error(
      paste0("The ", var_name, " argument contains NA or infite values."),
      error_class = "input_argument_error"
    )
    
  } else if (anyDuplicated(break_var) > 0L) {
    ..error(
      paste0("Some breaks in the ", var_name, " argument are duplicates."),
      error_class = "input_argument_error"
    )
  }
  
  return(invisible(TRUE))
}



.check_input_plot_label <- function(label_var, var_name) {
  # Generic label check
  if (is.null(label_var) || is.call(label_var) || is.expression(label_var)) {
    return(invisible(TRUE))
  }

  if (!(is.numeric(label_var) || is.character(label_var))) {
    ..error(
      paste0(
        "The ", var_name, " argument should be NULL, a single number, a single string, ",
        "or a call or expression interpreted by grDevices::plotmath."
      ),
      error_class = "input_argument_error"
    )
    
  } else if (length(label_var) != 1L) {
    ..error(
      paste0("The ", var_name, " argument contains multiple labels, whereas a single label is expected."),
      error_class = "input_argument_error"
    )
  }
  
  return(invisible(TRUE))
}



.check_input_plot_legend <- function(label_var, var_name) {
  # Generic label check
  if (is.null(label_var) || is.call(label_var) || is.expression(label_var)) {
    return(invisible(TRUE))
  }

  # Legend labels can be a list
  if (is.list(label_var)) {
    sapply(
      label_var, 
      .check_input_plot_label, 
      var_name = var_name
    )

    return(invisible(TRUE))
  }

  if (!(is.numeric(label_var) || is.character(label_var))) {
    ..error(
      paste0(
        "The ", var_name, " argument should be NULL, a single number, a single string, ",
        "or a call or expression interpreted by grDevices::plotmath."
      ),
      error_class = "input_argument_error"
    )
    
  } else if (length(label_var) != 1L) {
    ..error(
      paste0("The ", var_name, " argument contains multiple labels, whereas a single label is expected."),
      error_class = "input_argument_error"
    )
  }
  
  return(invisible(TRUE))
}



.check_plot_grid_unit <- function(x, var_name) {
  if (!grid::is.unit(x)) {
    ..error(
      paste0(
        "The ", var_name, " argument should be a grid unit (see ?grid::unit), ",
        "but instead is ", paste_s(class(x))
      ),
      error_class = "input_argument_error"
    )
  }
  
  return(invisible(TRUE))
}

.compute_rank_borda <- function(
    x,
    rank_threshold,
    truncated = FALSE,
    enhanced = FALSE
) {
  # Borda selection methods, as in Wald R, Khoshgoftaar TM, Dittman D, Awada W,
  # Napolitano A. An extensive comparison of feature ranking aggregation
  # techniques in bioinformatics. 2012 IEEE 13th Int. Conf. Inf. Reuse Integr.,
  # IEEE; 2012, p. 377-84.
  #
  # The Borda score is defined sum(N-r_l), with r the rank in list l (of length
  # N). If all lists have the same length, this devolves into a mean scoring.
  # The enhanced borda method resolves this issue by multiplying the borda score
  # for each feature by the feature occurrence. The truncated borda method only
  # assigns a non-zero value if the feature is within the top features as
  # indicated by rank_threshold. Higher scores signify better features.

  # Suppress NOTES due to non-standard evaluation in data.table
  score <- occurrence <- rank <- max_rank <- borda_score <- sum_score <- NULL

  # Extract the variable importance table.
  vimp_table <- x@vimp_table

  # Determine the number of runs
  n_runs <- data.table::uniqueN(vimp_table, by = "run_id")

  # Make a local copy to avoid changing the concatenated input rank table by
  # reference.
  borda_count_table <- data.table::copy(vimp_table)

  if (truncated) {
    # Keep only those features with a rank equal or lower than rank_threshold.

    # Compute a normalised Borda score, with maximum 1 and minimum
    # 1/rank_threshold. In effect, the best feature always receives a score of
    # 1, down to the feature with rank equal to rank_threshold 1/rank_threshold.
    # Any remaining features receive a score of 0.
    borda_count_table[
      rank <= rank_threshold,
      "borda_score" := (rank_threshold - rank + 1L) / rank_threshold
    ]
    borda_count_table[is.na(borda_score), "borda_score" := 0.0]
    
  } else {
    # If not truncated, compute the standard borda score according to Wald et
    # al. Find the maximum rank within each ranking list subset (identified by
    # run_id).
    borda_count_table[, "max_rank" := max(rank), by = "run_id"]

    # Compute a normalised Borda score, with maximum 1 and minimum 1/max_rank.
    borda_count_table[, "borda_score" := (max_rank - rank + 1L) / max_rank]
  }

  # Sum all scores over the different lists
  borda_count_table <- borda_count_table[
    ,
    list("sum_score" = sum(borda_score)),
    by = "name"
  ]

  if (enhanced) {
    # Enhanced borda methods use the occurrence (as in stability selection) for
    # weighting the borda score.
    occurrence_table <- .compute_feature_occurrence(
      vimp_table = vimp_table,
      threshold = rank_threshold,
      n_runs = n_runs
    )

    # Merge tables by name
    vimp_table <- merge(
      x = borda_count_table,
      y = occurrence_table,
      all.x = FALSE,
      all.y = TRUE,
      by = c("name")
    )

    # Compute the enhanced borda score.
    vimp_table[, "score" := sum_score * occurrence]
    vimp_table[!is.finite(score), "score" := 0.0]
    
  } else {
    # No merging or further aggregation is required.
    vimp_table <- borda_count_table
    data.table::setnames(
      x = vimp_table, 
      old = "sum_score",
      new = "score"
    )
  }

  # Attach to vimpTable object.
  x@vimp_table <- vimp_table[, mget(c("name", "score"))]

  # Set correct invert value.
  x@invert <- TRUE

  return(x)
}

.get_available_data_elements <- function(
    check_has_estimation_type = FALSE,
    check_has_detail_level = FALSE, 
    check_has_sample_limit = FALSE,
    check_has_n_important_features = FALSE,
    check_from_prediction_table = FALSE,
    check_from_data_object = FALSE
) {
  
  # All data elements.
  all_data_elements <- c(
    "auc_data", "calibration_data", "calibration_info", "confusion_matrix",
    "decision_curve_analyis", "feature_expressions",
    "fs_vimp", "hyperparameters", "model_performance",
    "model_vimp", "permutation_vimp", "prediction_data",
    "risk_stratification_data", "risk_stratification_info",
    "univariate_analysis", "feature_similarity", "sample_similarity", "ice_data",
    "shap"
  )
  
  # Data elements that allow setting an estimation type.
  can_set_estimation_type <- c(
    "auc_data", "calibration_data", "decision_curve_analyis",
    "model_performance", "permutation_vimp",  "prediction_data", "ice_data"
  )
  
  # Data elements that allow setting a detail level.
  can_set_detail_level <- c(
    can_set_estimation_type, "calibration_info", "confusion_matrix",
    "risk_stratification_data", "risk_stratification_info", "shap"
  )
   
  # Data elements that allow for setting an estimation type but not detail
  # level.
  can_set_estimation_type <- c(can_set_estimation_type, "feature_similarity")
  
  # Data elements that allow for setting a sample limit.
  can_set_sample_limit <- c("sample_similarity", "ice_data", "shap", "permutation_vimp")
  
  # Data elements that allow for setting the number of important features.
  can_set_n_important_features <- c("permutation_vimp", "ice_data", "shap")
  
  # Data elements that can be computed from prediction table objects.
  can_use_prediction_table <- c(
    "prediction_data", "auc_data", "calibration_data", "decision_curve_analyis",
    "model_performance", "risk_stratification_data"
  )
  
  # Data elements that can be computed from data objects.
  can_use_data_object <- c("risk_stratification_data", "feature_similarity", "sample_similarity", "feature_expressions")
  
  if (check_has_sample_limit) {
    all_data_elements <- intersect(all_data_elements, can_set_sample_limit)
  }
  
  if (check_has_estimation_type) {
    all_data_elements <- intersect(all_data_elements, can_set_estimation_type)
  } 
  
  if (check_has_detail_level) {
    all_data_elements <- intersect(all_data_elements, can_set_detail_level)
  }
  
  if (check_from_prediction_table) {
    all_data_elements <- intersect(all_data_elements, can_use_prediction_table)
  }
  
  if (check_from_data_object) {
    all_data_elements <- intersect(all_data_elements, can_use_data_object)
  }
  
  return(all_data_elements)
}



.parse_detail_level <- function(
    x,
    object, 
    default, 
    data_element) {
  
  if (is.waive(x)) x <- object@settings$detail_level
  
  if (is.null(x)) return(default)
  
  # detail level is stored in a list, by data_element.
  if (is.list(x)) x <- x[[data_element]]
  
  if (is.null(x)) return(default)
  
  .check_parameter_value_is_valid(
    x = x,
    var_name = "detail_level",
    values = c("ensemble", "hybrid", "model")
  )
  
  return(x)
}



.parse_estimation_type <- function(
    x,
    object,
    default,
    data_element,
    detail_level,
    has_internal_bootstrap
) {
  
  # Change to default to point if the detail_level is model.
  if (detail_level == "model") default <- "point"
  
  # In case there is no internal bootstrap, we can only determine point
  # estimates for ensemble and model detail levels (but potentially more for
  # hybrid).
  if (
    !has_internal_bootstrap &&
    detail_level %in% c("ensemble", "model") &&
    default != "point"
  ) {
    default <- "point"
  }
  
  if (is.waive(x) && .hasSlot(object, "settings")) x <- object@settings$estimation_type
  
  if (is.null(x)) return(default)
  
  # detail level is stored in a list, by data_element.
  if (is.list(x)) x <- x[[data_element]]
  
  if (is.null(x)) return(default)
  
  .check_parameter_value_is_valid(
    x = x,
    var_name = "estimation_type",
    values = c(
      "point", "bias_correction", "bc",
      "bootstrap_confidence_interval", "bci"
    )
  )
  
  return(x)
}



.parse_aggregate_results <- function(
    x,
    object, 
    default, 
    data_element) {
  
  if (is.waive(x) && methods::.hasSlot(object, "settings")) {
    x <- object@settings$aggregate_results
    
  } else if (is.waive(x)) {
    return(default)
  }
  
  if (is.null(x)) return(default)
  
  # detail level is stored in a list, by data_element.
  if (is.list(x)) x <- x[[data_element]]
  
  if (is.null(x)) return(default)
  
  x <- tolower(x)
  .check_parameter_value_is_valid(
    x = x,
    var_name = "aggregate_results",
    values = c("true", "false", "none", "all", "default")
  )
  
  if (x == "default") return(default)
  if (x %in% c("true", "all")) return(TRUE)
  
  return(FALSE)
}



.parse_sample_limit <- function(
    x,
    object,
    default,
    data_element
) {
  if (is.waive(x) && .hasSlot(object, "settings")) x <- object@settings$sample_limit
  
  if (is.null(x)) return(default)
  
  # detail level is stored in a list, by data_element.
  if (is.list(x)) x <- x[[data_element]]
  
  if (is.null(x)) return(default)
  
  if (x == "default") return(default)
  
  .check_number_in_valid_range(
    x = x,
    var_name = "sample_limit",
    range = c(20L, Inf)
  )
  
  return(x)
}



.parse_n_important_features <- function(
    x,
    object,
    default,
    data_element
) {
  if (is.waive(x)) x <- object@settings$n_important_features
  
  if (is.null(x)) return(default)
  
  # detail level is stored in a list, by data_element.
  if (is.list(x)) x <- x[[data_element]]
  
  if (is.null(x)) return(default)
  
  if (x == "default") return(default)
  
  .check_number_in_valid_range(
    x = x,
    var_name = "n_important_features",
    range = c(1L, Inf)
  )
  
  return(x)
}



.select_important_features <- function(
    object,
    data,
    fallback_vimp_method = "mim",
    n_important_features = Inf
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  name <- rank <- NULL
  
  if (!(is(object, "familiarModel") || is(object, "familiarNoveltyDetector") || is(object, "familiarEnsemble"))) {
    ..error_reached_unreachable_code(paste0("invalid object class: ", class(object)))
  }
  
  # Check that the model has any features, i.e. is not naive.
  if (object@vimp_method %in% .get_available_no_features_vimp_methods()) return(NULL)
  if (length(object@model_features) == 0L) return(NULL)
  
  if (is(object, "familiarEnsemble")) {
    # Make sure that models are loaded:
    object <- load_models(object, suppress_auto_detach = TRUE)
    if (!model_is_trained(object)) return(NULL)
    
    vimp_table <- lapply(object@model_list, function(x) (x@vimp_table))
    vimp_aggregation_method <- object@model_list[[1L]]@vimp_aggregation_method
    vimp_rank_threshold <- object@model_list[[1L]]@vimp_rank_threshold
    
  } else {
    if (!model_is_trained(object)) return(NULL)
    
    vimp_table <- object@vimp_table
    vimp_aggregation_methpd <- object@vimp_aggregation_method
    vimp_rank_threshold <- object@vimp_rank_threshold
  }
  
  # Flatten lists, if necessary.
  if (rlang::is_bare_list(vimp_table)) {
    vimp_table <- unlist(vimp_table)
  }
  
  # Get available features for the model or ensemble.
  features <- features_after_clustering(
    features = object@model_features,
    feature_info_list = object@feature_info
  )
  if (length(features) <= n_important_features) return(features)
  
  # Determine which features are pre-assigned to the signature.
  signature_features <- names(object@feature_info)[sapply(object@feature_info, is_in_signature)]
  
  # Check that fallback is required: in case of none, random, or no features
  # being selected.
  use_fallback <- object@vimp_method %in% c(
    .get_available_none_vimp_methods(),
    .get_available_random_vimp_methods()
  )
  
  # Check that fallback is required: in case there is no variable importance
  # table.
  if (!use_fallback) {
    if (rlang::is_bare_list(vimp_table)) {
      use_fallback <- all(sapply(vimp_table, is_empty))
    } else {
      use_fallback <- is_empty(vimp_table)
    }
  }
  
  # If signature is used, don't use fall-back option.
  if (object@vimp_method %in% .get_available_signature_only_vimp_methods()) use_fallback <- FALSE
  
  # Set-up fallback vimp-table
  if (use_fallback) {
    # Spawn task to obtain variable importance tables.
    vimp_task <- methods::new(
      "familiarTaskVimp",
      project_id = object@project_id,
      vimp_method = fallback_vimp_method,
      data_id = object@data_id,
      run_id = object@run_id,
      file = NA_character_
    )
    
    # Fill details required to get the data, in case the data is delayed.
    # Note that training data is used for obtaining variable importance.
    if (is(data, "delayedDataObject")) {
      data@data_id <- object@data_id
      data@run_id <- object@run_id
    }
    
    # Create variable importance table.
    vimp_table <- .perform_task(
      object = vimp_task,
      feature_info_list = object@feature_info,
      vimp_aggregation_method = vimp_aggregation_method,
      vimp_rank_threshold = vimp_rank_threshold,
      data = data
    )
  }
  
  # For signature-only, return all signature features, with no preference.
  if (object@vimp_method %in% .get_available_signature_only_vimp_methods()) {
    # Only select signature.
    if (length(signature_features) == 0L) {
      ..error(
        "No signature was provided.",
        error_class = "input_argument_error"
      )
    }
    
    return(signature_features)
  }
    
  # Select signature and any additional features according to rank.
  selected_features <- signature_features
  
  # Get number remaining available features
  n_allowed_features <- n_important_features - length(signature_features)
  
  # Check that features may be added, and the rank table is not empty.
  if (n_allowed_features > 0L && !is_empty(vimp_table)) {

    # Remove signature features, if any, to prevent duplicates.
    features <- setdiff(features, signature_features)
    
    # Extract aggregated rank table. First, ensure that the associated cluster
    # table is correct.
    vimp_table <- update_vimp_table_to_reference(
      x = vimp_table,
      reference_cluster_table = .create_clustering_table(
        feature_info_list = object@feature_info
      )
    )
    
    # Recluster the data according to the clustering table corresponding to the
    # model. This ensures that the variable importance table has the features
    # that are seen by the model.
    vimp_table <- recluster_vimp_table(vimp_table)
    
    # Get aggregate variable importances
    vimp_table <- aggregate_vimp_table(
      vimp_table,
      aggregation_method = vimp_aggregation_method,
      rank_threshold = vimp_rank_threshold
    )
    
    if (is_empty(vimp_table)) return(signature_features)
    
    # Keep only feature ranks of feature corresponding to available
    # features, and order by rank.
    rank_table <- get_vimp_table(vimp_table)[name %in% features, ][order(rank)]
    
    # Add good features (low rank) to the selection
    selected_features <- c(
      signature_features,
      head(x = rank_table, n = n_allowed_features)$name
    )
  }
  
  if (length(selected_features) == 0L) return(NULL)
  
  return(selected_features)
}

.required_packages_backend <- function(backend_type) {
  
  backend_packages <- NULL
  if (backend_type == "socket_server") {
    backend_packages <- "callr"
  }
  
  return(backend_packages)
}


.get_available_backend_types <- function() {
  
  # All types of backend.
  all_backend_types <- c("none", "socket_server")
  
  return(all_backend_types)
}



.get_selected_backend_type <- function() {
  # Get the type of backend server from the familiar global environment or the
  # global environment. The latter happens for parallel processes.
  if (exists("familiar_global_env")) {
    if (exists("backend_type", where = familiar_global_env)) {
      data_env <- familiar_global_env
      
    } else if (exists("backend_type", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
    }
    
  } else if (exists("backend_type", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
    
  } else {
    ..error_reached_unreachable_code(
      ".get_selected_backend_type: backend_type was not found in familiar_global_env or .GlobalEnv"
    )
  }
  
  backend_type <- tryCatch(
    get("backend_type", envir = data_env),
    error = identity
  )
  
  if (inherits(backend_type, "error")) {
    .assign_backend_options_to_global(
      backend_type = "none",
      server_port = NULL
    )
    
    backend_type <- "none"
  }
  
  return(backend_type)
}



.get_backend_server_port <- function() {
  # Get the port for the backend server from the familiar global environment or
  # the global environment. The latter happens for parallel processes.
  if (exists("familiar_global_env")) {
    if (exists("server_port", where = familiar_global_env)) {
      data_env <- familiar_global_env
 
    } else if (exists("server_port", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
    }
    
  } else if (exists("server_port", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
    
  } else {
    ..error_reached_unreachable_code(
      ".get_backend_server_port: server_port was not found in familiar_global_env or .GlobalEnv"
    )
  }

  return(get("server_port", envir = data_env))
}



.assign_backend_options_to_global <- function(backend_type, server_port) {
  assign(
    "backend_type",
    value = backend_type,
    envir = familiar_global_env
  )
  assign(
    "server_port",
    value = server_port,
    envir = familiar_global_env
  )
  
  return(invisible(TRUE))
}



.assign_data_to_backend <- function(
    data, 
    backend_type = NULL, 
    server_port = NULL
) {

  # Find the server port and backend_type variables if not directly provided.
  if (is.null(server_port)) server_port <- .get_backend_server_port()
  if (is.null(backend_type)) backend_type <- .get_selected_backend_type()
  
  # Check if required packages are installed.
  require_package(
    x = .required_packages_backend(backend_type = backend_type),
    purpose = paste0("to use the requested backend (", backend_type, ")")
  )
  
  if (backend_type %in% c("none")) {
    # Put master_data in global environment
    assign(
      x = "master_data",
      value = data,
      envir = familiar_global_env
    )
    
  } else if (backend_type %in% c("socket_server")) {
    # Start the separate r session thread that will run the socket server.
    start_socket_server_process(server_port = server_port)
    
    # Obtain the handle.
    socket_server_process <- .get_socket_server_process_handle()
    
    # Assign the data to socket server process so that its available later.
    socket_server_process$run(
      function(data) {
        assign(
          x = "master_data",
          value = data,
          envir = familiar_global_env
        )
      },
      args = list("data" = data),
      package = TRUE
    )
    
    # Finally, activate the server subroutine.
    .activate_socket_server_routine(server_port = server_port)
    
  } else {
    ..error_reached_unreachable_code(".assign_data_to_backend: unknown backend encountered")
  }
  
  return(invisible(TRUE))
}



get_data_from_backend <- function(
    backend_type = NULL,
    server_port = NULL,
    sample_identifiers = NULL,
    column_names = NULL
) {

  # Find the server port and backend_type variables if not directly provided.
  if (is.null(server_port)) server_port <- .get_backend_server_port()
  if (is.null(backend_type)) backend_type <- .get_selected_backend_type()
  
  # Check if required packages are installed.
  require_package(
    x = .required_packages_backend(backend_type = backend_type),
    purpose = paste0("to use the requested backend (", backend_type, ")")
  )
  
  if (backend_type == "none") {
    # Absence of backend.
    x <- .get_data_from_backend(
      sample_identifiers = sample_identifiers,
      column_names = column_names
    )
    
  } else if (backend_type == "socket_server") {
    # Socket server backend.
    
    # Open connection to socket server.
    con <- socket_client_open_connection(port = server_port)
    on.exit(socket_client_close_connection(con))
    
    # Execute .get_data_back_end on the server side, and obtain the data view.
    x <- socket_client_do_call(
      con,
      .get_data_from_backend,
      args = list(
        "sample_identifiers" = sample_identifiers,
        "column_names" = column_names
      )
    )
    
  } else {
    ..error_reached_unreachable_code(paste0(
      "get_data_from_backend: encountered unknown type of backend: ", backend_type
    ))
  }
  
  return(x)
}



.get_data_from_backend <- function(sample_identifiers = NULL, column_names = NULL) {
  # This is the support function that operates on the processes 
  
  # Suppress NOTES due to non-standard evaluation in data.table
  .NATURAL <- NULL
  
  # Identify the environment where the master_data object lives.
  if (exists("familiar_global_env")) {
    if (exists("master_data", where = familiar_global_env)) {
      data_env <- familiar_global_env
      
    } else if (exists("master_data", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
    }
    
  } else if (exists("master_data", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
    
  } else {
    ..error_reached_unreachable_code(
      ".get_data_from_backend: master_data was not found in familiar_global_env or .GlobalEnv"
    )
  }
  
  # Obtain and export views of the master data set to the calling function.
  if (is.null(sample_identifiers) && is.null(column_names)) {
    # Get entire dataset.
    x <- data.table::copy(get("master_data", envir = data_env))
    
  } else if (is.null(sample_identifiers) && !is.null(column_names)) {
    # Get entire columns from the dataset.
    x <- data.table::copy(get("master_data", envir = data_env)[, mget(column_names)])
    
  } else if (!is.null(sample_identifiers) && is.null(column_names)) {
    # Get entire rows from the dataset.
    x <- data.table::copy(get("master_data", envir = data_env)[
      sample_identifiers, on = .NATURAL
    ])
    
  } else {
    # Get certain columns and rows from the dataset.
    x <- data.table::copy(get("master_data", envir = data_env)[
      sample_identifiers, mget(column_names), on = .NATURAL
    ])
  }
  
  return(x)
}



.assign_feature_info_to_backend <- function(
    feature_info_list,
    backend_type = NULL,
    server_port = NULL
) {

  # Find the server port and backend_type variables if not directly provided.
  if (is.null(server_port)) server_port <- .get_backend_server_port()
  if (is.null(backend_type)) backend_type <- .get_selected_backend_type()
  
  # Check if required packages are installed.
  require_package(
    x = .required_packages_backend(backend_type = backend_type),
    purpose = paste0("to use the requested backend (", backend_type, ")")
  )
  
  # Put master_feature_info_list in the familiar global environment.
  assign(
    x = "master_feature_info_list",
    value = feature_info_list,
    envir = familiar_global_env
  )
  
  if (backend_type %in% c("socket_server")) {
    # Stop the socket_server subroutine so that we can add or update the list of
    # feature information.
    .shutdown_active_socket_server_routine(server_port = server_port)
    
    # Obtain the handle.
    socket_server_process <- .get_socket_server_process_handle()
    
    # Assign the list of feature information to the socket server process so
    # that its accessible later.
    socket_server_process$run(
      function(feature_info_list) {
        # Assign locally.
        assign(
          x = "master_feature_info_list",
          value = feature_info_list,
          envir = familiar_global_env
        )
      },
      args = list("feature_info_list" = feature_info_list),
      package = TRUE
    )
    
    # Reactivate the server subroutine.
    .activate_socket_server_routine(server_port = server_port)
    
  } else if (backend_type != "none") {
    ..error_reached_unreachable_code(
      ".assign_feature_info_to_backend: unknown backend encountered"
    )
  }
}


get_feature_info_from_backend <- function(
    backend_type = NULL,
    server_port = NULL,
    data_id = NULL,
    run_id = NULL
) {
  
  # Find the server port and backend_type variables if not directly provided.
  if (is.null(server_port)) server_port <- .get_backend_server_port()
  if (is.null(backend_type)) backend_type <- .get_selected_backend_type()
  
  # Check if required packages are installed.
  require_package(
    x = .required_packages_backend(backend_type = backend_type),
    purpose = paste0("to use the requested backend (", backend_type, ")")
  )
  
  if (backend_type == "none") {
    # Absence of backend.
    x <- .get_feature_info_from_backend(data_id = data_id, run_id = run_id)
    
  } else if (backend_type == "socket_server") {
    # Socket server backend.
    
    # Open connection to socket server.
    con <- socket_client_open_connection(port = server_port)
    on.exit(socket_client_close_connection(con))
    
    # Execute .get_feature_info_from_backend on the server side, and obtain the
    # requested list of feature information.
    x <- socket_client_do_call(
      con,
      FUN = .get_feature_info_from_backend,
      args = list(
        "data_id" = data_id,
        "run_id" = run_id
      )
    )
    
  } else {
    ..error_reached_unreachable_code(paste0(
      "get_feature_info_from_backend: encountered unknown type of backend: ", backend_type
    ))
  }
  
  return(x)
}



.get_feature_info_from_backend <- function(data_id = NULL, run_id = NULL) {
  # This function is executed locally in the thread where
  # master_feature_info_list resides.
  
  # Identify the environment where the master_feature_info_list object lives.
  if (exists("familiar_global_env")) {
    if (exists("master_feature_info_list", where = familiar_global_env)) {
      data_env <- familiar_global_env
      
    } else if (exists("master_feature_info_list", where = .GlobalEnv)) {
      data_env <- .GlobalEnv
    }
    
  } else if (exists("master_feature_info_list", where = .GlobalEnv)) {
    data_env <- .GlobalEnv
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".get_feature_info_from_backend: master_feature_info_list was not found ", 
      "in familiar_global_env or .GlobalEnv."
    ))
  }
  
  # Now read the requested element.
  if (is.null(data_id) && is.null(run_id)) {
    # Retrieve the generic feature information.
    x <- get("master_feature_info_list", envir = data_env)[["generic"]]
    
  } else if (is.waive(data_id) && is.waive(run_id)) {
    # Retrieve all feature information, i.e. the complete list.
    x <- get("master_feature_info_list", envir = data_env)
    
  } else if (!is.null(data_id) && !is.null(run_id)) {
    # Retrieve run-specific feature information.
    x <- get("master_feature_info_list", envir = data_env)[[
      paste0(data_id, ".", run_id)
    ]]
    
  } else {
    # data_id and run_id are either both provided, or both NULL.
    ..error_reached_unreachable_code(paste0(
      ".get_feature_info_from_backend: one of data_id and run_id is NULL ", 
      "instead of both NULL or both not NULL."
    ))
  }
  
  # The feature info list should not be empty.
  if (is.null(x)) {
    ..error_reached_unreachable_code(paste0(
      ".get_feature_info_from_backend: the requested feature information is empty. ",
      "data_id = ", data_id,
      "; run_id = ", run_id
    ))
  }
  
  return(x)
}



get_os <- function() {
  # Gets the operating system name
  system_info <- Sys.info()

  os_name <- system_info["sysname"]
  if (os_name == "Darwin") {
    os_name <- "mac_os"
  }
  
  return(tolower(os_name))
}



start_socket_server_process <- function(server_port = NULL) {
  
  # Find the server port if not directly provided.
  if (is.null(server_port)) server_port <- .get_backend_server_port()
  
  # Check if required packages are installed.
  require_package(
    x = .required_packages_backend(backend_type = "socket_server"),
    purpose = paste0("to use the requested backend (socket_server)")
  )
  
  # Check if the process is already running.
  if (!.is_socket_server_process_started()) {
    # Start the socket_server process in a separate persistent thread using
    # the callr package.
    socket_server_process <- callr::r_session$new()
    
    # Assign initial data to the 
    socket_server_process$run(
      function(server_port) {
        # Load the familiar package locally.
        library(familiar)
        
        # Assign the socket server port to the familiar global environment.
        assign(
          x = "server_port",
          value = server_port,
          envir = familiar_global_env
        )
      },
      args = list("server_port" = server_port),
      package = TRUE
    )
    
    # Export socket_server_process to the familiar global environment.
    assign(
      "socket_server_process",
      socket_server_process,
      envir = familiar_global_env
    )
  }
  
  return(invisible(TRUE))
}



.get_socket_server_process_handle <- function() {
  # Return the R6 callr::r_session object that serves as the handle for the
  # socket server process thread. This object is normally stored in the familiar
  # global environment.
  if (!exists("socket_server_process", where = familiar_global_env)) {
    ..error_reached_unreachable_code(paste0(
      ".get_socket_server_process_handle: socket_server_process ",
      "does not exist in familiar_global_env"
    ))
  }
  
  return(get("socket_server_process", envir = familiar_global_env))
}



.is_socket_server_process_started <- function() {
  # Check if the process exists
  if (!exists("socket_server_process", where = familiar_global_env)) return(FALSE)
  
  # Get handle
  socket_server_process <- .get_socket_server_process_handle()
  
  # Check if the process is alive
  is_alive <- socket_server_process$is_alive()
  
  return(is_alive)
}



.activate_socket_server_routine <- function(server_port = NULL) {
  
  # Get handle
  socket_server_process <- .get_socket_server_process_handle()
  
  # Check whether the socket server process is running in a separate thread.
  if (!socket_server_process$is_alive()) {
    ..error_reached_unreachable_code(paste0(
      ".activate_socket_server: the socket_server_process that hosts the ",
      "socket_server routine is not active."
    ))
  }
  
  # Check if the socket_server_process is idle.
  if (!socket_server_process$get_state() == "idle") {
    ..error_reached_unreachable_code(paste0(
      ".activate_socket_server: the socket_server_process that hosts the ",
      "socket_server routine is busy."
    ))
  }
  
  # Find the server port if not directly provided.
  if (is.null(server_port)) server_port <- .get_backend_server_port()
  
  # Start subroutine.
  socket_server_process$call(
    socket_server,
    args = list("port" = server_port),
    package = TRUE
  )
  
  return()
}



.shutdown_active_socket_server_routine <- function(server_port = NULL) {
  # Shuts down the socket_server routine running on the separate
  # socket_server_process callr::r_session thread.
  
  # Check if the server thread is active prior to shutting down the
  # socket_server subroutine.
  if (.is_socket_server_process_started()) {
    
    # Get handle
    socket_server_process <- .get_socket_server_process_handle()
    
    # Check the state of the process (idle or busy)
    process_state <- socket_server_process$get_state()
    
    # The process state will be "busy" when the socket_server subroutine is
    # active, because it loops. If its "idle" the subroutine is not longer
    # running.
    if (process_state == "busy") {
      
      # Find the server port if not directly provided.
      if (is.null(server_port)) server_port <- .get_backend_server_port()
      
      # Close the socket_server routine
      socket_client_server_shutdown(port = server_port)

      # The process should be responsive within a few milliseconds after sending
      # the command to stop the subroutine. We leave the process 5 seconds to
      # respond.      
      process_ready <- socket_server_process$poll_process(5000.0)
      
      # Read from the socket_server_process to free up the process again,
      # thereby finalising the shutdown of the socket_server subroutine and
      # returning the socket_server_process to its idle state.
      output <- socket_server_process$read()
    }
  }
  
  return(invisible(TRUE))
}



shutdown_backend_server <- function(backend_type = NULL, server_port = NULL) {
  
  # Find the server port and backend_type variables if not directly provided.
  if (is.null(server_port)) server_port <- .get_backend_server_port()
  if (is.null(backend_type)) backend_type <- .get_selected_backend_type()
  
  # Check if required packages are installed.
  require_package(
    x = .required_packages_backend(backend_type = backend_type),
    purpose = paste0("to use the requested backend (", backend_type, ")")
  )
  
  if (backend_type %in% c("socket_server")) {
    
    # Check if the server thread is active prior to shutting it down.
    if (.is_socket_server_process_started()) {
      
      # Attempt to shutdown the socket_server subroutine itself, if it is
      # running.
      .shutdown_active_socket_server_routine(server_port = server_port)
      
      # Get handle
      socket_server_process <- .get_socket_server_process_handle()
      
      # Stop the socket_server_process thread.
      socket_server_process$close()
    }
  }
  
  return()
}

# Clock for process timing.
ProcessClock <- setRefClass(
  "processClock",
  fields = list(
    clock_process = "ANY",
    start_time = "numeric",
    interrupt_time = "ANY"
  ),
  methods = list(
    "initialize" = function() {
      # Make sure that the packages are installed and can their namespaces can
      # be attached.
      require_package(c("callr", "microbenchmark"),
        purpose = "to measure process time"
      )

      # Set-up fields.
      .self$initFields(
        clock_process = NULL,
        start_time = 0.0,
        interrupt_time = NULL
      )

      # Start the clock process to look for interrupts, such as by
      # hibernating the system.
      .self$clock_process <- callr::r_bg(function() {
        time_start <- microbenchmark::get_nanotime()
        while (TRUE) {
          # Sleep for some time (hypothetically, 50ms, but realistically
          # anywhere between 50 and 90ms). Sys.sleep is not really accurate.
          # Luckily accuracy is not really required. Its just that most process
          # time should be spend asleep.
          Sys.sleep(0.05)

          # Measure the current time.
          time_current <- microbenchmark::get_nanotime()

          # Warn for interrupts if, for some reason (e.g. process suspension, OS
          # hibernation, etc.) the process slept for at least 1 second.
          if ((time_current - time_start) > 1.0E9) {
            cat(format(time_start, scientific = FALSE),
              ":",
              format(time_current, scientific = FALSE),
              "\n",
              sep = ""
            )

            # Correct for longer writing times.
            time_start <- microbenchmark::get_nanotime()
          } else {
            # Assume that the if-statement completes instantly.
            time_start <- time_current
          }
        }
      })

      # Set the start time. This is set after creating the clock process,
      # because launching the callr::r_bg asynchronous R background environment
      # takes a non-trivial amount of time.
      .self$start_time <- microbenchmark::get_nanotime()
    },
    
    "update_interrupt" = function() {
      # Suppress NOTES due to non-standard evaluation in data.table
      time_end <- time_start <- NULL

      # Read any interrupted time from the clock process. For some unobvious
      # reason the first read may not generate anything, but the second will.
      interrupt <- c(
        .self$clock_process$read_output_lines(),
        .self$clock_process$read_output_lines()
      )

      if (length(interrupt) > 0L) {
        # Split into start and endtimes.
        interrupt <- strsplit(x = interrupt, split = ":")

        interrupt_data <- data.table::data.table(
          "time_start" = as.numeric(sapply(interrupt, function(x) x[1L], USE.NAMES = FALSE)),
          "time_end" = as.numeric(sapply(interrupt, function(x) x[2L], USE.NAMES = FALSE))
        )

        # Update interrupt_time field.
        .self$interrupt_time <- rbind(.self$interrupt_time, interrupt_data)
      }
    },
    
    "accumulate_interrupt" = function(ref_start_time, ref_end_time) {
      # Suppress NOTES due to non-standard evaluation in data.table
      time_end <- time_start <- NULL

      # If there are no recorded interrupts, rejoice and return 0.
      if (is.null(.self$interrupt_time)) return(0.0)

      # Select only data that contains (part of) the reference time interval.
      # Work on a copy of the data to avoid updating by reference.
      interrupt_data <- data.table::copy(
        .self$interrupt_time[ref_start_time < time_end & ref_end_time > time_start]
      )

      # If all interrupts fall outside the reference time interval, return 0.
      if (nrow(interrupt_data) == 0L) return(0.0)

      # Update partial interrupt windows.
      interrupt_data[ref_start_time > time_start, "time_start" := ref_start_time]
      interrupt_data[ref_end_time < time_end, "time_end" := ref_end_time]

      # Set measured interrupt durations.
      interrupt_data[, "duration" := time_end - time_start]

      # Sum durations.
      total_interrupt_duration <- sum(interrupt_data$duration)

      # Return 0 if the total interrupt was smaller than 1s.
      if (total_interrupt_duration < 1.0E9) return(0.0)

      return(total_interrupt_duration)
    },
    
    "time" = function(units = "s", reference_time = NULL) {
      # Find current time.
      current_time <- microbenchmark::get_nanotime()

      # If the reference time is not provided, use the process start time field
      # instead.
      if (is.null(reference_time)) reference_time <- .self$start_time

      # Update tables with interrupt times, if any.
      .self$update_interrupt()

      # Accumulate interrupt time.
      sum_interrupt_time <- .self$accumulate_interrupt(
        ref_start_time = reference_time,
        ref_end_time = current_time
      )

      conversion_factor <- switch(units,
        "secs" = 1.0E9,
        "sec" = 1.0E9,
        "s" = 1.0E9,
        "mins" = 60.0E9,
        "min" = 60.0E9,
        "m" = 60.0E9,
        "hours" = 3600.0E9,
        "hour" = 3600.0E9,
        "h" = 3600.0E9,
        "millisecs" = 1.0E6,
        "millisec" = 1.0E6,
        "milli" = 1.0E6,
        "ms" = 1.0E6,
        "microsecs" = 1.0E3,
        "microsec" = 1.0E3,
        "micro" = 1.0E3,
        "us" = 1.0E3,
        "nanosecs" = 1.0,
        "nanosec" = 1.0,
        "nano" = 1.0,
        "ns" = 1.0
      )

      return((current_time - sum_interrupt_time - reference_time) / conversion_factor)
    },
    
    "close" = function() {
      .self$clock_process$kill()
    },
    
    "suspend" = function() {
      .self$clock_process$suspend()
      invisible(TRUE)
    },
    
    "resume" = function() {
      .self$clock_process$resume()
      invisible(TRUE)
    }
  )
)

testthat::skip_on_cran()
testthat::skip_on_ci()

# Initiate process clock.
pc <- familiar:::ProcessClock()

initial_time <- pc$time()
Sys.sleep(1)
time_1 <- pc$time()
pc$suspend()
Sys.sleep(1)
time_2 <- pc$time()
time_mark <- microbenchmark::get_nanotime()
Sys.sleep(1)
pc$resume()
time_3 <- pc$time()
pc$suspend()
Sys.sleep(1)
pc$resume()
time_4 <- pc$time()
time_diff <- pc$time(reference_time = time_mark)
pc$close()

testthat::test_that("Assert that process time is correctly recorded.", {
  testthat::expect_equal(initial_time, 0.0, tolerance = 0.1)
  testthat::expect_equal(time_1, 1.0, tolerance = 0.1)
  testthat::expect_equal(time_2, 2.0, tolerance = 0.2)
  testthat::expect_equal(time_3, 3.0, tolerance = 0.2)
  testthat::expect_equal(time_4, 4.0, tolerance = 0.3)
  testthat::expect_equal(time_diff, 2.0, tolerance = 0.2)
})

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# .file_exists (generic task) --------------------------------------------------
setMethod(
  ".file_exists",
  signature(object = "familiarTask"),
  function(object, ...) {
    if (is.na(object@file) || is.null(object@file)) return(FALSE)
    
    return(file.exists(object@file))
  }
)



# .get_current_run_table (generic task) ----------------------------------------
setMethod(
  ".get_current_run_table",
  signature(object = "familiarTask"),
  function(object, ...) {
    if (is_empty(object@run_table)) return(NULL)
    if (is.na(object@data_id) || is.na(object@run_id)) return(NULL)
    
    run_table <- object@run_table[[paste0(object@data_id, ".", object@run_id)]]
    if (!data.table::is.data.table(run_table)) return(NULL)
    
    return(run_table)
  }
)



# .get_feature_info_list (general task) ----------------------------------------
setMethod(
  ".get_feature_info_list",
  signature(object = "familiarTask"),
  function(object, feature_info_list, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    can_pre_process <- NULL
    
    # Attempt to get the feature info list from the backend.
    if (is.null(feature_info_list) && !is.null(object@run_table)) {
      # Find the last entry that is available for pre-processing
      run_table <- object@run_table[[paste0(object@data_id, ".", object@run_id)]]
      pre_processing_run <- tail(run_table[can_pre_process == TRUE, ], n = 1L)
      
      feature_info_list <- tryCatch(
        get_feature_info_from_backend(
          data_id = pre_processing_run$data_id[1L],
          run_id = pre_processing_run$run_id[1L]
        ),
        error = NULL
      )
    }
    
    # If no feature list is present on the backend, check other options.
    if (is.null(feature_info_list) && is.na(object@feature_info_file)) {
      # Check that a feature info list is provided, otherwise create an ad-hoc
      # list as an template.
      
      # Set up task, and explicitly don't write to file.
      generic_feature_info_task <- methods::new(
        "familiarTaskFeatureInfo",
        project_id = object@project_id,
        file = NA_character_
      )
      
      # Execute the task.
      feature_info_list <- .perform_task(
        object = generic_feature_info_task,
        ...
      )
      
    } else if (is.null(feature_info_list)) {
      # Assume that the feature info file attribute contains the path to the
      # file containing feature info.
      if (!file.exists(object@feature_info_file)) {
        ..error(paste0("feature info file does not exist at location: ", object@feature_info_file))
      }
      feature_info_list <- readRDS(object@feature_info_file)
      feature_info_list <- update_object(feature_info_list)
      
    } else if (is.character(feature_info_list)) {
      # If the feature info list is a string, interpret this as a path to the
      # file containing the feature info.
      if (!file.exists(feature_info_list)) {
        ..error(paste0("feature info file does not exist at location: ", feature_info_list))
      }
      feature_info_list <- readRDS(feature_info_list)
      feature_info_list <- update_object(feature_info_list)
    }
    
    if (!rlang::is_bare_list(feature_info_list)) {
      ..error("no feature info objects were found.")
    }
    
    return(feature_info_list)
  }
)



# .get_variable_importance_table (general task) --------------------------------
setMethod(
  ".get_variable_importance_table",
  signature(object = "familiarTask"),
  function(
    object,
    vimp_table,
    experiment_data = NULL,
    feature_info_list = NULL,
    file_paths = NULL,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    vimp <- main_data_id <- data_id <- run_id <- NULL
    
    if (is.null(vimp_table) && !is.null(object@run_table)) {
      # This routine loads variable importances from disk, and is used when an
      # experiment is run using summon_familiar.
      
      # This check exists to make sure that the standard workflow passes the
      # correct objects.
      if (is.null(file_paths)) {
        ..error_reached_unreachable_code("file_paths was expected, but not provided.")
      }
      
      # Find the data and run ids corresponding to variable importance tables
      # relevant to the current run. First we will figure out the data id and
      # run id for ALL variable importance tables.
      
      # Check that the variable importance stage actually exists --> this may be
      # lacking if variable importance is determined during hyperparameter
      # optimisation.
      if (all(!experiment_data@experiment_setup$vimp)) {
        if (object@vimp_method %in% c(
          .get_available_random_vimp_methods(),
          .get_available_none_vimp_methods(),
          .get_available_signature_only_vimp_methods()
        )) {
          return(NULL)
        }
        
        ..error_reached_unreachable_code("Cannot form variable importance tables. Something is wrong.")
      }
      
      vimp_data_id <- experiment_data@experiment_setup[vimp == TRUE, ]$main_data_id[1L]
      vimp_run_ids <- seq_len(experiment_data@experiment_setup[main_data_id == vimp_data_id, ]$n_runs[1L])
      
      # Select all run tables related to variable importance computation.
      vimp_run_tables <- object@run_table[paste(vimp_data_id, vimp_run_ids, sep = ".")]
      
      # Get the run table for training.
      train_run_table <- object@run_table[[paste0(object@data_id, ".", object@run_id)]]
      
      # Iterate backwards on data ids for the train run table to find matching
      # vimp run tables.
      matching <- logical(length(vimp_run_tables))
      train_data_chain_ids <- rev(train_run_table$data_id)
      ii <- 1L
      
      while (!any(matching)) {
        current_data_id <- train_data_chain_ids[ii]
        current_run_id <- train_run_table[data_id == current_data_id, ]$run_id[1L]
        
        for (jj in seq_along(matching)) {
          matching[jj] <- !is_empty(vimp_run_tables[[jj]][data_id == current_data_id & run_id == current_run_id])
        }
        ii <- ii + 1L
      }
      
      # Select matching variable importance run tables.
      vimp_run_tables <- vimp_run_tables[matching]
      vimp_run_ids <- sapply(vimp_run_tables, function(x) (tail(x, n = 1L)$run_id), simplify = TRUE, USE.NAMES = FALSE)
      
      # Get variable importance tables from disk.
      vimp_table <- list()
      for (ii in seq_along(vimp_run_ids)) {
        vimp_table_file <- get_object_file_name(
          project_id = object@project_id,
          data_id = vimp_data_id,
          run_id = vimp_run_ids[ii],
          vimp_method = object@vimp_method,
          object_type = "vimpTable",
          dir_path = file_paths$vimp_dir
        )
        
        if (file.exists(vimp_table_file)) {
          vimp_table[[ii]] <- update_object(readRDS(vimp_table_file))
          
        } else {
          ..error(paste0(
            "A variable importance table object was expected on disk but was not found: ",
            vimp_table_file
          ))
        }
      }
    }
    
    if (is.null(vimp_table) && is.na(object@vimp_table_file)) {
      # Create an ad-hoc list of variable importances.
      
      # Set up task, and explicitly don't write to file.
      vimp_task <- methods::new(
        "familiarTaskVimp",
        project_id = object@project_id,
        vimp_method = object@vimp_method,
        file = NA_character_
      )
      
      # Execute the task.
      vimp_table <- .perform_task(
        object = vimp_task,
        feature_info_list = feature_info_list,
        ...
      )
      
    } else if (is.null(vimp_table)) {
      # Assume that the vimp_table_file attribute contains the path to the
      # file containing the variable importance table.
      if (!file.exists(object@vimp_table_file)) {
        ..error(paste0("variable importance table file does not exist at location: ", object@vimp_table_file))
      }
      vimp_table <- update_object(readRDS(object@vimp_table_file))
      
    } else if (is.character(vimp_table)) {
      # If hyperparameters is a string, interpret this as a path to the
      # file containing the vimp method hyperparameters.
      if (!file.exists(vimp_table)) {
        ..error(paste0("variable importance table file does not exist at location: ", vimp_table))
      }
      vimp_table <- update_object(readRDS(vimp_table))
    }
    
    if (!(rlang::is_bare_list(vimp_table) || is(vimp_table, "vimpTable"))) {
      ..error("No variable importance table was found.")
    }
    
    return(vimp_table)
  }
)



.generate_learner_data_preprocessing_tasks <- function(
    experiment_data,
    file_paths
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  train <- can_pre_process <- perturbation_level <- main_data_id <- NULL
  
  # Find the data_id related to training.
  train_data_id <- experiment_data@experiment_setup[train == TRUE, ]$main_data_id[1L]
  if (is_empty(train_data_id)) return(NULL)
  
  # Determine which parts of the experimental setup are used by training.
  run_table <- .get_run_table_from_experiment_setup(
    data_id = train_data_id,
    experiment_setup = experiment_data@experiment_setup
  )
  
  # Find the data_id and run_ids for preprocessing.
  pre_process_data_id <- tail(run_table[main_data_id <= train_data_id & can_pre_process == TRUE], n = 1L)$main_data_id[1L]
  pre_process_run_ids <- seq_len(run_table[main_data_id == pre_process_data_id]$n_runs[1L])
  
  # Set up tasks.
  task_list <- .generate_data_preprocessing_tasks(
    data_ids = pre_process_data_id,
    run_ids = pre_process_run_ids,
    file_paths = file_paths,
    project_id = experiment_data@project_id
  )
  
  return(task_list)
}



.generate_vimp_data_preprocessing_tasks <- function(
    experiment_data,
    file_paths
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  vimp <- can_pre_process <- main_data_id <- NULL
  
  # Find the data_id related to computing variable importance.
  vimp_data_id <- experiment_data@experiment_setup[vimp == TRUE, ]$main_data_id[1L]
  if (is_empty(vimp_data_id)) return(NULL)
  
  # Determine which parts of the experimental setup are used for assessing
  # variable importance..
  run_table <- .get_run_table_from_experiment_setup(
    data_id = vimp_data_id,
    experiment_setup = experiment_data@experiment_setup
  )
  
  # Find the data_id and run_ids for preprocessing.
  pre_process_data_id <- tail(run_table[main_data_id <= vimp_data_id & can_pre_process == TRUE], n = 1L)$main_data_id[1L]
  pre_process_run_ids <- seq_len(run_table[main_data_id == pre_process_data_id]$n_runs[1L])
  
  # Set up tasks.
  task_list <- .generate_data_preprocessing_tasks(
    data_ids = pre_process_data_id,
    run_ids = pre_process_run_ids,
    file_paths = file_paths,
    project_id = experiment_data@project_id
  )
  
  return(task_list)
}



.sort_tasks <- function(task_list) {
  # Select unique tasks.
  duplicate_tasks <- duplicated(sapply(task_list, FUN = .get_task_descriptor))
  task_list <- task_list[!duplicate_tasks]
  
  # Determine class of tasks.
  task_class <- sapply(task_list, class)
  
  task_list <- list(
    "generic_feature_info" = task_list[task_class == "familiarTaskGenericFeatureInfo"],
    "feature_info" = task_list[task_class == "familiarTaskFeatureInfo"],
    "hyperparameters_vimp" = task_list[task_class == "familiarTaskVimpHyperparameters"],
    "vimp" = task_list[task_class == "familiarTaskVimp"],
    "hyperparameters_learner" = task_list[task_class == "familiarTaskLearnerHyperparameters"],
    "train" = task_list[task_class == "familiarTaskTrain"],
    "evaluate" = task_list[task_class == "familiarTaskEvaluate"],
    "collect" = task_list[task_class == "familiarTaskCollect"]
  )
  
  # Update task_id and n_tasks attribute of the tasks.
  task_list <- lapply(
    task_list,
    function(x) {
      lapply(
        seq_along(x),
        function(ii, x, n) {
          object <- x[[ii]]
          object@task_id <- ii
          object@n_tasks <- n
          return(object)
        },
        x = x,
        n = length(x)
      )
    }
  )
  
  return(task_list)
}

..required_plotting_packages <- function(extended = FALSE) {
  plot_packages <- c("ggplot2", "labeling", "scales", "rlang")

  if (extended) plot_packages <- c(plot_packages, "gtable")

  return(plot_packages)
}



#' Familiar ggplot2 theme
#'
#' This is the default theme used for plots created by familiar. The theme uses
#' `ggplot2::theme_light` as the base template.
#'
#' @param base_size Base font size in points. Size of other plot text elements
#'   is based off this.
#' @param base_family Font family used for text elements.
#' @param base_line_size Base size for line elements, in points.
#' @param base_rect_size Base size for rectangular elements, in points.
#'
#' @return A complete plotting theme.
#' @export
theme_familiar <- function(
    base_size = 10.0,
    base_family = "",
    base_line_size = 0.5,
    base_rect_size = 0.5
) {
  
  # The default familiar theme is based on ggplot2::theme_light.
  ggtheme <- ggplot2::theme_light(
    base_size = base_size,
    base_family = base_family,
    base_line_size = base_line_size,
    base_rect_size = base_rect_size
  )

  # Set colour to black.
  ggtheme$axis.text$colour <- "black"
  ggtheme$axis.ticks$colour <- "black"
  ggtheme$axis.line <- ggplot2::element_line(
    colour = "black",
    lineend = "square"
  )

  # Legend does not have a legend or border, and reserves less space.
  ggtheme$legend.background <- ggplot2::element_blank()
  ggtheme$legend.key <- ggplot2::element_blank()
  ggtheme$legend.key.size <- grid::unit(base_size * 1.1, "pt")
  ggtheme$legend.justification <- c("left", "center")
  ggtheme$legend.margin <- ggplot2::margin(0.0, 0.0, 0.0, 0.0, "pt")

  # Panel does not have a background, grid, or border.
  ggtheme$panel.background <- ggplot2::element_blank()
  ggtheme$panel.border <- ggplot2::element_blank()
  ggtheme$panel.grid <- ggplot2::element_blank()

  # Avoid removing some elements altogether by directly assigning NULL.
  ggtheme["panel.grid.major"] <- list(NULL)
  ggtheme["panel.grid.minor"] <- list(NULL)

  # Minimal plot margins
  ggtheme$plot.margin <- ggplot2::margin(1.0, 1.0, 1.0, 1.0, "pt")

  # The plot does not have a background
  ggtheme$plot.background <- ggplot2::element_blank()

  # Make title a bit smaller, and bold.
  ggtheme$plot.title <- ggplot2::element_text(
    face = "bold",
    size = ggplot2::rel(1.1),
    hjust = 0.0,
    vjust = 1.0,
    margin = ggplot2::margin(b = base_size / 2.0)
  )

  # Make subtitle a bit smaller.
  ggtheme$plot.subtitle <- ggplot2::element_text(
    size = ggplot2::rel(0.8),
    hjust = 0.0,
    vjust = 1.0,
    margin = ggplot2::margin(b = base_size / 2.0)
  )

  # Make caption a bit smaller.
  ggtheme$plot.caption <- ggplot2::element_text(
    size = ggplot2::rel(0.7),
    hjust = 1.0,
    vjust = 1.0,
    margin = ggplot2::margin(b = base_size / 2.0)
  )

  # Make tag bold.
  ggtheme$plot.tag <- ggplot2::element_text(
    face = "bold",
    hjust = 0.0,
    vjust = 0.7
  )

  # Make strip text black.
  ggtheme$strip.text <- ggplot2::element_text(
    size = ggplot2::rel(0.8),
    colour = "grey10",
    margin = ggplot2::margin(
      t = base_size / 4.0, 
      r = base_size / 4.0, 
      b = base_size / 4.0,
      l = base_size / 4.0,
      unit = "pt"
    )
  )

  # Remove strip background
  ggtheme$strip.background <- ggplot2::element_blank()

  return(ggtheme)
}



.check_ggtheme <- function(ggtheme) {
  # Check if the provided theme is a suitable theme.
  if (inherits(ggtheme, "theme")) {
    # Check if the theme is complete.
    if (!attr(ggtheme, "complete")) {
      ..error(paste0(
        "The plotting theme is not complete. The most likely cause is lack ",
        "of a valid template, such as theme_familiar or ggplot2::theme_light. ",
        "Note that ggplot2::theme is designed to tweak existing themes when creating a plot."
      ))
    }
    
  } else if (is.null(ggtheme)) {
    ggtheme <- theme_familiar(base_size = 9.0)
    
  } else {
    # Get the specified theme.
    ggtheme_fun <- switch(
      ggtheme,
      "default" = theme_familiar,
      "theme_familiar" = theme_familiar,
      "theme_gray" = ggplot2::theme_gray,
      "theme_grey" = ggplot2::theme_grey,
      "theme_bw" = ggplot2::theme_bw,
      "theme_linedraw" = ggplot2::theme_linedraw,
      "theme_light" = ggplot2::theme_light,
      "theme_dark" = ggplot2::theme_dark,
      "theme_minimal" = ggplot2::theme_minimal,
      "theme_classic" = ggplot2::theme_classic
    )

    if (is.null(ggtheme_fun)) {
      ..error(paste0(
        "The selected theme is not the default theme, or a standard ggplot2 theme. Found: ", ggtheme
      ))
    }

    # Generate theme
    ggtheme <- ggtheme_fun(base_size = 9.0)
  }

  return(ggtheme)
}



#' Checks and sanitizes splitting variables for plotting.
#'
#' @param x data.table or data.frame containing the data used for splitting.
#' @param split_by (*optional*) Splitting variables. This refers to column names
#'   on which datasets are split. A separate figure is created for each split.
#'   See details for available variables.
#' @param color_by (*optional*) Variables used to determine fill colour of plot
#'   objects. The variables cannot overlap with those provided to the `split_by`
#'   argument, but may overlap with other arguments. See details for available
#'   variables.
#' @param linetype_by (*optional*) Variables that are used to determine the
#'   linetype of lines in a plot. The variables cannot overlap with those
#'   provided to the `split_by` argument, but may overlap with other arguments.
#'   Sett details for available variables.
#' @param facet_by (*optional*) Variables used to determine how and if facets of
#'   each figure appear. In case the `facet_wrap_cols` argument is `NULL`, the
#'   first variable is used to define columns, and the remaing variables are
#'   used to define rows of facets. The variables cannot overlap with those
#'   provided to the `split_by` argument, but may overlap with other arguments.
#'   See details for available variables.
#' @param x_axis_by (*optional*) Variable plotted along the x-axis of a plot.
#'   The variable cannot overlap with variables provided to the `split_by` and
#'   `y_axis_by` arguments (if used), but may overlap with other arguments. Only
#'   one variable is allowed for this argument. See details for available
#'   variables.
#' @param y_axis_by (*optional*) Variable plotted along the y-axis of a plot.
#'   The variable cannot overlap with variables provided to the `split_by` and
#'   `x_axis_by` arguments (if used), but may overlap with other arguments. Only
#'   one variable is allowed for this argument. See details for available
#'   variables.
#' @param available Names of columns available for splitting.
#'
#' @details This internal function allows some flexibility regarding the exact
#'   input. Allowed splitting variables should be defined by the available
#'   argument.
#'
#' @return A sanitized list of splitting variables.
#' @md
#' @keywords internal
.check_plot_splitting_variables <- function(
    x,
    split_by = NULL,
    color_by = NULL,
    linetype_by = NULL,
    facet_by = NULL,
    x_axis_by = NULL,
    y_axis_by = NULL,
    available = NULL
) {
  # Find unique variables
  splitting_vars <- c(split_by, color_by, linetype_by, facet_by, x_axis_by, y_axis_by)

  if (is.null(available) && length(splitting_vars) == 0L) {
    return(list())
    
  } else if (is.null(available) && length(splitting_vars) > 0L) {
    ..error(paste0(
      "The current plot has no required splitting variables defined, but ",
      paste_s(splitting_vars),
      ifelse(length(splitting_vars) == 1L, " was assigned.", " were assigned.")
    ))
  }

  # Filter available down to those present in the data.
  filter_available <- intersect(available, colnames(x))

  # Filter available down to those that have more than one variable
  filter_available <- filter_available[sapply(
    filter_available,
    function(ii, x) (data.table::uniqueN(x = x, by = ii) > 1L),
    x = x
  )]
  
  if (is.null(filter_available)) {
    return(list())
    
  } else if (!all(filter_available %in% splitting_vars)) {
    missing_vars <- filter_available[!filter_available %in% splitting_vars]
    ..error(paste0(
      "The current plot requires ",
      paste_s(filter_available),
      ifelse(length(filter_available) > 1L, " as splitting variables", " as a splitting_variable"),
      ", but ",
      paste_s(missing_vars),
      ifelse(length(missing_vars) == 1L, " was not assigned.", " were not assigned.")
    ))
  }

  # Update available
  available <- filter_available

  # Generate output
  output_list <- list()

  # Update split_by
  if (!is.null(split_by) && any(split_by %in% available)) {
    output_list$split_by <- intersect(split_by, available)
  }

  # Update color_by
  if (!is.null(color_by) && any(color_by %in% available)) {
    output_list$color_by <- intersect(color_by, available)
  }

  # update linetype_by
  if (!is.null(linetype_by) && any(linetype_by %in% available)) {
    output_list$linetype_by <- intersect(linetype_by, available)
  }

  # update facet_by
  if (!is.null(facet_by) && any(facet_by %in% available)) {
    output_list$facet_by <- intersect(facet_by, available)
  }

  # update x_axis_by
  if (!is.null(x_axis_by) && any(x_axis_by %in% available)) {
    output_list$x_axis_by <- intersect(x_axis_by, available)
  }

  # update y_axis_by
  if (!is.null(y_axis_by) && any(y_axis_by %in% available)) {
    output_list$y_axis_by <- intersect(y_axis_by, available)
  }

  # Check split_by variable.
  .check_value_not_shared(output_list$split_by, output_list$color_by, "split_by", "color_by")
  .check_value_not_shared(output_list$split_by, output_list$linetype_by, "split_by", "linetype_by")
  .check_value_not_shared(output_list$split_by, output_list$facet_by, "split_by", "facet_by")
  .check_value_not_shared(output_list$split_by, output_list$x_axis_by, "split_by", "x_axis_by")
  .check_value_not_shared(output_list$split_by, output_list$y_axis_by, "split_by", "y_axis_by")

  # Check x_axis_by variable and y_axis_by variables.
  .check_value_not_shared(output_list$x_axis_by, output_list$y_axis_by, "x_axis_by", "y_axis_by")

  # Check length of x_axis_by and y_axis_by variables.
  .check_argument_length(output_list$x_axis_by, "x_axis_by", min = 0L, max = 1L)
  .check_argument_length(output_list$y_axis_by, "y_axis_by", min = 0L, max = 1L)

  return(output_list)
}



.parse_plot_facet_by <- function(x, facet_by, facet_wrap_cols) {
  if (is.null(facet_by)) {
    return(list())
    
  } else if (length(facet_by) == 1L) {
    if (is.null(facet_wrap_cols)) {
      return(list("facet_cols" = quos(!!ensym(facet_by))))
    } else {
      return(list("facet_by" = quos(!!ensym(facet_by))))
    }
    
  } else {
    if (is.null(facet_wrap_cols)) {
      facet_col <- facet_by[1L]
      facet_rows <- facet_by[2L:length(facet_by)]
      return(list(
        "facet_cols" = quos(!!ensym(facet_col)),
        "facet_rows" = quos(!!!parse_exprs(facet_rows))
      ))
      
    } else {
      return(list("facet_by" = quos(!!!parse_exprs(facet_by))))
    }
  }
}



.create_plot_subtitle <- function(
    x, 
    split_by = NULL, 
    additional = NULL
) {
  # Do not create a subtitle if there is no subtitle to be created.
  subtitle <- NULL
  
  # Generate subtitle from splitting variables and data.
  if (!is.null(split_by)) {
    subtitle <- c(
      subtitle,
      sapply(
        split_by,
        function(name, x) {
          split_variable_name <- name

          if (split_variable_name == "vimp_method") {
            split_variable_name <- "VIMP method"
          } else if (split_variable_name == "learner") {
            split_variable_name <- "learner"
          } else if (split_variable_name == "data_set") {
            split_variable_name <- "data set"
          } else if (split_variable_name == "evaluation_time") {
            split_variable_name <- "time point"
          } else if (split_variable_name == "sample_id") {
            split_variable_name <- "sample"
          } else if (split_variable_name %in% c("feature_name", "feature")) {
            split_variable_name <- "feature"
          }

          # Remove all underscores.
          split_variable_name <- gsub(
            x = split_variable_name, 
            pattern = "_",
            replacement = " ",
            fixed = TRUE
          )

          # Parse to an elementary string.
          split_variable_name <- paste0(split_variable_name, ": ", x[[name]][1L])

          return(split_variable_name)
        },
        x = x
      )
    )
  }
  
  # Generate additional strings from additional.
  if (!is.null(additional)) {
    subtitle <- c(
      subtitle,
      mapply(
        function(name, value) {
          # Remove all underscores.
          split_variable_name <- gsub(
            x = name, 
            pattern = "_",
            replacement = " ",
            fixed = TRUE
          )

          # Parse to an elementary string.
          split_variable_name <- paste0(split_variable_name, ": ", value)
        },
        name = names(additional),
        value = additional
      )
    )
  }

  # Check if any subtitle was generated.
  if (is.null(subtitle)) return(NULL)

  # Combine into single string.
  subtitle <- paste_s(subtitle)

  return(subtitle)
}



.add_time_to_plot_subtitle <- function(value) {
  return(list("time point" = value))
}



.create_plot_subtype <- function(
    x,
    subtype = NULL, 
    split_by = NULL, 
    additional = NULL
) {
  # Generate additional terms for the subtype, based on splits.
  if (!is.null(split_by)) {
    subtype <- c(
      subtype,
      as.character(sapply(
        split_by,
        function(jj, x) (x[[jj]][1L]),
        x = x
      ))
    )
  }

  if (!is.null(additional)) {
    subtype <- c(
      subtype,
      sapply(additional, function(jj) as.character(jj[1L]))
    )
  }

  if (is.null(subtype)) return(NULL)

  # Combine into a single string.
  subtype <- paste0(subtype, collapse = "_")

  return(subtype)
}



.create_plot_legend_title <- function(
    user_label, 
    color_by = NULL,
    linetype_by = NULL, 
    combine_legend = FALSE
) {
  # Sent for inspection
  .check_input_plot_args(
    legend_label = user_label,
    combine_legend = combine_legend
  )
  
  if (is.null(color_by) && is.null(linetype_by)) {
    # No splitting variables are used
    return(list(
      "guide_color" = NULL,
      "guide_linetype" = NULL
    ))
  }
  
  # Collect required list entries
  req_entries <- character(0L)
  if (!is.null(color_by)) {
    req_entries <- c(req_entries, "guide_color")
  }
  if (!is.null(linetype_by)) {
    req_entries <- c(req_entries, "guide_linetype")
  }
  
  # Waiver: no user input
  if (is.waive(user_label)) {
    if (combine_legend) {
      legend_label <- gsub(
        x = paste0(unique(c(color_by, linetype_by)), collapse = " & "),
        pattern = "_",
        replacement = " ",
        fixed = TRUE
      )
      
      return(list(
        "guide_color" = legend_label,
        "guide_linetype" = legend_label
      ))
      
    } else {
      # Colour labels
      if (!is.null(color_by)) {
        color_guide_label <- gsub(
          x = paste0(color_by, collapse = " & "),
          pattern = "_",
          replacement = " ",
          fixed = TRUE
        )
        
      } else {
        color_guide_label <- NULL
      }
      
      # Linetype labels
      if (!is.null(linetype_by)) {
        linetype_guide_label <- gsub(
          x = paste0(linetype_by, collapse = " & "),
          pattern = "_",
          replacement = " ",
          fixed = TRUE
        )
        
      } else {
        linetype_guide_label <- NULL
      }
      
      return(list(
        "guide_color" = color_guide_label,
        "guide_linetype" = linetype_guide_label
      ))
    }
  } else if (is.null(user_label)) {
    # NULL input
    
    return(list(
      "guide_color" = NULL,
      "guide_linetype" = NULL
    ))
    
  } else if (is.list(user_label)) {
    # List input
    
    # Check entries for existence
    for (current_entry in req_entries) {
      if (!current_entry %in% names(user_label)) {
        ..error(paste0(
          "A legend name is missing for ", current_entry, 
          ". Please set this name to a \"", current_entry, 
          "\" list element, e.g. list(\"",  current_entry, 
          "\"=\"some name\", ...)."
        ))
      }
    }
    
    # Select required entries
    user_label <- user_label[names(user_label) %in% req_entries]
    
    # Check that all entries are the same
    if (combine_legend && length(req_entries) >= 2L) {
      if (!all(sapply(
        user_label[2L:length(user_label)],
        identical,
        user_label[[1L]]
      ))) {
        ..error(paste0(
          "Not all provided legend names are identical, but identical legend ",
          "names are required for combining the legend."
        ))
      }
    }
    
    return(user_label)
    
  } else if (length(req_entries) >= 2L && !combine_legend) {
    # Single input where multiple is required
    
    ..error(paste0(
      "Multiple legend names are required, but only one is provided. ",
      "Please return a list with ",
      paste0("\"", req_entries, "\"", collapse = ", "), " elements."
    ))
    
  } else {
    # Single input
    
    return(list(
      "guide_color" = user_label, 
      "guide_linetype" = user_label
    ))
  }
}



.create_plot_guide_table <- function(
    x, 
    color_by = NULL, 
    linetype_by = NULL, 
    discrete_palette = NULL, 
    combine_legend = TRUE
) {
  
  .get_guide_tables <- function(x, color_by, linetype_by, discrete_palette) {
    # Suppress NOTES due to non-standard evaluation in data.table
    color_id <- linetype_id <- NULL
    
    # Select unique variables
    unique_vars <- unique(c(color_by, linetype_by))
    
    # Check whether there are any unique splitting variables
    if (is.null(unique_vars)) return(NULL)
    
    # Generate a guide table
    guide_table <- data.table::data.table(expand.grid(lapply(
      rev(unique_vars), 
      function(ii, x) (levels(x[[ii]])), 
      x = x
    )))
    
    # Rename variables
    data.table::setnames(x = guide_table, rev(unique_vars))
    
    # Convert to factors
    for (ii in unique_vars) {
      guide_table[[ii]] <- factor(
        x = guide_table[[ii]],
        levels = levels(x[[ii]])
      )
    }
    
    # Order columns according to unique_vars
    data.table::setcolorder(
      x = guide_table, 
      neworder = unique_vars
    )
    
    # Order data set by columns
    data.table::setorderv(
      x = guide_table, 
      cols = unique_vars
    )
    
    # Set breaks
    breaks <- apply(guide_table, 1L, paste, collapse = ", ")
    
    # Extend guide table
    if (!is.null(color_by)) {
      # Generate breaks
      guide_table$color_breaks <- factor(
        x = breaks,
        levels = breaks
      )
      
      # Define colour groups
      guide_table[, "color_id" := .GRP, by = color_by]
      
      # Get the palette to use.
      discr_palette <- .get_palette(
        x = discrete_palette,
        n = max(guide_table$color_id),
        palette_type = "qualitative"
      )
      
      # Assign colour values
      guide_table[, "color_values" := discr_palette[color_id]]
    }
    
    if (!is.null(linetype_by)) {
      # Generate breaks
      guide_table$linetype_breaks <- factor(
        x = breaks,
        levels = breaks
      )
      
      # Define linetype groups
      guide_table[, "linetype_id" := .GRP, by = linetype_by]
      
      # Get the palette to use
      line_palette <- scales::linetype_pal()(max(guide_table$linetype_id))
      
      # Assign linetypes
      guide_table[, "linetype_values" := line_palette[linetype_id]]
    }
    
    return(guide_table)
  }
  
  if (is_empty(x)) return(list("data" = x))
  
  # Extract guide tables
  if (combine_legend) {
    guide_list <- list(
      "guide_color" = .get_guide_tables(
        x = x, 
        color_by = color_by, 
        linetype_by = linetype_by, 
        discrete_palette = discrete_palette
      ),
      "guide_linetype" = .get_guide_tables(
        x = x, 
        color_by = color_by, 
        linetype_by = linetype_by, 
        discrete_palette = discrete_palette
      )
    )
    
  } else {
    guide_list <- list(
      "guide_color" = .get_guide_tables(
        x = x, 
        color_by = color_by, 
        linetype_by = NULL, 
        discrete_palette = discrete_palette
      ),
      "guide_linetype" = .get_guide_tables(
        x = x, 
        color_by = NULL, 
        linetype_by = linetype_by, 
        discrete_palette = discrete_palette
      )
    )
  }
  
  # Filter out lists corresponding to missing split variables
  guide_list <- guide_list[!sapply(list(color_by, linetype_by), is.null)]
  
  if (length(guide_list) == 0L) return(list("data" = x))
  
  # Initialise return list
  return_list <- list()
  
  # Add break column of the remaining lists to x
  for (guide_type in names(guide_list)) {
    if (guide_type == "guide_color") {
      # Add color_breaks to x
      if (combine_legend) {
        x <- merge(
          x = x,
          y = guide_list[[guide_type]][, mget(c(unique(c(color_by, linetype_by)), "color_breaks"))],
          by = unique(c(color_by, linetype_by)),
          all.x = TRUE,
          all.y = FALSE
        )
        
      } else {
        x <- merge(
          x = x,
          y = guide_list[[guide_type]][, mget(c(color_by, "color_breaks"))],
          by = color_by,
          all.x = TRUE, 
          all.y = FALSE
        )
      }
      
      # Return guide_color
      return_list[[guide_type]] <- guide_list[[guide_type]]
      
    } else if (guide_type == "guide_linetype") {
      # Add linetype_breaks to x
      if (combine_legend) {
        x <- merge(
          x = x,
          y = guide_list[[guide_type]][, mget(c(unique(c(color_by, linetype_by)), "linetype_breaks"))],
          by = unique(c(color_by, linetype_by)),
          all.x = TRUE, 
          all.y = FALSE
        )
        
      } else {
        x <- merge(
          x = x,
          y = guide_list[[guide_type]][, mget(c(linetype_by, "linetype_breaks"))],
          by = linetype_by,
          all.x = TRUE, 
          all.y = FALSE
        )
      }
      
      # Return guide_linetype
      return_list[[guide_type]] <- guide_list[[guide_type]]
    }
  }
  
  # Add updated data
  return_list$data <- x
  
  return(return_list)
}



.add_plot_cluster_name <- function(
    x, 
    color_by = NULL, 
    facet_by = NULL, 
    singular_cluster_character = "\u2014"
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  cluster_size <- cluster_id <- feature <- new_cluster_id <- cluster_name <- NULL

  ..integer_to_char <- function(x) {
    # Initialise placeholders
    x_remain <- x
    new_string <- character(0L)

    while (as.integer(ceiling(x_remain / 26L)) > 0L) {
      # Determine the modulo.
      mod <- x_remain %% 26L

      # Find if mod is equal to 0, which would indicate Z.
      mod <- ifelse(mod == 0L, 26L, mod)

      # Add letter
      new_string <- c(new_string, LETTERS[mod])

      # Update the remain variable
      x_remain <- (x_remain - mod) / 26L
    }

    return(paste(rev(new_string), collapse = ""))
  }

  # Identify splitting variables
  splitting_vars <- unique(c(color_by, facet_by))

  # Split x by splitting variables.
  if (length(splitting_vars) > 0L) {
    x <- split(x, by = splitting_vars)
  } else {
    x <- list(x)
  }

  # Iterate and add cluster names.
  x <- lapply(
    x,
    function(y) {
      # Check if x is empty.
      if (is_empty(y)) return(NULL)
      
      # This is for backward compatibility.
      if (!all(c("cluster_id", "cluster_size") %in% colnames(y))) {
        # Add cluster name
        y[, "cluster_name" := singular_cluster_character]
        
        return(y)
      }
      
      # Only determine cluster_name for those clusters that have cluster_size >
      # 1. Also, the most important features should receive a higher replacement
      # cluster_id.
      y_short <- y[cluster_size > 1L, mget(c("feature", "cluster_id"))]
      
      if (!is_empty(y_short)) {
        # Remove unused levels for the name column. The levels of name are
        # ordered according to importance.
        y_short <- droplevels(y_short)
        
        # Set placeholder cluster id
        y_short[, "new_cluster_id" := NA_integer_]
        
        new_id <- 1L
        for (current_feature in levels(y_short$feature)) {
          # Provide new cluster id in case none exists.
          if (is.na(y_short[feature == current_feature, ]$new_cluster_id[1L])) {
            # Find the old cluster id.
            old_cluster_id <- y_short[feature == current_feature, ]$cluster_id[1L]
            
            # Update all entries with the same old cluster id.
            y_short[cluster_id == old_cluster_id, "new_cluster_id" := new_id]
            
            # Increment new cluster id.
            new_id <- new_id + 1L
          }
        }
        
        # Determine cluster name based on id.
        y_short[, "cluster_name" := ..integer_to_char(new_cluster_id), by = "feature"]
        
        # Drop redundant columns
        y_short[, ":="(
          "cluster_id" = NULL, 
          "new_cluster_id" = NULL
        )]
        
        # Merge with y.
        y <- merge(
          x = y, 
          y = y_short,
          by = "feature",
          all = TRUE
        )
        
        # Mark singular clusters
        y[is.na(cluster_name), "cluster_name" := singular_cluster_character]
        
      } else {
        # Mark singular clusters
        y[, "cluster_name" := singular_cluster_character]
      }
      
      return(y)
    }
  )
  
  x <- data.table::rbindlist(x, use.names = TRUE)
  
  return(x)
}


.split_data_by_plot_facet <- function(x, plot_layout_table = NULL, ...) {
  if (is_empty(x)) return(NULL)
  
  if (is.null(plot_layout_table)) {
    plot_layout_table <- do.call(
      .get_plot_layout_table,
      args = c(list("x" = x), list(...))
    )
  }

  # Derive facet_by
  facet_by <- setdiff(
    colnames(plot_layout_table),
    c("col_id", "row_id")
  )

  if (length(facet_by) > 0L) {
    # Merge the plot_layout_table into x. This will keep things in order. All
    # levels are kept.
    x <- merge(
      x = x,
      y = plot_layout_table,
      by = facet_by,
      all = TRUE
    )
    
  } else {
    x <- cbind(x, plot_layout_table)
  }

  # Split data by row, then column
  split_data <- split(
    x, 
    by = c("col_id", "row_id"), 
    sorted = TRUE
  )

  return(split_data)
}



.get_plot_layout_dims <- function(plot_layout_table = NULL, ...) {
  # Create the plot_layout_table if it is not provided.
  if (is.null(plot_layout_table)) {
    plot_layout_table <- do.call(
      .get_plot_layout_table, 
      args = list(...)
    )
  }

  # Return (nrows, ncols)
  return(c(
    max(plot_layout_table$row_id),
    max(plot_layout_table$col_id)
  ))
}



.get_plot_layout_table <- function(x, facet_by, facet_wrap_cols) {
  if (is.null(facet_by)) {
    # Simple 1x1 layout without facets.
    plot_layout_table <- data.table::data.table(
      "col_id" = 1L,
      "row_id" = 1L
    )
    
  } else if (is.null(facet_wrap_cols)) {
    # Generate a plot_layout_table and order it
    plot_layout_table <- expand.grid(
      lapply(
        facet_by,
        function(column, x) levels(x[[column]]),
        x = x
      ),
      KEEP.OUT.ATTRS = FALSE
    )
    
    plot_layout_table <- data.table::as.data.table(plot_layout_table)
    data.table::setnames(plot_layout_table, facet_by)
    data.table::setorderv(x = plot_layout_table, cols = facet_by)

    # Find the number of columns
    n_cols <- length(unique(x[[facet_by[1L]]]))

    # Add column id to the plot_layout_table
    plot_layout_table[, "col_id" := .GRP, by = get(facet_by[1L])]

    if (length(facet_by) > 1L) {
      # Find the number of rows
      n_levels <- sapply(
        facet_by[2L:length(facet_by)],
        function(ii, x) {
          if (is.factor(x[[ii]])) {
            return(nlevels(x[[ii]]))
          } else {
            return(length(unique(x[[ii]])))
          }
        },
        x = x
      )
      n_rows <- prod(n_levels)

      # Add row id to the plot_layout_table
      facet_row_cols <- facet_by[2L:length(facet_by)]
      plot_layout_table[, "row_id" := .GRP, by = mget(facet_row_cols)]
      
    } else {
      # There is only one row
      n_rows <- 1L
      plot_layout_table[, "row_id" := 1L]
    }
    
  } else {
    # Generate a plot_layout_table, and order
    plot_layout_table <- unique(x[, (facet_by), with = FALSE], by = facet_by)
    data.table::setorderv(x = plot_layout_table, cols = facet_by)

    # Number of columns is provided using facet_wrap_cols.
    len_table <- nrow(plot_layout_table)
    n_cols <- facet_wrap_cols
    n_rows <- ceiling(len_table / n_cols)

    # Generate the column and row positions.
    col_ids <- rep(seq_len(n_cols), times = n_rows)[seq_len(len_table)]
    row_ids <- rep(seq_len(n_rows), each = n_cols)[seq_len(len_table)]

    # Add column and row ids to the plot_layout_table.
    plot_layout_table[, ":="(
      "col_id" = col_ids,
      "row_id" = row_ids
    )]
  }

  return(plot_layout_table)
}



.combine_plot_elements <- function(
  g_main,
  g_new,
  element_name,
  spacer = NULL,
  stack_direction = "vertical"
) {
  
  for (current_element_name in element_name) {
    # Determine matching elements in both g_main and g_new.
    element_names_main <- g_main$layout$name
    present_elements_main <- element_names_main[sapply(
      element_names_main, 
      startswith_any, 
      prefix = current_element_name
    )]
    
    element_names_new <- g_new$layout$name
    present_elements_new <- element_names_new[sapply(
      element_names_new,
      startswith_any,
      prefix = current_element_name
    )]
    
    if (length(present_elements_main) != 1L || length(present_elements_new) != 1L) next
    
    grob_index_main <- which(g_main$layout$name == present_elements_main)
    grob_index_new <- which(g_new$layout$name == present_elements_new)
    
    # Determine if there is anything to add from g_new.
    if (any(sapply(c("zeroGrob", "nullGrob"), function(ii, x) (is(x, ii)), x = g_new$grobs[[grob_index_new]]))) next
    
    grob_main <- g_main$grobs[[grob_index_main]]
    grob_new <- g_new$grobs[[grob_index_new]]
    
    # If there is an empty grob in g_main, simply copy the element from g_new
    # into g_main.
    if (any(sapply(c("zeroGrob", "nullGrob"), function(ii, x) (is(x, ii)), x = g_main$grobs[[grob_index_main]]))) {
      g_main$grobs[[grob_index_main]] <- grob_new
      
    } else {
      g <- list(grob_main, grob_new)
      
      # Find widths and heights
      widths <- lapply(g, .gtable_get_grob_aspect_size, aspect = "width")
      if (!is.null(spacer) && stack_direction == "horizontal") {
        widths <- append(widths, 0L, after = 1L)
        # Direct insertion of spacer with append results in spacer losing its
        # simpleUnit class.
        widths[[2L]] <- spacer
      }
      widths <- do.call(grid::unit.c, widths)
      
      heights <- lapply(g, .gtable_get_grob_aspect_size, aspect = "height")
      if (!is.null(spacer) && stack_direction == "vertical") {
       heights <- append(heights, 0L, after = 1L)
       # Direct insertion of spacer with append results in spacer losing its
       # simpleUnit class.
       heights[[2L]] <- spacer
      }
      heights <- do.call(grid::unit.c, heights)
      
      if (stack_direction == "vertical") {
        # Concatenate the widths.
        widths <- max(widths)

      } else {
        # Concatenate the heights.
        heights <- max(heights)
      }
      
      # Set up basic gtable.
      g_combine <- gtable::gtable(
        widths = widths,
        heights = heights,
        name = present_elements_main
      )
      
      # Insert grob from main.
      g_combine <- gtable::gtable_add_grob(
        g_combine,
        grobs = g[[1L]],
        t = 1L,
        l = 1L
      )
      
      # Insert grob from new.
      g_combine <- gtable::gtable_add_grob(
        g_combine,
        grobs = g[[2L]],
        t = length(heights),
        l = length(widths)
      )
      
      # Insert spacer.
      if (!is.null(spacer) && stack_direction == "vertical") {
        spacer_position <- c(2L, 2L, 1L, 1L)
        names(spacer_position) <- c("t", "b", "l", "r")
        
        g_combine <- .gtable_insert_spacer(
          g_combine,
          position = spacer_position,
          height = spacer
        )
        
      } else if (!is.null(spacer) && stack_direction == "horizontal") {
        spacer_position <- c(1L, 1L, 2L, 2L)
        names(spacer_position) <- c("t", "b", "l", "r")
        
        g_combine <- .gtable_insert_spacer(
          g_combine,
          position = spacer_position,
          width = spacer
        )
      }
      
      g_main$grobs[[grob_index_main]] <- g_combine
    }
  }
  
  g_main <- .gtable_update_layout(g = g_main)
  
  return(g_main)
}



.compose_figure <- function(
    figure_list,
    plot_layout_table,
    x_text_shared = "overall",
    x_label_shared = "overall",
    y_text_shared = "overall",
    y_label_shared = "overall",
    facet_wrap_cols = NULL,
    ggtheme = NULL
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  col_id <- row_id <- is_present <- n_present <- NULL
  type <- NULL
  
  # Global layout --------------------------------------------------------------
  
  # Add figure names to plot_layout_table.
  plot_layout_table[, figure_name := paste0(row_id, ".", col_id)]
  plot_layout_table[, is_present := figure_name %in% names(figure_list)]
  
  # Drop columns with only missing information.
  empty_cols <- plot_layout_table[
    ,
    list("n_present" = sum(is_present)),
    by = "col_id"
  ]
  empty_cols <- empty_cols[n_present == 0L]$col_id
  if (length(empty_cols) > 0L) {
    plot_layout_table <- plot_layout_table[!col_id %in% empty_cols]
  }
  
  # Drop rows with only missing information.
  empty_rows <- plot_layout_table[
    ,
    list("n_present" = sum(is_present)),
    by = "row_id"
  ]
  empty_rows <- empty_rows[n_present == 0L]$row_id
  if (length(empty_rows) > 0L) {
    plot_layout_table <- plot_layout_table[!row_id %in% empty_rows]
  }

  # Check that any part of the plot is remaining
  if (is_empty(plot_layout_table)) return(NULL)
  
  # Create missing figure panels -----------------------------------------------
  
  # Identify and add missing figures.
  missing_figures <- plot_layout_table[is_present == FALSE]$figure_name
  for (missing_figure in missing_figures) {
    # Get col_id and row_id to identify template figures.
    current_row_id <- plot_layout_table[figure_name == missing_figure]$row_id
    current_col_id <- plot_layout_table[figure_name == missing_figure]$col_id
    
    # Identify figures to use as templates.
    template_figure_row_name <- head(plot_layout_table[row_id == current_row_id & is_present == TRUE]$figure_name, n = 1L)
    template_figure_col_name <- head(plot_layout_table[col_id == current_col_id & is_present == TRUE]$figure_name, n = 1L)
    
    # Add template.
    figure_list[[missing_figure]] <- .create_placeholder_figure(
      template_figure_row = figure_list[[template_figure_row_name]],
      template_figure_col = figure_list[[template_figure_col_name]],
      row_id = current_row_id,
      col_id = current_col_id
    )
  }
  
  # Check that no figures are missing now.
  plot_layout_table[, is_present := figure_name %in% names(figure_list)]
  if (!all(plot_layout_table$is_present)) {
    ..error_reached_unreachable_code("All panels of the figure should be present, but one or more are missing.")
  }
  
  # Shared elements between figure panels --------------------------------------
  
  # Determine top and bottom rows, and left and right columns.
  top_row_id <- min(plot_layout_table$row_id)
  bottom_row_id <- max(plot_layout_table$row_id)
  left_col_id <- min(plot_layout_table$col_id)
  right_col_id <- max(plot_layout_table$col_id)
  
  # Remove elements from figures.
  for (figure_name in names(figure_list)) {
    # Configure removal.
    figure_list[[figure_name]] <- .set_figure_element_removal(
      object = figure_list[[figure_name]],
      top_row_id = top_row_id,
      bottow_row_id = bottom_row_id,
      left_col_id = left_col_id,
      right_col_id = right_col_id,
      x_text_shared = x_text_shared,
      y_text_shared = y_text_shared,
      x_label_shared = x_label_shared,
      y_label_shared = y_label_shared
    )
    
    # Remove elements by replacing them with a zeroGrob. This maintains the
    # size of the figure's gtable.
    figure_list[[figure_name]] <- .remove_figure_elements(
      object = figure_list[[figure_name]],
      replace_by_zero_grob = TRUE
    )
  }
  
  # Compose figure panels ------------------------------------------------------
  
  # Form plot rows.
  g <- NULL
  spacer_width_x <- .get_plot_panel_spacing(ggtheme = ggtheme, axis = "x")
  spacer_width_y <- .get_plot_panel_spacing(ggtheme = ggtheme, axis = "y")
  unique_rows <- sort(unique(plot_layout_table$row_id))
  unique_cols <- sort(unique(plot_layout_table$col_id))
  for (current_row_id in unique_rows) {
    # Merge columns within each row.
    g_row <- NULL
    for (current_col_id in unique_cols) {
      if (is.null(g_row)) {
        # Use 
        g_row <- figure_list[[paste0(current_row_id, ".", current_col_id)]]@gtable
        
      } else {
        # Insert column for spacer.
        g_row <- gtable::gtable_add_cols(
          g_row,
          widths = spacer_width_x,
          pos = ncol(g_row)
        )
        
        # Add spacer.
        g_row <- .gtable_insert_spacer(
          g = g_row,
          position = c("t" = 1L, "b" = nrow(g_row), "l" = ncol(g_row), "r" = ncol(g_row)),
          width = spacer_width_x
        )
        
        # Combine gtable by columns.
        g_row <- cbind(g_row, figure_list[[paste0(current_row_id, ".", current_col_id)]]@gtable)
      }
    }
    
    # Merge with existing rows.
    if (is.null(g)) {
      g <- g_row
      
    } else {
      # Insert row for spacer.
      g_row <- gtable::gtable_add_rows(
        g_row,
        heights = spacer_width_y,
        pos = nrow(g_row)
      )
      
      # Add spacer element in the new row.
      g_row <- .gtable_insert_spacer(
        g = g_row,
        position = c("t" = nrow(g_row), "b" = nrow(g_row), "l" = 1L, "r" = ncol(g_row)),
        height = spacer_width_y
      )
      
      # Combine gtable by rows.
      g <- rbind(g, g_row)
    }
  }
  
  # Insert global elements -----------------------------------------------------
  
  # Insert global elements.
  global_element_names <- c(
    .all_gtable_title_names(),
    .all_gtable_guide_names()
  )
  if (x_text_shared %in% c("overall", "TRUE")) {
    global_element_names <- c(global_element_names, .all_gtable_axis_x_names())
  }
  if (x_label_shared %in% c("overall", "TRUE")) {
    global_element_names <- c(global_element_names, .all_gtable_label_x_names())
  }
  if (y_text_shared %in% c("overall", "TRUE")) {
    global_element_names <- c(global_element_names, .all_gtable_axis_y_names())
  }
  if (y_label_shared %in% c("overall", "TRUE")) {
    global_element_names <- c(global_element_names, .all_gtable_label_x_names())
  }
  
  # Isolate global elements that need to be updated.
  global_elements <- figure_list[[1L]]@global_elements
  global_element_name_list <- lapply(
    names(global_elements),
    function(x, y) {
      if (!startswith_any(x, y)) return(NULL)
      return(list(
        "name" = x,
        "type" =  y[sapply(y, function(y, x) {startsWith(x, y)}, x = x)]
      ))
    },
    y = global_element_names
  )
  global_element_name_list <- data.table::rbindlist(global_element_name_list)
  
  # For title-like objects, insert in place, at the top.
  if (any(global_element_name_list$type %in% .all_gtable_title_names("title"))) {
    x <- global_element_name_list[type %in% .all_gtable_title_names("title")]
    for (ii in seq_len(nrow(x))) {
      element_positions <- g$layout[g$layout$name == x$name[ii], c("t", "l", "b", "r"), drop = FALSE]
      position <- c("t" = 0L, "l" = 0L, "b" = 0L, "r" = 0L)
      position["t"] <- position["b"] <- min(element_positions$t)
      position["l"] <- min(element_positions$l)
      position["r"] <- max(element_positions$r)
      g <- .gtable_insert(
        g = g,
        g_new = global_elements[[x$name[ii]]],
        where = c("at", position),
        grob_name = x$name[ii]
      )
    }
  }
  
  # For caption-like objects, insert in place, at the bottom.
  if (any(global_element_name_list$type %in% .all_gtable_title_names("caption"))) {
    x <- global_element_name_list[type %in% .all_gtable_title_names("caption")]
    for (ii in seq_len(nrow(x))) {
      element_positions <- g$layout[g$layout$name == x$name[ii], c("t", "l", "b", "r"), drop = FALSE]
      position <- c("t" = 0L, "l" = 0L, "b" = 0L, "r" = 0L)
      position["t"] <- position["b"] <- max(element_positions$b)
      position["l"] <- min(element_positions$l)
      position["r"] <- max(element_positions$r)
      g <- .gtable_insert(
        g = g,
        g_new = global_elements[[x$name[ii]]],
        where = c("at", position),
        grob_name = x$name[ii]
      )
    }
  }

  # For guide-like objects, insert at the corresponding position.
  if (any(global_element_name_list$type %in% .all_gtable_guide_names())) {
    x <- global_element_name_list[type %in% .all_gtable_guide_names()]
    for (ii in seq_len(nrow(x))) {
      element_positions <- g$layout[g$layout$name == x$name[ii], c("t", "l", "b", "r"), drop = FALSE]
      position <- c("t" = 0L, "l" = 0L, "b" = 0L, "r" = 0L)
      
      # Positioning 
      if (x$type[ii] %in% .all_gtable_guide_names("top")) {
        position["t"] <- position["b"] <- min(element_positions$t)
        position["l"] <- min(element_positions$l)
        position["r"] <- max(element_positions$r)
        
        spacer_position <- position
        spacer_position[["t"]] <- spacer_position[["b"]] <- position[["b"]] + 1L
        g <- .gtable_insert_spacer(
          g = g,
          position = spacer_position,
          height = .get_plot_panel_spacing(ggtheme = ggtheme, axis = "y"),
          make_space = TRUE
        )

      } else if (x$type[ii] %in% .all_gtable_guide_names("bottom")) {
        position["t"] <- position["b"] <- max(element_positions$b)
        position["l"] <- min(element_positions$l)
        position["r"] <- max(element_positions$r)
        
        spacer_position <- position
        spacer_position[["t"]] <- spacer_position[["b"]] <- position[["t"]]
        g <- .gtable_insert_spacer(
          g = g,
          position = spacer_position,
          height = .get_plot_panel_spacing(ggtheme = ggtheme, axis = "y"),
          make_space = TRUE
        )
        
        # Inserting the spacer moves the guide object down.
        position[["t"]] <- position[["t"]] + 1L
        position[["b"]] <- position[["b"]] + 1L
        
      } else if (x$type[ii] %in% .all_gtable_guide_names("left")) {
        position["l"] <- position["r"] <- min(element_positions$l)
        position["t"] <- min(element_positions$t)
        position["b"] <- max(element_positions$b)
        
        spacer_position <- position
        spacer_position[["l"]] <- spacer_position[["r"]] <- position[["r"]] + 1L
        g <- .gtable_insert_spacer(
          g = g,
          position = spacer_position,
          width = .get_plot_panel_spacing(ggtheme = ggtheme, axis = "x"),
          make_space = TRUE
        )
        
      } else if (x$type[ii] %in% .all_gtable_guide_names("right")) {
        position["l"] <- position["r"] <- max(element_positions$r)
        position["t"] <- min(element_positions$t)
        position["b"] <- max(element_positions$b)
        
        spacer_position <- position
        spacer_position[["l"]] <- spacer_position[["r"]] <- position[["l"]]
        g <- .gtable_insert_spacer(
          g = g,
          position = spacer_position,
          width = .get_plot_panel_spacing(ggtheme = ggtheme, axis = "x"),
          make_space = TRUE
        )
        
        # Inserting the spacer moves the guide object right.
        position[["l"]] <- position[["l"]] + 1L
        position[["r"]] <- position[["r"]] + 1L
        
      } else {
        ..error_reached_unreachable_code(paste0("unknown type: ", x$type[1L]))
      }
      
      g <- .gtable_insert(
        g = g,
        g_new = global_elements[[x$name[ii]]],
        where = c("at", position),
        grob_name = x$name[ii]
      )
    }
  }
  
  # Labels
  if (any(global_element_name_list$type %in% .all_gtable_label_names())) {
    x <- global_element_name_list[type %in% .all_gtable_label_names()]
    for (ii in seq_len(nrow(x))) {
      element_positions <- g$layout[g$layout$name == x$name[ii], c("t", "l", "b", "r"), drop = FALSE]
      position <- c("t" = 0L, "l" = 0L, "b" = 0L, "r" = 0L)
      
      # Positioning 
      if (x$type[ii] %in% .all_gtable_label_names("top")) {
        position["t"] <- position["b"] <- min(element_positions$t)
        position["l"] <- min(element_positions$l)
        position["r"] <- max(element_positions$r)
        
      } else if (x$type[ii] %in% .all_gtable_label_names("bottom")) {
        position["t"] <- position["b"] <- max(element_positions$b)
        position["l"] <- min(element_positions$l)
        position["r"] <- max(element_positions$r)
        
      } else if (x$type[ii] %in% .all_gtable_label_names("left")) {
        position["l"] <- position["r"] <- min(element_positions$l)
        position["t"] <- min(element_positions$t)
        position["b"] <- max(element_positions$b)
        
      } else if (x$type[ii] %in% .all_gtable_label_names("right")) {
        position["l"] <- position["r"] <- max(element_positions$r)
        position["t"] <- min(element_positions$t)
        position["b"] <- max(element_positions$b)
        
      } else {
        ..error_reached_unreachable_code(paste0("unknown type: ", x$type[1L]))
      }
      
      g <- .gtable_insert(
        g = g,
        g_new = global_elements[[x$name[ii]]],
        where = c("at", position),
        grob_name = x$name[ii]
      )
    }
  }
  
  # Axis elements
  if (any(global_element_name_list$type %in% .all_gtable_axis_names())) {
    x <- global_element_name_list[type %in% .all_gtable_axis_names()]
    for (ii in seq_len(nrow(x))) {
      element_positions <- g$layout[g$layout$name == x$name[ii], c("t", "l", "b", "r"), drop = FALSE]
      position <- c("t" = 0L, "l" = 0L, "b" = 0L, "r" = 0L)
      
      # Positioning 
      if (x$type[ii] %in% .all_gtable_axis_names("top")) {
        position["t"] <- position["b"] <- min(element_positions$t)
        position["l"] <- min(element_positions$l)
        position["r"] <- max(element_positions$r)
        
      } else if (x$type[ii] %in% .all_gtable_axis_names("bottom")) {
        position["t"] <- position["b"] <- max(element_positions$b)
        position["l"] <- min(element_positions$l)
        position["r"] <- max(element_positions$r)
        
      } else if (x$type[ii] %in% .all_gtable_axis_names("left")) {
        position["l"] <- position["r"] <- min(element_positions$l)
        position["t"] <- min(element_positions$t)
        position["b"] <- max(element_positions$b)
        
      } else if (x$type[ii] %in% .all_gtable_axis_names("right")) {
        position["l"] <- position["r"] <- max(element_positions$r)
        position["t"] <- min(element_positions$t)
        position["b"] <- max(element_positions$b)
        
      } else {
        ..error_reached_unreachable_code(paste0("unknown type: ", x$type[1L]))
      }
      
      g <- .gtable_insert(
        g = g,
        g_new = global_elements[[x$name[ii]]],
        where = c("at", position),
        grob_name = x$name[ii]
      )
    }
  }
  
  # Clean-up -------------------------------------------------------------------
  
  # Drop all zeroGrob elements to prevent occlusion by empty objects.
  matched_elements <- !sapply(g$grobs, is, "zeroGrob")
  g$layout <- g$layout[matched_elements, , drop = FALSE]
  g$grobs <- g$grobs[matched_elements]
  g <- gtable::gtable_trim(g)
  
  # Update heights and widths.
  g <- .gtable_update_layout(g = g)

  return(g)
}



..get_plot_theme_linewidth <- function(ggtheme = NULL) {
  # Import default ggtheme in case none is provided.
  ggtheme <- .check_ggtheme(ggtheme)

  # Since ggplot 3.4.0, the width of a line is determined by linewidth instead
  # of size.
  linewidth <- ggtheme$line$linewidth

  return(linewidth)
}



..get_plot_element_spacing <- function(ggtheme = NULL, axis, theme_element) {
  # Obtain spacing from a ggtheme element

  # Import default ggtheme in case none is provided.
  ggtheme <- .check_ggtheme(ggtheme)
  
  # Basic spacing settings
  spacing <- ggtheme$spacing
  spacing_rel <- 1.0
  
  # Attempt to base the text size on the general axis.text attribute.
  if (!is.null(ggtheme[[theme_element]])) {
    if (inherits(ggtheme[[theme_element]], "rel")) {
      # Find the relative text size of axis text.
      spacing_rel <- as.numeric(ggtheme[[theme_element]])
      
    } else {
      # Set absolute text size.
      spacing <- ggtheme[[theme_element]]
      spacing_rel <- 1.0
    }
  }
  
  # Attempt to refine the text size using the axis.text.y attribute in
  # particular.
  if (!is.null(ggtheme[[paste0(theme_element, ".", axis)]])) {
    if (inherits(ggtheme[[paste0(theme_element, ".", axis)]], "rel")) {
      # Set relative text size of axis text
      spacing_rel <- as.numeric(ggtheme[[paste0(theme_element, ".", axis)]])
      
    } else {
      # Set absolute text size.
      spacing <- ggtheme[[paste0(theme_element, ".", axis)]]
      spacing_rel <- 1.0
    }
  }
  
  spacing <- spacing * spacing_rel

  # If no spacing is provided, produce 0.0 length spacing.
  if (!grid::is.unit(spacing)) spacing <- grid::unit(0.0, "pt")

  return(spacing)
}



.get_plot_panel_spacing <- function(ggtheme = NULL, axis) {
  # Obtain spacing between panels. This determines distance between facets.
  return(..get_plot_element_spacing(
    ggtheme = ggtheme,
    axis = axis,
    theme_element = "panel.spacing"
  ))
}



.get_plot_legend_spacing <- function(ggtheme = NULL, axis) {
  # Obtain spacing between legend and the main panel.
  return(..get_plot_element_spacing(
    ggtheme = ggtheme,
    axis = axis,
    theme_element = "legend.spacing"
  ))
}



.get_plot_geom_text_settings <- function(ggtheme = NULL) {
  # Import formatting settings from the provided ggtheme.

  # Import default ggtheme in case none is provided.
  ggtheme <- .check_ggtheme(ggtheme)

  # Find the text size for the table. This is based on text sizes in the
  # ggtheme.
  fontsize <- ggtheme$text$size
  fontsize_rel <- 1.0

  # Attempt to base the text size on the general axis.text attribute.
  if (!is.null(ggtheme$axis.text$size)) {
    if (inherits(ggtheme$axis.text$size, "rel")) {
      # Find the relative text size of axis text.
      fontsize_rel <- as.numeric(ggtheme$axis.text$size)
    } else {
      # Set absolute text size.
      fontsize <- ggtheme$axis.text$size
      fontsize_rel <- 1.0
    }
  }

  # Attempt to refine the text size using the axis.text.y attribute in
  # particular.
  if (!is.null(ggtheme$axis.text.y$size)) {
    if (inherits(ggtheme$axis.text.y$size, "rel")) {
      # Set relative text size of axis text
      fontsize_rel <- as.numeric(ggtheme$axis.text.y$size)
    } else {
      # Set absolute text size.
      fontsize <- as.numeric(ggtheme$axis.text.y$size)
      fontsize_rel <- 1.0
    }
  }

  # Update the text size using the magical ggplot2 point size (ggplot2:::.pt).
  geom_text_size <- fontsize * fontsize_rel / ggplot2::.pt

  # Obtain lineheight
  lineheight <- ggtheme$text$lineheight
  if (!is.null(ggtheme$axis.text$lineheight)) lineheight <- ggtheme$axis.text$lineheight
  if (!is.null(ggtheme$axis.text.y$lineheight)) lineheight <- ggtheme$axis.text.y$lineheight

  # Obtain family
  fontfamily <- ggtheme$text$family
  if (!is.null(ggtheme$axis.text$family)) fontfamily <- ggtheme$axis.text$family
  if (!is.null(ggtheme$axis.text.y$family)) fontfamily <- ggtheme$axis.text.y$family
  if (!is.null(ggtheme$axis.text.x$family)) fontfamily <- ggtheme$axis.text.x$family

  # Obtain face
  fontface <- ggtheme$text$face
  if (!is.null(ggtheme$axis.text$face)) fontface <- ggtheme$axis.text$face
  if (!is.null(ggtheme$axis.text.y$face)) fontface <- ggtheme$axis.text.y$face
  if (!is.null(ggtheme$axis.text.x$face)) fontface <- ggtheme$axis.text.x$face

  # Obtain colour
  colour <- ggtheme$text$colour
  if (!is.null(ggtheme$axis.text$colour)) colour <- ggtheme$axis.text$colour
  if (!is.null(ggtheme$axis.text.y$colour)) colour <- ggtheme$axis.text.y$colour
  if (!is.null(ggtheme$axis.text.x$colour)) colour <- ggtheme$axis.text.x$colour

  return(list(
    "geom_text_size" = geom_text_size,
    "fontsize" = fontsize,
    "fontsize_rel" = fontsize_rel,
    "colour" = colour,
    "family" = fontfamily,
    "face" = fontface,
    "lineheight" = lineheight
  ))
}











.rename_plot_grobs <- function(g, extension = "main", use_generic = TRUE) {
  if (is.null(g)) return(g)

  element_names <- c(
    .all_gtable_panel_names(),
    .all_gtable_label_x_names(),
    .all_gtable_label_y_names(),
    .all_gtable_axis_x_names(),
    .all_gtable_axis_y_names()
  )
  
  # Filter only elements that are present, and generate list of matching
  # local and global elements.
  renamable_elements <- g$layout$name
  element_list <- lapply(
    renamable_elements,
    function(x, y) {
      if (!startswith_any(x, prefix = y)) return(NULL)
      
      y <- y[sapply(y, function(y, x) {startsWith(x, y)}, x = x)]
      return(list("local" = x, "global" = y))
    },
    y = element_names
  )
  element_list <- data.table::rbindlist(element_list)
  
  # If generic names are not used, replace global (generic) names by the local
  # names.
  if (!use_generic) {
    element_list$global <- element_list$local
  }

  # Rename elements and add extension.
  for (ii in seq_len(nrow(element_list))) {
    g <- .gtable_rename_element(
      g = g,
      old = element_list$local[ii],
      new = paste0(element_list$global[ii], "-", extension),
      partial_match = FALSE,
      allow_missing = FALSE
    )
  }
  
  return(g)
}



.convert_to_grob <- function(plots_or_grobs) {
  # Convert to list if the input is a single grob or
  unlist_grobs <- FALSE
  if (grid::is.grob(plots_or_grobs) || ggplot2::is_ggplot(plots_or_grobs)) {
    plots_or_grobs <- list(plots_or_grobs)

    # Set a flag so that we unlist the results after conversion.
    unlist_grobs <- TRUE
  }

  # Initialise list of grobs
  grobs <- list()

  for (p in plots_or_grobs) {
    if (inherits(p, "familiar_ggplot")) {
      # Convert to grob
      g <- suppressWarnings(tryCatch(
        ggplot2::ggplotGrob(p),
        error = identity
      ))
      
      if (inherits(g, "error")) g <- NULL

      # Make changes to g according to p$custom_grob.
      if (!is.null(p$custom_grob)) {
        if (!is.null(p$custom_grob$heights)) {
          # Iterate over the elements that need to be updated.
          for (ii in seq_len(length(p$custom_grob$heights$name))) {
            # Extract name and height
            name <- p$custom_grob$heights$name[ii]
            height <- p$custom_grob$heights$height[ii]

            # Find the intended grob.
            grob_index <- which(g$layout$name == name)

            # Update the height in the layout table.
            g$heights[g$layout[grob_index, "t"]] <- height

            # Update (or set) the height of the grob.
            g$grobs[[grob_index]]$heights <- height
          }
        }

        if (!is.null(p$custom_grob$widths)) {
          # Iterate over the elements that need to be updated.
          for (ii in seq_len(length(p$custom_grob$widths$name))) {
            name <- p$custom_grob$widths$name[ii]
            width <- p$custom_grob$widths$width[ii]

            # Find the intended grob.
            grob_index <- which(g$layout$name == name)

            # Update the height in the layout table.
            g$widths[g$layout[grob_index, "l"]] <- width

            # Update (or set) the height of the grob.
            g$grobs[[grob_index]]$widths <- width
          }
        }
      }
      
    } else if (inherits(p, "ggplot")) {
      # Convert to grob
      g <- suppressWarnings(tryCatch(
        ggplot2::ggplotGrob(p),
        error = identity
      ))
      
      if (inherits(g, "error")) g <- NULL
      
    } else if (inherits(p, "grob")) {
      # Assign grob
      g <- p
      
    } else {
      ..warning(paste0(
        "Could not convert an object of class ", class(p), " to a grob."
      ))
      g <- NULL
    }
    
    # Set panel sizes.
    if (!is.null(g)) {
      # Make panels inherit heights and widths, if they don't have any. This is done
      # to ensure that panels retain heights and widths, even if supporting elements
      # such as the axis text and label elements are stripped on figure composition.
      g <- .gtable_update_panel_aspects(g = g)
    }
    
    grobs <- c(grobs, list(g))
  }

  if (unlist_grobs) grobs <- grobs[[1L]]
  
  return(grobs)
}



.draw_plot <- function(plot_or_grob) {
  suppress_warnings(
    ..draw_plot(plot_or_grob),
    regexp = c(
      "containing missing values",
      "containing non-finite values",
      "no non-missing arguments to min",
      "no non-missing arguments to max"
    )
  )
}



..draw_plot <- function(plot_or_grob) {
  if (ggplot2::is_ggplot(plot_or_grob)) {
    show(plot_or_grob)
    
  } else if (grid::is.grob(plot_or_grob)) {
    grid::grid.newpage()
    grid::grid.draw(plot_or_grob)
    
  } else {
    ..error("Plot could not be drawn.")
  }
  
  return(invisible(NULL))
}



.save_plot_to_file <- function(
    plot_or_grob,
    object,
    dir_path,
    type,
    x,
    subtype = NULL,
    split_by = NULL,
    additional = NULL,
    filename = NULL,
    device = "png",
    ...
) {
  # ... are passed to ggplot2::ggsave

  # Check if the plot object exists
  if (is.null(plot_or_grob)) return(NULL)

  # Check if directory exists
  if (is.encapsulated_path(dir_path)) {
    file_dir <- normalizePath(
      file.path(dir_path, type),
      mustWork = FALSE
    )
    
  } else {
    file_dir <- normalizePath(dir_path, mustWork = FALSE)
  }

  if (!dir.exists(file_dir)) {
    dir.create(file_dir, recursive = TRUE)
  }

  if (!is.null(filename)) {
    # These are the file extensions supported by ggsave (v3.4.0).
    file_extensions <- c(
      "eps", "ps", "tex", "pdf", "svg", "emf", "wmf",
      "png", "jpg", "jpeg", "bmp", "tiff"
    )

    # Test if a file extension is present.
    device_present <- endswith_any(
      filename, 
      suffix = paste0(".", file_extensions)
    )
    
    if (any(device_present)) {
      # Update device indicated by the filename.
      device <- head(file_extensions[device_present], n = 1L)

      # Remove device from filename.
      filename <- sub_last(
        pattern = paste0(".", device),
        replacement = "",
        x = filename
      )
    }

    # Extend the filename if multiple plots are created from the same data.
    if (!is.null(split_by)) {
      subtype <- paste0(
        as.character(sapply(split_by, function(jj, x) (x[[jj]][1L]), x = x)),
        collapse = "_"
      )

      filename <- paste0(filename, subtype, collapse = "_")
    }
    
  } else {
    # Set subtype.
    subtype <- .create_plot_subtype(
      x = x,
      subtype = subtype,
      split_by = split_by,
      additional = additional
    )

    # Combine type and subtype as the filename.
    filename <- paste0(
      type,
      ifelse(is.null(subtype), "", paste0("_", subtype))
    )
  }

  for (current_device in device) {
    # Add in extension again.
    filename <- paste0(filename, ".", current_device)

    # There may be an issue with a cold RStudio where the plotting devices have
    # not started.
    tryCatch(
      
      # Call ggsave to save the data
      suppressMessages(
        do.call(
          ggplot2::ggsave,
          args = c(
            list(
              "filename" = filename,
              "plot" = plot_or_grob,
              "device" = current_device,
              "path" = file_dir
            ),
            list(...)
          )
        )
      ),
      error = function(err) {
        logger_warning(
          paste0(
            "Could not create plot ",
            filename,
            ". The OS may not allow long file names."
          )
        )
      }
    )
  }
  
  return(invisible(NULL))
}



.get_plot_results <- function(
    dir_path = NULL,
    plot_list = NULL,
    export_collection = FALSE,
    object = NULL
) {
  
  # Do not return plot information.
  if (!is.null(dir_path)) plot_list <- NULL

  if (export_collection) {
    return(list(
      "collection" = object,
      "plot_list" = plot_list
    ))
    
  } else {
    return(plot_list)
  }
}



.format_plot_number <- function(
    x,
    digits = 3L,
    common_base = NULL,
    min_common_base = NULL,
    max_common_base = NULL,
    character_out = TRUE
) {
  # Determine the common base.
  if (is.null(common_base)) {
    common_base <- ..format_get_common_base(x)
    if (is.numeric(max_common_base)) {
      common_base <- ifelse(common_base > max_common_base, max_common_base, common_base)
    }
    if (is.numeric(min_common_base)) {
      common_base <- ifelse(common_base < min_common_base, min_common_base, common_base)
    }
  }

  # Round numbers.
  x <- as.integer(round(x / 10.0^(1.0 + common_base - digits))) * 10.0^(1.0 + common_base - digits)

  # Format output.
  if (character_out) {
    return(format(x, digits = digits, trim = TRUE))
    
  } else {
    return(x)
  }
}



..format_get_common_base <- function(x) {
  # Find the base-10 integer of the data.
  x_base <- floor(log10(abs(x)))
  x_base <- x_base[is.finite(x_base)]
  
  return(as.integer(max(x_base)))
}




.format_plot_number_nice_range <- function(input_range, x) {
  # Shrink input range to first and last value
  input_range <- c(
    head(input_range, n = 1L),
    tail(input_range, n = 1L)
  )

  # Find values in input_range that should be replaced.
  replace_index <- is.na(input_range)

  # Return input range if no updating is required.
  if (!any(replace_index)) return(input_range)

  # Find range of values in x.
  value_range <- range(x)

  # Replace missing elements of the input range.
  input_range[replace_index] <- value_range[replace_index]

  # Make the input range nice
  nice_range <- range(labeling::extended(
    dmin = input_range[1L],
    dmax = input_range[2L],
    m = 5L,
    only.loose = TRUE
  ))

  # Update the input_range with nice values
  input_range[replace_index] <- nice_range[replace_index]

  return(input_range)
}



.convert_dendrogram_to_table <- function(h, similarity_metric) {
  # Suppress NOTES due to non-standard evaluation in data.table
  x_1 <- y_1 <- x_2 <- NULL

  # Convert to dendrogram
  h <- stats::as.dendrogram(h)

  # Determine the metric range.
  metric_range <- get_similarity_range(
    similarity_metric = similarity_metric, 
    as_distance = TRUE
  )

  # Convert dendogram to a list of connectors that can later be used for
  # plotting. Note that we do not know where the origin should be located on the
  # x-axis. We will correct for that later.
  connectors <- .decompose_dendrogram(
    h = h,
    parent_height = max(metric_range)
  )

  # Combine into single data.table.
  connectors <- data.table::rbindlist(connectors)

  # Keep only nodes with finite parent height (y_1). Depending on the metric,
  # this means that the connector with the origin may not be drawn.
  connectors <- connectors[is.finite(y_1)]

  # Return null if there are no connectors to be drawn.
  if (is_empty(connectors)) return(NULL)

  # Reposition the left-most leaf to 0.0.
  min_leaf_pos <- min(c(connectors$x_1, connectors$x_2))
  connectors[, ":="(
    "x_1" = x_1 - min_leaf_pos,
    "x_2" = x_2 - min_leaf_pos
  )]
  
  return(connectors)
}



.decompose_dendrogram <- function(
    h, 
    parent_height = Inf, 
    parent_x = NA, 
    leafs_visited = 0L
) {
  # Decompose dendogram. The function is designed to iterate through a
  # dendogram, and obtain the connector between node (h) and its parent, as well
  # as the connectors between the node and its children h[[1]] and h[[2]],
  # unless it has no children.

  dend_attr <- attributes(h)

  if (is.na(parent_x)) {
    parent_x <- ifelse(is.null(dend_attr$midpoint), 0.0, dend_attr$midpoint)
  }

  if (is.null(dend_attr$midpoint)) {
    # This indicates that the node has no children.

    # Connector from parent.
    conn_parent_child <- data.table::data.table(
      "x_1" = parent_x,
      "y_1" = parent_height,
      "x_2" = parent_x,
      "y_2" = dend_attr$height,
      "feature" = dend_attr$label
    )

    return(list(conn_parent_child))
  }

  # Connector from parent.
  conn_parent_child <- data.table::data.table(
    "x_1" = parent_x,
    "y_1" = parent_height,
    "x_2" = parent_x,
    "y_2" = dend_attr$height,
    "feature" = NA_character_
  )
  
  # Left child node x-axis location
  if (!is.null(attributes(h[[1L]])$midpoint)) {
    child_l_pos <- leafs_visited + attributes(h[[1L]])$midpoint
  } else {
    child_l_pos <- leafs_visited
  }

  # Connector to left leaf.
  conn_child_l_leaf <- data.table::data.table(
    "x_1" = parent_x,
    "y_1" = dend_attr$height,
    "x_2" = child_l_pos,
    "y_2" = dend_attr$height,
    "feature" = NA_character_
  )

  # Right child node x-axis location
  if (
    !is.null(attributes(h[[2L]])$midpoint) &&
    !is.null(attributes(h[[1L]])$members)
  ) {
    child_r_pos <- leafs_visited + attributes(h[[1L]])$members + attributes(h[[2L]])$midpoint
  } else if (!is.null(attributes(h[[1L]])$members)) {
    child_r_pos <- leafs_visited + attributes(h[[1L]])$members
  } else {
    child_r_pos <- leafs_visited
  }

  # Connector to right leaf.
  conn_child_r_leaf <- data.table::data.table(
    "x_1" = parent_x,
    "y_1" = dend_attr$height,
    "x_2" = child_r_pos,
    "y_2" = dend_attr$height,
    "feature" = NA_character_
  )

  # Add data.tables as list elements.
  connector_list <- list(
    conn_parent_child,
    conn_child_l_leaf, 
    conn_child_r_leaf
  )

  # Left leaf
  if (!is.null(h[[1L]])) {
    left_leaf_connectors <- .decompose_dendrogram(
      h = h[[1L]],
      parent_height = dend_attr$height,
      parent_x = child_l_pos,
      leafs_visited = leafs_visited
    )

    # Append to list
    connector_list <- append(connector_list, left_leaf_connectors)
  }

  # Right leaf
  if (!is.null(h[[2L]])) {
    right_leaf_connectors <- .decompose_dendrogram(
      h = h[[2L]],
      parent_height = dend_attr$height,
      parent_x = child_r_pos,
      leafs_visited = ifelse(
        !is.null(attributes(h[[1L]])$members),
        leafs_visited + attributes(h[[1L]])$members, 
        leafs_visited
      )
    )

    # Append to list
    connector_list <- append(connector_list, right_leaf_connectors)
  }

  return(connector_list)
}



..set_edge_points <- function(x, range, type) {
  # Function used to determine edge points, such as used for ggplot2::geom_rect.
  if (!is.numeric(x)) {
    x <- as.numeric(x)
    range <- c(0.5, length(x) + 0.5)
  }

  if (length(x) > 1L) {
    # Make sure x is sorted ascendingly.
    sort_index <- sort(x, index.return = TRUE)$ix
    x <- x[sort_index]

    # Compute difference between subsequent values.
    diff_x <- diff(x)

    # Compute edges.
    xmax <- c(
      head(x, n = length(x) - 1L) + diff_x / 2.0,
      tail(x, n = 1L) + tail(diff_x, n = 1L) / 2.0
    )
    xmin <- c(
      head(x, n = 1L) - head(diff_x, n = 1L) / 2.0,
      tail(x, n = length(x) - 1L) - diff_x / 2.0
    )

    # Shuffle back to input order.
    xmax[sort_index] <- xmax
    xmin[sort_index] <- xmin
    
  } else {
    xmin <- range[1L]
    xmax <- range[2L]
  }

  edge_points <- list(xmin, xmax)
  if (type == "x") {
    names(edge_points) <- c("xmin", "xmax")
  } else if (type == "y") {
    names(edge_points) <- c("ymin", "ymax")
  } else {
    ..error_reached_unreachable_code(paste0(
      "..set_edge_points: unknown type specified: ", type
    ))
  }

  return(edge_points)
}



..get_luminosity <- function(col) {
  values <- as.vector(grDevices::col2rgb(col)) / 255.0
  return ((min(values) + max(values)) / 2.0)
}

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Default
familiar:::test_export(
  export_function = familiar:::export_model_performance,
  data_element = "model_performance",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = c("binomial", "multinomial", "survival"),
  not_available_single_sample = c("binomial", "multinomial", "survival"),
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  estimation_type = "point",
  create_novelty_detector = FALSE,
  use_prediction_table = TRUE,
  debug = debug_flag
)

results <- familiar:::test_export_specific(
  export_function = familiar:::export_model_performance,
  data_element = "model_performance",
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = c("binomial", "multinomial", "survival"),
  not_available_single_sample = c("binomial", "multinomial", "survival"),
  outcome_type_available = c("continuous", "binomial", "multinomial", "survival"),
  create_novelty_detector = FALSE,
  use_prediction_table = FALSE,
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL


# create_instance_weights (dataObject) -----------------------------------------
setMethod(
  "create_instance_weights",
  signature(data = "dataObject"),
  function(data, method, outcome_type = NULL, ...) {
    
    if (is.null(outcome_type)) outcome_type <- data@outcome_type
    
    # Pass to data.table.
    return(create_instance_weights(
      data = data@data,
      method = method,
      outcome_type = outcome_type,
      ...
    ))
  }
)



# create_instance_weights (NULL) -----------------------------------------------
setMethod(
  "create_instance_weights",
  signature(data = "NULL"),
  function(data, method, ...) {
    return(NULL)
  }
)



# create_instance_weights (data.table) -----------------------------------------
setMethod(
  "create_instance_weights",
  signature(data = "data.table"),
  function(
    data,
    outcome_type,
    method = NULL,
    normalisation = "none",
    ...
  ) {
    
    # Check that an outcome-type is provided.
    if (is.null(outcome_type)) {
      ..error_reached_unreachable_code("create_instance_weights: an outcome type should be provided.")
    }
    
    # Only generate weights for outcomes with classes.
    if (!outcome_type %in% c("binomial", "multinomial")) return(NULL)
    
    # Return NULL if method equals NULL.
    if (is.null(method)) return(NULL)
    
    # Obtain id columns.
    id_columns <- get_id_columns(id_depth = "repetition")
    
    # Obtain data columns.
    outcome_columns <- get_outcome_columns(outcome_type)
    
    # Remove all feature columns.
    data <- data.table::copy(data[, mget(c(id_columns, outcome_columns))])
    
    if (method == "inverse_number_of_samples") {
      weights <- .compute_inverse_number_of_samples(data)
      
    } else if (method == "effective_number_of_samples") {
      weights <- .compute_effective_number_of_samples(data, ...)
      
    } else if (method == "none") {
      weights <- rep(1.0, nrow(data)) / nrow(data)
      
    } else {
      # Unknown method.
      ..error_reached_unreachable_code(paste0(
        "create_instance_weights: unknown class weighting method. Found: ", method
      ))
    }
    
    # Normalisation.
    if (normalisation == "sum_one") {
      weights <- weights / sum(weights)
      
    } else if (normalisation == "average_one") {
      weights <- weights * length(weights) / sum(weights)
      
    } else if (normalisation == "none") {
      weights <- weights
      
    } else {
      # Unknown normalisation method.
      ..error_reached_unreachable_code(paste0(
        "create_instance_weights: unknown normalisation method. Found: ", normalisation
      ))
    }
    
    return(weights)
  }
)



.compute_inverse_number_of_samples <- function(x, ...) {
  # Compute the number of samples in each class, and then weight by inverse
  # weights.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  n <- .NATURAL <- NULL
  
  # Sum the number of instances for each outcome.
  y <- x[, list("n" = .N), by = "outcome"]
  
  # Compute the sample weight for each outcome.
  y[, "sample_weight" := 1.0 / n]
  
  # Merge with x.
  return(x[y, on = .NATURAL]$sample_weight)
}



.compute_effective_number_of_samples <- function(x, beta, ...) {
  # Compute the number of samples in each class, and create weights by effective
  # number of samples. Based on Cui Y, Jia M, Lin T-Y, Song Y, Belongie S.
  # Class-balanced loss based on effective number of samples. 2019 IEEE/CVF
  # Conference on Computer Vision and Pattern Recognition (CVPR). IEEE; 2019.
  # pp. 9268–9277. 10.1109/cvpr.2019.00949
  
  # Suppress NOTES due to non-standard evaluation in data.table
  n <- .NATURAL <- NULL
  
  # Sum the number of instances for each outcome.
  y <- x[, list("n" = .N), by = "outcome"]
  
  # Compute the sample weight for each outcome.
  y[, "sample_weight" := (1.0 - beta) / (1.0 - beta^n)]
  
  # Merge with x.
  return(x[y, on = .NATURAL]$sample_weight)
}



..compute_effective_number_of_samples_beta <- function(x) {
  # Convert a negative input number from the model hyperparameters to a beta
  # parameter.
  
  # Check that the number is negative.
  if (x >= 0.0) {
    ..error_reached_unreachable_code(paste0(
      "..compute_effective_number_of_samples_beta: parameter should be negative. Found: ", x
    ))
  } 
  
  return(1.0 - 10.0^x)
}



.get_default_sample_weighting_method <- function(outcome_type) {
  
  # Set default method.
  default_method <- ifelse(
    outcome_type %in% c("binomial", "multinomial"),
    "inverse_number_of_samples",
    "none"
  )
  
  # Set hyperparameter.
  return(.set_hyperparameter(
    default = default_method,
    type = "factor",
    range = c("inverse_number_of_samples", "effective_number_of_samples", "none"),
    randomise = FALSE
  ))
}



.get_default_sample_weighting_beta <- function(method, outcome_type) {
  
  # Default values.
  if (
    "effective_number_of_samples" %in% method &&
    outcome_type %in% c("binomial", "multinomial")
  ) {
    default_values <- c(-4L, -3L, -2L, -1L)
    
  } else {
    default_values <- -2L
  }
  
  return(.set_hyperparameter(
    default = default_values,
    type = "integer",
    range = c(-6L, -1L),
    randomise = length(default_values) > 1L
  ))
}

#' @include FamiliarS4Generics.R
NULL



# familiarSimilarityMetric -----------------------------------------------------
setClass(
  "familiarSimilarityMetric",
  slots = list(
    "similarity_metric" = "character",
    "range" = "numeric",
    "distance" = "logical",
    "require_normalisation" = "logical",
    "highly_similar_threshold" = "numeric"
  ),
  prototype = list(
    "similarity_metric" = character(),
    "range" = NA_real_,
    "distance" = logical(),
    "require_normalisation" = logical(),
    "highly_similar_threshold" = NA_real_
  )
)



# familiarSimilarityMetricPseudoR2 ---------------------------------------------
setClass(
  "familiarSimilarityMetricPseudoR2",
  contains = "familiarSimilarityMetric",
  prototype = methods::prototype(
    "range" = c(0.0, 1.0),
    "distance" = FALSE,
    "require_normalisation" = FALSE,
    "highly_similar_threshold" = 0.80
  )
)



# familiarSimilarityMetricCorrelation ------------------------------------------
setClass(
  "familiarSimilarityMetricCorrelation",
  contains = "familiarSimilarityMetric",
  prototype = methods::prototype(
    "range" = c(-1.0, 0.0, 1.0),
    "distance" = FALSE,
    "require_normalisation" = FALSE,
    "highly_similar_threshold" = 0.90
  )
)



# familiarSimilarityMetricMutualInformation ------------------------------------
setClass(
  "familiarSimilarityMetricMutualInformation",
  contains = "familiarSimilarityMetric",
  prototype = methods::prototype(
    "range" = c(0.0, 1.0),
    "distance" = FALSE,
    "require_normalisation" = FALSE,
    "highly_similar_threshold" = 0.45
  )
)



# familiarSimilarityMetricDistance ---------------------------------------------
setClass(
  "familiarSimilarityMetricDistance",
  contains = "familiarSimilarityMetric",
  prototype = methods::prototype(
    "range" = c(0.0, Inf),
    "distance" = TRUE,
    "require_normalisation" = TRUE,
    "highly_similar_threshold" = 0.01
  )
)



# is_available (general) -------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarSimilarityMetric"),
  function(object, ...) {
    return(TRUE)
  }
)



# is_available (distance) ------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarSimilarityMetricDistance"),
  function(object, any_categorical, ...) {
    return(!any_categorical)
  }
)



.create_similarity_metric_object <- function(similarity_metric) {
  # Strip some information from similarity_metric
  similarity_metric <- gsub(
    x = similarity_metric, 
    pattern = "_trim", 
    replacement = "",
    fixed = TRUE
  )
  similarity_metric <- gsub(
    x = similarity_metric,
    pattern = "_winsor",
    replacement = "", 
    fixed = TRUE
  )
  
  if (similarity_metric %in% c("cox_snell_r2", "nagelkerke_r2", "mcfadden_r2")) {
    object <- methods::new(
      "familiarSimilarityMetricPseudoR2",
      similarity_metric = similarity_metric
    )
    
  } else if (similarity_metric %in% c("pearson", "kendall", "spearman")) {
    object <- methods::new(
      "familiarSimilarityMetricCorrelation",
      similarity_metric = similarity_metric
    )
    
  } else if (similarity_metric %in% c("mutual_information")) {
    object <- methods::new(
      "familiarSimilarityMetricMutualInformation",
      similarity_metric = similarity_metric
    )

  } else if (similarity_metric %in% c(
    "gower", "euclidean", "manhattan", "chebyshev", "cosine", "canberra", "bray_curtis"
  )) {
    object <- methods::new(
      "familiarSimilarityMetricDistance",
      similarity_metric = similarity_metric
    )
    
  } else {
    ..error_reached_unreachable_code(
      "compute_similarity_metric: encountered unknown similarity metric"
    )
  }
  
  return(object)
}



# encode_categorical_variables methods -----------------------------------------

## encode_categorical_variables (general) --------------------------------------
setMethod(
  "encode_categorical_variables",
  signature(
    object = "familiarSimilarityMetric",
    data = "ANY"
  ),
  function(object, data, ...) {
    
    # Extract data table with contrasts.
    data <- encode_categorical_variables(
      data = data,
      object = NULL,
      encoding_method = "dummy",
      drop_levels = FALSE
    )
    
    return(data$encoded_data)
  }
)



## encode_categorical_variables (distance) -------------------------------------
setMethod(
  "encode_categorical_variables",
  signature(
    object = "familiarSimilarityMetricDistance",
    data = "ANY"
  ),
  function(object, data, ...) {
    
    # Leave data as is.
    return(data)
  }
)



# .compute_feature_similarity methods ------------------------------------------

compute_feature_similarity_metric <- function(
    data,
    similarity_metric,
    feature_info_list,
    cl = NULL,
    verbose = FALSE
) {
  
  if (is_empty(data)) return(NULL)
  if (!is(data, "dataObject")) ..error_reached_unreachable_code("data should be dataObject.")
  
  # Strip some information from similarity_metric
  similarity_metric_type <- gsub(
    x = similarity_metric, 
    pattern = "_trim", 
    replacement = "",
    fixed = TRUE
  )
  similarity_metric_type <- gsub(
    x = similarity_metric_type,
    pattern = "_winsor",
    replacement = "", 
    fixed = TRUE
  )
  
  # Find similarity object.
  object <- .create_similarity_metric_object(similarity_metric = similarity_metric_type)
  
  # Set feature columns.
  feature_columns <- get_feature_columns(data)
  
  # Check that the number of features is at least two. This is more of technical
  # requirement than anything else.
  if (length(feature_columns) < 2L) return(NULL)
  
  # Set the categorical mask.
  categorical_mask <- sapply(
    feature_info_list[feature_columns],
    function(x) (x@feature_type == "factor")
  )

  if (!is_available(object = object, any_categorical = any(categorical_mask))) {
    ..warning(
      message = paste0(
        "The ", similarity_metric, " metric can not be used to assess similarity between ",
        "categorical variables. We will use mutual_information instead."
      ),
      warning_class = "parameter_replacement_warning"
    )
    
    # Replace metric by mutual_information.
    object <- .create_similarity_metric_object(similarity_metric = "mutual_information")
  }
  
  # Normalise data, if required.
  if (object@require_normalisation) {
    # Identify numerical features
    numerical_features <- feature_columns[!categorical_mask]
    
    # Create a local copy of data.
    data@data <- data.table::copy(data@data)
    
    # Find the normalisation method.
    norm_method <- "normalisation"
    if (grepl(pattern = "_trim", x = similarity_metric, fixed = TRUE)) {
      norm_method <- "normalisation_trim"
    } else if (grepl(pattern = "_winsor", x = similarity_metric, fixed = TRUE)) {
      norm_method <- "normalisation_winsor"
    }
    
    # Perform normalisation.
    for (ii in numerical_features) {
      data.table::set(data@data, j = ii, value = .normalise(
        x = data@data[[ii]],
        normalisation_method = norm_method,
        range = c(0.0, 1.0)
      ))
    }
  }
  
  # Generate all combinations of features
  combinations <- utils::combn(sort(feature_columns), 2L)
  
  # Determine similarity measures for each feature pair.
  similarity <- fam_sapply(
    cl = cl,
    assign = NULL,
    X = seq_len(ncol(combinations)),
    function(ii, combinations, data, object, categorical_mask) {
      
      # Identify features that are being compared.
      feature_1 <- combinations[1L, ii]
      feature_2 <- combinations[2L, ii]
      
      # Compute pairwise similarity
      similarity <- .compute_feature_similarity(
        object = object,
        x = data[[feature_1]],
        y = data[[feature_2]],
        x_categorical = categorical_mask[feature_1],
        y_categorical = categorical_mask[feature_2]
      )
      
      return(similarity)
    },
    progress_bar = verbose,
    combinations = combinations,
    data = droplevels(data@data),
    object = object,
    categorical_mask = categorical_mask,
    chopchop = TRUE
  )
  
  # Transform similarity scores into a data.table.
  similarity_data  <- data.table::data.table(
    "feature_name_1" = combinations[1L, ],
    "feature_name_2" = combinations[2L, ],
    "value" = similarity
  )
  
  return(similarity_data)
}





## .compute_feature_similarity (generic) ---------------------------------------
setGeneric(
  ".compute_feature_similarity",
  function(object, ...) standardGeneric(".compute_feature_similarity")
)



## .compute_feature_similarity (general) ---------------------------------------
setMethod(
  ".compute_feature_similarity",
  signature(object = "familiarSimilarityMetric"),
  function(object, ...) {
    # Fall-back method.
    return(NA_real_)
  }
)


## .compute_feature_similarity (pseudo r2) -------------------------------------
setMethod(
  ".compute_feature_similarity",
  signature(object = "familiarSimilarityMetricPseudoR2"),
  function(
    object, 
    x, 
    y, 
    x_categorical, 
    y_categorical, 
    type = "approximate", 
    ...
  ) {
    
    .check_parameter_value_is_valid(
      x = type,
      var_name = "type",
      values = c("exact", "approximate")
    )
    
    ..encode_variable_to_list <- function(x, is_categorical, insert_intercept = FALSE) {
      if (is_categorical) {
        # Categorical variables are encoded as numeric levels (ordinal), or
        # using one-hot-encoding (nominal).
        if (is.ordered(x)) {
          x <- list(as.numeric(x))
          
        } else {
          # Dummy encoding of categorical variable.
          level_names <- levels(x)
          level_count <- nlevels(x)
          
          x <- lapply(
            level_names[2L:level_count],
            function(ii, x) (as.numeric(x == ii)),
            x = x
          )
        }
      } else {
        # Numeric variables are only stored as a list.
        x <- list(x)
      }
      
      # Set names
      names(x) <- paste0("name_", seq_along(x))
      
      if (insert_intercept) {
        x <- c(x, list("intercept__" = numeric(length(x)) + 1.0))
      }
      
      return(data.table::as.data.table(x))
    }
    
    # Remove missing elements.
    valid_elements <- is.finite(x) & is.finite(y)
    if (sum(valid_elements) <= 1L) return(callNextMethod())
    
    # Keep only valid elements.
    x <- x[valid_elements]
    y <- y[valid_elements]
    
    # Check if x and y are both invariant.
    x_invariant <- all(x == x[1L])
    y_invariant <- all(y == y[1L])
    if (x_invariant && y_invariant) return(1.0)
    if (x_invariant || y_invariant) return(0.0)
    
    if (type == "approximate") {
      # Select the number of samples for the computing approximate distances.
      n_samples <- as.integer(min(c(length(x), ceiling(1000.0^0.3 * length(x)^0.7))))
      
      # Set sample indices - avoid selecting samples randomly, as that is a
      # somewhat costly operation.
      sample_index <- as.integer(round(seq_len(n_samples) / n_samples * length(x)))
      
      # Select sample subset.
      x <- x[sample_index]
      if (x_categorical) x <- droplevels(x)
      y <- y[sample_index]
      if (y_categorical) y <- droplevels(y)
      
      # Check if x and y are now both invariant.
      x_invariant <- all(x == x[1L])
      y_invariant <- all(y == y[1L])
      if (x_invariant && y_invariant) return(1.0)
      if (x_invariant || y_invariant) return(0.0)
    }
    
    # Find analysis type and whether x and y should be swapped in the models,
    # based on information content.
    analysis_info <- .get_analysis_type_pairwise_similarity(
      x = x,
      y = y,
      x_categorical = x_categorical,
      y_categorical = y_categorical
    )
    
    x <- analysis_info$x
    y <- analysis_info$y
    x_categorical <- analysis_info$x_categorical
    y_categorical <- analysis_info$y_categorical
    
    if (analysis_info$type %in% c("gaussian", "binomial")) {
      
      if (analysis_info$type == "gaussian") {
        fitting_family <- stats::gaussian()
      } else {
        fitting_family <- stats::binomial()
      }
      
      if (require_package("fastglm", message_type = "silent")) {
        predictors <- ..encode_variable_to_list(
          x = x,
          is_categorical = x_categorical,
          insert_intercept = TRUE
        )
        
        # Fit informative model.
        model <- fastglm::fastglm(
          x = as.matrix(predictors),
          y = y,
          family = fitting_family,
          method = 3L
        )
        
        predictor_names <- setdiff(names(model$coefficients), "intercept__")
        
        if (!all(is.finite(model$coefficients[predictor_names]))) return(0.0)
        if (all(approximately(model$coefficients[predictor_names], 0.0))) return(0.0)
        if (approximately(model$deviance, 0.0)) return(1.0)
        
        # Fit uninformative model.
        null_model <- fastglm::fastglm(
          x = as.matrix(numeric(length(y)), ncol = 1L),
          y = y,
          family = fitting_family,
          method = 3L
        )
        
      } else {
        # Fit informative model.
        model <- stats::glm(
          "y ~ x",
          data = data.table::data.table("x" = x, "y" = y),
          family = fitting_family
        )
        
        predictor_names <- setdiff(names(coef(model)), "(Intercept)")
        
        if (!all(is.finite(model$coefficients[predictor_names]))) return(0.0)
        if (all(near(model$coefficients[predictor_names], 0.0, df = 2L * length(x)))) return(0.0)
        if (approximately(model$deviance, 0.0)) return(1.0)
        
        # Fit uninformative model.
        null_model <- stats::glm(
          "y ~ 1",
          data = data.table::data.table("x" = x, "y" = y),
          family = fitting_family
        )
      }
      
      # Compute log-likelihoods
      model_loglik <- stats::logLik(model)[1L]
      null_loglik <- stats::logLik(null_model)[1L]
      
    } else if (analysis_info$type == "multinomial") {
      
      require_package(
        x = "nnet",
        purpose = paste0(
          "to compute log-likelihood pseudo R2 similarity using the ",
          object@similarity_metric, " metric"
        )
      )
      
      # Set predictors and response.
      predictors <- ..encode_variable_to_list(
        x = x,
        is_categorical = x_categorical
      )
      response <- data.table::data.table("response" = y)
      
      # Set model formula.
      model_formula <- stats::reformulate(
        termlabels = names(predictors),
        response = quote(response)
      )
      null_formula <- stats::as.formula("response ~ 1")
      
      quiet(model <- nnet::multinom(
        model_formula,
        data = cbind(predictors, response),
        maxit = ifelse(type == "approximate", 100L, 500L),
        MaxNWts = Inf
      ))
      
      model_coefficients <- stats::coef(model)
      predictor_names <- setdiff(colnames(model_coefficients), "(Intercept)")
      
      if (!all(is.finite(model_coefficients[, predictor_names]))) return(0.0)
      if (all(near(as.vector(model_coefficients[, predictor_names]), 0.0, df = 2L * length(x)))) return(0.0)
      if (approximately(model$deviance, 0.0)) return(1.0)
      
      quiet(null_model <- nnet::multinom(
        null_formula,
        data = cbind(predictors, response),
        maxit = ifelse(type == "approximate", 100L, 500L)
      ))
      
      # Compute log-likelihoods
      model_loglik <- stats::logLik(model)[1L]
      null_loglik <- stats::logLik(null_model)[1L]
      
    } else {
      ..error_reached_unreachable_code(paste0(
        ".compute_similarity,familiarSimilarityMetricPseudoR2: ",
        "encountered unknown analysis type"
      ))
    }
    
    # Compute pseudo R-squared values from log-likelihoods
    if (object@similarity_metric == "mcfadden_r2") {
      # Compute McFadden's R-squared
      similarity <- 1.0 - model_loglik / null_loglik
      
    } else if (object@similarity_metric == "cox_snell_r2") {
      # Compute Cox and Snell's R-squared
      similarity <- 1.0 - (exp(null_loglik) / exp(model_loglik))^(2.0 / length(x))
      
    } else if (object@similarity_metric == "nagelkerke_r2") {
      # Compute Nagelkerke's R-squared
      similarity <- (1.0 - (exp(null_loglik) / exp(model_loglik))^(2.0 / length(x))) /
        (1.0 - exp(null_loglik)^(2.0 / length(x)))
      
    } else {
      ..error_reached_unreachable_code(paste0(
        ".compute_similarity,familiarSimilarityMetricPseudoR2: ",
        "encountered unknown similarity metric"
      ))
    }
    
    # Check if similarity is NaN or infinite.
    if (!is.finite(similarity)) similarity <- 0.0
    
    # Limit to range [0,1]
    if (similarity < 0.0) {
      similarity <- 0.0
    } else if (similarity > 1.0) {
      # Greater than 1.0 should be set to 1.0.
      similarity <- 1.0
    } else if (approximately(similarity, 1.0)) {
      # Check if values are very close to 1.0.
      similarity <- 1.0
    }
    
    return(similarity)
  }
)



## .compute_feature_similarity (correlation) -----------------------------------
setMethod(
  ".compute_feature_similarity",
  signature(object = "familiarSimilarityMetricCorrelation"),
  function(
    object, 
    x, 
    y, 
    x_categorical, 
    y_categorical, 
    ...
  ) {
   
    ..dummy_encode <- function(x) {
      # Dummy encoding of categorical variable.
      level_names <- levels(x)
      level_count <- nlevels(x)
      
      return(lapply(
        level_names[2L:level_count],
        function(ii, x) (as.numeric(x == ii)),
        x = x
      ))
    }
    
    ..encode_variable_to_list <- function(x, is_categorical) {
      if (is_categorical) {
        # Categorical variables are encoded as numeric levels (ordinal), or using
        # one-hot-encoding (nominal).
        if (is.ordered(x)) {
          x <- list(as.numeric(x))
          
        } else {
          x <- ..dummy_encode(x)
        }
        
      } else {
        # Numeric variables are only stored as a list.
        x <- list(x)
      }
      
      return(x)
    }
    
    # Cast x and y to lists. Categorical variables are encoded.
    x <- ..encode_variable_to_list(x = x, is_categorical = x_categorical)
    y <- ..encode_variable_to_list(x = y, is_categorical = y_categorical)
    
    # Get all list combinations; for numeric x and y this is always 1, but can
    # be more if x and/or y are factors
    combinations <- expand.grid(
      ii = seq_along(x), 
      jj = seq_along(y),
      KEEP.OUT.ATTRS = FALSE,
      stringsAsFactors = FALSE
    )
    
    # Calculate correlation coefficients
    correlation_coef <- sapply(
      seq_len(nrow(combinations)),
      function(kk, combinations, x, y, method) {
        stats::cor(
          x = x[[combinations$ii[kk]]],
          y = y[[combinations$jj[kk]]],
          use = "na.or.complete",
          method = method
        )
      },
      combinations = combinations,
      x = x,
      y = y,
      method = object@similarity_metric
    )
    
    if (x_categorical || y_categorical) {
      correlation_coef <- abs(correlation_coef)
    }
    
    # Compute similarity as the mean correlation coefficient.
    similarity <- mean(correlation_coef, na.rm = TRUE)
    
    # Check for NaN or infinite values.
    if (!is.finite(similarity)) similarity <- 0.0
    
    # Return similarity
    return(similarity)
  }
)



## .compute_feature_similarity (mutual information) ----------------------------
setMethod(
  ".compute_feature_similarity",
  signature(object = "familiarSimilarityMetricMutualInformation"),
  function(
    object, 
    x, 
    y, 
    x_categorical, 
    y_categorical, 
    ...
  ) {
    # mutual information can be mapped to spearman:
    # log(mi) = -5.244 + 4.183 * sp
    
    # Remove missing elements.
    valid_elements <- is.finite(x) & is.finite(y)
    if (sum(valid_elements) <= 1L) return(callNextMethod())
    
    # Keep only valid elements.
    x <- x[valid_elements]
    y <- y[valid_elements]
    
    # Check if there are more than one unique values in x and or y.
    if (all(x == x[1L]) && all(y == y[1L])) return(1.0)
    
    # Ensure that numeric values are actually encoded as numeric, because
    # praznik handles integers as categorical variables. This is not the
    # convention used by familiar.
    if (!x_categorical && is.integer(x)) x <- as.numeric(x)
    if (!y_categorical && is.integer(y)) y <- as.numeric(y)
    
    # require_package(
    #   x = "praznik",
    #   purpose = paste0(
    #     "to compute similarity using the ",
    #     object@similarity_metric, " metric"
    #   )
    # )
    
    # Compute normalised mutual information.
    return(
      praznik::miScores(x, y, threads = 1L) / praznik::jhScores(x, y, threads = 1L)
    )
  }
)



## .compute_feature_similarity (distance) --------------------------------------
setMethod(
  ".compute_feature_similarity",
  signature(object = "familiarSimilarityMetricDistance"),
  function(
    object, 
    x, 
    y, 
    x_categorical, 
    y_categorical, 
    ...
  ) {
    
    # For feature-wise comparison, the presence of categorical features is
    # problematic, because they don't have a "distance".
    if (x_categorical || y_categorical) return(callNextMethod())
    
    # Pass to sample similarity.
    return(.compute_sample_similarity(
      object = object,
      x = x,
      y = y,
      categorical = logical(length(x)),
      ...
    ))
  }
)



compute_sample_similarity_metric <- function(
    data,
    similarity_metric,
    feature_info_list,
    cl = NULL,
    verbose = FALSE
) {
  
  if (is_empty(data)) return(NULL)
  if (!is(data, "dataObject")) ..error_reached_unreachable_code("data should be dataObject.")
  
  # Strip some information from similarity_metric
  similarity_metric_type <- gsub(
    x = similarity_metric, 
    pattern = "_trim", 
    replacement = "",
    fixed = TRUE
  )
  similarity_metric_type <- gsub(
    x = similarity_metric_type,
    pattern = "_winsor",
    replacement = "", 
    fixed = TRUE
  )
  
  # Find similarity object.
  object <- .create_similarity_metric_object(
    similarity_metric = similarity_metric_type
  )
  
  # Set feature columns.
  feature_columns <- get_feature_columns(data)
  
  # Set the categorical mask.
  categorical_mask <- sapply(
    feature_info_list[feature_columns],
    function(x) (x@feature_type == "factor")
  )
  
  # Normalise data, if required.
  if (object@require_normalisation) {
    # Identify numerical features
    numerical_features <- feature_columns[!categorical_mask]
    
    # Create a local copy of data.
    data@data <- data.table::copy(data@data)
    
    # Find the normalisation method.
    norm_method <- "normalisation"
    if (grepl(pattern = "_trim", x = similarity_metric, fixed = TRUE)) {
      norm_method <- "normalisation_trim"
    } else if (grepl(pattern = "_winsor", x = similarity_metric, fixed = TRUE)) {
      norm_method <- "normalisation_winsor"
    }
    
    # Perform normalisation.
    for (ii in numerical_features) {
      data.table::set(data@data, j = ii, value = .normalise(
        x = data@data[[ii]],
        normalisation_method = norm_method,
        range = c(0.0, 1.0)
      ))
    }
  }
  
  # Check that the number of rows is at least two. This is more
  # of technical requirement than anything else.
  if (nrow(data@data) < 2L) return(NULL)
  
  # Generate all combinations of samples
  combinations <- utils::combn(seq_len(nrow(data@data)), 2L)
  
  # Determine similarity measures for each sample pair.
  similarity <- fam_sapply(
    cl = cl,
    assign = NULL,
    X = seq_len(ncol(combinations)),
    function(ii, combinations, data, object, categorical_mask) {
      
      # Identify features that are being compared.
      row_1 <- combinations[1L, ii]
      row_2 <- combinations[2L, ii]
      
      # Compute pairwise similarity
      similarity <- .compute_sample_similarity(
        object = object,
        x = as.numeric(data[row_1, ]),
        y = as.numeric(data[row_2, ]),
        categorical = categorical_mask
      )
      
      return(similarity)
    },
    progress_bar = verbose,
    combinations = combinations,
    data = data@data[, mget(feature_columns)],
    object = object,
    categorical_mask = categorical_mask,
    chopchop = TRUE
  )
  
  # Create unique row names.
  row_names <- get_unique_row_names(x = data)
  
  # Transform similarity scores into a data.table.
  similarity_data  <- data.table::data.table(
    "sample_1" = row_names[combinations[1L, ]],
    "sample_2" = row_names[combinations[2L, ]],
    "value" = similarity
  )
  
  return(similarity_data)
}



# .compute_sample_similarity (generic) -----------------------------------------
setGeneric(
  ".compute_sample_similarity",
  function(object, ...) standardGeneric(".compute_sample_similarity")
)



# .compute_sample_similarity (general) -----------------------------------------
setMethod(
  ".compute_sample_similarity",
  signature(object = "familiarSimilarityMetric"),
  function(object, ...) {
    # Fall-back method.
    return(NA_real_)
  }
)


# .compute_sample_similarity (pseudo r2) --------------------------------------
setMethod(
  ".compute_sample_similarity",
  signature(object = "familiarSimilarityMetricPseudoR2"),
  function(
    object, 
    x, 
    y, 
    categorical, 
    ...
  ) {
    
    # The presence of categorical variables is problematic for computing
    # pseudo R2, because we can only correlation
    # between vectors that have a consistent data type, in this case numeric.
    if (any(categorical)) return(callNextMethod())
    
    # Pass to feature similarity routine.
    return(.compute_feature_similarity(
      object = object,
      x = x,
      y = y,
      categorical_x = FALSE,
      categorical_y = FALSE,
      ...
    ))
  }
)



# .compute_sample_similarity (correlation) -------------------------------------
setMethod(
  ".compute_sample_similarity",
  signature(object = "familiarSimilarityMetricCorrelation"),
  function(
    object, 
    x, 
    y, 
    categorical, 
    ...
  ) {
    
    # The presence of categorical variables is problematic for computing
    # sample-wise similarity, because we can only correlation
    # between vectors that have a consistent data type, in this case numeric.
    if (any(categorical)) return(callNextMethod())
    
    # Pass to feature similarity routine.
    return(.compute_feature_similarity(
      object = object,
      x = x,
      y = y,
      categorical_x = FALSE,
      categorical_y = FALSE,
      ...
    ))
  }
)



# .compute_sample_similarity (mutual information) ------------------------------
setMethod(
  ".compute_sample_similarity",
  signature(object = "familiarSimilarityMetricMutualInformation"),
  function(
    object, 
    x, 
    y, 
    categorical, 
    ...
  ) {
    
    # The presence of categorical variables is problematic for computing
    # sample-wise similarity, because we can only compute mutual information
    # between vectors that have a consistent data type, in this case numeric.
    if (any(categorical)) return(callNextMethod())
    
    # Pass to feature similarity routine.
    return(.compute_feature_similarity(
      object = object,
      x = x,
      y = y,
      categorical_x = FALSE,
      categorical_y = FALSE,
      ...
    ))
  }
)



# .compute_sample_similarity (distance) ----------------------------------------
setMethod(
  ".compute_sample_similarity",
  signature(object = "familiarSimilarityMetricDistance"),
  function(
    object, 
    x, 
    y, 
    categorical,
    ...
  ) {
    
    # Ensure that data presented as numeric values.
    if (!is.numeric(x)) x <- as.numeric(x)
    if (!is.numeric(y)) y <- as.numeric(y)
    
    # Remove missing elements.
    valid_elements <- is.finite(x) & is.finite(y)
    if (sum(valid_elements) < 1.0) return(callNextMethod())
    
    # Keep only valid elements.
    x <- x[valid_elements]
    y <- y[valid_elements]
    
    # Handle categorical variables so that distance = 0 if they are the same,
    # and distance = 1 if they are not.
    if (any(categorical)) {
      # Determine where values match.
      matching_values <- x[categorical] == y[categorical]
      
      # Assign the same value in x and y in case of matching categorical values.
      x[categorical][matching_values] <- 1.0
      y[categorical][matching_values] <- 1.0
      
      # Assign different values in x and y in case of mismatching categorical
      # values.
      x[categorical][!matching_values] <- 0.0
      y[categorical][!matching_values] <- 1.0
    }
    
    if (object@similarity_metric == "gower") {
      distance <- sum(abs(x - y)) / length(x)
    } else if (object@similarity_metric == "euclidean") {
      distance <- sqrt(sum((x - y)^2.0))
    } else if (object@similarity_metric == "manhattan") {
      distance <- sum(abs(x - y))
    } else if (object@similarity_metric == "chebyshev") {
      distance <- max(abs(x - y))
    } else if (object@similarity_metric == "cosine") {
      distance <- 1.0 - sum(x * y) / (sqrt(sum(x^2.0)) * sqrt(sum(y^2.0)))
    } else if (object@similarity_metric == "canberra") {
      distance <- sum(abs(x - y) / (abs(x) + abs(y)))
    } else if (object@similarity_metric == "bray_curtis") {
      distance <- sum(abs(x - y)) / sum(abs(x + y))
    }
    
    return(distance)
  }
)



.get_analysis_type_pairwise_similarity <- function(x, y, x_categorical, y_categorical) {
  # Determine analysis types: note that we expect the outcome to be y: hence it
  # is recommended that predictor x contains the widest range of variables, i.e.
  # continuous or multinomial variables.
  #
  # This function checks the analysis type and whether x and y should be swapped
  # in the analysis.

  # Determine number of levels of categorical features.
  n_x <- ifelse(x_categorical, nlevels(x), data.table::uniqueN(x))
  n_y <- ifelse(y_categorical, nlevels(y), data.table::uniqueN(y))

  if (x_categorical && y_categorical) {
    # Case 1: Both x and y are categorical variables.

    # Determine analysis type.
    analysis_type <- ifelse(min(c(n_x, n_y)) <= 2L, "binomial", "multinomial")

    # Determine if swapping is required.
    requires_swap <- n_x < n_y
    
  } else if (x_categorical) {
    # Case 2: x is a categorical variable, and y numerical.

    # Determine analysis type.
    analysis_type <- ifelse(n_x <= 2L, "binomial", "multinomial")

    # Swapping is required.
    requires_swap <- TRUE
    
  } else if (y_categorical) {
    # Case 3: x is a numerical variable, and y categorical.

    # Determine analysis type.
    analysis_type <- ifelse(n_y <= 2L, "binomial", "multinomial")

    # Swapping is not required.
    requires_swap <- FALSE
    
  } else {
    # Case 4: both x and y are numerical variables.
    analysis_type <- "gaussian"
    requires_swap <- n_x < n_y
  }

  if (requires_swap) {
    x_out <- y
    y_out <- x
    x_categorical_out <- y_categorical
    y_categorical_out <- x_categorical
    
  } else {
    x_out <- x
    y_out <- y
    x_categorical_out <- x_categorical
    y_categorical_out <- y_categorical
  }
  
  return(list(
    "type" = analysis_type,
    "swap" = requires_swap,
    "x" = x_out,
    "y" = y_out,
    "x_categorical" = x_categorical_out,
    "y_categorical" = y_categorical_out
  ))
}


# .similarity_to_distance ------------------------------------------------------

convert_similarity_to_distance <- function(x, similarity_metric) {
  # Convert to distance.
  
  # Find similarity object.
  object <- .create_similarity_metric_object(similarity_metric = similarity_metric)
  
  return(.similarity_to_distance(
    object = object,
    x = x
  ))
}



## .similarity_to_distance (generic) -------------------------------------------
setGeneric(
  ".similarity_to_distance",
  function(object, ...) standardGeneric(".similarity_to_distance")
)



## .similarity_to_distance (pseudo r2) -----------------------------------------
setMethod(
  ".similarity_to_distance",
  signature(object = "familiarSimilarityMetricPseudoR2"),
  function(object, x, ...) {
    return(1.0 - sqrt(x))
  }
)



## .similarity_to_distance (correlation) ---------------------------------------
setMethod(
  ".similarity_to_distance",
  signature(object = "familiarSimilarityMetricCorrelation"),
  function(object, x, ...) {
    return(1.0 - abs(x))
  }
)



## .similarity_to_distance (mutual information) --------------------------------
setMethod(
  ".similarity_to_distance",
  signature(object = "familiarSimilarityMetricMutualInformation"),
  function(object, x, ...) {
    return(1.0 - x)
  }
)



## .similarity_to_distance (distance) ------------------------------------------
setMethod(
  ".similarity_to_distance",
  signature(object = "familiarSimilarityMetricDistance"),
  function(object, x, ...) {
    # Distance-based metrics don't require conversion to distance.
    return(x)
  }
)


# .distance_to_similarity ------------------------------------------------------

convert_distance_to_similarity <- function(x, similarity_metric) {
  # Convert distance to similarity.

  # Find similarity object.
  object <- .create_similarity_metric_object(similarity_metric = similarity_metric)
  
  return(.distance_to_similarity(
    object = object,
    x = x
  ))
}



## .distance_to_similarity (generic) -------------------------------------------
setGeneric(
  ".distance_to_similarity",
  function(object, ...) standardGeneric(".distance_to_similarity")
)



## .distance_to_similarity (pseudo r2) -----------------------------------------
setMethod(
  ".distance_to_similarity",
  signature(object = "familiarSimilarityMetricPseudoR2"),
  function(object, x, ...) {
    return((1.0 - x)^2.0)
  }
)



## .distance_to_similarity (correlation) ---------------------------------------
setMethod(
  ".distance_to_similarity",
  signature(object = "familiarSimilarityMetricCorrelation"),
  function(object, x, ...) {
    # Note that the sign is lost at conversion back to similarity.
    return(1.0 - abs(x))
  }
)



## .distance_to_similarity (mutual information) --------------------------------
setMethod(
  ".distance_to_similarity",
  signature(object = "familiarSimilarityMetricMutualInformation"),
  function(object, x, ...) {
    return(1.0 - x)
  }
)



## .distance_to_similarity (distance) ------------------------------------------
setMethod(
  ".distance_to_similarity",
  signature(object = "familiarSimilarityMetricDistance"),
  function(object, x, ...) {
    # Return normalised similarity.
    return(1.0 - x / max(x, na.rm = TRUE))
  }
)



get_high_similarity_threshold <- function(similarity_metric) {

  # Find similarity object.
  object <- .create_similarity_metric_object(similarity_metric = similarity_metric)
  
  return(object@highly_similar_threshold)
}



get_similarity_range <- function(similarity_metric, as_distance = FALSE) {
  
  # Find similarity object.
  object <- .create_similarity_metric_object(similarity_metric = similarity_metric)
  
  value_range <- object@range
  
  if (as_distance) {
    value_range <- .similarity_to_distance(
      object = object,
      x = value_range
    )
    
    value_range <- sort(unique(value_range))
  }
  
  return(value_range)
}



is_default_distance <- function(similarity_metric) {
  # Find similarity object.
  object <- .create_similarity_metric_object(similarity_metric = similarity_metric)
  
  return(object@distance)
}



.get_available_similarity_metrics <- function(data_type = "feature") {
  if (data_type %in% c("feature", "cluster")) {
    # Pair-wise comparison between features.
    return(c(
      "mcfadden_r2", "cox_snell_r2", "nagelkerke_r2", "spearman", 
      "kendall", "pearson", "mutual_information"
    ))
    
  } else {
    # Pair-wise comparison between samples.
    return(c(
      "gower", "gower_trim", "gower_winsor",
      "euclidean", "euclidean_trim", "euclidean_winsor"
    ))
  }
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
#' @include Clustering.R
#' @include ClusteringMethod.R
NULL



# add_feature_info_parameters (singularClusteringObject, dataObject) -----------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "singularClusteringObject",
    data = "dataObject"
  ),
  function(
    object,
    data,
    ...
  ) {
    
    # Sanity check: the cluster is singular.
    if (length(object@cluster_features) != 1L) {
      ..error_reached_unreachable_code(paste0(
        "add_feature_info_parameters,singularClusteringObject,dataObject: ",
        "the number of features is expected to be 1. Found: ", length(object@cluster_features)
      ))
    }
    
    # Create representation object for a singular cluster.
    representation_object <- methods::new(
      "clusterRepresentationObject",
      "name" = object@cluster_features,
      "weight" = 1.0,
      "invert" = FALSE,
      "cluster_name" = object@cluster_features,
      "cluster_size" = 1L,
      "cluster_features" = object@cluster_features,
      "required_features" = object@cluster_features
    )
    
    # Set as list.
    representation_object_list <- list(representation_object)
    
    # Add name to list elements.
    names(representation_object_list) <- sapply(
      representation_object_list,
      function(x) (x@name)
    )
    
    return(representation_object_list)
  }
)



# add_feature_info_parameters (clusteringObject, dataObject) -------------------
setMethod(
  "add_feature_info_parameters",
  signature(
    object = "clusteringObject",
    data = "dataObject"
  ),
  function(
    object,
    data,
    feature_info_list,
    cluster_method_object,
    ...
  ) {
    
    if (object@representation_method == "none") {
      # Form singular clusters
      representation_object_list <- .cluster_representation_by_singular_features(object = object)
      
    } else if (object@representation_method == "medioid") {
      # Represent clusters by the most central feature.
      representation_object_list <- .cluster_representation_by_medioid(
        object = object,
        cluster_object = cluster_method_object
      )
      
    } else if (object@representation_method == "best_predictor") {
      # Represent clusters by the best predictor.
      representation_object_list <- .cluster_representation_by_best_predictor(
        object = object,
        data = data
      )
      
    } else if (object@representation_method == "mean") {
      # Represent cluster as a meta-feature.
      representation_object_list <- .cluster_representation_as_mean_feature(
        object = object,
        data = data,
        feature_info_list = feature_info_list,
        cluster_object = cluster_method_object
      )
      
    } else if (object@representation_method == "concordance") {
      # TODO implement.
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "add_feature_info_parameters,clusteringObject,dataObject: ",
        "encountered an unknown representation method: ", object@representation_method
      ))
    }
    
    # Add name to list elements.
    names(representation_object_list) <- sapply(
      representation_object_list,
      function(x) (x@name)
    )
    
    return(representation_object_list)
  }
)



.cluster_representation_by_singular_features <- function(object, ...) {
  # This forms singular clusters.
  
  # Iterate over features to form singular clusters.
  representation_object_list <- lapply(
    object@cluster_features,
    function(current_feature) {
      # Create representation object for a singular cluster.
      representation_object <- methods::new(
        "clusterRepresentationObject",
        "name" = current_feature,
        "weight" = 1.0,
        "invert" = FALSE,
        "cluster_name" = current_feature,
        "cluster_size" = 1L,
        "cluster_features" = current_feature,
        "required_features" = current_feature
      )
      
      return(representation_object)
    }
  )
  
  return(representation_object_list)
}



.cluster_representation_by_medioid <- function(
    object,
    cluster_object,
    ...
) {
  # Represent clusters by their medioid feature.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  element_1 <- element_2 <- value <- average_distance <- NULL
  
  # Get distance table.
  distance_table <- get_distance_table(
    object = cluster_object@similarity_table,
    include_diagonal = FALSE
  )
  
  # Determine whether the similarity table is for features (columns) or samples
  # (rows).
  element_names <- .get_cluster_data_type_element_name(
    data_type = cluster_object@data_type
  )
  
  # Change names in similarity table to a generic name.
  data.table::setnames(
    x = distance_table,
    old = element_names,
    new = c("element_1", "element_2")
  )
  
  # Limit distance table to only features in the cluster.
  distance_table <- distance_table[
    element_1 %in% object@cluster_features & element_2 %in% object@cluster_features
  ]
  
  # Find mean distance for each feature.
  distance_table <- distance_table[
    ,
    list("average_distance" = mean(value)),
    by = "element_1"
  ]
  
  # Find the feature to be used as the representative feature.
  representative_feature <- distance_table[
    average_distance == min(average_distance)
  ]$element_1[1L]
  
  # Iterate over features to form representation objects.
  representation_object_list <- lapply(
    object@cluster_features,
    function(current_feature, cluster_features, representative_feature) {
      # Create representation object for a singular cluster.
      representation_object <- methods::new(
        "clusterRepresentationObject",
        "name" = current_feature,
        "weight" = ifelse(current_feature == representative_feature, 1.0, 0.0),
        "invert" = FALSE,
        "cluster_name" = paste0(representative_feature, "_cluster"),
        "cluster_size" = length(cluster_features),
        "cluster_features" = cluster_features,
        "required_features" = representative_feature
      )
      
      return(representation_object)
    },
    cluster_features = object@cluster_features,
    representative_feature = representative_feature
  )
  
  return(representation_object_list)
}



.cluster_representation_by_best_predictor <- function(
    object,
    data,
    ...
) {
  # Represent clusters by the most predictive feature.
  
  # Suppress NOTES due to non-standard evaluation in data.table
  p_value <-  NULL
  
  # Select features
  data <- select_features(
    data = data,
    features = object@cluster_features
  )
  
  # Aggregate data
  data <- aggregate_data(data = data)
  
  # Calculate p-values of the features in the clusters
  feature_p_values <- compute_univariable_p_values(
    data_obj = data,
    feature_columns = object@cluster_features
  )
  
  # Integrate in a data.table
  predictor_data <- data.table::data.table(
    "name" = names(feature_p_values),
    "p_value" = feature_p_values
  )
  
  # Replace non-finite p-values.
  predictor_data[!is.finite(p_value), "p_value" := 1.0]
  
  # Find the feature to be used as the representative feature.
  representative_feature <- predictor_data[p_value == min(p_value)]$name[1L]
  
  # Iterate over features to form representation objects.
  representation_object_list <- lapply(
    object@cluster_features,
    function(
      current_feature,
      cluster_features,
      representative_feature
    ) {
      # Create representation object for a singular cluster.
      representation_object <- methods::new(
        "clusterRepresentationObject",
        "name" = current_feature,
        "weight" = ifelse(current_feature == representative_feature, 1.0, 0.0),
        "invert" = FALSE,
        "cluster_name" = paste0(representative_feature, "_cluster"),
        "cluster_size" = length(cluster_features),
        "cluster_features" = cluster_features,
        "required_features" = representative_feature
      )
      
      return(representation_object)
    },
    cluster_features = object@cluster_features,
    representative_feature = representative_feature
  )
  
  return(representation_object_list)
}



.cluster_representation_as_mean_feature <- function(
    object,
    data,
    feature_info_list,
    ...
) {
  # Represent clusters by the mean
  
  # Check if any of the features is categorical - for these "mean" clusters can
  # not be realistically created.
  if (
    any(
      sapply(
        feature_info_list[object@cluster_features],
        function(x) (x@feature_type)
      ) == "factor"
    )
  ) {
    return(.cluster_representation_by_medioid(
      object = object,
      ...
    ))
  }
  
  ..cluster_representation_as_mean_feature <- function(
    current_feature,
    feature_correlation,
    cluster_features
  ) {
    
    # Create representation object for a cluster meta-feature.
    representation_object <- methods::new(
      "clusterRepresentationObject",
      "name" = current_feature,
      "weight" = 1.0 / length(cluster_features),
      "invert" = feature_correlation < 0.0,
      "cluster_name" = paste0(cluster_features[1L], "_meta_cluster"),
      "cluster_size" = length(cluster_features),
      "cluster_features" = cluster_features,
      "required_features" = cluster_features
    )
    
    return(representation_object)
  }
  
  # Select features
  data <- select_features(data = data, features = object@cluster_features)
  
  # Compute Spearman correlation between the first feature and the other
  # features.
  feature_correlation <- sapply(
    data@data[, mget(object@cluster_features)],
    stats::cor,
    y = data@data[[object@cluster_features[1L]]],
    use = "na.or.complete",
    method = "spearman"
  )
  
  # Iterate over features to form representation objects.
  representation_object_list <- mapply(
    ..cluster_representation_as_mean_feature,
    current_feature = object@cluster_features,
    feature_correlation = feature_correlation,
    MoreArgs = list("cluster_features" = object@cluster_features),
    SIMPLIFY = FALSE
  )
  
  return(representation_object_list)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R


##### is_empty-------------------------------------------------------------------
setMethod(
  "is_empty",
  signature(x = "NULL"),
  function(x, ...) {
    return(TRUE)
  }
)

setMethod(
  "is_empty",
  signature(x = "data.table"),
  function(x, ...) {
    return(ncol(x) == 0L || nrow(x) == 0L)
  }
)

setMethod(
  "is_empty",
  signature(x = "character"),
  function(x, ...) {
    if (length(x) == 0L) {
      return(TRUE)
    } else if (all(x == "")) {
      return(TRUE)
    } else {
      return(FALSE)
    }
  }
)

setMethod(
  "is_empty",
  signature(x = "dataObject"),
  function(x, allow_no_features = FALSE, ...) {
    if (is.null(x@data)) {
      # Data is empty if it is NULL
      return(TRUE)
    } else if (nrow(x@data) == 0L) {
      # Data is empty if the data table has no rows
      return(TRUE)
    } else if (!allow_no_features && !has_feature_data(x)) {
      # Data is empty if there are no features.
      return(TRUE)
    } else {
      # Data is present otherwise
      return(FALSE)
    }
  }
)

setMethod(
  "is_empty",
  signature(x = "delayedDataObject"),
  function(x, ...) {
    # Data is not empty until is has been properly loaded
    return(FALSE)
  }
)

setMethod(
  "is_empty",
  signature(x = "list"),
  function(x, ...) {
    return(length(x) == 0L)
  }
)

setMethod(
  "is_empty", 
  signature(x = "vimpTable"),
  function(x, ...) {
    return(is_empty(x@vimp_table))
  }
)

setMethod(
  "is_empty",
  signature(x = "vector"),
  function(x, ...) {
    return(length(x) == 0L)
  }
)

setMethod(
  "is_empty",
  signature(x = "familiarDataElement"),
  function(x, ...) {
    return(is_empty(x@data))
  }
)



# get_outcome_class_levels------------------------------------------------------
setMethod(
  "get_outcome_class_levels",
  signature(x = "outcomeInfo"),
  function(x) {
    if (x@outcome_type %in% c("binomial", "multinomial")) {
      return(as.character(x@levels))
    } else {
      return(character(0L))
    }
  }
)

setMethod("get_outcome_class_levels", signature(x = "familiarModel"), function(x) {
  return(get_outcome_class_levels(x@outcome_info))
})

setMethod("get_outcome_class_levels", signature(x = "familiarEnsemble"), function(x) {
  return(get_outcome_class_levels(x@outcome_info))
})

setMethod("get_outcome_class_levels", signature(x = "familiarData"), function(x) {
  return(get_outcome_class_levels(x@outcome_info))
})

setMethod("get_outcome_class_levels", signature(x = "familiarCollection"), function(x) {
  return(get_outcome_class_levels(x@outcome_info))
})

setMethod("get_outcome_class_levels", signature(x = "data.table"), function(x, outcome_type) {
  return(.get_outcome_class_levels(data = x, outcome_type = outcome_type))
})

setMethod("get_outcome_class_levels", signature(x = "dataObject"), function(x) {
  return(.get_outcome_class_levels(data = x@data, outcome_type = x@outcome_type))
})

.get_outcome_class_levels <- function(data, outcome_type) {
  if (outcome_type %in% c("survival", "continuous", "competing_risk")) {
    return(character(0L))
    
  } else if (outcome_type %in% c("binomial", "multinomial")) {
    return(as.character(levels(data[[get_outcome_columns(x = outcome_type)]])))
    
  } else {
    ..error_no_known_outcome_type(outcome_type)
  }
}



# get_class_probability_name----------------------------------------------------
setMethod(
  "get_class_probability_name",
  signature(x = "outcomeInfo"),
  function(x) {
    # Check outcome type
    if (x@outcome_type %in% c("binomial", "multinomial")) {
      # Find class levels
      class_levels <- get_outcome_class_levels(x = x)
      
      # Pass to the method for character strings.
      return(get_class_probability_name(x = class_levels))
    } else {
      return(NULL)
    }
  }
)

setMethod("get_class_probability_name", signature(x = "familiarModel"), function(x) {
  return(get_class_probability_name(x = x@outcome_info))
})

setMethod("get_class_probability_name", signature(x = "familiarEnsemble"), function(x) {
  return(get_class_probability_name(x = x@outcome_info))
})

setMethod("get_class_probability_name", signature(x = "familiarData"), function(x) {
  return(get_class_probability_name(x = x@outcome_info))
})

setMethod("get_class_probability_name", signature(x = "familiarCollection"), function(x) {
  return(get_class_probability_name(x = x@outcome_info))
})

setMethod("get_class_probability_name", signature(x = "ANY"), function(x) {
  # Backup for numeric and logical categories.
  return(get_class_probability_name(x = as.character(x)))
})

setMethod("get_class_probability_name", signature(x = "dataObject"), function(x) {
  outcome_info <- .get_outcome_info(x = x)

  return(get_class_probability_name(x = outcome_info))
})

setMethod(
  "get_class_probability_name",
  signature(x = "data.table"),
  function(x, outcome_type) {
    if (outcome_type %in% c("binomial", "multinomial")) {
      # Find class levels
      class_levels <- get_outcome_class_levels(x = x)
      
      # Pass to the method for character strings.
      return(get_class_probability_name(x = class_levels))
      
    } else {
      return(NULL)
    }
  }
)

setMethod(
  "get_class_probability_name", 
  signature(x = "character"),
  function(x) {
    # Create column names
    class_probability_columns <- .replace_illegal_column_name(
      column_name = paste0("predicted_class_probability_", x)
    )
    
    return(class_probability_columns)
  }
)



# get_outcome_name--------------------------------------------------------------
setMethod("get_outcome_name", signature(x = "familiarModel"), function(x) {
  return(get_outcome_name(x@outcome_info))
})

setMethod("get_outcome_name", signature(x = "familiarEnsemble"), function(x) {
  return(get_outcome_name(x@outcome_info))
})

setMethod("get_outcome_name", signature(x = "familiarData"), function(x) {
  return(get_outcome_name(x@outcome_info))
})

setMethod("get_outcome_name", signature(x = "familiarCollection"), function(x) {
  return(get_outcome_name(x@outcome_info))
})

setMethod("get_outcome_name", signature(x = "outcomeInfo"), function(x) {
  return(x@name)
})



# get_outcome_columns-----------------------------------------------------------
setMethod("get_outcome_columns", signature(x = "character"), function(x) {
  return(.get_outcome_columns(outcome_type = x))
})

setMethod("get_outcome_columns", signature(x = "dataObject"), function(x) {
  return(.get_outcome_columns(outcome_type = x@outcome_type))
})

setMethod("get_outcome_columns", signature(x = "outcomeInfo"), function(x) {
  return(.get_outcome_columns(outcome_type = x@outcome_type))
})

setMethod("get_outcome_columns", signature(x = "familiarModel"), function(x) {
  return(.get_outcome_columns(outcome_type = x@outcome_type))
})

setMethod("get_outcome_columns", signature(x = "familiarEnsemble"), function(x) {
  return(.get_outcome_columns(outcome_type = x@outcome_type))
})

# Internal function
.get_outcome_columns <- function(outcome_type) {
  # Set outcome columns
  if (outcome_type %in% c("survival")) {
    outcome_cols <- c("outcome_time", "outcome_event")
    
  } else if (outcome_type %in% c("binomial", "multinomial", "continuous")) {
    outcome_cols <- c("outcome")
    
  } else if (outcome_type == "unsupervised") {
    outcome_cols <- NULL
    
  } else if (outcome_type == "competing_risk") {
    ..error_outcome_type_not_implemented(outcome_type)
    
  } else {
    ..error_no_known_outcome_type(outcome_type)
  }

  return(outcome_cols)
}



# get_non_feature_columns-------------------------------------------------------
setMethod("get_non_feature_columns", signature(x = "character"), function(x, id_depth = "repetition") {
  return(.get_non_feature_columns(outcome_type = x, id_depth = id_depth))
})

setMethod("get_non_feature_columns", signature(x = "outcomeInfo"), function(x, id_depth = "repetition") {
  return(.get_non_feature_columns(outcome_type = x@outcome_type, id_depth = id_depth))
})

setMethod("get_non_feature_columns", signature(x = "dataObject"), function(x, id_depth = "repetition") {
  return(.get_non_feature_columns(outcome_type = x@outcome_type, id_depth = id_depth))
})

setMethod("get_non_feature_columns", signature(x = "familiarVimpMethod"), function(x, id_depth = "repetition") {
  return(.get_non_feature_columns(outcome_type = x@outcome_type, id_depth = id_depth))
})

setMethod("get_non_feature_columns", signature(x = "familiarNoveltyDetector"), function(x, id_depth = "repetition") {
  return(.get_non_feature_columns(outcome_type = "unsupervised", id_depth = id_depth))
})

setMethod("get_non_feature_columns", signature(x = "familiarModel"), function(x, id_depth = "repetition") {
  return(.get_non_feature_columns(outcome_type = x@outcome_type, id_depth = id_depth))
})

setMethod("get_non_feature_columns", signature(x = "familiarEnsemble"), function(x, id_depth = "repetition") {
  return(.get_non_feature_columns(outcome_type = x@outcome_type, id_depth = id_depth))
})

# Internal function
.get_non_feature_columns <- function(outcome_type, id_depth = "repetition") {
  # Returns column names in data table which are not features

  # Find outcome columns
  outcome_columns <- get_outcome_columns(x = outcome_type)

  # Find the id-columns
  id_columns <- get_id_columns(id_depth = id_depth)

  # Generate the names of the non-feature columns
  return(c(outcome_columns, id_columns))
}



# get_feature_columns-----------------------------------------------------------
setMethod(
  "get_feature_columns",
  signature(x = "data.table"),
  function(x, outcome_type) {
    return(.get_feature_columns(
      column_names = colnames(x),
      outcome_type = outcome_type
    ))
  }
)

setMethod(
  "get_feature_columns",
  signature(x = "dataObject"),
  function(x) {
    return(.get_feature_columns(
      column_names = colnames(x@data),
      outcome_type = x@outcome_type
    ))
  }
)

# Internal function
.get_feature_columns <- function(column_names, outcome_type) {
  if (is.null(column_names)) {
    # Return 0-length character in case column_names is NULL.
    return(character(0L))
    
  } else if (length(column_names) == 0L) {
    # Return 0-length character in case column_names has no length.
    return(character(0L))
    
  } else {
    # Return all column names that are not related to identifiers and outcome.
    return(setdiff(column_names, get_non_feature_columns(x = outcome_type)))
  }
}


# get_n_features----------------------------------------------------------------
setMethod("get_n_features", signature(x = "data.table"), function(x, outcome_type) {
  return(length(get_feature_columns(x = x, outcome_type = outcome_type)))
})

setMethod("get_n_features", signature(x = "dataObject"), function(x) {
  return(length(get_feature_columns(x = x)))
})


# has_feature_data--------------------------------------------------------------
setMethod("has_feature_data", signature(x = "data.table"), function(x, outcome_type) {
  return(get_n_features(x = x, outcome_type = outcome_type) > 0L)
})

setMethod("has_feature_data", signature(x = "dataObject"), function(x, outcome_type) {
  return(get_n_features(x = x) > 0L)
})


# get_unique_row_names----------------------------------------------------------
setMethod(
  "get_unique_row_names",
  signature(x = "data.table"),
  function(x, include_batch = NULL, include_series = NULL) {
    
    if (is_empty(x)) return(character(0L))
    
    if (is.null(include_batch)) {
      # Determine if batch identifiers should be included, i.e. the same sample
      # identifiers appear in different batches.
      include_batch <- any(
        unique(
          x[, mget(get_id_columns(id_depth = "sample"))]
        )[, list("n" = .N), by = mget(get_id_columns(single_column = "sample"))]$n > 1L
      )
    }
    
    if (is.null(include_series)) {
      # Determine if series identifiers should be included, i.e. any unique
      # sample has multiple series.
      include_series <- any(
        unique(
          x[, mget(get_id_columns(id_depth = "series"))]
        )[, list("n" = .N), by = mget(get_id_columns(id_depth = "sample"))]$n > 1L
      )
    }
    
    # Get sample names.
    row_names <- as.character(x[[get_id_columns(single_column = "sample")]])
    
    if (any(c(include_batch, include_series))) {
      add_string <- " ("
      
      if (include_batch) {
        add_string <- paste0(add_string, x[[get_id_columns(single_column = "batch")]])
      }
      if (include_batch && include_series) {
        add_string <- paste0(add_string, "; ")
      }
      if (include_series) {
        add_string <- paste0(add_string, x[[get_id_columns(single_column = "series")]])
      }
      
      row_names <- paste0(row_names, add_string, ")")
    }
    
    return(row_names)
  }
)

setMethod(
  "get_unique_row_names",
  signature(x = "dataObject"),
  function(x, include_batch = NULL, include_series = NULL) {
    return(get_unique_row_names(
      x = x@data,
      include_batch = include_batch,
      include_series = include_series
    ))
  }
)



# get_n_samples ----------------------------------------------------------------
setMethod(
  "get_n_samples",
  signature(x = "data.table"),
  function(x, id_depth = "sample", count_unique = TRUE) {
    return(.get_n_samples(
      x = x,
      id_depth = id_depth, 
      count_unique = count_unique
    ))
  }
)

setMethod(
  "get_n_samples",
  signature(x = "dataObject"),
  function(x, id_depth = "sample", count_unique = TRUE) {
    return(.get_n_samples(
      x = x@data,
      id_depth = id_depth,
      count_unique = count_unique
    ))
  }
)

setMethod(
  "get_n_samples",
  signature(x = "NULL"),
  function(x, id_depth = "sample", ...) {
    return(0L)
  }
)


.get_n_samples <- function(x, id_depth, count_unique) {
  # Check if x is empty.
  if (is_empty(x)) {
    return(0L)
  }

  # Find identifier columns.
  id_columns <- get_id_columns(id_depth = id_depth)

  # Return the number of rows with unique values for the combination of
  # identifier columns.
  if (count_unique) {
    return(nrow(unique(x[, mget(id_columns)])))
    
  } else {
    return(nrow(x[, mget(id_columns)]))
  }
}


# encode_categorical_variables--------------------------------------------------
setMethod(
  "encode_categorical_variables",
  signature(
    data = "dataObject",
    object = "ANY"
  ),
  function(
    object = NULL,
    data,
    encoding_method = "effect", 
    drop_levels = TRUE,
    feature_columns = NULL
  ) {
    # Obtain feature columns if not provided.
    if (is.null(feature_columns)) feature_columns <- get_feature_columns(data)
    
    # Find encoding data
    encoding_data <- encode_categorical_variables(
      data = data@data,
      object = object,
      encoding_method = encoding_method,
      drop_levels = drop_levels,
      feature_columns = feature_columns
    )
    
    # Update data
    data@data <- encoding_data$encoded_data
    
    return(list(
      "encoded_data" = data, 
      "reference_table" = encoding_data$reference_table
    ))
  }
)


setMethod(
  "encode_categorical_variables", 
  signature(
    data = "data.table",
    object = "ANY"
  ),
  function(
    object = NULL,
    data, 
    encoding_method = "effect",
    drop_levels = TRUE,
    feature_columns = NULL, 
    outcome_type = NULL
  ) {
    # Determine columns with factors
    if (is.null(feature_columns) && !is.null(outcome_type)) {
      feature_columns <- get_feature_columns(
        x = data,
        outcome_type = outcome_type
      )
      
    } else if (is.null(feature_columns)) {
      feature_columns <- colnames(data)
    }

    # Return original if no variables are available.
    if (length(feature_columns) == 0L) {
      return(list(
        "encoded_data" = data, 
        "reference_table" = NULL
      ))
    }

    # Find column classes
    column_class <- lapply(
      feature_columns, 
      function(ii, data) (class(data[[ii]])),
      data = data
    )

    # Identify categorical columns
    factor_columns <- sapply(
      column_class, 
      function(selected_column_class) {
        return(any(selected_column_class %in% c("logical", "character", "factor")))
      }
    )
    factor_columns <- feature_columns[factor_columns]

    # Return original if no categorical variables are available.
    if (length(factor_columns) == 0L) {
      return(list(
        "encoded_data" = data,
        "reference_table" = NULL
      ))
    }

    # Initialise contrast list and reference list
    contrast_list <- list()
    reference_list <- list()

    # Iterate over columns and encode the categorical variables.
    for (ii in seq_along(factor_columns)) {
      # Extract data from current column.
      current_col <- factor_columns[ii]
      x <- data[[current_col]]

      # Drop missing levels
      if (drop_levels) x <- droplevels(x)

      if (is.factor(x)) {
        # Get names of levels and number of levels
        level_names <- levels(x)
        level_count <- nlevels(x)
        
      } else if (is.character(x)) {
        level_names <- unique(x)
        level_count <- length(level_names)
        
      } else if (is.logical(x)) {
        level_names <- c(FALSE, TRUE)
        level_count <- 2L
      }

      # Apply coding scheme
      if (encoding_method == "effect" || level_count == 1L) {
        # Effect/one-hot encoding
        curr_contrast_list <- lapply(
          level_names,
          function(ii, x) (as.numeric(x == ii)),
          x = x
        )

        # Check level_names for inconsistencies (i.e. spaces etc)
        contrast_names <- .replace_illegal_column_name(
          column_name = paste0(current_col, "_", level_names)
        )
        names(curr_contrast_list) <- contrast_names
        
      } else if (encoding_method == "dummy") {
        # Dummy encoding. The first level is used as a reference (0).
        curr_contrast_list <- lapply(
          level_names[2L:level_count],
          function(ii, x) (as.numeric(x == ii)),
          x = x
        )

        # Check level_names for inconsistencies (i.e. spaces etc)
        contrast_names <- .replace_illegal_column_name(
          column_name = paste0(current_col, "_", level_names[2L:level_count])
        )
        names(curr_contrast_list) <- contrast_names
        
      } else if (encoding_method == "numeric") {
        # Numeric encoding
        browser()
        curr_contrast_list <- list(as.numeric(x))

        # Set names
        contrast_names <- current_col
        names(curr_contrast_list) <- contrast_names
        
      } else {
        ..error_reached_unreachable_code(
          "encode_categorical_variables: unknown encoding method encountered."
        )
      }

      # Add list of generated contrast to contrast list
      contrast_list <- c(contrast_list, curr_contrast_list)

      # Add data table with reference and contrast names to the reference list
      reference_list[[ii]] <- data.table::data.table(
        "original_name" = current_col,
        "reference_name" = contrast_names
      )
    }

    # Check if the original data set contained any features other than
    # categorical features. Combine original data and contrasts into data table
    # with added contrast data, and get a list with reference names.
    if (length(factor_columns) == ncol(data)) {
      contrast_table <- data.table::as.data.table(contrast_list)
      
    } else {
      contrast_table <- cbind(
        data[, !factor_columns, with = FALSE],
        data.table::as.data.table(contrast_list)
      )
    }

    # Return as list
    return(list(
      "encoded_data" = contrast_table,
      "reference_table" = data.table::rbindlist(reference_list)
    ))
  }
)



# get_bootstrap_sample----------------------------------------------------------
setMethod(
  "get_bootstrap_sample",
  signature(data = "dataObject"),
  function(
    data, 
    seed = NULL, 
    rstream_object = NULL, 
    n_samples = NULL, 
    ...
  ) {
    if (.as_preprocessing_level(data) > "none") {
      # Indicating that some preprocessing has taken please.

      # Bootstrap the data element.
      data@data <- get_bootstrap_sample(
        data = data@data,
        seed = seed,
        rstream_object = rstream_object,
        n_samples = n_samples
      )
      
    } else if (length(data@sample_set_on_load) > 0L) {
      if (data.table::is.data.table(data@sample_set_on_load)) {
        data@sample_set_on_load <- get_bootstrap_sample(
          data = data@sample_set_on_load,
          seed = seed,
          rstream_object = rstream_object,
          n_samples = n_samples
        )
      } else {
        # Reshuffle the samples -- This is for backward compatibility.
        data@sample_set_on_load <- fam_sample(
          x = unique(data@sample_set_on_load),
          size = length(data@sample_set_on_load),
          replace = TRUE,
          seed = seed,
          rstream_object = rstream_object,
        )
      }
    } else {
      ..error_reached_unreachable_code(paste0(
        "get_boostrap_sample,dataObject: could not identify a suitable ",
        "method for bootstrapping."
      ))
    }

    return(data)
  }
)

setMethod(
  "get_bootstrap_sample", 
  signature(data = "data.table"),
  function(
    data, 
    seed = NULL, 
    rstream_object = NULL, 
    n_samples = NULL,
    ...
  ) {
    # Find identifier columns at the sample level, i.e. excluding repetitions
    # and series.
    id_columns <- intersect(
      get_id_columns(id_depth = "sample"),
      colnames(data)
    )

    if (length(id_columns) == 0L) {
      size <- nrow(data)
      if (!is.null(n_samples)) size <- min(n_samples, size)
      
      # Sample rows.
      row_ids <- fam_sample(
        x = seq_len(nrow(data)),
        size = size,
        replace = TRUE,
        seed = seed,
        rstream_object = rstream_object
      )

      # Create a subset.
      data <- data[row_ids, ]
      
    } else {
      # List unique samples.
      id_table <- unique(data[, mget(id_columns)])

      size <- nrow(id_table)
      if (!is.null(n_samples)) size <- min(n_samples, size)
      
      # Sample the unique rows of the identifier table.
      row_ids <- fam_sample(
        x = seq_len(nrow(id_table)),
        size = size,
        replace = TRUE,
        seed = seed,
        rstream_object = rstream_object
      )

      # Create subsample.
      id_table <- id_table[row_ids, ]

      # Merge the subsampled identifier table with the data (removing
      # duplicates).
      data <- merge(
        x = id_table,
        y = unique(data),
        by = id_columns,
        all = FALSE,
        allow.cartesian = TRUE
      )
    }

    return(data)
  }
)

setMethod(
  "get_bootstrap_sample",
  signature(data = "NULL"), 
  function(data, seed = NULL, rstream_object = NULL, ...) {
    return(NULL)
  }
)



# get_subsample-----------------------------------------------------------------
setMethod(
  "get_subsample",
  signature(data = "dataObject"),
  function(
    data, 
    seed = NULL, 
    size = NULL, 
    outcome_type = NULL,
    ...
  ) {
    # This function randomly selects up to "size" samples from the data.
    # Set seed for reproducible results.
    if (!is.null(seed)) {
      rstream_object <- .start_random_number_stream(seed = seed)
    }

    if (is.null(outcome_type)) outcome_type <- data@outcome_type

    if (.as_preprocessing_level(data) > "none") {
      # Indicating that some preprocessing has taken please.

      # Bootstrap the data element.
      data@data <- get_subsample(
        data = data@data,
        size = size,
        outcome_type = outcome_type,
        rstream_object = rstream_object
      )
      
    } else if (length(data@sample_set_on_load) > 0L) {
      # Subsample data.
      data@sample_set_on_load <- get_subsample(
        data = data@sample_set_on_load,
        size = size,
        outcome_type = outcome_type,
        rstream_object = rstream_object
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "get_boostrap_sample,dataObject: could not identify a suitable ",
        "method for bootstrapping."
      ))
    }

    return(data)
  }
)


setMethod(
  "get_subsample",
  signature(data = "data.table"),
  function(
    data, 
    seed = NULL, 
    size = NULL,
    outcome_type = NULL,
    rstream_object = NULL,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    .NATURAL <- NULL

    # If no size is set,
    if (is.null(size)) return(data)

    # Update size, if required.
    if (size > nrow(data)) size <- nrow(data)

    # If size is equal to the number of data elements, return the entire data.
    if (size == nrow(data)) return(data)

    # Generate a random stream object.
    if (!is.null(seed) && is.null(rstream_object)) {
      rstream_object <- .start_random_number_stream(seed = seed)
    }

    # Find identifier columns at the sample level, i.e. excluding
    # repetitions and series.
    id_columns <- intersect(
      get_id_columns(id_depth = "sample"),
      colnames(data)
    )

    if (length(id_columns) == 0L) {
      # Sample rows.
      row_ids <- fam_sample(
        x = seq_len(nrow(data)),
        size = size,
        replace = FALSE,
        rstream_object = rstream_object
      )

      # Create a subset.
      data <- data[row_ids, ]
      
    } else if (is.null(outcome_type)) {
      # Sample rows.
      row_ids <- fam_sample(
        x = seq_len(nrow(data)),
        size = size,
        replace = FALSE,
        rstream_object = rstream_object
      )

      # Create a subset.
      data <- data[row_ids, ]
      
    } else {
      # Get subsample.
      id_table <- .create_subsample(data,
        size = size,
        n_iter = 1L,
        outcome_type = outcome_type,
        rstream_object = rstream_object
      )

      # Isolate identifier table.
      id_table <- id_table$train_list[[1L]]

      # Select data.
      data <- data[id_table, on = .NATURAL]
    }

    return(data)
  }
)


setMethod(
  "get_subsample",
  signature(data = "NULL"), 
  function(data, seed = NULL, ...) {
    return(NULL)
  }
)



# fam_sample--------------------------------------------------------------------
setMethod(
  "fam_sample", signature(x = "ANY"),
  function(
    x,
    size = NULL,
    replace = FALSE, 
    prob = NULL,
    seed = NULL, 
    rstream_object = NULL,
    ...
  ) {
    # This function prevents the documented behaviour of the sample function,
    # where if x is positive, numeric and only has one element, it interprets x
    # as a series of x, i.e. x=seq_len(x). That's bad news if x is a sample
    # identifier.

    # Set size if it is unset.
    if (is.null(size)) size <- length(x)

    if (length(x) == 1L) {
      # Check that size is not greater than 1, if items are to be drawn
      # without replacement.
      if (!replace & size > 1L) {
        ..error("cannot take a sample larger than the population when 'replace = FALSE'")
      }

      return(rep_len(x = x, length.out = size))
      
    } else {
      # If x is a vector, array or list with multiple elements, then all
      # of the above is not an issue, and we can make use of sample.

      if (is.null(seed) & is.null(rstream_object)) {
        return(sample(
          x = x,
          size = size,
          replace = replace,
          prob = prob
        ))
        
      } else {
        return(..fam_sample(
          x = x,
          size = size,
          replace = replace,
          seed = seed,
          rstream_object = rstream_object
        ))
      }
    }
  }
)

setMethod(
  "fam_sample",
  signature(x = "data.table"),
  function(
    x, 
    size = NULL,
    replace = FALSE, 
    prob = NULL,
    seed = NULL,
    rstream_object = NULL, 
    ...
  ) {
    # This function prevents the documented behaviour of the sample function,
    # where if x is positive, numeric and only has one element, it interprets x
    # as a series of x, i.e. x=seq_len(x). That's bad news if x is a sample
    # identifier.

    # Suppress NOTES due to non-standard evaluation in data.table
    ..prob <- NULL

    # Make a local copy of x to prevent changing by reference.
    x <- data.table::copy(x)

    # Get id columns
    id_columns <- get_id_columns("sample")

    # Check if prob equals NULL, and set to FALSE if it is.
    if (is.null(prob)) prob <- FALSE

    # Check that batch_id and sample_id columns are present.
    if (!all(id_columns %in% colnames(x))) {
      ..error_reached_unreachable_code(
        "fam_sample,data.table: batch_id or sample_id columns are not present in x."
      )
    }

    # Check that the prob column is present.
    if (is.logical(prob)) {
      if (prob && !"prob" %in% colnames(x)) {
        ..error_reached_unreachable_code(
          "fam_sample,data.table: prob column is not present in x."
        )
      }

      # Select unique samples.
      x <- unique(x, by = id_columns)

      # Set probability to 1.0 for all samples.
      if (!prob) x[, "prob" := 1.0]
      
    } else {
      # Assume that the prob argument is provided as a separate vector
      if (length(x) == nrow(x)) {
        # In this case, assume that prob can be inserted as a new column.

        # Assign prob.
        x[, "prob" := ..prob]

        # Select unique samples.
        x <- unique(x, by = id_columns)
        
      } else {
        ..error_reached_unreachable_code(
          "fam_sample,data.table: the length of prob does not equal the number of instances in x."
        )
      }
    }

    # Set size if it is unset.
    if (is.null(size)) size <- nrow(x)

    # Sample x.
    if (nrow(x) == 1L) {
      # Check that size is not greater than 1, if items are to be drawn without
      # replacement.
      if (!replace && size > 1L) {
        ..error("cannot take a sample larger than the population when 'replace = FALSE'")
      }

      # Get row-ids
      row_id <- rep_len(x = 1L, length.out = size)

      # Create output table y.
      y <- x[row_id, mget(id_columns)]
      
    } else {
      # If x has a multiple instances, then all of the above is not an issue,
      # and we can use the standard implementation of sample.

      # Use default prob=NULL if all probabilities in x are 1.0.
      if (all(x$prob == 1.0)) {
        prob <- NULL
      } else {
        prob <- x$prob
      }

      if (is.null(seed) && is.null(rstream_object)) {
        # Get sampled row identifiers.
        row_id <- sample(
          x = seq_len(nrow(x)),
          size = size,
          replace = replace,
          prob = prob
        )
        
      } else {
        row_id <- ..fam_sample(
          x = seq_len(nrow(x)),
          size = size,
          replace = replace,
          seed = seed,
          rstream_object = rstream_object
        )
      }

      # Create output table y.
      y <- x[row_id, mget(id_columns)]
    }

    return(y)
  }
)


# has_optimised_hyperparameters ------------------------------------------------
setMethod("has_optimised_hyperparameters", signature(object = "familiarModel"), function(object) {
  return(.has_optimised_hyperparameters(object = object))
})

setMethod("has_optimised_hyperparameters", signature(object = "familiarVimpMethod"), function(object) {
  return(.has_optimised_hyperparameters(object = object))
})

setMethod("has_optimised_hyperparameters", signature(object = "familiarNoveltyDetector"), function(object) {
  return(.has_optimised_hyperparameters(object = object))
})


.has_optimised_hyperparameters <- function(object) {
  # Check if the object has any default parameters.
  default_parameters <- get_default_hyperparameters(object)

  # If there are no default hyperparameters, return TRUE.
  if (length(default_parameters) == 0L) return(TRUE)

  # Check if any hyperparameters are present in the object.
  if (is_empty(object@hyperparameters)) return(FALSE)

  # Check if any hyperparameters need to be optimised.
  unoptimised_hyperparameter <- sapply(object@hyperparameters, function(x) {
    # If element x is not a list itself, return false.
    if (!rlang::is_bare_list(x)) return(FALSE)

    # If element x is a list, but does not have a "randomise" element return
    # false.
    if (is.null(x$randomise)) return(FALSE)

    # Else, the hyperparameter requires optimisation.
    return(TRUE)
  })

  # If any hyperparameter was unoptimised return false.
  if (any(unoptimised_hyperparameter)) return(FALSE)

  # If any hyperparameter is missing, return false.
  if (!all(names(default_parameters) %in% names(object@hyperparameters))) {
    return(FALSE)
  }

  # If none of the above cases apply, the hyperparameters can be assumed to be
  # optimised.
  return(TRUE)
}

#' Internal check and update of settings related to data set parsing
#'
#' This function updates and checks parameters related to data set parsing based
#' on the available data set.
#'
#' @param formula User-provided formula, may be absent (`NULL`).
#' @param data Data set as loaded using the `.load_data` function.
#' @param settings List of parameter settings for data set parsing.
#' @inheritParams as_data_object
#'
#' @return A verified and updated list of parameter settings.
#' @md
#' @keywords internal
.update_initial_settings <- function(
    formula = NULL,
    data,
    settings,
    check_stringency = "strict"
) {
  # Outcome columns, outcome type, signature, included_features,
  # excluded_features, class_levels.

  if (!is.null(formula)) {
    
    # Parse information from the formula
    term_object <- stats::terms(formula, data = data)

    # Derive predictors
    predictor_vars <- .replace_illegal_column_name(attributes(term_object)$term.labels)
    all_vars <- .replace_illegal_column_name(all.vars(stats::terms(term_object)))
    outcome_col <- all_vars[!all_vars %in% predictor_vars]
    
    # Set outcome_col if it was previously unknown
    if (is.null(settings$data$outcome_col)) {
      settings$data$outcome_col <- outcome_col
    }
    
    # Ignore include_features and exclude_features when using the formula
    # interface
    if (!is.null(settings$data$include_features)) {
      ..warning(
        "The include_features parameter is ignored when using the formula interface.",
        warning_class = "input_argument_error"
      )
      settings$data$include_features <- NULL
    }
    
    if (!is.null(settings$data$exclude_features)) {
      ..warning(
        "The exclude_features parameter is ignored when using the formula interface.",
        warning_class = "input_argument_error"
      )
      settings$data$exclude_features <- NULL
    }
    
  } else {
    
    all_vars <- colnames(data)
    if (is.null(settings$data$outcome_col)) {
      predictor_vars <- all_vars
      
    } else {
      predictor_vars <- all_vars[!all_vars %in% settings$data$outcome_col]
    }
  }
  
  # Identifier columns----------------------------------------------------------
  if (!is.null(settings$data$sample_col) && check_stringency != "strict") {
    # Check if the internal default is used instead.
    if (
      !any(settings$data$sample_col %in% colnames(data)) &&
      get_id_columns(single_column = "sample") %in% colnames(data)
    ) {
      settings$data$sample_col <- get_id_columns(single_column = "sample")
    }
  }
  
  if (!is.null(settings$data$batch_col) && check_stringency != "strict") {
    # Check if the internal default is used instead.
    if (
      !any(settings$data$batch_col %in% colnames(data)) &&
      get_id_columns(single_column = "batch") %in% colnames(data)
    ) {
      settings$data$batch_col <- get_id_columns(single_column = "batch")
    }
  }
  
  if (!is.null(settings$data$series_col) && check_stringency != "strict") {
    # Check if the internal default is used instead.
    if (
      !any(settings$data$series_col %in% colnames(data)) &&
      get_id_columns(single_column = "series") %in% colnames(data)
    ) {
      settings$data$series_col <- get_id_columns(single_column = "series")
    }
  }
  
  ## Sample column -------------------------------------------------------------
  if (!is.null(settings$data$sample_col)) {
    # Check input
    .check_input_identifier_column(
      id_column = settings$data$sample_col,
      data = data,
      signature = settings$data$signature,
      exclude_features = settings$data$exclude_features,
      include_features = settings$data$include_features,
      other_id_column = c(settings$data$batch_col, settings$data$series_col),
      outcome_column = settings$data$outcome_col,
      col_type = "sample",
      check_stringency = check_stringency
    )
    
    # Remove sample column identifier if it doesn't appear in the dataset, and
    # autopopulate the column. Only possible when checks are not strict.
    if (
      !settings$data$sample_col %in% colnames(data) && check_stringency != "strict"
    ) {
      settings$data$sample_col <- NULL
    }
    
    # Remove the sample identifier column from the set of predictors
    predictor_vars <- setdiff(predictor_vars, settings$data$sample_col)
  }
  
  ## Batch column --------------------------------------------------------------
  if (!is.null(settings$data$batch_col)) {
    # Check input
    .check_input_identifier_column(
      id_column = settings$data$batch_col,
      data = data,
      signature = settings$data$signature,
      exclude_features = settings$data$exclude_features,
      include_features = settings$data$include_features,
      other_id_column = c(settings$data$sample_col, settings$data$series_col), 
      outcome_column = settings$data$outcome_col,
      col_type = "batch",
      check_stringency = check_stringency
    )
    
    # Remove batch column identifier if it doesn't appear in the dataset, and
    # autopopulate the column. Only possible when checks are not strict.
    if (
      !settings$data$batch_col %in% colnames(data) && check_stringency != "strict"
    ) {
      settings$data$batch_col <- NULL
    }
    
    # Remove the batch identifier column from the set of predictors
    predictor_vars <- setdiff(predictor_vars, settings$data$batch_col)
  }
  
  
  ## Series column -------------------------------------------------------------
  if (!is.null(settings$data$series_col)) {
    # Check input
    .check_input_identifier_column(
      id_column = settings$data$series_col,
      data = data,
      signature = settings$data$signature,
      exclude_features = settings$data$exclude_features,
      include_features = settings$data$include_features,
      other_id_column = c(settings$data$batch_col, settings$data$sample_col),
      outcome_column = settings$data$outcome_col,
      col_type = "series",
      check_stringency = check_stringency
    )
    
    # Remove series column identifier if it doesn't appear in the dataset, and
    # autopopulate the column. Only possible when checks are not strict.
    if (
      !settings$data$series_col %in% colnames(data) && check_stringency != "strict"
    ) {
      settings$data$series_col <- NULL
    }
    
    # Remove the series identifier column from the set of predictors
    predictor_vars <- setdiff(predictor_vars, settings$data$series_col)
  }
  
  # Outcome column -------------------------------------------------------------
  outcome_type <- "unset"
  if (!is.null(settings$data$outcome_type)) outcome_type <- settings$data$outcome_type
  
  if (outcome_type != "unset" && check_stringency != "strict") {
    # Check if outcome columns are present, or internal defaults are present.
    if (
      !any(settings$data$outcome_col %in% colnames(data)) &&
      all(get_outcome_columns(outcome_type) %in% colnames(data))
    ) {
      settings$data$outcome_col <- get_outcome_columns(settings$data$outcome_type)
    }
    
    # Set outcome column to NULL for unsupervised data, and assign to
    # exclude_features.
    if (outcome_type == "unsupervised") {
      if (!is.null(settings$data$outcome_col)) {
        ..warning(
          paste0(
            paste_s(settings$data$outcome_col), " was selected as an outcome column, but ",
            "unsupervised learners and novelty detectors do not require ",
            "outcome columns."
          ),
          warning_class = "dataset_error"
        )
        
        settings$data$exclude_features <- c(
          settings$data$exclude_features, settings$data$outcome_col
        )
        settings$data$outcome_col <- NULL
      }
    }
  }
  
  # Check for presence of outcome column
  if (
    is.null(settings$data$outcome_col) &&
    outcome_type != "unsupervised" &&
    check_stringency == "strict"
  ) {
    
    # Attempt to determine the outcome_col from the set difference of all
    # features and the union of signature, include_features and
    # exclude_features.
    outcome_col <- setdiff(
      predictor_vars,
      c(
        settings$data$exclude_features,
        settings$data$include_features,
        settings$data$signature,
        settings$data$novelty_features
      )
    )
    
    if (length(outcome_col) == 1L) {
      # Only select outcome column with stringent checks.
      ..warning(
        paste0(
          paste_s(outcome_col), "was selected as an outcome column. It is recommended ",
          "to provide the column name manually to avoid selecting the wrong column."
        ),
        warning_class = "dataset_error"
      )
      
      # Set the outcome column
      settings$data$outcome_col <- outcome_col
      
      # Update predictor_vars
      predictor_vars <- predictor_vars[!predictor_vars %in% outcome_col]
      
    } else {
      ..error(
        paste0(
          "No column(s) that determine the outcome were provided and none could be imputed. ",
          " Please provide outcome columns."
        ),
        error_class = "dataset_error"
      )
    }
  }
  
  if (!is.null(settings$data$outcome_col)) {
    # For strict checks outcome columns must be present. Check if outcome
    # columns are present in the data
    if (!all(settings$data$outcome_col %in% colnames(data))) {
      missing_col <- setdiff(settings$data$outcome_col, colnames(data))
      
      if (check_stringency %in% c("strict", "external_warn")) {
        stop_or_warn(
          paste0(
            "The outcome column ", paste_s(missing_col),
            " does not appear in the provided data set."
          ),
          check_stringency == "strict"
        )
      }
      
      # Set NULL and auto-generate outcome columns later.
      if (check_stringency != "strict") settings$data$outcome_col <- NULL
    }
  }
  
  if (length(settings$data$outcome_col) > 2L) {
    ..error(
      "Only one or two (in case of survival endpoints), may be specified",
      error_class = "dataset_error"
    )
  }
  
  
  # Outcome type ---------------------------------------------------------------
  
  # Attempt to impute outcome settings
  if (outcome_type == "unset") {
    settings$data$outcome_type <- .impute_outcome_type(
      data = data,
      outcome_column = settings$data$outcome_col,
      class_levels = settings$data$class_levels,
      censoring_indicator = settings$data$censoring_indicator,
      event_indicator = settings$data$event_indicator,
      competing_risk_indicator = settings$data$competing_risk_indicator
    )
  }
  
  # Check whether the outcome type fits the data.
  .check_outcome_type_plausibility(
    data = data,
    outcome_type = settings$data$outcome_type,
    outcome_column = settings$data$outcome_col,
    censoring_indicator = settings$data$censoring_indicator,
    event_indicator = settings$data$event_indicator,
    competing_risk_indicator = settings$data$competing_risk_indicator,
    check_stringency = check_stringency
  )
  
  # Class levels ---------------------------------------------------------------
  .check_class_level_plausibility(
    data = data,
    outcome_type = settings$data$outcome_type,
    outcome_column = settings$data$outcome_col,
    class_levels = settings$data$class_levels,
    check_stringency = check_stringency
  )
  
  # Set class levels in settings, if appropriate.
  if (
    is.null(settings$data$class_levels) &&
    settings$data$outcome_type %in% c("binomial", "multinomial") &&
    !is.null(settings$data$outcome_col)
  ) {
    
    if (is.factor(data[[settings$data$outcome_col]])) {
      settings$data$class_levels <- levels(data[[settings$data$outcome_col]])
      
    } else {
      settings$data$class_levels <- sort(unique_na(data[[settings$data$outcome_col]]))
    }
  }
  
  # Survival event indicator ---------------------------------------------------
  if (
    settings$data$outcome_type %in% c("survival", "competing_risk") &&
    !is.null(settings$data$outcome_col)
  ) {
    settings <- .impute_survival_indicators(
      data = data,
      outcome_type = settings$data$outcome_type,
      settings = settings,
      check_stringency = check_stringency
    )
  }
  
  # Features -------------------------------------------------------------------
  
  ## Signature features --------------------------------------------------------
  if (!is.null(settings$data$signature)) {
    # Check for overlap with exclude_features
    overlap_cols <- intersect(settings$data$signature, settings$data$exclude_features)
    
    if (length(overlap_cols) > 0L) {
      ..error(
        paste0(
          "One or more columns were provided that both appear in the signature and ",
          "among the features that should be removed. There can be no overlap. Found: ",
          paste_s(overlap_cols)
        ),
        error_class = "dataset_error"
      )
    }
    
    # Check if all features in the signature appear in the data
    missing_cols <- settings$data$signature[!settings$data$signature %in% predictor_vars]
    if (length(missing_cols) > 0L) {
      ..error(
        paste0(
          "One or more features assigned to the signature were not found in the data set:",
          paste_s(missing_cols)
        ),
        error_class = "dataset_error"
      )
    }
  }
  
  ## Novelty features ----------------------------------------------------------
  if (!is.null(settings$data$novelty_features)) {
    # Check for overlap with exclude_features
    overlap_cols <- intersect(settings$data$novelty_features, settings$data$exclude_features)
    
    if (length(overlap_cols) > 0L) {
      ..error(
        paste0(
          "One or more columns were provided that both appear in novelty_features and",
          "among the features that should be removed. There can be no overlap. Found:",
          paste_s(overlap_cols)
        ),
        error_class = "dataset_error"
      )
    }
    
    # Check if all features in novelty_features appear in the data
    missing_cols <- settings$data$novelty_features[!settings$data$novelty_features %in% predictor_vars]
    if (length(missing_cols) > 0L) {
      ..error(
        paste0(
          "One or more features assigned to novelty_features were not found in the data set:",
          paste_s(missing_cols)
        ),
        error_class = "dataset_error"
      )
    }
  }
  
  ## Excluded features ---------------------------------------------------------
  if (!is.null(settings$data$exclude_features)) {
    
    # Check if all features marked for exclusion appear in the data
    missing_cols <- settings$data$exclude_features[!settings$data$exclude_features %in% predictor_vars]
    if (length(missing_cols) > 0L) {
      ..error(
        paste0(
          "One or more features marked for exclusion were not found in the data set:",
          paste_s(missing_cols)
        ),
        error_class = "dataset_error"
      )
    }
    
    # Check if include_features also exists.
    overlap_cols <- intersect(settings$data$include_features, settings$data$exclude_features)
    
    if (length(overlap_cols) > 0L) {
      ..error(
        paste0(
          "One or more columns were provided that appear in both the set of features ",
          "marked for exclusion and for inclusion. There can be no overlap. Found:",
          paste_s(overlap_cols)
        ),
        error_class = "dataset_error"
      )
    }
    
    # Notify the used that include_features parameter takes precedence.
    if (!is.null(settings$data$include_features)) {
      ..warning(
        paste0(
          "Features marked for inclusion take precedence over the features marked for inclusion, i.e. ",
          "the final data set will only contain those features marked for inclusion."
        ),
        warning_class = "dataset_error"
      )
      
      settings$data$exclude_features <- NULL
    }
  }
  
  ## Included features ---------------------------------------------------------
  if (!is.null(settings$data$include_features)) {
    
    missing_cols <- settings$data$include_features[!settings$data$include_features %in% predictor_vars]
    
    if (length(missing_cols) > 0L) {
      ..error(
        paste0(
          "One or more features marked for inclusion were not found in the data set:",
          paste_s(missing_cols)
        ),
        error_class = "dataset_error"
      )
    }
  }
  
  # Determine which features should go to include_features.
  if (!is.null(settings$data$exclude_features)) {
    # Select everything but the features marked for exclusion
    settings$data$include_features <- setdiff(
      predictor_vars,
      settings$data$exclude_features
    )
    
  } else if (!is.null(settings$data$include_features)) {
    # Select features marked for inclusion and signature, in so far as these do not overlap.
    settings$data$include_features <- unique(c(
      settings$data$include_features,
      settings$data$signature,
      settings$data$novelty_features
    ))
    
  } else {
    # Select all available predictor variables.
    settings$data$include_features <- predictor_vars
  }
  
  # Outcome name ---------------------------------------------------------------
  if (is.null(settings$data$outcome_name) &&
      settings$data$outcome_type %in% c("survival", "competing_risk")) {
    if (!is.null(settings$data$outcome_col)) {
      # Use the name of the event column, because its usually more indicative.
      settings$data$outcome_name <- tolower(settings$data$outcome_col[2L])
      
    } else {
      settings$data$outcome_name <- "survival"
    }
    
  } else if (is.null(settings$data$outcome_name)) {
    if (!is.null(settings$data$outcome_col)) {
      settings$data$outcome_name <- tolower(settings$data$outcome_col)
      
    } else {
      settings$data$outcome_name <- "outcome"
    }
  }
  
  return(settings)
}



#' Internal function to check batch assignment to development and validation
#'
#' This function checks which batches in the data set are assigned to model
#' development and external validation. Several errors may be raised if there
#' are inconsistencies such as an overlapping assignment, name mismatches etc.
#'
#' @param section_table data.table generated by the `extract_experimental_setup`
#'  function. Contains information regarding the experiment.
#' @param data Data set as loaded using the `.load_data` function.
#' @param settings List of parameter settings for data set parsing and setting
#'  up the experiment.
#'
#' @return A verified and updated list of parameter settings.
#' @md
#' @keywords internal
.update_experimental_design_settings <- function(
    section_table,
    data,
    settings,
    verbose = FALSE
) {

  # Find out if any external validation is performed as part of the workflow
  if (is.waive(section_table)) {
    # If the section table is waived, this means that iterations come from a
    # user-provided different file.
    return(settings)
    
  } else {
    perform_external_validation <- any(section_table$external_validation)
  }
  
  # Determine the available batch identifiers
  available_batch_ids <- unique(data[[get_id_columns(single_column = "batch")]])
  
  # Determine what happens if batch identifiers are not specified for both
  # development and validation.
  if (
    is.null(union(settings$data$train_cohorts, settings$data$valid_cohorts)) &&
    !perform_external_validation
  ) {
    # Use all cohorts for training if nothing is specifically provided, and
    # external validation is not necessary.
    settings$data$train_cohorts <- available_batch_ids
    
  } else if (
    is.null(union(settings$data$train_cohorts, settings$data$valid_cohorts)) &&
    perform_external_validation
  ) {
    # Validation cohort(s) should be provided, or be identifiable
    ..error(
      "The validation_batch_id variable should be set to perform external validation.",
      error_class = "dataset_error"
    )
  } 
  
  # Check if some cohorts are available for external validation, if required.
  if (
    all(available_batch_ids %in% settings$data$train_cohorts) &&
    perform_external_validation
  ) {
    # All cohorts are found in the set used for development. Prompt the user
    # to assign some cohorts to validation.
    ..error(
      paste0(
        "All batches/cohorts in the data set are assigned for model development. ",
        "External validation is not possible. ",
        "Please assign one or more batches/cohorts to external validation by ", 
        "passing their names as an argument to the validation_batch_id variable."
      ),
      error_class = "dataset_error"
    )
  }
  
  # Check if one or more cohorts are available for development.
  if (all(available_batch_ids %in% settings$data$valid_cohorts)) {
    # All cohorts are found in the set used for external validation. Prompt the
    # user to assign some development cohorts.
    ..error(
      paste0(
        "All batches/cohorts in the data set are assigned for external validation. ",
        "Model development is not possible. ",
        "Please assign one or more batches/cohorts to model development by ",
        "passing their names as an argument to the development_batch_id variable."
      ),
      error_class = "dataset_error"
    )
  }
  
  # Check whether there is an overlap between development and validation group
  # identifiers.
  overlapping_batches <- intersect(
    settings$data$train_cohorts,
    settings$data$valid_cohorts
  )
    
  if (length(overlapping_batches) > 0L) {
    ..error(
      paste0(
        "One or more batch/cohort names occur in both the development_batch_id ",
        "and validation_batch_id variables: ",
        paste_s(overlapping_batches)
      ),
      error_class = "dataset_error"
    )
  }
  
  # Check whether all specified development group identifiers actually appear in
  # the data.
  missing_batches <- setdiff(settings$data$train_cohorts, available_batch_ids)
    
  if (length(missing_batches) > 0L) {
    ..error(
      paste0(
        "One or more batch/cohort names specified in the development_batch_id ",
        "variable could not be found in the data: ",
        paste_s(missing_batches)
      ),
      error_class = "dataset_error"
    )
  }
  
  # Determine if there are any missing batches for validation cohorts
  missing_batches <- setdiff(settings$data$valid_cohorts, available_batch_ids)
  
  if (length(missing_batches) > 0L && perform_external_validation) {
    # Raise an error if external validation is expected.
    ..error(
      paste0(
        "One or more batch/cohort names specified in the validation_batch_id ",
        "variable could not be found in the data: ",
        paste_s(missing_batches)
      ),
      error_class = "dataset_error"
    )
    
  } else if (length(missing_batches) > 0L) {
    # Filter out the missing cohort names when external validation is not
    # performed.
    settings$data$valid_cohorts <- setdiff(settings$data$valid_cohorts, missing_batches)
    
    # If the remaining number of validation cohorts is zero, set valid_cohorts
    # to NULL
    if (length(settings$data$valid_cohorts) == 0L) settings$data$valid_cohorts <- NULL
  }
  
  # Infer validation cohorts.
  if (
    !is.null(settings$data$train_cohorts) &&
    is.null(settings$data$valid_cohorts)
  ) {
    # Identify batch ids that appear in the data set but are not included for
    # development. These are then assigned for validation.
    new_validation_batch_id <- setdiff(
      available_batch_ids,
      settings$data$train_cohorts
    )
    
    if (length(new_validation_batch_id) > 0L && perform_external_validation) {
      if (verbose) {
        message(paste0(
          "One or more batches/cohorts were not used for development and are now ",
          "used for external validation: ",
          paste_s(new_validation_batch_id)
        ))
      }
      
      settings$data$valid_cohorts <- new_validation_batch_id
      
    } else if (length(new_validation_batch_id)) {
      if (verbose) {
        message(paste0(
          "One or more batches/cohorts in the data are not used for development because they ",
          "were not provided in the development_batch_id variable."
        ))
      }
    }
  }
  
  # Infer training cohorts
  if (
    is.null(settings$data$train_cohorts) && !is.null(settings$data$valid_cohorts)
  ) {
    # Identify batch ids that appear in the data set but are not included for
    # validation. These are then assigned for development.
    new_development_batch_id <- setdiff(
      available_batch_ids,
      settings$data$valid_cohorts
    )
    
    if (length(new_development_batch_id) > 0L) {
      if (verbose) {
        message(paste0(
          "One or more batches/cohorts were not used for external validation and are now ",
          "used for development: ",
          paste_s(new_development_batch_id)
        ))
      }
      
      settings$data$train_cohorts <- new_development_batch_id
    }
  }
  
  # Determine if there are any unused groups
  unused_batch_id <- setdiff(
    available_batch_ids,
    union(
      settings$data$train_cohorts,
      settings$data$valid_cohorts
    )
  )
  
  if (length(unused_batch_id) > 0L) {
    if (verbose) {
      message(paste0(
        "One or more batches/cohorts are not used for development or external validation: ",
        paste_s(unused_batch_id)
      ))
    }
  }
  
  return(settings)
}



#' Internal imputation function for the outcome type.
#'
#' This function allows for imputation of the most plausible outcome type.
#' This imputation is only done for trivial cases, where there is little doubt.
#' As a consequence `continuous` outcome type is never imputed.
#'
#' @param data Data set as loaded using the `.load_data` function.
#' @param outcome_column Name of the outcome column in the data set. 
#' @param class_levels User-provided class levels for the outcome.
#' @param censoring_indicator Name of censoring indicator.
#' @param event_indicator Name of event indicator.
#' @param competing_risk_indicator Name of competing risk indicator.
#' 
#' @note It is highly recommended that the user provides the outcome type.
#'
#' @return The imputed outcome type.
#' @md
#' @keywords internal
.impute_outcome_type <- function(
    data,
    outcome_column,
    class_levels,
    censoring_indicator,
    event_indicator,
    competing_risk_indicator
) {

  if (length(outcome_column) > 2L) {
    ..error(
      paste0(
        "Only one or two (in case of survival endpoints) outcome columns are expected.",
        "However ", length(outcome_column), " were found."
      ),
      error_class = "dataset_error"
    )
  }
  
  # Test for survival
  if (length(outcome_column) == 2L) {
    # One column should contain continuous data >= 0.0 (time column). One column
    # should contain numerical or logical data with 0, 1 (event status column)
    
    # Check for presence of an event column
    event_survival_cols <- sapply(
      outcome_column,
      .is_survival_status_col, 
      data = data,
      censoring_indicator = censoring_indicator,
      event_indicator = event_indicator,
      competing_risk_indicator = NULL
    )
    
    event_competing_risk_cols <- sapply(
      outcome_column,
      .is_survival_status_col,
      data = data,
      censoring_indicator = censoring_indicator,
      event_indicator = event_indicator,
      competing_risk_indicator = competing_risk_indicator
    )
    
    # Check for presence of a time column
    time_cols <- sapply(
      outcome_column,
      .is_survival_time_col,
      data = data
    )
    
    if (
      !any(time_cols) ||
      (!any(event_survival_cols) && !any(event_competing_risk_cols))
    ) {
      # Check if there is at least one event_col and one time_col, and in case
      # there is only one event_col and one time_col, that these are not the
      # same.
      ..error(
        paste0(
          "Survival or competing risk outcomes were expected as two outcome columns were provided. ",
          "However, data in the columns did not strictly correspond to these outcomes. ",
          "Check if a survival outcome was intended, and provide a single outcome column otherwise. ",
          "Columns for survival or competing risk outcomes should contain time (numeric values >= 0) and event status ",
          "(0 and 1 or FALSE and TRUE, or values according to censoring_indicator, ",
          "event_indicator and competing_risk_indicator arguments) information."
        ),
        error_class = "dataset_error"
      )
      
    } else if (
      sum(time_cols) == 1L &&
      sum(event_survival_cols) == 1L &&
      sum(event_competing_risk_cols) == 1L
    ) {
      # Only one column of either type. The test below yields 0 if these are
      # different columns, and 1 if they are the same column, in which case an
      # error is raised.
      
      if (
        sum(time_cols * event_survival_cols) &&
        sum(time_cols * event_competing_risk_cols)
      ) {
        ..error(
          paste0(
            "Survival or competing risk outcome was expected as two outcome columns were provided. ",
            "However, data in the columns did not strictly correspond to these outcomes. ",
            "Check if a survival outcome was intended, and provide a single outcome column otherwise. ",
            "Columns for survival or competing risk outcomes should contain time ",
            "(numeric values >= 0) and event status ",
            "(0 and 1 or FALSE and TRUE, censoring_indicator, event_indicator ",
            "and competing_risk_indicator arguments) information."
          ),
          error_class = "dataset_error"
        )
      }
    }
    
    if (is.null(competing_risk_indicator)) {
      # Set outcome_type to survival as it is plausible
      message(paste0(
        "A survival outcome was imputed based on the data. If this is an incorrect type, ",
        "please provide an outcome_type manually."
      ))
      
      return("survival")
      
    } else {
      # Set outcome_type to competing_risk as it is plausible
      message(paste0(
        "A competing_risk outcome was imputed based on the data. If this is an incorrect type,",
        "please provide an outcome_type manually."
      ))
      
      return("competing_risk")
    }
  }
  
  if (length(outcome_column) == 1L) {
    # Extract data
    x <- data[[outcome_column]]
    
    # Test for binomial and multinomial outcomes based on class_levels
    if (!is.null(class_levels)) {
      if (all(x %in% class_levels, na.rm = TRUE)) {
        if (length(class_levels) == 2L) {
          message(paste0(
            "A binomial outcome was imputed based on the data. If this is an incorrect type, ",
            "please provide an outcome_type manually."
          ))
          
          return("binomial")
          
        } else if (length(class_levels) > 2L) {
          message(paste0(
            "A multinomial outcome was imputed based on the data. If this is an incorrect type, ",
            "please provide an outcome_type manually."
          ))
          
          return("multinomial")
        }
      }
    }
    
    # Test for categorical outcomes based on class of the data in the outcome
    # column, in this case factor.
    if (is.factor(x)) {
      
      if (nlevels(x) == 2L) {
        message(paste0(
          "A binomial outcome was imputed based on the data. If this is an incorrect type, ",
          "please provide an outcome_type manually."
        ))
        
        return("binomial")
        
      } else if (nlevels(x) > 2L) {
        message(paste0(
          "A multinomial outcome was imputed based on the data. If this is an incorrect type,",
          "please provide an outcome_type manually."
        ))
        
        return("multinomial")
      }
    }
    
    # Test for categorical outcomes based on class of the data in the outcome
    # column, in this case logical.
    if (is.logical(x)) {
      message(paste0(
        "A binomial outcome was imputed based on the data. If this is an incorrect type,",
        "please provide an outcome_type manually."
      ))
      
      return("binomial")
    }
    
    # Test for categorical outcomes based on class of the data in the outcome
    # column, in this case character.
    if (is.character(x)) {
      if (data.table::uniqueN(x, na.rm = TRUE) == 2L) {
        message(paste0(
          "A binomial outcome was imputed based on the data. If this is an incorrect type, ",
          "please provide an outcome_type manually."
        ))
        
        return("binomial")
        
      } else if (data.table::uniqueN(x, na.rm = TRUE) > 2L) {
        message(paste0(
          "A multinomial outcome was imputed based on the data. If this is an incorrect type, ",
          "please provide an outcome_type manually."
        ))
        
        return("multinomial")
      }
    }
    
    # Tests for continuous type data are not really possible. One could test on
    # is.numeric, and minimum value, but these data could still represent
    # binomial (e.g. 0s and 1s) or multinomial data. One may devise some tests
    # by counting the number of samples with a unique value, but this is
    # dangerous for smaller data sets.
    ..error(
      "Imputation of the outcome type was not possible. Please provide an outcome type manually.",
      error_class = "dataset_error"
    )
  }
}



.impute_survival_indicators <- function(
    data,
    outcome_type,
    settings,
    check_stringency = "strict"
) {
  
  # Skip checks if check stringency is external: no checks on outcome columns
  # are required.
  if (check_stringency == "external") return(settings)
  
  # Define standard indicators for censoring, event and competing risk.
  standard_censoring_indicator <- .get_available_default_censoring_indicator()
  if (!is.null(settings$data$censoring_indicator)) {
    standard_censoring_indicator <- settings$data$censoring_indicator
  }
  
  standard_event_indicator <- .get_available_default_event_indicator()
  if (!is.null(settings$data$event_indicator)) {
    standard_event_indicator <- settings$data$event_indicator
  }
  
  standard_competing_risk_indicator <- NULL
  if (!is.null(settings$data$competing_risk_indicator)) {
    standard_competing_risk_indicator <- settings$data$competing_risk_indicator
  }
  
  # Combine indicators
  all_indicators <- c(
    standard_censoring_indicator,
    standard_event_indicator,
    standard_competing_risk_indicator
  )
  
  # Find the number of matches with each of the two outcome columns.
  n_matches <- sapply(
    settings$data$outcome_col,
    function(column, data, all_indicators) {
      x <- data[[column]]
      x <- tolower(as.character(x[!is.na(x)]))
      
      return(sum(x %in% tolower(all_indicators)))
    },
    data = data,
    all_indicators = all_indicators
  )
  
  # For non-strict test, we assume that that outcome columns are already
  # organised as time, event, whereas for strict tests we need to identify it de
  # novo.
  if (check_stringency == "strict") {
    # Identify the column that is most likely to be the event column.
    event_column <- settings$data$outcome_col[which.max(n_matches)]
    
  } else {
    event_column <- settings$data$outcome_col[2L]
  }
  
  # Select unique values in the event column.
  event_values <- data[[event_column]]
  event_values <- unique_na(event_values)
  
  # Try to identify indicators present in the dataset.
  present_censoring_indicator <- event_values[
    tolower(as.character(event_values)) %in% tolower(standard_censoring_indicator)
  ]
  present_event_indicator <- event_values[
    tolower(as.character(event_values)) %in% tolower(standard_event_indicator)
  ]
  present_competing_risk_indicator <- event_values[
    tolower(as.character(event_values)) %in% tolower(standard_competing_risk_indicator)
  ]
  
  # Identify values in the event column which are not yet assigned.
  event_values <- setdiff(
    event_values,
    c(
      present_censoring_indicator, 
      present_event_indicator, 
      present_competing_risk_indicator
    )
  )
  
  # Check the presence of user-provided indicators.
  if (!is.null(settings$data$event_indicator) && length(event_values) > 0L) {
    
    # An event indicator should always be present.
    if (
      !all(tolower(as.character(settings$data$event_indicator)) %in% tolower(as.character(event_values))) &&
      check_stringency == "strict"
    ) {
      ..error(
        paste0(
          "The following provided event indicator(s) were not found in the data: ",
          paste_s(setdiff(
            tolower(as.character(settings$data$event_indicator)),
            tolower(as.character(event_values))
          )),
          " . Please check for spelling errors."
        ),
        error_class = "dataset_error"
      )
    }
  }
  
  if (!is.null(settings$data$censoring_indicator) && length(event_values) > 0L) {
    # An censoring indicator is not required to be present, but if the user
    # provides one, and it isn't found in the dataset, raise an error.
    if (
      !all(tolower(as.character(settings$data$censoring_indicator)) %in% tolower(as.character(event_values))) &&
      check_stringency == "strict"
    ) {
      ..error(
        paste0(
          "The following provided censoring indicator(s) were not found in the data: ",
          paste_s(setdiff(
            tolower(as.character(settings$data$censoring_indicator)),
            tolower(as.character(event_values))
          ))
        ),
        error_class = "dataset_error"
      )
    }
  }
  
  if (!is.null(settings$data$competing_risk_indicator) && length(event_values) > 0L) {
    # Competing risk indicators are only present in competing_risk outcomes.
    if (outcome_type == "survival" && check_stringency == "strict") {
      ..error(
        paste0(
          "One or indicators for competing risks were specified. ",
          "However, the outcome type was set to survival, which does not check competing risks. ",
          "Please change outcome_type to competing_risk, or remove the competing risk indicators."
        ),
        error_class = "dataset_error"
      )
    }
    
    if (
      !all(tolower(as.character(settings$data$competing_risk_indicator)) %in% tolower(as.character(event_values))) &&
      check_stringency == "strict"
    ) {
      ..error(
        paste0(
          "The following provided competing risk indicator(s) were not found in the data: ",
          paste_s(setdiff(
            tolower(as.character(settings$data$competing_risk_indicator)),
            tolower(as.character(event_values))
          ))
        ),
        error_class = "dataset_error"
      )
    }
  }

  if (length(event_values) > 0L) {
    # Event indicator can only be determined if there is one value left over.
    if (length(present_event_indicator) == 0L) {
      if (length(event_values) == 1L) {
        
        # Assign the remaining value to the event indicator.
        present_event_indicator <- event_values
        
        # Make event_values empty.
        event_values <- setdiff(event_values, event_values)
        
        if (check_stringency %in% c("strict", "external_warn")) {
          # Throw a warning, because the assumption may be false.
          ..warning(
            paste0(
              "The event indicator for survival/competing risk outcomes was neither ",
              "a standard value, nor provided. Based on the values found in ", event_column,
              ", ", present_event_indicator, " was selected as an event indicator."
            ),
            error_class = "dataset_error"
          )
        }
        
      } else if (check_stringency %in% c("strict", "external_warn")) {
        stop_or_warn(
          paste0(
            "The event indicator for survival/competing risk outcomes was neither ",
            "a standard value, nor provided. More than one unassigned value was found in ",
            event_column, ", preventing automatic assignment: ",
            paste_s(event_values)
          ),
          as_error = check_stringency == "strict"
        )
      }
    }
  }
  
  if (length(event_values) > 0L && outcome_type == "survival") {
    if (length(present_censoring_indicator) == 0L) {
      
      # Assign all remaining values.
      present_censoring_indicator <- event_values
      
      # Make event_values empty.
      event_values <- setdiff(event_values, event_values)
      
      if (check_stringency %in% c("strict", "external_warn")) {
        # Throw a warning, because the assumption may be false.
        ..warning(
          paste0(
            "The censoring indicator for the survival outcome was neither a standard value, nor provided. ",
            "Based on the values found in ", event_column, ", ",
            paste_s(present_censoring_indicator),
            ifelse(length(present_censoring_indicator) == 1L, " was", " were"),
            " selected as censoring indicator."
          ),
          warning_class = "dataset_error"
        )
      }
      
    } else if (check_stringency %in% c("strict", "external_warn")) {
      stop_or_warn(
        paste0(
          "One or more unassigned values for the survival outcome were found in ", event_column, ": ",
          paste_s(event_values),
          ". Please assign manually."
        ),
        as_error = check_stringency == "strict"
      )
    }
  }
  
  if (length(event_values) > 0L && outcome_type == "competing_risk") {
    if (
      length(present_censoring_indicator) == 0L && length(present_competing_risk_indicator) == 0L
    ) {
      if (check_stringency %in% c("strict", "external_warn")) {
        stop_or_warn(
          paste0(
            "One or more unassigned values for the survival outcome were found in ", event_column, ": ",
            paste_s(event_values), ". These could not be assigned automatically as indicators for ",
            "censoring and competing risks were both missing. Please assign manually."
          ),
          as_error = check_stringency == "strict"
        )
      }
      
    } else if (length(present_censoring_indicator) == 0L) {
      
      # Assign all remaining values.
      present_censoring_indicator <- event_values
      
      # Make event_values empty.
      event_values <- setdiff(event_values, event_values)
      
      if (check_stringency %in% c("strict", "external_warn")) {
        # Throw a warning, because the assumption may be false.
        ..warning(
          paste0(
            "The censoring indicator for the competing risk outcome was neither a standard value, nor provided. ",
            "Based on the values found in ", event_column, ", ",
            paste_s(present_censoring_indicator),
            ifelse(length(present_censoring_indicator) == 1L, " was", " were"),
            " selected as censoring indicator."
          ),
          warning_class = "dataset_error"
        )
      }
      
    } else if (length(present_competing_risk_indicator) == 0L) {
      
      # Assign all remaining values.
      present_competing_risk_indicator <- event_values
      
      # Make event_values empty.
      event_values <- setdiff(event_values, event_values)
      
      if (check_stringency %in% c("strict", "external_warn")) {
        # Throw a warning, because the assumption may be false.
        ..warning(
          paste0(
            "The competing risk indicator for the competing risk outcome was not provided. ",
            "Based on the values found in ", event_column, ", ",
            paste_s(present_competing_risk_indicator),
            ifelse(length(present_competing_risk_indicator) == 1L, " was", " were"),
            " selected as competing risk indicator."
          ),
          warning_class = "dataset_error"
        )
      }
      
    } else {
      if (check_stringency %in% c("strict", "external_warn")) {
        stop_or_warn(
          paste0(
            "One or more unassigned values for the competing risk outcome were found in ",
            event_column, ": ", paste_s(event_values), 
            ". Please assign manually to the indicators."
          ),
          as_error = check_stringency == "strict"
        )
      }
    }
  }
  
  # Update censoring indicator in settings.
  if (length(present_censoring_indicator) > 0L) {
    settings$data$censoring_indicator <- present_censoring_indicator
    
  } else {
    settings$data$censoring_indicator <- NULL
  } 
  
  # Update event indicator in settings.
  if (length(present_event_indicator) > 0L) {
    settings$data$event_indicator <- present_event_indicator 
    
  } else {
    settings$data$event_indicator <- NULL
  } 
  
  if (length(present_competing_risk_indicator) > 0L) {
    settings$data$competing_risk_indicator <- present_competing_risk_indicator
    
  } else {
    settings$data$competing_risk_indicator <- NULL
  }
  
  # Make sure that columns are organised as time, event. Only do this for strict
  # tests.
  if (check_stringency == "strict") {
    time_column <- setdiff(settings$data$outcome_col, event_column)
    settings$data$outcome_col <- c(time_column, event_column)
  }
  
  return(settings)
}



#' Internal function for checking if the outcome type fits well to the data
#'
#' This function may help identify if the outcome type is plausible
#' given the outcome data. In practice it also tests whether the outcome column
#' is actually correct given the outcome type.
#'
#' @param data Data set as loaded using the `.load_data` function.
#' @param outcome_type Character string indicating the type of outcome being
#'  assessed.
#' @param outcome_column Name of the outcome column in the data set.
#' @param censoring_indicator Name of censoring indicator.
#' @param event_indicator Name of event indicator.
#' @param competing_risk_indicator Name of competing risk indicator.
#' @inheritParams as_data_object
#' 
#' @return NULL
#' @md
#' @keywords internal
.check_outcome_type_plausibility <- function(
    data,
    outcome_type,
    outcome_column,
    censoring_indicator,
    event_indicator,
    competing_risk_indicator,
    check_stringency = "strict"
) {
  
  # Checks plausibility of the outcome type and identifies any errors 
  
  # Check if only a single outcome_type is specified.
  if (length(outcome_type) != 1L) {
    ..error(
      paste0("A single outcome type should be provided. Found: ", length(outcome_type)),
      error_class = "input_argument_error"
    )
  }
  
  # Check for the count outcome type -- this type of outcome was deprecated in
  # version 2.0.0.
  if (outcome_type == "count") {
    ..deprecation_count(as_error = TRUE)
  }
  
  # Check if the provided outcome_type is a valid choice.
  if (!outcome_type %in% c(
    "binomial", "multinomial", "continuous",
    "survival", "competing_risk", "unsupervised"
  )) {
    ..error_no_known_outcome_type(outcome_type)
  }
  
  # Check if two columns are provided for survival data
  if (outcome_type %in% c("survival", "competing_risk")) {
    
    if (check_stringency != "strict" && !is.null(outcome_column)) {
      # For non-strict checks only check in case any outcome columns are
      # present.
      if (length(outcome_column) != 2L) {
        ..error(
          paste0(
            "Two outcome columns are expected for data with survival outcomes. Found: ",
            length(outcome_column)
          ),
          error_class = "dataset_error"
        )
      }
      
    } else if (check_stringency == "strict") {
      # For strict checks always check.
      if (length(outcome_column) != 2L) {
        ..error(
          paste0(
            "Two outcome columns are expected for data with survival outcomes. Found: ",
            length(outcome_column)
          ),
          error_class = "dataset_error"
        )
      }
    }
  }
    
  # Check if one column is provided for binomial, multinomial, and continuous
  # outcome types.
  if (outcome_type %in% c("binomial", "multinomial", "continuous")) {
    
    if (check_stringency != "strict" && !is.null(outcome_column)) {
      # For non-strict checks only check in case any outcome columns are
      # present.
      if (length(outcome_column) != 1L) {
        ..error(
          paste0(
            "One outcome column is expected for data with ", outcome_type, " outcomes. ",
            "Found: ", length(outcome_column)
          ),
          error_class = "dataset_error"
        )
      }
      
    } else if (check_stringency == "strict") {
      # For strict checks always check.
      if (length(outcome_column) != 1L) {
        ..error(
          paste0(
            "One outcome column is expected for data with ", outcome_type, " outcomes. ",
            "Found: ", length(outcome_column)
          ),
          error_class = "dataset_error"
        )
      }
    }
  }
  
  # Check that no outcome columns are provided for unsupervised outcome types.
  if (outcome_type %in% "unsupervised") {
    
    if (check_stringency == "strict") {
      # For strict checks always check if outcome columns are present.
      if (length(outcome_column) > 0L) {
        ..error(
          paste0(
            "No outcome columns are expected for unsupervised analysis. Found: ",
            paste_s(outcome_column)
          ),
          error_class = "dataset_error"
        )
      }
    }
  }
  
  # Check for outcome columns with NA values.
  outcome_na <- TRUE
  if (!is.null(outcome_column)) {
    outcome_na <- any(sapply(
      outcome_column,
      function(ii, data) (!any(is_valid_data(data[[ii]]))),
      data = data
    ))
  }
  
  if (
    check_stringency == "strict" && outcome_na && outcome_type != "unsupervised"
  ) {
    # Under strict conditions, outcome data should be present.
    ..error("Outcome column(s) do not contain any data.", error_class = "dataset_error")
    
  } else if (
    check_stringency == "external_warn" && outcome_na && outcome_type != "unsupervised"
  ) {
    # Under less strict conditions, outcome data may be absent, but the user
    # should be warned.
    ..warning(
      "Outcome column(s) do not contain any data. Some evaluation steps may fail to produce results.",
      warning_class = "dataset_error"
    )
  }
  
  # Plausibility checks for binomial outcome type
  if (outcome_type == "binomial" && !outcome_na) {
    
    # Check which classes are present.
    classes_present <- unique_na(data[[outcome_column]])
    
    if (check_stringency == "strict") {
      if (length(classes_present) > 2L) {
        ..error(
          paste0(
            "More than two classes (", paste_s(classes_present),
            ") were found in the outcome column: ", outcome_column, ". ",
            "Exactly two classes are expected for the binomial outcome type. ",
            "Specify outcome_type=\"multinomial\" if this is intentional."
          ),
          error_class = "dataset_error"
        )
        
      } else if (length(classes_present) < 2L) {
        ..error(
          paste0(
            "Fewer than two classes were found in the outcome column: ", outcome_column,
            ". Exactly two classes are expected for the binomial outcome type."
          ),
          error_class = "dataset_error"
        )
      }
      
    } else if (check_stringency == "external_warn") {
      if (length(classes_present) > 2L) {
        ..error(
          paste0(
            "More than two classes (", paste_s(classes_present),
            ") were found in the outcome column: ", outcome_column,
            ". Exactly two classes are expected for the binomial outcome type."
          ),
          error_class = "dataset_error"
        )
      }
    }
  } 
  
  # Plausibility check for the multinomial outcome type
  if (outcome_type == "multinomial" && !outcome_na) {
    # Check which classes are present.
    classes_present <- unique_na(data[[outcome_column]])
    
    if (check_stringency == "strict") {
      if (length(classes_present) < 2L) {
        ..error(
          paste0(
            "Fewer than two classes were found in the outcome column:", outcome_column,
            ". Two or more classes are expected for the multinomial outcome type."
          ),
          error_class = "dataset_error"
        )
      }
    }
  }
  
  # Plausibility check for the continuous outcome type
  if (outcome_type == "continuous" && !outcome_na) {
    
    if (check_stringency %in% c("strict", "external_warn")) {
      if (!is.numeric(data[[outcome_column]])) {
        ..error(
          paste0(
            "The outcome column (", outcome_column, ") does not contain numeric data. ",
            "Numeric data are expected for the continuous outcome type."
          ),
          error_class = "dataset_error"
        )
      }
    }
  }
  
  # Plausibility check for the survival outcome type
  if (outcome_type %in% c("survival", "competing_risk") && !outcome_na) {
    # Check for presence of an event column and time columns
    
    if (check_stringency == "strict") {
      if (outcome_type == "survival") {
        # The competing risk indicator should be absent for survival outcomes.
        event_cols <- sapply(
          outcome_column,
          .is_survival_status_col,
          data = data,
          censoring_indicator = censoring_indicator,
          event_indicator = event_indicator,
          competing_risk_indicator = NULL
        )
        
      } else {
        event_cols <- sapply(
          outcome_column,
          .is_survival_status_col,
          data = data,
          censoring_indicator = censoring_indicator,
          event_indicator = event_indicator,
          competing_risk_indicator = competing_risk_indicator
        )
      }
      
      time_cols <- sapply(
        outcome_column,
        .is_survival_time_col,
        data = data
      )
      
      # Check if there is at least one event_col and one time_col, and in case
      # there is only one event_col and one time_col, that these are not the
      # same.
      if (!any(event_cols)) {
        ..error(
          paste0(
            "None of the outcome columns (", paste_s(outcome_column),
            ") contain event status information. This column may only contain values 0 and 1 or ",
            "FALSE and TRUE, or the value indicated by ",
            ifelse(
              outcome_type == "survival",
              "censoring_indicator and event_indicator",
              "censoring_indicator, event_indicator and competing_risk_indicator"
            ),
            " arguments."
          ),
          error_class = "dataset_error"
        )
        
      } else if (all(event_cols)) {
        ..error(
          paste0(
            "Both outcome columns (", paste_s(outcome_column),
            ") seem to contain event status information. One column with survival times should ",
            "be provided together with one column with survival event status information."
          ),
          error_class = "dataset_error"
        )
        
      } else if (!any(time_cols)) {
        ..error(
          paste0(
            "None of the outcome columns (", paste_s(outcome_column),
            ") contain survival time information. This column may only contain numeric values ",
            "greater or equal to 0."
          ),
          error_class = "dataset_error"
        )
        
      } else if (sum(time_cols) == 1L && sum(event_cols) == 1L) {
        # Only one column of either type. The test below yields 0 if these are
        # different columns, and 1 if they are the same column, in which case an
        # error is raised.
        if (sum(time_cols * event_cols)) {
          ..error(
            paste0(
              "None of the outcome columns (", paste_s(outcome_column),
              ") contain survival time information. This column may only contain numeric values ",
              "greater or equal to 0, and may not contain only 0s and 1s to avoid misinterpration ",
              "as event status information."
            ),
            error_class = "dataset_error"
          )
        }
      }
    }
  }
  
  return(invisible(TRUE))
}



#' Internal function for checking consistency of the identifier columns
#' 
#' This function checks whether an identifier column is consistent, i.e. appears
#' it exists, there is only one, and there is no overlap with any user-provided
#' feature columns, identifiers, or 
#'
#' @param id_column Character string indicating the currently inspected
#'  identifier column.
#' @param data Data set as loaded using the `.load_data` function.
#' @param col_type Character string indicating the type of column, i.e. `sample`
#'  or `batch`.
#' @param other_id_column Character string indicating another identifier column.
#' @param outcome_column Character string indicating the outcome column(s).
#' @inheritParams .parse_experiment_settings
#' @inheritParams as_data_object
#' 
#' @return NULL
#' @md
#' @keywords internal
.check_input_identifier_column <- function(
    id_column,
    data,
    signature = NULL,
    exclude_features = NULL,
    include_features = NULL,
    other_id_column = NULL,
    outcome_column = NULL,
    col_type,
    check_stringency = "strict"
) {
  
  # Check number of provided columns
  if (length(id_column) > 1L) {
    ..error(
      paste0(
        "Only one column is expected to be contain ", col_type, " identifiers. ",
        length(id_column), " columns were provided."
      ),
      error_class = "dataset_error"
    )
  }
  
  # End of checks for external.
  if (check_stringency == "external") return(invisible(TRUE))
  
  # Check whether the column exists in the data set. This may be turned into a
  # warning.
  if (!id_column %in% colnames(data)) {
    stop_or_warn(
      paste0(
        "The ", col_type, " identifier column ", id_column,
        " does not appear in the provided data set."
      ),
      check_stringency == "strict"
    )
  }

  # Check whether the id column is the same as another id_column
  if (length(intersect(id_column, other_id_column)) > 0L) {
    ..error(
      paste0(
        "The ", col_type, " identifier column ", id_column,
        " is also used as a different identifier column."
      ),
      error_class = "dataset_error"
    )
  }
  
  # Check whether the id column overlaps with the outcome column
  if (length(intersect(id_column, outcome_column)) > 0L) {
    ..error(
      paste0(
        "The ", col_type, " identifier column ", id_column,
        " is also used as an outcome column."
      ),
      error_class = "dataset_error"
    )
  }
  
  # End of checks for external_warn. Only strict remains.
  if (check_stringency == "external_warn") return()
  
  # Check whether the identifier column is erroneously included in the
  # signature.
  if (length(intersect(id_column, signature)) > 0L) {
    ..error(
      paste0(
        "The ", col_type, " identifier column ", id_column,
        " also appears in the signature."
      ),
      error_class = "dataset_error"
    )
  }
  
  # Check whether the identifier column is erroneously included in the set of
  # features marked for exclusion.
  if (length(intersect(id_column, exclude_features)) > 0L) {
    ..error(
      paste0(
        "The ", col_type, " identifier column ", id_column, 
        " also appears among the features marked for exclusion."
      ),
      error_class = "dataset_error"
    )
  }
  
  # Check whether the identifier column is erroneously included in the set of
  # features marked for inclusion.
  if (length(intersect(id_column, include_features)) > 0L) {
    ..error(
      paste0(
        "The ", col_type, " identifier column ", id_column,
        " also appears among the features marked for inclusion."
      ),
      error_class = "dataset_error"
    )
  }
  
  return(invisible(TRUE))
}



#' Internal function to test plausibility of provided survival times.
#'
#' This function checks whether non-positive outcome time is present in the
#' data. This may produce unexpected results for some packages. For example,
#' glmnet will not train if an instance has a survival time of 0 or lower.
#'
#' @param data Data set as loaded using the `.load_data` function.
#' @inheritParams .parse_experiment_settings
#'
#' @return NULL
#' @md
#' @keywords internal
.check_survival_time_plausibility <- function(
    data,
    outcome_type,
    outcome_column,
    check_stringency = "strict"
) {
  
  if (outcome_type %in% c("survival", "competing_risk")) {
    # Find the levels in the data
    outcome_time <- unique_na(data[[outcome_column]])
    
    if (length(outcome_time) > 0L) {
      if (
        any(outcome_time <= 0.0) && check_stringency %in% c("strict", "external_warn")
      ) {
        ..warning(
          paste0(
            "Survival data contain instances with non-positive (zero or negative) time. ",
            "Some packages that could be used during the analysis, such as glmnet, ",
            "will not be able to produce useful models."
          ),
          warning_class = "dataset_error"
        )
      }
    }
  }
  
  return(invisible(TRUE))
}



#' Internal function to test plausibility of provided class levels
#'
#' This function checks whether categorical levels are present in the data that
#' are not found in the user-provided class levels.
#'
#' @param data Data set as loaded using the `.load_data` function.
#' @inheritParams .parse_experiment_settings
#' @inheritParams as_data_object
#'
#' @return NULL
#' @md
#' @keywords internal
.check_class_level_plausibility <- function(
    data,
    outcome_type, 
    outcome_column,
    class_levels,
    check_stringency = "strict"
) {
  
  if (
    outcome_type %in% c("binomial", "multinomial") &&
    !is.null(class_levels) &&
    !is.null(outcome_column)
  ) {
    
    # Find the levels in the data
    unique_levels <- unique_na(data[[outcome_column]])
    
    # Find levels that are not provided
    missing_levels <- setdiff(unique_levels, class_levels)
    
    if (check_stringency %in% c("strict", "external_warn")) {
      if (length(missing_levels) > 0L) {
        ..error(
          paste0(
            "The outcome data contains levels that are not found among the provided class levels: ",
            paste_s(missing_levels)
          ),
          error_class = "dataset_error"
        )
      }
    }
  }
  
  return(invisible(TRUE))
}



#' Internal function to check whether feature columns are found in the data
#' 
#' This function checks whether feature columns can be found in the data set.
#' It will raise an error if any feature columns are missing from the data set.
#'
#' @param data Data set as loaded using the `.load_data` function.
#' @param feature Character string(s) indicating one or more features.
#'
#' @return NULL
#' @md
#' @keywords internal
.check_feature_availability <- function(data, feature) {
  # Check whether features are available in the data
  
  # Missing features are features that are not available in the provided data.
  missing_feature <- feature[!feature %in% colnames(data)]
  
  if (length(missing_feature) > 0L) {
    ..error(
      paste0(
        "One or more features could not be found in the data set: ",
        paste_s(missing_feature)
      ),
      error_class = "dataset_error"
    )
  }
  
  return(invisible(TRUE))
}



.is_survival_status_col <- function(
    column_name,
    data,
    censoring_indicator = NULL,
    event_indicator = NULL,
    competing_risk_indicator = NULL
) {
  # Identify if the column could contain survival status information.
  
  # Find all indicators
  present_indicators <- c(
    censoring_indicator,
    event_indicator,
    competing_risk_indicator
  )
  
  if (!is.null(present_indicators)) {
    x <- data[[column_name]]
    x <- x[!is.na(x)]
    
    # Check if all data belong to the indicators.
    return(all(x %in% present_indicators))
    
  } else if (is.logical(data[[column_name]])) {
    return(TRUE)
    
  } else if (is.numeric(data[[column_name]])) {
    x <- data[[column_name]]
    x <- x[is.finite(x)]
    
    # Check if all finite values are either 0 or 1
    return((sum(x == 0L) + sum(x == 1L)) == length(x))
    
  }
  
  return(FALSE)
}



.is_survival_time_col <- function(
    column_name,
    data) {
  # Identify if the column could contain survival time information.
  
  if (is.numeric(data[[column_name]])) {
    x <- data[[column_name]]
    x <- x[is.finite(x)]
    
    # Check if all finite values are 0 or positive
    return(all(x >= 0.0))
  }
  
  return(FALSE)
}


.get_available_default_censoring_indicator <- function() {
  return(c("0", "false", "f", "n", "no"))
}

.get_available_default_event_indicator <- function() {
  return(c("1", "true", "t", "y", "yes"))
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
#' @include FamiliarDataComputationPredictionData.R

# predictionTableRegression ----------------------------------------------------
setClass(
  "predictionTableRegression",
  contains = "familiarDataElementPredictionTable",
  slots = list(
    "observed_value_range" = "numeric"
  ),
  prototype = list(
    "observed_value_range" = NA_real_
  )
)

# predictionTableSurvival ------------------------------------------------------
setClass(
  "predictionTableSurvival",
  contains = "familiarDataElementPredictionTable",
  slots = list(
    "censored" = "character",
    "event" = "character",
    "time" = "numeric"
  ),
  prototype = list(
    "censored" = NA_character_,
    "event" = NA_character_,
    "time" = Inf
  )
)

# predictionTableSurvivalHazardRatio -------------------------------------------
setClass(
  "predictionTableSurvivalHazardRatio",
  contains = "predictionTableSurvival"
)

# predictionTableSurvivalCumulativeHazard --------------------------------------
setClass(
  "predictionTableSurvivalCumulativeHazard",
  contains = "predictionTableSurvival"
)

# predictionTableSurvivalTime --------------------------------------------------
setClass(
  "predictionTableSurvivalTime",
  contains = "predictionTableSurvival"
)

# predictionTableSurvivalProbability -------------------------------------------
setClass(
  "predictionTableSurvivalProbability",
  contains = "predictionTableSurvival"
)

# predictionTableGrouping ------------------------------------------------------
setClass(
  "predictionTableGrouping",
  contains = "familiarDataElementPredictionTable",
  slots = list(
    "groups" = "character"
  ),
  prototype = list(
    "groups" = NA_character_
  )
)

# predictionTableRiskGroups ----------------------------------------------------
setClass(
  "predictionTableRiskGroups",
  contains = "predictionTableGrouping"
)

# predictionTableClassification ------------------------------------------------
setClass(
  "predictionTableClassification",
  contains = "familiarDataElementPredictionTable",
  slots = list(
    "classes" = "character",
    "thresholds" = "ANY"
  ),
  prototype = list(
    "classes" = NA_character_,
    "thresholds" = NULL
  )
)

# predictionTableNovelty -------------------------------------------------------
setClass(
  "predictionTableNovelty",
  contains = "familiarDataElementPredictionTable"
)


#' Convert to prediction table object
#'
#' Creates a prediction table object from input data.
#'
#' @param x Values predicted using a learner. For all but `classification`
#'   problems, predicted values should be a single vector of values in any
#'   format that results in a single-column `data.table` using
#'   `data.table::as.data.table`. For `classification` problems, predicted
#'   values are probabilities for each class. Here, it is recommended to ensure
#'   probabilities can be mapped to their respective class, e.g. using a named
#'   list.
#' @param type The type of prediction table that should be created. The
#'   following types are available:
#'
#'   * `regression`: The predicted values are values for a regression.
#'
#'   * `classification`: The predicted values are probabilities for specific
#'   classes.
#'
#'   * `hazard_ratio`: The predicted values are hazard ratios.
#'
#'   * `cumulative_hazard`: The predicted values are cumulative hazards at
#'   time `time`.
#'
#'   * `expected_survival_time`: The predicted values are expected survival times.
#'
#'   * `survival_probability`: The predicted values are survival probabilities
#'   at time `time`.
#'
#' @param y Known outcome value corresponding to each entry in `x`. For
#'   survival-related outcomes, two sets of values are expected, corresponding
#'   to the observed time and event status, respectively. Alternatively, a
#'   `survival::Surv` object can be provided.
#'
#' @param batch_id (*optional*) Array of batch or cohort identifiers.
#'
#'   In familiar any row of data is organised by four identifiers:
#'
#'   * The batch identifier `batch_id`: This denotes the group to which a
#'   set of samples belongs, e.g. patients from a single study, samples measured
#'   in a batch, etc. The batch identifier is used for batch normalisation, as
#'   well as selection of development and validation datasets.
#'
#'   * The sample identifier `sample_id`: This denotes the sample level,
#'   e.g. data from a single individual. Subsets of data, e.g. bootstraps or
#'   cross-validation folds, are created at this level.
#'
#'   * The series identifier `series_id`: Indicates measurements on a
#'   single sample that may not share the same outcome value, e.g. a time
#'   series, or the number of cells in a view.
#'
#'   * The repetition identifier `repetition_id`: Indicates repeated measurements
#'   in a single series where any feature values may differ, but the outcome
#'   does not. Repetition identifiers are always implicitly set when multiple
#'   entries for the same series of the same sample in the same batch that share
#'   the same outcome are encountered.
#'
#' @param sample_id (*optional*) Array of sample or subject identifiers. See
#'   `batch_id` above for more details.
#'
#'   If unset, every row will be identified as a single sample.
#' @param series_id (*optional*) Array of series identifiers, which distinguish
#'   between measurements that are part of a series for a single sample. See
#'   `batch_id` above for more details.
#'
#' @param repetition_id (*optional*) Array of repetition identifiers, which
#'   distinguishes between repeated measurements within a single series. See
#'   `batch_id` above for more details.
#'
#' @param time Time point at which the predicted values are generated e.g. the
#'   cumulative risks generated by random forest.
#'
#'   This parameter is only relevant for `survival` outcomes.
#' @param value_range Range of observed, **not predicted**, values.
#'
#'   This parameter is only relevant for `continuous` outcomes.
#'
#' @param learner The type of learner that generated the predictions.
#' @param vimp_method The type of variable importance method for identifying the
#'   features included by the learner that generated the predictions.
#' @param model_object A familiarModel or familiarEnsemble that can be used (and
#'   is used internally) for setting several of the other arguments of this
#'   function.
#' @param data A familiar dataObject object that can be used (and is used
#'   internally) for setting many of the other arguments of this function.
#'
#' @inheritParams .parse_experiment_settings
#'
#' @return A prediction table object.
#' @export
as_prediction_table <- function(
    x,
    type,
    y = waiver(),
    batch_id = waiver(),
    sample_id = waiver(),
    series_id = waiver(),
    repetition_id = waiver(),
    time = waiver(),
    class_levels = waiver(),
    value_range = waiver(),
    event_indicator = waiver(),
    censoring_indicator = waiver(),
    learner = waiver(),
    vimp_method = waiver(),
    model_object = NULL,
    data = NULL
) {
  
  # Create skeleton of object based on the provided type.
  if (type == "regression") {
    object <- methods::new("predictionTableRegression")
    
  } else if (type == "classification") {
    object <- methods::new("predictionTableClassification")
  
  } else if (type == "survival") {
    object <- methods::new("predictionTableSurvival")
    
  } else if (type == "hazard_ratio") {
    object <- methods::new("predictionTableSurvivalHazardRatio")
    
  } else if (type == "cumulative_hazard") {
    object <- methods::new("predictionTableSurvivalCumulativeHazard")
  
  } else if (type == "expected_survival_time") {
    object <- methods::new("predictionTableSurvivalTime")
    
  } else if (type == "survival_probability") {
    object <- methods::new("predictionTableSurvivalProbability")
  
  } else if (type == "grouping") {
    object <- methods::new("predictionTableGrouping")
    
  } else if (type == "risk_stratification") {
    object <- methods::new("predictionTableRiskGroups")
    
  } else if (type == "novelty") {
    object <- methods::new("predictionTableNovelty")
    
  } else {
    ..error(
      paste0(
        "The type argument was not recognized: ", type,
        "One of regression, classification, hazard_ratio, cumulative_hazard, ",
        "expected_survival_time, survival_probability, grouping, 
        risk_stratification or novelty is expected."
      )
    )
  }
  
  # Infer outcome_type based on known information.
  if (is(data, "dataObject") || is(data, "familiarDataElementPredictionTable")) {
    object@outcome_type <- data@outcome_type
  } else if (is(model_object, "familiarModelUnion")) {
    object@outcome_type <- model_object@outcome_type
  }
  
  # Set learner.
  if (!is.waive(learner)) {
    object@learner <- learner
  } else if (is(model_object, "familiarModelUnion")) {
    object@learner <- model_object@learner
  }
  
  # Set vimp method.
  if (!is.waive(vimp_method)) {
    object@vimp_method <- vimp_method
  } else if (is(model_object, "familiarModelUnion")) {
    object@vimp_method <- model_object@vimp_method
  }
  
  if (is_empty(x)) {
    return(object)
  }
  
  # Use x to set prediction_data attribute.
  object@prediction_data <- data.table::as.data.table(x)
  if (is_empty(object@prediction_data)) return(object)
  
  # Use y to set reference_data attribute.
  if (!is.waive(y) && !is.null(y)) {
    if (survival::is.Surv(y)) {
      object@reference_data <- data.table::data.table(
        "outcome_time" = y[, 1L],
        "outcome_event" = y[, 2L]
      )
    } else {
      object@reference_data <- data.table::as.data.table(y)
    }
    
  } else if (is(data, "dataObject")) {
    outcome_columns <- .get_outcome_columns(data@outcome_type)
    if (length(outcome_columns) > 0L) {
      y <- data@data[, mget(outcome_columns)]
      object@reference_data <- data.table::copy(y)
    }
    
  } else if (is(data, "familiarDataElementPredictionTable")) {
    outcome_columns <- .get_outcome_columns(data@outcome_type)
    if (.is_merged_prediction_table(data) && length(outcome_columns) > 0L) {
      y <- data@data[, mget(outcome_columns)]
      object@reference_data <- data.table::copy(y)
      
    } else if (length(outcome_columns) > 0L) {
      y <- data@reference_data[, mget(outcome_columns)]
      object@reference_data <- data.table::copy(y)
    }
  }
  
  # Use batch_id, sample_id, series_id and repetition_id to set identifier_data
  # attribute.
  if (is.waive(batch_id) && is.waive(sample_id) && is.waive(series_id) && is.waive(repetition_id)) {
    if (is(data, "dataObject")) {
      data_slot <- "data"
    } else if (is(data, "familiarDataElementPredictionTable")) {
      data_slot <- ifelse(.is_merged_prediction_table(data), "data", "identifier_data")
    } else {
      data_slot <- NULL
    }
    
    if (!is.null(data_slot)) {
      batch_id <- methods::slot(data, data_slot)[[get_id_columns(single_column = "batch")]]
      sample_id <- methods::slot(data, data_slot)[[get_id_columns(single_column = "sample")]]
      series_id <- methods::slot(data, data_slot)[[get_id_columns(single_column = "series")]]
      repetition_id <- methods::slot(data, data_slot)[[get_id_columns(single_column = "repetition")]]
    }
  }
  
  if (is.waive(batch_id)) {
    batch_id <- rep_len("generic_batch", nrow(object@prediction_data))
  }
  
  if (is.waive(sample_id)) {
    sample_id <- paste0("generic_sample_", seq_len(nrow(object@prediction_data)))
  }
  
  if (is.waive(series_id)) {
    series_id <- rep_len("1", nrow(object@prediction_data))
  }
  
  if (is.waive(repetition_id)) {
    repetition_id <- rep_len("1", nrow(object@prediction_data))
  }
  
  object@identifier_data <- data.table::data.table(
    "V1" = batch_id, "V2" = sample_id, "V3" = series_id, "V4" = repetition_id
  )
  
  data.table::setnames(object@identifier_data, new = get_id_columns())
  
  # Use time to set time attribute of prediction tables that have it.
  if (methods::.hasSlot(object, "time") && !is.waive(time)) {
    sapply(
      time,
      .check_number_in_valid_range,
      var_name = "time",
      range = c(0.0, Inf),
      closed = c(TRUE, FALSE)
    )
    
    object <- add_data_element_identifier(
      x = object,
      evaluation_time = time
    )[[1L]]
    
    object@time <- time
  }
  
  # Use classes to set classes attribute of prediction tables that have it.
  if (methods::.hasSlot(object, "classes")) {
    if (is.unset(class_levels) && is(data, "familiarDataElementPredictionTable")) {
      if (methods::.hasSlot(data, "classes")) {
        class_levels <- data@classes
      }
    }
    if (is.unset(class_levels) && is(data, "dataObject")) {
      class_levels <- get_outcome_class_levels(data)
    }
    if (is.unset(class_levels) && is(model_object, "familiarModelUnion")) {
      class_levels <- get_outcome_class_levels(model_object)
    }
    if (is.unset(class_levels) && !is_empty(object@reference_data)) {
      if (is.factor(object@reference_data[[1L]])) {
        class_levels <- levels(object@reference_data[[1L]])
      } else {
        class_levels <- unique_na(object@reference_data[[1L]])
      }
    }
    if (is.unset(class_levels) && !is_empty(names(x))) {
      class_levels <- names(x)
    }
    if (is.unset(class_levels)) {
      class_levels <- paste0("class_", seq_len(ncol(object@prediction_data)))
    }
    
    object@classes <- class_levels
  }
  
  # Use value range to set observed_value_range of prediction tables that have
  # it.
  if (methods::.hasSlot(object, "observed_value_range")) {
    if (!is.unset(value_range)) {
      sapply(
        value_range,
        .check_number_in_valid_range,
        var_name = "value_range",
        range = c(-Inf, Inf)
      )
      
      .check_argument_length(
        x = value_range, 
        var_name = "value_range",
        min = 2L,
        max = 2L
      )
    }
    
    if (is.unset(value_range) && is(data, "familiarDataElementPredictionTable")) {
      if (methods::.hasSlot(data, "observed_value_range")) {
        value_range <- data@observed_value_range
      }
    }
    if ((is.unset(value_range) || !all(is.finite(value_range))) && is(data, "dataObject")) {
      if (is(data@outcome_info, "outcomeInfo")) {
        value_range <- suppressWarnings(range(
          data@outcome_info@distribution$fivenum, 
          na.rm = TRUE, 
          finite = TRUE
        ))
      }
    }
    if ((is.unset(value_range) || !all(is.finite(value_range))) && is(model_object, "familiarModelUnion")) {
      if (is(model_object@outcome_info, "outcomeInfo")) {
        value_range <- suppressWarnings(range(
          model_object@outcome_info@distribution$fivenum, 
          na.rm = TRUE, 
          finite = TRUE
        ))
      }
    }
    if ((is.unset(value_range) || !all(is.finite(value_range))) && !is_empty(object@reference_data)) {
      value_range <- suppressWarnings(range(
        object@reference_data[[1L]],
        na.rm = TRUE, 
        finite = TRUE
      ))
    }
    if (is.unset(value_range) || !all(is.finite(value_range))) {
      value_range <- suppressWarnings(range(
        object@prediction_data[[1L]],
        na.rm = TRUE, 
        finite = TRUE
      ))
    }
    
    object@observed_value_range <- value_range
  }
  
  
  # Use censored_label to set censored attribute of prediction tables that have
  # it.
  if (methods::.hasSlot(object, "censored")) {
    if (is.unset(censoring_indicator) && is(data, "familiarDataElementPredictionTable")) {
      if (methods::.hasSlot(data, "censored")) {
        censoring_indicator <- data@censored
      }
    }
    if (is.unset(censoring_indicator) && is(data, "dataObject")) {
      if (is(data@outcome_info, "outcomeInfo")) {
        censoring_indicator <- data@outcome_info@censored
      }
    }
    if (is.unset(censoring_indicator) && is(model_object, "familiarModelUnion")) {
      if (is(model_object@outcome_info, "outcomeInfo")) {
        censoring_indicator <- model_object@outcome_info@censored 
      }
    }
    if (is.unset(censoring_indicator)) {
      censoring_indicator <- .get_available_default_censoring_indicator()
    }
    
    object@censored <- censoring_indicator
  }
  
  # Use event_label to set event attribute of prediction tables that have it.
  if (methods::.hasSlot(object, "event")) {
    if (is.unset(event_indicator) && is(data, "familiarDataElementPredictionTable")) {
      if (methods::.hasSlot(data, "censored")) {
        event_indicator <- data@event
      }
    }
    if (is.unset(event_indicator) && is(data, "dataObject")) {
      if (is(data@outcome_info, "outcomeInfo")) {
        event_indicator <- data@outcome_info@event
      }
    } 
    if (is.unset(event_indicator) && is(model_object, "familiarModelUnion")) {
      if (is(model_object@outcome_info, "outcomeInfo")) {
        event_indicator <- model_object@outcome_info@event 
      }
    }
    if (is.unset(event_indicator)) {
      event_indicator <- .get_available_default_event_indicator()
    }
    
    object@event <- event_indicator
  }
  
  # Check the prediction table object for consistency.
  object <- .complete_new_prediction_table(object)
  
  return(object)
}


# .complete_new_prediction_table (generic) -------------------------------------

setGeneric(
  ".complete_new_prediction_table",
  function(object, ...) standardGeneric(".complete_new_prediction_table")
)

# .complete_new_prediction_table (general) -------------------------------------
setMethod(
  ".complete_new_prediction_table",
  signature(object = "familiarDataElementPredictionTable"),
  function(object) {
    # Check that prediction_data, identifier_data and reference_data have the
    # same number of instances.
    if (is_empty(object)) return(object)
    
    n_instances <- nrow(object@prediction_data)
    
    if (!is_empty(object@reference_data)) {
      if (n_instances != nrow(object@reference_data)) {
        ..error(
          paste0(
            "The number of instances of predicted values (",
            n_instances, ") do not match the number of reference values (",
            nrow(object@reference_data), ")."
          ),
          error_class = "prediction_table_error"
        )
      }
    }
    
    if (!is_empty(object@identifier_data)) {
      if (n_instances != nrow(object@identifier_data)) {
        ..error(
          paste0(
            "The number of instances of predicted values (",
            n_instances, ") do not match the number of instances determined ",
            "from identifier data (", nrow(object@reference_data), ")."
          ),
          error_class = "prediction_table_error"
        )
      }
    }
    
    return(object)
  }
)

# .complete_new_prediction_table (regression) ----------------------------------
setMethod(
  ".complete_new_prediction_table",
  signature(object = "predictionTableRegression"),
  function(object) {
    
    # Pass to superclass method first.
    object <- callNextMethod()
    
    if (is_empty(object)) return(object)
    
    # Check outcome type
    if (is.null(object@outcome_type)) object@outcome_type <- "continuous"
    if (object@outcome_type != "continuous") {
      ..error_reached_unreachable_code(
        "The outcome type of predictionTableRegression objects should be \"continuous\"."
      )
    }
    
    # Check that one set of prediction data are provided.
    if (ncol(object@prediction_data) != 1L) {
      ..error(
        paste0(
          "Only one set of predicted values was expected, but (",
          ncol(object@prediction_data), ") sets were provided."
        ),
        error_class = "prediction_table_error"
      )
    }
    
    # Update column name of prediction_data to the standard value.
    data.table::setnames(object@prediction_data, new = "predicted_outcome")
    
    # Check that prediction data are numeric.
    if (!is.numeric(object@prediction_data$predicted_outcome)) {
      ..error(
        paste0(
          "Predicted values are expected to be numeric, but data with class ",
          paste_s(class(object@prediction_data$predicted_outcome)),
          " were encountered."
        ),
        error_class = "prediction_table_error"
      )
    }
    
    if (!is_empty(object@reference_data)) {
      # Check that one set of reference data are provided.
      if (ncol(object@reference_data) != 1L) {
        ..error(
          paste0(
            "Only one set of reference values was expected, but (",
            ncol(object@reference_data), ") sets were provided."
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Update column name of reference_data to the standard value.
      outcome_column <- get_outcome_columns(object@outcome_type)
      data.table::setnames(object@reference_data, new = outcome_column)
      
      # Check that reference data are numeric.
      if (!is.numeric(object@reference_data[[outcome_column]])) {
        ..error(
          paste0(
            "Reference values are expected to be numeric, but data with class ",
            paste_s(class(object@reference_data[[outcome_column]])), " were encountered."
          ),
          error_class = "prediction_table_error"
        )
      }
    }
    
    return(object)
  }
)


# .complete_new_prediction_table (classification) ------------------------------
setMethod(
  ".complete_new_prediction_table",
  signature(object = "predictionTableClassification"),
  function(object) {
    
    ..in_valid_range <- function(x) {
      return(all(x >= 0.0, na.rm = TRUE) && all(x <= 1.0, na.rm = TRUE))
    }
    
    ..test_columns_numeric <- function(object) {
      if (!all(sapply(object@prediction_data, is.numeric))) {
        ..error(
          paste0(
            "Predicted values are expected to be numeric, but data with classes ",
            paste_s(unique(sapply(object@prediction_data, class))),
            " were encountered."
          ),
          error_class = "prediction_table_error"
        )
      }
    }
    
    ..test_columns_probabilities <- function(object) {
      if (!all(sapply(object@prediction_data, ..in_valid_range))) {
        ..error(
          paste0(
            "Predicted values are expected to be probabilities between 0.0 and 1.0, ",
            "but values outside this range were found."
          ),
          error_class = "prediction_table_error"
        )
      }
    }
    
    # Pass to superclass method first.
    object <- callNextMethod()
    
    if (is_empty(object)) return(object)
    
    # Check outcome type
    if (is.null(object@outcome_type)) {
      object@outcome_type <- ifelse(length(object@classes) == 2L, "binomial", "multinomial")
    }
    if (!object@outcome_type %in% c("binomial", "multinomial")) {
      ..error_reached_unreachable_code(paste0(
        "The outcome type of predictionTableClassification objects should be 
        \"binomial\ or \"multinomial\"."
      ))
    }
    
    if (object@outcome_type == "binomial") {
      # Under binomial, either 1 or 2 columns can be specified, both of which
      # should contain numeric values between 0.0 and 1.0.
      
      # Check that one or two sets of prediction data are provided.
      if (ncol(object@prediction_data) > 2L) {
        ..error(
          paste0(
            "For binomial (two-class) endpoints, either one or two sets of ",
            "predicted values are expected. However (",
            ncol(object@prediction_data), ") sets were provided."
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Check that all columns are numeric.
      ..test_columns_numeric(object)
      
      # Check that all columns contain values between 0.0 and 1.0.
      ..test_columns_probabilities(object)
      
      if (length(object@classes) != 2L) {
        ..error(
          paste0(
            "Two classes are expected for binomial (two-class) but ",
            length(object@classes), " were found: ",
            paste_s(object@classes)
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Insert complementary column. First we determine the name of the
      # associated class.
      if (ncol(object@prediction_data) == 1L) {
        if (all(colnames(object@prediction_data) %in% object@classes)) {
          positive_class <- colnames(object@prediction_data)
          
        } else {
          positive_class <- tail(object@classes, n = 1L)
          
          ..warning(
            paste0(
              "The positive class cannot be directly inferred from names of the ",
              "provided prediction data, and is assumed to be ", positive_class, "."
            ),
            warning_class = "prediction_table_warning"
          )
          
          data.table::setnames(object@prediction_data, new = positive_class)
        }
        
        negative_class <- setdiff(object@classes, positive_class)
        object@prediction_data[, (negative_class) := 1.0 - get(positive_class)]
      }
      
      # Update column names.
      if (!any(colnames(object@prediction_data) %in% object@classes)) {
        # Here neither of the sets of predicted values can be directly matched 
        # against a known class, and we infer that these appear in the order of 
        # classes.
        
        ..warning(
          paste0(
            "Neither set of predicted values (",
            paste_s(colnames(object@prediction_data)),
            ") can be directly assigned to a corresponding class (",
            paste_s(object@classes), "). We assume these sets ",
            "are ordered according to the provided classes."
          ),
          warning_class = "prediction_table_warning"
        )
        data.table::setnames(object@prediction_data, new = object@classes)
        
      } else if (!all(colnames(object@prediction_data) %in% object@classes)) {
        # Here one of the sets of predicted values can be matched against a
        # known class, but the other cannot. We assume that the unmatched class
        # corresponds to the unidentified set.
        
        unassigned_class <- setdiff(object@classes, colnames(object@prediction_data))
        unassigned_column <- setdiff(colnames(object@prediction_data), object@classes)
        ..warning(
          paste0(
            "One of the sets of predicted values (", unassigned_column,
            ") cannot be directly assigned to a corresponding class (",
            paste_s(object@classes), "). We assume that this ",
            "set corresponds to the ", unassigned_class,  " class."
          ),
          warning_class = "prediction_table_warning"
        )
        data.table::setnames(
          x = object@prediction_data, 
          old = unassigned_column, 
          new = unassigned_class
        )
      }
      
    } else {
      # Under multinomial, 2 or more columns can be specified, all of which
      # should contain numeric values between 0.0 and 1.0.
      
      # Check that all columns are numeric.
      ..test_columns_numeric(object)
      
      # Check that all columns contain values between 0.0 and 1.0.
      ..test_columns_probabilities(object)
      
      # Check that at least two sets of prediction data are provided.
      if (ncol(object@prediction_data) < 2L) {
        rlang::abort(
          paste0(
            "For multinomial (multi-class) endpoints, at least two sets of ",
            "predicted values are expected. However (",
            ncol(object@prediction_data), ") sets were provided."
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Check that at least two classes are provided.
      if (length(object@classes) < 2L) {
        ..error(
          paste0(
            "At least two classes are expected for multinomial (multi-class) ",
            "endpoints, but only ", length(object@classes), " were found: ",
            paste_s(object@classes)
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Check that the number of classes and provided sets match.
      if (length(object@classes) != length(colnames(object@prediction_data))) {
        ..error(
          paste0(
            "The same number of classes and sets of predicted values are ",
            "expected. Found: ", length(object@classes), " classes and ",
            length(colnames(object@prediction_data)), " sets."
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Update column names.
      if (!any(colnames(object@prediction_data) %in% object@classes)) {
        # Here neither of the sets of predicted values can be directly matched 
        # against a known class, and we infer that these appear in the order of 
        # classes.
        
        ..warning(
          paste0(
            "Neither set of predicted values (",
            paste_s(colnames(object@prediction_data)),
            ") can be directly assigned to a corresponding class (",
            paste_s(object@classes), "). We assume these sets ",
            "are ordered according to the provided classes."
          ),
          warning_class = "prediction_table_warning"
        )
        data.table::setnames(object@prediction_data, new = object@classes)
        
      } else if (!all(colnames(object@prediction_data) %in% object@classes)) {
        # Here at least one of the sets of predicted values can be matched
        # against a known class, but some others cannot. We cannot assume
        # anything.
        
        unassigned_class <- setdiff(object@classes, colnames(object@prediction_data))
        unassigned_column <- setdiff(colnames(object@prediction_data), object@classes)
        ..error(
          paste0(
            "Some sets of predicted values (",
            paste_s(unassigned_column),
            ") cannot be directly assigned to a corresponding class (",
            paste_s(object@classes), ". Please name all ",
            "sets of predicted values according to their class."
          ),
          error_class = "prediction_table_warning"
        )
        data.table::setnames(
          x = object@prediction_data, 
          old = unassigned_column, 
          new = unassigned_class
        )
      }
    }
    
    # Order columns in prediction data.
    data.table::setcolorder(object@prediction_data, object@classes)
    
    if (!is_empty(object@reference_data)) {
      # Check that one set of reference data are provided.
      if (ncol(object@reference_data) != 1L) {
        ..error(
          paste0(
            "Only one set of reference values was expected, but (",
            ncol(object@reference_data), ") sets were provided."
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Update column name of reference_data to the standard value.
      outcome_column <- get_outcome_columns(object@outcome_type)
      data.table::setnames(object@reference_data, new = outcome_column)
      
      # Check that reference data only contains the known classes.
      missing_classes <- setdiff(
        unique_na(object@reference_data[[outcome_column]]),
        object@classes
      )
        
      if (length(missing_classes) > 0L) {
        ..error(
          paste0(
            "One or more reference values are present that are not listed as ",
            "classes: ", paste_s(missing_classes)
          ),
          error_class = "prediction_table_error"
        )
      }
    }
    
    return(object)
  }
)


# .complete_new_prediction_table (survival) ------------------------------------
setMethod(
  ".complete_new_prediction_table",
  signature(object = "predictionTableSurvival"),
  function(object) {
    
    # Pass to superclass method first.
    object <- callNextMethod()
    
    if (is_empty(object)) return(object)
    
    # Check outcome type
    if (is.null(object@outcome_type)) object@outcome_type <- "survival"
    if (object@outcome_type != "survival") {
      ..error_reached_unreachable_code(
        "The outcome type of predictionTableSurvival objects should be \"survival\"."
      )
    }
    
    # Check that one set of prediction data are provided.
    if (ncol(object@prediction_data) != 1L) {
      ..error(
        paste0(
          "Only one set of predicted values was expected, but (",
          ncol(object@prediction_data), ") sets were provided."
        ),
        error_class = "prediction_table_error"
      )
    }
    
    # Update column name of prediction_data to the standard value.
    data.table::setnames(object@prediction_data, new = "predicted_outcome")
    
    # Check that prediction data are numeric.
    if (!is.numeric(object@prediction_data$predicted_outcome)) {
      ..error(
        paste0(
          "Predicted values are expected to be numeric, but data with class ",
          paste_s(class(object@prediction_data$predicted_outcome)),
          " were encountered."
        ),
        error_class = "prediction_table_error"
      )
    }
    
    if (!is_empty(object@reference_data)) {
      # Check that two sets of reference data are provided (time, status).
      if (ncol(object@reference_data) != 2L) {
        ..error(
          paste0(
            "Both time and status data are expected, but (",
            ncol(object@reference_data), ") set(s) of reference data were provided."
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Update column name of reference_data to the standard value.
      outcome_column <- get_outcome_columns(object@outcome_type)
      data.table::setnames(object@reference_data, new = outcome_column)
      
      # Check that the time column is numeric and positive.
      time_column <- object@reference_data[[outcome_column[1L]]]
      if (!is.numeric(time_column)) {
        ..error(
          paste0(
            "Observed time is expected to be numeric, but data with class ",
            paste_s(time_column), " were encountered."
          ),
          error_class = "prediction_table_error"
        )
      }
      
      if (any(time_column < 0.0, na.rm = TRUE)) {
        negative_values <- time_column
        negative_values <- negative_values[is.finite(negative_values)]
        negative_values <- negative_values[negative_values < 0.0]
        
        ..error(
          paste0(
            "Observed time can not be negative, but negative values were found: ",
            paste_sh(negative_values)
          ),
          error_class = "prediction_table_error"
        )
      }
      
      # Check that the status columns contains 0s and 1s, or can be converted to
      # such values.
      status_column <- object@reference_data[[outcome_column[2L]]]
      status_indicators <- as.character(unique_na(status_column))
      
      if (length(status_indicators) > 2L) {
        ..error(
          paste0(
            "Event status may only contain two values that indicate censoring ",
            "and event. Found: ", paste_sh(status_indicators)
          ),
          error_class = "prediction_table_error"
        )
      }
      
      if (all(status_indicators %in% c(object@censored, object@event))) {
        # Convert to 0s and 1s.
        new_status_column <- rep_len(1L, length(status_column))
        new_status_column[status_column %in% object@censored] <- 0L
        
        object@reference_data <- data.table::copy(object@reference_data)
        object@reference_data[[outcome_column[2L]]] <- new_status_column
        
      } else if (all(status_indicators %in% c("0", "1"))) {
        # Do nothing.
        
      } else if (all(status_indicators %in% c(
        .get_available_default_censoring_indicator(),
        .get_available_default_event_indicator()
      ))) {
        
        # Convert to 0s and 1s.
        new_status_column <- rep_len(1L, length(status_column))
        new_status_column[status_column %in% .get_available_default_censoring_indicator()] <- 0L
        
        object@reference_data <- data.table::copy(object@reference_data)
        object@reference_data[[outcome_column[2L]]] <- new_status_column
        
      } else {
        
        # Find indicator values
        if (!all(is.na(c(object@censored, object@event)))) {
          indicator_values <- c(object@censored, object@event)
        } else {
          indicator_values <- c(
            .get_available_default_censoring_indicator(),
            .get_available_default_event_indicator()
          )
        }
        
        ..error(
          paste0(
            "Could not map potential status and event indicators (",
            paste_s(indicator_values), ") to values present in the status column (",
            paste_sh(status_indicators), ")."
          ),
          error_class = "prediction_table_error"
        )
      }
    }
    
    return(object)
  }
)

# .complete_new_prediction_table (novelty) -------------------------------------
setMethod(
  ".complete_new_prediction_table",
  signature(object = "predictionTableNovelty"),
  function(object) {
    
    # Pass to superclass method first.
    object <- callNextMethod()
    
    if (is_empty(object)) return(object)
    
    # Check outcome type
    object@outcome_type <- "unsupervised"
    
    # Check that one set of prediction data are provided.
    if (ncol(object@prediction_data) != 1L) {
      ..error(
        paste0(
          "Only one set of predicted values was expected, but (",
          ncol(object@prediction_data), ") sets were provided."
        ),
        error_class = "prediction_table_error"
      )
    }
    
    # Update column name of prediction_data to the standard value.
    data.table::setnames(object@prediction_data, new = "novelty")
    
    # Check that prediction data are numeric.
    if (!is.numeric(object@prediction_data$novelty)) {
      ..error(
        paste0(
          "Predicted values are expected to be numeric, but data with class ",
          paste_s(class(object@prediction_data$novelty)),
          " were encountered."
        ),
        error_class = "prediction_table_error"
      )
    }
    
    return(object)
  }
)

# .complete_new_prediction_table (grouping) ------------------------------------
setMethod(
  ".complete_new_prediction_table",
  signature(object = "predictionTableGrouping"),
  function(object) {
    
    # Pass to superclass method first.
    object <- callNextMethod()
    
    if (is_empty(object)) return(object)
    
    # Set outcome type
    object@outcome_type <- "unsupervised"
    
    # Check that one set of prediction data are provided.
    if (ncol(object@prediction_data) != 1L) {
      ..error(
        paste0(
          "Only one set of predicted values was expected, but (",
          ncol(object@prediction_data), ") sets were provided."
        ),
        error_class = "prediction_table_error"
      )
    }
    
    # Update column name of prediction_data to the standard value.
    data.table::setnames(object@prediction_data, new = "group")
    
    # Check that prediction data contains the groups, or set groups if absent.
    if (!anyNA(object@groups)) {
      group_labels <- unique_na(object@prediction_data$group)
      
      if (!all(group_labels %in% object@groups)) {
        unknown_group_label <- setdiff(group_labels, object@groups)
        
        ..error(
          paste0(
            "The set of predicted values contains group labels that could ",
            "not be matched to provided labels. The following labels were not ",
            "recognised: ", paste_s(unknown_group_label), ". ",
            "The following labels were expected: ", paste_s(object@groups)
          ),
          error_class = "prediction_table_error"
        )
      }
    } else {
      
      if (is.factor(object@prediction_data$group)) {
        object@groups <- levels(object@prediction_data$group)
        
      } else {
        group_labels <- unique_na(object@prediction_data$group)
        if (length(group_labels) > 0L) object@groups <- sort(group_labels)
      }
    }
    
    # Set group to factor.
    if (!is.factor(object@prediction_data$group)) {
      object@prediction_data$group <- factor(
        x = object@prediction_data$group,
        levels = object@groups
      )
    }
    
    return(object)
  }
)


# .is_merged_prediction_table (generic) ----------------------------------------
setGeneric(".is_merged_prediction_table", function(x, ...) standardGeneric(".is_merged_prediction_table"))

## .is_merged_prediction_table (general) ---------------------------------------
setMethod(
  ".is_merged_prediction_table",
  signature(x = "familiarDataElementPredictionTable"),
  function(x, ...) {
    return(!is_empty(x@grouping_column))
  }
)


# is_empty ---------------------------------------------------------------------
setMethod(
  "is_empty",
  signature(x = "familiarDataElementPredictionTable"),
  function(x, ...) {
    if (.is_merged_prediction_table(x)) {
      return(is_empty(x@data))
    } else {
      return(is_empty(x@prediction_data))
    }
  }
)




all_predictions_valid <- function(prediction_table) {
  # Return results when checking with the "all" function.
  return(.check_predictions_valid(
    object = prediction_table,
    check_type = "all"
  ))
}



any_predictions_valid <- function(prediction_table) {
  # Return results when checking with the "any" function.
  
  return(.check_predictions_valid(
    object = prediction_table,
    check_type = "any"
  ))
}



# .check_predictions_valid (generic) -------------------------------------------
setGeneric(".check_predictions_valid", function(object, ...) standardGeneric(".check_predictions_valid"))


# .check_predictions_valid (general) ----------------------------------------
setMethod(
  ".check_predictions_valid",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, check_type) {
    
    if (is_empty(object)) return(FALSE)

    check_fun <- switch(
      check_type,
      "all" = all,
      "any" = any
    )
    
    data_slot <- ifelse(.is_merged_prediction_table(object), "data", "prediction_data")
    
    return(check_fun(is.finite(slot(object, data_slot)$predicted_outcome)))
  }
)


# .check_predictions_valid (NULL) ----------------------------------------------
setMethod(
  ".check_predictions_valid",
  signature(object = "NULL"),
  function(object, check_type) {
    return(FALSE)
  }
)


# .check_predictions_valid (classification) ------------------------------------
setMethod(
  ".check_predictions_valid",
  signature(object = "predictionTableClassification"),
  function(object, check_type) {

    if (is_empty(object)) return(FALSE)
    
    check_fun <- switch(
      check_type,
      "all" = all,
      "any" = any
    )
    
    if (.is_merged_prediction_table(object)) {
      return(all(sapply(
        object@data[, mget(object@value_column)],
        function(x, check_fun) (check_fun(is_valid_data(x))),
        check_fun = check_fun
      )))
    } else {
      return(all(sapply(
        object@prediction_data,
        function(x, check_fun) (check_fun(is_valid_data(x))),
        check_fun = check_fun
      )))
    }
  }
)


# .check_predictions_valid (novelty) ----------------------------------------
setMethod(
  ".check_predictions_valid",
  signature(object = "predictionTableNovelty"),
  function(object, check_type) {
    
    if (is_empty(object)) return(FALSE)
    
    check_fun <- switch(
      check_type,
      "all" = all,
      "any" = any
    )
    
    data_slot <- ifelse(.is_merged_prediction_table(object), "data", "prediction_data")
    
    return(check_fun(is.finite(slot(object, data_slot)$novelty)))
  }
)


# .check_predictions_valid (grouping) ----------------------------------------
setMethod(
  ".check_predictions_valid",
  signature(object = "predictionTableGrouping"),
  function(object, check_type) {
    
    if (is_empty(object)) return(FALSE)
    
    check_fun <- switch(
      check_type,
      "all" = all,
      "any" = any
    )
    
    data_slot <- ifelse(.is_merged_prediction_table(object), "data", "prediction_data")
    
    return(check_fun(!is.na(slot(object, data_slot)$group)))
  }
)



# .has_reference_data (generic) ------------------------------------------------
setGeneric(
  ".has_reference_data",
  function(object, ...) standardGeneric(".has_reference_data")
)



# .has_reference_date (general) ------------------------------------------------
setMethod(
  ".has_reference_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    if (is_empty(object)) return(FALSE)
    
    if (.is_merged_prediction_table(object)) {
      outcome_columns <- c("outcome", "outcome_time", "outcome_event")
      present_outcome_columns <- intersect(colnames(object@data), outcome_columns)
      
      return(length(present_outcome_columns) > 0L)
      
    } else {
      return(!is_empty(object@reference_data))
    }
  }
)



# .has_reference_date (novelty) ------------------------------------------------
setMethod(
  ".has_reference_data",
  signature(object = "predictionTableNovelty"),
  function(object, ...) {
    return(FALSE)
  }
)



# .has_reference_date (grouping) -----------------------------------------------
setMethod(
  ".has_reference_data",
  signature(object = "predictionTableGrouping"),
  function(object, ...) {
    return(FALSE)
  }
)



# .has_reference_date (risk groups) --------------------------------------------
setMethod(
  ".has_reference_data",
  signature(object = "predictionTableRiskGroups"),
  function(object, ...) {
    if (is_empty(object)) return(FALSE)
    
    if (.is_merged_prediction_table(object)) {
      outcome_columns <- c("outcome_time", "outcome_event")
      present_outcome_columns <- intersect(colnames(object@data), outcome_columns)
      
      return(length(present_outcome_columns) > 0L)
      
    } else {
      return(!is_empty(object@reference_data))
    }
  }
)



# .has_reference_date (NULL) ---------------------------------------------------
setMethod(
  ".has_reference_data",
  signature(object = "NULL"),
  function(object, ...) {
    return(FALSE)
  }
)



# .drop_reference_data (generic) -----------------------------------------------
setGeneric(
  ".drop_reference_data",
  function(object, ...) standardGeneric(".drop_reference_data")
) 



# .drop_reference_data (general) -----------------------------------------------
setMethod(
  ".drop_reference_data",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    if (is_empty(object)) return(object)
    
    outcome_columns <- c("outcome", "outcome_time", "outcome_event")
    if (.is_merged_prediction_table(object)) {
      present_outcome_columns <- intersect(colnames(object@data), outcome_columns)
      if (length(present_outcome_columns) == 0L) return(object)
      
      object@data[, (present_outcome_columns) := NULL]
      
      grouping_columns <- setdiff(object@grouping_column, present_outcome_columns)
      if (length(grouping_columns) == 0L) grouping_columns <- NULL
      object@grouping_column <- grouping_columns
      
    } else {
      object@reference_data <- NULL
    }
    
    return(object)
  }
)



# .drop_reference_data (null) --------------------------------------------------
setMethod(
  ".drop_reference_data",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



# remove_invalid_predictions (generic) -----------------------------------------
setGeneric("remove_invalid_predictions", function(object, ...) standardGeneric("remove_invalid_predictions"))


# remove_invalid_predictions (general) -----------------------------------------
setMethod(
  "remove_invalid_predictions",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    if (is_empty(object)) return(object)
    
    if (.is_merged_prediction_table(object)) {
      instance_mask <- is.finite(object@data$predicted_outcome)
      
      object@data <- object@data[instance_mask, ]
      
    } else {
      instance_mask <- is.finite(object@prediction_data$predicted_outcome)
      
      object@identifier_data <- object@identifier_data[instance_mask, ]
      object@prediction_data <- object@prediction_data[instance_mask, ]
      if (!is_empty(object@reference_data)) {
        object@reference_data <- object@reference_data[instance_mask, ]
      }
    }
    
    return(object)
  }
)


# remove_invalid_predictions (classification) ----------------------------------
setMethod(
  "remove_invalid_predictions",
  signature(object = "predictionTableClassification"),
  function(object, ...) {
    if (is_empty(object)) return(object)
    
    if (.is_merged_prediction_table(object)) {
      instance_mask <- as.logical(
        do.call(pmin, lapply(object@data[, mget(object@value_column)], is.finite))
      )
      
      object@data <- object@data[instance_mask, ]
      
    } else {
      instance_mask <- as.logical(
        do.call(pmin, lapply(object@prediction_data, is.finite))
      )
      
      object@identifier_data <- object@identifier_data[instance_mask, ]
      object@prediction_data <- object@prediction_data[instance_mask, ]
      if (!is_empty(object@reference_data)) {
        object@reference_data <- object@reference_data[instance_mask, ]
      }
    }
    
    return(object)
  }
)


# remove_invalid_predictions (novelty) -----------------------------------------
setMethod(
  "remove_invalid_predictions",
  signature(object = "predictionTableNovelty"),
  function(object, ...) {
    if (is_empty(object)) return(object)
    
    if (.is_merged_prediction_table(object)) {
      instance_mask <- is.finite(object@data$novelty)
      
      object@data <- object@data[instance_mask, ]
      
    } else {
      instance_mask <- is.finite(object@prediction_data$novelty)
      
      object@identifier_data <- object@identifier_data[instance_mask, ]
      object@prediction_data <- object@prediction_data[instance_mask, ]
      if (!is_empty(object@reference_data)) {
        object@reference_data <- object@reference_data[instance_mask, ]
      }
    }
    
    return(object)
  }
)


# remove_invalid_predictions (grouping) ----------------------------------------
setMethod(
  "remove_invalid_predictions",
  signature(object = "predictionTableGrouping"),
  function(object, ...) {
    if (is_empty(object)) return(object)
    
    if (.is_merged_prediction_table(object)) {
      instance_mask <- is.finite(object@data$group)
      
      object@data <- object@data[instance_mask, ]
      
    } else {
      instance_mask <- is.finite(object@prediction_data$group)
      
      object@identifier_data <- object@identifier_data[instance_mask, ]
      object@prediction_data <- object@prediction_data[instance_mask, ]
      if (!is_empty(object@reference_data)) {
        object@reference_data <- object@reference_data[instance_mask, ]
      }
    }
    
    return(object)
  }
)



# remove_invalid_predictions (null) --------------------------------------------
setMethod(
  "remove_invalid_predictions",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



# .complete (general) ----------------------------------------------------------
setMethod(
  ".complete",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, ...) {
    object <- .merge_slots_into_data(object)
    
    return(object)
  }
)



# .complete (null) -------------------------------------------------------------
setMethod(
  ".complete",
  signature(object = "NULL"),
  function(object, ...) {
    return(NULL)
  }
)



# .complete (classification) ---------------------------------------------------
setMethod(
  ".complete",
  signature(object = "predictionTableClassification"),
  function(object, ...) {
    # Call general method first.
    object <- callNextMethod()

    if (is_empty(object)) return(object)
    if (!any_predictions_valid(object)) return(object)
    
    # Add predicted class.
    if (!"predicted_class" %in% colnames(object@data)) {
      
      # First define the class with the highest probability.
      class_indices <- apply(object@data[, mget(object@classes)], 1L, which.max, simplify = FALSE)
      class_indices <- sapply(class_indices, function(x) (ifelse(length(x) == 1L, x, NA_integer_)))
      class_predictions <- factor(class_indices, levels = seq_along(object@classes), labels = object@classes)
      
      object@data[, "predicted_class" := class_predictions]
      
      # Updated value columns.
      object@value_column <- c(object@value_column, "predicted_class")
    }
    
    return(object)
  }
)



# filter_missing_outcome (general) ---------------------------------------------
setMethod(
  "filter_missing_outcome",
  signature(data = "familiarDataElementPredictionTable"),
  function(
    data, 
    is_validation = FALSE, 
    outcome_type = NULL, 
    ...
  ) {
    # This method removes instances where reference data are missing.
    # See other methods for filter_missing_outcome in DataObject.R.
    
    # Check if data itself or the reference data contained therein is empty.
    if (is_empty(data)) return(data)
    
    merged_table <- .is_merged_prediction_table(data)
    data_slot <- ifelse(merged_table, "data", "reference_data")
    
    if (is_empty(slot(data, data_slot))) return(data)
    
    # Determine outcome type. This is not merely a shortcut, e.g. for some
    # prediction tables (e.g. predictionTableRiskGroups) the outcome type is
    # unsupervised (which is correct), but outcome columns might still be
    # present.
    if (is.null(outcome_type)) {
      outcome_type <- data@outcome_type
    }
    
    if (outcome_type %in% c("survival", "competing_risk")) {
      if (merged_table && !all(c("outcome_time", "outcome_event") %in% data@grouping_column)) return(data)
      
      outcome_is_valid <- is_valid_data(slot(data, data_slot)$outcome_time) &
        is_valid_data(slot(data, data_slot)$outcome_event)
      
    } else if (outcome_type %in% c("binomial", "multinomial", "continuous")) {
      if (merged_table && !("outcome" %in% data@grouping_column)) return(data)
      
      outcome_is_valid <- is_valid_data(slot(data, data_slot)$outcome)
    } else {
      ..error_outcome_type_not_implemented(outcome_type)
    }
    
    if (is_validation) {
      # Check whether all outcome information is missing for validation. It may
      # be a prospective study. In that case, keep all data.
      if (!any(outcome_is_valid)) outcome_is_valid <- !outcome_is_valid
    }
    
    # Filter instances.
    if (merged_table) {
      data@data <- data@data[outcome_is_valid, ]
      
    } else {
      data@identifier_data <- data@identifier_data[outcome_is_valid, ]
      data@reference_data <- data@reference_data[outcome_is_valid, ]
      data@prediction_data <- data@prediction_data[outcome_is_valid, ]
    }
    
    return(data)
  }
)



setMethod(
  ".as_data_table",
  signature(x = "familiarDataElementPredictionTable"),
  function(x, ...) {
    x <- .merge_slots_into_data(x)
    
    return(data.table::copy(x@data))
  }
)



# .merge_slots_into_data (general) ---------------------------------------------
setMethod(
  ".merge_slots_into_data",
  signature(x = "familiarDataElementPredictionTable"),
  function(x, ...) {
    # Merge identifier_data, reference_data and prediction_data and set
    # value column and grouping columns.
    
    # Do not merge prediction tables that are already merged.
    if (.is_merged_prediction_table(x)) return(x)
    if (is_empty(x)) return(x)
    
    # Set grouping and value columns.
    x@grouping_column <- c(x@grouping_column, colnames(x@identifier_data), colnames(x@reference_data))
    x@value_column <- colnames(x@prediction_data)
    
    # Merge data and set data attribute.
    x@data <- cbind(x@identifier_data, x@reference_data, x@prediction_data)
    
    # Empty data slots.
    slot(x, "identifier_data", check = FALSE) <- NULL
    slot(x, "reference_data", check = FALSE) <- NULL
    slot(x, "prediction_data", check = FALSE) <- NULL

    return(x)
  }
)


# .merge_slots_into_data (null) ------------------------------------------------
setMethod(
  ".merge_slots_into_data",
  signature(x = "NULL"),
  function(x, ...) {
    return(NULL)
  }
)


# .extract_slots_from_data (general) -------------------------------------------
setMethod(
  ".extract_slots_from_data",
  signature(x = "familiarDataElementPredictionTable"),
  function(x, ...) {
    # Extract identifier_data, reference_data and prediction_data from data.
    
    if (!.is_merged_prediction_table(x)) return(x)
    
    # Get reference columns.
    reference_columns <- get_outcome_columns(x@outcome_type)
    
    # Get identifier columns.
    identifier_columns <- setdiff(x@grouping_column, reference_columns)
    
    # Set identifier data.
    x@identifier_data <- x@data[, mget(identifier_columns)]
    
    # Set or generate reference columns.
    if (is.null(reference_columns)) {
      # Pass. Do not set reference data if none are required.
      
    } else if (all(reference_columns %in% colnames(x@data))) {
      x@reference_data <- x@data[, mget(reference_columns)]
        
    } else {
      # Pass do not set reference data it is missing.
    }
    
    # Set prediction data.
    x@prediction_data <- x@data[, mget(x@value_column)]
    
    # Reset grouping and value column attributes, and empty data.
    slot(x, "grouping_column", check = FALSE) <- NULL
    slot(x, "value_column") <- NA_character_
    slot(x, "data", check = FALSE) <- NULL
    
    return(x)
  }
)


# .extract_slots_from_data (null) ----------------------------------------------
setMethod(
  ".extract_slots_from_data",
  signature(x = "NULL"),
  function(x, ...) {
    
    return(NULL)
  }
)


# ..compute_ensemble_prediction_estimates (general) ----------------------------
setMethod(
  "..compute_ensemble_prediction_estimates",
  signature(x = "familiarDataElementPredictionTable"),
  function(x, data, ensemble_method, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    estimation_type <- NULL
    
    # Check if ensembling is actually required -- computing ensemble values is 
    # somewhat expensive because it is done for each subset of data.
    if (data.table::uniqueN(data, by = x@grouping_column) == nrow(data)) {
      return(data[
        , mget(c(x@grouping_column, x@value_column))
      ])
    }
    
    # Check if ensembling uses standard functions where we don't actually need
    # do difficult stuff.
    if (ensemble_method == "median") {
      data <- data[
        estimation_type != "point",
        (x@value_column) := lapply(.SD, stats::median, na.rm = TRUE),
        by = c(x@grouping_column),
        .SDcols = c(x@value_column)
      ]
      data <- unique(data[, mget(c(x@grouping_column, x@value_column))])
      
    } else if (ensemble_method == "mean") {
      data <- data[
        estimation_type != "point",
        (x@value_column) := lapply(.SD, mean, na.rm = TRUE),
        by = c(x@grouping_column),
        .SDcols = c(x@value_column)
      ]
      data <- unique(data[, mget(c(x@grouping_column, x@value_column))])
      
    } else {
      # Compute bootstrap estimates.
      data <- data[
        ,
        ...compute_bootstrap_ensemble_estimates(
          x = .SD,
          prediction_columns = x@value_column,
          confidence_level = x@confidence_level,
          percentiles = x@percentiles,
          ensemble_method = ensemble_method,
          ...
        ),
        by = c(x@grouping_column),
        .SDcols = c("estimation_type", x@value_column)
      ]
    }
    
    return(data)
  }
)


# ..compute_ensemble_prediction_estimates (general groups) ---------------------
setMethod(
  "..compute_ensemble_prediction_estimates",
  signature(x = "predictionTableGrouping"),
  function(x, data, ensemble_method, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    estimation_type <- NULL
    
    # There is no natural way to estimate bootstrap intervals for groups, and we
    # therefore revert to the "median" method.
    if (ensemble_method %in% c("percentile", "bc")) {
      ensemble_method <- "median"
      data <- data[estimation_type != "point"]
    }
    
    # Check if ensembling is actually required -- the computation is somewhat
    # expensive because it is done on the subset of data.
    if (data.table::uniqueN(data, by = x@grouping_column) == nrow(data)) {
      return(data[, mget(c(x@grouping_column, x@value_column))])
      
    } else {
      # For median, we use 
      return(data[
        estimation_type != "point",
        (x@value_column) := lapply(.SD, get_mode),
        by = c(x@grouping_column),
        .SDcols = c(x@value_column)
      ][
        , mget(c(x@grouping_column, x@value_column))
      ])
    }
  }
)



# ..compute_ensemble_prediction_estimates (risk strata) ------------------------
setMethod(
  "..compute_ensemble_prediction_estimates",
  signature(x = "predictionTableRiskGroups"),
  function(x, data, ensemble_method, ...) {
    # Suppress NOTES due to non-standard evaluation in data.table
    estimation_type <- NULL
    
    # There is no natural way to estimate bootstrap intervals for groups, and we
    # therefore revert to the "median" method.
    if (ensemble_method %in% c("percentile", "bc")) {
      ensemble_method <- "median"
      data <- data[estimation_type != "point"]
    }

    # Check if ensembling is actually required -- the computation is somewhat
    # expensive because it is done on the subset of data.
    if (data.table::uniqueN(data, by = x@grouping_column) == nrow(data)) {
      return(data[, mget(c(x@grouping_column, x@value_column))])
      
    } else if (ensemble_method == "median") {
      # For median, we use the closest approximate: the most commonly selected
      # group.
      return(data[
        estimation_type != "point",
        (x@value_column) := lapply(.SD, get_mode),
        by = c(x@grouping_column),
        .SDcols = c(x@value_column)
      ][
        , mget(c(x@grouping_column, x@value_column))
      ])
      
    } else if (ensemble_method == "mean") {
      return(data[
        estimation_type != "point",
        (x@value_column) := lapply(.SD, get_mean_risk_group),
        by = c(x@grouping_column),
        .SDcols = c(x@value_column)
      ][
        , mget(c(x@grouping_column, x@value_column))
      ])
    }
  }
)



...compute_bootstrap_ensemble_estimates <- function(
    x,
    prediction_columns,
    ensemble_method,
    confidence_level = NULL,
    percentiles = NULL
) {

  # Suppress NOTES due to non-standard evaluation in data.table
  estimation_type <- NULL
  
  # Compute ensemble estimates using bootstraps.
  ensemble_values <- lapply(
    prediction_columns,
    function(ii_col, x) {
      return(..bootstrap_ci(
        x = x[estimation_type != "point"][[ii_col]],
        x0 = x[estimation_type == "point"][[ii_col]],
        bootstrap_ci_method = ensemble_method,
        confidence_level = confidence_level,
        percentiles = percentiles
      ))
    },
    x = x
  )
  
  # Determine column names
  column_names <- lapply(
    seq_along(prediction_columns),
    function(ii, prediction_columns, ensemble_values, parse_percentile) {
      # If the content of element ii is not a list itself, it contains only a
      # single value that should be named.
      if (!is.list(ensemble_values[[ii]])) return(prediction_columns[ii])
      
      if (!parse_percentile) {
        # If the content of element ii is a list, assign the prediction column
        # name to the first column, and add the name to the remainder.
        
        column_names <- prediction_columns[ii]
        
        if (length(ensemble_values[[ii]]) > 1L) {
          column_names <- c(
            column_names,
            paste0(prediction_columns[ii], "_", names(ensemble_values[[ii]])[2L:length(ensemble_values[[ii]])])
          )
        }
        
        return(column_names)
        
      } else {
        # Assign the prediction column name to every column in case percentiles
        # are returned.
        return(paste0(prediction_columns[ii], "_", names(ensemble_values[[ii]])))
      }
    },
    prediction_columns = prediction_columns,
    ensemble_values = ensemble_values,
    parse_percentile = is.null(confidence_level) && !is.null(percentiles)
  )
  
  # Flatten list of column names
  column_names <- unlist(column_names)
  
  # Flatten list of ensemble values.
  ensemble_values <- unlist(ensemble_values, recursive = FALSE)
  if (!is.list(ensemble_values)) ensemble_values <- as.list(ensemble_values)
  
  # Set column names.
  names(ensemble_values) <- column_names
  
  return(ensemble_values)
}



# add_model_name (familiarDataElement, prediction table)------------------------
setMethod(
  "add_model_name",
  signature(
    data = "familiarDataElement",
    object = "familiarDataElementPredictionTable"
  ),
  function(data, object) {
    
    data <- add_data_element_identifier(
      x = data,
      model_name = "custom_prediction_table"
    )[[1L]]
    
    return(data)
  }
)


# load_familiar_object (prediction table) --------------------------------------
setMethod(
  "load_familiar_object", 
  signature(object = "familiarDataElementPredictionTable"),
  function(object) {
    return(object)
  }
)


# get_outcome_class_levels (prediction table) ----------------------------------
setMethod(
  "get_outcome_class_levels",
  signature(x = "familiarDataElementPredictionTable"),
  function(x) {
    ..error_reached_unreachable_code("This prediction table object does not have a classes attribute.")
  }
)


# get_outcome_class_levels (classification) ------------------------------------
setMethod(
  "get_outcome_class_levels",
  signature(x = "predictionTableClassification"),
  function(x) {
    return(x@classes)
  }
)


# get_n_samples (prediction table) ---------------------------------------------
setMethod(
  "get_n_samples",
  signature(x = "familiarDataElementPredictionTable"),
  function(x, id_depth = "sample", count_unique = TRUE) {
    data_slot <- ifelse(.is_merged_prediction_table(x), "data", "identifier_data")
    return(get_n_samples(methods::slot(x, data_slot), id_depth = id_depth, count_unique = count_unique))
  }
)



# get_bootstrap_sample (prediction table) --------------------------------------
setMethod(
  "get_bootstrap_sample",
  signature(data = "familiarDataElementPredictionTable"),
  function(
    data, 
    seed = NULL, 
    rstream_object = NULL, 
    n_samples = NULL,
    ...
  ) {
    data <- .merge_slots_into_data(data)
    data@data <- get_bootstrap_sample(
      data = data@data,
      seed = seed,
      rstream_object = rstream_object,
      n_samples = n_samples
    )
    
    return(data)
  }
)



# get_object_name (prediction table) -------------------------------------------
setMethod(
  "get_object_name",
  signature(object = "familiarDataElementPredictionTable"),
  function(object, abbreviated = FALSE) {
    
    return("prediction_table")
  }
)



# #export_prediction_data (collection) -----------------------------------------

#'@rdname export_prediction_data-methods
setMethod(
  "export_prediction_data",
  signature(object = "familiarCollection"),
  function(
    object,
    dir_path = NULL,
    export_collection = FALSE,
    ...
  ) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    .export_looper <- function(object, object_class, subtype, dir_path, export_collection) {
      .export(
        x = object,
        data_slot = "prediction_data",
        dir_path = dir_path,
        object_class = object_class,
        type = "prediction",
        subtype = subtype,
        export_collection = export_collection
      )
    }
    
    object_class <- c(
      "predictionTableRegression",
      "predictionTableSurvival",
      "predictionTableSurvivalHazardRatio",
      "predictionTableSurvivalCumulativeHazard",
      "predictionTableSurvivalTime",
      "predictionTableSurvivalProbability",
      "predictionTableGrouping",
      "predictionTableRiskGroups",
      "predictionTableClassification",
      "predictionTableNovelty"
    )
    
    subtype <- c(
      "regression",
      "survival_generic",
      "hazard_ratio",
      "cumulative_hazard",
      "expected_survival_time",
      "survival_probability",
      "grouping",
      "risk_stratification",
      "classification",
      "novelty"
    )

    export_data <- mapply(
      FUN = .export_looper,
      object_class = object_class,
      subtype = subtype,
      MoreArgs = list(
        "object" = object,
        "dir_path" = dir_path,
        "export_collection" = export_collection
      ),
      SIMPLIFY = FALSE,
      USE.NAMES = FALSE
    )
    
    names(export_data) <- subtype
    export_data <- export_data[!sapply(export_data, is_empty)]
    
    return(export_data)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarTaskTrain -------------------------------------------------------------
setClass(
  "familiarTaskTrain",
  contains = "familiarTask",
  slots = list(
    "vimp_method" = "character",
    "learner" = "character",
    "vimp_table_file" = "character",
    "hyperparameter_file" = "character",
    "feature_info_file" = "character"
  ),
  prototype = methods::prototype(
    vimp_method = NA_character_,
    learner = NA_character_,
    vimp_table_file = NA_character_,
    hyperparameter_file = NA_character_,
    feature_info_file = NA_character_,
    task_name = "train_model"
  )
)



# .set_file_name (train task) --------------------------------------------------
setMethod(
  ".set_file_name",
  signature(object = "familiarTaskTrain"),
  function(object, file_paths = NULL) {
    if (is.null(file_paths)) return(object)
    
    # Generate file name of the model.
    object@file <- get_object_file_name(
      object_type = "familiarModel",
      data_id = object@data_id,
      run_id = object@run_id,
      learner = object@learner,
      vimp_method = object@vimp_method,
      project_id = object@project_id,
      dir_path = file_paths$mb_dir
    )
    
    return(object)
  }
)



# .get_task_descriptor (train task) --------------------------------------------
setMethod(
  ".get_task_descriptor",
  signature(object = "familiarTaskTrain"),
  function(object, ...) {
    return(paste0(
      object@task_name, "_",
      object@data_id, "_", 
      object@run_id, "_", 
      object@vimp_method, "_", 
      object@learner
    ))
  }
)



# .perform_task (train task , NULL) --------------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskTrain",
    data = "NULL"
  ),
  function(
    object,
    data,
    experiment_data = NULL,
    outcome_info = NULL,
    ...
  ) {
    # This method is called when "data" is expected to be available somewhere in
    # the backend.
    
    if (is.null(experiment_data)) {
      ..error_reached_unreachable_code("experiment_data is required for retrieving data from the backend.")
    }
    if (is.null(outcome_info)) {
      ..error_reached_unreachable_code("outcome_info is required.")
    }
    
    # Find the run list.
    run_list <- .get_run_list(
      iteration_list = experiment_data@iteration_list,
      data_id = object@data_id,
      run_id = object@run_id
    )
    
    # Select unique samples.
    sample_identifiers <- .get_sample_identifiers(
      run = run_list,
      train_or_validate = "train"
    )
    sample_identifiers <- unique(sample_identifiers)
    
    # Create a dataObject.
    data <- methods::new(
      "dataObject",
      data = get_data_from_backend(sample_identifiers = sample_identifiers),
      preprocessing_level = "none",
      outcome_type = outcome_info@outcome_type,
      outcome_info = outcome_info
    )
    
    # Pass to method that dispatches with dataObject for further processing.
    return(.perform_task(
      object = object,
      data = data,
      experiment_data = experiment_data,
      ...
    ))
  }
)


# .perform_task (train task, dataObject) ---------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskTrain",
    data = "dataObject"
  ),
  function(
    object,
    data,
    vimp_aggregation_method = NULL,
    vimp_rank_threshold = NULL,
    settings = NULL,
    feature_info_list = NULL,
    vimp_table = NULL,
    hyperparameters = NULL,
    novelty_detector = NULL,
    detector_parameters = NULL,
    message_indent = 0L,
    verbose = FALSE,
    cl = NULL,
    return_results = TRUE,
    ...
  ) {
    logger_message(
      paste0(
        "Training: Starting model training for the \"", object@learner,
        "\" learner and the \"", object@vimp_method,
        "\" variable importance method for run ",
        object@task_id, " of ",
        object@n_tasks, "."
      ),
      indent = message_indent,
      verbose = verbose
    )
    
    # Check that outcome_info is present on data
    if (!is(data@outcome_info, "outcomeInfo")) {
      ..error_reached_unreachable_code(
        "outcome_info attribute of data (dataObject) does not contain an outcomeInfo object"
      )
    }
    
    # Set vimp aggregation method and vimp_rank_threshold based on settings.
    if (!is.null(settings)) {
      if (is.null(vimp_aggregation_method)) {
        vimp_aggregation_method <- settings$vimp$aggregation
      }
      if (is.null(vimp_rank_threshold)) {
        vimp_rank_threshold <- settings$vimp$aggr_rank_threshold
      }
    }
    
    # Check and retrieve feature info list.
    feature_info_list <- .get_feature_info_list(
      object = object,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings,
      message_indent = message_indent,
      verbose = verbose,
      cl = cl,
      ...
    )
    
    # Check and retrieve hyperparameters. We do this prior to retrieving the
    # variable importance tables, as these may be attached to hyperparameter
    # object.
    hyperparameters <- .get_hyperparameters(
      object = object,
      hyperparameters = hyperparameters,
      vimp_aggregation_method = vimp_aggregation_method,
      vimp_rank_threshold = vimp_rank_threshold,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings,
      message_indent = message_indent,
      verbose = verbose,
      cl = cl,
      ...
    )
    
    if (is_empty(hyperparameters$vimp_table)) {
      # Check and retrieve variable importances from the drive, or generate in
      # place, if the hyperparameter object did not contain a variable
      # importance table.
      vimp_table <- .get_variable_importance_table(
        object = object,
        vimp_table = vimp_table,
        feature_info_list = feature_info_list,
        data = data,
        settings = settings,
        message_indent = message_indent,
        verbose = verbose,
        cl = cl,
        ...
      )
      
    } else {
      vimp_table <- hyperparameters$vimp_table
    }
    
    # Create the raw model object for training.
    model_object <- methods::new(
      "familiarModel",
      outcome_type = data@outcome_type,
      hyperparameters = hyperparameters$hyperparameters,
      hyperparameter_data = hyperparameters$hyperparameter_data,
      vimp_method = object@vimp_method,
      vimp_table = vimp_table,
      vimp_aggregation_method = vimp_aggregation_method,
      vimp_rank_threshold = vimp_rank_threshold,
      learner = object@learner,
      feature_info = feature_info_list,
      outcome_info = data@outcome_info,
      data_id = object@data_id,
      run_id = object@run_id,
      run_table = .get_current_run_table(object = object),
      settings = settings$eval,
      project_id = object@project_id
    )
    
    # Select features based on variable importances.
    model_object <- set_model_features(
      object = model_object,
      minimise_footprint = FALSE
    )
    
    # Train model.
    model_object <- .train(
      object = model_object,
      data = data,
      get_additional_info = TRUE,
      ...
    )
    
    # Set up task to train novelty detector
    detector_task <- methods::new(
      "familiarTaskTrainNovelty",
      learner = novelty_detector,
      vimp_method = object@vimp_method,
      data_id = object@data_id,
      run_id = object@run_id,
      project_id = object@project_id
    )
    
    # Train novelty detector and add to model.
    model_object@novelty_detector <- .perform_task(
      object = detector_task,
      data = data,
      selected_features = features_after_clustering(
        features = model_object@novelty_features,
        feature_info_list = feature_info_list),
      settings = settings,
      feature_info_list = feature_info_list,
      vimp_table = vimp_table,
      vimp_aggregation_method = vimp_aggregation_method,
      vimp_rank_threshold = vimp_rank_threshold,
      hyperparameters = detector_parameters,
      return_results = TRUE
    )
    
    # Add model name
    model_object <- set_object_name(model_object)
    
    if (!is.na(object@file)) {
      saveRDS(model_object, file = object@file)
    }
    
    if (return_results) {
      return(model_object)
    }
    
    return(TRUE)
  }
)



# .get_hyperparameters (train task) --------------------------------------------
setMethod(
  ".get_hyperparameters",
  signature(object = "familiarTaskTrain"),
  function(
    object,
    hyperparameters,
    file_paths = NULL,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    can_pre_process <- NULL
    
    hyperparameter_object <- NULL
    if (is.null(hyperparameters) && !is.null(object@run_table)) {
      # This routine loads hyperparameters from disk, and is used when an
      # experiment is run using summon_familiar.
      
      # This check exists to make sure that the standard workflow passes the
      # correct objects.
      if (is.null(file_paths)) {
        ..error_reached_unreachable_code("file_paths was expected, but not provided.")
      }
      
      # Find the last entry on the run table that is marked as available for
      # pre-processing. This is what hyperparameters are based on.
      hyperparameter_run <- tail(
        object@run_table[[paste0(object@data_id, ".", object@run_id)]][can_pre_process == TRUE, ],
        n = 1L
      )
      
      # Find the file name.
      hyperparameter_file <- get_object_file_name(
        project_id = object@project_id,
        data_id = hyperparameter_run$data_id[1L],
        run_id = hyperparameter_run$run_id[1L],
        learner = object@learner,
        vimp_method = object@vimp_method,
        object_type = "hyperparametersLearner",
        dir_path = file_paths$mb_dir
      )
      
      if (file.exists(hyperparameter_file)) {
        hyperparameter_object <- update_object(readRDS(hyperparameter_file))
      }
    }
    
    
    if (is.null(hyperparameter_object) && is.na(object@hyperparameter_file)) {
      # Create an ad-hoc list of hyperparameters
      
      # Set up task, and explicitly don't write to file.
      hyperparameter_task <- methods::new(
        "familiarTaskLearnerHyperparameters",
        project_id = object@project_id,
        vimp_method = object@vimp_method,
        learner = object@learner,
        file = NA_character_
      )
      
      # Execute the task.
      hyperparameter_object <- .perform_task(
        object = hyperparameter_task,
        hyperparameters = hyperparameters,
        ...
      )
      
    } else if (is.null(hyperparameter_object)) {
      # Assume that the hyperparameter_file attribute contains the path to the
      # file containing the vimp method hyperparameters.
      if (!file.exists(object@hyperparameter_file)) {
        ..error(paste0("hyperparameter file does not exist at location: ", object@hyperparameter_file))
      }
      hyperparameter_object <- update_object(readRDS(object@hyperparameter_file))

    } else if (is.character(hyperparameters)) {
      # If hyperparameters is a string, interpret this as a path to the
      # file containing the vimp method hyperparameters.
      if (!file.exists(hyperparameters)) {
        ..error(paste0("hyperparameter file does not exist at location: ", hyperparameters))
      }
      hyperparameter_object <- update_object(readRDS(hyperparameters))
    }
    
    if (is(hyperparameter_object, "familiarModel")) {
      hyperparameters <- list(
        "hyperparameters" =  hyperparameter_object@hyperparameters,
        "hyperparameter_data" = hyperparameter_object@hyperparameter_data,
        "vimp_table" = hyperparameter_object@vimp_table
      )
      
    } else {
      hyperparameters <- list("hyperparameters" = hyperparameters)
    }
    
    return(hyperparameters)
  }
)



.generate_trainer_tasks <- function(
    experiment_data,
    optimisation_determine_vimp,
    vimp_methods,
    learners,
    file_paths,
    skip_existing = FALSE
) {
  # Suppress NOTES due to non-standard evaluation in data.table
  train <- main_data_id <- can_pre_process <- vimp <- NULL
  
  # Find the data_id related to model training.
  data_id <- experiment_data@experiment_setup[train == TRUE, ]$main_data_id[1L]
  if (is_empty(data_id)) return(NULL)
  
  # Initialise empty list.
  task_list <- list()
  ii <- 1L
  run_tables <- .collect_run_tables(iteration_list = experiment_data@iteration_list)
  
  # train tasks ----------------------------------------------------------------
  
  # Get run ids.
  run_ids <- seq_len(experiment_data@experiment_setup[main_data_id == data_id]$n_runs[1L])
  
  # Set up variable importance computation task.
  for (learner in learners) {
    for (vimp_method in vimp_methods) {
      for (run_id in run_ids) {
        
        # Create task to generate run-specific feature info.
        train_task <- methods::new(
          "familiarTaskTrain",
          data_id = data_id,
          run_id = run_id,
          vimp_method = vimp_method,
          learner = learner,
          run_table = run_tables,
          project_id = experiment_data@project_id
        )
        
        # Add file names.
        train_task <- .set_file_name(
          object = train_task,
          file_paths = file_paths
        )
        
        # Add to list, if the file does not exist on disk.
        if (!skip_existing || !.file_exists(train_task)) {
          task_list[[ii]] <- train_task
          ii <- ii + 1L
        }
      }
    }
  }
  
  # Check if any train-related tasks are required.
  if (length(task_list) == 0L) return(NULL)
  
  # learner hyperparameter tasks -----------------------------------------------
  
  # Check how variable importance data should be handled.
  if (is_empty(experiment_data@experiment_setup[vimp == TRUE, ])) {
    use_vimp <- "return_hpo_vimp"
    
  } else if (optimisation_determine_vimp) {
    use_vimp <- "use_hpo_vimp"
    
  } else {
    use_vimp <- "use_main_vimp"
  }
  
  # Set up variable importance hyperparameter task.
  train_run_table <- .get_run_table_from_experiment_setup(
    data_id = data_id,
    experiment_setup = experiment_data@experiment_setup
  )
  learner_hyperparameter_data_id <- tail(
    train_run_table[main_data_id <= data_id & can_pre_process == TRUE, ],
    n = 1L
  )$main_data_id[1L]
  
  # Get run ids.
  run_ids <- seq_len(train_run_table[main_data_id == learner_hyperparameter_data_id, ]$n_runs[1L])
  
  for (learner in learners) {
    for (vimp_method in vimp_methods) {
      for (run_id in run_ids) {
        # Create task to generate run-specific feature info.
        learner_hyperparameter_task <- methods::new(
          "familiarTaskLearnerHyperparameters",
          data_id = learner_hyperparameter_data_id,
          run_id = run_id,
          use_vimp = use_vimp,
          vimp_method = vimp_method,
          learner = learner,
          run_table = run_tables,
          project_id = experiment_data@project_id
        )
        
        # Add file names.
        learner_hyperparameter_task <- .set_file_name(
          object = learner_hyperparameter_task,
          file_paths = file_paths
        )
        
        # Add to list, if the file does not exist on disk.
        if (!skip_existing || !.file_exists(learner_hyperparameter_task)) {
          task_list[[ii]] <- learner_hyperparameter_task
          ii <- ii + 1L
        }
      }
    }
  }
  
  # Add tasks related to data processing for learner methods.
  task_list <- c(
    task_list, 
    .generate_learner_data_preprocessing_tasks(
      experiment_data = experiment_data,
      file_paths = file_paths
    )
  )
  
  # variable importance tasks --------------------------------------------------
  task_list <- c(
    task_list,
    .generate_vimp_tasks(
      experiment_data = experiment_data,
      vimp_methods = vimp_methods,
      file_paths = file_paths,
      skip_existing = skip_existing
    )
  )
  
  return(task_list)
}



.run_learner <- function(
    cl,
    tasks,
    message_indent = 0L,
    verbose,
    ...
) {
  
  # Check that any tasks are available for processing.
  if (is_empty(tasks$hyperparameters_learner) || is_empty(tasks$train)) return(invisible(FALSE))
  
  # Determine which learner hyperparameter sets need to be found.
  finished_tasks <- sapply(tasks$hyperparameters_learner, .file_exists)
  unfinished_tasks <- tasks$hyperparameters_learner[!finished_tasks]
  finished_tasks <- tasks$hyperparameters_learner[finished_tasks]
  
  # Process any unfinished tasks.
  if (length(unfinished_tasks) > 0L) {
    ..run_learner_computation_hyperparameters(
      cl = cl,
      tasks = unfinished_tasks,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
  }
  
  # Determine which variable importance tasks are required.
  finished_tasks <- sapply(tasks$train, .file_exists)
  unfinished_tasks <- tasks$train[!finished_tasks]
  finished_tasks <- tasks$train[finished_tasks]
  
  # Process any unfinished tasks.
  if (length(unfinished_tasks) > 0L) {
    ..run_learner(
      cl = cl,
      tasks = unfinished_tasks,
      message_indent = message_indent,
      verbose = verbose,
      ...
    )
  }
  
  return(invisible(TRUE))
}



..run_learner <- function(
    tasks,
    cl,
    settings,
    message_indent = 0L,
    verbose,
    ...
) {
  
  # Message that variable importances computation is starting.
  logger_message(
    paste0(
      "Training: Starting model training."
    ),
    indent = message_indent,
    verbose = verbose
  )
  
  fam_mapply_lb(
    cl = cl,
    assign = "all",
    FUN = .perform_task,
    progress_bar = FALSE,
    object = tasks,
    MoreArgs = list(
      "data" = NULL,
      "return_results" = FALSE,
      "settings" = settings,
      "novelty_detector" = settings$mb$novelty_detector,
      "detector_paramaters" = settings$mb$detector_parameters[[settings$mb$novelty_detector]],
      "vimp_aggregation_method" = settings$vimp$aggregation,
      "vimp_rank_threshold" = settings$vimp$aggr_rank_threshold,
      "message_indent" = message_indent + 1L,
      "verbose" = verbose,
      ...
    )
  )
  
  # Message that variable importances have been computed.
  logger_message(
    paste0(
      "Training: Models were trained.\n"
    ),
    indent = message_indent,
    verbose = verbose
  )
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R


setClass("familiarHyperparameterLearnerLAGP",
  contains = "familiarHyperparameterLearner",
  slots = list("encoding_reference_table" = "ANY"),
  prototype = list("encoding_reference_table" = NULL)
)



# initialize -------------------------------------------------------------------
setMethod(
  "initialize",
  signature(.Object = "familiarHyperparameterLearnerLAGP"),
  function(.Object, ...) {
    # Update with parent class first.
    .Object <- callNextMethod()

    # Set required package
    .Object@package <- "laGP"

    # Set name
    .Object@name <- "Localised approximate Gaussian process (laGP) hyperparameter optimisation model"

    return(.Object)
  }
)



.get_available_lagp_hyperparameter_learners <- function() {
  return("gaussian_process")
}



# get_prediction_type ----------------------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarHyperparameterLearnerLAGP"),
  function(object, ...) {
    return(c("default", "sd"))
  }
)



# ..train ----------------------------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarHyperparameterLearnerLAGP",
    data = "data.table"
  ),
  function(object, data, ...) {
    # Check if the training data is ok.
    if (has_bad_training_data(object = object, data = data)) {
      return(callNextMethod())
    }

    # Check that required packages are loaded and installed.
    require_package(object, "train")

    # Encode categorical variables.
    encoded_data <- encode_categorical_variables(
      data = data[, mget(object@target_hyperparameters)],
      object = NULL,
      encoding_method = "dummy",
      drop_levels = FALSE
    )

    # Get optimisation score as response.
    y <- data$optimisation_score

    # Add model. Note that laGP basically builds and evaluates the
    # Gaussian process when predicting. We are just passing the
    # information at this point.
    object@model <- list(
      "X" = as.matrix(encoded_data$encoded_data),
      "Z" = y
    )

    # Set reference table for encoding.
    object@encoding_reference_table <- encoded_data$reference_table

    # Set learner version
    object <- set_package_version(object)

    return(object)
  }
)



# ..predict --------------------------------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarHyperparameterLearnerLAGP",
    data = "data.table"
  ),
  function(object, data, type = "default", ...) {
    # Check that required packages are loaded and installed.
    require_package(object, "predict")

    # Check if the model was trained.
    if (!model_is_trained(object)) return(callNextMethod())

    # Check if the data is empty.
    if (is_empty(data)) return(callNextMethod())

    # Get an empty prediction table.
    prediction_table <- .get_placeholder_hyperparameter_prediction_table(
      object = object,
      data = data,
      type = type
    )

    # Encode categorical variables.
    x_encoded <- encode_categorical_variables(
      data = data[, mget(object@target_hyperparameters)],
      object = NULL,
      encoding_method = "dummy",
      drop_levels = FALSE
    )$encoded_data

    # Update the end-size parameter, if the number of instances is
    # smaller than the default of 51. Note that we already capture
    # end_size of 7 or smaller using has_bad_training_data.
    end_size <- ifelse(nrow(object@model$X) < 51L, nrow(object@model$X) - 1L, 50L)

    # Infer scores for the hyperparameters.
    quiet(predicted_scores <- laGP::aGP(
      X = object@model$X,
      Z = object@model$Z,
      XX = as.matrix(x_encoded),
      end = end_size
    ))

    # Compute mean and standard deviation.
    score_mean <- predicted_scores$mean
    score_sd <- sqrt(predicted_scores$var)

    # Separate by type
    if (type == "default") {
      prediction_table[, "mu" := score_mean]
    } else if (type == "sd") {
      prediction_table[, ":="(
        "mu" = score_mean,
        "sigma" = score_sd
      )]
    } else if (type %in% c("percentile", "raw")) {
      ..error_reached_unreachable_code(paste0(
        "..predict,familiarHyperparameterLearnerLAGP,data.table: ",
        "Encountered a prediction type that is not supported by the model: ", type
      ))
    } else {
      ..error_reached_unreachable_code(paste0(
        "..predict,familiarHyperparameterLearnerLAGP,data.table: ",
        "Encountered an unknown prediction type: ", type
      ))
    }

    return(prediction_table)
  }
)



# has_bad_training_data --------------------------------------------------------
setMethod(
  "has_bad_training_data",
  signature(
    object = "familiarHyperparameterLearnerLAGP",
    data = "data.table"
  ),
  function(object, data, ...) {
    # Perform the general check first.
    if (callNextMethod()) return(TRUE)

    # For Gaussian process at least eight instances should be present.
    if (nrow(data) < 8L) return(TRUE)

    return(FALSE)
  }
)

#' @include FamiliarS4Generics.R
NULL


# require_package (character) --------------------------------------------------
setMethod(
  "require_package",
  signature(x = "character"),
  function(
    x,
    purpose = NULL,
    message_type = "error",
    ...
  ) {
    
    # Pass to .require_package
    return(invisible(.require_package(
      x = x,
      purpose = purpose,
      message_type = message_type
    )))
  }
)


# require_package (NULL) -------------------------------------------------------
setMethod(
  "require_package",
  signature(x = "NULL"),
  function(x, purpose = NULL, ...) {
    
    return(invisible(TRUE))
  }
)


.require_package <- function(
    x,
    purpose = NULL,
    message_type = "error",
    call = rlang::caller_env(),
    ...
) {
  
  if (message_type %in% c("error", "warning")) {
    # Attempt to load required packages.
    package_loaded <- sapply(x, requireNamespace, quietly = TRUE)
    
  } else {
    # Check whether packages are installed, without loading the packages.
    package_loaded <- sapply(x, is_package_installed)
  }
  
  # Skip further analysis if all packages could be loaded.
  if (all(package_loaded)) return(TRUE)
  
  # Find all packages that are missing.
  missing_packages <- x[!package_loaded]
  
  if (message_type == "error") {
    # Throw an error.
    ..error_package_not_installed(
      x = missing_packages,
      purpose = purpose,
      call = call
    )
    
  } else if (message_type == "warning") {
    # Raise a warning
    ..warning_package_not_installed(
      x = missing_packages,
      purpose = purpose,
      call = call
    )
    
  } else if (message_type == "backend_error") {
    # Add message to backend.
    ..message_package_not_installed_to_backend(
      x = missing_packages,
      purpose = purpose,
      message_type = "error"
    )
    
  } else if (message_type == "backend_warning") {
    # Add message to backend.
    ..message_package_not_installed_to_backend(
      x = missing_packages,
      purpose = purpose,
      message_type = "warning"
    )
  }
  
  return(FALSE)
}



# Check whether installed packages are outdated or newer.
.check_package_version <- function(
    name,
    version,
    when = NULL
) {
  
  # Do not check if package versions are missing completely.
  if (is_empty(version)) return(invisible(NULL))
  
  # Check for outdated packages.
  package_outdated <- mapply(
    is_package_outdated,
    name = name,
    version = version
  )
  
  # Check for newer packages.
  package_newer <- mapply(
    is_package_newer,
    name = name,
    version = version
  )
  
  # Skip if no packages are outdated.
  if (!any(package_outdated) && !any(package_newer)) return(invisible(NULL))
  
  # Check whether one or more packages do not have the correct
  # version.
  multiple_packages <- sum(package_outdated + package_newer)
  
  # Form "when" string to describe when and for what the packages were initially
  # used.
  if (!is.null(when)) {
    when_str <- ifelse(
      multiple_packages,
      paste0(" from those ", when),
      paste0(" from that ", when)
    )
    
  } else {
    when_str <- NULL
  }
  
  # Initial string.
  message_str <- paste0(
    "The following installed package",
    ifelse(
      multiple_packages,
      paste0("s have versions that differ", when_str, ":"),
      paste0(" has a version that differs", when_str, ":")
    )
  )
  
  for (ii in seq_along(name)) {
    
    # Skip if package is not newer or outdated.
    if (!package_outdated[ii] && !package_newer[ii]) next
    
    # Parse package information.
    message_str <- c(
      message_str,
      paste0(
        name[ii],
        ": ",
        as.character(utils::packageVersion(name[ii])),
        ifelse(package_outdated[ii], " < ", " > "),
        as.character(version[ii]),
        ifelse(package_outdated[ii], " (outdated)", " (newer)")
      )
    )
  }
  
  # Show as warning.
  ..warning(paste(message_str, sep = "\n"))
  
  return(invisible(TRUE))
}



is_package_installed <- function(name) {
  
  if (length(name) == 0L) return(TRUE)
  
  # Attempt to get the installed version of the package. If it fails, the
  # package is not installed.
  #
  # This perhaps the cleanest way to check whether a package exists. require and
  # requireNameSpace attach and load packages, which is not required here. The
  # find.package documentation actively discourages its use to identify whether
  # a package is installed.
  installed_version <- .get_installed_package_version(name)
  
  return(!inherits(installed_version, "error"))
}



is_package_outdated <- function(name, version) {
  
  if (length(name) == 0L) return(TRUE)
  
  # Obtain the installed version of the package.
  installed_version <- .get_installed_package_version(name)
  
  if (inherits(installed_version, "error")) {
    ..error_package_not_installed(name)
  }
  
  # Make sure version is a package version.
  version <- as.package_version(version)
  
  return(version > installed_version)
} 



is_package_newer <- function(name, version) {
  
  if (length(name) == 0L) return(TRUE)
  
  # Obtain the installed version of the package.
  installed_version <- .get_installed_package_version(name)
  
  if (inherits(installed_version, "error")) {
    ..error_package_not_installed(name)
  }
  
  # Make sure version is a package version.
  version <- as.package_version(version)
  
  return(version < installed_version)
}



.get_installed_package_version <- function(name) {
  # Lookup version of the package using utils::packageVersion. The result is
  # cached to reduce lookup times.
  installed_version <- rlang::env_cache(
    familiar_global_env,
    paste0(name, "_version"),
    default = tryCatch(
      utils::packageVersion(name),
      error = identity
    )
  )
  
  return(installed_version)
}



.bioconductor_packages <- function() {
  return(NULL)
}



..message_package_not_installed_to_backend <- function(
    x,
    purpose,
    message_type
) {
  
  # Generate message string.
  message_str <- ..message_missing_package(
    x = x,
    purpose = purpose
  )
  
  missing_packages <- NULL
  
  # Get existing missing package, if any.
  if (exists("missing_packages", where = familiar_global_env)) {
    missing_packages <- get(
      "missing_packages",
      envir = familiar_global_env
    )
  }
  
  # Append missing packages.
  missing_packages <- c(missing_packages, x)
  
  # Assign to backend.
  assign(
    "missing_packages",
    value = missing_packages,
    envir = familiar_global_env
  )
  
  
  missing_package_messages <- NULL
  
  # Get existing package messages, if any.
  if (exists("missing_package_messages", where = familiar_global_env)) {
    
    missing_package_messages <- get(
      "missing_package_messages",
      envir = familiar_global_env
    )
  }
  
  # Append package messages.
  missing_package_messages <- c(missing_package_messages, message_str)
  
  # Assign to backend.
  assign(
    "missing_package_messages",
    value = missing_package_messages,
    envir = familiar_global_env
  )
  
  missing_package_message_type <- NULL
  
  # Get existing message types for missing package.
  if (exists("missing_package_message_type", where = familiar_global_env)) {
    missing_package_message_type <- get(
      "missing_package_message_type",
      envir = familiar_global_env
    )
  }
  
  # Append missing packages.
  missing_package_message_type <- c(missing_package_message_type, message_type)
  
  # Assign to backend.
  assign(
    "missing_package_message_type",
    value = missing_package_message_type,
    envir = familiar_global_env
  )
  
  return(invisible(TRUE))
}


.report_missing_package_messages <- function() {
  
  # Do not report if either messages or message types are missing.
  if (!exists("missing_packages", where = familiar_global_env)) return()
  if (!exists("missing_package_message_type", where = familiar_global_env)) return()
  if (!exists("missing_package_messages", where = familiar_global_env)) return()
  
  # Get all missing packages.
  x <- get("missing_packages", envir = familiar_global_env)
  
  # Only unique packages.
  x <- unique(x)
  
  # Retrieve package messages.
  err_message <- get("missing_package_messages", envir = familiar_global_env)
  
  # Determine number of messages.
  n_messages <- length(err_message)
  
  # Combine error messages.
  if (n_messages > 1L) err_message <- paste0(err_message, collapse = "\n")
  
  # Add basic message to install packages.
  if (n_messages > 1L) {
    err_message <- c(err_message, "\n", ..message_missing_package(x = x))
  }
  
  # Instructions for CRAN packages.
  err_message <- c(err_message, ..message_install_from_cran(x = x))
  
  # Instructions for Bioconductor packages.
  err_message <- c(err_message, ..message_install_from_bioconductor(x = x))
  
  # Obtain message types.
  message_type <- get("missing_package_message_type", envir = familiar_global_env)
  
  # Throw error or warning.
  if (any(message_type == "error")) {
    ..error(
      paste0(err_message, collapse = ""),
      error_class = "package_missing"
    )
    
  } else {
    ..warning(
      paste0(err_message, collapse = ""),
      warning_class = "package_missing"
    )
  }
  
  # Clean up
  rm(
    "missing_packages",
    "missing_package_message_type",
    "missing_package_messages",
    envir = familiar_global_env
  )
  
  return(invisible(TRUE))
}



..message_missing_package <- function(x, purpose = NULL) {
  
  # Select unique packages.
  x <- unique(x)
  
  # Check whether packages are not actually installed.
  all_missing <- !any(sapply(x, is_package_installed))
  
  if (all_missing) {
    message_str <- paste0(
      "The following package",
      ifelse(length(x) > 1L, "s have", " has"),
      " to be installed",
      ifelse(is.null(purpose), ": ", paste0(" ", purpose, ": ")),
      paste_s(x), "."
    )
    
  } else {
    message_str <- paste0(
      "The following package",
      ifelse(length(x) > 1L, "s have", " has"),
      " to be installed, or installed again to update dependencies",
      ifelse(is.null(purpose), ": ", paste0(" ", purpose, ": ")),
      paste_s(x), "."
    )
  }
  
  return(message_str)
}



..message_install_from_cran <- function(x) {
  
  message_str <- NULL
  
  # Select unique packages.
  x <- unique(x)
  x_cran <- setdiff(x, .bioconductor_packages())
  
  if (length(x_cran) > 0L) {
    message_str <- paste0(
      "\n\nInstall from CRAN: ",
      "\n\n\tinstall.packages(c(",
      paste0("\"", x_cran, "\"", collapse = ", "), "))"
    )
  }
  
  return(message_str)
}



..message_install_from_bioconductor <- function(x) {
  
  message_str <- NULL
  
  # Select unique packages.
  x <- unique(x)
  x_bioc <- intersect(x, .bioconductor_packages())
  
  if (length(x_bioc) > 0L) {
    message_str <- paste0(
      "\n\nInstall from Bioconductor: ",
      ifelse(is_package_installed("BiocManager"), "", "\n\n\tinstall.packages(\"BiocManager\")"),
      "\n\tBiocManager::install(c(",
      paste0("\"", x_bioc, "\"", collapse = ", "), "))"
    )
  }
  
  return(message_str)
}



..message_package_version <- function(x, version) {
  return(paste0(x, " (v", as.character(version), ")"))
}

# Stability selection methods, as in Wald R, Khoshgoftaar TM, Dittman D, Awada
# W, Napolitano A. An extensive comparison of feature ranking aggregation
# techniques in bioinformatics. 2012 IEEE 13th Int. Conf. Inf. Reuse Integr.,
# IEEE; 2012, p. 377-84. These methods are based on Haury AC, Gestraud P, Vert
# JP. The influence of feature selection methods on accuracy, stability and
# interpretability of molecular signatures. PLoS One 2011;6:1-12.

.compute_rank_stability <- function(x, rank_threshold) {
  # Aggregate using stability Meinshausen N, Bühlmann P. Stability selection. J
  # R Stat Soc Ser B (Statistical Methodol. 2010 Jul 5;72(4):417-73.

  # Extract the variable importance table.
  vimp_table <- x@vimp_table

  # Compute the number of runs.
  n_runs <- data.table::uniqueN(vimp_table, by = "run_id")

  # Calculate occurrence with a rank l.eq rank_threshold
  vimp_table <- .compute_feature_occurrence(
    vimp_table = vimp_table,
    threshold = rank_threshold,
    n_runs = n_runs
  )

  # Rename "occurrence" column for consistency
  data.table::setnames(
    x = vimp_table,
    old = "occurrence",
    new = "score"
  )

  # Attach to vimpTable object.
  x@vimp_table <- vimp_table

  # Set correct invert value.
  x@invert <- TRUE

  return(x)
}



.compute_rank_exponential <- function(x, rank_threshold) {
  # Aggregate using exponential weighted stability

  # Suppress NOTES due to non-standard evaluation in data.table
  rank <- NULL

  # Calculate occurrence with a rank l.eq rank_threshold and weight by
  # exponential rank: better ranks receive higher scores
  x@vimp_table <- x@vimp_table[
    ,
    list("score" = sum(exp(-rank / rank_threshold) * (rank <= rank_threshold))),
    by = c("name")
  ]

  # Set correct invert value.
  x@invert <- TRUE

  return(x)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# set_data_set_names -----------------------------------------------------------

#' @title Name datasets for plotting and export
#'
#' @description Tabular exports and figures created from a familiarCollection
#'   object can be customised by setting data labels.
#'
#' @details Labels convert internal naming of data sets to the requested label
#'   at export or when plotting. Currently assigned labels can be found using
#'   the \code{get_data_set_names} method.
#'
#' @param x A familiarCollection object.
#' @param old (optional) Set of old labels to replace.
#' @param new Set of replacement labels. The number of replacement labels should
#'   be equal to the number of provided old labels or the full number of labels.
#'   If a subset of labels is to be replaced, both \code{old} and \code{new}
#'   should be provided.
#' @param order (optional) Ordered set of replacement labels. This is used to
#'   provide the order in which the labels should be placed, which affects e.g.
#'   levels in a plot. If the ordering is not explicitly provided, the old
#'   ordering is used.
#'
#' @return A familiarCollection object with custom names for the data sets.
#' @export
#' @aliases set_data_set_names
#' @seealso * \linkS4class{familiarCollection} for information concerning the
#' familiarCollection class. * \code{\link{get_data_set_names}} for obtaining
#' currently assigned labels.
#' @md
setMethod(
  "set_data_set_names",
  signature(x = "familiarCollection"),
  function(
    x,
    old = NULL,
    new = NULL,
    order = NULL
  ) {
    
    x <- .set_labels(
      x = x,
      old_label = old,
      new_label = new,
      new_order = order,
      upd_slot = "data_set_labels"
    )
    
    return(x)
  }
)



# set_learner_names ------------------------------------------------------------

#' @title Rename learners for plotting and export
#'
#' @description Tabular exports and figures created from a familiarCollection
#'   object can be customised by providing names for the learners.
#'
#' @details Labels convert the internal naming for learners to the requested
#'   label at export or when plotting. This enables the use of more specific
#'   naming, e.g. changing \code{random_forest_rfsrc} to \code{Random Forest}.
#'   Currently assigned labels can be found using the \code{get_learner_names}
#'   method.
#'
#' @inheritParams set_data_set_names,familiarCollection-method
#' @return A familiarCollection object with custom labels for the learners.
#' @export
#' @aliases set_learner_names
#' @seealso * \linkS4class{familiarCollection} for information concerning the
#' familiarCollection class. * \code{\link{get_learner_names}} for obtaining
#' currently assigned labels.
#' @md
setMethod(
  "set_learner_names",
  signature(x = "familiarCollection"),
  function(
    x,
    old = NULL,
    new = NULL,
    order = NULL
  ) {
    
    x <- .set_labels(
      x = x,
      old_label = old, 
      new_label = new,
      new_order = order,
      upd_slot = "learner_labels"
    )
    
    return(x)
  }
)



# set_vimp_method_names ----------------------------------------------------------

#' @title Rename variable importance methods for plotting and export
#'
#' @description Tabular exports and figures created from a familiarCollection
#'   object can be customised by providing names for the variable importance
#'   methods.
#'
#' @details Labels convert the internal naming for variable importance methods
#'   to the requested label at export or when plotting. This enables the use of
#'   more specific naming, e.g. changing \code{mim} to \code{Mutual Information
#'   Maximisation}. Currently assigned labels can be found using the
#'   \code{get_vimp_method_names} method.
#'
#' @inheritParams set_data_set_names,familiarCollection-method
#' @return A familiarCollection object with updated labels.
#' @export
#' @aliases set_vimp_method_names
#' @seealso * \linkS4class{familiarCollection} for information concerning the
#' familiarCollection class. * \code{\link{get_vimp_method_names}} for obtaining
#' currently assigned labels.
#' @md
setMethod(
  "set_vimp_method_names",
  signature(x = "familiarCollection"),
  function(
    x,
    old = NULL,
    new = NULL, 
    order = NULL
  ) {
    
    x  <- .set_labels(
      x = x,
      old_label = old,
      new_label = new,
      new_order = order,
      upd_slot = "vimp_method_labels"
    )
    
    return(x)
  }
)



# set_feature_names ------------------------------------------------------------

#' @title Rename features for plotting and export
#'
#' @description Tabular exports and figures created from a familiarCollection
#'   object can be customised by providing names for features.
#'
#' @details Labels convert the internal naming for features to the requested
#'   label at export or when plotting. This enables customisation without
#'   redoing the analysis with renamed input data. Currently assigned labels can
#'   be found using the \code{get_feature_names} method.
#'
#' @inheritParams set_data_set_names,familiarCollection-method
#' @return A familiarCollection object with updated labels.
#' @export
#' @aliases set_feature_names
#' @seealso * \linkS4class{familiarCollection} for information concerning the
#' familiarCollection class. * \code{\link{get_feature_names}} for obtaining
#' currently assigned feature names.
#' @md
setMethod(
  "set_feature_names",
  signature(x = "familiarCollection"),
  function(
    x,
    old = NULL,
    new = NULL,
    order = NULL
  ) {
    
    x <- .set_labels(
      x = x,
      old_label = old,
      new_label = new,
      new_order = order,
      upd_slot = "feature_labels"
    )
    
    return(x)
  }
)



# set_risk_group_names ---------------------------------------------------------

#' @title Rename risk groups for plotting and export
#'
#' @description Tabular exports and figures created from a familiarCollection
#'   object can be customised by providing names for risk groups in survival
#'   analysis.
#'
#' @details Labels convert the internal naming for risk groups to the requested
#'   label at export or when plotting. This enables customisation of risk group
#'   names. Currently assigned labels can be found using the
#'   \code{get_risk_group_names} method.
#'
#' @inheritParams set_data_set_names,familiarCollection-method
#' @return A familiarCollection object with updated labels.
#' @export
#' @aliases set_risk_group_names
#' @seealso * \linkS4class{familiarCollection} for information concerning the
#' familiarCollection class. * \code{\link{get_risk_group_names}} for obtaining
#' currently assigned risk group labels.
#' @md
setMethod(
  "set_risk_group_names",
  signature(x = "familiarCollection"),
  function(
    x,
    old = NULL,
    new = NULL,
    order = NULL
  ) {
    
    x <- .set_labels(
      x = x,
      old_label = old,
      new_label = new,
      new_order = order,
      upd_slot = "km_group_labels"
    )
    
    return(x)
  }
)



# set_class_names --------------------------------------------------------------

#' @title Rename outcome classes for plotting and export
#'
#' @description Tabular exports and figures created from a familiarCollection
#'   object can be customised by providing names for outcome classes.
#'
#' @details Labels convert the internal naming for class levels to the requested
#'   label at export or when plotting. This enables customisation of class
#'   names. Currently assigned labels can be found using the
#'   \code{get_class_names} method.
#'
#' @inheritParams set_data_set_names,familiarCollection-method
#' @return A familiarCollection object with updated labels.
#' @export
#' @aliases set_class_names
#' @seealso * \linkS4class{familiarCollection} for information concerning the
#' familiarCollection class. * \code{\link{get_class_names}} for obtaining
#' currently assigned class names.
#' @md
setMethod(
  "set_class_names",
  signature(x = "familiarCollection"),
  function(
    x,
    old = NULL,
    new = NULL,
    order = NULL
  ) {
    
    x <- .set_labels(
      x = x,
      old_label = old,
      new_label = new,
      new_order = order,
      upd_slot = "class_labels"
    )
    
    return(x)
  }
)



# get_data_set_names -----------------------------------------------------------

#' @title Get current name of datasets
#'
#' @description Datasets in familiarCollection objects can have custom names for
#'   export and plotting. This function retrieves the currently assigned names.
#'
#' @details Labels convert internal naming of data sets to the requested label
#'   at export or when plotting. Labels can be changed using the
#'   \code{set_data_set_names} method.
#'
#' @param x A familiarCollection object.
#'
#' @return An ordered array of dataset name labels.
#' @export
#' @aliases get_data_set_names
#' @seealso
#' * \linkS4class{familiarCollection} for information concerning the familiarCollection class.
#' * \code{\link{set_data_set_names}} for updating the name of datasets and their ordering.
#' @md
setMethod(
  "get_data_set_names",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "data_set_labels",
      get_levels = FALSE
    ))
  }
)



# get_learner_names ------------------------------------------------------------

#' @title Get current learner name labels
#'
#' @description Learners in familiarCollection objects can have custom names for
#'   export and plotting. This function retrieves the currently assigned names.
#'
#' @details Labels convert internal naming of learners to the requested label at
#'   export or when plotting. Labels can be changed using the
#'   \code{set_learner_names} method.
#'
#' @inheritParams get_data_set_names,familiarCollection-method
#'
#' @return An ordered array of learner name labels.
#' @export
#' @aliases get_learner_names
#' @seealso
#' * \linkS4class{familiarCollection} for information concerning the familiarCollection class.
#' * \code{\link{set_learner_names}} for updating the name of learners and their ordering.
#' @md
setMethod(
  "get_learner_names",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "learner_labels",
      get_levels = FALSE
    ))
  }
)



# get_vimp_method_names ----------------------------------------------------------

#' @title Get current variable importance method name labels
#'
#' @description Variable importance methods in familiarCollection objects can
#'   have custom names for export and plotting. This function retrieves the
#'   currently assigned names.
#' 
#' @details Labels convert internal naming of variable importance methods to the
#'   requested label at export or when plotting. Labels can be changed using the
#'   \code{set_vimp_method_names} method.
#'
#' @inheritParams get_data_set_names,familiarCollection-method
#'
#' @return An ordered array of variable importance method name labels.
#' @export
#' @aliases get_vimp_method_names
#' @seealso
#' * \linkS4class{familiarCollection} for information concerning the familiarCollection class.
#' * \code{\link{set_vimp_method_names}} for updating the name of variable importance methods and their ordering.
#' @md
setMethod(
  "get_vimp_method_names",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "vimp_method_labels",
      get_levels = FALSE
    ))
  }
)



# get_feature_names ------------------------------------------------------------

#' @title Get current feature labels
#'
#' @description Features in familiarCollection objects can have custom names for
#'   export and plotting. This function retrieves the currently assigned names.
#'
#' @details Labels convert internal naming of features to the requested label at
#'   export or when plotting. Labels can be changed using the
#'   \code{set_feature_names} method.
#'
#' @inheritParams get_data_set_names,familiarCollection-method
#'
#' @return An ordered array of feature labels.
#' @export
#' @aliases get_feature_names
#' @seealso
#' * \linkS4class{familiarCollection} for information concerning the familiarCollection class.
#' * \code{\link{set_feature_names}} for updating the name and ordering of features.
#' @md
setMethod(
  "get_feature_names",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "feature_labels",
      get_levels = FALSE
    ))
  }
)



# get_risk_group_names ---------------------------------------------------------

#' @title Get current risk group labels
#'
#' @description Risk groups in familiarCollection objects can have custom names
#'   for export and plotting. This function retrieves the currently assigned
#'   names.
#'
#' @details Labels convert internal naming of risk groups to the requested label
#'   at export or when plotting. Labels can be changed using the
#'   \code{set_risk_group_names} method.
#'
#' @inheritParams get_data_set_names,familiarCollection-method
#'
#' @return An ordered array of risk group labels.
#' @export
#' @aliases get_risk_group_names
#' @seealso
#' * \linkS4class{familiarCollection} for information concerning the familiarCollection class.
#' * \code{\link{set_risk_group_names}} for updating the name and ordering of risk groups.
#' @md
setMethod(
  "get_risk_group_names",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "km_group_labels",
      get_levels = FALSE
    ))
  }
)



# get_class_names --------------------------------------------------------------

#' @title Get outcome class labels
#'
#' @description Outcome classes in familiarCollection objects can have custom
#'   names for export and plotting. This function retrieves the currently
#'   assigned names.
#'
#' @details Labels convert internal class names to the requested label at export
#'   or when plotting. Labels can be changed using the \code{set_class_names}
#'   method.
#'
#' @inheritParams get_data_set_names,familiarCollection-method
#'
#' @return An ordered array of class labels.
#' @export
#' @aliases get_class_names
#' @seealso
#' * \linkS4class{familiarCollection} for information concerning the familiarCollection class.
#' * \code{\link{set_class_names}} for updating the name and ordering of classes.
#' @md
setMethod(
  "get_class_names",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "class_labels",
      get_levels = FALSE
    ))
  }
)



# get_data_set_name_levels -----------------------------------------------------
setMethod(
  "get_data_set_name_levels",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "data_set_labels",
      get_levels = TRUE
    ))
  }
)



# get_learner_name_levels ------------------------------------------------------
setMethod(
  "get_learner_name_levels",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "learner_labels",
      get_levels = TRUE
    ))
  }
)



# get_vimp_method_name_levels ----------------------------------------------------
setMethod(
  "get_vimp_method_name_levels",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "vimp_method_labels",
      get_levels = TRUE
    ))
  }
)



# get_feature_name_levels ------------------------------------------------------
setMethod(
  "get_feature_name_levels",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "feature_labels",
      get_levels = TRUE
    ))
  }
)



# get_risk_group_name_levels ---------------------------------------------------
setMethod(
  "get_risk_group_name_levels",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "km_group_labels",
      get_levels = TRUE
    )) 
  }
)



# get_class_name_levels --------------------------------------------------------
setMethod(
  "get_class_name_levels",
  signature(x = "familiarCollection"),
  function(x) {
    
    return(.get_labels(
      x = x,
      upd_slot = "class_labels",
      get_levels = TRUE
    ))
  }
)


# add_package_version (collection) ---------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "familiarCollection"),
  function(object) {
    
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)



# save (collection) ------------------------------------------------------------
setMethod(
  "save",
  signature(
    list = "familiarCollection",
    file = "character"
  ),
  function(list, file) {
    .save(object = list, dir_path = file)
  }
)



# get_object_name (collection) -------------------------------------------------
setMethod(
  "get_object_name",
  signature(object = "familiarCollection"),
  function(
    object,
    abbreviated = FALSE
  ) {
    
    # Get the full name of the object
    object_name <- object@name
    
    return(object_name)
  }
)



# show (collection) ------------------------------------------------------------
setMethod(
  "show",
  signature(object = "familiarCollection"),
  function(object) {
    
    # Make sure the collection object is updated.
    object <- update_object(object = object)
    
    # Create an initial descriptor.
    cat(paste0(
      "A collection of datasets (", object@name, "; ",
      .familiar_version_string(object), "):\n"
    ))
    
    lapply(
      object@data_sets,
      function(x) cat(paste0("  ", x, "\n"))
    )
    
    # Outcome details
    cat("\nThe collection contains data for the following outcome:\n")
    show(object@outcome_info)
  }
)



# .set_labels ------------------------------------------------------------------
setMethod(
  ".set_labels",
  signature(x = "familiarCollection"),
  function(
    x,
    old_label,
    new_label,
    new_order,
    upd_slot
  ) {
    
    # Suppress NOTES due to non-standard evaluation in data.table
    label <- NULL
    
    # Make sure the collection object is updated.
    x <- update_object(object = x)
    
    # Read the slot
    label_table <- slot(x, upd_slot)
    
    # Generate the slot
    if (is.null(label_table)) {
      x <- .construct_label_table(x = x, upd_slot = upd_slot)
      
      # Read the label table
      label_table <- slot(x, upd_slot)
    }
    
    ## Label update ------------------------------------------------------------
    if (!is.null(new_label)) {
      
      # Check that the length of new_label is not 0
      if (length(new_label) == 0L) {
        ..error(
          "The number of \"new\" labels should be larger than 0.",
          error_class = "input_argument_error"
        )
      }
      
      # Check for duplicates in new_label
      if (anyDuplicated(new_label)) {
        ..warning("The \"new\" labels contain duplicate values.")
      }
      
      # Get old labels, if they were not provided.
      if (is.null(old_label)) {
        
        # Check if we can safely assume that the new labels are matching 
        if (length(new_label) != nrow(label_table)) {
          ..error(
            paste0(
              "If \"old\" is not provided explicitly, \"new\" should have the ",
              "same length as the currently existing labels."
            ),
            error_class = "input_argument_error"
          )
        }
        
        # Assume that old is the currently known label
        old_label <- label_table$label
      }
      
      # Check for duplicates in old_label
      if (anyDuplicated(old_label)) {
        ..error(
          "The provided \"old\" labels contain duplicate values.",
          error_class = "input_argument_error"
        )
      }
      
      # Check that length of new_label and old_label are the same
      if (length(new_label) != length(old_label)) {
        ..error(
          "The same number of \"old\" and \"new\" labels should be provided.",
          error_class = "input_argument_error"
        )
      }
      
      # Check if all old_label entries are actually existing labels.
      unrecognised_labels <- old_label[!old_label %in% label_table$label]
      if (length(unrecognised_labels) > 0L) {
        ..error(
          paste0(
            "The following old labels were not found: ",
            paste_s(unrecognised_labels)
          ),
          error_class = "input_argument_error"
        )
      }
      
      # Generate a data.table using old_label and new_label
      replacement_table <- data.table::data.table(
        "label" = old_label,
        "new_label" = new_label
      )
      
      # Merge both tables
      label_table <- merge(
        x = label_table,
        y = replacement_table,
        by = "label",
        all.x = TRUE
      )
      
      # Copy old labels to the new_label column for labels that were not
      # provided previously
      label_table[is.na(new_label), "new_label" := label]
      
      # Rename colums
      data.table::setnames(
        x = label_table,
        old = c("label", "new_label"),
        new = c("old_label", "label")
      )
      
      # Drop the "old_label" column as it is no longer required
      label_table[, "old_label" := NULL]
    }
    
    ## Label order update ------------------------------------------------------
    if (!is.null(new_order)) {
      # Update label order based on new_order
      
      # Check that new_order has the same length as the label_table
      if (length(new_order) != length(unique(label_table$label))) {
        ..error(
          "\"order\" should match the number of labels.",
          error_class = "input_argument_error"
        )
      }
      
      # Check that there are no duplicates in new_order.
      if (anyDuplicated(new_order)) {
        ..warning("\"order\" contains duplicate entries.")
      }
      
      # Check that new_order has the same class the label.
      if (!is_any(new_order, class(label_table$label))) {
        ..error(
          paste0(
            "\"order\" should have the same class as the label, i.e. ",
            paste(class(label_table$label), collapse = " or "), "."
          ),
          error_class = "input_argument_error"
        )
      }
      
      # Check that there are no elements of new_order that do not appear as a label
      unrecognised_labels <- new_order[!new_order %in% label_table$label]
      if (length(unrecognised_labels) > 0L) {
        ..error(
          paste0(
            "The following labels were not found for ordering: ",
            paste_s(unrecognised_labels)
          ),
          error_class = "input_argument_error"
        )
      }
      
      # Find the order by matching the label column in the label table with the
      # provided labels in new_order
      new_label_order <- match(
        label_table$label,
        new_order
      )
      
      # Update the label order
      label_table[, "label_order" := new_label_order]
    }
    
    # Insert the updated label table into its slot
    slot(x, upd_slot) <- label_table
    
    # Return the input object
    return(x)
  }
)



# .construct_label_table -------------------------------------------------------
setMethod(
  ".construct_label_table",
  signature(x = "familiarCollection"),
  function(x, upd_slot) {
    # Constructs a new label table if required.
    
    # Do not generate a new table with labels if it already exists
    if (!is.null(slot(x, upd_slot))) return(x)
    
    if (upd_slot == "data_set_labels") {
      data <- slot(x, "data_sets")
      
    } else if (upd_slot == "learner_labels") {
      data <- slot(x, "learner")
      
    } else if (upd_slot == "vimp_method_labels") {
      data <- slot(x, "vimp_method")
      
    } else if (upd_slot == "feature_labels") {
      data <- unique(c(
        slot(x, "required_features"),
        x@fs_vimp$vimp_table$name
      ))
      
    } else if (upd_slot == "km_group_labels") {
      data <- unique(unlist(lapply(x@km_data, function(x) (levels(x@data$group)))))
      if (all(data %in% c("low", "moderate", "high"))) {
        # Standard labels for 2 or three groups.
        data <- c("low", "moderate", "high")
      } else {
        # Custom labels are sorted alphabetically.
        data <- sort(data)
      }
      
    } else if (upd_slot == "class_labels") {
      data <- get_outcome_class_levels(x)
      
    } else {
      ..error("Slot is not available for familiarCollection objects.")
    }
    
    # Add the new label table to the slot
    slot(x, upd_slot) <- data.table::data.table(
      "internal" = data,
      "label" = data,
      "label_order" = seq_along(data)
    )
    
    return(x)
  }
)



# .get_labels ------------------------------------------------------------------
setMethod(
  ".get_labels",
  signature(x = "familiarCollection"),
  function(
    x,
    upd_slot,
    get_levels = FALSE
  ) {
    # Get ordered levels (i.e. internal column from the label table) from the
    # specific slot
    
    # Suppress NOTES due to non-standard evaluation in data.table
    label_order <- NULL
    
    # Make sure the collection object is updated.
    x <- update_object(object = x)
    
    # Determine if the slot has been set.
    if (is.null(slot(x, upd_slot))) {
      x <- .construct_label_table(x = x, upd_slot = upd_slot)
    }
    
    # Read the label table
    label_table <- slot(x, upd_slot)
    
    # Order table by label order
    label_table <- label_table[order(label_order)]
    
    if (get_levels) {
      # Return ordered levels if get_levels is TRUE
      return(label_table$internal)
      
    } else {
      # Return ordered labels
      return(label_table$label)
    }
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



.is_available_prediction_type <- function(type) {
  return(type %in% c(.get_available_prediction_type_arguments()))
}



.get_available_prediction_type_arguments <- function() {
  return(c(
    .get_available_novelty_prediction_type_arguments(),
    "default", "survival_probability", "risk_stratification"
  ))
}



.get_available_novelty_prediction_type_arguments <- function() {
  return(c("novelty"))
}



# predict (generic) ------------------------------------------------------------

#' @title Model predictions for familiar models and model ensembles
#'
#' @description Fits the model or ensemble of models to the data and shows the
#'   result.
#'
#' @param object A familiar model or ensemble of models that should be used for
#'   prediction. This can also be a path to the ensemble model, one or more
#'   paths to models, or a list of models.
#' @param newdata Data to which the models are fitted. `familiar` performs
#'   checks on the data to ensure that all features required for fitting the
#'   model are present, and no additional levels are present in categorical
#'   features. Unlike other `predict` methods, `newdata` cannot be missing in
#'   `familiar`, as training data are not stored with the models.
#' @param type Type of prediction made. The following values are directly
#'   supported:
#'
#'  * `default`: Default prediction, i.e. value estimates for `continuous`
#'   outcomes, predicted class probabilities and class for `binomial` and
#'   `multinomial` and the model response for `survival` outcomes.
#'
#'  * `survival_probability`: Predicts survival probabilities at the time
#'   specified by `time`. Only applicable to `survival` outcomes. Some models
#'   may not allow for predicting survival probabilities based on their
#'   response.
#'
#'  * `novelty`: Predicts novelty of each sample, which can be used for
#'   out-of-distribution detection.
#'
#'  * `risk_stratification`: Predicts the strata to which the data belongs. Only
#'   for `survival` outcomes.
#'
#'   Other values for type are passed to the fitting method of the actual
#'   underlying model. For example for generalised linear models (`glm`) `type`
#'   can be `link`, `response` or `terms` as well. Some of these model-specific
#'   prediction types may fail to return results if the model has been trimmed.
#'
#' @param time Time at which the response (`default`) or survival probability
#'   (`survival_probability`) should be predicted for `survival` outcomes. Some
#'   models have a response that does not depend on `time`, e.g. `cox`, whereas
#'   others do, e.g. `random_forest`.
#' @param dir_path Path to the folder containing the models. Ensemble objects
#'   are stored with the models detached. In case the models were moved since
#'   creation, `dir_path` can be used to specify the current folder.
#'   Alternatively the `update_model_dir_path` method can be used to update the
#'   path.
#' @param stratification_method Selects the stratification method from which the
#'   threshold values should be selected. If the model or ensemble of models
#'   does not contain thresholds for the indicated method, an error is returned.
#'   In addition this argument is ignored if a `stratification_threshold` is
#'   set.
#' @param stratification_threshold Threshold value(s) used for stratifying
#'   instances into risk groups. If this parameter is specified,
#'   `stratification_method` and any threshold values that come with the model
#'   are ignored, and `stratification_threshold` is used instead.
#' @param percentiles Currently unused.
#' @param .as_prediction_table Selects whether a data.table or prediction table
#'   object should be returned.
#' @param ... to be documented.
#'
#' @inheritParams .extract_data
#'
#' @details This method is used to predict values for instances specified by the
#'   `newdata` using the model or ensemble of models specified by the `object`
#'   argument.
#'
#' @return A `data.table` with predicted values.
#' @exportMethod predict
#' @md
#' @rdname predict-methods
setGeneric("predict")



# predict (model) --------------------------------------------------------------

#' @rdname predict-methods
setMethod(
  "predict",
  signature(object = "familiarModel"),
  function(
    object,
    newdata,
    type = "default",
    time = NULL,
    dir_path = NULL,
    ensemble_method = "median",
    stratification_threshold = NULL,
    stratification_method = NULL,
    percentiles = NULL,
    .as_prediction_table = FALSE,
    ...
  ) {
    # Create ensemble.
    object <- as_familiar_ensemble(object = object)

    # Create predictions.
    predictions <- predict(
      object = object,
      newdata = newdata,
      type = type,
      time = time,
      dir_path = dir_path,
      ensemble_method = ensemble_method,
      stratification_threshold = stratification_threshold,
      stratification_method = stratification_method,
      percentiles = percentiles,
      .as_prediction_table = .as_prediction_table,
      ...
    )

    return(predictions)
  }
)



# predict (ensemble) -----------------------------------------------------------

#' @rdname predict-methods
setMethod(
  "predict",
  signature(object = "familiarEnsemble"),
  function(
    object,
    newdata,
    type = "default",
    time = NULL,
    dir_path = NULL,
    ensemble_method = "median",
    stratification_threshold = NULL,
    stratification_method = NULL,
    percentiles = NULL,
    .as_prediction_table = FALSE,
    ...
  ) {
    if (missing(newdata)) rlang::abort("newdata must be provided.")
    if (is_empty(newdata)) ..error_data_set_is_empty()

    # Make sure the ensemble model object is updated.
    object <- update_object(object = object)

    # Parse newdata to data object
    data <- as_data_object(
      data = newdata,
      object = object,
      check_stringency = "external"
    )

    # Propagate to .predict
    predictions <- .predict(
      object = object,
      data = data,
      type = type,
      time = time,
      dir_path = dir_path,
      ensemble_method = ensemble_method,
      stratification_threshold = stratification_threshold,
      stratification_method = stratification_method,
      percentiles = percentiles
    )
    
    if (is(predictions, "familiarDataElementPredictionTable")) {
      # Only return prediction columns for external predict calls.
      
      # Complete the object by adding any missing information, such as the
      # predicted class.
      predictions <- .complete(predictions)
      
      # Ensure that values are ordered the same as in the input data.
      predictions@data <- merge(
        x = data@data[, mget(get_id_columns())],
        y = predictions@data,
        by = c(familiar:::get_id_columns()),
        all.x = TRUE,
        sort = FALSE
      )
      
      # Isolate predicted values.
      if (!.as_prediction_table) {
        predictions <- .as_data_table(predictions)[, mget(predictions@value_column)]
      }
    }

    return(predictions)
  }
)


# predict (novelty detector) ---------------------------------------------------

#' @rdname predict-methods
setMethod(
  "predict",
  signature(object = "familiarNoveltyDetector"),
  function(
    object,
    newdata,
    type = "novelty",
    ensemble_method = "median",
    .as_prediction_table = FALSE,
    ...
  ) {
    # Create ensemble.
    object <- as_familiar_ensemble(object = object)
    
    # Create predictions.
    predictions <- predict(
      object = object,
      newdata = newdata,
      type = type,
      time = time,
      ensemble_method = ensemble_method,
      .as_prediction_table = .as_prediction_table,
      ...
    )

    return(predictions)
  }
)



# predict (list) ---------------------------------------------------------------
#' @rdname predict-methods
setMethod(
  "predict",
  signature(object = "list"),
  function(
    object,
    newdata,
    type = "default",
    time = NULL,
    dir_path = NULL,
    ensemble_method = "median",
    stratification_threshold = NULL,
    stratification_method = NULL,
    percentiles = NULL,
    .as_prediction_table = FALSE,
    ...
  ) {
    # Create ensemble.
    object <- as_familiar_ensemble(object = object)

    # Create predictions.
    predictions <- predict(
      object = object,
      newdata = newdata,
      type = type,
      time = time,
      dir_path = dir_path,
      ensemble_method = ensemble_method,
      stratification_threshold = stratification_threshold,
      stratification_method = stratification_method,
      percentiles = percentiles,
      .as_prediction_table = .as_prediction_table,
      ...
    )
    
    return(predictions)
  }
)



# predict (character) ----------------------------------------------------------

#' @rdname predict-methods
setMethod(
  "predict",
  signature(object = "character"),
  function(
    object,
    newdata,
    type = "default",
    time = NULL,
    dir_path = NULL,
    ensemble_method = "median",
    stratification_threshold = NULL,
    stratification_method = NULL,
    percentiles = NULL,
    .as_prediction_table = FALSE,
    ...
  ) {
    # Create ensemble.
    object <- as_familiar_ensemble(object = object)
    
    # Create predictions.
    predictions <- predict(
      object = object,
      newdata = newdata,
      type = type,
      time = time,
      dir_path = dir_path,
      ensemble_method = ensemble_method,
      stratification_threshold = stratification_threshold,
      stratification_method = stratification_method,
      percentiles = percentiles,
      .as_prediction_table = .as_prediction_table,
      ...
    )
    
    return(predictions)
  }
)


# .predict (ensemble) ----------------------------------------------------------
setMethod(
  ".predict",
  signature(object = "familiarEnsemble"),
  function(
    object,
    data,
    allow_recalibration = TRUE,
    is_pre_processed = FALSE,
    time = NULL,
    type = "default",
    dir_path = NULL,
    ensemble_method = "median",
    stratification_threshold = NULL,
    stratification_method = NULL,
    percentiles = NULL,
    ...,
    aggregate_results = TRUE
  ) {
    # Predict function for a model ensemble. This will always return ensemble
    # information, and not details corresponding to the individual models.

    # Test if the models are loaded, and load the models if they are not.
    object <- load_models(object, dir_path = dir_path)
    
    # Extract predictions for each individual model
    if (length(object@model_list) > 0L) {
      predict_list <- lapply(
        object@model_list,
        .predict,
        data = data,
        allow_recalibration = allow_recalibration,
        is_pre_processed = is_pre_processed,
        time = time,
        type = type,
        ensemble_method = ensemble_method,
        stratification_threshold = stratification_threshold,
        stratification_method = stratification_method,
        percentiles = percentiles,
        ...
      )
      
    } else {
      # Generate a placeholder table.
      predict_list <- NULL
    }

    # Generate ensemble predictions
    if (all(type %in% .get_available_prediction_type_arguments())) {
      if (aggregate_results) {
        
        # Merge prediction data etc. slots into the data slot.
        predict_list <- lapply(predict_list, .merge_slots_into_data)
        
        # Compute ensemble values.
        predict_list <- .compute_data_element_estimates(
          x = predict_list,
          object = object
        )
        
        # Extract the merged object.
        predict_list <- predict_list[[1L]]
      }
    }
    
    return(predict_list)
  }
)


# .predict (model) -------------------------------------------------------------
setMethod(
  ".predict",
  signature(object = "familiarModel"),
  function(
    object,
    data,
    allow_recalibration = TRUE,
    is_pre_processed = FALSE,
    time = NULL,
    type = "default",
    ensemble_method = "median",
    stratification_threshold = NULL,
    stratification_method = NULL,
    percentiles = NULL,
    ...
  ) {

    if (length(type) != 1L) {
      ..error_reached_unreachable_code(paste0(
        "Only one type of prediction is expected as argument to .predict. ",
        "Found: ", paste_s(type)
      ))
    }
    
    # Prepare input data
    data <- process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = "clustering",
      keep_novelty = "novelty" %in% type
    )
    
    # user specified type ------------------------------------------------------
    if (!all(type %in% .get_available_prediction_type_arguments())) {
      # Select the first non-standard type.
      type <- setdiff(type, .get_available_prediction_type_arguments())[1L]

      # Select only features used by the model.
      data <- select_features(
        data = data,
        features = features_after_clustering(
          features = object@model_features,
          feature_info_list = object@feature_info
        )
      )
      
      # Call predict from ..predict method with the given type and the
      # unspecified extra arguments in ... .
      return(..predict(
        object = object,
        data = data,
        time = time,
        type = type,
        ...
      ))
    }

    # novelty ------------------------------------------------------------------
    if (type == "novelty") {
      # Predict instance novelty.
      return(.predict_novelty(
        object = object,
        data = data
      ))
    }
    
    # Keep only model features in data for the remaining predictions.
    data <- select_features(
      data = data,
      features = features_after_clustering(
        features = object@model_features,
        feature_info_list = object@feature_info
      )
    )
    
    if (type == "default") {
      # default ----------------------------------------------------------------
      
      # Predict using the model and the standard type.
      prediction_table <- ..predict(
        object = object,
        data = data,
        type = type,
        time = time
      )
      
      # Recalibrate the default predictions.
      if (allow_recalibration) {
        prediction_table <- .apply_recalibration(
          object = object,
          prediction_table = prediction_table,
          data = data
        )
      }
      
    } else if (type == "survival_probability") {
      # survival probability ---------------------------------------------------
      
      # Predict using the model and the standard type.
      prediction_table <- ..predict(
        object = object,
        data = data,
        type = type,
        time = time
      )
      
    } else if (type == "risk_stratification") {
      # risk stratification ----------------------------------------------------
      
      # Prediction survival probabilities,
      prediction_table <- .predict_risk_stratification(
        object = object,
        data = data,
        time = time,
        stratification_threshold = stratification_threshold,
        stratification_method = stratification_method,
        ...
      )
      
    } else {
      ..error_reached_unreachable_code(paste0(
        "Type of prediction was not recognised: ", type
      ))
    }
  
    if (is(prediction_table, "familiarDataElementPredictionTable")) {
      prediction_table@detail_level <- "ensemble"
      prediction_table@percentiles <- percentiles
      prediction_table@ensemble_method <- ensemble_method
      prediction_table@estimation_type <- ifelse(is.null(percentiles), "point", "bootstrap_confidence_interval")
    }
    
    return(prediction_table)
  }
)



# .predict (novelty detector) --------------------------------------------------
setMethod(
  ".predict",
  signature(object = "familiarNoveltyDetector"),
  function(
    object,
    data, 
    type = "novelty",
    is_pre_processed = FALSE,
    ...
  ) {
    # Prepare input data.
    data <- process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = "clustering"
    )
    
    # Predict novelty.
    prediction_table <- ..predict(
      object = object,
      data = data,
      type = type
    )

    return(prediction_table)
  }
)



# .predict (character) ---------------------------------------------------------
setMethod(
  ".predict",
  signature(object = "character"),
  function(object, data, ...) {
    # Load object.
    object <- load_familiar_object(object)

    return(do.call(
      .predict,
      args = c(
        list(
          "object" = object,
          "data" = data
        ),
        list(...)
      )
    ))
  }
)



# .predict (list) --------------------------------------------------------------
setMethod(
  ".predict",
  signature(object = "list"),
  function(object, data, ...) {
    if (all(sapply(object, is, class2 = "familiarModel"))) {
      object <- as_familiar_ensemble(object = object)
      
      return(do.call(
        .predict,
        args = c(
          list(
            "object" = object,
            "data" = data
          ),
          list(...)
        )
      ))
      
    } else {
      ..error_reached_unreachable_code(".predict for list expects a list of familiarModel objects")
    }
  }
)



# .predict_novelty (model) -----------------------------------------------------
setMethod(
  ".predict_novelty",
  signature(object = "familiarModel"),
  function(
    object,
    data, 
    type = "novelty",
    is_pre_processed = FALSE,
    ...
  ) {
    return(.predict(
      object = object@novelty_detector,
      data = data,
      type = type,
      is_pre_processed = is_pre_processed
    ))
  }
)



# .predict_novelty (character) -------------------------------------------------
setMethod(
  ".predict_novelty",
  signature(object = "character"),
  function(object, ...) {
    # Load object.
    object <- load_familiar_object(object)

    return(do.call(
      .predict_novelty,
      args = c(
        list("object" = object),
        list(...)
      )
    ))
  }
)



# .predict_risk_stratification (model) -----------------------------------------
setMethod(
  ".predict_risk_stratification",
  signature(object = "familiarModel"),
  function(
    object,
    data,
    time, 
    stratification_threshold = NULL,
    stratification_method = NULL, 
    ...
  ) {
    # Only assess stratification for survival outcomes.
    if (!object@outcome_type %in% c("survival")) return(NULL)
    
    # Allow for settings the stratification threshold explicitly.
    if (is.null(stratification_threshold)) {
      if (!is.null(stratification_method)) {
        # Select threshold values for the given method.

        # First check if the method exists.
        if (!stratification_method %in% object@km_info$stratification_method) {
          ..error(paste0(
            "Data for stratification method ", stratification_method,
            " was not found with the object."
          ))
        }

        # Select stratification threshold for the given method.
        stratification_threshold <- object@km_info$parameters[[stratification_method]]$cutoff
      } else {
        # Select stratification threshold for the first method.
        stratification_threshold <- object@km_info$parameters[[1L]]$cutoff
      }
    }

    # Check again if a stratification threshold is now set.
    if (is.null(stratification_threshold)) return(NULL)

    # Predict for the instances in data.
    predictions <- .predict(
      object = object,
      data = data,
      time = time
    )

    # Find risk groups for each instance.
    assigned_risk_group <- .apply_risk_threshold(
      x = predictions,
      cutoff = stratification_threshold
    )

    # Convert to prediction table
    prediction_table <- as_prediction_table(
      x = assigned_risk_group,
      type = "risk_stratification",
      data = data,
      time = time,
      model_object = object
    )
    
    return(prediction_table)
  }
)


# .predict_risk_stratification (naive model) -----------------------------------
setMethod(
  ".predict_risk_stratification",
  signature(object = "familiarNaiveModel"),
  function(
    object,
    data,
    time,
    stratification_threshold = NULL,
    stratification_method = NULL,
    ...
  ) {
    return(NULL)
  }
)



# .predict_risk_stratification (character) -------------------------------------
setMethod(
  ".predict_risk_stratification",
  signature(object = "character"),
  function(object, ...) {
    # Load object.
    object <- load_familiar_object(object)

    return(do.call(
      .predict_risk_stratification, 
      args = c(
        list("object" = object),
        list(...)
      )
    ))
  }
)



.get_available_ensemble_prediction_methods <- function() {
  return(c("median", "mean"))
}

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# AUC-ROC plots ----------------------------------------------------------------

familiar:::test_plots(
  plot_function = familiar::plot_auc_roc_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  debug = debug_flag
)

# Test with step-wise confidence interval
familiar:::test_plots(
  plot_function = familiar::plot_auc_roc_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  plot_args = list("conf_int_style" = "step"),
  debug = debug_flag
)

# Test without confidence interval shown
familiar:::test_plots(
  plot_function = familiar::plot_auc_roc_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  plot_args = list("conf_int_style" = "none"),
  debug = debug_flag
)

# Test without confidence interval
familiar:::test_plots(
  plot_function = familiar::plot_auc_roc_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  estimation_type = "point",
  test_config = "normal",
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar::plot_auc_roc_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar::plot_auc_roc_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "facet_by" = c("vimp_method", "learner"),
    "color_by" = c("data_set", "positive_class")),
  debug = debug_flag
)

# AUC precision-recall plots ---------------------------------------------------

familiar:::test_plots(
  plot_function = familiar::plot_auc_precision_recall_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  debug = debug_flag
)

# Test with step-wise confidence interval
familiar:::test_plots(
  plot_function = familiar::plot_auc_precision_recall_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  plot_args = list("conf_int_style" = "step"),
  debug = debug_flag
)

# Test without confidence interval shown
familiar:::test_plots(
  plot_function = familiar::plot_auc_precision_recall_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  plot_args = list("conf_int_style" = "none"),
  debug = debug_flag
)

# Test without confidence interval
familiar:::test_plots(
  plot_function = familiar::plot_auc_precision_recall_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  estimation_type = "point",
  test_config = "normal",
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar::plot_auc_precision_recall_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar::plot_auc_roc_curve,
  data_element = "auc_data",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list(
    "facet_by" = c("vimp_method", "learner"),
    "color_by" = c("data_set", "positive_class")),
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# familiarTaskTrainNovelty -----------------------------------------------------
setClass(
  "familiarTaskTrainNovelty",
  contains = "familiarTask",
  slots = list(
    "vimp_method" = "character",
    "learner" = "character",
    "vimp_table_file" = "character",
    "hyperparameter_file" = "character",
    "feature_info_file" = "character"
  ),
  prototype = methods::prototype(
    vimp_method = "none",
    learner = NA_character_,
    vimp_table_file = NA_character_,
    hyperparameter_file = NA_character_,
    feature_info_file = NA_character_,
    task_name = "train_novelty_detector"
  )
)



# .get_task_descriptor (train task) --------------------------------------------
setMethod(
  ".get_task_descriptor",
  signature(object = "familiarTaskTrainNovelty"),
  function(object, ...) {
    return(paste0(
      object@task_name, "_",
      object@data_id, "_", 
      object@run_id, "_", 
      object@vimp_method, "_", 
      object@learner
    ))
  }
)



# .perform_task (train task , NULL) --------------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskTrainNovelty",
    data = "NULL"
  ),
  function(
    object,
    data,
    experiment_data = NULL,
    outcome_info = NULL,
    ...
  ) {
    # This method is called when "data" is expected to be available somewhere in
    # the backend.
    
    if (is.null(experiment_data)) {
      ..error_reached_unreachable_code("experiment_data is required for retrieving data from the backend.")
    }
    if (is.null(outcome_info)) {
      ..error_reached_unreachable_code("outcome_info is required.")
    }
    
    # Find the run list.
    run_list <- .get_run_list(
      iteration_list = experiment_data@iteration_list,
      data_id = object@data_id,
      run_id = object@run_id
    )
    
    # Select unique samples.
    sample_identifiers <- .get_sample_identifiers(
      run = run_list,
      train_or_validate = "train"
    )
    sample_identifiers <- unique(sample_identifiers)
    
    # Create a dataObject.
    data <- methods::new(
      "dataObject",
      data = get_data_from_backend(sample_identifiers = sample_identifiers),
      preprocessing_level = "none",
      outcome_type = outcome_info@outcome_type,
      outcome_info = outcome_info
    )
    
    # Pass to method that dispatches with dataObject for further processing.
    return(.perform_task(
      object = object,
      data = data,
      experiment_data = experiment_data,
      ...
    ))
  }
)


# .perform_task (train task, dataObject) ---------------------------------------
setMethod(
  ".perform_task",
  signature(
    object = "familiarTaskTrainNovelty",
    data = "dataObject"
  ),
  function(
    object,
    data,
    selected_features = NULL,
    vimp_aggregation_method = NULL,
    vimp_rank_threshold = NULL,
    settings = NULL,
    feature_info_list = NULL,
    vimp_table = NULL,
    hyperparameters = NULL,
    message_indent = 0L,
    verbose = FALSE,
    cl = NULL,
    return_results = TRUE,
    ...
  ) {
    logger_message(
      paste0(
        "Training: Starting training for the \"", object@learner,
        "\" out-of-distribution detector and the \"", object@vimp_method,
        "\" variable importance method for run ",
        object@task_id, " of ",
        object@n_tasks, "."
      ),
      indent = message_indent,
      verbose = verbose
    )
    
    # Check and retrieve feature info list.
    feature_info_list <- .get_feature_info_list(
      object = object,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings,
      message_indent = message_indent,
      verbose = verbose,
      cl = cl,
      ...
    )
    
    # Set vimp aggregation method and vimp_rank_threshold based on settings.
    if (!is.null(settings)) {
      if (is.null(vimp_aggregation_method)) {
        vimp_aggregation_method <- settings$vimp$aggregation
      }
      if (is.null(vimp_rank_threshold)) {
        vimp_rank_threshold <- settings$vimp$aggr_rank_threshold
      }
    }
    
    # Check and retrieve hyperparameters. We do this prior to retrieving the
    # variable importance tables, as these may be attached to hyperparameter
    # object.
    hyperparameters <- .get_hyperparameters(
      object = object,
      selected_features = selected_features,
      hyperparameters = hyperparameters,
      vimp_aggregation_method = vimp_aggregation_method,
      vimp_rank_threshold = vimp_rank_threshold,
      feature_info_list = feature_info_list,
      data = data,
      settings = settings,
      message_indent = message_indent,
      verbose = verbose,
      cl = cl,
      ...
    )
    
    # If selected features are not provided, create the variable importance
    # table.
    if (is.null(selected_features)) {
      if (is_empty(hyperparameters$vimp_table)) {
        # Check and retrieve variable importances from the drive, or generate in
        # place, if the hyperparameter object did not contain a variable
        # importance table.
        vimp_table <- .get_variable_importance_table(
          object = object,
          vimp_table = vimp_table,
          feature_info_list = feature_info_list,
          data = data,
          settings = settings,
          message_indent = message_indent,
          verbose = verbose,
          cl = cl,
          ...
        )
        
      } else {
        vimp_table <- hyperparameters$vimp_table
      }

    } else {
      vimp_table <- NULL
    }
    
    # Create the raw model object for training.
    model_object <- methods::new(
      "familiarNoveltyDetector",
      hyperparameters = hyperparameters$hyperparameters,
      learner = object@learner,
      vimp_method = object@vimp_method,
      vimp_table = vimp_table,
      vimp_aggregation_method = vimp_aggregation_method,
      vimp_rank_threshold = vimp_rank_threshold,
      feature_info = feature_info_list,
      data_id = object@data_id,
      run_id = object@run_id,
      run_table = .get_current_run_table(object = object),
      project_id = object@project_id
    )

    # Promote to the correct type of detector.
    model_object <- promote_detector(object = model_object)
    
    # Select features based on variable importances.
    model_object <- set_model_features(
      object = model_object,
      minimise_footprint = FALSE,
      signature_features = selected_features
    )
    
    # Process data, if required.
    data <- process_input_data(
      object = model_object,
      data = data,
      stop_at = "clustering",
      force_check = TRUE
    )
    
    # Train the novelty detector.
    model_object <- .train(
      object = model_object,
      data = data,
      get_additional_info = TRUE,
      ...
    )
    
    # Add model name
    model_object <- set_object_name(model_object)
    
    if (!is.na(object@file)) {
      saveRDS(model_object, file = object@file)
    }
    
    if (return_results) {
      return(model_object)
    }
    
    return(TRUE)
  }
)



# .get_hyperparameters (train task) --------------------------------------------
setMethod(
  ".get_hyperparameters",
  signature(object = "familiarTaskTrainNovelty"),
  function(
    object,
    hyperparameters,
    file_paths = NULL,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    can_pre_process <- NULL
    
    hyperparameter_object <- NULL
    if (is.null(hyperparameters) && !is.null(object@run_table)) {
      # This routine loads hyperparameters from disk, and is used when an
      # experiment is run using summon_familiar.
      
      # This check exists to make sure that the standard workflow passes the
      # correct objects.
      if (is.null(file_paths)) {
        ..error_reached_unreachable_code("file_paths was expected, but not provided.")
      }
      
      # Find the last entry on the run table that is marked as available for
      # pre-processing. This is what hyperparameters are based on.
      hyperparameter_run <- tail(
        object@run_table[[paste0(object@data_id, ".", object@run_id)]][can_pre_process == TRUE, ],
        n = 1L
      )
      
      # Find the file name.
      hyperparameter_file <- get_object_file_name(
        project_id = object@project_id,
        data_id = hyperparameter_run$data_id[1L],
        run_id = hyperparameter_run$run_id[1L],
        learner = object@learner,
        vimp_method = object@vimp_method,
        object_type = "hyperparametersNoveltyDetector",
        dir_path = file_paths$mb_dir
      )
      
      if (file.exists(hyperparameter_file)) {
        hyperparameter_object <- update_object(readRDS(hyperparameter_file))
      }
    }
    
    
    if (is.null(hyperparameter_object) && is.na(object@hyperparameter_file)) {
      # Create an ad-hoc list of hyperparameters
      
      # Set up task, and explicitly don't write to file.
      hyperparameter_task <- methods::new(
        "familiarTaskNoveltyDetectorHyperparameters",
        project_id = object@project_id,
        vimp_method = object@vimp_method,
        learner = object@learner,
        file = NA_character_
      )
      
      # Execute the task.
      hyperparameter_object <- .perform_task(
        object = hyperparameter_task,
        hyperparameters = hyperparameters,
        ...
      )
      
    } else if (is.null(hyperparameter_object)) {
      # Assume that the hyperparameter_file attribute contains the path to the
      # file containing the vimp method hyperparameters.
      if (!file.exists(object@hyperparameter_file)) {
        ..error(paste0("hyperparameter file does not exist at location: ", object@hyperparameter_file))
      }
      hyperparameter_object <- update_object(readRDS(object@hyperparameter_file))
      
    } else if (is.character(hyperparameters)) {
      # If hyperparameters is a string, interpret this as a path to the
      # file containing the vimp method hyperparameters.
      if (!file.exists(hyperparameters)) {
        ..error(paste0("hyperparameter file does not exist at location: ", hyperparameters))
      }
      hyperparameter_object <- update_object(readRDS(hyperparameters))
    }
    
    if (is(hyperparameter_object, "familiarNoveltyDetector")) {
      hyperparameters <- list(
        "hyperparameters" =  hyperparameter_object@hyperparameters,
        "hyperparameter_data" = hyperparameter_object@hyperparameter_data,
        "vimp_table" = hyperparameter_object@vimp_table
      )
      
    } else {
      hyperparameters <- list("hyperparameters" = hyperparameters)
    }
    
    return(hyperparameters)
  }
)

# test_create_vimp_method (data.table) -----------------------------------------
setMethod(
  "test_create_vimp_method",
  signature(data = "data.table"),
  function(
    data, 
    vimp_method, 
    vimp_method_parameter_list = list(),
    ...
  ) {
    # Convert data to dataObject.
    data <- do.call(
      as_data_object, 
      args = c(
        list("data" = data),
        list(...)
      )
    )
    
    return(do.call(
      test_create_vimp_method, 
      args = c(
        list(
          "data" = data,
          "vimp_method" = vimp_method,
          "vimp_method_parameter_list" = vimp_method_parameter_list
        ),
        list(...)
      )
    ))
  }
)



# test_create_vimp_method (dataObject) -----------------------------------------
setMethod(
  "test_create_vimp_method",
  signature(data = "dataObject"),
  function(
    data,
    data_bypass = NULL,
    vimp_method, 
    vimp_method_parameter_list = list(), 
    ...
  ) {
    # The bypass data allows for bypassing important aspects of the
    # pre-processing pipeline, e.g. the preprocessing checks. This enables
    # testing of very rare cases where preprocessing may run fine, but the
    # subsample does not allow for training.
    if (is.null(data_bypass)) data_bypass <- data
    
    # Prepare setting ----------------------------------------------------------
    
    # Reconstitute settings from the data.
    settings <- extract_settings_from_data(data)
    
    # Update some missing settings that can be fixed within this method.
    settings$data$train_cohorts <- unique(data@data[[get_id_columns(single_column = "batch")]])
    
    # Parse the remaining settings that are important. Remove outcome_type from
    # ... This prevents an error caused by multiple matching arguments.
    dots <- list(...)
    dots$parallel <- NULL
    dots$vimp_method <- NULL
    
    if (!is.null(dots$signature)) settings$data$signature <- dots$signature
    
    settings <- do.call(
      .parse_general_settings,
      args = c(
        list(
          "settings" = settings,
          "data" = data_bypass@data,
          "parallel" = FALSE,
          "vimp_method" = vimp_method,
          "learner" = "glm"
        ),
        dots
      )
    )
    
    # Push settings to the backend.
    .assign_settings_to_global(settings = settings)
    
    
    # Prepare hyperparameters --------------------------------------------------
    
    # Get default hyperparameters.
    param_list <- .get_preset_hyperparameters(
      data = data,
      vimp_method = vimp_method,
      names_only = FALSE
    )
    
    # Update with user-provided settings.
    param_list <- .update_hyperparameters(
      parameter_list = param_list,
      user_list = vimp_method_parameter_list
    )
    
    # Determine which hyperparameters still need to be specified.
    unset_parameters <- sapply(
      param_list,
      function(hyperparameter_entry) hyperparameter_entry$randomise
    )
    
    # Mark sign-size as set, as it is not used for variable importance.
    if ("sign_size" %in% names(unset_parameters)) {
      unset_parameters["sign_size"] <- FALSE
    }
    
    # Raise an error if any hyperparameters were not set.
    if (any(unset_parameters)) {
      ..error(paste0(
        "The following hyperparameters need to be specified: ",
        paste_s(names(unset_parameters)[unset_parameters])
      ))
    }
    
    # Obtain the final list of hyperparameters.
    param_list <- lapply(
      param_list, 
      function(hyperparameter_entry) hyperparameter_entry$init_config
    )
    
    
    # Create feature information list ------------------------------------------
    
    feature_info_task <- methods::new(
      "familiarTaskFeatureInfo"
    )
    
    # Feature information objects are created from the bypass dataset.
    feature_info <- .perform_task(
      object = feature_info_task,
      data = data_bypass,
      settings = settings
    )
    
    
    # Prepare vimp object ------------------------------------------------------
    
    # Get required features.
    required_features <- get_required_features(
      x = data,
      feature_info_list = feature_info
    )
    
    # Create a familiar variable importance method.
    object <- methods::new(
      "familiarVimpMethod",
      outcome_type = data@outcome_type,
      vimp_method = vimp_method,
      hyperparameters = param_list,
      outcome_info = data@outcome_info,
      feature_info = feature_info[required_features],
      required_features = required_features
    )
    
    # Promote object to correct subclass.
    object <- promote_vimp_method(object)
    
    # Return in list.
    return(object)
  }
)

.add_package_version <- function(object) {
  # Adds the version of the familiar package used to generate the object. This
  # allows for backward compatibility.

  # Look-up of the familiar version takes more time than one would expect. Use
  # caching.
  current_version <- rlang::env_cache(
    familiar_global_env,
    "familiar_version",
    default = utils::packageVersion("familiar")
  )
  
  if (is.null(object@familiar_version)) {
    # Set package version.
    object@familiar_version <- current_version
    
  } else if (tail(object@familiar_version, n = 1L) < current_version) {
    # Check if package version differs from the currently installed version.
    # This is usually done when updating the object.
    object@familiar_version <- c(
      object@familiar_version,
      current_version
    )
  }

  return(object)
}



.familiar_version_string <- function(object) {
  # Set version string.
  version_string <- paste0("v", head(object@familiar_version, n = 1L))

  # Add version string if the object was updated from a previous version.
  if (length(object@familiar_version) > 1L) {
    version_string <- paste0(
      version_string,
      " -> ",
      tail(object@familiar_version, n = 1L)
    )
  }

  return(version_string)
}



.save <- function(object, dir_path) {
  # Saves the object to the disk. Applicable to familiarModel, familiarEnsemble,
  # and familiarData objects.

  # Generate directory path
  if (is(object, "familiarModel")) {
    object_type <- "familiarModel"
  } else if (is(object, "familiarEnsemble")) {
    object_type <- "familiarEnsemble"
  } else if (is(object, "familiarData")) {
    object_type <- "familiarData"
  } else if (is(object, "familiarCollection")) {
    object_type <- "familiarCollection"
  } else {
    ..error_reached_unreachable_code(".save: unknown type of object encountered.")
  }

  # Add file name.
  if (is.null(object@project_id) && length(object@name) > 0L) {
    file_name <- object@name
  } else {
    file_name <- get_object_name(object = object)
  }

  # Add file extension
  file_name <- paste0(file_name, ".RDS")

  # Obtain the directory path
  dir_path <- get_object_dir_path(
    dir_path = dir_path,
    object_type = object_type,
    learner = object@learner,
    vimp_method = object@vimp_method
  )

  # Check if the directory exists, and create anew if not
  if (!dir.exists(dir_path)) dir.create(dir_path, recursive = TRUE)

  # Add package version
  object <- add_package_version(object = object)

  # Save to disk
  saveRDS(object, file = file.path(dir_path, file_name))
}



# has_bad_training_data --------------------------------------------------------
setMethod(
  "has_bad_training_data", 
  signature(
    object = "ANY",
    data = "dataObject"
  ),
  function(
    object,
    data,
    allow_no_features = FALSE,
    ...
  ) {
    # Checks the data for consistency and usability. Any errors are passed as
    # attributes

    if (!(
      is(object, "familiarModel") ||
      is(object, "familiarVimpMethod") ||
      is(object, "familiarNoveltyDetector")
    )) {
      ..error_reached_unreachable_code(paste0(
        "has_bad_training_data: object is not a familiarModel, ",
        "familiarVimpMethod or familiarNoveltyDetector."
      ))
    }

    # One cannot train without data or on a single sample.
    if (is_empty(data, allow_no_features = allow_no_features)) {
      return_value <- TRUE
      attr(return_value, "error") <- ..error_message_no_training_data_available()

      return(return_value)
    }

    if (get_n_samples(data) < 2L) {
      return_value <- TRUE
      attr(return_value, "error") <- paste0(
        "Only one sample was available to train the model."
      )

      return(return_value)
    }

    # For familiarNoveltyDetector objects outcome is not important.
    if (is(object, "familiarNoveltyDetector")) {
      return(FALSE)
    }

    # Retrieve outcomeInfo object.
    if (!is.null(object@outcome_info)) {
      outcome_info <- object@outcome_info
    } else if (!is.null(data@outcome_info)) {
      outcome_info <- data@outcome_info
    } else {
      ..error_reached_unreachable_code(paste0(
        "has_bad_training_data: could not find outcomeInfo object attached to ",
        "familiarModel/familiarVimpMethod or dataObject."
      ))
    }

    if (object@outcome_type == "survival") {
      # Check that not all data are censored.
      censoring_variable <- outcome_info@censored
      if (length(censoring_variable) > 0L) {
        if (all(data@data$outcome_event == censoring_variable)) {
          return_value <- TRUE
          attr(return_value, "error") <- paste0(
            "All instances in the data set were censored. ",
            "Events are required for training the model."
          )

          return(return_value)
        }
      }

      # The same as above.
      if (all(data@data$outcome_event == 0L)) {
        return_value <- TRUE
        attr(return_value, "error") <- paste0(
          "All instances in the data set were censored. ",
          "Events are required for training the model."
        )

        return(return_value)
      }

      # Check that not all data have the same survival time.
      if (all(data@data$outcome_time == data@data$outcome_time[1L])) {
        return_value <- TRUE
        attr(return_value, "error") <- paste0(
          "All instances in the data set had the same recorded survival time."
        )

        return(return_value)
      }
      
    } else if (object@outcome_type %in% c("binomial", "multinomial")) {
      # Check that not all data have the same class.
      if (data.table::uniqueN(data@data$outcome) == 1L) {
        return_value <- TRUE
        attr(return_value, "error") <- paste0(
          "All instances in the data set were of the same class."
        )

        return(return_value)
      }

      # Check that all classes are present at least once.
      if (data.table::uniqueN(data@data$outcome) < nlevels(data@data$outcome)) {
        return_value <- TRUE
        attr(return_value, "error") <- paste0(
          "Some expected classes were not found in the data set."
        )

        return(return_value)
      }
      
    } else if (object@outcome_type %in% c("continuous")) {
      # Check that not all data have the same outcome value.
      if (all(data@data$outcome == data@data$outcome[1L])) {
        return_value <- TRUE
        attr(return_value, "error") <- paste0(
          "All instances in the data set had the same outcome value."
        )

        return(return_value)
      }
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }

    return(FALSE)
  }
)



.why_bad_training_data <- function(object, reason) {
  # Determine the reason why training data was bad.

  # If the reason was FALSE, return the object as is.
  if (!reason) return(object)

  # Check that object is a familiarModel, because otherwise we have nowhere to
  # store the reason.
  if (!(is(object, "familiarModel"))) {
    ..error_reached_unreachable_code(
      ".why_bad_training_data: object is not a familiarModel."
    )
  }

  # Store error messages and return the object.
  return(..update_errors(object = object, attr(reason, "error")))
}

.check_dots_is_parameter <- function(dots, call = rlang::caller_env()) {
  
  if (length(dots) > 0L) {
    # Find unmatched arguments.
    unmatched_args <- setdiff(dots, .get_all_parameter_names())
    
    if (length(unmatched_args) > 0L) {
      ..error(
        paste0(
          "Configuration: one or more function arguments could not be matched ",
          "to arguments passed by summon_familiar as configuration parameters: ",
          paste_s(unmatched_args),
          "\nThese arguments may have been misspelled, or were deprecated or renamed."
        ),
        error_class = "input_argument_error",
        call = call
      )
    }
  }
  
  return(invisible(TRUE))
}



.check_configuration_tag_is_parameter <- function(config, call = rlang::caller_env()) {
  
  if (!is.null(config)) {
    # Find names of parent nodes.
    config_node_names <- names(config)
    
    # Find nodes that are specified differently by the user.
    unmatched_node_names <- setdiff(
      config_node_names,
      .get_all_configuration_parent_node_names()
    )
    
    if (length(unmatched_node_names) > 0L) {
      ..error(
        paste0(
          "Configuration: one or more parent nodes in the configuration file could not be matched ",
          "to node names used by summon_familiar to group configuration parameters: ",
          paste_s(unmatched_node_names),
          "\nThese node names may have been misspelled, or were deprecated or renamed."
        ),
        error_class = "input_argument_error",
        call = call
      )
    }
    
    # Find names of configuration arguments.
    config_args <- unique(unlist(sapply(config, names)))
    
    # Find unmatched arguments.
    unmatched_args <- setdiff(config_args, .get_all_parameter_names())
    
    if (length(unmatched_args) > 0L) {
      ..error(paste0(
        "Configuration: one or more parameters set in the configuration file could not be matched ",
        "to arguments passed by summon_familiar as configuration parameters: ",
        paste_s(unmatched_args),
        "\nThese parameters may have been misspelled, or were deprecated or renamed."
      ),
      error_class = "input_argument_error",
      call = call
      )
    }
  }
  
  return(invisible(TRUE))
}



.check_is_numeric <- function(
    x,
    var_name,
    call = rlang::caller_env()
) {
  if (!is.numeric(x)) {
    ..error_type_not_valid(
      x = x,
      var_name = var_name,
      valid_type = "numeric",
      call = call
    )
  }
  
  return(x)
}



.check_is_integer <- function(
    x,
    var_name,
    call = rlang::caller_env()
) {
  if (!is.integer(x)) {
    ..error_type_not_valid(
      x = x,
      var_name = var_name,
      valid_type = "integer",
      call = call
    )
  }
}



.check_is_integerish <- function(
    x,
    var_name,
    call = rlang::caller_env()
) {
  # Less strict than .check_is_integer, i.e. the check would return pass for
  # 2.0, but fail for 2.5 or anything not numeric.
  if (!rlang::is_integerish(x)) {
    ..error_type_not_valid(
      x = x,
      var_name = var_name,
      valid_type = "integer or numeric castable to integer (e.g. 3.0)",
      call = call
    )
  }
}



.check_number_in_valid_range <- function(
    x,
    var_name,
    range,
    closed = c(TRUE, TRUE),
    call = rlang::caller_env()
) {

  # Interpret single input range value as range containing only one value.
  if (length(range) == 1L) range <- c(range, range)
  
  # Set lower limit to -Inf
  if (is.na(range[1L])) range[1L] <- -Inf
  
  # Set upper limit to Inf
  if (is.na(range[2L])) range[2L] <- Inf

  # Some internal checks that should never be triggered.
  if (length(x) != 1L) {
    ..error_reached_unreachable_code(paste0(
      ".check_number_in_valid_range: x does not have length 1."
    ))
  }
  
  # Another internal checks that should never be triggered.
  if (range[2L] - range[1L] < 0.0) {
    ..error_reached_unreachable_code(paste0(
      ".check_number_in_valid_range: the range is inverted"
    ))
  }
  
  # Check that x is numeric or NA.
  if (!is.numeric(x) && !is.na(x)) {
    ..error_type_not_valid(
      x = x,
      var_name = var_name,
      valid_type = "numeric",
      call = call
    )
  }
  
  if (!is.na(x)) {
    is_outside_range <- ifelse(
      closed[1L],
      x < range[1L],
      x <= range[1L]
    ) || ifelse(
      closed[2L],
      x > range[2L],
      x >= range[2L]
    )
    
  } else {
    # NA-values are outside the valid range.
    is_outside_range <- TRUE
  }
  
  if (is_outside_range) {
    ..error_value_outside_allowed_range(
      x = x,
      var_name = var_name,
      range = range,
      call = call
    )
  }
  
  return(invisible(TRUE))
}



.check_value_not_shared <- function(
    x,
    y,
    var_name_x,
    var_name_y,
    call = rlang::caller_env()
) {
  # If either or both are NULL, return NULL
  if (is.null(x) || is.null(y)) return(NULL)
  
  overlap_values <- intersect(x, y)
  if (length(overlap_values) > 0L) {
    ..error_value_shared_between_variables(
      x = x,
      y = y,
      var_name_x = var_name_x,
      var_name_y = var_name_y,
      call = call
    )
  }
  
  return(invisible(TRUE))
}



.check_argument_length <- function(
    x,
    var_name,
    min = 0L,
    max = Inf,
    call = rlang::caller_env()
) {
  if (length(x) < min) {
    ..error_variable_has_too_few_values(
      x = x,
      var_name = var_name,
      req_length = c(min, max),
      call = call
    )
    
  } else if (length(x) > max) {
    ..error_variable_has_too_many_values(
      x = x,
      var_name = var_name,
      req_length = c(min, max),
      call = call
    )
  }
  
  return(invisible(TRUE))
}



.check_parameter_value_is_valid <- function(
    x,
    var_name,
    values,
    call = rlang::caller_env()
) {

  # Check if NULL is an allowed value
  null_allowed <- any(is.null(values))
  
  if (length(x) == 1L) {
  
    # Check if x is NULL
    if (is.null(x) && null_allowed) {
      # If x is NULL and this is allowed, return to parent function
      return(invisible(TRUE))
    }
  }
  
  if (any(is.null(x)) && !null_allowed) {
    # If any of x is NULL and this is not allowed, raise an error.
    ..error_value_not_allowed(
      x = x,
      var_name = var_name,
      values = values,
      call = call
    )
  }
  
  if (!all(x %in% values)) {
    # Throw an error if any x is not among the valid values
    ..error_value_not_allowed(
      x = x,
      var_name = var_name,
      values = values,
      call = call
    )
  }
  
  return(invisible(TRUE))
}



.perform_type_conversion <- function(
    x,
    to_type,
    var_name,
    req_length,
    allow_more = FALSE
) {

  # Specify conversion functions
  if (to_type %in% c("character", "factor")) {
    conv_function <- as.character
    
  } else if (to_type == "numeric") {
    conv_function <- as.numeric
    
  } else if (to_type == "integer") {
    conv_function <- as.integer
    
  } else if (to_type == "logical") {
    conv_function <- as.logical
    
  } else if (to_type == "list") {
    conv_function <- as.list
    
  } else {
    ..error_reached_unreachable_code(paste0(
      ".perform_type_conversion: the to_type argument was not recognised: ", to_type
    ))
  }
  
  # Attempt conversion
  x <- tryCatch(
    conv_function(x),
    warning = function(war) {
      ..error_type_conversion_not_possible(
        x = x,
        to_type = to_type,
        var_name = var_name,
        req_length = req_length,
        allow_more = allow_more
      )
    },
    error = function(err) {
      ..error_type_conversion_not_possible(
        x = x,
        to_type = to_type,
        var_name = var_name,
        req_length = req_length,
        allow_more = allow_more
      )
    }
  )
  
  # Test length of resulting vector
  if (length(x) < req_length) {
    ..error_variable_has_too_few_values(
      x = x,
      var_name = var_name,
      req_length = req_length,
      allow_more = allow_more
    )
    
  } else if (length(x) > req_length && !allow_more) {
    ..error_variable_has_too_many_values(
      x = x,
      var_name = var_name,
      req_length = req_length,
      allow_fewer = FALSE
    )
  }
  
  return(x)
}



.parse_arg <- function(
    x_config,
    x_var = waiver(),
    var_name,
    type,
    optional = FALSE,
    default = NULL
) {

  # There are two options for parsing a value:
  # 1. Using a configuration file (through x_config)
  # 2. Using a command-line variable (through x_var and ...)
  
  # If the variable with name x_var is provided in ..., this variable takes
  # precedence over x_config. Multiple x_var entries can be provided. The first
  # valid one is selected.
  if (!is.waive(x_var)) {
    # Variable set through a function argument
    x <- x_var
    
  } else if (!is.null(x_config)) {
    # Variable set through a configuration list
    x <- x_config
    
    # Trim whitespace and split variables
    if (type %in% c("character_list", "numeric_list", "integer_list", "logical_list") && length(x) > 0L) {
      # Divide by comma
      x <- strsplit_all(
        x = x,
        split = ",",
        fixed = TRUE
      )[[1L]]

      # Remove whitespace
      x <- gsub(
        x = x,
        pattern = " ",
        replacement = "",
        fixed = TRUE
      )
    }
    
  } else if (optional) {
    # Return default value for optional values that have not been set
    # externally.
    return(default)
    
  } else {
    # Required, but no default.
    ..error_input_missing_without_default(
      var_name = var_name,
      allow_config = TRUE
    )
  }
  
  # This point in the code can only be reached if the user provided a value.  
  # Check for presence of values
  if (length(x) == 0L && optional == FALSE) {
    # Throw an error as entry is required
    ..error_input_missing_without_default(
      var_name = var_name,
      allow_config = TRUE
    )
    
  } else if (isTRUE(all.equal("", x)) && optional == FALSE) {
    ..error_input_missing_without_default(
      var_name = var_name,
      allow_config = TRUE
    )
    
  } else if (length(x) == 0L || isTRUE(all.equal("", x))) {
    # Return the default value if no value is provided
    return(default)
  }

  # Convert logical and logical_list
  if (type %in% c("logical", "logical_list")) {
    # Replace entries
    x[tolower(x) %in% c("t", "true", "y", "yes", "1")] <- TRUE
    x[tolower(x) %in% c("f", "false", "n", "no", "0")] <- FALSE
  }
  
  if (type == "list") {
    # Convert to list. Note that lists can be empty.
    x <- .perform_type_conversion(
      x = x,
      to_type = type,
      var_name = var_name,
      req_length = 0L,
      allow_more = TRUE
    )
    
  } else if (type %in% c("character", "numeric", "integer", "logical")) {
    # Convert to type
    x <- .perform_type_conversion(
      x = x,
      to_type = type,
      var_name = var_name,
      req_length = 1L,
      allow_more = FALSE
    )
    
  } else if (type %in% c("character_list", "numeric_list", "integer_list", "logical_list")) {
    # Find basic type string.
    list_type <- sub_last(
      x = type,
      pattern = "_list",
      replacement = "",
      fixed = TRUE
    )
    
    # Convert to type
    x <- .perform_type_conversion(
      x = x,
      to_type = list_type,
      var_name = var_name,
      req_length = 1L,
      allow_more = TRUE
    )
    
    # Check for duplicates
    if (anyDuplicated(x)) {
      ..error_input_not_unique(
        x = x,
        var_name = var_name,
        allow_config = TRUE
      )
    }
    
  } else {
    # By design unreachable
    ..error_reached_unreachable_code(
      paste0(".parse_arg: the type argument was not recognised: ", type)
    )
  }
  
  return(x)
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# familiarNaiveModel objects ---------------------------------------------------
setClass(
  "familiarNaiveModel",
  contains = "familiarModel"
)

setClass(
  "familiarNaiveCoxModel",
  contains = "familiarNaiveModel"
)

setClass(
  "familiarNaiveSurvivalTimeModel",
  contains = "familiarNaiveModel"
)

setClass(
  "familiarNaiveCumulativeIncidenceModel",
  contains = "familiarNaiveModel"
)



# ..train (familiarNaiveModel) -------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarNaiveModel",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Check if training data is ok.
    if (reason <- has_bad_training_data(
      object = object,
      data = data, 
      allow_no_features = TRUE
    )) {
      
      return(callNextMethod(object = .why_bad_training_data(
        object = object,
        reason = reason
      )))
    }
    
    # Check if hyperparameters are set.
    if (is.null(object@hyperparameters)) {
      return(callNextMethod(object = ..update_errors(
        object = object, 
        ..error_message_no_optimised_hyperparameters_available()
      )))
    }
    
    if (object@outcome_type %in% c("continuous")) {
      model <- object@outcome_info@distribution
    } else if (object@outcome_type %in% c("binomial", "multinomial")) {
      model <- object@outcome_info@distribution
    } else if (object@outcome_type %in% c("survival")) {
      model <- object@outcome_info@distribution
    } else {
      ..error_outcome_type_not_implemented(object@outcome_type)
    }
    
    # Add model... well "model" really.
    object@model <- model
    
    # Set learner version
    object <- set_package_version(object)
    
    return(object)
  }
)



# ..predict (familiarNaiveModel) -----------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarNaiveModel",
    data = "dataObject"
  ),
  function(
    object,
    data,
    type = "default",
    time = NULL,
    ...
  ) {

    # Check n_samples, and then ensure that n_samples is equal to the number
    # of rows in the dataset.
    n_samples <- get_n_samples(data)
    if (n_samples == 0L) return(callNextMethod())
    n_samples <- nrow(data@data)
    
    if (type == "default") {
      # default ----------------------------------------------------------------
      
      if (object@outcome_type %in% c("binomial", "multinomial")) {
        # Get classes and their probabilities from the stored model data. In the
        # naive model, the probability of each class is its occurrence in the
        # development data.
        class_levels <- get_outcome_class_levels(object)
        class_probabilities <- object@model$frequency$count / object@model$n
        
        # Fill class probabilities.
        prediction_list <- list()
        for (ii in seq_along(class_levels)) {
          prediction_list[[class_levels[ii]]] <- rep.int(class_probabilities[ii], n_samples)
        }
        
        # Store as prediction table.
        prediction_table <- as_prediction_table(
          x = prediction_list,
          type = "classification",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("continuous")) {
        # Store as prediction table. In the naive model, return the median value
        # for regression problems.
        prediction_table <- as_prediction_table(
          x = rep.int(object@model$median, n_samples),
          type = "regression",
          data = data,
          model_object = object
        )
        
      } else if (object@outcome_type %in% c("survival")) {
        # For survival outcomes based on relative risks, predict the average
        # risk, i.e. 1.0.
        prediction_table <- as_prediction_table(
          x = rep.int(1.0, n_samples),
          type = "hazard_ratio",
          data = data,
          model_object = object
        )
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }
    } else if (type == "survival_probability" && object@outcome_type == "survival") {
      # survival probability ---------------------------------------------------
      
      # If time is unset, read the max time stored by the model.
      if (is.null(time)) time <- object@settings$time_max

      # Check for several issues that prevent survival probabilities from being
      # predicted.
      if (!has_calibration_info(object)) return(callNextMethod())
      
      # For naive models, survival probability is defined by the Kaplan-Meier
      # estimate of survival at the time point time.
      prediction_table <- .survival_probability_relative_risk(
        object = object,
        data = data,
        time = time
      )
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
    
    return(prediction_table)
  }
)



# ..predict (familiarNaiveSurvivalTimeModel) -----------------------------------
setMethod(
  "..predict", 
  signature(
    object = "familiarNaiveSurvivalTimeModel",
    data = "dataObject"
  ),
  function(
    object, 
    data,
    type = "default",
    time = NULL,
    ...
  ) {
    
    # Check n_samples, and then ensure that n_samples is equal to the number
    # of rows in the dataset.
    n_samples <- get_n_samples(data)
    if (n_samples == 0L) return(callNextMethod())
    n_samples <- nrow(data@data)

    if (object@outcome_type == "survival" && type == "default") {
      # For survival outcomes based on survival times, predict the average
      # survival time.

      # Check if the median survival time was measured.
      if (any(object@model$survival_probability$survival_probability <= 0.5)) {
        # Interpolate the survival function to find the median survival time.
        survival_time <- suppressWarnings(
          stats::approx(
            x = object@model$survival_probability$survival_probability,
            y = object@model$survival_probability$time,
            xout = 0.5,
            method = "linear",
            rule = 2L
          )$y
        )
        
      } else {
        # Use the last known time point.
        survival_time <- max(object@model$survival_probability$time)
      }

      # Set the survival time.
      prediction_table <- as_prediction_table(
        x = rep.int(survival_time, n_samples),
        type = "expected_survival_time",
        data = data,
        model_object = object
      )
      
    } else if (object@outcome_type == "survival" && type == "survival_probability") {
      # For naive models, survival probability is defined by the Kaplan-Meier
      # estimate of survival at the time point time.
      prediction_table <- as_prediction_table(
        x = rep.int(1.0, n_samples),
        type = "hazard_ratio",
        data = data,
        model_object = object
      )
      
      # If time is unset, read the max time stored by the model.
      if (is.null(time)) time <- object@settings$time_max
      
      # Check for several issues that prevent survival probabilities from being
      # predicted.
      if (!has_calibration_info(object)) return(callNextMethod())
      
      # For naive models, survival probability is defined by the Kaplan-Meier
      # estimate of survival at the time point time.
      prediction_table <- .survival_probability_relative_risk(
        object = prediction_table,
        data = data,
        time = time,
        model = object
      )
      
    } else {
      ..error_no_predictions_possible(object, type)
    }

    return(prediction_table)
  }
)



# ..predict (familiarNaiveCumulativeIncidenceModel) ----------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarNaiveCumulativeIncidenceModel",
    data = "dataObject"
  ),
  function(
    object, 
    data, 
    type = "default",
    time = NULL,
    ...
  ) {
    
    # Check n_samples, and then ensure that n_samples is equal to the number
    # of rows in the dataset.
    n_samples <- get_n_samples(data)
    if (n_samples == 0L) return(callNextMethod())
    n_samples <- nrow(data@data)

    if (object@outcome_type %in% c("survival") &&  type == "default") {
      # For survival outcomes based on survival times, predict the average
      # survival time.
      
      # If time is not provided, set the time at the last observed event.
      if (is.null(time)) time <- object@model$event_fivenum$max
      
      # Approximate cumulative incidence.
      cumulative_incidence <- suppressWarnings(
        stats::approx(
          x = object@model$cumulative_incidence$time,
          y = object@model$cumulative_incidence$cumulative_incidence,
          xout = time,
          method = "linear",
          rule = 2L
        )$y
      )
      
      # Compute the cumulative hazard at the indicated time point.
      # Set the survival time.
      prediction_table <- as_prediction_table(
        x = rep.int(cumulative_incidence, n_samples),
        type = "cumulative_hazard",
        data = data,
        model_object = object
      )
      
    } else if (object@outcome_type == "survival" && type == "survival_probability") {
      # For naive models, survival probability is defined by the Kaplan-Meier
      # estimate of survival at the time point time.
      prediction_table <- as_prediction_table(
        x = rep.int(1.0, n_samples),
        type = "hazard_ratio",
        data = data,
        model_object = object
      )
      
      # If time is unset, read the max time stored by the model.
      if (is.null(time)) time <- object@settings$time_max
      
      # Check for several issues that prevent survival probabilities from being
      # predicted.
      if (!has_calibration_info(object)) return(callNextMethod())
      
      # For naive models, survival probability is defined by the Kaplan-Meier
      # estimate of survival at the time point time.
      prediction_table <- .survival_probability_relative_risk(
        object = prediction_table,
        data = data,
        time = time,
        model = object
      )
      
    } else {
      ..error_no_predictions_possible(object, type)
    }

    return(prediction_table)
  }
)



# get_prediction_type (familiarNaiveCoxModel) ----------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarNaiveCoxModel"),
  function(object, type = "default") {
    # Cox proportional hazards models predict relative risks.
    if (type == "default") {
      return("hazard_ratio")
    } else if (type == "survival_probability") {
      return("survival_probability")
    } else {
      ..error_reached_unreachable_code(
        "get_prediction_type,familiarNaiveCoxModel: unknown type"
      )
    }
  }
)



# get_prediction_type (familiarNaiveSurvivalTimeModel) -------------------------
setMethod(
  "get_prediction_type", 
  signature(object = "familiarNaiveSurvivalTimeModel"),
  function(object, type = "default") {
    # These models predict an expected survival time by
    # default.
    if (type == "default") {
      return("expected_survival_time")
    } else if (type == "survival_probability") {
      return("survival_probability")
    } else {
      ..error_reached_unreachable_code(
        "get_prediction_type,familiarNaiveSurvivalTimeModel: unknown type"
      )
    }
  }
)



# get_prediction_type (familiarNaiveCumulativeIncidenceModel) ------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarNaiveCumulativeIncidenceModel"),
  function(object, type = "default") {
    # These models predict an cumulative hazard by default.
    if (type == "default") {
      return("cumulative_hazard")
    } else if (type %in% c("survival_probability")) {
      return("survival_probability")
    } else {
      ..error_reached_unreachable_code(
        "get_prediction_type,familiarNaiveCumulativeIncidenceModel: unknown type"
      )
    }
  }
)


# ..set_calibration_info -------------------------------------------------------
setMethod(
  "..set_calibration_info",
  signature(object = "familiarNaiveModel"),
  function(object, data) {
    # Check if calibration info already.
    if (has_calibration_info(object)) {
      return(object)
    }

    if (object@outcome_type == "survival") {
      # Determine baseline survival.
      object@calibration_info <- get_baseline_survival(data = data)
      
    } else {
      return(callNextMethod())
    }

    return(object)
  }
)



# show (familiarNaiveModel) ----------------------------------------------------
setMethod(
  "show",
  signature(object = "familiarNaiveModel"),
  function(object) {
    # Make sure the model object is updated.
    object <- update_object(object = object)

    if (!model_is_trained(object)) {
      cat(paste0(
        "A ", object@learner, " model (class: ", class(object)[1L],
        ") that was not successfully trained (", .familiar_version_string(object),
        ").\n"
      ))

      if (length(object@messages$warning) > 0L) {
        condition_messages <- condition_summary(object@messages$warning)
        cat(paste0(
          "\nThe following ",
          ifelse(length(condition_messages) == 1L, "warning was", "warnings were"),
          " generated while trying to train the model:\n",
          paste0(condition_messages, collapse = "\n"),
          "\n"
        ))
      }

      if (length(object@messages$error) > 0L) {
        condition_messages <- condition_summary(object@messages$error)
        cat(paste0(
          "\nThe following ",
          ifelse(length(condition_messages) == 1L, "error was", "errors were"),
          " encountered while trying to train the model:\n",
          paste0(condition_messages, collapse = "\n"),
          "\n"
        ))
      }
    }
    
    # Describe the learner and the version of familiar.
    message_str <- paste0(
      "A naive ", object@learner, " model (class: ", class(object)[1L],
      "; ", .familiar_version_string(object), ")"
    )
    
    # Describe the package(s), if any
    if (!is.null(object@package)) {
      message_str <- c(
        message_str,
        paste0(" trained using "),
        paste_s(mapply(
          ..message_package_version,
          x = object@package, 
          version = object@package_version
        )),
        ifelse(length(object@package) > 1L, " packages", " package")
      )
    }
    
    # Complete message and write.
    message_str <- paste0(c(message_str, ".\n"), collapse = "")
    cat(message_str)
    
    cat(paste0("\n--------------- Model details ---------------\n"))
    
    # Model details
    if (object@is_trimmed) {
      cat(object@trimmed_function$show, sep = "\n")
    } else {
      show(object@model)
    }
    
    cat(paste0("\n---------------------------------------------\n"))
    
    # Outcome details
    cat("\nThe following outcome was modelled:\n")
    show(object@outcome_info)
    
    # Details concerning hyperparameters.
    cat("\nThe model was trained using the following hyperparameters:\n")
    invisible(lapply(
      names(object@hyperparameters),
      function(x, object) {
        cat(paste0("  ", x, ": ", object@hyperparameters[[x]], "\n"))
      },
      object = object
    ))
    
    # Add note that naive models don't directly use hyperparameters.
    if (model_is_trained(object)) {
      cat("Note that the above hyperparameters are not directly used by the naive model.\n")
    }
    
    if (length(object@messages$warning) > 0L || length(object@messages$error) > 0L) {
      cat(paste0("\n------------ Warnings and errors ------------\n"))
      
      if (length(object@messages$warning) > 0L) {
        condition_messages <- condition_summary(object@messages$warning)
        cat(paste0(
          "\nThe following ",
          ifelse(length(condition_messages) == 1L, "warning was", "warnings were"),
          " generated while training the model:\n",
          paste0(condition_messages, collapse = "\n")
        ))
      }
      
      if (length(object@messages$error) > 0L) {
        condition_messages <- condition_summary(object@messages$error)
        cat(paste0(
          "\nThe following ",
          ifelse(length(condition_messages) == 1L, "error was", "errors were"),
          " encountered while training the model:\n",
          paste0(condition_messages, collapse = "\n")
        ))
      }
    }
    
    # Check package version.
    check_package_version(object)
  }
)

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Generic test. Note that confusion matrices can be created if only one sample
# is present.
familiar:::test_plots(
  plot_function = familiar:::plot_confusion_matrix,
  not_available_all_prospective = TRUE,
  data_element = "confusion_matrix",
  outcome_type_available = c("binomial", "multinomial"),
  debug = debug_flag
)

# Test confusion matrix without transparent cells.
familiar:::test_plots(
  plot_function = familiar:::plot_confusion_matrix,
  data_element = "confusion_matrix",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  plot_args = list("show_alpha" = "none"),
  debug = debug_flag
)

# Test by confusion matrix with transparency by class.
familiar:::test_plots(
  plot_function = familiar:::plot_confusion_matrix,
  data_element = "confusion_matrix",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  plot_args = list("show_alpha" = "by_class"),
  debug = debug_flag
)

# Test by confusion matrix with transparency by the entire plot.
familiar:::test_plots(
  plot_function = familiar:::plot_confusion_matrix,
  data_element = "confusion_matrix",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  plot_args = list("show_alpha" = "by_figure"),
  debug = debug_flag
)

# Test by confusion matrix with transparency by the entire plot.
familiar:::test_plots(
  plot_function = familiar:::plot_confusion_matrix,
  data_element = "confusion_matrix",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  plot_args = list("show_alpha" = "by_all"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_confusion_matrix,
  data_element = "confusion_matrix",
  outcome_type_available = c("binomial", "multinomial"),
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_confusion_matrix,
  data_element = "confusion_matrix",
  outcome_type_available = c("binomial", "multinomial"),
  test_config = "normal",
  debug = debug_flag
)

# Test alignment of different plots, with missing data.
familiar:::test_plot_ordering(
  plot_function = familiar:::plot_confusion_matrix,
  data_element = "confusion_matrix",
  outcome_type_available = c("binomial", "multinomial"),
  plot_args = list("facet_by" = c("data_set", "learner", "vimp_method")),
  debug = debug_flag
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarVimpMethod"),
  function(object, ...) {
    # This is a fall-back method.
    return(FALSE)
  }
)



# require_package (vimp method) ------------------------------------------------
setMethod(
  "require_package",
  signature(x = "familiarVimpMethod"),
  function(x, purpose = "vimp", message_type = "error", ...) {
    # Skip if no package is required.
    if (is_empty(x@package)) return(invisible(TRUE))

    # Set standard purposes for common uses.
    if (!is.null(purpose)) {
      if (purpose %in% c("vimp")) {
        purpose <- switch(
          purpose,
          "vimp" = "to determine variable importance"
        )
      }
    }

    return(invisible(.require_package(
      x = x@package,
      purpose = purpose, 
      message_type = message_type
    )))
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarVimpMethod"),
  function(object, ...) {
    # This is a fall-back method.
    return(list())
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarVimpMethod"),
  function(object, ...) {
    # This is a fall-back method.
    return(get_placeholder_vimp_table(
      vimp_method = object@vimp_method,
      run_table = object@run_table
    ))
  }
)



# .vimp (familiarVimpMethod) ---------------------------------------------------
setMethod(
  ".vimp",
  signature(object = "familiarVimpMethod"),
  function(
    object, 
    data, 
    is_pre_processed = FALSE,
    cl = NULL,
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    cluster_name <- NULL

    # Check if the class of object is a subclass of familiarVimpMethod
    if (!is_subclass(class(object)[1L], "familiarVimpMethod")) {
      object <- promote_vimp_method(object)
    }

    # Return empty table if data is absent.
    if (is_empty(data)) {
      return(get_placeholder_vimp_table(
        vimp_method = object@vimp_method,
        run_table = object@run_table
      ))
    }

    # Prepare input data
    data <- process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = "clustering"
    )

    # Work only with data that has known outcomes when determining variable
    # importance.
    data <- filter_missing_outcome(data = data)

    # Check again if data is absent because data may not have been loaded in the
    # check above.
    if (is_empty(data)) {
      return(get_placeholder_vimp_table(
        vimp_method = object@vimp_method,
        run_table = object@run_table
      ))
    }
    
    # Identify invariant features and remove them.
    available_features <- get_feature_columns(x = data)
    is_invariant <- fam_sapply(
      cl = cl,
      X = data@data[, mget(available_features)],
      FUN = is_singular_data,
      progress_bar = FALSE,
      chopchop = TRUE
    )
    invariant_features <- available_features[is_invariant]
    
    # Remove invariant features.
    if (length(invariant_features) > 0L) {
      data <- filter_features(
        data = data,
        remove_features = invariant_features
      )
    }

    # Check that the data is suitable for predictions.
    if (has_bad_training_data(object = object, data = data)) {
      return(get_placeholder_vimp_table(
        vimp_method = object@vimp_method,
        run_table = object@run_table
      ))
    }

    # Determine variable importance.
    vimp_object <- ..vimp(
      object = object,
      data = data,
      cl = cl
    )

    # Find signature features.
    signature_features <- names(object@feature_info)[sapply(object@feature_info, is_in_signature)]

    # Remove signature features.
    vimp_object <- remove_signature_features(
      vimp_object,
      features = signature_features
    )

    # Set up a table with clustering information.
    cluster_table <- .create_clustering_table(object@feature_info)

    # Remove invariant features from the cluster_table
    if (length(invariant_features) > 0L) {
      cluster_table <- cluster_table[!cluster_name %in% invariant_features]
    }

    # Remove features in the signature.
    if (length(signature_features) > 0L) {
      cluster_table <- cluster_table[!cluster_name %in% signature_features]
    }

    # Update variable importance table object.
    vimp_object@vimp_method <- object@vimp_method
    vimp_object@run_table <- object@run_table
    vimp_object@cluster_table <- cluster_table
    vimp_object@project_id <- object@project_id

    # Set package version.
    vimp_object <- add_package_version(vimp_object)

    return(vimp_object)
  }
)



# .vimp (familiarModel) --------------------------------------------------------
setMethod(
  ".vimp",
  signature(object = "familiarModel"),
  function(
    object, 
    data, 
    is_pre_processed = FALSE, 
    ...
  ) {
    # Suppress NOTES due to non-standard evaluation in data.table
    cluster_name <- NULL

    # Check if the class of object is a subclass of familiarModel.
    if (!is_subclass(class(object)[1L], "familiarModel")) {
      object <- promote_learner(object)
    }
    
    # Return empty table if data is absent.
    if (is_empty(data)) {
      return(get_placeholder_vimp_table(
        vimp_method = object@learner,
        run_table = object@run_table
      ))
    }

    # Prepare input data
    data <- process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = "clustering"
    )

    # Work only with data that has known outcomes when determining variable
    # importance.
    data <- filter_missing_outcome(data = data)

    # Check again if data is absent because data may not have been loaded in the
    # check above.
    if (is_empty(data)) {
      return(get_placeholder_vimp_table(
        vimp_method = object@learner,
        run_table = object@run_table
      ))
    }

    # Identify invariant features and remove them.
    available_features <- get_feature_columns(x = data)
    is_invariant <- sapply(
      available_features,
      function(feature, data) is_singular_data(data[, get(feature)]),
      data = data@data
    )
    
    invariant_features <- available_features[is_invariant]

    # Remove invariant features.
    if (length(invariant_features) > 0L) {
      data <- filter_features(
        data = data,
        remove_features = invariant_features
      )
    }

    # Check that the data is suitable for predictions.
    if (has_bad_training_data(object = object, data = data)) {
      return(get_placeholder_vimp_table(
        vimp_method = object@learner,
        run_table = object@run_table
      ))
    }

    # Determine variable importance.
    vimp_object <- ..vimp(
      object = object,
      data = data
    )

    # Find signature features.
    signature_features <- names(object@feature_info)[sapply(object@feature_info, is_in_signature)]

    # Remove signature features.
    vimp_object <- remove_signature_features(
      vimp_object,
      features = signature_features
    )

    # Set up a table with clustering information.
    cluster_table <- .create_clustering_table(object@feature_info)

    # Remove invariant features from the cluster_table
    if (length(invariant_features) > 0L) {
      cluster_table <- cluster_table[!cluster_name %in% invariant_features]
    }

    # Remove features in the signature.
    if (length(signature_features) > 0L) {
      cluster_table <- cluster_table[!cluster_name %in% signature_features]
    }

    # Update variable importance table object.
    vimp_object@vimp_method <- object@learner
    vimp_object@run_table <- object@run_table
    vimp_object@cluster_table <- cluster_table
    vimp_object@project_id <- object@project_id

    # Set package version.
    vimp_object <- add_package_version(vimp_object)

    return(vimp_object)
  }
)

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL

# complete_familiar_ensemble ---------------------------------------------------
setMethod(
  "complete_familiar_ensemble",
  signature(object = "familiarEnsemble"),
  function(
    object,
    dir_path = NULL,
    message_indent = 0L,
    verbose = FALSE
  ) {
    # Fills out missing data from a familiarEnsemble based on attached models
    # and internal logic.

    # Load models
    object <- ..update_model_list(
      object = object,
      dir_path = dir_path
    )
    model_list <- ..get_model(object = object)

    # Determine which models were trained.
    trained_mask <- sapply(model_list, model_is_trained)

    # If no models were trained, select all models, otherwise select only the
    # models that were trained.
    if (any(trained_mask)) model_list <- model_list[trained_mask]

    if (is(model_list[[1L]], "familiarModel")) {
      # Add outcome_type and outcome_info to object. These slots are required in
      # some of the other aggregators.
      object@outcome_type <- model_list[[1L]]@outcome_type
      object@outcome_info <- .aggregate_outcome_info(
        x = lapply(
          model_list,
          function(list_elem) (list_elem@outcome_info)
        )
      )
      
      # Find all important features for novelty detection.
      novelty_features <- unique(unlist(lapply(
        model_list, 
        function(fam_model) (fam_model@novelty_features)
      )))
      
      # Aggregate calibration info
      calibration_info <- extract_calibration_info(
        object = object,
        detail_level = "hybrid",
        message_indent = message_indent,
        verbose = verbose
      )
      calibration_info <- .compute_data_element_estimates(calibration_info)
      
      if (is_empty(calibration_info)) {
        calibration_info <- NULL
      } else {
        calibration_info <- calibration_info[[1L]]@data
      }
      
      # Get settings
      settings <- model_list[[1L]]@settings
      
    } else if (is(model_list[[1L]], "familiarNoveltyDetector")) {
      # Add outcome_type and outcome_info to object.
      object@outcome_type <- "unsupervised"
      object@outcome_info <- NULL
      
      # Find all important features for novelty detection.
      novelty_features <- unique(unlist(lapply(
        model_list, 
        function(fam_model) (fam_model@model_features)
      )))
      
      calibration_info <- NULL
      settings <- NULL
    }
    
    # Find all required features
    required_features <- unique(unlist(lapply(
      model_list,
      function(fam_model) (fam_model@required_features)
    )))

    # Find all important features for the model.
    model_features <- unique(unlist(lapply(
      model_list,
      function(fam_model) (fam_model@model_features)
    )))

    # Aggregate feature information
    if (length(required_features) > 0L) {
      feature_info_list <- lapply(required_features,
        .collect_and_aggregate_feature_info,
        object = object,
        model_list = model_list,
        stop_at = "imputation"
      )

      # Add name to features.
      names(feature_info_list) <- required_features
      
    } else {
      feature_info_list <- NULL
    }

    # Generate a new version of the ensemble to avoid unnecessary copying.
    fam_ensemble <- methods::new("familiarEnsemble",
      model_list = object@model_list,
      outcome_type = object@outcome_type,
      outcome_info = object@outcome_info,
      data_column_info = model_list[[1L]]@data_column_info,
      learner = object@learner,
      vimp_method = object@vimp_method,
      required_features = required_features,
      model_features = model_features,
      novelty_features = novelty_features,
      feature_info = feature_info_list,
      data_id = object@data_id,
      run_id = object@run_id,
      run_table = object@run_table,
      calibration_info = calibration_info,
      model_dir_path = object@model_dir_path,
      auto_detach = object@auto_detach,
      settings = settings,
      project_id = model_list[[1L]]@project_id
    )

    # Add package version to the ensemble
    fam_ensemble <- add_package_version(object = fam_ensemble)

    return(fam_ensemble)
  }
)



# show (familiarEnsemble) ------------------------------------------------------
setMethod(
  "show",
  signature(object = "familiarEnsemble"),
  function(object) {
    # Make sure the model object is updated.
    object <- update_object(object = object)

    # Flag to show additional information.
    show_additional_information <- TRUE

    if (!is_model_loaded(object)) {
      cat(paste0(
        "An ensemble of ", length(object@model_list), " ",
        object@learner, " ",
        ifelse(length(object@model_list) == 1L, "model", "models"),
        " (", .familiar_version_string(object), ").\n"
      ))
      
    } else {
      cat(paste0(
        "An ensemble of ", length(object@model_list), " ",
        object@learner, " ",
        ifelse(length(object@model_list) == 1L, "model", "models"),
        " (", .familiar_version_string(object), ").\n"
      ))

      # Determine how many models are trained.
      model_trained <- sapply(object@model_list, model_is_trained)

      if (length(model_trained) == 1L) {
        show(object@model_list[[1L]])

        # Update the flag to prevent showing redundant information.
        show_additional_information <- FALSE
        
      } else {
        cat(paste0("\n", .report_ensemble_models(object = object), "\n"))
      }
    }

    if (show_additional_information) {
      # Outcome details
      cat("\nThe following outcome was modelled:\n")
      show(object@outcome_info)

      # Details concerning variable importance.
      cat(paste0(
        "\nVariable importance was determined using the ",
        object@vimp_method, " variable importance method.\n"
      ))

      # Details concerning model features:
      cat("\nThe following features were used in the ensemble:\n")
      lapply(
        object@model_features,
        function(x, object) {
          cat(.show_simple_feature_info(
            object@feature_info[[x]],
            line_end = ".\n"
          ))
        },
        object = object
      )
    }
  }
)



# get_prediction_type (familiarEnsemble) ---------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "familiarEnsemble"),
  function(object, ...) {
    # This dispatches to get_prediction_type for the familiarModel subclass if a
    # model is attached, and to character if not. The latter option then loads
    # the familiarModel.
    return(do.call(
      get_prediction_type,
      args = c(
        list("object" = object@model_list[[1L]]),
        list(...)
      )
    ))
  }
)



# get_prediction_type (character) ----------------------------------------------
setMethod(
  "get_prediction_type",
  signature(object = "character"),
  function(object, ...) {
    object <- load_familiar_object(object)
    
    return(do.call(
      get_prediction_type,
      args = c(
        list("object" = object),
        list(...)
      )
    ))
  }
)



# ..get_model (numeric, familiarEnsemble) --------------------------------------
setMethod(
  "..get_model",
  signature(
    ii = "numeric",
    object = "familiarEnsemble"
  ),
  function(ii, object) {
    
    if (ii > length(object@model_list)) {
      ..error_reached_unreachable_code(paste0(
        "..get_model,familiarEnsemble: the requested index (", ii,
        ") exceeds the total number of models in the ensemble (",
        length(object@model_list)
      ))
    }
    
    # If the model is already attached, dispatch to calling function.
    if (is(object@model_list[[ii]], "familiarModel")) {
      return(object@model_list[[ii]])
    }
    if (is(object@model_list[[ii]], "familiarNoveltyDetector")) {
      return(object@model_list[[ii]])
    }
    
    # If the model has been detached, load, and then dispatch.
    # First check if the file exists.
    if (!file.exists(object@model_list[[ii]])) {
      # Obtain the file name of the model.
      model_file_name <- ..get_model_file_path(ii = ii, object = object)

      # Load the model if the file exists.
      if (!is.null(model_file_name)) {
        return(load_familiar_object(model_file_name))
      }

      # If no model could be found, throw an error.
      ..error(paste0(
        "..get_model,familiarEnsemble: cannot find the indicated familiarModel",
        object@model_list[[ii]]
      ))
    }
    
    return(load_familiar_object(object@model_list[[ii]]))
  }
)



# ..get_model (missing, familiarEnsemble) --------------------------------------
setMethod(
  "..get_model",
  signature(
    ii = "missing",
    object = "familiarEnsemble"
  ),
  function(ii, object, ...) {
    
    if (length(object@model_list) > 0L) {
      # Dispatch get_model.
      return(lapply(
        seq_along(object@model_list),
        ..get_model,
        object = object,
        ...
      ))
      
    } else {
      return(NULL)
    }
  }
)



# ..get_model_file_path (numeric, familiarEnsemble) ----------------------------
setMethod(
  "..get_model_file_path",
  signature(
    ii = "numeric",
    object = "familiarEnsemble"
  ),
  function(ii, object, dir_path = NULL) {
    
    if (ii > length(object@model_list)) {
      ..error_reached_unreachable_code(paste0(
        "..get_model_file_path,familiarEnsemble: the requested index (", ii,
        ") exceeds the total number of models in the ensemble (",
        length(object@model_list)
      ))
    }

    # There are three directories a file can be located, namely:
    #
    # * dir_path: a user-specified path
    #
    # * mb_dir_path: based on file_paths environment variable
    #
    # * model_dir_path: based on the directory path stored with the object.
    #
    # These possibilities are ordered by precedence.

    # Try to find dir_path
    if (!is.null(dir_path)) {
      # Check if the dir_path actually points to a file, and strip the file
      # instead. Note that file.exists also returns TRUE if the path is a
      # directory.
      if (file.exists(dir_path) && !dir.exists(dir_path)) {
        dir_path <- dirname(dir_path)
      }

      # If the directory does not exist, set to NULL.
      if (!dir.exists(dir_path)) dir_path <- NULL
    }

    # Obtain the directory path stored in the file_paths environment variable
    # (if any).
    mb_dir_path <- tryCatch(
      get_file_paths()$mb_dir,
      error = function(err) (return(NULL))
    )

    # Obtain the directory path from the familiarEnsemble object.
    model_dir_path <- object@model_dir_path

    # Identify the basic model file name.
    if (is(object@model_list[[ii]], "familiarModel")) {
      # Generate the file name from the model.
      model_file_name <- get_object_name(
        object = object@model_list[[ii]],
        abbreviated = FALSE
      )
      model_file_name <- paste0(model_file_name, ".RDS")
      
    } else if (is(object@model_list[[ii]], "familiarNoveltyDetector")) {
      return(NULL)
      
    } else {
      # Generate the file name from the stored string.
      model_file_name <- basename(object@model_list[[ii]])
    }

    # Using the user-provided dir_path
    if (!is.null(dir_path)) {
      # File is directly located in the given directory
      file_path_1 <- file.path(dir_path, model_file_name)

      if (file.exists(file_path_1)) return(file_path_1)

      # File is located in a subdirectory of dir_path.
      file_path_2 <- file.path(
        get_object_dir_path(
          dir_path = dir_path,
          object_type = "familiarModel",
          learner = object@learner,
          vimp_method = object@vimp_method
        ),
        model_file_name
      )

      if (file.exists(file_path_2)) return(file_path_2)
    }

    # Using the file directory indicated by the file_paths environment variable.
    if (!is.null(mb_dir_path)) {
      # File is directly located in the given directory
      file_path_3 <- file.path(mb_dir_path, model_file_name)

      if (file.exists(file_path_3)) return(file_path_3)

      # File is located in a subdirectory of dir_path.
      file_path_4 <- file.path(
        get_object_dir_path(
          dir_path = mb_dir_path,
          object_type = "familiarModel",
          learner = object@learner,
          vimp_method = object@vimp_method
        ),
        model_file_name
      )

      if (file.exists(file_path_4)) return(file_path_4)
    }

    # Using the model_dir_path slot of the ensemble. This takes least
    # precedence, as this might have changed after creating the ensemble method.
    if (!is.na(model_dir_path)) {
      # File is directly located in the given directory.
      file_path_5 <- file.path(model_dir_path, model_file_name)

      if (file.exists(file_path_5)) return(file_path_5)

      # File is located in a subdirectory of model_dir_path.
      file_path_6 <- file.path(
        get_object_dir_path(
          dir_path = model_dir_path,
          object_type = "familiarModel",
          learner = object@learner,
          vimp_method = object@vimp_method
        ),
        model_file_name
      )

      if (file.exists(file_path_6)) return(file_path_6)
    }

    # Explicitly return NULL if the model could not be found on any known paths.
    return(NULL)
  }
)



# ..get_model_file_path (missing, familiarEnsemble) ----------------------------
setMethod(
  "..get_model_file_path",
  signature(
    ii = "missing",
    object = "familiarEnsemble"
  ),
  function(ii, object, ...) {
    # Dispatch to ..get_model_file_path.
    return(lapply(
      seq_along(object@model_list),
      ..get_model_file_path,
      object = object,
      ...
    ))
  }
)



# ..update_model_list ----------------------------------------------------------
setMethod(
  "..update_model_list",
  signature(object = "familiarEnsemble"),
  function(
    object,
    dir_path = NULL,
    auto_detach = FALSE
  ) {
    # Determine if models can detach. Models cannot detach if detaching them
    # would lead them to be lost, i.e. there are no files on a known drive
    # location drive.
    if (!..can_detach_models(object = object, dir_path = dir_path)) {
      auto_detach <- FALSE
      object@auto_detach <- FALSE
      object@model_dir_path <- NA_character_
    }

    if (auto_detach || object@auto_detach) object <- detach_models(object)

    for (ii in seq_along(object@model_list)) {
      # Skip if the current entry is an attached familiarModel object.
      if (is(object@model_list[[ii]], "familiarModel")) next
      if (is(object@model_list[[ii]], "familiarNoveltyDetector")) next

      # Stop if the current entry is not a character.
      if (!is.character(object@model_list[[ii]])) {
        rlang::abort(paste0(
          "The current entry in the model list is not a character string ",
          "or a familiarModel object."
        ))
      }

      # Skip if the entry points to a file and not a directory.
      if (
        file.exists(object@model_list[[ii]]) &&
        !dir.exists(object@model_list[[ii]])
      ) {
        next
      }

      # Identify the file path, if any.
      model_file_path <- ..get_model_file_path(
        ii = ii,
        object = object,
        dir_path = dir_path
      )

      # If there is a file path, add this to the model list, and update
      # the model_dir_path attribute.
      if (!is.null(model_file_path)) {
        object@model_list[[ii]] <- model_file_path
        object@model_dir_path <- dirname(model_file_path)
      }
    }

    # Check if the models are either attached or are located on a known
    # drive location.
    if (length(object@model_list) > 0L) {
      model_exists <- sapply(
        object@model_list,
        function(list_entry) {
          if (is(list_entry, "familiarModel")) {
            return(TRUE)
          } else if (is(list_entry, "familiarNoveltyDetector")) {
            return(TRUE)
          } else if (file.exists(list_entry)) {
            return(TRUE)
          }
          return(FALSE)
        }
      )
      
      # Throw an error if not all models exist.
      if (!any(model_exists)) {
        ..error(paste0(
          "None of the models could be found: ",
          paste_s(unlist(object@model_list)),
          ". \n\nThis is likely because the models are no longer found in the ",
          "same location as they were created. ",
          "Use the update_model_dir_path method to update the path for the ",
          "directory containing the models."
        ))
      }

      if (!all(model_exists)) {
        ..error(paste0(
          "The following models in the ensemble could not be found: ",
          paste_s(unlist(object@model_list[!model_exists])), "."
        ))
      }
    }

    # Check the model_dir_path slot if auto_detach is on. In that case
    # model_dir_path is required to find the models.
    if ((auto_detach || object@auto_detach) && length(object@model_list) > 0L) {
      # Check if all models are attached, because in that case we may still need
      # to determine the drive location.
      if (is_model_loaded(object)) {
        # Identify the file path, if any.
        model_file_path <- ..get_model_file_path(
          ii = 1L,
          object = object, 
          dir_path = dir_path
        )

        # Set the model_file_path explicitly.
        if (!is.null(model_file_path)) object@model_dir_path <- dirname(model_file_path)
      }

      # Check that a model directory path has been set and exists.
      if (is.na(object@model_dir_path)) rlang::abort("The model directory could not be found.")
      if (!dir.exists(object@model_dir_path)) rlang::abort("The model directory could not be found.")
    }

    return(object)
  }
)



# ..can_detach_models (numeric, familiarEnsemble) ------------------------------
setMethod(
  "..can_detach_models",
  signature(
    ii = "numeric",
    object = "familiarEnsemble"
  ),
  function(ii, object, dir_path = NULL) {
    # Check whether a model can be detached without losing it. We do this by
    # checking whether a valid file path can be generated. If not,
    # ..get_model_file_path will return NULL.
    model_file_path <- ..get_model_file_path(
      ii = ii, 
      object = object, 
      dir_path = dir_path
    )

    return(!is.null(model_file_path))
  }
)



# ..can_detach_models (missing, familiarEnsemble) ------------------------------
setMethod(
  "..can_detach_models",
  signature(
    ii = "missing",
    object = "familiarEnsemble"
  ),
  function(ii, object, dir_path = NULL) {
    # Check if there are any models in the ensemble.
    if (length(object@model_list) > 0L) {
      # Check whether a model can be detached without losing it.
      can_detach <- sapply(
        seq_along(object@model_list),
        ..can_detach_models,
        object = object,
        dir_path = dir_path
      )

      # Return TRUE when all models can be detached.
      return(all(can_detach))
      
    } else {
      return(TRUE)
    }
  }
)



# update_model_dir_path (generic) ----------------------------------------------

#' @title Updates model directory path for ensemble objects.
#'
#' @description Updates the model directory path of a `familiarEnsemble` object.
#'
#' @param object A `familiarEnsemble` object, or one or more `familiarModel`
#'  objects that will be internally converted to a `familiarEnsemble` object.
#'  Paths to such objects can also be provided.
#' @param dir_path Path to the directory where models are stored.
#' @param ... Unused arguments.
#'
#' @details Ensemble models created by familiar are often written to a directory
#'  on a local drive or network. In such cases, the actual models are detached,
#'  and paths to the models are stored instead. When the models are moved from
#'  their original location, they can no longer be found and attached to the
#'  ensemble. This method allows for pointing to the new directory containing
#'  the models.
#'
#' @return A `familiarEnsemble` object.
#' @exportMethod update_model_dir_path
#' @md
#' @rdname update_model_dir_path-methods
setGeneric(
  "update_model_dir_path",
  function(object, dir_path, ...) {
    standardGeneric("update_model_dir_path")
  } 
)



## update_model_dir_path (familiarEnsemble) ------------------------------------

#' @rdname update_model_dir_path-methods
setMethod(
  "update_model_dir_path",
  signature(object = "familiarEnsemble"),
  function(object, dir_path) {
    # Check if the directory indicated by dir_path exists.
    if (!dir.exists(dir_path)) {
      # Check whether a path to a file may have accidentally been
      # provided.
      if (file.exists(dir_path)) {
        dir_path <- dirname(dir_path)
        
      } else {
        ..error(paste0(
          "The new model directory does not exist, or cannot be accessed. ",
          "Found: ", dir_path
        ))
      }
    }

    # Assign to attribute slot.
    object@model_dir_path <- dir_path

    return(object)
  }
)


## update_model_dir_path(ANY) --------------------------------------------------

#' @rdname update_model_dir_path-methods
setMethod(
  "update_model_dir_path",
  signature(object = "ANY"),
  function(object, dir_path) {
    # Attempt to convert object to familiar ensemble.
    object <- as_familiar_ensemble(object)

    # Update path to model directory
    return(update_model_dir_path(
      object = object,
      dir_path = dir_path
    ))
  }
)


# load_models ------------------------------------------------------------------
setMethod(
  "load_models",
  signature(object = "familiarEnsemble"),
  function(
    object,
    dir_path = NULL,
    suppress_auto_detach = FALSE, 
    drop_untrained = FALSE
  ) {
    # Skip if there no models on the list.
    if (length(object@model_list) == 0L) return(object)

    # Update model list as a precaution. This also checks that models
    # can actually be attached.
    object <- ..update_model_list(object = object, dir_path = dir_path)

    # Do not attach models if auto_detach is set to TRUE.
    if (!object@auto_detach || suppress_auto_detach) {
      object@model_list <- ..get_model(object = object)
    }

    # Drop models that were not trained.
    if (drop_untrained) {
      # Determine which models have been trained.
      trained_model_mask <- sapply(object@model_list, model_is_trained)

      # Drop untrained models.
      object@model_list <- object@model_list[trained_model_mask]
    }

    return(object)
  }
)



# is_model_loaded --------------------------------------------------------------
setMethod(
  "is_model_loaded",
  signature(object = "familiarEnsemble"),
  function(object) {
    if (length(object@model_list) > 0L && !object@auto_detach) {
      # Check that all models are present.
      return(all(sapply(object@model_list, is, class2 = "familiarModel")))
      
    } else if (length(object@model_list) > 0L & object@auto_detach) {
      # Return TRUE if all models are dynamically loaded.
      return(TRUE)
      
    } else {
      # If there are no models, return TRUE.
      return(TRUE)
    }
  }
)


# detach_models ----------------------------------------------------------------
setMethod(
  "detach_models",
  signature(object = "familiarEnsemble"),
  function(object) {
    # Unload the models in the familiarEnsemble
    model_list <- object@model_list

    # Iterate over the entries in the model list.
    for (ii in seq_along(object@model_list)) {
      # Check if the entry contains a familiarModel object that can be detached.
      if (is(object@model_list[[ii]], "familiarModel")) {
        # Check if the model can be detached.
        if (..can_detach_models(ii = ii, object = object)) {
          # Get the model file name.
          model_file_name <- ..get_model_file_path(
            ii = ii,
            object = object
          )

          # Update the entry
          model_list[[ii]] <- model_file_name
        }
      }
    }

    # Attach the model list to the object and return the object.
    object@model_list <- model_list

    return(object)
  }
)



# add_model_name (ANY, familiarEnsemble) ---------------------------------------
setMethod(
  "add_model_name",
  signature(
    data = "ANY",
    object = "familiarEnsemble"
  ),
  function(data, object) {
  
    if (is_empty(data)) return(NULL)

    ..error_reached_unreachable_code(
      "add_model_name,any,familiarEnsemble: no method for non-empty data."
    )
  }
)


# add_model_name (familiarDataElement, familiarEnsemble) -----------------------
setMethod(
  "add_model_name",
  signature(
    data = "familiarDataElement",
    object = "familiarEnsemble"
  ),
  function(data, object) {
    # Determine the model name
    if (length(object@name) == 0L) {
      model_name <- get_object_name(object = object, abbreviated = TRUE)
    } else {
      model_name <- object@name
    }

    if (is.null(data@identifiers)) {
      data@identifiers <- list("ensemble_model_name" = model_name)
    } else {
      data@identifiers[["ensemble_model_name"]] <- model_name
    }

    return(data)
  }
)


# set_object_name (familiarEnsemble) -------------------------------------------

#' @title Set the name of a `familiarEnsemble` object.
#'
#' @description Set the `name` slot using the object name.
#'
#' @param x A `familiarEnsemble` object.
#'
#' @return A `familiarEnsemble` object with a generated or a provided name.
#' @md
#' @keywords internal
setMethod(
  "set_object_name",
  signature(x = "familiarEnsemble"),
  function(x, new = NULL) {
    
    if (is.null(x@project_id) && is.null(new)) {
      # Generate a random object name. A project_id of 0 means that the objects
      # was auto-generated (i.e. through object conversion). We randomly
      # generate characters and add a time stamp, so that collision is
      # practically impossible.
      slot(object = x, name = "name") <- paste0(
        as.character(as.numeric(format(Sys.time(), "%H%M%S"))),
        "_", rstring(n = 20L)
      )
    } else if (is.null(new)) {
      # Generate a sensible object name.
      slot(object = x, name = "name") <- get_object_name(object = x)
      
    } else {
      slot(object = x, name = "name") <- new
    }

    return(x)
  }
)



# get_object_name (ensemble) ---------------------------------------------------
setMethod(
  "get_object_name",
  signature(object = "familiarEnsemble"),
  function(object, abbreviated = FALSE) {

    if (abbreviated) {
      # Create an abbreviated name
      model_name <- paste0(
        "ensemble", ".", object@data_id, ".", object@run_id
      )
      
    } else {
      # Create the full name of the model
      model_name <- get_object_file_name(
        learner = object@learner,
        vimp_method = object@vimp_method,
        project_id = object@project_id,
        data_id = object@data_id,
        run_id = object@run_id,
        object_type = "familiarEnsemble",
        with_extension = FALSE
      )
    }

    return(model_name)
  }
)



# model_is_trained (ensemble) --------------------------------------------------
setMethod(
  "model_is_trained",
  signature(object = "familiarEnsemble"),
  function(object) {
    # Check if a model was trained
    if (is.null(object@model_list)) {
      # Check if a model is present
      return(FALSE)
      
    } else if (length(object@model_list) == 0L) {
      # No models were attached
      return(FALSE)
      
    } else {
      # Iterate over the models and check if any was trained
      return(any(sapply(object@model_list, model_is_trained)))
    }
  }
)


# add_package_version (ensemble) -----------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "familiarEnsemble"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)

# save (ensemble) --------------------------------------------------------------
setMethod(
  "save", signature(list = "familiarEnsemble", file = "character"),
  function(list, file) {
    .save(object = list, dir_path = file)
  }
)



.report_ensemble_models <- function(object) {
  # Determine how many models are trained.
  model_trained <- sapply(object@model_list, model_is_trained)
  
  # Determine how many models are naive
  model_naive <- sapply(object@model_list, is, class2 = "familiarNaiveModel")
  
  # Determine numbers of models and type.
  n_models <- length(object@model_list)
  n_models_naive <- sum(model_naive)
  n_models_trained <- sum(model_trained)
  
  message_str <- paste0(
    "The ensemble contains ", n_models,
    ifelse(n_models == 1L, " model", " models")
  )
  
  if (n_models == 1L) {
    # Ensemble contains a single model.
    
    if (n_models_naive == 0L && n_models_trained == 1L) {
      message_str <- c(
        message_str,
        paste0(" which was successfully trained.")
      )
      
    } else if (n_models_naive == 1L) {
      message_str <- c(
        message_str,
        paste0(" which was successfully trained as a naive model.")
      )
      
    } else {
      message_str <- c(
        message_str,
        paste0(" which failed to successfully train.")
      )
    }
  } else if (n_models_trained > 0L) {
    # Ensemble contains multiple trained models.
    
    message_str <- c(
      message_str,
      paste0(
        ". Of these models, ", n_models_trained,
        ifelse(n_models_trained == 1L, " model was", " models were"),
        " successfully trained, of which ", n_models_naive,
        ifelse(n_models_naive == 1L, " model was", " models were"),
        " trained as a naive model."
      )
    )
    
    if (n_models > n_models_trained) {
      n_models_untrained <- n_models - n_models_trained
      
      message_str <- c(
        message_str,
        paste0(
          " The remaining ", n_models_untrained,
          ifelse(n_models_untrained == 1L, " model", " models"),
          " failed to successfully train."
        )
      )
    }
    
  } else {
    # Ensemble contains only models that failed to train.
    message_str <- c(message_str, ". All models failed to successfully train.")
  }
  
  return(paste0(message_str, collapse = ""))
}



.get_available_stratification_ensemble_methods <- function() {
  return(c("ensemble_mode", "ensemble_mean", "mean_threshold", "median_threshold"))
}

..warning <- function(
    ...,
    warning_class = NULL,
    call = rlang::caller_env()
) {
  # Generic error function.
  message_string <- paste0(list(...), collapse = "")
  rlang::warn(
    message = message_string,
    class = union("familiar_warning", warning_class),
    call = call
  )
}


..warning_missing_cohorts <- function(x, call = rlang::caller_env()) {
  message_string <- paste0(
    "Creating iterations: ",
    ifelse(length(x) > 1L, "Cohorts ", "Cohort "),
    paste_s(x),
    ifelse(length(x) > 1L, " were", " was"),
    " not found in the data table."
  )
  
  logger_warning(message_string, call = call)
  
  return(invisible(TRUE))
}



..warning_no_comparison_between_models <- function(
    call = rlang::caller_env()
) {
  logger_warning(
    paste0(
      "Cannot create plots to compare directly between models. ",
      "Please use the hybrid or ensemble detail levels."
    ),
    call = call
  )
  
  return(invisible(TRUE))
}



..warning_prediction_table_lacks_reference <- function(
    data_element_description,
    call = rlang::caller_env()
) {
  message_string <- paste0(
    "Unable to compute ", data_element_description, " using prediction tables as reference data are missing."
  )
  
  logger_warning(
    warn_str = message_string,
    warn_class = "prediction_table_no_data_extraction_warning",
    call = call
  )
  
  return(invisible(TRUE))
}



..warning_no_data_extraction_from_prediction_table <- function(
    data_element_description,
    call = rlang::caller_env()
) {
  message_string <- paste0(
    "Unable to compute ", data_element_description, " using prediction tables."
  )
  
  logger_warning(
    warn_str = message_string,
    warn_class = "prediction_table_no_data_extraction_warning",
    call = call
  )
  
  return(invisible(TRUE))
}



..warning_package_not_installed <- function(
    x, 
    purpose = NULL,
    call = rlang::caller_env()
) {
  
  # Only unique packages.
  x <- unique(x)
  
  # Basic error message.
  message_string <- ..message_missing_package(x = x, purpose = purpose)
  
  # Instructions for CRAN packages.
  message_string <- c(
    message_string,
    ..message_install_from_cran(x = x)
  )
  
  # Instructions for Bioconductor packages.
  message_string <- c(
    message_string,
    ..message_install_from_bioconductor(x = x)
  )
  
  rlang::warn(
    message = paste0(message_string, collapse = ""),
    class = c("familiar_warning", "package_missing"),
    call = rlang::caller_env()
  )
  
  return(invisible(TRUE))
}



..deprecation_count <- function(as_error = FALSE) {
  
  message_string <- "The \"count\" outcome type has been deprecated in familiar version 2.0.0."
  
  if (as_error) {
    rlang::abort(
      message = message_string,
      class = c("familiar_error", "deprecation_error")
    )
  }
  
  if (!.is_testing()) {
    rlang::warn(
      message = message_string,
      class = c("familiar_warning", "deprecation_warning"),
      .frequency = "once",
      .frequency_id = "deprecation_warning_count"
    )
  }
}



..deprecation_vgam <- function() {
  if (!.is_testing()) {
    rlang::warn(
      message = "The use of VGAM for multinomial logistic models has been deprecated since familiar version 2.0.0.",
      class = c("familiar_warning", "deprecation_warning"),
      .frequency = "once",
      .frequency_id = "deprecation_warning_vgam"
    )
  }
}



..deprecation_qvalue <- function() {
  if (!.is_testing()) {
    rlang::warn(
      message = "The qvalue package for computing q-values has been deprecated since familiar version 2.0.0.",
      class = c("familiar_warning", "deprecation_warning"),
      .frequency = "once",
      .frequency_id = "deprecation_warning_qvalue"
    )
  }
}



..deprecation_mboost <- function() {
  if (!.is_testing()) {
    rlang::warn(
      message = "The mboost package for gradient boosted models has been deprecated since familiar version 2.0.0.",
      class = c("familiar_warning", "deprecation_warning"),
      .frequency = "once",
      .frequency_id = "deprecation_warning_mboost"
    )
  }
}



..deprecation_rfsrc_variable_hunting <- function(as_error = FALSE) {
  
  message_string <- paste0(
    "The variable hunting feature selection method of randomForestSRC has been ",
    "deprecated in familiar version 2.0.0 due to stability issues."
  )
  
  if (as_error) {
    rlang::abort(
      message = message_string,
      class = c("familiar_error", "deprecation_error")
    )
  }
  
  if (!.is_testing()) {
    rlang::warn(
      message = message_string,
      class = c("familiar_warning", "deprecation_warning"),
      .frequency = "once",
      .frequency_id = "deprecation_warning_rfsrc_variable_hunting"
    )
  }
}



..deprecation_rfsrc_minimum_depth <- function(as_error = FALSE) {
  message_string <- paste0(
    "The minimum depth feature selection method of randomForestSRC has been ",
    "deprecated in familiar version 2.0.0 due removal in randomForestSRC version ",
    "3.4.0."
  )
  
  if (as_error) {
    rlang::abort(
      message = message_string,
      class = c("familiar_error", "deprecation_error")
    )
  }
  
  if (!.is_testing()) {
    rlang::warn(
      message = message_string,
      class = c("familiar_warning", "deprecation_warning"),
      .frequency = "once",
      .frequency_id = "deprecation_warning_rfsrc_minimum_depth"
    )
  }
  
}



..error <- function(
    ...,
    error_class = NULL,
    call = rlang::caller_env()
) {
  # Generic error function.
  message_string <- paste0(list(...), collapse = "")
  rlang::abort(
    message = message_string,
    class = union("familiar_error", error_class),
    call = call
  )
}



..error_no_predictions_possible <- function(
    object,
    prediction_type,
    call = rlang::caller_env()
) {
  message_string <- paste0(
    "Predictions of the ", prediction_type, " type are not possible ",
    "using the ", object@learner, " model (class: ", class(object)[1L],
    ") for ", object@outcome_type, " outcomes."
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "prediction_type_error"),
    call = call
  )
}



..error_no_known_outcome_type <- function(outcome_type, call = rlang::caller_env()) {
  message_string <- paste0(
    "Outcome type was not recognised. Found: ", outcome_type, ". ",
    "One of binomial, multinomial, continuous, or survival was expected."
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "outcome_type_error"),
    call = call
  )
}



..error_outcome_type_not_implemented <- function(outcome_type, call = rlang::caller_env()) {
  message_string <- paste0(
    outcome_type, " is currently not supported, but scheduled for future implementation."
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "outcome_type_error"),
    call = call
  )
}



..error_data_set_is_empty <- function(call = rlang::caller_env()) {
  rlang::abort(
    message = "The provided dataset does not contain any samples.",
    class = c("familiar_error", "dataset_error"),
    call = call
  )
}



..error_data_set_has_no_features <- function(call = rlang::caller_env()) {
  rlang::abort(
    message = "The provided dataset has no associated feature data.",
    class = c("familiar_error", "dataset_error"),
    call = call
  )
}



..error_data_set_has_no_outcome <- function(call = rlang::caller_env()) {
  rlang::abort(
    message = "The provided dataset lacks outcome data.",
    class = c("familiar_error", "dataset_error"),
    call = call
  )
}


..error_input_missing_without_default <- function(
    var_name, 
    allow_config = FALSE,
    call = rlang::caller_env()
) {
  message_string <- paste0(
    var_name, " ", ifelse(allow_config, "tag/argument", "argument"), " is missing ",
    "and does not have a default setting. Please provide the ", var_name,
    ifelse(allow_config, " tag/argument.", " argument.")
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_input_not_unique <- function(
    x, 
    var_name, 
    allow_config = FALSE,
    call = rlang::caller_env()
) {
  
  # Suppress NOTES due to non-standard evaluation in data.table
  n <- NULL
  
  # Identify duplicate values
  test_table <- data.table::data.table("value" = x)[, list("n" = .N), by = "value"]
  dupl_value <- test_table[n > 1L]$value
  
  message_string <- paste0(
    var_name, " ", ifelse(allow_config, "tag/argument", "argument"),
    " has ", ifelse(length(dupl_value) > 1L, "a duplicate value: ", "multiple duplicate values: "),
    paste_s(dupl_value)
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_type_conversion_not_possible <- function(
    x,
    to_type,
    var_name,
    req_length = 1L,
    allow_more = FALSE,
    call = rlang::caller_env()
) {
  
  message_string <- paste0(
    "Conversion of input for ", var_name, " argument to the desired ", to_type, " class failed. ",
    "Please provide ", req_length, ifelse(allow_more, " or more ", " "),
    ifelse(req_length != 1L || allow_more, "input values ", "input value "),
    "of the correct type. Found: ", paste_s(x)
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_variable_has_too_few_values <- function(
    x,
    var_name,
    req_length = 1L,
    allow_more = FALSE,
    call = rlang::caller_env()
) {
  
  # Handle req_length being a vector with two values.
  if (length(req_length) == 2L) {
    if (!is.finite(req_length[2L])) {
      req_length <- req_length[1L]
      allow_more <- TRUE
      
    } else if (diff(req_length) == 0L) {
      req_length <- req_length[1L]
      allow_more <- FALSE
    }
  }
  
  if (length(req_length) == 1L) {
    # The req_length argument specifies the exact number or the minimum number
    # of values.
    message_string <- paste0(
      "The ", var_name, " argument requires ",
      ifelse(allow_more, "at least ", "exactly "), req_length,
      ifelse(req_length == 1L, " value", " values"), ". ",
      length(x), ifelse(length(x) == 1L, " value was", " values were"), " found."
    )
    
  } else {
    # The req_length argument specifies a range for the number of values.
    message_string <- paste0(
      "The ", var_name, " argument requires between ",
      req_length[1L], " and ", req_length[2L], " values. ",
      ifelse(length(x) == 1L, " value was", " values were"), " found."
    )
  }
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_variable_has_too_many_values <- function(
    x,
    var_name,
    req_length = 1L,
    allow_fewer = FALSE,
    call = rlang::caller_env()
) {
  
  # Handle req_length being a vector with two values.
  if (length(req_length) == 2L) {
    if (!is.finite(req_length[1L])) {
      req_length <- req_length[2L]
      allow_fewer <- TRUE
      
    } else if (diff(req_length) == 0L) {
      req_length <- req_length[2L]
      allow_fewer <- FALSE
    }
  }
  
  if (length(req_length) == 1L) {
    # The req_length argument specifies the exact number or the maximum number
    # of values.
    message_string <- paste0(
      "The ", var_name, " argument requires ",
      ifelse(allow_fewer, "at most ", "exactly "), req_length,
      ifelse(req_length == 1L, " value", " values"), ". ",
      length(x), ifelse(length(x) == 1L, " value was", " values were"), " found."
    )
    
  } else {
    message_string <- paste0(
      "The ", var_name, " argument requires between ",
      req_length[1L], " and ", req_length[2L], " values. ",
      ifelse(length(x) == 1L, " value was", " values were"), " found."
    )
  }
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_value_outside_allowed_range <- function(
    x, 
    var_name, 
    range, 
    call = rlang::caller_env()
) {
  
  # Determine the string for the allowed value.
  if (is.infinite(range[1L]) && is.infinite(range[2L])) {
    allowed_value_text <- "be any real number"
    
  } else if (is.infinite(range[1L])) {
    allowed_value_text <- paste("have a value of", range[2L], "or less")
    
  } else if (is.infinite(range[2L])) {
    allowed_value_text <- paste("have a value of", range[1L], "or more")
    
  } else if (range[1L] - range[2L] == 0.0) {
    allowed_value_text <- paste("be exactly", range[1L])
    
  } else {
    allowed_value_text <- paste("have a value between", range[1L], "and", range[2L])
  }
  
  message_string <- paste0(
    var_name, " is expected to ", allowed_value_text,
    ". Found: ", paste_s(x)
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_value_not_allowed <- function(
    x, 
    var_name, 
    values,
    call = rlang::caller_env()
) {
  
  # Replace NULL in x by "NULL"
  x[is.null(x)] <- "NULL"
  
  # Replace NULL in values by "NULL"
  values[is.null(values)] <- "NULL"
  
  # Identify those values that are not allowed
  forbidden_values <- setdiff(x, values)
  
  message_string <- paste0(
    var_name, " expects one of ", paste_s(values), " as input. ",
    "Found the following unknown values: ", paste_s(forbidden_values)
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_type_not_valid <- function(
    x,
    var_name, 
    valid_type,
    call = rlang::caller_env()
) {
  
  message_string <- paste0(
    var_name, " has type ", paste_s(class(x)),
    " whereas ", paste_s(valid_type), " was expected."
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_value_shared_between_variables <- function(
    x, 
    y, 
    var_name_x, 
    var_name_y,
    call = rlang::caller_env()
) {
  
  # Determine overlap
  overlap <- intersect(x, y)
  
  message_string <- paste0(
    ifelse(length(overlap) > 1L, "Multiple values were", "One value was"),
    " shared between ", var_name_x, " and ", var_name_y, ": ", paste_s(overlap),
    ". No overlap is allowed."
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "input_argument_error"),
    call = call
  )
}



..error_reached_unreachable_code <- function(..., call = rlang::caller_env()) {
  message_string <- paste0(
    "This error should not occur, as this code should not be reachable by design. ",
    "Contact the package maintainer if you see this error.\n\t",
    paste0(list(...), collapse = "")
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "developer_error"),
    call = call
  )
}



..error_ensemble_models_not_loaded <- function(call = rlang::caller_env()) {
  message_string <- paste0(
    "familiarModel objects were not loaded and attached the familiarEnsemble. ",
    "Please use the load_models method to load the models."
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "object_load_error"),
    call = call
  )
}



..error_cannot_convert_to_familiar_object <- function(
    object, 
    expected_class, 
    call = rlang::caller_env()
) {
  
  # Determine what classes are allowed as object
  if (expected_class == "familiarModel") {
    allowed_class <- "familiarModel"
    
  } else if (expected_class == "familiarEnsemble") {
    allowed_class <- c("familiarModel", "familiarEnsemble")
    
  } else if (expected_class == "familiarData") {
    allowed_class <- c("familiarModel", "familiarEnsemble", "familiarData")
    
  } else if (expected_class == "familiarCollection") {
    allowed_class <- c("familiarModel", "familiarEnsemble", "familiarData", "familiarCollection")
  }
  
  message_string <- paste0(
    "The provided object cannot be converted from ", class(object), " to ", expected_class, ". ",
    "The object should have one of the following classes, or inherit it: ", paste_s(allowed_class)
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "object_conversion_error"),
    call = call
  )
}



..error_data_not_prediction_table <- function(
    x, 
    expected_class,
    call = rlang::caller_env()
) {
  message_string <- paste0(
    "A prediction table object of class ", expected_class,
    " was expected, but ", class(x)[1L], " was found."
  )
  
  rlang::abort(
    message = message_string,
    class = c("familiar_error", "incorrect_prediction_table"),
    call = call
  )
}



..error_package_not_installed <- function(
    x, 
    purpose = NULL,
    call = rlang::caller_env()
) {
  
  # Only unique packages.
  x <- unique(x)

  # Basic error message.
  err_message <- ..message_missing_package(x = x, purpose = purpose)
  
  # Instructions for CRAN packages.
  err_message <- c(err_message, ..message_install_from_cran(x = x))
  
  # Instructions for Bioconductor packages.
  err_message <- c(err_message, ..message_install_from_bioconductor(x = x))
  
  rlang::abort(
    message = paste0(err_message, collapse = ""),
    class = c("familiar_error", "package_missing"),
    call = call
  )
}



..error_cannot_update_object <- function(
    object,
    fam_version,
    call = rlang::caller_env()
) {
  # Version of the object.
  object_fam_version <- tail(object@familiar_version, n = 1L)
  
  err_message <- paste0(
    "Familiar version ", fam_version, " introduces breaking changes to ",
    paste_s(class(object)), " objects. ",
    "To use this object, you need to install version familiar ", 
    object_fam_version, ". ",
    "This can be achieved using devtools::install_version('familiar', version = '",
    object_fam_version, "') or remotes::install_version('familiar', version = '",
    object_fam_version, "')."
  )
  
  ..error(
    err_message,
    error_class = "object_update_error",
    call = call
  )
}



..error_updated_object_invalid <- function(
    object,
    call = rlang::caller_env()
) {
  err_message <- paste0(
    "The provided ", paste_s(class(object)), " failed consistency checks after updating. ",
    "Please contact the developers."
  )
  
  ..error(
    err_message,
    error_class = c("object_update_error", "developer_error"),
    call = call
  )
}



..error_message_no_training_data_available <- function() {
  return("No data was available to train the model.")
}



..error_message_no_features_selected_for_training <- function() {
  return("No features were selected to train the model on.")
}



..error_message_no_optimised_hyperparameters_available <- function() {
  return("Model hyperparameters were not optimised.")
}



..error_message_failed_model_coefficient_estimation <- function() {
  return("None of the model coefficients could be estimated.")
}

#' @include FamiliarS4Generics.R
#' @include FamiliarS4Classes.R
NULL



# .train (familiarModel, dataObject) -------------------------------------------
setMethod(
  ".train",
  signature(
    object = "familiarModel",
    data = "dataObject"
  ),
  function(
    object,
    data,
    get_additional_info = FALSE,
    is_pre_processed = FALSE,
    trim_model = TRUE,
    timeout = 60000.0,
    approximate = FALSE,
    ...
  ) {
    # Train method for model training

    # Check if the class of object is a subclass of familiarModel.
    if (!is_subclass(class(object)[1L], "familiarModel")) object <- promote_learner(object)
    
    # Process data, if required.
    data <- process_input_data(
      object = object,
      data = data,
      is_pre_processed = is_pre_processed,
      stop_at = "clustering"
    )

    # Work only with data that has known outcomes when training.
    data <- filter_missing_outcome(data = data)

    # Set the training flags
    can_train <- can_train_naive <- TRUE

    # Check if there are any data entries. The familiar model cannot be trained
    # otherwise. We do allow for no features being present.
    if (is_empty(x = data, allow_no_features = TRUE)) {
      can_train <- can_train_naive <- FALSE
      object <- ..update_errors(
        object = object,
        ..error_message_no_training_data_available()
      )
    }

    # Check the number of features in data; if it has no features, a standard
    # familiar model can not be trained. However, it might be possible to train
    # a naive model.
    if (!has_feature_data(x = data)) {
      can_train <- FALSE
      object <- ..update_errors(
        object = object,
        ..error_message_no_features_selected_for_training()
      )
    }

    # Check if the hyperparameters are plausible.
    if (!has_optimised_hyperparameters(object = object)) {
      can_train <- can_train_naive <- FALSE
      object <- ..update_errors(
        object = object,
        ..error_message_no_optimised_hyperparameters_available()
      )
    }

    # Check if a naive model should be trained.
    if (!requires_naive_model(object)) {
      can_train_naive <- FALSE
    }

    # Add outcome distribution data.
    object@outcome_info <- .compute_outcome_distribution_data(
      object = object@outcome_info,
      data = data
    )

    # Train a new model based on data. If a normal model cannot be trained,
    # attempt to train a naive model.
    if (can_train) {
      object <- ..train(
        object = object,
        data = data,
        approximate = approximate,
        ...
      )
      
    } else if (can_train_naive) {
      object <- ..train_naive(
        object = object,
        data = data,
        ...
      )
    }

    # Extract information required for assessing model performance, calibration
    # (e.g. baseline survival) etc.
    if (get_additional_info) {
      # Remove duplicate samples from the data prior to obtaining additional
      # data.
      data <- aggregate_data(data = data)

      # Create calibration models and add to the object. Not all models require
      # recalibration.
      if (can_train) {
        object <- ..set_recalibration_model(
          object = object,
          data = data
        )
      }

      # Extract data required for assessing calibration. Not all outcome types
      # require calibration info. Currently calibration information is only
      # retrieved for survival outcomes, in the form of baseline survival
      # curves.
      if (can_train || can_train_naive) {
        object <- ..set_calibration_info(
          object = object,
          data = data
        )
      }

      # Set stratification thresholds. This is currently only done for
      # survival outcomes.
      if (can_train) {
        object <- ..set_risk_stratification_thresholds(
          object = object,
          data = data
        )
      }

      # Add column data
      object <- add_data_column_info(
        object = object,
        data = data
      )
    }

    if (trim_model) object <- trim_model(object = object, timeout = timeout)

    # Empty slots if a model can not be trained.
    if (!can_train) {
      object@required_features <- NULL
      object@model_features <- NULL
      object@novelty_features <- NULL
    }

    return(object)
  }
)


# .train (familiarModel, NULL)--------------------------------------------------
setMethod(
  ".train",
  signature(
    object = "familiarModel",
    data = "NULL"
  ),
  function(
    object,
    data,
    ...
  ) {
    # The model cannot be trained, and is returned directly.
    return(object)
  }
)



# show (model) -----------------------------------------------------------------
setMethod(
  "show",
  signature(object = "familiarModel"),
  function(object) {
    # Make sure the model object is updated.
    object <- update_object(object = object)

    if (!model_is_trained(object)) {
      cat(paste0(
        "A ", object@learner, " model (class: ", class(object)[1L],
        ") that was not successfully trained (", 
        .familiar_version_string(object), ").\n"
      ))

      if (length(object@messages$warning) > 0L) {
        condition_messages <- condition_summary(object@messages$warning)
        cat(paste0(
          "\nThe following ",
          ifelse(length(condition_messages) == 1L, "warning was", "warnings were"),
          " generated while trying to train the model:\n",
          paste0(condition_messages, collapse = "\n"),
          "\n"
        ))
      }

      if (length(object@messages$error) > 0L) {
        condition_messages <- condition_summary(object@messages$error)
        cat(paste0(
          "\nThe following ",
          ifelse(length(condition_messages) == 1L, "error was", "errors were"),
          " encountered while trying to train the model:\n",
          paste0(condition_messages, collapse = "\n"),
          "\n"
        ))
      }
      
      return(invisible(NULL))
    }
    
    # Describe the learner and the version of familiar.
    message_str <- paste0(
      "A ", object@learner, " model (class: ", class(object)[1L],
      "; ", .familiar_version_string(object), ")"
    )
    
    # Describe the package(s), if any
    if (!is.null(object@package)) {
      message_str <- c(
        message_str,
        paste0(" trained using "),
        paste_s(mapply(
          ..message_package_version,
          x = object@package,
          version = object@package_version
        )),
        ifelse(length(object@package) > 1L, " packages", " package"
        )
      )
    }
    
    # Complete message and write.
    message_str <- paste0(c(message_str, ".\n"), collapse = "")
    cat(message_str)
    
    cat(paste0("\n--------------- Model details ---------------\n"))
    
    # Model details
    if (object@is_trimmed) {
      cat(object@trimmed_function$show, sep = "\n")
    } else {
      show(object@model)
    }
    
    cat(paste0("\n---------------------------------------------\n"))
    
    # Outcome details
    cat("\nThe following outcome was modelled:\n")
    show(object@outcome_info)
    
    # Details concerning hyperparameters.
    cat("\nThe model was trained using the following hyperparameters:\n")
    invisible(lapply(
      names(object@hyperparameters),
      function(x, object) {
        cat(paste0("  ", x, ": ", object@hyperparameters[[x]], "\n"))
      },
      object = object
    ))
    
    # Details concerning variable importance.
    cat(paste0(
      "\nVariable importance was determined using the ",
      object@vimp_method, " variable importance method.\n"
    ))
    vimp_data <- aggregate_vimp_table(
      object@vimp_table,
      aggregation_method = object@vimp_aggregation_method,
      rank_threshold = object@vimp_rank_threshold
    )
    if (!is.null(vimp_data)) {
      cat(show(get_vimp_table(vimp_data)[order(rank)]))
      cat("\n")
    }
    rm(vimp_data)
    
    # Details concerning model features:
    cat("\nThe following features were used in the model:\n")
    lapply(
      object@model_features,
      function(x, object) show(object@feature_info[[x]]),
      object = object
    )
    
    # Details concerning novelty features:
    if (!model_is_trained(object@novelty_detector)) {
      cat("\nNo novelty detector was trained.\n")
      
    } else if (setequal(object@model_features, object@novelty_features)) {
      cat("\nA novelty detector was trained using the model features.\n")
      
    } else {
      cat("\nA novelty detector was trained using the model features above, and additionally:\n\n")
      
      # Identify novelty features that were set in addition to model
      # features.
      novelty_features <- setdiff(object@novelty_features, object@model_features)
      
      lapply(
        novelty_features, 
        function(x, object) show(object@feature_info[[x]]), 
        object = object
      )
    }
    
    if (length(object@messages$warning) > 0L || length(object@messages$error) > 0L) {
      cat(paste0("\n------------ Warnings and errors ------------\n"))
      
      if (length(object@messages$warning) > 0L) {
        condition_messages <- condition_summary(object@messages$warning)
        cat(paste0(
          "\nThe following ",
          ifelse(length(condition_messages) == 1L, "warning was", "warnings were"),
          " generated while training the model:\n",
          paste0(condition_messages, collapse = "\n")
        ))
      }
      
      if (length(object@messages$error) > 0L) {
        condition_messages <- condition_summary(object@messages$error)
        cat(paste0(
          "\nThe following ",
          ifelse(length(condition_messages) == 1L, "error was", "errors were"),
          " encountered while training the model:\n",
          paste0(condition_messages, collapse = "\n")
        ))
      }
    }
    
    # Check package version.
    check_package_version(object)
  }
)


#' Model summaries
#'
#' `summary` produces model summaries.
#'
#' @param object a familiarModel object
#' @param ... additional arguments passed to `summary` methods for the underlying
#'   model, when available.
#'
#' @details This method extends the `summary` S3 method. For some models
#'   `summary` requires information that is trimmed from the model. In this case
#'   a copy of summary data is stored with the model, and returned.
#'
#' @return Depends on underlying model. See the documentation for the particular
#'   models.
#'
#' @exportMethod summary
#' @rdname summary-methods
#' @md
setGeneric("summary")

# summary-----------------------------------------------------------------------

#' @rdname summary-methods
setMethod(
  "summary",
  signature(object = "familiarModel"),
  function(object, ...) {
    # Make sure the model object is updated.
    object <- update_object(object = object)

    if (!model_is_trained(object)) {
      message("The model was not trained. A summary is not available.")
      return(invisible(NULL))
    }

    # Attempt to retrieve the summary from the model.
    if (object@is_trimmed) {
      if (!is.null(object@trimmed_function$summary)) {
        return(object@trimmed_function$summary)
      }
    }

    # Attempt to capture the summary directly.
    h <- tryCatch(
      summary(object@model, ...),
      error = identity
    )

    # If an error is generated, create a message and return an invisible NULL.
    if (inherits(h, "error")) {
      message("No summary is available for this model.")
      return(invisible(NULL))
    }

    # If the summary is a default one, create a message and return an invisible
    # NULL.
    if (inherits(h, "summaryDefault")) {
      message("No summary is available for this model.")
      return(invisible(NULL))
    }

    return(h)
  }
)



#' Extract model coefficients
#'
#' @param object a familiarModel object
#' @param ... additional arguments passed to `coef` methods for the underlying
#'   model, when available.
#'
#' @details This method extends the `coef` S3 method. For some models `coef`
#'   requires information that is trimmed from the model. In this case a copy of
#'   the model coefficient is stored with the model, and returned.
#'
#' @return Coefficients extracted from the model in the familiarModel object, if
#'   any.
#' @export
#' @rdname coef-methods
#' @md
setGeneric("coef")

# coef -------------------------------------------------------------------------

#' @rdname coef-methods
setMethod(
  "coef",
  signature(object = "familiarModel"),
  function(object, ...) {
    # Make sure the model object is updated.
    object <- update_object(object = object)

    # No training, no coefficients.
    if (!model_is_trained(object)) return(NULL)

    # Attempt to retrieve the coefficients from the model.
    if (object@is_trimmed) {
      if (!is.null(object@trimmed_function$coef)) {
        return(object@trimmed_function$coef)
      }
    }

    # Attempt to capture the coefficients directly.
    feature_coefficients <- tryCatch(
      coef(object@model, ...),
      error = identity
    )

    # If an error is generated by coef, return a NULL.
    if (inherits(feature_coefficients, "error")) return(NULL)
    
    return(feature_coefficients)
  }
)



#' Calculate variance-covariance matrix for a model
#'
#' @param object a familiarModel object
#' @param ... additional arguments passed to `vcov` methods for the underlying
#'   model, when available.
#'
#' @details This method extends the `vcov` S3 method. For some models `vcov`
#'   requires information that is trimmed from the model. In this case a copy of
#'   the variance-covariance matrix is stored with the model, and returned.
#'
#' @return Variance-covariance matrix of the model in the familiarModel object,
#'   if any.
#' @export
#' @rdname vcov-methods
#' @md
setGeneric("vcov")

# vcov -------------------------------------------------------------------------

#' @rdname vcov-methods
setMethod(
  "vcov", signature(object = "familiarModel"),
  function(object, ...) {
    # Make sure the model object is updated.
    object <- update_object(object = object)

    # No training, no variance-covariance matrix
    if (!model_is_trained(object)) return(NULL)

    # Attempt to retrieve the variance-covariance matrix from the model.
    if (object@is_trimmed) {
      if (!is.null(object@trimmed_function$vcov)) {
        return(object@trimmed_function$vcov)
      }
    }

    # Attempt to capture the variance-covariance matrix directly.
    variance_covariance_matrix <- tryCatch(
      vcov(object@model, ...),
      error = identity
    )

    # If an error is generated by vcov, return a NULL.
    if (inherits(variance_covariance_matrix, "error")) return(NULL)

    return(variance_covariance_matrix)
  }
)



# require_package (model) ------------------------------------------------------
setMethod(
  "require_package",
  signature(x = "familiarModel"),
  function(x, purpose = NULL, message_type = "error", ...) {
    # Skip if no package is required.
    if (is_empty(x@package)) return(invisible(TRUE))

    # Set standard purposes for common uses.
    if (!is.null(purpose)) {
      if (purpose %in% c("train", "vimp", "predict", "show", "distribution")) {
        purpose <- switch(purpose,
          "train" = "to train a model",
          "vimp" = "to determine variable importance",
          "predict" = "to create model predictions",
          "show" = "to capture output",
          "distribution" = "to set the model distribution"
        )
      }
    }

    return(invisible(.require_package(
      x = x@package,
      purpose = purpose,
      message_type = message_type
    )))
  }
)



# set_package_version (model) --------------------------------------------------
setMethod(
  "set_package_version",
  signature(object = "familiarModel"),
  function(object) {
    # Do not add package versions if there are no packages.
    if (is_empty(object@package)) return(object)

    # Obtain package versions.
    object@package_version <- sapply(
      object@package,
      function(x) (as.character(utils::packageVersion(x)))
    )

    return(object)
  }
)



# check_package_version (model) ------------------------------------------------
setMethod(
  "check_package_version",
  signature(object = "familiarModel"),
  function(object) {
    .check_package_version(
      name = object@package,
      version = object@package_version,
      when = "at model creation"
    )
  }
)



# save (model) -----------------------------------------------------------------
setMethod(
  "save",
  signature(
    list = "familiarModel",
    file = "character"
  ),
  function(list, file) {
    .save(object = list, dir_path = file)
  }
)



# add_model_name (ANY, familiarModel)-------------------------------------------
setMethod(
  "add_model_name",
  signature(
    data = "ANY",
    object = "familiarModel"
  ),
  function(data, object) {
    if (is_empty(data)) return(NULL)

    ..error_reached_unreachable_code(
      "add_model_name,any,familiarModel: no method for non-empty data."
    )
  }
)



# add_model_name (familiarDataElement, familiarModel)---------------------------
setMethod(
  "add_model_name",
  signature(
    data = "familiarDataElement",
    object = "familiarModel"
  ),
  function(data, object) {
    # Determine the model name
    if (length(object@name) == 0L) {
      model_name <- get_object_name(object = object, abbreviated = TRUE)
    } else {
      model_name <- object@name
    }

    if (is.null(data@identifiers)) {
      data@identifiers <- list("model_name" = model_name)
    } else {
      data@identifiers[["model_name"]] <- model_name
    }

    return(data)
  }
)



# add_model_name (familiarDataElement, character) ------------------------------
setMethod(
  "add_model_name",
  signature(
    data = "familiarDataElement",
    object = "character"
  ),
  function(data, object) {
    # Load object.
    object <- load_familiar_object(object)

    return(do.call(
      add_model_name,
      args = list(
        "data" = data,
        "object" = object
      )
    ))
  }
)



# set_object_name (familiarModel) ----------------------------------------------

#' @title Set the name of a `familiarModel` object.
#'
#' @description Set the `name` slot using the object name.
#'
#' @param x A `familiarModel` object.
#'
#' @return A `familiarModel` object with a generated or a provided name.
#' @md
#' @keywords internal
setMethod(
  "set_object_name",
  signature(x = "familiarModel"),
  function(x, new = NULL) {
    
    if (is.null(x@project_id) && is.null(new)) {
      # Generate a random object name. A project_id of 0 means that the objects
      # was auto-generated (i.e. through object conversion). We randomly
      # generate characters and add a time stamp, so that collision is
      # practically impossible.
      slot(object = x, name = "name") <- paste0(
        as.character(as.numeric(format(Sys.time(), "%H%M%S"))),
        "_", rstring(n = 20L)
      )
      
    } else if (is.null(new)) {
      # Generate a sensible object name.
      slot(object = x, name = "name") <- get_object_name(object = x)
      
    } else {
      slot(object = x, name = "name") <- new
    }

    return(x)
  }
)



# get_object_name (model) ------------------------------------------------------
setMethod(
  "get_object_name",
  signature(object = "familiarModel"),
  function(object, abbreviated = FALSE) {
    
    if (abbreviated) {
      # Create an abbreviated name
      model_name <- paste0("model.", object@data_id, ".", object@run_id)
      
    } else {
      # Create the full name of the model
      model_name <- get_object_file_name(
        learner = object@learner,
        vimp_method = object@vimp_method,
        project_id = object@project_id,
        data_id = object@data_id,
        run_id = object@run_id,
        object_type = "familiarModel",
        with_extension = FALSE
      )
    }

    return(model_name)
  }
)



# model_is_trained (familiarModel) ---------------------------------------------
setMethod(
  "model_is_trained",
  signature(object = "familiarModel"),
  function(object) {
    # Check if a model was trained
    if (is.null(object@model)) {
      # If no model is attached to the object, assume that no model was
      # trained.
      return(FALSE)
      
    } else {
      # If the the model is not NULL, assume that it is present.
      return(TRUE)
    }
  }
)



# model_is_trained (character) -------------------------------------------------
setMethod(
  "model_is_trained",
  signature(object = "character"),
  function(object) {
    # Load object.
    object <- load_familiar_object(object)

    return(do.call(
      model_is_trained,
      args = c(list("object" = object))
    ))
  }
)



# model_is_trained (NULL) ------------------------------------------------------
setMethod(
  "model_is_trained",
  signature(object = "NULL"),
  function(object) {
    return(FALSE)
  }
)



# add_package_version (familiarModel) ------------------------------------------
setMethod(
  "add_package_version",
  signature(object = "familiarModel"),
  function(object) {
    # Set version of familiar
    return(.add_package_version(object = object))
  }
)



# add_data_column_info (familiarModel) -----------------------------------------
setMethod(
  "add_data_column_info",
  signature(object = "familiarModel"),
  function(
    object, 
    data = NULL, 
    sample_id_column = NULL, 
    batch_id_column = NULL, 
    series_id_column = NULL
  ) {
    # Don't determine new column information if this information is already
    # present.
    if (!is.null(object@data_column_info)) return(object)

    # Don't determine new column information if this information can be
    # inherited from a dataObject.
    if (is(data, "dataObject")) {
      if (!is_empty(data@data_column_info)) {
        object@data_column_info <- data@data_column_info

        return(object)
      }
    }

    # Load settings to find identifier columns
    settings <- get_settings()

    # Read from settings. If not set, these will be NULL.
    if (is.null(sample_id_column)) sample_id_column <- settings$data$sample_col
    if (is.null(batch_id_column)) batch_id_column <- settings$data$batch_col
    if (is.null(series_id_column)) series_id_column <- settings$data$series_col

    # Replace any missing.
    if (is.null(sample_id_column)) sample_id_column <- NA_character_
    if (is.null(batch_id_column)) batch_id_column <- NA_character_
    if (is.null(series_id_column)) series_id_column <- NA_character_

    # Repetition column ids are only internal.
    repetition_id_column <- NA_character_

    # Create table
    data_info_table <- data.table::data.table(
      "type" = c("batch_id_column", "sample_id_column", "series_id_column", "repetition_id_column"),
      "internal" = get_id_columns(),
      "external" = c(batch_id_column, sample_id_column, series_id_column, repetition_id_column)
    )

    if (object@outcome_type %in% c("survival", "competing_risk")) {
      # Find internal and external outcome column names.
      internal_outcome_columns <- get_outcome_columns(object@outcome_type)
      external_outcome_columns <- object@outcome_info@outcome_column

      # Add to table
      outcome_info_table <- data.table::data.table(
        "type" = c("outcome_column", "outcome_column"),
        "internal" = internal_outcome_columns,
        "external" = external_outcome_columns
      )
      
    } else if (object@outcome_type %in% c("binomial", "multinomial", "continuous")) {
      # Find internal and external outcome column names.
      internal_outcome_columns <- get_outcome_columns(object@outcome_type)
      external_outcome_columns <- object@outcome_info@outcome_column

      # Add to table
      outcome_info_table <- data.table::data.table(
        "type" = "outcome_column",
        "internal" = internal_outcome_columns,
        "external" = external_outcome_columns
      )
      
    } else {
      ..error_no_known_outcome_type(outcome_type = object@outcome_type)
    }

    # Combine into one table and add to object
    object@data_column_info <- rbind(data_info_table, outcome_info_table)

    return(object)
  }
)



# is_available -----------------------------------------------------------------
setMethod(
  "is_available",
  signature(object = "familiarModel"),
  function(object, ...) {
    return(FALSE)
  }
)



# get_default_hyperparameters --------------------------------------------------
setMethod(
  "get_default_hyperparameters",
  signature(object = "familiarModel"),
  function(object, ...) {
    return(list())
  }
)



# ..train (familiarModel, dataObject) ------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarModel",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Set a NULL model
    object@model <- NULL

    return(object)
  }
)



# ..train (familiarModel, NULL) ------------------------------------------------
setMethod(
  "..train",
  signature(
    object = "familiarModel",
    data = "NULL"
  ),
  function(object, data, ...) {
    # Set a NULL model
    object@model <- NULL

    return(object)
  }
)



# ..train_naive (familiarModel, dataObject) ------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarModel",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Set a NULL model
    object@model <- NULL

    return(object)
  }
)



#### ..train_naive (familiarModel, NULL) ---------------------------------------
setMethod(
  "..train_naive",
  signature(
    object = "familiarModel",
    data = "NULL"
  ),
  function(object, data, ...) {
    # Set a NULL model.
    object@model <- NULL

    return(object)
  }
)



# ..predict (familiarModel, dataObject) ----------------------------------------
setMethod(
  "..predict",
  signature(
    object = "familiarModel",
    data = "dataObject"
  ),
  function(object, data, type = "default", time = NULL, ...) {
    # Fallback method for missing or failing predictions.
    
    # Find the type of prediction table that should be created.
    prediction_table_type <- ..get_prediction_table_type(
      object = object,
      type = type
    )
    
    return(as_prediction_table(
      x = NULL,
      type = prediction_table_type,
      data = data,
      time = time,
      model_object = object
    ))
  }
)



# ..predict (character, dataObject) --------------------------------------------
setMethod(
  "..predict",
  signature(
    object = "character",
    data = "dataObject"
  ),
  function(object, data, ...) {
    # Load object.
    object <- load_familiar_object(object)

    return(do.call(
      ..predict,
      args = c(list(
        "object" = object,
        "data" = data
      ))
    ))
  }
)



# ..get_prediction_table_type (familiarModel) ----------------------------------
setMethod(
  "..get_prediction_table_type",
  signature(object = "familiarModel"),
  function(object, type, ...) {
    # This is the default option for familiar models. Learners that require more
    # specialised prediction table types should specify these types in their own
    # methods before calling this default method.
    prediction_table_type <- NULL
    if (type == "default") {
      if (object@outcome_type %in% c("binomial", "multinomial")) {
        prediction_table_type <- "classification"
        
      } else if (object@outcome_type %in% c("continuous")) {
        prediction_table_type <- "regression"
        
      } else if (object@outcome_type %in% c("survival")) {
        prediction_table_type <- "hazard_ratio"
        
      } else {
        ..error_outcome_type_not_implemented(object@outcome_type)
      }
      
    } else if (type == "survival_probability" && object@outcome_type == "survival") {
      prediction_table_type <- "survival_probability"
      
    } else {
      ..error_no_predictions_possible(object, type)
    }
    
    return(prediction_table_type)
  }
)




# ..set_calibration_info -------------------------------------------------------
setMethod(
  "..set_calibration_info",
  signature(object = "familiarModel"),
  function(object, data) {
    # This is a fall-back option.
    if (is.null(object@calibration_info)) object@calibration_info <- NULL

    return(object)
  }
)



# ..set_recalibration_model ----------------------------------------------------
setMethod(
  "..set_recalibration_model",
  signature(
    object = "familiarModel",
    data = "dataObject"
  ),
  function(object, data) {
    # This is a fall-back option.
    object@calibration_model <- NULL

    return(object)
  }
)



# ..set_risk_stratification_thresholds -----------------------------------------
setMethod(
  "..set_risk_stratification_thresholds",
  signature(
    object = "familiarModel",
    data = "dataObject"
  ),
  function(object, data) {
    
    if (
      object@outcome_type %in% c("survival", "competing_risk") &&
      model_is_trained(object)
    ) {
      
      object@km_info <- .find_survival_grouping_thresholds(
        object = object,
        data = data
      )
      
    } else {
      object@km_info <- NULL
    }

    return(object)
  }
)



# ..set_vimp_parameters --------------------------------------------------------
setMethod(
  "..set_vimp_parameters",
  signature(object = "familiarModel"),
  function(object, ...) {
    # This is a fall-back option.
    return(object)
  }
)



# ..vimp -----------------------------------------------------------------------
setMethod(
  "..vimp",
  signature(object = "familiarModel"),
  function(object, ...) {
    # This is a fall-back option.
    return(get_placeholder_vimp_table(
      vimp_method = object@learner,
      run_table = object@run_table
    ))
  }
)



# trim_model (familiarModel)----------------------------------------------------
setMethod(
  "trim_model",
  signature(object = "familiarModel"),
  function(object, timeout = 60000.0, ...) {
    # Do not trim the model if there is nothing to trim.
    if (!model_is_trained(object)) return(object)

    # Trim the model.
    trimmed_object <- .trim_model(object = object)

    # Skip further processing if the model object was not trimmed.
    if (!trimmed_object@is_trimmed) return(object)

    # Go over different functions.
    trimmed_object <- .replace_broken_functions(
      object = object,
      trimmed_object = trimmed_object,
      timeout = timeout
    )

    return(trimmed_object)
  }
)



# .trim_model (familiarModel)---------------------------------------------------
setMethod(
  ".trim_model",
  signature(object = "familiarModel"),
  function(object, ...) {
    # Default method for models that lack a more specific method.
    return(object)
  }
)



# requires_naive_model ---------------------------------------------------------
setMethod(
  "requires_naive_model",
  signature(object = "familiarModel"),
  function(object, ...) {
    # Determine if the model hyperparameters specify that a naive model should
    # be trained. This obviously does not work if there no hyperparameters to
    # speak of.
    if (!has_optimised_hyperparameters(object = object)) return(FALSE)
    
    # Check if the signature size hyperparameter exists.
    if (is.null(object@hyperparameters$sign_size)) return(FALSE)

    # Check if the no-features variable importance.
    if (object@vimp_method %in% .get_available_no_features_vimp_methods()) return(TRUE)

    # Check if the signature size is 0.
    return(all(object@hyperparameters$sign_size == 0L))
  }
)



# ..update_warnings-------------------------------------------------------------
setMethod(
  "..update_warnings",
  signature(object = "familiarModel"),
  function(object, messages) {
    # Update warnings attached to the object.
    object@messages$warning <- c(
      object@messages$warning,
      messages
    )

    return(object)
  }
)


# ..update_errors---------------------------------------------------------------
setMethod(
  "..update_errors",
  signature(object = "familiarModel"),
  function(object, messages) {
    # Update error attached to the object.
    object@messages$error <- c(
      object@messages$error,
      messages
    )

    return(object)
  }
)


# has_calibration_info ---------------------------------------------------------
setMethod(
  "has_calibration_info",
  signature(object = "familiarModel"),
  function(object) {
    return(!is.null(object@calibration_info))
  }
)



# set_model_features (familiarModel)--------------------------------------------
setMethod(
  "set_model_features",
  signature(object = "familiarModel"),
  function(
    object,
    minimise_footprint = FALSE,
    ...
  ) {
    # Get signature features. This is a list features after clustering.
    signature_features <- get_signature_features(object = object)

    # Find important features, i.e. those that constitute the signature either
    # individually or as part of a cluster. The resulting list of features are
    # features prior to clustering.
    model_features <- get_model_features(
      x = signature_features,
      is_clustered = TRUE,
      feature_info_list = object@feature_info
    )
    
    # Find novelty features. The resulting list of features are features prior
    # to clustering.
    novelty_features <- find_novelty_features(
      model_features = model_features,
      feature_info_list = object@feature_info
    )

    if (minimise_footprint) {
      # Find only features that are required for running the model.
      required_features <- union(model_features, novelty_features)
      
    } else {
      # Find features that are required for processing the data.
      required_features <- get_required_features(
        x = union(model_features, novelty_features),
        is_clustered = FALSE,
        feature_info_list = object@feature_info
      )
    }

    # Select only necessary feature info objects.
    available_feature_info <- names(object@feature_info) %in% required_features
    object@feature_info <- object@feature_info[available_feature_info]

    # Set feature-related attribute slots
    object@required_features <- required_features
    object@model_features <- model_features
    object@novelty_features <- novelty_features

    return(object)
  }
)



# get_signature_features (familiarModel)----------------------------------------
setMethod(
  "get_signature_features",
  signature(object = "familiarModel"),
  function(
    object
  ) {
    # Attempt to get signature directly from the object.
    if (!is_empty(object@model_features)) {
      # If model features have been set (i.e. using set_model_features), 
      # return those features instead. Since the model_features attribute stores
      # feature names prior to clustering, make sure to update these.
      return(features_after_clustering(
        features = object@model_features,
        feature_info_list = object@feature_info
      ))
    }

    # Get signature based on the stored feature information.
    return(.get_signature_features(object))
  }
)



.get_signature_features <- function(object) {
  # Suppress NOTES due to non-standard evaluation in data.table
  name <- rank <- NULL
  
  if (!(is(object, "familiarModel") || is(object, "familiarNoveltyDetector"))) {
    ..error_reached_unreachable_code(paste0("invalid object class: ", class(object)))
  }
  
  # Get signature size. This derives directly from the "sign_size"
  # hyperparameter. If absent, set signature size to infinite.
  signature_size <- object@hyperparameters$sign_size
  if (is_empty(signature_size)) signature_size <- Inf
  
  # Determine which features are pre-assigned to the signature.
  signature_features <- names(object@feature_info)[sapply(object@feature_info, is_in_signature)]
  
  if (object@vimp_method %in% .get_available_signature_only_vimp_methods()) {
    # Only select signature.
    if (length(signature_features) == 0L) {
      ..error(
        "No signature was provided.",
        error_class = "input_argument_error"
      )
    }
    
    selected_features <- signature_features
    
  } else if (object@vimp_method %in% .get_available_none_vimp_methods()) {
    # Select all features.
    selected_features <- features_after_clustering(
      features = get_available_features(feature_info_list = object@feature_info),
      feature_info_list = object@feature_info
    )
    
    # Order randomly so that there is no accidental dependency on order.
    selected_features <- fam_sample(
      x = selected_features,
      size = length(selected_features),
      replace = FALSE
    )
    
  } else if (object@vimp_method %in% .get_available_random_vimp_methods()) {
    # Select all features.
    selected_features <- features_after_clustering(
      features = get_available_features(feature_info_list = object@feature_info),
      feature_info_list = object@feature_info
    )
    
    # Shrink signature sizes that are too large.
    if (signature_size > length(selected_features)) {
      signature_size <- length(selected_features)
    }
    
    # Randomly pick the signature.
    selected_features <- fam_sample(
      x = selected_features,
      size = signature_size,
      replace = FALSE
    )
    
  } else if (object@vimp_method %in% .get_available_no_features_vimp_methods()) {
    # No features are selected.
    selected_features <- NULL
    
  } else {
    # Select signature and any additional features according to rank.
    selected_features <- signature_features
    
    # Get number remaining available features
    n_allowed_features <- signature_size - length(signature_features)
    
    # Check that features may be added, and the rank table is not empty.
    if (n_allowed_features > 0L && !is_empty(object@vimp_table)) {
      # Get available features.
      features <- features_after_clustering(
        features = get_available_features(feature_info_list = object@feature_info),
        feature_info_list = object@feature_info
      )
      
      # Remove signature features, if any, to prevent duplicates.
      features <- setdiff(features, signature_features)
      
      # Extract aggregated rank table. First, ensure that the associated cluster
      # table is correct.
      vimp_table <- update_vimp_table_to_reference(
        x = object@vimp_table,
        reference_cluster_table = .create_clustering_table(
          feature_info_list = object@feature_info
        )
      )
    
      # Recluster the data according to the clustering table corresponding to the
      # model. This ensures that the variable importance table has the features
      # that are seen by the model.
      vimp_table <- recluster_vimp_table(vimp_table)
      
      # Get aggregate variable importances
      vimp_table <- aggregate_vimp_table(
        vimp_table,
        aggregation_method = object@vimp_aggregation_method,
        rank_threshold = object@vimp_rank_threshold
      )
      
      if (is_empty(vimp_table)) return(signature_features)
      
      # Keep only feature ranks of feature corresponding to available
      # features, and order by rank.
      rank_table <- get_vimp_table(vimp_table)[name %in% features, ][order(rank)]
      
      # Add good features (low rank) to the selection
      selected_features <- c(
        signature_features,
        head(x = rank_table, n = n_allowed_features)$name
      )
    }
  }
  
  return(selected_features)
}



.get_available_risklike_prediction_types <- function() {
  return(c("hazard_ratio", "cumulative_hazard", "survival_probability"))
}

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

familiar:::test_plots(
  plot_function = familiar:::plot_ice,
  not_available_some_predictions_fail = FALSE,
  data_element = "ice_data",
  debug = debug_flag
)

# The default limit for the number of important features exceeds the number of
# features in the data, which means that no actual feature selection has 
# to take place. To test all path ways, we set the number of important features
# to one here. Because the test routine relies on "none" as variable importance
# method, this triggers ad-hoc feature selection when computing the dataset.
# This will fail if outcome data are absent, mostly absent, or singular.
familiar:::test_plots(
  plot_function = familiar:::plot_ice,
  not_available_some_predictions_fail = FALSE,
  not_available_all_prospective = TRUE,
  not_available_mostly_prospective = TRUE,
  not_available_single_sample = TRUE,
  data_element = "ice_data",
  n_important_features = 1L,
  debug = debug_flag
)

testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Test that creates a single naive model.
familiar:::integrated_test(
  experimental_design = "fs+mb",
  vimp_method = "no_features",
  cluster_method = "none",
  imputation_method = "simple",
  estimation_type = "point",
  parallel = FALSE,
  debug = debug_flag
)

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

# Default
results <- familiar:::test_export_specific(
  export_function = familiar:::export_sample_similarity,
  data_element = "sample_similarity",
  sample_limit = 20L,
  debug = FALSE
)

testthat::test_that("Sample limit is correctly propagated", {
  for (data_set in results) {
    # Get sample names.
    sample_names <- unique(c(
      data_set[[1]]@data$sample_1,
      data_set[[1]]@data$sample_2))

    testthat::expect_equal(length(sample_names), 20)
  }
})

results <- familiar:::test_export_specific(
  export_function = familiar:::export_sample_similarity,
  data_element = "sample_similarity",
  export_args = list("sample_limit" = 20L),
  debug = FALSE)

testthat::test_that("Sample limit is correctly propagated", {
  for (data_set in results) {
    # Get sample names.
    sample_names <- unique(c(
      data_set[[1]]@data$sample_1,
      data_set[[1]]@data$sample_2))

    testthat::expect_equal(length(sample_names), 20)
  }
})

# Don't perform any further tests on CRAN due to running time.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Default
results <- familiar:::test_export_specific(
  export_function = familiar:::export_ice_data,
  data_element = "ice_data",
  sample_limit = 20L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that("Sample limit is correctly propagated", {
  for (data_set in results) {
    # Get sample names.
    sample_names <- unique(c(data_set[[1]]@data$sample))

    testthat::expect_equal(length(sample_names), 20)
  }
})

# Test that test_export_specific works when specifying a single feature.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_ice_data,
  outcome_type_available = "survival",
  data_element = "ice_data",
  features = c("feature_1"),
  sample_limit = 20L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that("A single feature can be specified.", {
  testthat::expect_equal(results$survival[[1]]@identifiers$feature_x, "feature_1")
  testthat::expect_equal(length(results$survival), 1L)
})

# Test that test_export_specific works when specifying two features.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_ice_data,
  outcome_type_available = "survival",
  data_element = "ice_data",
  features = c("feature_4", "feature_1"),
  sample_limit = 20L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that("Two features can be specified.", {
  testthat::expect_equal(results$survival[[1]]@identifiers$feature_x, "feature_4")
  testthat::expect_equal(results$survival[[1]]@identifiers$feature_y, "feature_1")
  testthat::expect_equal(length(results$survival), 1L)
})

# Test that test_export_specific works when specifying two features and their
# ranges.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_ice_data,
  outcome_type_available = "survival",
  data_element = "ice_data",
  features = c("feature_4", "feature_1"),
  feature_x_range = c("better", "best"),
  feature_y_range = c(0.25, 0.75),
  n_sample_points = 5L,
  sample_limit = 20L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that("Ranges and features can specified.", {
  testthat::expect_setequal(unique(results$survival[[1]]@data$feature_x_value), c("better", "best"))
  testthat::expect_true(all(results$survival[[1]]@data$feature_y_value <= 0.75))
  testthat::expect_true(all(results$survival[[1]]@data$feature_y_value >= 0.25))
})

# Test that test_export_specific works when specifying two features and their
# specific ranges.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_ice_data,
  outcome_type_available = "survival",
  data_element = "ice_data",
  features = c("feature_4", "feature_3a"),
  feature_x_range = c("good", "better"),
  feature_y_range = c("round", "square"),
  sample_limit = 20L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that("Ranges and features can specified.", {
  testthat::expect_setequal(unique(results$survival[[1]]@data$feature_x_value), c("good", "better"))
  testthat::expect_setequal(unique(results$survival[[1]]@data$feature_y_value), c("round", "square"))
})


# Test that test_export_specific works when specifying n_important_features.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_ice_data,
  outcome_type_available = "continuous",
  data_element = "ice_data",
  n_important_features = 2L,
  sample_limit = 20L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that("The number of important features can be set.", {
  testthat::expect_length(results$continuous, 2L)
  testthat::expect_setequal(
    sapply(results$continuous, function(x) (x@identifiers$feature_x)), 
    c("feature_1", "feature_2b")
  )
})

# Don't perform any further tests on CRAN due to time of running the complete
# test.
testthat::skip_on_cran()
testthat::skip_on_ci()

debug_flag <- FALSE

# Without setting the number of important features.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_permutation_vimp,
  data_element = "permutation_vimp",
  outcome_type_available = "continuous",
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that(
  "all feature are assessed",
  {
    data <- results$continuous[[1L]]@data
    features <- unique(data$feature)
    testthat::expect_length(features, 6L)
  }
)


# With setting the number of important features to 2.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_permutation_vimp,
  data_element = "permutation_vimp",
  outcome_type_available = "continuous",
  n_important_features = 2L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that(
  "the 2 most important features are assessed",
  {
    data <- results$continuous[[1L]]@data
    features <- unique(data$feature)
    testthat::expect_length(features, 2L)
    testthat::expect_setequal(features, c("feature_1", "feature_2b"))
  }
)


# With setting the number of important features to 2.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_permutation_vimp,
  data_element = "permutation_vimp",
  outcome_type_available = "continuous",
  n_important_features = 2L,
  n_models = 3L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)


# With setting the number of important features to 1.
results <- familiar:::test_export_specific(
  export_function = familiar:::export_permutation_vimp,
  data_element = "permutation_vimp",
  outcome_type_available = "continuous",
  n_important_features = 1L,
  create_novelty_detector = FALSE,
  debug = debug_flag
)

testthat::test_that(
  "the most important feature is assessed",
  {
    data <- results$continuous[[1L]]@data
    features <- unique(data$feature)
    testthat::expect_length(features, 1L)
    testthat::expect_setequal(features, "feature_1")
  }
)

---
title: "Performance metrics"
author: "Alex Zwanenburg"
date: "2026-04-22"
output:
  rmarkdown::html_vignette:
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
  rmarkdown::github_document:
    html_preview: FALSE
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
bibliography: "refs.bib"
vignette: >
  %\VignetteIndexEntry{Performance metrics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



When we train a model, we usually want to know how good the model is. Model
performance is assessed using different metrics that quantify how well a model
discriminates cases, stratifies groups or predicts values. `familiar` implements
metrics that are typically used to assess the performance of categorical,
regression and survival models.

# Categorical outcomes

Performance metrics for models with categorical outcomes, i.e. `binomial` and
`multinomial` are listed below.

| **method**                              | **tag**                                              | **averaging** |
|:----------------------------------------|:-----------------------------------------------------|:-------------:|
| accuracy                                | `accuracy`                                           |               |
| area under the receiver-operating curve | `auc`, `auc_roc`                                     |               |
| balanced accuracy                       | `bac`, `balanced_accuracy`                           |               |
| balanced error rate                     | `ber`, `balanced_error_rate`                         |               |
| Brier score                             | `brier`                                              |               |
| Cohen's kappa                           | `kappa`, `cohen_kappa`                               |               |
| f1 score                                | `f1_score`                                           |       ×       |
| false discovery rate                    | `fdr`, `false_discovery_rate`                        |       ×       |
| informedness                            | `informedness`                                       |       ×       |
| markedness                              | `markedness`                                         |       ×       |
| Matthews' correlation coefficient       | `mcc`, `matthews_correlation_coefficient`            |               |
| negative predictive value               | `npv`                                                |       ×       |
| positive predictive value               | `precision`, `ppv`                                   |       ×       |
| recall                                  | `recall`, `sensitivity`, `tpr`, `true_positive_rate` |       ×       |
| specificity                             | `specificity`, `tnr`, `true_negative_rate`           |       ×       |
| Youden's J                              | `youden_j`, `youden_index`                           |       ×       |

Table: Overview of performance metrics for categorical outcomes. Some contingency
table-based metrics require averaging for `multinomial` outcomes as their
original definition only covers `binomial` problems with two classes.

## Area under the receiver-operating curve

The area under the receiver-operating curve is quite literally that. It is the
area under the curve created by plotting the true positive rate (sensitivity)
against the false positive rate (1-specificity). TPR and FPR are derived from a
contingency table, which is created by comparing predicted class probabilities
against a threshold. The receiver-operating curve is created by iterating over
threshold values. The AUC of a model that predicts perfectly is $1.0$, while
$0.5$ indicates predictions that are no better than random.

The implementation in `familiar` does not directly use the ROC curve to compute the AUC.
Instead, an algebraic equation by @Hand2001-ij is used. For `multinomial`
outcomes the AUC is computed for each pairwise comparison of outcome classes,
i.e. one-versus-all, and averaged [@Hand2001-ij].

## Brier score

The Brier score [@Brier1950-lo] is a measure of deviation of predicted
probabilities from the ideal situation where the probability for class *a* is
$1.0$ if the observed class is *a* and $0.0$ if it is not *a*. Hence, it can be
viewed as a measure of calibration as well. A value of $0.0$ is optimal.

The implementation in `familiar` iterates over all outcome classes in a
one-versus-all approach, as originally devised by @Brier1950-lo. For `binomial`
outcomes, the score is divided by 2, so that it falls in the $[0.0, 1.0]$ range.

## Contingency table-based metrics

A contingency table or confusion matrix shows observed and predicted
classes. When dealing with two classes, e.g. *a* and *b*, one of the classes is
usually termed 'positive' and the other 'negative'. For example, let *b* be the
'positive' class. Then we can define the following four categories:

-   True positive ($TP$): *b* is predicted and observed.
-   True negative ($TN$): *a* is predicted and observed.
-   False negative ($FN$): *b* was observed, but *a* was predicted.
-   False positive ($FP$): *a* was observed, but *b* was predicted.

A contingency table contains the occurrence of each of the four cases. If a
model is good, most samples will be either true positive or true negative.
Models that are not as good may have larger numbers of false positives and/or
false negatives.

Metrics based on the contingency table use two or more of the four categories to
characterise model performance. The extension from two classes (`binomial`) to
more (`multinomial`) is often not trivial. For many metrics, `familiar` uses a
one-versus-all approach. Here, all classes are iteratively used as the
'positive' class, with the rest grouped as 'negative'. Three options can be
used to obtain performance values for `multinomial` problems, with an
implementation similar to that of `scikit.learn`:

-   `micro`: The number of true positives, true negatives, false positives and
    false negatives are computed for each class iteration, and then summed over
    all classes. The score is calculated afterwards.

-   `macro`: A score is computed for each class iteration, and then averaged.

-   `weighted`: A score is computed for each class iteration, and the averaged
    with a weight corresponding to the number of samples with the observed
    'positive' class, i.e. the prevalence.

By default, `familiar` uses `macro`, but the averaging procedure may be selected
through appending `_micro`, `_macro` or `_weighted` to the name of the metric.
For example, `recall_micro` will compute the recall metric using `micro`
averaging.

Averaging only applies to `multinomial` outcomes. No averaging is performed for
`binomial` problems with two classes. In this case `familiar` will always
consider the second class level to correspond to the 'positive' class.

### Accuracy

Accuracy quantifies the number of correctly predicted classes:
$s_{acc}=(TP + TN) / (TP + TN + FP + FN)$. The extension to more than 2 classes
is trivial. No averaging is performed for the accuracy metric.

### Balanced accuracy

Accuracy is known to be sensitive to imbalances in the class distribution. A
balanced accuracy was therefore defined [@Brodersen2010-vb], which is the
averaged within-class true positive rate (also known as recall or sensitivity):
$s_{bac}=0.5 (TP / (TP + FN) + TN / (TN + FP))$.

The extension to more than 2 classes involves summation of in-class true
positive rate $TP / (TP + FN)$ for each positive class and subsequent division
by the number of classes. No averaging is performed for balanced accuracy.

### Balanced error rate

The balanced error rate is closely related to balanced accuracy, i.e. instead of
the in-class true positive rate, the in-class false negative rate is used:
$s_{ber}=0.5 (FN / (TP + FN) + FP / (TN + FP))$.

The extension to more than 2 classes involves summation of in-class false
negative rate $FN / (TP + FN)$ for each positive class and subsequent division
by the number of classes. No averaging is performed for balanced error rate.

### F1 score

The F1 score is the harmonic mean of precision and sensitivity:
$s_{f1} = 2 \; TP / (2 \; TP + FP + FN)$.

The metric is not invariant to class permutation. Averaging is therefore
performed for `multinomial` outcomes.

### False discovery rate

The false discovery rate quantifies the proportion of false positives among all
predicted positives, i.e. the Type I error: $s_{fdr} = FP / (TP + FP)$.

The metric is not invariant to class permutation. Averaging is therefore
performed for `multinomial` outcomes.

### Informedness

Informedness is a generalisation of Youden's J statistic [@Powers2011-jt].
Informedness can be extended to multiple classes, and no averaging is therefore
required.

For `binomial` problems, informedness and the Youden J statistic are the same.

### Cohen's kappa

Cohen's kappa coefficient is a measure of correspondence between the observed
and predicted classes [@Cohen1960-kc]. Cohen's kappa coefficient is invariant to
class permutations and no averaging is performed for Cohen's kappa.

### Markedness

Markedness is related to the precision or positive predictive value
[@Powers2011-jt].

### Matthews correlation efficient

Matthews' correlation coefficient measures the correlation between observed and
predicted classes [Matthews1975-kh].

An extension to multiple classes, i.e. multinomial outcomes, was devised by
@Gorodkin2004-tx.

### Negative predictive value

The negative predictive value (NPV) is the fraction of predicted negative
classes that were also observed to be negative: $s_{npv} = TN / (TN + FN)$.

The NPV is not invariant to class permutations. Averaging is performed for
`multinomial` outcomes.

### Positive predictive value

The positive predictive value (PPV) is the fraction of predicted positive
classes that were also observed to be positive: $s_{ppv} = TP / (TP + FP)$.

The PPV is also referred to as precision. The PPV is not invariant to class
permutations. Averaging is performed for `multinomial` outcomes.
`micro`-averaging effectively computes the accuracy.

### Recall

Recall, also known as sensitivity or true positive rate, is the fraction of
observed positive classes that were also predicted to be positive:
$s_{recall} = TP / (TP + FN)$.

Recall is not invariant to class permutations and averaging is performed for
`multinomial` outcomes. Both `micro` and `weighted` averaging effectively
compute the accuracy.

### Specificity

Specificity, also known as the true negative rate, is the fraction of observed
negative classes that were also predicted to be negative:
$s_{spec} = TN / (TN + FP)$.

Specificity is not invariant to class permutations and averaging is performed
for `multinomial` outcomes.

### Youden's J statistic

Youden's J statistic [@Youden1950-no] is the sum of recall and specificity minus
1: $s_{youden} = TP / (TP + FN) + TN / (TN + FP) - 1$.

Youden's J statistic is not invariant to class permutations and averaging is
performed for `multinomial` outcomes.

For `binomial` problems, informedness and the Youden J statistic are the same.

# Regression outcomes

Performance metrics for models with regression outcomes, i.e. `continuous`,
are listed below.

| **method**                 | **tag**                               |
|:---------------------------|:--------------------------------------|
| explained variance         | `explained_variance`                  |
| mean absolute error        | `mae`, `mean_absolute_error`          |
| relative absolute error    | `rae`, `relative_absolutive_error`    |
| mean log absolute error    | `mlae`, `mean_log_absolute_error`     |
| mean squared error         | `mse`, `mean_squared_error`           |
| relative squared error     | `rse`, `relative_squared_error`       |
| mean squared log error     | `msle`, `mean_squared_log_error`      |
| median absolute error      | `medae`, `median_absolute_error`      |
| R<sup>2</sup> score        | `r2_score`, `r_squared`               |
| root mean square error     | `rmse`, `root_mean_square_error`      |
| root relative squared error| `rrse`, `root_relative_squared_error` |
| root mean square log error | `rmsle`, `root_mean_square_log_error` |

Each of the above metrics can be made more robust against rare outliers by
appending `_winsor` or `_trim` as a suffix to the metric name. This respectively
performs winsorising (clipping) and trimming (truncating) based on the absolute
prediction error, prior to computing the metric. Winsorising clips the predicted
values for 5% of the instances with the most extreme absolute errors prior to
computing the performance metric, whereas trimming removes these instances. For
example, winsorised and trimmed versions of the mean squared error metric are
specified as `mse_winsor` and `mse_trim` respectively.

Let $y$ be the set of observed values, and $\hat{y}$ the corresponding predicted
values. The error is then $\epsilon = y-\hat{y}$.

## Explained variance

The explained variance is defined as
$1 - \text{Var}\left(\epsilon\right) / \text{Var}\left(y\right)$. This metric is
not sensitive to differences in offset between observed and predicted values.

## Mean absolute error

The mean absolute error is defined as $1/N \sum_i^N \left|\epsilon_i\right|$,
with $N$ the number of samples.

## Relative absolute error

The relative absolute error is defined as $\sum_i^N \left|\epsilon_i\right|/ \sum_i^N \left|y_i - \bar{y}\right|$.

## Mean log absolute error

The mean log absolute error is defined as
$1/N \sum_i^N \log(\left|\epsilon_i\right| + 1)$.

## Mean squared error

The mean squared error is defined as $1/N \sum_i^N \left(\epsilon_i \right)^2$.

## Relative squared error

The relative squared error is defined as $\sum_i^N \left(\epsilon_i\right)^2/ \sum_i^N \left(y_i - \bar{y}\right)^2$.

## Mean squared log error

Mean squared log error is defined as
$1/N \sum_i^N \left(\log \left(y_i + 1\right) - \log\left(\hat{y}_i + 1\right)\right)^2$.
Note that this score only applies to observed and predicted values in the
positive domain. It is not defined for negative values.

## Median absolute error

The median absolute error is the median of absolute error
$\left|\epsilon\right|$.

## R<sup>2</sup> score

The R<sup>2</sup> score is defined as:
$$R^2 = 1 - \frac{\sum_i^N(\epsilon_i)^2}{\sum_i^N(y_i - \bar{y})^2}$$ Here
$\bar{y}$ denotes the mean value of $y$.

## Root mean square error

The root mean square error is defined as
$\sqrt{1/N \sum_i^N \left(\epsilon_i \right)^2}$.

## Root relative squared error

The root relative squared error is defined as $\sqrt{\sum_i^N \left(\epsilon_i\right)^2/ \sum_i^N \left(y_i - \bar{y}\right)^2}$.

## Root mean square log error

The root mean square log error is defined as
$\sqrt{1/N \sum_i^N \left(\log \left(y_i + 1\right) - \log\left(\hat{y}_i + 1\right)\right)^2}$.
Note that this score only applies to observed and predicted values in the
positive domain. It is not defined for negative values.

# Survival outcomes

Performance metrics for models with survival outcomes, i.e. `survival`, are
listed below.

+--------------------------------------------+---------------------------------+
| **method**                                 | **tag**                         |
+:===========================================+:================================+
| concordance index                          | `concordance_index`, `c_index`, |
|                                            | `concordance_index_harrell`,    |
|                                            | `c_index_harrell`               |
+--------------------------------------------+---------------------------------+

## Concordance index

The concordance index assesses ordering between observed and predicted values.
Let $T$ be observed times, $c$ the censoring status ($0$: no observed event;
$1$: event observed) and $\hat{T}$ predicted times. Concordance between all
pairs of values is determined as follows [@Pencina2004-ii]:

-   Concordant: a pair is concordant if $T_i < T_j$ and $\hat{T}_i < \hat{T}_j$
    (provided $c_i=1$), or if $T_i > T_j$ and $\hat{T}_i > \hat{T}_j$ (provided
    $c_j=1$).
-   Discordant: a pair is discordant if $T_i < T_j$ and $\hat{T}_i > \hat{T}_j$
    (provided $c_i=1$), or if $T_i > T_j$ and $\hat{T}_i < \hat{T}_j$ (provided
    $c_j=1$).
-   Tied: a pair is tied if $\hat{T}_i = \hat{T}_j$, provided that $c_i=c_j=1$.
-   Not comparable: otherwise. This occurs, for example, if sample $i$ was
    censored before an event was observed in sample $j$, or both samples were
    censored.

The concordance index is then computed as:
$$ci = \frac{n_{concord} + 0.5 n_{tied}}{n_{concord} + n_{discord} + n_{tied}}$$

# References

---
title: "Introducing familiar"
author: "Alex Zwanenburg"
date: "2026-04-22"
output:
  rmarkdown::html_vignette:
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
  rmarkdown::github_document:
    html_preview: FALSE
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
bibliography: "refs.bib"
vignette: >
  %\VignetteIndexEntry{Introducing familiar}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---




``` r
library(familiar)
library(data.table)
```

Familiar is a package that allows for end-to-end machine learning of tabular
data, with subsequent evaluation and explanation of models. This vignette
provides an overview of its functionality and how to configure and run an
experiment.

# Familiar in brief

This section provides installation instructions, a brief overview of the
package, and the pipeline encapsulated by the `summon_familiar` function that is
used to run an experiment.

## Installing familiar

Stable versions of familiar can be installed from CRAN. `dependencies=TRUE`
prevents being prompted to install packages when using familiar.


``` r
install.packages("familiar", dependencies = TRUE)
```

It can also be installed directly from the GitHub repository:


``` r
require(devtools)
devtools::install_github("https://github.com/alexzwanenburg/familiar", dependencies = TRUE)
```

## Pipeline

The pipeline implemented in familiar follows a standard machine learning
workflow. A development (training) dataset is used to perform the steps listed
below. Many aspects of these steps can be configured, but the overall process is
fixed:

- **Data processing**: Features in the development dataset are assessed during this step:

  - General feature information: Are features categorical (e.g. has the values
    `FALSE`, `TRUE`) or numeric? Which levels does a categorical or ordinal
     feature have?

  - Invariance: Which features are invariant and should be dropped?

  - Transformation: How should numeric features be transformed using a power
    transformation to make these features behave more according to a normal
    distribution?

  - Normalisation: How should numeric features be normalised to reduce
    differences in scale between features the dataset? Note that familiar also
    allows for normalisation at the batch level to remove systematic differences
    in feature values between different batches or cohorts.

  - Robustness: Should non-robust features, assessed using repeated 
    measurements, be filtered?

  - Importance: Should generally unimportant features be filtered after 
    univariate analysis?

  - Imputation: How should missing feature values be imputed?

  - Redundancy clustering: Which features are similar and should be clustered
    together?

- **Variable importance**: Which features are important for the endpoint of
  interest? Familiar supports various univariate and multivariate variable
  importance methods (see the *Variable importance methods* vignette). Features in
  the data are ranked according to their importance. Based on this information,
  features are selected during hyperparameter optimisation.

- **Hyperparameter optimisation**: Most learners have hyperparameters, which are
  parameters that determine a specific aspect of the model created by the learner.
  Examples are the number of trees in a random forest, the width of the radial
  kernel in support vector machines, and the number of features in the signature
  of a model. Such parameters may significantly influence model performance.
  During hyperparameter optimisation, the aim is to find the set of
  hyperparameters that leads to a generalisable model. Since hyperparameter spaces
  can be high-dimensional, familiar uses Bayesian optimisation for efficiently
  exploring hyperparameter space. The *learning algorithms and hyperparameter optimisation*
  vignette describes model-specific hyperparameters and hyperparameter 
  optimisation in more detail.

- **Model training**: During the final model training step, the development data
  are fitted using the previously determined set of hyperparameters. By default,
  the models are trimmed after creation to remove extraneous information such as
  copies of the development dataset. The model objects that are created in this
  step contain more than just the model. Notably, the following information is
  included to allow for prospective use and evaluation:

  - Feature metadata, as generated during the data processing step, is stored
    to allow for preparing datasets in the same manner as the development dataset
    and for checking if new datasets are formatted as expected. It is also used to
    create default ranges for individual conditional expectation and partial
    dependence plots.
    
  - Outcome metadata is stored. This is primarily used to check whether
    outcome data in new datasets are formatted in accordance with the development
    data. It is also used in computing several performance metrics.
    
  - A novelty detector is trained to detect out-of-distribution samples and
    assess when a model starts extrapolating. The novelty detector is currently
    based on extended isolation forests in the `isotree` package @Cortes2021-aa.
    
  - Models used to recalibrate the output of specific models (see
    *Learning algorithm* vignette) are stored.
    
  - Calibration information is added. This currently is only done for survival
    analysis, for which we store baseline survival curves @Royston2013-ch.
    
  - Risk stratification thresholds used for assigning risk strata are stored.

After training the models, the models are assessed using the development and any
validation datasets. Models, and results from this analysis are written to a
local directory.

## Supported outcomes

Familiar supports modelling and evaluation of several types of endpoints:

- Categorical endpoints, where the outcome consists of two or more classes.
  Familiar distinguishes between two-class (`binomial`) and multi-class
  (`multinomial`) outcomes. These differ in that fewer variable importance methods
  and learners are available for multi-class outcomes. Additionally some
  evaluation and explanation steps will assess all classes separately in a
  one-against-all fashion for multi-class outcomes, whereas for two-class outcomes
  only the *positive* class is assessed.

- Numerical endpoints, where the outcome consists of numeric values (`continuous`).

- Survival endpoints, where the outcome consists of a pair of time and event
  status variables. Familiar supports right-censored time-to-event data
  (`survival`).

Other endpoints are not supported. Handling of competing risk survival endpoints
is planned for future releases.

# Running familiar

The end-to-end pipeline is accessed through the `summon_familiar` function. This
is the main function to use.

In the example below, we use the *iris* dataset, specify some minimal
configuration parameters, and run the experiment. In practice, you may need to
specify some additional configuration parameters, see the [Configuring familiar]
section.


``` r
# Example experiment using the iris dataset. You may want to specify a different
# path for experiment_dir. This is where results are written to.
familiar::summon_familiar(
  data = iris,
  experiment_dir = file.path(tempdir(), "familiar_1"),
  outcome_type = "multinomial",
  outcome_column = "Species",
  experimental_design = "fs+mb",
  vimp_method = "mrmr",
  learner = "glm",
  parallel = FALSE
)
```

It is also possible to use a formula instead. This is generally feasible only
for datasets with few features:


``` r
# Example experiment using a formula interface. You may want to specify a
# different path for experiment_dir. This is where results are written to.
familiar::summon_familiar(
  Species ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
  data = iris,
  experiment_dir = file.path(tempdir(), "familiar_2"),
  outcome_type = "multinomial",
  experimental_design = "fs+mb",
  vimp_method = "mrmr",
  learner = "glm",
  parallel = FALSE
)
```

Data does not need to be loaded prior to calling `summon_familiar`. A path to a
`csv` file can also be provided. The data can also be a `data.frame` or
`data.table` contained in an `RDS` or `RData` file. Other data formats are
currently not supported.

If your dataset encodes categorical features using integers, it is recommended 
to load the data and manually encode them, as is explained in the
[Preparing your data] section.


``` r
# Example experiment using a csv datafile. Note that because the file does not
# exist, you will not be able to execute the code as is.
familiar::summon_familiar(
  data = "path_to_data/iris.csv",
  experiment_dir = file.path(tempdir(), "familiar_3"),
  outcome_type = "multinomial",
  outcome_column = "Species",
  class_levels = c("setosa", "versicolor", "virginica"),
  experimental_design = "fs+mb",
  vimp_method = "mrmr",
  learner = "glm",
  parallel = FALSE
)
```

For reproducibility purposes, it may be useful to configure `summon_familiar`
using the configuration `xml` file instead of using function arguments. In that
case, we will point to a data file using the `data_file` parameter in the `xml` 
file. For more information on configuring familiar, see the 
[Configuring familiar] section.


``` r
# Example experiment using a configuration file. Note that because the file does
# not exist, you will not be able to execute the code as is.
familiar::summon_familiar(config = "path_to_configuration_file/config.xml")
```

Configuration parameters may also consist of parameters specified in the
`xml` file and function arguments. Parameters set using function arguments 
supersede those specified using the `xml` file.


``` r
# Example experiment using an xml file, but with additional function arguments.
# Note that because the configuration file does not exist, you will not be able
# to execute the code as is.
familiar::summon_familiar(
  config = "path_to_configuration_file/config.xml",
  data = iris,
  parallel = FALSE
)
```

## Configuring familiar

As mentioned previously, familiar is highly configurable. Parameters can be
specified in two ways:

1. Using a configuration file. An empty copy of the configuration file can be
  obtained using the `familiar::get_xml_config` function. The
  `familiar::summon_familiar` function should subsequently be called by specifying
  the `config` argument, as shown in the earlier examples.

2. By specifying function arguments for the `familiar::summon_familiar` function.

All configuration parameters are documented in the help file of the
`familiar::summon_familiar` function. Often, the default settings suffice. The
parameters below should always be specified:

- `experimental_design`: Specifies the design of the experiment. This is
  described more extensively further in the vignette, in the [Experimental
  designs] section.

- `vimp_method`: Specify one or more variable importance methods. See the 
  *Variable importance methods* vignette for available methods.

- `learner`: Specify one or more learners used to create models. See the 
  *learning algorithms and hyperparameter optimisation* vignette for available
  learners.

Though not always required, specifying the following parameters is recommended
or situationally required:

- `experiment_dir`: This specifies the disk or network location where files generated
  during the experiment are written to. This includes files with the trained
  models, which we usually want to preserve. If this location is not specified,
  such files are temporarily written to the temporary R directory, and
  subsequently removed, or provided as output of the function.

- `outcome_column`: Specifies the name of the column of the data table that contains the outcome
  values. In case of survival outcomes two columns should be specified that
  indicate time and event status, respectively. For survival outcomes familiar
  determines which columns contain time and event data. The `outcome_column`
  parameter is not required in case the formula interface is used.

- `outcome_type`: Specifies the type of outcome being modeled. Should be one of
  the outcome types mentioned above in the [Supported outcomes] section. If not
  specified, it can potentially be inferred from the data contained in the
  column(s) specified by the `outcome_column` parameter.

- `class_levels`: Specify the class levels of two-class (`binomial`) and
  multi-class (`multinomial`) outcomes. For two-class outcomes, the second level
  specifies the class regarded as the *positive* class. The values should match
  values present in the outcome column. Specifying this argument is not necessary
  in case the outcome column is encoded as a factor. If left unspecified, the
  unique values in the outcome column are used as values. This can lead to the
  wrong class being used as *positive* class.

- `event_indicator`, `censoring_indicator`, `competing_risk_indicator`:
  Specifies the values that should be used as event, censoring, and competing risk
  indicators for survival analysis, respectively. Familiar uses default values for
  censoring (e.g. `0`, `FALSE`, `no`) and event (e.g. `1`, `TRUE`, `yes`) status
  otherwise. Note that the `competing_risk` outcome type will be fully implemented
  in a future release.

- `batch_id_column`, `sample_id_column`, `series_id_column`: Specifies the names
  of the columns containing batch, sample, and series identifiers respectively.
  These are described in more detail in the [Preparing your data] section.

## Preparing your data

Familiar processes tabular data. In this case, a table consists of rows that
represent instances, and columns that represent features and additional
information. This is a very common representation for tabular data. Let us look
at the *colon* dataset found in the survival package, which contains data from a
clinical trial to assess a new anti-cancer drug in patients with colon cancer:


``` r
# Get the colon dataset.
data <- data.table::as.data.table(survival::colon)[etype == 1]

# Drop some irrelevant columns.
data[, ":="("node4" = NULL, "etype" = NULL)]

knitr::kable(data[1:5])
```



| id| study|rx      | sex| age| obstruct| perfor| adhere| nodes| status| differ| extent| surg| time|
|--:|-----:|:-------|---:|---:|--------:|------:|------:|-----:|------:|------:|------:|----:|----:|
|  1|     1|Lev+5FU |   1|  43|        0|      0|      0|     5|      1|      2|      3|    0|  968|
|  2|     1|Lev+5FU |   1|  63|        0|      0|      0|     1|      0|      2|      3|    0| 3087|
|  3|     1|Obs     |   0|  71|        0|      0|      1|     7|      1|      2|      2|    0|  542|
|  4|     1|Lev+5FU |   0|  66|        1|      0|      0|     6|      1|      2|      3|    1|  245|
|  5|     1|Obs     |   1|  69|        0|      0|      0|    22|      1|      2|      3|    1|  523|



Here we see that each row contains a separate instance.

### Identifier columns

The `id` and `study` columns are identifier columns. Familiar distinguishes four
different types of identifiers:

- Batch identifiers are used to identify data belonging to a batch, cohort or
  specific dataset. This is typically used for specifying external validation
  datasets (using the `validation_batch_id` parameter). It also used to define the
  batches for batch normalisation. The name of the column containing batch
  identifiers (if any) can be specified using the `batch_id_column` parameter. If
  no column with batch identifiers is specified, all instances are assumed to
  belong to the same batch. In the *colon* dataset, the `study` column is a batch
  identifier column.

- Sample identifiers are used to identify data belonging to a single sample,
  such as a patient, subject, customer, etc. Sample identifiers are used to
  prevent instances from the same sample from being allocated to both development
  and validation data subsets created for cross-validation or bootstrapping. This
  prevents information leakage, as instances from the same sample are often
  related -- knowing one instance of a sample would make it easy to predict
  another, thus increasing the risk of overfitting. The name of the column
  containing sample identifiers can be specified using the `sample_id_column`
  parameter. If not specified, it is assumed that each instance forms a separate
  sample. In the *colon* dataset, the `id` column contains sample identifiers.

- Within a sample, it is possible to have multiple series, for example due to
  measurements at different locations in the same sample. A series differs from
  repeated measurements. While for series the outcome value may change, this is
  not allowed for repeated measurements. The column containing series identifiers
  may be specified by providing the column name as the `series_id_column`
  parameter. If not set, all instances of a sample with a different outcome value
  will be assigned a unique identifier.

- Within a sample, or series, it is possible to have repeated measurements,
  where one or more feature values may change but the outcome value does not. Such
  instances can for example used to assess feature robustness. Repeated
  measurement identifiers are automatically assigned for instances that have the
  same batch, sample and series identifiers.

### Outcome columns

The *colon* dataset also contains two outcome columns: `time` and `status` that
define (censoring) time and survival status respectively. Survival status are
encoded as `0` for alive, censored patients and `1` for patients that passed
away after treatment. Note that these correspond to default values present in
familiar. It is not necessary to pass these values as `censoring_indicator` and
`event_indicator` parameters.

### Feature columns

The remaining columns in the *colon* dataset represent features. There are two
numeric features, `age` and `nodes`, a categorical feature `rx` and several
categorical and ordinal features encoded with integer values. Familiar will
automatically detect and encode features that consist of `character`, `logical`
or `factor` type. However, it will not automatically convert features
encoded with integer values. This is by design -- familiar cannot determine
whether a feature with integer values is intended to be a categorical feature or
not. Should categorical features that are encoded with integers be present in
your dataset, you should manually encode such values in the data prior to
passing the data to familiar. For the *colon* dataset, this could be done as
follows:


``` r
# Categorical features
data$sex <- factor(x = data$sex, levels = c(0, 1), labels = c("female", "male"))
data$obstruct <- factor(data$obstruct, levels = c(0, 1), labels = c(FALSE, TRUE))
data$perfor <- factor(data$perfor, levels = c(0, 1), labels = c(FALSE, TRUE))
data$adhere <- factor(data$adhere, levels = c(0, 1), labels = c(FALSE, TRUE))
data$surg <- factor(data$surg, levels = c(0, 1), labels = c("short", "long"))

# Ordinal features
data$differ <- factor(
  data$differ,
  levels = c(1, 2, 3),
  labels = c("well", "moderate", "poor"),
  ordered = TRUE
)
data$extent <- factor(
  data$extent, 
  levels = c(1, 2, 3, 4), 
  labels = c("submucosa", "muscle", "serosa", "contiguous_structures"),
  ordered = TRUE
)

knitr::kable(data[1:5])
```



| id| study|rx      |sex    | age|obstruct |perfor |adhere | nodes| status|differ   |extent |surg  | time|
|--:|-----:|:-------|:------|---:|:--------|:------|:------|-----:|------:|:--------|:------|:-----|----:|
|  1|     1|Lev+5FU |male   |  43|FALSE    |FALSE  |FALSE  |     5|      1|moderate |serosa |short |  968|
|  2|     1|Lev+5FU |male   |  63|FALSE    |FALSE  |FALSE  |     1|      0|moderate |serosa |short | 3087|
|  3|     1|Obs     |female |  71|FALSE    |FALSE  |TRUE   |     7|      1|moderate |muscle |short |  542|
|  4|     1|Lev+5FU |female |  66|TRUE     |FALSE  |FALSE  |     6|      1|moderate |serosa |long  |  245|
|  5|     1|Obs     |male   |  69|FALSE    |FALSE  |FALSE  |    22|      1|moderate |serosa |long  |  523|



Manual encoding also has the advantage that ordinal features can be specified.
Familiar cannot determine whether features with `character` type values have an
associated order and will encode these as regular categorical variables. Another
advantage is that manual encoding allows for specifying the reference level,
i.e. the level to which other levels of a feature are compared in regression
models. Otherwise, the reference level is taken as the first level after sorting
the levels.

## Experimental designs

The experimental design defines how data analysis is performed. Familiar allows
for various designs, from very straightforward training on a single dataset, to
complex nested cross-validation with external validation. Experimental design is
defined using the `experimental_design` parameter and consists of basic workflow
components and subsampling methods. The basic workflow components are:

- `fs`: positions the variable importance computation step. If absent, variable 
  importances are determined just in time to optimise model hyperparameters.

- `mb`: positions the model building step. This component should always be present.

- `ev`: positions the external validation step. This should be used in 
  conjunction with the `validation_batch_id` parameter to specify
  which batches/cohorts should be used for external validation. This element can
  be omitted.

Each basic workflow component can only appear once in the experimental design.
It is possible to form an experiment using just the basic workflow components,
i.e. `fs+mb`, `mb`, `mb+ev` or `fs+mb+ev`. In these experiments, variable importance
computation is directly followed by modelling, with external validation of the
model on one or more validation cohorts for `mb+ev` and `fs+mb+ev`. These 
options correspond to TRIPOD type 1a and 3, respectively. TRIPOD analysis types
1b and 2 require more complicated experimental designs, which are facilitated by 
subsampling.

Subsampling methods are used to (randomly) sample the data that are not used for
external validation, and divide these data into internal development and
validation sets. Thus the dataset as a whole is at most divided into three
parts: internal development, internal validation and external validation.
Familiar implements the following subsampling methods:

- `bs(x,n)`: (stratified) .632 bootstrap, with `n` the number of bootstraps.
  Bootstrapping randomly samples the data with replacement, and on average assigns
  63.2% of the samples to the new subsampled subset to form the in-bag dataset
  with the same size as the original dataset. Remaining, unselected samples form
  the out-of-bag dataset. All pre-processing steps and hyperparameter optimisation
  (if any) are performed using the in-bag data.

- `bt(x,n)`: (stratified) .632 bootstrap, with `n` the number of bootstraps.
  Functions like `bs`, but pre-processing parameters and hyperparameters (if any)
  are inherited from the enveloping layer. That is, for `bt(fs+mb,20)+ev` twenty
  bootstraps are created from the development dataset, and variable importance
  computation and modelling are performed on the in-bag data. However,
  pre-processing parameters and hyperparameters are determined on the 
  **main development** dataset. The most practical application of `bt` is for computing
  variable importance multiple times (e.g. `bt(fs,50)+mb+ev`), allowing for
  aggregating variable importance and reducing the effect of random selection.

- `cv(x,n,p)`: (stratified) `n`-fold cross-validation, repeated `p` times. `p`
  equals 1 by default. Cross-validation randomly assigns samples to `n` folds.
  Cross-validation forms `n` experiments where one fold is assigned as a
  validation fold, and the remainder as training folds. All pre-processing steps
  and hyperparameter optimisation (if any) are performed using data in the
  training folds.

- `lv(x)`: leave-one-out-cross-validation. This is the same as `n`-fold
  cross-validations with `n` the number of samples.

- `ip(x)`: imbalance partitioning for addressing class imbalances in the
  dataset. This creates subsets of the data with balanced classes and can be used
  in conjunction with `binomial` and `multinomial` outcomes. All pre-processing
  steps and hyperparameter optimisation are determined within the partitions. The
  number of partitions generated depends on the imbalance correction method
  (specified using the `imbalance_correction_method` parameter). Imbalance
  partitioning does not generate validation sets.

The `x` argument of subsample methods can contain one or more of the workflow
components. Moreover, it is possible to nest subsample methods. For example,
`experimental_design="cv(bt(fs,50)+mb,5)+ev"` would create a 5-fold
cross-validation of the development dataset, with each set of training folds
again subsampled for computing variable importances. After aggregating variable 
importance obtained over 50 bootstraps, a model is trained within each set of 
training folds, resulting in 5 models overall. The ensemble of these models is 
then evaluated on an external dataset.

Other designs, such as `experimental_design="bs(fs+mb,400)+ev"` allow for building
large ensembles, and capturing the posterior distribution of the model
predictions.

Subsampling involves randomisation. By default, every new experiment will
generate its own random assignment. However, samples will be assigned in the
same way across experiments by setting the `iteration_seed` parameter, provided
the underlying data doesn't change between experiments.

As a final remark: Though it is possible to encapsulate the external validation
(`ev`) workflow component in a subsampler, this is completely unnecessary.
Unlike the variable importance computation (`fs`) and modelling (`mb`)
components, `ev` is passive, and only indicates whether external validation
should be performed.

## Warm start

Calling `summon_familiar` as described above, provides a cold start of the
process. For some purposes, such as to ensure that the same data splits are
used, a (partial) warm start may be required across different experiments. Three
functions allow for generating data for a warm start:

- `precompute_data_assignment`: Generates data assignment.

- `precompute_feature_info`: Generates data assignment and corresponding
  feature information.

- `precompute_vimp`: Generates data assignment, corresponding feature
  information and variable importance.

All of functions above create an `experimentData` object, which contains
data that can be used to warm-start other familiar experiments that use the same
data. This object can then be supplied as the `experiment_data` argument for
`summon_familiar`.


``` r
# This creates both data assignment (5 bootstraps) and the corresponding feature
# information.
experiment_data <- familiar::precompute_feature_info(
  data = iris,
  experiment_dir = file.path(tempdir(), "familiar_1"),
  outcome_type = "multinomial",
  outcome_column = "Species",
  experimental_design = "bs(fs+mb,5)",
  parallel = FALSE
)

# Now we can warm-start a new experiment using the precomputed data.
familiar::summon_familiar(
  data = iris,
  experiment_data = experiment_data,
  experiment_dir = file.path(tempdir(), "familiar_2"),
  outcome_type = "multinomial",
  outcome_column = "Species",
  vimp_method = "mrmr",
  learner = "glm",
  parallel = FALSE
)
```

## Training models only

`summon_familiar` also incorporates a model evaluation process, which may not be
that useful if the aim is to simply train a model. In that case,
`train_familiar` can be used instead. `train_familiar` and `summon_familiar`
share the same configuration parameters. However, `train_familiar` will simply
return the model(s) after training, and forgo their evaluation.

# References

---
title: "Evaluation and explanation"
author: "Alex Zwanenburg"
date: "2026-04-22"
output:
  rmarkdown::html_vignette:
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
  rmarkdown::github_document:
    html_preview: FALSE
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
bibliography: "refs.bib"
vignette: >
  %\VignetteIndexEntry{Evaluation and explanation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---




``` r
library(familiar)
library(data.table)

set.seed(19)
```

Familiar will automatically evaluate created models, and export corresponding
tables and plots. The table below lists all available analyses. Several
analyses allow for determining distributions and confidence intervals to provide
an indication of the value spread for a given dataset and models (if
applicable). The `estimation_type` argument defines this behaviour:

-   `point`: Compute point estimates, i.e. single values.

-   `bias_correction` or `bc`: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates.

-   `bootstrap_confidence_interval` or `bci`: Bias-corrected estimates with
bootstrap confidence intervals [@Efron2016-ws]. This estimation type allows for
plotting confidence intervals in plots. The number of point estimates required
depends on the `confidence_level` parameter. Familiar uses a rule of thumb of
n=20/(1-`confidence_level`), i.e. at least 400 points for
`confidence_level=0.95`.

Another argument, `detail_level` determines how the required point estimates are
formed:

-   `ensemble`: Point estimates are computed at the ensemble level, i.e. over
all models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20 bootstraps
and computing the model performance of the ensemble model for each bootstrap.

-   `hybrid`: Point estimates are computed from the models in an ensemble.
This means that, for example, bias-corrected estimates of model performance are
directly computed using the models in the ensemble. If there are at least 20
trained models in the ensemble, performance is computed for each model, in
contrast to `ensemble` where performance is computed for the ensemble of models.
If there are less than 20 trained models in the ensemble, bootstraps are created
so that at least 20 point estimates can be made.

-   `model`: Point estimates are computed at the model level. This means that,
for example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model for
each bootstrap. Familiar will not automatically create plots in this case.

The more point estimates are required, the more computationally
intensive the analysis is. Some analyses also become computationally more
expensive with more samples. The `sample_limit` argument can be used to specify
the maximum number of samples that should be used.

Some analyses, such as model predictions and individual conditional expectation
plots, do not benefit from bootstraps as they are solely based on values
predicted by the (ensemble) models. If you want to compute bias-corrected
estimates or bootstrap confidence intervals for these analyses you should ensure
that sufficient models are created. Experimental designs such as
`experimental_design="bs(fs+mb,400)+ev"` allow for this, but are computationally
expensive.

Several methods do not require any more information than that provided in a 
prediction table or dataset.

| Name                                     | Plot function                                                                         | Export function                       | Est. type | Detail level | Sample limit | Model | Pred. table | Dataset |
|------------------------------------------|---------------------------------------------------------------------------------------|---------------------------------------|:---------:|:------------:|:------------:|:-----:|:-----------:|:-------:|
| AUC-PR^a^                                | `plot_auc_precision_recall_curve`                                                     | `export_auc_data`                     | ×         |      ×       |              | ×     |      ×      |         |
| AUC-ROC^a^                               | `plot_auc_roc_curve`                                                                  | `export_auc_data`                     | ×         |      ×       |              | ×     |      ×      |         |
| Calibration info^c^                      |                                                                                       | `export_calibration_info`             |           |      ×       |              | ×     |             |         |
| Confusion matrix^a^                      | `plot_confusion_matrix`                                                               | `export_confusion_matrix_data`        |           |      ×       |              | ×     |      ×      |         |
| Decision curve analysis^ab^              | `plot_decision_curve`                                                                 | `export_decision_curve_analysis_data` | ×         |      ×       |              | ×     |      ×      |         |
| Feature expression^c^                    | `plot_sample_clustering`                                                              | `export_feature_expressions`          |           |              |              | ×     |             | ×       |
| Feature selection variable importance^c^ | `plot_feature_selection_occurrence`; `plot_feature_selection_variable_importance`     | `export_fs_vimp`                      |           |              |              | ×     |             |         |
| Feature similarity^c^                    | `plot_feature_similarity`                                                             | `export_feature_similarity`           | ×         |              |              | ×     |             | ×       |
| Hyperparameters^c^                       |                                                                                       | `export_hyperparameters`              |           |              |              | ×     |             |         |
| Individual conditional expectation^c^    | `plot_ice`                                                                            | `export_ice_data`                     | ×^d^      |      ×       |      ×       | ×     |             |         |
| Model calibration^ce^                    | `plot_calibration_data`                                                               | `export_calibration_data`             | ×         |      ×       |              | ×     |             |         |
| Model performance^c^                     | `plot_model_performance`                                                              | `export_model_performance`            | ×         |      ×       |              | ×     |      ×      |         |
| Model predictions^c^                     |                                                                                       | `export_prediction_data`              | ×^d^      |      ×       |              | ×     |      x      |         |
| Model variable importance^ce^            | `plot_model_signature_occurrence`; `plot_model_signature_variable_importance`         | `export_model_vimp`                   |           |              |              | ×     |             |         |
| Partial dependence^c^                    | `plot_pd`                                                                             | `export_partial_dependence_data`      | ×^d^      |      ×       |      ×       | ×     |             |         |
| Permutation variable importance^c^       | `plot_permutation_variable_importance`                                                | `export_permutation_vimp`             | ×         |      ×       |              | ×     |             |         |
| Risk stratification info^b^              |                                                                                       | `export_risk_stratification_info`     |           |      ×       |              | ×     |             |         |
| Risk stratification data^b^              | `plot_kaplan_meier`                                                                   | `export_risk_stratification_data`     |           |              |              | ×     |      ×      |         |
| Sample similarity^c^                     |                                                                                       | `export_sample_similarity`            | ×         |              |      ×       | ×     |             | ×       |
| SHAP values                              | `plot_shap_summary`, `plot_shap_waterfall`, `plot_shap_force`, `plot_shap_dependence` | `export_shap`                         |           |      ×       |      ×       | ×     |             |         |
| Univariate analysis^c^                   | `plot_univariate_importance`                                                          | `export_univariate_analysis_data`     |           |              |              | ×     |             |         |

: Evaluation and explanation steps.\
^a^ Available for binomial and multinomial outcomes.\
^b^ Available for survival outcomes.\
^c^ Available for all outcomes.\
^d^ Estimation types other than `point` require sufficient models.\
^e^ May not be available for all models.

When familiar is run from `summon_familiar` plots and tables are exported
automatically. The corresponding plot and export functions can all be used externally as well.
This vignette shows their use.

We will first create two models. The first model aims to predict the risk for
ischemic stroke based on clinical and image features. This is a classification problem.
For illustration purposes, we will use the first 99 samples as development data,
and the remaining as a holdout set.


``` r
ischemic_stroke_data <- data.table::as.data.table(modeldata::ischemic_stroke)

# Set categorical variables.
ischemic_stroke_data$male <- factor(
  ischemic_stroke_data$male, levels = c(0, 1), labels = c("no", "yes")
)
ischemic_stroke_data$smoking_history <- factor(
  ischemic_stroke_data$smoking_history, levels = c(0, 1), labels = c("no", "yes")
)
ischemic_stroke_data$atrial_fibrillation <- factor(
  ischemic_stroke_data$atrial_fibrillation, levels = c(0, 1), labels = c("no", "yes")
)
ischemic_stroke_data$coronary_artery_disease <- factor(
  ischemic_stroke_data$coronary_artery_disease, levels = c(0, 1), labels = c("no", "yes")
)
ischemic_stroke_data$diabetes_history  <- factor(
  ischemic_stroke_data$diabetes_history, levels = c(0, 1), labels = c("no", "yes")
)
ischemic_stroke_data$hypercholesterolemia_history   <- factor(
  ischemic_stroke_data$hypercholesterolemia_history, levels = c(0, 1), labels = c("no", "yes")
)
ischemic_stroke_data$hypertension_history  <- factor(
  ischemic_stroke_data$hypertension_history, levels = c(0, 1), labels = c("no", "yes")
)

# Create training and hold-out test set.
ischemic_stroke_data_train <- ischemic_stroke_data[1L:99L, ]
ischemic_stroke_data <- ischemic_stroke_data[100L:126L, ]

# Call familiar to create models. We don't use summon_familiar here, but call
# train_familiar instead, which returns familiar models.
ischemic_stroke_model <- familiar::train_familiar(
  data = ischemic_stroke_data_train,
  outcome_column = "stroke",
  class_levels = c("no", "yes"),
  vimp_method = "none",
  learner = "random_forest_ranger_default",
  parallel = FALSE,
  verbose = FALSE
)

```

The second model is a survival model, and is trained on a colon
chemotherapy clinical trial dataset [@Moertel1995-bh] to predict recurrence 
after treatment.


``` r
survival_data <- data.table::as.data.table(survival::colon)

# Select data corresponding to tumour recurrence.
survival_data <- survival_data[etype == 1L]

# Remove unncessary columns.
survival_data[, ":=" ("id" = NULL, "study" = NULL, "etype" = NULL, "node4" = NULL)]

# Set categorical variables.
survival_data$sex <- factor(
  survival_data$sex, levels = c(0, 1), labels = c("female", "male")
)
survival_data$obstruct <- factor(
  survival_data$obstruct, levels = c(0, 1), labels = c("no", "yes")
)
survival_data$adhere <- factor(
  survival_data$adhere, levels = c(0, 1), labels = c("no", "yes")
)
survival_data$differ <- factor(
  survival_data$differ,
  levels = c(1, 2, 3),
  labels = c("well", "moderate", "poor"),
  ordered = TRUE
)
survival_data$extent <- factor(
  survival_data$extent,
  levels = c(1, 2, 3, 4),
  labels = c("submucosa", "muscle", "serosa", "contiguous structures"),
  ordered = TRUE
)
survival_data$surg <- factor(
  survival_data$surg, levels = c(0, 1), labels = c("short", "long")
)

# Create training and hold-out test set.
survival_data_train <- survival_data[1L:799L, ]
survival_data <- survival_data[800L:929L, ]

survival_model <- familiar::train_familiar(
  data = survival_data_train,
  outcome_column = c("status", "time"),
  vimp_method = "none",
  learner = "random_forest_ranger_default",
  parallel = FALSE,
  verbose = FALSE
)
```

# Model performance

We assess model performance to evaluate accuracy of a model. Model accuracy is
quantified using performance metrics that are documented in the *Performance metrics*
vignette, and can be set using the `evaluation_metric` argument of the
`summon_familiar` function or `metric` for plot and export functions. For
example, this results in the following plot for the ischemic stroke dataset:


``` r
plots <- familiar::plot_model_performance(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  metric = c("auc", "brier", "f1_score"),
  x_axis_by = "metric"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/model-performance-plot-1.png" alt="Performance of the ischemic stroke model on hold-out test data.
 The model is a classifier, and was assessed using three metrics:
 area under the receiver-operating characteristic curve (AUC), the Brier 
 score, and the F1-score. Higher AUC and F1-scores, and lower Brier score, 
 indicate better models. The graph shows the distribution of performance 
 metrics obtained from bootstraps of the dataset, along with 2.5th, 50.0th
 and 97.5th percentiles."  />
<p class="caption">Performance of the ischemic stroke model on hold-out test data.
 The model is a classifier, and was assessed using three metrics:
 area under the receiver-operating characteristic curve (AUC), the Brier 
 score, and the F1-score. Higher AUC and F1-scores, and lower Brier score, 
 indicate better models. The graph shows the distribution of performance 
 metrics obtained from bootstraps of the dataset, along with 2.5th, 50.0th
 and 97.5th percentiles.</p>
</div>

The plot shows distributions of the area under the receiver operating
characteristic curve, the brier score and the f1-score, based on bootstrap
resampling of model predictions.

Model performance can also be determined directly from predictions, which makes
it possible to assess predictions made by non-familiar models. This involves
conversion of the predictions to a prediction table object using `as_prediction_table`,
as shown below:


``` r
predictions <- predict(
  object = ischemic_stroke_model,
  newdata = ischemic_stroke_data
)

# Create a prediction table object using as_prediction_table. The minimum
# you need to provide here are: prediction probabilities for the positive class
# (yes), the reference labels, the type of prediction (classification), 
# and providing the class levels (no, yes).
predictions <- familiar::as_prediction_table(
  x = predictions$yes,
  y = ischemic_stroke_data$stroke,
  type = "classification",
  class_levels = c("no", "yes")
)
#> Error in `familiar::as_prediction_table()`:
#> ! (converted from warning) The positive class cannot be directly inferred from names of the provided prediction data, and is assumed to be yes.

# The prediction table object is then provided as object
plots <- familiar::plot_model_performance(
  object = predictions,
  metric = c("auc", "brier", "f1_score"),
  x_axis_by = "metric"
)
#> Error in `.local()`:
#> ! data_element expects one of feature_expressions, risk_stratification_data, feature_similarity and sample_similarity as input. Found the following unknown values: model_performance
```

<div class="figure" style="text-align: center">
<img src="eval_and_explain/model-performance-plot-standalone-1.png" alt="Performance for the ischemic strok model obtained using a 
  prediction table. This results in the same figure as above, obtained
  using the model itself."  />
<p class="caption">Performance for the ischemic strok model obtained using a 
  prediction table. This results in the same figure as above, obtained
  using the model itself.</p>
</div>

This produces the exact same plot. Model performance can also be exported to a 
table using using `export_model_performance`.


``` r
results <- familiar::export_model_performance(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  metric = c("auc", "brier", "f1_score"),
  aggregate_results = TRUE
)
results[[1L]]@data
#>                       data_set vimp_method                      learner ensemble_model_name   metric     value    ci_low     ci_up
#>                         <fctr>      <fctr>                       <fctr>              <fctr>   <fctr>     <num>     <num>     <num>
#> 1: 173529_2wFvaYlrMM6XXSNOt9rE        none random_forest_ranger_default      ensemble.NA.NA      auc 0.7387626 0.5166193 0.9146674
#> 2: 173529_2wFvaYlrMM6XXSNOt9rE        none random_forest_ranger_default      ensemble.NA.NA    brier 0.2041370 0.1639073 0.2492457
#> 3: 173529_2wFvaYlrMM6XXSNOt9rE        none random_forest_ranger_default      ensemble.NA.NA f1_score 0.7407407 0.5152258 0.8967186
```
In the above table, `data_set` is randomly generated because we call this method
on a (technically) *new* dataset. `ensemble_model_name` contains a placeholder
for the same reason. `vimp_method` and `learner` are derived from the model, and
metrics with 95% confidence intervals are shown.

## Receiver-operating characteristic curve

The receiver-operating characteristic (ROC) curve visualises concordance between
predicted class probabilities and the observed classes [@Hand2001-il]. It shows
the trade-off between sensitivity and specificity of the model. The
receiver-operating characteristic curve for the ischemic stroke dataset is as
follows:


``` r
plots <- familiar::plot_auc_roc_curve(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/auc-roc-curve-bci-1.png" alt="Receiver-operating characteristic curve for the ischemic stroke 
 model on hold-out test data. The 95% confidence interval is shown (grey)."  />
<p class="caption">Receiver-operating characteristic curve for the ischemic stroke 
 model on hold-out test data. The 95% confidence interval is shown (grey).</p>
</div>

Under ideal circumstances, the curve would lie in the top-left corner of the
plot. If a model does not predict better than at random, the curve would lie
along the diagonal, which corresponds to an AUC-ROC value of 0.50. The plot
above indicates that the model can somewhat differentiate between patients with
and without stroke.

Familiar creates an ROC curve by ascendingly sorting instances by the predicted
probability of the positive class, and computing the number of true positive,
true negative, false positive, false negative cases observed at each instance.
These are then used to compute sensitivity and specificity.

The above plot shows the average of 400 single curves, with 95% confidence
intervals. To achieve this, familiar interrogates each curve over the complete
range of potential specificity values ([0.000, 1.000]) with a step size of
0.005. Linear interpolation is used for this purpose [@Davis2006-xb]. This
allows for assessing the distribution of sensitivity at fixed specificity
values.

The above approach is the general approach used by familiar. However, familiar
will plot an exact ROC curve if the point estimate (`estimation_type="point"`)
of a single model or the complete ensemble of models is evaluated
(`detail_level="ensemble"`):


``` r
plots <- familiar::plot_auc_roc_curve(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  estimation_type = "point"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/auc-roc-curve-point-1.png" alt="Receiver-operating characteristic curve for the ischemic stroke 
 model on hold-out test data. Whereas the previous figure shows the bootstrap
 estimate and confidence interval, the current curve is formed
 directly from the data."  />
<p class="caption">Receiver-operating characteristic curve for the ischemic stroke 
 model on hold-out test data. Whereas the previous figure shows the bootstrap
 estimate and confidence interval, the current curve is formed
 directly from the data.</p>
</div>

Note that we did not specify the detail level to create the above plot because
only a single model is evaluated.

## Precision-recall curve

The precision-recall (PR) curve shows the trade-off between precision (positive
predictive value) and recall (sensitivity). An important distinction between the
PR curve and the ROC curve described above is that precision does not assess
true negatives but false positives. In case the number of negative instances
greatly exceeds the number of positive instances, a large change in the number
of false positives reduces the ROC curve somewhat, but this change is far easier
observed in the PR curve [@Davis2006-xb]. Hence the PR curve is commonly
assessed if the positive class is more important than and rarer compared to the
negative class. For the ischemic stroke dataset, the precision-recall curve is as
follows:


``` r
plots <- familiar::plot_auc_precision_recall_curve(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/auc-pr-curve-bci-1.png" alt="Precision-recall curve for the ischemic stroke model on hold-out 
test data. The 95% confidence interval is shown (grey)."  />
<p class="caption">Precision-recall curve for the ischemic stroke model on hold-out 
test data. The 95% confidence interval is shown (grey).</p>
</div>

Under ideal circumstances, the curve would lie in the top-right part of the
plot. A random classifier would yield a curve that is mostly horizontal and
located at the fraction of positive instances. This would equal a precision of
56% for the current dataset.

Like the ROC curve, familiar forms the PR curve by ascendingly sorting instances
by the predicted probability of the positive class and computing the number of
true positive and false positive negative cases observed at each instance. These
values are then used to compute recall and precision.

The above plot shows the average of 400 single curves, with 95% confidence
intervals. To achieve this, familiar interrogates each curve over the complete
range of recall values ([0.000, 1.000]) with a step size of 0.005. We use
linear interpolation for this purpose, treating the PR curve as a piecewise
function to manage the situation identified by Davis and Goadrich
[@Davis2006-xb]. Interpolating in this manner allows for assessing the
distribution of precision values at fixed recall values.

Again, this is the general approach used by familiar. Familiar can plot an exact
PR curve when evaluating the point estimate (`estimation_type="point"`) of a
single model or of the complete ensemble of models (`detail_level="ensemble"`).
In the ischemic strok dataset, this produces the following curve:


``` r
plots <- familiar::plot_auc_precision_recall_curve(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  estimation_type = "point"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/auc-pr-curve-point-1.png" alt="Precision-recall curve for the ischemic stroke 
 model on hold-out test data. Whereas the previous figure shows the bootstrap
 estimate and confidence interval, the current curve is formed
 directly from the data."  />
<p class="caption">Precision-recall curve for the ischemic stroke 
 model on hold-out test data. Whereas the previous figure shows the bootstrap
 estimate and confidence interval, the current curve is formed
 directly from the data.</p>
</div>

## Confusion matrix

A confusion matrix schematically represents the number of predicted (expected)
class instances and their actual (observed) values. This may help identify
potential weaknesses of classifiers, i.e. false positives or negatives.


``` r
plots <- familiar::plot_confusion_matrix(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/confusion-matrix-1.png" alt="Confusion matrix for the ischemic stroke 
 model on hold-out test data. The number of instances with the prediction
 of ischemic stroke (expected sroke), and actually observed stroke, is shown."  />
<p class="caption">Confusion matrix for the ischemic stroke 
 model on hold-out test data. The number of instances with the prediction
 of ischemic stroke (expected sroke), and actually observed stroke, is shown.</p>
</div>

Familiar selects the class with the highest predicted class probability as the
expected class. It is currently not possible in familiar to obtain confusion
matrices for set probability thresholds, such as the threshold that maximises
the Youden's Index.

## Kaplan-Meier survival curves

Kaplan-Meier survival curves are a standard method for assessing survival over time in
a population, or for assessing survival differences between groups.

Familiar applies one or more thresholds to stratify instances to risk groups.
These thresholds are created during training to avoid bias. The number
and value of these thresholds is determined by two parameters:
`stratification_method` and `stratification_threshold`. Both parameters are set
when calling `summon_familiar` or `train_familiar`.
`stratification_method` has three options:

-   `median`: Familiar uses the median predicted value in the development cohort
to stratify instances into two risk groups.

-   `fixed`: Instances are stratified based on the sample quantiles of the
predicted values. These quantiles are defined using the
`stratification_threshold` parameter.

-   `optimised`: Use maximally selected rank statistics to determine the optimal
threshold [@Lausen1992-qh; @Hothorn2003-qn] to stratify instances into two
optimally separated risk groups.

The `stratification_threshold` parameter is only used when
`stratification_method = "fixed"`. It allows for specifying the sample quantiles
that are used to determine the thresholds. For example
`stratification_threshold = c(0.25,0.75)` creates two thresholds that stratifies
the development dataset into three groups: one with 25% of the instances with
the highest risk, a group with 50% of the instances with medium risk, and a
third group with 25% of instances with lowest risk.

Applying the survival model to the colon cancer dataset yields the following
Kaplan-Meier plot:


``` r
plots <- familiar::plot_kaplan_meier(
  object = survival_model,
  data = survival_data
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/kaplan-meier-plot-1.png" alt="Kaplan-Meier plot for the colon cancer survival model on hold-out
 test data. The hold-out test data were split into two risk groups by
 applying a threshold that was defined during training. 95% confidence
 intervals are shown. The number of patients surviving without tumour recurrence
 is shown below the main plot."  />
<p class="caption">Kaplan-Meier plot for the colon cancer survival model on hold-out
 test data. The hold-out test data were split into two risk groups by
 applying a threshold that was defined during training. 95% confidence
 intervals are shown. The number of patients surviving without tumour recurrence
 is shown below the main plot.</p>
</div>

In the plot above, familiar divides instances into two groups by the median
threshold in the development dataset. The risk groups are shown with 95%
confidence intervals. Familiar uses the `survival::survfit` function to create
both the survival curves and the confidence intervals. The low-risk group has
significantly better survival than the high-risk group. This is also indicated
in the bottom-left of the main plot, which shows the result of a logrank test
between the two groups. Crosses indicate right-censored patients, of which there
were few. The number of surviving patients in both groups at the time points
along the x-axis are shown below the main plot.

# Model calibration

Evaluating model performance is important as it allows us to assess how 
accurate a model is. However, it often does not tell us how well we can rely on
predictions for individual instances, e.g. how accurate an instance class
probability is. Model calibration is assessed for this purpose. The model for
the ischemic stroke dataset produces the following calibration plot:


``` r
plots <- familiar::plot_calibration_data(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/calibration-data-bci-1.png" alt="Calibration plot for the ischemic stroke model on hold-out
 test data. The plot shows the predicted (expected) probability of ischemic 
 stroke, against the observed occurrence within smaller groupings of data. 
 The plot shows the ideal calibration (dashed line), the linear calibration
 fit (thin line), and estimated calibration (thick line). 95% confidence
 intervals, based on bootrstraps, are shown around the estimated calibration. 
 The line above the plot shows the density of predicted probabilities. 
 Calibration outside the main distribution is more uncertain. The inset 
 describes the intercept and slope of the linear fit, and the results
 of the Hosmer-Lemeshow goodness-of-fit test."  />
<p class="caption">Calibration plot for the ischemic stroke model on hold-out
 test data. The plot shows the predicted (expected) probability of ischemic 
 stroke, against the observed occurrence within smaller groupings of data. 
 The plot shows the ideal calibration (dashed line), the linear calibration
 fit (thin line), and estimated calibration (thick line). 95% confidence
 intervals, based on bootrstraps, are shown around the estimated calibration. 
 The line above the plot shows the density of predicted probabilities. 
 Calibration outside the main distribution is more uncertain. The inset 
 describes the intercept and slope of the linear fit, and the results
 of the Hosmer-Lemeshow goodness-of-fit test.</p>
</div>

The calibration curve shows the expected (predicted) probability of ischemic
stroke versus the observed proportion of ischemic stroke, together with a 95%
confidence interval. In an ideal calibration plot the calibration curve should
lie on the diagonal (dashed) line. Low expected probabilities tend to
overestimate the observed probability, whereas high expected probabilities tend
to underestimate the observed probability. The density plot in the top panel
indicates that predicted probabilities were distributed around 0.5, with
relatively few low- and high-probability cases being predicted.

Familiar also performs several calibration tests to assess various degrees of
calibration [@Van_Calster2016-ok]. First of all, we assess two aspects of the
linear fit (solid line in the plot): the intercept (calibration-in-the-large)
and slope. These are shown in the top-left of the main plot.

The intercept ideally is 0.00. In our example, it lies below this point, but is
still contained in the 95% confidence interval, owing to a small number of
samples in the test set.

In an ideal case, the slope of the linear fit is 1.00. In our example, the slope
of the linear fit lies above this value.

One issue with assessing calibration through a linear fit, is that the
calibration curve can be decidedly non-linear and still produce good values for
the intercept and slope. Therefore, we also assess calibration using statistical
goodness-of-fit tests: the Hosmer-Lemeshow (HL) test for categorical and
regression outcomes [@Hosmer1997-ov], and the Nam-D'Agostino (ND)
[@DAgostino2003-ot] and Greenwood-Nam-D'Agostino (GND) [@Demler2015-yx] tests
for survival outcomes.

## Implementation details

Here we detail the implementation of the model calibration in familiar. The
statistical tests mentioned above, as well as the linear fit, rely on grouping
instances. First we order the instances in the dataset by their expected
(predicted) values. Within each group, the expected values and observed values
are compared:

* Categorical outcomes: We compare the fraction instances with the (positive)
observed class in the group against the mean expected probability for that same
class.

* Regression outcomes: Both expected and observed values are first normalised to
a $[0, 1]$ range using the value range observed in the development dataset. We
then compare the mean observed value in the group against the mean expected
value.

* Survival outcomes: We predict survival probabilities at time points indicated
by `evaluation_times`. The mean expected survival probability in the group is
then compared against the observed probability (Kaplan-Meier estimator) at the
same time points.

We can observe the points thus created in a calibration plot when
`estimation_type="point"`:


``` r
plots <- familiar::plot_calibration_data(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  estimation_type = "point"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/calibration-data-point-1.png" alt="Calibration plot for the ischemic stroke model on hold-out
 test data. Whereas in the earlier figure the calibration curve was the
 bootstrap estimate, here the expected probability and observed proportions 
 in several (random) groupings is shown."  />
<p class="caption">Calibration plot for the ischemic stroke model on hold-out
 test data. Whereas in the earlier figure the calibration curve was the
 bootstrap estimate, here the expected probability and observed proportions 
 in several (random) groupings is shown.</p>
</div>

The plot with point estimates and the earlier calibration plot with bootstrap
confidence intervals lead to the same assessment. Moreover, you may not arrive
at the same plot. This happens because familiar introduces randomness into the
assessment of calibration. First, familiar randomises the number of groups drawn
because using a fixed number of 10 groups is arbitrary. Second, the actual group
sizes are randomised by drawing group identifiers with replacement for all
instances. We then sort the set of drawn identifiers, and the instances, ordered
by increasing expected value, are then assigned to their respective group.

Linear fit intercept and slope as well as the goodness-of-fit test p-value are
determined for each single set of groups. Intercept and slopes are then averaged
(`point` and `bias_correction`) or assessed as a distribution
(`bootstrap_confidence_interval`). p-values from goodness-of-fit tests need to
be combined. These tests are not fully independent because they derive from the
same dataset. We therefore compute a harmonic mean p-value [@Wilson2019-bi] from
the individual p-values.

# Decision curve analysis

Decision curve analysis is a method to assess the net (clinical) benefit of
models, introduced by Vickers and Elkin [@Vickers2006-be]. The decision curve
compares the model benefit for different threshold probabilities against two or
more options. The two standard options are that all instances receive an
intervention, and none receive an intervention. What constitutes an intervention
is model-dependent [@Vickers2019-ar]. For our ischemic stroke cohort, intervention
might constitute treatment of all patients, versus no treatment and treatment for
patients with malignancy predicted by the model. This produces the following
figure:


``` r
plots <- familiar::plot_decision_curve(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  y_range = c(-0.2, 0.8)
)
```

```
#> Error in `ggplot2::geom_line()`:
#> ! Problem while converting geom to grob.
#> ℹ Error occurred in the 1st layer.
#> Caused by error:
#> ! (converted from warning) Removed 33 rows containing missing values or values outside the scale range (`geom_line()`).
```

The shaded curve is the decision curve produced by the model, with 95%
confidence intervals. The declining curve represents the intervention for all
strategy, which offers no benefit for a probability of ischemic stroke of 56%.
The horizontal line at a net benefit of 0.0 represent the no intervention
strategy. As may be observed, the model offers an expected net benefit compared
to either strategy, but not always significantly.

For regression problems no definition of decision curves exists. However,
familiar can perform decision curve analysis for survival outcomes, based on
Vickers et al. [@Vickers2008-uu].

# Variable importance

Not all features are strongly associated with the outcome of interest. One of
the most critical points in machine learning is selecting those features that
are important enough to incorporate into a model. Familiar assesses variable
importance using the methods detailed in the *Variable importance methods*
vignette during training. Later, during evaluation, familiar employs both
model-specific and model-agnostic methods to explaining which features are
considered important by the model itself.

## Model-specific methods

Model-specific methods rely on aspects of the models themselves to determine
variable importance. For instance, the coefficients of regression models are a
measure of variable importance when features are normalised. In the case of the
random forest model for the ischemic stroke dataset, permutation variable
importance is determined. This produces the following plot:


``` r
plots <- familiar::plot_model_signature_variable_importance(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  rotate_x_tick_labels = TRUE
)
```

```
#> Error in `ggplot2::geom_bar()`:
#> ! Problem while converting geom to grob.
#> ℹ Error occurred in the 1st layer.
#> Caused by error:
#> ! (converted from warning) Removed 13 rows containing missing values or values outside the scale range (`geom_bar()`).
```
In such plots, we always order features so that the most important features are
on the left, and the least important features on the right. Letters indicate
features that are clustered together.

In case an ensemble of models is assessed, scores are aggregated using an
aggregation method (see the *variable importance methods* vignette). This method
can be set using the `aggregation_method` argument, or the
`eval_aggregation_method` configuration parameter (`summon_familiar`).

The `plot_model_signature_occurrence` method plots the occurrence of features
among the first 5 ranks across an ensemble of models (if available). The rank
threshold can be specified using the `rank_threshold` argument or the
`eval_aggregation_rank_threshold` parameter (`summon_familiar`).

## Permutation variable importance

Permutation variable importance is a model-agnostic variable importance method.
It assesses importance by measuring the decrease in model performance caused by
shuffling values of a feature across the instances in a dataset.

For the ischemic stroke dataset, this results in the following figure:


``` r
plots <- familiar::plot_permutation_variable_importance(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  estimation_type = "bias_correction"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/permutation-variable-importance-bc-1.png" alt="Permutation variable importance plot for the ischemic stroke model on hold-out
 test data. Features are sorted from highest (top) to least importance (bottom).
 Permutation variable importance measures the model performance difference 
 between normal and shuffled (informationless) input values for each feature.
 At the threshold of 1.00 every feature is shuffled individually, whereas at 0.23
 groups of similar features are shuffled simultaneously."  />
<p class="caption">Permutation variable importance plot for the ischemic stroke model on hold-out
 test data. Features are sorted from highest (top) to least importance (bottom).
 Permutation variable importance measures the model performance difference 
 between normal and shuffled (informationless) input values for each feature.
 At the threshold of 1.00 every feature is shuffled individually, whereas at 0.23
 groups of similar features are shuffled simultaneously.</p>
</div>

The figure shows the features along the y-axis, and the loss in AUC-ROC caused
by permuting the feature values along the x-axis. Accordingly,
`max_remodeling_ratio` is considered to be the most important feature in the
model.

The figure shows two bars for each feature. A well-known issue with permutation
variable importance is that the presence of highly correlated features causes
permutation variable importance to become unreliable [@Hooker2019-pv]. If two
features are both important and highly correlated, shuffling one of them does
not decrease model performance as much. For this reason, familiar will permute
features together depending on their mutual correlation, and the provided
threshold values.

By default, similarity between features is assessed using McFadden's
pseudo-$R^2$. This metric can assess similarity between different types of
features, i.e. numeric and categorical. The default thresholds are 0.23 and 1.0
that respectively cluster similar features and permute all features
individually. The dataset contains many features that are highly
similar and exceed the relatively stringent threshold of 0.23.

## SHAP summary plots

A SHAP value is the marginal contributions of a feature value to the predicted
value. For example, if you were to predict the number of daily ice cream sales,
season could be a feature. A SHAP value for the summer season is likely
positive, whereas the SHAP value for the winter season would likely be negative.
Important features have larger SHAP values than less important features. Thus
a SHAP summary plot can be used to evaluate feature importance.


``` r
plots <- familiar::plot_shap_summary(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap_summary-1.png" alt="SHAP summary plot for the ischemic stroke model on hold-out
 test data. Features are sorted from highest (top) to least importance (bottom),
 based on SHAP values. Each point represents the marginal contribution 
 of a feature value to the predicted probability of the model, and is coloured
 according to its (normalised) feature value."  />
<p class="caption">SHAP summary plot for the ischemic stroke model on hold-out
 test data. Features are sorted from highest (top) to least importance (bottom),
 based on SHAP values. Each point represents the marginal contribution 
 of a feature value to the predicted probability of the model, and is coloured
 according to its (normalised) feature value.</p>
</div>

The SHAP summary plot shows several things: features are listed along the
y-axis, with the most important feature on top, i.e. the feature that has the
highest mean absolute SHAP value. Each point corresponds to the value in an
instance (a patient), and each point is coloured according to its (normalised)
value. For the purpose of representation, all feature values are mapped to the
[-1.0, 1.0] range.

The default presentation might provide too much detail. It is therefore possible
to change the `plot_type` or `value_representation`. For example,
the above data could be presented as violin plots:


``` r
plots <- familiar::plot_shap_summary(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  plot_type = "violinplot"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap_summary_violin-1.png" alt="SHAP summary plot for the ischemic stroke model on hold-out
 test data. Features are sorted from highest (top) to least importance (bottom),
 based on SHAP values. The distribution of SHAP values for each feature
 is shown as a violin plot."  />
<p class="caption">SHAP summary plot for the ischemic stroke model on hold-out
 test data. Features are sorted from highest (top) to least importance (bottom),
 based on SHAP values. The distribution of SHAP values for each feature
 is shown as a violin plot.</p>
</div>

The violin plot shows the distribution of values more clearly, at the cost 
of losing information concerning underlying feature values.

If just variable importance is relevant, the SHAP summary plot can also be shown
as a bar plot (`plot_type = "barplot"`) showing the mean absolute SHAP value by
default (`value_representation = "abs_mean"`):


``` r
plots <- familiar::plot_shap_summary(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  value_representation = "abs_mean"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap_summary_bar-1.png" alt="SHAP summary plot for the ischemic stroke model on hold-out
 test data. Features are sorted from highest (top) to least importance (bottom),
 based on the mean of the absolute SHAP values in the underlying data."  />
<p class="caption">SHAP summary plot for the ischemic stroke model on hold-out
 test data. Features are sorted from highest (top) to least importance (bottom),
 based on the mean of the absolute SHAP values in the underlying data.</p>
</div>

# Feature effects

Variable importance does not explain how features influence the model, although
the SHAP summary plot already provided some hints. Familiar offers multiple
methods for assessing the effect of features.

## Partial dependence and individual conditional expectation plots

Partial dependence (PD) [@Friedman2001-jf] and individual conditional
expectation (ICE) [@Goldstein2015-rv] plots can be used to visualise how
features influence the model. As an example, we can create an ICE plot for the
`max_remodeling_ratio` feature that was found to be important for predicting
ischemic stroke.


``` r
plots <- familiar::plot_ice(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  features = "max_remodeling_ratio"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/ice-max-remodeling-ratio-1.png" alt="Individual conditional expectation plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data. 
 Each curve follows the predicted probability for an instance where only
 the value of the max_remodeling_ratio is changed. The thick line corresponds to
 the average of all individual curves, i.e. partial dependence. High novelty scores
 signify values that are more likely to be outliers."  />
<p class="caption">Individual conditional expectation plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data. 
 Each curve follows the predicted probability for an instance where only
 the value of the max_remodeling_ratio is changed. The thick line corresponds to
 the average of all individual curves, i.e. partial dependence. High novelty scores
 signify values that are more likely to be outliers.</p>
</div>

The ICE plot shows multiple curves for individual samples, as well as their
average, the partial dependence plot as a somewhat thicker curve. These curves
are created by iteratively updating the value of `max_remodeling_ratio` and then
observing the predicted outcome. To generate the sample points, familiar samples
the feature value distribution, as observed in the training dataset, at regular
percentile intervals. Also note that by default, ICE plots only show a limited
number of samples to prevent clutter. This can be specified through
`n_max_samples_shown`.

The plot also shows novelty, which is a measure of how out-of-distribution the value
is. Novelty is computed using extended isolation forests [@Hariri2019-xz] and is
the average normalised height at which a sample is found in a terminal leaf node
of the trees in the isolation forest [@Liu2008-kw]. Novelty values range between
`0.00` and `1.00`. Novelty values should be interpreted with care. While increasing
novelty values represent instances that lie further from the distribution, the
values that represent in-distribution instances may vary. In the ICE plot above,
novelty values range between `0.40` and `0.50`. None of the points in ICE plot
can therefore be considered out-of-distribution. Move to a conformal prediction
framework is planned for future versions.

It is possible to anchor the curves in the ICE plot to a specific feature value.
This allows for assessing the model response for individual instances versus a
fixed value. In effect this can reduce some of the offset caused by other
(fixed) features in each instance, and may help to elucidate the observed
behaviour:


``` r
plots <- familiar::plot_ice(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  features = "max_remodeling_ratio",
  anchor_values = list("max_remodeling_ratio" = 2.5)
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/ice-max-remodeling-ratio-anchored-1.png" alt="Individual conditional expectation plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data. 
 Each curve follows the predicted probability for an instance where only
 the value of the max_remodeling_ratio is changed. This curve was anchored at
 max_remodeling_ratio=2.5, i.e. at this value, every curve has a (relative)
 predicted probability of 0.0. The thick line corresponds to the average of all
 individual curves, i.e. partial dependence. High novelty scores
 signify values that are more likely to be outliers."  />
<p class="caption">Individual conditional expectation plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data. 
 Each curve follows the predicted probability for an instance where only
 the value of the max_remodeling_ratio is changed. This curve was anchored at
 max_remodeling_ratio=2.5, i.e. at this value, every curve has a (relative)
 predicted probability of 0.0. The thick line corresponds to the average of all
 individual curves, i.e. partial dependence. High novelty scores
 signify values that are more likely to be outliers.</p>
</div>

In the plot above, we have anchored the values at `max-remodeling-ratio = 2.5`.
All curves are then drawn relative to the predicted probability found for this
particular value.

It is also possible to show the partial dependence of two features at the same
time:


``` r
plots <- familiar::plot_ice(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  features = c("max_remodeling_ratio", "coronary_artery_disease")
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/ice-2d-1.png" alt="Partial depence plot for the max_remodeling_ratio and coronary_artery_disease
 features in the ischemic stroke model using instances from the hold-out test data. 
 Individual curves are not shown for 2D plots. High novelty scores,
 indicated by smaller points, signify values that are more likely to be outliers."  />
<p class="caption">Partial depence plot for the max_remodeling_ratio and coronary_artery_disease
 features in the ischemic stroke model using instances from the hold-out test data. 
 Individual curves are not shown for 2D plots. High novelty scores,
 indicated by smaller points, signify values that are more likely to be outliers.</p>
</div>


The partial dependence plot above shows `max_remodeling_ratio` along the x-axis
and `coronary_artery_disease` along the y-axis. Higher intensities are
associated with a higher expected probability of malignancy. The figure thus
indicates that higher values of `max_remodeling_ratio` clearly affect the
predicted probability of ischemic stroke. The presence of
`coronary_artery_disease` can be seen to decrease the probability of ischemic
stroke - probably because these patients arrived at the hospital due to this
condition, not necessarily cerebral stroke. The average novelty score of the
underlying instances determines the size of each point.

It also possible to anchor 2D partial dependence plots:


``` r
plots <- familiar::plot_ice(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  features = c("max_remodeling_ratio", "coronary_artery_disease"),
  anchor_values = list(
    "max_remodeling_ratio" = 2.5,
    "coronary_artery_disease" = "no"
  ),
  show_novelty = FALSE
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/ice-2d-anchored-1.png" alt="Partial depence plot for the max_remodeling_ratio and coronary_artery_disease
 features in the ischemic stroke model using instances from the hold-out test data.
 Without novelty, elements are shown as rectangles to better show probability.
 Moreover, probabilities were anchored at max_remodeling_ratio = 2.5, and
 coronary_artery_disease = no. Individual curves are not shown for 2D plots."  />
<p class="caption">Partial depence plot for the max_remodeling_ratio and coronary_artery_disease
 features in the ischemic stroke model using instances from the hold-out test data.
 Without novelty, elements are shown as rectangles to better show probability.
 Moreover, probabilities were anchored at max_remodeling_ratio = 2.5, and
 coronary_artery_disease = no. Individual curves are not shown for 2D plots.</p>
</div>

We disabled novelty (`show_novelty=FALSE`) for plotting the 2D anchored plot.
This changes the appearance of the plot, which now consists of coloured
rectangles. The white dots show the position at which familiar samples the
feature values.

Familiar automatically determines the values at which numerical features are
sampled based on their distribution in the development dataset. Such values can
also be manually set using the `feature_x_range` and `feature_y_range`
arguments.

## SHAP dependence plots

SHAP dependence plots function similar to dependence plots. The SHAP dependence
plot below shows how the SHAP values of `max_remodeling_ratio` are related
to its feature values.


``` r
plots <- familiar::plot_shap_dependence(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  shap_feature = "max_remodeling_ratio"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap-dependence-1.png" alt="SHAP dependence plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data."  />
<p class="caption">SHAP dependence plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data.</p>
</div>
The above plot shows that with increasing value of `max_remodeling_ratio`, the
marginal contribution to the prediction changes from negative (low values
decrease the probability of the ischemic stroke) to positive (high values
increase the probability of the ischemic stroke).

SHAP dependence plots can also show interaction with other features. For
example, we can show the interaction with the `coronary_artery_disease`
feature. 


``` r
plots <- familiar::plot_shap_dependence(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  shap_feature = "max_remodeling_ratio",
  interaction_feature = "coronary_artery_disease"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap-dependence-interaction-1.png" alt="SHAP dependence plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data.
 Each point shows the corresponding value for coronary_artery_disease."  />
<p class="caption">SHAP dependence plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data.
 Each point shows the corresponding value for coronary_artery_disease.</p>
</div>
The addition of an interaction feature does not alter the shape of the plot, but
now individual values are coloured by the value of the `coronary_artery_disease`
feature for the same sample. It appears that `coronary_artery_disease` is not
strongly interacting with `max_remodeling_ratio`.


``` r
plots <- familiar::plot_shap_dependence(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  shap_feature = "max_remodeling_ratio",
  interaction_feature = "max_dilation_by_area"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap-dependence_interaction_2-1.png" alt="SHAP dependence plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data.
 Each point shows the corresponding value for max_dilation_by_area."  />
<p class="caption">SHAP dependence plot for the max_remodeling_ratio feature
 in the ischemic stroke model using instances from the hold-out test data.
 Each point shows the corresponding value for max_dilation_by_area.</p>
</div>
The `max_dilation_by_area` feature, however, is much more strongly correlated,
as shown in the SHAP dependence plot above.

## SHAP force plots

SHAP force plots show the *force* exerted by features for individual samples.
Samples are (by default) ordered by increasing predicted value, and the 
positive and negative marginal contributions are shown for each sample.


``` r
plots <- familiar::plot_shap_force(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap_force_plot-1.png" alt="SHAP force plot for the ischemic stroke model using all instances
 from the hold-out test data. The force plot show the stacked positive and 
 negative marginal contributions for all features."  />
<p class="caption">SHAP force plot for the ischemic stroke model using all instances
 from the hold-out test data. The force plot show the stacked positive and 
 negative marginal contributions for all features.</p>
</div>
Force plots are not highly informative. However, one can set one or more highlight
features. For example, setting `max_remodeling_ratio` as a highlight yields the
following figure:


``` r
plots <- familiar::plot_shap_force(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  highlight_feature = "max_remodeling_ratio"
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap_force_plot_highlight-1.png" alt="SHAP force plot for the ischemic stroke model using all instances
 from the hold-out test data. The force plot show the stacked positive and 
 negative marginal contributions for all features. The contributions by
 the max_remodeling_ratio feature are highlighted."  />
<p class="caption">SHAP force plot for the ischemic stroke model using all instances
 from the hold-out test data. The force plot show the stacked positive and 
 negative marginal contributions for all features. The contributions by
 the max_remodeling_ratio feature are highlighted.</p>
</div>

## SHAP waterfall plots

SHAP waterfall plots show how SHAP values for each feature value contribute to
the predicted value for an individual sample.


``` r
plots <- familiar::plot_shap_waterfall(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/shap_waterfall-1.png" alt="SHAP waterfall plot for the ischemic stroke model for a single sample
 from the hold-out test data. The waterfall plot shows the marginal
 contributions of each feature for a sample. Features with the largest
 contributions are shown at the top."  />
<p class="caption">SHAP waterfall plot for the ischemic stroke model for a single sample
 from the hold-out test data. The waterfall plot shows the marginal
 contributions of each feature for a sample. Features with the largest
 contributions are shown at the top.</p>
</div>
The above plot shows a waterfall plot for the first sample, which has a
predicted value of $0.43$, where the average predicted value in
`ischemic_stroke_data` (i.e. the hold-out test set that we initially defined) is
$0.54$. Feature values are shown on the left, and all features are ordered by
the magnitude of their corresponding SHAP value.

# Feature and sample similarity

Highly similar features can carry redundant information for a model. This can be
visualised using similarity heatmaps:


``` r
plots <- familiar::plot_feature_similarity(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  feature_similarity_metric = "spearman",
  rotate_x_tick_labels = TRUE
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/feature-similarity-1.png" alt="Feature similarity heatmap for features used in the ischemic
 stroke model. Similarity is computed using Spearman's rank correlation
 coefficient. Features are clustered by similarity, with strongly (anti-)correlated
 features being grouped together more closely."  />
<p class="caption">Feature similarity heatmap for features used in the ischemic
 stroke model. Similarity is computed using Spearman's rank correlation
 coefficient. Features are clustered by similarity, with strongly (anti-)correlated
 features being grouped together more closely.</p>
</div>

Because we specify `feature_similarity_metric = "spearman"` the heatmap displays
Spearman's correlation coefficient between the different features (even
categorical ones). The features are also ordered by similarity, so that clusters
of similar features can be identified. Note that the dendrograms depict cluster
distance, i.e. $1-|\rho|$ for Spearman's correlation coefficient.

It is also possible to view the sample clustering. This can be used to identify
if samples are readily stratified into clusters, which may or may not correlate
to the endpoint of interest. For the ischemic stroke dataset, this yields the
following heatmap:


``` r
plots <- familiar::plot_sample_clustering(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  feature_similarity_metric = "spearman",
  rotate_x_tick_labels = TRUE
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/sample-similarity-1.png" alt="Sample similarity heatmap of samples in the hold-out test set based on 
 features used in the ischemic stroke model. Similarity between features is
 computed using Spearman's rank correlation coefficient, whereas similarity
 between samples is computed using Gower's distance. Strongly (anti-)correlated
 features and samples being grouped together more closely, respectively.
 Feature values are normalised."  />
<p class="caption">Sample similarity heatmap of samples in the hold-out test set based on 
 features used in the ischemic stroke model. Similarity between features is
 computed using Spearman's rank correlation coefficient, whereas similarity
 between samples is computed using Gower's distance. Strongly (anti-)correlated
 features and samples being grouped together more closely, respectively.
 Feature values are normalised.</p>
</div>

The sample clustering figure shows an absence of clear structure among the
samples, nor a clear ordering of features with their observed class label
(ischemic stroke). The top dendrogram is the same as for feature clustering, as
it should be. The right dendrogram is formed after computing Gower's distance
between samples (by default), and shows clustering of samples. The heatmap
itself shows normalised values for the features, based on normalisation and
transformation parameters obtained during model development.

Normalisation options can be altered by changing the `show_normalised_data`
argument. However, in the ischemic stroke dataset features do not have the same
value ranges, so without normalisation the figure looks less interpretable:


``` r
plots <- familiar::plot_sample_clustering(
  object = ischemic_stroke_model,
  data = ischemic_stroke_data,
  feature_similarity_metric = "spearman",
  show_normalised_data = "none",
  rotate_x_tick_labels = TRUE
)
```
<div class="figure" style="text-align: center">
<img src="eval_and_explain/sample-similarity-unnormalised-1.png" alt="Sample similarity heatmap of samples in the hold-out test set based on 
 features used in the ischemic stroke model. Similarity between features is
 computed using Spearman's rank correlation coefficient, whereas similarity
 between samples is computed using Gower's distance. Strongly (anti-)correlated
 features and samples being grouped together more closely, respectively.
 Feature values were not normalised (for plotting), and are coloured according to
 their actual value."  />
<p class="caption">Sample similarity heatmap of samples in the hold-out test set based on 
 features used in the ischemic stroke model. Similarity between features is
 computed using Spearman's rank correlation coefficient, whereas similarity
 between samples is computed using Gower's distance. Strongly (anti-)correlated
 features and samples being grouped together more closely, respectively.
 Feature values were not normalised (for plotting), and are coloured according to
 their actual value.</p>
</div>

# References

---
title: "Variable importance methods"
author: "Alex Zwanenburg"
date: "2026-04-22"
output:
  rmarkdown::html_vignette:
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
  rmarkdown::github_document:
    html_preview: FALSE
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
bibliography: "refs.bib"
vignette: >
  %\VignetteIndexEntry{Variable importance methods}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---



Variable importance methods score the importance of each feature.
This information is then used by the learning algorithm to determine which and
how many features should be included in the model.

| **method** | **tag** | **binomial** | **multinomial** | **continuous** | **survival** |
|:------------|:------------|:-----:|:-----:|:-----:|:-----:|:-----:|
| **correlation** |
| Pearson's *r* | `pearson`                                             |   |   | × | × |
| Spearman's *ρ* | `spearman`                                           |   |   | × | × |
| Kendall's *τ* | `kendall`                                             |   |   | × | × |
| **concordance** |
| concordance^a^ | `concordance`                                        | × | × | × | × |
| **CORElearn** |
| information gain ratio | `gain_ratio`                                 | × | × |   |   |
| gini-index | `gini`                                                   | × | × |   |   |
| minimum description length | `mdl`                                    | × | × |   |   |
| ReliefF with exponential weighting of distance ranks | `relieff_exp_rank`       | × | × | × |   |
| **mutual information**|
| mutual information maximisation | `mim`                               | × | × | × | × |
| mutual information features selection | `mifs`                        | × | × | × | × |
| minimum redundancy maximum relevance | `mrmr`                         | × | × | × | × |
| **univariate regression** |
| univariate regression | `univariate_regression`                       | × | × | × | × |
| **multivariate regression**| 
| multivariate regression | `multivariate_regression`                   | × | × | × | × |
| **lasso regression**| 
| general^a^ | `lasso`                                                  | × | × | × | × |
| logistic | `lasso_binomial`                                           | × |   |   |   |
| multi-logistic | `lasso_multinomial`                                  |   | × |   |   |
| normal (gaussian) | `lasso_gaussian`                                  |   |   | × |   |
| cox | `lasso_cox`                                                     |   |   |   | × |
| **ridge regression**|
| general^a^ | `ridge`                                                  | × | × | × | × |
| logistic | `ridge_binomial`                                           | × |   |   |   |
| multi-logistic | `ridge_multinomial`                                  |   | × |   |   |
| normal (gaussian) | `ridge_gaussian`                                  |   |   | × |   |
| cox | `ridge_cox`                                                     |   |   |   | × |
| **elastic net regression**| 
| general^a,b^ | `elastic_net`                                          | × | × | × | × |
| logistic^b^ | `elastic_net_binomial`                                  | × |   |   |   |
| multi-logistic^b^ | `elastic_net_multinomial`                         |   | × |   |   |
| normal (gaussian)^b^ | `elastic_net_gaussian`                         |   |   | × |   |
| cox^b^ | `elastic_net_cox`                                            |   |   |   | × |
| **random forest (RFSRC) variable importance**|
| permutation^b^ | `random_forest_permutation`                                     | × | × | × | × |
| permutation (unoptimised) | `random_forest_permutation_default`                  | × | × | × | × |
| hold-out ^b^ | `random_forest_holdout`                                           | × | × | × | × |
| hold-out (unoptimised) | `random_forest_holdout_default`                         | × | × | × | × |
| **random forest (ranger) variable importance**|
| permutation^b^ | `random_forest_ranger_permutation`                              | × | × | × | × |
| permutation (unoptimised) | `random_forest_ranger_permutation_default`           | × | × | × | × |
| hold-out permutation^b^ | `random_forest_ranger_holdout_permutation`             | × | × | × | × |
| hold-out perm. (unoptimised) |`random_forest_ranger_holdout_permutation_default` | × | × | × | × | 
| impurity^b^ | `random_forest_ranger_impurity`                                    | × | × | × | x |
| impurity (unoptimised) | `random_forest_ranger_impurity_default`                 | × | × | × | x |
| **special methods**|
| no selection | `none`                                                 | × | × | × | × |
| random selection | `random`                                           | × | × | × | × |
| signature only | `signature_only`                                     | × | × | × | × |

Table: Overview of variable importance methods. ^a^ This is a general method where
an appropriate specific method will be chosen, or multiple distributions or
linking families are tested in an attempt to find the best option. ^b^ This
method requires hyperparameter optimisation.

## Configuration options

Variable importance methods and related options can be set using the
`variable_importance` parameter in the *xml* file or as function argument.

| **tag** / **argument** | **description** | **default** |
|:----------:|:-------------------------|:-----------:|
| `vimp_method` | The desired variable importance method. Multiple methods may be provided at the same time. This setting has no default and must be provided. | -- (required) |
| `vimp_method_parameter` | Several variable importance methods have hyperparameters that can be set and/or optimised. | -- (optional) |
| `vimp_aggregation_method` | The aggregation method used to aggregate feature ranks over different bootstraps. | `borda` |
| `vimp_aggregation_rank_threshold` | Several aggregation methods count features if they have a rank below the threshold, i.e. are among the most important features. If `NULL`, a dynamic threshold is decided through Otsu-thresholding. | `5` | 
| `parallel_vimp` | Enables parallel processing for variable importance. Ignored if `parallel = FALSE`. Overrides `parallel = TRUE`. | `TRUE` |

## Providing parameters for variable importance computation

Some of the variable importance methods, notably those based on random forests and
(penalised) regression, have parameters that can be configured. These parameters are
mentioned under the respective entries in the [Overview of variable importance
methods] section. Moreover, some of these parameters are model parameters. In
this case, these parameters are optimised using hyperparameter optimisation,
which is described in the *learning algorithms and hyperparameter optimisation* vignette.

The syntax for such parameters is the same as for hyperparameter optimisation.
For the `multivariate_regression` variable importance method the `alpha` parameter
(which determines feature elimination during forward selection) may be provided as
follows using the configuration file:
```
<vimp_method_parameter>
  <multivariate_regression>
    <alpha>0.05</alpha>
  </multivariate_regression>
</vimp_method_parameter>
```

Or as a nested list passed as the `vimp_method_parameter` argument to `summon_familiar`:
```
vimp_method_parameter = list("multivariate_regression"=list("alpha"=0.05))
```

# Overview of variable importance methods

The variable importance methods implemented in familiar are described in more
detail in this section.

## Correlation methods

Correlation methods determine variable importance by assessing the correlation
between a feature and the outcome of interest. High (anti-)correlation indicates
an important feature, whereas low (anti-)correlation indicates that a feature is
not directly related to the outcome. Correlation-based variable importance is
determined using the `cor` function of the `stats` package that is part of the R
core distribution [@rcore2018].

Three correlation coefficients can be computed:

- `pearson`: Pearson’s $r$
- `spearman`: Spearman’s $\rho$
- `kendall`: Kendall’s $\tau$

To compute correlation of features with survival outcomes, only samples with an
event are considered.

## Concordance methods

Concordance methods assess how well the ordering of feature values corresponds
to the ordering of the outcome. The method internally refers to the `kendall`
method for continuous outcomes. For categorical outcomes, i.e. `binomial` and
`multinomial`, concordance is assessed using the area under the receiver
operating characteristic curve. For survival outcomes, concordance is measured
using the `concordance_index`.

## CORElearn methods

Familiar provides an interface to several variable importance methods
implemented in the `CORElearn` package. These methods are the *Information Gain Ratio* 
(`gain_ratio`), the *Gini-index* (`gini`), *Minimum Description Length* (`mdl`) 
and *ReliefF* and *rReliefF* with exponential distance rank weighting
(`relieff_exp_rank`).

## Mutual information-based methods

Mutual information $I$ is a measure of interdependency between two variables $x$
and $y$. In the context of variable importance, $x$ is a feature vector and $y$ is
the outcome vector.

Computing mutual information requires that probability distributions of $x$ and
$y$ are known. In practice, we don't know either one. For categorical $x$ and $y$
we can use the sample estimates instead. For continuous or mixed data, the
situation is more complex, and is often resolved through binning.

In familiar we therefore use the following two approaches to compute mutual
information:

1. For binomial, multinomial and continuous outcomes mutual information is
computed using the `praznik` package.

2. For survival outcomes we adapted the approximation proposed by De 
Jay et al. after Gel'fand and Yaglom [@Gelfand1959-de; @De_Jay2013-yl] for
use with a concordance index: $I  = -0.5 \log(1 - (2 * (C-0.5))^2 + \epsilon)$, 
with $C$ the concordance index.

### Mutual information maximisation

The `mim` method is a univariate method that ranks each feature by its mutual
information with the outcome.

### Mutual information variable importance

Mutual information variable importance (MIFS) finds a feature set that maximises
mutual information [@Battiti1994-ja]. This is done using forward selection. As
in mutual information maximisation, mutual information $I_{y,j}$ between each
feature and the outcome is computed. Starting from a potential pool of all
features, the feature with the highest mutual information is selected and
removed from the pool.

The rest proceeds iteratively. The mutual information $I_{s,1j}$ between
the previously selected feature and the remaining features is computed. This
mutual information is also called *redundancy*. The feature with the highest 
mutual information with the outcome and least redundancy (i.e. maximum
$I_{y,j} - I_{s,1j}$) is selected next, and removed from the pool of remaining
features. Then the mutual information $I_{s,2j}$ between this feature and
remaining features is computed, and the feature that maximises
$I_{y,j} - I{s,1j} - I_{s,2j}$ is selected, and so forth.

The iterative process stops if there is no feature $j$ for which
$I_{y,j} - \sum_{i\in S}  I_{s,ij} > 0$, with $S$ being the subset of selected
features, or all features have been exhausted.

To reduce the number of required computations, the implementation in familiar
actively eliminates any feature $j$ for which
$I_{y,j} - \sum_{i\in S}  I_{s,ij} \leq 0$ at the earliest instance, as the
$\sum_{i\in S}  I_{s,ij}$ term will monotonously increase.

### Minimum redundancy maximum relevance

Minimum redundancy maximum relevance (mRMR) is similar to
MIFS but differs in the way redundancy is used during optimisation
[@Peng2005-oo]. Whereas for MIFS the optimisation criterion is
$I_{y,j} - \sum_{i\in S} I_{s,ij}$, in mRMR the optimisation criterion is
$I_{y,j} - \frac{1} {\left| S \right|} \sum_{i\in S} I_{s,ij}$,
with $\left| S \right|$ the number of features already selected.

To limit computational complexity, features for which
$I_{y,j} - \frac{1} {\left| S \right|} \sum_{i\in S} I_{s,ij} \leq 0$,
are eliminated.

## Univariate and multivariate regression methods

Univariate and multivariate regression use a single feature or set of features
as predictors for regression models, respectively. The performance of the
regression model is then measured using a metric. Training and testing of
regression models are repeated multiple times using bootstraps. For each
bootstrap, the in-bag samples are used for training and the out-of-bag samples
are using for testing.

This also defines the parameters of both methods, which are shown in the table
below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| regression learner | `learner` | dependent on outcome | no | Any generalised linear regression model from the *learning algorithms and hyperparameter optimisation* vignette can be selected. Default values are `glm_logistic` for binomial, `glm_multinomial` for multinomial, `glm_gaussian` for continuous, and `cox` for survival outcomes. |
| performance metric | `metric` | dependent on outcome | no | Any metric from the *performance metrics* vignette can be selected. Default values are `auc_roc` for binomial and multinomial, `mse` for continuous, and `concordance_index` for survival outcomes |
| number of bootstraps | `n_bootstrap` | $\mathbb{Z} \in \left[1, \infty\right)$ | no | The default value is $10$. |
| drop-out alpha level | `alpha` | $\mathbb{R} \in \left[0, 1\right]$ | no | The default value is $0.33$. Only used in multivariate regression. |

### Univariate regression

In the univariate regression method, a regression model is built with each
feature separately using the in-bag data of the bootstrap. Then this model is
evaluated using the metric, expressed as an objective score with standardised optimum (see
*computing the objective score* in the *learning algorithms and hyperparameter optimisation* vignette). 

Then, the median of objective scores on out-of-bag data is used as variable
importance for each feature.

### Multivariate regression

The procedure described for univariate regression forms the first step in
multivariate regression. The rest follows forward selection. The most
important feature is assigned to the subset of selected features and removed
from the set of available features. Separate regression models are then built
with each remaining feature and all the feature(s) in the selected feature
subset as predictors. Thus, the subset of selected features iteratively increases
in size until no features are remaining or the objective score no longer
increases.

To limit mostly redundant computation, features that are unlikely to be selected
are actively removed. To do so, the standard deviation of the objective score
over the bootstraps is computed for each feature. The fraction `alpha` of the 
lowest scoring features is eliminated each iteration.

## Lasso, ridge and elastic net regression

Penalised regression is also a form of feature selection, as it selects an
'optimal' set of features to create a regression model. As features are usually
normalised as part of pre-processing, the magnitude of each coefficient can be
interpreted as its importance. All three shrinkage methods are implemented using
the `glmnet` package [@Hastie2009-ed; @Simon2011-ih].

Only elastic net regression has a model hyperparameter that requires
optimisation, but other parameters may be set as well, as shown in the table
below:

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| family | `family` | `gaussian`, `binomial`, `poisson`, `multinomial`, `cox` | continuous outcomes | For continuous outcomes `gaussian` and `poisson` may be tested. The family is not optimised when it is specified, e.g. `lasso_gaussian`. For other outcomes only one applicable family exists.|
| elastic net penalty | `alpha` | $\mathbb{R} \in \left[0,1\right]$ | elastic net | This penalty is fixed for ridge regression (`alpha = 0`) and lasso (`alpha = 1`). |
| optimal lambda | `lambda_min` | `lambda.1se`, `lambda.min` | no | Default is `lambda.min`.|
| number of CV folds | `n_folds` | $\mathbb{Z} \in \left[3,n\right]$ | no | Default is $3$ if $n<30$, $\lfloor n/10\rfloor$ if $30\leq n \leq 200$ and $20$ if $n>200$.|
| normalisation | `normalise` | `FALSE`, `TRUE` | no | Default is `FALSE`, as normalisation is part of pre-processing in familiar.|

## Random forest-based methods

Several feature selection methods are based on random forests. All these methods
require training a random forest. Hence, `familiar` will train a random
forest based on the training data prior to determining variable importance.
Random forest learners have a set of hyperparameters that can be optimised prior
to training. These parameters, which are slightly different
for `ranger`-based and `randomForestSRC`-based methods, are shown below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| number of trees | `n_tree` | $\mathbb{Z} \in \left[0,\infty\right)$ | yes |  This parameter is expressed on the $\log_{2}$ scale, i.e. the actual input value will be $2^\texttt{n_tree}$ [@Oshiro2012-mq]. The default range is $\left[4, 10\right]$.|
| subsampling fraction | `sample_size` | $\mathbb{R} \in \left(0, 1.0\right]$ | yes | Fraction of available data that is used for to create a single tree. The default range is $\left[2 / m, 1.0\right]$, with $m$ the number of samples. |
| number of features at each node | `m_try` | $\mathbb{R} \in \left[0.0, 1.0\right]$ | yes | Familiar ensures that there is always at least one candidate feature. |
| node size | `node_size` | $\mathbb{Z} \in \left[1, \infty\right)$ | yes | Minimum number of unique samples in terminal nodes. The default range is $\left[5, \lfloor m / 3\rfloor\right]$, with $m$ the number of samples.|
| maximum tree depth | `tree_depth` | $\mathbb{Z} \in \left[1,\infty\right)$ | yes | Maximum depth to which trees are allowed to grow. The default range is $\left[1, 10\right]$. |
| number of split points | `n_split` | $\mathbb{Z} \in \left[0, \infty\right)$ | no | By default, splitting is deterministic and has one split point ($0$).|
| splitting rule (`randomForestSRC` only) | `split_rule` | `gini`, `auc`, `entropy`, `mse`, `quantile.regr`, `la.quantile.regr`, `logrank`, `logrankscore`, `bs.gradient` | no | Default splitting rules are `gini` for `binomial` and `multinonial` outcomes, `mse` for `continuous` outcomes and `logrank` for `survival` outcomes.|
| splitting rule (`ranger` only) | `split_rule` | `gini`, `hellinger`, `extratrees`, `beta`, `variance`, `logrank`, `C`, `maxstat`| no | Default splitting rules are `gini` for `binomial` and `multinomial` outcomes and `maxstat` for `continuous` and `survival` outcomes.|
| significance split threshold (`ranger` only) | `alpha` | $\mathbb{R} \in \left(0.0, 1.0\right]$ | `maxstat` | Minimum significance level for further splitting. The default range is $\left[10^{-6}, 1.0\right]$ |

The unoptimised methods do not require hyperparameter optimisation, and use
default values from the `ranger` and `randomForestSRC`.

### Permutation importance

The permutation importance method is implemented by `random_forest_permutation`
and `random_forest_permutation_default` (`randomForestSRC` package) and
`random_forest_ranger_permutation` and
`random_forest_ranger_permutation_default` (`ranger` package). In short, this
method functions as follows [Ishwaran2007-va]. As usual, each tree in the random
forest is constructed using the in-bag samples of a bootstrap of the data. The
predictive performance of each model is first measured using the out-of-bag
data. Subsequently, the out-of-bag instances for each feature are randomly
permuted, and predictive performance is assessed again. The difference between
the normal performance and the permuted performance is used as a measure of the
variable importance. For important features, this difference is large, whereas
for irrelevant features the difference is negligible or even negative.

### Holdout permutation importance

This variant on permutation importance
(`random_forest_ranger_holdout_permutation` and
`random_forest_ranger_holdout_permutation_default`) is implemented using
`ranger::holdoutRF`. Instead of using out-of-bag to compute feature importance,
two cross-validation folds are used. A random forest is trained on either fold,
and variable importance determined on the other [@Janitza2018-kl].

The hold-out variable importance method implemented in the `randomForestSRC`
package (`random_forest_holdout` and `random_forest_holdout_default`) is
implemented using `randomForestSRC::holdout.vimp`. It is similar to the previous
variant, but does not use cross-validation folds. Instead, out-of-bag prediction
errors for models trained with and without each feature are compared.

### Minimum depth variable selection

Important features tend to appear closer to the root of trees in random forests.
Therefore, the position of each feature within a tree is assessed in minimum
depth variable selection [@Ishwaran2010-zv].

### Impurity importance

At each node, the data is split into (two) subsets, which connects to two
branches. After splitting, each single subset is purer than the parent dataset.
As a concrete example, in regression problems the variance of each of the
subsets is lower than that of the data prior to splitting. The decrease in
variance specifically, or the decrease of impurity generally, is then used to
assess feature importance.

`familiar` uses the `impurity_corrected` importance measure, which is unbiased
to the number of split points of a feature and its distribution
[@Nembrini2018-ay].

## Special methods

Familiar offers several methods that are special in that they are not feature
selection methods in the sense that they determine a variable importance that
can be used for establishing feature rankings.

### No variable importance

As the name suggests, the `none` method avoids computing variable importances
altogether. All features are passed into a model. Feature order is randomly
shuffled prior to building a model to avoid influence of the order of provided
features.

### Random variable importance

The `random` method randomly draws features prior to model building. It does not
assign a random variable importance to a feature. New features are drawn each
time a model is built. All features are available for the draw, but only $m$
features are drawn. Here $m$ is the signature size that is usually optimised by
hyperparameter optimisation.

### Signature only

When configuring familiar, any number of features can be set as a model
signature using the `signature` configuration parameter. However, more features
may be added to this signature after computing variable importances. To make
sure that only the provided features enter a model, the `signature_only` method
may be used.

# Aggregating variable importance

In case of variable importance or modelling in the presence of subsampling (e.g.
bootstraps), the ranks of features may need to be aggregated across the
different instances [@Wald2012-zk]. The rank aggregation methods shown in the
table below can be used for this purpose. Several methods require a threshold to
indicate the size of the set of most highly ranked features, which can be set by
specifying the `vimp_aggregation_rank_threshold` configuration parameter.

| **aggregation method** | **tag** | **comments** |
|:--------------|:--------|:----------:|
| none | `none` | |
| mean rank | `mean` | |
| median rank | `median` | |
| best rank | `best` | |
| worst rank | `worst` | |
| stability selection | `stability` | uses threshold |
| exponential selection | `exponential` | uses threshold  |
| borda ranking | `borda` | |
| enhanced borda ranking | `enhanced_borda`| uses threshold |
| truncated borda ranking | `truncated_borda` | uses threshold |
| enhanced truncated borda ranking | `enhanced_truncated_borda` | uses threshold |

## Notation

Let $N$ be the number of ranking experiments that should be aggregated. Feature
$i$ for experiment $j$ of $N$ then has rank $r_{ij}$. A lower rank indicates a
more important feature. Some features may not receive a score during a ranking
experiment, for example for multivariate variable importance methods such as
lasso regression, or through use of a threshold $\tau$. This is designated by
$\delta_{ij}$, which is $0$ if the feature is absent, and $1$ if it is present.

In case a threshold is used, $\delta_{ij} = 1$ if $r_{ij} \leq \tau$, and $0$
otherwise.

Thus, for each experiment $m_j = \sum^M_{i=1} \delta_{ij}$ features are ranked,
out of $M$ features. $m_j$ is then also the maximum rank found in experiment
$j$.

Aggregating ranks for each feature results in an aggregate rank score $s_i$.
Features are subsequently ranked according to this method-specific score to
arrive at an aggregate feature rank $r_i$.

## No rank aggregation

The `none` option does not aggregate ranks. Rather, scores are aggregated by
computing the average score of a feature over all experiments that contain it.
Ranks are then computed from the aggregated scores.

## Mean rank aggregation

The mean rank aggregation method ranks features by computing the mean rank of a
feature across all experiments that contain it.

$$s_i = \frac{\sum^{N}_{j=1} \delta_{ij} r_{ij}}{\sum^{N}_{j=1} \delta_{ij}}$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in ascending order.

## Median rank aggregation

The median rank aggregation method ranks features by computing the median rank
of a feature across all experiments that contain it.

$$s_i = \underset{j \in N, \, \delta_{ij}=1}{\textrm{median}}(r_{ij})$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in ascending order.

## Best rank aggregation

The best rank aggregation method ranks features by the best rank that a feature
has across all experiments that contain it.

$$s_i =  \underset{j \in N, \, \delta_{ij}=1}{\textrm{min}} (r_{ij})$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in ascending order.

## Worst rank aggregation

The worst rank aggregation method ranks features by the worst rank that a
feature has across all instances that contain it.

$$s_i =  \underset{j \in N, \, \delta_{ij}=1}{\textrm{max}} (r_{ij})$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in ascending order.

## Stability rank aggregation

The stability aggregation method ranks features by their occurrence within the
set of highly ranked features across all experiments. Our implementation
generalises the method originally proposed by Meinshausen and Bühlmann
[@Meinshausen2010-do].

This method uses threshold $\tau$ to designate the highly ranked features. Thus
$\delta_{ij} = 1$ if $r_{ij} \leq \tau$, and $0$ otherwise.

The aggregate rank score is computed as:

$$s_i = \frac{1}{N} \sum^N_{j=1} \delta_{ij}$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in descending order, as more commonly occurring features are considered
more important.

## Exponential rank aggregation

The exponential aggregation method ranks features by the sum of the negative
exponentials of their normalised ranks in instances where they occur within the
set of highly ranked features. This method was originally suggested by Haury et
al. [@Haury2011-zd].

This method uses threshold $\tau$ to designate the highly ranked features. Thus
$\delta_{ij} = 1$ if $r_{ij} \leq \tau$, and $0$ otherwise.

$$s_i = \sum^N_{j=1} \delta_{ij} \exp({-r_{ij} / \tau)}$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in descending order.

## Borda rank aggregation

Borda rank aggregation ranks a feature by the sum of normalised ranks (the borda
score) across all experiments that contain it. In case every experiment contains all
features, the result is equivalent to the mean aggregation method [@Wald2012-zk].

$$s_i = \sum^N_{j=1} \frac{m_j - r_{ij} + 1}{m_j}$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in descending order.

## Enhanced borda rank aggregation

Enhanced borda rank aggregation combines borda rank aggregation with stability
rank aggregation. The borda score is multiplied by the occurrence of the feature
within the set of highly ranked features across all experiments [@Wald2012-zk].

This method uses threshold $\tau$ to designate the highly ranked features for
the purpose of computing the occurrence. Thus $\delta_{ij} = 1$ if
$r_{ij} \leq \tau$, and $0$ otherwise.

$$s_i = \left( \frac{1}{N} \sum^N_{j=1} \delta_{ij} \right) \left( \sum^N_{j=1}
\frac{m_j - r_{ij} + 1}{m_j} \right)$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in descending order.

## Truncated borda rank aggregation

Truncated borda rank aggregation is borda rank aggregation performed with only
the set of most highly ranked features in each instance.

This method uses threshold $\tau$ to designate the highly ranked features. Thus
$\delta_{ij} = 1$ if $r_{ij} \leq \tau$, and $0$ otherwise.

$$s_i = \sum^N_{j=1} \delta_{ij} \frac{\tau - r_{ij} + 1}{\tau}$$

Note that compared to the borda method, the number of ranked features in an
experiment $m_j$ is replaced by threshold $\tau$.

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in descending order.

## Truncated enhanced borda rank aggregation

Truncated enhanced borda rank aggregation is enhanced borda aggregation
performed with only the set of most highly ranked features in each experiment.

This method uses threshold $\tau$ to designate the highly ranked features. Thus
$\delta_{ij} = 1$ if $r_{ij} \leq \tau$, and $0$ otherwise.

$$s_i = \left( \frac{1}{N} \sum^N_{j=1} \delta_{ij} \right) \left( \sum^N_{j=1}
\delta_{ij} \frac{\tau - r_{ij} + 1}{\tau} \right)$$

The aggregate rank of features is then determined by sorting aggregate scores
$s_i$ in descending order.

# References

---
title: "Learning algorithms and hyperparameter optimisation"
author: "Alex Zwanenburg"
date: "2026-04-22"
output:
  rmarkdown::html_vignette:
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
  rmarkdown::github_document:
    html_preview: FALSE
    includes:
      in_header: familiar_logo.html
      after_body: license.html
    toc: TRUE
bibliography: "refs.bib"
vignette: >
  %\VignetteIndexEntry{Learning algorithms and hyperparameter optimisation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


Learning algorithms create models that relate input data to the outcome.
Development data is used to train models, after which they can be used to
predict outcomes for new data. Familiar implements several commonly used
algorithms. This vignette first describes the learners and their
hyperparameters. Then, hyperparameter optimisation is described in more detail.

|**learner** | **tag** | **binomial** | **multinomial** | **continuous** | **survival** |
|:------------|:------------|:-----:|:-----:|:-----:|:-----:|:-----:|
|**generalised linear models**|
| general^a^ | `glm`                        	    | × | × | × | × |
| logistic | `glm_logistic`              		      | × |   |   |   |
| cauchy | `glm_cauchy` 						              | × |   |   |   |
| complementary log-log | `glm_loglog` 			      | × | 	| 	|  	|
| normal | `glm_probit` 						              | × | 	| 	| 	|
| multinomial | `glm_multinomial` 				        | 	| × | 	| 	|
| log-normal | `glm_log`, `glm_log_gaussian` 	    | 	| 	| × |  	|
| normal (gaussian) | `glm_gaussian` 			        | 	| 	| × |  	|
| inverse gaussian | `glm_inv_gaussian` 		      | 	| 	| × |  	|
| log-poisson | `glm_log_poisson` 				        | 	| 	| x	|  	|
| poisson | `glm_poisson` 						            | 	| 	| x	|  	|
| cox | `cox` 									                  | 	| 	| 	|	× |
| exponential | `survival_regr_exponential` 	    | 	| 	| 	|	× |
| gaussian | `survival_regr_gaussian` 			      | 	| 	| 	| × |
| logistic | `survival_regr_logistic` 			      | 	| 	| 	| × |
| log-logistic | `survival_regr_loglogistic` 	    | 	| 	| 	| × |
| log-normal | `survival_regr_lognormal` 		      | 	| 	| 	| × |
| survival regression^a^ | `survival_regr` 		    | 	| 	| 	| × |
| weibull | `survival_regr_weibull` 			        | 	| 	| 	| × |
|**lasso regression models**|
| general^a^ | `lasso` 							              | × | × | × | × |
| logistic | `lasso_binomial` 					          | × | 	| 	| 	|
| multi-logistic | `lasso_multinomial` 			      | 	| × | 	| 	|
| normal (gaussian) | `lasso_gaussian` 			      | 	| 	| × |  	|
| poisson | `lasso_poisson`						            | 	| 	| 	|   |
| cox | `lasso_cox` 							                | 	| 	| 	| × |
|**ridge regression models**|
| general^a^ | `ridge`                            | × | × | × | × |
| logistic | `ridge_binomial`                     | × |   |   |   |
| multi-logistic | `ridge_multinomial`            |   | × |   |   |
| normal (gaussian) | `ridge_gaussian`            |   |   | × |   |
| poisson | `ridge_poisson`                       |   |   |   |   |
| cox | `ridge_cox`                               |   |   |   | × |
|**elastic net regression models**|
| general^a^ | `elastic_net`                      | × | × | × | × |
| logistic | `elastic_net_binomial`               | × |   |   |   |
| multi-logistic | `elastic_net_multinomial`      |   | × |   |   |
| normal (gaussian) | `elastic_net_gaussian`      |   |   | × |   |
| poisson | `elastic_net_poisson`                 |   |   |   |   |
| cox | `elastic_net_cox`                         |   |   |   | × |
|**extreme gradient boosted linear models**|
| general^a^ | `xgboost_lm`                       | × | × | × | × |
| logistic | `xgboost_lm_logistic`                | × | × |   |   |
| gamma | `xgboost_lm_gamma`                      |   |   | × |   | 
| gaussian | `xgboost_lm_gaussian`                |   |   | × |   |
| poisson | `xgboost_lm_poisson`                  |   |   | × |   |
| cox | `xgboost_lm_cox`                          |   |   |   | × |
|**random forests**|
| random forest (RFSRC) | `random_forest_rfsrc`   | × | × | × | × |
| random forest (ranger) | `random_forest_ranger` | × | × | × | × | 
|**extreme gradient boosted trees**|
| general^a^ | `xgboost_tree`                     | × | × | × | × |
| logistic | `xgboost_tree_logistic`              | × | × |   |   |
| gamma | `xgboost_tree_gamma`                    |   |   | × |   | 
| gaussian | `xgboost_tree_gaussian`              |   |   | × |   |
| poisson | `xgboost_tree_poisson`                |   |   | × |   |
| cox | `xgboost_tree_cox`                        |   |   |   | × |
|**extreme gradient boosted DART trees**|
| general^a^ | `xgboost_dart`                     | × | × | × | × |
| logistic | `xgboost_dart_logistic`              | × | × |   |   |
| gamma | `xgboost_dart_gamma`                    |   |   | × |   | 
| gaussian | `xgboost_dart_gaussian`              |   |   | × |   |
| poisson | `xgboost_dart_poisson`                |   |   | × |   |
| cox | `xgboost_dart_cox`                        |   |   |   | × |
|**bayesian models**|
| naive bayes | `naive_bayes`                     | × | × |   |   |
|**nearest neighbour models**|
| k-nearest neighbours^b^ | `k_nearest_neighbours_*`, `knn_*`           | × | × | x |   |
|**support vector machines**|
| $C$-classification^c^ | `svm_c_*`                                     | × | × |   |   |
| $\nu$-classification/ regression^c^ | `svm_nu_*`                      | × | × | × |   |
| $\epsilon$-regression^c^ | `svm_eps_*`                                |   |   | × |   |
| linear kernel^d^ | `svm_*_linear`               | × | × | × |   |
| polynomial kernel^d^ | `svm_*_polynomial`       | × | × | × |   |
| radial kernel^d^ | `svm_*_radial`               | × | × | × |   |
| sigmoid kernel^d^ | `svm_*_sigmoid`             | × | × | × |   |

Table: Overview of learners implemented in familiar. ^a^ These learners test
multiple distributions or linking families and attempt to find the best option.
^b^ k-nearest neighbours learners allow for setting the distance metric using
`*`. If omitted (e.g. `knn`), the kernel is either `euclidean` (numeric
features) or `gower` (mixed features). ^c^ The SVM kernel is indicated by `*`.
If omitted (e.g. `svm_c`), the radial basis function is used as kernel. ^d^ SVM
type is indicated by `*`, and must be one of `c`, `nu`, or `epsilon`.

## Configuration options

Learners, their hyperparameters, and parallelisation options can be specified
using the tags or arguments below:

| **tag** / **argument** | **description** | **default** |
|:----------:|:-------------------------|:-----------:|
| `learner` | The desired learner. Multiple learners may be provided at the same time. This setting has no default and must be provided. | -- (required) |
| `hyperparameter` | Each learner has one or more hyperparameters which can be specified. Sequential model-based optimisation is used to identify hyperparameter values from the data unless these are specifically specified here. See the section on hyperparameter optimisation for more details. | -- (optional) |
| `parallel_model_development` | Enables parallel processing for model development. Ignored if `parallel = FALSE`. Overrides `parallel = TRUE`. | `TRUE` |

Table: Configuration options for model development.

## Generalised linear models

Generalised linear models (GLM) are easy to understand, use and share. In many
situations, GLM may be preferred over more complex models as they are easier to
report and validate. The most basic GLM is the linear model. The linear model
for an outcome variable $y_j$ and a single predictor variable $x_j$ for sample
$j$ is:

$$y_j=\beta_0 + \beta_1 x_j + \epsilon_j$$
Here, $\beta_0$ and $\beta_1$ are both model coefficients. $\beta_0$ is also
called the model intercept. $\epsilon_j$ is an error term that describes the
difference between the predicted and the observed outcome for sample $j$. When a
linear model is developed, the $\beta$-coefficients are fitted so that
the mean-squared error over the sample population is minimised.

The above model is a univariate model as it includes only a single predictor. A
multivariate linear model includes multiple predictors $\mathbf{X_j}$ and is
denoted as:

$$y_j= \mathbf{\beta}\mathbf{X_j} +  \epsilon_j$$

$\mathbf{\beta}\mathbf{X_j}$ is the linear predictor. The GLM generalises this
linear predictor through a linking function $g$ [@Nelder1972-rs]:

$$y_j=g\left(\mathbf{\beta}\mathbf{X}\right) + \epsilon_j$$

For example, binomial outcomes are commonly modelled using logistic regression
with the `logit` linking function. Multiple linking functions are available in
familiar and are detailed below.

### Linear models for binomial outcomes

Linear models for binomial outcomes derive from the `stats` package which is
part of the R core distribution [@rcore2018]. Hyperparameters for these models
include linkage functions and are shown in the table below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size |`sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| linking function |`family` | `logistic`, `probit`, `loglog`, `cauchy` | `glm` only | The linking function is not optimised when it is specified, e.g. `glm_logistic`.|
| sample weighting | `sample_weighting` | `inverse_number_of_samples`, `effective_number_of_samples`, `none` | no | Sample weighting allows for mitigating class imbalances. The default is `inverse_number_of_samples`. Instances with the majority class receive less weight. |
| beta | `sample_weighting_beta` | $\mathbb{Z} \in \left[-6,-1\right]$ | `effective_number_of_samples` only | Specifies the $\beta$ parameter for effective number of samples weighting [@Cui2019-ys]. It is expressed on the $\log_{10}$ scale: $\beta=1-10^\texttt{beta}$. |

### Linear models for multinomial outcomes

The linear model for multinomial outcomes is implemented using `nnet::multinom`
[@Venables2002-ma]. Hyperparameters for this model are shown in the table below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size |`sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| linking function |`family` | `multinomial` | no | There is only one linking function available.|
| sample weighting | `sample_weighting` | `inverse_number_of_samples`, `effective_number_of_samples`, `none` | no | Sample weighting allows for mitigating class imbalances. The default is `inverse_number_of_samples`. Instances with the majority class receive less weight. |
| beta | `sample_weighting_beta` | $\mathbb{Z} \in \left[-6,-1\right]$ | `effective_number_of_samples` only | Specifies the $\beta$ parameter for effective number of samples weighting [@Cui2019-ys]. It is expressed on the $\log_{10}$ scale: $\beta=1-10^\texttt{beta}$. |

### Linear models for continuous outcomes

Linear models for continuous outcomes derive from the `stats`
package of the R core distribution [@rcore2018]. Hyperparameters for these
models include linkage functions and are shown in the table below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size |`sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| linking function |`family` | `gaussian`, `log_gaussian`, `inv_gaussian`, `poisson`, `log_poisson` | `glm` only | The linking function is not optimised when it is specified, e.g. `glm_poisson`.|

### Linear models for survival outcomes

Linear models for survival outcomes are divided into semi-parametric and
parametric models. The semi-parametric Cox proportional hazards model
[@Cox1972-fc] is based on the `survival::coxph` function [@Therneau2000-jv].
Tied survival times are resolved using the default method by @Efron1977-ww.

Various fully parametric models are also available, which differ in the assumed
distribution of the outcome, i.e. the linking function. The parametric models
are all based on `survival::survreg` [@Therneau2000-jv].

Hyperparameters for semi-parametric and parametric survival models are shown in
the table below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size |`sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| linking function |`family` | `weibull`, `exponential`, `gaussian`, `logistic`, `lognormal`, `loglogistic` | `survival_regr` only | The linking function is not optimised when it is specified, e.g. `survival_regr_weibull`. The non-parametric `cox` learner does not have a linking function.|

## Lasso, ridge and elastic net regression

Generalised linear models can be problematic as there is no inherent limit to
model complexity, which can easily lead to overfitting. Penalised regression, or
shrinkage, methods address this issue by penalising model complexity.

Three shrinkage methods are implemented in `familiar`, namely ridge regression,
lasso and elastic net, which are all implemented using the `glmnet` package
[@Hastie2009-ed; @Simon2011-ih]. The set of hyperparameters is shown in the
table below. The optimal `lambda` parameter is determined by cross-validation as
part of the `glmnet::cv.glmnet` function, and is not directly determined using
hyperparameter optimisation.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size |`sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| elastic net penalty | `alpha` | $\mathbb{R} \in \left[0,1\right]$ | elastic net | This penalty is fixed for ridge regression (`alpha = 0`) and lasso (`alpha = 1`). |
| optimal lambda | `lambda_min` | `lambda.1se`, `lambda.min` | no | Default is `lambda.min`.|
| number of CV folds | `n_folds` | $\mathbb{Z} \in \left[3,n\right]$ | no | Default is $3$ if $n<30$, $\lfloor n/10\rfloor$ if $30\leq n \leq 200$ and $20$ if $n>200$.|
| sample weighting | `sample_weighting` | `inverse_number_of_samples`, `effective_number_of_samples`, `none` | no | Sample weighting allows for mitigating class imbalances. The default is `inverse_number_of_samples` for `binomial` and `multinomial` outcomes, and `none` otherwise. Instances with the majority class receive less weight. |
| beta | `sample_weighting_beta` | $\mathbb{Z} \in \left[-6,-1\right]$ | `effective_number_of_samples` only | Specifies the $\beta$ parameter for effective number of samples weighting [@Cui2019-ys]. It is expressed on the $\log_{10}$ scale: $\beta=1-10^\texttt{beta}$. |
| normalisation | `normalise` | `FALSE`, `TRUE` | no | Default is `FALSE`, as normalisation is part of pre-processing in familiar.|

## Extreme gradient boosted linear models and trees

Boosting is a procedure which combines many weak learners to form a more
powerful panel [@Schapire1990-vn; @Hastie2009-ed]. Extreme gradient boosting is
a gradient boosting implementation that was highly successful in several machine
learning competitions. Learners are implemented using the `xgboost` package
[@Chen2016-lo]. Three types are implemented: regression based boosting
(`xgboost_lm`), regression-tree based boosting (`xgboost_tree`) and
regression-tree based boosting with drop-outs (`xgboost_dart`).

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size | `sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| family | `learn_objective` | `gaussian`, `multinomial_logistic`, `binomial_logistic`, `poisson`, `gamma`, `cox` | general tags | The family is not optimised when it is specified, e.g. `xgboost_lm_linear`.|
| boosting iterations | `n_boost` | $\mathbb{R} \in \left[0,\infty \right)$ | yes | This parameter is expressed on the $\log_{10}$ scale, i.e. the actual value will be $10^\texttt{n_boost}$. The default range is  $\left[0, 3 \right]$.|
| learning rate | `learning_rate` | $\mathbb{R} \in \left(-\infty,0\right]$ | yes | This parameter is expressed on the $\log_{10}$ scale. The default range is $\left[-5, 0 \right]$.|
| L1 regularisation | `alpha` | $\mathbb{R} \in \left[-6, \infty\right)$ | yes | This parameter is expressed on the $\log_{10}$ scale with a $10^{-6}$ offset. The default range is $\left[-6, 3\right]$.|
| L2 regularisation | `lambda` | $\mathbb{R} \in \left[-6, \infty\right)$ | yes | This parameter is expressed on the $\log_{10}$ scale with a $10^{-6}$ offset. The default range is $\left[-6, 3\right]$.|
| sample weighting | `sample_weighting` | `inverse_number_of_samples`, `effective_number_of_samples`, `none` | no | Sample weighting allows for mitigating class imbalances. The default is `inverse_number_of_samples` for `binomial` and `multinomial` outcomes, and `none` otherwise. Instances with the majority class receive less weight. |
| beta | `sample_weighting_beta` | $\mathbb{Z} \in \left[-6,-1\right]$ | `effective_number_of_samples` only | Specifies the $\beta$ parameter for effective number of samples weighting [@Cui2019-ys]. It is expressed on the $\log_{10}$ scale: $\beta=1-10^\texttt{beta}$. |
| maximum tree depth | `tree_depth` | $\mathbb{Z} \in \left[1,\infty\right)$ | `gbtree`, `dart` | Maximum depth to which trees are allowed to grow. The default range is $\left[1, 10\right]$. |
| subsampling fraction | `sample_size` | $\mathbb{R} \in \left(0, 1.0\right]$ | `gbtree`, `dart` | Fraction of available data that is used for to create a single tree. The default range is $\left[2 / m, 1.0\right]$, with $m$ the number of samples. |
| minimum sum of instance weight | `min_child_weight` | $\mathbb{R} \in \left[0,\infty\right)$ | `gbtree`, `dart` | Minimal instance weight required for further branch partitioning, or the number of instances required in each node. This parameter is expressed on the $\log_{10}$ scale with a $-1$ offset. The default range is $\left[0, 2\right]$.|
| min. splitting error reduction | `gamma` | $\mathbb{R} \in \left[-6, \infty\right)$ | `gbtree`, `dart` | Minimum error reduction required to allow splitting. This parameter is expressed on the $\log_{10}$ scale with a $10^{-6}$ offset. The default range is $\left[-6, 3\right]$. `continuous`-type outcomes are normalised to the $\left[0, 1\right]$ range prior to model fitting to deal with scaling issues. Values are converted back to the original scale after prediction. |
| DART sampling algorithm | `sample_type` | `uniform`, `weighted` | `dart` | -- |
| drop-out rate | `rate_drop` | $\mathbb{R} \in \left[0,1\right)$ | `dart` | -- |

## Random forest (RFSRC)

Random forests (explain) [@Breiman2001-ao]. The random forest learner is
implemented through the `randomForestSRC` package, which provides a unified
interface for different types of forests [@Ishwaran2008-hz; @Ishwaran2011-gu].

Hyperparameters for random forest learners are shown in the table below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size | `sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| number of trees | `n_tree` | $\mathbb{Z} \in \left[0,\infty\right)$ | yes |  This parameter is expressed on the $\log_{2}$ scale, i.e. the actual input value will be $2^\texttt{n_tree}$ [@Oshiro2012-mq]. The default range is $\left[4, 10\right]$.|
| subsampling fraction | `sample_size` | $\mathbb{R} \in \left(0, 1.0\right]$ | yes | Fraction of available data that is used for to create a single tree. The default range is $\left[2 / m, 1.0\right]$, with $m$ the number of samples. |
| number of features at each node | `m_try` | $\mathbb{R} \in \left[0.0, 1.0\right]$ | yes | Familiar ensures that there is always at least one candidate feature. |
| node size | `node_size` | $\mathbb{Z} \in \left[1, \infty\right)$ | yes | Minimum number of unique samples in terminal nodes. The default range is $\left[5, \lfloor m / 3\rfloor\right]$, with $m$ the number of samples.|
| maximum tree depth | `tree_depth` | $\mathbb{Z} \in \left[1,\infty\right)$ | yes | Maximum depth to which trees are allowed to grow. The default range is $\left[1, 10\right]$. |
| number of split points | `n_split` | $\mathbb{Z} \in \left[0, \infty\right)$ | no | By default, splitting is deterministic and has one split point ($0$).|
| splitting rule | `split_rule` | `gini`, `auc`, `entropy`, `mse`, `quantile.regr`, `la.quantile.regr`, `logrank`, `logrankscore`, `bs.gradient` | no | Default splitting rules are `gini` for `binomial` and `multinonial` outcomes, `mse` for `continuous` outcomes and `logrank` for `survival` outcomes.|
| sample weighting | `sample_weighting` | `inverse_number_of_samples`, `effective_number_of_samples`, `none` | no | Sample weighting allows for mitigating class imbalances. The default is `inverse_number_of_samples` for `binomial` and `multinomial` outcomes, and `none` otherwise. Instances with the majority class receive less weight. |
| beta | `sample_weighting_beta` | $\mathbb{Z} \in \left[-6,-1\right]$ | `effective_number_of_samples` only | Specifies the $\beta$ parameter for effective number of samples weighting [@Cui2019-ys]. It is expressed on the $\log_{10}$ scale: $\beta=1-10^\texttt{beta}$. |

Note that optimising that optimising the number of trees can be slow, as big
forests take longer to construct and perform more computations for predictions.
Hence hyperparameter optimisation may be sped up by limiting the range of the
`n_tree` parameter, or setting a single value.

## Random forest (ranger)

The second implementation of random forests comes from the `ranger` package
[@Wright2017-wc]. It is generally faster than the implementation in
`randomForestSRC` and allows for different splitting rules, such as maximally
selected rank statistics [@Lausen1992-qh; @Wright2017-sj] and concordance
index-based splitting [@Schmid2016-ie]. Hyperparameters for the random forest
are shown in the table below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size | `sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| number of trees | `n_tree` | $\mathbb{Z} \in \left[0,\infty\right)$ | yes |  This parameter is expressed on the $\log_{2}$ scale, i.e. the actual input value will be $2^\texttt{n_tree}$ [@Oshiro2012-mq]. The default range is $\left[4, 10\right]$.|
| subsampling fraction | `sample_size` | $\mathbb{R} \in \left(0, 1.0\right]$ | yes | Fraction of available data that is used for to create a single tree. The default range is $\left[2 / m, 1.0\right]$, with $m$ the number of samples. |
| number of features at each node | `m_try` | $\mathbb{R} \in \left[0.0, 1.0\right]$ | yes | Familiar ensures that there is always at least one candidate feature. |
| node size | `node_size` | $\mathbb{Z} \in \left[1, \infty\right)$ | yes | Minimum number of unique samples in terminal nodes. The default range is $\left[5, \lfloor m / 3\rfloor\right]$, with $m$ the number of samples.|
| maximum tree depth | `tree_depth` | $\mathbb{Z} \in \left[1,\infty\right)$ | yes | Maximum depth to which trees are allowed to grow. The default range is $\left[1, 10\right]$. |
| splitting rule | `split_rule` | `gini`, `extratrees`, `variance`, `logrank`, `C`, `maxstat`| no | Default splitting rules are `gini` for `binomial` and `multinomial` outcomes and `maxstat` for `continuous` and `survival` outcomes.|
| significance split threshold | `alpha` | $\mathbb{R} \in \left(0.0, 1.0\right]$ | `maxstat` | Minimum significance level for further splitting. The default range is $\left[10^{-6}, 1.0\right]$ |
| sample weighting | `sample_weighting` | `inverse_number_of_samples`, `effective_number_of_samples`, `none` | no | Sample weighting allows for mitigating class imbalances. The default is `inverse_number_of_samples` for `binomial` and `multinomial` outcomes, and `none` otherwise. Instances with the majority class receive less weight. |
| beta | `sample_weighting_beta` | $\mathbb{Z} \in \left[-6,-1\right]$ | `effective_number_of_samples` only | Specifies the $\beta$ parameter for effective number of samples weighting [@Cui2019-ys]. It is expressed on the $\log_{10}$ scale: $\beta=1-10^\texttt{beta}$. |

Note that optimising the number of trees can be slow, as big forests take longer
to construct and perform more computations for predictions. Hence hyperparameter
optimisation may be sped up by limiting the range of the `n_tree` parameter or
setting a single value.

## Naive Bayes

The naive Bayes classifier uses Bayes rule to predict posterior probabilities.
The naive Bayes classifier uses the `e1071::naiveBayes` function
[@Meyer2021-aq]. The  hyperparameters of the classifier are shown in the table
below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size |`sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| laplace smoothing | `laplace` | $\mathbb{R} \in \left[0.0, \infty \right)$ | yes | The default range is $\left[0, 10\right]$. |

## *k*-nearest neighbours

*k*-nearest neighbours is a simple clustering algorithm that classifies samples
based on the classes of their *k* nearest neighbours in feature space. The
`e1071::gknn` function is used to implement *k*-nearest neighbours
[@Meyer2021-aq]. The hyperparameters are shown in the table below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size |`sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| number of nearest neighbours | `k` | $\mathbb{Z} \in \left[1,m\right]$ | yes | The default range is $\left[1, \lceil 2 m^{1/3} \rceil \right]$, with $m$ the number of samples. |
| distance metric | `distance_metric` | all metrics supported by `proxy::dist` | yes | The default set of metrics is `gower`, `euclidean` and `manhattan`.|

## Support vector machines

Support vector machines were originally defined to find optimal margins
between classes for classification problems. Support vector machines were
popularized after the use of the kernel trick was described. Using the kernel
trick the calculations are performed in an implicit high-dimensional feature
space, which is considerably more efficient than explicit calculations
[@Boser1992-nk].

Familiar implements SVM using `e1071::svm`. By default, both
features and outcome data are scaled internally. This was used to derive the
default hyperparameter ranges specified in the table below.

| **parameter** | **tag** | **values** | **optimised** | **comments** |
|:--------------|:--------|:----------:|:----------:|:-------------|
| signature size | `sign_size` | $\mathbb{Z} \in \left[1,n\right]$ | yes | -- |
| SVM kernel | `kernel` | `linear`, `polynomial`, `radial`, `sigmoid` | no | Default is `radial`, unless specified as part of the learner's name. |
| $C$ | `c` | $\mathbb{R}$ | yes | The cost for violating constraints is expressed on a $\log_{10}$ scale, i.e. the actual value is $10^\texttt{c}$. The default range is $\left[-5, 3\right]$. |
| $\epsilon$ | `epsilon` | $\mathbb{R}$ | `continuous` outcomes | The error tolerance $\epsilon$ is only used for regression SVM. $\epsilon$ is expressed on a $\log_{10}$ scale. The default range is $\left[-5, 1\right]$. |
| $\nu$ | `nu` | $\mathbb{R}$ | `nu` SVM type | `nu` provides upper bounds for the fraction of training errors and lower bounds for the fraction of support vectors [@Chang2011-ck]. It is expressed on a $\log_{10}$ scale. The default range is $\left[-5, 1\right]$. |
| inverse kernel width | `gamma` | $\mathbb{R}$ | non-linear kernels | This parameter specifies the inverse of the kernel width. It is expressed on the $\log_{10}$ scale. The default range is $\left[-9, 3\right]$. |
| polynomial degree | `degree` | $\mathbb{Z} \in \left[1, \infty\right)$ | polynomial kernel | The default range is $\left[1, 5\right]$. |
| kernel offset | `offset` | $\mathbb{R} \in \left[0, \infty\right)$ | polynomial and sigmoid kernels |  Negative values are not allowed. The default range is $\left[0, 1 \right]$. |

Several types of SVM algorithms exist for classification and regression. The
following SVM types are implemented:

* $C$-classification: `c`
* $\nu$-classification: `nu`
* $\nu$-regression: `nu`
* $\epsilon$-regression: `eps`

The following kernels are implemented:

* Linear kernel: `linear`
* Polynomial kernel: `polynomial`
* Radial basis function kernel: `radial`
* Hyperbolic tangent (sigmoid) kernel: `sigmoid`


# Hyperparameter optimization

Hyperparameter optimisation is conducted to select model parameters lead to
good, generalisable models. The main hyperparameter optimisation framework used
by familiar is based on sequential model-based optimisation (SMBO)
[@Hutter2011-ea], but with significant updates and extensions.

Overall, the following steps are conducted to optimise hyperparameters for a
given learner and dataset:

- A set of bootstraps of the development dataset is made. This allows for
training the models using the in-bag data and assessing performance using
out-of-bag data. This is independent of any bootstraps that may have been
defined as part of the experimental design. In case a subsampling method is
specified in the experimental design, the corresponding internal development
dataset is bootstrapped.

- Variable importance is determined for each of the in-bag datasets to avoid
positively biasing the optimisation process through information leakage.

- An initial exploration of the hyperparameter space is conducted. A model is
trained using each hyperparameter set and a bootstrap subsample. It is then
assessed by computing an optimisation score based on model performance in the
out-of-bag, and optionally, in-bag data.

- Exploration of the hyperparameter space then continues by iteratively
comparing the best known hyperparameter set against challenger hyperparameter
sets in so-called intensify steps. The challengers are selected by predicting
the expected model performance for a new hyperparameter set and computing its
utility [@Shahriari2016-kx]. In addition, we predict model run times, and
actively prune potential slow challengers. Utility is computed using an
acquisition function, of which several are available in familiar. The set of
challenger hyperparameter sets may moreover be iteratively pruned, dependent on
the exploration method.

- Exploration stops if no significant improvement to model performance could be
established by further exploring the hyperparameter space, the total number of
iterations has been reached, or bootstraps samples have been exhausted.

- If the best set of hyperparameters does not lead to a model that exceeds
performance of a naive model, e.g. one that always predicts the majority class
or median value, a naive model is trained instead.

## Predicting run time of model

Models take a certain time to train. Familiar is actively measuring this time
during hyperparameter optimisation for two reasons: first, to optimise
assignment of jobs to parallel nodes; and secondly, to prune potential challenger
sets that produce models that tend to run longer than the best-known model. The
strictness increases with increased exploration.

Let $m$ be the number of bootstraps used to quantify the most visited
hyperparameter sets, and $M$ the total number of bootstrap samples that can be
visited. Then let $t_{\texttt{opt}}$ be the runtime of the best known model. The
maximum time that a challenger hyperparameter set is allowed for training is
then empirically set to:

$$t_{\texttt{max}} = \left(5 - 4 \frac{m}{M} \right) t_{\texttt{opt}} $$

Hence, $t_{\texttt{max}}$ converges to $t_{\texttt{opt}}$ for $m \rightarrow M$.

If maximum runtime is relatively insignificant, i.e. $t_{\texttt{max}} < 10.0$
seconds, a threshold of $t_{\texttt{max}}= 10.0$ is used.

If the maximum runtime is not known, i.e. none of the hyperparameter sets
evaluated so far produced a valid model, the maximum time threshold is set to
infinite. In effect, no hyperparameter sets are pruned based on expected run
time.

A random forest is trained to infer runtimes for (new) hyperparameter sets,
based on the runtimes observed for visited hyperparameter sets. The random
forest subsequently infers runtime for a challenger hyperparameter set. The
runtime estimate is compared against $t_{\texttt{max}}$, and if it exceeds the
threshold, it is rejected and not evaluated.

## Assessing goodness of hyperparameter sets

A summary score $s$ determines how good a set of hyperparameters $\mathbf{x}$
is. This score is computed in two steps. First an objective score is computed to
assess model performance for a hyperparameter set in a bootstrap subsample.
Subsequently an optimisation score is computed within the subsample. Finally a
summary score is determined from the optimisation score over all subsamples.

An objective score $s''$ is computed from the performance metric value for a
specific hyperparameter set for in-bag (IB; $s''_{\textrm{IB}}$) and out-of-bag
data (OOB; $s''_{\textrm{OOB}}$). The objective score always lies in the
interval $[-1.0, 1.0]$. An objective score of $1.0$ always indicates the best
possible score. A score of $0.0$ indicates that the hyperparameter set leads to
the same degree of performance as the best educated guess, i.e. the majority
class (for binomial and multinomial outcomes), the median value (for continuous
outcomes), or tied risk or survival times (for survival outcomes). Objective
scores that are either missing or below $-1.0$ are truncated to the value of
$-1.0$. In case multiple metrics are used to assess model performance, the mean
of the respective objective scores is used.

The optimisation scores $s'$ and summary score $s$ of each hyperparameter set
are subsequently computed using one of the following functions, which can be set
using the `optimisation_function` parameter:

1. `validation`, `max_validation` (default): The optimisation score for each
bootstrap subsample is $s'=s''_{\textrm{OOB}}$. The summary score $s$ is then
the average of optimisation scores $s'$. This is a commonly used criterion that
tries to maximise the score on the OOB validation data.

2. `balanced`:  The optimisation score for each bootstrap subsample is
$s'=s''_{\textrm{OOB}} - \left| s''_{\textrm{OOB}} - s''_{\textrm{IB}} \right|$.
The summary score $s$ is then the average of optimisation scores $s'$. A
variation on `max_validation` with a penalty for differences in performance
between the IB and OOB data. The underlying idea is that a good set of
hyperparameters should lead to models that perform well on both development and
validation data.

3. `strong_balance`: The optimisation score for each bootstrap subsample is
$s'=s''_{\textrm{OOB}} - 2\left| s''_{\textrm{OOB}}-s''_{\textrm{IB}} \right|$.
The summary score $s$ is then the average of optimisation scores $s'$. A variant
of `balanced` with a stronger penalty term.

4. `validation_minus_sd`: The optimisation score for each bootstrap subsample is
$s'=s''_{\textrm{OOB}}$. The summary score $s$ is then the average of
optimisation scores $s'$ minus the standard deviation of $s'$. This penalises
hyperparameter sets that lead to wide variance on OOB data.

5. `validation_25th_percentile`: The optimisation score for each bootstrap
subsample is $s'=s''_{\textrm{OOB}}$. The summary score $s$ is the 25th
percentile of optimisation scores $s'$. Like `validation_minus_sd` this
penalises hyperparameter sets that lead to wide variance on OOB data.

6. `model_estimate`: The optimisation score for each bootstrap subsample is
$s'=s''_{\textrm{OOB}}$. The model used to infer utility of new hyperparameter
sets (see the next section [Predicting optimisation score for new hyperparameter
sets]) is used to predict the expected optimisation score $\mu(\mathbf{x})$.
This score is used as summary score $s$. Not available for random search.

7. `model_estimate_minus_sd`: The optimisation score for each bootstrap
subsample is $s'=s''_{\textrm{OOB}}$. The model used to infer utility of new
hyperparameter sets (see the next section [Predicting optimisation score for new
hyperparameter sets]) is used to predict the expected optimisation score
$\mu(\mathbf{x})$ and its standard deviation $\sigma(\mathbf{x})$. Then
$s = \mu(\mathbf{x}) - \sigma(\mathbf{x})$. This penalises hyperparameter sets
for which a large variance is expected. Not available for random search.

8. `model_balanced_estimate`: The optimisation score for each bootstrap
subsample is $s'=s''_{\textrm{OOB}} - \left| s''_{\textrm{OOB}} - s''_{\textrm{IB}} \right|$.
The model used to infer utility of new hyperparameter sets (see the next section
[Predicting optimisation score for new hyperparameter sets]) is used to predict
the expected optimisation score $\mu(\mathbf{x})$. This score is used as summary
score $s$. The method is similar to `model_estimate`, but predicts the balanced
score instead of the validation score. Not available for random search.

9. `model_balanced_estimate_minus_sd`: The optimisation score for each bootstrap subsample is
$s'=s''_{\textrm{OOB}} - \left| s''_{\textrm{OOB}} - s''_{\textrm{IB}} \right|$. 
The model used to infer utility of new hyperparameter sets (see the next section
[Predicting optimisation score for new hyperparameter sets]) is used to predict
the expected optimisation score $\mu(\mathbf{x})$ and its standard deviation
$\sigma(\mathbf{x})$. Then $s = \mu(\mathbf{x}) - \sigma(\mathbf{x})$. This
penalises hyperparameter sets for which a large variance is expected. The method
is similar to `model_estimate_minus_sd`, but predicts the balanced score and its
variance instead. Not available for random search.

The summary score is then used to select the best known hyperparameter set, i.e.
the set that maximises the summary score. Moreover, the optimisation scores are
used to identify candidate hyperparameter sets as described in the following
section. The optimisation function is set using the `optimisation_function`
argument.

## Predicting optimisation score for new hyperparameter sets

Any point in the hyperparameter space has a single, scalar, optimisation
score value that is *a priori* unknown. During the optimisation process,
the algorithm samples from the hyperparameter space by selecting
hyperparameter sets and computing the optimisation score for one or
more bootstraps. For each hyperparameter set the resulting scores are
distributed around the actual value.

A key point of Bayesian optimisation is the ability to estimate the usefulness
or utility of new hyperparameter sets. This ability is facilitated by modelling
the optimisation score of new hyperparameter sets using the optimisation scores
of observed hyperparameter sets.

The following models can be used for this purpose:

- `gaussian_process` (default): Creates a localised approximate Gaussian
deterministic Gaussian Process. This is implemented using the `laGP` package
[@Gramacy2016-aa].

- `bayesian_additive_regression_trees` or `bart`: Uses Bayesian Additive
Regression Trees for inference. Unlike standard random forests, BART allows for
estimating posterior distributions directly and can extrapolate. BART is
implemented using the `BART` package [@Sparapani2021-aa].

- `random_forest`: Creates a random forest for inference. A random forest was
originally used in the SMBO algorithm by Hutter et al. [@Hutter2011-ea]. A
weakness of random forests is their lack of extrapolation beyond observed
values, which limits their usefulness in exploiting promising areas of
hyperparameter space somewhat.

In addition, familiar can perform random search (`random` or `random_search`).
This forgoes the use of models to steer optimisation. Instead, the
hyperparameter space is sampled at random.

The learner used to predict optimisation scores can be specified using the
`hyperparameter_learner` parameter.

## Acquisition functions for utility of hyperparameter sets

The expected values and the posterior values of optimisation scores from the
models are used to compute utility of new hyperparameter sets using an
acquisition function. The following acquisition functions are available in
familiar [@Shahriari2016-kx]. Let $\nu=f(\mathbf{x})$ be the posterior
distribution of hyperparameter set $\mathbf{x}$, and $\tau$ the best observed
optimisation score. Let also $\mu(\mathbf{x})$ and $\sigma(\mathbf{x})$ be the
sample mean and sample standard deviation for set $\mathbf{x}$ (implicitly for
round $m$):

* `improvement_probability`: The probability of improvement quantifies the
  probability that the expected optimisation score for a set $\mathbf{x}$ is
  better than the best observed optimisation score $\tau$:

  $$\alpha(\mathbf{x}) = P(\nu > \tau) = \Phi \left(\frac{\mu(\mathbf{x}) - \tau} {\sigma(\mathbf{x})}\right)$$

  Here $\Phi$ is the cumulative distribution function of a normal distribution.
  Note that this acquisition function is typically prone to convergence to local
  minima, although in our algorithm this may be mitigated by the alternating
  random search strategy.

* `improvement_empirical_probability`: Similar to `improvement_probability`, but based
  directly on optimisation scores predicted by the individual decision trees:

  $$\alpha(\mathbf{x}) = P(\nu > \tau) = \frac{1}{N} \sum_{i=1}^N \left[\nu_i > \tau\right]$$

  with $\left[\ldots\right]$ denoting an Iverson bracket, which takes the value
  1 if the condition specified by $\ldots$ is true, and 0 otherwise.

* `expected_improvement` (default): Expected improvement is based on an improvement
  function, and is be computed as follows:

  $$\alpha(\mathbf{x})=(\mu(\mathbf{x})-\tau) \Phi \left(\frac{\mu(\mathbf{x}) - \tau}{\sigma(\mathbf{x})}\right) + \sigma(\mathbf{x}) \phi\left(\frac{\mu(\mathbf{x}) - \tau}{\sigma(\mathbf{x})}\right)$$

  with $\phi$ denoting the normal probability density function. If $\sigma(\mathbf{x}) = 0$,
  $\alpha(\mathbf{x})=0$.

* `upper_confidence_bound`: This acquisition function is based on the upper
  confidence bound of the distribution, and is defined as [@Srinivas2012-hq]:

  $$\alpha(\mathbf{x}, t)=\mu(\mathbf{x}) + \beta_t \sigma(\mathbf{x})$$

  Here the parameter $\beta_t$ depends on the current round $t$, which in our
  implementation would be roughly equivalent to the maximum number of bootstraps
  performed for any of the hyperparameter sets, minus $1$. The $\beta_t$ parameter
  is then defined as $\beta_t= 2 \log(|D|(t+1)^2\pi^2/6\delta)$
  [@Srinivas2012-hq]. Here, $\delta \in [0,1]$ and $|D|$ the dimensionality of the
  hyperparameter space. For their experiments Srinivas et al. used $\delta=0.1$,
  and noted that scaling down $\beta_t$ by a factor $5$ improves the algorithm.
  Hence, we use $\beta_t= \frac{2}{5} \log\left(\frac{5}{3} |D|(t+1)^2\pi^2\right)$.

* `bayes_upper_confidence_bound`: Proposed by Kaufmann et al. [@Kaufmann2012-kh],
  the Bayesian upper confidence bound is defined as:

  $$\alpha(\mathbf{x}, t)=Q \left(1-\frac{1}{(t + 1) \log(n)^c}, \nu_t\right)$$

  with $Q$ being a quantile function. Note that we here deviate from the original
  definition by substituting the original $t$ by $t+1$, and not constraining the
  posterior distribution $\nu$ to a known function. In the original, $n$ defines
  the number of rounds, with $t$ the current round. In our algorithm, this is
  interpreted as $n$ being the total number of bootstraps, and $t$ the maximum
  number of bootstraps already performed for any of the hyperparameter sets.
  Hence our implementation is the same, aside from the lack of constraint on the
  posterior distribution. Based on the simulations and recommendations by
  Kaufmann et al. we choose parameter $c=0$.

The acquisition function can be specified using the `acquisition_function`
parameter.

## Exploring challenger sets

After selecting challenger hyperparameter sets, the optimisation score of models
trained using these hyperparameter sets are compared against the best-known
score in a run-off intensify step. During this step, models will be trained on
new bootstrap data, compared, and then trained on another set of bootstraps, and
so on. Pruning the set of challenger hyperparameters after each iteration
reduces computational load. Familiar implements the following methods:

- `single_shot` (default): The set of alternative parameter sets is not pruned,
and each intensify iteration contains only a single (initial) intensification
step that only uses a single bootstrap. This is the exploration method that
under default settings does the least in-depth exploration, with the advantage
of being fast. Extensive internal tests show that this exploration method yields
similarly performant models as more extensive exploration methods.

- `successive_halving`: The set of alternative parameter sets is pruned by
removing the worst performing half of the sets after each step
[@Jamieson2016-fq]. The set of investigated parameter sets gets progressively
smaller.

- `stochastic_reject`: The set of alternative parameter sets is pruned by
comparing the performance of each parameter set with that of the incumbent best
parameter set using a paired Wilcoxon test.

- `none`: The set of alternative parameter sets is not pruned. Compared to
`single_shot`, hyperparameter sets are assessed using multiple bootstraps
(`smbo_step_bootstraps` * `smbo_intensify_steps`).

The method used to steer exploration can be set using the `exploration_method`
parameter.

## Providing hyperparameters manually

It is possible to set hyperparameters manually. This can be used to change
hyperparameters that are fixed by default, to set a fixed value for randomised
hyperparameters, or to provide a different search range for randomised
hyperparameters.

Hyperparameters can be provided using the `hyperparameter` tag. For the
`glm_logistic` learner an example tag may for example look as follows:
```
<hyperparameter>
  <glm_logistic>
    <sign_size>5</sign_size>
  </glm_logistic>
</hyperparameter>
```

Or as a nested list passed as the `hyperparameter` argument to `summon_familiar`:
```
hyperparameter = list("glm_logistic"=list("sign_size"=5))
```

More than one value can be provided. The behaviour changes depending on whether
the hyperparameter is categorical or numeric variable. In case of categorical
hyperparameters, the provided values define the search range. For numerical
hyperparameters, providing two values sets the bounds of the search range,
whereas providing more than two values will define the search range itself.

## Configuration options for hyperparameter optimisation

Hyperparameter optimization may be configured using the tags/arguments in the table below.

| **tag** / **argument** | **description** | **default** |
|:----------:|:-------------------------|:-----------:|
| `optimisation_bootstraps`  | Maximum number of bootstraps created for optimisation. | `20` |
| `optimisation_determine_vimp` | If `TRUE`, compute variable importance for each bootstrap. If `FALSE` use variable importance computed during the feature selection step. | `TRUE` |
| `smbo_random_initialisation` | If `random` initial parameter sets are generated randomly from default ranges. If `fixed` or `fixed_subsample`, the initial parameter sets are based on a grid in parameter space. | `fixed_subsample` |
| `smbo_n_random_sets` | Sets the number of hyperparameter sets drawn for initialisation. Ignored if `smbo_random_initialisation` is `fixed`. If `smbo_random_initialisation` is `fixed_subsample`, the number of selected hyperparameters may be lower. | `100` |
| `max_smbo_iterations` | Maximum number of intensify iterations of the SMBO algorithm. | `20` |
| `smbo_stop_convergent_iterations` | Number of subsequent convergent SMBO iterations required to stop hyperparameter optimisation early. | `3` |
| `smbo_stop_tolerance` | Tolerance for recent optimisation scores to determine convergence. | `0.1 * 1 / sqrt(n_samples)` |
| `smbo_time_limit` | Time limit (in minutes) for the optimisation process. | `NULL` |
| `smbo_initial_bootstraps` | Number of bootstraps assessed initially. | `1` |
| `smbo_step_bootstraps` | Number of bootstraps used within each step of an intensify iteration. | `3` |
| `smbo_intensify_steps` | Number of intensify steps within each intensify iteration. | `5` |
| `optimisation_metric` | The metric used for optimisation, e.g. `auc`. See the vignette on performance metrics for available options. More than one metric may be specified. | `auc_roc` (`binomial`, `multinomial`); `mse` (`continuous`); and `concordance_index` (`survival`) |
| `optimisation_function` | The optimisation function used (see [Assessing goodness of hyperparameter sets]). | `validation` |
| `hyperparameter_learner` | Learner used to predict optimisation scores for new hyperparameter sets (see [Predicting optimisation score for new hyperparameter sets]). | `gaussian_process` |
| `acquisition_function` | The function used to quantify utility of hyperparameter sets (see [Acquisition functions for utility of hyperparameter sets]). | `expected_improvement` |
| `exploration_method` | Method used to explore challenger hyperparameter sets (see [Exploring challenger sets]). | `single_shot` |
| `smbo_stochastic_reject_p_value` | The p-value level for stochastic pruning. | `0.05` |
| `parallel_hyperparameter_optimisation` | Enables parallel processing for hyperparameter optimisation. Ignored if `parallel = FALSE`. Can be `outer` to process multiple optimisation processes simultaneously. Overrides `parallel = TRUE`. | `TRUE` |

Table: Configuration options for hyperparameter optimisation.

# Model recalibration

Even though learners may be good at discrimination, model calibration can be
lacklustre. Some learners are therefore recalibrated as follows:

1.  3-fold cross-validation is performed. Responses are predicted using a new
model trained on each training fold.

2.  Three recalibration models are trained using the responses for the training
folds as input [@Niculescu-Mizil2005-kj].

3.  The three recalibration models are then applied to (new) responses of the
full model, and the resulting values averaged for each sample.

The following learners currently undergo recalibration:

| **learner** | **outcome** | **recalibration model** |
|:------------|:-----------:|:-------------|
|`xgboost_lm`, `xgboost_lm_logistic`| binomial, multinomial| `glm_logistic`|
|`xgboost_tree`, `xgboost_tree_logistic`| binomial, multinomial| `glm_logistic`|
|`xgboost_lm_cox`| survival | `glm_cox`|
|`xgboost_tree_cox`| survival | `glm_cox`|
|`boosted_glm_cindex`, `boosted_glm_gehan`| survival | `glm_cox`|
|`boosted_tree_cindex`, `boosted_tree_gehan`| survival | `glm_cox`|

Familiar currently does not recalibrate models by inverting the model
calibration curves, but may do so in the future.

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationSampleSimilarity.R
\name{extract_sample_similarity}
\alias{extract_sample_similarity}
\title{Internal function to extract the sample distance table.}
\usage{
extract_sample_similarity(
  object,
  data,
  cl = NULL,
  is_pre_processed = FALSE,
  sample_limit = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  sample_similarity_metric = waiver(),
  verbose = FALSE,
  message_indent = 0L,
  ...
)
}
\arguments{
\item{object}{A \code{familiarEnsemble}, which is an ensemble of one or more
\code{familiarModel} objects, or a \code{familiarDataElementPredictionTable} object
that contains prediction data.}

\item{data}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}

\item{cl}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}

\item{is_pre_processed}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}

\item{sample_limit}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}

\item{sample_cluster_method}{The method used to perform clustering based on
distance between samples. These are the same methods as for the
\code{cluster_method} configuration parameter: \code{hclust}, \code{agnes}, \code{diana} and
\code{pam}.

\code{none} cannot be used when extracting data for feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_similarity_metric}{Metric to determine pairwise similarity
between samples. Similarity is computed in the same manner as for
clustering, but \code{sample_similarity_metric} has different options that are
better suited to computing distance between samples instead of between
features: \code{gower}, \code{euclidean}.

The underlying feature data is scaled to the \eqn{[0, 1]} range (for
numerical features) using the feature values across the samples. The
normalisation parameters required can optionally be computed from feature
data with the outer 5\% (on both sides) of feature values trimmed or
winsorised. To do so append \verb{_trim} (trimming) or \verb{_winsor} (winsorising) to
the metric name. This reduces the effect of outliers somewhat.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{...}{Unused arguments.}
}
\value{
An object containing pairwise distance between samples. This data
is only the upper triangular of the complete matrix (i.e. the sparse
unitriangular representation). Diagonals will always be 0.0 and the lower
triangular is mirrored.
}
\description{
Computes and extracts the sample distance table for samples
analysed using a \code{familiarEnsemble} object to form a \code{familiarData} object.
This table can be used to cluster samples, and is exported directly by
\code{extract_feature_expression}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationRiskStratificationInfo.R
\name{export_risk_stratification_info}
\alias{export_risk_stratification_info}
\alias{export_risk_stratification_info,familiarCollection-method}
\alias{export_risk_stratification_info,ANY-method}
\title{Extract and export cut-off values for risk group stratification.}
\usage{
export_risk_stratification_info(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)

\S4method{export_risk_stratification_info}{familiarCollection}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)

\S4method{export_risk_stratification_info}{ANY}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{aggregate_results}{Flag that signifies whether results should be
aggregated for export.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A data.table (if \code{dir_path} is not provided), or nothing, as all data
is exported to \code{csv} files.
}
\description{
Extract and export cut-off values for risk group stratification
by models in a familiarCollection.
}
\details{
Data is usually collected from a \code{familiarCollection} object.
However, you can also provide one or more \code{familiarData} objects, that will
be internally converted to a \code{familiarCollection} object. It is also
possible to provide a \code{familiarEnsemble} or one or more \code{familiarModel}
objects together with the data from which data is computed prior to export.
Paths to the previous files can also be provided.

All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Stratification cut-off values are determined when creating a model, using
one of several methods set by the \code{stratification_method} parameter. These
values are then used to stratify samples in any new dataset. The available
methods are:
\itemize{
\item \code{median} (default): The median predicted value in the development cohort
is used to stratify the samples into two risk groups.
\item \code{fixed}: Samples are stratified based on the sample quantiles of the
predicted values. These quantiles are defined using the
\code{stratification_threshold} parameter.
\item \code{optimised}: Use maximally selected rank statistics to determine the
optimal threshold (Lausen and Schumacher, 1992; Hothorn et al., 2003) to
stratify samples into two optimally separated risk groups.
}
}
\references{
\enumerate{
\item Lausen, B. & Schumacher, M. Maximally Selected Rank Statistics.
Biometrics 48, 73 (1992).
\item Hothorn, T. & Lausen, B. On the exact distribution of maximally selected
rank statistics. Comput. Stat. Data Anal. 43, 121–137 (2003).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarCollection-class}
\alias{familiarCollection-class}
\title{Collection of familiar data.}
\description{
A familiarCollection object aggregates data from one or more familiarData
objects.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the collection.}

\item{\code{data_sets}}{Name of the individual underlying datasets.}

\item{\code{outcome_type}}{Outcome type for which the collection was created.}

\item{\code{outcome_info}}{Outcome information object, which contains information
concerning the outcome, such as class levels.}

\item{\code{fs_vimp}}{collected for variable importance methods.}

\item{\code{model_vimp}}{Variable importance data collected from model-specific
algorithms implemented by models created by familiar.}

\item{\code{permutation_vimp}}{Data collected for permutation variable importance.}

\item{\code{hyperparameters}}{Hyperparameters collected from created models.}

\item{\code{hyperparameter_data}}{Additional data concerning hyperparameters. This is
currently not used yet.}

\item{\code{required_features}}{The set of features required for complete
reproduction, i.e. with imputation.}

\item{\code{model_features}}{The set of features that are required for using the
model, but without imputation.}

\item{\code{learner}}{Learning algorithm(s) used for data in the collection.}

\item{\code{vimp_method}}{Variable importance method(s) used for data in the
collection.}

\item{\code{prediction_data}}{Model predictions for the data in the collection.}

\item{\code{confusion_matrix}}{Confusion matrix information for the data in the
collection.}

\item{\code{decision_curve_data}}{Decision curve analysis data for the data in the
collection.}

\item{\code{calibration_info}}{Calibration information, e.g. baseline survival in the
development cohort.}

\item{\code{calibration_data}}{Model calibration data collected from data in the
collection.}

\item{\code{model_performance}}{Collection of model performance data for data in the
collection.}

\item{\code{km_info}}{Information concerning risk-stratification cut-off values for
data in the collection.}

\item{\code{km_data}}{Kaplan-Meier survival data for data in the collection.}

\item{\code{auc_data}}{AUC-ROC and AUC-PR data for data in the collection.}

\item{\code{ice_data}}{Individual conditional expectation data for data in the
collection. Partial dependence data are computed on the fly from these
data.}

\item{\code{shap_data}}{SHAP values for features included in a model or ensemble of
models.}

\item{\code{univariate_analysis}}{Univariate analysis results of data in the
collection.}

\item{\code{feature_expressions}}{Feature expression values for data in the
collection.}

\item{\code{feature_similarity}}{Feature similarity information for data in the
collection.}

\item{\code{sample_similarity}}{Sample similarity information for data in the
collection.}

\item{\code{data_set_labels}}{Labels for the different datasets in the collection.
See \code{get_data_set_names} and \code{set_data_set_names}.}

\item{\code{learner_labels}}{Labels for the different learning algorithms used to
create the collection. See \code{get_learner_names} and \code{set_learner_names}.}

\item{\code{vimp_method_labels}}{Labels for the different variable importance methods
used to create the collection. See \code{get_vimp_method_names} and
\code{set_vimp_method_names}.}

\item{\code{feature_labels}}{Labels for the features in this collection. See
\code{get_feature_names} and \code{set_feature_names}.}

\item{\code{km_group_labels}}{Labels for the risk strata in this collection. See
\code{get_risk_group_names} and \code{set_risk_group_names}.}

\item{\code{class_labels}}{Labels of the response variable. See \code{get_class_names} and
\code{set_class_names}.}

\item{\code{project_id}}{Identifier of the project that generated this collection.}

\item{\code{familiar_version}}{Version of the familiar package.

familiarCollection objects collect data from one or more familiarData
objects. This objects are important, as all plotting and export functions
use it. The fact that one can supply familiarModel, familiarEnsemble and
familiarData objects as arguments for these methods, is because familiar
internally converts these into familiarCollection objects prior to
executing the method.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.impute_outcome_type}
\alias{.impute_outcome_type}
\title{Internal imputation function for the outcome type.}
\usage{
.impute_outcome_type(
  data,
  outcome_column,
  class_levels,
  censoring_indicator,
  event_indicator,
  competing_risk_indicator
)
}
\arguments{
\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{outcome_column}{Name of the outcome column in the data set.}

\item{class_levels}{User-provided class levels for the outcome.}

\item{censoring_indicator}{Name of censoring indicator.}

\item{event_indicator}{Name of event indicator.}

\item{competing_risk_indicator}{Name of competing risk indicator.}
}
\value{
The imputed outcome type.
}
\description{
This function allows for imputation of the most plausible outcome type.
This imputation is only done for trivial cases, where there is little doubt.
As a consequence \code{continuous} outcome type is never imputed.
}
\note{
It is highly recommended that the user provides the outcome type.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PredictS4Methods.R
\name{predict}
\alias{predict}
\alias{predict,familiarModel-method}
\alias{predict,familiarEnsemble-method}
\alias{predict,familiarNoveltyDetector-method}
\alias{predict,list-method}
\alias{predict,character-method}
\title{Model predictions for familiar models and model ensembles}
\usage{
predict(object, ...)

\S4method{predict}{familiarModel}(
  object,
  newdata,
  type = "default",
  time = NULL,
  dir_path = NULL,
  ensemble_method = "median",
  stratification_threshold = NULL,
  stratification_method = NULL,
  percentiles = NULL,
  .as_prediction_table = FALSE,
  ...
)

\S4method{predict}{familiarEnsemble}(
  object,
  newdata,
  type = "default",
  time = NULL,
  dir_path = NULL,
  ensemble_method = "median",
  stratification_threshold = NULL,
  stratification_method = NULL,
  percentiles = NULL,
  .as_prediction_table = FALSE,
  ...
)

\S4method{predict}{familiarNoveltyDetector}(
  object,
  newdata,
  type = "novelty",
  ensemble_method = "median",
  .as_prediction_table = FALSE,
  ...
)

\S4method{predict}{list}(
  object,
  newdata,
  type = "default",
  time = NULL,
  dir_path = NULL,
  ensemble_method = "median",
  stratification_threshold = NULL,
  stratification_method = NULL,
  percentiles = NULL,
  .as_prediction_table = FALSE,
  ...
)

\S4method{predict}{character}(
  object,
  newdata,
  type = "default",
  time = NULL,
  dir_path = NULL,
  ensemble_method = "median",
  stratification_threshold = NULL,
  stratification_method = NULL,
  percentiles = NULL,
  .as_prediction_table = FALSE,
  ...
)
}
\arguments{
\item{object}{A familiar model or ensemble of models that should be used for
prediction. This can also be a path to the ensemble model, one or more
paths to models, or a list of models.}

\item{...}{to be documented.}

\item{newdata}{Data to which the models are fitted. \code{familiar} performs
checks on the data to ensure that all features required for fitting the
model are present, and no additional levels are present in categorical
features. Unlike other \code{predict} methods, \code{newdata} cannot be missing in
\code{familiar}, as training data are not stored with the models.}

\item{type}{Type of prediction made. The following values are directly
supported:
\itemize{
\item \code{default}: Default prediction, i.e. value estimates for \code{continuous}
outcomes, predicted class probabilities and class for \code{binomial} and
\code{multinomial} and the model response for \code{survival} outcomes.
\item \code{survival_probability}: Predicts survival probabilities at the time
specified by \code{time}. Only applicable to \code{survival} outcomes. Some models
may not allow for predicting survival probabilities based on their
response.
\item \code{novelty}: Predicts novelty of each sample, which can be used for
out-of-distribution detection.
\item \code{risk_stratification}: Predicts the strata to which the data belongs. Only
for \code{survival} outcomes.
}

Other values for type are passed to the fitting method of the actual
underlying model. For example for generalised linear models (\code{glm}) \code{type}
can be \code{link}, \code{response} or \code{terms} as well. Some of these model-specific
prediction types may fail to return results if the model has been trimmed.}

\item{time}{Time at which the response (\code{default}) or survival probability
(\code{survival_probability}) should be predicted for \code{survival} outcomes. Some
models have a response that does not depend on \code{time}, e.g. \code{cox}, whereas
others do, e.g. \code{random_forest}.}

\item{dir_path}{Path to the folder containing the models. Ensemble objects
are stored with the models detached. In case the models were moved since
creation, \code{dir_path} can be used to specify the current folder.
Alternatively the \code{update_model_dir_path} method can be used to update the
path.}

\item{ensemble_method}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}

\item{stratification_threshold}{Threshold value(s) used for stratifying
instances into risk groups. If this parameter is specified,
\code{stratification_method} and any threshold values that come with the model
are ignored, and \code{stratification_threshold} is used instead.}

\item{stratification_method}{Selects the stratification method from which the
threshold values should be selected. If the model or ensemble of models
does not contain thresholds for the indicated method, an error is returned.
In addition this argument is ignored if a \code{stratification_threshold} is
set.}

\item{percentiles}{Currently unused.}

\item{.as_prediction_table}{Selects whether a data.table or prediction table
object should be returned.}
}
\value{
A \code{data.table} with predicted values.
}
\description{
Fits the model or ensemble of models to the data and shows the
result.
}
\details{
This method is used to predict values for instances specified by the
\code{newdata} using the model or ensemble of models specified by the \code{object}
argument.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarHyperparameterLearner-class}
\alias{familiarHyperparameterLearner-class}
\title{Hyperparameter learner.}
\description{
A familiarHyperparameterLearner object is a self-contained model that can be
applied to predict optimisation scores for a set of hyperparameters.
}
\details{
Hyperparameter learners are used to infer the optimisation score for
sets of hyperparameters. These are then used to either infer utility using
acquisition functions or to generate summary scores to identify the optimal
model.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the familiarHyperparameterLearner object.}

\item{\code{learner}}{Algorithm used to create the hyperparameter learner.}

\item{\code{target_learner}}{Algorithm for which the hyperparameters are being
learned.}

\item{\code{target_outcome_type}}{Outcome type of the learner for which
hyperparameters are being modeled. Used to determine the target
hyperparameters.}

\item{\code{optimisation_metric}}{One or metrics used to generate the optimisation
score.}

\item{\code{optimisation_function}}{Function used to generate the optimisation score.}

\item{\code{model}}{The actual model trained using the specific algorithm, e.g. a
isolation forest from the \code{isotree} package.}

\item{\code{target_hyperparameters}}{The names of the hyperparameters that are used
to train the hyperparameter learner.}

\item{\code{project_id}}{Identifier of the project that generated the
familiarHyperparameterLearner object.}

\item{\code{familiar_version}}{Version of the familiar package.}

\item{\code{package}}{Name of package(s) required to executed the hyperparameter
learner itself, e.g. \code{laGP}.}

\item{\code{package_version}}{Version of the packages mentioned in the \code{package}
attribute.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.check_input_identifier_column}
\alias{.check_input_identifier_column}
\title{Internal function for checking consistency of the identifier columns}
\usage{
.check_input_identifier_column(
  id_column,
  data,
  signature = NULL,
  exclude_features = NULL,
  include_features = NULL,
  other_id_column = NULL,
  outcome_column = NULL,
  col_type,
  check_stringency = "strict"
)
}
\arguments{
\item{id_column}{Character string indicating the currently inspected
identifier column.}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{signature}{(\emph{optional}) One or more names of feature columns that are
considered part of a specific signature. Features specified here will
always be used for modelling. Ranking of variable importances has no effect
for these features.}

\item{exclude_features}{(\emph{optional}) Feature columns that will be removed
from the data set. Cannot overlap with features in \code{signature},
\code{novelty_features} or \code{include_features}.}

\item{include_features}{(\emph{optional}) Feature columns that are specifically
included in the data set. By default all features are included. Cannot
overlap with \code{exclude_features}, but may overlap \code{signature}. Features in
\code{signature} and \code{novelty_features} are always included. If both
\code{exclude_features} and \code{include_features} are provided, \code{include_features}
takes precedence, provided that there is no overlap between the two.}

\item{other_id_column}{Character string indicating another identifier column.}

\item{outcome_column}{Character string indicating the outcome column(s).}

\item{col_type}{Character string indicating the type of column, i.e. \code{sample}
or \code{batch}.}

\item{check_stringency}{Specifies stringency of various checks. This is mostly:
\itemize{
\item \code{strict}: default value used for \code{summon_familiar}. Thoroughly checks
input data. Used internally for checking development data.
\item \code{external_warn}: value used for \code{extract_data} and related methods. Less
stringent checks, but will warn for possible issues. Used internally for
checking data for evaluation and explanation.
\item \code{external}: value used for external methods such as \code{predict}. Less
stringent checks, particularly for identifier and outcome columns, which may
be completely absent. Used internally for \code{predict}.
}}
}
\description{
This function checks whether an identifier column is consistent, i.e. appears
it exists, there is only one, and there is no overlap with any user-provided
feature columns, identifiers, or
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{get_data_set_names,familiarCollection-method}
\alias{get_data_set_names,familiarCollection-method}
\alias{get_data_set_names}
\title{Get current name of datasets}
\usage{
\S4method{get_data_set_names}{familiarCollection}(x)
}
\arguments{
\item{x}{A familiarCollection object.}
}
\value{
An ordered array of dataset name labels.
}
\description{
Datasets in familiarCollection objects can have custom names for
export and plotting. This function retrieves the currently assigned names.
}
\details{
Labels convert internal naming of data sets to the requested label
at export or when plotting. Labels can be changed using the
\code{set_data_set_names} method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the familiarCollection class.
\item \code{\link{set_data_set_names}} for updating the name of datasets and their ordering.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{get_class_names,familiarCollection-method}
\alias{get_class_names,familiarCollection-method}
\alias{get_class_names}
\title{Get outcome class labels}
\usage{
\S4method{get_class_names}{familiarCollection}(x)
}
\arguments{
\item{x}{A familiarCollection object.}
}
\value{
An ordered array of class labels.
}
\description{
Outcome classes in familiarCollection objects can have custom
names for export and plotting. This function retrieves the currently
assigned names.
}
\details{
Labels convert internal class names to the requested label at export
or when plotting. Labels can be changed using the \code{set_class_names}
method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the familiarCollection class.
\item \code{\link{set_class_names}} for updating the name and ordering of classes.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationUnivariateAnalysis.R
\name{export_univariate_analysis_data}
\alias{export_univariate_analysis_data}
\alias{export_univariate_analysis_data,familiarCollection-method}
\alias{export_univariate_analysis_data,ANY-method}
\title{Extract and export univariate analysis data of features.}
\usage{
export_univariate_analysis_data(
  object,
  dir_path = NULL,
  p_adjustment_method = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{export_univariate_analysis_data}{familiarCollection}(
  object,
  dir_path = NULL,
  p_adjustment_method = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{export_univariate_analysis_data}{ANY}(
  object,
  dir_path = NULL,
  p_adjustment_method = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{p_adjustment_method}{(\emph{optional}) Indicates type of p-value that is
shown. One of \code{holm}, \code{hochberg}, \code{hommel}, \code{bonferroni}, \code{BH}, \code{BY},
\code{fdr}, \code{none}, \code{p_value} or \code{q_value} for adjusted p-values, uncorrected
p-values}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=extract_univariate_analysis]{extract_univariate_analysis}}, \code{\link[=as_familiar_collection]{as_familiar_collection}}
  \describe{
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{feature_cluster_method}}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_linkage_method}}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_cluster_cut_method}}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_threshold}}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_metric}}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{icc_type}}{String indicating the type of intraclass correlation
coefficient (\code{1}, \code{2} or \code{3}) that should be used to compute robustness for
features in repeated measurements during the evaluation of univariate
importance. These types correspond to the types in Shrout and Fleiss (1979).
If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A data.table (if \code{dir_path} is not provided), or nothing, as
all data is exported to \code{csv} files.
}
\description{
Extract and export univariate analysis data of features for data
in a familiarCollection.
}
\details{
Data is usually collected from a \code{familiarCollection} object.
However, you can also provide one or more \code{familiarData} objects, that will
be internally converted to a \code{familiarCollection} object. It is also
possible to provide a \code{familiarEnsemble} or one or more \code{familiarModel}
objects together with the data from which data is computed prior to export.
Paths to the previous files can also be provided.

All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Univariate analysis includes the computation of p-values, as well as
robustness (in case of repeated measurements). p-values are derived from
Wald's test.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{delayedDataObject-class}
\alias{delayedDataObject-class}
\title{Data object with delayed loading}
\description{
The delayed loading object provides an interface to the backend data. This
data object is typically used within the evaluation pipeline to load data
when needed.
}
\section{Slots}{

\describe{
\item{\code{data}}{NULL or data table containing the data. If present (not \code{NULL}),
data is considered loaded. This should not happen -- load_data_object auto-
matically creates a dataObject from the delayedDataObject.}

\item{\code{preprocessing_level}}{character indicating the level of pre-processing
already conducted. \code{"none"} by default.}

\item{\code{outcome_type}}{character, determines the outcome type.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters. Optional.}

\item{\code{data_column_info}}{Object containing column information.}

\item{\code{data_id}}{integer. Defines the data_id of the dataset that should be
loaded.}

\item{\code{run_id}}{integer. Defines the run_id of the dataset that should be load.
Together with data_id, run_id and validation allows for looking up the
sample set. If run_id is left unset (NA_integer_), this will force the
run_id to be set using the model, vimp_method or ensemble object. This is
used during the evaluation process to load data specifically related to
training, internal validation and external validation. The run-tables
(which contain information about data partitioning) associated with these
objects are used to look-up the run_id based on the data_id (that is always
explicitly set). The perform_task method for familiarTaskEvaluate uses this
aspect explicitly.}

\item{\code{validation}}{logical. This determines which internal data set will be
loaded. If TRUE, the validation data will be loaded, whereas FALSE loads
the development data.}

\item{\code{aggregate_on_load}}{logical. Determines whether data is aggregated after
loading.}

\item{\code{sample_set_on_load}}{NULL or vector of sample identifiers to be loaded.
Overrides any \code{sample_seed} that may have been provided.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarEnsemble.R
\name{update_model_dir_path}
\alias{update_model_dir_path}
\alias{update_model_dir_path,familiarEnsemble-method}
\alias{update_model_dir_path,ANY-method}
\title{Updates model directory path for ensemble objects.}
\usage{
update_model_dir_path(object, dir_path, ...)

\S4method{update_model_dir_path}{familiarEnsemble}(object, dir_path)

\S4method{update_model_dir_path}{ANY}(object, dir_path)
}
\arguments{
\item{object}{A \code{familiarEnsemble} object, or one or more \code{familiarModel}
objects that will be internally converted to a \code{familiarEnsemble} object.
Paths to such objects can also be provided.}

\item{dir_path}{Path to the directory where models are stored.}

\item{...}{Unused arguments.}
}
\value{
A \code{familiarEnsemble} object.
}
\description{
Updates the model directory path of a \code{familiarEnsemble} object.
}
\details{
Ensemble models created by familiar are often written to a directory
on a local drive or network. In such cases, the actual models are detached,
and paths to the models are stored instead. When the models are moved from
their original location, they can no longer be found and attached to the
ensemble. This method allows for pointing to the new directory containing
the models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotUnivariateImportance.R
\name{.plot_univariate_importance}
\alias{.plot_univariate_importance}
\title{Internal plotting function for univariate plots}
\usage{
.plot_univariate_importance(
  x,
  color_by,
  facet_by,
  facet_wrap_cols,
  ggtheme,
  show_cluster,
  discrete_palette,
  gradient_palette,
  x_label,
  y_label,
  legend_label,
  plot_title,
  plot_sub_title,
  caption,
  limit_n_features,
  x_range,
  x_breaks,
  significance_level_shown
)
}
\arguments{
\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{show_cluster}{(\emph{optional}) Show which features were clustered together.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring the plot elements
according to the groupings indicated by the \code{color_by} argument (if any).
\code{familiar} has a default palette. Other palettes are supported by
\code{paletteer}, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{gradient_palette}{(\emph{optional}) Palette to use for filling the bars in
case the \code{color_by} argument is not set. The bars are then coloured
according to their importance. By default, no gradient is used, and the
bars are not filled according to importance. Use \code{NULL} to fill the bars
using the default palette in \code{familiar}. Other palettes are supported by
the \code{paletteer} package, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{limit_n_features}{(\emph{optional}) The number of features that should be
included in the plot. Only the most important features are shown. By
default, the number of features is not limited.}

\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{significance_level_shown}{Position(s) to draw vertical lines indicating
a significance level, e.g. 0.05. Can be NULL to not draw anything.}
}
\value{
ggplot plot object.
}
\description{
Internal plotting function for univariate plots
}
\seealso{
\itemize{
\item \code{\link{plot_univariate_importance}} for the user interface.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationPermutationVimp.R
\name{export_permutation_vimp}
\alias{export_permutation_vimp}
\alias{export_permutation_vimp,familiarCollection-method}
\alias{export_permutation_vimp,ANY-method}
\title{Extract and export permutation variable importance.}
\usage{
export_permutation_vimp(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)

\S4method{export_permutation_vimp}{familiarCollection}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)

\S4method{export_permutation_vimp}{ANY}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{aggregate_results}{Flag that signifies whether results should be
aggregated for export.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=extract_permutation_vimp]{extract_permutation_vimp}}, \code{\link[=as_familiar_collection]{as_familiar_collection}}
  \describe{
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{evaluation_times}}{One or more time points that are used for in analysis
of survival problems when data has to be assessed at a set time, e.g.
calibration. If not provided explicitly, this parameter is read from
settings used at creation of the underlying \code{familiarModel} objects. Only
used for \code{survival} outcomes.}
    \item{\code{ensemble_method}}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}
    \item{\code{metric}}{One or more metrics for assessing model performance. See the
vignette on performance metrics for the available metrics. If not provided
explicitly, this parameter is read from settings used at creation of the
underlying \code{familiarModel} objects.}
    \item{\code{feature_cluster_method}}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_linkage_method}}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_cluster_cut_method}}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_threshold}}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_metric}}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{sample_limit}}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}
    \item{\code{n_important_features}}{(\emph{optional}) Set the number of features that are
evaluated in evaluation steps. Cannot be 0 or fewer.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("ice_data"=10, "permutation_vimp"=5)}.

This parameter can be set for the following data elements:
\code{ice_data}, \code{permutation_vimp}, and \code{shap}.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
    \item{\code{estimation_type}}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}
    \item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}
    \item{\code{bootstrap_ci_method}}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A data.table (if \code{dir_path} is not provided), or nothing, as all data
is exported to \code{csv} files.
}
\description{
Extract and export model-based variable importance from a
familiarCollection.
}
\details{
Data, such as permutation variable importance and calibration
information, is usually collected from a \code{familiarCollection} object.
However, you can also provide one or more \code{familiarData} objects, that will
be internally converted to a \code{familiarCollection} object. It is also
possible to provide a \code{familiarEnsemble} or one or more \code{familiarModel}
objects together with the data from which data is computed prior to export.
Paths to the previously mentioned files can also be provided.

All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Permutation Variable importance assesses the improvement in model
performance due to a feature. For this purpose, the performance of the model
is measured as normal, and is measured again with a dataset where the values
of the feature in question have been randomly permuted. The difference
between both performance measurements is the permutation variable
importance.

In familiar, this basic concept is extended in several ways:
\itemize{
\item Point estimates of variable importance are based on multiple (21) random
permutations. The difference between model performance on the normal dataset
and the median performance measurement of the randomly permuted datasets is
used as permutation variable importance.
\item Confidence intervals for the ensemble model are determined using bootstrap
methods.
\item Permutation variable importance is assessed for any metric specified using
the \code{metric} argument.
\item Permutation variable importance can take into account similarity between
features and permute similar features simultaneously.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.check_class_level_plausibility}
\alias{.check_class_level_plausibility}
\title{Internal function to test plausibility of provided class levels}
\usage{
.check_class_level_plausibility(
  data,
  outcome_type,
  outcome_column,
  class_levels,
  check_stringency = "strict"
)
}
\arguments{
\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{outcome_type}{(\strong{recommended}) Type of outcome found in the outcome
column. The outcome type determines many aspects of the overall process,
e.g. the available variable importance methods and learners, but also the
type of assessments that can be conducted to evaluate the resulting models.
Implemented outcome types are:
\itemize{
\item \code{binomial}: categorical outcome with 2 levels.
\item \code{multinomial}: categorical outcome with 2 or more levels.
\item \code{continuous}: general continuous numeric outcomes.
\item \code{survival}: survival outcome for time-to-event data.
}

If not provided, the algorithm will attempt to obtain outcome_type from
contents of the outcome column. This may lead to unexpected results, and we
therefore advise to provide this information manually.

Note that \code{competing_risk} survival analysis are not fully supported, and
is currently not a valid choice for \code{outcome_type}. The \code{count} outcome
type was deprecated in version 2.0.0, and superseded by \code{continuous}.}

\item{outcome_column}{(\strong{recommended}) Name of the column containing the
outcome of interest. May be identified from a formula, if a formula is
provided as an argument. Otherwise an error is raised. Note that \code{survival}
and \code{competing_risk} outcome type outcomes require two columns that
indicate the time-to-event or the time of last follow-up and the event
status.}

\item{class_levels}{(\emph{optional}) Class levels for \code{binomial} or \code{multinomial}
outcomes. This argument can be used to specify the ordering of levels for
categorical outcomes. These class levels must exactly match the levels
present in the outcome column.}

\item{check_stringency}{Specifies stringency of various checks. This is mostly:
\itemize{
\item \code{strict}: default value used for \code{summon_familiar}. Thoroughly checks
input data. Used internally for checking development data.
\item \code{external_warn}: value used for \code{extract_data} and related methods. Less
stringent checks, but will warn for possible issues. Used internally for
checking data for evaluation and explanation.
\item \code{external}: value used for external methods such as \code{predict}. Less
stringent checks, particularly for identifier and outcome columns, which may
be completely absent. Used internally for \code{predict}.
}}
}
\description{
This function checks whether categorical levels are present in the data that
are not found in the user-provided class levels.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationDecisionCurveAnalysis.R
\name{export_decision_curve_analysis_data}
\alias{export_decision_curve_analysis_data}
\alias{export_decision_curve_analysis_data,familiarCollection-method}
\alias{export_decision_curve_analysis_data,ANY-method}
\title{Extract and export decision curve analysis data.}
\usage{
export_decision_curve_analysis_data(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  ...
)

\S4method{export_decision_curve_analysis_data}{familiarCollection}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  ...
)

\S4method{export_decision_curve_analysis_data}{ANY}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{aggregate_results}{Flag that signifies whether results should be
aggregated for export.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A list of data.table (if \code{dir_path} is not provided), or nothing, as
all data is exported to \code{csv} files.
}
\description{
Extract and export decision curve analysis data in a
familiarCollection.
}
\details{
Data is usually collected from a \code{familiarCollection} object.
However, you can also provide one or more \code{familiarData} objects, that will
be internally converted to a \code{familiarCollection} object. It is also
possible to provide a \code{familiarEnsemble} or one or more \code{familiarModel}
objects together with the data from which data is computed prior to export.
Paths to the previous files can also be provided.

All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Decision curve analysis data is computed for categorical outcomes, i.e.
binomial and multinomial, as well as survival outcomes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{featureInfo-class}
\alias{featureInfo-class}
\title{Feature information object.}
\description{
A featureInfo object contains information for a single feature. This
information is used to check data prospectively for consistency and for data
preparation. These objects are, for instance, attached to a familiarModel
object so that data can be pre-processed in the same way as the development
data.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the feature, which by default is the column name of the
feature.}

\item{\code{set_descriptor}}{Character string describing the set to which the feature
belongs. Currently not used.}

\item{\code{feature_type}}{Describes the feature type, i.e. \code{factor} or \code{numeric}.}

\item{\code{levels}}{The class levels of categorical features. This is used to check
prospective datasets.}

\item{\code{ordered}}{Specifies whether the}

\item{\code{distribution}}{Five-number summary (numeric) or class frequency
(categorical).}

\item{\code{data_id}}{Internal identifier for the dataset used to derive the feature
information.}

\item{\code{run_id}}{Internal identifier for the specific subset of the dataset used
to derive the feature information.}

\item{\code{in_signature}}{Specifies whether the feature is included in the model
signature.}

\item{\code{in_novelty}}{Specifies whether the feature is included in the novelty
detector.}

\item{\code{removed}}{Specifies whether the feature was removed during
pre-processing.}

\item{\code{removed_unknown_type}}{Specifies whether the feature was removed during
pre-processing because the type was neither factor nor numeric..}

\item{\code{removed_missing_values}}{Specifies whether the feature was removed during
pre-processing because it contained too many missing values.}

\item{\code{removed_no_variance}}{Specifies whether the feature was removed during
pre-processing because it did not contain more than 1 unique value.}

\item{\code{removed_low_variance}}{Specifies whether the feature was removed during
pre-processing because the variance was too low. Requires applying
\code{low_variance} as a \code{filter_method}.}

\item{\code{removed_low_robustness}}{Specifies whether the feature was removed during
pre-processing because it lacks robustness. Requires applying \code{robustness}
as a \code{filter_method}, as well as repeated measurement.}

\item{\code{removed_low_importance}}{Specifies whether the feature was removed during
pre-processing because it lacks relevance. Requires applying
\code{univariate_test} as a \code{filter_method}.}

\item{\code{fraction_missing}}{Specifies the fraction of missing values.}

\item{\code{robustness}}{Specifies robustness of the feature, if measured.}

\item{\code{univariate_importance}}{Specifies the univariate p-value of the feature,
if measured.}

\item{\code{transformation_parameters}}{Details parameters for power transformation
of numeric features.}

\item{\code{normalisation_parameters}}{Details parameters for (global) normalisation
of numeric features.}

\item{\code{batch_normalisation_parameters}}{Details parameters for batch
normalisation of numeric features.}

\item{\code{imputation_parameters}}{Details parameters or models for imputation of
missing values.}

\item{\code{cluster_parameters}}{Details parameters for forming clusters with other
features.}

\item{\code{required_features}}{Details features required for clustering or
imputation.}

\item{\code{project_id}}{Identifier of the project that generated this collection.}

\item{\code{familiar_version}}{Version of the familiar package.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{get_risk_group_names,familiarCollection-method}
\alias{get_risk_group_names,familiarCollection-method}
\alias{get_risk_group_names}
\title{Get current risk group labels}
\usage{
\S4method{get_risk_group_names}{familiarCollection}(x)
}
\arguments{
\item{x}{A familiarCollection object.}
}
\value{
An ordered array of risk group labels.
}
\description{
Risk groups in familiarCollection objects can have custom names
for export and plotting. This function retrieves the currently assigned
names.
}
\details{
Labels convert internal naming of risk groups to the requested label
at export or when plotting. Labels can be changed using the
\code{set_risk_group_names} method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the familiarCollection class.
\item \code{\link{set_risk_group_names}} for updating the name and ordering of risk groups.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseSettings.R
\name{.parse_evaluation_settings}
\alias{.parse_evaluation_settings}
\title{Internal function for parsing settings related to model evaluation}
\usage{
.parse_evaluation_settings(
  config = NULL,
  data,
  parallel,
  outcome_type,
  hpo_metric,
  development_batch_id,
  vimp_aggregation_method,
  vimp_aggregation_rank_threshold,
  prep_cluster_method,
  prep_cluster_linkage_method,
  prep_cluster_cut_method,
  prep_cluster_similarity_threshold,
  prep_cluster_similarity_metric,
  evaluate_top_level_only = waiver(),
  evaluation_elements = waiver(),
  skip_evaluation_elements = waiver(),
  ensemble_method = waiver(),
  evaluation_metric = waiver(),
  sample_limit = waiver(),
  n_important_features = waiver(),
  detail_level = waiver(),
  estimation_type = waiver(),
  aggregate_results = waiver(),
  confidence_level = waiver(),
  bootstrap_ci_method = waiver(),
  shap_tolerance = waiver(),
  shap_max_iterations = waiver(),
  shap_phi_0 = waiver(),
  feature_cluster_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_linkage_method = waiver(),
  feature_similarity_metric = waiver(),
  feature_similarity_threshold = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  sample_similarity_metric = waiver(),
  eval_aggregation_method = waiver(),
  eval_aggregation_rank_threshold = waiver(),
  eval_icc_type = waiver(),
  stratification_method = waiver(),
  stratification_threshold = waiver(),
  time_max = waiver(),
  evaluation_times = waiver(),
  dynamic_model_loading = waiver(),
  parallel_evaluation = waiver(),
  ...
)
}
\arguments{
\item{config}{A list of settings, e.g. from an xml file.}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{parallel}{Logical value that whether familiar uses parallelisation. If
\code{FALSE} it will override \code{parallel_evaluation}.}

\item{outcome_type}{Type of outcome found in the data set.}

\item{hpo_metric}{Metric defined for hyperparameter optimisation.}

\item{development_batch_id}{Identifiers of batches used for model development.
These identifiers are used to determine the cohorts used to determine a
setting for \code{time_max}, if the \code{outcome_type} is \code{survival}, and both
\code{time_max} and \code{evaluation_times} are not provided.}

\item{vimp_aggregation_method}{Variable importance aggregation method.}

\item{vimp_aggregation_rank_threshold}{Rank threshold used for variable
importance aggregation.}

\item{prep_cluster_method}{Cluster method used during pre-processing.}

\item{prep_cluster_linkage_method}{Cluster linkage method used during
pre-processing.}

\item{prep_cluster_cut_method}{Cluster cut method used during pre-processing.}

\item{prep_cluster_similarity_threshold}{Cluster similarity threshold used
during pre-processing.}

\item{prep_cluster_similarity_metric}{Cluster similarity metric used during
pre-processing.}

\item{evaluate_top_level_only}{(\emph{optional}) Flag that signals that only
evaluation at the most global experiment level is required. Consider a
cross-validation experiment with additional external validation. The global
experiment level consists of data that are used for development, internal
validation and external validation. The next lower experiment level are the
individual cross-validation iterations.

When the flag is \code{true}, evaluations take place on the global level only,
and no results are generated for the next lower experiment levels. In our
example, this means that results from individual cross-validation iterations
are not computed and shown. When the flag is \code{false}, results are computed
from both the global layer and the next lower level.

Setting the flag to \code{true} saves computation time.}

\item{evaluation_elements}{(\emph{optional}) Specifies which evaluation steps
should be performed during the evaluation process. Defaults to \code{all},
indicating that all evaluation steps should be performed. It can have one or
more of the following values:
\itemize{
\item \code{all}, \code{true}: all evaluation steps are performed.
\item \code{none}, \code{false}: no evaluation is performed.
\item \code{auc_data}: data for assessing and plotting the area under the receiver
operating characteristic curve are computed.
\item \code{calibration_data}: data for assessing and plotting model calibration are
computed.
\item \code{calibration_info}: data required to assess calibration, such as baseline
survival curves, are collected. These data will still be present in the
models.
\item \code{confusion_matrix}: data for assessing and plotting a confusion matrix are
collected.
\item \code{decision_curve_analyis}: data for performing a decision curve analysis
are computed.
\item \code{feature_expressions}: data for assessing and plotting sample clustering
are computed.
\item \code{feature_similarity}: data for assessing and plotting feature clusters
not computed.
\item \code{fs_vimp}: data for assessing and plotting feature selection-based
variable importance are collected.
\item \code{hyperparameters}: data for assessing model hyperparameters are
collected. These data will still be present in the models.
\item \code{ice_data}: data for individual conditional expectation and partial
dependence plots are created.
\item \code{model_performance}: data for assessing and visualising model performance
are created.
\item \code{model_vimp}: data for assessing and plotting model-based variable
importance are collected.
\item \code{permutation_vimp}: data for assessing and plotting model-agnostic
permutation variable importance are computed.
\item \code{prediction_data}: predictions for each sample are made and exported.
\item \code{risk_stratification_data}: data for assessing and plotting Kaplan-Meier
survival curves are collected.
\item \code{risk_stratification_info}: data for assessing stratification into risk
groups are computed.
\item \code{sample_similarity}: data for assessing sample similarity are computed.
\item \code{shap}: data for assessing marginal contributions of feature values in
a SHAP analysis.
\item \code{univariate_analysis}: data for assessing and plotting univariate feature
importance are computed.
}

The logical intersect of the evaluation elements specified by
\code{evaluation_elements} and \code{skip_evaluation_elements} is used to define which
evaluation steps are performed.}

\item{skip_evaluation_elements}{(\emph{optional}) Specifies which evaluation steps,
if any, should be skipped as part of the evaluation process. Defaults to
\code{none}, which means that all relevant evaluation steps are performed. It can
have one or more of the same values as \code{evaluation_elements}.

The logical intersect of the evaluation elements specified by
\code{evaluation_elements} and \code{skip_evaluation_elements} is used to define which
evaluation steps are performed.}

\item{ensemble_method}{(\emph{optional}) Method for ensembling predictions from
models for the same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}

This parameter is only used if \code{detail_level} is \code{ensemble}.}

\item{evaluation_metric}{(\emph{optional}) One or more metrics for assessing model
performance. See the vignette on performance metrics for the available
metrics.

Confidence intervals (or rather credibility intervals) are computed for each
metric during evaluation. This is done using bootstraps, the number of which
depends on the value of \code{confidence_level} (Davison and Hinkley, 1997).

If unset, the metric in the \code{optimisation_metric} variable is used.}

\item{sample_limit}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}

\item{n_important_features}{(\emph{optional}) Set the number of features that are
evaluated in evaluation steps. Cannot be 0 or fewer.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("ice_data"=10, "permutation_vimp"=5)}.

This parameter can be set for the following data elements:
\code{ice_data}, \code{permutation_vimp}, and \code{shap}.}

\item{detail_level}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}

\item{estimation_type}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}

\item{aggregate_results}{(\emph{optional}) Flag that signifies whether results
should be aggregated during evaluation. If \code{estimation_type} is
\code{bias_correction} or \code{bc}, aggregation leads to a single bias-corrected
estimate. If \code{estimation_type} is \code{bootstrap_confidence_interval} or \code{bci},
aggregation leads to a single bias-corrected estimate with lower and upper
boundaries of the confidence interval. This has no effect if
\code{estimation_type} is \code{point}.

The default value is equal to \code{TRUE} except when assessing metrics to assess
model performance, as the default violin plot requires underlying data.

As with \code{detail_level} and \code{estimation_type}, a non-default
\code{aggregate_results} parameter can be specified for separate evaluation steps
by providing a parameter value in a named list with data elements, e.g.
\code{list("auc_data"=TRUE, , "model_performance"=FALSE)}. This parameter exists
for the same elements as \code{estimation_type}.}

\item{confidence_level}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}

\item{bootstrap_ci_method}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}

\item{shap_tolerance}{(\emph{optional}) Relative tolerance for convergence of SHAP
values. The tolerance is scaled with the range in SHAP values.

The default value is \code{0.05}.}

\item{shap_max_iterations}{(\emph{optional}) Maximum iterations for convergence of
SHAP values.

The default value is \code{1000} iterations.}

\item{shap_phi_0}{(\emph{optional}) Value(s) to use for computing marginal
contributions. By default, the average predicted value in the population is
used. For \code{multinomial} and \code{survival} outcomes, this parameter requires a
value for each class and evaluation time (\code{evaluation_times}),
respectively.}

\item{feature_cluster_method}{(\emph{optional}) Method used to perform clustering
of features. The same methods as for the \code{cluster_method} configuration
parameter are available: \code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

The value for the \code{cluster_method} configuration parameter is used by
default. When generating clusters for the purpose of determining mutual
correlation and ordering feature expressions, \code{none} is ignored and \code{hclust}
is used instead.}

\item{feature_cluster_cut_method}{(\emph{optional}) Method used to divide features
into separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

The value for the \code{cluster_cut_method} configuration parameter is used by
default.}

\item{feature_linkage_method}{(\emph{optional}) Method used for agglomerative
clustering with \code{hclust} and \code{agnes}. Linkage determines how features are
sequentially combined into clusters based on distance. The methods are
shared with the \code{cluster_linkage_method} configuration parameter: \code{average},
\code{single}, \code{complete}, \code{weighted}, and \code{ward}.

The value for the \code{cluster_linkage_method} configuration parameters is used
by default.}

\item{feature_similarity_metric}{(\emph{optional}) Metric to determine pairwise
similarity between features. Similarity is computed in the same manner as
for clustering, and \code{feature_similarity_metric} therefore has the same
options as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{mutual_information}, \code{spearman}, \code{kendall} and \code{pearson}.

The value used for the \code{cluster_similarity_metric} configuration parameter
is used by default.}

\item{feature_similarity_threshold}{(\emph{optional}) The threshold level for
pair-wise similarity that is required to form feature clusters with the
\code{fixed_cut} method. This threshold functions in the same manner as the one
defined using the \code{cluster_similarity_threshold} parameter.

By default, the value for the \code{cluster_similarity_threshold} configuration
parameter is used.

Unlike for \code{cluster_similarity_threshold}, more than one value can be
supplied here.}

\item{sample_cluster_method}{(\emph{optional}) The method used to perform
clustering based on distance between samples. These are the same methods as
for the \code{cluster_method} configuration parameter: \code{hclust}, \code{agnes}, \code{diana}
and \code{pam}.

The value for the \code{cluster_method} configuration parameter is used by
default. When generating clusters for the purpose of ordering samples in
feature expressions, \code{none} is ignored and \code{hclust} is used instead.}

\item{sample_linkage_method}{(\emph{optional}) The method used for agglomerative
clustering in \code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

The value for the \code{cluster_linkage_method} configuration parameters is used
by default.}

\item{sample_similarity_metric}{(\emph{optional}) Metric to determine pairwise
similarity between samples. Similarity is computed in the same manner as for
clustering, but \code{sample_similarity_metric} has different options that are
better suited to computing distance between samples instead of between
features. The following metrics are available.
\itemize{
\item \code{gower} (default): compute Gower's distance between samples. By default,
Gower's distance is computed based on winsorised data to reduce the effect
of outliers (see below).
\item \code{euclidean}: compute the Euclidean distance between samples.
}

The underlying feature data for numerical features is scaled to the
\eqn{[0,1]} range using the feature values across the samples. The
normalisation parameters required can optionally be computed from feature
data with the outer 5\% (on both sides) of feature values trimmed or
winsorised. To do so append \verb{_trim} (trimming) or \verb{_winsor} (winsorising) to
the metric name. This reduces the effect of outliers somewhat.

Regardless of metric, all categorical features are handled as for the
Gower's distance: distance is 0 if the values in a pair of samples match,
and 1 if they do not.}

\item{eval_aggregation_method}{(\emph{optional}) Method for aggregating variable
importances for the purpose of evaluation. Variable importances are
determined both during initial variable importance computation steps and
after training the model. Both types are evaluated after training the
model, but only the initial variable importances are evaluated at run-time.

See the documentation for the \code{vimp_aggregation_method} argument for
information concerning the different methods available.}

\item{eval_aggregation_rank_threshold}{(\emph{optional}) The threshold used to
define the subset of highly important features during evaluation.

See the documentation for the \code{vimp_aggregation_rank_threshold} argument for
more information.}

\item{eval_icc_type}{(\emph{optional}) String indicating the type of intraclass
correlation coefficient (\code{1}, \code{2} or \code{3}) that should be used to compute
robustness for features in repeated measurements during the evaluation of
univariate importance. These types correspond to the types in Shrout and
Fleiss (1979). The default value is \code{1}.}

\item{stratification_method}{(\emph{optional}) Method for determining the
stratification threshold for creating survival groups. The actual,
model-dependent, threshold value is obtained from the development data, and
can afterwards be used to perform stratification on validation data.

The following stratification methods are available:
\itemize{
\item \code{median} (default): The median predicted value in the development cohort
is used to stratify the samples into two risk groups. For predicted outcome
values that build a continuous spectrum, the two risk groups in the
development cohort will be roughly equal in size.
\item \code{mean}: The mean predicted value in the development cohort is used to
stratify the samples into two risk groups.
\item \code{mean_trim}: As \code{mean}, but based on the set of predicted values
where the 5\% lowest and 5\% highest values are discarded. This reduces the
effect of outliers.
\item \code{mean_winsor}: As \code{mean}, but based on the set of predicted values where
the 5\% lowest and 5\% highest values are winsorised. This reduces the effect
of outliers.
\item \code{fixed}: Samples are stratified based on the sample quantiles of the
predicted values. These quantiles are defined using the
\code{stratification_threshold} parameter.
\item \code{optimised}: Use maximally selected rank statistics to determine the
optimal threshold (Lausen and Schumacher, 1992; Hothorn et al., 2003) to
stratify samples into two optimally separated risk groups.
}

One or more stratification methods can be selected simultaneously.

This parameter is only relevant for \code{survival} outcomes.}

\item{stratification_threshold}{(\emph{optional}) Numeric value(s) signifying the
sample quantiles for stratification using the \code{fixed} method. The number of
risk groups will be the number of values +1.

The default value is \code{c(1/3, 2/3)}, which will yield two thresholds that
divide samples into three equally sized groups. If \code{fixed} is not among the
selected stratification methods, this parameter is ignored.

This parameter is only relevant for \code{survival} outcomes.}

\item{time_max}{(\emph{optional}) Time point which is used as the benchmark for
e.g. cumulative risks generated by random forest, or the cutoff for Uno's
concordance index.

If \code{time_max} is not provided, but \code{evaluation_times} is, the largest value
of \code{evaluation_times} is used. If both are not provided, \code{time_max} is set
to the 98th percentile of the distribution of survival times for samples
with an event in the development data set.

This parameter is only relevant for \code{survival} outcomes.}

\item{evaluation_times}{(\emph{optional}) One or more time points that are used for
assessing calibration in survival problems. This is done as expected and
observed survival probabilities depend on time.

If unset, \code{evaluation_times} will be equal to \code{time_max}.

This parameter is only relevant for \code{survival} outcomes.}

\item{dynamic_model_loading}{(\emph{optional}) Enables dynamic loading of models
during the evaluation process, if \code{TRUE}. Defaults to \code{FALSE}. Dynamic
loading of models may reduce the overall memory footprint, at the cost of
increased disk or network IO. Models can only be dynamically loaded if they
are found at an accessible disk or network location. Setting this parameter
to \code{TRUE} may help if parallel processing causes out-of-memory issues during
evaluation.}

\item{parallel_evaluation}{(\emph{optional}) Enable parallel processing for
hyperparameter optimisation. Defaults to \code{TRUE}. When set to \code{FALSE}, this
will disable the use of parallel processing while performing optimisation,
regardless of the settings of the \code{parallel} parameter. The parameter
moreover specifies whether parallelisation takes place within the evaluation
process steps (\code{inner}, default), or in an outer loop ( \code{outer}) over
learners, data subsamples, etc.

\code{parallel_evaluation} is ignored if \code{parallel=FALSE}.}

\item{...}{Unused arguments.}
}
\value{
List of parameters related to model evaluation.
}
\description{
Internal function for parsing settings related to model evaluation
}
\references{
\enumerate{
\item Davison, A. C. & Hinkley, D. V. Bootstrap methods and their
application. (Cambridge University Press, 1997).
\item Efron, B. & Hastie, T. Computer Age Statistical Inference. (Cambridge
University Press, 2016).
\item Lausen, B. & Schumacher, M. Maximally Selected Rank Statistics.
Biometrics 48, 73 (1992).
\item Hothorn, T. & Lausen, B. On the exact distribution of maximally selected
rank statistics. Comput. Stat. Data Anal. 43, 121–137 (2003).
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotICE.R
\name{plot_ice}
\alias{plot_ice}
\alias{plot_ice,ANY-method}
\alias{plot_ice,familiarCollection-method}
\title{Plot individual conditional expectation plots.}
\usage{
plot_ice(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = NULL,
  gradient_palette_range = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_range = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  novelty_range = NULL,
  value_scales = waiver(),
  novelty_scales = waiver(),
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  ice_default_alpha = 0.6,
  n_max_samples_shown = 50L,
  show_ice = TRUE,
  show_pd = TRUE,
  show_novelty = TRUE,
  anchor_values = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_ice}{ANY}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = NULL,
  gradient_palette_range = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_range = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  novelty_range = NULL,
  value_scales = waiver(),
  novelty_scales = waiver(),
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  ice_default_alpha = 0.6,
  n_max_samples_shown = 50L,
  show_ice = TRUE,
  show_pd = TRUE,
  show_novelty = TRUE,
  anchor_values = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_ice}{familiarCollection}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = NULL,
  gradient_palette_range = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_range = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  novelty_range = NULL,
  value_scales = waiver(),
  novelty_scales = waiver(),
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  ice_default_alpha = 0.6,
  n_max_samples_shown = 50L,
  show_ice = TRUE,
  show_pd = TRUE,
  show_novelty = TRUE,
  anchor_values = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{\code{familiarCollection} object, or one or more \code{familiarData}
objects, that will be internally converted to a \code{familiarCollection}
object. It is also possible to provide a \code{familiarEnsemble} or one or more
\code{familiarModel} objects together with the data from which data is computed
prior to export. Paths to such files can also be provided.

Additionally, some \code{familiarData} objects can be created from prediction
tables (\code{familiarDataElementPredictionTable}). Other \code{familiarData} objects
can be created from data (\code{dataObject}, or \code{data.table}). Please
check \emph{details} for more information.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created individual
conditional expectation plots are saved to. Output is saved in the
\code{explanation} subdirectory. If \code{NULL}, figures are written to the folder,
but are returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring plot elements
indicated by the \code{color_by} argument (if any). For 2D individual
conditional expectation plots without novelty, the initial colour
determines the colour of the points indicating sample values. \code{familiar}
has a default palette. Other palettes are supported by the \code{paletteer}
package, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{gradient_palette}{(\emph{optional}) Sequential or divergent palette used to
colour the raster in 2D individual conditional expectation or partial
dependence plots. This argument is not used for 1D plots. \code{familiar} has a
default palette. Other palettes are supported by the \code{paletteer} package,
\code{grDevices::palette.pals()} (requires R >= 4.0.0), \code{grDevices::hcl.pals()}
(requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors}, \code{terrain.colors},
\code{topo.colors} and \code{cm.colors}, which correspond to the palettes of the same
name in \code{grDevices}. You may also specify your own palette by providing a
vector of colour names listed by \code{grDevices::colors()} or through
hexadecimal RGB strings.}

\item{gradient_palette_range}{(\emph{optional}) Numerical range used to span the
gradient for 2D plots. This should be a range of two values, e.g. \code{c(0, 1)}. By default, values are determined from the data, dependent on the
\code{value_scales} parameter. This parameter is ignored for 1D plots.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{x_n_breaks}{(\emph{optional}) Number of breaks to show on the x-axis of the
plot. \code{x_n_breaks} is used to determine the \code{x_breaks} argument in case it
is unset.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{y_range}{(\emph{optional}) Value range for the y-axis.}

\item{y_n_breaks}{(\emph{optional}) Number of breaks to show on the y-axis of the
plot. \code{y_n_breaks} is used to determine the \code{y_breaks} argument in case it
is unset.}

\item{y_breaks}{(\emph{optional}) Break points on the y-axis of the plot.}

\item{novelty_range}{(\emph{optional}) Numerical range used to span the range of
novelty values. This determines the size of the bubbles in 2D, and
transparency of lines in 1D. This should be a range of two values, e.g.
\code{c(0, 1)}. By default, values are determined from the data, dependent on
the \code{value_scales} parameter. This parameter is ignored if
\code{show_novelty=FALSE}.}

\item{value_scales}{(\emph{optional}) Sets scaling of predicted values. This
parameter has several options:
\itemize{
\item \code{fixed} (default): The value axis for all features will have the same
range.
\item \code{feature}: The value axis for each feature will have the same range. This
option is unavailable for 2D plots.
\item \code{figure}: The value axis for all facets in a figure will have the same
range.
\item \code{facet}: Each facet has its own range. This option is unavailable for 2D
plots.
}

For 1D plots, this option is ignored if the \code{y_range} is provided, whereas
for 2D it is ignored if the \code{gradient_palette_range} is provided.}

\item{novelty_scales}{(\emph{optional}) Sets scaling of novelty values, similar to
the \code{value_scales} parameter, but with more limited options:
\itemize{
\item \code{fixed} (default): The novelty will have the same range for all features.
\item \code{figure}: The novelty will have the same range for all facets in a figure.
}}

\item{conf_int_style}{(\emph{optional}) Confidence interval style. See details for
allowed styles.}

\item{conf_int_alpha}{(\emph{optional}) Alpha value to determine transparency of
confidence intervals or, alternatively, other plot elements with which the
confidence interval overlaps. Only values between 0.0 (fully transparent)
and 1.0 (fully opaque) are allowed.}

\item{ice_default_alpha}{(\emph{optional}) Default transparency (value) of sample
lines in an 1D plot. When novelty is shown, this is the transparency
corresponding to the least novel points. The confidence interval alpha
values is scaled by this value.}

\item{n_max_samples_shown}{(\emph{optional}) Maximum number of samples shown in an
individual conditional expectation plot. Defaults to 50. These samples are
randomly picked from the samples present in the ICE data, but the same
samples are consistently picked. Partial dependence is nonetheless computed
from all available samples.}

\item{show_ice}{(\emph{optional}) Sets whether individual conditional expectation
plots should be created.}

\item{show_pd}{(\emph{optional}) Sets whether partial dependence plots should be
created. Note that if an anchor is set for a particular feature, its
partial dependence cannot be shown.}

\item{show_novelty}{(\emph{optional}) Sets whether novelty is shown in plots.}

\item{anchor_values}{(\emph{optional}) A single value or a named list or array of
values that are used to centre the individual conditional expectation plot.
A single value is valid if and only if only a single feature is assessed.
Otherwise, values Has no effect if the plot is not shown, i.e.
\code{show_ice=FALSE}. A partial dependence plot cannot be shown for those
features.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from the number of features and the number of facets.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=export_ice_data]{export_ice_data}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}, \code{\link[=extract_ice]{extract_ice}}
  \describe{
    \item{\code{aggregate_results}}{Flag that signifies whether results should be
aggregated for export.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
    \item{\code{features}}{Names of the feature or features (2) assessed simultaneously.
By default \code{NULL}, which means that all features are assessed one-by-one.}
    \item{\code{feature_x_range}}{When one or two features are defined using \code{features},
\code{feature_x_range} can be used to set the range of values for the first
feature. For numeric features, a vector of two values is assumed to indicate
a range from which \code{n_sample_points} are uniformly sampled. A vector of more
than two values is interpreted as is, i.e. these represent the values to be
sampled. For categorical features, values should represent a (sub)set of
available levels.}
    \item{\code{feature_y_range}}{As \code{feature_x_range}, but for the second feature in
case two features are defined.}
    \item{\code{n_sample_points}}{Number of points used to sample continuous features.}
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{evaluation_times}}{One or more time points that are used for in analysis
of survival problems when data has to be assessed at a set time, e.g.
calibration. If not provided explicitly, this parameter is read from
settings used at creation of the underlying \code{familiarModel} objects. Only
used for \code{survival} outcomes.}
    \item{\code{ensemble_method}}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{sample_limit}}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}
    \item{\code{n_important_features}}{(\emph{optional}) Set the number of features that are
evaluated in evaluation steps. Cannot be 0 or fewer.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("ice_data"=10, "permutation_vimp"=5)}.

This parameter can be set for the following data elements:
\code{ice_data}, \code{permutation_vimp}, and \code{shap}.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
    \item{\code{estimation_type}}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}
    \item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}
    \item{\code{bootstrap_ci_method}}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This method creates individual conditional expectation plots
based on data in a familiarCollection object.
}
\details{
This function generates individual conditional expectation plots.
These plots come in two varieties, namely 1D and 2D. 1D plots show the
predicted value as function of a single feature, whereas 2D plots show the
predicted value as a function of two features.

Available splitting variables are: \code{feature_x}, \code{feature_y} (2D only),
\code{vimp_method}, \code{learner}, \code{data_set} and \code{positive_class} (categorical
outcomes) or \code{evaluation_time} (survival outcomes). By default, for 1D ICE
plots the data are split by \code{feature_x}, \code{vimp_method} and \code{learner}, with
faceting by \code{data_set}, \code{positive_class} or \code{evaluation_time}. If only
partial dependence is shown, \code{positive_class} and \code{evaluation_time} are
used to set colours instead. For 2D plots, by default the data are split by
\code{feature_x}, \code{vimp_method} and \code{learner}, with faceting by \code{data_set},
\code{positive_class} or \code{evaluation_time}. The \code{color_by} argument cannot be
used with 2D plots, and attempting to do so causes an error. Attempting to
specify \code{feature_x} or \code{feature_y} for \code{color_by} will likewise result in
an error, as multiple features cannot be shown in the same facet.

Bootstrap confidence intervals of the partial dependence plots can be shown
using various styles set by \code{conf_int_style}:
\itemize{
\item \code{ribbon} (default): confidence intervals are shown as a ribbon with an
opacity of \code{conf_int_alpha} around the point estimate of the partial
dependence.
\item \code{step} (default): confidence intervals are shown as a step function around
the point estimate of the partial dependence.
\item \code{none}: confidence intervals are not shown. The point estimate of the
partial dependence is shown as usual.
}

Note that when bootstrap confidence intervals were computed, they were also
computed for individual samples in individual conditional expectation
plots. To avoid clutter, only point estimates for individual samples are
shown.

Labelling methods such as \code{set_vimp_method_names} or \code{set_data_set_names} can
be applied to the \code{familiarCollection} object to update labels, and order
the output in the figure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utilities.R
\name{is.encapsulated_path}
\alias{is.encapsulated_path}
\title{Internal test for encapsulated_path}
\usage{
is.encapsulated_path(x)
}
\arguments{
\item{x}{Object to be tested.}
}
\value{
\code{TRUE} for objects that are \code{encapsulated_path}, \code{FALSE} otherwise.
}
\description{
This function tests if the object is an \code{encapsulated_path} object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotShapSummary.R
\name{plot_shap_summary}
\alias{plot_shap_summary}
\alias{plot_shap_summary,ANY-method}
\alias{plot_shap_summary,familiarCollection-method}
\title{Plot SHAP summary.}
\usage{
plot_shap_summary(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  x_axis_by = NULL,
  y_axis_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  plot_type = NULL,
  value_representation = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_shap_summary}{ANY}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  x_axis_by = NULL,
  y_axis_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  plot_type = NULL,
  value_representation = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_shap_summary}{familiarCollection}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  x_axis_by = NULL,
  y_axis_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  plot_type = NULL,
  value_representation = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{\code{familiarCollection} object, or one or more \code{familiarData}
objects, that will be internally converted to a \code{familiarCollection}
object. It is also possible to provide a \code{familiarEnsemble} or one or more
\code{familiarModel} objects together with the data from which data is computed
prior to export. Paths to such files can also be provided.

Additionally, some \code{familiarData} objects can be created from prediction
tables (\code{familiarDataElementPredictionTable}). Other \code{familiarData} objects
can be created from data (\code{dataObject}, or \code{data.table}). Please
check \emph{details} for more information.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created SHAP summary
plots are saved to. Output is saved in the \code{explanation} subdirectory. If
\code{NULL} no figures are saved, but are returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{x_axis_by}{(\emph{optional}) Variable plotted along the x-axis of a plot.
The variable cannot overlap with variables provided to the \code{split_by} and
\code{y_axis_by} arguments (if used), but may overlap with other arguments. Only
one variable is allowed for this argument. See details for available
variables.}

\item{y_axis_by}{(\emph{optional}) Variable plotted along the y-axis of a plot.
The variable cannot overlap with variables provided to the \code{split_by} and
\code{x_axis_by} arguments (if used), but may overlap with other arguments. Only
one variable is allowed for this argument. See details for available
variables.}

\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{plot_type}{(\emph{optional}) Type of plot to draw. This is one of
\code{swarmplot} (draws a beeswarm plot), \code{barplot} (draws a barplot),
\code{boxplot} (draws a boxplot) and \code{violinplot} (draws a violin plot).
Defaults to \code{boxplot} if a single SHAP value is available for each feature,
and \code{swarmplot} otherwise.

The choice for \code{plot_type} affects several other arguments.}

\item{value_representation}{(\emph{optional}) Indicates how SHAP values are
represented, with the following options:
\itemize{
\item \code{raw} (default for \code{swarmplot}, \code{boxplot}, and \code{violinplot} plot
types): uses SHAP values as they are.
\item \code{abs}: uses absolute value of SHAP values.
\item \code{abs_mean} (default for \code{barplot} plot type): uses mean absolute value of
SHAP values. Only used by \code{barplot}.
\item \code{abs_max}: uses maximum absolute value of SHAP values. Only used by
\code{barplot}.
\item \code{abs_min}: uses minimum absolute value of SHAP values. Only used by
\code{barplot}.
}

If \code{abs_mean}, \code{abs_max} or \code{abs_min} are chosen, \code{plot_type} automatically
switches to \code{barplot}.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring plot elements
indicated by the \code{color_by} argument (if any). Only used if \code{plot_type} is
not \code{swarmplot}. \code{familiar} has a default palette. Other palettes are
supported by the \code{paletteer} package, \code{grDevices::palette.pals()} (requires
R >= 4.0.0), \code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow},
\code{heat.colors}, \code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which
correspond to the palettes of the same name in \code{grDevices}. You may also
specify your own palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{gradient_palette}{(\emph{optional}) Sequential or divergent palette used to
colour the points  the raster in the default \code{swarmplot} plots. This
argument is not used for other \code{plot_type} value. \code{familiar} has a default
palette. Other palettes are supported by the \code{paletteer} package,
\code{grDevices::palette.pals()} (requires R >= 4.0.0), \code{grDevices::hcl.pals()}
(requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors}, \code{terrain.colors},
\code{topo.colors} and \code{cm.colors}, which correspond to the palettes of the same
name in \code{grDevices}. You may also specify your own palette by providing a
vector of colour names listed by \code{grDevices::colors()} or through
hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{limit_n_features}{(\emph{optional}) The number of features that should be
included in the plot. Only the most important features are shown. By
default, the number of features is not limited.}

\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{x_n_breaks}{(\emph{optional}) Number of breaks to show on the x-axis of the
plot. \code{x_n_breaks} is used to determine the \code{x_breaks} argument in case it
is unset.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from the number of features and the number of facets.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link{extract_performance}}, \code{\link{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}
  \describe{
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{evaluation_times}}{One or more time points that are used for in analysis
of survival problems when data has to be assessed at a set time, e.g.
calibration. If not provided explicitly, this parameter is read from
settings used at creation of the underlying \code{familiarModel} objects. Only
used for \code{survival} outcomes.}
    \item{\code{ensemble_method}}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}
    \item{\code{metric}}{One or more metrics for assessing model performance. See the
vignette on performance metrics for the available metrics. If not provided
explicitly, this parameter is read from settings used at creation of the
underlying \code{familiarModel} objects.}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
    \item{\code{estimation_type}}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}
    \item{\code{aggregate_results}}{(\emph{optional}) Flag that signifies whether results
should be aggregated during evaluation. If \code{estimation_type} is
\code{bias_correction} or \code{bc}, aggregation leads to a single bias-corrected
estimate. If \code{estimation_type} is \code{bootstrap_confidence_interval} or \code{bci},
aggregation leads to a single bias-corrected estimate with lower and upper
boundaries of the confidence interval. This has no effect if
\code{estimation_type} is \code{point}.

The default value is equal to \code{TRUE} except when assessing metrics to assess
model performance, as the default violin plot requires underlying data.

As with \code{detail_level} and \code{estimation_type}, a non-default
\code{aggregate_results} parameter can be specified for separate evaluation steps
by providing a parameter value in a named list with data elements, e.g.
\code{list("auc_data"=TRUE, , "model_performance"=FALSE)}. This parameter exists
for the same elements as \code{estimation_type}.}
    \item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}
    \item{\code{bootstrap_ci_method}}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This method creates plots that show a summary of SHAP values
obtained from the data stored in a familiarCollection object.
}
\details{
This function creates SHAP summary plots, which provide an overview
of marginal contributions of feature values to the predicted values.

Available splitting variables are: \code{vimp_method}, \code{learner}, \code{data_set},
\code{evaluation_time} (survival outcome only) and \code{positive_class} (categorical
outcomes). The default for \code{value_representation = "raw"} is to facet by
\code{evaluation_time} or \code{positive_class}, and split by \code{vimp_method} and
\code{learner}. \code{color_by} is not used. The default for other
\code{value_representation} is to \code{color_by} \code{evaluation_time} or
\code{positive_class}, and split by \code{vimp_method}, \code{learner} and \code{data_set}.

Labelling methods such as \code{set_vimp_method_names} or \code{set_learner_names}
can be applied to the \code{familiarCollection} object to update labels, and
order the output in the figure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{set_learner_names,familiarCollection-method}
\alias{set_learner_names,familiarCollection-method}
\alias{set_learner_names}
\title{Rename learners for plotting and export}
\usage{
\S4method{set_learner_names}{familiarCollection}(x, old = NULL, new = NULL, order = NULL)
}
\arguments{
\item{x}{A familiarCollection object.}

\item{old}{(optional) Set of old labels to replace.}

\item{new}{Set of replacement labels. The number of replacement labels should
be equal to the number of provided old labels or the full number of labels.
If a subset of labels is to be replaced, both \code{old} and \code{new}
should be provided.}

\item{order}{(optional) Ordered set of replacement labels. This is used to
provide the order in which the labels should be placed, which affects e.g.
levels in a plot. If the ordering is not explicitly provided, the old
ordering is used.}
}
\value{
A familiarCollection object with custom labels for the learners.
}
\description{
Tabular exports and figures created from a familiarCollection
object can be customised by providing names for the learners.
}
\details{
Labels convert the internal naming for learners to the requested
label at export or when plotting. This enables the use of more specific
naming, e.g. changing \code{random_forest_rfsrc} to \code{Random Forest}.
Currently assigned labels can be found using the \code{get_learner_names}
method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the
familiarCollection class. * \code{\link{get_learner_names}} for obtaining
currently assigned labels.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Familiar.R
\docType{package}
\name{familiar}
\alias{familiar-package}
\alias{familiar}
\title{familiar: Fully Automated Machine Learning with Interpretable Analysis of Results}
\description{
End-to-end, automated machine learning package for creating
trustworthy and interpretable machine learning models. Familiar supports
modelling of regression, categorical and time-to-event (survival) outcomes.
Models created using familiar are self-containing, and their use does not
require additional information such as baseline survival, feature
clustering, or feature transformation and normalisation parameters. In
addition, an novelty or out-of-distribution detector is trained
simultaneously and contained with every model. Model performance,
calibration, risk group stratification, (permutation) variable importance,
individual conditional expectation, partial dependence, and more, are
assessed automatically as part of the evaluation process and exported in
tabular format and plotted, and may also be computed manually using export
and plot functions. Where possible, metrics and values obtained during the
evaluation process come with confidence intervals.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/oncoray/familiar}
  \item Report bugs at \url{https://github.com/oncoray/familiar/issues}
}

}
\author{
\strong{Maintainer}: Alex Zwanenburg \email{alexander.zwanenburg@nct-dresden.de} (\href{https://orcid.org/0000-0002-0342-9545}{ORCID})

Authors:
\itemize{
  \item Steffen Löck
}

Other contributors:
\itemize{
  \item German Cancer Research Center (DKFZ) [copyright holder]
  \item Technische Universität Dresden [copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotSampleClustering.R
\name{plot_sample_clustering}
\alias{plot_sample_clustering}
\alias{plot_sample_clustering,ANY-method}
\alias{plot_sample_clustering,familiarCollection-method}
\title{Plot heatmaps for pairwise similarity between features.}
\usage{
plot_sample_clustering(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  sample_limit = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  x_axis_by = NULL,
  y_axis_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  gradient_palette = NULL,
  gradient_palette_range = waiver(),
  outcome_palette = NULL,
  outcome_palette_range = waiver(),
  x_label = waiver(),
  x_label_shared = "column",
  y_label = waiver(),
  y_label_shared = "row",
  legend_label = waiver(),
  outcome_legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 3L,
  x_breaks = NULL,
  y_range = NULL,
  y_n_breaks = 3L,
  y_breaks = NULL,
  rotate_x_tick_labels = waiver(),
  show_feature_dendrogram = TRUE,
  show_sample_dendrogram = TRUE,
  show_normalised_data = TRUE,
  show_outcome = TRUE,
  dendrogram_height = grid::unit(1.5, "cm"),
  outcome_height = grid::unit(0.3, "cm"),
  evaluation_times = waiver(),
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  verbose = TRUE,
  ...
)

\S4method{plot_sample_clustering}{ANY}(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  sample_limit = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  x_axis_by = NULL,
  y_axis_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  gradient_palette = NULL,
  gradient_palette_range = waiver(),
  outcome_palette = NULL,
  outcome_palette_range = waiver(),
  x_label = waiver(),
  x_label_shared = "column",
  y_label = waiver(),
  y_label_shared = "row",
  legend_label = waiver(),
  outcome_legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 3L,
  x_breaks = NULL,
  y_range = NULL,
  y_n_breaks = 3L,
  y_breaks = NULL,
  rotate_x_tick_labels = waiver(),
  show_feature_dendrogram = TRUE,
  show_sample_dendrogram = TRUE,
  show_normalised_data = TRUE,
  show_outcome = TRUE,
  dendrogram_height = grid::unit(1.5, "cm"),
  outcome_height = grid::unit(0.3, "cm"),
  evaluation_times = waiver(),
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  verbose = TRUE,
  ...
)

\S4method{plot_sample_clustering}{familiarCollection}(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  sample_limit = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  x_axis_by = NULL,
  y_axis_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  gradient_palette = NULL,
  gradient_palette_range = waiver(),
  outcome_palette = NULL,
  outcome_palette_range = waiver(),
  x_label = waiver(),
  x_label_shared = "column",
  y_label = waiver(),
  y_label_shared = "row",
  legend_label = waiver(),
  outcome_legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 3L,
  x_breaks = NULL,
  y_range = NULL,
  y_n_breaks = 3L,
  y_breaks = NULL,
  rotate_x_tick_labels = waiver(),
  show_feature_dendrogram = TRUE,
  show_sample_dendrogram = TRUE,
  show_normalised_data = TRUE,
  show_outcome = TRUE,
  dendrogram_height = grid::unit(1.5, "cm"),
  outcome_height = grid::unit(0.3, "cm"),
  evaluation_times = waiver(),
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{feature_cluster_method}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_cluster_method}{The method used to perform clustering based on
distance between samples. These are the same methods as for the
\code{cluster_method} configuration parameter: \code{hclust}, \code{agnes}, \code{diana} and
\code{pam}.

\code{none} cannot be used when extracting data for feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_limit}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created performance
plots are saved to. Output is saved in the \code{feature_similarity}
subdirectory. If \code{NULL} no figures are saved, but are returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{x_axis_by}{(\emph{optional}) Variable plotted along the x-axis of a plot.
The variable cannot overlap with variables provided to the \code{split_by} and
\code{y_axis_by} arguments (if used), but may overlap with other arguments. Only
one variable is allowed for this argument. See details for available
variables.}

\item{y_axis_by}{(\emph{optional}) Variable plotted along the y-axis of a plot.
The variable cannot overlap with variables provided to the \code{split_by} and
\code{x_axis_by} arguments (if used), but may overlap with other arguments. Only
one variable is allowed for this argument. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{gradient_palette}{(\emph{optional}) Sequential or divergent palette used to
colour the similarity or distance between features in a heatmap. \code{familiar}
has a default palette. Other palettes are supported by the \code{paletteer}
package, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{gradient_palette_range}{(\emph{optional}) Numerical range used to span the
gradient. This should be a range of two values, e.g. \code{c(0, 1)}. Lower or
upper boundary can be unset by using \code{NA}. If not set, the full
metric-specific range is used.}

\item{outcome_palette}{(\emph{optional}) Sequential (\code{continuous} outcomes) or
qualitative (other outcome types) palette used to show outcome values. This
argument is ignored if the outcome is not shown.}

\item{outcome_palette_range}{(\emph{optional}) Numerical range used to span the
gradient of numeric (\code{continuous}) outcome values. This argument is ignored
for other outcome types or if the outcome is not shown.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{x_label_shared}{(\emph{optional}) Sharing of x-axis labels between facets.
One of three values:
\itemize{
\item \code{overall}: A single label is placed at the bottom of the figure. Tick
text (but not the ticks themselves) is removed for all but the bottom facet
plot(s).
\item \code{column}: A label is placed at the bottom of each column. Tick text (but
not the ticks themselves) is removed for all but the bottom facet plot(s).
\item \code{individual}: A label is placed below each facet plot. Tick text is kept.
}}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{y_label_shared}{(\emph{optional}) Sharing of y-axis labels between facets.
One of three values:
\itemize{
\item \code{overall}: A single label is placed to the left of the figure. Tick text
(but not the ticks themselves) is removed for all but the left-most facet
plot(s).
\item \code{row}: A label is placed to the left of each row. Tick text (but not the
ticks themselves) is removed for all but the left-most facet plot(s).
\item \code{individual}: A label is placed below each facet plot. Tick text is kept.
}}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{outcome_legend_label}{(\emph{optional}) Label to provide to the legend for
outcome data. If NULL, the legend will not have a name. By default,
\code{class}, \code{value} and \code{event} are used for \code{binomial} and \code{multinomial},
\code{continuous}, and \code{survival} outcome types, respectively.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{x_n_breaks}{(\emph{optional}) Number of breaks to show on the x-axis of the
plot. \code{x_n_breaks} is used to determine the \code{x_breaks} argument in case it
is unset.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{y_range}{(\emph{optional}) Value range for the y-axis.}

\item{y_n_breaks}{(\emph{optional}) Number of breaks to show on the y-axis of the
plot. \code{y_n_breaks} is used to determine the \code{y_breaks} argument in case it
is unset.}

\item{y_breaks}{(\emph{optional}) Break points on the y-axis of the plot.}

\item{rotate_x_tick_labels}{(\emph{optional}) Rotate tick labels on the x-axis by
90 degrees. Defaults to \code{TRUE}. Rotation of x-axis tick labels may also be
controlled through the \code{ggtheme}. In this case, \code{FALSE} should be provided
explicitly.}

\item{show_feature_dendrogram}{(\emph{optional}) Show feature dendrogram around
the main panel. Can be \code{TRUE}, \code{FALSE}, \code{NULL}, or a position, i.e. \code{top},
\code{bottom}, \code{left} and \code{right}.

If a position is specified, it should be appropriate with regard to the
\code{x_axis_by} or \code{y_axis_by} argument. If \code{x_axis_by} is \code{sample} (default),
the only valid positions are \code{top} (default) and \code{bottom}. Alternatively,
if \code{y_axis_by} is \code{feature}, the only valid positions are \code{right} (default)
and \code{left}.

A dendrogram can only be drawn from cluster methods that produce
dendograms, such as \code{hclust}. A dendogram can for example not be
constructed using the partioning around medioids method (\code{pam}).}

\item{show_sample_dendrogram}{(\emph{optional}) Show sample dendrogram around the
main panel. Can be \code{TRUE}, \code{FALSE}, \code{NULL}, or a position, i.e. \code{top},
\code{bottom}, \code{left} and \code{right}.

If a position is specified, it should be appropriate with regard to the
\code{x_axis_by} or \code{y_axis_by} argument. If \code{y_axis_by} is \code{sample} (default),
the only valid positions are \code{right} (default) and \code{left}. Alternatively,
if \code{x_axis_by} is \code{sample}, the only valid positions are \code{top} (default)
and \code{bottom}.

A dendrogram can only be drawn from cluster methods that produce
dendograms, such as \code{hclust}. A dendogram can for example not be
constructed using the partioning around medioids method (\code{pam}).}

\item{show_normalised_data}{(\emph{optional}) Flag that determines whether the
data shown in the main heatmap is normalised using the same settings as
within the analysis (\code{fixed}; default), using a standardisation method
(\code{set_normalisation}) that is applied separately to each dataset, or not at
all (\code{none}), which shows the data at the original scale, albeit with
batch-corrections.

Categorial variables are plotted to span 90\% of the entire numerical value
range, i.e. the levels of categorical variables with 2 levels are
represented at 5\% and 95\% of the range, with 3 levels at 5\%, 50\%, and 95\%,
etc.}

\item{show_outcome}{(\emph{optional}) Show outcome column(s) or row(s) in the
graph. Can be \code{TRUE}, \code{FALSE}, \code{NULL} or a poistion, i.e. \code{top}, \code{bottom},
\code{left} and \code{right}.

If a position is specified, it should be appropriate with regard to the
\code{x_axis_by} or \code{y_axis_by} argument. If \code{y_axis_by} is \code{sample} (default),
the only valid positions are \code{left} (default) and \code{right}. Alternatively,
if \code{x_axis_by} is \code{sample}, the only valid positions are \code{top} (default)
and \code{bottom}.

The outcome data will be drawn between the main panel and the sample
dendrogram (if any).}

\item{dendrogram_height}{(\emph{optional}) Height of the dendrogram. The height is
1.5 cm by default. Height is expected to be grid unit (see \code{grid::unit}),
which also allows for specifying relative heights.}

\item{outcome_height}{(\emph{optional}) Height of an outcome data column/row. The
height is 0.3 cm by default. Height is expected to be a grid unit (see
\code{grid::unit}), which also allows for specifying relative heights. In case
of \code{survival} outcome data with multipe \code{evaluation_times}, this height is
multiplied by the number of time points.}

\item{evaluation_times}{(\emph{optional}) Times at which the event status of
time-to-event survival outcomes are determined. Only used for \code{survival}
outcome. If not specified, the values used when creating the underlying
\code{familiarData} objects are used.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from the number of features and the number of facets.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{verbose}{Flag to indicate whether feedback should be provided for the
plotting.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}, \code{\link[=extract_feature_expression]{extract_feature_expression}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
    \item{\code{feature_similarity}}{Table containing pairwise distance between
sample. This is used to determine cluster information, and indicate which
samples are similar. The table is created by the
\code{extract_sample_similarity} method.}
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{feature_similarity_metric}}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{sample_similarity_metric}}{Metric to determine pairwise similarity
between samples. Similarity is computed in the same manner as for
clustering, but \code{sample_similarity_metric} has different options that are
better suited to computing distance between samples instead of between
features: \code{gower}, \code{euclidean}.

The underlying feature data is scaled to the \eqn{[0, 1]} range (for
numerical features) using the feature values across the samples. The
normalisation parameters required can optionally be computed from feature
data with the outer 5\% (on both sides) of feature values trimmed or
winsorised. To do so append \verb{_trim} (trimming) or \verb{_winsor} (winsorising) to
the metric name. This reduces the effect of outliers somewhat.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This method creates a heatmap based on data stored in a
\code{familiarCollection} object. Features in the heatmap are ordered so that
more similar features appear together.
}
\details{
This function generates area under the ROC curve plots.

Available splitting variables are: \code{vimp_method}, \code{learner}, and \code{data_set}.
By default, the data is split by \code{vimp_method} and \code{learner} and \code{data_set},
since the number of samples will typically differ between data sets, even
for the same variable importance method and learner.

The \code{x_axis_by} and \code{y_axis_by} arguments determine what data are shown
along which axis. Each argument takes one of \code{feature} and \code{sample}, and
both arguments should be unique. By default, features are shown along the
x-axis and samples along the y-axis.

Note that similarity is determined based on the underlying data. Hence the
ordering of features may differ between facets, and tick labels are
maintained for each panel.

Labeling methods such as \code{set_vimp_method_names} or \code{set_data_set_names} can
be applied to the \code{familiarCollection} object to update labels, and order
the output in the figure.

This plot can be created from \code{dataObject}, or \code{data.table} objects.
For \code{data.table}, see \code{\link{as_data_object}} for additional arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{experimentData-class}
\alias{experimentData-class}
\title{Experiment data}
\description{
An experimentData object contains information concerning the experiment.
These objects can be used to instantiate multiple experiments using the same
iterations, feature information and variable importance.
}
\details{
experimentData objects are primarily used to improve
reproducibility, since these allow for training models on a shared
foundation.
}
\section{Slots}{

\describe{
\item{\code{experiment_setup}}{Contains regarding the experimental setup that is used
to generate the iteration list.}

\item{\code{iteration_list}}{List of iteration data that determines which instances
are assigned to training, validation and test sets.}

\item{\code{feature_info}}{Feature information objects. Only available if the
experimentData object was generated using the \code{precompute_feature_info} or
\code{precompute_vimp} functions.}

\item{\code{vimp_hyperparameter_list}}{List of hyperparameters for variable importance
objects.  Only available if the experimentData object was created using the
\code{precompute_vimp} function.}

\item{\code{vimp_table_list}}{List of variable importance table objects. Only
available if the experimentData object was created using the
\code{precompute_vimp} function.}

\item{\code{project_id}}{Identifier of the project that generated the experimentData
object.}

\item{\code{familiar_version}}{Version of the familiar package used to create this
experimentData.}
}}

\seealso{
\code{\link{precompute_data_assignment}}
\code{\link{precompute_feature_info}}, \code{\link{precompute_vimp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PredictionTable.R
\name{as_prediction_table}
\alias{as_prediction_table}
\title{Convert to prediction table object}
\usage{
as_prediction_table(
  x,
  type,
  y = waiver(),
  batch_id = waiver(),
  sample_id = waiver(),
  series_id = waiver(),
  repetition_id = waiver(),
  time = waiver(),
  class_levels = waiver(),
  value_range = waiver(),
  event_indicator = waiver(),
  censoring_indicator = waiver(),
  learner = waiver(),
  vimp_method = waiver(),
  model_object = NULL,
  data = NULL
)
}
\arguments{
\item{x}{Values predicted using a learner. For all but \code{classification}
problems, predicted values should be a single vector of values in any
format that results in a single-column \code{data.table} using
\code{data.table::as.data.table}. For \code{classification} problems, predicted
values are probabilities for each class. Here, it is recommended to ensure
probabilities can be mapped to their respective class, e.g. using a named
list.}

\item{type}{The type of prediction table that should be created. The
following types are available:
\itemize{
\item \code{regression}: The predicted values are values for a regression.
\item \code{classification}: The predicted values are probabilities for specific
classes.
\item \code{hazard_ratio}: The predicted values are hazard ratios.
\item \code{cumulative_hazard}: The predicted values are cumulative hazards at
time \code{time}.
\item \code{expected_survival_time}: The predicted values are expected survival times.
\item \code{survival_probability}: The predicted values are survival probabilities
at time \code{time}.
}}

\item{y}{Known outcome value corresponding to each entry in \code{x}. For
survival-related outcomes, two sets of values are expected, corresponding
to the observed time and event status, respectively. Alternatively, a
\code{survival::Surv} object can be provided.}

\item{batch_id}{(\emph{optional}) Array of batch or cohort identifiers.

In familiar any row of data is organised by four identifiers:
\itemize{
\item The batch identifier \code{batch_id}: This denotes the group to which a
set of samples belongs, e.g. patients from a single study, samples measured
in a batch, etc. The batch identifier is used for batch normalisation, as
well as selection of development and validation datasets.
\item The sample identifier \code{sample_id}: This denotes the sample level,
e.g. data from a single individual. Subsets of data, e.g. bootstraps or
cross-validation folds, are created at this level.
\item The series identifier \code{series_id}: Indicates measurements on a
single sample that may not share the same outcome value, e.g. a time
series, or the number of cells in a view.
\item The repetition identifier \code{repetition_id}: Indicates repeated measurements
in a single series where any feature values may differ, but the outcome
does not. Repetition identifiers are always implicitly set when multiple
entries for the same series of the same sample in the same batch that share
the same outcome are encountered.
}}

\item{sample_id}{(\emph{optional}) Array of sample or subject identifiers. See
\code{batch_id} above for more details.

If unset, every row will be identified as a single sample.}

\item{series_id}{(\emph{optional}) Array of series identifiers, which distinguish
between measurements that are part of a series for a single sample. See
\code{batch_id} above for more details.}

\item{repetition_id}{(\emph{optional}) Array of repetition identifiers, which
distinguishes between repeated measurements within a single series. See
\code{batch_id} above for more details.}

\item{time}{Time point at which the predicted values are generated e.g. the
cumulative risks generated by random forest.

This parameter is only relevant for \code{survival} outcomes.}

\item{class_levels}{(\emph{optional}) Class levels for \code{binomial} or \code{multinomial}
outcomes. This argument can be used to specify the ordering of levels for
categorical outcomes. These class levels must exactly match the levels
present in the outcome column.}

\item{value_range}{Range of observed, \strong{not predicted}, values.

This parameter is only relevant for \code{continuous} outcomes.}

\item{event_indicator}{(\strong{recommended}) Indicator for events in \code{survival}
and \code{competing_risk} analyses. \code{familiar} will automatically recognise \code{1},
\code{true}, \code{t}, \code{y} and \code{yes} as event indicators, including different
capitalisations. If this parameter is set, it replaces the default values.}

\item{censoring_indicator}{(\strong{recommended}) Indicator for right-censoring in
\code{survival} and \code{competing_risk} analyses. \code{familiar} will automatically
recognise \code{0}, \code{false}, \code{f}, \code{n}, \code{no} as censoring indicators, including
different capitalisations. If this parameter is set, it replaces the
default values.}

\item{learner}{The type of learner that generated the predictions.}

\item{vimp_method}{The type of variable importance method for identifying the
features included by the learner that generated the predictions.}

\item{model_object}{A familiarModel or familiarEnsemble that can be used (and
is used internally) for setting several of the other arguments of this
function.}

\item{data}{A familiar dataObject object that can be used (and is used
internally) for setting many of the other arguments of this function.}
}
\value{
A prediction table object.
}
\description{
Creates a prediction table object from input data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VimpTable.R
\name{aggregate_vimp_table}
\alias{aggregate_vimp_table}
\alias{aggregate_vimp_table,list-method}
\alias{aggregate_vimp_table,character-method}
\alias{aggregate_vimp_table,vimpTable-method}
\alias{aggregate_vimp_table,NULL-method}
\alias{aggregate_vimp_table,experimentData-method}
\title{Aggregate variable importance from multiple variable importance
objects.}
\usage{
aggregate_vimp_table(x, aggregation_method, rank_threshold = NULL, ...)

\S4method{aggregate_vimp_table}{list}(x, aggregation_method, rank_threshold = NULL, ...)

\S4method{aggregate_vimp_table}{character}(x, aggregation_method, rank_threshold = NULL, ...)

\S4method{aggregate_vimp_table}{vimpTable}(x, aggregation_method, rank_threshold = NULL, ...)

\S4method{aggregate_vimp_table}{NULL}(x, aggregation_method, rank_threshold = NULL, ...)

\S4method{aggregate_vimp_table}{experimentData}(x, aggregation_method, rank_threshold = NULL, ...)
}
\arguments{
\item{x}{Variable importance (\code{vimpTable}) object, a list thereof, or one or
more paths to these objects.}

\item{aggregation_method}{Method used to aggregate variable importance. The
available methods are described in the \emph{feature selection methods} vignette.}

\item{rank_threshold}{Rank threshold used within several aggregation methods.
See the \emph{feature selection methods} vignette for more details.}

\item{...}{unused parameters.}
}
\value{
A \code{vimpTable} object with aggregated variable importance data.
}
\description{
This methods aggregates variable importance from one or more
\code{vimpTable} objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationPredictionData.R,
%   R/PredictionTable.R
\name{export_prediction_data}
\alias{export_prediction_data}
\alias{export_prediction_data,ANY-method}
\alias{export_prediction_data,familiarCollection-method}
\title{Extract and export predicted values.}
\usage{
export_prediction_data(object, dir_path = NULL, export_collection = FALSE, ...)

\S4method{export_prediction_data}{ANY}(object, dir_path = NULL, export_collection = FALSE, ...)

\S4method{export_prediction_data}{familiarCollection}(object, dir_path = NULL, export_collection = FALSE, ...)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link{extract_predictions}}, \code{\link{as_familiar_collection}}
  \describe{
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{evaluation_times}}{One or more time points that are used for in analysis
of survival problems when data has to be assessed at a set time, e.g.
calibration. If not provided explicitly, this parameter is read from
settings used at creation of the underlying \code{familiarModel} objects. Only
used for \code{survival} outcomes.}
    \item{\code{ensemble_method}}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
    \item{\code{estimation_type}}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}
    \item{\code{aggregate_results}}{(\emph{optional}) Flag that signifies whether results
should be aggregated during evaluation. If \code{estimation_type} is
\code{bias_correction} or \code{bc}, aggregation leads to a single bias-corrected
estimate. If \code{estimation_type} is \code{bootstrap_confidence_interval} or \code{bci},
aggregation leads to a single bias-corrected estimate with lower and upper
boundaries of the confidence interval. This has no effect if
\code{estimation_type} is \code{point}.

The default value is equal to \code{TRUE} except when assessing metrics to assess
model performance, as the default violin plot requires underlying data.

As with \code{detail_level} and \code{estimation_type}, a non-default
\code{aggregate_results} parameter can be specified for separate evaluation steps
by providing a parameter value in a named list with data elements, e.g.
\code{list("auc_data"=TRUE, , "model_performance"=FALSE)}. This parameter exists
for the same elements as \code{estimation_type}.}
    \item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A list of data.tables (if \code{dir_path} is not provided), or nothing, as
all data is exported to \code{csv} files.
}
\description{
Extract and export the values predicted by single and ensemble
models in a familiarCollection.
}
\details{
Data, such as model performance and calibration information, is
usually collected from a \code{familiarCollection} object. However, you can also
provide one or more \code{familiarData} objects, that will be internally
converted to a \code{familiarCollection} object. It is also possible to provide a
\code{familiarEnsemble} or one or more \code{familiarModel} objects together with the
data from which data is computed prior to export. Paths to the previous
files can also be provided.

All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Both single and ensemble predictions are exported.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{set_data_set_names,familiarCollection-method}
\alias{set_data_set_names,familiarCollection-method}
\alias{set_data_set_names}
\title{Name datasets for plotting and export}
\usage{
\S4method{set_data_set_names}{familiarCollection}(x, old = NULL, new = NULL, order = NULL)
}
\arguments{
\item{x}{A familiarCollection object.}

\item{old}{(optional) Set of old labels to replace.}

\item{new}{Set of replacement labels. The number of replacement labels should
be equal to the number of provided old labels or the full number of labels.
If a subset of labels is to be replaced, both \code{old} and \code{new}
should be provided.}

\item{order}{(optional) Ordered set of replacement labels. This is used to
provide the order in which the labels should be placed, which affects e.g.
levels in a plot. If the ordering is not explicitly provided, the old
ordering is used.}
}
\value{
A familiarCollection object with custom names for the data sets.
}
\description{
Tabular exports and figures created from a familiarCollection
object can be customised by setting data labels.
}
\details{
Labels convert internal naming of data sets to the requested label
at export or when plotting. Currently assigned labels can be found using
the \code{get_data_set_names} method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the
familiarCollection class. * \code{\link{get_data_set_names}} for obtaining
currently assigned labels.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarObjectUpdate.R
\name{update_object}
\alias{update_object}
\alias{update_object,familiarModel-method}
\alias{update_object,familiarEnsemble-method}
\alias{update_object,familiarData-method}
\alias{update_object,familiarCollection-method}
\alias{update_object,vimpTable-method}
\alias{update_object,familiarNoveltyDetector-method}
\alias{update_object,featureInfo-method}
\alias{update_object,featureInfoParametersTransformationPowerTransform-method}
\alias{update_object,experimentData-method}
\alias{update_object,list-method}
\alias{update_object,ANY-method}
\title{Update familiar S4 objects to the most recent version.}
\usage{
update_object(object, ...)

\S4method{update_object}{familiarModel}(object, ...)

\S4method{update_object}{familiarEnsemble}(object, ...)

\S4method{update_object}{familiarData}(object, ...)

\S4method{update_object}{familiarCollection}(object, ...)

\S4method{update_object}{vimpTable}(object, ...)

\S4method{update_object}{familiarNoveltyDetector}(object, ...)

\S4method{update_object}{featureInfo}(object, ...)

\S4method{update_object}{featureInfoParametersTransformationPowerTransform}(object, ...)

\S4method{update_object}{experimentData}(object, ...)

\S4method{update_object}{list}(object, ...)

\S4method{update_object}{ANY}(object, ...)
}
\arguments{
\item{object}{A \code{familiarModel}, a \code{familiarEnsemble}, a \code{familiarData} or
\code{familiarCollection} object.}

\item{...}{Unused arguments.}
}
\value{
An up-to-date version of the respective S4 object.
}
\description{
Provides backward compatibility for familiar objects exported to
a file. This mitigates compatibility issues when working with files that
become outdated as new versions of familiar are released, e.g. because
slots have been removed.

Major version releases (e.g., versions 1.0.0, 2.0.0) introduce breaking
changes which can result in objects that cannot be updated.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseSettings.R
\name{.parse_model_development_settings}
\alias{.parse_model_development_settings}
\title{Internal function for parsing settings related to model development}
\usage{
.parse_model_development_settings(
  config = NULL,
  data,
  parallel,
  outcome_type,
  learner = waiver(),
  hyperparameter = waiver(),
  novelty_detector = waiver(),
  detector_parameters = waiver(),
  parallel_model_development = waiver(),
  ...
)
}
\arguments{
\item{config}{A list of settings, e.g. from an xml file.}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{parallel}{Logical value that whether familiar uses parallelisation. If
\code{FALSE} it will override \code{parallel_model_development}.}

\item{outcome_type}{Type of outcome found in the data set.}

\item{learner}{(\strong{required}) One or more algorithms used for model
development. A sizeable number learners is supported in \code{familiar}. Please
see the vignette on learners for more information concerning the available
learners.}

\item{hyperparameter}{(\emph{optional}) List of lists containing hyperparameters
for learners. Each sublist should have the name of the learner method it
corresponds to, with list elements being named after the intended
hyperparameter, e.g. \code{"glm_logistic"=list("sign_size"=3)}

All learners have hyperparameters. Please refer to the vignette on learners
for more details. If no parameters are provided, sequential model-based
optimisation is used to determine optimal hyperparameters.

Hyperparameters provided by the user are never optimised. However, if more
than one value is provided for a single hyperparameter, optimisation will
be conducted using these values.}

\item{novelty_detector}{(\emph{optional}) Specify the algorithm used for training
a novelty detector. This detector can be used to identify
out-of-distribution data prospectively.}

\item{detector_parameters}{(\emph{optional}) List lists containing hyperparameters
for novelty detectors. Currently not used.}

\item{parallel_model_development}{(\emph{optional}) Enable parallel processing for
the model development workflow. Defaults to \code{TRUE}. When set to \code{FALSE},
this will disable the use of parallel processing while developing models,
regardless of the settings of the \code{parallel} parameter.
\code{parallel_model_development} is ignored if \code{parallel=FALSE}.}

\item{...}{Unused arguments.}
}
\value{
List of parameters related to model development.
}
\description{
Internal function for parsing settings related to model development
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseData.R
\name{.parse_categorical_features}
\alias{.parse_categorical_features}
\title{Internal function for setting categorical features}
\usage{
.parse_categorical_features(data, outcome_type, reference_method = "auto")
}
\arguments{
\item{data}{data.table with feature data}

\item{outcome_type}{character, indicating the type of outcome}

\item{reference_method}{character, indicating the type of method used to set
the reference level.}
}
\value{
data.table with several features converted to factor.
}
\description{
Internal function for setting categorical features
}
\details{
This function parses columns containing feature data to factors if
the data contained therein have logical (TRUE, FALSE), character, or factor
classes.  Unless passed as feature names with \code{reference}, numerical data,
including integers, are not converted to factors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarModel.R
\name{summary}
\alias{summary}
\alias{summary,familiarModel-method}
\title{Model summaries}
\usage{
summary(object, ...)

\S4method{summary}{familiarModel}(object, ...)
}
\arguments{
\item{object}{a familiarModel object}

\item{...}{additional arguments passed to \code{summary} methods for the underlying
model, when available.}
}
\value{
Depends on underlying model. See the documentation for the particular
models.
}
\description{
\code{summary} produces model summaries.
}
\details{
This method extends the \code{summary} S3 method. For some models
\code{summary} requires information that is trimmed from the model. In this case
a copy of summary data is stored with the model, and returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationHyperparameters.R
\name{extract_hyperparameters}
\alias{extract_hyperparameters}
\title{Internal function to extract hyperparameters from models.}
\usage{
extract_hyperparameters(object, message_indent = 0L, verbose = FALSE, ...)
}
\arguments{
\item{object}{A \code{familiarEnsemble}, which is an ensemble of one or more
\code{familiarModel} objects, or a \code{familiarDataElementPredictionTable} object
that contains prediction data.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}

\item{...}{Unused arguments.}
}
\value{
A list of familiarDataElements with hyperparameters.
}
\description{
Collects hyperparameters from models in a \code{familiarEnsemble}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VimpTable.R
\name{get_vimp_table}
\alias{get_vimp_table}
\alias{get_vimp_table,list-method}
\alias{get_vimp_table,character-method}
\alias{get_vimp_table,vimpTable-method}
\alias{get_vimp_table,NULL-method}
\alias{get_vimp_table,experimentData-method}
\alias{get_vimp_table,familiarModel-method}
\title{Extract variable importance table.}
\usage{
get_vimp_table(x, state = "ranked", ...)

\S4method{get_vimp_table}{list}(x, state = "ranked", ...)

\S4method{get_vimp_table}{character}(x, state = "ranked", ...)

\S4method{get_vimp_table}{vimpTable}(x, state = "ranked", ...)

\S4method{get_vimp_table}{NULL}(x, state = "ranked", ...)

\S4method{get_vimp_table}{experimentData}(x, state = "ranked", ...)

\S4method{get_vimp_table}{familiarModel}(x, state = "ranked", data = NULL, as_object = FALSE, ...)
}
\arguments{
\item{x}{Variable importance (\code{vimpTable}) object, a list thereof, or one or
more paths to these objects. This method extracts the variable importance
table from such objects.}

\item{state}{State of the returned variable importance table. This affects
what contents are shown, and in which format. The variable importance table
can be returned with the following states:
\itemize{
\item \code{initial}: initial state, directly after the variable importance table is
filled. The returned variable importance table shows the raw, un-processed
data.
\item \code{decoded}: depending on the variable importance method, the initial
variable importance table may contain the scores of individual contrasts
for categorical variables. When decoded, scores from all contrasts are
aggregated to a single score for each feature.
\item \code{declustered}: variable importance is determined from fully processed
features, which includes clustering. This means that a single feature in
the variable importance table may represent multiple original features.
When a variable importance table has been declustered, all clusters have
been turned into their constituent features.
\item \code{ranked} (default): The scores have been used to create ranks, with lower
ranks indicating better features.
}

Internally, the variable importance table will go through each state, i.e.
an variable importance table in the initial state will be decoded,
declustered and then ranked prior to returning the variable importance
table.}

\item{...}{Unused arguments.}

\item{data}{Internally used argument for use with \code{familiarModel} objects.}

\item{as_object}{Internally used argument for use with \code{familiarModel}
objects.}
}
\value{
A \code{data.table} with variable importance scores and, with
\code{state="ranked"}, the respective ranks.
}
\description{
This method retrieves and parses variable importance tables from
their respective \code{vimpTable} objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseSettings.R
\name{.parse_file_paths}
\alias{.parse_file_paths}
\title{Internal function for parsing file paths}
\usage{
.parse_file_paths(
  config = NULL,
  project_dir = waiver(),
  experiment_dir = waiver(),
  data_file = waiver(),
  verbose = TRUE,
  ...
)
}
\arguments{
\item{config}{A list of settings, e.g. from an xml file.}

\item{project_dir}{(\emph{optional}) Path to the project directory. \code{familiar}
checks if the directory indicated by \code{experiment_dir} and data files in
\code{data_file} are relative to the \code{project_dir}.}

\item{experiment_dir}{(\strong{recommended}) Path to the directory where all
intermediate and final results produced by \code{familiar} are written to.

The \code{experiment_dir} can be a path relative to \code{project_dir} or an absolute
path.

In case no project directory is provided and the experiment directory is
not on an absolute path, a directory will be created in the temporary R
directory indicated by \code{tempdir()}. This directory is deleted after closing
the R session or once data analysis has finished. All information will be
lost afterwards. Hence, it is recommended to provide either
\code{experiment_dir} as an absolute path, or provide both \code{project_dir} and
\code{experiment_dir}.}

\item{data_file}{(\emph{optional}) Path to files containing data that should be
analysed. The paths can be relative to \code{project_dir} or absolute paths. An
error will be raised if the file cannot be found.

The following types of data are supported.
\itemize{
\item \code{csv} files containing column headers on the first row, and samples per
row. \code{csv} files are read using \code{data.table::fread}.
\item \code{rds} files that contain a \code{data.table} or \code{data.frame} object. \code{rds}
files are imported using \code{base::readRDS}.
\item \code{RData} files that contain a single \code{data.table} or \code{data.frame} object.
\code{RData} files are imported using \code{base::load}.
}

All data are expected in wide format, with sample information organised
row-wise.

More than one data file can be provided. \code{familiar} will try to combine
data files based on column names and identifier columns.

Alternatively, data can be provided using the \code{data} argument. These data
are expected to be \code{data.frame} or \code{data.table} objects or paths to data
files. The latter are handled in the same way as file paths provided to
\code{data_file}.}

\item{verbose}{Sets verbosity.}

\item{...}{Unused arguments}
}
\value{
List of paths to important directories and files.
}
\description{
Internal function for parsing file paths
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.check_survival_time_plausibility}
\alias{.check_survival_time_plausibility}
\title{Internal function to test plausibility of provided survival times.}
\usage{
.check_survival_time_plausibility(
  data,
  outcome_type,
  outcome_column,
  check_stringency = "strict"
)
}
\arguments{
\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{outcome_type}{(\strong{recommended}) Type of outcome found in the outcome
column. The outcome type determines many aspects of the overall process,
e.g. the available variable importance methods and learners, but also the
type of assessments that can be conducted to evaluate the resulting models.
Implemented outcome types are:
\itemize{
\item \code{binomial}: categorical outcome with 2 levels.
\item \code{multinomial}: categorical outcome with 2 or more levels.
\item \code{continuous}: general continuous numeric outcomes.
\item \code{survival}: survival outcome for time-to-event data.
}

If not provided, the algorithm will attempt to obtain outcome_type from
contents of the outcome column. This may lead to unexpected results, and we
therefore advise to provide this information manually.

Note that \code{competing_risk} survival analysis are not fully supported, and
is currently not a valid choice for \code{outcome_type}. The \code{count} outcome
type was deprecated in version 2.0.0, and superseded by \code{continuous}.}

\item{outcome_column}{(\strong{recommended}) Name of the column containing the
outcome of interest. May be identified from a formula, if a formula is
provided as an argument. Otherwise an error is raised. Note that \code{survival}
and \code{competing_risk} outcome type outcomes require two columns that
indicate the time-to-event or the time of last follow-up and the event
status.}
}
\description{
This function checks whether non-positive outcome time is present in the
data. This may produce unexpected results for some packages. For example,
glmnet will not train if an instance has a survival time of 0 or lower.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarNoveltyDetector.R
\name{set_object_name,familiarNoveltyDetector-method}
\alias{set_object_name,familiarNoveltyDetector-method}
\title{Set the name of a \code{familiarNoveltyDetector} object.}
\usage{
\S4method{set_object_name}{familiarNoveltyDetector}(x, new = NULL)
}
\arguments{
\item{x}{A \code{familiarNoveltyDetector} object.}
}
\value{
A \code{familiarNoveltyDetector} object with a generated or a provided name.
}
\description{
Set the \code{name} slot using the object name.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotConfusionMatrix.R
\name{plot_confusion_matrix}
\alias{plot_confusion_matrix}
\alias{plot_confusion_matrix,ANY-method}
\alias{plot_confusion_matrix,familiarCollection-method}
\title{Plot confusion matrix.}
\usage{
plot_confusion_matrix(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  rotate_x_tick_labels = waiver(),
  show_alpha = TRUE,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_confusion_matrix}{ANY}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  rotate_x_tick_labels = waiver(),
  show_alpha = TRUE,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_confusion_matrix}{familiarCollection}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  rotate_x_tick_labels = waiver(),
  show_alpha = TRUE,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{\code{familiarCollection} object, or one or more \code{familiarData}
objects, that will be internally converted to a \code{familiarCollection}
object. It is also possible to provide a \code{familiarEnsemble} or one or more
\code{familiarModel} objects together with the data from which data is computed
prior to export. Paths to such files can also be provided.

Additionally, some \code{familiarData} objects can be created from prediction
tables (\code{familiarDataElementPredictionTable}). Other \code{familiarData} objects
can be created from data (\code{dataObject}, or \code{data.table}). Please
check \emph{details} for more information.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created confusion
matrixes are saved to. Output is saved in the \code{performance} subdirectory.
If \code{NULL} no figures are saved, but are returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette Palette for colouring the cells
of the confusion matrix. The colour depends on whether each cell of the
confusion matrix is on the diagonal (observed outcome matched expected
outcome) or not. \code{familiar} has a default palette. Other palettes are
supported by the \code{paletteer} package, \code{grDevices::palette.pals()} (requires
R >= 4.0.0), \code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow},
\code{heat.colors}, \code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which
correspond to the palettes of the same name in \code{grDevices}. You may also
specify your own palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{rotate_x_tick_labels}{(\emph{optional}) Rotate tick labels on the x-axis by
90 degrees. Defaults to \code{TRUE}. Rotation of x-axis tick labels may also be
controlled through the \code{ggtheme}. In this case, \code{FALSE} should be provided
explicitly.}

\item{show_alpha}{(\emph{optional}) Interpreting confusion matrices is made easier
by setting the opacity of the cells. \code{show_alpha} takes the following
values:
\itemize{
\item \code{none}: Cell opacity is not altered. Diagonal and off-diagonal cells are
completely opaque and transparent, respectively. Same as
\code{show_alpha=FALSE}.
\item \code{by_class}: Cell opacity is normalised by the number of instances for each
observed outcome class in each confusion matrix.
\item \code{by_matrix} (default): Cell opacity is normalised by the number of
instances in the largest observed outcome class in each confusion matrix.
Same as \code{show_alpha=TRUE}
\item \code{by_figure}: Cell opacity is normalised by the number of instances in the
largest observed outcome class across confusion matrices in different
facets.
\item \code{by_all}: Cell opacity is normalised by the number of instances in the
largest observed outcome class across all confusion matrices.
}}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from the number of features and the number of facets.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}, \code{\link[=extract_confusion_matrix]{extract_confusion_matrix}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{ensemble_method}}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This method creates confusion matrices based on data in a
familiarCollection object.
}
\details{
This function generates area under the ROC curve plots.

Available splitting variables are: \code{vimp_method}, \code{learner} and \code{data_set}.
By default, the data is split by \code{vimp_method} and \code{learner}, with facetting
by \code{data_set}.

Labelling methods such as \code{set_vimp_method_names} or \code{set_data_set_names} can
be applied to the \code{familiarCollection} object to update labels, and order
the output in the figure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ExperimentSetup.R
\name{extract_experimental_setup}
\alias{extract_experimental_setup}
\title{Parse experimental design}
\usage{
extract_experimental_setup(
  experimental_design,
  file_dir,
  message_indent = 0L,
  verbose = TRUE
)
}
\arguments{
\item{experimental_design}{(\strong{required}) Defines what the experiment looks
like, e.g. \code{cv(bt(fs,20)+mb,3,2)+ev} for 2 times repeated 3-fold
cross-validation with nested variable importance computation on 20
bootstraps and model-building, and external validation. The basic workflow
components are:
\itemize{
\item \code{fs}: (optional) variable importance computation step. If not explicitly
declared, feature selection will be done just in time for hyperparameter
optimisation.
\item \code{mb}: (required) model building step.
\item \code{ev}: (optional) external validation. Note that internal validation due
to subsampling will always be conducted if the subsampling methods create
any validation data sets.
}

The different components are linked using \code{+}.

Different subsampling methods can be used in conjunction with the basic
workflow components:
\itemize{
\item \code{bs(x,n)}: (stratified) .632 bootstrap, with \code{n} the number of
bootstraps. In contrast to \code{bt}, feature pre-processing parameters and
hyperparameter optimisation are conducted on individual bootstraps.
\item \code{bt(x,n)}: (stratified) .632 bootstrap, with \code{n} the number of
bootstraps. Unlike \code{bs} and other subsampling methods, no separate
pre-processing parameters or optimised hyperparameters will be determined
for each bootstrap.
\item \code{cv(x,n,p)}: (stratified) \code{n}-fold cross-validation, repeated \code{p} times.
Pre-processing parameters are determined for each iteration.
\item \code{lv(x)}: leave-one-out-cross-validation. Pre-processing parameters are
determined for each iteration.
\item \code{ip(x)}: imbalance partitioning for addressing class imbalances on the
data set. Pre-processing parameters are determined for each partition. The
number of partitions generated depends on the imbalance correction method
(see the \code{imbalance_correction_method} parameter). Imbalance partitioning
does not generate validation sets.
}

As shown in the example above, sampling algorithms can be nested.

The simplest valid experimental design is \code{fs+mb}, which corresponds to a
TRIPOD type 1a analysis. Type 1b analyses are only possible using
bootstraps, e.g. \code{bt(fs+mb,100)}. Type 2a analyses can be conducted using
cross-validation, e.g. \code{cv(bt(fs,100)+mb,10,1)}. Depending on the origin of
the external validation data, designs such as \code{fs+mb+ev} or
\code{cv(bt(fs,100)+mb,10,1)+ev} constitute type 2b or type 3 analyses. Type 4
analyses can be done by obtaining one or more \code{familiarModel} objects from
others and applying them to your own data set.

Alternatively, the \code{experimental_design} parameter may be used to provide a
path to a file containing iterations, which is named \verb{####_iterations.RDS}
by convention. This path can be relative to the directory of the current
experiment (\code{experiment_dir}), or an absolute path. The absolute path may
thus also point to a file from a different experiment.}

\item{message_indent}{Spacing inserted before messages.}

\item{verbose}{Sets verbosity.}
}
\value{
data.table with subsampler information at different levels of the
experimental design.
}
\description{
Parse experimental design
}
\details{
This function converts the experimental_design string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationSampleSimilarity.R
\name{export_sample_similarity}
\alias{export_sample_similarity}
\alias{export_sample_similarity,familiarCollection-method}
\alias{export_sample_similarity,ANY-method}
\title{Extract and export mutual correlation between features.}
\usage{
export_sample_similarity(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  sample_limit = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  export_dendrogram = FALSE,
  export_collection = FALSE,
  ...
)

\S4method{export_sample_similarity}{familiarCollection}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  sample_limit = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  export_dendrogram = FALSE,
  export_collection = FALSE,
  ...
)

\S4method{export_sample_similarity}{ANY}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  sample_limit = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  export_dendrogram = FALSE,
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{aggregate_results}{Flag that signifies whether results should be
aggregated for export.}

\item{sample_limit}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}

\item{sample_cluster_method}{The method used to perform clustering based on
distance between samples. These are the same methods as for the
\code{cluster_method} configuration parameter: \code{hclust}, \code{agnes}, \code{diana} and
\code{pam}.

\code{none} cannot be used when extracting data for feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{export_dendrogram}{Add dendrogram in the data element objects.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[=as_data_object]{as_data_object}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{data}}{A \code{data.frame} or \code{data.table}, a path to such tables on a local
or network drive, or a path to tabular data that may be converted to these
formats.}
  }}
}
\value{
A list containing a data.table (if \code{dir_path} is not provided), or
nothing, as all data is exported to \code{csv} files.
}
\description{
Extract and export mutual correlation between features in a
familiarCollection.
}
\details{
All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Sample similarity data can be created from \code{dataObject}, or \code{data.table} objects.
For \code{data.table}, see \code{\link{as_data_object}} for additional arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFeatureRanking.R
\name{plot_variable_importance}
\alias{plot_variable_importance}
\alias{plot_variable_importance,ANY-method}
\alias{plot_variable_importance,familiarCollection-method}
\alias{plot_feature_selection_occurrence}
\alias{plot_feature_selection_variable_importance}
\alias{plot_model_signature_occurrence}
\alias{plot_model_signature_variable_importance}
\title{Plot variable importance scores of features during feature selection
or after training a model.}
\usage{
plot_variable_importance(
  object,
  type,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  show_cluster = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = waiver(),
  x_label = "feature",
  rotate_x_tick_labels = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  y_range = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_variable_importance}{ANY}(
  object,
  type,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  show_cluster = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = waiver(),
  x_label = "feature",
  rotate_x_tick_labels = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  y_range = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_variable_importance}{familiarCollection}(
  object,
  type,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  show_cluster = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = waiver(),
  x_label = "feature",
  rotate_x_tick_labels = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  y_range = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

plot_feature_selection_occurrence(...)

plot_feature_selection_variable_importance(...)

plot_model_signature_occurrence(...)

plot_model_signature_variable_importance(...)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{type}{Determine what variable importance should be shown. Can be
\code{feature_selection} or \code{model} for the variable importance after the
feature selection step and after the model training step, respectively.}

\item{feature_cluster_method}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_cluster_cut_method}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_threshold}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{aggregation_method}{(\emph{optional}) The method used to aggregate variable
importances over different data subsets, e.g. bootstraps. The following
methods can be selected:
\itemize{
\item \code{mean} (default): Use the mean rank of a feature over the subsets to
determine the aggregated feature rank.
\item \code{median}: Use the median rank of a feature over the subsets to determine
the aggregated feature rank.
\item \code{best}: Use the best rank the feature obtained in any subset to determine
the aggregated feature rank.
\item \code{worst}: Use the worst rank the feature obtained in any subset to
determine the aggregated feature rank.
\item \code{stability}: Use the frequency of the feature being in the subset of
highly ranked features as measure for the aggregated feature rank
(Meinshausen and Buehlmann, 2010).
\item \code{exponential}: Use a rank-weighted frequence of occurrence in the subset
of highly ranked features as measure for the aggregated feature rank (Haury
et al., 2011).
\item \code{borda}: Use the borda count as measure for the aggregated feature rank
(Wald et al., 2012).
\item \code{enhanced_borda}: Use an occurrence frequency-weighted borda count as
measure for the aggregated feature rank (Wald et al., 2012).
\item \code{truncated_borda}: Use borda count computed only on features within the
subset of highly ranked features.
\item \code{enhanced_truncated_borda}: Apply both the enhanced borda method and the
truncated borda method and use the resulting borda count as the aggregated
feature rank.
}}

\item{rank_threshold}{(\emph{optional}) The threshold used to define the subset of
highly important features. If not set, this threshold is determined by
maximising the variance in the occurrence value over all features over the
subset size.

This parameter is only relevant for \code{stability}, \code{exponential},
\code{enhanced_borda}, \code{truncated_borda} and \code{enhanced_truncated_borda} methods.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created figures are
saved to. Output is saved in the \code{variable_importance} subdirectory. If
\code{NULL} no figures are saved, but are returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{show_cluster}{(\emph{optional}) Show which features were clustered together.
Currently not available in combination with variable importance obtained
during feature selection.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring the plot elements
according to the groupings indicated by the \code{color_by} argument (if any).
\code{familiar} has a default palette. Other palettes are supported by
\code{paletteer}, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{gradient_palette}{(\emph{optional}) Palette for filling bars if the
\code{color_by} argument is not set. By default, bars are not coloured. If
\code{gradient_palette} is set, the palette will colour bars according to
feature importance. Use \code{NULL} to fill the bars using the \code{familiar}
default palette. Other palettes are supported by the \code{paletteer} package,
\code{grDevices::palette.pals()} (requires R >= 4.0.0), \code{grDevices::hcl.pals()}
(requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors}, \code{terrain.colors},
\code{topo.colors} and \code{cm.colors}, which correspond to the palettes of the same
name in \code{grDevices}. You may also specify your own palette by providing a
vector of colour names listed by \code{grDevices::colors()} or through
hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{rotate_x_tick_labels}{(\emph{optional}) Rotate tick labels on the x-axis by
90 degrees. Defaults to \code{TRUE}. Rotation of x-axis tick labels may also be
controlled through the \code{ggtheme}. In this case, \code{FALSE} should be provided
explicitly.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{limit_n_features}{(\emph{optional}) The number of features that should be
included in the plot. Only the most important features are shown. By
default, the number of features is not limited.}

\item{y_range}{(\emph{optional}) Value range for the y-axis.}

\item{y_n_breaks}{(\emph{optional}) Number of breaks to show on the y-axis of the
plot. \code{y_n_breaks} is used to determine the \code{y_breaks} argument in case it
is unset.}

\item{y_breaks}{(\emph{optional}) Break points on the y-axis of the plot.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets and the number of features.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from number of facets, and the length of the longest feature name (if
\code{rotate_x_tick_labels} is \code{TRUE}).}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}, \code{\link[=extract_fs_vimp]{extract_fs_vimp}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This function plots variable importance based data obtained
during feature selection or after training a model, which are stored in a
\code{familiarCollection} object.
}
\details{
This function generates a barplot based on variable importance of
features.

The only allowed values for \code{split_by}, \code{color_by} or \code{facet_by} are
\code{vimp_method} and \code{learner}, but note that \code{learner} has no effect when
plotting variable importance of features acquired during feature selection.

Labeling methods such as \code{set_feature_names} or \code{set_vimp_method_names} can
be applied to the \code{familiarCollection} object to update labels, and order
the output in the figure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseSettings.R
\name{.parse_general_settings}
\alias{.parse_general_settings}
\title{Internal function for parsing settings that configure various aspects of the
worklow}
\usage{
.parse_general_settings(settings, config = NULL, data, ...)
}
\arguments{
\item{settings}{List of settings that was previously generated by
\code{.parse_initial_settings}.}

\item{config}{A list of settings, e.g. from an xml file.}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{...}{
  Arguments passed on to \code{\link[=.parse_setup_settings]{.parse_setup_settings}}, \code{\link[=.parse_preprocessing_settings]{.parse_preprocessing_settings}}, \code{\link[=.parse_variable_importance_settings]{.parse_variable_importance_settings}}, \code{\link[=.parse_model_development_settings]{.parse_model_development_settings}}, \code{\link[=.parse_hyperparameter_optimisation_settings]{.parse_hyperparameter_optimisation_settings}}, \code{\link[=.parse_evaluation_settings]{.parse_evaluation_settings}}
  \describe{
    \item{\code{parallel}}{(\emph{optional}) Enable parallel processing. Defaults to \code{TRUE}.
When set to \code{FALSE}, this disables all parallel processing, regardless of
specific parameters such as \code{parallel_preprocessing}. However, when
\code{parallel} is \code{TRUE}, parallel processing of different parts of the
workflow can be disabled by setting respective flags to \code{FALSE}.}
    \item{\code{parallel_nr_cores}}{(\emph{optional}) Number of cores available for
parallelisation. Defaults to 2. This setting does nothing if
parallelisation is disabled.}
    \item{\code{restart_cluster}}{(\emph{optional}) Restart nodes used for parallel computing
to free up memory prior to starting a parallel process. Note that it does
take time to set up the clusters. Therefore setting this argument to \code{TRUE}
may impact processing speed. This argument is ignored if \code{parallel} is
\code{FALSE} or the cluster was initialised outside of familiar. Default is
\code{FALSE}, which causes the clusters to be initialised only once.}
    \item{\code{cluster_type}}{(\emph{optional}) Selection of the cluster type for parallel
processing. Available types are the ones supported by the parallel package
that is part of the base R distribution: \code{psock} (default), \code{fork}, \code{mpi},
\code{nws}, \code{sock}. In addition, \code{none} is available, which also disables
parallel processing.}
    \item{\code{backend_type}}{(\emph{optional}) Selection of the backend for distributing
copies of the data. This backend ensures that only a single master copy is
kept in memory. This limits memory usage during parallel processing.

Several backend options are available, notably \code{socket_server}, and \code{none}
(default). \code{socket_server} is based on the callr package and R sockets,
comes with \code{familiar} and is available for any OS. \code{none} uses the package
environment of familiar to store data, and is available for any OS.
However, \code{none} requires copying of data to any parallel process, and has a
larger memory footprint.}
    \item{\code{server_port}}{(\emph{optional}) Integer indicating the port on which the
socket server or RServe process should communicate. Defaults to port 6311.
Note that ports 0 to 1024 and 49152 to 65535 cannot be used.}
    \item{\code{feature_max_fraction_missing}}{(\emph{optional}) Numeric value between \code{0.0}
and \code{0.95} that determines the meximum fraction of missing values that
still allows a feature to be included in the data set. All features with a
missing value fraction over this threshold are not processed further. The
default value is \code{0.30}.}
    \item{\code{sample_max_fraction_missing}}{(\emph{optional}) Numeric value between \code{0.0}
and \code{0.95} that determines the maximum fraction of missing values that
still allows a sample to be included in the data set. All samples with a
missing value fraction over this threshold are excluded and not processed
further. The default value is \code{0.30}.}
    \item{\code{filter_method}}{(\emph{optional}) One or methods used to reduce
dimensionality of the data set by removing irrelevant or poorly
reproducible features.

Several method are available:
\itemize{
\item \code{none} (default): None of the features will be filtered.
\item \code{low_variance}: Features with a variance below the
\code{low_var_minimum_variance_threshold} are filtered. This can be useful to
filter, for example, genes that are not differentially expressed.
\item \code{univariate_test}: Features undergo a univariate regression using an
outcome-appropriate regression model. The p-value of the model coefficient
is collected. Features with coefficient p-value above the
\code{univariate_test_threshold} are subsequently filtered.
\item \code{robustness}: Features that are not sufficiently robust according to the
intraclass correlation coefficient are filtered. Use of this method
requires that repeated measurements are present in the data set, i.e. there
should be entries for which the sample and cohort identifiers are the same.
}

More than one method can be used simultaneously. Features with singular
values are always filtered, as these do not contain information.}
    \item{\code{univariate_test_threshold}}{(\emph{optional}) Numeric value between \code{1.0} and
\code{0.0} that determines which features are irrelevant and will be filtered by
the \code{univariate_test}. p-values are compared to this threshold.
All features with values above the threshold are filtered. The default
value is \code{0.20}.}
    \item{\code{univariate_test_threshold_metric}}{(\emph{optional}) Metric used with the to
compare the \code{univariate_test_threshold} against. The following metric can
be chosen:
\itemize{
\item \code{p_value} (default): The unadjusted p-value of each feature is used for
to filter features.
}}
    \item{\code{univariate_test_max_feature_set_size}}{(\emph{optional}) Maximum size of the
feature set after the univariate test. P or q values of features are
compared against the threshold, but if the resulting data set would be
larger than this setting, only the most relevant features up to the desired
feature set size are selected.

The default value is \code{NULL}, which causes features to be filtered based on
their relevance only.}
    \item{\code{low_var_minimum_variance_threshold}}{(required, if used) Numeric value
that determines which features will be filtered by the \code{low_variance}
method. The variance of each feature is computed and compared to the
threshold. If it is below the threshold, the feature is removed.

This parameter has no default value and should be set if \code{low_variance} is
used.}
    \item{\code{low_var_max_feature_set_size}}{(\emph{optional}) Maximum size of the feature
set after filtering features with a low variance. All features are first
compared against \code{low_var_minimum_variance_threshold}. If the resulting
feature set would be larger than specified, only the most strongly varying
features will be selected, up to the desired size of the feature set.

The default value is \code{NULL}, which causes features to be filtered based on
their variance only.}
    \item{\code{robustness_icc_type}}{(\emph{optional}) String indicating the type of
intraclass correlation coefficient (\code{1}, \code{2} or \code{3}) that should be used to
compute robustness for features in repeated measurements. These types
correspond to the types in Shrout and Fleiss (1979). The default value is
\code{1}.}
    \item{\code{robustness_threshold_metric}}{(\emph{optional}) String indicating which
specific intraclass correlation coefficient (ICC) metric should be used to
filter features. This should be one of:
\itemize{
\item \code{icc}: The estimated ICC value itself.
\item \code{icc_low} (default): The estimated lower limit of the 95\% confidence
interval of the ICC, as suggested by Koo and Li (2016).
\item \code{icc_panel}: The estimated ICC value over the panel average, i.e. the ICC
that would be obtained if all repeated measurements were averaged.
\item \code{icc_panel_low}: The estimated lower limit of the 95\% confidence interval
of the panel ICC.
}}
    \item{\code{robustness_threshold_value}}{(\emph{optional}) The intraclass correlation
coefficient value that is as threshold. The default value is \code{0.70}.}
    \item{\code{transformation_method}}{(\emph{optional}) The transformation method used to
change the distribution of the data to be more normal-like. The following
methods are available:
\itemize{
\item \code{none}: This disables transformation of features.
\item \code{yeo_johnson}: Transformation using the location and scale invariant
version of the Yeo-Johnson transformation (Yeo and Johnson, 2000;
Zwanenburg and Löck, 2026).
\item \code{yeo_johnson_robust} (default): A robust version of \code{yeo_johnson}.
This method is less sensitive to outliers.
\item \code{yeo_johnson_conventional}: As \code{yeo_johnson}, but without optimisation of
location and scale parameters. This method is equivalent to the original
transformation proposed by Yeo and Johnson (2001).
\item \code{box_cox}: Transformation using the location and scale invariant version
of the Box-Cox transformation (Box and Cox, 1964; Zwanenburg and Löck,
2026).
\item \code{box_cox_robust}: A robust version of \code{yeo_johnson}. This method is less
sensitive to outliers.
\item \code{box_cox_conventional}: As \code{box_cox}, but without optimisation of
location and scale parameters. This method is equivalent to the original
transformation proposed by Box and Cox (1964). This method requires
strictly positive feature values.
}

Transformation requires the \code{power.transform} package. Only features that
contain numerical data are transformed. Transformation parameters obtained
in development data are stored within \code{featureInfo} objects for later use
with validation data sets.}
    \item{\code{transformation_optimisation_criterion}}{(\emph{optional}) Transformation
parameters are optimised using a criterion, conventionally
maximum-likelihood-estimation. \code{power.transform} implements multiple
optimisation criteria, of which the following are available:
\itemize{
\item \code{mle} (default): Optimisation using maximum likelihood estimation.
\item \code{cramer_von_mises}: Optimisation using the Cramér-von Mises
criterion. Zwanenburg and Löck (2026) found that this criterion was
relatively robust against outliers.
}}
    \item{\code{transformation_gof_test_p_value}}{(\emph{optional}) Not all transformations
will lead to features that are roughly normally distributed. Zwanenburg and
Löck (2026) established a empirical goodness-of-fit test for central
normality. This parameter sets the significance for rejecting the
null-hypothesis that a feature distribution is centrally normal. When the
null-hypothesis is rejected, no transformation is performed. The default
value is \code{NULL}, which disables the test.}
    \item{\code{normalisation_method}}{(\emph{optional}) The normalisation method used to
improve the comparability between numerical features that may have very
different scales. The following normalisation methods can be chosen:
\itemize{
\item \code{none}: This disables feature normalisation.
\item \code{standardisation}: Features are normalised by subtraction of their mean
values and division by their standard deviations. This causes every feature
to be have a center value of 0.0 and standard deviation of 1.0.
\item \code{standardisation_trim}: As \code{standardisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are discarded.
This reduces the effect of outliers.
\item \code{standardisation_winsor}: As \code{standardisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are winsorised.
This reduces the effect of outliers.
\item \code{standardisation_robust} (default): A robust version of \code{standardisation}
that relies on computing Huber's M-estimators for location and scale.
\item \code{normalisation}: Features are normalised by subtraction of their minimum
values and division by their ranges. This maps all feature values to a
\eqn{[0, 1]} interval.
\item \code{normalisation_trim}: As \code{normalisation}, but based on the set of feature
values where the 5\% lowest and 5\% highest values are discarded. This
reduces the effect of outliers.
\item \code{normalisation_winsor}: As \code{normalisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are winsorised.
This reduces the effect of outliers.
\item \code{quantile}: Features are normalised by subtraction of their median values
and division by their interquartile range.
\item \code{mean_centering}: Features are centered by substracting the mean, but do
not undergo rescaling.
}

Only features that contain numerical data are normalised. Normalisation
parameters obtained in development data are stored within \code{featureInfo}
objects for later use with validation data sets.}
    \item{\code{batch_normalisation_method}}{(\emph{optional}) The method used for batch
normalisation. Available methods are:
\itemize{
\item \code{none} (default): This disables batch normalisation of features.
\item \code{standardisation}: Features within each batch are normalised by
subtraction of the mean value and division by the standard deviation in
each batch.
\item \code{standardisation_trim}: As \code{standardisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are discarded.
This reduces the effect of outliers.
\item \code{standardisation_winsor}: As \code{standardisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are winsorised.
This reduces the effect of outliers.
\item \code{standardisation_robust}: A robust version of \code{standardisation} that
relies on computing Huber's M-estimators for location and scale within each
batch.
\item \code{normalisation}: Features within each batch are normalised by subtraction
of their minimum values and division by their range in each batch. This
maps all feature values in each batch to a \eqn{[0, 1]} interval.
\item \code{normalisation_trim}: As \code{normalisation}, but based on the set of feature
values where the 5\% lowest and 5\% highest values are discarded. This
reduces the effect of outliers.
\item \code{normalisation_winsor}: As \code{normalisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are winsorised.
This reduces the effect of outliers.
\item \code{quantile}: Features in each batch are normalised by subtraction of the
median value and division by the interquartile range of each batch.
\item \code{mean_centering}: Features in each batch are centered on 0.0 by
substracting the mean value in each batch, but are not rescaled.
\item \code{combat_parametric}: Batch adjustments using parametric empirical Bayes
(Johnson et al, 2007). \code{combat_p} leads to the same method.
\item \code{combat_non_parametric}: Batch adjustments using non-parametric empirical
Bayes (Johnson et al, 2007). \code{combat_np} and \code{combat} lead to the same
method. Note that we reduced complexity from O(\eqn{n^2}) to O(\eqn{n}) by
only computing batch adjustment parameters for each feature on a subset of
50 randomly selected features, instead of all features.
}

Only features that contain numerical data are normalised using batch
normalisation. Batch normalisation parameters obtained in development data
are stored within \code{featureInfo} objects for later use with validation data
sets, in case the validation data is from the same batch.

If validation data contains data from unknown batches, normalisation
parameters are separately determined for these batches.

Note that for both empirical Bayes methods, the batch effect is assumed to
produce results across the features. This is often true for things such as
gene expressions, but the assumption may not hold generally.

When performing batch normalisation, it is moreover important to check that
differences between batches or cohorts are not related to the studied
endpoint.}
    \item{\code{imputation_method}}{(\emph{optional}) Method used for imputing missing
feature values. Two methods are implemented:
\itemize{
\item \code{simple}: Simple replacement of a missing value by the median value (for
numeric features) or the modal value (for categorical features).
\item \code{lasso}: Imputation of missing value by lasso regression (using \code{glmnet})
based on information contained in other features.
}

\code{simple} imputation precedes \code{lasso} imputation to ensure that any missing
values in predictors required for \code{lasso} regression are resolved. The
\code{lasso} estimate is then used to replace the missing value.

The default value depends on the number of features in the dataset. If the
number is lower than 100, \code{lasso} is used by default, and \code{simple}
otherwise.

Only single imputation is performed. Imputation models and parameters are
stored within \code{featureInfo} objects for later use with validation data
sets.}
    \item{\code{cluster_method}}{(\emph{optional}) Clustering is performed to identify and
replace redundant features, for example those that are highly correlated.
Such features do not carry much additional information and may be removed
or replaced instead (Park et al., 2007; Tolosi and Lengauer, 2011).

The cluster method determines the algorithm used to form the clusters. The
following cluster methods are implemented:
\itemize{
\item \code{none}: No clustering is performed.
\item \code{hclust} (default): Hierarchical agglomerative clustering. If the
\code{fastcluster} package is installed, \code{fastcluster::hclust} is used (Muellner
2013), otherwise \code{stats::hclust} is used.
\item \code{agnes}: Hierarchical clustering using agglomerative nesting (Kaufman and
Rousseeuw, 1990). This algorithm is similar to \code{hclust}, but uses the
\code{cluster::agnes} implementation.
\item \code{diana}: Divisive analysis hierarchical clustering. This method uses
divisive instead of agglomerative clustering (Kaufman and Rousseeuw, 1990).
\code{cluster::diana} is used.
\item \code{pam}: Partioning around medioids. This partitions the data into $k$
clusters around medioids (Kaufman and Rousseeuw, 1990). $k$ is selected
using the \code{silhouette} metric. \code{pam} is implemented using the
\code{cluster::pam} function.
}

Clusters and cluster information is stored within \code{featureInfo} objects for
later use with validation data sets. This enables reproduction of the same
clusters as formed in the development data set.}
    \item{\code{cluster_linkage_method}}{(\emph{optional}) Linkage method used for
agglomerative clustering in \code{hclust} and \code{agnes}. The following linkage
methods can be used:
\itemize{
\item \code{average} (default): Average linkage.
\item \code{single}: Single linkage.
\item \code{complete}: Complete linkage.
\item \code{weighted}: Weighted linkage, also known as McQuitty linkage.
\item \code{ward}: Linkage using Ward's minimum variance method.
}

\code{diana} and \code{pam} do not require a linkage method.}
    \item{\code{cluster_cut_method}}{(\emph{optional}) The method used to define the actual
clusters. The following methods can be used:
\itemize{
\item \code{silhouette}: Clusters are formed based on the silhouette score
(Rousseeuw, 1987). The average silhouette score is computed from 2 to
\eqn{n} clusters, with \eqn{n} the number of features. Clusters are only
formed if the average silhouette exceeds 0.50, which indicates reasonable
evidence for structure. This procedure may be slow if the number of
features is large (>100s).
\item \code{fixed_cut}: Clusters are formed by cutting the hierarchical tree at the
point indicated by the \code{cluster_similarity_threshold}, e.g. where features
in a cluster have an average Spearman correlation of 0.90. \code{fixed_cut} is
only available for \code{agnes}, \code{diana} and \code{hclust}.
\item \code{dynamic_cut}: Dynamic cluster formation using the cutting algorithm in
the \code{dynamicTreeCut} package. This package should be installed to select
this option. \code{dynamic_cut} can only be used with \code{agnes} and \code{hclust}.
}

The default options are \code{silhouette} for partioning around medioids (\code{pam})
and \code{fixed_cut} otherwise.}
    \item{\code{cluster_similarity_metric}}{(\emph{optional}) Clusters are formed based on
feature similarity. All features are compared in a pair-wise fashion to
compute similarity, for example correlation. The resulting similarity grid
is converted into a distance matrix that is subsequently used for
clustering. The following metrics are supported to compute pairwise
similarities:
\itemize{
\item \code{mutual_information} (default): normalised mutual information.
\item \code{mcfadden_r2}: McFadden's pseudo R-squared (McFadden, 1974).
\item \code{cox_snell_r2}: Cox and Snell's pseudo R-squared (Cox and Snell, 1989).
\item \code{nagelkerke_r2}: Nagelkerke's pseudo R-squared (Nagelkerke, 1991).
\item \code{spearman}: Spearman's rank order correlation.
\item \code{kendall}: Kendall rank correlation.
\item \code{pearson}: Pearson product-moment correlation.
}

The pseudo R-squared metrics can be used to assess similarity between mixed
pairs of numeric and categorical features, as these are based on the
log-likelihood of regression models. In \code{familiar}, the more informative
feature is used as the predictor and the other feature as the reponse
variable. In numeric-categorical pairs, the numeric feature is considered
to be more informative and is thus used as the predictor. In
categorical-categorical pairs, the feature with most levels is used as the
predictor.

In case any of the classical correlation coefficients (\code{pearson},
\code{spearman} and \code{kendall}) are used with (mixed) categorical features, the
categorical features are one-hot encoded and the mean correlation over all
resulting pairs is used as similarity.}
    \item{\code{cluster_similarity_threshold}}{(\emph{optional}) The threshold level for
pair-wise similarity that is required to form clusters using \code{fixed_cut}.
This should be a numerical value between 0.0 and 1.0. Note however, that a
reasonable threshold value depends strongly on the similarity metric. The
following are the default values used:
\itemize{
\item \code{mcfadden_r2} and \code{mutual_information}: \code{0.30}
\item \code{cox_snell_r2} and \code{nagelkerke_r2}: \code{0.75}
\item \code{spearman}, \code{kendall} and \code{pearson}: \code{0.90}
}

Alternatively, if the \verb{fixed cut} method is not used, this value determines
whether any clustering should be performed, because the data may not
contain highly similar features. The default values in this situation are:
\itemize{
\item \code{mcfadden_r2}  and \code{mutual_information}: \code{0.25}
\item \code{cox_snell_r2} and \code{nagelkerke_r2}: \code{0.40}
\item \code{spearman}, \code{kendall} and \code{pearson}: \code{0.70}
}

The threshold value is converted to a distance (1-similarity) prior to
cutting hierarchical trees.}
    \item{\code{cluster_representation_method}}{(\emph{optional}) Method used to determine
how the information of co-clustered features is summarised and used to
represent the cluster. The following methods can be selected:
\itemize{
\item \code{best_predictor} (default): The feature with the highest importance
according to univariate regression with the outcome is used to represent
the cluster.
\item \code{medioid}: The feature closest to the cluster center, i.e. the feature
that is most similar to the remaining features in the cluster, is used to
represent the feature.
\item \code{mean}: A meta-feature is generated by averaging the feature values for
all features in a cluster. This method aligns all features so that all
features will be positively correlated prior to averaging. Should a cluster
contain one or more categorical features, the \code{medioid} method will be used
instead, as averaging is not possible. Note that if this method is chosen,
the \code{normalisation_method} parameter should be one of \code{standardisation},
\code{standardisation_trim}, \code{standardisation_winsor} or \code{quantile}.`
}

If the \code{pam} cluster method is selected, only the \code{medioid} method can be
used. In that case 1 medioid is used by default.}
    \item{\code{parallel_preprocessing}}{(\emph{optional}) Enable parallel processing for the
preprocessing workflow. Defaults to \code{TRUE}. When set to \code{FALSE}, this will
disable the use of parallel processing while preprocessing, regardless of
the settings of the \code{parallel} parameter. \code{parallel_preprocessing} is
ignored if \code{parallel=FALSE}.}
    \item{\code{vimp_method}}{(\strong{required}) Variable importance method. \code{familiar}
implements various variable importance methods. Please refer to the
vignette on variable importance methods for more details.

More than one variable importance method can be chosen. The experiment will
then repeated for each feature selection method.

Variable importance methods determine the ranking of features. Actual
selection of features is done by optimising the signature size model
hyperparameter during the hyperparameter optimisation step.}
    \item{\code{vimp_method_parameter}}{(\emph{optional}) List of lists containing parameters
for feature selection methods. Each sublist should have the name of the
feature selection method it corresponds to.

Most feature selection methods do not have parameters that can be set.
Please refer to the vignette on feature selection methods for more details.
Note that if the feature selection method is based on a learner (e.g. lasso
regression), hyperparameter optimisation may be performed prior to
assessing variable importance.}
    \item{\code{vimp_aggregation_method}}{(\emph{optional}) The method used to aggregate
variable importances over different data subsets, e.g. bootstraps. The
following methods can be selected:
\itemize{
\item \code{none}: Don't aggregate ranks, but rather aggregate the variable
importance scores themselves.
\item \code{mean}: Use the mean rank of a feature over the subsets to
determine the aggregated feature rank.
\item \code{median}: Use the median rank of a feature over the subsets to determine
the aggregated feature rank.
\item \code{best}: Use the best rank the feature obtained in any subset to determine
the aggregated feature rank.
\item \code{worst}: Use the worst rank the feature obtained in any subset to
determine the aggregated feature rank.
\item \code{stability}: Use the frequency of the feature being in the subset of
highly ranked features as measure for the aggregated feature rank
(Meinshausen and Buehlmann, 2010).
\item \code{exponential}: Use a rank-weighted frequence of occurrence in the subset
of highly ranked features as measure for the aggregated feature rank (Haury
et al., 2011).
\item \code{borda} (default): Use the borda count as measure for the aggregated
feature rank (Wald et al., 2012).
\item \code{enhanced_borda}: Use an occurrence frequency-weighted borda count as
measure for the aggregated feature rank (Wald et al., 2012).
\item \code{truncated_borda}: Use borda count computed only on features within the
subset of highly ranked features.
\item \code{enhanced_truncated_borda}: Apply both the enhanced borda method and the
truncated borda method and use the resulting borda count as the aggregated
feature rank.
}

The \emph{feature selection methods} vignette provides additional information.}
    \item{\code{vimp_aggregation_rank_threshold}}{(\emph{optional}) The threshold used to
define the subset of highly important features. If set to \code{NULL}, this
threshold is determined by maximising the variance in the occurrence value
over all features over the subset size. The default value is \code{5}.

This parameter is only relevant for \code{stability}, \code{exponential},
\code{enhanced_borda}, \code{truncated_borda} and \code{enhanced_truncated_borda} methods.}
    \item{\code{parallel_vimp}}{(\emph{optional}) Enable parallel processing for
variable importance tasks. Defaults to \code{TRUE}. When set to \code{FALSE},
this will disable the use of parallel processing while performing feature
selection, regardless of the settings of the \code{parallel} parameter.
\code{parallel_vimp} is ignored if \code{parallel = FALSE}.}
    \item{\code{learner}}{(\strong{required}) One or more algorithms used for model
development. A sizeable number learners is supported in \code{familiar}. Please
see the vignette on learners for more information concerning the available
learners.}
    \item{\code{hyperparameter}}{(\emph{optional}) List of lists containing hyperparameters
for learners. Each sublist should have the name of the learner method it
corresponds to, with list elements being named after the intended
hyperparameter, e.g. \code{"glm_logistic"=list("sign_size"=3)}

All learners have hyperparameters. Please refer to the vignette on learners
for more details. If no parameters are provided, sequential model-based
optimisation is used to determine optimal hyperparameters.

Hyperparameters provided by the user are never optimised. However, if more
than one value is provided for a single hyperparameter, optimisation will
be conducted using these values.}
    \item{\code{novelty_detector}}{(\emph{optional}) Specify the algorithm used for training
a novelty detector. This detector can be used to identify
out-of-distribution data prospectively.}
    \item{\code{detector_parameters}}{(\emph{optional}) List lists containing hyperparameters
for novelty detectors. Currently not used.}
    \item{\code{parallel_model_development}}{(\emph{optional}) Enable parallel processing for
the model development workflow. Defaults to \code{TRUE}. When set to \code{FALSE},
this will disable the use of parallel processing while developing models,
regardless of the settings of the \code{parallel} parameter.
\code{parallel_model_development} is ignored if \code{parallel=FALSE}.}
    \item{\code{optimisation_bootstraps}}{(\emph{optional}) Number of bootstraps that should
be generated from the development data set. During the optimisation
procedure one or more of these bootstraps (indicated by
\code{smbo_step_bootstraps}) are used for model development using different
combinations of hyperparameters. The effect of the hyperparameters is then
assessed by comparing in-bag and out-of-bag model performance.

The default number of bootstraps is \code{50}. Hyperparameter optimisation may
finish before exhausting the set of bootstraps.}
    \item{\code{optimisation_determine_vimp}}{(\emph{optional}) Logical value that indicates
whether variable importance is determined separately for each of the
bootstraps created during the optimisation process (\code{TRUE}) or the
applicable results from the variable importance computation step are used
(\code{FALSE}).

Determining variable importance increases the initial computational
overhead. However, it prevents positive biases for the out-of-bag data due
to overlap of these data with the development data set used for the feature
selection step. In this case, any hyperparameters of the variable
importance method are not determined separately for each bootstrap, but
those obtained during the variable importance computation step are used
instead. In case multiple of such hyperparameter sets could be applicable,
the set that will be used is randomly selected for each bootstrap.

This parameter only affects hyperparameter optimisation of learners. The
default is \code{TRUE}.}
    \item{\code{smbo_random_initialisation}}{(\emph{optional}) String indicating the
initialisation method for the hyperparameter space. Can be one of
\code{fixed_subsample} (default), \code{fixed}, or \code{random}. \code{fixed} and
\code{fixed_subsample} first create hyperparameter sets from a range of default
values set by familiar. \code{fixed_subsample} then randomly draws up to
\code{smbo_n_random_sets} from the grid. \code{random} does not rely upon a fixed
grid, and randomly draws up to \code{smbo_n_random_sets} hyperparameter sets
from the hyperparameter space.}
    \item{\code{smbo_n_random_sets}}{(\emph{optional}) Number of random or subsampled
hyperparameters drawn during the initialisation process. Default: \code{100}.
Cannot be smaller than \code{10}. The parameter is not used when
\code{smbo_random_initialisation} is \code{fixed}, as the entire pre-defined grid
will be explored.}
    \item{\code{max_smbo_iterations}}{(\emph{optional}) Maximum number of intensify
iterations of the SMBO algorithm. During an intensify iteration a run-off
occurs between the current \emph{best} hyperparameter combination and either 10
challenger combination with the highest expected improvement or a set of 20
random combinations.

Run-off with random combinations is used to force exploration of the
hyperparameter space, and is performed every second intensify iteration, or
if there is no expected improvement for any challenger combination.

If a combination of hyperparameters leads to better performance on the same
data than the incumbent \emph{best} set of hyperparameters, it replaces the
incumbent set at the end of the intensify iteration.

The default number of intensify iteration is \code{20}. Iterations may be
stopped early if the incumbent set of hyperparameters remains the same for
\code{smbo_stop_convergent_iterations} iterations, or performance improvement is
minimal. This behaviour is suppressed during the first 4 iterations to
enable the algorithm to explore the hyperparameter space.}
    \item{\code{smbo_stop_convergent_iterations}}{(\emph{optional}) The number of subsequent
convergent SMBO iterations required to stop hyperparameter optimisation
early. An iteration is convergent if the \emph{best} parameter set has not
changed or the optimisation score over the 4 most recent iterations has not
changed beyond the tolerance level in \code{smbo_stop_tolerance}.

The default value is \code{3}.}
    \item{\code{smbo_stop_tolerance}}{(\emph{optional}) Tolerance for early stopping due to
convergent optimisation score.

The default value depends on the square root of the number of samples (at
the series level), and is \code{0.01} for 100 samples. This value is computed as
\code{0.1 * 1 / sqrt(n_samples)}. The upper limit is \code{0.0001} for 1M or more
samples.}
    \item{\code{smbo_time_limit}}{(\emph{optional}) Time limit (in minutes) for the
optimisation process. Optimisation is stopped after this limit is exceeded.
Time taken to determine variable importance for the optimisation process
(see the \code{optimisation_determine_vimp} parameter) does not count.

The default is \code{NULL}, indicating that there is no time limit for the
optimisation process. The time limit cannot be less than 1 minute.}
    \item{\code{smbo_initial_bootstraps}}{(\emph{optional}) The number of bootstraps taken
from the set of \code{optimisation_bootstraps} as the bootstraps assessed
initially.

The default value is \code{1}. The value cannot be larger than
\code{optimisation_bootstraps}.}
    \item{\code{smbo_step_bootstraps}}{(\emph{optional}) The number of bootstraps taken from
the set of \code{optimisation_bootstraps} bootstraps as the bootstraps assessed
during the steps of each intensify iteration.

The default value is \code{3}. The value cannot be larger than
\code{optimisation_bootstraps}.}
    \item{\code{smbo_intensify_steps}}{(\emph{optional}) The number of steps in each SMBO
intensify iteration. Each step a new set of \code{smbo_step_bootstraps}
bootstraps is drawn and used in the run-off between the incumbent \emph{best}
hyperparameter combination and its challengers.

The default value is \code{5}. Higher numbers allow for a more detailed
comparison, but this comes with added computational cost.}
    \item{\code{optimisation_metric}}{(\emph{optional}) One or more metrics used to compute
performance scores. See the vignette on performance metrics for the
available metrics.

If unset, the following metrics are used by default:
\itemize{
\item \code{auc_roc}: For \code{binomial} and \code{multinomial} models.
\item \code{mse}: Mean squared error for \code{continuous} models.
\item \code{concordance_index}: For \code{survival} models.
}

Multiple optimisation metrics can be specified. Actual metric values are
converted to an objective value by comparison with a baseline metric value
that derives from a trivial model, i.e. majority class for binomial and
multinomial outcomes, the median outcome for count and continuous outcomes
and a fixed risk or time for survival outcomes.}
    \item{\code{optimisation_function}}{(\emph{optional}) Type of optimisation function used
to quantify the performance of a hyperparameter set. Model performance is
assessed using the metric(s) specified by \code{optimisation_metric} on the
in-bag (IB) and out-of-bag (OOB) samples of a bootstrap. These values are
converted to objective scores with a standardised interval of
\eqn{[-1.0, 1.0]}. Each pair of objective is subsequently used to compute an
optimisation score. The optimisation score across different bootstraps is
than aggregated to a summary score. This summary score is used to rank
hyperparameter sets, and select the optimal set.

The combination of optimisation score and summary score is determined by
the optimisation function indicated by this parameter:
\itemize{
\item \code{validation} or \code{max_validation} (default): seeks to maximise OOB score.
\item \code{balanced}: seeks to balance IB and OOB score.
\item \code{stronger_balance}: similar to \code{balanced}, but with stronger penalty for
differences between IB and OOB scores.
\item \code{validation_minus_sd}: seeks to optimise the average OOB score minus its
standard deviation.
\item \code{validation_25th_percentile}: seeks to optimise the 25th percentile of
OOB scores, and is conceptually similar to \code{validation_minus_sd}.
\item \code{model_estimate}: seeks to maximise the OOB score estimate predicted by
the hyperparameter learner (not available for random search).
\item \code{model_estimate_minus_sd}: seeks to maximise the OOB score estimate minus
its estimated standard deviation, as predicted by the hyperparameter
learner (not available for random search).
\item \code{model_balanced_estimate}: seeks to maximise the estimate of the balanced
IB and OOB score. This is similar to the \code{balanced} score, and in fact uses
a hyperparameter learner to predict said score (not available for random
search).
\item \code{model_balanced_estimate_minus_sd}: seeks to maximise the estimate of the
balanced IB and OOB score, minus its estimated standard deviation. This is
similar to the \code{balanced} score, but takes into account its estimated
spread.
}

Additional detail are provided in the \emph{Learning algorithms and
hyperparameter optimisation} vignette.}
    \item{\code{hyperparameter_learner}}{(\emph{optional}) Any point in the hyperparameter
space has a single, scalar, optimisation score value that is \emph{a priori}
unknown. During the optimisation process, the algorithm samples from the
hyperparameter space by selecting hyperparameter sets and computing the
optimisation score value for one or more bootstraps. For each
hyperparameter set the resulting values are distributed around the actual
value. The learner indicated by \code{hyperparameter_learner} is then used to
infer optimisation score estimates for unsampled parts of the
hyperparameter space.

The following models are available:
\itemize{
\item \code{bayesian_additive_regression_trees} or \code{bart}: Uses Bayesian Additive
Regression Trees (Sparapani et al., 2021) for inference. Unlike standard
random forests, BART allows for estimating posterior distributions directly
and can extrapolate.
\item \code{gaussian_process} (default): Creates a localised approximate Gaussian
process for inference (Gramacy, 2016). This allows for better scaling than
deterministic Gaussian Processes.
\item \code{random_forest}: Creates a random forest for inference. Originally
suggested by Hutter et al. (2011). A weakness of random forests is their
lack of extrapolation beyond observed values, which limits their usefulness
in exploiting promising areas of hyperparameter space.
\item \code{random} or \code{random_search}: Forgoes the use of models to steer
optimisation. Instead, a random search is performed.
}}
    \item{\code{acquisition_function}}{(\emph{optional}) The acquisition function influences
how new hyperparameter sets are selected. The algorithm uses the model
learned by the learner indicated by \code{hyperparameter_learner} to search the
hyperparameter space for hyperparameter sets that are either likely better
than the best known set (\emph{exploitation}) or where there is considerable
uncertainty (\emph{exploration}). The acquisition function quantifies this
(Shahriari et al., 2016).

The following acquisition functions are available, and are described in
more detail in the \emph{learner algorithms} vignette:
\itemize{
\item \code{improvement_probability}: The probability of improvement quantifies the
probability that the expected optimisation score for a set is better than
the best observed optimisation score
\item \code{improvement_empirical_probability}: Similar to
\code{improvement_probability}, but based directly on optimisation scores
predicted by the individual decision trees.
\item \code{expected_improvement} (default): Computes expected improvement.
\item \code{upper_confidence_bound}: This acquisition function is based on the upper
confidence bound of the distribution (Srinivas et al., 2012).
\item \code{bayes_upper_confidence_bound}: This acquisition function is based on the
upper confidence bound of the distribution (Kaufmann et al., 2012).
}}
    \item{\code{exploration_method}}{(\emph{optional}) Method used to steer exploration in
post-initialisation intensive searching steps. As stated earlier, each SMBO
iteration step compares suggested alternative parameter sets with an
incumbent \strong{best} set in a series of steps. The exploration method
controls how the set of alternative parameter sets is pruned after each
step in an iteration. Can be one of the following:
\itemize{
\item \code{single_shot} (default): The set of alternative parameter sets is not
pruned, and each intensification iteration contains only a single
intensification step that only uses a single bootstrap. This is the fastest
exploration method, but only superficially tests each parameter set.
\item \code{successive_halving}: The set of alternative parameter sets is
pruned by removing the worst performing half of the sets after each step
(Jamieson and Talwalkar, 2016).
\item \code{stochastic_reject}: The set of alternative parameter sets is pruned by
comparing the performance of each parameter set with that of the incumbent
\strong{best} parameter set using a paired Wilcoxon test based on shared
bootstraps. Parameter sets that perform significantly worse, at an alpha
level indicated by \code{smbo_stochastic_reject_p_value}, are pruned.
\item \code{none}: The set of alternative parameter sets is not pruned.
}}
    \item{\code{smbo_stochastic_reject_p_value}}{(\emph{optional}) The p-value threshold used
for the \code{stochastic_reject} exploration method.

The default value is \code{0.05}.}
    \item{\code{parallel_hyperparameter_optimisation}}{(\emph{optional}) Enable parallel
processing for hyperparameter optimisation. Defaults to \code{TRUE}. When set to
\code{FALSE}, this will disable the use of parallel processing while performing
optimisation, regardless of the settings of the \code{parallel} parameter. The
parameter moreover specifies whether parallelisation takes place within the
optimisation algorithm (\code{inner}, default), or in an outer loop ( \code{outer})
over learners, data subsamples, etc.

\code{parallel_hyperparameter_optimisation} is ignored if \code{parallel=FALSE}.}
    \item{\code{evaluate_top_level_only}}{(\emph{optional}) Flag that signals that only
evaluation at the most global experiment level is required. Consider a
cross-validation experiment with additional external validation. The global
experiment level consists of data that are used for development, internal
validation and external validation. The next lower experiment level are the
individual cross-validation iterations.

When the flag is \code{true}, evaluations take place on the global level only,
and no results are generated for the next lower experiment levels. In our
example, this means that results from individual cross-validation iterations
are not computed and shown. When the flag is \code{false}, results are computed
from both the global layer and the next lower level.

Setting the flag to \code{true} saves computation time.}
    \item{\code{evaluation_elements}}{(\emph{optional}) Specifies which evaluation steps
should be performed during the evaluation process. Defaults to \code{all},
indicating that all evaluation steps should be performed. It can have one or
more of the following values:
\itemize{
\item \code{all}, \code{true}: all evaluation steps are performed.
\item \code{none}, \code{false}: no evaluation is performed.
\item \code{auc_data}: data for assessing and plotting the area under the receiver
operating characteristic curve are computed.
\item \code{calibration_data}: data for assessing and plotting model calibration are
computed.
\item \code{calibration_info}: data required to assess calibration, such as baseline
survival curves, are collected. These data will still be present in the
models.
\item \code{confusion_matrix}: data for assessing and plotting a confusion matrix are
collected.
\item \code{decision_curve_analyis}: data for performing a decision curve analysis
are computed.
\item \code{feature_expressions}: data for assessing and plotting sample clustering
are computed.
\item \code{feature_similarity}: data for assessing and plotting feature clusters
not computed.
\item \code{fs_vimp}: data for assessing and plotting feature selection-based
variable importance are collected.
\item \code{hyperparameters}: data for assessing model hyperparameters are
collected. These data will still be present in the models.
\item \code{ice_data}: data for individual conditional expectation and partial
dependence plots are created.
\item \code{model_performance}: data for assessing and visualising model performance
are created.
\item \code{model_vimp}: data for assessing and plotting model-based variable
importance are collected.
\item \code{permutation_vimp}: data for assessing and plotting model-agnostic
permutation variable importance are computed.
\item \code{prediction_data}: predictions for each sample are made and exported.
\item \code{risk_stratification_data}: data for assessing and plotting Kaplan-Meier
survival curves are collected.
\item \code{risk_stratification_info}: data for assessing stratification into risk
groups are computed.
\item \code{sample_similarity}: data for assessing sample similarity are computed.
\item \code{shap}: data for assessing marginal contributions of feature values in
a SHAP analysis.
\item \code{univariate_analysis}: data for assessing and plotting univariate feature
importance are computed.
}

The logical intersect of the evaluation elements specified by
\code{evaluation_elements} and \code{skip_evaluation_elements} is used to define which
evaluation steps are performed.}
    \item{\code{skip_evaluation_elements}}{(\emph{optional}) Specifies which evaluation steps,
if any, should be skipped as part of the evaluation process. Defaults to
\code{none}, which means that all relevant evaluation steps are performed. It can
have one or more of the same values as \code{evaluation_elements}.

The logical intersect of the evaluation elements specified by
\code{evaluation_elements} and \code{skip_evaluation_elements} is used to define which
evaluation steps are performed.}
    \item{\code{ensemble_method}}{(\emph{optional}) Method for ensembling predictions from
models for the same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}

This parameter is only used if \code{detail_level} is \code{ensemble}.}
    \item{\code{evaluation_metric}}{(\emph{optional}) One or more metrics for assessing model
performance. See the vignette on performance metrics for the available
metrics.

Confidence intervals (or rather credibility intervals) are computed for each
metric during evaluation. This is done using bootstraps, the number of which
depends on the value of \code{confidence_level} (Davison and Hinkley, 1997).

If unset, the metric in the \code{optimisation_metric} variable is used.}
    \item{\code{sample_limit}}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}
    \item{\code{n_important_features}}{(\emph{optional}) Set the number of features that are
evaluated in evaluation steps. Cannot be 0 or fewer.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("ice_data"=10, "permutation_vimp"=5)}.

This parameter can be set for the following data elements:
\code{ice_data}, \code{permutation_vimp}, and \code{shap}.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
    \item{\code{estimation_type}}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}
    \item{\code{aggregate_results}}{(\emph{optional}) Flag that signifies whether results
should be aggregated during evaluation. If \code{estimation_type} is
\code{bias_correction} or \code{bc}, aggregation leads to a single bias-corrected
estimate. If \code{estimation_type} is \code{bootstrap_confidence_interval} or \code{bci},
aggregation leads to a single bias-corrected estimate with lower and upper
boundaries of the confidence interval. This has no effect if
\code{estimation_type} is \code{point}.

The default value is equal to \code{TRUE} except when assessing metrics to assess
model performance, as the default violin plot requires underlying data.

As with \code{detail_level} and \code{estimation_type}, a non-default
\code{aggregate_results} parameter can be specified for separate evaluation steps
by providing a parameter value in a named list with data elements, e.g.
\code{list("auc_data"=TRUE, , "model_performance"=FALSE)}. This parameter exists
for the same elements as \code{estimation_type}.}
    \item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}
    \item{\code{bootstrap_ci_method}}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}
    \item{\code{shap_tolerance}}{(\emph{optional}) Relative tolerance for convergence of SHAP
values. The tolerance is scaled with the range in SHAP values.

The default value is \code{0.05}.}
    \item{\code{shap_max_iterations}}{(\emph{optional}) Maximum iterations for convergence of
SHAP values.

The default value is \code{1000} iterations.}
    \item{\code{shap_phi_0}}{(\emph{optional}) Value(s) to use for computing marginal
contributions. By default, the average predicted value in the population is
used. For \code{multinomial} and \code{survival} outcomes, this parameter requires a
value for each class and evaluation time (\code{evaluation_times}),
respectively.}
    \item{\code{feature_cluster_method}}{(\emph{optional}) Method used to perform clustering
of features. The same methods as for the \code{cluster_method} configuration
parameter are available: \code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

The value for the \code{cluster_method} configuration parameter is used by
default. When generating clusters for the purpose of determining mutual
correlation and ordering feature expressions, \code{none} is ignored and \code{hclust}
is used instead.}
    \item{\code{feature_linkage_method}}{(\emph{optional}) Method used for agglomerative
clustering with \code{hclust} and \code{agnes}. Linkage determines how features are
sequentially combined into clusters based on distance. The methods are
shared with the \code{cluster_linkage_method} configuration parameter: \code{average},
\code{single}, \code{complete}, \code{weighted}, and \code{ward}.

The value for the \code{cluster_linkage_method} configuration parameters is used
by default.}
    \item{\code{feature_cluster_cut_method}}{(\emph{optional}) Method used to divide features
into separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

The value for the \code{cluster_cut_method} configuration parameter is used by
default.}
    \item{\code{feature_similarity_metric}}{(\emph{optional}) Metric to determine pairwise
similarity between features. Similarity is computed in the same manner as
for clustering, and \code{feature_similarity_metric} therefore has the same
options as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{mutual_information}, \code{spearman}, \code{kendall} and \code{pearson}.

The value used for the \code{cluster_similarity_metric} configuration parameter
is used by default.}
    \item{\code{feature_similarity_threshold}}{(\emph{optional}) The threshold level for
pair-wise similarity that is required to form feature clusters with the
\code{fixed_cut} method. This threshold functions in the same manner as the one
defined using the \code{cluster_similarity_threshold} parameter.

By default, the value for the \code{cluster_similarity_threshold} configuration
parameter is used.

Unlike for \code{cluster_similarity_threshold}, more than one value can be
supplied here.}
    \item{\code{sample_cluster_method}}{(\emph{optional}) The method used to perform
clustering based on distance between samples. These are the same methods as
for the \code{cluster_method} configuration parameter: \code{hclust}, \code{agnes}, \code{diana}
and \code{pam}.

The value for the \code{cluster_method} configuration parameter is used by
default. When generating clusters for the purpose of ordering samples in
feature expressions, \code{none} is ignored and \code{hclust} is used instead.}
    \item{\code{sample_linkage_method}}{(\emph{optional}) The method used for agglomerative
clustering in \code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

The value for the \code{cluster_linkage_method} configuration parameters is used
by default.}
    \item{\code{sample_similarity_metric}}{(\emph{optional}) Metric to determine pairwise
similarity between samples. Similarity is computed in the same manner as for
clustering, but \code{sample_similarity_metric} has different options that are
better suited to computing distance between samples instead of between
features. The following metrics are available.
\itemize{
\item \code{gower} (default): compute Gower's distance between samples. By default,
Gower's distance is computed based on winsorised data to reduce the effect
of outliers (see below).
\item \code{euclidean}: compute the Euclidean distance between samples.
}

The underlying feature data for numerical features is scaled to the
\eqn{[0,1]} range using the feature values across the samples. The
normalisation parameters required can optionally be computed from feature
data with the outer 5\% (on both sides) of feature values trimmed or
winsorised. To do so append \verb{_trim} (trimming) or \verb{_winsor} (winsorising) to
the metric name. This reduces the effect of outliers somewhat.

Regardless of metric, all categorical features are handled as for the
Gower's distance: distance is 0 if the values in a pair of samples match,
and 1 if they do not.}
    \item{\code{eval_aggregation_method}}{(\emph{optional}) Method for aggregating variable
importances for the purpose of evaluation. Variable importances are
determined both during initial variable importance computation steps and
after training the model. Both types are evaluated after training the
model, but only the initial variable importances are evaluated at run-time.

See the documentation for the \code{vimp_aggregation_method} argument for
information concerning the different methods available.}
    \item{\code{eval_aggregation_rank_threshold}}{(\emph{optional}) The threshold used to
define the subset of highly important features during evaluation.

See the documentation for the \code{vimp_aggregation_rank_threshold} argument for
more information.}
    \item{\code{eval_icc_type}}{(\emph{optional}) String indicating the type of intraclass
correlation coefficient (\code{1}, \code{2} or \code{3}) that should be used to compute
robustness for features in repeated measurements during the evaluation of
univariate importance. These types correspond to the types in Shrout and
Fleiss (1979). The default value is \code{1}.}
    \item{\code{stratification_method}}{(\emph{optional}) Method for determining the
stratification threshold for creating survival groups. The actual,
model-dependent, threshold value is obtained from the development data, and
can afterwards be used to perform stratification on validation data.

The following stratification methods are available:
\itemize{
\item \code{median} (default): The median predicted value in the development cohort
is used to stratify the samples into two risk groups. For predicted outcome
values that build a continuous spectrum, the two risk groups in the
development cohort will be roughly equal in size.
\item \code{mean}: The mean predicted value in the development cohort is used to
stratify the samples into two risk groups.
\item \code{mean_trim}: As \code{mean}, but based on the set of predicted values
where the 5\% lowest and 5\% highest values are discarded. This reduces the
effect of outliers.
\item \code{mean_winsor}: As \code{mean}, but based on the set of predicted values where
the 5\% lowest and 5\% highest values are winsorised. This reduces the effect
of outliers.
\item \code{fixed}: Samples are stratified based on the sample quantiles of the
predicted values. These quantiles are defined using the
\code{stratification_threshold} parameter.
\item \code{optimised}: Use maximally selected rank statistics to determine the
optimal threshold (Lausen and Schumacher, 1992; Hothorn et al., 2003) to
stratify samples into two optimally separated risk groups.
}

One or more stratification methods can be selected simultaneously.

This parameter is only relevant for \code{survival} outcomes.}
    \item{\code{stratification_threshold}}{(\emph{optional}) Numeric value(s) signifying the
sample quantiles for stratification using the \code{fixed} method. The number of
risk groups will be the number of values +1.

The default value is \code{c(1/3, 2/3)}, which will yield two thresholds that
divide samples into three equally sized groups. If \code{fixed} is not among the
selected stratification methods, this parameter is ignored.

This parameter is only relevant for \code{survival} outcomes.}
    \item{\code{time_max}}{(\emph{optional}) Time point which is used as the benchmark for
e.g. cumulative risks generated by random forest, or the cutoff for Uno's
concordance index.

If \code{time_max} is not provided, but \code{evaluation_times} is, the largest value
of \code{evaluation_times} is used. If both are not provided, \code{time_max} is set
to the 98th percentile of the distribution of survival times for samples
with an event in the development data set.

This parameter is only relevant for \code{survival} outcomes.}
    \item{\code{evaluation_times}}{(\emph{optional}) One or more time points that are used for
assessing calibration in survival problems. This is done as expected and
observed survival probabilities depend on time.

If unset, \code{evaluation_times} will be equal to \code{time_max}.

This parameter is only relevant for \code{survival} outcomes.}
    \item{\code{dynamic_model_loading}}{(\emph{optional}) Enables dynamic loading of models
during the evaluation process, if \code{TRUE}. Defaults to \code{FALSE}. Dynamic
loading of models may reduce the overall memory footprint, at the cost of
increased disk or network IO. Models can only be dynamically loaded if they
are found at an accessible disk or network location. Setting this parameter
to \code{TRUE} may help if parallel processing causes out-of-memory issues during
evaluation.}
    \item{\code{parallel_evaluation}}{(\emph{optional}) Enable parallel processing for
hyperparameter optimisation. Defaults to \code{TRUE}. When set to \code{FALSE}, this
will disable the use of parallel processing while performing optimisation,
regardless of the settings of the \code{parallel} parameter. The parameter
moreover specifies whether parallelisation takes place within the evaluation
process steps (\code{inner}, default), or in an outer loop ( \code{outer}) over
learners, data subsamples, etc.

\code{parallel_evaluation} is ignored if \code{parallel=FALSE}.}
  }}
}
\value{
A list of settings to be used within the workflow
}
\description{
Internal function for parsing settings that configure various aspects of the
worklow
}
\references{
\enumerate{
\item Shrout, P. E. & Fleiss, J. L. Intraclass correlations: uses in assessing
rater reliability. Psychol. Bull. 86, 420–428 (1979).
\item Koo, T. K. & Li, M. Y. A guideline of selecting and reporting intraclass
correlation coefficients for reliability research. J. Chiropr. Med. 15,
155–163 (2016).
\item Yeo, I. & Johnson, R. A. A new family of power transformations to
improve normality or symmetry. Biometrika 87, 954–959 (2000).
\item Box, G. E. P. & Cox, D. R. An analysis of transformations. J. R. Stat.
Soc. Series B Stat. Methodol. 26, 211–252 (1964).
\item Raymaekers, J., Rousseeuw,  P. J. Transforming variables to central
normality. Mach Learn. (2021).
\item Zwanenburg, A., & Löck, S. (2026). Location and scale-invariant power
transformations for transforming data to normality. Machine Learning,
115(3), 34.
\item Park, M. Y., Hastie, T. & Tibshirani, R. Averaged gene expressions for
regression. Biostatistics 8, 212–227 (2007).
\item Tolosi, L. & Lengauer, T. Classification with correlated features:
unreliability of feature ranking and solutions. Bioinformatics 27,
1986–1994 (2011).
\item Johnson, W. E., Li, C. & Rabinovic, A. Adjusting batch effects in
microarray expression data using empirical Bayes methods. Biostatistics 8,
118–127 (2007)
\item Kaufman, L. & Rousseeuw, P. J. Finding groups in data: an introduction
to cluster analysis. (John Wiley & Sons, 2009).
\item Muellner, D. fastcluster: fast hierarchical, agglomerative clustering
routines for R and Python. J. Stat. Softw. 53, 1–18 (2013).
\item Rousseeuw, P. J. Silhouettes: A graphical aid to the interpretation and
validation of cluster analysis. J. Comput. Appl. Math. 20, 53–65 (1987).
\item Langfelder, P., Zhang, B. & Horvath, S. Defining clusters from a
hierarchical cluster tree: the Dynamic Tree Cut package for R.
Bioinformatics 24, 719–720 (2008).
\item McFadden, D. Conditional logit analysis of qualitative choice behavior.
in Frontiers in Econometrics (ed. Zarembka, P.) 105–142 (Academic Press,
1974).
\item Cox, D. R. & Snell, E. J. Analysis of binary data. (Chapman and Hall,
1989).
\item Nagelkerke, N. J. D. A note on a general definition of the coefficient
of determination. Biometrika 78, 691–692 (1991).
\item Meinshausen, N. & Buehlmann, P. Stability selection. J. R. Stat. Soc.
Series B Stat. Methodol. 72, 417–473 (2010).
\item Haury, A.-C., Gestraud, P. & Vert, J.-P. The influence of feature
selection methods on accuracy, stability and interpretability of molecular
signatures. PLoS One 6, e28210 (2011).
\item Wald, R., Khoshgoftaar, T. M., Dittman, D., Awada, W. & Napolitano,A. An
extensive comparison of feature ranking aggregation techniques in
bioinformatics. in 2012 IEEE 13th International Conference on Information
Reuse Integration (IRI) 377–384 (2012).
\item Hutter, F., Hoos, H. H. & Leyton-Brown, K. Sequential model-based
optimization for general algorithm configuration. in Learning and
Intelligent Optimization (ed. Coello, C. A. C.) 6683, 507–523 (Springer
Berlin Heidelberg, 2011).
\item Shahriari, B., Swersky, K., Wang, Z., Adams, R. P. & de Freitas, N.
Taking the Human Out of the Loop: A Review of Bayesian Optimization. Proc.
IEEE 104, 148–175 (2016)
\item Srinivas, N., Krause, A., Kakade, S. M. & Seeger, M. W.
Information-Theoretic Regret Bounds for Gaussian Process Optimization in
the Bandit Setting. IEEE Trans. Inf. Theory 58, 3250–3265 (2012)
\item Kaufmann, E., Cappé, O. & Garivier, A. On Bayesian upper confidence
bounds for bandit problems. in Artificial intelligence and statistics
592–600 (2012).
\item Jamieson, K. & Talwalkar, A. Non-stochastic Best Arm Identification and
Hyperparameter Optimization. in Proceedings of the 19th International
Conference on Artificial Intelligence and Statistics (eds. Gretton, A. &
Robert, C. C.) vol. 51 240–248 (PMLR, 2016).
\item Gramacy, R. B. laGP: Large-Scale Spatial Modeling via Local Approximate
Gaussian Processes in R. Journal of Statistical Software 72, 1–46 (2016)
\item Sparapani, R., Spanbauer, C. & McCulloch, R. Nonparametric Machine
Learning and Efficient Computation with Bayesian Additive Regression Trees:
The BART R Package. Journal of Statistical Software 97, 1–66 (2021)
\item Davison, A. C. & Hinkley, D. V. Bootstrap methods and their application.
(Cambridge University Press, 1997).
\item Efron, B. & Hastie, T. Computer Age Statistical Inference. (Cambridge
University Press, 2016).
\item Lausen, B. & Schumacher, M. Maximally Selected Rank Statistics.
Biometrics 48, 73 (1992).
\item Hothorn, T. & Lausen, B. On the exact distribution of maximally selected
rank statistics. Comput. Stat. Data Anal. 43, 121–137 (2003).
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Iterations.R
\name{.get_iteration_data}
\alias{.get_iteration_data}
\title{Internal function for creating or retrieving iteration data}
\usage{
.get_iteration_data(
  file_paths,
  data,
  experiment_setup,
  settings,
  message_indent = 0L,
  verbose = TRUE
)
}
\arguments{
\item{file_paths}{Path to directory where iteration files are stored.}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{experiment_setup}{data.table with subsampler information at different
levels of the experimental design.}

\item{settings}{List of parameter settings. Some of these parameters are
relevant to creating iterations.}

\item{message_indent}{Indenting of messages.}

\item{verbose}{Sets verbosity.}
}
\value{
A list with the following elements:
\itemize{
\item \code{iter_list}: A list containing iteration data at the different levels of
the experiment.
\item \code{project_id}: The unique project identifier.
\item \code{experiment_setup}: data.table with subsampler information at different
levels of the experimental design.
}
}
\description{
Internal function for creating or retrieving iteration data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseSettings.R
\name{.parse_variable_importance_settings}
\alias{.parse_variable_importance_settings}
\title{Internal function for parsing settings related to variable importance
computation.}
\usage{
.parse_variable_importance_settings(
  config = NULL,
  data,
  parallel,
  outcome_type,
  vimp_method = waiver(),
  vimp_method_parameter = waiver(),
  vimp_aggregation_method = waiver(),
  vimp_aggregation_rank_threshold = waiver(),
  parallel_vimp = waiver(),
  ...
)
}
\arguments{
\item{config}{A list of settings, e.g. from an xml file.}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{parallel}{Logical value that whether familiar uses parallelisation. If
\code{FALSE} it will override \code{parallel_vimp}.}

\item{outcome_type}{Type of outcome found in the data set.}

\item{vimp_method}{(\strong{required}) Variable importance method. \code{familiar}
implements various variable importance methods. Please refer to the
vignette on variable importance methods for more details.

More than one variable importance method can be chosen. The experiment will
then repeated for each feature selection method.

Variable importance methods determine the ranking of features. Actual
selection of features is done by optimising the signature size model
hyperparameter during the hyperparameter optimisation step.}

\item{vimp_method_parameter}{(\emph{optional}) List of lists containing parameters
for feature selection methods. Each sublist should have the name of the
feature selection method it corresponds to.

Most feature selection methods do not have parameters that can be set.
Please refer to the vignette on feature selection methods for more details.
Note that if the feature selection method is based on a learner (e.g. lasso
regression), hyperparameter optimisation may be performed prior to
assessing variable importance.}

\item{vimp_aggregation_method}{(\emph{optional}) The method used to aggregate
variable importances over different data subsets, e.g. bootstraps. The
following methods can be selected:
\itemize{
\item \code{none}: Don't aggregate ranks, but rather aggregate the variable
importance scores themselves.
\item \code{mean}: Use the mean rank of a feature over the subsets to
determine the aggregated feature rank.
\item \code{median}: Use the median rank of a feature over the subsets to determine
the aggregated feature rank.
\item \code{best}: Use the best rank the feature obtained in any subset to determine
the aggregated feature rank.
\item \code{worst}: Use the worst rank the feature obtained in any subset to
determine the aggregated feature rank.
\item \code{stability}: Use the frequency of the feature being in the subset of
highly ranked features as measure for the aggregated feature rank
(Meinshausen and Buehlmann, 2010).
\item \code{exponential}: Use a rank-weighted frequence of occurrence in the subset
of highly ranked features as measure for the aggregated feature rank (Haury
et al., 2011).
\item \code{borda} (default): Use the borda count as measure for the aggregated
feature rank (Wald et al., 2012).
\item \code{enhanced_borda}: Use an occurrence frequency-weighted borda count as
measure for the aggregated feature rank (Wald et al., 2012).
\item \code{truncated_borda}: Use borda count computed only on features within the
subset of highly ranked features.
\item \code{enhanced_truncated_borda}: Apply both the enhanced borda method and the
truncated borda method and use the resulting borda count as the aggregated
feature rank.
}

The \emph{feature selection methods} vignette provides additional information.}

\item{vimp_aggregation_rank_threshold}{(\emph{optional}) The threshold used to
define the subset of highly important features. If set to \code{NULL}, this
threshold is determined by maximising the variance in the occurrence value
over all features over the subset size. The default value is \code{5}.

This parameter is only relevant for \code{stability}, \code{exponential},
\code{enhanced_borda}, \code{truncated_borda} and \code{enhanced_truncated_borda} methods.}

\item{parallel_vimp}{(\emph{optional}) Enable parallel processing for
variable importance tasks. Defaults to \code{TRUE}. When set to \code{FALSE},
this will disable the use of parallel processing while performing feature
selection, regardless of the settings of the \code{parallel} parameter.
\code{parallel_vimp} is ignored if \code{parallel = FALSE}.}

\item{...}{Unused arguments.}
}
\value{
List of parameters related to variable importance computation.
}
\description{
Internal function for parsing settings related to variable importance
computation.
}
\references{
\enumerate{
\item Wald, R., Khoshgoftaar, T. M., Dittman, D., Awada, W. &
Napolitano, A. An extensive comparison of feature ranking aggregation
techniques in bioinformatics. in 2012 IEEE 13th International Conference on
Information Reuse Integration (IRI) 377–384 (2012).
\item Meinshausen, N. & Buehlmann, P. Stability selection. J. R. Stat. Soc.
Series B Stat. Methodol. 72, 417–473 (2010).
\item Haury, A.-C., Gestraud, P. & Vert, J.-P. The influence of feature
selection methods on accuracy, stability and interpretability of molecular
signatures. PLoS One 6, e28210 (2011).
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utilities.R
\name{is.waive}
\alias{is.waive}
\title{Internal test to see if an object is a waiver}
\usage{
is.waive(x)
}
\arguments{
\item{x}{Object to be tested.}
}
\value{
\code{TRUE} for objects that are waivers, \code{FALSE} otherwise.
}
\description{
This function tests if the object was created by the \code{waiver} function. This
function is functionally identical to \code{ggplot2:::is.waive()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseSettings.R
\name{.parse_hyperparameter_optimisation_settings}
\alias{.parse_hyperparameter_optimisation_settings}
\title{Internal function for parsing settings related to hyperparameter optimisation}
\usage{
.parse_hyperparameter_optimisation_settings(
  config = NULL,
  parallel,
  outcome_type,
  optimisation_bootstraps = waiver(),
  optimisation_determine_vimp = waiver(),
  smbo_random_initialisation = waiver(),
  smbo_n_random_sets = waiver(),
  max_smbo_iterations = waiver(),
  smbo_stop_convergent_iterations = waiver(),
  smbo_stop_tolerance = waiver(),
  smbo_time_limit = waiver(),
  smbo_initial_bootstraps = waiver(),
  smbo_step_bootstraps = waiver(),
  smbo_intensify_steps = waiver(),
  smbo_stochastic_reject_p_value = waiver(),
  optimisation_function = waiver(),
  optimisation_metric = waiver(),
  acquisition_function = waiver(),
  exploration_method = waiver(),
  hyperparameter_learner = waiver(),
  parallel_hyperparameter_optimisation = waiver(),
  ...
)
}
\arguments{
\item{config}{A list of settings, e.g. from an xml file.}

\item{parallel}{Logical value that whether familiar uses parallelisation. If
\code{FALSE} it will override \code{parallel_hyperparameter_optimisation}.}

\item{outcome_type}{Type of outcome found in the data set.}

\item{optimisation_bootstraps}{(\emph{optional}) Number of bootstraps that should
be generated from the development data set. During the optimisation
procedure one or more of these bootstraps (indicated by
\code{smbo_step_bootstraps}) are used for model development using different
combinations of hyperparameters. The effect of the hyperparameters is then
assessed by comparing in-bag and out-of-bag model performance.

The default number of bootstraps is \code{50}. Hyperparameter optimisation may
finish before exhausting the set of bootstraps.}

\item{optimisation_determine_vimp}{(\emph{optional}) Logical value that indicates
whether variable importance is determined separately for each of the
bootstraps created during the optimisation process (\code{TRUE}) or the
applicable results from the variable importance computation step are used
(\code{FALSE}).

Determining variable importance increases the initial computational
overhead. However, it prevents positive biases for the out-of-bag data due
to overlap of these data with the development data set used for the feature
selection step. In this case, any hyperparameters of the variable
importance method are not determined separately for each bootstrap, but
those obtained during the variable importance computation step are used
instead. In case multiple of such hyperparameter sets could be applicable,
the set that will be used is randomly selected for each bootstrap.

This parameter only affects hyperparameter optimisation of learners. The
default is \code{TRUE}.}

\item{smbo_random_initialisation}{(\emph{optional}) String indicating the
initialisation method for the hyperparameter space. Can be one of
\code{fixed_subsample} (default), \code{fixed}, or \code{random}. \code{fixed} and
\code{fixed_subsample} first create hyperparameter sets from a range of default
values set by familiar. \code{fixed_subsample} then randomly draws up to
\code{smbo_n_random_sets} from the grid. \code{random} does not rely upon a fixed
grid, and randomly draws up to \code{smbo_n_random_sets} hyperparameter sets
from the hyperparameter space.}

\item{smbo_n_random_sets}{(\emph{optional}) Number of random or subsampled
hyperparameters drawn during the initialisation process. Default: \code{100}.
Cannot be smaller than \code{10}. The parameter is not used when
\code{smbo_random_initialisation} is \code{fixed}, as the entire pre-defined grid
will be explored.}

\item{max_smbo_iterations}{(\emph{optional}) Maximum number of intensify
iterations of the SMBO algorithm. During an intensify iteration a run-off
occurs between the current \emph{best} hyperparameter combination and either 10
challenger combination with the highest expected improvement or a set of 20
random combinations.

Run-off with random combinations is used to force exploration of the
hyperparameter space, and is performed every second intensify iteration, or
if there is no expected improvement for any challenger combination.

If a combination of hyperparameters leads to better performance on the same
data than the incumbent \emph{best} set of hyperparameters, it replaces the
incumbent set at the end of the intensify iteration.

The default number of intensify iteration is \code{20}. Iterations may be
stopped early if the incumbent set of hyperparameters remains the same for
\code{smbo_stop_convergent_iterations} iterations, or performance improvement is
minimal. This behaviour is suppressed during the first 4 iterations to
enable the algorithm to explore the hyperparameter space.}

\item{smbo_stop_convergent_iterations}{(\emph{optional}) The number of subsequent
convergent SMBO iterations required to stop hyperparameter optimisation
early. An iteration is convergent if the \emph{best} parameter set has not
changed or the optimisation score over the 4 most recent iterations has not
changed beyond the tolerance level in \code{smbo_stop_tolerance}.

The default value is \code{3}.}

\item{smbo_stop_tolerance}{(\emph{optional}) Tolerance for early stopping due to
convergent optimisation score.

The default value depends on the square root of the number of samples (at
the series level), and is \code{0.01} for 100 samples. This value is computed as
\code{0.1 * 1 / sqrt(n_samples)}. The upper limit is \code{0.0001} for 1M or more
samples.}

\item{smbo_time_limit}{(\emph{optional}) Time limit (in minutes) for the
optimisation process. Optimisation is stopped after this limit is exceeded.
Time taken to determine variable importance for the optimisation process
(see the \code{optimisation_determine_vimp} parameter) does not count.

The default is \code{NULL}, indicating that there is no time limit for the
optimisation process. The time limit cannot be less than 1 minute.}

\item{smbo_initial_bootstraps}{(\emph{optional}) The number of bootstraps taken
from the set of \code{optimisation_bootstraps} as the bootstraps assessed
initially.

The default value is \code{1}. The value cannot be larger than
\code{optimisation_bootstraps}.}

\item{smbo_step_bootstraps}{(\emph{optional}) The number of bootstraps taken from
the set of \code{optimisation_bootstraps} bootstraps as the bootstraps assessed
during the steps of each intensify iteration.

The default value is \code{3}. The value cannot be larger than
\code{optimisation_bootstraps}.}

\item{smbo_intensify_steps}{(\emph{optional}) The number of steps in each SMBO
intensify iteration. Each step a new set of \code{smbo_step_bootstraps}
bootstraps is drawn and used in the run-off between the incumbent \emph{best}
hyperparameter combination and its challengers.

The default value is \code{5}. Higher numbers allow for a more detailed
comparison, but this comes with added computational cost.}

\item{smbo_stochastic_reject_p_value}{(\emph{optional}) The p-value threshold used
for the \code{stochastic_reject} exploration method.

The default value is \code{0.05}.}

\item{optimisation_function}{(\emph{optional}) Type of optimisation function used
to quantify the performance of a hyperparameter set. Model performance is
assessed using the metric(s) specified by \code{optimisation_metric} on the
in-bag (IB) and out-of-bag (OOB) samples of a bootstrap. These values are
converted to objective scores with a standardised interval of
\eqn{[-1.0, 1.0]}. Each pair of objective is subsequently used to compute an
optimisation score. The optimisation score across different bootstraps is
than aggregated to a summary score. This summary score is used to rank
hyperparameter sets, and select the optimal set.

The combination of optimisation score and summary score is determined by
the optimisation function indicated by this parameter:
\itemize{
\item \code{validation} or \code{max_validation} (default): seeks to maximise OOB score.
\item \code{balanced}: seeks to balance IB and OOB score.
\item \code{stronger_balance}: similar to \code{balanced}, but with stronger penalty for
differences between IB and OOB scores.
\item \code{validation_minus_sd}: seeks to optimise the average OOB score minus its
standard deviation.
\item \code{validation_25th_percentile}: seeks to optimise the 25th percentile of
OOB scores, and is conceptually similar to \code{validation_minus_sd}.
\item \code{model_estimate}: seeks to maximise the OOB score estimate predicted by
the hyperparameter learner (not available for random search).
\item \code{model_estimate_minus_sd}: seeks to maximise the OOB score estimate minus
its estimated standard deviation, as predicted by the hyperparameter
learner (not available for random search).
\item \code{model_balanced_estimate}: seeks to maximise the estimate of the balanced
IB and OOB score. This is similar to the \code{balanced} score, and in fact uses
a hyperparameter learner to predict said score (not available for random
search).
\item \code{model_balanced_estimate_minus_sd}: seeks to maximise the estimate of the
balanced IB and OOB score, minus its estimated standard deviation. This is
similar to the \code{balanced} score, but takes into account its estimated
spread.
}

Additional detail are provided in the \emph{Learning algorithms and
hyperparameter optimisation} vignette.}

\item{optimisation_metric}{(\emph{optional}) One or more metrics used to compute
performance scores. See the vignette on performance metrics for the
available metrics.

If unset, the following metrics are used by default:
\itemize{
\item \code{auc_roc}: For \code{binomial} and \code{multinomial} models.
\item \code{mse}: Mean squared error for \code{continuous} models.
\item \code{concordance_index}: For \code{survival} models.
}

Multiple optimisation metrics can be specified. Actual metric values are
converted to an objective value by comparison with a baseline metric value
that derives from a trivial model, i.e. majority class for binomial and
multinomial outcomes, the median outcome for count and continuous outcomes
and a fixed risk or time for survival outcomes.}

\item{acquisition_function}{(\emph{optional}) The acquisition function influences
how new hyperparameter sets are selected. The algorithm uses the model
learned by the learner indicated by \code{hyperparameter_learner} to search the
hyperparameter space for hyperparameter sets that are either likely better
than the best known set (\emph{exploitation}) or where there is considerable
uncertainty (\emph{exploration}). The acquisition function quantifies this
(Shahriari et al., 2016).

The following acquisition functions are available, and are described in
more detail in the \emph{learner algorithms} vignette:
\itemize{
\item \code{improvement_probability}: The probability of improvement quantifies the
probability that the expected optimisation score for a set is better than
the best observed optimisation score
\item \code{improvement_empirical_probability}: Similar to
\code{improvement_probability}, but based directly on optimisation scores
predicted by the individual decision trees.
\item \code{expected_improvement} (default): Computes expected improvement.
\item \code{upper_confidence_bound}: This acquisition function is based on the upper
confidence bound of the distribution (Srinivas et al., 2012).
\item \code{bayes_upper_confidence_bound}: This acquisition function is based on the
upper confidence bound of the distribution (Kaufmann et al., 2012).
}}

\item{exploration_method}{(\emph{optional}) Method used to steer exploration in
post-initialisation intensive searching steps. As stated earlier, each SMBO
iteration step compares suggested alternative parameter sets with an
incumbent \strong{best} set in a series of steps. The exploration method
controls how the set of alternative parameter sets is pruned after each
step in an iteration. Can be one of the following:
\itemize{
\item \code{single_shot} (default): The set of alternative parameter sets is not
pruned, and each intensification iteration contains only a single
intensification step that only uses a single bootstrap. This is the fastest
exploration method, but only superficially tests each parameter set.
\item \code{successive_halving}: The set of alternative parameter sets is
pruned by removing the worst performing half of the sets after each step
(Jamieson and Talwalkar, 2016).
\item \code{stochastic_reject}: The set of alternative parameter sets is pruned by
comparing the performance of each parameter set with that of the incumbent
\strong{best} parameter set using a paired Wilcoxon test based on shared
bootstraps. Parameter sets that perform significantly worse, at an alpha
level indicated by \code{smbo_stochastic_reject_p_value}, are pruned.
\item \code{none}: The set of alternative parameter sets is not pruned.
}}

\item{hyperparameter_learner}{(\emph{optional}) Any point in the hyperparameter
space has a single, scalar, optimisation score value that is \emph{a priori}
unknown. During the optimisation process, the algorithm samples from the
hyperparameter space by selecting hyperparameter sets and computing the
optimisation score value for one or more bootstraps. For each
hyperparameter set the resulting values are distributed around the actual
value. The learner indicated by \code{hyperparameter_learner} is then used to
infer optimisation score estimates for unsampled parts of the
hyperparameter space.

The following models are available:
\itemize{
\item \code{bayesian_additive_regression_trees} or \code{bart}: Uses Bayesian Additive
Regression Trees (Sparapani et al., 2021) for inference. Unlike standard
random forests, BART allows for estimating posterior distributions directly
and can extrapolate.
\item \code{gaussian_process} (default): Creates a localised approximate Gaussian
process for inference (Gramacy, 2016). This allows for better scaling than
deterministic Gaussian Processes.
\item \code{random_forest}: Creates a random forest for inference. Originally
suggested by Hutter et al. (2011). A weakness of random forests is their
lack of extrapolation beyond observed values, which limits their usefulness
in exploiting promising areas of hyperparameter space.
\item \code{random} or \code{random_search}: Forgoes the use of models to steer
optimisation. Instead, a random search is performed.
}}

\item{parallel_hyperparameter_optimisation}{(\emph{optional}) Enable parallel
processing for hyperparameter optimisation. Defaults to \code{TRUE}. When set to
\code{FALSE}, this will disable the use of parallel processing while performing
optimisation, regardless of the settings of the \code{parallel} parameter. The
parameter moreover specifies whether parallelisation takes place within the
optimisation algorithm (\code{inner}, default), or in an outer loop ( \code{outer})
over learners, data subsamples, etc.

\code{parallel_hyperparameter_optimisation} is ignored if \code{parallel=FALSE}.}

\item{...}{Unused arguments.}
}
\value{
List of parameters related to model hyperparameter optimisation.
}
\description{
Internal function for parsing settings related to hyperparameter optimisation
}
\references{
\enumerate{
\item Hutter, F., Hoos, H. H. & Leyton-Brown, K. Sequential
model-based optimization for general algorithm configuration. in Learning
and Intelligent Optimization (ed. Coello, C. A. C.) 6683, 507–523 (Springer
Berlin Heidelberg, 2011).
\item Shahriari, B., Swersky, K., Wang, Z., Adams, R. P. & de Freitas, N.
Taking the Human Out of the Loop: A Review of Bayesian Optimization. Proc.
IEEE 104, 148–175 (2016)
\item Srinivas, N., Krause, A., Kakade, S. M. & Seeger, M. W.
Information-Theoretic Regret Bounds for Gaussian Process Optimization in
the Bandit Setting. IEEE Trans. Inf. Theory 58, 3250–3265 (2012)
\item Kaufmann, E., Cappé, O. & Garivier, A. On Bayesian upper confidence
bounds for bandit problems. in Artificial intelligence and statistics
592–600 (2012).
\item Jamieson, K. & Talwalkar, A. Non-stochastic Best Arm Identification and
Hyperparameter Optimization. in Proceedings of the 19th International
Conference on Artificial Intelligence and Statistics (eds. Gretton, A. &
Robert, C. C.) vol. 51 240–248 (PMLR, 2016).
\item Gramacy, R. B. laGP: Large-Scale Spatial Modeling via Local Approximate
Gaussian Processes in R. Journal of Statistical Software 72, 1–46 (2016)
\item Sparapani, R., Spanbauer, C. & McCulloch, R. Nonparametric Machine
Learning and Efficient Computation with Bayesian Additive Regression Trees:
The BART R Package. Journal of Statistical Software 97, 1–66 (2021)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseData.R
\name{.parse_integer_features}
\alias{.parse_integer_features}
\title{Internal function for converting integer features}
\usage{
.parse_integer_features(data, outcome_type)
}
\arguments{
\item{data}{data.table with feature data}

\item{outcome_type}{character, indicating the type of outcome}
}
\value{
data.table with integer features converted to double.
}
\description{
Internal function for converting integer features
}
\details{
This function parses columns containing integer feature data to
features to double. This prevents, e.g., errors when the result of an
operation on the feature data yields a non-integer (i.e. floating point)
result.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarEnsemble-class}
\alias{familiarEnsemble-class}
\title{Ensemble of familiar models.}
\description{
A familiarEnsemble object contains one or more familiarModel objects.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the familiarEnsemble object.}

\item{\code{model_list}}{List of attached familiarModel objects, or paths to these
objects. Familiar attaches familiarModel objects when required.}

\item{\code{outcome_type}}{Outcome type of the data used to create the object.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{data_column_info}}{Data information object containing information
regarding identifier column names and outcome column names.}

\item{\code{learner}}{Learning algorithm used to create the models in the ensemble.}

\item{\code{vimp_method}}{Method used to determine variable importance for the models
in the ensemble.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters.}

\item{\code{required_features}}{The set of features required for complete
reproduction, i.e. with imputation.}

\item{\code{model_features}}{The combined set of features that is used to train the
models in the ensemble,}

\item{\code{novelty_features}}{The combined set of features that is used to train all
novelty detectors in the ensemble.}

\item{\code{data_id}}{Internal identifier for the dataset used to train or
evaluate the ensemble.}

\item{\code{run_id}}{Internal identifier for the specific subset of the dataset used
used to train or evaluate the ensemble.}

\item{\code{run_table}}{Run table for the data used to train the ensemble. Used
internally.}

\item{\code{calibration_info}}{Calibration information, e.g. baseline survival in the
development cohort.}

\item{\code{model_dir_path}}{Path to folder containing the familiarModel objects. Can
be updated using the \code{update_model_dir_path} method.}

\item{\code{auto_detach}}{Flag used to determine whether models should be detached
from the model after use, or not. Used internally.}

\item{\code{settings}}{A copy of the evaluation configuration parameters used at
model creation. These are used as default parameters when evaluating the
ensemble to create a familiarData object.}

\item{\code{project_id}}{Identifier of the project that generated the underlying
familiarModel object(s).}

\item{\code{familiar_version}}{Version of the familiar package.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationVimp.R
\name{export_model_vimp}
\alias{export_model_vimp}
\alias{export_model_vimp,familiarCollection-method}
\alias{export_model_vimp,ANY-method}
\title{Extract and export model-based variable importance.}
\usage{
export_model_vimp(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{export_model_vimp}{familiarCollection}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{export_model_vimp}{ANY}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{aggregate_results}{Flag that signifies whether results should be
aggregated for export.}

\item{aggregation_method}{(\emph{optional}) The method used to aggregate variable
importances over different data subsets, e.g. bootstraps. The following
methods can be selected:
\itemize{
\item \code{mean} (default): Use the mean rank of a feature over the subsets to
determine the aggregated feature rank.
\item \code{median}: Use the median rank of a feature over the subsets to determine
the aggregated feature rank.
\item \code{best}: Use the best rank the feature obtained in any subset to determine
the aggregated feature rank.
\item \code{worst}: Use the worst rank the feature obtained in any subset to
determine the aggregated feature rank.
\item \code{stability}: Use the frequency of the feature being in the subset of
highly ranked features as measure for the aggregated feature rank
(Meinshausen and Buehlmann, 2010).
\item \code{exponential}: Use a rank-weighted frequence of occurrence in the subset
of highly ranked features as measure for the aggregated feature rank (Haury
et al., 2011).
\item \code{borda}: Use the borda count as measure for the aggregated feature rank
(Wald et al., 2012).
\item \code{enhanced_borda}: Use an occurrence frequency-weighted borda count as
measure for the aggregated feature rank (Wald et al., 2012).
\item \code{truncated_borda}: Use borda count computed only on features within the
subset of highly ranked features.
\item \code{enhanced_truncated_borda}: Apply both the enhanced borda method and the
truncated borda method and use the resulting borda count as the aggregated
feature rank.
}}

\item{rank_threshold}{(\emph{optional}) The threshold used to define the subset of
highly important features. If not set, this threshold is determined by
maximising the variance in the occurrence value over all features over the
subset size.

This parameter is only relevant for \code{stability}, \code{exponential},
\code{enhanced_borda}, \code{truncated_borda} and \code{enhanced_truncated_borda} methods.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A data.table (if \code{dir_path} is not provided), or nothing, as all data
is exported to \code{csv} files.
}
\description{
Extract and export model-based variable importance from a
familiarCollection.
}
\details{
Data, such as model performance and calibration information, is
usually collected from a \code{familiarCollection} object. However, you can also
provide one or more \code{familiarData} objects, that will be internally
converted to a \code{familiarCollection} object. It is also possible to provide a
\code{familiarEnsemble} or one or more \code{familiarModel} objects together with the
data from which data is computed prior to export. Paths to the previous
files can also be provided.

All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Variable importance is based on the ranking produced by model-specific
variable importance routines, e.g. permutation for random forests. If such a
routine is absent, variable importance is based on the feature selection
method that led to the features included in the model. In case multiple
models (\code{familiarModel} objects) are combined, feature ranks are first
aggregated using the method defined by the \code{aggregation_method}, some of
which require a \code{rank_threshold} to indicate a subset of most important
features.

Information concerning highly similar features that form clusters is
provided as well. This information is based on consensus clustering of the
features that were used in the signatures of the underlying models. This
clustering information is also used during aggregation to ensure that
co-clustered features are only taken into account once.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Familiar.R
\name{precompute_feature_info}
\alias{precompute_feature_info}
\title{Pre-compute feature information}
\usage{
precompute_feature_info(
  formula = NULL,
  data = NULL,
  experiment_data = NULL,
  cl = NULL,
  experimental_design = "fs+mb",
  verbose = TRUE,
  ...
)
}
\arguments{
\item{formula}{An R formula. The formula can only contain feature names and
dot (\code{.}). The \code{*} and \code{+1} operators are not supported as these refer to
columns that are not present in the data set.

Use of the formula interface is optional.}

\item{data}{A \code{data.table} object, a \code{data.frame} object, list containing
multiple \code{data.table} or \code{data.frame} objects, or paths to data files.

\code{data} should be provided if no file paths are provided to the \code{data_files}
argument. If both are provided, only \code{data} will be used.

All data is expected to be in wide format, and ideally has a sample
identifier (see \code{sample_id_column}), batch identifier (see \code{cohort_column})
and outcome columns (see \code{outcome_column}).

In case paths are provided, the data should be stored as \code{csv}, \code{rds} or
\code{RData} files. See documentation for the \code{data_files} argument for more
information.}

\item{experiment_data}{Experimental data may provided in the form of the
output of \code{precompute_data_assignment}, \code{precompute_feature_info} or
\code{precompute_vimp}. This allows for warm-starting experiments.}

\item{cl}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallelisation. When a cluster is not
provided, parallelisation is performed by setting up a cluster on the local
machine.

This parameter has no effect if the \code{parallel} argument is set to \code{FALSE}.}

\item{experimental_design}{(\strong{required}) Defines what the experiment looks
like, e.g. \code{cv(bt(fs,20)+mb,3,2)} for 2 times repeated 3-fold
cross-validation with nested variable importance computation on 20
bootstraps and model-building. The basic workflow components are:
\itemize{
\item \code{fs}: (optional) variable importance computation step. If not explicitly
declared, feature selection will be done just in time for hyperparameter
optimisation.
\item \code{mb}: (required) model building step.
\item \code{ev}: (optional) external validation. If validation batches or cohorts
are present in the dataset (\code{data}), these should be indicated in the
\code{validation_batch_id} argument.
}

The different components are linked using \code{+}.

Different subsampling methods can be used in conjunction with the basic
workflow components:
\itemize{
\item \code{bs(x,n)}: (stratified) .632 bootstrap, with \code{n} the number of
bootstraps. In contrast to \code{bt}, feature pre-processing parameters and
hyperparameter optimisation are conducted on individual bootstraps.
\item \code{bt(x,n)}: (stratified) .632 bootstrap, with \code{n} the number of
bootstraps. Unlike \code{bs} and other subsampling methods, no separate
pre-processing parameters or optimised hyperparameters will be determined
for each bootstrap.
\item \code{cv(x,n,p)}: (stratified) \code{n}-fold cross-validation, repeated \code{p} times.
Pre-processing parameters are determined for each iteration.
\item \code{lv(x)}: leave-one-out-cross-validation. Pre-processing parameters are
determined for each iteration.
\item \code{ip(x)}: imbalance partitioning for addressing class imbalances on the
data set. Pre-processing parameters are determined for each partition. The
number of partitions generated depends on the imbalance correction method
(see the \code{imbalance_correction_method} parameter).
}

As shown in the example above, sampling algorithms can be nested.

Though neither variable importance is determined nor models are learned
within \code{precompute_feature_info}, the corresponding elements are still
required to prevent issues when using the resulting \code{experimentData} object
to warm-start the experiments.

The simplest valid experimental design is \code{fs+mb}. This is the default in
\code{precompute_feature_info}, and will determine feature parameters over the
entire dataset.

This argument is ignored if the \code{experiment_data} argument is set.}

\item{verbose}{Indicates verbosity of the results. Default is TRUE, and all
messages and warnings are returned.}

\item{...}{
  Arguments passed on to \code{\link[=.parse_experiment_settings]{.parse_experiment_settings}}, \code{\link[=.parse_setup_settings]{.parse_setup_settings}}, \code{\link[=.parse_preprocessing_settings]{.parse_preprocessing_settings}}
  \describe{
    \item{\code{batch_id_column}}{(\strong{recommended}) Name of the column containing batch
or cohort identifiers. This parameter is required if more than one dataset
is provided, or if external validation is performed.

In familiar any row of data is organised by four identifiers:
\itemize{
\item The batch identifier \code{batch_id_column}: This denotes the group to which a
set of samples belongs, e.g. patients from a single study, samples measured
in a batch, etc. The batch identifier is used for batch normalisation, as
well as selection of development and validation datasets.
\item The sample identifier \code{sample_id_column}: This denotes the sample level,
e.g. data from a single individual. Subsets of data, e.g. bootstraps or
cross-validation folds, are created at this level.
\item The series identifier \code{series_id_column}: Indicates measurements on a
single sample that may not share the same outcome value, e.g. a time
series, or the number of cells in a view.
\item The repetition identifier: Indicates repeated measurements in a single
series where any feature values may differ, but the outcome does not.
Repetition identifiers are always implicitly set when multiple entries for
the same series of the same sample in the same batch that share the same
outcome are encountered.
}}
    \item{\code{sample_id_column}}{(\strong{recommended}) Name of the column containing
sample or subject identifiers. See \code{batch_id_column} above for more
details.

If unset, every row will be identified as a single sample.}
    \item{\code{series_id_column}}{(\emph{optional}) Name of the column containing series
identifiers, which distinguish between measurements that are part of a
series for a single sample. See \code{batch_id_column} above for more details.

If unset, rows which share the same batch and sample identifiers but have a
different outcome are assigned unique series identifiers.}
    \item{\code{development_batch_id}}{(\emph{optional}) One or more batch or cohort
identifiers to constitute data sets for development. Defaults to all, or
all minus the identifiers in \code{validation_batch_id} for external validation.
Required if external validation is performed and \code{validation_batch_id} is
not provided.}
    \item{\code{validation_batch_id}}{(\emph{optional}) One or more batch or cohort
identifiers to constitute data sets for external validation. Defaults to
all data sets except those in \code{development_batch_id} for external
validation, or none if not. Required if \code{development_batch_id} is not
provided.}
    \item{\code{outcome_name}}{(\emph{optional}) Name of the modelled outcome. This name will
be used in figures created by \code{familiar}.

If not set, the column name in \code{outcome_column} will be used for
\code{binomial}, \code{multinomial}, and \code{continuous} outcomes. For other outcomes
(\code{survival} and \code{competing_risk}) no default is used.}
    \item{\code{outcome_column}}{(\strong{recommended}) Name of the column containing the
outcome of interest. May be identified from a formula, if a formula is
provided as an argument. Otherwise an error is raised. Note that \code{survival}
and \code{competing_risk} outcome type outcomes require two columns that
indicate the time-to-event or the time of last follow-up and the event
status.}
    \item{\code{outcome_type}}{(\strong{recommended}) Type of outcome found in the outcome
column. The outcome type determines many aspects of the overall process,
e.g. the available variable importance methods and learners, but also the
type of assessments that can be conducted to evaluate the resulting models.
Implemented outcome types are:
\itemize{
\item \code{binomial}: categorical outcome with 2 levels.
\item \code{multinomial}: categorical outcome with 2 or more levels.
\item \code{continuous}: general continuous numeric outcomes.
\item \code{survival}: survival outcome for time-to-event data.
}

If not provided, the algorithm will attempt to obtain outcome_type from
contents of the outcome column. This may lead to unexpected results, and we
therefore advise to provide this information manually.

Note that \code{competing_risk} survival analysis are not fully supported, and
is currently not a valid choice for \code{outcome_type}. The \code{count} outcome
type was deprecated in version 2.0.0, and superseded by \code{continuous}.}
    \item{\code{class_levels}}{(\emph{optional}) Class levels for \code{binomial} or \code{multinomial}
outcomes. This argument can be used to specify the ordering of levels for
categorical outcomes. These class levels must exactly match the levels
present in the outcome column.}
    \item{\code{event_indicator}}{(\strong{recommended}) Indicator for events in \code{survival}
and \code{competing_risk} analyses. \code{familiar} will automatically recognise \code{1},
\code{true}, \code{t}, \code{y} and \code{yes} as event indicators, including different
capitalisations. If this parameter is set, it replaces the default values.}
    \item{\code{censoring_indicator}}{(\strong{recommended}) Indicator for right-censoring in
\code{survival} and \code{competing_risk} analyses. \code{familiar} will automatically
recognise \code{0}, \code{false}, \code{f}, \code{n}, \code{no} as censoring indicators, including
different capitalisations. If this parameter is set, it replaces the
default values.}
    \item{\code{competing_risk_indicator}}{(\strong{recommended}) Indicator for competing
risks in \code{competing_risk} analyses. There are no default values, and if
unset, all values other than those specified by the \code{event_indicator} and
\code{censoring_indicator} parameters are considered to indicate competing
risks.}
    \item{\code{signature}}{(\emph{optional}) One or more names of feature columns that are
considered part of a specific signature. Features specified here will
always be used for modelling. Ranking of variable importances has no effect
for these features.}
    \item{\code{novelty_features}}{(\emph{optional}) One or more names of feature columns
that should be included for the purpose of novelty detection.}
    \item{\code{exclude_features}}{(\emph{optional}) Feature columns that will be removed
from the data set. Cannot overlap with features in \code{signature},
\code{novelty_features} or \code{include_features}.}
    \item{\code{include_features}}{(\emph{optional}) Feature columns that are specifically
included in the data set. By default all features are included. Cannot
overlap with \code{exclude_features}, but may overlap \code{signature}. Features in
\code{signature} and \code{novelty_features} are always included. If both
\code{exclude_features} and \code{include_features} are provided, \code{include_features}
takes precedence, provided that there is no overlap between the two.}
    \item{\code{reference_method}}{(\emph{optional}) Method used to set reference levels for
categorical features. There are several options:
\itemize{
\item \code{auto} (default): Categorical features that are not explicitly set by the
user, i.e. columns containing boolean values or characters, use the most
frequent level as reference. Categorical features that are explicitly set,
i.e. as factors, are used as is.
\item \code{always}: Both automatically detected and user-specified categorical
features have the reference level set to the most frequent level. Ordinal
features are not altered, but are used as is.
\item \code{never}: User-specified categorical features are used as is.
Automatically detected categorical features are simply sorted, and the
first level is then used as the reference level. This was the behaviour
prior to familiar version 1.3.0.
}}
    \item{\code{imbalance_correction_method}}{(\emph{optional}) Type of method used to
address class imbalances. Available options are:
\itemize{
\item \code{full_undersampling} (default): All data will be used in an ensemble
fashion. The full minority class will appear in each partition, but
majority classes are undersampled until all data have been used.
\item \code{random_undersampling}: Randomly undersamples majority classes. This is
useful in cases where full undersampling would lead to the formation of
many models due major overrepresentation of the largest class.
}

This parameter is only used in combination with imbalance partitioning in
the experimental design, and \code{ip} should therefore appear in the string
that defines the design.}
    \item{\code{imbalance_n_partitions}}{(\emph{optional}) Number of times random
undersampling should be repeated. 10 undersampled subsets with balanced
classes are formed by default.}
    \item{\code{iteration_seed}}{(\emph{optional}) Integer seed used in sampling algorithms
specified by the \code{experimental_design} argument. This allows for creating
the same sample assignments across different experiments -- of course
provided that the same dataset is used. By default a random seed is used.}
    \item{\code{parallel}}{(\emph{optional}) Enable parallel processing. Defaults to \code{TRUE}.
When set to \code{FALSE}, this disables all parallel processing, regardless of
specific parameters such as \code{parallel_preprocessing}. However, when
\code{parallel} is \code{TRUE}, parallel processing of different parts of the
workflow can be disabled by setting respective flags to \code{FALSE}.}
    \item{\code{parallel_nr_cores}}{(\emph{optional}) Number of cores available for
parallelisation. Defaults to 2. This setting does nothing if
parallelisation is disabled.}
    \item{\code{restart_cluster}}{(\emph{optional}) Restart nodes used for parallel computing
to free up memory prior to starting a parallel process. Note that it does
take time to set up the clusters. Therefore setting this argument to \code{TRUE}
may impact processing speed. This argument is ignored if \code{parallel} is
\code{FALSE} or the cluster was initialised outside of familiar. Default is
\code{FALSE}, which causes the clusters to be initialised only once.}
    \item{\code{cluster_type}}{(\emph{optional}) Selection of the cluster type for parallel
processing. Available types are the ones supported by the parallel package
that is part of the base R distribution: \code{psock} (default), \code{fork}, \code{mpi},
\code{nws}, \code{sock}. In addition, \code{none} is available, which also disables
parallel processing.}
    \item{\code{backend_type}}{(\emph{optional}) Selection of the backend for distributing
copies of the data. This backend ensures that only a single master copy is
kept in memory. This limits memory usage during parallel processing.

Several backend options are available, notably \code{socket_server}, and \code{none}
(default). \code{socket_server} is based on the callr package and R sockets,
comes with \code{familiar} and is available for any OS. \code{none} uses the package
environment of familiar to store data, and is available for any OS.
However, \code{none} requires copying of data to any parallel process, and has a
larger memory footprint.}
    \item{\code{server_port}}{(\emph{optional}) Integer indicating the port on which the
socket server or RServe process should communicate. Defaults to port 6311.
Note that ports 0 to 1024 and 49152 to 65535 cannot be used.}
    \item{\code{feature_max_fraction_missing}}{(\emph{optional}) Numeric value between \code{0.0}
and \code{0.95} that determines the meximum fraction of missing values that
still allows a feature to be included in the data set. All features with a
missing value fraction over this threshold are not processed further. The
default value is \code{0.30}.}
    \item{\code{sample_max_fraction_missing}}{(\emph{optional}) Numeric value between \code{0.0}
and \code{0.95} that determines the maximum fraction of missing values that
still allows a sample to be included in the data set. All samples with a
missing value fraction over this threshold are excluded and not processed
further. The default value is \code{0.30}.}
    \item{\code{filter_method}}{(\emph{optional}) One or methods used to reduce
dimensionality of the data set by removing irrelevant or poorly
reproducible features.

Several method are available:
\itemize{
\item \code{none} (default): None of the features will be filtered.
\item \code{low_variance}: Features with a variance below the
\code{low_var_minimum_variance_threshold} are filtered. This can be useful to
filter, for example, genes that are not differentially expressed.
\item \code{univariate_test}: Features undergo a univariate regression using an
outcome-appropriate regression model. The p-value of the model coefficient
is collected. Features with coefficient p-value above the
\code{univariate_test_threshold} are subsequently filtered.
\item \code{robustness}: Features that are not sufficiently robust according to the
intraclass correlation coefficient are filtered. Use of this method
requires that repeated measurements are present in the data set, i.e. there
should be entries for which the sample and cohort identifiers are the same.
}

More than one method can be used simultaneously. Features with singular
values are always filtered, as these do not contain information.}
    \item{\code{univariate_test_threshold}}{(\emph{optional}) Numeric value between \code{1.0} and
\code{0.0} that determines which features are irrelevant and will be filtered by
the \code{univariate_test}. p-values are compared to this threshold.
All features with values above the threshold are filtered. The default
value is \code{0.20}.}
    \item{\code{univariate_test_threshold_metric}}{(\emph{optional}) Metric used with the to
compare the \code{univariate_test_threshold} against. The following metric can
be chosen:
\itemize{
\item \code{p_value} (default): The unadjusted p-value of each feature is used for
to filter features.
}}
    \item{\code{univariate_test_max_feature_set_size}}{(\emph{optional}) Maximum size of the
feature set after the univariate test. P or q values of features are
compared against the threshold, but if the resulting data set would be
larger than this setting, only the most relevant features up to the desired
feature set size are selected.

The default value is \code{NULL}, which causes features to be filtered based on
their relevance only.}
    \item{\code{low_var_minimum_variance_threshold}}{(required, if used) Numeric value
that determines which features will be filtered by the \code{low_variance}
method. The variance of each feature is computed and compared to the
threshold. If it is below the threshold, the feature is removed.

This parameter has no default value and should be set if \code{low_variance} is
used.}
    \item{\code{low_var_max_feature_set_size}}{(\emph{optional}) Maximum size of the feature
set after filtering features with a low variance. All features are first
compared against \code{low_var_minimum_variance_threshold}. If the resulting
feature set would be larger than specified, only the most strongly varying
features will be selected, up to the desired size of the feature set.

The default value is \code{NULL}, which causes features to be filtered based on
their variance only.}
    \item{\code{robustness_icc_type}}{(\emph{optional}) String indicating the type of
intraclass correlation coefficient (\code{1}, \code{2} or \code{3}) that should be used to
compute robustness for features in repeated measurements. These types
correspond to the types in Shrout and Fleiss (1979). The default value is
\code{1}.}
    \item{\code{robustness_threshold_metric}}{(\emph{optional}) String indicating which
specific intraclass correlation coefficient (ICC) metric should be used to
filter features. This should be one of:
\itemize{
\item \code{icc}: The estimated ICC value itself.
\item \code{icc_low} (default): The estimated lower limit of the 95\% confidence
interval of the ICC, as suggested by Koo and Li (2016).
\item \code{icc_panel}: The estimated ICC value over the panel average, i.e. the ICC
that would be obtained if all repeated measurements were averaged.
\item \code{icc_panel_low}: The estimated lower limit of the 95\% confidence interval
of the panel ICC.
}}
    \item{\code{robustness_threshold_value}}{(\emph{optional}) The intraclass correlation
coefficient value that is as threshold. The default value is \code{0.70}.}
    \item{\code{transformation_method}}{(\emph{optional}) The transformation method used to
change the distribution of the data to be more normal-like. The following
methods are available:
\itemize{
\item \code{none}: This disables transformation of features.
\item \code{yeo_johnson}: Transformation using the location and scale invariant
version of the Yeo-Johnson transformation (Yeo and Johnson, 2000;
Zwanenburg and Löck, 2026).
\item \code{yeo_johnson_robust} (default): A robust version of \code{yeo_johnson}.
This method is less sensitive to outliers.
\item \code{yeo_johnson_conventional}: As \code{yeo_johnson}, but without optimisation of
location and scale parameters. This method is equivalent to the original
transformation proposed by Yeo and Johnson (2001).
\item \code{box_cox}: Transformation using the location and scale invariant version
of the Box-Cox transformation (Box and Cox, 1964; Zwanenburg and Löck,
2026).
\item \code{box_cox_robust}: A robust version of \code{yeo_johnson}. This method is less
sensitive to outliers.
\item \code{box_cox_conventional}: As \code{box_cox}, but without optimisation of
location and scale parameters. This method is equivalent to the original
transformation proposed by Box and Cox (1964). This method requires
strictly positive feature values.
}

Transformation requires the \code{power.transform} package. Only features that
contain numerical data are transformed. Transformation parameters obtained
in development data are stored within \code{featureInfo} objects for later use
with validation data sets.}
    \item{\code{transformation_optimisation_criterion}}{(\emph{optional}) Transformation
parameters are optimised using a criterion, conventionally
maximum-likelihood-estimation. \code{power.transform} implements multiple
optimisation criteria, of which the following are available:
\itemize{
\item \code{mle} (default): Optimisation using maximum likelihood estimation.
\item \code{cramer_von_mises}: Optimisation using the Cramér-von Mises
criterion. Zwanenburg and Löck (2026) found that this criterion was
relatively robust against outliers.
}}
    \item{\code{transformation_gof_test_p_value}}{(\emph{optional}) Not all transformations
will lead to features that are roughly normally distributed. Zwanenburg and
Löck (2026) established a empirical goodness-of-fit test for central
normality. This parameter sets the significance for rejecting the
null-hypothesis that a feature distribution is centrally normal. When the
null-hypothesis is rejected, no transformation is performed. The default
value is \code{NULL}, which disables the test.}
    \item{\code{normalisation_method}}{(\emph{optional}) The normalisation method used to
improve the comparability between numerical features that may have very
different scales. The following normalisation methods can be chosen:
\itemize{
\item \code{none}: This disables feature normalisation.
\item \code{standardisation}: Features are normalised by subtraction of their mean
values and division by their standard deviations. This causes every feature
to be have a center value of 0.0 and standard deviation of 1.0.
\item \code{standardisation_trim}: As \code{standardisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are discarded.
This reduces the effect of outliers.
\item \code{standardisation_winsor}: As \code{standardisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are winsorised.
This reduces the effect of outliers.
\item \code{standardisation_robust} (default): A robust version of \code{standardisation}
that relies on computing Huber's M-estimators for location and scale.
\item \code{normalisation}: Features are normalised by subtraction of their minimum
values and division by their ranges. This maps all feature values to a
\eqn{[0, 1]} interval.
\item \code{normalisation_trim}: As \code{normalisation}, but based on the set of feature
values where the 5\% lowest and 5\% highest values are discarded. This
reduces the effect of outliers.
\item \code{normalisation_winsor}: As \code{normalisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are winsorised.
This reduces the effect of outliers.
\item \code{quantile}: Features are normalised by subtraction of their median values
and division by their interquartile range.
\item \code{mean_centering}: Features are centered by substracting the mean, but do
not undergo rescaling.
}

Only features that contain numerical data are normalised. Normalisation
parameters obtained in development data are stored within \code{featureInfo}
objects for later use with validation data sets.}
    \item{\code{batch_normalisation_method}}{(\emph{optional}) The method used for batch
normalisation. Available methods are:
\itemize{
\item \code{none} (default): This disables batch normalisation of features.
\item \code{standardisation}: Features within each batch are normalised by
subtraction of the mean value and division by the standard deviation in
each batch.
\item \code{standardisation_trim}: As \code{standardisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are discarded.
This reduces the effect of outliers.
\item \code{standardisation_winsor}: As \code{standardisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are winsorised.
This reduces the effect of outliers.
\item \code{standardisation_robust}: A robust version of \code{standardisation} that
relies on computing Huber's M-estimators for location and scale within each
batch.
\item \code{normalisation}: Features within each batch are normalised by subtraction
of their minimum values and division by their range in each batch. This
maps all feature values in each batch to a \eqn{[0, 1]} interval.
\item \code{normalisation_trim}: As \code{normalisation}, but based on the set of feature
values where the 5\% lowest and 5\% highest values are discarded. This
reduces the effect of outliers.
\item \code{normalisation_winsor}: As \code{normalisation}, but based on the set of
feature values where the 5\% lowest and 5\% highest values are winsorised.
This reduces the effect of outliers.
\item \code{quantile}: Features in each batch are normalised by subtraction of the
median value and division by the interquartile range of each batch.
\item \code{mean_centering}: Features in each batch are centered on 0.0 by
substracting the mean value in each batch, but are not rescaled.
\item \code{combat_parametric}: Batch adjustments using parametric empirical Bayes
(Johnson et al, 2007). \code{combat_p} leads to the same method.
\item \code{combat_non_parametric}: Batch adjustments using non-parametric empirical
Bayes (Johnson et al, 2007). \code{combat_np} and \code{combat} lead to the same
method. Note that we reduced complexity from O(\eqn{n^2}) to O(\eqn{n}) by
only computing batch adjustment parameters for each feature on a subset of
50 randomly selected features, instead of all features.
}

Only features that contain numerical data are normalised using batch
normalisation. Batch normalisation parameters obtained in development data
are stored within \code{featureInfo} objects for later use with validation data
sets, in case the validation data is from the same batch.

If validation data contains data from unknown batches, normalisation
parameters are separately determined for these batches.

Note that for both empirical Bayes methods, the batch effect is assumed to
produce results across the features. This is often true for things such as
gene expressions, but the assumption may not hold generally.

When performing batch normalisation, it is moreover important to check that
differences between batches or cohorts are not related to the studied
endpoint.}
    \item{\code{imputation_method}}{(\emph{optional}) Method used for imputing missing
feature values. Two methods are implemented:
\itemize{
\item \code{simple}: Simple replacement of a missing value by the median value (for
numeric features) or the modal value (for categorical features).
\item \code{lasso}: Imputation of missing value by lasso regression (using \code{glmnet})
based on information contained in other features.
}

\code{simple} imputation precedes \code{lasso} imputation to ensure that any missing
values in predictors required for \code{lasso} regression are resolved. The
\code{lasso} estimate is then used to replace the missing value.

The default value depends on the number of features in the dataset. If the
number is lower than 100, \code{lasso} is used by default, and \code{simple}
otherwise.

Only single imputation is performed. Imputation models and parameters are
stored within \code{featureInfo} objects for later use with validation data
sets.}
    \item{\code{cluster_method}}{(\emph{optional}) Clustering is performed to identify and
replace redundant features, for example those that are highly correlated.
Such features do not carry much additional information and may be removed
or replaced instead (Park et al., 2007; Tolosi and Lengauer, 2011).

The cluster method determines the algorithm used to form the clusters. The
following cluster methods are implemented:
\itemize{
\item \code{none}: No clustering is performed.
\item \code{hclust} (default): Hierarchical agglomerative clustering. If the
\code{fastcluster} package is installed, \code{fastcluster::hclust} is used (Muellner
2013), otherwise \code{stats::hclust} is used.
\item \code{agnes}: Hierarchical clustering using agglomerative nesting (Kaufman and
Rousseeuw, 1990). This algorithm is similar to \code{hclust}, but uses the
\code{cluster::agnes} implementation.
\item \code{diana}: Divisive analysis hierarchical clustering. This method uses
divisive instead of agglomerative clustering (Kaufman and Rousseeuw, 1990).
\code{cluster::diana} is used.
\item \code{pam}: Partioning around medioids. This partitions the data into $k$
clusters around medioids (Kaufman and Rousseeuw, 1990). $k$ is selected
using the \code{silhouette} metric. \code{pam} is implemented using the
\code{cluster::pam} function.
}

Clusters and cluster information is stored within \code{featureInfo} objects for
later use with validation data sets. This enables reproduction of the same
clusters as formed in the development data set.}
    \item{\code{cluster_linkage_method}}{(\emph{optional}) Linkage method used for
agglomerative clustering in \code{hclust} and \code{agnes}. The following linkage
methods can be used:
\itemize{
\item \code{average} (default): Average linkage.
\item \code{single}: Single linkage.
\item \code{complete}: Complete linkage.
\item \code{weighted}: Weighted linkage, also known as McQuitty linkage.
\item \code{ward}: Linkage using Ward's minimum variance method.
}

\code{diana} and \code{pam} do not require a linkage method.}
    \item{\code{cluster_cut_method}}{(\emph{optional}) The method used to define the actual
clusters. The following methods can be used:
\itemize{
\item \code{silhouette}: Clusters are formed based on the silhouette score
(Rousseeuw, 1987). The average silhouette score is computed from 2 to
\eqn{n} clusters, with \eqn{n} the number of features. Clusters are only
formed if the average silhouette exceeds 0.50, which indicates reasonable
evidence for structure. This procedure may be slow if the number of
features is large (>100s).
\item \code{fixed_cut}: Clusters are formed by cutting the hierarchical tree at the
point indicated by the \code{cluster_similarity_threshold}, e.g. where features
in a cluster have an average Spearman correlation of 0.90. \code{fixed_cut} is
only available for \code{agnes}, \code{diana} and \code{hclust}.
\item \code{dynamic_cut}: Dynamic cluster formation using the cutting algorithm in
the \code{dynamicTreeCut} package. This package should be installed to select
this option. \code{dynamic_cut} can only be used with \code{agnes} and \code{hclust}.
}

The default options are \code{silhouette} for partioning around medioids (\code{pam})
and \code{fixed_cut} otherwise.}
    \item{\code{cluster_similarity_metric}}{(\emph{optional}) Clusters are formed based on
feature similarity. All features are compared in a pair-wise fashion to
compute similarity, for example correlation. The resulting similarity grid
is converted into a distance matrix that is subsequently used for
clustering. The following metrics are supported to compute pairwise
similarities:
\itemize{
\item \code{mutual_information} (default): normalised mutual information.
\item \code{mcfadden_r2}: McFadden's pseudo R-squared (McFadden, 1974).
\item \code{cox_snell_r2}: Cox and Snell's pseudo R-squared (Cox and Snell, 1989).
\item \code{nagelkerke_r2}: Nagelkerke's pseudo R-squared (Nagelkerke, 1991).
\item \code{spearman}: Spearman's rank order correlation.
\item \code{kendall}: Kendall rank correlation.
\item \code{pearson}: Pearson product-moment correlation.
}

The pseudo R-squared metrics can be used to assess similarity between mixed
pairs of numeric and categorical features, as these are based on the
log-likelihood of regression models. In \code{familiar}, the more informative
feature is used as the predictor and the other feature as the reponse
variable. In numeric-categorical pairs, the numeric feature is considered
to be more informative and is thus used as the predictor. In
categorical-categorical pairs, the feature with most levels is used as the
predictor.

In case any of the classical correlation coefficients (\code{pearson},
\code{spearman} and \code{kendall}) are used with (mixed) categorical features, the
categorical features are one-hot encoded and the mean correlation over all
resulting pairs is used as similarity.}
    \item{\code{cluster_similarity_threshold}}{(\emph{optional}) The threshold level for
pair-wise similarity that is required to form clusters using \code{fixed_cut}.
This should be a numerical value between 0.0 and 1.0. Note however, that a
reasonable threshold value depends strongly on the similarity metric. The
following are the default values used:
\itemize{
\item \code{mcfadden_r2} and \code{mutual_information}: \code{0.30}
\item \code{cox_snell_r2} and \code{nagelkerke_r2}: \code{0.75}
\item \code{spearman}, \code{kendall} and \code{pearson}: \code{0.90}
}

Alternatively, if the \verb{fixed cut} method is not used, this value determines
whether any clustering should be performed, because the data may not
contain highly similar features. The default values in this situation are:
\itemize{
\item \code{mcfadden_r2}  and \code{mutual_information}: \code{0.25}
\item \code{cox_snell_r2} and \code{nagelkerke_r2}: \code{0.40}
\item \code{spearman}, \code{kendall} and \code{pearson}: \code{0.70}
}

The threshold value is converted to a distance (1-similarity) prior to
cutting hierarchical trees.}
    \item{\code{cluster_representation_method}}{(\emph{optional}) Method used to determine
how the information of co-clustered features is summarised and used to
represent the cluster. The following methods can be selected:
\itemize{
\item \code{best_predictor} (default): The feature with the highest importance
according to univariate regression with the outcome is used to represent
the cluster.
\item \code{medioid}: The feature closest to the cluster center, i.e. the feature
that is most similar to the remaining features in the cluster, is used to
represent the feature.
\item \code{mean}: A meta-feature is generated by averaging the feature values for
all features in a cluster. This method aligns all features so that all
features will be positively correlated prior to averaging. Should a cluster
contain one or more categorical features, the \code{medioid} method will be used
instead, as averaging is not possible. Note that if this method is chosen,
the \code{normalisation_method} parameter should be one of \code{standardisation},
\code{standardisation_trim}, \code{standardisation_winsor} or \code{quantile}.`
}

If the \code{pam} cluster method is selected, only the \code{medioid} method can be
used. In that case 1 medioid is used by default.}
    \item{\code{parallel_preprocessing}}{(\emph{optional}) Enable parallel processing for the
preprocessing workflow. Defaults to \code{TRUE}. When set to \code{FALSE}, this will
disable the use of parallel processing while preprocessing, regardless of
the settings of the \code{parallel} parameter. \code{parallel_preprocessing} is
ignored if \code{parallel=FALSE}.}
  }}
}
\value{
An \code{experimentData} object.
}
\description{
Creates data assignment and subsequently extracts feature
information such as normalisation and clustering parameters.
}
\details{
This is a thin wrapper around \code{summon_familiar}, and functions like
it, but automatically skips computation of variable importance, learning
and subsequent evaluation steps.

The function returns an \code{experimentData} object, which can be used to
warm-start other experiments by providing it to the \code{experiment_data}
argument.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseSettings.R
\name{.parse_initial_settings}
\alias{.parse_initial_settings}
\title{Internal function for parsing settings required to parse the input data and
define the experiment}
\usage{
.parse_initial_settings(config = NULL, ...)
}
\arguments{
\item{config}{A list of settings, e.g. from an xml file.}

\item{...}{
  Arguments passed on to \code{\link[=.parse_experiment_settings]{.parse_experiment_settings}}, \code{\link[=.parse_setup_settings]{.parse_setup_settings}}
  \describe{
    \item{\code{batch_id_column}}{(\strong{recommended}) Name of the column containing batch
or cohort identifiers. This parameter is required if more than one dataset
is provided, or if external validation is performed.

In familiar any row of data is organised by four identifiers:
\itemize{
\item The batch identifier \code{batch_id_column}: This denotes the group to which a
set of samples belongs, e.g. patients from a single study, samples measured
in a batch, etc. The batch identifier is used for batch normalisation, as
well as selection of development and validation datasets.
\item The sample identifier \code{sample_id_column}: This denotes the sample level,
e.g. data from a single individual. Subsets of data, e.g. bootstraps or
cross-validation folds, are created at this level.
\item The series identifier \code{series_id_column}: Indicates measurements on a
single sample that may not share the same outcome value, e.g. a time
series, or the number of cells in a view.
\item The repetition identifier: Indicates repeated measurements in a single
series where any feature values may differ, but the outcome does not.
Repetition identifiers are always implicitly set when multiple entries for
the same series of the same sample in the same batch that share the same
outcome are encountered.
}}
    \item{\code{sample_id_column}}{(\strong{recommended}) Name of the column containing
sample or subject identifiers. See \code{batch_id_column} above for more
details.

If unset, every row will be identified as a single sample.}
    \item{\code{series_id_column}}{(\emph{optional}) Name of the column containing series
identifiers, which distinguish between measurements that are part of a
series for a single sample. See \code{batch_id_column} above for more details.

If unset, rows which share the same batch and sample identifiers but have a
different outcome are assigned unique series identifiers.}
    \item{\code{development_batch_id}}{(\emph{optional}) One or more batch or cohort
identifiers to constitute data sets for development. Defaults to all, or
all minus the identifiers in \code{validation_batch_id} for external validation.
Required if external validation is performed and \code{validation_batch_id} is
not provided.}
    \item{\code{validation_batch_id}}{(\emph{optional}) One or more batch or cohort
identifiers to constitute data sets for external validation. Defaults to
all data sets except those in \code{development_batch_id} for external
validation, or none if not. Required if \code{development_batch_id} is not
provided.}
    \item{\code{outcome_name}}{(\emph{optional}) Name of the modelled outcome. This name will
be used in figures created by \code{familiar}.

If not set, the column name in \code{outcome_column} will be used for
\code{binomial}, \code{multinomial}, and \code{continuous} outcomes. For other outcomes
(\code{survival} and \code{competing_risk}) no default is used.}
    \item{\code{outcome_column}}{(\strong{recommended}) Name of the column containing the
outcome of interest. May be identified from a formula, if a formula is
provided as an argument. Otherwise an error is raised. Note that \code{survival}
and \code{competing_risk} outcome type outcomes require two columns that
indicate the time-to-event or the time of last follow-up and the event
status.}
    \item{\code{outcome_type}}{(\strong{recommended}) Type of outcome found in the outcome
column. The outcome type determines many aspects of the overall process,
e.g. the available variable importance methods and learners, but also the
type of assessments that can be conducted to evaluate the resulting models.
Implemented outcome types are:
\itemize{
\item \code{binomial}: categorical outcome with 2 levels.
\item \code{multinomial}: categorical outcome with 2 or more levels.
\item \code{continuous}: general continuous numeric outcomes.
\item \code{survival}: survival outcome for time-to-event data.
}

If not provided, the algorithm will attempt to obtain outcome_type from
contents of the outcome column. This may lead to unexpected results, and we
therefore advise to provide this information manually.

Note that \code{competing_risk} survival analysis are not fully supported, and
is currently not a valid choice for \code{outcome_type}. The \code{count} outcome
type was deprecated in version 2.0.0, and superseded by \code{continuous}.}
    \item{\code{class_levels}}{(\emph{optional}) Class levels for \code{binomial} or \code{multinomial}
outcomes. This argument can be used to specify the ordering of levels for
categorical outcomes. These class levels must exactly match the levels
present in the outcome column.}
    \item{\code{event_indicator}}{(\strong{recommended}) Indicator for events in \code{survival}
and \code{competing_risk} analyses. \code{familiar} will automatically recognise \code{1},
\code{true}, \code{t}, \code{y} and \code{yes} as event indicators, including different
capitalisations. If this parameter is set, it replaces the default values.}
    \item{\code{censoring_indicator}}{(\strong{recommended}) Indicator for right-censoring in
\code{survival} and \code{competing_risk} analyses. \code{familiar} will automatically
recognise \code{0}, \code{false}, \code{f}, \code{n}, \code{no} as censoring indicators, including
different capitalisations. If this parameter is set, it replaces the
default values.}
    \item{\code{competing_risk_indicator}}{(\strong{recommended}) Indicator for competing
risks in \code{competing_risk} analyses. There are no default values, and if
unset, all values other than those specified by the \code{event_indicator} and
\code{censoring_indicator} parameters are considered to indicate competing
risks.}
    \item{\code{signature}}{(\emph{optional}) One or more names of feature columns that are
considered part of a specific signature. Features specified here will
always be used for modelling. Ranking of variable importances has no effect
for these features.}
    \item{\code{novelty_features}}{(\emph{optional}) One or more names of feature columns
that should be included for the purpose of novelty detection.}
    \item{\code{exclude_features}}{(\emph{optional}) Feature columns that will be removed
from the data set. Cannot overlap with features in \code{signature},
\code{novelty_features} or \code{include_features}.}
    \item{\code{include_features}}{(\emph{optional}) Feature columns that are specifically
included in the data set. By default all features are included. Cannot
overlap with \code{exclude_features}, but may overlap \code{signature}. Features in
\code{signature} and \code{novelty_features} are always included. If both
\code{exclude_features} and \code{include_features} are provided, \code{include_features}
takes precedence, provided that there is no overlap between the two.}
    \item{\code{reference_method}}{(\emph{optional}) Method used to set reference levels for
categorical features. There are several options:
\itemize{
\item \code{auto} (default): Categorical features that are not explicitly set by the
user, i.e. columns containing boolean values or characters, use the most
frequent level as reference. Categorical features that are explicitly set,
i.e. as factors, are used as is.
\item \code{always}: Both automatically detected and user-specified categorical
features have the reference level set to the most frequent level. Ordinal
features are not altered, but are used as is.
\item \code{never}: User-specified categorical features are used as is.
Automatically detected categorical features are simply sorted, and the
first level is then used as the reference level. This was the behaviour
prior to familiar version 1.3.0.
}}
    \item{\code{experimental_design}}{(\strong{required}) Defines what the experiment looks
like, e.g. \code{cv(bt(fs,20)+mb,3,2)+ev} for 2 times repeated 3-fold
cross-validation with nested variable importance computation on 20
bootstraps and model-building, and external validation. The basic workflow
components are:
\itemize{
\item \code{fs}: (optional) variable importance computation step. If not explicitly
declared, feature selection will be done just in time for hyperparameter
optimisation.
\item \code{mb}: (required) model building step.
\item \code{ev}: (optional) external validation. Note that internal validation due
to subsampling will always be conducted if the subsampling methods create
any validation data sets.
}

The different components are linked using \code{+}.

Different subsampling methods can be used in conjunction with the basic
workflow components:
\itemize{
\item \code{bs(x,n)}: (stratified) .632 bootstrap, with \code{n} the number of
bootstraps. In contrast to \code{bt}, feature pre-processing parameters and
hyperparameter optimisation are conducted on individual bootstraps.
\item \code{bt(x,n)}: (stratified) .632 bootstrap, with \code{n} the number of
bootstraps. Unlike \code{bs} and other subsampling methods, no separate
pre-processing parameters or optimised hyperparameters will be determined
for each bootstrap.
\item \code{cv(x,n,p)}: (stratified) \code{n}-fold cross-validation, repeated \code{p} times.
Pre-processing parameters are determined for each iteration.
\item \code{lv(x)}: leave-one-out-cross-validation. Pre-processing parameters are
determined for each iteration.
\item \code{ip(x)}: imbalance partitioning for addressing class imbalances on the
data set. Pre-processing parameters are determined for each partition. The
number of partitions generated depends on the imbalance correction method
(see the \code{imbalance_correction_method} parameter). Imbalance partitioning
does not generate validation sets.
}

As shown in the example above, sampling algorithms can be nested.

The simplest valid experimental design is \code{fs+mb}, which corresponds to a
TRIPOD type 1a analysis. Type 1b analyses are only possible using
bootstraps, e.g. \code{bt(fs+mb,100)}. Type 2a analyses can be conducted using
cross-validation, e.g. \code{cv(bt(fs,100)+mb,10,1)}. Depending on the origin of
the external validation data, designs such as \code{fs+mb+ev} or
\code{cv(bt(fs,100)+mb,10,1)+ev} constitute type 2b or type 3 analyses. Type 4
analyses can be done by obtaining one or more \code{familiarModel} objects from
others and applying them to your own data set.

Alternatively, the \code{experimental_design} parameter may be used to provide a
path to a file containing iterations, which is named \verb{####_iterations.RDS}
by convention. This path can be relative to the directory of the current
experiment (\code{experiment_dir}), or an absolute path. The absolute path may
thus also point to a file from a different experiment.}
    \item{\code{imbalance_correction_method}}{(\emph{optional}) Type of method used to
address class imbalances. Available options are:
\itemize{
\item \code{full_undersampling} (default): All data will be used in an ensemble
fashion. The full minority class will appear in each partition, but
majority classes are undersampled until all data have been used.
\item \code{random_undersampling}: Randomly undersamples majority classes. This is
useful in cases where full undersampling would lead to the formation of
many models due major overrepresentation of the largest class.
}

This parameter is only used in combination with imbalance partitioning in
the experimental design, and \code{ip} should therefore appear in the string
that defines the design.}
    \item{\code{imbalance_n_partitions}}{(\emph{optional}) Number of times random
undersampling should be repeated. 10 undersampled subsets with balanced
classes are formed by default.}
    \item{\code{iteration_seed}}{(\emph{optional}) Integer seed used in sampling algorithms
specified by the \code{experimental_design} argument. This allows for creating
the same sample assignments across different experiments -- of course
provided that the same dataset is used. By default a random seed is used.}
    \item{\code{parallel}}{(\emph{optional}) Enable parallel processing. Defaults to \code{TRUE}.
When set to \code{FALSE}, this disables all parallel processing, regardless of
specific parameters such as \code{parallel_preprocessing}. However, when
\code{parallel} is \code{TRUE}, parallel processing of different parts of the
workflow can be disabled by setting respective flags to \code{FALSE}.}
    \item{\code{parallel_nr_cores}}{(\emph{optional}) Number of cores available for
parallelisation. Defaults to 2. This setting does nothing if
parallelisation is disabled.}
    \item{\code{restart_cluster}}{(\emph{optional}) Restart nodes used for parallel computing
to free up memory prior to starting a parallel process. Note that it does
take time to set up the clusters. Therefore setting this argument to \code{TRUE}
may impact processing speed. This argument is ignored if \code{parallel} is
\code{FALSE} or the cluster was initialised outside of familiar. Default is
\code{FALSE}, which causes the clusters to be initialised only once.}
    \item{\code{cluster_type}}{(\emph{optional}) Selection of the cluster type for parallel
processing. Available types are the ones supported by the parallel package
that is part of the base R distribution: \code{psock} (default), \code{fork}, \code{mpi},
\code{nws}, \code{sock}. In addition, \code{none} is available, which also disables
parallel processing.}
    \item{\code{backend_type}}{(\emph{optional}) Selection of the backend for distributing
copies of the data. This backend ensures that only a single master copy is
kept in memory. This limits memory usage during parallel processing.

Several backend options are available, notably \code{socket_server}, and \code{none}
(default). \code{socket_server} is based on the callr package and R sockets,
comes with \code{familiar} and is available for any OS. \code{none} uses the package
environment of familiar to store data, and is available for any OS.
However, \code{none} requires copying of data to any parallel process, and has a
larger memory footprint.}
    \item{\code{server_port}}{(\emph{optional}) Integer indicating the port on which the
socket server or RServe process should communicate. Defaults to port 6311.
Note that ports 0 to 1024 and 49152 to 65535 cannot be used.}
  }}
}
\value{
A list of settings to be used for configuring the experiments.
}
\description{
This function parses settings required to parse the data set, e.g. determine
which columns are identfier columns, what column contains outcome data, which
type of outcome is it?
}
\details{
Three variants of parameters exist:
\itemize{
\item required: this parameter is required and must be set by the user.
\item recommended: not setting this parameter might cause an error to be thrown,
dependent on other input.
\item optional: these parameters have default values that may be altered if
required.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationRiskStratificationInfo.R
\name{extract_risk_stratification_info}
\alias{extract_risk_stratification_info}
\title{Internal function to extract risk stratification info from data.}
\usage{
extract_risk_stratification_info(
  object,
  detail_level = waiver(),
  message_indent = 0L,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarEnsemble}, which is an ensemble of one or more
\code{familiarModel} objects, or a \code{familiarDataElementPredictionTable} object
that contains prediction data.}

\item{detail_level}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}

\item{...}{Unused arguments.}
}
\value{
A list of familiarDataElements with risk stratification information.
}
\description{
Collects risk stratification information.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationCalibrationData.R
\name{.calibration_create_randomised_groups}
\alias{.calibration_create_randomised_groups}
\title{Create randomised groups Creates randomised groups, e.g. for tests that
depend on splitting (continuous) data into groups, such as the
Hosmer-Lemeshow test}
\usage{
.calibration_create_randomised_groups(
  x,
  y = NULL,
  sample_identifiers,
  n_max_groups = NULL,
  n_min_groups = NULL,
  n_min_y_in_group = NULL,
  unique_samples_only = TRUE,
  rstream_object = NULL
)
}
\arguments{
\item{x}{Vector with data used for sorting. Groups are formed based on
adjacent values.}

\item{y}{Vector with markers, e.g. the events. Should be 0 or 1 (for an
event).}

\item{sample_identifiers}{data.table with sample_identifiers. If provide, a
list of grouped sample_identifiers will be returned, and integers
otherwise.}

\item{n_max_groups}{Maximum number of groups that need to be formed.}

\item{n_min_groups}{Minimum number of groups that need to be formed.}

\item{n_min_y_in_group}{Minimum number of y=1 in each group for a valid
group.}
}
\value{
data.table with sample identifiers and random group ids.
}
\description{
The default fast mode is based on random sampling, whereas the slow mode is
based on probabilistic joining of adjacent groups. As the name suggests, fast
mode operates considerably more efficient.
}
\details{
Creates randomised groups, e.g. for tests that depend on splitting
(continuous) data into groups, such as the Hosmer-Lemeshow test
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseSettings.R
\name{.parse_setup_settings}
\alias{.parse_setup_settings}
\title{Internal function for parsing settings related to the computational setup}
\usage{
.parse_setup_settings(
  config = NULL,
  parallel = waiver(),
  parallel_nr_cores = waiver(),
  restart_cluster = waiver(),
  cluster_type = waiver(),
  backend_type = waiver(),
  server_port = waiver(),
  ...
)
}
\arguments{
\item{config}{A list of settings, e.g. from an xml file.}

\item{parallel}{(\emph{optional}) Enable parallel processing. Defaults to \code{TRUE}.
When set to \code{FALSE}, this disables all parallel processing, regardless of
specific parameters such as \code{parallel_preprocessing}. However, when
\code{parallel} is \code{TRUE}, parallel processing of different parts of the
workflow can be disabled by setting respective flags to \code{FALSE}.}

\item{parallel_nr_cores}{(\emph{optional}) Number of cores available for
parallelisation. Defaults to 2. This setting does nothing if
parallelisation is disabled.}

\item{restart_cluster}{(\emph{optional}) Restart nodes used for parallel computing
to free up memory prior to starting a parallel process. Note that it does
take time to set up the clusters. Therefore setting this argument to \code{TRUE}
may impact processing speed. This argument is ignored if \code{parallel} is
\code{FALSE} or the cluster was initialised outside of familiar. Default is
\code{FALSE}, which causes the clusters to be initialised only once.}

\item{cluster_type}{(\emph{optional}) Selection of the cluster type for parallel
processing. Available types are the ones supported by the parallel package
that is part of the base R distribution: \code{psock} (default), \code{fork}, \code{mpi},
\code{nws}, \code{sock}. In addition, \code{none} is available, which also disables
parallel processing.}

\item{backend_type}{(\emph{optional}) Selection of the backend for distributing
copies of the data. This backend ensures that only a single master copy is
kept in memory. This limits memory usage during parallel processing.

Several backend options are available, notably \code{socket_server}, and \code{none}
(default). \code{socket_server} is based on the callr package and R sockets,
comes with \code{familiar} and is available for any OS. \code{none} uses the package
environment of familiar to store data, and is available for any OS.
However, \code{none} requires copying of data to any parallel process, and has a
larger memory footprint.}

\item{server_port}{(\emph{optional}) Integer indicating the port on which the
socket server or RServe process should communicate. Defaults to port 6311.
Note that ports 0 to 1024 and 49152 to 65535 cannot be used.}

\item{...}{Unused arguments.}
}
\value{
List of parameters related to the computational setup.
}
\description{
Internal function for parsing settings related to the computational setup
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{dataObject-class}
\alias{dataObject-class}
\title{Data object}
\description{
The dataObject class is used to resolve the issue of keeping track of
pre-processing status and data loading inside complex workflows, e.g. nested
predict functions inside a calibration function.
}
\section{Slots}{

\describe{
\item{\code{data}}{NULL or data table containing the data. This is the data which
will be read and used.}

\item{\code{preprocessing_level}}{character indicating the level of pre-processing
already conducted.}

\item{\code{outcome_type}}{character, determines the outcome type.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters. Optional.}

\item{\code{data_column_info}}{Object containing column information.}

\item{\code{data_id}}{Data identifier for dataset. Set using internal routines if the
\code{dataObject} was created from a \code{delayedDataObject}}

\item{\code{run_id}}{Run identifier for dataset. Set using internal routines if the
\code{dataObject} was created from a \code{delayedDataObject}}

\item{\code{validation}}{Identifies if validation or development samples were loaded.
Set using internal routines if the \code{dataObject} was created from a
\code{delayedDataObject}.}

\item{\code{sample_seed}}{Seed used for creating a bootstrap of the data.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationFeatureExpression.R
\name{export_feature_expressions}
\alias{export_feature_expressions}
\alias{export_feature_expressions,familiarCollection-method}
\alias{export_feature_expressions,ANY-method}
\title{Extract and export feature expressions.}
\usage{
export_feature_expressions(
  object,
  dir_path = NULL,
  evaluation_time = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{export_feature_expressions}{familiarCollection}(
  object,
  dir_path = NULL,
  evaluation_time = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{export_feature_expressions}{ANY}(
  object,
  dir_path = NULL,
  evaluation_time = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{evaluation_time}{One or more time points that are used to create the
outcome columns in expression plots. If not provided explicitly, this
parameter is read from settings used at creation of the underlying
\code{familiarData} objects. Only used for \code{survival} outcomes.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=extract_feature_expression]{extract_feature_expression}}, \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[=as_data_object]{as_data_object}}
  \describe{
    \item{\code{feature_similarity}}{Table containing pairwise distance between
sample. This is used to determine cluster information, and indicate which
samples are similar. The table is created by the
\code{extract_sample_similarity} method.}
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{evaluation_times}}{One or more time points that are used for in analysis
of survival problems when data has to be assessed at a set time, e.g.
calibration. If not provided explicitly, this parameter is read from
settings used at creation of the underlying \code{familiarModel} objects. Only
used for \code{survival} outcomes.}
    \item{\code{feature_cluster_method}}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_linkage_method}}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_metric}}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{sample_cluster_method}}{The method used to perform clustering based on
distance between samples. These are the same methods as for the
\code{cluster_method} configuration parameter: \code{hclust}, \code{agnes}, \code{diana} and
\code{pam}.

\code{none} cannot be used when extracting data for feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{sample_linkage_method}}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{sample_similarity_metric}}{Metric to determine pairwise similarity
between samples. Similarity is computed in the same manner as for
clustering, but \code{sample_similarity_metric} has different options that are
better suited to computing distance between samples instead of between
features: \code{gower}, \code{euclidean}.

The underlying feature data is scaled to the \eqn{[0, 1]} range (for
numerical features) using the feature values across the samples. The
normalisation parameters required can optionally be computed from feature
data with the outer 5\% (on both sides) of feature values trimmed or
winsorised. To do so append \verb{_trim} (trimming) or \verb{_winsor} (winsorising) to
the metric name. This reduces the effect of outliers somewhat.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A data.table (if \code{dir_path} is not provided), or nothing, as all data
is exported to \code{csv} files.
}
\description{
Extract and export feature expressions for the features in a
familiarCollection.
}
\details{
All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Feature similarity data can be created from \code{dataObject}, or \code{data.table} objects.
For \code{data.table}, see \code{\link{as_data_object}} for additional arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.check_feature_availability}
\alias{.check_feature_availability}
\title{Internal function to check whether feature columns are found in the data}
\usage{
.check_feature_availability(data, feature)
}
\arguments{
\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{feature}{Character string(s) indicating one or more features.}
}
\description{
This function checks whether feature columns can be found in the data set.
It will raise an error if any feature columns are missing from the data set.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotKaplanMeier.R
\name{plot_kaplan_meier}
\alias{plot_kaplan_meier}
\alias{plot_kaplan_meier,ANY-method}
\alias{plot_kaplan_meier,familiarCollection-method}
\title{Plot Kaplan-Meier survival curves.}
\usage{
plot_kaplan_meier(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  linetype_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  combine_legend = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = "time",
  x_label_shared = "column",
  y_label = "survival probability",
  y_label_shared = "row",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_range = c(0, 1),
  y_n_breaks = 5L,
  y_breaks = NULL,
  confidence_level = NULL,
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  censoring = TRUE,
  censor_shape = "plus",
  show_logrank = TRUE,
  show_survival_table = TRUE,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_kaplan_meier}{ANY}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  linetype_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  combine_legend = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = "time",
  x_label_shared = "column",
  y_label = "survival probability",
  y_label_shared = "row",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_range = c(0, 1),
  y_n_breaks = 5L,
  y_breaks = NULL,
  confidence_level = NULL,
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  censoring = TRUE,
  censor_shape = "plus",
  show_logrank = TRUE,
  show_survival_table = TRUE,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_kaplan_meier}{familiarCollection}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  linetype_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  combine_legend = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = "time",
  x_label_shared = "column",
  y_label = "survival probability",
  y_label_shared = "row",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_range = c(0, 1),
  y_n_breaks = 5L,
  y_breaks = NULL,
  confidence_level = NULL,
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  censoring = TRUE,
  censor_shape = "plus",
  show_logrank = TRUE,
  show_survival_table = TRUE,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{\code{familiarCollection} object, or one or more \code{familiarData}
objects, that will be internally converted to a \code{familiarCollection}
object. It is also possible to provide a \code{familiarEnsemble} or one or more
\code{familiarModel} objects together with the data from which data is computed
prior to export. Paths to such files can also be provided.

Additionally, some \code{familiarData} objects can be created from prediction
tables (\code{familiarDataElementPredictionTable}). Other \code{familiarData} objects
can be created from data (\code{dataObject}, or \code{data.table}). Please
check \emph{details} for more information.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created figures are
saved to. Output is saved in the \code{stratification} subdirectory. If \code{NULL}
no figures are saved, but are returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{linetype_by}{(\emph{optional}) Variables that are used to determine the
linetype of lines in a plot. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
Sett details for available variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{combine_legend}{(\emph{optional}) Flag to indicate whether the same legend
is to be shared by multiple aesthetics, such as those specified by
\code{color_by} and \code{linetype_by} arguments.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring the plot elements
according to the groupings indicated by the \code{color_by} argument (if any).
\code{familiar} has a default palette. Other palettes are supported by
\code{paletteer}, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{x_label_shared}{(\emph{optional}) Sharing of x-axis labels between facets.
One of three values:
\itemize{
\item \code{overall}: A single label is placed at the bottom of the figure. Tick
text (but not the ticks themselves) is removed for all but the bottom facet
plot(s).
\item \code{column}: A label is placed at the bottom of each column. Tick text (but
not the ticks themselves) is removed for all but the bottom facet plot(s).
\item \code{individual}: A label is placed below each facet plot. Tick text is kept.
}}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{y_label_shared}{(\emph{optional}) Sharing of y-axis labels between facets.
One of three values:
\itemize{
\item \code{overall}: A single label is placed to the left of the figure. Tick text
(but not the ticks themselves) is removed for all but the left-most facet
plot(s).
\item \code{row}: A label is placed to the left of each row. Tick text (but not the
ticks themselves) is removed for all but the left-most facet plot(s).
\item \code{individual}: A label is placed below each facet plot. Tick text is kept.
}}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{x_n_breaks}{(\emph{optional}) Number of breaks to show on the x-axis of the
plot. \code{x_n_breaks} is used to determine the \code{x_breaks} argument in case it
is unset.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{y_range}{(\emph{optional}) Value range for the y-axis.}

\item{y_n_breaks}{(\emph{optional}) Number of breaks to show on the y-axis of the
plot. \code{y_n_breaks} is used to determine the \code{y_breaks} argument in case it
is unset.}

\item{y_breaks}{(\emph{optional}) Break points on the y-axis of the plot.}

\item{confidence_level}{(\emph{optional}) Confidence level for the strata in the
plot.}

\item{conf_int_style}{(\emph{optional}) Confidence interval style. See details for
allowed styles.}

\item{conf_int_alpha}{(\emph{optional}) Alpha value to determine transparency of
confidence intervals or, alternatively, other plot elements with which the
confidence interval overlaps. Only values between 0.0 (fully transparent)
and 1.0 (fully opaque) are allowed.}

\item{censoring}{(\emph{optional}) Flag to indicate whether censored samples
should be indicated on the survival curve.}

\item{censor_shape}{(\emph{optional}) Shape used to indicate censored samples on
the survival curve. Available shapes are documented in the \code{ggplot2}
vignette \emph{Aesthetic specifications}. By default a plus shape is used.}

\item{show_logrank}{(\emph{optional}) Specifies whether the results of a logrank
test to assess differences between the risk strata is annotated in the
plot. A log-rank test can only be shown when \code{color_by} and \code{linestyle_by}
are either unset, or only contain \code{group}.}

\item{show_survival_table}{(\emph{optional}) Specifies whether a survival table is
shown below the Kaplan-Meier survival curves. Survival in the risk strata
is assessed for each of the breaks in \code{x_breaks}.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from number of facets and the inclusion of survival tables.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}, \code{\link{extract_risk_stratification_data}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{ensemble_method}}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This function creates Kaplan-Meier survival curves from
stratification data stored in a familiarCollection object.
}
\details{
This function generates a Kaplan-Meier survival plot based on risk
group stratification by the learners.

\code{familiar} does not determine what units the x-axis has or what kind of
survival the y-axis represents. It is therefore recommended to provide
\code{x_label} and \code{y_label} arguments.

Available splitting variables are: \code{vimp_method}, \code{learner}, \code{data_set},
\code{group} and \code{stratification_method}. By default, separate figures are
created for each combination of \code{vimp_method} and \code{learner}, with faceting by
\code{data_set}, colouring of the strata in each individual plot by
\code{group}.

Greenwood confidence intervals of the Kaplan-Meier curve can be shown using
various styles set by \code{conf_int_style}:
\itemize{
\item \code{ribbon} (default): confidence intervals are shown as a ribbon with an
opacity of \code{conf_int_alpha} around the point estimate of the Kaplan-Meier
curve.
\item \code{step} (default): confidence intervals are shown as a step function around
the point estimate of the Kaplan-Meier curve.
\item \code{none}: confidence intervals are not shown. The point estimate of the ROC
curve is shown as usual.
}

Labelling methods such as \code{set_risk_group_names} or \code{set_data_set_names}
can be applied to the \code{familiarCollection} object to update labels, and
order the output in the figure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarModel.R
\name{vcov}
\alias{vcov}
\alias{vcov,familiarModel-method}
\title{Calculate variance-covariance matrix for a model}
\usage{
vcov(object, ...)

\S4method{vcov}{familiarModel}(object, ...)
}
\arguments{
\item{object}{a familiarModel object}

\item{...}{additional arguments passed to \code{vcov} methods for the underlying
model, when available.}
}
\value{
Variance-covariance matrix of the model in the familiarModel object,
if any.
}
\description{
Calculate variance-covariance matrix for a model
}
\details{
This method extends the \code{vcov} S3 method. For some models \code{vcov}
requires information that is trimmed from the model. In this case a copy of
the variance-covariance matrix is stored with the model, and returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotUnivariateImportance.R
\name{plot_univariate_importance}
\alias{plot_univariate_importance}
\alias{plot_univariate_importance,ANY-method}
\alias{plot_univariate_importance,familiarCollection-method}
\title{Plot univariate importance.}
\usage{
plot_univariate_importance(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  p_adjustment_method = waiver(),
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  show_cluster = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = waiver(),
  x_label = waiver(),
  y_label = "feature",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  significance_level_shown = 0.05,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  verbose = TRUE,
  export_collection = FALSE,
  ...
)

\S4method{plot_univariate_importance}{ANY}(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  p_adjustment_method = waiver(),
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  show_cluster = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = waiver(),
  x_label = waiver(),
  y_label = "feature",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  significance_level_shown = 0.05,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  verbose = TRUE,
  export_collection = FALSE,
  ...
)

\S4method{plot_univariate_importance}{familiarCollection}(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  p_adjustment_method = waiver(),
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  show_cluster = TRUE,
  ggtheme = NULL,
  discrete_palette = NULL,
  gradient_palette = waiver(),
  x_label = waiver(),
  y_label = "feature",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  significance_level_shown = 0.05,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  verbose = TRUE,
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{feature_cluster_method}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_cluster_cut_method}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_threshold}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created figures are
saved to. Output is saved in the \code{variable_importance} subdirectory. If
NULL no figures are saved, but are returned instead.}

\item{p_adjustment_method}{(\emph{optional}) Indicates type of p-value that is
shown. One of \code{holm}, \code{hochberg}, \code{hommel}, \code{bonferroni}, \code{BH}, \code{BY},
\code{fdr}, \code{none}, \code{p_value} or \code{q_value} for adjusted p-values, uncorrected
p-values}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{show_cluster}{(\emph{optional}) Show which features were clustered together.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring the plot elements
according to the groupings indicated by the \code{color_by} argument (if any).
\code{familiar} has a default palette. Other palettes are supported by
\code{paletteer}, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{gradient_palette}{(\emph{optional}) Palette to use for filling the bars in
case the \code{color_by} argument is not set. The bars are then coloured
according to their importance. By default, no gradient is used, and the
bars are not filled according to importance. Use \code{NULL} to fill the bars
using the default palette in \code{familiar}. Other palettes are supported by
the \code{paletteer} package, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{limit_n_features}{(\emph{optional}) The number of features that should be
included in the plot. Only the most important features are shown. By
default, the number of features is not limited.}

\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{x_n_breaks}{(\emph{optional}) Number of breaks to show on the x-axis of the
plot. \code{x_n_breaks} is used to determine the \code{x_breaks} argument in case it
is unset.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{significance_level_shown}{Position(s) to draw vertical lines indicating
a significance level, e.g. 0.05. Can be NULL to not draw anything.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from the number of features and the number of facets.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{verbose}{Flag to indicate whether feedback should be provided for the
plotting.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}, \code{\link[=extract_univariate_analysis]{extract_univariate_analysis}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{feature_similarity_metric}}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{icc_type}}{String indicating the type of intraclass correlation
coefficient (\code{1}, \code{2} or \code{3}) that should be used to compute robustness for
features in repeated measurements during the evaluation of univariate
importance. These types correspond to the types in Shrout and Fleiss (1979).
If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This function plots the univariate analysis data stored in a
familiarCollection object.
}
\details{
This function generates a horizontal barplot with the length of the
bars corresponding to the 10-logarithm of the (multiple-testing corrected)
p-value.

Features are assessed univariately using one-sample location t-tests after
fitting a suitable regression model. The fitted model coefficient and the
covariance matrix are then used to compute a p-value.

The following splitting variables are available for \code{split_by}, \code{color_by}
and \code{facet_by}:
\itemize{
\item \code{vimp_method}: variable importance methods
\item \code{learner}: learners
\item \code{data_set}: data sets
}

Unlike for plots of feature ranking in initial variable importance
computation and after modelling (as assessed by model-specific routines),
clusters of features are now found during creation of underlying
\code{familiarData} objects, instead of through consensus clustering. Hence,
clustering results may differ due to differences in the underlying
datasets.

Labelling methods such as \code{set_vimp_method_names} or \code{set_feature_names} can
be applied to the \code{familiarCollection} object to update labels, and order
the output in the figure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParseData.R
\name{.finish_data_preparation}
\alias{.finish_data_preparation}
\title{Internal function for finalising generic data processing}
\usage{
.finish_data_preparation(
  data,
  sample_id_column,
  batch_id_column,
  series_id_column,
  outcome_column,
  outcome_type,
  include_features,
  class_levels,
  censoring_indicator,
  event_indicator,
  competing_risk_indicator,
  check_stringency = "strict",
  reference_method = "auto",
  .no_features_required = FALSE
)
}
\arguments{
\item{data}{data.table with feature data}

\item{sample_id_column}{(\strong{recommended}) Name of the column containing
sample or subject identifiers. See \code{batch_id_column} above for more
details.

If unset, every row will be identified as a single sample.}

\item{batch_id_column}{(\strong{recommended}) Name of the column containing batch
or cohort identifiers. This parameter is required if more than one dataset
is provided, or if external validation is performed.

In familiar any row of data is organised by four identifiers:
\itemize{
\item The batch identifier \code{batch_id_column}: This denotes the group to which a
set of samples belongs, e.g. patients from a single study, samples measured
in a batch, etc. The batch identifier is used for batch normalisation, as
well as selection of development and validation datasets.
\item The sample identifier \code{sample_id_column}: This denotes the sample level,
e.g. data from a single individual. Subsets of data, e.g. bootstraps or
cross-validation folds, are created at this level.
\item The series identifier \code{series_id_column}: Indicates measurements on a
single sample that may not share the same outcome value, e.g. a time
series, or the number of cells in a view.
\item The repetition identifier: Indicates repeated measurements in a single
series where any feature values may differ, but the outcome does not.
Repetition identifiers are always implicitly set when multiple entries for
the same series of the same sample in the same batch that share the same
outcome are encountered.
}}

\item{series_id_column}{(\emph{optional}) Name of the column containing series
identifiers, which distinguish between measurements that are part of a
series for a single sample. See \code{batch_id_column} above for more details.

If unset, rows which share the same batch and sample identifiers but have a
different outcome are assigned unique series identifiers.}

\item{outcome_column}{(\strong{recommended}) Name of the column containing the
outcome of interest. May be identified from a formula, if a formula is
provided as an argument. Otherwise an error is raised. Note that \code{survival}
and \code{competing_risk} outcome type outcomes require two columns that
indicate the time-to-event or the time of last follow-up and the event
status.}

\item{outcome_type}{(\strong{recommended}) Type of outcome found in the outcome
column. The outcome type determines many aspects of the overall process,
e.g. the available variable importance methods and learners, but also the
type of assessments that can be conducted to evaluate the resulting models.
Implemented outcome types are:
\itemize{
\item \code{binomial}: categorical outcome with 2 levels.
\item \code{multinomial}: categorical outcome with 2 or more levels.
\item \code{continuous}: general continuous numeric outcomes.
\item \code{survival}: survival outcome for time-to-event data.
}

If not provided, the algorithm will attempt to obtain outcome_type from
contents of the outcome column. This may lead to unexpected results, and we
therefore advise to provide this information manually.

Note that \code{competing_risk} survival analysis are not fully supported, and
is currently not a valid choice for \code{outcome_type}. The \code{count} outcome
type was deprecated in version 2.0.0, and superseded by \code{continuous}.}

\item{include_features}{(\emph{optional}) Feature columns that are specifically
included in the data set. By default all features are included. Cannot
overlap with \code{exclude_features}, but may overlap \code{signature}. Features in
\code{signature} and \code{novelty_features} are always included. If both
\code{exclude_features} and \code{include_features} are provided, \code{include_features}
takes precedence, provided that there is no overlap between the two.}

\item{class_levels}{(\emph{optional}) Class levels for \code{binomial} or \code{multinomial}
outcomes. This argument can be used to specify the ordering of levels for
categorical outcomes. These class levels must exactly match the levels
present in the outcome column.}

\item{censoring_indicator}{(\strong{recommended}) Indicator for right-censoring in
\code{survival} and \code{competing_risk} analyses. \code{familiar} will automatically
recognise \code{0}, \code{false}, \code{f}, \code{n}, \code{no} as censoring indicators, including
different capitalisations. If this parameter is set, it replaces the
default values.}

\item{event_indicator}{(\strong{recommended}) Indicator for events in \code{survival}
and \code{competing_risk} analyses. \code{familiar} will automatically recognise \code{1},
\code{true}, \code{t}, \code{y} and \code{yes} as event indicators, including different
capitalisations. If this parameter is set, it replaces the default values.}

\item{competing_risk_indicator}{(\strong{recommended}) Indicator for competing
risks in \code{competing_risk} analyses. There are no default values, and if
unset, all values other than those specified by the \code{event_indicator} and
\code{censoring_indicator} parameters are considered to indicate competing
risks.}

\item{check_stringency}{Specifies stringency of various checks. This is mostly:
\itemize{
\item \code{strict}: default value used for \code{summon_familiar}. Thoroughly checks
input data. Used internally for checking development data.
\item \code{external_warn}: value used for \code{extract_data} and related methods. Less
stringent checks, but will warn for possible issues. Used internally for
checking data for evaluation and explanation.
\item \code{external}: value used for external methods such as \code{predict}. Less
stringent checks, particularly for identifier and outcome columns, which may
be completely absent. Used internally for \code{predict}.
}}

\item{reference_method}{(\emph{optional}) Method used to set reference levels for
categorical features. There are several options:
\itemize{
\item \code{auto} (default): Categorical features that are not explicitly set by the
user, i.e. columns containing boolean values or characters, use the most
frequent level as reference. Categorical features that are explicitly set,
i.e. as factors, are used as is.
\item \code{always}: Both automatically detected and user-specified categorical
features have the reference level set to the most frequent level. Ordinal
features are not altered, but are used as is.
\item \code{never}: User-specified categorical features are used as is.
Automatically detected categorical features are simply sorted, and the
first level is then used as the reference level. This was the behaviour
prior to familiar version 1.3.0.
}}

\item{.no_features_required}{Internal flag to signify that data without
features is allowed. Default: FALSE (most processing steps require features).}
}
\value{
data.table with expected column names.
}
\description{
Internal function for finalising generic data processing
}
\details{
This function is used to update data.table provided by loading the
data. When part of the main familiar workflow, this function is used after
.parse_initial_settings --> .load_data --> .update_initial_settings.

When used to parse external data (e.g. in conjunction with familiarModel)
it follows after .load_data. Hence the function contains several checks
which are otherwise part of .update_initial_settings.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationVimp.R
\name{extract_model_vimp}
\alias{extract_model_vimp}
\title{Internal function to extract variable importance from models.}
\usage{
extract_model_vimp(
  object,
  data,
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  message_indent = 0L,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarEnsemble}, which is an ensemble of one or more
\code{familiarModel} objects, or a \code{familiarDataElementPredictionTable} object
that contains prediction data.}

\item{data}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}

\item{aggregation_method}{Method for aggregating variable importances for the
purpose of evaluation. Variable importances are determined during feature
selection steps and after training the model. Both types are evaluated, but
feature selection variable importance is only evaluated at run-time.

See the documentation for the \code{vimp_aggregation_method} argument in
\code{summon_familiar} for information concerning the different available
methods.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{rank_threshold}{The threshold used to  define the subset of highly
important features during evaluation.

See the documentation for the \code{vimp_aggregation_rank_threshold} argument in
\code{summon_familiar} for more information.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}

\item{...}{Unused arguments.}
}
\value{
A list containing variable importance information.
}
\description{
Aggregate variable importance from models in a
\code{familiarEnsemble}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utilities.R
\name{waiver}
\alias{waiver}
\title{Create a waiver object}
\usage{
waiver()
}
\value{
waiver object
}
\description{
This function is functionally identical to \code{ggplot2::waiver()} function and
creates a waiver object. A waiver object is an otherwise empty object that
serves the same purpose as \code{NULL}, i.e. as placeholder for a default value.
Because \code{NULL} can sometimes be a valid input argument, it can therefore not
be used to switch to an internal default value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{set_feature_names,familiarCollection-method}
\alias{set_feature_names,familiarCollection-method}
\alias{set_feature_names}
\title{Rename features for plotting and export}
\usage{
\S4method{set_feature_names}{familiarCollection}(x, old = NULL, new = NULL, order = NULL)
}
\arguments{
\item{x}{A familiarCollection object.}

\item{old}{(optional) Set of old labels to replace.}

\item{new}{Set of replacement labels. The number of replacement labels should
be equal to the number of provided old labels or the full number of labels.
If a subset of labels is to be replaced, both \code{old} and \code{new}
should be provided.}

\item{order}{(optional) Ordered set of replacement labels. This is used to
provide the order in which the labels should be placed, which affects e.g.
levels in a plot. If the ordering is not explicitly provided, the old
ordering is used.}
}
\value{
A familiarCollection object with updated labels.
}
\description{
Tabular exports and figures created from a familiarCollection
object can be customised by providing names for features.
}
\details{
Labels convert the internal naming for features to the requested
label at export or when plotting. This enables customisation without
redoing the analysis with renamed input data. Currently assigned labels can
be found using the \code{get_feature_names} method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the
familiarCollection class. * \code{\link{get_feature_names}} for obtaining
currently assigned feature names.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarNoveltyDetector-class}
\alias{familiarNoveltyDetector-class}
\title{Novelty detector.}
\description{
A familiarNoveltyDetector object is a self-contained model that can be
applied to generate out-of-distribution predictions for instances in a
dataset.
}
\details{
Note that these objects do not contain any data concerning outcome,
as this not relevant for (prospective) out-of-distribution detection.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the familiarNoveltyDetector object.}

\item{\code{model}}{The actual novelty detector trained using a specific algorithm,
e.g. a isolation forest from the \code{isotree} package.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters.}

\item{\code{data_column_info}}{Data information object containing information
regarding identifier column names.}

\item{\code{hyperparameters}}{Set of hyperparameters used to train the detector.}

\item{\code{hyperparameter_data}}{Information generated during hyperparameter
optimisation. Currently not used.}

\item{\code{calibration_model}}{Model used to convert raw output to statistical
probability of being out-of-distribution. Currently not used.}

\item{\code{learner}}{Learning algorithm used to create the novelty detector.}

\item{\code{vimp_method}}{Method used to determine variable importance for the
novelty detector.}

\item{\code{vimp_table}}{Variable importance table or list of variable importance
tables for the model.}

\item{\code{vimp_aggregation_method}}{Method used for aggregating variable importance
tables if more than one is present.)}

\item{\code{vimp_rank_threshold}}{Threshold used for some variable importance
aggregation methods.}

\item{\code{required_features}}{The set of features required for complete
reproduction, i.e. with imputation.}

\item{\code{model_features}}{The set of features that is used to train the detector.}

\item{\code{data_id}}{Internal identifier for the dataset used to train the detector.}

\item{\code{run_id}}{Internal identifier for the specific subset of the dataset used
used to train the detector.}

\item{\code{run_table}}{Run table for the data used to train the detector. Used
internally.}

\item{\code{is_trimmed}}{Flag that indicates whether the detector, stored in the
\code{model} slot, has been trimmed.}

\item{\code{trimmed_function}}{List of functions whose output has been captured prior
to trimming the model.}

\item{\code{messages}}{List of warning and error messages generated during training.}

\item{\code{project_id}}{Identifier of the project that generated the
familiarNoveltyDetector object.}

\item{\code{familiar_version}}{Version of the familiar package.}

\item{\code{package}}{Name of package(s) required to executed the detector itself,
e.g. \code{isotree}.}

\item{\code{package_version}}{Version of the packages mentioned in the \code{package}
attribute.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarCollection.R
\name{get_vimp_method_names,familiarCollection-method}
\alias{get_vimp_method_names,familiarCollection-method}
\alias{get_vimp_method_names}
\title{Get current variable importance method name labels}
\usage{
\S4method{get_vimp_method_names}{familiarCollection}(x)
}
\arguments{
\item{x}{A familiarCollection object.}
}
\value{
An ordered array of variable importance method name labels.
}
\description{
Variable importance methods in familiarCollection objects can
have custom names for export and plotting. This function retrieves the
currently assigned names.
}
\details{
Labels convert internal naming of variable importance methods to the
requested label at export or when plotting. Labels can be changed using the
\code{set_vimp_method_names} method.
}
\seealso{
\itemize{
\item \linkS4class{familiarCollection} for information concerning the familiarCollection class.
\item \code{\link{set_vimp_method_names}} for updating the name of variable importance methods and their ordering.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.check_outcome_type_plausibility}
\alias{.check_outcome_type_plausibility}
\title{Internal function for checking if the outcome type fits well to the data}
\usage{
.check_outcome_type_plausibility(
  data,
  outcome_type,
  outcome_column,
  censoring_indicator,
  event_indicator,
  competing_risk_indicator,
  check_stringency = "strict"
)
}
\arguments{
\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{outcome_type}{Character string indicating the type of outcome being
assessed.}

\item{outcome_column}{Name of the outcome column in the data set.}

\item{censoring_indicator}{Name of censoring indicator.}

\item{event_indicator}{Name of event indicator.}

\item{competing_risk_indicator}{Name of competing risk indicator.}

\item{check_stringency}{Specifies stringency of various checks. This is mostly:
\itemize{
\item \code{strict}: default value used for \code{summon_familiar}. Thoroughly checks
input data. Used internally for checking development data.
\item \code{external_warn}: value used for \code{extract_data} and related methods. Less
stringent checks, but will warn for possible issues. Used internally for
checking data for evaluation and explanation.
\item \code{external}: value used for external methods such as \code{predict}. Less
stringent checks, particularly for identifier and outcome columns, which may
be completely absent. Used internally for \code{predict}.
}}
}
\description{
This function may help identify if the outcome type is plausible
given the outcome data. In practice it also tests whether the outcome column
is actually correct given the outcome type.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarMetric-class}
\alias{familiarMetric-class}
\title{Model performance metric.}
\description{
Superclass for model performance objects.
}
\section{Slots}{

\describe{
\item{\code{metric}}{Performance metric.}

\item{\code{outcome_type}}{Type of outcome being predicted.}

\item{\code{name}}{Name of the performance metric.}

\item{\code{value_range}}{Range of the performance metric. Can be half-open.}

\item{\code{baseline_value}}{Value of the metric for trivial models, e.g. models that
always predict the median value, the majority class, or the mean hazard,
etc.}

\item{\code{higher_better}}{States whether higher metric values correspond to better
predictive model performance (e.g. accuracy) or not (e.g. root mean squared
error).}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Iterations.R
\name{.load_iterations}
\alias{.load_iterations}
\title{Internal function for loading iteration data from the file system}
\usage{
.load_iterations(file_dir, iteration_file = NULL)
}
\arguments{
\item{file_dir}{Path to directory where iteration files are stored.}
}
\value{
List containing:
\itemize{
\item \code{iteration_file_exists}: An indicator whether an iteration file was found.
\item \code{iteration_list}: The list of iterations (if available).
\item \code{project_id}: The unique project identifier (if available).
}
}
\description{
Loads iterations generated by \code{.create_iterations} that were created in a
previous session. If these are not available, this is indicated by setting a
return flag.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationVimp.R
\name{export_fs_vimp}
\alias{export_fs_vimp}
\alias{export_fs_vimp,familiarCollection-method}
\alias{export_fs_vimp,ANY-method}
\title{Extract and export feature selection variable importance.}
\usage{
export_fs_vimp(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{export_fs_vimp}{familiarCollection}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{export_fs_vimp}{ANY}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  aggregation_method = waiver(),
  rank_threshold = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{aggregate_results}{Flag that signifies whether results should be
aggregated for export.}

\item{aggregation_method}{(\emph{optional}) The method used to aggregate variable
importances over different data subsets, e.g. bootstraps. The following
methods can be selected:
\itemize{
\item \code{mean} (default): Use the mean rank of a feature over the subsets to
determine the aggregated feature rank.
\item \code{median}: Use the median rank of a feature over the subsets to determine
the aggregated feature rank.
\item \code{best}: Use the best rank the feature obtained in any subset to determine
the aggregated feature rank.
\item \code{worst}: Use the worst rank the feature obtained in any subset to
determine the aggregated feature rank.
\item \code{stability}: Use the frequency of the feature being in the subset of
highly ranked features as measure for the aggregated feature rank
(Meinshausen and Buehlmann, 2010).
\item \code{exponential}: Use a rank-weighted frequence of occurrence in the subset
of highly ranked features as measure for the aggregated feature rank (Haury
et al., 2011).
\item \code{borda}: Use the borda count as measure for the aggregated feature rank
(Wald et al., 2012).
\item \code{enhanced_borda}: Use an occurrence frequency-weighted borda count as
measure for the aggregated feature rank (Wald et al., 2012).
\item \code{truncated_borda}: Use borda count computed only on features within the
subset of highly ranked features.
\item \code{enhanced_truncated_borda}: Apply both the enhanced borda method and the
truncated borda method and use the resulting borda count as the aggregated
feature rank.
}}

\item{rank_threshold}{(\emph{optional}) The threshold used to define the subset of
highly important features. If not set, this threshold is determined by
maximising the variance in the occurrence value over all features over the
subset size.

This parameter is only relevant for \code{stability}, \code{exponential},
\code{enhanced_borda}, \code{truncated_borda} and \code{enhanced_truncated_borda} methods.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A data.table (if \code{dir_path} is not provided), or nothing, as all data
is exported to \code{csv} files.
}
\description{
Extract and export feature selection variable importance from a
familiarCollection.
}
\details{
Data, such as model performance and calibration information, is
usually collected from a \code{familiarCollection} object. However, you can also
provide one or more \code{familiarData} objects, that will be internally
converted to a \code{familiarCollection} object. Paths to the previous files can
also be provided.

Unlike other export function, export using \code{familiarEnsemble} or
\code{familiarModel} objects is not possible. This is because feature selection
variable importance is not stored within \code{familiarModel} objects.

All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Variable importance is based on the ranking produced by feature selection
routines. In case feature selection was performed repeatedly, e.g. using
bootstraps, feature ranks are first aggregated using the method defined by
the \code{aggregation_method}, some of which require a \code{rank_threshold} to
indicate a subset of most important features.

Information concerning highly similar features that form clusters is
provided as well. This information is based on consensus clustering of the
features. This clustering information is also used during aggregation to
ensure that co-clustered features are only taken into account once.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotPermutationVariableImportance.R
\name{plot_permutation_variable_importance}
\alias{plot_permutation_variable_importance}
\alias{plot_permutation_variable_importance,ANY-method}
\alias{plot_permutation_variable_importance,familiarCollection-method}
\title{Plot permutation variable importance.}
\usage{
plot_permutation_variable_importance(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = "feature",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  conf_int_style = c("point_line", "line", "bar_line", "none"),
  conf_int_alpha = 0.2,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_permutation_variable_importance}{ANY}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = "feature",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  conf_int_style = c("point_line", "line", "bar_line", "none"),
  conf_int_alpha = 0.2,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_permutation_variable_importance}{familiarCollection}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = "feature",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  limit_n_features = waiver(),
  x_range = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  conf_int_style = c("point_line", "line", "bar_line", "none"),
  conf_int_alpha = 0.2,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{\code{familiarCollection} object, or one or more \code{familiarData}
objects, that will be internally converted to a \code{familiarCollection}
object. It is also possible to provide a \code{familiarEnsemble} or one or more
\code{familiarModel} objects together with the data from which data is computed
prior to export. Paths to such files can also be provided.

Additionally, some \code{familiarData} objects can be created from prediction
tables (\code{familiarDataElementPredictionTable}). Other \code{familiarData} objects
can be created from data (\code{dataObject}, or \code{data.table}). Please
check \emph{details} for more information.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created figures are
saved to. Output is saved in the \code{variable_importance} subdirectory. If
NULL no figures are saved, but are returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring the plot elements
according to the groupings indicated by the \code{color_by} argument (if any).
\code{familiar} has a default palette. Other palettes are supported by
\code{paletteer}, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{limit_n_features}{(\emph{optional}) The number of features that should be
included in the plot. Only the most important features are shown. By
default, the number of features is not limited.}

\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{x_n_breaks}{(\emph{optional}) Number of breaks to show on the x-axis of the
plot. \code{x_n_breaks} is used to determine the \code{x_breaks} argument in case it
is unset.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{conf_int_style}{(\emph{optional}) Confidence interval style. See details for
allowed styles.}

\item{conf_int_alpha}{(\emph{optional}) Alpha value to determine transparency of
confidence intervals or, alternatively, other plot elements with which the
confidence interval overlaps. Only values between 0.0 (fully transparent)
and 1.0 (fully opaque) are allowed.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from the number of features and the number of facets.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}, \code{\link[=extract_permutation_vimp]{extract_permutation_vimp}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{evaluation_times}}{One or more time points that are used for in analysis
of survival problems when data has to be assessed at a set time, e.g.
calibration. If not provided explicitly, this parameter is read from
settings used at creation of the underlying \code{familiarModel} objects. Only
used for \code{survival} outcomes.}
    \item{\code{ensemble_method}}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}
    \item{\code{metric}}{One or more metrics for assessing model performance. See the
vignette on performance metrics for the available metrics. If not provided
explicitly, this parameter is read from settings used at creation of the
underlying \code{familiarModel} objects.}
    \item{\code{feature_cluster_method}}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_linkage_method}}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_cluster_cut_method}}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_threshold}}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_metric}}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{sample_limit}}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}
    \item{\code{n_important_features}}{(\emph{optional}) Set the number of features that are
evaluated in evaluation steps. Cannot be 0 or fewer.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("ice_data"=10, "permutation_vimp"=5)}.

This parameter can be set for the following data elements:
\code{ice_data}, \code{permutation_vimp}, and \code{shap}.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
    \item{\code{estimation_type}}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}
    \item{\code{aggregate_results}}{(\emph{optional}) Flag that signifies whether results
should be aggregated during evaluation. If \code{estimation_type} is
\code{bias_correction} or \code{bc}, aggregation leads to a single bias-corrected
estimate. If \code{estimation_type} is \code{bootstrap_confidence_interval} or \code{bci},
aggregation leads to a single bias-corrected estimate with lower and upper
boundaries of the confidence interval. This has no effect if
\code{estimation_type} is \code{point}.

The default value is equal to \code{TRUE} except when assessing metrics to assess
model performance, as the default violin plot requires underlying data.

As with \code{detail_level} and \code{estimation_type}, a non-default
\code{aggregate_results} parameter can be specified for separate evaluation steps
by providing a parameter value in a named list with data elements, e.g.
\code{list("auc_data"=TRUE, , "model_performance"=FALSE)}. This parameter exists
for the same elements as \code{estimation_type}.}
    \item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}
    \item{\code{bootstrap_ci_method}}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This function plots the data on permutation variable importance
stored in a familiarCollection object.
}
\details{
This function generates a horizontal barplot that lists features by
the estimated model improvement over that of a dataset where the respective
feature is randomly permuted.

The following splitting variables are available for \code{split_by}, \code{color_by}
and \code{facet_by}:
\itemize{
\item \code{vimp_method}: variable importance methods.
\item \code{learner}: learners.
\item \code{data_set}: data sets.
\item \code{metric}: the model performance metrics.
\item \code{evaluation_time}: the evaluation times (survival outcomes only).
\item \code{similarity_threshold}: the similarity threshold used to identify groups
of features to permute simultaneously.
}

By default, the data is split by \code{vimp_method}, \code{learner} and \code{metric},
faceted by \code{data_set} and \code{evaluation_time}, and coloured by
\code{similarity_threshold}.

Labelling methods such as \code{set_vimp_method_names} or \code{set_feature_names} can
be applied to the \code{familiarCollection} object to update labels, and order
the output in the figure.

Bootstrap confidence intervals (if present) can be shown using various
styles set by \code{conf_int_style}:
\itemize{
\item \code{point_line} (default): confidence intervals are shown as lines, on which
the point estimate is likewise shown.
\item \code{line} (default): confidence intervals are shown as lines, but the point
estimate is not shown.
\item \code{bar_line}: confidence intervals are shown as lines, with the point
estimate shown as a bar plot with the opacity of \code{conf_int_alpha}.
\item \code{none}: confidence intervals are not shown. The point estimate is shown as
a bar plot.
}

For metrics where lower values indicate better model performance, more
negative permutation variable importance values indicate features that are
more important. Because this may cause confusion, values obtained for these
metrics are mirrored around 0.0 for plotting (but not any tabular data
export).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationFeatureSimilarity.R
\name{extract_feature_similarity}
\alias{extract_feature_similarity}
\title{Internal function to extract the feature distance table.}
\usage{
extract_feature_similarity(
  object,
  data,
  cl = NULL,
  estimation_type = waiver(),
  aggregate_results = waiver(),
  confidence_level = waiver(),
  bootstrap_ci_method = waiver(),
  is_pre_processed = FALSE,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  feature_similarity_metric = waiver(),
  verbose = FALSE,
  message_indent = 0L,
  ...
)
}
\arguments{
\item{object}{A \code{familiarEnsemble}, which is an ensemble of one or more
\code{familiarModel} objects, or a \code{familiarDataElementPredictionTable} object
that contains prediction data.}

\item{data}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}

\item{cl}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}

\item{estimation_type}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}

\item{aggregate_results}{(\emph{optional}) Flag that signifies whether results
should be aggregated during evaluation. If \code{estimation_type} is
\code{bias_correction} or \code{bc}, aggregation leads to a single bias-corrected
estimate. If \code{estimation_type} is \code{bootstrap_confidence_interval} or \code{bci},
aggregation leads to a single bias-corrected estimate with lower and upper
boundaries of the confidence interval. This has no effect if
\code{estimation_type} is \code{point}.

The default value is equal to \code{TRUE} except when assessing metrics to assess
model performance, as the default violin plot requires underlying data.

As with \code{detail_level} and \code{estimation_type}, a non-default
\code{aggregate_results} parameter can be specified for separate evaluation steps
by providing a parameter value in a named list with data elements, e.g.
\code{list("auc_data"=TRUE, , "model_performance"=FALSE)}. This parameter exists
for the same elements as \code{estimation_type}.}

\item{confidence_level}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}

\item{bootstrap_ci_method}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}

\item{is_pre_processed}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}

\item{feature_cluster_method}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_cluster_cut_method}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_threshold}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_metric}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{...}{Unused arguments.}
}
\value{
A data.table containing pairwise distance between features. This data
is only the upper triangular of the complete matrix (i.e. the sparse
unitriangular representation). Diagonals will always be 0.0 and the lower
triangular is mirrored.
}
\description{
Computes and extracts the feature distance table for features
used in a \code{familiarEnsemble} object. This table can be used to cluster
features, and is exported directly by \code{export_feature_similarity}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotPermutationVariableImportance.R
\name{.plot_permutation_variable_importance}
\alias{.plot_permutation_variable_importance}
\title{Internal plotting function for permutation variable importance plots}
\usage{
.plot_permutation_variable_importance(
  x,
  color_by,
  facet_by,
  facet_wrap_cols,
  ggtheme,
  discrete_palette,
  x_label,
  y_label,
  legend_label,
  plot_title,
  plot_sub_title,
  caption,
  limit_n_features,
  conf_int_style,
  conf_int_alpha,
  x_range,
  x_breaks
)
}
\arguments{
\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring the plot elements
according to the groupings indicated by the \code{color_by} argument (if any).
\code{familiar} has a default palette. Other palettes are supported by
\code{paletteer}, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{limit_n_features}{(\emph{optional}) The number of features that should be
included in the plot. Only the most important features are shown. By
default, the number of features is not limited.}

\item{conf_int_style}{(\emph{optional}) Confidence interval style. See details for
allowed styles.}

\item{conf_int_alpha}{(\emph{optional}) Alpha value to determine transparency of
confidence intervals or, alternatively, other plot elements with which the
confidence interval overlaps. Only values between 0.0 (fully transparent)
and 1.0 (fully opaque) are allowed.}

\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}
}
\value{
ggplot plot object.
}
\description{
Internal plotting function for permutation variable importance plots
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{outcomeInfo-class}
\alias{outcomeInfo-class}
\title{Outcome information object.}
\description{
An outcome information object stores data concerning an outcome. This is used
to prospectively check data.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the outcome, inherited from the original column name by
default.}

\item{\code{outcome_type}}{Type of outcome.}

\item{\code{outcome_column}}{Name of the outcome column in data.}

\item{\code{levels}}{Specifies class levels of categorical outcomes.}

\item{\code{ordered}}{Specifies whether categorical outcomes are ordered.}

\item{\code{reference}}{Class level used as reference.}

\item{\code{time}}{Maximum time, as set by the \code{time_max} configuration parameter.}

\item{\code{censored}}{Censoring indicators for survival outcomes.}

\item{\code{event}}{Event indicators for survival outcomes.}

\item{\code{competing_risk}}{Indicators for competing risks in survival outcomes.}

\item{\code{distribution}}{Five-number summary (numeric outcomes), class frequency
(categorical outcomes), or survival distributions.}

\item{\code{data_id}}{Internal identifier for the dataset used to derive the outcome
information.}

\item{\code{run_id}}{Internal identifier for the specific subset of the dataset used
to derive the outcome information.}

\item{\code{transformation_parameters}}{Parameters used for transforming a numeric
outcomes. Currently unused.}

\item{\code{normalisation_parameters}}{Parameters used for normalising numeric
outcomes. Currently unused.}

\item{\code{familiar_version}}{Version of the familiar package used to create this
object.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotAUCcurves.R
\name{plot_auc_precision_recall_curve}
\alias{plot_auc_precision_recall_curve}
\alias{plot_auc_precision_recall_curve,ANY-method}
\alias{plot_auc_precision_recall_curve,familiarCollection-method}
\title{Plot the precision-recall curve.}
\usage{
plot_auc_precision_recall_curve(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_auc_precision_recall_curve}{ANY}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_auc_precision_recall_curve}{familiarCollection}(
  object,
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  color_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  discrete_palette = NULL,
  x_label = waiver(),
  y_label = waiver(),
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  x_n_breaks = 5L,
  x_breaks = NULL,
  y_n_breaks = 5L,
  y_breaks = NULL,
  conf_int_style = c("ribbon", "step", "none"),
  conf_int_alpha = 0.2,
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{\code{familiarCollection} object, or one or more \code{familiarData}
objects, that will be internally converted to a \code{familiarCollection}
object. It is also possible to provide a \code{familiarEnsemble} or one or more
\code{familiarModel} objects together with the data from which data is computed
prior to export. Paths to such files can also be provided.

Additionally, some \code{familiarData} objects can be created from prediction
tables (\code{familiarDataElementPredictionTable}). Other \code{familiarData} objects
can be created from data (\code{dataObject}, or \code{data.table}). Please
check \emph{details} for more information.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where the plots of
receiver operating characteristic curves are saved to. Output is saved in
the \code{performance} subdirectory. If \code{NULL} no figures are saved, but are
returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{discrete_palette}{(\emph{optional}) Palette for colouring the plot elements
according to the groupings indicated by the \code{color_by} argument (if any).
\code{familiar} has a default palette. Other palettes are supported by
\code{paletteer}, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{x_n_breaks}{(\emph{optional}) Number of breaks to show on the x-axis of the
plot. \code{x_n_breaks} is used to determine the \code{x_breaks} argument in case it
is unset.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{y_n_breaks}{(\emph{optional}) Number of breaks to show on the y-axis of the
plot. \code{y_n_breaks} is used to determine the \code{y_breaks} argument in case it
is unset.}

\item{y_breaks}{(\emph{optional}) Break points on the y-axis of the plot.}

\item{conf_int_style}{(\emph{optional}) Confidence interval style. See details for
allowed styles.}

\item{conf_int_alpha}{(\emph{optional}) Alpha value to determine transparency of
confidence intervals or, alternatively, other plot elements with which the
confidence interval overlaps. Only values between 0.0 (fully transparent)
and 1.0 (fully opaque) are allowed.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from the number of features and the number of facets.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This method creates precision-recall curves based on data in a
familiarCollection object.
}
\details{
This function generates area under the precision-recall curve plots.

Available splitting variables are: \code{vimp_method}, \code{learner}, \code{data_set} and
\code{positive_class}. By default, the data is split by \code{vimp_method} and \code{learner},
with faceting by \code{data_set} and colouring by \code{positive_class}.

Available palettes for \code{discrete_palette} are those listed by
\code{grDevices::palette.pals()} (requires R >= 4.0.0), \code{grDevices::hcl.pals()}
(requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors}, \code{terrain.colors},
\code{topo.colors} and \code{cm.colors}, which correspond to the palettes of the same
name in \code{grDevices}. If not specified, a default palette based on palettes
in Tableau are used. You may also specify your own palette by using colour
names listed by \code{grDevices::colors()} or through hexadecimal RGB strings.

Bootstrap confidence intervals of the ROC curve (if present) can be shown
using various styles set by \code{conf_int_style}:
\itemize{
\item \code{ribbon} (default): confidence intervals are shown as a ribbon with an
opacity of \code{conf_int_alpha} around the point estimate of the ROC curve.
\item \code{step} (default): confidence intervals are shown as a step function around
the point estimate of the ROC curve.
\item \code{none}: confidence intervals are not shown. The point estimate of the ROC
curve is shown as usual.
}

Labelling methods such as \code{set_vimp_method_names} or \code{set_data_set_names} can
be applied to the \code{familiarCollection} object to update labels, and order
the output in the figure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarObjectConversion.R
\name{as_familiar_collection}
\alias{as_familiar_collection}
\alias{as_familiar_collection,familiarCollection-method}
\alias{as_familiar_collection,familiarData-method}
\alias{as_familiar_collection,familiarEnsemble-method}
\alias{as_familiar_collection,familiarModel-method}
\alias{as_familiar_collection,familiarDataElementPredictionTable-method}
\alias{as_familiar_collection,dataObject-method}
\alias{as_familiar_collection,data.table-method}
\alias{as_familiar_collection,list-method}
\alias{as_familiar_collection,character-method}
\alias{as_familiar_collection,ANY-method}
\title{Conversion to familiarCollection object.}
\usage{
as_familiar_collection(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{familiarCollection}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{familiarData}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{familiarEnsemble}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{familiarModel}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{familiarDataElementPredictionTable}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{dataObject}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{data.table}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{list}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{character}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)

\S4method{as_familiar_collection}{ANY}(
  object,
  familiar_data_names = NULL,
  collection_name = NULL,
  ...
)
}
\arguments{
\item{object}{\code{familiarCollection} object, or one or more \code{familiarData}
objects, that will be internally converted to a \code{familiarCollection}
object. It is also possible to provide a \code{familiarEnsemble} or one or more
\code{familiarModel} objects together with the data from which data is computed
prior to export. Paths to such files can also be provided.

Additionally, some \code{familiarData} objects can be created from prediction
tables (\code{familiarDataElementPredictionTable}). Other \code{familiarData} objects
can be created from data (\code{dataObject}, or \code{data.table}). Please
check \emph{details} for more information.}

\item{familiar_data_names}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}

\item{collection_name}{Name of the collection.}

\item{...}{
  Arguments passed on to \code{\link{.extract_data}}
  \describe{
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{time_max}}{Time point which is used as the benchmark for e.g. cumulative
risks generated by random forest, or the cut-off value for Uno's concordance
index. If not provided explicitly, this parameter is read from settings used
at creation of the underlying \code{familiarModel} objects. Only used for
\code{survival} outcomes.}
    \item{\code{evaluation_times}}{One or more time points that are used for in analysis
of survival problems when data has to be assessed at a set time, e.g.
calibration. If not provided explicitly, this parameter is read from
settings used at creation of the underlying \code{familiarModel} objects. Only
used for \code{survival} outcomes.}
    \item{\code{aggregation_method}}{Method for aggregating variable importances for the
purpose of evaluation. Variable importances are determined during feature
selection steps and after training the model. Both types are evaluated, but
feature selection variable importance is only evaluated at run-time.

See the documentation for the \code{vimp_aggregation_method} argument in
\code{summon_familiar} for information concerning the different available
methods.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{rank_threshold}}{The threshold used to  define the subset of highly
important features during evaluation.

See the documentation for the \code{vimp_aggregation_rank_threshold} argument in
\code{summon_familiar} for more information.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{ensemble_method}}{Method for ensembling predictions from models for the
same sample. Available methods are:
\itemize{
\item \code{median} (default): Use the median of the predicted values as the ensemble
value for a sample.
\item \code{mean}: Use the mean of the predicted values as the ensemble value for a
sample.
}}
    \item{\code{metric}}{One or more metrics for assessing model performance. See the
vignette on performance metrics for the available metrics. If not provided
explicitly, this parameter is read from settings used at creation of the
underlying \code{familiarModel} objects.}
    \item{\code{features}}{Features that should be considered for extracting information
from. Typically called in external workflows, e.g. for plotting. Internally,
i.e. from summon_familiar, this variable is not used.}
    \item{\code{feature_cluster_method}}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_linkage_method}}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_cluster_cut_method}}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_threshold}}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{feature_similarity_metric}}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{sample_cluster_method}}{The method used to perform clustering based on
distance between samples. These are the same methods as for the
\code{cluster_method} configuration parameter: \code{hclust}, \code{agnes}, \code{diana} and
\code{pam}.

\code{none} cannot be used when extracting data for feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{sample_linkage_method}}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{sample_similarity_metric}}{Metric to determine pairwise similarity
between samples. Similarity is computed in the same manner as for
clustering, but \code{sample_similarity_metric} has different options that are
better suited to computing distance between samples instead of between
features: \code{gower}, \code{euclidean}.

The underlying feature data is scaled to the \eqn{[0, 1]} range (for
numerical features) using the feature values across the samples. The
normalisation parameters required can optionally be computed from feature
data with the outer 5\% (on both sides) of feature values trimmed or
winsorised. To do so append \verb{_trim} (trimming) or \verb{_winsor} (winsorising) to
the metric name. This reduces the effect of outliers somewhat.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{icc_type}}{String indicating the type of intraclass correlation
coefficient (\code{1}, \code{2} or \code{3}) that should be used to compute robustness for
features in repeated measurements during the evaluation of univariate
importance. These types correspond to the types in Shrout and Fleiss (1979).
If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{data_element}}{String indicating which data elements are to be extracted.
Default is \code{all}, but specific elements can be specified to speed up
computations if not all elements are to be computed. This is an internal
parameter that is set by, e.g. the \code{export_model_vimp} method.}
    \item{\code{sample_limit}}{(\emph{optional}) Set the upper limit of the number of samples
that are used during evaluation steps. Cannot be fewer than 20.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("sample_similarity"=100, "permutation_vimp"=1000)}.

This parameter can be set for the following data elements:
\code{sample_similarity}, \code{shap}, \code{permutation_vimp}, and \code{ice_data}.}
    \item{\code{n_important_features}}{(\emph{optional}) Set the number of features that are
evaluated in evaluation steps. Cannot be 0 or fewer.

This setting can be specified per data element by providing a parameter
value in a named list with data elements, e.g.
\code{list("ice_data"=10, "permutation_vimp"=5)}.

This parameter can be set for the following data elements:
\code{ice_data}, \code{permutation_vimp}, and \code{shap}.}
    \item{\code{detail_level}}{(\emph{optional}) Sets the level at which results are computed
and aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If there
are at least 20 trained models in the ensemble, performance is computed for
each model, in contrast to \code{ensemble} where performance is computed for the
ensemble of models. If there are less than 20 trained models in the
ensemble, bootstraps are created so that at least 20 point estimates can be
made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the model
for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

\code{hybrid} is generally computationally less expensive then \code{ensemble}, which
in turn is somewhat less expensive than \code{model}.

A non-default \code{detail_level} parameter can be specified for separate
evaluation steps by providing a parameter value in a named list with data
elements, e.g. \code{list("auc_data"="ensemble", "model_performance"="hybrid")}.
This parameter can be set for the following data elements: \code{auc_data},
\code{decision_curve_analyis}, \code{model_performance}, \code{permutation_vimp},
\code{ice_data}, \code{prediction_data} and \code{confusion_matrix}.

If results are computed from 10 samples or fewer, \code{ensemble} is
automatically used. This prevents issues where evaluation steps do not have
a required minimum number of samples for \code{hybrid} or \code{model}.}
    \item{\code{estimation_type}}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}
    \item{\code{aggregate_results}}{(\emph{optional}) Flag that signifies whether results
should be aggregated during evaluation. If \code{estimation_type} is
\code{bias_correction} or \code{bc}, aggregation leads to a single bias-corrected
estimate. If \code{estimation_type} is \code{bootstrap_confidence_interval} or \code{bci},
aggregation leads to a single bias-corrected estimate with lower and upper
boundaries of the confidence interval. This has no effect if
\code{estimation_type} is \code{point}.

The default value is equal to \code{TRUE} except when assessing metrics to assess
model performance, as the default violin plot requires underlying data.

As with \code{detail_level} and \code{estimation_type}, a non-default
\code{aggregate_results} parameter can be specified for separate evaluation steps
by providing a parameter value in a named list with data elements, e.g.
\code{list("auc_data"=TRUE, , "model_performance"=FALSE)}. This parameter exists
for the same elements as \code{estimation_type}.}
    \item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}
    \item{\code{bootstrap_ci_method}}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}
    \item{\code{stratification_method}}{(\emph{optional}) Method for determining the
stratification threshold for creating survival groups. The actual,
model-dependent, threshold value is obtained from the development data, and
can afterwards be used to perform stratification on validation data.

The following stratification methods are available:
\itemize{
\item \code{median} (default): The median predicted value in the development cohort
is used to stratify the samples into two risk groups. For predicted outcome
values that build a continuous spectrum, the two risk groups in the
development cohort will be roughly equal in size.
\item \code{mean}: The mean predicted value in the development cohort is used to
stratify the samples into two risk groups.
\item \code{mean_trim}: As \code{mean}, but based on the set of predicted values
where the 5\% lowest and 5\% highest values are discarded. This reduces the
effect of outliers.
\item \code{mean_winsor}: As \code{mean}, but based on the set of predicted values where
the 5\% lowest and 5\% highest values are winsorised. This reduces the effect
of outliers.
\item \code{fixed}: Samples are stratified based on the sample quantiles of the
predicted values. These quantiles are defined using the
\code{stratification_threshold} parameter.
\item \code{optimised}: Use maximally selected rank statistics to determine the
optimal threshold (Lausen and Schumacher, 1992; Hothorn et al., 2003) to
stratify samples into two optimally separated risk groups.
}

One or more stratification methods can be selected simultaneously.

This parameter is only relevant for \code{survival} outcomes.}
  }}
}
\value{
A \code{familiarCollection} object.
}
\description{
Creates a \code{familiarCollection} objects from \code{familiarData},
\code{familiarEnsemble} or \code{familiarModel} objects.
}
\details{
A \code{data} argument is expected if the \code{object} argument is a
\code{familiarEnsemble} object or one or more \code{familiarModel} objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LearnerMain.R
\name{.get_default_sign_size}
\alias{.get_default_sign_size}
\title{Internal function for obtaining a default signature size parameter}
\usage{
.get_default_sign_size(data, restrict_samples = FALSE)
}
\arguments{
\item{data}{dataObject class object which contains the data on which the
preset parameters are determined.}

\item{restrict_samples}{Logical indicating whether the signature size should
be limited by the number of samples in addition to the number of available
features. This may help convergence of OLS-based methods.}
}
\value{
List containing the preset values for the signature size parameter.
}
\description{
Internal function for obtaining a default signature size parameter
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utilities.R
\name{encapsulate_path}
\alias{encapsulate_path}
\title{Encapsulate path}
\usage{
encapsulate_path(path)
}
\value{
encapsulated_path object
}
\description{
This function is used to encapsulate paths to allow for behaviour switches.
One use is for example when plotting. The plot_all method will encapsulate a
path so that plots may be saved to a directory structure. Other plot methods,
e.g. plot_model_performance do not encapsulate a path, and if the user calls
these functions directly, the plot may be written to the provided path
instead of a directory structure.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationUnivariateAnalysis.R
\name{extract_univariate_analysis}
\alias{extract_univariate_analysis}
\title{Internal function to extract data from a univariate analysis.}
\usage{
extract_univariate_analysis(
  object,
  data,
  cl = NULL,
  icc_type = waiver(),
  feature_similarity = NULL,
  feature_cluster_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_linkage_method = waiver(),
  feature_similarity_threshold = waiver(),
  feature_similarity_metric = waiver(),
  message_indent = 0L,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarEnsemble}, which is an ensemble of one or more
\code{familiarModel} objects, or a \code{familiarDataElementPredictionTable} object
that contains prediction data.}

\item{data}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}

\item{cl}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}

\item{icc_type}{String indicating the type of intraclass correlation
coefficient (\code{1}, \code{2} or \code{3}) that should be used to compute robustness for
features in repeated measurements during the evaluation of univariate
importance. These types correspond to the types in Shrout and Fleiss (1979).
If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_cluster_method}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_cluster_cut_method}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_threshold}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_metric}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}

\item{...}{Unused arguments.}
}
\value{
A list with a data.table containing information concerning the
univariate analysis of important features.
}
\description{
Computes and extracts univariate analysis for the features used
in a \code{familiarEnsemble} object. This assessment includes the computation of
p-values, as well as robustness (in case of repeated measurements).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{vimpTable-class}
\alias{vimpTable-class}
\title{Variable importance table}
\description{
A vimpTable object contains information concerning variable importance of one
or more features. These objects are created during variable importance
computation..
}
\details{
vimpTable objects exists in various states. These states are
generally incremental, i.e. one cannot turn a declustered table into the
initial version. Some methods such as aggregation internally do some state
reshuffling.

This object replaces the ad-hoc lists with information that were used in
versions prior to familiar 1.2.0.
}
\section{Slots}{

\describe{
\item{\code{vimp_table}}{Table containing features with corresponding scores.}

\item{\code{vimp_method}}{Method used to compute variable importance scores for each
feature.}

\item{\code{data_id}}{Internal identifier for the dataset used to derive the variable
importance table.}

\item{\code{run_id}}{Internal identifier for the specific subset of the dataset used
to derive the variable importance table.}

\item{\code{run_table}}{Run table for the data used to compute variable importances
from. Used internally.}

\item{\code{score_aggregation}}{Method used to aggregate the score of contrasts for
each categorical feature, if any,}

\item{\code{encoding_table}}{Table used to relate categorical features to their
contrasts, if any. Not used for all variable importance methods.}

\item{\code{cluster_table}}{Table used to relate original features with features
after clustering. Variable importance is determined after feature
processing, which includes clustering.}

\item{\code{invert}}{Determines whether increasing score corresponds to increasing
(\code{FALSE}) or decreasing rank (\code{TRUE}). Used internally to determine how
ranks should be formed.}

\item{\code{project_id}}{Identifier of the project that generated the vimpTable
object.}

\item{\code{familiar_version}}{Version of the familiar package used to create this
table.}

\item{\code{state}}{State of the variable importance table. The object can have the
following states:
\itemize{
\item \code{initial}: initial state, directly after the variable importance table is
filled.
\item \code{decoded}: depending on the variable importance method, the initial
variable importance table may contain the scores of individual contrasts
for categorical variables. When decoded, data in the \code{encoding_table}
attribute has been used to aggregate scores from all contrasts into a
single score for each feature.
\item \code{declustered}: variable importance is determined from fully processed
features, which includes clustering. This means that a single feature in
the variable importance table may represent multiple original features.
When a variable importance table has been declustered, all clusters have
been turned into their constituent features.
\item \code{reclustered}: When the table is reclustered, features are replaced by
their respective clusters. This is actually used when updating the cluster
table to ensure it fits to a local context. This prevents issues when
attempting to aggregate or apply variable importance tables in data with
different feature preprocessing, and as a result, different clusters.
\item \code{ranked}: The scores have been used to create ranks, with lower ranks
indicating better features.
\item \code{aggregated}: Score and ranks from multiple variable importance tables
were aggregated.
}}
}}

\seealso{
\code{\link{get_vimp_table}}, \code{\link{aggregate_vimp_table}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationFeatureExpression.R
\name{extract_feature_expression}
\alias{extract_feature_expression}
\title{Internal function to extract feature expressions.}
\usage{
extract_feature_expression(
  object,
  data,
  feature_similarity,
  sample_similarity,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_similarity_metric = waiver(),
  sample_cluster_method = waiver(),
  sample_linkage_method = waiver(),
  sample_similarity_metric = waiver(),
  evaluation_times = waiver(),
  message_indent = 0L,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarEnsemble}, which is an ensemble of one or more
\code{familiarModel} objects, or a \code{familiarDataElementPredictionTable} object
that contains prediction data.}

\item{data}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}

\item{feature_similarity}{Table containing pairwise distance between
sample. This is used to determine cluster information, and indicate which
samples are similar. The table is created by the
\code{extract_sample_similarity} method.}

\item{feature_cluster_method}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_metric}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_cluster_method}{The method used to perform clustering based on
distance between samples. These are the same methods as for the
\code{cluster_method} configuration parameter: \code{hclust}, \code{agnes}, \code{diana} and
\code{pam}.

\code{none} cannot be used when extracting data for feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{sample_similarity_metric}{Metric to determine pairwise similarity
between samples. Similarity is computed in the same manner as for
clustering, but \code{sample_similarity_metric} has different options that are
better suited to computing distance between samples instead of between
features: \code{gower}, \code{euclidean}.

The underlying feature data is scaled to the \eqn{[0, 1]} range (for
numerical features) using the feature values across the samples. The
normalisation parameters required can optionally be computed from feature
data with the outer 5\% (on both sides) of feature values trimmed or
winsorised. To do so append \verb{_trim} (trimming) or \verb{_winsor} (winsorising) to
the metric name. This reduces the effect of outliers somewhat.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{evaluation_times}{One or more time points that are used for in analysis
of survival problems when data has to be assessed at a set time, e.g.
calibration. If not provided explicitly, this parameter is read from
settings used at creation of the underlying \code{familiarModel} objects. Only
used for \code{survival} outcomes.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}

\item{...}{Unused arguments.}
}
\value{
A list with a data.table containing feature expressions.
}
\description{
Computes and extracts feature expressions for features
used in a \code{familiarEnsemble} object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarDataElement-class}
\alias{familiarDataElement-class}
\title{Data container for evaluation data.}
\description{
Most attributes of the familiarData object are objects of the
familiarDataElement class. This (super-)class is used to allow for
standardised aggregation and processing of evaluation data.
}
\section{Slots}{

\describe{
\item{\code{data}}{Evaluation data, typically a data.table or list.}

\item{\code{identifiers}}{Identifiers of the data, e.g. the generating model name,
learner, etc.}

\item{\code{detail_level}}{Sets the level at which results are computed and
aggregated.
\itemize{
\item \code{ensemble}: Results are computed at the ensemble level, i.e. over all
models in the ensemble. This means that, for example, bias-corrected
estimates of model performance are assessed by creating (at least) 20
bootstraps and computing the model performance of the ensemble model for
each bootstrap.
\item \code{hybrid} (default): Results are computed at the level of models in an
ensemble. This means that, for example, bias-corrected estimates of model
performance are directly computed using the models in the ensemble. If
there are at least 20 trained models in the ensemble, performance is
computed for each model, in contrast to \code{ensemble} where performance is
computed for the ensemble of models. If there are less than 20 trained
models in the ensemble, bootstraps are created so that at least 20 point
estimates can be made.
\item \code{model}: Results are computed at the model level. This means that, for
example, bias-corrected estimates of model performance are assessed by
creating (at least) 20 bootstraps and computing the performance of the
model for each bootstrap.
}

Note that each level of detail has a different interpretation for bootstrap
confidence intervals. For \code{ensemble} and \code{model} these are the confidence
intervals for the ensemble and an individual model, respectively. That is,
the confidence interval describes the range where an estimate produced by a
respective ensemble or model trained on a repeat of the experiment may be
found with the probability of the confidence level. For \code{hybrid}, it
represents the range where any single model trained on a repeat of the
experiment may be found with the probability of the confidence level. By
definition, confidence intervals obtained using \code{hybrid} are at least as
wide as those for \code{ensemble}. \code{hybrid} offers the correct interpretation if
the goal of the analysis is to assess the result of a single, unspecified,
model.

Some child classes do not use this parameter.}

\item{\code{estimation_type}}{Sets the type of estimation that should be possible.
This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

Some child classes do not use this parameter.}

\item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses
the rule of thumb \eqn{n = 20 / ci.level} to determine the number of
required bootstraps.}

\item{\code{bootstrap_ci_method}}{Method used to determine bootstrap confidence
intervals (Efron and Hastie, 2016). The following methods are implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}

\item{\code{value_column}}{Identifies column(s) in the \code{data} attribute presenting
values.}

\item{\code{grouping_column}}{Identifies column(s) in the \code{data} attribute presenting
identifier columns for grouping during aggregation. Familiar will
automatically assign items from the \code{identifiers} attribute to the data and
this attribute when combining multiple familiarDataElements of the same
(child) class.}

\item{\code{is_aggregated}}{Defines whether the object was aggregated.}
}}

\references{
\enumerate{
\item Efron, B. & Hastie, T. Computer Age Statistical Inference.
(Cambridge University Press, 2016).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationCalibrationInfo.R
\name{export_calibration_info}
\alias{export_calibration_info}
\alias{export_calibration_info,familiarCollection-method}
\alias{export_calibration_info,ANY-method}
\title{Extract and export calibration information.}
\usage{
export_calibration_info(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)

\S4method{export_calibration_info}{familiarCollection}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)

\S4method{export_calibration_info}{ANY}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{aggregate_results}{Flag that signifies whether results should be
aggregated for export.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
  }}
}
\value{
A data.table (if \code{dir_path} is not provided), or nothing, as all data
is exported to \code{csv} files.
}
\description{
Extract and export calibration information (e.g. baseline
survival) for data in a familiarCollection.
}
\details{
Data is usually collected from a \code{familiarCollection} object.
However, you can also provide one or more \code{familiarData} objects, that will
be internally converted to a \code{familiarCollection} object. It is also
possible to provide a \code{familiarEnsemble} or one or more \code{familiarModel}
objects together with the data from which data is computed prior to export.
Paths to the previous files can also be provided.

All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Currently only baseline survival is exported as supporting calibration
information. See \code{export_calibration_data} for export of direct assessment
of calibration, including calibration and goodness-of-fit tests.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotUtilities.R
\name{theme_familiar}
\alias{theme_familiar}
\title{Familiar ggplot2 theme}
\usage{
theme_familiar(
  base_size = 10,
  base_family = "",
  base_line_size = 0.5,
  base_rect_size = 0.5
)
}
\arguments{
\item{base_size}{Base font size in points. Size of other plot text elements
is based off this.}

\item{base_family}{Font family used for text elements.}

\item{base_line_size}{Base size for line elements, in points.}

\item{base_rect_size}{Base size for rectangular elements, in points.}
}
\value{
A complete plotting theme.
}
\description{
This is the default theme used for plots created by familiar. The theme uses
\code{ggplot2::theme_light} as the base template.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarObjectConversion.R
\name{as_familiar_ensemble}
\alias{as_familiar_ensemble}
\alias{as_familiar_ensemble,familiarEnsemble-method}
\alias{as_familiar_ensemble,familiarModel-method}
\alias{as_familiar_ensemble,familiarNoveltyDetector-method}
\alias{as_familiar_ensemble,list-method}
\alias{as_familiar_ensemble,character-method}
\alias{as_familiar_ensemble,ANY-method}
\title{Conversion to familiarEnsemble object.}
\usage{
as_familiar_ensemble(object, ...)

\S4method{as_familiar_ensemble}{familiarEnsemble}(object, ...)

\S4method{as_familiar_ensemble}{familiarModel}(object, ...)

\S4method{as_familiar_ensemble}{familiarNoveltyDetector}(object, ...)

\S4method{as_familiar_ensemble}{list}(object, ...)

\S4method{as_familiar_ensemble}{character}(object, ...)

\S4method{as_familiar_ensemble}{ANY}(object, ...)
}
\arguments{
\item{object}{A \code{familiarEnsemble} object, or one or more \code{familiarModel}
objects that will be internally converted to a \code{familiarEnsemble} object.
Paths to such objects can also be provided.}

\item{...}{Unused arguments.}
}
\value{
A \code{familiarEnsemble} object.
}
\description{
Creates \code{familiarEnsemble} a object from \code{familiarModel}
objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarVimpMethod-class}
\alias{familiarVimpMethod-class}
\title{Variable importance method object.}
\description{
The familiarVimpMethod class is the parent class for all variable importance
methods in familiar.
}
\section{Slots}{

\describe{
\item{\code{outcome_type}}{Outcome type of the data to be evaluated using the object.}

\item{\code{hyperparameters}}{Set of hyperparameters for the variable importance
method.}

\item{\code{vimp_method}}{The character string indicating the variable importance
method.}

\item{\code{vimp_aggregation_method}}{Method used for aggregating variable importance
tables if more than one is present.)}

\item{\code{vimp_rank_threshold}}{Threshold used for some variable importance
aggregation methods.}

\item{\code{multivariate}}{Flags whether the variable importance method is
multivariate vs. univariate.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters.}

\item{\code{required_features}}{The set of features to be assessed by the variable
importance method.}

\item{\code{package}}{Name of the package(s) required to execute the variable
importance method.}

\item{\code{run_table}}{Run table for the data to be assessed by the variable
importance method. Used internally.}

\item{\code{project_id}}{Identifier of the project that generated the
familiarVimpMethod object.}

\item{\code{familiar_version}}{Version of the familiar package used to create this
object.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Familiar.R
\name{get_xml_config}
\alias{get_xml_config}
\title{Create an empty xml configuration file}
\usage{
get_xml_config(dir_path)
}
\arguments{
\item{dir_path}{Path to the directory where the configuration file should be
created. The directory should exist, and no file named \code{config.xml} should
be present.}
}
\value{
Nothing. A file named \code{config.xml} is created in the directory
indicated by \code{dir_path}.
}
\description{
This function creates an empty configuration xml file in the directory
specified by \code{dir_path}. This provides an alternative to the use of input
arguments for familiar.
}
\examples{
\dontrun{
# Creates a config.xml file in the working directory
get_xml_config(dir_path=getwd())
}
}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{featureInfoParameters-class}
\alias{featureInfoParameters-class}
\title{Feature information parameters object.}
\description{
A featureInfo object contains information for a single feature. Some
information, for example concerning clustering and transformation contains
various parameters that allow for applying the data transformation correctly.
These are stored in featureInfoParameters objects.
}
\details{
featureInfoParameters is normally a parent class for specific
classes, such as featureInfoParametersTransformation.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the feature, which by default is the column name of the
feature. Typically used to correctly assign the data.}

\item{\code{complete}}{Flags whether the parameters have been completely set.}

\item{\code{familiar_version}}{Version of the familiar package.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarData-class}
\alias{familiarData-class}
\title{Dataset obtained after evaluating models on a dataset.}
\description{
A familiarData object is created by evaluating familiarEnsemble or
familiarModel objects on a dataset. Multiple familiarData objects are
aggregated in a familiarCollection object.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the dataset, e.g. training or internal validation.}

\item{\code{outcome_type}}{Outcome type of the data used to create the object.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{fs_vimp}}{Data collected for variable importance methods.}

\item{\code{model_vimp}}{Variable importance data collected from model-specific
algorithms implemented by models created by familiar.}

\item{\code{permutation_vimp}}{Data collected for permutation variable importance.}

\item{\code{hyperparameters}}{Hyperparameters collected from created models.}

\item{\code{hyperparameter_data}}{Additional data concerning hyperparameters. This is
currently not used yet.}

\item{\code{required_features}}{The set of features required for complete
reproduction, i.e. with imputation.}

\item{\code{model_features}}{The set of features that are required for using the
model or ensemble of models, but without imputation.}

\item{\code{learner}}{Learning algorithm used to create the model or ensemble of
models.}

\item{\code{vimp_method}}{Method used to determine variable importance for the model
or ensemble of models.}

\item{\code{pooling_table}}{Run table for the data underlying the familiarData
object. Used internally.}

\item{\code{prediction_data}}{Model predictions for a model or ensemble of models for
the underlying dataset.}

\item{\code{confusion_matrix}}{Confusion matrix for a model or ensemble of models,
based on the underlying dataset.}

\item{\code{decision_curve_data}}{Decision curve analysis data for a model or
ensemble of models, based on the underlying dataset.}

\item{\code{calibration_info}}{Calibration information, e.g. baseline survival in the
development cohort.}

\item{\code{calibration_data}}{Calibration data for a model or ensemble of models,
based on the underlying dataset.}

\item{\code{model_performance}}{Model performance data for a model or ensemble of
models, based on the underlying dataset.}

\item{\code{km_info}}{Information concerning risk-stratification cut-off values..}

\item{\code{km_data}}{Kaplan-Meier survival data for a model or ensemble of models,
based on the underlying dataset.}

\item{\code{auc_data}}{AUC-ROC and AUC-PR data for a model or ensemble of models,
based on the underlying dataset.}

\item{\code{ice_data}}{Individual conditional expectation data for features included
in a model or ensemble of models, based on the underlying dataset. Partial
dependence data are computed on the fly from these data.}

\item{\code{shap_data}}{SHAP values for features included in a model or ensemble of
models.}

\item{\code{univariate_analysis}}{Univariate analysis of the underlying dataset.}

\item{\code{feature_expressions}}{Feature expression values of the underlying
dataset.}

\item{\code{feature_similarity}}{Feature similarity information of the underlying
dataset.}

\item{\code{sample_similarity}}{Sample similarity information of the underlying
dataset.}

\item{\code{project_id}}{Identifier of the project that generated the familiarData
object.}

\item{\code{familiar_version}}{Version of the familiar package.

familiarData objects contain information obtained by evaluating a single
model or single ensemble of models on a dataset.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataComputationFeatureSimilarity.R
\name{export_feature_similarity}
\alias{export_feature_similarity}
\alias{export_feature_similarity,familiarCollection-method}
\alias{export_feature_similarity,ANY-method}
\title{Extract and export mutual correlation between features.}
\usage{
export_feature_similarity(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  export_dendrogram = FALSE,
  export_ordered_data = FALSE,
  export_clustering = FALSE,
  export_collection = FALSE,
  ...
)

\S4method{export_feature_similarity}{familiarCollection}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  export_dendrogram = FALSE,
  export_ordered_data = FALSE,
  export_clustering = FALSE,
  export_collection = FALSE,
  ...
)

\S4method{export_feature_similarity}{ANY}(
  object,
  dir_path = NULL,
  aggregate_results = TRUE,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  export_dendrogram = FALSE,
  export_ordered_data = FALSE,
  export_clustering = FALSE,
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{dir_path}{Path to folder where extracted data should be saved. \code{NULL}
will allow export as a structured list of data.tables.}

\item{aggregate_results}{Flag that signifies whether results should be
aggregated for export.}

\item{feature_cluster_method}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_cluster_cut_method}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_threshold}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{export_dendrogram}{Add dendrogram in the data element objects.}

\item{export_ordered_data}{Add feature label ordering to data in the data
element objects.}

\item{export_clustering}{Add clustering information to data.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[=as_data_object]{as_data_object}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{data}}{A \code{data.frame} or \code{data.table}, a path to such tables on a local
or network drive, or a path to tabular data that may be converted to these
formats.}
  }}
}
\value{
A list containing a data.table (if \code{dir_path} is not provided), or
nothing, as all data is exported to \code{csv} files.
}
\description{
Extract and export mutual correlation between features in a
familiarCollection.
}
\details{
All parameters aside from \code{object} and \code{dir_path} are only used if \code{object}
is not a \code{familiarCollection} object, or a path to one.

Feature similarity data can be created from \code{dataObject}, or \code{data.table} objects.
For \code{data.table}, see \code{\link{as_data_object}} for additional arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.update_experimental_design_settings}
\alias{.update_experimental_design_settings}
\title{Internal function to check batch assignment to development and validation}
\usage{
.update_experimental_design_settings(
  section_table,
  data,
  settings,
  verbose = FALSE
)
}
\arguments{
\item{section_table}{data.table generated by the \code{extract_experimental_setup}
function. Contains information regarding the experiment.}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{settings}{List of parameter settings for data set parsing and setting
up the experiment.}
}
\value{
A verified and updated list of parameter settings.
}
\description{
This function checks which batches in the data set are assigned to model
development and external validation. Several errors may be raised if there
are inconsistencies such as an overlapping assignment, name mismatches etc.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotUtilities.R
\name{.check_plot_splitting_variables}
\alias{.check_plot_splitting_variables}
\title{Checks and sanitizes splitting variables for plotting.}
\usage{
.check_plot_splitting_variables(
  x,
  split_by = NULL,
  color_by = NULL,
  linetype_by = NULL,
  facet_by = NULL,
  x_axis_by = NULL,
  y_axis_by = NULL,
  available = NULL
)
}
\arguments{
\item{x}{data.table or data.frame containing the data used for splitting.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{color_by}{(\emph{optional}) Variables used to determine fill colour of plot
objects. The variables cannot overlap with those provided to the \code{split_by}
argument, but may overlap with other arguments. See details for available
variables.}

\item{linetype_by}{(\emph{optional}) Variables that are used to determine the
linetype of lines in a plot. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
Sett details for available variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{x_axis_by}{(\emph{optional}) Variable plotted along the x-axis of a plot.
The variable cannot overlap with variables provided to the \code{split_by} and
\code{y_axis_by} arguments (if used), but may overlap with other arguments. Only
one variable is allowed for this argument. See details for available
variables.}

\item{y_axis_by}{(\emph{optional}) Variable plotted along the y-axis of a plot.
The variable cannot overlap with variables provided to the \code{split_by} and
\code{x_axis_by} arguments (if used), but may overlap with other arguments. Only
one variable is allowed for this argument. See details for available
variables.}

\item{available}{Names of columns available for splitting.}
}
\value{
A sanitized list of splitting variables.
}
\description{
Checks and sanitizes splitting variables for plotting.
}
\details{
This internal function allows some flexibility regarding the exact
input. Allowed splitting variables should be defined by the available
argument.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarDataElement.R
\name{extract_dispatcher,familiarEnsemble,familiarDataElement-method}
\alias{extract_dispatcher,familiarEnsemble,familiarDataElement-method}
\title{Internal function to dispatch extraction functions.}
\usage{
\S4method{extract_dispatcher}{familiarEnsemble,familiarDataElement}(
  cl = NULL,
  FUN,
  object,
  proto_data_element,
  aggregate_results,
  has_internal_bootstrap,
  ...,
  message_indent = 0L,
  verbose = TRUE
)
}
\arguments{
\item{cl}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}

\item{FUN}{Extraction function or method to which data and parameters are
dispatched.}

\item{object}{A \code{familiarEnsemble} object.}

\item{proto_data_element}{A \code{familiarDataElement} object, or an object that
inherits from it.}

\item{aggregate_results}{A logical flag indicating whether results should be
aggregated.}

\item{has_internal_bootstrap}{A logical flag that indicates whether \code{FUN} has
internal bootstrapping capabilities.}

\item{...}{Unused arguments.}

\item{message_indent}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}

\item{verbose}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
}
\value{
A list of \code{familiarDataElement} objects.
}
\description{
This function provides a unified access point to extraction
functions. Some of these functions require bootstrapping and result
aggregation, which are handled here.
}
\details{
This function first determines how many data points need to be
evaluated to complete the desired estimation, i.e. 1 for point estimates, 20
for bias-corrected estimates, and 20 / (1 - confidence level) for bootstrap
confidence intervals.

Subsequently, we determine the number of models. This number is used to set
external or internal clusters, the number of bootstraps, and to evaluate
whether the estimation can be done in case \code{FUN} does not support
bootstrapping.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarModel.R
\name{coef}
\alias{coef}
\alias{coef,familiarModel-method}
\title{Extract model coefficients}
\usage{
coef(object, ...)

\S4method{coef}{familiarModel}(object, ...)
}
\arguments{
\item{object}{a familiarModel object}

\item{...}{additional arguments passed to \code{coef} methods for the underlying
model, when available.}
}
\value{
Coefficients extracted from the model in the familiarModel object, if
any.
}
\description{
Extract model coefficients
}
\details{
This method extends the \code{coef} S3 method. For some models \code{coef}
requires information that is trimmed from the model. In this case a copy of
the model coefficient is stored with the model, and returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotFeatureSimilarity.R
\name{plot_feature_similarity}
\alias{plot_feature_similarity}
\alias{plot_feature_similarity,ANY-method}
\alias{plot_feature_similarity,familiarCollection-method}
\title{Plot heatmaps for pairwise similarity between features.}
\usage{
plot_feature_similarity(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  gradient_palette = NULL,
  gradient_palette_range = NULL,
  x_label = waiver(),
  x_label_shared = "column",
  y_label = waiver(),
  y_label_shared = "row",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  y_range = NULL,
  y_n_breaks = 3L,
  y_breaks = NULL,
  rotate_x_tick_labels = waiver(),
  show_dendrogram = c("top", "right"),
  dendrogram_height = grid::unit(1.5, "cm"),
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_feature_similarity}{ANY}(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  gradient_palette = NULL,
  gradient_palette_range = NULL,
  x_label = waiver(),
  x_label_shared = "column",
  y_label = waiver(),
  y_label_shared = "row",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  y_range = NULL,
  y_n_breaks = 3L,
  y_breaks = NULL,
  rotate_x_tick_labels = waiver(),
  show_dendrogram = c("top", "right"),
  dendrogram_height = grid::unit(1.5, "cm"),
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)

\S4method{plot_feature_similarity}{familiarCollection}(
  object,
  feature_cluster_method = waiver(),
  feature_linkage_method = waiver(),
  feature_cluster_cut_method = waiver(),
  feature_similarity_threshold = waiver(),
  draw = FALSE,
  dir_path = NULL,
  split_by = NULL,
  facet_by = NULL,
  facet_wrap_cols = NULL,
  ggtheme = NULL,
  gradient_palette = NULL,
  gradient_palette_range = NULL,
  x_label = waiver(),
  x_label_shared = "column",
  y_label = waiver(),
  y_label_shared = "row",
  legend_label = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = NULL,
  y_range = NULL,
  y_n_breaks = 3L,
  y_breaks = NULL,
  rotate_x_tick_labels = waiver(),
  show_dendrogram = c("top", "right"),
  dendrogram_height = grid::unit(1.5, "cm"),
  width = waiver(),
  height = waiver(),
  units = waiver(),
  export_collection = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{familiarCollection} object, or other other objects from which
a \code{familiarCollection} can be extracted. See details for more information.}

\item{feature_cluster_method}{The method used to perform clustering. These are
the same methods as for the \code{cluster_method} configuration parameter:
\code{none}, \code{hclust}, \code{agnes}, \code{diana} and \code{pam}.

\code{none} cannot be used when extracting data regarding mutual correlation or
feature expressions.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_linkage_method}{The method used for agglomerative clustering in
\code{hclust} and \code{agnes}. These are the same methods as for the
\code{cluster_linkage_method} configuration parameter: \code{average}, \code{single},
\code{complete}, \code{weighted}, and \code{ward}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_cluster_cut_method}{The method used to divide features into
separate clusters. The available methods are the same as for the
\code{cluster_cut_method} configuration parameter: \code{silhouette}, \code{fixed_cut} and
\code{dynamic_cut}.

\code{silhouette} is available for all cluster methods, but \code{fixed_cut} only
applies to methods that create hierarchical trees (\code{hclust}, \code{agnes} and
\code{diana}). \code{dynamic_cut} requires the \code{dynamicTreeCut} package and can only
be used with \code{agnes} and \code{hclust}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{feature_similarity_threshold}{The threshold level for pair-wise
similarity that is required to form feature clusters with the \code{fixed_cut}
method.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}

\item{draw}{(\emph{optional}) Draws the plot if TRUE.}

\item{dir_path}{(\emph{optional}) Path to the directory where created performance
plots are saved to. Output is saved in the \code{feature_similarity}
subdirectory. If \code{NULL} no figures are saved, but are returned instead.}

\item{split_by}{(\emph{optional}) Splitting variables. This refers to column names
on which datasets are split. A separate figure is created for each split.
See details for available variables.}

\item{facet_by}{(\emph{optional}) Variables used to determine how and if facets of
each figure appear. In case the \code{facet_wrap_cols} argument is \code{NULL}, the
first variable is used to define columns, and the remaing variables are
used to define rows of facets. The variables cannot overlap with those
provided to the \code{split_by} argument, but may overlap with other arguments.
See details for available variables.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{ggtheme}{(\emph{optional}) \code{ggplot} theme to use for plotting.}

\item{gradient_palette}{(\emph{optional}) Sequential or divergent palette used to
colour the similarity or distance between features in a heatmap.
\code{familiar} has a default palette. Other palettes are supported by the
\code{paletteer} package, \code{grDevices::palette.pals()} (requires R >= 4.0.0),
\code{grDevices::hcl.pals()} (requires R >= 3.6.0) and \code{rainbow}, \code{heat.colors},
\code{terrain.colors}, \code{topo.colors} and \code{cm.colors}, which correspond to the
palettes of the same name in \code{grDevices}. You may also specify your own
palette by providing a vector of colour names listed by
\code{grDevices::colors()} or through hexadecimal RGB strings.}

\item{gradient_palette_range}{(\emph{optional}) Numerical range used to span the
gradient. This should be a range of two values, e.g. \code{c(0, 1)}. Lower or
upper boundary can be unset by using \code{NA}. If not set, the full
metric-specific range is used.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{x_label_shared}{(\emph{optional}) Sharing of x-axis labels between facets.
One of three values:
\itemize{
\item \code{overall}: A single label is placed at the bottom of the figure. Tick
text (but not the ticks themselves) is removed for all but the bottom facet
plot(s).
\item \code{column}: A label is placed at the bottom of each column. Tick text (but
not the ticks themselves) is removed for all but the bottom facet plot(s).
\item \code{individual}: A label is placed below each facet plot. Tick text is kept.
}}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{y_label_shared}{(\emph{optional}) Sharing of y-axis labels between facets.
One of three values:
\itemize{
\item \code{overall}: A single label is placed to the left of the figure. Tick text
(but not the ticks themselves) is removed for all but the left-most facet
plot(s).
\item \code{row}: A label is placed to the left of each row. Tick text (but not the
ticks themselves) is removed for all but the left-most facet plot(s).
\item \code{individual}: A label is placed below each facet plot. Tick text is kept.
}}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}

\item{y_range}{(\emph{optional}) Value range for the y-axis.}

\item{y_n_breaks}{(\emph{optional}) Number of breaks to show on the y-axis of the
plot. \code{y_n_breaks} is used to determine the \code{y_breaks} argument in case it
is unset.}

\item{y_breaks}{(\emph{optional}) Break points on the y-axis of the plot.}

\item{rotate_x_tick_labels}{(\emph{optional}) Rotate tick labels on the x-axis by
90 degrees. Defaults to \code{TRUE}. Rotation of x-axis tick labels may also be
controlled through the \code{ggtheme}. In this case, \code{FALSE} should be provided
explicitly.}

\item{show_dendrogram}{(\emph{optional}) Show dendrogram around the main panel.
Can be \code{TRUE}, \code{FALSE}, \code{NULL}, or a position, i.e. \code{top}, \code{bottom}, \code{left}
and \code{right}. Up to two positions may be provided, but only as long as the
dendrograms are not on opposite sides of the heatmap: \code{top} and \code{bottom},
and \code{left} and \code{right} cannot be used together.

A dendrogram can only be drawn from cluster methods that produce
dendrograms, such as \code{hclust}. A dendrogram can for example not be
constructed using the partitioning around medioids method (\code{pam}).

By default, a dendrogram is drawn to the top and right of the panel.}

\item{dendrogram_height}{(\emph{optional}) Height of the dendrogram. The height is
1.5 cm by default. Height is expected to be grid unit (see \code{grid::unit}),
which also allows for specifying relative heights.}

\item{width}{(\emph{optional}) Width of the plot. A default value is derived from
the number of facets.}

\item{height}{(\emph{optional}) Height of the plot. A default value is derived
from the number of features and the number of facets.}

\item{units}{(\emph{optional}) Plot size unit. Either \code{cm} (default), \code{mm} or
\verb{in}.}

\item{export_collection}{(\emph{optional}) Exports the collection if TRUE.}

\item{...}{
  Arguments passed on to \code{\link[=as_familiar_collection]{as_familiar_collection}}, \code{\link[ggplot2:ggsave]{ggplot2::ggsave}}, \code{\link[=extract_feature_similarity]{extract_feature_similarity}}
  \describe{
    \item{\code{familiar_data_names}}{Names of the dataset(s). Only used if the \code{object}
parameter is one or more \code{familiarData} objects.}
    \item{\code{collection_name}}{Name of the collection.}
    \item{\code{device}}{Device to use. Can either be a device function
(e.g. \link{png}), or one of "eps", "ps", "tex" (pictex),
"pdf", "jpeg", "tiff", "png", "bmp", "svg" or "wmf" (windows only). If
\code{NULL} (default), the device is guessed based on the \code{filename} extension.}
    \item{\code{scale}}{Multiplicative scaling factor.}
    \item{\code{dpi}}{Plot resolution. Also accepts a string input: "retina" (320),
"print" (300), or "screen" (72). Only applies when converting pixel units,
as is typical for raster output types.}
    \item{\code{limitsize}}{When \code{TRUE} (the default), \code{ggsave()} will not
save images larger than 50x50 inches, to prevent the common error of
specifying dimensions in pixels.}
    \item{\code{bg}}{Background colour. If \code{NULL}, uses the \code{plot.background} fill value
from the plot theme.}
    \item{\code{create.dir}}{Whether to create new directories if a non-existing
directory is specified in the \code{filename} or \code{path} (\code{TRUE}) or return an
error (\code{FALSE}, default). If \code{FALSE} and run in an interactive session,
a prompt will appear asking to create a new directory when necessary.}
    \item{\code{data}}{A \code{dataObject} object, \code{data.table} or \code{data.frame} that
constitutes the data that are assessed.}
    \item{\code{is_pre_processed}}{Flag that indicates whether the data was already
pre-processed externally, e.g. normalised and clustered. Only used if the
\code{data} argument is a \code{data.table} or \code{data.frame}.}
    \item{\code{cl}}{Cluster created using the \code{parallel} package. This cluster is then
used to speed up computation through parallellisation.}
    \item{\code{feature_similarity_metric}}{Metric to determine pairwise similarity
between features. Similarity is computed in the same manner as for
clustering, and \code{feature_similarity_metric} therefore has the same options
as \code{cluster_similarity_metric}: \code{mcfadden_r2}, \code{cox_snell_r2},
\code{nagelkerke_r2}, \code{spearman}, \code{kendall} and \code{pearson}.

If not provided explicitly, this parameter is read from settings used at
creation of the underlying \code{familiarModel} objects.}
    \item{\code{verbose}}{Flag to indicate whether feedback should be provided on the
computation and extraction of various data elements.}
    \item{\code{message_indent}}{Number of indentation steps for messages shown during
computation and extraction of various data elements.}
    \item{\code{estimation_type}}{(\emph{optional}) Sets the type of estimation that should be
possible. This has the following options:
\itemize{
\item \code{point}: Point estimates.
\item \code{bias_correction} or \code{bc}: Bias-corrected estimates. A bias-corrected
estimate is computed from (at least) 20 point estimates, and \code{familiar} may
bootstrap the data to create them.
\item \code{bootstrap_confidence_interval} or \code{bci} (default): Bias-corrected
estimates with bootstrap confidence intervals (Efron and Hastie, 2016). The
number of point estimates required depends on the \code{confidence_level}
parameter, and \code{familiar} may bootstrap the data to create them.
}

As with \code{detail_level}, a non-default \code{estimation_type} parameter can be
specified for separate evaluation steps by providing a parameter value in a
named list with data elements, e.g. \code{list("auc_data"="bci", "model_performance"="point")}. This parameter can be set for the following
data elements: \code{auc_data}, \code{decision_curve_analyis}, \code{model_performance},
\code{permutation_vimp}, \code{ice_data}, and \code{prediction_data}.}
    \item{\code{aggregate_results}}{(\emph{optional}) Flag that signifies whether results
should be aggregated during evaluation. If \code{estimation_type} is
\code{bias_correction} or \code{bc}, aggregation leads to a single bias-corrected
estimate. If \code{estimation_type} is \code{bootstrap_confidence_interval} or \code{bci},
aggregation leads to a single bias-corrected estimate with lower and upper
boundaries of the confidence interval. This has no effect if
\code{estimation_type} is \code{point}.

The default value is equal to \code{TRUE} except when assessing metrics to assess
model performance, as the default violin plot requires underlying data.

As with \code{detail_level} and \code{estimation_type}, a non-default
\code{aggregate_results} parameter can be specified for separate evaluation steps
by providing a parameter value in a named list with data elements, e.g.
\code{list("auc_data"=TRUE, , "model_performance"=FALSE)}. This parameter exists
for the same elements as \code{estimation_type}.}
    \item{\code{confidence_level}}{(\emph{optional}) Numeric value for the level at which
confidence intervals are determined. In the case bootstraps are used to
determine the confidence intervals bootstrap estimation, \code{familiar} uses the
rule of thumb \eqn{n = 20 / ci.level} to determine the number of required
bootstraps.

The default value is \code{0.95}.}
    \item{\code{bootstrap_ci_method}}{(\emph{optional}) Method used to determine bootstrap
confidence intervals (Efron and Hastie, 2016). The following methods are
implemented:
\itemize{
\item \code{percentile} (default): Confidence intervals obtained using the percentile
method.
\item \code{bc}: Bias-corrected confidence intervals.
}

Note that the standard method is not implemented because this method is
often not suitable due to non-normal distributions. The bias-corrected and
accelerated (BCa) method is not implemented yet.}
  }}
}
\value{
\code{NULL} or list of plot objects, if \code{dir_path} is \code{NULL}.
}
\description{
This method creates a heatmap based on data stored in a
\code{familiarCollection} object. Features in the heatmap are ordered so that
more similar features appear together.
}
\details{
This function generates area under the ROC curve plots.

Available splitting variables are: \code{vimp_method}, \code{learner}, and \code{data_set}.
By default, the data is split by \code{vimp_method}, \code{learner} and \code{data_set},
since the features may be ordered differently for each data set.

Note that similarity is determined based on the underlying data. Hence the
ordering of features may differ between facets, and tick labels are
maintained for each panel.

Labeling methods such as \code{set_vimp_method_names} or \code{set_data_set_names} can
be applied to the \code{familiarCollection} object to update labels, and order
the output in the figure.

This plot can be created from \code{dataObject}, or \code{data.table} objects.
For \code{data.table}, see \code{\link{as_data_object}} for additional arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataParameterChecks.R
\name{.update_initial_settings}
\alias{.update_initial_settings}
\title{Internal check and update of settings related to data set parsing}
\usage{
.update_initial_settings(
  formula = NULL,
  data,
  settings,
  check_stringency = "strict"
)
}
\arguments{
\item{formula}{User-provided formula, may be absent (\code{NULL}).}

\item{data}{Data set as loaded using the \code{.load_data} function.}

\item{settings}{List of parameter settings for data set parsing.}

\item{check_stringency}{Specifies stringency of various checks. This is mostly:
\itemize{
\item \code{strict}: default value used for \code{summon_familiar}. Thoroughly checks
input data. Used internally for checking development data.
\item \code{external_warn}: value used for \code{extract_data} and related methods. Less
stringent checks, but will warn for possible issues. Used internally for
checking data for evaluation and explanation.
\item \code{external}: value used for external methods such as \code{predict}. Less
stringent checks, particularly for identifier and outcome columns, which may
be completely absent. Used internally for \code{predict}.
}}
}
\value{
A verified and updated list of parameter settings.
}
\description{
This function updates and checks parameters related to data set parsing based
on the available data set.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PlotInputArguments.R
\name{.check_input_plot_args}
\alias{.check_input_plot_args}
\title{Internal checks on common plot input arguments}
\usage{
.check_input_plot_args(
  x_range = waiver(),
  y_range = waiver(),
  x_n_breaks = waiver(),
  y_n_breaks = waiver(),
  x_breaks = waiver(),
  y_breaks = waiver(),
  conf_int = waiver(),
  conf_int_alpha = waiver(),
  conf_int_style = waiver(),
  conf_int_default = c("step", "ribbon", "none"),
  limit_n_features = waiver(),
  facet_wrap_cols = waiver(),
  x_label = waiver(),
  y_label = waiver(),
  x_label_shared = waiver(),
  y_label_shared = waiver(),
  rotate_x_tick_labels = waiver(),
  rotate_y_tick_labels = waiver(),
  legend_label = waiver(),
  combine_legend = waiver(),
  plot_title = waiver(),
  plot_sub_title = waiver(),
  caption = waiver()
)
}
\arguments{
\item{x_range}{(\emph{optional}) Value range for the x-axis.}

\item{y_range}{(\emph{optional}) Value range for the y-axis.}

\item{x_n_breaks}{(\emph{optional}) Number of breaks to show on the x-axis of the
plot. \code{x_n_breaks} is used to determine the \code{x_breaks} argument in case it
is unset.}

\item{y_n_breaks}{(\emph{optional}) Number of breaks to show on the y-axis of the
plot. \code{y_n_breaks} is used to determine the \code{y_breaks} argument in case it
is unset.}

\item{x_breaks}{(\emph{optional}) Break points on the x-axis of the plot.}

\item{y_breaks}{(\emph{optional}) Break points on the y-axis of the plot.}

\item{conf_int}{(\emph{optional})}

\item{conf_int_alpha}{(\emph{optional}) Alpha value to determine transparency of
confidence intervals or, alternatively, other plot elements with which the
confidence interval overlaps. Only values between 0.0 (fully transparent)
and 1.0 (fully opaque) are allowed.}

\item{conf_int_style}{(\emph{optional}) Confidence interval style. See details for
allowed styles.}

\item{conf_int_default}{Sets the default options for the confidence interval.}

\item{limit_n_features}{(\emph{optional}) The number of features that should be
included in the plot. Only the most important features are shown. By
default, the number of features is not limited.}

\item{facet_wrap_cols}{(\emph{optional}) Number of columns to generate when facet
wrapping. If NULL, a facet grid is produced instead.}

\item{x_label}{(\emph{optional}) Label to provide to the x-axis. If NULL, no label
is shown.}

\item{y_label}{(\emph{optional}) Label to provide to the y-axis. If NULL, no label
is shown.}

\item{x_label_shared}{(\emph{optional}) Sharing of x-axis labels between facets.
One of three values:
\itemize{
\item \code{overall}: A single label is placed at the bottom of the figure. Tick
text (but not the ticks themselves) is removed for all but the bottom facet
plot(s).
\item \code{column}: A label is placed at the bottom of each column. Tick text (but
not the ticks themselves) is removed for all but the bottom facet plot(s).
\item \code{individual}: A label is placed below each facet plot. Tick text is kept.
}}

\item{y_label_shared}{(\emph{optional}) Sharing of y-axis labels between facets.
One of three values:
\itemize{
\item \code{overall}: A single label is placed to the left of the figure. Tick text
(but not the ticks themselves) is removed for all but the left-most facet
plot(s).
\item \code{row}: A label is placed to the left of each row. Tick text (but not the
ticks themselves) is removed for all but the left-most facet plot(s).
\item \code{individual}: A label is placed below each facet plot. Tick text is kept.
}}

\item{rotate_x_tick_labels}{(\emph{optional}) Rotate tick labels on the x-axis by
90 degrees. Defaults to \code{TRUE}. Rotation of x-axis tick labels may also be
controlled through the \code{ggtheme}. In this case, \code{FALSE} should be provided
explicitly.}

\item{rotate_y_tick_labels}{(\emph{optional}) Rotate tick labels on the y-axis by
45 degrees.}

\item{legend_label}{(\emph{optional}) Label to provide to the legend. If NULL, the
legend will not have a name.}

\item{combine_legend}{(\emph{optional}) Flag to indicate whether the same legend
is to be shared by multiple aesthetics, such as those specified by
\code{color_by} and \code{linetype_by} arguments.}

\item{plot_title}{(\emph{optional}) Label to provide as figure title. If NULL, no
title is shown.}

\item{plot_sub_title}{(\emph{optional}) Label to provide as figure subtitle. If
NULL, no subtitle is shown.}

\item{caption}{(\emph{optional}) Label to provide as figure caption. If NULL, no
caption is shown.}
}
\description{
Internal checks on common plot input arguments
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FamiliarS4Classes.R
\docType{class}
\name{familiarModel-class}
\alias{familiarModel-class}
\title{Familiar model.}
\description{
A familiarModel object is a self-contained model that can be applied to
generate predictions for a dataset. familiarModel objects form the parent
class of learner-specific child classes.
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the familiarModel object.}

\item{\code{model}}{The actual model trained using a specific algorithm, e.g. a
random forest from the \code{ranger} package, or a LASSO model from \code{glmnet}.}

\item{\code{outcome_type}}{Outcome type of the data used to create the object.}

\item{\code{outcome_info}}{Outcome information object, which contains additional
information concerning the outcome, such as class levels.}

\item{\code{feature_info}}{List of objects containing feature information, e.g.,
name, class levels, transformation, normalisation and clustering
parameters.}

\item{\code{data_column_info}}{Data information object containing information
regarding identifier column names and outcome column names.}

\item{\code{hyperparameters}}{Set of hyperparameters used to train the model.}

\item{\code{hyperparameter_data}}{Information generated during hyperparameter
optimisation.}

\item{\code{calibration_model}}{One or more models used to recalibrate the model
output. Currently only used by some models.}

\item{\code{novelty_detector}}{A familiarNoveltyDetector object that can be used to
detect out-of-distribution samples.}

\item{\code{learner}}{Learning algorithm used to create the model.}

\item{\code{vimp_method}}{Method used to determine variable importance for the model.}

\item{\code{vimp_table}}{Variable importance table or list of variable importance
tables for the model.}

\item{\code{vimp_aggregation_method}}{Method used for aggregating variable importance
tables if more than one is present.)}

\item{\code{vimp_rank_threshold}}{Threshold used for some variable importance
aggregation methods.}

\item{\code{required_features}}{The set of features required for complete
reproduction, i.e. with imputation.}

\item{\code{model_features}}{The set of features that is used to train the model,}

\item{\code{novelty_features}}{The set of features that is used to train all novelty
detectors in the ensemble.}

\item{\code{calibration_info}}{Calibration information, e.g. baseline survival in the
development cohort.}

\item{\code{km_info}}{Data concerning stratification into risk groups.}

\item{\code{data_id}}{Internal identifier for the dataset used to train the model.}

\item{\code{run_id}}{Internal identifier for the specific subset of the dataset used
used to train the model.}

\item{\code{run_table}}{Run table for the data used to train the model. Used
internally.}

\item{\code{settings}}{A copy of the evaluation configuration parameters used at
model creation. These are used as default parameters when evaluating the
model (technically, familiarEnsemble) to create a familiarData object.}

\item{\code{is_trimmed}}{Flag that indicates whether the model, stored in the \code{model}
slot, has been trimmed.}

\item{\code{trimmed_function}}{List of functions whose output has been captured prior
to trimming the model.}

\item{\code{messages}}{List of warning and error messages generated during training.}

\item{\code{project_id}}{Identifier of the project that generated the familiarModel
object.}

\item{\code{familiar_version}}{Version of the familiar package.}

\item{\code{package}}{Name of package(s) required to executed the model itself, e.g.
\code{ranger} or \code{glmnet}.}

\item{\code{package_version}}{Version of the packages mentioned in the \code{package}
attribute.}
}}
